;;; phps-mode-parser.el --- Exported Emacs Parser Generator -*- lexical-binding: t -*-

;; Copyright (C) 2018-2025  Free Software Foundation, Inc.

;; This file is not part of GNU Emacs.


;;; Commentary:


;;; Code:


(require 'phps-mode-lexer)

(define-error
 'phps-parser-error
 "PHPs Parser Error")


;;; Variables:


(defvar
  phps-mode-parser--action-tables
  #s(hash-table size 42072 test equal rehash-size 1.5 rehash-threshold 0.8125 data (0 0 1 1 2 2 3 3 4 4 5 2 6 2 7 2 8 8 9 2 10 10 11 11 12 2 13 13 14 2 15 15 16 16 17 17 18 2 19 16 20 20 21 15 22 22 23 11 24 24 25 25 26 26 27 2 28 2 29 11 30 16 31 11 32 32 33 33 34 34 35 35 36 11 37 11 38 38 39 39 40 40 41 16 42 11 43 11 44 22 45 2 46 2 47 47 48 16 49 2 50 11 51 51 52 11 53 53 54 11 55 55 56 56 57 57 58 58 59 59 60 60 61 61 62 2 63 2 64 64 65 65 66 2 67 2 68 15 69 69 70 70 71 71 72 2 73 11 74 2 75 16 76 76 77 77 78 11 79 2 80 80 81 81 82 11 83 83 84 2 85 85 86 86 87 87 88 88 89 89 90 90 91 91 92 92 93 11 94 94 95 95 96 96 97 97 98 98 99 99 100 100 101 101 102 102 103 103 104 104 105 105 106 106 107 107 108 108 109 109 110 110 111 111 112 112 113 113 114 114 115 115 116 116 117 117 118 118 119 119 120 120 121 121 122 122 123 123 124 124 125 99 126 126 127 2 128 13 129 129 130 11 131 131 132 132 133 133 134 134 135 135 136 104 137 137 138 138 139 139 140 140 141 11 142 142 143 143 144 13 145 145 146 146 147 147 148 148 149 149 150 150 151 151 152 152 153 153 154 154 155 155 156 156 157 157 158 158 159 159 160 160 161 161 162 162 163 163 164 11 165 165 166 166 167 167 168 77 169 126 170 170 171 171 172 15 173 16 174 15 175 11 176 26 177 2 178 16 179 11 180 11 181 40 182 16 183 11 184 11 185 185 186 16 187 15 188 70 189 71 190 11 191 16 192 77 193 11 194 11 195 195 196 88 197 197 198 92 199 199 200 98 201 201 202 103 203 106 204 204 205 205 206 206 207 207 208 208 209 209 210 210 211 126 212 212 213 213 214 170 215 215 216 26 217 11 218 2 219 2 220 3 221 4 222 2 223 2 224 2 225 2 226 11 227 2 228 2 229 229 230 2 231 231 232 22 233 233 234 234 235 2 236 11 237 11 238 238 239 239 240 240 241 22 242 2 243 2 244 2 245 11 246 246 247 11 248 248 249 11 250 250 251 251 252 252 253 253 254 60 255 255 256 2 257 2 258 258 259 11 260 2 261 2 262 69 263 131 264 264 265 2 266 2 267 267 268 2 269 269 270 270 271 2 272 85 273 86 274 89 275 134 276 11 277 277 278 278 279 99 280 280 281 281 282 282 283 104 284 104 285 285 286 286 287 287 288 288 289 289 290 290 291 291 292 292 293 293 294 294 295 295 296 99 297 2 298 298 299 2 300 2 301 2 302 2 303 2 304 2 305 2 306 2 307 307 308 2 309 2 310 2 311 2 312 2 313 313 314 2 315 2 316 2 317 2 318 2 319 2 320 2 321 2 322 2 323 2 324 2 325 2 326 2 327 2 328 2 329 329 330 330 331 331 332 332 333 333 334 334 335 335 336 336 337 337 338 338 339 339 340 340 341 341 342 342 343 343 344 4 345 2 346 346 347 347 348 348 349 349 350 350 351 351 352 352 353 353 354 354 355 355 356 356 357 357 358 357 359 40 360 360 361 2 362 3 363 4 364 2 365 2 366 2 367 2 368 11 369 2 370 2 371 371 372 2 373 373 374 22 375 375 376 376 377 2 378 11 379 11 380 380 381 381 382 382 383 22 384 2 385 2 386 2 387 11 388 388 389 11 390 390 391 11 392 392 393 393 394 394 395 395 396 60 397 397 398 2 399 2 400 400 401 11 402 2 403 2 404 69 405 131 406 406 407 2 408 2 409 409 410 2 411 411 412 412 413 2 414 85 415 86 416 89 417 134 418 11 419 419 420 420 421 99 422 422 423 423 424 424 425 104 426 104 427 427 428 428 429 429 430 430 431 431 432 432 433 433 434 434 435 435 436 436 437 437 438 438 439 99 440 2 441 441 442 2 443 2 444 2 445 2 446 2 447 2 448 2 449 2 450 307 451 2 452 2 453 2 454 2 455 2 456 313 457 2 458 2 459 2 460 2 461 2 462 2 463 2 464 2 465 2 466 2 467 2 468 2 469 2 470 2 471 2 472 472 473 473 474 474 475 475 476 476 477 477 478 478 479 479 480 480 481 481 482 482 483 483 484 484 485 485 486 486 487 4 488 2 489 489 490 490 491 491 492 492 493 493 494 494 495 495 496 496 497 497 498 498 499 499 500 357 501 357 502 40 503 360 504 434 505 505 506 506 507 507 508 508 509 509 510 2 511 2 512 3 513 4 514 2 515 2 516 2 517 2 518 11 519 2 520 2 521 521 522 2 523 523 524 22 525 525 526 526 527 2 528 11 529 11 530 530 531 531 532 532 533 22 534 2 535 2 536 2 537 11 538 538 539 11 540 540 541 11 542 542 543 543 544 544 545 545 546 60 547 547 548 2 549 2 550 550 551 11 552 2 553 2 554 69 555 131 556 556 557 2 558 2 559 559 560 2 561 561 562 562 563 2 564 85 565 86 566 89 567 134 568 11 569 569 570 570 571 99 572 572 573 573 574 574 575 104 576 104 577 577 578 578 579 579 580 580 581 581 582 582 583 583 584 584 585 585 586 586 587 587 588 99 589 2 590 590 591 2 592 2 593 2 594 2 595 2 596 2 597 2 598 2 599 307 600 2 601 2 602 2 603 2 604 2 605 313 606 2 607 2 608 2 609 2 610 2 611 2 612 2 613 2 614 2 615 2 616 2 617 2 618 2 619 2 620 2 621 621 622 622 623 623 624 624 625 625 626 626 627 627 628 628 629 629 630 630 631 631 632 632 633 633 634 634 635 635 636 4 637 2 638 638 639 639 640 640 641 641 642 642 643 643 644 644 645 645 646 646 647 647 648 648 649 357 650 357 651 40 652 360 653 434 654 654 655 655 656 656 657 657 658 658 659 2 660 574 661 661 662 662 663 40 664 664 665 282 666 666 667 667 668 2 669 574 670 670 671 671 672 672 673 673 674 674 675 675 676 2 677 3 678 4 679 2 680 2 681 2 682 2 683 2 684 11 685 2 686 2 687 687 688 2 689 689 690 22 691 691 692 692 693 2 694 11 695 11 696 696 697 697 698 698 699 22 700 2 701 2 702 2 703 11 704 704 705 11 706 706 707 11 708 708 709 709 710 710 711 711 712 60 713 713 714 2 715 2 716 716 717 11 718 2 719 2 720 69 721 131 722 722 723 2 724 2 725 725 726 2 727 727 728 728 729 2 730 85 731 86 732 89 733 134 734 11 735 735 736 736 737 99 738 738 739 739 740 740 741 104 742 104 743 743 744 744 745 745 746 746 747 747 748 748 749 749 750 750 751 751 752 752 753 753 754 99 755 2 756 756 757 2 758 2 759 2 760 2 761 2 762 2 763 2 764 2 765 307 766 2 767 2 768 2 769 2 770 2 771 313 772 2 773 2 774 2 775 2 776 2 777 2 778 2 779 2 780 2 781 2 782 2 783 2 784 2 785 2 786 2 787 787 788 788 789 789 790 790 791 791 792 792 793 793 794 794 795 795 796 796 797 797 798 798 799 799 800 800 801 801 802 4 803 2 804 804 805 805 806 806 807 807 808 808 809 809 810 810 811 811 812 812 813 813 814 814 815 357 816 357 817 40 818 360 819 434 820 820 821 821 822 822 823 823 824 824 825 2 826 574 827 827 828 828 829 40 830 830 831 282 832 832 833 833 834 2 835 574 836 836 837 837 838 838 839 839 840 840 841 841 842 2 843 740 844 2 845 845 846 846 847 847 848 848 849 849 850 850 851 851 852 852 853 853 854 854 855 855 856 856 857 857 858 858 859 859 860 860 861 861 862 862 863 863 864 864 865 865 866 866 867 867 868 868 869 869 870 870 871 871 872 872 873 873 874 874 875 875 876 876 877 877 878 878 879 879 880 880 881 881 882 882 883 883 884 884 885 885 886 886 887 887 888 888 889 889 890 890 891 891 892 892 893 893 894 894 895 895 896 896 897 897 898 898 899 899 900 900 901 901 902 902 903 903 904 904 905 905 906 906 907 907 908 908 909 909 910 910 911 911 912 912 913 913 914 914 915 915 916 916 917 917 918 918 919 919 920 920 921 921 922 922 923 923 924 924 925 925 926 926 927 927 928 928 929 929 930 930 931 931 932 932 933 933 934 934 935 935 936 11 937 937 938 938 939 939 940 2 941 574 942 942 943 943 944 944 945 2 946 3 947 4 948 2 949 949 950 2 951 2 952 2 953 953 954 954 955 2 956 956 957 2 958 958 959 959 960 960 961 961 962 962 963 963 964 964 965 965 966 966 967 967 968 22 969 969 970 970 971 971 972 972 973 973 974 2 975 975 976 976 977 977 978 978 979 979 980 980 981 981 982 982 983 983 984 984 985 985 986 986 987 987 988 988 989 989 990 990 991 991 992 992 993 993 994 994 995 995 996 996 997 997 998 998 999 999 1000 1000 1001 1001 1002 22 1003 1003 1004 1004 1005 1005 1006 1006 1007 1007 1008 2 1009 1009 1010 1010 1011 1011 1012 1012 1013 1013 1014 1014 1015 1015 1016 1016 1017 1017 1018 1018 1019 1019 1020 1020 1021 1021 1022 1022 1023 1023 1024 2 1025 1025 1026 1026 1027 1027 1028 1028 1029 1029 1030 1030 1031 1031 1032 1032 1033 1033 1034 69 1035 1035 1036 1036 1037 2 1038 1038 1039 1039 1040 1040 1041 1041 1042 1042 1043 1043 1044 2 1045 1045 1046 1046 1047 1047 1048 1048 1049 1049 1050 2 1051 85 1052 86 1053 1053 1054 89 1055 134 1056 11 1057 1057 1058 1058 1059 99 1060 1060 1061 1061 1062 1062 1063 104 1064 104 1065 1065 1066 1066 1067 1067 1068 1068 1069 1069 1070 1070 1071 1071 1072 1072 1073 159 1074 1074 1075 1075 1076 1076 1077 1077 1078 1078 1079 1079 1080 99 1081 2 1082 11 1083 1083 1084 2 1085 1085 1086 11 1087 11 1088 1088 1089 1089 1090 1090 1091 2 1092 2 1093 11 1094 1094 1095 11 1096 11 1097 1097 1098 60 1099 1099 1100 2 1101 1101 1102 11 1103 2 1104 2 1105 131 1106 1106 1107 2 1108 1108 1109 1109 1110 1110 1111 2 1112 2 1113 2 1114 2 1115 2 1116 2 1117 2 1118 2 1119 307 1120 2 1121 2 1122 2 1123 2 1124 2 1125 313 1126 2 1127 2 1128 2 1129 2 1130 2 1131 2 1132 2 1133 2 1134 2 1135 2 1136 2 1137 2 1138 2 1139 2 1140 2 1141 1141 1142 1142 1143 1143 1144 1144 1145 1145 1146 1146 1147 1147 1148 1148 1149 1149 1150 1150 1151 1151 1152 1152 1153 1153 1154 1154 1155 1155 1156 4 1157 2 1158 1158 1159 1159 1160 1160 1161 1161 1162 1162 1163 1163 1164 1164 1165 1165 1166 1166 1167 1167 1168 1168 1169 357 1170 357 1171 40 1172 360 1173 434 1174 1174 1175 1175 1176 1176 1177 1177 1178 1178 1179 2 1180 574 1181 1181 1182 1182 1183 40 1184 1184 1185 282 1186 1186 1187 1187 1188 2 1189 574 1190 1190 1191 1191 1192 1192 1193 1193 1194 1194 1195 1195 1196 2 1197 740 1198 2 1199 1199 1200 1200 1201 1201 1202 1202 1203 1203 1204 1204 1205 1205 1206 1206 1207 1207 1208 1208 1209 2 1210 3 1211 4 1212 2 1213 2 1214 2 1215 2 1216 11 1217 2 1218 2 1219 1219 1220 2 1221 1221 1222 22 1223 1223 1224 1224 1225 2 1226 11 1227 11 1228 1228 1229 1229 1230 1230 1231 22 1232 2 1233 2 1234 2 1235 11 1236 1236 1237 11 1238 1238 1239 11 1240 1240 1241 1241 1242 1242 1243 1243 1244 60 1245 1245 1246 2 1247 2 1248 1248 1249 11 1250 2 1251 2 1252 69 1253 131 1254 1254 1255 2 1256 2 1257 1257 1258 2 1259 1259 1260 1260 1261 2 1262 85 1263 86 1264 89 1265 134 1266 11 1267 1267 1268 1268 1269 99 1270 1270 1271 1271 1272 1272 1273 104 1274 104 1275 1275 1276 1276 1277 1277 1278 1278 1279 1279 1280 1280 1281 1281 1282 1282 1283 1283 1284 1284 1285 1285 1286 99 1287 2 1288 1288 1289 2 1290 2 1291 2 1292 2 1293 2 1294 2 1295 2 1296 2 1297 307 1298 2 1299 2 1300 2 1301 2 1302 2 1303 313 1304 2 1305 2 1306 2 1307 2 1308 2 1309 2 1310 2 1311 2 1312 2 1313 2 1314 2 1315 2 1316 2 1317 2 1318 2 1319 1319 1320 1320 1321 1321 1322 1322 1323 1323 1324 1324 1325 1325 1326 1326 1327 1327 1328 1328 1329 1329 1330 1330 1331 1331 1332 1332 1333 1333 1334 4 1335 2 1336 1336 1337 1337 1338 1338 1339 1339 1340 1340 1341 1341 1342 1342 1343 1343 1344 1344 1345 1345 1346 1346 1347 357 1348 357 1349 40 1350 360 1351 434 1352 1352 1353 1353 1354 1354 1355 1355 1356 1356 1357 2 1358 574 1359 1359 1360 1360 1361 40 1362 1362 1363 282 1364 1364 1365 1365 1366 2 1367 574 1368 1368 1369 1369 1370 1370 1371 1371 1372 1372 1373 1373 1374 2 1375 740 1376 2 1377 1377 1378 1378 1379 1379 1380 1380 1381 1381 1382 1382 1383 1383 1384 1384 1385 1385 1386 1386 1387 855 1388 1388 1389 1389 1390 1390 1391 1391 1392 1392 1393 1393 1394 1394 1395 1395 1396 1396 1397 1397 1398 1398 1399 1399 1400 1400 1401 1401 1402 1402 1403 1403 1404 1404 1405 1405 1406 1406 1407 1407 1408 1408 1409 1409 1410 1410 1411 1411 1412 1412 1413 1413 1414 1414 1415 1415 1416 1416 1417 1417 1418 1418 1419 1419 1420 1420 1421 1421 1422 1422 1423 1423 1424 1424 1425 1425 1426 1426 1427 1427 1428 1428 1429 1429 1430 1430 1431 1431 1432 1432 1433 1433 1434 1434 1435 1435 1436 1436 1437 1437 1438 1438 1439 1439 1440 1440 1441 1441 1442 1442 1443 1443 1444 1444 1445 1445 1446 1446 1447 1447 1448 1448 1449 1449 1450 1450 1451 1451 1452 1452 1453 1453 1454 1454 1455 1455 1456 1456 1457 1457 1458 1458 1459 1459 1460 1460 1461 1461 1462 1462 1463 1463 1464 1464 1465 1465 1466 1466 1467 1467 1468 11 1469 1469 1470 1470 1471 1471 1472 2 1473 574 1474 1474 1475 943 1476 1476 1477 1477 1478 976 1479 159 1480 1480 1481 162 1482 1482 1483 2 1484 1484 1485 1485 1486 1486 1487 1487 1488 2 1489 2 1490 2 1491 1491 1492 2 1493 1493 1494 2 1495 2 1496 2 1497 2 1498 2 1499 2 1500 2 1501 2 1502 2 1503 1503 1504 1504 1505 1505 1506 1506 1507 1507 1508 1508 1509 1509 1510 1510 1511 1511 1512 1512 1513 1513 1514 1514 1515 1515 1516 1516 1517 1517 1518 22 1519 1519 1520 3 1521 4 1522 2 1523 11 1524 1524 1525 1525 1526 1526 1527 1527 1528 1528 1529 1529 1530 1530 1531 1531 1532 1532 1533 1533 1534 60 1535 1535 1536 1536 1537 11 1538 1538 1539 1539 1540 1540 1541 1541 1542 85 1543 89 1544 11 1545 1545 1546 1546 1547 99 1548 1548 1549 1549 1550 1550 1551 1551 1552 1552 1553 1553 1554 1554 1555 1555 1556 99 1557 855 1558 1558 1559 1559 1560 1560 1561 1561 1562 1562 1563 1563 1564 1564 1565 1565 1566 1566 1567 1567 1568 1568 1569 1569 1570 1570 1571 1571 1572 1572 1573 1573 1574 1574 1575 1575 1576 1576 1577 1577 1578 1578 1579 1579 1580 1580 1581 1581 1582 1582 1583 1583 1584 1584 1585 1585 1586 1586 1587 1587 1588 1588 1589 1589 1590 1590 1591 1591 1592 1592 1593 1593 1594 1594 1595 1595 1596 1596 1597 1597 1598 1598 1599 1599 1600 1600 1601 1601 1602 1602 1603 1603 1604 1604 1605 1605 1606 1606 1607 1607 1608 1608 1609 1609 1610 1610 1611 1611 1612 1612 1613 1613 1614 1614 1615 1615 1616 1616 1617 1617 1618 1618 1619 1619 1620 1620 1621 1621 1622 1622 1623 1623 1624 1624 1625 1625 1626 1626 1627 1627 1628 1628 1629 1629 1630 1630 1631 1631 1632 1632 1633 1633 1634 1634 1635 1635 1636 1636 1637 1637 1638 11 1639 1639 1640 1640 1641 1641 1642 2 1643 574 1644 1644 1645 943 1646 1646 1647 1647 1648 976 1649 159 1650 162 1651 1651 1652 1652 1653 1653 1654 855 1655 1655 1656 11 1657 1657 1658 2 1659 574 1660 1660 1661 1661 1662 1662 1663 357 1664 357 1665 360 1666 434 1667 1667 1668 1668 1669 1669 1670 1670 1671 2 1672 574 1673 1673 1674 1674 1675 1675 1676 1676 1677 2 1678 3 1679 4 1680 2 1681 2 1682 2 1683 2 1684 11 1685 2 1686 2 1687 1687 1688 2 1689 1689 1690 22 1691 1691 1692 1692 1693 2 1694 2 1695 11 1696 11 1697 1697 1698 1698 1699 1699 1700 22 1701 2 1702 2 1703 2 1704 11 1705 1705 1706 11 1707 1707 1708 11 1709 1709 1710 1710 1711 1711 1712 1712 1713 60 1714 1714 1715 2 1716 2 1717 1717 1718 11 1719 2 1720 2 1721 69 1722 131 1723 1723 1724 2 1725 2 1726 1726 1727 2 1728 1728 1729 1729 1730 2 1731 85 1732 86 1733 22 1734 89 1735 1735 1736 434 1737 134 1738 11 1739 1739 1740 1740 1741 99 1742 1742 1743 1743 1744 1744 1745 104 1746 104 1747 1747 1748 1748 1749 1749 1750 1750 1751 1751 1752 1752 1753 1753 1754 1754 1755 1755 1756 1756 1757 1757 1758 1758 1759 1759 1760 99 1761 2 1762 11 1763 1763 1764 2 1765 2 1766 2 1767 2 1768 2 1769 2 1770 2 1771 2 1772 307 1773 2 1774 2 1775 2 1776 2 1777 2 1778 313 1779 2 1780 2 1781 2 1782 2 1783 2 1784 2 1785 2 1786 2 1787 2 1788 2 1789 2 1790 2 1791 2 1792 2 1793 2 1794 1794 1795 1795 1796 1796 1797 1797 1798 1798 1799 1799 1800 1800 1801 1801 1802 1802 1803 1803 1804 1804 1805 1805 1806 1806 1807 1807 1808 1808 1809 4 1810 2 1811 1811 1812 1812 1813 1813 1814 1814 1815 1815 1816 1816 1817 1817 1818 1818 1819 1819 1820 1820 1821 1821 1822 357 1823 357 1824 40 1825 360 1826 434 1827 1827 1828 1828 1829 1829 1830 1830 1831 1831 1832 2 1833 574 1834 1834 1835 1835 1836 40 1837 1837 1838 282 1839 1839 1840 1840 1841 2 1842 574 1843 1843 1844 1844 1845 1845 1846 1846 1847 1847 1848 1848 1849 2 1850 740 1851 2 1852 1852 1853 1853 1854 1854 1855 1855 1856 1856 1857 1857 1858 1858 1859 1859 1860 1860 1861 1861 1862 1862 1863 2 1864 3 1865 4 1866 2 1867 2 1868 2 1869 2 1870 11 1871 2 1872 2 1873 1219 1874 2 1875 1221 1876 22 1877 1223 1878 1224 1879 2 1880 2 1881 11 1882 11 1883 1228 1884 1229 1885 1230 1886 22 1887 2 1888 2 1889 2 1890 11 1891 1236 1892 11 1893 1238 1894 11 1895 1240 1896 1241 1897 1242 1898 1243 1899 60 1900 1245 1901 2 1902 2 1903 1248 1904 11 1905 2 1906 2 1907 69 1908 131 1909 1254 1910 2 1911 2 1912 1257 1913 2 1914 1259 1915 1260 1916 2 1917 85 1918 86 1919 22 1920 89 1921 1921 1922 162 1923 134 1924 11 1925 1267 1926 1268 1927 99 1928 1270 1929 1271 1930 1930 1931 104 1932 104 1933 1275 1934 1276 1935 1277 1936 1278 1937 1279 1938 1280 1939 1281 1940 1940 1941 1941 1942 1282 1943 1283 1944 1284 1945 1285 1946 99 1947 2 1948 11 1949 1288 1950 2 1951 2 1952 2 1953 2 1954 2 1955 2 1956 2 1957 2 1958 307 1959 2 1960 2 1961 2 1962 2 1963 2 1964 313 1965 2 1966 2 1967 2 1968 2 1969 2 1970 2 1971 2 1972 2 1973 2 1974 2 1975 2 1976 2 1977 2 1978 2 1979 2 1980 1319 1981 1320 1982 1321 1983 1322 1984 1323 1985 1324 1986 1325 1987 1326 1988 1327 1989 1328 1990 1329 1991 1330 1992 1331 1993 1332 1994 1333 1995 4 1996 2 1997 1336 1998 1337 1999 1338 2000 1339 2001 1340 2002 1341 2003 1342 2004 1343 2005 1344 2006 1345 2007 1346 2008 357 2009 357 2010 40 2011 360 2012 434 2013 1352 2014 1353 2015 1354 2016 1355 2017 1356 2018 2 2019 574 2020 1359 2021 1360 2022 40 2023 1362 2024 282 2025 1364 2026 1365 2027 2 2028 574 2029 1368 2030 1369 2031 1370 2032 1371 2033 1372 2034 1373 2035 2 2036 740 2037 2 2038 1377 2039 1378 2040 1379 2041 1380 2042 1381 2043 1382 2044 1383 2045 1384 2046 1385 2047 1386 2048 1862 2049 162 2050 2050 2051 2 2052 2052 2053 855 2054 1388 2055 1389 2056 1390 2057 1391 2058 1392 2059 1393 2060 1394 2061 1395 2062 1396 2063 1397 2064 1398 2065 1399 2066 1400 2067 1401 2068 1402 2069 1403 2070 1404 2071 1405 2072 1406 2073 1407 2074 1408 2075 1409 2076 1410 2077 1411 2078 1412 2079 1413 2080 1414 2081 1415 2082 1416 2083 1417 2084 1418 2085 1419 2086 1420 2087 1421 2088 1422 2089 1423 2090 1424 2091 1425 2092 1426 2093 1427 2094 1428 2095 1429 2096 1430 2097 1431 2098 1432 2099 1433 2100 1434 2101 1435 2102 1436 2103 1437 2104 1438 2105 1439 2106 1440 2107 1441 2108 1442 2109 1443 2110 1444 2111 1445 2112 1446 2113 1447 2114 1448 2115 1449 2116 1450 2117 1451 2118 1452 2119 1453 2120 1454 2121 1455 2122 1456 2123 1457 2124 1458 2125 1459 2126 1460 2127 1461 2128 1462 2129 1463 2130 1464 2131 1465 2132 1466 2133 1467 2134 11 2135 1469 2136 1470 2137 1471 2138 2 2139 574 2140 1474 2141 943 2142 1476 2143 1477 2144 976 2145 159 2146 162 2147 1482 2148 1486 2149 1487 2150 2 2151 2 2152 2 2153 1491 2154 2 2155 1493 2156 2 2157 2 2158 2 2159 2 2160 2 2161 2 2162 2 2163 2 2164 2 2165 1503 2166 1504 2167 1505 2168 1506 2169 1507 2170 1508 2171 1509 2172 1510 2173 1511 2174 1512 2175 1513 2176 1514 2177 1515 2178 22 2179 1519 2180 4 2181 11 2182 1541 2183 89 2184 11 2185 1545 2186 99 2187 1550 2188 1551 2189 1553 2190 1554 2191 1555 2192 99 2193 855 2194 11 2195 1641 2196 943 2197 1646 2198 1647 2199 976 2200 159 2201 162 2202 1651 2203 1652 2204 1653 2205 855 2206 11 2207 1657 2208 1661 2209 1662 2210 357 2211 357 2212 360 2213 434 2214 1667 2215 1668 2216 1669 2217 1670 2218 2 2219 574 2220 1673 2221 1674 2222 1675 2223 1676 2224 2224 2225 2225 2226 2 2227 574 2228 2228 2229 1862 2230 2230 2231 2231 2232 140 2233 11 2234 142 2235 162 2236 163 2237 165 2238 167 2239 77 2240 126 2241 170 2242 2242 2243 2243 2244 140 2245 11 2246 142 2247 162 2248 2248 2249 166 2250 2250 2251 2251 2252 2252 2253 2 2254 2254 2255 2255 2256 2256 2257 2256 2258 2258 2259 2259 2260 2260 2261 2261 2262 2262 2263 2263 2264 2264 2265 2265 2266 2266 2267 2267 2268 2268 2269 2269 2270 2270 2271 2271 2272 2272 2273 2256 2274 2274 2275 2275 2276 2276 2277 2277 2278 2278 2279 2279 2280 2280 2281 2281 2282 2282 2283 2283 2284 2284 2285 2285 2286 2286 2287 2287 2288 2288 2289 2289 2290 2290 2291 2291 2292 2292 2293 2293 2294 2294 2295 2295 2296 2256 2297 2297 2298 2298 2299 2299 2300 2300 2301 2301 2302 2302 2303 2303 2304 2304 2305 2305 2306 2306 2307 2307 2308 2256 2309 2309 2310 2272 2311 2311 2312 2256 2313 2313 2314 2314 2315 2315 2316 2316 2317 2317 2318 2318 2319 2319 2320 2320 2321 2321 2322 2322 2323 2256 2324 2324 2325 2325 2326 2326 2327 2327 2328 2328 2329 2329 2330 2330 2331 2331 2332 2332 2333 2333 2334 2334 2335 2292 2336 2336 2337 1487 2338 2 2339 3 2340 4 2341 2 2342 2 2343 2 2344 2 2345 11 2346 2 2347 2 2348 1083 2349 2 2350 966 2351 22 2352 1085 2353 972 2354 2 2355 11 2356 11 2357 1088 2358 1089 2359 1090 2360 22 2361 2 2362 2 2363 2 2364 11 2365 1094 2366 11 2367 1012 2368 11 2369 1097 2370 1019 2371 1020 2372 1021 2373 60 2374 1099 2375 2 2376 2 2377 1101 2378 11 2379 2 2380 2 2381 69 2382 131 2383 1106 2384 2 2385 2 2386 1108 2387 2 2388 1047 2389 1109 2390 2 2391 85 2392 86 2393 22 2394 89 2395 134 2396 11 2397 1057 2398 1058 2399 99 2400 1060 2401 1061 2402 2402 2403 104 2404 104 2405 1065 2406 1067 2407 1068 2408 1069 2409 1070 2410 1071 2411 1072 2412 1075 2413 1077 2414 1078 2415 1079 2416 99 2417 2 2418 11 2419 1110 2420 2 2421 2 2422 2 2423 2 2424 2 2425 2 2426 2 2427 2 2428 307 2429 2 2430 2 2431 2 2432 2 2433 2 2434 313 2435 2 2436 2 2437 2 2438 2 2439 2 2440 2 2441 2 2442 2 2443 2 2444 2 2445 2 2446 2 2447 2 2448 2 2449 2 2450 1141 2451 1142 2452 1143 2453 1144 2454 1145 2455 1146 2456 1147 2457 1148 2458 1149 2459 1150 2460 1151 2461 1152 2462 1153 2463 1154 2464 1155 2465 4 2466 2 2467 1158 2468 1159 2469 1160 2470 1161 2471 1162 2472 1163 2473 1164 2474 1165 2475 1166 2476 1167 2477 1168 2478 357 2479 357 2480 40 2481 360 2482 434 2483 1174 2484 1175 2485 1176 2486 1177 2487 1178 2488 2 2489 574 2490 1181 2491 1182 2492 40 2493 1184 2494 282 2495 1186 2496 1187 2497 2 2498 574 2499 1190 2500 1191 2501 1192 2502 1193 2503 1194 2504 1195 2505 2 2506 740 2507 2 2508 1199 2509 1200 2510 1201 2511 1202 2512 1203 2513 1204 2514 1205 2515 1206 2516 1207 2517 1208 2518 1862 2519 162 2520 2050 2521 2 2522 2522 2523 855 2524 2524 2525 2525 2526 2526 2527 2527 2528 2528 2529 2529 2530 2530 2531 2531 2532 2532 2533 2533 2534 2534 2535 2535 2536 2536 2537 2537 2538 2538 2539 2539 2540 2540 2541 2541 2542 2542 2543 2543 2544 2544 2545 2545 2546 2546 2547 2547 2548 2548 2549 2549 2550 2550 2551 2551 2552 2552 2553 2553 2554 2554 2555 2555 2556 2556 2557 2557 2558 2558 2559 2559 2560 2560 2561 2561 2562 2562 2563 2563 2564 2564 2565 2565 2566 2566 2567 2567 2568 2568 2569 2569 2570 2570 2571 2571 2572 2572 2573 2573 2574 2574 2575 2575 2576 2576 2577 2577 2578 2578 2579 2579 2580 2580 2581 2581 2582 2582 2583 2583 2584 2584 2585 2585 2586 2586 2587 2587 2588 2588 2589 2589 2590 2590 2591 2591 2592 2592 2593 2593 2594 2594 2595 2595 2596 2596 2597 2597 2598 2598 2599 2599 2600 2600 2601 2601 2602 2602 2603 2603 2604 11 2605 2605 2606 2606 2607 2607 2608 2 2609 574 2610 2610 2611 943 2612 2612 2613 2613 2614 976 2615 159 2616 162 2617 2617 2618 2618 2619 1487 2620 2 2621 2 2622 2 2623 2623 2624 2 2625 2625 2626 2 2627 2 2628 2 2629 2 2630 2 2631 2 2632 2 2633 2 2634 2 2635 2635 2636 2636 2637 2637 2638 2638 2639 2639 2640 2640 2641 2641 2642 2642 2643 2643 2644 2644 2645 2645 2646 2646 2647 2647 2648 22 2649 2649 2650 4 2651 11 2652 2652 2653 89 2654 11 2655 2655 2656 99 2657 2657 2658 1551 2659 2659 2660 2660 2661 2661 2662 99 2663 855 2664 11 2665 2665 2666 943 2667 2667 2668 2668 2669 976 2670 159 2671 162 2672 2672 2673 2673 2674 2674 2675 855 2676 11 2677 2677 2678 2678 2679 2679 2680 357 2681 357 2682 360 2683 434 2684 2684 2685 2685 2686 2686 2687 2687 2688 2 2689 574 2690 2690 2691 2691 2692 2692 2693 2693 2694 2694 2695 2695 2696 2 2697 574 2698 2698 2699 2699 2700 140 2701 11 2702 142 2703 162 2704 163 2705 165 2706 167 2707 77 2708 126 2709 170 2710 2710 2711 2711 2712 140 2713 11 2714 142 2715 162 2716 2248 2717 2250 2718 2251 2719 2252 2720 2 2721 2721 2722 2722 2723 855 2724 2724 2725 11 2726 2726 2727 2 2728 574 2729 2729 2730 2730 2731 2731 2732 2732 2733 2733 2734 2734 2735 2735 2736 357 2737 357 2738 360 2739 434 2740 2740 2741 2741 2742 2742 2743 2743 2744 2 2745 574 2746 2746 2747 2747 2748 2748 2749 2749 2750 4 2751 11 2752 2752 2753 89 2754 11 2755 2755 2756 99 2757 2757 2758 1551 2759 2759 2760 2760 2761 2761 2762 99 2763 855 2764 11 2765 2765 2766 943 2767 2767 2768 2768 2769 976 2770 159 2771 162 2772 2772 2773 2773 2774 2774 2775 855 2776 11 2777 2777 2778 2778 2779 2779 2780 357 2781 357 2782 360 2783 434 2784 2784 2785 2785 2786 2786 2787 2787 2788 2 2789 574 2790 2790 2791 2791 2792 2792 2793 2793 2794 2794 2795 2795 2796 2 2797 574 2798 2798 2799 22 2800 2800 2801 2801 2802 2802 2803 2803 2804 2804 2805 2805 2806 2806 2807 2807 2808 2808 2809 16 2810 16 2811 2811 2812 2812 2813 2813 2814 2814 2815 2815 2816 434 2817 2817 2818 2818 2819 2819 2820 2820 2821 2821 2822 2822 2823 574 2824 2824 2825 2 2826 2826 2827 2827 2828 2828 2829 2829 2830 4 2831 11 2832 2832 2833 89 2834 11 2835 2835 2836 99 2837 2837 2838 1551 2839 2839 2840 2840 2841 2841 2842 99 2843 855 2844 11 2845 2845 2846 943 2847 2847 2848 2848 2849 976 2850 159 2851 162 2852 2852 2853 2853 2854 2854 2855 2855 2856 855 2857 11 2858 2858 2859 2859 2860 2860 2861 357 2862 357 2863 360 2864 434 2865 2865 2866 2866 2867 2867 2868 2868 2869 2 2870 574 2871 2871 2872 2872 2873 2873 2874 2874 2875 2875 2876 2876 2877 2 2878 574 2879 2879 2880 434 2881 2881 2882 2 2883 2883 2884 2884 2885 1272 2886 2 2887 2887 2888 2888 2889 2889 2890 2890 2891 2891 2892 22 2893 2800 2894 2894 2895 2895 2896 2896 2897 69 2898 2898 2899 2899 2900 2900 2901 2901 2902 16 2903 16 2904 2811 2905 434 2906 2906 2907 2907 2908 2908 2909 2909 2910 2910 2911 2822 2912 574 2913 2913 2914 2 2915 2915 2916 2827 2917 2828 2918 2918 2919 2841 2920 2920 2921 2921 2922 2922 2923 2923 2924 2924 2925 2925 2926 2 2927 13 2928 2928 2929 2929 2930 2930 2931 2931 2932 2932 2933 2933 2934 2934 2935 2935 2936 2936 2937 2937 2938 2938 2939 2939 2940 2940 2941 2941 2942 2942 2943 2943 2944 2944 2945 357 2946 357 2947 40 2948 360 2949 434 2950 2950 2951 2951 2952 2952 2953 2953 2954 2954 2955 40 2956 2956 2957 2957 2958 2958 2959 2959 2960 943 2961 2961 2962 2962 2963 13 2964 2964 2965 13 2966 2966 2967 77 2968 2968 2969 2969 2970 2970 2971 13 2972 2972 2973 2973 2974 2974 2975 2975 2976 2976 2977 2977 2978 2978 2979 2979 2980 2980 2981 2981 2982 13 2983 2983 2984 2984 2985 2985 2986 2986 2987 2987 2988 2988 2989 2989 2990 2990 2991 2991 2992 2992 2993 2993 2994 2994 2995 2995 2996 2995 2997 2997 2998 2998 2999 2999 3000 3000 3001 3001 3002 3002 3003 3003 3004 3004 3005 3005 3006 3006 3007 3007 3008 3008 3009 2271 3010 3010 3011 2995 3012 3012 3013 3013 3014 3014 3015 3015 3016 3016 3017 3017 3018 3018 3019 3019 3020 3020 3021 2292 3022 2294 3023 2995 3024 3024 3025 3025 3026 2995 3027 3027 3028 3010 3029 3029 3030 2995 3031 3031 3032 3032 3033 3033 3034 3034 3035 3035 3036 3036 3037 3037 3038 3038 3039 3039 3040 3040 3041 3041 3042 3042 3043 3005 3044 3044 3045 3045 3046 3046 3047 2 3048 3048 3049 2 3050 3 3051 4 3052 2 3053 2 3054 2 3055 2 3056 11 3057 2 3058 2 3059 3059 3060 2 3061 3061 3062 22 3063 3063 3064 3064 3065 2 3066 11 3067 11 3068 3068 3069 3069 3070 3070 3071 22 3072 2 3073 2 3074 2 3075 11 3076 3076 3077 11 3078 3078 3079 11 3080 3080 3081 3081 3082 3082 3083 3083 3084 60 3085 3085 3086 2 3087 2 3088 3088 3089 11 3090 2 3091 2 3092 69 3093 131 3094 3094 3095 2 3096 2 3097 3097 3098 2 3099 3099 3100 3100 3101 2 3102 85 3103 86 3104 89 3105 134 3106 11 3107 3107 3108 3108 3109 99 3110 3110 3111 3111 3112 3112 3113 104 3114 104 3115 3115 3116 3116 3117 3117 3118 3118 3119 3119 3120 3120 3121 3121 3122 3122 3123 3123 3124 3124 3125 3125 3126 99 3127 2 3128 3128 3129 2 3130 2 3131 2 3132 2 3133 2 3134 2 3135 2 3136 2 3137 307 3138 2 3139 2 3140 2 3141 2 3142 2 3143 313 3144 2 3145 2 3146 2 3147 2 3148 2 3149 2 3150 2 3151 2 3152 2 3153 2 3154 2 3155 2 3156 2 3157 2 3158 2 3159 3159 3160 3160 3161 3161 3162 3162 3163 3163 3164 3164 3165 3165 3166 3166 3167 3167 3168 3168 3169 3169 3170 3170 3171 3171 3172 3172 3173 3173 3174 4 3175 2 3176 3176 3177 3177 3178 3178 3179 3179 3180 3180 3181 3181 3182 3182 3183 3183 3184 3184 3185 3185 3186 3186 3187 357 3188 357 3189 40 3190 360 3191 434 3192 3192 3193 3193 3194 3194 3195 3195 3196 3196 3197 2 3198 574 3199 3199 3200 3200 3201 40 3202 3202 3203 282 3204 3204 3205 3205 3206 2 3207 574 3208 3208 3209 3209 3210 3210 3211 3211 3212 3212 3213 3213 3214 2 3215 740 3216 2 3217 3217 3218 3218 3219 3219 3220 3220 3221 3221 3222 3222 3223 3223 3224 3224 3225 3225 3226 3226 3227 855 3228 3228 3229 3229 3230 3230 3231 3231 3232 3232 3233 3233 3234 3234 3235 3235 3236 3236 3237 3237 3238 3238 3239 3239 3240 3240 3241 3241 3242 3242 3243 3243 3244 3244 3245 3245 3246 3246 3247 3247 3248 3248 3249 3249 3250 3250 3251 3251 3252 3252 3253 3253 3254 3254 3255 3255 3256 3256 3257 3257 3258 3258 3259 3259 3260 3260 3261 3261 3262 3262 3263 3263 3264 3264 3265 3265 3266 3266 3267 3267 3268 3268 3269 3269 3270 3270 3271 3271 3272 3272 3273 3273 3274 3274 3275 3275 3276 3276 3277 3277 3278 3278 3279 3279 3280 3280 3281 3281 3282 3282 3283 3283 3284 3284 3285 3285 3286 3286 3287 3287 3288 3288 3289 3289 3290 3290 3291 3291 3292 3292 3293 3293 3294 3294 3295 3295 3296 3296 3297 3297 3298 3298 3299 3299 3300 3300 3301 3301 3302 3302 3303 3303 3304 3304 3305 3305 3306 3306 3307 3307 3308 11 3309 3309 3310 3310 3311 3311 3312 2 3313 574 3314 3314 3315 943 3316 3316 3317 3317 3318 976 3319 159 3320 162 3321 3321 3322 3322 3323 1487 3324 2 3325 2 3326 2 3327 3327 3328 2 3329 3329 3330 2 3331 2 3332 2 3333 2 3334 2 3335 2 3336 2 3337 2 3338 2 3339 3339 3340 3340 3341 3341 3342 3342 3343 3343 3344 3344 3345 3345 3346 3346 3347 3347 3348 3348 3349 3349 3350 3350 3351 3351 3352 22 3353 3353 3354 4 3355 11 3356 3356 3357 89 3358 11 3359 3359 3360 99 3361 3361 3362 1551 3363 3363 3364 3364 3365 3365 3366 99 3367 855 3368 11 3369 3369 3370 943 3371 3371 3372 3372 3373 976 3374 159 3375 162 3376 3376 3377 3377 3378 3378 3379 855 3380 11 3381 3381 3382 3382 3383 3383 3384 357 3385 357 3386 360 3387 434 3388 3388 3389 3389 3390 3390 3391 3391 3392 2 3393 574 3394 3394 3395 3395 3396 3396 3397 3397 3398 3398 3399 3399 3400 2 3401 574 3402 3402 3403 3403 3404 140 3405 11 3406 142 3407 162 3408 163 3409 165 3410 167 3411 77 3412 126 3413 170 3414 3414 3415 3415 3416 140 3417 11 3418 142 3419 162 3420 2248 3421 2250 3422 2251 3423 2252 3424 2 3425 3425 3426 3426 3427 3427 3428 855 3429 3429 3430 11 3431 3431 3432 2 3433 574 3434 3434 3435 3435 3436 3436 3437 2732 3438 3438 3439 3439 3440 357 3441 357 3442 360 3443 434 3444 3444 3445 3445 3446 3446 3447 3447 3448 2 3449 574 3450 3450 3451 3451 3452 3452 3453 3453 3454 2803 3455 3455 3456 434 3457 3457 3458 2 3459 3459 3460 3460 3461 2 3462 3 3463 4 3464 2 3465 2 3466 2 3467 2 3468 11 3469 2 3470 2 3471 3471 3472 2 3473 3473 3474 22 3475 3475 3476 3476 3477 2 3478 11 3479 11 3480 3480 3481 3481 3482 3482 3483 22 3484 2 3485 2 3486 2 3487 11 3488 3488 3489 11 3490 3490 3491 11 3492 3492 3493 3493 3494 3494 3495 3495 3496 60 3497 3497 3498 2 3499 2 3500 3500 3501 11 3502 2 3503 2 3504 69 3505 131 3506 3506 3507 2 3508 2 3509 3509 3510 2 3511 3511 3512 3512 3513 2 3514 85 3515 86 3516 89 3517 134 3518 11 3519 3519 3520 3520 3521 99 3522 3522 3523 3523 3524 3524 3525 104 3526 104 3527 3527 3528 3528 3529 3529 3530 3530 3531 3531 3532 3532 3533 3533 3534 3534 3535 3535 3536 3536 3537 3537 3538 99 3539 2 3540 3540 3541 2 3542 2 3543 2 3544 2 3545 2 3546 2 3547 2 3548 2 3549 307 3550 2 3551 2 3552 2 3553 2 3554 2 3555 313 3556 2 3557 2 3558 2 3559 2 3560 2 3561 2 3562 2 3563 2 3564 2 3565 2 3566 2 3567 2 3568 2 3569 2 3570 2 3571 3571 3572 3572 3573 3573 3574 3574 3575 3575 3576 3576 3577 3577 3578 3578 3579 3579 3580 3580 3581 3581 3582 3582 3583 3583 3584 3584 3585 3585 3586 4 3587 2 3588 3588 3589 3589 3590 3590 3591 3591 3592 3592 3593 3593 3594 3594 3595 3595 3596 3596 3597 3597 3598 3598 3599 357 3600 357 3601 40 3602 360 3603 434 3604 3604 3605 3605 3606 3606 3607 3607 3608 3608 3609 2 3610 574 3611 3611 3612 3612 3613 40 3614 3614 3615 282 3616 3616 3617 3617 3618 2 3619 574 3620 3620 3621 3621 3622 3622 3623 3623 3624 3624 3625 3625 3626 2 3627 740 3628 2 3629 3629 3630 3630 3631 3631 3632 3632 3633 3633 3634 3634 3635 3635 3636 3636 3637 3637 3638 3638 3639 855 3640 3640 3641 3641 3642 3642 3643 3643 3644 3644 3645 3645 3646 3646 3647 3647 3648 3648 3649 3649 3650 3650 3651 3651 3652 3652 3653 3653 3654 3654 3655 3655 3656 3656 3657 3657 3658 3658 3659 3659 3660 3660 3661 3661 3662 3662 3663 3663 3664 3664 3665 3665 3666 3666 3667 3667 3668 3668 3669 3669 3670 3670 3671 3671 3672 3672 3673 3673 3674 3674 3675 3675 3676 3676 3677 3677 3678 3678 3679 3679 3680 3680 3681 3681 3682 3682 3683 3683 3684 3684 3685 3685 3686 3686 3687 3687 3688 3688 3689 3689 3690 3690 3691 3691 3692 3692 3693 3693 3694 3694 3695 3695 3696 3696 3697 3697 3698 3698 3699 3699 3700 3700 3701 3701 3702 3702 3703 3703 3704 3704 3705 3705 3706 3706 3707 3707 3708 3708 3709 3709 3710 3710 3711 3711 3712 3712 3713 3713 3714 3714 3715 3715 3716 3716 3717 3717 3718 3718 3719 3719 3720 11 3721 3721 3722 3722 3723 3723 3724 2 3725 574 3726 3726 3727 943 3728 3728 3729 3729 3730 976 3731 159 3732 162 3733 3733 3734 3734 3735 1487 3736 2 3737 2 3738 2 3739 3739 3740 2 3741 3741 3742 2 3743 2 3744 2 3745 2 3746 2 3747 2 3748 2 3749 2 3750 2 3751 3751 3752 3752 3753 3753 3754 3754 3755 3755 3756 3756 3757 3757 3758 3758 3759 3759 3760 3760 3761 3761 3762 3762 3763 3763 3764 22 3765 3765 3766 4 3767 11 3768 3768 3769 89 3770 11 3771 3771 3772 99 3773 3773 3774 1551 3775 3775 3776 3776 3777 3777 3778 99 3779 855 3780 11 3781 3781 3782 943 3783 3783 3784 3784 3785 976 3786 159 3787 162 3788 3788 3789 3789 3790 3790 3791 855 3792 11 3793 3793 3794 3794 3795 3795 3796 357 3797 357 3798 360 3799 434 3800 3800 3801 3801 3802 3802 3803 3803 3804 2 3805 574 3806 3806 3807 3807 3808 3808 3809 3809 3810 3810 3811 3811 3812 2 3813 574 3814 3814 3815 3815 3816 140 3817 11 3818 142 3819 162 3820 163 3821 165 3822 167 3823 77 3824 126 3825 170 3826 3826 3827 3827 3828 140 3829 11 3830 142 3831 162 3832 2248 3833 2250 3834 2251 3835 2252 3836 2 3837 3837 3838 3838 3839 2 3840 3840 3841 855 3842 3842 3843 11 3844 3844 3845 2 3846 574 3847 3847 3848 3848 3849 3849 3850 2732 3851 3851 3852 3852 3853 357 3854 357 3855 360 3856 434 3857 3857 3858 3858 3859 3859 3860 3860 3861 2 3862 574 3863 3863 3864 3864 3865 3865 3866 3866 3867 2803 3868 3868 3869 434 3870 3870 3871 2 3872 3872 3873 3873 3874 3524 3875 2 3876 3876 3877 3877 3878 3878 3879 3879 3880 3880 3881 2894 3882 3882 3883 69 3884 3884 3885 3885 3886 3886 3887 3887 3888 3888 3889 3889 3890 3890 3891 2 3892 3892 3893 3893 3894 3894 3895 3895 3896 3896 3897 3897 3898 2936 3899 2938 3900 3900 3901 3901 3902 3902 3903 3903 3904 3904 3905 357 3906 357 3907 40 3908 360 3909 434 3910 3910 3911 3911 3912 3912 3913 3913 3914 3914 3915 40 3916 3916 3917 3917 3918 2959 3919 2962 3920 2964 3921 2966 3922 77 3923 2968 3924 2969 3925 3925 3926 282 3927 3927 3928 2 3929 282 3930 77 3931 3931 3932 2 3933 3 3934 4 3935 2 3936 2 3937 2 3938 2 3939 11 3940 2 3941 2 3942 3942 3943 2 3944 3944 3945 22 3946 3946 3947 3947 3948 3948 3949 2 3950 11 3951 11 3952 3952 3953 3953 3954 3954 3955 22 3956 2 3957 2 3958 2 3959 11 3960 3960 3961 11 3962 3962 3963 11 3964 3964 3965 3965 3966 3966 3967 3967 3968 60 3969 3969 3970 2 3971 2 3972 3972 3973 11 3974 2 3975 2 3976 69 3977 131 3978 3978 3979 2 3980 2 3981 3981 3982 2 3983 3983 3984 3984 3985 2 3986 85 3987 86 3988 89 3989 134 3990 11 3991 3991 3992 3992 3993 99 3994 3994 3995 3995 3996 3996 3997 104 3998 104 3999 3999 4000 4000 4001 4001 4002 4002 4003 4003 4004 3946 4005 2828 4006 4006 4007 4007 4008 4008 4009 4009 4010 4010 4011 4011 4012 4012 4013 4013 4014 99 4015 2 4016 4016 4017 2 4018 2 4019 2 4020 2 4021 2 4022 2 4023 2 4024 2 4025 307 4026 2 4027 2 4028 2 4029 2 4030 2 4031 313 4032 2 4033 2 4034 2 4035 2 4036 2 4037 2 4038 2 4039 2 4040 2 4041 2 4042 2 4043 2 4044 2 4045 2 4046 2 4047 4047 4048 4048 4049 4049 4050 4050 4051 4051 4052 4052 4053 4053 4054 4054 4055 4055 4056 4056 4057 4057 4058 4058 4059 4059 4060 4060 4061 4061 4062 4 4063 2 4064 4064 4065 4065 4066 4066 4067 4067 4068 4068 4069 4069 4070 4070 4071 4071 4072 4072 4073 4073 4074 4074 4075 357 4076 357 4077 40 4078 360 4079 434 4080 4080 4081 4081 4082 4082 4083 4083 4084 4084 4085 2 4086 574 4087 4087 4088 4088 4089 40 4090 4090 4091 282 4092 4092 4093 4093 4094 2 4095 574 4096 4096 4097 4097 4098 4098 4099 4099 4100 4100 4101 4101 4102 2 4103 740 4104 2 4105 4105 4106 4106 4107 4107 4108 4108 4109 4109 4110 4110 4111 4111 4112 4112 4113 4113 4114 4114 4115 855 4116 4116 4117 4117 4118 4118 4119 4119 4120 4120 4121 4121 4122 4122 4123 4123 4124 4124 4125 4125 4126 4126 4127 4127 4128 4128 4129 4129 4130 4130 4131 4131 4132 4132 4133 4133 4134 4134 4135 4135 4136 4136 4137 4137 4138 4138 4139 4139 4140 4140 4141 4141 4142 4142 4143 4143 4144 4144 4145 4145 4146 4146 4147 4147 4148 4148 4149 4149 4150 4150 4151 4151 4152 4152 4153 4153 4154 4154 4155 4155 4156 4156 4157 4157 4158 4158 4159 4159 4160 4160 4161 4161 4162 4162 4163 4163 4164 4164 4165 4165 4166 4166 4167 4167 4168 4168 4169 4169 4170 4170 4171 4171 4172 4172 4173 4173 4174 4174 4175 4175 4176 4176 4177 4177 4178 4178 4179 4179 4180 4180 4181 4181 4182 4182 4183 4183 4184 4184 4185 4185 4186 4186 4187 4187 4188 4188 4189 4189 4190 4190 4191 4191 4192 4192 4193 4193 4194 4194 4195 4195 4196 11 4197 4197 4198 4198 4199 4199 4200 2 4201 574 4202 4202 4203 943 4204 4204 4205 4205 4206 976 4207 159 4208 162 4209 4209 4210 4210 4211 1487 4212 2 4213 2 4214 2 4215 4215 4216 2 4217 4217 4218 2 4219 2 4220 2 4221 2 4222 2 4223 2 4224 2 4225 2 4226 2 4227 4227 4228 4228 4229 4229 4230 4230 4231 4231 4232 4232 4233 4233 4234 4234 4235 4235 4236 4236 4237 4237 4238 4238 4239 4239 4240 22 4241 4241 4242 4 4243 11 4244 4244 4245 89 4246 11 4247 4247 4248 99 4249 4249 4250 1551 4251 4251 4252 4252 4253 4253 4254 99 4255 855 4256 11 4257 4257 4258 943 4259 4259 4260 4260 4261 976 4262 159 4263 162 4264 4264 4265 4265 4266 4266 4267 855 4268 11 4269 4269 4270 4270 4271 4271 4272 357 4273 357 4274 360 4275 434 4276 4276 4277 4277 4278 4278 4279 4279 4280 2 4281 574 4282 4282 4283 4283 4284 4284 4285 4285 4286 4286 4287 4287 4288 2 4289 574 4290 4290 4291 4291 4292 4292 4293 4293 4294 4294 4295 4295 4296 4296 4297 2250 4298 2 4299 2 4300 3 4301 4 4302 2 4303 2 4304 2 4305 2 4306 11 4307 2 4308 2 4309 4309 4310 2 4311 4311 4312 22 4313 4313 4314 4314 4315 2 4316 11 4317 11 4318 4318 4319 4319 4320 4320 4321 22 4322 2 4323 2 4324 2 4325 11 4326 4326 4327 11 4328 4328 4329 11 4330 4330 4331 4331 4332 4332 4333 4333 4334 60 4335 4335 4336 2 4337 2 4338 4338 4339 11 4340 2 4341 2 4342 69 4343 131 4344 4344 4345 2 4346 2 4347 4347 4348 2 4349 4349 4350 4350 4351 2 4352 85 4353 86 4354 89 4355 134 4356 11 4357 4357 4358 4358 4359 99 4360 4360 4361 4361 4362 4362 4363 104 4364 104 4365 4365 4366 4366 4367 4367 4368 4368 4369 4369 4370 4370 4371 4371 4372 4372 4373 4373 4374 4374 4375 4375 4376 99 4377 2 4378 4378 4379 2 4380 2 4381 2 4382 2 4383 2 4384 2 4385 2 4386 2 4387 307 4388 2 4389 2 4390 2 4391 2 4392 2 4393 313 4394 2 4395 2 4396 2 4397 2 4398 2 4399 2 4400 2 4401 2 4402 2 4403 2 4404 2 4405 2 4406 2 4407 2 4408 2 4409 4409 4410 4410 4411 4411 4412 4412 4413 4413 4414 4414 4415 4415 4416 4416 4417 4417 4418 4418 4419 4419 4420 4420 4421 4421 4422 4422 4423 4423 4424 4 4425 2 4426 4426 4427 4427 4428 4428 4429 4429 4430 4430 4431 4431 4432 4432 4433 4433 4434 4434 4435 4435 4436 4436 4437 357 4438 357 4439 40 4440 360 4441 434 4442 4442 4443 4443 4444 4444 4445 4445 4446 4446 4447 2 4448 574 4449 4449 4450 4450 4451 40 4452 4452 4453 282 4454 4454 4455 4455 4456 2 4457 574 4458 4458 4459 4459 4460 4460 4461 4461 4462 4462 4463 4463 4464 2 4465 740 4466 2 4467 4467 4468 4468 4469 4469 4470 4470 4471 4471 4472 4472 4473 4473 4474 4474 4475 4475 4476 4476 4477 855 4478 4478 4479 4479 4480 4480 4481 4481 4482 4482 4483 4483 4484 4484 4485 4485 4486 4486 4487 4487 4488 4488 4489 4489 4490 4490 4491 4491 4492 4492 4493 4493 4494 4494 4495 4495 4496 4496 4497 4497 4498 4498 4499 4499 4500 4500 4501 4501 4502 4502 4503 4503 4504 4504 4505 4505 4506 4506 4507 4507 4508 4508 4509 4509 4510 4510 4511 4511 4512 4512 4513 4513 4514 4514 4515 4515 4516 4516 4517 4517 4518 4518 4519 4519 4520 4520 4521 4521 4522 4522 4523 4523 4524 4524 4525 4525 4526 4526 4527 4527 4528 4528 4529 4529 4530 4530 4531 4531 4532 4532 4533 4533 4534 4534 4535 4535 4536 4536 4537 4537 4538 4538 4539 4539 4540 4540 4541 4541 4542 4542 4543 4543 4544 4544 4545 4545 4546 4546 4547 4547 4548 4548 4549 4549 4550 4550 4551 4551 4552 4552 4553 4553 4554 4554 4555 4555 4556 4556 4557 4557 4558 11 4559 4559 4560 4560 4561 4561 4562 2 4563 574 4564 4564 4565 943 4566 4566 4567 4567 4568 976 4569 159 4570 162 4571 4571 4572 4572 4573 1487 4574 2 4575 2 4576 2 4577 4577 4578 2 4579 4579 4580 2 4581 2 4582 2 4583 2 4584 2 4585 2 4586 2 4587 2 4588 2 4589 4589 4590 4590 4591 4591 4592 4592 4593 4593 4594 4594 4595 4595 4596 4596 4597 4597 4598 4598 4599 4599 4600 4600 4601 4601 4602 22 4603 4603 4604 4 4605 11 4606 4606 4607 89 4608 11 4609 4609 4610 99 4611 4611 4612 1551 4613 4613 4614 4614 4615 4615 4616 99 4617 855 4618 11 4619 4619 4620 943 4621 4621 4622 4622 4623 976 4624 159 4625 162 4626 4626 4627 4627 4628 4628 4629 855 4630 11 4631 4631 4632 4632 4633 4633 4634 357 4635 357 4636 360 4637 434 4638 4638 4639 4639 4640 4640 4641 4641 4642 2 4643 574 4644 4644 4645 4645 4646 4646 4647 4647 4648 4648 4649 4649 4650 2 4651 574 4652 4652 4653 4653 4654 140 4655 11 4656 142 4657 162 4658 163 4659 165 4660 167 4661 77 4662 126 4663 170 4664 4664 4665 4665 4666 140 4667 11 4668 142 4669 162 4670 2248 4671 2250 4672 2251 4673 2252 4674 2 4675 4675 4676 4676 4677 855 4678 4678 4679 11 4680 4680 4681 2 4682 574 4683 4683 4684 4684 4685 4685 4686 2732 4687 4687 4688 4688 4689 357 4690 357 4691 360 4692 434 4693 4693 4694 4694 4695 4695 4696 4696 4697 2 4698 574 4699 4699 4700 4700 4701 4701 4702 4702 4703 2803 4704 4704 4705 434 4706 4706 4707 2 4708 4708 4709 4709 4710 2 4711 3 4712 4 4713 2 4714 2 4715 2 4716 2 4717 11 4718 2 4719 2 4720 4720 4721 2 4722 4722 4723 22 4724 4724 4725 4725 4726 2 4727 11 4728 11 4729 4729 4730 4730 4731 4731 4732 22 4733 2 4734 2 4735 2 4736 11 4737 4737 4738 11 4739 4739 4740 11 4741 4741 4742 4742 4743 4743 4744 4744 4745 60 4746 4746 4747 2 4748 2 4749 4749 4750 11 4751 2 4752 2 4753 69 4754 131 4755 4755 4756 2 4757 2 4758 4758 4759 2 4760 4760 4761 4761 4762 2 4763 85 4764 86 4765 89 4766 134 4767 11 4768 4768 4769 4769 4770 99 4771 4771 4772 4772 4773 4773 4774 104 4775 104 4776 4776 4777 4777 4778 4778 4779 4779 4780 4780 4781 4781 4782 4782 4783 4783 4784 4784 4785 4785 4786 4786 4787 99 4788 2 4789 4789 4790 2 4791 2 4792 2 4793 2 4794 2 4795 2 4796 2 4797 2 4798 307 4799 2 4800 2 4801 2 4802 2 4803 2 4804 313 4805 2 4806 2 4807 2 4808 2 4809 2 4810 2 4811 2 4812 2 4813 2 4814 2 4815 2 4816 2 4817 2 4818 2 4819 2 4820 4820 4821 4821 4822 4822 4823 4823 4824 4824 4825 4825 4826 4826 4827 4827 4828 4828 4829 4829 4830 4830 4831 4831 4832 4832 4833 4833 4834 4834 4835 4 4836 2 4837 4837 4838 4838 4839 4839 4840 4840 4841 4841 4842 4842 4843 4843 4844 4844 4845 4845 4846 4846 4847 4847 4848 357 4849 357 4850 40 4851 360 4852 434 4853 4853 4854 4854 4855 4855 4856 4856 4857 4857 4858 2 4859 574 4860 4860 4861 4861 4862 40 4863 4863 4864 282 4865 4865 4866 4866 4867 2 4868 574 4869 4869 4870 4870 4871 4871 4872 4872 4873 4873 4874 4874 4875 2 4876 740 4877 2 4878 4878 4879 4879 4880 4880 4881 4881 4882 4882 4883 4883 4884 4884 4885 4885 4886 4886 4887 4887 4888 855 4889 4889 4890 4890 4891 4891 4892 4892 4893 4893 4894 4894 4895 4895 4896 4896 4897 4897 4898 4898 4899 4899 4900 4900 4901 4901 4902 4902 4903 4903 4904 4904 4905 4905 4906 4906 4907 4907 4908 4908 4909 4909 4910 4910 4911 4911 4912 4912 4913 4913 4914 4914 4915 4915 4916 4916 4917 4917 4918 4918 4919 4919 4920 4920 4921 4921 4922 4922 4923 4923 4924 4924 4925 4925 4926 4926 4927 4927 4928 4928 4929 4929 4930 4930 4931 4931 4932 4932 4933 4933 4934 4934 4935 4935 4936 4936 4937 4937 4938 4938 4939 4939 4940 4940 4941 4941 4942 4942 4943 4943 4944 4944 4945 4945 4946 4946 4947 4947 4948 4948 4949 4949 4950 4950 4951 4951 4952 4952 4953 4953 4954 4954 4955 4955 4956 4956 4957 4957 4958 4958 4959 4959 4960 4960 4961 4961 4962 4962 4963 4963 4964 4964 4965 4965 4966 4966 4967 4967 4968 4968 4969 11 4970 4970 4971 4971 4972 4972 4973 2 4974 574 4975 4975 4976 943 4977 4977 4978 4978 4979 976 4980 159 4981 162 4982 4982 4983 4983 4984 1487 4985 2 4986 2 4987 2 4988 4988 4989 2 4990 4990 4991 2 4992 2 4993 2 4994 2 4995 2 4996 2 4997 2 4998 2 4999 2 5000 5000 5001 5001 5002 5002 5003 5003 5004 5004 5005 5005 5006 5006 5007 5007 5008 5008 5009 5009 5010 5010 5011 5011 5012 5012 5013 22 5014 5014 5015 4 5016 11 5017 5017 5018 89 5019 11 5020 5020 5021 99 5022 5022 5023 1551 5024 5024 5025 5025 5026 5026 5027 99 5028 855 5029 11 5030 5030 5031 943 5032 5032 5033 5033 5034 976 5035 159 5036 162 5037 5037 5038 5038 5039 5039 5040 855 5041 11 5042 5042 5043 5043 5044 5044 5045 357 5046 357 5047 360 5048 434 5049 5049 5050 5050 5051 5051 5052 5052 5053 2 5054 574 5055 5055 5056 5056 5057 5057 5058 5058 5059 5059 5060 5060 5061 2 5062 574 5063 5063 5064 5064 5065 140 5066 11 5067 142 5068 162 5069 163 5070 165 5071 167 5072 77 5073 126 5074 170 5075 5075 5076 5076 5077 140 5078 11 5079 142 5080 162 5081 2248 5082 2250 5083 2251 5084 2252 5085 2 5086 5086 5087 5087 5088 2 5089 5089 5090 855 5091 5091 5092 11 5093 5093 5094 2 5095 574 5096 5096 5097 5097 5098 5098 5099 2732 5100 5100 5101 5101 5102 357 5103 357 5104 360 5105 434 5106 5106 5107 5107 5108 5108 5109 5109 5110 2 5111 574 5112 5112 5113 5113 5114 5114 5115 5115 5116 2803 5117 5117 5118 434 5119 5119 5120 2 5121 5121 5122 5122 5123 4773 5124 2 5125 5125 5126 5126 5127 5127 5128 5128 5129 5129 5130 2894 5131 5131 5132 69 5133 5133 5134 5134 5135 5135 5136 5136 5137 5137 5138 5138 5139 5139 5140 2 5141 5141 5142 5142 5143 5143 5144 5144 5145 5145 5146 5146 5147 2936 5148 2938 5149 5149 5150 5150 5151 5151 5152 5152 5153 5153 5154 357 5155 357 5156 40 5157 360 5158 434 5159 5159 5160 5160 5161 5161 5162 5162 5163 5163 5164 40 5165 5165 5166 5166 5167 2959 5168 2962 5169 2964 5170 2966 5171 77 5172 2968 5173 2969 5174 5174 5175 282 5176 5176 5177 2 5178 282 5179 77 5180 3931 5181 2828 5182 5182 5183 1862 5184 162 5185 2050 5186 2 5187 5187 5188 2 5189 5189 5190 5190 5191 159 5192 5192 5193 162 5194 5194 5195 5195 5196 5196 5197 5197 5198 5198 5199 5199 5200 943 5201 5201 5202 5202 5203 5203 5204 976 5205 159 5206 162 5207 5207 5208 5208 5209 2 5210 282 5211 5211 5212 2 5213 282 5214 5214 5215 5215 5216 5216 5217 5217 5218 5218 5219 5219 5220 1862 5221 162 5222 5222 5223 5223 5224 5224 5225 5225 5226 282 5227 5227 5228 5228 5229 2 5230 574 5231 5231 5232 22 5233 2800 5234 5234 5235 5235 5236 5236 5237 5237 5238 5238 5239 5239 5240 5240 5241 16 5242 16 5243 2811 5244 434 5245 5245 5246 5246 5247 5247 5248 5248 5249 2822 5250 574 5251 5251 5252 2 5253 5253 5254 2827 5255 2828 5256 5256 5257 2841 5258 5258 5259 5259 5260 5260 5261 5261 5262 5262 5263 5263 5264 2894 5265 5265 5266 69 5267 5267 5268 5268 5269 5269 5270 5270 5271 5271 5272 5272 5273 5273 5274 2 5275 5275 5276 5276 5277 5277 5278 5278 5279 5279 5280 5280 5281 2936 5282 2938 5283 5283 5284 5284 5285 5285 5286 5286 5287 5287 5288 357 5289 357 5290 40 5291 360 5292 434 5293 5293 5294 5294 5295 5295 5296 5296 5297 5297 5298 40 5299 5299 5300 5300 5301 2959 5302 2962 5303 2964 5304 2966 5305 77 5306 2968 5307 2969 5308 5308 5309 282 5310 5310 5311 2 5312 282 5313 77 5314 3931 5315 2828 5316 5316 5317 1862 5318 162 5319 2050 5320 2 5321 5321 5322 2 5323 159 5324 162 5325 5325 5326 5326 5327 5327 5328 5328 5329 5329 5330 943 5331 5331 5332 5332 5333 5333 5334 976 5335 159 5336 162 5337 5337 5338 5338 5339 2 5340 282 5341 5341 5342 2 5343 282 5344 5344 5345 5345 5346 5346 5347 5347 5348 5348 5349 5349 5350 1862 5351 162 5352 5352 5353 5353 5354 5354 5355 5355 5356 282 5357 5357 5358 5358 5359 2 5360 574 5361 5361 5362 5236 5363 5363 5364 5364 5365 5365 5366 140 5367 11 5368 142 5369 162 5370 163 5371 165 5372 167 5373 77 5374 126 5375 170 5376 5376 5377 5377 5378 140 5379 11 5380 142 5381 162 5382 2248 5383 2250 5384 2251 5385 2252 5386 2 5387 5387 5388 5388 5389 855 5390 5390 5391 11 5392 5392 5393 2 5394 574 5395 5395 5396 5396 5397 5397 5398 2732 5399 5399 5400 5400 5401 357 5402 357 5403 360 5404 434 5405 5405 5406 5406 5407 5407 5408 5408 5409 2 5410 574 5411 5411 5412 5412 5413 5413 5414 5414 5415 2803 5416 5416 5417 434 5418 5418 5419 2 5420 5420 5421 5421 5422 5422 5423 2 5424 5424 5425 5425 5426 5426 5427 5427 5428 5428 5429 2894 5430 5430 5431 69 5432 5432 5433 5433 5434 5434 5435 5435 5436 5436 5437 5437 5438 5438 5439 2 5440 5440 5441 5441 5442 5442 5443 5443 5444 5444 5445 5445 5446 2936 5447 2938 5448 5448 5449 5449 5450 5450 5451 5451 5452 5452 5453 357 5454 357 5455 40 5456 360 5457 434 5458 5458 5459 5459 5460 5460 5461 5461 5462 5462 5463 40 5464 5464 5465 5465 5466 2959 5467 2962 5468 2964 5469 2966 5470 77 5471 2968 5472 2969 5473 5473 5474 282 5475 5475 5476 2 5477 282 5478 77 5479 3931 5480 2828 5481 5481 5482 1862 5483 162 5484 2050 5485 2 5486 5486 5487 2 5488 159 5489 162 5490 5490 5491 5491 5492 5492 5493 5493 5494 5494 5495 943 5496 5496 5497 5497 5498 5498 5499 976 5500 159 5501 162 5502 5502 5503 5503 5504 2 5505 282 5506 5506 5507 2 5508 282 5509 5509 5510 5510 5511 5511 5512 2250 5513 2 5514 5514 5515 5515 5516 5516 5517 5517 5518 5518 5519 1862 5520 162 5521 5521 5522 5522 5523 5523 5524 5524 5525 282 5526 5526 5527 5527 5528 2 5529 574 5530 5530 5531 5236 5532 5532 5533 5533 5534 1862 5535 162 5536 2050 5537 2 5538 5538 5539 2 5540 159 5541 162 5542 5542 5543 5543 5544 5544 5545 5545 5546 5546 5547 943 5548 5548 5549 5549 5550 5550 5551 976 5552 159 5553 162 5554 5554 5555 5555 5556 2 5557 282 5558 5558 5559 2 5560 282 5561 5561 5562 5562 5563 5563 5564 5564 5565 5565 5566 5566 5567 1862 5568 162 5569 5569 5570 5570 5571 5571 5572 5572 5573 282 5574 5574 5575 5575 5576 2 5577 574 5578 5578 5579 5236 5580 5580 5581 5581 5582 5582 5583 5583 5584 5584 5585 5585 5586 2894 5587 5587 5588 69 5589 5589 5590 5590 5591 5591 5592 5592 5593 5593 5594 5594 5595 5595 5596 2 5597 5597 5598 5598 5599 5599 5600 5600 5601 5601 5602 5602 5603 2936 5604 2938 5605 5605 5606 5606 5607 5607 5608 5608 5609 5609 5610 357 5611 357 5612 40 5613 360 5614 434 5615 5615 5616 5616 5617 5617 5618 5618 5619 5619 5620 40 5621 5621 5622 5622 5623 2959 5624 2962 5625 2964 5626 2966 5627 77 5628 2968 5629 2969 5630 5630 5631 282 5632 5632 5633 2 5634 282 5635 77 5636 3931 5637 2828 5638 5638 5639 1862 5640 162 5641 2050 5642 2 5643 5643 5644 2 5645 159 5646 162 5647 5647 5648 5648 5649 5649 5650 5650 5651 5651 5652 943 5653 5653 5654 5654 5655 5655 5656 976 5657 159 5658 162 5659 5659 5660 5660 5661 2 5662 282 5663 5663 5664 2 5665 282 5666 5666 5667 5667 5668 5668 5669 5669 5670 5670 5671 5671 5672 1862 5673 162 5674 5674 5675 5675 5676 5676 5677 5677 5678 282 5679 5679 5680 5680 5681 2 5682 574 5683 5683 5684 5236 5685 5685 5686 5686 5687 2 5688 5688 5689 5689 5690 5690 5691 5691 5692 13 5693 5693 5694 5694 5695 5695 5696 5696 5697 5697 5698 5698 5699 5699 5700 5700 5701 5701 5702 5702 5703 5703 5704 5704 5705 5705 5706 5706 5707 5707 5708 5708 5709 5709 5710 5710 5711 5711 5712 16 5713 5713 5714 5714 5715 142 5716 5716 5717 5717 5718 5718 5719 2 5720 5720 5721 126 5722 170 5723 5723 5724 5724 5725 103 5726 2 5727 2 5728 2 5729 2 5730 2 5731 2 5732 5732 5733 2 5734 2 5735 307 5736 2 5737 2 5738 2 5739 2 5740 2 5741 313 5742 2 5743 2 5744 2 5745 2 5746 2 5747 2 5748 2 5749 2 5750 2 5751 2 5752 2 5753 2 5754 2 5755 2 5756 2 5757 5757 5758 5758 5759 5759 5760 5760 5761 5761 5762 5762 5763 5763 5764 5764 5765 5765 5766 5766 5767 5767 5768 5768 5769 5769 5770 5770 5771 5771 5772 4 5773 2 5774 5774 5775 5775 5776 5776 5777 5777 5778 5778 5779 5779 5780 5780 5781 5781 5782 5782 5783 5783 5784 5784 5785 357 5786 357 5787 40 5788 360 5789 434 5790 5790 5791 5791 5792 5792 5793 5793 5794 5794 5795 2 5796 574 5797 5797 5798 5798 5799 40 5800 5800 5801 282 5802 5802 5803 5803 5804 2 5805 574 5806 5806 5807 5807 5808 5808 5809 5809 5810 5810 5811 5811 5812 2 5813 740 5814 2 5815 5815 5816 5816 5817 5817 5818 5818 5819 5819 5820 5820 5821 5821 5822 5822 5823 5823 5824 5824 5825 162 5826 5826 5827 5827 5828 5828 5829 5829 5830 5830 5831 5831 5832 5832 5833 5833 5834 5834 5835 5835 5836 5836 5837 5837 5838 5838 5839 943 5840 5840 5841 5841 5842 976 5843 159 5844 162 5845 5845 5846 5846 5847 5847 5848 434 5849 5849 5850 5850 5851 2 5852 3 5853 4 5854 2 5855 2 5856 2 5857 2 5858 11 5859 2 5860 2 5861 5861 5862 2 5863 5863 5864 22 5865 5865 5866 5866 5867 2 5868 11 5869 11 5870 5870 5871 5871 5872 5872 5873 22 5874 2 5875 2 5876 2 5877 11 5878 5878 5879 11 5880 5880 5881 11 5882 5882 5883 5883 5884 5884 5885 5885 5886 60 5887 5887 5888 2 5889 2 5890 5890 5891 11 5892 2 5893 2 5894 69 5895 131 5896 5896 5897 2 5898 2 5899 5899 5900 2 5901 5901 5902 5902 5903 2 5904 85 5905 86 5906 89 5907 134 5908 11 5909 5909 5910 5910 5911 99 5912 5912 5913 5913 5914 5914 5915 104 5916 104 5917 5917 5918 5918 5919 5919 5920 5920 5921 5921 5922 5922 5923 5923 5924 5924 5925 5925 5926 5926 5927 5927 5928 99 5929 2 5930 5930 5931 2 5932 2 5933 2 5934 2 5935 2 5936 2 5937 2 5938 2 5939 307 5940 2 5941 2 5942 2 5943 2 5944 2 5945 313 5946 2 5947 2 5948 2 5949 2 5950 2 5951 2 5952 2 5953 2 5954 2 5955 2 5956 2 5957 2 5958 2 5959 2 5960 2 5961 5961 5962 5962 5963 5963 5964 5964 5965 5965 5966 5966 5967 5967 5968 5968 5969 5969 5970 5970 5971 5971 5972 5972 5973 5973 5974 5974 5975 5975 5976 4 5977 2 5978 5978 5979 5979 5980 5980 5981 5981 5982 5982 5983 5983 5984 5984 5985 5985 5986 5986 5987 5987 5988 5988 5989 357 5990 357 5991 40 5992 360 5993 434 5994 5994 5995 5995 5996 5996 5997 5997 5998 5998 5999 2 6000 574 6001 6001 6002 6002 6003 40 6004 6004 6005 282 6006 6006 6007 6007 6008 2 6009 574 6010 6010 6011 6011 6012 6012 6013 6013 6014 6014 6015 6015 6016 2 6017 740 6018 2 6019 6019 6020 6020 6021 6021 6022 6022 6023 6023 6024 6024 6025 6025 6026 6026 6027 6027 6028 6028 6029 855 6030 6030 6031 6031 6032 6032 6033 6033 6034 6034 6035 6035 6036 6036 6037 6037 6038 6038 6039 6039 6040 6040 6041 6041 6042 6042 6043 6043 6044 6044 6045 6045 6046 6046 6047 6047 6048 6048 6049 6049 6050 6050 6051 6051 6052 6052 6053 6053 6054 6054 6055 6055 6056 6056 6057 6057 6058 6058 6059 6059 6060 6060 6061 6061 6062 6062 6063 6063 6064 6064 6065 6065 6066 6066 6067 6067 6068 6068 6069 6069 6070 6070 6071 6071 6072 6072 6073 6073 6074 6074 6075 6075 6076 6076 6077 6077 6078 6078 6079 6079 6080 6080 6081 6081 6082 6082 6083 6083 6084 6084 6085 6085 6086 6086 6087 6087 6088 6088 6089 6089 6090 6090 6091 6091 6092 6092 6093 6093 6094 6094 6095 6095 6096 6096 6097 6097 6098 6098 6099 6099 6100 6100 6101 6101 6102 6102 6103 6103 6104 6104 6105 6105 6106 6106 6107 6107 6108 6108 6109 6109 6110 11 6111 6111 6112 6112 6113 6113 6114 2 6115 574 6116 6116 6117 943 6118 6118 6119 6119 6120 976 6121 159 6122 162 6123 6123 6124 6124 6125 1487 6126 2 6127 2 6128 2 6129 6129 6130 2 6131 6131 6132 2 6133 2 6134 2 6135 2 6136 2 6137 2 6138 2 6139 2 6140 2 6141 6141 6142 6142 6143 6143 6144 6144 6145 6145 6146 6146 6147 6147 6148 6148 6149 6149 6150 6150 6151 6151 6152 6152 6153 6153 6154 22 6155 6155 6156 4 6157 11 6158 6158 6159 89 6160 11 6161 6161 6162 99 6163 6163 6164 1551 6165 6165 6166 6166 6167 6167 6168 99 6169 855 6170 11 6171 6171 6172 943 6173 6173 6174 6174 6175 976 6176 159 6177 162 6178 6178 6179 6179 6180 6180 6181 855 6182 11 6183 6183 6184 6184 6185 6185 6186 357 6187 357 6188 360 6189 434 6190 6190 6191 6191 6192 6192 6193 6193 6194 2 6195 574 6196 6196 6197 6197 6198 6198 6199 6199 6200 6200 6201 6201 6202 2 6203 574 6204 6204 6205 6205 6206 140 6207 11 6208 142 6209 162 6210 163 6211 165 6212 167 6213 77 6214 126 6215 170 6216 6216 6217 6217 6218 140 6219 11 6220 142 6221 162 6222 2248 6223 2250 6224 2251 6225 2252 6226 2 6227 6227 6228 6228 6229 6229 6230 5689 6231 5690 6232 6232 6233 855 6234 6234 6235 11 6236 6236 6237 2 6238 574 6239 6239 6240 6240 6241 6241 6242 2732 6243 6243 6244 6244 6245 357 6246 357 6247 360 6248 434 6249 6249 6250 6250 6251 6251 6252 6252 6253 2 6254 574 6255 6255 6256 6256 6257 6257 6258 6258 6259 2803 6260 6260 6261 434 6262 6262 6263 2 6264 6264 6265 6265 6266 2 6267 3 6268 4 6269 2 6270 2 6271 2 6272 2 6273 11 6274 2 6275 2 6276 6276 6277 2 6278 6278 6279 22 6280 6280 6281 6281 6282 2 6283 11 6284 11 6285 6285 6286 6286 6287 6287 6288 22 6289 2 6290 2 6291 2 6292 11 6293 6293 6294 11 6295 6295 6296 11 6297 6297 6298 6298 6299 6299 6300 6300 6301 60 6302 6302 6303 2 6304 2 6305 6305 6306 11 6307 2 6308 2 6309 69 6310 131 6311 6311 6312 2 6313 2 6314 6314 6315 2 6316 6316 6317 6317 6318 2 6319 85 6320 86 6321 89 6322 134 6323 11 6324 6324 6325 6325 6326 99 6327 6327 6328 6328 6329 6329 6330 104 6331 104 6332 6332 6333 6333 6334 6334 6335 6335 6336 6336 6337 6337 6338 6338 6339 6339 6340 6340 6341 6341 6342 6342 6343 99 6344 2 6345 6345 6346 2 6347 2 6348 2 6349 2 6350 2 6351 2 6352 2 6353 2 6354 307 6355 2 6356 2 6357 2 6358 2 6359 2 6360 313 6361 2 6362 2 6363 2 6364 2 6365 2 6366 2 6367 2 6368 2 6369 2 6370 2 6371 2 6372 2 6373 2 6374 2 6375 2 6376 6376 6377 6377 6378 6378 6379 6379 6380 6380 6381 6381 6382 6382 6383 6383 6384 6384 6385 6385 6386 6386 6387 6387 6388 6388 6389 6389 6390 6390 6391 4 6392 2 6393 6393 6394 6394 6395 6395 6396 6396 6397 6397 6398 6398 6399 6399 6400 6400 6401 6401 6402 6402 6403 6403 6404 357 6405 357 6406 40 6407 360 6408 434 6409 6409 6410 6410 6411 6411 6412 6412 6413 6413 6414 2 6415 574 6416 6416 6417 6417 6418 40 6419 6419 6420 282 6421 6421 6422 6422 6423 2 6424 574 6425 6425 6426 6426 6427 6427 6428 6428 6429 6429 6430 6430 6431 2 6432 740 6433 2 6434 6434 6435 6435 6436 6436 6437 6437 6438 6438 6439 6439 6440 6440 6441 6441 6442 6442 6443 6443 6444 855 6445 6445 6446 6446 6447 6447 6448 6448 6449 6449 6450 6450 6451 6451 6452 6452 6453 6453 6454 6454 6455 6455 6456 6456 6457 6457 6458 6458 6459 6459 6460 6460 6461 6461 6462 6462 6463 6463 6464 6464 6465 6465 6466 6466 6467 6467 6468 6468 6469 6469 6470 6470 6471 6471 6472 6472 6473 6473 6474 6474 6475 6475 6476 6476 6477 6477 6478 6478 6479 6479 6480 6480 6481 6481 6482 6482 6483 6483 6484 6484 6485 6485 6486 6486 6487 6487 6488 6488 6489 6489 6490 6490 6491 6491 6492 6492 6493 6493 6494 6494 6495 6495 6496 6496 6497 6497 6498 6498 6499 6499 6500 6500 6501 6501 6502 6502 6503 6503 6504 6504 6505 6505 6506 6506 6507 6507 6508 6508 6509 6509 6510 6510 6511 6511 6512 6512 6513 6513 6514 6514 6515 6515 6516 6516 6517 6517 6518 6518 6519 6519 6520 6520 6521 6521 6522 6522 6523 6523 6524 6524 6525 11 6526 6526 6527 6527 6528 6528 6529 2 6530 574 6531 6531 6532 943 6533 6533 6534 6534 6535 976 6536 159 6537 162 6538 6538 6539 6539 6540 1487 6541 2 6542 2 6543 2 6544 6544 6545 2 6546 6546 6547 2 6548 2 6549 2 6550 2 6551 2 6552 2 6553 2 6554 2 6555 2 6556 6556 6557 6557 6558 6558 6559 6559 6560 6560 6561 6561 6562 6562 6563 6563 6564 6564 6565 6565 6566 6566 6567 6567 6568 6568 6569 22 6570 6570 6571 4 6572 11 6573 6573 6574 89 6575 11 6576 6576 6577 99 6578 6578 6579 1551 6580 6580 6581 6581 6582 6582 6583 99 6584 855 6585 11 6586 6586 6587 943 6588 6588 6589 6589 6590 976 6591 159 6592 162 6593 6593 6594 6594 6595 6595 6596 855 6597 11 6598 6598 6599 6599 6600 6600 6601 357 6602 357 6603 360 6604 434 6605 6605 6606 6606 6607 6607 6608 6608 6609 2 6610 574 6611 6611 6612 6612 6613 6613 6614 6614 6615 6615 6616 6616 6617 2 6618 574 6619 6619 6620 6620 6621 140 6622 11 6623 142 6624 162 6625 163 6626 165 6627 167 6628 77 6629 126 6630 170 6631 6631 6632 6632 6633 140 6634 11 6635 142 6636 162 6637 2248 6638 2250 6639 2251 6640 2252 6641 2 6642 6642 6643 6643 6644 2 6645 6645 6646 855 6647 6647 6648 11 6649 6649 6650 2 6651 574 6652 6652 6653 6653 6654 6654 6655 2732 6656 6656 6657 6657 6658 357 6659 357 6660 360 6661 434 6662 6662 6663 6663 6664 6664 6665 6665 6666 2 6667 574 6668 6668 6669 6669 6670 6670 6671 6671 6672 2803 6673 6673 6674 434 6675 6675 6676 2 6677 6677 6678 6678 6679 6329 6680 2 6681 6681 6682 6682 6683 6683 6684 6684 6685 6685 6686 2894 6687 6687 6688 69 6689 6689 6690 6690 6691 6691 6692 6692 6693 6693 6694 6694 6695 6695 6696 2 6697 6697 6698 6698 6699 6699 6700 6700 6701 6701 6702 6702 6703 2936 6704 2938 6705 6705 6706 6706 6707 6707 6708 6708 6709 6709 6710 357 6711 357 6712 40 6713 360 6714 434 6715 6715 6716 6716 6717 6717 6718 6718 6719 6719 6720 40 6721 6721 6722 6722 6723 2959 6724 2962 6725 2964 6726 2966 6727 77 6728 2968 6729 2969 6730 6730 6731 282 6732 6732 6733 2 6734 282 6735 77 6736 3931 6737 2828 6738 6738 6739 1862 6740 162 6741 2050 6742 2 6743 6743 6744 2 6745 159 6746 162 6747 6747 6748 6748 6749 6749 6750 6750 6751 6751 6752 943 6753 6753 6754 6754 6755 6755 6756 976 6757 159 6758 162 6759 6759 6760 6760 6761 2 6762 282 6763 6763 6764 2 6765 282 6766 6766 6767 6767 6768 6768 6769 6769 6770 6770 6771 6771 6772 1862 6773 162 6774 6774 6775 6775 6776 6776 6777 6777 6778 282 6779 6779 6780 6780 6781 2 6782 574 6783 6783 6784 5236 6785 6785 6786 6786 6787 6787 6788 6788 6789 6789 6790 6790 6791 2894 6792 6792 6793 69 6794 6794 6795 6795 6796 6796 6797 6797 6798 6798 6799 6799 6800 6800 6801 2 6802 6802 6803 6803 6804 6804 6805 6805 6806 6806 6807 6807 6808 2936 6809 2938 6810 6810 6811 6811 6812 6812 6813 6813 6814 6814 6815 357 6816 357 6817 40 6818 360 6819 434 6820 6820 6821 6821 6822 6822 6823 6823 6824 6824 6825 40 6826 6826 6827 6827 6828 2959 6829 2962 6830 2964 6831 2966 6832 77 6833 2968 6834 2969 6835 6835 6836 282 6837 6837 6838 2 6839 282 6840 77 6841 3931 6842 2828 6843 6843 6844 1862 6845 162 6846 2050 6847 2 6848 6848 6849 2 6850 159 6851 162 6852 6852 6853 6853 6854 6854 6855 6855 6856 6856 6857 943 6858 6858 6859 6859 6860 6860 6861 976 6862 159 6863 162 6864 6864 6865 6865 6866 2 6867 282 6868 6868 6869 2 6870 282 6871 6871 6872 6872 6873 6873 6874 6874 6875 6875 6876 6876 6877 1862 6878 162 6879 6879 6880 6880 6881 6881 6882 6882 6883 282 6884 6884 6885 6885 6886 2 6887 574 6888 6888 6889 5236 6890 6890 6891 6891 6892 2995 6893 6893 6894 6894 6895 6895 6896 6896 6897 6897 6898 6898 6899 6899 6900 6900 6901 6901 6902 2292 6903 6903 6904 6904 6905 6905 6906 6906 6907 6907 6908 6908 6909 6909 6910 6910 6911 6911 6912 6912 6913 6913 6914 6914 6915 6915 6916 6916 6917 6917 6918 6918 6919 6919 6920 6920 6921 6921 6922 6922 6923 6923 6924 6924 6925 6925 6926 6926 6927 6927 6928 6928 6929 6929 6930 6930 6931 6931 6932 6932 6933 6933 6934 6934 6935 6935 6936 6936 6937 6937 6938 6938 6939 6939 6940 6940 6941 6941 6942 6942 6943 6943 6944 6944 6945 6945 6946 6946 6947 6947 6948 6948 6949 6949 6950 6950 6951 6951 6952 6952 6953 6953 6954 6954 6955 6955 6956 6956 6957 6957 6958 6958 6959 6959 6960 6960 6961 6961 6962 6962 6963 6963 6964 6964 6965 6965 6966 6966 6967 6967 6968 6968 6969 6969 6970 6970 6971 6971 6972 6972 6973 6973 6974 6974 6975 6975 6976 6976 6977 6977 6978 6978 6979 6979 6980 6980 6981 6981 6982 6982 6983 6983 6984 6984 6985 6985 6986 6986 6987 6987 6988 6988 6989 6989 6990 140 6991 6991 6992 6992 6993 11 6994 142 6995 162 6996 6996 6997 166 6998 6998 6999 6999 7000 7000 7001 7001 7002 126 7003 170 7004 7004 7005 7005 7006 2256 7007 7007 7008 7008 7009 7009 7010 7010 7011 7011 7012 7012 7013 7013 7014 7014 7015 7015 7016 7016 7017 7017 7018 7018 7019 7019 7020 2271 7021 2272 7022 2256 7023 7023 7024 7024 7025 7025 7026 7026 7027 7027 7028 7028 7029 7029 7030 7030 7031 7031 7032 7032 7033 7033 7034 2292 7035 7035 7036 2272 7037 7037 7038 2256 7039 7039 7040 7040 7041 7041 7042 7042 7043 7043 7044 7044 7045 7045 7046 7046 7047 7047 7048 7048 7049 2256 7050 7050 7051 7051 7052 7052 7053 7053 7054 7054 7055 7055 7056 7056 7057 7057 7058 7058 7059 7059 7060 7060 7061 7061 7062 2256 7063 7063 7064 7064 7065 7065 7066 7066 7067 7067 7068 7068 7069 7069 7070 7070 7071 7071 7072 7072 7073 7073 7074 7074 7075 7075 7076 7076 7077 7077 7078 7078 7079 7079 7080 7080 7081 7081 7082 7082 7083 7083 7084 7084 7085 7085 7086 7086 7087 7087 7088 7088 7089 7089 7090 7090 7091 7091 7092 7092 7093 7093 7094 7094 7095 7095 7096 7096 7097 7097 7098 7098 7099 7099 7100 7100 7101 7101 7102 7102 7103 7103 7104 7104 7105 7105 7106 7106 7107 7107 7108 7108 7109 7109 7110 7110 7111 7111 7112 7112 7113 7113 7114 7114 7115 7115 7116 7116 7117 7117 7118 7118 7119 7119 7120 7120 7121 7121 7122 7122 7123 7123 7124 7124 7125 7125 7126 7126 7127 7127 7128 7128 7129 7129 7130 7130 7131 7131 7132 7132 7133 7133 7134 7134 7135 7135 7136 7136 7137 7137 7138 7138 7139 7139 7140 7140 7141 7141 7142 7142 7143 7143 7144 7144 7145 7145 7146 3042 7147 7147 7148 7148 7149 7149 7150 2050 7151 7151 7152 6910 7153 7153 7154 7154 7155 2271 7156 2272 7157 2256 7158 7158 7159 7159 7160 7160 7161 7161 7162 7162 7163 7163 7164 7164 7165 7165 7166 7166 7167 7167 7168 7168 7169 2292 7170 7170 7171 2272 7172 7172 7173 7173 7174 7174 7175 7175 7176 2050 7177 3042 7178 6910 7179 7179 7180 7180 7181 2 7182 3112 7183 7183 7184 2256 7185 7185 7186 7186 7187 7187 7188 7188 7189 7189 7190 7190 7191 7191 7192 7192 7193 7193 7194 7194 7195 2 7196 3112 7197 7197 7198 2256 7199 7199 7200 7200 7201 7201 7202 7202 7203 7203 7204 7204 7205 7205 7206 7206 7207 7207 7208 7208 7209 7209 7210 7210 7211 7211 7212 7212 7213 7213 7214 7214 7215 7215 7216 7216 7217 7217 7218 7218 7219 7219 7220 7220 7221 13 7222 7222 7223 7223 7224 7224 7225 7225 7226 7226 7227 7227 7228 7228 7229 7229 7230 7230 7231 7231 7232 7232 7233 7233 7234 7234 7235 7235 7236 7236 7237 7237 7238 7238 7239 7239 7240 7240 7241 7241 7242 7242 7243 7243 7244 7244 7245 7245 7246 7246 7247 7247 7248 7248 7249 7249 7250 7250 7251 7251 7252 7252 7253 7253 7254 7254 7255 7255 7256 7256 7257 7257 7258 7258 7259 7259 7260 7260 7261 7261 7262 7262 7263 7263 7264 7264 7265 7265 7266 7266 7267 7267 7268 7268 7269 7269 7270 7270 7271 7271 7272 7272 7273 7273 7274 7274 7275 7275 7276 7276 7277 7277 7278 7278 7279 7279 7280 7280 7281 7281 7282 7282 7283 7283 7284 7284 7285 7285 7286 7286 7287 7287 7288 7288 7289 7289 7290 7290 7291 7291 7292 7292 7293 7293 7294 7294 7295 7295 7296 7296 7297 7297 7298 7298 7299 7299 7300 7300 7301 7301 7302 7302 7303 7303 7304 7304 7305 7305 7306 7306 7307 7307 7308 99 7309 7309 7310 7310 7311 7311 7312 7312 7313 7313 7314 7224 7315 7315 7316 7316 7317 7315 7318 7318 7319 7319 7320 7320 7321 7321 7322 7322 7323 7323 7324 7324 7325 7325 7326 7326 7327 7327 7328 7328 7329 7329 7330 7330 7331 7331 7332 7332 7333 7333 7334 7334 7335 7335 7336 7336 7337 7337 7338 7338 7339 7339 7340 7340 7341 7341 7342 7342 7343 7343 7344 7344 7345 7345 7346 7346 7347 7347 7348 7348 7349 7349 7350 7350 7351 7351 7352 7352 7353 7353 7354 7354 7355 7355 7356 7356 7357 7357 7358 7358 7359 7359 7360 7360 7361 7361 7362 7362 7363 7363 7364 7364 7365 7365 7366 7366 7367 7367 7368 7368 7369 7369 7370 7370 7371 7371 7372 7372 7373 7373 7374 7374 7375 7375 7376 7376 7377 7377 7378 7378 7379 7379 7380 7380 7381 7381 7382 7382 7383 7383 7384 7384 7385 7385 7386 7386 7387 7387 7388 7388 7389 7389 7390 7390 7391 7391 7392 7392 7393 7393 7394 7394 7395 7395 7396 7396 7397 7397 7398 7398 7399 7399 7400 7400 7401 7401 7402 7402 7403 7403 7404 7404 7405 7405 7406 7406 7407 7407 7408 7408 7409 7409 7410 7410 7411 7411 7412 7412 7413 7413 7414 7414 7415 7415 7416 7416 7417 7417 7418 7418 7419 6910 7420 7420 7421 7421 7422 7422 7423 7423 7424 7424 7425 7425 7426 7426 7427 7427 7428 7428 7429 7429 7430 7430 7431 7431 7432 7432 7433 7433 7434 7434 7435 7435 7436 7436 7437 7437 7438 7438 7439 7439 7440 7440 7441 7441 7442 7442 7443 7443 7444 7444 7445 7445 7446 7446 7447 7447 7448 7448 7449 7449 7450 7450 7451 7451 7452 7452 7453 7453 7454 7454 7455 7455 7456 7456 7457 7457 7458 7458 7459 7459 7460 7460 7461 7461 7462 7462 7463 7463 7464 7464 7465 7465 7466 7466 7467 7467 7468 7468 7469 7469 7470 7470 7471 7471 7472 7472 7473 7473 7474 7474 7475 7475 7476 7476 7477 7477 7478 7478 7479 7479 7480 7480 7481 7481 7482 7482 7483 7483 7484 7484 7485 7485 7486 7486 7487 7487 7488 7488 7489 7489 7490 7490 7491 7491 7492 7492 7493 7493 7494 7494 7495 7495 7496 7496 7497 7497 7498 7498 7499 7499 7500 7500 7501 7501 7502 13 7503 7503 7504 7504 7505 7505 7506 7506 7507 7507 7508 7508 7509 7509 7510 7510 7511 13 7512 7512 7513 7513 7514 6910 7515 7515 7516 7516 7517 7517 7518 7518 7519 7519 7520 7520 7521 7521 7522 7522 7523 7523 7524 7524 7525 7525 7526 7526 7527 7527 7528 7528 7529 7529 7530 7530 7531 7531 7532 7532 7533 7533 7534 7534 7535 7535 7536 7536 7537 7537 7538 7538 7539 7539 7540 7540 7541 7541 7542 7542 7543 7543 7544 7544 7545 7545 7546 7546 7547 7547 7548 7548 7549 7549 7550 7550 7551 7551 7552 7552 7553 7553 7554 7554 7555 7555 7556 7556 7557 7557 7558 7558 7559 7559 7560 7560 7561 7561 7562 7562 7563 7563 7564 7564 7565 7565 7566 7566 7567 7567 7568 7568 7569 7569 7570 7570 7571 7571 7572 7572 7573 7573 7574 7574 7575 7575 7576 7576 7577 7577 7578 7578 7579 7579 7580 7580 7581 7581 7582 7582 7583 7583 7584 7584 7585 7585 7586 7586 7587 7587 7588 7588 7589 7589 7590 7590 7591 7591 7592 7592 7593 7593 7594 7594 7595 7595 7596 7596 7597 2 7598 7315 7599 7599 7600 7600 7601 5836 7602 434 7603 7603 7604 7604 7605 7605 7606 7606 7607 7607 7608 7608 7609 7609 7610 7610 7611 7611 7612 13 7613 7613 7614 7614 7615 7615 7616 7616 7617 7617 7618 7618 7619 7619 7620 7620 7621 7621 7622 976 7623 159 7624 162 7625 7625 7626 7626 7627 7627 7628 5836 7629 434 7630 7630 7631 282 7632 7632 7633 2 7634 282 7635 77 7636 3931 7637 2828 7638 7638 7639 1862 7640 162 7641 7641 7642 2 7643 159 7644 162 7645 7645 7646 7646 7647 7647 7648 7648 7649 7649 7650 943 7651 7651 7652 7652 7653 7653 7654 976 7655 159 7656 162 7657 7657 7658 7658 7659 2 7660 282 7661 7661 7662 2 7663 282 7664 7664 7665 7665 7666 7666 7667 7667 7668 7668 7669 7669 7670 1862 7671 162 7672 7672 7673 7673 7674 7674 7675 7675 7676 282 7677 7677 7678 7678 7679 2 7680 574 7681 7681 7682 5236 7683 7683 7684 7684 7685 855 7686 7686 7687 11 7688 7688 7689 2 7690 574 7691 7691 7692 7692 7693 7693 7694 2732 7695 7695 7696 7696 7697 2050 7698 2 7699 7699 7700 357 7701 357 7702 360 7703 434 7704 7704 7705 7705 7706 7706 7707 7707 7708 2 7709 574 7710 7710 7711 7711 7712 7712 7713 7713 7714 7714 7715 2803 7716 7716 7717 434 7718 7718 7719 2 7720 7720 7721 7721 7722 1272 7723 2 7724 7724 7725 7725 7726 7726 7727 7727 7728 7728 7729 2894 7730 7730 7731 69 7732 7732 7733 7733 7734 7734 7735 7735 7736 7736 7737 7737 7738 7738 7739 2 7740 7740 7741 7741 7742 7742 7743 7743 7744 7744 7745 7745 7746 2936 7747 2938 7748 7748 7749 7749 7750 7750 7751 7751 7752 7752 7753 357 7754 357 7755 40 7756 360 7757 434 7758 7758 7759 7759 7760 7760 7761 7761 7762 7762 7763 40 7764 7764 7765 7765 7766 2959 7767 2962 7768 2964 7769 2966 7770 77 7771 2968 7772 2969 7773 7773 7774 282 7775 7775 7776 2 7777 282 7778 77 7779 3931 7780 2828 7781 7781 7782 1862 7783 162 7784 7784 7785 2 7786 159 7787 162 7788 7788 7789 7789 7790 7790 7791 7791 7792 7792 7793 943 7794 7794 7795 7795 7796 7796 7797 976 7798 159 7799 162 7800 7800 7801 7801 7802 2 7803 282 7804 7804 7805 2 7806 282 7807 7807 7808 7808 7809 7809 7810 7810 7811 7811 7812 7812 7813 7813 7814 1862 7815 162 7816 7816 7817 7817 7818 7818 7819 7819 7820 282 7821 7821 7822 7822 7823 2 7824 574 7825 7825 7826 5236 7827 7827 7828 7828 7829 855 7830 7830 7831 7831 7832 7832 7833 7833 7834 7834 7835 7835 7836 7836 7837 7837 7838 7838 7839 7839 7840 7840 7841 7841 7842 7842 7843 7843 7844 7844 7845 7845 7846 7846 7847 7847 7848 7848 7849 7849 7850 7850 7851 7851 7852 7852 7853 7853 7854 7854 7855 7855 7856 7856 7857 7857 7858 7858 7859 7859 7860 7860 7861 7861 7862 7862 7863 7863 7864 7864 7865 7865 7866 7866 7867 7867 7868 7868 7869 7869 7870 7870 7871 7871 7872 7872 7873 7873 7874 7874 7875 7875 7876 7876 7877 7877 7878 7878 7879 7879 7880 7880 7881 7881 7882 7882 7883 7883 7884 7884 7885 7885 7886 7886 7887 7887 7888 7888 7889 7889 7890 7890 7891 7891 7892 7892 7893 7893 7894 7894 7895 7895 7896 7896 7897 7897 7898 7898 7899 7899 7900 7900 7901 7901 7902 7902 7903 7903 7904 7904 7905 7905 7906 7906 7907 7907 7908 7908 7909 7909 7910 11 7911 7911 7912 7912 7913 7913 7914 2 7915 574 7916 7916 7917 943 7918 7918 7919 7919 7920 976 7921 159 7922 162 7923 7923 7924 7924 7925 1487 7926 2 7927 2 7928 2 7929 7929 7930 2 7931 7931 7932 2 7933 2 7934 2 7935 2 7936 2 7937 2 7938 2 7939 2 7940 2 7941 7941 7942 7942 7943 7943 7944 7944 7945 7945 7946 7946 7947 7947 7948 7948 7949 7949 7950 7950 7951 7951 7952 7952 7953 7953 7954 22 7955 7955 7956 4 7957 11 7958 7958 7959 89 7960 11 7961 7961 7962 99 7963 7963 7964 1551 7965 7965 7966 7966 7967 7967 7968 99 7969 855 7970 11 7971 7971 7972 943 7973 7973 7974 7974 7975 976 7976 159 7977 162 7978 7978 7979 7979 7980 7980 7981 855 7982 11 7983 7983 7984 7984 7985 7985 7986 357 7987 357 7988 360 7989 434 7990 7990 7991 7991 7992 7992 7993 7993 7994 2 7995 574 7996 7996 7997 7997 7998 7998 7999 7999 8000 8000 8001 8001 8002 2 8003 574 8004 8004 8005 85 8006 8006 8007 8007 8008 140 8009 11 8010 142 8011 162 8012 163 8013 165 8014 167 8015 77 8016 126 8017 170 8018 8018 8019 8019 8020 140 8021 11 8022 142 8023 162 8024 2248 8025 2250 8026 2251 8027 2252 8028 2 8029 2 8030 3 8031 4 8032 2 8033 2 8034 2 8035 2 8036 11 8037 2 8038 2 8039 1687 8040 2 8041 1689 8042 22 8043 1691 8044 1692 8045 2 8046 11 8047 11 8048 1697 8049 1698 8050 1699 8051 22 8052 2 8053 2 8054 2 8055 11 8056 1705 8057 11 8058 1707 8059 11 8060 1709 8061 1710 8062 1711 8063 1712 8064 60 8065 1714 8066 2 8067 2 8068 1717 8069 11 8070 2 8071 2 8072 69 8073 131 8074 1723 8075 2 8076 2 8077 1726 8078 2 8079 1728 8080 1729 8081 2 8082 85 8083 86 8084 89 8085 134 8086 11 8087 1739 8088 1740 8089 99 8090 1742 8091 1743 8092 8092 8093 104 8094 104 8095 1747 8096 1748 8097 1749 8098 1750 8099 1751 8100 1752 8101 1753 8102 1756 8103 1757 8104 1758 8105 1759 8106 99 8107 2 8108 1763 8109 2 8110 2 8111 2 8112 2 8113 2 8114 2 8115 2 8116 2 8117 307 8118 2 8119 2 8120 2 8121 2 8122 2 8123 313 8124 2 8125 2 8126 2 8127 2 8128 2 8129 2 8130 2 8131 2 8132 2 8133 2 8134 2 8135 2 8136 2 8137 2 8138 2 8139 1794 8140 1795 8141 1796 8142 1797 8143 1798 8144 1799 8145 1800 8146 1801 8147 1802 8148 1803 8149 1804 8150 1805 8151 1806 8152 1807 8153 1808 8154 4 8155 2 8156 1811 8157 1812 8158 1813 8159 1814 8160 1815 8161 1816 8162 1817 8163 1818 8164 1819 8165 1820 8166 1821 8167 357 8168 357 8169 40 8170 360 8171 434 8172 1827 8173 1828 8174 1829 8175 1830 8176 1831 8177 2 8178 574 8179 1834 8180 1835 8181 40 8182 1837 8183 282 8184 1839 8185 1840 8186 2 8187 574 8188 1843 8189 1844 8190 1845 8191 1846 8192 1847 8193 1848 8194 2 8195 740 8196 2 8197 1852 8198 1853 8199 1854 8200 1855 8201 1856 8202 1857 8203 1858 8204 1859 8205 1860 8206 1861 8207 855 8208 7830 8209 7831 8210 7832 8211 7833 8212 7834 8213 7835 8214 7836 8215 7837 8216 7838 8217 7839 8218 7840 8219 7841 8220 7842 8221 7843 8222 7844 8223 7845 8224 7846 8225 7847 8226 7848 8227 7849 8228 7850 8229 7851 8230 7852 8231 7853 8232 7854 8233 7855 8234 7856 8235 7857 8236 7858 8237 7859 8238 7860 8239 7861 8240 7862 8241 7863 8242 7864 8243 7865 8244 7866 8245 7867 8246 7868 8247 7869 8248 7870 8249 7871 8250 7872 8251 7873 8252 7874 8253 7875 8254 7876 8255 7877 8256 7878 8257 7879 8258 7880 8259 7881 8260 7882 8261 7883 8262 7884 8263 7885 8264 7886 8265 7887 8266 7888 8267 7889 8268 7890 8269 7891 8270 7892 8271 7893 8272 7894 8273 7895 8274 7896 8275 7897 8276 7898 8277 7899 8278 7900 8279 7901 8280 7902 8281 7903 8282 7904 8283 7905 8284 7906 8285 7907 8286 7908 8287 7909 8288 11 8289 7911 8290 7912 8291 7913 8292 2 8293 574 8294 7916 8295 943 8296 7918 8297 7919 8298 976 8299 159 8300 162 8301 7923 8302 7924 8303 1487 8304 2 8305 2 8306 2 8307 7929 8308 2 8309 7931 8310 2 8311 2 8312 2 8313 2 8314 2 8315 2 8316 2 8317 2 8318 2 8319 7941 8320 7942 8321 7943 8322 7944 8323 7945 8324 7946 8325 7947 8326 7948 8327 7949 8328 7950 8329 7951 8330 7952 8331 7953 8332 22 8333 7955 8334 4 8335 11 8336 7958 8337 89 8338 11 8339 7961 8340 99 8341 7963 8342 1551 8343 7965 8344 7966 8345 7967 8346 99 8347 855 8348 11 8349 7971 8350 943 8351 7973 8352 7974 8353 976 8354 159 8355 162 8356 7978 8357 7979 8358 7980 8359 855 8360 11 8361 7983 8362 7984 8363 7985 8364 357 8365 357 8366 360 8367 434 8368 7990 8369 7991 8370 7992 8371 7993 8372 2 8373 574 8374 7996 8375 7997 8376 7998 8377 7999 8378 8000 8379 8001 8380 2 8381 574 8382 8004 8383 8007 8384 140 8385 11 8386 142 8387 162 8388 163 8389 165 8390 167 8391 77 8392 126 8393 170 8394 8018 8395 8019 8396 140 8397 11 8398 142 8399 162 8400 2248 8401 2250 8402 2251 8403 2252 8404 2 8405 8092 8406 8406 8407 8406 8408 855 8409 8409 8410 11 8411 8411 8412 2 8413 574 8414 8414 8415 8415 8416 8416 8417 2732 8418 8418 8419 8419 8420 357 8421 357 8422 360 8423 434 8424 8424 8425 8425 8426 8426 8427 8427 8428 2 8429 574 8430 8430 8431 8431 8432 8432 8433 8433 8434 2803 8435 8435 8436 434 8437 8437 8438 2 8439 8439 8440 8440 8441 8441 8442 2 8443 8443 8444 8444 8445 8445 8446 8446 8447 8447 8448 2894 8449 8449 8450 69 8451 8451 8452 8452 8453 8453 8454 8454 8455 8455 8456 8456 8457 8457 8458 2 8459 8459 8460 8460 8461 8461 8462 8462 8463 8463 8464 8464 8465 2936 8466 2938 8467 8467 8468 8468 8469 8469 8470 8470 8471 8471 8472 357 8473 357 8474 40 8475 360 8476 434 8477 8477 8478 8478 8479 8479 8480 8480 8481 8481 8482 40 8483 8483 8484 8484 8485 2959 8486 2962 8487 2964 8488 2966 8489 77 8490 2968 8491 2969 8492 8492 8493 282 8494 8494 8495 2 8496 282 8497 77 8498 3931 8499 2828 8500 8500 8501 1862 8502 162 8503 2050 8504 2 8505 7699 8506 2 8507 159 8508 162 8509 8509 8510 8510 8511 8511 8512 8512 8513 8513 8514 943 8515 8515 8516 8516 8517 8517 8518 976 8519 159 8520 162 8521 8521 8522 8522 8523 2 8524 282 8525 8525 8526 2 8527 282 8528 8528 8529 8529 8530 8530 8531 8531 8532 8532 8533 8533 8534 1862 8535 162 8536 8536 8537 8537 8538 8538 8539 8539 8540 282 8541 8541 8542 8542 8543 2 8544 574 8545 8545 8546 5236 8547 8547 8548 8548 8549 1487 8550 2 8551 3 8552 4 8553 2 8554 2 8555 2 8556 2 8557 11 8558 2 8559 2 8560 8560 8561 2 8562 8562 8563 22 8564 8564 8565 8565 8566 2 8567 11 8568 11 8569 8569 8570 8570 8571 8571 8572 22 8573 2 8574 2 8575 2 8576 11 8577 8577 8578 11 8579 8579 8580 11 8581 8581 8582 8582 8583 8583 8584 8584 8585 60 8586 8586 8587 2 8588 2 8589 8589 8590 11 8591 2 8592 2 8593 69 8594 131 8595 8595 8596 2 8597 2 8598 8598 8599 2 8600 8600 8601 8601 8602 2 8603 85 8604 86 8605 22 8606 89 8607 134 8608 11 8609 8609 8610 8610 8611 99 8612 8612 8613 8613 8614 8614 8615 104 8616 104 8617 8617 8618 8618 8619 8619 8620 8620 8621 8621 8622 8622 8623 8623 8624 8624 8625 8625 8626 8626 8627 8627 8628 99 8629 2 8630 11 8631 8631 8632 2 8633 2 8634 2 8635 2 8636 2 8637 2 8638 2 8639 2 8640 307 8641 2 8642 2 8643 2 8644 2 8645 2 8646 313 8647 2 8648 2 8649 2 8650 2 8651 2 8652 2 8653 2 8654 2 8655 2 8656 2 8657 2 8658 2 8659 2 8660 2 8661 2 8662 8662 8663 8663 8664 8664 8665 8665 8666 8666 8667 8667 8668 8668 8669 8669 8670 8670 8671 8671 8672 8672 8673 8673 8674 8674 8675 8675 8676 8676 8677 4 8678 2 8679 8679 8680 8680 8681 8681 8682 8682 8683 8683 8684 8684 8685 8685 8686 8686 8687 8687 8688 8688 8689 8689 8690 357 8691 357 8692 40 8693 360 8694 434 8695 8695 8696 8696 8697 8697 8698 8698 8699 8699 8700 2 8701 574 8702 8702 8703 8703 8704 40 8705 8705 8706 282 8707 8707 8708 8708 8709 2 8710 574 8711 8711 8712 8712 8713 8713 8714 8714 8715 8715 8716 8716 8717 2 8718 740 8719 2 8720 8720 8721 8721 8722 8722 8723 8723 8724 8724 8725 8725 8726 8726 8727 8727 8728 8728 8729 8729 8730 1862 8731 162 8732 2050 8733 2 8734 8734 8735 855 8736 8736 8737 8737 8738 8738 8739 8739 8740 8740 8741 8741 8742 8742 8743 8743 8744 8744 8745 8745 8746 8746 8747 8747 8748 8748 8749 8749 8750 8750 8751 8751 8752 8752 8753 8753 8754 8754 8755 8755 8756 8756 8757 8757 8758 8758 8759 8759 8760 8760 8761 8761 8762 8762 8763 8763 8764 8764 8765 8765 8766 8766 8767 8767 8768 8768 8769 8769 8770 8770 8771 8771 8772 8772 8773 8773 8774 8774 8775 8775 8776 8776 8777 8777 8778 8778 8779 8779 8780 8780 8781 8781 8782 8782 8783 8783 8784 8784 8785 8785 8786 8786 8787 8787 8788 8788 8789 8789 8790 8790 8791 8791 8792 8792 8793 8793 8794 8794 8795 8795 8796 8796 8797 8797 8798 8798 8799 8799 8800 8800 8801 8801 8802 8802 8803 8803 8804 8804 8805 8805 8806 8806 8807 8807 8808 8808 8809 8809 8810 8810 8811 8811 8812 8812 8813 8813 8814 8814 8815 8815 8816 11 8817 8817 8818 8818 8819 8819 8820 2 8821 574 8822 8822 8823 943 8824 8824 8825 8825 8826 976 8827 159 8828 162 8829 8829 8830 8830 8831 1487 8832 2 8833 2 8834 2 8835 8835 8836 2 8837 8837 8838 2 8839 2 8840 2 8841 2 8842 2 8843 2 8844 2 8845 2 8846 2 8847 8847 8848 8848 8849 8849 8850 8850 8851 8851 8852 8852 8853 8853 8854 8854 8855 8855 8856 8856 8857 8857 8858 8858 8859 8859 8860 22 8861 8861 8862 4 8863 11 8864 8864 8865 89 8866 11 8867 8867 8868 99 8869 8869 8870 1551 8871 8871 8872 8872 8873 8873 8874 99 8875 855 8876 11 8877 8877 8878 943 8879 8879 8880 8880 8881 976 8882 159 8883 162 8884 8884 8885 8885 8886 8886 8887 855 8888 11 8889 8889 8890 8890 8891 8891 8892 357 8893 357 8894 360 8895 434 8896 8896 8897 8897 8898 8898 8899 8899 8900 2 8901 574 8902 8902 8903 8903 8904 8904 8905 8905 8906 8906 8907 8907 8908 2 8909 574 8910 8910 8911 8911 8912 140 8913 11 8914 142 8915 162 8916 163 8917 165 8918 167 8919 77 8920 126 8921 170 8922 8922 8923 8923 8924 140 8925 11 8926 142 8927 162 8928 2248 8929 2250 8930 2251 8931 2252 8932 2 8933 2 8934 3 8935 4 8936 2 8937 2 8938 2 8939 2 8940 11 8941 2 8942 2 8943 8560 8944 2 8945 8562 8946 22 8947 8564 8948 8565 8949 2 8950 11 8951 11 8952 8569 8953 8570 8954 8571 8955 22 8956 2 8957 2 8958 2 8959 11 8960 8577 8961 11 8962 8579 8963 11 8964 8581 8965 8582 8966 8583 8967 8584 8968 60 8969 8586 8970 2 8971 2 8972 8589 8973 11 8974 2 8975 2 8976 69 8977 131 8978 8595 8979 2 8980 2 8981 8598 8982 2 8983 8600 8984 8601 8985 2 8986 85 8987 86 8988 89 8989 134 8990 11 8991 8609 8992 8610 8993 99 8994 8612 8995 8613 8996 8996 8997 104 8998 104 8999 8617 9000 8618 9001 8619 9002 8620 9003 8621 9004 8622 9005 8623 9006 8624 9007 8625 9008 8626 9009 8627 9010 99 9011 2 9012 8631 9013 2 9014 2 9015 2 9016 2 9017 2 9018 2 9019 2 9020 2 9021 307 9022 2 9023 2 9024 2 9025 2 9026 2 9027 313 9028 2 9029 2 9030 2 9031 2 9032 2 9033 2 9034 2 9035 2 9036 2 9037 2 9038 2 9039 2 9040 2 9041 2 9042 2 9043 8662 9044 8663 9045 8664 9046 8665 9047 8666 9048 8667 9049 8668 9050 8669 9051 8670 9052 8671 9053 8672 9054 8673 9055 8674 9056 8675 9057 8676 9058 4 9059 2 9060 8679 9061 8680 9062 8681 9063 8682 9064 8683 9065 8684 9066 8685 9067 8686 9068 8687 9069 8688 9070 8689 9071 357 9072 357 9073 40 9074 360 9075 434 9076 8695 9077 8696 9078 8697 9079 8698 9080 8699 9081 2 9082 574 9083 8702 9084 8703 9085 40 9086 8705 9087 282 9088 8707 9089 8708 9090 2 9091 574 9092 8711 9093 8712 9094 8713 9095 8714 9096 8715 9097 8716 9098 2 9099 740 9100 2 9101 8720 9102 8721 9103 8722 9104 8723 9105 8724 9106 8725 9107 8726 9108 8727 9109 8728 9110 8729 9111 855 9112 8736 9113 8737 9114 8738 9115 8739 9116 8740 9117 8741 9118 8742 9119 8743 9120 8744 9121 8745 9122 8746 9123 8747 9124 8748 9125 8749 9126 8750 9127 8751 9128 8752 9129 8753 9130 8754 9131 8755 9132 8756 9133 8757 9134 8758 9135 8759 9136 8760 9137 8761 9138 8762 9139 8763 9140 8764 9141 8765 9142 8766 9143 8767 9144 8768 9145 8769 9146 8770 9147 8771 9148 8772 9149 8773 9150 8774 9151 8775 9152 8776 9153 8777 9154 8778 9155 8779 9156 8780 9157 8781 9158 8782 9159 8783 9160 8784 9161 8785 9162 8786 9163 8787 9164 8788 9165 8789 9166 8790 9167 8791 9168 8792 9169 8793 9170 8794 9171 8795 9172 8796 9173 8797 9174 8798 9175 8799 9176 8800 9177 8801 9178 8802 9179 8803 9180 8804 9181 8805 9182 8806 9183 8807 9184 8808 9185 8809 9186 8810 9187 8811 9188 8812 9189 8813 9190 8814 9191 8815 9192 11 9193 8817 9194 8818 9195 8819 9196 2 9197 574 9198 8822 9199 943 9200 8824 9201 8825 9202 976 9203 159 9204 162 9205 8829 9206 8830 9207 1487 9208 2 9209 2 9210 2 9211 8835 9212 2 9213 8837 9214 2 9215 2 9216 2 9217 2 9218 2 9219 2 9220 2 9221 2 9222 2 9223 8847 9224 8848 9225 8849 9226 8850 9227 8851 9228 8852 9229 8853 9230 8854 9231 8855 9232 8856 9233 8857 9234 8858 9235 8859 9236 22 9237 8861 9238 4 9239 11 9240 8864 9241 89 9242 11 9243 8867 9244 99 9245 8869 9246 1551 9247 8871 9248 8872 9249 8873 9250 99 9251 855 9252 11 9253 8877 9254 943 9255 8879 9256 8880 9257 976 9258 159 9259 162 9260 8884 9261 8885 9262 8886 9263 855 9264 11 9265 8889 9266 8890 9267 8891 9268 357 9269 357 9270 360 9271 434 9272 8896 9273 8897 9274 8898 9275 8899 9276 2 9277 574 9278 8902 9279 8903 9280 8904 9281 8905 9282 8906 9283 8907 9284 2 9285 574 9286 8910 9287 8911 9288 140 9289 11 9290 142 9291 162 9292 163 9293 165 9294 167 9295 77 9296 126 9297 170 9298 8922 9299 8923 9300 140 9301 11 9302 142 9303 162 9304 2248 9305 2250 9306 2251 9307 2252 9308 2 9309 8996 9310 9310 9311 9310 9312 855 9313 9313 9314 11 9315 9315 9316 2 9317 574 9318 9318 9319 9319 9320 9320 9321 2732 9322 9322 9323 9323 9324 357 9325 357 9326 360 9327 434 9328 9328 9329 9329 9330 9330 9331 9331 9332 2 9333 574 9334 9334 9335 9335 9336 9336 9337 9337 9338 2803 9339 9339 9340 434 9341 9341 9342 2 9343 9343 9344 9344 9345 8441 9346 2 9347 9347 9348 9348 9349 9349 9350 9350 9351 9351 9352 2894 9353 9353 9354 69 9355 9355 9356 9356 9357 9357 9358 9358 9359 9359 9360 9360 9361 9361 9362 2 9363 9363 9364 9364 9365 9365 9366 9366 9367 9367 9368 9368 9369 2936 9370 2938 9371 9371 9372 9372 9373 9373 9374 9374 9375 9375 9376 357 9377 357 9378 40 9379 360 9380 434 9381 9381 9382 9382 9383 9383 9384 9384 9385 9385 9386 40 9387 9387 9388 9388 9389 2959 9390 2962 9391 2964 9392 2966 9393 77 9394 2968 9395 2969 9396 9396 9397 282 9398 9398 9399 2 9400 282 9401 77 9402 3931 9403 2828 9404 9404 9405 1862 9406 162 9407 2050 9408 2 9409 8734 9410 2 9411 159 9412 162 9413 9413 9414 9414 9415 9415 9416 9416 9417 9417 9418 943 9419 9419 9420 9420 9421 9421 9422 976 9423 159 9424 162 9425 9425 9426 9426 9427 2 9428 282 9429 9429 9430 2 9431 282 9432 9432 9433 9433 9434 9434 9435 9435 9436 9436 9437 9437 9438 1862 9439 162 9440 9440 9441 9441 9442 9442 9443 9443 9444 282 9445 9445 9446 9446 9447 2 9448 574 9449 9449 9450 5236 9451 9451 9452 9452 9453 855 9454 9313 9455 11 9456 9315 9457 2 9458 574 9459 9318 9460 9319 9461 9320 9462 2732 9463 9322 9464 9323 9465 357 9466 357 9467 360 9468 434 9469 9328 9470 9329 9471 9330 9472 9331 9473 2 9474 574 9475 9334 9476 9335 9477 9336 9478 9337 9479 4 9480 11 9481 9481 9482 89 9483 11 9484 9484 9485 99 9486 9486 9487 1551 9488 9488 9489 9489 9490 9490 9491 99 9492 855 9493 11 9494 9494 9495 943 9496 9496 9497 9497 9498 976 9499 159 9500 162 9501 9501 9502 9502 9503 9503 9504 855 9505 11 9506 9506 9507 9507 9508 9508 9509 357 9510 357 9511 360 9512 434 9513 9513 9514 9514 9515 9515 9516 9516 9517 2 9518 574 9519 9519 9520 9520 9521 9521 9522 9522 9523 9523 9524 9524 9525 2 9526 574 9527 9527 9528 2803 9529 9339 9530 434 9531 9341 9532 2 9533 9343 9534 9344 9535 8441 9536 2 9537 9347 9538 9348 9539 9349 9540 9350 9541 9351 9542 2894 9543 9353 9544 69 9545 9355 9546 9356 9547 9357 9548 9358 9549 9359 9550 9360 9551 9361 9552 2 9553 9363 9554 9364 9555 9365 9556 9366 9557 9367 9558 9368 9559 2936 9560 2938 9561 9371 9562 9372 9563 9373 9564 9374 9565 9375 9566 357 9567 357 9568 40 9569 360 9570 434 9571 9381 9572 9382 9573 9383 9574 9384 9575 9385 9576 40 9577 9387 9578 9388 9579 2959 9580 2962 9581 2964 9582 2966 9583 77 9584 2968 9585 2969 9586 9396 9587 282 9588 9398 9589 2 9590 282 9591 77 9592 3931 9593 2828 9594 9404 9595 1862 9596 162 9597 9597 9598 2 9599 159 9600 162 9601 9413 9602 9414 9603 9415 9604 9416 9605 9417 9606 943 9607 9419 9608 9420 9609 9421 9610 976 9611 159 9612 162 9613 9425 9614 9426 9615 2 9616 282 9617 9429 9618 2 9619 282 9620 9432 9621 9433 9622 9434 9623 9435 9624 9436 9625 9437 9626 1862 9627 162 9628 9440 9629 9441 9630 9442 9631 9443 9632 282 9633 9445 9634 9446 9635 2 9636 574 9637 9449 9638 5236 9639 9451 9640 9452 9641 855 9642 8409 9643 11 9644 8411 9645 2 9646 574 9647 8414 9648 8415 9649 8416 9650 2732 9651 8418 9652 8419 9653 9653 9654 357 9655 357 9656 360 9657 434 9658 8424 9659 8425 9660 8426 9661 8427 9662 2 9663 574 9664 8430 9665 8431 9666 8432 9667 8433 9668 9668 9669 2803 9670 8435 9671 434 9672 8437 9673 2 9674 8439 9675 8440 9676 8441 9677 2 9678 8443 9679 8444 9680 8445 9681 8446 9682 8447 9683 2894 9684 8449 9685 69 9686 8451 9687 8452 9688 8453 9689 8454 9690 8455 9691 8456 9692 8457 9693 2 9694 8459 9695 8460 9696 8461 9697 8462 9698 8463 9699 8464 9700 2936 9701 2938 9702 8467 9703 8468 9704 8469 9705 8470 9706 8471 9707 357 9708 357 9709 40 9710 360 9711 434 9712 8477 9713 8478 9714 8479 9715 8480 9716 8481 9717 40 9718 8483 9719 8484 9720 2959 9721 2962 9722 2964 9723 2966 9724 77 9725 2968 9726 2969 9727 8492 9728 282 9729 8494 9730 2 9731 282 9732 77 9733 3931 9734 2828 9735 8500 9736 1862 9737 162 9738 9738 9739 2 9740 159 9741 162 9742 8509 9743 8510 9744 8511 9745 8512 9746 8513 9747 943 9748 8515 9749 8516 9750 8517 9751 976 9752 159 9753 162 9754 8521 9755 8522 9756 2 9757 282 9758 8525 9759 2 9760 282 9761 8528 9762 9762 9763 8529 9764 8530 9765 8531 9766 8532 9767 8533 9768 1862 9769 162 9770 8536 9771 8537 9772 8538 9773 8539 9774 282 9775 8541 9776 8542 9777 2 9778 574 9779 8545 9780 5236 9781 8547 9782 8548 9783 2224 9784 4 9785 2 9786 9786 9787 9787 9788 9788 9789 9789 9790 9790 9791 9791 9792 9792 9793 2936 9794 2938 9795 9795 9796 11 9797 9797 9798 9798 9799 9799 9800 357 9801 357 9802 40 9803 360 9804 434 9805 9805 9806 9806 9807 9807 9808 9808 9809 9809 9810 2 9811 574 9812 9812 9813 9813 9814 943 9815 9815 9816 9816 9817 976 9818 159 9819 162 9820 9820 9821 9821 9822 40 9823 9823 9824 9824 9825 2959 9826 2962 9827 2964 9828 2966 9829 77 9830 2968 9831 2969 9832 9832 9833 282 9834 9834 9835 9835 9836 2 9837 574 9838 9838 9839 1862 9840 162 9841 9841 9842 282 9843 9843 9844 2225 9845 2 9846 574 9847 2228 9848 5236 9849 9849 9850 2231 9851 140 9852 11 9853 142 9854 162 9855 163 9856 165 9857 167 9858 77 9859 126 9860 170 9861 2242 9862 2243 9863 140 9864 11 9865 142 9866 162 9867 2248 9868 2250 9869 2251 9870 2252 9871 2 9872 2254 9873 2255 9874 2 9875 2887 9876 855 9877 7686 9878 11 9879 7688 9880 2 9881 574 9882 7691 9883 7692 9884 7693 9885 2732 9886 7695 9887 7696 9888 357 9889 357 9890 360 9891 434 9892 7704 9893 7705 9894 7706 9895 7707 9896 2 9897 574 9898 7710 9899 7711 9900 7712 9901 7713 9902 2803 9903 7716 9904 434 9905 7718 9906 2 9907 7720 9908 7721 9909 1272 9910 2 9911 7724 9912 7725 9913 7726 9914 7727 9915 7728 9916 2894 9917 7730 9918 69 9919 7732 9920 7733 9921 7734 9922 7735 9923 7736 9924 7737 9925 7738 9926 2 9927 7740 9928 7741 9929 7742 9930 7743 9931 7744 9932 7745 9933 2936 9934 2938 9935 7748 9936 7749 9937 7750 9938 7751 9939 7752 9940 357 9941 357 9942 40 9943 360 9944 434 9945 7758 9946 7759 9947 7760 9948 7761 9949 7762 9950 40 9951 7764 9952 7765 9953 2959 9954 2962 9955 2964 9956 2966 9957 77 9958 2968 9959 2969 9960 7773 9961 282 9962 7775 9963 2 9964 282 9965 77 9966 3931 9967 2828 9968 7781 9969 1862 9970 162 9971 2050 9972 2 9973 2052 9974 2 9975 159 9976 162 9977 7788 9978 7789 9979 7790 9980 7791 9981 7792 9982 943 9983 7794 9984 7795 9985 7796 9986 976 9987 159 9988 162 9989 7800 9990 7801 9991 2 9992 282 9993 7804 9994 2 9995 282 9996 7807 9997 7809 9998 7810 9999 7811 10000 7812 10001 7813 10002 1862 10003 162 10004 7816 10005 7817 10006 7818 10007 7819 10008 282 10009 7821 10010 7822 10011 2 10012 574 10013 7825 10014 5236 10015 7827 10016 7828 10017 2889 10018 2891 10019 2921 10020 2922 10021 943 10022 2923 10023 2613 10024 976 10025 159 10026 162 10027 2617 10028 2618 10029 2924 10030 4 10031 2 10032 2928 10033 2929 10034 2930 10035 2932 10036 2933 10037 2652 10038 2934 10039 2936 10040 2938 10041 2940 10042 2941 10043 2942 10044 2943 10045 2944 10046 357 10047 357 10048 40 10049 360 10050 434 10051 2950 10052 2951 10053 2685 10054 2952 10055 2687 10056 2 10057 574 10058 2690 10059 2953 10060 943 10061 2954 10062 2668 10063 976 10064 159 10065 162 10066 2672 10067 2673 10068 40 10069 2956 10070 2957 10071 2959 10072 2962 10073 2964 10074 2966 10075 77 10076 2968 10077 2969 10078 2994 10079 282 10080 7632 10081 2695 10082 2 10083 574 10084 2698 10085 2 10086 282 10087 77 10088 3931 10089 2828 10090 7638 10091 1862 10092 162 10093 2050 10094 2 10095 2522 10096 2 10097 159 10098 162 10099 7645 10100 7647 10101 7648 10102 943 10103 7651 10104 7652 10105 7653 10106 976 10107 159 10108 162 10109 7657 10110 7658 10111 2 10112 282 10113 7661 10114 2 10115 282 10116 7664 10117 7667 10118 1862 10119 162 10120 7672 10121 855 10122 2524 10123 2525 10124 2526 10125 2527 10126 2528 10127 2529 10128 2530 10129 2531 10130 2532 10131 2533 10132 2534 10133 2535 10134 2536 10135 2537 10136 2538 10137 2539 10138 2540 10139 2541 10140 2542 10141 2543 10142 2544 10143 2545 10144 2546 10145 2547 10146 2548 10147 2549 10148 2550 10149 2551 10150 2552 10151 2553 10152 2554 10153 2555 10154 2556 10155 2557 10156 2558 10157 2559 10158 2560 10159 2561 10160 2562 10161 2563 10162 2564 10163 2565 10164 2566 10165 2567 10166 2568 10167 2569 10168 2570 10169 2571 10170 2572 10171 2573 10172 2574 10173 2575 10174 2576 10175 2577 10176 2578 10177 2579 10178 2580 10179 2581 10180 2582 10181 2583 10182 2584 10183 2585 10184 2586 10185 2587 10186 2588 10187 2589 10188 2590 10189 2591 10190 2592 10191 2593 10192 2594 10193 2595 10194 2596 10195 2597 10196 2598 10197 2599 10198 2600 10199 2601 10200 2602 10201 2603 10202 11 10203 2605 10204 2606 10205 2607 10206 2 10207 574 10208 2610 10209 2612 10210 1487 10211 2 10212 2 10213 2 10214 2623 10215 2 10216 2625 10217 2 10218 2 10219 2 10220 2 10221 2 10222 2 10223 2 10224 2 10225 2 10226 2635 10227 2636 10228 2637 10229 2638 10230 2639 10231 2640 10232 2641 10233 2642 10234 2643 10235 2644 10236 2645 10237 2646 10238 2647 10239 22 10240 2649 10241 11 10242 89 10243 11 10244 2655 10245 99 10246 2657 10247 1551 10248 2659 10249 2660 10250 2661 10251 99 10252 855 10253 11 10254 2665 10255 2667 10256 2674 10257 855 10258 11 10259 2677 10260 2678 10261 2679 10262 357 10263 357 10264 360 10265 434 10266 2684 10267 2686 10268 2691 10269 2692 10270 2693 10271 2694 10272 162 10273 10273 10274 2699 10275 2 10276 10276 10277 140 10278 11 10279 142 10280 162 10281 163 10282 165 10283 167 10284 77 10285 126 10286 170 10287 2710 10288 2711 10289 140 10290 11 10291 142 10292 162 10293 2248 10294 2250 10295 2251 10296 2252 10297 2 10298 2721 10299 2722 10300 855 10301 2724 10302 11 10303 2726 10304 2 10305 574 10306 2729 10307 2730 10308 2731 10309 2732 10310 2734 10311 2735 10312 357 10313 357 10314 360 10315 434 10316 2740 10317 2741 10318 2742 10319 2743 10320 2 10321 574 10322 2746 10323 2747 10324 2748 10325 2749 10326 2803 10327 2808 10328 434 10329 2881 10330 2 10331 2883 10332 2884 10333 2888 10334 2890 10335 2894 10336 2895 10337 69 10338 2900 10339 2909 10340 2925 10341 7646 10342 7649 10343 10343 10344 7665 10345 7666 10346 7668 10347 7669 10348 7673 10349 7674 10350 7675 10351 282 10352 7677 10353 7678 10354 2 10355 574 10356 7681 10357 5236 10358 7683 10359 7684 10360 1487 10361 2 10362 2 10363 2 10364 10364 10365 2 10366 10366 10367 2 10368 2 10369 2 10370 2 10371 2 10372 2 10373 2 10374 2 10375 2 10376 10376 10377 10377 10378 10378 10379 10379 10380 10380 10381 10381 10382 10382 10383 10383 10384 10384 10385 10385 10386 10386 10387 10387 10388 10388 10389 22 10390 10390 10391 4 10392 11 10393 10393 10394 89 10395 11 10396 10396 10397 99 10398 10398 10399 1551 10400 10400 10401 10401 10402 10402 10403 99 10404 855 10405 11 10406 10406 10407 943 10408 10408 10409 10409 10410 976 10411 159 10412 162 10413 10413 10414 10414 10415 10415 10416 855 10417 11 10418 10418 10419 10419 10420 10420 10421 357 10422 357 10423 360 10424 434 10425 10425 10426 10426 10427 10427 10428 10428 10429 2 10430 574 10431 10431 10432 10432 10433 10433 10434 10434 10435 10435 10436 10436 10437 2 10438 574 10439 10439 10440 10440 10441 140 10442 11 10443 142 10444 162 10445 163 10446 165 10447 167 10448 77 10449 126 10450 170 10451 10451 10452 10452 10453 140 10454 11 10455 142 10456 162 10457 2248 10458 2250 10459 2251 10460 2252 10461 2 10462 2 10463 3 10464 4 10465 2 10466 2 10467 2 10468 2 10469 11 10470 2 10471 2 10472 687 10473 2 10474 689 10475 22 10476 691 10477 692 10478 2 10479 11 10480 11 10481 696 10482 697 10483 698 10484 22 10485 2 10486 2 10487 2 10488 11 10489 704 10490 11 10491 706 10492 11 10493 708 10494 709 10495 710 10496 711 10497 60 10498 713 10499 2 10500 2 10501 716 10502 11 10503 2 10504 2 10505 69 10506 131 10507 722 10508 2 10509 2 10510 725 10511 2 10512 727 10513 728 10514 2 10515 85 10516 86 10517 89 10518 134 10519 11 10520 735 10521 736 10522 99 10523 738 10524 739 10525 10525 10526 104 10527 104 10528 743 10529 744 10530 745 10531 746 10532 747 10533 748 10534 749 10535 750 10536 751 10537 752 10538 753 10539 99 10540 2 10541 756 10542 2 10543 2 10544 2 10545 2 10546 2 10547 2 10548 2 10549 2 10550 307 10551 2 10552 2 10553 2 10554 2 10555 2 10556 313 10557 2 10558 2 10559 2 10560 2 10561 2 10562 2 10563 2 10564 2 10565 2 10566 2 10567 2 10568 2 10569 2 10570 2 10571 2 10572 787 10573 788 10574 789 10575 790 10576 791 10577 792 10578 793 10579 794 10580 795 10581 796 10582 797 10583 798 10584 799 10585 800 10586 801 10587 4 10588 2 10589 804 10590 805 10591 806 10592 807 10593 808 10594 809 10595 810 10596 811 10597 812 10598 813 10599 814 10600 357 10601 357 10602 40 10603 360 10604 434 10605 820 10606 821 10607 822 10608 823 10609 824 10610 2 10611 574 10612 827 10613 828 10614 40 10615 830 10616 282 10617 832 10618 833 10619 2 10620 574 10621 836 10622 837 10623 838 10624 839 10625 840 10626 841 10627 2 10628 740 10629 2 10630 845 10631 846 10632 847 10633 848 10634 849 10635 850 10636 851 10637 852 10638 853 10639 854 10640 855 10641 856 10642 857 10643 858 10644 859 10645 860 10646 861 10647 862 10648 863 10649 864 10650 865 10651 866 10652 867 10653 868 10654 869 10655 870 10656 871 10657 872 10658 873 10659 874 10660 875 10661 876 10662 877 10663 878 10664 879 10665 880 10666 881 10667 882 10668 883 10669 884 10670 885 10671 886 10672 887 10673 888 10674 889 10675 890 10676 891 10677 892 10678 893 10679 894 10680 895 10681 896 10682 897 10683 898 10684 899 10685 900 10686 901 10687 902 10688 903 10689 904 10690 905 10691 906 10692 907 10693 908 10694 909 10695 910 10696 911 10697 912 10698 913 10699 914 10700 915 10701 916 10702 917 10703 918 10704 919 10705 920 10706 921 10707 922 10708 923 10709 924 10710 925 10711 926 10712 927 10713 928 10714 929 10715 930 10716 931 10717 932 10718 933 10719 934 10720 935 10721 11 10722 937 10723 938 10724 939 10725 2 10726 574 10727 942 10728 943 10729 944 10730 949 10731 976 10732 159 10733 162 10734 10273 10735 10343 10736 1487 10737 2 10738 2 10739 2 10740 10364 10741 2 10742 10366 10743 2 10744 2 10745 2 10746 2 10747 2 10748 2 10749 2 10750 2 10751 2 10752 10376 10753 10377 10754 10378 10755 10379 10756 10380 10757 10381 10758 10382 10759 10383 10760 10384 10761 10385 10762 10386 10763 10387 10764 10388 10765 22 10766 10390 10767 4 10768 11 10769 10393 10770 89 10771 11 10772 10396 10773 99 10774 10398 10775 1551 10776 10400 10777 10401 10778 10402 10779 99 10780 855 10781 11 10782 10406 10783 943 10784 10408 10785 10409 10786 976 10787 159 10788 162 10789 10413 10790 10414 10791 10415 10792 855 10793 11 10794 10418 10795 10419 10796 10420 10797 357 10798 357 10799 360 10800 434 10801 10425 10802 10426 10803 10427 10804 10428 10805 2 10806 574 10807 10431 10808 10432 10809 10433 10810 10434 10811 10435 10812 10436 10813 2 10814 574 10815 10439 10816 10440 10817 140 10818 11 10819 142 10820 162 10821 163 10822 165 10823 167 10824 77 10825 126 10826 170 10827 10451 10828 10452 10829 140 10830 11 10831 142 10832 162 10833 2248 10834 2250 10835 2251 10836 2252 10837 2 10838 10525 10839 10839 10840 10839 10841 855 10842 10842 10843 11 10844 10844 10845 2 10846 574 10847 10847 10848 10848 10849 10849 10850 2732 10851 10851 10852 10852 10853 357 10854 357 10855 360 10856 434 10857 10857 10858 10858 10859 10859 10860 10860 10861 2 10862 574 10863 10863 10864 10864 10865 10865 10866 10866 10867 2803 10868 10868 10869 434 10870 10870 10871 2 10872 10872 10873 10873 10874 2 10875 3 10876 4 10877 2 10878 2 10879 2 10880 2 10881 11 10882 2 10883 2 10884 10884 10885 2 10886 10886 10887 22 10888 10888 10889 10889 10890 2 10891 11 10892 11 10893 10893 10894 10894 10895 10895 10896 22 10897 2 10898 2 10899 2 10900 11 10901 10901 10902 11 10903 10903 10904 11 10905 10905 10906 10906 10907 10907 10908 10908 10909 60 10910 10910 10911 2 10912 2 10913 10913 10914 11 10915 2 10916 2 10917 69 10918 131 10919 10919 10920 2 10921 2 10922 10922 10923 2 10924 10924 10925 10925 10926 2 10927 85 10928 86 10929 89 10930 134 10931 11 10932 10932 10933 10933 10934 99 10935 10935 10936 10936 10937 10937 10938 104 10939 104 10940 10940 10941 10941 10942 10942 10943 10943 10944 10944 10945 10945 10946 10946 10947 10947 10948 10948 10949 10949 10950 10950 10951 99 10952 2 10953 10953 10954 2 10955 2 10956 2 10957 2 10958 2 10959 2 10960 2 10961 2 10962 307 10963 2 10964 2 10965 2 10966 2 10967 2 10968 313 10969 2 10970 2 10971 2 10972 2 10973 2 10974 2 10975 2 10976 2 10977 2 10978 2 10979 2 10980 2 10981 2 10982 2 10983 2 10984 10984 10985 10985 10986 10986 10987 10987 10988 10988 10989 10989 10990 10990 10991 10991 10992 10992 10993 10993 10994 10994 10995 10995 10996 10996 10997 10997 10998 10998 10999 4 11000 2 11001 11001 11002 11002 11003 11003 11004 11004 11005 11005 11006 11006 11007 11007 11008 11008 11009 11009 11010 11010 11011 11011 11012 357 11013 357 11014 40 11015 360 11016 434 11017 11017 11018 11018 11019 11019 11020 11020 11021 11021 11022 2 11023 574 11024 11024 11025 11025 11026 40 11027 11027 11028 282 11029 11029 11030 11030 11031 2 11032 574 11033 11033 11034 11034 11035 11035 11036 11036 11037 11037 11038 11038 11039 2 11040 740 11041 2 11042 11042 11043 11043 11044 11044 11045 11045 11046 11046 11047 11047 11048 11048 11049 11049 11050 11050 11051 11051 11052 855 11053 11053 11054 11054 11055 11055 11056 11056 11057 11057 11058 11058 11059 11059 11060 11060 11061 11061 11062 11062 11063 11063 11064 11064 11065 11065 11066 11066 11067 11067 11068 11068 11069 11069 11070 11070 11071 11071 11072 11072 11073 11073 11074 11074 11075 11075 11076 11076 11077 11077 11078 11078 11079 11079 11080 11080 11081 11081 11082 11082 11083 11083 11084 11084 11085 11085 11086 11086 11087 11087 11088 11088 11089 11089 11090 11090 11091 11091 11092 11092 11093 11093 11094 11094 11095 11095 11096 11096 11097 11097 11098 11098 11099 11099 11100 11100 11101 11101 11102 11102 11103 11103 11104 11104 11105 11105 11106 11106 11107 11107 11108 11108 11109 11109 11110 11110 11111 11111 11112 11112 11113 11113 11114 11114 11115 11115 11116 11116 11117 11117 11118 11118 11119 11119 11120 11120 11121 11121 11122 11122 11123 11123 11124 11124 11125 11125 11126 11126 11127 11127 11128 11128 11129 11129 11130 11130 11131 11131 11132 11132 11133 11 11134 11134 11135 11135 11136 11136 11137 2 11138 574 11139 11139 11140 943 11141 11141 11142 11142 11143 976 11144 159 11145 162 11146 11146 11147 11147 11148 1487 11149 2 11150 2 11151 2 11152 11152 11153 2 11154 11154 11155 2 11156 2 11157 2 11158 2 11159 2 11160 2 11161 2 11162 2 11163 2 11164 11164 11165 11165 11166 11166 11167 11167 11168 11168 11169 11169 11170 11170 11171 11171 11172 11172 11173 11173 11174 11174 11175 11175 11176 11176 11177 22 11178 11178 11179 4 11180 11 11181 11181 11182 89 11183 11 11184 11184 11185 99 11186 11186 11187 1551 11188 11188 11189 11189 11190 11190 11191 99 11192 855 11193 11 11194 11194 11195 943 11196 11196 11197 11197 11198 976 11199 159 11200 162 11201 11201 11202 11202 11203 11203 11204 855 11205 11 11206 11206 11207 11207 11208 11208 11209 357 11210 357 11211 360 11212 434 11213 11213 11214 11214 11215 11215 11216 11216 11217 2 11218 574 11219 11219 11220 11220 11221 11221 11222 11222 11223 11223 11224 11224 11225 2 11226 574 11227 11227 11228 11228 11229 140 11230 11 11231 142 11232 162 11233 163 11234 165 11235 167 11236 77 11237 126 11238 170 11239 11239 11240 11240 11241 140 11242 11 11243 142 11244 162 11245 2248 11246 2250 11247 2251 11248 2252 11249 2 11250 11250 11251 11251 11252 2 11253 11253 11254 855 11255 11255 11256 11 11257 11257 11258 2 11259 574 11260 11260 11261 11261 11262 11262 11263 2732 11264 11264 11265 11265 11266 357 11267 357 11268 360 11269 434 11270 11270 11271 11271 11272 11272 11273 11273 11274 2 11275 574 11276 11276 11277 11277 11278 11278 11279 11279 11280 2803 11281 11281 11282 434 11283 11283 11284 2 11285 11285 11286 11286 11287 10937 11288 2 11289 11289 11290 11290 11291 11291 11292 11292 11293 11293 11294 2894 11295 11295 11296 69 11297 11297 11298 11298 11299 11299 11300 11300 11301 11301 11302 11302 11303 11303 11304 2 11305 11305 11306 11306 11307 11307 11308 11308 11309 11309 11310 11310 11311 2936 11312 2938 11313 11313 11314 11314 11315 11315 11316 11316 11317 11317 11318 357 11319 357 11320 40 11321 360 11322 434 11323 11323 11324 11324 11325 11325 11326 11326 11327 11327 11328 40 11329 11329 11330 11330 11331 2959 11332 2962 11333 2964 11334 2966 11335 77 11336 2968 11337 2969 11338 11338 11339 282 11340 11340 11341 2 11342 282 11343 77 11344 3931 11345 2828 11346 11346 11347 1862 11348 162 11349 2050 11350 2 11351 11351 11352 2 11353 159 11354 162 11355 11355 11356 11356 11357 11357 11358 11358 11359 11359 11360 943 11361 11361 11362 11362 11363 11363 11364 976 11365 159 11366 162 11367 11367 11368 11368 11369 2 11370 282 11371 11371 11372 2 11373 282 11374 11374 11375 11375 11376 11376 11377 11377 11378 11378 11379 11379 11380 1862 11381 162 11382 11382 11383 11383 11384 11384 11385 11385 11386 282 11387 11387 11388 11388 11389 2 11390 574 11391 11391 11392 5236 11393 11393 11394 11394 11395 11395 11396 11396 11397 11397 11398 11398 11399 2894 11400 11400 11401 69 11402 11402 11403 11403 11404 11404 11405 11405 11406 11406 11407 11407 11408 11408 11409 2 11410 11410 11411 11411 11412 11412 11413 11413 11414 11414 11415 11415 11416 2936 11417 2938 11418 11418 11419 11419 11420 11420 11421 11421 11422 11422 11423 357 11424 357 11425 40 11426 360 11427 434 11428 11428 11429 11429 11430 11430 11431 11431 11432 11432 11433 40 11434 11434 11435 11435 11436 2959 11437 2962 11438 2964 11439 2966 11440 77 11441 2968 11442 2969 11443 11443 11444 282 11445 11445 11446 2 11447 282 11448 77 11449 3931 11450 2828 11451 11451 11452 1862 11453 162 11454 2050 11455 2 11456 11456 11457 2 11458 159 11459 162 11460 11460 11461 11461 11462 11462 11463 11463 11464 11464 11465 943 11466 11466 11467 11467 11468 11468 11469 976 11470 159 11471 162 11472 11472 11473 11473 11474 2 11475 282 11476 11476 11477 2 11478 282 11479 11479 11480 11480 11481 11481 11482 11482 11483 11483 11484 11484 11485 1862 11486 162 11487 11487 11488 11488 11489 11489 11490 11490 11491 282 11492 11492 11493 11493 11494 2 11495 574 11496 11496 11497 5236 11498 11498 11499 11499 11500 2 11501 11501 11502 855 11503 10842 11504 11 11505 10844 11506 2 11507 574 11508 10847 11509 10848 11510 10849 11511 2732 11512 10851 11513 10852 11514 357 11515 357 11516 360 11517 434 11518 10857 11519 10858 11520 10859 11521 10860 11522 2 11523 574 11524 10863 11525 10864 11526 10865 11527 10866 11528 2803 11529 10868 11530 434 11531 10870 11532 2 11533 10872 11534 10873 11535 2 11536 3 11537 4 11538 2 11539 2 11540 2 11541 2 11542 11 11543 2 11544 2 11545 10884 11546 2 11547 10886 11548 22 11549 10888 11550 10889 11551 2 11552 11 11553 11 11554 10893 11555 10894 11556 10895 11557 22 11558 2 11559 2 11560 2 11561 11 11562 10901 11563 11 11564 10903 11565 11 11566 10905 11567 10906 11568 10907 11569 10908 11570 60 11571 10910 11572 2 11573 2 11574 10913 11575 11 11576 2 11577 2 11578 69 11579 131 11580 10919 11581 2 11582 2 11583 10922 11584 2 11585 10924 11586 10925 11587 2 11588 85 11589 86 11590 89 11591 134 11592 11 11593 10932 11594 10933 11595 99 11596 10935 11597 10936 11598 10937 11599 104 11600 104 11601 10940 11602 10941 11603 10942 11604 10943 11605 10944 11606 10945 11607 10946 11608 10947 11609 10948 11610 10949 11611 10950 11612 99 11613 2 11614 10953 11615 2 11616 2 11617 2 11618 2 11619 2 11620 2 11621 2 11622 2 11623 307 11624 2 11625 2 11626 2 11627 2 11628 2 11629 313 11630 2 11631 2 11632 2 11633 2 11634 2 11635 2 11636 2 11637 2 11638 2 11639 2 11640 2 11641 2 11642 2 11643 2 11644 2 11645 10984 11646 10985 11647 10986 11648 10987 11649 10988 11650 10989 11651 10990 11652 10991 11653 10992 11654 10993 11655 10994 11656 10995 11657 10996 11658 10997 11659 10998 11660 4 11661 2 11662 11001 11663 11002 11664 11003 11665 11004 11666 11005 11667 11006 11668 11007 11669 11008 11670 11009 11671 11010 11672 11011 11673 357 11674 357 11675 40 11676 360 11677 434 11678 11017 11679 11018 11680 11019 11681 11020 11682 11021 11683 2 11684 574 11685 11024 11686 11025 11687 40 11688 11027 11689 282 11690 11029 11691 11030 11692 2 11693 574 11694 11033 11695 11034 11696 11035 11697 11036 11698 11037 11699 11038 11700 2 11701 740 11702 2 11703 11042 11704 11043 11705 11044 11706 11045 11707 11046 11708 11047 11709 11048 11710 11049 11711 11050 11712 11051 11713 855 11714 11053 11715 11054 11716 11055 11717 11056 11718 11057 11719 11058 11720 11059 11721 11060 11722 11061 11723 11062 11724 11063 11725 11064 11726 11065 11727 11066 11728 11067 11729 11068 11730 11069 11731 11070 11732 11071 11733 11072 11734 11073 11735 11074 11736 11075 11737 11076 11738 11077 11739 11078 11740 11079 11741 11080 11742 11081 11743 11082 11744 11083 11745 11084 11746 11085 11747 11086 11748 11087 11749 11088 11750 11089 11751 11090 11752 11091 11753 11092 11754 11093 11755 11094 11756 11095 11757 11096 11758 11097 11759 11098 11760 11099 11761 11100 11762 11101 11763 11102 11764 11103 11765 11104 11766 11105 11767 11106 11768 11107 11769 11108 11770 11109 11771 11110 11772 11111 11773 11112 11774 11113 11775 11114 11776 11115 11777 11116 11778 11117 11779 11118 11780 11119 11781 11120 11782 11121 11783 11122 11784 11123 11785 11124 11786 11125 11787 11126 11788 11127 11789 11128 11790 11129 11791 11130 11792 11131 11793 11132 11794 11 11795 11134 11796 11135 11797 11136 11798 2 11799 574 11800 11139 11801 943 11802 11141 11803 11142 11804 976 11805 159 11806 162 11807 11146 11808 11147 11809 1487 11810 2 11811 2 11812 2 11813 11152 11814 2 11815 11154 11816 2 11817 2 11818 2 11819 2 11820 2 11821 2 11822 2 11823 2 11824 2 11825 11164 11826 11165 11827 11166 11828 11167 11829 11168 11830 11169 11831 11170 11832 11171 11833 11172 11834 11173 11835 11174 11836 11175 11837 11176 11838 22 11839 11178 11840 4 11841 11 11842 11181 11843 89 11844 11 11845 11184 11846 99 11847 11186 11848 1551 11849 11188 11850 11189 11851 11190 11852 99 11853 855 11854 11 11855 11194 11856 943 11857 11196 11858 11197 11859 976 11860 159 11861 162 11862 11201 11863 11202 11864 11203 11865 855 11866 11 11867 11206 11868 11207 11869 11208 11870 357 11871 357 11872 360 11873 434 11874 11213 11875 11214 11876 11215 11877 11216 11878 2 11879 574 11880 11219 11881 11220 11882 11221 11883 11222 11884 11223 11885 11224 11886 2 11887 574 11888 11227 11889 11228 11890 140 11891 11 11892 142 11893 162 11894 163 11895 165 11896 167 11897 77 11898 126 11899 170 11900 11239 11901 11240 11902 140 11903 11 11904 142 11905 162 11906 2248 11907 2250 11908 2251 11909 2252 11910 2 11911 11250 11912 11251 11913 2 11914 11253 11915 855 11916 11255 11917 11 11918 11257 11919 2 11920 574 11921 11260 11922 11261 11923 11262 11924 2732 11925 11264 11926 11265 11927 357 11928 357 11929 360 11930 434 11931 11270 11932 11271 11933 11272 11934 11273 11935 2 11936 574 11937 11276 11938 11277 11939 11278 11940 11279 11941 2803 11942 11281 11943 434 11944 11283 11945 2 11946 11285 11947 11286 11948 10937 11949 2 11950 11289 11951 11290 11952 11291 11953 11292 11954 11293 11955 2894 11956 11295 11957 69 11958 11297 11959 11298 11960 11299 11961 11300 11962 11301 11963 11302 11964 11303 11965 2 11966 11305 11967 11306 11968 11307 11969 11308 11970 11309 11971 11310 11972 2936 11973 2938 11974 11313 11975 11314 11976 11315 11977 11316 11978 11317 11979 357 11980 357 11981 40 11982 360 11983 434 11984 11323 11985 11324 11986 11325 11987 11326 11988 11327 11989 40 11990 11329 11991 11330 11992 2959 11993 2962 11994 2964 11995 2966 11996 77 11997 2968 11998 2969 11999 11338 12000 282 12001 11340 12002 2 12003 282 12004 77 12005 3931 12006 2828 12007 11346 12008 1862 12009 162 12010 2050 12011 2 12012 11351 12013 2 12014 159 12015 162 12016 11355 12017 11356 12018 11357 12019 11358 12020 11359 12021 943 12022 11361 12023 11362 12024 11363 12025 976 12026 159 12027 162 12028 11367 12029 11368 12030 2 12031 282 12032 11371 12033 2 12034 282 12035 11374 12036 11375 12037 11376 12038 11377 12039 11378 12040 11379 12041 1862 12042 162 12043 11382 12044 11383 12045 11384 12046 11385 12047 282 12048 11387 12049 11388 12050 2 12051 574 12052 11391 12053 5236 12054 11393 12055 11394 12056 11395 12057 11396 12058 11397 12059 11398 12060 2894 12061 11400 12062 69 12063 11402 12064 11403 12065 11404 12066 11405 12067 11406 12068 11407 12069 11408 12070 2 12071 11410 12072 11411 12073 11412 12074 11413 12075 11414 12076 11415 12077 2936 12078 2938 12079 11418 12080 11419 12081 11420 12082 11421 12083 11422 12084 357 12085 357 12086 40 12087 360 12088 434 12089 11428 12090 11429 12091 11430 12092 11431 12093 11432 12094 40 12095 11434 12096 11435 12097 2959 12098 2962 12099 2964 12100 2966 12101 77 12102 2968 12103 2969 12104 11443 12105 282 12106 11445 12107 2 12108 282 12109 77 12110 3931 12111 2828 12112 11451 12113 1862 12114 162 12115 2050 12116 2 12117 11456 12118 2 12119 159 12120 162 12121 11460 12122 11461 12123 11462 12124 11463 12125 11464 12126 943 12127 11466 12128 11467 12129 11468 12130 976 12131 159 12132 162 12133 11472 12134 11473 12135 2 12136 282 12137 11476 12138 2 12139 282 12140 11479 12141 11480 12142 11481 12143 11482 12144 11483 12145 11484 12146 1862 12147 162 12148 11487 12149 11488 12150 12150 12151 11489 12152 11490 12153 282 12154 11492 12155 11493 12156 2 12157 574 12158 11496 12159 5236 12160 11498 12161 11499 12162 12162 12163 12163 12164 12164 12165 12165 12166 12166 12167 12167 12168 12168 12169 12169 12170 855 12171 12171 12172 12172 12173 12173 12174 12174 12175 12175 12176 12176 12177 12177 12178 12178 12179 12179 12180 12180 12181 12181 12182 12182 12183 12183 12184 12184 12185 12185 12186 12186 12187 12187 12188 12188 12189 12189 12190 12190 12191 12191 12192 12192 12193 12193 12194 12194 12195 12195 12196 12196 12197 12197 12198 12198 12199 12199 12200 12200 12201 12201 12202 12202 12203 12203 12204 12204 12205 12205 12206 12206 12207 12207 12208 12208 12209 12209 12210 12210 12211 12211 12212 12212 12213 12213 12214 12214 12215 12215 12216 12216 12217 12217 12218 12218 12219 12219 12220 12220 12221 12221 12222 12222 12223 12223 12224 12224 12225 12225 12226 12226 12227 12227 12228 12228 12229 12229 12230 12230 12231 12231 12232 12232 12233 12233 12234 12234 12235 12235 12236 12236 12237 12237 12238 12238 12239 12239 12240 12240 12241 12241 12242 12242 12243 12243 12244 12244 12245 12245 12246 12246 12247 12247 12248 12248 12249 12249 12250 12250 12251 11 12252 12252 12253 12253 12254 12254 12255 2 12256 574 12257 12257 12258 943 12259 12259 12260 12260 12261 976 12262 159 12263 162 12264 12264 12265 12265 12266 1487 12267 2 12268 2 12269 2 12270 12270 12271 2 12272 12272 12273 2 12274 2 12275 2 12276 2 12277 2 12278 2 12279 2 12280 2 12281 2 12282 12282 12283 12283 12284 12284 12285 12285 12286 12286 12287 12287 12288 12288 12289 12289 12290 12290 12291 12291 12292 12292 12293 12293 12294 12294 12295 22 12296 12296 12297 4 12298 11 12299 12299 12300 89 12301 11 12302 12302 12303 99 12304 12304 12305 1551 12306 12306 12307 12307 12308 12308 12309 99 12310 855 12311 11 12312 12312 12313 943 12314 12314 12315 12315 12316 976 12317 159 12318 162 12319 12319 12320 12320 12321 12321 12322 855 12323 11 12324 12324 12325 12325 12326 12326 12327 357 12328 357 12329 360 12330 434 12331 12331 12332 12332 12333 12333 12334 12334 12335 2 12336 574 12337 12337 12338 12338 12339 12339 12340 12340 12341 12341 12342 12342 12343 2 12344 574 12345 12345 12346 12346 12347 140 12348 11 12349 142 12350 162 12351 163 12352 165 12353 167 12354 77 12355 126 12356 170 12357 12357 12358 12358 12359 140 12360 11 12361 142 12362 162 12363 2248 12364 2250 12365 2251 12366 2252 12367 2 12368 2 12369 3 12370 4 12371 2 12372 2 12373 2 12374 2 12375 11 12376 2 12377 2 12378 521 12379 2 12380 523 12381 22 12382 525 12383 526 12384 2 12385 11 12386 11 12387 530 12388 531 12389 532 12390 22 12391 2 12392 2 12393 2 12394 11 12395 538 12396 11 12397 540 12398 11 12399 542 12400 543 12401 544 12402 545 12403 60 12404 547 12405 2 12406 2 12407 550 12408 11 12409 2 12410 2 12411 69 12412 131 12413 556 12414 2 12415 2 12416 559 12417 2 12418 561 12419 562 12420 2 12421 85 12422 86 12423 89 12424 134 12425 11 12426 569 12427 570 12428 99 12429 572 12430 573 12431 12431 12432 104 12433 104 12434 577 12435 578 12436 579 12437 580 12438 581 12439 582 12440 583 12441 584 12442 585 12443 586 12444 587 12445 99 12446 2 12447 590 12448 2 12449 2 12450 2 12451 2 12452 2 12453 2 12454 2 12455 2 12456 307 12457 2 12458 2 12459 2 12460 2 12461 2 12462 313 12463 2 12464 2 12465 2 12466 2 12467 2 12468 2 12469 2 12470 2 12471 2 12472 2 12473 2 12474 2 12475 2 12476 2 12477 2 12478 621 12479 622 12480 623 12481 624 12482 625 12483 626 12484 627 12485 628 12486 629 12487 630 12488 631 12489 632 12490 633 12491 634 12492 635 12493 4 12494 2 12495 638 12496 639 12497 640 12498 641 12499 642 12500 643 12501 644 12502 645 12503 646 12504 647 12505 648 12506 357 12507 357 12508 40 12509 360 12510 434 12511 654 12512 655 12513 656 12514 657 12515 658 12516 2 12517 574 12518 661 12519 662 12520 40 12521 664 12522 282 12523 666 12524 667 12525 2 12526 574 12527 670 12528 671 12529 672 12530 673 12531 674 12532 675 12533 2 12534 740 12535 2 12536 11501 12537 12150 12538 12162 12539 12163 12540 12164 12541 12165 12542 12166 12543 12167 12544 12168 12545 12169 12546 855 12547 12171 12548 12172 12549 12173 12550 12174 12551 12175 12552 12176 12553 12177 12554 12178 12555 12179 12556 12180 12557 12181 12558 12182 12559 12183 12560 12184 12561 12185 12562 12186 12563 12187 12564 12188 12565 12189 12566 12190 12567 12191 12568 12192 12569 12193 12570 12194 12571 12195 12572 12196 12573 12197 12574 12198 12575 12199 12576 12200 12577 12201 12578 12202 12579 12203 12580 12204 12581 12205 12582 12206 12583 12207 12584 12208 12585 12209 12586 12210 12587 12211 12588 12212 12589 12213 12590 12214 12591 12215 12592 12216 12593 12217 12594 12218 12595 12219 12596 12220 12597 12221 12598 12222 12599 12223 12600 12224 12601 12225 12602 12226 12603 12227 12604 12228 12605 12229 12606 12230 12607 12231 12608 12232 12609 12233 12610 12234 12611 12235 12612 12236 12613 12237 12614 12238 12615 12239 12616 12240 12617 12241 12618 12242 12619 12243 12620 12244 12621 12245 12622 12246 12623 12247 12624 12248 12625 12249 12626 12250 12627 11 12628 12252 12629 12253 12630 12254 12631 2 12632 574 12633 12257 12634 943 12635 12259 12636 12260 12637 976 12638 159 12639 162 12640 12264 12641 12265 12642 1487 12643 2 12644 2 12645 2 12646 12270 12647 2 12648 12272 12649 2 12650 2 12651 2 12652 2 12653 2 12654 2 12655 2 12656 2 12657 2 12658 12282 12659 12283 12660 12284 12661 12285 12662 12286 12663 12287 12664 12288 12665 12289 12666 12290 12667 12291 12668 12292 12669 12293 12670 12294 12671 22 12672 12296 12673 4 12674 11 12675 12299 12676 89 12677 11 12678 12302 12679 99 12680 12304 12681 1551 12682 12306 12683 12307 12684 12308 12685 99 12686 855 12687 11 12688 12312 12689 943 12690 12314 12691 12315 12692 976 12693 159 12694 162 12695 12319 12696 12320 12697 12321 12698 855 12699 11 12700 12324 12701 12325 12702 12326 12703 357 12704 357 12705 360 12706 434 12707 12331 12708 12332 12709 12333 12710 12334 12711 2 12712 574 12713 12337 12714 12338 12715 12339 12716 12340 12717 12341 12718 12342 12719 2 12720 574 12721 12345 12722 12346 12723 140 12724 11 12725 142 12726 162 12727 163 12728 165 12729 167 12730 77 12731 126 12732 170 12733 12357 12734 12358 12735 140 12736 11 12737 142 12738 162 12739 2248 12740 2250 12741 2251 12742 2252 12743 2 12744 12431 12745 12745 12746 12745 12747 855 12748 12748 12749 11 12750 12750 12751 2 12752 574 12753 12753 12754 12754 12755 12755 12756 2732 12757 12757 12758 12758 12759 357 12760 357 12761 360 12762 434 12763 12763 12764 12764 12765 12765 12766 12766 12767 2 12768 574 12769 12769 12770 12770 12771 12771 12772 12772 12773 2803 12774 12774 12775 434 12776 12776 12777 2 12778 12778 12779 12779 12780 2 12781 3 12782 4 12783 2 12784 2 12785 2 12786 2 12787 11 12788 2 12789 2 12790 12790 12791 2 12792 12792 12793 22 12794 12794 12795 12795 12796 2 12797 11 12798 11 12799 12799 12800 12800 12801 12801 12802 22 12803 2 12804 2 12805 2 12806 11 12807 12807 12808 11 12809 12809 12810 11 12811 12811 12812 12812 12813 12813 12814 12814 12815 60 12816 12816 12817 2 12818 2 12819 12819 12820 11 12821 2 12822 2 12823 69 12824 131 12825 12825 12826 2 12827 2 12828 12828 12829 2 12830 12830 12831 12831 12832 2 12833 85 12834 86 12835 89 12836 134 12837 11 12838 12838 12839 12839 12840 99 12841 12841 12842 12842 12843 12843 12844 104 12845 104 12846 12846 12847 12847 12848 12848 12849 12849 12850 12850 12851 12851 12852 12852 12853 12853 12854 12854 12855 12855 12856 12856 12857 99 12858 2 12859 12859 12860 2 12861 2 12862 2 12863 2 12864 2 12865 2 12866 2 12867 2 12868 307 12869 2 12870 2 12871 2 12872 2 12873 2 12874 313 12875 2 12876 2 12877 2 12878 2 12879 2 12880 2 12881 2 12882 2 12883 2 12884 2 12885 2 12886 2 12887 2 12888 2 12889 2 12890 12890 12891 12891 12892 12892 12893 12893 12894 12894 12895 12895 12896 12896 12897 12897 12898 12898 12899 12899 12900 12900 12901 12901 12902 12902 12903 12903 12904 12904 12905 4 12906 2 12907 12907 12908 12908 12909 12909 12910 12910 12911 12911 12912 12912 12913 12913 12914 12914 12915 12915 12916 12916 12917 12917 12918 357 12919 357 12920 40 12921 360 12922 434 12923 12923 12924 12924 12925 12925 12926 12926 12927 12927 12928 2 12929 574 12930 12930 12931 12931 12932 40 12933 12933 12934 282 12935 12935 12936 12936 12937 2 12938 574 12939 12939 12940 12940 12941 12941 12942 12942 12943 12943 12944 12944 12945 2 12946 740 12947 2 12948 12948 12949 12949 12950 12950 12951 12951 12952 12952 12953 12953 12954 12954 12955 12955 12956 12956 12957 12957 12958 855 12959 12959 12960 12960 12961 12961 12962 12962 12963 12963 12964 12964 12965 12965 12966 12966 12967 12967 12968 12968 12969 12969 12970 12970 12971 12971 12972 12972 12973 12973 12974 12974 12975 12975 12976 12976 12977 12977 12978 12978 12979 12979 12980 12980 12981 12981 12982 12982 12983 12983 12984 12984 12985 12985 12986 12986 12987 12987 12988 12988 12989 12989 12990 12990 12991 12991 12992 12992 12993 12993 12994 12994 12995 12995 12996 12996 12997 12997 12998 12998 12999 12999 13000 13000 13001 13001 13002 13002 13003 13003 13004 13004 13005 13005 13006 13006 13007 13007 13008 13008 13009 13009 13010 13010 13011 13011 13012 13012 13013 13013 13014 13014 13015 13015 13016 13016 13017 13017 13018 13018 13019 13019 13020 13020 13021 13021 13022 13022 13023 13023 13024 13024 13025 13025 13026 13026 13027 13027 13028 13028 13029 13029 13030 13030 13031 13031 13032 13032 13033 13033 13034 13034 13035 13035 13036 13036 13037 13037 13038 13038 13039 11 13040 13040 13041 13041 13042 13042 13043 2 13044 574 13045 13045 13046 943 13047 13047 13048 13048 13049 976 13050 159 13051 162 13052 13052 13053 13053 13054 1487 13055 2 13056 2 13057 2 13058 13058 13059 2 13060 13060 13061 2 13062 2 13063 2 13064 2 13065 2 13066 2 13067 2 13068 2 13069 2 13070 13070 13071 13071 13072 13072 13073 13073 13074 13074 13075 13075 13076 13076 13077 13077 13078 13078 13079 13079 13080 13080 13081 13081 13082 13082 13083 22 13084 13084 13085 4 13086 11 13087 13087 13088 89 13089 11 13090 13090 13091 99 13092 13092 13093 1551 13094 13094 13095 13095 13096 13096 13097 99 13098 855 13099 11 13100 13100 13101 943 13102 13102 13103 13103 13104 976 13105 159 13106 162 13107 13107 13108 13108 13109 13109 13110 855 13111 11 13112 13112 13113 13113 13114 13114 13115 357 13116 357 13117 360 13118 434 13119 13119 13120 13120 13121 13121 13122 13122 13123 2 13124 574 13125 13125 13126 13126 13127 13127 13128 13128 13129 13129 13130 13130 13131 2 13132 574 13133 13133 13134 13134 13135 140 13136 11 13137 142 13138 162 13139 163 13140 165 13141 167 13142 77 13143 126 13144 170 13145 13145 13146 13146 13147 140 13148 11 13149 142 13150 162 13151 2248 13152 2250 13153 2251 13154 2252 13155 2 13156 13156 13157 13157 13158 2 13159 13159 13160 855 13161 13161 13162 11 13163 13163 13164 2 13165 574 13166 13166 13167 13167 13168 13168 13169 2732 13170 13170 13171 13171 13172 357 13173 357 13174 360 13175 434 13176 13176 13177 13177 13178 13178 13179 13179 13180 2 13181 574 13182 13182 13183 13183 13184 13184 13185 13185 13186 2803 13187 13187 13188 434 13189 13189 13190 2 13191 13191 13192 13192 13193 12843 13194 2 13195 13195 13196 13196 13197 13197 13198 13198 13199 13199 13200 2894 13201 13201 13202 69 13203 13203 13204 13204 13205 13205 13206 13206 13207 13207 13208 13208 13209 13209 13210 2 13211 13211 13212 13212 13213 13213 13214 13214 13215 13215 13216 13216 13217 2936 13218 2938 13219 13219 13220 13220 13221 13221 13222 13222 13223 13223 13224 357 13225 357 13226 40 13227 360 13228 434 13229 13229 13230 13230 13231 13231 13232 13232 13233 13233 13234 40 13235 13235 13236 13236 13237 2959 13238 2962 13239 2964 13240 2966 13241 77 13242 2968 13243 2969 13244 13244 13245 282 13246 13246 13247 2 13248 282 13249 77 13250 3931 13251 2828 13252 13252 13253 1862 13254 162 13255 2050 13256 2 13257 13257 13258 2 13259 159 13260 162 13261 13261 13262 13262 13263 13263 13264 13264 13265 13265 13266 943 13267 13267 13268 13268 13269 13269 13270 976 13271 159 13272 162 13273 13273 13274 13274 13275 2 13276 282 13277 13277 13278 2 13279 282 13280 13280 13281 13281 13282 13282 13283 13283 13284 13284 13285 13285 13286 1862 13287 162 13288 13288 13289 13289 13290 13290 13291 13291 13292 282 13293 13293 13294 13294 13295 2 13296 574 13297 13297 13298 5236 13299 13299 13300 13300 13301 13301 13302 13302 13303 13303 13304 13304 13305 2894 13306 13306 13307 69 13308 13308 13309 13309 13310 13310 13311 13311 13312 13312 13313 13313 13314 13314 13315 2 13316 13316 13317 13317 13318 13318 13319 13319 13320 13320 13321 13321 13322 2936 13323 2938 13324 13324 13325 13325 13326 13326 13327 13327 13328 13328 13329 357 13330 357 13331 40 13332 360 13333 434 13334 13334 13335 13335 13336 13336 13337 13337 13338 13338 13339 40 13340 13340 13341 13341 13342 2959 13343 2962 13344 2964 13345 2966 13346 77 13347 2968 13348 2969 13349 13349 13350 282 13351 13351 13352 2 13353 282 13354 77 13355 3931 13356 2828 13357 13357 13358 1862 13359 162 13360 2050 13361 2 13362 13362 13363 2 13364 159 13365 162 13366 13366 13367 13367 13368 13368 13369 13369 13370 13370 13371 943 13372 13372 13373 13373 13374 13374 13375 976 13376 159 13377 162 13378 13378 13379 13379 13380 2 13381 282 13382 13382 13383 2 13384 282 13385 13385 13386 13386 13387 13387 13388 13388 13389 13389 13390 13390 13391 1862 13392 162 13393 13393 13394 13394 13395 13395 13396 13396 13397 282 13398 13398 13399 13399 13400 2 13401 574 13402 13402 13403 5236 13404 13404 13405 13405 13406 13406 13407 855 13408 12748 13409 11 13410 12750 13411 2 13412 574 13413 12753 13414 12754 13415 12755 13416 2732 13417 12757 13418 12758 13419 357 13420 357 13421 360 13422 434 13423 12763 13424 12764 13425 12765 13426 12766 13427 2 13428 574 13429 12769 13430 12770 13431 12771 13432 12772 13433 2803 13434 12774 13435 434 13436 12776 13437 2 13438 12778 13439 12779 13440 2 13441 3 13442 4 13443 2 13444 2 13445 2 13446 2 13447 11 13448 2 13449 2 13450 12790 13451 2 13452 12792 13453 22 13454 12794 13455 12795 13456 2 13457 11 13458 11 13459 12799 13460 12800 13461 12801 13462 22 13463 2 13464 2 13465 2 13466 11 13467 12807 13468 11 13469 12809 13470 11 13471 12811 13472 12812 13473 12813 13474 12814 13475 60 13476 12816 13477 2 13478 2 13479 12819 13480 11 13481 2 13482 2 13483 69 13484 131 13485 12825 13486 2 13487 2 13488 12828 13489 2 13490 12830 13491 12831 13492 2 13493 85 13494 86 13495 89 13496 134 13497 11 13498 12838 13499 12839 13500 99 13501 12841 13502 12842 13503 12843 13504 104 13505 104 13506 12846 13507 12847 13508 12848 13509 12849 13510 12850 13511 12851 13512 12852 13513 12853 13514 12854 13515 12855 13516 12856 13517 99 13518 2 13519 12859 13520 2 13521 2 13522 2 13523 2 13524 2 13525 2 13526 2 13527 2 13528 307 13529 2 13530 2 13531 2 13532 2 13533 2 13534 313 13535 2 13536 2 13537 2 13538 2 13539 2 13540 2 13541 2 13542 2 13543 2 13544 2 13545 2 13546 2 13547 2 13548 2 13549 2 13550 12890 13551 12891 13552 12892 13553 12893 13554 12894 13555 12895 13556 12896 13557 12897 13558 12898 13559 12899 13560 12900 13561 12901 13562 12902 13563 12903 13564 12904 13565 4 13566 2 13567 12907 13568 12908 13569 12909 13570 12910 13571 12911 13572 12912 13573 12913 13574 12914 13575 12915 13576 12916 13577 12917 13578 357 13579 357 13580 40 13581 360 13582 434 13583 12923 13584 12924 13585 12925 13586 12926 13587 12927 13588 2 13589 574 13590 12930 13591 12931 13592 40 13593 12933 13594 282 13595 12935 13596 12936 13597 2 13598 574 13599 12939 13600 12940 13601 12941 13602 12942 13603 12943 13604 12944 13605 2 13606 740 13607 2 13608 12948 13609 12949 13610 12950 13611 12951 13612 12952 13613 12953 13614 12954 13615 12955 13616 12956 13617 12957 13618 855 13619 12959 13620 12960 13621 12961 13622 12962 13623 12963 13624 12964 13625 12965 13626 12966 13627 12967 13628 12968 13629 12969 13630 12970 13631 12971 13632 12972 13633 12973 13634 12974 13635 12975 13636 12976 13637 12977 13638 12978 13639 12979 13640 12980 13641 12981 13642 12982 13643 12983 13644 12984 13645 12985 13646 12986 13647 12987 13648 12988 13649 12989 13650 12990 13651 12991 13652 12992 13653 12993 13654 12994 13655 12995 13656 12996 13657 12997 13658 12998 13659 12999 13660 13000 13661 13001 13662 13002 13663 13003 13664 13004 13665 13005 13666 13006 13667 13007 13668 13008 13669 13009 13670 13010 13671 13011 13672 13012 13673 13013 13674 13014 13675 13015 13676 13016 13677 13017 13678 13018 13679 13019 13680 13020 13681 13021 13682 13022 13683 13023 13684 13024 13685 13025 13686 13026 13687 13027 13688 13028 13689 13029 13690 13030 13691 13031 13692 13032 13693 13033 13694 13034 13695 13035 13696 13036 13697 13037 13698 13038 13699 11 13700 13040 13701 13041 13702 13042 13703 2 13704 574 13705 13045 13706 943 13707 13047 13708 13048 13709 976 13710 159 13711 162 13712 13052 13713 13053 13714 1487 13715 2 13716 2 13717 2 13718 13058 13719 2 13720 13060 13721 2 13722 2 13723 2 13724 2 13725 2 13726 2 13727 2 13728 2 13729 2 13730 13070 13731 13071 13732 13072 13733 13073 13734 13074 13735 13075 13736 13076 13737 13077 13738 13078 13739 13079 13740 13080 13741 13081 13742 13082 13743 22 13744 13084 13745 4 13746 11 13747 13087 13748 89 13749 11 13750 13090 13751 99 13752 13092 13753 1551 13754 13094 13755 13095 13756 13096 13757 99 13758 855 13759 11 13760 13100 13761 943 13762 13102 13763 13103 13764 976 13765 159 13766 162 13767 13107 13768 13108 13769 13109 13770 855 13771 11 13772 13112 13773 13113 13774 13114 13775 357 13776 357 13777 360 13778 434 13779 13119 13780 13120 13781 13121 13782 13122 13783 2 13784 574 13785 13125 13786 13126 13787 13127 13788 13128 13789 13129 13790 13130 13791 2 13792 574 13793 13133 13794 13134 13795 140 13796 11 13797 142 13798 162 13799 163 13800 165 13801 167 13802 77 13803 126 13804 170 13805 13145 13806 13146 13807 140 13808 11 13809 142 13810 162 13811 2248 13812 2250 13813 2251 13814 2252 13815 2 13816 13156 13817 13157 13818 2 13819 13159 13820 855 13821 13161 13822 11 13823 13163 13824 2 13825 574 13826 13166 13827 13167 13828 13168 13829 2732 13830 13170 13831 13171 13832 357 13833 357 13834 360 13835 434 13836 13176 13837 13177 13838 13178 13839 13179 13840 2 13841 574 13842 13182 13843 13183 13844 13184 13845 13185 13846 2803 13847 13187 13848 434 13849 13189 13850 2 13851 13191 13852 13192 13853 12843 13854 2 13855 13195 13856 13196 13857 13197 13858 13198 13859 13199 13860 2894 13861 13201 13862 69 13863 13203 13864 13204 13865 13205 13866 13206 13867 13207 13868 13208 13869 13209 13870 2 13871 13211 13872 13212 13873 13213 13874 13214 13875 13215 13876 13216 13877 2936 13878 2938 13879 13219 13880 13220 13881 13221 13882 13222 13883 13223 13884 357 13885 357 13886 40 13887 360 13888 434 13889 13229 13890 13230 13891 13231 13892 13232 13893 13233 13894 40 13895 13235 13896 13236 13897 2959 13898 2962 13899 2964 13900 2966 13901 77 13902 2968 13903 2969 13904 13244 13905 282 13906 13246 13907 2 13908 282 13909 77 13910 3931 13911 2828 13912 13252 13913 1862 13914 162 13915 2050 13916 2 13917 13257 13918 2 13919 159 13920 162 13921 13261 13922 13262 13923 13263 13924 13264 13925 13265 13926 943 13927 13267 13928 13268 13929 13269 13930 976 13931 159 13932 162 13933 13273 13934 13274 13935 2 13936 282 13937 13277 13938 2 13939 282 13940 13280 13941 13281 13942 13282 13943 13283 13944 13284 13945 13285 13946 1862 13947 162 13948 13288 13949 13289 13950 13290 13951 13291 13952 282 13953 13293 13954 13294 13955 2 13956 574 13957 13297 13958 5236 13959 13299 13960 13300 13961 13301 13962 13302 13963 13303 13964 13304 13965 2894 13966 13306 13967 69 13968 13308 13969 13309 13970 13310 13971 13311 13972 13312 13973 13313 13974 13314 13975 2 13976 13316 13977 13317 13978 13318 13979 13319 13980 13320 13981 13321 13982 2936 13983 2938 13984 13324 13985 13325 13986 13326 13987 13327 13988 13328 13989 357 13990 357 13991 40 13992 360 13993 434 13994 13334 13995 13335 13996 13336 13997 13337 13998 13338 13999 40 14000 13340 14001 13341 14002 2959 14003 2962 14004 2964 14005 2966 14006 77 14007 2968 14008 2969 14009 13349 14010 282 14011 13351 14012 2 14013 282 14014 77 14015 3931 14016 2828 14017 13357 14018 1862 14019 162 14020 2050 14021 2 14022 13362 14023 2 14024 159 14025 162 14026 13366 14027 13367 14028 13368 14029 13369 14030 13370 14031 943 14032 13372 14033 13373 14034 13374 14035 976 14036 159 14037 162 14038 13378 14039 13379 14040 2 14041 282 14042 13382 14043 2 14044 282 14045 13385 14046 13386 14047 13387 14048 13388 14049 13389 14050 13390 14051 1862 14052 162 14053 13393 14054 13394 14055 13395 14056 13396 14057 282 14058 13398 14059 13399 14060 2 14061 574 14062 13402 14063 5236 14064 13404 14065 13405 14066 14066 14067 40 14068 14068 14069 282 14070 14070 14071 14071 14072 2 14073 574 14074 14074 14075 14075 14076 14076 14077 14077 14078 14078 14079 14079 14080 2 14081 740 14082 2 14083 14083 14084 14084 14085 14085 14086 14086 14087 14087 14088 14088 14089 14089 14090 14090 14091 14091 14092 14092 14093 855 14094 14094 14095 14095 14096 14096 14097 14097 14098 14098 14099 14099 14100 14100 14101 14101 14102 14102 14103 14103 14104 14104 14105 14105 14106 14106 14107 14107 14108 14108 14109 14109 14110 14110 14111 14111 14112 14112 14113 14113 14114 14114 14115 14115 14116 14116 14117 14117 14118 14118 14119 14119 14120 14120 14121 14121 14122 14122 14123 14123 14124 14124 14125 14125 14126 14126 14127 14127 14128 14128 14129 14129 14130 14130 14131 14131 14132 14132 14133 14133 14134 14134 14135 14135 14136 14136 14137 14137 14138 14138 14139 14139 14140 14140 14141 14141 14142 14142 14143 14143 14144 14144 14145 14145 14146 14146 14147 14147 14148 14148 14149 14149 14150 14150 14151 14151 14152 14152 14153 14153 14154 14154 14155 14155 14156 14156 14157 14157 14158 14158 14159 14159 14160 14160 14161 14161 14162 14162 14163 14163 14164 14164 14165 14165 14166 14166 14167 14167 14168 14168 14169 14169 14170 14170 14171 14171 14172 14172 14173 14173 14174 11 14175 14175 14176 14176 14177 14177 14178 2 14179 574 14180 14180 14181 943 14182 14182 14183 14183 14184 976 14185 159 14186 162 14187 14187 14188 14188 14189 1487 14190 2 14191 2 14192 2 14193 14193 14194 2 14195 14195 14196 2 14197 2 14198 2 14199 2 14200 2 14201 2 14202 2 14203 2 14204 2 14205 14205 14206 14206 14207 14207 14208 14208 14209 14209 14210 14210 14211 14211 14212 14212 14213 14213 14214 14214 14215 14215 14216 14216 14217 14217 14218 22 14219 14219 14220 4 14221 11 14222 14222 14223 89 14224 11 14225 14225 14226 99 14227 14227 14228 1551 14229 14229 14230 14230 14231 14231 14232 99 14233 855 14234 11 14235 14235 14236 943 14237 14237 14238 14238 14239 976 14240 159 14241 162 14242 14242 14243 14243 14244 14244 14245 855 14246 11 14247 14247 14248 14248 14249 14249 14250 357 14251 357 14252 360 14253 434 14254 14254 14255 14255 14256 14256 14257 14257 14258 2 14259 574 14260 14260 14261 14261 14262 14262 14263 14263 14264 14264 14265 14265 14266 2 14267 574 14268 14268 14269 14269 14270 14270 14271 140 14272 11 14273 142 14274 162 14275 163 14276 165 14277 167 14278 77 14279 126 14280 170 14281 14281 14282 14282 14283 140 14284 11 14285 142 14286 162 14287 2248 14288 2250 14289 2251 14290 2252 14291 2 14292 2 14293 3 14294 4 14295 2 14296 2 14297 2 14298 2 14299 11 14300 2 14301 2 14302 371 14303 2 14304 373 14305 22 14306 375 14307 376 14308 2 14309 11 14310 11 14311 380 14312 381 14313 382 14314 22 14315 2 14316 2 14317 2 14318 11 14319 388 14320 11 14321 390 14322 11 14323 392 14324 393 14325 394 14326 395 14327 60 14328 397 14329 2 14330 2 14331 400 14332 11 14333 2 14334 2 14335 69 14336 131 14337 406 14338 2 14339 2 14340 409 14341 2 14342 411 14343 412 14344 2 14345 85 14346 86 14347 89 14348 134 14349 11 14350 419 14351 420 14352 99 14353 422 14354 423 14355 14355 14356 104 14357 104 14358 427 14359 428 14360 429 14361 430 14362 431 14363 432 14364 433 14365 435 14366 436 14367 437 14368 438 14369 99 14370 2 14371 441 14372 2 14373 2 14374 2 14375 2 14376 2 14377 2 14378 2 14379 2 14380 307 14381 2 14382 2 14383 2 14384 2 14385 2 14386 313 14387 2 14388 2 14389 2 14390 2 14391 2 14392 2 14393 2 14394 2 14395 2 14396 2 14397 2 14398 2 14399 2 14400 2 14401 2 14402 472 14403 473 14404 474 14405 475 14406 476 14407 477 14408 478 14409 479 14410 480 14411 481 14412 482 14413 483 14414 484 14415 485 14416 486 14417 4 14418 2 14419 489 14420 490 14421 491 14422 492 14423 493 14424 494 14425 495 14426 496 14427 497 14428 498 14429 499 14430 357 14431 357 14432 40 14433 360 14434 434 14435 505 14436 506 14437 507 14438 508 14439 509 14440 2 14441 574 14442 13406 14443 14066 14444 40 14445 14068 14446 282 14447 14070 14448 14071 14449 2 14450 574 14451 14074 14452 14075 14453 14076 14454 14077 14455 14078 14456 14079 14457 2 14458 740 14459 2 14460 14083 14461 14084 14462 14085 14463 14086 14464 14087 14465 14088 14466 14089 14467 14090 14468 14091 14469 14092 14470 855 14471 14094 14472 14095 14473 14096 14474 14097 14475 14098 14476 14099 14477 14100 14478 14101 14479 14102 14480 14103 14481 14104 14482 14105 14483 14106 14484 14107 14485 14108 14486 14109 14487 14110 14488 14111 14489 14112 14490 14113 14491 14114 14492 14115 14493 14116 14494 14117 14495 14118 14496 14119 14497 14120 14498 14121 14499 14122 14500 14123 14501 14124 14502 14125 14503 14126 14504 14127 14505 14128 14506 14129 14507 14130 14508 14131 14509 14132 14510 14133 14511 14134 14512 14135 14513 14136 14514 14137 14515 14138 14516 14139 14517 14140 14518 14141 14519 14142 14520 14143 14521 14144 14522 14145 14523 14146 14524 14147 14525 14148 14526 14149 14527 14150 14528 14151 14529 14152 14530 14153 14531 14154 14532 14155 14533 14156 14534 14157 14535 14158 14536 14159 14537 14160 14538 14161 14539 14162 14540 14163 14541 14164 14542 14165 14543 14166 14544 14167 14545 14168 14546 14169 14547 14170 14548 14171 14549 14172 14550 14173 14551 11 14552 14175 14553 14176 14554 14177 14555 2 14556 574 14557 14180 14558 943 14559 14182 14560 14183 14561 976 14562 159 14563 162 14564 14187 14565 14188 14566 1487 14567 2 14568 2 14569 2 14570 14193 14571 2 14572 14195 14573 2 14574 2 14575 2 14576 2 14577 2 14578 2 14579 2 14580 2 14581 2 14582 14205 14583 14206 14584 14207 14585 14208 14586 14209 14587 14210 14588 14211 14589 14212 14590 14213 14591 14214 14592 14215 14593 14216 14594 14217 14595 22 14596 14219 14597 4 14598 11 14599 14222 14600 89 14601 11 14602 14225 14603 99 14604 14227 14605 1551 14606 14229 14607 14230 14608 14231 14609 99 14610 855 14611 11 14612 14235 14613 943 14614 14237 14615 14238 14616 976 14617 159 14618 162 14619 14242 14620 14243 14621 14244 14622 855 14623 11 14624 14247 14625 14248 14626 14249 14627 357 14628 357 14629 360 14630 434 14631 14254 14632 14255 14633 14256 14634 14257 14635 2 14636 574 14637 14260 14638 14261 14639 14262 14640 14263 14641 14264 14642 14265 14643 2 14644 574 14645 14268 14646 14270 14647 140 14648 11 14649 142 14650 162 14651 163 14652 165 14653 167 14654 77 14655 126 14656 170 14657 14281 14658 14282 14659 140 14660 11 14661 142 14662 162 14663 2248 14664 2250 14665 2251 14666 2252 14667 2 14668 14355 14669 14669 14670 14669 14671 855 14672 14672 14673 11 14674 14674 14675 2 14676 574 14677 14677 14678 14678 14679 14679 14680 2732 14681 14681 14682 14682 14683 357 14684 357 14685 360 14686 434 14687 14687 14688 14688 14689 14689 14690 14690 14691 2 14692 574 14693 14693 14694 14694 14695 14695 14696 14696 14697 2803 14698 14698 14699 434 14700 14700 14701 2 14702 14702 14703 14703 14704 2 14705 3 14706 4 14707 2 14708 2 14709 2 14710 2 14711 11 14712 2 14713 2 14714 14714 14715 2 14716 14716 14717 22 14718 14718 14719 14719 14720 2 14721 11 14722 11 14723 14723 14724 14724 14725 14725 14726 22 14727 2 14728 2 14729 2 14730 11 14731 14731 14732 11 14733 14733 14734 11 14735 14735 14736 14736 14737 14737 14738 14738 14739 60 14740 14740 14741 2 14742 2 14743 14743 14744 11 14745 2 14746 2 14747 69 14748 131 14749 14749 14750 2 14751 2 14752 14752 14753 2 14754 14754 14755 14755 14756 2 14757 85 14758 86 14759 89 14760 134 14761 11 14762 14762 14763 14763 14764 99 14765 14765 14766 14766 14767 14767 14768 104 14769 104 14770 14770 14771 14771 14772 14772 14773 14773 14774 14774 14775 14775 14776 14776 14777 14777 14778 14778 14779 14779 14780 14780 14781 99 14782 2 14783 14783 14784 2 14785 2 14786 2 14787 2 14788 2 14789 2 14790 2 14791 2 14792 307 14793 2 14794 2 14795 2 14796 2 14797 2 14798 313 14799 2 14800 2 14801 2 14802 2 14803 2 14804 2 14805 2 14806 2 14807 2 14808 2 14809 2 14810 2 14811 2 14812 2 14813 2 14814 14814 14815 14815 14816 14816 14817 14817 14818 14818 14819 14819 14820 14820 14821 14821 14822 14822 14823 14823 14824 14824 14825 14825 14826 14826 14827 14827 14828 14828 14829 4 14830 2 14831 14831 14832 14832 14833 14833 14834 14834 14835 14835 14836 14836 14837 14837 14838 14838 14839 14839 14840 14840 14841 14841 14842 357 14843 357 14844 40 14845 360 14846 434 14847 14847 14848 14848 14849 14849 14850 14850 14851 14851 14852 2 14853 574 14854 14854 14855 14855 14856 40 14857 14857 14858 282 14859 14859 14860 14860 14861 2 14862 574 14863 14863 14864 14864 14865 14865 14866 14866 14867 14867 14868 14868 14869 2 14870 740 14871 2 14872 14872 14873 14873 14874 14874 14875 14875 14876 14876 14877 14877 14878 14878 14879 14879 14880 14880 14881 14881 14882 855 14883 14883 14884 14884 14885 14885 14886 14886 14887 14887 14888 14888 14889 14889 14890 14890 14891 14891 14892 14892 14893 14893 14894 14894 14895 14895 14896 14896 14897 14897 14898 14898 14899 14899 14900 14900 14901 14901 14902 14902 14903 14903 14904 14904 14905 14905 14906 14906 14907 14907 14908 14908 14909 14909 14910 14910 14911 14911 14912 14912 14913 14913 14914 14914 14915 14915 14916 14916 14917 14917 14918 14918 14919 14919 14920 14920 14921 14921 14922 14922 14923 14923 14924 14924 14925 14925 14926 14926 14927 14927 14928 14928 14929 14929 14930 14930 14931 14931 14932 14932 14933 14933 14934 14934 14935 14935 14936 14936 14937 14937 14938 14938 14939 14939 14940 14940 14941 14941 14942 14942 14943 14943 14944 14944 14945 14945 14946 14946 14947 14947 14948 14948 14949 14949 14950 14950 14951 14951 14952 14952 14953 14953 14954 14954 14955 14955 14956 14956 14957 14957 14958 14958 14959 14959 14960 14960 14961 14961 14962 14962 14963 11 14964 14964 14965 14965 14966 14966 14967 2 14968 574 14969 14969 14970 943 14971 14971 14972 14972 14973 976 14974 159 14975 162 14976 14976 14977 14977 14978 1487 14979 2 14980 2 14981 2 14982 14982 14983 2 14984 14984 14985 2 14986 2 14987 2 14988 2 14989 2 14990 2 14991 2 14992 2 14993 2 14994 14994 14995 14995 14996 14996 14997 14997 14998 14998 14999 14999 15000 15000 15001 15001 15002 15002 15003 15003 15004 15004 15005 15005 15006 15006 15007 22 15008 15008 15009 4 15010 11 15011 15011 15012 89 15013 11 15014 15014 15015 99 15016 15016 15017 1551 15018 15018 15019 15019 15020 15020 15021 99 15022 855 15023 11 15024 15024 15025 943 15026 15026 15027 15027 15028 976 15029 159 15030 162 15031 15031 15032 15032 15033 15033 15034 855 15035 11 15036 15036 15037 15037 15038 15038 15039 357 15040 357 15041 360 15042 434 15043 15043 15044 15044 15045 15045 15046 15046 15047 2 15048 574 15049 15049 15050 15050 15051 15051 15052 15052 15053 15053 15054 15054 15055 2 15056 574 15057 15057 15058 15058 15059 140 15060 11 15061 142 15062 162 15063 163 15064 165 15065 167 15066 77 15067 126 15068 170 15069 15069 15070 15070 15071 140 15072 11 15073 142 15074 162 15075 2248 15076 2250 15077 2251 15078 2252 15079 2 15080 15080 15081 15081 15082 2 15083 15083 15084 855 15085 15085 15086 11 15087 15087 15088 2 15089 574 15090 15090 15091 15091 15092 15092 15093 2732 15094 15094 15095 15095 15096 357 15097 357 15098 360 15099 434 15100 15100 15101 15101 15102 15102 15103 15103 15104 2 15105 574 15106 15106 15107 15107 15108 15108 15109 15109 15110 2803 15111 15111 15112 434 15113 15113 15114 2 15115 15115 15116 15116 15117 14767 15118 2 15119 15119 15120 15120 15121 15121 15122 15122 15123 15123 15124 2894 15125 15125 15126 69 15127 15127 15128 15128 15129 15129 15130 15130 15131 15131 15132 15132 15133 15133 15134 2 15135 15135 15136 15136 15137 15137 15138 15138 15139 15139 15140 15140 15141 2936 15142 2938 15143 15143 15144 15144 15145 15145 15146 15146 15147 15147 15148 357 15149 357 15150 40 15151 360 15152 434 15153 15153 15154 15154 15155 15155 15156 15156 15157 15157 15158 40 15159 15159 15160 15160 15161 2959 15162 2962 15163 2964 15164 2966 15165 77 15166 2968 15167 2969 15168 15168 15169 282 15170 15170 15171 2 15172 282 15173 77 15174 3931 15175 2828 15176 15176 15177 1862 15178 162 15179 2050 15180 2 15181 15181 15182 2 15183 159 15184 162 15185 15185 15186 15186 15187 15187 15188 15188 15189 15189 15190 943 15191 15191 15192 15192 15193 15193 15194 976 15195 159 15196 162 15197 15197 15198 15198 15199 2 15200 282 15201 15201 15202 2 15203 282 15204 15204 15205 15205 15206 15206 15207 15207 15208 15208 15209 15209 15210 1862 15211 162 15212 15212 15213 15213 15214 15214 15215 15215 15216 282 15217 15217 15218 15218 15219 2 15220 574 15221 15221 15222 5236 15223 15223 15224 15224 15225 15225 15226 15226 15227 15227 15228 15228 15229 2894 15230 15230 15231 69 15232 15232 15233 15233 15234 15234 15235 15235 15236 15236 15237 15237 15238 15238 15239 2 15240 15240 15241 15241 15242 15242 15243 15243 15244 15244 15245 15245 15246 2936 15247 2938 15248 15248 15249 15249 15250 15250 15251 15251 15252 15252 15253 357 15254 357 15255 40 15256 360 15257 434 15258 15258 15259 15259 15260 15260 15261 15261 15262 15262 15263 40 15264 15264 15265 15265 15266 2959 15267 2962 15268 2964 15269 2966 15270 77 15271 2968 15272 2969 15273 15273 15274 282 15275 15275 15276 2 15277 282 15278 77 15279 3931 15280 2828 15281 15281 15282 1862 15283 162 15284 2050 15285 2 15286 15286 15287 2 15288 159 15289 162 15290 15290 15291 15291 15292 15292 15293 15293 15294 15294 15295 943 15296 15296 15297 15297 15298 15298 15299 976 15300 159 15301 162 15302 15302 15303 15303 15304 2 15305 282 15306 15306 15307 2 15308 282 15309 15309 15310 15310 15311 15311 15312 15312 15313 15313 15314 15314 15315 1862 15316 162 15317 15317 15318 15318 15319 15319 15320 15320 15321 282 15322 15322 15323 15323 15324 2 15325 574 15326 15326 15327 5236 15328 15328 15329 15329 15330 855 15331 14672 15332 11 15333 14674 15334 2 15335 574 15336 14677 15337 14678 15338 14679 15339 2732 15340 14681 15341 14682 15342 357 15343 357 15344 360 15345 434 15346 14687 15347 14688 15348 14689 15349 14690 15350 2 15351 574 15352 14693 15353 14694 15354 14695 15355 14696 15356 2803 15357 14698 15358 434 15359 14700 15360 2 15361 14702 15362 14703 15363 2 15364 3 15365 4 15366 2 15367 2 15368 2 15369 2 15370 11 15371 2 15372 2 15373 14714 15374 2 15375 14716 15376 22 15377 14718 15378 14719 15379 2 15380 11 15381 11 15382 14723 15383 14724 15384 14725 15385 22 15386 2 15387 2 15388 2 15389 11 15390 14731 15391 11 15392 14733 15393 11 15394 14735 15395 14736 15396 14737 15397 14738 15398 60 15399 14740 15400 2 15401 2 15402 14743 15403 11 15404 2 15405 2 15406 69 15407 131 15408 14749 15409 2 15410 2 15411 14752 15412 2 15413 14754 15414 14755 15415 2 15416 85 15417 86 15418 89 15419 134 15420 11 15421 14762 15422 14763 15423 99 15424 14765 15425 14766 15426 14767 15427 104 15428 104 15429 14770 15430 14771 15431 14772 15432 14773 15433 14774 15434 14775 15435 14776 15436 14777 15437 14778 15438 14779 15439 14780 15440 99 15441 2 15442 14783 15443 2 15444 2 15445 2 15446 2 15447 2 15448 2 15449 2 15450 2 15451 307 15452 2 15453 2 15454 2 15455 2 15456 2 15457 313 15458 2 15459 2 15460 2 15461 2 15462 2 15463 2 15464 2 15465 2 15466 2 15467 2 15468 2 15469 2 15470 2 15471 2 15472 2 15473 14814 15474 14815 15475 14816 15476 14817 15477 14818 15478 14819 15479 14820 15480 14821 15481 14822 15482 14823 15483 14824 15484 14825 15485 14826 15486 14827 15487 14828 15488 4 15489 2 15490 14831 15491 14832 15492 14833 15493 14834 15494 14835 15495 14836 15496 14837 15497 14838 15498 14839 15499 14840 15500 14841 15501 357 15502 357 15503 40 15504 360 15505 434 15506 14847 15507 14848 15508 14849 15509 14850 15510 14851 15511 2 15512 574 15513 14854 15514 14855 15515 40 15516 14857 15517 282 15518 14859 15519 14860 15520 2 15521 574 15522 14863 15523 14864 15524 14865 15525 14866 15526 14867 15527 14868 15528 2 15529 740 15530 2 15531 14872 15532 14873 15533 14874 15534 14875 15535 14876 15536 14877 15537 14878 15538 14879 15539 14880 15540 14881 15541 855 15542 14883 15543 14884 15544 14885 15545 14886 15546 14887 15547 14888 15548 14889 15549 14890 15550 14891 15551 14892 15552 14893 15553 14894 15554 14895 15555 14896 15556 14897 15557 14898 15558 14899 15559 14900 15560 14901 15561 14902 15562 14903 15563 14904 15564 14905 15565 14906 15566 14907 15567 14908 15568 14909 15569 14910 15570 14911 15571 14912 15572 14913 15573 14914 15574 14915 15575 14916 15576 14917 15577 14918 15578 14919 15579 14920 15580 14921 15581 14922 15582 14923 15583 14924 15584 14925 15585 14926 15586 14927 15587 14928 15588 14929 15589 14930 15590 14931 15591 14932 15592 14933 15593 14934 15594 14935 15595 14936 15596 14937 15597 14938 15598 14939 15599 14940 15600 14941 15601 14942 15602 14943 15603 14944 15604 14945 15605 14946 15606 14947 15607 14948 15608 14949 15609 14950 15610 14951 15611 14952 15612 14953 15613 14954 15614 14955 15615 14956 15616 14957 15617 14958 15618 14959 15619 14960 15620 14961 15621 14962 15622 11 15623 14964 15624 14965 15625 14966 15626 2 15627 574 15628 14969 15629 943 15630 14971 15631 14972 15632 976 15633 159 15634 162 15635 14976 15636 14977 15637 1487 15638 2 15639 2 15640 2 15641 14982 15642 2 15643 14984 15644 2 15645 2 15646 2 15647 2 15648 2 15649 2 15650 2 15651 2 15652 2 15653 14994 15654 14995 15655 14996 15656 14997 15657 14998 15658 14999 15659 15000 15660 15001 15661 15002 15662 15003 15663 15004 15664 15005 15665 15006 15666 22 15667 15008 15668 4 15669 11 15670 15011 15671 89 15672 11 15673 15014 15674 99 15675 15016 15676 1551 15677 15018 15678 15019 15679 15020 15680 99 15681 855 15682 11 15683 15024 15684 943 15685 15026 15686 15027 15687 976 15688 159 15689 162 15690 15031 15691 15032 15692 15033 15693 855 15694 11 15695 15036 15696 15037 15697 15038 15698 357 15699 357 15700 360 15701 434 15702 15043 15703 15044 15704 15045 15705 15046 15706 2 15707 574 15708 15049 15709 15050 15710 15051 15711 15052 15712 15053 15713 15054 15714 2 15715 574 15716 15057 15717 15058 15718 140 15719 11 15720 142 15721 162 15722 163 15723 165 15724 167 15725 77 15726 126 15727 170 15728 15069 15729 15070 15730 140 15731 11 15732 142 15733 162 15734 2248 15735 2250 15736 2251 15737 2252 15738 2 15739 15080 15740 15081 15741 2 15742 15083 15743 855 15744 15085 15745 11 15746 15087 15747 2 15748 574 15749 15090 15750 15091 15751 15092 15752 2732 15753 15094 15754 15095 15755 357 15756 357 15757 360 15758 434 15759 15100 15760 15101 15761 15102 15762 15103 15763 2 15764 574 15765 15106 15766 15107 15767 15108 15768 15109 15769 2803 15770 15111 15771 434 15772 15113 15773 2 15774 15115 15775 15116 15776 14767 15777 2 15778 15119 15779 15120 15780 15121 15781 15122 15782 15123 15783 2894 15784 15125 15785 69 15786 15127 15787 15128 15788 15129 15789 15130 15790 15131 15791 15132 15792 15133 15793 2 15794 15135 15795 15136 15796 15137 15797 15138 15798 15139 15799 15140 15800 2936 15801 2938 15802 15143 15803 15144 15804 15145 15805 15146 15806 15147 15807 357 15808 357 15809 40 15810 360 15811 434 15812 15153 15813 15154 15814 15155 15815 15156 15816 15157 15817 40 15818 15159 15819 15160 15820 2959 15821 2962 15822 2964 15823 2966 15824 77 15825 2968 15826 2969 15827 15168 15828 282 15829 15170 15830 2 15831 282 15832 77 15833 3931 15834 2828 15835 15176 15836 1862 15837 162 15838 2050 15839 2 15840 15181 15841 2 15842 159 15843 162 15844 15185 15845 15186 15846 15187 15847 15188 15848 15189 15849 943 15850 15191 15851 15192 15852 15193 15853 976 15854 159 15855 162 15856 15197 15857 15198 15858 2 15859 282 15860 15201 15861 2 15862 282 15863 15204 15864 15205 15865 15206 15866 15207 15867 15208 15868 15209 15869 1862 15870 162 15871 15212 15872 15213 15873 15214 15874 15215 15875 282 15876 15217 15877 15218 15878 2 15879 574 15880 15221 15881 5236 15882 15223 15883 15224 15884 15225 15885 15226 15886 15227 15887 15228 15888 2894 15889 15230 15890 69 15891 15232 15892 15233 15893 15234 15894 15235 15895 15236 15896 15237 15897 15238 15898 2 15899 15240 15900 15241 15901 15242 15902 15243 15903 15244 15904 15245 15905 2936 15906 2938 15907 15248 15908 15249 15909 15250 15910 15251 15911 15252 15912 357 15913 357 15914 40 15915 360 15916 434 15917 15258 15918 15259 15919 15260 15920 15261 15921 15262 15922 40 15923 15264 15924 15265 15925 2959 15926 2962 15927 2964 15928 2966 15929 77 15930 2968 15931 2969 15932 15273 15933 282 15934 15275 15935 2 15936 282 15937 77 15938 3931 15939 2828 15940 15281 15941 1862 15942 162 15943 2050 15944 2 15945 15286 15946 2 15947 159 15948 162 15949 15290 15950 15291 15951 15292 15952 15293 15953 15294 15954 943 15955 15296 15956 15297 15957 15298 15958 976 15959 159 15960 162 15961 15302 15962 15303 15963 2 15964 282 15965 15306 15966 2 15967 282 15968 15309 15969 15310 15970 15311 15971 15312 15972 15313 15973 15314 15974 1862 15975 162 15976 15317 15977 15318 15978 15319 15979 15320 15980 282 15981 15322 15982 15323 15983 2 15984 574 15985 15326 15986 5236 15987 15328 15988 15329 15989 15989 15990 15990 15991 15991 15992 15992 15993 2 15994 574 15995 15995 15996 15996 15997 40 15998 15998 15999 282 16000 16000 16001 16001 16002 2 16003 574 16004 16004 16005 16005 16006 16006 16007 16007 16008 16008 16009 16009 16010 2 16011 740 16012 2 16013 16013 16014 16014 16015 16015 16016 16016 16017 16017 16018 16018 16019 16019 16020 16020 16021 16021 16022 16022 16023 855 16024 16024 16025 16025 16026 16026 16027 16027 16028 16028 16029 16029 16030 16030 16031 16031 16032 16032 16033 16033 16034 16034 16035 16035 16036 16036 16037 16037 16038 16038 16039 16039 16040 16040 16041 16041 16042 16042 16043 16043 16044 16044 16045 16045 16046 16046 16047 16047 16048 16048 16049 16049 16050 16050 16051 16051 16052 16052 16053 16053 16054 16054 16055 16055 16056 16056 16057 16057 16058 16058 16059 16059 16060 16060 16061 16061 16062 16062 16063 16063 16064 16064 16065 16065 16066 16066 16067 16067 16068 16068 16069 16069 16070 16070 16071 16071 16072 16072 16073 16073 16074 16074 16075 16075 16076 16076 16077 16077 16078 16078 16079 16079 16080 16080 16081 16081 16082 16082 16083 16083 16084 16084 16085 16085 16086 16086 16087 16087 16088 16088 16089 16089 16090 16090 16091 16091 16092 16092 16093 16093 16094 16094 16095 16095 16096 16096 16097 16097 16098 16098 16099 16099 16100 16100 16101 16101 16102 16102 16103 16103 16104 11 16105 16105 16106 16106 16107 16107 16108 2 16109 574 16110 16110 16111 943 16112 16112 16113 16113 16114 976 16115 159 16116 162 16117 16117 16118 16118 16119 1487 16120 2 16121 2 16122 2 16123 16123 16124 2 16125 16125 16126 2 16127 2 16128 2 16129 2 16130 2 16131 2 16132 2 16133 2 16134 2 16135 16135 16136 16136 16137 16137 16138 16138 16139 16139 16140 16140 16141 16141 16142 16142 16143 16143 16144 16144 16145 16145 16146 16146 16147 16147 16148 22 16149 16149 16150 4 16151 11 16152 16152 16153 89 16154 11 16155 16155 16156 99 16157 16157 16158 1551 16159 16159 16160 16160 16161 16161 16162 99 16163 855 16164 11 16165 16165 16166 943 16167 16167 16168 16168 16169 976 16170 159 16171 162 16172 16172 16173 16173 16174 16174 16175 855 16176 11 16177 16177 16178 16178 16179 16179 16180 357 16181 357 16182 360 16183 434 16184 16184 16185 16185 16186 16186 16187 16187 16188 2 16189 574 16190 16190 16191 16191 16192 16192 16193 16193 16194 16194 16195 16195 16196 2 16197 574 16198 16198 16199 16199 16200 140 16201 11 16202 142 16203 162 16204 163 16205 165 16206 167 16207 77 16208 126 16209 170 16210 16210 16211 16211 16212 140 16213 11 16214 142 16215 162 16216 2248 16217 2250 16218 2251 16219 2252 16220 2 16221 2 16222 3 16223 4 16224 2 16225 2 16226 2 16227 2 16228 11 16229 2 16230 2 16231 229 16232 2 16233 231 16234 22 16235 233 16236 234 16237 2 16238 11 16239 11 16240 238 16241 239 16242 240 16243 22 16244 2 16245 2 16246 2 16247 11 16248 246 16249 11 16250 248 16251 11 16252 250 16253 251 16254 252 16255 253 16256 60 16257 255 16258 2 16259 2 16260 258 16261 11 16262 2 16263 2 16264 69 16265 131 16266 264 16267 2 16268 2 16269 267 16270 2 16271 269 16272 270 16273 2 16274 85 16275 86 16276 89 16277 134 16278 11 16279 277 16280 278 16281 99 16282 280 16283 281 16284 16284 16285 104 16286 104 16287 285 16288 286 16289 287 16290 288 16291 289 16292 290 16293 291 16294 292 16295 293 16296 294 16297 295 16298 99 16299 2 16300 298 16301 2 16302 2 16303 2 16304 2 16305 2 16306 2 16307 2 16308 2 16309 307 16310 2 16311 2 16312 2 16313 2 16314 2 16315 313 16316 2 16317 2 16318 2 16319 2 16320 2 16321 2 16322 2 16323 2 16324 2 16325 2 16326 2 16327 2 16328 2 16329 2 16330 2 16331 329 16332 330 16333 331 16334 332 16335 333 16336 334 16337 335 16338 336 16339 337 16340 338 16341 339 16342 340 16343 341 16344 342 16345 343 16346 4 16347 2 16348 346 16349 347 16350 348 16351 349 16352 350 16353 351 16354 352 16355 353 16356 354 16357 355 16358 356 16359 357 16360 357 16361 40 16362 360 16363 434 16364 14269 16365 15989 16366 15990 16367 15991 16368 15992 16369 2 16370 574 16371 15995 16372 15996 16373 40 16374 15998 16375 282 16376 16000 16377 16001 16378 2 16379 574 16380 16004 16381 16005 16382 16006 16383 16007 16384 16008 16385 16009 16386 2 16387 740 16388 2 16389 16013 16390 16014 16391 16015 16392 16016 16393 16017 16394 16018 16395 16019 16396 16020 16397 16021 16398 16022 16399 855 16400 16024 16401 16025 16402 16026 16403 16027 16404 16028 16405 16029 16406 16030 16407 16031 16408 16032 16409 16033 16410 16034 16411 16035 16412 16036 16413 16037 16414 16038 16415 16039 16416 16040 16417 16041 16418 16042 16419 16043 16420 16044 16421 16045 16422 16046 16423 16047 16424 16048 16425 16049 16426 16050 16427 16051 16428 16052 16429 16053 16430 16054 16431 16055 16432 16056 16433 16057 16434 16058 16435 16059 16436 16060 16437 16061 16438 16062 16439 16063 16440 16064 16441 16065 16442 16066 16443 16067 16444 16068 16445 16069 16446 16070 16447 16071 16448 16072 16449 16073 16450 16074 16451 16075 16452 16076 16453 16077 16454 16078 16455 16079 16456 16080 16457 16081 16458 16082 16459 16083 16460 16084 16461 16085 16462 16086 16463 16087 16464 16088 16465 16089 16466 16090 16467 16091 16468 16092 16469 16093 16470 16094 16471 16095 16472 16096 16473 16097 16474 16098 16475 16099 16476 16100 16477 16101 16478 16102 16479 16103 16480 11 16481 16105 16482 16106 16483 16107 16484 2 16485 574 16486 16110 16487 943 16488 16112 16489 16113 16490 976 16491 159 16492 162 16493 16117 16494 16118 16495 1487 16496 2 16497 2 16498 2 16499 16123 16500 2 16501 16125 16502 2 16503 2 16504 2 16505 2 16506 2 16507 2 16508 2 16509 2 16510 2 16511 16135 16512 16136 16513 16137 16514 16138 16515 16139 16516 16140 16517 16141 16518 16142 16519 16143 16520 16144 16521 16145 16522 16146 16523 16147 16524 22 16525 16149 16526 4 16527 11 16528 16152 16529 89 16530 11 16531 16155 16532 99 16533 16157 16534 1551 16535 16159 16536 16160 16537 16161 16538 99 16539 855 16540 11 16541 16165 16542 943 16543 16167 16544 16168 16545 976 16546 159 16547 162 16548 16172 16549 16173 16550 16174 16551 855 16552 11 16553 16177 16554 16178 16555 16179 16556 357 16557 357 16558 360 16559 434 16560 16184 16561 16185 16562 16186 16563 16187 16564 2 16565 574 16566 16190 16567 16191 16568 16192 16569 16193 16570 16194 16571 16195 16572 2 16573 574 16574 16198 16575 16199 16576 140 16577 11 16578 142 16579 162 16580 163 16581 165 16582 167 16583 77 16584 126 16585 170 16586 16210 16587 16211 16588 140 16589 11 16590 142 16591 162 16592 2248 16593 2250 16594 2251 16595 2252 16596 2 16597 16284 16598 16598 16599 16598 16600 855 16601 16601 16602 11 16603 16603 16604 2 16605 574 16606 16606 16607 16607 16608 16608 16609 2732 16610 16610 16611 16611 16612 357 16613 357 16614 360 16615 434 16616 16616 16617 16617 16618 16618 16619 16619 16620 2 16621 574 16622 16622 16623 16623 16624 16624 16625 16625 16626 2803 16627 16627 16628 434 16629 16629 16630 2 16631 16631 16632 16632 16633 2 16634 3 16635 4 16636 2 16637 2 16638 2 16639 2 16640 11 16641 2 16642 2 16643 16643 16644 2 16645 16645 16646 22 16647 16647 16648 16648 16649 2 16650 11 16651 11 16652 16652 16653 16653 16654 16654 16655 22 16656 2 16657 2 16658 2 16659 11 16660 16660 16661 11 16662 16662 16663 11 16664 16664 16665 16665 16666 16666 16667 16667 16668 60 16669 16669 16670 2 16671 2 16672 16672 16673 11 16674 2 16675 2 16676 69 16677 131 16678 16678 16679 2 16680 2 16681 16681 16682 2 16683 16683 16684 16684 16685 2 16686 85 16687 86 16688 89 16689 134 16690 11 16691 16691 16692 16692 16693 99 16694 16694 16695 16695 16696 16696 16697 104 16698 104 16699 16699 16700 16700 16701 16701 16702 16702 16703 16703 16704 16704 16705 16705 16706 16706 16707 16707 16708 16708 16709 16709 16710 99 16711 2 16712 16712 16713 2 16714 2 16715 2 16716 2 16717 2 16718 2 16719 2 16720 2 16721 307 16722 2 16723 2 16724 2 16725 2 16726 2 16727 313 16728 2 16729 2 16730 2 16731 2 16732 2 16733 2 16734 2 16735 2 16736 2 16737 2 16738 2 16739 2 16740 2 16741 2 16742 2 16743 16743 16744 16744 16745 16745 16746 16746 16747 16747 16748 16748 16749 16749 16750 16750 16751 16751 16752 16752 16753 16753 16754 16754 16755 16755 16756 16756 16757 16757 16758 4 16759 2 16760 16760 16761 16761 16762 16762 16763 16763 16764 16764 16765 16765 16766 16766 16767 16767 16768 16768 16769 16769 16770 16770 16771 357 16772 357 16773 40 16774 360 16775 434 16776 16776 16777 16777 16778 16778 16779 16779 16780 16780 16781 2 16782 574 16783 16783 16784 16784 16785 40 16786 16786 16787 282 16788 16788 16789 16789 16790 2 16791 574 16792 16792 16793 16793 16794 16794 16795 16795 16796 16796 16797 16797 16798 2 16799 740 16800 2 16801 16801 16802 16802 16803 16803 16804 16804 16805 16805 16806 16806 16807 16807 16808 16808 16809 16809 16810 16810 16811 855 16812 16812 16813 16813 16814 16814 16815 16815 16816 16816 16817 16817 16818 16818 16819 16819 16820 16820 16821 16821 16822 16822 16823 16823 16824 16824 16825 16825 16826 16826 16827 16827 16828 16828 16829 16829 16830 16830 16831 16831 16832 16832 16833 16833 16834 16834 16835 16835 16836 16836 16837 16837 16838 16838 16839 16839 16840 16840 16841 16841 16842 16842 16843 16843 16844 16844 16845 16845 16846 16846 16847 16847 16848 16848 16849 16849 16850 16850 16851 16851 16852 16852 16853 16853 16854 16854 16855 16855 16856 16856 16857 16857 16858 16858 16859 16859 16860 16860 16861 16861 16862 16862 16863 16863 16864 16864 16865 16865 16866 16866 16867 16867 16868 16868 16869 16869 16870 16870 16871 16871 16872 16872 16873 16873 16874 16874 16875 16875 16876 16876 16877 16877 16878 16878 16879 16879 16880 16880 16881 16881 16882 16882 16883 16883 16884 16884 16885 16885 16886 16886 16887 16887 16888 16888 16889 16889 16890 16890 16891 16891 16892 11 16893 16893 16894 16894 16895 16895 16896 2 16897 574 16898 16898 16899 943 16900 16900 16901 16901 16902 976 16903 159 16904 162 16905 16905 16906 16906 16907 1487 16908 2 16909 2 16910 2 16911 16911 16912 2 16913 16913 16914 2 16915 2 16916 2 16917 2 16918 2 16919 2 16920 2 16921 2 16922 2 16923 16923 16924 16924 16925 16925 16926 16926 16927 16927 16928 16928 16929 16929 16930 16930 16931 16931 16932 16932 16933 16933 16934 16934 16935 16935 16936 22 16937 16937 16938 4 16939 11 16940 16940 16941 89 16942 11 16943 16943 16944 99 16945 16945 16946 1551 16947 16947 16948 16948 16949 16949 16950 99 16951 855 16952 11 16953 16953 16954 943 16955 16955 16956 16956 16957 976 16958 159 16959 162 16960 16960 16961 16961 16962 16962 16963 855 16964 11 16965 16965 16966 16966 16967 16967 16968 357 16969 357 16970 360 16971 434 16972 16972 16973 16973 16974 16974 16975 16975 16976 2 16977 574 16978 16978 16979 16979 16980 16980 16981 16981 16982 16982 16983 16983 16984 2 16985 574 16986 16986 16987 16987 16988 140 16989 11 16990 142 16991 162 16992 163 16993 165 16994 167 16995 77 16996 126 16997 170 16998 16998 16999 16999 17000 140 17001 11 17002 142 17003 162 17004 2248 17005 2250 17006 2251 17007 2252 17008 2 17009 17009 17010 17010 17011 2 17012 17012 17013 855 17014 17014 17015 11 17016 17016 17017 2 17018 574 17019 17019 17020 17020 17021 17021 17022 2732 17023 17023 17024 17024 17025 357 17026 357 17027 360 17028 434 17029 17029 17030 17030 17031 17031 17032 17032 17033 2 17034 574 17035 17035 17036 17036 17037 17037 17038 17038 17039 2803 17040 17040 17041 434 17042 17042 17043 2 17044 17044 17045 17045 17046 16696 17047 2 17048 17048 17049 17049 17050 17050 17051 17051 17052 17052 17053 2894 17054 17054 17055 69 17056 17056 17057 17057 17058 17058 17059 17059 17060 17060 17061 17061 17062 17062 17063 2 17064 17064 17065 17065 17066 17066 17067 17067 17068 17068 17069 17069 17070 2936 17071 2938 17072 17072 17073 17073 17074 17074 17075 17075 17076 17076 17077 357 17078 357 17079 40 17080 360 17081 434 17082 17082 17083 17083 17084 17084 17085 17085 17086 17086 17087 40 17088 17088 17089 17089 17090 2959 17091 2962 17092 2964 17093 2966 17094 77 17095 2968 17096 2969 17097 17097 17098 282 17099 17099 17100 2 17101 282 17102 77 17103 3931 17104 2828 17105 17105 17106 1862 17107 162 17108 2050 17109 2 17110 17110 17111 2 17112 159 17113 162 17114 17114 17115 17115 17116 17116 17117 17117 17118 17118 17119 943 17120 17120 17121 17121 17122 17122 17123 976 17124 159 17125 162 17126 17126 17127 17127 17128 2 17129 282 17130 17130 17131 2 17132 282 17133 17133 17134 17134 17135 17135 17136 17136 17137 17137 17138 17138 17139 1862 17140 162 17141 17141 17142 17142 17143 17143 17144 17144 17145 282 17146 17146 17147 17147 17148 2 17149 574 17150 17150 17151 5236 17152 17152 17153 17153 17154 17154 17155 17155 17156 17156 17157 17157 17158 2894 17159 17159 17160 69 17161 17161 17162 17162 17163 17163 17164 17164 17165 17165 17166 17166 17167 17167 17168 2 17169 17169 17170 17170 17171 17171 17172 17172 17173 17173 17174 17174 17175 2936 17176 2938 17177 17177 17178 17178 17179 17179 17180 17180 17181 17181 17182 357 17183 357 17184 40 17185 360 17186 434 17187 17187 17188 17188 17189 17189 17190 17190 17191 17191 17192 40 17193 17193 17194 17194 17195 2959 17196 2962 17197 2964 17198 2966 17199 77 17200 2968 17201 2969 17202 17202 17203 282 17204 17204 17205 2 17206 282 17207 77 17208 3931 17209 2828 17210 17210 17211 1862 17212 162 17213 2050 17214 2 17215 17215 17216 2 17217 159 17218 162 17219 17219 17220 17220 17221 17221 17222 17222 17223 17223 17224 943 17225 17225 17226 17226 17227 17227 17228 976 17229 159 17230 162 17231 17231 17232 17232 17233 2 17234 282 17235 17235 17236 2 17237 282 17238 17238 17239 17239 17240 17240 17241 17241 17242 17242 17243 17243 17244 1862 17245 162 17246 17246 17247 17247 17248 17248 17249 17249 17250 282 17251 17251 17252 17252 17253 2 17254 574 17255 17255 17256 5236 17257 17257 17258 17258 17259 26 17260 17260 17261 15 17262 15 17263 11 17264 26 17265 2 17266 11 17267 11 17268 40 17269 16 17270 17270 17271 15 17272 70 17273 71 17274 11 17275 77 17276 11 17277 11 17278 17278 17279 88 17280 103 17281 17281 17282 206 17283 17283 17284 126 17285 170 17286 17286 17287 26 17288 17288 17289 17289 17290 1066 17291 11 17292 7315 17293 17293 17294 2 17295 282 17296 1066 17297 17297 17298 17298 17299 17299 17300 15 17301 16 17302 15 17303 11 17304 26 17305 2 17306 16 17307 11 17308 11 17309 40 17310 16 17311 11 17312 11 17313 17313 17314 16 17315 15 17316 70 17317 71 17318 11 17319 16 17320 77 17321 11 17322 11 17323 17323 17324 88 17325 17325 17326 92 17327 17327 17328 98 17329 17329 17330 103 17331 106 17332 17332 17333 17333 17334 17334 17335 17335 17336 17336 17337 17337 17338 17338 17339 126 17340 170 17341 17341 17342 26 17343 11 17344 2 17345 282 17346 26 17347 17347 17348 17348 17349 17349 17350 17350 17351 17351 17352 17352 17353 17353 17354 17354 17355 140 17356 11 17357 142 17358 162 17359 165 17360 167 17361 77 17362 126 17363 170 17364 17364 17365 17365 17366 17366 17367 16 17368 17368 17369 2962 17370 2964 17371 2966 17372 77 17373 2968 17374 2969 17375 17375 17376 2732 17377 17377 17378 17378 17379 17379 17380 17380 17381 1066 17382 7315 17383 17383 17384 17384 17385 17385 17386 17386 17387 15 17388 16 17389 15 17390 11 17391 26 17392 2 17393 7315 17394 16 17395 11 17396 11 17397 40 17398 16 17399 11 17400 17400 17401 16 17402 15 17403 70 17404 71 17405 11 17406 16 17407 77 17408 11 17409 11 17410 17410 17411 88 17412 17412 17413 92 17414 17414 17415 98 17416 17416 17417 103 17418 106 17419 17419 17420 17420 17421 17334 17422 17422 17423 17423 17424 17424 17425 17425 17426 126 17427 170 17428 17428 17429 26 17430 17430 17431 17352 17432 140 17433 11 17434 142 17435 162 17436 165 17437 167 17438 77 17439 126 17440 170 17441 17441 17442 17442 17443 17443 17444 16 17445 2962 17446 2964 17447 2966 17448 77 17449 2968 17450 2969 17451 17451 17452 17452 17453 1066 17454 7315 17455 17455 17456 17384 17457 17385 17458 7315 17459 17459 17460 2 17461 282 17462 17462 17463 17463 17464 17464 17465 17465 17466 17466 17467 17467 17468 15 17469 16 17470 15 17471 11 17472 26 17473 2 17474 7315 17475 16 17476 11 17477 11 17478 40 17479 16 17480 11 17481 11 17482 17482 17483 16 17484 15 17485 70 17486 71 17487 11 17488 16 17489 77 17490 11 17491 11 17492 17492 17493 88 17494 17494 17495 92 17496 17496 17497 98 17498 17498 17499 103 17500 106 17501 17501 17502 17502 17503 17503 17504 17504 17505 17505 17506 17506 17507 17507 17508 126 17509 170 17510 17510 17511 26 17512 11 17513 2 17514 282 17515 26 17516 17516 17517 15 17518 15 17519 11 17520 26 17521 2 17522 11 17523 11 17524 40 17525 16 17526 17526 17527 15 17528 70 17529 71 17530 11 17531 77 17532 11 17533 11 17534 17534 17535 88 17536 103 17537 17537 17538 17503 17539 17539 17540 126 17541 170 17542 17542 17543 26 17544 17544 17545 17545 17546 1066 17547 7315 17548 17548 17549 17384 17550 17385 17551 7315 17552 17552 17553 2 17554 282 17555 17462 17556 17463 17557 17557 17558 17558 17559 17466 17560 7315 17561 17561 17562 22 17563 4 17564 11 17565 2752 17566 89 17567 11 17568 2755 17569 99 17570 2757 17571 1551 17572 2759 17573 2760 17574 17574 17575 159 17576 17576 17577 99 17578 855 17579 11 17580 2765 17581 943 17582 2767 17583 2768 17584 976 17585 159 17586 162 17587 2772 17588 2773 17589 17589 17590 162 17591 7315 17592 17592 17593 17593 17594 2774 17595 855 17596 11 17597 2777 17598 2778 17599 2779 17600 357 17601 357 17602 360 17603 434 17604 2784 17605 2785 17606 2786 17607 2787 17608 2 17609 574 17610 2790 17611 2791 17612 2792 17613 2793 17614 2794 17615 2795 17616 2 17617 574 17618 2798 17619 126 17620 170 17621 17621 17622 17622 17623 11 17624 77 17625 17625 17626 126 17627 170 17628 17628 17629 13 17630 17630 17631 17631 17632 17632 17633 17633 17634 17634 17635 17635 17636 17636 17637 17637 17638 17638 17639 162 17640 13 17641 17641 17642 77 17643 126 17644 170 17645 17645 17646 2 17647 282 17648 17648 17649 17649 17650 17650 17651 17651 17652 17652 17653 17653 17654 2 17655 17655 17656 17656 17657 17657 17658 17658 17659 17659 17660 17660 17661 17661 17662 15 17663 16 17664 15 17665 11 17666 26 17667 2 17668 16 17669 11 17670 11 17671 40 17672 16 17673 11 17674 11 17675 17675 17676 16 17677 15 17678 70 17679 71 17680 11 17681 16 17682 77 17683 11 17684 11 17685 17685 17686 88 17687 17687 17688 92 17689 17689 17690 98 17691 17691 17692 103 17693 106 17694 17694 17695 17695 17696 17696 17697 17697 17698 17698 17699 17699 17700 17700 17701 126 17702 170 17703 17703 17704 26 17705 11 17706 2 17707 282 17708 26 17709 17709 17710 15 17711 15 17712 11 17713 26 17714 2 17715 11 17716 11 17717 40 17718 16 17719 17719 17720 15 17721 70 17722 71 17723 11 17724 77 17725 11 17726 11 17727 17727 17728 88 17729 103 17730 17730 17731 17696 17732 17732 17733 126 17734 170 17735 17735 17736 26 17737 17737 17738 17738 17739 1066 17740 7315 17741 17741 17742 17384 17743 17385 17744 7315 17745 17745 17746 2 17747 282 17748 17462 17749 17463 17750 17750 17751 17751 17752 17466 17753 7315 17754 17754 17755 22 17756 159 17757 162 17758 7315 17759 17759 17760 126 17761 170 17762 17762 17763 17763 17764 11 17765 77 17766 17766 17767 126 17768 170 17769 17769 17770 13 17771 17635 17772 162 17773 77 17774 126 17775 170 17776 17776 17777 2 17778 282 17779 17648 17780 17649 17781 17781 17782 17651 17783 17652 17784 17653 17785 17785 17786 17653 17787 17787 17788 17788 17789 17789 17790 2 17791 17655 17792 7315 17793 17793 17794 17794 17795 17795 17796 17796 17797 17797 17798 17798 17799 15 17800 16 17801 15 17802 11 17803 26 17804 2 17805 16 17806 11 17807 11 17808 40 17809 16 17810 11 17811 11 17812 17812 17813 16 17814 15 17815 70 17816 71 17817 11 17818 16 17819 77 17820 11 17821 11 17822 17822 17823 88 17824 17824 17825 92 17826 17826 17827 98 17828 17828 17829 103 17830 106 17831 17831 17832 17832 17833 17833 17834 17834 17835 17835 17836 17836 17837 17837 17838 126 17839 170 17840 17840 17841 26 17842 11 17843 2 17844 282 17845 26 17846 17846 17847 15 17848 15 17849 11 17850 26 17851 2 17852 11 17853 11 17854 40 17855 16 17856 17856 17857 15 17858 70 17859 71 17860 11 17861 77 17862 11 17863 11 17864 17864 17865 88 17866 103 17867 17867 17868 17833 17869 17869 17870 126 17871 170 17872 17872 17873 26 17874 17874 17875 17875 17876 1066 17877 7315 17878 17878 17879 17384 17880 17385 17881 7315 17882 17882 17883 2 17884 282 17885 17462 17886 17463 17887 17887 17888 17888 17889 17466 17890 7315 17891 17891 17892 22 17893 159 17894 162 17895 7315 17896 17896 17897 126 17898 170 17899 17899 17900 17900 17901 11 17902 77 17903 17903 17904 126 17905 170 17906 17906 17907 13 17908 17635 17909 162 17910 77 17911 126 17912 170 17913 17913 17914 2 17915 282 17916 17648 17917 17649 17918 17918 17919 17651 17920 17652 17921 17653 17922 17922 17923 17653 17924 17924 17925 17788 17926 17789 17927 7315 17928 17928 17929 17789 17930 7315 17931 17931 17932 17932 17933 17933 17934 17934 17935 17935 17936 3042 17937 3005 17938 17938 17939 17939 17940 2 17941 2 17942 3 17943 4 17944 2 17945 2 17946 2 17947 2 17948 11 17949 2 17950 2 17951 3059 17952 2 17953 3061 17954 22 17955 3063 17956 3064 17957 2 17958 11 17959 11 17960 3068 17961 3069 17962 3070 17963 22 17964 2 17965 2 17966 2 17967 11 17968 3076 17969 11 17970 3078 17971 11 17972 3080 17973 3081 17974 3082 17975 3083 17976 60 17977 3085 17978 2 17979 2 17980 3088 17981 11 17982 2 17983 2 17984 69 17985 131 17986 3094 17987 2 17988 2 17989 3097 17990 2 17991 3099 17992 3100 17993 2 17994 85 17995 86 17996 89 17997 134 17998 11 17999 3107 18000 3108 18001 99 18002 3110 18003 3111 18004 18004 18005 104 18006 104 18007 3115 18008 3116 18009 3117 18010 3118 18011 3119 18012 3120 18013 3121 18014 3122 18015 3123 18016 3124 18017 3125 18018 99 18019 2 18020 3128 18021 2 18022 2 18023 2 18024 2 18025 2 18026 2 18027 2 18028 2 18029 307 18030 2 18031 2 18032 2 18033 2 18034 2 18035 313 18036 2 18037 2 18038 2 18039 2 18040 2 18041 2 18042 2 18043 2 18044 2 18045 2 18046 2 18047 2 18048 2 18049 2 18050 2 18051 3159 18052 3160 18053 3161 18054 3162 18055 3163 18056 3164 18057 3165 18058 3166 18059 3167 18060 3168 18061 3169 18062 3170 18063 3171 18064 3172 18065 3173 18066 4 18067 2 18068 3176 18069 3177 18070 3178 18071 3179 18072 3180 18073 3181 18074 3182 18075 3183 18076 3184 18077 3185 18078 3186 18079 357 18080 357 18081 40 18082 360 18083 434 18084 3192 18085 3193 18086 3194 18087 3195 18088 3196 18089 2 18090 574 18091 3199 18092 3200 18093 40 18094 3202 18095 282 18096 3204 18097 3205 18098 2 18099 574 18100 3208 18101 3209 18102 3210 18103 3211 18104 3212 18105 3213 18106 2 18107 740 18108 2 18109 3217 18110 3218 18111 3219 18112 3220 18113 3221 18114 3222 18115 3223 18116 3224 18117 3225 18118 3226 18119 855 18120 3228 18121 3229 18122 3230 18123 3231 18124 3232 18125 3233 18126 3234 18127 3235 18128 3236 18129 3237 18130 3238 18131 3239 18132 3240 18133 3241 18134 3242 18135 3243 18136 3244 18137 3245 18138 3246 18139 3247 18140 3248 18141 3249 18142 3250 18143 3251 18144 3252 18145 3253 18146 3254 18147 3255 18148 3256 18149 3257 18150 3258 18151 3259 18152 3260 18153 3261 18154 3262 18155 3263 18156 3264 18157 3265 18158 3266 18159 3267 18160 3268 18161 3269 18162 3270 18163 3271 18164 3272 18165 3273 18166 3274 18167 3275 18168 3276 18169 3277 18170 3278 18171 3279 18172 3280 18173 3281 18174 3282 18175 3283 18176 3284 18177 3285 18178 3286 18179 3287 18180 3288 18181 3289 18182 3290 18183 3291 18184 3292 18185 3293 18186 3294 18187 3295 18188 3296 18189 3297 18190 3298 18191 3299 18192 3300 18193 3301 18194 3302 18195 3303 18196 3304 18197 3305 18198 3306 18199 3307 18200 11 18201 3309 18202 3310 18203 3311 18204 2 18205 574 18206 3314 18207 943 18208 3316 18209 3317 18210 976 18211 159 18212 162 18213 3321 18214 3322 18215 1487 18216 2 18217 2 18218 2 18219 3327 18220 2 18221 3329 18222 2 18223 2 18224 2 18225 2 18226 2 18227 2 18228 2 18229 2 18230 2 18231 3339 18232 3340 18233 3341 18234 3342 18235 3343 18236 3344 18237 3345 18238 3346 18239 3347 18240 3348 18241 3349 18242 3350 18243 3351 18244 22 18245 3353 18246 4 18247 11 18248 3356 18249 89 18250 11 18251 3359 18252 99 18253 3361 18254 1551 18255 3363 18256 3364 18257 3365 18258 99 18259 855 18260 11 18261 3369 18262 943 18263 3371 18264 3372 18265 976 18266 159 18267 162 18268 3376 18269 3377 18270 3378 18271 855 18272 11 18273 3381 18274 3382 18275 3383 18276 357 18277 357 18278 360 18279 434 18280 3388 18281 3389 18282 3390 18283 3391 18284 2 18285 574 18286 3394 18287 3395 18288 3396 18289 3397 18290 3398 18291 3399 18292 2 18293 574 18294 3402 18295 3403 18296 140 18297 11 18298 142 18299 162 18300 163 18301 165 18302 167 18303 77 18304 126 18305 170 18306 3414 18307 3415 18308 140 18309 11 18310 142 18311 162 18312 2248 18313 2250 18314 2251 18315 2252 18316 2 18317 3425 18318 3426 18319 855 18320 3429 18321 11 18322 3431 18323 2 18324 574 18325 3434 18326 3435 18327 3436 18328 2732 18329 3438 18330 3439 18331 357 18332 357 18333 360 18334 434 18335 3444 18336 3445 18337 3446 18338 3447 18339 2 18340 574 18341 3450 18342 3451 18343 3452 18344 3453 18345 2803 18346 3455 18347 434 18348 3457 18349 2 18350 3459 18351 3460 18352 2 18353 3 18354 4 18355 2 18356 2 18357 2 18358 2 18359 11 18360 2 18361 2 18362 3471 18363 2 18364 3473 18365 22 18366 3475 18367 3476 18368 2 18369 11 18370 11 18371 3480 18372 3481 18373 3482 18374 22 18375 2 18376 2 18377 2 18378 11 18379 3488 18380 11 18381 3490 18382 11 18383 3492 18384 3493 18385 3494 18386 3495 18387 60 18388 3497 18389 2 18390 2 18391 3500 18392 11 18393 2 18394 2 18395 69 18396 131 18397 3506 18398 2 18399 2 18400 3509 18401 2 18402 3511 18403 3512 18404 2 18405 85 18406 86 18407 89 18408 134 18409 11 18410 3519 18411 3520 18412 99 18413 3522 18414 3523 18415 3524 18416 104 18417 104 18418 3527 18419 3528 18420 3529 18421 3530 18422 3531 18423 3532 18424 3533 18425 3534 18426 3535 18427 3536 18428 3537 18429 99 18430 2 18431 3540 18432 2 18433 2 18434 2 18435 2 18436 2 18437 2 18438 2 18439 2 18440 307 18441 2 18442 2 18443 2 18444 2 18445 2 18446 313 18447 2 18448 2 18449 2 18450 2 18451 2 18452 2 18453 2 18454 2 18455 2 18456 2 18457 2 18458 2 18459 2 18460 2 18461 2 18462 3571 18463 3572 18464 3573 18465 3574 18466 3575 18467 3576 18468 3577 18469 3578 18470 3579 18471 3580 18472 3581 18473 3582 18474 3583 18475 3584 18476 3585 18477 4 18478 2 18479 3588 18480 3589 18481 3590 18482 3591 18483 3592 18484 3593 18485 3594 18486 3595 18487 3596 18488 3597 18489 3598 18490 357 18491 357 18492 40 18493 360 18494 434 18495 3604 18496 3605 18497 3606 18498 3607 18499 3608 18500 2 18501 574 18502 3611 18503 3612 18504 40 18505 3614 18506 282 18507 3616 18508 3617 18509 2 18510 574 18511 3620 18512 3621 18513 3622 18514 3623 18515 3624 18516 3625 18517 2 18518 740 18519 2 18520 3629 18521 3630 18522 3631 18523 3632 18524 3633 18525 3634 18526 3635 18527 3636 18528 3637 18529 3638 18530 855 18531 3640 18532 3641 18533 3642 18534 3643 18535 3644 18536 3645 18537 3646 18538 3647 18539 3648 18540 3649 18541 3650 18542 3651 18543 3652 18544 3653 18545 3654 18546 3655 18547 3656 18548 3657 18549 3658 18550 3659 18551 3660 18552 3661 18553 3662 18554 3663 18555 3664 18556 3665 18557 3666 18558 3667 18559 3668 18560 3669 18561 3670 18562 3671 18563 3672 18564 3673 18565 3674 18566 3675 18567 3676 18568 3677 18569 3678 18570 3679 18571 3680 18572 3681 18573 3682 18574 3683 18575 3684 18576 3685 18577 3686 18578 3687 18579 3688 18580 3689 18581 3690 18582 3691 18583 3692 18584 3693 18585 3694 18586 3695 18587 3696 18588 3697 18589 3698 18590 3699 18591 3700 18592 3701 18593 3702 18594 3703 18595 3704 18596 3705 18597 3706 18598 3707 18599 3708 18600 3709 18601 3710 18602 3711 18603 3712 18604 3713 18605 3714 18606 3715 18607 3716 18608 3717 18609 3718 18610 3719 18611 11 18612 3721 18613 3722 18614 3723 18615 2 18616 574 18617 3726 18618 943 18619 3728 18620 3729 18621 976 18622 159 18623 162 18624 3733 18625 3734 18626 1487 18627 2 18628 2 18629 2 18630 3739 18631 2 18632 3741 18633 2 18634 2 18635 2 18636 2 18637 2 18638 2 18639 2 18640 2 18641 2 18642 3751 18643 3752 18644 3753 18645 3754 18646 3755 18647 3756 18648 3757 18649 3758 18650 3759 18651 3760 18652 3761 18653 3762 18654 3763 18655 22 18656 3765 18657 4 18658 11 18659 3768 18660 89 18661 11 18662 3771 18663 99 18664 3773 18665 1551 18666 3775 18667 3776 18668 3777 18669 99 18670 855 18671 11 18672 3781 18673 943 18674 3783 18675 3784 18676 976 18677 159 18678 162 18679 3788 18680 3789 18681 3790 18682 855 18683 11 18684 3793 18685 3794 18686 3795 18687 357 18688 357 18689 360 18690 434 18691 3800 18692 3801 18693 3802 18694 3803 18695 2 18696 574 18697 3806 18698 3807 18699 3808 18700 3809 18701 3810 18702 3811 18703 2 18704 574 18705 3814 18706 3815 18707 140 18708 11 18709 142 18710 162 18711 163 18712 165 18713 167 18714 77 18715 126 18716 170 18717 3826 18718 3827 18719 140 18720 11 18721 142 18722 162 18723 2248 18724 2250 18725 2251 18726 2252 18727 2 18728 3837 18729 3838 18730 2 18731 3840 18732 855 18733 3842 18734 11 18735 3844 18736 2 18737 574 18738 3847 18739 3848 18740 3849 18741 2732 18742 3851 18743 3852 18744 357 18745 357 18746 360 18747 434 18748 3857 18749 3858 18750 3859 18751 3860 18752 2 18753 574 18754 3863 18755 3864 18756 3865 18757 3866 18758 2803 18759 3868 18760 434 18761 3870 18762 2 18763 3872 18764 3873 18765 3524 18766 2 18767 3876 18768 3877 18769 3878 18770 3879 18771 3880 18772 2894 18773 3882 18774 69 18775 3884 18776 3885 18777 3886 18778 3887 18779 3888 18780 3889 18781 3890 18782 2 18783 3892 18784 3893 18785 3894 18786 3895 18787 3896 18788 3897 18789 2936 18790 2938 18791 3900 18792 3901 18793 3902 18794 3903 18795 3904 18796 357 18797 357 18798 40 18799 360 18800 434 18801 3910 18802 3911 18803 3912 18804 3913 18805 3914 18806 40 18807 3916 18808 3917 18809 2959 18810 2962 18811 2964 18812 2966 18813 77 18814 2968 18815 2969 18816 3925 18817 282 18818 3927 18819 2 18820 282 18821 77 18822 3931 18823 2828 18824 4295 18825 1862 18826 162 18827 2050 18828 2 18829 5538 18830 2 18831 159 18832 162 18833 5542 18834 5543 18835 5544 18836 5545 18837 5546 18838 943 18839 5548 18840 5549 18841 5550 18842 976 18843 159 18844 162 18845 5554 18846 5555 18847 2 18848 282 18849 5558 18850 2 18851 282 18852 5561 18853 5562 18854 5563 18855 5564 18856 5565 18857 5566 18858 1862 18859 162 18860 5569 18861 5570 18862 5571 18863 5572 18864 282 18865 5574 18866 5575 18867 2 18868 574 18869 5578 18870 5236 18871 5580 18872 5581 18873 5582 18874 5583 18875 5584 18876 5585 18877 2894 18878 5587 18879 69 18880 5589 18881 5590 18882 5591 18883 5592 18884 5593 18885 5594 18886 5595 18887 2 18888 5597 18889 5598 18890 5599 18891 5600 18892 5601 18893 5602 18894 2936 18895 2938 18896 5605 18897 5606 18898 5607 18899 5608 18900 5609 18901 357 18902 357 18903 40 18904 360 18905 434 18906 5615 18907 5616 18908 5617 18909 5618 18910 5619 18911 40 18912 5621 18913 5622 18914 2959 18915 2962 18916 2964 18917 2966 18918 77 18919 2968 18920 2969 18921 5630 18922 282 18923 5632 18924 2 18925 282 18926 77 18927 3931 18928 2828 18929 5638 18930 1862 18931 162 18932 2050 18933 2 18934 5643 18935 2 18936 159 18937 162 18938 5647 18939 5648 18940 5649 18941 5650 18942 5651 18943 943 18944 5653 18945 5654 18946 5655 18947 976 18948 159 18949 162 18950 5659 18951 5660 18952 2 18953 282 18954 5663 18955 2 18956 282 18957 5666 18958 5667 18959 5668 18960 5669 18961 5670 18962 5671 18963 1862 18964 162 18965 5674 18966 5675 18967 5676 18968 5677 18969 282 18970 5679 18971 5680 18972 2 18973 574 18974 5683 18975 5236 18976 5685 18977 5686 18978 18978 18979 7315 18980 18980 18981 7315 18982 18982 18983 4 18984 18984 18985 18985 18986 3042 18987 18987 18988 40 18989 18989 18990 18990 18991 18991 18992 2 18993 574 18994 18994 18995 2 18996 2 18997 3 18998 4 18999 2 19000 2 19001 2 19002 2 19003 11 19004 2 19005 2 19006 19006 19007 2 19008 19008 19009 22 19010 19010 19011 19011 19012 2 19013 11 19014 11 19015 19015 19016 19016 19017 19017 19018 22 19019 2 19020 2 19021 2 19022 11 19023 19023 19024 11 19025 19025 19026 11 19027 19027 19028 19028 19029 19029 19030 19030 19031 60 19032 19032 19033 2 19034 2 19035 19035 19036 11 19037 2 19038 2 19039 69 19040 131 19041 19041 19042 2 19043 2 19044 19044 19045 2 19046 19046 19047 19047 19048 2 19049 85 19050 86 19051 89 19052 134 19053 11 19054 19054 19055 19055 19056 99 19057 19057 19058 19058 19059 19059 19060 104 19061 104 19062 19062 19063 19063 19064 19064 19065 19065 19066 19066 19067 19067 19068 19068 19069 19069 19070 19070 19071 19071 19072 19072 19073 99 19074 2 19075 19075 19076 2 19077 2 19078 2 19079 2 19080 2 19081 2 19082 2 19083 2 19084 307 19085 2 19086 2 19087 2 19088 2 19089 2 19090 313 19091 2 19092 2 19093 2 19094 2 19095 2 19096 2 19097 2 19098 2 19099 2 19100 2 19101 2 19102 2 19103 2 19104 2 19105 2 19106 19106 19107 19107 19108 19108 19109 19109 19110 19110 19111 19111 19112 19112 19113 19113 19114 19114 19115 19115 19116 19116 19117 19117 19118 19118 19119 19119 19120 19120 19121 4 19122 2 19123 19123 19124 19124 19125 19125 19126 19126 19127 19127 19128 19128 19129 19129 19130 19130 19131 19131 19132 19132 19133 19133 19134 357 19135 357 19136 40 19137 360 19138 434 19139 19139 19140 19140 19141 19141 19142 19142 19143 19143 19144 2 19145 574 19146 19146 19147 19147 19148 40 19149 19149 19150 282 19151 19151 19152 19152 19153 2 19154 574 19155 19155 19156 19156 19157 19157 19158 19158 19159 19159 19160 19160 19161 2 19162 740 19163 2 19164 19164 19165 19165 19166 19166 19167 19167 19168 19168 19169 19169 19170 19170 19171 19171 19172 19172 19173 19173 19174 855 19175 19175 19176 19176 19177 19177 19178 19178 19179 19179 19180 19180 19181 19181 19182 19182 19183 19183 19184 19184 19185 19185 19186 19186 19187 19187 19188 19188 19189 19189 19190 19190 19191 19191 19192 19192 19193 19193 19194 19194 19195 19195 19196 19196 19197 19197 19198 19198 19199 19199 19200 19200 19201 19201 19202 19202 19203 19203 19204 19204 19205 19205 19206 19206 19207 19207 19208 19208 19209 19209 19210 19210 19211 19211 19212 19212 19213 19213 19214 19214 19215 19215 19216 19216 19217 19217 19218 19218 19219 19219 19220 19220 19221 19221 19222 19222 19223 19223 19224 19224 19225 19225 19226 19226 19227 19227 19228 19228 19229 19229 19230 19230 19231 19231 19232 19232 19233 19233 19234 19234 19235 19235 19236 19236 19237 19237 19238 19238 19239 19239 19240 19240 19241 19241 19242 19242 19243 19243 19244 19244 19245 19245 19246 19246 19247 19247 19248 19248 19249 19249 19250 19250 19251 19251 19252 19252 19253 19253 19254 19254 19255 11 19256 19256 19257 19257 19258 19258 19259 2 19260 574 19261 19261 19262 943 19263 19263 19264 19264 19265 976 19266 159 19267 162 19268 19268 19269 19269 19270 1487 19271 2 19272 2 19273 2 19274 19274 19275 2 19276 19276 19277 2 19278 2 19279 2 19280 2 19281 2 19282 2 19283 2 19284 2 19285 2 19286 19286 19287 19287 19288 19288 19289 19289 19290 19290 19291 19291 19292 19292 19293 19293 19294 19294 19295 19295 19296 19296 19297 19297 19298 19298 19299 22 19300 19300 19301 4 19302 11 19303 19303 19304 89 19305 11 19306 19306 19307 99 19308 19308 19309 1551 19310 19310 19311 19311 19312 19312 19313 99 19314 855 19315 11 19316 19316 19317 943 19318 19318 19319 19319 19320 976 19321 159 19322 162 19323 19323 19324 19324 19325 19325 19326 855 19327 11 19328 19328 19329 19329 19330 19330 19331 357 19332 357 19333 360 19334 434 19335 19335 19336 19336 19337 19337 19338 19338 19339 2 19340 574 19341 19341 19342 19342 19343 19343 19344 19344 19345 19345 19346 19346 19347 2 19348 574 19349 19349 19350 19350 19351 140 19352 11 19353 142 19354 162 19355 163 19356 165 19357 167 19358 77 19359 126 19360 170 19361 19361 19362 19362 19363 140 19364 11 19365 142 19366 162 19367 2248 19368 2250 19369 2251 19370 2252 19371 2 19372 2 19373 3 19374 4 19375 2 19376 2 19377 2 19378 2 19379 11 19380 2 19381 2 19382 19006 19383 2 19384 19008 19385 22 19386 19010 19387 19011 19388 2 19389 11 19390 11 19391 19015 19392 19016 19393 19017 19394 22 19395 2 19396 2 19397 2 19398 11 19399 19023 19400 11 19401 19025 19402 11 19403 19027 19404 19028 19405 19029 19406 19030 19407 60 19408 19032 19409 2 19410 2 19411 19035 19412 11 19413 2 19414 2 19415 69 19416 131 19417 19041 19418 2 19419 2 19420 19044 19421 2 19422 19046 19423 19047 19424 2 19425 85 19426 86 19427 89 19428 134 19429 11 19430 19054 19431 19055 19432 99 19433 19057 19434 19058 19435 19435 19436 104 19437 104 19438 19062 19439 19063 19440 19064 19441 19065 19442 19066 19443 19067 19444 19068 19445 19069 19446 19070 19447 19071 19448 19072 19449 99 19450 2 19451 19075 19452 2 19453 2 19454 2 19455 2 19456 2 19457 2 19458 2 19459 2 19460 307 19461 2 19462 2 19463 2 19464 2 19465 2 19466 313 19467 2 19468 2 19469 2 19470 2 19471 2 19472 2 19473 2 19474 2 19475 2 19476 2 19477 2 19478 2 19479 2 19480 2 19481 2 19482 19106 19483 19107 19484 19108 19485 19109 19486 19110 19487 19111 19488 19112 19489 19113 19490 19114 19491 19115 19492 19116 19493 19117 19494 19118 19495 19119 19496 19120 19497 4 19498 2 19499 19123 19500 19124 19501 19125 19502 19126 19503 19127 19504 19128 19505 19129 19506 19130 19507 19131 19508 19132 19509 19133 19510 357 19511 357 19512 40 19513 360 19514 434 19515 19139 19516 19140 19517 19141 19518 19142 19519 19143 19520 2 19521 574 19522 19146 19523 19147 19524 40 19525 19149 19526 282 19527 19151 19528 19152 19529 2 19530 574 19531 19155 19532 19156 19533 19157 19534 19158 19535 19159 19536 19160 19537 2 19538 740 19539 2 19540 19164 19541 19165 19542 19166 19543 19167 19544 19168 19545 19169 19546 19170 19547 19171 19548 19172 19549 19173 19550 855 19551 19175 19552 19176 19553 19177 19554 19178 19555 19179 19556 19180 19557 19181 19558 19182 19559 19183 19560 19184 19561 19185 19562 19186 19563 19187 19564 19188 19565 19189 19566 19190 19567 19191 19568 19192 19569 19193 19570 19194 19571 19195 19572 19196 19573 19197 19574 19198 19575 19199 19576 19200 19577 19201 19578 19202 19579 19203 19580 19204 19581 19205 19582 19206 19583 19207 19584 19208 19585 19209 19586 19210 19587 19211 19588 19212 19589 19213 19590 19214 19591 19215 19592 19216 19593 19217 19594 19218 19595 19219 19596 19220 19597 19221 19598 19222 19599 19223 19600 19224 19601 19225 19602 19226 19603 19227 19604 19228 19605 19229 19606 19230 19607 19231 19608 19232 19609 19233 19610 19234 19611 19235 19612 19236 19613 19237 19614 19238 19615 19239 19616 19240 19617 19241 19618 19242 19619 19243 19620 19244 19621 19245 19622 19246 19623 19247 19624 19248 19625 19249 19626 19250 19627 19251 19628 19252 19629 19253 19630 19254 19631 11 19632 19256 19633 19257 19634 19258 19635 2 19636 574 19637 19261 19638 943 19639 19263 19640 19264 19641 976 19642 159 19643 162 19644 19268 19645 19269 19646 1487 19647 2 19648 2 19649 2 19650 19274 19651 2 19652 19276 19653 2 19654 2 19655 2 19656 2 19657 2 19658 2 19659 2 19660 2 19661 2 19662 19286 19663 19287 19664 19288 19665 19289 19666 19290 19667 19291 19668 19292 19669 19293 19670 19294 19671 19295 19672 19296 19673 19297 19674 19298 19675 22 19676 19300 19677 4 19678 11 19679 19303 19680 89 19681 11 19682 19306 19683 99 19684 19308 19685 1551 19686 19310 19687 19311 19688 19312 19689 99 19690 855 19691 11 19692 19316 19693 943 19694 19318 19695 19319 19696 976 19697 159 19698 162 19699 19323 19700 19324 19701 19325 19702 855 19703 11 19704 19328 19705 19329 19706 19330 19707 357 19708 357 19709 360 19710 434 19711 19335 19712 19336 19713 19337 19714 19338 19715 2 19716 574 19717 19341 19718 19342 19719 19343 19720 19344 19721 19345 19722 19346 19723 2 19724 574 19725 19349 19726 19350 19727 140 19728 11 19729 142 19730 162 19731 163 19732 165 19733 167 19734 77 19735 126 19736 170 19737 19361 19738 19362 19739 140 19740 11 19741 142 19742 162 19743 2248 19744 2250 19745 2251 19746 2252 19747 2 19748 19435 19749 19749 19750 19749 19751 855 19752 19752 19753 11 19754 19754 19755 2 19756 574 19757 19757 19758 19758 19759 19759 19760 2732 19761 19761 19762 19762 19763 357 19764 357 19765 360 19766 434 19767 19767 19768 19768 19769 19769 19770 19770 19771 2 19772 574 19773 19773 19774 19774 19775 19775 19776 19776 19777 2803 19778 19778 19779 434 19780 19780 19781 2 19782 19782 19783 19783 19784 2 19785 3 19786 4 19787 2 19788 2 19789 2 19790 2 19791 11 19792 2 19793 2 19794 19794 19795 2 19796 19796 19797 22 19798 19798 19799 19799 19800 2 19801 11 19802 11 19803 19803 19804 19804 19805 19805 19806 22 19807 2 19808 2 19809 2 19810 11 19811 19811 19812 11 19813 19813 19814 11 19815 19815 19816 19816 19817 19817 19818 19818 19819 60 19820 19820 19821 2 19822 2 19823 19823 19824 11 19825 2 19826 2 19827 69 19828 131 19829 19829 19830 2 19831 2 19832 19832 19833 2 19834 19834 19835 19835 19836 2 19837 85 19838 86 19839 89 19840 134 19841 11 19842 19842 19843 19843 19844 99 19845 19845 19846 19846 19847 19847 19848 104 19849 104 19850 19850 19851 19851 19852 19852 19853 19853 19854 19854 19855 19855 19856 19856 19857 19857 19858 19858 19859 19859 19860 19860 19861 99 19862 2 19863 19863 19864 2 19865 2 19866 2 19867 2 19868 2 19869 2 19870 2 19871 2 19872 307 19873 2 19874 2 19875 2 19876 2 19877 2 19878 313 19879 2 19880 2 19881 2 19882 2 19883 2 19884 2 19885 2 19886 2 19887 2 19888 2 19889 2 19890 2 19891 2 19892 2 19893 2 19894 19894 19895 19895 19896 19896 19897 19897 19898 19898 19899 19899 19900 19900 19901 19901 19902 19902 19903 19903 19904 19904 19905 19905 19906 19906 19907 19907 19908 19908 19909 4 19910 2 19911 19911 19912 19912 19913 19913 19914 19914 19915 19915 19916 19916 19917 19917 19918 19918 19919 19919 19920 19920 19921 19921 19922 357 19923 357 19924 40 19925 360 19926 434 19927 19927 19928 19928 19929 19929 19930 19930 19931 19931 19932 2 19933 574 19934 19934 19935 19935 19936 40 19937 19937 19938 282 19939 19939 19940 19940 19941 2 19942 574 19943 19943 19944 19944 19945 19945 19946 19946 19947 19947 19948 19948 19949 2 19950 740 19951 2 19952 19952 19953 19953 19954 19954 19955 19955 19956 19956 19957 19957 19958 19958 19959 19959 19960 19960 19961 19961 19962 855 19963 19963 19964 19964 19965 19965 19966 19966 19967 19967 19968 19968 19969 19969 19970 19970 19971 19971 19972 19972 19973 19973 19974 19974 19975 19975 19976 19976 19977 19977 19978 19978 19979 19979 19980 19980 19981 19981 19982 19982 19983 19983 19984 19984 19985 19985 19986 19986 19987 19987 19988 19988 19989 19989 19990 19990 19991 19991 19992 19992 19993 19993 19994 19994 19995 19995 19996 19996 19997 19997 19998 19998 19999 19999 20000 20000 20001 20001 20002 20002 20003 20003 20004 20004 20005 20005 20006 20006 20007 20007 20008 20008 20009 20009 20010 20010 20011 20011 20012 20012 20013 20013 20014 20014 20015 20015 20016 20016 20017 20017 20018 20018 20019 20019 20020 20020 20021 20021 20022 20022 20023 20023 20024 20024 20025 20025 20026 20026 20027 20027 20028 20028 20029 20029 20030 20030 20031 20031 20032 20032 20033 20033 20034 20034 20035 20035 20036 20036 20037 20037 20038 20038 20039 20039 20040 20040 20041 20041 20042 20042 20043 11 20044 20044 20045 20045 20046 20046 20047 2 20048 574 20049 20049 20050 943 20051 20051 20052 20052 20053 976 20054 159 20055 162 20056 20056 20057 20057 20058 1487 20059 2 20060 2 20061 2 20062 20062 20063 2 20064 20064 20065 2 20066 2 20067 2 20068 2 20069 2 20070 2 20071 2 20072 2 20073 2 20074 20074 20075 20075 20076 20076 20077 20077 20078 20078 20079 20079 20080 20080 20081 20081 20082 20082 20083 20083 20084 20084 20085 20085 20086 20086 20087 22 20088 20088 20089 4 20090 11 20091 20091 20092 89 20093 11 20094 20094 20095 99 20096 20096 20097 1551 20098 20098 20099 20099 20100 20100 20101 99 20102 855 20103 11 20104 20104 20105 943 20106 20106 20107 20107 20108 976 20109 159 20110 162 20111 20111 20112 20112 20113 20113 20114 855 20115 11 20116 20116 20117 20117 20118 20118 20119 357 20120 357 20121 360 20122 434 20123 20123 20124 20124 20125 20125 20126 20126 20127 2 20128 574 20129 20129 20130 20130 20131 20131 20132 20132 20133 20133 20134 20134 20135 2 20136 574 20137 20137 20138 20138 20139 140 20140 11 20141 142 20142 162 20143 163 20144 165 20145 167 20146 77 20147 126 20148 170 20149 20149 20150 20150 20151 140 20152 11 20153 142 20154 162 20155 2248 20156 2250 20157 2251 20158 2252 20159 2 20160 20160 20161 20161 20162 2 20163 20163 20164 855 20165 20165 20166 11 20167 20167 20168 2 20169 574 20170 20170 20171 20171 20172 20172 20173 2732 20174 20174 20175 20175 20176 357 20177 357 20178 360 20179 434 20180 20180 20181 20181 20182 20182 20183 20183 20184 2 20185 574 20186 20186 20187 20187 20188 20188 20189 20189 20190 2803 20191 20191 20192 434 20193 20193 20194 2 20195 20195 20196 20196 20197 19847 20198 2 20199 20199 20200 20200 20201 20201 20202 20202 20203 20203 20204 2894 20205 20205 20206 69 20207 20207 20208 20208 20209 20209 20210 20210 20211 20211 20212 20212 20213 20213 20214 2 20215 20215 20216 20216 20217 20217 20218 20218 20219 20219 20220 20220 20221 2936 20222 2938 20223 20223 20224 20224 20225 20225 20226 20226 20227 20227 20228 357 20229 357 20230 40 20231 360 20232 434 20233 20233 20234 20234 20235 20235 20236 20236 20237 20237 20238 40 20239 20239 20240 20240 20241 2959 20242 2962 20243 2964 20244 2966 20245 77 20246 2968 20247 2969 20248 20248 20249 282 20250 20250 20251 2 20252 282 20253 77 20254 3931 20255 2828 20256 20256 20257 1862 20258 162 20259 2050 20260 2 20261 20261 20262 2 20263 159 20264 162 20265 20265 20266 20266 20267 20267 20268 20268 20269 20269 20270 943 20271 20271 20272 20272 20273 20273 20274 976 20275 159 20276 162 20277 20277 20278 20278 20279 2 20280 282 20281 20281 20282 2 20283 282 20284 20284 20285 20285 20286 20286 20287 20287 20288 20288 20289 20289 20290 1862 20291 162 20292 20292 20293 20293 20294 20294 20295 20295 20296 282 20297 20297 20298 20298 20299 2 20300 574 20301 20301 20302 5236 20303 20303 20304 20304 20305 20305 20306 20306 20307 20307 20308 20308 20309 2894 20310 20310 20311 69 20312 20312 20313 20313 20314 20314 20315 20315 20316 20316 20317 20317 20318 20318 20319 2 20320 20320 20321 20321 20322 20322 20323 20323 20324 20324 20325 20325 20326 2936 20327 2938 20328 20328 20329 20329 20330 20330 20331 20331 20332 20332 20333 357 20334 357 20335 40 20336 360 20337 434 20338 20338 20339 20339 20340 20340 20341 20341 20342 20342 20343 40 20344 20344 20345 20345 20346 2959 20347 2962 20348 2964 20349 2966 20350 77 20351 2968 20352 2969 20353 20353 20354 282 20355 20355 20356 2 20357 282 20358 77 20359 3931 20360 2828 20361 20361 20362 1862 20363 162 20364 2050 20365 2 20366 20366 20367 2 20368 159 20369 162 20370 20370 20371 20371 20372 20372 20373 20373 20374 20374 20375 943 20376 20376 20377 20377 20378 20378 20379 976 20380 159 20381 162 20382 20382 20383 20383 20384 2 20385 282 20386 20386 20387 2 20388 282 20389 20389 20390 20390 20391 20391 20392 20392 20393 20393 20394 20394 20395 1862 20396 162 20397 20397 20398 20398 20399 20399 20400 20400 20401 282 20402 20402 20403 20403 20404 2 20405 574 20406 20406 20407 5236 20408 20408 20409 20409 20410 20410 20411 4 20412 11 20413 11 20414 20414 20415 85 20416 22 20417 89 20418 11 20419 20419 20420 99 20421 20421 20422 20422 20423 1551 20424 20424 20425 20425 20426 20426 20427 99 20428 855 20429 11 20430 20430 20431 943 20432 20432 20433 20433 20434 976 20435 159 20436 162 20437 20437 20438 20438 20439 20439 20440 17462 20441 20410 20442 4 20443 11 20444 11 20445 20445 20446 85 20447 22 20448 89 20449 11 20450 20450 20451 99 20452 162 20453 20453 20454 1551 20455 20455 20456 20456 20457 20457 20458 99 20459 855 20460 11 20461 20461 20462 943 20463 20463 20464 20464 20465 976 20466 159 20467 162 20468 20468 20469 20469 20470 20470 20471 17462 20472 20472 20473 20473 20474 20474 20475 20475 20476 20476 20477 15 20478 16 20479 15 20480 11 20481 26 20482 2 20483 7315 20484 16 20485 11 20486 11 20487 40 20488 16 20489 11 20490 11 20491 20491 20492 16 20493 15 20494 70 20495 71 20496 11 20497 16 20498 77 20499 11 20500 11 20501 20501 20502 88 20503 20503 20504 92 20505 20505 20506 98 20507 20507 20508 103 20509 106 20510 20510 20511 20511 20512 20512 20513 20513 20514 20514 20515 20515 20516 20516 20517 126 20518 170 20519 20519 20520 26 20521 11 20522 2 20523 282 20524 26 20525 20525 20526 15 20527 15 20528 11 20529 26 20530 2 20531 11 20532 11 20533 40 20534 16 20535 20535 20536 15 20537 70 20538 71 20539 11 20540 77 20541 11 20542 11 20543 20543 20544 88 20545 103 20546 20546 20547 20512 20548 20548 20549 126 20550 170 20551 20551 20552 26 20553 20553 20554 20554 20555 1066 20556 7315 20557 20557 20558 17384 20559 17385 20560 7315 20561 20561 20562 2 20563 282 20564 17462 20565 17463 20566 20566 20567 20567 20568 17466 20569 7315 20570 20570 20571 22 20572 159 20573 162 20574 7315 20575 20575 20576 126 20577 170 20578 20578 20579 20579 20580 11 20581 77 20582 20582 20583 126 20584 170 20585 20585 20586 13 20587 17635 20588 162 20589 77 20590 126 20591 170 20592 20592 20593 2 20594 282 20595 17648 20596 17649 20597 20597 20598 17651 20599 17652 20600 17653 20601 20601 20602 17653 20603 20603 20604 17788 20605 17789 20606 7315 20607 20607 20608 17789 20609 7315 20610 20610 20611 20611 20612 3042 20613 20613 20614 7315 20615 20615 20616 7315 20617 20617 20618 3042 20619 20619 20620 2 20621 19059 20622 20410 20623 20421 20624 17462 20625 20410 20626 162 20627 17462 20628 20472 20629 20629 20630 20630 20631 20475 20632 7315 20633 20633 20634 20634 20635 20635 20636 20636 20637 7315 20638 20638 20639 2 20640 20640 20641 20635 20642 7315 20643 20643 20644 20644 20645 162 20646 20646 20647 2 20648 20648 20649 17462 20650 20650 20651 20651 20652 20652 20653 20653 20654 20654 20655 15 20656 16 20657 15 20658 11 20659 26 20660 2 20661 7315 20662 16 20663 11 20664 11 20665 40 20666 16 20667 11 20668 11 20669 20669 20670 16 20671 15 20672 70 20673 71 20674 11 20675 16 20676 77 20677 11 20678 11 20679 20679 20680 88 20681 20681 20682 92 20683 20683 20684 98 20685 20685 20686 103 20687 106 20688 20688 20689 20689 20690 20690 20691 20691 20692 20692 20693 20693 20694 20694 20695 126 20696 170 20697 20697 20698 26 20699 11 20700 2 20701 282 20702 26 20703 20703 20704 15 20705 15 20706 11 20707 26 20708 2 20709 11 20710 11 20711 40 20712 16 20713 20713 20714 15 20715 70 20716 71 20717 11 20718 77 20719 11 20720 11 20721 20721 20722 88 20723 103 20724 20724 20725 20690 20726 20726 20727 126 20728 170 20729 20729 20730 26 20731 20731 20732 20732 20733 1066 20734 7315 20735 20735 20736 17384 20737 17385 20738 7315 20739 20739 20740 2 20741 282 20742 17462 20743 17463 20744 20744 20745 20745 20746 17466 20747 7315 20748 20748 20749 22 20750 159 20751 162 20752 7315 20753 20753 20754 126 20755 170 20756 20756 20757 20757 20758 11 20759 77 20760 20760 20761 126 20762 170 20763 20763 20764 13 20765 17635 20766 162 20767 77 20768 126 20769 170 20770 20770 20771 2 20772 282 20773 17648 20774 17649 20775 20775 20776 17651 20777 17652 20778 17653 20779 20779 20780 17653 20781 20781 20782 17788 20783 17789 20784 7315 20785 20785 20786 17789 20787 7315 20788 20788 20789 20789 20790 3042 20791 20791 20792 7315 20793 20793 20794 7315 20795 20795 20796 3042 20797 20797 20798 2 20799 19059 20800 20410 20801 20421 20802 17462 20803 20410 20804 162 20805 17462 20806 20472 20807 20807 20808 20808 20809 20475 20810 7315 20811 20811 20812 20812 20813 20635 20814 7315 20815 20635 20816 7315 20817 20643 20818 162 20819 17462 20820 20650 20821 20821 20822 20822 20823 20653 20824 7315 20825 20825 20826 20826 20827 3042 20828 20828 20829 2 20830 20830 20831 20831 20832 20832 20833 15 20834 15 20835 11 20836 26 20837 2 20838 11 20839 11 20840 40 20841 16 20842 11 20843 20843 20844 15 20845 70 20846 71 20847 11 20848 77 20849 11 20850 11 20851 20851 20852 88 20853 103 20854 20854 20855 20855 20856 20856 20857 126 20858 170 20859 20859 20860 11 20861 2 20862 282 20863 7315 20864 20864 20865 26 20866 11 20867 2 20868 282 20869 26 20870 20870 20871 15 20872 15 20873 11 20874 26 20875 2 20876 11 20877 11 20878 40 20879 16 20880 20880 20881 15 20882 70 20883 71 20884 11 20885 77 20886 11 20887 11 20888 20888 20889 88 20890 103 20891 20891 20892 20855 20893 20893 20894 126 20895 170 20896 20896 20897 26 20898 20898 20899 20899 20900 1066 20901 7315 20902 20902 20903 17384 20904 17385 20905 7315 20906 20906 20907 2 20908 282 20909 17462 20910 17463 20911 20911 20912 20912 20913 17466 20914 7315 20915 20915 20916 22 20917 159 20918 162 20919 7315 20920 20920 20921 126 20922 170 20923 20923 20924 20924 20925 11 20926 77 20927 20927 20928 126 20929 170 20930 20930 20931 13 20932 17635 20933 162 20934 77 20935 126 20936 170 20937 20937 20938 2 20939 282 20940 17648 20941 17649 20942 20942 20943 17651 20944 17652 20945 17653 20946 20946 20947 17653 20948 20948 20949 17788 20950 17789 20951 7315 20952 20952 20953 17789 20954 7315 20955 20955 20956 20956 20957 3042 20958 20958 20959 7315 20960 20960 20961 7315 20962 20962 20963 3042 20964 20964 20965 2 20966 19059 20967 20410 20968 20421 20969 17462 20970 20410 20971 162 20972 17462 20973 20472 20974 20974 20975 20975 20976 20475 20977 7315 20978 20978 20979 20979 20980 20635 20981 7315 20982 20635 20983 7315 20984 20643 20985 162 20986 17462 20987 20650 20988 20988 20989 20989 20990 20653 20991 7315 20992 20992 20993 3042 20994 20994 20995 20856 20996 11 20997 2 20998 282 20999 7315 21000 21000 21001 16 21002 2050 21003 21003 21004 21004 21005 17462 21006 16 21007 21007 21008 21008 21009 21009 21010 21010 21011 21011 21012 21012 21013 15 21014 16 21015 15 21016 11 21017 26 21018 2 21019 7315 21020 16 21021 11 21022 11 21023 40 21024 16 21025 11 21026 11 21027 21027 21028 16 21029 15 21030 70 21031 71 21032 11 21033 16 21034 77 21035 11 21036 11 21037 21037 21038 88 21039 21039 21040 92 21041 21041 21042 98 21043 21043 21044 103 21045 106 21046 21046 21047 21047 21048 21048 21049 21049 21050 21050 21051 21051 21052 21052 21053 126 21054 170 21055 21055 21056 26 21057 11 21058 2 21059 282 21060 26 21061 21061 21062 15 21063 15 21064 11 21065 26 21066 2 21067 11 21068 11 21069 40 21070 16 21071 21071 21072 15 21073 70 21074 71 21075 11 21076 77 21077 11 21078 11 21079 21079 21080 88 21081 103 21082 21082 21083 21048 21084 21084 21085 126 21086 170 21087 21087 21088 26 21089 21089 21090 21090 21091 1066 21092 7315 21093 21093 21094 17384 21095 17385 21096 7315 21097 21097 21098 2 21099 282 21100 17462 21101 17463 21102 21102 21103 21103 21104 17466 21105 7315 21106 21106 21107 22 21108 159 21109 162 21110 7315 21111 21111 21112 126 21113 170 21114 21114 21115 21115 21116 11 21117 77 21118 21118 21119 126 21120 170 21121 21121 21122 13 21123 17635 21124 162 21125 77 21126 126 21127 170 21128 21128 21129 2 21130 282 21131 17648 21132 17649 21133 21133 21134 17651 21135 17652 21136 17653 21137 21137 21138 17653 21139 21139 21140 17788 21141 17789 21142 7315 21143 21143 21144 17789 21145 7315 21146 21146 21147 21147 21148 3042 21149 21149 21150 7315 21151 21151 21152 7315 21153 21153 21154 3042 21155 21155 21156 2 21157 19059 21158 20410 21159 20421 21160 17462 21161 20410 21162 162 21163 17462 21164 20472 21165 21165 21166 21166 21167 20475 21168 7315 21169 21169 21170 21170 21171 20635 21172 7315 21173 20635 21174 7315 21175 20643 21176 162 21177 17462 21178 20650 21179 21179 21180 21180 21181 20653 21182 7315 21183 21183 21184 3042 21185 21185 21186 20856 21187 11 21188 2 21189 282 21190 7315 21191 21191 21192 16 21193 21004 21194 17462 21195 21008 21196 21196 21197 21197 21198 21011 21199 7315 21200 21200 21201 7315 21202 21202 21203 7315 21204 21204 21205 21205 21206 17352 21207 140 21208 11 21209 142 21210 162 21211 165 21212 167 21213 77 21214 126 21215 170 21216 21216 21217 21217 21218 21218 21219 16 21220 2962 21221 2964 21222 2966 21223 77 21224 2968 21225 2969 21226 21226 21227 21227 21228 1066 21229 7315 21230 21230 21231 17384 21232 17385 21233 7315 21234 21234 21235 2 21236 282 21237 17462 21238 17463 21239 21239 21240 21240 21241 17466 21242 7315 21243 21243 21244 22 21245 159 21246 162 21247 7315 21248 21248 21249 126 21250 170 21251 21251 21252 21252 21253 11 21254 77 21255 21255 21256 126 21257 170 21258 21258 21259 13 21260 17635 21261 162 21262 77 21263 126 21264 170 21265 21265 21266 2966 21267 77 21268 2968 21269 2969 21270 21270 21271 2 21272 282 21273 17648 21274 17649 21275 21275 21276 17651 21277 17652 21278 17653 21279 21279 21280 17653 21281 21281 21282 17788 21283 17789 21284 7315 21285 21285 21286 17789 21287 7315 21288 21288 21289 21289 21290 3042 21291 21291 21292 7315 21293 21293 21294 21294 21295 13 21296 2966 21297 77 21298 2968 21299 2969 21300 21300 21301 21301 21302 2 21303 282 21304 17462 21305 17297 21306 21306 21307 21307 21308 162 21309 7315 21310 21310 21311 7315 21312 21312 21313 3042 21314 21314 21315 2 21316 19059 21317 20410 21318 20421 21319 17462 21320 20410 21321 162 21322 17462 21323 20472 21324 21324 21325 21325 21326 20475 21327 7315 21328 21328 21329 21329 21330 20635 21331 7315 21332 20635 21333 7315 21334 20643 21335 162 21336 17462 21337 20650 21338 21338 21339 21339 21340 20653 21341 7315 21342 21342 21343 21343 21344 166 21345 2964 21346 2966 21347 77 21348 2968 21349 2969 21350 21350 21351 2256 21352 21352 21353 21353 21354 21354 21355 21355 21356 21356 21357 21357 21358 21358 21359 21359 21360 21360 21361 21361 21362 21362 21363 21363 21364 21364 21365 2271 21366 2272 21367 2256 21368 21368 21369 21369 21370 21370 21371 21371 21372 21372 21373 21373 21374 21374 21375 21375 21376 21376 21377 21377 21378 21378 21379 2292 21380 21380 21381 2272 21382 21382 21383 2256 21384 21384 21385 21385 21386 21386 21387 21387 21388 21388 21389 21389 21390 21390 21391 21391 21392 21392 21393 21393 21394 2256 21395 21395 21396 21396 21397 21397 21398 21398 21399 21399 21400 21400 21401 21401 21402 21402 21403 3042 21404 21404 21405 20856 21406 11 21407 2 21408 282 21409 7315 21410 21410 21411 16 21412 21004 21413 17462 21414 21008 21415 21415 21416 21416 21417 21011 21418 7315 21419 21419 21420 7315 21421 21421 21422 2962 21423 2964 21424 2966 21425 77 21426 2968 21427 2969 21428 21428 21429 7315 21430 21430 21431 2 21432 21432 21433 21433 21434 7315 21435 21435 21436 7315 21437 21437 21438 21438 21439 21439 21440 1066 21441 7315 21442 21442 21443 17384 21444 17385 21445 7315 21446 21446 21447 2 21448 282 21449 17462 21450 17463 21451 21451 21452 21452 21453 17466 21454 7315 21455 21455 21456 22 21457 159 21458 162 21459 7315 21460 21460 21461 126 21462 170 21463 21463 21464 21464 21465 11 21466 77 21467 21467 21468 126 21469 170 21470 21470 21471 13 21472 17635 21473 162 21474 77 21475 126 21476 170 21477 21477 21478 2 21479 282 21480 17648 21481 17649 21482 21482 21483 17651 21484 17652 21485 17653 21486 21486 21487 17653 21488 21488 21489 17788 21490 17789 21491 7315 21492 21492 21493 17789 21494 7315 21495 21495 21496 21496 21497 3042 21498 21498 21499 7315 21500 21500 21501 2 21502 282 21503 17462 21504 21504 21505 7315 21506 21506 21507 3042 21508 21508 21509 2 21510 19059 21511 20410 21512 20421 21513 17462 21514 20410 21515 162 21516 17462 21517 20472 21518 21518 21519 21519 21520 20475 21521 7315 21522 21522 21523 21523 21524 20635 21525 7315 21526 20635 21527 7315 21528 20643 21529 162 21530 17462 21531 20650 21532 21532 21533 21533 21534 20653 21535 7315 21536 21536 21537 3042 21538 21538 21539 20856 21540 11 21541 2 21542 282 21543 7315 21544 21544 21545 16 21546 21004 21547 17462 21548 21008 21549 21549 21550 21550 21551 21011 21552 7315 21553 21553 21554 7315 21555 21555 21556 7315 21557 21557 21558 16 21559 21004 21560 17462 21561 21008 21562 21562 21563 21563 21564 21011 21565 7315 21566 21566 21567 7315 21568 21568 21569 7315 21570 21570 21571 21571 21572 17352 21573 140 21574 11 21575 142 21576 162 21577 165 21578 167 21579 77 21580 126 21581 170 21582 21582 21583 21583 21584 21584 21585 16 21586 2962 21587 2964 21588 2966 21589 77 21590 2968 21591 2969 21592 21592 21593 21593 21594 1066 21595 7315 21596 21596 21597 17384 21598 17385 21599 7315 21600 21600 21601 2 21602 282 21603 17462 21604 17463 21605 21605 21606 21606 21607 17466 21608 7315 21609 21609 21610 22 21611 159 21612 162 21613 7315 21614 21614 21615 126 21616 170 21617 21617 21618 21618 21619 11 21620 77 21621 21621 21622 126 21623 170 21624 21624 21625 13 21626 17635 21627 162 21628 77 21629 126 21630 170 21631 21631 21632 2966 21633 77 21634 2968 21635 2969 21636 21636 21637 2 21638 282 21639 17648 21640 17649 21641 21641 21642 17651 21643 17652 21644 17653 21645 21645 21646 17653 21647 21647 21648 17788 21649 17789 21650 7315 21651 21651 21652 17789 21653 7315 21654 21654 21655 21655 21656 3042 21657 21657 21658 7315 21659 21659 21660 21294 21661 2966 21662 77 21663 2968 21664 2969 21665 21665 21666 2 21667 282 21668 17462 21669 21669 21670 162 21671 7315 21672 21672 21673 7315 21674 21674 21675 3042 21676 21676 21677 2 21678 19059 21679 20410 21680 20421 21681 17462 21682 20410 21683 162 21684 17462 21685 20472 21686 21686 21687 21687 21688 20475 21689 7315 21690 21690 21691 21691 21692 20635 21693 7315 21694 20635 21695 7315 21696 20643 21697 162 21698 17462 21699 20650 21700 21700 21701 21701 21702 20653 21703 7315 21704 21704 21705 21343 21706 2964 21707 2966 21708 77 21709 2968 21710 2969 21711 21711 21712 21712 21713 3042 21714 21714 21715 20856 21716 11 21717 2 21718 282 21719 7315 21720 21720 21721 16 21722 21004 21723 17462 21724 21008 21725 21725 21726 21726 21727 21011 21728 7315 21729 21729 21730 7315 21731 21731 21732 2962 21733 2964 21734 2966 21735 77 21736 2968 21737 2969 21738 21738 21739 7315 21740 21740 21741 3042 21742 21742 21743 20856 21744 11 21745 2 21746 282 21747 7315 21748 21748 21749 16 21750 21004 21751 17462 21752 21008 21753 21753 21754 21754 21755 21011 21756 7315 21757 21757 21758 7315 21759 21759 21760 7315 21761 21761 21762 21762 21763 17352 21764 140 21765 11 21766 142 21767 162 21768 165 21769 167 21770 77 21771 126 21772 170 21773 21773 21774 21774 21775 21775 21776 16 21777 2962 21778 2964 21779 2966 21780 77 21781 2968 21782 2969 21783 21783 21784 21784 21785 1066 21786 7315 21787 21787 21788 17384 21789 17385 21790 7315 21791 21791 21792 2 21793 282 21794 17462 21795 17463 21796 21796 21797 21797 21798 17466 21799 7315 21800 21800 21801 22 21802 159 21803 162 21804 7315 21805 21805 21806 126 21807 170 21808 21808 21809 21809 21810 11 21811 77 21812 21812 21813 126 21814 170 21815 21815 21816 13 21817 17635 21818 162 21819 77 21820 126 21821 170 21822 21822 21823 2966 21824 77 21825 2968 21826 2969 21827 21827 21828 2 21829 282 21830 17648 21831 17649 21832 21832 21833 17651 21834 17652 21835 17653 21836 21836 21837 17653 21838 21838 21839 17788 21840 17789 21841 7315 21842 21842 21843 17789 21844 7315 21845 21845 21846 21846 21847 3042 21848 21848 21849 7315 21850 21850 21851 21294 21852 2966 21853 77 21854 2968 21855 2969 21856 21856 21857 2 21858 282 21859 17462 21860 21860 21861 162 21862 7315 21863 21863 21864 7315 21865 21865 21866 3042 21867 21867 21868 2 21869 19059 21870 20410 21871 20421 21872 17462 21873 20410 21874 162 21875 17462 21876 20472 21877 21877 21878 21878 21879 20475 21880 7315 21881 21881 21882 21882 21883 20635 21884 7315 21885 20635 21886 7315 21887 20643 21888 162 21889 17462 21890 20650 21891 21891 21892 21892 21893 20653 21894 7315 21895 21895 21896 21343 21897 2964 21898 2966 21899 77 21900 2968 21901 2969 21902 21902 21903 21903 21904 3042 21905 21905 21906 20856 21907 11 21908 2 21909 282 21910 7315 21911 21911 21912 16 21913 21004 21914 17462 21915 21008 21916 21916 21917 21917 21918 21011 21919 7315 21920 21920 21921 7315 21922 21922 21923 2962 21924 2964 21925 2966 21926 77 21927 2968 21928 2969 21929 21929 21930 7315 21931 21931 21932 855 21933 11 21934 21934 21935 21935 21936 21936 21937 357 21938 357 21939 360 21940 434 21941 21941 21942 21942 21943 21943 21944 21944 21945 2 21946 574 21947 21947 21948 21948 21949 21949 21950 21950 21951 21951 21952 434 21953 21953 21954 21954 21955 1862 21956 162 21957 21957 21958 21958 21959 2 21960 574 21961 21961 21962 21962 21963 855 21964 11 21965 21965 21966 21966 21967 21967 21968 357 21969 357 21970 360 21971 434 21972 21972 21973 21973 21974 21974 21975 21975 21976 2 21977 574 21978 21978 21979 21979 21980 21980 21981 21981 21982 21982 21983 434 21984 21984 21985 21985 21986 1862 21987 162 21988 21988 21989 21989 21990 2 21991 574 21992 21992 21993 855 21994 19752 21995 11 21996 19754 21997 2 21998 574 21999 19757 22000 19758 22001 19759 22002 2732 22003 19761 22004 19762 22005 357 22006 357 22007 360 22008 434 22009 19767 22010 19768 22011 19769 22012 19770 22013 2 22014 574 22015 19773 22016 19774 22017 19775 22018 19776 22019 2803 22020 19778 22021 434 22022 19780 22023 2 22024 19782 22025 19783 22026 2 22027 3 22028 4 22029 2 22030 2 22031 2 22032 2 22033 11 22034 2 22035 2 22036 19794 22037 2 22038 19796 22039 22 22040 19798 22041 19799 22042 2 22043 11 22044 11 22045 19803 22046 19804 22047 19805 22048 22 22049 2 22050 2 22051 2 22052 11 22053 19811 22054 11 22055 19813 22056 11 22057 19815 22058 19816 22059 19817 22060 19818 22061 60 22062 19820 22063 2 22064 2 22065 19823 22066 11 22067 2 22068 2 22069 69 22070 131 22071 19829 22072 2 22073 2 22074 19832 22075 2 22076 19834 22077 19835 22078 2 22079 85 22080 86 22081 89 22082 134 22083 11 22084 19842 22085 19843 22086 99 22087 19845 22088 19846 22089 19847 22090 104 22091 104 22092 19850 22093 19851 22094 19852 22095 19853 22096 19854 22097 19855 22098 19856 22099 19857 22100 19858 22101 19859 22102 19860 22103 99 22104 2 22105 19863 22106 2 22107 2 22108 2 22109 2 22110 2 22111 2 22112 2 22113 2 22114 307 22115 2 22116 2 22117 2 22118 2 22119 2 22120 313 22121 2 22122 2 22123 2 22124 2 22125 2 22126 2 22127 2 22128 2 22129 2 22130 2 22131 2 22132 2 22133 2 22134 2 22135 2 22136 19894 22137 19895 22138 19896 22139 19897 22140 19898 22141 19899 22142 19900 22143 19901 22144 19902 22145 19903 22146 19904 22147 19905 22148 19906 22149 19907 22150 19908 22151 4 22152 2 22153 19911 22154 19912 22155 19913 22156 19914 22157 19915 22158 19916 22159 19917 22160 19918 22161 19919 22162 19920 22163 19921 22164 357 22165 357 22166 40 22167 360 22168 434 22169 19927 22170 19928 22171 19929 22172 19930 22173 19931 22174 2 22175 574 22176 19934 22177 19935 22178 40 22179 19937 22180 282 22181 19939 22182 19940 22183 2 22184 574 22185 19943 22186 19944 22187 19945 22188 19946 22189 19947 22190 19948 22191 2 22192 740 22193 2 22194 19952 22195 19953 22196 19954 22197 19955 22198 19956 22199 19957 22200 19958 22201 19959 22202 19960 22203 19961 22204 855 22205 19963 22206 19964 22207 19965 22208 19966 22209 19967 22210 19968 22211 19969 22212 19970 22213 19971 22214 19972 22215 19973 22216 19974 22217 19975 22218 19976 22219 19977 22220 19978 22221 19979 22222 19980 22223 19981 22224 19982 22225 19983 22226 19984 22227 19985 22228 19986 22229 19987 22230 19988 22231 19989 22232 19990 22233 19991 22234 19992 22235 19993 22236 19994 22237 19995 22238 19996 22239 19997 22240 19998 22241 19999 22242 20000 22243 20001 22244 20002 22245 20003 22246 20004 22247 20005 22248 20006 22249 20007 22250 20008 22251 20009 22252 20010 22253 20011 22254 20012 22255 20013 22256 20014 22257 20015 22258 20016 22259 20017 22260 20018 22261 20019 22262 20020 22263 20021 22264 20022 22265 20023 22266 20024 22267 20025 22268 20026 22269 20027 22270 20028 22271 20029 22272 20030 22273 20031 22274 20032 22275 20033 22276 20034 22277 20035 22278 20036 22279 20037 22280 20038 22281 20039 22282 20040 22283 20041 22284 20042 22285 11 22286 20044 22287 20045 22288 20046 22289 2 22290 574 22291 20049 22292 943 22293 20051 22294 20052 22295 976 22296 159 22297 162 22298 20056 22299 20057 22300 1487 22301 2 22302 2 22303 2 22304 20062 22305 2 22306 20064 22307 2 22308 2 22309 2 22310 2 22311 2 22312 2 22313 2 22314 2 22315 2 22316 20074 22317 20075 22318 20076 22319 20077 22320 20078 22321 20079 22322 20080 22323 20081 22324 20082 22325 20083 22326 20084 22327 20085 22328 20086 22329 22 22330 20088 22331 4 22332 11 22333 20091 22334 89 22335 11 22336 20094 22337 99 22338 20096 22339 1551 22340 20098 22341 20099 22342 20100 22343 99 22344 855 22345 11 22346 20104 22347 943 22348 20106 22349 20107 22350 976 22351 159 22352 162 22353 20111 22354 20112 22355 20113 22356 855 22357 11 22358 20116 22359 20117 22360 20118 22361 357 22362 357 22363 360 22364 434 22365 20123 22366 20124 22367 20125 22368 20126 22369 2 22370 574 22371 20129 22372 20130 22373 20131 22374 20132 22375 20133 22376 20134 22377 2 22378 574 22379 20137 22380 20138 22381 140 22382 11 22383 142 22384 162 22385 163 22386 165 22387 167 22388 77 22389 126 22390 170 22391 20149 22392 20150 22393 140 22394 11 22395 142 22396 162 22397 2248 22398 2250 22399 2251 22400 2252 22401 2 22402 20160 22403 20161 22404 2 22405 20163 22406 855 22407 20165 22408 11 22409 20167 22410 2 22411 574 22412 20170 22413 20171 22414 20172 22415 2732 22416 20174 22417 20175 22418 357 22419 357 22420 360 22421 434 22422 20180 22423 20181 22424 20182 22425 20183 22426 2 22427 574 22428 20186 22429 20187 22430 20188 22431 20189 22432 2803 22433 20191 22434 434 22435 20193 22436 2 22437 20195 22438 20196 22439 19847 22440 2 22441 20199 22442 20200 22443 20201 22444 20202 22445 20203 22446 2894 22447 20205 22448 69 22449 20207 22450 20208 22451 20209 22452 20210 22453 20211 22454 20212 22455 20213 22456 2 22457 20215 22458 20216 22459 20217 22460 20218 22461 20219 22462 20220 22463 2936 22464 2938 22465 20223 22466 20224 22467 20225 22468 20226 22469 20227 22470 357 22471 357 22472 40 22473 360 22474 434 22475 20233 22476 20234 22477 20235 22478 20236 22479 20237 22480 40 22481 20239 22482 20240 22483 2959 22484 2962 22485 2964 22486 2966 22487 77 22488 2968 22489 2969 22490 20248 22491 282 22492 20250 22493 2 22494 282 22495 77 22496 3931 22497 2828 22498 20256 22499 1862 22500 162 22501 2050 22502 2 22503 20261 22504 2 22505 159 22506 162 22507 20265 22508 20266 22509 20267 22510 20268 22511 20269 22512 943 22513 20271 22514 20272 22515 20273 22516 976 22517 159 22518 162 22519 20277 22520 20278 22521 2 22522 282 22523 20281 22524 2 22525 282 22526 20284 22527 20285 22528 20286 22529 20287 22530 20288 22531 20289 22532 1862 22533 162 22534 20292 22535 20293 22536 20294 22537 20295 22538 282 22539 20297 22540 20298 22541 2 22542 574 22543 20301 22544 5236 22545 20303 22546 20304 22547 20305 22548 20306 22549 20307 22550 20308 22551 2894 22552 20310 22553 69 22554 20312 22555 20313 22556 20314 22557 20315 22558 20316 22559 20317 22560 20318 22561 2 22562 20320 22563 20321 22564 20322 22565 20323 22566 20324 22567 20325 22568 2936 22569 2938 22570 20328 22571 20329 22572 20330 22573 20331 22574 20332 22575 357 22576 357 22577 40 22578 360 22579 434 22580 20338 22581 20339 22582 20340 22583 20341 22584 20342 22585 40 22586 20344 22587 20345 22588 2959 22589 2962 22590 2964 22591 2966 22592 77 22593 2968 22594 2969 22595 20353 22596 282 22597 20355 22598 2 22599 282 22600 77 22601 3931 22602 2828 22603 20361 22604 1862 22605 162 22606 2050 22607 2 22608 20366 22609 2 22610 159 22611 162 22612 20370 22613 20371 22614 20372 22615 20373 22616 20374 22617 943 22618 20376 22619 20377 22620 20378 22621 976 22622 159 22623 162 22624 20382 22625 20383 22626 2 22627 282 22628 20386 22629 2 22630 282 22631 20389 22632 20390 22633 20391 22634 20392 22635 20393 22636 20394 22637 1862 22638 162 22639 20397 22640 20398 22641 20399 22642 20400 22643 282 22644 20402 22645 20403 22646 2 22647 574 22648 20406 22649 5236 22650 20408 22651 20409 22652 20635 22653 7315 22654 20635 22655 7315 22656 20643 22657 162 22658 17462 22659 20650 22660 22660 22661 22661 22662 20653 22663 7315 22664 22664 22665 3042 22666 22666 22667 20856 22668 11 22669 2 22670 282 22671 7315 22672 22672 22673 16 22674 21004 22675 17462 22676 21008 22677 22677 22678 22678 22679 21011 22680 7315 22681 22681 22682 7315 22683 22683 22684 7315 22685 22685 22686 22686 22687 17352 22688 140 22689 11 22690 142 22691 162 22692 165 22693 167 22694 77 22695 126 22696 170 22697 22697 22698 22698 22699 22699 22700 16 22701 2962 22702 2964 22703 2966 22704 77 22705 2968 22706 2969 22707 22707 22708 22708 22709 1066 22710 7315 22711 22711 22712 17384 22713 17385 22714 7315 22715 22715 22716 2 22717 282 22718 17462 22719 17463 22720 22720 22721 22721 22722 17466 22723 7315 22724 22724 22725 22 22726 159 22727 162 22728 7315 22729 22729 22730 126 22731 170 22732 22732 22733 22733 22734 11 22735 77 22736 22736 22737 126 22738 170 22739 22739 22740 13 22741 17635 22742 162 22743 77 22744 126 22745 170 22746 22746 22747 2966 22748 77 22749 2968 22750 2969 22751 22751 22752 2 22753 282 22754 17648 22755 17649 22756 22756 22757 17651 22758 17652 22759 17653 22760 22760 22761 17653 22762 22762 22763 17788 22764 17789 22765 7315 22766 22766 22767 17789 22768 7315 22769 22769 22770 22770 22771 3042 22772 22772 22773 7315 22774 22774 22775 21294 22776 2966 22777 77 22778 2968 22779 2969 22780 22780 22781 2 22782 282 22783 17462 22784 22784 22785 162 22786 7315 22787 22787 22788 7315 22789 22789 22790 3042 22791 22791 22792 2 22793 19059 22794 20410 22795 20421 22796 17462 22797 20410 22798 162 22799 17462 22800 20472 22801 22801 22802 22802 22803 20475 22804 7315 22805 22805 22806 22806 22807 20635 22808 7315 22809 20635 22810 7315 22811 20643 22812 162 22813 17462 22814 20650 22815 22815 22816 22816 22817 20653 22818 7315 22819 22819 22820 21343 22821 2964 22822 2966 22823 77 22824 2968 22825 2969 22826 22826 22827 3042 22828 22828 22829 20856 22830 11 22831 2 22832 282 22833 7315 22834 22834 22835 16 22836 21004 22837 17462 22838 21008 22839 22839 22840 22840 22841 21011 22842 7315 22843 22843 22844 7315 22845 22845 22846 2962 22847 2964 22848 2966 22849 77 22850 2968 22851 2969 22852 22852 22853 7315 22854 22854 22855 2 22856 3 22857 4 22858 2 22859 2 22860 2 22861 2 22862 11 22863 2 22864 2 22865 22865 22866 2 22867 22867 22868 22 22869 22869 22870 22870 22871 2 22872 11 22873 11 22874 22874 22875 22875 22876 22876 22877 22 22878 2 22879 2 22880 2 22881 11 22882 22882 22883 11 22884 22884 22885 11 22886 22886 22887 22887 22888 22888 22889 22889 22890 60 22891 22891 22892 2 22893 2 22894 22894 22895 11 22896 2 22897 2 22898 69 22899 131 22900 22900 22901 2 22902 2 22903 22903 22904 2 22905 22905 22906 22906 22907 2 22908 85 22909 86 22910 89 22911 134 22912 11 22913 22913 22914 22914 22915 99 22916 22916 22917 22917 22918 22918 22919 104 22920 104 22921 22921 22922 22922 22923 22923 22924 22924 22925 22925 22926 22926 22927 22927 22928 22928 22929 22929 22930 22930 22931 22931 22932 99 22933 2 22934 22934 22935 2 22936 2 22937 2 22938 2 22939 2 22940 2 22941 2 22942 2 22943 307 22944 2 22945 2 22946 2 22947 2 22948 2 22949 313 22950 2 22951 2 22952 2 22953 2 22954 2 22955 2 22956 2 22957 2 22958 2 22959 2 22960 2 22961 2 22962 2 22963 2 22964 2 22965 22965 22966 22966 22967 22967 22968 22968 22969 22969 22970 22970 22971 22971 22972 22972 22973 22973 22974 22974 22975 22975 22976 22976 22977 22977 22978 22978 22979 22979 22980 4 22981 2 22982 22982 22983 22983 22984 22984 22985 22985 22986 22986 22987 22987 22988 22988 22989 22989 22990 22990 22991 22991 22992 22992 22993 357 22994 357 22995 40 22996 360 22997 434 22998 22998 22999 22999 23000 23000 23001 23001 23002 23002 23003 2 23004 574 23005 23005 23006 23006 23007 40 23008 23008 23009 282 23010 23010 23011 23011 23012 2 23013 574 23014 23014 23015 23015 23016 23016 23017 23017 23018 23018 23019 23019 23020 2 23021 740 23022 2 23023 23023 23024 23024 23025 23025 23026 23026 23027 23027 23028 23028 23029 23029 23030 23030 23031 23031 23032 23032 23033 855 23034 23034 23035 23035 23036 23036 23037 23037 23038 23038 23039 23039 23040 23040 23041 23041 23042 23042 23043 23043 23044 23044 23045 23045 23046 23046 23047 23047 23048 23048 23049 23049 23050 23050 23051 23051 23052 23052 23053 23053 23054 23054 23055 23055 23056 23056 23057 23057 23058 23058 23059 23059 23060 23060 23061 23061 23062 23062 23063 23063 23064 23064 23065 23065 23066 23066 23067 23067 23068 23068 23069 23069 23070 23070 23071 23071 23072 23072 23073 23073 23074 23074 23075 23075 23076 23076 23077 23077 23078 23078 23079 23079 23080 23080 23081 23081 23082 23082 23083 23083 23084 23084 23085 23085 23086 23086 23087 23087 23088 23088 23089 23089 23090 23090 23091 23091 23092 23092 23093 23093 23094 23094 23095 23095 23096 23096 23097 23097 23098 23098 23099 23099 23100 23100 23101 23101 23102 23102 23103 23103 23104 23104 23105 23105 23106 23106 23107 23107 23108 23108 23109 23109 23110 23110 23111 23111 23112 23112 23113 23113 23114 11 23115 23115 23116 23116 23117 23117 23118 2 23119 574 23120 23120 23121 943 23122 23122 23123 23123 23124 976 23125 159 23126 162 23127 23127 23128 23128 23129 1487 23130 2 23131 2 23132 2 23133 23133 23134 2 23135 23135 23136 2 23137 2 23138 2 23139 2 23140 2 23141 2 23142 2 23143 2 23144 2 23145 23145 23146 23146 23147 23147 23148 23148 23149 23149 23150 23150 23151 23151 23152 23152 23153 23153 23154 23154 23155 23155 23156 23156 23157 23157 23158 22 23159 23159 23160 4 23161 11 23162 23162 23163 89 23164 11 23165 23165 23166 99 23167 23167 23168 1551 23169 23169 23170 23170 23171 23171 23172 99 23173 855 23174 11 23175 23175 23176 943 23177 23177 23178 23178 23179 976 23180 159 23181 162 23182 23182 23183 23183 23184 23184 23185 855 23186 11 23187 23187 23188 23188 23189 23189 23190 357 23191 357 23192 360 23193 434 23194 23194 23195 23195 23196 23196 23197 23197 23198 2 23199 574 23200 23200 23201 23201 23202 23202 23203 23203 23204 23204 23205 23205 23206 2 23207 574 23208 23208 23209 23209 23210 140 23211 11 23212 142 23213 162 23214 163 23215 165 23216 167 23217 77 23218 126 23219 170 23220 23220 23221 23221 23222 140 23223 11 23224 142 23225 162 23226 2248 23227 2250 23228 2251 23229 2252 23230 2 23231 2 23232 3 23233 4 23234 2 23235 2 23236 2 23237 2 23238 11 23239 2 23240 2 23241 22865 23242 2 23243 22867 23244 22 23245 22869 23246 22870 23247 2 23248 11 23249 11 23250 22874 23251 22875 23252 22876 23253 22 23254 2 23255 2 23256 2 23257 11 23258 22882 23259 11 23260 22884 23261 11 23262 22886 23263 22887 23264 22888 23265 22889 23266 60 23267 22891 23268 2 23269 2 23270 22894 23271 11 23272 2 23273 2 23274 69 23275 131 23276 22900 23277 2 23278 2 23279 22903 23280 2 23281 22905 23282 22906 23283 2 23284 85 23285 86 23286 89 23287 134 23288 11 23289 22913 23290 22914 23291 99 23292 22916 23293 22917 23294 23294 23295 104 23296 104 23297 22921 23298 22922 23299 22923 23300 22924 23301 22925 23302 22926 23303 22927 23304 22928 23305 22929 23306 22930 23307 22931 23308 99 23309 2 23310 22934 23311 2 23312 2 23313 2 23314 2 23315 2 23316 2 23317 2 23318 2 23319 307 23320 2 23321 2 23322 2 23323 2 23324 2 23325 313 23326 2 23327 2 23328 2 23329 2 23330 2 23331 2 23332 2 23333 2 23334 2 23335 2 23336 2 23337 2 23338 2 23339 2 23340 2 23341 22965 23342 22966 23343 22967 23344 22968 23345 22969 23346 22970 23347 22971 23348 22972 23349 22973 23350 22974 23351 22975 23352 22976 23353 22977 23354 22978 23355 22979 23356 4 23357 2 23358 22982 23359 22983 23360 22984 23361 22985 23362 22986 23363 22987 23364 22988 23365 22989 23366 22990 23367 22991 23368 22992 23369 357 23370 357 23371 40 23372 360 23373 434 23374 22998 23375 22999 23376 23000 23377 23001 23378 23002 23379 2 23380 574 23381 23005 23382 23006 23383 40 23384 23008 23385 282 23386 23010 23387 23011 23388 2 23389 574 23390 23014 23391 23015 23392 23016 23393 23017 23394 23018 23395 23019 23396 2 23397 740 23398 2 23399 23023 23400 23024 23401 23025 23402 23026 23403 23027 23404 23028 23405 23029 23406 23030 23407 23031 23408 23032 23409 855 23410 23034 23411 23035 23412 23036 23413 23037 23414 23038 23415 23039 23416 23040 23417 23041 23418 23042 23419 23043 23420 23044 23421 23045 23422 23046 23423 23047 23424 23048 23425 23049 23426 23050 23427 23051 23428 23052 23429 23053 23430 23054 23431 23055 23432 23056 23433 23057 23434 23058 23435 23059 23436 23060 23437 23061 23438 23062 23439 23063 23440 23064 23441 23065 23442 23066 23443 23067 23444 23068 23445 23069 23446 23070 23447 23071 23448 23072 23449 23073 23450 23074 23451 23075 23452 23076 23453 23077 23454 23078 23455 23079 23456 23080 23457 23081 23458 23082 23459 23083 23460 23084 23461 23085 23462 23086 23463 23087 23464 23088 23465 23089 23466 23090 23467 23091 23468 23092 23469 23093 23470 23094 23471 23095 23472 23096 23473 23097 23474 23098 23475 23099 23476 23100 23477 23101 23478 23102 23479 23103 23480 23104 23481 23105 23482 23106 23483 23107 23484 23108 23485 23109 23486 23110 23487 23111 23488 23112 23489 23113 23490 11 23491 23115 23492 23116 23493 23117 23494 2 23495 574 23496 23120 23497 943 23498 23122 23499 23123 23500 976 23501 159 23502 162 23503 23127 23504 23128 23505 1487 23506 2 23507 2 23508 2 23509 23133 23510 2 23511 23135 23512 2 23513 2 23514 2 23515 2 23516 2 23517 2 23518 2 23519 2 23520 2 23521 23145 23522 23146 23523 23147 23524 23148 23525 23149 23526 23150 23527 23151 23528 23152 23529 23153 23530 23154 23531 23155 23532 23156 23533 23157 23534 22 23535 23159 23536 4 23537 11 23538 23162 23539 89 23540 11 23541 23165 23542 99 23543 23167 23544 1551 23545 23169 23546 23170 23547 23171 23548 99 23549 855 23550 11 23551 23175 23552 943 23553 23177 23554 23178 23555 976 23556 159 23557 162 23558 23182 23559 23183 23560 23184 23561 855 23562 11 23563 23187 23564 23188 23565 23189 23566 357 23567 357 23568 360 23569 434 23570 23194 23571 23195 23572 23196 23573 23197 23574 2 23575 574 23576 23200 23577 23201 23578 23202 23579 23203 23580 23204 23581 23205 23582 2 23583 574 23584 23208 23585 23209 23586 140 23587 11 23588 142 23589 162 23590 163 23591 165 23592 167 23593 77 23594 126 23595 170 23596 23220 23597 23221 23598 140 23599 11 23600 142 23601 162 23602 2248 23603 2250 23604 2251 23605 2252 23606 2 23607 23294 23608 23608 23609 23608 23610 855 23611 23611 23612 11 23613 23613 23614 2 23615 574 23616 23616 23617 23617 23618 23618 23619 2732 23620 23620 23621 23621 23622 357 23623 357 23624 360 23625 434 23626 23626 23627 23627 23628 23628 23629 23629 23630 2 23631 574 23632 23632 23633 23633 23634 23634 23635 23635 23636 2803 23637 23637 23638 434 23639 23639 23640 2 23641 23641 23642 23642 23643 2 23644 3 23645 4 23646 2 23647 2 23648 2 23649 2 23650 11 23651 2 23652 2 23653 23653 23654 2 23655 23655 23656 22 23657 23657 23658 23658 23659 2 23660 11 23661 11 23662 23662 23663 23663 23664 23664 23665 22 23666 2 23667 2 23668 2 23669 11 23670 23670 23671 11 23672 23672 23673 11 23674 23674 23675 23675 23676 23676 23677 23677 23678 60 23679 23679 23680 2 23681 2 23682 23682 23683 11 23684 2 23685 2 23686 69 23687 131 23688 23688 23689 2 23690 2 23691 23691 23692 2 23693 23693 23694 23694 23695 2 23696 85 23697 86 23698 89 23699 134 23700 11 23701 23701 23702 23702 23703 99 23704 23704 23705 23705 23706 23706 23707 104 23708 104 23709 23709 23710 23710 23711 23711 23712 23712 23713 23713 23714 23714 23715 23715 23716 23716 23717 23717 23718 23718 23719 23719 23720 99 23721 2 23722 23722 23723 2 23724 2 23725 2 23726 2 23727 2 23728 2 23729 2 23730 2 23731 307 23732 2 23733 2 23734 2 23735 2 23736 2 23737 313 23738 2 23739 2 23740 2 23741 2 23742 2 23743 2 23744 2 23745 2 23746 2 23747 2 23748 2 23749 2 23750 2 23751 2 23752 2 23753 23753 23754 23754 23755 23755 23756 23756 23757 23757 23758 23758 23759 23759 23760 23760 23761 23761 23762 23762 23763 23763 23764 23764 23765 23765 23766 23766 23767 23767 23768 4 23769 2 23770 23770 23771 23771 23772 23772 23773 23773 23774 23774 23775 23775 23776 23776 23777 23777 23778 23778 23779 23779 23780 23780 23781 357 23782 357 23783 40 23784 360 23785 434 23786 23786 23787 23787 23788 23788 23789 23789 23790 23790 23791 2 23792 574 23793 23793 23794 23794 23795 40 23796 23796 23797 282 23798 23798 23799 23799 23800 2 23801 574 23802 23802 23803 23803 23804 23804 23805 23805 23806 23806 23807 23807 23808 2 23809 740 23810 2 23811 23811 23812 23812 23813 23813 23814 23814 23815 23815 23816 23816 23817 23817 23818 23818 23819 23819 23820 23820 23821 855 23822 23822 23823 23823 23824 23824 23825 23825 23826 23826 23827 23827 23828 23828 23829 23829 23830 23830 23831 23831 23832 23832 23833 23833 23834 23834 23835 23835 23836 23836 23837 23837 23838 23838 23839 23839 23840 23840 23841 23841 23842 23842 23843 23843 23844 23844 23845 23845 23846 23846 23847 23847 23848 23848 23849 23849 23850 23850 23851 23851 23852 23852 23853 23853 23854 23854 23855 23855 23856 23856 23857 23857 23858 23858 23859 23859 23860 23860 23861 23861 23862 23862 23863 23863 23864 23864 23865 23865 23866 23866 23867 23867 23868 23868 23869 23869 23870 23870 23871 23871 23872 23872 23873 23873 23874 23874 23875 23875 23876 23876 23877 23877 23878 23878 23879 23879 23880 23880 23881 23881 23882 23882 23883 23883 23884 23884 23885 23885 23886 23886 23887 23887 23888 23888 23889 23889 23890 23890 23891 23891 23892 23892 23893 23893 23894 23894 23895 23895 23896 23896 23897 23897 23898 23898 23899 23899 23900 23900 23901 23901 23902 11 23903 23903 23904 23904 23905 23905 23906 2 23907 574 23908 23908 23909 943 23910 23910 23911 23911 23912 976 23913 159 23914 162 23915 23915 23916 23916 23917 1487 23918 2 23919 2 23920 2 23921 23921 23922 2 23923 23923 23924 2 23925 2 23926 2 23927 2 23928 2 23929 2 23930 2 23931 2 23932 2 23933 23933 23934 23934 23935 23935 23936 23936 23937 23937 23938 23938 23939 23939 23940 23940 23941 23941 23942 23942 23943 23943 23944 23944 23945 23945 23946 22 23947 23947 23948 4 23949 11 23950 23950 23951 89 23952 11 23953 23953 23954 99 23955 23955 23956 1551 23957 23957 23958 23958 23959 23959 23960 99 23961 855 23962 11 23963 23963 23964 943 23965 23965 23966 23966 23967 976 23968 159 23969 162 23970 23970 23971 23971 23972 23972 23973 855 23974 11 23975 23975 23976 23976 23977 23977 23978 357 23979 357 23980 360 23981 434 23982 23982 23983 23983 23984 23984 23985 23985 23986 2 23987 574 23988 23988 23989 23989 23990 23990 23991 23991 23992 23992 23993 23993 23994 2 23995 574 23996 23996 23997 23997 23998 140 23999 11 24000 142 24001 162 24002 163 24003 165 24004 167 24005 77 24006 126 24007 170 24008 24008 24009 24009 24010 140 24011 11 24012 142 24013 162 24014 2248 24015 2250 24016 2251 24017 2252 24018 2 24019 24019 24020 24020 24021 2 24022 24022 24023 855 24024 24024 24025 11 24026 24026 24027 2 24028 574 24029 24029 24030 24030 24031 24031 24032 2732 24033 24033 24034 24034 24035 357 24036 357 24037 360 24038 434 24039 24039 24040 24040 24041 24041 24042 24042 24043 2 24044 574 24045 24045 24046 24046 24047 24047 24048 24048 24049 2803 24050 24050 24051 434 24052 24052 24053 2 24054 24054 24055 24055 24056 23706 24057 2 24058 24058 24059 24059 24060 24060 24061 24061 24062 24062 24063 2894 24064 24064 24065 69 24066 24066 24067 24067 24068 24068 24069 24069 24070 24070 24071 24071 24072 24072 24073 2 24074 24074 24075 24075 24076 24076 24077 24077 24078 24078 24079 24079 24080 2936 24081 2938 24082 24082 24083 24083 24084 24084 24085 24085 24086 24086 24087 357 24088 357 24089 40 24090 360 24091 434 24092 24092 24093 24093 24094 24094 24095 24095 24096 24096 24097 40 24098 24098 24099 24099 24100 2959 24101 2962 24102 2964 24103 2966 24104 77 24105 2968 24106 2969 24107 24107 24108 282 24109 24109 24110 2 24111 282 24112 77 24113 3931 24114 2828 24115 24115 24116 1862 24117 162 24118 2050 24119 2 24120 24120 24121 2 24122 159 24123 162 24124 24124 24125 24125 24126 24126 24127 24127 24128 24128 24129 943 24130 24130 24131 24131 24132 24132 24133 976 24134 159 24135 162 24136 24136 24137 24137 24138 2 24139 282 24140 24140 24141 2 24142 282 24143 24143 24144 24144 24145 24145 24146 24146 24147 24147 24148 24148 24149 1862 24150 162 24151 24151 24152 24152 24153 24153 24154 24154 24155 282 24156 24156 24157 24157 24158 2 24159 574 24160 24160 24161 5236 24162 24162 24163 24163 24164 24164 24165 24165 24166 24166 24167 24167 24168 2894 24169 24169 24170 69 24171 24171 24172 24172 24173 24173 24174 24174 24175 24175 24176 24176 24177 24177 24178 2 24179 24179 24180 24180 24181 24181 24182 24182 24183 24183 24184 24184 24185 2936 24186 2938 24187 24187 24188 24188 24189 24189 24190 24190 24191 24191 24192 357 24193 357 24194 40 24195 360 24196 434 24197 24197 24198 24198 24199 24199 24200 24200 24201 24201 24202 40 24203 24203 24204 24204 24205 2959 24206 2962 24207 2964 24208 2966 24209 77 24210 2968 24211 2969 24212 24212 24213 282 24214 24214 24215 2 24216 282 24217 77 24218 3931 24219 2828 24220 24220 24221 1862 24222 162 24223 2050 24224 2 24225 24225 24226 2 24227 159 24228 162 24229 24229 24230 24230 24231 24231 24232 24232 24233 24233 24234 943 24235 24235 24236 24236 24237 24237 24238 976 24239 159 24240 162 24241 24241 24242 24242 24243 2 24244 282 24245 24245 24246 2 24247 282 24248 24248 24249 24249 24250 24250 24251 24251 24252 24252 24253 24253 24254 1862 24255 162 24256 24256 24257 24257 24258 24258 24259 24259 24260 282 24261 24261 24262 24262 24263 2 24264 574 24265 24265 24266 5236 24267 24267 24268 24268 24269 17796 24270 24270 24271 855 24272 23611 24273 11 24274 23613 24275 2 24276 574 24277 23616 24278 23617 24279 23618 24280 2732 24281 23620 24282 23621 24283 357 24284 357 24285 360 24286 434 24287 23626 24288 23627 24289 23628 24290 23629 24291 2 24292 574 24293 23632 24294 23633 24295 23634 24296 23635 24297 2803 24298 23637 24299 434 24300 23639 24301 2 24302 23641 24303 23642 24304 2 24305 3 24306 4 24307 2 24308 2 24309 2 24310 2 24311 11 24312 2 24313 2 24314 23653 24315 2 24316 23655 24317 22 24318 23657 24319 23658 24320 2 24321 11 24322 11 24323 23662 24324 23663 24325 23664 24326 22 24327 2 24328 2 24329 2 24330 11 24331 23670 24332 11 24333 23672 24334 11 24335 23674 24336 23675 24337 23676 24338 23677 24339 60 24340 23679 24341 2 24342 2 24343 23682 24344 11 24345 2 24346 2 24347 69 24348 131 24349 23688 24350 2 24351 2 24352 23691 24353 2 24354 23693 24355 23694 24356 2 24357 85 24358 86 24359 89 24360 134 24361 11 24362 23701 24363 23702 24364 99 24365 23704 24366 23705 24367 23706 24368 104 24369 104 24370 23709 24371 23710 24372 23711 24373 23712 24374 23713 24375 23714 24376 23715 24377 23716 24378 23717 24379 23718 24380 23719 24381 99 24382 2 24383 23722 24384 2 24385 2 24386 2 24387 2 24388 2 24389 2 24390 2 24391 2 24392 307 24393 2 24394 2 24395 2 24396 2 24397 2 24398 313 24399 2 24400 2 24401 2 24402 2 24403 2 24404 2 24405 2 24406 2 24407 2 24408 2 24409 2 24410 2 24411 2 24412 2 24413 2 24414 23753 24415 23754 24416 23755 24417 23756 24418 23757 24419 23758 24420 23759 24421 23760 24422 23761 24423 23762 24424 23763 24425 23764 24426 23765 24427 23766 24428 23767 24429 4 24430 2 24431 23770 24432 23771 24433 23772 24434 23773 24435 23774 24436 23775 24437 23776 24438 23777 24439 23778 24440 23779 24441 23780 24442 357 24443 357 24444 40 24445 360 24446 434 24447 23786 24448 23787 24449 23788 24450 23789 24451 23790 24452 2 24453 574 24454 23793 24455 23794 24456 40 24457 23796 24458 282 24459 23798 24460 23799 24461 2 24462 574 24463 23802 24464 23803 24465 23804 24466 23805 24467 23806 24468 23807 24469 2 24470 740 24471 2 24472 23811 24473 23812 24474 23813 24475 23814 24476 23815 24477 23816 24478 23817 24479 23818 24480 23819 24481 23820 24482 855 24483 23822 24484 23823 24485 23824 24486 23825 24487 23826 24488 23827 24489 23828 24490 23829 24491 23830 24492 23831 24493 23832 24494 23833 24495 23834 24496 23835 24497 23836 24498 23837 24499 23838 24500 23839 24501 23840 24502 23841 24503 23842 24504 23843 24505 23844 24506 23845 24507 23846 24508 23847 24509 23848 24510 23849 24511 23850 24512 23851 24513 23852 24514 23853 24515 23854 24516 23855 24517 23856 24518 23857 24519 23858 24520 23859 24521 23860 24522 23861 24523 23862 24524 23863 24525 23864 24526 23865 24527 23866 24528 23867 24529 23868 24530 23869 24531 23870 24532 23871 24533 23872 24534 23873 24535 23874 24536 23875 24537 23876 24538 23877 24539 23878 24540 23879 24541 23880 24542 23881 24543 23882 24544 23883 24545 23884 24546 23885 24547 23886 24548 23887 24549 23888 24550 23889 24551 23890 24552 23891 24553 23892 24554 23893 24555 23894 24556 23895 24557 23896 24558 23897 24559 23898 24560 23899 24561 23900 24562 23901 24563 11 24564 23903 24565 23904 24566 23905 24567 2 24568 574 24569 23908 24570 943 24571 23910 24572 23911 24573 976 24574 159 24575 162 24576 23915 24577 23916 24578 1487 24579 2 24580 2 24581 2 24582 23921 24583 2 24584 23923 24585 2 24586 2 24587 2 24588 2 24589 2 24590 2 24591 2 24592 2 24593 2 24594 23933 24595 23934 24596 23935 24597 23936 24598 23937 24599 23938 24600 23939 24601 23940 24602 23941 24603 23942 24604 23943 24605 23944 24606 23945 24607 22 24608 23947 24609 4 24610 11 24611 23950 24612 89 24613 11 24614 23953 24615 99 24616 23955 24617 1551 24618 23957 24619 23958 24620 23959 24621 99 24622 855 24623 11 24624 23963 24625 943 24626 23965 24627 23966 24628 976 24629 159 24630 162 24631 23970 24632 23971 24633 23972 24634 855 24635 11 24636 23975 24637 23976 24638 23977 24639 357 24640 357 24641 360 24642 434 24643 23982 24644 23983 24645 23984 24646 23985 24647 2 24648 574 24649 23988 24650 23989 24651 23990 24652 23991 24653 23992 24654 23993 24655 2 24656 574 24657 23996 24658 23997 24659 140 24660 11 24661 142 24662 162 24663 163 24664 165 24665 167 24666 77 24667 126 24668 170 24669 24008 24670 24009 24671 140 24672 11 24673 142 24674 162 24675 2248 24676 2250 24677 2251 24678 2252 24679 2 24680 24019 24681 24020 24682 2 24683 24022 24684 855 24685 24024 24686 11 24687 24026 24688 2 24689 574 24690 24029 24691 24030 24692 24031 24693 2732 24694 24033 24695 24034 24696 357 24697 357 24698 360 24699 434 24700 24039 24701 24040 24702 24041 24703 24042 24704 2 24705 574 24706 24045 24707 24046 24708 24047 24709 24048 24710 2803 24711 24050 24712 434 24713 24052 24714 2 24715 24054 24716 24055 24717 23706 24718 2 24719 24058 24720 24059 24721 24060 24722 24061 24723 24062 24724 2894 24725 24064 24726 69 24727 24066 24728 24067 24729 24068 24730 24069 24731 24070 24732 24071 24733 24072 24734 2 24735 24074 24736 24075 24737 24076 24738 24077 24739 24078 24740 24079 24741 2936 24742 2938 24743 24082 24744 24083 24745 24084 24746 24085 24747 24086 24748 357 24749 357 24750 40 24751 360 24752 434 24753 24092 24754 24093 24755 24094 24756 24095 24757 24096 24758 40 24759 24098 24760 24099 24761 2959 24762 2962 24763 2964 24764 2966 24765 77 24766 2968 24767 2969 24768 24107 24769 282 24770 24109 24771 2 24772 282 24773 77 24774 3931 24775 2828 24776 24115 24777 1862 24778 162 24779 2050 24780 2 24781 24120 24782 2 24783 159 24784 162 24785 24124 24786 24125 24787 24126 24788 24127 24789 24128 24790 943 24791 24130 24792 24131 24793 24132 24794 976 24795 159 24796 162 24797 24136 24798 24137 24799 2 24800 282 24801 24140 24802 2 24803 282 24804 24143 24805 24144 24806 24145 24807 24146 24808 24147 24809 24148 24810 1862 24811 162 24812 24151 24813 24152 24814 24153 24815 24154 24816 282 24817 24156 24818 24157 24819 2 24820 574 24821 24160 24822 5236 24823 24162 24824 24163 24825 24164 24826 24165 24827 24166 24828 24167 24829 2894 24830 24169 24831 69 24832 24171 24833 24172 24834 24173 24835 24174 24836 24175 24837 24176 24838 24177 24839 2 24840 24179 24841 24180 24842 24181 24843 24182 24844 24183 24845 24184 24846 2936 24847 2938 24848 24187 24849 24188 24850 24189 24851 24190 24852 24191 24853 357 24854 357 24855 40 24856 360 24857 434 24858 24197 24859 24198 24860 24199 24861 24200 24862 24201 24863 40 24864 24203 24865 24204 24866 2959 24867 2962 24868 2964 24869 2966 24870 77 24871 2968 24872 2969 24873 24212 24874 282 24875 24214 24876 2 24877 282 24878 77 24879 3931 24880 2828 24881 24220 24882 1862 24883 162 24884 2050 24885 2 24886 24225 24887 2 24888 159 24889 162 24890 24229 24891 24230 24892 24231 24893 24232 24894 24233 24895 943 24896 24235 24897 24236 24898 24237 24899 976 24900 159 24901 162 24902 24241 24903 24242 24904 2 24905 282 24906 24245 24907 2 24908 282 24909 24248 24910 24249 24911 24250 24912 24251 24913 24252 24914 24253 24915 1862 24916 162 24917 24256 24918 24257 24919 24258 24920 24259 24921 282 24922 24261 24923 24262 24924 2 24925 574 24926 24265 24927 5236 24928 24267 24929 24268 24930 17789 24931 7315 24932 24932 24933 24933 24934 3042 24935 24935 24936 7315 24937 24937 24938 7315 24939 24939 24940 3042 24941 24941 24942 2 24943 19059 24944 20410 24945 20421 24946 17462 24947 20410 24948 162 24949 17462 24950 20472 24951 24951 24952 24952 24953 20475 24954 7315 24955 24955 24956 24956 24957 20635 24958 7315 24959 20635 24960 7315 24961 20643 24962 162 24963 17462 24964 20650 24965 24965 24966 24966 24967 20653 24968 7315 24969 24969 24970 3042 24971 24971 24972 20856 24973 11 24974 2 24975 282 24976 7315 24977 24977 24978 16 24979 21004 24980 17462 24981 21008 24982 24982 24983 24983 24984 21011 24985 7315 24986 24986 24987 7315 24988 24988 24989 7315 24990 24990 24991 24991 24992 17352 24993 140 24994 11 24995 142 24996 162 24997 165 24998 167 24999 77 25000 126 25001 170 25002 25002 25003 25003 25004 25004 25005 16 25006 2962 25007 2964 25008 2966 25009 77 25010 2968 25011 2969 25012 25012 25013 25013 25014 1066 25015 7315 25016 25016 25017 17384 25018 17385 25019 7315 25020 25020 25021 2 25022 282 25023 17462 25024 17463 25025 25025 25026 25026 25027 17466 25028 7315 25029 25029 25030 22 25031 159 25032 162 25033 7315 25034 25034 25035 126 25036 170 25037 25037 25038 25038 25039 11 25040 77 25041 25041 25042 126 25043 170 25044 25044 25045 13 25046 17635 25047 162 25048 77 25049 126 25050 170 25051 25051 25052 2966 25053 77 25054 2968 25055 2969 25056 25056 25057 2 25058 282 25059 17648 25060 17649 25061 25061 25062 17651 25063 17652 25064 17653 25065 25065 25066 17653 25067 25067 25068 17788 25069 17789 25070 7315 25071 25071 25072 17789 25073 7315 25074 25074 25075 25075 25076 3042 25077 25077 25078 7315 25079 25079 25080 21294 25081 2966 25082 77 25083 2968 25084 2969 25085 25085 25086 2 25087 282 25088 17462 25089 25089 25090 162 25091 7315 25092 25092 25093 7315 25094 25094 25095 3042 25096 25096 25097 2 25098 19059 25099 20410 25100 20421 25101 17462 25102 20410 25103 162 25104 17462 25105 20472 25106 25106 25107 25107 25108 20475 25109 7315 25110 25110 25111 25111 25112 20635 25113 7315 25114 20635 25115 7315 25116 20643 25117 162 25118 17462 25119 20650 25120 25120 25121 25121 25122 20653 25123 7315 25124 25124 25125 21343 25126 2964 25127 2966 25128 77 25129 2968 25130 2969 25131 25131 25132 3042 25133 25133 25134 20856 25135 11 25136 2 25137 282 25138 7315 25139 25139 25140 16 25141 21004 25142 17462 25143 21008 25144 25144 25145 25145 25146 21011 25147 7315 25148 25148 25149 7315 25150 25150 25151 2962 25152 2964 25153 2966 25154 77 25155 2968 25156 2969 25157 25157 25158 7315 25159 25159 25160 22918 25161 17659 25162 25162 25163 17653 25164 25164 25165 17788 25166 17789 25167 7315 25168 25168 25169 17789 25170 7315 25171 25171 25172 25172 25173 3042 25174 25174 25175 7315 25176 25176 25177 7315 25178 25178 25179 3042 25180 25180 25181 2 25182 19059 25183 20410 25184 20421 25185 17462 25186 20410 25187 162 25188 17462 25189 20472 25190 25190 25191 25191 25192 20475 25193 7315 25194 25194 25195 25195 25196 20635 25197 7315 25198 20635 25199 7315 25200 20643 25201 162 25202 17462 25203 20650 25204 25204 25205 25205 25206 20653 25207 7315 25208 25208 25209 3042 25210 25210 25211 20856 25212 11 25213 2 25214 282 25215 7315 25216 25216 25217 16 25218 21004 25219 17462 25220 21008 25221 25221 25222 25222 25223 21011 25224 7315 25225 25225 25226 7315 25227 25227 25228 7315 25229 25229 25230 25230 25231 17352 25232 140 25233 11 25234 142 25235 162 25236 165 25237 167 25238 77 25239 126 25240 170 25241 25241 25242 25242 25243 25243 25244 16 25245 2962 25246 2964 25247 2966 25248 77 25249 2968 25250 2969 25251 25251 25252 25252 25253 1066 25254 7315 25255 25255 25256 17384 25257 17385 25258 7315 25259 25259 25260 2 25261 282 25262 17462 25263 17463 25264 25264 25265 25265 25266 17466 25267 7315 25268 25268 25269 22 25270 159 25271 162 25272 7315 25273 25273 25274 126 25275 170 25276 25276 25277 25277 25278 11 25279 77 25280 25280 25281 126 25282 170 25283 25283 25284 13 25285 17635 25286 162 25287 77 25288 126 25289 170 25290 25290 25291 2966 25292 77 25293 2968 25294 2969 25295 25295 25296 2 25297 282 25298 17648 25299 17649 25300 25300 25301 17651 25302 17652 25303 17653 25304 25304 25305 17653 25306 25306 25307 17788 25308 17789 25309 7315 25310 25310 25311 17789 25312 7315 25313 25313 25314 25314 25315 3042 25316 25316 25317 7315 25318 25318 25319 21294 25320 2966 25321 77 25322 2968 25323 2969 25324 25324 25325 2 25326 282 25327 17462 25328 25328 25329 162 25330 7315 25331 25331 25332 7315 25333 25333 25334 3042 25335 25335 25336 2 25337 19059 25338 20410 25339 20421 25340 17462 25341 20410 25342 162 25343 17462 25344 20472 25345 25345 25346 25346 25347 20475 25348 7315 25349 25349 25350 25350 25351 20635 25352 7315 25353 20635 25354 7315 25355 20643 25356 162 25357 17462 25358 20650 25359 25359 25360 25360 25361 20653 25362 7315 25363 25363 25364 21343 25365 2964 25366 2966 25367 77 25368 2968 25369 2969 25370 25370 25371 25371 25372 3042 25373 25373 25374 20856 25375 11 25376 2 25377 282 25378 7315 25379 25379 25380 16 25381 21004 25382 17462 25383 21008 25384 25384 25385 25385 25386 21011 25387 7315 25388 25388 25389 7315 25390 25390 25391 2962 25392 2964 25393 2966 25394 77 25395 2968 25396 2969 25397 25397 25398 7315 25399 25399 25400 22 25401 159 25402 162 25403 7315 25404 25404 25405 126 25406 170 25407 25407 25408 25408 25409 11 25410 77 25411 25411 25412 126 25413 170 25414 25414 25415 13 25416 17635 25417 162 25418 77 25419 126 25420 170 25421 25421 25422 2966 25423 77 25424 2968 25425 2969 25426 25426 25427 2 25428 282 25429 17648 25430 17649 25431 25431 25432 17651 25433 17652 25434 17653 25435 25435 25436 17653 25437 25437 25438 17788 25439 17789 25440 7315 25441 25441 25442 17789 25443 7315 25444 25444 25445 25445 25446 3042 25447 25447 25448 7315 25449 25449 25450 21294 25451 2966 25452 77 25453 2968 25454 2969 25455 25455 25456 162 25457 7315 25458 25458 25459 7315 25460 25460 25461 3042 25462 25462 25463 2 25464 19059 25465 20410 25466 20421 25467 17462 25468 20410 25469 162 25470 17462 25471 20472 25472 25472 25473 25473 25474 20475 25475 7315 25476 25476 25477 25477 25478 20635 25479 7315 25480 20635 25481 7315 25482 20643 25483 162 25484 17462 25485 20650 25486 25486 25487 25487 25488 20653 25489 7315 25490 25490 25491 21343 25492 2964 25493 2966 25494 77 25495 2968 25496 2969 25497 25497 25498 25498 25499 3042 25500 25500 25501 20856 25502 11 25503 2 25504 282 25505 7315 25506 25506 25507 16 25508 21004 25509 17462 25510 21008 25511 25511 25512 25512 25513 21011 25514 7315 25515 25515 25516 7315 25517 25517 25518 2962 25519 2964 25520 2966 25521 77 25522 2968 25523 2969 25524 25524 25525 7315 25526 25526 25527 2 25528 282 25529 17462 25530 17463 25531 25531 25532 25532 25533 17466 25534 7315 25535 25535 25536 22 25537 159 25538 162 25539 7315 25540 25540 25541 126 25542 170 25543 25543 25544 25544 25545 11 25546 77 25547 25547 25548 126 25549 170 25550 25550 25551 13 25552 17635 25553 162 25554 77 25555 126 25556 170 25557 25557 25558 2966 25559 77 25560 2968 25561 2969 25562 25562 25563 2 25564 282 25565 17648 25566 17649 25567 25567 25568 17651 25569 17652 25570 17653 25571 25571 25572 17653 25573 25573 25574 17788 25575 17789 25576 7315 25577 25577 25578 17789 25579 7315 25580 25580 25581 25581 25582 3042 25583 25583 25584 7315 25585 25585 25586 21294 25587 2966 25588 77 25589 2968 25590 2969 25591 25591 25592 2 25593 282 25594 17462 25595 25595 25596 162 25597 7315 25598 25598 25599 7315 25600 25600 25601 3042 25602 25602 25603 2 25604 19059 25605 20410 25606 20421 25607 17462 25608 20410 25609 162 25610 17462 25611 20472 25612 25612 25613 25613 25614 20475 25615 7315 25616 25616 25617 25617 25618 20635 25619 7315 25620 20635 25621 7315 25622 20643 25623 162 25624 17462 25625 20650 25626 25626 25627 25627 25628 20653 25629 7315 25630 25630 25631 21343 25632 2964 25633 2966 25634 77 25635 2968 25636 2969 25637 25637 25638 3042 25639 25639 25640 20856 25641 11 25642 2 25643 282 25644 7315 25645 25645 25646 16 25647 21004 25648 17462 25649 21008 25650 25650 25651 25651 25652 21011 25653 7315 25654 25654 25655 7315 25656 25656 25657 2962 25658 2964 25659 2966 25660 77 25661 2968 25662 2969 25663 25663 25664 7315 25665 25665 25666 17384 25667 17385 25668 7315 25669 25669 25670 2 25671 282 25672 17462 25673 17463 25674 25674 25675 25675 25676 17466 25677 7315 25678 25678 25679 22 25680 159 25681 162 25682 7315 25683 25683 25684 126 25685 170 25686 25686 25687 25687 25688 11 25689 77 25690 25690 25691 126 25692 170 25693 25693 25694 13 25695 17635 25696 162 25697 77 25698 126 25699 170 25700 25700 25701 2 25702 282 25703 17648 25704 17649 25705 25705 25706 17651 25707 17652 25708 17653 25709 25709 25710 17653 25711 25711 25712 17788 25713 17789 25714 7315 25715 25715 25716 17789 25717 7315 25718 25718 25719 25719 25720 3042 25721 25721 25722 7315 25723 25723 25724 7315 25725 25725 25726 3042 25727 25727 25728 2 25729 19059 25730 20410 25731 20421 25732 17462 25733 20410 25734 162 25735 17462 25736 20472 25737 25737 25738 25738 25739 20475 25740 7315 25741 25741 25742 25742 25743 20635 25744 7315 25745 20635 25746 7315 25747 20643 25748 162 25749 17462 25750 20650 25751 25751 25752 25752 25753 20653 25754 7315 25755 25755 25756 3042 25757 25757 25758 20856 25759 11 25760 2 25761 282 25762 7315 25763 25763 25764 16 25765 21004 25766 17462 25767 21008 25768 25768 25769 25769 25770 21011 25771 7315 25772 25772 25773 7315 25774 25774 25775 7315 25776 25776 25777 855 25778 16601 25779 11 25780 16603 25781 2 25782 574 25783 16606 25784 16607 25785 16608 25786 2732 25787 16610 25788 16611 25789 357 25790 357 25791 360 25792 434 25793 16616 25794 16617 25795 16618 25796 16619 25797 2 25798 574 25799 16622 25800 16623 25801 16624 25802 16625 25803 2803 25804 16627 25805 434 25806 16629 25807 2 25808 16631 25809 16632 25810 2 25811 3 25812 4 25813 2 25814 2 25815 2 25816 2 25817 11 25818 2 25819 2 25820 16643 25821 2 25822 16645 25823 22 25824 16647 25825 16648 25826 2 25827 11 25828 11 25829 16652 25830 16653 25831 16654 25832 22 25833 2 25834 2 25835 2 25836 11 25837 16660 25838 11 25839 16662 25840 11 25841 16664 25842 16665 25843 16666 25844 16667 25845 60 25846 16669 25847 2 25848 2 25849 16672 25850 11 25851 2 25852 2 25853 69 25854 131 25855 16678 25856 2 25857 2 25858 16681 25859 2 25860 16683 25861 16684 25862 2 25863 85 25864 86 25865 89 25866 134 25867 11 25868 16691 25869 16692 25870 99 25871 16694 25872 16695 25873 16696 25874 104 25875 104 25876 16699 25877 16700 25878 16701 25879 16702 25880 16703 25881 16704 25882 16705 25883 16706 25884 16707 25885 16708 25886 16709 25887 99 25888 2 25889 16712 25890 2 25891 2 25892 2 25893 2 25894 2 25895 2 25896 2 25897 2 25898 307 25899 2 25900 2 25901 2 25902 2 25903 2 25904 313 25905 2 25906 2 25907 2 25908 2 25909 2 25910 2 25911 2 25912 2 25913 2 25914 2 25915 2 25916 2 25917 2 25918 2 25919 2 25920 16743 25921 16744 25922 16745 25923 16746 25924 16747 25925 16748 25926 16749 25927 16750 25928 16751 25929 16752 25930 16753 25931 16754 25932 16755 25933 16756 25934 16757 25935 4 25936 2 25937 16760 25938 16761 25939 16762 25940 16763 25941 16764 25942 16765 25943 16766 25944 16767 25945 16768 25946 16769 25947 16770 25948 357 25949 357 25950 40 25951 360 25952 434 25953 16776 25954 16777 25955 16778 25956 16779 25957 16780 25958 2 25959 574 25960 16783 25961 16784 25962 40 25963 16786 25964 282 25965 16788 25966 16789 25967 2 25968 574 25969 16792 25970 16793 25971 16794 25972 16795 25973 16796 25974 16797 25975 2 25976 740 25977 2 25978 16801 25979 16802 25980 16803 25981 16804 25982 16805 25983 16806 25984 16807 25985 16808 25986 16809 25987 16810 25988 855 25989 16812 25990 16813 25991 16814 25992 16815 25993 16816 25994 16817 25995 16818 25996 16819 25997 16820 25998 16821 25999 16822 26000 16823 26001 16824 26002 16825 26003 16826 26004 16827 26005 16828 26006 16829 26007 16830 26008 16831 26009 16832 26010 16833 26011 16834 26012 16835 26013 16836 26014 16837 26015 16838 26016 16839 26017 16840 26018 16841 26019 16842 26020 16843 26021 16844 26022 16845 26023 16846 26024 16847 26025 16848 26026 16849 26027 16850 26028 16851 26029 16852 26030 16853 26031 16854 26032 16855 26033 16856 26034 16857 26035 16858 26036 16859 26037 16860 26038 16861 26039 16862 26040 16863 26041 16864 26042 16865 26043 16866 26044 16867 26045 16868 26046 16869 26047 16870 26048 16871 26049 16872 26050 16873 26051 16874 26052 16875 26053 16876 26054 16877 26055 16878 26056 16879 26057 16880 26058 16881 26059 16882 26060 16883 26061 16884 26062 16885 26063 16886 26064 16887 26065 16888 26066 16889 26067 16890 26068 16891 26069 11 26070 16893 26071 16894 26072 16895 26073 2 26074 574 26075 16898 26076 943 26077 16900 26078 16901 26079 976 26080 159 26081 162 26082 16905 26083 16906 26084 1487 26085 2 26086 2 26087 2 26088 16911 26089 2 26090 16913 26091 2 26092 2 26093 2 26094 2 26095 2 26096 2 26097 2 26098 2 26099 2 26100 16923 26101 16924 26102 16925 26103 16926 26104 16927 26105 16928 26106 16929 26107 16930 26108 16931 26109 16932 26110 16933 26111 16934 26112 16935 26113 22 26114 16937 26115 4 26116 11 26117 16940 26118 89 26119 11 26120 16943 26121 99 26122 16945 26123 1551 26124 16947 26125 16948 26126 16949 26127 99 26128 855 26129 11 26130 16953 26131 943 26132 16955 26133 16956 26134 976 26135 159 26136 162 26137 16960 26138 16961 26139 16962 26140 855 26141 11 26142 16965 26143 16966 26144 16967 26145 357 26146 357 26147 360 26148 434 26149 16972 26150 16973 26151 16974 26152 16975 26153 2 26154 574 26155 16978 26156 16979 26157 16980 26158 16981 26159 16982 26160 16983 26161 2 26162 574 26163 16986 26164 16987 26165 140 26166 11 26167 142 26168 162 26169 163 26170 165 26171 167 26172 77 26173 126 26174 170 26175 16998 26176 16999 26177 140 26178 11 26179 142 26180 162 26181 2248 26182 2250 26183 2251 26184 2252 26185 2 26186 17009 26187 17010 26188 2 26189 17012 26190 855 26191 17014 26192 11 26193 17016 26194 2 26195 574 26196 17019 26197 17020 26198 17021 26199 2732 26200 17023 26201 17024 26202 357 26203 357 26204 360 26205 434 26206 17029 26207 17030 26208 17031 26209 17032 26210 2 26211 574 26212 17035 26213 17036 26214 17037 26215 17038 26216 2803 26217 17040 26218 434 26219 17042 26220 2 26221 17044 26222 17045 26223 16696 26224 2 26225 17048 26226 17049 26227 17050 26228 17051 26229 17052 26230 2894 26231 17054 26232 69 26233 17056 26234 17057 26235 17058 26236 17059 26237 17060 26238 17061 26239 17062 26240 2 26241 17064 26242 17065 26243 17066 26244 17067 26245 17068 26246 17069 26247 2936 26248 2938 26249 17072 26250 17073 26251 17074 26252 17075 26253 17076 26254 357 26255 357 26256 40 26257 360 26258 434 26259 17082 26260 17083 26261 17084 26262 17085 26263 17086 26264 40 26265 17088 26266 17089 26267 2959 26268 2962 26269 2964 26270 2966 26271 77 26272 2968 26273 2969 26274 17097 26275 282 26276 17099 26277 2 26278 282 26279 77 26280 3931 26281 2828 26282 17105 26283 1862 26284 162 26285 2050 26286 2 26287 17110 26288 2 26289 159 26290 162 26291 17114 26292 17115 26293 17116 26294 17117 26295 17118 26296 943 26297 17120 26298 17121 26299 17122 26300 976 26301 159 26302 162 26303 17126 26304 17127 26305 2 26306 282 26307 17130 26308 2 26309 282 26310 17133 26311 17134 26312 17135 26313 17136 26314 17137 26315 17138 26316 1862 26317 162 26318 17141 26319 17142 26320 17143 26321 17144 26322 282 26323 17146 26324 17147 26325 2 26326 574 26327 17150 26328 5236 26329 17152 26330 17153 26331 17154 26332 17155 26333 17156 26334 17157 26335 2894 26336 17159 26337 69 26338 17161 26339 17162 26340 17163 26341 17164 26342 17165 26343 17166 26344 17167 26345 2 26346 17169 26347 17170 26348 17171 26349 17172 26350 17173 26351 17174 26352 2936 26353 2938 26354 17177 26355 17178 26356 17179 26357 17180 26358 17181 26359 357 26360 357 26361 40 26362 360 26363 434 26364 17187 26365 17188 26366 17189 26367 17190 26368 17191 26369 40 26370 17193 26371 17194 26372 2959 26373 2962 26374 2964 26375 2966 26376 77 26377 2968 26378 2969 26379 17202 26380 282 26381 17204 26382 2 26383 282 26384 77 26385 3931 26386 2828 26387 17210 26388 1862 26389 162 26390 2050 26391 2 26392 17215 26393 2 26394 159 26395 162 26396 17219 26397 17220 26398 17221 26399 17222 26400 17223 26401 943 26402 17225 26403 17226 26404 17227 26405 976 26406 159 26407 162 26408 17231 26409 17232 26410 2 26411 282 26412 17235 26413 2 26414 282 26415 17238 26416 17239 26417 17240 26418 17241 26419 17242 26420 17243 26421 1862 26422 162 26423 17246 26424 17247 26425 17248 26426 17249 26427 282 26428 17251 26429 17252 26430 2 26431 574 26432 17255 26433 5236 26434 17257 26435 17258 26436 26436 26437 17352 26438 140 26439 11 26440 142 26441 162 26442 165 26443 167 26444 77 26445 126 26446 170 26447 26447 26448 26448 26449 26449 26450 16 26451 2962 26452 2964 26453 2966 26454 77 26455 2968 26456 2969 26457 26457 26458 26458 26459 1066 26460 7315 26461 26461 26462 17384 26463 17385 26464 7315 26465 26465 26466 2 26467 282 26468 17462 26469 17463 26470 26470 26471 26471 26472 17466 26473 7315 26474 26474 26475 22 26476 159 26477 162 26478 7315 26479 26479 26480 126 26481 170 26482 26482 26483 26483 26484 11 26485 77 26486 26486 26487 126 26488 170 26489 26489 26490 13 26491 17635 26492 162 26493 77 26494 126 26495 170 26496 26496 26497 2966 26498 77 26499 2968 26500 2969 26501 26501 26502 2 26503 282 26504 17648 26505 17649 26506 26506 26507 17651 26508 17652 26509 17653 26510 26510 26511 17653 26512 26512 26513 17788 26514 17789 26515 7315 26516 26516 26517 17789 26518 7315 26519 26519 26520 26520 26521 3042 26522 26522 26523 7315 26524 26524 26525 21294 26526 2966 26527 77 26528 2968 26529 2969 26530 26530 26531 2 26532 282 26533 17462 26534 26534 26535 162 26536 7315 26537 26537 26538 7315 26539 26539 26540 3042 26541 26541 26542 2 26543 19059 26544 20410 26545 20421 26546 17462 26547 20410 26548 162 26549 17462 26550 20472 26551 26551 26552 26552 26553 20475 26554 7315 26555 26555 26556 26556 26557 20635 26558 7315 26559 20635 26560 7315 26561 20643 26562 162 26563 17462 26564 20650 26565 26565 26566 26566 26567 20653 26568 7315 26569 26569 26570 21343 26571 2964 26572 2966 26573 77 26574 2968 26575 2969 26576 26576 26577 3042 26578 26578 26579 20856 26580 11 26581 2 26582 282 26583 7315 26584 26584 26585 16 26586 21004 26587 17462 26588 21008 26589 26589 26590 26590 26591 21011 26592 7315 26593 26593 26594 7315 26595 26595 26596 2962 26597 2964 26598 2966 26599 77 26600 2968 26601 2969 26602 26602 26603 7315 26604 26604 26605 2256 26606 26606 26607 26607 26608 26608 26609 26609 26610 26610 26611 26611 26612 26612 26613 26613 26614 26614 26615 26615 26616 26616 26617 26617 26618 26618 26619 2271 26620 2272 26621 2256 26622 26622 26623 26623 26624 26624 26625 26625 26626 26626 26627 26627 26628 26628 26629 26629 26630 26630 26631 26631 26632 26632 26633 2292 26634 26634 26635 2272 26636 26636 26637 2256 26638 26638 26639 26639 26640 26640 26641 26641 26642 26642 26643 26643 26644 26644 26645 26645 26646 26646 26647 26647 26648 2256 26649 26649 26650 26650 26651 26651 26652 26652 26653 26653 26654 26654 26655 26655 26656 26656 26657 26657 26658 26658 26659 26659 26660 26660 26661 26661 26662 26662 26663 26663 26664 3005 26665 26665 26666 159 26667 26667 26668 162 26669 26669 26670 26670 26671 26671 26672 2995 26673 26673 26674 26674 26675 26675 26676 26676 26677 26677 26678 26678 26679 26679 26680 26680 26681 26681 26682 26682 26683 26683 26684 26684 26685 2271 26686 3010 26687 2995 26688 26688 26689 26689 26690 26690 26691 26691 26692 26692 26693 26693 26694 26694 26695 26695 26696 26696 26697 2292 26698 26698 26699 3010 26700 26700 26701 2995 26702 26702 26703 26703 26704 26704 26705 26705 26706 26706 26707 26707 26708 26708 26709 26709 26710 26710 26711 26711 26712 26712 26713 3005 26714 26714 26715 26715 26716 2 26717 26717 26718 5689 26719 5690 26720 26720 26721 2 26722 3 26723 4 26724 2 26725 2 26726 2 26727 2 26728 11 26729 2 26730 2 26731 26731 26732 2 26733 26733 26734 22 26735 26735 26736 26736 26737 2 26738 11 26739 11 26740 26740 26741 26741 26742 26742 26743 22 26744 2 26745 2 26746 2 26747 11 26748 26748 26749 11 26750 26750 26751 11 26752 26752 26753 26753 26754 26754 26755 26755 26756 60 26757 26757 26758 2 26759 2 26760 26760 26761 11 26762 2 26763 2 26764 69 26765 131 26766 26766 26767 2 26768 2 26769 26769 26770 2 26771 26771 26772 26772 26773 2 26774 85 26775 86 26776 89 26777 134 26778 11 26779 26779 26780 26780 26781 99 26782 26782 26783 26783 26784 26784 26785 104 26786 104 26787 26787 26788 26788 26789 26789 26790 26790 26791 26791 26792 26792 26793 26793 26794 26794 26795 26795 26796 26796 26797 26797 26798 99 26799 2 26800 26800 26801 2 26802 2 26803 2 26804 2 26805 2 26806 2 26807 2 26808 2 26809 307 26810 2 26811 2 26812 2 26813 2 26814 2 26815 313 26816 2 26817 2 26818 2 26819 2 26820 2 26821 2 26822 2 26823 2 26824 2 26825 2 26826 2 26827 2 26828 2 26829 2 26830 2 26831 26831 26832 26832 26833 26833 26834 26834 26835 26835 26836 26836 26837 26837 26838 26838 26839 26839 26840 26840 26841 26841 26842 26842 26843 26843 26844 26844 26845 26845 26846 4 26847 2 26848 26848 26849 26849 26850 26850 26851 26851 26852 26852 26853 26853 26854 26854 26855 26855 26856 26856 26857 26857 26858 26858 26859 357 26860 357 26861 40 26862 360 26863 434 26864 26864 26865 26865 26866 26866 26867 26867 26868 26868 26869 2 26870 574 26871 26871 26872 26872 26873 40 26874 26874 26875 282 26876 26876 26877 26877 26878 2 26879 574 26880 26880 26881 26881 26882 26882 26883 26883 26884 26884 26885 26885 26886 2 26887 740 26888 2 26889 26889 26890 26890 26891 26891 26892 26892 26893 26893 26894 26894 26895 26895 26896 26896 26897 26897 26898 26898 26899 855 26900 26900 26901 26901 26902 26902 26903 26903 26904 26904 26905 26905 26906 26906 26907 26907 26908 26908 26909 26909 26910 26910 26911 26911 26912 26912 26913 26913 26914 26914 26915 26915 26916 26916 26917 26917 26918 26918 26919 26919 26920 26920 26921 26921 26922 26922 26923 26923 26924 26924 26925 26925 26926 26926 26927 26927 26928 26928 26929 26929 26930 26930 26931 26931 26932 26932 26933 26933 26934 26934 26935 26935 26936 26936 26937 26937 26938 26938 26939 26939 26940 26940 26941 26941 26942 26942 26943 26943 26944 26944 26945 26945 26946 26946 26947 26947 26948 26948 26949 26949 26950 26950 26951 26951 26952 26952 26953 26953 26954 26954 26955 26955 26956 26956 26957 26957 26958 26958 26959 26959 26960 26960 26961 26961 26962 26962 26963 26963 26964 26964 26965 26965 26966 26966 26967 26967 26968 26968 26969 26969 26970 26970 26971 26971 26972 26972 26973 26973 26974 26974 26975 26975 26976 26976 26977 26977 26978 26978 26979 26979 26980 11 26981 26981 26982 26982 26983 26983 26984 2 26985 574 26986 26986 26987 943 26988 26988 26989 26989 26990 976 26991 159 26992 162 26993 26993 26994 26994 26995 1487 26996 2 26997 2 26998 2 26999 26999 27000 2 27001 27001 27002 2 27003 2 27004 2 27005 2 27006 2 27007 2 27008 2 27009 2 27010 2 27011 27011 27012 27012 27013 27013 27014 27014 27015 27015 27016 27016 27017 27017 27018 27018 27019 27019 27020 27020 27021 27021 27022 27022 27023 27023 27024 22 27025 27025 27026 4 27027 11 27028 27028 27029 89 27030 11 27031 27031 27032 99 27033 27033 27034 1551 27035 27035 27036 27036 27037 27037 27038 99 27039 855 27040 11 27041 27041 27042 943 27043 27043 27044 27044 27045 976 27046 159 27047 162 27048 27048 27049 27049 27050 27050 27051 855 27052 11 27053 27053 27054 27054 27055 27055 27056 357 27057 357 27058 360 27059 434 27060 27060 27061 27061 27062 27062 27063 27063 27064 2 27065 574 27066 27066 27067 27067 27068 27068 27069 27069 27070 27070 27071 27071 27072 2 27073 574 27074 27074 27075 27075 27076 140 27077 11 27078 142 27079 162 27080 163 27081 165 27082 167 27083 77 27084 126 27085 170 27086 27086 27087 27087 27088 140 27089 11 27090 142 27091 162 27092 2248 27093 2250 27094 2251 27095 2252 27096 2 27097 27097 27098 27098 27099 27099 27100 855 27101 27101 27102 11 27103 27103 27104 2 27105 574 27106 27106 27107 27107 27108 27108 27109 2732 27110 27110 27111 27111 27112 357 27113 357 27114 360 27115 434 27116 27116 27117 27117 27118 27118 27119 27119 27120 2 27121 574 27122 27122 27123 27123 27124 27124 27125 27125 27126 2803 27127 27127 27128 434 27129 27129 27130 2 27131 27131 27132 27132 27133 2 27134 3 27135 4 27136 2 27137 2 27138 2 27139 2 27140 11 27141 2 27142 2 27143 27143 27144 2 27145 27145 27146 22 27147 27147 27148 27148 27149 2 27150 11 27151 11 27152 27152 27153 27153 27154 27154 27155 22 27156 2 27157 2 27158 2 27159 11 27160 27160 27161 11 27162 27162 27163 11 27164 27164 27165 27165 27166 27166 27167 27167 27168 60 27169 27169 27170 2 27171 2 27172 27172 27173 11 27174 2 27175 2 27176 69 27177 131 27178 27178 27179 2 27180 2 27181 27181 27182 2 27183 27183 27184 27184 27185 2 27186 85 27187 86 27188 89 27189 134 27190 11 27191 27191 27192 27192 27193 99 27194 27194 27195 27195 27196 27196 27197 104 27198 104 27199 27199 27200 27200 27201 27201 27202 27202 27203 27203 27204 27204 27205 27205 27206 27206 27207 27207 27208 27208 27209 27209 27210 99 27211 2 27212 27212 27213 2 27214 2 27215 2 27216 2 27217 2 27218 2 27219 2 27220 2 27221 307 27222 2 27223 2 27224 2 27225 2 27226 2 27227 313 27228 2 27229 2 27230 2 27231 2 27232 2 27233 2 27234 2 27235 2 27236 2 27237 2 27238 2 27239 2 27240 2 27241 2 27242 2 27243 27243 27244 27244 27245 27245 27246 27246 27247 27247 27248 27248 27249 27249 27250 27250 27251 27251 27252 27252 27253 27253 27254 27254 27255 27255 27256 27256 27257 27257 27258 4 27259 2 27260 27260 27261 27261 27262 27262 27263 27263 27264 27264 27265 27265 27266 27266 27267 27267 27268 27268 27269 27269 27270 27270 27271 357 27272 357 27273 40 27274 360 27275 434 27276 27276 27277 27277 27278 27278 27279 27279 27280 27280 27281 2 27282 574 27283 27283 27284 27284 27285 40 27286 27286 27287 282 27288 27288 27289 27289 27290 2 27291 574 27292 27292 27293 27293 27294 27294 27295 27295 27296 27296 27297 27297 27298 2 27299 740 27300 2 27301 27301 27302 27302 27303 27303 27304 27304 27305 27305 27306 27306 27307 27307 27308 27308 27309 27309 27310 27310 27311 855 27312 27312 27313 27313 27314 27314 27315 27315 27316 27316 27317 27317 27318 27318 27319 27319 27320 27320 27321 27321 27322 27322 27323 27323 27324 27324 27325 27325 27326 27326 27327 27327 27328 27328 27329 27329 27330 27330 27331 27331 27332 27332 27333 27333 27334 27334 27335 27335 27336 27336 27337 27337 27338 27338 27339 27339 27340 27340 27341 27341 27342 27342 27343 27343 27344 27344 27345 27345 27346 27346 27347 27347 27348 27348 27349 27349 27350 27350 27351 27351 27352 27352 27353 27353 27354 27354 27355 27355 27356 27356 27357 27357 27358 27358 27359 27359 27360 27360 27361 27361 27362 27362 27363 27363 27364 27364 27365 27365 27366 27366 27367 27367 27368 27368 27369 27369 27370 27370 27371 27371 27372 27372 27373 27373 27374 27374 27375 27375 27376 27376 27377 27377 27378 27378 27379 27379 27380 27380 27381 27381 27382 27382 27383 27383 27384 27384 27385 27385 27386 27386 27387 27387 27388 27388 27389 27389 27390 27390 27391 27391 27392 11 27393 27393 27394 27394 27395 27395 27396 2 27397 574 27398 27398 27399 943 27400 27400 27401 27401 27402 976 27403 159 27404 162 27405 27405 27406 27406 27407 1487 27408 2 27409 2 27410 2 27411 27411 27412 2 27413 27413 27414 2 27415 2 27416 2 27417 2 27418 2 27419 2 27420 2 27421 2 27422 2 27423 27423 27424 27424 27425 27425 27426 27426 27427 27427 27428 27428 27429 27429 27430 27430 27431 27431 27432 27432 27433 27433 27434 27434 27435 27435 27436 22 27437 27437 27438 4 27439 11 27440 27440 27441 89 27442 11 27443 27443 27444 99 27445 27445 27446 1551 27447 27447 27448 27448 27449 27449 27450 99 27451 855 27452 11 27453 27453 27454 943 27455 27455 27456 27456 27457 976 27458 159 27459 162 27460 27460 27461 27461 27462 27462 27463 855 27464 11 27465 27465 27466 27466 27467 27467 27468 357 27469 357 27470 360 27471 434 27472 27472 27473 27473 27474 27474 27475 27475 27476 2 27477 574 27478 27478 27479 27479 27480 27480 27481 27481 27482 27482 27483 27483 27484 2 27485 574 27486 27486 27487 27487 27488 140 27489 11 27490 142 27491 162 27492 163 27493 165 27494 167 27495 77 27496 126 27497 170 27498 27498 27499 27499 27500 140 27501 11 27502 142 27503 162 27504 2248 27505 2250 27506 2251 27507 2252 27508 2 27509 27509 27510 27510 27511 2 27512 27512 27513 855 27514 27514 27515 11 27516 27516 27517 2 27518 574 27519 27519 27520 27520 27521 27521 27522 2732 27523 27523 27524 27524 27525 357 27526 357 27527 360 27528 434 27529 27529 27530 27530 27531 27531 27532 27532 27533 2 27534 574 27535 27535 27536 27536 27537 27537 27538 27538 27539 2803 27540 27540 27541 434 27542 27542 27543 2 27544 27544 27545 27545 27546 27196 27547 2 27548 27548 27549 27549 27550 27550 27551 27551 27552 27552 27553 2894 27554 27554 27555 69 27556 27556 27557 27557 27558 27558 27559 27559 27560 27560 27561 27561 27562 27562 27563 2 27564 27564 27565 27565 27566 27566 27567 27567 27568 27568 27569 27569 27570 2936 27571 2938 27572 27572 27573 27573 27574 27574 27575 27575 27576 27576 27577 357 27578 357 27579 40 27580 360 27581 434 27582 27582 27583 27583 27584 27584 27585 27585 27586 27586 27587 40 27588 27588 27589 27589 27590 2959 27591 2962 27592 2964 27593 2966 27594 77 27595 2968 27596 2969 27597 27597 27598 282 27599 27599 27600 2 27601 282 27602 77 27603 3931 27604 2828 27605 27605 27606 1862 27607 162 27608 2050 27609 2 27610 27610 27611 2 27612 159 27613 162 27614 27614 27615 27615 27616 27616 27617 27617 27618 27618 27619 943 27620 27620 27621 27621 27622 27622 27623 976 27624 159 27625 162 27626 27626 27627 27627 27628 2 27629 282 27630 27630 27631 2 27632 282 27633 27633 27634 27634 27635 27635 27636 27636 27637 27637 27638 27638 27639 1862 27640 162 27641 27641 27642 27642 27643 27643 27644 27644 27645 282 27646 27646 27647 27647 27648 2 27649 574 27650 27650 27651 5236 27652 27652 27653 27653 27654 27654 27655 27655 27656 27656 27657 27657 27658 2894 27659 27659 27660 69 27661 27661 27662 27662 27663 27663 27664 27664 27665 27665 27666 27666 27667 27667 27668 2 27669 27669 27670 27670 27671 27671 27672 27672 27673 27673 27674 27674 27675 2936 27676 2938 27677 27677 27678 27678 27679 27679 27680 27680 27681 27681 27682 357 27683 357 27684 40 27685 360 27686 434 27687 27687 27688 27688 27689 27689 27690 27690 27691 27691 27692 40 27693 27693 27694 27694 27695 2959 27696 2962 27697 2964 27698 2966 27699 77 27700 2968 27701 2969 27702 27702 27703 282 27704 27704 27705 2 27706 282 27707 77 27708 3931 27709 2828 27710 27710 27711 1862 27712 162 27713 2050 27714 2 27715 27715 27716 2 27717 159 27718 162 27719 27719 27720 27720 27721 27721 27722 27722 27723 27723 27724 943 27725 27725 27726 27726 27727 27727 27728 976 27729 159 27730 162 27731 27731 27732 27732 27733 2 27734 282 27735 27735 27736 2 27737 282 27738 27738 27739 27739 27740 27740 27741 27741 27742 27742 27743 27743 27744 1862 27745 162 27746 27746 27747 27747 27748 27748 27749 27749 27750 282 27751 27751 27752 27752 27753 2 27754 574 27755 27755 27756 5236 27757 27757 27758 27758 27759 2995 27760 27760 27761 27761 27762 27762 27763 27763 27764 27764 27765 27765 27766 27766 27767 27767 27768 27768 27769 27769 27770 27770 27771 27771 27772 27772 27773 27773 27774 27774 27775 5836 27776 434 27777 27777 27778 943 27779 27779 27780 27780 27781 976 27782 159 27783 162 27784 27784 27785 27785 27786 5836 27787 434 27788 27788 27789 170 27790 27790 27791 855 27792 27792 27793 27793 27794 27794 27795 27795 27796 27796 27797 27797 27798 27798 27799 27799 27800 27800 27801 27801 27802 27802 27803 27803 27804 27804 27805 27805 27806 27806 27807 27807 27808 27808 27809 27809 27810 27810 27811 27811 27812 27812 27813 27813 27814 27814 27815 27815 27816 27816 27817 27817 27818 27818 27819 27819 27820 27820 27821 27821 27822 27822 27823 27823 27824 27824 27825 27825 27826 27826 27827 27827 27828 27828 27829 27829 27830 27830 27831 27831 27832 27832 27833 27833 27834 27834 27835 27835 27836 27836 27837 27837 27838 27838 27839 27839 27840 27840 27841 27841 27842 27842 27843 27843 27844 27844 27845 27845 27846 27846 27847 27847 27848 27848 27849 27849 27850 27850 27851 27851 27852 27852 27853 27853 27854 27854 27855 27855 27856 27856 27857 27857 27858 27858 27859 27859 27860 27860 27861 27861 27862 27862 27863 27863 27864 27864 27865 27865 27866 27866 27867 27867 27868 27868 27869 27869 27870 27870 27871 27871 27872 11 27873 27873 27874 27874 27875 27875 27876 2 27877 574 27878 27878 27879 27879 27880 1487 27881 2 27882 2 27883 2 27884 27884 27885 2 27886 27886 27887 2 27888 2 27889 2 27890 2 27891 2 27892 2 27893 2 27894 2 27895 2 27896 27896 27897 27897 27898 27898 27899 27899 27900 27900 27901 27901 27902 27902 27903 27903 27904 27904 27905 27905 27906 27906 27907 27907 27908 27908 27909 22 27910 27910 27911 4 27912 11 27913 27913 27914 89 27915 11 27916 27916 27917 99 27918 27918 27919 1551 27920 27920 27921 27921 27922 27922 27923 99 27924 855 27925 11 27926 27926 27927 943 27928 27928 27929 27929 27930 976 27931 159 27932 162 27933 27933 27934 27934 27935 27935 27936 855 27937 11 27938 27938 27939 27939 27940 27940 27941 357 27942 357 27943 360 27944 434 27945 27945 27946 27946 27947 27947 27948 27948 27949 2 27950 574 27951 27951 27952 27952 27953 27953 27954 27954 27955 27955 27956 27956 27957 2 27958 574 27959 27959 27960 27960 27961 26 27962 11 27963 2 27964 282 27965 26 27966 27966 27967 15 27968 15 27969 11 27970 26 27971 2 27972 11 27973 11 27974 40 27975 16 27976 27976 27977 15 27978 70 27979 71 27980 11 27981 77 27982 11 27983 11 27984 27984 27985 88 27986 103 27987 27987 27988 110 27989 27989 27990 126 27991 170 27992 27992 27993 26 27994 27994 27995 27995 27996 1066 27997 7315 27998 27998 27999 17384 28000 17385 28001 7315 28002 28002 28003 2 28004 282 28005 17462 28006 17463 28007 28007 28008 28008 28009 17466 28010 7315 28011 28011 28012 22 28013 159 28014 162 28015 7315 28016 28016 28017 126 28018 170 28019 28019 28020 28020 28021 11 28022 77 28023 28023 28024 126 28025 170 28026 28026 28027 13 28028 17635 28029 162 28030 77 28031 126 28032 170 28033 28033 28034 2 28035 282 28036 17648 28037 17649 28038 28038 28039 17651 28040 17652 28041 17653 28042 28042 28043 17653 28044 28044 28045 17788 28046 17789 28047 7315 28048 28048 28049 17789 28050 7315 28051 28051 28052 28052 28053 3042 28054 28054 28055 7315 28056 28056 28057 7315 28058 28058 28059 3042 28060 28060 28061 2 28062 19059 28063 20410 28064 20421 28065 17462 28066 20410 28067 162 28068 17462 28069 20472 28070 28070 28071 28071 28072 20475 28073 7315 28074 28074 28075 28075 28076 20635 28077 7315 28078 20635 28079 7315 28080 20643 28081 162 28082 17462 28083 20650 28084 28084 28085 28085 28086 20653 28087 7315 28088 28088 28089 3042 28090 28090 28091 20856 28092 11 28093 2 28094 282 28095 7315 28096 28096 28097 16 28098 21004 28099 17462 28100 21008 28101 28101 28102 28102 28103 21011 28104 7315 28105 28105 28106 7315 28107 28107 28108 7315 28109 28109 28110 28110 28111 17352 28112 140 28113 11 28114 142 28115 162 28116 165 28117 167 28118 77 28119 126 28120 170 28121 28121 28122 28122 28123 140 28124 11 28125 142 28126 162 28127 2248 28128 2250 28129 2251 28130 2252 28131 2 28132 2 28133 3 28134 4 28135 2 28136 2 28137 2 28138 2 28139 11 28140 2 28141 2 28142 17 28143 2 28144 20 28145 22 28146 24 28147 25 28148 2 28149 11 28150 11 28151 32 28152 33 28153 39 28154 22 28155 2 28156 2 28157 2 28158 11 28159 51 28160 11 28161 53 28162 11 28163 55 28164 57 28165 58 28166 59 28167 60 28168 61 28169 2 28170 2 28171 64 28172 11 28173 2 28174 2 28175 69 28176 131 28177 132 28178 2 28179 2 28180 76 28181 2 28182 81 28183 83 28184 2 28185 85 28186 86 28187 89 28188 134 28189 11 28190 94 28191 95 28192 99 28193 100 28194 101 28195 28195 28196 104 28197 104 28198 107 28199 111 28200 113 28201 114 28202 115 28203 116 28204 117 28205 118 28206 119 28207 121 28208 124 28209 99 28210 2 28211 135 28212 2 28213 2 28214 2 28215 2 28216 2 28217 2 28218 2 28219 2 28220 307 28221 2 28222 2 28223 2 28224 2 28225 2 28226 313 28227 2 28228 2 28229 2 28230 2 28231 2 28232 2 28233 2 28234 2 28235 2 28236 2 28237 2 28238 2 28239 2 28240 2 28241 2 28242 5757 28243 5758 28244 5759 28245 5760 28246 5761 28247 5762 28248 5763 28249 5764 28250 5765 28251 5766 28252 5767 28253 5768 28254 5769 28255 5770 28256 5771 28257 4 28258 2 28259 5774 28260 5775 28261 5776 28262 5777 28263 5778 28264 5779 28265 5780 28266 5781 28267 5782 28268 5783 28269 5784 28270 357 28271 357 28272 40 28273 360 28274 434 28275 5790 28276 5791 28277 5792 28278 5793 28279 5794 28280 2 28281 574 28282 5797 28283 5798 28284 40 28285 5800 28286 282 28287 5802 28288 5803 28289 2 28290 574 28291 5806 28292 5807 28293 5808 28294 5809 28295 5810 28296 5811 28297 2 28298 740 28299 2 28300 5815 28301 5816 28302 5817 28303 5818 28304 5819 28305 5820 28306 5821 28307 5822 28308 5823 28309 5824 28310 855 28311 27792 28312 27793 28313 27794 28314 27795 28315 27796 28316 27797 28317 27798 28318 27799 28319 27800 28320 27801 28321 27802 28322 27803 28323 27804 28324 27805 28325 27806 28326 27807 28327 27808 28328 27809 28329 27810 28330 27811 28331 27812 28332 27813 28333 27814 28334 27815 28335 27816 28336 27817 28337 27818 28338 27819 28339 27820 28340 27821 28341 27822 28342 27823 28343 27824 28344 27825 28345 27826 28346 27827 28347 27828 28348 27829 28349 27830 28350 27831 28351 27832 28352 27833 28353 27834 28354 27835 28355 27836 28356 27837 28357 27838 28358 27839 28359 27840 28360 27841 28361 27842 28362 27843 28363 27844 28364 27845 28365 27846 28366 27847 28367 27848 28368 27849 28369 27850 28370 27851 28371 27852 28372 27853 28373 27854 28374 27855 28375 27856 28376 27857 28377 27858 28378 27859 28379 27860 28380 27861 28381 27862 28382 27863 28383 27864 28384 27865 28385 27866 28386 27867 28387 27868 28388 27869 28389 27870 28390 27871 28391 11 28392 27873 28393 27874 28394 27875 28395 2 28396 574 28397 27878 28398 943 28399 27879 28400 27780 28401 976 28402 159 28403 162 28404 27784 28405 27785 28406 1487 28407 2 28408 2 28409 2 28410 27884 28411 2 28412 27886 28413 2 28414 2 28415 2 28416 2 28417 2 28418 2 28419 2 28420 2 28421 2 28422 27896 28423 27897 28424 27898 28425 27899 28426 27900 28427 27901 28428 27902 28429 27903 28430 27904 28431 27905 28432 27906 28433 27907 28434 27908 28435 22 28436 27910 28437 4 28438 11 28439 27913 28440 89 28441 11 28442 27916 28443 99 28444 27918 28445 1551 28446 27920 28447 27921 28448 27922 28449 99 28450 855 28451 11 28452 27926 28453 943 28454 27928 28455 27929 28456 976 28457 159 28458 162 28459 27933 28460 27934 28461 27935 28462 855 28463 11 28464 27938 28465 27939 28466 27940 28467 357 28468 357 28469 360 28470 434 28471 27945 28472 27946 28473 27947 28474 27948 28475 2 28476 574 28477 27951 28478 27952 28479 27953 28480 27954 28481 27955 28482 27956 28483 2 28484 574 28485 27959 28486 27960 28487 140 28488 11 28489 142 28490 162 28491 163 28492 165 28493 167 28494 77 28495 126 28496 170 28497 212 28498 213 28499 140 28500 11 28501 142 28502 162 28503 2248 28504 2250 28505 2251 28506 2252 28507 2 28508 28195 28509 28509 28510 28509 28511 855 28512 28512 28513 11 28514 28514 28515 2 28516 574 28517 28517 28518 28518 28519 28519 28520 2732 28521 17379 28522 17380 28523 357 28524 357 28525 360 28526 434 28527 28527 28528 28528 28529 28529 28530 28530 28531 2 28532 574 28533 28533 28534 28534 28535 28535 28536 28536 28537 2803 28538 28538 28539 434 28540 28540 28541 2 28542 28542 28543 28543 28544 2 28545 3 28546 4 28547 2 28548 2 28549 2 28550 2 28551 11 28552 2 28553 2 28554 28554 28555 2 28556 28556 28557 22 28558 28558 28559 28559 28560 2 28561 11 28562 11 28563 28563 28564 28564 28565 28565 28566 22 28567 2 28568 2 28569 2 28570 11 28571 28571 28572 11 28573 28573 28574 11 28575 28575 28576 28576 28577 28577 28578 28578 28579 60 28580 28580 28581 2 28582 2 28583 28583 28584 11 28585 2 28586 2 28587 69 28588 131 28589 28589 28590 2 28591 2 28592 28592 28593 2 28594 28594 28595 28595 28596 2 28597 85 28598 86 28599 89 28600 134 28601 11 28602 28602 28603 28603 28604 99 28605 28605 28606 28606 28607 28607 28608 104 28609 104 28610 28610 28611 28611 28612 28612 28613 28613 28614 28614 28615 28615 28616 28616 28617 28617 28618 28618 28619 28619 28620 28620 28621 99 28622 2 28623 28623 28624 2 28625 2 28626 2 28627 2 28628 2 28629 2 28630 2 28631 2 28632 307 28633 2 28634 2 28635 2 28636 2 28637 2 28638 313 28639 2 28640 2 28641 2 28642 2 28643 2 28644 2 28645 2 28646 2 28647 2 28648 2 28649 2 28650 2 28651 2 28652 2 28653 2 28654 28654 28655 28655 28656 28656 28657 28657 28658 28658 28659 28659 28660 28660 28661 28661 28662 28662 28663 28663 28664 28664 28665 28665 28666 28666 28667 28667 28668 28668 28669 4 28670 2 28671 28671 28672 28672 28673 28673 28674 28674 28675 28675 28676 28676 28677 28677 28678 28678 28679 28679 28680 28680 28681 28681 28682 357 28683 357 28684 40 28685 360 28686 434 28687 28687 28688 28688 28689 28689 28690 28690 28691 28691 28692 2 28693 574 28694 28694 28695 28695 28696 40 28697 28697 28698 282 28699 28699 28700 28700 28701 2 28702 574 28703 28703 28704 28704 28705 28705 28706 28706 28707 28707 28708 28708 28709 2 28710 740 28711 2 28712 28712 28713 28713 28714 28714 28715 28715 28716 28716 28717 28717 28718 28718 28719 28719 28720 28720 28721 28721 28722 855 28723 28723 28724 28724 28725 28725 28726 28726 28727 28727 28728 28728 28729 28729 28730 28730 28731 28731 28732 28732 28733 28733 28734 28734 28735 28735 28736 28736 28737 28737 28738 28738 28739 28739 28740 28740 28741 28741 28742 28742 28743 28743 28744 28744 28745 28745 28746 28746 28747 28747 28748 28748 28749 28749 28750 28750 28751 28751 28752 28752 28753 28753 28754 28754 28755 28755 28756 28756 28757 28757 28758 28758 28759 28759 28760 28760 28761 28761 28762 28762 28763 28763 28764 28764 28765 28765 28766 28766 28767 28767 28768 28768 28769 28769 28770 28770 28771 28771 28772 28772 28773 28773 28774 28774 28775 28775 28776 28776 28777 28777 28778 28778 28779 28779 28780 28780 28781 28781 28782 28782 28783 28783 28784 28784 28785 28785 28786 28786 28787 28787 28788 28788 28789 28789 28790 28790 28791 28791 28792 28792 28793 28793 28794 28794 28795 28795 28796 28796 28797 28797 28798 28798 28799 28799 28800 28800 28801 28801 28802 28802 28803 11 28804 28804 28805 28805 28806 28806 28807 2 28808 574 28809 28809 28810 943 28811 28811 28812 28812 28813 976 28814 159 28815 162 28816 28816 28817 28817 28818 1487 28819 2 28820 2 28821 2 28822 28822 28823 2 28824 28824 28825 2 28826 2 28827 2 28828 2 28829 2 28830 2 28831 2 28832 2 28833 2 28834 28834 28835 28835 28836 28836 28837 28837 28838 28838 28839 28839 28840 28840 28841 28841 28842 28842 28843 28843 28844 28844 28845 28845 28846 28846 28847 22 28848 28848 28849 4 28850 11 28851 28851 28852 89 28853 11 28854 28854 28855 99 28856 28856 28857 1551 28858 28858 28859 28859 28860 28860 28861 99 28862 855 28863 11 28864 28864 28865 943 28866 28866 28867 28867 28868 976 28869 159 28870 162 28871 28871 28872 28872 28873 28873 28874 855 28875 11 28876 28876 28877 28877 28878 28878 28879 357 28880 357 28881 360 28882 434 28883 28883 28884 28884 28885 28885 28886 28886 28887 2 28888 574 28889 28889 28890 28890 28891 28891 28892 28892 28893 28893 28894 28894 28895 2 28896 574 28897 28897 28898 28898 28899 140 28900 11 28901 142 28902 162 28903 163 28904 165 28905 167 28906 77 28907 126 28908 170 28909 28909 28910 28910 28911 140 28912 11 28913 142 28914 162 28915 2248 28916 2250 28917 2251 28918 2252 28919 2 28920 28920 28921 28921 28922 2 28923 28923 28924 855 28925 28925 28926 11 28927 28927 28928 2 28929 574 28930 28930 28931 28931 28932 28932 28933 2732 28934 28934 28935 28935 28936 357 28937 357 28938 360 28939 434 28940 28940 28941 28941 28942 28942 28943 28943 28944 2 28945 574 28946 28946 28947 28947 28948 28948 28949 28949 28950 2803 28951 28951 28952 434 28953 28953 28954 2 28955 28955 28956 28956 28957 28607 28958 2 28959 28959 28960 28960 28961 28961 28962 28962 28963 28963 28964 2894 28965 28965 28966 69 28967 28967 28968 28968 28969 28969 28970 28970 28971 28971 28972 28972 28973 28973 28974 2 28975 28975 28976 28976 28977 28977 28978 28978 28979 28979 28980 28980 28981 2936 28982 2938 28983 28983 28984 28984 28985 28985 28986 28986 28987 28987 28988 357 28989 357 28990 40 28991 360 28992 434 28993 28993 28994 28994 28995 28995 28996 28996 28997 28997 28998 40 28999 28999 29000 29000 29001 2959 29002 2962 29003 2964 29004 2966 29005 77 29006 2968 29007 2969 29008 29008 29009 282 29010 29010 29011 2 29012 282 29013 77 29014 3931 29015 2828 29016 29016 29017 1862 29018 162 29019 2050 29020 2 29021 29021 29022 2 29023 159 29024 162 29025 29025 29026 29026 29027 29027 29028 29028 29029 29029 29030 943 29031 29031 29032 29032 29033 29033 29034 976 29035 159 29036 162 29037 29037 29038 29038 29039 2 29040 282 29041 29041 29042 2 29043 282 29044 29044 29045 29045 29046 29046 29047 29047 29048 29048 29049 29049 29050 1862 29051 162 29052 29052 29053 29053 29054 29054 29055 29055 29056 282 29057 29057 29058 29058 29059 2 29060 574 29061 29061 29062 5236 29063 29063 29064 29064 29065 29065 29066 29066 29067 29067 29068 17934 29069 2894 29070 29070 29071 69 29072 29072 29073 29073 29074 29074 29075 29075 29076 27779 29077 29077 29078 29078 29079 2 29080 29080 29081 29081 29082 29082 29083 29083 29084 29084 29085 29085 29086 2936 29087 2938 29088 29088 29089 29089 29090 29090 29091 29091 29092 29092 29093 357 29094 357 29095 40 29096 360 29097 434 29098 29098 29099 29099 29100 29100 29101 29101 29102 29102 29103 40 29104 29104 29105 29105 29106 2959 29107 2962 29108 2964 29109 2966 29110 77 29111 2968 29112 2969 29113 29113 29114 282 29115 29115 29116 2 29117 282 29118 77 29119 3931 29120 2828 29121 29121 29122 1862 29123 162 29124 2050 29125 2 29126 29126 29127 2 29128 159 29129 162 29130 29130 29131 29131 29132 29132 29133 29133 29134 29134 29135 943 29136 29136 29137 29137 29138 29138 29139 976 29140 159 29141 162 29142 29142 29143 29143 29144 2 29145 282 29146 29146 29147 2 29148 282 29149 29149 29150 29150 29151 29151 29152 29152 29153 29153 29154 29154 29155 1862 29156 162 29157 29157 29158 29158 29159 29159 29160 29160 29161 282 29162 29162 29163 29163 29164 2 29165 574 29166 29166 29167 5236 29168 29168 29169 29169 29170 29170 29171 855 29172 28512 29173 11 29174 28514 29175 2 29176 574 29177 28517 29178 28518 29179 16 29180 2962 29181 2964 29182 2966 29183 77 29184 2968 29185 2969 29186 29186 29187 28519 29188 16 29189 16 29190 16 29191 16 29192 29192 29193 29193 29194 98 29195 29195 29196 106 29197 29197 29198 29198 29199 29199 29200 17352 29201 140 29202 11 29203 142 29204 162 29205 165 29206 167 29207 77 29208 126 29209 170 29210 29210 29211 29211 29212 16 29213 2962 29214 2964 29215 2966 29216 77 29217 2968 29218 2969 29219 29219 29220 2966 29221 77 29222 2968 29223 2969 29224 29224 29225 21294 29226 2966 29227 77 29228 2968 29229 2969 29230 29230 29231 21343 29232 2964 29233 2966 29234 77 29235 2968 29236 2969 29237 29237 29238 2962 29239 2964 29240 2966 29241 77 29242 2968 29243 2969 29244 29244 29245 357 29246 357 29247 360 29248 434 29249 28527 29250 28528 29251 28529 29252 28530 29253 2 29254 574 29255 28533 29256 28534 29257 28535 29258 28536 29259 1066 29260 7315 29261 29261 29262 17384 29263 17385 29264 7315 29265 29265 29266 2803 29267 28538 29268 434 29269 28540 29270 2 29271 28542 29272 28543 29273 2 29274 3 29275 4 29276 2 29277 2 29278 2 29279 2 29280 11 29281 2 29282 2 29283 28554 29284 2 29285 28556 29286 22 29287 28558 29288 28559 29289 2 29290 11 29291 11 29292 28563 29293 28564 29294 28565 29295 22 29296 2 29297 2 29298 2 29299 11 29300 28571 29301 11 29302 28573 29303 11 29304 28575 29305 28576 29306 28577 29307 28578 29308 60 29309 28580 29310 2 29311 2 29312 28583 29313 11 29314 2 29315 2 29316 69 29317 131 29318 28589 29319 2 29320 2 29321 28592 29322 2 29323 28594 29324 28595 29325 2 29326 85 29327 86 29328 89 29329 134 29330 11 29331 28602 29332 28603 29333 99 29334 28605 29335 28606 29336 28607 29337 104 29338 104 29339 28610 29340 28611 29341 28612 29342 28613 29343 28614 29344 28615 29345 28616 29346 28617 29347 28618 29348 28619 29349 28620 29350 99 29351 2 29352 28623 29353 2 29354 2 29355 2 29356 2 29357 2 29358 2 29359 2 29360 2 29361 307 29362 2 29363 2 29364 2 29365 2 29366 2 29367 313 29368 2 29369 2 29370 2 29371 2 29372 2 29373 2 29374 2 29375 2 29376 2 29377 2 29378 2 29379 2 29380 2 29381 2 29382 2 29383 28654 29384 28655 29385 28656 29386 28657 29387 28658 29388 28659 29389 28660 29390 28661 29391 28662 29392 28663 29393 28664 29394 28665 29395 28666 29396 28667 29397 28668 29398 4 29399 2 29400 28671 29401 28672 29402 28673 29403 28674 29404 28675 29405 28676 29406 28677 29407 28678 29408 28679 29409 28680 29410 28681 29411 357 29412 357 29413 40 29414 360 29415 434 29416 28687 29417 28688 29418 28689 29419 28690 29420 28691 29421 2 29422 574 29423 28694 29424 28695 29425 40 29426 28697 29427 282 29428 28699 29429 28700 29430 2 29431 574 29432 28703 29433 28704 29434 28705 29435 28706 29436 28707 29437 28708 29438 2 29439 740 29440 2 29441 28712 29442 28713 29443 28714 29444 28715 29445 28716 29446 28717 29447 28718 29448 28719 29449 28720 29450 28721 29451 855 29452 28723 29453 28724 29454 28725 29455 28726 29456 28727 29457 28728 29458 28729 29459 28730 29460 28731 29461 28732 29462 28733 29463 28734 29464 28735 29465 28736 29466 28737 29467 28738 29468 28739 29469 28740 29470 28741 29471 28742 29472 28743 29473 28744 29474 28745 29475 28746 29476 28747 29477 28748 29478 28749 29479 28750 29480 28751 29481 28752 29482 28753 29483 28754 29484 28755 29485 28756 29486 28757 29487 28758 29488 28759 29489 28760 29490 28761 29491 28762 29492 28763 29493 28764 29494 28765 29495 28766 29496 28767 29497 28768 29498 28769 29499 28770 29500 28771 29501 28772 29502 28773 29503 28774 29504 28775 29505 28776 29506 28777 29507 28778 29508 28779 29509 28780 29510 28781 29511 28782 29512 28783 29513 28784 29514 28785 29515 28786 29516 28787 29517 28788 29518 28789 29519 28790 29520 28791 29521 28792 29522 28793 29523 28794 29524 28795 29525 28796 29526 28797 29527 28798 29528 28799 29529 28800 29530 28801 29531 28802 29532 11 29533 28804 29534 28805 29535 28806 29536 2 29537 574 29538 28809 29539 943 29540 28811 29541 28812 29542 976 29543 159 29544 162 29545 28816 29546 28817 29547 1487 29548 2 29549 2 29550 2 29551 28822 29552 2 29553 28824 29554 2 29555 2 29556 2 29557 2 29558 2 29559 2 29560 2 29561 2 29562 2 29563 28834 29564 28835 29565 28836 29566 28837 29567 28838 29568 28839 29569 28840 29570 28841 29571 28842 29572 28843 29573 28844 29574 28845 29575 28846 29576 22 29577 28848 29578 4 29579 11 29580 28851 29581 89 29582 11 29583 28854 29584 99 29585 28856 29586 1551 29587 28858 29588 28859 29589 28860 29590 99 29591 855 29592 11 29593 28864 29594 943 29595 28866 29596 28867 29597 976 29598 159 29599 162 29600 28871 29601 28872 29602 28873 29603 855 29604 11 29605 28876 29606 28877 29607 28878 29608 357 29609 357 29610 360 29611 434 29612 28883 29613 28884 29614 28885 29615 28886 29616 2 29617 574 29618 28889 29619 28890 29620 28891 29621 28892 29622 28893 29623 28894 29624 2 29625 574 29626 28897 29627 28898 29628 140 29629 11 29630 142 29631 162 29632 163 29633 165 29634 167 29635 77 29636 126 29637 170 29638 28909 29639 28910 29640 140 29641 11 29642 142 29643 162 29644 2248 29645 2250 29646 2251 29647 2252 29648 2 29649 28920 29650 28921 29651 2 29652 28923 29653 855 29654 28925 29655 11 29656 28927 29657 2 29658 574 29659 28930 29660 28931 29661 28932 29662 2732 29663 28934 29664 28935 29665 357 29666 357 29667 360 29668 434 29669 28940 29670 28941 29671 28942 29672 28943 29673 2 29674 574 29675 28946 29676 28947 29677 28948 29678 28949 29679 2803 29680 28951 29681 434 29682 28953 29683 2 29684 28955 29685 28956 29686 28607 29687 2 29688 28959 29689 28960 29690 28961 29691 28962 29692 28963 29693 2894 29694 28965 29695 69 29696 28967 29697 28968 29698 28969 29699 28970 29700 28971 29701 28972 29702 28973 29703 2 29704 28975 29705 28976 29706 28977 29707 28978 29708 28979 29709 28980 29710 2936 29711 2938 29712 28983 29713 28984 29714 28985 29715 28986 29716 28987 29717 357 29718 357 29719 40 29720 360 29721 434 29722 28993 29723 28994 29724 28995 29725 28996 29726 28997 29727 40 29728 28999 29729 29000 29730 2959 29731 2962 29732 2964 29733 2966 29734 77 29735 2968 29736 2969 29737 29008 29738 282 29739 29010 29740 2 29741 282 29742 77 29743 3931 29744 2828 29745 29016 29746 1862 29747 162 29748 2050 29749 2 29750 29021 29751 2 29752 159 29753 162 29754 29025 29755 29026 29756 29027 29757 29028 29758 29029 29759 943 29760 29031 29761 29032 29762 29033 29763 976 29764 159 29765 162 29766 29037 29767 29038 29768 2 29769 282 29770 29041 29771 2 29772 282 29773 29044 29774 29045 29775 29046 29776 29047 29777 29048 29778 29049 29779 1862 29780 162 29781 29052 29782 29053 29783 29054 29784 29055 29785 282 29786 29057 29787 29058 29788 2 29789 574 29790 29061 29791 5236 29792 29063 29793 29064 29794 2 29795 282 29796 17462 29797 17463 29798 29798 29799 29799 29800 17466 29801 7315 29802 29802 29803 29803 29804 29804 29805 29805 29806 29806 29807 29807 29808 29808 29809 7315 29810 29810 29811 29811 29812 3042 29813 29813 29814 7315 29815 29808 29816 3042 29817 29813 29818 29818 29819 29819 29820 29820 29821 29821 29822 29822 29823 29823 29824 29824 29825 16 29826 29826 29827 77 29828 29828 29829 29829 29830 29830 29831 29831 29832 29832 29833 4009 29834 29834 29835 2828 29836 29836 29837 29837 29838 16 29839 29839 29840 29840 29841 29841 29842 29842 29843 77 29844 29844 29845 29845 29846 29846 29847 29847 29848 4009 29849 29849 29850 29828 29851 29851 29852 29852 29853 2828 29854 29854 29855 29855 29856 29065 29857 22 29858 159 29859 162 29860 7315 29861 29861 29862 126 29863 170 29864 29864 29865 29865 29866 11 29867 77 29868 29868 29869 126 29870 170 29871 29871 29872 13 29873 17635 29874 162 29875 77 29876 126 29877 170 29878 29878 29879 2966 29880 77 29881 2968 29882 2969 29883 29883 29884 29066 29885 2 29886 282 29887 17648 29888 17649 29889 29889 29890 17651 29891 17652 29892 17653 29893 29893 29894 17653 29895 29895 29896 17788 29897 17789 29898 7315 29899 29899 29900 17789 29901 7315 29902 29902 29903 29067 29904 29904 29905 3042 29906 29906 29907 2894 29908 29070 29909 69 29910 29072 29911 29073 29912 7315 29913 29913 29914 29074 29915 29075 29916 29077 29917 29078 29918 2 29919 29080 29920 29081 29921 29082 29922 29083 29923 29084 29924 29085 29925 2936 29926 2938 29927 29088 29928 29089 29929 29090 29930 29091 29931 29092 29932 357 29933 357 29934 40 29935 360 29936 434 29937 29098 29938 29099 29939 29100 29940 29101 29941 29102 29942 40 29943 29104 29944 29105 29945 2959 29946 2962 29947 2964 29948 2966 29949 77 29950 2968 29951 2969 29952 29113 29953 282 29954 29115 29955 29955 29956 29956 29957 29957 29958 29958 29959 29959 29960 29960 29961 29961 29962 29962 29963 29963 29964 29964 29965 29965 29966 29966 29967 29967 29968 29968 29969 29969 29970 29970 29971 29971 29972 29972 29973 29973 29974 29974 29975 29975 29976 29976 29977 29977 29978 29978 29979 29979 29980 29980 29981 29981 29982 29982 29983 29983 29984 29984 29985 29985 29986 29986 29987 29987 29988 29988 29989 29989 29990 29990 29991 29991 29992 29992 29993 29993 29994 29994 29995 29995 29996 29996 29997 29997 29998 29998 29999 29999 30000 30000 30001 30001 30002 30002 30003 30003 30004 30004 30005 30005 30006 30006 30007 30007 30008 30008 30009 30009 30010 30010 30011 30011 30012 30012 30013 30013 30014 30014 30015 30015 30016 30016 30017 30017 30018 30018 30019 30019 30020 30020 30021 30021 30022 30022 30023 30023 30024 30024 30025 30025 30026 30026 30027 30027 30028 30028 30029 30029 30030 30030 30031 30031 30032 30032 30033 30033 30034 30034 30035 30035 30036 6999 30037 30037 30038 30038 30039 30039 30040 30040 30041 30041 30042 15 30043 16 30044 16 30045 15 30046 11 30047 26 30048 2 30049 16 30050 11 30051 11 30052 40 30053 16 30054 11 30055 11 30056 30056 30057 16 30058 56 30059 15 30060 70 30061 71 30062 11 30063 16 30064 77 30065 11 30066 80 30067 11 30068 30068 30069 88 30070 30070 30071 92 30072 30072 30073 98 30074 30074 30075 103 30076 106 30077 30077 30078 30078 30079 30079 30080 30080 30081 30081 30082 30082 30083 30083 30084 126 30085 30085 30086 170 30087 30087 30088 26 30089 11 30090 2 30091 282 30092 26 30093 30093 30094 15 30095 15 30096 11 30097 26 30098 2 30099 11 30100 11 30101 40 30102 16 30103 30103 30104 15 30105 70 30106 71 30107 11 30108 77 30109 11 30110 11 30111 30111 30112 88 30113 103 30114 30114 30115 30079 30116 30116 30117 126 30118 170 30119 30119 30120 26 30121 30121 30122 30122 30123 1066 30124 7315 30125 30125 30126 17384 30127 17385 30128 7315 30129 30129 30130 2 30131 282 30132 17462 30133 17463 30134 30134 30135 30135 30136 17466 30137 7315 30138 30138 30139 22 30140 159 30141 162 30142 7315 30143 30143 30144 126 30145 170 30146 30146 30147 30147 30148 11 30149 77 30150 30150 30151 126 30152 170 30153 30153 30154 13 30155 17635 30156 162 30157 77 30158 126 30159 170 30160 30160 30161 2 30162 282 30163 17648 30164 17649 30165 30165 30166 17651 30167 17652 30168 17653 30169 30169 30170 17653 30171 30171 30172 17788 30173 17789 30174 7315 30175 30175 30176 17789 30177 7315 30178 30178 30179 30179 30180 3042 30181 30181 30182 7315 30183 30183 30184 7315 30185 30185 30186 3042 30187 30187 30188 2 30189 19059 30190 20410 30191 20421 30192 17462 30193 20410 30194 162 30195 17462 30196 20472 30197 30197 30198 30198 30199 20475 30200 7315 30201 30201 30202 30202 30203 20635 30204 7315 30205 20635 30206 7315 30207 20643 30208 162 30209 17462 30210 20650 30211 30211 30212 30212 30213 20653 30214 7315 30215 30215 30216 3042 30217 30217 30218 20856 30219 11 30220 2 30221 282 30222 7315 30223 30223 30224 16 30225 21004 30226 17462 30227 21008 30228 30228 30229 30229 30230 21011 30231 7315 30232 30232 30233 7315 30234 30234 30235 7315 30236 30236 30237 30237 30238 17352 30239 140 30240 11 30241 142 30242 162 30243 165 30244 167 30245 77 30246 126 30247 170 30248 30248 30249 30249 30250 30250 30251 16 30252 2962 30253 2964 30254 2966 30255 77 30256 2968 30257 2969 30258 30258 30259 30259 30260 1066 30261 7315 30262 30262 30263 17384 30264 17385 30265 7315 30266 30266 30267 2 30268 282 30269 17462 30270 17463 30271 30271 30272 30272 30273 17466 30274 7315 30275 30275 30276 7315 30277 3042 30278 29813 30279 7315 30280 3042 30281 30281 30282 30282 30283 30283 30284 30284 30285 22 30286 159 30287 162 30288 7315 30289 30289 30290 126 30291 170 30292 30292 30293 30293 30294 11 30295 77 30296 30296 30297 126 30298 170 30299 30299 30300 13 30301 17635 30302 162 30303 77 30304 126 30305 170 30306 30306 30307 2966 30308 77 30309 2968 30310 2969 30311 30311 30312 2 30313 282 30314 17648 30315 17649 30316 30316 30317 17651 30318 17652 30319 17653 30320 30320 30321 17653 30322 30322 30323 17788 30324 17789 30325 7315 30326 30326 30327 17789 30328 7315 30329 30329 30330 30330 30331 3042 30332 30332 30333 7315 30334 30334 30335 6999 30336 30039 30337 30040 30338 30338 30339 30339 30340 30039 30341 30040 30342 30342 30343 21294 30344 2966 30345 77 30346 2968 30347 2969 30348 30348 30349 2 30350 282 30351 17462 30352 30352 30353 162 30354 7315 30355 30355 30356 7315 30357 30357 30358 3042 30359 30359 30360 2 30361 19059 30362 20410 30363 20421 30364 17462 30365 20410 30366 162 30367 17462 30368 20472 30369 30369 30370 30370 30371 20475 30372 7315 30373 30373 30374 30374 30375 20635 30376 7315 30377 20635 30378 7315 30379 20643 30380 162 30381 17462 30382 20650 30383 30383 30384 30384 30385 20653 30386 7315 30387 30387 30388 21343 30389 2964 30390 2966 30391 77 30392 2968 30393 2969 30394 30394 30395 3042 30396 30396 30397 20856 30398 11 30399 2 30400 282 30401 7315 30402 30402 30403 16 30404 21004 30405 17462 30406 21008 30407 30407 30408 30408 30409 21011 30410 7315 30411 30411 30412 7315 30413 30413 30414 2050 30415 30415 30416 3042 30417 16 30418 30418 30419 30419 30420 2 30421 3112 30422 30422 30423 2962 30424 2964 30425 2966 30426 77 30427 2968 30428 2969 30429 30429 30430 7315 30431 30431 30432 30432 30433 30039 30434 30040 30435 30435 30436 2 30437 282 30438 77 30439 3931 30440 2828 30441 29121 30442 1862 30443 162 30444 2050 30445 2 30446 29126 30447 2 30448 159 30449 162 30450 29130 30451 29131 30452 21294 30453 2966 30454 77 30455 2968 30456 2969 30457 30457 30458 29132 30459 29133 30460 29134 30461 2 30462 282 30463 17462 30464 30464 30465 15 30466 15 30467 11 30468 26 30469 2 30470 11 30471 11 30472 40 30473 16 30474 11 30475 30475 30476 15 30477 70 30478 71 30479 11 30480 77 30481 11 30482 11 30483 30483 30484 88 30485 103 30486 30486 30487 30487 30488 30488 30489 126 30490 170 30491 30491 30492 26 30493 11 30494 2 30495 282 30496 26 30497 30497 30498 30498 30499 30499 30500 1066 30501 7315 30502 30502 30503 17384 30504 17385 30505 7315 30506 30506 30507 2 30508 282 30509 17462 30510 17463 30511 30511 30512 30512 30513 17466 30514 7315 30515 30515 30516 22 30517 159 30518 162 30519 7315 30520 30520 30521 126 30522 170 30523 30523 30524 30524 30525 11 30526 77 30527 30527 30528 126 30529 170 30530 30530 30531 13 30532 17635 30533 162 30534 77 30535 126 30536 170 30537 30537 30538 2 30539 282 30540 17648 30541 17649 30542 30542 30543 17651 30544 17652 30545 17653 30546 30546 30547 17653 30548 30548 30549 17788 30550 17789 30551 7315 30552 30552 30553 17789 30554 7315 30555 30555 30556 30556 30557 3042 30558 30558 30559 7315 30560 30560 30561 2 30562 282 30563 17462 30564 30564 30565 7315 30566 30566 30567 3042 30568 30568 30569 2 30570 19059 30571 20410 30572 20421 30573 17462 30574 20410 30575 162 30576 17462 30577 20472 30578 30578 30579 30579 30580 20475 30581 7315 30582 30582 30583 30583 30584 20635 30585 7315 30586 20635 30587 7315 30588 20643 30589 162 30590 17462 30591 20650 30592 30592 30593 30593 30594 20653 30595 7315 30596 30596 30597 3042 30598 30598 30599 20856 30600 11 30601 2 30602 282 30603 7315 30604 30604 30605 16 30606 21004 30607 17462 30608 21008 30609 30609 30610 30610 30611 21011 30612 7315 30613 30613 30614 7315 30615 30615 30616 7315 30617 30617 30618 162 30619 7315 30620 30620 30621 7315 30622 30622 30623 3042 30624 30624 30625 2 30626 19059 30627 20410 30628 20421 30629 17462 30630 20410 30631 162 30632 17462 30633 20472 30634 30634 30635 30635 30636 20475 30637 7315 30638 30638 30639 30639 30640 20635 30641 7315 30642 20635 30643 7315 30644 20643 30645 162 30646 17462 30647 20650 30648 30648 30649 30649 30650 20653 30651 7315 30652 30652 30653 943 30654 29136 30655 29137 30656 29138 30657 976 30658 159 30659 162 30660 29142 30661 29143 30662 2 30663 282 30664 29146 30665 21343 30666 2964 30667 2966 30668 77 30669 2968 30670 2969 30671 30671 30672 2 30673 282 30674 29149 30675 3042 30676 30676 30677 29150 30678 20856 30679 11 30680 2 30681 282 30682 7315 30683 30683 30684 16 30685 21004 30686 17462 30687 21008 30688 30688 30689 30689 30690 21011 30691 7315 30692 30692 30693 29151 30694 7315 30695 30695 30696 3042 30697 30697 30698 29152 30699 2962 30700 2964 30701 2966 30702 77 30703 2968 30704 2969 30705 30705 30706 7315 30707 30707 30708 29153 30709 5836 30710 434 30711 30711 30712 29154 30713 1862 30714 162 30715 29157 30716 29158 30717 29159 30718 29160 30719 282 30720 29162 30721 29163 30722 2 30723 574 30724 29166 30725 5236 30726 29168 30727 29169))
  "The generated action-tables.")

(defvar
  phps-mode-parser--distinct-action-tables
  #s(hash-table size 12466 test equal rehash-size 1.5 rehash-threshold 0.8125 data (0 ((("!") reduce 85) (("\"") reduce 85) (($) reduce 85) (("$") reduce 85) (("(") reduce 85) (("+") reduce 85) (("-") reduce 85) ((";") reduce 85) (("@") reduce 85) ((T_ABSTRACT) reduce 85) ((T_ARRAY) reduce 85) ((T_ARRAY_CAST) reduce 85) ((T_ATTRIBUTE) reduce 85) ((T_BOOL_CAST) reduce 85) ((T_BREAK) reduce 85) ((T_CLASS) reduce 85) ((T_CLASS_C) reduce 85) ((T_CLONE) reduce 85) ((T_CONST) reduce 85) ((T_CONSTANT_ENCAPSED_STRING) reduce 85) ((T_CONTINUE) reduce 85) ((T_DEC) reduce 85) ((T_DECLARE) reduce 85) ((T_DIR) reduce 85) ((T_DNUMBER) reduce 85) ((T_DO) reduce 85) ((T_DOUBLE_CAST) reduce 85) ((T_ECHO) reduce 85) ((T_EMPTY) reduce 85) ((T_ENUM) reduce 85) ((T_EVAL) reduce 85) ((T_EXIT) reduce 85) ((T_FILE) reduce 85) ((T_FINAL) reduce 85) ((T_FN) reduce 85) ((T_FOR) reduce 85) ((T_FOREACH) reduce 85) ((T_FUNCTION) reduce 85) ((T_FUNC_C) reduce 85) ((T_GLOBAL) reduce 85) ((T_GOTO) reduce 85) ((T_HALT_COMPILER) reduce 85) ((T_IF) reduce 85) ((T_INC) reduce 85) ((T_INCLUDE) reduce 85) ((T_INCLUDE_ONCE) reduce 85) ((T_INLINE_HTML) reduce 85) ((T_INTERFACE) reduce 85) ((T_INT_CAST) reduce 85) ((T_ISSET) reduce 85) ((T_LINE) reduce 85) ((T_LIST) reduce 85) ((T_LNUMBER) reduce 85) ((T_MATCH) reduce 85) ((T_METHOD_C) reduce 85) ((T_NAMESPACE) reduce 85) ((T_NAME_FULLY_QUALIFIED) reduce 85) ((T_NAME_QUALIFIED) reduce 85) ((T_NAME_RELATIVE) reduce 85) ((T_NEW) reduce 85) ((T_NS_C) reduce 85) ((T_OBJECT_CAST) reduce 85) ((T_PRINT) reduce 85) ((T_PROPERTY_C) reduce 85) ((T_READONLY) reduce 85) ((T_REQUIRE) reduce 85) ((T_REQUIRE_ONCE) reduce 85) ((T_RETURN) reduce 85) ((T_START_HEREDOC) reduce 85) ((T_STATIC) reduce 85) ((T_STRING) reduce 85) ((T_STRING_CAST) reduce 85) ((T_SWITCH) reduce 85) ((T_THROW) reduce 85) ((T_TRAIT) reduce 85) ((T_TRAIT_C) reduce 85) ((T_TRY) reduce 85) ((T_UNSET) reduce 85) ((T_UNSET_CAST) reduce 85) ((T_USE) reduce 85) ((T_VARIABLE) reduce 85) ((T_WHILE) reduce 85) ((T_YIELD) reduce 85) ((T_YIELD_FROM) reduce 85) (("[") reduce 85) (("`") reduce 85) (("{") reduce 85) (("~") reduce 85)) 1 ((("!") shift) (("\"") shift) (($) accept) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_USE) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 2 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 3 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_ENCAPSED_AND_WHITESPACE) shift) ((T_VARIABLE) shift)) 4 ((("$") shift) ((T_VARIABLE) shift) (("{") shift)) 8 ((("!") reduce 165) (("\"") reduce 165) (($) reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONST) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAMESPACE) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_USE) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 10 (((T_ABSTRACT) reduce 195) ((T_CLASS) reduce 195) ((T_FINAL) reduce 195) ((T_READONLY) reduce 195)) 11 ((("(") shift)) 13 (((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift)) 15 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") reduce 528) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 16 (((T_STRING) shift)) 17 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 20 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 22 ((("\"") shift) (("$") shift) (("(") shift) ((T_ARRAY) shift) ((T_CLASS_C) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DIR) shift) ((T_FILE) shift) ((T_FUNC_C) shift) ((T_LINE) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_TRAIT_C) shift) ((T_VARIABLE) shift) (("[") shift)) 24 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 25 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507)) 26 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 32 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) ((";") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 33 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 34 (((T_ABSTRACT) reduce 196) ((T_CLASS) reduce 196) ((T_FINAL) reduce 196) ((T_READONLY) reduce 196)) 35 ((("(") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474)) 38 ((("(") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_READONLY) reduce 475) ((T_STRING) reduce 475)) 39 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 40 ((("$") shift) ((T_VARIABLE) shift)) 47 ((("!") reduce 159) (("\"") reduce 159) (($) reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONST) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAMESPACE) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_USE) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 51 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 53 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506)) 55 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 56 (((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_QUALIFIED) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift) (("{") shift)) 57 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 58 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 59 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 60 ((("$") shift) (("(") shift) ((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CLASS) reduce 193) ((T_FINAL) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_VARIABLE) shift)) 61 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 64 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 65 ((("(") shift) ((T_ABSTRACT) reduce 197) ((T_CLASS) reduce 197) ((T_FINAL) reduce 197) ((T_READONLY) reduce 197)) 69 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_ENCAPSED_AND_WHITESPACE) shift) ((T_END_HEREDOC) shift) ((T_VARIABLE) shift)) 70 (((T_FN) shift) ((T_FUNCTION) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_VARIABLE) shift)) 71 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") shift) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 76 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 77 ((("{") shift)) 80 (((T_CONST) shift) ((T_FUNCTION) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_STRING) shift)) 81 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 83 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) (("-") shift) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 85 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) ((",") reduce 569) (("-") shift) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_ELLIPSIS) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 569) (("`") shift) (("~") shift)) 86 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_ENCAPSED_AND_WHITESPACE) shift) ((T_VARIABLE) shift) (("`") reduce 497)) 87 ((("!") reduce 148) (("\"") reduce 148) (($) reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONST) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAMESPACE) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_USE) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 88 (((T_ELSE) shift) ((T_ELSEIF) shift) ((T_ENDIF) shift)) 89 (((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) (("[") shift)) 90 (((T_ABSTRACT) reduce 101) ((T_ATTRIBUTE) reduce 101) ((T_CLASS) reduce 101) ((T_ENUM) reduce 101) ((T_FINAL) reduce 101) ((T_FN) reduce 101) ((T_FUNCTION) reduce 101) ((T_INTERFACE) reduce 101) ((T_READONLY) reduce 101) ((T_STATIC) reduce 101) ((T_TRAIT) reduce 101)) 91 ((("!") reduce 109) (("\"") reduce 109) (("$") reduce 109) (($) reduce 109) (("(") reduce 109) (("+") reduce 109) (("-") reduce 109) ((";") reduce 109) (("@") reduce 109) ((T_ABSTRACT) reduce 109) ((T_ARRAY) reduce 109) ((T_ARRAY_CAST) reduce 109) ((T_ATTRIBUTE) reduce 109) ((T_BOOL_CAST) reduce 109) ((T_BREAK) reduce 109) ((T_CLASS) reduce 109) ((T_CLASS_C) reduce 109) ((T_CLONE) reduce 109) ((T_CONST) reduce 109) ((T_CONSTANT_ENCAPSED_STRING) reduce 109) ((T_CONTINUE) reduce 109) ((T_DEC) reduce 109) ((T_DECLARE) reduce 109) ((T_DIR) reduce 109) ((T_DNUMBER) reduce 109) ((T_DO) reduce 109) ((T_DOUBLE_CAST) reduce 109) ((T_ECHO) reduce 109) ((T_EMPTY) reduce 109) ((T_ENUM) reduce 109) ((T_EVAL) reduce 109) ((T_EXIT) reduce 109) ((T_FILE) reduce 109) ((T_FINAL) reduce 109) ((T_FN) reduce 109) ((T_FOR) reduce 109) ((T_FOREACH) reduce 109) ((T_FUNCTION) reduce 109) ((T_FUNC_C) reduce 109) ((T_GLOBAL) reduce 109) ((T_GOTO) reduce 109) ((T_HALT_COMPILER) reduce 109) ((T_IF) reduce 109) ((T_INC) reduce 109) ((T_INCLUDE) reduce 109) ((T_INCLUDE_ONCE) reduce 109) ((T_INLINE_HTML) reduce 109) ((T_INTERFACE) reduce 109) ((T_INT_CAST) reduce 109) ((T_ISSET) reduce 109) ((T_LINE) reduce 109) ((T_LIST) reduce 109) ((T_LNUMBER) reduce 109) ((T_MATCH) reduce 109) ((T_METHOD_C) reduce 109) ((T_NAMESPACE) reduce 109) ((T_NAME_FULLY_QUALIFIED) reduce 109) ((T_NAME_QUALIFIED) reduce 109) ((T_NAME_RELATIVE) reduce 109) ((T_NEW) reduce 109) ((T_NS_C) reduce 109) ((T_OBJECT_CAST) reduce 109) ((T_PRINT) reduce 109) ((T_PROPERTY_C) reduce 109) ((T_READONLY) reduce 109) ((T_REQUIRE) reduce 109) ((T_REQUIRE_ONCE) reduce 109) ((T_RETURN) reduce 109) ((T_START_HEREDOC) reduce 109) ((T_STATIC) reduce 109) ((T_STRING) reduce 109) ((T_STRING_CAST) reduce 109) ((T_SWITCH) reduce 109) ((T_THROW) reduce 109) ((T_TRAIT) reduce 109) ((T_TRAIT_C) reduce 109) ((T_TRY) reduce 109) ((T_UNSET) reduce 109) ((T_UNSET_CAST) reduce 109) ((T_USE) reduce 109) ((T_VARIABLE) reduce 109) ((T_WHILE) reduce 109) ((T_YIELD) reduce 109) ((T_YIELD_FROM) reduce 109) (("[") reduce 109) (("`") reduce 109) (("{") reduce 109) (("~") reduce 109)) 92 (((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CLASS) shift) ((T_ENUM) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_INTERFACE) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_TRAIT) shift)) 94 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((";") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 95 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513)) 96 ((("!") reduce 104) (("\"") reduce 104) (($) reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONST) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAMESPACE) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_USE) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 97 (((T_ABSTRACT) reduce 189) ((T_CLASS) reduce 189) ((T_FINAL) reduce 189) ((T_READONLY) reduce 189)) 98 (((T_ABSTRACT) shift) ((T_CLASS) shift) ((T_FINAL) shift) ((T_READONLY) shift)) 99 (((T_PAAMAYIM_NEKUDOTAYIM) shift)) 100 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512)) 101 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511)) 102 ((("!") reduce 107) (("\"") reduce 107) (($) reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONST) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAMESPACE) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_USE) reduce 107) ((T_VARIABLE) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 103 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 104 ((("(") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift)) 105 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) (("[") reduce 536)) 106 ((("(") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_READONLY) reduce 479) ((T_STRING) reduce 479)) 107 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((";") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 108 ((("!") reduce 103) (("\"") reduce 103) (($) reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONST) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAMESPACE) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_USE) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 109 ((("!") reduce 147) (("\"") reduce 147) (($) reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONST) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAMESPACE) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_USE) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 110 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (($) reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONST) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAMESPACE) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_USE) reduce 244) ((T_VARIABLE) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 111 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 112 ((("!") reduce 106) (("\"") reduce 106) (($) reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONST) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAMESPACE) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_USE) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 113 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) ((";") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449)) 114 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 115 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((";") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514)) 116 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444)) 117 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 118 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 119 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 120 ((("!") reduce 108) (("\"") reduce 108) (("$") reduce 108) (($) reduce 108) (("(") reduce 108) (("+") reduce 108) (("-") reduce 108) ((";") reduce 108) (("@") reduce 108) ((T_ABSTRACT) reduce 108) ((T_ARRAY) reduce 108) ((T_ARRAY_CAST) reduce 108) ((T_ATTRIBUTE) reduce 108) ((T_BOOL_CAST) reduce 108) ((T_BREAK) reduce 108) ((T_CLASS) reduce 108) ((T_CLASS_C) reduce 108) ((T_CLONE) reduce 108) ((T_CONST) reduce 108) ((T_CONSTANT_ENCAPSED_STRING) reduce 108) ((T_CONTINUE) reduce 108) ((T_DEC) reduce 108) ((T_DECLARE) reduce 108) ((T_DIR) reduce 108) ((T_DNUMBER) reduce 108) ((T_DO) reduce 108) ((T_DOUBLE_CAST) reduce 108) ((T_ECHO) reduce 108) ((T_EMPTY) reduce 108) ((T_ENUM) reduce 108) ((T_EVAL) reduce 108) ((T_EXIT) reduce 108) ((T_FILE) reduce 108) ((T_FINAL) reduce 108) ((T_FN) reduce 108) ((T_FOR) reduce 108) ((T_FOREACH) reduce 108) ((T_FUNCTION) reduce 108) ((T_FUNC_C) reduce 108) ((T_GLOBAL) reduce 108) ((T_GOTO) reduce 108) ((T_HALT_COMPILER) reduce 108) ((T_IF) reduce 108) ((T_INC) reduce 108) ((T_INCLUDE) reduce 108) ((T_INCLUDE_ONCE) reduce 108) ((T_INLINE_HTML) reduce 108) ((T_INTERFACE) reduce 108) ((T_INT_CAST) reduce 108) ((T_ISSET) reduce 108) ((T_LINE) reduce 108) ((T_LIST) reduce 108) ((T_LNUMBER) reduce 108) ((T_MATCH) reduce 108) ((T_METHOD_C) reduce 108) ((T_NAMESPACE) reduce 108) ((T_NAME_FULLY_QUALIFIED) reduce 108) ((T_NAME_QUALIFIED) reduce 108) ((T_NAME_RELATIVE) reduce 108) ((T_NEW) reduce 108) ((T_NS_C) reduce 108) ((T_OBJECT_CAST) reduce 108) ((T_PRINT) reduce 108) ((T_PROPERTY_C) reduce 108) ((T_READONLY) reduce 108) ((T_REQUIRE) reduce 108) ((T_REQUIRE_ONCE) reduce 108) ((T_RETURN) reduce 108) ((T_START_HEREDOC) reduce 108) ((T_STATIC) reduce 108) ((T_STRING) reduce 108) ((T_STRING_CAST) reduce 108) ((T_SWITCH) reduce 108) ((T_THROW) reduce 108) ((T_TRAIT) reduce 108) ((T_TRAIT_C) reduce 108) ((T_TRY) reduce 108) ((T_UNSET) reduce 108) ((T_UNSET_CAST) reduce 108) ((T_USE) reduce 108) ((T_VARIABLE) reduce 108) ((T_WHILE) reduce 108) ((T_YIELD) reduce 108) ((T_YIELD_FROM) reduce 108) (("[") reduce 108) (("`") reduce 108) (("{") reduce 108) (("~") reduce 108)) 121 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((";") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 122 ((("!") reduce 84) (("\"") reduce 84) (($) reduce 84) (("$") reduce 84) (("(") reduce 84) (("+") reduce 84) (("-") reduce 84) ((";") reduce 84) (("@") reduce 84) ((T_ABSTRACT) reduce 84) ((T_ARRAY) reduce 84) ((T_ARRAY_CAST) reduce 84) ((T_ATTRIBUTE) reduce 84) ((T_BOOL_CAST) reduce 84) ((T_BREAK) reduce 84) ((T_CLASS) reduce 84) ((T_CLASS_C) reduce 84) ((T_CLONE) reduce 84) ((T_CONST) reduce 84) ((T_CONSTANT_ENCAPSED_STRING) reduce 84) ((T_CONTINUE) reduce 84) ((T_DEC) reduce 84) ((T_DECLARE) reduce 84) ((T_DIR) reduce 84) ((T_DNUMBER) reduce 84) ((T_DO) reduce 84) ((T_DOUBLE_CAST) reduce 84) ((T_ECHO) reduce 84) ((T_EMPTY) reduce 84) ((T_ENUM) reduce 84) ((T_EVAL) reduce 84) ((T_EXIT) reduce 84) ((T_FILE) reduce 84) ((T_FINAL) reduce 84) ((T_FN) reduce 84) ((T_FOR) reduce 84) ((T_FOREACH) reduce 84) ((T_FUNCTION) reduce 84) ((T_FUNC_C) reduce 84) ((T_GLOBAL) reduce 84) ((T_GOTO) reduce 84) ((T_HALT_COMPILER) reduce 84) ((T_IF) reduce 84) ((T_INC) reduce 84) ((T_INCLUDE) reduce 84) ((T_INCLUDE_ONCE) reduce 84) ((T_INLINE_HTML) reduce 84) ((T_INTERFACE) reduce 84) ((T_INT_CAST) reduce 84) ((T_ISSET) reduce 84) ((T_LINE) reduce 84) ((T_LIST) reduce 84) ((T_LNUMBER) reduce 84) ((T_MATCH) reduce 84) ((T_METHOD_C) reduce 84) ((T_NAMESPACE) reduce 84) ((T_NAME_FULLY_QUALIFIED) reduce 84) ((T_NAME_QUALIFIED) reduce 84) ((T_NAME_RELATIVE) reduce 84) ((T_NEW) reduce 84) ((T_NS_C) reduce 84) ((T_OBJECT_CAST) reduce 84) ((T_PRINT) reduce 84) ((T_PROPERTY_C) reduce 84) ((T_READONLY) reduce 84) ((T_REQUIRE) reduce 84) ((T_REQUIRE_ONCE) reduce 84) ((T_RETURN) reduce 84) ((T_START_HEREDOC) reduce 84) ((T_STATIC) reduce 84) ((T_STRING) reduce 84) ((T_STRING_CAST) reduce 84) ((T_SWITCH) reduce 84) ((T_THROW) reduce 84) ((T_TRAIT) reduce 84) ((T_TRAIT_C) reduce 84) ((T_TRY) reduce 84) ((T_UNSET) reduce 84) ((T_UNSET_CAST) reduce 84) ((T_USE) reduce 84) ((T_VARIABLE) reduce 84) ((T_WHILE) reduce 84) ((T_YIELD) reduce 84) ((T_YIELD_FROM) reduce 84) (("[") reduce 84) (("`") reduce 84) (("{") reduce 84) (("~") reduce 84)) 123 ((("!") reduce 105) (("\"") reduce 105) (($) reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONST) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAMESPACE) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_USE) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 124 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) ((";") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388)) 126 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("}") reduce 141) (("~") reduce 141)) 129 ((("(") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475)) 131 (((T_FN) shift) ((T_FUNCTION) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492)) 132 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 133 (((T_ATTRIBUTE) reduce 101) ((T_FN) reduce 101) ((T_FUNCTION) reduce 101) ((T_STATIC) reduce 101)) 134 (((T_ATTRIBUTE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_STATIC) shift)) 135 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) ((";") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432)) 137 ((("(") reduce 80)) 138 ((("(") reduce 81)) 139 ((("(") reduce 480)) 140 ((("(") reduce 476)) 142 ((("(") reduce 256) ((")") reduce 251) (("?") reduce 256) ((T_ABSTRACT) shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 256) ((T_ARRAY) reduce 256) ((T_ATTRIBUTE) shift) ((T_CALLABLE) reduce 256) ((T_ELLIPSIS) reduce 256) ((T_FINAL) shift) ((T_NAME_FULLY_QUALIFIED) reduce 256) ((T_NAME_QUALIFIED) reduce 256) ((T_NAME_RELATIVE) reduce 256) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 256) ((T_VARIABLE) reduce 256)) 143 ((("(") reduce 349) (("?") reduce 349) ((T_ABSTRACT) reduce 349) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 349) ((T_ARRAY) reduce 349) ((T_CALLABLE) reduce 349) ((T_ELLIPSIS) reduce 349) ((T_FINAL) reduce 349) ((T_NAME_FULLY_QUALIFIED) reduce 349) ((T_NAME_QUALIFIED) reduce 349) ((T_NAME_RELATIVE) reduce 349) ((T_PRIVATE) reduce 349) ((T_PRIVATE_SET) reduce 349) ((T_PROTECTED) reduce 349) ((T_PROTECTED_SET) reduce 349) ((T_PUBLIC) reduce 349) ((T_PUBLIC_SET) reduce 349) ((T_READONLY) reduce 349) ((T_STATIC) reduce 349) ((T_STRING) reduce 349) ((T_VARIABLE) reduce 349)) 145 ((("(") reduce 350) (("?") reduce 350) ((T_ABSTRACT) reduce 350) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 350) ((T_ARRAY) reduce 350) ((T_CALLABLE) reduce 350) ((T_ELLIPSIS) reduce 350) ((T_FINAL) reduce 350) ((T_NAME_FULLY_QUALIFIED) reduce 350) ((T_NAME_QUALIFIED) reduce 350) ((T_NAME_RELATIVE) reduce 350) ((T_PRIVATE) reduce 350) ((T_PRIVATE_SET) reduce 350) ((T_PROTECTED) reduce 350) ((T_PROTECTED_SET) reduce 350) ((T_PUBLIC) reduce 350) ((T_PUBLIC_SET) reduce 350) ((T_READONLY) reduce 350) ((T_STATIC) reduce 350) ((T_STRING) reduce 350) ((T_VARIABLE) reduce 350)) 146 ((("(") reduce 344) (("?") reduce 344) ((T_ABSTRACT) reduce 344) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 344) ((T_ARRAY) reduce 344) ((T_CALLABLE) reduce 344) ((T_ELLIPSIS) reduce 344) ((T_FINAL) reduce 344) ((T_NAME_FULLY_QUALIFIED) reduce 344) ((T_NAME_QUALIFIED) reduce 344) ((T_NAME_RELATIVE) reduce 344) ((T_PRIVATE) reduce 344) ((T_PRIVATE_SET) reduce 344) ((T_PROTECTED) reduce 344) ((T_PROTECTED_SET) reduce 344) ((T_PUBLIC) reduce 344) ((T_PUBLIC_SET) reduce 344) ((T_READONLY) reduce 344) ((T_STATIC) reduce 344) ((T_STRING) reduce 344) ((T_VARIABLE) reduce 344)) 147 ((("(") reduce 347) (("?") reduce 347) ((T_ABSTRACT) reduce 347) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 347) ((T_ARRAY) reduce 347) ((T_CALLABLE) reduce 347) ((T_ELLIPSIS) reduce 347) ((T_FINAL) reduce 347) ((T_NAME_FULLY_QUALIFIED) reduce 347) ((T_NAME_QUALIFIED) reduce 347) ((T_NAME_RELATIVE) reduce 347) ((T_PRIVATE) reduce 347) ((T_PRIVATE_SET) reduce 347) ((T_PROTECTED) reduce 347) ((T_PROTECTED_SET) reduce 347) ((T_PUBLIC) reduce 347) ((T_PUBLIC_SET) reduce 347) ((T_READONLY) reduce 347) ((T_STATIC) reduce 347) ((T_STRING) reduce 347) ((T_VARIABLE) reduce 347)) 148 ((("(") reduce 343) (("?") reduce 343) ((T_ABSTRACT) reduce 343) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 343) ((T_ARRAY) reduce 343) ((T_CALLABLE) reduce 343) ((T_ELLIPSIS) reduce 343) ((T_FINAL) reduce 343) ((T_NAME_FULLY_QUALIFIED) reduce 343) ((T_NAME_QUALIFIED) reduce 343) ((T_NAME_RELATIVE) reduce 343) ((T_PRIVATE) reduce 343) ((T_PRIVATE_SET) reduce 343) ((T_PROTECTED) reduce 343) ((T_PROTECTED_SET) reduce 343) ((T_PUBLIC) reduce 343) ((T_PUBLIC_SET) reduce 343) ((T_READONLY) reduce 343) ((T_STATIC) reduce 343) ((T_STRING) reduce 343) ((T_VARIABLE) reduce 343)) 149 ((("(") reduce 346) (("?") reduce 346) ((T_ABSTRACT) reduce 346) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 346) ((T_ARRAY) reduce 346) ((T_CALLABLE) reduce 346) ((T_ELLIPSIS) reduce 346) ((T_FINAL) reduce 346) ((T_NAME_FULLY_QUALIFIED) reduce 346) ((T_NAME_QUALIFIED) reduce 346) ((T_NAME_RELATIVE) reduce 346) ((T_PRIVATE) reduce 346) ((T_PRIVATE_SET) reduce 346) ((T_PROTECTED) reduce 346) ((T_PROTECTED_SET) reduce 346) ((T_PUBLIC) reduce 346) ((T_PUBLIC_SET) reduce 346) ((T_READONLY) reduce 346) ((T_STATIC) reduce 346) ((T_STRING) reduce 346) ((T_VARIABLE) reduce 346)) 150 ((("(") reduce 342) (("?") reduce 342) ((T_ABSTRACT) reduce 342) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 342) ((T_ARRAY) reduce 342) ((T_CALLABLE) reduce 342) ((T_ELLIPSIS) reduce 342) ((T_FINAL) reduce 342) ((T_NAME_FULLY_QUALIFIED) reduce 342) ((T_NAME_QUALIFIED) reduce 342) ((T_NAME_RELATIVE) reduce 342) ((T_PRIVATE) reduce 342) ((T_PRIVATE_SET) reduce 342) ((T_PROTECTED) reduce 342) ((T_PROTECTED_SET) reduce 342) ((T_PUBLIC) reduce 342) ((T_PUBLIC_SET) reduce 342) ((T_READONLY) reduce 342) ((T_STATIC) reduce 342) ((T_STRING) reduce 342) ((T_VARIABLE) reduce 342)) 151 ((("(") reduce 345) (("?") reduce 345) ((T_ABSTRACT) reduce 345) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 345) ((T_ARRAY) reduce 345) ((T_CALLABLE) reduce 345) ((T_ELLIPSIS) reduce 345) ((T_FINAL) reduce 345) ((T_NAME_FULLY_QUALIFIED) reduce 345) ((T_NAME_QUALIFIED) reduce 345) ((T_NAME_RELATIVE) reduce 345) ((T_PRIVATE) reduce 345) ((T_PRIVATE_SET) reduce 345) ((T_PROTECTED) reduce 345) ((T_PROTECTED_SET) reduce 345) ((T_PUBLIC) reduce 345) ((T_PUBLIC_SET) reduce 345) ((T_READONLY) reduce 345) ((T_STATIC) reduce 345) ((T_STRING) reduce 345) ((T_VARIABLE) reduce 345)) 152 ((("(") reduce 351) (("?") reduce 351) ((T_ABSTRACT) reduce 351) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 351) ((T_ARRAY) reduce 351) ((T_CALLABLE) reduce 351) ((T_ELLIPSIS) reduce 351) ((T_FINAL) reduce 351) ((T_NAME_FULLY_QUALIFIED) reduce 351) ((T_NAME_QUALIFIED) reduce 351) ((T_NAME_RELATIVE) reduce 351) ((T_PRIVATE) reduce 351) ((T_PRIVATE_SET) reduce 351) ((T_PROTECTED) reduce 351) ((T_PROTECTED_SET) reduce 351) ((T_PUBLIC) reduce 351) ((T_PUBLIC_SET) reduce 351) ((T_READONLY) reduce 351) ((T_STATIC) reduce 351) ((T_STRING) reduce 351) ((T_VARIABLE) reduce 351)) 153 ((("(") reduce 348) (("?") reduce 348) ((T_ABSTRACT) reduce 348) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 348) ((T_ARRAY) reduce 348) ((T_CALLABLE) reduce 348) ((T_ELLIPSIS) reduce 348) ((T_FINAL) reduce 348) ((T_NAME_FULLY_QUALIFIED) reduce 348) ((T_NAME_QUALIFIED) reduce 348) ((T_NAME_RELATIVE) reduce 348) ((T_PRIVATE) reduce 348) ((T_PRIVATE_SET) reduce 348) ((T_PROTECTED) reduce 348) ((T_PROTECTED_SET) reduce 348) ((T_PUBLIC) reduce 348) ((T_PUBLIC_SET) reduce 348) ((T_READONLY) reduce 348) ((T_STATIC) reduce 348) ((T_STRING) reduce 348) ((T_VARIABLE) reduce 348)) 154 ((("(") reduce 101) (("?") reduce 101) ((T_ABSTRACT) reduce 101) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 101) ((T_ARRAY) reduce 101) ((T_ATTRIBUTE) reduce 101) ((T_CALLABLE) reduce 101) ((T_ELLIPSIS) reduce 101) ((T_FINAL) reduce 101) ((T_NAME_FULLY_QUALIFIED) reduce 101) ((T_NAME_QUALIFIED) reduce 101) ((T_NAME_RELATIVE) reduce 101) ((T_PRIVATE) reduce 101) ((T_PRIVATE_SET) reduce 101) ((T_PROTECTED) reduce 101) ((T_PROTECTED_SET) reduce 101) ((T_PUBLIC) reduce 101) ((T_PUBLIC_SET) reduce 101) ((T_READONLY) reduce 101) ((T_STATIC) reduce 101) ((T_STRING) reduce 101) ((T_VARIABLE) reduce 101)) 155 (((")") reduce 252) ((",") reduce 252)) 156 ((("(") reduce 256) (("?") reduce 256) ((T_ABSTRACT) shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 256) ((T_ARRAY) reduce 256) ((T_ATTRIBUTE) shift) ((T_CALLABLE) reduce 256) ((T_ELLIPSIS) reduce 256) ((T_FINAL) shift) ((T_NAME_FULLY_QUALIFIED) reduce 256) ((T_NAME_QUALIFIED) reduce 256) ((T_NAME_RELATIVE) reduce 256) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 256) ((T_VARIABLE) reduce 256)) 157 ((("(") reduce 340) (("?") reduce 340) ((T_ABSTRACT) reduce 340) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 340) ((T_ARRAY) reduce 340) ((T_CALLABLE) reduce 340) ((T_ELLIPSIS) reduce 340) ((T_FINAL) reduce 340) ((T_NAME_FULLY_QUALIFIED) reduce 340) ((T_NAME_QUALIFIED) reduce 340) ((T_NAME_RELATIVE) reduce 340) ((T_PRIVATE) reduce 340) ((T_PRIVATE_SET) reduce 340) ((T_PROTECTED) reduce 340) ((T_PROTECTED_SET) reduce 340) ((T_PUBLIC) reduce 340) ((T_PUBLIC_SET) reduce 340) ((T_READONLY) reduce 340) ((T_STATIC) reduce 340) ((T_STRING) reduce 340) ((T_VARIABLE) reduce 340)) 158 ((("(") reduce 257) (("?") reduce 257) ((T_ABSTRACT) shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 257) ((T_ARRAY) reduce 257) ((T_CALLABLE) reduce 257) ((T_ELLIPSIS) reduce 257) ((T_FINAL) shift) ((T_NAME_FULLY_QUALIFIED) reduce 257) ((T_NAME_QUALIFIED) reduce 257) ((T_NAME_RELATIVE) reduce 257) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 257) ((T_VARIABLE) reduce 257)) 159 (((")") reduce 124) ((",") shift)) 160 ((("(") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 260) ((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_ELLIPSIS) reduce 260) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STRING) shift) ((T_VARIABLE) reduce 260)) 161 (((")") reduce 255) ((",") reduce 255)) 162 (((")") shift)) 163 (((":") reduce 481) ((T_USE) shift) (("{") reduce 481)) 165 (((":") shift) (("{") reduce 287)) 166 ((("(") shift) (("?") shift) ((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift)) 167 ((("{") reduce 477)) 170 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("}") shift) (("~") shift)) 171 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 185 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 195 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 197 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("}") reduce 143) (("~") reduce 143)) 199 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("}") reduce 104) (("~") reduce 104)) 201 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("}") reduce 107) (("~") reduce 107)) 204 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("}") reduce 103) (("~") reduce 103)) 205 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 206 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("}") reduce 244) (("~") reduce 244)) 207 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_PROPERTY_C) reduce 140) ((T_READONLY) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("}") reduce 140) (("~") reduce 140)) 208 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("}") reduce 106) (("~") reduce 106)) 209 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("}") reduce 142) (("~") reduce 142)) 210 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("}") reduce 105) (("~") reduce 105)) 212 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 213 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 215 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("}") reduce 146) (("~") reduce 146)) 229 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 231 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 233 ((("%") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 234 ((("%") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507)) 238 ((("%") reduce 500) (("(") shift) ((")") reduce 500) (("*") reduce 500) (("+") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 239 ((("%") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 240 ((("%") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 246 ((("%") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 248 ((("%") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506)) 250 ((("%") reduce 519) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 251 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 252 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 253 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 255 ((("%") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 258 ((("%") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 264 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 267 ((("%") reduce 518) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 269 ((("%") reduce 551) (("(") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 270 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) ((")") reduce 462) (("*") reduce 462) (("+") shift) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 277 ((("%") reduce 547) (("(") reduce 538) ((")") reduce 547) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 278 ((("%") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513)) 280 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512)) 281 ((("%") reduce 511) (("(") reduce 540) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511)) 282 ((("%") shift) ((")") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 285 ((("%") reduce 546) (("(") reduce 546) ((")") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 286 ((("%") reduce 467) ((")") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 287 ((("%") reduce 449) ((")") reduce 449) (("*") reduce 449) (("+") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449)) 288 ((("%") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 289 ((("%") reduce 514) (("(") shift) ((")") reduce 514) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514)) 290 ((("%") reduce 444) (("(") reduce 541) ((")") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444)) 291 ((("%") reduce 445) ((")") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 292 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 293 ((("%") reduce 542) (("(") reduce 542) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 294 ((("%") reduce 548) ((")") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 295 ((("%") reduce 388) ((")") reduce 388) (("*") reduce 388) (("+") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388)) 298 ((("%") reduce 432) ((")") reduce 432) (("*") reduce 432) (("+") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432)) 307 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 313 ((("$") shift) (("(") shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_VARIABLE) shift)) 329 ((("%") shift) ((")") reduce 416) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416)) 330 ((("%") shift) ((")") reduce 419) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419)) 331 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428)) 332 ((("%") shift) ((")") reduce 441) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441)) 333 ((("%") shift) ((")") reduce 427) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 334 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 335 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 336 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 337 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 338 ((("%") shift) ((")") reduce 438) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438)) 339 ((("%") shift) ((")") reduce 434) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 340 ((("%") shift) ((")") reduce 436) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 341 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 342 ((("%") shift) ((")") reduce 440) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440)) 343 ((("%") shift) ((")") reduce 435) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 346 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 347 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 348 ((("%") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 349 ((("%") reduce 492) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 350 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 351 ((("%") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 352 ((("%") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 353 ((("%") reduce 442) ((")") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 354 ((("%") reduce 493) ((")") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 355 ((("%") reduce 495) ((")") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 356 ((("%") reduce 556) ((")") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 357 ((("$") shift) ((T_STRING) shift) ((T_VARIABLE) shift) (("{") shift)) 360 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 528) (("`") shift) (("~") shift)) 371 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("]") reduce 523) (("^") reduce 523) (("|") reduce 523)) 373 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("]") reduce 504) (("^") reduce 504) (("|") reduce 504)) 375 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("|") reduce 517)) 376 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("]") reduce 507) (("^") reduce 507) (("|") reduce 507)) 380 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("]") reduce 500) (("^") reduce 500) (("|") reduce 500)) 381 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("|") reduce 516)) 382 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("|") reduce 520)) 388 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("|") reduce 515)) 390 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("]") reduce 506) (("^") reduce 506) (("|") reduce 506)) 392 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("|") reduce 519)) 393 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 394 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 395 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 397 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("|") reduce 522)) 400 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("|") reduce 521)) 406 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 409 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("]") reduce 518) (("^") reduce 518) (("|") reduce 518)) 411 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("]") reduce 551) (("^") reduce 551) (("|") reduce 551)) 412 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 462) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 419 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("]") reduce 547) (("^") reduce 547) (("|") reduce 547)) 420 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("]") reduce 513) (("^") reduce 513) (("|") reduce 513)) 422 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("]") reduce 512) (("^") reduce 512) (("|") reduce 512)) 423 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("]") reduce 511) (("^") reduce 511) (("|") reduce 511)) 424 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 529) (("^") shift) (("|") shift)) 427 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("]") reduce 546) (("^") reduce 546) (("|") reduce 546)) 428 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("]") reduce 467) (("^") reduce 467) (("|") reduce 467)) 429 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("]") reduce 449) (("^") reduce 449) (("|") reduce 449)) 430 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("]") reduce 471) (("^") reduce 471) (("|") reduce 471)) 431 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("]") reduce 514) (("^") reduce 514) (("|") reduce 514)) 432 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("]") reduce 444) (("^") reduce 444) (("|") reduce 444)) 433 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("]") reduce 445) (("^") reduce 445) (("|") reduce 445)) 434 ((("]") shift)) 435 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 436 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("]") reduce 542) (("^") reduce 542) (("|") reduce 542)) 437 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("]") reduce 548) (("^") reduce 548) (("|") reduce 548)) 438 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("]") reduce 388) (("^") reduce 388) (("|") reduce 388)) 441 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("]") reduce 432) (("^") reduce 432) (("|") reduce 432)) 472 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 416) (("^") shift) (("|") reduce 416)) 473 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 419) (("^") reduce 419) (("|") reduce 419)) 474 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("]") reduce 428) (("^") reduce 428) (("|") reduce 428)) 475 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 441) (("^") reduce 441) (("|") reduce 441)) 476 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("]") reduce 427) (("^") reduce 427) (("|") reduce 427)) 477 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("]") reduce 424) (("^") reduce 424) (("|") reduce 424)) 478 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 415) (("^") shift) (("|") shift)) 479 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 413) (("^") shift) (("|") shift)) 480 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 414) (("^") shift) (("|") shift)) 481 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("]") reduce 438) (("^") reduce 438) (("|") reduce 438)) 482 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 434) (("^") reduce 434) (("|") reduce 434)) 483 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 436) (("^") reduce 436) (("|") reduce 436)) 484 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 433) (("^") reduce 433) (("|") reduce 433)) 485 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("]") reduce 440) (("^") reduce 440) (("|") reduce 440)) 486 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 435) (("^") reduce 435) (("|") reduce 435)) 489 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 490 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 491 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 492 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("]") reduce 492) (("^") reduce 492) (("|") reduce 492)) 493 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 494 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("]") reduce 551) (("^") reduce 551) (("|") reduce 551)) 495 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("]") reduce 494) (("^") reduce 494) (("|") reduce 494)) 496 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("]") reduce 442) (("^") reduce 442) (("|") reduce 442)) 497 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("]") reduce 493) (("^") reduce 493) (("|") reduce 493)) 498 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("]") reduce 495) (("^") reduce 495) (("|") reduce 495)) 499 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("]") reduce 556) (("^") reduce 556) (("|") reduce 556)) 505 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("]") reduce 557) (("^") reduce 557) (("|") reduce 557)) 506 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 507 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 508 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("]") reduce 558) (("^") reduce 558) (("|") reduce 558)) 509 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 521 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523) (("}") reduce 523)) 523 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504) (("}") reduce 504)) 525 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517) (("}") reduce 517)) 526 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507) (("}") reduce 507)) 530 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500) (("}") reduce 500)) 531 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516) (("}") reduce 516)) 532 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520) (("}") reduce 520)) 538 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515) (("}") reduce 515)) 540 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506) (("}") reduce 506)) 542 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519) (("}") reduce 519)) 543 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 544 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 545 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 547 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522) (("}") reduce 522)) 550 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521) (("}") reduce 521)) 556 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 559 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518) (("}") reduce 518)) 561 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551) (("}") reduce 551)) 562 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("}") reduce 462) (("~") shift)) 569 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547) (("}") reduce 547)) 570 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513) (("}") reduce 513)) 572 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512) (("}") reduce 512)) 573 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511) (("}") reduce 511)) 574 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") shift)) 577 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546) (("}") reduce 546)) 578 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467) (("}") reduce 467)) 579 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449) (("}") reduce 449)) 580 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471) (("}") reduce 471)) 581 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514) (("}") reduce 514)) 582 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444) (("}") reduce 444)) 583 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445) (("}") reduce 445)) 584 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 585 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542) (("}") reduce 542)) 586 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548) (("}") reduce 548)) 587 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388) (("}") reduce 388)) 590 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432) (("}") reduce 432)) 621 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416) (("}") reduce 416)) 622 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419) (("}") reduce 419)) 623 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428) (("}") reduce 428)) 624 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441) (("}") reduce 441)) 625 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427) (("}") reduce 427)) 626 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424) (("}") reduce 424)) 627 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 415)) 628 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 413)) 629 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 414)) 630 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438) (("}") reduce 438)) 631 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434) (("}") reduce 434)) 632 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436) (("}") reduce 436)) 633 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433) (("}") reduce 433)) 634 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440) (("}") reduce 440)) 635 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435) (("}") reduce 435)) 638 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 639 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 640 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 641 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492) (("}") reduce 492)) 642 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 643 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551) (("}") reduce 551)) 644 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494) (("}") reduce 494)) 645 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442) (("}") reduce 442)) 646 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493) (("}") reduce 493)) 647 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495) (("}") reduce 495)) 648 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556) (("}") reduce 556)) 654 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557) (("}") reduce 557)) 655 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 656 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 657 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558) (("}") reduce 558)) 658 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 661 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 662 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 664 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 666 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496) (("}") reduce 496)) 667 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553) (("}") reduce 553)) 670 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552) (("}") reduce 552)) 671 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 448)) 672 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 411)) 673 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 412)) 674 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417) (("}") reduce 417)) 675 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418) (("}") reduce 418)) 687 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 689 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((":") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 691 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 692 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507)) 696 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) ((":") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 697 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 698 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 704 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 706 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506)) 708 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 709 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 710 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 711 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 713 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 716 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 722 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 725 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 727 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((":") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 728 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) (("-") shift) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 735 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((":") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 736 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513)) 738 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512)) 739 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511)) 740 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 743 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((":") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 744 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((":") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 745 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) ((":") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449)) 746 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((":") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 747 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((":") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514)) 748 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((":") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444)) 749 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((":") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 750 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 751 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((":") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 752 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((":") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 753 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) ((":") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388)) 756 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) ((":") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432)) 787 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 416) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416)) 788 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 419) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419)) 789 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 428) (("/") shift) ((":") reduce 428) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428)) 790 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 441) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441)) 791 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 427) (("/") shift) ((":") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 792 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((":") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 793 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 794 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 795 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 796 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 438) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438)) 797 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 798 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 436) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 799 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 800 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 440) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440)) 801 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 804 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 805 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 806 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 807 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((":") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 808 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 809 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((":") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 810 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((":") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 811 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((":") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 812 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((":") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 813 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((":") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 814 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((":") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 820 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((":") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 821 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 822 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 823 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((":") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 824 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 827 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 828 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 830 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 832 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((":") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 833 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((":") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 836 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((":") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 837 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 448) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 838 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 411) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 839 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 412) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 840 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 417) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417)) 841 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 418) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418)) 845 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 446) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 846 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 447) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 847 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 439) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439)) 848 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 437) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437)) 849 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((":") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 850 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 420) (("/") shift) ((":") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420)) 851 ((("%") shift) (("*") shift) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) ((":") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 852 ((("%") shift) (("*") shift) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) ((":") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 853 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((":") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 854 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((":") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 855 ((("$") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) (("{") shift)) 856 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((":") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 857 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((":") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 858 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((":") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 859 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((":") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 860 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((":") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 861 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((":") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 862 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((":") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 863 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((":") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 864 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((":") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 865 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((":") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 866 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((":") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 867 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((":") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 868 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((":") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 869 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((":") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 870 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((":") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 871 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((":") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 872 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((":") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 873 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((":") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 874 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((":") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 875 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((":") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 876 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((":") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 877 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((":") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 878 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((":") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 879 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((":") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 880 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((":") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 881 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((":") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 882 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((":") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 883 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((":") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 884 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((":") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 885 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((":") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 886 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((":") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 887 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((":") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 888 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((":") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 889 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((":") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 890 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((":") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 891 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((":") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 892 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((":") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 893 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((":") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 894 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((":") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 895 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((":") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 896 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((":") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 897 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((":") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 898 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((":") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 899 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((":") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 900 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((":") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 901 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((":") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 902 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((":") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 903 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((":") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 904 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((":") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 905 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((":") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 906 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((":") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 907 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((":") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 908 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((":") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 909 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((":") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 910 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((":") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 911 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((":") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 912 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((":") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 913 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((":") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 914 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((":") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 915 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((":") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 916 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((":") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 917 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((":") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 918 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((":") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 919 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((":") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 920 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((":") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 921 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((":") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 922 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((":") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 923 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((":") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 924 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((":") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 925 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((":") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 926 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((":") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 927 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((":") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 928 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((":") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 929 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((":") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 930 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((":") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 931 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((":") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 932 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((":") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 933 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((":") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 934 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((":") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 935 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525)) 937 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((":") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 938 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((":") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 939 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((":") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 942 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((":") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527)) 943 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_AS) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_ELLIPSIS) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_USE) shift) ((T_VAR) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 944 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 949 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((":") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 953 (((":") reduce 74)) 954 ((("(") shift) ((":") reduce 52)) 956 (((":") reduce 27)) 958 (((":") reduce 51)) 959 (((":") reduce 53)) 960 (((":") reduce 49)) 961 (((":") reduce 29)) 962 (((":") reduce 59)) 963 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 60) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 964 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 11) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 965 (((":") reduce 42)) 966 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 967 (((":") reduce 39)) 969 (((":") reduce 25)) 970 (((":") reduce 50)) 971 ((("%") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 66) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 972 ((("%") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507)) 973 (((":") reduce 18)) 975 (((":") reduce 17)) 976 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 977 (((":") reduce 15)) 978 (((":") reduce 14)) 979 ((("(") shift) ((":") reduce 38)) 980 (((":") reduce 26)) 981 (((":") reduce 22)) 982 (((":") reduce 24)) 983 (((":") reduce 16)) 984 (((":") reduce 48)) 985 (((":") reduce 20)) 986 (((":") reduce 70)) 987 ((("(") shift) ((":") reduce 3)) 988 ((("%") reduce 500) (("(") shift) ((")") reduce 500) (("*") reduce 500) (("+") reduce 500) ((",") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) ((":") reduce 12) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 989 (((":") reduce 54)) 990 ((("%") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 65) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 991 (((":") reduce 75)) 992 (((":") reduce 30)) 993 ((("(") reduce 474) ((":") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474)) 994 (((":") reduce 21)) 995 (((":") reduce 23)) 996 ((("(") reduce 475) ((":") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475)) 997 ((("%") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 62) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 998 (((":") reduce 34)) 999 (((":") reduce 40)) 1000 (((":") reduce 13)) 1001 (((":") reduce 55)) 1003 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 1) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1004 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 2) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1005 (((":") reduce 9)) 1006 (((":") reduce 33)) 1007 (((":") reduce 58)) 1009 ((("(") shift) ((":") reduce 37)) 1010 ((("%") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 64) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 1011 ((("(") shift) ((":") reduce 46)) 1012 ((("%") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506)) 1013 (((":") reduce 8)) 1014 (((":") reduce 6)) 1015 (((":") reduce 7)) 1016 ((("(") shift) ((":") reduce 69)) 1017 ((("%") reduce 519) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 63) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 1018 (((":") reduce 56)) 1019 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 1020 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 1021 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 1022 ((("$") shift) (("(") shift) ((":") reduce 10) ((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CLASS) reduce 193) ((T_FINAL) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_VARIABLE) shift)) 1023 ((("%") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 67) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 1025 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 44) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1026 (((":") reduce 76)) 1027 ((("%") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 71) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 1028 (((":") reduce 77)) 1029 (((":") reduce 78)) 1030 ((("(") shift) ((":") reduce 79)) 1031 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 4) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1032 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 5) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1033 (((":") reduce 43)) 1035 (((":") reduce 73) ((T_FN) shift) ((T_FUNCTION) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492)) 1036 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 82) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 1038 (((":") reduce 47)) 1039 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 31) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1040 (((":") reduce 57)) 1041 ((("%") reduce 518) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 61) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 1042 (((":") reduce 28)) 1043 (((":") reduce 36)) 1045 (((":") reduce 32)) 1046 (((":") reduce 35)) 1047 ((("%") reduce 551) (("(") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 1048 (((":") reduce 19)) 1049 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) ((")") reduce 462) (("*") reduce 462) (("+") shift) ((",") reduce 462) (("-") shift) ((".") reduce 462) (("/") reduce 462) ((":") reduce 45) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 1053 (((")") reduce 292) ((",") reduce 292)) 1057 ((("%") reduce 547) (("(") reduce 538) ((")") reduce 547) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 1058 ((("%") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513)) 1060 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512)) 1061 ((("%") reduce 511) (("(") reduce 540) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511)) 1062 ((("%") shift) ((")") reduce 294) (("*") shift) (("+") shift) ((",") reduce 294) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1065 ((("%") reduce 546) (("(") reduce 546) ((")") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 1066 (((":") shift)) 1067 ((("%") reduce 467) ((")") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 1068 ((("%") reduce 449) ((")") reduce 449) (("*") reduce 449) (("+") reduce 449) ((",") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449)) 1069 ((("%") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 1070 ((("%") reduce 514) (("(") shift) ((")") reduce 514) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514)) 1071 ((("%") reduce 444) (("(") reduce 541) ((")") reduce 444) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444)) 1072 ((("%") reduce 445) ((")") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 1074 (((":") reduce 72)) 1075 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 1076 (((":") reduce 83)) 1077 ((("%") reduce 542) (("(") reduce 542) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 1078 ((("%") reduce 548) ((")") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 1079 ((("%") reduce 388) ((")") reduce 388) (("*") reduce 388) (("+") reduce 388) ((",") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388)) 1083 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 1085 ((("%") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 1088 ((("%") reduce 500) (("(") shift) ((")") reduce 500) (("*") reduce 500) (("+") reduce 500) ((",") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 1089 ((("%") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 1090 ((("%") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 1094 ((("%") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 1097 ((("%") reduce 519) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 1099 ((("%") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 1101 ((("%") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 1106 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 1108 ((("%") reduce 518) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 1109 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) ((")") reduce 462) (("*") reduce 462) (("+") shift) ((",") reduce 462) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 1110 ((("%") reduce 432) ((")") reduce 432) (("*") reduce 432) (("+") reduce 432) ((",") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432)) 1141 ((("%") shift) ((")") reduce 416) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416)) 1142 ((("%") shift) ((")") reduce 419) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419)) 1143 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428)) 1144 ((("%") shift) ((")") reduce 441) (("*") shift) (("+") shift) ((",") reduce 441) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441)) 1145 ((("%") shift) ((")") reduce 427) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 1146 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 1147 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1148 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1149 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1150 ((("%") shift) ((")") reduce 438) (("*") shift) (("+") shift) ((",") reduce 438) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438)) 1151 ((("%") shift) ((")") reduce 434) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 1152 ((("%") shift) ((")") reduce 436) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 1153 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 1154 ((("%") shift) ((")") reduce 440) (("*") shift) (("+") shift) ((",") reduce 440) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440)) 1155 ((("%") shift) ((")") reduce 435) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 1158 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 1159 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 1160 ((("%") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 1161 ((("%") reduce 492) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 1162 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 1163 ((("%") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 1164 ((("%") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 1165 ((("%") reduce 442) ((")") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 1166 ((("%") reduce 493) ((")") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 1167 ((("%") reduce 495) ((")") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 1168 ((("%") reduce 556) ((")") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 1174 ((("%") reduce 557) ((")") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 1175 ((("%") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 1176 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 1177 ((("%") reduce 558) ((")") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 1178 ((("%") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 1181 ((("%") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 1182 ((("%") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 1184 ((("%") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 1186 ((("%") reduce 496) ((")") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 1187 ((("%") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 1190 ((("%") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 1191 ((("%") shift) ((")") reduce 448) (("*") shift) (("+") shift) ((",") reduce 448) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1192 ((("%") shift) ((")") reduce 411) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1193 ((("%") shift) ((")") reduce 412) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1194 ((("%") shift) ((")") reduce 417) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417)) 1195 ((("%") shift) ((")") reduce 418) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418)) 1199 ((("%") shift) ((")") reduce 446) (("*") shift) (("+") shift) ((",") reduce 446) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1200 ((("%") shift) ((")") reduce 447) (("*") shift) (("+") shift) ((",") reduce 447) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1201 ((("%") shift) ((")") reduce 439) (("*") shift) (("+") shift) ((",") reduce 439) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439)) 1202 ((("%") shift) ((")") reduce 437) (("*") shift) (("+") shift) ((",") reduce 437) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437)) 1203 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 1204 ((("%") shift) ((")") reduce 420) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420)) 1205 ((("%") shift) ((")") reduce 422) (("*") shift) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 1206 ((("%") shift) ((")") reduce 421) (("*") shift) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 1207 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 1208 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 1219 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 1221 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 1223 ((("%") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 1224 ((("%") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507)) 1228 ((("%") reduce 500) (("(") shift) ((")") reduce 500) (("*") reduce 500) (("+") reduce 500) ((",") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_DOUBLE_ARROW) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 1229 ((("%") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 1230 ((("%") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 1236 ((("%") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 1238 ((("%") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506)) 1240 ((("%") reduce 519) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 1241 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 1242 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 1243 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 1245 ((("%") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 1248 ((("%") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 1254 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 1257 ((("%") reduce 518) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 1259 ((("%") reduce 551) (("(") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 1260 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) ((")") reduce 462) (("*") reduce 462) (("+") shift) ((",") reduce 462) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 462) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 1267 ((("%") reduce 547) (("(") reduce 538) ((")") reduce 547) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 1268 ((("%") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513)) 1270 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512)) 1271 ((("%") reduce 511) (("(") reduce 540) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511)) 1272 ((("%") shift) ((")") reduce 463) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1275 ((("%") reduce 546) (("(") reduce 546) ((")") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 1276 ((("%") reduce 467) ((")") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 1277 ((("%") reduce 449) ((")") reduce 449) (("*") reduce 449) (("+") reduce 449) ((",") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449)) 1278 ((("%") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 1279 ((("%") reduce 514) (("(") shift) ((")") reduce 514) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514)) 1280 ((("%") reduce 444) (("(") reduce 541) ((")") reduce 444) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444)) 1281 ((("%") reduce 445) ((")") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 1282 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 1283 ((("%") reduce 542) (("(") reduce 542) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 1284 ((("%") reduce 548) ((")") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 1285 ((("%") reduce 388) ((")") reduce 388) (("*") reduce 388) (("+") reduce 388) ((",") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 388) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388)) 1288 ((("%") reduce 432) ((")") reduce 432) (("*") reduce 432) (("+") reduce 432) ((",") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432)) 1319 ((("%") shift) ((")") reduce 416) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416)) 1320 ((("%") shift) ((")") reduce 419) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419)) 1321 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428)) 1322 ((("%") shift) ((")") reduce 441) (("*") shift) (("+") shift) ((",") reduce 441) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441)) 1323 ((("%") shift) ((")") reduce 427) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 1324 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 1325 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1326 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1327 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1328 ((("%") shift) ((")") reduce 438) (("*") shift) (("+") shift) ((",") reduce 438) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438)) 1329 ((("%") shift) ((")") reduce 434) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 1330 ((("%") shift) ((")") reduce 436) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 1331 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 1332 ((("%") shift) ((")") reduce 440) (("*") shift) (("+") shift) ((",") reduce 440) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440)) 1333 ((("%") shift) ((")") reduce 435) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 1336 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 1337 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 1338 ((("%") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 1339 ((("%") reduce 492) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 1340 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 1341 ((("%") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 1342 ((("%") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 1343 ((("%") reduce 442) ((")") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 1344 ((("%") reduce 493) ((")") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 1345 ((("%") reduce 495) ((")") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 1346 ((("%") reduce 556) ((")") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 1352 ((("%") reduce 557) ((")") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 1353 ((("%") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 1354 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 1355 ((("%") reduce 558) ((")") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 1356 ((("%") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 1359 ((("%") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 1360 ((("%") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 1362 ((("%") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 1364 ((("%") reduce 496) ((")") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 1365 ((("%") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 1368 ((("%") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 1369 ((("%") shift) ((")") reduce 448) (("*") shift) (("+") shift) ((",") reduce 448) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1370 ((("%") shift) ((")") reduce 411) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1371 ((("%") shift) ((")") reduce 412) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1372 ((("%") shift) ((")") reduce 417) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417)) 1373 ((("%") shift) ((")") reduce 418) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418)) 1377 ((("%") shift) ((")") reduce 446) (("*") shift) (("+") shift) ((",") reduce 446) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1378 ((("%") shift) ((")") reduce 447) (("*") shift) (("+") shift) ((",") reduce 447) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1379 ((("%") shift) ((")") reduce 439) (("*") shift) (("+") shift) ((",") reduce 439) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439)) 1380 ((("%") shift) ((")") reduce 437) (("*") shift) (("+") shift) ((",") reduce 437) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437)) 1381 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 1382 ((("%") shift) ((")") reduce 420) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420)) 1383 ((("%") shift) ((")") reduce 422) (("*") shift) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 1384 ((("%") shift) ((")") reduce 421) (("*") shift) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 1385 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 1386 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 1388 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 1389 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 1390 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 1391 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 1392 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 1393 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 1394 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 1395 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 1396 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 1397 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 1398 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 1399 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 1400 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 1401 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 1402 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 1403 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 1404 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 1405 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 1406 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 1407 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 1408 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 1409 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 1410 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 1411 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 1412 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 1413 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 1414 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 1415 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 1416 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 1417 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 1418 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 1419 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 1420 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 1421 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 1422 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 1423 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 1424 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 1425 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 1426 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 1427 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 1428 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 1429 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 1430 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 1431 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 1432 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 1433 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 1434 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 1435 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 1436 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 1437 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 1438 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 1439 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 1440 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 1441 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 1442 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 1443 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 1444 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 1445 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 1446 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 1447 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 1448 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 1449 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 1450 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 1451 ((("%") reduce 79) (("(") reduce 79) ((")") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 1452 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 1453 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 1454 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 1455 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 1456 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 1457 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 1458 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 1459 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 1460 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 1461 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 1462 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 1463 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 1464 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 1465 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 1466 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 1467 ((("%") reduce 525) (("(") reduce 562) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525)) 1469 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 1470 ((("%") reduce 83) (("(") reduce 83) ((")") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 1471 ((("%") reduce 555) (("(") reduce 564) ((")") reduce 555) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 1474 ((("%") reduce 527) (("(") reduce 563) ((")") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527)) 1476 ((("%") reduce 490) (("(") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 1477 ((("%") reduce 289) (("(") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 1480 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") reduce 125) (("+") shift) (("-") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_AS) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_ELLIPSIS) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_USE) shift) ((T_VAR) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1482 ((("%") reduce 290) (("(") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 1484 (((")") reduce 293) ((",") reduce 293)) 1485 ((("%") shift) ((")") reduce 296) (("*") shift) (("+") shift) ((",") reduce 296) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1486 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 1487 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1491 ((("%") reduce 409) ((")") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 1493 ((("%") reduce 407) ((")") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407)) 1503 ((("%") shift) ((")") reduce 403) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1504 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1505 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1506 ((("%") shift) ((")") reduce 397) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1507 ((("%") shift) ((")") reduce 394) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1508 ((("%") shift) ((")") reduce 402) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1509 ((("%") shift) ((")") reduce 396) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1510 ((("%") shift) ((")") reduce 400) (("*") shift) (("+") shift) ((",") reduce 400) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1511 ((("%") shift) ((")") reduce 395) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1512 ((("%") shift) ((")") reduce 398) (("*") shift) (("+") shift) ((",") reduce 398) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1513 ((("%") shift) ((")") reduce 399) (("*") shift) (("+") shift) ((",") reduce 399) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1514 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1515 ((("%") shift) ((")") reduce 401) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1516 ((("\"") reduce 80) (("$") reduce 80) (("(") reduce 80) ((T_ARRAY) reduce 80) ((T_CLASS_C) reduce 80) ((T_CONSTANT_ENCAPSED_STRING) reduce 80) ((T_DIR) reduce 80) ((T_FILE) reduce 80) ((T_FUNC_C) reduce 80) ((T_LINE) reduce 80) ((T_METHOD_C) reduce 80) ((T_NAME_FULLY_QUALIFIED) reduce 80) ((T_NAME_QUALIFIED) reduce 80) ((T_NAME_RELATIVE) reduce 80) ((T_NEW) reduce 80) ((T_NS_C) reduce 80) ((T_PROPERTY_C) reduce 80) ((T_READONLY) reduce 80) ((T_STATIC) reduce 80) ((T_STRING) reduce 80) ((T_TRAIT_C) reduce 80) ((T_VARIABLE) reduce 80) (("[") reduce 80)) 1517 ((("\"") reduce 81) (("$") reduce 81) (("(") reduce 81) ((T_ARRAY) reduce 81) ((T_CLASS_C) reduce 81) ((T_CONSTANT_ENCAPSED_STRING) reduce 81) ((T_DIR) reduce 81) ((T_FILE) reduce 81) ((T_FUNC_C) reduce 81) ((T_LINE) reduce 81) ((T_METHOD_C) reduce 81) ((T_NAME_FULLY_QUALIFIED) reduce 81) ((T_NAME_QUALIFIED) reduce 81) ((T_NAME_RELATIVE) reduce 81) ((T_NEW) reduce 81) ((T_NS_C) reduce 81) ((T_PROPERTY_C) reduce 81) ((T_READONLY) reduce 81) ((T_STATIC) reduce 81) ((T_STRING) reduce 81) ((T_TRAIT_C) reduce 81) ((T_VARIABLE) reduce 81) (("[") reduce 81)) 1519 ((("%") shift) ((")") reduce 391) (("*") shift) (("+") shift) ((",") reduce 391) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1524 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) (("[") reduce 523)) 1525 ((("(") reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) (("[") reduce 504)) 1526 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) (("[") reduce 517)) 1527 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) (("[") reduce 516)) 1528 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) (("[") reduce 520)) 1529 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) (("[") reduce 515)) 1530 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) (("[") reduce 519)) 1531 ((("(") reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) (("[") reduce 94)) 1532 ((("(") reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) (("[") reduce 93)) 1533 ((("(") reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) (("[") reduce 95)) 1535 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) (("[") reduce 522)) 1536 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) (("[") reduce 521)) 1538 (((T_PAAMAYIM_NEKUDOTAYIM) reduce 492)) 1539 ((("(") reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) (("[") reduce 92)) 1540 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) (("[") reduce 518)) 1541 ((("%") reduce 551) (("(") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 1545 ((("%") reduce 547) (("(") reduce 538) ((")") reduce 547) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 1546 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) (("[") reduce 534)) 1548 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) (("[") reduce 537)) 1549 ((("(") reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) (("[") reduce 533)) 1550 ((("%") reduce 546) (("(") reduce 546) ((")") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 1551 ((("(") shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) (("[") reduce 514)) 1552 ((("(") reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) (("[") reduce 535)) 1553 ((("%") reduce 542) (("(") reduce 542) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 1554 ((("%") reduce 548) ((")") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 1555 ((("%") reduce 392) ((")") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392)) 1558 ((("(") reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) (("[") reduce 74)) 1559 ((("(") reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) (("[") reduce 52)) 1560 ((("(") reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) (("[") reduce 27)) 1561 ((("(") reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) (("[") reduce 51)) 1562 ((("(") reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) (("[") reduce 53)) 1563 ((("(") reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) (("[") reduce 49)) 1564 ((("(") reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) (("[") reduce 29)) 1565 ((("(") reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) (("[") reduce 59)) 1566 ((("(") reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) (("[") reduce 60)) 1567 ((("(") reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) (("[") reduce 11)) 1568 ((("(") reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) (("[") reduce 42)) 1569 ((("(") reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) (("[") reduce 39)) 1570 ((("(") reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) (("[") reduce 25)) 1571 ((("(") reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) (("[") reduce 50)) 1572 ((("(") reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) (("[") reduce 66)) 1573 ((("(") reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) (("[") reduce 18)) 1574 ((("(") reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) (("[") reduce 17)) 1575 ((("(") reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) (("[") reduce 15)) 1576 ((("(") reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) (("[") reduce 14)) 1577 ((("(") reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) (("[") reduce 38)) 1578 ((("(") reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) (("[") reduce 26)) 1579 ((("(") reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) (("[") reduce 22)) 1580 ((("(") reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) (("[") reduce 24)) 1581 ((("(") reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) (("[") reduce 16)) 1582 ((("(") reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) (("[") reduce 48)) 1583 ((("(") reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) (("[") reduce 20)) 1584 ((("(") reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) (("[") reduce 70)) 1585 ((("(") reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) (("[") reduce 3)) 1586 ((("(") reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) (("[") reduce 12)) 1587 ((("(") reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) (("[") reduce 54)) 1588 ((("(") reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) (("[") reduce 65)) 1589 ((("(") reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) (("[") reduce 75)) 1590 ((("(") reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) (("[") reduce 30)) 1591 ((("(") reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) (("[") reduce 68)) 1592 ((("(") reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) (("[") reduce 21)) 1593 ((("(") reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) (("[") reduce 23)) 1594 ((("(") reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) (("[") reduce 41)) 1595 ((("(") reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) (("[") reduce 62)) 1596 ((("(") reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) (("[") reduce 34)) 1597 ((("(") reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) (("[") reduce 40)) 1598 ((("(") reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) (("[") reduce 13)) 1599 ((("(") reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) (("[") reduce 55)) 1600 ((("(") reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) (("[") reduce 1)) 1601 ((("(") reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) (("[") reduce 2)) 1602 ((("(") reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) (("[") reduce 9)) 1603 ((("(") reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) (("[") reduce 33)) 1604 ((("(") reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) (("[") reduce 58)) 1605 ((("(") reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) (("[") reduce 37)) 1606 ((("(") reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) (("[") reduce 64)) 1607 ((("(") reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) (("[") reduce 46)) 1608 ((("(") reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) (("[") reduce 8)) 1609 ((("(") reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) (("[") reduce 6)) 1610 ((("(") reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) (("[") reduce 7)) 1611 ((("(") reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) (("[") reduce 69)) 1612 ((("(") reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) (("[") reduce 63)) 1613 ((("(") reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) (("[") reduce 56)) 1614 ((("(") reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) (("[") reduce 10)) 1615 ((("(") reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) (("[") reduce 67)) 1616 ((("(") reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) (("[") reduce 44)) 1617 ((("(") reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) (("[") reduce 76)) 1618 ((("(") reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) (("[") reduce 71)) 1619 ((("(") reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) (("[") reduce 77)) 1620 ((("(") reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) (("[") reduce 78)) 1621 ((("(") reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) (("[") reduce 79)) 1622 ((("(") reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) (("[") reduce 4)) 1623 ((("(") reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) (("[") reduce 5)) 1624 ((("(") reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) (("[") reduce 43)) 1625 ((("(") reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) (("[") reduce 73)) 1626 ((("(") reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) (("[") reduce 82)) 1627 ((("(") reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) (("[") reduce 47)) 1628 ((("(") reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) (("[") reduce 31)) 1629 ((("(") reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) (("[") reduce 57)) 1630 ((("(") reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) (("[") reduce 61)) 1631 ((("(") reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) (("[") reduce 28)) 1632 ((("(") reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) (("[") reduce 36)) 1633 ((("(") reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) (("[") reduce 32)) 1634 ((("(") reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) (("[") reduce 35)) 1635 ((("(") reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) (("[") reduce 19)) 1636 ((("(") reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) (("[") reduce 45)) 1637 ((("(") reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) (("[") reduce 525)) 1639 ((("(") reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) (("[") reduce 72)) 1640 ((("(") reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) (("[") reduce 83)) 1641 ((("%") reduce 555) (("(") reduce 564) ((")") reduce 555) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 1644 ((("(") reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) (("[") reduce 527)) 1646 ((("%") reduce 490) (("(") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 1647 ((("%") reduce 289) (("(") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 1651 ((("%") reduce 290) (("(") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 1652 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 1653 ((("%") reduce 487) (("(") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 1655 ((("(") reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) (("[") reduce 524)) 1657 ((("%") reduce 554) (("(") reduce 564) ((")") reduce 554) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 1660 ((("(") reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) (("[") reduce 526)) 1661 ((("%") reduce 489) (("(") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 1662 ((("%") reduce 491) (("(") reduce 491) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 1667 ((("%") reduce 543) (("(") reduce 543) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 1668 ((("%") reduce 565) (("(") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 1669 ((("%") reduce 549) (("(") shift) ((")") reduce 549) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 1670 ((("%") reduce 567) (("(") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 1673 ((("%") reduce 566) (("(") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 1674 ((("%") reduce 544) (("(") reduce 544) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 1675 ((("%") reduce 550) (("(") shift) ((")") reduce 550) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 1676 ((("%") reduce 545) (("(") reduce 545) ((")") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 1687 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("]") reduce 523) (("^") reduce 523) (("|") reduce 523)) 1689 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("]") reduce 504) (("^") reduce 504) (("|") reduce 504)) 1691 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("|") reduce 517)) 1692 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("]") reduce 507) (("^") reduce 507) (("|") reduce 507)) 1697 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) ((",") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_DOUBLE_ARROW) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("]") reduce 500) (("^") reduce 500) (("|") reduce 500)) 1698 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("|") reduce 516)) 1699 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("|") reduce 520)) 1705 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("|") reduce 515)) 1707 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("]") reduce 506) (("^") reduce 506) (("|") reduce 506)) 1709 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("|") reduce 519)) 1710 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 1711 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 1712 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 1714 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("|") reduce 522)) 1717 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("|") reduce 521)) 1723 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 1726 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("]") reduce 518) (("^") reduce 518) (("|") reduce 518)) 1728 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("]") reduce 551) (("^") reduce 551) (("|") reduce 551)) 1729 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) ((",") reduce 462) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 462) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 462) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 1735 (((",") reduce 570) (("]") reduce 570)) 1739 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("]") reduce 547) (("^") reduce 547) (("|") reduce 547)) 1740 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("]") reduce 513) (("^") reduce 513) (("|") reduce 513)) 1742 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("]") reduce 512) (("^") reduce 512) (("|") reduce 512)) 1743 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("]") reduce 511) (("^") reduce 511) (("|") reduce 511)) 1744 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 574) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 574) (("^") shift) (("|") shift)) 1747 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("]") reduce 546) (("^") reduce 546) (("|") reduce 546)) 1748 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("]") reduce 467) (("^") reduce 467) (("|") reduce 467)) 1749 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) ((",") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("]") reduce 449) (("^") reduce 449) (("|") reduce 449)) 1750 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("]") reduce 471) (("^") reduce 471) (("|") reduce 471)) 1751 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("]") reduce 514) (("^") reduce 514) (("|") reduce 514)) 1752 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("]") reduce 444) (("^") reduce 444) (("|") reduce 444)) 1753 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("]") reduce 445) (("^") reduce 445) (("|") reduce 445)) 1754 (((",") shift) (("]") reduce 568)) 1755 (((",") reduce 572) (("]") reduce 572)) 1756 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 1757 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("]") reduce 542) (("^") reduce 542) (("|") reduce 542)) 1758 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("]") reduce 548) (("^") reduce 548) (("|") reduce 548)) 1759 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) ((",") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 388) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("]") reduce 388) (("^") reduce 388) (("|") reduce 388)) 1763 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) ((",") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("]") reduce 432) (("^") reduce 432) (("|") reduce 432)) 1794 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 416) (("^") shift) (("|") reduce 416)) 1795 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 419) (("^") reduce 419) (("|") reduce 419)) 1796 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("]") reduce 428) (("^") reduce 428) (("|") reduce 428)) 1797 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 441) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 441) (("^") reduce 441) (("|") reduce 441)) 1798 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("]") reduce 427) (("^") reduce 427) (("|") reduce 427)) 1799 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("]") reduce 424) (("^") reduce 424) (("|") reduce 424)) 1800 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 415) (("^") shift) (("|") shift)) 1801 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 413) (("^") shift) (("|") shift)) 1802 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 414) (("^") shift) (("|") shift)) 1803 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 438) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("]") reduce 438) (("^") reduce 438) (("|") reduce 438)) 1804 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 434) (("^") reduce 434) (("|") reduce 434)) 1805 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 436) (("^") reduce 436) (("|") reduce 436)) 1806 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 433) (("^") reduce 433) (("|") reduce 433)) 1807 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 440) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("]") reduce 440) (("^") reduce 440) (("|") reduce 440)) 1808 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 435) (("^") reduce 435) (("|") reduce 435)) 1811 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 1812 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 1813 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 1814 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("]") reduce 492) (("^") reduce 492) (("|") reduce 492)) 1815 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 1816 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("]") reduce 551) (("^") reduce 551) (("|") reduce 551)) 1817 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("]") reduce 494) (("^") reduce 494) (("|") reduce 494)) 1818 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("]") reduce 442) (("^") reduce 442) (("|") reduce 442)) 1819 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("]") reduce 493) (("^") reduce 493) (("|") reduce 493)) 1820 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("]") reduce 495) (("^") reduce 495) (("|") reduce 495)) 1821 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("]") reduce 556) (("^") reduce 556) (("|") reduce 556)) 1827 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("]") reduce 557) (("^") reduce 557) (("|") reduce 557)) 1828 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 1829 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 1830 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("]") reduce 558) (("^") reduce 558) (("|") reduce 558)) 1831 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 1834 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 1835 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 1837 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 1839 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("]") reduce 496) (("^") reduce 496) (("|") reduce 496)) 1840 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("]") reduce 553) (("^") reduce 553) (("|") reduce 553)) 1843 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("]") reduce 552) (("^") reduce 552) (("|") reduce 552)) 1844 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 448) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 448) (("^") shift) (("|") shift)) 1845 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 411) (("^") shift) (("|") shift)) 1846 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 412) (("^") shift) (("|") shift)) 1847 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 417) (("^") reduce 417) (("|") reduce 417)) 1848 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 418) (("^") reduce 418) (("|") reduce 418)) 1852 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 446) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 446) (("^") shift) (("|") shift)) 1853 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 447) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 447) (("^") shift) (("|") shift)) 1854 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 439) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("]") reduce 439) (("^") reduce 439) (("|") reduce 439)) 1855 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 437) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("]") reduce 437) (("^") reduce 437) (("|") reduce 437)) 1856 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("]") reduce 425) (("^") reduce 425) (("|") reduce 425)) 1857 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("]") reduce 420) (("^") reduce 420) (("|") reduce 420)) 1858 ((("%") shift) (("*") shift) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("]") reduce 422) (("^") reduce 422) (("|") reduce 422)) 1859 ((("%") shift) (("*") shift) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("]") reduce 421) (("^") reduce 421) (("|") reduce 421)) 1860 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("]") reduce 423) (("^") reduce 423) (("|") reduce 423)) 1861 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("]") reduce 426) (("^") reduce 426) (("|") reduce 426)) 1862 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") reduce 569) (("+") shift) ((",") reduce 569) (("-") shift) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_ELLIPSIS) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1921 (((")") reduce 570) ((",") reduce 570)) 1930 ((("%") shift) ((")") reduce 574) (("*") shift) (("+") shift) ((",") reduce 574) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1940 (((")") reduce 568) ((",") shift)) 1941 (((")") reduce 572) ((",") reduce 572)) 2050 ((("=") shift)) 2052 ((("%") shift) ((")") reduce 389) (("*") shift) (("+") shift) ((",") reduce 389) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2224 ((("%") reduce 488) (("(") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 2225 ((("%") reduce 553) (("(") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 2228 ((("%") reduce 552) (("(") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 2230 (((")") reduce 571) ((",") reduce 571)) 2231 ((("%") reduce 487) (("(") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 2242 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 2243 ((("%") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 2248 (((":") shift) ((T_DOUBLE_ARROW) reduce 287)) 2250 (((T_DOUBLE_ARROW) shift)) 2251 ((("!") reduce 477) (("\"") reduce 477) (("$") reduce 477) (("(") reduce 477) (("+") reduce 477) (("-") reduce 477) (("@") reduce 477) ((T_ARRAY) reduce 477) ((T_ARRAY_CAST) reduce 477) ((T_ATTRIBUTE) reduce 477) ((T_BOOL_CAST) reduce 477) ((T_CLASS_C) reduce 477) ((T_CLONE) reduce 477) ((T_CONSTANT_ENCAPSED_STRING) reduce 477) ((T_DEC) reduce 477) ((T_DIR) reduce 477) ((T_DNUMBER) reduce 477) ((T_DOUBLE_CAST) reduce 477) ((T_EMPTY) reduce 477) ((T_EVAL) reduce 477) ((T_EXIT) reduce 477) ((T_FILE) reduce 477) ((T_FN) reduce 477) ((T_FUNCTION) reduce 477) ((T_FUNC_C) reduce 477) ((T_INC) reduce 477) ((T_INCLUDE) reduce 477) ((T_INCLUDE_ONCE) reduce 477) ((T_INT_CAST) reduce 477) ((T_ISSET) reduce 477) ((T_LINE) reduce 477) ((T_LIST) reduce 477) ((T_LNUMBER) reduce 477) ((T_MATCH) reduce 477) ((T_METHOD_C) reduce 477) ((T_NAME_FULLY_QUALIFIED) reduce 477) ((T_NAME_QUALIFIED) reduce 477) ((T_NAME_RELATIVE) reduce 477) ((T_NEW) reduce 477) ((T_NS_C) reduce 477) ((T_OBJECT_CAST) reduce 477) ((T_PRINT) reduce 477) ((T_PROPERTY_C) reduce 477) ((T_READONLY) reduce 477) ((T_REQUIRE) reduce 477) ((T_REQUIRE_ONCE) reduce 477) ((T_START_HEREDOC) reduce 477) ((T_STATIC) reduce 477) ((T_STRING) reduce 477) ((T_STRING_CAST) reduce 477) ((T_THROW) reduce 477) ((T_TRAIT_C) reduce 477) ((T_UNSET_CAST) reduce 477) ((T_VARIABLE) reduce 477) ((T_YIELD) reduce 477) ((T_YIELD_FROM) reduce 477) (("[") reduce 477) (("`") reduce 477) (("~") reduce 477)) 2252 ((("!") reduce 478) (("\"") reduce 478) (("$") reduce 478) (("(") reduce 478) (("+") reduce 478) (("-") reduce 478) (("@") reduce 478) ((T_ARRAY) reduce 478) ((T_ARRAY_CAST) reduce 478) ((T_ATTRIBUTE) reduce 478) ((T_BOOL_CAST) reduce 478) ((T_CLASS_C) reduce 478) ((T_CLONE) reduce 478) ((T_CONSTANT_ENCAPSED_STRING) reduce 478) ((T_DEC) reduce 478) ((T_DIR) reduce 478) ((T_DNUMBER) reduce 478) ((T_DOUBLE_CAST) reduce 478) ((T_EMPTY) reduce 478) ((T_EVAL) reduce 478) ((T_EXIT) reduce 478) ((T_FILE) reduce 478) ((T_FN) reduce 478) ((T_FUNCTION) reduce 478) ((T_FUNC_C) reduce 478) ((T_INC) reduce 478) ((T_INCLUDE) reduce 478) ((T_INCLUDE_ONCE) reduce 478) ((T_INT_CAST) reduce 478) ((T_ISSET) reduce 478) ((T_LINE) reduce 478) ((T_LIST) reduce 478) ((T_LNUMBER) reduce 478) ((T_MATCH) reduce 478) ((T_METHOD_C) reduce 478) ((T_NAME_FULLY_QUALIFIED) reduce 478) ((T_NAME_QUALIFIED) reduce 478) ((T_NAME_RELATIVE) reduce 478) ((T_NEW) reduce 478) ((T_NS_C) reduce 478) ((T_OBJECT_CAST) reduce 478) ((T_PRINT) reduce 478) ((T_PROPERTY_C) reduce 478) ((T_READONLY) reduce 478) ((T_REQUIRE) reduce 478) ((T_REQUIRE_ONCE) reduce 478) ((T_START_HEREDOC) reduce 478) ((T_STATIC) reduce 478) ((T_STRING) reduce 478) ((T_STRING_CAST) reduce 478) ((T_THROW) reduce 478) ((T_TRAIT_C) reduce 478) ((T_UNSET_CAST) reduce 478) ((T_VARIABLE) reduce 478) ((T_YIELD) reduce 478) ((T_YIELD_FROM) reduce 478) (("[") reduce 478) (("`") reduce 478) (("~") reduce 478)) 2254 ((("%") shift) ((")") reduce 477) (("*") shift) (("+") shift) ((",") reduce 477) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2255 ((("%") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 2256 (((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift)) 2258 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_DOUBLE_ARROW) reduce 278) (("|") reduce 278)) 2259 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) ((T_DOUBLE_ARROW) reduce 279) (("|") reduce 279)) 2260 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_DOUBLE_ARROW) reduce 94) (("|") reduce 94)) 2261 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_DOUBLE_ARROW) reduce 93) (("|") reduce 93)) 2262 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_DOUBLE_ARROW) reduce 95) (("|") reduce 95)) 2263 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267) ((T_DOUBLE_ARROW) reduce 267) (("|") reduce 267)) 2264 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_DOUBLE_ARROW) reduce 92) (("|") reduce 92)) 2265 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_DOUBLE_ARROW) reduce 265)) 2266 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_DOUBLE_ARROW) reduce 280) (("|") reduce 280)) 2267 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_DOUBLE_ARROW) reduce 262) (("|") reduce 268)) 2268 (((T_DOUBLE_ARROW) reduce 288)) 2269 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266) ((T_DOUBLE_ARROW) reduce 266) (("|") reduce 266)) 2270 (((T_DOUBLE_ARROW) reduce 264) (("|") shift)) 2271 ((("|") shift)) 2272 ((("(") shift) ((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift)) 2274 (((T_DOUBLE_ARROW) reduce 278) (("|") reduce 278)) 2275 (((T_DOUBLE_ARROW) reduce 279) (("|") reduce 279)) 2276 (((T_DOUBLE_ARROW) reduce 94) (("|") reduce 94)) 2277 (((T_DOUBLE_ARROW) reduce 93) (("|") reduce 93)) 2278 (((T_DOUBLE_ARROW) reduce 95) (("|") reduce 95)) 2279 (((T_DOUBLE_ARROW) reduce 267) (("|") reduce 267)) 2280 (((T_DOUBLE_ARROW) reduce 92) (("|") reduce 92)) 2281 (((T_DOUBLE_ARROW) reduce 280) (("|") reduce 280)) 2282 (((T_DOUBLE_ARROW) reduce 268) (("|") reduce 268)) 2283 (((T_DOUBLE_ARROW) reduce 266) (("|") reduce 266)) 2284 (((T_DOUBLE_ARROW) reduce 270) (("|") reduce 270)) 2285 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278)) 2286 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279)) 2287 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94)) 2288 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93)) 2289 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95)) 2290 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267)) 2291 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92)) 2292 (((")") shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift)) 2293 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280)) 2294 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift)) 2295 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266)) 2297 (((")") reduce 278) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278)) 2298 (((")") reduce 279) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279)) 2299 (((")") reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94)) 2300 (((")") reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93)) 2301 (((")") reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95)) 2302 (((")") reduce 267) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267)) 2303 (((")") reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92)) 2304 (((")") reduce 280) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280)) 2305 (((")") reduce 272) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272)) 2306 (((")") reduce 266) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266)) 2307 (((T_DOUBLE_ARROW) reduce 269) (("|") reduce 269)) 2309 (((")") reduce 273) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273)) 2311 (((T_DOUBLE_ARROW) reduce 271) (("|") reduce 271)) 2313 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_DOUBLE_ARROW) reduce 278)) 2314 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) ((T_DOUBLE_ARROW) reduce 279)) 2315 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_DOUBLE_ARROW) reduce 94)) 2316 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_DOUBLE_ARROW) reduce 93)) 2317 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_DOUBLE_ARROW) reduce 95)) 2318 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267) ((T_DOUBLE_ARROW) reduce 267)) 2319 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_DOUBLE_ARROW) reduce 92)) 2320 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_DOUBLE_ARROW) reduce 280)) 2321 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) ((T_DOUBLE_ARROW) reduce 272)) 2322 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266) ((T_DOUBLE_ARROW) reduce 266)) 2324 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) ((T_DOUBLE_ARROW) reduce 273)) 2325 (((T_DOUBLE_ARROW) reduce 278)) 2326 (((T_DOUBLE_ARROW) reduce 279)) 2327 (((T_DOUBLE_ARROW) reduce 94)) 2328 (((T_DOUBLE_ARROW) reduce 93)) 2329 (((T_DOUBLE_ARROW) reduce 95)) 2330 (((T_DOUBLE_ARROW) reduce 267)) 2331 (((T_DOUBLE_ARROW) reduce 92)) 2332 (((T_DOUBLE_ARROW) reduce 280)) 2333 (((T_DOUBLE_ARROW) reduce 263)) 2334 (((T_DOUBLE_ARROW) reduce 266)) 2336 ((("|") reduce 269)) 2402 ((("%") shift) ((")") reduce 573) (("*") shift) (("+") shift) ((",") reduce 573) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2522 ((("%") shift) ((")") reduce 389) (("*") shift) (("+") shift) ((",") reduce 389) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2524 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 2525 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 2526 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 2527 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 2528 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 2529 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 2530 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 2531 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 2532 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 2533 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 2534 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 2535 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 2536 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 2537 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 2538 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 2539 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 2540 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 2541 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 2542 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 2543 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 2544 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 2545 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 2546 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 2547 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 2548 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 2549 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 2550 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 2551 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 2552 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 2553 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 2554 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 2555 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 2556 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 2557 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 2558 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 2559 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 2560 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 2561 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 2562 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 2563 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 2564 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 2565 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 2566 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 2567 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 2568 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 2569 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 2570 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 2571 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 2572 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 2573 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 2574 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 2575 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 2576 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 2577 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 2578 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 2579 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 2580 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 2581 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 2582 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 2583 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 2584 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 2585 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 2586 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 2587 ((("%") reduce 79) (("(") reduce 79) ((")") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 2588 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 2589 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 2590 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 2591 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 2592 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 2593 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 2594 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 2595 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 2596 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 2597 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 2598 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 2599 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 2600 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 2601 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 2602 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 2603 ((("%") reduce 525) (("(") reduce 562) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525)) 2605 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 2606 ((("%") reduce 83) (("(") reduce 83) ((")") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 2607 ((("%") reduce 555) (("(") reduce 564) ((")") reduce 555) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 2610 ((("%") reduce 527) (("(") reduce 563) ((")") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527)) 2612 ((("%") reduce 490) (("(") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 2613 ((("%") reduce 289) (("(") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 2617 ((("%") reduce 290) (("(") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 2618 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 2623 ((("%") reduce 409) ((")") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 2625 ((("%") reduce 407) ((")") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407)) 2635 ((("%") shift) ((")") reduce 403) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2636 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2637 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2638 ((("%") shift) ((")") reduce 397) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2639 ((("%") shift) ((")") reduce 394) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2640 ((("%") shift) ((")") reduce 402) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2641 ((("%") shift) ((")") reduce 396) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2642 ((("%") shift) ((")") reduce 400) (("*") shift) (("+") shift) ((",") reduce 400) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2643 ((("%") shift) ((")") reduce 395) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2644 ((("%") shift) ((")") reduce 398) (("*") shift) (("+") shift) ((",") reduce 398) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2645 ((("%") shift) ((")") reduce 399) (("*") shift) (("+") shift) ((",") reduce 399) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2646 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2647 ((("%") shift) ((")") reduce 401) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2649 ((("%") shift) ((")") reduce 391) (("*") shift) (("+") shift) ((",") reduce 391) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2652 ((("%") reduce 551) (("(") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 2655 ((("%") reduce 547) (("(") reduce 538) ((")") reduce 547) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 2657 ((("%") reduce 546) (("(") reduce 546) ((")") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 2659 ((("%") reduce 542) (("(") reduce 542) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 2660 ((("%") reduce 548) ((")") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 2661 ((("%") reduce 392) ((")") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392)) 2665 ((("%") reduce 555) (("(") reduce 564) ((")") reduce 555) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 2667 ((("%") reduce 490) (("(") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 2668 ((("%") reduce 289) (("(") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 2672 ((("%") reduce 290) (("(") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 2673 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 2674 ((("%") reduce 487) (("(") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 2677 ((("%") reduce 554) (("(") reduce 564) ((")") reduce 554) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 2678 ((("%") reduce 489) (("(") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 2679 ((("%") reduce 491) (("(") reduce 491) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 2684 ((("%") reduce 543) (("(") reduce 543) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 2685 ((("%") reduce 565) (("(") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 2686 ((("%") reduce 549) (("(") shift) ((")") reduce 549) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 2687 ((("%") reduce 567) (("(") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 2690 ((("%") reduce 566) (("(") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 2691 ((("%") reduce 544) (("(") reduce 544) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 2692 ((("%") reduce 550) (("(") shift) ((")") reduce 550) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 2693 ((("%") reduce 545) (("(") reduce 545) ((")") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 2694 ((("%") reduce 488) (("(") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 2695 ((("%") reduce 553) (("(") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 2698 ((("%") reduce 552) (("(") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 2699 ((("%") reduce 487) (("(") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 2710 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 2711 ((("%") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 2721 ((("%") shift) ((")") reduce 477) (("*") shift) (("+") shift) ((",") reduce 477) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2722 ((("%") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 2724 ((("%") reduce 524) (("(") reduce 562) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524)) 2726 ((("%") reduce 554) (("(") reduce 564) ((")") reduce 554) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 2729 ((("%") reduce 526) (("(") reduce 563) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526)) 2730 ((("%") reduce 489) (("(") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 2731 ((("%") reduce 491) (("(") reduce 491) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 2732 (((T_FN) shift) ((T_FUNCTION) shift)) 2733 (((T_ATTRIBUTE) reduce 102) ((T_FN) reduce 102) ((T_FUNCTION) reduce 102) ((T_STATIC) reduce 102)) 2734 ((("%") reduce 468) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 2735 ((("%") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 2740 ((("%") reduce 543) (("(") reduce 543) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 2741 ((("%") reduce 565) (("(") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 2742 ((("%") reduce 549) (("(") shift) ((")") reduce 549) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 2743 ((("%") reduce 567) (("(") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 2746 ((("%") reduce 566) (("(") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 2747 ((("%") reduce 544) (("(") reduce 544) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 2748 ((("%") reduce 550) (("(") shift) ((")") reduce 550) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 2749 ((("%") reduce 545) (("(") reduce 545) ((")") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 2752 ((("(") reduce 551) ((")") reduce 551) ((",") reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) (("[") reduce 551)) 2755 ((("(") reduce 538) ((")") reduce 547) ((",") reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) (("[") reduce 547)) 2757 ((("(") reduce 546) ((")") reduce 546) ((",") reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) (("[") reduce 546)) 2759 ((("(") reduce 542) ((")") reduce 542) ((",") reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) (("[") reduce 542)) 2760 (((")") reduce 548) ((",") reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) (("[") reduce 548)) 2761 (((")") reduce 575) ((",") reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) (("[") reduce 531)) 2765 ((("(") reduce 564) ((")") reduce 555) ((",") reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) (("[") reduce 555)) 2767 ((("(") reduce 490) ((")") reduce 490) ((",") reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) (("[") reduce 490)) 2768 ((("(") reduce 289) ((")") reduce 289) ((",") reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) (("[") reduce 289)) 2772 ((("(") reduce 290) ((")") reduce 290) ((",") reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) (("[") reduce 290)) 2773 ((("(") reduce 291) ((")") reduce 291) ((",") reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) (("[") reduce 291)) 2774 ((("(") reduce 487) ((")") reduce 487) ((",") reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) (("[") reduce 487)) 2777 ((("(") reduce 564) ((")") reduce 554) ((",") reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) (("[") reduce 554)) 2778 ((("(") reduce 489) ((")") reduce 489) ((",") reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) (("[") reduce 489)) 2779 ((("(") reduce 491) ((")") reduce 491) ((",") reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) (("[") reduce 491)) 2784 ((("(") reduce 543) ((")") reduce 543) ((",") reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) (("[") reduce 543)) 2785 ((("(") reduce 565) ((")") reduce 565) ((",") reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) (("[") reduce 565)) 2786 ((("(") shift) ((")") reduce 549) ((",") reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) (("[") reduce 549)) 2787 ((("(") reduce 567) ((")") reduce 567) ((",") reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) (("[") reduce 567)) 2790 ((("(") reduce 566) ((")") reduce 566) ((",") reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) (("[") reduce 566)) 2791 ((("(") reduce 544) ((")") reduce 544) ((",") reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) (("[") reduce 544)) 2792 ((("(") shift) ((")") reduce 550) ((",") reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) (("[") reduce 550)) 2793 ((("(") reduce 545) ((")") reduce 545) ((",") reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) (("[") reduce 545)) 2794 ((("(") reduce 488) ((")") reduce 488) ((",") reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) (("[") reduce 488)) 2795 ((("(") reduce 553) ((")") reduce 553) ((",") reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) (("[") reduce 553)) 2798 ((("(") reduce 552) ((")") reduce 552) ((",") reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) (("[") reduce 552)) 2800 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_STRING_VARNAME) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 2801 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_VARIABLE) shift) (("`") reduce 498)) 2802 (((T_CURLY_OPEN) reduce 584) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 584) ((T_ENCAPSED_AND_WHITESPACE) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_VARIABLE) reduce 584) (("[") shift) (("`") reduce 584)) 2803 ((("`") shift)) 2804 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_ENCAPSED_AND_WHITESPACE) shift) ((T_VARIABLE) shift) (("`") reduce 499)) 2805 (((T_CURLY_OPEN) reduce 582) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 582) ((T_ENCAPSED_AND_WHITESPACE) reduce 582) ((T_VARIABLE) reduce 582) (("`") reduce 582)) 2806 (((T_CURLY_OPEN) reduce 581) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 581) ((T_ENCAPSED_AND_WHITESPACE) reduce 581) ((T_VARIABLE) reduce 581) (("`") reduce 581)) 2807 (((T_CURLY_OPEN) reduce 580) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 580) ((T_ENCAPSED_AND_WHITESPACE) reduce 580) ((T_VARIABLE) reduce 580) (("`") reduce 580)) 2808 ((("%") reduce 460) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 2811 ((("-") shift) ((T_NUM_STRING) shift) ((T_STRING) shift) ((T_VARIABLE) shift)) 2812 (((T_NUM_STRING) shift)) 2813 ((("]") reduce 593)) 2814 ((("]") reduce 592)) 2815 ((("]") reduce 595)) 2817 (((T_CURLY_OPEN) reduce 585) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 585) ((T_ENCAPSED_AND_WHITESPACE) reduce 585) ((T_VARIABLE) reduce 585) (("`") reduce 585)) 2818 ((("]") reduce 594)) 2819 (((T_CURLY_OPEN) reduce 586) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 586) ((T_ENCAPSED_AND_WHITESPACE) reduce 586) ((T_VARIABLE) reduce 586) (("`") reduce 586)) 2820 (((T_CURLY_OPEN) reduce 587) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 587) ((T_ENCAPSED_AND_WHITESPACE) reduce 587) ((T_VARIABLE) reduce 587) (("`") reduce 587)) 2821 (((T_CURLY_OPEN) reduce 583) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 583) ((T_ENCAPSED_AND_WHITESPACE) reduce 583) ((T_VARIABLE) reduce 583) (("`") reduce 583)) 2822 ((("[") shift) (("}") shift)) 2824 (((T_CURLY_OPEN) reduce 588) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 588) ((T_ENCAPSED_AND_WHITESPACE) reduce 588) ((T_VARIABLE) reduce 588) (("`") reduce 588)) 2826 (((T_CURLY_OPEN) reduce 589) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 589) ((T_ENCAPSED_AND_WHITESPACE) reduce 589) ((T_VARIABLE) reduce 589) (("`") reduce 589)) 2827 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") shift) (("^") shift) (("|") shift)) 2828 ((("}") shift)) 2829 (((T_CURLY_OPEN) reduce 590) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 590) ((T_ENCAPSED_AND_WHITESPACE) reduce 590) ((T_VARIABLE) reduce 590) (("`") reduce 590)) 2832 ((("(") reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) (("[") reduce 551) (("}") reduce 551)) 2835 ((("(") reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) (("[") reduce 547) (("}") reduce 547)) 2837 ((("(") reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) (("[") reduce 546) (("}") reduce 546)) 2839 ((("(") reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) (("[") reduce 542) (("}") reduce 542)) 2840 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) (("[") reduce 548) (("}") reduce 548)) 2841 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) (("[") reduce 531) (("}") shift)) 2845 ((("(") reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) (("[") reduce 555) (("}") reduce 555)) 2847 ((("(") reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) (("[") reduce 490) (("}") reduce 490)) 2848 ((("(") reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) (("[") reduce 289) (("}") reduce 289)) 2852 ((("(") reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) (("[") reduce 290) (("}") reduce 290)) 2853 ((("(") reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) (("[") reduce 291) (("}") reduce 291)) 2854 (((T_CURLY_OPEN) reduce 591) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 591) ((T_ENCAPSED_AND_WHITESPACE) reduce 591) ((T_VARIABLE) reduce 591) (("`") reduce 591)) 2855 ((("(") reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) (("[") reduce 487) (("}") reduce 487)) 2858 ((("(") reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) (("[") reduce 554) (("}") reduce 554)) 2859 ((("(") reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) (("[") reduce 489) (("}") reduce 489)) 2860 ((("(") reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) (("[") reduce 491) (("}") reduce 491)) 2865 ((("(") reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) (("[") reduce 543) (("}") reduce 543)) 2866 ((("(") reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) (("[") reduce 565) (("}") reduce 565)) 2867 ((("(") shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) (("[") reduce 549) (("}") reduce 549)) 2868 ((("(") reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) (("[") reduce 567) (("}") reduce 567)) 2871 ((("(") reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) (("[") reduce 566) (("}") reduce 566)) 2872 ((("(") reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) (("[") reduce 544) (("}") reduce 544)) 2873 ((("(") shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) (("[") reduce 550) (("}") reduce 550)) 2874 ((("(") reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) (("[") reduce 545) (("}") reduce 545)) 2875 ((("(") reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) (("[") reduce 488) (("}") reduce 488)) 2876 ((("(") reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) (("[") reduce 553) (("}") reduce 553)) 2879 ((("(") reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) (("[") reduce 552) (("}") reduce 552)) 2881 ((("%") reduce 503) (("(") reduce 503) ((")") reduce 503) (("*") reduce 503) (("+") reduce 503) ((",") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503)) 2883 ((("%") shift) ((")") reduce 390) (("*") shift) (("+") shift) ((",") reduce 390) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2884 ((("%") shift) ((")") reduce 465) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2887 ((("%") shift) ((")") reduce 464) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2888 ((("%") reduce 456) ((")") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456)) 2889 ((("%") shift) ((")") reduce 466) (("*") shift) (("+") shift) ((",") reduce 466) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2890 ((("%") reduce 452) ((")") reduce 452) (("*") reduce 452) (("+") reduce 452) ((",") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452)) 2891 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 2894 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_END_HEREDOC) shift) ((T_VARIABLE) shift)) 2895 ((("%") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 2896 (((T_CURLY_OPEN) reduce 584) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 584) ((T_ENCAPSED_AND_WHITESPACE) reduce 584) ((T_END_HEREDOC) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_VARIABLE) reduce 584) (("[") shift)) 2898 (((T_CURLY_OPEN) reduce 582) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 582) ((T_ENCAPSED_AND_WHITESPACE) reduce 582) ((T_END_HEREDOC) reduce 582) ((T_VARIABLE) reduce 582)) 2899 (((T_CURLY_OPEN) reduce 581) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 581) ((T_ENCAPSED_AND_WHITESPACE) reduce 581) ((T_END_HEREDOC) reduce 581) ((T_VARIABLE) reduce 581)) 2900 ((("%") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 2901 (((T_CURLY_OPEN) reduce 580) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 580) ((T_ENCAPSED_AND_WHITESPACE) reduce 580) ((T_END_HEREDOC) reduce 580) ((T_VARIABLE) reduce 580)) 2906 (((T_CURLY_OPEN) reduce 585) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 585) ((T_ENCAPSED_AND_WHITESPACE) reduce 585) ((T_END_HEREDOC) reduce 585) ((T_VARIABLE) reduce 585)) 2907 (((T_CURLY_OPEN) reduce 586) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 586) ((T_ENCAPSED_AND_WHITESPACE) reduce 586) ((T_END_HEREDOC) reduce 586) ((T_VARIABLE) reduce 586)) 2908 (((T_CURLY_OPEN) reduce 587) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 587) ((T_ENCAPSED_AND_WHITESPACE) reduce 587) ((T_END_HEREDOC) reduce 587) ((T_VARIABLE) reduce 587)) 2909 ((("%") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508)) 2910 (((T_CURLY_OPEN) reduce 583) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 583) ((T_ENCAPSED_AND_WHITESPACE) reduce 583) ((T_END_HEREDOC) reduce 583) ((T_VARIABLE) reduce 583)) 2913 (((T_CURLY_OPEN) reduce 588) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 588) ((T_ENCAPSED_AND_WHITESPACE) reduce 588) ((T_END_HEREDOC) reduce 588) ((T_VARIABLE) reduce 588)) 2915 (((T_CURLY_OPEN) reduce 589) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 589) ((T_ENCAPSED_AND_WHITESPACE) reduce 589) ((T_END_HEREDOC) reduce 589) ((T_VARIABLE) reduce 589)) 2918 (((T_CURLY_OPEN) reduce 590) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 590) ((T_ENCAPSED_AND_WHITESPACE) reduce 590) ((T_END_HEREDOC) reduce 590) ((T_VARIABLE) reduce 590)) 2920 (((T_CURLY_OPEN) reduce 591) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 591) ((T_ENCAPSED_AND_WHITESPACE) reduce 591) ((T_END_HEREDOC) reduce 591) ((T_VARIABLE) reduce 591)) 2921 ((("%") shift) ((")") reduce 602) (("*") shift) (("+") shift) ((",") reduce 602) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2922 ((("%") shift) ((")") reduce 601) (("*") shift) (("+") shift) ((",") reduce 601) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2923 ((("%") reduce 488) (("(") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 2924 ((("%") shift) ((")") reduce 461) (("*") shift) (("+") shift) ((",") reduce 461) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2925 ((("%") reduce 454) ((")") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454)) 2928 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 2929 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 2930 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 2931 (((T_ABSTRACT) reduce 197) ((T_CLASS) reduce 197) ((T_FINAL) reduce 197) ((T_READONLY) reduce 197)) 2932 ((("%") reduce 492) (("(") reduce 492) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 2933 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 2934 ((("%") reduce 385) (("(") reduce 385) ((")") reduce 385) (("*") reduce 385) (("+") reduce 385) ((",") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385)) 2935 (((T_ABSTRACT) shift) ((T_CLASS) reduce 194) ((T_FINAL) shift) ((T_READONLY) shift)) 2936 (((T_CLASS) shift)) 2937 (((T_ABSTRACT) reduce 101) ((T_ATTRIBUTE) reduce 101) ((T_CLASS) reduce 101) ((T_FINAL) reduce 101) ((T_READONLY) reduce 101)) 2938 (((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CLASS) reduce 193) ((T_FINAL) shift) ((T_READONLY) shift)) 2939 (((T_ABSTRACT) reduce 191) ((T_CLASS) reduce 191) ((T_FINAL) reduce 191) ((T_READONLY) reduce 191)) 2940 ((("%") reduce 494) (("(") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 2941 ((("%") reduce 387) (("(") shift) ((")") reduce 387) (("*") reduce 387) (("+") reduce 387) ((",") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387)) 2942 ((("%") reduce 493) (("(") reduce 493) ((")") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 2943 ((("%") reduce 495) (("(") reduce 495) ((")") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 2944 ((("%") reduce 556) (("(") reduce 556) ((")") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 2950 ((("%") reduce 557) (("(") reduce 557) ((")") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 2951 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 2952 ((("%") reduce 558) (("(") reduce 558) ((")") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 2953 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 2954 ((("%") reduce 384) (("(") reduce 384) ((")") reduce 384) (("*") reduce 384) (("+") reduce 384) ((",") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384)) 2956 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 2957 ((("%") reduce 386) (("(") reduce 386) ((")") reduce 386) (("*") reduce 386) (("+") reduce 386) ((",") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386)) 2958 (((T_ABSTRACT) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_CLASS) reduce 102) ((T_FINAL) reduce 102) ((T_READONLY) reduce 102)) 2959 ((("(") shift) ((T_EXTENDS) reduce 500) ((T_IMPLEMENTS) reduce 500) (("{") reduce 500)) 2961 (((T_EXTENDS) reduce 501) ((T_IMPLEMENTS) reduce 501) (("{") reduce 501)) 2962 (((T_EXTENDS) shift) ((T_IMPLEMENTS) reduce 206) (("{") reduce 206)) 2964 (((T_IMPLEMENTS) shift) (("{") reduce 210)) 2966 ((("{") reduce 476)) 2968 (((T_ABSTRACT) reduce 305) ((T_ATTRIBUTE) reduce 305) ((T_CASE) reduce 305) ((T_CONST) reduce 305) ((T_FINAL) reduce 305) ((T_FUNCTION) reduce 305) ((T_PRIVATE) reduce 305) ((T_PRIVATE_SET) reduce 305) ((T_PROTECTED) reduce 305) ((T_PROTECTED_SET) reduce 305) ((T_PUBLIC) reduce 305) ((T_PUBLIC_SET) reduce 305) ((T_READONLY) reduce 305) ((T_STATIC) reduce 305) ((T_USE) reduce 305) ((T_VAR) reduce 305) (("}") reduce 305)) 2969 (((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CASE) shift) ((T_CONST) reduce 338) ((T_FINAL) shift) ((T_FUNCTION) reduce 336) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_USE) shift) ((T_VAR) shift) (("}") shift)) 2970 ((("(") reduce 349) (("?") reduce 349) ((T_ABSTRACT) reduce 349) ((T_ARRAY) reduce 349) ((T_CALLABLE) reduce 349) ((T_CONST) reduce 349) ((T_FINAL) reduce 349) ((T_FUNCTION) reduce 349) ((T_NAME_FULLY_QUALIFIED) reduce 349) ((T_NAME_QUALIFIED) reduce 349) ((T_NAME_RELATIVE) reduce 349) ((T_PRIVATE) reduce 349) ((T_PRIVATE_SET) reduce 349) ((T_PROTECTED) reduce 349) ((T_PROTECTED_SET) reduce 349) ((T_PUBLIC) reduce 349) ((T_PUBLIC_SET) reduce 349) ((T_READONLY) reduce 349) ((T_STATIC) reduce 349) ((T_STRING) reduce 349) ((T_VARIABLE) reduce 349)) 2972 (((T_ABSTRACT) reduce 476) ((T_ARRAY) reduce 476) ((T_AS) reduce 476) ((T_BREAK) reduce 476) ((T_CALLABLE) reduce 476) ((T_CASE) reduce 476) ((T_CATCH) reduce 476) ((T_CLASS) reduce 476) ((T_CLASS_C) reduce 476) ((T_CLONE) reduce 476) ((T_CONST) reduce 476) ((T_CONTINUE) reduce 476) ((T_DECLARE) reduce 476) ((T_DEFAULT) reduce 476) ((T_DIR) reduce 476) ((T_DO) reduce 476) ((T_ECHO) reduce 476) ((T_ELSE) reduce 476) ((T_ELSEIF) reduce 476) ((T_EMPTY) reduce 476) ((T_ENDDECLARE) reduce 476) ((T_ENDFOR) reduce 476) ((T_ENDFOREACH) reduce 476) ((T_ENDIF) reduce 476) ((T_ENDSWITCH) reduce 476) ((T_ENDWHILE) reduce 476) ((T_ENUM) reduce 476) ((T_EVAL) reduce 476) ((T_EXIT) reduce 476) ((T_EXTENDS) reduce 476) ((T_FILE) reduce 476) ((T_FINAL) reduce 476) ((T_FINALLY) reduce 476) ((T_FN) reduce 476) ((T_FOR) reduce 476) ((T_FOREACH) reduce 476) ((T_FUNCTION) reduce 476) ((T_FUNC_C) reduce 476) ((T_GLOBAL) reduce 476) ((T_GOTO) reduce 476) ((T_IF) reduce 476) ((T_IMPLEMENTS) reduce 476) ((T_INCLUDE) reduce 476) ((T_INCLUDE_ONCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_INSTEADOF) reduce 476) ((T_INTERFACE) reduce 476) ((T_ISSET) reduce 476) ((T_LINE) reduce 476) ((T_LIST) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_MATCH) reduce 476) ((T_METHOD_C) reduce 476) ((T_NAMESPACE) reduce 476) ((T_NEW) reduce 476) ((T_NS_C) reduce 476) ((T_PRINT) reduce 476) ((T_PRIVATE) reduce 476) ((T_PROPERTY_C) reduce 476) ((T_PROTECTED) reduce 476) ((T_PUBLIC) reduce 476) ((T_READONLY) reduce 476) ((T_REQUIRE) reduce 476) ((T_REQUIRE_ONCE) reduce 476) ((T_RETURN) reduce 476) ((T_STATIC) reduce 476) ((T_STRING) reduce 476) ((T_SWITCH) reduce 476) ((T_THROW) reduce 476) ((T_TRAIT) reduce 476) ((T_TRAIT_C) reduce 476) ((T_TRY) reduce 476) ((T_UNSET) reduce 476) ((T_USE) reduce 476) ((T_VAR) reduce 476) ((T_WHILE) reduce 476) ((T_YIELD) reduce 476)) 2973 ((("(") reduce 350) (("?") reduce 350) ((T_ABSTRACT) reduce 350) ((T_ARRAY) reduce 350) ((T_CALLABLE) reduce 350) ((T_CONST) reduce 350) ((T_FINAL) reduce 350) ((T_FUNCTION) reduce 350) ((T_NAME_FULLY_QUALIFIED) reduce 350) ((T_NAME_QUALIFIED) reduce 350) ((T_NAME_RELATIVE) reduce 350) ((T_PRIVATE) reduce 350) ((T_PRIVATE_SET) reduce 350) ((T_PROTECTED) reduce 350) ((T_PROTECTED_SET) reduce 350) ((T_PUBLIC) reduce 350) ((T_PUBLIC_SET) reduce 350) ((T_READONLY) reduce 350) ((T_STATIC) reduce 350) ((T_STRING) reduce 350) ((T_VARIABLE) reduce 350)) 2974 ((("(") reduce 344) (("?") reduce 344) ((T_ABSTRACT) reduce 344) ((T_ARRAY) reduce 344) ((T_CALLABLE) reduce 344) ((T_CONST) reduce 344) ((T_FINAL) reduce 344) ((T_FUNCTION) reduce 344) ((T_NAME_FULLY_QUALIFIED) reduce 344) ((T_NAME_QUALIFIED) reduce 344) ((T_NAME_RELATIVE) reduce 344) ((T_PRIVATE) reduce 344) ((T_PRIVATE_SET) reduce 344) ((T_PROTECTED) reduce 344) ((T_PROTECTED_SET) reduce 344) ((T_PUBLIC) reduce 344) ((T_PUBLIC_SET) reduce 344) ((T_READONLY) reduce 344) ((T_STATIC) reduce 344) ((T_STRING) reduce 344) ((T_VARIABLE) reduce 344)) 2975 ((("(") reduce 347) (("?") reduce 347) ((T_ABSTRACT) reduce 347) ((T_ARRAY) reduce 347) ((T_CALLABLE) reduce 347) ((T_CONST) reduce 347) ((T_FINAL) reduce 347) ((T_FUNCTION) reduce 347) ((T_NAME_FULLY_QUALIFIED) reduce 347) ((T_NAME_QUALIFIED) reduce 347) ((T_NAME_RELATIVE) reduce 347) ((T_PRIVATE) reduce 347) ((T_PRIVATE_SET) reduce 347) ((T_PROTECTED) reduce 347) ((T_PROTECTED_SET) reduce 347) ((T_PUBLIC) reduce 347) ((T_PUBLIC_SET) reduce 347) ((T_READONLY) reduce 347) ((T_STATIC) reduce 347) ((T_STRING) reduce 347) ((T_VARIABLE) reduce 347)) 2976 ((("(") reduce 343) (("?") reduce 343) ((T_ABSTRACT) reduce 343) ((T_ARRAY) reduce 343) ((T_CALLABLE) reduce 343) ((T_CONST) reduce 343) ((T_FINAL) reduce 343) ((T_FUNCTION) reduce 343) ((T_NAME_FULLY_QUALIFIED) reduce 343) ((T_NAME_QUALIFIED) reduce 343) ((T_NAME_RELATIVE) reduce 343) ((T_PRIVATE) reduce 343) ((T_PRIVATE_SET) reduce 343) ((T_PROTECTED) reduce 343) ((T_PROTECTED_SET) reduce 343) ((T_PUBLIC) reduce 343) ((T_PUBLIC_SET) reduce 343) ((T_READONLY) reduce 343) ((T_STATIC) reduce 343) ((T_STRING) reduce 343) ((T_VARIABLE) reduce 343)) 2977 ((("(") reduce 346) (("?") reduce 346) ((T_ABSTRACT) reduce 346) ((T_ARRAY) reduce 346) ((T_CALLABLE) reduce 346) ((T_CONST) reduce 346) ((T_FINAL) reduce 346) ((T_FUNCTION) reduce 346) ((T_NAME_FULLY_QUALIFIED) reduce 346) ((T_NAME_QUALIFIED) reduce 346) ((T_NAME_RELATIVE) reduce 346) ((T_PRIVATE) reduce 346) ((T_PRIVATE_SET) reduce 346) ((T_PROTECTED) reduce 346) ((T_PROTECTED_SET) reduce 346) ((T_PUBLIC) reduce 346) ((T_PUBLIC_SET) reduce 346) ((T_READONLY) reduce 346) ((T_STATIC) reduce 346) ((T_STRING) reduce 346) ((T_VARIABLE) reduce 346)) 2978 ((("(") reduce 342) (("?") reduce 342) ((T_ABSTRACT) reduce 342) ((T_ARRAY) reduce 342) ((T_CALLABLE) reduce 342) ((T_CONST) reduce 342) ((T_FINAL) reduce 342) ((T_FUNCTION) reduce 342) ((T_NAME_FULLY_QUALIFIED) reduce 342) ((T_NAME_QUALIFIED) reduce 342) ((T_NAME_RELATIVE) reduce 342) ((T_PRIVATE) reduce 342) ((T_PRIVATE_SET) reduce 342) ((T_PROTECTED) reduce 342) ((T_PROTECTED_SET) reduce 342) ((T_PUBLIC) reduce 342) ((T_PUBLIC_SET) reduce 342) ((T_READONLY) reduce 342) ((T_STATIC) reduce 342) ((T_STRING) reduce 342) ((T_VARIABLE) reduce 342)) 2979 ((("(") reduce 345) (("?") reduce 345) ((T_ABSTRACT) reduce 345) ((T_ARRAY) reduce 345) ((T_CALLABLE) reduce 345) ((T_CONST) reduce 345) ((T_FINAL) reduce 345) ((T_FUNCTION) reduce 345) ((T_NAME_FULLY_QUALIFIED) reduce 345) ((T_NAME_QUALIFIED) reduce 345) ((T_NAME_RELATIVE) reduce 345) ((T_PRIVATE) reduce 345) ((T_PRIVATE_SET) reduce 345) ((T_PROTECTED) reduce 345) ((T_PROTECTED_SET) reduce 345) ((T_PUBLIC) reduce 345) ((T_PUBLIC_SET) reduce 345) ((T_READONLY) reduce 345) ((T_STATIC) reduce 345) ((T_STRING) reduce 345) ((T_VARIABLE) reduce 345)) 2980 ((("(") reduce 351) (("?") reduce 351) ((T_ABSTRACT) reduce 351) ((T_ARRAY) reduce 351) ((T_CALLABLE) reduce 351) ((T_CONST) reduce 351) ((T_FINAL) reduce 351) ((T_FUNCTION) reduce 351) ((T_NAME_FULLY_QUALIFIED) reduce 351) ((T_NAME_QUALIFIED) reduce 351) ((T_NAME_RELATIVE) reduce 351) ((T_PRIVATE) reduce 351) ((T_PRIVATE_SET) reduce 351) ((T_PROTECTED) reduce 351) ((T_PROTECTED_SET) reduce 351) ((T_PUBLIC) reduce 351) ((T_PUBLIC_SET) reduce 351) ((T_READONLY) reduce 351) ((T_STATIC) reduce 351) ((T_STRING) reduce 351) ((T_VARIABLE) reduce 351)) 2981 ((("(") reduce 348) (("?") reduce 348) ((T_ABSTRACT) reduce 348) ((T_ARRAY) reduce 348) ((T_CALLABLE) reduce 348) ((T_CONST) reduce 348) ((T_FINAL) reduce 348) ((T_FUNCTION) reduce 348) ((T_NAME_FULLY_QUALIFIED) reduce 348) ((T_NAME_QUALIFIED) reduce 348) ((T_NAME_RELATIVE) reduce 348) ((T_PRIVATE) reduce 348) ((T_PRIVATE_SET) reduce 348) ((T_PROTECTED) reduce 348) ((T_PROTECTED_SET) reduce 348) ((T_PUBLIC) reduce 348) ((T_PUBLIC_SET) reduce 348) ((T_READONLY) reduce 348) ((T_STATIC) reduce 348) ((T_STRING) reduce 348) ((T_VARIABLE) reduce 348)) 2983 ((("(") reduce 335) (("?") reduce 335) ((T_ARRAY) reduce 335) ((T_CALLABLE) reduce 335) ((T_NAME_FULLY_QUALIFIED) reduce 335) ((T_NAME_QUALIFIED) reduce 335) ((T_NAME_RELATIVE) reduce 335) ((T_STRING) reduce 335) ((T_VARIABLE) reduce 335)) 2984 (((T_ABSTRACT) reduce 101) ((T_ATTRIBUTE) reduce 101) ((T_CASE) reduce 101) ((T_CONST) reduce 101) ((T_FINAL) reduce 101) ((T_FUNCTION) reduce 101) ((T_PRIVATE) reduce 101) ((T_PRIVATE_SET) reduce 101) ((T_PROTECTED) reduce 101) ((T_PROTECTED_SET) reduce 101) ((T_PUBLIC) reduce 101) ((T_PUBLIC_SET) reduce 101) ((T_READONLY) reduce 101) ((T_STATIC) reduce 101) ((T_VAR) reduce 101)) 2985 (((T_ABSTRACT) reduce 312) ((T_ATTRIBUTE) reduce 312) ((T_CASE) reduce 312) ((T_CONST) reduce 312) ((T_FINAL) reduce 312) ((T_FUNCTION) reduce 312) ((T_PRIVATE) reduce 312) ((T_PRIVATE_SET) reduce 312) ((T_PROTECTED) reduce 312) ((T_PROTECTED_SET) reduce 312) ((T_PUBLIC) reduce 312) ((T_PUBLIC_SET) reduce 312) ((T_READONLY) reduce 312) ((T_STATIC) reduce 312) ((T_USE) reduce 312) ((T_VAR) reduce 312) (("}") reduce 312)) 2986 (((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CASE) shift) ((T_CONST) reduce 338) ((T_FINAL) shift) ((T_FUNCTION) reduce 336) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_VAR) shift)) 2987 (((T_CONST) shift)) 2988 (((T_ABSTRACT) reduce 304) ((T_ATTRIBUTE) reduce 304) ((T_CASE) reduce 304) ((T_CONST) reduce 304) ((T_FINAL) reduce 304) ((T_FUNCTION) reduce 304) ((T_PRIVATE) reduce 304) ((T_PRIVATE_SET) reduce 304) ((T_PROTECTED) reduce 304) ((T_PROTECTED_SET) reduce 304) ((T_PUBLIC) reduce 304) ((T_PUBLIC_SET) reduce 304) ((T_READONLY) reduce 304) ((T_STATIC) reduce 304) ((T_USE) reduce 304) ((T_VAR) reduce 304) (("}") reduce 304)) 2989 (((T_ABSTRACT) reduce 311) ((T_ATTRIBUTE) reduce 311) ((T_CASE) reduce 311) ((T_CONST) reduce 311) ((T_FINAL) reduce 311) ((T_FUNCTION) reduce 311) ((T_PRIVATE) reduce 311) ((T_PRIVATE_SET) reduce 311) ((T_PROTECTED) reduce 311) ((T_PROTECTED_SET) reduce 311) ((T_PUBLIC) reduce 311) ((T_PUBLIC_SET) reduce 311) ((T_READONLY) reduce 311) ((T_STATIC) reduce 311) ((T_USE) reduce 311) ((T_VAR) reduce 311) (("}") reduce 311)) 2990 ((("(") reduce 340) (("?") reduce 340) ((T_ABSTRACT) reduce 340) ((T_ARRAY) reduce 340) ((T_CALLABLE) reduce 340) ((T_CONST) reduce 340) ((T_FINAL) reduce 340) ((T_FUNCTION) reduce 340) ((T_NAME_FULLY_QUALIFIED) reduce 340) ((T_NAME_QUALIFIED) reduce 340) ((T_NAME_RELATIVE) reduce 340) ((T_PRIVATE) reduce 340) ((T_PRIVATE_SET) reduce 340) ((T_PROTECTED) reduce 340) ((T_PROTECTED_SET) reduce 340) ((T_PUBLIC) reduce 340) ((T_PUBLIC_SET) reduce 340) ((T_READONLY) reduce 340) ((T_STATIC) reduce 340) ((T_STRING) reduce 340) ((T_VARIABLE) reduce 340)) 2991 (((T_FUNCTION) shift)) 2992 ((("(") reduce 334) (("?") reduce 334) ((T_ABSTRACT) shift) ((T_ARRAY) reduce 334) ((T_CALLABLE) reduce 334) ((T_CONST) reduce 339) ((T_FINAL) shift) ((T_FUNCTION) reduce 337) ((T_NAME_FULLY_QUALIFIED) reduce 334) ((T_NAME_QUALIFIED) reduce 334) ((T_NAME_RELATIVE) reduce 334) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 334) ((T_VARIABLE) reduce 334)) 2993 ((("(") shift) (("?") shift) ((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STRING) shift) ((T_VARIABLE) reduce 260)) 2994 ((("%") reduce 383) (("(") reduce 383) ((")") reduce 383) (("*") reduce 383) (("+") reduce 383) ((",") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383)) 2995 (((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STRING) shift)) 2997 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_VARIABLE) reduce 278) (("|") reduce 278)) 2998 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) ((T_VARIABLE) reduce 279) (("|") reduce 279)) 2999 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_VARIABLE) reduce 94) (("|") reduce 94)) 3000 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_VARIABLE) reduce 93) (("|") reduce 93)) 3001 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_VARIABLE) reduce 95) (("|") reduce 95)) 3002 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_VARIABLE) reduce 92) (("|") reduce 92)) 3003 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_VARIABLE) reduce 277)) 3004 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_VARIABLE) reduce 280) (("|") reduce 280)) 3005 (((T_VARIABLE) shift)) 3006 (((T_VARIABLE) reduce 261)) 3007 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_VARIABLE) reduce 274) (("|") reduce 281)) 3008 (((T_VARIABLE) reduce 276) (("|") shift)) 3010 ((("(") shift) ((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STRING) shift)) 3012 (((T_VARIABLE) reduce 278) (("|") reduce 278)) 3013 (((T_VARIABLE) reduce 279) (("|") reduce 279)) 3014 (((T_VARIABLE) reduce 94) (("|") reduce 94)) 3015 (((T_VARIABLE) reduce 93) (("|") reduce 93)) 3016 (((T_VARIABLE) reduce 95) (("|") reduce 95)) 3017 (((T_VARIABLE) reduce 92) (("|") reduce 92)) 3018 (((T_VARIABLE) reduce 280) (("|") reduce 280)) 3019 (((T_VARIABLE) reduce 281) (("|") reduce 281)) 3020 (((T_VARIABLE) reduce 283) (("|") reduce 283)) 3024 (((")") reduce 285) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 285)) 3025 (((T_VARIABLE) reduce 282) (("|") reduce 282)) 3027 (((")") reduce 286) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 286)) 3029 (((T_VARIABLE) reduce 284) (("|") reduce 284)) 3031 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_VARIABLE) reduce 278)) 3032 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) ((T_VARIABLE) reduce 279)) 3033 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_VARIABLE) reduce 94)) 3034 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_VARIABLE) reduce 93)) 3035 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_VARIABLE) reduce 95)) 3036 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_VARIABLE) reduce 92)) 3037 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_VARIABLE) reduce 280)) 3038 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 285) ((T_VARIABLE) reduce 285)) 3039 (((",") reduce 476) ((";") reduce 476) (("=") shift) (("{") reduce 476)) 3040 (((T_ABSTRACT) reduce 307) ((T_ATTRIBUTE) reduce 307) ((T_CASE) reduce 307) ((T_CONST) reduce 307) ((T_FINAL) reduce 307) ((T_FUNCTION) reduce 307) ((T_PRIVATE) reduce 307) ((T_PRIVATE_SET) reduce 307) ((T_PROTECTED) reduce 307) ((T_PROTECTED_SET) reduce 307) ((T_PUBLIC) reduce 307) ((T_PUBLIC_SET) reduce 307) ((T_READONLY) reduce 307) ((T_STATIC) reduce 307) ((T_USE) reduce 307) ((T_VAR) reduce 307) (("}") reduce 307)) 3041 (((",") reduce 353) ((";") reduce 353)) 3042 (((",") shift) ((";") shift)) 3044 (((T_ABSTRACT) reduce 306) ((T_ATTRIBUTE) reduce 306) ((T_CASE) reduce 306) ((T_CONST) reduce 306) ((T_FINAL) reduce 306) ((T_FUNCTION) reduce 306) ((T_PRIVATE) reduce 306) ((T_PRIVATE_SET) reduce 306) ((T_PROTECTED) reduce 306) ((T_PROTECTED_SET) reduce 306) ((T_PUBLIC) reduce 306) ((T_PUBLIC_SET) reduce 306) ((T_READONLY) reduce 306) ((T_STATIC) reduce 306) ((T_USE) reduce 306) ((T_VAR) reduce 306) (("}") reduce 306)) 3045 (((",") reduce 476) ((";") reduce 476) (("=") shift)) 3046 (((",") reduce 352) ((";") reduce 352)) 3048 (((",") reduce 354) ((";") reduce 354)) 3059 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 3061 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 3063 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 3064 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507)) 3068 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) ((",") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) ((";") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 3069 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 3070 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 3076 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 3078 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506)) 3080 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 3081 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 3082 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 3083 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 3085 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 3088 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 3094 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 3097 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 3099 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 3100 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) ((",") reduce 462) (("-") shift) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 3107 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((";") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 3108 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513)) 3110 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512)) 3111 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511)) 3112 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 476) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 476) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3115 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((";") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 3116 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 3117 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) ((",") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) ((";") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449)) 3118 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 3119 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((";") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514)) 3120 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444)) 3121 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 3122 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 3123 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 3124 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((";") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 3125 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) ((",") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) ((";") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388)) 3128 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) ((",") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) ((";") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432)) 3159 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 416) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416)) 3160 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 419) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419)) 3161 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") reduce 428) (("/") shift) ((";") reduce 428) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428)) 3162 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 441) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 441) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441)) 3163 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") reduce 427) (("/") shift) ((";") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 3164 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 3165 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3166 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3167 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3168 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 438) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 438) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438)) 3169 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 3170 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 436) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 3171 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 3172 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 440) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 440) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440)) 3173 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 3176 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 3177 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 3178 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 3179 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 3180 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 3181 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 3182 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 3183 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 3184 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 3185 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 3186 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 3192 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 3193 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 3194 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 3195 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 3196 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 3199 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 3200 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 3202 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 3204 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 3205 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 3208 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 3209 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 448) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 448) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3210 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 411) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3211 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 412) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3212 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 417) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417)) 3213 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 418) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418)) 3217 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 446) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 446) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3218 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 447) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 447) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3219 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 439) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 439) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439)) 3220 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 437) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 437) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437)) 3221 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 3222 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") reduce 420) (("/") shift) ((";") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420)) 3223 ((("%") shift) (("*") shift) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) ((";") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 3224 ((("%") shift) (("*") shift) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) ((";") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 3225 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 3226 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 3228 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 3229 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 3230 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 3231 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 3232 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 3233 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 3234 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 3235 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 3236 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 3237 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 3238 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 3239 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 3240 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 3241 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 3242 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 3243 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 3244 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 3245 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 3246 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 3247 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 3248 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 3249 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 3250 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 3251 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 3252 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 3253 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 3254 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 3255 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 3256 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 3257 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 3258 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 3259 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 3260 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 3261 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 3262 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 3263 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 3264 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 3265 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 3266 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 3267 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 3268 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 3269 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 3270 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 3271 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 3272 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 3273 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 3274 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 3275 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 3276 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 3277 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 3278 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 3279 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 3280 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 3281 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 3282 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 3283 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 3284 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 3285 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 3286 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 3287 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 3288 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 3289 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 3290 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 3291 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 3292 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 3293 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 3294 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 3295 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 3296 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 3297 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 3298 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 3299 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 3300 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 3301 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 3302 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 3303 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 3304 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 3305 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 3306 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 3307 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525)) 3309 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 3310 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 3311 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((";") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 3314 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527)) 3316 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 3317 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 3321 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 3322 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 3327 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 3329 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407)) 3339 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 403) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3340 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3341 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3342 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 397) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3343 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 394) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3344 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 402) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3345 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 396) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3346 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 400) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 400) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3347 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 395) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3348 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 398) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 398) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3349 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 399) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 399) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3350 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3351 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 401) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3353 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 391) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 391) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3356 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 3359 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((";") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 3361 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((";") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 3363 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 3364 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((";") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 3365 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((";") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392)) 3369 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((";") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 3371 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 3372 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 3376 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 3377 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 3378 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 3381 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((";") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 3382 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 3383 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 3388 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 3389 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 3390 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((";") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 3391 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 3394 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 3395 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 3396 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((";") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 3397 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((";") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 3398 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 3399 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 3402 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 3403 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 3414 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 3415 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 3425 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 477) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 477) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3426 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 3427 (((",") reduce 355) ((";") reduce 355)) 3429 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524)) 3431 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((";") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 3434 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526)) 3435 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 3436 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 3438 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 3439 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 3444 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 3445 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 3446 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((";") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 3447 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 3450 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 3451 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 3452 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((";") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 3453 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((";") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 3455 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 3457 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) ((",") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) ((";") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503)) 3459 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 390) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 390) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3460 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 465) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3471 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 3473 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 3475 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 3476 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507)) 3480 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) ((",") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) ((";") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_DOUBLE_ARROW) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 3481 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 3482 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 3488 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 3490 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506)) 3492 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 3493 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 3494 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 3495 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 3497 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 3500 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 3506 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 3509 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 3511 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 3512 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) ((",") reduce 462) (("-") shift) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 462) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 3519 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((";") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 3520 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513)) 3522 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512)) 3523 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511)) 3524 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 463) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3527 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((";") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 3528 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 3529 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) ((",") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) ((";") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449)) 3530 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 3531 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((";") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514)) 3532 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444)) 3533 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 3534 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 3535 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 3536 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((";") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 3537 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) ((",") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) ((";") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 388) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388)) 3540 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) ((",") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) ((";") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432)) 3571 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 416) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416)) 3572 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 419) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419)) 3573 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") reduce 428) (("/") shift) ((";") reduce 428) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428)) 3574 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 441) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 441) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441)) 3575 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") reduce 427) (("/") shift) ((";") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 3576 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 3577 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3578 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3579 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3580 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 438) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 438) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438)) 3581 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 3582 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 436) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 3583 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 3584 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 440) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 440) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440)) 3585 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 3588 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 3589 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 3590 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 3591 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 3592 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 3593 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 3594 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 3595 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 3596 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 3597 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 3598 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 3604 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 3605 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 3606 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 3607 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 3608 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 3611 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 3612 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 3614 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 3616 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 3617 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 3620 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 3621 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 448) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 448) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3622 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 411) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3623 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 412) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3624 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 417) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417)) 3625 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 418) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418)) 3629 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 446) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 446) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3630 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 447) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 447) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3631 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 439) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 439) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439)) 3632 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 437) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 437) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437)) 3633 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 3634 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") reduce 420) (("/") shift) ((";") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420)) 3635 ((("%") shift) (("*") shift) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) ((";") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 3636 ((("%") shift) (("*") shift) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) ((";") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 3637 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 3638 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 3640 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 3641 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 3642 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 3643 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 3644 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 3645 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 3646 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 3647 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 3648 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 3649 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 3650 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 3651 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 3652 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 3653 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 3654 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 3655 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 3656 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 3657 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 3658 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 3659 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 3660 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 3661 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 3662 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 3663 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 3664 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 3665 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 3666 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 3667 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 3668 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 3669 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 3670 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 3671 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 3672 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 3673 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 3674 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 3675 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 3676 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 3677 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 3678 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 3679 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 3680 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 3681 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 3682 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 3683 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 3684 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 3685 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 3686 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 3687 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 3688 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 3689 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 3690 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 3691 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 3692 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 3693 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 3694 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 3695 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 3696 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 3697 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 3698 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 3699 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 3700 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 3701 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 3702 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 3703 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 3704 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 3705 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 3706 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 3707 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 3708 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 3709 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 3710 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 3711 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 3712 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 3713 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 3714 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 3715 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 3716 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 3717 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 3718 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 3719 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525)) 3721 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 3722 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 3723 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((";") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 3726 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527)) 3728 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 3729 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 3733 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 3734 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 3739 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 3741 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407)) 3751 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 403) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3752 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3753 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3754 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 397) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3755 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 394) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3756 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 402) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3757 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 396) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3758 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 400) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 400) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3759 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 395) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3760 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 398) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 398) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3761 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 399) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 399) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3762 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3763 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 401) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3765 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 391) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 391) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3768 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 3771 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((";") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 3773 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((";") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 3775 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 3776 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((";") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 3777 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((";") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392)) 3781 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((";") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 3783 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 3784 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 3788 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 3789 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 3790 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 3793 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((";") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 3794 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 3795 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 3800 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 3801 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 3802 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((";") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 3803 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 3806 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 3807 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 3808 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((";") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 3809 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((";") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 3810 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 3811 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 3814 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 3815 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 3826 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 3827 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 3837 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 477) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 477) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3838 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 3840 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 464) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3842 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524)) 3844 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((";") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 3847 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526)) 3848 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 3849 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 3851 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 3852 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 3857 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 3858 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 3859 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((";") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 3860 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 3863 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 3864 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 3865 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((";") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 3866 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((";") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 3868 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 3870 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) ((",") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) ((";") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_DOUBLE_ARROW) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503)) 3872 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 390) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 390) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3873 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 465) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3876 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 464) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3877 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((";") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456)) 3878 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 466) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 466) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3879 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) ((",") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) ((";") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452)) 3880 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 3882 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 3884 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 3885 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508)) 3886 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 602) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 602) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 602) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3887 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 601) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 601) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 601) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3888 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 3889 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 461) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 461) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3890 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((";") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454)) 3892 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 3893 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 3894 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 3895 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 3896 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 3897 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) ((",") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) ((";") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385)) 3900 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 3901 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) ((",") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) ((";") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387)) 3902 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 3903 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 3904 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 3910 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 3911 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 3912 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 3913 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 3914 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) ((",") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) ((";") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384)) 3916 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 3917 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) ((",") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) ((";") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386)) 3925 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) ((",") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) ((";") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383)) 3927 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 3931 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("}") reduce 232) (("~") shift)) 3942 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 3944 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 3946 (((",") shift) ((T_DOUBLE_ARROW) reduce 124)) 3947 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 3948 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507)) 3952 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) ((",") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_DOUBLE_ARROW) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 3953 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 3954 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 3960 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 3962 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506)) 3964 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 3965 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 3966 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 3967 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 3969 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 3972 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 3978 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 3981 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 3983 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 3984 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) ((",") reduce 462) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 462) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 3991 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 3992 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513)) 3994 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512)) 3995 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511)) 3996 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 238) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 238) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3999 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 4000 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 4001 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) ((",") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449)) 4002 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 4003 (((",") reduce 234) (("}") reduce 234)) 4006 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514)) 4007 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444)) 4008 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 4009 (((",") shift) (("}") reduce 124)) 4010 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 4011 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 4012 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 4013 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) ((",") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 388) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388)) 4016 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) ((",") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432)) 4047 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416)) 4048 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419)) 4049 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428)) 4050 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 441) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441)) 4051 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 4052 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 4053 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4054 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4055 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4056 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 438) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438)) 4057 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 4058 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 4059 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 4060 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 440) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440)) 4061 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 4064 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 4065 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 4066 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 4067 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 4068 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 4069 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 4070 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 4071 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 4072 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 4073 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 4074 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 4080 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 4081 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 4082 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 4083 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 4084 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 4087 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 4088 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 4090 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 4092 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 4093 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 4096 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 4097 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 448) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4098 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4099 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4100 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417)) 4101 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418)) 4105 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 446) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4106 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 447) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4107 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 439) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439)) 4108 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 437) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437)) 4109 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 4110 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420)) 4111 ((("%") shift) (("*") shift) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 4112 ((("%") shift) (("*") shift) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 4113 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 4114 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 4116 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 4117 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 4118 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 4119 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 4120 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 4121 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 4122 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 4123 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 4124 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 4125 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 4126 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 4127 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 4128 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 4129 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 4130 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 4131 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 4132 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 4133 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 4134 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 4135 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 4136 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 4137 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 4138 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 4139 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 4140 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 4141 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 4142 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 4143 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 4144 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 4145 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 4146 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 4147 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 4148 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 4149 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 4150 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 4151 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 4152 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 4153 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 4154 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 4155 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 4156 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 4157 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 4158 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 4159 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 4160 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 4161 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 4162 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 4163 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 4164 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 4165 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 4166 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 4167 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 4168 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 4169 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 4170 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 4171 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 4172 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 4173 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 4174 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 4175 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 4176 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 4177 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 4178 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 4179 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 4180 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 4181 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 4182 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 4183 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 4184 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 4185 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 4186 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 4187 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 4188 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 4189 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 4190 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 4191 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 4192 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 4193 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 4194 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 4195 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525)) 4197 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 4198 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 4199 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 4202 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527)) 4204 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 4205 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 4209 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 4210 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 4215 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 4217 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407)) 4227 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4228 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4229 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4230 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4231 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4232 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4233 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4234 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 400) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4235 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4236 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 398) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4237 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 399) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4238 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4239 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4241 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 391) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4244 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 4247 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 4249 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 4251 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 4252 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 4253 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392)) 4257 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 4259 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 4260 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 4264 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 4265 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 4266 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 4269 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 4270 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 4271 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 4276 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 4277 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 4278 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 4279 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 4282 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 4283 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 4284 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 4285 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 4286 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 4287 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 4290 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 4291 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("}") reduce 125) (("~") shift)) 4292 ((("}") reduce 233)) 4293 (((",") reduce 235) (("}") reduce 235)) 4294 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 4295 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) ((",") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) ((";") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_DOUBLE_ARROW) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231)) 4296 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 125) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 4309 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523) (("}") reduce 523)) 4311 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504) (("}") reduce 504)) 4313 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517) (("}") reduce 517)) 4314 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507) (("}") reduce 507)) 4318 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) ((",") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500) (("}") reduce 500)) 4319 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516) (("}") reduce 516)) 4320 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520) (("}") reduce 520)) 4326 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515) (("}") reduce 515)) 4328 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506) (("}") reduce 506)) 4330 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519) (("}") reduce 519)) 4331 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 4332 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 4333 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 4335 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522) (("}") reduce 522)) 4338 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521) (("}") reduce 521)) 4344 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 4347 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518) (("}") reduce 518)) 4349 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551) (("}") reduce 551)) 4350 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) ((",") reduce 462) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("}") reduce 462) (("~") shift)) 4357 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547) (("}") reduce 547)) 4358 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513) (("}") reduce 513)) 4360 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512) (("}") reduce 512)) 4361 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511) (("}") reduce 511)) 4362 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 236) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 236)) 4365 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546) (("}") reduce 546)) 4366 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467) (("}") reduce 467)) 4367 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) ((",") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449) (("}") reduce 449)) 4368 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471) (("}") reduce 471)) 4369 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514) (("}") reduce 514)) 4370 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444) (("}") reduce 444)) 4371 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445) (("}") reduce 445)) 4372 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 4373 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542) (("}") reduce 542)) 4374 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548) (("}") reduce 548)) 4375 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) ((",") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388) (("}") reduce 388)) 4378 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) ((",") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432) (("}") reduce 432)) 4409 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416) (("}") reduce 416)) 4410 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419) (("}") reduce 419)) 4411 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428) (("}") reduce 428)) 4412 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 441) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441) (("}") reduce 441)) 4413 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427) (("}") reduce 427)) 4414 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424) (("}") reduce 424)) 4415 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 415)) 4416 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 413)) 4417 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 414)) 4418 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 438) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438) (("}") reduce 438)) 4419 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434) (("}") reduce 434)) 4420 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436) (("}") reduce 436)) 4421 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433) (("}") reduce 433)) 4422 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 440) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440) (("}") reduce 440)) 4423 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435) (("}") reduce 435)) 4426 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 4427 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 4428 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 4429 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492) (("}") reduce 492)) 4430 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 4431 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551) (("}") reduce 551)) 4432 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494) (("}") reduce 494)) 4433 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442) (("}") reduce 442)) 4434 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493) (("}") reduce 493)) 4435 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495) (("}") reduce 495)) 4436 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556) (("}") reduce 556)) 4442 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557) (("}") reduce 557)) 4443 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 4444 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 4445 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558) (("}") reduce 558)) 4446 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 4449 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 4450 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 4452 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 4454 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496) (("}") reduce 496)) 4455 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553) (("}") reduce 553)) 4458 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552) (("}") reduce 552)) 4459 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 448) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 448)) 4460 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 411)) 4461 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 412)) 4462 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417) (("}") reduce 417)) 4463 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418) (("}") reduce 418)) 4467 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 446) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 446)) 4468 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 447) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 447)) 4469 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 439) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439) (("}") reduce 439)) 4470 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 437) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437) (("}") reduce 437)) 4471 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425) (("}") reduce 425)) 4472 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420) (("}") reduce 420)) 4473 ((("%") shift) (("*") shift) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422) (("}") reduce 422)) 4474 ((("%") shift) (("*") shift) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421) (("}") reduce 421)) 4475 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423) (("}") reduce 423)) 4476 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426) (("}") reduce 426)) 4478 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74) (("}") reduce 74)) 4479 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52) (("}") reduce 52)) 4480 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27) (("}") reduce 27)) 4481 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51) (("}") reduce 51)) 4482 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53) (("}") reduce 53)) 4483 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49) (("}") reduce 49)) 4484 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29) (("}") reduce 29)) 4485 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59) (("}") reduce 59)) 4486 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60) (("}") reduce 60)) 4487 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11) (("}") reduce 11)) 4488 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42) (("}") reduce 42)) 4489 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39) (("}") reduce 39)) 4490 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25) (("}") reduce 25)) 4491 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50) (("}") reduce 50)) 4492 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66) (("}") reduce 66)) 4493 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18) (("}") reduce 18)) 4494 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17) (("}") reduce 17)) 4495 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15) (("}") reduce 15)) 4496 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14) (("}") reduce 14)) 4497 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38) (("}") reduce 38)) 4498 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26) (("}") reduce 26)) 4499 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22) (("}") reduce 22)) 4500 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24) (("}") reduce 24)) 4501 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16) (("}") reduce 16)) 4502 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48) (("}") reduce 48)) 4503 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20) (("}") reduce 20)) 4504 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70) (("}") reduce 70)) 4505 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3) (("}") reduce 3)) 4506 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12) (("}") reduce 12)) 4507 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54) (("}") reduce 54)) 4508 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65) (("}") reduce 65)) 4509 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75) (("}") reduce 75)) 4510 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30) (("}") reduce 30)) 4511 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68) (("}") reduce 68)) 4512 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21) (("}") reduce 21)) 4513 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23) (("}") reduce 23)) 4514 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41) (("}") reduce 41)) 4515 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62) (("}") reduce 62)) 4516 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34) (("}") reduce 34)) 4517 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40) (("}") reduce 40)) 4518 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13) (("}") reduce 13)) 4519 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55) (("}") reduce 55)) 4520 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1) (("}") reduce 1)) 4521 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2) (("}") reduce 2)) 4522 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9) (("}") reduce 9)) 4523 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33) (("}") reduce 33)) 4524 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58) (("}") reduce 58)) 4525 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37) (("}") reduce 37)) 4526 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64) (("}") reduce 64)) 4527 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46) (("}") reduce 46)) 4528 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8) (("}") reduce 8)) 4529 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6) (("}") reduce 6)) 4530 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7) (("}") reduce 7)) 4531 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69) (("}") reduce 69)) 4532 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63) (("}") reduce 63)) 4533 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56) (("}") reduce 56)) 4534 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10) (("}") reduce 10)) 4535 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67) (("}") reduce 67)) 4536 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44) (("}") reduce 44)) 4537 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76) (("}") reduce 76)) 4538 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71) (("}") reduce 71)) 4539 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77) (("}") reduce 77)) 4540 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78) (("}") reduce 78)) 4541 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79) (("}") reduce 79)) 4542 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4) (("}") reduce 4)) 4543 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5) (("}") reduce 5)) 4544 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43) (("}") reduce 43)) 4545 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73) (("}") reduce 73)) 4546 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82) (("}") reduce 82)) 4547 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47) (("}") reduce 47)) 4548 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31) (("}") reduce 31)) 4549 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57) (("}") reduce 57)) 4550 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61) (("}") reduce 61)) 4551 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28) (("}") reduce 28)) 4552 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36) (("}") reduce 36)) 4553 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32) (("}") reduce 32)) 4554 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35) (("}") reduce 35)) 4555 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19) (("}") reduce 19)) 4556 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45) (("}") reduce 45)) 4557 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525) (("}") reduce 525)) 4559 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72) (("}") reduce 72)) 4560 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83) (("}") reduce 83)) 4561 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555) (("}") reduce 555)) 4564 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527) (("}") reduce 527)) 4566 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490) (("}") reduce 490)) 4567 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289) (("}") reduce 289)) 4571 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290) (("}") reduce 290)) 4572 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 4577 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409) (("}") reduce 409)) 4579 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407) (("}") reduce 407)) 4589 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 403)) 4590 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 405)) 4591 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 404)) 4592 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 397)) 4593 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 394)) 4594 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 402)) 4595 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 396)) 4596 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 400) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 400)) 4597 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 395)) 4598 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 398) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 398)) 4599 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 399) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 399)) 4600 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 406)) 4601 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 401)) 4603 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 391) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 391)) 4606 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551) (("}") reduce 551)) 4609 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547) (("}") reduce 547)) 4611 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546) (("}") reduce 546)) 4613 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542) (("}") reduce 542)) 4614 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548) (("}") reduce 548)) 4615 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392) (("}") reduce 392)) 4619 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555) (("}") reduce 555)) 4621 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490) (("}") reduce 490)) 4622 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289) (("}") reduce 289)) 4626 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290) (("}") reduce 290)) 4627 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 4628 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487) (("}") reduce 487)) 4631 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554) (("}") reduce 554)) 4632 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489) (("}") reduce 489)) 4633 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491) (("}") reduce 491)) 4638 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543) (("}") reduce 543)) 4639 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 4640 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549) (("}") reduce 549)) 4641 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 4644 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 4645 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544) (("}") reduce 544)) 4646 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550) (("}") reduce 550)) 4647 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545) (("}") reduce 545)) 4648 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488) (("}") reduce 488)) 4649 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553) (("}") reduce 553)) 4652 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552) (("}") reduce 552)) 4653 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487) (("}") reduce 487)) 4664 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477) (("}") reduce 477)) 4665 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472) (("}") reduce 472)) 4675 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 477) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 477)) 4676 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473) (("}") reduce 473)) 4678 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524) (("}") reduce 524)) 4680 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554) (("}") reduce 554)) 4683 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526) (("}") reduce 526)) 4684 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489) (("}") reduce 489)) 4685 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491) (("}") reduce 491)) 4687 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468) (("}") reduce 468)) 4688 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470) (("}") reduce 470)) 4693 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543) (("}") reduce 543)) 4694 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 4695 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549) (("}") reduce 549)) 4696 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 4699 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 4700 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544) (("}") reduce 544)) 4701 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550) (("}") reduce 550)) 4702 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545) (("}") reduce 545)) 4704 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 4706 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) ((",") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503) (("}") reduce 503)) 4708 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 390) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 390)) 4709 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 465)) 4720 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523) (("}") reduce 523)) 4722 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504) (("}") reduce 504)) 4724 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517) (("}") reduce 517)) 4725 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507) (("}") reduce 507)) 4729 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) ((",") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_DOUBLE_ARROW) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500) (("}") reduce 500)) 4730 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516) (("}") reduce 516)) 4731 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520) (("}") reduce 520)) 4737 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515) (("}") reduce 515)) 4739 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506) (("}") reduce 506)) 4741 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519) (("}") reduce 519)) 4742 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 4743 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 4744 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 4746 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522) (("}") reduce 522)) 4749 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521) (("}") reduce 521)) 4755 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 4758 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518) (("}") reduce 518)) 4760 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551) (("}") reduce 551)) 4761 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) ((",") reduce 462) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 462) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("}") reduce 462) (("~") shift)) 4768 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547) (("}") reduce 547)) 4769 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513) (("}") reduce 513)) 4771 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512) (("}") reduce 512)) 4772 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511) (("}") reduce 511)) 4773 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 463)) 4776 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546) (("}") reduce 546)) 4777 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467) (("}") reduce 467)) 4778 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) ((",") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449) (("}") reduce 449)) 4779 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471) (("}") reduce 471)) 4780 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514) (("}") reduce 514)) 4781 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444) (("}") reduce 444)) 4782 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445) (("}") reduce 445)) 4783 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 4784 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542) (("}") reduce 542)) 4785 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548) (("}") reduce 548)) 4786 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) ((",") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 388) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388) (("}") reduce 388)) 4789 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) ((",") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432) (("}") reduce 432)) 4820 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416) (("}") reduce 416)) 4821 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419) (("}") reduce 419)) 4822 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428) (("}") reduce 428)) 4823 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 441) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441) (("}") reduce 441)) 4824 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427) (("}") reduce 427)) 4825 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424) (("}") reduce 424)) 4826 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 415)) 4827 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 413)) 4828 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 414)) 4829 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 438) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438) (("}") reduce 438)) 4830 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434) (("}") reduce 434)) 4831 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436) (("}") reduce 436)) 4832 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433) (("}") reduce 433)) 4833 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 440) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440) (("}") reduce 440)) 4834 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435) (("}") reduce 435)) 4837 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 4838 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 4839 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 4840 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492) (("}") reduce 492)) 4841 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 4842 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551) (("}") reduce 551)) 4843 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494) (("}") reduce 494)) 4844 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442) (("}") reduce 442)) 4845 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493) (("}") reduce 493)) 4846 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495) (("}") reduce 495)) 4847 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556) (("}") reduce 556)) 4853 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557) (("}") reduce 557)) 4854 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 4855 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 4856 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558) (("}") reduce 558)) 4857 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 4860 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 4861 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 4863 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 4865 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496) (("}") reduce 496)) 4866 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553) (("}") reduce 553)) 4869 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552) (("}") reduce 552)) 4870 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 448) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 448)) 4871 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 411)) 4872 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 412)) 4873 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417) (("}") reduce 417)) 4874 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418) (("}") reduce 418)) 4878 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 446) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 446)) 4879 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 447) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 447)) 4880 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 439) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439) (("}") reduce 439)) 4881 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 437) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437) (("}") reduce 437)) 4882 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425) (("}") reduce 425)) 4883 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420) (("}") reduce 420)) 4884 ((("%") shift) (("*") shift) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422) (("}") reduce 422)) 4885 ((("%") shift) (("*") shift) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421) (("}") reduce 421)) 4886 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423) (("}") reduce 423)) 4887 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426) (("}") reduce 426)) 4889 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74) (("}") reduce 74)) 4890 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52) (("}") reduce 52)) 4891 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27) (("}") reduce 27)) 4892 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51) (("}") reduce 51)) 4893 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53) (("}") reduce 53)) 4894 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49) (("}") reduce 49)) 4895 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29) (("}") reduce 29)) 4896 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59) (("}") reduce 59)) 4897 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60) (("}") reduce 60)) 4898 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11) (("}") reduce 11)) 4899 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42) (("}") reduce 42)) 4900 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39) (("}") reduce 39)) 4901 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25) (("}") reduce 25)) 4902 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50) (("}") reduce 50)) 4903 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66) (("}") reduce 66)) 4904 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18) (("}") reduce 18)) 4905 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17) (("}") reduce 17)) 4906 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15) (("}") reduce 15)) 4907 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14) (("}") reduce 14)) 4908 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38) (("}") reduce 38)) 4909 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26) (("}") reduce 26)) 4910 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22) (("}") reduce 22)) 4911 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24) (("}") reduce 24)) 4912 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16) (("}") reduce 16)) 4913 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48) (("}") reduce 48)) 4914 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20) (("}") reduce 20)) 4915 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70) (("}") reduce 70)) 4916 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3) (("}") reduce 3)) 4917 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12) (("}") reduce 12)) 4918 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54) (("}") reduce 54)) 4919 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65) (("}") reduce 65)) 4920 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75) (("}") reduce 75)) 4921 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30) (("}") reduce 30)) 4922 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68) (("}") reduce 68)) 4923 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21) (("}") reduce 21)) 4924 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23) (("}") reduce 23)) 4925 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41) (("}") reduce 41)) 4926 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62) (("}") reduce 62)) 4927 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34) (("}") reduce 34)) 4928 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40) (("}") reduce 40)) 4929 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13) (("}") reduce 13)) 4930 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55) (("}") reduce 55)) 4931 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1) (("}") reduce 1)) 4932 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2) (("}") reduce 2)) 4933 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9) (("}") reduce 9)) 4934 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33) (("}") reduce 33)) 4935 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58) (("}") reduce 58)) 4936 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37) (("}") reduce 37)) 4937 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64) (("}") reduce 64)) 4938 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46) (("}") reduce 46)) 4939 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8) (("}") reduce 8)) 4940 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6) (("}") reduce 6)) 4941 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7) (("}") reduce 7)) 4942 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69) (("}") reduce 69)) 4943 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63) (("}") reduce 63)) 4944 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56) (("}") reduce 56)) 4945 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10) (("}") reduce 10)) 4946 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67) (("}") reduce 67)) 4947 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44) (("}") reduce 44)) 4948 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76) (("}") reduce 76)) 4949 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71) (("}") reduce 71)) 4950 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77) (("}") reduce 77)) 4951 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78) (("}") reduce 78)) 4952 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79) (("}") reduce 79)) 4953 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4) (("}") reduce 4)) 4954 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5) (("}") reduce 5)) 4955 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43) (("}") reduce 43)) 4956 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73) (("}") reduce 73)) 4957 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82) (("}") reduce 82)) 4958 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47) (("}") reduce 47)) 4959 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31) (("}") reduce 31)) 4960 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57) (("}") reduce 57)) 4961 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61) (("}") reduce 61)) 4962 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28) (("}") reduce 28)) 4963 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36) (("}") reduce 36)) 4964 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32) (("}") reduce 32)) 4965 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35) (("}") reduce 35)) 4966 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19) (("}") reduce 19)) 4967 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45) (("}") reduce 45)) 4968 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525) (("}") reduce 525)) 4970 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72) (("}") reduce 72)) 4971 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83) (("}") reduce 83)) 4972 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555) (("}") reduce 555)) 4975 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527) (("}") reduce 527)) 4977 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490) (("}") reduce 490)) 4978 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289) (("}") reduce 289)) 4982 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290) (("}") reduce 290)) 4983 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 4988 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409) (("}") reduce 409)) 4990 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407) (("}") reduce 407)) 5000 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 403)) 5001 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 405)) 5002 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 404)) 5003 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 397)) 5004 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 394)) 5005 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 402)) 5006 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 396)) 5007 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 400) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 400)) 5008 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 395)) 5009 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 398) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 398)) 5010 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 399) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 399)) 5011 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 406)) 5012 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 401)) 5014 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 391) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 391)) 5017 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551) (("}") reduce 551)) 5020 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547) (("}") reduce 547)) 5022 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546) (("}") reduce 546)) 5024 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542) (("}") reduce 542)) 5025 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548) (("}") reduce 548)) 5026 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392) (("}") reduce 392)) 5030 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555) (("}") reduce 555)) 5032 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490) (("}") reduce 490)) 5033 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289) (("}") reduce 289)) 5037 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290) (("}") reduce 290)) 5038 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 5039 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487) (("}") reduce 487)) 5042 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554) (("}") reduce 554)) 5043 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489) (("}") reduce 489)) 5044 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491) (("}") reduce 491)) 5049 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543) (("}") reduce 543)) 5050 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 5051 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549) (("}") reduce 549)) 5052 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 5055 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 5056 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544) (("}") reduce 544)) 5057 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550) (("}") reduce 550)) 5058 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545) (("}") reduce 545)) 5059 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488) (("}") reduce 488)) 5060 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553) (("}") reduce 553)) 5063 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552) (("}") reduce 552)) 5064 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487) (("}") reduce 487)) 5075 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477) (("}") reduce 477)) 5076 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472) (("}") reduce 472)) 5086 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 477) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 477)) 5087 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473) (("}") reduce 473)) 5089 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 464)) 5091 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524) (("}") reduce 524)) 5093 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554) (("}") reduce 554)) 5096 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526) (("}") reduce 526)) 5097 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489) (("}") reduce 489)) 5098 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491) (("}") reduce 491)) 5100 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468) (("}") reduce 468)) 5101 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470) (("}") reduce 470)) 5106 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543) (("}") reduce 543)) 5107 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 5108 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549) (("}") reduce 549)) 5109 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 5112 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 5113 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544) (("}") reduce 544)) 5114 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550) (("}") reduce 550)) 5115 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545) (("}") reduce 545)) 5117 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 5119 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) ((",") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_DOUBLE_ARROW) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503) (("}") reduce 503)) 5121 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 390) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 390)) 5122 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 465)) 5125 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 464)) 5126 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456) (("}") reduce 456)) 5127 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 466) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 466)) 5128 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) ((",") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452) (("}") reduce 452)) 5129 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469) (("}") reduce 469)) 5131 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509) (("}") reduce 509)) 5133 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510) (("}") reduce 510)) 5134 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508) (("}") reduce 508)) 5135 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 602) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 602) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 602)) 5136 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 601) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 601) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 601)) 5137 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488) (("}") reduce 488)) 5138 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 461) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 461)) 5139 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454) (("}") reduce 454)) 5141 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 5142 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 5143 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 5144 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492) (("}") reduce 492)) 5145 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 5146 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) ((",") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385) (("}") reduce 385)) 5149 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494) (("}") reduce 494)) 5150 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) ((",") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387) (("}") reduce 387)) 5151 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493) (("}") reduce 493)) 5152 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495) (("}") reduce 495)) 5153 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556) (("}") reduce 556)) 5159 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557) (("}") reduce 557)) 5160 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 5161 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558) (("}") reduce 558)) 5162 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 5163 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) ((",") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384) (("}") reduce 384)) 5165 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 5166 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) ((",") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386) (("}") reduce 386)) 5174 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) ((",") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383) (("}") reduce 383)) 5176 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496) (("}") reduce 496)) 5182 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) ((",") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_DOUBLE_ARROW) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231) (("}") reduce 231)) 5187 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 389) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 389)) 5189 ((("%") shift) ((")") reduce 605) (("*") shift) (("+") shift) ((",") reduce 605) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5190 (((")") reduce 603) ((",") reduce 603)) 5192 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") reduce 125) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 5194 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) ((",") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_DOUBLE_ARROW) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596) (("}") reduce 596)) 5195 (((")") reduce 604) ((",") reduce 604)) 5196 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) ((",") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450) (("}") reduce 450)) 5197 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 599) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 599) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 599)) 5198 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 598) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 598) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 598)) 5199 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408) (("}") reduce 408)) 5201 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) ((",") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_DOUBLE_ARROW) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501) (("}") reduce 501)) 5202 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457) (("}") reduce 457)) 5203 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289) (("}") reduce 289)) 5207 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290) (("}") reduce 290)) 5208 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 5211 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) ((",") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_DOUBLE_ARROW) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600) (("}") reduce 600)) 5214 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) ((",") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_DOUBLE_ARROW) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597) (("}") reduce 597)) 5215 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) ((",") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451) (("}") reduce 451)) 5216 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410) (("}") reduce 410)) 5217 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393) (("}") reduce 393)) 5218 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455) (("}") reduce 455)) 5219 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) ((",") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453) (("}") reduce 453)) 5222 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) ((",") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_DOUBLE_ARROW) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502) (("}") reduce 502)) 5223 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 5224 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) ((",") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430) (("}") reduce 430)) 5225 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) ((",") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429) (("}") reduce 429)) 5227 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) ((",") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443) (("}") reduce 443)) 5228 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553) (("}") reduce 553)) 5231 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552) (("}") reduce 552)) 5234 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_VARIABLE) shift)) 5235 ((("\"") reduce 584) ((T_CURLY_OPEN) reduce 584) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 584) ((T_ENCAPSED_AND_WHITESPACE) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_VARIABLE) reduce 584) (("[") shift)) 5236 ((("\"") shift) ((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_ENCAPSED_AND_WHITESPACE) shift) ((T_VARIABLE) shift)) 5237 ((("\"") reduce 582) ((T_CURLY_OPEN) reduce 582) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 582) ((T_ENCAPSED_AND_WHITESPACE) reduce 582) ((T_VARIABLE) reduce 582)) 5238 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505) (("}") reduce 505)) 5239 ((("\"") reduce 581) ((T_CURLY_OPEN) reduce 581) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 581) ((T_ENCAPSED_AND_WHITESPACE) reduce 581) ((T_VARIABLE) reduce 581)) 5240 ((("\"") reduce 580) ((T_CURLY_OPEN) reduce 580) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 580) ((T_ENCAPSED_AND_WHITESPACE) reduce 580) ((T_VARIABLE) reduce 580)) 5245 ((("\"") reduce 585) ((T_CURLY_OPEN) reduce 585) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 585) ((T_ENCAPSED_AND_WHITESPACE) reduce 585) ((T_VARIABLE) reduce 585)) 5246 ((("\"") reduce 586) ((T_CURLY_OPEN) reduce 586) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 586) ((T_ENCAPSED_AND_WHITESPACE) reduce 586) ((T_VARIABLE) reduce 586)) 5247 ((("\"") reduce 587) ((T_CURLY_OPEN) reduce 587) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 587) ((T_ENCAPSED_AND_WHITESPACE) reduce 587) ((T_VARIABLE) reduce 587)) 5248 ((("\"") reduce 583) ((T_CURLY_OPEN) reduce 583) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 583) ((T_ENCAPSED_AND_WHITESPACE) reduce 583) ((T_VARIABLE) reduce 583)) 5251 ((("\"") reduce 588) ((T_CURLY_OPEN) reduce 588) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 588) ((T_ENCAPSED_AND_WHITESPACE) reduce 588) ((T_VARIABLE) reduce 588)) 5253 ((("\"") reduce 589) ((T_CURLY_OPEN) reduce 589) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 589) ((T_ENCAPSED_AND_WHITESPACE) reduce 589) ((T_VARIABLE) reduce 589)) 5256 ((("\"") reduce 590) ((T_CURLY_OPEN) reduce 590) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 590) ((T_ENCAPSED_AND_WHITESPACE) reduce 590) ((T_VARIABLE) reduce 590)) 5258 ((("\"") reduce 591) ((T_CURLY_OPEN) reduce 591) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 591) ((T_ENCAPSED_AND_WHITESPACE) reduce 591) ((T_VARIABLE) reduce 591)) 5259 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) ((",") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431) (("}") reduce 431)) 5260 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456) (("}") reduce 456)) 5261 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 466) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 466)) 5262 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) ((",") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452) (("}") reduce 452)) 5263 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469) (("}") reduce 469)) 5265 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509) (("}") reduce 509)) 5267 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510) (("}") reduce 510)) 5268 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508) (("}") reduce 508)) 5269 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 602) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 602)) 5270 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 601) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 601)) 5271 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488) (("}") reduce 488)) 5272 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 461) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 461)) 5273 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454) (("}") reduce 454)) 5275 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 5276 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 5277 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 5278 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492) (("}") reduce 492)) 5279 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 5280 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) ((",") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385) (("}") reduce 385)) 5283 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494) (("}") reduce 494)) 5284 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) ((",") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387) (("}") reduce 387)) 5285 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493) (("}") reduce 493)) 5286 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495) (("}") reduce 495)) 5287 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556) (("}") reduce 556)) 5293 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557) (("}") reduce 557)) 5294 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 5295 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558) (("}") reduce 558)) 5296 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 5297 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) ((",") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384) (("}") reduce 384)) 5299 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 5300 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) ((",") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386) (("}") reduce 386)) 5308 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) ((",") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383) (("}") reduce 383)) 5310 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496) (("}") reduce 496)) 5316 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) ((",") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231) (("}") reduce 231)) 5321 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 389) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 389)) 5325 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) ((",") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596) (("}") reduce 596)) 5326 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) ((",") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450) (("}") reduce 450)) 5327 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 599) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 599)) 5328 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 598) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 598)) 5329 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408) (("}") reduce 408)) 5331 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) ((",") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501) (("}") reduce 501)) 5332 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457) (("}") reduce 457)) 5333 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289) (("}") reduce 289)) 5337 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290) (("}") reduce 290)) 5338 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 5341 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) ((",") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600) (("}") reduce 600)) 5344 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) ((",") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597) (("}") reduce 597)) 5345 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) ((",") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451) (("}") reduce 451)) 5346 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410) (("}") reduce 410)) 5347 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393) (("}") reduce 393)) 5348 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455) (("}") reduce 455)) 5349 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) ((",") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453) (("}") reduce 453)) 5352 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) ((",") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502) (("}") reduce 502)) 5353 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 5354 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) ((",") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430) (("}") reduce 430)) 5355 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) ((",") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429) (("}") reduce 429)) 5357 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) ((",") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443) (("}") reduce 443)) 5358 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553) (("}") reduce 553)) 5361 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552) (("}") reduce 552)) 5363 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505) (("}") reduce 505)) 5364 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) ((",") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431) (("}") reduce 431)) 5365 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 239) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 239) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5376 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 5377 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 5387 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 477) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5388 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 5390 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524)) 5392 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 5395 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526)) 5396 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 5397 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 5399 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 5400 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 5405 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 5406 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 5407 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 5408 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 5411 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 5412 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 5413 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 5414 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 5416 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 5418 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) ((",") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_DOUBLE_ARROW) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503)) 5420 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 390) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5421 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5422 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5424 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5425 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456)) 5426 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 466) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5427 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) ((",") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452)) 5428 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 5430 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 5432 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 5433 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508)) 5434 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 602) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 602) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5435 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 601) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 601) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5436 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 5437 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 461) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5438 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454)) 5440 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 5441 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 5442 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 5443 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 5444 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 5445 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) ((",") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385)) 5448 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 5449 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) ((",") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387)) 5450 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 5451 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 5452 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 5458 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 5459 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 5460 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 5461 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 5462 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) ((",") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384)) 5464 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 5465 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) ((",") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386)) 5473 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) ((",") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383)) 5475 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 5481 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) ((",") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_DOUBLE_ARROW) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231)) 5486 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 389) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5490 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) ((",") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_DOUBLE_ARROW) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596)) 5491 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) ((",") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450)) 5492 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 599) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 599) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5493 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 598) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 598) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5494 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408)) 5496 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) ((",") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_DOUBLE_ARROW) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501)) 5497 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457)) 5498 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289)) 5502 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290)) 5503 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 5506 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) ((",") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_DOUBLE_ARROW) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600)) 5509 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) ((",") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_DOUBLE_ARROW) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597)) 5510 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) ((",") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451)) 5511 (((T_DOUBLE_ARROW) reduce 125)) 5514 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 237) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 237)) 5515 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410)) 5516 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 5517 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455)) 5518 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) ((",") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453)) 5521 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) ((",") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_DOUBLE_ARROW) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502)) 5522 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 5523 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) ((",") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430)) 5524 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) ((",") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429)) 5526 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) ((",") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443)) 5527 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 5530 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 5532 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 5533 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) ((",") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431)) 5538 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 389) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 389) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5542 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) ((",") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) ((";") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_DOUBLE_ARROW) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596)) 5543 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) ((",") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) ((";") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450)) 5544 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 599) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 599) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 599) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5545 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 598) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 598) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 598) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5546 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408)) 5548 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) ((",") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) ((";") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_DOUBLE_ARROW) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501)) 5549 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((";") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457)) 5550 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289)) 5554 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290)) 5555 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 5558 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) ((",") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) ((";") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_DOUBLE_ARROW) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600)) 5561 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) ((",") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) ((";") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_DOUBLE_ARROW) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597)) 5562 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) ((",") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) ((";") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451)) 5563 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((";") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410)) 5564 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((";") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 5565 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((";") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455)) 5566 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) ((",") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453)) 5569 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) ((",") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) ((";") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_DOUBLE_ARROW) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502)) 5570 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 5571 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) ((",") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) ((";") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430)) 5572 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) ((",") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) ((";") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429)) 5574 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) ((",") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443)) 5575 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 5578 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 5580 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 5581 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) ((",") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) ((";") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431)) 5582 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((";") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456)) 5583 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 466) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 466) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5584 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) ((",") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) ((";") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452)) 5585 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 5587 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 5589 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 5590 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508)) 5591 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 602) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 602) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5592 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 601) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 601) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5593 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 5594 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 461) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 461) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5595 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((";") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454)) 5597 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 5598 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 5599 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 5600 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 5601 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 5602 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) ((",") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) ((";") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385)) 5605 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 5606 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) ((",") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) ((";") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387)) 5607 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 5608 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 5609 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 5615 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 5616 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 5617 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 5618 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 5619 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) ((",") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) ((";") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384)) 5621 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 5622 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) ((",") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) ((";") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386)) 5630 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) ((",") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) ((";") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383)) 5632 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 5638 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) ((",") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) ((";") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231)) 5643 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 389) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 389) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5647 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) ((",") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) ((";") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596)) 5648 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) ((",") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) ((";") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450)) 5649 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 599) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 599) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5650 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 598) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 598) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5651 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408)) 5653 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) ((",") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) ((";") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501)) 5654 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((";") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457)) 5655 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289)) 5659 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290)) 5660 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 5663 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) ((",") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) ((";") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600)) 5666 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) ((",") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) ((";") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597)) 5667 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) ((",") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) ((";") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451)) 5668 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((";") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410)) 5669 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((";") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 5670 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((";") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455)) 5671 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) ((",") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453)) 5674 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) ((",") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) ((";") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502)) 5675 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 5676 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) ((",") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) ((";") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430)) 5677 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) ((",") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) ((";") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429)) 5679 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) ((",") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443)) 5680 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 5683 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 5685 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 5686 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) ((",") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) ((";") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431)) 5688 (((",") reduce 354) ((";") reduce 354) (("{") shift)) 5689 (((T_ABSTRACT) reduce 358) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 358) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 358) ((T_ATTRIBUTE) reduce 358) ((T_FINAL) reduce 358) ((T_PRIVATE) reduce 358) ((T_PRIVATE_SET) reduce 358) ((T_PROTECTED) reduce 358) ((T_PROTECTED_SET) reduce 358) ((T_PUBLIC) reduce 358) ((T_PUBLIC_SET) reduce 358) ((T_READONLY) reduce 358) ((T_STATIC) reduce 358) ((T_STRING) reduce 358) (("}") reduce 358)) 5690 (((T_ABSTRACT) shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 363) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 363) ((T_ATTRIBUTE) shift) ((T_FINAL) shift) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 363) (("}") shift)) 5691 (((T_ABSTRACT) reduce 349) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 349) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 349) ((T_FINAL) reduce 349) ((T_PRIVATE) reduce 349) ((T_PRIVATE_SET) reduce 349) ((T_PROTECTED) reduce 349) ((T_PROTECTED_SET) reduce 349) ((T_PUBLIC) reduce 349) ((T_PUBLIC_SET) reduce 349) ((T_READONLY) reduce 349) ((T_STATIC) reduce 349) ((T_STRING) reduce 349)) 5693 (((T_ABSTRACT) reduce 350) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 350) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 350) ((T_FINAL) reduce 350) ((T_PRIVATE) reduce 350) ((T_PRIVATE_SET) reduce 350) ((T_PROTECTED) reduce 350) ((T_PROTECTED_SET) reduce 350) ((T_PUBLIC) reduce 350) ((T_PUBLIC_SET) reduce 350) ((T_READONLY) reduce 350) ((T_STATIC) reduce 350) ((T_STRING) reduce 350)) 5694 (((T_ABSTRACT) reduce 344) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 344) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 344) ((T_FINAL) reduce 344) ((T_PRIVATE) reduce 344) ((T_PRIVATE_SET) reduce 344) ((T_PROTECTED) reduce 344) ((T_PROTECTED_SET) reduce 344) ((T_PUBLIC) reduce 344) ((T_PUBLIC_SET) reduce 344) ((T_READONLY) reduce 344) ((T_STATIC) reduce 344) ((T_STRING) reduce 344)) 5695 (((T_ABSTRACT) reduce 347) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 347) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 347) ((T_FINAL) reduce 347) ((T_PRIVATE) reduce 347) ((T_PRIVATE_SET) reduce 347) ((T_PROTECTED) reduce 347) ((T_PROTECTED_SET) reduce 347) ((T_PUBLIC) reduce 347) ((T_PUBLIC_SET) reduce 347) ((T_READONLY) reduce 347) ((T_STATIC) reduce 347) ((T_STRING) reduce 347)) 5696 (((T_ABSTRACT) reduce 343) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 343) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 343) ((T_FINAL) reduce 343) ((T_PRIVATE) reduce 343) ((T_PRIVATE_SET) reduce 343) ((T_PROTECTED) reduce 343) ((T_PROTECTED_SET) reduce 343) ((T_PUBLIC) reduce 343) ((T_PUBLIC_SET) reduce 343) ((T_READONLY) reduce 343) ((T_STATIC) reduce 343) ((T_STRING) reduce 343)) 5697 (((T_ABSTRACT) reduce 346) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 346) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 346) ((T_FINAL) reduce 346) ((T_PRIVATE) reduce 346) ((T_PRIVATE_SET) reduce 346) ((T_PROTECTED) reduce 346) ((T_PROTECTED_SET) reduce 346) ((T_PUBLIC) reduce 346) ((T_PUBLIC_SET) reduce 346) ((T_READONLY) reduce 346) ((T_STATIC) reduce 346) ((T_STRING) reduce 346)) 5698 (((T_ABSTRACT) reduce 342) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 342) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 342) ((T_FINAL) reduce 342) ((T_PRIVATE) reduce 342) ((T_PRIVATE_SET) reduce 342) ((T_PROTECTED) reduce 342) ((T_PROTECTED_SET) reduce 342) ((T_PUBLIC) reduce 342) ((T_PUBLIC_SET) reduce 342) ((T_READONLY) reduce 342) ((T_STATIC) reduce 342) ((T_STRING) reduce 342)) 5699 (((T_ABSTRACT) reduce 345) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 345) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 345) ((T_FINAL) reduce 345) ((T_PRIVATE) reduce 345) ((T_PRIVATE_SET) reduce 345) ((T_PROTECTED) reduce 345) ((T_PROTECTED_SET) reduce 345) ((T_PUBLIC) reduce 345) ((T_PUBLIC_SET) reduce 345) ((T_READONLY) reduce 345) ((T_STATIC) reduce 345) ((T_STRING) reduce 345)) 5700 (((T_ABSTRACT) reduce 351) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 351) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 351) ((T_FINAL) reduce 351) ((T_PRIVATE) reduce 351) ((T_PRIVATE_SET) reduce 351) ((T_PROTECTED) reduce 351) ((T_PROTECTED_SET) reduce 351) ((T_PUBLIC) reduce 351) ((T_PUBLIC_SET) reduce 351) ((T_READONLY) reduce 351) ((T_STATIC) reduce 351) ((T_STRING) reduce 351)) 5701 (((T_ABSTRACT) reduce 348) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 348) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 348) ((T_FINAL) reduce 348) ((T_PRIVATE) reduce 348) ((T_PRIVATE_SET) reduce 348) ((T_PROTECTED) reduce 348) ((T_PROTECTED_SET) reduce 348) ((T_PUBLIC) reduce 348) ((T_PUBLIC_SET) reduce 348) ((T_READONLY) reduce 348) ((T_STATIC) reduce 348) ((T_STRING) reduce 348)) 5702 (((T_ABSTRACT) reduce 101) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 101) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 101) ((T_ATTRIBUTE) reduce 101) ((T_FINAL) reduce 101) ((T_PRIVATE) reduce 101) ((T_PRIVATE_SET) reduce 101) ((T_PROTECTED) reduce 101) ((T_PROTECTED_SET) reduce 101) ((T_PUBLIC) reduce 101) ((T_PUBLIC_SET) reduce 101) ((T_READONLY) reduce 101) ((T_STATIC) reduce 101) ((T_STRING) reduce 101)) 5703 (((T_ABSTRACT) shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 363) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 363) ((T_ATTRIBUTE) shift) ((T_FINAL) shift) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 363)) 5704 (((T_ABSTRACT) reduce 340) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 340) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 340) ((T_FINAL) reduce 340) ((T_PRIVATE) reduce 340) ((T_PRIVATE_SET) reduce 340) ((T_PROTECTED) reduce 340) ((T_PROTECTED_SET) reduce 340) ((T_PUBLIC) reduce 340) ((T_PUBLIC_SET) reduce 340) ((T_READONLY) reduce 340) ((T_STATIC) reduce 340) ((T_STRING) reduce 340)) 5705 (((T_ABSTRACT) shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 364) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 364) ((T_FINAL) shift) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 364)) 5706 (((T_ABSTRACT) reduce 359) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 359) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 359) ((T_ATTRIBUTE) reduce 359) ((T_FINAL) reduce 359) ((T_PRIVATE) reduce 359) ((T_PRIVATE_SET) reduce 359) ((T_PROTECTED) reduce 359) ((T_PROTECTED_SET) reduce 359) ((T_PUBLIC) reduce 359) ((T_PUBLIC_SET) reduce 359) ((T_READONLY) reduce 359) ((T_STATIC) reduce 359) ((T_STRING) reduce 359) (("}") reduce 359)) 5707 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_STRING) reduce 479)) 5708 (((T_ABSTRACT) reduce 356) ((T_ATTRIBUTE) reduce 356) ((T_CASE) reduce 356) ((T_CONST) reduce 356) ((T_FINAL) reduce 356) ((T_FUNCTION) reduce 356) ((T_PRIVATE) reduce 356) ((T_PRIVATE_SET) reduce 356) ((T_PROTECTED) reduce 356) ((T_PROTECTED_SET) reduce 356) ((T_PUBLIC) reduce 356) ((T_PUBLIC_SET) reduce 356) ((T_READONLY) reduce 356) ((T_STATIC) reduce 356) ((T_USE) reduce 356) ((T_VAR) reduce 356) (("}") reduce 356)) 5709 (((T_STRING) reduce 80)) 5710 (((T_STRING) reduce 81)) 5711 (((T_STRING) reduce 480)) 5713 ((("(") reduce 476) ((";") reduce 476) ((T_DOUBLE_ARROW) reduce 476) (("{") reduce 476)) 5714 ((("(") shift) ((";") reduce 369) ((T_DOUBLE_ARROW) reduce 369) (("{") reduce 369)) 5716 (((";") reduce 477) ((T_DOUBLE_ARROW) reduce 477) (("{") reduce 477)) 5717 (((";") shift) ((T_DOUBLE_ARROW) shift) (("{") shift)) 5718 (((T_ABSTRACT) reduce 366) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 366) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 366) ((T_ATTRIBUTE) reduce 366) ((T_FINAL) reduce 366) ((T_PRIVATE) reduce 366) ((T_PRIVATE_SET) reduce 366) ((T_PROTECTED) reduce 366) ((T_PROTECTED_SET) reduce 366) ((T_PUBLIC) reduce 366) ((T_PUBLIC_SET) reduce 366) ((T_READONLY) reduce 366) ((T_STATIC) reduce 366) ((T_STRING) reduce 366) (("}") reduce 366)) 5720 (((T_ABSTRACT) reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_ATTRIBUTE) reduce 477) ((T_FINAL) reduce 477) ((T_PRIVATE) reduce 477) ((T_PRIVATE_SET) reduce 477) ((T_PROTECTED) reduce 477) ((T_PROTECTED_SET) reduce 477) ((T_PUBLIC) reduce 477) ((T_PUBLIC_SET) reduce 477) ((T_READONLY) reduce 477) ((T_STATIC) reduce 477) ((T_STRING) reduce 477) (("}") reduce 477)) 5723 (((T_ABSTRACT) reduce 367) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 367) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 367) ((T_ATTRIBUTE) reduce 367) ((T_FINAL) reduce 367) ((T_PRIVATE) reduce 367) ((T_PRIVATE_SET) reduce 367) ((T_PROTECTED) reduce 367) ((T_PROTECTED_SET) reduce 367) ((T_PUBLIC) reduce 367) ((T_PUBLIC_SET) reduce 367) ((T_READONLY) reduce 367) ((T_STATIC) reduce 367) ((T_STRING) reduce 367) (("}") reduce 367)) 5724 (((T_ABSTRACT) reduce 365) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 365) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 365) ((T_ATTRIBUTE) reduce 365) ((T_FINAL) reduce 365) ((T_PRIVATE) reduce 365) ((T_PRIVATE_SET) reduce 365) ((T_PROTECTED) reduce 365) ((T_PROTECTED_SET) reduce 365) ((T_PUBLIC) reduce 365) ((T_PUBLIC_SET) reduce 365) ((T_READONLY) reduce 365) ((T_STATIC) reduce 365) ((T_STRING) reduce 365) (("}") reduce 365)) 5732 (((T_ABSTRACT) reduce 368) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 368) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 368) ((T_ATTRIBUTE) reduce 368) ((T_FINAL) reduce 368) ((T_PRIVATE) reduce 368) ((T_PRIVATE_SET) reduce 368) ((T_PROTECTED) reduce 368) ((T_PROTECTED_SET) reduce 368) ((T_PUBLIC) reduce 368) ((T_PUBLIC_SET) reduce 368) ((T_READONLY) reduce 368) ((T_STATIC) reduce 368) ((T_STRING) reduce 368) (("}") reduce 368)) 5757 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 416) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416)) 5758 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 419) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419)) 5759 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 428) (("/") shift) ((";") reduce 428) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428)) 5760 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 441) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441)) 5761 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 427) (("/") shift) ((";") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 5762 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 5763 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5764 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5765 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5766 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 438) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438)) 5767 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 5768 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 436) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 5769 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 5770 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 440) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440)) 5771 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 5774 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 5775 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 5776 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 5777 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 5778 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 5779 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 5780 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 5781 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 5782 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 5783 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 5784 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 5790 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 5791 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 5792 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 5793 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 5794 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 5797 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 5798 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 5800 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 5802 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 5803 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 5806 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 5807 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 448) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5808 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 411) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5809 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 412) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5810 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 417) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417)) 5811 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 418) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418)) 5815 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 446) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5816 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 447) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5817 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 439) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439)) 5818 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 437) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437)) 5819 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 5820 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 420) (("/") shift) ((";") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420)) 5821 ((("%") shift) (("*") shift) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) ((";") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 5822 ((("%") shift) (("*") shift) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) ((";") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 5823 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 5824 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 5826 (((";") reduce 370) ((T_DOUBLE_ARROW) reduce 370) (("{") reduce 370)) 5827 (((T_ABSTRACT) reduce 341) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 341) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 341) ((T_FINAL) reduce 341) ((T_PRIVATE) reduce 341) ((T_PRIVATE_SET) reduce 341) ((T_PROTECTED) reduce 341) ((T_PROTECTED_SET) reduce 341) ((T_PUBLIC) reduce 341) ((T_PUBLIC_SET) reduce 341) ((T_READONLY) reduce 341) ((T_STATIC) reduce 341) ((T_STRING) reduce 341)) 5828 (((T_ABSTRACT) reduce 102) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 102) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_FINAL) reduce 102) ((T_PRIVATE) reduce 102) ((T_PRIVATE_SET) reduce 102) ((T_PROTECTED) reduce 102) ((T_PROTECTED_SET) reduce 102) ((T_PUBLIC) reduce 102) ((T_PUBLIC_SET) reduce 102) ((T_READONLY) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102)) 5829 (((T_ABSTRACT) reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_ATTRIBUTE) reduce 360) ((T_FINAL) reduce 360) ((T_PRIVATE) reduce 360) ((T_PRIVATE_SET) reduce 360) ((T_PROTECTED) reduce 360) ((T_PROTECTED_SET) reduce 360) ((T_PUBLIC) reduce 360) ((T_PUBLIC_SET) reduce 360) ((T_READONLY) reduce 360) ((T_STATIC) reduce 360) ((T_STRING) reduce 360) (("}") reduce 360)) 5830 ((("(") reduce 94) ((",") reduce 94) (("]") reduce 94)) 5831 ((("(") reduce 93) ((",") reduce 93) (("]") reduce 93)) 5832 ((("(") reduce 95) ((",") reduce 95) (("]") reduce 95)) 5833 ((("(") reduce 492) ((",") reduce 492) (("]") reduce 492)) 5834 ((("(") reduce 92) ((",") reduce 92) (("]") reduce 92)) 5835 (((",") reduce 98) (("]") reduce 98)) 5836 (((",") shift) (("]") reduce 124)) 5837 ((("(") shift) ((",") reduce 96) (("]") reduce 96)) 5838 ((("(") reduce 493) ((",") reduce 493) (("]") reduce 493)) 5840 (((",") reduce 97) (("]") reduce 97)) 5841 (((",") reduce 289) (("]") reduce 289)) 5845 (((",") reduce 290) (("]") reduce 290)) 5846 (((",") reduce 291) (("]") reduce 291)) 5847 (((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift) (("]") reduce 125)) 5849 (((T_ABSTRACT) reduce 100) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 100) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 100) ((T_ATTRIBUTE) reduce 100) ((T_FINAL) reduce 100) ((T_PRIVATE) reduce 100) ((T_PRIVATE_SET) reduce 100) ((T_PROTECTED) reduce 100) ((T_PROTECTED_SET) reduce 100) ((T_PUBLIC) reduce 100) ((T_PUBLIC_SET) reduce 100) ((T_READONLY) reduce 100) ((T_STATIC) reduce 100) ((T_STRING) reduce 100)) 5850 (((",") reduce 99) (("]") reduce 99)) 5861 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 5863 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("{") reduce 504) (("|") reduce 504)) 5865 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 5866 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 5870 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) ((",") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) ((";") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("{") reduce 500) (("|") reduce 500)) 5871 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 5872 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 5878 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 5880 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 5882 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 5883 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 5884 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 5885 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 5887 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 5890 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 5896 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 5899 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 5901 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("{") reduce 551) (("|") reduce 551)) 5902 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) ((",") reduce 462) (("-") shift) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("{") reduce 462) (("|") reduce 462) (("~") shift)) 5909 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((";") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("{") reduce 547) (("|") reduce 547)) 5910 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 5912 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 5913 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 5914 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 476) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 476) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 476) (("|") shift)) 5917 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((";") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("{") reduce 546) (("|") reduce 546)) 5918 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("{") reduce 467) (("|") reduce 467)) 5919 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) ((",") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) ((";") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("{") reduce 449) (("|") reduce 449)) 5920 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 5921 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((";") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 5922 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("{") reduce 444) (("|") reduce 444)) 5923 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("{") reduce 445) (("|") reduce 445)) 5924 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("{") reduce 459) (("|") reduce 459)) 5925 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("{") reduce 542) (("|") reduce 542)) 5926 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((";") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("{") reduce 548) (("|") reduce 548)) 5927 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) ((",") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) ((";") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("{") reduce 388) (("|") reduce 388)) 5930 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) ((",") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) ((";") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("{") reduce 432) (("|") reduce 432)) 5961 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 416) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 416) (("|") reduce 416)) 5962 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 419) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("{") reduce 419) (("|") reduce 419)) 5963 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") reduce 428) (("/") shift) ((";") reduce 428) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("{") reduce 428) (("|") reduce 428)) 5964 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 441) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 441) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("{") reduce 441) (("|") reduce 441)) 5965 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") reduce 427) (("/") shift) ((";") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("{") reduce 427) (("|") reduce 427)) 5966 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("{") reduce 424) (("|") reduce 424)) 5967 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 415) (("|") shift)) 5968 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 413) (("|") shift)) 5969 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 414) (("|") shift)) 5970 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 438) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 438) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("{") reduce 438) (("|") reduce 438)) 5971 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("{") reduce 434) (("|") reduce 434)) 5972 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 436) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("{") reduce 436) (("|") reduce 436)) 5973 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("{") reduce 433) (("|") reduce 433)) 5974 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 440) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 440) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("{") reduce 440) (("|") reduce 440)) 5975 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("{") reduce 435) (("|") reduce 435)) 5978 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 5979 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 5980 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 5981 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 5982 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 5983 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("{") reduce 551) (("|") reduce 551)) 5984 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("{") reduce 494) (("|") reduce 494)) 5985 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("{") reduce 442) (("|") reduce 442)) 5986 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("{") reduce 493) (("|") reduce 493)) 5987 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("{") reduce 495) (("|") reduce 495)) 5988 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("{") reduce 556) (("|") reduce 556)) 5994 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("{") reduce 557) (("|") reduce 557)) 5995 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 5996 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 5997 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("{") reduce 558) (("|") reduce 558)) 5998 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 6001 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 6002 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 6004 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 6006 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("{") reduce 496) (("|") reduce 496)) 6007 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("{") reduce 553) (("|") reduce 553)) 6010 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("{") reduce 552) (("|") reduce 552)) 6011 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 448) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 448) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 448) (("|") shift)) 6012 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 411) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 411) (("|") shift)) 6013 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 412) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 412) (("|") shift)) 6014 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 417) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("{") reduce 417) (("|") reduce 417)) 6015 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 418) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("{") reduce 418) (("|") reduce 418)) 6019 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 446) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 446) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 446) (("|") shift)) 6020 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 447) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 447) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 447) (("|") shift)) 6021 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 439) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 439) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("{") reduce 439) (("|") reduce 439)) 6022 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 437) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 437) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("{") reduce 437) (("|") reduce 437)) 6023 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("{") reduce 425) (("|") reduce 425)) 6024 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") reduce 420) (("/") shift) ((";") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("{") reduce 420) (("|") reduce 420)) 6025 ((("%") shift) (("*") shift) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) ((";") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("{") reduce 422) (("|") reduce 422)) 6026 ((("%") shift) (("*") shift) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) ((";") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("{") reduce 421) (("|") reduce 421)) 6027 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("{") reduce 423) (("|") reduce 423)) 6028 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("{") reduce 426) (("|") reduce 426)) 6030 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 6031 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 6032 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 6033 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 6034 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 6035 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 6036 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 6037 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 6038 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 6039 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 6040 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 6041 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 6042 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 6043 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 6044 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 6045 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 6046 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 6047 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 6048 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 6049 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 6050 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 6051 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 6052 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 6053 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 6054 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 6055 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 6056 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 6057 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 6058 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 6059 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 6060 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 6061 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 6062 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 6063 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 6064 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 6065 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 6066 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 6067 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 6068 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 6069 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 6070 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 6071 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 6072 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 6073 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 6074 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 6075 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 6076 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 6077 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 6078 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 6079 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 6080 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 6081 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 6082 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 6083 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 6084 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 6085 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 6086 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 6087 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 6088 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 6089 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 6090 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 6091 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 6092 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 6093 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("{") reduce 79) (("|") reduce 79)) 6094 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 6095 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 6096 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 6097 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 6098 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 6099 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 6100 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 6101 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 6102 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 6103 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 6104 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 6105 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 6106 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 6107 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 6108 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 6109 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 6111 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 6112 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("{") reduce 83) (("|") reduce 83)) 6113 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((";") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("{") reduce 555) (("|") reduce 555)) 6116 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("{") reduce 527) (("|") reduce 527)) 6118 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 6119 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("{") reduce 289) (("|") reduce 289)) 6123 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("{") reduce 290) (("|") reduce 290)) 6124 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 6129 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("{") reduce 409) (("|") reduce 409)) 6131 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("{") reduce 407) (("|") reduce 407)) 6141 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 403) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 403) (("|") shift)) 6142 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 405) (("|") shift)) 6143 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 404) (("|") shift)) 6144 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 397) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 397) (("|") shift)) 6145 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 394) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 394) (("|") shift)) 6146 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 402) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 402) (("|") shift)) 6147 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 396) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 396) (("|") shift)) 6148 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 400) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 400) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 400) (("|") shift)) 6149 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 395) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 395) (("|") shift)) 6150 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 398) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 398) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 398) (("|") shift)) 6151 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 399) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 399) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 399) (("|") shift)) 6152 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 406) (("|") shift)) 6153 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 401) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 401) (("|") shift)) 6155 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 391) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 391) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 391) (("|") shift)) 6158 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("{") reduce 551) (("|") reduce 551)) 6161 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((";") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("{") reduce 547) (("|") reduce 547)) 6163 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((";") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("{") reduce 546) (("|") reduce 546)) 6165 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("{") reduce 542) (("|") reduce 542)) 6166 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((";") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("{") reduce 548) (("|") reduce 548)) 6167 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((";") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("{") reduce 392) (("|") reduce 392)) 6171 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((";") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("{") reduce 555) (("|") reduce 555)) 6173 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 6174 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("{") reduce 289) (("|") reduce 289)) 6178 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("{") reduce 290) (("|") reduce 290)) 6179 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 6180 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 6183 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((";") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("{") reduce 554) (("|") reduce 554)) 6184 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 6185 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 6190 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("{") reduce 543) (("|") reduce 543)) 6191 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 6192 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((";") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("{") reduce 549) (("|") reduce 549)) 6193 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 6196 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 6197 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("{") reduce 544) (("|") reduce 544)) 6198 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((";") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("{") reduce 550) (("|") reduce 550)) 6199 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((";") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("{") reduce 545) (("|") reduce 545)) 6200 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 6201 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("{") reduce 553) (("|") reduce 553)) 6204 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("{") reduce 552) (("|") reduce 552)) 6205 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 6216 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("{") reduce 477) (("|") reduce 477)) 6217 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 6227 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 477) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 477) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 477) (("|") shift)) 6228 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 6229 (((",") reduce 355) ((";") reduce 355) (("{") shift)) 6232 (((T_ABSTRACT) reduce 357) ((T_ATTRIBUTE) reduce 357) ((T_CASE) reduce 357) ((T_CONST) reduce 357) ((T_FINAL) reduce 357) ((T_FUNCTION) reduce 357) ((T_PRIVATE) reduce 357) ((T_PRIVATE_SET) reduce 357) ((T_PROTECTED) reduce 357) ((T_PROTECTED_SET) reduce 357) ((T_PUBLIC) reduce 357) ((T_PUBLIC_SET) reduce 357) ((T_READONLY) reduce 357) ((T_STATIC) reduce 357) ((T_USE) reduce 357) ((T_VAR) reduce 357) (("}") reduce 357)) 6234 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 6236 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((";") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("{") reduce 554) (("|") reduce 554)) 6239 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 6240 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 6241 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 6243 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("{") reduce 468) (("|") reduce 468)) 6244 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 6249 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("{") reduce 543) (("|") reduce 543)) 6250 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 6251 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((";") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("{") reduce 549) (("|") reduce 549)) 6252 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 6255 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 6256 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("{") reduce 544) (("|") reduce 544)) 6257 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((";") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("{") reduce 550) (("|") reduce 550)) 6258 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((";") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("{") reduce 545) (("|") reduce 545)) 6260 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("{") reduce 460) (("|") reduce 460)) 6262 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) ((",") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) ((";") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("{") reduce 503) (("|") reduce 503)) 6264 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 390) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 390) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 390) (("|") shift)) 6265 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 465) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 465) (("|") shift)) 6276 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 6278 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("{") reduce 504) (("|") reduce 504)) 6280 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 6281 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 6285 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) ((",") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) ((";") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_DOUBLE_ARROW) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("{") reduce 500) (("|") reduce 500)) 6286 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 6287 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 6293 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 6295 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 6297 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 6298 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 6299 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 6300 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 6302 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 6305 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 6311 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 6314 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 6316 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("{") reduce 551) (("|") reduce 551)) 6317 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) ((",") reduce 462) (("-") shift) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 462) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("{") reduce 462) (("|") reduce 462) (("~") shift)) 6324 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((";") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("{") reduce 547) (("|") reduce 547)) 6325 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 6327 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 6328 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 6329 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 463) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 463) (("|") shift)) 6332 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((";") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("{") reduce 546) (("|") reduce 546)) 6333 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("{") reduce 467) (("|") reduce 467)) 6334 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) ((",") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) ((";") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("{") reduce 449) (("|") reduce 449)) 6335 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 6336 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((";") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 6337 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("{") reduce 444) (("|") reduce 444)) 6338 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("{") reduce 445) (("|") reduce 445)) 6339 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("{") reduce 459) (("|") reduce 459)) 6340 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("{") reduce 542) (("|") reduce 542)) 6341 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((";") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("{") reduce 548) (("|") reduce 548)) 6342 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) ((",") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) ((";") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 388) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("{") reduce 388) (("|") reduce 388)) 6345 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) ((",") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) ((";") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("{") reduce 432) (("|") reduce 432)) 6376 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 416) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 416) (("|") reduce 416)) 6377 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 419) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("{") reduce 419) (("|") reduce 419)) 6378 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") reduce 428) (("/") shift) ((";") reduce 428) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("{") reduce 428) (("|") reduce 428)) 6379 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 441) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 441) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("{") reduce 441) (("|") reduce 441)) 6380 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") reduce 427) (("/") shift) ((";") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("{") reduce 427) (("|") reduce 427)) 6381 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("{") reduce 424) (("|") reduce 424)) 6382 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 415) (("|") shift)) 6383 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 413) (("|") shift)) 6384 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 414) (("|") shift)) 6385 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 438) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 438) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("{") reduce 438) (("|") reduce 438)) 6386 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("{") reduce 434) (("|") reduce 434)) 6387 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 436) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("{") reduce 436) (("|") reduce 436)) 6388 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("{") reduce 433) (("|") reduce 433)) 6389 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 440) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 440) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("{") reduce 440) (("|") reduce 440)) 6390 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("{") reduce 435) (("|") reduce 435)) 6393 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 6394 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 6395 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 6396 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 6397 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 6398 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("{") reduce 551) (("|") reduce 551)) 6399 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("{") reduce 494) (("|") reduce 494)) 6400 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("{") reduce 442) (("|") reduce 442)) 6401 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("{") reduce 493) (("|") reduce 493)) 6402 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("{") reduce 495) (("|") reduce 495)) 6403 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("{") reduce 556) (("|") reduce 556)) 6409 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("{") reduce 557) (("|") reduce 557)) 6410 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 6411 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 6412 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("{") reduce 558) (("|") reduce 558)) 6413 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 6416 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 6417 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 6419 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 6421 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("{") reduce 496) (("|") reduce 496)) 6422 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("{") reduce 553) (("|") reduce 553)) 6425 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("{") reduce 552) (("|") reduce 552)) 6426 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 448) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 448) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 448) (("|") shift)) 6427 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 411) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 411) (("|") shift)) 6428 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 412) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 412) (("|") shift)) 6429 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 417) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("{") reduce 417) (("|") reduce 417)) 6430 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 418) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("{") reduce 418) (("|") reduce 418)) 6434 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 446) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 446) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 446) (("|") shift)) 6435 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 447) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 447) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 447) (("|") shift)) 6436 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 439) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 439) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("{") reduce 439) (("|") reduce 439)) 6437 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 437) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 437) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("{") reduce 437) (("|") reduce 437)) 6438 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("{") reduce 425) (("|") reduce 425)) 6439 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") reduce 420) (("/") shift) ((";") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("{") reduce 420) (("|") reduce 420)) 6440 ((("%") shift) (("*") shift) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) ((";") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("{") reduce 422) (("|") reduce 422)) 6441 ((("%") shift) (("*") shift) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) ((";") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("{") reduce 421) (("|") reduce 421)) 6442 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("{") reduce 423) (("|") reduce 423)) 6443 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("{") reduce 426) (("|") reduce 426)) 6445 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 6446 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 6447 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 6448 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 6449 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 6450 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 6451 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 6452 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 6453 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 6454 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 6455 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 6456 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 6457 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 6458 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 6459 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 6460 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 6461 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 6462 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 6463 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 6464 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 6465 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 6466 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 6467 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 6468 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 6469 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 6470 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 6471 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 6472 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 6473 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 6474 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 6475 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 6476 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 6477 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 6478 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 6479 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 6480 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 6481 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 6482 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 6483 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 6484 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 6485 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 6486 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 6487 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 6488 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 6489 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 6490 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 6491 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 6492 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 6493 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 6494 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 6495 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 6496 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 6497 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 6498 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 6499 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 6500 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 6501 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 6502 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 6503 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 6504 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 6505 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 6506 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 6507 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 6508 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("{") reduce 79) (("|") reduce 79)) 6509 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 6510 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 6511 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 6512 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 6513 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 6514 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 6515 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 6516 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 6517 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 6518 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 6519 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 6520 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 6521 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 6522 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 6523 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 6524 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 6526 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 6527 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("{") reduce 83) (("|") reduce 83)) 6528 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((";") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("{") reduce 555) (("|") reduce 555)) 6531 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("{") reduce 527) (("|") reduce 527)) 6533 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 6534 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("{") reduce 289) (("|") reduce 289)) 6538 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("{") reduce 290) (("|") reduce 290)) 6539 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 6544 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("{") reduce 409) (("|") reduce 409)) 6546 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("{") reduce 407) (("|") reduce 407)) 6556 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 403) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 403) (("|") shift)) 6557 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 405) (("|") shift)) 6558 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 404) (("|") shift)) 6559 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 397) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 397) (("|") shift)) 6560 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 394) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 394) (("|") shift)) 6561 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 402) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 402) (("|") shift)) 6562 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 396) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 396) (("|") shift)) 6563 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 400) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 400) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 400) (("|") shift)) 6564 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 395) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 395) (("|") shift)) 6565 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 398) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 398) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 398) (("|") shift)) 6566 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 399) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 399) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 399) (("|") shift)) 6567 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 406) (("|") shift)) 6568 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 401) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 401) (("|") shift)) 6570 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 391) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 391) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 391) (("|") shift)) 6573 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("{") reduce 551) (("|") reduce 551)) 6576 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((";") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("{") reduce 547) (("|") reduce 547)) 6578 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((";") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("{") reduce 546) (("|") reduce 546)) 6580 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("{") reduce 542) (("|") reduce 542)) 6581 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((";") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("{") reduce 548) (("|") reduce 548)) 6582 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((";") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("{") reduce 392) (("|") reduce 392)) 6586 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((";") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("{") reduce 555) (("|") reduce 555)) 6588 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 6589 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("{") reduce 289) (("|") reduce 289)) 6593 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("{") reduce 290) (("|") reduce 290)) 6594 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 6595 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 6598 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((";") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("{") reduce 554) (("|") reduce 554)) 6599 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 6600 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 6605 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("{") reduce 543) (("|") reduce 543)) 6606 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 6607 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((";") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("{") reduce 549) (("|") reduce 549)) 6608 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 6611 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 6612 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("{") reduce 544) (("|") reduce 544)) 6613 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((";") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("{") reduce 550) (("|") reduce 550)) 6614 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((";") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("{") reduce 545) (("|") reduce 545)) 6615 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 6616 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("{") reduce 553) (("|") reduce 553)) 6619 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("{") reduce 552) (("|") reduce 552)) 6620 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 6631 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("{") reduce 477) (("|") reduce 477)) 6632 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 6642 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 477) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 477) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 477) (("|") shift)) 6643 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 6645 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 464) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 464) (("|") shift)) 6647 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 6649 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((";") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("{") reduce 554) (("|") reduce 554)) 6652 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 6653 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 6654 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 6656 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("{") reduce 468) (("|") reduce 468)) 6657 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 6662 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("{") reduce 543) (("|") reduce 543)) 6663 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 6664 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((";") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("{") reduce 549) (("|") reduce 549)) 6665 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 6668 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 6669 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("{") reduce 544) (("|") reduce 544)) 6670 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((";") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("{") reduce 550) (("|") reduce 550)) 6671 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((";") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("{") reduce 545) (("|") reduce 545)) 6673 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("{") reduce 460) (("|") reduce 460)) 6675 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) ((",") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) ((";") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_DOUBLE_ARROW) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("{") reduce 503) (("|") reduce 503)) 6677 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 390) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 390) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 390) (("|") shift)) 6678 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 465) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 465) (("|") shift)) 6681 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 464) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 464) (("|") shift)) 6682 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((";") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 6683 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 466) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 466) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 466) (("|") shift)) 6684 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) ((",") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) ((";") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("{") reduce 452) (("|") reduce 452)) 6685 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("{") reduce 469) (("|") reduce 469)) 6687 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 6689 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 6690 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 6691 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 602) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 602) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 602) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 602) (("|") shift)) 6692 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 601) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 601) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 601) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 601) (("|") shift)) 6693 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 6694 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 461) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 461) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 461) (("|") shift)) 6695 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((";") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 6697 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 6698 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 6699 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 6700 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 6701 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 6702 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) ((",") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) ((";") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("{") reduce 385) (("|") reduce 385)) 6705 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("{") reduce 494) (("|") reduce 494)) 6706 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) ((",") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) ((";") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("{") reduce 387) (("|") reduce 387)) 6707 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("{") reduce 493) (("|") reduce 493)) 6708 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("{") reduce 495) (("|") reduce 495)) 6709 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("{") reduce 556) (("|") reduce 556)) 6715 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("{") reduce 557) (("|") reduce 557)) 6716 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 6717 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("{") reduce 558) (("|") reduce 558)) 6718 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 6719 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) ((",") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) ((";") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("{") reduce 384) (("|") reduce 384)) 6721 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 6722 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) ((",") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) ((";") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("{") reduce 386) (("|") reduce 386)) 6730 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) ((",") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) ((";") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("{") reduce 383) (("|") reduce 383)) 6732 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("{") reduce 496) (("|") reduce 496)) 6738 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) ((",") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) ((";") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_DOUBLE_ARROW) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("{") reduce 231) (("|") reduce 231)) 6743 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 389) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 389) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 389) (("|") shift)) 6747 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) ((",") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) ((";") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_DOUBLE_ARROW) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("{") reduce 596) (("|") reduce 596)) 6748 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) ((",") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) ((";") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("{") reduce 450) (("|") reduce 450)) 6749 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 599) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 599) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 599) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 599) (("|") shift)) 6750 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 598) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 598) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 598) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 598) (("|") shift)) 6751 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("{") reduce 408) (("|") reduce 408)) 6753 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) ((",") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) ((";") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_DOUBLE_ARROW) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("{") reduce 501) (("|") reduce 501)) 6754 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((";") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 6755 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("{") reduce 289) (("|") reduce 289)) 6759 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("{") reduce 290) (("|") reduce 290)) 6760 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 6763 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) ((",") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) ((";") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_DOUBLE_ARROW) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("{") reduce 600) (("|") reduce 600)) 6766 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) ((",") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) ((";") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_DOUBLE_ARROW) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("{") reduce 597) (("|") reduce 597)) 6767 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) ((",") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) ((";") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("{") reduce 451) (("|") reduce 451)) 6768 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((";") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("{") reduce 410) (("|") reduce 410)) 6769 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((";") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("{") reduce 393) (("|") reduce 393)) 6770 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((";") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 6771 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) ((",") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("{") reduce 453) (("|") reduce 453)) 6774 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) ((",") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) ((";") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_DOUBLE_ARROW) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("{") reduce 502) (("|") reduce 502)) 6775 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("{") reduce 458) (("|") reduce 458)) 6776 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) ((",") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) ((";") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("{") reduce 430) (("|") reduce 430)) 6777 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) ((",") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) ((";") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("{") reduce 429) (("|") reduce 429)) 6779 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) ((",") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("{") reduce 443) (("|") reduce 443)) 6780 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("{") reduce 553) (("|") reduce 553)) 6783 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("{") reduce 552) (("|") reduce 552)) 6785 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 6786 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) ((",") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) ((";") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("{") reduce 431) (("|") reduce 431)) 6787 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((";") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 6788 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 466) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 466) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 466) (("|") shift)) 6789 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) ((",") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) ((";") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("{") reduce 452) (("|") reduce 452)) 6790 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("{") reduce 469) (("|") reduce 469)) 6792 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 6794 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 6795 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 6796 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 602) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 602) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 602) (("|") shift)) 6797 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 601) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 601) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 601) (("|") shift)) 6798 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 6799 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 461) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 461) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 461) (("|") shift)) 6800 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((";") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 6802 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 6803 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 6804 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 6805 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 6806 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 6807 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) ((",") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) ((";") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("{") reduce 385) (("|") reduce 385)) 6810 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("{") reduce 494) (("|") reduce 494)) 6811 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) ((",") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) ((";") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("{") reduce 387) (("|") reduce 387)) 6812 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("{") reduce 493) (("|") reduce 493)) 6813 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("{") reduce 495) (("|") reduce 495)) 6814 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("{") reduce 556) (("|") reduce 556)) 6820 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("{") reduce 557) (("|") reduce 557)) 6821 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 6822 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("{") reduce 558) (("|") reduce 558)) 6823 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 6824 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) ((",") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) ((";") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("{") reduce 384) (("|") reduce 384)) 6826 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 6827 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) ((",") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) ((";") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("{") reduce 386) (("|") reduce 386)) 6835 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) ((",") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) ((";") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("{") reduce 383) (("|") reduce 383)) 6837 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("{") reduce 496) (("|") reduce 496)) 6843 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) ((",") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) ((";") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("{") reduce 231) (("|") reduce 231)) 6848 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 389) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 389) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 389) (("|") shift)) 6852 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) ((",") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) ((";") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("{") reduce 596) (("|") reduce 596)) 6853 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) ((",") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) ((";") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("{") reduce 450) (("|") reduce 450)) 6854 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 599) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 599) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 599) (("|") shift)) 6855 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 598) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 598) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 598) (("|") shift)) 6856 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("{") reduce 408) (("|") reduce 408)) 6858 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) ((",") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) ((";") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("{") reduce 501) (("|") reduce 501)) 6859 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((";") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 6860 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("{") reduce 289) (("|") reduce 289)) 6864 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("{") reduce 290) (("|") reduce 290)) 6865 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 6868 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) ((",") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) ((";") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("{") reduce 600) (("|") reduce 600)) 6871 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) ((",") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) ((";") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("{") reduce 597) (("|") reduce 597)) 6872 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) ((",") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) ((";") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("{") reduce 451) (("|") reduce 451)) 6873 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((";") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("{") reduce 410) (("|") reduce 410)) 6874 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((";") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("{") reduce 393) (("|") reduce 393)) 6875 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((";") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 6876 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) ((",") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("{") reduce 453) (("|") reduce 453)) 6879 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) ((",") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) ((";") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("{") reduce 502) (("|") reduce 502)) 6880 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("{") reduce 458) (("|") reduce 458)) 6881 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) ((",") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) ((";") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("{") reduce 430) (("|") reduce 430)) 6882 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) ((",") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) ((";") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("{") reduce 429) (("|") reduce 429)) 6884 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) ((",") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("{") reduce 443) (("|") reduce 443)) 6885 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("{") reduce 553) (("|") reduce 553)) 6888 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("{") reduce 552) (("|") reduce 552)) 6890 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 6891 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) ((",") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) ((";") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("{") reduce 431) (("|") reduce 431)) 6893 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 286) ((T_VARIABLE) reduce 286)) 6894 (((T_VARIABLE) reduce 278)) 6895 (((T_VARIABLE) reduce 279)) 6896 (((T_VARIABLE) reduce 94)) 6897 (((T_VARIABLE) reduce 93)) 6898 (((T_VARIABLE) reduce 95)) 6899 (((T_VARIABLE) reduce 92)) 6900 (((T_VARIABLE) reduce 280)) 6901 (((T_VARIABLE) reduce 275)) 6903 ((("|") reduce 282)) 6904 ((("(") reduce 341) (("?") reduce 341) ((T_ABSTRACT) reduce 341) ((T_ARRAY) reduce 341) ((T_CALLABLE) reduce 341) ((T_CONST) reduce 341) ((T_FINAL) reduce 341) ((T_FUNCTION) reduce 341) ((T_NAME_FULLY_QUALIFIED) reduce 341) ((T_NAME_QUALIFIED) reduce 341) ((T_NAME_RELATIVE) reduce 341) ((T_PRIVATE) reduce 341) ((T_PRIVATE_SET) reduce 341) ((T_PROTECTED) reduce 341) ((T_PROTECTED_SET) reduce 341) ((T_PUBLIC) reduce 341) ((T_PUBLIC_SET) reduce 341) ((T_READONLY) reduce 341) ((T_STATIC) reduce 341) ((T_STRING) reduce 341) ((T_VARIABLE) reduce 341)) 6905 (((T_ABSTRACT) reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_ARRAY) reduce 475) ((T_AS) reduce 475) ((T_BREAK) reduce 475) ((T_CALLABLE) reduce 475) ((T_CASE) reduce 475) ((T_CATCH) reduce 475) ((T_CLASS) reduce 475) ((T_CLASS_C) reduce 475) ((T_CLONE) reduce 475) ((T_CONST) reduce 475) ((T_CONTINUE) reduce 475) ((T_DECLARE) reduce 475) ((T_DEFAULT) reduce 475) ((T_DIR) reduce 475) ((T_DO) reduce 475) ((T_ECHO) reduce 475) ((T_ELSE) reduce 475) ((T_ELSEIF) reduce 475) ((T_EMPTY) reduce 475) ((T_ENDDECLARE) reduce 475) ((T_ENDFOR) reduce 475) ((T_ENDFOREACH) reduce 475) ((T_ENDIF) reduce 475) ((T_ENDSWITCH) reduce 475) ((T_ENDWHILE) reduce 475) ((T_ENUM) reduce 475) ((T_EVAL) reduce 475) ((T_EXIT) reduce 475) ((T_EXTENDS) reduce 475) ((T_FILE) reduce 475) ((T_FINAL) reduce 475) ((T_FINALLY) reduce 475) ((T_FN) reduce 475) ((T_FOR) reduce 475) ((T_FOREACH) reduce 475) ((T_FUNCTION) reduce 475) ((T_FUNC_C) reduce 475) ((T_GLOBAL) reduce 475) ((T_GOTO) reduce 475) ((T_IF) reduce 475) ((T_IMPLEMENTS) reduce 475) ((T_INCLUDE) reduce 475) ((T_INCLUDE_ONCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_INSTEADOF) reduce 475) ((T_INTERFACE) reduce 475) ((T_ISSET) reduce 475) ((T_LINE) reduce 475) ((T_LIST) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_MATCH) reduce 475) ((T_METHOD_C) reduce 475) ((T_NAMESPACE) reduce 475) ((T_NEW) reduce 475) ((T_NS_C) reduce 475) ((T_PRINT) reduce 475) ((T_PRIVATE) reduce 475) ((T_PROPERTY_C) reduce 475) ((T_PROTECTED) reduce 475) ((T_PUBLIC) reduce 475) ((T_READONLY) reduce 475) ((T_REQUIRE) reduce 475) ((T_REQUIRE_ONCE) reduce 475) ((T_RETURN) reduce 475) ((T_STATIC) reduce 475) ((T_STRING) reduce 475) ((T_SWITCH) reduce 475) ((T_THROW) reduce 475) ((T_TRAIT) reduce 475) ((T_TRAIT_C) reduce 475) ((T_TRY) reduce 475) ((T_UNSET) reduce 475) ((T_USE) reduce 475) ((T_VAR) reduce 475) ((T_WHILE) reduce 475) ((T_YIELD) reduce 475)) 6906 (((T_ABSTRACT) reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) reduce 479) ((T_AS) reduce 479) ((T_BREAK) reduce 479) ((T_CALLABLE) reduce 479) ((T_CASE) reduce 479) ((T_CATCH) reduce 479) ((T_CLASS) reduce 479) ((T_CLASS_C) reduce 479) ((T_CLONE) reduce 479) ((T_CONST) reduce 479) ((T_CONTINUE) reduce 479) ((T_DECLARE) reduce 479) ((T_DEFAULT) reduce 479) ((T_DIR) reduce 479) ((T_DO) reduce 479) ((T_ECHO) reduce 479) ((T_ELSE) reduce 479) ((T_ELSEIF) reduce 479) ((T_EMPTY) reduce 479) ((T_ENDDECLARE) reduce 479) ((T_ENDFOR) reduce 479) ((T_ENDFOREACH) reduce 479) ((T_ENDIF) reduce 479) ((T_ENDSWITCH) reduce 479) ((T_ENDWHILE) reduce 479) ((T_ENUM) reduce 479) ((T_EVAL) reduce 479) ((T_EXIT) reduce 479) ((T_EXTENDS) reduce 479) ((T_FILE) reduce 479) ((T_FINAL) reduce 479) ((T_FINALLY) reduce 479) ((T_FN) reduce 479) ((T_FOR) reduce 479) ((T_FOREACH) reduce 479) ((T_FUNCTION) reduce 479) ((T_FUNC_C) reduce 479) ((T_GLOBAL) reduce 479) ((T_GOTO) reduce 479) ((T_IF) reduce 479) ((T_IMPLEMENTS) reduce 479) ((T_INCLUDE) reduce 479) ((T_INCLUDE_ONCE) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_INSTEADOF) reduce 479) ((T_INTERFACE) reduce 479) ((T_ISSET) reduce 479) ((T_LINE) reduce 479) ((T_LIST) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) reduce 479) ((T_METHOD_C) reduce 479) ((T_NAMESPACE) reduce 479) ((T_NEW) reduce 479) ((T_NS_C) reduce 479) ((T_PRINT) reduce 479) ((T_PRIVATE) reduce 479) ((T_PROPERTY_C) reduce 479) ((T_PROTECTED) reduce 479) ((T_PUBLIC) reduce 479) ((T_READONLY) reduce 479) ((T_REQUIRE) reduce 479) ((T_REQUIRE_ONCE) reduce 479) ((T_RETURN) reduce 479) ((T_STATIC) reduce 479) ((T_STRING) reduce 479) ((T_SWITCH) reduce 479) ((T_THROW) reduce 479) ((T_TRAIT) reduce 479) ((T_TRAIT_C) reduce 479) ((T_TRY) reduce 479) ((T_UNSET) reduce 479) ((T_USE) reduce 479) ((T_VAR) reduce 479) ((T_WHILE) reduce 479) ((T_YIELD) reduce 479)) 6907 (((T_ABSTRACT) reduce 80) ((T_ARRAY) reduce 80) ((T_AS) reduce 80) ((T_BREAK) reduce 80) ((T_CALLABLE) reduce 80) ((T_CASE) reduce 80) ((T_CATCH) reduce 80) ((T_CLASS) reduce 80) ((T_CLASS_C) reduce 80) ((T_CLONE) reduce 80) ((T_CONST) reduce 80) ((T_CONTINUE) reduce 80) ((T_DECLARE) reduce 80) ((T_DEFAULT) reduce 80) ((T_DIR) reduce 80) ((T_DO) reduce 80) ((T_ECHO) reduce 80) ((T_ELSE) reduce 80) ((T_ELSEIF) reduce 80) ((T_EMPTY) reduce 80) ((T_ENDDECLARE) reduce 80) ((T_ENDFOR) reduce 80) ((T_ENDFOREACH) reduce 80) ((T_ENDIF) reduce 80) ((T_ENDSWITCH) reduce 80) ((T_ENDWHILE) reduce 80) ((T_ENUM) reduce 80) ((T_EVAL) reduce 80) ((T_EXIT) reduce 80) ((T_EXTENDS) reduce 80) ((T_FILE) reduce 80) ((T_FINAL) reduce 80) ((T_FINALLY) reduce 80) ((T_FN) reduce 80) ((T_FOR) reduce 80) ((T_FOREACH) reduce 80) ((T_FUNCTION) reduce 80) ((T_FUNC_C) reduce 80) ((T_GLOBAL) reduce 80) ((T_GOTO) reduce 80) ((T_IF) reduce 80) ((T_IMPLEMENTS) reduce 80) ((T_INCLUDE) reduce 80) ((T_INCLUDE_ONCE) reduce 80) ((T_INSTANCEOF) reduce 80) ((T_INSTEADOF) reduce 80) ((T_INTERFACE) reduce 80) ((T_ISSET) reduce 80) ((T_LINE) reduce 80) ((T_LIST) reduce 80) ((T_LOGICAL_AND) reduce 80) ((T_LOGICAL_OR) reduce 80) ((T_LOGICAL_XOR) reduce 80) ((T_MATCH) reduce 80) ((T_METHOD_C) reduce 80) ((T_NAMESPACE) reduce 80) ((T_NEW) reduce 80) ((T_NS_C) reduce 80) ((T_PRINT) reduce 80) ((T_PRIVATE) reduce 80) ((T_PROPERTY_C) reduce 80) ((T_PROTECTED) reduce 80) ((T_PUBLIC) reduce 80) ((T_READONLY) reduce 80) ((T_REQUIRE) reduce 80) ((T_REQUIRE_ONCE) reduce 80) ((T_RETURN) reduce 80) ((T_STATIC) reduce 80) ((T_STRING) reduce 80) ((T_SWITCH) reduce 80) ((T_THROW) reduce 80) ((T_TRAIT) reduce 80) ((T_TRAIT_C) reduce 80) ((T_TRY) reduce 80) ((T_UNSET) reduce 80) ((T_USE) reduce 80) ((T_VAR) reduce 80) ((T_WHILE) reduce 80) ((T_YIELD) reduce 80)) 6908 (((T_ABSTRACT) reduce 81) ((T_ARRAY) reduce 81) ((T_AS) reduce 81) ((T_BREAK) reduce 81) ((T_CALLABLE) reduce 81) ((T_CASE) reduce 81) ((T_CATCH) reduce 81) ((T_CLASS) reduce 81) ((T_CLASS_C) reduce 81) ((T_CLONE) reduce 81) ((T_CONST) reduce 81) ((T_CONTINUE) reduce 81) ((T_DECLARE) reduce 81) ((T_DEFAULT) reduce 81) ((T_DIR) reduce 81) ((T_DO) reduce 81) ((T_ECHO) reduce 81) ((T_ELSE) reduce 81) ((T_ELSEIF) reduce 81) ((T_EMPTY) reduce 81) ((T_ENDDECLARE) reduce 81) ((T_ENDFOR) reduce 81) ((T_ENDFOREACH) reduce 81) ((T_ENDIF) reduce 81) ((T_ENDSWITCH) reduce 81) ((T_ENDWHILE) reduce 81) ((T_ENUM) reduce 81) ((T_EVAL) reduce 81) ((T_EXIT) reduce 81) ((T_EXTENDS) reduce 81) ((T_FILE) reduce 81) ((T_FINAL) reduce 81) ((T_FINALLY) reduce 81) ((T_FN) reduce 81) ((T_FOR) reduce 81) ((T_FOREACH) reduce 81) ((T_FUNCTION) reduce 81) ((T_FUNC_C) reduce 81) ((T_GLOBAL) reduce 81) ((T_GOTO) reduce 81) ((T_IF) reduce 81) ((T_IMPLEMENTS) reduce 81) ((T_INCLUDE) reduce 81) ((T_INCLUDE_ONCE) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_INSTEADOF) reduce 81) ((T_INTERFACE) reduce 81) ((T_ISSET) reduce 81) ((T_LINE) reduce 81) ((T_LIST) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_MATCH) reduce 81) ((T_METHOD_C) reduce 81) ((T_NAMESPACE) reduce 81) ((T_NEW) reduce 81) ((T_NS_C) reduce 81) ((T_PRINT) reduce 81) ((T_PRIVATE) reduce 81) ((T_PROPERTY_C) reduce 81) ((T_PROTECTED) reduce 81) ((T_PUBLIC) reduce 81) ((T_READONLY) reduce 81) ((T_REQUIRE) reduce 81) ((T_REQUIRE_ONCE) reduce 81) ((T_RETURN) reduce 81) ((T_STATIC) reduce 81) ((T_STRING) reduce 81) ((T_SWITCH) reduce 81) ((T_THROW) reduce 81) ((T_TRAIT) reduce 81) ((T_TRAIT_C) reduce 81) ((T_TRY) reduce 81) ((T_UNSET) reduce 81) ((T_USE) reduce 81) ((T_VAR) reduce 81) ((T_WHILE) reduce 81) ((T_YIELD) reduce 81)) 6909 (((T_ABSTRACT) reduce 480) ((T_ARRAY) reduce 480) ((T_AS) reduce 480) ((T_BREAK) reduce 480) ((T_CALLABLE) reduce 480) ((T_CASE) reduce 480) ((T_CATCH) reduce 480) ((T_CLASS) reduce 480) ((T_CLASS_C) reduce 480) ((T_CLONE) reduce 480) ((T_CONST) reduce 480) ((T_CONTINUE) reduce 480) ((T_DECLARE) reduce 480) ((T_DEFAULT) reduce 480) ((T_DIR) reduce 480) ((T_DO) reduce 480) ((T_ECHO) reduce 480) ((T_ELSE) reduce 480) ((T_ELSEIF) reduce 480) ((T_EMPTY) reduce 480) ((T_ENDDECLARE) reduce 480) ((T_ENDFOR) reduce 480) ((T_ENDFOREACH) reduce 480) ((T_ENDIF) reduce 480) ((T_ENDSWITCH) reduce 480) ((T_ENDWHILE) reduce 480) ((T_ENUM) reduce 480) ((T_EVAL) reduce 480) ((T_EXIT) reduce 480) ((T_EXTENDS) reduce 480) ((T_FILE) reduce 480) ((T_FINAL) reduce 480) ((T_FINALLY) reduce 480) ((T_FN) reduce 480) ((T_FOR) reduce 480) ((T_FOREACH) reduce 480) ((T_FUNCTION) reduce 480) ((T_FUNC_C) reduce 480) ((T_GLOBAL) reduce 480) ((T_GOTO) reduce 480) ((T_IF) reduce 480) ((T_IMPLEMENTS) reduce 480) ((T_INCLUDE) reduce 480) ((T_INCLUDE_ONCE) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_INSTEADOF) reduce 480) ((T_INTERFACE) reduce 480) ((T_ISSET) reduce 480) ((T_LINE) reduce 480) ((T_LIST) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_MATCH) reduce 480) ((T_METHOD_C) reduce 480) ((T_NAMESPACE) reduce 480) ((T_NEW) reduce 480) ((T_NS_C) reduce 480) ((T_PRINT) reduce 480) ((T_PRIVATE) reduce 480) ((T_PROPERTY_C) reduce 480) ((T_PROTECTED) reduce 480) ((T_PUBLIC) reduce 480) ((T_READONLY) reduce 480) ((T_REQUIRE) reduce 480) ((T_REQUIRE_ONCE) reduce 480) ((T_RETURN) reduce 480) ((T_STATIC) reduce 480) ((T_STRING) reduce 480) ((T_SWITCH) reduce 480) ((T_THROW) reduce 480) ((T_TRAIT) reduce 480) ((T_TRAIT_C) reduce 480) ((T_TRY) reduce 480) ((T_UNSET) reduce 480) ((T_USE) reduce 480) ((T_VAR) reduce 480) ((T_WHILE) reduce 480) ((T_YIELD) reduce 480)) 6910 (((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift)) 6911 ((("(") reduce 74)) 6912 ((("(") reduce 52)) 6913 ((("(") reduce 27)) 6914 ((("(") reduce 51)) 6915 ((("(") reduce 53)) 6916 ((("(") reduce 49)) 6917 ((("(") reduce 29)) 6918 ((("(") reduce 59)) 6919 ((("(") reduce 60)) 6920 ((("(") reduce 11)) 6921 ((("(") reduce 42)) 6922 ((("(") reduce 39)) 6923 ((("(") reduce 25)) 6924 ((("(") reduce 50)) 6925 ((("(") reduce 66)) 6926 ((("(") reduce 18)) 6927 ((("(") reduce 17)) 6928 ((("(") reduce 15)) 6929 ((("(") reduce 14)) 6930 ((("(") reduce 38)) 6931 ((("(") reduce 26)) 6932 ((("(") reduce 22)) 6933 ((("(") reduce 24)) 6934 ((("(") reduce 16)) 6935 ((("(") reduce 48)) 6936 ((("(") reduce 20)) 6937 ((("(") reduce 70)) 6938 ((("(") reduce 3)) 6939 ((("(") reduce 12)) 6940 ((("(") reduce 54)) 6941 ((("(") reduce 65)) 6942 ((("(") reduce 75)) 6943 ((("(") reduce 30)) 6944 ((("(") reduce 68)) 6945 ((("(") reduce 21)) 6946 ((("(") reduce 23)) 6947 ((("(") reduce 41)) 6948 ((("(") reduce 62)) 6949 ((("(") reduce 34)) 6950 ((("(") reduce 40)) 6951 ((("(") reduce 13)) 6952 ((("(") reduce 55)) 6953 ((("(") reduce 1)) 6954 ((("(") reduce 2)) 6955 ((("(") reduce 9)) 6956 ((("(") reduce 33)) 6957 ((("(") reduce 58)) 6958 ((("(") reduce 37)) 6959 ((("(") reduce 64)) 6960 ((("(") reduce 46)) 6961 ((("(") reduce 8)) 6962 ((("(") reduce 6)) 6963 ((("(") reduce 7)) 6964 ((("(") reduce 69)) 6965 ((("(") reduce 63)) 6966 ((("(") reduce 56)) 6967 ((("(") reduce 10)) 6968 ((("(") reduce 67)) 6969 ((("(") reduce 44)) 6970 ((("(") reduce 76)) 6971 ((("(") reduce 71)) 6972 ((("(") reduce 77)) 6973 ((("(") reduce 78)) 6974 ((("(") reduce 79)) 6975 ((("(") reduce 4)) 6976 ((("(") reduce 5)) 6977 ((("(") reduce 43)) 6978 ((("(") reduce 73)) 6979 ((("(") reduce 82)) 6980 ((("(") reduce 47)) 6981 ((("(") reduce 31)) 6982 ((("(") reduce 57)) 6983 ((("(") reduce 61)) 6984 ((("(") reduce 28)) 6985 ((("(") reduce 36)) 6986 ((("(") reduce 32)) 6987 ((("(") reduce 35)) 6988 ((("(") reduce 19)) 6989 ((("(") reduce 45)) 6991 ((("(") reduce 72)) 6992 ((("(") reduce 83)) 6996 (((":") shift) ((";") reduce 287) (("{") reduce 287)) 6998 (((";") reduce 477) (("{") reduce 477)) 6999 (((";") shift) (("{") shift)) 7000 (((T_ABSTRACT) reduce 332) ((T_ATTRIBUTE) reduce 332) ((T_CASE) reduce 332) ((T_CONST) reduce 332) ((T_FINAL) reduce 332) ((T_FUNCTION) reduce 332) ((T_PRIVATE) reduce 332) ((T_PRIVATE_SET) reduce 332) ((T_PROTECTED) reduce 332) ((T_PROTECTED_SET) reduce 332) ((T_PUBLIC) reduce 332) ((T_PUBLIC_SET) reduce 332) ((T_READONLY) reduce 332) ((T_STATIC) reduce 332) ((T_USE) reduce 332) ((T_VAR) reduce 332) (("}") reduce 332)) 7001 (((T_ABSTRACT) reduce 477) ((T_ATTRIBUTE) reduce 477) ((T_CASE) reduce 477) ((T_CONST) reduce 477) ((T_FINAL) reduce 477) ((T_FUNCTION) reduce 477) ((T_PRIVATE) reduce 477) ((T_PRIVATE_SET) reduce 477) ((T_PROTECTED) reduce 477) ((T_PROTECTED_SET) reduce 477) ((T_PUBLIC) reduce 477) ((T_PUBLIC_SET) reduce 477) ((T_READONLY) reduce 477) ((T_STATIC) reduce 477) ((T_USE) reduce 477) ((T_VAR) reduce 477) (("}") reduce 477)) 7004 (((T_ABSTRACT) reduce 333) ((T_ATTRIBUTE) reduce 333) ((T_CASE) reduce 333) ((T_CONST) reduce 333) ((T_FINAL) reduce 333) ((T_FUNCTION) reduce 333) ((T_PRIVATE) reduce 333) ((T_PRIVATE_SET) reduce 333) ((T_PROTECTED) reduce 333) ((T_PROTECTED_SET) reduce 333) ((T_PUBLIC) reduce 333) ((T_PUBLIC_SET) reduce 333) ((T_READONLY) reduce 333) ((T_STATIC) reduce 333) ((T_USE) reduce 333) ((T_VAR) reduce 333) (("}") reduce 333)) 7005 (((T_ABSTRACT) reduce 310) ((T_ATTRIBUTE) reduce 310) ((T_CASE) reduce 310) ((T_CONST) reduce 310) ((T_FINAL) reduce 310) ((T_FUNCTION) reduce 310) ((T_PRIVATE) reduce 310) ((T_PRIVATE_SET) reduce 310) ((T_PROTECTED) reduce 310) ((T_PROTECTED_SET) reduce 310) ((T_PUBLIC) reduce 310) ((T_PUBLIC_SET) reduce 310) ((T_READONLY) reduce 310) ((T_STATIC) reduce 310) ((T_USE) reduce 310) ((T_VAR) reduce 310) (("}") reduce 310)) 7007 (((";") reduce 278) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) (("{") reduce 278) (("|") reduce 278)) 7008 (((";") reduce 279) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) (("{") reduce 279) (("|") reduce 279)) 7009 (((";") reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) (("{") reduce 94) (("|") reduce 94)) 7010 (((";") reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) (("{") reduce 93) (("|") reduce 93)) 7011 (((";") reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) (("{") reduce 95) (("|") reduce 95)) 7012 (((";") reduce 267) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267) (("{") reduce 267) (("|") reduce 267)) 7013 (((";") reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) (("{") reduce 92) (("|") reduce 92)) 7014 (((";") reduce 265) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) (("{") reduce 265)) 7015 (((";") reduce 280) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) (("{") reduce 280) (("|") reduce 280)) 7016 (((";") reduce 262) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) (("{") reduce 262) (("|") reduce 268)) 7017 (((";") reduce 288) (("{") reduce 288)) 7018 (((";") reduce 266) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266) (("{") reduce 266) (("|") reduce 266)) 7019 (((";") reduce 264) (("{") reduce 264) (("|") shift)) 7023 (((";") reduce 278) (("{") reduce 278) (("|") reduce 278)) 7024 (((";") reduce 279) (("{") reduce 279) (("|") reduce 279)) 7025 (((";") reduce 94) (("{") reduce 94) (("|") reduce 94)) 7026 (((";") reduce 93) (("{") reduce 93) (("|") reduce 93)) 7027 (((";") reduce 95) (("{") reduce 95) (("|") reduce 95)) 7028 (((";") reduce 267) (("{") reduce 267) (("|") reduce 267)) 7029 (((";") reduce 92) (("{") reduce 92) (("|") reduce 92)) 7030 (((";") reduce 280) (("{") reduce 280) (("|") reduce 280)) 7031 (((";") reduce 268) (("{") reduce 268) (("|") reduce 268)) 7032 (((";") reduce 266) (("{") reduce 266) (("|") reduce 266)) 7033 (((";") reduce 270) (("{") reduce 270) (("|") reduce 270)) 7035 (((";") reduce 269) (("{") reduce 269) (("|") reduce 269)) 7037 (((";") reduce 271) (("{") reduce 271) (("|") reduce 271)) 7039 (((";") reduce 278) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) (("{") reduce 278)) 7040 (((";") reduce 279) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) (("{") reduce 279)) 7041 (((";") reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) (("{") reduce 94)) 7042 (((";") reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) (("{") reduce 93)) 7043 (((";") reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) (("{") reduce 95)) 7044 (((";") reduce 267) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267) (("{") reduce 267)) 7045 (((";") reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) (("{") reduce 92)) 7046 (((";") reduce 280) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) (("{") reduce 280)) 7047 (((";") reduce 272) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) (("{") reduce 272)) 7048 (((";") reduce 266) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266) (("{") reduce 266)) 7050 (((";") reduce 273) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) (("{") reduce 273)) 7051 (((";") reduce 278) (("{") reduce 278)) 7052 (((";") reduce 279) (("{") reduce 279)) 7053 (((";") reduce 94) (("{") reduce 94)) 7054 (((";") reduce 93) (("{") reduce 93)) 7055 (((";") reduce 95) (("{") reduce 95)) 7056 (((";") reduce 267) (("{") reduce 267)) 7057 (((";") reduce 92) (("{") reduce 92)) 7058 (((";") reduce 280) (("{") reduce 280)) 7059 (((";") reduce 263) (("{") reduce 263)) 7060 (((";") reduce 266) (("{") reduce 266)) 7061 ((("(") shift) (("?") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift)) 7063 ((("=") reduce 74)) 7064 ((("=") reduce 52) ((T_ABSTRACT) reduce 278) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_ARRAY) reduce 278) ((T_AS) reduce 278) ((T_BREAK) reduce 278) ((T_CALLABLE) reduce 278) ((T_CASE) reduce 278) ((T_CATCH) reduce 278) ((T_CLASS) reduce 278) ((T_CLASS_C) reduce 278) ((T_CLONE) reduce 278) ((T_CONST) reduce 278) ((T_CONTINUE) reduce 278) ((T_DECLARE) reduce 278) ((T_DEFAULT) reduce 278) ((T_DIR) reduce 278) ((T_DO) reduce 278) ((T_ECHO) reduce 278) ((T_ELSE) reduce 278) ((T_ELSEIF) reduce 278) ((T_EMPTY) reduce 278) ((T_ENDDECLARE) reduce 278) ((T_ENDFOR) reduce 278) ((T_ENDFOREACH) reduce 278) ((T_ENDIF) reduce 278) ((T_ENDSWITCH) reduce 278) ((T_ENDWHILE) reduce 278) ((T_ENUM) reduce 278) ((T_EVAL) reduce 278) ((T_EXIT) reduce 278) ((T_EXTENDS) reduce 278) ((T_FILE) reduce 278) ((T_FINAL) reduce 278) ((T_FINALLY) reduce 278) ((T_FN) reduce 278) ((T_FOR) reduce 278) ((T_FOREACH) reduce 278) ((T_FUNCTION) reduce 278) ((T_FUNC_C) reduce 278) ((T_GLOBAL) reduce 278) ((T_GOTO) reduce 278) ((T_IF) reduce 278) ((T_IMPLEMENTS) reduce 278) ((T_INCLUDE) reduce 278) ((T_INCLUDE_ONCE) reduce 278) ((T_INSTANCEOF) reduce 278) ((T_INSTEADOF) reduce 278) ((T_INTERFACE) reduce 278) ((T_ISSET) reduce 278) ((T_LINE) reduce 278) ((T_LIST) reduce 278) ((T_LOGICAL_AND) reduce 278) ((T_LOGICAL_OR) reduce 278) ((T_LOGICAL_XOR) reduce 278) ((T_MATCH) reduce 278) ((T_METHOD_C) reduce 278) ((T_NAMESPACE) reduce 278) ((T_NEW) reduce 278) ((T_NS_C) reduce 278) ((T_PRINT) reduce 278) ((T_PRIVATE) reduce 278) ((T_PROPERTY_C) reduce 278) ((T_PROTECTED) reduce 278) ((T_PUBLIC) reduce 278) ((T_READONLY) reduce 278) ((T_REQUIRE) reduce 278) ((T_REQUIRE_ONCE) reduce 278) ((T_RETURN) reduce 278) ((T_STATIC) reduce 278) ((T_STRING) reduce 278) ((T_SWITCH) reduce 278) ((T_THROW) reduce 278) ((T_TRAIT) reduce 278) ((T_TRAIT_C) reduce 278) ((T_TRY) reduce 278) ((T_UNSET) reduce 278) ((T_USE) reduce 278) ((T_VAR) reduce 278) ((T_WHILE) reduce 278) ((T_YIELD) reduce 278) (("|") reduce 278)) 7065 ((("=") reduce 27)) 7066 ((("=") reduce 51)) 7067 ((("=") reduce 53) ((T_ABSTRACT) reduce 279) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) ((T_ARRAY) reduce 279) ((T_AS) reduce 279) ((T_BREAK) reduce 279) ((T_CALLABLE) reduce 279) ((T_CASE) reduce 279) ((T_CATCH) reduce 279) ((T_CLASS) reduce 279) ((T_CLASS_C) reduce 279) ((T_CLONE) reduce 279) ((T_CONST) reduce 279) ((T_CONTINUE) reduce 279) ((T_DECLARE) reduce 279) ((T_DEFAULT) reduce 279) ((T_DIR) reduce 279) ((T_DO) reduce 279) ((T_ECHO) reduce 279) ((T_ELSE) reduce 279) ((T_ELSEIF) reduce 279) ((T_EMPTY) reduce 279) ((T_ENDDECLARE) reduce 279) ((T_ENDFOR) reduce 279) ((T_ENDFOREACH) reduce 279) ((T_ENDIF) reduce 279) ((T_ENDSWITCH) reduce 279) ((T_ENDWHILE) reduce 279) ((T_ENUM) reduce 279) ((T_EVAL) reduce 279) ((T_EXIT) reduce 279) ((T_EXTENDS) reduce 279) ((T_FILE) reduce 279) ((T_FINAL) reduce 279) ((T_FINALLY) reduce 279) ((T_FN) reduce 279) ((T_FOR) reduce 279) ((T_FOREACH) reduce 279) ((T_FUNCTION) reduce 279) ((T_FUNC_C) reduce 279) ((T_GLOBAL) reduce 279) ((T_GOTO) reduce 279) ((T_IF) reduce 279) ((T_IMPLEMENTS) reduce 279) ((T_INCLUDE) reduce 279) ((T_INCLUDE_ONCE) reduce 279) ((T_INSTANCEOF) reduce 279) ((T_INSTEADOF) reduce 279) ((T_INTERFACE) reduce 279) ((T_ISSET) reduce 279) ((T_LINE) reduce 279) ((T_LIST) reduce 279) ((T_LOGICAL_AND) reduce 279) ((T_LOGICAL_OR) reduce 279) ((T_LOGICAL_XOR) reduce 279) ((T_MATCH) reduce 279) ((T_METHOD_C) reduce 279) ((T_NAMESPACE) reduce 279) ((T_NEW) reduce 279) ((T_NS_C) reduce 279) ((T_PRINT) reduce 279) ((T_PRIVATE) reduce 279) ((T_PROPERTY_C) reduce 279) ((T_PROTECTED) reduce 279) ((T_PUBLIC) reduce 279) ((T_READONLY) reduce 279) ((T_REQUIRE) reduce 279) ((T_REQUIRE_ONCE) reduce 279) ((T_RETURN) reduce 279) ((T_STATIC) reduce 279) ((T_STRING) reduce 279) ((T_SWITCH) reduce 279) ((T_THROW) reduce 279) ((T_TRAIT) reduce 279) ((T_TRAIT_C) reduce 279) ((T_TRY) reduce 279) ((T_UNSET) reduce 279) ((T_USE) reduce 279) ((T_VAR) reduce 279) ((T_WHILE) reduce 279) ((T_YIELD) reduce 279) (("|") reduce 279)) 7068 ((("=") reduce 49)) 7069 ((("=") reduce 29)) 7070 ((("=") reduce 59)) 7071 ((("=") reduce 60)) 7072 ((("=") reduce 11)) 7073 ((("=") reduce 42)) 7074 ((("=") reduce 39)) 7075 ((("=") reduce 25)) 7076 ((("=") reduce 50)) 7077 ((("=") reduce 66)) 7078 ((("=") reduce 18)) 7079 ((("=") reduce 17)) 7080 ((("=") reduce 15)) 7081 ((("=") reduce 14)) 7082 ((("=") reduce 38)) 7083 ((("=") reduce 26)) 7084 ((("=") reduce 22)) 7085 ((("=") reduce 24)) 7086 ((("=") reduce 16)) 7087 ((("=") reduce 48)) 7088 ((("=") reduce 20)) 7089 ((("=") reduce 70)) 7090 ((("=") reduce 3)) 7091 ((("=") reduce 12)) 7092 ((("=") reduce 54)) 7093 ((("=") reduce 65)) 7094 ((("=") reduce 75)) 7095 ((("=") reduce 30)) 7096 ((("=") reduce 68)) 7097 ((("=") reduce 21)) 7098 ((("=") reduce 23)) 7099 ((("=") reduce 41)) 7100 ((("=") reduce 62)) 7101 ((("=") reduce 34)) 7102 ((("=") reduce 40)) 7103 ((("=") reduce 13)) 7104 ((("=") reduce 55)) 7105 ((("=") reduce 1)) 7106 ((("=") reduce 2)) 7107 ((("=") reduce 9)) 7108 ((("=") reduce 33)) 7109 ((("=") reduce 58)) 7110 ((("=") reduce 37)) 7111 ((("=") reduce 64)) 7112 ((("=") reduce 46)) 7113 ((("=") reduce 8)) 7114 ((("=") reduce 6)) 7115 ((("=") reduce 7)) 7116 ((("=") reduce 69)) 7117 ((("=") reduce 63)) 7118 ((("=") reduce 56)) 7119 (((T_ABSTRACT) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ARRAY) reduce 94) ((T_AS) reduce 94) ((T_BREAK) reduce 94) ((T_CALLABLE) reduce 94) ((T_CASE) reduce 94) ((T_CATCH) reduce 94) ((T_CLASS) reduce 94) ((T_CLASS_C) reduce 94) ((T_CLONE) reduce 94) ((T_CONST) reduce 94) ((T_CONTINUE) reduce 94) ((T_DECLARE) reduce 94) ((T_DEFAULT) reduce 94) ((T_DIR) reduce 94) ((T_DO) reduce 94) ((T_ECHO) reduce 94) ((T_ELSE) reduce 94) ((T_ELSEIF) reduce 94) ((T_EMPTY) reduce 94) ((T_ENDDECLARE) reduce 94) ((T_ENDFOR) reduce 94) ((T_ENDFOREACH) reduce 94) ((T_ENDIF) reduce 94) ((T_ENDSWITCH) reduce 94) ((T_ENDWHILE) reduce 94) ((T_ENUM) reduce 94) ((T_EVAL) reduce 94) ((T_EXIT) reduce 94) ((T_EXTENDS) reduce 94) ((T_FILE) reduce 94) ((T_FINAL) reduce 94) ((T_FINALLY) reduce 94) ((T_FN) reduce 94) ((T_FOR) reduce 94) ((T_FOREACH) reduce 94) ((T_FUNCTION) reduce 94) ((T_FUNC_C) reduce 94) ((T_GLOBAL) reduce 94) ((T_GOTO) reduce 94) ((T_IF) reduce 94) ((T_IMPLEMENTS) reduce 94) ((T_INCLUDE) reduce 94) ((T_INCLUDE_ONCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_INSTEADOF) reduce 94) ((T_INTERFACE) reduce 94) ((T_ISSET) reduce 94) ((T_LINE) reduce 94) ((T_LIST) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_MATCH) reduce 94) ((T_METHOD_C) reduce 94) ((T_NAMESPACE) reduce 94) ((T_NEW) reduce 94) ((T_NS_C) reduce 94) ((T_PRINT) reduce 94) ((T_PRIVATE) reduce 94) ((T_PROPERTY_C) reduce 94) ((T_PROTECTED) reduce 94) ((T_PUBLIC) reduce 94) ((T_READONLY) reduce 94) ((T_REQUIRE) reduce 94) ((T_REQUIRE_ONCE) reduce 94) ((T_RETURN) reduce 94) ((T_STATIC) reduce 94) ((T_STRING) reduce 94) ((T_SWITCH) reduce 94) ((T_THROW) reduce 94) ((T_TRAIT) reduce 94) ((T_TRAIT_C) reduce 94) ((T_TRY) reduce 94) ((T_UNSET) reduce 94) ((T_USE) reduce 94) ((T_VAR) reduce 94) ((T_WHILE) reduce 94) ((T_YIELD) reduce 94) (("|") reduce 94)) 7120 (((T_ABSTRACT) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ARRAY) reduce 93) ((T_AS) reduce 93) ((T_BREAK) reduce 93) ((T_CALLABLE) reduce 93) ((T_CASE) reduce 93) ((T_CATCH) reduce 93) ((T_CLASS) reduce 93) ((T_CLASS_C) reduce 93) ((T_CLONE) reduce 93) ((T_CONST) reduce 93) ((T_CONTINUE) reduce 93) ((T_DECLARE) reduce 93) ((T_DEFAULT) reduce 93) ((T_DIR) reduce 93) ((T_DO) reduce 93) ((T_ECHO) reduce 93) ((T_ELSE) reduce 93) ((T_ELSEIF) reduce 93) ((T_EMPTY) reduce 93) ((T_ENDDECLARE) reduce 93) ((T_ENDFOR) reduce 93) ((T_ENDFOREACH) reduce 93) ((T_ENDIF) reduce 93) ((T_ENDSWITCH) reduce 93) ((T_ENDWHILE) reduce 93) ((T_ENUM) reduce 93) ((T_EVAL) reduce 93) ((T_EXIT) reduce 93) ((T_EXTENDS) reduce 93) ((T_FILE) reduce 93) ((T_FINAL) reduce 93) ((T_FINALLY) reduce 93) ((T_FN) reduce 93) ((T_FOR) reduce 93) ((T_FOREACH) reduce 93) ((T_FUNCTION) reduce 93) ((T_FUNC_C) reduce 93) ((T_GLOBAL) reduce 93) ((T_GOTO) reduce 93) ((T_IF) reduce 93) ((T_IMPLEMENTS) reduce 93) ((T_INCLUDE) reduce 93) ((T_INCLUDE_ONCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_INSTEADOF) reduce 93) ((T_INTERFACE) reduce 93) ((T_ISSET) reduce 93) ((T_LINE) reduce 93) ((T_LIST) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_MATCH) reduce 93) ((T_METHOD_C) reduce 93) ((T_NAMESPACE) reduce 93) ((T_NEW) reduce 93) ((T_NS_C) reduce 93) ((T_PRINT) reduce 93) ((T_PRIVATE) reduce 93) ((T_PROPERTY_C) reduce 93) ((T_PROTECTED) reduce 93) ((T_PUBLIC) reduce 93) ((T_READONLY) reduce 93) ((T_REQUIRE) reduce 93) ((T_REQUIRE_ONCE) reduce 93) ((T_RETURN) reduce 93) ((T_STATIC) reduce 93) ((T_STRING) reduce 93) ((T_SWITCH) reduce 93) ((T_THROW) reduce 93) ((T_TRAIT) reduce 93) ((T_TRAIT_C) reduce 93) ((T_TRY) reduce 93) ((T_UNSET) reduce 93) ((T_USE) reduce 93) ((T_VAR) reduce 93) ((T_WHILE) reduce 93) ((T_YIELD) reduce 93) (("|") reduce 93)) 7121 (((T_ABSTRACT) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_ARRAY) reduce 95) ((T_AS) reduce 95) ((T_BREAK) reduce 95) ((T_CALLABLE) reduce 95) ((T_CASE) reduce 95) ((T_CATCH) reduce 95) ((T_CLASS) reduce 95) ((T_CLASS_C) reduce 95) ((T_CLONE) reduce 95) ((T_CONST) reduce 95) ((T_CONTINUE) reduce 95) ((T_DECLARE) reduce 95) ((T_DEFAULT) reduce 95) ((T_DIR) reduce 95) ((T_DO) reduce 95) ((T_ECHO) reduce 95) ((T_ELSE) reduce 95) ((T_ELSEIF) reduce 95) ((T_EMPTY) reduce 95) ((T_ENDDECLARE) reduce 95) ((T_ENDFOR) reduce 95) ((T_ENDFOREACH) reduce 95) ((T_ENDIF) reduce 95) ((T_ENDSWITCH) reduce 95) ((T_ENDWHILE) reduce 95) ((T_ENUM) reduce 95) ((T_EVAL) reduce 95) ((T_EXIT) reduce 95) ((T_EXTENDS) reduce 95) ((T_FILE) reduce 95) ((T_FINAL) reduce 95) ((T_FINALLY) reduce 95) ((T_FN) reduce 95) ((T_FOR) reduce 95) ((T_FOREACH) reduce 95) ((T_FUNCTION) reduce 95) ((T_FUNC_C) reduce 95) ((T_GLOBAL) reduce 95) ((T_GOTO) reduce 95) ((T_IF) reduce 95) ((T_IMPLEMENTS) reduce 95) ((T_INCLUDE) reduce 95) ((T_INCLUDE_ONCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_INSTEADOF) reduce 95) ((T_INTERFACE) reduce 95) ((T_ISSET) reduce 95) ((T_LINE) reduce 95) ((T_LIST) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_MATCH) reduce 95) ((T_METHOD_C) reduce 95) ((T_NAMESPACE) reduce 95) ((T_NEW) reduce 95) ((T_NS_C) reduce 95) ((T_PRINT) reduce 95) ((T_PRIVATE) reduce 95) ((T_PROPERTY_C) reduce 95) ((T_PROTECTED) reduce 95) ((T_PUBLIC) reduce 95) ((T_READONLY) reduce 95) ((T_REQUIRE) reduce 95) ((T_REQUIRE_ONCE) reduce 95) ((T_RETURN) reduce 95) ((T_STATIC) reduce 95) ((T_STRING) reduce 95) ((T_SWITCH) reduce 95) ((T_THROW) reduce 95) ((T_TRAIT) reduce 95) ((T_TRAIT_C) reduce 95) ((T_TRY) reduce 95) ((T_UNSET) reduce 95) ((T_USE) reduce 95) ((T_VAR) reduce 95) ((T_WHILE) reduce 95) ((T_YIELD) reduce 95) (("|") reduce 95)) 7122 ((("=") reduce 10)) 7123 ((("=") reduce 67)) 7124 ((("=") reduce 44)) 7125 ((("=") reduce 76)) 7126 ((("=") reduce 71)) 7127 ((("=") reduce 77)) 7128 ((("=") reduce 78)) 7129 ((("=") reduce 79)) 7130 ((("=") reduce 4)) 7131 ((("=") reduce 5)) 7132 ((("=") reduce 43)) 7133 ((("=") reduce 73) ((T_ABSTRACT) reduce 267) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267) ((T_ARRAY) reduce 267) ((T_AS) reduce 267) ((T_BREAK) reduce 267) ((T_CALLABLE) reduce 267) ((T_CASE) reduce 267) ((T_CATCH) reduce 267) ((T_CLASS) reduce 267) ((T_CLASS_C) reduce 267) ((T_CLONE) reduce 267) ((T_CONST) reduce 267) ((T_CONTINUE) reduce 267) ((T_DECLARE) reduce 267) ((T_DEFAULT) reduce 267) ((T_DIR) reduce 267) ((T_DO) reduce 267) ((T_ECHO) reduce 267) ((T_ELSE) reduce 267) ((T_ELSEIF) reduce 267) ((T_EMPTY) reduce 267) ((T_ENDDECLARE) reduce 267) ((T_ENDFOR) reduce 267) ((T_ENDFOREACH) reduce 267) ((T_ENDIF) reduce 267) ((T_ENDSWITCH) reduce 267) ((T_ENDWHILE) reduce 267) ((T_ENUM) reduce 267) ((T_EVAL) reduce 267) ((T_EXIT) reduce 267) ((T_EXTENDS) reduce 267) ((T_FILE) reduce 267) ((T_FINAL) reduce 267) ((T_FINALLY) reduce 267) ((T_FN) reduce 267) ((T_FOR) reduce 267) ((T_FOREACH) reduce 267) ((T_FUNCTION) reduce 267) ((T_FUNC_C) reduce 267) ((T_GLOBAL) reduce 267) ((T_GOTO) reduce 267) ((T_IF) reduce 267) ((T_IMPLEMENTS) reduce 267) ((T_INCLUDE) reduce 267) ((T_INCLUDE_ONCE) reduce 267) ((T_INSTANCEOF) reduce 267) ((T_INSTEADOF) reduce 267) ((T_INTERFACE) reduce 267) ((T_ISSET) reduce 267) ((T_LINE) reduce 267) ((T_LIST) reduce 267) ((T_LOGICAL_AND) reduce 267) ((T_LOGICAL_OR) reduce 267) ((T_LOGICAL_XOR) reduce 267) ((T_MATCH) reduce 267) ((T_METHOD_C) reduce 267) ((T_NAMESPACE) reduce 267) ((T_NEW) reduce 267) ((T_NS_C) reduce 267) ((T_PRINT) reduce 267) ((T_PRIVATE) reduce 267) ((T_PROPERTY_C) reduce 267) ((T_PROTECTED) reduce 267) ((T_PUBLIC) reduce 267) ((T_READONLY) reduce 267) ((T_REQUIRE) reduce 267) ((T_REQUIRE_ONCE) reduce 267) ((T_RETURN) reduce 267) ((T_STATIC) reduce 267) ((T_STRING) reduce 267) ((T_SWITCH) reduce 267) ((T_THROW) reduce 267) ((T_TRAIT) reduce 267) ((T_TRAIT_C) reduce 267) ((T_TRY) reduce 267) ((T_UNSET) reduce 267) ((T_USE) reduce 267) ((T_VAR) reduce 267) ((T_WHILE) reduce 267) ((T_YIELD) reduce 267) (("|") reduce 267)) 7134 ((("=") shift) ((T_ABSTRACT) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ARRAY) reduce 92) ((T_AS) reduce 92) ((T_BREAK) reduce 92) ((T_CALLABLE) reduce 92) ((T_CASE) reduce 92) ((T_CATCH) reduce 92) ((T_CLASS) reduce 92) ((T_CLASS_C) reduce 92) ((T_CLONE) reduce 92) ((T_CONST) reduce 92) ((T_CONTINUE) reduce 92) ((T_DECLARE) reduce 92) ((T_DEFAULT) reduce 92) ((T_DIR) reduce 92) ((T_DO) reduce 92) ((T_ECHO) reduce 92) ((T_ELSE) reduce 92) ((T_ELSEIF) reduce 92) ((T_EMPTY) reduce 92) ((T_ENDDECLARE) reduce 92) ((T_ENDFOR) reduce 92) ((T_ENDFOREACH) reduce 92) ((T_ENDIF) reduce 92) ((T_ENDSWITCH) reduce 92) ((T_ENDWHILE) reduce 92) ((T_ENUM) reduce 92) ((T_EVAL) reduce 92) ((T_EXIT) reduce 92) ((T_EXTENDS) reduce 92) ((T_FILE) reduce 92) ((T_FINAL) reduce 92) ((T_FINALLY) reduce 92) ((T_FN) reduce 92) ((T_FOR) reduce 92) ((T_FOREACH) reduce 92) ((T_FUNCTION) reduce 92) ((T_FUNC_C) reduce 92) ((T_GLOBAL) reduce 92) ((T_GOTO) reduce 92) ((T_IF) reduce 92) ((T_IMPLEMENTS) reduce 92) ((T_INCLUDE) reduce 92) ((T_INCLUDE_ONCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_INSTEADOF) reduce 92) ((T_INTERFACE) reduce 92) ((T_ISSET) reduce 92) ((T_LINE) reduce 92) ((T_LIST) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_MATCH) reduce 92) ((T_METHOD_C) reduce 92) ((T_NAMESPACE) reduce 92) ((T_NEW) reduce 92) ((T_NS_C) reduce 92) ((T_PRINT) reduce 92) ((T_PRIVATE) reduce 92) ((T_PROPERTY_C) reduce 92) ((T_PROTECTED) reduce 92) ((T_PUBLIC) reduce 92) ((T_READONLY) reduce 92) ((T_REQUIRE) reduce 92) ((T_REQUIRE_ONCE) reduce 92) ((T_RETURN) reduce 92) ((T_STATIC) reduce 92) ((T_STRING) reduce 92) ((T_SWITCH) reduce 92) ((T_THROW) reduce 92) ((T_TRAIT) reduce 92) ((T_TRAIT_C) reduce 92) ((T_TRY) reduce 92) ((T_UNSET) reduce 92) ((T_USE) reduce 92) ((T_VAR) reduce 92) ((T_WHILE) reduce 92) ((T_YIELD) reduce 92) (("|") reduce 92)) 7135 ((("=") reduce 47)) 7136 ((("=") reduce 31)) 7137 ((("=") reduce 57)) 7138 ((("=") reduce 61)) 7139 ((("=") reduce 28)) 7140 ((("=") reduce 36)) 7141 ((("=") reduce 32)) 7142 ((("=") reduce 35)) 7143 ((("=") reduce 19)) 7144 ((("=") reduce 45)) 7145 (((",") reduce 372) ((";") reduce 372)) 7147 (((T_ABSTRACT) reduce 265) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) reduce 265) ((T_AS) reduce 265) ((T_BREAK) reduce 265) ((T_CALLABLE) reduce 265) ((T_CASE) reduce 265) ((T_CATCH) reduce 265) ((T_CLASS) reduce 265) ((T_CLASS_C) reduce 265) ((T_CLONE) reduce 265) ((T_CONST) reduce 265) ((T_CONTINUE) reduce 265) ((T_DECLARE) reduce 265) ((T_DEFAULT) reduce 265) ((T_DIR) reduce 265) ((T_DO) reduce 265) ((T_ECHO) reduce 265) ((T_ELSE) reduce 265) ((T_ELSEIF) reduce 265) ((T_EMPTY) reduce 265) ((T_ENDDECLARE) reduce 265) ((T_ENDFOR) reduce 265) ((T_ENDFOREACH) reduce 265) ((T_ENDIF) reduce 265) ((T_ENDSWITCH) reduce 265) ((T_ENDWHILE) reduce 265) ((T_ENUM) reduce 265) ((T_EVAL) reduce 265) ((T_EXIT) reduce 265) ((T_EXTENDS) reduce 265) ((T_FILE) reduce 265) ((T_FINAL) reduce 265) ((T_FINALLY) reduce 265) ((T_FN) reduce 265) ((T_FOR) reduce 265) ((T_FOREACH) reduce 265) ((T_FUNCTION) reduce 265) ((T_FUNC_C) reduce 265) ((T_GLOBAL) reduce 265) ((T_GOTO) reduce 265) ((T_IF) reduce 265) ((T_IMPLEMENTS) reduce 265) ((T_INCLUDE) reduce 265) ((T_INCLUDE_ONCE) reduce 265) ((T_INSTANCEOF) reduce 265) ((T_INSTEADOF) reduce 265) ((T_INTERFACE) reduce 265) ((T_ISSET) reduce 265) ((T_LINE) reduce 265) ((T_LIST) reduce 265) ((T_LOGICAL_AND) reduce 265) ((T_LOGICAL_OR) reduce 265) ((T_LOGICAL_XOR) reduce 265) ((T_MATCH) reduce 265) ((T_METHOD_C) reduce 265) ((T_NAMESPACE) reduce 265) ((T_NEW) reduce 265) ((T_NS_C) reduce 265) ((T_PRINT) reduce 265) ((T_PRIVATE) reduce 265) ((T_PROPERTY_C) reduce 265) ((T_PROTECTED) reduce 265) ((T_PUBLIC) reduce 265) ((T_READONLY) reduce 265) ((T_REQUIRE) reduce 265) ((T_REQUIRE_ONCE) reduce 265) ((T_RETURN) reduce 265) ((T_STATIC) reduce 265) ((T_STRING) reduce 265) ((T_SWITCH) reduce 265) ((T_THROW) reduce 265) ((T_TRAIT) reduce 265) ((T_TRAIT_C) reduce 265) ((T_TRY) reduce 265) ((T_UNSET) reduce 265) ((T_USE) reduce 265) ((T_VAR) reduce 265) ((T_WHILE) reduce 265) ((T_YIELD) reduce 265)) 7148 (((T_ABSTRACT) reduce 280) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_ARRAY) reduce 280) ((T_AS) reduce 280) ((T_BREAK) reduce 280) ((T_CALLABLE) reduce 280) ((T_CASE) reduce 280) ((T_CATCH) reduce 280) ((T_CLASS) reduce 280) ((T_CLASS_C) reduce 280) ((T_CLONE) reduce 280) ((T_CONST) reduce 280) ((T_CONTINUE) reduce 280) ((T_DECLARE) reduce 280) ((T_DEFAULT) reduce 280) ((T_DIR) reduce 280) ((T_DO) reduce 280) ((T_ECHO) reduce 280) ((T_ELSE) reduce 280) ((T_ELSEIF) reduce 280) ((T_EMPTY) reduce 280) ((T_ENDDECLARE) reduce 280) ((T_ENDFOR) reduce 280) ((T_ENDFOREACH) reduce 280) ((T_ENDIF) reduce 280) ((T_ENDSWITCH) reduce 280) ((T_ENDWHILE) reduce 280) ((T_ENUM) reduce 280) ((T_EVAL) reduce 280) ((T_EXIT) reduce 280) ((T_EXTENDS) reduce 280) ((T_FILE) reduce 280) ((T_FINAL) reduce 280) ((T_FINALLY) reduce 280) ((T_FN) reduce 280) ((T_FOR) reduce 280) ((T_FOREACH) reduce 280) ((T_FUNCTION) reduce 280) ((T_FUNC_C) reduce 280) ((T_GLOBAL) reduce 280) ((T_GOTO) reduce 280) ((T_IF) reduce 280) ((T_IMPLEMENTS) reduce 280) ((T_INCLUDE) reduce 280) ((T_INCLUDE_ONCE) reduce 280) ((T_INSTANCEOF) reduce 280) ((T_INSTEADOF) reduce 280) ((T_INTERFACE) reduce 280) ((T_ISSET) reduce 280) ((T_LINE) reduce 280) ((T_LIST) reduce 280) ((T_LOGICAL_AND) reduce 280) ((T_LOGICAL_OR) reduce 280) ((T_LOGICAL_XOR) reduce 280) ((T_MATCH) reduce 280) ((T_METHOD_C) reduce 280) ((T_NAMESPACE) reduce 280) ((T_NEW) reduce 280) ((T_NS_C) reduce 280) ((T_PRINT) reduce 280) ((T_PRIVATE) reduce 280) ((T_PROPERTY_C) reduce 280) ((T_PROTECTED) reduce 280) ((T_PUBLIC) reduce 280) ((T_READONLY) reduce 280) ((T_REQUIRE) reduce 280) ((T_REQUIRE_ONCE) reduce 280) ((T_RETURN) reduce 280) ((T_STATIC) reduce 280) ((T_STRING) reduce 280) ((T_SWITCH) reduce 280) ((T_THROW) reduce 280) ((T_TRAIT) reduce 280) ((T_TRAIT_C) reduce 280) ((T_TRY) reduce 280) ((T_UNSET) reduce 280) ((T_USE) reduce 280) ((T_VAR) reduce 280) ((T_WHILE) reduce 280) ((T_YIELD) reduce 280) (("|") reduce 280)) 7149 ((("=") reduce 72)) 7151 (((T_ABSTRACT) reduce 262) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) reduce 262) ((T_AS) reduce 262) ((T_BREAK) reduce 262) ((T_CALLABLE) reduce 262) ((T_CASE) reduce 262) ((T_CATCH) reduce 262) ((T_CLASS) reduce 262) ((T_CLASS_C) reduce 262) ((T_CLONE) reduce 262) ((T_CONST) reduce 262) ((T_CONTINUE) reduce 262) ((T_DECLARE) reduce 262) ((T_DEFAULT) reduce 262) ((T_DIR) reduce 262) ((T_DO) reduce 262) ((T_ECHO) reduce 262) ((T_ELSE) reduce 262) ((T_ELSEIF) reduce 262) ((T_EMPTY) reduce 262) ((T_ENDDECLARE) reduce 262) ((T_ENDFOR) reduce 262) ((T_ENDFOREACH) reduce 262) ((T_ENDIF) reduce 262) ((T_ENDSWITCH) reduce 262) ((T_ENDWHILE) reduce 262) ((T_ENUM) reduce 262) ((T_EVAL) reduce 262) ((T_EXIT) reduce 262) ((T_EXTENDS) reduce 262) ((T_FILE) reduce 262) ((T_FINAL) reduce 262) ((T_FINALLY) reduce 262) ((T_FN) reduce 262) ((T_FOR) reduce 262) ((T_FOREACH) reduce 262) ((T_FUNCTION) reduce 262) ((T_FUNC_C) reduce 262) ((T_GLOBAL) reduce 262) ((T_GOTO) reduce 262) ((T_IF) reduce 262) ((T_IMPLEMENTS) reduce 262) ((T_INCLUDE) reduce 262) ((T_INCLUDE_ONCE) reduce 262) ((T_INSTANCEOF) reduce 262) ((T_INSTEADOF) reduce 262) ((T_INTERFACE) reduce 262) ((T_ISSET) reduce 262) ((T_LINE) reduce 262) ((T_LIST) reduce 262) ((T_LOGICAL_AND) reduce 262) ((T_LOGICAL_OR) reduce 262) ((T_LOGICAL_XOR) reduce 262) ((T_MATCH) reduce 262) ((T_METHOD_C) reduce 262) ((T_NAMESPACE) reduce 262) ((T_NEW) reduce 262) ((T_NS_C) reduce 262) ((T_PRINT) reduce 262) ((T_PRIVATE) reduce 262) ((T_PROPERTY_C) reduce 262) ((T_PROTECTED) reduce 262) ((T_PUBLIC) reduce 262) ((T_READONLY) reduce 262) ((T_REQUIRE) reduce 262) ((T_REQUIRE_ONCE) reduce 262) ((T_RETURN) reduce 262) ((T_STATIC) reduce 262) ((T_STRING) reduce 262) ((T_SWITCH) reduce 262) ((T_THROW) reduce 262) ((T_TRAIT) reduce 262) ((T_TRAIT_C) reduce 262) ((T_TRY) reduce 262) ((T_UNSET) reduce 262) ((T_USE) reduce 262) ((T_VAR) reduce 262) ((T_WHILE) reduce 262) ((T_YIELD) reduce 262) (("|") reduce 268)) 7153 (((T_ABSTRACT) reduce 266) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266) ((T_ARRAY) reduce 266) ((T_AS) reduce 266) ((T_BREAK) reduce 266) ((T_CALLABLE) reduce 266) ((T_CASE) reduce 266) ((T_CATCH) reduce 266) ((T_CLASS) reduce 266) ((T_CLASS_C) reduce 266) ((T_CLONE) reduce 266) ((T_CONST) reduce 266) ((T_CONTINUE) reduce 266) ((T_DECLARE) reduce 266) ((T_DEFAULT) reduce 266) ((T_DIR) reduce 266) ((T_DO) reduce 266) ((T_ECHO) reduce 266) ((T_ELSE) reduce 266) ((T_ELSEIF) reduce 266) ((T_EMPTY) reduce 266) ((T_ENDDECLARE) reduce 266) ((T_ENDFOR) reduce 266) ((T_ENDFOREACH) reduce 266) ((T_ENDIF) reduce 266) ((T_ENDSWITCH) reduce 266) ((T_ENDWHILE) reduce 266) ((T_ENUM) reduce 266) ((T_EVAL) reduce 266) ((T_EXIT) reduce 266) ((T_EXTENDS) reduce 266) ((T_FILE) reduce 266) ((T_FINAL) reduce 266) ((T_FINALLY) reduce 266) ((T_FN) reduce 266) ((T_FOR) reduce 266) ((T_FOREACH) reduce 266) ((T_FUNCTION) reduce 266) ((T_FUNC_C) reduce 266) ((T_GLOBAL) reduce 266) ((T_GOTO) reduce 266) ((T_IF) reduce 266) ((T_IMPLEMENTS) reduce 266) ((T_INCLUDE) reduce 266) ((T_INCLUDE_ONCE) reduce 266) ((T_INSTANCEOF) reduce 266) ((T_INSTEADOF) reduce 266) ((T_INTERFACE) reduce 266) ((T_ISSET) reduce 266) ((T_LINE) reduce 266) ((T_LIST) reduce 266) ((T_LOGICAL_AND) reduce 266) ((T_LOGICAL_OR) reduce 266) ((T_LOGICAL_XOR) reduce 266) ((T_MATCH) reduce 266) ((T_METHOD_C) reduce 266) ((T_NAMESPACE) reduce 266) ((T_NEW) reduce 266) ((T_NS_C) reduce 266) ((T_PRINT) reduce 266) ((T_PRIVATE) reduce 266) ((T_PROPERTY_C) reduce 266) ((T_PROTECTED) reduce 266) ((T_PUBLIC) reduce 266) ((T_READONLY) reduce 266) ((T_REQUIRE) reduce 266) ((T_REQUIRE_ONCE) reduce 266) ((T_RETURN) reduce 266) ((T_STATIC) reduce 266) ((T_STRING) reduce 266) ((T_SWITCH) reduce 266) ((T_THROW) reduce 266) ((T_TRAIT) reduce 266) ((T_TRAIT_C) reduce 266) ((T_TRY) reduce 266) ((T_UNSET) reduce 266) ((T_USE) reduce 266) ((T_VAR) reduce 266) ((T_WHILE) reduce 266) ((T_YIELD) reduce 266) (("|") reduce 266)) 7154 (((T_ABSTRACT) reduce 264) ((T_ARRAY) reduce 264) ((T_AS) reduce 264) ((T_BREAK) reduce 264) ((T_CALLABLE) reduce 264) ((T_CASE) reduce 264) ((T_CATCH) reduce 264) ((T_CLASS) reduce 264) ((T_CLASS_C) reduce 264) ((T_CLONE) reduce 264) ((T_CONST) reduce 264) ((T_CONTINUE) reduce 264) ((T_DECLARE) reduce 264) ((T_DEFAULT) reduce 264) ((T_DIR) reduce 264) ((T_DO) reduce 264) ((T_ECHO) reduce 264) ((T_ELSE) reduce 264) ((T_ELSEIF) reduce 264) ((T_EMPTY) reduce 264) ((T_ENDDECLARE) reduce 264) ((T_ENDFOR) reduce 264) ((T_ENDFOREACH) reduce 264) ((T_ENDIF) reduce 264) ((T_ENDSWITCH) reduce 264) ((T_ENDWHILE) reduce 264) ((T_ENUM) reduce 264) ((T_EVAL) reduce 264) ((T_EXIT) reduce 264) ((T_EXTENDS) reduce 264) ((T_FILE) reduce 264) ((T_FINAL) reduce 264) ((T_FINALLY) reduce 264) ((T_FN) reduce 264) ((T_FOR) reduce 264) ((T_FOREACH) reduce 264) ((T_FUNCTION) reduce 264) ((T_FUNC_C) reduce 264) ((T_GLOBAL) reduce 264) ((T_GOTO) reduce 264) ((T_IF) reduce 264) ((T_IMPLEMENTS) reduce 264) ((T_INCLUDE) reduce 264) ((T_INCLUDE_ONCE) reduce 264) ((T_INSTANCEOF) reduce 264) ((T_INSTEADOF) reduce 264) ((T_INTERFACE) reduce 264) ((T_ISSET) reduce 264) ((T_LINE) reduce 264) ((T_LIST) reduce 264) ((T_LOGICAL_AND) reduce 264) ((T_LOGICAL_OR) reduce 264) ((T_LOGICAL_XOR) reduce 264) ((T_MATCH) reduce 264) ((T_METHOD_C) reduce 264) ((T_NAMESPACE) reduce 264) ((T_NEW) reduce 264) ((T_NS_C) reduce 264) ((T_PRINT) reduce 264) ((T_PRIVATE) reduce 264) ((T_PROPERTY_C) reduce 264) ((T_PROTECTED) reduce 264) ((T_PUBLIC) reduce 264) ((T_READONLY) reduce 264) ((T_REQUIRE) reduce 264) ((T_REQUIRE_ONCE) reduce 264) ((T_RETURN) reduce 264) ((T_STATIC) reduce 264) ((T_STRING) reduce 264) ((T_SWITCH) reduce 264) ((T_THROW) reduce 264) ((T_TRAIT) reduce 264) ((T_TRAIT_C) reduce 264) ((T_TRY) reduce 264) ((T_UNSET) reduce 264) ((T_USE) reduce 264) ((T_VAR) reduce 264) ((T_WHILE) reduce 264) ((T_YIELD) reduce 264) (("|") shift)) 7158 (((T_ABSTRACT) reduce 278) ((T_ARRAY) reduce 278) ((T_AS) reduce 278) ((T_BREAK) reduce 278) ((T_CALLABLE) reduce 278) ((T_CASE) reduce 278) ((T_CATCH) reduce 278) ((T_CLASS) reduce 278) ((T_CLASS_C) reduce 278) ((T_CLONE) reduce 278) ((T_CONST) reduce 278) ((T_CONTINUE) reduce 278) ((T_DECLARE) reduce 278) ((T_DEFAULT) reduce 278) ((T_DIR) reduce 278) ((T_DO) reduce 278) ((T_ECHO) reduce 278) ((T_ELSE) reduce 278) ((T_ELSEIF) reduce 278) ((T_EMPTY) reduce 278) ((T_ENDDECLARE) reduce 278) ((T_ENDFOR) reduce 278) ((T_ENDFOREACH) reduce 278) ((T_ENDIF) reduce 278) ((T_ENDSWITCH) reduce 278) ((T_ENDWHILE) reduce 278) ((T_ENUM) reduce 278) ((T_EVAL) reduce 278) ((T_EXIT) reduce 278) ((T_EXTENDS) reduce 278) ((T_FILE) reduce 278) ((T_FINAL) reduce 278) ((T_FINALLY) reduce 278) ((T_FN) reduce 278) ((T_FOR) reduce 278) ((T_FOREACH) reduce 278) ((T_FUNCTION) reduce 278) ((T_FUNC_C) reduce 278) ((T_GLOBAL) reduce 278) ((T_GOTO) reduce 278) ((T_IF) reduce 278) ((T_IMPLEMENTS) reduce 278) ((T_INCLUDE) reduce 278) ((T_INCLUDE_ONCE) reduce 278) ((T_INSTANCEOF) reduce 278) ((T_INSTEADOF) reduce 278) ((T_INTERFACE) reduce 278) ((T_ISSET) reduce 278) ((T_LINE) reduce 278) ((T_LIST) reduce 278) ((T_LOGICAL_AND) reduce 278) ((T_LOGICAL_OR) reduce 278) ((T_LOGICAL_XOR) reduce 278) ((T_MATCH) reduce 278) ((T_METHOD_C) reduce 278) ((T_NAMESPACE) reduce 278) ((T_NEW) reduce 278) ((T_NS_C) reduce 278) ((T_PRINT) reduce 278) ((T_PRIVATE) reduce 278) ((T_PROPERTY_C) reduce 278) ((T_PROTECTED) reduce 278) ((T_PUBLIC) reduce 278) ((T_READONLY) reduce 278) ((T_REQUIRE) reduce 278) ((T_REQUIRE_ONCE) reduce 278) ((T_RETURN) reduce 278) ((T_STATIC) reduce 278) ((T_STRING) reduce 278) ((T_SWITCH) reduce 278) ((T_THROW) reduce 278) ((T_TRAIT) reduce 278) ((T_TRAIT_C) reduce 278) ((T_TRY) reduce 278) ((T_UNSET) reduce 278) ((T_USE) reduce 278) ((T_VAR) reduce 278) ((T_WHILE) reduce 278) ((T_YIELD) reduce 278) (("|") reduce 278)) 7159 (((T_ABSTRACT) reduce 279) ((T_ARRAY) reduce 279) ((T_AS) reduce 279) ((T_BREAK) reduce 279) ((T_CALLABLE) reduce 279) ((T_CASE) reduce 279) ((T_CATCH) reduce 279) ((T_CLASS) reduce 279) ((T_CLASS_C) reduce 279) ((T_CLONE) reduce 279) ((T_CONST) reduce 279) ((T_CONTINUE) reduce 279) ((T_DECLARE) reduce 279) ((T_DEFAULT) reduce 279) ((T_DIR) reduce 279) ((T_DO) reduce 279) ((T_ECHO) reduce 279) ((T_ELSE) reduce 279) ((T_ELSEIF) reduce 279) ((T_EMPTY) reduce 279) ((T_ENDDECLARE) reduce 279) ((T_ENDFOR) reduce 279) ((T_ENDFOREACH) reduce 279) ((T_ENDIF) reduce 279) ((T_ENDSWITCH) reduce 279) ((T_ENDWHILE) reduce 279) ((T_ENUM) reduce 279) ((T_EVAL) reduce 279) ((T_EXIT) reduce 279) ((T_EXTENDS) reduce 279) ((T_FILE) reduce 279) ((T_FINAL) reduce 279) ((T_FINALLY) reduce 279) ((T_FN) reduce 279) ((T_FOR) reduce 279) ((T_FOREACH) reduce 279) ((T_FUNCTION) reduce 279) ((T_FUNC_C) reduce 279) ((T_GLOBAL) reduce 279) ((T_GOTO) reduce 279) ((T_IF) reduce 279) ((T_IMPLEMENTS) reduce 279) ((T_INCLUDE) reduce 279) ((T_INCLUDE_ONCE) reduce 279) ((T_INSTANCEOF) reduce 279) ((T_INSTEADOF) reduce 279) ((T_INTERFACE) reduce 279) ((T_ISSET) reduce 279) ((T_LINE) reduce 279) ((T_LIST) reduce 279) ((T_LOGICAL_AND) reduce 279) ((T_LOGICAL_OR) reduce 279) ((T_LOGICAL_XOR) reduce 279) ((T_MATCH) reduce 279) ((T_METHOD_C) reduce 279) ((T_NAMESPACE) reduce 279) ((T_NEW) reduce 279) ((T_NS_C) reduce 279) ((T_PRINT) reduce 279) ((T_PRIVATE) reduce 279) ((T_PROPERTY_C) reduce 279) ((T_PROTECTED) reduce 279) ((T_PUBLIC) reduce 279) ((T_READONLY) reduce 279) ((T_REQUIRE) reduce 279) ((T_REQUIRE_ONCE) reduce 279) ((T_RETURN) reduce 279) ((T_STATIC) reduce 279) ((T_STRING) reduce 279) ((T_SWITCH) reduce 279) ((T_THROW) reduce 279) ((T_TRAIT) reduce 279) ((T_TRAIT_C) reduce 279) ((T_TRY) reduce 279) ((T_UNSET) reduce 279) ((T_USE) reduce 279) ((T_VAR) reduce 279) ((T_WHILE) reduce 279) ((T_YIELD) reduce 279) (("|") reduce 279)) 7160 (((T_ABSTRACT) reduce 94) ((T_ARRAY) reduce 94) ((T_AS) reduce 94) ((T_BREAK) reduce 94) ((T_CALLABLE) reduce 94) ((T_CASE) reduce 94) ((T_CATCH) reduce 94) ((T_CLASS) reduce 94) ((T_CLASS_C) reduce 94) ((T_CLONE) reduce 94) ((T_CONST) reduce 94) ((T_CONTINUE) reduce 94) ((T_DECLARE) reduce 94) ((T_DEFAULT) reduce 94) ((T_DIR) reduce 94) ((T_DO) reduce 94) ((T_ECHO) reduce 94) ((T_ELSE) reduce 94) ((T_ELSEIF) reduce 94) ((T_EMPTY) reduce 94) ((T_ENDDECLARE) reduce 94) ((T_ENDFOR) reduce 94) ((T_ENDFOREACH) reduce 94) ((T_ENDIF) reduce 94) ((T_ENDSWITCH) reduce 94) ((T_ENDWHILE) reduce 94) ((T_ENUM) reduce 94) ((T_EVAL) reduce 94) ((T_EXIT) reduce 94) ((T_EXTENDS) reduce 94) ((T_FILE) reduce 94) ((T_FINAL) reduce 94) ((T_FINALLY) reduce 94) ((T_FN) reduce 94) ((T_FOR) reduce 94) ((T_FOREACH) reduce 94) ((T_FUNCTION) reduce 94) ((T_FUNC_C) reduce 94) ((T_GLOBAL) reduce 94) ((T_GOTO) reduce 94) ((T_IF) reduce 94) ((T_IMPLEMENTS) reduce 94) ((T_INCLUDE) reduce 94) ((T_INCLUDE_ONCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_INSTEADOF) reduce 94) ((T_INTERFACE) reduce 94) ((T_ISSET) reduce 94) ((T_LINE) reduce 94) ((T_LIST) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_MATCH) reduce 94) ((T_METHOD_C) reduce 94) ((T_NAMESPACE) reduce 94) ((T_NEW) reduce 94) ((T_NS_C) reduce 94) ((T_PRINT) reduce 94) ((T_PRIVATE) reduce 94) ((T_PROPERTY_C) reduce 94) ((T_PROTECTED) reduce 94) ((T_PUBLIC) reduce 94) ((T_READONLY) reduce 94) ((T_REQUIRE) reduce 94) ((T_REQUIRE_ONCE) reduce 94) ((T_RETURN) reduce 94) ((T_STATIC) reduce 94) ((T_STRING) reduce 94) ((T_SWITCH) reduce 94) ((T_THROW) reduce 94) ((T_TRAIT) reduce 94) ((T_TRAIT_C) reduce 94) ((T_TRY) reduce 94) ((T_UNSET) reduce 94) ((T_USE) reduce 94) ((T_VAR) reduce 94) ((T_WHILE) reduce 94) ((T_YIELD) reduce 94) (("|") reduce 94)) 7161 (((T_ABSTRACT) reduce 93) ((T_ARRAY) reduce 93) ((T_AS) reduce 93) ((T_BREAK) reduce 93) ((T_CALLABLE) reduce 93) ((T_CASE) reduce 93) ((T_CATCH) reduce 93) ((T_CLASS) reduce 93) ((T_CLASS_C) reduce 93) ((T_CLONE) reduce 93) ((T_CONST) reduce 93) ((T_CONTINUE) reduce 93) ((T_DECLARE) reduce 93) ((T_DEFAULT) reduce 93) ((T_DIR) reduce 93) ((T_DO) reduce 93) ((T_ECHO) reduce 93) ((T_ELSE) reduce 93) ((T_ELSEIF) reduce 93) ((T_EMPTY) reduce 93) ((T_ENDDECLARE) reduce 93) ((T_ENDFOR) reduce 93) ((T_ENDFOREACH) reduce 93) ((T_ENDIF) reduce 93) ((T_ENDSWITCH) reduce 93) ((T_ENDWHILE) reduce 93) ((T_ENUM) reduce 93) ((T_EVAL) reduce 93) ((T_EXIT) reduce 93) ((T_EXTENDS) reduce 93) ((T_FILE) reduce 93) ((T_FINAL) reduce 93) ((T_FINALLY) reduce 93) ((T_FN) reduce 93) ((T_FOR) reduce 93) ((T_FOREACH) reduce 93) ((T_FUNCTION) reduce 93) ((T_FUNC_C) reduce 93) ((T_GLOBAL) reduce 93) ((T_GOTO) reduce 93) ((T_IF) reduce 93) ((T_IMPLEMENTS) reduce 93) ((T_INCLUDE) reduce 93) ((T_INCLUDE_ONCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_INSTEADOF) reduce 93) ((T_INTERFACE) reduce 93) ((T_ISSET) reduce 93) ((T_LINE) reduce 93) ((T_LIST) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_MATCH) reduce 93) ((T_METHOD_C) reduce 93) ((T_NAMESPACE) reduce 93) ((T_NEW) reduce 93) ((T_NS_C) reduce 93) ((T_PRINT) reduce 93) ((T_PRIVATE) reduce 93) ((T_PROPERTY_C) reduce 93) ((T_PROTECTED) reduce 93) ((T_PUBLIC) reduce 93) ((T_READONLY) reduce 93) ((T_REQUIRE) reduce 93) ((T_REQUIRE_ONCE) reduce 93) ((T_RETURN) reduce 93) ((T_STATIC) reduce 93) ((T_STRING) reduce 93) ((T_SWITCH) reduce 93) ((T_THROW) reduce 93) ((T_TRAIT) reduce 93) ((T_TRAIT_C) reduce 93) ((T_TRY) reduce 93) ((T_UNSET) reduce 93) ((T_USE) reduce 93) ((T_VAR) reduce 93) ((T_WHILE) reduce 93) ((T_YIELD) reduce 93) (("|") reduce 93)) 7162 (((T_ABSTRACT) reduce 95) ((T_ARRAY) reduce 95) ((T_AS) reduce 95) ((T_BREAK) reduce 95) ((T_CALLABLE) reduce 95) ((T_CASE) reduce 95) ((T_CATCH) reduce 95) ((T_CLASS) reduce 95) ((T_CLASS_C) reduce 95) ((T_CLONE) reduce 95) ((T_CONST) reduce 95) ((T_CONTINUE) reduce 95) ((T_DECLARE) reduce 95) ((T_DEFAULT) reduce 95) ((T_DIR) reduce 95) ((T_DO) reduce 95) ((T_ECHO) reduce 95) ((T_ELSE) reduce 95) ((T_ELSEIF) reduce 95) ((T_EMPTY) reduce 95) ((T_ENDDECLARE) reduce 95) ((T_ENDFOR) reduce 95) ((T_ENDFOREACH) reduce 95) ((T_ENDIF) reduce 95) ((T_ENDSWITCH) reduce 95) ((T_ENDWHILE) reduce 95) ((T_ENUM) reduce 95) ((T_EVAL) reduce 95) ((T_EXIT) reduce 95) ((T_EXTENDS) reduce 95) ((T_FILE) reduce 95) ((T_FINAL) reduce 95) ((T_FINALLY) reduce 95) ((T_FN) reduce 95) ((T_FOR) reduce 95) ((T_FOREACH) reduce 95) ((T_FUNCTION) reduce 95) ((T_FUNC_C) reduce 95) ((T_GLOBAL) reduce 95) ((T_GOTO) reduce 95) ((T_IF) reduce 95) ((T_IMPLEMENTS) reduce 95) ((T_INCLUDE) reduce 95) ((T_INCLUDE_ONCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_INSTEADOF) reduce 95) ((T_INTERFACE) reduce 95) ((T_ISSET) reduce 95) ((T_LINE) reduce 95) ((T_LIST) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_MATCH) reduce 95) ((T_METHOD_C) reduce 95) ((T_NAMESPACE) reduce 95) ((T_NEW) reduce 95) ((T_NS_C) reduce 95) ((T_PRINT) reduce 95) ((T_PRIVATE) reduce 95) ((T_PROPERTY_C) reduce 95) ((T_PROTECTED) reduce 95) ((T_PUBLIC) reduce 95) ((T_READONLY) reduce 95) ((T_REQUIRE) reduce 95) ((T_REQUIRE_ONCE) reduce 95) ((T_RETURN) reduce 95) ((T_STATIC) reduce 95) ((T_STRING) reduce 95) ((T_SWITCH) reduce 95) ((T_THROW) reduce 95) ((T_TRAIT) reduce 95) ((T_TRAIT_C) reduce 95) ((T_TRY) reduce 95) ((T_UNSET) reduce 95) ((T_USE) reduce 95) ((T_VAR) reduce 95) ((T_WHILE) reduce 95) ((T_YIELD) reduce 95) (("|") reduce 95)) 7163 (((T_ABSTRACT) reduce 267) ((T_ARRAY) reduce 267) ((T_AS) reduce 267) ((T_BREAK) reduce 267) ((T_CALLABLE) reduce 267) ((T_CASE) reduce 267) ((T_CATCH) reduce 267) ((T_CLASS) reduce 267) ((T_CLASS_C) reduce 267) ((T_CLONE) reduce 267) ((T_CONST) reduce 267) ((T_CONTINUE) reduce 267) ((T_DECLARE) reduce 267) ((T_DEFAULT) reduce 267) ((T_DIR) reduce 267) ((T_DO) reduce 267) ((T_ECHO) reduce 267) ((T_ELSE) reduce 267) ((T_ELSEIF) reduce 267) ((T_EMPTY) reduce 267) ((T_ENDDECLARE) reduce 267) ((T_ENDFOR) reduce 267) ((T_ENDFOREACH) reduce 267) ((T_ENDIF) reduce 267) ((T_ENDSWITCH) reduce 267) ((T_ENDWHILE) reduce 267) ((T_ENUM) reduce 267) ((T_EVAL) reduce 267) ((T_EXIT) reduce 267) ((T_EXTENDS) reduce 267) ((T_FILE) reduce 267) ((T_FINAL) reduce 267) ((T_FINALLY) reduce 267) ((T_FN) reduce 267) ((T_FOR) reduce 267) ((T_FOREACH) reduce 267) ((T_FUNCTION) reduce 267) ((T_FUNC_C) reduce 267) ((T_GLOBAL) reduce 267) ((T_GOTO) reduce 267) ((T_IF) reduce 267) ((T_IMPLEMENTS) reduce 267) ((T_INCLUDE) reduce 267) ((T_INCLUDE_ONCE) reduce 267) ((T_INSTANCEOF) reduce 267) ((T_INSTEADOF) reduce 267) ((T_INTERFACE) reduce 267) ((T_ISSET) reduce 267) ((T_LINE) reduce 267) ((T_LIST) reduce 267) ((T_LOGICAL_AND) reduce 267) ((T_LOGICAL_OR) reduce 267) ((T_LOGICAL_XOR) reduce 267) ((T_MATCH) reduce 267) ((T_METHOD_C) reduce 267) ((T_NAMESPACE) reduce 267) ((T_NEW) reduce 267) ((T_NS_C) reduce 267) ((T_PRINT) reduce 267) ((T_PRIVATE) reduce 267) ((T_PROPERTY_C) reduce 267) ((T_PROTECTED) reduce 267) ((T_PUBLIC) reduce 267) ((T_READONLY) reduce 267) ((T_REQUIRE) reduce 267) ((T_REQUIRE_ONCE) reduce 267) ((T_RETURN) reduce 267) ((T_STATIC) reduce 267) ((T_STRING) reduce 267) ((T_SWITCH) reduce 267) ((T_THROW) reduce 267) ((T_TRAIT) reduce 267) ((T_TRAIT_C) reduce 267) ((T_TRY) reduce 267) ((T_UNSET) reduce 267) ((T_USE) reduce 267) ((T_VAR) reduce 267) ((T_WHILE) reduce 267) ((T_YIELD) reduce 267) (("|") reduce 267)) 7164 (((T_ABSTRACT) reduce 92) ((T_ARRAY) reduce 92) ((T_AS) reduce 92) ((T_BREAK) reduce 92) ((T_CALLABLE) reduce 92) ((T_CASE) reduce 92) ((T_CATCH) reduce 92) ((T_CLASS) reduce 92) ((T_CLASS_C) reduce 92) ((T_CLONE) reduce 92) ((T_CONST) reduce 92) ((T_CONTINUE) reduce 92) ((T_DECLARE) reduce 92) ((T_DEFAULT) reduce 92) ((T_DIR) reduce 92) ((T_DO) reduce 92) ((T_ECHO) reduce 92) ((T_ELSE) reduce 92) ((T_ELSEIF) reduce 92) ((T_EMPTY) reduce 92) ((T_ENDDECLARE) reduce 92) ((T_ENDFOR) reduce 92) ((T_ENDFOREACH) reduce 92) ((T_ENDIF) reduce 92) ((T_ENDSWITCH) reduce 92) ((T_ENDWHILE) reduce 92) ((T_ENUM) reduce 92) ((T_EVAL) reduce 92) ((T_EXIT) reduce 92) ((T_EXTENDS) reduce 92) ((T_FILE) reduce 92) ((T_FINAL) reduce 92) ((T_FINALLY) reduce 92) ((T_FN) reduce 92) ((T_FOR) reduce 92) ((T_FOREACH) reduce 92) ((T_FUNCTION) reduce 92) ((T_FUNC_C) reduce 92) ((T_GLOBAL) reduce 92) ((T_GOTO) reduce 92) ((T_IF) reduce 92) ((T_IMPLEMENTS) reduce 92) ((T_INCLUDE) reduce 92) ((T_INCLUDE_ONCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_INSTEADOF) reduce 92) ((T_INTERFACE) reduce 92) ((T_ISSET) reduce 92) ((T_LINE) reduce 92) ((T_LIST) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_MATCH) reduce 92) ((T_METHOD_C) reduce 92) ((T_NAMESPACE) reduce 92) ((T_NEW) reduce 92) ((T_NS_C) reduce 92) ((T_PRINT) reduce 92) ((T_PRIVATE) reduce 92) ((T_PROPERTY_C) reduce 92) ((T_PROTECTED) reduce 92) ((T_PUBLIC) reduce 92) ((T_READONLY) reduce 92) ((T_REQUIRE) reduce 92) ((T_REQUIRE_ONCE) reduce 92) ((T_RETURN) reduce 92) ((T_STATIC) reduce 92) ((T_STRING) reduce 92) ((T_SWITCH) reduce 92) ((T_THROW) reduce 92) ((T_TRAIT) reduce 92) ((T_TRAIT_C) reduce 92) ((T_TRY) reduce 92) ((T_UNSET) reduce 92) ((T_USE) reduce 92) ((T_VAR) reduce 92) ((T_WHILE) reduce 92) ((T_YIELD) reduce 92) (("|") reduce 92)) 7165 (((T_ABSTRACT) reduce 280) ((T_ARRAY) reduce 280) ((T_AS) reduce 280) ((T_BREAK) reduce 280) ((T_CALLABLE) reduce 280) ((T_CASE) reduce 280) ((T_CATCH) reduce 280) ((T_CLASS) reduce 280) ((T_CLASS_C) reduce 280) ((T_CLONE) reduce 280) ((T_CONST) reduce 280) ((T_CONTINUE) reduce 280) ((T_DECLARE) reduce 280) ((T_DEFAULT) reduce 280) ((T_DIR) reduce 280) ((T_DO) reduce 280) ((T_ECHO) reduce 280) ((T_ELSE) reduce 280) ((T_ELSEIF) reduce 280) ((T_EMPTY) reduce 280) ((T_ENDDECLARE) reduce 280) ((T_ENDFOR) reduce 280) ((T_ENDFOREACH) reduce 280) ((T_ENDIF) reduce 280) ((T_ENDSWITCH) reduce 280) ((T_ENDWHILE) reduce 280) ((T_ENUM) reduce 280) ((T_EVAL) reduce 280) ((T_EXIT) reduce 280) ((T_EXTENDS) reduce 280) ((T_FILE) reduce 280) ((T_FINAL) reduce 280) ((T_FINALLY) reduce 280) ((T_FN) reduce 280) ((T_FOR) reduce 280) ((T_FOREACH) reduce 280) ((T_FUNCTION) reduce 280) ((T_FUNC_C) reduce 280) ((T_GLOBAL) reduce 280) ((T_GOTO) reduce 280) ((T_IF) reduce 280) ((T_IMPLEMENTS) reduce 280) ((T_INCLUDE) reduce 280) ((T_INCLUDE_ONCE) reduce 280) ((T_INSTANCEOF) reduce 280) ((T_INSTEADOF) reduce 280) ((T_INTERFACE) reduce 280) ((T_ISSET) reduce 280) ((T_LINE) reduce 280) ((T_LIST) reduce 280) ((T_LOGICAL_AND) reduce 280) ((T_LOGICAL_OR) reduce 280) ((T_LOGICAL_XOR) reduce 280) ((T_MATCH) reduce 280) ((T_METHOD_C) reduce 280) ((T_NAMESPACE) reduce 280) ((T_NEW) reduce 280) ((T_NS_C) reduce 280) ((T_PRINT) reduce 280) ((T_PRIVATE) reduce 280) ((T_PROPERTY_C) reduce 280) ((T_PROTECTED) reduce 280) ((T_PUBLIC) reduce 280) ((T_READONLY) reduce 280) ((T_REQUIRE) reduce 280) ((T_REQUIRE_ONCE) reduce 280) ((T_RETURN) reduce 280) ((T_STATIC) reduce 280) ((T_STRING) reduce 280) ((T_SWITCH) reduce 280) ((T_THROW) reduce 280) ((T_TRAIT) reduce 280) ((T_TRAIT_C) reduce 280) ((T_TRY) reduce 280) ((T_UNSET) reduce 280) ((T_USE) reduce 280) ((T_VAR) reduce 280) ((T_WHILE) reduce 280) ((T_YIELD) reduce 280) (("|") reduce 280)) 7166 (((T_ABSTRACT) reduce 268) ((T_ARRAY) reduce 268) ((T_AS) reduce 268) ((T_BREAK) reduce 268) ((T_CALLABLE) reduce 268) ((T_CASE) reduce 268) ((T_CATCH) reduce 268) ((T_CLASS) reduce 268) ((T_CLASS_C) reduce 268) ((T_CLONE) reduce 268) ((T_CONST) reduce 268) ((T_CONTINUE) reduce 268) ((T_DECLARE) reduce 268) ((T_DEFAULT) reduce 268) ((T_DIR) reduce 268) ((T_DO) reduce 268) ((T_ECHO) reduce 268) ((T_ELSE) reduce 268) ((T_ELSEIF) reduce 268) ((T_EMPTY) reduce 268) ((T_ENDDECLARE) reduce 268) ((T_ENDFOR) reduce 268) ((T_ENDFOREACH) reduce 268) ((T_ENDIF) reduce 268) ((T_ENDSWITCH) reduce 268) ((T_ENDWHILE) reduce 268) ((T_ENUM) reduce 268) ((T_EVAL) reduce 268) ((T_EXIT) reduce 268) ((T_EXTENDS) reduce 268) ((T_FILE) reduce 268) ((T_FINAL) reduce 268) ((T_FINALLY) reduce 268) ((T_FN) reduce 268) ((T_FOR) reduce 268) ((T_FOREACH) reduce 268) ((T_FUNCTION) reduce 268) ((T_FUNC_C) reduce 268) ((T_GLOBAL) reduce 268) ((T_GOTO) reduce 268) ((T_IF) reduce 268) ((T_IMPLEMENTS) reduce 268) ((T_INCLUDE) reduce 268) ((T_INCLUDE_ONCE) reduce 268) ((T_INSTANCEOF) reduce 268) ((T_INSTEADOF) reduce 268) ((T_INTERFACE) reduce 268) ((T_ISSET) reduce 268) ((T_LINE) reduce 268) ((T_LIST) reduce 268) ((T_LOGICAL_AND) reduce 268) ((T_LOGICAL_OR) reduce 268) ((T_LOGICAL_XOR) reduce 268) ((T_MATCH) reduce 268) ((T_METHOD_C) reduce 268) ((T_NAMESPACE) reduce 268) ((T_NEW) reduce 268) ((T_NS_C) reduce 268) ((T_PRINT) reduce 268) ((T_PRIVATE) reduce 268) ((T_PROPERTY_C) reduce 268) ((T_PROTECTED) reduce 268) ((T_PUBLIC) reduce 268) ((T_READONLY) reduce 268) ((T_REQUIRE) reduce 268) ((T_REQUIRE_ONCE) reduce 268) ((T_RETURN) reduce 268) ((T_STATIC) reduce 268) ((T_STRING) reduce 268) ((T_SWITCH) reduce 268) ((T_THROW) reduce 268) ((T_TRAIT) reduce 268) ((T_TRAIT_C) reduce 268) ((T_TRY) reduce 268) ((T_UNSET) reduce 268) ((T_USE) reduce 268) ((T_VAR) reduce 268) ((T_WHILE) reduce 268) ((T_YIELD) reduce 268) (("|") reduce 268)) 7167 (((T_ABSTRACT) reduce 266) ((T_ARRAY) reduce 266) ((T_AS) reduce 266) ((T_BREAK) reduce 266) ((T_CALLABLE) reduce 266) ((T_CASE) reduce 266) ((T_CATCH) reduce 266) ((T_CLASS) reduce 266) ((T_CLASS_C) reduce 266) ((T_CLONE) reduce 266) ((T_CONST) reduce 266) ((T_CONTINUE) reduce 266) ((T_DECLARE) reduce 266) ((T_DEFAULT) reduce 266) ((T_DIR) reduce 266) ((T_DO) reduce 266) ((T_ECHO) reduce 266) ((T_ELSE) reduce 266) ((T_ELSEIF) reduce 266) ((T_EMPTY) reduce 266) ((T_ENDDECLARE) reduce 266) ((T_ENDFOR) reduce 266) ((T_ENDFOREACH) reduce 266) ((T_ENDIF) reduce 266) ((T_ENDSWITCH) reduce 266) ((T_ENDWHILE) reduce 266) ((T_ENUM) reduce 266) ((T_EVAL) reduce 266) ((T_EXIT) reduce 266) ((T_EXTENDS) reduce 266) ((T_FILE) reduce 266) ((T_FINAL) reduce 266) ((T_FINALLY) reduce 266) ((T_FN) reduce 266) ((T_FOR) reduce 266) ((T_FOREACH) reduce 266) ((T_FUNCTION) reduce 266) ((T_FUNC_C) reduce 266) ((T_GLOBAL) reduce 266) ((T_GOTO) reduce 266) ((T_IF) reduce 266) ((T_IMPLEMENTS) reduce 266) ((T_INCLUDE) reduce 266) ((T_INCLUDE_ONCE) reduce 266) ((T_INSTANCEOF) reduce 266) ((T_INSTEADOF) reduce 266) ((T_INTERFACE) reduce 266) ((T_ISSET) reduce 266) ((T_LINE) reduce 266) ((T_LIST) reduce 266) ((T_LOGICAL_AND) reduce 266) ((T_LOGICAL_OR) reduce 266) ((T_LOGICAL_XOR) reduce 266) ((T_MATCH) reduce 266) ((T_METHOD_C) reduce 266) ((T_NAMESPACE) reduce 266) ((T_NEW) reduce 266) ((T_NS_C) reduce 266) ((T_PRINT) reduce 266) ((T_PRIVATE) reduce 266) ((T_PROPERTY_C) reduce 266) ((T_PROTECTED) reduce 266) ((T_PUBLIC) reduce 266) ((T_READONLY) reduce 266) ((T_REQUIRE) reduce 266) ((T_REQUIRE_ONCE) reduce 266) ((T_RETURN) reduce 266) ((T_STATIC) reduce 266) ((T_STRING) reduce 266) ((T_SWITCH) reduce 266) ((T_THROW) reduce 266) ((T_TRAIT) reduce 266) ((T_TRAIT_C) reduce 266) ((T_TRY) reduce 266) ((T_UNSET) reduce 266) ((T_USE) reduce 266) ((T_VAR) reduce 266) ((T_WHILE) reduce 266) ((T_YIELD) reduce 266) (("|") reduce 266)) 7168 (((T_ABSTRACT) reduce 270) ((T_ARRAY) reduce 270) ((T_AS) reduce 270) ((T_BREAK) reduce 270) ((T_CALLABLE) reduce 270) ((T_CASE) reduce 270) ((T_CATCH) reduce 270) ((T_CLASS) reduce 270) ((T_CLASS_C) reduce 270) ((T_CLONE) reduce 270) ((T_CONST) reduce 270) ((T_CONTINUE) reduce 270) ((T_DECLARE) reduce 270) ((T_DEFAULT) reduce 270) ((T_DIR) reduce 270) ((T_DO) reduce 270) ((T_ECHO) reduce 270) ((T_ELSE) reduce 270) ((T_ELSEIF) reduce 270) ((T_EMPTY) reduce 270) ((T_ENDDECLARE) reduce 270) ((T_ENDFOR) reduce 270) ((T_ENDFOREACH) reduce 270) ((T_ENDIF) reduce 270) ((T_ENDSWITCH) reduce 270) ((T_ENDWHILE) reduce 270) ((T_ENUM) reduce 270) ((T_EVAL) reduce 270) ((T_EXIT) reduce 270) ((T_EXTENDS) reduce 270) ((T_FILE) reduce 270) ((T_FINAL) reduce 270) ((T_FINALLY) reduce 270) ((T_FN) reduce 270) ((T_FOR) reduce 270) ((T_FOREACH) reduce 270) ((T_FUNCTION) reduce 270) ((T_FUNC_C) reduce 270) ((T_GLOBAL) reduce 270) ((T_GOTO) reduce 270) ((T_IF) reduce 270) ((T_IMPLEMENTS) reduce 270) ((T_INCLUDE) reduce 270) ((T_INCLUDE_ONCE) reduce 270) ((T_INSTANCEOF) reduce 270) ((T_INSTEADOF) reduce 270) ((T_INTERFACE) reduce 270) ((T_ISSET) reduce 270) ((T_LINE) reduce 270) ((T_LIST) reduce 270) ((T_LOGICAL_AND) reduce 270) ((T_LOGICAL_OR) reduce 270) ((T_LOGICAL_XOR) reduce 270) ((T_MATCH) reduce 270) ((T_METHOD_C) reduce 270) ((T_NAMESPACE) reduce 270) ((T_NEW) reduce 270) ((T_NS_C) reduce 270) ((T_PRINT) reduce 270) ((T_PRIVATE) reduce 270) ((T_PROPERTY_C) reduce 270) ((T_PROTECTED) reduce 270) ((T_PUBLIC) reduce 270) ((T_READONLY) reduce 270) ((T_REQUIRE) reduce 270) ((T_REQUIRE_ONCE) reduce 270) ((T_RETURN) reduce 270) ((T_STATIC) reduce 270) ((T_STRING) reduce 270) ((T_SWITCH) reduce 270) ((T_THROW) reduce 270) ((T_TRAIT) reduce 270) ((T_TRAIT_C) reduce 270) ((T_TRY) reduce 270) ((T_UNSET) reduce 270) ((T_USE) reduce 270) ((T_VAR) reduce 270) ((T_WHILE) reduce 270) ((T_YIELD) reduce 270) (("|") reduce 270)) 7170 (((T_ABSTRACT) reduce 269) ((T_ARRAY) reduce 269) ((T_AS) reduce 269) ((T_BREAK) reduce 269) ((T_CALLABLE) reduce 269) ((T_CASE) reduce 269) ((T_CATCH) reduce 269) ((T_CLASS) reduce 269) ((T_CLASS_C) reduce 269) ((T_CLONE) reduce 269) ((T_CONST) reduce 269) ((T_CONTINUE) reduce 269) ((T_DECLARE) reduce 269) ((T_DEFAULT) reduce 269) ((T_DIR) reduce 269) ((T_DO) reduce 269) ((T_ECHO) reduce 269) ((T_ELSE) reduce 269) ((T_ELSEIF) reduce 269) ((T_EMPTY) reduce 269) ((T_ENDDECLARE) reduce 269) ((T_ENDFOR) reduce 269) ((T_ENDFOREACH) reduce 269) ((T_ENDIF) reduce 269) ((T_ENDSWITCH) reduce 269) ((T_ENDWHILE) reduce 269) ((T_ENUM) reduce 269) ((T_EVAL) reduce 269) ((T_EXIT) reduce 269) ((T_EXTENDS) reduce 269) ((T_FILE) reduce 269) ((T_FINAL) reduce 269) ((T_FINALLY) reduce 269) ((T_FN) reduce 269) ((T_FOR) reduce 269) ((T_FOREACH) reduce 269) ((T_FUNCTION) reduce 269) ((T_FUNC_C) reduce 269) ((T_GLOBAL) reduce 269) ((T_GOTO) reduce 269) ((T_IF) reduce 269) ((T_IMPLEMENTS) reduce 269) ((T_INCLUDE) reduce 269) ((T_INCLUDE_ONCE) reduce 269) ((T_INSTANCEOF) reduce 269) ((T_INSTEADOF) reduce 269) ((T_INTERFACE) reduce 269) ((T_ISSET) reduce 269) ((T_LINE) reduce 269) ((T_LIST) reduce 269) ((T_LOGICAL_AND) reduce 269) ((T_LOGICAL_OR) reduce 269) ((T_LOGICAL_XOR) reduce 269) ((T_MATCH) reduce 269) ((T_METHOD_C) reduce 269) ((T_NAMESPACE) reduce 269) ((T_NEW) reduce 269) ((T_NS_C) reduce 269) ((T_PRINT) reduce 269) ((T_PRIVATE) reduce 269) ((T_PROPERTY_C) reduce 269) ((T_PROTECTED) reduce 269) ((T_PUBLIC) reduce 269) ((T_READONLY) reduce 269) ((T_REQUIRE) reduce 269) ((T_REQUIRE_ONCE) reduce 269) ((T_RETURN) reduce 269) ((T_STATIC) reduce 269) ((T_STRING) reduce 269) ((T_SWITCH) reduce 269) ((T_THROW) reduce 269) ((T_TRAIT) reduce 269) ((T_TRAIT_C) reduce 269) ((T_TRY) reduce 269) ((T_UNSET) reduce 269) ((T_USE) reduce 269) ((T_VAR) reduce 269) ((T_WHILE) reduce 269) ((T_YIELD) reduce 269) (("|") reduce 269)) 7172 (((T_ABSTRACT) reduce 271) ((T_ARRAY) reduce 271) ((T_AS) reduce 271) ((T_BREAK) reduce 271) ((T_CALLABLE) reduce 271) ((T_CASE) reduce 271) ((T_CATCH) reduce 271) ((T_CLASS) reduce 271) ((T_CLASS_C) reduce 271) ((T_CLONE) reduce 271) ((T_CONST) reduce 271) ((T_CONTINUE) reduce 271) ((T_DECLARE) reduce 271) ((T_DEFAULT) reduce 271) ((T_DIR) reduce 271) ((T_DO) reduce 271) ((T_ECHO) reduce 271) ((T_ELSE) reduce 271) ((T_ELSEIF) reduce 271) ((T_EMPTY) reduce 271) ((T_ENDDECLARE) reduce 271) ((T_ENDFOR) reduce 271) ((T_ENDFOREACH) reduce 271) ((T_ENDIF) reduce 271) ((T_ENDSWITCH) reduce 271) ((T_ENDWHILE) reduce 271) ((T_ENUM) reduce 271) ((T_EVAL) reduce 271) ((T_EXIT) reduce 271) ((T_EXTENDS) reduce 271) ((T_FILE) reduce 271) ((T_FINAL) reduce 271) ((T_FINALLY) reduce 271) ((T_FN) reduce 271) ((T_FOR) reduce 271) ((T_FOREACH) reduce 271) ((T_FUNCTION) reduce 271) ((T_FUNC_C) reduce 271) ((T_GLOBAL) reduce 271) ((T_GOTO) reduce 271) ((T_IF) reduce 271) ((T_IMPLEMENTS) reduce 271) ((T_INCLUDE) reduce 271) ((T_INCLUDE_ONCE) reduce 271) ((T_INSTANCEOF) reduce 271) ((T_INSTEADOF) reduce 271) ((T_INTERFACE) reduce 271) ((T_ISSET) reduce 271) ((T_LINE) reduce 271) ((T_LIST) reduce 271) ((T_LOGICAL_AND) reduce 271) ((T_LOGICAL_OR) reduce 271) ((T_LOGICAL_XOR) reduce 271) ((T_MATCH) reduce 271) ((T_METHOD_C) reduce 271) ((T_NAMESPACE) reduce 271) ((T_NEW) reduce 271) ((T_NS_C) reduce 271) ((T_PRINT) reduce 271) ((T_PRIVATE) reduce 271) ((T_PROPERTY_C) reduce 271) ((T_PROTECTED) reduce 271) ((T_PUBLIC) reduce 271) ((T_READONLY) reduce 271) ((T_REQUIRE) reduce 271) ((T_REQUIRE_ONCE) reduce 271) ((T_RETURN) reduce 271) ((T_STATIC) reduce 271) ((T_STRING) reduce 271) ((T_SWITCH) reduce 271) ((T_THROW) reduce 271) ((T_TRAIT) reduce 271) ((T_TRAIT_C) reduce 271) ((T_TRY) reduce 271) ((T_UNSET) reduce 271) ((T_USE) reduce 271) ((T_VAR) reduce 271) ((T_WHILE) reduce 271) ((T_YIELD) reduce 271) (("|") reduce 271)) 7173 ((("=") reduce 52)) 7174 ((("=") reduce 53)) 7175 ((("=") reduce 73)) 7179 (((T_ABSTRACT) reduce 309) ((T_ATTRIBUTE) reduce 309) ((T_CASE) reduce 309) ((T_CONST) reduce 309) ((T_FINAL) reduce 309) ((T_FUNCTION) reduce 309) ((T_PRIVATE) reduce 309) ((T_PRIVATE_SET) reduce 309) ((T_PROTECTED) reduce 309) ((T_PROTECTED_SET) reduce 309) ((T_PUBLIC) reduce 309) ((T_PUBLIC_SET) reduce 309) ((T_READONLY) reduce 309) ((T_STATIC) reduce 309) ((T_USE) reduce 309) ((T_VAR) reduce 309) (("}") reduce 309)) 7180 (((",") reduce 371) ((";") reduce 371)) 7183 (((",") reduce 373) ((";") reduce 373)) 7185 (((T_ABSTRACT) reduce 278) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_ARRAY) reduce 278) ((T_AS) reduce 278) ((T_BREAK) reduce 278) ((T_CALLABLE) reduce 278) ((T_CASE) reduce 278) ((T_CATCH) reduce 278) ((T_CLASS) reduce 278) ((T_CLASS_C) reduce 278) ((T_CLONE) reduce 278) ((T_CONST) reduce 278) ((T_CONTINUE) reduce 278) ((T_DECLARE) reduce 278) ((T_DEFAULT) reduce 278) ((T_DIR) reduce 278) ((T_DO) reduce 278) ((T_ECHO) reduce 278) ((T_ELSE) reduce 278) ((T_ELSEIF) reduce 278) ((T_EMPTY) reduce 278) ((T_ENDDECLARE) reduce 278) ((T_ENDFOR) reduce 278) ((T_ENDFOREACH) reduce 278) ((T_ENDIF) reduce 278) ((T_ENDSWITCH) reduce 278) ((T_ENDWHILE) reduce 278) ((T_ENUM) reduce 278) ((T_EVAL) reduce 278) ((T_EXIT) reduce 278) ((T_EXTENDS) reduce 278) ((T_FILE) reduce 278) ((T_FINAL) reduce 278) ((T_FINALLY) reduce 278) ((T_FN) reduce 278) ((T_FOR) reduce 278) ((T_FOREACH) reduce 278) ((T_FUNCTION) reduce 278) ((T_FUNC_C) reduce 278) ((T_GLOBAL) reduce 278) ((T_GOTO) reduce 278) ((T_IF) reduce 278) ((T_IMPLEMENTS) reduce 278) ((T_INCLUDE) reduce 278) ((T_INCLUDE_ONCE) reduce 278) ((T_INSTANCEOF) reduce 278) ((T_INSTEADOF) reduce 278) ((T_INTERFACE) reduce 278) ((T_ISSET) reduce 278) ((T_LINE) reduce 278) ((T_LIST) reduce 278) ((T_LOGICAL_AND) reduce 278) ((T_LOGICAL_OR) reduce 278) ((T_LOGICAL_XOR) reduce 278) ((T_MATCH) reduce 278) ((T_METHOD_C) reduce 278) ((T_NAMESPACE) reduce 278) ((T_NEW) reduce 278) ((T_NS_C) reduce 278) ((T_PRINT) reduce 278) ((T_PRIVATE) reduce 278) ((T_PROPERTY_C) reduce 278) ((T_PROTECTED) reduce 278) ((T_PUBLIC) reduce 278) ((T_READONLY) reduce 278) ((T_REQUIRE) reduce 278) ((T_REQUIRE_ONCE) reduce 278) ((T_RETURN) reduce 278) ((T_STATIC) reduce 278) ((T_STRING) reduce 278) ((T_SWITCH) reduce 278) ((T_THROW) reduce 278) ((T_TRAIT) reduce 278) ((T_TRAIT_C) reduce 278) ((T_TRY) reduce 278) ((T_UNSET) reduce 278) ((T_USE) reduce 278) ((T_VAR) reduce 278) ((T_WHILE) reduce 278) ((T_YIELD) reduce 278)) 7186 (((T_ABSTRACT) reduce 279) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) ((T_ARRAY) reduce 279) ((T_AS) reduce 279) ((T_BREAK) reduce 279) ((T_CALLABLE) reduce 279) ((T_CASE) reduce 279) ((T_CATCH) reduce 279) ((T_CLASS) reduce 279) ((T_CLASS_C) reduce 279) ((T_CLONE) reduce 279) ((T_CONST) reduce 279) ((T_CONTINUE) reduce 279) ((T_DECLARE) reduce 279) ((T_DEFAULT) reduce 279) ((T_DIR) reduce 279) ((T_DO) reduce 279) ((T_ECHO) reduce 279) ((T_ELSE) reduce 279) ((T_ELSEIF) reduce 279) ((T_EMPTY) reduce 279) ((T_ENDDECLARE) reduce 279) ((T_ENDFOR) reduce 279) ((T_ENDFOREACH) reduce 279) ((T_ENDIF) reduce 279) ((T_ENDSWITCH) reduce 279) ((T_ENDWHILE) reduce 279) ((T_ENUM) reduce 279) ((T_EVAL) reduce 279) ((T_EXIT) reduce 279) ((T_EXTENDS) reduce 279) ((T_FILE) reduce 279) ((T_FINAL) reduce 279) ((T_FINALLY) reduce 279) ((T_FN) reduce 279) ((T_FOR) reduce 279) ((T_FOREACH) reduce 279) ((T_FUNCTION) reduce 279) ((T_FUNC_C) reduce 279) ((T_GLOBAL) reduce 279) ((T_GOTO) reduce 279) ((T_IF) reduce 279) ((T_IMPLEMENTS) reduce 279) ((T_INCLUDE) reduce 279) ((T_INCLUDE_ONCE) reduce 279) ((T_INSTANCEOF) reduce 279) ((T_INSTEADOF) reduce 279) ((T_INTERFACE) reduce 279) ((T_ISSET) reduce 279) ((T_LINE) reduce 279) ((T_LIST) reduce 279) ((T_LOGICAL_AND) reduce 279) ((T_LOGICAL_OR) reduce 279) ((T_LOGICAL_XOR) reduce 279) ((T_MATCH) reduce 279) ((T_METHOD_C) reduce 279) ((T_NAMESPACE) reduce 279) ((T_NEW) reduce 279) ((T_NS_C) reduce 279) ((T_PRINT) reduce 279) ((T_PRIVATE) reduce 279) ((T_PROPERTY_C) reduce 279) ((T_PROTECTED) reduce 279) ((T_PUBLIC) reduce 279) ((T_READONLY) reduce 279) ((T_REQUIRE) reduce 279) ((T_REQUIRE_ONCE) reduce 279) ((T_RETURN) reduce 279) ((T_STATIC) reduce 279) ((T_STRING) reduce 279) ((T_SWITCH) reduce 279) ((T_THROW) reduce 279) ((T_TRAIT) reduce 279) ((T_TRAIT_C) reduce 279) ((T_TRY) reduce 279) ((T_UNSET) reduce 279) ((T_USE) reduce 279) ((T_VAR) reduce 279) ((T_WHILE) reduce 279) ((T_YIELD) reduce 279)) 7187 (((T_ABSTRACT) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ARRAY) reduce 94) ((T_AS) reduce 94) ((T_BREAK) reduce 94) ((T_CALLABLE) reduce 94) ((T_CASE) reduce 94) ((T_CATCH) reduce 94) ((T_CLASS) reduce 94) ((T_CLASS_C) reduce 94) ((T_CLONE) reduce 94) ((T_CONST) reduce 94) ((T_CONTINUE) reduce 94) ((T_DECLARE) reduce 94) ((T_DEFAULT) reduce 94) ((T_DIR) reduce 94) ((T_DO) reduce 94) ((T_ECHO) reduce 94) ((T_ELSE) reduce 94) ((T_ELSEIF) reduce 94) ((T_EMPTY) reduce 94) ((T_ENDDECLARE) reduce 94) ((T_ENDFOR) reduce 94) ((T_ENDFOREACH) reduce 94) ((T_ENDIF) reduce 94) ((T_ENDSWITCH) reduce 94) ((T_ENDWHILE) reduce 94) ((T_ENUM) reduce 94) ((T_EVAL) reduce 94) ((T_EXIT) reduce 94) ((T_EXTENDS) reduce 94) ((T_FILE) reduce 94) ((T_FINAL) reduce 94) ((T_FINALLY) reduce 94) ((T_FN) reduce 94) ((T_FOR) reduce 94) ((T_FOREACH) reduce 94) ((T_FUNCTION) reduce 94) ((T_FUNC_C) reduce 94) ((T_GLOBAL) reduce 94) ((T_GOTO) reduce 94) ((T_IF) reduce 94) ((T_IMPLEMENTS) reduce 94) ((T_INCLUDE) reduce 94) ((T_INCLUDE_ONCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_INSTEADOF) reduce 94) ((T_INTERFACE) reduce 94) ((T_ISSET) reduce 94) ((T_LINE) reduce 94) ((T_LIST) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_MATCH) reduce 94) ((T_METHOD_C) reduce 94) ((T_NAMESPACE) reduce 94) ((T_NEW) reduce 94) ((T_NS_C) reduce 94) ((T_PRINT) reduce 94) ((T_PRIVATE) reduce 94) ((T_PROPERTY_C) reduce 94) ((T_PROTECTED) reduce 94) ((T_PUBLIC) reduce 94) ((T_READONLY) reduce 94) ((T_REQUIRE) reduce 94) ((T_REQUIRE_ONCE) reduce 94) ((T_RETURN) reduce 94) ((T_STATIC) reduce 94) ((T_STRING) reduce 94) ((T_SWITCH) reduce 94) ((T_THROW) reduce 94) ((T_TRAIT) reduce 94) ((T_TRAIT_C) reduce 94) ((T_TRY) reduce 94) ((T_UNSET) reduce 94) ((T_USE) reduce 94) ((T_VAR) reduce 94) ((T_WHILE) reduce 94) ((T_YIELD) reduce 94)) 7188 (((T_ABSTRACT) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ARRAY) reduce 93) ((T_AS) reduce 93) ((T_BREAK) reduce 93) ((T_CALLABLE) reduce 93) ((T_CASE) reduce 93) ((T_CATCH) reduce 93) ((T_CLASS) reduce 93) ((T_CLASS_C) reduce 93) ((T_CLONE) reduce 93) ((T_CONST) reduce 93) ((T_CONTINUE) reduce 93) ((T_DECLARE) reduce 93) ((T_DEFAULT) reduce 93) ((T_DIR) reduce 93) ((T_DO) reduce 93) ((T_ECHO) reduce 93) ((T_ELSE) reduce 93) ((T_ELSEIF) reduce 93) ((T_EMPTY) reduce 93) ((T_ENDDECLARE) reduce 93) ((T_ENDFOR) reduce 93) ((T_ENDFOREACH) reduce 93) ((T_ENDIF) reduce 93) ((T_ENDSWITCH) reduce 93) ((T_ENDWHILE) reduce 93) ((T_ENUM) reduce 93) ((T_EVAL) reduce 93) ((T_EXIT) reduce 93) ((T_EXTENDS) reduce 93) ((T_FILE) reduce 93) ((T_FINAL) reduce 93) ((T_FINALLY) reduce 93) ((T_FN) reduce 93) ((T_FOR) reduce 93) ((T_FOREACH) reduce 93) ((T_FUNCTION) reduce 93) ((T_FUNC_C) reduce 93) ((T_GLOBAL) reduce 93) ((T_GOTO) reduce 93) ((T_IF) reduce 93) ((T_IMPLEMENTS) reduce 93) ((T_INCLUDE) reduce 93) ((T_INCLUDE_ONCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_INSTEADOF) reduce 93) ((T_INTERFACE) reduce 93) ((T_ISSET) reduce 93) ((T_LINE) reduce 93) ((T_LIST) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_MATCH) reduce 93) ((T_METHOD_C) reduce 93) ((T_NAMESPACE) reduce 93) ((T_NEW) reduce 93) ((T_NS_C) reduce 93) ((T_PRINT) reduce 93) ((T_PRIVATE) reduce 93) ((T_PROPERTY_C) reduce 93) ((T_PROTECTED) reduce 93) ((T_PUBLIC) reduce 93) ((T_READONLY) reduce 93) ((T_REQUIRE) reduce 93) ((T_REQUIRE_ONCE) reduce 93) ((T_RETURN) reduce 93) ((T_STATIC) reduce 93) ((T_STRING) reduce 93) ((T_SWITCH) reduce 93) ((T_THROW) reduce 93) ((T_TRAIT) reduce 93) ((T_TRAIT_C) reduce 93) ((T_TRY) reduce 93) ((T_UNSET) reduce 93) ((T_USE) reduce 93) ((T_VAR) reduce 93) ((T_WHILE) reduce 93) ((T_YIELD) reduce 93)) 7189 (((T_ABSTRACT) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_ARRAY) reduce 95) ((T_AS) reduce 95) ((T_BREAK) reduce 95) ((T_CALLABLE) reduce 95) ((T_CASE) reduce 95) ((T_CATCH) reduce 95) ((T_CLASS) reduce 95) ((T_CLASS_C) reduce 95) ((T_CLONE) reduce 95) ((T_CONST) reduce 95) ((T_CONTINUE) reduce 95) ((T_DECLARE) reduce 95) ((T_DEFAULT) reduce 95) ((T_DIR) reduce 95) ((T_DO) reduce 95) ((T_ECHO) reduce 95) ((T_ELSE) reduce 95) ((T_ELSEIF) reduce 95) ((T_EMPTY) reduce 95) ((T_ENDDECLARE) reduce 95) ((T_ENDFOR) reduce 95) ((T_ENDFOREACH) reduce 95) ((T_ENDIF) reduce 95) ((T_ENDSWITCH) reduce 95) ((T_ENDWHILE) reduce 95) ((T_ENUM) reduce 95) ((T_EVAL) reduce 95) ((T_EXIT) reduce 95) ((T_EXTENDS) reduce 95) ((T_FILE) reduce 95) ((T_FINAL) reduce 95) ((T_FINALLY) reduce 95) ((T_FN) reduce 95) ((T_FOR) reduce 95) ((T_FOREACH) reduce 95) ((T_FUNCTION) reduce 95) ((T_FUNC_C) reduce 95) ((T_GLOBAL) reduce 95) ((T_GOTO) reduce 95) ((T_IF) reduce 95) ((T_IMPLEMENTS) reduce 95) ((T_INCLUDE) reduce 95) ((T_INCLUDE_ONCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_INSTEADOF) reduce 95) ((T_INTERFACE) reduce 95) ((T_ISSET) reduce 95) ((T_LINE) reduce 95) ((T_LIST) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_MATCH) reduce 95) ((T_METHOD_C) reduce 95) ((T_NAMESPACE) reduce 95) ((T_NEW) reduce 95) ((T_NS_C) reduce 95) ((T_PRINT) reduce 95) ((T_PRIVATE) reduce 95) ((T_PROPERTY_C) reduce 95) ((T_PROTECTED) reduce 95) ((T_PUBLIC) reduce 95) ((T_READONLY) reduce 95) ((T_REQUIRE) reduce 95) ((T_REQUIRE_ONCE) reduce 95) ((T_RETURN) reduce 95) ((T_STATIC) reduce 95) ((T_STRING) reduce 95) ((T_SWITCH) reduce 95) ((T_THROW) reduce 95) ((T_TRAIT) reduce 95) ((T_TRAIT_C) reduce 95) ((T_TRY) reduce 95) ((T_UNSET) reduce 95) ((T_USE) reduce 95) ((T_VAR) reduce 95) ((T_WHILE) reduce 95) ((T_YIELD) reduce 95)) 7190 (((T_ABSTRACT) reduce 267) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267) ((T_ARRAY) reduce 267) ((T_AS) reduce 267) ((T_BREAK) reduce 267) ((T_CALLABLE) reduce 267) ((T_CASE) reduce 267) ((T_CATCH) reduce 267) ((T_CLASS) reduce 267) ((T_CLASS_C) reduce 267) ((T_CLONE) reduce 267) ((T_CONST) reduce 267) ((T_CONTINUE) reduce 267) ((T_DECLARE) reduce 267) ((T_DEFAULT) reduce 267) ((T_DIR) reduce 267) ((T_DO) reduce 267) ((T_ECHO) reduce 267) ((T_ELSE) reduce 267) ((T_ELSEIF) reduce 267) ((T_EMPTY) reduce 267) ((T_ENDDECLARE) reduce 267) ((T_ENDFOR) reduce 267) ((T_ENDFOREACH) reduce 267) ((T_ENDIF) reduce 267) ((T_ENDSWITCH) reduce 267) ((T_ENDWHILE) reduce 267) ((T_ENUM) reduce 267) ((T_EVAL) reduce 267) ((T_EXIT) reduce 267) ((T_EXTENDS) reduce 267) ((T_FILE) reduce 267) ((T_FINAL) reduce 267) ((T_FINALLY) reduce 267) ((T_FN) reduce 267) ((T_FOR) reduce 267) ((T_FOREACH) reduce 267) ((T_FUNCTION) reduce 267) ((T_FUNC_C) reduce 267) ((T_GLOBAL) reduce 267) ((T_GOTO) reduce 267) ((T_IF) reduce 267) ((T_IMPLEMENTS) reduce 267) ((T_INCLUDE) reduce 267) ((T_INCLUDE_ONCE) reduce 267) ((T_INSTANCEOF) reduce 267) ((T_INSTEADOF) reduce 267) ((T_INTERFACE) reduce 267) ((T_ISSET) reduce 267) ((T_LINE) reduce 267) ((T_LIST) reduce 267) ((T_LOGICAL_AND) reduce 267) ((T_LOGICAL_OR) reduce 267) ((T_LOGICAL_XOR) reduce 267) ((T_MATCH) reduce 267) ((T_METHOD_C) reduce 267) ((T_NAMESPACE) reduce 267) ((T_NEW) reduce 267) ((T_NS_C) reduce 267) ((T_PRINT) reduce 267) ((T_PRIVATE) reduce 267) ((T_PROPERTY_C) reduce 267) ((T_PROTECTED) reduce 267) ((T_PUBLIC) reduce 267) ((T_READONLY) reduce 267) ((T_REQUIRE) reduce 267) ((T_REQUIRE_ONCE) reduce 267) ((T_RETURN) reduce 267) ((T_STATIC) reduce 267) ((T_STRING) reduce 267) ((T_SWITCH) reduce 267) ((T_THROW) reduce 267) ((T_TRAIT) reduce 267) ((T_TRAIT_C) reduce 267) ((T_TRY) reduce 267) ((T_UNSET) reduce 267) ((T_USE) reduce 267) ((T_VAR) reduce 267) ((T_WHILE) reduce 267) ((T_YIELD) reduce 267)) 7191 (((T_ABSTRACT) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ARRAY) reduce 92) ((T_AS) reduce 92) ((T_BREAK) reduce 92) ((T_CALLABLE) reduce 92) ((T_CASE) reduce 92) ((T_CATCH) reduce 92) ((T_CLASS) reduce 92) ((T_CLASS_C) reduce 92) ((T_CLONE) reduce 92) ((T_CONST) reduce 92) ((T_CONTINUE) reduce 92) ((T_DECLARE) reduce 92) ((T_DEFAULT) reduce 92) ((T_DIR) reduce 92) ((T_DO) reduce 92) ((T_ECHO) reduce 92) ((T_ELSE) reduce 92) ((T_ELSEIF) reduce 92) ((T_EMPTY) reduce 92) ((T_ENDDECLARE) reduce 92) ((T_ENDFOR) reduce 92) ((T_ENDFOREACH) reduce 92) ((T_ENDIF) reduce 92) ((T_ENDSWITCH) reduce 92) ((T_ENDWHILE) reduce 92) ((T_ENUM) reduce 92) ((T_EVAL) reduce 92) ((T_EXIT) reduce 92) ((T_EXTENDS) reduce 92) ((T_FILE) reduce 92) ((T_FINAL) reduce 92) ((T_FINALLY) reduce 92) ((T_FN) reduce 92) ((T_FOR) reduce 92) ((T_FOREACH) reduce 92) ((T_FUNCTION) reduce 92) ((T_FUNC_C) reduce 92) ((T_GLOBAL) reduce 92) ((T_GOTO) reduce 92) ((T_IF) reduce 92) ((T_IMPLEMENTS) reduce 92) ((T_INCLUDE) reduce 92) ((T_INCLUDE_ONCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_INSTEADOF) reduce 92) ((T_INTERFACE) reduce 92) ((T_ISSET) reduce 92) ((T_LINE) reduce 92) ((T_LIST) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_MATCH) reduce 92) ((T_METHOD_C) reduce 92) ((T_NAMESPACE) reduce 92) ((T_NEW) reduce 92) ((T_NS_C) reduce 92) ((T_PRINT) reduce 92) ((T_PRIVATE) reduce 92) ((T_PROPERTY_C) reduce 92) ((T_PROTECTED) reduce 92) ((T_PUBLIC) reduce 92) ((T_READONLY) reduce 92) ((T_REQUIRE) reduce 92) ((T_REQUIRE_ONCE) reduce 92) ((T_RETURN) reduce 92) ((T_STATIC) reduce 92) ((T_STRING) reduce 92) ((T_SWITCH) reduce 92) ((T_THROW) reduce 92) ((T_TRAIT) reduce 92) ((T_TRAIT_C) reduce 92) ((T_TRY) reduce 92) ((T_UNSET) reduce 92) ((T_USE) reduce 92) ((T_VAR) reduce 92) ((T_WHILE) reduce 92) ((T_YIELD) reduce 92)) 7192 (((T_ABSTRACT) reduce 280) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_ARRAY) reduce 280) ((T_AS) reduce 280) ((T_BREAK) reduce 280) ((T_CALLABLE) reduce 280) ((T_CASE) reduce 280) ((T_CATCH) reduce 280) ((T_CLASS) reduce 280) ((T_CLASS_C) reduce 280) ((T_CLONE) reduce 280) ((T_CONST) reduce 280) ((T_CONTINUE) reduce 280) ((T_DECLARE) reduce 280) ((T_DEFAULT) reduce 280) ((T_DIR) reduce 280) ((T_DO) reduce 280) ((T_ECHO) reduce 280) ((T_ELSE) reduce 280) ((T_ELSEIF) reduce 280) ((T_EMPTY) reduce 280) ((T_ENDDECLARE) reduce 280) ((T_ENDFOR) reduce 280) ((T_ENDFOREACH) reduce 280) ((T_ENDIF) reduce 280) ((T_ENDSWITCH) reduce 280) ((T_ENDWHILE) reduce 280) ((T_ENUM) reduce 280) ((T_EVAL) reduce 280) ((T_EXIT) reduce 280) ((T_EXTENDS) reduce 280) ((T_FILE) reduce 280) ((T_FINAL) reduce 280) ((T_FINALLY) reduce 280) ((T_FN) reduce 280) ((T_FOR) reduce 280) ((T_FOREACH) reduce 280) ((T_FUNCTION) reduce 280) ((T_FUNC_C) reduce 280) ((T_GLOBAL) reduce 280) ((T_GOTO) reduce 280) ((T_IF) reduce 280) ((T_IMPLEMENTS) reduce 280) ((T_INCLUDE) reduce 280) ((T_INCLUDE_ONCE) reduce 280) ((T_INSTANCEOF) reduce 280) ((T_INSTEADOF) reduce 280) ((T_INTERFACE) reduce 280) ((T_ISSET) reduce 280) ((T_LINE) reduce 280) ((T_LIST) reduce 280) ((T_LOGICAL_AND) reduce 280) ((T_LOGICAL_OR) reduce 280) ((T_LOGICAL_XOR) reduce 280) ((T_MATCH) reduce 280) ((T_METHOD_C) reduce 280) ((T_NAMESPACE) reduce 280) ((T_NEW) reduce 280) ((T_NS_C) reduce 280) ((T_PRINT) reduce 280) ((T_PRIVATE) reduce 280) ((T_PROPERTY_C) reduce 280) ((T_PROTECTED) reduce 280) ((T_PUBLIC) reduce 280) ((T_READONLY) reduce 280) ((T_REQUIRE) reduce 280) ((T_REQUIRE_ONCE) reduce 280) ((T_RETURN) reduce 280) ((T_STATIC) reduce 280) ((T_STRING) reduce 280) ((T_SWITCH) reduce 280) ((T_THROW) reduce 280) ((T_TRAIT) reduce 280) ((T_TRAIT_C) reduce 280) ((T_TRY) reduce 280) ((T_UNSET) reduce 280) ((T_USE) reduce 280) ((T_VAR) reduce 280) ((T_WHILE) reduce 280) ((T_YIELD) reduce 280)) 7193 (((T_ABSTRACT) reduce 272) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) ((T_ARRAY) reduce 272) ((T_AS) reduce 272) ((T_BREAK) reduce 272) ((T_CALLABLE) reduce 272) ((T_CASE) reduce 272) ((T_CATCH) reduce 272) ((T_CLASS) reduce 272) ((T_CLASS_C) reduce 272) ((T_CLONE) reduce 272) ((T_CONST) reduce 272) ((T_CONTINUE) reduce 272) ((T_DECLARE) reduce 272) ((T_DEFAULT) reduce 272) ((T_DIR) reduce 272) ((T_DO) reduce 272) ((T_ECHO) reduce 272) ((T_ELSE) reduce 272) ((T_ELSEIF) reduce 272) ((T_EMPTY) reduce 272) ((T_ENDDECLARE) reduce 272) ((T_ENDFOR) reduce 272) ((T_ENDFOREACH) reduce 272) ((T_ENDIF) reduce 272) ((T_ENDSWITCH) reduce 272) ((T_ENDWHILE) reduce 272) ((T_ENUM) reduce 272) ((T_EVAL) reduce 272) ((T_EXIT) reduce 272) ((T_EXTENDS) reduce 272) ((T_FILE) reduce 272) ((T_FINAL) reduce 272) ((T_FINALLY) reduce 272) ((T_FN) reduce 272) ((T_FOR) reduce 272) ((T_FOREACH) reduce 272) ((T_FUNCTION) reduce 272) ((T_FUNC_C) reduce 272) ((T_GLOBAL) reduce 272) ((T_GOTO) reduce 272) ((T_IF) reduce 272) ((T_IMPLEMENTS) reduce 272) ((T_INCLUDE) reduce 272) ((T_INCLUDE_ONCE) reduce 272) ((T_INSTANCEOF) reduce 272) ((T_INSTEADOF) reduce 272) ((T_INTERFACE) reduce 272) ((T_ISSET) reduce 272) ((T_LINE) reduce 272) ((T_LIST) reduce 272) ((T_LOGICAL_AND) reduce 272) ((T_LOGICAL_OR) reduce 272) ((T_LOGICAL_XOR) reduce 272) ((T_MATCH) reduce 272) ((T_METHOD_C) reduce 272) ((T_NAMESPACE) reduce 272) ((T_NEW) reduce 272) ((T_NS_C) reduce 272) ((T_PRINT) reduce 272) ((T_PRIVATE) reduce 272) ((T_PROPERTY_C) reduce 272) ((T_PROTECTED) reduce 272) ((T_PUBLIC) reduce 272) ((T_READONLY) reduce 272) ((T_REQUIRE) reduce 272) ((T_REQUIRE_ONCE) reduce 272) ((T_RETURN) reduce 272) ((T_STATIC) reduce 272) ((T_STRING) reduce 272) ((T_SWITCH) reduce 272) ((T_THROW) reduce 272) ((T_TRAIT) reduce 272) ((T_TRAIT_C) reduce 272) ((T_TRY) reduce 272) ((T_UNSET) reduce 272) ((T_USE) reduce 272) ((T_VAR) reduce 272) ((T_WHILE) reduce 272) ((T_YIELD) reduce 272)) 7194 (((T_ABSTRACT) reduce 266) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266) ((T_ARRAY) reduce 266) ((T_AS) reduce 266) ((T_BREAK) reduce 266) ((T_CALLABLE) reduce 266) ((T_CASE) reduce 266) ((T_CATCH) reduce 266) ((T_CLASS) reduce 266) ((T_CLASS_C) reduce 266) ((T_CLONE) reduce 266) ((T_CONST) reduce 266) ((T_CONTINUE) reduce 266) ((T_DECLARE) reduce 266) ((T_DEFAULT) reduce 266) ((T_DIR) reduce 266) ((T_DO) reduce 266) ((T_ECHO) reduce 266) ((T_ELSE) reduce 266) ((T_ELSEIF) reduce 266) ((T_EMPTY) reduce 266) ((T_ENDDECLARE) reduce 266) ((T_ENDFOR) reduce 266) ((T_ENDFOREACH) reduce 266) ((T_ENDIF) reduce 266) ((T_ENDSWITCH) reduce 266) ((T_ENDWHILE) reduce 266) ((T_ENUM) reduce 266) ((T_EVAL) reduce 266) ((T_EXIT) reduce 266) ((T_EXTENDS) reduce 266) ((T_FILE) reduce 266) ((T_FINAL) reduce 266) ((T_FINALLY) reduce 266) ((T_FN) reduce 266) ((T_FOR) reduce 266) ((T_FOREACH) reduce 266) ((T_FUNCTION) reduce 266) ((T_FUNC_C) reduce 266) ((T_GLOBAL) reduce 266) ((T_GOTO) reduce 266) ((T_IF) reduce 266) ((T_IMPLEMENTS) reduce 266) ((T_INCLUDE) reduce 266) ((T_INCLUDE_ONCE) reduce 266) ((T_INSTANCEOF) reduce 266) ((T_INSTEADOF) reduce 266) ((T_INTERFACE) reduce 266) ((T_ISSET) reduce 266) ((T_LINE) reduce 266) ((T_LIST) reduce 266) ((T_LOGICAL_AND) reduce 266) ((T_LOGICAL_OR) reduce 266) ((T_LOGICAL_XOR) reduce 266) ((T_MATCH) reduce 266) ((T_METHOD_C) reduce 266) ((T_NAMESPACE) reduce 266) ((T_NEW) reduce 266) ((T_NS_C) reduce 266) ((T_PRINT) reduce 266) ((T_PRIVATE) reduce 266) ((T_PROPERTY_C) reduce 266) ((T_PROTECTED) reduce 266) ((T_PUBLIC) reduce 266) ((T_READONLY) reduce 266) ((T_REQUIRE) reduce 266) ((T_REQUIRE_ONCE) reduce 266) ((T_RETURN) reduce 266) ((T_STATIC) reduce 266) ((T_STRING) reduce 266) ((T_SWITCH) reduce 266) ((T_THROW) reduce 266) ((T_TRAIT) reduce 266) ((T_TRAIT_C) reduce 266) ((T_TRY) reduce 266) ((T_UNSET) reduce 266) ((T_USE) reduce 266) ((T_VAR) reduce 266) ((T_WHILE) reduce 266) ((T_YIELD) reduce 266)) 7197 (((",") reduce 374) ((";") reduce 374)) 7199 (((T_ABSTRACT) reduce 273) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) ((T_ARRAY) reduce 273) ((T_AS) reduce 273) ((T_BREAK) reduce 273) ((T_CALLABLE) reduce 273) ((T_CASE) reduce 273) ((T_CATCH) reduce 273) ((T_CLASS) reduce 273) ((T_CLASS_C) reduce 273) ((T_CLONE) reduce 273) ((T_CONST) reduce 273) ((T_CONTINUE) reduce 273) ((T_DECLARE) reduce 273) ((T_DEFAULT) reduce 273) ((T_DIR) reduce 273) ((T_DO) reduce 273) ((T_ECHO) reduce 273) ((T_ELSE) reduce 273) ((T_ELSEIF) reduce 273) ((T_EMPTY) reduce 273) ((T_ENDDECLARE) reduce 273) ((T_ENDFOR) reduce 273) ((T_ENDFOREACH) reduce 273) ((T_ENDIF) reduce 273) ((T_ENDSWITCH) reduce 273) ((T_ENDWHILE) reduce 273) ((T_ENUM) reduce 273) ((T_EVAL) reduce 273) ((T_EXIT) reduce 273) ((T_EXTENDS) reduce 273) ((T_FILE) reduce 273) ((T_FINAL) reduce 273) ((T_FINALLY) reduce 273) ((T_FN) reduce 273) ((T_FOR) reduce 273) ((T_FOREACH) reduce 273) ((T_FUNCTION) reduce 273) ((T_FUNC_C) reduce 273) ((T_GLOBAL) reduce 273) ((T_GOTO) reduce 273) ((T_IF) reduce 273) ((T_IMPLEMENTS) reduce 273) ((T_INCLUDE) reduce 273) ((T_INCLUDE_ONCE) reduce 273) ((T_INSTANCEOF) reduce 273) ((T_INSTEADOF) reduce 273) ((T_INTERFACE) reduce 273) ((T_ISSET) reduce 273) ((T_LINE) reduce 273) ((T_LIST) reduce 273) ((T_LOGICAL_AND) reduce 273) ((T_LOGICAL_OR) reduce 273) ((T_LOGICAL_XOR) reduce 273) ((T_MATCH) reduce 273) ((T_METHOD_C) reduce 273) ((T_NAMESPACE) reduce 273) ((T_NEW) reduce 273) ((T_NS_C) reduce 273) ((T_PRINT) reduce 273) ((T_PRIVATE) reduce 273) ((T_PROPERTY_C) reduce 273) ((T_PROTECTED) reduce 273) ((T_PUBLIC) reduce 273) ((T_READONLY) reduce 273) ((T_REQUIRE) reduce 273) ((T_REQUIRE_ONCE) reduce 273) ((T_RETURN) reduce 273) ((T_STATIC) reduce 273) ((T_STRING) reduce 273) ((T_SWITCH) reduce 273) ((T_THROW) reduce 273) ((T_TRAIT) reduce 273) ((T_TRAIT_C) reduce 273) ((T_TRY) reduce 273) ((T_UNSET) reduce 273) ((T_USE) reduce 273) ((T_VAR) reduce 273) ((T_WHILE) reduce 273) ((T_YIELD) reduce 273)) 7200 (((T_ABSTRACT) reduce 308) ((T_ATTRIBUTE) reduce 308) ((T_CASE) reduce 308) ((T_CONST) reduce 308) ((T_FINAL) reduce 308) ((T_FUNCTION) reduce 308) ((T_PRIVATE) reduce 308) ((T_PRIVATE_SET) reduce 308) ((T_PROTECTED) reduce 308) ((T_PROTECTED_SET) reduce 308) ((T_PUBLIC) reduce 308) ((T_PUBLIC_SET) reduce 308) ((T_READONLY) reduce 308) ((T_STATIC) reduce 308) ((T_USE) reduce 308) ((T_VAR) reduce 308) (("}") reduce 308)) 7201 (((T_ABSTRACT) reduce 278) ((T_ARRAY) reduce 278) ((T_AS) reduce 278) ((T_BREAK) reduce 278) ((T_CALLABLE) reduce 278) ((T_CASE) reduce 278) ((T_CATCH) reduce 278) ((T_CLASS) reduce 278) ((T_CLASS_C) reduce 278) ((T_CLONE) reduce 278) ((T_CONST) reduce 278) ((T_CONTINUE) reduce 278) ((T_DECLARE) reduce 278) ((T_DEFAULT) reduce 278) ((T_DIR) reduce 278) ((T_DO) reduce 278) ((T_ECHO) reduce 278) ((T_ELSE) reduce 278) ((T_ELSEIF) reduce 278) ((T_EMPTY) reduce 278) ((T_ENDDECLARE) reduce 278) ((T_ENDFOR) reduce 278) ((T_ENDFOREACH) reduce 278) ((T_ENDIF) reduce 278) ((T_ENDSWITCH) reduce 278) ((T_ENDWHILE) reduce 278) ((T_ENUM) reduce 278) ((T_EVAL) reduce 278) ((T_EXIT) reduce 278) ((T_EXTENDS) reduce 278) ((T_FILE) reduce 278) ((T_FINAL) reduce 278) ((T_FINALLY) reduce 278) ((T_FN) reduce 278) ((T_FOR) reduce 278) ((T_FOREACH) reduce 278) ((T_FUNCTION) reduce 278) ((T_FUNC_C) reduce 278) ((T_GLOBAL) reduce 278) ((T_GOTO) reduce 278) ((T_IF) reduce 278) ((T_IMPLEMENTS) reduce 278) ((T_INCLUDE) reduce 278) ((T_INCLUDE_ONCE) reduce 278) ((T_INSTANCEOF) reduce 278) ((T_INSTEADOF) reduce 278) ((T_INTERFACE) reduce 278) ((T_ISSET) reduce 278) ((T_LINE) reduce 278) ((T_LIST) reduce 278) ((T_LOGICAL_AND) reduce 278) ((T_LOGICAL_OR) reduce 278) ((T_LOGICAL_XOR) reduce 278) ((T_MATCH) reduce 278) ((T_METHOD_C) reduce 278) ((T_NAMESPACE) reduce 278) ((T_NEW) reduce 278) ((T_NS_C) reduce 278) ((T_PRINT) reduce 278) ((T_PRIVATE) reduce 278) ((T_PROPERTY_C) reduce 278) ((T_PROTECTED) reduce 278) ((T_PUBLIC) reduce 278) ((T_READONLY) reduce 278) ((T_REQUIRE) reduce 278) ((T_REQUIRE_ONCE) reduce 278) ((T_RETURN) reduce 278) ((T_STATIC) reduce 278) ((T_STRING) reduce 278) ((T_SWITCH) reduce 278) ((T_THROW) reduce 278) ((T_TRAIT) reduce 278) ((T_TRAIT_C) reduce 278) ((T_TRY) reduce 278) ((T_UNSET) reduce 278) ((T_USE) reduce 278) ((T_VAR) reduce 278) ((T_WHILE) reduce 278) ((T_YIELD) reduce 278)) 7202 (((T_ABSTRACT) reduce 279) ((T_ARRAY) reduce 279) ((T_AS) reduce 279) ((T_BREAK) reduce 279) ((T_CALLABLE) reduce 279) ((T_CASE) reduce 279) ((T_CATCH) reduce 279) ((T_CLASS) reduce 279) ((T_CLASS_C) reduce 279) ((T_CLONE) reduce 279) ((T_CONST) reduce 279) ((T_CONTINUE) reduce 279) ((T_DECLARE) reduce 279) ((T_DEFAULT) reduce 279) ((T_DIR) reduce 279) ((T_DO) reduce 279) ((T_ECHO) reduce 279) ((T_ELSE) reduce 279) ((T_ELSEIF) reduce 279) ((T_EMPTY) reduce 279) ((T_ENDDECLARE) reduce 279) ((T_ENDFOR) reduce 279) ((T_ENDFOREACH) reduce 279) ((T_ENDIF) reduce 279) ((T_ENDSWITCH) reduce 279) ((T_ENDWHILE) reduce 279) ((T_ENUM) reduce 279) ((T_EVAL) reduce 279) ((T_EXIT) reduce 279) ((T_EXTENDS) reduce 279) ((T_FILE) reduce 279) ((T_FINAL) reduce 279) ((T_FINALLY) reduce 279) ((T_FN) reduce 279) ((T_FOR) reduce 279) ((T_FOREACH) reduce 279) ((T_FUNCTION) reduce 279) ((T_FUNC_C) reduce 279) ((T_GLOBAL) reduce 279) ((T_GOTO) reduce 279) ((T_IF) reduce 279) ((T_IMPLEMENTS) reduce 279) ((T_INCLUDE) reduce 279) ((T_INCLUDE_ONCE) reduce 279) ((T_INSTANCEOF) reduce 279) ((T_INSTEADOF) reduce 279) ((T_INTERFACE) reduce 279) ((T_ISSET) reduce 279) ((T_LINE) reduce 279) ((T_LIST) reduce 279) ((T_LOGICAL_AND) reduce 279) ((T_LOGICAL_OR) reduce 279) ((T_LOGICAL_XOR) reduce 279) ((T_MATCH) reduce 279) ((T_METHOD_C) reduce 279) ((T_NAMESPACE) reduce 279) ((T_NEW) reduce 279) ((T_NS_C) reduce 279) ((T_PRINT) reduce 279) ((T_PRIVATE) reduce 279) ((T_PROPERTY_C) reduce 279) ((T_PROTECTED) reduce 279) ((T_PUBLIC) reduce 279) ((T_READONLY) reduce 279) ((T_REQUIRE) reduce 279) ((T_REQUIRE_ONCE) reduce 279) ((T_RETURN) reduce 279) ((T_STATIC) reduce 279) ((T_STRING) reduce 279) ((T_SWITCH) reduce 279) ((T_THROW) reduce 279) ((T_TRAIT) reduce 279) ((T_TRAIT_C) reduce 279) ((T_TRY) reduce 279) ((T_UNSET) reduce 279) ((T_USE) reduce 279) ((T_VAR) reduce 279) ((T_WHILE) reduce 279) ((T_YIELD) reduce 279)) 7203 (((T_ABSTRACT) reduce 94) ((T_ARRAY) reduce 94) ((T_AS) reduce 94) ((T_BREAK) reduce 94) ((T_CALLABLE) reduce 94) ((T_CASE) reduce 94) ((T_CATCH) reduce 94) ((T_CLASS) reduce 94) ((T_CLASS_C) reduce 94) ((T_CLONE) reduce 94) ((T_CONST) reduce 94) ((T_CONTINUE) reduce 94) ((T_DECLARE) reduce 94) ((T_DEFAULT) reduce 94) ((T_DIR) reduce 94) ((T_DO) reduce 94) ((T_ECHO) reduce 94) ((T_ELSE) reduce 94) ((T_ELSEIF) reduce 94) ((T_EMPTY) reduce 94) ((T_ENDDECLARE) reduce 94) ((T_ENDFOR) reduce 94) ((T_ENDFOREACH) reduce 94) ((T_ENDIF) reduce 94) ((T_ENDSWITCH) reduce 94) ((T_ENDWHILE) reduce 94) ((T_ENUM) reduce 94) ((T_EVAL) reduce 94) ((T_EXIT) reduce 94) ((T_EXTENDS) reduce 94) ((T_FILE) reduce 94) ((T_FINAL) reduce 94) ((T_FINALLY) reduce 94) ((T_FN) reduce 94) ((T_FOR) reduce 94) ((T_FOREACH) reduce 94) ((T_FUNCTION) reduce 94) ((T_FUNC_C) reduce 94) ((T_GLOBAL) reduce 94) ((T_GOTO) reduce 94) ((T_IF) reduce 94) ((T_IMPLEMENTS) reduce 94) ((T_INCLUDE) reduce 94) ((T_INCLUDE_ONCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_INSTEADOF) reduce 94) ((T_INTERFACE) reduce 94) ((T_ISSET) reduce 94) ((T_LINE) reduce 94) ((T_LIST) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_MATCH) reduce 94) ((T_METHOD_C) reduce 94) ((T_NAMESPACE) reduce 94) ((T_NEW) reduce 94) ((T_NS_C) reduce 94) ((T_PRINT) reduce 94) ((T_PRIVATE) reduce 94) ((T_PROPERTY_C) reduce 94) ((T_PROTECTED) reduce 94) ((T_PUBLIC) reduce 94) ((T_READONLY) reduce 94) ((T_REQUIRE) reduce 94) ((T_REQUIRE_ONCE) reduce 94) ((T_RETURN) reduce 94) ((T_STATIC) reduce 94) ((T_STRING) reduce 94) ((T_SWITCH) reduce 94) ((T_THROW) reduce 94) ((T_TRAIT) reduce 94) ((T_TRAIT_C) reduce 94) ((T_TRY) reduce 94) ((T_UNSET) reduce 94) ((T_USE) reduce 94) ((T_VAR) reduce 94) ((T_WHILE) reduce 94) ((T_YIELD) reduce 94)) 7204 (((T_ABSTRACT) reduce 93) ((T_ARRAY) reduce 93) ((T_AS) reduce 93) ((T_BREAK) reduce 93) ((T_CALLABLE) reduce 93) ((T_CASE) reduce 93) ((T_CATCH) reduce 93) ((T_CLASS) reduce 93) ((T_CLASS_C) reduce 93) ((T_CLONE) reduce 93) ((T_CONST) reduce 93) ((T_CONTINUE) reduce 93) ((T_DECLARE) reduce 93) ((T_DEFAULT) reduce 93) ((T_DIR) reduce 93) ((T_DO) reduce 93) ((T_ECHO) reduce 93) ((T_ELSE) reduce 93) ((T_ELSEIF) reduce 93) ((T_EMPTY) reduce 93) ((T_ENDDECLARE) reduce 93) ((T_ENDFOR) reduce 93) ((T_ENDFOREACH) reduce 93) ((T_ENDIF) reduce 93) ((T_ENDSWITCH) reduce 93) ((T_ENDWHILE) reduce 93) ((T_ENUM) reduce 93) ((T_EVAL) reduce 93) ((T_EXIT) reduce 93) ((T_EXTENDS) reduce 93) ((T_FILE) reduce 93) ((T_FINAL) reduce 93) ((T_FINALLY) reduce 93) ((T_FN) reduce 93) ((T_FOR) reduce 93) ((T_FOREACH) reduce 93) ((T_FUNCTION) reduce 93) ((T_FUNC_C) reduce 93) ((T_GLOBAL) reduce 93) ((T_GOTO) reduce 93) ((T_IF) reduce 93) ((T_IMPLEMENTS) reduce 93) ((T_INCLUDE) reduce 93) ((T_INCLUDE_ONCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_INSTEADOF) reduce 93) ((T_INTERFACE) reduce 93) ((T_ISSET) reduce 93) ((T_LINE) reduce 93) ((T_LIST) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_MATCH) reduce 93) ((T_METHOD_C) reduce 93) ((T_NAMESPACE) reduce 93) ((T_NEW) reduce 93) ((T_NS_C) reduce 93) ((T_PRINT) reduce 93) ((T_PRIVATE) reduce 93) ((T_PROPERTY_C) reduce 93) ((T_PROTECTED) reduce 93) ((T_PUBLIC) reduce 93) ((T_READONLY) reduce 93) ((T_REQUIRE) reduce 93) ((T_REQUIRE_ONCE) reduce 93) ((T_RETURN) reduce 93) ((T_STATIC) reduce 93) ((T_STRING) reduce 93) ((T_SWITCH) reduce 93) ((T_THROW) reduce 93) ((T_TRAIT) reduce 93) ((T_TRAIT_C) reduce 93) ((T_TRY) reduce 93) ((T_UNSET) reduce 93) ((T_USE) reduce 93) ((T_VAR) reduce 93) ((T_WHILE) reduce 93) ((T_YIELD) reduce 93)) 7205 (((T_ABSTRACT) reduce 95) ((T_ARRAY) reduce 95) ((T_AS) reduce 95) ((T_BREAK) reduce 95) ((T_CALLABLE) reduce 95) ((T_CASE) reduce 95) ((T_CATCH) reduce 95) ((T_CLASS) reduce 95) ((T_CLASS_C) reduce 95) ((T_CLONE) reduce 95) ((T_CONST) reduce 95) ((T_CONTINUE) reduce 95) ((T_DECLARE) reduce 95) ((T_DEFAULT) reduce 95) ((T_DIR) reduce 95) ((T_DO) reduce 95) ((T_ECHO) reduce 95) ((T_ELSE) reduce 95) ((T_ELSEIF) reduce 95) ((T_EMPTY) reduce 95) ((T_ENDDECLARE) reduce 95) ((T_ENDFOR) reduce 95) ((T_ENDFOREACH) reduce 95) ((T_ENDIF) reduce 95) ((T_ENDSWITCH) reduce 95) ((T_ENDWHILE) reduce 95) ((T_ENUM) reduce 95) ((T_EVAL) reduce 95) ((T_EXIT) reduce 95) ((T_EXTENDS) reduce 95) ((T_FILE) reduce 95) ((T_FINAL) reduce 95) ((T_FINALLY) reduce 95) ((T_FN) reduce 95) ((T_FOR) reduce 95) ((T_FOREACH) reduce 95) ((T_FUNCTION) reduce 95) ((T_FUNC_C) reduce 95) ((T_GLOBAL) reduce 95) ((T_GOTO) reduce 95) ((T_IF) reduce 95) ((T_IMPLEMENTS) reduce 95) ((T_INCLUDE) reduce 95) ((T_INCLUDE_ONCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_INSTEADOF) reduce 95) ((T_INTERFACE) reduce 95) ((T_ISSET) reduce 95) ((T_LINE) reduce 95) ((T_LIST) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_MATCH) reduce 95) ((T_METHOD_C) reduce 95) ((T_NAMESPACE) reduce 95) ((T_NEW) reduce 95) ((T_NS_C) reduce 95) ((T_PRINT) reduce 95) ((T_PRIVATE) reduce 95) ((T_PROPERTY_C) reduce 95) ((T_PROTECTED) reduce 95) ((T_PUBLIC) reduce 95) ((T_READONLY) reduce 95) ((T_REQUIRE) reduce 95) ((T_REQUIRE_ONCE) reduce 95) ((T_RETURN) reduce 95) ((T_STATIC) reduce 95) ((T_STRING) reduce 95) ((T_SWITCH) reduce 95) ((T_THROW) reduce 95) ((T_TRAIT) reduce 95) ((T_TRAIT_C) reduce 95) ((T_TRY) reduce 95) ((T_UNSET) reduce 95) ((T_USE) reduce 95) ((T_VAR) reduce 95) ((T_WHILE) reduce 95) ((T_YIELD) reduce 95)) 7206 (((T_ABSTRACT) reduce 267) ((T_ARRAY) reduce 267) ((T_AS) reduce 267) ((T_BREAK) reduce 267) ((T_CALLABLE) reduce 267) ((T_CASE) reduce 267) ((T_CATCH) reduce 267) ((T_CLASS) reduce 267) ((T_CLASS_C) reduce 267) ((T_CLONE) reduce 267) ((T_CONST) reduce 267) ((T_CONTINUE) reduce 267) ((T_DECLARE) reduce 267) ((T_DEFAULT) reduce 267) ((T_DIR) reduce 267) ((T_DO) reduce 267) ((T_ECHO) reduce 267) ((T_ELSE) reduce 267) ((T_ELSEIF) reduce 267) ((T_EMPTY) reduce 267) ((T_ENDDECLARE) reduce 267) ((T_ENDFOR) reduce 267) ((T_ENDFOREACH) reduce 267) ((T_ENDIF) reduce 267) ((T_ENDSWITCH) reduce 267) ((T_ENDWHILE) reduce 267) ((T_ENUM) reduce 267) ((T_EVAL) reduce 267) ((T_EXIT) reduce 267) ((T_EXTENDS) reduce 267) ((T_FILE) reduce 267) ((T_FINAL) reduce 267) ((T_FINALLY) reduce 267) ((T_FN) reduce 267) ((T_FOR) reduce 267) ((T_FOREACH) reduce 267) ((T_FUNCTION) reduce 267) ((T_FUNC_C) reduce 267) ((T_GLOBAL) reduce 267) ((T_GOTO) reduce 267) ((T_IF) reduce 267) ((T_IMPLEMENTS) reduce 267) ((T_INCLUDE) reduce 267) ((T_INCLUDE_ONCE) reduce 267) ((T_INSTANCEOF) reduce 267) ((T_INSTEADOF) reduce 267) ((T_INTERFACE) reduce 267) ((T_ISSET) reduce 267) ((T_LINE) reduce 267) ((T_LIST) reduce 267) ((T_LOGICAL_AND) reduce 267) ((T_LOGICAL_OR) reduce 267) ((T_LOGICAL_XOR) reduce 267) ((T_MATCH) reduce 267) ((T_METHOD_C) reduce 267) ((T_NAMESPACE) reduce 267) ((T_NEW) reduce 267) ((T_NS_C) reduce 267) ((T_PRINT) reduce 267) ((T_PRIVATE) reduce 267) ((T_PROPERTY_C) reduce 267) ((T_PROTECTED) reduce 267) ((T_PUBLIC) reduce 267) ((T_READONLY) reduce 267) ((T_REQUIRE) reduce 267) ((T_REQUIRE_ONCE) reduce 267) ((T_RETURN) reduce 267) ((T_STATIC) reduce 267) ((T_STRING) reduce 267) ((T_SWITCH) reduce 267) ((T_THROW) reduce 267) ((T_TRAIT) reduce 267) ((T_TRAIT_C) reduce 267) ((T_TRY) reduce 267) ((T_UNSET) reduce 267) ((T_USE) reduce 267) ((T_VAR) reduce 267) ((T_WHILE) reduce 267) ((T_YIELD) reduce 267)) 7207 (((T_ABSTRACT) reduce 92) ((T_ARRAY) reduce 92) ((T_AS) reduce 92) ((T_BREAK) reduce 92) ((T_CALLABLE) reduce 92) ((T_CASE) reduce 92) ((T_CATCH) reduce 92) ((T_CLASS) reduce 92) ((T_CLASS_C) reduce 92) ((T_CLONE) reduce 92) ((T_CONST) reduce 92) ((T_CONTINUE) reduce 92) ((T_DECLARE) reduce 92) ((T_DEFAULT) reduce 92) ((T_DIR) reduce 92) ((T_DO) reduce 92) ((T_ECHO) reduce 92) ((T_ELSE) reduce 92) ((T_ELSEIF) reduce 92) ((T_EMPTY) reduce 92) ((T_ENDDECLARE) reduce 92) ((T_ENDFOR) reduce 92) ((T_ENDFOREACH) reduce 92) ((T_ENDIF) reduce 92) ((T_ENDSWITCH) reduce 92) ((T_ENDWHILE) reduce 92) ((T_ENUM) reduce 92) ((T_EVAL) reduce 92) ((T_EXIT) reduce 92) ((T_EXTENDS) reduce 92) ((T_FILE) reduce 92) ((T_FINAL) reduce 92) ((T_FINALLY) reduce 92) ((T_FN) reduce 92) ((T_FOR) reduce 92) ((T_FOREACH) reduce 92) ((T_FUNCTION) reduce 92) ((T_FUNC_C) reduce 92) ((T_GLOBAL) reduce 92) ((T_GOTO) reduce 92) ((T_IF) reduce 92) ((T_IMPLEMENTS) reduce 92) ((T_INCLUDE) reduce 92) ((T_INCLUDE_ONCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_INSTEADOF) reduce 92) ((T_INTERFACE) reduce 92) ((T_ISSET) reduce 92) ((T_LINE) reduce 92) ((T_LIST) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_MATCH) reduce 92) ((T_METHOD_C) reduce 92) ((T_NAMESPACE) reduce 92) ((T_NEW) reduce 92) ((T_NS_C) reduce 92) ((T_PRINT) reduce 92) ((T_PRIVATE) reduce 92) ((T_PROPERTY_C) reduce 92) ((T_PROTECTED) reduce 92) ((T_PUBLIC) reduce 92) ((T_READONLY) reduce 92) ((T_REQUIRE) reduce 92) ((T_REQUIRE_ONCE) reduce 92) ((T_RETURN) reduce 92) ((T_STATIC) reduce 92) ((T_STRING) reduce 92) ((T_SWITCH) reduce 92) ((T_THROW) reduce 92) ((T_TRAIT) reduce 92) ((T_TRAIT_C) reduce 92) ((T_TRY) reduce 92) ((T_UNSET) reduce 92) ((T_USE) reduce 92) ((T_VAR) reduce 92) ((T_WHILE) reduce 92) ((T_YIELD) reduce 92)) 7208 (((T_ABSTRACT) reduce 280) ((T_ARRAY) reduce 280) ((T_AS) reduce 280) ((T_BREAK) reduce 280) ((T_CALLABLE) reduce 280) ((T_CASE) reduce 280) ((T_CATCH) reduce 280) ((T_CLASS) reduce 280) ((T_CLASS_C) reduce 280) ((T_CLONE) reduce 280) ((T_CONST) reduce 280) ((T_CONTINUE) reduce 280) ((T_DECLARE) reduce 280) ((T_DEFAULT) reduce 280) ((T_DIR) reduce 280) ((T_DO) reduce 280) ((T_ECHO) reduce 280) ((T_ELSE) reduce 280) ((T_ELSEIF) reduce 280) ((T_EMPTY) reduce 280) ((T_ENDDECLARE) reduce 280) ((T_ENDFOR) reduce 280) ((T_ENDFOREACH) reduce 280) ((T_ENDIF) reduce 280) ((T_ENDSWITCH) reduce 280) ((T_ENDWHILE) reduce 280) ((T_ENUM) reduce 280) ((T_EVAL) reduce 280) ((T_EXIT) reduce 280) ((T_EXTENDS) reduce 280) ((T_FILE) reduce 280) ((T_FINAL) reduce 280) ((T_FINALLY) reduce 280) ((T_FN) reduce 280) ((T_FOR) reduce 280) ((T_FOREACH) reduce 280) ((T_FUNCTION) reduce 280) ((T_FUNC_C) reduce 280) ((T_GLOBAL) reduce 280) ((T_GOTO) reduce 280) ((T_IF) reduce 280) ((T_IMPLEMENTS) reduce 280) ((T_INCLUDE) reduce 280) ((T_INCLUDE_ONCE) reduce 280) ((T_INSTANCEOF) reduce 280) ((T_INSTEADOF) reduce 280) ((T_INTERFACE) reduce 280) ((T_ISSET) reduce 280) ((T_LINE) reduce 280) ((T_LIST) reduce 280) ((T_LOGICAL_AND) reduce 280) ((T_LOGICAL_OR) reduce 280) ((T_LOGICAL_XOR) reduce 280) ((T_MATCH) reduce 280) ((T_METHOD_C) reduce 280) ((T_NAMESPACE) reduce 280) ((T_NEW) reduce 280) ((T_NS_C) reduce 280) ((T_PRINT) reduce 280) ((T_PRIVATE) reduce 280) ((T_PROPERTY_C) reduce 280) ((T_PROTECTED) reduce 280) ((T_PUBLIC) reduce 280) ((T_READONLY) reduce 280) ((T_REQUIRE) reduce 280) ((T_REQUIRE_ONCE) reduce 280) ((T_RETURN) reduce 280) ((T_STATIC) reduce 280) ((T_STRING) reduce 280) ((T_SWITCH) reduce 280) ((T_THROW) reduce 280) ((T_TRAIT) reduce 280) ((T_TRAIT_C) reduce 280) ((T_TRY) reduce 280) ((T_UNSET) reduce 280) ((T_USE) reduce 280) ((T_VAR) reduce 280) ((T_WHILE) reduce 280) ((T_YIELD) reduce 280)) 7209 (((T_ABSTRACT) reduce 263) ((T_ARRAY) reduce 263) ((T_AS) reduce 263) ((T_BREAK) reduce 263) ((T_CALLABLE) reduce 263) ((T_CASE) reduce 263) ((T_CATCH) reduce 263) ((T_CLASS) reduce 263) ((T_CLASS_C) reduce 263) ((T_CLONE) reduce 263) ((T_CONST) reduce 263) ((T_CONTINUE) reduce 263) ((T_DECLARE) reduce 263) ((T_DEFAULT) reduce 263) ((T_DIR) reduce 263) ((T_DO) reduce 263) ((T_ECHO) reduce 263) ((T_ELSE) reduce 263) ((T_ELSEIF) reduce 263) ((T_EMPTY) reduce 263) ((T_ENDDECLARE) reduce 263) ((T_ENDFOR) reduce 263) ((T_ENDFOREACH) reduce 263) ((T_ENDIF) reduce 263) ((T_ENDSWITCH) reduce 263) ((T_ENDWHILE) reduce 263) ((T_ENUM) reduce 263) ((T_EVAL) reduce 263) ((T_EXIT) reduce 263) ((T_EXTENDS) reduce 263) ((T_FILE) reduce 263) ((T_FINAL) reduce 263) ((T_FINALLY) reduce 263) ((T_FN) reduce 263) ((T_FOR) reduce 263) ((T_FOREACH) reduce 263) ((T_FUNCTION) reduce 263) ((T_FUNC_C) reduce 263) ((T_GLOBAL) reduce 263) ((T_GOTO) reduce 263) ((T_IF) reduce 263) ((T_IMPLEMENTS) reduce 263) ((T_INCLUDE) reduce 263) ((T_INCLUDE_ONCE) reduce 263) ((T_INSTANCEOF) reduce 263) ((T_INSTEADOF) reduce 263) ((T_INTERFACE) reduce 263) ((T_ISSET) reduce 263) ((T_LINE) reduce 263) ((T_LIST) reduce 263) ((T_LOGICAL_AND) reduce 263) ((T_LOGICAL_OR) reduce 263) ((T_LOGICAL_XOR) reduce 263) ((T_MATCH) reduce 263) ((T_METHOD_C) reduce 263) ((T_NAMESPACE) reduce 263) ((T_NEW) reduce 263) ((T_NS_C) reduce 263) ((T_PRINT) reduce 263) ((T_PRIVATE) reduce 263) ((T_PROPERTY_C) reduce 263) ((T_PROTECTED) reduce 263) ((T_PUBLIC) reduce 263) ((T_READONLY) reduce 263) ((T_REQUIRE) reduce 263) ((T_REQUIRE_ONCE) reduce 263) ((T_RETURN) reduce 263) ((T_STATIC) reduce 263) ((T_STRING) reduce 263) ((T_SWITCH) reduce 263) ((T_THROW) reduce 263) ((T_TRAIT) reduce 263) ((T_TRAIT_C) reduce 263) ((T_TRY) reduce 263) ((T_UNSET) reduce 263) ((T_USE) reduce 263) ((T_VAR) reduce 263) ((T_WHILE) reduce 263) ((T_YIELD) reduce 263)) 7210 (((T_ABSTRACT) reduce 266) ((T_ARRAY) reduce 266) ((T_AS) reduce 266) ((T_BREAK) reduce 266) ((T_CALLABLE) reduce 266) ((T_CASE) reduce 266) ((T_CATCH) reduce 266) ((T_CLASS) reduce 266) ((T_CLASS_C) reduce 266) ((T_CLONE) reduce 266) ((T_CONST) reduce 266) ((T_CONTINUE) reduce 266) ((T_DECLARE) reduce 266) ((T_DEFAULT) reduce 266) ((T_DIR) reduce 266) ((T_DO) reduce 266) ((T_ECHO) reduce 266) ((T_ELSE) reduce 266) ((T_ELSEIF) reduce 266) ((T_EMPTY) reduce 266) ((T_ENDDECLARE) reduce 266) ((T_ENDFOR) reduce 266) ((T_ENDFOREACH) reduce 266) ((T_ENDIF) reduce 266) ((T_ENDSWITCH) reduce 266) ((T_ENDWHILE) reduce 266) ((T_ENUM) reduce 266) ((T_EVAL) reduce 266) ((T_EXIT) reduce 266) ((T_EXTENDS) reduce 266) ((T_FILE) reduce 266) ((T_FINAL) reduce 266) ((T_FINALLY) reduce 266) ((T_FN) reduce 266) ((T_FOR) reduce 266) ((T_FOREACH) reduce 266) ((T_FUNCTION) reduce 266) ((T_FUNC_C) reduce 266) ((T_GLOBAL) reduce 266) ((T_GOTO) reduce 266) ((T_IF) reduce 266) ((T_IMPLEMENTS) reduce 266) ((T_INCLUDE) reduce 266) ((T_INCLUDE_ONCE) reduce 266) ((T_INSTANCEOF) reduce 266) ((T_INSTEADOF) reduce 266) ((T_INTERFACE) reduce 266) ((T_ISSET) reduce 266) ((T_LINE) reduce 266) ((T_LIST) reduce 266) ((T_LOGICAL_AND) reduce 266) ((T_LOGICAL_OR) reduce 266) ((T_LOGICAL_XOR) reduce 266) ((T_MATCH) reduce 266) ((T_METHOD_C) reduce 266) ((T_NAMESPACE) reduce 266) ((T_NEW) reduce 266) ((T_NS_C) reduce 266) ((T_PRINT) reduce 266) ((T_PRIVATE) reduce 266) ((T_PROPERTY_C) reduce 266) ((T_PROTECTED) reduce 266) ((T_PUBLIC) reduce 266) ((T_READONLY) reduce 266) ((T_REQUIRE) reduce 266) ((T_REQUIRE_ONCE) reduce 266) ((T_RETURN) reduce 266) ((T_STATIC) reduce 266) ((T_STRING) reduce 266) ((T_SWITCH) reduce 266) ((T_THROW) reduce 266) ((T_TRAIT) reduce 266) ((T_TRAIT_C) reduce 266) ((T_TRY) reduce 266) ((T_UNSET) reduce 266) ((T_USE) reduce 266) ((T_VAR) reduce 266) ((T_WHILE) reduce 266) ((T_YIELD) reduce 266)) 7211 (((T_ABSTRACT) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_CASE) reduce 102) ((T_CONST) reduce 102) ((T_FINAL) reduce 102) ((T_FUNCTION) reduce 102) ((T_PRIVATE) reduce 102) ((T_PRIVATE_SET) reduce 102) ((T_PROTECTED) reduce 102) ((T_PROTECTED_SET) reduce 102) ((T_PUBLIC) reduce 102) ((T_PUBLIC_SET) reduce 102) ((T_READONLY) reduce 102) ((T_STATIC) reduce 102) ((T_VAR) reduce 102)) 7212 (((T_ABSTRACT) reduce 313) ((T_ATTRIBUTE) reduce 313) ((T_CASE) reduce 313) ((T_CONST) reduce 313) ((T_FINAL) reduce 313) ((T_FUNCTION) reduce 313) ((T_PRIVATE) reduce 313) ((T_PRIVATE_SET) reduce 313) ((T_PROTECTED) reduce 313) ((T_PROTECTED_SET) reduce 313) ((T_PUBLIC) reduce 313) ((T_PUBLIC_SET) reduce 313) ((T_READONLY) reduce 313) ((T_STATIC) reduce 313) ((T_USE) reduce 313) ((T_VAR) reduce 313) (("}") reduce 313)) 7213 (((",") reduce 94) ((";") reduce 94) (("{") reduce 94)) 7214 (((",") reduce 93) ((";") reduce 93) (("{") reduce 93)) 7215 (((",") reduce 95) ((";") reduce 95) (("{") reduce 95)) 7216 (((",") reduce 492) ((";") reduce 492) (("{") reduce 492)) 7217 (((",") reduce 92) ((";") reduce 92) (("{") reduce 92)) 7218 (((",") reduce 315) ((";") reduce 315) (("{") reduce 315)) 7219 (((",") shift) ((";") shift) (("{") shift)) 7220 (((",") reduce 493) ((";") reduce 493) (("{") reduce 493)) 7222 (((T_ABSTRACT) reduce 317) ((T_ATTRIBUTE) reduce 317) ((T_CASE) reduce 317) ((T_CONST) reduce 317) ((T_FINAL) reduce 317) ((T_FUNCTION) reduce 317) ((T_PRIVATE) reduce 317) ((T_PRIVATE_SET) reduce 317) ((T_PROTECTED) reduce 317) ((T_PROTECTED_SET) reduce 317) ((T_PUBLIC) reduce 317) ((T_PUBLIC_SET) reduce 317) ((T_READONLY) reduce 317) ((T_STATIC) reduce 317) ((T_USE) reduce 317) ((T_VAR) reduce 317) (("}") reduce 317)) 7223 (((T_ABSTRACT) reduce 314) ((T_ATTRIBUTE) reduce 314) ((T_CASE) reduce 314) ((T_CONST) reduce 314) ((T_FINAL) reduce 314) ((T_FUNCTION) reduce 314) ((T_PRIVATE) reduce 314) ((T_PRIVATE_SET) reduce 314) ((T_PROTECTED) reduce 314) ((T_PROTECTED_SET) reduce 314) ((T_PUBLIC) reduce 314) ((T_PUBLIC_SET) reduce 314) ((T_READONLY) reduce 314) ((T_STATIC) reduce 314) ((T_USE) reduce 314) ((T_VAR) reduce 314) (("}") reduce 314)) 7224 (((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift) (("}") shift)) 7225 (((T_AS) reduce 74)) 7226 (((T_AS) reduce 52)) 7227 (((T_AS) reduce 27)) 7228 (((T_AS) reduce 51)) 7229 (((T_AS) reduce 53)) 7230 (((T_AS) reduce 49)) 7231 (((T_AS) reduce 29)) 7232 (((T_AS) reduce 59)) 7233 (((T_AS) reduce 60)) 7234 (((T_AS) reduce 11)) 7235 (((T_AS) reduce 42)) 7236 (((T_AS) reduce 39)) 7237 (((T_AS) reduce 25)) 7238 (((T_AS) reduce 50)) 7239 (((T_AS) reduce 66)) 7240 (((T_AS) reduce 18)) 7241 (((T_AS) reduce 17)) 7242 (((T_AS) reduce 15)) 7243 (((T_AS) reduce 14)) 7244 (((T_AS) reduce 38)) 7245 (((T_AS) reduce 26)) 7246 (((T_AS) reduce 22)) 7247 (((T_AS) reduce 24)) 7248 (((T_AS) reduce 16)) 7249 (((T_AS) reduce 48)) 7250 (((T_AS) reduce 20)) 7251 (((T_AS) reduce 70)) 7252 (((T_AS) reduce 3)) 7253 (((T_AS) reduce 12)) 7254 (((T_AS) reduce 54)) 7255 (((T_AS) reduce 65)) 7256 (((T_AS) reduce 75)) 7257 (((T_AS) reduce 30)) 7258 (((T_AS) reduce 68)) 7259 (((T_AS) reduce 21)) 7260 (((T_AS) reduce 23)) 7261 (((T_AS) reduce 41)) 7262 (((T_AS) reduce 62)) 7263 (((T_AS) reduce 34)) 7264 (((T_AS) reduce 40)) 7265 (((T_AS) reduce 13)) 7266 (((T_AS) reduce 55)) 7267 (((T_AS) reduce 1)) 7268 (((T_AS) reduce 2)) 7269 (((T_AS) reduce 9)) 7270 (((T_AS) reduce 33)) 7271 (((T_AS) reduce 58)) 7272 (((T_AS) reduce 37)) 7273 (((T_AS) reduce 64)) 7274 (((T_AS) reduce 46)) 7275 (((T_AS) reduce 8)) 7276 (((T_AS) reduce 6)) 7277 (((T_AS) reduce 7)) 7278 (((T_AS) reduce 69)) 7279 (((T_AS) reduce 63)) 7280 (((T_AS) reduce 56)) 7281 (((T_PAAMAYIM_NEKUDOTAYIM) reduce 94)) 7282 (((T_PAAMAYIM_NEKUDOTAYIM) reduce 93)) 7283 (((T_PAAMAYIM_NEKUDOTAYIM) reduce 95)) 7284 (((T_AS) reduce 10)) 7285 (((T_AS) reduce 67)) 7286 (((T_AS) reduce 44)) 7287 (((T_AS) reduce 76)) 7288 (((T_AS) reduce 71)) 7289 (((T_AS) reduce 77)) 7290 (((T_AS) reduce 78)) 7291 (((T_AS) reduce 79)) 7292 (((T_AS) reduce 4)) 7293 (((T_AS) reduce 5)) 7294 (((T_AS) reduce 43)) 7295 (((T_AS) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492)) 7296 (((T_AS) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92)) 7297 (((T_AS) reduce 47)) 7298 (((T_AS) reduce 31)) 7299 (((T_AS) reduce 57)) 7300 (((T_AS) reduce 61)) 7301 (((T_AS) reduce 28)) 7302 (((T_AS) reduce 36)) 7303 (((T_AS) reduce 32)) 7304 (((T_AS) reduce 35)) 7305 (((T_AS) reduce 19)) 7306 (((T_AS) reduce 45)) 7307 (((T_AS) reduce 330) ((T_INSTEADOF) shift)) 7309 (((T_AS) reduce 329)) 7310 (((T_PAAMAYIM_NEKUDOTAYIM) reduce 493)) 7311 (((T_AS) reduce 72)) 7312 (((T_AS) reduce 83)) 7313 (((T_ABSTRACT) reduce 320) ((T_ARRAY) reduce 320) ((T_AS) reduce 320) ((T_BREAK) reduce 320) ((T_CALLABLE) reduce 320) ((T_CASE) reduce 320) ((T_CATCH) reduce 320) ((T_CLASS) reduce 320) ((T_CLASS_C) reduce 320) ((T_CLONE) reduce 320) ((T_CONST) reduce 320) ((T_CONTINUE) reduce 320) ((T_DECLARE) reduce 320) ((T_DEFAULT) reduce 320) ((T_DIR) reduce 320) ((T_DO) reduce 320) ((T_ECHO) reduce 320) ((T_ELSE) reduce 320) ((T_ELSEIF) reduce 320) ((T_EMPTY) reduce 320) ((T_ENDDECLARE) reduce 320) ((T_ENDFOR) reduce 320) ((T_ENDFOREACH) reduce 320) ((T_ENDIF) reduce 320) ((T_ENDSWITCH) reduce 320) ((T_ENDWHILE) reduce 320) ((T_ENUM) reduce 320) ((T_EVAL) reduce 320) ((T_EXIT) reduce 320) ((T_EXTENDS) reduce 320) ((T_FILE) reduce 320) ((T_FINAL) reduce 320) ((T_FINALLY) reduce 320) ((T_FN) reduce 320) ((T_FOR) reduce 320) ((T_FOREACH) reduce 320) ((T_FUNCTION) reduce 320) ((T_FUNC_C) reduce 320) ((T_GLOBAL) reduce 320) ((T_GOTO) reduce 320) ((T_IF) reduce 320) ((T_IMPLEMENTS) reduce 320) ((T_INCLUDE) reduce 320) ((T_INCLUDE_ONCE) reduce 320) ((T_INSTANCEOF) reduce 320) ((T_INSTEADOF) reduce 320) ((T_INTERFACE) reduce 320) ((T_ISSET) reduce 320) ((T_LINE) reduce 320) ((T_LIST) reduce 320) ((T_LOGICAL_AND) reduce 320) ((T_LOGICAL_OR) reduce 320) ((T_LOGICAL_XOR) reduce 320) ((T_MATCH) reduce 320) ((T_METHOD_C) reduce 320) ((T_NAMESPACE) reduce 320) ((T_NAME_FULLY_QUALIFIED) reduce 320) ((T_NAME_QUALIFIED) reduce 320) ((T_NAME_RELATIVE) reduce 320) ((T_NEW) reduce 320) ((T_NS_C) reduce 320) ((T_PRINT) reduce 320) ((T_PRIVATE) reduce 320) ((T_PROPERTY_C) reduce 320) ((T_PROTECTED) reduce 320) ((T_PUBLIC) reduce 320) ((T_READONLY) reduce 320) ((T_REQUIRE) reduce 320) ((T_REQUIRE_ONCE) reduce 320) ((T_RETURN) reduce 320) ((T_STATIC) reduce 320) ((T_STRING) reduce 320) ((T_SWITCH) reduce 320) ((T_THROW) reduce 320) ((T_TRAIT) reduce 320) ((T_TRAIT_C) reduce 320) ((T_TRY) reduce 320) ((T_UNSET) reduce 320) ((T_USE) reduce 320) ((T_VAR) reduce 320) ((T_WHILE) reduce 320) ((T_YIELD) reduce 320) (("}") reduce 320)) 7315 (((";") shift)) 7316 (((T_AS) shift)) 7318 (((T_ABSTRACT) reduce 318) ((T_ATTRIBUTE) reduce 318) ((T_CASE) reduce 318) ((T_CONST) reduce 318) ((T_FINAL) reduce 318) ((T_FUNCTION) reduce 318) ((T_PRIVATE) reduce 318) ((T_PRIVATE_SET) reduce 318) ((T_PROTECTED) reduce 318) ((T_PROTECTED_SET) reduce 318) ((T_PUBLIC) reduce 318) ((T_PUBLIC_SET) reduce 318) ((T_READONLY) reduce 318) ((T_STATIC) reduce 318) ((T_USE) reduce 318) ((T_VAR) reduce 318) (("}") reduce 318)) 7319 (((T_ABSTRACT) reduce 322) ((T_ARRAY) reduce 322) ((T_AS) reduce 322) ((T_BREAK) reduce 322) ((T_CALLABLE) reduce 322) ((T_CASE) reduce 322) ((T_CATCH) reduce 322) ((T_CLASS) reduce 322) ((T_CLASS_C) reduce 322) ((T_CLONE) reduce 322) ((T_CONST) reduce 322) ((T_CONTINUE) reduce 322) ((T_DECLARE) reduce 322) ((T_DEFAULT) reduce 322) ((T_DIR) reduce 322) ((T_DO) reduce 322) ((T_ECHO) reduce 322) ((T_ELSE) reduce 322) ((T_ELSEIF) reduce 322) ((T_EMPTY) reduce 322) ((T_ENDDECLARE) reduce 322) ((T_ENDFOR) reduce 322) ((T_ENDFOREACH) reduce 322) ((T_ENDIF) reduce 322) ((T_ENDSWITCH) reduce 322) ((T_ENDWHILE) reduce 322) ((T_ENUM) reduce 322) ((T_EVAL) reduce 322) ((T_EXIT) reduce 322) ((T_EXTENDS) reduce 322) ((T_FILE) reduce 322) ((T_FINAL) reduce 322) ((T_FINALLY) reduce 322) ((T_FN) reduce 322) ((T_FOR) reduce 322) ((T_FOREACH) reduce 322) ((T_FUNCTION) reduce 322) ((T_FUNC_C) reduce 322) ((T_GLOBAL) reduce 322) ((T_GOTO) reduce 322) ((T_IF) reduce 322) ((T_IMPLEMENTS) reduce 322) ((T_INCLUDE) reduce 322) ((T_INCLUDE_ONCE) reduce 322) ((T_INSTANCEOF) reduce 322) ((T_INSTEADOF) reduce 322) ((T_INTERFACE) reduce 322) ((T_ISSET) reduce 322) ((T_LINE) reduce 322) ((T_LIST) reduce 322) ((T_LOGICAL_AND) reduce 322) ((T_LOGICAL_OR) reduce 322) ((T_LOGICAL_XOR) reduce 322) ((T_MATCH) reduce 322) ((T_METHOD_C) reduce 322) ((T_NAMESPACE) reduce 322) ((T_NAME_FULLY_QUALIFIED) reduce 322) ((T_NAME_QUALIFIED) reduce 322) ((T_NAME_RELATIVE) reduce 322) ((T_NEW) reduce 322) ((T_NS_C) reduce 322) ((T_PRINT) reduce 322) ((T_PRIVATE) reduce 322) ((T_PROPERTY_C) reduce 322) ((T_PROTECTED) reduce 322) ((T_PUBLIC) reduce 322) ((T_READONLY) reduce 322) ((T_REQUIRE) reduce 322) ((T_REQUIRE_ONCE) reduce 322) ((T_RETURN) reduce 322) ((T_STATIC) reduce 322) ((T_STRING) reduce 322) ((T_SWITCH) reduce 322) ((T_THROW) reduce 322) ((T_TRAIT) reduce 322) ((T_TRAIT_C) reduce 322) ((T_TRY) reduce 322) ((T_UNSET) reduce 322) ((T_USE) reduce 322) ((T_VAR) reduce 322) ((T_WHILE) reduce 322) ((T_YIELD) reduce 322) (("}") reduce 322)) 7320 (((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift)) 7321 (((";") reduce 349) ((T_ABSTRACT) reduce 349) ((T_ARRAY) reduce 349) ((T_AS) reduce 349) ((T_BREAK) reduce 349) ((T_CALLABLE) reduce 349) ((T_CASE) reduce 349) ((T_CATCH) reduce 349) ((T_CLASS) reduce 349) ((T_CLASS_C) reduce 349) ((T_CLONE) reduce 349) ((T_CONST) reduce 349) ((T_CONTINUE) reduce 349) ((T_DECLARE) reduce 349) ((T_DEFAULT) reduce 349) ((T_DIR) reduce 349) ((T_DO) reduce 349) ((T_ECHO) reduce 349) ((T_ELSE) reduce 349) ((T_ELSEIF) reduce 349) ((T_EMPTY) reduce 349) ((T_ENDDECLARE) reduce 349) ((T_ENDFOR) reduce 349) ((T_ENDFOREACH) reduce 349) ((T_ENDIF) reduce 349) ((T_ENDSWITCH) reduce 349) ((T_ENDWHILE) reduce 349) ((T_ENUM) reduce 349) ((T_EVAL) reduce 349) ((T_EXIT) reduce 349) ((T_EXTENDS) reduce 349) ((T_FILE) reduce 349) ((T_FINAL) reduce 349) ((T_FINALLY) reduce 349) ((T_FN) reduce 349) ((T_FOR) reduce 349) ((T_FOREACH) reduce 349) ((T_FUNCTION) reduce 349) ((T_FUNC_C) reduce 349) ((T_GLOBAL) reduce 349) ((T_GOTO) reduce 349) ((T_IF) reduce 349) ((T_IMPLEMENTS) reduce 349) ((T_INCLUDE) reduce 349) ((T_INCLUDE_ONCE) reduce 349) ((T_INSTANCEOF) reduce 349) ((T_INSTEADOF) reduce 349) ((T_INTERFACE) reduce 349) ((T_ISSET) reduce 349) ((T_LINE) reduce 349) ((T_LIST) reduce 349) ((T_LOGICAL_AND) reduce 349) ((T_LOGICAL_OR) reduce 349) ((T_LOGICAL_XOR) reduce 349) ((T_MATCH) reduce 349) ((T_METHOD_C) reduce 349) ((T_NAMESPACE) reduce 349) ((T_NEW) reduce 349) ((T_NS_C) reduce 349) ((T_PRINT) reduce 349) ((T_PRIVATE) reduce 349) ((T_PROPERTY_C) reduce 349) ((T_PROTECTED) reduce 349) ((T_PUBLIC) reduce 349) ((T_READONLY) reduce 349) ((T_REQUIRE) reduce 349) ((T_REQUIRE_ONCE) reduce 349) ((T_RETURN) reduce 349) ((T_STATIC) reduce 349) ((T_STRING) reduce 349) ((T_SWITCH) reduce 349) ((T_THROW) reduce 349) ((T_TRAIT) reduce 349) ((T_TRAIT_C) reduce 349) ((T_TRY) reduce 349) ((T_UNSET) reduce 349) ((T_USE) reduce 349) ((T_VAR) reduce 349) ((T_WHILE) reduce 349) ((T_YIELD) reduce 349)) 7322 (((";") reduce 52)) 7323 (((";") reduce 27)) 7324 (((";") reduce 51)) 7325 (((";") reduce 53)) 7326 (((";") reduce 49)) 7327 (((";") reduce 29)) 7328 (((";") reduce 59)) 7329 (((";") reduce 60)) 7330 (((";") reduce 11)) 7331 (((";") reduce 42)) 7332 (((";") reduce 39)) 7333 (((";") reduce 25)) 7334 (((";") reduce 50)) 7335 (((";") reduce 66)) 7336 (((";") reduce 18)) 7337 (((";") reduce 17)) 7338 (((";") reduce 15)) 7339 (((";") reduce 14)) 7340 (((";") reduce 38)) 7341 (((";") reduce 26)) 7342 (((";") reduce 22)) 7343 (((";") reduce 24)) 7344 (((";") reduce 16)) 7345 (((";") reduce 48)) 7346 (((";") reduce 20)) 7347 (((";") reduce 70)) 7348 (((";") reduce 3)) 7349 (((";") reduce 12)) 7350 (((";") reduce 54)) 7351 (((";") reduce 65)) 7352 (((";") reduce 350) ((T_ABSTRACT) reduce 350) ((T_ARRAY) reduce 350) ((T_AS) reduce 350) ((T_BREAK) reduce 350) ((T_CALLABLE) reduce 350) ((T_CASE) reduce 350) ((T_CATCH) reduce 350) ((T_CLASS) reduce 350) ((T_CLASS_C) reduce 350) ((T_CLONE) reduce 350) ((T_CONST) reduce 350) ((T_CONTINUE) reduce 350) ((T_DECLARE) reduce 350) ((T_DEFAULT) reduce 350) ((T_DIR) reduce 350) ((T_DO) reduce 350) ((T_ECHO) reduce 350) ((T_ELSE) reduce 350) ((T_ELSEIF) reduce 350) ((T_EMPTY) reduce 350) ((T_ENDDECLARE) reduce 350) ((T_ENDFOR) reduce 350) ((T_ENDFOREACH) reduce 350) ((T_ENDIF) reduce 350) ((T_ENDSWITCH) reduce 350) ((T_ENDWHILE) reduce 350) ((T_ENUM) reduce 350) ((T_EVAL) reduce 350) ((T_EXIT) reduce 350) ((T_EXTENDS) reduce 350) ((T_FILE) reduce 350) ((T_FINAL) reduce 350) ((T_FINALLY) reduce 350) ((T_FN) reduce 350) ((T_FOR) reduce 350) ((T_FOREACH) reduce 350) ((T_FUNCTION) reduce 350) ((T_FUNC_C) reduce 350) ((T_GLOBAL) reduce 350) ((T_GOTO) reduce 350) ((T_IF) reduce 350) ((T_IMPLEMENTS) reduce 350) ((T_INCLUDE) reduce 350) ((T_INCLUDE_ONCE) reduce 350) ((T_INSTANCEOF) reduce 350) ((T_INSTEADOF) reduce 350) ((T_INTERFACE) reduce 350) ((T_ISSET) reduce 350) ((T_LINE) reduce 350) ((T_LIST) reduce 350) ((T_LOGICAL_AND) reduce 350) ((T_LOGICAL_OR) reduce 350) ((T_LOGICAL_XOR) reduce 350) ((T_MATCH) reduce 350) ((T_METHOD_C) reduce 350) ((T_NAMESPACE) reduce 350) ((T_NEW) reduce 350) ((T_NS_C) reduce 350) ((T_PRINT) reduce 350) ((T_PRIVATE) reduce 350) ((T_PROPERTY_C) reduce 350) ((T_PROTECTED) reduce 350) ((T_PUBLIC) reduce 350) ((T_READONLY) reduce 350) ((T_REQUIRE) reduce 350) ((T_REQUIRE_ONCE) reduce 350) ((T_RETURN) reduce 350) ((T_STATIC) reduce 350) ((T_STRING) reduce 350) ((T_SWITCH) reduce 350) ((T_THROW) reduce 350) ((T_TRAIT) reduce 350) ((T_TRAIT_C) reduce 350) ((T_TRY) reduce 350) ((T_UNSET) reduce 350) ((T_USE) reduce 350) ((T_VAR) reduce 350) ((T_WHILE) reduce 350) ((T_YIELD) reduce 350)) 7353 (((";") reduce 30)) 7354 (((";") reduce 68)) 7355 (((";") reduce 21)) 7356 (((";") reduce 23)) 7357 (((";") reduce 41)) 7358 (((";") reduce 62)) 7359 (((";") reduce 34)) 7360 (((";") reduce 40)) 7361 (((";") reduce 13)) 7362 (((";") reduce 55)) 7363 (((";") reduce 1)) 7364 (((";") reduce 2)) 7365 (((";") reduce 9)) 7366 (((";") reduce 33)) 7367 (((";") reduce 58)) 7368 (((";") reduce 37)) 7369 (((";") reduce 64)) 7370 (((";") reduce 46)) 7371 (((";") reduce 8)) 7372 (((";") reduce 6)) 7373 (((";") reduce 7)) 7374 (((";") reduce 69)) 7375 (((";") reduce 63)) 7376 (((";") reduce 56)) 7377 (((";") reduce 10)) 7378 (((";") reduce 67)) 7379 (((";") reduce 44)) 7380 (((";") reduce 344) ((T_ABSTRACT) reduce 344) ((T_ARRAY) reduce 344) ((T_AS) reduce 344) ((T_BREAK) reduce 344) ((T_CALLABLE) reduce 344) ((T_CASE) reduce 344) ((T_CATCH) reduce 344) ((T_CLASS) reduce 344) ((T_CLASS_C) reduce 344) ((T_CLONE) reduce 344) ((T_CONST) reduce 344) ((T_CONTINUE) reduce 344) ((T_DECLARE) reduce 344) ((T_DEFAULT) reduce 344) ((T_DIR) reduce 344) ((T_DO) reduce 344) ((T_ECHO) reduce 344) ((T_ELSE) reduce 344) ((T_ELSEIF) reduce 344) ((T_EMPTY) reduce 344) ((T_ENDDECLARE) reduce 344) ((T_ENDFOR) reduce 344) ((T_ENDFOREACH) reduce 344) ((T_ENDIF) reduce 344) ((T_ENDSWITCH) reduce 344) ((T_ENDWHILE) reduce 344) ((T_ENUM) reduce 344) ((T_EVAL) reduce 344) ((T_EXIT) reduce 344) ((T_EXTENDS) reduce 344) ((T_FILE) reduce 344) ((T_FINAL) reduce 344) ((T_FINALLY) reduce 344) ((T_FN) reduce 344) ((T_FOR) reduce 344) ((T_FOREACH) reduce 344) ((T_FUNCTION) reduce 344) ((T_FUNC_C) reduce 344) ((T_GLOBAL) reduce 344) ((T_GOTO) reduce 344) ((T_IF) reduce 344) ((T_IMPLEMENTS) reduce 344) ((T_INCLUDE) reduce 344) ((T_INCLUDE_ONCE) reduce 344) ((T_INSTANCEOF) reduce 344) ((T_INSTEADOF) reduce 344) ((T_INTERFACE) reduce 344) ((T_ISSET) reduce 344) ((T_LINE) reduce 344) ((T_LIST) reduce 344) ((T_LOGICAL_AND) reduce 344) ((T_LOGICAL_OR) reduce 344) ((T_LOGICAL_XOR) reduce 344) ((T_MATCH) reduce 344) ((T_METHOD_C) reduce 344) ((T_NAMESPACE) reduce 344) ((T_NEW) reduce 344) ((T_NS_C) reduce 344) ((T_PRINT) reduce 344) ((T_PRIVATE) reduce 344) ((T_PROPERTY_C) reduce 344) ((T_PROTECTED) reduce 344) ((T_PUBLIC) reduce 344) ((T_READONLY) reduce 344) ((T_REQUIRE) reduce 344) ((T_REQUIRE_ONCE) reduce 344) ((T_RETURN) reduce 344) ((T_STATIC) reduce 344) ((T_STRING) reduce 344) ((T_SWITCH) reduce 344) ((T_THROW) reduce 344) ((T_TRAIT) reduce 344) ((T_TRAIT_C) reduce 344) ((T_TRY) reduce 344) ((T_UNSET) reduce 344) ((T_USE) reduce 344) ((T_VAR) reduce 344) ((T_WHILE) reduce 344) ((T_YIELD) reduce 344)) 7381 (((";") reduce 347) ((T_ABSTRACT) reduce 347) ((T_ARRAY) reduce 347) ((T_AS) reduce 347) ((T_BREAK) reduce 347) ((T_CALLABLE) reduce 347) ((T_CASE) reduce 347) ((T_CATCH) reduce 347) ((T_CLASS) reduce 347) ((T_CLASS_C) reduce 347) ((T_CLONE) reduce 347) ((T_CONST) reduce 347) ((T_CONTINUE) reduce 347) ((T_DECLARE) reduce 347) ((T_DEFAULT) reduce 347) ((T_DIR) reduce 347) ((T_DO) reduce 347) ((T_ECHO) reduce 347) ((T_ELSE) reduce 347) ((T_ELSEIF) reduce 347) ((T_EMPTY) reduce 347) ((T_ENDDECLARE) reduce 347) ((T_ENDFOR) reduce 347) ((T_ENDFOREACH) reduce 347) ((T_ENDIF) reduce 347) ((T_ENDSWITCH) reduce 347) ((T_ENDWHILE) reduce 347) ((T_ENUM) reduce 347) ((T_EVAL) reduce 347) ((T_EXIT) reduce 347) ((T_EXTENDS) reduce 347) ((T_FILE) reduce 347) ((T_FINAL) reduce 347) ((T_FINALLY) reduce 347) ((T_FN) reduce 347) ((T_FOR) reduce 347) ((T_FOREACH) reduce 347) ((T_FUNCTION) reduce 347) ((T_FUNC_C) reduce 347) ((T_GLOBAL) reduce 347) ((T_GOTO) reduce 347) ((T_IF) reduce 347) ((T_IMPLEMENTS) reduce 347) ((T_INCLUDE) reduce 347) ((T_INCLUDE_ONCE) reduce 347) ((T_INSTANCEOF) reduce 347) ((T_INSTEADOF) reduce 347) ((T_INTERFACE) reduce 347) ((T_ISSET) reduce 347) ((T_LINE) reduce 347) ((T_LIST) reduce 347) ((T_LOGICAL_AND) reduce 347) ((T_LOGICAL_OR) reduce 347) ((T_LOGICAL_XOR) reduce 347) ((T_MATCH) reduce 347) ((T_METHOD_C) reduce 347) ((T_NAMESPACE) reduce 347) ((T_NEW) reduce 347) ((T_NS_C) reduce 347) ((T_PRINT) reduce 347) ((T_PRIVATE) reduce 347) ((T_PROPERTY_C) reduce 347) ((T_PROTECTED) reduce 347) ((T_PUBLIC) reduce 347) ((T_READONLY) reduce 347) ((T_REQUIRE) reduce 347) ((T_REQUIRE_ONCE) reduce 347) ((T_RETURN) reduce 347) ((T_STATIC) reduce 347) ((T_STRING) reduce 347) ((T_SWITCH) reduce 347) ((T_THROW) reduce 347) ((T_TRAIT) reduce 347) ((T_TRAIT_C) reduce 347) ((T_TRY) reduce 347) ((T_UNSET) reduce 347) ((T_USE) reduce 347) ((T_VAR) reduce 347) ((T_WHILE) reduce 347) ((T_YIELD) reduce 347)) 7382 (((";") reduce 71)) 7383 (((";") reduce 343) ((T_ABSTRACT) reduce 343) ((T_ARRAY) reduce 343) ((T_AS) reduce 343) ((T_BREAK) reduce 343) ((T_CALLABLE) reduce 343) ((T_CASE) reduce 343) ((T_CATCH) reduce 343) ((T_CLASS) reduce 343) ((T_CLASS_C) reduce 343) ((T_CLONE) reduce 343) ((T_CONST) reduce 343) ((T_CONTINUE) reduce 343) ((T_DECLARE) reduce 343) ((T_DEFAULT) reduce 343) ((T_DIR) reduce 343) ((T_DO) reduce 343) ((T_ECHO) reduce 343) ((T_ELSE) reduce 343) ((T_ELSEIF) reduce 343) ((T_EMPTY) reduce 343) ((T_ENDDECLARE) reduce 343) ((T_ENDFOR) reduce 343) ((T_ENDFOREACH) reduce 343) ((T_ENDIF) reduce 343) ((T_ENDSWITCH) reduce 343) ((T_ENDWHILE) reduce 343) ((T_ENUM) reduce 343) ((T_EVAL) reduce 343) ((T_EXIT) reduce 343) ((T_EXTENDS) reduce 343) ((T_FILE) reduce 343) ((T_FINAL) reduce 343) ((T_FINALLY) reduce 343) ((T_FN) reduce 343) ((T_FOR) reduce 343) ((T_FOREACH) reduce 343) ((T_FUNCTION) reduce 343) ((T_FUNC_C) reduce 343) ((T_GLOBAL) reduce 343) ((T_GOTO) reduce 343) ((T_IF) reduce 343) ((T_IMPLEMENTS) reduce 343) ((T_INCLUDE) reduce 343) ((T_INCLUDE_ONCE) reduce 343) ((T_INSTANCEOF) reduce 343) ((T_INSTEADOF) reduce 343) ((T_INTERFACE) reduce 343) ((T_ISSET) reduce 343) ((T_LINE) reduce 343) ((T_LIST) reduce 343) ((T_LOGICAL_AND) reduce 343) ((T_LOGICAL_OR) reduce 343) ((T_LOGICAL_XOR) reduce 343) ((T_MATCH) reduce 343) ((T_METHOD_C) reduce 343) ((T_NAMESPACE) reduce 343) ((T_NEW) reduce 343) ((T_NS_C) reduce 343) ((T_PRINT) reduce 343) ((T_PRIVATE) reduce 343) ((T_PROPERTY_C) reduce 343) ((T_PROTECTED) reduce 343) ((T_PUBLIC) reduce 343) ((T_READONLY) reduce 343) ((T_REQUIRE) reduce 343) ((T_REQUIRE_ONCE) reduce 343) ((T_RETURN) reduce 343) ((T_STATIC) reduce 343) ((T_STRING) reduce 343) ((T_SWITCH) reduce 343) ((T_THROW) reduce 343) ((T_TRAIT) reduce 343) ((T_TRAIT_C) reduce 343) ((T_TRY) reduce 343) ((T_UNSET) reduce 343) ((T_USE) reduce 343) ((T_VAR) reduce 343) ((T_WHILE) reduce 343) ((T_YIELD) reduce 343)) 7384 (((";") reduce 346) ((T_ABSTRACT) reduce 346) ((T_ARRAY) reduce 346) ((T_AS) reduce 346) ((T_BREAK) reduce 346) ((T_CALLABLE) reduce 346) ((T_CASE) reduce 346) ((T_CATCH) reduce 346) ((T_CLASS) reduce 346) ((T_CLASS_C) reduce 346) ((T_CLONE) reduce 346) ((T_CONST) reduce 346) ((T_CONTINUE) reduce 346) ((T_DECLARE) reduce 346) ((T_DEFAULT) reduce 346) ((T_DIR) reduce 346) ((T_DO) reduce 346) ((T_ECHO) reduce 346) ((T_ELSE) reduce 346) ((T_ELSEIF) reduce 346) ((T_EMPTY) reduce 346) ((T_ENDDECLARE) reduce 346) ((T_ENDFOR) reduce 346) ((T_ENDFOREACH) reduce 346) ((T_ENDIF) reduce 346) ((T_ENDSWITCH) reduce 346) ((T_ENDWHILE) reduce 346) ((T_ENUM) reduce 346) ((T_EVAL) reduce 346) ((T_EXIT) reduce 346) ((T_EXTENDS) reduce 346) ((T_FILE) reduce 346) ((T_FINAL) reduce 346) ((T_FINALLY) reduce 346) ((T_FN) reduce 346) ((T_FOR) reduce 346) ((T_FOREACH) reduce 346) ((T_FUNCTION) reduce 346) ((T_FUNC_C) reduce 346) ((T_GLOBAL) reduce 346) ((T_GOTO) reduce 346) ((T_IF) reduce 346) ((T_IMPLEMENTS) reduce 346) ((T_INCLUDE) reduce 346) ((T_INCLUDE_ONCE) reduce 346) ((T_INSTANCEOF) reduce 346) ((T_INSTEADOF) reduce 346) ((T_INTERFACE) reduce 346) ((T_ISSET) reduce 346) ((T_LINE) reduce 346) ((T_LIST) reduce 346) ((T_LOGICAL_AND) reduce 346) ((T_LOGICAL_OR) reduce 346) ((T_LOGICAL_XOR) reduce 346) ((T_MATCH) reduce 346) ((T_METHOD_C) reduce 346) ((T_NAMESPACE) reduce 346) ((T_NEW) reduce 346) ((T_NS_C) reduce 346) ((T_PRINT) reduce 346) ((T_PRIVATE) reduce 346) ((T_PROPERTY_C) reduce 346) ((T_PROTECTED) reduce 346) ((T_PUBLIC) reduce 346) ((T_READONLY) reduce 346) ((T_REQUIRE) reduce 346) ((T_REQUIRE_ONCE) reduce 346) ((T_RETURN) reduce 346) ((T_STATIC) reduce 346) ((T_STRING) reduce 346) ((T_SWITCH) reduce 346) ((T_THROW) reduce 346) ((T_TRAIT) reduce 346) ((T_TRAIT_C) reduce 346) ((T_TRY) reduce 346) ((T_UNSET) reduce 346) ((T_USE) reduce 346) ((T_VAR) reduce 346) ((T_WHILE) reduce 346) ((T_YIELD) reduce 346)) 7385 (((";") reduce 342) ((T_ABSTRACT) reduce 342) ((T_ARRAY) reduce 342) ((T_AS) reduce 342) ((T_BREAK) reduce 342) ((T_CALLABLE) reduce 342) ((T_CASE) reduce 342) ((T_CATCH) reduce 342) ((T_CLASS) reduce 342) ((T_CLASS_C) reduce 342) ((T_CLONE) reduce 342) ((T_CONST) reduce 342) ((T_CONTINUE) reduce 342) ((T_DECLARE) reduce 342) ((T_DEFAULT) reduce 342) ((T_DIR) reduce 342) ((T_DO) reduce 342) ((T_ECHO) reduce 342) ((T_ELSE) reduce 342) ((T_ELSEIF) reduce 342) ((T_EMPTY) reduce 342) ((T_ENDDECLARE) reduce 342) ((T_ENDFOR) reduce 342) ((T_ENDFOREACH) reduce 342) ((T_ENDIF) reduce 342) ((T_ENDSWITCH) reduce 342) ((T_ENDWHILE) reduce 342) ((T_ENUM) reduce 342) ((T_EVAL) reduce 342) ((T_EXIT) reduce 342) ((T_EXTENDS) reduce 342) ((T_FILE) reduce 342) ((T_FINAL) reduce 342) ((T_FINALLY) reduce 342) ((T_FN) reduce 342) ((T_FOR) reduce 342) ((T_FOREACH) reduce 342) ((T_FUNCTION) reduce 342) ((T_FUNC_C) reduce 342) ((T_GLOBAL) reduce 342) ((T_GOTO) reduce 342) ((T_IF) reduce 342) ((T_IMPLEMENTS) reduce 342) ((T_INCLUDE) reduce 342) ((T_INCLUDE_ONCE) reduce 342) ((T_INSTANCEOF) reduce 342) ((T_INSTEADOF) reduce 342) ((T_INTERFACE) reduce 342) ((T_ISSET) reduce 342) ((T_LINE) reduce 342) ((T_LIST) reduce 342) ((T_LOGICAL_AND) reduce 342) ((T_LOGICAL_OR) reduce 342) ((T_LOGICAL_XOR) reduce 342) ((T_MATCH) reduce 342) ((T_METHOD_C) reduce 342) ((T_NAMESPACE) reduce 342) ((T_NEW) reduce 342) ((T_NS_C) reduce 342) ((T_PRINT) reduce 342) ((T_PRIVATE) reduce 342) ((T_PROPERTY_C) reduce 342) ((T_PROTECTED) reduce 342) ((T_PUBLIC) reduce 342) ((T_READONLY) reduce 342) ((T_REQUIRE) reduce 342) ((T_REQUIRE_ONCE) reduce 342) ((T_RETURN) reduce 342) ((T_STATIC) reduce 342) ((T_STRING) reduce 342) ((T_SWITCH) reduce 342) ((T_THROW) reduce 342) ((T_TRAIT) reduce 342) ((T_TRAIT_C) reduce 342) ((T_TRY) reduce 342) ((T_UNSET) reduce 342) ((T_USE) reduce 342) ((T_VAR) reduce 342) ((T_WHILE) reduce 342) ((T_YIELD) reduce 342)) 7386 (((";") reduce 345) ((T_ABSTRACT) reduce 345) ((T_ARRAY) reduce 345) ((T_AS) reduce 345) ((T_BREAK) reduce 345) ((T_CALLABLE) reduce 345) ((T_CASE) reduce 345) ((T_CATCH) reduce 345) ((T_CLASS) reduce 345) ((T_CLASS_C) reduce 345) ((T_CLONE) reduce 345) ((T_CONST) reduce 345) ((T_CONTINUE) reduce 345) ((T_DECLARE) reduce 345) ((T_DEFAULT) reduce 345) ((T_DIR) reduce 345) ((T_DO) reduce 345) ((T_ECHO) reduce 345) ((T_ELSE) reduce 345) ((T_ELSEIF) reduce 345) ((T_EMPTY) reduce 345) ((T_ENDDECLARE) reduce 345) ((T_ENDFOR) reduce 345) ((T_ENDFOREACH) reduce 345) ((T_ENDIF) reduce 345) ((T_ENDSWITCH) reduce 345) ((T_ENDWHILE) reduce 345) ((T_ENUM) reduce 345) ((T_EVAL) reduce 345) ((T_EXIT) reduce 345) ((T_EXTENDS) reduce 345) ((T_FILE) reduce 345) ((T_FINAL) reduce 345) ((T_FINALLY) reduce 345) ((T_FN) reduce 345) ((T_FOR) reduce 345) ((T_FOREACH) reduce 345) ((T_FUNCTION) reduce 345) ((T_FUNC_C) reduce 345) ((T_GLOBAL) reduce 345) ((T_GOTO) reduce 345) ((T_IF) reduce 345) ((T_IMPLEMENTS) reduce 345) ((T_INCLUDE) reduce 345) ((T_INCLUDE_ONCE) reduce 345) ((T_INSTANCEOF) reduce 345) ((T_INSTEADOF) reduce 345) ((T_INTERFACE) reduce 345) ((T_ISSET) reduce 345) ((T_LINE) reduce 345) ((T_LIST) reduce 345) ((T_LOGICAL_AND) reduce 345) ((T_LOGICAL_OR) reduce 345) ((T_LOGICAL_XOR) reduce 345) ((T_MATCH) reduce 345) ((T_METHOD_C) reduce 345) ((T_NAMESPACE) reduce 345) ((T_NEW) reduce 345) ((T_NS_C) reduce 345) ((T_PRINT) reduce 345) ((T_PRIVATE) reduce 345) ((T_PROPERTY_C) reduce 345) ((T_PROTECTED) reduce 345) ((T_PUBLIC) reduce 345) ((T_READONLY) reduce 345) ((T_REQUIRE) reduce 345) ((T_REQUIRE_ONCE) reduce 345) ((T_RETURN) reduce 345) ((T_STATIC) reduce 345) ((T_STRING) reduce 345) ((T_SWITCH) reduce 345) ((T_THROW) reduce 345) ((T_TRAIT) reduce 345) ((T_TRAIT_C) reduce 345) ((T_TRY) reduce 345) ((T_UNSET) reduce 345) ((T_USE) reduce 345) ((T_VAR) reduce 345) ((T_WHILE) reduce 345) ((T_YIELD) reduce 345)) 7387 (((";") reduce 351) ((T_ABSTRACT) reduce 351) ((T_ARRAY) reduce 351) ((T_AS) reduce 351) ((T_BREAK) reduce 351) ((T_CALLABLE) reduce 351) ((T_CASE) reduce 351) ((T_CATCH) reduce 351) ((T_CLASS) reduce 351) ((T_CLASS_C) reduce 351) ((T_CLONE) reduce 351) ((T_CONST) reduce 351) ((T_CONTINUE) reduce 351) ((T_DECLARE) reduce 351) ((T_DEFAULT) reduce 351) ((T_DIR) reduce 351) ((T_DO) reduce 351) ((T_ECHO) reduce 351) ((T_ELSE) reduce 351) ((T_ELSEIF) reduce 351) ((T_EMPTY) reduce 351) ((T_ENDDECLARE) reduce 351) ((T_ENDFOR) reduce 351) ((T_ENDFOREACH) reduce 351) ((T_ENDIF) reduce 351) ((T_ENDSWITCH) reduce 351) ((T_ENDWHILE) reduce 351) ((T_ENUM) reduce 351) ((T_EVAL) reduce 351) ((T_EXIT) reduce 351) ((T_EXTENDS) reduce 351) ((T_FILE) reduce 351) ((T_FINAL) reduce 351) ((T_FINALLY) reduce 351) ((T_FN) reduce 351) ((T_FOR) reduce 351) ((T_FOREACH) reduce 351) ((T_FUNCTION) reduce 351) ((T_FUNC_C) reduce 351) ((T_GLOBAL) reduce 351) ((T_GOTO) reduce 351) ((T_IF) reduce 351) ((T_IMPLEMENTS) reduce 351) ((T_INCLUDE) reduce 351) ((T_INCLUDE_ONCE) reduce 351) ((T_INSTANCEOF) reduce 351) ((T_INSTEADOF) reduce 351) ((T_INTERFACE) reduce 351) ((T_ISSET) reduce 351) ((T_LINE) reduce 351) ((T_LIST) reduce 351) ((T_LOGICAL_AND) reduce 351) ((T_LOGICAL_OR) reduce 351) ((T_LOGICAL_XOR) reduce 351) ((T_MATCH) reduce 351) ((T_METHOD_C) reduce 351) ((T_NAMESPACE) reduce 351) ((T_NEW) reduce 351) ((T_NS_C) reduce 351) ((T_PRINT) reduce 351) ((T_PRIVATE) reduce 351) ((T_PROPERTY_C) reduce 351) ((T_PROTECTED) reduce 351) ((T_PUBLIC) reduce 351) ((T_READONLY) reduce 351) ((T_REQUIRE) reduce 351) ((T_REQUIRE_ONCE) reduce 351) ((T_RETURN) reduce 351) ((T_STATIC) reduce 351) ((T_STRING) reduce 351) ((T_SWITCH) reduce 351) ((T_THROW) reduce 351) ((T_TRAIT) reduce 351) ((T_TRAIT_C) reduce 351) ((T_TRY) reduce 351) ((T_UNSET) reduce 351) ((T_USE) reduce 351) ((T_VAR) reduce 351) ((T_WHILE) reduce 351) ((T_YIELD) reduce 351)) 7388 (((";") reduce 4)) 7389 (((";") reduce 5)) 7390 (((";") reduce 43)) 7391 (((";") reduce 348) ((T_ABSTRACT) reduce 348) ((T_ARRAY) reduce 348) ((T_AS) reduce 348) ((T_BREAK) reduce 348) ((T_CALLABLE) reduce 348) ((T_CASE) reduce 348) ((T_CATCH) reduce 348) ((T_CLASS) reduce 348) ((T_CLASS_C) reduce 348) ((T_CLONE) reduce 348) ((T_CONST) reduce 348) ((T_CONTINUE) reduce 348) ((T_DECLARE) reduce 348) ((T_DEFAULT) reduce 348) ((T_DIR) reduce 348) ((T_DO) reduce 348) ((T_ECHO) reduce 348) ((T_ELSE) reduce 348) ((T_ELSEIF) reduce 348) ((T_EMPTY) reduce 348) ((T_ENDDECLARE) reduce 348) ((T_ENDFOR) reduce 348) ((T_ENDFOREACH) reduce 348) ((T_ENDIF) reduce 348) ((T_ENDSWITCH) reduce 348) ((T_ENDWHILE) reduce 348) ((T_ENUM) reduce 348) ((T_EVAL) reduce 348) ((T_EXIT) reduce 348) ((T_EXTENDS) reduce 348) ((T_FILE) reduce 348) ((T_FINAL) reduce 348) ((T_FINALLY) reduce 348) ((T_FN) reduce 348) ((T_FOR) reduce 348) ((T_FOREACH) reduce 348) ((T_FUNCTION) reduce 348) ((T_FUNC_C) reduce 348) ((T_GLOBAL) reduce 348) ((T_GOTO) reduce 348) ((T_IF) reduce 348) ((T_IMPLEMENTS) reduce 348) ((T_INCLUDE) reduce 348) ((T_INCLUDE_ONCE) reduce 348) ((T_INSTANCEOF) reduce 348) ((T_INSTEADOF) reduce 348) ((T_INTERFACE) reduce 348) ((T_ISSET) reduce 348) ((T_LINE) reduce 348) ((T_LIST) reduce 348) ((T_LOGICAL_AND) reduce 348) ((T_LOGICAL_OR) reduce 348) ((T_LOGICAL_XOR) reduce 348) ((T_MATCH) reduce 348) ((T_METHOD_C) reduce 348) ((T_NAMESPACE) reduce 348) ((T_NEW) reduce 348) ((T_NS_C) reduce 348) ((T_PRINT) reduce 348) ((T_PRIVATE) reduce 348) ((T_PROPERTY_C) reduce 348) ((T_PROTECTED) reduce 348) ((T_PUBLIC) reduce 348) ((T_READONLY) reduce 348) ((T_REQUIRE) reduce 348) ((T_REQUIRE_ONCE) reduce 348) ((T_RETURN) reduce 348) ((T_STATIC) reduce 348) ((T_STRING) reduce 348) ((T_SWITCH) reduce 348) ((T_THROW) reduce 348) ((T_TRAIT) reduce 348) ((T_TRAIT_C) reduce 348) ((T_TRY) reduce 348) ((T_UNSET) reduce 348) ((T_USE) reduce 348) ((T_VAR) reduce 348) ((T_WHILE) reduce 348) ((T_YIELD) reduce 348)) 7392 (((";") reduce 325)) 7393 (((";") reduce 47)) 7394 (((";") reduce 31)) 7395 (((";") reduce 57)) 7396 (((";") reduce 61)) 7397 (((";") reduce 28)) 7398 (((";") reduce 36)) 7399 (((";") reduce 32)) 7400 (((";") reduce 35)) 7401 (((";") reduce 19)) 7402 (((";") reduce 45)) 7403 (((";") reduce 328) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift)) 7404 (((";") reduce 326)) 7405 (((";") reduce 74)) 7406 (((";") reduce 75)) 7407 (((";") reduce 76)) 7408 (((";") reduce 77)) 7409 (((";") reduce 78)) 7410 (((";") reduce 79)) 7411 (((";") reduce 73)) 7412 (((";") reduce 82)) 7413 (((";") reduce 327)) 7414 (((";") reduce 72)) 7415 (((";") reduce 83)) 7416 (((T_ABSTRACT) reduce 323) ((T_ARRAY) reduce 323) ((T_AS) reduce 323) ((T_BREAK) reduce 323) ((T_CALLABLE) reduce 323) ((T_CASE) reduce 323) ((T_CATCH) reduce 323) ((T_CLASS) reduce 323) ((T_CLASS_C) reduce 323) ((T_CLONE) reduce 323) ((T_CONST) reduce 323) ((T_CONTINUE) reduce 323) ((T_DECLARE) reduce 323) ((T_DEFAULT) reduce 323) ((T_DIR) reduce 323) ((T_DO) reduce 323) ((T_ECHO) reduce 323) ((T_ELSE) reduce 323) ((T_ELSEIF) reduce 323) ((T_EMPTY) reduce 323) ((T_ENDDECLARE) reduce 323) ((T_ENDFOR) reduce 323) ((T_ENDFOREACH) reduce 323) ((T_ENDIF) reduce 323) ((T_ENDSWITCH) reduce 323) ((T_ENDWHILE) reduce 323) ((T_ENUM) reduce 323) ((T_EVAL) reduce 323) ((T_EXIT) reduce 323) ((T_EXTENDS) reduce 323) ((T_FILE) reduce 323) ((T_FINAL) reduce 323) ((T_FINALLY) reduce 323) ((T_FN) reduce 323) ((T_FOR) reduce 323) ((T_FOREACH) reduce 323) ((T_FUNCTION) reduce 323) ((T_FUNC_C) reduce 323) ((T_GLOBAL) reduce 323) ((T_GOTO) reduce 323) ((T_IF) reduce 323) ((T_IMPLEMENTS) reduce 323) ((T_INCLUDE) reduce 323) ((T_INCLUDE_ONCE) reduce 323) ((T_INSTANCEOF) reduce 323) ((T_INSTEADOF) reduce 323) ((T_INTERFACE) reduce 323) ((T_ISSET) reduce 323) ((T_LINE) reduce 323) ((T_LIST) reduce 323) ((T_LOGICAL_AND) reduce 323) ((T_LOGICAL_OR) reduce 323) ((T_LOGICAL_XOR) reduce 323) ((T_MATCH) reduce 323) ((T_METHOD_C) reduce 323) ((T_NAMESPACE) reduce 323) ((T_NAME_FULLY_QUALIFIED) reduce 323) ((T_NAME_QUALIFIED) reduce 323) ((T_NAME_RELATIVE) reduce 323) ((T_NEW) reduce 323) ((T_NS_C) reduce 323) ((T_PRINT) reduce 323) ((T_PRIVATE) reduce 323) ((T_PROPERTY_C) reduce 323) ((T_PROTECTED) reduce 323) ((T_PUBLIC) reduce 323) ((T_READONLY) reduce 323) ((T_REQUIRE) reduce 323) ((T_REQUIRE_ONCE) reduce 323) ((T_RETURN) reduce 323) ((T_STATIC) reduce 323) ((T_STRING) reduce 323) ((T_SWITCH) reduce 323) ((T_THROW) reduce 323) ((T_TRAIT) reduce 323) ((T_TRAIT_C) reduce 323) ((T_TRY) reduce 323) ((T_UNSET) reduce 323) ((T_USE) reduce 323) ((T_VAR) reduce 323) ((T_WHILE) reduce 323) ((T_YIELD) reduce 323) (("}") reduce 323)) 7417 (((T_ABSTRACT) reduce 321) ((T_ARRAY) reduce 321) ((T_AS) reduce 321) ((T_BREAK) reduce 321) ((T_CALLABLE) reduce 321) ((T_CASE) reduce 321) ((T_CATCH) reduce 321) ((T_CLASS) reduce 321) ((T_CLASS_C) reduce 321) ((T_CLONE) reduce 321) ((T_CONST) reduce 321) ((T_CONTINUE) reduce 321) ((T_DECLARE) reduce 321) ((T_DEFAULT) reduce 321) ((T_DIR) reduce 321) ((T_DO) reduce 321) ((T_ECHO) reduce 321) ((T_ELSE) reduce 321) ((T_ELSEIF) reduce 321) ((T_EMPTY) reduce 321) ((T_ENDDECLARE) reduce 321) ((T_ENDFOR) reduce 321) ((T_ENDFOREACH) reduce 321) ((T_ENDIF) reduce 321) ((T_ENDSWITCH) reduce 321) ((T_ENDWHILE) reduce 321) ((T_ENUM) reduce 321) ((T_EVAL) reduce 321) ((T_EXIT) reduce 321) ((T_EXTENDS) reduce 321) ((T_FILE) reduce 321) ((T_FINAL) reduce 321) ((T_FINALLY) reduce 321) ((T_FN) reduce 321) ((T_FOR) reduce 321) ((T_FOREACH) reduce 321) ((T_FUNCTION) reduce 321) ((T_FUNC_C) reduce 321) ((T_GLOBAL) reduce 321) ((T_GOTO) reduce 321) ((T_IF) reduce 321) ((T_IMPLEMENTS) reduce 321) ((T_INCLUDE) reduce 321) ((T_INCLUDE_ONCE) reduce 321) ((T_INSTANCEOF) reduce 321) ((T_INSTEADOF) reduce 321) ((T_INTERFACE) reduce 321) ((T_ISSET) reduce 321) ((T_LINE) reduce 321) ((T_LIST) reduce 321) ((T_LOGICAL_AND) reduce 321) ((T_LOGICAL_OR) reduce 321) ((T_LOGICAL_XOR) reduce 321) ((T_MATCH) reduce 321) ((T_METHOD_C) reduce 321) ((T_NAMESPACE) reduce 321) ((T_NAME_FULLY_QUALIFIED) reduce 321) ((T_NAME_QUALIFIED) reduce 321) ((T_NAME_RELATIVE) reduce 321) ((T_NEW) reduce 321) ((T_NS_C) reduce 321) ((T_PRINT) reduce 321) ((T_PRIVATE) reduce 321) ((T_PROPERTY_C) reduce 321) ((T_PROTECTED) reduce 321) ((T_PUBLIC) reduce 321) ((T_READONLY) reduce 321) ((T_REQUIRE) reduce 321) ((T_REQUIRE_ONCE) reduce 321) ((T_RETURN) reduce 321) ((T_STATIC) reduce 321) ((T_STRING) reduce 321) ((T_SWITCH) reduce 321) ((T_THROW) reduce 321) ((T_TRAIT) reduce 321) ((T_TRAIT_C) reduce 321) ((T_TRY) reduce 321) ((T_UNSET) reduce 321) ((T_USE) reduce 321) ((T_VAR) reduce 321) ((T_WHILE) reduce 321) ((T_YIELD) reduce 321) (("}") reduce 321)) 7418 (((T_ABSTRACT) reduce 319) ((T_ATTRIBUTE) reduce 319) ((T_CASE) reduce 319) ((T_CONST) reduce 319) ((T_FINAL) reduce 319) ((T_FUNCTION) reduce 319) ((T_PRIVATE) reduce 319) ((T_PRIVATE_SET) reduce 319) ((T_PROTECTED) reduce 319) ((T_PROTECTED_SET) reduce 319) ((T_PUBLIC) reduce 319) ((T_PUBLIC_SET) reduce 319) ((T_READONLY) reduce 319) ((T_STATIC) reduce 319) ((T_USE) reduce 319) ((T_VAR) reduce 319) (("}") reduce 319)) 7420 (((T_AS) reduce 74) ((T_INSTEADOF) reduce 74)) 7421 (((T_AS) reduce 52) ((T_INSTEADOF) reduce 52)) 7422 (((T_AS) reduce 27) ((T_INSTEADOF) reduce 27)) 7423 (((T_AS) reduce 51) ((T_INSTEADOF) reduce 51)) 7424 (((T_AS) reduce 53) ((T_INSTEADOF) reduce 53)) 7425 (((T_AS) reduce 49) ((T_INSTEADOF) reduce 49)) 7426 (((T_AS) reduce 29) ((T_INSTEADOF) reduce 29)) 7427 (((T_AS) reduce 59) ((T_INSTEADOF) reduce 59)) 7428 (((T_AS) reduce 60) ((T_INSTEADOF) reduce 60)) 7429 (((T_AS) reduce 11) ((T_INSTEADOF) reduce 11)) 7430 (((T_AS) reduce 42) ((T_INSTEADOF) reduce 42)) 7431 (((T_AS) reduce 39) ((T_INSTEADOF) reduce 39)) 7432 (((T_AS) reduce 25) ((T_INSTEADOF) reduce 25)) 7433 (((T_AS) reduce 50) ((T_INSTEADOF) reduce 50)) 7434 (((T_AS) reduce 66) ((T_INSTEADOF) reduce 66)) 7435 (((T_AS) reduce 18) ((T_INSTEADOF) reduce 18)) 7436 (((T_AS) reduce 17) ((T_INSTEADOF) reduce 17)) 7437 (((T_AS) reduce 15) ((T_INSTEADOF) reduce 15)) 7438 (((T_AS) reduce 14) ((T_INSTEADOF) reduce 14)) 7439 (((T_AS) reduce 38) ((T_INSTEADOF) reduce 38)) 7440 (((T_AS) reduce 26) ((T_INSTEADOF) reduce 26)) 7441 (((T_AS) reduce 22) ((T_INSTEADOF) reduce 22)) 7442 (((T_AS) reduce 24) ((T_INSTEADOF) reduce 24)) 7443 (((T_AS) reduce 16) ((T_INSTEADOF) reduce 16)) 7444 (((T_AS) reduce 48) ((T_INSTEADOF) reduce 48)) 7445 (((T_AS) reduce 20) ((T_INSTEADOF) reduce 20)) 7446 (((T_AS) reduce 70) ((T_INSTEADOF) reduce 70)) 7447 (((T_AS) reduce 3) ((T_INSTEADOF) reduce 3)) 7448 (((T_AS) reduce 12) ((T_INSTEADOF) reduce 12)) 7449 (((T_AS) reduce 54) ((T_INSTEADOF) reduce 54)) 7450 (((T_AS) reduce 65) ((T_INSTEADOF) reduce 65)) 7451 (((T_AS) reduce 75) ((T_INSTEADOF) reduce 75)) 7452 (((T_AS) reduce 30) ((T_INSTEADOF) reduce 30)) 7453 (((T_AS) reduce 68) ((T_INSTEADOF) reduce 68)) 7454 (((T_AS) reduce 21) ((T_INSTEADOF) reduce 21)) 7455 (((T_AS) reduce 23) ((T_INSTEADOF) reduce 23)) 7456 (((T_AS) reduce 41) ((T_INSTEADOF) reduce 41)) 7457 (((T_AS) reduce 62) ((T_INSTEADOF) reduce 62)) 7458 (((T_AS) reduce 34) ((T_INSTEADOF) reduce 34)) 7459 (((T_AS) reduce 40) ((T_INSTEADOF) reduce 40)) 7460 (((T_AS) reduce 13) ((T_INSTEADOF) reduce 13)) 7461 (((T_AS) reduce 55) ((T_INSTEADOF) reduce 55)) 7462 (((T_AS) reduce 1) ((T_INSTEADOF) reduce 1)) 7463 (((T_AS) reduce 2) ((T_INSTEADOF) reduce 2)) 7464 (((T_AS) reduce 9) ((T_INSTEADOF) reduce 9)) 7465 (((T_AS) reduce 33) ((T_INSTEADOF) reduce 33)) 7466 (((T_AS) reduce 58) ((T_INSTEADOF) reduce 58)) 7467 (((T_AS) reduce 37) ((T_INSTEADOF) reduce 37)) 7468 (((T_AS) reduce 64) ((T_INSTEADOF) reduce 64)) 7469 (((T_AS) reduce 46) ((T_INSTEADOF) reduce 46)) 7470 (((T_AS) reduce 8) ((T_INSTEADOF) reduce 8)) 7471 (((T_AS) reduce 6) ((T_INSTEADOF) reduce 6)) 7472 (((T_AS) reduce 7) ((T_INSTEADOF) reduce 7)) 7473 (((T_AS) reduce 69) ((T_INSTEADOF) reduce 69)) 7474 (((T_AS) reduce 63) ((T_INSTEADOF) reduce 63)) 7475 (((T_AS) reduce 56) ((T_INSTEADOF) reduce 56)) 7476 (((T_AS) reduce 10) ((T_INSTEADOF) reduce 10)) 7477 (((T_AS) reduce 67) ((T_INSTEADOF) reduce 67)) 7478 (((T_AS) reduce 44) ((T_INSTEADOF) reduce 44)) 7479 (((T_AS) reduce 76) ((T_INSTEADOF) reduce 76)) 7480 (((T_AS) reduce 71) ((T_INSTEADOF) reduce 71)) 7481 (((T_AS) reduce 77) ((T_INSTEADOF) reduce 77)) 7482 (((T_AS) reduce 78) ((T_INSTEADOF) reduce 78)) 7483 (((T_AS) reduce 79) ((T_INSTEADOF) reduce 79)) 7484 (((T_AS) reduce 4) ((T_INSTEADOF) reduce 4)) 7485 (((T_AS) reduce 5) ((T_INSTEADOF) reduce 5)) 7486 (((T_AS) reduce 43) ((T_INSTEADOF) reduce 43)) 7487 (((T_AS) reduce 73) ((T_INSTEADOF) reduce 73)) 7488 (((T_AS) reduce 82) ((T_INSTEADOF) reduce 82)) 7489 (((T_AS) reduce 47) ((T_INSTEADOF) reduce 47)) 7490 (((T_AS) reduce 31) ((T_INSTEADOF) reduce 31)) 7491 (((T_AS) reduce 57) ((T_INSTEADOF) reduce 57)) 7492 (((T_AS) reduce 61) ((T_INSTEADOF) reduce 61)) 7493 (((T_AS) reduce 28) ((T_INSTEADOF) reduce 28)) 7494 (((T_AS) reduce 36) ((T_INSTEADOF) reduce 36)) 7495 (((T_AS) reduce 32) ((T_INSTEADOF) reduce 32)) 7496 (((T_AS) reduce 35) ((T_INSTEADOF) reduce 35)) 7497 (((T_AS) reduce 19) ((T_INSTEADOF) reduce 19)) 7498 (((T_AS) reduce 45) ((T_INSTEADOF) reduce 45)) 7499 (((T_AS) reduce 331) ((T_INSTEADOF) reduce 331)) 7500 (((T_AS) reduce 72) ((T_INSTEADOF) reduce 72)) 7501 (((T_AS) reduce 83) ((T_INSTEADOF) reduce 83)) 7503 (((",") reduce 94) ((";") reduce 94)) 7504 (((",") reduce 93) ((";") reduce 93)) 7505 (((",") reduce 95) ((";") reduce 95)) 7506 (((",") reduce 492) ((";") reduce 492)) 7507 (((",") reduce 92) ((";") reduce 92)) 7508 (((",") reduce 315) ((";") reduce 315)) 7509 (((",") shift) ((";") reduce 324)) 7510 (((",") reduce 493) ((";") reduce 493)) 7512 (((",") reduce 316) ((";") reduce 316)) 7513 (((",") reduce 316) ((";") reduce 316) (("{") reduce 316)) 7515 (((";") reduce 74) (("=") reduce 74)) 7516 (((";") reduce 52) (("=") reduce 52)) 7517 (((";") reduce 27) (("=") reduce 27)) 7518 (((";") reduce 51) (("=") reduce 51)) 7519 (((";") reduce 53) (("=") reduce 53)) 7520 (((";") reduce 49) (("=") reduce 49)) 7521 (((";") reduce 29) (("=") reduce 29)) 7522 (((";") reduce 59) (("=") reduce 59)) 7523 (((";") reduce 60) (("=") reduce 60)) 7524 (((";") reduce 11) (("=") reduce 11)) 7525 (((";") reduce 42) (("=") reduce 42)) 7526 (((";") reduce 39) (("=") reduce 39)) 7527 (((";") reduce 25) (("=") reduce 25)) 7528 (((";") reduce 50) (("=") reduce 50)) 7529 (((";") reduce 66) (("=") reduce 66)) 7530 (((";") reduce 18) (("=") reduce 18)) 7531 (((";") reduce 17) (("=") reduce 17)) 7532 (((";") reduce 15) (("=") reduce 15)) 7533 (((";") reduce 14) (("=") reduce 14)) 7534 (((";") reduce 38) (("=") reduce 38)) 7535 (((";") reduce 26) (("=") reduce 26)) 7536 (((";") reduce 22) (("=") reduce 22)) 7537 (((";") reduce 24) (("=") reduce 24)) 7538 (((";") reduce 16) (("=") reduce 16)) 7539 (((";") reduce 48) (("=") reduce 48)) 7540 (((";") reduce 20) (("=") reduce 20)) 7541 (((";") reduce 70) (("=") reduce 70)) 7542 (((";") reduce 3) (("=") reduce 3)) 7543 (((";") reduce 12) (("=") reduce 12)) 7544 (((";") reduce 54) (("=") reduce 54)) 7545 (((";") reduce 65) (("=") reduce 65)) 7546 (((";") reduce 75) (("=") reduce 75)) 7547 (((";") reduce 30) (("=") reduce 30)) 7548 (((";") reduce 68) (("=") reduce 68)) 7549 (((";") reduce 21) (("=") reduce 21)) 7550 (((";") reduce 23) (("=") reduce 23)) 7551 (((";") reduce 41) (("=") reduce 41)) 7552 (((";") reduce 62) (("=") reduce 62)) 7553 (((";") reduce 34) (("=") reduce 34)) 7554 (((";") reduce 40) (("=") reduce 40)) 7555 (((";") reduce 13) (("=") reduce 13)) 7556 (((";") reduce 55) (("=") reduce 55)) 7557 (((";") reduce 1) (("=") reduce 1)) 7558 (((";") reduce 2) (("=") reduce 2)) 7559 (((";") reduce 9) (("=") reduce 9)) 7560 (((";") reduce 33) (("=") reduce 33)) 7561 (((";") reduce 58) (("=") reduce 58)) 7562 (((";") reduce 37) (("=") reduce 37)) 7563 (((";") reduce 64) (("=") reduce 64)) 7564 (((";") reduce 46) (("=") reduce 46)) 7565 (((";") reduce 8) (("=") reduce 8)) 7566 (((";") reduce 6) (("=") reduce 6)) 7567 (((";") reduce 7) (("=") reduce 7)) 7568 (((";") reduce 69) (("=") reduce 69)) 7569 (((";") reduce 63) (("=") reduce 63)) 7570 (((";") reduce 56) (("=") reduce 56)) 7571 (((";") reduce 10) (("=") reduce 10)) 7572 (((";") reduce 67) (("=") reduce 67)) 7573 (((";") reduce 44) (("=") reduce 44)) 7574 (((";") reduce 76) (("=") reduce 76)) 7575 (((";") reduce 71) (("=") reduce 71)) 7576 (((";") reduce 77) (("=") reduce 77)) 7577 (((";") reduce 78) (("=") reduce 78)) 7578 (((";") reduce 79) (("=") reduce 79)) 7579 (((";") reduce 4) (("=") reduce 4)) 7580 (((";") reduce 5) (("=") reduce 5)) 7581 (((";") reduce 43) (("=") reduce 43)) 7582 (((";") reduce 73) (("=") reduce 73)) 7583 (((";") reduce 82) (("=") reduce 82)) 7584 (((";") reduce 47) (("=") reduce 47)) 7585 (((";") reduce 31) (("=") reduce 31)) 7586 (((";") reduce 57) (("=") reduce 57)) 7587 (((";") reduce 61) (("=") reduce 61)) 7588 (((";") reduce 28) (("=") reduce 28)) 7589 (((";") reduce 36) (("=") reduce 36)) 7590 (((";") reduce 32) (("=") reduce 32)) 7591 (((";") reduce 35) (("=") reduce 35)) 7592 (((";") reduce 19) (("=") reduce 19)) 7593 (((";") reduce 45) (("=") reduce 45)) 7594 (((";") reduce 204) (("=") shift)) 7595 (((";") reduce 72) (("=") reduce 72)) 7596 (((";") reduce 83) (("=") reduce 83)) 7599 (((T_ABSTRACT) reduce 203) ((T_ATTRIBUTE) reduce 203) ((T_CASE) reduce 203) ((T_CONST) reduce 203) ((T_FINAL) reduce 203) ((T_FUNCTION) reduce 203) ((T_PRIVATE) reduce 203) ((T_PRIVATE_SET) reduce 203) ((T_PROTECTED) reduce 203) ((T_PROTECTED_SET) reduce 203) ((T_PUBLIC) reduce 203) ((T_PUBLIC_SET) reduce 203) ((T_READONLY) reduce 203) ((T_STATIC) reduce 203) ((T_USE) reduce 203) ((T_VAR) reduce 203) (("}") reduce 203)) 7600 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 205) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7603 (((T_ABSTRACT) reduce 100) ((T_ATTRIBUTE) reduce 100) ((T_CASE) reduce 100) ((T_CONST) reduce 100) ((T_FINAL) reduce 100) ((T_FUNCTION) reduce 100) ((T_PRIVATE) reduce 100) ((T_PRIVATE_SET) reduce 100) ((T_PROTECTED) reduce 100) ((T_PROTECTED_SET) reduce 100) ((T_PUBLIC) reduce 100) ((T_PUBLIC_SET) reduce 100) ((T_READONLY) reduce 100) ((T_STATIC) reduce 100) ((T_VAR) reduce 100)) 7604 (((",") reduce 94) (("{") reduce 94)) 7605 (((",") reduce 93) (("{") reduce 93)) 7606 (((",") reduce 95) (("{") reduce 95)) 7607 (((",") reduce 492) (("{") reduce 492)) 7608 (((",") reduce 92) (("{") reduce 92)) 7609 (((",") reduce 315) (("{") reduce 315)) 7610 (((",") shift) (("{") reduce 211)) 7611 (((",") reduce 493) (("{") reduce 493)) 7613 (((",") reduce 316) (("{") reduce 316)) 7614 (((T_IMPLEMENTS) reduce 94) (("{") reduce 94)) 7615 (((T_IMPLEMENTS) reduce 93) (("{") reduce 93)) 7616 (((T_IMPLEMENTS) reduce 95) (("{") reduce 95)) 7617 (((T_IMPLEMENTS) reduce 492) (("{") reduce 492)) 7618 (((T_IMPLEMENTS) reduce 92) (("{") reduce 92)) 7619 (((T_IMPLEMENTS) reduce 207) (("{") reduce 207)) 7620 (((T_IMPLEMENTS) reduce 493) (("{") reduce 493)) 7621 (((T_EXTENDS) reduce 289) ((T_IMPLEMENTS) reduce 289) (("{") reduce 289)) 7625 (((T_EXTENDS) reduce 290) ((T_IMPLEMENTS) reduce 290) (("{") reduce 290)) 7626 (((T_EXTENDS) reduce 291) ((T_IMPLEMENTS) reduce 291) (("{") reduce 291)) 7627 (((T_ABSTRACT) reduce 192) ((T_CLASS) reduce 192) ((T_FINAL) reduce 192) ((T_READONLY) reduce 192)) 7630 (((T_ABSTRACT) reduce 100) ((T_ATTRIBUTE) reduce 100) ((T_CLASS) reduce 100) ((T_FINAL) reduce 100) ((T_READONLY) reduce 100)) 7632 ((("%") reduce 496) (("(") reduce 496) ((")") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 7638 ((("%") reduce 231) ((")") reduce 231) (("*") reduce 231) (("+") reduce 231) ((",") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231)) 7641 (((")") reduce 578) ((",") reduce 578) (("=") shift)) 7645 ((("%") reduce 596) ((")") reduce 596) (("*") reduce 596) (("+") reduce 596) ((",") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596)) 7646 ((("%") reduce 450) ((")") reduce 450) (("*") reduce 450) (("+") reduce 450) ((",") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450)) 7647 ((("%") shift) ((")") reduce 599) (("*") shift) (("+") shift) ((",") reduce 599) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7648 ((("%") shift) ((")") reduce 598) (("*") shift) (("+") shift) ((",") reduce 598) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7649 ((("%") reduce 408) ((")") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408)) 7651 ((("%") reduce 501) ((")") reduce 501) (("*") reduce 501) (("+") reduce 501) ((",") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501)) 7652 ((("%") reduce 457) ((")") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457)) 7653 ((("%") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289)) 7657 ((("%") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290)) 7658 ((("%") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 7661 ((("%") reduce 600) ((")") reduce 600) (("*") reduce 600) (("+") reduce 600) ((",") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600)) 7664 ((("%") reduce 597) ((")") reduce 597) (("*") reduce 597) (("+") reduce 597) ((",") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597)) 7665 ((("%") reduce 451) ((")") reduce 451) (("*") reduce 451) (("+") reduce 451) ((",") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451)) 7666 ((("%") reduce 410) ((")") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410)) 7667 ((("%") reduce 393) ((")") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 7668 ((("%") reduce 455) ((")") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455)) 7669 ((("%") reduce 453) ((")") reduce 453) (("*") reduce 453) (("+") reduce 453) ((",") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453)) 7672 ((("%") reduce 502) (("(") reduce 502) ((")") reduce 502) (("*") reduce 502) (("+") reduce 502) ((",") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502)) 7673 ((("%") reduce 458) ((")") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 7674 ((("%") reduce 430) ((")") reduce 430) (("*") reduce 430) (("+") reduce 430) ((",") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430)) 7675 ((("%") reduce 429) ((")") reduce 429) (("*") reduce 429) (("+") reduce 429) ((",") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429)) 7677 ((("%") reduce 443) (("(") reduce 539) ((")") reduce 443) (("*") reduce 443) (("+") reduce 443) ((",") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443)) 7678 ((("%") reduce 553) (("(") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 7681 ((("%") reduce 552) (("(") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 7683 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 7684 ((("%") reduce 431) ((")") reduce 431) (("*") reduce 431) (("+") reduce 431) ((",") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431)) 7686 ((("%") reduce 524) (("(") reduce 562) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524)) 7688 ((("%") reduce 554) (("(") reduce 564) ((")") reduce 554) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 7691 ((("%") reduce 526) (("(") reduce 563) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526)) 7692 ((("%") reduce 489) (("(") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 7693 ((("%") reduce 491) (("(") reduce 491) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 7695 ((("%") reduce 468) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 7696 ((("%") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 7699 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 389) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 389) (("^") shift) (("|") shift)) 7704 ((("%") reduce 543) (("(") reduce 543) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 7705 ((("%") reduce 565) (("(") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 7706 ((("%") reduce 549) (("(") shift) ((")") reduce 549) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 7707 ((("%") reduce 567) (("(") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 7710 ((("%") reduce 566) (("(") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 7711 ((("%") reduce 544) (("(") reduce 544) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 7712 ((("%") reduce 550) (("(") shift) ((")") reduce 550) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 7713 ((("%") reduce 545) (("(") reduce 545) ((")") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 7714 (((")") reduce 576) ((",") reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) (("[") reduce 531)) 7716 ((("%") reduce 460) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 7718 ((("%") reduce 503) (("(") reduce 503) ((")") reduce 503) (("*") reduce 503) (("+") reduce 503) ((",") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_DOUBLE_ARROW) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503)) 7720 ((("%") shift) ((")") reduce 390) (("*") shift) (("+") shift) ((",") reduce 390) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7721 ((("%") shift) ((")") reduce 465) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7724 ((("%") shift) ((")") reduce 464) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7725 ((("%") reduce 456) ((")") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456)) 7726 ((("%") shift) ((")") reduce 466) (("*") shift) (("+") shift) ((",") reduce 466) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7727 ((("%") reduce 452) ((")") reduce 452) (("*") reduce 452) (("+") reduce 452) ((",") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452)) 7728 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 7730 ((("%") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 7732 ((("%") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 7733 ((("%") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508)) 7734 ((("%") shift) ((")") reduce 602) (("*") shift) (("+") shift) ((",") reduce 602) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 602) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7735 ((("%") shift) ((")") reduce 601) (("*") shift) (("+") shift) ((",") reduce 601) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 601) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7736 ((("%") reduce 488) (("(") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 7737 ((("%") shift) ((")") reduce 461) (("*") shift) (("+") shift) ((",") reduce 461) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7738 ((("%") reduce 454) ((")") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454)) 7740 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 7741 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 7742 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 7743 ((("%") reduce 492) (("(") reduce 492) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 7744 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 7745 ((("%") reduce 385) (("(") reduce 385) ((")") reduce 385) (("*") reduce 385) (("+") reduce 385) ((",") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385)) 7748 ((("%") reduce 494) (("(") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 7749 ((("%") reduce 387) (("(") shift) ((")") reduce 387) (("*") reduce 387) (("+") reduce 387) ((",") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387)) 7750 ((("%") reduce 493) (("(") reduce 493) ((")") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 7751 ((("%") reduce 495) (("(") reduce 495) ((")") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 7752 ((("%") reduce 556) (("(") reduce 556) ((")") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 7758 ((("%") reduce 557) (("(") reduce 557) ((")") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 7759 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 7760 ((("%") reduce 558) (("(") reduce 558) ((")") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 7761 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 7762 ((("%") reduce 384) (("(") reduce 384) ((")") reduce 384) (("*") reduce 384) (("+") reduce 384) ((",") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384)) 7764 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 7765 ((("%") reduce 386) (("(") reduce 386) ((")") reduce 386) (("*") reduce 386) (("+") reduce 386) ((",") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386)) 7773 ((("%") reduce 383) (("(") reduce 383) ((")") reduce 383) (("*") reduce 383) (("+") reduce 383) ((",") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383)) 7775 ((("%") reduce 496) (("(") reduce 496) ((")") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 7781 ((("%") reduce 231) ((")") reduce 231) (("*") reduce 231) (("+") reduce 231) ((",") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_DOUBLE_ARROW) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231)) 7784 (((")") reduce 579) ((",") reduce 579) (("=") shift)) 7788 ((("%") reduce 596) ((")") reduce 596) (("*") reduce 596) (("+") reduce 596) ((",") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_DOUBLE_ARROW) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596)) 7789 ((("%") reduce 450) ((")") reduce 450) (("*") reduce 450) (("+") reduce 450) ((",") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450)) 7790 ((("%") shift) ((")") reduce 599) (("*") shift) (("+") shift) ((",") reduce 599) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 599) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7791 ((("%") shift) ((")") reduce 598) (("*") shift) (("+") shift) ((",") reduce 598) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 598) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7792 ((("%") reduce 408) ((")") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408)) 7794 ((("%") reduce 501) ((")") reduce 501) (("*") reduce 501) (("+") reduce 501) ((",") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_DOUBLE_ARROW) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501)) 7795 ((("%") reduce 457) ((")") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457)) 7796 ((("%") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289)) 7800 ((("%") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290)) 7801 ((("%") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 7804 ((("%") reduce 600) ((")") reduce 600) (("*") reduce 600) (("+") reduce 600) ((",") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_DOUBLE_ARROW) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600)) 7807 ((("%") reduce 597) ((")") reduce 597) (("*") reduce 597) (("+") reduce 597) ((",") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_DOUBLE_ARROW) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597)) 7808 ((("%") shift) ((")") reduce 577) (("*") shift) (("+") shift) ((",") reduce 577) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7809 ((("%") reduce 451) ((")") reduce 451) (("*") reduce 451) (("+") reduce 451) ((",") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451)) 7810 ((("%") reduce 410) ((")") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410)) 7811 ((("%") reduce 393) ((")") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 7812 ((("%") reduce 455) ((")") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455)) 7813 ((("%") reduce 453) ((")") reduce 453) (("*") reduce 453) (("+") reduce 453) ((",") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453)) 7816 ((("%") reduce 502) (("(") reduce 502) ((")") reduce 502) (("*") reduce 502) (("+") reduce 502) ((",") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_DOUBLE_ARROW) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502)) 7817 ((("%") reduce 458) ((")") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 7818 ((("%") reduce 430) ((")") reduce 430) (("*") reduce 430) (("+") reduce 430) ((",") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430)) 7819 ((("%") reduce 429) ((")") reduce 429) (("*") reduce 429) (("+") reduce 429) ((",") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429)) 7821 ((("%") reduce 443) (("(") reduce 539) ((")") reduce 443) (("*") reduce 443) (("+") reduce 443) ((",") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443)) 7822 ((("%") reduce 553) (("(") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 7825 ((("%") reduce 552) (("(") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 7827 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 7828 ((("%") reduce 431) ((")") reduce 431) (("*") reduce 431) (("+") reduce 431) ((",") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431)) 7830 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("]") reduce 74) (("^") reduce 74) (("|") reduce 74)) 7831 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("]") reduce 52) (("^") reduce 52) (("|") reduce 52)) 7832 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("]") reduce 27) (("^") reduce 27) (("|") reduce 27)) 7833 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("]") reduce 51) (("^") reduce 51) (("|") reduce 51)) 7834 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("]") reduce 53) (("^") reduce 53) (("|") reduce 53)) 7835 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("]") reduce 49) (("^") reduce 49) (("|") reduce 49)) 7836 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("]") reduce 29) (("^") reduce 29) (("|") reduce 29)) 7837 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("]") reduce 59) (("^") reduce 59) (("|") reduce 59)) 7838 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("]") reduce 60) (("^") reduce 60) (("|") reduce 60)) 7839 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("]") reduce 11) (("^") reduce 11) (("|") reduce 11)) 7840 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("]") reduce 42) (("^") reduce 42) (("|") reduce 42)) 7841 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("]") reduce 39) (("^") reduce 39) (("|") reduce 39)) 7842 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("]") reduce 25) (("^") reduce 25) (("|") reduce 25)) 7843 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("]") reduce 50) (("^") reduce 50) (("|") reduce 50)) 7844 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("]") reduce 66) (("^") reduce 66) (("|") reduce 66)) 7845 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("]") reduce 18) (("^") reduce 18) (("|") reduce 18)) 7846 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("]") reduce 17) (("^") reduce 17) (("|") reduce 17)) 7847 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("]") reduce 15) (("^") reduce 15) (("|") reduce 15)) 7848 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("]") reduce 14) (("^") reduce 14) (("|") reduce 14)) 7849 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("]") reduce 38) (("^") reduce 38) (("|") reduce 38)) 7850 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("]") reduce 26) (("^") reduce 26) (("|") reduce 26)) 7851 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("]") reduce 22) (("^") reduce 22) (("|") reduce 22)) 7852 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("]") reduce 24) (("^") reduce 24) (("|") reduce 24)) 7853 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("]") reduce 16) (("^") reduce 16) (("|") reduce 16)) 7854 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("]") reduce 48) (("^") reduce 48) (("|") reduce 48)) 7855 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("]") reduce 20) (("^") reduce 20) (("|") reduce 20)) 7856 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("]") reduce 70) (("^") reduce 70) (("|") reduce 70)) 7857 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("]") reduce 3) (("^") reduce 3) (("|") reduce 3)) 7858 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("]") reduce 12) (("^") reduce 12) (("|") reduce 12)) 7859 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("]") reduce 54) (("^") reduce 54) (("|") reduce 54)) 7860 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("]") reduce 65) (("^") reduce 65) (("|") reduce 65)) 7861 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("]") reduce 75) (("^") reduce 75) (("|") reduce 75)) 7862 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("]") reduce 30) (("^") reduce 30) (("|") reduce 30)) 7863 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("]") reduce 68) (("^") reduce 68) (("|") reduce 68)) 7864 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("]") reduce 21) (("^") reduce 21) (("|") reduce 21)) 7865 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("]") reduce 23) (("^") reduce 23) (("|") reduce 23)) 7866 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("]") reduce 41) (("^") reduce 41) (("|") reduce 41)) 7867 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("]") reduce 62) (("^") reduce 62) (("|") reduce 62)) 7868 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("]") reduce 34) (("^") reduce 34) (("|") reduce 34)) 7869 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("]") reduce 40) (("^") reduce 40) (("|") reduce 40)) 7870 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("]") reduce 13) (("^") reduce 13) (("|") reduce 13)) 7871 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("]") reduce 55) (("^") reduce 55) (("|") reduce 55)) 7872 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("]") reduce 1) (("^") reduce 1) (("|") reduce 1)) 7873 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("]") reduce 2) (("^") reduce 2) (("|") reduce 2)) 7874 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("]") reduce 9) (("^") reduce 9) (("|") reduce 9)) 7875 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("]") reduce 33) (("^") reduce 33) (("|") reduce 33)) 7876 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("]") reduce 58) (("^") reduce 58) (("|") reduce 58)) 7877 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("]") reduce 37) (("^") reduce 37) (("|") reduce 37)) 7878 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("]") reduce 64) (("^") reduce 64) (("|") reduce 64)) 7879 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("]") reduce 46) (("^") reduce 46) (("|") reduce 46)) 7880 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("]") reduce 8) (("^") reduce 8) (("|") reduce 8)) 7881 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("]") reduce 6) (("^") reduce 6) (("|") reduce 6)) 7882 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("]") reduce 7) (("^") reduce 7) (("|") reduce 7)) 7883 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("]") reduce 69) (("^") reduce 69) (("|") reduce 69)) 7884 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("]") reduce 63) (("^") reduce 63) (("|") reduce 63)) 7885 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("]") reduce 56) (("^") reduce 56) (("|") reduce 56)) 7886 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("]") reduce 10) (("^") reduce 10) (("|") reduce 10)) 7887 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("]") reduce 67) (("^") reduce 67) (("|") reduce 67)) 7888 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("]") reduce 44) (("^") reduce 44) (("|") reduce 44)) 7889 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("]") reduce 76) (("^") reduce 76) (("|") reduce 76)) 7890 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("]") reduce 71) (("^") reduce 71) (("|") reduce 71)) 7891 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("]") reduce 77) (("^") reduce 77) (("|") reduce 77)) 7892 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("]") reduce 78) (("^") reduce 78) (("|") reduce 78)) 7893 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("]") reduce 79) (("^") reduce 79) (("|") reduce 79)) 7894 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("]") reduce 4) (("^") reduce 4) (("|") reduce 4)) 7895 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("]") reduce 5) (("^") reduce 5) (("|") reduce 5)) 7896 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("]") reduce 43) (("^") reduce 43) (("|") reduce 43)) 7897 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("]") reduce 73) (("^") reduce 73) (("|") reduce 73)) 7898 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("]") reduce 82) (("^") reduce 82) (("|") reduce 82)) 7899 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("]") reduce 47) (("^") reduce 47) (("|") reduce 47)) 7900 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("]") reduce 31) (("^") reduce 31) (("|") reduce 31)) 7901 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("]") reduce 57) (("^") reduce 57) (("|") reduce 57)) 7902 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("]") reduce 61) (("^") reduce 61) (("|") reduce 61)) 7903 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("]") reduce 28) (("^") reduce 28) (("|") reduce 28)) 7904 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("]") reduce 36) (("^") reduce 36) (("|") reduce 36)) 7905 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("]") reduce 32) (("^") reduce 32) (("|") reduce 32)) 7906 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("]") reduce 35) (("^") reduce 35) (("|") reduce 35)) 7907 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("]") reduce 19) (("^") reduce 19) (("|") reduce 19)) 7908 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("]") reduce 45) (("^") reduce 45) (("|") reduce 45)) 7909 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("]") reduce 525) (("^") reduce 525) (("|") reduce 525)) 7911 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("]") reduce 72) (("^") reduce 72) (("|") reduce 72)) 7912 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("]") reduce 83) (("^") reduce 83) (("|") reduce 83)) 7913 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("]") reduce 555) (("^") reduce 555) (("|") reduce 555)) 7916 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("]") reduce 527) (("^") reduce 527) (("|") reduce 527)) 7918 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("]") reduce 490) (("^") reduce 490) (("|") reduce 490)) 7919 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("]") reduce 289) (("^") reduce 289) (("|") reduce 289)) 7923 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("]") reduce 290) (("^") reduce 290) (("|") reduce 290)) 7924 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 7929 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("]") reduce 409) (("^") reduce 409) (("|") reduce 409)) 7931 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("]") reduce 407) (("^") reduce 407) (("|") reduce 407)) 7941 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 403) (("^") shift) (("|") shift)) 7942 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 405) (("^") shift) (("|") shift)) 7943 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 404) (("^") shift) (("|") shift)) 7944 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 397) (("^") shift) (("|") shift)) 7945 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 394) (("^") shift) (("|") shift)) 7946 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 402) (("^") shift) (("|") shift)) 7947 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 396) (("^") shift) (("|") shift)) 7948 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 400) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 400) (("^") shift) (("|") shift)) 7949 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 395) (("^") shift) (("|") shift)) 7950 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 398) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 398) (("^") shift) (("|") shift)) 7951 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 399) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 399) (("^") shift) (("|") shift)) 7952 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 406) (("^") shift) (("|") shift)) 7953 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 401) (("^") shift) (("|") shift)) 7955 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 391) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 391) (("^") shift) (("|") shift)) 7958 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("]") reduce 551) (("^") reduce 551) (("|") reduce 551)) 7961 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("]") reduce 547) (("^") reduce 547) (("|") reduce 547)) 7963 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("]") reduce 546) (("^") reduce 546) (("|") reduce 546)) 7965 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("]") reduce 542) (("^") reduce 542) (("|") reduce 542)) 7966 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("]") reduce 548) (("^") reduce 548) (("|") reduce 548)) 7967 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("]") reduce 392) (("^") reduce 392) (("|") reduce 392)) 7971 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("]") reduce 555) (("^") reduce 555) (("|") reduce 555)) 7973 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("]") reduce 490) (("^") reduce 490) (("|") reduce 490)) 7974 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("]") reduce 289) (("^") reduce 289) (("|") reduce 289)) 7978 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("]") reduce 290) (("^") reduce 290) (("|") reduce 290)) 7979 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 7980 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("]") reduce 487) (("^") reduce 487) (("|") reduce 487)) 7983 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("]") reduce 554) (("^") reduce 554) (("|") reduce 554)) 7984 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("]") reduce 489) (("^") reduce 489) (("|") reduce 489)) 7985 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("]") reduce 491) (("^") reduce 491) (("|") reduce 491)) 7990 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("]") reduce 543) (("^") reduce 543) (("|") reduce 543)) 7991 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 7992 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("]") reduce 549) (("^") reduce 549) (("|") reduce 549)) 7993 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 7996 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 7997 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("]") reduce 544) (("^") reduce 544) (("|") reduce 544)) 7998 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("]") reduce 550) (("^") reduce 550) (("|") reduce 550)) 7999 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("]") reduce 545) (("^") reduce 545) (("|") reduce 545)) 8000 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("]") reduce 488) (("^") reduce 488) (("|") reduce 488)) 8001 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("]") reduce 553) (("^") reduce 553) (("|") reduce 553)) 8004 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("]") reduce 552) (("^") reduce 552) (("|") reduce 552)) 8006 (((",") reduce 571) (("]") reduce 571)) 8007 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("]") reduce 487) (("^") reduce 487) (("|") reduce 487)) 8018 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("]") reduce 477) (("^") reduce 477) (("|") reduce 477)) 8019 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("]") reduce 472) (("^") reduce 472) (("|") reduce 472)) 8092 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 477) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 477) (("^") shift) (("|") shift)) 8406 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("]") reduce 473) (("^") reduce 473) (("|") reduce 473)) 8409 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("]") reduce 524) (("^") reduce 524) (("|") reduce 524)) 8411 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("]") reduce 554) (("^") reduce 554) (("|") reduce 554)) 8414 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("]") reduce 526) (("^") reduce 526) (("|") reduce 526)) 8415 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("]") reduce 489) (("^") reduce 489) (("|") reduce 489)) 8416 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("]") reduce 491) (("^") reduce 491) (("|") reduce 491)) 8418 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("]") reduce 468) (("^") reduce 468) (("|") reduce 468)) 8419 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("]") reduce 470) (("^") reduce 470) (("|") reduce 470)) 8424 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("]") reduce 543) (("^") reduce 543) (("|") reduce 543)) 8425 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 8426 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("]") reduce 549) (("^") reduce 549) (("|") reduce 549)) 8427 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 8430 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 8431 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("]") reduce 544) (("^") reduce 544) (("|") reduce 544)) 8432 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("]") reduce 550) (("^") reduce 550) (("|") reduce 550)) 8433 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("]") reduce 545) (("^") reduce 545) (("|") reduce 545)) 8435 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 8437 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) ((",") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_DOUBLE_ARROW) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("]") reduce 503) (("^") reduce 503) (("|") reduce 503)) 8439 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 390) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 390) (("^") shift) (("|") shift)) 8440 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 465) (("^") shift) (("|") shift)) 8441 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 463) (("^") shift) (("|") shift)) 8443 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 464) (("^") shift) (("|") shift)) 8444 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("]") reduce 456) (("^") reduce 456) (("|") reduce 456)) 8445 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 466) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 466) (("^") shift) (("|") shift)) 8446 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) ((",") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("]") reduce 452) (("^") reduce 452) (("|") reduce 452)) 8447 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("]") reduce 469) (("^") reduce 469) (("|") reduce 469)) 8449 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("]") reduce 509) (("^") reduce 509) (("|") reduce 509)) 8451 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("]") reduce 510) (("^") reduce 510) (("|") reduce 510)) 8452 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("]") reduce 508) (("^") reduce 508) (("|") reduce 508)) 8453 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 602) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 602) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 602) (("^") shift) (("|") shift)) 8454 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 601) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 601) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 601) (("^") shift) (("|") shift)) 8455 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("]") reduce 488) (("^") reduce 488) (("|") reduce 488)) 8456 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 461) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 461) (("^") shift) (("|") shift)) 8457 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("]") reduce 454) (("^") reduce 454) (("|") reduce 454)) 8459 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 8460 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 8461 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 8462 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("]") reduce 492) (("^") reduce 492) (("|") reduce 492)) 8463 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 8464 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) ((",") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("]") reduce 385) (("^") reduce 385) (("|") reduce 385)) 8467 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("]") reduce 494) (("^") reduce 494) (("|") reduce 494)) 8468 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) ((",") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("]") reduce 387) (("^") reduce 387) (("|") reduce 387)) 8469 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("]") reduce 493) (("^") reduce 493) (("|") reduce 493)) 8470 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("]") reduce 495) (("^") reduce 495) (("|") reduce 495)) 8471 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("]") reduce 556) (("^") reduce 556) (("|") reduce 556)) 8477 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("]") reduce 557) (("^") reduce 557) (("|") reduce 557)) 8478 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 8479 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("]") reduce 558) (("^") reduce 558) (("|") reduce 558)) 8480 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 8481 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) ((",") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("]") reduce 384) (("^") reduce 384) (("|") reduce 384)) 8483 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 8484 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) ((",") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("]") reduce 386) (("^") reduce 386) (("|") reduce 386)) 8492 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) ((",") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("]") reduce 383) (("^") reduce 383) (("|") reduce 383)) 8494 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("]") reduce 496) (("^") reduce 496) (("|") reduce 496)) 8500 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) ((",") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_DOUBLE_ARROW) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("]") reduce 231) (("^") reduce 231) (("|") reduce 231)) 8509 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) ((",") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_DOUBLE_ARROW) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("]") reduce 596) (("^") reduce 596) (("|") reduce 596)) 8510 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) ((",") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("]") reduce 450) (("^") reduce 450) (("|") reduce 450)) 8511 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 599) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 599) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 599) (("^") shift) (("|") shift)) 8512 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 598) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 598) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 598) (("^") shift) (("|") shift)) 8513 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("]") reduce 408) (("^") reduce 408) (("|") reduce 408)) 8515 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) ((",") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_DOUBLE_ARROW) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("]") reduce 501) (("^") reduce 501) (("|") reduce 501)) 8516 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("]") reduce 457) (("^") reduce 457) (("|") reduce 457)) 8517 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("]") reduce 289) (("^") reduce 289) (("|") reduce 289)) 8521 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("]") reduce 290) (("^") reduce 290) (("|") reduce 290)) 8522 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 8525 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) ((",") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_DOUBLE_ARROW) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("]") reduce 600) (("^") reduce 600) (("|") reduce 600)) 8528 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) ((",") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_DOUBLE_ARROW) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("]") reduce 597) (("^") reduce 597) (("|") reduce 597)) 8529 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) ((",") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("]") reduce 451) (("^") reduce 451) (("|") reduce 451)) 8530 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("]") reduce 410) (("^") reduce 410) (("|") reduce 410)) 8531 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("]") reduce 393) (("^") reduce 393) (("|") reduce 393)) 8532 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("]") reduce 455) (("^") reduce 455) (("|") reduce 455)) 8533 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) ((",") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("]") reduce 453) (("^") reduce 453) (("|") reduce 453)) 8536 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) ((",") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_DOUBLE_ARROW) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("]") reduce 502) (("^") reduce 502) (("|") reduce 502)) 8537 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 8538 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) ((",") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("]") reduce 430) (("^") reduce 430) (("|") reduce 430)) 8539 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) ((",") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("]") reduce 429) (("^") reduce 429) (("|") reduce 429)) 8541 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) ((",") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("]") reduce 443) (("^") reduce 443) (("|") reduce 443)) 8542 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("]") reduce 553) (("^") reduce 553) (("|") reduce 553)) 8545 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("]") reduce 552) (("^") reduce 552) (("|") reduce 552)) 8547 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("|") reduce 505)) 8548 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) ((",") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("]") reduce 431) (("^") reduce 431) (("|") reduce 431)) 8560 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("]") reduce 523) (("^") reduce 523) (("|") reduce 523)) 8562 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("]") reduce 504) (("^") reduce 504) (("|") reduce 504)) 8564 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("|") reduce 517)) 8565 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("]") reduce 507) (("^") reduce 507) (("|") reduce 507)) 8569 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) ((",") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("]") reduce 500) (("^") reduce 500) (("|") reduce 500)) 8570 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("|") reduce 516)) 8571 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("|") reduce 520)) 8577 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("|") reduce 515)) 8579 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("]") reduce 506) (("^") reduce 506) (("|") reduce 506)) 8581 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("|") reduce 519)) 8582 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 8583 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 8584 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 8586 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("|") reduce 522)) 8589 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("|") reduce 521)) 8595 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 8598 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("]") reduce 518) (("^") reduce 518) (("|") reduce 518)) 8600 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("]") reduce 551) (("^") reduce 551) (("|") reduce 551)) 8601 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) ((",") reduce 462) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 462) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 8609 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("]") reduce 547) (("^") reduce 547) (("|") reduce 547)) 8610 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("]") reduce 513) (("^") reduce 513) (("|") reduce 513)) 8612 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("]") reduce 512) (("^") reduce 512) (("|") reduce 512)) 8613 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("]") reduce 511) (("^") reduce 511) (("|") reduce 511)) 8614 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 573) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 573) (("^") shift) (("|") shift)) 8617 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("]") reduce 546) (("^") reduce 546) (("|") reduce 546)) 8618 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("]") reduce 467) (("^") reduce 467) (("|") reduce 467)) 8619 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) ((",") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("]") reduce 449) (("^") reduce 449) (("|") reduce 449)) 8620 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("]") reduce 471) (("^") reduce 471) (("|") reduce 471)) 8621 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("]") reduce 514) (("^") reduce 514) (("|") reduce 514)) 8622 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("]") reduce 444) (("^") reduce 444) (("|") reduce 444)) 8623 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("]") reduce 445) (("^") reduce 445) (("|") reduce 445)) 8624 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 8625 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("]") reduce 542) (("^") reduce 542) (("|") reduce 542)) 8626 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("]") reduce 548) (("^") reduce 548) (("|") reduce 548)) 8627 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) ((",") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("]") reduce 388) (("^") reduce 388) (("|") reduce 388)) 8631 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) ((",") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("]") reduce 432) (("^") reduce 432) (("|") reduce 432)) 8662 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 416) (("^") shift) (("|") reduce 416)) 8663 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 419) (("^") reduce 419) (("|") reduce 419)) 8664 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("]") reduce 428) (("^") reduce 428) (("|") reduce 428)) 8665 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 441) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 441) (("^") reduce 441) (("|") reduce 441)) 8666 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("]") reduce 427) (("^") reduce 427) (("|") reduce 427)) 8667 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("]") reduce 424) (("^") reduce 424) (("|") reduce 424)) 8668 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 415) (("^") shift) (("|") shift)) 8669 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 413) (("^") shift) (("|") shift)) 8670 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 414) (("^") shift) (("|") shift)) 8671 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 438) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("]") reduce 438) (("^") reduce 438) (("|") reduce 438)) 8672 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 434) (("^") reduce 434) (("|") reduce 434)) 8673 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 436) (("^") reduce 436) (("|") reduce 436)) 8674 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 433) (("^") reduce 433) (("|") reduce 433)) 8675 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 440) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("]") reduce 440) (("^") reduce 440) (("|") reduce 440)) 8676 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 435) (("^") reduce 435) (("|") reduce 435)) 8679 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 8680 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 8681 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 8682 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("]") reduce 492) (("^") reduce 492) (("|") reduce 492)) 8683 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 8684 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("]") reduce 551) (("^") reduce 551) (("|") reduce 551)) 8685 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("]") reduce 494) (("^") reduce 494) (("|") reduce 494)) 8686 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("]") reduce 442) (("^") reduce 442) (("|") reduce 442)) 8687 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("]") reduce 493) (("^") reduce 493) (("|") reduce 493)) 8688 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("]") reduce 495) (("^") reduce 495) (("|") reduce 495)) 8689 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("]") reduce 556) (("^") reduce 556) (("|") reduce 556)) 8695 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("]") reduce 557) (("^") reduce 557) (("|") reduce 557)) 8696 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 8697 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 8698 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("]") reduce 558) (("^") reduce 558) (("|") reduce 558)) 8699 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 8702 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 8703 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 8705 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 8707 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("]") reduce 496) (("^") reduce 496) (("|") reduce 496)) 8708 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("]") reduce 553) (("^") reduce 553) (("|") reduce 553)) 8711 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("]") reduce 552) (("^") reduce 552) (("|") reduce 552)) 8712 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 448) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 448) (("^") shift) (("|") shift)) 8713 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 411) (("^") shift) (("|") shift)) 8714 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 412) (("^") shift) (("|") shift)) 8715 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 417) (("^") reduce 417) (("|") reduce 417)) 8716 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 418) (("^") reduce 418) (("|") reduce 418)) 8720 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 446) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 446) (("^") shift) (("|") shift)) 8721 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 447) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 447) (("^") shift) (("|") shift)) 8722 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 439) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("]") reduce 439) (("^") reduce 439) (("|") reduce 439)) 8723 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 437) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("]") reduce 437) (("^") reduce 437) (("|") reduce 437)) 8724 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("]") reduce 425) (("^") reduce 425) (("|") reduce 425)) 8725 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("]") reduce 420) (("^") reduce 420) (("|") reduce 420)) 8726 ((("%") shift) (("*") shift) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("]") reduce 422) (("^") reduce 422) (("|") reduce 422)) 8727 ((("%") shift) (("*") shift) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("]") reduce 421) (("^") reduce 421) (("|") reduce 421)) 8728 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("]") reduce 423) (("^") reduce 423) (("|") reduce 423)) 8729 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("]") reduce 426) (("^") reduce 426) (("|") reduce 426)) 8734 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 389) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 389) (("^") shift) (("|") shift)) 8736 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("]") reduce 74) (("^") reduce 74) (("|") reduce 74)) 8737 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("]") reduce 52) (("^") reduce 52) (("|") reduce 52)) 8738 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("]") reduce 27) (("^") reduce 27) (("|") reduce 27)) 8739 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("]") reduce 51) (("^") reduce 51) (("|") reduce 51)) 8740 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("]") reduce 53) (("^") reduce 53) (("|") reduce 53)) 8741 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("]") reduce 49) (("^") reduce 49) (("|") reduce 49)) 8742 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("]") reduce 29) (("^") reduce 29) (("|") reduce 29)) 8743 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("]") reduce 59) (("^") reduce 59) (("|") reduce 59)) 8744 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("]") reduce 60) (("^") reduce 60) (("|") reduce 60)) 8745 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("]") reduce 11) (("^") reduce 11) (("|") reduce 11)) 8746 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("]") reduce 42) (("^") reduce 42) (("|") reduce 42)) 8747 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("]") reduce 39) (("^") reduce 39) (("|") reduce 39)) 8748 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("]") reduce 25) (("^") reduce 25) (("|") reduce 25)) 8749 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("]") reduce 50) (("^") reduce 50) (("|") reduce 50)) 8750 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("]") reduce 66) (("^") reduce 66) (("|") reduce 66)) 8751 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("]") reduce 18) (("^") reduce 18) (("|") reduce 18)) 8752 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("]") reduce 17) (("^") reduce 17) (("|") reduce 17)) 8753 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("]") reduce 15) (("^") reduce 15) (("|") reduce 15)) 8754 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("]") reduce 14) (("^") reduce 14) (("|") reduce 14)) 8755 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("]") reduce 38) (("^") reduce 38) (("|") reduce 38)) 8756 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("]") reduce 26) (("^") reduce 26) (("|") reduce 26)) 8757 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("]") reduce 22) (("^") reduce 22) (("|") reduce 22)) 8758 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("]") reduce 24) (("^") reduce 24) (("|") reduce 24)) 8759 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("]") reduce 16) (("^") reduce 16) (("|") reduce 16)) 8760 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("]") reduce 48) (("^") reduce 48) (("|") reduce 48)) 8761 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("]") reduce 20) (("^") reduce 20) (("|") reduce 20)) 8762 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("]") reduce 70) (("^") reduce 70) (("|") reduce 70)) 8763 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("]") reduce 3) (("^") reduce 3) (("|") reduce 3)) 8764 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("]") reduce 12) (("^") reduce 12) (("|") reduce 12)) 8765 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("]") reduce 54) (("^") reduce 54) (("|") reduce 54)) 8766 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("]") reduce 65) (("^") reduce 65) (("|") reduce 65)) 8767 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("]") reduce 75) (("^") reduce 75) (("|") reduce 75)) 8768 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("]") reduce 30) (("^") reduce 30) (("|") reduce 30)) 8769 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("]") reduce 68) (("^") reduce 68) (("|") reduce 68)) 8770 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("]") reduce 21) (("^") reduce 21) (("|") reduce 21)) 8771 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("]") reduce 23) (("^") reduce 23) (("|") reduce 23)) 8772 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("]") reduce 41) (("^") reduce 41) (("|") reduce 41)) 8773 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("]") reduce 62) (("^") reduce 62) (("|") reduce 62)) 8774 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("]") reduce 34) (("^") reduce 34) (("|") reduce 34)) 8775 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("]") reduce 40) (("^") reduce 40) (("|") reduce 40)) 8776 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("]") reduce 13) (("^") reduce 13) (("|") reduce 13)) 8777 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("]") reduce 55) (("^") reduce 55) (("|") reduce 55)) 8778 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("]") reduce 1) (("^") reduce 1) (("|") reduce 1)) 8779 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("]") reduce 2) (("^") reduce 2) (("|") reduce 2)) 8780 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("]") reduce 9) (("^") reduce 9) (("|") reduce 9)) 8781 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("]") reduce 33) (("^") reduce 33) (("|") reduce 33)) 8782 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("]") reduce 58) (("^") reduce 58) (("|") reduce 58)) 8783 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("]") reduce 37) (("^") reduce 37) (("|") reduce 37)) 8784 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("]") reduce 64) (("^") reduce 64) (("|") reduce 64)) 8785 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("]") reduce 46) (("^") reduce 46) (("|") reduce 46)) 8786 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("]") reduce 8) (("^") reduce 8) (("|") reduce 8)) 8787 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("]") reduce 6) (("^") reduce 6) (("|") reduce 6)) 8788 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("]") reduce 7) (("^") reduce 7) (("|") reduce 7)) 8789 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("]") reduce 69) (("^") reduce 69) (("|") reduce 69)) 8790 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("]") reduce 63) (("^") reduce 63) (("|") reduce 63)) 8791 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("]") reduce 56) (("^") reduce 56) (("|") reduce 56)) 8792 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("]") reduce 10) (("^") reduce 10) (("|") reduce 10)) 8793 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("]") reduce 67) (("^") reduce 67) (("|") reduce 67)) 8794 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("]") reduce 44) (("^") reduce 44) (("|") reduce 44)) 8795 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("]") reduce 76) (("^") reduce 76) (("|") reduce 76)) 8796 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("]") reduce 71) (("^") reduce 71) (("|") reduce 71)) 8797 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("]") reduce 77) (("^") reduce 77) (("|") reduce 77)) 8798 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("]") reduce 78) (("^") reduce 78) (("|") reduce 78)) 8799 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("]") reduce 79) (("^") reduce 79) (("|") reduce 79)) 8800 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("]") reduce 4) (("^") reduce 4) (("|") reduce 4)) 8801 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("]") reduce 5) (("^") reduce 5) (("|") reduce 5)) 8802 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("]") reduce 43) (("^") reduce 43) (("|") reduce 43)) 8803 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("]") reduce 73) (("^") reduce 73) (("|") reduce 73)) 8804 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("]") reduce 82) (("^") reduce 82) (("|") reduce 82)) 8805 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("]") reduce 47) (("^") reduce 47) (("|") reduce 47)) 8806 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("]") reduce 31) (("^") reduce 31) (("|") reduce 31)) 8807 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("]") reduce 57) (("^") reduce 57) (("|") reduce 57)) 8808 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("]") reduce 61) (("^") reduce 61) (("|") reduce 61)) 8809 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("]") reduce 28) (("^") reduce 28) (("|") reduce 28)) 8810 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("]") reduce 36) (("^") reduce 36) (("|") reduce 36)) 8811 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("]") reduce 32) (("^") reduce 32) (("|") reduce 32)) 8812 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("]") reduce 35) (("^") reduce 35) (("|") reduce 35)) 8813 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("]") reduce 19) (("^") reduce 19) (("|") reduce 19)) 8814 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("]") reduce 45) (("^") reduce 45) (("|") reduce 45)) 8815 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("]") reduce 525) (("^") reduce 525) (("|") reduce 525)) 8817 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("]") reduce 72) (("^") reduce 72) (("|") reduce 72)) 8818 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("]") reduce 83) (("^") reduce 83) (("|") reduce 83)) 8819 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("]") reduce 555) (("^") reduce 555) (("|") reduce 555)) 8822 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("]") reduce 527) (("^") reduce 527) (("|") reduce 527)) 8824 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("]") reduce 490) (("^") reduce 490) (("|") reduce 490)) 8825 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("]") reduce 289) (("^") reduce 289) (("|") reduce 289)) 8829 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("]") reduce 290) (("^") reduce 290) (("|") reduce 290)) 8830 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 8835 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("]") reduce 409) (("^") reduce 409) (("|") reduce 409)) 8837 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("]") reduce 407) (("^") reduce 407) (("|") reduce 407)) 8847 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 403) (("^") shift) (("|") shift)) 8848 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 405) (("^") shift) (("|") shift)) 8849 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 404) (("^") shift) (("|") shift)) 8850 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 397) (("^") shift) (("|") shift)) 8851 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 394) (("^") shift) (("|") shift)) 8852 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 402) (("^") shift) (("|") shift)) 8853 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 396) (("^") shift) (("|") shift)) 8854 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 400) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 400) (("^") shift) (("|") shift)) 8855 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 395) (("^") shift) (("|") shift)) 8856 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 398) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 398) (("^") shift) (("|") shift)) 8857 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 399) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 399) (("^") shift) (("|") shift)) 8858 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 406) (("^") shift) (("|") shift)) 8859 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 401) (("^") shift) (("|") shift)) 8861 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 391) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 391) (("^") shift) (("|") shift)) 8864 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("]") reduce 551) (("^") reduce 551) (("|") reduce 551)) 8867 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("]") reduce 547) (("^") reduce 547) (("|") reduce 547)) 8869 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("]") reduce 546) (("^") reduce 546) (("|") reduce 546)) 8871 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("]") reduce 542) (("^") reduce 542) (("|") reduce 542)) 8872 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("]") reduce 548) (("^") reduce 548) (("|") reduce 548)) 8873 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("]") reduce 392) (("^") reduce 392) (("|") reduce 392)) 8877 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("]") reduce 555) (("^") reduce 555) (("|") reduce 555)) 8879 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("]") reduce 490) (("^") reduce 490) (("|") reduce 490)) 8880 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("]") reduce 289) (("^") reduce 289) (("|") reduce 289)) 8884 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("]") reduce 290) (("^") reduce 290) (("|") reduce 290)) 8885 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 8886 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("]") reduce 487) (("^") reduce 487) (("|") reduce 487)) 8889 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("]") reduce 554) (("^") reduce 554) (("|") reduce 554)) 8890 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("]") reduce 489) (("^") reduce 489) (("|") reduce 489)) 8891 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("]") reduce 491) (("^") reduce 491) (("|") reduce 491)) 8896 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("]") reduce 543) (("^") reduce 543) (("|") reduce 543)) 8897 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 8898 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("]") reduce 549) (("^") reduce 549) (("|") reduce 549)) 8899 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 8902 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 8903 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("]") reduce 544) (("^") reduce 544) (("|") reduce 544)) 8904 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("]") reduce 550) (("^") reduce 550) (("|") reduce 550)) 8905 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("]") reduce 545) (("^") reduce 545) (("|") reduce 545)) 8906 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("]") reduce 488) (("^") reduce 488) (("|") reduce 488)) 8907 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("]") reduce 553) (("^") reduce 553) (("|") reduce 553)) 8910 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("]") reduce 552) (("^") reduce 552) (("|") reduce 552)) 8911 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("]") reduce 487) (("^") reduce 487) (("|") reduce 487)) 8922 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("]") reduce 477) (("^") reduce 477) (("|") reduce 477)) 8923 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("]") reduce 472) (("^") reduce 472) (("|") reduce 472)) 8996 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 477) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 477) (("^") shift) (("|") shift)) 9310 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("]") reduce 473) (("^") reduce 473) (("|") reduce 473)) 9313 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("]") reduce 524) (("^") reduce 524) (("|") reduce 524)) 9315 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("]") reduce 554) (("^") reduce 554) (("|") reduce 554)) 9318 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("]") reduce 526) (("^") reduce 526) (("|") reduce 526)) 9319 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("]") reduce 489) (("^") reduce 489) (("|") reduce 489)) 9320 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("]") reduce 491) (("^") reduce 491) (("|") reduce 491)) 9322 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("]") reduce 468) (("^") reduce 468) (("|") reduce 468)) 9323 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("]") reduce 470) (("^") reduce 470) (("|") reduce 470)) 9328 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("]") reduce 543) (("^") reduce 543) (("|") reduce 543)) 9329 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 9330 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("]") reduce 549) (("^") reduce 549) (("|") reduce 549)) 9331 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 9334 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 9335 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("]") reduce 544) (("^") reduce 544) (("|") reduce 544)) 9336 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("]") reduce 550) (("^") reduce 550) (("|") reduce 550)) 9337 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("]") reduce 545) (("^") reduce 545) (("|") reduce 545)) 9339 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 9341 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) ((",") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("]") reduce 503) (("^") reduce 503) (("|") reduce 503)) 9343 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 390) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 390) (("^") shift) (("|") shift)) 9344 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 465) (("^") shift) (("|") shift)) 9347 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 464) (("^") shift) (("|") shift)) 9348 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("]") reduce 456) (("^") reduce 456) (("|") reduce 456)) 9349 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 466) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 466) (("^") shift) (("|") shift)) 9350 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) ((",") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("]") reduce 452) (("^") reduce 452) (("|") reduce 452)) 9351 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("]") reduce 469) (("^") reduce 469) (("|") reduce 469)) 9353 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("]") reduce 509) (("^") reduce 509) (("|") reduce 509)) 9355 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("]") reduce 510) (("^") reduce 510) (("|") reduce 510)) 9356 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("]") reduce 508) (("^") reduce 508) (("|") reduce 508)) 9357 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 602) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 602) (("^") shift) (("|") shift)) 9358 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 601) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 601) (("^") shift) (("|") shift)) 9359 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("]") reduce 488) (("^") reduce 488) (("|") reduce 488)) 9360 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 461) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 461) (("^") shift) (("|") shift)) 9361 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("]") reduce 454) (("^") reduce 454) (("|") reduce 454)) 9363 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 9364 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 9365 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 9366 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("]") reduce 492) (("^") reduce 492) (("|") reduce 492)) 9367 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 9368 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) ((",") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("]") reduce 385) (("^") reduce 385) (("|") reduce 385)) 9371 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("]") reduce 494) (("^") reduce 494) (("|") reduce 494)) 9372 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) ((",") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("]") reduce 387) (("^") reduce 387) (("|") reduce 387)) 9373 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("]") reduce 493) (("^") reduce 493) (("|") reduce 493)) 9374 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("]") reduce 495) (("^") reduce 495) (("|") reduce 495)) 9375 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("]") reduce 556) (("^") reduce 556) (("|") reduce 556)) 9381 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("]") reduce 557) (("^") reduce 557) (("|") reduce 557)) 9382 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 9383 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("]") reduce 558) (("^") reduce 558) (("|") reduce 558)) 9384 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 9385 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) ((",") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("]") reduce 384) (("^") reduce 384) (("|") reduce 384)) 9387 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 9388 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) ((",") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("]") reduce 386) (("^") reduce 386) (("|") reduce 386)) 9396 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) ((",") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("]") reduce 383) (("^") reduce 383) (("|") reduce 383)) 9398 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("]") reduce 496) (("^") reduce 496) (("|") reduce 496)) 9404 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) ((",") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("]") reduce 231) (("^") reduce 231) (("|") reduce 231)) 9413 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) ((",") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("]") reduce 596) (("^") reduce 596) (("|") reduce 596)) 9414 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) ((",") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("]") reduce 450) (("^") reduce 450) (("|") reduce 450)) 9415 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 599) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 599) (("^") shift) (("|") shift)) 9416 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 598) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 598) (("^") shift) (("|") shift)) 9417 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("]") reduce 408) (("^") reduce 408) (("|") reduce 408)) 9419 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) ((",") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("]") reduce 501) (("^") reduce 501) (("|") reduce 501)) 9420 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("]") reduce 457) (("^") reduce 457) (("|") reduce 457)) 9421 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("]") reduce 289) (("^") reduce 289) (("|") reduce 289)) 9425 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("]") reduce 290) (("^") reduce 290) (("|") reduce 290)) 9426 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 9429 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) ((",") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("]") reduce 600) (("^") reduce 600) (("|") reduce 600)) 9432 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) ((",") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("]") reduce 597) (("^") reduce 597) (("|") reduce 597)) 9433 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) ((",") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("]") reduce 451) (("^") reduce 451) (("|") reduce 451)) 9434 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("]") reduce 410) (("^") reduce 410) (("|") reduce 410)) 9435 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("]") reduce 393) (("^") reduce 393) (("|") reduce 393)) 9436 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("]") reduce 455) (("^") reduce 455) (("|") reduce 455)) 9437 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) ((",") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("]") reduce 453) (("^") reduce 453) (("|") reduce 453)) 9440 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) ((",") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("]") reduce 502) (("^") reduce 502) (("|") reduce 502)) 9441 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 9442 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) ((",") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("]") reduce 430) (("^") reduce 430) (("|") reduce 430)) 9443 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) ((",") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("]") reduce 429) (("^") reduce 429) (("|") reduce 429)) 9445 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) ((",") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("]") reduce 443) (("^") reduce 443) (("|") reduce 443)) 9446 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("]") reduce 553) (("^") reduce 553) (("|") reduce 553)) 9449 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("]") reduce 552) (("^") reduce 552) (("|") reduce 552)) 9451 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("|") reduce 505)) 9452 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) ((",") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("]") reduce 431) (("^") reduce 431) (("|") reduce 431)) 9481 ((("(") reduce 551) ((",") reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) (("[") reduce 551) (("]") reduce 551)) 9484 ((("(") reduce 538) ((",") reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) (("[") reduce 547) (("]") reduce 547)) 9486 ((("(") reduce 546) ((",") reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) (("[") reduce 546) (("]") reduce 546)) 9488 ((("(") reduce 542) ((",") reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) (("[") reduce 542) (("]") reduce 542)) 9489 (((",") reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) (("[") reduce 548) (("]") reduce 548)) 9490 (((",") reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) (("[") reduce 531) (("]") reduce 575)) 9494 ((("(") reduce 564) ((",") reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) (("[") reduce 555) (("]") reduce 555)) 9496 ((("(") reduce 490) ((",") reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) (("[") reduce 490) (("]") reduce 490)) 9497 ((("(") reduce 289) ((",") reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) (("[") reduce 289) (("]") reduce 289)) 9501 ((("(") reduce 290) ((",") reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) (("[") reduce 290) (("]") reduce 290)) 9502 ((("(") reduce 291) ((",") reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) (("[") reduce 291) (("]") reduce 291)) 9503 ((("(") reduce 487) ((",") reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) (("[") reduce 487) (("]") reduce 487)) 9506 ((("(") reduce 564) ((",") reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) (("[") reduce 554) (("]") reduce 554)) 9507 ((("(") reduce 489) ((",") reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) (("[") reduce 489) (("]") reduce 489)) 9508 ((("(") reduce 491) ((",") reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) (("[") reduce 491) (("]") reduce 491)) 9513 ((("(") reduce 543) ((",") reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) (("[") reduce 543) (("]") reduce 543)) 9514 ((("(") reduce 565) ((",") reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) (("[") reduce 565) (("]") reduce 565)) 9515 ((("(") shift) ((",") reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) (("[") reduce 549) (("]") reduce 549)) 9516 ((("(") reduce 567) ((",") reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) (("[") reduce 567) (("]") reduce 567)) 9519 ((("(") reduce 566) ((",") reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) (("[") reduce 566) (("]") reduce 566)) 9520 ((("(") reduce 544) ((",") reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) (("[") reduce 544) (("]") reduce 544)) 9521 ((("(") shift) ((",") reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) (("[") reduce 550) (("]") reduce 550)) 9522 ((("(") reduce 545) ((",") reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) (("[") reduce 545) (("]") reduce 545)) 9523 ((("(") reduce 488) ((",") reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) (("[") reduce 488) (("]") reduce 488)) 9524 ((("(") reduce 553) ((",") reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) (("[") reduce 553) (("]") reduce 553)) 9527 ((("(") reduce 552) ((",") reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) (("[") reduce 552) (("]") reduce 552)) 9597 (((",") reduce 578) (("=") shift) (("]") reduce 578)) 9653 ((("(") reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) (("[") reduce 503)) 9668 (((",") reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) (("[") reduce 531) (("]") reduce 576)) 9738 (((",") reduce 579) (("=") shift) (("]") reduce 579)) 9762 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 577) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 577) (("^") shift) (("|") shift)) 9786 ((("(") reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94)) 9787 ((("(") reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93)) 9788 ((("(") reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95)) 9789 ((("(") reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492)) 9790 ((("(") reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92)) 9791 ((("(") reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) (("[") reduce 551)) 9792 ((("(") reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) (("[") reduce 385)) 9795 ((("(") reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift)) 9797 ((("(") reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493)) 9798 ((("(") reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) (("[") shift)) 9799 ((("(") reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) (("[") reduce 556)) 9805 ((("(") reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) (("[") reduce 557)) 9806 ((("(") reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) (("[") reduce 561)) 9807 ((("(") reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) (("[") reduce 565)) 9808 ((("(") reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) (("[") reduce 558)) 9809 ((("(") reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) (("[") reduce 567)) 9812 ((("(") reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) (("[") reduce 566)) 9813 ((("(") reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) (("[") reduce 559)) 9815 ((("(") reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) (("[") reduce 384)) 9816 ((("(") reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) (("[") reduce 289)) 9820 ((("(") reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) (("[") reduce 290)) 9821 ((("(") reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) (("[") reduce 291)) 9823 ((("(") reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) (("[") reduce 560)) 9824 ((("(") reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) (("[") reduce 386)) 9832 ((("(") reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) (("[") reduce 383)) 9834 ((("(") reduce 496)) 9835 ((("(") reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) (("[") reduce 553)) 9838 ((("(") reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) (("[") reduce 552)) 9841 ((("(") reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) (("[") reduce 502)) 9843 ((("(") reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) (("[") reduce 532)) 9849 ((("(") reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) (("[") reduce 505)) 10273 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((":") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 10276 ((("%") shift) ((")") reduce 295) (("*") shift) (("+") shift) ((",") reduce 295) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10343 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 10364 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((":") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 10366 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((":") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407)) 10376 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 403) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10377 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10378 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10379 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 397) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10380 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 394) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10381 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 402) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10382 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 396) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10383 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 400) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10384 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 395) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10385 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 398) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10386 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 399) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10387 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10388 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 401) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10390 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 391) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10393 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((":") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 10396 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((":") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 10398 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((":") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 10400 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((":") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 10401 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((":") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 10402 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((":") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392)) 10406 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((":") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 10408 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 10409 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((":") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 10413 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((":") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 10414 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 10415 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 10418 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((":") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 10419 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 10420 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((":") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 10425 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((":") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 10426 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 10427 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((":") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 10428 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 10431 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 10432 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((":") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 10433 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((":") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 10434 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((":") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 10435 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 10436 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((":") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 10439 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((":") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 10440 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 10451 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((":") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 10452 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((":") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 10525 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 477) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10839 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((":") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 10842 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524)) 10844 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((":") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 10847 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526)) 10848 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 10849 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((":") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 10851 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((":") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 10852 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((":") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 10857 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((":") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 10858 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 10859 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((":") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 10860 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 10863 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 10864 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((":") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 10865 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((":") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 10866 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((":") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 10868 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 10870 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) ((":") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503)) 10872 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 390) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10873 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 465) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10884 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 10886 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((":") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 10888 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 10889 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507)) 10893 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) ((":") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_DOUBLE_ARROW) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 10894 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 10895 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 10901 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 10903 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506)) 10905 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 10906 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 10907 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 10908 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 10910 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 10913 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 10919 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 10922 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 10924 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((":") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 10925 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) (("-") shift) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 462) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 10932 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((":") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 10933 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513)) 10935 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512)) 10936 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511)) 10937 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 463) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10940 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((":") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 10941 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((":") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 10942 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) ((":") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449)) 10943 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((":") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 10944 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((":") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514)) 10945 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((":") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444)) 10946 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((":") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 10947 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 10948 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((":") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 10949 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((":") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 10950 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) ((":") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 388) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388)) 10953 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) ((":") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432)) 10984 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 416) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416)) 10985 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 419) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419)) 10986 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 428) (("/") shift) ((":") reduce 428) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428)) 10987 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 441) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441)) 10988 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 427) (("/") shift) ((":") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 10989 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((":") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 10990 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10991 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10992 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10993 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 438) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438)) 10994 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 10995 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 436) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 10996 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 10997 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 440) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440)) 10998 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 11001 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 11002 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 11003 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 11004 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((":") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 11005 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 11006 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((":") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 11007 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((":") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 11008 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((":") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 11009 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((":") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 11010 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((":") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 11011 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((":") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 11017 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((":") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 11018 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 11019 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 11020 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((":") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 11021 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 11024 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 11025 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 11027 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 11029 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((":") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 11030 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((":") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 11033 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((":") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 11034 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 448) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11035 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 411) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11036 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 412) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11037 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 417) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417)) 11038 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 418) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418)) 11042 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 446) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11043 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 447) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11044 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 439) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439)) 11045 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 437) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437)) 11046 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((":") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 11047 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 420) (("/") shift) ((":") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420)) 11048 ((("%") shift) (("*") shift) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) ((":") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 11049 ((("%") shift) (("*") shift) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) ((":") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 11050 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((":") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 11051 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((":") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 11053 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((":") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 11054 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((":") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 11055 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((":") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 11056 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((":") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 11057 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((":") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 11058 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((":") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 11059 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((":") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 11060 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((":") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 11061 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((":") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 11062 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((":") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 11063 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((":") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 11064 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((":") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 11065 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((":") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 11066 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((":") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 11067 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((":") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 11068 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((":") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 11069 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((":") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 11070 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((":") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 11071 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((":") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 11072 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((":") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 11073 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((":") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 11074 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((":") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 11075 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((":") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 11076 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((":") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 11077 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((":") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 11078 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((":") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 11079 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((":") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 11080 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((":") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 11081 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((":") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 11082 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((":") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 11083 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((":") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 11084 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((":") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 11085 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((":") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 11086 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((":") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 11087 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((":") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 11088 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((":") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 11089 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((":") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 11090 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((":") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 11091 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((":") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 11092 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((":") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 11093 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((":") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 11094 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((":") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 11095 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((":") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 11096 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((":") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 11097 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((":") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 11098 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((":") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 11099 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((":") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 11100 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((":") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 11101 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((":") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 11102 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((":") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 11103 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((":") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 11104 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((":") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 11105 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((":") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 11106 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((":") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 11107 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((":") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 11108 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((":") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 11109 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((":") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 11110 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((":") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 11111 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((":") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 11112 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((":") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 11113 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((":") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 11114 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((":") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 11115 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((":") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 11116 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((":") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 11117 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((":") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 11118 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((":") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 11119 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((":") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 11120 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((":") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 11121 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((":") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 11122 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((":") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 11123 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((":") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 11124 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((":") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 11125 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((":") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 11126 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((":") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 11127 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((":") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 11128 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((":") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 11129 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((":") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 11130 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((":") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 11131 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((":") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 11132 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525)) 11134 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((":") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 11135 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((":") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 11136 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((":") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 11139 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((":") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527)) 11141 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 11142 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((":") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 11146 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((":") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 11147 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 11152 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((":") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 11154 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((":") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407)) 11164 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 403) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11165 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11166 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11167 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 397) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11168 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 394) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11169 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 402) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11170 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 396) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11171 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 400) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11172 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 395) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11173 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 398) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11174 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 399) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11175 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11176 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 401) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11178 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 391) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11181 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((":") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 11184 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((":") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 11186 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((":") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 11188 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((":") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 11189 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((":") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 11190 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((":") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392)) 11194 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((":") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 11196 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 11197 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((":") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 11201 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((":") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 11202 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 11203 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 11206 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((":") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 11207 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 11208 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((":") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 11213 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((":") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 11214 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 11215 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((":") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 11216 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 11219 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 11220 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((":") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 11221 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((":") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 11222 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((":") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 11223 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 11224 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((":") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 11227 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((":") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 11228 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 11239 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((":") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 11240 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((":") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 11250 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 477) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11251 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((":") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 11253 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 464) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11255 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524)) 11257 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((":") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 11260 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526)) 11261 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 11262 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((":") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 11264 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((":") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 11265 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((":") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 11270 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((":") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 11271 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 11272 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((":") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 11273 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 11276 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 11277 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((":") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 11278 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((":") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 11279 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((":") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 11281 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 11283 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) ((":") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_DOUBLE_ARROW) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503)) 11285 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 390) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11286 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 465) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11289 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 464) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11290 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((":") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456)) 11291 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 466) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11292 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) ((":") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452)) 11293 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((":") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 11295 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 11297 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 11298 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508)) 11299 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 602) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 602) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11300 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 601) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 601) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11301 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 11302 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 461) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11303 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((":") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454)) 11305 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 11306 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 11307 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 11308 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((":") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 11309 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 11310 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) ((":") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385)) 11313 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((":") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 11314 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) ((":") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387)) 11315 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((":") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 11316 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((":") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 11317 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((":") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 11323 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((":") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 11324 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 11325 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((":") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 11326 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 11327 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) ((":") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384)) 11329 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 11330 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) ((":") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386)) 11338 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) ((":") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383)) 11340 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((":") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 11346 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) ((":") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_DOUBLE_ARROW) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231)) 11351 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 389) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11355 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) ((":") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_DOUBLE_ARROW) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596)) 11356 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) ((":") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450)) 11357 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 599) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 599) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11358 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 598) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 598) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11359 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((":") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408)) 11361 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) ((":") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_DOUBLE_ARROW) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501)) 11362 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((":") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457)) 11363 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((":") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289)) 11367 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((":") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290)) 11368 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 11371 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) ((":") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_DOUBLE_ARROW) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600)) 11374 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) ((":") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_DOUBLE_ARROW) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597)) 11375 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) ((":") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451)) 11376 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((":") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410)) 11377 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((":") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 11378 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((":") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455)) 11379 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) ((":") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453)) 11382 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) ((":") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_DOUBLE_ARROW) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502)) 11383 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((":") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 11384 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) ((":") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430)) 11385 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) ((":") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429)) 11387 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) ((":") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443)) 11388 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((":") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 11391 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((":") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 11393 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 11394 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) ((":") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431)) 11395 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((":") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456)) 11396 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 466) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11397 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) ((":") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452)) 11398 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((":") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 11400 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 11402 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 11403 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508)) 11404 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 602) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11405 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 601) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11406 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 11407 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 461) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11408 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((":") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454)) 11410 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 11411 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 11412 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 11413 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((":") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 11414 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 11415 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) ((":") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385)) 11418 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((":") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 11419 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) ((":") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387)) 11420 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((":") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 11421 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((":") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 11422 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((":") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 11428 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((":") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 11429 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 11430 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((":") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 11431 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 11432 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) ((":") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384)) 11434 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 11435 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) ((":") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386)) 11443 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) ((":") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383)) 11445 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((":") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 11451 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) ((":") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231)) 11456 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 389) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11460 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) ((":") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596)) 11461 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) ((":") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450)) 11462 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 599) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11463 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 598) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11464 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((":") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408)) 11466 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) ((":") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501)) 11467 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((":") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457)) 11468 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((":") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289)) 11472 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((":") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290)) 11473 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 11476 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) ((":") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600)) 11479 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) ((":") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597)) 11480 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) ((":") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451)) 11481 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((":") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410)) 11482 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((":") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 11483 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((":") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455)) 11484 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) ((":") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453)) 11487 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) ((":") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502)) 11488 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((":") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 11489 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) ((":") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430)) 11490 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) ((":") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429)) 11492 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) ((":") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443)) 11493 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((":") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 11496 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((":") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 11498 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 11499 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) ((":") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431)) 11501 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 446)) 12150 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 447)) 12162 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439) (("}") reduce 439)) 12163 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437) (("}") reduce 437)) 12164 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425) (("}") reduce 425)) 12165 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420) (("}") reduce 420)) 12166 ((("%") shift) (("*") shift) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422) (("}") reduce 422)) 12167 ((("%") shift) (("*") shift) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421) (("}") reduce 421)) 12168 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423) (("}") reduce 423)) 12169 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426) (("}") reduce 426)) 12171 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74) (("}") reduce 74)) 12172 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52) (("}") reduce 52)) 12173 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27) (("}") reduce 27)) 12174 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51) (("}") reduce 51)) 12175 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53) (("}") reduce 53)) 12176 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49) (("}") reduce 49)) 12177 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29) (("}") reduce 29)) 12178 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59) (("}") reduce 59)) 12179 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60) (("}") reduce 60)) 12180 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11) (("}") reduce 11)) 12181 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42) (("}") reduce 42)) 12182 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39) (("}") reduce 39)) 12183 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25) (("}") reduce 25)) 12184 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50) (("}") reduce 50)) 12185 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66) (("}") reduce 66)) 12186 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18) (("}") reduce 18)) 12187 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17) (("}") reduce 17)) 12188 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15) (("}") reduce 15)) 12189 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14) (("}") reduce 14)) 12190 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38) (("}") reduce 38)) 12191 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26) (("}") reduce 26)) 12192 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22) (("}") reduce 22)) 12193 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24) (("}") reduce 24)) 12194 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16) (("}") reduce 16)) 12195 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48) (("}") reduce 48)) 12196 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20) (("}") reduce 20)) 12197 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70) (("}") reduce 70)) 12198 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3) (("}") reduce 3)) 12199 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12) (("}") reduce 12)) 12200 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54) (("}") reduce 54)) 12201 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65) (("}") reduce 65)) 12202 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75) (("}") reduce 75)) 12203 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30) (("}") reduce 30)) 12204 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68) (("}") reduce 68)) 12205 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21) (("}") reduce 21)) 12206 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23) (("}") reduce 23)) 12207 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41) (("}") reduce 41)) 12208 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62) (("}") reduce 62)) 12209 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34) (("}") reduce 34)) 12210 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40) (("}") reduce 40)) 12211 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13) (("}") reduce 13)) 12212 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55) (("}") reduce 55)) 12213 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1) (("}") reduce 1)) 12214 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2) (("}") reduce 2)) 12215 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9) (("}") reduce 9)) 12216 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33) (("}") reduce 33)) 12217 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58) (("}") reduce 58)) 12218 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37) (("}") reduce 37)) 12219 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64) (("}") reduce 64)) 12220 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46) (("}") reduce 46)) 12221 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8) (("}") reduce 8)) 12222 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6) (("}") reduce 6)) 12223 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7) (("}") reduce 7)) 12224 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69) (("}") reduce 69)) 12225 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63) (("}") reduce 63)) 12226 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56) (("}") reduce 56)) 12227 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10) (("}") reduce 10)) 12228 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67) (("}") reduce 67)) 12229 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44) (("}") reduce 44)) 12230 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76) (("}") reduce 76)) 12231 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71) (("}") reduce 71)) 12232 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77) (("}") reduce 77)) 12233 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78) (("}") reduce 78)) 12234 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79) (("}") reduce 79)) 12235 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4) (("}") reduce 4)) 12236 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5) (("}") reduce 5)) 12237 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43) (("}") reduce 43)) 12238 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73) (("}") reduce 73)) 12239 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82) (("}") reduce 82)) 12240 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47) (("}") reduce 47)) 12241 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31) (("}") reduce 31)) 12242 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57) (("}") reduce 57)) 12243 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61) (("}") reduce 61)) 12244 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28) (("}") reduce 28)) 12245 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36) (("}") reduce 36)) 12246 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32) (("}") reduce 32)) 12247 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35) (("}") reduce 35)) 12248 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19) (("}") reduce 19)) 12249 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45) (("}") reduce 45)) 12250 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525) (("}") reduce 525)) 12252 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72) (("}") reduce 72)) 12253 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83) (("}") reduce 83)) 12254 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555) (("}") reduce 555)) 12257 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527) (("}") reduce 527)) 12259 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490) (("}") reduce 490)) 12260 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289) (("}") reduce 289)) 12264 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290) (("}") reduce 290)) 12265 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 12270 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409) (("}") reduce 409)) 12272 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407) (("}") reduce 407)) 12282 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 403)) 12283 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 405)) 12284 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 404)) 12285 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 397)) 12286 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 394)) 12287 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 402)) 12288 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 396)) 12289 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 400)) 12290 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 395)) 12291 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 398)) 12292 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 399)) 12293 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 406)) 12294 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 401)) 12296 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 391)) 12299 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551) (("}") reduce 551)) 12302 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547) (("}") reduce 547)) 12304 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546) (("}") reduce 546)) 12306 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542) (("}") reduce 542)) 12307 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548) (("}") reduce 548)) 12308 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392) (("}") reduce 392)) 12312 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555) (("}") reduce 555)) 12314 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490) (("}") reduce 490)) 12315 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289) (("}") reduce 289)) 12319 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290) (("}") reduce 290)) 12320 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 12321 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487) (("}") reduce 487)) 12324 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554) (("}") reduce 554)) 12325 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489) (("}") reduce 489)) 12326 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491) (("}") reduce 491)) 12331 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543) (("}") reduce 543)) 12332 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 12333 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549) (("}") reduce 549)) 12334 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 12337 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 12338 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544) (("}") reduce 544)) 12339 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550) (("}") reduce 550)) 12340 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545) (("}") reduce 545)) 12341 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488) (("}") reduce 488)) 12342 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553) (("}") reduce 553)) 12345 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552) (("}") reduce 552)) 12346 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487) (("}") reduce 487)) 12357 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477) (("}") reduce 477)) 12358 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472) (("}") reduce 472)) 12431 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 477)) 12745 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473) (("}") reduce 473)) 12748 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524) (("}") reduce 524)) 12750 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554) (("}") reduce 554)) 12753 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526) (("}") reduce 526)) 12754 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489) (("}") reduce 489)) 12755 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491) (("}") reduce 491)) 12757 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468) (("}") reduce 468)) 12758 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470) (("}") reduce 470)) 12763 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543) (("}") reduce 543)) 12764 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 12765 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549) (("}") reduce 549)) 12766 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 12769 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 12770 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544) (("}") reduce 544)) 12771 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550) (("}") reduce 550)) 12772 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545) (("}") reduce 545)) 12774 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 12776 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503) (("}") reduce 503)) 12778 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 390)) 12779 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 465)) 12790 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523) (("}") reduce 523)) 12792 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504) (("}") reduce 504)) 12794 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517) (("}") reduce 517)) 12795 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507) (("}") reduce 507)) 12799 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_DOUBLE_ARROW) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500) (("}") reduce 500)) 12800 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516) (("}") reduce 516)) 12801 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520) (("}") reduce 520)) 12807 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515) (("}") reduce 515)) 12809 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506) (("}") reduce 506)) 12811 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519) (("}") reduce 519)) 12812 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 12813 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 12814 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 12816 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522) (("}") reduce 522)) 12819 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521) (("}") reduce 521)) 12825 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 12828 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518) (("}") reduce 518)) 12830 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551) (("}") reduce 551)) 12831 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 462) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("}") reduce 462) (("~") shift)) 12838 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547) (("}") reduce 547)) 12839 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513) (("}") reduce 513)) 12841 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512) (("}") reduce 512)) 12842 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511) (("}") reduce 511)) 12843 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 463)) 12846 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546) (("}") reduce 546)) 12847 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467) (("}") reduce 467)) 12848 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449) (("}") reduce 449)) 12849 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471) (("}") reduce 471)) 12850 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514) (("}") reduce 514)) 12851 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444) (("}") reduce 444)) 12852 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445) (("}") reduce 445)) 12853 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 12854 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542) (("}") reduce 542)) 12855 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548) (("}") reduce 548)) 12856 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 388) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388) (("}") reduce 388)) 12859 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432) (("}") reduce 432)) 12890 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416) (("}") reduce 416)) 12891 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419) (("}") reduce 419)) 12892 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428) (("}") reduce 428)) 12893 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441) (("}") reduce 441)) 12894 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427) (("}") reduce 427)) 12895 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424) (("}") reduce 424)) 12896 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 415)) 12897 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 413)) 12898 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 414)) 12899 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438) (("}") reduce 438)) 12900 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434) (("}") reduce 434)) 12901 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436) (("}") reduce 436)) 12902 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433) (("}") reduce 433)) 12903 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440) (("}") reduce 440)) 12904 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435) (("}") reduce 435)) 12907 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 12908 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 12909 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 12910 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492) (("}") reduce 492)) 12911 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 12912 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551) (("}") reduce 551)) 12913 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494) (("}") reduce 494)) 12914 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442) (("}") reduce 442)) 12915 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493) (("}") reduce 493)) 12916 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495) (("}") reduce 495)) 12917 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556) (("}") reduce 556)) 12923 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557) (("}") reduce 557)) 12924 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 12925 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 12926 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558) (("}") reduce 558)) 12927 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 12930 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 12931 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 12933 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 12935 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496) (("}") reduce 496)) 12936 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553) (("}") reduce 553)) 12939 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552) (("}") reduce 552)) 12940 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 448)) 12941 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 411)) 12942 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 412)) 12943 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417) (("}") reduce 417)) 12944 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418) (("}") reduce 418)) 12948 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 446)) 12949 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 447)) 12950 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439) (("}") reduce 439)) 12951 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437) (("}") reduce 437)) 12952 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425) (("}") reduce 425)) 12953 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420) (("}") reduce 420)) 12954 ((("%") shift) (("*") shift) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422) (("}") reduce 422)) 12955 ((("%") shift) (("*") shift) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421) (("}") reduce 421)) 12956 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423) (("}") reduce 423)) 12957 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426) (("}") reduce 426)) 12959 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74) (("}") reduce 74)) 12960 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52) (("}") reduce 52)) 12961 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27) (("}") reduce 27)) 12962 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51) (("}") reduce 51)) 12963 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53) (("}") reduce 53)) 12964 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49) (("}") reduce 49)) 12965 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29) (("}") reduce 29)) 12966 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59) (("}") reduce 59)) 12967 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60) (("}") reduce 60)) 12968 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11) (("}") reduce 11)) 12969 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42) (("}") reduce 42)) 12970 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39) (("}") reduce 39)) 12971 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25) (("}") reduce 25)) 12972 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50) (("}") reduce 50)) 12973 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66) (("}") reduce 66)) 12974 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18) (("}") reduce 18)) 12975 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17) (("}") reduce 17)) 12976 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15) (("}") reduce 15)) 12977 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14) (("}") reduce 14)) 12978 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38) (("}") reduce 38)) 12979 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26) (("}") reduce 26)) 12980 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22) (("}") reduce 22)) 12981 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24) (("}") reduce 24)) 12982 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16) (("}") reduce 16)) 12983 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48) (("}") reduce 48)) 12984 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20) (("}") reduce 20)) 12985 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70) (("}") reduce 70)) 12986 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3) (("}") reduce 3)) 12987 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12) (("}") reduce 12)) 12988 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54) (("}") reduce 54)) 12989 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65) (("}") reduce 65)) 12990 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75) (("}") reduce 75)) 12991 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30) (("}") reduce 30)) 12992 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68) (("}") reduce 68)) 12993 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21) (("}") reduce 21)) 12994 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23) (("}") reduce 23)) 12995 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41) (("}") reduce 41)) 12996 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62) (("}") reduce 62)) 12997 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34) (("}") reduce 34)) 12998 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40) (("}") reduce 40)) 12999 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13) (("}") reduce 13)) 13000 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55) (("}") reduce 55)) 13001 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1) (("}") reduce 1)) 13002 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2) (("}") reduce 2)) 13003 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9) (("}") reduce 9)) 13004 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33) (("}") reduce 33)) 13005 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58) (("}") reduce 58)) 13006 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37) (("}") reduce 37)) 13007 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64) (("}") reduce 64)) 13008 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46) (("}") reduce 46)) 13009 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8) (("}") reduce 8)) 13010 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6) (("}") reduce 6)) 13011 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7) (("}") reduce 7)) 13012 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69) (("}") reduce 69)) 13013 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63) (("}") reduce 63)) 13014 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56) (("}") reduce 56)) 13015 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10) (("}") reduce 10)) 13016 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67) (("}") reduce 67)) 13017 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44) (("}") reduce 44)) 13018 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76) (("}") reduce 76)) 13019 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71) (("}") reduce 71)) 13020 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77) (("}") reduce 77)) 13021 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78) (("}") reduce 78)) 13022 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79) (("}") reduce 79)) 13023 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4) (("}") reduce 4)) 13024 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5) (("}") reduce 5)) 13025 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43) (("}") reduce 43)) 13026 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73) (("}") reduce 73)) 13027 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82) (("}") reduce 82)) 13028 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47) (("}") reduce 47)) 13029 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31) (("}") reduce 31)) 13030 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57) (("}") reduce 57)) 13031 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61) (("}") reduce 61)) 13032 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28) (("}") reduce 28)) 13033 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36) (("}") reduce 36)) 13034 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32) (("}") reduce 32)) 13035 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35) (("}") reduce 35)) 13036 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19) (("}") reduce 19)) 13037 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45) (("}") reduce 45)) 13038 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525) (("}") reduce 525)) 13040 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72) (("}") reduce 72)) 13041 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83) (("}") reduce 83)) 13042 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555) (("}") reduce 555)) 13045 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527) (("}") reduce 527)) 13047 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490) (("}") reduce 490)) 13048 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289) (("}") reduce 289)) 13052 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290) (("}") reduce 290)) 13053 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 13058 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409) (("}") reduce 409)) 13060 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407) (("}") reduce 407)) 13070 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 403)) 13071 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 405)) 13072 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 404)) 13073 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 397)) 13074 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 394)) 13075 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 402)) 13076 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 396)) 13077 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 400)) 13078 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 395)) 13079 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 398)) 13080 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 399)) 13081 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 406)) 13082 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 401)) 13084 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 391)) 13087 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551) (("}") reduce 551)) 13090 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547) (("}") reduce 547)) 13092 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546) (("}") reduce 546)) 13094 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542) (("}") reduce 542)) 13095 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548) (("}") reduce 548)) 13096 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392) (("}") reduce 392)) 13100 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555) (("}") reduce 555)) 13102 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490) (("}") reduce 490)) 13103 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289) (("}") reduce 289)) 13107 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290) (("}") reduce 290)) 13108 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 13109 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487) (("}") reduce 487)) 13112 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554) (("}") reduce 554)) 13113 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489) (("}") reduce 489)) 13114 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491) (("}") reduce 491)) 13119 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543) (("}") reduce 543)) 13120 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 13121 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549) (("}") reduce 549)) 13122 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 13125 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 13126 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544) (("}") reduce 544)) 13127 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550) (("}") reduce 550)) 13128 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545) (("}") reduce 545)) 13129 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488) (("}") reduce 488)) 13130 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553) (("}") reduce 553)) 13133 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552) (("}") reduce 552)) 13134 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487) (("}") reduce 487)) 13145 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477) (("}") reduce 477)) 13146 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472) (("}") reduce 472)) 13156 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 477)) 13157 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473) (("}") reduce 473)) 13159 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 464)) 13161 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524) (("}") reduce 524)) 13163 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554) (("}") reduce 554)) 13166 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526) (("}") reduce 526)) 13167 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489) (("}") reduce 489)) 13168 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491) (("}") reduce 491)) 13170 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468) (("}") reduce 468)) 13171 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470) (("}") reduce 470)) 13176 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543) (("}") reduce 543)) 13177 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 13178 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549) (("}") reduce 549)) 13179 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 13182 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 13183 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544) (("}") reduce 544)) 13184 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550) (("}") reduce 550)) 13185 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545) (("}") reduce 545)) 13187 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 13189 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_DOUBLE_ARROW) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503) (("}") reduce 503)) 13191 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 390)) 13192 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 465)) 13195 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 464)) 13196 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456) (("}") reduce 456)) 13197 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 466)) 13198 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452) (("}") reduce 452)) 13199 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469) (("}") reduce 469)) 13201 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509) (("}") reduce 509)) 13203 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510) (("}") reduce 510)) 13204 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508) (("}") reduce 508)) 13205 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 602) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 602)) 13206 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 601) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 601)) 13207 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488) (("}") reduce 488)) 13208 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 461)) 13209 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454) (("}") reduce 454)) 13211 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 13212 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 13213 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 13214 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492) (("}") reduce 492)) 13215 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 13216 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385) (("}") reduce 385)) 13219 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494) (("}") reduce 494)) 13220 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387) (("}") reduce 387)) 13221 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493) (("}") reduce 493)) 13222 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495) (("}") reduce 495)) 13223 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556) (("}") reduce 556)) 13229 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557) (("}") reduce 557)) 13230 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 13231 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558) (("}") reduce 558)) 13232 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 13233 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384) (("}") reduce 384)) 13235 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 13236 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386) (("}") reduce 386)) 13244 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383) (("}") reduce 383)) 13246 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496) (("}") reduce 496)) 13252 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_DOUBLE_ARROW) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231) (("}") reduce 231)) 13257 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 389)) 13261 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_DOUBLE_ARROW) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596) (("}") reduce 596)) 13262 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450) (("}") reduce 450)) 13263 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 599) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 599)) 13264 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 598) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 598)) 13265 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408) (("}") reduce 408)) 13267 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_DOUBLE_ARROW) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501) (("}") reduce 501)) 13268 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457) (("}") reduce 457)) 13269 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289) (("}") reduce 289)) 13273 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290) (("}") reduce 290)) 13274 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 13277 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_DOUBLE_ARROW) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600) (("}") reduce 600)) 13280 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_DOUBLE_ARROW) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597) (("}") reduce 597)) 13281 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451) (("}") reduce 451)) 13282 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410) (("}") reduce 410)) 13283 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393) (("}") reduce 393)) 13284 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455) (("}") reduce 455)) 13285 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453) (("}") reduce 453)) 13288 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_DOUBLE_ARROW) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502) (("}") reduce 502)) 13289 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 13290 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430) (("}") reduce 430)) 13291 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429) (("}") reduce 429)) 13293 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443) (("}") reduce 443)) 13294 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553) (("}") reduce 553)) 13297 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552) (("}") reduce 552)) 13299 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505) (("}") reduce 505)) 13300 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431) (("}") reduce 431)) 13301 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456) (("}") reduce 456)) 13302 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 466)) 13303 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452) (("}") reduce 452)) 13304 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469) (("}") reduce 469)) 13306 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509) (("}") reduce 509)) 13308 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510) (("}") reduce 510)) 13309 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508) (("}") reduce 508)) 13310 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 602)) 13311 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 601)) 13312 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488) (("}") reduce 488)) 13313 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 461)) 13314 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454) (("}") reduce 454)) 13316 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 13317 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 13318 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 13319 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492) (("}") reduce 492)) 13320 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 13321 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385) (("}") reduce 385)) 13324 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494) (("}") reduce 494)) 13325 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387) (("}") reduce 387)) 13326 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493) (("}") reduce 493)) 13327 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495) (("}") reduce 495)) 13328 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556) (("}") reduce 556)) 13334 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557) (("}") reduce 557)) 13335 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 13336 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558) (("}") reduce 558)) 13337 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 13338 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384) (("}") reduce 384)) 13340 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 13341 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386) (("}") reduce 386)) 13349 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383) (("}") reduce 383)) 13351 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496) (("}") reduce 496)) 13357 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231) (("}") reduce 231)) 13362 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 389)) 13366 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596) (("}") reduce 596)) 13367 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450) (("}") reduce 450)) 13368 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 599)) 13369 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 598)) 13370 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408) (("}") reduce 408)) 13372 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501) (("}") reduce 501)) 13373 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457) (("}") reduce 457)) 13374 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289) (("}") reduce 289)) 13378 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290) (("}") reduce 290)) 13379 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 13382 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600) (("}") reduce 600)) 13385 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597) (("}") reduce 597)) 13386 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451) (("}") reduce 451)) 13387 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410) (("}") reduce 410)) 13388 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393) (("}") reduce 393)) 13389 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455) (("}") reduce 455)) 13390 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453) (("}") reduce 453)) 13393 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502) (("}") reduce 502)) 13394 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 13395 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430) (("}") reduce 430)) 13396 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429) (("}") reduce 429)) 13398 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443) (("}") reduce 443)) 13399 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553) (("}") reduce 553)) 13402 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552) (("}") reduce 552)) 13404 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505) (("}") reduce 505)) 13405 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431) (("}") reduce 431)) 13406 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 14066 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 14068 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 14070 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("]") reduce 496) (("^") reduce 496) (("|") reduce 496)) 14071 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("]") reduce 553) (("^") reduce 553) (("|") reduce 553)) 14074 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("]") reduce 552) (("^") reduce 552) (("|") reduce 552)) 14075 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 448) (("^") shift) (("|") shift)) 14076 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 411) (("^") shift) (("|") shift)) 14077 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 412) (("^") shift) (("|") shift)) 14078 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 417) (("^") reduce 417) (("|") reduce 417)) 14079 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 418) (("^") reduce 418) (("|") reduce 418)) 14083 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 446) (("^") shift) (("|") shift)) 14084 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 447) (("^") shift) (("|") shift)) 14085 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("]") reduce 439) (("^") reduce 439) (("|") reduce 439)) 14086 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("]") reduce 437) (("^") reduce 437) (("|") reduce 437)) 14087 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("]") reduce 425) (("^") reduce 425) (("|") reduce 425)) 14088 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("]") reduce 420) (("^") reduce 420) (("|") reduce 420)) 14089 ((("%") shift) (("*") shift) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("]") reduce 422) (("^") reduce 422) (("|") reduce 422)) 14090 ((("%") shift) (("*") shift) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("]") reduce 421) (("^") reduce 421) (("|") reduce 421)) 14091 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("]") reduce 423) (("^") reduce 423) (("|") reduce 423)) 14092 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("]") reduce 426) (("^") reduce 426) (("|") reduce 426)) 14094 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("]") reduce 74) (("^") reduce 74) (("|") reduce 74)) 14095 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("]") reduce 52) (("^") reduce 52) (("|") reduce 52)) 14096 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("]") reduce 27) (("^") reduce 27) (("|") reduce 27)) 14097 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("]") reduce 51) (("^") reduce 51) (("|") reduce 51)) 14098 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("]") reduce 53) (("^") reduce 53) (("|") reduce 53)) 14099 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("]") reduce 49) (("^") reduce 49) (("|") reduce 49)) 14100 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("]") reduce 29) (("^") reduce 29) (("|") reduce 29)) 14101 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("]") reduce 59) (("^") reduce 59) (("|") reduce 59)) 14102 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("]") reduce 60) (("^") reduce 60) (("|") reduce 60)) 14103 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("]") reduce 11) (("^") reduce 11) (("|") reduce 11)) 14104 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("]") reduce 42) (("^") reduce 42) (("|") reduce 42)) 14105 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("]") reduce 39) (("^") reduce 39) (("|") reduce 39)) 14106 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("]") reduce 25) (("^") reduce 25) (("|") reduce 25)) 14107 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("]") reduce 50) (("^") reduce 50) (("|") reduce 50)) 14108 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("]") reduce 66) (("^") reduce 66) (("|") reduce 66)) 14109 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("]") reduce 18) (("^") reduce 18) (("|") reduce 18)) 14110 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("]") reduce 17) (("^") reduce 17) (("|") reduce 17)) 14111 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("]") reduce 15) (("^") reduce 15) (("|") reduce 15)) 14112 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("]") reduce 14) (("^") reduce 14) (("|") reduce 14)) 14113 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("]") reduce 38) (("^") reduce 38) (("|") reduce 38)) 14114 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("]") reduce 26) (("^") reduce 26) (("|") reduce 26)) 14115 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("]") reduce 22) (("^") reduce 22) (("|") reduce 22)) 14116 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("]") reduce 24) (("^") reduce 24) (("|") reduce 24)) 14117 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("]") reduce 16) (("^") reduce 16) (("|") reduce 16)) 14118 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("]") reduce 48) (("^") reduce 48) (("|") reduce 48)) 14119 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("]") reduce 20) (("^") reduce 20) (("|") reduce 20)) 14120 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("]") reduce 70) (("^") reduce 70) (("|") reduce 70)) 14121 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("]") reduce 3) (("^") reduce 3) (("|") reduce 3)) 14122 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("]") reduce 12) (("^") reduce 12) (("|") reduce 12)) 14123 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("]") reduce 54) (("^") reduce 54) (("|") reduce 54)) 14124 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("]") reduce 65) (("^") reduce 65) (("|") reduce 65)) 14125 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("]") reduce 75) (("^") reduce 75) (("|") reduce 75)) 14126 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("]") reduce 30) (("^") reduce 30) (("|") reduce 30)) 14127 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("]") reduce 68) (("^") reduce 68) (("|") reduce 68)) 14128 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("]") reduce 21) (("^") reduce 21) (("|") reduce 21)) 14129 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("]") reduce 23) (("^") reduce 23) (("|") reduce 23)) 14130 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("]") reduce 41) (("^") reduce 41) (("|") reduce 41)) 14131 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("]") reduce 62) (("^") reduce 62) (("|") reduce 62)) 14132 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("]") reduce 34) (("^") reduce 34) (("|") reduce 34)) 14133 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("]") reduce 40) (("^") reduce 40) (("|") reduce 40)) 14134 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("]") reduce 13) (("^") reduce 13) (("|") reduce 13)) 14135 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("]") reduce 55) (("^") reduce 55) (("|") reduce 55)) 14136 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("]") reduce 1) (("^") reduce 1) (("|") reduce 1)) 14137 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("]") reduce 2) (("^") reduce 2) (("|") reduce 2)) 14138 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("]") reduce 9) (("^") reduce 9) (("|") reduce 9)) 14139 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("]") reduce 33) (("^") reduce 33) (("|") reduce 33)) 14140 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("]") reduce 58) (("^") reduce 58) (("|") reduce 58)) 14141 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("]") reduce 37) (("^") reduce 37) (("|") reduce 37)) 14142 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("]") reduce 64) (("^") reduce 64) (("|") reduce 64)) 14143 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("]") reduce 46) (("^") reduce 46) (("|") reduce 46)) 14144 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("]") reduce 8) (("^") reduce 8) (("|") reduce 8)) 14145 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("]") reduce 6) (("^") reduce 6) (("|") reduce 6)) 14146 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("]") reduce 7) (("^") reduce 7) (("|") reduce 7)) 14147 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("]") reduce 69) (("^") reduce 69) (("|") reduce 69)) 14148 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("]") reduce 63) (("^") reduce 63) (("|") reduce 63)) 14149 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("]") reduce 56) (("^") reduce 56) (("|") reduce 56)) 14150 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("]") reduce 10) (("^") reduce 10) (("|") reduce 10)) 14151 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("]") reduce 67) (("^") reduce 67) (("|") reduce 67)) 14152 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("]") reduce 44) (("^") reduce 44) (("|") reduce 44)) 14153 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("]") reduce 76) (("^") reduce 76) (("|") reduce 76)) 14154 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("]") reduce 71) (("^") reduce 71) (("|") reduce 71)) 14155 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("]") reduce 77) (("^") reduce 77) (("|") reduce 77)) 14156 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("]") reduce 78) (("^") reduce 78) (("|") reduce 78)) 14157 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("]") reduce 79) (("^") reduce 79) (("|") reduce 79)) 14158 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("]") reduce 4) (("^") reduce 4) (("|") reduce 4)) 14159 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("]") reduce 5) (("^") reduce 5) (("|") reduce 5)) 14160 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("]") reduce 43) (("^") reduce 43) (("|") reduce 43)) 14161 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("]") reduce 73) (("^") reduce 73) (("|") reduce 73)) 14162 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("]") reduce 82) (("^") reduce 82) (("|") reduce 82)) 14163 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("]") reduce 47) (("^") reduce 47) (("|") reduce 47)) 14164 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("]") reduce 31) (("^") reduce 31) (("|") reduce 31)) 14165 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("]") reduce 57) (("^") reduce 57) (("|") reduce 57)) 14166 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("]") reduce 61) (("^") reduce 61) (("|") reduce 61)) 14167 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("]") reduce 28) (("^") reduce 28) (("|") reduce 28)) 14168 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("]") reduce 36) (("^") reduce 36) (("|") reduce 36)) 14169 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("]") reduce 32) (("^") reduce 32) (("|") reduce 32)) 14170 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("]") reduce 35) (("^") reduce 35) (("|") reduce 35)) 14171 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("]") reduce 19) (("^") reduce 19) (("|") reduce 19)) 14172 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("]") reduce 45) (("^") reduce 45) (("|") reduce 45)) 14173 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("]") reduce 525) (("^") reduce 525) (("|") reduce 525)) 14175 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("]") reduce 72) (("^") reduce 72) (("|") reduce 72)) 14176 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("]") reduce 83) (("^") reduce 83) (("|") reduce 83)) 14177 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("]") reduce 555) (("^") reduce 555) (("|") reduce 555)) 14180 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("]") reduce 527) (("^") reduce 527) (("|") reduce 527)) 14182 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("]") reduce 490) (("^") reduce 490) (("|") reduce 490)) 14183 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("]") reduce 289) (("^") reduce 289) (("|") reduce 289)) 14187 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("]") reduce 290) (("^") reduce 290) (("|") reduce 290)) 14188 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 14193 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("]") reduce 409) (("^") reduce 409) (("|") reduce 409)) 14195 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("]") reduce 407) (("^") reduce 407) (("|") reduce 407)) 14205 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 403) (("^") shift) (("|") shift)) 14206 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 405) (("^") shift) (("|") shift)) 14207 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 404) (("^") shift) (("|") shift)) 14208 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 397) (("^") shift) (("|") shift)) 14209 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 394) (("^") shift) (("|") shift)) 14210 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 402) (("^") shift) (("|") shift)) 14211 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 396) (("^") shift) (("|") shift)) 14212 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 400) (("^") shift) (("|") shift)) 14213 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 395) (("^") shift) (("|") shift)) 14214 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 398) (("^") shift) (("|") shift)) 14215 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 399) (("^") shift) (("|") shift)) 14216 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 406) (("^") shift) (("|") shift)) 14217 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 401) (("^") shift) (("|") shift)) 14219 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 391) (("^") shift) (("|") shift)) 14222 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("]") reduce 551) (("^") reduce 551) (("|") reduce 551)) 14225 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("]") reduce 547) (("^") reduce 547) (("|") reduce 547)) 14227 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("]") reduce 546) (("^") reduce 546) (("|") reduce 546)) 14229 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("]") reduce 542) (("^") reduce 542) (("|") reduce 542)) 14230 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("]") reduce 548) (("^") reduce 548) (("|") reduce 548)) 14231 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("]") reduce 392) (("^") reduce 392) (("|") reduce 392)) 14235 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("]") reduce 555) (("^") reduce 555) (("|") reduce 555)) 14237 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("]") reduce 490) (("^") reduce 490) (("|") reduce 490)) 14238 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("]") reduce 289) (("^") reduce 289) (("|") reduce 289)) 14242 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("]") reduce 290) (("^") reduce 290) (("|") reduce 290)) 14243 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 14244 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("]") reduce 487) (("^") reduce 487) (("|") reduce 487)) 14247 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("]") reduce 554) (("^") reduce 554) (("|") reduce 554)) 14248 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("]") reduce 489) (("^") reduce 489) (("|") reduce 489)) 14249 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("]") reduce 491) (("^") reduce 491) (("|") reduce 491)) 14254 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("]") reduce 543) (("^") reduce 543) (("|") reduce 543)) 14255 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 14256 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("]") reduce 549) (("^") reduce 549) (("|") reduce 549)) 14257 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 14260 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 14261 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("]") reduce 544) (("^") reduce 544) (("|") reduce 544)) 14262 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("]") reduce 550) (("^") reduce 550) (("|") reduce 550)) 14263 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("]") reduce 545) (("^") reduce 545) (("|") reduce 545)) 14264 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("]") reduce 488) (("^") reduce 488) (("|") reduce 488)) 14265 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("]") reduce 553) (("^") reduce 553) (("|") reduce 553)) 14268 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("]") reduce 552) (("^") reduce 552) (("|") reduce 552)) 14269 ((("%") reduce 557) ((")") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 14270 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("]") reduce 487) (("^") reduce 487) (("|") reduce 487)) 14281 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("]") reduce 477) (("^") reduce 477) (("|") reduce 477)) 14282 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("]") reduce 472) (("^") reduce 472) (("|") reduce 472)) 14355 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 477) (("^") shift) (("|") shift)) 14669 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("]") reduce 473) (("^") reduce 473) (("|") reduce 473)) 14672 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("]") reduce 524) (("^") reduce 524) (("|") reduce 524)) 14674 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("]") reduce 554) (("^") reduce 554) (("|") reduce 554)) 14677 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("]") reduce 526) (("^") reduce 526) (("|") reduce 526)) 14678 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("]") reduce 489) (("^") reduce 489) (("|") reduce 489)) 14679 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("]") reduce 491) (("^") reduce 491) (("|") reduce 491)) 14681 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("]") reduce 468) (("^") reduce 468) (("|") reduce 468)) 14682 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("]") reduce 470) (("^") reduce 470) (("|") reduce 470)) 14687 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("]") reduce 543) (("^") reduce 543) (("|") reduce 543)) 14688 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 14689 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("]") reduce 549) (("^") reduce 549) (("|") reduce 549)) 14690 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 14693 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 14694 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("]") reduce 544) (("^") reduce 544) (("|") reduce 544)) 14695 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("]") reduce 550) (("^") reduce 550) (("|") reduce 550)) 14696 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("]") reduce 545) (("^") reduce 545) (("|") reduce 545)) 14698 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 14700 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("]") reduce 503) (("^") reduce 503) (("|") reduce 503)) 14702 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 390) (("^") shift) (("|") shift)) 14703 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 465) (("^") shift) (("|") shift)) 14714 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("]") reduce 523) (("^") reduce 523) (("|") reduce 523)) 14716 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("]") reduce 504) (("^") reduce 504) (("|") reduce 504)) 14718 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("|") reduce 517)) 14719 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("]") reduce 507) (("^") reduce 507) (("|") reduce 507)) 14723 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_DOUBLE_ARROW) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("]") reduce 500) (("^") reduce 500) (("|") reduce 500)) 14724 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("|") reduce 516)) 14725 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("|") reduce 520)) 14731 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("|") reduce 515)) 14733 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("]") reduce 506) (("^") reduce 506) (("|") reduce 506)) 14735 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("|") reduce 519)) 14736 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 14737 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 14738 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 14740 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("|") reduce 522)) 14743 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("|") reduce 521)) 14749 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 14752 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("]") reduce 518) (("^") reduce 518) (("|") reduce 518)) 14754 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("]") reduce 551) (("^") reduce 551) (("|") reduce 551)) 14755 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 462) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 462) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 14762 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("]") reduce 547) (("^") reduce 547) (("|") reduce 547)) 14763 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("]") reduce 513) (("^") reduce 513) (("|") reduce 513)) 14765 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("]") reduce 512) (("^") reduce 512) (("|") reduce 512)) 14766 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("]") reduce 511) (("^") reduce 511) (("|") reduce 511)) 14767 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 463) (("^") shift) (("|") shift)) 14770 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("]") reduce 546) (("^") reduce 546) (("|") reduce 546)) 14771 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("]") reduce 467) (("^") reduce 467) (("|") reduce 467)) 14772 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("]") reduce 449) (("^") reduce 449) (("|") reduce 449)) 14773 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("]") reduce 471) (("^") reduce 471) (("|") reduce 471)) 14774 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("]") reduce 514) (("^") reduce 514) (("|") reduce 514)) 14775 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("]") reduce 444) (("^") reduce 444) (("|") reduce 444)) 14776 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("]") reduce 445) (("^") reduce 445) (("|") reduce 445)) 14777 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 14778 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("]") reduce 542) (("^") reduce 542) (("|") reduce 542)) 14779 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("]") reduce 548) (("^") reduce 548) (("|") reduce 548)) 14780 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 388) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("]") reduce 388) (("^") reduce 388) (("|") reduce 388)) 14783 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("]") reduce 432) (("^") reduce 432) (("|") reduce 432)) 14814 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 416) (("^") shift) (("|") reduce 416)) 14815 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 419) (("^") reduce 419) (("|") reduce 419)) 14816 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("]") reduce 428) (("^") reduce 428) (("|") reduce 428)) 14817 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 441) (("^") reduce 441) (("|") reduce 441)) 14818 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("]") reduce 427) (("^") reduce 427) (("|") reduce 427)) 14819 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("]") reduce 424) (("^") reduce 424) (("|") reduce 424)) 14820 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 415) (("^") shift) (("|") shift)) 14821 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 413) (("^") shift) (("|") shift)) 14822 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 414) (("^") shift) (("|") shift)) 14823 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("]") reduce 438) (("^") reduce 438) (("|") reduce 438)) 14824 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 434) (("^") reduce 434) (("|") reduce 434)) 14825 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 436) (("^") reduce 436) (("|") reduce 436)) 14826 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 433) (("^") reduce 433) (("|") reduce 433)) 14827 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("]") reduce 440) (("^") reduce 440) (("|") reduce 440)) 14828 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 435) (("^") reduce 435) (("|") reduce 435)) 14831 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 14832 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 14833 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 14834 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("]") reduce 492) (("^") reduce 492) (("|") reduce 492)) 14835 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 14836 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("]") reduce 551) (("^") reduce 551) (("|") reduce 551)) 14837 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("]") reduce 494) (("^") reduce 494) (("|") reduce 494)) 14838 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("]") reduce 442) (("^") reduce 442) (("|") reduce 442)) 14839 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("]") reduce 493) (("^") reduce 493) (("|") reduce 493)) 14840 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("]") reduce 495) (("^") reduce 495) (("|") reduce 495)) 14841 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("]") reduce 556) (("^") reduce 556) (("|") reduce 556)) 14847 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("]") reduce 557) (("^") reduce 557) (("|") reduce 557)) 14848 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 14849 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 14850 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("]") reduce 558) (("^") reduce 558) (("|") reduce 558)) 14851 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 14854 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 14855 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 14857 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 14859 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("]") reduce 496) (("^") reduce 496) (("|") reduce 496)) 14860 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("]") reduce 553) (("^") reduce 553) (("|") reduce 553)) 14863 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("]") reduce 552) (("^") reduce 552) (("|") reduce 552)) 14864 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 448) (("^") shift) (("|") shift)) 14865 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 411) (("^") shift) (("|") shift)) 14866 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 412) (("^") shift) (("|") shift)) 14867 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 417) (("^") reduce 417) (("|") reduce 417)) 14868 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 418) (("^") reduce 418) (("|") reduce 418)) 14872 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 446) (("^") shift) (("|") shift)) 14873 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 447) (("^") shift) (("|") shift)) 14874 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("]") reduce 439) (("^") reduce 439) (("|") reduce 439)) 14875 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("]") reduce 437) (("^") reduce 437) (("|") reduce 437)) 14876 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("]") reduce 425) (("^") reduce 425) (("|") reduce 425)) 14877 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("]") reduce 420) (("^") reduce 420) (("|") reduce 420)) 14878 ((("%") shift) (("*") shift) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("]") reduce 422) (("^") reduce 422) (("|") reduce 422)) 14879 ((("%") shift) (("*") shift) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("]") reduce 421) (("^") reduce 421) (("|") reduce 421)) 14880 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("]") reduce 423) (("^") reduce 423) (("|") reduce 423)) 14881 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("]") reduce 426) (("^") reduce 426) (("|") reduce 426)) 14883 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("]") reduce 74) (("^") reduce 74) (("|") reduce 74)) 14884 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("]") reduce 52) (("^") reduce 52) (("|") reduce 52)) 14885 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("]") reduce 27) (("^") reduce 27) (("|") reduce 27)) 14886 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("]") reduce 51) (("^") reduce 51) (("|") reduce 51)) 14887 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("]") reduce 53) (("^") reduce 53) (("|") reduce 53)) 14888 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("]") reduce 49) (("^") reduce 49) (("|") reduce 49)) 14889 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("]") reduce 29) (("^") reduce 29) (("|") reduce 29)) 14890 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("]") reduce 59) (("^") reduce 59) (("|") reduce 59)) 14891 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("]") reduce 60) (("^") reduce 60) (("|") reduce 60)) 14892 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("]") reduce 11) (("^") reduce 11) (("|") reduce 11)) 14893 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("]") reduce 42) (("^") reduce 42) (("|") reduce 42)) 14894 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("]") reduce 39) (("^") reduce 39) (("|") reduce 39)) 14895 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("]") reduce 25) (("^") reduce 25) (("|") reduce 25)) 14896 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("]") reduce 50) (("^") reduce 50) (("|") reduce 50)) 14897 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("]") reduce 66) (("^") reduce 66) (("|") reduce 66)) 14898 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("]") reduce 18) (("^") reduce 18) (("|") reduce 18)) 14899 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("]") reduce 17) (("^") reduce 17) (("|") reduce 17)) 14900 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("]") reduce 15) (("^") reduce 15) (("|") reduce 15)) 14901 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("]") reduce 14) (("^") reduce 14) (("|") reduce 14)) 14902 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("]") reduce 38) (("^") reduce 38) (("|") reduce 38)) 14903 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("]") reduce 26) (("^") reduce 26) (("|") reduce 26)) 14904 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("]") reduce 22) (("^") reduce 22) (("|") reduce 22)) 14905 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("]") reduce 24) (("^") reduce 24) (("|") reduce 24)) 14906 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("]") reduce 16) (("^") reduce 16) (("|") reduce 16)) 14907 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("]") reduce 48) (("^") reduce 48) (("|") reduce 48)) 14908 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("]") reduce 20) (("^") reduce 20) (("|") reduce 20)) 14909 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("]") reduce 70) (("^") reduce 70) (("|") reduce 70)) 14910 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("]") reduce 3) (("^") reduce 3) (("|") reduce 3)) 14911 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("]") reduce 12) (("^") reduce 12) (("|") reduce 12)) 14912 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("]") reduce 54) (("^") reduce 54) (("|") reduce 54)) 14913 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("]") reduce 65) (("^") reduce 65) (("|") reduce 65)) 14914 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("]") reduce 75) (("^") reduce 75) (("|") reduce 75)) 14915 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("]") reduce 30) (("^") reduce 30) (("|") reduce 30)) 14916 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("]") reduce 68) (("^") reduce 68) (("|") reduce 68)) 14917 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("]") reduce 21) (("^") reduce 21) (("|") reduce 21)) 14918 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("]") reduce 23) (("^") reduce 23) (("|") reduce 23)) 14919 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("]") reduce 41) (("^") reduce 41) (("|") reduce 41)) 14920 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("]") reduce 62) (("^") reduce 62) (("|") reduce 62)) 14921 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("]") reduce 34) (("^") reduce 34) (("|") reduce 34)) 14922 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("]") reduce 40) (("^") reduce 40) (("|") reduce 40)) 14923 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("]") reduce 13) (("^") reduce 13) (("|") reduce 13)) 14924 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("]") reduce 55) (("^") reduce 55) (("|") reduce 55)) 14925 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("]") reduce 1) (("^") reduce 1) (("|") reduce 1)) 14926 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("]") reduce 2) (("^") reduce 2) (("|") reduce 2)) 14927 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("]") reduce 9) (("^") reduce 9) (("|") reduce 9)) 14928 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("]") reduce 33) (("^") reduce 33) (("|") reduce 33)) 14929 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("]") reduce 58) (("^") reduce 58) (("|") reduce 58)) 14930 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("]") reduce 37) (("^") reduce 37) (("|") reduce 37)) 14931 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("]") reduce 64) (("^") reduce 64) (("|") reduce 64)) 14932 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("]") reduce 46) (("^") reduce 46) (("|") reduce 46)) 14933 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("]") reduce 8) (("^") reduce 8) (("|") reduce 8)) 14934 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("]") reduce 6) (("^") reduce 6) (("|") reduce 6)) 14935 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("]") reduce 7) (("^") reduce 7) (("|") reduce 7)) 14936 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("]") reduce 69) (("^") reduce 69) (("|") reduce 69)) 14937 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("]") reduce 63) (("^") reduce 63) (("|") reduce 63)) 14938 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("]") reduce 56) (("^") reduce 56) (("|") reduce 56)) 14939 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("]") reduce 10) (("^") reduce 10) (("|") reduce 10)) 14940 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("]") reduce 67) (("^") reduce 67) (("|") reduce 67)) 14941 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("]") reduce 44) (("^") reduce 44) (("|") reduce 44)) 14942 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("]") reduce 76) (("^") reduce 76) (("|") reduce 76)) 14943 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("]") reduce 71) (("^") reduce 71) (("|") reduce 71)) 14944 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("]") reduce 77) (("^") reduce 77) (("|") reduce 77)) 14945 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("]") reduce 78) (("^") reduce 78) (("|") reduce 78)) 14946 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("]") reduce 79) (("^") reduce 79) (("|") reduce 79)) 14947 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("]") reduce 4) (("^") reduce 4) (("|") reduce 4)) 14948 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("]") reduce 5) (("^") reduce 5) (("|") reduce 5)) 14949 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("]") reduce 43) (("^") reduce 43) (("|") reduce 43)) 14950 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("]") reduce 73) (("^") reduce 73) (("|") reduce 73)) 14951 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("]") reduce 82) (("^") reduce 82) (("|") reduce 82)) 14952 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("]") reduce 47) (("^") reduce 47) (("|") reduce 47)) 14953 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("]") reduce 31) (("^") reduce 31) (("|") reduce 31)) 14954 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("]") reduce 57) (("^") reduce 57) (("|") reduce 57)) 14955 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("]") reduce 61) (("^") reduce 61) (("|") reduce 61)) 14956 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("]") reduce 28) (("^") reduce 28) (("|") reduce 28)) 14957 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("]") reduce 36) (("^") reduce 36) (("|") reduce 36)) 14958 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("]") reduce 32) (("^") reduce 32) (("|") reduce 32)) 14959 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("]") reduce 35) (("^") reduce 35) (("|") reduce 35)) 14960 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("]") reduce 19) (("^") reduce 19) (("|") reduce 19)) 14961 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("]") reduce 45) (("^") reduce 45) (("|") reduce 45)) 14962 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("]") reduce 525) (("^") reduce 525) (("|") reduce 525)) 14964 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("]") reduce 72) (("^") reduce 72) (("|") reduce 72)) 14965 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("]") reduce 83) (("^") reduce 83) (("|") reduce 83)) 14966 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("]") reduce 555) (("^") reduce 555) (("|") reduce 555)) 14969 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("]") reduce 527) (("^") reduce 527) (("|") reduce 527)) 14971 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("]") reduce 490) (("^") reduce 490) (("|") reduce 490)) 14972 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("]") reduce 289) (("^") reduce 289) (("|") reduce 289)) 14976 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("]") reduce 290) (("^") reduce 290) (("|") reduce 290)) 14977 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 14982 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("]") reduce 409) (("^") reduce 409) (("|") reduce 409)) 14984 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("]") reduce 407) (("^") reduce 407) (("|") reduce 407)) 14994 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 403) (("^") shift) (("|") shift)) 14995 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 405) (("^") shift) (("|") shift)) 14996 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 404) (("^") shift) (("|") shift)) 14997 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 397) (("^") shift) (("|") shift)) 14998 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 394) (("^") shift) (("|") shift)) 14999 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 402) (("^") shift) (("|") shift)) 15000 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 396) (("^") shift) (("|") shift)) 15001 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 400) (("^") shift) (("|") shift)) 15002 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 395) (("^") shift) (("|") shift)) 15003 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 398) (("^") shift) (("|") shift)) 15004 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 399) (("^") shift) (("|") shift)) 15005 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 406) (("^") shift) (("|") shift)) 15006 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 401) (("^") shift) (("|") shift)) 15008 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 391) (("^") shift) (("|") shift)) 15011 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("]") reduce 551) (("^") reduce 551) (("|") reduce 551)) 15014 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("]") reduce 547) (("^") reduce 547) (("|") reduce 547)) 15016 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("]") reduce 546) (("^") reduce 546) (("|") reduce 546)) 15018 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("]") reduce 542) (("^") reduce 542) (("|") reduce 542)) 15019 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("]") reduce 548) (("^") reduce 548) (("|") reduce 548)) 15020 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("]") reduce 392) (("^") reduce 392) (("|") reduce 392)) 15024 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("]") reduce 555) (("^") reduce 555) (("|") reduce 555)) 15026 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("]") reduce 490) (("^") reduce 490) (("|") reduce 490)) 15027 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("]") reduce 289) (("^") reduce 289) (("|") reduce 289)) 15031 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("]") reduce 290) (("^") reduce 290) (("|") reduce 290)) 15032 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 15033 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("]") reduce 487) (("^") reduce 487) (("|") reduce 487)) 15036 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("]") reduce 554) (("^") reduce 554) (("|") reduce 554)) 15037 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("]") reduce 489) (("^") reduce 489) (("|") reduce 489)) 15038 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("]") reduce 491) (("^") reduce 491) (("|") reduce 491)) 15043 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("]") reduce 543) (("^") reduce 543) (("|") reduce 543)) 15044 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 15045 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("]") reduce 549) (("^") reduce 549) (("|") reduce 549)) 15046 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 15049 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 15050 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("]") reduce 544) (("^") reduce 544) (("|") reduce 544)) 15051 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("]") reduce 550) (("^") reduce 550) (("|") reduce 550)) 15052 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("]") reduce 545) (("^") reduce 545) (("|") reduce 545)) 15053 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("]") reduce 488) (("^") reduce 488) (("|") reduce 488)) 15054 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("]") reduce 553) (("^") reduce 553) (("|") reduce 553)) 15057 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("]") reduce 552) (("^") reduce 552) (("|") reduce 552)) 15058 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("]") reduce 487) (("^") reduce 487) (("|") reduce 487)) 15069 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("]") reduce 477) (("^") reduce 477) (("|") reduce 477)) 15070 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("]") reduce 472) (("^") reduce 472) (("|") reduce 472)) 15080 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 477) (("^") shift) (("|") shift)) 15081 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("]") reduce 473) (("^") reduce 473) (("|") reduce 473)) 15083 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 464) (("^") shift) (("|") shift)) 15085 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("]") reduce 524) (("^") reduce 524) (("|") reduce 524)) 15087 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("]") reduce 554) (("^") reduce 554) (("|") reduce 554)) 15090 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("]") reduce 526) (("^") reduce 526) (("|") reduce 526)) 15091 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("]") reduce 489) (("^") reduce 489) (("|") reduce 489)) 15092 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("]") reduce 491) (("^") reduce 491) (("|") reduce 491)) 15094 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("]") reduce 468) (("^") reduce 468) (("|") reduce 468)) 15095 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("]") reduce 470) (("^") reduce 470) (("|") reduce 470)) 15100 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("]") reduce 543) (("^") reduce 543) (("|") reduce 543)) 15101 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 15102 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("]") reduce 549) (("^") reduce 549) (("|") reduce 549)) 15103 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 15106 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 15107 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("]") reduce 544) (("^") reduce 544) (("|") reduce 544)) 15108 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("]") reduce 550) (("^") reduce 550) (("|") reduce 550)) 15109 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("]") reduce 545) (("^") reduce 545) (("|") reduce 545)) 15111 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 15113 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_DOUBLE_ARROW) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("]") reduce 503) (("^") reduce 503) (("|") reduce 503)) 15115 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 390) (("^") shift) (("|") shift)) 15116 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 465) (("^") shift) (("|") shift)) 15119 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 464) (("^") shift) (("|") shift)) 15120 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("]") reduce 456) (("^") reduce 456) (("|") reduce 456)) 15121 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 466) (("^") shift) (("|") shift)) 15122 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("]") reduce 452) (("^") reduce 452) (("|") reduce 452)) 15123 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("]") reduce 469) (("^") reduce 469) (("|") reduce 469)) 15125 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("]") reduce 509) (("^") reduce 509) (("|") reduce 509)) 15127 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("]") reduce 510) (("^") reduce 510) (("|") reduce 510)) 15128 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("]") reduce 508) (("^") reduce 508) (("|") reduce 508)) 15129 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 602) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 602) (("^") shift) (("|") shift)) 15130 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 601) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 601) (("^") shift) (("|") shift)) 15131 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("]") reduce 488) (("^") reduce 488) (("|") reduce 488)) 15132 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 461) (("^") shift) (("|") shift)) 15133 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("]") reduce 454) (("^") reduce 454) (("|") reduce 454)) 15135 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 15136 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 15137 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 15138 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("]") reduce 492) (("^") reduce 492) (("|") reduce 492)) 15139 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 15140 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("]") reduce 385) (("^") reduce 385) (("|") reduce 385)) 15143 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("]") reduce 494) (("^") reduce 494) (("|") reduce 494)) 15144 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("]") reduce 387) (("^") reduce 387) (("|") reduce 387)) 15145 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("]") reduce 493) (("^") reduce 493) (("|") reduce 493)) 15146 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("]") reduce 495) (("^") reduce 495) (("|") reduce 495)) 15147 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("]") reduce 556) (("^") reduce 556) (("|") reduce 556)) 15153 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("]") reduce 557) (("^") reduce 557) (("|") reduce 557)) 15154 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 15155 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("]") reduce 558) (("^") reduce 558) (("|") reduce 558)) 15156 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 15157 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("]") reduce 384) (("^") reduce 384) (("|") reduce 384)) 15159 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 15160 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("]") reduce 386) (("^") reduce 386) (("|") reduce 386)) 15168 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("]") reduce 383) (("^") reduce 383) (("|") reduce 383)) 15170 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("]") reduce 496) (("^") reduce 496) (("|") reduce 496)) 15176 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_DOUBLE_ARROW) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("]") reduce 231) (("^") reduce 231) (("|") reduce 231)) 15181 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 389) (("^") shift) (("|") shift)) 15185 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_DOUBLE_ARROW) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("]") reduce 596) (("^") reduce 596) (("|") reduce 596)) 15186 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("]") reduce 450) (("^") reduce 450) (("|") reduce 450)) 15187 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 599) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 599) (("^") shift) (("|") shift)) 15188 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 598) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 598) (("^") shift) (("|") shift)) 15189 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("]") reduce 408) (("^") reduce 408) (("|") reduce 408)) 15191 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_DOUBLE_ARROW) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("]") reduce 501) (("^") reduce 501) (("|") reduce 501)) 15192 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("]") reduce 457) (("^") reduce 457) (("|") reduce 457)) 15193 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("]") reduce 289) (("^") reduce 289) (("|") reduce 289)) 15197 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("]") reduce 290) (("^") reduce 290) (("|") reduce 290)) 15198 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 15201 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_DOUBLE_ARROW) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("]") reduce 600) (("^") reduce 600) (("|") reduce 600)) 15204 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_DOUBLE_ARROW) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("]") reduce 597) (("^") reduce 597) (("|") reduce 597)) 15205 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("]") reduce 451) (("^") reduce 451) (("|") reduce 451)) 15206 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("]") reduce 410) (("^") reduce 410) (("|") reduce 410)) 15207 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("]") reduce 393) (("^") reduce 393) (("|") reduce 393)) 15208 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("]") reduce 455) (("^") reduce 455) (("|") reduce 455)) 15209 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("]") reduce 453) (("^") reduce 453) (("|") reduce 453)) 15212 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_DOUBLE_ARROW) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("]") reduce 502) (("^") reduce 502) (("|") reduce 502)) 15213 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 15214 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("]") reduce 430) (("^") reduce 430) (("|") reduce 430)) 15215 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("]") reduce 429) (("^") reduce 429) (("|") reduce 429)) 15217 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("]") reduce 443) (("^") reduce 443) (("|") reduce 443)) 15218 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("]") reduce 553) (("^") reduce 553) (("|") reduce 553)) 15221 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("]") reduce 552) (("^") reduce 552) (("|") reduce 552)) 15223 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("|") reduce 505)) 15224 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("]") reduce 431) (("^") reduce 431) (("|") reduce 431)) 15225 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("]") reduce 456) (("^") reduce 456) (("|") reduce 456)) 15226 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 466) (("^") shift) (("|") shift)) 15227 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("]") reduce 452) (("^") reduce 452) (("|") reduce 452)) 15228 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("]") reduce 469) (("^") reduce 469) (("|") reduce 469)) 15230 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("]") reduce 509) (("^") reduce 509) (("|") reduce 509)) 15232 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("]") reduce 510) (("^") reduce 510) (("|") reduce 510)) 15233 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("]") reduce 508) (("^") reduce 508) (("|") reduce 508)) 15234 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 602) (("^") shift) (("|") shift)) 15235 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 601) (("^") shift) (("|") shift)) 15236 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("]") reduce 488) (("^") reduce 488) (("|") reduce 488)) 15237 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 461) (("^") shift) (("|") shift)) 15238 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("]") reduce 454) (("^") reduce 454) (("|") reduce 454)) 15240 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 15241 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 15242 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 15243 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("]") reduce 492) (("^") reduce 492) (("|") reduce 492)) 15244 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 15245 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("]") reduce 385) (("^") reduce 385) (("|") reduce 385)) 15248 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("]") reduce 494) (("^") reduce 494) (("|") reduce 494)) 15249 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("]") reduce 387) (("^") reduce 387) (("|") reduce 387)) 15250 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("]") reduce 493) (("^") reduce 493) (("|") reduce 493)) 15251 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("]") reduce 495) (("^") reduce 495) (("|") reduce 495)) 15252 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("]") reduce 556) (("^") reduce 556) (("|") reduce 556)) 15258 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("]") reduce 557) (("^") reduce 557) (("|") reduce 557)) 15259 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 15260 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("]") reduce 558) (("^") reduce 558) (("|") reduce 558)) 15261 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 15262 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("]") reduce 384) (("^") reduce 384) (("|") reduce 384)) 15264 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 15265 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("]") reduce 386) (("^") reduce 386) (("|") reduce 386)) 15273 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("]") reduce 383) (("^") reduce 383) (("|") reduce 383)) 15275 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("]") reduce 496) (("^") reduce 496) (("|") reduce 496)) 15281 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("]") reduce 231) (("^") reduce 231) (("|") reduce 231)) 15286 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 389) (("^") shift) (("|") shift)) 15290 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("]") reduce 596) (("^") reduce 596) (("|") reduce 596)) 15291 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("]") reduce 450) (("^") reduce 450) (("|") reduce 450)) 15292 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 599) (("^") shift) (("|") shift)) 15293 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 598) (("^") shift) (("|") shift)) 15294 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("]") reduce 408) (("^") reduce 408) (("|") reduce 408)) 15296 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("]") reduce 501) (("^") reduce 501) (("|") reduce 501)) 15297 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("]") reduce 457) (("^") reduce 457) (("|") reduce 457)) 15298 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("]") reduce 289) (("^") reduce 289) (("|") reduce 289)) 15302 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("]") reduce 290) (("^") reduce 290) (("|") reduce 290)) 15303 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 15306 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("]") reduce 600) (("^") reduce 600) (("|") reduce 600)) 15309 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("]") reduce 597) (("^") reduce 597) (("|") reduce 597)) 15310 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("]") reduce 451) (("^") reduce 451) (("|") reduce 451)) 15311 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("]") reduce 410) (("^") reduce 410) (("|") reduce 410)) 15312 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("]") reduce 393) (("^") reduce 393) (("|") reduce 393)) 15313 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("]") reduce 455) (("^") reduce 455) (("|") reduce 455)) 15314 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("]") reduce 453) (("^") reduce 453) (("|") reduce 453)) 15317 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("]") reduce 502) (("^") reduce 502) (("|") reduce 502)) 15318 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 15319 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("]") reduce 430) (("^") reduce 430) (("|") reduce 430)) 15320 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("]") reduce 429) (("^") reduce 429) (("|") reduce 429)) 15322 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("]") reduce 443) (("^") reduce 443) (("|") reduce 443)) 15323 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("]") reduce 553) (("^") reduce 553) (("|") reduce 553)) 15326 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("]") reduce 552) (("^") reduce 552) (("|") reduce 552)) 15328 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("|") reduce 505)) 15329 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("]") reduce 431) (("^") reduce 431) (("|") reduce 431)) 15989 ((("%") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 15990 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 15991 ((("%") reduce 558) ((")") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 15992 ((("%") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 15995 ((("%") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 15996 ((("%") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 15998 ((("%") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 16000 ((("%") reduce 496) ((")") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 16001 ((("%") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 16004 ((("%") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 16005 ((("%") shift) ((")") reduce 448) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16006 ((("%") shift) ((")") reduce 411) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16007 ((("%") shift) ((")") reduce 412) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16008 ((("%") shift) ((")") reduce 417) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417)) 16009 ((("%") shift) ((")") reduce 418) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418)) 16013 ((("%") shift) ((")") reduce 446) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16014 ((("%") shift) ((")") reduce 447) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16015 ((("%") shift) ((")") reduce 439) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439)) 16016 ((("%") shift) ((")") reduce 437) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437)) 16017 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 16018 ((("%") shift) ((")") reduce 420) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420)) 16019 ((("%") shift) ((")") reduce 422) (("*") shift) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 16020 ((("%") shift) ((")") reduce 421) (("*") shift) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 16021 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 16022 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 16024 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 16025 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 16026 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 16027 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 16028 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 16029 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 16030 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 16031 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 16032 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 16033 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 16034 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 16035 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 16036 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 16037 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 16038 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 16039 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 16040 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 16041 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 16042 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 16043 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 16044 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 16045 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 16046 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 16047 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 16048 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 16049 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 16050 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 16051 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 16052 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 16053 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 16054 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 16055 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 16056 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 16057 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 16058 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 16059 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 16060 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 16061 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 16062 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 16063 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 16064 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 16065 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 16066 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 16067 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 16068 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 16069 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 16070 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 16071 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 16072 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 16073 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 16074 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 16075 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 16076 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 16077 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 16078 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 16079 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 16080 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 16081 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 16082 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 16083 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 16084 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 16085 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 16086 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 16087 ((("%") reduce 79) (("(") reduce 79) ((")") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 16088 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 16089 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 16090 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 16091 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 16092 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 16093 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 16094 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 16095 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 16096 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 16097 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 16098 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 16099 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 16100 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 16101 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 16102 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 16103 ((("%") reduce 525) (("(") reduce 562) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525)) 16105 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 16106 ((("%") reduce 83) (("(") reduce 83) ((")") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 16107 ((("%") reduce 555) (("(") reduce 564) ((")") reduce 555) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 16110 ((("%") reduce 527) (("(") reduce 563) ((")") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527)) 16112 ((("%") reduce 490) (("(") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 16113 ((("%") reduce 289) (("(") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 16117 ((("%") reduce 290) (("(") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 16118 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 16123 ((("%") reduce 409) ((")") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 16125 ((("%") reduce 407) ((")") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407)) 16135 ((("%") shift) ((")") reduce 403) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16136 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16137 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16138 ((("%") shift) ((")") reduce 397) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16139 ((("%") shift) ((")") reduce 394) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16140 ((("%") shift) ((")") reduce 402) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16141 ((("%") shift) ((")") reduce 396) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16142 ((("%") shift) ((")") reduce 400) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16143 ((("%") shift) ((")") reduce 395) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16144 ((("%") shift) ((")") reduce 398) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16145 ((("%") shift) ((")") reduce 399) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16146 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16147 ((("%") shift) ((")") reduce 401) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16149 ((("%") shift) ((")") reduce 391) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16152 ((("%") reduce 551) (("(") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 16155 ((("%") reduce 547) (("(") reduce 538) ((")") reduce 547) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 16157 ((("%") reduce 546) (("(") reduce 546) ((")") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 16159 ((("%") reduce 542) (("(") reduce 542) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 16160 ((("%") reduce 548) ((")") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 16161 ((("%") reduce 392) ((")") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392)) 16165 ((("%") reduce 555) (("(") reduce 564) ((")") reduce 555) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 16167 ((("%") reduce 490) (("(") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 16168 ((("%") reduce 289) (("(") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 16172 ((("%") reduce 290) (("(") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 16173 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 16174 ((("%") reduce 487) (("(") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 16177 ((("%") reduce 554) (("(") reduce 564) ((")") reduce 554) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 16178 ((("%") reduce 489) (("(") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 16179 ((("%") reduce 491) (("(") reduce 491) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 16184 ((("%") reduce 543) (("(") reduce 543) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 16185 ((("%") reduce 565) (("(") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 16186 ((("%") reduce 549) (("(") shift) ((")") reduce 549) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 16187 ((("%") reduce 567) (("(") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 16190 ((("%") reduce 566) (("(") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 16191 ((("%") reduce 544) (("(") reduce 544) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 16192 ((("%") reduce 550) (("(") shift) ((")") reduce 550) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 16193 ((("%") reduce 545) (("(") reduce 545) ((")") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 16194 ((("%") reduce 488) (("(") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 16195 ((("%") reduce 553) (("(") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 16198 ((("%") reduce 552) (("(") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 16199 ((("%") reduce 487) (("(") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 16210 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 16211 ((("%") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 16284 ((("%") shift) ((")") reduce 477) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16598 ((("%") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 16601 ((("%") reduce 524) (("(") reduce 562) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524)) 16603 ((("%") reduce 554) (("(") reduce 564) ((")") reduce 554) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 16606 ((("%") reduce 526) (("(") reduce 563) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526)) 16607 ((("%") reduce 489) (("(") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 16608 ((("%") reduce 491) (("(") reduce 491) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 16610 ((("%") reduce 468) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 16611 ((("%") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 16616 ((("%") reduce 543) (("(") reduce 543) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 16617 ((("%") reduce 565) (("(") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 16618 ((("%") reduce 549) (("(") shift) ((")") reduce 549) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 16619 ((("%") reduce 567) (("(") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 16622 ((("%") reduce 566) (("(") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 16623 ((("%") reduce 544) (("(") reduce 544) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 16624 ((("%") reduce 550) (("(") shift) ((")") reduce 550) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 16625 ((("%") reduce 545) (("(") reduce 545) ((")") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 16627 ((("%") reduce 460) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 16629 ((("%") reduce 503) (("(") reduce 503) ((")") reduce 503) (("*") reduce 503) (("+") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503)) 16631 ((("%") shift) ((")") reduce 390) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16632 ((("%") shift) ((")") reduce 465) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16643 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 16645 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 16647 ((("%") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 16648 ((("%") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507)) 16652 ((("%") reduce 500) (("(") shift) ((")") reduce 500) (("*") reduce 500) (("+") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_DOUBLE_ARROW) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 16653 ((("%") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 16654 ((("%") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 16660 ((("%") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 16662 ((("%") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506)) 16664 ((("%") reduce 519) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 16665 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 16666 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 16667 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 16669 ((("%") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 16672 ((("%") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 16678 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 16681 ((("%") reduce 518) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 16683 ((("%") reduce 551) (("(") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 16684 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) ((")") reduce 462) (("*") reduce 462) (("+") shift) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 462) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 16691 ((("%") reduce 547) (("(") reduce 538) ((")") reduce 547) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 16692 ((("%") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513)) 16694 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512)) 16695 ((("%") reduce 511) (("(") reduce 540) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511)) 16696 ((("%") shift) ((")") reduce 463) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16699 ((("%") reduce 546) (("(") reduce 546) ((")") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 16700 ((("%") reduce 467) ((")") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 16701 ((("%") reduce 449) ((")") reduce 449) (("*") reduce 449) (("+") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449)) 16702 ((("%") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 16703 ((("%") reduce 514) (("(") shift) ((")") reduce 514) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514)) 16704 ((("%") reduce 444) (("(") reduce 541) ((")") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444)) 16705 ((("%") reduce 445) ((")") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 16706 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 16707 ((("%") reduce 542) (("(") reduce 542) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 16708 ((("%") reduce 548) ((")") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 16709 ((("%") reduce 388) ((")") reduce 388) (("*") reduce 388) (("+") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 388) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388)) 16712 ((("%") reduce 432) ((")") reduce 432) (("*") reduce 432) (("+") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432)) 16743 ((("%") shift) ((")") reduce 416) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416)) 16744 ((("%") shift) ((")") reduce 419) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419)) 16745 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428)) 16746 ((("%") shift) ((")") reduce 441) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441)) 16747 ((("%") shift) ((")") reduce 427) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 16748 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 16749 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16750 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16751 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16752 ((("%") shift) ((")") reduce 438) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438)) 16753 ((("%") shift) ((")") reduce 434) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 16754 ((("%") shift) ((")") reduce 436) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 16755 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 16756 ((("%") shift) ((")") reduce 440) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440)) 16757 ((("%") shift) ((")") reduce 435) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 16760 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 16761 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 16762 ((("%") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 16763 ((("%") reduce 492) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 16764 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 16765 ((("%") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 16766 ((("%") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 16767 ((("%") reduce 442) ((")") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 16768 ((("%") reduce 493) ((")") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 16769 ((("%") reduce 495) ((")") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 16770 ((("%") reduce 556) ((")") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 16776 ((("%") reduce 557) ((")") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 16777 ((("%") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 16778 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 16779 ((("%") reduce 558) ((")") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 16780 ((("%") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 16783 ((("%") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 16784 ((("%") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 16786 ((("%") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 16788 ((("%") reduce 496) ((")") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 16789 ((("%") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 16792 ((("%") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 16793 ((("%") shift) ((")") reduce 448) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16794 ((("%") shift) ((")") reduce 411) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16795 ((("%") shift) ((")") reduce 412) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16796 ((("%") shift) ((")") reduce 417) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417)) 16797 ((("%") shift) ((")") reduce 418) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418)) 16801 ((("%") shift) ((")") reduce 446) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16802 ((("%") shift) ((")") reduce 447) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16803 ((("%") shift) ((")") reduce 439) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439)) 16804 ((("%") shift) ((")") reduce 437) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437)) 16805 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 16806 ((("%") shift) ((")") reduce 420) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420)) 16807 ((("%") shift) ((")") reduce 422) (("*") shift) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 16808 ((("%") shift) ((")") reduce 421) (("*") shift) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 16809 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 16810 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 16812 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 16813 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 16814 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 16815 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 16816 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 16817 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 16818 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 16819 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 16820 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 16821 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 16822 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 16823 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 16824 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 16825 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 16826 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 16827 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 16828 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 16829 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 16830 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 16831 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 16832 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 16833 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 16834 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 16835 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 16836 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 16837 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 16838 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 16839 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 16840 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 16841 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 16842 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 16843 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 16844 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 16845 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 16846 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 16847 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 16848 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 16849 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 16850 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 16851 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 16852 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 16853 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 16854 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 16855 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 16856 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 16857 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 16858 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 16859 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 16860 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 16861 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 16862 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 16863 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 16864 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 16865 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 16866 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 16867 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 16868 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 16869 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 16870 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 16871 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 16872 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 16873 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 16874 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 16875 ((("%") reduce 79) (("(") reduce 79) ((")") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 16876 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 16877 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 16878 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 16879 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 16880 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 16881 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 16882 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 16883 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 16884 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 16885 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 16886 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 16887 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 16888 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 16889 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 16890 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 16891 ((("%") reduce 525) (("(") reduce 562) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525)) 16893 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 16894 ((("%") reduce 83) (("(") reduce 83) ((")") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 16895 ((("%") reduce 555) (("(") reduce 564) ((")") reduce 555) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 16898 ((("%") reduce 527) (("(") reduce 563) ((")") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527)) 16900 ((("%") reduce 490) (("(") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 16901 ((("%") reduce 289) (("(") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 16905 ((("%") reduce 290) (("(") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 16906 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 16911 ((("%") reduce 409) ((")") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 16913 ((("%") reduce 407) ((")") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407)) 16923 ((("%") shift) ((")") reduce 403) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16924 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16925 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16926 ((("%") shift) ((")") reduce 397) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16927 ((("%") shift) ((")") reduce 394) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16928 ((("%") shift) ((")") reduce 402) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16929 ((("%") shift) ((")") reduce 396) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16930 ((("%") shift) ((")") reduce 400) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16931 ((("%") shift) ((")") reduce 395) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16932 ((("%") shift) ((")") reduce 398) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16933 ((("%") shift) ((")") reduce 399) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16934 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16935 ((("%") shift) ((")") reduce 401) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16937 ((("%") shift) ((")") reduce 391) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16940 ((("%") reduce 551) (("(") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 16943 ((("%") reduce 547) (("(") reduce 538) ((")") reduce 547) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 16945 ((("%") reduce 546) (("(") reduce 546) ((")") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 16947 ((("%") reduce 542) (("(") reduce 542) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 16948 ((("%") reduce 548) ((")") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 16949 ((("%") reduce 392) ((")") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392)) 16953 ((("%") reduce 555) (("(") reduce 564) ((")") reduce 555) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 16955 ((("%") reduce 490) (("(") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 16956 ((("%") reduce 289) (("(") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 16960 ((("%") reduce 290) (("(") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 16961 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 16962 ((("%") reduce 487) (("(") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 16965 ((("%") reduce 554) (("(") reduce 564) ((")") reduce 554) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 16966 ((("%") reduce 489) (("(") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 16967 ((("%") reduce 491) (("(") reduce 491) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 16972 ((("%") reduce 543) (("(") reduce 543) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 16973 ((("%") reduce 565) (("(") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 16974 ((("%") reduce 549) (("(") shift) ((")") reduce 549) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 16975 ((("%") reduce 567) (("(") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 16978 ((("%") reduce 566) (("(") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 16979 ((("%") reduce 544) (("(") reduce 544) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 16980 ((("%") reduce 550) (("(") shift) ((")") reduce 550) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 16981 ((("%") reduce 545) (("(") reduce 545) ((")") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 16982 ((("%") reduce 488) (("(") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 16983 ((("%") reduce 553) (("(") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 16986 ((("%") reduce 552) (("(") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 16987 ((("%") reduce 487) (("(") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 16998 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 16999 ((("%") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 17009 ((("%") shift) ((")") reduce 477) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17010 ((("%") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 17012 ((("%") shift) ((")") reduce 464) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17014 ((("%") reduce 524) (("(") reduce 562) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524)) 17016 ((("%") reduce 554) (("(") reduce 564) ((")") reduce 554) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 17019 ((("%") reduce 526) (("(") reduce 563) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526)) 17020 ((("%") reduce 489) (("(") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 17021 ((("%") reduce 491) (("(") reduce 491) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 17023 ((("%") reduce 468) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 17024 ((("%") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 17029 ((("%") reduce 543) (("(") reduce 543) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 17030 ((("%") reduce 565) (("(") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 17031 ((("%") reduce 549) (("(") shift) ((")") reduce 549) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 17032 ((("%") reduce 567) (("(") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 17035 ((("%") reduce 566) (("(") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 17036 ((("%") reduce 544) (("(") reduce 544) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 17037 ((("%") reduce 550) (("(") shift) ((")") reduce 550) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 17038 ((("%") reduce 545) (("(") reduce 545) ((")") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 17040 ((("%") reduce 460) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 17042 ((("%") reduce 503) (("(") reduce 503) ((")") reduce 503) (("*") reduce 503) (("+") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_DOUBLE_ARROW) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503)) 17044 ((("%") shift) ((")") reduce 390) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17045 ((("%") shift) ((")") reduce 465) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17048 ((("%") shift) ((")") reduce 464) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17049 ((("%") reduce 456) ((")") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456)) 17050 ((("%") shift) ((")") reduce 466) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17051 ((("%") reduce 452) ((")") reduce 452) (("*") reduce 452) (("+") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452)) 17052 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 17054 ((("%") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 17056 ((("%") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 17057 ((("%") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508)) 17058 ((("%") shift) ((")") reduce 602) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 602) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17059 ((("%") shift) ((")") reduce 601) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 601) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17060 ((("%") reduce 488) (("(") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 17061 ((("%") shift) ((")") reduce 461) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17062 ((("%") reduce 454) ((")") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454)) 17064 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 17065 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 17066 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 17067 ((("%") reduce 492) (("(") reduce 492) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 17068 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 17069 ((("%") reduce 385) (("(") reduce 385) ((")") reduce 385) (("*") reduce 385) (("+") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385)) 17072 ((("%") reduce 494) (("(") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 17073 ((("%") reduce 387) (("(") shift) ((")") reduce 387) (("*") reduce 387) (("+") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387)) 17074 ((("%") reduce 493) (("(") reduce 493) ((")") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 17075 ((("%") reduce 495) (("(") reduce 495) ((")") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 17076 ((("%") reduce 556) (("(") reduce 556) ((")") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 17082 ((("%") reduce 557) (("(") reduce 557) ((")") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 17083 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 17084 ((("%") reduce 558) (("(") reduce 558) ((")") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 17085 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 17086 ((("%") reduce 384) (("(") reduce 384) ((")") reduce 384) (("*") reduce 384) (("+") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384)) 17088 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 17089 ((("%") reduce 386) (("(") reduce 386) ((")") reduce 386) (("*") reduce 386) (("+") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386)) 17097 ((("%") reduce 383) (("(") reduce 383) ((")") reduce 383) (("*") reduce 383) (("+") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383)) 17099 ((("%") reduce 496) (("(") reduce 496) ((")") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 17105 ((("%") reduce 231) ((")") reduce 231) (("*") reduce 231) (("+") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_DOUBLE_ARROW) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231)) 17110 ((("%") shift) ((")") reduce 389) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17114 ((("%") reduce 596) ((")") reduce 596) (("*") reduce 596) (("+") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_DOUBLE_ARROW) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596)) 17115 ((("%") reduce 450) ((")") reduce 450) (("*") reduce 450) (("+") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450)) 17116 ((("%") shift) ((")") reduce 599) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 599) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17117 ((("%") shift) ((")") reduce 598) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 598) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17118 ((("%") reduce 408) ((")") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408)) 17120 ((("%") reduce 501) ((")") reduce 501) (("*") reduce 501) (("+") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_DOUBLE_ARROW) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501)) 17121 ((("%") reduce 457) ((")") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457)) 17122 ((("%") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289)) 17126 ((("%") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290)) 17127 ((("%") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 17130 ((("%") reduce 600) ((")") reduce 600) (("*") reduce 600) (("+") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_DOUBLE_ARROW) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600)) 17133 ((("%") reduce 597) ((")") reduce 597) (("*") reduce 597) (("+") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_DOUBLE_ARROW) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597)) 17134 ((("%") reduce 451) ((")") reduce 451) (("*") reduce 451) (("+") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451)) 17135 ((("%") reduce 410) ((")") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410)) 17136 ((("%") reduce 393) ((")") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 17137 ((("%") reduce 455) ((")") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455)) 17138 ((("%") reduce 453) ((")") reduce 453) (("*") reduce 453) (("+") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453)) 17141 ((("%") reduce 502) (("(") reduce 502) ((")") reduce 502) (("*") reduce 502) (("+") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_DOUBLE_ARROW) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502)) 17142 ((("%") reduce 458) ((")") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 17143 ((("%") reduce 430) ((")") reduce 430) (("*") reduce 430) (("+") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430)) 17144 ((("%") reduce 429) ((")") reduce 429) (("*") reduce 429) (("+") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429)) 17146 ((("%") reduce 443) (("(") reduce 539) ((")") reduce 443) (("*") reduce 443) (("+") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443)) 17147 ((("%") reduce 553) (("(") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 17150 ((("%") reduce 552) (("(") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 17152 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 17153 ((("%") reduce 431) ((")") reduce 431) (("*") reduce 431) (("+") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431)) 17154 ((("%") reduce 456) ((")") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456)) 17155 ((("%") shift) ((")") reduce 466) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17156 ((("%") reduce 452) ((")") reduce 452) (("*") reduce 452) (("+") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452)) 17157 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 17159 ((("%") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 17161 ((("%") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 17162 ((("%") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508)) 17163 ((("%") shift) ((")") reduce 602) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17164 ((("%") shift) ((")") reduce 601) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17165 ((("%") reduce 488) (("(") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 17166 ((("%") shift) ((")") reduce 461) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17167 ((("%") reduce 454) ((")") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454)) 17169 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 17170 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 17171 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 17172 ((("%") reduce 492) (("(") reduce 492) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 17173 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 17174 ((("%") reduce 385) (("(") reduce 385) ((")") reduce 385) (("*") reduce 385) (("+") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385)) 17177 ((("%") reduce 494) (("(") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 17178 ((("%") reduce 387) (("(") shift) ((")") reduce 387) (("*") reduce 387) (("+") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387)) 17179 ((("%") reduce 493) (("(") reduce 493) ((")") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 17180 ((("%") reduce 495) (("(") reduce 495) ((")") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 17181 ((("%") reduce 556) (("(") reduce 556) ((")") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 17187 ((("%") reduce 557) (("(") reduce 557) ((")") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 17188 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 17189 ((("%") reduce 558) (("(") reduce 558) ((")") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 17190 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 17191 ((("%") reduce 384) (("(") reduce 384) ((")") reduce 384) (("*") reduce 384) (("+") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384)) 17193 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 17194 ((("%") reduce 386) (("(") reduce 386) ((")") reduce 386) (("*") reduce 386) (("+") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386)) 17202 ((("%") reduce 383) (("(") reduce 383) ((")") reduce 383) (("*") reduce 383) (("+") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383)) 17204 ((("%") reduce 496) (("(") reduce 496) ((")") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 17210 ((("%") reduce 231) ((")") reduce 231) (("*") reduce 231) (("+") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231)) 17215 ((("%") shift) ((")") reduce 389) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17219 ((("%") reduce 596) ((")") reduce 596) (("*") reduce 596) (("+") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596)) 17220 ((("%") reduce 450) ((")") reduce 450) (("*") reduce 450) (("+") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450)) 17221 ((("%") shift) ((")") reduce 599) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17222 ((("%") shift) ((")") reduce 598) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17223 ((("%") reduce 408) ((")") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408)) 17225 ((("%") reduce 501) ((")") reduce 501) (("*") reduce 501) (("+") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501)) 17226 ((("%") reduce 457) ((")") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457)) 17227 ((("%") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289)) 17231 ((("%") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290)) 17232 ((("%") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 17235 ((("%") reduce 600) ((")") reduce 600) (("*") reduce 600) (("+") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600)) 17238 ((("%") reduce 597) ((")") reduce 597) (("*") reduce 597) (("+") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597)) 17239 ((("%") reduce 451) ((")") reduce 451) (("*") reduce 451) (("+") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451)) 17240 ((("%") reduce 410) ((")") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410)) 17241 ((("%") reduce 393) ((")") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 17242 ((("%") reduce 455) ((")") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455)) 17243 ((("%") reduce 453) ((")") reduce 453) (("*") reduce 453) (("+") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453)) 17246 ((("%") reduce 502) (("(") reduce 502) ((")") reduce 502) (("*") reduce 502) (("+") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502)) 17247 ((("%") reduce 458) ((")") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 17248 ((("%") reduce 430) ((")") reduce 430) (("*") reduce 430) (("+") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430)) 17249 ((("%") reduce 429) ((")") reduce 429) (("*") reduce 429) (("+") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429)) 17251 ((("%") reduce 443) (("(") reduce 539) ((")") reduce 443) (("*") reduce 443) (("+") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443)) 17252 ((("%") reduce 553) (("(") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 17255 ((("%") reduce 552) (("(") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 17257 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 17258 ((("%") reduce 431) ((")") reduce 431) (("*") reduce 431) (("+") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431)) 17260 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 17270 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 17278 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 17281 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 17283 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("}") reduce 243) (("~") reduce 243)) 17286 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("}") reduce 146) (("~") reduce 146)) 17288 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("}") reduce 245) (("~") reduce 245)) 17289 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 17293 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_ELSE) reduce 248) ((T_ELSEIF) reduce 248) ((T_EMPTY) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("}") reduce 248) (("~") reduce 248)) 17297 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_ELSE) reduce 141) ((T_ELSEIF) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDIF) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 17298 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_ELSE) reduce 247) ((T_ELSEIF) reduce 247) ((T_EMPTY) shift) ((T_ENDIF) reduce 247) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 17299 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDIF) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 17313 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDIF) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 17323 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDIF) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 17325 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_ELSE) reduce 143) ((T_ELSEIF) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDIF) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 17327 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_ELSE) reduce 104) ((T_ELSEIF) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDIF) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 17329 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_ELSE) reduce 107) ((T_ELSEIF) reduce 107) ((T_EMPTY) reduce 107) ((T_ENDIF) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 17332 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_ELSE) reduce 103) ((T_ELSEIF) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDIF) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 17333 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDIF) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 17334 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 244) ((T_ENDIF) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 17335 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_ELSE) reduce 140) ((T_ELSEIF) reduce 140) ((T_EMPTY) reduce 140) ((T_ENDIF) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_PROPERTY_C) reduce 140) ((T_READONLY) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("~") reduce 140)) 17336 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_ELSE) reduce 106) ((T_ELSEIF) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDIF) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 17337 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_ELSE) reduce 142) ((T_ELSEIF) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDIF) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 17338 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_ELSE) reduce 105) ((T_ELSEIF) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDIF) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 17341 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDIF) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 17347 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDIF) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 17348 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDIF) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 17349 ((("(") reduce 80) ((T_READONLY) reduce 80) ((T_STRING) reduce 80)) 17350 ((("(") reduce 81) ((T_READONLY) reduce 81) ((T_STRING) reduce 81)) 17351 ((("(") reduce 480) ((T_READONLY) reduce 480) ((T_STRING) reduce 480)) 17352 ((("(") reduce 476) ((T_READONLY) shift) ((T_STRING) shift)) 17353 ((("(") reduce 181)) 17354 ((("(") reduce 180)) 17364 ((("!") reduce 477) (("\"") reduce 477) (("$") reduce 477) (("(") reduce 477) (("+") reduce 477) (("-") reduce 477) ((";") reduce 477) (("@") reduce 477) ((T_ABSTRACT) reduce 477) ((T_ARRAY) reduce 477) ((T_ARRAY_CAST) reduce 477) ((T_ATTRIBUTE) reduce 477) ((T_BOOL_CAST) reduce 477) ((T_BREAK) reduce 477) ((T_CLASS) reduce 477) ((T_CLASS_C) reduce 477) ((T_CLONE) reduce 477) ((T_CONSTANT_ENCAPSED_STRING) reduce 477) ((T_CONTINUE) reduce 477) ((T_DEC) reduce 477) ((T_DECLARE) reduce 477) ((T_DIR) reduce 477) ((T_DNUMBER) reduce 477) ((T_DO) reduce 477) ((T_DOUBLE_CAST) reduce 477) ((T_ECHO) reduce 477) ((T_ELSE) reduce 477) ((T_ELSEIF) reduce 477) ((T_EMPTY) reduce 477) ((T_ENDIF) reduce 477) ((T_ENUM) reduce 477) ((T_EVAL) reduce 477) ((T_EXIT) reduce 477) ((T_FILE) reduce 477) ((T_FINAL) reduce 477) ((T_FN) reduce 477) ((T_FOR) reduce 477) ((T_FOREACH) reduce 477) ((T_FUNCTION) reduce 477) ((T_FUNC_C) reduce 477) ((T_GLOBAL) reduce 477) ((T_GOTO) reduce 477) ((T_HALT_COMPILER) reduce 477) ((T_IF) reduce 477) ((T_INC) reduce 477) ((T_INCLUDE) reduce 477) ((T_INCLUDE_ONCE) reduce 477) ((T_INLINE_HTML) reduce 477) ((T_INTERFACE) reduce 477) ((T_INT_CAST) reduce 477) ((T_ISSET) reduce 477) ((T_LINE) reduce 477) ((T_LIST) reduce 477) ((T_LNUMBER) reduce 477) ((T_MATCH) reduce 477) ((T_METHOD_C) reduce 477) ((T_NAME_FULLY_QUALIFIED) reduce 477) ((T_NAME_QUALIFIED) reduce 477) ((T_NAME_RELATIVE) reduce 477) ((T_NEW) reduce 477) ((T_NS_C) reduce 477) ((T_OBJECT_CAST) reduce 477) ((T_PRINT) reduce 477) ((T_PROPERTY_C) reduce 477) ((T_READONLY) reduce 477) ((T_REQUIRE) reduce 477) ((T_REQUIRE_ONCE) reduce 477) ((T_RETURN) reduce 477) ((T_START_HEREDOC) reduce 477) ((T_STATIC) reduce 477) ((T_STRING) reduce 477) ((T_STRING_CAST) reduce 477) ((T_SWITCH) reduce 477) ((T_THROW) reduce 477) ((T_TRAIT) reduce 477) ((T_TRAIT_C) reduce 477) ((T_TRY) reduce 477) ((T_UNSET) reduce 477) ((T_UNSET_CAST) reduce 477) ((T_VARIABLE) reduce 477) ((T_WHILE) reduce 477) ((T_YIELD) reduce 477) ((T_YIELD_FROM) reduce 477) (("[") reduce 477) (("`") reduce 477) (("{") reduce 477) (("~") reduce 477)) 17365 ((("!") reduce 182) (("\"") reduce 182) (("$") reduce 182) (("(") reduce 182) (("+") reduce 182) (("-") reduce 182) ((";") reduce 182) (("@") reduce 182) ((T_ABSTRACT) reduce 182) ((T_ARRAY) reduce 182) ((T_ARRAY_CAST) reduce 182) ((T_ATTRIBUTE) reduce 182) ((T_BOOL_CAST) reduce 182) ((T_BREAK) reduce 182) ((T_CLASS) reduce 182) ((T_CLASS_C) reduce 182) ((T_CLONE) reduce 182) ((T_CONSTANT_ENCAPSED_STRING) reduce 182) ((T_CONTINUE) reduce 182) ((T_DEC) reduce 182) ((T_DECLARE) reduce 182) ((T_DIR) reduce 182) ((T_DNUMBER) reduce 182) ((T_DO) reduce 182) ((T_DOUBLE_CAST) reduce 182) ((T_ECHO) reduce 182) ((T_ELSE) reduce 182) ((T_ELSEIF) reduce 182) ((T_EMPTY) reduce 182) ((T_ENDIF) reduce 182) ((T_ENUM) reduce 182) ((T_EVAL) reduce 182) ((T_EXIT) reduce 182) ((T_FILE) reduce 182) ((T_FINAL) reduce 182) ((T_FN) reduce 182) ((T_FOR) reduce 182) ((T_FOREACH) reduce 182) ((T_FUNCTION) reduce 182) ((T_FUNC_C) reduce 182) ((T_GLOBAL) reduce 182) ((T_GOTO) reduce 182) ((T_HALT_COMPILER) reduce 182) ((T_IF) reduce 182) ((T_INC) reduce 182) ((T_INCLUDE) reduce 182) ((T_INCLUDE_ONCE) reduce 182) ((T_INLINE_HTML) reduce 182) ((T_INTERFACE) reduce 182) ((T_INT_CAST) reduce 182) ((T_ISSET) reduce 182) ((T_LINE) reduce 182) ((T_LIST) reduce 182) ((T_LNUMBER) reduce 182) ((T_MATCH) reduce 182) ((T_METHOD_C) reduce 182) ((T_NAME_FULLY_QUALIFIED) reduce 182) ((T_NAME_QUALIFIED) reduce 182) ((T_NAME_RELATIVE) reduce 182) ((T_NEW) reduce 182) ((T_NS_C) reduce 182) ((T_OBJECT_CAST) reduce 182) ((T_PRINT) reduce 182) ((T_PROPERTY_C) reduce 182) ((T_READONLY) reduce 182) ((T_REQUIRE) reduce 182) ((T_REQUIRE_ONCE) reduce 182) ((T_RETURN) reduce 182) ((T_START_HEREDOC) reduce 182) ((T_STATIC) reduce 182) ((T_STRING) reduce 182) ((T_STRING_CAST) reduce 182) ((T_SWITCH) reduce 182) ((T_THROW) reduce 182) ((T_TRAIT) reduce 182) ((T_TRAIT_C) reduce 182) ((T_TRY) reduce 182) ((T_UNSET) reduce 182) ((T_UNSET_CAST) reduce 182) ((T_VARIABLE) reduce 182) ((T_WHILE) reduce 182) ((T_YIELD) reduce 182) ((T_YIELD_FROM) reduce 182) (("[") reduce 182) (("`") reduce 182) (("{") reduce 182) (("~") reduce 182)) 17366 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDIF) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 17368 (((T_ABSTRACT) reduce 190) ((T_CLASS) reduce 190) ((T_FINAL) reduce 190) ((T_READONLY) reduce 190)) 17375 ((("!") reduce 187) (("\"") reduce 187) (("$") reduce 187) (("(") reduce 187) (("+") reduce 187) (("-") reduce 187) ((";") reduce 187) (("@") reduce 187) ((T_ABSTRACT) reduce 187) ((T_ARRAY) reduce 187) ((T_ARRAY_CAST) reduce 187) ((T_ATTRIBUTE) reduce 187) ((T_BOOL_CAST) reduce 187) ((T_BREAK) reduce 187) ((T_CLASS) reduce 187) ((T_CLASS_C) reduce 187) ((T_CLONE) reduce 187) ((T_CONSTANT_ENCAPSED_STRING) reduce 187) ((T_CONTINUE) reduce 187) ((T_DEC) reduce 187) ((T_DECLARE) reduce 187) ((T_DIR) reduce 187) ((T_DNUMBER) reduce 187) ((T_DO) reduce 187) ((T_DOUBLE_CAST) reduce 187) ((T_ECHO) reduce 187) ((T_ELSE) reduce 187) ((T_ELSEIF) reduce 187) ((T_EMPTY) reduce 187) ((T_ENDIF) reduce 187) ((T_ENUM) reduce 187) ((T_EVAL) reduce 187) ((T_EXIT) reduce 187) ((T_FILE) reduce 187) ((T_FINAL) reduce 187) ((T_FN) reduce 187) ((T_FOR) reduce 187) ((T_FOREACH) reduce 187) ((T_FUNCTION) reduce 187) ((T_FUNC_C) reduce 187) ((T_GLOBAL) reduce 187) ((T_GOTO) reduce 187) ((T_HALT_COMPILER) reduce 187) ((T_IF) reduce 187) ((T_INC) reduce 187) ((T_INCLUDE) reduce 187) ((T_INCLUDE_ONCE) reduce 187) ((T_INLINE_HTML) reduce 187) ((T_INTERFACE) reduce 187) ((T_INT_CAST) reduce 187) ((T_ISSET) reduce 187) ((T_LINE) reduce 187) ((T_LIST) reduce 187) ((T_LNUMBER) reduce 187) ((T_MATCH) reduce 187) ((T_METHOD_C) reduce 187) ((T_NAME_FULLY_QUALIFIED) reduce 187) ((T_NAME_QUALIFIED) reduce 187) ((T_NAME_RELATIVE) reduce 187) ((T_NEW) reduce 187) ((T_NS_C) reduce 187) ((T_OBJECT_CAST) reduce 187) ((T_PRINT) reduce 187) ((T_PROPERTY_C) reduce 187) ((T_READONLY) reduce 187) ((T_REQUIRE) reduce 187) ((T_REQUIRE_ONCE) reduce 187) ((T_RETURN) reduce 187) ((T_START_HEREDOC) reduce 187) ((T_STATIC) reduce 187) ((T_STRING) reduce 187) ((T_STRING_CAST) reduce 187) ((T_SWITCH) reduce 187) ((T_THROW) reduce 187) ((T_TRAIT) reduce 187) ((T_TRAIT_C) reduce 187) ((T_TRY) reduce 187) ((T_UNSET) reduce 187) ((T_UNSET_CAST) reduce 187) ((T_VARIABLE) reduce 187) ((T_WHILE) reduce 187) ((T_YIELD) reduce 187) ((T_YIELD_FROM) reduce 187) (("[") reduce 187) (("`") reduce 187) (("{") reduce 187) (("~") reduce 187)) 17377 (((T_ABSTRACT) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_CLASS) reduce 102) ((T_ENUM) reduce 102) ((T_FINAL) reduce 102) ((T_FN) reduce 102) ((T_FUNCTION) reduce 102) ((T_INTERFACE) reduce 102) ((T_READONLY) reduce 102) ((T_STATIC) reduce 102) ((T_TRAIT) reduce 102)) 17378 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_ELSE) reduce 144) ((T_ELSEIF) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDIF) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 17379 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 17380 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 17383 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_ELSE) reduce 248) ((T_ELSEIF) reduce 248) ((T_EMPTY) reduce 248) ((T_ENDIF) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("~") reduce 248)) 17384 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDIF) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 17385 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDIF) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 17386 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDIF) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 17400 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDIF) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 17410 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDIF) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 17412 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDIF) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 17414 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDIF) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 17416 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENDIF) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 17419 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDIF) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 17420 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDIF) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 17422 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENDIF) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_PROPERTY_C) reduce 140) ((T_READONLY) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("~") reduce 140)) 17423 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDIF) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 17424 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDIF) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 17425 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDIF) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 17428 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDIF) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 17430 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDIF) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 17441 ((("!") reduce 477) (("\"") reduce 477) (("$") reduce 477) (("(") reduce 477) (("+") reduce 477) (("-") reduce 477) ((";") reduce 477) (("@") reduce 477) ((T_ABSTRACT) reduce 477) ((T_ARRAY) reduce 477) ((T_ARRAY_CAST) reduce 477) ((T_ATTRIBUTE) reduce 477) ((T_BOOL_CAST) reduce 477) ((T_BREAK) reduce 477) ((T_CLASS) reduce 477) ((T_CLASS_C) reduce 477) ((T_CLONE) reduce 477) ((T_CONSTANT_ENCAPSED_STRING) reduce 477) ((T_CONTINUE) reduce 477) ((T_DEC) reduce 477) ((T_DECLARE) reduce 477) ((T_DIR) reduce 477) ((T_DNUMBER) reduce 477) ((T_DO) reduce 477) ((T_DOUBLE_CAST) reduce 477) ((T_ECHO) reduce 477) ((T_EMPTY) reduce 477) ((T_ENDIF) reduce 477) ((T_ENUM) reduce 477) ((T_EVAL) reduce 477) ((T_EXIT) reduce 477) ((T_FILE) reduce 477) ((T_FINAL) reduce 477) ((T_FN) reduce 477) ((T_FOR) reduce 477) ((T_FOREACH) reduce 477) ((T_FUNCTION) reduce 477) ((T_FUNC_C) reduce 477) ((T_GLOBAL) reduce 477) ((T_GOTO) reduce 477) ((T_HALT_COMPILER) reduce 477) ((T_IF) reduce 477) ((T_INC) reduce 477) ((T_INCLUDE) reduce 477) ((T_INCLUDE_ONCE) reduce 477) ((T_INLINE_HTML) reduce 477) ((T_INTERFACE) reduce 477) ((T_INT_CAST) reduce 477) ((T_ISSET) reduce 477) ((T_LINE) reduce 477) ((T_LIST) reduce 477) ((T_LNUMBER) reduce 477) ((T_MATCH) reduce 477) ((T_METHOD_C) reduce 477) ((T_NAME_FULLY_QUALIFIED) reduce 477) ((T_NAME_QUALIFIED) reduce 477) ((T_NAME_RELATIVE) reduce 477) ((T_NEW) reduce 477) ((T_NS_C) reduce 477) ((T_OBJECT_CAST) reduce 477) ((T_PRINT) reduce 477) ((T_PROPERTY_C) reduce 477) ((T_READONLY) reduce 477) ((T_REQUIRE) reduce 477) ((T_REQUIRE_ONCE) reduce 477) ((T_RETURN) reduce 477) ((T_START_HEREDOC) reduce 477) ((T_STATIC) reduce 477) ((T_STRING) reduce 477) ((T_STRING_CAST) reduce 477) ((T_SWITCH) reduce 477) ((T_THROW) reduce 477) ((T_TRAIT) reduce 477) ((T_TRAIT_C) reduce 477) ((T_TRY) reduce 477) ((T_UNSET) reduce 477) ((T_UNSET_CAST) reduce 477) ((T_VARIABLE) reduce 477) ((T_WHILE) reduce 477) ((T_YIELD) reduce 477) ((T_YIELD_FROM) reduce 477) (("[") reduce 477) (("`") reduce 477) (("{") reduce 477) (("~") reduce 477)) 17442 ((("!") reduce 182) (("\"") reduce 182) (("$") reduce 182) (("(") reduce 182) (("+") reduce 182) (("-") reduce 182) ((";") reduce 182) (("@") reduce 182) ((T_ABSTRACT) reduce 182) ((T_ARRAY) reduce 182) ((T_ARRAY_CAST) reduce 182) ((T_ATTRIBUTE) reduce 182) ((T_BOOL_CAST) reduce 182) ((T_BREAK) reduce 182) ((T_CLASS) reduce 182) ((T_CLASS_C) reduce 182) ((T_CLONE) reduce 182) ((T_CONSTANT_ENCAPSED_STRING) reduce 182) ((T_CONTINUE) reduce 182) ((T_DEC) reduce 182) ((T_DECLARE) reduce 182) ((T_DIR) reduce 182) ((T_DNUMBER) reduce 182) ((T_DO) reduce 182) ((T_DOUBLE_CAST) reduce 182) ((T_ECHO) reduce 182) ((T_EMPTY) reduce 182) ((T_ENDIF) reduce 182) ((T_ENUM) reduce 182) ((T_EVAL) reduce 182) ((T_EXIT) reduce 182) ((T_FILE) reduce 182) ((T_FINAL) reduce 182) ((T_FN) reduce 182) ((T_FOR) reduce 182) ((T_FOREACH) reduce 182) ((T_FUNCTION) reduce 182) ((T_FUNC_C) reduce 182) ((T_GLOBAL) reduce 182) ((T_GOTO) reduce 182) ((T_HALT_COMPILER) reduce 182) ((T_IF) reduce 182) ((T_INC) reduce 182) ((T_INCLUDE) reduce 182) ((T_INCLUDE_ONCE) reduce 182) ((T_INLINE_HTML) reduce 182) ((T_INTERFACE) reduce 182) ((T_INT_CAST) reduce 182) ((T_ISSET) reduce 182) ((T_LINE) reduce 182) ((T_LIST) reduce 182) ((T_LNUMBER) reduce 182) ((T_MATCH) reduce 182) ((T_METHOD_C) reduce 182) ((T_NAME_FULLY_QUALIFIED) reduce 182) ((T_NAME_QUALIFIED) reduce 182) ((T_NAME_RELATIVE) reduce 182) ((T_NEW) reduce 182) ((T_NS_C) reduce 182) ((T_OBJECT_CAST) reduce 182) ((T_PRINT) reduce 182) ((T_PROPERTY_C) reduce 182) ((T_READONLY) reduce 182) ((T_REQUIRE) reduce 182) ((T_REQUIRE_ONCE) reduce 182) ((T_RETURN) reduce 182) ((T_START_HEREDOC) reduce 182) ((T_STATIC) reduce 182) ((T_STRING) reduce 182) ((T_STRING_CAST) reduce 182) ((T_SWITCH) reduce 182) ((T_THROW) reduce 182) ((T_TRAIT) reduce 182) ((T_TRAIT_C) reduce 182) ((T_TRY) reduce 182) ((T_UNSET) reduce 182) ((T_UNSET_CAST) reduce 182) ((T_VARIABLE) reduce 182) ((T_WHILE) reduce 182) ((T_YIELD) reduce 182) ((T_YIELD_FROM) reduce 182) (("[") reduce 182) (("`") reduce 182) (("{") reduce 182) (("~") reduce 182)) 17443 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDIF) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 17451 ((("!") reduce 187) (("\"") reduce 187) (("$") reduce 187) (("(") reduce 187) (("+") reduce 187) (("-") reduce 187) ((";") reduce 187) (("@") reduce 187) ((T_ABSTRACT) reduce 187) ((T_ARRAY) reduce 187) ((T_ARRAY_CAST) reduce 187) ((T_ATTRIBUTE) reduce 187) ((T_BOOL_CAST) reduce 187) ((T_BREAK) reduce 187) ((T_CLASS) reduce 187) ((T_CLASS_C) reduce 187) ((T_CLONE) reduce 187) ((T_CONSTANT_ENCAPSED_STRING) reduce 187) ((T_CONTINUE) reduce 187) ((T_DEC) reduce 187) ((T_DECLARE) reduce 187) ((T_DIR) reduce 187) ((T_DNUMBER) reduce 187) ((T_DO) reduce 187) ((T_DOUBLE_CAST) reduce 187) ((T_ECHO) reduce 187) ((T_EMPTY) reduce 187) ((T_ENDIF) reduce 187) ((T_ENUM) reduce 187) ((T_EVAL) reduce 187) ((T_EXIT) reduce 187) ((T_FILE) reduce 187) ((T_FINAL) reduce 187) ((T_FN) reduce 187) ((T_FOR) reduce 187) ((T_FOREACH) reduce 187) ((T_FUNCTION) reduce 187) ((T_FUNC_C) reduce 187) ((T_GLOBAL) reduce 187) ((T_GOTO) reduce 187) ((T_HALT_COMPILER) reduce 187) ((T_IF) reduce 187) ((T_INC) reduce 187) ((T_INCLUDE) reduce 187) ((T_INCLUDE_ONCE) reduce 187) ((T_INLINE_HTML) reduce 187) ((T_INTERFACE) reduce 187) ((T_INT_CAST) reduce 187) ((T_ISSET) reduce 187) ((T_LINE) reduce 187) ((T_LIST) reduce 187) ((T_LNUMBER) reduce 187) ((T_MATCH) reduce 187) ((T_METHOD_C) reduce 187) ((T_NAME_FULLY_QUALIFIED) reduce 187) ((T_NAME_QUALIFIED) reduce 187) ((T_NAME_RELATIVE) reduce 187) ((T_NEW) reduce 187) ((T_NS_C) reduce 187) ((T_OBJECT_CAST) reduce 187) ((T_PRINT) reduce 187) ((T_PROPERTY_C) reduce 187) ((T_READONLY) reduce 187) ((T_REQUIRE) reduce 187) ((T_REQUIRE_ONCE) reduce 187) ((T_RETURN) reduce 187) ((T_START_HEREDOC) reduce 187) ((T_STATIC) reduce 187) ((T_STRING) reduce 187) ((T_STRING_CAST) reduce 187) ((T_SWITCH) reduce 187) ((T_THROW) reduce 187) ((T_TRAIT) reduce 187) ((T_TRAIT_C) reduce 187) ((T_TRY) reduce 187) ((T_UNSET) reduce 187) ((T_UNSET_CAST) reduce 187) ((T_VARIABLE) reduce 187) ((T_WHILE) reduce 187) ((T_YIELD) reduce 187) ((T_YIELD_FROM) reduce 187) (("[") reduce 187) (("`") reduce 187) (("{") reduce 187) (("~") reduce 187)) 17452 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDIF) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 17455 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_EMPTY) reduce 248) ((T_ENDIF) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("~") reduce 248)) 17459 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_EMPTY) reduce 249) ((T_ENDIF) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("~") reduce 249)) 17462 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") shift) ((";") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 17463 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDWHILE) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 17464 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDIF) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 17465 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDIF) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 17466 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 17467 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDWHILE) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 17482 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDWHILE) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 17492 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDWHILE) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 17494 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDWHILE) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 17496 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDWHILE) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 17498 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENDWHILE) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 17501 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDWHILE) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 17502 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDWHILE) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 17503 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 244) ((T_ENDWHILE) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 17504 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENDWHILE) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_PROPERTY_C) reduce 140) ((T_READONLY) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("~") reduce 140)) 17505 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDWHILE) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 17506 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDWHILE) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 17507 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDWHILE) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 17510 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDWHILE) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 17516 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDWHILE) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 17526 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDWHILE) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 17534 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDWHILE) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 17537 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDWHILE) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 17539 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDWHILE) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 17542 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDWHILE) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 17544 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDWHILE) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 17545 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDWHILE) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 17548 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_ELSE) reduce 248) ((T_ELSEIF) reduce 248) ((T_EMPTY) reduce 248) ((T_ENDWHILE) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("~") reduce 248)) 17552 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_ELSE) reduce 249) ((T_ELSEIF) reduce 249) ((T_EMPTY) reduce 249) ((T_ENDWHILE) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("~") reduce 249)) 17557 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDWHILE) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 17558 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDWHILE) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 17561 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDWHILE) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 17574 (((")") reduce 177) ((",") reduce 177)) 17576 (((")") reduce 179) ((",") reduce 179) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) (("[") reduce 531)) 17589 ((("\"") shift) (("$") shift) (("(") shift) ((")") reduce 125) ((T_ARRAY) shift) ((T_CLASS_C) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DIR) shift) ((T_FILE) shift) ((T_FUNC_C) shift) ((T_LINE) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_TRAIT_C) shift) ((T_VARIABLE) shift) (("[") shift)) 17592 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDWHILE) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 17593 (((")") reduce 178) ((",") reduce 178)) 17621 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDWHILE) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 17622 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDWHILE) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 17625 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDWHILE) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 17628 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_ELSE) reduce 176) ((T_ELSEIF) reduce 176) ((T_EMPTY) reduce 176) ((T_ENDWHILE) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("~") reduce 176)) 17630 (((")") reduce 94) ((T_VARIABLE) reduce 94) (("|") reduce 94)) 17631 (((")") reduce 93) ((T_VARIABLE) reduce 93) (("|") reduce 93)) 17632 (((")") reduce 95) ((T_VARIABLE) reduce 95) (("|") reduce 95)) 17633 (((")") reduce 492) ((T_VARIABLE) reduce 492) (("|") reduce 492)) 17634 (((")") reduce 92) ((T_VARIABLE) reduce 92) (("|") reduce 92)) 17635 (((")") reduce 173) ((T_VARIABLE) shift) (("|") shift)) 17636 (((")") reduce 171) ((T_VARIABLE) reduce 171) (("|") reduce 171)) 17637 (((")") reduce 493) ((T_VARIABLE) reduce 493) (("|") reduce 493)) 17638 (((")") reduce 174)) 17641 (((")") reduce 172) ((T_VARIABLE) reduce 172) (("|") reduce 172)) 17645 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDWHILE) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 17648 (((":") shift) (("{") shift)) 17649 (((";") shift) ((T_CASE) reduce 226) ((T_DEFAULT) reduce 226) ((T_ENDSWITCH) reduce 226)) 17650 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDWHILE) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 17651 (((";") shift) ((T_CASE) reduce 226) ((T_DEFAULT) reduce 226) (("}") reduce 226)) 17652 (((T_CASE) reduce 226) ((T_DEFAULT) reduce 226) (("}") reduce 226)) 17653 (((T_CASE) shift) ((T_DEFAULT) shift) (("}") shift)) 17655 (((":") shift) ((";") shift)) 17656 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDWHILE) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 17657 ((("!") reduce 229) (("\"") reduce 229) (("$") reduce 229) (("(") reduce 229) (("+") reduce 229) (("-") reduce 229) ((";") reduce 229) (("@") reduce 229) ((T_ABSTRACT) reduce 229) ((T_ARRAY) reduce 229) ((T_ARRAY_CAST) reduce 229) ((T_ATTRIBUTE) reduce 229) ((T_BOOL_CAST) reduce 229) ((T_BREAK) reduce 229) ((T_CASE) reduce 229) ((T_CLASS) reduce 229) ((T_CLASS_C) reduce 229) ((T_CLONE) reduce 229) ((T_CONSTANT_ENCAPSED_STRING) reduce 229) ((T_CONTINUE) reduce 229) ((T_DEC) reduce 229) ((T_DECLARE) reduce 229) ((T_DEFAULT) reduce 229) ((T_DIR) reduce 229) ((T_DNUMBER) reduce 229) ((T_DO) reduce 229) ((T_DOUBLE_CAST) reduce 229) ((T_ECHO) reduce 229) ((T_EMPTY) reduce 229) ((T_ENUM) reduce 229) ((T_EVAL) reduce 229) ((T_EXIT) reduce 229) ((T_FILE) reduce 229) ((T_FINAL) reduce 229) ((T_FN) reduce 229) ((T_FOR) reduce 229) ((T_FOREACH) reduce 229) ((T_FUNCTION) reduce 229) ((T_FUNC_C) reduce 229) ((T_GLOBAL) reduce 229) ((T_GOTO) reduce 229) ((T_HALT_COMPILER) reduce 229) ((T_IF) reduce 229) ((T_INC) reduce 229) ((T_INCLUDE) reduce 229) ((T_INCLUDE_ONCE) reduce 229) ((T_INLINE_HTML) reduce 229) ((T_INTERFACE) reduce 229) ((T_INT_CAST) reduce 229) ((T_ISSET) reduce 229) ((T_LINE) reduce 229) ((T_LIST) reduce 229) ((T_LNUMBER) reduce 229) ((T_MATCH) reduce 229) ((T_METHOD_C) reduce 229) ((T_NAME_FULLY_QUALIFIED) reduce 229) ((T_NAME_QUALIFIED) reduce 229) ((T_NAME_RELATIVE) reduce 229) ((T_NEW) reduce 229) ((T_NS_C) reduce 229) ((T_OBJECT_CAST) reduce 229) ((T_PRINT) reduce 229) ((T_PROPERTY_C) reduce 229) ((T_READONLY) reduce 229) ((T_REQUIRE) reduce 229) ((T_REQUIRE_ONCE) reduce 229) ((T_RETURN) reduce 229) ((T_START_HEREDOC) reduce 229) ((T_STATIC) reduce 229) ((T_STRING) reduce 229) ((T_STRING_CAST) reduce 229) ((T_SWITCH) reduce 229) ((T_THROW) reduce 229) ((T_TRAIT) reduce 229) ((T_TRAIT_C) reduce 229) ((T_TRY) reduce 229) ((T_UNSET) reduce 229) ((T_UNSET_CAST) reduce 229) ((T_VARIABLE) reduce 229) ((T_WHILE) reduce 229) ((T_YIELD) reduce 229) ((T_YIELD_FROM) reduce 229) (("[") reduce 229) (("`") reduce 229) (("{") reduce 229) (("}") reduce 229) (("~") reduce 229)) 17658 ((("!") reduce 230) (("\"") reduce 230) (("$") reduce 230) (("(") reduce 230) (("+") reduce 230) (("-") reduce 230) ((";") reduce 230) (("@") reduce 230) ((T_ABSTRACT) reduce 230) ((T_ARRAY) reduce 230) ((T_ARRAY_CAST) reduce 230) ((T_ATTRIBUTE) reduce 230) ((T_BOOL_CAST) reduce 230) ((T_BREAK) reduce 230) ((T_CASE) reduce 230) ((T_CLASS) reduce 230) ((T_CLASS_C) reduce 230) ((T_CLONE) reduce 230) ((T_CONSTANT_ENCAPSED_STRING) reduce 230) ((T_CONTINUE) reduce 230) ((T_DEC) reduce 230) ((T_DECLARE) reduce 230) ((T_DEFAULT) reduce 230) ((T_DIR) reduce 230) ((T_DNUMBER) reduce 230) ((T_DO) reduce 230) ((T_DOUBLE_CAST) reduce 230) ((T_ECHO) reduce 230) ((T_EMPTY) reduce 230) ((T_ENUM) reduce 230) ((T_EVAL) reduce 230) ((T_EXIT) reduce 230) ((T_FILE) reduce 230) ((T_FINAL) reduce 230) ((T_FN) reduce 230) ((T_FOR) reduce 230) ((T_FOREACH) reduce 230) ((T_FUNCTION) reduce 230) ((T_FUNC_C) reduce 230) ((T_GLOBAL) reduce 230) ((T_GOTO) reduce 230) ((T_HALT_COMPILER) reduce 230) ((T_IF) reduce 230) ((T_INC) reduce 230) ((T_INCLUDE) reduce 230) ((T_INCLUDE_ONCE) reduce 230) ((T_INLINE_HTML) reduce 230) ((T_INTERFACE) reduce 230) ((T_INT_CAST) reduce 230) ((T_ISSET) reduce 230) ((T_LINE) reduce 230) ((T_LIST) reduce 230) ((T_LNUMBER) reduce 230) ((T_MATCH) reduce 230) ((T_METHOD_C) reduce 230) ((T_NAME_FULLY_QUALIFIED) reduce 230) ((T_NAME_QUALIFIED) reduce 230) ((T_NAME_RELATIVE) reduce 230) ((T_NEW) reduce 230) ((T_NS_C) reduce 230) ((T_OBJECT_CAST) reduce 230) ((T_PRINT) reduce 230) ((T_PROPERTY_C) reduce 230) ((T_READONLY) reduce 230) ((T_REQUIRE) reduce 230) ((T_REQUIRE_ONCE) reduce 230) ((T_RETURN) reduce 230) ((T_START_HEREDOC) reduce 230) ((T_STATIC) reduce 230) ((T_STRING) reduce 230) ((T_STRING_CAST) reduce 230) ((T_SWITCH) reduce 230) ((T_THROW) reduce 230) ((T_TRAIT) reduce 230) ((T_TRAIT_C) reduce 230) ((T_TRY) reduce 230) ((T_UNSET) reduce 230) ((T_UNSET_CAST) reduce 230) ((T_VARIABLE) reduce 230) ((T_WHILE) reduce 230) ((T_YIELD) reduce 230) ((T_YIELD_FROM) reduce 230) (("[") reduce 230) (("`") reduce 230) (("{") reduce 230) (("}") reduce 230) (("~") reduce 230)) 17659 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CASE) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DEFAULT) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("}") reduce 141) (("~") reduce 141)) 17660 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 228) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 228) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("}") reduce 228) (("~") shift)) 17661 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CASE) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DEFAULT) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 17675 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CASE) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DEFAULT) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 17685 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CASE) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DEFAULT) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 17687 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CASE) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DEFAULT) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("}") reduce 143) (("~") reduce 143)) 17689 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CASE) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DEFAULT) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("}") reduce 104) (("~") reduce 104)) 17691 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CASE) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DEFAULT) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("}") reduce 107) (("~") reduce 107)) 17694 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CASE) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DEFAULT) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("}") reduce 103) (("~") reduce 103)) 17695 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CASE) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DEFAULT) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 17696 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CASE) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DEFAULT) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("}") reduce 244) (("~") reduce 244)) 17697 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CASE) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DEFAULT) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_PROPERTY_C) reduce 140) ((T_READONLY) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("}") reduce 140) (("~") reduce 140)) 17698 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CASE) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DEFAULT) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("}") reduce 106) (("~") reduce 106)) 17699 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CASE) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DEFAULT) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("}") reduce 142) (("~") reduce 142)) 17700 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CASE) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DEFAULT) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("}") reduce 105) (("~") reduce 105)) 17703 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CASE) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DEFAULT) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("}") reduce 146) (("~") reduce 146)) 17709 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CASE) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DEFAULT) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 17719 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CASE) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DEFAULT) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 17727 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CASE) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DEFAULT) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 17730 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CASE) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DEFAULT) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 17732 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CASE) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DEFAULT) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("}") reduce 243) (("~") reduce 243)) 17735 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CASE) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DEFAULT) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("}") reduce 146) (("~") reduce 146)) 17737 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CASE) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DEFAULT) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("}") reduce 245) (("~") reduce 245)) 17738 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CASE) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DEFAULT) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 17741 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CASE) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DEFAULT) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_ELSE) reduce 248) ((T_ELSEIF) reduce 248) ((T_EMPTY) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("}") reduce 248) (("~") reduce 248)) 17745 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CASE) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DEFAULT) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_ELSE) reduce 249) ((T_ELSEIF) reduce 249) ((T_EMPTY) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("}") reduce 249) (("~") reduce 249)) 17750 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CASE) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DEFAULT) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("}") reduce 240) (("~") reduce 240)) 17751 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CASE) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DEFAULT) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 17754 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CASE) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DEFAULT) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("}") reduce 241) (("~") reduce 241)) 17759 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CASE) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DEFAULT) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 17762 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CASE) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DEFAULT) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 17763 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CASE) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DEFAULT) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("}") reduce 175) (("~") reduce 175)) 17766 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CASE) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DEFAULT) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 17769 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CASE) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DEFAULT) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_ELSE) reduce 176) ((T_ELSEIF) reduce 176) ((T_EMPTY) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("}") reduce 176) (("~") reduce 176)) 17776 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CASE) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DEFAULT) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("}") reduce 170) (("~") reduce 170)) 17781 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CASE) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DEFAULT) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 17785 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CASE) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DEFAULT) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("}") reduce 222) (("~") reduce 222)) 17787 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CASE) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DEFAULT) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("}") reduce 223) (("~") reduce 223)) 17788 (((T_CASE) reduce 226) ((T_DEFAULT) reduce 226) ((T_ENDSWITCH) reduce 226)) 17789 (((T_CASE) shift) ((T_DEFAULT) shift) ((T_ENDSWITCH) shift)) 17793 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CASE) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DEFAULT) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("}") reduce 224) (("~") reduce 224)) 17794 ((("!") reduce 229) (("\"") reduce 229) (("$") reduce 229) (("(") reduce 229) (("+") reduce 229) (("-") reduce 229) ((";") reduce 229) (("@") reduce 229) ((T_ABSTRACT) reduce 229) ((T_ARRAY) reduce 229) ((T_ARRAY_CAST) reduce 229) ((T_ATTRIBUTE) reduce 229) ((T_BOOL_CAST) reduce 229) ((T_BREAK) reduce 229) ((T_CASE) reduce 229) ((T_CLASS) reduce 229) ((T_CLASS_C) reduce 229) ((T_CLONE) reduce 229) ((T_CONSTANT_ENCAPSED_STRING) reduce 229) ((T_CONTINUE) reduce 229) ((T_DEC) reduce 229) ((T_DECLARE) reduce 229) ((T_DEFAULT) reduce 229) ((T_DIR) reduce 229) ((T_DNUMBER) reduce 229) ((T_DO) reduce 229) ((T_DOUBLE_CAST) reduce 229) ((T_ECHO) reduce 229) ((T_EMPTY) reduce 229) ((T_ENDSWITCH) reduce 229) ((T_ENUM) reduce 229) ((T_EVAL) reduce 229) ((T_EXIT) reduce 229) ((T_FILE) reduce 229) ((T_FINAL) reduce 229) ((T_FN) reduce 229) ((T_FOR) reduce 229) ((T_FOREACH) reduce 229) ((T_FUNCTION) reduce 229) ((T_FUNC_C) reduce 229) ((T_GLOBAL) reduce 229) ((T_GOTO) reduce 229) ((T_HALT_COMPILER) reduce 229) ((T_IF) reduce 229) ((T_INC) reduce 229) ((T_INCLUDE) reduce 229) ((T_INCLUDE_ONCE) reduce 229) ((T_INLINE_HTML) reduce 229) ((T_INTERFACE) reduce 229) ((T_INT_CAST) reduce 229) ((T_ISSET) reduce 229) ((T_LINE) reduce 229) ((T_LIST) reduce 229) ((T_LNUMBER) reduce 229) ((T_MATCH) reduce 229) ((T_METHOD_C) reduce 229) ((T_NAME_FULLY_QUALIFIED) reduce 229) ((T_NAME_QUALIFIED) reduce 229) ((T_NAME_RELATIVE) reduce 229) ((T_NEW) reduce 229) ((T_NS_C) reduce 229) ((T_OBJECT_CAST) reduce 229) ((T_PRINT) reduce 229) ((T_PROPERTY_C) reduce 229) ((T_READONLY) reduce 229) ((T_REQUIRE) reduce 229) ((T_REQUIRE_ONCE) reduce 229) ((T_RETURN) reduce 229) ((T_START_HEREDOC) reduce 229) ((T_STATIC) reduce 229) ((T_STRING) reduce 229) ((T_STRING_CAST) reduce 229) ((T_SWITCH) reduce 229) ((T_THROW) reduce 229) ((T_TRAIT) reduce 229) ((T_TRAIT_C) reduce 229) ((T_TRY) reduce 229) ((T_UNSET) reduce 229) ((T_UNSET_CAST) reduce 229) ((T_VARIABLE) reduce 229) ((T_WHILE) reduce 229) ((T_YIELD) reduce 229) ((T_YIELD_FROM) reduce 229) (("[") reduce 229) (("`") reduce 229) (("{") reduce 229) (("~") reduce 229)) 17795 ((("!") reduce 230) (("\"") reduce 230) (("$") reduce 230) (("(") reduce 230) (("+") reduce 230) (("-") reduce 230) ((";") reduce 230) (("@") reduce 230) ((T_ABSTRACT) reduce 230) ((T_ARRAY) reduce 230) ((T_ARRAY_CAST) reduce 230) ((T_ATTRIBUTE) reduce 230) ((T_BOOL_CAST) reduce 230) ((T_BREAK) reduce 230) ((T_CASE) reduce 230) ((T_CLASS) reduce 230) ((T_CLASS_C) reduce 230) ((T_CLONE) reduce 230) ((T_CONSTANT_ENCAPSED_STRING) reduce 230) ((T_CONTINUE) reduce 230) ((T_DEC) reduce 230) ((T_DECLARE) reduce 230) ((T_DEFAULT) reduce 230) ((T_DIR) reduce 230) ((T_DNUMBER) reduce 230) ((T_DO) reduce 230) ((T_DOUBLE_CAST) reduce 230) ((T_ECHO) reduce 230) ((T_EMPTY) reduce 230) ((T_ENDSWITCH) reduce 230) ((T_ENUM) reduce 230) ((T_EVAL) reduce 230) ((T_EXIT) reduce 230) ((T_FILE) reduce 230) ((T_FINAL) reduce 230) ((T_FN) reduce 230) ((T_FOR) reduce 230) ((T_FOREACH) reduce 230) ((T_FUNCTION) reduce 230) ((T_FUNC_C) reduce 230) ((T_GLOBAL) reduce 230) ((T_GOTO) reduce 230) ((T_HALT_COMPILER) reduce 230) ((T_IF) reduce 230) ((T_INC) reduce 230) ((T_INCLUDE) reduce 230) ((T_INCLUDE_ONCE) reduce 230) ((T_INLINE_HTML) reduce 230) ((T_INTERFACE) reduce 230) ((T_INT_CAST) reduce 230) ((T_ISSET) reduce 230) ((T_LINE) reduce 230) ((T_LIST) reduce 230) ((T_LNUMBER) reduce 230) ((T_MATCH) reduce 230) ((T_METHOD_C) reduce 230) ((T_NAME_FULLY_QUALIFIED) reduce 230) ((T_NAME_QUALIFIED) reduce 230) ((T_NAME_RELATIVE) reduce 230) ((T_NEW) reduce 230) ((T_NS_C) reduce 230) ((T_OBJECT_CAST) reduce 230) ((T_PRINT) reduce 230) ((T_PROPERTY_C) reduce 230) ((T_READONLY) reduce 230) ((T_REQUIRE) reduce 230) ((T_REQUIRE_ONCE) reduce 230) ((T_RETURN) reduce 230) ((T_START_HEREDOC) reduce 230) ((T_STATIC) reduce 230) ((T_STRING) reduce 230) ((T_STRING_CAST) reduce 230) ((T_SWITCH) reduce 230) ((T_THROW) reduce 230) ((T_TRAIT) reduce 230) ((T_TRAIT_C) reduce 230) ((T_TRY) reduce 230) ((T_UNSET) reduce 230) ((T_UNSET_CAST) reduce 230) ((T_VARIABLE) reduce 230) ((T_WHILE) reduce 230) ((T_YIELD) reduce 230) ((T_YIELD_FROM) reduce 230) (("[") reduce 230) (("`") reduce 230) (("{") reduce 230) (("~") reduce 230)) 17796 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CASE) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DEFAULT) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDSWITCH) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 17797 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 228) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 228) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDSWITCH) reduce 228) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 17798 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CASE) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DEFAULT) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDSWITCH) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 17812 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CASE) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DEFAULT) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDSWITCH) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 17822 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CASE) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DEFAULT) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDSWITCH) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 17824 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CASE) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DEFAULT) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDSWITCH) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 17826 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CASE) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DEFAULT) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDSWITCH) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 17828 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CASE) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DEFAULT) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENDSWITCH) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 17831 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CASE) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DEFAULT) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDSWITCH) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 17832 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CASE) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DEFAULT) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDSWITCH) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 17833 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CASE) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DEFAULT) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 244) ((T_ENDSWITCH) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 17834 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CASE) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DEFAULT) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENDSWITCH) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_PROPERTY_C) reduce 140) ((T_READONLY) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("~") reduce 140)) 17835 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CASE) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DEFAULT) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDSWITCH) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 17836 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CASE) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DEFAULT) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDSWITCH) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 17837 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CASE) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DEFAULT) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDSWITCH) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 17840 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CASE) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DEFAULT) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDSWITCH) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 17846 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CASE) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DEFAULT) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDSWITCH) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 17856 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CASE) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DEFAULT) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDSWITCH) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 17864 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CASE) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DEFAULT) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDSWITCH) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 17867 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CASE) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DEFAULT) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDSWITCH) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 17869 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CASE) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DEFAULT) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDSWITCH) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 17872 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CASE) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DEFAULT) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDSWITCH) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 17874 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CASE) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DEFAULT) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDSWITCH) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 17875 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CASE) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DEFAULT) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDSWITCH) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 17878 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CASE) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DEFAULT) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_ELSE) reduce 248) ((T_ELSEIF) reduce 248) ((T_EMPTY) reduce 248) ((T_ENDSWITCH) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("~") reduce 248)) 17882 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CASE) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DEFAULT) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_ELSE) reduce 249) ((T_ELSEIF) reduce 249) ((T_EMPTY) reduce 249) ((T_ENDSWITCH) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("~") reduce 249)) 17887 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CASE) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DEFAULT) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDSWITCH) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 17888 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CASE) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DEFAULT) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDSWITCH) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 17891 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CASE) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DEFAULT) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDSWITCH) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 17896 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CASE) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DEFAULT) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDSWITCH) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 17899 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CASE) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DEFAULT) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDSWITCH) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 17900 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CASE) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DEFAULT) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDSWITCH) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 17903 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CASE) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DEFAULT) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDSWITCH) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 17906 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CASE) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DEFAULT) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_ELSE) reduce 176) ((T_ELSEIF) reduce 176) ((T_EMPTY) reduce 176) ((T_ENDSWITCH) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("~") reduce 176)) 17913 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CASE) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DEFAULT) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDSWITCH) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 17918 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CASE) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DEFAULT) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDSWITCH) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 17922 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CASE) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DEFAULT) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDSWITCH) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 17924 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CASE) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DEFAULT) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDSWITCH) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 17928 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CASE) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DEFAULT) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDSWITCH) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 17931 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CASE) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DEFAULT) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDSWITCH) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 17932 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CASE) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DEFAULT) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDSWITCH) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 17933 (((",") reduce 302) ((";") reduce 302) (("=") shift)) 17934 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 17935 (((",") reduce 301) ((";") reduce 301)) 17938 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CASE) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DEFAULT) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDSWITCH) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 17939 (((",") reduce 300) ((";") reduce 300)) 18004 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 303) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 303) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18978 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 529) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18980 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CASE) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DEFAULT) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDSWITCH) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 18982 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CASE) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DEFAULT) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDSWITCH) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 18984 (((",") reduce 551) ((";") reduce 551)) 18985 (((",") reduce 298) ((";") reduce 298)) 18987 (((",") reduce 299) ((";") reduce 299)) 18989 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CASE) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DEFAULT) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDSWITCH) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 18990 (((",") reduce 297) ((";") reduce 297)) 18991 (((",") reduce 553) ((";") reduce 553)) 18994 (((",") reduce 552) ((";") reduce 552)) 19006 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AS) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 19008 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AS) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 19010 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AS) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 19011 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AS) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507)) 19015 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AS) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 19016 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AS) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 19017 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AS) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 19023 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AS) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 19025 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AS) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506)) 19027 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AS) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 19028 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 19029 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 19030 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AS) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 19032 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AS) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 19035 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AS) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 19041 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 19044 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AS) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 19046 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_AS) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 19047 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_AS) reduce 462) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 19054 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_AS) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 19055 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AS) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513)) 19057 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AS) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512)) 19058 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AS) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511)) 19059 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19062 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_AS) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 19063 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AS) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 19064 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AS) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449)) 19065 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AS) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 19066 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AS) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514)) 19067 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AS) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444)) 19068 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AS) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 19069 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AS) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 19070 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_AS) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 19071 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_AS) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 19072 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_AS) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388)) 19075 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AS) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432)) 19106 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416)) 19107 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419)) 19108 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AS) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428)) 19109 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AS) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441)) 19110 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AS) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 19111 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AS) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 19112 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 415) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19113 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 413) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19114 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 414) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19115 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AS) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438)) 19116 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AS) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 19117 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AS) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 19118 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AS) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 19119 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AS) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440)) 19120 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AS) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 19123 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 19124 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 19125 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AS) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 19126 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AS) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 19127 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 19128 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AS) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 19129 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AS) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 19130 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AS) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 19131 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AS) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 19132 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AS) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 19133 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AS) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 19139 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AS) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 19140 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AS) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 19141 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AS) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 19142 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AS) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 19143 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AS) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 19146 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AS) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 19147 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AS) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 19149 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AS) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 19151 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AS) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 19152 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AS) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 19155 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AS) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 19156 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 448) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19157 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 411) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19158 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19159 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AS) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417)) 19160 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AS) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418)) 19164 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 446) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19165 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 447) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19166 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AS) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439)) 19167 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AS) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437)) 19168 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AS) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 19169 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AS) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420)) 19170 ((("%") shift) (("*") shift) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AS) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 19171 ((("%") shift) (("*") shift) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AS) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 19172 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AS) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 19173 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AS) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 19175 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AS) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 19176 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AS) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 19177 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AS) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 19178 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AS) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 19179 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AS) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 19180 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AS) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 19181 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AS) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 19182 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AS) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 19183 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AS) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 19184 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AS) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 19185 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AS) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 19186 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AS) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 19187 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AS) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 19188 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AS) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 19189 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AS) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 19190 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AS) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 19191 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AS) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 19192 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AS) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 19193 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AS) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 19194 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AS) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 19195 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AS) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 19196 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AS) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 19197 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AS) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 19198 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AS) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 19199 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AS) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 19200 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AS) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 19201 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AS) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 19202 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AS) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 19203 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AS) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 19204 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AS) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 19205 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AS) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 19206 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AS) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 19207 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AS) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 19208 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AS) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 19209 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AS) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 19210 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AS) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 19211 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AS) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 19212 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AS) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 19213 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AS) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 19214 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AS) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 19215 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AS) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 19216 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AS) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 19217 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AS) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 19218 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AS) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 19219 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AS) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 19220 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AS) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 19221 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AS) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 19222 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AS) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 19223 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AS) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 19224 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AS) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 19225 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AS) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 19226 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AS) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 19227 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AS) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 19228 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AS) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 19229 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AS) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 19230 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AS) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 19231 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AS) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 19232 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AS) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 19233 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AS) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 19234 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AS) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 19235 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AS) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 19236 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AS) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 19237 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AS) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 19238 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AS) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 19239 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AS) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 19240 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AS) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 19241 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AS) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 19242 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AS) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 19243 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AS) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 19244 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AS) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 19245 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AS) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 19246 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AS) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 19247 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AS) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 19248 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AS) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 19249 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AS) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 19250 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AS) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 19251 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AS) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 19252 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AS) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 19253 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AS) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 19254 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AS) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525)) 19256 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AS) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 19257 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AS) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 19258 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_AS) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 19261 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AS) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527)) 19263 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_AS) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 19264 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_AS) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 19268 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_AS) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 19269 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_AS) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 19274 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AS) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 19276 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AS) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407)) 19286 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 403) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19287 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 405) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19288 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 404) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19289 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 397) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19290 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 394) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19291 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 402) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19292 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 396) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19293 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 400) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19294 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 395) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19295 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 398) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19296 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 399) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19297 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 406) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19298 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 401) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19300 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 391) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19303 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AS) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 19306 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AS) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 19308 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AS) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 19310 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AS) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 19311 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AS) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 19312 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AS) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392)) 19316 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AS) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 19318 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AS) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 19319 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AS) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 19323 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AS) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 19324 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AS) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 19325 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AS) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 19328 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AS) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 19329 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AS) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 19330 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AS) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 19335 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AS) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 19336 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AS) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 19337 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AS) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 19338 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AS) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 19341 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AS) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 19342 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AS) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 19343 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AS) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 19344 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AS) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 19345 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AS) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 19346 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AS) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 19349 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AS) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 19350 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_AS) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 19361 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AS) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 19362 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AS) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 19435 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 477) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19749 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AS) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 19752 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AS) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524)) 19754 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_AS) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 19757 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AS) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526)) 19758 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_AS) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 19759 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_AS) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 19761 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AS) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 19762 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AS) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 19767 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_AS) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 19768 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_AS) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 19769 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_AS) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 19770 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_AS) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 19773 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_AS) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 19774 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_AS) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 19775 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_AS) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 19776 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_AS) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 19778 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AS) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 19780 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AS) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503)) 19782 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 390) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19783 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 465) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19794 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AS) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 19796 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AS) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 19798 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AS) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 19799 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AS) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507)) 19803 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AS) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_DOUBLE_ARROW) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 19804 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AS) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 19805 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AS) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 19811 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AS) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 19813 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AS) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506)) 19815 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AS) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 19816 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 19817 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 19818 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AS) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 19820 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AS) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 19823 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AS) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 19829 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 19832 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AS) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 19834 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_AS) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 19835 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_AS) reduce 462) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 462) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 19842 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_AS) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 19843 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AS) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513)) 19845 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AS) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512)) 19846 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AS) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511)) 19847 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 463) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19850 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_AS) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 19851 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AS) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 19852 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AS) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449)) 19853 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AS) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 19854 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AS) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514)) 19855 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AS) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444)) 19856 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AS) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 19857 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AS) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 19858 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_AS) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 19859 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_AS) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 19860 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_AS) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 388) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388)) 19863 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AS) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432)) 19894 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416)) 19895 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419)) 19896 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AS) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428)) 19897 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AS) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441)) 19898 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AS) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 19899 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AS) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 19900 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 415) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19901 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 413) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19902 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 414) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19903 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AS) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438)) 19904 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AS) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 19905 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AS) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 19906 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AS) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 19907 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AS) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440)) 19908 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AS) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 19911 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 19912 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 19913 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AS) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 19914 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AS) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 19915 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 19916 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AS) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 19917 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AS) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 19918 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AS) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 19919 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AS) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 19920 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AS) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 19921 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AS) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 19927 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AS) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 19928 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AS) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 19929 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AS) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 19930 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AS) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 19931 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AS) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 19934 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AS) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 19935 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AS) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 19937 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AS) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 19939 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AS) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 19940 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AS) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 19943 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AS) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 19944 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 448) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19945 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 411) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19946 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19947 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AS) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417)) 19948 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AS) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418)) 19952 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 446) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19953 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 447) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19954 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AS) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439)) 19955 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AS) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437)) 19956 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AS) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 19957 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AS) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420)) 19958 ((("%") shift) (("*") shift) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AS) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 19959 ((("%") shift) (("*") shift) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AS) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 19960 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AS) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 19961 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AS) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 19963 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AS) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 19964 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AS) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 19965 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AS) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 19966 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AS) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 19967 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AS) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 19968 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AS) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 19969 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AS) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 19970 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AS) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 19971 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AS) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 19972 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AS) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 19973 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AS) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 19974 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AS) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 19975 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AS) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 19976 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AS) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 19977 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AS) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 19978 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AS) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 19979 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AS) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 19980 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AS) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 19981 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AS) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 19982 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AS) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 19983 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AS) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 19984 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AS) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 19985 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AS) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 19986 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AS) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 19987 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AS) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 19988 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AS) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 19989 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AS) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 19990 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AS) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 19991 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AS) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 19992 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AS) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 19993 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AS) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 19994 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AS) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 19995 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AS) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 19996 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AS) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 19997 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AS) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 19998 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AS) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 19999 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AS) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 20000 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AS) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 20001 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AS) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 20002 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AS) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 20003 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AS) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 20004 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AS) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 20005 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AS) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 20006 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AS) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 20007 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AS) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 20008 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AS) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 20009 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AS) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 20010 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AS) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 20011 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AS) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 20012 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AS) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 20013 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AS) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 20014 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AS) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 20015 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AS) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 20016 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AS) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 20017 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AS) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 20018 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AS) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 20019 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AS) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 20020 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AS) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 20021 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AS) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 20022 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AS) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 20023 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AS) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 20024 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AS) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 20025 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AS) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 20026 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AS) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 20027 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AS) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 20028 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AS) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 20029 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AS) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 20030 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AS) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 20031 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AS) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 20032 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AS) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 20033 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AS) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 20034 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AS) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 20035 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AS) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 20036 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AS) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 20037 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AS) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 20038 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AS) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 20039 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AS) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 20040 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AS) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 20041 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AS) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 20042 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AS) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525)) 20044 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AS) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 20045 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AS) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 20046 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_AS) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 20049 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AS) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527)) 20051 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_AS) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 20052 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_AS) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 20056 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_AS) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 20057 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_AS) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 20062 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AS) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 20064 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AS) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407)) 20074 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 403) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20075 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 405) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20076 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 404) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20077 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 397) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20078 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 394) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20079 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 402) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20080 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 396) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20081 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 400) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20082 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 395) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20083 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 398) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20084 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 399) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20085 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 406) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20086 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 401) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20088 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 391) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20091 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AS) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 20094 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AS) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 20096 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AS) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 20098 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AS) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 20099 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AS) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 20100 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AS) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392)) 20104 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AS) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 20106 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AS) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 20107 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AS) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 20111 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AS) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 20112 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AS) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 20113 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AS) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 20116 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AS) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 20117 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AS) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 20118 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AS) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 20123 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AS) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 20124 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AS) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 20125 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AS) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 20126 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AS) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 20129 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AS) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 20130 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AS) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 20131 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AS) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 20132 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AS) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 20133 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AS) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 20134 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AS) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 20137 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AS) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 20138 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_AS) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 20149 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AS) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 20150 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AS) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 20160 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 477) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20161 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AS) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 20163 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 464) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20165 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AS) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524)) 20167 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_AS) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 20170 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AS) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526)) 20171 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_AS) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 20172 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_AS) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 20174 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AS) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 20175 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AS) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 20180 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_AS) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 20181 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_AS) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 20182 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_AS) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 20183 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_AS) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 20186 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_AS) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 20187 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_AS) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 20188 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_AS) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 20189 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_AS) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 20191 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AS) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 20193 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AS) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_DOUBLE_ARROW) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503)) 20195 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 390) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20196 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 465) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20199 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 464) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20200 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AS) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456)) 20201 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 466) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20202 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AS) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452)) 20203 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AS) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 20205 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AS) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 20207 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AS) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 20208 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AS) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508)) 20209 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 602) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 602) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20210 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 601) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 601) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20211 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_AS) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 20212 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 461) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20213 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AS) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454)) 20215 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 20216 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 20217 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AS) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 20218 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AS) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 20219 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 20220 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AS) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385)) 20223 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AS) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 20224 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AS) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387)) 20225 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AS) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 20226 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AS) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 20227 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AS) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 20233 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AS) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 20234 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AS) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 20235 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AS) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 20236 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AS) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 20237 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AS) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384)) 20239 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AS) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 20240 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AS) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386)) 20248 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AS) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383)) 20250 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AS) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 20256 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AS) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_DOUBLE_ARROW) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231)) 20261 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 389) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20265 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AS) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_DOUBLE_ARROW) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596)) 20266 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AS) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450)) 20267 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 599) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 599) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20268 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 598) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 598) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20269 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AS) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408)) 20271 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AS) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_DOUBLE_ARROW) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501)) 20272 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AS) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457)) 20273 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AS) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289)) 20277 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AS) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290)) 20278 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AS) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 20281 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AS) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_DOUBLE_ARROW) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600)) 20284 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AS) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_DOUBLE_ARROW) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597)) 20285 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AS) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451)) 20286 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AS) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410)) 20287 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AS) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 20288 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AS) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455)) 20289 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AS) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453)) 20292 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AS) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_DOUBLE_ARROW) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502)) 20293 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AS) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 20294 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AS) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430)) 20295 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AS) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429)) 20297 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AS) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443)) 20298 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_AS) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 20301 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_AS) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 20303 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AS) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 20304 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AS) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431)) 20305 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AS) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456)) 20306 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 466) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20307 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AS) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452)) 20308 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AS) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 20310 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AS) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 20312 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AS) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 20313 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AS) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508)) 20314 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 602) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20315 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 601) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20316 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_AS) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 20317 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 461) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20318 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AS) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454)) 20320 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 20321 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 20322 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AS) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 20323 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AS) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 20324 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 20325 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AS) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385)) 20328 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AS) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 20329 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AS) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387)) 20330 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AS) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 20331 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AS) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 20332 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AS) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 20338 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AS) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 20339 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AS) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 20340 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AS) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 20341 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AS) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 20342 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AS) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384)) 20344 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AS) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 20345 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AS) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386)) 20353 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AS) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383)) 20355 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AS) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 20361 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AS) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231)) 20366 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 389) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20370 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AS) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596)) 20371 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AS) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450)) 20372 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 599) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20373 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 598) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20374 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AS) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408)) 20376 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AS) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501)) 20377 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AS) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457)) 20378 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AS) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289)) 20382 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AS) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290)) 20383 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AS) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 20386 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AS) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600)) 20389 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AS) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597)) 20390 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AS) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451)) 20391 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AS) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410)) 20392 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AS) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 20393 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AS) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455)) 20394 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AS) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453)) 20397 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AS) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502)) 20398 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AS) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 20399 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AS) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430)) 20400 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AS) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429)) 20402 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AS) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443)) 20403 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_AS) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 20406 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_AS) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 20408 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AS) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 20409 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AS) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431)) 20410 ((("\"") shift) (("$") shift) (("(") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) shift) ((T_CLASS_C) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DIR) shift) ((T_FILE) shift) ((T_FUNC_C) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_TRAIT_C) shift) ((T_VARIABLE) shift) (("[") shift)) 20414 ((("(") reduce 551) ((")") reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) (("[") reduce 551)) 20419 ((("(") reduce 538) ((")") reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) (("[") reduce 547)) 20421 (((")") shift) ((T_DOUBLE_ARROW) shift)) 20422 ((("(") reduce 546) ((")") reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) (("[") reduce 546)) 20424 ((("(") reduce 542) ((")") reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) (("[") reduce 542)) 20425 (((")") reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) (("[") reduce 548)) 20426 (((")") reduce 212) ((T_DOUBLE_ARROW) reduce 212) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) (("[") reduce 531)) 20430 ((("(") reduce 564) ((")") reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) (("[") reduce 555)) 20432 ((("(") reduce 490) ((")") reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) (("[") reduce 490)) 20433 ((("(") reduce 289) ((")") reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) (("[") reduce 289)) 20437 ((("(") reduce 290) ((")") reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) (("[") reduce 290)) 20438 ((("(") reduce 291) ((")") reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) (("[") reduce 291)) 20439 ((("(") reduce 487) ((")") reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) (("[") reduce 487)) 20445 ((("(") reduce 551) ((")") reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) (("[") reduce 551)) 20450 ((("(") reduce 538) ((")") reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) (("[") reduce 547)) 20453 ((("(") reduce 546) ((")") reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) (("[") reduce 546)) 20455 ((("(") reduce 542) ((")") reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) (("[") reduce 542)) 20456 (((")") reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) (("[") reduce 548)) 20457 (((")") reduce 212) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) (("[") reduce 531)) 20461 ((("(") reduce 564) ((")") reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) (("[") reduce 555)) 20463 ((("(") reduce 490) ((")") reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) (("[") reduce 490)) 20464 ((("(") reduce 289) ((")") reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) (("[") reduce 289)) 20468 ((("(") reduce 290) ((")") reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) (("[") reduce 290)) 20469 ((("(") reduce 291) ((")") reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) (("[") reduce 291)) 20470 ((("(") reduce 487) ((")") reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) (("[") reduce 487)) 20472 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDFOREACH) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 20473 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CASE) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DEFAULT) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDSWITCH) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 20474 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CASE) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DEFAULT) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDSWITCH) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 20475 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDFOREACH) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 20476 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDFOREACH) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 20491 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDFOREACH) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 20501 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDFOREACH) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 20503 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDFOREACH) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 20505 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDFOREACH) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 20507 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENDFOREACH) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 20510 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDFOREACH) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 20511 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDFOREACH) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 20512 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 244) ((T_ENDFOREACH) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 20513 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENDFOREACH) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_PROPERTY_C) reduce 140) ((T_READONLY) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("~") reduce 140)) 20514 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDFOREACH) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 20515 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDFOREACH) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 20516 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDFOREACH) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 20519 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDFOREACH) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 20525 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDFOREACH) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 20535 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDFOREACH) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 20543 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDFOREACH) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 20546 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDFOREACH) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 20548 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDFOREACH) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 20551 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDFOREACH) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 20553 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDFOREACH) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 20554 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDFOREACH) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 20557 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_ELSE) reduce 248) ((T_ELSEIF) reduce 248) ((T_EMPTY) reduce 248) ((T_ENDFOREACH) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("~") reduce 248)) 20561 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_ELSE) reduce 249) ((T_ELSEIF) reduce 249) ((T_EMPTY) reduce 249) ((T_ENDFOREACH) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("~") reduce 249)) 20566 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDFOREACH) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 20567 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDFOREACH) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 20570 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDFOREACH) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 20575 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDFOREACH) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 20578 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDFOREACH) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 20579 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDFOREACH) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 20582 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDFOREACH) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 20585 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_ELSE) reduce 176) ((T_ELSEIF) reduce 176) ((T_EMPTY) reduce 176) ((T_ENDFOREACH) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("~") reduce 176)) 20592 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDFOREACH) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 20597 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDFOREACH) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 20601 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDFOREACH) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 20603 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDFOREACH) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 20607 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDFOREACH) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 20610 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDFOREACH) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 20611 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDFOREACH) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 20613 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDFOREACH) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 20615 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDFOREACH) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 20617 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDFOREACH) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 20619 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDFOREACH) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 20629 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDFOREACH) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 20630 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDFOREACH) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 20633 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDFOREACH) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 20634 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDFOREACH) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 20635 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") reduce 379) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 20636 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 382) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 382) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20638 (((",") shift) ((";") reduce 380)) 20640 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 381) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 381) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20643 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") reduce 379) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 20644 ((("%") shift) ((")") reduce 382) (("*") shift) (("+") shift) ((",") reduce 382) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20646 (((")") reduce 380) ((",") shift)) 20648 ((("%") shift) ((")") reduce 381) (("*") shift) (("+") shift) ((",") reduce 381) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20650 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDFOR) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 20651 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDFOREACH) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 20652 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDFOREACH) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 20653 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDFOR) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 20654 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDFOR) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 20669 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDFOR) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 20679 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDFOR) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 20681 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDFOR) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 20683 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDFOR) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 20685 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENDFOR) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 20688 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDFOR) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 20689 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDFOR) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 20690 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 244) ((T_ENDFOR) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 20691 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENDFOR) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_PROPERTY_C) reduce 140) ((T_READONLY) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("~") reduce 140)) 20692 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDFOR) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 20693 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDFOR) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 20694 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDFOR) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 20697 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDFOR) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 20703 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDFOR) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 20713 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDFOR) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 20721 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDFOR) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 20724 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDFOR) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 20726 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDFOR) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 20729 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDFOR) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 20731 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDFOR) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 20732 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDFOR) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 20735 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_ELSE) reduce 248) ((T_ELSEIF) reduce 248) ((T_EMPTY) reduce 248) ((T_ENDFOR) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("~") reduce 248)) 20739 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_ELSE) reduce 249) ((T_ELSEIF) reduce 249) ((T_EMPTY) reduce 249) ((T_ENDFOR) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("~") reduce 249)) 20744 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDFOR) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 20745 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDFOR) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 20748 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDFOR) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 20753 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDFOR) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 20756 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDFOR) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 20757 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDFOR) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 20760 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDFOR) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 20763 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_ELSE) reduce 176) ((T_ELSEIF) reduce 176) ((T_EMPTY) reduce 176) ((T_ENDFOR) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("~") reduce 176)) 20770 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDFOR) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 20775 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDFOR) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 20779 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDFOR) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 20781 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDFOR) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 20785 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDFOR) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 20788 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDFOR) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 20789 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDFOR) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 20791 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDFOR) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 20793 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDFOR) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 20795 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDFOR) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 20797 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDFOR) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 20807 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDFOR) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 20808 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDFOR) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 20811 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDFOR) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 20812 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDFOR) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 20821 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDFOR) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 20822 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDFOR) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 20825 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDFOR) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 20826 (((",") reduce 377) ((";") reduce 377)) 20828 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 378) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 378) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20830 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDFOR) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 20831 (((",") reduce 376) ((";") reduce 376)) 20832 (((T_WHILE) reduce 165)) 20843 (((T_WHILE) reduce 159)) 20851 (((T_WHILE) reduce 148)) 20854 (((T_WHILE) reduce 147)) 20855 (((T_ELSE) shift) ((T_ELSEIF) shift) ((T_WHILE) reduce 244)) 20856 (((T_WHILE) shift)) 20859 (((T_WHILE) reduce 146)) 20864 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDFOR) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 20870 (((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_WHILE) reduce 165)) 20880 (((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_WHILE) reduce 159)) 20888 (((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_WHILE) reduce 148)) 20891 (((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_WHILE) reduce 147)) 20893 (((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_WHILE) reduce 243)) 20896 (((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_WHILE) reduce 146)) 20898 (((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_WHILE) reduce 245)) 20899 (((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_WHILE) reduce 160)) 20902 (((T_ELSE) reduce 248) ((T_ELSEIF) reduce 248) ((T_WHILE) reduce 248)) 20906 (((T_ELSE) reduce 249) ((T_ELSEIF) reduce 249) ((T_WHILE) reduce 249)) 20911 (((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_WHILE) reduce 240)) 20912 (((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_WHILE) reduce 149)) 20915 (((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_WHILE) reduce 241)) 20920 (((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_WHILE) reduce 161)) 20923 (((T_CATCH) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_FINALLY) reduce 169) ((T_WHILE) reduce 169)) 20924 (((T_CATCH) shift) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_FINALLY) shift) ((T_WHILE) reduce 175)) 20927 (((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_WHILE) reduce 166)) 20930 (((T_ELSE) reduce 176) ((T_ELSEIF) reduce 176) ((T_WHILE) reduce 176)) 20937 (((T_CATCH) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_FINALLY) reduce 170) ((T_WHILE) reduce 170)) 20942 (((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_WHILE) reduce 152)) 20946 (((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_WHILE) reduce 222)) 20948 (((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_WHILE) reduce 223)) 20952 (((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_WHILE) reduce 224)) 20955 (((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_WHILE) reduce 225)) 20956 (((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_WHILE) reduce 168)) 20958 (((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_WHILE) reduce 157)) 20960 (((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_WHILE) reduce 155)) 20962 (((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_WHILE) reduce 167)) 20964 (((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_WHILE) reduce 156)) 20974 (((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_WHILE) reduce 163)) 20975 (((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_WHILE) reduce 218)) 20978 (((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_WHILE) reduce 219)) 20979 (((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_WHILE) reduce 162)) 20988 (((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_WHILE) reduce 151)) 20989 (((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_WHILE) reduce 216)) 20992 (((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_WHILE) reduce 217)) 20994 (((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_WHILE) reduce 158)) 21000 (((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_WHILE) reduce 150)) 21003 (((")") reduce 139) ((",") reduce 139)) 21004 (((")") shift) ((",") shift)) 21007 (((")") reduce 138) ((",") reduce 138)) 21008 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDDECLARE) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 21009 (((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_WHILE) reduce 164)) 21010 (((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_WHILE) reduce 220)) 21011 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 21012 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDDECLARE) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 21027 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDDECLARE) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 21037 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDDECLARE) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 21039 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDDECLARE) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 21041 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDDECLARE) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 21043 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENDDECLARE) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 21046 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDDECLARE) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 21047 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDDECLARE) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 21048 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 244) ((T_ENDDECLARE) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 21049 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENDDECLARE) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_PROPERTY_C) reduce 140) ((T_READONLY) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("~") reduce 140)) 21050 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDDECLARE) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 21051 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDDECLARE) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 21052 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDDECLARE) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 21055 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDDECLARE) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 21061 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDDECLARE) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 21071 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDDECLARE) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 21079 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDDECLARE) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 21082 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDDECLARE) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 21084 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDDECLARE) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 21087 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDDECLARE) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 21089 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDDECLARE) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 21090 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDDECLARE) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 21093 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_ELSE) reduce 248) ((T_ELSEIF) reduce 248) ((T_EMPTY) reduce 248) ((T_ENDDECLARE) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("~") reduce 248)) 21097 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_ELSE) reduce 249) ((T_ELSEIF) reduce 249) ((T_EMPTY) reduce 249) ((T_ENDDECLARE) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("~") reduce 249)) 21102 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDDECLARE) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 21103 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDDECLARE) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 21106 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDDECLARE) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 21111 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDDECLARE) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 21114 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDDECLARE) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 21115 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDDECLARE) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 21118 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDDECLARE) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 21121 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_ELSE) reduce 176) ((T_ELSEIF) reduce 176) ((T_EMPTY) reduce 176) ((T_ENDDECLARE) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("~") reduce 176)) 21128 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDDECLARE) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 21133 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDDECLARE) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 21137 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDDECLARE) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 21139 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDDECLARE) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 21143 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDDECLARE) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 21146 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDDECLARE) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 21147 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDDECLARE) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 21149 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDDECLARE) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 21151 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDDECLARE) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 21153 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDDECLARE) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 21155 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDDECLARE) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 21165 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDDECLARE) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 21166 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDDECLARE) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 21169 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDDECLARE) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 21170 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDDECLARE) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 21179 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDDECLARE) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 21180 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDDECLARE) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 21183 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDDECLARE) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 21185 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDDECLARE) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 21191 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDDECLARE) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 21196 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDDECLARE) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 21197 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDDECLARE) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 21200 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDDECLARE) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 21202 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDDECLARE) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 21204 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDDECLARE) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 21205 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDDECLARE) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 21216 ((("!") reduce 477) (("\"") reduce 477) (("$") reduce 477) (("(") reduce 477) (("+") reduce 477) (("-") reduce 477) ((";") reduce 477) (("@") reduce 477) ((T_ABSTRACT) reduce 477) ((T_ARRAY) reduce 477) ((T_ARRAY_CAST) reduce 477) ((T_ATTRIBUTE) reduce 477) ((T_BOOL_CAST) reduce 477) ((T_BREAK) reduce 477) ((T_CLASS) reduce 477) ((T_CLASS_C) reduce 477) ((T_CLONE) reduce 477) ((T_CONSTANT_ENCAPSED_STRING) reduce 477) ((T_CONTINUE) reduce 477) ((T_DEC) reduce 477) ((T_DECLARE) reduce 477) ((T_DIR) reduce 477) ((T_DNUMBER) reduce 477) ((T_DO) reduce 477) ((T_DOUBLE_CAST) reduce 477) ((T_ECHO) reduce 477) ((T_EMPTY) reduce 477) ((T_ENDDECLARE) reduce 477) ((T_ENUM) reduce 477) ((T_EVAL) reduce 477) ((T_EXIT) reduce 477) ((T_FILE) reduce 477) ((T_FINAL) reduce 477) ((T_FN) reduce 477) ((T_FOR) reduce 477) ((T_FOREACH) reduce 477) ((T_FUNCTION) reduce 477) ((T_FUNC_C) reduce 477) ((T_GLOBAL) reduce 477) ((T_GOTO) reduce 477) ((T_HALT_COMPILER) reduce 477) ((T_IF) reduce 477) ((T_INC) reduce 477) ((T_INCLUDE) reduce 477) ((T_INCLUDE_ONCE) reduce 477) ((T_INLINE_HTML) reduce 477) ((T_INTERFACE) reduce 477) ((T_INT_CAST) reduce 477) ((T_ISSET) reduce 477) ((T_LINE) reduce 477) ((T_LIST) reduce 477) ((T_LNUMBER) reduce 477) ((T_MATCH) reduce 477) ((T_METHOD_C) reduce 477) ((T_NAME_FULLY_QUALIFIED) reduce 477) ((T_NAME_QUALIFIED) reduce 477) ((T_NAME_RELATIVE) reduce 477) ((T_NEW) reduce 477) ((T_NS_C) reduce 477) ((T_OBJECT_CAST) reduce 477) ((T_PRINT) reduce 477) ((T_PROPERTY_C) reduce 477) ((T_READONLY) reduce 477) ((T_REQUIRE) reduce 477) ((T_REQUIRE_ONCE) reduce 477) ((T_RETURN) reduce 477) ((T_START_HEREDOC) reduce 477) ((T_STATIC) reduce 477) ((T_STRING) reduce 477) ((T_STRING_CAST) reduce 477) ((T_SWITCH) reduce 477) ((T_THROW) reduce 477) ((T_TRAIT) reduce 477) ((T_TRAIT_C) reduce 477) ((T_TRY) reduce 477) ((T_UNSET) reduce 477) ((T_UNSET_CAST) reduce 477) ((T_VARIABLE) reduce 477) ((T_WHILE) reduce 477) ((T_YIELD) reduce 477) ((T_YIELD_FROM) reduce 477) (("[") reduce 477) (("`") reduce 477) (("{") reduce 477) (("~") reduce 477)) 21217 ((("!") reduce 182) (("\"") reduce 182) (("$") reduce 182) (("(") reduce 182) (("+") reduce 182) (("-") reduce 182) ((";") reduce 182) (("@") reduce 182) ((T_ABSTRACT) reduce 182) ((T_ARRAY) reduce 182) ((T_ARRAY_CAST) reduce 182) ((T_ATTRIBUTE) reduce 182) ((T_BOOL_CAST) reduce 182) ((T_BREAK) reduce 182) ((T_CLASS) reduce 182) ((T_CLASS_C) reduce 182) ((T_CLONE) reduce 182) ((T_CONSTANT_ENCAPSED_STRING) reduce 182) ((T_CONTINUE) reduce 182) ((T_DEC) reduce 182) ((T_DECLARE) reduce 182) ((T_DIR) reduce 182) ((T_DNUMBER) reduce 182) ((T_DO) reduce 182) ((T_DOUBLE_CAST) reduce 182) ((T_ECHO) reduce 182) ((T_EMPTY) reduce 182) ((T_ENDDECLARE) reduce 182) ((T_ENUM) reduce 182) ((T_EVAL) reduce 182) ((T_EXIT) reduce 182) ((T_FILE) reduce 182) ((T_FINAL) reduce 182) ((T_FN) reduce 182) ((T_FOR) reduce 182) ((T_FOREACH) reduce 182) ((T_FUNCTION) reduce 182) ((T_FUNC_C) reduce 182) ((T_GLOBAL) reduce 182) ((T_GOTO) reduce 182) ((T_HALT_COMPILER) reduce 182) ((T_IF) reduce 182) ((T_INC) reduce 182) ((T_INCLUDE) reduce 182) ((T_INCLUDE_ONCE) reduce 182) ((T_INLINE_HTML) reduce 182) ((T_INTERFACE) reduce 182) ((T_INT_CAST) reduce 182) ((T_ISSET) reduce 182) ((T_LINE) reduce 182) ((T_LIST) reduce 182) ((T_LNUMBER) reduce 182) ((T_MATCH) reduce 182) ((T_METHOD_C) reduce 182) ((T_NAME_FULLY_QUALIFIED) reduce 182) ((T_NAME_QUALIFIED) reduce 182) ((T_NAME_RELATIVE) reduce 182) ((T_NEW) reduce 182) ((T_NS_C) reduce 182) ((T_OBJECT_CAST) reduce 182) ((T_PRINT) reduce 182) ((T_PROPERTY_C) reduce 182) ((T_READONLY) reduce 182) ((T_REQUIRE) reduce 182) ((T_REQUIRE_ONCE) reduce 182) ((T_RETURN) reduce 182) ((T_START_HEREDOC) reduce 182) ((T_STATIC) reduce 182) ((T_STRING) reduce 182) ((T_STRING_CAST) reduce 182) ((T_SWITCH) reduce 182) ((T_THROW) reduce 182) ((T_TRAIT) reduce 182) ((T_TRAIT_C) reduce 182) ((T_TRY) reduce 182) ((T_UNSET) reduce 182) ((T_UNSET_CAST) reduce 182) ((T_VARIABLE) reduce 182) ((T_WHILE) reduce 182) ((T_YIELD) reduce 182) ((T_YIELD_FROM) reduce 182) (("[") reduce 182) (("`") reduce 182) (("{") reduce 182) (("~") reduce 182)) 21218 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDDECLARE) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 21226 ((("!") reduce 187) (("\"") reduce 187) (("$") reduce 187) (("(") reduce 187) (("+") reduce 187) (("-") reduce 187) ((";") reduce 187) (("@") reduce 187) ((T_ABSTRACT) reduce 187) ((T_ARRAY) reduce 187) ((T_ARRAY_CAST) reduce 187) ((T_ATTRIBUTE) reduce 187) ((T_BOOL_CAST) reduce 187) ((T_BREAK) reduce 187) ((T_CLASS) reduce 187) ((T_CLASS_C) reduce 187) ((T_CLONE) reduce 187) ((T_CONSTANT_ENCAPSED_STRING) reduce 187) ((T_CONTINUE) reduce 187) ((T_DEC) reduce 187) ((T_DECLARE) reduce 187) ((T_DIR) reduce 187) ((T_DNUMBER) reduce 187) ((T_DO) reduce 187) ((T_DOUBLE_CAST) reduce 187) ((T_ECHO) reduce 187) ((T_EMPTY) reduce 187) ((T_ENDDECLARE) reduce 187) ((T_ENUM) reduce 187) ((T_EVAL) reduce 187) ((T_EXIT) reduce 187) ((T_FILE) reduce 187) ((T_FINAL) reduce 187) ((T_FN) reduce 187) ((T_FOR) reduce 187) ((T_FOREACH) reduce 187) ((T_FUNCTION) reduce 187) ((T_FUNC_C) reduce 187) ((T_GLOBAL) reduce 187) ((T_GOTO) reduce 187) ((T_HALT_COMPILER) reduce 187) ((T_IF) reduce 187) ((T_INC) reduce 187) ((T_INCLUDE) reduce 187) ((T_INCLUDE_ONCE) reduce 187) ((T_INLINE_HTML) reduce 187) ((T_INTERFACE) reduce 187) ((T_INT_CAST) reduce 187) ((T_ISSET) reduce 187) ((T_LINE) reduce 187) ((T_LIST) reduce 187) ((T_LNUMBER) reduce 187) ((T_MATCH) reduce 187) ((T_METHOD_C) reduce 187) ((T_NAME_FULLY_QUALIFIED) reduce 187) ((T_NAME_QUALIFIED) reduce 187) ((T_NAME_RELATIVE) reduce 187) ((T_NEW) reduce 187) ((T_NS_C) reduce 187) ((T_OBJECT_CAST) reduce 187) ((T_PRINT) reduce 187) ((T_PROPERTY_C) reduce 187) ((T_READONLY) reduce 187) ((T_REQUIRE) reduce 187) ((T_REQUIRE_ONCE) reduce 187) ((T_RETURN) reduce 187) ((T_START_HEREDOC) reduce 187) ((T_STATIC) reduce 187) ((T_STRING) reduce 187) ((T_STRING_CAST) reduce 187) ((T_SWITCH) reduce 187) ((T_THROW) reduce 187) ((T_TRAIT) reduce 187) ((T_TRAIT_C) reduce 187) ((T_TRY) reduce 187) ((T_UNSET) reduce 187) ((T_UNSET_CAST) reduce 187) ((T_VARIABLE) reduce 187) ((T_WHILE) reduce 187) ((T_YIELD) reduce 187) ((T_YIELD_FROM) reduce 187) (("[") reduce 187) (("`") reduce 187) (("{") reduce 187) (("~") reduce 187)) 21227 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDDECLARE) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 21230 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_EMPTY) reduce 248) ((T_ENDDECLARE) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("~") reduce 248)) 21234 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_EMPTY) reduce 249) ((T_ENDDECLARE) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("~") reduce 249)) 21239 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDDECLARE) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 21240 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDDECLARE) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 21243 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDDECLARE) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 21248 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDDECLARE) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 21251 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDDECLARE) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 21252 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDDECLARE) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 21255 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDDECLARE) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 21258 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_EMPTY) reduce 176) ((T_ENDDECLARE) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("~") reduce 176)) 21265 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDDECLARE) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 21270 ((("!") reduce 198) (("\"") reduce 198) (("$") reduce 198) (("(") reduce 198) (("+") reduce 198) (("-") reduce 198) ((";") reduce 198) (("@") reduce 198) ((T_ABSTRACT) reduce 198) ((T_ARRAY) reduce 198) ((T_ARRAY_CAST) reduce 198) ((T_ATTRIBUTE) reduce 198) ((T_BOOL_CAST) reduce 198) ((T_BREAK) reduce 198) ((T_CLASS) reduce 198) ((T_CLASS_C) reduce 198) ((T_CLONE) reduce 198) ((T_CONSTANT_ENCAPSED_STRING) reduce 198) ((T_CONTINUE) reduce 198) ((T_DEC) reduce 198) ((T_DECLARE) reduce 198) ((T_DIR) reduce 198) ((T_DNUMBER) reduce 198) ((T_DO) reduce 198) ((T_DOUBLE_CAST) reduce 198) ((T_ECHO) reduce 198) ((T_EMPTY) reduce 198) ((T_ENDDECLARE) reduce 198) ((T_ENUM) reduce 198) ((T_EVAL) reduce 198) ((T_EXIT) reduce 198) ((T_FILE) reduce 198) ((T_FINAL) reduce 198) ((T_FN) reduce 198) ((T_FOR) reduce 198) ((T_FOREACH) reduce 198) ((T_FUNCTION) reduce 198) ((T_FUNC_C) reduce 198) ((T_GLOBAL) reduce 198) ((T_GOTO) reduce 198) ((T_HALT_COMPILER) reduce 198) ((T_IF) reduce 198) ((T_INC) reduce 198) ((T_INCLUDE) reduce 198) ((T_INCLUDE_ONCE) reduce 198) ((T_INLINE_HTML) reduce 198) ((T_INTERFACE) reduce 198) ((T_INT_CAST) reduce 198) ((T_ISSET) reduce 198) ((T_LINE) reduce 198) ((T_LIST) reduce 198) ((T_LNUMBER) reduce 198) ((T_MATCH) reduce 198) ((T_METHOD_C) reduce 198) ((T_NAME_FULLY_QUALIFIED) reduce 198) ((T_NAME_QUALIFIED) reduce 198) ((T_NAME_RELATIVE) reduce 198) ((T_NEW) reduce 198) ((T_NS_C) reduce 198) ((T_OBJECT_CAST) reduce 198) ((T_PRINT) reduce 198) ((T_PROPERTY_C) reduce 198) ((T_READONLY) reduce 198) ((T_REQUIRE) reduce 198) ((T_REQUIRE_ONCE) reduce 198) ((T_RETURN) reduce 198) ((T_START_HEREDOC) reduce 198) ((T_STATIC) reduce 198) ((T_STRING) reduce 198) ((T_STRING_CAST) reduce 198) ((T_SWITCH) reduce 198) ((T_THROW) reduce 198) ((T_TRAIT) reduce 198) ((T_TRAIT_C) reduce 198) ((T_TRY) reduce 198) ((T_UNSET) reduce 198) ((T_UNSET_CAST) reduce 198) ((T_VARIABLE) reduce 198) ((T_WHILE) reduce 198) ((T_YIELD) reduce 198) ((T_YIELD_FROM) reduce 198) (("[") reduce 198) (("`") reduce 198) (("{") reduce 198) (("~") reduce 198)) 21275 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDDECLARE) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 21279 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDDECLARE) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 21281 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDDECLARE) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 21285 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDDECLARE) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 21288 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDDECLARE) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 21289 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDDECLARE) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 21291 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDDECLARE) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 21293 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDDECLARE) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 21294 (((T_EXTENDS) shift) (("{") reduce 208)) 21300 ((("!") reduce 199) (("\"") reduce 199) (("$") reduce 199) (("(") reduce 199) (("+") reduce 199) (("-") reduce 199) ((";") reduce 199) (("@") reduce 199) ((T_ABSTRACT) reduce 199) ((T_ARRAY) reduce 199) ((T_ARRAY_CAST) reduce 199) ((T_ATTRIBUTE) reduce 199) ((T_BOOL_CAST) reduce 199) ((T_BREAK) reduce 199) ((T_CLASS) reduce 199) ((T_CLASS_C) reduce 199) ((T_CLONE) reduce 199) ((T_CONSTANT_ENCAPSED_STRING) reduce 199) ((T_CONTINUE) reduce 199) ((T_DEC) reduce 199) ((T_DECLARE) reduce 199) ((T_DIR) reduce 199) ((T_DNUMBER) reduce 199) ((T_DO) reduce 199) ((T_DOUBLE_CAST) reduce 199) ((T_ECHO) reduce 199) ((T_EMPTY) reduce 199) ((T_ENDDECLARE) reduce 199) ((T_ENUM) reduce 199) ((T_EVAL) reduce 199) ((T_EXIT) reduce 199) ((T_FILE) reduce 199) ((T_FINAL) reduce 199) ((T_FN) reduce 199) ((T_FOR) reduce 199) ((T_FOREACH) reduce 199) ((T_FUNCTION) reduce 199) ((T_FUNC_C) reduce 199) ((T_GLOBAL) reduce 199) ((T_GOTO) reduce 199) ((T_HALT_COMPILER) reduce 199) ((T_IF) reduce 199) ((T_INC) reduce 199) ((T_INCLUDE) reduce 199) ((T_INCLUDE_ONCE) reduce 199) ((T_INLINE_HTML) reduce 199) ((T_INTERFACE) reduce 199) ((T_INT_CAST) reduce 199) ((T_ISSET) reduce 199) ((T_LINE) reduce 199) ((T_LIST) reduce 199) ((T_LNUMBER) reduce 199) ((T_MATCH) reduce 199) ((T_METHOD_C) reduce 199) ((T_NAME_FULLY_QUALIFIED) reduce 199) ((T_NAME_QUALIFIED) reduce 199) ((T_NAME_RELATIVE) reduce 199) ((T_NEW) reduce 199) ((T_NS_C) reduce 199) ((T_OBJECT_CAST) reduce 199) ((T_PRINT) reduce 199) ((T_PROPERTY_C) reduce 199) ((T_READONLY) reduce 199) ((T_REQUIRE) reduce 199) ((T_REQUIRE_ONCE) reduce 199) ((T_RETURN) reduce 199) ((T_START_HEREDOC) reduce 199) ((T_STATIC) reduce 199) ((T_STRING) reduce 199) ((T_STRING_CAST) reduce 199) ((T_SWITCH) reduce 199) ((T_THROW) reduce 199) ((T_TRAIT) reduce 199) ((T_TRAIT_C) reduce 199) ((T_TRY) reduce 199) ((T_UNSET) reduce 199) ((T_UNSET_CAST) reduce 199) ((T_VARIABLE) reduce 199) ((T_WHILE) reduce 199) ((T_YIELD) reduce 199) ((T_YIELD_FROM) reduce 199) (("[") reduce 199) (("`") reduce 199) (("{") reduce 199) (("~") reduce 199)) 21301 (((",") shift) (("{") reduce 209)) 21306 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDDECLARE) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 21307 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_ELSE) reduce 246) ((T_ELSEIF) reduce 246) ((T_EMPTY) shift) ((T_ENDIF) reduce 246) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 21310 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDDECLARE) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 21312 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDDECLARE) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 21314 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDDECLARE) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 21324 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDDECLARE) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 21325 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDDECLARE) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 21328 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDDECLARE) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 21329 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDDECLARE) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 21338 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDDECLARE) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 21339 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDDECLARE) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 21342 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDDECLARE) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 21343 (((":") shift) ((T_IMPLEMENTS) reduce 201) (("{") reduce 201)) 21350 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENDDECLARE) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 21352 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_IMPLEMENTS) reduce 278) (("{") reduce 278) (("|") reduce 278)) 21353 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) ((T_IMPLEMENTS) reduce 279) (("{") reduce 279) (("|") reduce 279)) 21354 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_IMPLEMENTS) reduce 94) (("{") reduce 94) (("|") reduce 94)) 21355 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_IMPLEMENTS) reduce 93) (("{") reduce 93) (("|") reduce 93)) 21356 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_IMPLEMENTS) reduce 95) (("{") reduce 95) (("|") reduce 95)) 21357 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267) ((T_IMPLEMENTS) reduce 267) (("{") reduce 267) (("|") reduce 267)) 21358 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_IMPLEMENTS) reduce 92) (("{") reduce 92) (("|") reduce 92)) 21359 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_IMPLEMENTS) reduce 265) (("{") reduce 265)) 21360 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_IMPLEMENTS) reduce 280) (("{") reduce 280) (("|") reduce 280)) 21361 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_IMPLEMENTS) reduce 262) (("{") reduce 262) (("|") reduce 268)) 21362 (((T_IMPLEMENTS) reduce 202) (("{") reduce 202)) 21363 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266) ((T_IMPLEMENTS) reduce 266) (("{") reduce 266) (("|") reduce 266)) 21364 (((T_IMPLEMENTS) reduce 264) (("{") reduce 264) (("|") shift)) 21368 (((T_IMPLEMENTS) reduce 278) (("{") reduce 278) (("|") reduce 278)) 21369 (((T_IMPLEMENTS) reduce 279) (("{") reduce 279) (("|") reduce 279)) 21370 (((T_IMPLEMENTS) reduce 94) (("{") reduce 94) (("|") reduce 94)) 21371 (((T_IMPLEMENTS) reduce 93) (("{") reduce 93) (("|") reduce 93)) 21372 (((T_IMPLEMENTS) reduce 95) (("{") reduce 95) (("|") reduce 95)) 21373 (((T_IMPLEMENTS) reduce 267) (("{") reduce 267) (("|") reduce 267)) 21374 (((T_IMPLEMENTS) reduce 92) (("{") reduce 92) (("|") reduce 92)) 21375 (((T_IMPLEMENTS) reduce 280) (("{") reduce 280) (("|") reduce 280)) 21376 (((T_IMPLEMENTS) reduce 268) (("{") reduce 268) (("|") reduce 268)) 21377 (((T_IMPLEMENTS) reduce 266) (("{") reduce 266) (("|") reduce 266)) 21378 (((T_IMPLEMENTS) reduce 270) (("{") reduce 270) (("|") reduce 270)) 21380 (((T_IMPLEMENTS) reduce 269) (("{") reduce 269) (("|") reduce 269)) 21382 (((T_IMPLEMENTS) reduce 271) (("{") reduce 271) (("|") reduce 271)) 21384 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_IMPLEMENTS) reduce 278) (("{") reduce 278)) 21385 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) ((T_IMPLEMENTS) reduce 279) (("{") reduce 279)) 21386 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_IMPLEMENTS) reduce 94) (("{") reduce 94)) 21387 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_IMPLEMENTS) reduce 93) (("{") reduce 93)) 21388 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_IMPLEMENTS) reduce 95) (("{") reduce 95)) 21389 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267) ((T_IMPLEMENTS) reduce 267) (("{") reduce 267)) 21390 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_IMPLEMENTS) reduce 92) (("{") reduce 92)) 21391 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_IMPLEMENTS) reduce 280) (("{") reduce 280)) 21392 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) ((T_IMPLEMENTS) reduce 272) (("{") reduce 272)) 21393 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266) ((T_IMPLEMENTS) reduce 266) (("{") reduce 266)) 21395 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) ((T_IMPLEMENTS) reduce 273) (("{") reduce 273)) 21396 (((T_IMPLEMENTS) reduce 278) (("{") reduce 278)) 21397 (((T_IMPLEMENTS) reduce 279) (("{") reduce 279)) 21398 (((T_IMPLEMENTS) reduce 267) (("{") reduce 267)) 21399 (((T_IMPLEMENTS) reduce 280) (("{") reduce 280)) 21400 (((T_IMPLEMENTS) reduce 263) (("{") reduce 263)) 21401 (((T_IMPLEMENTS) reduce 266) (("{") reduce 266)) 21402 (((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_WHILE) reduce 221)) 21404 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDDECLARE) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 21410 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDDECLARE) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 21415 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDDECLARE) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 21416 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDDECLARE) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 21419 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDDECLARE) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 21421 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDDECLARE) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 21428 ((("!") reduce 188) (("\"") reduce 188) (("$") reduce 188) (("(") reduce 188) (("+") reduce 188) (("-") reduce 188) ((";") reduce 188) (("@") reduce 188) ((T_ABSTRACT) reduce 188) ((T_ARRAY) reduce 188) ((T_ARRAY_CAST) reduce 188) ((T_ATTRIBUTE) reduce 188) ((T_BOOL_CAST) reduce 188) ((T_BREAK) reduce 188) ((T_CLASS) reduce 188) ((T_CLASS_C) reduce 188) ((T_CLONE) reduce 188) ((T_CONSTANT_ENCAPSED_STRING) reduce 188) ((T_CONTINUE) reduce 188) ((T_DEC) reduce 188) ((T_DECLARE) reduce 188) ((T_DIR) reduce 188) ((T_DNUMBER) reduce 188) ((T_DO) reduce 188) ((T_DOUBLE_CAST) reduce 188) ((T_ECHO) reduce 188) ((T_EMPTY) reduce 188) ((T_ENDDECLARE) reduce 188) ((T_ENUM) reduce 188) ((T_EVAL) reduce 188) ((T_EXIT) reduce 188) ((T_FILE) reduce 188) ((T_FINAL) reduce 188) ((T_FN) reduce 188) ((T_FOR) reduce 188) ((T_FOREACH) reduce 188) ((T_FUNCTION) reduce 188) ((T_FUNC_C) reduce 188) ((T_GLOBAL) reduce 188) ((T_GOTO) reduce 188) ((T_HALT_COMPILER) reduce 188) ((T_IF) reduce 188) ((T_INC) reduce 188) ((T_INCLUDE) reduce 188) ((T_INCLUDE_ONCE) reduce 188) ((T_INLINE_HTML) reduce 188) ((T_INTERFACE) reduce 188) ((T_INT_CAST) reduce 188) ((T_ISSET) reduce 188) ((T_LINE) reduce 188) ((T_LIST) reduce 188) ((T_LNUMBER) reduce 188) ((T_MATCH) reduce 188) ((T_METHOD_C) reduce 188) ((T_NAME_FULLY_QUALIFIED) reduce 188) ((T_NAME_QUALIFIED) reduce 188) ((T_NAME_RELATIVE) reduce 188) ((T_NEW) reduce 188) ((T_NS_C) reduce 188) ((T_OBJECT_CAST) reduce 188) ((T_PRINT) reduce 188) ((T_PROPERTY_C) reduce 188) ((T_READONLY) reduce 188) ((T_REQUIRE) reduce 188) ((T_REQUIRE_ONCE) reduce 188) ((T_RETURN) reduce 188) ((T_START_HEREDOC) reduce 188) ((T_STATIC) reduce 188) ((T_STRING) reduce 188) ((T_STRING_CAST) reduce 188) ((T_SWITCH) reduce 188) ((T_THROW) reduce 188) ((T_TRAIT) reduce 188) ((T_TRAIT_C) reduce 188) ((T_TRY) reduce 188) ((T_UNSET) reduce 188) ((T_UNSET_CAST) reduce 188) ((T_VARIABLE) reduce 188) ((T_WHILE) reduce 188) ((T_YIELD) reduce 188) ((T_YIELD_FROM) reduce 188) (("[") reduce 188) (("`") reduce 188) (("{") reduce 188) (("~") reduce 188)) 21430 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDDECLARE) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 21432 ((("%") shift) ((")") reduce 476) (("*") shift) (("+") shift) ((",") reduce 476) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21433 (((")") reduce 375) ((",") reduce 375)) 21435 (((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_WHILE) reduce 154)) 21437 (((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_WHILE) reduce 153)) 21438 (((T_WHILE) reduce 245)) 21439 (((T_WHILE) reduce 160)) 21442 (((T_WHILE) reduce 248)) 21446 (((T_WHILE) reduce 249)) 21451 (((T_WHILE) reduce 240)) 21452 (((T_WHILE) reduce 149)) 21455 (((T_WHILE) reduce 241)) 21460 (((T_WHILE) reduce 161)) 21463 (((T_CATCH) reduce 169) ((T_FINALLY) reduce 169) ((T_WHILE) reduce 169)) 21464 (((T_CATCH) shift) ((T_FINALLY) shift) ((T_WHILE) reduce 175)) 21467 (((T_WHILE) reduce 166)) 21470 (((T_WHILE) reduce 176)) 21477 (((T_CATCH) reduce 170) ((T_FINALLY) reduce 170) ((T_WHILE) reduce 170)) 21482 (((T_WHILE) reduce 152)) 21486 (((T_WHILE) reduce 222)) 21488 (((T_WHILE) reduce 223)) 21492 (((T_WHILE) reduce 224)) 21495 (((T_WHILE) reduce 225)) 21496 (((T_WHILE) reduce 168)) 21498 (((T_WHILE) reduce 157)) 21500 (((T_WHILE) reduce 155)) 21504 (((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_WHILE) reduce 242)) 21506 (((T_WHILE) reduce 167)) 21508 (((T_WHILE) reduce 156)) 21518 (((T_WHILE) reduce 163)) 21519 (((T_WHILE) reduce 218)) 21522 (((T_WHILE) reduce 219)) 21523 (((T_WHILE) reduce 162)) 21532 (((T_WHILE) reduce 151)) 21533 (((T_WHILE) reduce 216)) 21536 (((T_WHILE) reduce 217)) 21538 (((T_WHILE) reduce 158)) 21544 (((T_WHILE) reduce 150)) 21549 (((T_WHILE) reduce 164)) 21550 (((T_WHILE) reduce 220)) 21553 (((T_WHILE) reduce 221)) 21555 (((T_WHILE) reduce 154)) 21557 (((T_WHILE) reduce 153)) 21562 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDFOR) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 21563 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDFOR) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 21566 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDFOR) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 21568 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDFOR) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 21570 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDFOR) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 21571 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDFOR) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 21582 ((("!") reduce 477) (("\"") reduce 477) (("$") reduce 477) (("(") reduce 477) (("+") reduce 477) (("-") reduce 477) ((";") reduce 477) (("@") reduce 477) ((T_ABSTRACT) reduce 477) ((T_ARRAY) reduce 477) ((T_ARRAY_CAST) reduce 477) ((T_ATTRIBUTE) reduce 477) ((T_BOOL_CAST) reduce 477) ((T_BREAK) reduce 477) ((T_CLASS) reduce 477) ((T_CLASS_C) reduce 477) ((T_CLONE) reduce 477) ((T_CONSTANT_ENCAPSED_STRING) reduce 477) ((T_CONTINUE) reduce 477) ((T_DEC) reduce 477) ((T_DECLARE) reduce 477) ((T_DIR) reduce 477) ((T_DNUMBER) reduce 477) ((T_DO) reduce 477) ((T_DOUBLE_CAST) reduce 477) ((T_ECHO) reduce 477) ((T_EMPTY) reduce 477) ((T_ENDFOR) reduce 477) ((T_ENUM) reduce 477) ((T_EVAL) reduce 477) ((T_EXIT) reduce 477) ((T_FILE) reduce 477) ((T_FINAL) reduce 477) ((T_FN) reduce 477) ((T_FOR) reduce 477) ((T_FOREACH) reduce 477) ((T_FUNCTION) reduce 477) ((T_FUNC_C) reduce 477) ((T_GLOBAL) reduce 477) ((T_GOTO) reduce 477) ((T_HALT_COMPILER) reduce 477) ((T_IF) reduce 477) ((T_INC) reduce 477) ((T_INCLUDE) reduce 477) ((T_INCLUDE_ONCE) reduce 477) ((T_INLINE_HTML) reduce 477) ((T_INTERFACE) reduce 477) ((T_INT_CAST) reduce 477) ((T_ISSET) reduce 477) ((T_LINE) reduce 477) ((T_LIST) reduce 477) ((T_LNUMBER) reduce 477) ((T_MATCH) reduce 477) ((T_METHOD_C) reduce 477) ((T_NAME_FULLY_QUALIFIED) reduce 477) ((T_NAME_QUALIFIED) reduce 477) ((T_NAME_RELATIVE) reduce 477) ((T_NEW) reduce 477) ((T_NS_C) reduce 477) ((T_OBJECT_CAST) reduce 477) ((T_PRINT) reduce 477) ((T_PROPERTY_C) reduce 477) ((T_READONLY) reduce 477) ((T_REQUIRE) reduce 477) ((T_REQUIRE_ONCE) reduce 477) ((T_RETURN) reduce 477) ((T_START_HEREDOC) reduce 477) ((T_STATIC) reduce 477) ((T_STRING) reduce 477) ((T_STRING_CAST) reduce 477) ((T_SWITCH) reduce 477) ((T_THROW) reduce 477) ((T_TRAIT) reduce 477) ((T_TRAIT_C) reduce 477) ((T_TRY) reduce 477) ((T_UNSET) reduce 477) ((T_UNSET_CAST) reduce 477) ((T_VARIABLE) reduce 477) ((T_WHILE) reduce 477) ((T_YIELD) reduce 477) ((T_YIELD_FROM) reduce 477) (("[") reduce 477) (("`") reduce 477) (("{") reduce 477) (("~") reduce 477)) 21583 ((("!") reduce 182) (("\"") reduce 182) (("$") reduce 182) (("(") reduce 182) (("+") reduce 182) (("-") reduce 182) ((";") reduce 182) (("@") reduce 182) ((T_ABSTRACT) reduce 182) ((T_ARRAY) reduce 182) ((T_ARRAY_CAST) reduce 182) ((T_ATTRIBUTE) reduce 182) ((T_BOOL_CAST) reduce 182) ((T_BREAK) reduce 182) ((T_CLASS) reduce 182) ((T_CLASS_C) reduce 182) ((T_CLONE) reduce 182) ((T_CONSTANT_ENCAPSED_STRING) reduce 182) ((T_CONTINUE) reduce 182) ((T_DEC) reduce 182) ((T_DECLARE) reduce 182) ((T_DIR) reduce 182) ((T_DNUMBER) reduce 182) ((T_DO) reduce 182) ((T_DOUBLE_CAST) reduce 182) ((T_ECHO) reduce 182) ((T_EMPTY) reduce 182) ((T_ENDFOR) reduce 182) ((T_ENUM) reduce 182) ((T_EVAL) reduce 182) ((T_EXIT) reduce 182) ((T_FILE) reduce 182) ((T_FINAL) reduce 182) ((T_FN) reduce 182) ((T_FOR) reduce 182) ((T_FOREACH) reduce 182) ((T_FUNCTION) reduce 182) ((T_FUNC_C) reduce 182) ((T_GLOBAL) reduce 182) ((T_GOTO) reduce 182) ((T_HALT_COMPILER) reduce 182) ((T_IF) reduce 182) ((T_INC) reduce 182) ((T_INCLUDE) reduce 182) ((T_INCLUDE_ONCE) reduce 182) ((T_INLINE_HTML) reduce 182) ((T_INTERFACE) reduce 182) ((T_INT_CAST) reduce 182) ((T_ISSET) reduce 182) ((T_LINE) reduce 182) ((T_LIST) reduce 182) ((T_LNUMBER) reduce 182) ((T_MATCH) reduce 182) ((T_METHOD_C) reduce 182) ((T_NAME_FULLY_QUALIFIED) reduce 182) ((T_NAME_QUALIFIED) reduce 182) ((T_NAME_RELATIVE) reduce 182) ((T_NEW) reduce 182) ((T_NS_C) reduce 182) ((T_OBJECT_CAST) reduce 182) ((T_PRINT) reduce 182) ((T_PROPERTY_C) reduce 182) ((T_READONLY) reduce 182) ((T_REQUIRE) reduce 182) ((T_REQUIRE_ONCE) reduce 182) ((T_RETURN) reduce 182) ((T_START_HEREDOC) reduce 182) ((T_STATIC) reduce 182) ((T_STRING) reduce 182) ((T_STRING_CAST) reduce 182) ((T_SWITCH) reduce 182) ((T_THROW) reduce 182) ((T_TRAIT) reduce 182) ((T_TRAIT_C) reduce 182) ((T_TRY) reduce 182) ((T_UNSET) reduce 182) ((T_UNSET_CAST) reduce 182) ((T_VARIABLE) reduce 182) ((T_WHILE) reduce 182) ((T_YIELD) reduce 182) ((T_YIELD_FROM) reduce 182) (("[") reduce 182) (("`") reduce 182) (("{") reduce 182) (("~") reduce 182)) 21584 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDFOR) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 21592 ((("!") reduce 187) (("\"") reduce 187) (("$") reduce 187) (("(") reduce 187) (("+") reduce 187) (("-") reduce 187) ((";") reduce 187) (("@") reduce 187) ((T_ABSTRACT) reduce 187) ((T_ARRAY) reduce 187) ((T_ARRAY_CAST) reduce 187) ((T_ATTRIBUTE) reduce 187) ((T_BOOL_CAST) reduce 187) ((T_BREAK) reduce 187) ((T_CLASS) reduce 187) ((T_CLASS_C) reduce 187) ((T_CLONE) reduce 187) ((T_CONSTANT_ENCAPSED_STRING) reduce 187) ((T_CONTINUE) reduce 187) ((T_DEC) reduce 187) ((T_DECLARE) reduce 187) ((T_DIR) reduce 187) ((T_DNUMBER) reduce 187) ((T_DO) reduce 187) ((T_DOUBLE_CAST) reduce 187) ((T_ECHO) reduce 187) ((T_EMPTY) reduce 187) ((T_ENDFOR) reduce 187) ((T_ENUM) reduce 187) ((T_EVAL) reduce 187) ((T_EXIT) reduce 187) ((T_FILE) reduce 187) ((T_FINAL) reduce 187) ((T_FN) reduce 187) ((T_FOR) reduce 187) ((T_FOREACH) reduce 187) ((T_FUNCTION) reduce 187) ((T_FUNC_C) reduce 187) ((T_GLOBAL) reduce 187) ((T_GOTO) reduce 187) ((T_HALT_COMPILER) reduce 187) ((T_IF) reduce 187) ((T_INC) reduce 187) ((T_INCLUDE) reduce 187) ((T_INCLUDE_ONCE) reduce 187) ((T_INLINE_HTML) reduce 187) ((T_INTERFACE) reduce 187) ((T_INT_CAST) reduce 187) ((T_ISSET) reduce 187) ((T_LINE) reduce 187) ((T_LIST) reduce 187) ((T_LNUMBER) reduce 187) ((T_MATCH) reduce 187) ((T_METHOD_C) reduce 187) ((T_NAME_FULLY_QUALIFIED) reduce 187) ((T_NAME_QUALIFIED) reduce 187) ((T_NAME_RELATIVE) reduce 187) ((T_NEW) reduce 187) ((T_NS_C) reduce 187) ((T_OBJECT_CAST) reduce 187) ((T_PRINT) reduce 187) ((T_PROPERTY_C) reduce 187) ((T_READONLY) reduce 187) ((T_REQUIRE) reduce 187) ((T_REQUIRE_ONCE) reduce 187) ((T_RETURN) reduce 187) ((T_START_HEREDOC) reduce 187) ((T_STATIC) reduce 187) ((T_STRING) reduce 187) ((T_STRING_CAST) reduce 187) ((T_SWITCH) reduce 187) ((T_THROW) reduce 187) ((T_TRAIT) reduce 187) ((T_TRAIT_C) reduce 187) ((T_TRY) reduce 187) ((T_UNSET) reduce 187) ((T_UNSET_CAST) reduce 187) ((T_VARIABLE) reduce 187) ((T_WHILE) reduce 187) ((T_YIELD) reduce 187) ((T_YIELD_FROM) reduce 187) (("[") reduce 187) (("`") reduce 187) (("{") reduce 187) (("~") reduce 187)) 21593 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDFOR) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 21596 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_EMPTY) reduce 248) ((T_ENDFOR) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("~") reduce 248)) 21600 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_EMPTY) reduce 249) ((T_ENDFOR) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("~") reduce 249)) 21605 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDFOR) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 21606 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDFOR) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 21609 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDFOR) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 21614 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDFOR) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 21617 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDFOR) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 21618 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDFOR) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 21621 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDFOR) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 21624 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_EMPTY) reduce 176) ((T_ENDFOR) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("~") reduce 176)) 21631 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDFOR) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 21636 ((("!") reduce 198) (("\"") reduce 198) (("$") reduce 198) (("(") reduce 198) (("+") reduce 198) (("-") reduce 198) ((";") reduce 198) (("@") reduce 198) ((T_ABSTRACT) reduce 198) ((T_ARRAY) reduce 198) ((T_ARRAY_CAST) reduce 198) ((T_ATTRIBUTE) reduce 198) ((T_BOOL_CAST) reduce 198) ((T_BREAK) reduce 198) ((T_CLASS) reduce 198) ((T_CLASS_C) reduce 198) ((T_CLONE) reduce 198) ((T_CONSTANT_ENCAPSED_STRING) reduce 198) ((T_CONTINUE) reduce 198) ((T_DEC) reduce 198) ((T_DECLARE) reduce 198) ((T_DIR) reduce 198) ((T_DNUMBER) reduce 198) ((T_DO) reduce 198) ((T_DOUBLE_CAST) reduce 198) ((T_ECHO) reduce 198) ((T_EMPTY) reduce 198) ((T_ENDFOR) reduce 198) ((T_ENUM) reduce 198) ((T_EVAL) reduce 198) ((T_EXIT) reduce 198) ((T_FILE) reduce 198) ((T_FINAL) reduce 198) ((T_FN) reduce 198) ((T_FOR) reduce 198) ((T_FOREACH) reduce 198) ((T_FUNCTION) reduce 198) ((T_FUNC_C) reduce 198) ((T_GLOBAL) reduce 198) ((T_GOTO) reduce 198) ((T_HALT_COMPILER) reduce 198) ((T_IF) reduce 198) ((T_INC) reduce 198) ((T_INCLUDE) reduce 198) ((T_INCLUDE_ONCE) reduce 198) ((T_INLINE_HTML) reduce 198) ((T_INTERFACE) reduce 198) ((T_INT_CAST) reduce 198) ((T_ISSET) reduce 198) ((T_LINE) reduce 198) ((T_LIST) reduce 198) ((T_LNUMBER) reduce 198) ((T_MATCH) reduce 198) ((T_METHOD_C) reduce 198) ((T_NAME_FULLY_QUALIFIED) reduce 198) ((T_NAME_QUALIFIED) reduce 198) ((T_NAME_RELATIVE) reduce 198) ((T_NEW) reduce 198) ((T_NS_C) reduce 198) ((T_OBJECT_CAST) reduce 198) ((T_PRINT) reduce 198) ((T_PROPERTY_C) reduce 198) ((T_READONLY) reduce 198) ((T_REQUIRE) reduce 198) ((T_REQUIRE_ONCE) reduce 198) ((T_RETURN) reduce 198) ((T_START_HEREDOC) reduce 198) ((T_STATIC) reduce 198) ((T_STRING) reduce 198) ((T_STRING_CAST) reduce 198) ((T_SWITCH) reduce 198) ((T_THROW) reduce 198) ((T_TRAIT) reduce 198) ((T_TRAIT_C) reduce 198) ((T_TRY) reduce 198) ((T_UNSET) reduce 198) ((T_UNSET_CAST) reduce 198) ((T_VARIABLE) reduce 198) ((T_WHILE) reduce 198) ((T_YIELD) reduce 198) ((T_YIELD_FROM) reduce 198) (("[") reduce 198) (("`") reduce 198) (("{") reduce 198) (("~") reduce 198)) 21641 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDFOR) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 21645 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDFOR) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 21647 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDFOR) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 21651 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDFOR) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 21654 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDFOR) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 21655 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDFOR) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 21657 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDFOR) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 21659 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDFOR) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 21665 ((("!") reduce 199) (("\"") reduce 199) (("$") reduce 199) (("(") reduce 199) (("+") reduce 199) (("-") reduce 199) ((";") reduce 199) (("@") reduce 199) ((T_ABSTRACT) reduce 199) ((T_ARRAY) reduce 199) ((T_ARRAY_CAST) reduce 199) ((T_ATTRIBUTE) reduce 199) ((T_BOOL_CAST) reduce 199) ((T_BREAK) reduce 199) ((T_CLASS) reduce 199) ((T_CLASS_C) reduce 199) ((T_CLONE) reduce 199) ((T_CONSTANT_ENCAPSED_STRING) reduce 199) ((T_CONTINUE) reduce 199) ((T_DEC) reduce 199) ((T_DECLARE) reduce 199) ((T_DIR) reduce 199) ((T_DNUMBER) reduce 199) ((T_DO) reduce 199) ((T_DOUBLE_CAST) reduce 199) ((T_ECHO) reduce 199) ((T_EMPTY) reduce 199) ((T_ENDFOR) reduce 199) ((T_ENUM) reduce 199) ((T_EVAL) reduce 199) ((T_EXIT) reduce 199) ((T_FILE) reduce 199) ((T_FINAL) reduce 199) ((T_FN) reduce 199) ((T_FOR) reduce 199) ((T_FOREACH) reduce 199) ((T_FUNCTION) reduce 199) ((T_FUNC_C) reduce 199) ((T_GLOBAL) reduce 199) ((T_GOTO) reduce 199) ((T_HALT_COMPILER) reduce 199) ((T_IF) reduce 199) ((T_INC) reduce 199) ((T_INCLUDE) reduce 199) ((T_INCLUDE_ONCE) reduce 199) ((T_INLINE_HTML) reduce 199) ((T_INTERFACE) reduce 199) ((T_INT_CAST) reduce 199) ((T_ISSET) reduce 199) ((T_LINE) reduce 199) ((T_LIST) reduce 199) ((T_LNUMBER) reduce 199) ((T_MATCH) reduce 199) ((T_METHOD_C) reduce 199) ((T_NAME_FULLY_QUALIFIED) reduce 199) ((T_NAME_QUALIFIED) reduce 199) ((T_NAME_RELATIVE) reduce 199) ((T_NEW) reduce 199) ((T_NS_C) reduce 199) ((T_OBJECT_CAST) reduce 199) ((T_PRINT) reduce 199) ((T_PROPERTY_C) reduce 199) ((T_READONLY) reduce 199) ((T_REQUIRE) reduce 199) ((T_REQUIRE_ONCE) reduce 199) ((T_RETURN) reduce 199) ((T_START_HEREDOC) reduce 199) ((T_STATIC) reduce 199) ((T_STRING) reduce 199) ((T_STRING_CAST) reduce 199) ((T_SWITCH) reduce 199) ((T_THROW) reduce 199) ((T_TRAIT) reduce 199) ((T_TRAIT_C) reduce 199) ((T_TRY) reduce 199) ((T_UNSET) reduce 199) ((T_UNSET_CAST) reduce 199) ((T_VARIABLE) reduce 199) ((T_WHILE) reduce 199) ((T_YIELD) reduce 199) ((T_YIELD_FROM) reduce 199) (("[") reduce 199) (("`") reduce 199) (("{") reduce 199) (("~") reduce 199)) 21669 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDFOR) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 21672 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDFOR) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 21674 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDFOR) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 21676 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDFOR) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 21686 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDFOR) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 21687 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDFOR) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 21690 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDFOR) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 21691 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDFOR) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 21700 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDFOR) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 21701 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDFOR) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 21704 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDFOR) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 21711 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENDFOR) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 21712 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDFOREACH) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 21714 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDFOR) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 21720 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDFOR) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 21725 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDFOR) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 21726 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDFOR) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 21729 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDFOR) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 21731 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDFOR) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 21738 ((("!") reduce 188) (("\"") reduce 188) (("$") reduce 188) (("(") reduce 188) (("+") reduce 188) (("-") reduce 188) ((";") reduce 188) (("@") reduce 188) ((T_ABSTRACT) reduce 188) ((T_ARRAY) reduce 188) ((T_ARRAY_CAST) reduce 188) ((T_ATTRIBUTE) reduce 188) ((T_BOOL_CAST) reduce 188) ((T_BREAK) reduce 188) ((T_CLASS) reduce 188) ((T_CLASS_C) reduce 188) ((T_CLONE) reduce 188) ((T_CONSTANT_ENCAPSED_STRING) reduce 188) ((T_CONTINUE) reduce 188) ((T_DEC) reduce 188) ((T_DECLARE) reduce 188) ((T_DIR) reduce 188) ((T_DNUMBER) reduce 188) ((T_DO) reduce 188) ((T_DOUBLE_CAST) reduce 188) ((T_ECHO) reduce 188) ((T_EMPTY) reduce 188) ((T_ENDFOR) reduce 188) ((T_ENUM) reduce 188) ((T_EVAL) reduce 188) ((T_EXIT) reduce 188) ((T_FILE) reduce 188) ((T_FINAL) reduce 188) ((T_FN) reduce 188) ((T_FOR) reduce 188) ((T_FOREACH) reduce 188) ((T_FUNCTION) reduce 188) ((T_FUNC_C) reduce 188) ((T_GLOBAL) reduce 188) ((T_GOTO) reduce 188) ((T_HALT_COMPILER) reduce 188) ((T_IF) reduce 188) ((T_INC) reduce 188) ((T_INCLUDE) reduce 188) ((T_INCLUDE_ONCE) reduce 188) ((T_INLINE_HTML) reduce 188) ((T_INTERFACE) reduce 188) ((T_INT_CAST) reduce 188) ((T_ISSET) reduce 188) ((T_LINE) reduce 188) ((T_LIST) reduce 188) ((T_LNUMBER) reduce 188) ((T_MATCH) reduce 188) ((T_METHOD_C) reduce 188) ((T_NAME_FULLY_QUALIFIED) reduce 188) ((T_NAME_QUALIFIED) reduce 188) ((T_NAME_RELATIVE) reduce 188) ((T_NEW) reduce 188) ((T_NS_C) reduce 188) ((T_OBJECT_CAST) reduce 188) ((T_PRINT) reduce 188) ((T_PROPERTY_C) reduce 188) ((T_READONLY) reduce 188) ((T_REQUIRE) reduce 188) ((T_REQUIRE_ONCE) reduce 188) ((T_RETURN) reduce 188) ((T_START_HEREDOC) reduce 188) ((T_STATIC) reduce 188) ((T_STRING) reduce 188) ((T_STRING_CAST) reduce 188) ((T_SWITCH) reduce 188) ((T_THROW) reduce 188) ((T_TRAIT) reduce 188) ((T_TRAIT_C) reduce 188) ((T_TRY) reduce 188) ((T_UNSET) reduce 188) ((T_UNSET_CAST) reduce 188) ((T_VARIABLE) reduce 188) ((T_WHILE) reduce 188) ((T_YIELD) reduce 188) ((T_YIELD_FROM) reduce 188) (("[") reduce 188) (("`") reduce 188) (("{") reduce 188) (("~") reduce 188)) 21740 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDFOR) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 21742 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDFOREACH) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 21748 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDFOREACH) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 21753 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDFOREACH) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 21754 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDFOREACH) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 21757 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDFOREACH) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 21759 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDFOREACH) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 21761 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDFOREACH) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 21762 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDFOREACH) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 21773 ((("!") reduce 477) (("\"") reduce 477) (("$") reduce 477) (("(") reduce 477) (("+") reduce 477) (("-") reduce 477) ((";") reduce 477) (("@") reduce 477) ((T_ABSTRACT) reduce 477) ((T_ARRAY) reduce 477) ((T_ARRAY_CAST) reduce 477) ((T_ATTRIBUTE) reduce 477) ((T_BOOL_CAST) reduce 477) ((T_BREAK) reduce 477) ((T_CLASS) reduce 477) ((T_CLASS_C) reduce 477) ((T_CLONE) reduce 477) ((T_CONSTANT_ENCAPSED_STRING) reduce 477) ((T_CONTINUE) reduce 477) ((T_DEC) reduce 477) ((T_DECLARE) reduce 477) ((T_DIR) reduce 477) ((T_DNUMBER) reduce 477) ((T_DO) reduce 477) ((T_DOUBLE_CAST) reduce 477) ((T_ECHO) reduce 477) ((T_EMPTY) reduce 477) ((T_ENDFOREACH) reduce 477) ((T_ENUM) reduce 477) ((T_EVAL) reduce 477) ((T_EXIT) reduce 477) ((T_FILE) reduce 477) ((T_FINAL) reduce 477) ((T_FN) reduce 477) ((T_FOR) reduce 477) ((T_FOREACH) reduce 477) ((T_FUNCTION) reduce 477) ((T_FUNC_C) reduce 477) ((T_GLOBAL) reduce 477) ((T_GOTO) reduce 477) ((T_HALT_COMPILER) reduce 477) ((T_IF) reduce 477) ((T_INC) reduce 477) ((T_INCLUDE) reduce 477) ((T_INCLUDE_ONCE) reduce 477) ((T_INLINE_HTML) reduce 477) ((T_INTERFACE) reduce 477) ((T_INT_CAST) reduce 477) ((T_ISSET) reduce 477) ((T_LINE) reduce 477) ((T_LIST) reduce 477) ((T_LNUMBER) reduce 477) ((T_MATCH) reduce 477) ((T_METHOD_C) reduce 477) ((T_NAME_FULLY_QUALIFIED) reduce 477) ((T_NAME_QUALIFIED) reduce 477) ((T_NAME_RELATIVE) reduce 477) ((T_NEW) reduce 477) ((T_NS_C) reduce 477) ((T_OBJECT_CAST) reduce 477) ((T_PRINT) reduce 477) ((T_PROPERTY_C) reduce 477) ((T_READONLY) reduce 477) ((T_REQUIRE) reduce 477) ((T_REQUIRE_ONCE) reduce 477) ((T_RETURN) reduce 477) ((T_START_HEREDOC) reduce 477) ((T_STATIC) reduce 477) ((T_STRING) reduce 477) ((T_STRING_CAST) reduce 477) ((T_SWITCH) reduce 477) ((T_THROW) reduce 477) ((T_TRAIT) reduce 477) ((T_TRAIT_C) reduce 477) ((T_TRY) reduce 477) ((T_UNSET) reduce 477) ((T_UNSET_CAST) reduce 477) ((T_VARIABLE) reduce 477) ((T_WHILE) reduce 477) ((T_YIELD) reduce 477) ((T_YIELD_FROM) reduce 477) (("[") reduce 477) (("`") reduce 477) (("{") reduce 477) (("~") reduce 477)) 21774 ((("!") reduce 182) (("\"") reduce 182) (("$") reduce 182) (("(") reduce 182) (("+") reduce 182) (("-") reduce 182) ((";") reduce 182) (("@") reduce 182) ((T_ABSTRACT) reduce 182) ((T_ARRAY) reduce 182) ((T_ARRAY_CAST) reduce 182) ((T_ATTRIBUTE) reduce 182) ((T_BOOL_CAST) reduce 182) ((T_BREAK) reduce 182) ((T_CLASS) reduce 182) ((T_CLASS_C) reduce 182) ((T_CLONE) reduce 182) ((T_CONSTANT_ENCAPSED_STRING) reduce 182) ((T_CONTINUE) reduce 182) ((T_DEC) reduce 182) ((T_DECLARE) reduce 182) ((T_DIR) reduce 182) ((T_DNUMBER) reduce 182) ((T_DO) reduce 182) ((T_DOUBLE_CAST) reduce 182) ((T_ECHO) reduce 182) ((T_EMPTY) reduce 182) ((T_ENDFOREACH) reduce 182) ((T_ENUM) reduce 182) ((T_EVAL) reduce 182) ((T_EXIT) reduce 182) ((T_FILE) reduce 182) ((T_FINAL) reduce 182) ((T_FN) reduce 182) ((T_FOR) reduce 182) ((T_FOREACH) reduce 182) ((T_FUNCTION) reduce 182) ((T_FUNC_C) reduce 182) ((T_GLOBAL) reduce 182) ((T_GOTO) reduce 182) ((T_HALT_COMPILER) reduce 182) ((T_IF) reduce 182) ((T_INC) reduce 182) ((T_INCLUDE) reduce 182) ((T_INCLUDE_ONCE) reduce 182) ((T_INLINE_HTML) reduce 182) ((T_INTERFACE) reduce 182) ((T_INT_CAST) reduce 182) ((T_ISSET) reduce 182) ((T_LINE) reduce 182) ((T_LIST) reduce 182) ((T_LNUMBER) reduce 182) ((T_MATCH) reduce 182) ((T_METHOD_C) reduce 182) ((T_NAME_FULLY_QUALIFIED) reduce 182) ((T_NAME_QUALIFIED) reduce 182) ((T_NAME_RELATIVE) reduce 182) ((T_NEW) reduce 182) ((T_NS_C) reduce 182) ((T_OBJECT_CAST) reduce 182) ((T_PRINT) reduce 182) ((T_PROPERTY_C) reduce 182) ((T_READONLY) reduce 182) ((T_REQUIRE) reduce 182) ((T_REQUIRE_ONCE) reduce 182) ((T_RETURN) reduce 182) ((T_START_HEREDOC) reduce 182) ((T_STATIC) reduce 182) ((T_STRING) reduce 182) ((T_STRING_CAST) reduce 182) ((T_SWITCH) reduce 182) ((T_THROW) reduce 182) ((T_TRAIT) reduce 182) ((T_TRAIT_C) reduce 182) ((T_TRY) reduce 182) ((T_UNSET) reduce 182) ((T_UNSET_CAST) reduce 182) ((T_VARIABLE) reduce 182) ((T_WHILE) reduce 182) ((T_YIELD) reduce 182) ((T_YIELD_FROM) reduce 182) (("[") reduce 182) (("`") reduce 182) (("{") reduce 182) (("~") reduce 182)) 21775 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDFOREACH) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 21783 ((("!") reduce 187) (("\"") reduce 187) (("$") reduce 187) (("(") reduce 187) (("+") reduce 187) (("-") reduce 187) ((";") reduce 187) (("@") reduce 187) ((T_ABSTRACT) reduce 187) ((T_ARRAY) reduce 187) ((T_ARRAY_CAST) reduce 187) ((T_ATTRIBUTE) reduce 187) ((T_BOOL_CAST) reduce 187) ((T_BREAK) reduce 187) ((T_CLASS) reduce 187) ((T_CLASS_C) reduce 187) ((T_CLONE) reduce 187) ((T_CONSTANT_ENCAPSED_STRING) reduce 187) ((T_CONTINUE) reduce 187) ((T_DEC) reduce 187) ((T_DECLARE) reduce 187) ((T_DIR) reduce 187) ((T_DNUMBER) reduce 187) ((T_DO) reduce 187) ((T_DOUBLE_CAST) reduce 187) ((T_ECHO) reduce 187) ((T_EMPTY) reduce 187) ((T_ENDFOREACH) reduce 187) ((T_ENUM) reduce 187) ((T_EVAL) reduce 187) ((T_EXIT) reduce 187) ((T_FILE) reduce 187) ((T_FINAL) reduce 187) ((T_FN) reduce 187) ((T_FOR) reduce 187) ((T_FOREACH) reduce 187) ((T_FUNCTION) reduce 187) ((T_FUNC_C) reduce 187) ((T_GLOBAL) reduce 187) ((T_GOTO) reduce 187) ((T_HALT_COMPILER) reduce 187) ((T_IF) reduce 187) ((T_INC) reduce 187) ((T_INCLUDE) reduce 187) ((T_INCLUDE_ONCE) reduce 187) ((T_INLINE_HTML) reduce 187) ((T_INTERFACE) reduce 187) ((T_INT_CAST) reduce 187) ((T_ISSET) reduce 187) ((T_LINE) reduce 187) ((T_LIST) reduce 187) ((T_LNUMBER) reduce 187) ((T_MATCH) reduce 187) ((T_METHOD_C) reduce 187) ((T_NAME_FULLY_QUALIFIED) reduce 187) ((T_NAME_QUALIFIED) reduce 187) ((T_NAME_RELATIVE) reduce 187) ((T_NEW) reduce 187) ((T_NS_C) reduce 187) ((T_OBJECT_CAST) reduce 187) ((T_PRINT) reduce 187) ((T_PROPERTY_C) reduce 187) ((T_READONLY) reduce 187) ((T_REQUIRE) reduce 187) ((T_REQUIRE_ONCE) reduce 187) ((T_RETURN) reduce 187) ((T_START_HEREDOC) reduce 187) ((T_STATIC) reduce 187) ((T_STRING) reduce 187) ((T_STRING_CAST) reduce 187) ((T_SWITCH) reduce 187) ((T_THROW) reduce 187) ((T_TRAIT) reduce 187) ((T_TRAIT_C) reduce 187) ((T_TRY) reduce 187) ((T_UNSET) reduce 187) ((T_UNSET_CAST) reduce 187) ((T_VARIABLE) reduce 187) ((T_WHILE) reduce 187) ((T_YIELD) reduce 187) ((T_YIELD_FROM) reduce 187) (("[") reduce 187) (("`") reduce 187) (("{") reduce 187) (("~") reduce 187)) 21784 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDFOREACH) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 21787 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_EMPTY) reduce 248) ((T_ENDFOREACH) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("~") reduce 248)) 21791 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_EMPTY) reduce 249) ((T_ENDFOREACH) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("~") reduce 249)) 21796 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDFOREACH) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 21797 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDFOREACH) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 21800 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDFOREACH) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 21805 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDFOREACH) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 21808 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDFOREACH) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 21809 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDFOREACH) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 21812 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDFOREACH) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 21815 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_EMPTY) reduce 176) ((T_ENDFOREACH) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("~") reduce 176)) 21822 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDFOREACH) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 21827 ((("!") reduce 198) (("\"") reduce 198) (("$") reduce 198) (("(") reduce 198) (("+") reduce 198) (("-") reduce 198) ((";") reduce 198) (("@") reduce 198) ((T_ABSTRACT) reduce 198) ((T_ARRAY) reduce 198) ((T_ARRAY_CAST) reduce 198) ((T_ATTRIBUTE) reduce 198) ((T_BOOL_CAST) reduce 198) ((T_BREAK) reduce 198) ((T_CLASS) reduce 198) ((T_CLASS_C) reduce 198) ((T_CLONE) reduce 198) ((T_CONSTANT_ENCAPSED_STRING) reduce 198) ((T_CONTINUE) reduce 198) ((T_DEC) reduce 198) ((T_DECLARE) reduce 198) ((T_DIR) reduce 198) ((T_DNUMBER) reduce 198) ((T_DO) reduce 198) ((T_DOUBLE_CAST) reduce 198) ((T_ECHO) reduce 198) ((T_EMPTY) reduce 198) ((T_ENDFOREACH) reduce 198) ((T_ENUM) reduce 198) ((T_EVAL) reduce 198) ((T_EXIT) reduce 198) ((T_FILE) reduce 198) ((T_FINAL) reduce 198) ((T_FN) reduce 198) ((T_FOR) reduce 198) ((T_FOREACH) reduce 198) ((T_FUNCTION) reduce 198) ((T_FUNC_C) reduce 198) ((T_GLOBAL) reduce 198) ((T_GOTO) reduce 198) ((T_HALT_COMPILER) reduce 198) ((T_IF) reduce 198) ((T_INC) reduce 198) ((T_INCLUDE) reduce 198) ((T_INCLUDE_ONCE) reduce 198) ((T_INLINE_HTML) reduce 198) ((T_INTERFACE) reduce 198) ((T_INT_CAST) reduce 198) ((T_ISSET) reduce 198) ((T_LINE) reduce 198) ((T_LIST) reduce 198) ((T_LNUMBER) reduce 198) ((T_MATCH) reduce 198) ((T_METHOD_C) reduce 198) ((T_NAME_FULLY_QUALIFIED) reduce 198) ((T_NAME_QUALIFIED) reduce 198) ((T_NAME_RELATIVE) reduce 198) ((T_NEW) reduce 198) ((T_NS_C) reduce 198) ((T_OBJECT_CAST) reduce 198) ((T_PRINT) reduce 198) ((T_PROPERTY_C) reduce 198) ((T_READONLY) reduce 198) ((T_REQUIRE) reduce 198) ((T_REQUIRE_ONCE) reduce 198) ((T_RETURN) reduce 198) ((T_START_HEREDOC) reduce 198) ((T_STATIC) reduce 198) ((T_STRING) reduce 198) ((T_STRING_CAST) reduce 198) ((T_SWITCH) reduce 198) ((T_THROW) reduce 198) ((T_TRAIT) reduce 198) ((T_TRAIT_C) reduce 198) ((T_TRY) reduce 198) ((T_UNSET) reduce 198) ((T_UNSET_CAST) reduce 198) ((T_VARIABLE) reduce 198) ((T_WHILE) reduce 198) ((T_YIELD) reduce 198) ((T_YIELD_FROM) reduce 198) (("[") reduce 198) (("`") reduce 198) (("{") reduce 198) (("~") reduce 198)) 21832 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDFOREACH) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 21836 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDFOREACH) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 21838 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDFOREACH) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 21842 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDFOREACH) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 21845 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDFOREACH) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 21846 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDFOREACH) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 21848 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDFOREACH) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 21850 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDFOREACH) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 21856 ((("!") reduce 199) (("\"") reduce 199) (("$") reduce 199) (("(") reduce 199) (("+") reduce 199) (("-") reduce 199) ((";") reduce 199) (("@") reduce 199) ((T_ABSTRACT) reduce 199) ((T_ARRAY) reduce 199) ((T_ARRAY_CAST) reduce 199) ((T_ATTRIBUTE) reduce 199) ((T_BOOL_CAST) reduce 199) ((T_BREAK) reduce 199) ((T_CLASS) reduce 199) ((T_CLASS_C) reduce 199) ((T_CLONE) reduce 199) ((T_CONSTANT_ENCAPSED_STRING) reduce 199) ((T_CONTINUE) reduce 199) ((T_DEC) reduce 199) ((T_DECLARE) reduce 199) ((T_DIR) reduce 199) ((T_DNUMBER) reduce 199) ((T_DO) reduce 199) ((T_DOUBLE_CAST) reduce 199) ((T_ECHO) reduce 199) ((T_EMPTY) reduce 199) ((T_ENDFOREACH) reduce 199) ((T_ENUM) reduce 199) ((T_EVAL) reduce 199) ((T_EXIT) reduce 199) ((T_FILE) reduce 199) ((T_FINAL) reduce 199) ((T_FN) reduce 199) ((T_FOR) reduce 199) ((T_FOREACH) reduce 199) ((T_FUNCTION) reduce 199) ((T_FUNC_C) reduce 199) ((T_GLOBAL) reduce 199) ((T_GOTO) reduce 199) ((T_HALT_COMPILER) reduce 199) ((T_IF) reduce 199) ((T_INC) reduce 199) ((T_INCLUDE) reduce 199) ((T_INCLUDE_ONCE) reduce 199) ((T_INLINE_HTML) reduce 199) ((T_INTERFACE) reduce 199) ((T_INT_CAST) reduce 199) ((T_ISSET) reduce 199) ((T_LINE) reduce 199) ((T_LIST) reduce 199) ((T_LNUMBER) reduce 199) ((T_MATCH) reduce 199) ((T_METHOD_C) reduce 199) ((T_NAME_FULLY_QUALIFIED) reduce 199) ((T_NAME_QUALIFIED) reduce 199) ((T_NAME_RELATIVE) reduce 199) ((T_NEW) reduce 199) ((T_NS_C) reduce 199) ((T_OBJECT_CAST) reduce 199) ((T_PRINT) reduce 199) ((T_PROPERTY_C) reduce 199) ((T_READONLY) reduce 199) ((T_REQUIRE) reduce 199) ((T_REQUIRE_ONCE) reduce 199) ((T_RETURN) reduce 199) ((T_START_HEREDOC) reduce 199) ((T_STATIC) reduce 199) ((T_STRING) reduce 199) ((T_STRING_CAST) reduce 199) ((T_SWITCH) reduce 199) ((T_THROW) reduce 199) ((T_TRAIT) reduce 199) ((T_TRAIT_C) reduce 199) ((T_TRY) reduce 199) ((T_UNSET) reduce 199) ((T_UNSET_CAST) reduce 199) ((T_VARIABLE) reduce 199) ((T_WHILE) reduce 199) ((T_YIELD) reduce 199) ((T_YIELD_FROM) reduce 199) (("[") reduce 199) (("`") reduce 199) (("{") reduce 199) (("~") reduce 199)) 21860 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDFOREACH) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 21863 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDFOREACH) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 21865 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDFOREACH) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 21867 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDFOREACH) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 21877 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDFOREACH) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 21878 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDFOREACH) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 21881 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDFOREACH) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 21882 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDFOREACH) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 21891 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDFOREACH) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 21892 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDFOREACH) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 21895 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDFOREACH) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 21902 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENDFOREACH) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 21903 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CASE) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DEFAULT) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDSWITCH) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 21905 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDFOREACH) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 21911 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDFOREACH) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 21916 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDFOREACH) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 21917 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDFOREACH) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 21920 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDFOREACH) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 21922 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDFOREACH) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 21929 ((("!") reduce 188) (("\"") reduce 188) (("$") reduce 188) (("(") reduce 188) (("+") reduce 188) (("-") reduce 188) ((";") reduce 188) (("@") reduce 188) ((T_ABSTRACT) reduce 188) ((T_ARRAY) reduce 188) ((T_ARRAY_CAST) reduce 188) ((T_ATTRIBUTE) reduce 188) ((T_BOOL_CAST) reduce 188) ((T_BREAK) reduce 188) ((T_CLASS) reduce 188) ((T_CLASS_C) reduce 188) ((T_CLONE) reduce 188) ((T_CONSTANT_ENCAPSED_STRING) reduce 188) ((T_CONTINUE) reduce 188) ((T_DEC) reduce 188) ((T_DECLARE) reduce 188) ((T_DIR) reduce 188) ((T_DNUMBER) reduce 188) ((T_DO) reduce 188) ((T_DOUBLE_CAST) reduce 188) ((T_ECHO) reduce 188) ((T_EMPTY) reduce 188) ((T_ENDFOREACH) reduce 188) ((T_ENUM) reduce 188) ((T_EVAL) reduce 188) ((T_EXIT) reduce 188) ((T_FILE) reduce 188) ((T_FINAL) reduce 188) ((T_FN) reduce 188) ((T_FOR) reduce 188) ((T_FOREACH) reduce 188) ((T_FUNCTION) reduce 188) ((T_FUNC_C) reduce 188) ((T_GLOBAL) reduce 188) ((T_GOTO) reduce 188) ((T_HALT_COMPILER) reduce 188) ((T_IF) reduce 188) ((T_INC) reduce 188) ((T_INCLUDE) reduce 188) ((T_INCLUDE_ONCE) reduce 188) ((T_INLINE_HTML) reduce 188) ((T_INTERFACE) reduce 188) ((T_INT_CAST) reduce 188) ((T_ISSET) reduce 188) ((T_LINE) reduce 188) ((T_LIST) reduce 188) ((T_LNUMBER) reduce 188) ((T_MATCH) reduce 188) ((T_METHOD_C) reduce 188) ((T_NAME_FULLY_QUALIFIED) reduce 188) ((T_NAME_QUALIFIED) reduce 188) ((T_NAME_RELATIVE) reduce 188) ((T_NEW) reduce 188) ((T_NS_C) reduce 188) ((T_OBJECT_CAST) reduce 188) ((T_PRINT) reduce 188) ((T_PROPERTY_C) reduce 188) ((T_READONLY) reduce 188) ((T_REQUIRE) reduce 188) ((T_REQUIRE_ONCE) reduce 188) ((T_RETURN) reduce 188) ((T_START_HEREDOC) reduce 188) ((T_STATIC) reduce 188) ((T_STRING) reduce 188) ((T_STRING_CAST) reduce 188) ((T_SWITCH) reduce 188) ((T_THROW) reduce 188) ((T_TRAIT) reduce 188) ((T_TRAIT_C) reduce 188) ((T_TRY) reduce 188) ((T_UNSET) reduce 188) ((T_UNSET_CAST) reduce 188) ((T_VARIABLE) reduce 188) ((T_WHILE) reduce 188) ((T_YIELD) reduce 188) ((T_YIELD_FROM) reduce 188) (("[") reduce 188) (("`") reduce 188) (("{") reduce 188) (("~") reduce 188)) 21931 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDFOREACH) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 21934 ((("(") reduce 564) ((")") reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) (("[") reduce 554)) 21935 ((("(") reduce 489) ((")") reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) (("[") reduce 489)) 21936 ((("(") reduce 491) ((")") reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) (("[") reduce 491)) 21941 ((("(") reduce 543) ((")") reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) (("[") reduce 543)) 21942 ((("(") reduce 565) ((")") reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) (("[") reduce 565)) 21943 ((("(") shift) ((")") reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) (("[") reduce 549)) 21944 ((("(") reduce 567) ((")") reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) (("[") reduce 567)) 21947 ((("(") reduce 566) ((")") reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) (("[") reduce 566)) 21948 ((("(") reduce 544) ((")") reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) (("[") reduce 544)) 21949 ((("(") shift) ((")") reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) (("[") reduce 550)) 21950 ((("(") reduce 545) ((")") reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) (("[") reduce 545)) 21951 (((")") reduce 213) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) (("[") reduce 531)) 21953 ((("(") reduce 503) ((")") reduce 215) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) (("[") reduce 503)) 21954 ((("(") reduce 488) ((")") reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) (("[") reduce 488)) 21957 (((")") reduce 214)) 21958 ((("(") reduce 553) ((")") reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) (("[") reduce 553)) 21961 ((("(") reduce 552) ((")") reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) (("[") reduce 552)) 21962 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CASE) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DEFAULT) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDSWITCH) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 21965 ((("(") reduce 564) ((")") reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) (("[") reduce 554)) 21966 ((("(") reduce 489) ((")") reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) (("[") reduce 489)) 21967 ((("(") reduce 491) ((")") reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) (("[") reduce 491)) 21972 ((("(") reduce 543) ((")") reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) (("[") reduce 543)) 21973 ((("(") reduce 565) ((")") reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) (("[") reduce 565)) 21974 ((("(") shift) ((")") reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) (("[") reduce 549)) 21975 ((("(") reduce 567) ((")") reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) (("[") reduce 567)) 21978 ((("(") reduce 566) ((")") reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) (("[") reduce 566)) 21979 ((("(") reduce 544) ((")") reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) (("[") reduce 544)) 21980 ((("(") shift) ((")") reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) (("[") reduce 550)) 21981 ((("(") reduce 545) ((")") reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) (("[") reduce 545)) 21982 (((")") reduce 213) ((T_DOUBLE_ARROW) reduce 213) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) (("[") reduce 531)) 21984 ((("(") reduce 503) ((")") reduce 215) ((T_DOUBLE_ARROW) reduce 215) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) (("[") reduce 503)) 21985 ((("(") reduce 488) ((")") reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) (("[") reduce 488)) 21988 (((")") reduce 214) ((T_DOUBLE_ARROW) reduce 214)) 21989 ((("(") reduce 553) ((")") reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) (("[") reduce 553)) 21992 ((("(") reduce 552) ((")") reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) (("[") reduce 552)) 22660 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CASE) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DEFAULT) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDSWITCH) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 22661 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CASE) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DEFAULT) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDSWITCH) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 22664 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CASE) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DEFAULT) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDSWITCH) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 22666 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CASE) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DEFAULT) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDSWITCH) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 22672 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CASE) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DEFAULT) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDSWITCH) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 22677 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CASE) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DEFAULT) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDSWITCH) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 22678 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CASE) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DEFAULT) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDSWITCH) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 22681 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CASE) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DEFAULT) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDSWITCH) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 22683 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CASE) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DEFAULT) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDSWITCH) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 22685 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CASE) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DEFAULT) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDSWITCH) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 22686 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CASE) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DEFAULT) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDSWITCH) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 22697 ((("!") reduce 477) (("\"") reduce 477) (("$") reduce 477) (("(") reduce 477) (("+") reduce 477) (("-") reduce 477) ((";") reduce 477) (("@") reduce 477) ((T_ABSTRACT) reduce 477) ((T_ARRAY) reduce 477) ((T_ARRAY_CAST) reduce 477) ((T_ATTRIBUTE) reduce 477) ((T_BOOL_CAST) reduce 477) ((T_BREAK) reduce 477) ((T_CASE) reduce 477) ((T_CLASS) reduce 477) ((T_CLASS_C) reduce 477) ((T_CLONE) reduce 477) ((T_CONSTANT_ENCAPSED_STRING) reduce 477) ((T_CONTINUE) reduce 477) ((T_DEC) reduce 477) ((T_DECLARE) reduce 477) ((T_DEFAULT) reduce 477) ((T_DIR) reduce 477) ((T_DNUMBER) reduce 477) ((T_DO) reduce 477) ((T_DOUBLE_CAST) reduce 477) ((T_ECHO) reduce 477) ((T_EMPTY) reduce 477) ((T_ENDSWITCH) reduce 477) ((T_ENUM) reduce 477) ((T_EVAL) reduce 477) ((T_EXIT) reduce 477) ((T_FILE) reduce 477) ((T_FINAL) reduce 477) ((T_FN) reduce 477) ((T_FOR) reduce 477) ((T_FOREACH) reduce 477) ((T_FUNCTION) reduce 477) ((T_FUNC_C) reduce 477) ((T_GLOBAL) reduce 477) ((T_GOTO) reduce 477) ((T_HALT_COMPILER) reduce 477) ((T_IF) reduce 477) ((T_INC) reduce 477) ((T_INCLUDE) reduce 477) ((T_INCLUDE_ONCE) reduce 477) ((T_INLINE_HTML) reduce 477) ((T_INTERFACE) reduce 477) ((T_INT_CAST) reduce 477) ((T_ISSET) reduce 477) ((T_LINE) reduce 477) ((T_LIST) reduce 477) ((T_LNUMBER) reduce 477) ((T_MATCH) reduce 477) ((T_METHOD_C) reduce 477) ((T_NAME_FULLY_QUALIFIED) reduce 477) ((T_NAME_QUALIFIED) reduce 477) ((T_NAME_RELATIVE) reduce 477) ((T_NEW) reduce 477) ((T_NS_C) reduce 477) ((T_OBJECT_CAST) reduce 477) ((T_PRINT) reduce 477) ((T_PROPERTY_C) reduce 477) ((T_READONLY) reduce 477) ((T_REQUIRE) reduce 477) ((T_REQUIRE_ONCE) reduce 477) ((T_RETURN) reduce 477) ((T_START_HEREDOC) reduce 477) ((T_STATIC) reduce 477) ((T_STRING) reduce 477) ((T_STRING_CAST) reduce 477) ((T_SWITCH) reduce 477) ((T_THROW) reduce 477) ((T_TRAIT) reduce 477) ((T_TRAIT_C) reduce 477) ((T_TRY) reduce 477) ((T_UNSET) reduce 477) ((T_UNSET_CAST) reduce 477) ((T_VARIABLE) reduce 477) ((T_WHILE) reduce 477) ((T_YIELD) reduce 477) ((T_YIELD_FROM) reduce 477) (("[") reduce 477) (("`") reduce 477) (("{") reduce 477) (("~") reduce 477)) 22698 ((("!") reduce 182) (("\"") reduce 182) (("$") reduce 182) (("(") reduce 182) (("+") reduce 182) (("-") reduce 182) ((";") reduce 182) (("@") reduce 182) ((T_ABSTRACT) reduce 182) ((T_ARRAY) reduce 182) ((T_ARRAY_CAST) reduce 182) ((T_ATTRIBUTE) reduce 182) ((T_BOOL_CAST) reduce 182) ((T_BREAK) reduce 182) ((T_CASE) reduce 182) ((T_CLASS) reduce 182) ((T_CLASS_C) reduce 182) ((T_CLONE) reduce 182) ((T_CONSTANT_ENCAPSED_STRING) reduce 182) ((T_CONTINUE) reduce 182) ((T_DEC) reduce 182) ((T_DECLARE) reduce 182) ((T_DEFAULT) reduce 182) ((T_DIR) reduce 182) ((T_DNUMBER) reduce 182) ((T_DO) reduce 182) ((T_DOUBLE_CAST) reduce 182) ((T_ECHO) reduce 182) ((T_EMPTY) reduce 182) ((T_ENDSWITCH) reduce 182) ((T_ENUM) reduce 182) ((T_EVAL) reduce 182) ((T_EXIT) reduce 182) ((T_FILE) reduce 182) ((T_FINAL) reduce 182) ((T_FN) reduce 182) ((T_FOR) reduce 182) ((T_FOREACH) reduce 182) ((T_FUNCTION) reduce 182) ((T_FUNC_C) reduce 182) ((T_GLOBAL) reduce 182) ((T_GOTO) reduce 182) ((T_HALT_COMPILER) reduce 182) ((T_IF) reduce 182) ((T_INC) reduce 182) ((T_INCLUDE) reduce 182) ((T_INCLUDE_ONCE) reduce 182) ((T_INLINE_HTML) reduce 182) ((T_INTERFACE) reduce 182) ((T_INT_CAST) reduce 182) ((T_ISSET) reduce 182) ((T_LINE) reduce 182) ((T_LIST) reduce 182) ((T_LNUMBER) reduce 182) ((T_MATCH) reduce 182) ((T_METHOD_C) reduce 182) ((T_NAME_FULLY_QUALIFIED) reduce 182) ((T_NAME_QUALIFIED) reduce 182) ((T_NAME_RELATIVE) reduce 182) ((T_NEW) reduce 182) ((T_NS_C) reduce 182) ((T_OBJECT_CAST) reduce 182) ((T_PRINT) reduce 182) ((T_PROPERTY_C) reduce 182) ((T_READONLY) reduce 182) ((T_REQUIRE) reduce 182) ((T_REQUIRE_ONCE) reduce 182) ((T_RETURN) reduce 182) ((T_START_HEREDOC) reduce 182) ((T_STATIC) reduce 182) ((T_STRING) reduce 182) ((T_STRING_CAST) reduce 182) ((T_SWITCH) reduce 182) ((T_THROW) reduce 182) ((T_TRAIT) reduce 182) ((T_TRAIT_C) reduce 182) ((T_TRY) reduce 182) ((T_UNSET) reduce 182) ((T_UNSET_CAST) reduce 182) ((T_VARIABLE) reduce 182) ((T_WHILE) reduce 182) ((T_YIELD) reduce 182) ((T_YIELD_FROM) reduce 182) (("[") reduce 182) (("`") reduce 182) (("{") reduce 182) (("~") reduce 182)) 22699 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CASE) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DEFAULT) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDSWITCH) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 22707 ((("!") reduce 187) (("\"") reduce 187) (("$") reduce 187) (("(") reduce 187) (("+") reduce 187) (("-") reduce 187) ((";") reduce 187) (("@") reduce 187) ((T_ABSTRACT) reduce 187) ((T_ARRAY) reduce 187) ((T_ARRAY_CAST) reduce 187) ((T_ATTRIBUTE) reduce 187) ((T_BOOL_CAST) reduce 187) ((T_BREAK) reduce 187) ((T_CASE) reduce 187) ((T_CLASS) reduce 187) ((T_CLASS_C) reduce 187) ((T_CLONE) reduce 187) ((T_CONSTANT_ENCAPSED_STRING) reduce 187) ((T_CONTINUE) reduce 187) ((T_DEC) reduce 187) ((T_DECLARE) reduce 187) ((T_DEFAULT) reduce 187) ((T_DIR) reduce 187) ((T_DNUMBER) reduce 187) ((T_DO) reduce 187) ((T_DOUBLE_CAST) reduce 187) ((T_ECHO) reduce 187) ((T_EMPTY) reduce 187) ((T_ENDSWITCH) reduce 187) ((T_ENUM) reduce 187) ((T_EVAL) reduce 187) ((T_EXIT) reduce 187) ((T_FILE) reduce 187) ((T_FINAL) reduce 187) ((T_FN) reduce 187) ((T_FOR) reduce 187) ((T_FOREACH) reduce 187) ((T_FUNCTION) reduce 187) ((T_FUNC_C) reduce 187) ((T_GLOBAL) reduce 187) ((T_GOTO) reduce 187) ((T_HALT_COMPILER) reduce 187) ((T_IF) reduce 187) ((T_INC) reduce 187) ((T_INCLUDE) reduce 187) ((T_INCLUDE_ONCE) reduce 187) ((T_INLINE_HTML) reduce 187) ((T_INTERFACE) reduce 187) ((T_INT_CAST) reduce 187) ((T_ISSET) reduce 187) ((T_LINE) reduce 187) ((T_LIST) reduce 187) ((T_LNUMBER) reduce 187) ((T_MATCH) reduce 187) ((T_METHOD_C) reduce 187) ((T_NAME_FULLY_QUALIFIED) reduce 187) ((T_NAME_QUALIFIED) reduce 187) ((T_NAME_RELATIVE) reduce 187) ((T_NEW) reduce 187) ((T_NS_C) reduce 187) ((T_OBJECT_CAST) reduce 187) ((T_PRINT) reduce 187) ((T_PROPERTY_C) reduce 187) ((T_READONLY) reduce 187) ((T_REQUIRE) reduce 187) ((T_REQUIRE_ONCE) reduce 187) ((T_RETURN) reduce 187) ((T_START_HEREDOC) reduce 187) ((T_STATIC) reduce 187) ((T_STRING) reduce 187) ((T_STRING_CAST) reduce 187) ((T_SWITCH) reduce 187) ((T_THROW) reduce 187) ((T_TRAIT) reduce 187) ((T_TRAIT_C) reduce 187) ((T_TRY) reduce 187) ((T_UNSET) reduce 187) ((T_UNSET_CAST) reduce 187) ((T_VARIABLE) reduce 187) ((T_WHILE) reduce 187) ((T_YIELD) reduce 187) ((T_YIELD_FROM) reduce 187) (("[") reduce 187) (("`") reduce 187) (("{") reduce 187) (("~") reduce 187)) 22708 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CASE) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DEFAULT) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDSWITCH) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 22711 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CASE) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DEFAULT) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_EMPTY) reduce 248) ((T_ENDSWITCH) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("~") reduce 248)) 22715 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CASE) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DEFAULT) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_EMPTY) reduce 249) ((T_ENDSWITCH) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("~") reduce 249)) 22720 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CASE) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DEFAULT) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDSWITCH) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 22721 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CASE) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DEFAULT) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDSWITCH) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 22724 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CASE) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DEFAULT) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDSWITCH) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 22729 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CASE) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DEFAULT) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDSWITCH) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 22732 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CASE) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DEFAULT) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDSWITCH) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 22733 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CASE) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DEFAULT) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDSWITCH) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 22736 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CASE) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DEFAULT) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDSWITCH) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 22739 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CASE) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DEFAULT) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_EMPTY) reduce 176) ((T_ENDSWITCH) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("~") reduce 176)) 22746 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CASE) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DEFAULT) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDSWITCH) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 22751 ((("!") reduce 198) (("\"") reduce 198) (("$") reduce 198) (("(") reduce 198) (("+") reduce 198) (("-") reduce 198) ((";") reduce 198) (("@") reduce 198) ((T_ABSTRACT) reduce 198) ((T_ARRAY) reduce 198) ((T_ARRAY_CAST) reduce 198) ((T_ATTRIBUTE) reduce 198) ((T_BOOL_CAST) reduce 198) ((T_BREAK) reduce 198) ((T_CASE) reduce 198) ((T_CLASS) reduce 198) ((T_CLASS_C) reduce 198) ((T_CLONE) reduce 198) ((T_CONSTANT_ENCAPSED_STRING) reduce 198) ((T_CONTINUE) reduce 198) ((T_DEC) reduce 198) ((T_DECLARE) reduce 198) ((T_DEFAULT) reduce 198) ((T_DIR) reduce 198) ((T_DNUMBER) reduce 198) ((T_DO) reduce 198) ((T_DOUBLE_CAST) reduce 198) ((T_ECHO) reduce 198) ((T_EMPTY) reduce 198) ((T_ENDSWITCH) reduce 198) ((T_ENUM) reduce 198) ((T_EVAL) reduce 198) ((T_EXIT) reduce 198) ((T_FILE) reduce 198) ((T_FINAL) reduce 198) ((T_FN) reduce 198) ((T_FOR) reduce 198) ((T_FOREACH) reduce 198) ((T_FUNCTION) reduce 198) ((T_FUNC_C) reduce 198) ((T_GLOBAL) reduce 198) ((T_GOTO) reduce 198) ((T_HALT_COMPILER) reduce 198) ((T_IF) reduce 198) ((T_INC) reduce 198) ((T_INCLUDE) reduce 198) ((T_INCLUDE_ONCE) reduce 198) ((T_INLINE_HTML) reduce 198) ((T_INTERFACE) reduce 198) ((T_INT_CAST) reduce 198) ((T_ISSET) reduce 198) ((T_LINE) reduce 198) ((T_LIST) reduce 198) ((T_LNUMBER) reduce 198) ((T_MATCH) reduce 198) ((T_METHOD_C) reduce 198) ((T_NAME_FULLY_QUALIFIED) reduce 198) ((T_NAME_QUALIFIED) reduce 198) ((T_NAME_RELATIVE) reduce 198) ((T_NEW) reduce 198) ((T_NS_C) reduce 198) ((T_OBJECT_CAST) reduce 198) ((T_PRINT) reduce 198) ((T_PROPERTY_C) reduce 198) ((T_READONLY) reduce 198) ((T_REQUIRE) reduce 198) ((T_REQUIRE_ONCE) reduce 198) ((T_RETURN) reduce 198) ((T_START_HEREDOC) reduce 198) ((T_STATIC) reduce 198) ((T_STRING) reduce 198) ((T_STRING_CAST) reduce 198) ((T_SWITCH) reduce 198) ((T_THROW) reduce 198) ((T_TRAIT) reduce 198) ((T_TRAIT_C) reduce 198) ((T_TRY) reduce 198) ((T_UNSET) reduce 198) ((T_UNSET_CAST) reduce 198) ((T_VARIABLE) reduce 198) ((T_WHILE) reduce 198) ((T_YIELD) reduce 198) ((T_YIELD_FROM) reduce 198) (("[") reduce 198) (("`") reduce 198) (("{") reduce 198) (("~") reduce 198)) 22756 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CASE) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DEFAULT) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDSWITCH) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 22760 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CASE) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DEFAULT) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDSWITCH) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 22762 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CASE) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DEFAULT) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDSWITCH) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 22766 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CASE) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DEFAULT) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDSWITCH) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 22769 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CASE) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DEFAULT) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDSWITCH) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 22770 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CASE) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DEFAULT) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDSWITCH) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 22772 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CASE) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DEFAULT) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDSWITCH) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 22774 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CASE) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DEFAULT) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDSWITCH) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 22780 ((("!") reduce 199) (("\"") reduce 199) (("$") reduce 199) (("(") reduce 199) (("+") reduce 199) (("-") reduce 199) ((";") reduce 199) (("@") reduce 199) ((T_ABSTRACT) reduce 199) ((T_ARRAY) reduce 199) ((T_ARRAY_CAST) reduce 199) ((T_ATTRIBUTE) reduce 199) ((T_BOOL_CAST) reduce 199) ((T_BREAK) reduce 199) ((T_CASE) reduce 199) ((T_CLASS) reduce 199) ((T_CLASS_C) reduce 199) ((T_CLONE) reduce 199) ((T_CONSTANT_ENCAPSED_STRING) reduce 199) ((T_CONTINUE) reduce 199) ((T_DEC) reduce 199) ((T_DECLARE) reduce 199) ((T_DEFAULT) reduce 199) ((T_DIR) reduce 199) ((T_DNUMBER) reduce 199) ((T_DO) reduce 199) ((T_DOUBLE_CAST) reduce 199) ((T_ECHO) reduce 199) ((T_EMPTY) reduce 199) ((T_ENDSWITCH) reduce 199) ((T_ENUM) reduce 199) ((T_EVAL) reduce 199) ((T_EXIT) reduce 199) ((T_FILE) reduce 199) ((T_FINAL) reduce 199) ((T_FN) reduce 199) ((T_FOR) reduce 199) ((T_FOREACH) reduce 199) ((T_FUNCTION) reduce 199) ((T_FUNC_C) reduce 199) ((T_GLOBAL) reduce 199) ((T_GOTO) reduce 199) ((T_HALT_COMPILER) reduce 199) ((T_IF) reduce 199) ((T_INC) reduce 199) ((T_INCLUDE) reduce 199) ((T_INCLUDE_ONCE) reduce 199) ((T_INLINE_HTML) reduce 199) ((T_INTERFACE) reduce 199) ((T_INT_CAST) reduce 199) ((T_ISSET) reduce 199) ((T_LINE) reduce 199) ((T_LIST) reduce 199) ((T_LNUMBER) reduce 199) ((T_MATCH) reduce 199) ((T_METHOD_C) reduce 199) ((T_NAME_FULLY_QUALIFIED) reduce 199) ((T_NAME_QUALIFIED) reduce 199) ((T_NAME_RELATIVE) reduce 199) ((T_NEW) reduce 199) ((T_NS_C) reduce 199) ((T_OBJECT_CAST) reduce 199) ((T_PRINT) reduce 199) ((T_PROPERTY_C) reduce 199) ((T_READONLY) reduce 199) ((T_REQUIRE) reduce 199) ((T_REQUIRE_ONCE) reduce 199) ((T_RETURN) reduce 199) ((T_START_HEREDOC) reduce 199) ((T_STATIC) reduce 199) ((T_STRING) reduce 199) ((T_STRING_CAST) reduce 199) ((T_SWITCH) reduce 199) ((T_THROW) reduce 199) ((T_TRAIT) reduce 199) ((T_TRAIT_C) reduce 199) ((T_TRY) reduce 199) ((T_UNSET) reduce 199) ((T_UNSET_CAST) reduce 199) ((T_VARIABLE) reduce 199) ((T_WHILE) reduce 199) ((T_YIELD) reduce 199) ((T_YIELD_FROM) reduce 199) (("[") reduce 199) (("`") reduce 199) (("{") reduce 199) (("~") reduce 199)) 22784 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CASE) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DEFAULT) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDSWITCH) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 22787 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CASE) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DEFAULT) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDSWITCH) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 22789 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CASE) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DEFAULT) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDSWITCH) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 22791 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CASE) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DEFAULT) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDSWITCH) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 22801 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CASE) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DEFAULT) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDSWITCH) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 22802 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CASE) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DEFAULT) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDSWITCH) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 22805 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CASE) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DEFAULT) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDSWITCH) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 22806 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CASE) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DEFAULT) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDSWITCH) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 22815 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CASE) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DEFAULT) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDSWITCH) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 22816 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CASE) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DEFAULT) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDSWITCH) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 22819 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CASE) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DEFAULT) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDSWITCH) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 22826 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CASE) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DEFAULT) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENDSWITCH) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 22828 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CASE) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DEFAULT) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDSWITCH) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 22834 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CASE) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DEFAULT) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDSWITCH) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 22839 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CASE) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DEFAULT) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDSWITCH) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 22840 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CASE) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DEFAULT) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDSWITCH) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 22843 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CASE) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DEFAULT) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDSWITCH) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 22845 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CASE) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DEFAULT) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDSWITCH) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 22852 ((("!") reduce 188) (("\"") reduce 188) (("$") reduce 188) (("(") reduce 188) (("+") reduce 188) (("-") reduce 188) ((";") reduce 188) (("@") reduce 188) ((T_ABSTRACT) reduce 188) ((T_ARRAY) reduce 188) ((T_ARRAY_CAST) reduce 188) ((T_ATTRIBUTE) reduce 188) ((T_BOOL_CAST) reduce 188) ((T_BREAK) reduce 188) ((T_CASE) reduce 188) ((T_CLASS) reduce 188) ((T_CLASS_C) reduce 188) ((T_CLONE) reduce 188) ((T_CONSTANT_ENCAPSED_STRING) reduce 188) ((T_CONTINUE) reduce 188) ((T_DEC) reduce 188) ((T_DECLARE) reduce 188) ((T_DEFAULT) reduce 188) ((T_DIR) reduce 188) ((T_DNUMBER) reduce 188) ((T_DO) reduce 188) ((T_DOUBLE_CAST) reduce 188) ((T_ECHO) reduce 188) ((T_EMPTY) reduce 188) ((T_ENDSWITCH) reduce 188) ((T_ENUM) reduce 188) ((T_EVAL) reduce 188) ((T_EXIT) reduce 188) ((T_FILE) reduce 188) ((T_FINAL) reduce 188) ((T_FN) reduce 188) ((T_FOR) reduce 188) ((T_FOREACH) reduce 188) ((T_FUNCTION) reduce 188) ((T_FUNC_C) reduce 188) ((T_GLOBAL) reduce 188) ((T_GOTO) reduce 188) ((T_HALT_COMPILER) reduce 188) ((T_IF) reduce 188) ((T_INC) reduce 188) ((T_INCLUDE) reduce 188) ((T_INCLUDE_ONCE) reduce 188) ((T_INLINE_HTML) reduce 188) ((T_INTERFACE) reduce 188) ((T_INT_CAST) reduce 188) ((T_ISSET) reduce 188) ((T_LINE) reduce 188) ((T_LIST) reduce 188) ((T_LNUMBER) reduce 188) ((T_MATCH) reduce 188) ((T_METHOD_C) reduce 188) ((T_NAME_FULLY_QUALIFIED) reduce 188) ((T_NAME_QUALIFIED) reduce 188) ((T_NAME_RELATIVE) reduce 188) ((T_NEW) reduce 188) ((T_NS_C) reduce 188) ((T_OBJECT_CAST) reduce 188) ((T_PRINT) reduce 188) ((T_PROPERTY_C) reduce 188) ((T_READONLY) reduce 188) ((T_REQUIRE) reduce 188) ((T_REQUIRE_ONCE) reduce 188) ((T_RETURN) reduce 188) ((T_START_HEREDOC) reduce 188) ((T_STATIC) reduce 188) ((T_STRING) reduce 188) ((T_STRING_CAST) reduce 188) ((T_SWITCH) reduce 188) ((T_THROW) reduce 188) ((T_TRAIT) reduce 188) ((T_TRAIT_C) reduce 188) ((T_TRY) reduce 188) ((T_UNSET) reduce 188) ((T_UNSET_CAST) reduce 188) ((T_VARIABLE) reduce 188) ((T_WHILE) reduce 188) ((T_YIELD) reduce 188) ((T_YIELD_FROM) reduce 188) (("[") reduce 188) (("`") reduce 188) (("{") reduce 188) (("~") reduce 188)) 22854 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CASE) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DEFAULT) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDSWITCH) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 22865 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 22867 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((":") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 22869 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 22870 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507)) 22874 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) ((":") reduce 500) ((";") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 22875 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 22876 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 22882 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 22884 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506)) 22886 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 22887 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 22888 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 22889 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 22891 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 22894 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 22900 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 22903 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 22905 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((":") reduce 551) ((";") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 22906 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) (("-") shift) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 22913 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((":") reduce 547) ((";") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 22914 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513)) 22916 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512)) 22917 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511)) 22918 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") shift) ((";") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22921 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((":") reduce 546) ((";") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 22922 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((":") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 22923 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) ((":") reduce 449) ((";") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449)) 22924 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((":") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 22925 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((":") reduce 514) ((";") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514)) 22926 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((":") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444)) 22927 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((":") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 22928 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 22929 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((":") reduce 542) ((";") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 22930 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((":") reduce 548) ((";") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 22931 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) ((":") reduce 388) ((";") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388)) 22934 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) ((":") reduce 432) ((";") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432)) 22965 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 416) ((";") reduce 416) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416)) 22966 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 419) ((";") reduce 419) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419)) 22967 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 428) (("/") shift) ((":") reduce 428) ((";") reduce 428) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428)) 22968 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 441) ((";") reduce 441) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441)) 22969 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 427) (("/") shift) ((":") reduce 427) ((";") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 22970 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((":") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 22971 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 415) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22972 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 413) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22973 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 414) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22974 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 438) ((";") reduce 438) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438)) 22975 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 434) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 22976 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 436) ((";") reduce 436) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 22977 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 433) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 22978 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 440) ((";") reduce 440) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440)) 22979 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 435) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 22982 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 22983 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 22984 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 22985 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((":") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 22986 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 22987 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((":") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 22988 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((":") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 22989 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((":") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 22990 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((":") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 22991 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((":") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 22992 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((":") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 22998 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((":") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 22999 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 23000 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 23001 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((":") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 23002 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 23005 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 23006 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 23008 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 23010 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((":") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 23011 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((":") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 23014 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((":") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 23015 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 448) ((";") reduce 448) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23016 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 411) ((";") reduce 411) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23017 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 412) ((";") reduce 412) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23018 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 417) ((";") reduce 417) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417)) 23019 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 418) ((";") reduce 418) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418)) 23023 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 446) ((";") reduce 446) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23024 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 447) ((";") reduce 447) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23025 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 439) ((";") reduce 439) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439)) 23026 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 437) ((";") reduce 437) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437)) 23027 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((":") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 23028 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 420) (("/") shift) ((":") reduce 420) ((";") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420)) 23029 ((("%") shift) (("*") shift) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) ((":") reduce 422) ((";") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 23030 ((("%") shift) (("*") shift) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) ((":") reduce 421) ((";") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 23031 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((":") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 23032 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((":") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 23034 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((":") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 23035 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((":") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 23036 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((":") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 23037 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((":") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 23038 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((":") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 23039 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((":") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 23040 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((":") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 23041 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((":") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 23042 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((":") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 23043 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((":") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 23044 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((":") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 23045 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((":") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 23046 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((":") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 23047 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((":") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 23048 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((":") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 23049 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((":") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 23050 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((":") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 23051 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((":") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 23052 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((":") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 23053 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((":") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 23054 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((":") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 23055 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((":") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 23056 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((":") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 23057 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((":") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 23058 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((":") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 23059 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((":") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 23060 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((":") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 23061 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((":") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 23062 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((":") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 23063 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((":") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 23064 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((":") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 23065 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((":") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 23066 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((":") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 23067 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((":") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 23068 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((":") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 23069 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((":") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 23070 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((":") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 23071 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((":") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 23072 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((":") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 23073 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((":") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 23074 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((":") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 23075 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((":") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 23076 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((":") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 23077 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((":") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 23078 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((":") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 23079 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((":") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 23080 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((":") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 23081 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((":") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 23082 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((":") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 23083 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((":") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 23084 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((":") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 23085 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((":") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 23086 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((":") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 23087 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((":") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 23088 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((":") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 23089 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((":") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 23090 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((":") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 23091 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((":") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 23092 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((":") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 23093 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((":") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 23094 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((":") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 23095 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((":") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 23096 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((":") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 23097 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((":") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 23098 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((":") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 23099 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((":") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 23100 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((":") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 23101 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((":") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 23102 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((":") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 23103 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((":") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 23104 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((":") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 23105 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((":") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 23106 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((":") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 23107 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((":") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 23108 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((":") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 23109 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((":") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 23110 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((":") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 23111 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((":") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 23112 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((":") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 23113 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525)) 23115 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((":") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 23116 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((":") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 23117 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((":") reduce 555) ((";") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 23120 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((":") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527)) 23122 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) ((";") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 23123 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((":") reduce 289) ((";") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 23127 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((":") reduce 290) ((";") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 23128 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 23133 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((":") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 23135 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((":") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407)) 23145 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 403) ((";") reduce 403) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23146 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 405) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23147 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 404) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23148 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 397) ((";") reduce 397) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23149 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 394) ((";") reduce 394) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23150 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 402) ((";") reduce 402) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23151 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 396) ((";") reduce 396) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23152 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 400) ((";") reduce 400) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23153 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 395) ((";") reduce 395) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23154 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 398) ((";") reduce 398) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23155 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 399) ((";") reduce 399) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23156 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 406) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23157 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 401) ((";") reduce 401) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23159 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 391) ((";") reduce 391) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23162 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((":") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 23165 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((":") reduce 547) ((";") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 23167 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((":") reduce 546) ((";") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 23169 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((":") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 23170 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((":") reduce 548) ((";") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 23171 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((":") reduce 392) ((";") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392)) 23175 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((":") reduce 555) ((";") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 23177 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 23178 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((":") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 23182 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((":") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 23183 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 23184 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 23187 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((":") reduce 554) ((";") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 23188 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 23189 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((":") reduce 491) ((";") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 23194 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((":") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 23195 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 23196 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((":") reduce 549) ((";") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 23197 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 23200 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 23201 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((":") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 23202 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((":") reduce 550) ((";") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 23203 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((":") reduce 545) ((";") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 23204 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 23205 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((":") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 23208 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((":") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 23209 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) ((";") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 23220 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((":") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 23221 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((":") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 23294 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 477) ((";") reduce 477) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23608 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((":") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 23611 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524)) 23613 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((":") reduce 554) ((";") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 23616 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526)) 23617 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) ((";") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 23618 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((":") reduce 491) ((";") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 23620 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((":") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 23621 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((":") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 23626 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((":") reduce 543) ((";") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 23627 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 23628 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((":") reduce 549) ((";") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 23629 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 23632 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 23633 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((":") reduce 544) ((";") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 23634 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((":") reduce 550) ((";") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 23635 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((":") reduce 545) ((";") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 23637 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 23639 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) ((":") reduce 503) ((";") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503)) 23641 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 390) ((";") reduce 390) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23642 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 465) ((";") reduce 465) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23653 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 23655 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((":") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 23657 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 23658 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507)) 23662 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) ((":") reduce 500) ((";") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_DOUBLE_ARROW) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 23663 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 23664 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 23670 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 23672 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506)) 23674 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 23675 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 23676 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 23677 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 23679 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 23682 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 23688 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 23691 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 23693 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((":") reduce 551) ((";") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 23694 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) (("-") shift) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 462) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 23701 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((":") reduce 547) ((";") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 23702 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513)) 23704 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512)) 23705 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511)) 23706 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 463) ((";") reduce 463) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23709 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((":") reduce 546) ((";") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 23710 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((":") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 23711 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) ((":") reduce 449) ((";") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449)) 23712 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((":") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 23713 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((":") reduce 514) ((";") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514)) 23714 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((":") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444)) 23715 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((":") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 23716 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 23717 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((":") reduce 542) ((";") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 23718 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((":") reduce 548) ((";") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 23719 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) ((":") reduce 388) ((";") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 388) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388)) 23722 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) ((":") reduce 432) ((";") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432)) 23753 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 416) ((";") reduce 416) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416)) 23754 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 419) ((";") reduce 419) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419)) 23755 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 428) (("/") shift) ((":") reduce 428) ((";") reduce 428) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428)) 23756 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 441) ((";") reduce 441) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441)) 23757 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 427) (("/") shift) ((":") reduce 427) ((";") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 23758 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((":") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 23759 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 415) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23760 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 413) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23761 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 414) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23762 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 438) ((";") reduce 438) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438)) 23763 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 434) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 23764 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 436) ((";") reduce 436) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 23765 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 433) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 23766 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 440) ((";") reduce 440) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440)) 23767 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 435) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 23770 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 23771 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 23772 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 23773 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((":") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 23774 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 23775 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((":") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 23776 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((":") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 23777 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((":") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 23778 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((":") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 23779 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((":") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 23780 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((":") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 23786 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((":") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 23787 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 23788 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 23789 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((":") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 23790 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 23793 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 23794 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 23796 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 23798 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((":") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 23799 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((":") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 23802 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((":") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 23803 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 448) ((";") reduce 448) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23804 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 411) ((";") reduce 411) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23805 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 412) ((";") reduce 412) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23806 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 417) ((";") reduce 417) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417)) 23807 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 418) ((";") reduce 418) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418)) 23811 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 446) ((";") reduce 446) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23812 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 447) ((";") reduce 447) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23813 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 439) ((";") reduce 439) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439)) 23814 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 437) ((";") reduce 437) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437)) 23815 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((":") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 23816 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 420) (("/") shift) ((":") reduce 420) ((";") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420)) 23817 ((("%") shift) (("*") shift) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) ((":") reduce 422) ((";") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 23818 ((("%") shift) (("*") shift) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) ((":") reduce 421) ((";") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 23819 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((":") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 23820 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((":") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 23822 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((":") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 23823 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((":") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 23824 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((":") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 23825 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((":") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 23826 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((":") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 23827 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((":") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 23828 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((":") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 23829 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((":") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 23830 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((":") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 23831 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((":") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 23832 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((":") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 23833 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((":") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 23834 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((":") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 23835 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((":") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 23836 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((":") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 23837 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((":") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 23838 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((":") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 23839 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((":") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 23840 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((":") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 23841 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((":") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 23842 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((":") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 23843 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((":") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 23844 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((":") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 23845 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((":") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 23846 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((":") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 23847 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((":") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 23848 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((":") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 23849 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((":") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 23850 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((":") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 23851 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((":") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 23852 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((":") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 23853 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((":") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 23854 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((":") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 23855 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((":") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 23856 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((":") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 23857 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((":") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 23858 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((":") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 23859 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((":") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 23860 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((":") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 23861 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((":") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 23862 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((":") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 23863 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((":") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 23864 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((":") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 23865 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((":") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 23866 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((":") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 23867 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((":") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 23868 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((":") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 23869 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((":") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 23870 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((":") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 23871 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((":") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 23872 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((":") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 23873 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((":") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 23874 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((":") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 23875 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((":") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 23876 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((":") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 23877 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((":") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 23878 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((":") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 23879 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((":") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 23880 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((":") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 23881 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((":") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 23882 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((":") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 23883 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((":") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 23884 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((":") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 23885 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((":") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 23886 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((":") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 23887 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((":") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 23888 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((":") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 23889 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((":") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 23890 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((":") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 23891 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((":") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 23892 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((":") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 23893 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((":") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 23894 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((":") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 23895 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((":") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 23896 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((":") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 23897 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((":") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 23898 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((":") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 23899 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((":") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 23900 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((":") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 23901 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525)) 23903 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((":") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 23904 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((":") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 23905 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((":") reduce 555) ((";") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 23908 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((":") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527)) 23910 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) ((";") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 23911 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((":") reduce 289) ((";") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 23915 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((":") reduce 290) ((";") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 23916 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 23921 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((":") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 23923 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((":") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407)) 23933 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 403) ((";") reduce 403) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23934 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 405) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23935 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 404) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23936 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 397) ((";") reduce 397) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23937 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 394) ((";") reduce 394) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23938 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 402) ((";") reduce 402) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23939 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 396) ((";") reduce 396) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23940 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 400) ((";") reduce 400) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23941 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 395) ((";") reduce 395) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23942 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 398) ((";") reduce 398) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23943 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 399) ((";") reduce 399) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23944 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 406) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23945 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 401) ((";") reduce 401) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23947 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 391) ((";") reduce 391) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23950 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((":") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 23953 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((":") reduce 547) ((";") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 23955 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((":") reduce 546) ((";") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 23957 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((":") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 23958 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((":") reduce 548) ((";") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 23959 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((":") reduce 392) ((";") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392)) 23963 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((":") reduce 555) ((";") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 23965 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 23966 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((":") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 23970 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((":") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 23971 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 23972 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 23975 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((":") reduce 554) ((";") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 23976 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 23977 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((":") reduce 491) ((";") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 23982 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((":") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 23983 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 23984 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((":") reduce 549) ((";") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 23985 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 23988 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 23989 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((":") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 23990 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((":") reduce 550) ((";") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 23991 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((":") reduce 545) ((";") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 23992 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 23993 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((":") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 23996 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((":") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 23997 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) ((";") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 24008 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((":") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 24009 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((":") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 24019 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 477) ((";") reduce 477) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24020 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((":") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 24022 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 464) ((";") reduce 464) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24024 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524)) 24026 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((":") reduce 554) ((";") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 24029 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526)) 24030 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) ((";") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 24031 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((":") reduce 491) ((";") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 24033 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((":") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 24034 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((":") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 24039 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((":") reduce 543) ((";") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 24040 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 24041 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((":") reduce 549) ((";") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 24042 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 24045 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 24046 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((":") reduce 544) ((";") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 24047 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((":") reduce 550) ((";") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 24048 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((":") reduce 545) ((";") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 24050 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 24052 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) ((":") reduce 503) ((";") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_DOUBLE_ARROW) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503)) 24054 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 390) ((";") reduce 390) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24055 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 465) ((";") reduce 465) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24058 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 464) ((";") reduce 464) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24059 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((":") reduce 456) ((";") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456)) 24060 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 466) ((";") reduce 466) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24061 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) ((":") reduce 452) ((";") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452)) 24062 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((":") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 24064 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 24066 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 24067 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508)) 24068 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 602) ((";") reduce 602) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 602) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24069 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 601) ((";") reduce 601) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 601) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24070 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) ((";") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 24071 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 461) ((";") reduce 461) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24072 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((":") reduce 454) ((";") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454)) 24074 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 24075 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 24076 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 24077 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((":") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 24078 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 24079 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) ((":") reduce 385) ((";") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385)) 24082 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((":") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 24083 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) ((":") reduce 387) ((";") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387)) 24084 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((":") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 24085 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((":") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 24086 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((":") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 24092 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((":") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 24093 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 24094 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((":") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 24095 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 24096 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) ((":") reduce 384) ((";") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384)) 24098 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 24099 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) ((":") reduce 386) ((";") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386)) 24107 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) ((":") reduce 383) ((";") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383)) 24109 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((":") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 24115 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) ((":") reduce 231) ((";") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_DOUBLE_ARROW) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231)) 24120 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 389) ((";") reduce 389) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24124 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) ((":") reduce 596) ((";") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_DOUBLE_ARROW) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596)) 24125 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) ((":") reduce 450) ((";") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450)) 24126 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 599) ((";") reduce 599) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 599) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24127 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 598) ((";") reduce 598) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 598) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24128 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((":") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408)) 24130 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) ((":") reduce 501) ((";") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_DOUBLE_ARROW) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501)) 24131 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((":") reduce 457) ((";") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457)) 24132 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((":") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289)) 24136 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((":") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290)) 24137 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 24140 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) ((":") reduce 600) ((";") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_DOUBLE_ARROW) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600)) 24143 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) ((":") reduce 597) ((";") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_DOUBLE_ARROW) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597)) 24144 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) ((":") reduce 451) ((";") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451)) 24145 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((":") reduce 410) ((";") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410)) 24146 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((":") reduce 393) ((";") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 24147 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((":") reduce 455) ((";") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455)) 24148 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) ((":") reduce 453) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453)) 24151 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) ((":") reduce 502) ((";") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_DOUBLE_ARROW) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502)) 24152 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((":") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 24153 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) ((":") reduce 430) ((";") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430)) 24154 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) ((":") reduce 429) ((";") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429)) 24156 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) ((":") reduce 443) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443)) 24157 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((":") reduce 553) ((";") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 24160 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((":") reduce 552) ((";") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 24162 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 24163 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) ((":") reduce 431) ((";") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431)) 24164 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((":") reduce 456) ((";") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456)) 24165 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 466) ((";") reduce 466) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24166 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) ((":") reduce 452) ((";") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452)) 24167 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((":") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 24169 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 24171 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 24172 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508)) 24173 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 602) ((";") reduce 602) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24174 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 601) ((";") reduce 601) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24175 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) ((";") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 24176 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 461) ((";") reduce 461) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24177 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((":") reduce 454) ((";") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454)) 24179 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 24180 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 24181 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 24182 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((":") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 24183 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 24184 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) ((":") reduce 385) ((";") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385)) 24187 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((":") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 24188 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) ((":") reduce 387) ((";") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387)) 24189 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((":") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 24190 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((":") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 24191 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((":") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 24197 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((":") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 24198 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 24199 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((":") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 24200 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 24201 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) ((":") reduce 384) ((";") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384)) 24203 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 24204 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) ((":") reduce 386) ((";") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386)) 24212 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) ((":") reduce 383) ((";") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383)) 24214 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((":") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 24220 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) ((":") reduce 231) ((";") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231)) 24225 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 389) ((";") reduce 389) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24229 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) ((":") reduce 596) ((";") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596)) 24230 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) ((":") reduce 450) ((";") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450)) 24231 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 599) ((";") reduce 599) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24232 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 598) ((";") reduce 598) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24233 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((":") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408)) 24235 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) ((":") reduce 501) ((";") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501)) 24236 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((":") reduce 457) ((";") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457)) 24237 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((":") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289)) 24241 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((":") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290)) 24242 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 24245 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) ((":") reduce 600) ((";") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600)) 24248 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) ((":") reduce 597) ((";") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597)) 24249 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) ((":") reduce 451) ((";") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451)) 24250 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((":") reduce 410) ((";") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410)) 24251 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((":") reduce 393) ((";") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 24252 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((":") reduce 455) ((";") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455)) 24253 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) ((":") reduce 453) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453)) 24256 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) ((":") reduce 502) ((";") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502)) 24257 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((":") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 24258 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) ((":") reduce 430) ((";") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430)) 24259 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) ((":") reduce 429) ((";") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429)) 24261 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) ((":") reduce 443) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443)) 24262 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((":") reduce 553) ((";") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 24265 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((":") reduce 552) ((";") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 24267 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 24268 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) ((":") reduce 431) ((";") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431)) 24270 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 227) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 227) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDSWITCH) reduce 227) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 24932 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CASE) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DEFAULT) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("}") reduce 225) (("~") reduce 225)) 24933 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CASE) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DEFAULT) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 24935 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CASE) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DEFAULT) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 24937 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CASE) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DEFAULT) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 24939 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CASE) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DEFAULT) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 24941 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CASE) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DEFAULT) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 24951 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CASE) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DEFAULT) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 24952 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CASE) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DEFAULT) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("}") reduce 218) (("~") reduce 218)) 24955 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CASE) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DEFAULT) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("}") reduce 219) (("~") reduce 219)) 24956 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CASE) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DEFAULT) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 24965 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CASE) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DEFAULT) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 24966 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CASE) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DEFAULT) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("}") reduce 216) (("~") reduce 216)) 24969 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CASE) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DEFAULT) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("}") reduce 217) (("~") reduce 217)) 24971 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CASE) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DEFAULT) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 24977 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CASE) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DEFAULT) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 24982 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CASE) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DEFAULT) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 24983 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CASE) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DEFAULT) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("}") reduce 220) (("~") reduce 220)) 24986 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CASE) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DEFAULT) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("}") reduce 221) (("~") reduce 221)) 24988 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CASE) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DEFAULT) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 24990 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CASE) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DEFAULT) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 24991 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CASE) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DEFAULT) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_EMPTY) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("}") reduce 245) (("~") reduce 245)) 25002 ((("!") reduce 477) (("\"") reduce 477) (("$") reduce 477) (("(") reduce 477) (("+") reduce 477) (("-") reduce 477) ((";") reduce 477) (("@") reduce 477) ((T_ABSTRACT) reduce 477) ((T_ARRAY) reduce 477) ((T_ARRAY_CAST) reduce 477) ((T_ATTRIBUTE) reduce 477) ((T_BOOL_CAST) reduce 477) ((T_BREAK) reduce 477) ((T_CASE) reduce 477) ((T_CLASS) reduce 477) ((T_CLASS_C) reduce 477) ((T_CLONE) reduce 477) ((T_CONSTANT_ENCAPSED_STRING) reduce 477) ((T_CONTINUE) reduce 477) ((T_DEC) reduce 477) ((T_DECLARE) reduce 477) ((T_DEFAULT) reduce 477) ((T_DIR) reduce 477) ((T_DNUMBER) reduce 477) ((T_DO) reduce 477) ((T_DOUBLE_CAST) reduce 477) ((T_ECHO) reduce 477) ((T_EMPTY) reduce 477) ((T_ENUM) reduce 477) ((T_EVAL) reduce 477) ((T_EXIT) reduce 477) ((T_FILE) reduce 477) ((T_FINAL) reduce 477) ((T_FN) reduce 477) ((T_FOR) reduce 477) ((T_FOREACH) reduce 477) ((T_FUNCTION) reduce 477) ((T_FUNC_C) reduce 477) ((T_GLOBAL) reduce 477) ((T_GOTO) reduce 477) ((T_HALT_COMPILER) reduce 477) ((T_IF) reduce 477) ((T_INC) reduce 477) ((T_INCLUDE) reduce 477) ((T_INCLUDE_ONCE) reduce 477) ((T_INLINE_HTML) reduce 477) ((T_INTERFACE) reduce 477) ((T_INT_CAST) reduce 477) ((T_ISSET) reduce 477) ((T_LINE) reduce 477) ((T_LIST) reduce 477) ((T_LNUMBER) reduce 477) ((T_MATCH) reduce 477) ((T_METHOD_C) reduce 477) ((T_NAME_FULLY_QUALIFIED) reduce 477) ((T_NAME_QUALIFIED) reduce 477) ((T_NAME_RELATIVE) reduce 477) ((T_NEW) reduce 477) ((T_NS_C) reduce 477) ((T_OBJECT_CAST) reduce 477) ((T_PRINT) reduce 477) ((T_PROPERTY_C) reduce 477) ((T_READONLY) reduce 477) ((T_REQUIRE) reduce 477) ((T_REQUIRE_ONCE) reduce 477) ((T_RETURN) reduce 477) ((T_START_HEREDOC) reduce 477) ((T_STATIC) reduce 477) ((T_STRING) reduce 477) ((T_STRING_CAST) reduce 477) ((T_SWITCH) reduce 477) ((T_THROW) reduce 477) ((T_TRAIT) reduce 477) ((T_TRAIT_C) reduce 477) ((T_TRY) reduce 477) ((T_UNSET) reduce 477) ((T_UNSET_CAST) reduce 477) ((T_VARIABLE) reduce 477) ((T_WHILE) reduce 477) ((T_YIELD) reduce 477) ((T_YIELD_FROM) reduce 477) (("[") reduce 477) (("`") reduce 477) (("{") reduce 477) (("}") reduce 477) (("~") reduce 477)) 25003 ((("!") reduce 182) (("\"") reduce 182) (("$") reduce 182) (("(") reduce 182) (("+") reduce 182) (("-") reduce 182) ((";") reduce 182) (("@") reduce 182) ((T_ABSTRACT) reduce 182) ((T_ARRAY) reduce 182) ((T_ARRAY_CAST) reduce 182) ((T_ATTRIBUTE) reduce 182) ((T_BOOL_CAST) reduce 182) ((T_BREAK) reduce 182) ((T_CASE) reduce 182) ((T_CLASS) reduce 182) ((T_CLASS_C) reduce 182) ((T_CLONE) reduce 182) ((T_CONSTANT_ENCAPSED_STRING) reduce 182) ((T_CONTINUE) reduce 182) ((T_DEC) reduce 182) ((T_DECLARE) reduce 182) ((T_DEFAULT) reduce 182) ((T_DIR) reduce 182) ((T_DNUMBER) reduce 182) ((T_DO) reduce 182) ((T_DOUBLE_CAST) reduce 182) ((T_ECHO) reduce 182) ((T_EMPTY) reduce 182) ((T_ENUM) reduce 182) ((T_EVAL) reduce 182) ((T_EXIT) reduce 182) ((T_FILE) reduce 182) ((T_FINAL) reduce 182) ((T_FN) reduce 182) ((T_FOR) reduce 182) ((T_FOREACH) reduce 182) ((T_FUNCTION) reduce 182) ((T_FUNC_C) reduce 182) ((T_GLOBAL) reduce 182) ((T_GOTO) reduce 182) ((T_HALT_COMPILER) reduce 182) ((T_IF) reduce 182) ((T_INC) reduce 182) ((T_INCLUDE) reduce 182) ((T_INCLUDE_ONCE) reduce 182) ((T_INLINE_HTML) reduce 182) ((T_INTERFACE) reduce 182) ((T_INT_CAST) reduce 182) ((T_ISSET) reduce 182) ((T_LINE) reduce 182) ((T_LIST) reduce 182) ((T_LNUMBER) reduce 182) ((T_MATCH) reduce 182) ((T_METHOD_C) reduce 182) ((T_NAME_FULLY_QUALIFIED) reduce 182) ((T_NAME_QUALIFIED) reduce 182) ((T_NAME_RELATIVE) reduce 182) ((T_NEW) reduce 182) ((T_NS_C) reduce 182) ((T_OBJECT_CAST) reduce 182) ((T_PRINT) reduce 182) ((T_PROPERTY_C) reduce 182) ((T_READONLY) reduce 182) ((T_REQUIRE) reduce 182) ((T_REQUIRE_ONCE) reduce 182) ((T_RETURN) reduce 182) ((T_START_HEREDOC) reduce 182) ((T_STATIC) reduce 182) ((T_STRING) reduce 182) ((T_STRING_CAST) reduce 182) ((T_SWITCH) reduce 182) ((T_THROW) reduce 182) ((T_TRAIT) reduce 182) ((T_TRAIT_C) reduce 182) ((T_TRY) reduce 182) ((T_UNSET) reduce 182) ((T_UNSET_CAST) reduce 182) ((T_VARIABLE) reduce 182) ((T_WHILE) reduce 182) ((T_YIELD) reduce 182) ((T_YIELD_FROM) reduce 182) (("[") reduce 182) (("`") reduce 182) (("{") reduce 182) (("}") reduce 182) (("~") reduce 182)) 25004 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CASE) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DEFAULT) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 25012 ((("!") reduce 187) (("\"") reduce 187) (("$") reduce 187) (("(") reduce 187) (("+") reduce 187) (("-") reduce 187) ((";") reduce 187) (("@") reduce 187) ((T_ABSTRACT) reduce 187) ((T_ARRAY) reduce 187) ((T_ARRAY_CAST) reduce 187) ((T_ATTRIBUTE) reduce 187) ((T_BOOL_CAST) reduce 187) ((T_BREAK) reduce 187) ((T_CASE) reduce 187) ((T_CLASS) reduce 187) ((T_CLASS_C) reduce 187) ((T_CLONE) reduce 187) ((T_CONSTANT_ENCAPSED_STRING) reduce 187) ((T_CONTINUE) reduce 187) ((T_DEC) reduce 187) ((T_DECLARE) reduce 187) ((T_DEFAULT) reduce 187) ((T_DIR) reduce 187) ((T_DNUMBER) reduce 187) ((T_DO) reduce 187) ((T_DOUBLE_CAST) reduce 187) ((T_ECHO) reduce 187) ((T_EMPTY) reduce 187) ((T_ENUM) reduce 187) ((T_EVAL) reduce 187) ((T_EXIT) reduce 187) ((T_FILE) reduce 187) ((T_FINAL) reduce 187) ((T_FN) reduce 187) ((T_FOR) reduce 187) ((T_FOREACH) reduce 187) ((T_FUNCTION) reduce 187) ((T_FUNC_C) reduce 187) ((T_GLOBAL) reduce 187) ((T_GOTO) reduce 187) ((T_HALT_COMPILER) reduce 187) ((T_IF) reduce 187) ((T_INC) reduce 187) ((T_INCLUDE) reduce 187) ((T_INCLUDE_ONCE) reduce 187) ((T_INLINE_HTML) reduce 187) ((T_INTERFACE) reduce 187) ((T_INT_CAST) reduce 187) ((T_ISSET) reduce 187) ((T_LINE) reduce 187) ((T_LIST) reduce 187) ((T_LNUMBER) reduce 187) ((T_MATCH) reduce 187) ((T_METHOD_C) reduce 187) ((T_NAME_FULLY_QUALIFIED) reduce 187) ((T_NAME_QUALIFIED) reduce 187) ((T_NAME_RELATIVE) reduce 187) ((T_NEW) reduce 187) ((T_NS_C) reduce 187) ((T_OBJECT_CAST) reduce 187) ((T_PRINT) reduce 187) ((T_PROPERTY_C) reduce 187) ((T_READONLY) reduce 187) ((T_REQUIRE) reduce 187) ((T_REQUIRE_ONCE) reduce 187) ((T_RETURN) reduce 187) ((T_START_HEREDOC) reduce 187) ((T_STATIC) reduce 187) ((T_STRING) reduce 187) ((T_STRING_CAST) reduce 187) ((T_SWITCH) reduce 187) ((T_THROW) reduce 187) ((T_TRAIT) reduce 187) ((T_TRAIT_C) reduce 187) ((T_TRY) reduce 187) ((T_UNSET) reduce 187) ((T_UNSET_CAST) reduce 187) ((T_VARIABLE) reduce 187) ((T_WHILE) reduce 187) ((T_YIELD) reduce 187) ((T_YIELD_FROM) reduce 187) (("[") reduce 187) (("`") reduce 187) (("{") reduce 187) (("}") reduce 187) (("~") reduce 187)) 25013 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CASE) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DEFAULT) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("}") reduce 144) (("~") reduce 144)) 25016 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CASE) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DEFAULT) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_EMPTY) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("}") reduce 248) (("~") reduce 248)) 25020 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CASE) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DEFAULT) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_EMPTY) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("}") reduce 249) (("~") reduce 249)) 25025 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CASE) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DEFAULT) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("}") reduce 240) (("~") reduce 240)) 25026 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CASE) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DEFAULT) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 25029 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CASE) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DEFAULT) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("}") reduce 241) (("~") reduce 241)) 25034 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CASE) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DEFAULT) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 25037 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CASE) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DEFAULT) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 25038 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CASE) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DEFAULT) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("}") reduce 175) (("~") reduce 175)) 25041 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CASE) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DEFAULT) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 25044 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CASE) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DEFAULT) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_EMPTY) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("}") reduce 176) (("~") reduce 176)) 25051 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CASE) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DEFAULT) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("}") reduce 170) (("~") reduce 170)) 25056 ((("!") reduce 198) (("\"") reduce 198) (("$") reduce 198) (("(") reduce 198) (("+") reduce 198) (("-") reduce 198) ((";") reduce 198) (("@") reduce 198) ((T_ABSTRACT) reduce 198) ((T_ARRAY) reduce 198) ((T_ARRAY_CAST) reduce 198) ((T_ATTRIBUTE) reduce 198) ((T_BOOL_CAST) reduce 198) ((T_BREAK) reduce 198) ((T_CASE) reduce 198) ((T_CLASS) reduce 198) ((T_CLASS_C) reduce 198) ((T_CLONE) reduce 198) ((T_CONSTANT_ENCAPSED_STRING) reduce 198) ((T_CONTINUE) reduce 198) ((T_DEC) reduce 198) ((T_DECLARE) reduce 198) ((T_DEFAULT) reduce 198) ((T_DIR) reduce 198) ((T_DNUMBER) reduce 198) ((T_DO) reduce 198) ((T_DOUBLE_CAST) reduce 198) ((T_ECHO) reduce 198) ((T_EMPTY) reduce 198) ((T_ENUM) reduce 198) ((T_EVAL) reduce 198) ((T_EXIT) reduce 198) ((T_FILE) reduce 198) ((T_FINAL) reduce 198) ((T_FN) reduce 198) ((T_FOR) reduce 198) ((T_FOREACH) reduce 198) ((T_FUNCTION) reduce 198) ((T_FUNC_C) reduce 198) ((T_GLOBAL) reduce 198) ((T_GOTO) reduce 198) ((T_HALT_COMPILER) reduce 198) ((T_IF) reduce 198) ((T_INC) reduce 198) ((T_INCLUDE) reduce 198) ((T_INCLUDE_ONCE) reduce 198) ((T_INLINE_HTML) reduce 198) ((T_INTERFACE) reduce 198) ((T_INT_CAST) reduce 198) ((T_ISSET) reduce 198) ((T_LINE) reduce 198) ((T_LIST) reduce 198) ((T_LNUMBER) reduce 198) ((T_MATCH) reduce 198) ((T_METHOD_C) reduce 198) ((T_NAME_FULLY_QUALIFIED) reduce 198) ((T_NAME_QUALIFIED) reduce 198) ((T_NAME_RELATIVE) reduce 198) ((T_NEW) reduce 198) ((T_NS_C) reduce 198) ((T_OBJECT_CAST) reduce 198) ((T_PRINT) reduce 198) ((T_PROPERTY_C) reduce 198) ((T_READONLY) reduce 198) ((T_REQUIRE) reduce 198) ((T_REQUIRE_ONCE) reduce 198) ((T_RETURN) reduce 198) ((T_START_HEREDOC) reduce 198) ((T_STATIC) reduce 198) ((T_STRING) reduce 198) ((T_STRING_CAST) reduce 198) ((T_SWITCH) reduce 198) ((T_THROW) reduce 198) ((T_TRAIT) reduce 198) ((T_TRAIT_C) reduce 198) ((T_TRY) reduce 198) ((T_UNSET) reduce 198) ((T_UNSET_CAST) reduce 198) ((T_VARIABLE) reduce 198) ((T_WHILE) reduce 198) ((T_YIELD) reduce 198) ((T_YIELD_FROM) reduce 198) (("[") reduce 198) (("`") reduce 198) (("{") reduce 198) (("}") reduce 198) (("~") reduce 198)) 25061 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CASE) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DEFAULT) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 25065 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CASE) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DEFAULT) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("}") reduce 222) (("~") reduce 222)) 25067 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CASE) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DEFAULT) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("}") reduce 223) (("~") reduce 223)) 25071 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CASE) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DEFAULT) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("}") reduce 224) (("~") reduce 224)) 25074 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CASE) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DEFAULT) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("}") reduce 225) (("~") reduce 225)) 25075 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CASE) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DEFAULT) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 25077 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CASE) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DEFAULT) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 25079 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CASE) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DEFAULT) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 25085 ((("!") reduce 199) (("\"") reduce 199) (("$") reduce 199) (("(") reduce 199) (("+") reduce 199) (("-") reduce 199) ((";") reduce 199) (("@") reduce 199) ((T_ABSTRACT) reduce 199) ((T_ARRAY) reduce 199) ((T_ARRAY_CAST) reduce 199) ((T_ATTRIBUTE) reduce 199) ((T_BOOL_CAST) reduce 199) ((T_BREAK) reduce 199) ((T_CASE) reduce 199) ((T_CLASS) reduce 199) ((T_CLASS_C) reduce 199) ((T_CLONE) reduce 199) ((T_CONSTANT_ENCAPSED_STRING) reduce 199) ((T_CONTINUE) reduce 199) ((T_DEC) reduce 199) ((T_DECLARE) reduce 199) ((T_DEFAULT) reduce 199) ((T_DIR) reduce 199) ((T_DNUMBER) reduce 199) ((T_DO) reduce 199) ((T_DOUBLE_CAST) reduce 199) ((T_ECHO) reduce 199) ((T_EMPTY) reduce 199) ((T_ENUM) reduce 199) ((T_EVAL) reduce 199) ((T_EXIT) reduce 199) ((T_FILE) reduce 199) ((T_FINAL) reduce 199) ((T_FN) reduce 199) ((T_FOR) reduce 199) ((T_FOREACH) reduce 199) ((T_FUNCTION) reduce 199) ((T_FUNC_C) reduce 199) ((T_GLOBAL) reduce 199) ((T_GOTO) reduce 199) ((T_HALT_COMPILER) reduce 199) ((T_IF) reduce 199) ((T_INC) reduce 199) ((T_INCLUDE) reduce 199) ((T_INCLUDE_ONCE) reduce 199) ((T_INLINE_HTML) reduce 199) ((T_INTERFACE) reduce 199) ((T_INT_CAST) reduce 199) ((T_ISSET) reduce 199) ((T_LINE) reduce 199) ((T_LIST) reduce 199) ((T_LNUMBER) reduce 199) ((T_MATCH) reduce 199) ((T_METHOD_C) reduce 199) ((T_NAME_FULLY_QUALIFIED) reduce 199) ((T_NAME_QUALIFIED) reduce 199) ((T_NAME_RELATIVE) reduce 199) ((T_NEW) reduce 199) ((T_NS_C) reduce 199) ((T_OBJECT_CAST) reduce 199) ((T_PRINT) reduce 199) ((T_PROPERTY_C) reduce 199) ((T_READONLY) reduce 199) ((T_REQUIRE) reduce 199) ((T_REQUIRE_ONCE) reduce 199) ((T_RETURN) reduce 199) ((T_START_HEREDOC) reduce 199) ((T_STATIC) reduce 199) ((T_STRING) reduce 199) ((T_STRING_CAST) reduce 199) ((T_SWITCH) reduce 199) ((T_THROW) reduce 199) ((T_TRAIT) reduce 199) ((T_TRAIT_C) reduce 199) ((T_TRY) reduce 199) ((T_UNSET) reduce 199) ((T_UNSET_CAST) reduce 199) ((T_VARIABLE) reduce 199) ((T_WHILE) reduce 199) ((T_YIELD) reduce 199) ((T_YIELD_FROM) reduce 199) (("[") reduce 199) (("`") reduce 199) (("{") reduce 199) (("}") reduce 199) (("~") reduce 199)) 25089 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CASE) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DEFAULT) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("}") reduce 242) (("~") reduce 242)) 25092 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CASE) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DEFAULT) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("}") reduce 145) (("~") reduce 145)) 25094 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CASE) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DEFAULT) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 25096 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CASE) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DEFAULT) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 25106 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CASE) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DEFAULT) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 25107 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CASE) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DEFAULT) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("}") reduce 218) (("~") reduce 218)) 25110 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CASE) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DEFAULT) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("}") reduce 219) (("~") reduce 219)) 25111 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CASE) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DEFAULT) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 25120 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CASE) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DEFAULT) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 25121 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CASE) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DEFAULT) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("}") reduce 216) (("~") reduce 216)) 25124 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CASE) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DEFAULT) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("}") reduce 217) (("~") reduce 217)) 25131 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CASE) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DEFAULT) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("}") reduce 200) (("~") reduce 200)) 25133 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CASE) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DEFAULT) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 25139 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CASE) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DEFAULT) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 25144 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CASE) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DEFAULT) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 25145 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CASE) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DEFAULT) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("}") reduce 220) (("~") reduce 220)) 25148 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CASE) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DEFAULT) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("}") reduce 221) (("~") reduce 221)) 25150 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CASE) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DEFAULT) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 25157 ((("!") reduce 188) (("\"") reduce 188) (("$") reduce 188) (("(") reduce 188) (("+") reduce 188) (("-") reduce 188) ((";") reduce 188) (("@") reduce 188) ((T_ABSTRACT) reduce 188) ((T_ARRAY) reduce 188) ((T_ARRAY_CAST) reduce 188) ((T_ATTRIBUTE) reduce 188) ((T_BOOL_CAST) reduce 188) ((T_BREAK) reduce 188) ((T_CASE) reduce 188) ((T_CLASS) reduce 188) ((T_CLASS_C) reduce 188) ((T_CLONE) reduce 188) ((T_CONSTANT_ENCAPSED_STRING) reduce 188) ((T_CONTINUE) reduce 188) ((T_DEC) reduce 188) ((T_DECLARE) reduce 188) ((T_DEFAULT) reduce 188) ((T_DIR) reduce 188) ((T_DNUMBER) reduce 188) ((T_DO) reduce 188) ((T_DOUBLE_CAST) reduce 188) ((T_ECHO) reduce 188) ((T_EMPTY) reduce 188) ((T_ENUM) reduce 188) ((T_EVAL) reduce 188) ((T_EXIT) reduce 188) ((T_FILE) reduce 188) ((T_FINAL) reduce 188) ((T_FN) reduce 188) ((T_FOR) reduce 188) ((T_FOREACH) reduce 188) ((T_FUNCTION) reduce 188) ((T_FUNC_C) reduce 188) ((T_GLOBAL) reduce 188) ((T_GOTO) reduce 188) ((T_HALT_COMPILER) reduce 188) ((T_IF) reduce 188) ((T_INC) reduce 188) ((T_INCLUDE) reduce 188) ((T_INCLUDE_ONCE) reduce 188) ((T_INLINE_HTML) reduce 188) ((T_INTERFACE) reduce 188) ((T_INT_CAST) reduce 188) ((T_ISSET) reduce 188) ((T_LINE) reduce 188) ((T_LIST) reduce 188) ((T_LNUMBER) reduce 188) ((T_MATCH) reduce 188) ((T_METHOD_C) reduce 188) ((T_NAME_FULLY_QUALIFIED) reduce 188) ((T_NAME_QUALIFIED) reduce 188) ((T_NAME_RELATIVE) reduce 188) ((T_NEW) reduce 188) ((T_NS_C) reduce 188) ((T_OBJECT_CAST) reduce 188) ((T_PRINT) reduce 188) ((T_PROPERTY_C) reduce 188) ((T_READONLY) reduce 188) ((T_REQUIRE) reduce 188) ((T_REQUIRE_ONCE) reduce 188) ((T_RETURN) reduce 188) ((T_START_HEREDOC) reduce 188) ((T_STATIC) reduce 188) ((T_STRING) reduce 188) ((T_STRING_CAST) reduce 188) ((T_SWITCH) reduce 188) ((T_THROW) reduce 188) ((T_TRAIT) reduce 188) ((T_TRAIT_C) reduce 188) ((T_TRY) reduce 188) ((T_UNSET) reduce 188) ((T_UNSET_CAST) reduce 188) ((T_VARIABLE) reduce 188) ((T_WHILE) reduce 188) ((T_YIELD) reduce 188) ((T_YIELD_FROM) reduce 188) (("[") reduce 188) (("`") reduce 188) (("{") reduce 188) (("}") reduce 188) (("~") reduce 188)) 25159 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CASE) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DEFAULT) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 25162 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 227) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 227) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("}") reduce 227) (("~") shift)) 25164 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDWHILE) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 25168 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDWHILE) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 25171 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDWHILE) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 25172 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDWHILE) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 25174 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDWHILE) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 25176 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDWHILE) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 25178 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDWHILE) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 25180 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDWHILE) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 25190 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDWHILE) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 25191 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDWHILE) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 25194 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDWHILE) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 25195 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDWHILE) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 25204 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDWHILE) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 25205 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDWHILE) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 25208 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDWHILE) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 25210 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDWHILE) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 25216 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDWHILE) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 25221 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDWHILE) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 25222 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDWHILE) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 25225 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDWHILE) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 25227 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDWHILE) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 25229 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDWHILE) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 25230 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDWHILE) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 25241 ((("!") reduce 477) (("\"") reduce 477) (("$") reduce 477) (("(") reduce 477) (("+") reduce 477) (("-") reduce 477) ((";") reduce 477) (("@") reduce 477) ((T_ABSTRACT) reduce 477) ((T_ARRAY) reduce 477) ((T_ARRAY_CAST) reduce 477) ((T_ATTRIBUTE) reduce 477) ((T_BOOL_CAST) reduce 477) ((T_BREAK) reduce 477) ((T_CLASS) reduce 477) ((T_CLASS_C) reduce 477) ((T_CLONE) reduce 477) ((T_CONSTANT_ENCAPSED_STRING) reduce 477) ((T_CONTINUE) reduce 477) ((T_DEC) reduce 477) ((T_DECLARE) reduce 477) ((T_DIR) reduce 477) ((T_DNUMBER) reduce 477) ((T_DO) reduce 477) ((T_DOUBLE_CAST) reduce 477) ((T_ECHO) reduce 477) ((T_EMPTY) reduce 477) ((T_ENDWHILE) reduce 477) ((T_ENUM) reduce 477) ((T_EVAL) reduce 477) ((T_EXIT) reduce 477) ((T_FILE) reduce 477) ((T_FINAL) reduce 477) ((T_FN) reduce 477) ((T_FOR) reduce 477) ((T_FOREACH) reduce 477) ((T_FUNCTION) reduce 477) ((T_FUNC_C) reduce 477) ((T_GLOBAL) reduce 477) ((T_GOTO) reduce 477) ((T_HALT_COMPILER) reduce 477) ((T_IF) reduce 477) ((T_INC) reduce 477) ((T_INCLUDE) reduce 477) ((T_INCLUDE_ONCE) reduce 477) ((T_INLINE_HTML) reduce 477) ((T_INTERFACE) reduce 477) ((T_INT_CAST) reduce 477) ((T_ISSET) reduce 477) ((T_LINE) reduce 477) ((T_LIST) reduce 477) ((T_LNUMBER) reduce 477) ((T_MATCH) reduce 477) ((T_METHOD_C) reduce 477) ((T_NAME_FULLY_QUALIFIED) reduce 477) ((T_NAME_QUALIFIED) reduce 477) ((T_NAME_RELATIVE) reduce 477) ((T_NEW) reduce 477) ((T_NS_C) reduce 477) ((T_OBJECT_CAST) reduce 477) ((T_PRINT) reduce 477) ((T_PROPERTY_C) reduce 477) ((T_READONLY) reduce 477) ((T_REQUIRE) reduce 477) ((T_REQUIRE_ONCE) reduce 477) ((T_RETURN) reduce 477) ((T_START_HEREDOC) reduce 477) ((T_STATIC) reduce 477) ((T_STRING) reduce 477) ((T_STRING_CAST) reduce 477) ((T_SWITCH) reduce 477) ((T_THROW) reduce 477) ((T_TRAIT) reduce 477) ((T_TRAIT_C) reduce 477) ((T_TRY) reduce 477) ((T_UNSET) reduce 477) ((T_UNSET_CAST) reduce 477) ((T_VARIABLE) reduce 477) ((T_WHILE) reduce 477) ((T_YIELD) reduce 477) ((T_YIELD_FROM) reduce 477) (("[") reduce 477) (("`") reduce 477) (("{") reduce 477) (("~") reduce 477)) 25242 ((("!") reduce 182) (("\"") reduce 182) (("$") reduce 182) (("(") reduce 182) (("+") reduce 182) (("-") reduce 182) ((";") reduce 182) (("@") reduce 182) ((T_ABSTRACT) reduce 182) ((T_ARRAY) reduce 182) ((T_ARRAY_CAST) reduce 182) ((T_ATTRIBUTE) reduce 182) ((T_BOOL_CAST) reduce 182) ((T_BREAK) reduce 182) ((T_CLASS) reduce 182) ((T_CLASS_C) reduce 182) ((T_CLONE) reduce 182) ((T_CONSTANT_ENCAPSED_STRING) reduce 182) ((T_CONTINUE) reduce 182) ((T_DEC) reduce 182) ((T_DECLARE) reduce 182) ((T_DIR) reduce 182) ((T_DNUMBER) reduce 182) ((T_DO) reduce 182) ((T_DOUBLE_CAST) reduce 182) ((T_ECHO) reduce 182) ((T_EMPTY) reduce 182) ((T_ENDWHILE) reduce 182) ((T_ENUM) reduce 182) ((T_EVAL) reduce 182) ((T_EXIT) reduce 182) ((T_FILE) reduce 182) ((T_FINAL) reduce 182) ((T_FN) reduce 182) ((T_FOR) reduce 182) ((T_FOREACH) reduce 182) ((T_FUNCTION) reduce 182) ((T_FUNC_C) reduce 182) ((T_GLOBAL) reduce 182) ((T_GOTO) reduce 182) ((T_HALT_COMPILER) reduce 182) ((T_IF) reduce 182) ((T_INC) reduce 182) ((T_INCLUDE) reduce 182) ((T_INCLUDE_ONCE) reduce 182) ((T_INLINE_HTML) reduce 182) ((T_INTERFACE) reduce 182) ((T_INT_CAST) reduce 182) ((T_ISSET) reduce 182) ((T_LINE) reduce 182) ((T_LIST) reduce 182) ((T_LNUMBER) reduce 182) ((T_MATCH) reduce 182) ((T_METHOD_C) reduce 182) ((T_NAME_FULLY_QUALIFIED) reduce 182) ((T_NAME_QUALIFIED) reduce 182) ((T_NAME_RELATIVE) reduce 182) ((T_NEW) reduce 182) ((T_NS_C) reduce 182) ((T_OBJECT_CAST) reduce 182) ((T_PRINT) reduce 182) ((T_PROPERTY_C) reduce 182) ((T_READONLY) reduce 182) ((T_REQUIRE) reduce 182) ((T_REQUIRE_ONCE) reduce 182) ((T_RETURN) reduce 182) ((T_START_HEREDOC) reduce 182) ((T_STATIC) reduce 182) ((T_STRING) reduce 182) ((T_STRING_CAST) reduce 182) ((T_SWITCH) reduce 182) ((T_THROW) reduce 182) ((T_TRAIT) reduce 182) ((T_TRAIT_C) reduce 182) ((T_TRY) reduce 182) ((T_UNSET) reduce 182) ((T_UNSET_CAST) reduce 182) ((T_VARIABLE) reduce 182) ((T_WHILE) reduce 182) ((T_YIELD) reduce 182) ((T_YIELD_FROM) reduce 182) (("[") reduce 182) (("`") reduce 182) (("{") reduce 182) (("~") reduce 182)) 25243 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDWHILE) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 25251 ((("!") reduce 187) (("\"") reduce 187) (("$") reduce 187) (("(") reduce 187) (("+") reduce 187) (("-") reduce 187) ((";") reduce 187) (("@") reduce 187) ((T_ABSTRACT) reduce 187) ((T_ARRAY) reduce 187) ((T_ARRAY_CAST) reduce 187) ((T_ATTRIBUTE) reduce 187) ((T_BOOL_CAST) reduce 187) ((T_BREAK) reduce 187) ((T_CLASS) reduce 187) ((T_CLASS_C) reduce 187) ((T_CLONE) reduce 187) ((T_CONSTANT_ENCAPSED_STRING) reduce 187) ((T_CONTINUE) reduce 187) ((T_DEC) reduce 187) ((T_DECLARE) reduce 187) ((T_DIR) reduce 187) ((T_DNUMBER) reduce 187) ((T_DO) reduce 187) ((T_DOUBLE_CAST) reduce 187) ((T_ECHO) reduce 187) ((T_EMPTY) reduce 187) ((T_ENDWHILE) reduce 187) ((T_ENUM) reduce 187) ((T_EVAL) reduce 187) ((T_EXIT) reduce 187) ((T_FILE) reduce 187) ((T_FINAL) reduce 187) ((T_FN) reduce 187) ((T_FOR) reduce 187) ((T_FOREACH) reduce 187) ((T_FUNCTION) reduce 187) ((T_FUNC_C) reduce 187) ((T_GLOBAL) reduce 187) ((T_GOTO) reduce 187) ((T_HALT_COMPILER) reduce 187) ((T_IF) reduce 187) ((T_INC) reduce 187) ((T_INCLUDE) reduce 187) ((T_INCLUDE_ONCE) reduce 187) ((T_INLINE_HTML) reduce 187) ((T_INTERFACE) reduce 187) ((T_INT_CAST) reduce 187) ((T_ISSET) reduce 187) ((T_LINE) reduce 187) ((T_LIST) reduce 187) ((T_LNUMBER) reduce 187) ((T_MATCH) reduce 187) ((T_METHOD_C) reduce 187) ((T_NAME_FULLY_QUALIFIED) reduce 187) ((T_NAME_QUALIFIED) reduce 187) ((T_NAME_RELATIVE) reduce 187) ((T_NEW) reduce 187) ((T_NS_C) reduce 187) ((T_OBJECT_CAST) reduce 187) ((T_PRINT) reduce 187) ((T_PROPERTY_C) reduce 187) ((T_READONLY) reduce 187) ((T_REQUIRE) reduce 187) ((T_REQUIRE_ONCE) reduce 187) ((T_RETURN) reduce 187) ((T_START_HEREDOC) reduce 187) ((T_STATIC) reduce 187) ((T_STRING) reduce 187) ((T_STRING_CAST) reduce 187) ((T_SWITCH) reduce 187) ((T_THROW) reduce 187) ((T_TRAIT) reduce 187) ((T_TRAIT_C) reduce 187) ((T_TRY) reduce 187) ((T_UNSET) reduce 187) ((T_UNSET_CAST) reduce 187) ((T_VARIABLE) reduce 187) ((T_WHILE) reduce 187) ((T_YIELD) reduce 187) ((T_YIELD_FROM) reduce 187) (("[") reduce 187) (("`") reduce 187) (("{") reduce 187) (("~") reduce 187)) 25252 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDWHILE) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 25255 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_EMPTY) reduce 248) ((T_ENDWHILE) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("~") reduce 248)) 25259 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_EMPTY) reduce 249) ((T_ENDWHILE) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("~") reduce 249)) 25264 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDWHILE) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 25265 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDWHILE) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 25268 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDWHILE) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 25273 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDWHILE) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 25276 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDWHILE) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 25277 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDWHILE) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 25280 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDWHILE) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 25283 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_EMPTY) reduce 176) ((T_ENDWHILE) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("~") reduce 176)) 25290 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDWHILE) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 25295 ((("!") reduce 198) (("\"") reduce 198) (("$") reduce 198) (("(") reduce 198) (("+") reduce 198) (("-") reduce 198) ((";") reduce 198) (("@") reduce 198) ((T_ABSTRACT) reduce 198) ((T_ARRAY) reduce 198) ((T_ARRAY_CAST) reduce 198) ((T_ATTRIBUTE) reduce 198) ((T_BOOL_CAST) reduce 198) ((T_BREAK) reduce 198) ((T_CLASS) reduce 198) ((T_CLASS_C) reduce 198) ((T_CLONE) reduce 198) ((T_CONSTANT_ENCAPSED_STRING) reduce 198) ((T_CONTINUE) reduce 198) ((T_DEC) reduce 198) ((T_DECLARE) reduce 198) ((T_DIR) reduce 198) ((T_DNUMBER) reduce 198) ((T_DO) reduce 198) ((T_DOUBLE_CAST) reduce 198) ((T_ECHO) reduce 198) ((T_EMPTY) reduce 198) ((T_ENDWHILE) reduce 198) ((T_ENUM) reduce 198) ((T_EVAL) reduce 198) ((T_EXIT) reduce 198) ((T_FILE) reduce 198) ((T_FINAL) reduce 198) ((T_FN) reduce 198) ((T_FOR) reduce 198) ((T_FOREACH) reduce 198) ((T_FUNCTION) reduce 198) ((T_FUNC_C) reduce 198) ((T_GLOBAL) reduce 198) ((T_GOTO) reduce 198) ((T_HALT_COMPILER) reduce 198) ((T_IF) reduce 198) ((T_INC) reduce 198) ((T_INCLUDE) reduce 198) ((T_INCLUDE_ONCE) reduce 198) ((T_INLINE_HTML) reduce 198) ((T_INTERFACE) reduce 198) ((T_INT_CAST) reduce 198) ((T_ISSET) reduce 198) ((T_LINE) reduce 198) ((T_LIST) reduce 198) ((T_LNUMBER) reduce 198) ((T_MATCH) reduce 198) ((T_METHOD_C) reduce 198) ((T_NAME_FULLY_QUALIFIED) reduce 198) ((T_NAME_QUALIFIED) reduce 198) ((T_NAME_RELATIVE) reduce 198) ((T_NEW) reduce 198) ((T_NS_C) reduce 198) ((T_OBJECT_CAST) reduce 198) ((T_PRINT) reduce 198) ((T_PROPERTY_C) reduce 198) ((T_READONLY) reduce 198) ((T_REQUIRE) reduce 198) ((T_REQUIRE_ONCE) reduce 198) ((T_RETURN) reduce 198) ((T_START_HEREDOC) reduce 198) ((T_STATIC) reduce 198) ((T_STRING) reduce 198) ((T_STRING_CAST) reduce 198) ((T_SWITCH) reduce 198) ((T_THROW) reduce 198) ((T_TRAIT) reduce 198) ((T_TRAIT_C) reduce 198) ((T_TRY) reduce 198) ((T_UNSET) reduce 198) ((T_UNSET_CAST) reduce 198) ((T_VARIABLE) reduce 198) ((T_WHILE) reduce 198) ((T_YIELD) reduce 198) ((T_YIELD_FROM) reduce 198) (("[") reduce 198) (("`") reduce 198) (("{") reduce 198) (("~") reduce 198)) 25300 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDWHILE) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 25304 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDWHILE) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 25306 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDWHILE) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 25310 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDWHILE) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 25313 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDWHILE) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 25314 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDWHILE) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 25316 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDWHILE) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 25318 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDWHILE) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 25324 ((("!") reduce 199) (("\"") reduce 199) (("$") reduce 199) (("(") reduce 199) (("+") reduce 199) (("-") reduce 199) ((";") reduce 199) (("@") reduce 199) ((T_ABSTRACT) reduce 199) ((T_ARRAY) reduce 199) ((T_ARRAY_CAST) reduce 199) ((T_ATTRIBUTE) reduce 199) ((T_BOOL_CAST) reduce 199) ((T_BREAK) reduce 199) ((T_CLASS) reduce 199) ((T_CLASS_C) reduce 199) ((T_CLONE) reduce 199) ((T_CONSTANT_ENCAPSED_STRING) reduce 199) ((T_CONTINUE) reduce 199) ((T_DEC) reduce 199) ((T_DECLARE) reduce 199) ((T_DIR) reduce 199) ((T_DNUMBER) reduce 199) ((T_DO) reduce 199) ((T_DOUBLE_CAST) reduce 199) ((T_ECHO) reduce 199) ((T_EMPTY) reduce 199) ((T_ENDWHILE) reduce 199) ((T_ENUM) reduce 199) ((T_EVAL) reduce 199) ((T_EXIT) reduce 199) ((T_FILE) reduce 199) ((T_FINAL) reduce 199) ((T_FN) reduce 199) ((T_FOR) reduce 199) ((T_FOREACH) reduce 199) ((T_FUNCTION) reduce 199) ((T_FUNC_C) reduce 199) ((T_GLOBAL) reduce 199) ((T_GOTO) reduce 199) ((T_HALT_COMPILER) reduce 199) ((T_IF) reduce 199) ((T_INC) reduce 199) ((T_INCLUDE) reduce 199) ((T_INCLUDE_ONCE) reduce 199) ((T_INLINE_HTML) reduce 199) ((T_INTERFACE) reduce 199) ((T_INT_CAST) reduce 199) ((T_ISSET) reduce 199) ((T_LINE) reduce 199) ((T_LIST) reduce 199) ((T_LNUMBER) reduce 199) ((T_MATCH) reduce 199) ((T_METHOD_C) reduce 199) ((T_NAME_FULLY_QUALIFIED) reduce 199) ((T_NAME_QUALIFIED) reduce 199) ((T_NAME_RELATIVE) reduce 199) ((T_NEW) reduce 199) ((T_NS_C) reduce 199) ((T_OBJECT_CAST) reduce 199) ((T_PRINT) reduce 199) ((T_PROPERTY_C) reduce 199) ((T_READONLY) reduce 199) ((T_REQUIRE) reduce 199) ((T_REQUIRE_ONCE) reduce 199) ((T_RETURN) reduce 199) ((T_START_HEREDOC) reduce 199) ((T_STATIC) reduce 199) ((T_STRING) reduce 199) ((T_STRING_CAST) reduce 199) ((T_SWITCH) reduce 199) ((T_THROW) reduce 199) ((T_TRAIT) reduce 199) ((T_TRAIT_C) reduce 199) ((T_TRY) reduce 199) ((T_UNSET) reduce 199) ((T_UNSET_CAST) reduce 199) ((T_VARIABLE) reduce 199) ((T_WHILE) reduce 199) ((T_YIELD) reduce 199) ((T_YIELD_FROM) reduce 199) (("[") reduce 199) (("`") reduce 199) (("{") reduce 199) (("~") reduce 199)) 25328 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDWHILE) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 25331 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDWHILE) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 25333 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDWHILE) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 25335 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDWHILE) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 25345 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDWHILE) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 25346 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDWHILE) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 25349 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDWHILE) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 25350 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDWHILE) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 25359 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDWHILE) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 25360 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDWHILE) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 25363 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDWHILE) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 25370 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENDWHILE) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 25371 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDIF) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 25373 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDWHILE) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 25379 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDWHILE) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 25384 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDWHILE) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 25385 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDWHILE) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 25388 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDWHILE) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 25390 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDWHILE) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 25397 ((("!") reduce 188) (("\"") reduce 188) (("$") reduce 188) (("(") reduce 188) (("+") reduce 188) (("-") reduce 188) ((";") reduce 188) (("@") reduce 188) ((T_ABSTRACT) reduce 188) ((T_ARRAY) reduce 188) ((T_ARRAY_CAST) reduce 188) ((T_ATTRIBUTE) reduce 188) ((T_BOOL_CAST) reduce 188) ((T_BREAK) reduce 188) ((T_CLASS) reduce 188) ((T_CLASS_C) reduce 188) ((T_CLONE) reduce 188) ((T_CONSTANT_ENCAPSED_STRING) reduce 188) ((T_CONTINUE) reduce 188) ((T_DEC) reduce 188) ((T_DECLARE) reduce 188) ((T_DIR) reduce 188) ((T_DNUMBER) reduce 188) ((T_DO) reduce 188) ((T_DOUBLE_CAST) reduce 188) ((T_ECHO) reduce 188) ((T_EMPTY) reduce 188) ((T_ENDWHILE) reduce 188) ((T_ENUM) reduce 188) ((T_EVAL) reduce 188) ((T_EXIT) reduce 188) ((T_FILE) reduce 188) ((T_FINAL) reduce 188) ((T_FN) reduce 188) ((T_FOR) reduce 188) ((T_FOREACH) reduce 188) ((T_FUNCTION) reduce 188) ((T_FUNC_C) reduce 188) ((T_GLOBAL) reduce 188) ((T_GOTO) reduce 188) ((T_HALT_COMPILER) reduce 188) ((T_IF) reduce 188) ((T_INC) reduce 188) ((T_INCLUDE) reduce 188) ((T_INCLUDE_ONCE) reduce 188) ((T_INLINE_HTML) reduce 188) ((T_INTERFACE) reduce 188) ((T_INT_CAST) reduce 188) ((T_ISSET) reduce 188) ((T_LINE) reduce 188) ((T_LIST) reduce 188) ((T_LNUMBER) reduce 188) ((T_MATCH) reduce 188) ((T_METHOD_C) reduce 188) ((T_NAME_FULLY_QUALIFIED) reduce 188) ((T_NAME_QUALIFIED) reduce 188) ((T_NAME_RELATIVE) reduce 188) ((T_NEW) reduce 188) ((T_NS_C) reduce 188) ((T_OBJECT_CAST) reduce 188) ((T_PRINT) reduce 188) ((T_PROPERTY_C) reduce 188) ((T_READONLY) reduce 188) ((T_REQUIRE) reduce 188) ((T_REQUIRE_ONCE) reduce 188) ((T_RETURN) reduce 188) ((T_START_HEREDOC) reduce 188) ((T_STATIC) reduce 188) ((T_STRING) reduce 188) ((T_STRING_CAST) reduce 188) ((T_SWITCH) reduce 188) ((T_THROW) reduce 188) ((T_TRAIT) reduce 188) ((T_TRAIT_C) reduce 188) ((T_TRY) reduce 188) ((T_UNSET) reduce 188) ((T_UNSET_CAST) reduce 188) ((T_VARIABLE) reduce 188) ((T_WHILE) reduce 188) ((T_YIELD) reduce 188) ((T_YIELD_FROM) reduce 188) (("[") reduce 188) (("`") reduce 188) (("{") reduce 188) (("~") reduce 188)) 25399 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDWHILE) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 25404 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDIF) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 25407 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDIF) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 25408 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDIF) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 25411 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDIF) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 25414 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_EMPTY) reduce 176) ((T_ENDIF) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("~") reduce 176)) 25421 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDIF) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 25426 ((("!") reduce 198) (("\"") reduce 198) (("$") reduce 198) (("(") reduce 198) (("+") reduce 198) (("-") reduce 198) ((";") reduce 198) (("@") reduce 198) ((T_ABSTRACT) reduce 198) ((T_ARRAY) reduce 198) ((T_ARRAY_CAST) reduce 198) ((T_ATTRIBUTE) reduce 198) ((T_BOOL_CAST) reduce 198) ((T_BREAK) reduce 198) ((T_CLASS) reduce 198) ((T_CLASS_C) reduce 198) ((T_CLONE) reduce 198) ((T_CONSTANT_ENCAPSED_STRING) reduce 198) ((T_CONTINUE) reduce 198) ((T_DEC) reduce 198) ((T_DECLARE) reduce 198) ((T_DIR) reduce 198) ((T_DNUMBER) reduce 198) ((T_DO) reduce 198) ((T_DOUBLE_CAST) reduce 198) ((T_ECHO) reduce 198) ((T_EMPTY) reduce 198) ((T_ENDIF) reduce 198) ((T_ENUM) reduce 198) ((T_EVAL) reduce 198) ((T_EXIT) reduce 198) ((T_FILE) reduce 198) ((T_FINAL) reduce 198) ((T_FN) reduce 198) ((T_FOR) reduce 198) ((T_FOREACH) reduce 198) ((T_FUNCTION) reduce 198) ((T_FUNC_C) reduce 198) ((T_GLOBAL) reduce 198) ((T_GOTO) reduce 198) ((T_HALT_COMPILER) reduce 198) ((T_IF) reduce 198) ((T_INC) reduce 198) ((T_INCLUDE) reduce 198) ((T_INCLUDE_ONCE) reduce 198) ((T_INLINE_HTML) reduce 198) ((T_INTERFACE) reduce 198) ((T_INT_CAST) reduce 198) ((T_ISSET) reduce 198) ((T_LINE) reduce 198) ((T_LIST) reduce 198) ((T_LNUMBER) reduce 198) ((T_MATCH) reduce 198) ((T_METHOD_C) reduce 198) ((T_NAME_FULLY_QUALIFIED) reduce 198) ((T_NAME_QUALIFIED) reduce 198) ((T_NAME_RELATIVE) reduce 198) ((T_NEW) reduce 198) ((T_NS_C) reduce 198) ((T_OBJECT_CAST) reduce 198) ((T_PRINT) reduce 198) ((T_PROPERTY_C) reduce 198) ((T_READONLY) reduce 198) ((T_REQUIRE) reduce 198) ((T_REQUIRE_ONCE) reduce 198) ((T_RETURN) reduce 198) ((T_START_HEREDOC) reduce 198) ((T_STATIC) reduce 198) ((T_STRING) reduce 198) ((T_STRING_CAST) reduce 198) ((T_SWITCH) reduce 198) ((T_THROW) reduce 198) ((T_TRAIT) reduce 198) ((T_TRAIT_C) reduce 198) ((T_TRY) reduce 198) ((T_UNSET) reduce 198) ((T_UNSET_CAST) reduce 198) ((T_VARIABLE) reduce 198) ((T_WHILE) reduce 198) ((T_YIELD) reduce 198) ((T_YIELD_FROM) reduce 198) (("[") reduce 198) (("`") reduce 198) (("{") reduce 198) (("~") reduce 198)) 25431 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDIF) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 25435 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDIF) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 25437 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDIF) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 25441 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDIF) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 25444 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDIF) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 25445 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDIF) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 25447 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDIF) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 25449 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDIF) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 25455 ((("!") reduce 199) (("\"") reduce 199) (("$") reduce 199) (("(") reduce 199) (("+") reduce 199) (("-") reduce 199) ((";") reduce 199) (("@") reduce 199) ((T_ABSTRACT) reduce 199) ((T_ARRAY) reduce 199) ((T_ARRAY_CAST) reduce 199) ((T_ATTRIBUTE) reduce 199) ((T_BOOL_CAST) reduce 199) ((T_BREAK) reduce 199) ((T_CLASS) reduce 199) ((T_CLASS_C) reduce 199) ((T_CLONE) reduce 199) ((T_CONSTANT_ENCAPSED_STRING) reduce 199) ((T_CONTINUE) reduce 199) ((T_DEC) reduce 199) ((T_DECLARE) reduce 199) ((T_DIR) reduce 199) ((T_DNUMBER) reduce 199) ((T_DO) reduce 199) ((T_DOUBLE_CAST) reduce 199) ((T_ECHO) reduce 199) ((T_EMPTY) reduce 199) ((T_ENDIF) reduce 199) ((T_ENUM) reduce 199) ((T_EVAL) reduce 199) ((T_EXIT) reduce 199) ((T_FILE) reduce 199) ((T_FINAL) reduce 199) ((T_FN) reduce 199) ((T_FOR) reduce 199) ((T_FOREACH) reduce 199) ((T_FUNCTION) reduce 199) ((T_FUNC_C) reduce 199) ((T_GLOBAL) reduce 199) ((T_GOTO) reduce 199) ((T_HALT_COMPILER) reduce 199) ((T_IF) reduce 199) ((T_INC) reduce 199) ((T_INCLUDE) reduce 199) ((T_INCLUDE_ONCE) reduce 199) ((T_INLINE_HTML) reduce 199) ((T_INTERFACE) reduce 199) ((T_INT_CAST) reduce 199) ((T_ISSET) reduce 199) ((T_LINE) reduce 199) ((T_LIST) reduce 199) ((T_LNUMBER) reduce 199) ((T_MATCH) reduce 199) ((T_METHOD_C) reduce 199) ((T_NAME_FULLY_QUALIFIED) reduce 199) ((T_NAME_QUALIFIED) reduce 199) ((T_NAME_RELATIVE) reduce 199) ((T_NEW) reduce 199) ((T_NS_C) reduce 199) ((T_OBJECT_CAST) reduce 199) ((T_PRINT) reduce 199) ((T_PROPERTY_C) reduce 199) ((T_READONLY) reduce 199) ((T_REQUIRE) reduce 199) ((T_REQUIRE_ONCE) reduce 199) ((T_RETURN) reduce 199) ((T_START_HEREDOC) reduce 199) ((T_STATIC) reduce 199) ((T_STRING) reduce 199) ((T_STRING_CAST) reduce 199) ((T_SWITCH) reduce 199) ((T_THROW) reduce 199) ((T_TRAIT) reduce 199) ((T_TRAIT_C) reduce 199) ((T_TRY) reduce 199) ((T_UNSET) reduce 199) ((T_UNSET_CAST) reduce 199) ((T_VARIABLE) reduce 199) ((T_WHILE) reduce 199) ((T_YIELD) reduce 199) ((T_YIELD_FROM) reduce 199) (("[") reduce 199) (("`") reduce 199) (("{") reduce 199) (("~") reduce 199)) 25458 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDIF) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 25460 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDIF) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 25462 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDIF) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 25472 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDIF) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 25473 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDIF) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 25476 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDIF) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 25477 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDIF) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 25486 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDIF) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 25487 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDIF) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 25490 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDIF) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 25497 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENDIF) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 25498 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_ELSE) reduce 249) ((T_ELSEIF) reduce 249) ((T_EMPTY) reduce 249) ((T_ENDIF) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("~") reduce 249)) 25500 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDIF) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 25506 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDIF) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 25511 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDIF) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 25512 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDIF) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 25515 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDIF) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 25517 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDIF) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 25524 ((("!") reduce 188) (("\"") reduce 188) (("$") reduce 188) (("(") reduce 188) (("+") reduce 188) (("-") reduce 188) ((";") reduce 188) (("@") reduce 188) ((T_ABSTRACT) reduce 188) ((T_ARRAY) reduce 188) ((T_ARRAY_CAST) reduce 188) ((T_ATTRIBUTE) reduce 188) ((T_BOOL_CAST) reduce 188) ((T_BREAK) reduce 188) ((T_CLASS) reduce 188) ((T_CLASS_C) reduce 188) ((T_CLONE) reduce 188) ((T_CONSTANT_ENCAPSED_STRING) reduce 188) ((T_CONTINUE) reduce 188) ((T_DEC) reduce 188) ((T_DECLARE) reduce 188) ((T_DIR) reduce 188) ((T_DNUMBER) reduce 188) ((T_DO) reduce 188) ((T_DOUBLE_CAST) reduce 188) ((T_ECHO) reduce 188) ((T_EMPTY) reduce 188) ((T_ENDIF) reduce 188) ((T_ENUM) reduce 188) ((T_EVAL) reduce 188) ((T_EXIT) reduce 188) ((T_FILE) reduce 188) ((T_FINAL) reduce 188) ((T_FN) reduce 188) ((T_FOR) reduce 188) ((T_FOREACH) reduce 188) ((T_FUNCTION) reduce 188) ((T_FUNC_C) reduce 188) ((T_GLOBAL) reduce 188) ((T_GOTO) reduce 188) ((T_HALT_COMPILER) reduce 188) ((T_IF) reduce 188) ((T_INC) reduce 188) ((T_INCLUDE) reduce 188) ((T_INCLUDE_ONCE) reduce 188) ((T_INLINE_HTML) reduce 188) ((T_INTERFACE) reduce 188) ((T_INT_CAST) reduce 188) ((T_ISSET) reduce 188) ((T_LINE) reduce 188) ((T_LIST) reduce 188) ((T_LNUMBER) reduce 188) ((T_MATCH) reduce 188) ((T_METHOD_C) reduce 188) ((T_NAME_FULLY_QUALIFIED) reduce 188) ((T_NAME_QUALIFIED) reduce 188) ((T_NAME_RELATIVE) reduce 188) ((T_NEW) reduce 188) ((T_NS_C) reduce 188) ((T_OBJECT_CAST) reduce 188) ((T_PRINT) reduce 188) ((T_PROPERTY_C) reduce 188) ((T_READONLY) reduce 188) ((T_REQUIRE) reduce 188) ((T_REQUIRE_ONCE) reduce 188) ((T_RETURN) reduce 188) ((T_START_HEREDOC) reduce 188) ((T_STATIC) reduce 188) ((T_STRING) reduce 188) ((T_STRING_CAST) reduce 188) ((T_SWITCH) reduce 188) ((T_THROW) reduce 188) ((T_TRAIT) reduce 188) ((T_TRAIT_C) reduce 188) ((T_TRY) reduce 188) ((T_UNSET) reduce 188) ((T_UNSET_CAST) reduce 188) ((T_VARIABLE) reduce 188) ((T_WHILE) reduce 188) ((T_YIELD) reduce 188) ((T_YIELD_FROM) reduce 188) (("[") reduce 188) (("`") reduce 188) (("{") reduce 188) (("~") reduce 188)) 25526 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDIF) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 25531 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDIF) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 25532 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDIF) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 25535 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDIF) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 25540 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDIF) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 25543 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDIF) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 25544 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDIF) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 25547 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDIF) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 25550 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_ELSE) reduce 176) ((T_ELSEIF) reduce 176) ((T_EMPTY) reduce 176) ((T_ENDIF) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("~") reduce 176)) 25557 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDIF) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 25562 ((("!") reduce 198) (("\"") reduce 198) (("$") reduce 198) (("(") reduce 198) (("+") reduce 198) (("-") reduce 198) ((";") reduce 198) (("@") reduce 198) ((T_ABSTRACT) reduce 198) ((T_ARRAY) reduce 198) ((T_ARRAY_CAST) reduce 198) ((T_ATTRIBUTE) reduce 198) ((T_BOOL_CAST) reduce 198) ((T_BREAK) reduce 198) ((T_CLASS) reduce 198) ((T_CLASS_C) reduce 198) ((T_CLONE) reduce 198) ((T_CONSTANT_ENCAPSED_STRING) reduce 198) ((T_CONTINUE) reduce 198) ((T_DEC) reduce 198) ((T_DECLARE) reduce 198) ((T_DIR) reduce 198) ((T_DNUMBER) reduce 198) ((T_DO) reduce 198) ((T_DOUBLE_CAST) reduce 198) ((T_ECHO) reduce 198) ((T_ELSE) reduce 198) ((T_ELSEIF) reduce 198) ((T_EMPTY) reduce 198) ((T_ENDIF) reduce 198) ((T_ENUM) reduce 198) ((T_EVAL) reduce 198) ((T_EXIT) reduce 198) ((T_FILE) reduce 198) ((T_FINAL) reduce 198) ((T_FN) reduce 198) ((T_FOR) reduce 198) ((T_FOREACH) reduce 198) ((T_FUNCTION) reduce 198) ((T_FUNC_C) reduce 198) ((T_GLOBAL) reduce 198) ((T_GOTO) reduce 198) ((T_HALT_COMPILER) reduce 198) ((T_IF) reduce 198) ((T_INC) reduce 198) ((T_INCLUDE) reduce 198) ((T_INCLUDE_ONCE) reduce 198) ((T_INLINE_HTML) reduce 198) ((T_INTERFACE) reduce 198) ((T_INT_CAST) reduce 198) ((T_ISSET) reduce 198) ((T_LINE) reduce 198) ((T_LIST) reduce 198) ((T_LNUMBER) reduce 198) ((T_MATCH) reduce 198) ((T_METHOD_C) reduce 198) ((T_NAME_FULLY_QUALIFIED) reduce 198) ((T_NAME_QUALIFIED) reduce 198) ((T_NAME_RELATIVE) reduce 198) ((T_NEW) reduce 198) ((T_NS_C) reduce 198) ((T_OBJECT_CAST) reduce 198) ((T_PRINT) reduce 198) ((T_PROPERTY_C) reduce 198) ((T_READONLY) reduce 198) ((T_REQUIRE) reduce 198) ((T_REQUIRE_ONCE) reduce 198) ((T_RETURN) reduce 198) ((T_START_HEREDOC) reduce 198) ((T_STATIC) reduce 198) ((T_STRING) reduce 198) ((T_STRING_CAST) reduce 198) ((T_SWITCH) reduce 198) ((T_THROW) reduce 198) ((T_TRAIT) reduce 198) ((T_TRAIT_C) reduce 198) ((T_TRY) reduce 198) ((T_UNSET) reduce 198) ((T_UNSET_CAST) reduce 198) ((T_VARIABLE) reduce 198) ((T_WHILE) reduce 198) ((T_YIELD) reduce 198) ((T_YIELD_FROM) reduce 198) (("[") reduce 198) (("`") reduce 198) (("{") reduce 198) (("~") reduce 198)) 25567 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDIF) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 25571 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDIF) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 25573 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDIF) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 25577 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDIF) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 25580 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDIF) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 25581 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDIF) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 25583 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDIF) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 25585 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDIF) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 25591 ((("!") reduce 199) (("\"") reduce 199) (("$") reduce 199) (("(") reduce 199) (("+") reduce 199) (("-") reduce 199) ((";") reduce 199) (("@") reduce 199) ((T_ABSTRACT) reduce 199) ((T_ARRAY) reduce 199) ((T_ARRAY_CAST) reduce 199) ((T_ATTRIBUTE) reduce 199) ((T_BOOL_CAST) reduce 199) ((T_BREAK) reduce 199) ((T_CLASS) reduce 199) ((T_CLASS_C) reduce 199) ((T_CLONE) reduce 199) ((T_CONSTANT_ENCAPSED_STRING) reduce 199) ((T_CONTINUE) reduce 199) ((T_DEC) reduce 199) ((T_DECLARE) reduce 199) ((T_DIR) reduce 199) ((T_DNUMBER) reduce 199) ((T_DO) reduce 199) ((T_DOUBLE_CAST) reduce 199) ((T_ECHO) reduce 199) ((T_ELSE) reduce 199) ((T_ELSEIF) reduce 199) ((T_EMPTY) reduce 199) ((T_ENDIF) reduce 199) ((T_ENUM) reduce 199) ((T_EVAL) reduce 199) ((T_EXIT) reduce 199) ((T_FILE) reduce 199) ((T_FINAL) reduce 199) ((T_FN) reduce 199) ((T_FOR) reduce 199) ((T_FOREACH) reduce 199) ((T_FUNCTION) reduce 199) ((T_FUNC_C) reduce 199) ((T_GLOBAL) reduce 199) ((T_GOTO) reduce 199) ((T_HALT_COMPILER) reduce 199) ((T_IF) reduce 199) ((T_INC) reduce 199) ((T_INCLUDE) reduce 199) ((T_INCLUDE_ONCE) reduce 199) ((T_INLINE_HTML) reduce 199) ((T_INTERFACE) reduce 199) ((T_INT_CAST) reduce 199) ((T_ISSET) reduce 199) ((T_LINE) reduce 199) ((T_LIST) reduce 199) ((T_LNUMBER) reduce 199) ((T_MATCH) reduce 199) ((T_METHOD_C) reduce 199) ((T_NAME_FULLY_QUALIFIED) reduce 199) ((T_NAME_QUALIFIED) reduce 199) ((T_NAME_RELATIVE) reduce 199) ((T_NEW) reduce 199) ((T_NS_C) reduce 199) ((T_OBJECT_CAST) reduce 199) ((T_PRINT) reduce 199) ((T_PROPERTY_C) reduce 199) ((T_READONLY) reduce 199) ((T_REQUIRE) reduce 199) ((T_REQUIRE_ONCE) reduce 199) ((T_RETURN) reduce 199) ((T_START_HEREDOC) reduce 199) ((T_STATIC) reduce 199) ((T_STRING) reduce 199) ((T_STRING_CAST) reduce 199) ((T_SWITCH) reduce 199) ((T_THROW) reduce 199) ((T_TRAIT) reduce 199) ((T_TRAIT_C) reduce 199) ((T_TRY) reduce 199) ((T_UNSET) reduce 199) ((T_UNSET_CAST) reduce 199) ((T_VARIABLE) reduce 199) ((T_WHILE) reduce 199) ((T_YIELD) reduce 199) ((T_YIELD_FROM) reduce 199) (("[") reduce 199) (("`") reduce 199) (("{") reduce 199) (("~") reduce 199)) 25595 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDIF) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 25598 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_ELSE) reduce 145) ((T_ELSEIF) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDIF) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 25600 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDIF) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 25602 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDIF) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 25612 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDIF) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 25613 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDIF) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 25616 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDIF) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 25617 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDIF) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 25626 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDIF) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 25627 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDIF) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 25630 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDIF) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 25637 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_ELSE) reduce 200) ((T_ELSEIF) reduce 200) ((T_EMPTY) reduce 200) ((T_ENDIF) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 25639 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDIF) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 25645 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDIF) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 25650 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDIF) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 25651 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDIF) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 25654 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDIF) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 25656 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDIF) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 25663 ((("!") reduce 188) (("\"") reduce 188) (("$") reduce 188) (("(") reduce 188) (("+") reduce 188) (("-") reduce 188) ((";") reduce 188) (("@") reduce 188) ((T_ABSTRACT) reduce 188) ((T_ARRAY) reduce 188) ((T_ARRAY_CAST) reduce 188) ((T_ATTRIBUTE) reduce 188) ((T_BOOL_CAST) reduce 188) ((T_BREAK) reduce 188) ((T_CLASS) reduce 188) ((T_CLASS_C) reduce 188) ((T_CLONE) reduce 188) ((T_CONSTANT_ENCAPSED_STRING) reduce 188) ((T_CONTINUE) reduce 188) ((T_DEC) reduce 188) ((T_DECLARE) reduce 188) ((T_DIR) reduce 188) ((T_DNUMBER) reduce 188) ((T_DO) reduce 188) ((T_DOUBLE_CAST) reduce 188) ((T_ECHO) reduce 188) ((T_ELSE) reduce 188) ((T_ELSEIF) reduce 188) ((T_EMPTY) reduce 188) ((T_ENDIF) reduce 188) ((T_ENUM) reduce 188) ((T_EVAL) reduce 188) ((T_EXIT) reduce 188) ((T_FILE) reduce 188) ((T_FINAL) reduce 188) ((T_FN) reduce 188) ((T_FOR) reduce 188) ((T_FOREACH) reduce 188) ((T_FUNCTION) reduce 188) ((T_FUNC_C) reduce 188) ((T_GLOBAL) reduce 188) ((T_GOTO) reduce 188) ((T_HALT_COMPILER) reduce 188) ((T_IF) reduce 188) ((T_INC) reduce 188) ((T_INCLUDE) reduce 188) ((T_INCLUDE_ONCE) reduce 188) ((T_INLINE_HTML) reduce 188) ((T_INTERFACE) reduce 188) ((T_INT_CAST) reduce 188) ((T_ISSET) reduce 188) ((T_LINE) reduce 188) ((T_LIST) reduce 188) ((T_LNUMBER) reduce 188) ((T_MATCH) reduce 188) ((T_METHOD_C) reduce 188) ((T_NAME_FULLY_QUALIFIED) reduce 188) ((T_NAME_QUALIFIED) reduce 188) ((T_NAME_RELATIVE) reduce 188) ((T_NEW) reduce 188) ((T_NS_C) reduce 188) ((T_OBJECT_CAST) reduce 188) ((T_PRINT) reduce 188) ((T_PROPERTY_C) reduce 188) ((T_READONLY) reduce 188) ((T_REQUIRE) reduce 188) ((T_REQUIRE_ONCE) reduce 188) ((T_RETURN) reduce 188) ((T_START_HEREDOC) reduce 188) ((T_STATIC) reduce 188) ((T_STRING) reduce 188) ((T_STRING_CAST) reduce 188) ((T_SWITCH) reduce 188) ((T_THROW) reduce 188) ((T_TRAIT) reduce 188) ((T_TRAIT_C) reduce 188) ((T_TRY) reduce 188) ((T_UNSET) reduce 188) ((T_UNSET_CAST) reduce 188) ((T_VARIABLE) reduce 188) ((T_WHILE) reduce 188) ((T_YIELD) reduce 188) ((T_YIELD_FROM) reduce 188) (("[") reduce 188) (("`") reduce 188) (("{") reduce 188) (("~") reduce 188)) 25665 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDIF) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 25669 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_ELSE) reduce 249) ((T_ELSEIF) reduce 249) ((T_EMPTY) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("}") reduce 249) (("~") reduce 249)) 25674 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("}") reduce 240) (("~") reduce 240)) 25675 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 25678 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("}") reduce 241) (("~") reduce 241)) 25683 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 25686 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 25687 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("}") reduce 175) (("~") reduce 175)) 25690 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 25693 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_ELSE) reduce 176) ((T_ELSEIF) reduce 176) ((T_EMPTY) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("}") reduce 176) (("~") reduce 176)) 25700 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("}") reduce 170) (("~") reduce 170)) 25705 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 25709 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("}") reduce 222) (("~") reduce 222)) 25711 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("}") reduce 223) (("~") reduce 223)) 25715 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("}") reduce 224) (("~") reduce 224)) 25718 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("}") reduce 225) (("~") reduce 225)) 25719 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 25721 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 25723 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 25725 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 25727 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 25737 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 25738 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("}") reduce 218) (("~") reduce 218)) 25741 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("}") reduce 219) (("~") reduce 219)) 25742 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 25751 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 25752 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("}") reduce 216) (("~") reduce 216)) 25755 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("}") reduce 217) (("~") reduce 217)) 25757 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 25763 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 25768 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 25769 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("}") reduce 220) (("~") reduce 220)) 25772 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("}") reduce 221) (("~") reduce 221)) 25774 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 25776 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 26436 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_EMPTY) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("}") reduce 245) (("~") reduce 245)) 26447 ((("!") reduce 477) (("\"") reduce 477) (("$") reduce 477) (("(") reduce 477) (("+") reduce 477) (("-") reduce 477) ((";") reduce 477) (("@") reduce 477) ((T_ABSTRACT) reduce 477) ((T_ARRAY) reduce 477) ((T_ARRAY_CAST) reduce 477) ((T_ATTRIBUTE) reduce 477) ((T_BOOL_CAST) reduce 477) ((T_BREAK) reduce 477) ((T_CLASS) reduce 477) ((T_CLASS_C) reduce 477) ((T_CLONE) reduce 477) ((T_CONSTANT_ENCAPSED_STRING) reduce 477) ((T_CONTINUE) reduce 477) ((T_DEC) reduce 477) ((T_DECLARE) reduce 477) ((T_DIR) reduce 477) ((T_DNUMBER) reduce 477) ((T_DO) reduce 477) ((T_DOUBLE_CAST) reduce 477) ((T_ECHO) reduce 477) ((T_EMPTY) reduce 477) ((T_ENUM) reduce 477) ((T_EVAL) reduce 477) ((T_EXIT) reduce 477) ((T_FILE) reduce 477) ((T_FINAL) reduce 477) ((T_FN) reduce 477) ((T_FOR) reduce 477) ((T_FOREACH) reduce 477) ((T_FUNCTION) reduce 477) ((T_FUNC_C) reduce 477) ((T_GLOBAL) reduce 477) ((T_GOTO) reduce 477) ((T_HALT_COMPILER) reduce 477) ((T_IF) reduce 477) ((T_INC) reduce 477) ((T_INCLUDE) reduce 477) ((T_INCLUDE_ONCE) reduce 477) ((T_INLINE_HTML) reduce 477) ((T_INTERFACE) reduce 477) ((T_INT_CAST) reduce 477) ((T_ISSET) reduce 477) ((T_LINE) reduce 477) ((T_LIST) reduce 477) ((T_LNUMBER) reduce 477) ((T_MATCH) reduce 477) ((T_METHOD_C) reduce 477) ((T_NAME_FULLY_QUALIFIED) reduce 477) ((T_NAME_QUALIFIED) reduce 477) ((T_NAME_RELATIVE) reduce 477) ((T_NEW) reduce 477) ((T_NS_C) reduce 477) ((T_OBJECT_CAST) reduce 477) ((T_PRINT) reduce 477) ((T_PROPERTY_C) reduce 477) ((T_READONLY) reduce 477) ((T_REQUIRE) reduce 477) ((T_REQUIRE_ONCE) reduce 477) ((T_RETURN) reduce 477) ((T_START_HEREDOC) reduce 477) ((T_STATIC) reduce 477) ((T_STRING) reduce 477) ((T_STRING_CAST) reduce 477) ((T_SWITCH) reduce 477) ((T_THROW) reduce 477) ((T_TRAIT) reduce 477) ((T_TRAIT_C) reduce 477) ((T_TRY) reduce 477) ((T_UNSET) reduce 477) ((T_UNSET_CAST) reduce 477) ((T_VARIABLE) reduce 477) ((T_WHILE) reduce 477) ((T_YIELD) reduce 477) ((T_YIELD_FROM) reduce 477) (("[") reduce 477) (("`") reduce 477) (("{") reduce 477) (("}") reduce 477) (("~") reduce 477)) 26448 ((("!") reduce 182) (("\"") reduce 182) (("$") reduce 182) (("(") reduce 182) (("+") reduce 182) (("-") reduce 182) ((";") reduce 182) (("@") reduce 182) ((T_ABSTRACT) reduce 182) ((T_ARRAY) reduce 182) ((T_ARRAY_CAST) reduce 182) ((T_ATTRIBUTE) reduce 182) ((T_BOOL_CAST) reduce 182) ((T_BREAK) reduce 182) ((T_CLASS) reduce 182) ((T_CLASS_C) reduce 182) ((T_CLONE) reduce 182) ((T_CONSTANT_ENCAPSED_STRING) reduce 182) ((T_CONTINUE) reduce 182) ((T_DEC) reduce 182) ((T_DECLARE) reduce 182) ((T_DIR) reduce 182) ((T_DNUMBER) reduce 182) ((T_DO) reduce 182) ((T_DOUBLE_CAST) reduce 182) ((T_ECHO) reduce 182) ((T_EMPTY) reduce 182) ((T_ENUM) reduce 182) ((T_EVAL) reduce 182) ((T_EXIT) reduce 182) ((T_FILE) reduce 182) ((T_FINAL) reduce 182) ((T_FN) reduce 182) ((T_FOR) reduce 182) ((T_FOREACH) reduce 182) ((T_FUNCTION) reduce 182) ((T_FUNC_C) reduce 182) ((T_GLOBAL) reduce 182) ((T_GOTO) reduce 182) ((T_HALT_COMPILER) reduce 182) ((T_IF) reduce 182) ((T_INC) reduce 182) ((T_INCLUDE) reduce 182) ((T_INCLUDE_ONCE) reduce 182) ((T_INLINE_HTML) reduce 182) ((T_INTERFACE) reduce 182) ((T_INT_CAST) reduce 182) ((T_ISSET) reduce 182) ((T_LINE) reduce 182) ((T_LIST) reduce 182) ((T_LNUMBER) reduce 182) ((T_MATCH) reduce 182) ((T_METHOD_C) reduce 182) ((T_NAME_FULLY_QUALIFIED) reduce 182) ((T_NAME_QUALIFIED) reduce 182) ((T_NAME_RELATIVE) reduce 182) ((T_NEW) reduce 182) ((T_NS_C) reduce 182) ((T_OBJECT_CAST) reduce 182) ((T_PRINT) reduce 182) ((T_PROPERTY_C) reduce 182) ((T_READONLY) reduce 182) ((T_REQUIRE) reduce 182) ((T_REQUIRE_ONCE) reduce 182) ((T_RETURN) reduce 182) ((T_START_HEREDOC) reduce 182) ((T_STATIC) reduce 182) ((T_STRING) reduce 182) ((T_STRING_CAST) reduce 182) ((T_SWITCH) reduce 182) ((T_THROW) reduce 182) ((T_TRAIT) reduce 182) ((T_TRAIT_C) reduce 182) ((T_TRY) reduce 182) ((T_UNSET) reduce 182) ((T_UNSET_CAST) reduce 182) ((T_VARIABLE) reduce 182) ((T_WHILE) reduce 182) ((T_YIELD) reduce 182) ((T_YIELD_FROM) reduce 182) (("[") reduce 182) (("`") reduce 182) (("{") reduce 182) (("}") reduce 182) (("~") reduce 182)) 26449 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 26457 ((("!") reduce 187) (("\"") reduce 187) (("$") reduce 187) (("(") reduce 187) (("+") reduce 187) (("-") reduce 187) ((";") reduce 187) (("@") reduce 187) ((T_ABSTRACT) reduce 187) ((T_ARRAY) reduce 187) ((T_ARRAY_CAST) reduce 187) ((T_ATTRIBUTE) reduce 187) ((T_BOOL_CAST) reduce 187) ((T_BREAK) reduce 187) ((T_CLASS) reduce 187) ((T_CLASS_C) reduce 187) ((T_CLONE) reduce 187) ((T_CONSTANT_ENCAPSED_STRING) reduce 187) ((T_CONTINUE) reduce 187) ((T_DEC) reduce 187) ((T_DECLARE) reduce 187) ((T_DIR) reduce 187) ((T_DNUMBER) reduce 187) ((T_DO) reduce 187) ((T_DOUBLE_CAST) reduce 187) ((T_ECHO) reduce 187) ((T_EMPTY) reduce 187) ((T_ENUM) reduce 187) ((T_EVAL) reduce 187) ((T_EXIT) reduce 187) ((T_FILE) reduce 187) ((T_FINAL) reduce 187) ((T_FN) reduce 187) ((T_FOR) reduce 187) ((T_FOREACH) reduce 187) ((T_FUNCTION) reduce 187) ((T_FUNC_C) reduce 187) ((T_GLOBAL) reduce 187) ((T_GOTO) reduce 187) ((T_HALT_COMPILER) reduce 187) ((T_IF) reduce 187) ((T_INC) reduce 187) ((T_INCLUDE) reduce 187) ((T_INCLUDE_ONCE) reduce 187) ((T_INLINE_HTML) reduce 187) ((T_INTERFACE) reduce 187) ((T_INT_CAST) reduce 187) ((T_ISSET) reduce 187) ((T_LINE) reduce 187) ((T_LIST) reduce 187) ((T_LNUMBER) reduce 187) ((T_MATCH) reduce 187) ((T_METHOD_C) reduce 187) ((T_NAME_FULLY_QUALIFIED) reduce 187) ((T_NAME_QUALIFIED) reduce 187) ((T_NAME_RELATIVE) reduce 187) ((T_NEW) reduce 187) ((T_NS_C) reduce 187) ((T_OBJECT_CAST) reduce 187) ((T_PRINT) reduce 187) ((T_PROPERTY_C) reduce 187) ((T_READONLY) reduce 187) ((T_REQUIRE) reduce 187) ((T_REQUIRE_ONCE) reduce 187) ((T_RETURN) reduce 187) ((T_START_HEREDOC) reduce 187) ((T_STATIC) reduce 187) ((T_STRING) reduce 187) ((T_STRING_CAST) reduce 187) ((T_SWITCH) reduce 187) ((T_THROW) reduce 187) ((T_TRAIT) reduce 187) ((T_TRAIT_C) reduce 187) ((T_TRY) reduce 187) ((T_UNSET) reduce 187) ((T_UNSET_CAST) reduce 187) ((T_VARIABLE) reduce 187) ((T_WHILE) reduce 187) ((T_YIELD) reduce 187) ((T_YIELD_FROM) reduce 187) (("[") reduce 187) (("`") reduce 187) (("{") reduce 187) (("}") reduce 187) (("~") reduce 187)) 26458 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("}") reduce 144) (("~") reduce 144)) 26461 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_EMPTY) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("}") reduce 248) (("~") reduce 248)) 26465 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_EMPTY) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("}") reduce 249) (("~") reduce 249)) 26470 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("}") reduce 240) (("~") reduce 240)) 26471 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 26474 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("}") reduce 241) (("~") reduce 241)) 26479 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 26482 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 26483 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("}") reduce 175) (("~") reduce 175)) 26486 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 26489 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_EMPTY) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("}") reduce 176) (("~") reduce 176)) 26496 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("}") reduce 170) (("~") reduce 170)) 26501 ((("!") reduce 198) (("\"") reduce 198) (("$") reduce 198) (("(") reduce 198) (("+") reduce 198) (("-") reduce 198) ((";") reduce 198) (("@") reduce 198) ((T_ABSTRACT) reduce 198) ((T_ARRAY) reduce 198) ((T_ARRAY_CAST) reduce 198) ((T_ATTRIBUTE) reduce 198) ((T_BOOL_CAST) reduce 198) ((T_BREAK) reduce 198) ((T_CLASS) reduce 198) ((T_CLASS_C) reduce 198) ((T_CLONE) reduce 198) ((T_CONSTANT_ENCAPSED_STRING) reduce 198) ((T_CONTINUE) reduce 198) ((T_DEC) reduce 198) ((T_DECLARE) reduce 198) ((T_DIR) reduce 198) ((T_DNUMBER) reduce 198) ((T_DO) reduce 198) ((T_DOUBLE_CAST) reduce 198) ((T_ECHO) reduce 198) ((T_EMPTY) reduce 198) ((T_ENUM) reduce 198) ((T_EVAL) reduce 198) ((T_EXIT) reduce 198) ((T_FILE) reduce 198) ((T_FINAL) reduce 198) ((T_FN) reduce 198) ((T_FOR) reduce 198) ((T_FOREACH) reduce 198) ((T_FUNCTION) reduce 198) ((T_FUNC_C) reduce 198) ((T_GLOBAL) reduce 198) ((T_GOTO) reduce 198) ((T_HALT_COMPILER) reduce 198) ((T_IF) reduce 198) ((T_INC) reduce 198) ((T_INCLUDE) reduce 198) ((T_INCLUDE_ONCE) reduce 198) ((T_INLINE_HTML) reduce 198) ((T_INTERFACE) reduce 198) ((T_INT_CAST) reduce 198) ((T_ISSET) reduce 198) ((T_LINE) reduce 198) ((T_LIST) reduce 198) ((T_LNUMBER) reduce 198) ((T_MATCH) reduce 198) ((T_METHOD_C) reduce 198) ((T_NAME_FULLY_QUALIFIED) reduce 198) ((T_NAME_QUALIFIED) reduce 198) ((T_NAME_RELATIVE) reduce 198) ((T_NEW) reduce 198) ((T_NS_C) reduce 198) ((T_OBJECT_CAST) reduce 198) ((T_PRINT) reduce 198) ((T_PROPERTY_C) reduce 198) ((T_READONLY) reduce 198) ((T_REQUIRE) reduce 198) ((T_REQUIRE_ONCE) reduce 198) ((T_RETURN) reduce 198) ((T_START_HEREDOC) reduce 198) ((T_STATIC) reduce 198) ((T_STRING) reduce 198) ((T_STRING_CAST) reduce 198) ((T_SWITCH) reduce 198) ((T_THROW) reduce 198) ((T_TRAIT) reduce 198) ((T_TRAIT_C) reduce 198) ((T_TRY) reduce 198) ((T_UNSET) reduce 198) ((T_UNSET_CAST) reduce 198) ((T_VARIABLE) reduce 198) ((T_WHILE) reduce 198) ((T_YIELD) reduce 198) ((T_YIELD_FROM) reduce 198) (("[") reduce 198) (("`") reduce 198) (("{") reduce 198) (("}") reduce 198) (("~") reduce 198)) 26506 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 26510 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("}") reduce 222) (("~") reduce 222)) 26512 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("}") reduce 223) (("~") reduce 223)) 26516 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("}") reduce 224) (("~") reduce 224)) 26519 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("}") reduce 225) (("~") reduce 225)) 26520 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 26522 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 26524 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 26530 ((("!") reduce 199) (("\"") reduce 199) (("$") reduce 199) (("(") reduce 199) (("+") reduce 199) (("-") reduce 199) ((";") reduce 199) (("@") reduce 199) ((T_ABSTRACT) reduce 199) ((T_ARRAY) reduce 199) ((T_ARRAY_CAST) reduce 199) ((T_ATTRIBUTE) reduce 199) ((T_BOOL_CAST) reduce 199) ((T_BREAK) reduce 199) ((T_CLASS) reduce 199) ((T_CLASS_C) reduce 199) ((T_CLONE) reduce 199) ((T_CONSTANT_ENCAPSED_STRING) reduce 199) ((T_CONTINUE) reduce 199) ((T_DEC) reduce 199) ((T_DECLARE) reduce 199) ((T_DIR) reduce 199) ((T_DNUMBER) reduce 199) ((T_DO) reduce 199) ((T_DOUBLE_CAST) reduce 199) ((T_ECHO) reduce 199) ((T_EMPTY) reduce 199) ((T_ENUM) reduce 199) ((T_EVAL) reduce 199) ((T_EXIT) reduce 199) ((T_FILE) reduce 199) ((T_FINAL) reduce 199) ((T_FN) reduce 199) ((T_FOR) reduce 199) ((T_FOREACH) reduce 199) ((T_FUNCTION) reduce 199) ((T_FUNC_C) reduce 199) ((T_GLOBAL) reduce 199) ((T_GOTO) reduce 199) ((T_HALT_COMPILER) reduce 199) ((T_IF) reduce 199) ((T_INC) reduce 199) ((T_INCLUDE) reduce 199) ((T_INCLUDE_ONCE) reduce 199) ((T_INLINE_HTML) reduce 199) ((T_INTERFACE) reduce 199) ((T_INT_CAST) reduce 199) ((T_ISSET) reduce 199) ((T_LINE) reduce 199) ((T_LIST) reduce 199) ((T_LNUMBER) reduce 199) ((T_MATCH) reduce 199) ((T_METHOD_C) reduce 199) ((T_NAME_FULLY_QUALIFIED) reduce 199) ((T_NAME_QUALIFIED) reduce 199) ((T_NAME_RELATIVE) reduce 199) ((T_NEW) reduce 199) ((T_NS_C) reduce 199) ((T_OBJECT_CAST) reduce 199) ((T_PRINT) reduce 199) ((T_PROPERTY_C) reduce 199) ((T_READONLY) reduce 199) ((T_REQUIRE) reduce 199) ((T_REQUIRE_ONCE) reduce 199) ((T_RETURN) reduce 199) ((T_START_HEREDOC) reduce 199) ((T_STATIC) reduce 199) ((T_STRING) reduce 199) ((T_STRING_CAST) reduce 199) ((T_SWITCH) reduce 199) ((T_THROW) reduce 199) ((T_TRAIT) reduce 199) ((T_TRAIT_C) reduce 199) ((T_TRY) reduce 199) ((T_UNSET) reduce 199) ((T_UNSET_CAST) reduce 199) ((T_VARIABLE) reduce 199) ((T_WHILE) reduce 199) ((T_YIELD) reduce 199) ((T_YIELD_FROM) reduce 199) (("[") reduce 199) (("`") reduce 199) (("{") reduce 199) (("}") reduce 199) (("~") reduce 199)) 26534 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("}") reduce 242) (("~") reduce 242)) 26537 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("}") reduce 145) (("~") reduce 145)) 26539 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 26541 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 26551 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 26552 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("}") reduce 218) (("~") reduce 218)) 26555 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("}") reduce 219) (("~") reduce 219)) 26556 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 26565 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 26566 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("}") reduce 216) (("~") reduce 216)) 26569 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("}") reduce 217) (("~") reduce 217)) 26576 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("}") reduce 200) (("~") reduce 200)) 26578 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 26584 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 26589 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 26590 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("}") reduce 220) (("~") reduce 220)) 26593 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("}") reduce 221) (("~") reduce 221)) 26595 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 26602 ((("!") reduce 188) (("\"") reduce 188) (("$") reduce 188) (("(") reduce 188) (("+") reduce 188) (("-") reduce 188) ((";") reduce 188) (("@") reduce 188) ((T_ABSTRACT) reduce 188) ((T_ARRAY) reduce 188) ((T_ARRAY_CAST) reduce 188) ((T_ATTRIBUTE) reduce 188) ((T_BOOL_CAST) reduce 188) ((T_BREAK) reduce 188) ((T_CLASS) reduce 188) ((T_CLASS_C) reduce 188) ((T_CLONE) reduce 188) ((T_CONSTANT_ENCAPSED_STRING) reduce 188) ((T_CONTINUE) reduce 188) ((T_DEC) reduce 188) ((T_DECLARE) reduce 188) ((T_DIR) reduce 188) ((T_DNUMBER) reduce 188) ((T_DO) reduce 188) ((T_DOUBLE_CAST) reduce 188) ((T_ECHO) reduce 188) ((T_EMPTY) reduce 188) ((T_ENUM) reduce 188) ((T_EVAL) reduce 188) ((T_EXIT) reduce 188) ((T_FILE) reduce 188) ((T_FINAL) reduce 188) ((T_FN) reduce 188) ((T_FOR) reduce 188) ((T_FOREACH) reduce 188) ((T_FUNCTION) reduce 188) ((T_FUNC_C) reduce 188) ((T_GLOBAL) reduce 188) ((T_GOTO) reduce 188) ((T_HALT_COMPILER) reduce 188) ((T_IF) reduce 188) ((T_INC) reduce 188) ((T_INCLUDE) reduce 188) ((T_INCLUDE_ONCE) reduce 188) ((T_INLINE_HTML) reduce 188) ((T_INTERFACE) reduce 188) ((T_INT_CAST) reduce 188) ((T_ISSET) reduce 188) ((T_LINE) reduce 188) ((T_LIST) reduce 188) ((T_LNUMBER) reduce 188) ((T_MATCH) reduce 188) ((T_METHOD_C) reduce 188) ((T_NAME_FULLY_QUALIFIED) reduce 188) ((T_NAME_QUALIFIED) reduce 188) ((T_NAME_RELATIVE) reduce 188) ((T_NEW) reduce 188) ((T_NS_C) reduce 188) ((T_OBJECT_CAST) reduce 188) ((T_PRINT) reduce 188) ((T_PROPERTY_C) reduce 188) ((T_READONLY) reduce 188) ((T_REQUIRE) reduce 188) ((T_REQUIRE_ONCE) reduce 188) ((T_RETURN) reduce 188) ((T_START_HEREDOC) reduce 188) ((T_STATIC) reduce 188) ((T_STRING) reduce 188) ((T_STRING_CAST) reduce 188) ((T_SWITCH) reduce 188) ((T_THROW) reduce 188) ((T_TRAIT) reduce 188) ((T_TRAIT_C) reduce 188) ((T_TRY) reduce 188) ((T_UNSET) reduce 188) ((T_UNSET_CAST) reduce 188) ((T_VARIABLE) reduce 188) ((T_WHILE) reduce 188) ((T_YIELD) reduce 188) ((T_YIELD_FROM) reduce 188) (("[") reduce 188) (("`") reduce 188) (("{") reduce 188) (("}") reduce 188) (("~") reduce 188)) 26604 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 26606 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) (("{") reduce 278) (("|") reduce 278)) 26607 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) (("{") reduce 279) (("|") reduce 279)) 26608 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) (("{") reduce 94) (("|") reduce 94)) 26609 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) (("{") reduce 93) (("|") reduce 93)) 26610 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) (("{") reduce 95) (("|") reduce 95)) 26611 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267) (("{") reduce 267) (("|") reduce 267)) 26612 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) (("{") reduce 92) (("|") reduce 92)) 26613 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) (("{") reduce 265)) 26614 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) (("{") reduce 280) (("|") reduce 280)) 26615 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) (("{") reduce 262) (("|") reduce 268)) 26616 ((("{") reduce 288)) 26617 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266) (("{") reduce 266) (("|") reduce 266)) 26618 ((("{") reduce 264) (("|") shift)) 26622 ((("{") reduce 278) (("|") reduce 278)) 26623 ((("{") reduce 279) (("|") reduce 279)) 26624 ((("{") reduce 94) (("|") reduce 94)) 26625 ((("{") reduce 93) (("|") reduce 93)) 26626 ((("{") reduce 95) (("|") reduce 95)) 26627 ((("{") reduce 267) (("|") reduce 267)) 26628 ((("{") reduce 92) (("|") reduce 92)) 26629 ((("{") reduce 280) (("|") reduce 280)) 26630 ((("{") reduce 268) (("|") reduce 268)) 26631 ((("{") reduce 266) (("|") reduce 266)) 26632 ((("{") reduce 270) (("|") reduce 270)) 26634 ((("{") reduce 269) (("|") reduce 269)) 26636 ((("{") reduce 271) (("|") reduce 271)) 26638 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) (("{") reduce 278)) 26639 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) (("{") reduce 279)) 26640 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) (("{") reduce 94)) 26641 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) (("{") reduce 93)) 26642 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) (("{") reduce 95)) 26643 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267) (("{") reduce 267)) 26644 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) (("{") reduce 92)) 26645 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) (("{") reduce 280)) 26646 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) (("{") reduce 272)) 26647 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266) (("{") reduce 266)) 26649 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) (("{") reduce 273)) 26650 ((("{") reduce 278)) 26651 ((("{") reduce 279)) 26652 ((("{") reduce 94)) 26653 ((("{") reduce 93)) 26654 ((("{") reduce 95)) 26655 ((("{") reduce 267)) 26656 ((("{") reduce 92)) 26657 ((("{") reduce 280)) 26658 ((("{") reduce 263)) 26659 ((("{") reduce 266)) 26660 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_VARIABLE) shift)) 26661 (((T_VARIABLE) reduce 80)) 26662 (((T_VARIABLE) reduce 81)) 26663 (((")") reduce 485) ((",") reduce 485)) 26665 (((")") reduce 484) ((",") reduce 484)) 26667 (((")") reduce 125) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_VARIABLE) shift)) 26669 (((":") reduce 482) (("{") reduce 482)) 26670 (((")") reduce 483) ((",") reduce 483)) 26671 (((")") reduce 486) ((",") reduce 486)) 26673 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_ELLIPSIS) reduce 278) ((T_VARIABLE) reduce 278) (("|") reduce 278)) 26674 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) ((T_ELLIPSIS) reduce 279) ((T_VARIABLE) reduce 279) (("|") reduce 279)) 26675 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ELLIPSIS) reduce 94) ((T_VARIABLE) reduce 94) (("|") reduce 94)) 26676 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ELLIPSIS) reduce 93) ((T_VARIABLE) reduce 93) (("|") reduce 93)) 26677 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_ELLIPSIS) reduce 95) ((T_VARIABLE) reduce 95) (("|") reduce 95)) 26678 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ELLIPSIS) reduce 92) ((T_VARIABLE) reduce 92) (("|") reduce 92)) 26679 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 277) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ELLIPSIS) reduce 277) ((T_VARIABLE) reduce 277)) 26680 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_ELLIPSIS) reduce 280) ((T_VARIABLE) reduce 280) (("|") reduce 280)) 26681 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ELLIPSIS) reduce 183) ((T_VARIABLE) reduce 183)) 26682 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 261) ((T_ELLIPSIS) reduce 261) ((T_VARIABLE) reduce 261)) 26683 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ELLIPSIS) reduce 274) ((T_VARIABLE) reduce 274) (("|") reduce 281)) 26684 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 276) ((T_ELLIPSIS) reduce 276) ((T_VARIABLE) reduce 276) (("|") shift)) 26688 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_ELLIPSIS) reduce 278) ((T_VARIABLE) reduce 278) (("|") reduce 278)) 26689 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) ((T_ELLIPSIS) reduce 279) ((T_VARIABLE) reduce 279) (("|") reduce 279)) 26690 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ELLIPSIS) reduce 94) ((T_VARIABLE) reduce 94) (("|") reduce 94)) 26691 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ELLIPSIS) reduce 93) ((T_VARIABLE) reduce 93) (("|") reduce 93)) 26692 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_ELLIPSIS) reduce 95) ((T_VARIABLE) reduce 95) (("|") reduce 95)) 26693 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ELLIPSIS) reduce 92) ((T_VARIABLE) reduce 92) (("|") reduce 92)) 26694 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_ELLIPSIS) reduce 280) ((T_VARIABLE) reduce 280) (("|") reduce 280)) 26695 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_ELLIPSIS) reduce 281) ((T_VARIABLE) reduce 281) (("|") reduce 281)) 26696 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_ELLIPSIS) reduce 283) ((T_VARIABLE) reduce 283) (("|") reduce 283)) 26698 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_ELLIPSIS) reduce 282) ((T_VARIABLE) reduce 282) (("|") reduce 282)) 26700 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 284) ((T_ELLIPSIS) reduce 284) ((T_VARIABLE) reduce 284) (("|") reduce 284)) 26702 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_ELLIPSIS) reduce 278) ((T_VARIABLE) reduce 278)) 26703 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) ((T_ELLIPSIS) reduce 279) ((T_VARIABLE) reduce 279)) 26704 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ELLIPSIS) reduce 94) ((T_VARIABLE) reduce 94)) 26705 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ELLIPSIS) reduce 93) ((T_VARIABLE) reduce 93)) 26706 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_ELLIPSIS) reduce 95) ((T_VARIABLE) reduce 95)) 26707 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ELLIPSIS) reduce 92) ((T_VARIABLE) reduce 92)) 26708 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_ELLIPSIS) reduce 280) ((T_VARIABLE) reduce 280)) 26709 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 285) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 285) ((T_ELLIPSIS) reduce 285) ((T_VARIABLE) reduce 285)) 26710 (((T_ELLIPSIS) reduce 184) ((T_VARIABLE) reduce 184)) 26711 (((T_ELLIPSIS) shift) ((T_VARIABLE) reduce 185)) 26712 (((T_VARIABLE) reduce 186)) 26714 (((")") reduce 476) ((",") reduce 476) (("=") reduce 476) (("{") reduce 476)) 26715 (((")") reduce 361) ((",") reduce 361) (("=") shift) (("{") shift)) 26717 (((")") reduce 258) ((",") reduce 258)) 26720 (((")") reduce 362) ((",") reduce 362)) 26731 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 26733 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("{") reduce 504) (("|") reduce 504)) 26735 ((("%") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 26736 ((("%") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 26740 ((("%") reduce 500) (("(") shift) ((")") reduce 500) (("*") reduce 500) (("+") reduce 500) ((",") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("{") reduce 500) (("|") reduce 500)) 26741 ((("%") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 26742 ((("%") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 26748 ((("%") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 26750 ((("%") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 26752 ((("%") reduce 519) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 26753 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 26754 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 26755 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 26757 ((("%") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 26760 ((("%") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 26766 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 26769 ((("%") reduce 518) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 26771 ((("%") reduce 551) (("(") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("{") reduce 551) (("|") reduce 551)) 26772 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) ((")") reduce 462) (("*") reduce 462) (("+") shift) ((",") reduce 462) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("{") reduce 462) (("|") reduce 462) (("~") shift)) 26779 ((("%") reduce 547) (("(") reduce 538) ((")") reduce 547) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("{") reduce 547) (("|") reduce 547)) 26780 ((("%") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 26782 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 26783 ((("%") reduce 511) (("(") reduce 540) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 26784 ((("%") shift) ((")") reduce 361) (("*") shift) (("+") shift) ((",") reduce 361) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") shift) (("|") shift)) 26787 ((("%") reduce 546) (("(") reduce 546) ((")") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("{") reduce 546) (("|") reduce 546)) 26788 ((("%") reduce 467) ((")") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("{") reduce 467) (("|") reduce 467)) 26789 ((("%") reduce 449) ((")") reduce 449) (("*") reduce 449) (("+") reduce 449) ((",") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("{") reduce 449) (("|") reduce 449)) 26790 ((("%") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 26791 ((("%") reduce 514) (("(") shift) ((")") reduce 514) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 26792 ((("%") reduce 444) (("(") reduce 541) ((")") reduce 444) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("{") reduce 444) (("|") reduce 444)) 26793 ((("%") reduce 445) ((")") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("{") reduce 445) (("|") reduce 445)) 26794 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("{") reduce 459) (("|") reduce 459)) 26795 ((("%") reduce 542) (("(") reduce 542) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("{") reduce 542) (("|") reduce 542)) 26796 ((("%") reduce 548) ((")") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("{") reduce 548) (("|") reduce 548)) 26797 ((("%") reduce 388) ((")") reduce 388) (("*") reduce 388) (("+") reduce 388) ((",") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("{") reduce 388) (("|") reduce 388)) 26800 ((("%") reduce 432) ((")") reduce 432) (("*") reduce 432) (("+") reduce 432) ((",") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("{") reduce 432) (("|") reduce 432)) 26831 ((("%") shift) ((")") reduce 416) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 416) (("|") reduce 416)) 26832 ((("%") shift) ((")") reduce 419) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("{") reduce 419) (("|") reduce 419)) 26833 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("{") reduce 428) (("|") reduce 428)) 26834 ((("%") shift) ((")") reduce 441) (("*") shift) (("+") shift) ((",") reduce 441) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("{") reduce 441) (("|") reduce 441)) 26835 ((("%") shift) ((")") reduce 427) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("{") reduce 427) (("|") reduce 427)) 26836 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("{") reduce 424) (("|") reduce 424)) 26837 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 415) (("|") shift)) 26838 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 413) (("|") shift)) 26839 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 414) (("|") shift)) 26840 ((("%") shift) ((")") reduce 438) (("*") shift) (("+") shift) ((",") reduce 438) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("{") reduce 438) (("|") reduce 438)) 26841 ((("%") shift) ((")") reduce 434) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("{") reduce 434) (("|") reduce 434)) 26842 ((("%") shift) ((")") reduce 436) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("{") reduce 436) (("|") reduce 436)) 26843 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("{") reduce 433) (("|") reduce 433)) 26844 ((("%") shift) ((")") reduce 440) (("*") shift) (("+") shift) ((",") reduce 440) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("{") reduce 440) (("|") reduce 440)) 26845 ((("%") shift) ((")") reduce 435) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("{") reduce 435) (("|") reduce 435)) 26848 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 26849 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 26850 ((("%") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 26851 ((("%") reduce 492) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 26852 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 26853 ((("%") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("{") reduce 551) (("|") reduce 551)) 26854 ((("%") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("{") reduce 494) (("|") reduce 494)) 26855 ((("%") reduce 442) ((")") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("{") reduce 442) (("|") reduce 442)) 26856 ((("%") reduce 493) ((")") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("{") reduce 493) (("|") reduce 493)) 26857 ((("%") reduce 495) ((")") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("{") reduce 495) (("|") reduce 495)) 26858 ((("%") reduce 556) ((")") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("{") reduce 556) (("|") reduce 556)) 26864 ((("%") reduce 557) ((")") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("{") reduce 557) (("|") reduce 557)) 26865 ((("%") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 26866 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 26867 ((("%") reduce 558) ((")") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("{") reduce 558) (("|") reduce 558)) 26868 ((("%") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 26871 ((("%") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 26872 ((("%") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 26874 ((("%") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 26876 ((("%") reduce 496) ((")") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("{") reduce 496) (("|") reduce 496)) 26877 ((("%") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("{") reduce 553) (("|") reduce 553)) 26880 ((("%") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("{") reduce 552) (("|") reduce 552)) 26881 ((("%") shift) ((")") reduce 448) (("*") shift) (("+") shift) ((",") reduce 448) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 448) (("|") shift)) 26882 ((("%") shift) ((")") reduce 411) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 411) (("|") shift)) 26883 ((("%") shift) ((")") reduce 412) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 412) (("|") shift)) 26884 ((("%") shift) ((")") reduce 417) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("{") reduce 417) (("|") reduce 417)) 26885 ((("%") shift) ((")") reduce 418) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("{") reduce 418) (("|") reduce 418)) 26889 ((("%") shift) ((")") reduce 446) (("*") shift) (("+") shift) ((",") reduce 446) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 446) (("|") shift)) 26890 ((("%") shift) ((")") reduce 447) (("*") shift) (("+") shift) ((",") reduce 447) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 447) (("|") shift)) 26891 ((("%") shift) ((")") reduce 439) (("*") shift) (("+") shift) ((",") reduce 439) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("{") reduce 439) (("|") reduce 439)) 26892 ((("%") shift) ((")") reduce 437) (("*") shift) (("+") shift) ((",") reduce 437) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("{") reduce 437) (("|") reduce 437)) 26893 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("{") reduce 425) (("|") reduce 425)) 26894 ((("%") shift) ((")") reduce 420) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("{") reduce 420) (("|") reduce 420)) 26895 ((("%") shift) ((")") reduce 422) (("*") shift) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("{") reduce 422) (("|") reduce 422)) 26896 ((("%") shift) ((")") reduce 421) (("*") shift) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("{") reduce 421) (("|") reduce 421)) 26897 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("{") reduce 423) (("|") reduce 423)) 26898 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("{") reduce 426) (("|") reduce 426)) 26900 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 26901 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 26902 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 26903 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 26904 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 26905 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 26906 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 26907 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 26908 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 26909 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 26910 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 26911 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 26912 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 26913 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 26914 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 26915 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 26916 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 26917 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 26918 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 26919 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 26920 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 26921 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 26922 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 26923 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 26924 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 26925 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 26926 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 26927 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 26928 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 26929 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 26930 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 26931 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 26932 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 26933 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 26934 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 26935 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 26936 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 26937 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 26938 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 26939 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 26940 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 26941 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 26942 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 26943 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 26944 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 26945 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 26946 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 26947 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 26948 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 26949 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 26950 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 26951 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 26952 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 26953 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 26954 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 26955 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 26956 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 26957 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 26958 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 26959 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 26960 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 26961 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 26962 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 26963 ((("%") reduce 79) (("(") reduce 79) ((")") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("{") reduce 79) (("|") reduce 79)) 26964 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 26965 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 26966 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 26967 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 26968 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 26969 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 26970 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 26971 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 26972 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 26973 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 26974 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 26975 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 26976 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 26977 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 26978 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 26979 ((("%") reduce 525) (("(") reduce 562) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 26981 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 26982 ((("%") reduce 83) (("(") reduce 83) ((")") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("{") reduce 83) (("|") reduce 83)) 26983 ((("%") reduce 555) (("(") reduce 564) ((")") reduce 555) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("{") reduce 555) (("|") reduce 555)) 26986 ((("%") reduce 527) (("(") reduce 563) ((")") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("{") reduce 527) (("|") reduce 527)) 26988 ((("%") reduce 490) (("(") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 26989 ((("%") reduce 289) (("(") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("{") reduce 289) (("|") reduce 289)) 26993 ((("%") reduce 290) (("(") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("{") reduce 290) (("|") reduce 290)) 26994 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 26999 ((("%") reduce 409) ((")") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("{") reduce 409) (("|") reduce 409)) 27001 ((("%") reduce 407) ((")") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("{") reduce 407) (("|") reduce 407)) 27011 ((("%") shift) ((")") reduce 403) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 403) (("|") shift)) 27012 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 405) (("|") shift)) 27013 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 404) (("|") shift)) 27014 ((("%") shift) ((")") reduce 397) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 397) (("|") shift)) 27015 ((("%") shift) ((")") reduce 394) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 394) (("|") shift)) 27016 ((("%") shift) ((")") reduce 402) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 402) (("|") shift)) 27017 ((("%") shift) ((")") reduce 396) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 396) (("|") shift)) 27018 ((("%") shift) ((")") reduce 400) (("*") shift) (("+") shift) ((",") reduce 400) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 400) (("|") shift)) 27019 ((("%") shift) ((")") reduce 395) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 395) (("|") shift)) 27020 ((("%") shift) ((")") reduce 398) (("*") shift) (("+") shift) ((",") reduce 398) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 398) (("|") shift)) 27021 ((("%") shift) ((")") reduce 399) (("*") shift) (("+") shift) ((",") reduce 399) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 399) (("|") shift)) 27022 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 406) (("|") shift)) 27023 ((("%") shift) ((")") reduce 401) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 401) (("|") shift)) 27025 ((("%") shift) ((")") reduce 391) (("*") shift) (("+") shift) ((",") reduce 391) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 391) (("|") shift)) 27028 ((("%") reduce 551) (("(") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("{") reduce 551) (("|") reduce 551)) 27031 ((("%") reduce 547) (("(") reduce 538) ((")") reduce 547) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("{") reduce 547) (("|") reduce 547)) 27033 ((("%") reduce 546) (("(") reduce 546) ((")") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("{") reduce 546) (("|") reduce 546)) 27035 ((("%") reduce 542) (("(") reduce 542) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("{") reduce 542) (("|") reduce 542)) 27036 ((("%") reduce 548) ((")") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("{") reduce 548) (("|") reduce 548)) 27037 ((("%") reduce 392) ((")") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("{") reduce 392) (("|") reduce 392)) 27041 ((("%") reduce 555) (("(") reduce 564) ((")") reduce 555) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("{") reduce 555) (("|") reduce 555)) 27043 ((("%") reduce 490) (("(") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 27044 ((("%") reduce 289) (("(") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("{") reduce 289) (("|") reduce 289)) 27048 ((("%") reduce 290) (("(") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("{") reduce 290) (("|") reduce 290)) 27049 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 27050 ((("%") reduce 487) (("(") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 27053 ((("%") reduce 554) (("(") reduce 564) ((")") reduce 554) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("{") reduce 554) (("|") reduce 554)) 27054 ((("%") reduce 489) (("(") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 27055 ((("%") reduce 491) (("(") reduce 491) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 27060 ((("%") reduce 543) (("(") reduce 543) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("{") reduce 543) (("|") reduce 543)) 27061 ((("%") reduce 565) (("(") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 27062 ((("%") reduce 549) (("(") shift) ((")") reduce 549) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("{") reduce 549) (("|") reduce 549)) 27063 ((("%") reduce 567) (("(") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 27066 ((("%") reduce 566) (("(") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 27067 ((("%") reduce 544) (("(") reduce 544) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("{") reduce 544) (("|") reduce 544)) 27068 ((("%") reduce 550) (("(") shift) ((")") reduce 550) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("{") reduce 550) (("|") reduce 550)) 27069 ((("%") reduce 545) (("(") reduce 545) ((")") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("{") reduce 545) (("|") reduce 545)) 27070 ((("%") reduce 488) (("(") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 27071 ((("%") reduce 553) (("(") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("{") reduce 553) (("|") reduce 553)) 27074 ((("%") reduce 552) (("(") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("{") reduce 552) (("|") reduce 552)) 27075 ((("%") reduce 487) (("(") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 27086 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("{") reduce 477) (("|") reduce 477)) 27087 ((("%") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 27097 ((("%") shift) ((")") reduce 477) (("*") shift) (("+") shift) ((",") reduce 477) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 477) (("|") shift)) 27098 ((("%") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 27099 (((")") reduce 259) ((",") reduce 259)) 27101 ((("%") reduce 524) (("(") reduce 562) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 27103 ((("%") reduce 554) (("(") reduce 564) ((")") reduce 554) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("{") reduce 554) (("|") reduce 554)) 27106 ((("%") reduce 526) (("(") reduce 563) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 27107 ((("%") reduce 489) (("(") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 27108 ((("%") reduce 491) (("(") reduce 491) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 27110 ((("%") reduce 468) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("{") reduce 468) (("|") reduce 468)) 27111 ((("%") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 27116 ((("%") reduce 543) (("(") reduce 543) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("{") reduce 543) (("|") reduce 543)) 27117 ((("%") reduce 565) (("(") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 27118 ((("%") reduce 549) (("(") shift) ((")") reduce 549) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("{") reduce 549) (("|") reduce 549)) 27119 ((("%") reduce 567) (("(") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 27122 ((("%") reduce 566) (("(") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 27123 ((("%") reduce 544) (("(") reduce 544) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("{") reduce 544) (("|") reduce 544)) 27124 ((("%") reduce 550) (("(") shift) ((")") reduce 550) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("{") reduce 550) (("|") reduce 550)) 27125 ((("%") reduce 545) (("(") reduce 545) ((")") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("{") reduce 545) (("|") reduce 545)) 27127 ((("%") reduce 460) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("{") reduce 460) (("|") reduce 460)) 27129 ((("%") reduce 503) (("(") reduce 503) ((")") reduce 503) (("*") reduce 503) (("+") reduce 503) ((",") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("{") reduce 503) (("|") reduce 503)) 27131 ((("%") shift) ((")") reduce 390) (("*") shift) (("+") shift) ((",") reduce 390) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 390) (("|") shift)) 27132 ((("%") shift) ((")") reduce 465) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 465) (("|") shift)) 27143 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 27145 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("{") reduce 504) (("|") reduce 504)) 27147 ((("%") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 27148 ((("%") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 27152 ((("%") reduce 500) (("(") shift) ((")") reduce 500) (("*") reduce 500) (("+") reduce 500) ((",") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_DOUBLE_ARROW) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("{") reduce 500) (("|") reduce 500)) 27153 ((("%") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 27154 ((("%") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 27160 ((("%") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 27162 ((("%") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 27164 ((("%") reduce 519) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 27165 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 27166 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 27167 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 27169 ((("%") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 27172 ((("%") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 27178 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 27181 ((("%") reduce 518) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 27183 ((("%") reduce 551) (("(") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("{") reduce 551) (("|") reduce 551)) 27184 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) ((")") reduce 462) (("*") reduce 462) (("+") shift) ((",") reduce 462) (("-") shift) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 462) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("{") reduce 462) (("|") reduce 462) (("~") shift)) 27191 ((("%") reduce 547) (("(") reduce 538) ((")") reduce 547) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("{") reduce 547) (("|") reduce 547)) 27192 ((("%") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 27194 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 27195 ((("%") reduce 511) (("(") reduce 540) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 27196 ((("%") shift) ((")") reduce 463) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 463) (("|") shift)) 27199 ((("%") reduce 546) (("(") reduce 546) ((")") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("{") reduce 546) (("|") reduce 546)) 27200 ((("%") reduce 467) ((")") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("{") reduce 467) (("|") reduce 467)) 27201 ((("%") reduce 449) ((")") reduce 449) (("*") reduce 449) (("+") reduce 449) ((",") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("{") reduce 449) (("|") reduce 449)) 27202 ((("%") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 27203 ((("%") reduce 514) (("(") shift) ((")") reduce 514) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 27204 ((("%") reduce 444) (("(") reduce 541) ((")") reduce 444) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("{") reduce 444) (("|") reduce 444)) 27205 ((("%") reduce 445) ((")") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("{") reduce 445) (("|") reduce 445)) 27206 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("{") reduce 459) (("|") reduce 459)) 27207 ((("%") reduce 542) (("(") reduce 542) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("{") reduce 542) (("|") reduce 542)) 27208 ((("%") reduce 548) ((")") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("{") reduce 548) (("|") reduce 548)) 27209 ((("%") reduce 388) ((")") reduce 388) (("*") reduce 388) (("+") reduce 388) ((",") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 388) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("{") reduce 388) (("|") reduce 388)) 27212 ((("%") reduce 432) ((")") reduce 432) (("*") reduce 432) (("+") reduce 432) ((",") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("{") reduce 432) (("|") reduce 432)) 27243 ((("%") shift) ((")") reduce 416) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 416) (("|") reduce 416)) 27244 ((("%") shift) ((")") reduce 419) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("{") reduce 419) (("|") reduce 419)) 27245 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") reduce 428) (("/") shift) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("{") reduce 428) (("|") reduce 428)) 27246 ((("%") shift) ((")") reduce 441) (("*") shift) (("+") shift) ((",") reduce 441) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("{") reduce 441) (("|") reduce 441)) 27247 ((("%") shift) ((")") reduce 427) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") reduce 427) (("/") shift) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("{") reduce 427) (("|") reduce 427)) 27248 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("{") reduce 424) (("|") reduce 424)) 27249 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 415) (("|") shift)) 27250 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 413) (("|") shift)) 27251 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 414) (("|") shift)) 27252 ((("%") shift) ((")") reduce 438) (("*") shift) (("+") shift) ((",") reduce 438) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("{") reduce 438) (("|") reduce 438)) 27253 ((("%") shift) ((")") reduce 434) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("{") reduce 434) (("|") reduce 434)) 27254 ((("%") shift) ((")") reduce 436) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("{") reduce 436) (("|") reduce 436)) 27255 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("{") reduce 433) (("|") reduce 433)) 27256 ((("%") shift) ((")") reduce 440) (("*") shift) (("+") shift) ((",") reduce 440) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("{") reduce 440) (("|") reduce 440)) 27257 ((("%") shift) ((")") reduce 435) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("{") reduce 435) (("|") reduce 435)) 27260 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 27261 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 27262 ((("%") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 27263 ((("%") reduce 492) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 27264 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 27265 ((("%") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("{") reduce 551) (("|") reduce 551)) 27266 ((("%") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("{") reduce 494) (("|") reduce 494)) 27267 ((("%") reduce 442) ((")") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("{") reduce 442) (("|") reduce 442)) 27268 ((("%") reduce 493) ((")") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("{") reduce 493) (("|") reduce 493)) 27269 ((("%") reduce 495) ((")") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("{") reduce 495) (("|") reduce 495)) 27270 ((("%") reduce 556) ((")") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("{") reduce 556) (("|") reduce 556)) 27276 ((("%") reduce 557) ((")") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("{") reduce 557) (("|") reduce 557)) 27277 ((("%") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 27278 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 27279 ((("%") reduce 558) ((")") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("{") reduce 558) (("|") reduce 558)) 27280 ((("%") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 27283 ((("%") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 27284 ((("%") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 27286 ((("%") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 27288 ((("%") reduce 496) ((")") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("{") reduce 496) (("|") reduce 496)) 27289 ((("%") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("{") reduce 553) (("|") reduce 553)) 27292 ((("%") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("{") reduce 552) (("|") reduce 552)) 27293 ((("%") shift) ((")") reduce 448) (("*") shift) (("+") shift) ((",") reduce 448) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 448) (("|") shift)) 27294 ((("%") shift) ((")") reduce 411) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 411) (("|") shift)) 27295 ((("%") shift) ((")") reduce 412) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 412) (("|") shift)) 27296 ((("%") shift) ((")") reduce 417) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("{") reduce 417) (("|") reduce 417)) 27297 ((("%") shift) ((")") reduce 418) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("{") reduce 418) (("|") reduce 418)) 27301 ((("%") shift) ((")") reduce 446) (("*") shift) (("+") shift) ((",") reduce 446) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 446) (("|") shift)) 27302 ((("%") shift) ((")") reduce 447) (("*") shift) (("+") shift) ((",") reduce 447) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 447) (("|") shift)) 27303 ((("%") shift) ((")") reduce 439) (("*") shift) (("+") shift) ((",") reduce 439) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("{") reduce 439) (("|") reduce 439)) 27304 ((("%") shift) ((")") reduce 437) (("*") shift) (("+") shift) ((",") reduce 437) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("{") reduce 437) (("|") reduce 437)) 27305 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("{") reduce 425) (("|") reduce 425)) 27306 ((("%") shift) ((")") reduce 420) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") reduce 420) (("/") shift) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("{") reduce 420) (("|") reduce 420)) 27307 ((("%") shift) ((")") reduce 422) (("*") shift) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("{") reduce 422) (("|") reduce 422)) 27308 ((("%") shift) ((")") reduce 421) (("*") shift) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("{") reduce 421) (("|") reduce 421)) 27309 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("{") reduce 423) (("|") reduce 423)) 27310 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("{") reduce 426) (("|") reduce 426)) 27312 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 27313 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 27314 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 27315 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 27316 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 27317 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 27318 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 27319 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 27320 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 27321 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 27322 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 27323 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 27324 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 27325 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 27326 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 27327 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 27328 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 27329 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 27330 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 27331 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 27332 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 27333 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 27334 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 27335 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 27336 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 27337 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 27338 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 27339 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 27340 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 27341 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 27342 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 27343 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 27344 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 27345 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 27346 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 27347 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 27348 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 27349 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 27350 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 27351 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 27352 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 27353 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 27354 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 27355 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 27356 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 27357 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 27358 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 27359 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 27360 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 27361 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 27362 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 27363 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 27364 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 27365 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 27366 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 27367 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 27368 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 27369 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 27370 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 27371 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 27372 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 27373 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 27374 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 27375 ((("%") reduce 79) (("(") reduce 79) ((")") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("{") reduce 79) (("|") reduce 79)) 27376 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 27377 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 27378 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 27379 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 27380 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 27381 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 27382 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 27383 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 27384 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 27385 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 27386 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 27387 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 27388 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 27389 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 27390 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 27391 ((("%") reduce 525) (("(") reduce 562) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 27393 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 27394 ((("%") reduce 83) (("(") reduce 83) ((")") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("{") reduce 83) (("|") reduce 83)) 27395 ((("%") reduce 555) (("(") reduce 564) ((")") reduce 555) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("{") reduce 555) (("|") reduce 555)) 27398 ((("%") reduce 527) (("(") reduce 563) ((")") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("{") reduce 527) (("|") reduce 527)) 27400 ((("%") reduce 490) (("(") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 27401 ((("%") reduce 289) (("(") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("{") reduce 289) (("|") reduce 289)) 27405 ((("%") reduce 290) (("(") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("{") reduce 290) (("|") reduce 290)) 27406 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 27411 ((("%") reduce 409) ((")") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("{") reduce 409) (("|") reduce 409)) 27413 ((("%") reduce 407) ((")") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("{") reduce 407) (("|") reduce 407)) 27423 ((("%") shift) ((")") reduce 403) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 403) (("|") shift)) 27424 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 405) (("|") shift)) 27425 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 404) (("|") shift)) 27426 ((("%") shift) ((")") reduce 397) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 397) (("|") shift)) 27427 ((("%") shift) ((")") reduce 394) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 394) (("|") shift)) 27428 ((("%") shift) ((")") reduce 402) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 402) (("|") shift)) 27429 ((("%") shift) ((")") reduce 396) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 396) (("|") shift)) 27430 ((("%") shift) ((")") reduce 400) (("*") shift) (("+") shift) ((",") reduce 400) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 400) (("|") shift)) 27431 ((("%") shift) ((")") reduce 395) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 395) (("|") shift)) 27432 ((("%") shift) ((")") reduce 398) (("*") shift) (("+") shift) ((",") reduce 398) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 398) (("|") shift)) 27433 ((("%") shift) ((")") reduce 399) (("*") shift) (("+") shift) ((",") reduce 399) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 399) (("|") shift)) 27434 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 406) (("|") shift)) 27435 ((("%") shift) ((")") reduce 401) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 401) (("|") shift)) 27437 ((("%") shift) ((")") reduce 391) (("*") shift) (("+") shift) ((",") reduce 391) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 391) (("|") shift)) 27440 ((("%") reduce 551) (("(") reduce 551) ((")") reduce 551) (("*") reduce 551) (("+") reduce 551) ((",") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("{") reduce 551) (("|") reduce 551)) 27443 ((("%") reduce 547) (("(") reduce 538) ((")") reduce 547) (("*") reduce 547) (("+") reduce 547) ((",") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("{") reduce 547) (("|") reduce 547)) 27445 ((("%") reduce 546) (("(") reduce 546) ((")") reduce 546) (("*") reduce 546) (("+") reduce 546) ((",") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("{") reduce 546) (("|") reduce 546)) 27447 ((("%") reduce 542) (("(") reduce 542) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("{") reduce 542) (("|") reduce 542)) 27448 ((("%") reduce 548) ((")") reduce 548) (("*") reduce 548) (("+") reduce 548) ((",") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("{") reduce 548) (("|") reduce 548)) 27449 ((("%") reduce 392) ((")") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("{") reduce 392) (("|") reduce 392)) 27453 ((("%") reduce 555) (("(") reduce 564) ((")") reduce 555) (("*") reduce 555) (("+") reduce 555) ((",") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("{") reduce 555) (("|") reduce 555)) 27455 ((("%") reduce 490) (("(") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 27456 ((("%") reduce 289) (("(") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("{") reduce 289) (("|") reduce 289)) 27460 ((("%") reduce 290) (("(") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("{") reduce 290) (("|") reduce 290)) 27461 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 27462 ((("%") reduce 487) (("(") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 27465 ((("%") reduce 554) (("(") reduce 564) ((")") reduce 554) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("{") reduce 554) (("|") reduce 554)) 27466 ((("%") reduce 489) (("(") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 27467 ((("%") reduce 491) (("(") reduce 491) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 27472 ((("%") reduce 543) (("(") reduce 543) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("{") reduce 543) (("|") reduce 543)) 27473 ((("%") reduce 565) (("(") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 27474 ((("%") reduce 549) (("(") shift) ((")") reduce 549) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("{") reduce 549) (("|") reduce 549)) 27475 ((("%") reduce 567) (("(") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 27478 ((("%") reduce 566) (("(") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 27479 ((("%") reduce 544) (("(") reduce 544) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("{") reduce 544) (("|") reduce 544)) 27480 ((("%") reduce 550) (("(") shift) ((")") reduce 550) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("{") reduce 550) (("|") reduce 550)) 27481 ((("%") reduce 545) (("(") reduce 545) ((")") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("{") reduce 545) (("|") reduce 545)) 27482 ((("%") reduce 488) (("(") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 27483 ((("%") reduce 553) (("(") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("{") reduce 553) (("|") reduce 553)) 27486 ((("%") reduce 552) (("(") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("{") reduce 552) (("|") reduce 552)) 27487 ((("%") reduce 487) (("(") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 27498 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("{") reduce 477) (("|") reduce 477)) 27499 ((("%") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 27509 ((("%") shift) ((")") reduce 477) (("*") shift) (("+") shift) ((",") reduce 477) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 477) (("|") shift)) 27510 ((("%") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 27512 ((("%") shift) ((")") reduce 464) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 464) (("|") shift)) 27514 ((("%") reduce 524) (("(") reduce 562) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 27516 ((("%") reduce 554) (("(") reduce 564) ((")") reduce 554) (("*") reduce 554) (("+") reduce 554) ((",") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("{") reduce 554) (("|") reduce 554)) 27519 ((("%") reduce 526) (("(") reduce 563) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 27520 ((("%") reduce 489) (("(") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 27521 ((("%") reduce 491) (("(") reduce 491) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 27523 ((("%") reduce 468) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("{") reduce 468) (("|") reduce 468)) 27524 ((("%") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 27529 ((("%") reduce 543) (("(") reduce 543) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("{") reduce 543) (("|") reduce 543)) 27530 ((("%") reduce 565) (("(") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 27531 ((("%") reduce 549) (("(") shift) ((")") reduce 549) (("*") reduce 549) (("+") reduce 549) ((",") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("{") reduce 549) (("|") reduce 549)) 27532 ((("%") reduce 567) (("(") reduce 567) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 27535 ((("%") reduce 566) (("(") reduce 566) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 27536 ((("%") reduce 544) (("(") reduce 544) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("{") reduce 544) (("|") reduce 544)) 27537 ((("%") reduce 550) (("(") shift) ((")") reduce 550) (("*") reduce 550) (("+") reduce 550) ((",") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("{") reduce 550) (("|") reduce 550)) 27538 ((("%") reduce 545) (("(") reduce 545) ((")") reduce 545) (("*") reduce 545) (("+") reduce 545) ((",") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("{") reduce 545) (("|") reduce 545)) 27540 ((("%") reduce 460) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("{") reduce 460) (("|") reduce 460)) 27542 ((("%") reduce 503) (("(") reduce 503) ((")") reduce 503) (("*") reduce 503) (("+") reduce 503) ((",") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_DOUBLE_ARROW) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("{") reduce 503) (("|") reduce 503)) 27544 ((("%") shift) ((")") reduce 390) (("*") shift) (("+") shift) ((",") reduce 390) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 390) (("|") shift)) 27545 ((("%") shift) ((")") reduce 465) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 465) (("|") shift)) 27548 ((("%") shift) ((")") reduce 464) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 464) (("|") shift)) 27549 ((("%") reduce 456) ((")") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 27550 ((("%") shift) ((")") reduce 466) (("*") shift) (("+") shift) ((",") reduce 466) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 466) (("|") shift)) 27551 ((("%") reduce 452) ((")") reduce 452) (("*") reduce 452) (("+") reduce 452) ((",") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("{") reduce 452) (("|") reduce 452)) 27552 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("{") reduce 469) (("|") reduce 469)) 27554 ((("%") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 27556 ((("%") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 27557 ((("%") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 27558 ((("%") shift) ((")") reduce 602) (("*") shift) (("+") shift) ((",") reduce 602) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 602) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 602) (("|") shift)) 27559 ((("%") shift) ((")") reduce 601) (("*") shift) (("+") shift) ((",") reduce 601) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 601) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 601) (("|") shift)) 27560 ((("%") reduce 488) (("(") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 27561 ((("%") shift) ((")") reduce 461) (("*") shift) (("+") shift) ((",") reduce 461) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 461) (("|") shift)) 27562 ((("%") reduce 454) ((")") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 27564 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 27565 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 27566 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 27567 ((("%") reduce 492) (("(") reduce 492) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 27568 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 27569 ((("%") reduce 385) (("(") reduce 385) ((")") reduce 385) (("*") reduce 385) (("+") reduce 385) ((",") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("{") reduce 385) (("|") reduce 385)) 27572 ((("%") reduce 494) (("(") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("{") reduce 494) (("|") reduce 494)) 27573 ((("%") reduce 387) (("(") shift) ((")") reduce 387) (("*") reduce 387) (("+") reduce 387) ((",") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("{") reduce 387) (("|") reduce 387)) 27574 ((("%") reduce 493) (("(") reduce 493) ((")") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("{") reduce 493) (("|") reduce 493)) 27575 ((("%") reduce 495) (("(") reduce 495) ((")") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("{") reduce 495) (("|") reduce 495)) 27576 ((("%") reduce 556) (("(") reduce 556) ((")") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("{") reduce 556) (("|") reduce 556)) 27582 ((("%") reduce 557) (("(") reduce 557) ((")") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("{") reduce 557) (("|") reduce 557)) 27583 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 27584 ((("%") reduce 558) (("(") reduce 558) ((")") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("{") reduce 558) (("|") reduce 558)) 27585 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 27586 ((("%") reduce 384) (("(") reduce 384) ((")") reduce 384) (("*") reduce 384) (("+") reduce 384) ((",") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("{") reduce 384) (("|") reduce 384)) 27588 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 27589 ((("%") reduce 386) (("(") reduce 386) ((")") reduce 386) (("*") reduce 386) (("+") reduce 386) ((",") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("{") reduce 386) (("|") reduce 386)) 27597 ((("%") reduce 383) (("(") reduce 383) ((")") reduce 383) (("*") reduce 383) (("+") reduce 383) ((",") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("{") reduce 383) (("|") reduce 383)) 27599 ((("%") reduce 496) (("(") reduce 496) ((")") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("{") reduce 496) (("|") reduce 496)) 27605 ((("%") reduce 231) ((")") reduce 231) (("*") reduce 231) (("+") reduce 231) ((",") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_DOUBLE_ARROW) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("{") reduce 231) (("|") reduce 231)) 27610 ((("%") shift) ((")") reduce 389) (("*") shift) (("+") shift) ((",") reduce 389) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 389) (("|") shift)) 27614 ((("%") reduce 596) ((")") reduce 596) (("*") reduce 596) (("+") reduce 596) ((",") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_DOUBLE_ARROW) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("{") reduce 596) (("|") reduce 596)) 27615 ((("%") reduce 450) ((")") reduce 450) (("*") reduce 450) (("+") reduce 450) ((",") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("{") reduce 450) (("|") reduce 450)) 27616 ((("%") shift) ((")") reduce 599) (("*") shift) (("+") shift) ((",") reduce 599) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 599) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 599) (("|") shift)) 27617 ((("%") shift) ((")") reduce 598) (("*") shift) (("+") shift) ((",") reduce 598) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 598) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 598) (("|") shift)) 27618 ((("%") reduce 408) ((")") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("{") reduce 408) (("|") reduce 408)) 27620 ((("%") reduce 501) ((")") reduce 501) (("*") reduce 501) (("+") reduce 501) ((",") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_DOUBLE_ARROW) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("{") reduce 501) (("|") reduce 501)) 27621 ((("%") reduce 457) ((")") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 27622 ((("%") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("{") reduce 289) (("|") reduce 289)) 27626 ((("%") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("{") reduce 290) (("|") reduce 290)) 27627 ((("%") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 27630 ((("%") reduce 600) ((")") reduce 600) (("*") reduce 600) (("+") reduce 600) ((",") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_DOUBLE_ARROW) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("{") reduce 600) (("|") reduce 600)) 27633 ((("%") reduce 597) ((")") reduce 597) (("*") reduce 597) (("+") reduce 597) ((",") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_DOUBLE_ARROW) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("{") reduce 597) (("|") reduce 597)) 27634 ((("%") reduce 451) ((")") reduce 451) (("*") reduce 451) (("+") reduce 451) ((",") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("{") reduce 451) (("|") reduce 451)) 27635 ((("%") reduce 410) ((")") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("{") reduce 410) (("|") reduce 410)) 27636 ((("%") reduce 393) ((")") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("{") reduce 393) (("|") reduce 393)) 27637 ((("%") reduce 455) ((")") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 27638 ((("%") reduce 453) ((")") reduce 453) (("*") reduce 453) (("+") reduce 453) ((",") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("{") reduce 453) (("|") reduce 453)) 27641 ((("%") reduce 502) (("(") reduce 502) ((")") reduce 502) (("*") reduce 502) (("+") reduce 502) ((",") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_DOUBLE_ARROW) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("{") reduce 502) (("|") reduce 502)) 27642 ((("%") reduce 458) ((")") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("{") reduce 458) (("|") reduce 458)) 27643 ((("%") reduce 430) ((")") reduce 430) (("*") reduce 430) (("+") reduce 430) ((",") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("{") reduce 430) (("|") reduce 430)) 27644 ((("%") reduce 429) ((")") reduce 429) (("*") reduce 429) (("+") reduce 429) ((",") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("{") reduce 429) (("|") reduce 429)) 27646 ((("%") reduce 443) (("(") reduce 539) ((")") reduce 443) (("*") reduce 443) (("+") reduce 443) ((",") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("{") reduce 443) (("|") reduce 443)) 27647 ((("%") reduce 553) (("(") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("{") reduce 553) (("|") reduce 553)) 27650 ((("%") reduce 552) (("(") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("{") reduce 552) (("|") reduce 552)) 27652 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 27653 ((("%") reduce 431) ((")") reduce 431) (("*") reduce 431) (("+") reduce 431) ((",") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("{") reduce 431) (("|") reduce 431)) 27654 ((("%") reduce 456) ((")") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 27655 ((("%") shift) ((")") reduce 466) (("*") shift) (("+") shift) ((",") reduce 466) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 466) (("|") shift)) 27656 ((("%") reduce 452) ((")") reduce 452) (("*") reduce 452) (("+") reduce 452) ((",") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("{") reduce 452) (("|") reduce 452)) 27657 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("{") reduce 469) (("|") reduce 469)) 27659 ((("%") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 27661 ((("%") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 27662 ((("%") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 27663 ((("%") shift) ((")") reduce 602) (("*") shift) (("+") shift) ((",") reduce 602) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 602) (("|") shift)) 27664 ((("%") shift) ((")") reduce 601) (("*") shift) (("+") shift) ((",") reduce 601) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 601) (("|") shift)) 27665 ((("%") reduce 488) (("(") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 27666 ((("%") shift) ((")") reduce 461) (("*") shift) (("+") shift) ((",") reduce 461) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 461) (("|") shift)) 27667 ((("%") reduce 454) ((")") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 27669 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 27670 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 27671 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 27672 ((("%") reduce 492) (("(") reduce 492) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 27673 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 27674 ((("%") reduce 385) (("(") reduce 385) ((")") reduce 385) (("*") reduce 385) (("+") reduce 385) ((",") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("{") reduce 385) (("|") reduce 385)) 27677 ((("%") reduce 494) (("(") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("{") reduce 494) (("|") reduce 494)) 27678 ((("%") reduce 387) (("(") shift) ((")") reduce 387) (("*") reduce 387) (("+") reduce 387) ((",") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("{") reduce 387) (("|") reduce 387)) 27679 ((("%") reduce 493) (("(") reduce 493) ((")") reduce 493) (("*") reduce 493) (("+") reduce 493) ((",") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("{") reduce 493) (("|") reduce 493)) 27680 ((("%") reduce 495) (("(") reduce 495) ((")") reduce 495) (("*") reduce 495) (("+") reduce 495) ((",") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("{") reduce 495) (("|") reduce 495)) 27681 ((("%") reduce 556) (("(") reduce 556) ((")") reduce 556) (("*") reduce 556) (("+") reduce 556) ((",") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("{") reduce 556) (("|") reduce 556)) 27687 ((("%") reduce 557) (("(") reduce 557) ((")") reduce 557) (("*") reduce 557) (("+") reduce 557) ((",") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("{") reduce 557) (("|") reduce 557)) 27688 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 27689 ((("%") reduce 558) (("(") reduce 558) ((")") reduce 558) (("*") reduce 558) (("+") reduce 558) ((",") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("{") reduce 558) (("|") reduce 558)) 27690 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 27691 ((("%") reduce 384) (("(") reduce 384) ((")") reduce 384) (("*") reduce 384) (("+") reduce 384) ((",") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("{") reduce 384) (("|") reduce 384)) 27693 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 27694 ((("%") reduce 386) (("(") reduce 386) ((")") reduce 386) (("*") reduce 386) (("+") reduce 386) ((",") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("{") reduce 386) (("|") reduce 386)) 27702 ((("%") reduce 383) (("(") reduce 383) ((")") reduce 383) (("*") reduce 383) (("+") reduce 383) ((",") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("{") reduce 383) (("|") reduce 383)) 27704 ((("%") reduce 496) (("(") reduce 496) ((")") reduce 496) (("*") reduce 496) (("+") reduce 496) ((",") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("{") reduce 496) (("|") reduce 496)) 27710 ((("%") reduce 231) ((")") reduce 231) (("*") reduce 231) (("+") reduce 231) ((",") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("{") reduce 231) (("|") reduce 231)) 27715 ((("%") shift) ((")") reduce 389) (("*") shift) (("+") shift) ((",") reduce 389) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 389) (("|") shift)) 27719 ((("%") reduce 596) ((")") reduce 596) (("*") reduce 596) (("+") reduce 596) ((",") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("{") reduce 596) (("|") reduce 596)) 27720 ((("%") reduce 450) ((")") reduce 450) (("*") reduce 450) (("+") reduce 450) ((",") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("{") reduce 450) (("|") reduce 450)) 27721 ((("%") shift) ((")") reduce 599) (("*") shift) (("+") shift) ((",") reduce 599) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 599) (("|") shift)) 27722 ((("%") shift) ((")") reduce 598) (("*") shift) (("+") shift) ((",") reduce 598) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 598) (("|") shift)) 27723 ((("%") reduce 408) ((")") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("{") reduce 408) (("|") reduce 408)) 27725 ((("%") reduce 501) ((")") reduce 501) (("*") reduce 501) (("+") reduce 501) ((",") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("{") reduce 501) (("|") reduce 501)) 27726 ((("%") reduce 457) ((")") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 27727 ((("%") reduce 289) ((")") reduce 289) (("*") reduce 289) (("+") reduce 289) ((",") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("{") reduce 289) (("|") reduce 289)) 27731 ((("%") reduce 290) ((")") reduce 290) (("*") reduce 290) (("+") reduce 290) ((",") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("{") reduce 290) (("|") reduce 290)) 27732 ((("%") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 27735 ((("%") reduce 600) ((")") reduce 600) (("*") reduce 600) (("+") reduce 600) ((",") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("{") reduce 600) (("|") reduce 600)) 27738 ((("%") reduce 597) ((")") reduce 597) (("*") reduce 597) (("+") reduce 597) ((",") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("{") reduce 597) (("|") reduce 597)) 27739 ((("%") reduce 451) ((")") reduce 451) (("*") reduce 451) (("+") reduce 451) ((",") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("{") reduce 451) (("|") reduce 451)) 27740 ((("%") reduce 410) ((")") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("{") reduce 410) (("|") reduce 410)) 27741 ((("%") reduce 393) ((")") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("{") reduce 393) (("|") reduce 393)) 27742 ((("%") reduce 455) ((")") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 27743 ((("%") reduce 453) ((")") reduce 453) (("*") reduce 453) (("+") reduce 453) ((",") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("{") reduce 453) (("|") reduce 453)) 27746 ((("%") reduce 502) (("(") reduce 502) ((")") reduce 502) (("*") reduce 502) (("+") reduce 502) ((",") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("{") reduce 502) (("|") reduce 502)) 27747 ((("%") reduce 458) ((")") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("{") reduce 458) (("|") reduce 458)) 27748 ((("%") reduce 430) ((")") reduce 430) (("*") reduce 430) (("+") reduce 430) ((",") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("{") reduce 430) (("|") reduce 430)) 27749 ((("%") reduce 429) ((")") reduce 429) (("*") reduce 429) (("+") reduce 429) ((",") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("{") reduce 429) (("|") reduce 429)) 27751 ((("%") reduce 443) (("(") reduce 539) ((")") reduce 443) (("*") reduce 443) (("+") reduce 443) ((",") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("{") reduce 443) (("|") reduce 443)) 27752 ((("%") reduce 553) (("(") reduce 553) ((")") reduce 553) (("*") reduce 553) (("+") reduce 553) ((",") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("{") reduce 553) (("|") reduce 553)) 27755 ((("%") reduce 552) (("(") reduce 552) ((")") reduce 552) (("*") reduce 552) (("+") reduce 552) ((",") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("{") reduce 552) (("|") reduce 552)) 27757 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 27758 ((("%") reduce 431) ((")") reduce 431) (("*") reduce 431) (("+") reduce 431) ((",") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("{") reduce 431) (("|") reduce 431)) 27760 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 286) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 286) ((T_ELLIPSIS) reduce 286) ((T_VARIABLE) reduce 286)) 27761 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_ELLIPSIS) reduce 278) ((T_VARIABLE) reduce 278)) 27762 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) ((T_ELLIPSIS) reduce 279) ((T_VARIABLE) reduce 279)) 27763 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ELLIPSIS) reduce 94) ((T_VARIABLE) reduce 94)) 27764 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ELLIPSIS) reduce 93) ((T_VARIABLE) reduce 93)) 27765 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_ELLIPSIS) reduce 95) ((T_VARIABLE) reduce 95)) 27766 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ELLIPSIS) reduce 92) ((T_VARIABLE) reduce 92)) 27767 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_ELLIPSIS) reduce 280) ((T_VARIABLE) reduce 280)) 27768 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 275) ((T_ELLIPSIS) reduce 275) ((T_VARIABLE) reduce 275)) 27769 ((("(") reduce 256) ((")") reduce 125) (("?") reduce 256) ((T_ABSTRACT) shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 256) ((T_ARRAY) reduce 256) ((T_ATTRIBUTE) shift) ((T_CALLABLE) reduce 256) ((T_ELLIPSIS) reduce 256) ((T_FINAL) shift) ((T_NAME_FULLY_QUALIFIED) reduce 256) ((T_NAME_QUALIFIED) reduce 256) ((T_NAME_RELATIVE) reduce 256) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 256) ((T_VARIABLE) reduce 256)) 27770 (((")") reduce 250)) 27771 (((")") reduce 253) ((",") reduce 253)) 27772 ((("(") reduce 341) (("?") reduce 341) ((T_ABSTRACT) reduce 341) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 341) ((T_ARRAY) reduce 341) ((T_CALLABLE) reduce 341) ((T_ELLIPSIS) reduce 341) ((T_FINAL) reduce 341) ((T_NAME_FULLY_QUALIFIED) reduce 341) ((T_NAME_QUALIFIED) reduce 341) ((T_NAME_RELATIVE) reduce 341) ((T_PRIVATE) reduce 341) ((T_PRIVATE_SET) reduce 341) ((T_PROTECTED) reduce 341) ((T_PROTECTED_SET) reduce 341) ((T_PUBLIC) reduce 341) ((T_PUBLIC_SET) reduce 341) ((T_READONLY) reduce 341) ((T_STATIC) reduce 341) ((T_STRING) reduce 341) ((T_VARIABLE) reduce 341)) 27773 ((("(") reduce 102) (("?") reduce 102) ((T_ABSTRACT) reduce 102) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 102) ((T_ARRAY) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_CALLABLE) reduce 102) ((T_ELLIPSIS) reduce 102) ((T_FINAL) reduce 102) ((T_NAME_FULLY_QUALIFIED) reduce 102) ((T_NAME_QUALIFIED) reduce 102) ((T_NAME_RELATIVE) reduce 102) ((T_PRIVATE) reduce 102) ((T_PRIVATE_SET) reduce 102) ((T_PROTECTED) reduce 102) ((T_PROTECTED_SET) reduce 102) ((T_PUBLIC) reduce 102) ((T_PUBLIC_SET) reduce 102) ((T_READONLY) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102) ((T_VARIABLE) reduce 102)) 27774 (((")") reduce 254) ((",") reduce 254)) 27777 ((("(") reduce 100) (("?") reduce 100) ((T_ABSTRACT) reduce 100) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 100) ((T_ARRAY) reduce 100) ((T_ATTRIBUTE) reduce 100) ((T_CALLABLE) reduce 100) ((T_ELLIPSIS) reduce 100) ((T_FINAL) reduce 100) ((T_NAME_FULLY_QUALIFIED) reduce 100) ((T_NAME_QUALIFIED) reduce 100) ((T_NAME_RELATIVE) reduce 100) ((T_PRIVATE) reduce 100) ((T_PRIVATE_SET) reduce 100) ((T_PROTECTED) reduce 100) ((T_PROTECTED_SET) reduce 100) ((T_PUBLIC) reduce 100) ((T_PUBLIC_SET) reduce 100) ((T_READONLY) reduce 100) ((T_STATIC) reduce 100) ((T_STRING) reduce 100) ((T_VARIABLE) reduce 100)) 27779 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 27780 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 27784 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 27785 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 27788 (((T_ATTRIBUTE) reduce 100) ((T_FN) reduce 100) ((T_FUNCTION) reduce 100) ((T_STATIC) reduce 100)) 27790 ((("!") reduce 146) (("\"") reduce 146) (($) reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONST) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAMESPACE) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_USE) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 27792 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 27793 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 27794 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 27795 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 27796 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 27797 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 27798 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 27799 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 27800 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 27801 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 27802 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 27803 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 27804 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 27805 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 27806 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 27807 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 27808 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 27809 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 27810 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 27811 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 27812 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 27813 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 27814 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 27815 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 27816 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 27817 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 27818 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 27819 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 27820 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 27821 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 27822 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 27823 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 27824 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 27825 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 27826 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 27827 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 27828 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 27829 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 27830 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 27831 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 27832 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 27833 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 27834 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 27835 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 27836 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 27837 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 27838 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 27839 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 27840 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 27841 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 27842 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 27843 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 27844 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 27845 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 27846 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 27847 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 27848 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 27849 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 27850 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 27851 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 27852 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 27853 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 27854 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 27855 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 27856 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 27857 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 27858 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 27859 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 27860 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 27861 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 27862 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 27863 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 27864 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 27865 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 27866 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 27867 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 27868 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 27869 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 27870 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 27871 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525)) 27873 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 27874 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 27875 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((";") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 27878 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527)) 27879 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 27884 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 27886 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407)) 27896 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 403) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 27897 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 27898 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 27899 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 397) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 27900 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 394) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 27901 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 402) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 27902 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 396) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 27903 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 400) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 27904 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 395) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 27905 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 398) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 27906 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 399) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 27907 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 27908 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 401) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 27910 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 391) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 27913 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 27916 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((";") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 27918 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((";") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 27920 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 27921 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((";") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 27922 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((";") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392)) 27926 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((";") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 27928 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 27929 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 27933 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 27934 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 27935 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 27938 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((";") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 27939 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 27940 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 27945 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 27946 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 27947 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((";") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 27948 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 27951 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 27952 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 27953 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((";") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 27954 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((";") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 27955 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 27956 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 27959 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 27960 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 27966 ((("!") reduce 165) (("\"") reduce 165) (($) reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONST) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAMESPACE) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_USE) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 27976 ((("!") reduce 159) (("\"") reduce 159) (($) reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONST) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAMESPACE) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_USE) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 27984 ((("!") reduce 148) (("\"") reduce 148) (($) reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONST) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAMESPACE) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_USE) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 27987 ((("!") reduce 147) (("\"") reduce 147) (($) reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONST) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAMESPACE) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_USE) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 27989 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (($) reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONST) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAMESPACE) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_USE) reduce 243) ((T_VARIABLE) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 27992 ((("!") reduce 146) (("\"") reduce 146) (($) reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONST) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAMESPACE) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_USE) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 27994 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (($) reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONST) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAMESPACE) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_USE) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 27995 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (($) reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONST) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAMESPACE) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_USE) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 27998 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (($) reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONST) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_ELSE) reduce 248) ((T_ELSEIF) reduce 248) ((T_EMPTY) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAMESPACE) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_USE) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("~") reduce 248)) 28002 ((("!") reduce 249) (("\"") reduce 249) (($) reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONST) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_ELSE) reduce 249) ((T_ELSEIF) reduce 249) ((T_EMPTY) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAMESPACE) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_USE) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("~") reduce 249)) 28007 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (($) reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONST) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAMESPACE) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_USE) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 28008 ((("!") reduce 149) (("\"") reduce 149) (($) reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONST) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAMESPACE) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_USE) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 28011 ((("!") reduce 241) (("\"") reduce 241) (($) reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONST) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAMESPACE) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_USE) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 28016 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (($) reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONST) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAMESPACE) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_USE) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 28019 ((("!") reduce 169) (("\"") reduce 169) (($) reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONST) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAMESPACE) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_USE) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 28020 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (($) reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONST) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAMESPACE) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_USE) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 28023 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (($) reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONST) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAMESPACE) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_USE) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 28026 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (($) reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONST) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_ELSE) reduce 176) ((T_ELSEIF) reduce 176) ((T_EMPTY) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAMESPACE) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_USE) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("~") reduce 176)) 28033 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (($) reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONST) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAMESPACE) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_USE) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 28038 ((("!") reduce 152) (("\"") reduce 152) (($) reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONST) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAMESPACE) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_USE) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 28042 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (($) reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONST) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAMESPACE) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_USE) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 28044 ((("!") reduce 223) (("\"") reduce 223) (($) reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONST) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAMESPACE) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_USE) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 28048 ((("!") reduce 224) (("\"") reduce 224) (($) reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONST) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAMESPACE) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_USE) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 28051 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (($) reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONST) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAMESPACE) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_USE) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 28052 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (($) reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONST) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAMESPACE) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_USE) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 28054 ((("!") reduce 157) (("\"") reduce 157) (($) reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONST) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAMESPACE) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_USE) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 28056 ((("!") reduce 155) (("\"") reduce 155) (($) reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONST) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAMESPACE) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_USE) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 28058 ((("!") reduce 167) (("\"") reduce 167) (($) reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONST) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAMESPACE) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_USE) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 28060 ((("!") reduce 156) (("\"") reduce 156) (($) reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONST) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAMESPACE) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_USE) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 28070 ((("!") reduce 163) (("\"") reduce 163) (($) reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONST) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAMESPACE) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_USE) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 28071 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (($) reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONST) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAMESPACE) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_USE) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 28074 ((("!") reduce 219) (("\"") reduce 219) (($) reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONST) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAMESPACE) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_USE) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 28075 ((("!") reduce 162) (("\"") reduce 162) (($) reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONST) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAMESPACE) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_USE) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 28084 ((("!") reduce 151) (("\"") reduce 151) (($) reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONST) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAMESPACE) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_USE) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 28085 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (($) reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONST) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAMESPACE) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_USE) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 28088 ((("!") reduce 217) (("\"") reduce 217) (($) reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONST) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAMESPACE) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_USE) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 28090 ((("!") reduce 158) (("\"") reduce 158) (($) reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONST) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAMESPACE) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_USE) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 28096 ((("!") reduce 150) (("\"") reduce 150) (($) reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONST) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAMESPACE) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_USE) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 28101 ((("!") reduce 164) (("\"") reduce 164) (($) reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONST) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAMESPACE) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_USE) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 28102 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (($) reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONST) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAMESPACE) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_USE) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 28105 ((("!") reduce 221) (("\"") reduce 221) (($) reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONST) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAMESPACE) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_USE) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 28107 ((("!") reduce 154) (("\"") reduce 154) (($) reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONST) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAMESPACE) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_USE) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 28109 ((("!") reduce 153) (("\"") reduce 153) (($) reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONST) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAMESPACE) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_USE) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 28110 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (($) reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONST) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_EMPTY) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAMESPACE) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_USE) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 28121 ((("!") reduce 477) (("\"") reduce 477) (("$") reduce 477) (($) reduce 477) (("(") reduce 477) (("+") reduce 477) (("-") reduce 477) ((";") reduce 477) (("@") reduce 477) ((T_ABSTRACT) reduce 477) ((T_ARRAY) reduce 477) ((T_ARRAY_CAST) reduce 477) ((T_ATTRIBUTE) reduce 477) ((T_BOOL_CAST) reduce 477) ((T_BREAK) reduce 477) ((T_CLASS) reduce 477) ((T_CLASS_C) reduce 477) ((T_CLONE) reduce 477) ((T_CONST) reduce 477) ((T_CONSTANT_ENCAPSED_STRING) reduce 477) ((T_CONTINUE) reduce 477) ((T_DEC) reduce 477) ((T_DECLARE) reduce 477) ((T_DIR) reduce 477) ((T_DNUMBER) reduce 477) ((T_DO) reduce 477) ((T_DOUBLE_CAST) reduce 477) ((T_ECHO) reduce 477) ((T_EMPTY) reduce 477) ((T_ENUM) reduce 477) ((T_EVAL) reduce 477) ((T_EXIT) reduce 477) ((T_FILE) reduce 477) ((T_FINAL) reduce 477) ((T_FN) reduce 477) ((T_FOR) reduce 477) ((T_FOREACH) reduce 477) ((T_FUNCTION) reduce 477) ((T_FUNC_C) reduce 477) ((T_GLOBAL) reduce 477) ((T_GOTO) reduce 477) ((T_HALT_COMPILER) reduce 477) ((T_IF) reduce 477) ((T_INC) reduce 477) ((T_INCLUDE) reduce 477) ((T_INCLUDE_ONCE) reduce 477) ((T_INLINE_HTML) reduce 477) ((T_INTERFACE) reduce 477) ((T_INT_CAST) reduce 477) ((T_ISSET) reduce 477) ((T_LINE) reduce 477) ((T_LIST) reduce 477) ((T_LNUMBER) reduce 477) ((T_MATCH) reduce 477) ((T_METHOD_C) reduce 477) ((T_NAMESPACE) reduce 477) ((T_NAME_FULLY_QUALIFIED) reduce 477) ((T_NAME_QUALIFIED) reduce 477) ((T_NAME_RELATIVE) reduce 477) ((T_NEW) reduce 477) ((T_NS_C) reduce 477) ((T_OBJECT_CAST) reduce 477) ((T_PRINT) reduce 477) ((T_PROPERTY_C) reduce 477) ((T_READONLY) reduce 477) ((T_REQUIRE) reduce 477) ((T_REQUIRE_ONCE) reduce 477) ((T_RETURN) reduce 477) ((T_START_HEREDOC) reduce 477) ((T_STATIC) reduce 477) ((T_STRING) reduce 477) ((T_STRING_CAST) reduce 477) ((T_SWITCH) reduce 477) ((T_THROW) reduce 477) ((T_TRAIT) reduce 477) ((T_TRAIT_C) reduce 477) ((T_TRY) reduce 477) ((T_UNSET) reduce 477) ((T_UNSET_CAST) reduce 477) ((T_USE) reduce 477) ((T_VARIABLE) reduce 477) ((T_WHILE) reduce 477) ((T_YIELD) reduce 477) ((T_YIELD_FROM) reduce 477) (("[") reduce 477) (("`") reduce 477) (("{") reduce 477) (("~") reduce 477)) 28122 ((("!") reduce 182) (("\"") reduce 182) (("$") reduce 182) (($) reduce 182) (("(") reduce 182) (("+") reduce 182) (("-") reduce 182) ((";") reduce 182) (("@") reduce 182) ((T_ABSTRACT) reduce 182) ((T_ARRAY) reduce 182) ((T_ARRAY_CAST) reduce 182) ((T_ATTRIBUTE) reduce 182) ((T_BOOL_CAST) reduce 182) ((T_BREAK) reduce 182) ((T_CLASS) reduce 182) ((T_CLASS_C) reduce 182) ((T_CLONE) reduce 182) ((T_CONST) reduce 182) ((T_CONSTANT_ENCAPSED_STRING) reduce 182) ((T_CONTINUE) reduce 182) ((T_DEC) reduce 182) ((T_DECLARE) reduce 182) ((T_DIR) reduce 182) ((T_DNUMBER) reduce 182) ((T_DO) reduce 182) ((T_DOUBLE_CAST) reduce 182) ((T_ECHO) reduce 182) ((T_EMPTY) reduce 182) ((T_ENUM) reduce 182) ((T_EVAL) reduce 182) ((T_EXIT) reduce 182) ((T_FILE) reduce 182) ((T_FINAL) reduce 182) ((T_FN) reduce 182) ((T_FOR) reduce 182) ((T_FOREACH) reduce 182) ((T_FUNCTION) reduce 182) ((T_FUNC_C) reduce 182) ((T_GLOBAL) reduce 182) ((T_GOTO) reduce 182) ((T_HALT_COMPILER) reduce 182) ((T_IF) reduce 182) ((T_INC) reduce 182) ((T_INCLUDE) reduce 182) ((T_INCLUDE_ONCE) reduce 182) ((T_INLINE_HTML) reduce 182) ((T_INTERFACE) reduce 182) ((T_INT_CAST) reduce 182) ((T_ISSET) reduce 182) ((T_LINE) reduce 182) ((T_LIST) reduce 182) ((T_LNUMBER) reduce 182) ((T_MATCH) reduce 182) ((T_METHOD_C) reduce 182) ((T_NAMESPACE) reduce 182) ((T_NAME_FULLY_QUALIFIED) reduce 182) ((T_NAME_QUALIFIED) reduce 182) ((T_NAME_RELATIVE) reduce 182) ((T_NEW) reduce 182) ((T_NS_C) reduce 182) ((T_OBJECT_CAST) reduce 182) ((T_PRINT) reduce 182) ((T_PROPERTY_C) reduce 182) ((T_READONLY) reduce 182) ((T_REQUIRE) reduce 182) ((T_REQUIRE_ONCE) reduce 182) ((T_RETURN) reduce 182) ((T_START_HEREDOC) reduce 182) ((T_STATIC) reduce 182) ((T_STRING) reduce 182) ((T_STRING_CAST) reduce 182) ((T_SWITCH) reduce 182) ((T_THROW) reduce 182) ((T_TRAIT) reduce 182) ((T_TRAIT_C) reduce 182) ((T_TRY) reduce 182) ((T_UNSET) reduce 182) ((T_UNSET_CAST) reduce 182) ((T_USE) reduce 182) ((T_VARIABLE) reduce 182) ((T_WHILE) reduce 182) ((T_YIELD) reduce 182) ((T_YIELD_FROM) reduce 182) (("[") reduce 182) (("`") reduce 182) (("{") reduce 182) (("~") reduce 182)) 28195 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 477) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28509 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 28512 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524)) 28514 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((";") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 28517 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526)) 28518 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 28519 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 28527 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 28528 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 28529 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((";") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 28530 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 28533 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 28534 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 28535 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((";") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 28536 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((";") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 28538 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 28540 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) ((";") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503)) 28542 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 390) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28543 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 465) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28554 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("|") reduce 523)) 28556 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 28558 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("|") reduce 517)) 28559 ((("%") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("^") reduce 507) (("|") reduce 507)) 28563 ((("%") reduce 500) (("(") shift) (("*") reduce 500) (("+") reduce 500) (("-") reduce 500) ((".") reduce 500) (("/") reduce 500) ((";") reduce 500) (("<") reduce 500) ((">") reduce 500) (("?") reduce 500) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 500) ((T_BOOLEAN_AND) reduce 500) ((T_BOOLEAN_OR) reduce 500) ((T_COALESCE) reduce 500) ((T_DOUBLE_ARROW) reduce 500) ((T_INSTANCEOF) reduce 500) ((T_IS_EQUAL) reduce 500) ((T_IS_GREATER_OR_EQUAL) reduce 500) ((T_IS_IDENTICAL) reduce 500) ((T_IS_NOT_EQUAL) reduce 500) ((T_IS_NOT_IDENTICAL) reduce 500) ((T_IS_SMALLER_OR_EQUAL) reduce 500) ((T_LOGICAL_AND) reduce 500) ((T_LOGICAL_OR) reduce 500) ((T_LOGICAL_XOR) reduce 500) ((T_POW) reduce 500) ((T_SL) reduce 500) ((T_SPACESHIP) reduce 500) ((T_SR) reduce 500) (("^") reduce 500) (("|") reduce 500)) 28564 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("|") reduce 516)) 28565 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 28571 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("|") reduce 515)) 28573 ((("%") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("^") reduce 506) (("|") reduce 506)) 28575 ((("%") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 28576 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 28577 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 28578 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 28580 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 28583 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 28589 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 28592 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("|") reduce 518)) 28594 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) (("=") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AND_EQUAL) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_COALESCE_EQUAL) reduce 551) ((T_CONCAT_EQUAL) reduce 551) ((T_DEC) reduce 551) ((T_DIV_EQUAL) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INC) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_MINUS_EQUAL) reduce 551) ((T_MOD_EQUAL) reduce 551) ((T_MUL_EQUAL) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_OR_EQUAL) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PLUS_EQUAL) reduce 551) ((T_POW) reduce 551) ((T_POW_EQUAL) reduce 551) ((T_SL) reduce 551) ((T_SL_EQUAL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) ((T_SR_EQUAL) reduce 551) ((T_XOR_EQUAL) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 28595 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 462) (("(") shift) (("*") reduce 462) (("+") shift) (("-") shift) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 462) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 462) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 462) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 462) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 462) (("`") shift) (("|") reduce 462) (("~") shift)) 28602 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((";") reduce 547) (("<") reduce 547) (("=") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AND_EQUAL) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_COALESCE_EQUAL) reduce 547) ((T_CONCAT_EQUAL) reduce 547) ((T_DEC) reduce 547) ((T_DIV_EQUAL) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INC) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_MINUS_EQUAL) reduce 547) ((T_MOD_EQUAL) reduce 547) ((T_MUL_EQUAL) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_OR_EQUAL) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_PLUS_EQUAL) reduce 547) ((T_POW) reduce 547) ((T_POW_EQUAL) reduce 547) ((T_SL) reduce 547) ((T_SL_EQUAL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) ((T_SR_EQUAL) reduce 547) ((T_XOR_EQUAL) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 28603 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 534) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 534) (("^") reduce 513) (("|") reduce 513)) 28605 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 537) (("^") reduce 512) (("|") reduce 512)) 28606 ((("%") reduce 511) (("(") reduce 540) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 533) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 533) (("^") reduce 511) (("|") reduce 511)) 28607 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 463) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28610 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((";") reduce 546) (("<") reduce 546) (("=") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AND_EQUAL) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_COALESCE_EQUAL) reduce 546) ((T_CONCAT_EQUAL) reduce 546) ((T_DEC) reduce 546) ((T_DIV_EQUAL) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INC) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_MINUS_EQUAL) reduce 546) ((T_MOD_EQUAL) reduce 546) ((T_MUL_EQUAL) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_OR_EQUAL) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_PLUS_EQUAL) reduce 546) ((T_POW) reduce 546) ((T_POW_EQUAL) reduce 546) ((T_SL) reduce 546) ((T_SL_EQUAL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) ((T_SR_EQUAL) reduce 546) ((T_XOR_EQUAL) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 28611 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_POW) reduce 467) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 28612 ((("%") reduce 449) (("*") reduce 449) (("+") reduce 449) (("-") reduce 449) ((".") reduce 449) (("/") reduce 449) ((";") reduce 449) (("<") reduce 449) ((">") reduce 449) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) reduce 449) ((T_IS_EQUAL) reduce 449) ((T_IS_GREATER_OR_EQUAL) reduce 449) ((T_IS_IDENTICAL) reduce 449) ((T_IS_NOT_EQUAL) reduce 449) ((T_IS_NOT_IDENTICAL) reduce 449) ((T_IS_SMALLER_OR_EQUAL) reduce 449) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_POW) reduce 449) ((T_SL) reduce 449) ((T_SPACESHIP) reduce 449) ((T_SR) reduce 449) (("^") reduce 449) (("|") reduce 449)) 28613 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 28614 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((";") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("|") reduce 514)) 28615 ((("%") reduce 444) (("(") reduce 541) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 535) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("[") reduce 535) (("^") reduce 444) (("|") reduce 444)) 28616 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_POW) reduce 445) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 28617 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 28618 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) (("=") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AND_EQUAL) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_COALESCE_EQUAL) reduce 542) ((T_CONCAT_EQUAL) reduce 542) ((T_DEC) reduce 542) ((T_DIV_EQUAL) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INC) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_MINUS_EQUAL) reduce 542) ((T_MOD_EQUAL) reduce 542) ((T_MUL_EQUAL) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_OR_EQUAL) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PLUS_EQUAL) reduce 542) ((T_POW) reduce 542) ((T_POW_EQUAL) reduce 542) ((T_SL) reduce 542) ((T_SL_EQUAL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) ((T_SR_EQUAL) reduce 542) ((T_XOR_EQUAL) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 28619 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((";") reduce 548) (("<") reduce 548) (("=") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AND_EQUAL) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_COALESCE_EQUAL) reduce 548) ((T_CONCAT_EQUAL) reduce 548) ((T_DEC) reduce 548) ((T_DIV_EQUAL) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INC) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_MINUS_EQUAL) reduce 548) ((T_MOD_EQUAL) reduce 548) ((T_MUL_EQUAL) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PLUS_EQUAL) reduce 548) ((T_POW) reduce 548) ((T_POW_EQUAL) reduce 548) ((T_SL) reduce 548) ((T_SL_EQUAL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) ((T_SR_EQUAL) reduce 548) ((T_XOR_EQUAL) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 28620 ((("%") reduce 388) (("*") reduce 388) (("+") reduce 388) (("-") reduce 388) ((".") reduce 388) (("/") reduce 388) ((";") reduce 388) (("<") reduce 388) (("=") shift) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 388) ((T_INC) shift) ((T_INSTANCEOF) reduce 388) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 388) ((T_POW_EQUAL) shift) ((T_SL) reduce 388) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) reduce 388) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 531) (("^") reduce 388) (("|") reduce 388)) 28623 ((("%") reduce 432) (("*") reduce 432) (("+") reduce 432) (("-") reduce 432) ((".") reduce 432) (("/") reduce 432) ((";") reduce 432) (("<") reduce 432) ((">") reduce 432) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) reduce 432) ((T_IS_EQUAL) reduce 432) ((T_IS_GREATER_OR_EQUAL) reduce 432) ((T_IS_IDENTICAL) reduce 432) ((T_IS_NOT_EQUAL) reduce 432) ((T_IS_NOT_IDENTICAL) reduce 432) ((T_IS_SMALLER_OR_EQUAL) reduce 432) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) reduce 432) ((T_SPACESHIP) reduce 432) ((T_SR) reduce 432) (("^") reduce 432) (("|") reduce 432)) 28654 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 416) (("<") shift) ((">") shift) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 416)) 28655 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 419) (("<") shift) ((">") shift) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 419) (("|") reduce 419)) 28656 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 428) (("/") shift) ((";") reduce 428) (("<") reduce 428) ((">") reduce 428) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 428) ((T_IS_GREATER_OR_EQUAL) reduce 428) ((T_IS_IDENTICAL) reduce 428) ((T_IS_NOT_EQUAL) reduce 428) ((T_IS_NOT_IDENTICAL) reduce 428) ((T_IS_SMALLER_OR_EQUAL) reduce 428) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) reduce 428) ((T_SPACESHIP) reduce 428) ((T_SR) reduce 428) (("^") reduce 428) (("|") reduce 428)) 28657 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 441) (("<") shift) ((">") shift) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 441) (("|") reduce 441)) 28658 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 427) (("/") shift) ((";") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) shift) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 28659 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) shift) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 28660 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28661 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28662 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28663 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 438) (("<") shift) ((">") shift) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 438) ((T_SR) shift) (("^") reduce 438) (("|") reduce 438)) 28664 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 28665 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 436) (("<") shift) ((">") shift) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 28666 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 28667 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 440) (("<") shift) ((">") shift) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 440) ((T_SR) shift) (("^") reduce 440) (("|") reduce 440)) 28668 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 28671 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 28672 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 28673 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 28674 ((("%") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 28675 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 28676 ((("%") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 28677 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 28678 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_POW) reduce 442) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 28679 ((("%") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 28680 ((("%") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 28681 ((("%") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 28687 ((("%") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 28688 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 28689 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 28690 ((("%") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 28691 ((("%") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 28694 ((("%") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 28695 ((("%") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 28697 ((("%") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 28699 ((("%") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 28700 ((("%") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 28703 ((("%") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 28704 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 448) (("<") shift) ((">") shift) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28705 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 411) (("<") shift) ((">") shift) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28706 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 412) (("<") shift) ((">") shift) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28707 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 417) (("<") shift) ((">") shift) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 417) (("|") reduce 417)) 28708 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 418) (("<") shift) ((">") shift) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 418) (("|") reduce 418)) 28712 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 446) (("<") shift) ((">") shift) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28713 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 447) (("<") shift) ((">") shift) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28714 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 439) (("<") shift) ((">") shift) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 439) ((T_SR) shift) (("^") reduce 439) (("|") reduce 439)) 28715 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 437) (("<") shift) ((">") shift) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 437) ((T_SR) shift) (("^") reduce 437) (("|") reduce 437)) 28716 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 28717 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 420) (("/") shift) ((";") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 420) ((T_SR) shift) (("^") reduce 420) (("|") reduce 420)) 28718 ((("%") shift) (("*") shift) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") shift) ((";") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 28719 ((("%") shift) (("*") shift) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") shift) ((";") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 28720 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 28721 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) shift) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 28723 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 28724 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 28725 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 28726 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 28727 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 28728 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 28729 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 28730 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 28731 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 28732 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 28733 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 28734 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 28735 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 28736 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 28737 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 28738 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 28739 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 28740 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 28741 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 28742 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 28743 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 28744 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 28745 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 28746 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 28747 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 28748 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 28749 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 28750 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 28751 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 28752 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 28753 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 28754 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 28755 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 28756 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 28757 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 28758 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 28759 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 28760 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 28761 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 28762 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 28763 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 28764 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 28765 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 28766 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 28767 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 28768 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 28769 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 28770 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 28771 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 28772 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 28773 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 28774 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 28775 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 28776 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 28777 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 28778 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 28779 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 28780 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 28781 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 28782 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 28783 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 28784 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 28785 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 28786 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 28787 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 28788 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 28789 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 28790 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 28791 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 28792 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 28793 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 28794 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 28795 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 28796 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 28797 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 28798 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 28799 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 28800 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 28801 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 28802 ((("%") reduce 525) (("(") reduce 562) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("|") reduce 525)) 28804 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 28805 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 28806 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((";") reduce 555) (("<") reduce 555) (("=") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AND_EQUAL) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_COALESCE_EQUAL) reduce 555) ((T_CONCAT_EQUAL) reduce 555) ((T_DEC) reduce 555) ((T_DIV_EQUAL) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INC) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_MINUS_EQUAL) reduce 555) ((T_MOD_EQUAL) reduce 555) ((T_MUL_EQUAL) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_OR_EQUAL) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_PLUS_EQUAL) reduce 555) ((T_POW) reduce 555) ((T_POW_EQUAL) reduce 555) ((T_SL) reduce 555) ((T_SL_EQUAL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) ((T_SR_EQUAL) reduce 555) ((T_XOR_EQUAL) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 28809 ((("%") reduce 527) (("(") reduce 563) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 527) ((T_OBJECT_OPERATOR) reduce 527) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("[") reduce 527) (("^") reduce 527) (("|") reduce 527)) 28811 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) (("=") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AND_EQUAL) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_COALESCE_EQUAL) reduce 490) ((T_CONCAT_EQUAL) reduce 490) ((T_DEC) reduce 490) ((T_DIV_EQUAL) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INC) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_MINUS_EQUAL) reduce 490) ((T_MOD_EQUAL) reduce 490) ((T_MUL_EQUAL) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_OR_EQUAL) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_PLUS_EQUAL) reduce 490) ((T_POW) reduce 490) ((T_POW_EQUAL) reduce 490) ((T_SL) reduce 490) ((T_SL_EQUAL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) ((T_SR_EQUAL) reduce 490) ((T_XOR_EQUAL) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 28812 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) (("=") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AND_EQUAL) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_COALESCE_EQUAL) reduce 289) ((T_CONCAT_EQUAL) reduce 289) ((T_DEC) reduce 289) ((T_DIV_EQUAL) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INC) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_MINUS_EQUAL) reduce 289) ((T_MOD_EQUAL) reduce 289) ((T_MUL_EQUAL) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_OR_EQUAL) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_PLUS_EQUAL) reduce 289) ((T_POW) reduce 289) ((T_POW_EQUAL) reduce 289) ((T_SL) reduce 289) ((T_SL_EQUAL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) ((T_SR_EQUAL) reduce 289) ((T_XOR_EQUAL) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 28816 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) (("=") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AND_EQUAL) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_COALESCE_EQUAL) reduce 290) ((T_CONCAT_EQUAL) reduce 290) ((T_DEC) reduce 290) ((T_DIV_EQUAL) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INC) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_MINUS_EQUAL) reduce 290) ((T_MOD_EQUAL) reduce 290) ((T_MUL_EQUAL) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_OR_EQUAL) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_PLUS_EQUAL) reduce 290) ((T_POW) reduce 290) ((T_POW_EQUAL) reduce 290) ((T_SL) reduce 290) ((T_SL_EQUAL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) ((T_SR_EQUAL) reduce 290) ((T_XOR_EQUAL) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 28817 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 28822 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 28824 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("^") reduce 407) (("|") reduce 407)) 28834 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 403) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28835 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28836 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28837 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 397) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28838 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 394) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28839 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 402) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28840 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 396) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28841 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 400) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28842 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 395) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28843 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 398) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28844 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 399) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28845 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28846 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 401) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28848 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 391) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28851 ((("%") reduce 551) (("(") reduce 551) (("*") reduce 551) (("+") reduce 551) (("-") reduce 551) ((".") reduce 551) (("/") reduce 551) ((";") reduce 551) (("<") reduce 551) ((">") reduce 551) (("?") reduce 551) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 551) ((T_BOOLEAN_AND) reduce 551) ((T_BOOLEAN_OR) reduce 551) ((T_COALESCE) reduce 551) ((T_DOUBLE_ARROW) reduce 551) ((T_INSTANCEOF) reduce 551) ((T_IS_EQUAL) reduce 551) ((T_IS_GREATER_OR_EQUAL) reduce 551) ((T_IS_IDENTICAL) reduce 551) ((T_IS_NOT_EQUAL) reduce 551) ((T_IS_NOT_IDENTICAL) reduce 551) ((T_IS_SMALLER_OR_EQUAL) reduce 551) ((T_LOGICAL_AND) reduce 551) ((T_LOGICAL_OR) reduce 551) ((T_LOGICAL_XOR) reduce 551) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_POW) reduce 551) ((T_SL) reduce 551) ((T_SPACESHIP) reduce 551) ((T_SR) reduce 551) (("[") reduce 551) (("^") reduce 551) (("|") reduce 551)) 28854 ((("%") reduce 547) (("(") reduce 538) (("*") reduce 547) (("+") reduce 547) (("-") reduce 547) ((".") reduce 547) (("/") reduce 547) ((";") reduce 547) (("<") reduce 547) ((">") reduce 547) (("?") reduce 547) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 547) ((T_BOOLEAN_AND) reduce 547) ((T_BOOLEAN_OR) reduce 547) ((T_COALESCE) reduce 547) ((T_DOUBLE_ARROW) reduce 547) ((T_INSTANCEOF) reduce 547) ((T_IS_EQUAL) reduce 547) ((T_IS_GREATER_OR_EQUAL) reduce 547) ((T_IS_IDENTICAL) reduce 547) ((T_IS_NOT_EQUAL) reduce 547) ((T_IS_NOT_IDENTICAL) reduce 547) ((T_IS_SMALLER_OR_EQUAL) reduce 547) ((T_LOGICAL_AND) reduce 547) ((T_LOGICAL_OR) reduce 547) ((T_LOGICAL_XOR) reduce 547) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 547) ((T_OBJECT_OPERATOR) reduce 547) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) ((T_POW) reduce 547) ((T_SL) reduce 547) ((T_SPACESHIP) reduce 547) ((T_SR) reduce 547) (("[") reduce 547) (("^") reduce 547) (("|") reduce 547)) 28856 ((("%") reduce 546) (("(") reduce 546) (("*") reduce 546) (("+") reduce 546) (("-") reduce 546) ((".") reduce 546) (("/") reduce 546) ((";") reduce 546) (("<") reduce 546) ((">") reduce 546) (("?") reduce 546) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 546) ((T_BOOLEAN_AND) reduce 546) ((T_BOOLEAN_OR) reduce 546) ((T_COALESCE) reduce 546) ((T_DOUBLE_ARROW) reduce 546) ((T_INSTANCEOF) reduce 546) ((T_IS_EQUAL) reduce 546) ((T_IS_GREATER_OR_EQUAL) reduce 546) ((T_IS_IDENTICAL) reduce 546) ((T_IS_NOT_EQUAL) reduce 546) ((T_IS_NOT_IDENTICAL) reduce 546) ((T_IS_SMALLER_OR_EQUAL) reduce 546) ((T_LOGICAL_AND) reduce 546) ((T_LOGICAL_OR) reduce 546) ((T_LOGICAL_XOR) reduce 546) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 546) ((T_OBJECT_OPERATOR) reduce 546) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 546) ((T_POW) reduce 546) ((T_SL) reduce 546) ((T_SPACESHIP) reduce 546) ((T_SR) reduce 546) (("[") reduce 546) (("^") reduce 546) (("|") reduce 546)) 28858 ((("%") reduce 542) (("(") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 28859 ((("%") reduce 548) (("*") reduce 548) (("+") reduce 548) (("-") reduce 548) ((".") reduce 548) (("/") reduce 548) ((";") reduce 548) (("<") reduce 548) ((">") reduce 548) (("?") reduce 548) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 548) ((T_BOOLEAN_AND) reduce 548) ((T_BOOLEAN_OR) reduce 548) ((T_COALESCE) reduce 548) ((T_DOUBLE_ARROW) reduce 548) ((T_INSTANCEOF) reduce 548) ((T_IS_EQUAL) reduce 548) ((T_IS_GREATER_OR_EQUAL) reduce 548) ((T_IS_IDENTICAL) reduce 548) ((T_IS_NOT_EQUAL) reduce 548) ((T_IS_NOT_IDENTICAL) reduce 548) ((T_IS_SMALLER_OR_EQUAL) reduce 548) ((T_LOGICAL_AND) reduce 548) ((T_LOGICAL_OR) reduce 548) ((T_LOGICAL_XOR) reduce 548) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_POW) reduce 548) ((T_SL) reduce 548) ((T_SPACESHIP) reduce 548) ((T_SR) reduce 548) (("[") reduce 548) (("^") reduce 548) (("|") reduce 548)) 28860 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((";") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 392) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("[") reduce 531) (("^") reduce 392) (("|") reduce 392)) 28864 ((("%") reduce 555) (("(") reduce 564) (("*") reduce 555) (("+") reduce 555) (("-") reduce 555) ((".") reduce 555) (("/") reduce 555) ((";") reduce 555) (("<") reduce 555) ((">") reduce 555) (("?") reduce 555) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 555) ((T_BOOLEAN_AND) reduce 555) ((T_BOOLEAN_OR) reduce 555) ((T_COALESCE) reduce 555) ((T_DOUBLE_ARROW) reduce 555) ((T_INSTANCEOF) reduce 555) ((T_IS_EQUAL) reduce 555) ((T_IS_GREATER_OR_EQUAL) reduce 555) ((T_IS_IDENTICAL) reduce 555) ((T_IS_NOT_EQUAL) reduce 555) ((T_IS_NOT_IDENTICAL) reduce 555) ((T_IS_SMALLER_OR_EQUAL) reduce 555) ((T_LOGICAL_AND) reduce 555) ((T_LOGICAL_OR) reduce 555) ((T_LOGICAL_XOR) reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 555) ((T_OBJECT_OPERATOR) reduce 555) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 555) ((T_POW) reduce 555) ((T_SL) reduce 555) ((T_SPACESHIP) reduce 555) ((T_SR) reduce 555) (("[") reduce 555) (("^") reduce 555) (("|") reduce 555)) 28866 ((("%") reduce 490) (("(") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("|") reduce 490)) 28867 ((("%") reduce 289) (("(") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 289) ((T_OBJECT_OPERATOR) reduce 289) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("[") reduce 289) (("^") reduce 289) (("|") reduce 289)) 28871 ((("%") reduce 290) (("(") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 290) ((T_OBJECT_OPERATOR) reduce 290) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("[") reduce 290) (("^") reduce 290) (("|") reduce 290)) 28872 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 28873 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 28876 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((";") reduce 554) (("<") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_POW) reduce 554) ((T_SL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 28877 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 28878 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 28883 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 28884 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 28885 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((";") reduce 549) (("<") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_POW) reduce 549) ((T_SL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 28886 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 28889 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 28890 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 28891 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((";") reduce 550) (("<") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_POW) reduce 550) ((T_SL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 28892 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((";") reduce 545) (("<") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_POW) reduce 545) ((T_SL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 28893 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 28894 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_POW) reduce 553) ((T_SL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 28897 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_POW) reduce 552) ((T_SL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 28898 ((("%") reduce 487) (("(") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) (("=") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AND_EQUAL) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_COALESCE_EQUAL) reduce 487) ((T_CONCAT_EQUAL) reduce 487) ((T_DEC) reduce 487) ((T_DIV_EQUAL) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INC) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_MINUS_EQUAL) reduce 487) ((T_MOD_EQUAL) reduce 487) ((T_MUL_EQUAL) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_OR_EQUAL) reduce 487) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 487) ((T_PLUS_EQUAL) reduce 487) ((T_POW) reduce 487) ((T_POW_EQUAL) reduce 487) ((T_SL) reduce 487) ((T_SL_EQUAL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) ((T_SR_EQUAL) reduce 487) ((T_XOR_EQUAL) reduce 487) (("[") reduce 487) (("^") reduce 487) (("|") reduce 487)) 28909 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 28910 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 28920 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 477) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28921 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 28923 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 464) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28925 ((("%") reduce 524) (("(") reduce 562) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("|") reduce 524)) 28927 ((("%") reduce 554) (("(") reduce 564) (("*") reduce 554) (("+") reduce 554) (("-") reduce 554) ((".") reduce 554) (("/") reduce 554) ((";") reduce 554) (("<") reduce 554) (("=") reduce 554) ((">") reduce 554) (("?") reduce 554) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 554) ((T_AND_EQUAL) reduce 554) ((T_BOOLEAN_AND) reduce 554) ((T_BOOLEAN_OR) reduce 554) ((T_COALESCE) reduce 554) ((T_COALESCE_EQUAL) reduce 554) ((T_CONCAT_EQUAL) reduce 554) ((T_DEC) reduce 554) ((T_DIV_EQUAL) reduce 554) ((T_DOUBLE_ARROW) reduce 554) ((T_INC) reduce 554) ((T_INSTANCEOF) reduce 554) ((T_IS_EQUAL) reduce 554) ((T_IS_GREATER_OR_EQUAL) reduce 554) ((T_IS_IDENTICAL) reduce 554) ((T_IS_NOT_EQUAL) reduce 554) ((T_IS_NOT_IDENTICAL) reduce 554) ((T_IS_SMALLER_OR_EQUAL) reduce 554) ((T_LOGICAL_AND) reduce 554) ((T_LOGICAL_OR) reduce 554) ((T_LOGICAL_XOR) reduce 554) ((T_MINUS_EQUAL) reduce 554) ((T_MOD_EQUAL) reduce 554) ((T_MUL_EQUAL) reduce 554) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_OR_EQUAL) reduce 554) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 554) ((T_PLUS_EQUAL) reduce 554) ((T_POW) reduce 554) ((T_POW_EQUAL) reduce 554) ((T_SL) reduce 554) ((T_SL_EQUAL) reduce 554) ((T_SPACESHIP) reduce 554) ((T_SR) reduce 554) ((T_SR_EQUAL) reduce 554) ((T_XOR_EQUAL) reduce 554) (("[") reduce 554) (("^") reduce 554) (("|") reduce 554)) 28930 ((("%") reduce 526) (("(") reduce 563) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("|") reduce 526)) 28931 ((("%") reduce 489) (("(") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) (("=") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AND_EQUAL) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_COALESCE_EQUAL) reduce 489) ((T_CONCAT_EQUAL) reduce 489) ((T_DEC) reduce 489) ((T_DIV_EQUAL) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INC) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_MINUS_EQUAL) reduce 489) ((T_MOD_EQUAL) reduce 489) ((T_MUL_EQUAL) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_OR_EQUAL) reduce 489) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 489) ((T_PLUS_EQUAL) reduce 489) ((T_POW) reduce 489) ((T_POW_EQUAL) reduce 489) ((T_SL) reduce 489) ((T_SL_EQUAL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) ((T_SR_EQUAL) reduce 489) ((T_XOR_EQUAL) reduce 489) (("[") reduce 489) (("^") reduce 489) (("|") reduce 489)) 28932 ((("%") reduce 491) (("(") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) (("=") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AND_EQUAL) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_COALESCE_EQUAL) reduce 491) ((T_CONCAT_EQUAL) reduce 491) ((T_DEC) reduce 491) ((T_DIV_EQUAL) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INC) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_MINUS_EQUAL) reduce 491) ((T_MOD_EQUAL) reduce 491) ((T_MUL_EQUAL) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_OR_EQUAL) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_PLUS_EQUAL) reduce 491) ((T_POW) reduce 491) ((T_POW_EQUAL) reduce 491) ((T_SL) reduce 491) ((T_SL_EQUAL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) ((T_SR_EQUAL) reduce 491) ((T_XOR_EQUAL) reduce 491) (("[") reduce 491) (("^") reduce 491) (("|") reduce 491)) 28934 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 28935 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 28940 ((("%") reduce 543) (("(") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) (("=") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AND_EQUAL) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_COALESCE_EQUAL) reduce 543) ((T_CONCAT_EQUAL) reduce 543) ((T_DEC) reduce 543) ((T_DIV_EQUAL) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INC) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_MINUS_EQUAL) reduce 543) ((T_MOD_EQUAL) reduce 543) ((T_MUL_EQUAL) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_OR_EQUAL) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PLUS_EQUAL) reduce 543) ((T_POW) reduce 543) ((T_POW_EQUAL) reduce 543) ((T_SL) reduce 543) ((T_SL_EQUAL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) ((T_SR_EQUAL) reduce 543) ((T_XOR_EQUAL) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 28941 ((("%") reduce 565) (("(") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 28942 ((("%") reduce 549) (("(") shift) (("*") reduce 549) (("+") reduce 549) (("-") reduce 549) ((".") reduce 549) (("/") reduce 549) ((";") reduce 549) (("<") reduce 549) (("=") reduce 549) ((">") reduce 549) (("?") reduce 549) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 549) ((T_AND_EQUAL) reduce 549) ((T_BOOLEAN_AND) reduce 549) ((T_BOOLEAN_OR) reduce 549) ((T_COALESCE) reduce 549) ((T_COALESCE_EQUAL) reduce 549) ((T_CONCAT_EQUAL) reduce 549) ((T_DEC) reduce 549) ((T_DIV_EQUAL) reduce 549) ((T_DOUBLE_ARROW) reduce 549) ((T_INC) reduce 549) ((T_INSTANCEOF) reduce 549) ((T_IS_EQUAL) reduce 549) ((T_IS_GREATER_OR_EQUAL) reduce 549) ((T_IS_IDENTICAL) reduce 549) ((T_IS_NOT_EQUAL) reduce 549) ((T_IS_NOT_IDENTICAL) reduce 549) ((T_IS_SMALLER_OR_EQUAL) reduce 549) ((T_LOGICAL_AND) reduce 549) ((T_LOGICAL_OR) reduce 549) ((T_LOGICAL_XOR) reduce 549) ((T_MINUS_EQUAL) reduce 549) ((T_MOD_EQUAL) reduce 549) ((T_MUL_EQUAL) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_OR_EQUAL) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PLUS_EQUAL) reduce 549) ((T_POW) reduce 549) ((T_POW_EQUAL) reduce 549) ((T_SL) reduce 549) ((T_SL_EQUAL) reduce 549) ((T_SPACESHIP) reduce 549) ((T_SR) reduce 549) ((T_SR_EQUAL) reduce 549) ((T_XOR_EQUAL) reduce 549) (("[") reduce 549) (("^") reduce 549) (("|") reduce 549)) 28943 ((("%") reduce 567) (("(") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 28946 ((("%") reduce 566) (("(") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 28947 ((("%") reduce 544) (("(") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) (("=") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AND_EQUAL) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_COALESCE_EQUAL) reduce 544) ((T_CONCAT_EQUAL) reduce 544) ((T_DEC) reduce 544) ((T_DIV_EQUAL) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INC) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_MINUS_EQUAL) reduce 544) ((T_MOD_EQUAL) reduce 544) ((T_MUL_EQUAL) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_OR_EQUAL) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PLUS_EQUAL) reduce 544) ((T_POW) reduce 544) ((T_POW_EQUAL) reduce 544) ((T_SL) reduce 544) ((T_SL_EQUAL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) ((T_SR_EQUAL) reduce 544) ((T_XOR_EQUAL) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 28948 ((("%") reduce 550) (("(") shift) (("*") reduce 550) (("+") reduce 550) (("-") reduce 550) ((".") reduce 550) (("/") reduce 550) ((";") reduce 550) (("<") reduce 550) (("=") reduce 550) ((">") reduce 550) (("?") reduce 550) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 550) ((T_AND_EQUAL) reduce 550) ((T_BOOLEAN_AND) reduce 550) ((T_BOOLEAN_OR) reduce 550) ((T_COALESCE) reduce 550) ((T_COALESCE_EQUAL) reduce 550) ((T_CONCAT_EQUAL) reduce 550) ((T_DEC) reduce 550) ((T_DIV_EQUAL) reduce 550) ((T_DOUBLE_ARROW) reduce 550) ((T_INC) reduce 550) ((T_INSTANCEOF) reduce 550) ((T_IS_EQUAL) reduce 550) ((T_IS_GREATER_OR_EQUAL) reduce 550) ((T_IS_IDENTICAL) reduce 550) ((T_IS_NOT_EQUAL) reduce 550) ((T_IS_NOT_IDENTICAL) reduce 550) ((T_IS_SMALLER_OR_EQUAL) reduce 550) ((T_LOGICAL_AND) reduce 550) ((T_LOGICAL_OR) reduce 550) ((T_LOGICAL_XOR) reduce 550) ((T_MINUS_EQUAL) reduce 550) ((T_MOD_EQUAL) reduce 550) ((T_MUL_EQUAL) reduce 550) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_OR_EQUAL) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PLUS_EQUAL) reduce 550) ((T_POW) reduce 550) ((T_POW_EQUAL) reduce 550) ((T_SL) reduce 550) ((T_SL_EQUAL) reduce 550) ((T_SPACESHIP) reduce 550) ((T_SR) reduce 550) ((T_SR_EQUAL) reduce 550) ((T_XOR_EQUAL) reduce 550) (("[") reduce 550) (("^") reduce 550) (("|") reduce 550)) 28949 ((("%") reduce 545) (("(") reduce 545) (("*") reduce 545) (("+") reduce 545) (("-") reduce 545) ((".") reduce 545) (("/") reduce 545) ((";") reduce 545) (("<") reduce 545) (("=") reduce 545) ((">") reduce 545) (("?") reduce 545) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 545) ((T_AND_EQUAL) reduce 545) ((T_BOOLEAN_AND) reduce 545) ((T_BOOLEAN_OR) reduce 545) ((T_COALESCE) reduce 545) ((T_COALESCE_EQUAL) reduce 545) ((T_CONCAT_EQUAL) reduce 545) ((T_DEC) reduce 545) ((T_DIV_EQUAL) reduce 545) ((T_DOUBLE_ARROW) reduce 545) ((T_INC) reduce 545) ((T_INSTANCEOF) reduce 545) ((T_IS_EQUAL) reduce 545) ((T_IS_GREATER_OR_EQUAL) reduce 545) ((T_IS_IDENTICAL) reduce 545) ((T_IS_NOT_EQUAL) reduce 545) ((T_IS_NOT_IDENTICAL) reduce 545) ((T_IS_SMALLER_OR_EQUAL) reduce 545) ((T_LOGICAL_AND) reduce 545) ((T_LOGICAL_OR) reduce 545) ((T_LOGICAL_XOR) reduce 545) ((T_MINUS_EQUAL) reduce 545) ((T_MOD_EQUAL) reduce 545) ((T_MUL_EQUAL) reduce 545) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 545) ((T_OBJECT_OPERATOR) reduce 545) ((T_OR_EQUAL) reduce 545) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 545) ((T_PLUS_EQUAL) reduce 545) ((T_POW) reduce 545) ((T_POW_EQUAL) reduce 545) ((T_SL) reduce 545) ((T_SL_EQUAL) reduce 545) ((T_SPACESHIP) reduce 545) ((T_SR) reduce 545) ((T_SR_EQUAL) reduce 545) ((T_XOR_EQUAL) reduce 545) (("[") reduce 545) (("^") reduce 545) (("|") reduce 545)) 28951 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 28953 ((("%") reduce 503) (("(") reduce 503) (("*") reduce 503) (("+") reduce 503) (("-") reduce 503) ((".") reduce 503) (("/") reduce 503) ((";") reduce 503) (("<") reduce 503) (("=") shift) ((">") reduce 503) (("?") reduce 503) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 503) ((T_BOOLEAN_AND) reduce 503) ((T_BOOLEAN_OR) reduce 503) ((T_COALESCE) reduce 503) ((T_DOUBLE_ARROW) reduce 503) ((T_INSTANCEOF) reduce 503) ((T_IS_EQUAL) reduce 503) ((T_IS_GREATER_OR_EQUAL) reduce 503) ((T_IS_IDENTICAL) reduce 503) ((T_IS_NOT_EQUAL) reduce 503) ((T_IS_NOT_IDENTICAL) reduce 503) ((T_IS_SMALLER_OR_EQUAL) reduce 503) ((T_LOGICAL_AND) reduce 503) ((T_LOGICAL_OR) reduce 503) ((T_LOGICAL_XOR) reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 503) ((T_OBJECT_OPERATOR) reduce 503) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 503) ((T_POW) reduce 503) ((T_SL) reduce 503) ((T_SPACESHIP) reduce 503) ((T_SR) reduce 503) (("[") reduce 503) (("^") reduce 503) (("|") reduce 503)) 28955 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 390) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28956 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 465) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28959 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 464) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28960 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((";") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456)) 28961 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 466) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28962 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) ((";") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452)) 28963 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 28965 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 28967 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 28968 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508)) 28969 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 602) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 602) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28970 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 601) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 601) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28971 ((("%") reduce 488) (("(") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) (("=") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AND_EQUAL) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_COALESCE_EQUAL) reduce 488) ((T_CONCAT_EQUAL) reduce 488) ((T_DEC) reduce 488) ((T_DIV_EQUAL) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INC) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_MINUS_EQUAL) reduce 488) ((T_MOD_EQUAL) reduce 488) ((T_MUL_EQUAL) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_OR_EQUAL) reduce 488) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 488) ((T_PLUS_EQUAL) reduce 488) ((T_POW) reduce 488) ((T_POW_EQUAL) reduce 488) ((T_SL) reduce 488) ((T_SL_EQUAL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) ((T_SR_EQUAL) reduce 488) ((T_XOR_EQUAL) reduce 488) (("[") reduce 488) (("^") reduce 488) (("|") reduce 488)) 28972 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 461) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 28973 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((";") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454)) 28975 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 28976 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 28977 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 28978 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 28979 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 28980 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) ((";") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385)) 28983 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 28984 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) ((";") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387)) 28985 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_DOUBLE_ARROW) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 28986 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_DOUBLE_ARROW) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 28987 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_DOUBLE_ARROW) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 28993 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_DOUBLE_ARROW) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 28994 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 28995 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_DOUBLE_ARROW) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 28996 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 28997 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) ((";") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384)) 28999 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 29000 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) ((";") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386)) 29008 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) ((";") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383)) 29010 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_DOUBLE_ARROW) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 29016 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) ((";") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_DOUBLE_ARROW) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231)) 29021 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 389) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 29025 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) ((";") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_DOUBLE_ARROW) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596)) 29026 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) ((";") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450)) 29027 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 599) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 599) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 29028 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 598) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 598) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 29029 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408)) 29031 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) ((";") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_DOUBLE_ARROW) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501)) 29032 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((";") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457)) 29033 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_DOUBLE_ARROW) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289)) 29037 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_DOUBLE_ARROW) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290)) 29038 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 29041 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) ((";") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_DOUBLE_ARROW) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600)) 29044 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) ((";") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_DOUBLE_ARROW) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597)) 29045 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) ((";") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451)) 29046 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((";") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410)) 29047 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((";") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 29048 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((";") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455)) 29049 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453)) 29052 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) ((";") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_DOUBLE_ARROW) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502)) 29053 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 29054 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) ((";") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430)) 29055 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) ((";") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429)) 29057 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443)) 29058 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_DOUBLE_ARROW) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 29061 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_DOUBLE_ARROW) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 29063 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 29064 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) ((";") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431)) 29065 ((("%") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((";") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_POW) shift) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("^") reduce 456) (("|") reduce 456)) 29066 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 466) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 29067 ((("%") reduce 452) (("*") reduce 452) (("+") reduce 452) (("-") reduce 452) ((".") reduce 452) (("/") reduce 452) ((";") reduce 452) (("<") reduce 452) ((">") reduce 452) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) reduce 452) ((T_IS_EQUAL) reduce 452) ((T_IS_GREATER_OR_EQUAL) reduce 452) ((T_IS_IDENTICAL) reduce 452) ((T_IS_NOT_EQUAL) reduce 452) ((T_IS_NOT_IDENTICAL) reduce 452) ((T_IS_SMALLER_OR_EQUAL) reduce 452) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_POW) shift) ((T_SL) reduce 452) ((T_SPACESHIP) reduce 452) ((T_SR) reduce 452) (("^") reduce 452) (("|") reduce 452)) 29070 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 29072 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 29073 ((("%") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("^") reduce 508) (("|") reduce 508)) 29074 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 602) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 29075 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 601) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 29077 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 461) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 29078 ((("%") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((";") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_POW) shift) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("^") reduce 454) (("|") reduce 454)) 29080 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 29081 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 29082 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 29083 ((("%") reduce 492) (("(") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("^") reduce 492) (("|") reduce 492)) 29084 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 29085 ((("%") reduce 385) (("(") reduce 385) (("*") reduce 385) (("+") reduce 385) (("-") reduce 385) ((".") reduce 385) (("/") reduce 385) ((";") reduce 385) (("<") reduce 385) ((">") reduce 385) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) reduce 385) ((T_IS_EQUAL) reduce 385) ((T_IS_GREATER_OR_EQUAL) reduce 385) ((T_IS_IDENTICAL) reduce 385) ((T_IS_NOT_EQUAL) reduce 385) ((T_IS_NOT_IDENTICAL) reduce 385) ((T_IS_SMALLER_OR_EQUAL) reduce 385) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 385) ((T_OBJECT_OPERATOR) reduce 385) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 385) ((T_POW) reduce 385) ((T_SL) reduce 385) ((T_SPACESHIP) reduce 385) ((T_SR) reduce 385) (("[") reduce 385) (("^") reduce 385) (("|") reduce 385)) 29088 ((("%") reduce 494) (("(") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 29089 ((("%") reduce 387) (("(") shift) (("*") reduce 387) (("+") reduce 387) (("-") reduce 387) ((".") reduce 387) (("/") reduce 387) ((";") reduce 387) (("<") reduce 387) ((">") reduce 387) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) reduce 387) ((T_IS_EQUAL) reduce 387) ((T_IS_GREATER_OR_EQUAL) reduce 387) ((T_IS_IDENTICAL) reduce 387) ((T_IS_NOT_EQUAL) reduce 387) ((T_IS_NOT_IDENTICAL) reduce 387) ((T_IS_SMALLER_OR_EQUAL) reduce 387) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) reduce 387) ((T_SL) reduce 387) ((T_SPACESHIP) reduce 387) ((T_SR) reduce 387) (("^") reduce 387) (("|") reduce 387)) 29090 ((("%") reduce 493) (("(") reduce 493) (("*") reduce 493) (("+") reduce 493) (("-") reduce 493) ((".") reduce 493) (("/") reduce 493) ((";") reduce 493) (("<") reduce 493) ((">") reduce 493) (("?") reduce 493) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 493) ((T_BOOLEAN_AND) reduce 493) ((T_BOOLEAN_OR) reduce 493) ((T_COALESCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_IS_EQUAL) reduce 493) ((T_IS_GREATER_OR_EQUAL) reduce 493) ((T_IS_IDENTICAL) reduce 493) ((T_IS_NOT_EQUAL) reduce 493) ((T_IS_NOT_IDENTICAL) reduce 493) ((T_IS_SMALLER_OR_EQUAL) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 493) ((T_POW) reduce 493) ((T_SL) reduce 493) ((T_SPACESHIP) reduce 493) ((T_SR) reduce 493) (("^") reduce 493) (("|") reduce 493)) 29091 ((("%") reduce 495) (("(") reduce 495) (("*") reduce 495) (("+") reduce 495) (("-") reduce 495) ((".") reduce 495) (("/") reduce 495) ((";") reduce 495) (("<") reduce 495) ((">") reduce 495) (("?") reduce 495) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 495) ((T_BOOLEAN_AND) reduce 495) ((T_BOOLEAN_OR) reduce 495) ((T_COALESCE) reduce 495) ((T_INSTANCEOF) reduce 495) ((T_IS_EQUAL) reduce 495) ((T_IS_GREATER_OR_EQUAL) reduce 495) ((T_IS_IDENTICAL) reduce 495) ((T_IS_NOT_EQUAL) reduce 495) ((T_IS_NOT_IDENTICAL) reduce 495) ((T_IS_SMALLER_OR_EQUAL) reduce 495) ((T_LOGICAL_AND) reduce 495) ((T_LOGICAL_OR) reduce 495) ((T_LOGICAL_XOR) reduce 495) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 495) ((T_SL) reduce 495) ((T_SPACESHIP) reduce 495) ((T_SR) reduce 495) (("[") shift) (("^") reduce 495) (("|") reduce 495)) 29092 ((("%") reduce 556) (("(") reduce 556) (("*") reduce 556) (("+") reduce 556) (("-") reduce 556) ((".") reduce 556) (("/") reduce 556) ((";") reduce 556) (("<") reduce 556) ((">") reduce 556) (("?") reduce 556) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 556) ((T_BOOLEAN_AND) reduce 556) ((T_BOOLEAN_OR) reduce 556) ((T_COALESCE) reduce 556) ((T_INSTANCEOF) reduce 556) ((T_IS_EQUAL) reduce 556) ((T_IS_GREATER_OR_EQUAL) reduce 556) ((T_IS_IDENTICAL) reduce 556) ((T_IS_NOT_EQUAL) reduce 556) ((T_IS_NOT_IDENTICAL) reduce 556) ((T_IS_SMALLER_OR_EQUAL) reduce 556) ((T_LOGICAL_AND) reduce 556) ((T_LOGICAL_OR) reduce 556) ((T_LOGICAL_XOR) reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 556) ((T_OBJECT_OPERATOR) reduce 556) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 556) ((T_POW) reduce 556) ((T_SL) reduce 556) ((T_SPACESHIP) reduce 556) ((T_SR) reduce 556) (("[") reduce 556) (("^") reduce 556) (("|") reduce 556)) 29098 ((("%") reduce 557) (("(") reduce 557) (("*") reduce 557) (("+") reduce 557) (("-") reduce 557) ((".") reduce 557) (("/") reduce 557) ((";") reduce 557) (("<") reduce 557) ((">") reduce 557) (("?") reduce 557) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 557) ((T_BOOLEAN_AND) reduce 557) ((T_BOOLEAN_OR) reduce 557) ((T_COALESCE) reduce 557) ((T_INSTANCEOF) reduce 557) ((T_IS_EQUAL) reduce 557) ((T_IS_GREATER_OR_EQUAL) reduce 557) ((T_IS_IDENTICAL) reduce 557) ((T_IS_NOT_EQUAL) reduce 557) ((T_IS_NOT_IDENTICAL) reduce 557) ((T_IS_SMALLER_OR_EQUAL) reduce 557) ((T_LOGICAL_AND) reduce 557) ((T_LOGICAL_OR) reduce 557) ((T_LOGICAL_XOR) reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 557) ((T_OBJECT_OPERATOR) reduce 557) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 557) ((T_POW) reduce 557) ((T_SL) reduce 557) ((T_SPACESHIP) reduce 557) ((T_SR) reduce 557) (("[") reduce 557) (("^") reduce 557) (("|") reduce 557)) 29099 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 29100 ((("%") reduce 558) (("(") reduce 558) (("*") reduce 558) (("+") reduce 558) (("-") reduce 558) ((".") reduce 558) (("/") reduce 558) ((";") reduce 558) (("<") reduce 558) ((">") reduce 558) (("?") reduce 558) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 558) ((T_BOOLEAN_AND) reduce 558) ((T_BOOLEAN_OR) reduce 558) ((T_COALESCE) reduce 558) ((T_INSTANCEOF) reduce 558) ((T_IS_EQUAL) reduce 558) ((T_IS_GREATER_OR_EQUAL) reduce 558) ((T_IS_IDENTICAL) reduce 558) ((T_IS_NOT_EQUAL) reduce 558) ((T_IS_NOT_IDENTICAL) reduce 558) ((T_IS_SMALLER_OR_EQUAL) reduce 558) ((T_LOGICAL_AND) reduce 558) ((T_LOGICAL_OR) reduce 558) ((T_LOGICAL_XOR) reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 558) ((T_OBJECT_OPERATOR) reduce 558) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 558) ((T_POW) reduce 558) ((T_SL) reduce 558) ((T_SPACESHIP) reduce 558) ((T_SR) reduce 558) (("[") reduce 558) (("^") reduce 558) (("|") reduce 558)) 29101 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 29102 ((("%") reduce 384) (("(") reduce 384) (("*") reduce 384) (("+") reduce 384) (("-") reduce 384) ((".") reduce 384) (("/") reduce 384) ((";") reduce 384) (("<") reduce 384) ((">") reduce 384) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) reduce 384) ((T_IS_EQUAL) reduce 384) ((T_IS_GREATER_OR_EQUAL) reduce 384) ((T_IS_IDENTICAL) reduce 384) ((T_IS_NOT_EQUAL) reduce 384) ((T_IS_NOT_IDENTICAL) reduce 384) ((T_IS_SMALLER_OR_EQUAL) reduce 384) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 384) ((T_OBJECT_OPERATOR) reduce 384) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 384) ((T_POW) reduce 384) ((T_SL) reduce 384) ((T_SPACESHIP) reduce 384) ((T_SR) reduce 384) (("[") reduce 384) (("^") reduce 384) (("|") reduce 384)) 29104 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 29105 ((("%") reduce 386) (("(") reduce 386) (("*") reduce 386) (("+") reduce 386) (("-") reduce 386) ((".") reduce 386) (("/") reduce 386) ((";") reduce 386) (("<") reduce 386) ((">") reduce 386) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) reduce 386) ((T_IS_EQUAL) reduce 386) ((T_IS_GREATER_OR_EQUAL) reduce 386) ((T_IS_IDENTICAL) reduce 386) ((T_IS_NOT_EQUAL) reduce 386) ((T_IS_NOT_IDENTICAL) reduce 386) ((T_IS_SMALLER_OR_EQUAL) reduce 386) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 386) ((T_OBJECT_OPERATOR) reduce 386) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 386) ((T_POW) reduce 386) ((T_SL) reduce 386) ((T_SPACESHIP) reduce 386) ((T_SR) reduce 386) (("[") reduce 386) (("^") reduce 386) (("|") reduce 386)) 29113 ((("%") reduce 383) (("(") reduce 383) (("*") reduce 383) (("+") reduce 383) (("-") reduce 383) ((".") reduce 383) (("/") reduce 383) ((";") reduce 383) (("<") reduce 383) ((">") reduce 383) (("?") reduce 383) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 383) ((T_BOOLEAN_AND) reduce 383) ((T_BOOLEAN_OR) reduce 383) ((T_COALESCE) reduce 383) ((T_INSTANCEOF) reduce 383) ((T_IS_EQUAL) reduce 383) ((T_IS_GREATER_OR_EQUAL) reduce 383) ((T_IS_IDENTICAL) reduce 383) ((T_IS_NOT_EQUAL) reduce 383) ((T_IS_NOT_IDENTICAL) reduce 383) ((T_IS_SMALLER_OR_EQUAL) reduce 383) ((T_LOGICAL_AND) reduce 383) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 383) ((T_OBJECT_OPERATOR) reduce 383) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 383) ((T_POW) reduce 383) ((T_SL) reduce 383) ((T_SPACESHIP) reduce 383) ((T_SR) reduce 383) (("[") reduce 383) (("^") reduce 383) (("|") reduce 383)) 29115 ((("%") reduce 496) (("(") reduce 496) (("*") reduce 496) (("+") reduce 496) (("-") reduce 496) ((".") reduce 496) (("/") reduce 496) ((";") reduce 496) (("<") reduce 496) ((">") reduce 496) (("?") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 496) ((T_BOOLEAN_AND) reduce 496) ((T_BOOLEAN_OR) reduce 496) ((T_COALESCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_IS_EQUAL) reduce 496) ((T_IS_GREATER_OR_EQUAL) reduce 496) ((T_IS_IDENTICAL) reduce 496) ((T_IS_NOT_EQUAL) reduce 496) ((T_IS_NOT_IDENTICAL) reduce 496) ((T_IS_SMALLER_OR_EQUAL) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_POW) reduce 496) ((T_SL) reduce 496) ((T_SPACESHIP) reduce 496) ((T_SR) reduce 496) (("^") reduce 496) (("|") reduce 496)) 29121 ((("%") reduce 231) (("*") reduce 231) (("+") reduce 231) (("-") reduce 231) ((".") reduce 231) (("/") reduce 231) ((";") reduce 231) (("<") reduce 231) ((">") reduce 231) (("?") reduce 231) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 231) ((T_BOOLEAN_AND) reduce 231) ((T_BOOLEAN_OR) reduce 231) ((T_COALESCE) reduce 231) ((T_INSTANCEOF) reduce 231) ((T_IS_EQUAL) reduce 231) ((T_IS_GREATER_OR_EQUAL) reduce 231) ((T_IS_IDENTICAL) reduce 231) ((T_IS_NOT_EQUAL) reduce 231) ((T_IS_NOT_IDENTICAL) reduce 231) ((T_IS_SMALLER_OR_EQUAL) reduce 231) ((T_LOGICAL_AND) reduce 231) ((T_LOGICAL_OR) reduce 231) ((T_LOGICAL_XOR) reduce 231) ((T_POW) reduce 231) ((T_SL) reduce 231) ((T_SPACESHIP) reduce 231) ((T_SR) reduce 231) (("^") reduce 231) (("|") reduce 231)) 29126 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 389) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 29130 ((("%") reduce 596) (("*") reduce 596) (("+") reduce 596) (("-") reduce 596) ((".") reduce 596) (("/") reduce 596) ((";") reduce 596) (("<") reduce 596) ((">") reduce 596) (("?") reduce 596) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 596) ((T_BOOLEAN_AND) reduce 596) ((T_BOOLEAN_OR) reduce 596) ((T_COALESCE) reduce 596) ((T_INSTANCEOF) reduce 596) ((T_IS_EQUAL) reduce 596) ((T_IS_GREATER_OR_EQUAL) reduce 596) ((T_IS_IDENTICAL) reduce 596) ((T_IS_NOT_EQUAL) reduce 596) ((T_IS_NOT_IDENTICAL) reduce 596) ((T_IS_SMALLER_OR_EQUAL) reduce 596) ((T_LOGICAL_AND) reduce 596) ((T_LOGICAL_OR) reduce 596) ((T_LOGICAL_XOR) reduce 596) ((T_POW) reduce 596) ((T_SL) reduce 596) ((T_SPACESHIP) reduce 596) ((T_SR) reduce 596) (("^") reduce 596) (("|") reduce 596)) 29131 ((("%") reduce 450) (("*") reduce 450) (("+") reduce 450) (("-") reduce 450) ((".") reduce 450) (("/") reduce 450) ((";") reduce 450) (("<") reduce 450) ((">") reduce 450) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) reduce 450) ((T_IS_EQUAL) reduce 450) ((T_IS_GREATER_OR_EQUAL) reduce 450) ((T_IS_IDENTICAL) reduce 450) ((T_IS_NOT_EQUAL) reduce 450) ((T_IS_NOT_IDENTICAL) reduce 450) ((T_IS_SMALLER_OR_EQUAL) reduce 450) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_POW) shift) ((T_SL) reduce 450) ((T_SPACESHIP) reduce 450) ((T_SR) reduce 450) (("^") reduce 450) (("|") reduce 450)) 29132 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 599) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 29133 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 598) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 29134 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("[") reduce 531) (("^") reduce 408) (("|") reduce 408)) 29136 ((("%") reduce 501) (("*") reduce 501) (("+") reduce 501) (("-") reduce 501) ((".") reduce 501) (("/") reduce 501) ((";") reduce 501) (("<") reduce 501) ((">") reduce 501) (("?") reduce 501) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 501) ((T_BOOLEAN_AND) reduce 501) ((T_BOOLEAN_OR) reduce 501) ((T_COALESCE) reduce 501) ((T_INSTANCEOF) reduce 501) ((T_IS_EQUAL) reduce 501) ((T_IS_GREATER_OR_EQUAL) reduce 501) ((T_IS_IDENTICAL) reduce 501) ((T_IS_NOT_EQUAL) reduce 501) ((T_IS_NOT_IDENTICAL) reduce 501) ((T_IS_SMALLER_OR_EQUAL) reduce 501) ((T_LOGICAL_AND) reduce 501) ((T_LOGICAL_OR) reduce 501) ((T_LOGICAL_XOR) reduce 501) ((T_POW) reduce 501) ((T_SL) reduce 501) ((T_SPACESHIP) reduce 501) ((T_SR) reduce 501) (("^") reduce 501) (("|") reduce 501)) 29137 ((("%") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((";") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("^") reduce 457) (("|") reduce 457)) 29138 ((("%") reduce 289) (("*") reduce 289) (("+") reduce 289) (("-") reduce 289) ((".") reduce 289) (("/") reduce 289) ((";") reduce 289) (("<") reduce 289) ((">") reduce 289) (("?") reduce 289) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 289) ((T_BOOLEAN_AND) reduce 289) ((T_BOOLEAN_OR) reduce 289) ((T_COALESCE) reduce 289) ((T_INSTANCEOF) reduce 289) ((T_IS_EQUAL) reduce 289) ((T_IS_GREATER_OR_EQUAL) reduce 289) ((T_IS_IDENTICAL) reduce 289) ((T_IS_NOT_EQUAL) reduce 289) ((T_IS_NOT_IDENTICAL) reduce 289) ((T_IS_SMALLER_OR_EQUAL) reduce 289) ((T_LOGICAL_AND) reduce 289) ((T_LOGICAL_OR) reduce 289) ((T_LOGICAL_XOR) reduce 289) ((T_POW) reduce 289) ((T_SL) reduce 289) ((T_SPACESHIP) reduce 289) ((T_SR) reduce 289) (("^") reduce 289) (("|") reduce 289)) 29142 ((("%") reduce 290) (("*") reduce 290) (("+") reduce 290) (("-") reduce 290) ((".") reduce 290) (("/") reduce 290) ((";") reduce 290) (("<") reduce 290) ((">") reduce 290) (("?") reduce 290) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 290) ((T_BOOLEAN_AND) reduce 290) ((T_BOOLEAN_OR) reduce 290) ((T_COALESCE) reduce 290) ((T_INSTANCEOF) reduce 290) ((T_IS_EQUAL) reduce 290) ((T_IS_GREATER_OR_EQUAL) reduce 290) ((T_IS_IDENTICAL) reduce 290) ((T_IS_NOT_EQUAL) reduce 290) ((T_IS_NOT_IDENTICAL) reduce 290) ((T_IS_SMALLER_OR_EQUAL) reduce 290) ((T_LOGICAL_AND) reduce 290) ((T_LOGICAL_OR) reduce 290) ((T_LOGICAL_XOR) reduce 290) ((T_POW) reduce 290) ((T_SL) reduce 290) ((T_SPACESHIP) reduce 290) ((T_SR) reduce 290) (("^") reduce 290) (("|") reduce 290)) 29143 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 29146 ((("%") reduce 600) (("*") reduce 600) (("+") reduce 600) (("-") reduce 600) ((".") reduce 600) (("/") reduce 600) ((";") reduce 600) (("<") reduce 600) ((">") reduce 600) (("?") reduce 600) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 600) ((T_BOOLEAN_AND) reduce 600) ((T_BOOLEAN_OR) reduce 600) ((T_COALESCE) reduce 600) ((T_INSTANCEOF) reduce 600) ((T_IS_EQUAL) reduce 600) ((T_IS_GREATER_OR_EQUAL) reduce 600) ((T_IS_IDENTICAL) reduce 600) ((T_IS_NOT_EQUAL) reduce 600) ((T_IS_NOT_IDENTICAL) reduce 600) ((T_IS_SMALLER_OR_EQUAL) reduce 600) ((T_LOGICAL_AND) reduce 600) ((T_LOGICAL_OR) reduce 600) ((T_LOGICAL_XOR) reduce 600) ((T_POW) reduce 600) ((T_SL) reduce 600) ((T_SPACESHIP) reduce 600) ((T_SR) reduce 600) (("^") reduce 600) (("|") reduce 600)) 29149 ((("%") reduce 597) (("*") reduce 597) (("+") reduce 597) (("-") reduce 597) ((".") reduce 597) (("/") reduce 597) ((";") reduce 597) (("<") reduce 597) ((">") reduce 597) (("?") reduce 597) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 597) ((T_BOOLEAN_AND) reduce 597) ((T_BOOLEAN_OR) reduce 597) ((T_COALESCE) reduce 597) ((T_INSTANCEOF) reduce 597) ((T_IS_EQUAL) reduce 597) ((T_IS_GREATER_OR_EQUAL) reduce 597) ((T_IS_IDENTICAL) reduce 597) ((T_IS_NOT_EQUAL) reduce 597) ((T_IS_NOT_IDENTICAL) reduce 597) ((T_IS_SMALLER_OR_EQUAL) reduce 597) ((T_LOGICAL_AND) reduce 597) ((T_LOGICAL_OR) reduce 597) ((T_LOGICAL_XOR) reduce 597) ((T_POW) reduce 597) ((T_SL) reduce 597) ((T_SPACESHIP) reduce 597) ((T_SR) reduce 597) (("^") reduce 597) (("|") reduce 597)) 29150 ((("%") reduce 451) (("*") reduce 451) (("+") reduce 451) (("-") reduce 451) ((".") reduce 451) (("/") reduce 451) ((";") reduce 451) (("<") reduce 451) ((">") reduce 451) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) reduce 451) ((T_IS_EQUAL) reduce 451) ((T_IS_GREATER_OR_EQUAL) reduce 451) ((T_IS_IDENTICAL) reduce 451) ((T_IS_NOT_EQUAL) reduce 451) ((T_IS_NOT_IDENTICAL) reduce 451) ((T_IS_SMALLER_OR_EQUAL) reduce 451) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_POW) shift) ((T_SL) reduce 451) ((T_SPACESHIP) reduce 451) ((T_SR) reduce 451) (("^") reduce 451) (("|") reduce 451)) 29151 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((";") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("[") reduce 531) (("^") reduce 410) (("|") reduce 410)) 29152 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((";") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) reduce 393) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) reduce 393) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 29153 ((("%") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((";") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_POW) shift) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("^") reduce 455) (("|") reduce 455)) 29154 ((("%") reduce 453) (("*") reduce 453) (("+") reduce 453) (("-") reduce 453) ((".") reduce 453) (("/") reduce 453) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) reduce 453) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_POW) shift) ((T_SL) reduce 453) ((T_SPACESHIP) reduce 453) ((T_SR) reduce 453) (("^") reduce 453) (("|") reduce 453)) 29157 ((("%") reduce 502) (("(") reduce 502) (("*") reduce 502) (("+") reduce 502) (("-") reduce 502) ((".") reduce 502) (("/") reduce 502) ((";") reduce 502) (("<") reduce 502) ((">") reduce 502) (("?") reduce 502) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 502) ((T_BOOLEAN_AND) reduce 502) ((T_BOOLEAN_OR) reduce 502) ((T_COALESCE) reduce 502) ((T_INSTANCEOF) reduce 502) ((T_IS_EQUAL) reduce 502) ((T_IS_GREATER_OR_EQUAL) reduce 502) ((T_IS_IDENTICAL) reduce 502) ((T_IS_NOT_EQUAL) reduce 502) ((T_IS_NOT_IDENTICAL) reduce 502) ((T_IS_SMALLER_OR_EQUAL) reduce 502) ((T_LOGICAL_AND) reduce 502) ((T_LOGICAL_OR) reduce 502) ((T_LOGICAL_XOR) reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 502) ((T_OBJECT_OPERATOR) reduce 502) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 502) ((T_POW) reduce 502) ((T_SL) reduce 502) ((T_SPACESHIP) reduce 502) ((T_SR) reduce 502) (("[") reduce 502) (("^") reduce 502) (("|") reduce 502)) 29158 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_POW) shift) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 29159 ((("%") reduce 430) (("*") reduce 430) (("+") reduce 430) (("-") reduce 430) ((".") reduce 430) (("/") reduce 430) ((";") reduce 430) (("<") reduce 430) ((">") reduce 430) (("?") reduce 430) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 430) ((T_BOOLEAN_AND) reduce 430) ((T_BOOLEAN_OR) reduce 430) ((T_COALESCE) reduce 430) ((T_INSTANCEOF) reduce 430) ((T_IS_EQUAL) reduce 430) ((T_IS_GREATER_OR_EQUAL) reduce 430) ((T_IS_IDENTICAL) reduce 430) ((T_IS_NOT_EQUAL) reduce 430) ((T_IS_NOT_IDENTICAL) reduce 430) ((T_IS_SMALLER_OR_EQUAL) reduce 430) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) reduce 430) ((T_SPACESHIP) reduce 430) ((T_SR) reduce 430) (("^") reduce 430) (("|") reduce 430)) 29160 ((("%") reduce 429) (("*") reduce 429) (("+") reduce 429) (("-") reduce 429) ((".") reduce 429) (("/") reduce 429) ((";") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_COALESCE) reduce 429) ((T_INSTANCEOF) reduce 429) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_POW) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) (("^") reduce 429) (("|") reduce 429)) 29162 ((("%") reduce 443) (("(") reduce 539) (("*") reduce 443) (("+") reduce 443) (("-") reduce 443) ((".") reduce 443) (("/") reduce 443) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 443) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("[") reduce 532) (("^") reduce 443) (("|") reduce 443)) 29163 ((("%") reduce 553) (("(") reduce 553) (("*") reduce 553) (("+") reduce 553) (("-") reduce 553) ((".") reduce 553) (("/") reduce 553) ((";") reduce 553) (("<") reduce 553) (("=") reduce 553) ((">") reduce 553) (("?") reduce 553) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 553) ((T_AND_EQUAL) reduce 553) ((T_BOOLEAN_AND) reduce 553) ((T_BOOLEAN_OR) reduce 553) ((T_COALESCE) reduce 553) ((T_COALESCE_EQUAL) reduce 553) ((T_CONCAT_EQUAL) reduce 553) ((T_DEC) reduce 553) ((T_DIV_EQUAL) reduce 553) ((T_INC) reduce 553) ((T_INSTANCEOF) reduce 553) ((T_IS_EQUAL) reduce 553) ((T_IS_GREATER_OR_EQUAL) reduce 553) ((T_IS_IDENTICAL) reduce 553) ((T_IS_NOT_EQUAL) reduce 553) ((T_IS_NOT_IDENTICAL) reduce 553) ((T_IS_SMALLER_OR_EQUAL) reduce 553) ((T_LOGICAL_AND) reduce 553) ((T_LOGICAL_OR) reduce 553) ((T_LOGICAL_XOR) reduce 553) ((T_MINUS_EQUAL) reduce 553) ((T_MOD_EQUAL) reduce 553) ((T_MUL_EQUAL) reduce 553) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_OR_EQUAL) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 553) ((T_PLUS_EQUAL) reduce 553) ((T_POW) reduce 553) ((T_POW_EQUAL) reduce 553) ((T_SL) reduce 553) ((T_SL_EQUAL) reduce 553) ((T_SPACESHIP) reduce 553) ((T_SR) reduce 553) ((T_SR_EQUAL) reduce 553) ((T_XOR_EQUAL) reduce 553) (("[") reduce 553) (("^") reduce 553) (("|") reduce 553)) 29166 ((("%") reduce 552) (("(") reduce 552) (("*") reduce 552) (("+") reduce 552) (("-") reduce 552) ((".") reduce 552) (("/") reduce 552) ((";") reduce 552) (("<") reduce 552) (("=") reduce 552) ((">") reduce 552) (("?") reduce 552) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 552) ((T_AND_EQUAL) reduce 552) ((T_BOOLEAN_AND) reduce 552) ((T_BOOLEAN_OR) reduce 552) ((T_COALESCE) reduce 552) ((T_COALESCE_EQUAL) reduce 552) ((T_CONCAT_EQUAL) reduce 552) ((T_DEC) reduce 552) ((T_DIV_EQUAL) reduce 552) ((T_INC) reduce 552) ((T_INSTANCEOF) reduce 552) ((T_IS_EQUAL) reduce 552) ((T_IS_GREATER_OR_EQUAL) reduce 552) ((T_IS_IDENTICAL) reduce 552) ((T_IS_NOT_EQUAL) reduce 552) ((T_IS_NOT_IDENTICAL) reduce 552) ((T_IS_SMALLER_OR_EQUAL) reduce 552) ((T_LOGICAL_AND) reduce 552) ((T_LOGICAL_OR) reduce 552) ((T_LOGICAL_XOR) reduce 552) ((T_MINUS_EQUAL) reduce 552) ((T_MOD_EQUAL) reduce 552) ((T_MUL_EQUAL) reduce 552) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_OR_EQUAL) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PLUS_EQUAL) reduce 552) ((T_POW) reduce 552) ((T_POW_EQUAL) reduce 552) ((T_SL) reduce 552) ((T_SL_EQUAL) reduce 552) ((T_SPACESHIP) reduce 552) ((T_SR) reduce 552) ((T_SR_EQUAL) reduce 552) ((T_XOR_EQUAL) reduce 552) (("[") reduce 552) (("^") reduce 552) (("|") reduce 552)) 29168 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 29169 ((("%") reduce 431) (("*") reduce 431) (("+") reduce 431) (("-") reduce 431) ((".") reduce 431) (("/") reduce 431) ((";") reduce 431) (("<") reduce 431) ((">") reduce 431) (("?") reduce 431) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 431) ((T_BOOLEAN_AND) reduce 431) ((T_BOOLEAN_OR) reduce 431) ((T_COALESCE) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 431) ((T_IS_GREATER_OR_EQUAL) reduce 431) ((T_IS_IDENTICAL) reduce 431) ((T_IS_NOT_EQUAL) reduce 431) ((T_IS_NOT_IDENTICAL) reduce 431) ((T_IS_SMALLER_OR_EQUAL) reduce 431) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) reduce 431) ((T_SPACESHIP) reduce 431) ((T_SR) reduce 431) (("^") reduce 431) (("|") reduce 431)) 29170 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (($) reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONST) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAMESPACE) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_USE) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 29186 ((("!") reduce 187) (("\"") reduce 187) (("$") reduce 187) (($) reduce 187) (("(") reduce 187) (("+") reduce 187) (("-") reduce 187) ((";") reduce 187) (("@") reduce 187) ((T_ABSTRACT) reduce 187) ((T_ARRAY) reduce 187) ((T_ARRAY_CAST) reduce 187) ((T_ATTRIBUTE) reduce 187) ((T_BOOL_CAST) reduce 187) ((T_BREAK) reduce 187) ((T_CLASS) reduce 187) ((T_CLASS_C) reduce 187) ((T_CLONE) reduce 187) ((T_CONST) reduce 187) ((T_CONSTANT_ENCAPSED_STRING) reduce 187) ((T_CONTINUE) reduce 187) ((T_DEC) reduce 187) ((T_DECLARE) reduce 187) ((T_DIR) reduce 187) ((T_DNUMBER) reduce 187) ((T_DO) reduce 187) ((T_DOUBLE_CAST) reduce 187) ((T_ECHO) reduce 187) ((T_EMPTY) reduce 187) ((T_ENUM) reduce 187) ((T_EVAL) reduce 187) ((T_EXIT) reduce 187) ((T_FILE) reduce 187) ((T_FINAL) reduce 187) ((T_FN) reduce 187) ((T_FOR) reduce 187) ((T_FOREACH) reduce 187) ((T_FUNCTION) reduce 187) ((T_FUNC_C) reduce 187) ((T_GLOBAL) reduce 187) ((T_GOTO) reduce 187) ((T_HALT_COMPILER) reduce 187) ((T_IF) reduce 187) ((T_INC) reduce 187) ((T_INCLUDE) reduce 187) ((T_INCLUDE_ONCE) reduce 187) ((T_INLINE_HTML) reduce 187) ((T_INTERFACE) reduce 187) ((T_INT_CAST) reduce 187) ((T_ISSET) reduce 187) ((T_LINE) reduce 187) ((T_LIST) reduce 187) ((T_LNUMBER) reduce 187) ((T_MATCH) reduce 187) ((T_METHOD_C) reduce 187) ((T_NAMESPACE) reduce 187) ((T_NAME_FULLY_QUALIFIED) reduce 187) ((T_NAME_QUALIFIED) reduce 187) ((T_NAME_RELATIVE) reduce 187) ((T_NEW) reduce 187) ((T_NS_C) reduce 187) ((T_OBJECT_CAST) reduce 187) ((T_PRINT) reduce 187) ((T_PROPERTY_C) reduce 187) ((T_READONLY) reduce 187) ((T_REQUIRE) reduce 187) ((T_REQUIRE_ONCE) reduce 187) ((T_RETURN) reduce 187) ((T_START_HEREDOC) reduce 187) ((T_STATIC) reduce 187) ((T_STRING) reduce 187) ((T_STRING_CAST) reduce 187) ((T_SWITCH) reduce 187) ((T_THROW) reduce 187) ((T_TRAIT) reduce 187) ((T_TRAIT_C) reduce 187) ((T_TRY) reduce 187) ((T_UNSET) reduce 187) ((T_UNSET_CAST) reduce 187) ((T_USE) reduce 187) ((T_VARIABLE) reduce 187) ((T_WHILE) reduce 187) ((T_YIELD) reduce 187) ((T_YIELD_FROM) reduce 187) (("[") reduce 187) (("`") reduce 187) (("{") reduce 187) (("~") reduce 187)) 29192 ((("!") reduce 110) (("\"") reduce 110) (($) reduce 110) (("$") reduce 110) (("(") reduce 110) (("+") reduce 110) (("-") reduce 110) ((";") reduce 110) (("@") reduce 110) ((T_ABSTRACT) reduce 110) ((T_ARRAY) reduce 110) ((T_ARRAY_CAST) reduce 110) ((T_ATTRIBUTE) reduce 110) ((T_BOOL_CAST) reduce 110) ((T_BREAK) reduce 110) ((T_CLASS) reduce 110) ((T_CLASS_C) reduce 110) ((T_CLONE) reduce 110) ((T_CONST) reduce 110) ((T_CONSTANT_ENCAPSED_STRING) reduce 110) ((T_CONTINUE) reduce 110) ((T_DEC) reduce 110) ((T_DECLARE) reduce 110) ((T_DIR) reduce 110) ((T_DNUMBER) reduce 110) ((T_DO) reduce 110) ((T_DOUBLE_CAST) reduce 110) ((T_ECHO) reduce 110) ((T_EMPTY) reduce 110) ((T_ENUM) reduce 110) ((T_EVAL) reduce 110) ((T_EXIT) reduce 110) ((T_FILE) reduce 110) ((T_FINAL) reduce 110) ((T_FN) reduce 110) ((T_FOR) reduce 110) ((T_FOREACH) reduce 110) ((T_FUNCTION) reduce 110) ((T_FUNC_C) reduce 110) ((T_GLOBAL) reduce 110) ((T_GOTO) reduce 110) ((T_HALT_COMPILER) reduce 110) ((T_IF) reduce 110) ((T_INC) reduce 110) ((T_INCLUDE) reduce 110) ((T_INCLUDE_ONCE) reduce 110) ((T_INLINE_HTML) reduce 110) ((T_INTERFACE) reduce 110) ((T_INT_CAST) reduce 110) ((T_ISSET) reduce 110) ((T_LINE) reduce 110) ((T_LIST) reduce 110) ((T_LNUMBER) reduce 110) ((T_MATCH) reduce 110) ((T_METHOD_C) reduce 110) ((T_NAMESPACE) reduce 110) ((T_NAME_FULLY_QUALIFIED) reduce 110) ((T_NAME_QUALIFIED) reduce 110) ((T_NAME_RELATIVE) reduce 110) ((T_NEW) reduce 110) ((T_NS_C) reduce 110) ((T_OBJECT_CAST) reduce 110) ((T_PRINT) reduce 110) ((T_PROPERTY_C) reduce 110) ((T_READONLY) reduce 110) ((T_REQUIRE) reduce 110) ((T_REQUIRE_ONCE) reduce 110) ((T_RETURN) reduce 110) ((T_START_HEREDOC) reduce 110) ((T_STATIC) reduce 110) ((T_STRING) reduce 110) ((T_STRING_CAST) reduce 110) ((T_SWITCH) reduce 110) ((T_THROW) reduce 110) ((T_TRAIT) reduce 110) ((T_TRAIT_C) reduce 110) ((T_TRY) reduce 110) ((T_UNSET) reduce 110) ((T_UNSET_CAST) reduce 110) ((T_USE) reduce 110) ((T_VARIABLE) reduce 110) ((T_WHILE) reduce 110) ((T_YIELD) reduce 110) ((T_YIELD_FROM) reduce 110) (("[") reduce 110) (("`") reduce 110) (("{") reduce 110) (("~") reduce 110)) 29193 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (($) reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONST) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAMESPACE) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_USE) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 29195 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (($) reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONST) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAMESPACE) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_USE) reduce 107) ((T_VARIABLE) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 29197 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (($) reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONST) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAMESPACE) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_USE) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 29198 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (($) reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONST) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAMESPACE) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_USE) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 29199 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (($) reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONST) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAMESPACE) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_USE) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 29210 ((("!") reduce 477) (("\"") reduce 477) (($) reduce 477) (("$") reduce 477) (("(") reduce 477) (("+") reduce 477) (("-") reduce 477) ((";") reduce 477) (("@") reduce 477) ((T_ABSTRACT) reduce 477) ((T_ARRAY) reduce 477) ((T_ARRAY_CAST) reduce 477) ((T_ATTRIBUTE) reduce 477) ((T_BOOL_CAST) reduce 477) ((T_BREAK) reduce 477) ((T_CLASS) reduce 477) ((T_CLASS_C) reduce 477) ((T_CLONE) reduce 477) ((T_CONST) reduce 477) ((T_CONSTANT_ENCAPSED_STRING) reduce 477) ((T_CONTINUE) reduce 477) ((T_DEC) reduce 477) ((T_DECLARE) reduce 477) ((T_DIR) reduce 477) ((T_DNUMBER) reduce 477) ((T_DO) reduce 477) ((T_DOUBLE_CAST) reduce 477) ((T_ECHO) reduce 477) ((T_EMPTY) reduce 477) ((T_ENUM) reduce 477) ((T_EVAL) reduce 477) ((T_EXIT) reduce 477) ((T_FILE) reduce 477) ((T_FINAL) reduce 477) ((T_FN) reduce 477) ((T_FOR) reduce 477) ((T_FOREACH) reduce 477) ((T_FUNCTION) reduce 477) ((T_FUNC_C) reduce 477) ((T_GLOBAL) reduce 477) ((T_GOTO) reduce 477) ((T_HALT_COMPILER) reduce 477) ((T_IF) reduce 477) ((T_INC) reduce 477) ((T_INCLUDE) reduce 477) ((T_INCLUDE_ONCE) reduce 477) ((T_INLINE_HTML) reduce 477) ((T_INTERFACE) reduce 477) ((T_INT_CAST) reduce 477) ((T_ISSET) reduce 477) ((T_LINE) reduce 477) ((T_LIST) reduce 477) ((T_LNUMBER) reduce 477) ((T_MATCH) reduce 477) ((T_METHOD_C) reduce 477) ((T_NAMESPACE) reduce 477) ((T_NAME_FULLY_QUALIFIED) reduce 477) ((T_NAME_QUALIFIED) reduce 477) ((T_NAME_RELATIVE) reduce 477) ((T_NEW) reduce 477) ((T_NS_C) reduce 477) ((T_OBJECT_CAST) reduce 477) ((T_PRINT) reduce 477) ((T_PROPERTY_C) reduce 477) ((T_READONLY) reduce 477) ((T_REQUIRE) reduce 477) ((T_REQUIRE_ONCE) reduce 477) ((T_RETURN) reduce 477) ((T_START_HEREDOC) reduce 477) ((T_STATIC) reduce 477) ((T_STRING) reduce 477) ((T_STRING_CAST) reduce 477) ((T_SWITCH) reduce 477) ((T_THROW) reduce 477) ((T_TRAIT) reduce 477) ((T_TRAIT_C) reduce 477) ((T_TRY) reduce 477) ((T_UNSET) reduce 477) ((T_UNSET_CAST) reduce 477) ((T_USE) reduce 477) ((T_VARIABLE) reduce 477) ((T_WHILE) reduce 477) ((T_YIELD) reduce 477) ((T_YIELD_FROM) reduce 477) (("[") reduce 477) (("`") reduce 477) (("{") reduce 477) (("~") reduce 477)) 29211 ((("!") reduce 182) (("\"") reduce 182) (($) reduce 182) (("$") reduce 182) (("(") reduce 182) (("+") reduce 182) (("-") reduce 182) ((";") reduce 182) (("@") reduce 182) ((T_ABSTRACT) reduce 182) ((T_ARRAY) reduce 182) ((T_ARRAY_CAST) reduce 182) ((T_ATTRIBUTE) reduce 182) ((T_BOOL_CAST) reduce 182) ((T_BREAK) reduce 182) ((T_CLASS) reduce 182) ((T_CLASS_C) reduce 182) ((T_CLONE) reduce 182) ((T_CONST) reduce 182) ((T_CONSTANT_ENCAPSED_STRING) reduce 182) ((T_CONTINUE) reduce 182) ((T_DEC) reduce 182) ((T_DECLARE) reduce 182) ((T_DIR) reduce 182) ((T_DNUMBER) reduce 182) ((T_DO) reduce 182) ((T_DOUBLE_CAST) reduce 182) ((T_ECHO) reduce 182) ((T_EMPTY) reduce 182) ((T_ENUM) reduce 182) ((T_EVAL) reduce 182) ((T_EXIT) reduce 182) ((T_FILE) reduce 182) ((T_FINAL) reduce 182) ((T_FN) reduce 182) ((T_FOR) reduce 182) ((T_FOREACH) reduce 182) ((T_FUNCTION) reduce 182) ((T_FUNC_C) reduce 182) ((T_GLOBAL) reduce 182) ((T_GOTO) reduce 182) ((T_HALT_COMPILER) reduce 182) ((T_IF) reduce 182) ((T_INC) reduce 182) ((T_INCLUDE) reduce 182) ((T_INCLUDE_ONCE) reduce 182) ((T_INLINE_HTML) reduce 182) ((T_INTERFACE) reduce 182) ((T_INT_CAST) reduce 182) ((T_ISSET) reduce 182) ((T_LINE) reduce 182) ((T_LIST) reduce 182) ((T_LNUMBER) reduce 182) ((T_MATCH) reduce 182) ((T_METHOD_C) reduce 182) ((T_NAMESPACE) reduce 182) ((T_NAME_FULLY_QUALIFIED) reduce 182) ((T_NAME_QUALIFIED) reduce 182) ((T_NAME_RELATIVE) reduce 182) ((T_NEW) reduce 182) ((T_NS_C) reduce 182) ((T_OBJECT_CAST) reduce 182) ((T_PRINT) reduce 182) ((T_PROPERTY_C) reduce 182) ((T_READONLY) reduce 182) ((T_REQUIRE) reduce 182) ((T_REQUIRE_ONCE) reduce 182) ((T_RETURN) reduce 182) ((T_START_HEREDOC) reduce 182) ((T_STATIC) reduce 182) ((T_STRING) reduce 182) ((T_STRING_CAST) reduce 182) ((T_SWITCH) reduce 182) ((T_THROW) reduce 182) ((T_TRAIT) reduce 182) ((T_TRAIT_C) reduce 182) ((T_TRY) reduce 182) ((T_UNSET) reduce 182) ((T_UNSET_CAST) reduce 182) ((T_USE) reduce 182) ((T_VARIABLE) reduce 182) ((T_WHILE) reduce 182) ((T_YIELD) reduce 182) ((T_YIELD_FROM) reduce 182) (("[") reduce 182) (("`") reduce 182) (("{") reduce 182) (("~") reduce 182)) 29219 ((("!") reduce 187) (("\"") reduce 187) (($) reduce 187) (("$") reduce 187) (("(") reduce 187) (("+") reduce 187) (("-") reduce 187) ((";") reduce 187) (("@") reduce 187) ((T_ABSTRACT) reduce 187) ((T_ARRAY) reduce 187) ((T_ARRAY_CAST) reduce 187) ((T_ATTRIBUTE) reduce 187) ((T_BOOL_CAST) reduce 187) ((T_BREAK) reduce 187) ((T_CLASS) reduce 187) ((T_CLASS_C) reduce 187) ((T_CLONE) reduce 187) ((T_CONST) reduce 187) ((T_CONSTANT_ENCAPSED_STRING) reduce 187) ((T_CONTINUE) reduce 187) ((T_DEC) reduce 187) ((T_DECLARE) reduce 187) ((T_DIR) reduce 187) ((T_DNUMBER) reduce 187) ((T_DO) reduce 187) ((T_DOUBLE_CAST) reduce 187) ((T_ECHO) reduce 187) ((T_EMPTY) reduce 187) ((T_ENUM) reduce 187) ((T_EVAL) reduce 187) ((T_EXIT) reduce 187) ((T_FILE) reduce 187) ((T_FINAL) reduce 187) ((T_FN) reduce 187) ((T_FOR) reduce 187) ((T_FOREACH) reduce 187) ((T_FUNCTION) reduce 187) ((T_FUNC_C) reduce 187) ((T_GLOBAL) reduce 187) ((T_GOTO) reduce 187) ((T_HALT_COMPILER) reduce 187) ((T_IF) reduce 187) ((T_INC) reduce 187) ((T_INCLUDE) reduce 187) ((T_INCLUDE_ONCE) reduce 187) ((T_INLINE_HTML) reduce 187) ((T_INTERFACE) reduce 187) ((T_INT_CAST) reduce 187) ((T_ISSET) reduce 187) ((T_LINE) reduce 187) ((T_LIST) reduce 187) ((T_LNUMBER) reduce 187) ((T_MATCH) reduce 187) ((T_METHOD_C) reduce 187) ((T_NAMESPACE) reduce 187) ((T_NAME_FULLY_QUALIFIED) reduce 187) ((T_NAME_QUALIFIED) reduce 187) ((T_NAME_RELATIVE) reduce 187) ((T_NEW) reduce 187) ((T_NS_C) reduce 187) ((T_OBJECT_CAST) reduce 187) ((T_PRINT) reduce 187) ((T_PROPERTY_C) reduce 187) ((T_READONLY) reduce 187) ((T_REQUIRE) reduce 187) ((T_REQUIRE_ONCE) reduce 187) ((T_RETURN) reduce 187) ((T_START_HEREDOC) reduce 187) ((T_STATIC) reduce 187) ((T_STRING) reduce 187) ((T_STRING_CAST) reduce 187) ((T_SWITCH) reduce 187) ((T_THROW) reduce 187) ((T_TRAIT) reduce 187) ((T_TRAIT_C) reduce 187) ((T_TRY) reduce 187) ((T_UNSET) reduce 187) ((T_UNSET_CAST) reduce 187) ((T_USE) reduce 187) ((T_VARIABLE) reduce 187) ((T_WHILE) reduce 187) ((T_YIELD) reduce 187) ((T_YIELD_FROM) reduce 187) (("[") reduce 187) (("`") reduce 187) (("{") reduce 187) (("~") reduce 187)) 29224 ((("!") reduce 198) (("\"") reduce 198) (("$") reduce 198) (($) reduce 198) (("(") reduce 198) (("+") reduce 198) (("-") reduce 198) ((";") reduce 198) (("@") reduce 198) ((T_ABSTRACT) reduce 198) ((T_ARRAY) reduce 198) ((T_ARRAY_CAST) reduce 198) ((T_ATTRIBUTE) reduce 198) ((T_BOOL_CAST) reduce 198) ((T_BREAK) reduce 198) ((T_CLASS) reduce 198) ((T_CLASS_C) reduce 198) ((T_CLONE) reduce 198) ((T_CONST) reduce 198) ((T_CONSTANT_ENCAPSED_STRING) reduce 198) ((T_CONTINUE) reduce 198) ((T_DEC) reduce 198) ((T_DECLARE) reduce 198) ((T_DIR) reduce 198) ((T_DNUMBER) reduce 198) ((T_DO) reduce 198) ((T_DOUBLE_CAST) reduce 198) ((T_ECHO) reduce 198) ((T_EMPTY) reduce 198) ((T_ENUM) reduce 198) ((T_EVAL) reduce 198) ((T_EXIT) reduce 198) ((T_FILE) reduce 198) ((T_FINAL) reduce 198) ((T_FN) reduce 198) ((T_FOR) reduce 198) ((T_FOREACH) reduce 198) ((T_FUNCTION) reduce 198) ((T_FUNC_C) reduce 198) ((T_GLOBAL) reduce 198) ((T_GOTO) reduce 198) ((T_HALT_COMPILER) reduce 198) ((T_IF) reduce 198) ((T_INC) reduce 198) ((T_INCLUDE) reduce 198) ((T_INCLUDE_ONCE) reduce 198) ((T_INLINE_HTML) reduce 198) ((T_INTERFACE) reduce 198) ((T_INT_CAST) reduce 198) ((T_ISSET) reduce 198) ((T_LINE) reduce 198) ((T_LIST) reduce 198) ((T_LNUMBER) reduce 198) ((T_MATCH) reduce 198) ((T_METHOD_C) reduce 198) ((T_NAMESPACE) reduce 198) ((T_NAME_FULLY_QUALIFIED) reduce 198) ((T_NAME_QUALIFIED) reduce 198) ((T_NAME_RELATIVE) reduce 198) ((T_NEW) reduce 198) ((T_NS_C) reduce 198) ((T_OBJECT_CAST) reduce 198) ((T_PRINT) reduce 198) ((T_PROPERTY_C) reduce 198) ((T_READONLY) reduce 198) ((T_REQUIRE) reduce 198) ((T_REQUIRE_ONCE) reduce 198) ((T_RETURN) reduce 198) ((T_START_HEREDOC) reduce 198) ((T_STATIC) reduce 198) ((T_STRING) reduce 198) ((T_STRING_CAST) reduce 198) ((T_SWITCH) reduce 198) ((T_THROW) reduce 198) ((T_TRAIT) reduce 198) ((T_TRAIT_C) reduce 198) ((T_TRY) reduce 198) ((T_UNSET) reduce 198) ((T_UNSET_CAST) reduce 198) ((T_USE) reduce 198) ((T_VARIABLE) reduce 198) ((T_WHILE) reduce 198) ((T_YIELD) reduce 198) ((T_YIELD_FROM) reduce 198) (("[") reduce 198) (("`") reduce 198) (("{") reduce 198) (("~") reduce 198)) 29230 ((("!") reduce 199) (("\"") reduce 199) (($) reduce 199) (("$") reduce 199) (("(") reduce 199) (("+") reduce 199) (("-") reduce 199) ((";") reduce 199) (("@") reduce 199) ((T_ABSTRACT) reduce 199) ((T_ARRAY) reduce 199) ((T_ARRAY_CAST) reduce 199) ((T_ATTRIBUTE) reduce 199) ((T_BOOL_CAST) reduce 199) ((T_BREAK) reduce 199) ((T_CLASS) reduce 199) ((T_CLASS_C) reduce 199) ((T_CLONE) reduce 199) ((T_CONST) reduce 199) ((T_CONSTANT_ENCAPSED_STRING) reduce 199) ((T_CONTINUE) reduce 199) ((T_DEC) reduce 199) ((T_DECLARE) reduce 199) ((T_DIR) reduce 199) ((T_DNUMBER) reduce 199) ((T_DO) reduce 199) ((T_DOUBLE_CAST) reduce 199) ((T_ECHO) reduce 199) ((T_EMPTY) reduce 199) ((T_ENUM) reduce 199) ((T_EVAL) reduce 199) ((T_EXIT) reduce 199) ((T_FILE) reduce 199) ((T_FINAL) reduce 199) ((T_FN) reduce 199) ((T_FOR) reduce 199) ((T_FOREACH) reduce 199) ((T_FUNCTION) reduce 199) ((T_FUNC_C) reduce 199) ((T_GLOBAL) reduce 199) ((T_GOTO) reduce 199) ((T_HALT_COMPILER) reduce 199) ((T_IF) reduce 199) ((T_INC) reduce 199) ((T_INCLUDE) reduce 199) ((T_INCLUDE_ONCE) reduce 199) ((T_INLINE_HTML) reduce 199) ((T_INTERFACE) reduce 199) ((T_INT_CAST) reduce 199) ((T_ISSET) reduce 199) ((T_LINE) reduce 199) ((T_LIST) reduce 199) ((T_LNUMBER) reduce 199) ((T_MATCH) reduce 199) ((T_METHOD_C) reduce 199) ((T_NAMESPACE) reduce 199) ((T_NAME_FULLY_QUALIFIED) reduce 199) ((T_NAME_QUALIFIED) reduce 199) ((T_NAME_RELATIVE) reduce 199) ((T_NEW) reduce 199) ((T_NS_C) reduce 199) ((T_OBJECT_CAST) reduce 199) ((T_PRINT) reduce 199) ((T_PROPERTY_C) reduce 199) ((T_READONLY) reduce 199) ((T_REQUIRE) reduce 199) ((T_REQUIRE_ONCE) reduce 199) ((T_RETURN) reduce 199) ((T_START_HEREDOC) reduce 199) ((T_STATIC) reduce 199) ((T_STRING) reduce 199) ((T_STRING_CAST) reduce 199) ((T_SWITCH) reduce 199) ((T_THROW) reduce 199) ((T_TRAIT) reduce 199) ((T_TRAIT_C) reduce 199) ((T_TRY) reduce 199) ((T_UNSET) reduce 199) ((T_UNSET_CAST) reduce 199) ((T_USE) reduce 199) ((T_VARIABLE) reduce 199) ((T_WHILE) reduce 199) ((T_YIELD) reduce 199) ((T_YIELD_FROM) reduce 199) (("[") reduce 199) (("`") reduce 199) (("{") reduce 199) (("~") reduce 199)) 29237 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (($) reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONST) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAMESPACE) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_USE) reduce 200) ((T_VARIABLE) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 29244 ((("!") reduce 188) (("\"") reduce 188) (("$") reduce 188) (($) reduce 188) (("(") reduce 188) (("+") reduce 188) (("-") reduce 188) ((";") reduce 188) (("@") reduce 188) ((T_ABSTRACT) reduce 188) ((T_ARRAY) reduce 188) ((T_ARRAY_CAST) reduce 188) ((T_ATTRIBUTE) reduce 188) ((T_BOOL_CAST) reduce 188) ((T_BREAK) reduce 188) ((T_CLASS) reduce 188) ((T_CLASS_C) reduce 188) ((T_CLONE) reduce 188) ((T_CONST) reduce 188) ((T_CONSTANT_ENCAPSED_STRING) reduce 188) ((T_CONTINUE) reduce 188) ((T_DEC) reduce 188) ((T_DECLARE) reduce 188) ((T_DIR) reduce 188) ((T_DNUMBER) reduce 188) ((T_DO) reduce 188) ((T_DOUBLE_CAST) reduce 188) ((T_ECHO) reduce 188) ((T_EMPTY) reduce 188) ((T_ENUM) reduce 188) ((T_EVAL) reduce 188) ((T_EXIT) reduce 188) ((T_FILE) reduce 188) ((T_FINAL) reduce 188) ((T_FN) reduce 188) ((T_FOR) reduce 188) ((T_FOREACH) reduce 188) ((T_FUNCTION) reduce 188) ((T_FUNC_C) reduce 188) ((T_GLOBAL) reduce 188) ((T_GOTO) reduce 188) ((T_HALT_COMPILER) reduce 188) ((T_IF) reduce 188) ((T_INC) reduce 188) ((T_INCLUDE) reduce 188) ((T_INCLUDE_ONCE) reduce 188) ((T_INLINE_HTML) reduce 188) ((T_INTERFACE) reduce 188) ((T_INT_CAST) reduce 188) ((T_ISSET) reduce 188) ((T_LINE) reduce 188) ((T_LIST) reduce 188) ((T_LNUMBER) reduce 188) ((T_MATCH) reduce 188) ((T_METHOD_C) reduce 188) ((T_NAMESPACE) reduce 188) ((T_NAME_FULLY_QUALIFIED) reduce 188) ((T_NAME_QUALIFIED) reduce 188) ((T_NAME_RELATIVE) reduce 188) ((T_NEW) reduce 188) ((T_NS_C) reduce 188) ((T_OBJECT_CAST) reduce 188) ((T_PRINT) reduce 188) ((T_PROPERTY_C) reduce 188) ((T_READONLY) reduce 188) ((T_REQUIRE) reduce 188) ((T_REQUIRE_ONCE) reduce 188) ((T_RETURN) reduce 188) ((T_START_HEREDOC) reduce 188) ((T_STATIC) reduce 188) ((T_STRING) reduce 188) ((T_STRING_CAST) reduce 188) ((T_SWITCH) reduce 188) ((T_THROW) reduce 188) ((T_TRAIT) reduce 188) ((T_TRAIT_C) reduce 188) ((T_TRY) reduce 188) ((T_UNSET) reduce 188) ((T_UNSET_CAST) reduce 188) ((T_USE) reduce 188) ((T_VARIABLE) reduce 188) ((T_WHILE) reduce 188) ((T_YIELD) reduce 188) ((T_YIELD_FROM) reduce 188) (("[") reduce 188) (("`") reduce 188) (("{") reduce 188) (("~") reduce 188)) 29261 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (($) reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONST) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_EMPTY) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAMESPACE) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_USE) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("~") reduce 248)) 29265 ((("!") reduce 249) (("\"") reduce 249) (($) reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONST) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_EMPTY) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAMESPACE) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_USE) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("~") reduce 249)) 29798 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (($) reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONST) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAMESPACE) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_USE) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 29799 ((("!") reduce 149) (("\"") reduce 149) (($) reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONST) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAMESPACE) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_USE) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 29802 ((("!") reduce 241) (("\"") reduce 241) (($) reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONST) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAMESPACE) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_USE) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 29803 (((T_NAME_FULLY_QUALIFIED) reduce 121) ((T_NAME_QUALIFIED) reduce 121) ((T_STRING) reduce 121)) 29804 (((T_NAME_FULLY_QUALIFIED) reduce 120) ((T_NAME_QUALIFIED) reduce 120) ((T_STRING) reduce 120)) 29805 (((",") reduce 91) ((";") reduce 91) ((T_AS) reduce 91) ((T_NS_SEPARATOR) reduce 91)) 29806 (((",") reduce 89) ((";") reduce 89) ((T_AS) reduce 89) ((T_NS_SEPARATOR) reduce 89)) 29807 (((",") reduce 88) ((";") reduce 88) ((T_AS) reduce 88) ((T_NS_SEPARATOR) reduce 88)) 29808 (((",") reduce 136) ((";") reduce 136) ((T_AS) shift) ((T_NS_SEPARATOR) shift)) 29810 (((",") reduce 90) ((";") reduce 90) ((T_AS) reduce 90) ((T_NS_SEPARATOR) reduce 90)) 29811 (((",") reduce 131) ((";") reduce 131)) 29813 (((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_STRING) shift)) 29818 ((("!") reduce 118) (("\"") reduce 118) (($) reduce 118) (("$") reduce 118) (("(") reduce 118) (("+") reduce 118) (("-") reduce 118) ((";") reduce 118) (("@") reduce 118) ((T_ABSTRACT) reduce 118) ((T_ARRAY) reduce 118) ((T_ARRAY_CAST) reduce 118) ((T_ATTRIBUTE) reduce 118) ((T_BOOL_CAST) reduce 118) ((T_BREAK) reduce 118) ((T_CLASS) reduce 118) ((T_CLASS_C) reduce 118) ((T_CLONE) reduce 118) ((T_CONST) reduce 118) ((T_CONSTANT_ENCAPSED_STRING) reduce 118) ((T_CONTINUE) reduce 118) ((T_DEC) reduce 118) ((T_DECLARE) reduce 118) ((T_DIR) reduce 118) ((T_DNUMBER) reduce 118) ((T_DO) reduce 118) ((T_DOUBLE_CAST) reduce 118) ((T_ECHO) reduce 118) ((T_EMPTY) reduce 118) ((T_ENUM) reduce 118) ((T_EVAL) reduce 118) ((T_EXIT) reduce 118) ((T_FILE) reduce 118) ((T_FINAL) reduce 118) ((T_FN) reduce 118) ((T_FOR) reduce 118) ((T_FOREACH) reduce 118) ((T_FUNCTION) reduce 118) ((T_FUNC_C) reduce 118) ((T_GLOBAL) reduce 118) ((T_GOTO) reduce 118) ((T_HALT_COMPILER) reduce 118) ((T_IF) reduce 118) ((T_INC) reduce 118) ((T_INCLUDE) reduce 118) ((T_INCLUDE_ONCE) reduce 118) ((T_INLINE_HTML) reduce 118) ((T_INTERFACE) reduce 118) ((T_INT_CAST) reduce 118) ((T_ISSET) reduce 118) ((T_LINE) reduce 118) ((T_LIST) reduce 118) ((T_LNUMBER) reduce 118) ((T_MATCH) reduce 118) ((T_METHOD_C) reduce 118) ((T_NAMESPACE) reduce 118) ((T_NAME_FULLY_QUALIFIED) reduce 118) ((T_NAME_QUALIFIED) reduce 118) ((T_NAME_RELATIVE) reduce 118) ((T_NEW) reduce 118) ((T_NS_C) reduce 118) ((T_OBJECT_CAST) reduce 118) ((T_PRINT) reduce 118) ((T_PROPERTY_C) reduce 118) ((T_READONLY) reduce 118) ((T_REQUIRE) reduce 118) ((T_REQUIRE_ONCE) reduce 118) ((T_RETURN) reduce 118) ((T_START_HEREDOC) reduce 118) ((T_STATIC) reduce 118) ((T_STRING) reduce 118) ((T_STRING_CAST) reduce 118) ((T_SWITCH) reduce 118) ((T_THROW) reduce 118) ((T_TRAIT) reduce 118) ((T_TRAIT_C) reduce 118) ((T_TRY) reduce 118) ((T_UNSET) reduce 118) ((T_UNSET_CAST) reduce 118) ((T_USE) reduce 118) ((T_VARIABLE) reduce 118) ((T_WHILE) reduce 118) ((T_YIELD) reduce 118) ((T_YIELD_FROM) reduce 118) (("[") reduce 118) (("`") reduce 118) (("{") reduce 118) (("~") reduce 118)) 29819 (((",") reduce 91) ((";") reduce 91) ((T_AS) reduce 91)) 29820 (((",") reduce 89) ((";") reduce 89) ((T_AS) reduce 89)) 29821 (((",") reduce 88) ((";") reduce 88) ((T_AS) reduce 88)) 29822 (((",") reduce 136) ((";") reduce 136) ((T_AS) shift)) 29823 (((",") reduce 90) ((";") reduce 90) ((T_AS) reduce 90)) 29824 (((",") reduce 130) ((";") reduce 130)) 29826 (((",") reduce 137) ((";") reduce 137)) 29828 (((T_NAME_QUALIFIED) shift) ((T_STRING) shift)) 29829 (((",") reduce 89) ((T_AS) reduce 89) (("}") reduce 89)) 29830 (((",") reduce 88) ((T_AS) reduce 88) (("}") reduce 88)) 29831 (((",") reduce 134) ((T_AS) shift) (("}") reduce 134)) 29832 (((",") reduce 129) (("}") reduce 129)) 29834 (((T_NAME_QUALIFIED) shift) ((T_STRING) shift) (("}") reduce 125)) 29836 (((";") reduce 122)) 29837 (((",") reduce 128) (("}") reduce 128)) 29839 (((",") reduce 135) (("}") reduce 135)) 29840 ((("!") reduce 116) (("\"") reduce 116) (($) reduce 116) (("$") reduce 116) (("(") reduce 116) (("+") reduce 116) (("-") reduce 116) ((";") reduce 116) (("@") reduce 116) ((T_ABSTRACT) reduce 116) ((T_ARRAY) reduce 116) ((T_ARRAY_CAST) reduce 116) ((T_ATTRIBUTE) reduce 116) ((T_BOOL_CAST) reduce 116) ((T_BREAK) reduce 116) ((T_CLASS) reduce 116) ((T_CLASS_C) reduce 116) ((T_CLONE) reduce 116) ((T_CONST) reduce 116) ((T_CONSTANT_ENCAPSED_STRING) reduce 116) ((T_CONTINUE) reduce 116) ((T_DEC) reduce 116) ((T_DECLARE) reduce 116) ((T_DIR) reduce 116) ((T_DNUMBER) reduce 116) ((T_DO) reduce 116) ((T_DOUBLE_CAST) reduce 116) ((T_ECHO) reduce 116) ((T_EMPTY) reduce 116) ((T_ENUM) reduce 116) ((T_EVAL) reduce 116) ((T_EXIT) reduce 116) ((T_FILE) reduce 116) ((T_FINAL) reduce 116) ((T_FN) reduce 116) ((T_FOR) reduce 116) ((T_FOREACH) reduce 116) ((T_FUNCTION) reduce 116) ((T_FUNC_C) reduce 116) ((T_GLOBAL) reduce 116) ((T_GOTO) reduce 116) ((T_HALT_COMPILER) reduce 116) ((T_IF) reduce 116) ((T_INC) reduce 116) ((T_INCLUDE) reduce 116) ((T_INCLUDE_ONCE) reduce 116) ((T_INLINE_HTML) reduce 116) ((T_INTERFACE) reduce 116) ((T_INT_CAST) reduce 116) ((T_ISSET) reduce 116) ((T_LINE) reduce 116) ((T_LIST) reduce 116) ((T_LNUMBER) reduce 116) ((T_MATCH) reduce 116) ((T_METHOD_C) reduce 116) ((T_NAMESPACE) reduce 116) ((T_NAME_FULLY_QUALIFIED) reduce 116) ((T_NAME_QUALIFIED) reduce 116) ((T_NAME_RELATIVE) reduce 116) ((T_NEW) reduce 116) ((T_NS_C) reduce 116) ((T_OBJECT_CAST) reduce 116) ((T_PRINT) reduce 116) ((T_PROPERTY_C) reduce 116) ((T_READONLY) reduce 116) ((T_REQUIRE) reduce 116) ((T_REQUIRE_ONCE) reduce 116) ((T_RETURN) reduce 116) ((T_START_HEREDOC) reduce 116) ((T_STATIC) reduce 116) ((T_STRING) reduce 116) ((T_STRING_CAST) reduce 116) ((T_SWITCH) reduce 116) ((T_THROW) reduce 116) ((T_TRAIT) reduce 116) ((T_TRAIT_C) reduce 116) ((T_TRY) reduce 116) ((T_UNSET) reduce 116) ((T_UNSET_CAST) reduce 116) ((T_USE) reduce 116) ((T_VARIABLE) reduce 116) ((T_WHILE) reduce 116) ((T_YIELD) reduce 116) ((T_YIELD_FROM) reduce 116) (("[") reduce 116) (("`") reduce 116) (("{") reduce 116) (("~") reduce 116)) 29841 ((("!") reduce 117) (("\"") reduce 117) (("$") reduce 117) (($) reduce 117) (("(") reduce 117) (("+") reduce 117) (("-") reduce 117) ((";") reduce 117) (("@") reduce 117) ((T_ABSTRACT) reduce 117) ((T_ARRAY) reduce 117) ((T_ARRAY_CAST) reduce 117) ((T_ATTRIBUTE) reduce 117) ((T_BOOL_CAST) reduce 117) ((T_BREAK) reduce 117) ((T_CLASS) reduce 117) ((T_CLASS_C) reduce 117) ((T_CLONE) reduce 117) ((T_CONST) reduce 117) ((T_CONSTANT_ENCAPSED_STRING) reduce 117) ((T_CONTINUE) reduce 117) ((T_DEC) reduce 117) ((T_DECLARE) reduce 117) ((T_DIR) reduce 117) ((T_DNUMBER) reduce 117) ((T_DO) reduce 117) ((T_DOUBLE_CAST) reduce 117) ((T_ECHO) reduce 117) ((T_EMPTY) reduce 117) ((T_ENUM) reduce 117) ((T_EVAL) reduce 117) ((T_EXIT) reduce 117) ((T_FILE) reduce 117) ((T_FINAL) reduce 117) ((T_FN) reduce 117) ((T_FOR) reduce 117) ((T_FOREACH) reduce 117) ((T_FUNCTION) reduce 117) ((T_FUNC_C) reduce 117) ((T_GLOBAL) reduce 117) ((T_GOTO) reduce 117) ((T_HALT_COMPILER) reduce 117) ((T_IF) reduce 117) ((T_INC) reduce 117) ((T_INCLUDE) reduce 117) ((T_INCLUDE_ONCE) reduce 117) ((T_INLINE_HTML) reduce 117) ((T_INTERFACE) reduce 117) ((T_INT_CAST) reduce 117) ((T_ISSET) reduce 117) ((T_LINE) reduce 117) ((T_LIST) reduce 117) ((T_LNUMBER) reduce 117) ((T_MATCH) reduce 117) ((T_METHOD_C) reduce 117) ((T_NAMESPACE) reduce 117) ((T_NAME_FULLY_QUALIFIED) reduce 117) ((T_NAME_QUALIFIED) reduce 117) ((T_NAME_RELATIVE) reduce 117) ((T_NEW) reduce 117) ((T_NS_C) reduce 117) ((T_OBJECT_CAST) reduce 117) ((T_PRINT) reduce 117) ((T_PROPERTY_C) reduce 117) ((T_READONLY) reduce 117) ((T_REQUIRE) reduce 117) ((T_REQUIRE_ONCE) reduce 117) ((T_RETURN) reduce 117) ((T_START_HEREDOC) reduce 117) ((T_STATIC) reduce 117) ((T_STRING) reduce 117) ((T_STRING_CAST) reduce 117) ((T_SWITCH) reduce 117) ((T_THROW) reduce 117) ((T_TRAIT) reduce 117) ((T_TRAIT_C) reduce 117) ((T_TRY) reduce 117) ((T_UNSET) reduce 117) ((T_UNSET_CAST) reduce 117) ((T_USE) reduce 117) ((T_VARIABLE) reduce 117) ((T_WHILE) reduce 117) ((T_YIELD) reduce 117) ((T_YIELD_FROM) reduce 117) (("[") reduce 117) (("`") reduce 117) (("{") reduce 117) (("~") reduce 117)) 29842 ((("!") reduce 115) (("\"") reduce 115) (("$") reduce 115) (($) reduce 115) (("(") reduce 115) (("+") reduce 115) (("-") reduce 115) ((";") reduce 115) (("@") reduce 115) ((T_ABSTRACT) reduce 115) ((T_ARRAY) reduce 115) ((T_ARRAY_CAST) reduce 115) ((T_ATTRIBUTE) reduce 115) ((T_BOOL_CAST) reduce 115) ((T_BREAK) reduce 115) ((T_CLASS) reduce 115) ((T_CLASS_C) reduce 115) ((T_CLONE) reduce 115) ((T_CONST) reduce 115) ((T_CONSTANT_ENCAPSED_STRING) reduce 115) ((T_CONTINUE) reduce 115) ((T_DEC) reduce 115) ((T_DECLARE) reduce 115) ((T_DIR) reduce 115) ((T_DNUMBER) reduce 115) ((T_DO) reduce 115) ((T_DOUBLE_CAST) reduce 115) ((T_ECHO) reduce 115) ((T_EMPTY) reduce 115) ((T_ENUM) reduce 115) ((T_EVAL) reduce 115) ((T_EXIT) reduce 115) ((T_FILE) reduce 115) ((T_FINAL) reduce 115) ((T_FN) reduce 115) ((T_FOR) reduce 115) ((T_FOREACH) reduce 115) ((T_FUNCTION) reduce 115) ((T_FUNC_C) reduce 115) ((T_GLOBAL) reduce 115) ((T_GOTO) reduce 115) ((T_HALT_COMPILER) reduce 115) ((T_IF) reduce 115) ((T_INC) reduce 115) ((T_INCLUDE) reduce 115) ((T_INCLUDE_ONCE) reduce 115) ((T_INLINE_HTML) reduce 115) ((T_INTERFACE) reduce 115) ((T_INT_CAST) reduce 115) ((T_ISSET) reduce 115) ((T_LINE) reduce 115) ((T_LIST) reduce 115) ((T_LNUMBER) reduce 115) ((T_MATCH) reduce 115) ((T_METHOD_C) reduce 115) ((T_NAMESPACE) reduce 115) ((T_NAME_FULLY_QUALIFIED) reduce 115) ((T_NAME_QUALIFIED) reduce 115) ((T_NAME_RELATIVE) reduce 115) ((T_NEW) reduce 115) ((T_NS_C) reduce 115) ((T_OBJECT_CAST) reduce 115) ((T_PRINT) reduce 115) ((T_PROPERTY_C) reduce 115) ((T_READONLY) reduce 115) ((T_REQUIRE) reduce 115) ((T_REQUIRE_ONCE) reduce 115) ((T_RETURN) reduce 115) ((T_START_HEREDOC) reduce 115) ((T_STATIC) reduce 115) ((T_STRING) reduce 115) ((T_STRING_CAST) reduce 115) ((T_SWITCH) reduce 115) ((T_THROW) reduce 115) ((T_TRAIT) reduce 115) ((T_TRAIT_C) reduce 115) ((T_TRY) reduce 115) ((T_UNSET) reduce 115) ((T_UNSET_CAST) reduce 115) ((T_USE) reduce 115) ((T_VARIABLE) reduce 115) ((T_WHILE) reduce 115) ((T_YIELD) reduce 115) ((T_YIELD_FROM) reduce 115) (("[") reduce 115) (("`") reduce 115) (("{") reduce 115) (("~") reduce 115)) 29844 (((T_CONST) shift) ((T_FUNCTION) shift) ((T_NAME_QUALIFIED) shift) ((T_STRING) shift)) 29845 (((T_NAME_QUALIFIED) reduce 121) ((T_STRING) reduce 121)) 29846 (((T_NAME_QUALIFIED) reduce 120) ((T_STRING) reduce 120)) 29847 (((",") reduce 127) (("}") reduce 127)) 29849 (((",") reduce 132) (("}") reduce 132)) 29851 (((",") reduce 133) (("}") reduce 133)) 29852 (((T_CONST) shift) ((T_FUNCTION) shift) ((T_NAME_QUALIFIED) shift) ((T_STRING) shift) (("}") reduce 125)) 29854 (((";") reduce 123)) 29855 (((",") reduce 126) (("}") reduce 126)) 29861 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (($) reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONST) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAMESPACE) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_USE) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 29864 ((("!") reduce 169) (("\"") reduce 169) (($) reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONST) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAMESPACE) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_USE) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 29865 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (($) reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONST) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAMESPACE) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_USE) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 29868 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (($) reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONST) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAMESPACE) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_USE) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 29871 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (($) reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONST) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_EMPTY) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAMESPACE) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_USE) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("~") reduce 176)) 29878 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (($) reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONST) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAMESPACE) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_USE) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 29883 ((("!") reduce 198) (("\"") reduce 198) (($) reduce 198) (("$") reduce 198) (("(") reduce 198) (("+") reduce 198) (("-") reduce 198) ((";") reduce 198) (("@") reduce 198) ((T_ABSTRACT) reduce 198) ((T_ARRAY) reduce 198) ((T_ARRAY_CAST) reduce 198) ((T_ATTRIBUTE) reduce 198) ((T_BOOL_CAST) reduce 198) ((T_BREAK) reduce 198) ((T_CLASS) reduce 198) ((T_CLASS_C) reduce 198) ((T_CLONE) reduce 198) ((T_CONST) reduce 198) ((T_CONSTANT_ENCAPSED_STRING) reduce 198) ((T_CONTINUE) reduce 198) ((T_DEC) reduce 198) ((T_DECLARE) reduce 198) ((T_DIR) reduce 198) ((T_DNUMBER) reduce 198) ((T_DO) reduce 198) ((T_DOUBLE_CAST) reduce 198) ((T_ECHO) reduce 198) ((T_EMPTY) reduce 198) ((T_ENUM) reduce 198) ((T_EVAL) reduce 198) ((T_EXIT) reduce 198) ((T_FILE) reduce 198) ((T_FINAL) reduce 198) ((T_FN) reduce 198) ((T_FOR) reduce 198) ((T_FOREACH) reduce 198) ((T_FUNCTION) reduce 198) ((T_FUNC_C) reduce 198) ((T_GLOBAL) reduce 198) ((T_GOTO) reduce 198) ((T_HALT_COMPILER) reduce 198) ((T_IF) reduce 198) ((T_INC) reduce 198) ((T_INCLUDE) reduce 198) ((T_INCLUDE_ONCE) reduce 198) ((T_INLINE_HTML) reduce 198) ((T_INTERFACE) reduce 198) ((T_INT_CAST) reduce 198) ((T_ISSET) reduce 198) ((T_LINE) reduce 198) ((T_LIST) reduce 198) ((T_LNUMBER) reduce 198) ((T_MATCH) reduce 198) ((T_METHOD_C) reduce 198) ((T_NAMESPACE) reduce 198) ((T_NAME_FULLY_QUALIFIED) reduce 198) ((T_NAME_QUALIFIED) reduce 198) ((T_NAME_RELATIVE) reduce 198) ((T_NEW) reduce 198) ((T_NS_C) reduce 198) ((T_OBJECT_CAST) reduce 198) ((T_PRINT) reduce 198) ((T_PROPERTY_C) reduce 198) ((T_READONLY) reduce 198) ((T_REQUIRE) reduce 198) ((T_REQUIRE_ONCE) reduce 198) ((T_RETURN) reduce 198) ((T_START_HEREDOC) reduce 198) ((T_STATIC) reduce 198) ((T_STRING) reduce 198) ((T_STRING_CAST) reduce 198) ((T_SWITCH) reduce 198) ((T_THROW) reduce 198) ((T_TRAIT) reduce 198) ((T_TRAIT_C) reduce 198) ((T_TRY) reduce 198) ((T_UNSET) reduce 198) ((T_UNSET_CAST) reduce 198) ((T_USE) reduce 198) ((T_VARIABLE) reduce 198) ((T_WHILE) reduce 198) ((T_YIELD) reduce 198) ((T_YIELD_FROM) reduce 198) (("[") reduce 198) (("`") reduce 198) (("{") reduce 198) (("~") reduce 198)) 29889 ((("!") reduce 152) (("\"") reduce 152) (($) reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONST) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAMESPACE) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_USE) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 29893 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (($) reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONST) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAMESPACE) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_USE) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 29895 ((("!") reduce 223) (("\"") reduce 223) (($) reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONST) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAMESPACE) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_USE) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 29899 ((("!") reduce 224) (("\"") reduce 224) (($) reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONST) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAMESPACE) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_USE) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 29902 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (($) reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONST) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAMESPACE) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_USE) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 29904 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (($) reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONST) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAMESPACE) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_USE) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 29906 ((("!") reduce 157) (("\"") reduce 157) (($) reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONST) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAMESPACE) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_USE) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 29913 ((("!") reduce 155) (("\"") reduce 155) (($) reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONST) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAMESPACE) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_USE) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 29955 (((";") reduce 74) (("{") reduce 74)) 29956 (((";") reduce 52) (("{") reduce 52)) 29957 (((";") reduce 27) (("{") reduce 27)) 29958 (((";") reduce 51) (("{") reduce 51)) 29959 (((";") reduce 53) (("{") reduce 53)) 29960 (((";") reduce 49) (("{") reduce 49)) 29961 (((";") reduce 29) (("{") reduce 29)) 29962 (((";") reduce 59) (("{") reduce 59)) 29963 (((";") reduce 60) (("{") reduce 60)) 29964 (((";") reduce 11) (("{") reduce 11)) 29965 (((";") reduce 42) (("{") reduce 42)) 29966 (((";") reduce 39) (("{") reduce 39)) 29967 (((";") reduce 25) (("{") reduce 25)) 29968 (((";") reduce 50) (("{") reduce 50)) 29969 (((";") reduce 66) (("{") reduce 66)) 29970 (((";") reduce 18) (("{") reduce 18)) 29971 (((";") reduce 17) (("{") reduce 17)) 29972 (((";") reduce 15) (("{") reduce 15)) 29973 (((";") reduce 14) (("{") reduce 14)) 29974 (((";") reduce 38) (("{") reduce 38)) 29975 (((";") reduce 26) (("{") reduce 26)) 29976 (((";") reduce 22) (("{") reduce 22)) 29977 (((";") reduce 24) (("{") reduce 24)) 29978 (((";") reduce 16) (("{") reduce 16)) 29979 (((";") reduce 48) (("{") reduce 48)) 29980 (((";") reduce 20) (("{") reduce 20)) 29981 (((";") reduce 70) (("{") reduce 70)) 29982 (((";") reduce 3) (("{") reduce 3)) 29983 (((";") reduce 12) (("{") reduce 12)) 29984 (((";") reduce 54) (("{") reduce 54)) 29985 (((";") reduce 65) (("{") reduce 65)) 29986 (((";") reduce 75) (("{") reduce 75)) 29987 (((";") reduce 30) (("{") reduce 30)) 29988 (((";") reduce 68) (("{") reduce 68)) 29989 (((";") reduce 21) (("{") reduce 21)) 29990 (((";") reduce 23) (("{") reduce 23)) 29991 (((";") reduce 41) (("{") reduce 41)) 29992 (((";") reduce 62) (("{") reduce 62)) 29993 (((";") reduce 34) (("{") reduce 34)) 29994 (((";") reduce 40) (("{") reduce 40)) 29995 (((";") reduce 13) (("{") reduce 13)) 29996 (((";") reduce 55) (("{") reduce 55)) 29997 (((";") reduce 1) (("{") reduce 1)) 29998 (((";") reduce 2) (("{") reduce 2)) 29999 (((";") reduce 9) (("{") reduce 9)) 30000 (((";") reduce 33) (("{") reduce 33)) 30001 (((";") reduce 58) (("{") reduce 58)) 30002 (((";") reduce 37) (("{") reduce 37)) 30003 (((";") reduce 64) (("{") reduce 64)) 30004 (((";") reduce 46) (("{") reduce 46)) 30005 (((";") reduce 8) (("{") reduce 8)) 30006 (((";") reduce 6) (("{") reduce 6)) 30007 (((";") reduce 7) (("{") reduce 7)) 30008 (((";") reduce 69) (("{") reduce 69)) 30009 (((";") reduce 63) (("{") reduce 63)) 30010 (((";") reduce 56) (("{") reduce 56)) 30011 (((";") reduce 87) (("{") reduce 87)) 30012 (((";") reduce 10) (("{") reduce 10)) 30013 (((";") reduce 67) (("{") reduce 67)) 30014 (((";") reduce 44) (("{") reduce 44)) 30015 (((";") reduce 76) (("{") reduce 76)) 30016 (((";") reduce 71) (("{") reduce 71)) 30017 (((";") reduce 77) (("{") reduce 77)) 30018 (((";") reduce 78) (("{") reduce 78)) 30019 (((";") reduce 79) (("{") reduce 79)) 30020 (((";") reduce 4) (("{") reduce 4)) 30021 (((";") reduce 5) (("{") reduce 5)) 30022 (((";") reduce 43) (("{") reduce 43)) 30023 (((";") reduce 73) (("{") reduce 73)) 30024 (((";") reduce 82) (("{") reduce 82)) 30025 (((";") reduce 47) (("{") reduce 47)) 30026 (((";") reduce 31) (("{") reduce 31)) 30027 (((";") reduce 57) (("{") reduce 57)) 30028 (((";") reduce 61) (("{") reduce 61)) 30029 (((";") reduce 28) (("{") reduce 28)) 30030 (((";") reduce 36) (("{") reduce 36)) 30031 (((";") reduce 32) (("{") reduce 32)) 30032 (((";") reduce 35) (("{") reduce 35)) 30033 (((";") reduce 19) (("{") reduce 19)) 30034 (((";") reduce 45) (("{") reduce 45)) 30035 (((";") reduce 86) (("{") reduce 86)) 30037 (((";") reduce 72) (("{") reduce 72)) 30038 (((";") reduce 83) (("{") reduce 83)) 30039 ((("!") reduce 85) (("\"") reduce 85) (("$") reduce 85) (("(") reduce 85) (("+") reduce 85) (("-") reduce 85) ((";") reduce 85) (("@") reduce 85) ((T_ABSTRACT) reduce 85) ((T_ARRAY) reduce 85) ((T_ARRAY_CAST) reduce 85) ((T_ATTRIBUTE) reduce 85) ((T_BOOL_CAST) reduce 85) ((T_BREAK) reduce 85) ((T_CLASS) reduce 85) ((T_CLASS_C) reduce 85) ((T_CLONE) reduce 85) ((T_CONST) reduce 85) ((T_CONSTANT_ENCAPSED_STRING) reduce 85) ((T_CONTINUE) reduce 85) ((T_DEC) reduce 85) ((T_DECLARE) reduce 85) ((T_DIR) reduce 85) ((T_DNUMBER) reduce 85) ((T_DO) reduce 85) ((T_DOUBLE_CAST) reduce 85) ((T_ECHO) reduce 85) ((T_EMPTY) reduce 85) ((T_ENUM) reduce 85) ((T_EVAL) reduce 85) ((T_EXIT) reduce 85) ((T_FILE) reduce 85) ((T_FINAL) reduce 85) ((T_FN) reduce 85) ((T_FOR) reduce 85) ((T_FOREACH) reduce 85) ((T_FUNCTION) reduce 85) ((T_FUNC_C) reduce 85) ((T_GLOBAL) reduce 85) ((T_GOTO) reduce 85) ((T_HALT_COMPILER) reduce 85) ((T_IF) reduce 85) ((T_INC) reduce 85) ((T_INCLUDE) reduce 85) ((T_INCLUDE_ONCE) reduce 85) ((T_INLINE_HTML) reduce 85) ((T_INTERFACE) reduce 85) ((T_INT_CAST) reduce 85) ((T_ISSET) reduce 85) ((T_LINE) reduce 85) ((T_LIST) reduce 85) ((T_LNUMBER) reduce 85) ((T_MATCH) reduce 85) ((T_METHOD_C) reduce 85) ((T_NAMESPACE) reduce 85) ((T_NAME_FULLY_QUALIFIED) reduce 85) ((T_NAME_QUALIFIED) reduce 85) ((T_NAME_RELATIVE) reduce 85) ((T_NEW) reduce 85) ((T_NS_C) reduce 85) ((T_OBJECT_CAST) reduce 85) ((T_PRINT) reduce 85) ((T_PROPERTY_C) reduce 85) ((T_READONLY) reduce 85) ((T_REQUIRE) reduce 85) ((T_REQUIRE_ONCE) reduce 85) ((T_RETURN) reduce 85) ((T_START_HEREDOC) reduce 85) ((T_STATIC) reduce 85) ((T_STRING) reduce 85) ((T_STRING_CAST) reduce 85) ((T_SWITCH) reduce 85) ((T_THROW) reduce 85) ((T_TRAIT) reduce 85) ((T_TRAIT_C) reduce 85) ((T_TRY) reduce 85) ((T_UNSET) reduce 85) ((T_UNSET_CAST) reduce 85) ((T_USE) reduce 85) ((T_VARIABLE) reduce 85) ((T_WHILE) reduce 85) ((T_YIELD) reduce 85) ((T_YIELD_FROM) reduce 85) (("[") reduce 85) (("`") reduce 85) (("{") reduce 85) (("}") reduce 85) (("~") reduce 85)) 30040 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_USE) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("}") shift) (("~") shift)) 30041 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONST) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAMESPACE) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_USE) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 30056 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONST) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAMESPACE) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_USE) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 30068 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONST) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAMESPACE) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_USE) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 30070 ((("!") reduce 109) (("\"") reduce 109) (("$") reduce 109) (("(") reduce 109) (("+") reduce 109) (("-") reduce 109) ((";") reduce 109) (("@") reduce 109) ((T_ABSTRACT) reduce 109) ((T_ARRAY) reduce 109) ((T_ARRAY_CAST) reduce 109) ((T_ATTRIBUTE) reduce 109) ((T_BOOL_CAST) reduce 109) ((T_BREAK) reduce 109) ((T_CLASS) reduce 109) ((T_CLASS_C) reduce 109) ((T_CLONE) reduce 109) ((T_CONST) reduce 109) ((T_CONSTANT_ENCAPSED_STRING) reduce 109) ((T_CONTINUE) reduce 109) ((T_DEC) reduce 109) ((T_DECLARE) reduce 109) ((T_DIR) reduce 109) ((T_DNUMBER) reduce 109) ((T_DO) reduce 109) ((T_DOUBLE_CAST) reduce 109) ((T_ECHO) reduce 109) ((T_EMPTY) reduce 109) ((T_ENUM) reduce 109) ((T_EVAL) reduce 109) ((T_EXIT) reduce 109) ((T_FILE) reduce 109) ((T_FINAL) reduce 109) ((T_FN) reduce 109) ((T_FOR) reduce 109) ((T_FOREACH) reduce 109) ((T_FUNCTION) reduce 109) ((T_FUNC_C) reduce 109) ((T_GLOBAL) reduce 109) ((T_GOTO) reduce 109) ((T_HALT_COMPILER) reduce 109) ((T_IF) reduce 109) ((T_INC) reduce 109) ((T_INCLUDE) reduce 109) ((T_INCLUDE_ONCE) reduce 109) ((T_INLINE_HTML) reduce 109) ((T_INTERFACE) reduce 109) ((T_INT_CAST) reduce 109) ((T_ISSET) reduce 109) ((T_LINE) reduce 109) ((T_LIST) reduce 109) ((T_LNUMBER) reduce 109) ((T_MATCH) reduce 109) ((T_METHOD_C) reduce 109) ((T_NAMESPACE) reduce 109) ((T_NAME_FULLY_QUALIFIED) reduce 109) ((T_NAME_QUALIFIED) reduce 109) ((T_NAME_RELATIVE) reduce 109) ((T_NEW) reduce 109) ((T_NS_C) reduce 109) ((T_OBJECT_CAST) reduce 109) ((T_PRINT) reduce 109) ((T_PROPERTY_C) reduce 109) ((T_READONLY) reduce 109) ((T_REQUIRE) reduce 109) ((T_REQUIRE_ONCE) reduce 109) ((T_RETURN) reduce 109) ((T_START_HEREDOC) reduce 109) ((T_STATIC) reduce 109) ((T_STRING) reduce 109) ((T_STRING_CAST) reduce 109) ((T_SWITCH) reduce 109) ((T_THROW) reduce 109) ((T_TRAIT) reduce 109) ((T_TRAIT_C) reduce 109) ((T_TRY) reduce 109) ((T_UNSET) reduce 109) ((T_UNSET_CAST) reduce 109) ((T_USE) reduce 109) ((T_VARIABLE) reduce 109) ((T_WHILE) reduce 109) ((T_YIELD) reduce 109) ((T_YIELD_FROM) reduce 109) (("[") reduce 109) (("`") reduce 109) (("{") reduce 109) (("}") reduce 109) (("~") reduce 109)) 30072 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONST) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAMESPACE) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_USE) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("}") reduce 104) (("~") reduce 104)) 30074 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONST) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAMESPACE) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_USE) reduce 107) ((T_VARIABLE) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("}") reduce 107) (("~") reduce 107)) 30077 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONST) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAMESPACE) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_USE) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("}") reduce 103) (("~") reduce 103)) 30078 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONST) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAMESPACE) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_USE) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 30079 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONST) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAMESPACE) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_USE) reduce 244) ((T_VARIABLE) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("}") reduce 244) (("~") reduce 244)) 30080 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONST) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAMESPACE) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_USE) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("}") reduce 106) (("~") reduce 106)) 30081 ((("!") reduce 108) (("\"") reduce 108) (("$") reduce 108) (("(") reduce 108) (("+") reduce 108) (("-") reduce 108) ((";") reduce 108) (("@") reduce 108) ((T_ABSTRACT) reduce 108) ((T_ARRAY) reduce 108) ((T_ARRAY_CAST) reduce 108) ((T_ATTRIBUTE) reduce 108) ((T_BOOL_CAST) reduce 108) ((T_BREAK) reduce 108) ((T_CLASS) reduce 108) ((T_CLASS_C) reduce 108) ((T_CLONE) reduce 108) ((T_CONST) reduce 108) ((T_CONSTANT_ENCAPSED_STRING) reduce 108) ((T_CONTINUE) reduce 108) ((T_DEC) reduce 108) ((T_DECLARE) reduce 108) ((T_DIR) reduce 108) ((T_DNUMBER) reduce 108) ((T_DO) reduce 108) ((T_DOUBLE_CAST) reduce 108) ((T_ECHO) reduce 108) ((T_EMPTY) reduce 108) ((T_ENUM) reduce 108) ((T_EVAL) reduce 108) ((T_EXIT) reduce 108) ((T_FILE) reduce 108) ((T_FINAL) reduce 108) ((T_FN) reduce 108) ((T_FOR) reduce 108) ((T_FOREACH) reduce 108) ((T_FUNCTION) reduce 108) ((T_FUNC_C) reduce 108) ((T_GLOBAL) reduce 108) ((T_GOTO) reduce 108) ((T_HALT_COMPILER) reduce 108) ((T_IF) reduce 108) ((T_INC) reduce 108) ((T_INCLUDE) reduce 108) ((T_INCLUDE_ONCE) reduce 108) ((T_INLINE_HTML) reduce 108) ((T_INTERFACE) reduce 108) ((T_INT_CAST) reduce 108) ((T_ISSET) reduce 108) ((T_LINE) reduce 108) ((T_LIST) reduce 108) ((T_LNUMBER) reduce 108) ((T_MATCH) reduce 108) ((T_METHOD_C) reduce 108) ((T_NAMESPACE) reduce 108) ((T_NAME_FULLY_QUALIFIED) reduce 108) ((T_NAME_QUALIFIED) reduce 108) ((T_NAME_RELATIVE) reduce 108) ((T_NEW) reduce 108) ((T_NS_C) reduce 108) ((T_OBJECT_CAST) reduce 108) ((T_PRINT) reduce 108) ((T_PROPERTY_C) reduce 108) ((T_READONLY) reduce 108) ((T_REQUIRE) reduce 108) ((T_REQUIRE_ONCE) reduce 108) ((T_RETURN) reduce 108) ((T_START_HEREDOC) reduce 108) ((T_STATIC) reduce 108) ((T_STRING) reduce 108) ((T_STRING_CAST) reduce 108) ((T_SWITCH) reduce 108) ((T_THROW) reduce 108) ((T_TRAIT) reduce 108) ((T_TRAIT_C) reduce 108) ((T_TRY) reduce 108) ((T_UNSET) reduce 108) ((T_UNSET_CAST) reduce 108) ((T_USE) reduce 108) ((T_VARIABLE) reduce 108) ((T_WHILE) reduce 108) ((T_YIELD) reduce 108) ((T_YIELD_FROM) reduce 108) (("[") reduce 108) (("`") reduce 108) (("{") reduce 108) (("}") reduce 108) (("~") reduce 108)) 30082 ((("!") reduce 84) (("\"") reduce 84) (("$") reduce 84) (("(") reduce 84) (("+") reduce 84) (("-") reduce 84) ((";") reduce 84) (("@") reduce 84) ((T_ABSTRACT) reduce 84) ((T_ARRAY) reduce 84) ((T_ARRAY_CAST) reduce 84) ((T_ATTRIBUTE) reduce 84) ((T_BOOL_CAST) reduce 84) ((T_BREAK) reduce 84) ((T_CLASS) reduce 84) ((T_CLASS_C) reduce 84) ((T_CLONE) reduce 84) ((T_CONST) reduce 84) ((T_CONSTANT_ENCAPSED_STRING) reduce 84) ((T_CONTINUE) reduce 84) ((T_DEC) reduce 84) ((T_DECLARE) reduce 84) ((T_DIR) reduce 84) ((T_DNUMBER) reduce 84) ((T_DO) reduce 84) ((T_DOUBLE_CAST) reduce 84) ((T_ECHO) reduce 84) ((T_EMPTY) reduce 84) ((T_ENUM) reduce 84) ((T_EVAL) reduce 84) ((T_EXIT) reduce 84) ((T_FILE) reduce 84) ((T_FINAL) reduce 84) ((T_FN) reduce 84) ((T_FOR) reduce 84) ((T_FOREACH) reduce 84) ((T_FUNCTION) reduce 84) ((T_FUNC_C) reduce 84) ((T_GLOBAL) reduce 84) ((T_GOTO) reduce 84) ((T_HALT_COMPILER) reduce 84) ((T_IF) reduce 84) ((T_INC) reduce 84) ((T_INCLUDE) reduce 84) ((T_INCLUDE_ONCE) reduce 84) ((T_INLINE_HTML) reduce 84) ((T_INTERFACE) reduce 84) ((T_INT_CAST) reduce 84) ((T_ISSET) reduce 84) ((T_LINE) reduce 84) ((T_LIST) reduce 84) ((T_LNUMBER) reduce 84) ((T_MATCH) reduce 84) ((T_METHOD_C) reduce 84) ((T_NAMESPACE) reduce 84) ((T_NAME_FULLY_QUALIFIED) reduce 84) ((T_NAME_QUALIFIED) reduce 84) ((T_NAME_RELATIVE) reduce 84) ((T_NEW) reduce 84) ((T_NS_C) reduce 84) ((T_OBJECT_CAST) reduce 84) ((T_PRINT) reduce 84) ((T_PROPERTY_C) reduce 84) ((T_READONLY) reduce 84) ((T_REQUIRE) reduce 84) ((T_REQUIRE_ONCE) reduce 84) ((T_RETURN) reduce 84) ((T_START_HEREDOC) reduce 84) ((T_STATIC) reduce 84) ((T_STRING) reduce 84) ((T_STRING_CAST) reduce 84) ((T_SWITCH) reduce 84) ((T_THROW) reduce 84) ((T_TRAIT) reduce 84) ((T_TRAIT_C) reduce 84) ((T_TRY) reduce 84) ((T_UNSET) reduce 84) ((T_UNSET_CAST) reduce 84) ((T_USE) reduce 84) ((T_VARIABLE) reduce 84) ((T_WHILE) reduce 84) ((T_YIELD) reduce 84) ((T_YIELD_FROM) reduce 84) (("[") reduce 84) (("`") reduce 84) (("{") reduce 84) (("}") reduce 84) (("~") reduce 84)) 30083 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONST) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAMESPACE) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_USE) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("}") reduce 105) (("~") reduce 105)) 30085 ((("!") reduce 114) (("\"") reduce 114) (($) reduce 114) (("$") reduce 114) (("(") reduce 114) (("+") reduce 114) (("-") reduce 114) ((";") reduce 114) (("@") reduce 114) ((T_ABSTRACT) reduce 114) ((T_ARRAY) reduce 114) ((T_ARRAY_CAST) reduce 114) ((T_ATTRIBUTE) reduce 114) ((T_BOOL_CAST) reduce 114) ((T_BREAK) reduce 114) ((T_CLASS) reduce 114) ((T_CLASS_C) reduce 114) ((T_CLONE) reduce 114) ((T_CONST) reduce 114) ((T_CONSTANT_ENCAPSED_STRING) reduce 114) ((T_CONTINUE) reduce 114) ((T_DEC) reduce 114) ((T_DECLARE) reduce 114) ((T_DIR) reduce 114) ((T_DNUMBER) reduce 114) ((T_DO) reduce 114) ((T_DOUBLE_CAST) reduce 114) ((T_ECHO) reduce 114) ((T_EMPTY) reduce 114) ((T_ENUM) reduce 114) ((T_EVAL) reduce 114) ((T_EXIT) reduce 114) ((T_FILE) reduce 114) ((T_FINAL) reduce 114) ((T_FN) reduce 114) ((T_FOR) reduce 114) ((T_FOREACH) reduce 114) ((T_FUNCTION) reduce 114) ((T_FUNC_C) reduce 114) ((T_GLOBAL) reduce 114) ((T_GOTO) reduce 114) ((T_HALT_COMPILER) reduce 114) ((T_IF) reduce 114) ((T_INC) reduce 114) ((T_INCLUDE) reduce 114) ((T_INCLUDE_ONCE) reduce 114) ((T_INLINE_HTML) reduce 114) ((T_INTERFACE) reduce 114) ((T_INT_CAST) reduce 114) ((T_ISSET) reduce 114) ((T_LINE) reduce 114) ((T_LIST) reduce 114) ((T_LNUMBER) reduce 114) ((T_MATCH) reduce 114) ((T_METHOD_C) reduce 114) ((T_NAMESPACE) reduce 114) ((T_NAME_FULLY_QUALIFIED) reduce 114) ((T_NAME_QUALIFIED) reduce 114) ((T_NAME_RELATIVE) reduce 114) ((T_NEW) reduce 114) ((T_NS_C) reduce 114) ((T_OBJECT_CAST) reduce 114) ((T_PRINT) reduce 114) ((T_PROPERTY_C) reduce 114) ((T_READONLY) reduce 114) ((T_REQUIRE) reduce 114) ((T_REQUIRE_ONCE) reduce 114) ((T_RETURN) reduce 114) ((T_START_HEREDOC) reduce 114) ((T_STATIC) reduce 114) ((T_STRING) reduce 114) ((T_STRING_CAST) reduce 114) ((T_SWITCH) reduce 114) ((T_THROW) reduce 114) ((T_TRAIT) reduce 114) ((T_TRAIT_C) reduce 114) ((T_TRY) reduce 114) ((T_UNSET) reduce 114) ((T_UNSET_CAST) reduce 114) ((T_USE) reduce 114) ((T_VARIABLE) reduce 114) ((T_WHILE) reduce 114) ((T_YIELD) reduce 114) ((T_YIELD_FROM) reduce 114) (("[") reduce 114) (("`") reduce 114) (("{") reduce 114) (("~") reduce 114)) 30087 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONST) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAMESPACE) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_USE) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("}") reduce 146) (("~") reduce 146)) 30093 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONST) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAMESPACE) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_USE) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 30103 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONST) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAMESPACE) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_USE) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 30111 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONST) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAMESPACE) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_USE) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 30114 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONST) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAMESPACE) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_USE) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 30116 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONST) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAMESPACE) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_USE) reduce 243) ((T_VARIABLE) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("}") reduce 243) (("~") reduce 243)) 30119 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONST) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAMESPACE) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_USE) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("}") reduce 146) (("~") reduce 146)) 30121 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONST) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAMESPACE) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_USE) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("}") reduce 245) (("~") reduce 245)) 30122 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONST) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAMESPACE) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_USE) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 30125 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONST) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_ELSE) reduce 248) ((T_ELSEIF) reduce 248) ((T_EMPTY) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAMESPACE) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_USE) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("}") reduce 248) (("~") reduce 248)) 30129 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONST) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_ELSE) reduce 249) ((T_ELSEIF) reduce 249) ((T_EMPTY) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAMESPACE) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_USE) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("}") reduce 249) (("~") reduce 249)) 30134 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONST) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAMESPACE) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_USE) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("}") reduce 240) (("~") reduce 240)) 30135 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONST) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAMESPACE) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_USE) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 30138 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONST) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAMESPACE) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_USE) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("}") reduce 241) (("~") reduce 241)) 30143 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONST) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAMESPACE) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_USE) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 30146 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONST) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAMESPACE) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_USE) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 30147 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONST) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAMESPACE) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_USE) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("}") reduce 175) (("~") reduce 175)) 30150 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONST) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAMESPACE) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_USE) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 30153 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONST) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_ELSE) reduce 176) ((T_ELSEIF) reduce 176) ((T_EMPTY) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAMESPACE) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_USE) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("}") reduce 176) (("~") reduce 176)) 30160 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONST) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAMESPACE) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_USE) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("}") reduce 170) (("~") reduce 170)) 30165 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONST) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAMESPACE) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_USE) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 30169 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONST) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAMESPACE) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_USE) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("}") reduce 222) (("~") reduce 222)) 30171 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONST) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAMESPACE) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_USE) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("}") reduce 223) (("~") reduce 223)) 30175 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONST) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAMESPACE) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_USE) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("}") reduce 224) (("~") reduce 224)) 30178 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONST) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAMESPACE) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_USE) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("}") reduce 225) (("~") reduce 225)) 30179 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONST) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAMESPACE) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_USE) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 30181 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONST) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAMESPACE) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_USE) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 30183 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONST) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAMESPACE) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_USE) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 30185 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONST) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAMESPACE) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_USE) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 30187 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONST) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAMESPACE) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_USE) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 30197 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONST) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAMESPACE) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_USE) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 30198 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONST) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAMESPACE) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_USE) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("}") reduce 218) (("~") reduce 218)) 30201 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONST) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAMESPACE) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_USE) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("}") reduce 219) (("~") reduce 219)) 30202 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONST) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAMESPACE) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_USE) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 30211 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONST) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAMESPACE) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_USE) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 30212 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONST) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAMESPACE) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_USE) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("}") reduce 216) (("~") reduce 216)) 30215 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONST) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAMESPACE) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_USE) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("}") reduce 217) (("~") reduce 217)) 30217 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONST) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAMESPACE) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_USE) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 30223 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONST) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAMESPACE) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_USE) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 30228 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONST) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAMESPACE) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_USE) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 30229 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONST) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAMESPACE) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_USE) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("}") reduce 220) (("~") reduce 220)) 30232 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONST) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAMESPACE) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_USE) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("}") reduce 221) (("~") reduce 221)) 30234 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONST) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAMESPACE) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_USE) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 30236 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONST) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAMESPACE) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_USE) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 30237 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONST) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_EMPTY) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAMESPACE) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_USE) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("}") reduce 245) (("~") reduce 245)) 30248 ((("!") reduce 477) (("\"") reduce 477) (("$") reduce 477) (("(") reduce 477) (("+") reduce 477) (("-") reduce 477) ((";") reduce 477) (("@") reduce 477) ((T_ABSTRACT) reduce 477) ((T_ARRAY) reduce 477) ((T_ARRAY_CAST) reduce 477) ((T_ATTRIBUTE) reduce 477) ((T_BOOL_CAST) reduce 477) ((T_BREAK) reduce 477) ((T_CLASS) reduce 477) ((T_CLASS_C) reduce 477) ((T_CLONE) reduce 477) ((T_CONST) reduce 477) ((T_CONSTANT_ENCAPSED_STRING) reduce 477) ((T_CONTINUE) reduce 477) ((T_DEC) reduce 477) ((T_DECLARE) reduce 477) ((T_DIR) reduce 477) ((T_DNUMBER) reduce 477) ((T_DO) reduce 477) ((T_DOUBLE_CAST) reduce 477) ((T_ECHO) reduce 477) ((T_EMPTY) reduce 477) ((T_ENUM) reduce 477) ((T_EVAL) reduce 477) ((T_EXIT) reduce 477) ((T_FILE) reduce 477) ((T_FINAL) reduce 477) ((T_FN) reduce 477) ((T_FOR) reduce 477) ((T_FOREACH) reduce 477) ((T_FUNCTION) reduce 477) ((T_FUNC_C) reduce 477) ((T_GLOBAL) reduce 477) ((T_GOTO) reduce 477) ((T_HALT_COMPILER) reduce 477) ((T_IF) reduce 477) ((T_INC) reduce 477) ((T_INCLUDE) reduce 477) ((T_INCLUDE_ONCE) reduce 477) ((T_INLINE_HTML) reduce 477) ((T_INTERFACE) reduce 477) ((T_INT_CAST) reduce 477) ((T_ISSET) reduce 477) ((T_LINE) reduce 477) ((T_LIST) reduce 477) ((T_LNUMBER) reduce 477) ((T_MATCH) reduce 477) ((T_METHOD_C) reduce 477) ((T_NAMESPACE) reduce 477) ((T_NAME_FULLY_QUALIFIED) reduce 477) ((T_NAME_QUALIFIED) reduce 477) ((T_NAME_RELATIVE) reduce 477) ((T_NEW) reduce 477) ((T_NS_C) reduce 477) ((T_OBJECT_CAST) reduce 477) ((T_PRINT) reduce 477) ((T_PROPERTY_C) reduce 477) ((T_READONLY) reduce 477) ((T_REQUIRE) reduce 477) ((T_REQUIRE_ONCE) reduce 477) ((T_RETURN) reduce 477) ((T_START_HEREDOC) reduce 477) ((T_STATIC) reduce 477) ((T_STRING) reduce 477) ((T_STRING_CAST) reduce 477) ((T_SWITCH) reduce 477) ((T_THROW) reduce 477) ((T_TRAIT) reduce 477) ((T_TRAIT_C) reduce 477) ((T_TRY) reduce 477) ((T_UNSET) reduce 477) ((T_UNSET_CAST) reduce 477) ((T_USE) reduce 477) ((T_VARIABLE) reduce 477) ((T_WHILE) reduce 477) ((T_YIELD) reduce 477) ((T_YIELD_FROM) reduce 477) (("[") reduce 477) (("`") reduce 477) (("{") reduce 477) (("}") reduce 477) (("~") reduce 477)) 30249 ((("!") reduce 182) (("\"") reduce 182) (("$") reduce 182) (("(") reduce 182) (("+") reduce 182) (("-") reduce 182) ((";") reduce 182) (("@") reduce 182) ((T_ABSTRACT) reduce 182) ((T_ARRAY) reduce 182) ((T_ARRAY_CAST) reduce 182) ((T_ATTRIBUTE) reduce 182) ((T_BOOL_CAST) reduce 182) ((T_BREAK) reduce 182) ((T_CLASS) reduce 182) ((T_CLASS_C) reduce 182) ((T_CLONE) reduce 182) ((T_CONST) reduce 182) ((T_CONSTANT_ENCAPSED_STRING) reduce 182) ((T_CONTINUE) reduce 182) ((T_DEC) reduce 182) ((T_DECLARE) reduce 182) ((T_DIR) reduce 182) ((T_DNUMBER) reduce 182) ((T_DO) reduce 182) ((T_DOUBLE_CAST) reduce 182) ((T_ECHO) reduce 182) ((T_EMPTY) reduce 182) ((T_ENUM) reduce 182) ((T_EVAL) reduce 182) ((T_EXIT) reduce 182) ((T_FILE) reduce 182) ((T_FINAL) reduce 182) ((T_FN) reduce 182) ((T_FOR) reduce 182) ((T_FOREACH) reduce 182) ((T_FUNCTION) reduce 182) ((T_FUNC_C) reduce 182) ((T_GLOBAL) reduce 182) ((T_GOTO) reduce 182) ((T_HALT_COMPILER) reduce 182) ((T_IF) reduce 182) ((T_INC) reduce 182) ((T_INCLUDE) reduce 182) ((T_INCLUDE_ONCE) reduce 182) ((T_INLINE_HTML) reduce 182) ((T_INTERFACE) reduce 182) ((T_INT_CAST) reduce 182) ((T_ISSET) reduce 182) ((T_LINE) reduce 182) ((T_LIST) reduce 182) ((T_LNUMBER) reduce 182) ((T_MATCH) reduce 182) ((T_METHOD_C) reduce 182) ((T_NAMESPACE) reduce 182) ((T_NAME_FULLY_QUALIFIED) reduce 182) ((T_NAME_QUALIFIED) reduce 182) ((T_NAME_RELATIVE) reduce 182) ((T_NEW) reduce 182) ((T_NS_C) reduce 182) ((T_OBJECT_CAST) reduce 182) ((T_PRINT) reduce 182) ((T_PROPERTY_C) reduce 182) ((T_READONLY) reduce 182) ((T_REQUIRE) reduce 182) ((T_REQUIRE_ONCE) reduce 182) ((T_RETURN) reduce 182) ((T_START_HEREDOC) reduce 182) ((T_STATIC) reduce 182) ((T_STRING) reduce 182) ((T_STRING_CAST) reduce 182) ((T_SWITCH) reduce 182) ((T_THROW) reduce 182) ((T_TRAIT) reduce 182) ((T_TRAIT_C) reduce 182) ((T_TRY) reduce 182) ((T_UNSET) reduce 182) ((T_UNSET_CAST) reduce 182) ((T_USE) reduce 182) ((T_VARIABLE) reduce 182) ((T_WHILE) reduce 182) ((T_YIELD) reduce 182) ((T_YIELD_FROM) reduce 182) (("[") reduce 182) (("`") reduce 182) (("{") reduce 182) (("}") reduce 182) (("~") reduce 182)) 30250 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONST) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAMESPACE) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_USE) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 30258 ((("!") reduce 187) (("\"") reduce 187) (("$") reduce 187) (("(") reduce 187) (("+") reduce 187) (("-") reduce 187) ((";") reduce 187) (("@") reduce 187) ((T_ABSTRACT) reduce 187) ((T_ARRAY) reduce 187) ((T_ARRAY_CAST) reduce 187) ((T_ATTRIBUTE) reduce 187) ((T_BOOL_CAST) reduce 187) ((T_BREAK) reduce 187) ((T_CLASS) reduce 187) ((T_CLASS_C) reduce 187) ((T_CLONE) reduce 187) ((T_CONST) reduce 187) ((T_CONSTANT_ENCAPSED_STRING) reduce 187) ((T_CONTINUE) reduce 187) ((T_DEC) reduce 187) ((T_DECLARE) reduce 187) ((T_DIR) reduce 187) ((T_DNUMBER) reduce 187) ((T_DO) reduce 187) ((T_DOUBLE_CAST) reduce 187) ((T_ECHO) reduce 187) ((T_EMPTY) reduce 187) ((T_ENUM) reduce 187) ((T_EVAL) reduce 187) ((T_EXIT) reduce 187) ((T_FILE) reduce 187) ((T_FINAL) reduce 187) ((T_FN) reduce 187) ((T_FOR) reduce 187) ((T_FOREACH) reduce 187) ((T_FUNCTION) reduce 187) ((T_FUNC_C) reduce 187) ((T_GLOBAL) reduce 187) ((T_GOTO) reduce 187) ((T_HALT_COMPILER) reduce 187) ((T_IF) reduce 187) ((T_INC) reduce 187) ((T_INCLUDE) reduce 187) ((T_INCLUDE_ONCE) reduce 187) ((T_INLINE_HTML) reduce 187) ((T_INTERFACE) reduce 187) ((T_INT_CAST) reduce 187) ((T_ISSET) reduce 187) ((T_LINE) reduce 187) ((T_LIST) reduce 187) ((T_LNUMBER) reduce 187) ((T_MATCH) reduce 187) ((T_METHOD_C) reduce 187) ((T_NAMESPACE) reduce 187) ((T_NAME_FULLY_QUALIFIED) reduce 187) ((T_NAME_QUALIFIED) reduce 187) ((T_NAME_RELATIVE) reduce 187) ((T_NEW) reduce 187) ((T_NS_C) reduce 187) ((T_OBJECT_CAST) reduce 187) ((T_PRINT) reduce 187) ((T_PROPERTY_C) reduce 187) ((T_READONLY) reduce 187) ((T_REQUIRE) reduce 187) ((T_REQUIRE_ONCE) reduce 187) ((T_RETURN) reduce 187) ((T_START_HEREDOC) reduce 187) ((T_STATIC) reduce 187) ((T_STRING) reduce 187) ((T_STRING_CAST) reduce 187) ((T_SWITCH) reduce 187) ((T_THROW) reduce 187) ((T_TRAIT) reduce 187) ((T_TRAIT_C) reduce 187) ((T_TRY) reduce 187) ((T_UNSET) reduce 187) ((T_UNSET_CAST) reduce 187) ((T_USE) reduce 187) ((T_VARIABLE) reduce 187) ((T_WHILE) reduce 187) ((T_YIELD) reduce 187) ((T_YIELD_FROM) reduce 187) (("[") reduce 187) (("`") reduce 187) (("{") reduce 187) (("}") reduce 187) (("~") reduce 187)) 30259 ((("!") reduce 110) (("\"") reduce 110) (("$") reduce 110) (("(") reduce 110) (("+") reduce 110) (("-") reduce 110) ((";") reduce 110) (("@") reduce 110) ((T_ABSTRACT) reduce 110) ((T_ARRAY) reduce 110) ((T_ARRAY_CAST) reduce 110) ((T_ATTRIBUTE) reduce 110) ((T_BOOL_CAST) reduce 110) ((T_BREAK) reduce 110) ((T_CLASS) reduce 110) ((T_CLASS_C) reduce 110) ((T_CLONE) reduce 110) ((T_CONST) reduce 110) ((T_CONSTANT_ENCAPSED_STRING) reduce 110) ((T_CONTINUE) reduce 110) ((T_DEC) reduce 110) ((T_DECLARE) reduce 110) ((T_DIR) reduce 110) ((T_DNUMBER) reduce 110) ((T_DO) reduce 110) ((T_DOUBLE_CAST) reduce 110) ((T_ECHO) reduce 110) ((T_EMPTY) reduce 110) ((T_ENUM) reduce 110) ((T_EVAL) reduce 110) ((T_EXIT) reduce 110) ((T_FILE) reduce 110) ((T_FINAL) reduce 110) ((T_FN) reduce 110) ((T_FOR) reduce 110) ((T_FOREACH) reduce 110) ((T_FUNCTION) reduce 110) ((T_FUNC_C) reduce 110) ((T_GLOBAL) reduce 110) ((T_GOTO) reduce 110) ((T_HALT_COMPILER) reduce 110) ((T_IF) reduce 110) ((T_INC) reduce 110) ((T_INCLUDE) reduce 110) ((T_INCLUDE_ONCE) reduce 110) ((T_INLINE_HTML) reduce 110) ((T_INTERFACE) reduce 110) ((T_INT_CAST) reduce 110) ((T_ISSET) reduce 110) ((T_LINE) reduce 110) ((T_LIST) reduce 110) ((T_LNUMBER) reduce 110) ((T_MATCH) reduce 110) ((T_METHOD_C) reduce 110) ((T_NAMESPACE) reduce 110) ((T_NAME_FULLY_QUALIFIED) reduce 110) ((T_NAME_QUALIFIED) reduce 110) ((T_NAME_RELATIVE) reduce 110) ((T_NEW) reduce 110) ((T_NS_C) reduce 110) ((T_OBJECT_CAST) reduce 110) ((T_PRINT) reduce 110) ((T_PROPERTY_C) reduce 110) ((T_READONLY) reduce 110) ((T_REQUIRE) reduce 110) ((T_REQUIRE_ONCE) reduce 110) ((T_RETURN) reduce 110) ((T_START_HEREDOC) reduce 110) ((T_STATIC) reduce 110) ((T_STRING) reduce 110) ((T_STRING_CAST) reduce 110) ((T_SWITCH) reduce 110) ((T_THROW) reduce 110) ((T_TRAIT) reduce 110) ((T_TRAIT_C) reduce 110) ((T_TRY) reduce 110) ((T_UNSET) reduce 110) ((T_UNSET_CAST) reduce 110) ((T_USE) reduce 110) ((T_VARIABLE) reduce 110) ((T_WHILE) reduce 110) ((T_YIELD) reduce 110) ((T_YIELD_FROM) reduce 110) (("[") reduce 110) (("`") reduce 110) (("{") reduce 110) (("}") reduce 110) (("~") reduce 110)) 30262 ((("!") reduce 248) (("\"") reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONST) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_EMPTY) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAMESPACE) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_USE) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("}") reduce 248) (("~") reduce 248)) 30266 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONST) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_EMPTY) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAMESPACE) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_USE) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("}") reduce 249) (("~") reduce 249)) 30271 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONST) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAMESPACE) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_USE) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("}") reduce 240) (("~") reduce 240)) 30272 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONST) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAMESPACE) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_USE) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 30275 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONST) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAMESPACE) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_USE) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("}") reduce 241) (("~") reduce 241)) 30281 ((("!") reduce 118) (("\"") reduce 118) (("$") reduce 118) (("(") reduce 118) (("+") reduce 118) (("-") reduce 118) ((";") reduce 118) (("@") reduce 118) ((T_ABSTRACT) reduce 118) ((T_ARRAY) reduce 118) ((T_ARRAY_CAST) reduce 118) ((T_ATTRIBUTE) reduce 118) ((T_BOOL_CAST) reduce 118) ((T_BREAK) reduce 118) ((T_CLASS) reduce 118) ((T_CLASS_C) reduce 118) ((T_CLONE) reduce 118) ((T_CONST) reduce 118) ((T_CONSTANT_ENCAPSED_STRING) reduce 118) ((T_CONTINUE) reduce 118) ((T_DEC) reduce 118) ((T_DECLARE) reduce 118) ((T_DIR) reduce 118) ((T_DNUMBER) reduce 118) ((T_DO) reduce 118) ((T_DOUBLE_CAST) reduce 118) ((T_ECHO) reduce 118) ((T_EMPTY) reduce 118) ((T_ENUM) reduce 118) ((T_EVAL) reduce 118) ((T_EXIT) reduce 118) ((T_FILE) reduce 118) ((T_FINAL) reduce 118) ((T_FN) reduce 118) ((T_FOR) reduce 118) ((T_FOREACH) reduce 118) ((T_FUNCTION) reduce 118) ((T_FUNC_C) reduce 118) ((T_GLOBAL) reduce 118) ((T_GOTO) reduce 118) ((T_HALT_COMPILER) reduce 118) ((T_IF) reduce 118) ((T_INC) reduce 118) ((T_INCLUDE) reduce 118) ((T_INCLUDE_ONCE) reduce 118) ((T_INLINE_HTML) reduce 118) ((T_INTERFACE) reduce 118) ((T_INT_CAST) reduce 118) ((T_ISSET) reduce 118) ((T_LINE) reduce 118) ((T_LIST) reduce 118) ((T_LNUMBER) reduce 118) ((T_MATCH) reduce 118) ((T_METHOD_C) reduce 118) ((T_NAMESPACE) reduce 118) ((T_NAME_FULLY_QUALIFIED) reduce 118) ((T_NAME_QUALIFIED) reduce 118) ((T_NAME_RELATIVE) reduce 118) ((T_NEW) reduce 118) ((T_NS_C) reduce 118) ((T_OBJECT_CAST) reduce 118) ((T_PRINT) reduce 118) ((T_PROPERTY_C) reduce 118) ((T_READONLY) reduce 118) ((T_REQUIRE) reduce 118) ((T_REQUIRE_ONCE) reduce 118) ((T_RETURN) reduce 118) ((T_START_HEREDOC) reduce 118) ((T_STATIC) reduce 118) ((T_STRING) reduce 118) ((T_STRING_CAST) reduce 118) ((T_SWITCH) reduce 118) ((T_THROW) reduce 118) ((T_TRAIT) reduce 118) ((T_TRAIT_C) reduce 118) ((T_TRY) reduce 118) ((T_UNSET) reduce 118) ((T_UNSET_CAST) reduce 118) ((T_USE) reduce 118) ((T_VARIABLE) reduce 118) ((T_WHILE) reduce 118) ((T_YIELD) reduce 118) ((T_YIELD_FROM) reduce 118) (("[") reduce 118) (("`") reduce 118) (("{") reduce 118) (("}") reduce 118) (("~") reduce 118)) 30282 ((("!") reduce 116) (("\"") reduce 116) (("$") reduce 116) (("(") reduce 116) (("+") reduce 116) (("-") reduce 116) ((";") reduce 116) (("@") reduce 116) ((T_ABSTRACT) reduce 116) ((T_ARRAY) reduce 116) ((T_ARRAY_CAST) reduce 116) ((T_ATTRIBUTE) reduce 116) ((T_BOOL_CAST) reduce 116) ((T_BREAK) reduce 116) ((T_CLASS) reduce 116) ((T_CLASS_C) reduce 116) ((T_CLONE) reduce 116) ((T_CONST) reduce 116) ((T_CONSTANT_ENCAPSED_STRING) reduce 116) ((T_CONTINUE) reduce 116) ((T_DEC) reduce 116) ((T_DECLARE) reduce 116) ((T_DIR) reduce 116) ((T_DNUMBER) reduce 116) ((T_DO) reduce 116) ((T_DOUBLE_CAST) reduce 116) ((T_ECHO) reduce 116) ((T_EMPTY) reduce 116) ((T_ENUM) reduce 116) ((T_EVAL) reduce 116) ((T_EXIT) reduce 116) ((T_FILE) reduce 116) ((T_FINAL) reduce 116) ((T_FN) reduce 116) ((T_FOR) reduce 116) ((T_FOREACH) reduce 116) ((T_FUNCTION) reduce 116) ((T_FUNC_C) reduce 116) ((T_GLOBAL) reduce 116) ((T_GOTO) reduce 116) ((T_HALT_COMPILER) reduce 116) ((T_IF) reduce 116) ((T_INC) reduce 116) ((T_INCLUDE) reduce 116) ((T_INCLUDE_ONCE) reduce 116) ((T_INLINE_HTML) reduce 116) ((T_INTERFACE) reduce 116) ((T_INT_CAST) reduce 116) ((T_ISSET) reduce 116) ((T_LINE) reduce 116) ((T_LIST) reduce 116) ((T_LNUMBER) reduce 116) ((T_MATCH) reduce 116) ((T_METHOD_C) reduce 116) ((T_NAMESPACE) reduce 116) ((T_NAME_FULLY_QUALIFIED) reduce 116) ((T_NAME_QUALIFIED) reduce 116) ((T_NAME_RELATIVE) reduce 116) ((T_NEW) reduce 116) ((T_NS_C) reduce 116) ((T_OBJECT_CAST) reduce 116) ((T_PRINT) reduce 116) ((T_PROPERTY_C) reduce 116) ((T_READONLY) reduce 116) ((T_REQUIRE) reduce 116) ((T_REQUIRE_ONCE) reduce 116) ((T_RETURN) reduce 116) ((T_START_HEREDOC) reduce 116) ((T_STATIC) reduce 116) ((T_STRING) reduce 116) ((T_STRING_CAST) reduce 116) ((T_SWITCH) reduce 116) ((T_THROW) reduce 116) ((T_TRAIT) reduce 116) ((T_TRAIT_C) reduce 116) ((T_TRY) reduce 116) ((T_UNSET) reduce 116) ((T_UNSET_CAST) reduce 116) ((T_USE) reduce 116) ((T_VARIABLE) reduce 116) ((T_WHILE) reduce 116) ((T_YIELD) reduce 116) ((T_YIELD_FROM) reduce 116) (("[") reduce 116) (("`") reduce 116) (("{") reduce 116) (("}") reduce 116) (("~") reduce 116)) 30283 ((("!") reduce 117) (("\"") reduce 117) (("$") reduce 117) (("(") reduce 117) (("+") reduce 117) (("-") reduce 117) ((";") reduce 117) (("@") reduce 117) ((T_ABSTRACT) reduce 117) ((T_ARRAY) reduce 117) ((T_ARRAY_CAST) reduce 117) ((T_ATTRIBUTE) reduce 117) ((T_BOOL_CAST) reduce 117) ((T_BREAK) reduce 117) ((T_CLASS) reduce 117) ((T_CLASS_C) reduce 117) ((T_CLONE) reduce 117) ((T_CONST) reduce 117) ((T_CONSTANT_ENCAPSED_STRING) reduce 117) ((T_CONTINUE) reduce 117) ((T_DEC) reduce 117) ((T_DECLARE) reduce 117) ((T_DIR) reduce 117) ((T_DNUMBER) reduce 117) ((T_DO) reduce 117) ((T_DOUBLE_CAST) reduce 117) ((T_ECHO) reduce 117) ((T_EMPTY) reduce 117) ((T_ENUM) reduce 117) ((T_EVAL) reduce 117) ((T_EXIT) reduce 117) ((T_FILE) reduce 117) ((T_FINAL) reduce 117) ((T_FN) reduce 117) ((T_FOR) reduce 117) ((T_FOREACH) reduce 117) ((T_FUNCTION) reduce 117) ((T_FUNC_C) reduce 117) ((T_GLOBAL) reduce 117) ((T_GOTO) reduce 117) ((T_HALT_COMPILER) reduce 117) ((T_IF) reduce 117) ((T_INC) reduce 117) ((T_INCLUDE) reduce 117) ((T_INCLUDE_ONCE) reduce 117) ((T_INLINE_HTML) reduce 117) ((T_INTERFACE) reduce 117) ((T_INT_CAST) reduce 117) ((T_ISSET) reduce 117) ((T_LINE) reduce 117) ((T_LIST) reduce 117) ((T_LNUMBER) reduce 117) ((T_MATCH) reduce 117) ((T_METHOD_C) reduce 117) ((T_NAMESPACE) reduce 117) ((T_NAME_FULLY_QUALIFIED) reduce 117) ((T_NAME_QUALIFIED) reduce 117) ((T_NAME_RELATIVE) reduce 117) ((T_NEW) reduce 117) ((T_NS_C) reduce 117) ((T_OBJECT_CAST) reduce 117) ((T_PRINT) reduce 117) ((T_PROPERTY_C) reduce 117) ((T_READONLY) reduce 117) ((T_REQUIRE) reduce 117) ((T_REQUIRE_ONCE) reduce 117) ((T_RETURN) reduce 117) ((T_START_HEREDOC) reduce 117) ((T_STATIC) reduce 117) ((T_STRING) reduce 117) ((T_STRING_CAST) reduce 117) ((T_SWITCH) reduce 117) ((T_THROW) reduce 117) ((T_TRAIT) reduce 117) ((T_TRAIT_C) reduce 117) ((T_TRY) reduce 117) ((T_UNSET) reduce 117) ((T_UNSET_CAST) reduce 117) ((T_USE) reduce 117) ((T_VARIABLE) reduce 117) ((T_WHILE) reduce 117) ((T_YIELD) reduce 117) ((T_YIELD_FROM) reduce 117) (("[") reduce 117) (("`") reduce 117) (("{") reduce 117) (("}") reduce 117) (("~") reduce 117)) 30284 ((("!") reduce 115) (("\"") reduce 115) (("$") reduce 115) (("(") reduce 115) (("+") reduce 115) (("-") reduce 115) ((";") reduce 115) (("@") reduce 115) ((T_ABSTRACT) reduce 115) ((T_ARRAY) reduce 115) ((T_ARRAY_CAST) reduce 115) ((T_ATTRIBUTE) reduce 115) ((T_BOOL_CAST) reduce 115) ((T_BREAK) reduce 115) ((T_CLASS) reduce 115) ((T_CLASS_C) reduce 115) ((T_CLONE) reduce 115) ((T_CONST) reduce 115) ((T_CONSTANT_ENCAPSED_STRING) reduce 115) ((T_CONTINUE) reduce 115) ((T_DEC) reduce 115) ((T_DECLARE) reduce 115) ((T_DIR) reduce 115) ((T_DNUMBER) reduce 115) ((T_DO) reduce 115) ((T_DOUBLE_CAST) reduce 115) ((T_ECHO) reduce 115) ((T_EMPTY) reduce 115) ((T_ENUM) reduce 115) ((T_EVAL) reduce 115) ((T_EXIT) reduce 115) ((T_FILE) reduce 115) ((T_FINAL) reduce 115) ((T_FN) reduce 115) ((T_FOR) reduce 115) ((T_FOREACH) reduce 115) ((T_FUNCTION) reduce 115) ((T_FUNC_C) reduce 115) ((T_GLOBAL) reduce 115) ((T_GOTO) reduce 115) ((T_HALT_COMPILER) reduce 115) ((T_IF) reduce 115) ((T_INC) reduce 115) ((T_INCLUDE) reduce 115) ((T_INCLUDE_ONCE) reduce 115) ((T_INLINE_HTML) reduce 115) ((T_INTERFACE) reduce 115) ((T_INT_CAST) reduce 115) ((T_ISSET) reduce 115) ((T_LINE) reduce 115) ((T_LIST) reduce 115) ((T_LNUMBER) reduce 115) ((T_MATCH) reduce 115) ((T_METHOD_C) reduce 115) ((T_NAMESPACE) reduce 115) ((T_NAME_FULLY_QUALIFIED) reduce 115) ((T_NAME_QUALIFIED) reduce 115) ((T_NAME_RELATIVE) reduce 115) ((T_NEW) reduce 115) ((T_NS_C) reduce 115) ((T_OBJECT_CAST) reduce 115) ((T_PRINT) reduce 115) ((T_PROPERTY_C) reduce 115) ((T_READONLY) reduce 115) ((T_REQUIRE) reduce 115) ((T_REQUIRE_ONCE) reduce 115) ((T_RETURN) reduce 115) ((T_START_HEREDOC) reduce 115) ((T_STATIC) reduce 115) ((T_STRING) reduce 115) ((T_STRING_CAST) reduce 115) ((T_SWITCH) reduce 115) ((T_THROW) reduce 115) ((T_TRAIT) reduce 115) ((T_TRAIT_C) reduce 115) ((T_TRY) reduce 115) ((T_UNSET) reduce 115) ((T_UNSET_CAST) reduce 115) ((T_USE) reduce 115) ((T_VARIABLE) reduce 115) ((T_WHILE) reduce 115) ((T_YIELD) reduce 115) ((T_YIELD_FROM) reduce 115) (("[") reduce 115) (("`") reduce 115) (("{") reduce 115) (("}") reduce 115) (("~") reduce 115)) 30289 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONST) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAMESPACE) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_USE) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 30292 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONST) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAMESPACE) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_USE) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 30293 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONST) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAMESPACE) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_USE) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("}") reduce 175) (("~") reduce 175)) 30296 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONST) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAMESPACE) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_USE) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 30299 ((("!") reduce 176) (("\"") reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONST) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_EMPTY) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAMESPACE) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_USE) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("}") reduce 176) (("~") reduce 176)) 30306 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONST) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAMESPACE) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_USE) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("}") reduce 170) (("~") reduce 170)) 30311 ((("!") reduce 198) (("\"") reduce 198) (("$") reduce 198) (("(") reduce 198) (("+") reduce 198) (("-") reduce 198) ((";") reduce 198) (("@") reduce 198) ((T_ABSTRACT) reduce 198) ((T_ARRAY) reduce 198) ((T_ARRAY_CAST) reduce 198) ((T_ATTRIBUTE) reduce 198) ((T_BOOL_CAST) reduce 198) ((T_BREAK) reduce 198) ((T_CLASS) reduce 198) ((T_CLASS_C) reduce 198) ((T_CLONE) reduce 198) ((T_CONST) reduce 198) ((T_CONSTANT_ENCAPSED_STRING) reduce 198) ((T_CONTINUE) reduce 198) ((T_DEC) reduce 198) ((T_DECLARE) reduce 198) ((T_DIR) reduce 198) ((T_DNUMBER) reduce 198) ((T_DO) reduce 198) ((T_DOUBLE_CAST) reduce 198) ((T_ECHO) reduce 198) ((T_EMPTY) reduce 198) ((T_ENUM) reduce 198) ((T_EVAL) reduce 198) ((T_EXIT) reduce 198) ((T_FILE) reduce 198) ((T_FINAL) reduce 198) ((T_FN) reduce 198) ((T_FOR) reduce 198) ((T_FOREACH) reduce 198) ((T_FUNCTION) reduce 198) ((T_FUNC_C) reduce 198) ((T_GLOBAL) reduce 198) ((T_GOTO) reduce 198) ((T_HALT_COMPILER) reduce 198) ((T_IF) reduce 198) ((T_INC) reduce 198) ((T_INCLUDE) reduce 198) ((T_INCLUDE_ONCE) reduce 198) ((T_INLINE_HTML) reduce 198) ((T_INTERFACE) reduce 198) ((T_INT_CAST) reduce 198) ((T_ISSET) reduce 198) ((T_LINE) reduce 198) ((T_LIST) reduce 198) ((T_LNUMBER) reduce 198) ((T_MATCH) reduce 198) ((T_METHOD_C) reduce 198) ((T_NAMESPACE) reduce 198) ((T_NAME_FULLY_QUALIFIED) reduce 198) ((T_NAME_QUALIFIED) reduce 198) ((T_NAME_RELATIVE) reduce 198) ((T_NEW) reduce 198) ((T_NS_C) reduce 198) ((T_OBJECT_CAST) reduce 198) ((T_PRINT) reduce 198) ((T_PROPERTY_C) reduce 198) ((T_READONLY) reduce 198) ((T_REQUIRE) reduce 198) ((T_REQUIRE_ONCE) reduce 198) ((T_RETURN) reduce 198) ((T_START_HEREDOC) reduce 198) ((T_STATIC) reduce 198) ((T_STRING) reduce 198) ((T_STRING_CAST) reduce 198) ((T_SWITCH) reduce 198) ((T_THROW) reduce 198) ((T_TRAIT) reduce 198) ((T_TRAIT_C) reduce 198) ((T_TRY) reduce 198) ((T_UNSET) reduce 198) ((T_UNSET_CAST) reduce 198) ((T_USE) reduce 198) ((T_VARIABLE) reduce 198) ((T_WHILE) reduce 198) ((T_YIELD) reduce 198) ((T_YIELD_FROM) reduce 198) (("[") reduce 198) (("`") reduce 198) (("{") reduce 198) (("}") reduce 198) (("~") reduce 198)) 30316 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONST) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAMESPACE) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_USE) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 30320 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONST) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAMESPACE) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_USE) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("}") reduce 222) (("~") reduce 222)) 30322 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONST) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAMESPACE) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_USE) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("}") reduce 223) (("~") reduce 223)) 30326 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONST) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAMESPACE) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_USE) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("}") reduce 224) (("~") reduce 224)) 30329 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONST) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAMESPACE) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_USE) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("}") reduce 225) (("~") reduce 225)) 30330 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONST) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAMESPACE) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_USE) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 30332 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONST) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAMESPACE) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_USE) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 30334 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONST) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAMESPACE) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_USE) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 30338 ((("!") reduce 114) (("\"") reduce 114) (("$") reduce 114) (("(") reduce 114) (("+") reduce 114) (("-") reduce 114) ((";") reduce 114) (("@") reduce 114) ((T_ABSTRACT) reduce 114) ((T_ARRAY) reduce 114) ((T_ARRAY_CAST) reduce 114) ((T_ATTRIBUTE) reduce 114) ((T_BOOL_CAST) reduce 114) ((T_BREAK) reduce 114) ((T_CLASS) reduce 114) ((T_CLASS_C) reduce 114) ((T_CLONE) reduce 114) ((T_CONST) reduce 114) ((T_CONSTANT_ENCAPSED_STRING) reduce 114) ((T_CONTINUE) reduce 114) ((T_DEC) reduce 114) ((T_DECLARE) reduce 114) ((T_DIR) reduce 114) ((T_DNUMBER) reduce 114) ((T_DO) reduce 114) ((T_DOUBLE_CAST) reduce 114) ((T_ECHO) reduce 114) ((T_EMPTY) reduce 114) ((T_ENUM) reduce 114) ((T_EVAL) reduce 114) ((T_EXIT) reduce 114) ((T_FILE) reduce 114) ((T_FINAL) reduce 114) ((T_FN) reduce 114) ((T_FOR) reduce 114) ((T_FOREACH) reduce 114) ((T_FUNCTION) reduce 114) ((T_FUNC_C) reduce 114) ((T_GLOBAL) reduce 114) ((T_GOTO) reduce 114) ((T_HALT_COMPILER) reduce 114) ((T_IF) reduce 114) ((T_INC) reduce 114) ((T_INCLUDE) reduce 114) ((T_INCLUDE_ONCE) reduce 114) ((T_INLINE_HTML) reduce 114) ((T_INTERFACE) reduce 114) ((T_INT_CAST) reduce 114) ((T_ISSET) reduce 114) ((T_LINE) reduce 114) ((T_LIST) reduce 114) ((T_LNUMBER) reduce 114) ((T_MATCH) reduce 114) ((T_METHOD_C) reduce 114) ((T_NAMESPACE) reduce 114) ((T_NAME_FULLY_QUALIFIED) reduce 114) ((T_NAME_QUALIFIED) reduce 114) ((T_NAME_RELATIVE) reduce 114) ((T_NEW) reduce 114) ((T_NS_C) reduce 114) ((T_OBJECT_CAST) reduce 114) ((T_PRINT) reduce 114) ((T_PROPERTY_C) reduce 114) ((T_READONLY) reduce 114) ((T_REQUIRE) reduce 114) ((T_REQUIRE_ONCE) reduce 114) ((T_RETURN) reduce 114) ((T_START_HEREDOC) reduce 114) ((T_STATIC) reduce 114) ((T_STRING) reduce 114) ((T_STRING_CAST) reduce 114) ((T_SWITCH) reduce 114) ((T_THROW) reduce 114) ((T_TRAIT) reduce 114) ((T_TRAIT_C) reduce 114) ((T_TRY) reduce 114) ((T_UNSET) reduce 114) ((T_UNSET_CAST) reduce 114) ((T_USE) reduce 114) ((T_VARIABLE) reduce 114) ((T_WHILE) reduce 114) ((T_YIELD) reduce 114) ((T_YIELD_FROM) reduce 114) (("[") reduce 114) (("`") reduce 114) (("{") reduce 114) (("}") reduce 114) (("~") reduce 114)) 30339 ((("!") reduce 112) (("\"") reduce 112) (("$") reduce 112) (("(") reduce 112) (("+") reduce 112) (("-") reduce 112) ((";") reduce 112) (("@") reduce 112) ((T_ABSTRACT) reduce 112) ((T_ARRAY) reduce 112) ((T_ARRAY_CAST) reduce 112) ((T_ATTRIBUTE) reduce 112) ((T_BOOL_CAST) reduce 112) ((T_BREAK) reduce 112) ((T_CLASS) reduce 112) ((T_CLASS_C) reduce 112) ((T_CLONE) reduce 112) ((T_CONST) reduce 112) ((T_CONSTANT_ENCAPSED_STRING) reduce 112) ((T_CONTINUE) reduce 112) ((T_DEC) reduce 112) ((T_DECLARE) reduce 112) ((T_DIR) reduce 112) ((T_DNUMBER) reduce 112) ((T_DO) reduce 112) ((T_DOUBLE_CAST) reduce 112) ((T_ECHO) reduce 112) ((T_EMPTY) reduce 112) ((T_ENUM) reduce 112) ((T_EVAL) reduce 112) ((T_EXIT) reduce 112) ((T_FILE) reduce 112) ((T_FINAL) reduce 112) ((T_FN) reduce 112) ((T_FOR) reduce 112) ((T_FOREACH) reduce 112) ((T_FUNCTION) reduce 112) ((T_FUNC_C) reduce 112) ((T_GLOBAL) reduce 112) ((T_GOTO) reduce 112) ((T_HALT_COMPILER) reduce 112) ((T_IF) reduce 112) ((T_INC) reduce 112) ((T_INCLUDE) reduce 112) ((T_INCLUDE_ONCE) reduce 112) ((T_INLINE_HTML) reduce 112) ((T_INTERFACE) reduce 112) ((T_INT_CAST) reduce 112) ((T_ISSET) reduce 112) ((T_LINE) reduce 112) ((T_LIST) reduce 112) ((T_LNUMBER) reduce 112) ((T_MATCH) reduce 112) ((T_METHOD_C) reduce 112) ((T_NAMESPACE) reduce 112) ((T_NAME_FULLY_QUALIFIED) reduce 112) ((T_NAME_QUALIFIED) reduce 112) ((T_NAME_RELATIVE) reduce 112) ((T_NEW) reduce 112) ((T_NS_C) reduce 112) ((T_OBJECT_CAST) reduce 112) ((T_PRINT) reduce 112) ((T_PROPERTY_C) reduce 112) ((T_READONLY) reduce 112) ((T_REQUIRE) reduce 112) ((T_REQUIRE_ONCE) reduce 112) ((T_RETURN) reduce 112) ((T_START_HEREDOC) reduce 112) ((T_STATIC) reduce 112) ((T_STRING) reduce 112) ((T_STRING_CAST) reduce 112) ((T_SWITCH) reduce 112) ((T_THROW) reduce 112) ((T_TRAIT) reduce 112) ((T_TRAIT_C) reduce 112) ((T_TRY) reduce 112) ((T_UNSET) reduce 112) ((T_UNSET_CAST) reduce 112) ((T_USE) reduce 112) ((T_VARIABLE) reduce 112) ((T_WHILE) reduce 112) ((T_YIELD) reduce 112) ((T_YIELD_FROM) reduce 112) (("[") reduce 112) (("`") reduce 112) (("{") reduce 112) (("}") reduce 112) (("~") reduce 112)) 30342 ((("!") reduce 113) (("\"") reduce 113) (("$") reduce 113) (("(") reduce 113) (("+") reduce 113) (("-") reduce 113) ((";") reduce 113) (("@") reduce 113) ((T_ABSTRACT) reduce 113) ((T_ARRAY) reduce 113) ((T_ARRAY_CAST) reduce 113) ((T_ATTRIBUTE) reduce 113) ((T_BOOL_CAST) reduce 113) ((T_BREAK) reduce 113) ((T_CLASS) reduce 113) ((T_CLASS_C) reduce 113) ((T_CLONE) reduce 113) ((T_CONST) reduce 113) ((T_CONSTANT_ENCAPSED_STRING) reduce 113) ((T_CONTINUE) reduce 113) ((T_DEC) reduce 113) ((T_DECLARE) reduce 113) ((T_DIR) reduce 113) ((T_DNUMBER) reduce 113) ((T_DO) reduce 113) ((T_DOUBLE_CAST) reduce 113) ((T_ECHO) reduce 113) ((T_EMPTY) reduce 113) ((T_ENUM) reduce 113) ((T_EVAL) reduce 113) ((T_EXIT) reduce 113) ((T_FILE) reduce 113) ((T_FINAL) reduce 113) ((T_FN) reduce 113) ((T_FOR) reduce 113) ((T_FOREACH) reduce 113) ((T_FUNCTION) reduce 113) ((T_FUNC_C) reduce 113) ((T_GLOBAL) reduce 113) ((T_GOTO) reduce 113) ((T_HALT_COMPILER) reduce 113) ((T_IF) reduce 113) ((T_INC) reduce 113) ((T_INCLUDE) reduce 113) ((T_INCLUDE_ONCE) reduce 113) ((T_INLINE_HTML) reduce 113) ((T_INTERFACE) reduce 113) ((T_INT_CAST) reduce 113) ((T_ISSET) reduce 113) ((T_LINE) reduce 113) ((T_LIST) reduce 113) ((T_LNUMBER) reduce 113) ((T_MATCH) reduce 113) ((T_METHOD_C) reduce 113) ((T_NAMESPACE) reduce 113) ((T_NAME_FULLY_QUALIFIED) reduce 113) ((T_NAME_QUALIFIED) reduce 113) ((T_NAME_RELATIVE) reduce 113) ((T_NEW) reduce 113) ((T_NS_C) reduce 113) ((T_OBJECT_CAST) reduce 113) ((T_PRINT) reduce 113) ((T_PROPERTY_C) reduce 113) ((T_READONLY) reduce 113) ((T_REQUIRE) reduce 113) ((T_REQUIRE_ONCE) reduce 113) ((T_RETURN) reduce 113) ((T_START_HEREDOC) reduce 113) ((T_STATIC) reduce 113) ((T_STRING) reduce 113) ((T_STRING_CAST) reduce 113) ((T_SWITCH) reduce 113) ((T_THROW) reduce 113) ((T_TRAIT) reduce 113) ((T_TRAIT_C) reduce 113) ((T_TRY) reduce 113) ((T_UNSET) reduce 113) ((T_UNSET_CAST) reduce 113) ((T_USE) reduce 113) ((T_VARIABLE) reduce 113) ((T_WHILE) reduce 113) ((T_YIELD) reduce 113) ((T_YIELD_FROM) reduce 113) (("[") reduce 113) (("`") reduce 113) (("{") reduce 113) (("}") reduce 113) (("~") reduce 113)) 30348 ((("!") reduce 199) (("\"") reduce 199) (("$") reduce 199) (("(") reduce 199) (("+") reduce 199) (("-") reduce 199) ((";") reduce 199) (("@") reduce 199) ((T_ABSTRACT) reduce 199) ((T_ARRAY) reduce 199) ((T_ARRAY_CAST) reduce 199) ((T_ATTRIBUTE) reduce 199) ((T_BOOL_CAST) reduce 199) ((T_BREAK) reduce 199) ((T_CLASS) reduce 199) ((T_CLASS_C) reduce 199) ((T_CLONE) reduce 199) ((T_CONST) reduce 199) ((T_CONSTANT_ENCAPSED_STRING) reduce 199) ((T_CONTINUE) reduce 199) ((T_DEC) reduce 199) ((T_DECLARE) reduce 199) ((T_DIR) reduce 199) ((T_DNUMBER) reduce 199) ((T_DO) reduce 199) ((T_DOUBLE_CAST) reduce 199) ((T_ECHO) reduce 199) ((T_EMPTY) reduce 199) ((T_ENUM) reduce 199) ((T_EVAL) reduce 199) ((T_EXIT) reduce 199) ((T_FILE) reduce 199) ((T_FINAL) reduce 199) ((T_FN) reduce 199) ((T_FOR) reduce 199) ((T_FOREACH) reduce 199) ((T_FUNCTION) reduce 199) ((T_FUNC_C) reduce 199) ((T_GLOBAL) reduce 199) ((T_GOTO) reduce 199) ((T_HALT_COMPILER) reduce 199) ((T_IF) reduce 199) ((T_INC) reduce 199) ((T_INCLUDE) reduce 199) ((T_INCLUDE_ONCE) reduce 199) ((T_INLINE_HTML) reduce 199) ((T_INTERFACE) reduce 199) ((T_INT_CAST) reduce 199) ((T_ISSET) reduce 199) ((T_LINE) reduce 199) ((T_LIST) reduce 199) ((T_LNUMBER) reduce 199) ((T_MATCH) reduce 199) ((T_METHOD_C) reduce 199) ((T_NAMESPACE) reduce 199) ((T_NAME_FULLY_QUALIFIED) reduce 199) ((T_NAME_QUALIFIED) reduce 199) ((T_NAME_RELATIVE) reduce 199) ((T_NEW) reduce 199) ((T_NS_C) reduce 199) ((T_OBJECT_CAST) reduce 199) ((T_PRINT) reduce 199) ((T_PROPERTY_C) reduce 199) ((T_READONLY) reduce 199) ((T_REQUIRE) reduce 199) ((T_REQUIRE_ONCE) reduce 199) ((T_RETURN) reduce 199) ((T_START_HEREDOC) reduce 199) ((T_STATIC) reduce 199) ((T_STRING) reduce 199) ((T_STRING_CAST) reduce 199) ((T_SWITCH) reduce 199) ((T_THROW) reduce 199) ((T_TRAIT) reduce 199) ((T_TRAIT_C) reduce 199) ((T_TRY) reduce 199) ((T_UNSET) reduce 199) ((T_UNSET_CAST) reduce 199) ((T_USE) reduce 199) ((T_VARIABLE) reduce 199) ((T_WHILE) reduce 199) ((T_YIELD) reduce 199) ((T_YIELD_FROM) reduce 199) (("[") reduce 199) (("`") reduce 199) (("{") reduce 199) (("}") reduce 199) (("~") reduce 199)) 30352 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONST) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAMESPACE) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_USE) reduce 242) ((T_VARIABLE) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("}") reduce 242) (("~") reduce 242)) 30355 ((("!") reduce 111) (("\"") reduce 111) (("$") reduce 111) (("(") reduce 111) (("+") reduce 111) (("-") reduce 111) ((";") reduce 111) (("@") reduce 111) ((T_ABSTRACT) reduce 111) ((T_ARRAY) reduce 111) ((T_ARRAY_CAST) reduce 111) ((T_ATTRIBUTE) reduce 111) ((T_BOOL_CAST) reduce 111) ((T_BREAK) reduce 111) ((T_CLASS) reduce 111) ((T_CLASS_C) reduce 111) ((T_CLONE) reduce 111) ((T_CONST) reduce 111) ((T_CONSTANT_ENCAPSED_STRING) reduce 111) ((T_CONTINUE) reduce 111) ((T_DEC) reduce 111) ((T_DECLARE) reduce 111) ((T_DIR) reduce 111) ((T_DNUMBER) reduce 111) ((T_DO) reduce 111) ((T_DOUBLE_CAST) reduce 111) ((T_ECHO) reduce 111) ((T_EMPTY) reduce 111) ((T_ENUM) reduce 111) ((T_EVAL) reduce 111) ((T_EXIT) reduce 111) ((T_FILE) reduce 111) ((T_FINAL) reduce 111) ((T_FN) reduce 111) ((T_FOR) reduce 111) ((T_FOREACH) reduce 111) ((T_FUNCTION) reduce 111) ((T_FUNC_C) reduce 111) ((T_GLOBAL) reduce 111) ((T_GOTO) reduce 111) ((T_HALT_COMPILER) reduce 111) ((T_IF) reduce 111) ((T_INC) reduce 111) ((T_INCLUDE) reduce 111) ((T_INCLUDE_ONCE) reduce 111) ((T_INLINE_HTML) reduce 111) ((T_INTERFACE) reduce 111) ((T_INT_CAST) reduce 111) ((T_ISSET) reduce 111) ((T_LINE) reduce 111) ((T_LIST) reduce 111) ((T_LNUMBER) reduce 111) ((T_MATCH) reduce 111) ((T_METHOD_C) reduce 111) ((T_NAMESPACE) reduce 111) ((T_NAME_FULLY_QUALIFIED) reduce 111) ((T_NAME_QUALIFIED) reduce 111) ((T_NAME_RELATIVE) reduce 111) ((T_NEW) reduce 111) ((T_NS_C) reduce 111) ((T_OBJECT_CAST) reduce 111) ((T_PRINT) reduce 111) ((T_PROPERTY_C) reduce 111) ((T_READONLY) reduce 111) ((T_REQUIRE) reduce 111) ((T_REQUIRE_ONCE) reduce 111) ((T_RETURN) reduce 111) ((T_START_HEREDOC) reduce 111) ((T_STATIC) reduce 111) ((T_STRING) reduce 111) ((T_STRING_CAST) reduce 111) ((T_SWITCH) reduce 111) ((T_THROW) reduce 111) ((T_TRAIT) reduce 111) ((T_TRAIT_C) reduce 111) ((T_TRY) reduce 111) ((T_UNSET) reduce 111) ((T_UNSET_CAST) reduce 111) ((T_USE) reduce 111) ((T_VARIABLE) reduce 111) ((T_WHILE) reduce 111) ((T_YIELD) reduce 111) ((T_YIELD_FROM) reduce 111) (("[") reduce 111) (("`") reduce 111) (("{") reduce 111) (("}") reduce 111) (("~") reduce 111)) 30357 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONST) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAMESPACE) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_USE) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 30359 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONST) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAMESPACE) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_USE) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 30369 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONST) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAMESPACE) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_USE) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 30370 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONST) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAMESPACE) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_USE) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("}") reduce 218) (("~") reduce 218)) 30373 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONST) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAMESPACE) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_USE) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("}") reduce 219) (("~") reduce 219)) 30374 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONST) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAMESPACE) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_USE) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 30383 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONST) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAMESPACE) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_USE) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 30384 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONST) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAMESPACE) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_USE) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("}") reduce 216) (("~") reduce 216)) 30387 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONST) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAMESPACE) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_USE) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("}") reduce 217) (("~") reduce 217)) 30394 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONST) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAMESPACE) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_USE) reduce 200) ((T_VARIABLE) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("}") reduce 200) (("~") reduce 200)) 30396 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONST) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAMESPACE) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_USE) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 30402 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONST) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAMESPACE) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_USE) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 30407 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONST) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAMESPACE) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_USE) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 30408 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONST) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAMESPACE) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_USE) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("}") reduce 220) (("~") reduce 220)) 30411 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONST) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAMESPACE) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_USE) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("}") reduce 221) (("~") reduce 221)) 30413 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONST) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAMESPACE) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_USE) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 30415 (((",") reduce 139) ((";") reduce 139)) 30418 ((("!") reduce 119) (("\"") reduce 119) (("$") reduce 119) (("(") reduce 119) (("+") reduce 119) (("-") reduce 119) ((";") reduce 119) (("@") reduce 119) ((T_ABSTRACT) reduce 119) ((T_ARRAY) reduce 119) ((T_ARRAY_CAST) reduce 119) ((T_ATTRIBUTE) reduce 119) ((T_BOOL_CAST) reduce 119) ((T_BREAK) reduce 119) ((T_CLASS) reduce 119) ((T_CLASS_C) reduce 119) ((T_CLONE) reduce 119) ((T_CONST) reduce 119) ((T_CONSTANT_ENCAPSED_STRING) reduce 119) ((T_CONTINUE) reduce 119) ((T_DEC) reduce 119) ((T_DECLARE) reduce 119) ((T_DIR) reduce 119) ((T_DNUMBER) reduce 119) ((T_DO) reduce 119) ((T_DOUBLE_CAST) reduce 119) ((T_ECHO) reduce 119) ((T_EMPTY) reduce 119) ((T_ENUM) reduce 119) ((T_EVAL) reduce 119) ((T_EXIT) reduce 119) ((T_FILE) reduce 119) ((T_FINAL) reduce 119) ((T_FN) reduce 119) ((T_FOR) reduce 119) ((T_FOREACH) reduce 119) ((T_FUNCTION) reduce 119) ((T_FUNC_C) reduce 119) ((T_GLOBAL) reduce 119) ((T_GOTO) reduce 119) ((T_HALT_COMPILER) reduce 119) ((T_IF) reduce 119) ((T_INC) reduce 119) ((T_INCLUDE) reduce 119) ((T_INCLUDE_ONCE) reduce 119) ((T_INLINE_HTML) reduce 119) ((T_INTERFACE) reduce 119) ((T_INT_CAST) reduce 119) ((T_ISSET) reduce 119) ((T_LINE) reduce 119) ((T_LIST) reduce 119) ((T_LNUMBER) reduce 119) ((T_MATCH) reduce 119) ((T_METHOD_C) reduce 119) ((T_NAMESPACE) reduce 119) ((T_NAME_FULLY_QUALIFIED) reduce 119) ((T_NAME_QUALIFIED) reduce 119) ((T_NAME_RELATIVE) reduce 119) ((T_NEW) reduce 119) ((T_NS_C) reduce 119) ((T_OBJECT_CAST) reduce 119) ((T_PRINT) reduce 119) ((T_PROPERTY_C) reduce 119) ((T_READONLY) reduce 119) ((T_REQUIRE) reduce 119) ((T_REQUIRE_ONCE) reduce 119) ((T_RETURN) reduce 119) ((T_START_HEREDOC) reduce 119) ((T_STATIC) reduce 119) ((T_STRING) reduce 119) ((T_STRING_CAST) reduce 119) ((T_SWITCH) reduce 119) ((T_THROW) reduce 119) ((T_TRAIT) reduce 119) ((T_TRAIT_C) reduce 119) ((T_TRY) reduce 119) ((T_UNSET) reduce 119) ((T_UNSET_CAST) reduce 119) ((T_USE) reduce 119) ((T_VARIABLE) reduce 119) ((T_WHILE) reduce 119) ((T_YIELD) reduce 119) ((T_YIELD_FROM) reduce 119) (("[") reduce 119) (("`") reduce 119) (("{") reduce 119) (("}") reduce 119) (("~") reduce 119)) 30419 (((",") reduce 138) ((";") reduce 138)) 30422 (((",") reduce 375) ((";") reduce 375)) 30429 ((("!") reduce 188) (("\"") reduce 188) (("$") reduce 188) (("(") reduce 188) (("+") reduce 188) (("-") reduce 188) ((";") reduce 188) (("@") reduce 188) ((T_ABSTRACT) reduce 188) ((T_ARRAY) reduce 188) ((T_ARRAY_CAST) reduce 188) ((T_ATTRIBUTE) reduce 188) ((T_BOOL_CAST) reduce 188) ((T_BREAK) reduce 188) ((T_CLASS) reduce 188) ((T_CLASS_C) reduce 188) ((T_CLONE) reduce 188) ((T_CONST) reduce 188) ((T_CONSTANT_ENCAPSED_STRING) reduce 188) ((T_CONTINUE) reduce 188) ((T_DEC) reduce 188) ((T_DECLARE) reduce 188) ((T_DIR) reduce 188) ((T_DNUMBER) reduce 188) ((T_DO) reduce 188) ((T_DOUBLE_CAST) reduce 188) ((T_ECHO) reduce 188) ((T_EMPTY) reduce 188) ((T_ENUM) reduce 188) ((T_EVAL) reduce 188) ((T_EXIT) reduce 188) ((T_FILE) reduce 188) ((T_FINAL) reduce 188) ((T_FN) reduce 188) ((T_FOR) reduce 188) ((T_FOREACH) reduce 188) ((T_FUNCTION) reduce 188) ((T_FUNC_C) reduce 188) ((T_GLOBAL) reduce 188) ((T_GOTO) reduce 188) ((T_HALT_COMPILER) reduce 188) ((T_IF) reduce 188) ((T_INC) reduce 188) ((T_INCLUDE) reduce 188) ((T_INCLUDE_ONCE) reduce 188) ((T_INLINE_HTML) reduce 188) ((T_INTERFACE) reduce 188) ((T_INT_CAST) reduce 188) ((T_ISSET) reduce 188) ((T_LINE) reduce 188) ((T_LIST) reduce 188) ((T_LNUMBER) reduce 188) ((T_MATCH) reduce 188) ((T_METHOD_C) reduce 188) ((T_NAMESPACE) reduce 188) ((T_NAME_FULLY_QUALIFIED) reduce 188) ((T_NAME_QUALIFIED) reduce 188) ((T_NAME_RELATIVE) reduce 188) ((T_NEW) reduce 188) ((T_NS_C) reduce 188) ((T_OBJECT_CAST) reduce 188) ((T_PRINT) reduce 188) ((T_PROPERTY_C) reduce 188) ((T_READONLY) reduce 188) ((T_REQUIRE) reduce 188) ((T_REQUIRE_ONCE) reduce 188) ((T_RETURN) reduce 188) ((T_START_HEREDOC) reduce 188) ((T_STATIC) reduce 188) ((T_STRING) reduce 188) ((T_STRING_CAST) reduce 188) ((T_SWITCH) reduce 188) ((T_THROW) reduce 188) ((T_TRAIT) reduce 188) ((T_TRAIT_C) reduce 188) ((T_TRY) reduce 188) ((T_UNSET) reduce 188) ((T_UNSET_CAST) reduce 188) ((T_USE) reduce 188) ((T_VARIABLE) reduce 188) ((T_WHILE) reduce 188) ((T_YIELD) reduce 188) ((T_YIELD_FROM) reduce 188) (("[") reduce 188) (("`") reduce 188) (("{") reduce 188) (("}") reduce 188) (("~") reduce 188)) 30431 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONST) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAMESPACE) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_USE) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 30432 ((("!") reduce 112) (("\"") reduce 112) (("$") reduce 112) (($) reduce 112) (("(") reduce 112) (("+") reduce 112) (("-") reduce 112) ((";") reduce 112) (("@") reduce 112) ((T_ABSTRACT) reduce 112) ((T_ARRAY) reduce 112) ((T_ARRAY_CAST) reduce 112) ((T_ATTRIBUTE) reduce 112) ((T_BOOL_CAST) reduce 112) ((T_BREAK) reduce 112) ((T_CLASS) reduce 112) ((T_CLASS_C) reduce 112) ((T_CLONE) reduce 112) ((T_CONST) reduce 112) ((T_CONSTANT_ENCAPSED_STRING) reduce 112) ((T_CONTINUE) reduce 112) ((T_DEC) reduce 112) ((T_DECLARE) reduce 112) ((T_DIR) reduce 112) ((T_DNUMBER) reduce 112) ((T_DO) reduce 112) ((T_DOUBLE_CAST) reduce 112) ((T_ECHO) reduce 112) ((T_EMPTY) reduce 112) ((T_ENUM) reduce 112) ((T_EVAL) reduce 112) ((T_EXIT) reduce 112) ((T_FILE) reduce 112) ((T_FINAL) reduce 112) ((T_FN) reduce 112) ((T_FOR) reduce 112) ((T_FOREACH) reduce 112) ((T_FUNCTION) reduce 112) ((T_FUNC_C) reduce 112) ((T_GLOBAL) reduce 112) ((T_GOTO) reduce 112) ((T_HALT_COMPILER) reduce 112) ((T_IF) reduce 112) ((T_INC) reduce 112) ((T_INCLUDE) reduce 112) ((T_INCLUDE_ONCE) reduce 112) ((T_INLINE_HTML) reduce 112) ((T_INTERFACE) reduce 112) ((T_INT_CAST) reduce 112) ((T_ISSET) reduce 112) ((T_LINE) reduce 112) ((T_LIST) reduce 112) ((T_LNUMBER) reduce 112) ((T_MATCH) reduce 112) ((T_METHOD_C) reduce 112) ((T_NAMESPACE) reduce 112) ((T_NAME_FULLY_QUALIFIED) reduce 112) ((T_NAME_QUALIFIED) reduce 112) ((T_NAME_RELATIVE) reduce 112) ((T_NEW) reduce 112) ((T_NS_C) reduce 112) ((T_OBJECT_CAST) reduce 112) ((T_PRINT) reduce 112) ((T_PROPERTY_C) reduce 112) ((T_READONLY) reduce 112) ((T_REQUIRE) reduce 112) ((T_REQUIRE_ONCE) reduce 112) ((T_RETURN) reduce 112) ((T_START_HEREDOC) reduce 112) ((T_STATIC) reduce 112) ((T_STRING) reduce 112) ((T_STRING_CAST) reduce 112) ((T_SWITCH) reduce 112) ((T_THROW) reduce 112) ((T_TRAIT) reduce 112) ((T_TRAIT_C) reduce 112) ((T_TRY) reduce 112) ((T_UNSET) reduce 112) ((T_UNSET_CAST) reduce 112) ((T_USE) reduce 112) ((T_VARIABLE) reduce 112) ((T_WHILE) reduce 112) ((T_YIELD) reduce 112) ((T_YIELD_FROM) reduce 112) (("[") reduce 112) (("`") reduce 112) (("{") reduce 112) (("~") reduce 112)) 30435 ((("!") reduce 113) (("\"") reduce 113) (("$") reduce 113) (($) reduce 113) (("(") reduce 113) (("+") reduce 113) (("-") reduce 113) ((";") reduce 113) (("@") reduce 113) ((T_ABSTRACT) reduce 113) ((T_ARRAY) reduce 113) ((T_ARRAY_CAST) reduce 113) ((T_ATTRIBUTE) reduce 113) ((T_BOOL_CAST) reduce 113) ((T_BREAK) reduce 113) ((T_CLASS) reduce 113) ((T_CLASS_C) reduce 113) ((T_CLONE) reduce 113) ((T_CONST) reduce 113) ((T_CONSTANT_ENCAPSED_STRING) reduce 113) ((T_CONTINUE) reduce 113) ((T_DEC) reduce 113) ((T_DECLARE) reduce 113) ((T_DIR) reduce 113) ((T_DNUMBER) reduce 113) ((T_DO) reduce 113) ((T_DOUBLE_CAST) reduce 113) ((T_ECHO) reduce 113) ((T_EMPTY) reduce 113) ((T_ENUM) reduce 113) ((T_EVAL) reduce 113) ((T_EXIT) reduce 113) ((T_FILE) reduce 113) ((T_FINAL) reduce 113) ((T_FN) reduce 113) ((T_FOR) reduce 113) ((T_FOREACH) reduce 113) ((T_FUNCTION) reduce 113) ((T_FUNC_C) reduce 113) ((T_GLOBAL) reduce 113) ((T_GOTO) reduce 113) ((T_HALT_COMPILER) reduce 113) ((T_IF) reduce 113) ((T_INC) reduce 113) ((T_INCLUDE) reduce 113) ((T_INCLUDE_ONCE) reduce 113) ((T_INLINE_HTML) reduce 113) ((T_INTERFACE) reduce 113) ((T_INT_CAST) reduce 113) ((T_ISSET) reduce 113) ((T_LINE) reduce 113) ((T_LIST) reduce 113) ((T_LNUMBER) reduce 113) ((T_MATCH) reduce 113) ((T_METHOD_C) reduce 113) ((T_NAMESPACE) reduce 113) ((T_NAME_FULLY_QUALIFIED) reduce 113) ((T_NAME_QUALIFIED) reduce 113) ((T_NAME_RELATIVE) reduce 113) ((T_NEW) reduce 113) ((T_NS_C) reduce 113) ((T_OBJECT_CAST) reduce 113) ((T_PRINT) reduce 113) ((T_PROPERTY_C) reduce 113) ((T_READONLY) reduce 113) ((T_REQUIRE) reduce 113) ((T_REQUIRE_ONCE) reduce 113) ((T_RETURN) reduce 113) ((T_START_HEREDOC) reduce 113) ((T_STATIC) reduce 113) ((T_STRING) reduce 113) ((T_STRING_CAST) reduce 113) ((T_SWITCH) reduce 113) ((T_THROW) reduce 113) ((T_TRAIT) reduce 113) ((T_TRAIT_C) reduce 113) ((T_TRY) reduce 113) ((T_UNSET) reduce 113) ((T_UNSET_CAST) reduce 113) ((T_USE) reduce 113) ((T_VARIABLE) reduce 113) ((T_WHILE) reduce 113) ((T_YIELD) reduce 113) ((T_YIELD_FROM) reduce 113) (("[") reduce 113) (("`") reduce 113) (("{") reduce 113) (("~") reduce 113)) 30457 ((("!") reduce 199) (("\"") reduce 199) (("$") reduce 199) (($) reduce 199) (("(") reduce 199) (("+") reduce 199) (("-") reduce 199) ((";") reduce 199) (("@") reduce 199) ((T_ABSTRACT) reduce 199) ((T_ARRAY) reduce 199) ((T_ARRAY_CAST) reduce 199) ((T_ATTRIBUTE) reduce 199) ((T_BOOL_CAST) reduce 199) ((T_BREAK) reduce 199) ((T_CLASS) reduce 199) ((T_CLASS_C) reduce 199) ((T_CLONE) reduce 199) ((T_CONST) reduce 199) ((T_CONSTANT_ENCAPSED_STRING) reduce 199) ((T_CONTINUE) reduce 199) ((T_DEC) reduce 199) ((T_DECLARE) reduce 199) ((T_DIR) reduce 199) ((T_DNUMBER) reduce 199) ((T_DO) reduce 199) ((T_DOUBLE_CAST) reduce 199) ((T_ECHO) reduce 199) ((T_EMPTY) reduce 199) ((T_ENUM) reduce 199) ((T_EVAL) reduce 199) ((T_EXIT) reduce 199) ((T_FILE) reduce 199) ((T_FINAL) reduce 199) ((T_FN) reduce 199) ((T_FOR) reduce 199) ((T_FOREACH) reduce 199) ((T_FUNCTION) reduce 199) ((T_FUNC_C) reduce 199) ((T_GLOBAL) reduce 199) ((T_GOTO) reduce 199) ((T_HALT_COMPILER) reduce 199) ((T_IF) reduce 199) ((T_INC) reduce 199) ((T_INCLUDE) reduce 199) ((T_INCLUDE_ONCE) reduce 199) ((T_INLINE_HTML) reduce 199) ((T_INTERFACE) reduce 199) ((T_INT_CAST) reduce 199) ((T_ISSET) reduce 199) ((T_LINE) reduce 199) ((T_LIST) reduce 199) ((T_LNUMBER) reduce 199) ((T_MATCH) reduce 199) ((T_METHOD_C) reduce 199) ((T_NAMESPACE) reduce 199) ((T_NAME_FULLY_QUALIFIED) reduce 199) ((T_NAME_QUALIFIED) reduce 199) ((T_NAME_RELATIVE) reduce 199) ((T_NEW) reduce 199) ((T_NS_C) reduce 199) ((T_OBJECT_CAST) reduce 199) ((T_PRINT) reduce 199) ((T_PROPERTY_C) reduce 199) ((T_READONLY) reduce 199) ((T_REQUIRE) reduce 199) ((T_REQUIRE_ONCE) reduce 199) ((T_RETURN) reduce 199) ((T_START_HEREDOC) reduce 199) ((T_STATIC) reduce 199) ((T_STRING) reduce 199) ((T_STRING_CAST) reduce 199) ((T_SWITCH) reduce 199) ((T_THROW) reduce 199) ((T_TRAIT) reduce 199) ((T_TRAIT_C) reduce 199) ((T_TRY) reduce 199) ((T_UNSET) reduce 199) ((T_UNSET_CAST) reduce 199) ((T_USE) reduce 199) ((T_VARIABLE) reduce 199) ((T_WHILE) reduce 199) ((T_YIELD) reduce 199) ((T_YIELD_FROM) reduce 199) (("[") reduce 199) (("`") reduce 199) (("{") reduce 199) (("~") reduce 199)) 30464 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (($) reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONST) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAMESPACE) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_USE) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 30475 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (($) reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONST) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAMESPACE) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_USE) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 30483 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (($) reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONST) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAMESPACE) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_USE) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 30486 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (($) reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONST) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAMESPACE) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_USE) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 30487 ((("!") reduce 244) (("\"") reduce 244) (($) reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONST) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAMESPACE) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_USE) reduce 244) ((T_VARIABLE) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 30488 ((("!") reduce 242) (("\"") reduce 242) (($) reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONST) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAMESPACE) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_USE) reduce 242) ((T_VARIABLE) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 30491 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (($) reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONST) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAMESPACE) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_USE) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 30497 ((("!") reduce 243) (("\"") reduce 243) (($) reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONST) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAMESPACE) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_USE) reduce 243) ((T_VARIABLE) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 30498 ((("!") reduce 245) (("\"") reduce 245) (($) reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONST) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAMESPACE) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_USE) reduce 245) ((T_VARIABLE) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 30499 ((("!") reduce 160) (("\"") reduce 160) (($) reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONST) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAMESPACE) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_USE) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 30502 ((("!") reduce 248) (("\"") reduce 248) (($) reduce 248) (("$") reduce 248) (("(") reduce 248) (("+") reduce 248) (("-") reduce 248) ((";") reduce 248) (("@") reduce 248) ((T_ABSTRACT) reduce 248) ((T_ARRAY) reduce 248) ((T_ARRAY_CAST) reduce 248) ((T_ATTRIBUTE) reduce 248) ((T_BOOL_CAST) reduce 248) ((T_BREAK) reduce 248) ((T_CLASS) reduce 248) ((T_CLASS_C) reduce 248) ((T_CLONE) reduce 248) ((T_CONST) reduce 248) ((T_CONSTANT_ENCAPSED_STRING) reduce 248) ((T_CONTINUE) reduce 248) ((T_DEC) reduce 248) ((T_DECLARE) reduce 248) ((T_DIR) reduce 248) ((T_DNUMBER) reduce 248) ((T_DO) reduce 248) ((T_DOUBLE_CAST) reduce 248) ((T_ECHO) reduce 248) ((T_ELSE) reduce 248) ((T_ELSEIF) reduce 248) ((T_EMPTY) reduce 248) ((T_ENUM) reduce 248) ((T_EVAL) reduce 248) ((T_EXIT) reduce 248) ((T_FILE) reduce 248) ((T_FINAL) reduce 248) ((T_FN) reduce 248) ((T_FOR) reduce 248) ((T_FOREACH) reduce 248) ((T_FUNCTION) reduce 248) ((T_FUNC_C) reduce 248) ((T_GLOBAL) reduce 248) ((T_GOTO) reduce 248) ((T_HALT_COMPILER) reduce 248) ((T_IF) reduce 248) ((T_INC) reduce 248) ((T_INCLUDE) reduce 248) ((T_INCLUDE_ONCE) reduce 248) ((T_INLINE_HTML) reduce 248) ((T_INTERFACE) reduce 248) ((T_INT_CAST) reduce 248) ((T_ISSET) reduce 248) ((T_LINE) reduce 248) ((T_LIST) reduce 248) ((T_LNUMBER) reduce 248) ((T_MATCH) reduce 248) ((T_METHOD_C) reduce 248) ((T_NAMESPACE) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_NEW) reduce 248) ((T_NS_C) reduce 248) ((T_OBJECT_CAST) reduce 248) ((T_PRINT) reduce 248) ((T_PROPERTY_C) reduce 248) ((T_READONLY) reduce 248) ((T_REQUIRE) reduce 248) ((T_REQUIRE_ONCE) reduce 248) ((T_RETURN) reduce 248) ((T_START_HEREDOC) reduce 248) ((T_STATIC) reduce 248) ((T_STRING) reduce 248) ((T_STRING_CAST) reduce 248) ((T_SWITCH) reduce 248) ((T_THROW) reduce 248) ((T_TRAIT) reduce 248) ((T_TRAIT_C) reduce 248) ((T_TRY) reduce 248) ((T_UNSET) reduce 248) ((T_UNSET_CAST) reduce 248) ((T_USE) reduce 248) ((T_VARIABLE) reduce 248) ((T_WHILE) reduce 248) ((T_YIELD) reduce 248) ((T_YIELD_FROM) reduce 248) (("[") reduce 248) (("`") reduce 248) (("{") reduce 248) (("~") reduce 248)) 30506 ((("!") reduce 249) (("\"") reduce 249) (("$") reduce 249) (($) reduce 249) (("(") reduce 249) (("+") reduce 249) (("-") reduce 249) ((";") reduce 249) (("@") reduce 249) ((T_ABSTRACT) reduce 249) ((T_ARRAY) reduce 249) ((T_ARRAY_CAST) reduce 249) ((T_ATTRIBUTE) reduce 249) ((T_BOOL_CAST) reduce 249) ((T_BREAK) reduce 249) ((T_CLASS) reduce 249) ((T_CLASS_C) reduce 249) ((T_CLONE) reduce 249) ((T_CONST) reduce 249) ((T_CONSTANT_ENCAPSED_STRING) reduce 249) ((T_CONTINUE) reduce 249) ((T_DEC) reduce 249) ((T_DECLARE) reduce 249) ((T_DIR) reduce 249) ((T_DNUMBER) reduce 249) ((T_DO) reduce 249) ((T_DOUBLE_CAST) reduce 249) ((T_ECHO) reduce 249) ((T_ELSE) reduce 249) ((T_ELSEIF) reduce 249) ((T_EMPTY) reduce 249) ((T_ENUM) reduce 249) ((T_EVAL) reduce 249) ((T_EXIT) reduce 249) ((T_FILE) reduce 249) ((T_FINAL) reduce 249) ((T_FN) reduce 249) ((T_FOR) reduce 249) ((T_FOREACH) reduce 249) ((T_FUNCTION) reduce 249) ((T_FUNC_C) reduce 249) ((T_GLOBAL) reduce 249) ((T_GOTO) reduce 249) ((T_HALT_COMPILER) reduce 249) ((T_IF) reduce 249) ((T_INC) reduce 249) ((T_INCLUDE) reduce 249) ((T_INCLUDE_ONCE) reduce 249) ((T_INLINE_HTML) reduce 249) ((T_INTERFACE) reduce 249) ((T_INT_CAST) reduce 249) ((T_ISSET) reduce 249) ((T_LINE) reduce 249) ((T_LIST) reduce 249) ((T_LNUMBER) reduce 249) ((T_MATCH) reduce 249) ((T_METHOD_C) reduce 249) ((T_NAMESPACE) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_NEW) reduce 249) ((T_NS_C) reduce 249) ((T_OBJECT_CAST) reduce 249) ((T_PRINT) reduce 249) ((T_PROPERTY_C) reduce 249) ((T_READONLY) reduce 249) ((T_REQUIRE) reduce 249) ((T_REQUIRE_ONCE) reduce 249) ((T_RETURN) reduce 249) ((T_START_HEREDOC) reduce 249) ((T_STATIC) reduce 249) ((T_STRING) reduce 249) ((T_STRING_CAST) reduce 249) ((T_SWITCH) reduce 249) ((T_THROW) reduce 249) ((T_TRAIT) reduce 249) ((T_TRAIT_C) reduce 249) ((T_TRY) reduce 249) ((T_UNSET) reduce 249) ((T_UNSET_CAST) reduce 249) ((T_USE) reduce 249) ((T_VARIABLE) reduce 249) ((T_WHILE) reduce 249) ((T_YIELD) reduce 249) ((T_YIELD_FROM) reduce 249) (("[") reduce 249) (("`") reduce 249) (("{") reduce 249) (("~") reduce 249)) 30511 ((("!") reduce 240) (("\"") reduce 240) (($) reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONST) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAMESPACE) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_PROPERTY_C) reduce 240) ((T_READONLY) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_USE) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 30512 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (($) reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONST) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAMESPACE) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_USE) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 30515 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (($) reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONST) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAMESPACE) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_PROPERTY_C) reduce 241) ((T_READONLY) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_USE) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 30520 ((("!") reduce 161) (("\"") reduce 161) (($) reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONST) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAMESPACE) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_USE) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 30523 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (($) reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONST) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAMESPACE) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_USE) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 30524 ((("!") reduce 175) (("\"") reduce 175) (($) reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CATCH) shift) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONST) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FINALLY) shift) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAMESPACE) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_PROPERTY_C) reduce 175) ((T_READONLY) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_USE) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 30527 ((("!") reduce 166) (("\"") reduce 166) (($) reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONST) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAMESPACE) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_USE) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 30530 ((("!") reduce 176) (("\"") reduce 176) (($) reduce 176) (("$") reduce 176) (("(") reduce 176) (("+") reduce 176) (("-") reduce 176) ((";") reduce 176) (("@") reduce 176) ((T_ABSTRACT) reduce 176) ((T_ARRAY) reduce 176) ((T_ARRAY_CAST) reduce 176) ((T_ATTRIBUTE) reduce 176) ((T_BOOL_CAST) reduce 176) ((T_BREAK) reduce 176) ((T_CLASS) reduce 176) ((T_CLASS_C) reduce 176) ((T_CLONE) reduce 176) ((T_CONST) reduce 176) ((T_CONSTANT_ENCAPSED_STRING) reduce 176) ((T_CONTINUE) reduce 176) ((T_DEC) reduce 176) ((T_DECLARE) reduce 176) ((T_DIR) reduce 176) ((T_DNUMBER) reduce 176) ((T_DO) reduce 176) ((T_DOUBLE_CAST) reduce 176) ((T_ECHO) reduce 176) ((T_ELSE) reduce 176) ((T_ELSEIF) reduce 176) ((T_EMPTY) reduce 176) ((T_ENUM) reduce 176) ((T_EVAL) reduce 176) ((T_EXIT) reduce 176) ((T_FILE) reduce 176) ((T_FINAL) reduce 176) ((T_FN) reduce 176) ((T_FOR) reduce 176) ((T_FOREACH) reduce 176) ((T_FUNCTION) reduce 176) ((T_FUNC_C) reduce 176) ((T_GLOBAL) reduce 176) ((T_GOTO) reduce 176) ((T_HALT_COMPILER) reduce 176) ((T_IF) reduce 176) ((T_INC) reduce 176) ((T_INCLUDE) reduce 176) ((T_INCLUDE_ONCE) reduce 176) ((T_INLINE_HTML) reduce 176) ((T_INTERFACE) reduce 176) ((T_INT_CAST) reduce 176) ((T_ISSET) reduce 176) ((T_LINE) reduce 176) ((T_LIST) reduce 176) ((T_LNUMBER) reduce 176) ((T_MATCH) reduce 176) ((T_METHOD_C) reduce 176) ((T_NAMESPACE) reduce 176) ((T_NAME_FULLY_QUALIFIED) reduce 176) ((T_NAME_QUALIFIED) reduce 176) ((T_NAME_RELATIVE) reduce 176) ((T_NEW) reduce 176) ((T_NS_C) reduce 176) ((T_OBJECT_CAST) reduce 176) ((T_PRINT) reduce 176) ((T_PROPERTY_C) reduce 176) ((T_READONLY) reduce 176) ((T_REQUIRE) reduce 176) ((T_REQUIRE_ONCE) reduce 176) ((T_RETURN) reduce 176) ((T_START_HEREDOC) reduce 176) ((T_STATIC) reduce 176) ((T_STRING) reduce 176) ((T_STRING_CAST) reduce 176) ((T_SWITCH) reduce 176) ((T_THROW) reduce 176) ((T_TRAIT) reduce 176) ((T_TRAIT_C) reduce 176) ((T_TRY) reduce 176) ((T_UNSET) reduce 176) ((T_UNSET_CAST) reduce 176) ((T_USE) reduce 176) ((T_VARIABLE) reduce 176) ((T_WHILE) reduce 176) ((T_YIELD) reduce 176) ((T_YIELD_FROM) reduce 176) (("[") reduce 176) (("`") reduce 176) (("{") reduce 176) (("~") reduce 176)) 30537 ((("!") reduce 170) (("\"") reduce 170) (($) reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CATCH) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONST) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FINALLY) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAMESPACE) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_USE) reduce 170) ((T_VARIABLE) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 30542 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (($) reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONST) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAMESPACE) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_USE) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 30546 ((("!") reduce 222) (("\"") reduce 222) (($) reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONST) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAMESPACE) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_USE) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 30548 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (($) reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONST) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAMESPACE) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_USE) reduce 223) ((T_VARIABLE) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 30552 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (($) reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONST) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAMESPACE) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_USE) reduce 224) ((T_VARIABLE) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 30555 ((("!") reduce 225) (("\"") reduce 225) (($) reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONST) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAMESPACE) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_USE) reduce 225) ((T_VARIABLE) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 30556 ((("!") reduce 168) (("\"") reduce 168) (($) reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONST) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAMESPACE) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_USE) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 30558 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (($) reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONST) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAMESPACE) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_USE) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 30560 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (($) reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONST) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAMESPACE) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_USE) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 30564 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (($) reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONST) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAMESPACE) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_USE) reduce 242) ((T_VARIABLE) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 30566 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (($) reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONST) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAMESPACE) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_USE) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 30568 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (($) reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONST) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAMESPACE) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_USE) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 30578 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (($) reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONST) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAMESPACE) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_USE) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 30579 ((("!") reduce 218) (("\"") reduce 218) (($) reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONST) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAMESPACE) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_USE) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 30582 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (($) reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONST) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAMESPACE) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_USE) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 30583 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (($) reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONST) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAMESPACE) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_USE) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 30592 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (($) reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONST) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAMESPACE) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_USE) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 30593 ((("!") reduce 216) (("\"") reduce 216) (($) reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONST) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAMESPACE) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_USE) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 30596 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (($) reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONST) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAMESPACE) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_USE) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 30598 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (($) reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONST) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAMESPACE) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_USE) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 30604 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (($) reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONST) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAMESPACE) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_USE) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 30609 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (($) reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONST) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAMESPACE) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_USE) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 30610 ((("!") reduce 220) (("\"") reduce 220) (($) reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONST) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAMESPACE) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_USE) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 30613 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (($) reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONST) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAMESPACE) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_USE) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 30615 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (($) reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONST) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAMESPACE) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_USE) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 30617 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (($) reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONST) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAMESPACE) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_USE) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 30620 ((("!") reduce 111) (("\"") reduce 111) (($) reduce 111) (("$") reduce 111) (("(") reduce 111) (("+") reduce 111) (("-") reduce 111) ((";") reduce 111) (("@") reduce 111) ((T_ABSTRACT) reduce 111) ((T_ARRAY) reduce 111) ((T_ARRAY_CAST) reduce 111) ((T_ATTRIBUTE) reduce 111) ((T_BOOL_CAST) reduce 111) ((T_BREAK) reduce 111) ((T_CLASS) reduce 111) ((T_CLASS_C) reduce 111) ((T_CLONE) reduce 111) ((T_CONST) reduce 111) ((T_CONSTANT_ENCAPSED_STRING) reduce 111) ((T_CONTINUE) reduce 111) ((T_DEC) reduce 111) ((T_DECLARE) reduce 111) ((T_DIR) reduce 111) ((T_DNUMBER) reduce 111) ((T_DO) reduce 111) ((T_DOUBLE_CAST) reduce 111) ((T_ECHO) reduce 111) ((T_EMPTY) reduce 111) ((T_ENUM) reduce 111) ((T_EVAL) reduce 111) ((T_EXIT) reduce 111) ((T_FILE) reduce 111) ((T_FINAL) reduce 111) ((T_FN) reduce 111) ((T_FOR) reduce 111) ((T_FOREACH) reduce 111) ((T_FUNCTION) reduce 111) ((T_FUNC_C) reduce 111) ((T_GLOBAL) reduce 111) ((T_GOTO) reduce 111) ((T_HALT_COMPILER) reduce 111) ((T_IF) reduce 111) ((T_INC) reduce 111) ((T_INCLUDE) reduce 111) ((T_INCLUDE_ONCE) reduce 111) ((T_INLINE_HTML) reduce 111) ((T_INTERFACE) reduce 111) ((T_INT_CAST) reduce 111) ((T_ISSET) reduce 111) ((T_LINE) reduce 111) ((T_LIST) reduce 111) ((T_LNUMBER) reduce 111) ((T_MATCH) reduce 111) ((T_METHOD_C) reduce 111) ((T_NAMESPACE) reduce 111) ((T_NAME_FULLY_QUALIFIED) reduce 111) ((T_NAME_QUALIFIED) reduce 111) ((T_NAME_RELATIVE) reduce 111) ((T_NEW) reduce 111) ((T_NS_C) reduce 111) ((T_OBJECT_CAST) reduce 111) ((T_PRINT) reduce 111) ((T_PROPERTY_C) reduce 111) ((T_READONLY) reduce 111) ((T_REQUIRE) reduce 111) ((T_REQUIRE_ONCE) reduce 111) ((T_RETURN) reduce 111) ((T_START_HEREDOC) reduce 111) ((T_STATIC) reduce 111) ((T_STRING) reduce 111) ((T_STRING_CAST) reduce 111) ((T_SWITCH) reduce 111) ((T_THROW) reduce 111) ((T_TRAIT) reduce 111) ((T_TRAIT_C) reduce 111) ((T_TRY) reduce 111) ((T_UNSET) reduce 111) ((T_UNSET_CAST) reduce 111) ((T_USE) reduce 111) ((T_VARIABLE) reduce 111) ((T_WHILE) reduce 111) ((T_YIELD) reduce 111) ((T_YIELD_FROM) reduce 111) (("[") reduce 111) (("`") reduce 111) (("{") reduce 111) (("~") reduce 111)) 30622 ((("!") reduce 167) (("\"") reduce 167) (($) reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONST) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAMESPACE) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_USE) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 30624 ((("!") reduce 156) (("\"") reduce 156) (($) reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONST) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAMESPACE) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_USE) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 30634 ((("!") reduce 163) (("\"") reduce 163) (($) reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONST) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAMESPACE) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_USE) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 30635 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (($) reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONST) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAMESPACE) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_USE) reduce 218) ((T_VARIABLE) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 30638 ((("!") reduce 219) (("\"") reduce 219) (($) reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONST) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAMESPACE) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_USE) reduce 219) ((T_VARIABLE) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 30639 ((("!") reduce 162) (("\"") reduce 162) (($) reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONST) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAMESPACE) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_USE) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 30648 ((("!") reduce 151) (("\"") reduce 151) (($) reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONST) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAMESPACE) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_USE) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 30649 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (($) reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONST) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAMESPACE) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_PROPERTY_C) reduce 216) ((T_READONLY) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_USE) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 30652 ((("!") reduce 217) (("\"") reduce 217) (($) reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONST) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAMESPACE) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_PROPERTY_C) reduce 217) ((T_READONLY) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_USE) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 30671 ((("!") reduce 200) (("\"") reduce 200) (($) reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONST) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAMESPACE) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_USE) reduce 200) ((T_VARIABLE) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 30676 ((("!") reduce 158) (("\"") reduce 158) (($) reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONST) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAMESPACE) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_USE) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 30683 ((("!") reduce 150) (("\"") reduce 150) (($) reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONST) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAMESPACE) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_USE) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 30688 ((("!") reduce 164) (("\"") reduce 164) (($) reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONST) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAMESPACE) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_USE) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 30689 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (($) reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONST) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAMESPACE) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_USE) reduce 220) ((T_VARIABLE) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 30692 ((("!") reduce 221) (("\"") reduce 221) (($) reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONST) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAMESPACE) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_USE) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 30695 ((("!") reduce 154) (("\"") reduce 154) (($) reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONST) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAMESPACE) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_USE) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 30697 ((("!") reduce 119) (("\"") reduce 119) (("$") reduce 119) (($) reduce 119) (("(") reduce 119) (("+") reduce 119) (("-") reduce 119) ((";") reduce 119) (("@") reduce 119) ((T_ABSTRACT) reduce 119) ((T_ARRAY) reduce 119) ((T_ARRAY_CAST) reduce 119) ((T_ATTRIBUTE) reduce 119) ((T_BOOL_CAST) reduce 119) ((T_BREAK) reduce 119) ((T_CLASS) reduce 119) ((T_CLASS_C) reduce 119) ((T_CLONE) reduce 119) ((T_CONST) reduce 119) ((T_CONSTANT_ENCAPSED_STRING) reduce 119) ((T_CONTINUE) reduce 119) ((T_DEC) reduce 119) ((T_DECLARE) reduce 119) ((T_DIR) reduce 119) ((T_DNUMBER) reduce 119) ((T_DO) reduce 119) ((T_DOUBLE_CAST) reduce 119) ((T_ECHO) reduce 119) ((T_EMPTY) reduce 119) ((T_ENUM) reduce 119) ((T_EVAL) reduce 119) ((T_EXIT) reduce 119) ((T_FILE) reduce 119) ((T_FINAL) reduce 119) ((T_FN) reduce 119) ((T_FOR) reduce 119) ((T_FOREACH) reduce 119) ((T_FUNCTION) reduce 119) ((T_FUNC_C) reduce 119) ((T_GLOBAL) reduce 119) ((T_GOTO) reduce 119) ((T_HALT_COMPILER) reduce 119) ((T_IF) reduce 119) ((T_INC) reduce 119) ((T_INCLUDE) reduce 119) ((T_INCLUDE_ONCE) reduce 119) ((T_INLINE_HTML) reduce 119) ((T_INTERFACE) reduce 119) ((T_INT_CAST) reduce 119) ((T_ISSET) reduce 119) ((T_LINE) reduce 119) ((T_LIST) reduce 119) ((T_LNUMBER) reduce 119) ((T_MATCH) reduce 119) ((T_METHOD_C) reduce 119) ((T_NAMESPACE) reduce 119) ((T_NAME_FULLY_QUALIFIED) reduce 119) ((T_NAME_QUALIFIED) reduce 119) ((T_NAME_RELATIVE) reduce 119) ((T_NEW) reduce 119) ((T_NS_C) reduce 119) ((T_OBJECT_CAST) reduce 119) ((T_PRINT) reduce 119) ((T_PROPERTY_C) reduce 119) ((T_READONLY) reduce 119) ((T_REQUIRE) reduce 119) ((T_REQUIRE_ONCE) reduce 119) ((T_RETURN) reduce 119) ((T_START_HEREDOC) reduce 119) ((T_STATIC) reduce 119) ((T_STRING) reduce 119) ((T_STRING_CAST) reduce 119) ((T_SWITCH) reduce 119) ((T_THROW) reduce 119) ((T_TRAIT) reduce 119) ((T_TRAIT_C) reduce 119) ((T_TRY) reduce 119) ((T_UNSET) reduce 119) ((T_UNSET_CAST) reduce 119) ((T_USE) reduce 119) ((T_VARIABLE) reduce 119) ((T_WHILE) reduce 119) ((T_YIELD) reduce 119) ((T_YIELD_FROM) reduce 119) (("[") reduce 119) (("`") reduce 119) (("{") reduce 119) (("~") reduce 119)) 30705 ((("!") reduce 188) (("\"") reduce 188) (($) reduce 188) (("$") reduce 188) (("(") reduce 188) (("+") reduce 188) (("-") reduce 188) ((";") reduce 188) (("@") reduce 188) ((T_ABSTRACT) reduce 188) ((T_ARRAY) reduce 188) ((T_ARRAY_CAST) reduce 188) ((T_ATTRIBUTE) reduce 188) ((T_BOOL_CAST) reduce 188) ((T_BREAK) reduce 188) ((T_CLASS) reduce 188) ((T_CLASS_C) reduce 188) ((T_CLONE) reduce 188) ((T_CONST) reduce 188) ((T_CONSTANT_ENCAPSED_STRING) reduce 188) ((T_CONTINUE) reduce 188) ((T_DEC) reduce 188) ((T_DECLARE) reduce 188) ((T_DIR) reduce 188) ((T_DNUMBER) reduce 188) ((T_DO) reduce 188) ((T_DOUBLE_CAST) reduce 188) ((T_ECHO) reduce 188) ((T_EMPTY) reduce 188) ((T_ENUM) reduce 188) ((T_EVAL) reduce 188) ((T_EXIT) reduce 188) ((T_FILE) reduce 188) ((T_FINAL) reduce 188) ((T_FN) reduce 188) ((T_FOR) reduce 188) ((T_FOREACH) reduce 188) ((T_FUNCTION) reduce 188) ((T_FUNC_C) reduce 188) ((T_GLOBAL) reduce 188) ((T_GOTO) reduce 188) ((T_HALT_COMPILER) reduce 188) ((T_IF) reduce 188) ((T_INC) reduce 188) ((T_INCLUDE) reduce 188) ((T_INCLUDE_ONCE) reduce 188) ((T_INLINE_HTML) reduce 188) ((T_INTERFACE) reduce 188) ((T_INT_CAST) reduce 188) ((T_ISSET) reduce 188) ((T_LINE) reduce 188) ((T_LIST) reduce 188) ((T_LNUMBER) reduce 188) ((T_MATCH) reduce 188) ((T_METHOD_C) reduce 188) ((T_NAMESPACE) reduce 188) ((T_NAME_FULLY_QUALIFIED) reduce 188) ((T_NAME_QUALIFIED) reduce 188) ((T_NAME_RELATIVE) reduce 188) ((T_NEW) reduce 188) ((T_NS_C) reduce 188) ((T_OBJECT_CAST) reduce 188) ((T_PRINT) reduce 188) ((T_PROPERTY_C) reduce 188) ((T_READONLY) reduce 188) ((T_REQUIRE) reduce 188) ((T_REQUIRE_ONCE) reduce 188) ((T_RETURN) reduce 188) ((T_START_HEREDOC) reduce 188) ((T_STATIC) reduce 188) ((T_STRING) reduce 188) ((T_STRING_CAST) reduce 188) ((T_SWITCH) reduce 188) ((T_THROW) reduce 188) ((T_TRAIT) reduce 188) ((T_TRAIT_C) reduce 188) ((T_TRY) reduce 188) ((T_UNSET) reduce 188) ((T_UNSET_CAST) reduce 188) ((T_USE) reduce 188) ((T_VARIABLE) reduce 188) ((T_WHILE) reduce 188) ((T_YIELD) reduce 188) ((T_YIELD_FROM) reduce 188) (("[") reduce 188) (("`") reduce 188) (("{") reduce 188) (("~") reduce 188)) 30707 ((("!") reduce 153) (("\"") reduce 153) (($) reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONST) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAMESPACE) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_USE) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 30711 (((T_ABSTRACT) reduce 100) ((T_ATTRIBUTE) reduce 100) ((T_CLASS) reduce 100) ((T_ENUM) reduce 100) ((T_FINAL) reduce 100) ((T_FN) reduce 100) ((T_FUNCTION) reduce 100) ((T_INTERFACE) reduce 100) ((T_READONLY) reduce 100) ((T_STATIC) reduce 100) ((T_TRAIT) reduce 100))))
  "The generated distinct action-tables.")

(defvar
  phps-mode-parser--goto-tables
  #s(hash-table size 42072 test equal rehash-size 1.5 rehash-threshold 0.8125 data (0 0 1 1 2 2 3 3 4 4 5 5 6 6 7 7 8 8 9 9 10 8 11 10 12 11 13 12 14 13 15 14 16 15 17 8 18 16 19 17 20 8 21 18 22 19 23 20 24 8 25 8 26 21 27 22 28 23 29 24 30 25 31 26 32 27 33 8 34 8 35 8 36 28 37 29 38 8 39 8 40 30 41 31 42 32 43 33 44 34 45 35 46 36 47 8 48 37 49 38 50 39 51 8 52 40 53 8 54 41 55 8 56 42 57 8 58 8 59 8 60 43 61 8 62 44 63 45 64 8 65 46 66 47 67 48 68 49 69 50 70 51 71 52 72 53 73 54 74 55 75 56 76 8 77 57 78 58 79 59 80 60 81 8 82 61 83 62 84 63 85 64 86 65 87 8 88 66 89 67 90 8 91 8 92 68 93 69 94 8 95 8 96 8 97 8 98 70 99 71 100 8 101 8 102 8 103 72 104 73 105 8 106 74 107 8 108 8 109 8 110 75 111 8 112 8 113 8 114 8 115 76 116 8 117 8 118 8 119 8 120 8 121 8 122 8 123 8 124 77 125 78 126 79 127 80 128 81 129 8 130 46 131 82 132 8 133 8 134 83 135 84 136 85 137 8 138 8 139 8 140 86 141 87 142 88 143 8 144 89 145 8 146 8 147 8 148 8 149 8 150 8 151 8 152 8 153 8 154 8 155 8 156 90 157 8 158 91 159 92 160 93 161 8 162 94 163 95 164 96 165 97 166 98 167 99 168 100 169 101 170 102 171 8 172 103 173 104 174 105 175 106 176 107 177 108 178 109 179 110 180 111 181 112 182 113 183 114 184 115 185 8 186 116 187 117 188 118 189 119 190 120 191 121 192 122 193 123 194 124 195 8 196 125 197 8 198 126 199 8 200 127 201 8 202 128 203 129 204 8 205 8 206 130 207 8 208 8 209 8 210 8 211 131 212 132 213 8 214 133 215 8 216 134 217 135 218 136 219 137 220 138 221 139 222 140 223 141 224 142 225 143 226 144 227 145 228 146 229 8 230 147 231 8 232 148 233 8 234 8 235 149 236 150 237 151 238 152 239 8 240 8 241 153 242 154 243 155 244 156 245 157 246 8 247 158 248 8 249 159 250 8 251 8 252 8 253 8 254 160 255 8 256 161 257 162 258 8 259 163 260 164 261 165 262 166 263 167 264 8 265 168 266 169 267 8 268 170 269 8 270 171 271 172 272 173 273 174 274 175 275 176 276 177 277 8 278 8 279 178 280 8 281 8 282 179 283 180 284 181 285 8 286 8 287 8 288 8 289 182 290 8 291 8 292 8 293 8 294 8 295 183 296 184 297 185 298 186 299 187 300 188 301 189 302 190 303 191 304 192 305 193 306 194 307 195 308 196 309 197 310 198 311 199 312 200 313 201 314 202 315 203 316 204 317 205 318 206 319 207 320 208 321 209 322 210 323 211 324 212 325 213 326 214 327 215 328 216 329 186 330 186 331 186 332 186 333 186 334 186 335 186 336 186 337 186 338 186 339 186 340 186 341 186 342 186 343 186 344 217 345 218 346 8 347 8 348 8 349 8 350 8 351 8 352 219 353 8 354 8 355 220 356 8 357 221 358 222 359 223 360 224 361 225 362 226 363 227 364 228 365 229 366 230 367 231 368 232 369 233 370 234 371 8 372 235 373 8 374 236 375 8 376 8 377 237 378 238 379 239 380 240 381 8 382 8 383 241 384 242 385 243 386 244 387 245 388 8 389 246 390 8 391 247 392 8 393 8 394 8 395 8 396 248 397 8 398 249 399 250 400 8 401 251 402 252 403 253 404 254 405 255 406 8 407 256 408 257 409 8 410 258 411 8 412 259 413 260 414 261 415 262 416 263 417 264 418 265 419 8 420 8 421 266 422 8 423 8 424 267 425 268 426 269 427 8 428 8 429 8 430 8 431 270 432 8 433 8 434 271 435 8 436 8 437 8 438 272 439 273 440 274 441 267 442 275 443 276 444 277 445 278 446 279 447 280 448 281 449 282 450 283 451 284 452 285 453 286 454 287 455 288 456 289 457 290 458 291 459 292 460 293 461 294 462 295 463 296 464 297 465 298 466 299 467 300 468 301 469 302 470 303 471 304 472 267 473 267 474 267 475 267 476 267 477 267 478 267 479 267 480 267 481 267 482 267 483 267 484 267 485 267 486 267 487 305 488 306 489 8 490 8 491 8 492 8 493 8 494 8 495 307 496 8 497 8 498 308 499 8 500 309 501 310 502 311 503 312 504 313 505 8 506 8 507 8 508 8 509 8 510 314 511 315 512 316 513 317 514 318 515 319 516 320 517 321 518 322 519 323 520 324 521 8 522 325 523 8 524 326 525 8 526 8 527 327 528 328 529 329 530 330 531 8 532 8 533 331 534 332 535 333 536 334 537 335 538 8 539 336 540 8 541 337 542 8 543 8 544 8 545 8 546 338 547 8 548 339 549 340 550 8 551 341 552 342 553 343 554 344 555 345 556 8 557 346 558 347 559 8 560 348 561 8 562 349 563 350 564 351 565 352 566 353 567 354 568 355 569 8 570 8 571 356 572 8 573 8 574 357 575 358 576 359 577 8 578 8 579 8 580 8 581 360 582 8 583 8 584 8 585 8 586 8 587 361 588 362 589 363 590 364 591 365 592 366 593 367 594 368 595 369 596 370 597 371 598 372 599 373 600 374 601 375 602 376 603 377 604 378 605 379 606 380 607 381 608 382 609 383 610 384 611 385 612 386 613 387 614 388 615 389 616 390 617 391 618 392 619 393 620 394 621 364 622 364 623 364 624 364 625 364 626 364 627 364 628 364 629 364 630 364 631 364 632 364 633 364 634 364 635 364 636 395 637 396 638 8 639 8 640 8 641 8 642 8 643 8 644 397 645 8 646 8 647 398 648 8 649 399 650 400 651 401 652 402 653 403 654 8 655 8 656 8 657 8 658 8 659 404 660 405 661 8 662 8 663 406 664 8 665 407 666 8 667 8 668 408 669 409 670 8 671 364 672 364 673 364 674 364 675 364 676 410 677 411 678 412 679 413 680 414 681 415 682 416 683 417 684 418 685 419 686 420 687 8 688 421 689 8 690 422 691 8 692 8 693 423 694 424 695 425 696 426 697 8 698 8 699 427 700 428 701 429 702 430 703 431 704 8 705 432 706 8 707 433 708 8 709 8 710 8 711 8 712 434 713 8 714 435 715 436 716 8 717 437 718 438 719 439 720 440 721 441 722 8 723 442 724 443 725 8 726 444 727 8 728 445 729 446 730 447 731 448 732 449 733 450 734 451 735 8 736 8 737 452 738 8 739 8 740 453 741 454 742 455 743 8 744 8 745 8 746 8 747 456 748 8 749 8 750 8 751 8 752 8 753 457 754 458 755 459 756 460 757 461 758 462 759 463 760 464 761 465 762 466 763 467 764 468 765 469 766 470 767 471 768 472 769 473 770 474 771 475 772 476 773 477 774 478 775 479 776 480 777 481 778 482 779 483 780 484 781 485 782 486 783 487 784 488 785 489 786 490 787 460 788 460 789 460 790 460 791 460 792 460 793 460 794 460 795 460 796 460 797 460 798 460 799 460 800 460 801 460 802 491 803 492 804 8 805 8 806 8 807 8 808 8 809 8 810 493 811 8 812 8 813 494 814 8 815 495 816 496 817 497 818 498 819 499 820 8 821 8 822 8 823 8 824 8 825 500 826 501 827 8 828 8 829 502 830 8 831 503 832 8 833 8 834 504 835 505 836 8 837 460 838 460 839 460 840 460 841 460 842 506 843 507 844 508 845 460 846 460 847 460 848 460 849 460 850 460 851 460 852 460 853 460 854 460 855 509 856 8 857 8 858 8 859 8 860 8 861 8 862 8 863 8 864 8 865 8 866 8 867 8 868 8 869 8 870 8 871 8 872 8 873 8 874 8 875 8 876 8 877 8 878 8 879 8 880 8 881 8 882 8 883 8 884 8 885 8 886 8 887 8 888 8 889 8 890 8 891 8 892 8 893 8 894 8 895 8 896 8 897 8 898 8 899 8 900 8 901 8 902 8 903 8 904 8 905 8 906 8 907 8 908 8 909 8 910 8 911 8 912 8 913 8 914 8 915 8 916 8 917 8 918 8 919 8 920 8 921 8 922 8 923 8 924 8 925 8 926 8 927 8 928 8 929 8 930 8 931 8 932 8 933 8 934 8 935 8 936 510 937 8 938 8 939 8 940 511 941 512 942 8 943 513 944 8 945 514 946 515 947 516 948 517 949 8 950 518 951 519 952 520 953 8 954 521 955 522 956 8 957 523 958 8 959 8 960 8 961 8 962 8 963 8 964 524 965 8 966 8 967 8 968 525 969 8 970 8 971 8 972 8 973 8 974 526 975 8 976 527 977 8 978 8 979 528 980 8 981 8 982 8 983 8 984 8 985 8 986 8 987 529 988 530 989 8 990 8 991 8 992 8 993 8 994 8 995 8 996 8 997 8 998 8 999 8 1000 8 1001 8 1002 531 1003 532 1004 533 1005 8 1006 8 1007 8 1008 534 1009 535 1010 8 1011 536 1012 8 1013 8 1014 8 1015 8 1016 537 1017 8 1018 8 1019 8 1020 8 1021 8 1022 538 1023 8 1024 539 1025 540 1026 8 1027 8 1028 8 1029 8 1030 541 1031 542 1032 543 1033 8 1034 544 1035 545 1036 8 1037 546 1038 8 1039 547 1040 8 1041 8 1042 8 1043 8 1044 548 1045 8 1046 8 1047 8 1048 8 1049 549 1050 550 1051 551 1052 552 1053 8 1054 553 1055 554 1056 555 1057 8 1058 8 1059 556 1060 8 1061 8 1062 557 1063 558 1064 559 1065 8 1066 560 1067 8 1068 8 1069 8 1070 561 1071 8 1072 8 1073 562 1074 8 1075 8 1076 8 1077 8 1078 8 1079 563 1080 564 1081 565 1082 521 1083 8 1084 524 1085 8 1086 528 1087 529 1088 530 1089 8 1090 8 1091 532 1092 533 1093 535 1094 8 1095 536 1096 537 1097 8 1098 538 1099 8 1100 540 1101 8 1102 541 1103 542 1104 543 1105 545 1106 8 1107 547 1108 8 1109 549 1110 557 1111 566 1112 567 1113 568 1114 569 1115 570 1116 571 1117 572 1118 573 1119 574 1120 575 1121 576 1122 577 1123 578 1124 579 1125 580 1126 581 1127 582 1128 583 1129 584 1130 585 1131 586 1132 587 1133 588 1134 589 1135 590 1136 591 1137 592 1138 593 1139 594 1140 595 1141 557 1142 557 1143 557 1144 557 1145 557 1146 557 1147 557 1148 557 1149 557 1150 557 1151 557 1152 557 1153 557 1154 557 1155 557 1156 596 1157 597 1158 8 1159 8 1160 8 1161 8 1162 8 1163 8 1164 598 1165 8 1166 8 1167 599 1168 8 1169 600 1170 601 1171 602 1172 603 1173 604 1174 8 1175 8 1176 8 1177 8 1178 8 1179 605 1180 606 1181 8 1182 8 1183 607 1184 8 1185 608 1186 8 1187 8 1188 609 1189 610 1190 8 1191 557 1192 557 1193 557 1194 557 1195 557 1196 611 1197 612 1198 613 1199 557 1200 557 1201 557 1202 557 1203 557 1204 557 1205 557 1206 557 1207 557 1208 557 1209 614 1210 615 1211 616 1212 617 1213 618 1214 619 1215 620 1216 621 1217 622 1218 623 1219 8 1220 624 1221 8 1222 625 1223 8 1224 8 1225 626 1226 627 1227 628 1228 629 1229 8 1230 8 1231 630 1232 631 1233 632 1234 633 1235 634 1236 8 1237 635 1238 8 1239 636 1240 8 1241 8 1242 8 1243 8 1244 637 1245 8 1246 638 1247 639 1248 8 1249 640 1250 641 1251 642 1252 643 1253 644 1254 8 1255 645 1256 646 1257 8 1258 647 1259 8 1260 648 1261 649 1262 650 1263 651 1264 652 1265 653 1266 654 1267 8 1268 8 1269 655 1270 8 1271 8 1272 656 1273 657 1274 658 1275 8 1276 8 1277 8 1278 8 1279 659 1280 8 1281 8 1282 8 1283 8 1284 8 1285 660 1286 661 1287 662 1288 663 1289 664 1290 665 1291 666 1292 667 1293 668 1294 669 1295 670 1296 671 1297 672 1298 673 1299 674 1300 675 1301 676 1302 677 1303 678 1304 679 1305 680 1306 681 1307 682 1308 683 1309 684 1310 685 1311 686 1312 687 1313 688 1314 689 1315 690 1316 691 1317 692 1318 693 1319 663 1320 663 1321 663 1322 663 1323 663 1324 663 1325 663 1326 663 1327 663 1328 663 1329 663 1330 663 1331 663 1332 663 1333 663 1334 694 1335 695 1336 8 1337 8 1338 8 1339 8 1340 8 1341 8 1342 696 1343 8 1344 8 1345 697 1346 8 1347 698 1348 699 1349 700 1350 701 1351 702 1352 8 1353 8 1354 8 1355 8 1356 8 1357 703 1358 704 1359 8 1360 8 1361 705 1362 8 1363 706 1364 8 1365 8 1366 707 1367 708 1368 8 1369 663 1370 663 1371 663 1372 663 1373 663 1374 709 1375 710 1376 711 1377 663 1378 663 1379 663 1380 663 1381 663 1382 663 1383 663 1384 663 1385 663 1386 663 1387 712 1388 8 1389 8 1390 8 1391 8 1392 8 1393 8 1394 8 1395 8 1396 8 1397 8 1398 8 1399 8 1400 8 1401 8 1402 8 1403 8 1404 8 1405 8 1406 8 1407 8 1408 8 1409 8 1410 8 1411 8 1412 8 1413 8 1414 8 1415 8 1416 8 1417 8 1418 8 1419 8 1420 8 1421 8 1422 8 1423 8 1424 8 1425 8 1426 8 1427 8 1428 8 1429 8 1430 8 1431 8 1432 8 1433 8 1434 8 1435 8 1436 8 1437 8 1438 8 1439 8 1440 8 1441 8 1442 8 1443 8 1444 8 1445 8 1446 8 1447 8 1448 8 1449 8 1450 8 1451 8 1452 8 1453 8 1454 8 1455 8 1456 8 1457 8 1458 8 1459 8 1460 8 1461 8 1462 8 1463 8 1464 8 1465 8 1466 8 1467 8 1468 713 1469 8 1470 8 1471 8 1472 714 1473 715 1474 8 1475 716 1476 8 1477 8 1478 717 1479 718 1480 719 1481 720 1482 8 1483 721 1484 8 1485 557 1486 8 1487 722 1488 723 1489 724 1490 725 1491 8 1492 726 1493 8 1494 727 1495 728 1496 729 1497 730 1498 731 1499 732 1500 733 1501 734 1502 735 1503 663 1504 663 1505 663 1506 663 1507 663 1508 663 1509 663 1510 663 1511 663 1512 663 1513 663 1514 663 1515 663 1516 8 1517 8 1518 736 1519 663 1520 737 1521 738 1522 739 1523 740 1524 8 1525 8 1526 8 1527 8 1528 8 1529 8 1530 8 1531 8 1532 8 1533 8 1534 741 1535 8 1536 8 1537 742 1538 8 1539 8 1540 8 1541 8 1542 743 1543 744 1544 745 1545 8 1546 8 1547 746 1548 8 1549 8 1550 8 1551 747 1552 8 1553 8 1554 8 1555 8 1556 748 1557 749 1558 8 1559 8 1560 8 1561 8 1562 8 1563 8 1564 8 1565 8 1566 8 1567 8 1568 8 1569 8 1570 8 1571 8 1572 8 1573 8 1574 8 1575 8 1576 8 1577 8 1578 8 1579 8 1580 8 1581 8 1582 8 1583 8 1584 8 1585 8 1586 8 1587 8 1588 8 1589 8 1590 8 1591 8 1592 8 1593 8 1594 8 1595 8 1596 8 1597 8 1598 8 1599 8 1600 8 1601 8 1602 8 1603 8 1604 8 1605 8 1606 8 1607 8 1608 8 1609 8 1610 8 1611 8 1612 8 1613 8 1614 8 1615 8 1616 8 1617 8 1618 8 1619 8 1620 8 1621 8 1622 8 1623 8 1624 8 1625 8 1626 8 1627 8 1628 8 1629 8 1630 8 1631 8 1632 8 1633 8 1634 8 1635 8 1636 8 1637 8 1638 750 1639 8 1640 8 1641 8 1642 751 1643 752 1644 8 1645 753 1646 8 1647 8 1648 754 1649 755 1650 756 1651 8 1652 8 1653 8 1654 757 1655 8 1656 758 1657 8 1658 759 1659 760 1660 8 1661 8 1662 8 1663 761 1664 762 1665 763 1666 764 1667 8 1668 8 1669 765 1670 8 1671 766 1672 767 1673 8 1674 8 1675 768 1676 8 1677 769 1678 770 1679 771 1680 772 1681 773 1682 774 1683 775 1684 776 1685 777 1686 778 1687 8 1688 779 1689 8 1690 780 1691 8 1692 8 1693 781 1694 782 1695 783 1696 784 1697 785 1698 8 1699 8 1700 786 1701 787 1702 788 1703 789 1704 790 1705 8 1706 791 1707 8 1708 792 1709 8 1710 8 1711 8 1712 8 1713 793 1714 8 1715 794 1716 795 1717 8 1718 796 1719 797 1720 798 1721 799 1722 800 1723 8 1724 801 1725 802 1726 8 1727 803 1728 8 1729 804 1730 805 1731 806 1732 807 1733 808 1734 809 1735 8 1736 810 1737 811 1738 812 1739 8 1740 8 1741 813 1742 8 1743 8 1744 814 1745 815 1746 816 1747 8 1748 8 1749 8 1750 8 1751 817 1752 8 1753 8 1754 818 1755 8 1756 8 1757 8 1758 8 1759 819 1760 820 1761 821 1762 822 1763 823 1764 824 1765 825 1766 826 1767 827 1768 828 1769 829 1770 830 1771 831 1772 832 1773 833 1774 834 1775 835 1776 836 1777 837 1778 838 1779 839 1780 840 1781 841 1782 842 1783 843 1784 844 1785 845 1786 846 1787 847 1788 848 1789 849 1790 850 1791 851 1792 852 1793 853 1794 823 1795 823 1796 823 1797 823 1798 823 1799 823 1800 823 1801 823 1802 823 1803 823 1804 823 1805 823 1806 823 1807 823 1808 823 1809 854 1810 855 1811 8 1812 8 1813 8 1814 8 1815 8 1816 8 1817 856 1818 8 1819 8 1820 857 1821 8 1822 858 1823 859 1824 860 1825 861 1826 862 1827 8 1828 8 1829 8 1830 8 1831 8 1832 863 1833 864 1834 8 1835 8 1836 865 1837 8 1838 866 1839 8 1840 8 1841 867 1842 868 1843 8 1844 823 1845 823 1846 823 1847 823 1848 823 1849 869 1850 870 1851 871 1852 823 1853 823 1854 823 1855 823 1856 823 1857 823 1858 823 1859 823 1860 823 1861 823 1862 872 1863 873 1864 874 1865 875 1866 876 1867 877 1868 878 1869 879 1870 880 1871 881 1872 882 1873 8 1874 883 1875 8 1876 884 1877 8 1878 8 1879 885 1880 886 1881 887 1882 888 1883 889 1884 8 1885 8 1886 890 1887 891 1888 892 1889 893 1890 894 1891 8 1892 895 1893 8 1894 896 1895 8 1896 8 1897 8 1898 8 1899 897 1900 8 1901 898 1902 899 1903 8 1904 900 1905 901 1906 902 1907 903 1908 904 1909 8 1910 905 1911 906 1912 8 1913 907 1914 8 1915 908 1916 909 1917 910 1918 911 1919 912 1920 913 1921 8 1922 914 1923 915 1924 916 1925 8 1926 8 1927 917 1928 8 1929 8 1930 918 1931 919 1932 920 1933 8 1934 8 1935 8 1936 8 1937 921 1938 8 1939 8 1940 922 1941 8 1942 8 1943 8 1944 8 1945 923 1946 924 1947 925 1948 926 1949 927 1950 928 1951 929 1952 930 1953 931 1954 932 1955 933 1956 934 1957 935 1958 936 1959 937 1960 938 1961 939 1962 940 1963 941 1964 942 1965 943 1966 944 1967 945 1968 946 1969 947 1970 948 1971 949 1972 950 1973 951 1974 952 1975 953 1976 954 1977 955 1978 956 1979 957 1980 927 1981 927 1982 927 1983 927 1984 927 1985 927 1986 927 1987 927 1988 927 1989 927 1990 927 1991 927 1992 927 1993 927 1994 927 1995 958 1996 959 1997 8 1998 8 1999 8 2000 8 2001 8 2002 8 2003 960 2004 8 2005 8 2006 961 2007 8 2008 962 2009 963 2010 964 2011 965 2012 966 2013 8 2014 8 2015 8 2016 8 2017 8 2018 967 2019 968 2020 8 2021 8 2022 969 2023 8 2024 970 2025 8 2026 8 2027 971 2028 972 2029 8 2030 927 2031 927 2032 927 2033 927 2034 927 2035 973 2036 974 2037 975 2038 927 2039 927 2040 927 2041 927 2042 927 2043 927 2044 927 2045 927 2046 927 2047 927 2048 976 2049 977 2050 978 2051 979 2052 927 2053 980 2054 8 2055 8 2056 8 2057 8 2058 8 2059 8 2060 8 2061 8 2062 8 2063 8 2064 8 2065 8 2066 8 2067 8 2068 8 2069 8 2070 8 2071 8 2072 8 2073 8 2074 8 2075 8 2076 8 2077 8 2078 8 2079 8 2080 8 2081 8 2082 8 2083 8 2084 8 2085 8 2086 8 2087 8 2088 8 2089 8 2090 8 2091 8 2092 8 2093 8 2094 8 2095 8 2096 8 2097 8 2098 8 2099 8 2100 8 2101 8 2102 8 2103 8 2104 8 2105 8 2106 8 2107 8 2108 8 2109 8 2110 8 2111 8 2112 8 2113 8 2114 8 2115 8 2116 8 2117 8 2118 8 2119 8 2120 8 2121 8 2122 8 2123 8 2124 8 2125 8 2126 8 2127 8 2128 8 2129 8 2130 8 2131 8 2132 8 2133 8 2134 981 2135 8 2136 8 2137 8 2138 982 2139 983 2140 8 2141 984 2142 8 2143 8 2144 985 2145 986 2146 987 2147 8 2148 8 2149 988 2150 989 2151 990 2152 991 2153 8 2154 992 2155 8 2156 993 2157 994 2158 995 2159 996 2160 997 2161 998 2162 999 2163 1000 2164 1001 2165 927 2166 927 2167 927 2168 927 2169 927 2170 927 2171 927 2172 927 2173 927 2174 927 2175 927 2176 927 2177 927 2178 1002 2179 927 2180 1003 2181 1004 2182 8 2183 1005 2184 1006 2185 8 2186 1007 2187 8 2188 1008 2189 8 2190 8 2191 8 2192 1009 2193 1010 2194 1011 2195 8 2196 1012 2197 8 2198 8 2199 1013 2200 1014 2201 1015 2202 8 2203 8 2204 8 2205 1016 2206 1017 2207 8 2208 8 2209 8 2210 1018 2211 1019 2212 1020 2213 1021 2214 8 2215 8 2216 1022 2217 8 2218 1023 2219 1024 2220 8 2221 8 2222 1025 2223 8 2224 8 2225 8 2226 1026 2227 1027 2228 8 2229 1028 2230 8 2231 8 2232 1029 2233 1030 2234 1031 2235 1032 2236 1033 2237 1034 2238 1035 2239 1036 2240 1037 2241 1038 2242 1039 2243 8 2244 1040 2245 1041 2246 1042 2247 1043 2248 1044 2249 1045 2250 1046 2251 1047 2252 1048 2253 1049 2254 1050 2255 8 2256 1051 2257 1052 2258 8 2259 8 2260 8 2261 8 2262 8 2263 8 2264 8 2265 1053 2266 8 2267 1054 2268 8 2269 8 2270 1055 2271 1056 2272 1057 2273 1058 2274 8 2275 8 2276 8 2277 8 2278 8 2279 8 2280 8 2281 8 2282 8 2283 8 2284 8 2285 8 2286 8 2287 8 2288 8 2289 8 2290 8 2291 8 2292 1059 2293 8 2294 1060 2295 8 2296 1061 2297 8 2298 8 2299 8 2300 8 2301 8 2302 8 2303 8 2304 8 2305 8 2306 8 2307 8 2308 1062 2309 8 2310 1063 2311 8 2312 1064 2313 8 2314 8 2315 8 2316 8 2317 8 2318 8 2319 8 2320 8 2321 8 2322 8 2323 1065 2324 8 2325 8 2326 8 2327 8 2328 8 2329 8 2330 8 2331 8 2332 8 2333 8 2334 8 2335 1066 2336 8 2337 1067 2338 1068 2339 1069 2340 1070 2341 1071 2342 1072 2343 1073 2344 1074 2345 1075 2346 1076 2347 1077 2348 8 2349 1078 2350 8 2351 1079 2352 8 2353 8 2354 1080 2355 1081 2356 1082 2357 1083 2358 8 2359 8 2360 1084 2361 1085 2362 1086 2363 1087 2364 1088 2365 8 2366 1089 2367 8 2368 1090 2369 8 2370 8 2371 8 2372 8 2373 1091 2374 8 2375 1092 2376 1093 2377 8 2378 1094 2379 1095 2380 1096 2381 1097 2382 1098 2383 8 2384 1099 2385 1100 2386 8 2387 1101 2388 8 2389 1102 2390 1103 2391 1104 2392 1105 2393 1106 2394 1107 2395 1108 2396 1109 2397 8 2398 8 2399 1110 2400 8 2401 8 2402 1111 2403 1112 2404 1113 2405 8 2406 8 2407 8 2408 8 2409 1114 2410 8 2411 8 2412 8 2413 8 2414 8 2415 1115 2416 1116 2417 1117 2418 1118 2419 1111 2420 1119 2421 1120 2422 1121 2423 1122 2424 1123 2425 1124 2426 1125 2427 1126 2428 1127 2429 1128 2430 1129 2431 1130 2432 1131 2433 1132 2434 1133 2435 1134 2436 1135 2437 1136 2438 1137 2439 1138 2440 1139 2441 1140 2442 1141 2443 1142 2444 1143 2445 1144 2446 1145 2447 1146 2448 1147 2449 1148 2450 1111 2451 1111 2452 1111 2453 1111 2454 1111 2455 1111 2456 1111 2457 1111 2458 1111 2459 1111 2460 1111 2461 1111 2462 1111 2463 1111 2464 1111 2465 1149 2466 1150 2467 8 2468 8 2469 8 2470 8 2471 8 2472 8 2473 1151 2474 8 2475 8 2476 1152 2477 8 2478 1153 2479 1154 2480 1155 2481 1156 2482 1157 2483 8 2484 8 2485 8 2486 8 2487 8 2488 1158 2489 1159 2490 8 2491 8 2492 1160 2493 8 2494 1161 2495 8 2496 8 2497 1162 2498 1163 2499 8 2500 1111 2501 1111 2502 1111 2503 1111 2504 1111 2505 1164 2506 1165 2507 1166 2508 1111 2509 1111 2510 1111 2511 1111 2512 1111 2513 1111 2514 1111 2515 1111 2516 1111 2517 1111 2518 1167 2519 1168 2520 1169 2521 1170 2522 1111 2523 1171 2524 8 2525 8 2526 8 2527 8 2528 8 2529 8 2530 8 2531 8 2532 8 2533 8 2534 8 2535 8 2536 8 2537 8 2538 8 2539 8 2540 8 2541 8 2542 8 2543 8 2544 8 2545 8 2546 8 2547 8 2548 8 2549 8 2550 8 2551 8 2552 8 2553 8 2554 8 2555 8 2556 8 2557 8 2558 8 2559 8 2560 8 2561 8 2562 8 2563 8 2564 8 2565 8 2566 8 2567 8 2568 8 2569 8 2570 8 2571 8 2572 8 2573 8 2574 8 2575 8 2576 8 2577 8 2578 8 2579 8 2580 8 2581 8 2582 8 2583 8 2584 8 2585 8 2586 8 2587 8 2588 8 2589 8 2590 8 2591 8 2592 8 2593 8 2594 8 2595 8 2596 8 2597 8 2598 8 2599 8 2600 8 2601 8 2602 8 2603 8 2604 1172 2605 8 2606 8 2607 8 2608 1173 2609 1174 2610 8 2611 1175 2612 8 2613 8 2614 1176 2615 1177 2616 1178 2617 8 2618 8 2619 1179 2620 1180 2621 1181 2622 1182 2623 8 2624 1183 2625 8 2626 1184 2627 1185 2628 1186 2629 1187 2630 1188 2631 1189 2632 1190 2633 1191 2634 1192 2635 1111 2636 1111 2637 1111 2638 1111 2639 1111 2640 1111 2641 1111 2642 1111 2643 1111 2644 1111 2645 1111 2646 1111 2647 1111 2648 1193 2649 1111 2650 1194 2651 1195 2652 8 2653 1196 2654 1197 2655 8 2656 1198 2657 8 2658 1199 2659 8 2660 8 2661 8 2662 1200 2663 1201 2664 1202 2665 8 2666 1203 2667 8 2668 8 2669 1204 2670 1205 2671 1206 2672 8 2673 8 2674 8 2675 1207 2676 1208 2677 8 2678 8 2679 8 2680 1209 2681 1210 2682 1211 2683 1212 2684 8 2685 8 2686 1213 2687 8 2688 1214 2689 1215 2690 8 2691 8 2692 1216 2693 8 2694 8 2695 8 2696 1217 2697 1218 2698 8 2699 8 2700 1219 2701 1220 2702 1221 2703 1222 2704 1223 2705 1224 2706 1225 2707 1226 2708 1227 2709 1228 2710 1229 2711 8 2712 1230 2713 1231 2714 1232 2715 1233 2716 1234 2717 1235 2718 1236 2719 1237 2720 1238 2721 1239 2722 8 2723 1240 2724 8 2725 1241 2726 8 2727 1242 2728 1243 2729 8 2730 8 2731 8 2732 1244 2733 8 2734 8 2735 8 2736 1245 2737 1246 2738 1247 2739 1248 2740 8 2741 8 2742 1249 2743 8 2744 1250 2745 1251 2746 8 2747 8 2748 1252 2749 8 2750 1253 2751 1254 2752 8 2753 1255 2754 1256 2755 8 2756 1257 2757 8 2758 1258 2759 8 2760 8 2761 8 2762 1259 2763 1260 2764 1261 2765 8 2766 1262 2767 8 2768 8 2769 1263 2770 1264 2771 1265 2772 8 2773 8 2774 8 2775 1266 2776 1267 2777 8 2778 8 2779 8 2780 1268 2781 1269 2782 1270 2783 1271 2784 8 2785 8 2786 1272 2787 8 2788 1273 2789 1274 2790 8 2791 8 2792 1275 2793 8 2794 8 2795 8 2796 1276 2797 1277 2798 8 2799 1278 2800 1279 2801 1280 2802 1281 2803 1282 2804 1283 2805 8 2806 8 2807 8 2808 8 2809 1284 2810 1285 2811 1286 2812 1287 2813 8 2814 8 2815 8 2816 1288 2817 8 2818 8 2819 8 2820 8 2821 8 2822 1289 2823 1290 2824 8 2825 1291 2826 8 2827 1292 2828 1293 2829 8 2830 1294 2831 1295 2832 8 2833 1296 2834 1297 2835 8 2836 1298 2837 8 2838 1299 2839 8 2840 8 2841 1300 2842 1301 2843 1302 2844 1303 2845 8 2846 1304 2847 8 2848 8 2849 1305 2850 1306 2851 1307 2852 8 2853 8 2854 8 2855 8 2856 1308 2857 1309 2858 8 2859 8 2860 8 2861 1310 2862 1311 2863 1312 2864 1313 2865 8 2866 8 2867 1314 2868 8 2869 1315 2870 1316 2871 8 2872 8 2873 1317 2874 8 2875 8 2876 8 2877 1318 2878 1319 2879 8 2880 1320 2881 1321 2882 1322 2883 1111 2884 1111 2885 1323 2886 1324 2887 1111 2888 1111 2889 1111 2890 1111 2891 8 2892 1325 2893 1326 2894 1327 2895 8 2896 1328 2897 1329 2898 8 2899 8 2900 8 2901 8 2902 1330 2903 1331 2904 1332 2905 1333 2906 8 2907 8 2908 8 2909 8 2910 8 2911 1334 2912 1335 2913 8 2914 1336 2915 8 2916 1337 2917 1338 2918 8 2919 1339 2920 8 2921 1111 2922 1111 2923 8 2924 1111 2925 1111 2926 1340 2927 1341 2928 8 2929 8 2930 8 2931 8 2932 8 2933 8 2934 8 2935 1342 2936 1343 2937 8 2938 1344 2939 8 2940 1345 2941 1346 2942 8 2943 1347 2944 8 2945 1348 2946 1349 2947 1350 2948 1351 2949 1352 2950 8 2951 8 2952 8 2953 8 2954 8 2955 1353 2956 8 2957 8 2958 8 2959 1354 2960 1355 2961 8 2962 1356 2963 1357 2964 1358 2965 1359 2966 1360 2967 1361 2968 1362 2969 1363 2970 8 2971 1364 2972 1365 2973 8 2974 8 2975 8 2976 8 2977 8 2978 8 2979 8 2980 8 2981 8 2982 1366 2983 8 2984 8 2985 8 2986 1367 2987 1368 2988 8 2989 8 2990 8 2991 1369 2992 1370 2993 1371 2994 8 2995 1372 2996 1373 2997 8 2998 8 2999 8 3000 8 3001 8 3002 8 3003 1374 3004 8 3005 1375 3006 8 3007 1376 3008 1377 3009 1378 3010 1379 3011 1380 3012 8 3013 8 3014 8 3015 8 3016 8 3017 8 3018 8 3019 8 3020 8 3021 1381 3022 1382 3023 1383 3024 8 3025 8 3026 1384 3027 8 3028 1385 3029 8 3030 1386 3031 8 3032 8 3033 8 3034 8 3035 8 3036 8 3037 8 3038 8 3039 1387 3040 8 3041 8 3042 1388 3043 1389 3044 8 3045 1390 3046 8 3047 1391 3048 8 3049 1392 3050 1393 3051 1394 3052 1395 3053 1396 3054 1397 3055 1398 3056 1399 3057 1400 3058 1401 3059 8 3060 1402 3061 8 3062 1403 3063 8 3064 8 3065 1404 3066 1405 3067 1406 3068 1407 3069 8 3070 8 3071 1408 3072 1409 3073 1410 3074 1411 3075 1412 3076 8 3077 1413 3078 8 3079 1414 3080 8 3081 8 3082 8 3083 8 3084 1415 3085 8 3086 1416 3087 1417 3088 8 3089 1418 3090 1419 3091 1420 3092 1421 3093 1422 3094 8 3095 1423 3096 1424 3097 8 3098 1425 3099 8 3100 1426 3101 1427 3102 1428 3103 1429 3104 1430 3105 1431 3106 1432 3107 8 3108 8 3109 1433 3110 8 3111 8 3112 1434 3113 1435 3114 1436 3115 8 3116 8 3117 8 3118 8 3119 1437 3120 8 3121 8 3122 8 3123 8 3124 8 3125 1438 3126 1439 3127 1440 3128 1441 3129 1442 3130 1443 3131 1444 3132 1445 3133 1446 3134 1447 3135 1448 3136 1449 3137 1450 3138 1451 3139 1452 3140 1453 3141 1454 3142 1455 3143 1456 3144 1457 3145 1458 3146 1459 3147 1460 3148 1461 3149 1462 3150 1463 3151 1464 3152 1465 3153 1466 3154 1467 3155 1468 3156 1469 3157 1470 3158 1471 3159 1441 3160 1441 3161 1441 3162 1441 3163 1441 3164 1441 3165 1441 3166 1441 3167 1441 3168 1441 3169 1441 3170 1441 3171 1441 3172 1441 3173 1441 3174 1472 3175 1473 3176 8 3177 8 3178 8 3179 8 3180 8 3181 8 3182 1474 3183 8 3184 8 3185 1475 3186 8 3187 1476 3188 1477 3189 1478 3190 1479 3191 1480 3192 8 3193 8 3194 8 3195 8 3196 8 3197 1481 3198 1482 3199 8 3200 8 3201 1483 3202 8 3203 1484 3204 8 3205 8 3206 1485 3207 1486 3208 8 3209 1441 3210 1441 3211 1441 3212 1441 3213 1441 3214 1487 3215 1488 3216 1489 3217 1441 3218 1441 3219 1441 3220 1441 3221 1441 3222 1441 3223 1441 3224 1441 3225 1441 3226 1441 3227 1490 3228 8 3229 8 3230 8 3231 8 3232 8 3233 8 3234 8 3235 8 3236 8 3237 8 3238 8 3239 8 3240 8 3241 8 3242 8 3243 8 3244 8 3245 8 3246 8 3247 8 3248 8 3249 8 3250 8 3251 8 3252 8 3253 8 3254 8 3255 8 3256 8 3257 8 3258 8 3259 8 3260 8 3261 8 3262 8 3263 8 3264 8 3265 8 3266 8 3267 8 3268 8 3269 8 3270 8 3271 8 3272 8 3273 8 3274 8 3275 8 3276 8 3277 8 3278 8 3279 8 3280 8 3281 8 3282 8 3283 8 3284 8 3285 8 3286 8 3287 8 3288 8 3289 8 3290 8 3291 8 3292 8 3293 8 3294 8 3295 8 3296 8 3297 8 3298 8 3299 8 3300 8 3301 8 3302 8 3303 8 3304 8 3305 8 3306 8 3307 8 3308 1491 3309 8 3310 8 3311 8 3312 1492 3313 1493 3314 8 3315 1494 3316 8 3317 8 3318 1495 3319 1496 3320 1497 3321 8 3322 8 3323 1498 3324 1499 3325 1500 3326 1501 3327 8 3328 1502 3329 8 3330 1503 3331 1504 3332 1505 3333 1506 3334 1507 3335 1508 3336 1509 3337 1510 3338 1511 3339 1441 3340 1441 3341 1441 3342 1441 3343 1441 3344 1441 3345 1441 3346 1441 3347 1441 3348 1441 3349 1441 3350 1441 3351 1441 3352 1512 3353 1441 3354 1513 3355 1514 3356 8 3357 1515 3358 1516 3359 8 3360 1517 3361 8 3362 1518 3363 8 3364 8 3365 8 3366 1519 3367 1520 3368 1521 3369 8 3370 1522 3371 8 3372 8 3373 1523 3374 1524 3375 1525 3376 8 3377 8 3378 8 3379 1526 3380 1527 3381 8 3382 8 3383 8 3384 1528 3385 1529 3386 1530 3387 1531 3388 8 3389 8 3390 1532 3391 8 3392 1533 3393 1534 3394 8 3395 8 3396 1535 3397 8 3398 8 3399 8 3400 1536 3401 1537 3402 8 3403 8 3404 1538 3405 1539 3406 1540 3407 1541 3408 1542 3409 1543 3410 1544 3411 1545 3412 1546 3413 1547 3414 1548 3415 8 3416 1549 3417 1550 3418 1551 3419 1552 3420 1553 3421 1554 3422 1555 3423 1556 3424 1557 3425 1558 3426 8 3427 8 3428 1559 3429 8 3430 1560 3431 8 3432 1561 3433 1562 3434 8 3435 8 3436 8 3437 1563 3438 8 3439 8 3440 1564 3441 1565 3442 1566 3443 1567 3444 8 3445 8 3446 1568 3447 8 3448 1569 3449 1570 3450 8 3451 8 3452 1571 3453 8 3454 1572 3455 8 3456 1573 3457 1574 3458 1575 3459 1441 3460 1441 3461 1576 3462 1577 3463 1578 3464 1579 3465 1580 3466 1581 3467 1582 3468 1583 3469 1584 3470 1585 3471 8 3472 1586 3473 8 3474 1587 3475 8 3476 8 3477 1588 3478 1589 3479 1590 3480 1591 3481 8 3482 8 3483 1592 3484 1593 3485 1594 3486 1595 3487 1596 3488 8 3489 1597 3490 8 3491 1598 3492 8 3493 8 3494 8 3495 8 3496 1599 3497 8 3498 1600 3499 1601 3500 8 3501 1602 3502 1603 3503 1604 3504 1605 3505 1606 3506 8 3507 1607 3508 1608 3509 8 3510 1609 3511 8 3512 1610 3513 1611 3514 1612 3515 1613 3516 1614 3517 1615 3518 1616 3519 8 3520 8 3521 1617 3522 8 3523 8 3524 1618 3525 1619 3526 1620 3527 8 3528 8 3529 8 3530 8 3531 1621 3532 8 3533 8 3534 8 3535 8 3536 8 3537 1622 3538 1623 3539 1624 3540 1625 3541 1626 3542 1627 3543 1628 3544 1629 3545 1630 3546 1631 3547 1632 3548 1633 3549 1634 3550 1635 3551 1636 3552 1637 3553 1638 3554 1639 3555 1640 3556 1641 3557 1642 3558 1643 3559 1644 3560 1645 3561 1646 3562 1647 3563 1648 3564 1649 3565 1650 3566 1651 3567 1652 3568 1653 3569 1654 3570 1655 3571 1625 3572 1625 3573 1625 3574 1625 3575 1625 3576 1625 3577 1625 3578 1625 3579 1625 3580 1625 3581 1625 3582 1625 3583 1625 3584 1625 3585 1625 3586 1656 3587 1657 3588 8 3589 8 3590 8 3591 8 3592 8 3593 8 3594 1658 3595 8 3596 8 3597 1659 3598 8 3599 1660 3600 1661 3601 1662 3602 1663 3603 1664 3604 8 3605 8 3606 8 3607 8 3608 8 3609 1665 3610 1666 3611 8 3612 8 3613 1667 3614 8 3615 1668 3616 8 3617 8 3618 1669 3619 1670 3620 8 3621 1625 3622 1625 3623 1625 3624 1625 3625 1625 3626 1671 3627 1672 3628 1673 3629 1625 3630 1625 3631 1625 3632 1625 3633 1625 3634 1625 3635 1625 3636 1625 3637 1625 3638 1625 3639 1674 3640 8 3641 8 3642 8 3643 8 3644 8 3645 8 3646 8 3647 8 3648 8 3649 8 3650 8 3651 8 3652 8 3653 8 3654 8 3655 8 3656 8 3657 8 3658 8 3659 8 3660 8 3661 8 3662 8 3663 8 3664 8 3665 8 3666 8 3667 8 3668 8 3669 8 3670 8 3671 8 3672 8 3673 8 3674 8 3675 8 3676 8 3677 8 3678 8 3679 8 3680 8 3681 8 3682 8 3683 8 3684 8 3685 8 3686 8 3687 8 3688 8 3689 8 3690 8 3691 8 3692 8 3693 8 3694 8 3695 8 3696 8 3697 8 3698 8 3699 8 3700 8 3701 8 3702 8 3703 8 3704 8 3705 8 3706 8 3707 8 3708 8 3709 8 3710 8 3711 8 3712 8 3713 8 3714 8 3715 8 3716 8 3717 8 3718 8 3719 8 3720 1675 3721 8 3722 8 3723 8 3724 1676 3725 1677 3726 8 3727 1678 3728 8 3729 8 3730 1679 3731 1680 3732 1681 3733 8 3734 8 3735 1682 3736 1683 3737 1684 3738 1685 3739 8 3740 1686 3741 8 3742 1687 3743 1688 3744 1689 3745 1690 3746 1691 3747 1692 3748 1693 3749 1694 3750 1695 3751 1625 3752 1625 3753 1625 3754 1625 3755 1625 3756 1625 3757 1625 3758 1625 3759 1625 3760 1625 3761 1625 3762 1625 3763 1625 3764 1696 3765 1625 3766 1697 3767 1698 3768 8 3769 1699 3770 1700 3771 8 3772 1701 3773 8 3774 1702 3775 8 3776 8 3777 8 3778 1703 3779 1704 3780 1705 3781 8 3782 1706 3783 8 3784 8 3785 1707 3786 1708 3787 1709 3788 8 3789 8 3790 8 3791 1710 3792 1711 3793 8 3794 8 3795 8 3796 1712 3797 1713 3798 1714 3799 1715 3800 8 3801 8 3802 1716 3803 8 3804 1717 3805 1718 3806 8 3807 8 3808 1719 3809 8 3810 8 3811 8 3812 1720 3813 1721 3814 8 3815 8 3816 1722 3817 1723 3818 1724 3819 1725 3820 1726 3821 1727 3822 1728 3823 1729 3824 1730 3825 1731 3826 1732 3827 8 3828 1733 3829 1734 3830 1735 3831 1736 3832 1737 3833 1738 3834 1739 3835 1740 3836 1741 3837 1742 3838 8 3839 1743 3840 1441 3841 1744 3842 8 3843 1745 3844 8 3845 1746 3846 1747 3847 8 3848 8 3849 8 3850 1748 3851 8 3852 8 3853 1749 3854 1750 3855 1751 3856 1752 3857 8 3858 8 3859 1753 3860 8 3861 1754 3862 1755 3863 8 3864 8 3865 1756 3866 8 3867 1757 3868 8 3869 1758 3870 1759 3871 1760 3872 1625 3873 1625 3874 1761 3875 1762 3876 1625 3877 1625 3878 1625 3879 1625 3880 8 3881 1763 3882 8 3883 1764 3884 8 3885 8 3886 1625 3887 1625 3888 8 3889 1625 3890 1625 3891 1765 3892 8 3893 8 3894 8 3895 8 3896 8 3897 8 3898 1766 3899 1767 3900 1768 3901 1769 3902 8 3903 1770 3904 8 3905 1771 3906 1772 3907 1773 3908 1774 3909 1775 3910 8 3911 8 3912 8 3913 8 3914 8 3915 1776 3916 8 3917 8 3918 1777 3919 1778 3920 1779 3921 1780 3922 1781 3923 1782 3924 1783 3925 8 3926 1784 3927 8 3928 1785 3929 1786 3930 1787 3931 1788 3932 1789 3933 1790 3934 1791 3935 1792 3936 1793 3937 1794 3938 1795 3939 1796 3940 1797 3941 1798 3942 8 3943 1799 3944 8 3945 1800 3946 1801 3947 8 3948 8 3949 1802 3950 1803 3951 1804 3952 1805 3953 8 3954 8 3955 1806 3956 1807 3957 1808 3958 1809 3959 1810 3960 8 3961 1811 3962 8 3963 1812 3964 8 3965 8 3966 8 3967 8 3968 1813 3969 8 3970 1814 3971 1815 3972 8 3973 1816 3974 1817 3975 1818 3976 1819 3977 1820 3978 8 3979 1821 3980 1822 3981 8 3982 1823 3983 8 3984 1824 3985 1825 3986 1826 3987 1827 3988 1828 3989 1829 3990 1830 3991 8 3992 8 3993 1831 3994 8 3995 8 3996 1832 3997 1833 3998 1834 3999 8 4000 8 4001 8 4002 8 4003 8 4004 1835 4005 1836 4006 1837 4007 8 4008 8 4009 1838 4010 8 4011 8 4012 8 4013 1839 4014 1840 4015 1841 4016 1832 4017 1842 4018 1843 4019 1844 4020 1845 4021 1846 4022 1847 4023 1848 4024 1849 4025 1850 4026 1851 4027 1852 4028 1853 4029 1854 4030 1855 4031 1856 4032 1857 4033 1858 4034 1859 4035 1860 4036 1861 4037 1862 4038 1863 4039 1864 4040 1865 4041 1866 4042 1867 4043 1868 4044 1869 4045 1870 4046 1871 4047 1832 4048 1832 4049 1832 4050 1832 4051 1832 4052 1832 4053 1832 4054 1832 4055 1832 4056 1832 4057 1832 4058 1832 4059 1832 4060 1832 4061 1832 4062 1872 4063 1873 4064 8 4065 8 4066 8 4067 8 4068 8 4069 8 4070 1874 4071 8 4072 8 4073 1875 4074 8 4075 1876 4076 1877 4077 1878 4078 1879 4079 1880 4080 8 4081 8 4082 8 4083 8 4084 8 4085 1881 4086 1882 4087 8 4088 8 4089 1883 4090 8 4091 1884 4092 8 4093 8 4094 1885 4095 1886 4096 8 4097 1832 4098 1832 4099 1832 4100 1832 4101 1832 4102 1887 4103 1888 4104 1889 4105 1832 4106 1832 4107 1832 4108 1832 4109 1832 4110 1832 4111 1832 4112 1832 4113 1832 4114 1832 4115 1890 4116 8 4117 8 4118 8 4119 8 4120 8 4121 8 4122 8 4123 8 4124 8 4125 8 4126 8 4127 8 4128 8 4129 8 4130 8 4131 8 4132 8 4133 8 4134 8 4135 8 4136 8 4137 8 4138 8 4139 8 4140 8 4141 8 4142 8 4143 8 4144 8 4145 8 4146 8 4147 8 4148 8 4149 8 4150 8 4151 8 4152 8 4153 8 4154 8 4155 8 4156 8 4157 8 4158 8 4159 8 4160 8 4161 8 4162 8 4163 8 4164 8 4165 8 4166 8 4167 8 4168 8 4169 8 4170 8 4171 8 4172 8 4173 8 4174 8 4175 8 4176 8 4177 8 4178 8 4179 8 4180 8 4181 8 4182 8 4183 8 4184 8 4185 8 4186 8 4187 8 4188 8 4189 8 4190 8 4191 8 4192 8 4193 8 4194 8 4195 8 4196 1891 4197 8 4198 8 4199 8 4200 1892 4201 1893 4202 8 4203 1894 4204 8 4205 8 4206 1895 4207 1896 4208 1897 4209 8 4210 8 4211 1898 4212 1899 4213 1900 4214 1901 4215 8 4216 1902 4217 8 4218 1903 4219 1904 4220 1905 4221 1906 4222 1907 4223 1908 4224 1909 4225 1910 4226 1911 4227 1832 4228 1832 4229 1832 4230 1832 4231 1832 4232 1832 4233 1832 4234 1832 4235 1832 4236 1832 4237 1832 4238 1832 4239 1832 4240 1912 4241 1832 4242 1913 4243 1914 4244 8 4245 1915 4246 1916 4247 8 4248 1917 4249 8 4250 1918 4251 8 4252 8 4253 8 4254 1919 4255 1920 4256 1921 4257 8 4258 1922 4259 8 4260 8 4261 1923 4262 1924 4263 1925 4264 8 4265 8 4266 8 4267 1926 4268 1927 4269 8 4270 8 4271 8 4272 1928 4273 1929 4274 1930 4275 1931 4276 8 4277 8 4278 1932 4279 8 4280 1933 4281 1934 4282 8 4283 8 4284 1935 4285 8 4286 8 4287 8 4288 1936 4289 1937 4290 8 4291 1938 4292 8 4293 8 4294 8 4295 8 4296 1939 4297 1940 4298 1941 4299 1942 4300 1943 4301 1944 4302 1945 4303 1946 4304 1947 4305 1948 4306 1949 4307 1950 4308 1951 4309 8 4310 1952 4311 8 4312 1953 4313 8 4314 8 4315 1954 4316 1955 4317 1956 4318 1957 4319 8 4320 8 4321 1958 4322 1959 4323 1960 4324 1961 4325 1962 4326 8 4327 1963 4328 8 4329 1964 4330 8 4331 8 4332 8 4333 8 4334 1965 4335 8 4336 1966 4337 1967 4338 8 4339 1968 4340 1969 4341 1970 4342 1971 4343 1972 4344 8 4345 1973 4346 1974 4347 8 4348 1975 4349 8 4350 1976 4351 1977 4352 1978 4353 1979 4354 1980 4355 1981 4356 1982 4357 8 4358 8 4359 1983 4360 8 4361 8 4362 1984 4363 1985 4364 1986 4365 8 4366 8 4367 8 4368 8 4369 1987 4370 8 4371 8 4372 8 4373 8 4374 8 4375 1988 4376 1989 4377 1990 4378 1984 4379 1991 4380 1992 4381 1993 4382 1994 4383 1995 4384 1996 4385 1997 4386 1998 4387 1999 4388 2000 4389 2001 4390 2002 4391 2003 4392 2004 4393 2005 4394 2006 4395 2007 4396 2008 4397 2009 4398 2010 4399 2011 4400 2012 4401 2013 4402 2014 4403 2015 4404 2016 4405 2017 4406 2018 4407 2019 4408 2020 4409 1984 4410 1984 4411 1984 4412 1984 4413 1984 4414 1984 4415 1984 4416 1984 4417 1984 4418 1984 4419 1984 4420 1984 4421 1984 4422 1984 4423 1984 4424 2021 4425 2022 4426 8 4427 8 4428 8 4429 8 4430 8 4431 8 4432 2023 4433 8 4434 8 4435 2024 4436 8 4437 2025 4438 2026 4439 2027 4440 2028 4441 2029 4442 8 4443 8 4444 8 4445 8 4446 8 4447 2030 4448 2031 4449 8 4450 8 4451 2032 4452 8 4453 2033 4454 8 4455 8 4456 2034 4457 2035 4458 8 4459 1984 4460 1984 4461 1984 4462 1984 4463 1984 4464 2036 4465 2037 4466 2038 4467 1984 4468 1984 4469 1984 4470 1984 4471 1984 4472 1984 4473 1984 4474 1984 4475 1984 4476 1984 4477 2039 4478 8 4479 8 4480 8 4481 8 4482 8 4483 8 4484 8 4485 8 4486 8 4487 8 4488 8 4489 8 4490 8 4491 8 4492 8 4493 8 4494 8 4495 8 4496 8 4497 8 4498 8 4499 8 4500 8 4501 8 4502 8 4503 8 4504 8 4505 8 4506 8 4507 8 4508 8 4509 8 4510 8 4511 8 4512 8 4513 8 4514 8 4515 8 4516 8 4517 8 4518 8 4519 8 4520 8 4521 8 4522 8 4523 8 4524 8 4525 8 4526 8 4527 8 4528 8 4529 8 4530 8 4531 8 4532 8 4533 8 4534 8 4535 8 4536 8 4537 8 4538 8 4539 8 4540 8 4541 8 4542 8 4543 8 4544 8 4545 8 4546 8 4547 8 4548 8 4549 8 4550 8 4551 8 4552 8 4553 8 4554 8 4555 8 4556 8 4557 8 4558 2040 4559 8 4560 8 4561 8 4562 2041 4563 2042 4564 8 4565 2043 4566 8 4567 8 4568 2044 4569 2045 4570 2046 4571 8 4572 8 4573 2047 4574 2048 4575 2049 4576 2050 4577 8 4578 2051 4579 8 4580 2052 4581 2053 4582 2054 4583 2055 4584 2056 4585 2057 4586 2058 4587 2059 4588 2060 4589 1984 4590 1984 4591 1984 4592 1984 4593 1984 4594 1984 4595 1984 4596 1984 4597 1984 4598 1984 4599 1984 4600 1984 4601 1984 4602 2061 4603 1984 4604 2062 4605 2063 4606 8 4607 2064 4608 2065 4609 8 4610 2066 4611 8 4612 2067 4613 8 4614 8 4615 8 4616 2068 4617 2069 4618 2070 4619 8 4620 2071 4621 8 4622 8 4623 2072 4624 2073 4625 2074 4626 8 4627 8 4628 8 4629 2075 4630 2076 4631 8 4632 8 4633 8 4634 2077 4635 2078 4636 2079 4637 2080 4638 8 4639 8 4640 2081 4641 8 4642 2082 4643 2083 4644 8 4645 8 4646 2084 4647 8 4648 8 4649 8 4650 2085 4651 2086 4652 8 4653 8 4654 2087 4655 2088 4656 2089 4657 2090 4658 2091 4659 2092 4660 2093 4661 2094 4662 2095 4663 2096 4664 2097 4665 8 4666 2098 4667 2099 4668 2100 4669 2101 4670 2102 4671 2103 4672 2104 4673 2105 4674 2106 4675 2107 4676 8 4677 2108 4678 8 4679 2109 4680 8 4681 2110 4682 2111 4683 8 4684 8 4685 8 4686 2112 4687 8 4688 8 4689 2113 4690 2114 4691 2115 4692 2116 4693 8 4694 8 4695 2117 4696 8 4697 2118 4698 2119 4699 8 4700 8 4701 2120 4702 8 4703 2121 4704 8 4705 2122 4706 2123 4707 2124 4708 1984 4709 1984 4710 2125 4711 2126 4712 2127 4713 2128 4714 2129 4715 2130 4716 2131 4717 2132 4718 2133 4719 2134 4720 8 4721 2135 4722 8 4723 2136 4724 8 4725 8 4726 2137 4727 2138 4728 2139 4729 2140 4730 8 4731 8 4732 2141 4733 2142 4734 2143 4735 2144 4736 2145 4737 8 4738 2146 4739 8 4740 2147 4741 8 4742 8 4743 8 4744 8 4745 2148 4746 8 4747 2149 4748 2150 4749 8 4750 2151 4751 2152 4752 2153 4753 2154 4754 2155 4755 8 4756 2156 4757 2157 4758 8 4759 2158 4760 8 4761 2159 4762 2160 4763 2161 4764 2162 4765 2163 4766 2164 4767 2165 4768 8 4769 8 4770 2166 4771 8 4772 8 4773 2167 4774 2168 4775 2169 4776 8 4777 8 4778 8 4779 8 4780 2170 4781 8 4782 8 4783 8 4784 8 4785 8 4786 2171 4787 2172 4788 2173 4789 2174 4790 2175 4791 2176 4792 2177 4793 2178 4794 2179 4795 2180 4796 2181 4797 2182 4798 2183 4799 2184 4800 2185 4801 2186 4802 2187 4803 2188 4804 2189 4805 2190 4806 2191 4807 2192 4808 2193 4809 2194 4810 2195 4811 2196 4812 2197 4813 2198 4814 2199 4815 2200 4816 2201 4817 2202 4818 2203 4819 2204 4820 2174 4821 2174 4822 2174 4823 2174 4824 2174 4825 2174 4826 2174 4827 2174 4828 2174 4829 2174 4830 2174 4831 2174 4832 2174 4833 2174 4834 2174 4835 2205 4836 2206 4837 8 4838 8 4839 8 4840 8 4841 8 4842 8 4843 2207 4844 8 4845 8 4846 2208 4847 8 4848 2209 4849 2210 4850 2211 4851 2212 4852 2213 4853 8 4854 8 4855 8 4856 8 4857 8 4858 2214 4859 2215 4860 8 4861 8 4862 2216 4863 8 4864 2217 4865 8 4866 8 4867 2218 4868 2219 4869 8 4870 2174 4871 2174 4872 2174 4873 2174 4874 2174 4875 2220 4876 2221 4877 2222 4878 2174 4879 2174 4880 2174 4881 2174 4882 2174 4883 2174 4884 2174 4885 2174 4886 2174 4887 2174 4888 2223 4889 8 4890 8 4891 8 4892 8 4893 8 4894 8 4895 8 4896 8 4897 8 4898 8 4899 8 4900 8 4901 8 4902 8 4903 8 4904 8 4905 8 4906 8 4907 8 4908 8 4909 8 4910 8 4911 8 4912 8 4913 8 4914 8 4915 8 4916 8 4917 8 4918 8 4919 8 4920 8 4921 8 4922 8 4923 8 4924 8 4925 8 4926 8 4927 8 4928 8 4929 8 4930 8 4931 8 4932 8 4933 8 4934 8 4935 8 4936 8 4937 8 4938 8 4939 8 4940 8 4941 8 4942 8 4943 8 4944 8 4945 8 4946 8 4947 8 4948 8 4949 8 4950 8 4951 8 4952 8 4953 8 4954 8 4955 8 4956 8 4957 8 4958 8 4959 8 4960 8 4961 8 4962 8 4963 8 4964 8 4965 8 4966 8 4967 8 4968 8 4969 2224 4970 8 4971 8 4972 8 4973 2225 4974 2226 4975 8 4976 2227 4977 8 4978 8 4979 2228 4980 2229 4981 2230 4982 8 4983 8 4984 2231 4985 2232 4986 2233 4987 2234 4988 8 4989 2235 4990 8 4991 2236 4992 2237 4993 2238 4994 2239 4995 2240 4996 2241 4997 2242 4998 2243 4999 2244 5000 2174 5001 2174 5002 2174 5003 2174 5004 2174 5005 2174 5006 2174 5007 2174 5008 2174 5009 2174 5010 2174 5011 2174 5012 2174 5013 2245 5014 2174 5015 2246 5016 2247 5017 8 5018 2248 5019 2249 5020 8 5021 2250 5022 8 5023 2251 5024 8 5025 8 5026 8 5027 2252 5028 2253 5029 2254 5030 8 5031 2255 5032 8 5033 8 5034 2256 5035 2257 5036 2258 5037 8 5038 8 5039 8 5040 2259 5041 2260 5042 8 5043 8 5044 8 5045 2261 5046 2262 5047 2263 5048 2264 5049 8 5050 8 5051 2265 5052 8 5053 2266 5054 2267 5055 8 5056 8 5057 2268 5058 8 5059 8 5060 8 5061 2269 5062 2270 5063 8 5064 8 5065 2271 5066 2272 5067 2273 5068 2274 5069 2275 5070 2276 5071 2277 5072 2278 5073 2279 5074 2280 5075 2281 5076 8 5077 2282 5078 2283 5079 2284 5080 2285 5081 2286 5082 2287 5083 2288 5084 2289 5085 2290 5086 2291 5087 8 5088 2292 5089 1984 5090 2293 5091 8 5092 2294 5093 8 5094 2295 5095 2296 5096 8 5097 8 5098 8 5099 2297 5100 8 5101 8 5102 2298 5103 2299 5104 2300 5105 2301 5106 8 5107 8 5108 2302 5109 8 5110 2303 5111 2304 5112 8 5113 8 5114 2305 5115 8 5116 2306 5117 8 5118 2307 5119 2308 5120 2309 5121 2174 5122 2174 5123 2310 5124 2311 5125 2174 5126 2174 5127 2174 5128 2174 5129 8 5130 2312 5131 8 5132 2313 5133 8 5134 8 5135 2174 5136 2174 5137 8 5138 2174 5139 2174 5140 2314 5141 8 5142 8 5143 8 5144 8 5145 8 5146 8 5147 2315 5148 2316 5149 2317 5150 2318 5151 8 5152 2319 5153 8 5154 2320 5155 2321 5156 2322 5157 2323 5158 2324 5159 8 5160 8 5161 8 5162 8 5163 8 5164 2325 5165 8 5166 8 5167 2326 5168 2327 5169 2328 5170 2329 5171 2330 5172 2331 5173 2332 5174 8 5175 2333 5176 8 5177 2334 5178 2335 5179 2336 5180 2337 5181 2338 5182 8 5183 2339 5184 2340 5185 2341 5186 2342 5187 2174 5188 2343 5189 557 5190 8 5191 2344 5192 2345 5193 2346 5194 8 5195 8 5196 2174 5197 2174 5198 2174 5199 8 5200 2347 5201 8 5202 8 5203 8 5204 2348 5205 2349 5206 2350 5207 8 5208 8 5209 2351 5210 2352 5211 8 5212 2353 5213 2354 5214 8 5215 2174 5216 8 5217 2174 5218 2174 5219 2174 5220 2355 5221 2356 5222 8 5223 2174 5224 2174 5225 2174 5226 2357 5227 8 5228 8 5229 2358 5230 2359 5231 8 5232 2360 5233 2361 5234 2362 5235 2363 5236 2364 5237 8 5238 8 5239 8 5240 8 5241 2365 5242 2366 5243 2367 5244 2368 5245 8 5246 8 5247 8 5248 8 5249 2369 5250 2370 5251 8 5252 2371 5253 8 5254 2372 5255 2373 5256 8 5257 2374 5258 8 5259 2174 5260 1984 5261 1984 5262 1984 5263 8 5264 2375 5265 8 5266 2376 5267 8 5268 8 5269 1984 5270 1984 5271 8 5272 1984 5273 1984 5274 2377 5275 8 5276 8 5277 8 5278 8 5279 8 5280 8 5281 2378 5282 2379 5283 2380 5284 2381 5285 8 5286 2382 5287 8 5288 2383 5289 2384 5290 2385 5291 2386 5292 2387 5293 8 5294 8 5295 8 5296 8 5297 8 5298 2388 5299 8 5300 8 5301 2389 5302 2390 5303 2391 5304 2392 5305 2393 5306 2394 5307 2395 5308 8 5309 2396 5310 8 5311 2397 5312 2398 5313 2399 5314 2400 5315 2401 5316 8 5317 2402 5318 2403 5319 2404 5320 2405 5321 1984 5322 2406 5323 2407 5324 2408 5325 8 5326 1984 5327 1984 5328 1984 5329 8 5330 2409 5331 8 5332 8 5333 8 5334 2410 5335 2411 5336 2412 5337 8 5338 8 5339 2413 5340 2414 5341 8 5342 2415 5343 2416 5344 8 5345 1984 5346 8 5347 1984 5348 1984 5349 1984 5350 2417 5351 2418 5352 8 5353 1984 5354 1984 5355 1984 5356 2419 5357 8 5358 8 5359 2420 5360 2421 5361 8 5362 2422 5363 8 5364 1984 5365 1832 5366 2423 5367 2424 5368 2425 5369 2426 5370 2427 5371 2428 5372 2429 5373 2430 5374 2431 5375 2432 5376 2433 5377 8 5378 2434 5379 2435 5380 2436 5381 2437 5382 2438 5383 2439 5384 2440 5385 2441 5386 2442 5387 2443 5388 8 5389 2444 5390 8 5391 2445 5392 8 5393 2446 5394 2447 5395 8 5396 8 5397 8 5398 2448 5399 8 5400 8 5401 2449 5402 2450 5403 2451 5404 2452 5405 8 5406 8 5407 2453 5408 8 5409 2454 5410 2455 5411 8 5412 8 5413 2456 5414 8 5415 2457 5416 8 5417 2458 5418 2459 5419 2460 5420 1832 5421 1832 5422 2461 5423 2462 5424 1832 5425 1832 5426 1832 5427 1832 5428 8 5429 2463 5430 8 5431 2464 5432 8 5433 8 5434 1832 5435 1832 5436 8 5437 1832 5438 1832 5439 2465 5440 8 5441 8 5442 8 5443 8 5444 8 5445 8 5446 2466 5447 2467 5448 2468 5449 2469 5450 8 5451 2470 5452 8 5453 2471 5454 2472 5455 2473 5456 2474 5457 2475 5458 8 5459 8 5460 8 5461 8 5462 8 5463 2476 5464 8 5465 8 5466 2477 5467 2478 5468 2479 5469 2480 5470 2481 5471 2482 5472 2483 5473 8 5474 2484 5475 8 5476 2485 5477 2486 5478 2487 5479 2488 5480 2489 5481 8 5482 2490 5483 2491 5484 2492 5485 2493 5486 1832 5487 2494 5488 2495 5489 2496 5490 8 5491 1832 5492 1832 5493 1832 5494 8 5495 2497 5496 8 5497 8 5498 8 5499 2498 5500 2499 5501 2500 5502 8 5503 8 5504 2501 5505 2502 5506 8 5507 2503 5508 2504 5509 8 5510 1832 5511 8 5512 2505 5513 2506 5514 1984 5515 8 5516 1832 5517 1832 5518 1832 5519 2507 5520 2508 5521 8 5522 1832 5523 1832 5524 1832 5525 2509 5526 8 5527 8 5528 2510 5529 2511 5530 8 5531 2512 5532 8 5533 1832 5534 2513 5535 2514 5536 2515 5537 2516 5538 1625 5539 2517 5540 2518 5541 2519 5542 8 5543 1625 5544 1625 5545 1625 5546 8 5547 2520 5548 8 5549 8 5550 8 5551 2521 5552 2522 5553 2523 5554 8 5555 8 5556 2524 5557 2525 5558 8 5559 2526 5560 2527 5561 8 5562 1625 5563 8 5564 1625 5565 1625 5566 1625 5567 2528 5568 2529 5569 8 5570 1625 5571 1625 5572 1625 5573 2530 5574 8 5575 8 5576 2531 5577 2532 5578 8 5579 2533 5580 8 5581 1625 5582 1441 5583 1441 5584 1441 5585 8 5586 2534 5587 8 5588 2535 5589 8 5590 8 5591 1441 5592 1441 5593 8 5594 1441 5595 1441 5596 2536 5597 8 5598 8 5599 8 5600 8 5601 8 5602 8 5603 2537 5604 2538 5605 2539 5606 2540 5607 8 5608 2541 5609 8 5610 2542 5611 2543 5612 2544 5613 2545 5614 2546 5615 8 5616 8 5617 8 5618 8 5619 8 5620 2547 5621 8 5622 8 5623 2548 5624 2549 5625 2550 5626 2551 5627 2552 5628 2553 5629 2554 5630 8 5631 2555 5632 8 5633 2556 5634 2557 5635 2558 5636 2559 5637 2560 5638 8 5639 2561 5640 2562 5641 2563 5642 2564 5643 1441 5644 2565 5645 2566 5646 2567 5647 8 5648 1441 5649 1441 5650 1441 5651 8 5652 2568 5653 8 5654 8 5655 8 5656 2569 5657 2570 5658 2571 5659 8 5660 8 5661 2572 5662 2573 5663 8 5664 2574 5665 2575 5666 8 5667 1441 5668 8 5669 1441 5670 1441 5671 1441 5672 2576 5673 2577 5674 8 5675 1441 5676 1441 5677 1441 5678 2578 5679 8 5680 8 5681 2579 5682 2580 5683 8 5684 2581 5685 8 5686 1441 5687 2582 5688 2583 5689 2584 5690 2585 5691 8 5692 2586 5693 8 5694 8 5695 8 5696 8 5697 8 5698 8 5699 8 5700 8 5701 8 5702 8 5703 2587 5704 8 5705 2588 5706 8 5707 2589 5708 8 5709 8 5710 8 5711 8 5712 2590 5713 2591 5714 2592 5715 2593 5716 2594 5717 2595 5718 8 5719 2596 5720 2597 5721 2598 5722 2599 5723 8 5724 8 5725 2600 5726 2601 5727 2602 5728 2603 5729 2604 5730 2605 5731 2606 5732 8 5733 2607 5734 2608 5735 2609 5736 2610 5737 2611 5738 2612 5739 2613 5740 2614 5741 2615 5742 2616 5743 2617 5744 2618 5745 2619 5746 2620 5747 2621 5748 2622 5749 2623 5750 2624 5751 2625 5752 2626 5753 2627 5754 2628 5755 2629 5756 2630 5757 84 5758 84 5759 84 5760 84 5761 84 5762 84 5763 84 5764 84 5765 84 5766 84 5767 84 5768 84 5769 84 5770 84 5771 84 5772 2631 5773 2632 5774 8 5775 8 5776 8 5777 8 5778 8 5779 8 5780 2633 5781 8 5782 8 5783 2634 5784 8 5785 2635 5786 2636 5787 2637 5788 2638 5789 2639 5790 8 5791 8 5792 8 5793 8 5794 8 5795 2640 5796 2641 5797 8 5798 8 5799 2642 5800 8 5801 2643 5802 8 5803 8 5804 2644 5805 2645 5806 8 5807 84 5808 84 5809 84 5810 84 5811 84 5812 2646 5813 2647 5814 2648 5815 84 5816 84 5817 84 5818 84 5819 84 5820 84 5821 84 5822 84 5823 84 5824 84 5825 2649 5826 8 5827 8 5828 8 5829 8 5830 8 5831 8 5832 8 5833 8 5834 8 5835 8 5836 2650 5837 2651 5838 8 5839 2652 5840 8 5841 8 5842 2653 5843 2654 5844 2655 5845 8 5846 8 5847 2656 5848 2657 5849 8 5850 8 5851 2658 5852 2659 5853 2660 5854 2661 5855 2662 5856 2663 5857 2664 5858 2665 5859 2666 5860 2667 5861 8 5862 2668 5863 8 5864 2669 5865 8 5866 8 5867 2670 5868 2671 5869 2672 5870 2673 5871 8 5872 8 5873 2674 5874 2675 5875 2676 5876 2677 5877 2678 5878 8 5879 2679 5880 8 5881 2680 5882 8 5883 8 5884 8 5885 8 5886 2681 5887 8 5888 2682 5889 2683 5890 8 5891 2684 5892 2685 5893 2686 5894 2687 5895 2688 5896 8 5897 2689 5898 2690 5899 8 5900 2691 5901 8 5902 2692 5903 2693 5904 2694 5905 2695 5906 2696 5907 2697 5908 2698 5909 8 5910 8 5911 2699 5912 8 5913 8 5914 2700 5915 2701 5916 2702 5917 8 5918 8 5919 8 5920 8 5921 2703 5922 8 5923 8 5924 8 5925 8 5926 8 5927 2704 5928 2705 5929 2706 5930 2707 5931 2708 5932 2709 5933 2710 5934 2711 5935 2712 5936 2713 5937 2714 5938 2715 5939 2716 5940 2717 5941 2718 5942 2719 5943 2720 5944 2721 5945 2722 5946 2723 5947 2724 5948 2725 5949 2726 5950 2727 5951 2728 5952 2729 5953 2730 5954 2731 5955 2732 5956 2733 5957 2734 5958 2735 5959 2736 5960 2737 5961 2707 5962 2707 5963 2707 5964 2707 5965 2707 5966 2707 5967 2707 5968 2707 5969 2707 5970 2707 5971 2707 5972 2707 5973 2707 5974 2707 5975 2707 5976 2738 5977 2739 5978 8 5979 8 5980 8 5981 8 5982 8 5983 8 5984 2740 5985 8 5986 8 5987 2741 5988 8 5989 2742 5990 2743 5991 2744 5992 2745 5993 2746 5994 8 5995 8 5996 8 5997 8 5998 8 5999 2747 6000 2748 6001 8 6002 8 6003 2749 6004 8 6005 2750 6006 8 6007 8 6008 2751 6009 2752 6010 8 6011 2707 6012 2707 6013 2707 6014 2707 6015 2707 6016 2753 6017 2754 6018 2755 6019 2707 6020 2707 6021 2707 6022 2707 6023 2707 6024 2707 6025 2707 6026 2707 6027 2707 6028 2707 6029 2756 6030 8 6031 8 6032 8 6033 8 6034 8 6035 8 6036 8 6037 8 6038 8 6039 8 6040 8 6041 8 6042 8 6043 8 6044 8 6045 8 6046 8 6047 8 6048 8 6049 8 6050 8 6051 8 6052 8 6053 8 6054 8 6055 8 6056 8 6057 8 6058 8 6059 8 6060 8 6061 8 6062 8 6063 8 6064 8 6065 8 6066 8 6067 8 6068 8 6069 8 6070 8 6071 8 6072 8 6073 8 6074 8 6075 8 6076 8 6077 8 6078 8 6079 8 6080 8 6081 8 6082 8 6083 8 6084 8 6085 8 6086 8 6087 8 6088 8 6089 8 6090 8 6091 8 6092 8 6093 8 6094 8 6095 8 6096 8 6097 8 6098 8 6099 8 6100 8 6101 8 6102 8 6103 8 6104 8 6105 8 6106 8 6107 8 6108 8 6109 8 6110 2757 6111 8 6112 8 6113 8 6114 2758 6115 2759 6116 8 6117 2760 6118 8 6119 8 6120 2761 6121 2762 6122 2763 6123 8 6124 8 6125 2764 6126 2765 6127 2766 6128 2767 6129 8 6130 2768 6131 8 6132 2769 6133 2770 6134 2771 6135 2772 6136 2773 6137 2774 6138 2775 6139 2776 6140 2777 6141 2707 6142 2707 6143 2707 6144 2707 6145 2707 6146 2707 6147 2707 6148 2707 6149 2707 6150 2707 6151 2707 6152 2707 6153 2707 6154 2778 6155 2707 6156 2779 6157 2780 6158 8 6159 2781 6160 2782 6161 8 6162 2783 6163 8 6164 2784 6165 8 6166 8 6167 8 6168 2785 6169 2786 6170 2787 6171 8 6172 2788 6173 8 6174 8 6175 2789 6176 2790 6177 2791 6178 8 6179 8 6180 8 6181 2792 6182 2793 6183 8 6184 8 6185 8 6186 2794 6187 2795 6188 2796 6189 2797 6190 8 6191 8 6192 2798 6193 8 6194 2799 6195 2800 6196 8 6197 8 6198 2801 6199 8 6200 8 6201 8 6202 2802 6203 2803 6204 8 6205 8 6206 2804 6207 2805 6208 2806 6209 2807 6210 2808 6211 2809 6212 2810 6213 2811 6214 2812 6215 2813 6216 2814 6217 8 6218 2815 6219 2816 6220 2817 6221 2818 6222 2819 6223 2820 6224 2821 6225 2822 6226 2823 6227 2824 6228 8 6229 2825 6230 2826 6231 2827 6232 8 6233 2828 6234 8 6235 2829 6236 8 6237 2830 6238 2831 6239 8 6240 8 6241 8 6242 2832 6243 8 6244 8 6245 2833 6246 2834 6247 2835 6248 2836 6249 8 6250 8 6251 2837 6252 8 6253 2838 6254 2839 6255 8 6256 8 6257 2840 6258 8 6259 2841 6260 8 6261 2842 6262 2843 6263 2844 6264 2707 6265 2707 6266 2845 6267 2846 6268 2847 6269 2848 6270 2849 6271 2850 6272 2851 6273 2852 6274 2853 6275 2854 6276 8 6277 2855 6278 8 6279 2856 6280 8 6281 8 6282 2857 6283 2858 6284 2859 6285 2860 6286 8 6287 8 6288 2861 6289 2862 6290 2863 6291 2864 6292 2865 6293 8 6294 2866 6295 8 6296 2867 6297 8 6298 8 6299 8 6300 8 6301 2868 6302 8 6303 2869 6304 2870 6305 8 6306 2871 6307 2872 6308 2873 6309 2874 6310 2875 6311 8 6312 2876 6313 2877 6314 8 6315 2878 6316 8 6317 2879 6318 2880 6319 2881 6320 2882 6321 2883 6322 2884 6323 2885 6324 8 6325 8 6326 2886 6327 8 6328 8 6329 2887 6330 2888 6331 2889 6332 8 6333 8 6334 8 6335 8 6336 2890 6337 8 6338 8 6339 8 6340 8 6341 8 6342 2891 6343 2892 6344 2893 6345 2894 6346 2895 6347 2896 6348 2897 6349 2898 6350 2899 6351 2900 6352 2901 6353 2902 6354 2903 6355 2904 6356 2905 6357 2906 6358 2907 6359 2908 6360 2909 6361 2910 6362 2911 6363 2912 6364 2913 6365 2914 6366 2915 6367 2916 6368 2917 6369 2918 6370 2919 6371 2920 6372 2921 6373 2922 6374 2923 6375 2924 6376 2894 6377 2894 6378 2894 6379 2894 6380 2894 6381 2894 6382 2894 6383 2894 6384 2894 6385 2894 6386 2894 6387 2894 6388 2894 6389 2894 6390 2894 6391 2925 6392 2926 6393 8 6394 8 6395 8 6396 8 6397 8 6398 8 6399 2927 6400 8 6401 8 6402 2928 6403 8 6404 2929 6405 2930 6406 2931 6407 2932 6408 2933 6409 8 6410 8 6411 8 6412 8 6413 8 6414 2934 6415 2935 6416 8 6417 8 6418 2936 6419 8 6420 2937 6421 8 6422 8 6423 2938 6424 2939 6425 8 6426 2894 6427 2894 6428 2894 6429 2894 6430 2894 6431 2940 6432 2941 6433 2942 6434 2894 6435 2894 6436 2894 6437 2894 6438 2894 6439 2894 6440 2894 6441 2894 6442 2894 6443 2894 6444 2943 6445 8 6446 8 6447 8 6448 8 6449 8 6450 8 6451 8 6452 8 6453 8 6454 8 6455 8 6456 8 6457 8 6458 8 6459 8 6460 8 6461 8 6462 8 6463 8 6464 8 6465 8 6466 8 6467 8 6468 8 6469 8 6470 8 6471 8 6472 8 6473 8 6474 8 6475 8 6476 8 6477 8 6478 8 6479 8 6480 8 6481 8 6482 8 6483 8 6484 8 6485 8 6486 8 6487 8 6488 8 6489 8 6490 8 6491 8 6492 8 6493 8 6494 8 6495 8 6496 8 6497 8 6498 8 6499 8 6500 8 6501 8 6502 8 6503 8 6504 8 6505 8 6506 8 6507 8 6508 8 6509 8 6510 8 6511 8 6512 8 6513 8 6514 8 6515 8 6516 8 6517 8 6518 8 6519 8 6520 8 6521 8 6522 8 6523 8 6524 8 6525 2944 6526 8 6527 8 6528 8 6529 2945 6530 2946 6531 8 6532 2947 6533 8 6534 8 6535 2948 6536 2949 6537 2950 6538 8 6539 8 6540 2951 6541 2952 6542 2953 6543 2954 6544 8 6545 2955 6546 8 6547 2956 6548 2957 6549 2958 6550 2959 6551 2960 6552 2961 6553 2962 6554 2963 6555 2964 6556 2894 6557 2894 6558 2894 6559 2894 6560 2894 6561 2894 6562 2894 6563 2894 6564 2894 6565 2894 6566 2894 6567 2894 6568 2894 6569 2965 6570 2894 6571 2966 6572 2967 6573 8 6574 2968 6575 2969 6576 8 6577 2970 6578 8 6579 2971 6580 8 6581 8 6582 8 6583 2972 6584 2973 6585 2974 6586 8 6587 2975 6588 8 6589 8 6590 2976 6591 2977 6592 2978 6593 8 6594 8 6595 8 6596 2979 6597 2980 6598 8 6599 8 6600 8 6601 2981 6602 2982 6603 2983 6604 2984 6605 8 6606 8 6607 2985 6608 8 6609 2986 6610 2987 6611 8 6612 8 6613 2988 6614 8 6615 8 6616 8 6617 2989 6618 2990 6619 8 6620 8 6621 2991 6622 2992 6623 2993 6624 2994 6625 2995 6626 2996 6627 2997 6628 2998 6629 2999 6630 3000 6631 3001 6632 8 6633 3002 6634 3003 6635 3004 6636 3005 6637 3006 6638 3007 6639 3008 6640 3009 6641 3010 6642 3011 6643 8 6644 3012 6645 2707 6646 3013 6647 8 6648 3014 6649 8 6650 3015 6651 3016 6652 8 6653 8 6654 8 6655 3017 6656 8 6657 8 6658 3018 6659 3019 6660 3020 6661 3021 6662 8 6663 8 6664 3022 6665 8 6666 3023 6667 3024 6668 8 6669 8 6670 3025 6671 8 6672 3026 6673 8 6674 3027 6675 3028 6676 3029 6677 2894 6678 2894 6679 3030 6680 3031 6681 2894 6682 2894 6683 2894 6684 2894 6685 8 6686 3032 6687 8 6688 3033 6689 8 6690 8 6691 2894 6692 2894 6693 8 6694 2894 6695 2894 6696 3034 6697 8 6698 8 6699 8 6700 8 6701 8 6702 8 6703 3035 6704 3036 6705 3037 6706 3038 6707 8 6708 3039 6709 8 6710 3040 6711 3041 6712 3042 6713 3043 6714 3044 6715 8 6716 8 6717 8 6718 8 6719 8 6720 3045 6721 8 6722 8 6723 3046 6724 3047 6725 3048 6726 3049 6727 3050 6728 3051 6729 3052 6730 8 6731 3053 6732 8 6733 3054 6734 3055 6735 3056 6736 3057 6737 3058 6738 8 6739 3059 6740 3060 6741 3061 6742 3062 6743 2894 6744 3063 6745 3064 6746 3065 6747 8 6748 2894 6749 2894 6750 2894 6751 8 6752 3066 6753 8 6754 8 6755 8 6756 3067 6757 3068 6758 3069 6759 8 6760 8 6761 3070 6762 3071 6763 8 6764 3072 6765 3073 6766 8 6767 2894 6768 8 6769 2894 6770 2894 6771 2894 6772 3074 6773 3075 6774 8 6775 2894 6776 2894 6777 2894 6778 3076 6779 8 6780 8 6781 3077 6782 3078 6783 8 6784 3079 6785 8 6786 2894 6787 2707 6788 2707 6789 2707 6790 8 6791 3080 6792 8 6793 3081 6794 8 6795 8 6796 2707 6797 2707 6798 8 6799 2707 6800 2707 6801 3082 6802 8 6803 8 6804 8 6805 8 6806 8 6807 8 6808 3083 6809 3084 6810 3085 6811 3086 6812 8 6813 3087 6814 8 6815 3088 6816 3089 6817 3090 6818 3091 6819 3092 6820 8 6821 8 6822 8 6823 8 6824 8 6825 3093 6826 8 6827 8 6828 3094 6829 3095 6830 3096 6831 3097 6832 3098 6833 3099 6834 3100 6835 8 6836 3101 6837 8 6838 3102 6839 3103 6840 3104 6841 3105 6842 3106 6843 8 6844 3107 6845 3108 6846 3109 6847 3110 6848 2707 6849 3111 6850 3112 6851 3113 6852 8 6853 2707 6854 2707 6855 2707 6856 8 6857 3114 6858 8 6859 8 6860 8 6861 3115 6862 3116 6863 3117 6864 8 6865 8 6866 3118 6867 3119 6868 8 6869 3120 6870 3121 6871 8 6872 2707 6873 8 6874 2707 6875 2707 6876 2707 6877 3122 6878 3123 6879 8 6880 2707 6881 2707 6882 2707 6883 3124 6884 8 6885 8 6886 3125 6887 3126 6888 8 6889 3127 6890 8 6891 2707 6892 3128 6893 8 6894 8 6895 8 6896 8 6897 8 6898 8 6899 8 6900 8 6901 8 6902 3129 6903 8 6904 8 6905 8 6906 3130 6907 8 6908 8 6909 8 6910 3131 6911 8 6912 8 6913 8 6914 8 6915 8 6916 8 6917 8 6918 8 6919 8 6920 8 6921 8 6922 8 6923 8 6924 8 6925 8 6926 8 6927 8 6928 8 6929 8 6930 8 6931 8 6932 8 6933 8 6934 8 6935 8 6936 8 6937 8 6938 8 6939 8 6940 8 6941 8 6942 8 6943 8 6944 8 6945 8 6946 8 6947 8 6948 8 6949 8 6950 8 6951 8 6952 8 6953 8 6954 8 6955 8 6956 8 6957 8 6958 8 6959 8 6960 8 6961 8 6962 8 6963 8 6964 8 6965 8 6966 8 6967 8 6968 8 6969 8 6970 8 6971 8 6972 8 6973 8 6974 8 6975 8 6976 8 6977 8 6978 8 6979 8 6980 8 6981 8 6982 8 6983 8 6984 8 6985 8 6986 8 6987 8 6988 8 6989 8 6990 3132 6991 8 6992 8 6993 3133 6994 3134 6995 3135 6996 3136 6997 3137 6998 3138 6999 3139 7000 8 7001 3140 7002 3141 7003 3142 7004 8 7005 8 7006 3143 7007 8 7008 8 7009 8 7010 8 7011 8 7012 8 7013 8 7014 3144 7015 8 7016 3145 7017 8 7018 8 7019 3146 7020 3147 7021 3148 7022 3149 7023 8 7024 8 7025 8 7026 8 7027 8 7028 8 7029 8 7030 8 7031 8 7032 8 7033 8 7034 3150 7035 8 7036 3151 7037 8 7038 3152 7039 8 7040 8 7041 8 7042 8 7043 8 7044 8 7045 8 7046 8 7047 8 7048 8 7049 3153 7050 8 7051 8 7052 8 7053 8 7054 8 7055 8 7056 8 7057 8 7058 8 7059 8 7060 8 7061 3154 7062 3155 7063 8 7064 8 7065 8 7066 8 7067 8 7068 8 7069 8 7070 8 7071 8 7072 8 7073 8 7074 8 7075 8 7076 8 7077 8 7078 8 7079 8 7080 8 7081 8 7082 8 7083 8 7084 8 7085 8 7086 8 7087 8 7088 8 7089 8 7090 8 7091 8 7092 8 7093 8 7094 8 7095 8 7096 8 7097 8 7098 8 7099 8 7100 8 7101 8 7102 8 7103 8 7104 8 7105 8 7106 8 7107 8 7108 8 7109 8 7110 8 7111 8 7112 8 7113 8 7114 8 7115 8 7116 8 7117 8 7118 8 7119 8 7120 8 7121 8 7122 8 7123 8 7124 8 7125 8 7126 8 7127 8 7128 8 7129 8 7130 8 7131 8 7132 8 7133 8 7134 3156 7135 8 7136 8 7137 8 7138 8 7139 8 7140 8 7141 8 7142 8 7143 8 7144 8 7145 8 7146 3157 7147 3158 7148 8 7149 8 7150 3159 7151 3160 7152 3161 7153 8 7154 3162 7155 3163 7156 3164 7157 3165 7158 8 7159 8 7160 8 7161 8 7162 8 7163 8 7164 8 7165 8 7166 8 7167 8 7168 8 7169 3166 7170 8 7171 3167 7172 8 7173 8 7174 8 7175 8 7176 3156 7177 3168 7178 3169 7179 8 7180 8 7181 3170 7182 3171 7183 8 7184 3172 7185 8 7186 8 7187 8 7188 8 7189 8 7190 8 7191 8 7192 8 7193 8 7194 8 7195 3173 7196 3174 7197 8 7198 3175 7199 8 7200 8 7201 8 7202 8 7203 8 7204 8 7205 8 7206 8 7207 8 7208 8 7209 8 7210 8 7211 8 7212 8 7213 8 7214 8 7215 8 7216 8 7217 8 7218 8 7219 3176 7220 8 7221 3177 7222 8 7223 8 7224 3178 7225 8 7226 8 7227 8 7228 8 7229 8 7230 8 7231 8 7232 8 7233 8 7234 8 7235 8 7236 8 7237 8 7238 8 7239 8 7240 8 7241 8 7242 8 7243 8 7244 8 7245 8 7246 8 7247 8 7248 8 7249 8 7250 8 7251 8 7252 8 7253 8 7254 8 7255 8 7256 8 7257 8 7258 8 7259 8 7260 8 7261 8 7262 8 7263 8 7264 8 7265 8 7266 8 7267 8 7268 8 7269 8 7270 8 7271 8 7272 8 7273 8 7274 8 7275 8 7276 8 7277 8 7278 8 7279 8 7280 8 7281 8 7282 8 7283 8 7284 8 7285 8 7286 8 7287 8 7288 8 7289 8 7290 8 7291 8 7292 8 7293 8 7294 8 7295 8 7296 8 7297 8 7298 8 7299 8 7300 8 7301 8 7302 8 7303 8 7304 8 7305 8 7306 8 7307 3179 7308 3180 7309 8 7310 8 7311 8 7312 8 7313 8 7314 3181 7315 3182 7316 3183 7317 3184 7318 8 7319 8 7320 3185 7321 8 7322 8 7323 8 7324 8 7325 8 7326 8 7327 8 7328 8 7329 8 7330 8 7331 8 7332 8 7333 8 7334 8 7335 8 7336 8 7337 8 7338 8 7339 8 7340 8 7341 8 7342 8 7343 8 7344 8 7345 8 7346 8 7347 8 7348 8 7349 8 7350 8 7351 8 7352 8 7353 8 7354 8 7355 8 7356 8 7357 8 7358 8 7359 8 7360 8 7361 8 7362 8 7363 8 7364 8 7365 8 7366 8 7367 8 7368 8 7369 8 7370 8 7371 8 7372 8 7373 8 7374 8 7375 8 7376 8 7377 8 7378 8 7379 8 7380 8 7381 8 7382 8 7383 8 7384 8 7385 8 7386 8 7387 8 7388 8 7389 8 7390 8 7391 8 7392 8 7393 8 7394 8 7395 8 7396 8 7397 8 7398 8 7399 8 7400 8 7401 8 7402 8 7403 3186 7404 8 7405 8 7406 8 7407 8 7408 8 7409 8 7410 8 7411 8 7412 8 7413 8 7414 8 7415 8 7416 8 7417 8 7418 8 7419 3187 7420 8 7421 8 7422 8 7423 8 7424 8 7425 8 7426 8 7427 8 7428 8 7429 8 7430 8 7431 8 7432 8 7433 8 7434 8 7435 8 7436 8 7437 8 7438 8 7439 8 7440 8 7441 8 7442 8 7443 8 7444 8 7445 8 7446 8 7447 8 7448 8 7449 8 7450 8 7451 8 7452 8 7453 8 7454 8 7455 8 7456 8 7457 8 7458 8 7459 8 7460 8 7461 8 7462 8 7463 8 7464 8 7465 8 7466 8 7467 8 7468 8 7469 8 7470 8 7471 8 7472 8 7473 8 7474 8 7475 8 7476 8 7477 8 7478 8 7479 8 7480 8 7481 8 7482 8 7483 8 7484 8 7485 8 7486 8 7487 8 7488 8 7489 8 7490 8 7491 8 7492 8 7493 8 7494 8 7495 8 7496 8 7497 8 7498 8 7499 8 7500 8 7501 8 7502 3188 7503 8 7504 8 7505 8 7506 8 7507 8 7508 8 7509 3189 7510 8 7511 3190 7512 8 7513 8 7514 3191 7515 8 7516 8 7517 8 7518 8 7519 8 7520 8 7521 8 7522 8 7523 8 7524 8 7525 8 7526 8 7527 8 7528 8 7529 8 7530 8 7531 8 7532 8 7533 8 7534 8 7535 8 7536 8 7537 8 7538 8 7539 8 7540 8 7541 8 7542 8 7543 8 7544 8 7545 8 7546 8 7547 8 7548 8 7549 8 7550 8 7551 8 7552 8 7553 8 7554 8 7555 8 7556 8 7557 8 7558 8 7559 8 7560 8 7561 8 7562 8 7563 8 7564 8 7565 8 7566 8 7567 8 7568 8 7569 8 7570 8 7571 8 7572 8 7573 8 7574 8 7575 8 7576 8 7577 8 7578 8 7579 8 7580 8 7581 8 7582 8 7583 8 7584 8 7585 8 7586 8 7587 8 7588 8 7589 8 7590 8 7591 8 7592 8 7593 8 7594 3192 7595 8 7596 8 7597 3193 7598 3194 7599 8 7600 84 7601 3195 7602 3196 7603 8 7604 8 7605 8 7606 8 7607 8 7608 8 7609 8 7610 3197 7611 8 7612 3198 7613 8 7614 8 7615 8 7616 8 7617 8 7618 8 7619 8 7620 8 7621 8 7622 3199 7623 3200 7624 3201 7625 8 7626 8 7627 8 7628 3202 7629 3203 7630 8 7631 3204 7632 8 7633 3205 7634 3206 7635 3207 7636 3208 7637 3209 7638 8 7639 3210 7640 3211 7641 1169 7642 3212 7643 3213 7644 3214 7645 8 7646 1111 7647 1111 7648 1111 7649 8 7650 3215 7651 8 7652 8 7653 8 7654 3216 7655 3217 7656 3218 7657 8 7658 8 7659 3219 7660 3220 7661 8 7662 3221 7663 3222 7664 8 7665 1111 7666 8 7667 1111 7668 1111 7669 1111 7670 3223 7671 3224 7672 8 7673 1111 7674 1111 7675 1111 7676 3225 7677 8 7678 8 7679 3226 7680 3227 7681 8 7682 3228 7683 8 7684 1111 7685 3229 7686 8 7687 3230 7688 8 7689 3231 7690 3232 7691 8 7692 8 7693 8 7694 3233 7695 8 7696 8 7697 3234 7698 3235 7699 823 7700 3236 7701 3237 7702 3238 7703 3239 7704 8 7705 8 7706 3240 7707 8 7708 3241 7709 3242 7710 8 7711 8 7712 3243 7713 8 7714 8 7715 3244 7716 8 7717 3245 7718 3246 7719 3247 7720 927 7721 927 7722 3248 7723 3249 7724 927 7725 927 7726 927 7727 927 7728 8 7729 3250 7730 8 7731 3251 7732 8 7733 8 7734 927 7735 927 7736 8 7737 927 7738 927 7739 3252 7740 8 7741 8 7742 8 7743 8 7744 8 7745 8 7746 3253 7747 3254 7748 3255 7749 3256 7750 8 7751 3257 7752 8 7753 3258 7754 3259 7755 3260 7756 3261 7757 3262 7758 8 7759 8 7760 8 7761 8 7762 8 7763 3263 7764 8 7765 8 7766 3264 7767 3265 7768 3266 7769 3267 7770 3268 7771 3269 7772 3270 7773 8 7774 3271 7775 8 7776 3272 7777 3273 7778 3274 7779 3275 7780 3276 7781 8 7782 3277 7783 3278 7784 978 7785 3279 7786 3280 7787 3281 7788 8 7789 927 7790 927 7791 927 7792 8 7793 3282 7794 8 7795 8 7796 8 7797 3283 7798 3284 7799 3285 7800 8 7801 8 7802 3286 7803 3287 7804 8 7805 3288 7806 3289 7807 8 7808 1111 7809 927 7810 8 7811 927 7812 927 7813 927 7814 3290 7815 3291 7816 8 7817 927 7818 927 7819 927 7820 3292 7821 8 7822 8 7823 3293 7824 3294 7825 8 7826 3295 7827 8 7828 927 7829 3296 7830 8 7831 8 7832 8 7833 8 7834 8 7835 8 7836 8 7837 8 7838 8 7839 8 7840 8 7841 8 7842 8 7843 8 7844 8 7845 8 7846 8 7847 8 7848 8 7849 8 7850 8 7851 8 7852 8 7853 8 7854 8 7855 8 7856 8 7857 8 7858 8 7859 8 7860 8 7861 8 7862 8 7863 8 7864 8 7865 8 7866 8 7867 8 7868 8 7869 8 7870 8 7871 8 7872 8 7873 8 7874 8 7875 8 7876 8 7877 8 7878 8 7879 8 7880 8 7881 8 7882 8 7883 8 7884 8 7885 8 7886 8 7887 8 7888 8 7889 8 7890 8 7891 8 7892 8 7893 8 7894 8 7895 8 7896 8 7897 8 7898 8 7899 8 7900 8 7901 8 7902 8 7903 8 7904 8 7905 8 7906 8 7907 8 7908 8 7909 8 7910 3297 7911 8 7912 8 7913 8 7914 3298 7915 3299 7916 8 7917 3300 7918 8 7919 8 7920 3301 7921 3302 7922 3303 7923 8 7924 8 7925 3304 7926 3305 7927 3306 7928 3307 7929 8 7930 3308 7931 8 7932 3309 7933 3310 7934 3311 7935 3312 7936 3313 7937 3314 7938 3315 7939 3316 7940 3317 7941 823 7942 823 7943 823 7944 823 7945 823 7946 823 7947 823 7948 823 7949 823 7950 823 7951 823 7952 823 7953 823 7954 3318 7955 823 7956 3319 7957 3320 7958 8 7959 3321 7960 3322 7961 8 7962 3323 7963 8 7964 3324 7965 8 7966 8 7967 8 7968 3325 7969 3326 7970 3327 7971 8 7972 3328 7973 8 7974 8 7975 3329 7976 3330 7977 3331 7978 8 7979 8 7980 8 7981 3332 7982 3333 7983 8 7984 8 7985 8 7986 3334 7987 3335 7988 3336 7989 3337 7990 8 7991 8 7992 3338 7993 8 7994 3339 7995 3340 7996 8 7997 8 7998 3341 7999 8 8000 8 8001 8 8002 3342 8003 3343 8004 8 8005 3344 8006 8 8007 8 8008 3345 8009 3346 8010 3347 8011 3348 8012 3349 8013 3350 8014 3351 8015 3352 8016 3353 8017 3354 8018 3355 8019 8 8020 3356 8021 3357 8022 3358 8023 3359 8024 3360 8025 3361 8026 3362 8027 3363 8028 3364 8029 3365 8030 3366 8031 3367 8032 3368 8033 3369 8034 3370 8035 3371 8036 3372 8037 3373 8038 3374 8039 8 8040 3375 8041 8 8042 3376 8043 8 8044 8 8045 3377 8046 3378 8047 3379 8048 3380 8049 8 8050 8 8051 3381 8052 3382 8053 3383 8054 3384 8055 3385 8056 8 8057 3386 8058 8 8059 3387 8060 8 8061 8 8062 8 8063 8 8064 3388 8065 8 8066 3389 8067 3390 8068 8 8069 3391 8070 3392 8071 3393 8072 3394 8073 3395 8074 8 8075 3396 8076 3397 8077 8 8078 3398 8079 8 8080 3399 8081 3400 8082 3401 8083 3402 8084 3403 8085 3404 8086 3405 8087 8 8088 8 8089 3406 8090 8 8091 8 8092 3407 8093 3408 8094 3409 8095 8 8096 8 8097 8 8098 8 8099 3410 8100 8 8101 8 8102 8 8103 8 8104 8 8105 3411 8106 3412 8107 3413 8108 3414 8109 3415 8110 3416 8111 3417 8112 3418 8113 3419 8114 3420 8115 3421 8116 3422 8117 3423 8118 3424 8119 3425 8120 3426 8121 3427 8122 3428 8123 3429 8124 3430 8125 3431 8126 3432 8127 3433 8128 3434 8129 3435 8130 3436 8131 3437 8132 3438 8133 3439 8134 3440 8135 3441 8136 3442 8137 3443 8138 3444 8139 3414 8140 3414 8141 3414 8142 3414 8143 3414 8144 3414 8145 3414 8146 3414 8147 3414 8148 3414 8149 3414 8150 3414 8151 3414 8152 3414 8153 3414 8154 3445 8155 3446 8156 8 8157 8 8158 8 8159 8 8160 8 8161 8 8162 3447 8163 8 8164 8 8165 3448 8166 8 8167 3449 8168 3450 8169 3451 8170 3452 8171 3453 8172 8 8173 8 8174 8 8175 8 8176 8 8177 3454 8178 3455 8179 8 8180 8 8181 3456 8182 8 8183 3457 8184 8 8185 8 8186 3458 8187 3459 8188 8 8189 3414 8190 3414 8191 3414 8192 3414 8193 3414 8194 3460 8195 3461 8196 3462 8197 3414 8198 3414 8199 3414 8200 3414 8201 3414 8202 3414 8203 3414 8204 3414 8205 3414 8206 3414 8207 3463 8208 8 8209 8 8210 8 8211 8 8212 8 8213 8 8214 8 8215 8 8216 8 8217 8 8218 8 8219 8 8220 8 8221 8 8222 8 8223 8 8224 8 8225 8 8226 8 8227 8 8228 8 8229 8 8230 8 8231 8 8232 8 8233 8 8234 8 8235 8 8236 8 8237 8 8238 8 8239 8 8240 8 8241 8 8242 8 8243 8 8244 8 8245 8 8246 8 8247 8 8248 8 8249 8 8250 8 8251 8 8252 8 8253 8 8254 8 8255 8 8256 8 8257 8 8258 8 8259 8 8260 8 8261 8 8262 8 8263 8 8264 8 8265 8 8266 8 8267 8 8268 8 8269 8 8270 8 8271 8 8272 8 8273 8 8274 8 8275 8 8276 8 8277 8 8278 8 8279 8 8280 8 8281 8 8282 8 8283 8 8284 8 8285 8 8286 8 8287 8 8288 3464 8289 8 8290 8 8291 8 8292 3465 8293 3466 8294 8 8295 3467 8296 8 8297 8 8298 3468 8299 3469 8300 3470 8301 8 8302 8 8303 3471 8304 3472 8305 3473 8306 3474 8307 8 8308 3475 8309 8 8310 3476 8311 3477 8312 3478 8313 3479 8314 3480 8315 3481 8316 3482 8317 3483 8318 3484 8319 3414 8320 3414 8321 3414 8322 3414 8323 3414 8324 3414 8325 3414 8326 3414 8327 3414 8328 3414 8329 3414 8330 3414 8331 3414 8332 3485 8333 3414 8334 3486 8335 3487 8336 8 8337 3488 8338 3489 8339 8 8340 3490 8341 8 8342 3491 8343 8 8344 8 8345 8 8346 3492 8347 3493 8348 3494 8349 8 8350 3495 8351 8 8352 8 8353 3496 8354 3497 8355 3498 8356 8 8357 8 8358 8 8359 3499 8360 3500 8361 8 8362 8 8363 8 8364 3501 8365 3502 8366 3503 8367 3504 8368 8 8369 8 8370 3505 8371 8 8372 3506 8373 3507 8374 8 8375 8 8376 3508 8377 8 8378 8 8379 8 8380 3509 8381 3510 8382 8 8383 8 8384 3511 8385 3512 8386 3513 8387 3514 8388 3515 8389 3516 8390 3517 8391 3518 8392 3519 8393 3520 8394 3521 8395 8 8396 3522 8397 3523 8398 3524 8399 3525 8400 3526 8401 3527 8402 3528 8403 3529 8404 3530 8405 3531 8406 8 8407 8 8408 3532 8409 8 8410 3533 8411 8 8412 3534 8413 3535 8414 8 8415 8 8416 8 8417 3536 8418 8 8419 8 8420 3537 8421 3538 8422 3539 8423 3540 8424 8 8425 8 8426 3541 8427 8 8428 3542 8429 3543 8430 8 8431 8 8432 3544 8433 8 8434 3545 8435 8 8436 3546 8437 3547 8438 3548 8439 3414 8440 3414 8441 3549 8442 3550 8443 3414 8444 3414 8445 3414 8446 3414 8447 8 8448 3551 8449 8 8450 3552 8451 8 8452 8 8453 3414 8454 3414 8455 8 8456 3414 8457 3414 8458 3553 8459 8 8460 8 8461 8 8462 8 8463 8 8464 8 8465 3554 8466 3555 8467 3556 8468 3557 8469 8 8470 3558 8471 8 8472 3559 8473 3560 8474 3561 8475 3562 8476 3563 8477 8 8478 8 8479 8 8480 8 8481 8 8482 3564 8483 8 8484 8 8485 3565 8486 3566 8487 3567 8488 3568 8489 3569 8490 3570 8491 3571 8492 8 8493 3572 8494 8 8495 3573 8496 3574 8497 3575 8498 3576 8499 3577 8500 8 8501 3578 8502 3579 8503 3580 8504 3581 8505 3414 8506 3582 8507 3583 8508 3584 8509 8 8510 3414 8511 3414 8512 3414 8513 8 8514 3585 8515 8 8516 8 8517 8 8518 3586 8519 3587 8520 3588 8521 8 8522 8 8523 3589 8524 3590 8525 8 8526 3591 8527 3592 8528 8 8529 3414 8530 8 8531 3414 8532 3414 8533 3414 8534 3593 8535 3594 8536 8 8537 3414 8538 3414 8539 3414 8540 3595 8541 8 8542 8 8543 3596 8544 3597 8545 8 8546 3598 8547 8 8548 3414 8549 3599 8550 3600 8551 3601 8552 3602 8553 3603 8554 3604 8555 3605 8556 3606 8557 3607 8558 3608 8559 3609 8560 8 8561 3610 8562 8 8563 3611 8564 8 8565 8 8566 3612 8567 3613 8568 3614 8569 3615 8570 8 8571 8 8572 3616 8573 3617 8574 3618 8575 3619 8576 3620 8577 8 8578 3621 8579 8 8580 3622 8581 8 8582 8 8583 8 8584 8 8585 3623 8586 8 8587 3624 8588 3625 8589 8 8590 3626 8591 3627 8592 3628 8593 3629 8594 3630 8595 8 8596 3631 8597 3632 8598 8 8599 3633 8600 8 8601 3634 8602 3635 8603 3636 8604 3637 8605 3638 8606 3639 8607 3640 8608 3641 8609 8 8610 8 8611 3642 8612 8 8613 8 8614 3643 8615 3644 8616 3645 8617 8 8618 8 8619 8 8620 8 8621 3646 8622 8 8623 8 8624 8 8625 8 8626 8 8627 3647 8628 3648 8629 3649 8630 3650 8631 3643 8632 3651 8633 3652 8634 3653 8635 3654 8636 3655 8637 3656 8638 3657 8639 3658 8640 3659 8641 3660 8642 3661 8643 3662 8644 3663 8645 3664 8646 3665 8647 3666 8648 3667 8649 3668 8650 3669 8651 3670 8652 3671 8653 3672 8654 3673 8655 3674 8656 3675 8657 3676 8658 3677 8659 3678 8660 3679 8661 3680 8662 3643 8663 3643 8664 3643 8665 3643 8666 3643 8667 3643 8668 3643 8669 3643 8670 3643 8671 3643 8672 3643 8673 3643 8674 3643 8675 3643 8676 3643 8677 3681 8678 3682 8679 8 8680 8 8681 8 8682 8 8683 8 8684 8 8685 3683 8686 8 8687 8 8688 3684 8689 8 8690 3685 8691 3686 8692 3687 8693 3688 8694 3689 8695 8 8696 8 8697 8 8698 8 8699 8 8700 3690 8701 3691 8702 8 8703 8 8704 3692 8705 8 8706 3693 8707 8 8708 8 8709 3694 8710 3695 8711 8 8712 3643 8713 3643 8714 3643 8715 3643 8716 3643 8717 3696 8718 3697 8719 3698 8720 3643 8721 3643 8722 3643 8723 3643 8724 3643 8725 3643 8726 3643 8727 3643 8728 3643 8729 3643 8730 3699 8731 3700 8732 3701 8733 3702 8734 3643 8735 3703 8736 8 8737 8 8738 8 8739 8 8740 8 8741 8 8742 8 8743 8 8744 8 8745 8 8746 8 8747 8 8748 8 8749 8 8750 8 8751 8 8752 8 8753 8 8754 8 8755 8 8756 8 8757 8 8758 8 8759 8 8760 8 8761 8 8762 8 8763 8 8764 8 8765 8 8766 8 8767 8 8768 8 8769 8 8770 8 8771 8 8772 8 8773 8 8774 8 8775 8 8776 8 8777 8 8778 8 8779 8 8780 8 8781 8 8782 8 8783 8 8784 8 8785 8 8786 8 8787 8 8788 8 8789 8 8790 8 8791 8 8792 8 8793 8 8794 8 8795 8 8796 8 8797 8 8798 8 8799 8 8800 8 8801 8 8802 8 8803 8 8804 8 8805 8 8806 8 8807 8 8808 8 8809 8 8810 8 8811 8 8812 8 8813 8 8814 8 8815 8 8816 3704 8817 8 8818 8 8819 8 8820 3705 8821 3706 8822 8 8823 3707 8824 8 8825 8 8826 3708 8827 3709 8828 3710 8829 8 8830 8 8831 3711 8832 3712 8833 3713 8834 3714 8835 8 8836 3715 8837 8 8838 3716 8839 3717 8840 3718 8841 3719 8842 3720 8843 3721 8844 3722 8845 3723 8846 3724 8847 3643 8848 3643 8849 3643 8850 3643 8851 3643 8852 3643 8853 3643 8854 3643 8855 3643 8856 3643 8857 3643 8858 3643 8859 3643 8860 3725 8861 3643 8862 3726 8863 3727 8864 8 8865 3728 8866 3729 8867 8 8868 3730 8869 8 8870 3731 8871 8 8872 8 8873 8 8874 3732 8875 3733 8876 3734 8877 8 8878 3735 8879 8 8880 8 8881 3736 8882 3737 8883 3738 8884 8 8885 8 8886 8 8887 3739 8888 3740 8889 8 8890 8 8891 8 8892 3741 8893 3742 8894 3743 8895 3744 8896 8 8897 8 8898 3745 8899 8 8900 3746 8901 3747 8902 8 8903 8 8904 3748 8905 8 8906 8 8907 8 8908 3749 8909 3750 8910 8 8911 8 8912 3751 8913 3752 8914 3753 8915 3754 8916 3755 8917 3756 8918 3757 8919 3758 8920 3759 8921 3760 8922 3761 8923 8 8924 3762 8925 3763 8926 3764 8927 3765 8928 3766 8929 3767 8930 3768 8931 3769 8932 3770 8933 3771 8934 3772 8935 3773 8936 3774 8937 3775 8938 3776 8939 3777 8940 3778 8941 3779 8942 3780 8943 8 8944 3781 8945 8 8946 3782 8947 8 8948 8 8949 3783 8950 3784 8951 3785 8952 3786 8953 8 8954 8 8955 3787 8956 3788 8957 3789 8958 3790 8959 3791 8960 8 8961 3792 8962 8 8963 3793 8964 8 8965 8 8966 8 8967 8 8968 3794 8969 8 8970 3795 8971 3796 8972 8 8973 3797 8974 3798 8975 3799 8976 3800 8977 3801 8978 8 8979 3802 8980 3803 8981 8 8982 3804 8983 8 8984 3805 8985 3806 8986 3807 8987 3808 8988 3809 8989 3810 8990 3811 8991 8 8992 8 8993 3812 8994 8 8995 8 8996 3813 8997 3814 8998 3815 8999 8 9000 8 9001 8 9002 8 9003 3816 9004 8 9005 8 9006 8 9007 8 9008 8 9009 3817 9010 3818 9011 3819 9012 3820 9013 3821 9014 3822 9015 3823 9016 3824 9017 3825 9018 3826 9019 3827 9020 3828 9021 3829 9022 3830 9023 3831 9024 3832 9025 3833 9026 3834 9027 3835 9028 3836 9029 3837 9030 3838 9031 3839 9032 3840 9033 3841 9034 3842 9035 3843 9036 3844 9037 3845 9038 3846 9039 3847 9040 3848 9041 3849 9042 3850 9043 3820 9044 3820 9045 3820 9046 3820 9047 3820 9048 3820 9049 3820 9050 3820 9051 3820 9052 3820 9053 3820 9054 3820 9055 3820 9056 3820 9057 3820 9058 3851 9059 3852 9060 8 9061 8 9062 8 9063 8 9064 8 9065 8 9066 3853 9067 8 9068 8 9069 3854 9070 8 9071 3855 9072 3856 9073 3857 9074 3858 9075 3859 9076 8 9077 8 9078 8 9079 8 9080 8 9081 3860 9082 3861 9083 8 9084 8 9085 3862 9086 8 9087 3863 9088 8 9089 8 9090 3864 9091 3865 9092 8 9093 3820 9094 3820 9095 3820 9096 3820 9097 3820 9098 3866 9099 3867 9100 3868 9101 3820 9102 3820 9103 3820 9104 3820 9105 3820 9106 3820 9107 3820 9108 3820 9109 3820 9110 3820 9111 3869 9112 8 9113 8 9114 8 9115 8 9116 8 9117 8 9118 8 9119 8 9120 8 9121 8 9122 8 9123 8 9124 8 9125 8 9126 8 9127 8 9128 8 9129 8 9130 8 9131 8 9132 8 9133 8 9134 8 9135 8 9136 8 9137 8 9138 8 9139 8 9140 8 9141 8 9142 8 9143 8 9144 8 9145 8 9146 8 9147 8 9148 8 9149 8 9150 8 9151 8 9152 8 9153 8 9154 8 9155 8 9156 8 9157 8 9158 8 9159 8 9160 8 9161 8 9162 8 9163 8 9164 8 9165 8 9166 8 9167 8 9168 8 9169 8 9170 8 9171 8 9172 8 9173 8 9174 8 9175 8 9176 8 9177 8 9178 8 9179 8 9180 8 9181 8 9182 8 9183 8 9184 8 9185 8 9186 8 9187 8 9188 8 9189 8 9190 8 9191 8 9192 3870 9193 8 9194 8 9195 8 9196 3871 9197 3872 9198 8 9199 3873 9200 8 9201 8 9202 3874 9203 3875 9204 3876 9205 8 9206 8 9207 3877 9208 3878 9209 3879 9210 3880 9211 8 9212 3881 9213 8 9214 3882 9215 3883 9216 3884 9217 3885 9218 3886 9219 3887 9220 3888 9221 3889 9222 3890 9223 3820 9224 3820 9225 3820 9226 3820 9227 3820 9228 3820 9229 3820 9230 3820 9231 3820 9232 3820 9233 3820 9234 3820 9235 3820 9236 3891 9237 3820 9238 3892 9239 3893 9240 8 9241 3894 9242 3895 9243 8 9244 3896 9245 8 9246 3897 9247 8 9248 8 9249 8 9250 3898 9251 3899 9252 3900 9253 8 9254 3901 9255 8 9256 8 9257 3902 9258 3903 9259 3904 9260 8 9261 8 9262 8 9263 3905 9264 3906 9265 8 9266 8 9267 8 9268 3907 9269 3908 9270 3909 9271 3910 9272 8 9273 8 9274 3911 9275 8 9276 3912 9277 3913 9278 8 9279 8 9280 3914 9281 8 9282 8 9283 8 9284 3915 9285 3916 9286 8 9287 8 9288 3917 9289 3918 9290 3919 9291 3920 9292 3921 9293 3922 9294 3923 9295 3924 9296 3925 9297 3926 9298 3927 9299 8 9300 3928 9301 3929 9302 3930 9303 3931 9304 3932 9305 3933 9306 3934 9307 3935 9308 3936 9309 3937 9310 8 9311 8 9312 3938 9313 8 9314 3939 9315 8 9316 3940 9317 3941 9318 8 9319 8 9320 8 9321 3942 9322 8 9323 8 9324 3943 9325 3944 9326 3945 9327 3946 9328 8 9329 8 9330 3947 9331 8 9332 3948 9333 3949 9334 8 9335 8 9336 3950 9337 8 9338 3951 9339 8 9340 3952 9341 3953 9342 3954 9343 3820 9344 3820 9345 3955 9346 3956 9347 3820 9348 3820 9349 3820 9350 3820 9351 8 9352 3957 9353 8 9354 3958 9355 8 9356 8 9357 3820 9358 3820 9359 8 9360 3820 9361 3820 9362 3959 9363 8 9364 8 9365 8 9366 8 9367 8 9368 8 9369 3960 9370 3961 9371 3962 9372 3963 9373 8 9374 3964 9375 8 9376 3965 9377 3966 9378 3967 9379 3968 9380 3969 9381 8 9382 8 9383 8 9384 8 9385 8 9386 3970 9387 8 9388 8 9389 3971 9390 3972 9391 3973 9392 3974 9393 3975 9394 3976 9395 3977 9396 8 9397 3978 9398 8 9399 3979 9400 3980 9401 3981 9402 3982 9403 3983 9404 8 9405 3984 9406 3985 9407 3986 9408 3987 9409 3820 9410 3988 9411 3989 9412 3990 9413 8 9414 3820 9415 3820 9416 3820 9417 8 9418 3991 9419 8 9420 8 9421 8 9422 3992 9423 3993 9424 3994 9425 8 9426 8 9427 3995 9428 3996 9429 8 9430 3997 9431 3998 9432 8 9433 3820 9434 8 9435 3820 9436 3820 9437 3820 9438 3999 9439 4000 9440 8 9441 3820 9442 3820 9443 3820 9444 4001 9445 8 9446 8 9447 4002 9448 4003 9449 8 9450 4004 9451 8 9452 3820 9453 4005 9454 8 9455 4006 9456 8 9457 4007 9458 4008 9459 8 9460 8 9461 8 9462 4009 9463 8 9464 8 9465 4010 9466 4011 9467 4012 9468 4013 9469 8 9470 8 9471 4014 9472 8 9473 4015 9474 4016 9475 8 9476 8 9477 4017 9478 8 9479 4018 9480 4019 9481 8 9482 4020 9483 4021 9484 8 9485 4022 9486 8 9487 4023 9488 8 9489 8 9490 8 9491 4024 9492 4025 9493 4026 9494 8 9495 4027 9496 8 9497 8 9498 4028 9499 4029 9500 4030 9501 8 9502 8 9503 8 9504 4031 9505 4032 9506 8 9507 8 9508 8 9509 4033 9510 4034 9511 4035 9512 4036 9513 8 9514 8 9515 4037 9516 8 9517 4038 9518 4039 9519 8 9520 8 9521 4040 9522 8 9523 8 9524 8 9525 4041 9526 4042 9527 8 9528 4043 9529 8 9530 4044 9531 4045 9532 4046 9533 3643 9534 3643 9535 4047 9536 4048 9537 3643 9538 3643 9539 3643 9540 3643 9541 8 9542 4049 9543 8 9544 4050 9545 8 9546 8 9547 3643 9548 3643 9549 8 9550 3643 9551 3643 9552 4051 9553 8 9554 8 9555 8 9556 8 9557 8 9558 8 9559 4052 9560 4053 9561 4054 9562 4055 9563 8 9564 4056 9565 8 9566 4057 9567 4058 9568 4059 9569 4060 9570 4061 9571 8 9572 8 9573 8 9574 8 9575 8 9576 4062 9577 8 9578 8 9579 4063 9580 4064 9581 4065 9582 4066 9583 4067 9584 4068 9585 4069 9586 8 9587 4070 9588 8 9589 4071 9590 4072 9591 4073 9592 4074 9593 4075 9594 8 9595 4076 9596 4077 9597 3701 9598 4078 9599 4079 9600 4080 9601 8 9602 3643 9603 3643 9604 3643 9605 8 9606 4081 9607 8 9608 8 9609 8 9610 4082 9611 4083 9612 4084 9613 8 9614 8 9615 4085 9616 4086 9617 8 9618 4087 9619 4088 9620 8 9621 3643 9622 8 9623 3643 9624 3643 9625 3643 9626 4089 9627 4090 9628 8 9629 3643 9630 3643 9631 3643 9632 4091 9633 8 9634 8 9635 4092 9636 4093 9637 8 9638 4094 9639 8 9640 3643 9641 4095 9642 8 9643 4096 9644 8 9645 4097 9646 4098 9647 8 9648 8 9649 8 9650 4099 9651 8 9652 8 9653 8 9654 4100 9655 4101 9656 4102 9657 4103 9658 8 9659 8 9660 4104 9661 8 9662 4105 9663 4106 9664 8 9665 8 9666 4107 9667 8 9668 8 9669 4108 9670 8 9671 4109 9672 4110 9673 4111 9674 823 9675 823 9676 4112 9677 4113 9678 823 9679 823 9680 823 9681 823 9682 8 9683 4114 9684 8 9685 4115 9686 8 9687 8 9688 823 9689 823 9690 8 9691 823 9692 823 9693 4116 9694 8 9695 8 9696 8 9697 8 9698 8 9699 8 9700 4117 9701 4118 9702 4119 9703 4120 9704 8 9705 4121 9706 8 9707 4122 9708 4123 9709 4124 9710 4125 9711 4126 9712 8 9713 8 9714 8 9715 8 9716 8 9717 4127 9718 8 9719 8 9720 4128 9721 4129 9722 4130 9723 4131 9724 4132 9725 4133 9726 4134 9727 8 9728 4135 9729 8 9730 4136 9731 4137 9732 4138 9733 4139 9734 4140 9735 8 9736 4141 9737 4142 9738 3234 9739 4143 9740 4144 9741 4145 9742 8 9743 823 9744 823 9745 823 9746 8 9747 4146 9748 8 9749 8 9750 8 9751 4147 9752 4148 9753 4149 9754 8 9755 8 9756 4150 9757 4151 9758 8 9759 4152 9760 4153 9761 8 9762 3643 9763 823 9764 8 9765 823 9766 823 9767 823 9768 4154 9769 4155 9770 8 9771 823 9772 823 9773 823 9774 4156 9775 8 9776 8 9777 4157 9778 4158 9779 8 9780 4159 9781 8 9782 823 9783 8 9784 4160 9785 4161 9786 8 9787 8 9788 8 9789 8 9790 8 9791 8 9792 8 9793 4162 9794 4163 9795 4164 9796 4165 9797 8 9798 4166 9799 8 9800 4167 9801 4168 9802 4169 9803 4170 9804 4171 9805 8 9806 8 9807 8 9808 8 9809 8 9810 4172 9811 4173 9812 8 9813 8 9814 4174 9815 8 9816 8 9817 4175 9818 4176 9819 4177 9820 8 9821 8 9822 4178 9823 8 9824 8 9825 4179 9826 4180 9827 4181 9828 4182 9829 4183 9830 4184 9831 4185 9832 8 9833 4186 9834 8 9835 8 9836 4187 9837 4188 9838 8 9839 4189 9840 4190 9841 8 9842 4191 9843 8 9844 8 9845 4192 9846 4193 9847 8 9848 4194 9849 8 9850 8 9851 4195 9852 4196 9853 4197 9854 4198 9855 4199 9856 4200 9857 4201 9858 4202 9859 4203 9860 4204 9861 4205 9862 8 9863 4206 9864 4207 9865 4208 9866 4209 9867 4210 9868 4211 9869 4212 9870 4213 9871 4214 9872 4215 9873 8 9874 4216 9875 557 9876 4217 9877 8 9878 4218 9879 8 9880 4219 9881 4220 9882 8 9883 8 9884 8 9885 4221 9886 8 9887 8 9888 4222 9889 4223 9890 4224 9891 4225 9892 8 9893 8 9894 4226 9895 8 9896 4227 9897 4228 9898 8 9899 8 9900 4229 9901 8 9902 4230 9903 8 9904 4231 9905 4232 9906 4233 9907 663 9908 663 9909 4234 9910 4235 9911 663 9912 663 9913 663 9914 663 9915 8 9916 4236 9917 8 9918 4237 9919 8 9920 8 9921 663 9922 663 9923 8 9924 663 9925 663 9926 4238 9927 8 9928 8 9929 8 9930 8 9931 8 9932 8 9933 4239 9934 4240 9935 4241 9936 4242 9937 8 9938 4243 9939 8 9940 4244 9941 4245 9942 4246 9943 4247 9944 4248 9945 8 9946 8 9947 8 9948 8 9949 8 9950 4249 9951 8 9952 8 9953 4250 9954 4251 9955 4252 9956 4253 9957 4254 9958 4255 9959 4256 9960 8 9961 4257 9962 8 9963 4258 9964 4259 9965 4260 9966 4261 9967 4262 9968 8 9969 4263 9970 4264 9971 4265 9972 4266 9973 663 9974 4267 9975 4268 9976 4269 9977 8 9978 663 9979 663 9980 663 9981 8 9982 4270 9983 8 9984 8 9985 8 9986 4271 9987 4272 9988 4273 9989 8 9990 8 9991 4274 9992 4275 9993 8 9994 4276 9995 4277 9996 8 9997 663 9998 8 9999 663 10000 663 10001 663 10002 4278 10003 4279 10004 8 10005 663 10006 663 10007 663 10008 4280 10009 8 10010 8 10011 4281 10012 4282 10013 8 10014 4283 10015 8 10016 663 10017 557 10018 8 10019 557 10020 557 10021 4284 10022 8 10023 8 10024 4285 10025 4286 10026 4287 10027 8 10028 8 10029 557 10030 4288 10031 4289 10032 8 10033 8 10034 8 10035 8 10036 8 10037 8 10038 8 10039 4290 10040 4291 10041 4292 10042 4293 10043 8 10044 4294 10045 8 10046 4295 10047 4296 10048 4297 10049 4298 10050 4299 10051 8 10052 8 10053 8 10054 8 10055 8 10056 4300 10057 4301 10058 8 10059 8 10060 4302 10061 8 10062 8 10063 4303 10064 4304 10065 4305 10066 8 10067 8 10068 4306 10069 8 10070 8 10071 4307 10072 4308 10073 4309 10074 4310 10075 4311 10076 4312 10077 4313 10078 8 10079 4314 10080 8 10081 8 10082 4315 10083 4316 10084 8 10085 4317 10086 4318 10087 4319 10088 4320 10089 4321 10090 8 10091 4322 10092 4323 10093 4324 10094 4325 10095 557 10096 4326 10097 4327 10098 4328 10099 8 10100 557 10101 557 10102 4329 10103 8 10104 8 10105 8 10106 4330 10107 4331 10108 4332 10109 8 10110 8 10111 4333 10112 4334 10113 8 10114 4335 10115 4336 10116 8 10117 557 10118 4337 10119 4338 10120 8 10121 4339 10122 8 10123 8 10124 8 10125 8 10126 8 10127 8 10128 8 10129 8 10130 8 10131 8 10132 8 10133 8 10134 8 10135 8 10136 8 10137 8 10138 8 10139 8 10140 8 10141 8 10142 8 10143 8 10144 8 10145 8 10146 8 10147 8 10148 8 10149 8 10150 8 10151 8 10152 8 10153 8 10154 8 10155 8 10156 8 10157 8 10158 8 10159 8 10160 8 10161 8 10162 8 10163 8 10164 8 10165 8 10166 8 10167 8 10168 8 10169 8 10170 8 10171 8 10172 8 10173 8 10174 8 10175 8 10176 8 10177 8 10178 8 10179 8 10180 8 10181 8 10182 8 10183 8 10184 8 10185 8 10186 8 10187 8 10188 8 10189 8 10190 8 10191 8 10192 8 10193 8 10194 8 10195 8 10196 8 10197 8 10198 8 10199 8 10200 8 10201 8 10202 4340 10203 8 10204 8 10205 8 10206 4341 10207 4342 10208 8 10209 8 10210 4343 10211 4344 10212 4345 10213 4346 10214 8 10215 4347 10216 8 10217 4348 10218 4349 10219 4350 10220 4351 10221 4352 10222 4353 10223 4354 10224 4355 10225 4356 10226 557 10227 557 10228 557 10229 557 10230 557 10231 557 10232 557 10233 557 10234 557 10235 557 10236 557 10237 557 10238 557 10239 4357 10240 557 10241 4358 10242 4359 10243 4360 10244 8 10245 4361 10246 8 10247 4362 10248 8 10249 8 10250 8 10251 4363 10252 4364 10253 4365 10254 8 10255 8 10256 8 10257 4366 10258 4367 10259 8 10260 8 10261 8 10262 4368 10263 4369 10264 4370 10265 4371 10266 8 10267 4372 10268 8 10269 4373 10270 8 10271 8 10272 4374 10273 8 10274 8 10275 4375 10276 557 10277 4376 10278 4377 10279 4378 10280 4379 10281 4380 10282 4381 10283 4382 10284 4383 10285 4384 10286 4385 10287 4386 10288 8 10289 4387 10290 4388 10291 4389 10292 4390 10293 4391 10294 4392 10295 4393 10296 4394 10297 4395 10298 4396 10299 8 10300 4397 10301 8 10302 4398 10303 8 10304 4399 10305 4400 10306 8 10307 8 10308 8 10309 4401 10310 8 10311 8 10312 4402 10313 4403 10314 4404 10315 4405 10316 8 10317 8 10318 4406 10319 8 10320 4407 10321 4408 10322 8 10323 8 10324 4409 10325 8 10326 4410 10327 8 10328 4411 10329 4412 10330 4413 10331 557 10332 557 10333 557 10334 557 10335 4414 10336 8 10337 4415 10338 8 10339 8 10340 557 10341 557 10342 8 10343 8 10344 557 10345 8 10346 557 10347 557 10348 557 10349 557 10350 557 10351 4416 10352 8 10353 8 10354 4417 10355 4418 10356 8 10357 4419 10358 8 10359 557 10360 4420 10361 4421 10362 4422 10363 4423 10364 8 10365 4424 10366 8 10367 4425 10368 4426 10369 4427 10370 4428 10371 4429 10372 4430 10373 4431 10374 4432 10375 4433 10376 460 10377 460 10378 460 10379 460 10380 460 10381 460 10382 460 10383 460 10384 460 10385 460 10386 460 10387 460 10388 460 10389 4434 10390 460 10391 4435 10392 4436 10393 8 10394 4437 10395 4438 10396 8 10397 4439 10398 8 10399 4440 10400 8 10401 8 10402 8 10403 4441 10404 4442 10405 4443 10406 8 10407 4444 10408 8 10409 8 10410 4445 10411 4446 10412 4447 10413 8 10414 8 10415 8 10416 4448 10417 4449 10418 8 10419 8 10420 8 10421 4450 10422 4451 10423 4452 10424 4453 10425 8 10426 8 10427 4454 10428 8 10429 4455 10430 4456 10431 8 10432 8 10433 4457 10434 8 10435 8 10436 8 10437 4458 10438 4459 10439 8 10440 8 10441 4460 10442 4461 10443 4462 10444 4463 10445 4464 10446 4465 10447 4466 10448 4467 10449 4468 10450 4469 10451 4470 10452 8 10453 4471 10454 4472 10455 4473 10456 4474 10457 4475 10458 4476 10459 4477 10460 4478 10461 4479 10462 4480 10463 4481 10464 4482 10465 4483 10466 4484 10467 4485 10468 4486 10469 4487 10470 4488 10471 4489 10472 8 10473 4490 10474 8 10475 4491 10476 8 10477 8 10478 4492 10479 4493 10480 4494 10481 4495 10482 8 10483 8 10484 4496 10485 4497 10486 4498 10487 4499 10488 4500 10489 8 10490 4501 10491 8 10492 4502 10493 8 10494 8 10495 8 10496 8 10497 4503 10498 8 10499 4504 10500 4505 10501 8 10502 4506 10503 4507 10504 4508 10505 4509 10506 4510 10507 8 10508 4511 10509 4512 10510 8 10511 4513 10512 8 10513 4514 10514 4515 10515 4516 10516 4517 10517 4518 10518 4519 10519 4520 10520 8 10521 8 10522 4521 10523 8 10524 8 10525 4522 10526 4523 10527 4524 10528 8 10529 8 10530 8 10531 8 10532 4525 10533 8 10534 8 10535 8 10536 8 10537 8 10538 4526 10539 4527 10540 4528 10541 4529 10542 4530 10543 4531 10544 4532 10545 4533 10546 4534 10547 4535 10548 4536 10549 4537 10550 4538 10551 4539 10552 4540 10553 4541 10554 4542 10555 4543 10556 4544 10557 4545 10558 4546 10559 4547 10560 4548 10561 4549 10562 4550 10563 4551 10564 4552 10565 4553 10566 4554 10567 4555 10568 4556 10569 4557 10570 4558 10571 4559 10572 4529 10573 4529 10574 4529 10575 4529 10576 4529 10577 4529 10578 4529 10579 4529 10580 4529 10581 4529 10582 4529 10583 4529 10584 4529 10585 4529 10586 4529 10587 4560 10588 4561 10589 8 10590 8 10591 8 10592 8 10593 8 10594 8 10595 4562 10596 8 10597 8 10598 4563 10599 8 10600 4564 10601 4565 10602 4566 10603 4567 10604 4568 10605 8 10606 8 10607 8 10608 8 10609 8 10610 4569 10611 4570 10612 8 10613 8 10614 4571 10615 8 10616 4572 10617 8 10618 8 10619 4573 10620 4574 10621 8 10622 4529 10623 4529 10624 4529 10625 4529 10626 4529 10627 4575 10628 4576 10629 4577 10630 4529 10631 4529 10632 4529 10633 4529 10634 4529 10635 4529 10636 4529 10637 4529 10638 4529 10639 4529 10640 4578 10641 8 10642 8 10643 8 10644 8 10645 8 10646 8 10647 8 10648 8 10649 8 10650 8 10651 8 10652 8 10653 8 10654 8 10655 8 10656 8 10657 8 10658 8 10659 8 10660 8 10661 8 10662 8 10663 8 10664 8 10665 8 10666 8 10667 8 10668 8 10669 8 10670 8 10671 8 10672 8 10673 8 10674 8 10675 8 10676 8 10677 8 10678 8 10679 8 10680 8 10681 8 10682 8 10683 8 10684 8 10685 8 10686 8 10687 8 10688 8 10689 8 10690 8 10691 8 10692 8 10693 8 10694 8 10695 8 10696 8 10697 8 10698 8 10699 8 10700 8 10701 8 10702 8 10703 8 10704 8 10705 8 10706 8 10707 8 10708 8 10709 8 10710 8 10711 8 10712 8 10713 8 10714 8 10715 8 10716 8 10717 8 10718 8 10719 8 10720 8 10721 4579 10722 8 10723 8 10724 8 10725 4580 10726 4581 10727 8 10728 4582 10729 8 10730 8 10731 4583 10732 4584 10733 4585 10734 8 10735 8 10736 4586 10737 4587 10738 4588 10739 4589 10740 8 10741 4590 10742 8 10743 4591 10744 4592 10745 4593 10746 4594 10747 4595 10748 4596 10749 4597 10750 4598 10751 4599 10752 4529 10753 4529 10754 4529 10755 4529 10756 4529 10757 4529 10758 4529 10759 4529 10760 4529 10761 4529 10762 4529 10763 4529 10764 4529 10765 4600 10766 4529 10767 4601 10768 4602 10769 8 10770 4603 10771 4604 10772 8 10773 4605 10774 8 10775 4606 10776 8 10777 8 10778 8 10779 4607 10780 4608 10781 4609 10782 8 10783 4610 10784 8 10785 8 10786 4611 10787 4612 10788 4613 10789 8 10790 8 10791 8 10792 4614 10793 4615 10794 8 10795 8 10796 8 10797 4616 10798 4617 10799 4618 10800 4619 10801 8 10802 8 10803 4620 10804 8 10805 4621 10806 4622 10807 8 10808 8 10809 4623 10810 8 10811 8 10812 8 10813 4624 10814 4625 10815 8 10816 8 10817 4626 10818 4627 10819 4628 10820 4629 10821 4630 10822 4631 10823 4632 10824 4633 10825 4634 10826 4635 10827 4636 10828 8 10829 4637 10830 4638 10831 4639 10832 4640 10833 4641 10834 4642 10835 4643 10836 4644 10837 4645 10838 4646 10839 8 10840 8 10841 4647 10842 8 10843 4648 10844 8 10845 4649 10846 4650 10847 8 10848 8 10849 8 10850 4651 10851 8 10852 8 10853 4652 10854 4653 10855 4654 10856 4655 10857 8 10858 8 10859 4656 10860 8 10861 4657 10862 4658 10863 8 10864 8 10865 4659 10866 8 10867 4660 10868 8 10869 4661 10870 4662 10871 4663 10872 4529 10873 4529 10874 4664 10875 4665 10876 4666 10877 4667 10878 4668 10879 4669 10880 4670 10881 4671 10882 4672 10883 4673 10884 8 10885 4674 10886 8 10887 4675 10888 8 10889 8 10890 4676 10891 4677 10892 4678 10893 4679 10894 8 10895 8 10896 4680 10897 4681 10898 4682 10899 4683 10900 4684 10901 8 10902 4685 10903 8 10904 4686 10905 8 10906 8 10907 8 10908 8 10909 4687 10910 8 10911 4688 10912 4689 10913 8 10914 4690 10915 4691 10916 4692 10917 4693 10918 4694 10919 8 10920 4695 10921 4696 10922 8 10923 4697 10924 8 10925 4698 10926 4699 10927 4700 10928 4701 10929 4702 10930 4703 10931 4704 10932 8 10933 8 10934 4705 10935 8 10936 8 10937 4706 10938 4707 10939 4708 10940 8 10941 8 10942 8 10943 8 10944 4709 10945 8 10946 8 10947 8 10948 8 10949 8 10950 4710 10951 4711 10952 4712 10953 4713 10954 4714 10955 4715 10956 4716 10957 4717 10958 4718 10959 4719 10960 4720 10961 4721 10962 4722 10963 4723 10964 4724 10965 4725 10966 4726 10967 4727 10968 4728 10969 4729 10970 4730 10971 4731 10972 4732 10973 4733 10974 4734 10975 4735 10976 4736 10977 4737 10978 4738 10979 4739 10980 4740 10981 4741 10982 4742 10983 4743 10984 4713 10985 4713 10986 4713 10987 4713 10988 4713 10989 4713 10990 4713 10991 4713 10992 4713 10993 4713 10994 4713 10995 4713 10996 4713 10997 4713 10998 4713 10999 4744 11000 4745 11001 8 11002 8 11003 8 11004 8 11005 8 11006 8 11007 4746 11008 8 11009 8 11010 4747 11011 8 11012 4748 11013 4749 11014 4750 11015 4751 11016 4752 11017 8 11018 8 11019 8 11020 8 11021 8 11022 4753 11023 4754 11024 8 11025 8 11026 4755 11027 8 11028 4756 11029 8 11030 8 11031 4757 11032 4758 11033 8 11034 4713 11035 4713 11036 4713 11037 4713 11038 4713 11039 4759 11040 4760 11041 4761 11042 4713 11043 4713 11044 4713 11045 4713 11046 4713 11047 4713 11048 4713 11049 4713 11050 4713 11051 4713 11052 4762 11053 8 11054 8 11055 8 11056 8 11057 8 11058 8 11059 8 11060 8 11061 8 11062 8 11063 8 11064 8 11065 8 11066 8 11067 8 11068 8 11069 8 11070 8 11071 8 11072 8 11073 8 11074 8 11075 8 11076 8 11077 8 11078 8 11079 8 11080 8 11081 8 11082 8 11083 8 11084 8 11085 8 11086 8 11087 8 11088 8 11089 8 11090 8 11091 8 11092 8 11093 8 11094 8 11095 8 11096 8 11097 8 11098 8 11099 8 11100 8 11101 8 11102 8 11103 8 11104 8 11105 8 11106 8 11107 8 11108 8 11109 8 11110 8 11111 8 11112 8 11113 8 11114 8 11115 8 11116 8 11117 8 11118 8 11119 8 11120 8 11121 8 11122 8 11123 8 11124 8 11125 8 11126 8 11127 8 11128 8 11129 8 11130 8 11131 8 11132 8 11133 4763 11134 8 11135 8 11136 8 11137 4764 11138 4765 11139 8 11140 4766 11141 8 11142 8 11143 4767 11144 4768 11145 4769 11146 8 11147 8 11148 4770 11149 4771 11150 4772 11151 4773 11152 8 11153 4774 11154 8 11155 4775 11156 4776 11157 4777 11158 4778 11159 4779 11160 4780 11161 4781 11162 4782 11163 4783 11164 4713 11165 4713 11166 4713 11167 4713 11168 4713 11169 4713 11170 4713 11171 4713 11172 4713 11173 4713 11174 4713 11175 4713 11176 4713 11177 4784 11178 4713 11179 4785 11180 4786 11181 8 11182 4787 11183 4788 11184 8 11185 4789 11186 8 11187 4790 11188 8 11189 8 11190 8 11191 4791 11192 4792 11193 4793 11194 8 11195 4794 11196 8 11197 8 11198 4795 11199 4796 11200 4797 11201 8 11202 8 11203 8 11204 4798 11205 4799 11206 8 11207 8 11208 8 11209 4800 11210 4801 11211 4802 11212 4803 11213 8 11214 8 11215 4804 11216 8 11217 4805 11218 4806 11219 8 11220 8 11221 4807 11222 8 11223 8 11224 8 11225 4808 11226 4809 11227 8 11228 8 11229 4810 11230 4811 11231 4812 11232 4813 11233 4814 11234 4815 11235 4816 11236 4817 11237 4818 11238 4819 11239 4820 11240 8 11241 4821 11242 4822 11243 4823 11244 4824 11245 4825 11246 4826 11247 4827 11248 4828 11249 4829 11250 4830 11251 8 11252 4831 11253 4529 11254 4832 11255 8 11256 4833 11257 8 11258 4834 11259 4835 11260 8 11261 8 11262 8 11263 4836 11264 8 11265 8 11266 4837 11267 4838 11268 4839 11269 4840 11270 8 11271 8 11272 4841 11273 8 11274 4842 11275 4843 11276 8 11277 8 11278 4844 11279 8 11280 4845 11281 8 11282 4846 11283 4847 11284 4848 11285 4713 11286 4713 11287 4849 11288 4850 11289 4713 11290 4713 11291 4713 11292 4713 11293 8 11294 4851 11295 8 11296 4852 11297 8 11298 8 11299 4713 11300 4713 11301 8 11302 4713 11303 4713 11304 4853 11305 8 11306 8 11307 8 11308 8 11309 8 11310 8 11311 4854 11312 4855 11313 4856 11314 4857 11315 8 11316 4858 11317 8 11318 4859 11319 4860 11320 4861 11321 4862 11322 4863 11323 8 11324 8 11325 8 11326 8 11327 8 11328 4864 11329 8 11330 8 11331 4865 11332 4866 11333 4867 11334 4868 11335 4869 11336 4870 11337 4871 11338 8 11339 4872 11340 8 11341 4873 11342 4874 11343 4875 11344 4876 11345 4877 11346 8 11347 4878 11348 4879 11349 4880 11350 4881 11351 4713 11352 4882 11353 4883 11354 4884 11355 8 11356 4713 11357 4713 11358 4713 11359 8 11360 4885 11361 8 11362 8 11363 8 11364 4886 11365 4887 11366 4888 11367 8 11368 8 11369 4889 11370 4890 11371 8 11372 4891 11373 4892 11374 8 11375 4713 11376 8 11377 4713 11378 4713 11379 4713 11380 4893 11381 4894 11382 8 11383 4713 11384 4713 11385 4713 11386 4895 11387 8 11388 8 11389 4896 11390 4897 11391 8 11392 4898 11393 8 11394 4713 11395 4529 11396 4529 11397 4529 11398 8 11399 4899 11400 8 11401 4900 11402 8 11403 8 11404 4529 11405 4529 11406 8 11407 4529 11408 4529 11409 4901 11410 8 11411 8 11412 8 11413 8 11414 8 11415 8 11416 4902 11417 4903 11418 4904 11419 4905 11420 8 11421 4906 11422 8 11423 4907 11424 4908 11425 4909 11426 4910 11427 4911 11428 8 11429 8 11430 8 11431 8 11432 8 11433 4912 11434 8 11435 8 11436 4913 11437 4914 11438 4915 11439 4916 11440 4917 11441 4918 11442 4919 11443 8 11444 4920 11445 8 11446 4921 11447 4922 11448 4923 11449 4924 11450 4925 11451 8 11452 4926 11453 4927 11454 4928 11455 4929 11456 4529 11457 4930 11458 4931 11459 4932 11460 8 11461 4529 11462 4529 11463 4529 11464 8 11465 4933 11466 8 11467 8 11468 8 11469 4934 11470 4935 11471 4936 11472 8 11473 8 11474 4937 11475 4938 11476 8 11477 4939 11478 4940 11479 8 11480 4529 11481 8 11482 4529 11483 4529 11484 4529 11485 4941 11486 4942 11487 8 11488 4529 11489 4529 11490 4529 11491 4943 11492 8 11493 8 11494 4944 11495 4945 11496 8 11497 4946 11498 8 11499 4529 11500 4947 11501 364 11502 4948 11503 8 11504 4949 11505 8 11506 4950 11507 4951 11508 8 11509 8 11510 8 11511 4952 11512 8 11513 8 11514 4953 11515 4954 11516 4955 11517 4956 11518 8 11519 8 11520 4957 11521 8 11522 4958 11523 4959 11524 8 11525 8 11526 4960 11527 8 11528 4961 11529 8 11530 4962 11531 4963 11532 4964 11533 460 11534 460 11535 4965 11536 4966 11537 4967 11538 4968 11539 4969 11540 4970 11541 4971 11542 4972 11543 4973 11544 4974 11545 8 11546 4975 11547 8 11548 4976 11549 8 11550 8 11551 4977 11552 4978 11553 4979 11554 4980 11555 8 11556 8 11557 4981 11558 4982 11559 4983 11560 4984 11561 4985 11562 8 11563 4986 11564 8 11565 4987 11566 8 11567 8 11568 8 11569 8 11570 4988 11571 8 11572 4989 11573 4990 11574 8 11575 4991 11576 4992 11577 4993 11578 4994 11579 4995 11580 8 11581 4996 11582 4997 11583 8 11584 4998 11585 8 11586 4999 11587 5000 11588 5001 11589 5002 11590 5003 11591 5004 11592 5005 11593 8 11594 8 11595 5006 11596 8 11597 8 11598 5007 11599 5008 11600 5009 11601 8 11602 8 11603 8 11604 8 11605 5010 11606 8 11607 8 11608 8 11609 8 11610 8 11611 5011 11612 5012 11613 5013 11614 5014 11615 5015 11616 5016 11617 5017 11618 5018 11619 5019 11620 5020 11621 5021 11622 5022 11623 5023 11624 5024 11625 5025 11626 5026 11627 5027 11628 5028 11629 5029 11630 5030 11631 5031 11632 5032 11633 5033 11634 5034 11635 5035 11636 5036 11637 5037 11638 5038 11639 5039 11640 5040 11641 5041 11642 5042 11643 5043 11644 5044 11645 5014 11646 5014 11647 5014 11648 5014 11649 5014 11650 5014 11651 5014 11652 5014 11653 5014 11654 5014 11655 5014 11656 5014 11657 5014 11658 5014 11659 5014 11660 5045 11661 5046 11662 8 11663 8 11664 8 11665 8 11666 8 11667 8 11668 5047 11669 8 11670 8 11671 5048 11672 8 11673 5049 11674 5050 11675 5051 11676 5052 11677 5053 11678 8 11679 8 11680 8 11681 8 11682 8 11683 5054 11684 5055 11685 8 11686 8 11687 5056 11688 8 11689 5057 11690 8 11691 8 11692 5058 11693 5059 11694 8 11695 5014 11696 5014 11697 5014 11698 5014 11699 5014 11700 5060 11701 5061 11702 5062 11703 5014 11704 5014 11705 5014 11706 5014 11707 5014 11708 5014 11709 5014 11710 5014 11711 5014 11712 5014 11713 5063 11714 8 11715 8 11716 8 11717 8 11718 8 11719 8 11720 8 11721 8 11722 8 11723 8 11724 8 11725 8 11726 8 11727 8 11728 8 11729 8 11730 8 11731 8 11732 8 11733 8 11734 8 11735 8 11736 8 11737 8 11738 8 11739 8 11740 8 11741 8 11742 8 11743 8 11744 8 11745 8 11746 8 11747 8 11748 8 11749 8 11750 8 11751 8 11752 8 11753 8 11754 8 11755 8 11756 8 11757 8 11758 8 11759 8 11760 8 11761 8 11762 8 11763 8 11764 8 11765 8 11766 8 11767 8 11768 8 11769 8 11770 8 11771 8 11772 8 11773 8 11774 8 11775 8 11776 8 11777 8 11778 8 11779 8 11780 8 11781 8 11782 8 11783 8 11784 8 11785 8 11786 8 11787 8 11788 8 11789 8 11790 8 11791 8 11792 8 11793 8 11794 5064 11795 8 11796 8 11797 8 11798 5065 11799 5066 11800 8 11801 5067 11802 8 11803 8 11804 5068 11805 5069 11806 5070 11807 8 11808 8 11809 5071 11810 5072 11811 5073 11812 5074 11813 8 11814 5075 11815 8 11816 5076 11817 5077 11818 5078 11819 5079 11820 5080 11821 5081 11822 5082 11823 5083 11824 5084 11825 5014 11826 5014 11827 5014 11828 5014 11829 5014 11830 5014 11831 5014 11832 5014 11833 5014 11834 5014 11835 5014 11836 5014 11837 5014 11838 5085 11839 5014 11840 5086 11841 5087 11842 8 11843 5088 11844 5089 11845 8 11846 5090 11847 8 11848 5091 11849 8 11850 8 11851 8 11852 5092 11853 5093 11854 5094 11855 8 11856 5095 11857 8 11858 8 11859 5096 11860 5097 11861 5098 11862 8 11863 8 11864 8 11865 5099 11866 5100 11867 8 11868 8 11869 8 11870 5101 11871 5102 11872 5103 11873 5104 11874 8 11875 8 11876 5105 11877 8 11878 5106 11879 5107 11880 8 11881 8 11882 5108 11883 8 11884 8 11885 8 11886 5109 11887 5110 11888 8 11889 8 11890 5111 11891 5112 11892 5113 11893 5114 11894 5115 11895 5116 11896 5117 11897 5118 11898 5119 11899 5120 11900 5121 11901 8 11902 5122 11903 5123 11904 5124 11905 5125 11906 5126 11907 5127 11908 5128 11909 5129 11910 5130 11911 5131 11912 8 11913 5132 11914 460 11915 5133 11916 8 11917 5134 11918 8 11919 5135 11920 5136 11921 8 11922 8 11923 8 11924 5137 11925 8 11926 8 11927 5138 11928 5139 11929 5140 11930 5141 11931 8 11932 8 11933 5142 11934 8 11935 5143 11936 5144 11937 8 11938 8 11939 5145 11940 8 11941 5146 11942 8 11943 5147 11944 5148 11945 5149 11946 5014 11947 5014 11948 5150 11949 5151 11950 5014 11951 5014 11952 5014 11953 5014 11954 8 11955 5152 11956 8 11957 5153 11958 8 11959 8 11960 5014 11961 5014 11962 8 11963 5014 11964 5014 11965 5154 11966 8 11967 8 11968 8 11969 8 11970 8 11971 8 11972 5155 11973 5156 11974 5157 11975 5158 11976 8 11977 5159 11978 8 11979 5160 11980 5161 11981 5162 11982 5163 11983 5164 11984 8 11985 8 11986 8 11987 8 11988 8 11989 5165 11990 8 11991 8 11992 5166 11993 5167 11994 5168 11995 5169 11996 5170 11997 5171 11998 5172 11999 8 12000 5173 12001 8 12002 5174 12003 5175 12004 5176 12005 5177 12006 5178 12007 8 12008 5179 12009 5180 12010 5181 12011 5182 12012 5014 12013 5183 12014 5184 12015 5185 12016 8 12017 5014 12018 5014 12019 5014 12020 8 12021 5186 12022 8 12023 8 12024 8 12025 5187 12026 5188 12027 5189 12028 8 12029 8 12030 5190 12031 5191 12032 8 12033 5192 12034 5193 12035 8 12036 5014 12037 8 12038 5014 12039 5014 12040 5014 12041 5194 12042 5195 12043 8 12044 5014 12045 5014 12046 5014 12047 5196 12048 8 12049 8 12050 5197 12051 5198 12052 8 12053 5199 12054 8 12055 5014 12056 460 12057 460 12058 460 12059 8 12060 5200 12061 8 12062 5201 12063 8 12064 8 12065 460 12066 460 12067 8 12068 460 12069 460 12070 5202 12071 8 12072 8 12073 8 12074 8 12075 8 12076 8 12077 5203 12078 5204 12079 5205 12080 5206 12081 8 12082 5207 12083 8 12084 5208 12085 5209 12086 5210 12087 5211 12088 5212 12089 8 12090 8 12091 8 12092 8 12093 8 12094 5213 12095 8 12096 8 12097 5214 12098 5215 12099 5216 12100 5217 12101 5218 12102 5219 12103 5220 12104 8 12105 5221 12106 8 12107 5222 12108 5223 12109 5224 12110 5225 12111 5226 12112 8 12113 5227 12114 5228 12115 5229 12116 5230 12117 460 12118 5231 12119 5232 12120 5233 12121 8 12122 460 12123 460 12124 460 12125 8 12126 5234 12127 8 12128 8 12129 8 12130 5235 12131 5236 12132 5237 12133 8 12134 8 12135 5238 12136 5239 12137 8 12138 5240 12139 5241 12140 8 12141 460 12142 8 12143 460 12144 460 12145 460 12146 5242 12147 5243 12148 8 12149 460 12150 364 12151 460 12152 460 12153 5244 12154 8 12155 8 12156 5245 12157 5246 12158 8 12159 5247 12160 8 12161 460 12162 364 12163 364 12164 364 12165 364 12166 364 12167 364 12168 364 12169 364 12170 5248 12171 8 12172 8 12173 8 12174 8 12175 8 12176 8 12177 8 12178 8 12179 8 12180 8 12181 8 12182 8 12183 8 12184 8 12185 8 12186 8 12187 8 12188 8 12189 8 12190 8 12191 8 12192 8 12193 8 12194 8 12195 8 12196 8 12197 8 12198 8 12199 8 12200 8 12201 8 12202 8 12203 8 12204 8 12205 8 12206 8 12207 8 12208 8 12209 8 12210 8 12211 8 12212 8 12213 8 12214 8 12215 8 12216 8 12217 8 12218 8 12219 8 12220 8 12221 8 12222 8 12223 8 12224 8 12225 8 12226 8 12227 8 12228 8 12229 8 12230 8 12231 8 12232 8 12233 8 12234 8 12235 8 12236 8 12237 8 12238 8 12239 8 12240 8 12241 8 12242 8 12243 8 12244 8 12245 8 12246 8 12247 8 12248 8 12249 8 12250 8 12251 5249 12252 8 12253 8 12254 8 12255 5250 12256 5251 12257 8 12258 5252 12259 8 12260 8 12261 5253 12262 5254 12263 5255 12264 8 12265 8 12266 5256 12267 5257 12268 5258 12269 5259 12270 8 12271 5260 12272 8 12273 5261 12274 5262 12275 5263 12276 5264 12277 5265 12278 5266 12279 5267 12280 5268 12281 5269 12282 364 12283 364 12284 364 12285 364 12286 364 12287 364 12288 364 12289 364 12290 364 12291 364 12292 364 12293 364 12294 364 12295 5270 12296 364 12297 5271 12298 5272 12299 8 12300 5273 12301 5274 12302 8 12303 5275 12304 8 12305 5276 12306 8 12307 8 12308 8 12309 5277 12310 5278 12311 5279 12312 8 12313 5280 12314 8 12315 8 12316 5281 12317 5282 12318 5283 12319 8 12320 8 12321 8 12322 5284 12323 5285 12324 8 12325 8 12326 8 12327 5286 12328 5287 12329 5288 12330 5289 12331 8 12332 8 12333 5290 12334 8 12335 5291 12336 5292 12337 8 12338 8 12339 5293 12340 8 12341 8 12342 8 12343 5294 12344 5295 12345 8 12346 8 12347 5296 12348 5297 12349 5298 12350 5299 12351 5300 12352 5301 12353 5302 12354 5303 12355 5304 12356 5305 12357 5306 12358 8 12359 5307 12360 5308 12361 5309 12362 5310 12363 5311 12364 5312 12365 5313 12366 5314 12367 5315 12368 5316 12369 5317 12370 5318 12371 5319 12372 5320 12373 5321 12374 5322 12375 5323 12376 5324 12377 5325 12378 8 12379 5326 12380 8 12381 5327 12382 8 12383 8 12384 5328 12385 5329 12386 5330 12387 5331 12388 8 12389 8 12390 5332 12391 5333 12392 5334 12393 5335 12394 5336 12395 8 12396 5337 12397 8 12398 5338 12399 8 12400 8 12401 8 12402 8 12403 5339 12404 8 12405 5340 12406 5341 12407 8 12408 5342 12409 5343 12410 5344 12411 5345 12412 5346 12413 8 12414 5347 12415 5348 12416 8 12417 5349 12418 8 12419 5350 12420 5351 12421 5352 12422 5353 12423 5354 12424 5355 12425 5356 12426 8 12427 8 12428 5357 12429 8 12430 8 12431 5358 12432 5359 12433 5360 12434 8 12435 8 12436 8 12437 8 12438 5361 12439 8 12440 8 12441 8 12442 8 12443 8 12444 5362 12445 5363 12446 5364 12447 5365 12448 5366 12449 5367 12450 5368 12451 5369 12452 5370 12453 5371 12454 5372 12455 5373 12456 5374 12457 5375 12458 5376 12459 5377 12460 5378 12461 5379 12462 5380 12463 5381 12464 5382 12465 5383 12466 5384 12467 5385 12468 5386 12469 5387 12470 5388 12471 5389 12472 5390 12473 5391 12474 5392 12475 5393 12476 5394 12477 5395 12478 5365 12479 5365 12480 5365 12481 5365 12482 5365 12483 5365 12484 5365 12485 5365 12486 5365 12487 5365 12488 5365 12489 5365 12490 5365 12491 5365 12492 5365 12493 5396 12494 5397 12495 8 12496 8 12497 8 12498 8 12499 8 12500 8 12501 5398 12502 8 12503 8 12504 5399 12505 8 12506 5400 12507 5401 12508 5402 12509 5403 12510 5404 12511 8 12512 8 12513 8 12514 8 12515 8 12516 5405 12517 5406 12518 8 12519 8 12520 5407 12521 8 12522 5408 12523 8 12524 8 12525 5409 12526 5410 12527 8 12528 5365 12529 5365 12530 5365 12531 5365 12532 5365 12533 5411 12534 5412 12535 5413 12536 5365 12537 5365 12538 5365 12539 5365 12540 5365 12541 5365 12542 5365 12543 5365 12544 5365 12545 5365 12546 5414 12547 8 12548 8 12549 8 12550 8 12551 8 12552 8 12553 8 12554 8 12555 8 12556 8 12557 8 12558 8 12559 8 12560 8 12561 8 12562 8 12563 8 12564 8 12565 8 12566 8 12567 8 12568 8 12569 8 12570 8 12571 8 12572 8 12573 8 12574 8 12575 8 12576 8 12577 8 12578 8 12579 8 12580 8 12581 8 12582 8 12583 8 12584 8 12585 8 12586 8 12587 8 12588 8 12589 8 12590 8 12591 8 12592 8 12593 8 12594 8 12595 8 12596 8 12597 8 12598 8 12599 8 12600 8 12601 8 12602 8 12603 8 12604 8 12605 8 12606 8 12607 8 12608 8 12609 8 12610 8 12611 8 12612 8 12613 8 12614 8 12615 8 12616 8 12617 8 12618 8 12619 8 12620 8 12621 8 12622 8 12623 8 12624 8 12625 8 12626 8 12627 5415 12628 8 12629 8 12630 8 12631 5416 12632 5417 12633 8 12634 5418 12635 8 12636 8 12637 5419 12638 5420 12639 5421 12640 8 12641 8 12642 5422 12643 5423 12644 5424 12645 5425 12646 8 12647 5426 12648 8 12649 5427 12650 5428 12651 5429 12652 5430 12653 5431 12654 5432 12655 5433 12656 5434 12657 5435 12658 5365 12659 5365 12660 5365 12661 5365 12662 5365 12663 5365 12664 5365 12665 5365 12666 5365 12667 5365 12668 5365 12669 5365 12670 5365 12671 5436 12672 5365 12673 5437 12674 5438 12675 8 12676 5439 12677 5440 12678 8 12679 5441 12680 8 12681 5442 12682 8 12683 8 12684 8 12685 5443 12686 5444 12687 5445 12688 8 12689 5446 12690 8 12691 8 12692 5447 12693 5448 12694 5449 12695 8 12696 8 12697 8 12698 5450 12699 5451 12700 8 12701 8 12702 8 12703 5452 12704 5453 12705 5454 12706 5455 12707 8 12708 8 12709 5456 12710 8 12711 5457 12712 5458 12713 8 12714 8 12715 5459 12716 8 12717 8 12718 8 12719 5460 12720 5461 12721 8 12722 8 12723 5462 12724 5463 12725 5464 12726 5465 12727 5466 12728 5467 12729 5468 12730 5469 12731 5470 12732 5471 12733 5472 12734 8 12735 5473 12736 5474 12737 5475 12738 5476 12739 5477 12740 5478 12741 5479 12742 5480 12743 5481 12744 5482 12745 8 12746 8 12747 5483 12748 8 12749 5484 12750 8 12751 5485 12752 5486 12753 8 12754 8 12755 8 12756 5487 12757 8 12758 8 12759 5488 12760 5489 12761 5490 12762 5491 12763 8 12764 8 12765 5492 12766 8 12767 5493 12768 5494 12769 8 12770 8 12771 5495 12772 8 12773 5496 12774 8 12775 5497 12776 5498 12777 5499 12778 5365 12779 5365 12780 5500 12781 5501 12782 5502 12783 5503 12784 5504 12785 5505 12786 5506 12787 5507 12788 5508 12789 5509 12790 8 12791 5510 12792 8 12793 5511 12794 8 12795 8 12796 5512 12797 5513 12798 5514 12799 5515 12800 8 12801 8 12802 5516 12803 5517 12804 5518 12805 5519 12806 5520 12807 8 12808 5521 12809 8 12810 5522 12811 8 12812 8 12813 8 12814 8 12815 5523 12816 8 12817 5524 12818 5525 12819 8 12820 5526 12821 5527 12822 5528 12823 5529 12824 5530 12825 8 12826 5531 12827 5532 12828 8 12829 5533 12830 8 12831 5534 12832 5535 12833 5536 12834 5537 12835 5538 12836 5539 12837 5540 12838 8 12839 8 12840 5541 12841 8 12842 8 12843 5542 12844 5543 12845 5544 12846 8 12847 8 12848 8 12849 8 12850 5545 12851 8 12852 8 12853 8 12854 8 12855 8 12856 5546 12857 5547 12858 5548 12859 5549 12860 5550 12861 5551 12862 5552 12863 5553 12864 5554 12865 5555 12866 5556 12867 5557 12868 5558 12869 5559 12870 5560 12871 5561 12872 5562 12873 5563 12874 5564 12875 5565 12876 5566 12877 5567 12878 5568 12879 5569 12880 5570 12881 5571 12882 5572 12883 5573 12884 5574 12885 5575 12886 5576 12887 5577 12888 5578 12889 5579 12890 5549 12891 5549 12892 5549 12893 5549 12894 5549 12895 5549 12896 5549 12897 5549 12898 5549 12899 5549 12900 5549 12901 5549 12902 5549 12903 5549 12904 5549 12905 5580 12906 5581 12907 8 12908 8 12909 8 12910 8 12911 8 12912 8 12913 5582 12914 8 12915 8 12916 5583 12917 8 12918 5584 12919 5585 12920 5586 12921 5587 12922 5588 12923 8 12924 8 12925 8 12926 8 12927 8 12928 5589 12929 5590 12930 8 12931 8 12932 5591 12933 8 12934 5592 12935 8 12936 8 12937 5593 12938 5594 12939 8 12940 5549 12941 5549 12942 5549 12943 5549 12944 5549 12945 5595 12946 5596 12947 5597 12948 5549 12949 5549 12950 5549 12951 5549 12952 5549 12953 5549 12954 5549 12955 5549 12956 5549 12957 5549 12958 5598 12959 8 12960 8 12961 8 12962 8 12963 8 12964 8 12965 8 12966 8 12967 8 12968 8 12969 8 12970 8 12971 8 12972 8 12973 8 12974 8 12975 8 12976 8 12977 8 12978 8 12979 8 12980 8 12981 8 12982 8 12983 8 12984 8 12985 8 12986 8 12987 8 12988 8 12989 8 12990 8 12991 8 12992 8 12993 8 12994 8 12995 8 12996 8 12997 8 12998 8 12999 8 13000 8 13001 8 13002 8 13003 8 13004 8 13005 8 13006 8 13007 8 13008 8 13009 8 13010 8 13011 8 13012 8 13013 8 13014 8 13015 8 13016 8 13017 8 13018 8 13019 8 13020 8 13021 8 13022 8 13023 8 13024 8 13025 8 13026 8 13027 8 13028 8 13029 8 13030 8 13031 8 13032 8 13033 8 13034 8 13035 8 13036 8 13037 8 13038 8 13039 5599 13040 8 13041 8 13042 8 13043 5600 13044 5601 13045 8 13046 5602 13047 8 13048 8 13049 5603 13050 5604 13051 5605 13052 8 13053 8 13054 5606 13055 5607 13056 5608 13057 5609 13058 8 13059 5610 13060 8 13061 5611 13062 5612 13063 5613 13064 5614 13065 5615 13066 5616 13067 5617 13068 5618 13069 5619 13070 5549 13071 5549 13072 5549 13073 5549 13074 5549 13075 5549 13076 5549 13077 5549 13078 5549 13079 5549 13080 5549 13081 5549 13082 5549 13083 5620 13084 5549 13085 5621 13086 5622 13087 8 13088 5623 13089 5624 13090 8 13091 5625 13092 8 13093 5626 13094 8 13095 8 13096 8 13097 5627 13098 5628 13099 5629 13100 8 13101 5630 13102 8 13103 8 13104 5631 13105 5632 13106 5633 13107 8 13108 8 13109 8 13110 5634 13111 5635 13112 8 13113 8 13114 8 13115 5636 13116 5637 13117 5638 13118 5639 13119 8 13120 8 13121 5640 13122 8 13123 5641 13124 5642 13125 8 13126 8 13127 5643 13128 8 13129 8 13130 8 13131 5644 13132 5645 13133 8 13134 8 13135 5646 13136 5647 13137 5648 13138 5649 13139 5650 13140 5651 13141 5652 13142 5653 13143 5654 13144 5655 13145 5656 13146 8 13147 5657 13148 5658 13149 5659 13150 5660 13151 5661 13152 5662 13153 5663 13154 5664 13155 5665 13156 5666 13157 8 13158 5667 13159 5365 13160 5668 13161 8 13162 5669 13163 8 13164 5670 13165 5671 13166 8 13167 8 13168 8 13169 5672 13170 8 13171 8 13172 5673 13173 5674 13174 5675 13175 5676 13176 8 13177 8 13178 5677 13179 8 13180 5678 13181 5679 13182 8 13183 8 13184 5680 13185 8 13186 5681 13187 8 13188 5682 13189 5683 13190 5684 13191 5549 13192 5549 13193 5685 13194 5686 13195 5549 13196 5549 13197 5549 13198 5549 13199 8 13200 5687 13201 8 13202 5688 13203 8 13204 8 13205 5549 13206 5549 13207 8 13208 5549 13209 5549 13210 5689 13211 8 13212 8 13213 8 13214 8 13215 8 13216 8 13217 5690 13218 5691 13219 5692 13220 5693 13221 8 13222 5694 13223 8 13224 5695 13225 5696 13226 5697 13227 5698 13228 5699 13229 8 13230 8 13231 8 13232 8 13233 8 13234 5700 13235 8 13236 8 13237 5701 13238 5702 13239 5703 13240 5704 13241 5705 13242 5706 13243 5707 13244 8 13245 5708 13246 8 13247 5709 13248 5710 13249 5711 13250 5712 13251 5713 13252 8 13253 5714 13254 5715 13255 5716 13256 5717 13257 5549 13258 5718 13259 5719 13260 5720 13261 8 13262 5549 13263 5549 13264 5549 13265 8 13266 5721 13267 8 13268 8 13269 8 13270 5722 13271 5723 13272 5724 13273 8 13274 8 13275 5725 13276 5726 13277 8 13278 5727 13279 5728 13280 8 13281 5549 13282 8 13283 5549 13284 5549 13285 5549 13286 5729 13287 5730 13288 8 13289 5549 13290 5549 13291 5549 13292 5731 13293 8 13294 8 13295 5732 13296 5733 13297 8 13298 5734 13299 8 13300 5549 13301 5365 13302 5365 13303 5365 13304 8 13305 5735 13306 8 13307 5736 13308 8 13309 8 13310 5365 13311 5365 13312 8 13313 5365 13314 5365 13315 5737 13316 8 13317 8 13318 8 13319 8 13320 8 13321 8 13322 5738 13323 5739 13324 5740 13325 5741 13326 8 13327 5742 13328 8 13329 5743 13330 5744 13331 5745 13332 5746 13333 5747 13334 8 13335 8 13336 8 13337 8 13338 8 13339 5748 13340 8 13341 8 13342 5749 13343 5750 13344 5751 13345 5752 13346 5753 13347 5754 13348 5755 13349 8 13350 5756 13351 8 13352 5757 13353 5758 13354 5759 13355 5760 13356 5761 13357 8 13358 5762 13359 5763 13360 5764 13361 5765 13362 5365 13363 5766 13364 5767 13365 5768 13366 8 13367 5365 13368 5365 13369 5365 13370 8 13371 5769 13372 8 13373 8 13374 8 13375 5770 13376 5771 13377 5772 13378 8 13379 8 13380 5773 13381 5774 13382 8 13383 5775 13384 5776 13385 8 13386 5365 13387 8 13388 5365 13389 5365 13390 5365 13391 5777 13392 5778 13393 8 13394 5365 13395 5365 13396 5365 13397 5779 13398 8 13399 8 13400 5780 13401 5781 13402 8 13403 5782 13404 8 13405 5365 13406 8 13407 5783 13408 8 13409 5784 13410 8 13411 5785 13412 5786 13413 8 13414 8 13415 8 13416 5787 13417 8 13418 8 13419 5788 13420 5789 13421 5790 13422 5791 13423 8 13424 8 13425 5792 13426 8 13427 5793 13428 5794 13429 8 13430 8 13431 5795 13432 8 13433 5796 13434 8 13435 5797 13436 5798 13437 5799 13438 364 13439 364 13440 5800 13441 5801 13442 5802 13443 5803 13444 5804 13445 5805 13446 5806 13447 5807 13448 5808 13449 5809 13450 8 13451 5810 13452 8 13453 5811 13454 8 13455 8 13456 5812 13457 5813 13458 5814 13459 5815 13460 8 13461 8 13462 5816 13463 5817 13464 5818 13465 5819 13466 5820 13467 8 13468 5821 13469 8 13470 5822 13471 8 13472 8 13473 8 13474 8 13475 5823 13476 8 13477 5824 13478 5825 13479 8 13480 5826 13481 5827 13482 5828 13483 5829 13484 5830 13485 8 13486 5831 13487 5832 13488 8 13489 5833 13490 8 13491 5834 13492 5835 13493 5836 13494 5837 13495 5838 13496 5839 13497 5840 13498 8 13499 8 13500 5841 13501 8 13502 8 13503 5842 13504 5843 13505 5844 13506 8 13507 8 13508 8 13509 8 13510 5845 13511 8 13512 8 13513 8 13514 8 13515 8 13516 5846 13517 5847 13518 5848 13519 5849 13520 5850 13521 5851 13522 5852 13523 5853 13524 5854 13525 5855 13526 5856 13527 5857 13528 5858 13529 5859 13530 5860 13531 5861 13532 5862 13533 5863 13534 5864 13535 5865 13536 5866 13537 5867 13538 5868 13539 5869 13540 5870 13541 5871 13542 5872 13543 5873 13544 5874 13545 5875 13546 5876 13547 5877 13548 5878 13549 5879 13550 5849 13551 5849 13552 5849 13553 5849 13554 5849 13555 5849 13556 5849 13557 5849 13558 5849 13559 5849 13560 5849 13561 5849 13562 5849 13563 5849 13564 5849 13565 5880 13566 5881 13567 8 13568 8 13569 8 13570 8 13571 8 13572 8 13573 5882 13574 8 13575 8 13576 5883 13577 8 13578 5884 13579 5885 13580 5886 13581 5887 13582 5888 13583 8 13584 8 13585 8 13586 8 13587 8 13588 5889 13589 5890 13590 8 13591 8 13592 5891 13593 8 13594 5892 13595 8 13596 8 13597 5893 13598 5894 13599 8 13600 5849 13601 5849 13602 5849 13603 5849 13604 5849 13605 5895 13606 5896 13607 5897 13608 5849 13609 5849 13610 5849 13611 5849 13612 5849 13613 5849 13614 5849 13615 5849 13616 5849 13617 5849 13618 5898 13619 8 13620 8 13621 8 13622 8 13623 8 13624 8 13625 8 13626 8 13627 8 13628 8 13629 8 13630 8 13631 8 13632 8 13633 8 13634 8 13635 8 13636 8 13637 8 13638 8 13639 8 13640 8 13641 8 13642 8 13643 8 13644 8 13645 8 13646 8 13647 8 13648 8 13649 8 13650 8 13651 8 13652 8 13653 8 13654 8 13655 8 13656 8 13657 8 13658 8 13659 8 13660 8 13661 8 13662 8 13663 8 13664 8 13665 8 13666 8 13667 8 13668 8 13669 8 13670 8 13671 8 13672 8 13673 8 13674 8 13675 8 13676 8 13677 8 13678 8 13679 8 13680 8 13681 8 13682 8 13683 8 13684 8 13685 8 13686 8 13687 8 13688 8 13689 8 13690 8 13691 8 13692 8 13693 8 13694 8 13695 8 13696 8 13697 8 13698 8 13699 5899 13700 8 13701 8 13702 8 13703 5900 13704 5901 13705 8 13706 5902 13707 8 13708 8 13709 5903 13710 5904 13711 5905 13712 8 13713 8 13714 5906 13715 5907 13716 5908 13717 5909 13718 8 13719 5910 13720 8 13721 5911 13722 5912 13723 5913 13724 5914 13725 5915 13726 5916 13727 5917 13728 5918 13729 5919 13730 5849 13731 5849 13732 5849 13733 5849 13734 5849 13735 5849 13736 5849 13737 5849 13738 5849 13739 5849 13740 5849 13741 5849 13742 5849 13743 5920 13744 5849 13745 5921 13746 5922 13747 8 13748 5923 13749 5924 13750 8 13751 5925 13752 8 13753 5926 13754 8 13755 8 13756 8 13757 5927 13758 5928 13759 5929 13760 8 13761 5930 13762 8 13763 8 13764 5931 13765 5932 13766 5933 13767 8 13768 8 13769 8 13770 5934 13771 5935 13772 8 13773 8 13774 8 13775 5936 13776 5937 13777 5938 13778 5939 13779 8 13780 8 13781 5940 13782 8 13783 5941 13784 5942 13785 8 13786 8 13787 5943 13788 8 13789 8 13790 8 13791 5944 13792 5945 13793 8 13794 8 13795 5946 13796 5947 13797 5948 13798 5949 13799 5950 13800 5951 13801 5952 13802 5953 13803 5954 13804 5955 13805 5956 13806 8 13807 5957 13808 5958 13809 5959 13810 5960 13811 5961 13812 5962 13813 5963 13814 5964 13815 5965 13816 5966 13817 8 13818 5967 13819 364 13820 5968 13821 8 13822 5969 13823 8 13824 5970 13825 5971 13826 8 13827 8 13828 8 13829 5972 13830 8 13831 8 13832 5973 13833 5974 13834 5975 13835 5976 13836 8 13837 8 13838 5977 13839 8 13840 5978 13841 5979 13842 8 13843 8 13844 5980 13845 8 13846 5981 13847 8 13848 5982 13849 5983 13850 5984 13851 5849 13852 5849 13853 5985 13854 5986 13855 5849 13856 5849 13857 5849 13858 5849 13859 8 13860 5987 13861 8 13862 5988 13863 8 13864 8 13865 5849 13866 5849 13867 8 13868 5849 13869 5849 13870 5989 13871 8 13872 8 13873 8 13874 8 13875 8 13876 8 13877 5990 13878 5991 13879 5992 13880 5993 13881 8 13882 5994 13883 8 13884 5995 13885 5996 13886 5997 13887 5998 13888 5999 13889 8 13890 8 13891 8 13892 8 13893 8 13894 6000 13895 8 13896 8 13897 6001 13898 6002 13899 6003 13900 6004 13901 6005 13902 6006 13903 6007 13904 8 13905 6008 13906 8 13907 6009 13908 6010 13909 6011 13910 6012 13911 6013 13912 8 13913 6014 13914 6015 13915 6016 13916 6017 13917 5849 13918 6018 13919 6019 13920 6020 13921 8 13922 5849 13923 5849 13924 5849 13925 8 13926 6021 13927 8 13928 8 13929 8 13930 6022 13931 6023 13932 6024 13933 8 13934 8 13935 6025 13936 6026 13937 8 13938 6027 13939 6028 13940 8 13941 5849 13942 8 13943 5849 13944 5849 13945 5849 13946 6029 13947 6030 13948 8 13949 5849 13950 5849 13951 5849 13952 6031 13953 8 13954 8 13955 6032 13956 6033 13957 8 13958 6034 13959 8 13960 5849 13961 364 13962 364 13963 364 13964 8 13965 6035 13966 8 13967 6036 13968 8 13969 8 13970 364 13971 364 13972 8 13973 364 13974 364 13975 6037 13976 8 13977 8 13978 8 13979 8 13980 8 13981 8 13982 6038 13983 6039 13984 6040 13985 6041 13986 8 13987 6042 13988 8 13989 6043 13990 6044 13991 6045 13992 6046 13993 6047 13994 8 13995 8 13996 8 13997 8 13998 8 13999 6048 14000 8 14001 8 14002 6049 14003 6050 14004 6051 14005 6052 14006 6053 14007 6054 14008 6055 14009 8 14010 6056 14011 8 14012 6057 14013 6058 14014 6059 14015 6060 14016 6061 14017 8 14018 6062 14019 6063 14020 6064 14021 6065 14022 364 14023 6066 14024 6067 14025 6068 14026 8 14027 364 14028 364 14029 364 14030 8 14031 6069 14032 8 14033 8 14034 8 14035 6070 14036 6071 14037 6072 14038 8 14039 8 14040 6073 14041 6074 14042 8 14043 6075 14044 6076 14045 8 14046 364 14047 8 14048 364 14049 364 14050 364 14051 6077 14052 6078 14053 8 14054 364 14055 364 14056 364 14057 6079 14058 8 14059 8 14060 6080 14061 6081 14062 8 14063 6082 14064 8 14065 364 14066 8 14067 6083 14068 8 14069 6084 14070 8 14071 8 14072 6085 14073 6086 14074 8 14075 267 14076 267 14077 267 14078 267 14079 267 14080 6087 14081 6088 14082 6089 14083 267 14084 267 14085 267 14086 267 14087 267 14088 267 14089 267 14090 267 14091 267 14092 267 14093 6090 14094 8 14095 8 14096 8 14097 8 14098 8 14099 8 14100 8 14101 8 14102 8 14103 8 14104 8 14105 8 14106 8 14107 8 14108 8 14109 8 14110 8 14111 8 14112 8 14113 8 14114 8 14115 8 14116 8 14117 8 14118 8 14119 8 14120 8 14121 8 14122 8 14123 8 14124 8 14125 8 14126 8 14127 8 14128 8 14129 8 14130 8 14131 8 14132 8 14133 8 14134 8 14135 8 14136 8 14137 8 14138 8 14139 8 14140 8 14141 8 14142 8 14143 8 14144 8 14145 8 14146 8 14147 8 14148 8 14149 8 14150 8 14151 8 14152 8 14153 8 14154 8 14155 8 14156 8 14157 8 14158 8 14159 8 14160 8 14161 8 14162 8 14163 8 14164 8 14165 8 14166 8 14167 8 14168 8 14169 8 14170 8 14171 8 14172 8 14173 8 14174 6091 14175 8 14176 8 14177 8 14178 6092 14179 6093 14180 8 14181 6094 14182 8 14183 8 14184 6095 14185 6096 14186 6097 14187 8 14188 8 14189 6098 14190 6099 14191 6100 14192 6101 14193 8 14194 6102 14195 8 14196 6103 14197 6104 14198 6105 14199 6106 14200 6107 14201 6108 14202 6109 14203 6110 14204 6111 14205 267 14206 267 14207 267 14208 267 14209 267 14210 267 14211 267 14212 267 14213 267 14214 267 14215 267 14216 267 14217 267 14218 6112 14219 267 14220 6113 14221 6114 14222 8 14223 6115 14224 6116 14225 8 14226 6117 14227 8 14228 6118 14229 8 14230 8 14231 8 14232 6119 14233 6120 14234 6121 14235 8 14236 6122 14237 8 14238 8 14239 6123 14240 6124 14241 6125 14242 8 14243 8 14244 8 14245 6126 14246 6127 14247 8 14248 8 14249 8 14250 6128 14251 6129 14252 6130 14253 6131 14254 8 14255 8 14256 6132 14257 8 14258 6133 14259 6134 14260 8 14261 8 14262 6135 14263 8 14264 8 14265 8 14266 6136 14267 6137 14268 8 14269 8 14270 8 14271 6138 14272 6139 14273 6140 14274 6141 14275 6142 14276 6143 14277 6144 14278 6145 14279 6146 14280 6147 14281 6148 14282 8 14283 6149 14284 6150 14285 6151 14286 6152 14287 6153 14288 6154 14289 6155 14290 6156 14291 6157 14292 6158 14293 6159 14294 6160 14295 6161 14296 6162 14297 6163 14298 6164 14299 6165 14300 6166 14301 6167 14302 8 14303 6168 14304 8 14305 6169 14306 8 14307 8 14308 6170 14309 6171 14310 6172 14311 6173 14312 8 14313 8 14314 6174 14315 6175 14316 6176 14317 6177 14318 6178 14319 8 14320 6179 14321 8 14322 6180 14323 8 14324 8 14325 8 14326 8 14327 6181 14328 8 14329 6182 14330 6183 14331 8 14332 6184 14333 6185 14334 6186 14335 6187 14336 6188 14337 8 14338 6189 14339 6190 14340 8 14341 6191 14342 8 14343 6192 14344 6193 14345 6194 14346 6195 14347 6196 14348 6197 14349 6198 14350 8 14351 8 14352 6199 14353 8 14354 8 14355 6200 14356 6201 14357 6202 14358 8 14359 8 14360 8 14361 8 14362 6203 14363 8 14364 8 14365 8 14366 8 14367 8 14368 6204 14369 6205 14370 6206 14371 6207 14372 6208 14373 6209 14374 6210 14375 6211 14376 6212 14377 6213 14378 6214 14379 6215 14380 6216 14381 6217 14382 6218 14383 6219 14384 6220 14385 6221 14386 6222 14387 6223 14388 6224 14389 6225 14390 6226 14391 6227 14392 6228 14393 6229 14394 6230 14395 6231 14396 6232 14397 6233 14398 6234 14399 6235 14400 6236 14401 6237 14402 6207 14403 6207 14404 6207 14405 6207 14406 6207 14407 6207 14408 6207 14409 6207 14410 6207 14411 6207 14412 6207 14413 6207 14414 6207 14415 6207 14416 6207 14417 6238 14418 6239 14419 8 14420 8 14421 8 14422 8 14423 8 14424 8 14425 6240 14426 8 14427 8 14428 6241 14429 8 14430 6242 14431 6243 14432 6244 14433 6245 14434 6246 14435 8 14436 8 14437 8 14438 8 14439 8 14440 6247 14441 6248 14442 8 14443 8 14444 6249 14445 8 14446 6250 14447 8 14448 8 14449 6251 14450 6252 14451 8 14452 6207 14453 6207 14454 6207 14455 6207 14456 6207 14457 6253 14458 6254 14459 6255 14460 6207 14461 6207 14462 6207 14463 6207 14464 6207 14465 6207 14466 6207 14467 6207 14468 6207 14469 6207 14470 6256 14471 8 14472 8 14473 8 14474 8 14475 8 14476 8 14477 8 14478 8 14479 8 14480 8 14481 8 14482 8 14483 8 14484 8 14485 8 14486 8 14487 8 14488 8 14489 8 14490 8 14491 8 14492 8 14493 8 14494 8 14495 8 14496 8 14497 8 14498 8 14499 8 14500 8 14501 8 14502 8 14503 8 14504 8 14505 8 14506 8 14507 8 14508 8 14509 8 14510 8 14511 8 14512 8 14513 8 14514 8 14515 8 14516 8 14517 8 14518 8 14519 8 14520 8 14521 8 14522 8 14523 8 14524 8 14525 8 14526 8 14527 8 14528 8 14529 8 14530 8 14531 8 14532 8 14533 8 14534 8 14535 8 14536 8 14537 8 14538 8 14539 8 14540 8 14541 8 14542 8 14543 8 14544 8 14545 8 14546 8 14547 8 14548 8 14549 8 14550 8 14551 6257 14552 8 14553 8 14554 8 14555 6258 14556 6259 14557 8 14558 6260 14559 8 14560 8 14561 6261 14562 6262 14563 6263 14564 8 14565 8 14566 6264 14567 6265 14568 6266 14569 6267 14570 8 14571 6268 14572 8 14573 6269 14574 6270 14575 6271 14576 6272 14577 6273 14578 6274 14579 6275 14580 6276 14581 6277 14582 6207 14583 6207 14584 6207 14585 6207 14586 6207 14587 6207 14588 6207 14589 6207 14590 6207 14591 6207 14592 6207 14593 6207 14594 6207 14595 6278 14596 6207 14597 6279 14598 6280 14599 8 14600 6281 14601 6282 14602 8 14603 6283 14604 8 14605 6284 14606 8 14607 8 14608 8 14609 6285 14610 6286 14611 6287 14612 8 14613 6288 14614 8 14615 8 14616 6289 14617 6290 14618 6291 14619 8 14620 8 14621 8 14622 6292 14623 6293 14624 8 14625 8 14626 8 14627 6294 14628 6295 14629 6296 14630 6297 14631 8 14632 8 14633 6298 14634 8 14635 6299 14636 6300 14637 8 14638 8 14639 6301 14640 8 14641 8 14642 8 14643 6302 14644 6303 14645 8 14646 8 14647 6304 14648 6305 14649 6306 14650 6307 14651 6308 14652 6309 14653 6310 14654 6311 14655 6312 14656 6313 14657 6314 14658 8 14659 6315 14660 6316 14661 6317 14662 6318 14663 6319 14664 6320 14665 6321 14666 6322 14667 6323 14668 6324 14669 8 14670 8 14671 6325 14672 8 14673 6326 14674 8 14675 6327 14676 6328 14677 8 14678 8 14679 8 14680 6329 14681 8 14682 8 14683 6330 14684 6331 14685 6332 14686 6333 14687 8 14688 8 14689 6334 14690 8 14691 6335 14692 6336 14693 8 14694 8 14695 6337 14696 8 14697 6338 14698 8 14699 6339 14700 6340 14701 6341 14702 6207 14703 6207 14704 6342 14705 6343 14706 6344 14707 6345 14708 6346 14709 6347 14710 6348 14711 6349 14712 6350 14713 6351 14714 8 14715 6352 14716 8 14717 6353 14718 8 14719 8 14720 6354 14721 6355 14722 6356 14723 6357 14724 8 14725 8 14726 6358 14727 6359 14728 6360 14729 6361 14730 6362 14731 8 14732 6363 14733 8 14734 6364 14735 8 14736 8 14737 8 14738 8 14739 6365 14740 8 14741 6366 14742 6367 14743 8 14744 6368 14745 6369 14746 6370 14747 6371 14748 6372 14749 8 14750 6373 14751 6374 14752 8 14753 6375 14754 8 14755 6376 14756 6377 14757 6378 14758 6379 14759 6380 14760 6381 14761 6382 14762 8 14763 8 14764 6383 14765 8 14766 8 14767 6384 14768 6385 14769 6386 14770 8 14771 8 14772 8 14773 8 14774 6387 14775 8 14776 8 14777 8 14778 8 14779 8 14780 6388 14781 6389 14782 6390 14783 6391 14784 6392 14785 6393 14786 6394 14787 6395 14788 6396 14789 6397 14790 6398 14791 6399 14792 6400 14793 6401 14794 6402 14795 6403 14796 6404 14797 6405 14798 6406 14799 6407 14800 6408 14801 6409 14802 6410 14803 6411 14804 6412 14805 6413 14806 6414 14807 6415 14808 6416 14809 6417 14810 6418 14811 6419 14812 6420 14813 6421 14814 6391 14815 6391 14816 6391 14817 6391 14818 6391 14819 6391 14820 6391 14821 6391 14822 6391 14823 6391 14824 6391 14825 6391 14826 6391 14827 6391 14828 6391 14829 6422 14830 6423 14831 8 14832 8 14833 8 14834 8 14835 8 14836 8 14837 6424 14838 8 14839 8 14840 6425 14841 8 14842 6426 14843 6427 14844 6428 14845 6429 14846 6430 14847 8 14848 8 14849 8 14850 8 14851 8 14852 6431 14853 6432 14854 8 14855 8 14856 6433 14857 8 14858 6434 14859 8 14860 8 14861 6435 14862 6436 14863 8 14864 6391 14865 6391 14866 6391 14867 6391 14868 6391 14869 6437 14870 6438 14871 6439 14872 6391 14873 6391 14874 6391 14875 6391 14876 6391 14877 6391 14878 6391 14879 6391 14880 6391 14881 6391 14882 6440 14883 8 14884 8 14885 8 14886 8 14887 8 14888 8 14889 8 14890 8 14891 8 14892 8 14893 8 14894 8 14895 8 14896 8 14897 8 14898 8 14899 8 14900 8 14901 8 14902 8 14903 8 14904 8 14905 8 14906 8 14907 8 14908 8 14909 8 14910 8 14911 8 14912 8 14913 8 14914 8 14915 8 14916 8 14917 8 14918 8 14919 8 14920 8 14921 8 14922 8 14923 8 14924 8 14925 8 14926 8 14927 8 14928 8 14929 8 14930 8 14931 8 14932 8 14933 8 14934 8 14935 8 14936 8 14937 8 14938 8 14939 8 14940 8 14941 8 14942 8 14943 8 14944 8 14945 8 14946 8 14947 8 14948 8 14949 8 14950 8 14951 8 14952 8 14953 8 14954 8 14955 8 14956 8 14957 8 14958 8 14959 8 14960 8 14961 8 14962 8 14963 6441 14964 8 14965 8 14966 8 14967 6442 14968 6443 14969 8 14970 6444 14971 8 14972 8 14973 6445 14974 6446 14975 6447 14976 8 14977 8 14978 6448 14979 6449 14980 6450 14981 6451 14982 8 14983 6452 14984 8 14985 6453 14986 6454 14987 6455 14988 6456 14989 6457 14990 6458 14991 6459 14992 6460 14993 6461 14994 6391 14995 6391 14996 6391 14997 6391 14998 6391 14999 6391 15000 6391 15001 6391 15002 6391 15003 6391 15004 6391 15005 6391 15006 6391 15007 6462 15008 6391 15009 6463 15010 6464 15011 8 15012 6465 15013 6466 15014 8 15015 6467 15016 8 15017 6468 15018 8 15019 8 15020 8 15021 6469 15022 6470 15023 6471 15024 8 15025 6472 15026 8 15027 8 15028 6473 15029 6474 15030 6475 15031 8 15032 8 15033 8 15034 6476 15035 6477 15036 8 15037 8 15038 8 15039 6478 15040 6479 15041 6480 15042 6481 15043 8 15044 8 15045 6482 15046 8 15047 6483 15048 6484 15049 8 15050 8 15051 6485 15052 8 15053 8 15054 8 15055 6486 15056 6487 15057 8 15058 8 15059 6488 15060 6489 15061 6490 15062 6491 15063 6492 15064 6493 15065 6494 15066 6495 15067 6496 15068 6497 15069 6498 15070 8 15071 6499 15072 6500 15073 6501 15074 6502 15075 6503 15076 6504 15077 6505 15078 6506 15079 6507 15080 6508 15081 8 15082 6509 15083 6207 15084 6510 15085 8 15086 6511 15087 8 15088 6512 15089 6513 15090 8 15091 8 15092 8 15093 6514 15094 8 15095 8 15096 6515 15097 6516 15098 6517 15099 6518 15100 8 15101 8 15102 6519 15103 8 15104 6520 15105 6521 15106 8 15107 8 15108 6522 15109 8 15110 6523 15111 8 15112 6524 15113 6525 15114 6526 15115 6391 15116 6391 15117 6527 15118 6528 15119 6391 15120 6391 15121 6391 15122 6391 15123 8 15124 6529 15125 8 15126 6530 15127 8 15128 8 15129 6391 15130 6391 15131 8 15132 6391 15133 6391 15134 6531 15135 8 15136 8 15137 8 15138 8 15139 8 15140 8 15141 6532 15142 6533 15143 6534 15144 6535 15145 8 15146 6536 15147 8 15148 6537 15149 6538 15150 6539 15151 6540 15152 6541 15153 8 15154 8 15155 8 15156 8 15157 8 15158 6542 15159 8 15160 8 15161 6543 15162 6544 15163 6545 15164 6546 15165 6547 15166 6548 15167 6549 15168 8 15169 6550 15170 8 15171 6551 15172 6552 15173 6553 15174 6554 15175 6555 15176 8 15177 6556 15178 6557 15179 6558 15180 6559 15181 6391 15182 6560 15183 6561 15184 6562 15185 8 15186 6391 15187 6391 15188 6391 15189 8 15190 6563 15191 8 15192 8 15193 8 15194 6564 15195 6565 15196 6566 15197 8 15198 8 15199 6567 15200 6568 15201 8 15202 6569 15203 6570 15204 8 15205 6391 15206 8 15207 6391 15208 6391 15209 6391 15210 6571 15211 6572 15212 8 15213 6391 15214 6391 15215 6391 15216 6573 15217 8 15218 8 15219 6574 15220 6575 15221 8 15222 6576 15223 8 15224 6391 15225 6207 15226 6207 15227 6207 15228 8 15229 6577 15230 8 15231 6578 15232 8 15233 8 15234 6207 15235 6207 15236 8 15237 6207 15238 6207 15239 6579 15240 8 15241 8 15242 8 15243 8 15244 8 15245 8 15246 6580 15247 6581 15248 6582 15249 6583 15250 8 15251 6584 15252 8 15253 6585 15254 6586 15255 6587 15256 6588 15257 6589 15258 8 15259 8 15260 8 15261 8 15262 8 15263 6590 15264 8 15265 8 15266 6591 15267 6592 15268 6593 15269 6594 15270 6595 15271 6596 15272 6597 15273 8 15274 6598 15275 8 15276 6599 15277 6600 15278 6601 15279 6602 15280 6603 15281 8 15282 6604 15283 6605 15284 6606 15285 6607 15286 6207 15287 6608 15288 6609 15289 6610 15290 8 15291 6207 15292 6207 15293 6207 15294 8 15295 6611 15296 8 15297 8 15298 8 15299 6612 15300 6613 15301 6614 15302 8 15303 8 15304 6615 15305 6616 15306 8 15307 6617 15308 6618 15309 8 15310 6207 15311 8 15312 6207 15313 6207 15314 6207 15315 6619 15316 6620 15317 8 15318 6207 15319 6207 15320 6207 15321 6621 15322 8 15323 8 15324 6622 15325 6623 15326 8 15327 6624 15328 8 15329 6207 15330 6625 15331 8 15332 6626 15333 8 15334 6627 15335 6628 15336 8 15337 8 15338 8 15339 6629 15340 8 15341 8 15342 6630 15343 6631 15344 6632 15345 6633 15346 8 15347 8 15348 6634 15349 8 15350 6635 15351 6636 15352 8 15353 8 15354 6637 15355 8 15356 6638 15357 8 15358 6639 15359 6640 15360 6641 15361 267 15362 267 15363 6642 15364 6643 15365 6644 15366 6645 15367 6646 15368 6647 15369 6648 15370 6649 15371 6650 15372 6651 15373 8 15374 6652 15375 8 15376 6653 15377 8 15378 8 15379 6654 15380 6655 15381 6656 15382 6657 15383 8 15384 8 15385 6658 15386 6659 15387 6660 15388 6661 15389 6662 15390 8 15391 6663 15392 8 15393 6664 15394 8 15395 8 15396 8 15397 8 15398 6665 15399 8 15400 6666 15401 6667 15402 8 15403 6668 15404 6669 15405 6670 15406 6671 15407 6672 15408 8 15409 6673 15410 6674 15411 8 15412 6675 15413 8 15414 6676 15415 6677 15416 6678 15417 6679 15418 6680 15419 6681 15420 6682 15421 8 15422 8 15423 6683 15424 8 15425 8 15426 6684 15427 6685 15428 6686 15429 8 15430 8 15431 8 15432 8 15433 6687 15434 8 15435 8 15436 8 15437 8 15438 8 15439 6688 15440 6689 15441 6690 15442 6691 15443 6692 15444 6693 15445 6694 15446 6695 15447 6696 15448 6697 15449 6698 15450 6699 15451 6700 15452 6701 15453 6702 15454 6703 15455 6704 15456 6705 15457 6706 15458 6707 15459 6708 15460 6709 15461 6710 15462 6711 15463 6712 15464 6713 15465 6714 15466 6715 15467 6716 15468 6717 15469 6718 15470 6719 15471 6720 15472 6721 15473 6691 15474 6691 15475 6691 15476 6691 15477 6691 15478 6691 15479 6691 15480 6691 15481 6691 15482 6691 15483 6691 15484 6691 15485 6691 15486 6691 15487 6691 15488 6722 15489 6723 15490 8 15491 8 15492 8 15493 8 15494 8 15495 8 15496 6724 15497 8 15498 8 15499 6725 15500 8 15501 6726 15502 6727 15503 6728 15504 6729 15505 6730 15506 8 15507 8 15508 8 15509 8 15510 8 15511 6731 15512 6732 15513 8 15514 8 15515 6733 15516 8 15517 6734 15518 8 15519 8 15520 6735 15521 6736 15522 8 15523 6691 15524 6691 15525 6691 15526 6691 15527 6691 15528 6737 15529 6738 15530 6739 15531 6691 15532 6691 15533 6691 15534 6691 15535 6691 15536 6691 15537 6691 15538 6691 15539 6691 15540 6691 15541 6740 15542 8 15543 8 15544 8 15545 8 15546 8 15547 8 15548 8 15549 8 15550 8 15551 8 15552 8 15553 8 15554 8 15555 8 15556 8 15557 8 15558 8 15559 8 15560 8 15561 8 15562 8 15563 8 15564 8 15565 8 15566 8 15567 8 15568 8 15569 8 15570 8 15571 8 15572 8 15573 8 15574 8 15575 8 15576 8 15577 8 15578 8 15579 8 15580 8 15581 8 15582 8 15583 8 15584 8 15585 8 15586 8 15587 8 15588 8 15589 8 15590 8 15591 8 15592 8 15593 8 15594 8 15595 8 15596 8 15597 8 15598 8 15599 8 15600 8 15601 8 15602 8 15603 8 15604 8 15605 8 15606 8 15607 8 15608 8 15609 8 15610 8 15611 8 15612 8 15613 8 15614 8 15615 8 15616 8 15617 8 15618 8 15619 8 15620 8 15621 8 15622 6741 15623 8 15624 8 15625 8 15626 6742 15627 6743 15628 8 15629 6744 15630 8 15631 8 15632 6745 15633 6746 15634 6747 15635 8 15636 8 15637 6748 15638 6749 15639 6750 15640 6751 15641 8 15642 6752 15643 8 15644 6753 15645 6754 15646 6755 15647 6756 15648 6757 15649 6758 15650 6759 15651 6760 15652 6761 15653 6691 15654 6691 15655 6691 15656 6691 15657 6691 15658 6691 15659 6691 15660 6691 15661 6691 15662 6691 15663 6691 15664 6691 15665 6691 15666 6762 15667 6691 15668 6763 15669 6764 15670 8 15671 6765 15672 6766 15673 8 15674 6767 15675 8 15676 6768 15677 8 15678 8 15679 8 15680 6769 15681 6770 15682 6771 15683 8 15684 6772 15685 8 15686 8 15687 6773 15688 6774 15689 6775 15690 8 15691 8 15692 8 15693 6776 15694 6777 15695 8 15696 8 15697 8 15698 6778 15699 6779 15700 6780 15701 6781 15702 8 15703 8 15704 6782 15705 8 15706 6783 15707 6784 15708 8 15709 8 15710 6785 15711 8 15712 8 15713 8 15714 6786 15715 6787 15716 8 15717 8 15718 6788 15719 6789 15720 6790 15721 6791 15722 6792 15723 6793 15724 6794 15725 6795 15726 6796 15727 6797 15728 6798 15729 8 15730 6799 15731 6800 15732 6801 15733 6802 15734 6803 15735 6804 15736 6805 15737 6806 15738 6807 15739 6808 15740 8 15741 6809 15742 267 15743 6810 15744 8 15745 6811 15746 8 15747 6812 15748 6813 15749 8 15750 8 15751 8 15752 6814 15753 8 15754 8 15755 6815 15756 6816 15757 6817 15758 6818 15759 8 15760 8 15761 6819 15762 8 15763 6820 15764 6821 15765 8 15766 8 15767 6822 15768 8 15769 6823 15770 8 15771 6824 15772 6825 15773 6826 15774 6691 15775 6691 15776 6827 15777 6828 15778 6691 15779 6691 15780 6691 15781 6691 15782 8 15783 6829 15784 8 15785 6830 15786 8 15787 8 15788 6691 15789 6691 15790 8 15791 6691 15792 6691 15793 6831 15794 8 15795 8 15796 8 15797 8 15798 8 15799 8 15800 6832 15801 6833 15802 6834 15803 6835 15804 8 15805 6836 15806 8 15807 6837 15808 6838 15809 6839 15810 6840 15811 6841 15812 8 15813 8 15814 8 15815 8 15816 8 15817 6842 15818 8 15819 8 15820 6843 15821 6844 15822 6845 15823 6846 15824 6847 15825 6848 15826 6849 15827 8 15828 6850 15829 8 15830 6851 15831 6852 15832 6853 15833 6854 15834 6855 15835 8 15836 6856 15837 6857 15838 6858 15839 6859 15840 6691 15841 6860 15842 6861 15843 6862 15844 8 15845 6691 15846 6691 15847 6691 15848 8 15849 6863 15850 8 15851 8 15852 8 15853 6864 15854 6865 15855 6866 15856 8 15857 8 15858 6867 15859 6868 15860 8 15861 6869 15862 6870 15863 8 15864 6691 15865 8 15866 6691 15867 6691 15868 6691 15869 6871 15870 6872 15871 8 15872 6691 15873 6691 15874 6691 15875 6873 15876 8 15877 8 15878 6874 15879 6875 15880 8 15881 6876 15882 8 15883 6691 15884 267 15885 267 15886 267 15887 8 15888 6877 15889 8 15890 6878 15891 8 15892 8 15893 267 15894 267 15895 8 15896 267 15897 267 15898 6879 15899 8 15900 8 15901 8 15902 8 15903 8 15904 8 15905 6880 15906 6881 15907 6882 15908 6883 15909 8 15910 6884 15911 8 15912 6885 15913 6886 15914 6887 15915 6888 15916 6889 15917 8 15918 8 15919 8 15920 8 15921 8 15922 6890 15923 8 15924 8 15925 6891 15926 6892 15927 6893 15928 6894 15929 6895 15930 6896 15931 6897 15932 8 15933 6898 15934 8 15935 6899 15936 6900 15937 6901 15938 6902 15939 6903 15940 8 15941 6904 15942 6905 15943 6906 15944 6907 15945 267 15946 6908 15947 6909 15948 6910 15949 8 15950 267 15951 267 15952 267 15953 8 15954 6911 15955 8 15956 8 15957 8 15958 6912 15959 6913 15960 6914 15961 8 15962 8 15963 6915 15964 6916 15965 8 15966 6917 15967 6918 15968 8 15969 267 15970 8 15971 267 15972 267 15973 267 15974 6919 15975 6920 15976 8 15977 267 15978 267 15979 267 15980 6921 15981 8 15982 8 15983 6922 15984 6923 15985 8 15986 6924 15987 8 15988 267 15989 8 15990 8 15991 8 15992 8 15993 6925 15994 6926 15995 8 15996 8 15997 6927 15998 8 15999 6928 16000 8 16001 8 16002 6929 16003 6930 16004 8 16005 186 16006 186 16007 186 16008 186 16009 186 16010 6931 16011 6932 16012 6933 16013 186 16014 186 16015 186 16016 186 16017 186 16018 186 16019 186 16020 186 16021 186 16022 186 16023 6934 16024 8 16025 8 16026 8 16027 8 16028 8 16029 8 16030 8 16031 8 16032 8 16033 8 16034 8 16035 8 16036 8 16037 8 16038 8 16039 8 16040 8 16041 8 16042 8 16043 8 16044 8 16045 8 16046 8 16047 8 16048 8 16049 8 16050 8 16051 8 16052 8 16053 8 16054 8 16055 8 16056 8 16057 8 16058 8 16059 8 16060 8 16061 8 16062 8 16063 8 16064 8 16065 8 16066 8 16067 8 16068 8 16069 8 16070 8 16071 8 16072 8 16073 8 16074 8 16075 8 16076 8 16077 8 16078 8 16079 8 16080 8 16081 8 16082 8 16083 8 16084 8 16085 8 16086 8 16087 8 16088 8 16089 8 16090 8 16091 8 16092 8 16093 8 16094 8 16095 8 16096 8 16097 8 16098 8 16099 8 16100 8 16101 8 16102 8 16103 8 16104 6935 16105 8 16106 8 16107 8 16108 6936 16109 6937 16110 8 16111 6938 16112 8 16113 8 16114 6939 16115 6940 16116 6941 16117 8 16118 8 16119 6942 16120 6943 16121 6944 16122 6945 16123 8 16124 6946 16125 8 16126 6947 16127 6948 16128 6949 16129 6950 16130 6951 16131 6952 16132 6953 16133 6954 16134 6955 16135 186 16136 186 16137 186 16138 186 16139 186 16140 186 16141 186 16142 186 16143 186 16144 186 16145 186 16146 186 16147 186 16148 6956 16149 186 16150 6957 16151 6958 16152 8 16153 6959 16154 6960 16155 8 16156 6961 16157 8 16158 6962 16159 8 16160 8 16161 8 16162 6963 16163 6964 16164 6965 16165 8 16166 6966 16167 8 16168 8 16169 6967 16170 6968 16171 6969 16172 8 16173 8 16174 8 16175 6970 16176 6971 16177 8 16178 8 16179 8 16180 6972 16181 6973 16182 6974 16183 6975 16184 8 16185 8 16186 6976 16187 8 16188 6977 16189 6978 16190 8 16191 8 16192 6979 16193 8 16194 8 16195 8 16196 6980 16197 6981 16198 8 16199 8 16200 6982 16201 6983 16202 6984 16203 6985 16204 6986 16205 6987 16206 6988 16207 6989 16208 6990 16209 6991 16210 6992 16211 8 16212 6993 16213 6994 16214 6995 16215 6996 16216 6997 16217 6998 16218 6999 16219 7000 16220 7001 16221 7002 16222 7003 16223 7004 16224 7005 16225 7006 16226 7007 16227 7008 16228 7009 16229 7010 16230 7011 16231 8 16232 7012 16233 8 16234 7013 16235 8 16236 8 16237 7014 16238 7015 16239 7016 16240 7017 16241 8 16242 8 16243 7018 16244 7019 16245 7020 16246 7021 16247 7022 16248 8 16249 7023 16250 8 16251 7024 16252 8 16253 8 16254 8 16255 8 16256 7025 16257 8 16258 7026 16259 7027 16260 8 16261 7028 16262 7029 16263 7030 16264 7031 16265 7032 16266 8 16267 7033 16268 7034 16269 8 16270 7035 16271 8 16272 7036 16273 7037 16274 7038 16275 7039 16276 7040 16277 7041 16278 7042 16279 8 16280 8 16281 7043 16282 8 16283 8 16284 7044 16285 7045 16286 7046 16287 8 16288 8 16289 8 16290 8 16291 7047 16292 8 16293 8 16294 8 16295 8 16296 8 16297 7048 16298 7049 16299 7050 16300 7051 16301 7052 16302 7053 16303 7054 16304 7055 16305 7056 16306 7057 16307 7058 16308 7059 16309 7060 16310 7061 16311 7062 16312 7063 16313 7064 16314 7065 16315 7066 16316 7067 16317 7068 16318 7069 16319 7070 16320 7071 16321 7072 16322 7073 16323 7074 16324 7075 16325 7076 16326 7077 16327 7078 16328 7079 16329 7080 16330 7081 16331 7051 16332 7051 16333 7051 16334 7051 16335 7051 16336 7051 16337 7051 16338 7051 16339 7051 16340 7051 16341 7051 16342 7051 16343 7051 16344 7051 16345 7051 16346 7082 16347 7083 16348 8 16349 8 16350 8 16351 8 16352 8 16353 8 16354 7084 16355 8 16356 8 16357 7085 16358 8 16359 7086 16360 7087 16361 7088 16362 7089 16363 7090 16364 8 16365 8 16366 8 16367 8 16368 8 16369 7091 16370 7092 16371 8 16372 8 16373 7093 16374 8 16375 7094 16376 8 16377 8 16378 7095 16379 7096 16380 8 16381 7051 16382 7051 16383 7051 16384 7051 16385 7051 16386 7097 16387 7098 16388 7099 16389 7051 16390 7051 16391 7051 16392 7051 16393 7051 16394 7051 16395 7051 16396 7051 16397 7051 16398 7051 16399 7100 16400 8 16401 8 16402 8 16403 8 16404 8 16405 8 16406 8 16407 8 16408 8 16409 8 16410 8 16411 8 16412 8 16413 8 16414 8 16415 8 16416 8 16417 8 16418 8 16419 8 16420 8 16421 8 16422 8 16423 8 16424 8 16425 8 16426 8 16427 8 16428 8 16429 8 16430 8 16431 8 16432 8 16433 8 16434 8 16435 8 16436 8 16437 8 16438 8 16439 8 16440 8 16441 8 16442 8 16443 8 16444 8 16445 8 16446 8 16447 8 16448 8 16449 8 16450 8 16451 8 16452 8 16453 8 16454 8 16455 8 16456 8 16457 8 16458 8 16459 8 16460 8 16461 8 16462 8 16463 8 16464 8 16465 8 16466 8 16467 8 16468 8 16469 8 16470 8 16471 8 16472 8 16473 8 16474 8 16475 8 16476 8 16477 8 16478 8 16479 8 16480 7101 16481 8 16482 8 16483 8 16484 7102 16485 7103 16486 8 16487 7104 16488 8 16489 8 16490 7105 16491 7106 16492 7107 16493 8 16494 8 16495 7108 16496 7109 16497 7110 16498 7111 16499 8 16500 7112 16501 8 16502 7113 16503 7114 16504 7115 16505 7116 16506 7117 16507 7118 16508 7119 16509 7120 16510 7121 16511 7051 16512 7051 16513 7051 16514 7051 16515 7051 16516 7051 16517 7051 16518 7051 16519 7051 16520 7051 16521 7051 16522 7051 16523 7051 16524 7122 16525 7051 16526 7123 16527 7124 16528 8 16529 7125 16530 7126 16531 8 16532 7127 16533 8 16534 7128 16535 8 16536 8 16537 8 16538 7129 16539 7130 16540 7131 16541 8 16542 7132 16543 8 16544 8 16545 7133 16546 7134 16547 7135 16548 8 16549 8 16550 8 16551 7136 16552 7137 16553 8 16554 8 16555 8 16556 7138 16557 7139 16558 7140 16559 7141 16560 8 16561 8 16562 7142 16563 8 16564 7143 16565 7144 16566 8 16567 8 16568 7145 16569 8 16570 8 16571 8 16572 7146 16573 7147 16574 8 16575 8 16576 7148 16577 7149 16578 7150 16579 7151 16580 7152 16581 7153 16582 7154 16583 7155 16584 7156 16585 7157 16586 7158 16587 8 16588 7159 16589 7160 16590 7161 16591 7162 16592 7163 16593 7164 16594 7165 16595 7166 16596 7167 16597 7168 16598 8 16599 8 16600 7169 16601 8 16602 7170 16603 8 16604 7171 16605 7172 16606 8 16607 8 16608 8 16609 7173 16610 8 16611 8 16612 7174 16613 7175 16614 7176 16615 7177 16616 8 16617 8 16618 7178 16619 8 16620 7179 16621 7180 16622 8 16623 8 16624 7181 16625 8 16626 7182 16627 8 16628 7183 16629 7184 16630 7185 16631 7051 16632 7051 16633 7186 16634 7187 16635 7188 16636 7189 16637 7190 16638 7191 16639 7192 16640 7193 16641 7194 16642 7195 16643 8 16644 7196 16645 8 16646 7197 16647 8 16648 8 16649 7198 16650 7199 16651 7200 16652 7201 16653 8 16654 8 16655 7202 16656 7203 16657 7204 16658 7205 16659 7206 16660 8 16661 7207 16662 8 16663 7208 16664 8 16665 8 16666 8 16667 8 16668 7209 16669 8 16670 7210 16671 7211 16672 8 16673 7212 16674 7213 16675 7214 16676 7215 16677 7216 16678 8 16679 7217 16680 7218 16681 8 16682 7219 16683 8 16684 7220 16685 7221 16686 7222 16687 7223 16688 7224 16689 7225 16690 7226 16691 8 16692 8 16693 7227 16694 8 16695 8 16696 7228 16697 7229 16698 7230 16699 8 16700 8 16701 8 16702 8 16703 7231 16704 8 16705 8 16706 8 16707 8 16708 8 16709 7232 16710 7233 16711 7234 16712 7235 16713 7236 16714 7237 16715 7238 16716 7239 16717 7240 16718 7241 16719 7242 16720 7243 16721 7244 16722 7245 16723 7246 16724 7247 16725 7248 16726 7249 16727 7250 16728 7251 16729 7252 16730 7253 16731 7254 16732 7255 16733 7256 16734 7257 16735 7258 16736 7259 16737 7260 16738 7261 16739 7262 16740 7263 16741 7264 16742 7265 16743 7235 16744 7235 16745 7235 16746 7235 16747 7235 16748 7235 16749 7235 16750 7235 16751 7235 16752 7235 16753 7235 16754 7235 16755 7235 16756 7235 16757 7235 16758 7266 16759 7267 16760 8 16761 8 16762 8 16763 8 16764 8 16765 8 16766 7268 16767 8 16768 8 16769 7269 16770 8 16771 7270 16772 7271 16773 7272 16774 7273 16775 7274 16776 8 16777 8 16778 8 16779 8 16780 8 16781 7275 16782 7276 16783 8 16784 8 16785 7277 16786 8 16787 7278 16788 8 16789 8 16790 7279 16791 7280 16792 8 16793 7235 16794 7235 16795 7235 16796 7235 16797 7235 16798 7281 16799 7282 16800 7283 16801 7235 16802 7235 16803 7235 16804 7235 16805 7235 16806 7235 16807 7235 16808 7235 16809 7235 16810 7235 16811 7284 16812 8 16813 8 16814 8 16815 8 16816 8 16817 8 16818 8 16819 8 16820 8 16821 8 16822 8 16823 8 16824 8 16825 8 16826 8 16827 8 16828 8 16829 8 16830 8 16831 8 16832 8 16833 8 16834 8 16835 8 16836 8 16837 8 16838 8 16839 8 16840 8 16841 8 16842 8 16843 8 16844 8 16845 8 16846 8 16847 8 16848 8 16849 8 16850 8 16851 8 16852 8 16853 8 16854 8 16855 8 16856 8 16857 8 16858 8 16859 8 16860 8 16861 8 16862 8 16863 8 16864 8 16865 8 16866 8 16867 8 16868 8 16869 8 16870 8 16871 8 16872 8 16873 8 16874 8 16875 8 16876 8 16877 8 16878 8 16879 8 16880 8 16881 8 16882 8 16883 8 16884 8 16885 8 16886 8 16887 8 16888 8 16889 8 16890 8 16891 8 16892 7285 16893 8 16894 8 16895 8 16896 7286 16897 7287 16898 8 16899 7288 16900 8 16901 8 16902 7289 16903 7290 16904 7291 16905 8 16906 8 16907 7292 16908 7293 16909 7294 16910 7295 16911 8 16912 7296 16913 8 16914 7297 16915 7298 16916 7299 16917 7300 16918 7301 16919 7302 16920 7303 16921 7304 16922 7305 16923 7235 16924 7235 16925 7235 16926 7235 16927 7235 16928 7235 16929 7235 16930 7235 16931 7235 16932 7235 16933 7235 16934 7235 16935 7235 16936 7306 16937 7235 16938 7307 16939 7308 16940 8 16941 7309 16942 7310 16943 8 16944 7311 16945 8 16946 7312 16947 8 16948 8 16949 8 16950 7313 16951 7314 16952 7315 16953 8 16954 7316 16955 8 16956 8 16957 7317 16958 7318 16959 7319 16960 8 16961 8 16962 8 16963 7320 16964 7321 16965 8 16966 8 16967 8 16968 7322 16969 7323 16970 7324 16971 7325 16972 8 16973 8 16974 7326 16975 8 16976 7327 16977 7328 16978 8 16979 8 16980 7329 16981 8 16982 8 16983 8 16984 7330 16985 7331 16986 8 16987 8 16988 7332 16989 7333 16990 7334 16991 7335 16992 7336 16993 7337 16994 7338 16995 7339 16996 7340 16997 7341 16998 7342 16999 8 17000 7343 17001 7344 17002 7345 17003 7346 17004 7347 17005 7348 17006 7349 17007 7350 17008 7351 17009 7352 17010 8 17011 7353 17012 7051 17013 7354 17014 8 17015 7355 17016 8 17017 7356 17018 7357 17019 8 17020 8 17021 8 17022 7358 17023 8 17024 8 17025 7359 17026 7360 17027 7361 17028 7362 17029 8 17030 8 17031 7363 17032 8 17033 7364 17034 7365 17035 8 17036 8 17037 7366 17038 8 17039 7367 17040 8 17041 7368 17042 7369 17043 7370 17044 7235 17045 7235 17046 7371 17047 7372 17048 7235 17049 7235 17050 7235 17051 7235 17052 8 17053 7373 17054 8 17055 7374 17056 8 17057 8 17058 7235 17059 7235 17060 8 17061 7235 17062 7235 17063 7375 17064 8 17065 8 17066 8 17067 8 17068 8 17069 8 17070 7376 17071 7377 17072 7378 17073 7379 17074 8 17075 7380 17076 8 17077 7381 17078 7382 17079 7383 17080 7384 17081 7385 17082 8 17083 8 17084 8 17085 8 17086 8 17087 7386 17088 8 17089 8 17090 7387 17091 7388 17092 7389 17093 7390 17094 7391 17095 7392 17096 7393 17097 8 17098 7394 17099 8 17100 7395 17101 7396 17102 7397 17103 7398 17104 7399 17105 8 17106 7400 17107 7401 17108 7402 17109 7403 17110 7235 17111 7404 17112 7405 17113 7406 17114 8 17115 7235 17116 7235 17117 7235 17118 8 17119 7407 17120 8 17121 8 17122 8 17123 7408 17124 7409 17125 7410 17126 8 17127 8 17128 7411 17129 7412 17130 8 17131 7413 17132 7414 17133 8 17134 7235 17135 8 17136 7235 17137 7235 17138 7235 17139 7415 17140 7416 17141 8 17142 7235 17143 7235 17144 7235 17145 7417 17146 8 17147 8 17148 7418 17149 7419 17150 8 17151 7420 17152 8 17153 7235 17154 7051 17155 7051 17156 7051 17157 8 17158 7421 17159 8 17160 7422 17161 8 17162 8 17163 7051 17164 7051 17165 8 17166 7051 17167 7051 17168 7423 17169 8 17170 8 17171 8 17172 8 17173 8 17174 8 17175 7424 17176 7425 17177 7426 17178 7427 17179 8 17180 7428 17181 8 17182 7429 17183 7430 17184 7431 17185 7432 17186 7433 17187 8 17188 8 17189 8 17190 8 17191 8 17192 7434 17193 8 17194 8 17195 7435 17196 7436 17197 7437 17198 7438 17199 7439 17200 7440 17201 7441 17202 8 17203 7442 17204 8 17205 7443 17206 7444 17207 7445 17208 7446 17209 7447 17210 8 17211 7448 17212 7449 17213 7450 17214 7451 17215 7051 17216 7452 17217 7453 17218 7454 17219 8 17220 7051 17221 7051 17222 7051 17223 8 17224 7455 17225 8 17226 8 17227 8 17228 7456 17229 7457 17230 7458 17231 8 17232 8 17233 7459 17234 7460 17235 8 17236 7461 17237 7462 17238 8 17239 7051 17240 8 17241 7051 17242 7051 17243 7051 17244 7463 17245 7464 17246 8 17247 7051 17248 7051 17249 7051 17250 7465 17251 8 17252 8 17253 7466 17254 7467 17255 8 17256 7468 17257 8 17258 7051 17259 7469 17260 8 17261 7470 17262 7471 17263 7472 17264 7473 17265 7474 17266 7475 17267 7476 17268 7477 17269 7478 17270 8 17271 7479 17272 7480 17273 7481 17274 7482 17275 7483 17276 7484 17277 7485 17278 8 17279 7486 17280 7487 17281 8 17282 7488 17283 8 17284 7489 17285 7490 17286 8 17287 7491 17288 8 17289 8 17290 7492 17291 7493 17292 7494 17293 8 17294 7495 17295 7496 17296 7497 17297 7498 17298 7499 17299 8 17300 7500 17301 7501 17302 7502 17303 7503 17304 7504 17305 7505 17306 7506 17307 7507 17308 7508 17309 7509 17310 7510 17311 7511 17312 7512 17313 8 17314 7513 17315 7514 17316 7515 17317 7516 17318 7517 17319 7518 17320 7519 17321 7520 17322 7521 17323 8 17324 7522 17325 8 17326 7523 17327 8 17328 7524 17329 8 17330 7525 17331 7526 17332 8 17333 8 17334 7527 17335 8 17336 8 17337 8 17338 8 17339 7528 17340 7529 17341 8 17342 7530 17343 7531 17344 7532 17345 7533 17346 7534 17347 8 17348 8 17349 8 17350 8 17351 8 17352 7535 17353 8 17354 8 17355 7536 17356 7537 17357 7538 17358 7539 17359 7540 17360 7541 17361 7542 17362 7543 17363 7544 17364 7545 17365 8 17366 8 17367 7546 17368 8 17369 7547 17370 7548 17371 7549 17372 7550 17373 7551 17374 7552 17375 8 17376 7553 17377 8 17378 8 17379 8 17380 8 17381 7554 17382 7555 17383 8 17384 7556 17385 7557 17386 8 17387 7558 17388 7559 17389 7560 17390 7561 17391 7562 17392 7563 17393 7564 17394 7565 17395 7566 17396 7567 17397 7568 17398 7569 17399 7570 17400 8 17401 7571 17402 7572 17403 7573 17404 7574 17405 7575 17406 7576 17407 7577 17408 7578 17409 7579 17410 8 17411 7580 17412 8 17413 7581 17414 8 17415 7582 17416 8 17417 7583 17418 7584 17419 8 17420 8 17421 7585 17422 8 17423 8 17424 8 17425 8 17426 7586 17427 7587 17428 8 17429 7588 17430 8 17431 7589 17432 7590 17433 7591 17434 7592 17435 7593 17436 7594 17437 7595 17438 7596 17439 7597 17440 7598 17441 7599 17442 8 17443 8 17444 7600 17445 7601 17446 7602 17447 7603 17448 7604 17449 7605 17450 7606 17451 8 17452 8 17453 7607 17454 7608 17455 8 17456 7609 17457 7610 17458 7611 17459 8 17460 7612 17461 7613 17462 7614 17463 7615 17464 8 17465 8 17466 7616 17467 8 17468 7617 17469 7618 17470 7619 17471 7620 17472 7621 17473 7622 17474 7623 17475 7624 17476 7625 17477 7626 17478 7627 17479 7628 17480 7629 17481 7630 17482 8 17483 7631 17484 7632 17485 7633 17486 7634 17487 7635 17488 7636 17489 7637 17490 7638 17491 7639 17492 8 17493 7640 17494 8 17495 7641 17496 8 17497 7642 17498 8 17499 7643 17500 7644 17501 8 17502 8 17503 7645 17504 8 17505 8 17506 8 17507 8 17508 7646 17509 7647 17510 8 17511 7648 17512 7649 17513 7650 17514 7651 17515 7652 17516 8 17517 7653 17518 7654 17519 7655 17520 7656 17521 7657 17522 7658 17523 7659 17524 7660 17525 7661 17526 8 17527 7662 17528 7663 17529 7664 17530 7665 17531 7666 17532 7667 17533 7668 17534 8 17535 7669 17536 7670 17537 8 17538 7671 17539 8 17540 7672 17541 7673 17542 8 17543 7674 17544 8 17545 8 17546 7675 17547 7676 17548 8 17549 7677 17550 7678 17551 7679 17552 8 17553 7680 17554 7681 17555 7682 17556 7683 17557 8 17558 8 17559 7684 17560 7685 17561 8 17562 7686 17563 7687 17564 7688 17565 8 17566 7689 17567 7690 17568 8 17569 7691 17570 8 17571 7692 17572 8 17573 8 17574 8 17575 7693 17576 8 17577 7694 17578 7695 17579 7696 17580 8 17581 7697 17582 8 17583 8 17584 7698 17585 7699 17586 7700 17587 8 17588 8 17589 7701 17590 7702 17591 7703 17592 8 17593 8 17594 8 17595 7704 17596 7705 17597 8 17598 8 17599 8 17600 7706 17601 7707 17602 7708 17603 7709 17604 8 17605 8 17606 7710 17607 8 17608 7711 17609 7712 17610 8 17611 8 17612 7713 17613 8 17614 8 17615 8 17616 7714 17617 7715 17618 8 17619 7716 17620 7717 17621 7718 17622 7719 17623 7720 17624 7721 17625 8 17626 7722 17627 7723 17628 8 17629 7724 17630 8 17631 8 17632 8 17633 8 17634 8 17635 7725 17636 8 17637 8 17638 8 17639 7726 17640 7727 17641 8 17642 7728 17643 7729 17644 7730 17645 8 17646 7731 17647 7732 17648 7733 17649 7734 17650 8 17651 7735 17652 7736 17653 7737 17654 7738 17655 7739 17656 8 17657 8 17658 8 17659 7740 17660 7741 17661 8 17662 7742 17663 7743 17664 7744 17665 7745 17666 7746 17667 7747 17668 7748 17669 7749 17670 7750 17671 7751 17672 7752 17673 7753 17674 7754 17675 8 17676 7755 17677 7756 17678 7757 17679 7758 17680 7759 17681 7760 17682 7761 17683 7762 17684 7763 17685 8 17686 7764 17687 8 17688 7765 17689 8 17690 7766 17691 8 17692 7767 17693 7768 17694 8 17695 8 17696 7769 17697 8 17698 8 17699 8 17700 8 17701 7770 17702 7771 17703 8 17704 7772 17705 7773 17706 7774 17707 7775 17708 7776 17709 8 17710 7777 17711 7778 17712 7779 17713 7780 17714 7781 17715 7782 17716 7783 17717 7784 17718 7785 17719 8 17720 7786 17721 7787 17722 7788 17723 7789 17724 7790 17725 7791 17726 7792 17727 8 17728 7793 17729 7794 17730 8 17731 7795 17732 8 17733 7796 17734 7797 17735 8 17736 7798 17737 8 17738 8 17739 7799 17740 7800 17741 8 17742 7801 17743 7802 17744 7803 17745 8 17746 7804 17747 7805 17748 7806 17749 7807 17750 8 17751 8 17752 7808 17753 7809 17754 8 17755 7810 17756 7811 17757 7812 17758 7813 17759 8 17760 7814 17761 7815 17762 7816 17763 7817 17764 7818 17765 7819 17766 8 17767 7820 17768 7821 17769 8 17770 7822 17771 7823 17772 7824 17773 7825 17774 7826 17775 7827 17776 8 17777 7828 17778 7829 17779 7830 17780 7831 17781 8 17782 7832 17783 7833 17784 7834 17785 8 17786 7835 17787 8 17788 7836 17789 7837 17790 7838 17791 7839 17792 7840 17793 8 17794 8 17795 8 17796 7841 17797 7842 17798 8 17799 7843 17800 7844 17801 7845 17802 7846 17803 7847 17804 7848 17805 7849 17806 7850 17807 7851 17808 7852 17809 7853 17810 7854 17811 7855 17812 8 17813 7856 17814 7857 17815 7858 17816 7859 17817 7860 17818 7861 17819 7862 17820 7863 17821 7864 17822 8 17823 7865 17824 8 17825 7866 17826 8 17827 7867 17828 8 17829 7868 17830 7869 17831 8 17832 8 17833 7870 17834 8 17835 8 17836 8 17837 8 17838 7871 17839 7872 17840 8 17841 7873 17842 7874 17843 7875 17844 7876 17845 7877 17846 8 17847 7878 17848 7879 17849 7880 17850 7881 17851 7882 17852 7883 17853 7884 17854 7885 17855 7886 17856 8 17857 7887 17858 7888 17859 7889 17860 7890 17861 7891 17862 7892 17863 7893 17864 8 17865 7894 17866 7895 17867 8 17868 7896 17869 8 17870 7897 17871 7898 17872 8 17873 7899 17874 8 17875 8 17876 7900 17877 7901 17878 8 17879 7902 17880 7903 17881 7904 17882 8 17883 7905 17884 7906 17885 7907 17886 7908 17887 8 17888 8 17889 7909 17890 7910 17891 8 17892 7911 17893 7912 17894 7913 17895 7914 17896 8 17897 7915 17898 7916 17899 7917 17900 7918 17901 7919 17902 7920 17903 8 17904 7921 17905 7922 17906 8 17907 7923 17908 7924 17909 7925 17910 7926 17911 7927 17912 7928 17913 8 17914 7929 17915 7930 17916 7931 17917 7932 17918 8 17919 7933 17920 7934 17921 7935 17922 8 17923 7936 17924 8 17925 7937 17926 7938 17927 7939 17928 8 17929 7940 17930 7941 17931 8 17932 8 17933 7942 17934 8 17935 8 17936 7943 17937 7944 17938 8 17939 8 17940 7945 17941 7946 17942 7947 17943 7948 17944 7949 17945 7950 17946 7951 17947 7952 17948 7953 17949 7954 17950 7955 17951 8 17952 7956 17953 8 17954 7957 17955 8 17956 8 17957 7958 17958 7959 17959 7960 17960 7961 17961 8 17962 8 17963 7962 17964 7963 17965 7964 17966 7965 17967 7966 17968 8 17969 7967 17970 8 17971 7968 17972 8 17973 8 17974 8 17975 8 17976 7969 17977 8 17978 7970 17979 7971 17980 8 17981 7972 17982 7973 17983 7974 17984 7975 17985 7976 17986 8 17987 7977 17988 7978 17989 8 17990 7979 17991 8 17992 7980 17993 7981 17994 7982 17995 7983 17996 7984 17997 7985 17998 7986 17999 8 18000 8 18001 7987 18002 8 18003 8 18004 7988 18005 7989 18006 7990 18007 8 18008 8 18009 8 18010 8 18011 7991 18012 8 18013 8 18014 8 18015 8 18016 8 18017 7992 18018 7993 18019 7994 18020 7988 18021 7995 18022 7996 18023 7997 18024 7998 18025 7999 18026 8000 18027 8001 18028 8002 18029 8003 18030 8004 18031 8005 18032 8006 18033 8007 18034 8008 18035 8009 18036 8010 18037 8011 18038 8012 18039 8013 18040 8014 18041 8015 18042 8016 18043 8017 18044 8018 18045 8019 18046 8020 18047 8021 18048 8022 18049 8023 18050 8024 18051 7988 18052 7988 18053 7988 18054 7988 18055 7988 18056 7988 18057 7988 18058 7988 18059 7988 18060 7988 18061 7988 18062 7988 18063 7988 18064 7988 18065 7988 18066 8025 18067 8026 18068 8 18069 8 18070 8 18071 8 18072 8 18073 8 18074 8027 18075 8 18076 8 18077 8028 18078 8 18079 8029 18080 8030 18081 8031 18082 8032 18083 8033 18084 8 18085 8 18086 8 18087 8 18088 8 18089 8034 18090 8035 18091 8 18092 8 18093 8036 18094 8 18095 8037 18096 8 18097 8 18098 8038 18099 8039 18100 8 18101 7988 18102 7988 18103 7988 18104 7988 18105 7988 18106 8040 18107 8041 18108 8042 18109 7988 18110 7988 18111 7988 18112 7988 18113 7988 18114 7988 18115 7988 18116 7988 18117 7988 18118 7988 18119 8043 18120 8 18121 8 18122 8 18123 8 18124 8 18125 8 18126 8 18127 8 18128 8 18129 8 18130 8 18131 8 18132 8 18133 8 18134 8 18135 8 18136 8 18137 8 18138 8 18139 8 18140 8 18141 8 18142 8 18143 8 18144 8 18145 8 18146 8 18147 8 18148 8 18149 8 18150 8 18151 8 18152 8 18153 8 18154 8 18155 8 18156 8 18157 8 18158 8 18159 8 18160 8 18161 8 18162 8 18163 8 18164 8 18165 8 18166 8 18167 8 18168 8 18169 8 18170 8 18171 8 18172 8 18173 8 18174 8 18175 8 18176 8 18177 8 18178 8 18179 8 18180 8 18181 8 18182 8 18183 8 18184 8 18185 8 18186 8 18187 8 18188 8 18189 8 18190 8 18191 8 18192 8 18193 8 18194 8 18195 8 18196 8 18197 8 18198 8 18199 8 18200 8044 18201 8 18202 8 18203 8 18204 8045 18205 8046 18206 8 18207 8047 18208 8 18209 8 18210 8048 18211 8049 18212 8050 18213 8 18214 8 18215 8051 18216 8052 18217 8053 18218 8054 18219 8 18220 8055 18221 8 18222 8056 18223 8057 18224 8058 18225 8059 18226 8060 18227 8061 18228 8062 18229 8063 18230 8064 18231 7988 18232 7988 18233 7988 18234 7988 18235 7988 18236 7988 18237 7988 18238 7988 18239 7988 18240 7988 18241 7988 18242 7988 18243 7988 18244 8065 18245 7988 18246 8066 18247 8067 18248 8 18249 8068 18250 8069 18251 8 18252 8070 18253 8 18254 8071 18255 8 18256 8 18257 8 18258 8072 18259 8073 18260 8074 18261 8 18262 8075 18263 8 18264 8 18265 8076 18266 8077 18267 8078 18268 8 18269 8 18270 8 18271 8079 18272 8080 18273 8 18274 8 18275 8 18276 8081 18277 8082 18278 8083 18279 8084 18280 8 18281 8 18282 8085 18283 8 18284 8086 18285 8087 18286 8 18287 8 18288 8088 18289 8 18290 8 18291 8 18292 8089 18293 8090 18294 8 18295 8 18296 8091 18297 8092 18298 8093 18299 8094 18300 8095 18301 8096 18302 8097 18303 8098 18304 8099 18305 8100 18306 8101 18307 8 18308 8102 18309 8103 18310 8104 18311 8105 18312 8106 18313 8107 18314 8108 18315 8109 18316 8110 18317 8111 18318 8 18319 8112 18320 8 18321 8113 18322 8 18323 8114 18324 8115 18325 8 18326 8 18327 8 18328 8116 18329 8 18330 8 18331 8117 18332 8118 18333 8119 18334 8120 18335 8 18336 8 18337 8121 18338 8 18339 8122 18340 8123 18341 8 18342 8 18343 8124 18344 8 18345 8125 18346 8 18347 8126 18348 8127 18349 8128 18350 7988 18351 7988 18352 8129 18353 8130 18354 8131 18355 8132 18356 8133 18357 8134 18358 8135 18359 8136 18360 8137 18361 8138 18362 8 18363 8139 18364 8 18365 8140 18366 8 18367 8 18368 8141 18369 8142 18370 8143 18371 8144 18372 8 18373 8 18374 8145 18375 8146 18376 8147 18377 8148 18378 8149 18379 8 18380 8150 18381 8 18382 8151 18383 8 18384 8 18385 8 18386 8 18387 8152 18388 8 18389 8153 18390 8154 18391 8 18392 8155 18393 8156 18394 8157 18395 8158 18396 8159 18397 8 18398 8160 18399 8161 18400 8 18401 8162 18402 8 18403 8163 18404 8164 18405 8165 18406 8166 18407 8167 18408 8168 18409 8169 18410 8 18411 8 18412 8170 18413 8 18414 8 18415 8171 18416 8172 18417 8173 18418 8 18419 8 18420 8 18421 8 18422 8174 18423 8 18424 8 18425 8 18426 8 18427 8 18428 8175 18429 8176 18430 8177 18431 8178 18432 8179 18433 8180 18434 8181 18435 8182 18436 8183 18437 8184 18438 8185 18439 8186 18440 8187 18441 8188 18442 8189 18443 8190 18444 8191 18445 8192 18446 8193 18447 8194 18448 8195 18449 8196 18450 8197 18451 8198 18452 8199 18453 8200 18454 8201 18455 8202 18456 8203 18457 8204 18458 8205 18459 8206 18460 8207 18461 8208 18462 8178 18463 8178 18464 8178 18465 8178 18466 8178 18467 8178 18468 8178 18469 8178 18470 8178 18471 8178 18472 8178 18473 8178 18474 8178 18475 8178 18476 8178 18477 8209 18478 8210 18479 8 18480 8 18481 8 18482 8 18483 8 18484 8 18485 8211 18486 8 18487 8 18488 8212 18489 8 18490 8213 18491 8214 18492 8215 18493 8216 18494 8217 18495 8 18496 8 18497 8 18498 8 18499 8 18500 8218 18501 8219 18502 8 18503 8 18504 8220 18505 8 18506 8221 18507 8 18508 8 18509 8222 18510 8223 18511 8 18512 8178 18513 8178 18514 8178 18515 8178 18516 8178 18517 8224 18518 8225 18519 8226 18520 8178 18521 8178 18522 8178 18523 8178 18524 8178 18525 8178 18526 8178 18527 8178 18528 8178 18529 8178 18530 8227 18531 8 18532 8 18533 8 18534 8 18535 8 18536 8 18537 8 18538 8 18539 8 18540 8 18541 8 18542 8 18543 8 18544 8 18545 8 18546 8 18547 8 18548 8 18549 8 18550 8 18551 8 18552 8 18553 8 18554 8 18555 8 18556 8 18557 8 18558 8 18559 8 18560 8 18561 8 18562 8 18563 8 18564 8 18565 8 18566 8 18567 8 18568 8 18569 8 18570 8 18571 8 18572 8 18573 8 18574 8 18575 8 18576 8 18577 8 18578 8 18579 8 18580 8 18581 8 18582 8 18583 8 18584 8 18585 8 18586 8 18587 8 18588 8 18589 8 18590 8 18591 8 18592 8 18593 8 18594 8 18595 8 18596 8 18597 8 18598 8 18599 8 18600 8 18601 8 18602 8 18603 8 18604 8 18605 8 18606 8 18607 8 18608 8 18609 8 18610 8 18611 8228 18612 8 18613 8 18614 8 18615 8229 18616 8230 18617 8 18618 8231 18619 8 18620 8 18621 8232 18622 8233 18623 8234 18624 8 18625 8 18626 8235 18627 8236 18628 8237 18629 8238 18630 8 18631 8239 18632 8 18633 8240 18634 8241 18635 8242 18636 8243 18637 8244 18638 8245 18639 8246 18640 8247 18641 8248 18642 8178 18643 8178 18644 8178 18645 8178 18646 8178 18647 8178 18648 8178 18649 8178 18650 8178 18651 8178 18652 8178 18653 8178 18654 8178 18655 8249 18656 8178 18657 8250 18658 8251 18659 8 18660 8252 18661 8253 18662 8 18663 8254 18664 8 18665 8255 18666 8 18667 8 18668 8 18669 8256 18670 8257 18671 8258 18672 8 18673 8259 18674 8 18675 8 18676 8260 18677 8261 18678 8262 18679 8 18680 8 18681 8 18682 8263 18683 8264 18684 8 18685 8 18686 8 18687 8265 18688 8266 18689 8267 18690 8268 18691 8 18692 8 18693 8269 18694 8 18695 8270 18696 8271 18697 8 18698 8 18699 8272 18700 8 18701 8 18702 8 18703 8273 18704 8274 18705 8 18706 8 18707 8275 18708 8276 18709 8277 18710 8278 18711 8279 18712 8280 18713 8281 18714 8282 18715 8283 18716 8284 18717 8285 18718 8 18719 8286 18720 8287 18721 8288 18722 8289 18723 8290 18724 8291 18725 8292 18726 8293 18727 8294 18728 8295 18729 8 18730 8296 18731 7988 18732 8297 18733 8 18734 8298 18735 8 18736 8299 18737 8300 18738 8 18739 8 18740 8 18741 8301 18742 8 18743 8 18744 8302 18745 8303 18746 8304 18747 8305 18748 8 18749 8 18750 8306 18751 8 18752 8307 18753 8308 18754 8 18755 8 18756 8309 18757 8 18758 8310 18759 8 18760 8311 18761 8312 18762 8313 18763 8178 18764 8178 18765 8314 18766 8315 18767 8178 18768 8178 18769 8178 18770 8178 18771 8 18772 8316 18773 8 18774 8317 18775 8 18776 8 18777 8178 18778 8178 18779 8 18780 8178 18781 8178 18782 8318 18783 8 18784 8 18785 8 18786 8 18787 8 18788 8 18789 8319 18790 8320 18791 8321 18792 8322 18793 8 18794 8323 18795 8 18796 8324 18797 8325 18798 8326 18799 8327 18800 8328 18801 8 18802 8 18803 8 18804 8 18805 8 18806 8329 18807 8 18808 8 18809 8330 18810 8331 18811 8332 18812 8333 18813 8334 18814 8335 18815 8336 18816 8 18817 8337 18818 8 18819 8338 18820 8339 18821 8340 18822 8341 18823 8342 18824 8 18825 8343 18826 8344 18827 8345 18828 8346 18829 8178 18830 8347 18831 8348 18832 8349 18833 8 18834 8178 18835 8178 18836 8178 18837 8 18838 8350 18839 8 18840 8 18841 8 18842 8351 18843 8352 18844 8353 18845 8 18846 8 18847 8354 18848 8355 18849 8 18850 8356 18851 8357 18852 8 18853 8178 18854 8 18855 8178 18856 8178 18857 8178 18858 8358 18859 8359 18860 8 18861 8178 18862 8178 18863 8178 18864 8360 18865 8 18866 8 18867 8361 18868 8362 18869 8 18870 8363 18871 8 18872 8178 18873 7988 18874 7988 18875 7988 18876 8 18877 8364 18878 8 18879 8365 18880 8 18881 8 18882 7988 18883 7988 18884 8 18885 7988 18886 7988 18887 8366 18888 8 18889 8 18890 8 18891 8 18892 8 18893 8 18894 8367 18895 8368 18896 8369 18897 8370 18898 8 18899 8371 18900 8 18901 8372 18902 8373 18903 8374 18904 8375 18905 8376 18906 8 18907 8 18908 8 18909 8 18910 8 18911 8377 18912 8 18913 8 18914 8378 18915 8379 18916 8380 18917 8381 18918 8382 18919 8383 18920 8384 18921 8 18922 8385 18923 8 18924 8386 18925 8387 18926 8388 18927 8389 18928 8390 18929 8 18930 8391 18931 8392 18932 8393 18933 8394 18934 7988 18935 8395 18936 8396 18937 8397 18938 8 18939 7988 18940 7988 18941 7988 18942 8 18943 8398 18944 8 18945 8 18946 8 18947 8399 18948 8400 18949 8401 18950 8 18951 8 18952 8402 18953 8403 18954 8 18955 8404 18956 8405 18957 8 18958 7988 18959 8 18960 7988 18961 7988 18962 7988 18963 8406 18964 8407 18965 8 18966 7988 18967 7988 18968 7988 18969 8408 18970 8 18971 8 18972 8409 18973 8410 18974 8 18975 8411 18976 8 18977 7988 18978 84 18979 8412 18980 8 18981 8413 18982 8 18983 8414 18984 8 18985 8 18986 8415 18987 8 18988 8416 18989 8 18990 8 18991 8 18992 8417 18993 8418 18994 8 18995 8419 18996 8420 18997 8421 18998 8422 18999 8423 19000 8424 19001 8425 19002 8426 19003 8427 19004 8428 19005 8429 19006 8 19007 8430 19008 8 19009 8431 19010 8 19011 8 19012 8432 19013 8433 19014 8434 19015 8435 19016 8 19017 8 19018 8436 19019 8437 19020 8438 19021 8439 19022 8440 19023 8 19024 8441 19025 8 19026 8442 19027 8 19028 8 19029 8 19030 8 19031 8443 19032 8 19033 8444 19034 8445 19035 8 19036 8446 19037 8447 19038 8448 19039 8449 19040 8450 19041 8 19042 8451 19043 8452 19044 8 19045 8453 19046 8 19047 8454 19048 8455 19049 8456 19050 8457 19051 8458 19052 8459 19053 8460 19054 8 19055 8 19056 8461 19057 8 19058 8 19059 8462 19060 8463 19061 8464 19062 8 19063 8 19064 8 19065 8 19066 8465 19067 8 19068 8 19069 8 19070 8 19071 8 19072 8466 19073 8467 19074 8468 19075 8469 19076 8470 19077 8471 19078 8472 19079 8473 19080 8474 19081 8475 19082 8476 19083 8477 19084 8478 19085 8479 19086 8480 19087 8481 19088 8482 19089 8483 19090 8484 19091 8485 19092 8486 19093 8487 19094 8488 19095 8489 19096 8490 19097 8491 19098 8492 19099 8493 19100 8494 19101 8495 19102 8496 19103 8497 19104 8498 19105 8499 19106 8469 19107 8469 19108 8469 19109 8469 19110 8469 19111 8469 19112 8469 19113 8469 19114 8469 19115 8469 19116 8469 19117 8469 19118 8469 19119 8469 19120 8469 19121 8500 19122 8501 19123 8 19124 8 19125 8 19126 8 19127 8 19128 8 19129 8502 19130 8 19131 8 19132 8503 19133 8 19134 8504 19135 8505 19136 8506 19137 8507 19138 8508 19139 8 19140 8 19141 8 19142 8 19143 8 19144 8509 19145 8510 19146 8 19147 8 19148 8511 19149 8 19150 8512 19151 8 19152 8 19153 8513 19154 8514 19155 8 19156 8469 19157 8469 19158 8469 19159 8469 19160 8469 19161 8515 19162 8516 19163 8517 19164 8469 19165 8469 19166 8469 19167 8469 19168 8469 19169 8469 19170 8469 19171 8469 19172 8469 19173 8469 19174 8518 19175 8 19176 8 19177 8 19178 8 19179 8 19180 8 19181 8 19182 8 19183 8 19184 8 19185 8 19186 8 19187 8 19188 8 19189 8 19190 8 19191 8 19192 8 19193 8 19194 8 19195 8 19196 8 19197 8 19198 8 19199 8 19200 8 19201 8 19202 8 19203 8 19204 8 19205 8 19206 8 19207 8 19208 8 19209 8 19210 8 19211 8 19212 8 19213 8 19214 8 19215 8 19216 8 19217 8 19218 8 19219 8 19220 8 19221 8 19222 8 19223 8 19224 8 19225 8 19226 8 19227 8 19228 8 19229 8 19230 8 19231 8 19232 8 19233 8 19234 8 19235 8 19236 8 19237 8 19238 8 19239 8 19240 8 19241 8 19242 8 19243 8 19244 8 19245 8 19246 8 19247 8 19248 8 19249 8 19250 8 19251 8 19252 8 19253 8 19254 8 19255 8519 19256 8 19257 8 19258 8 19259 8520 19260 8521 19261 8 19262 8522 19263 8 19264 8 19265 8523 19266 8524 19267 8525 19268 8 19269 8 19270 8526 19271 8527 19272 8528 19273 8529 19274 8 19275 8530 19276 8 19277 8531 19278 8532 19279 8533 19280 8534 19281 8535 19282 8536 19283 8537 19284 8538 19285 8539 19286 8469 19287 8469 19288 8469 19289 8469 19290 8469 19291 8469 19292 8469 19293 8469 19294 8469 19295 8469 19296 8469 19297 8469 19298 8469 19299 8540 19300 8469 19301 8541 19302 8542 19303 8 19304 8543 19305 8544 19306 8 19307 8545 19308 8 19309 8546 19310 8 19311 8 19312 8 19313 8547 19314 8548 19315 8549 19316 8 19317 8550 19318 8 19319 8 19320 8551 19321 8552 19322 8553 19323 8 19324 8 19325 8 19326 8554 19327 8555 19328 8 19329 8 19330 8 19331 8556 19332 8557 19333 8558 19334 8559 19335 8 19336 8 19337 8560 19338 8 19339 8561 19340 8562 19341 8 19342 8 19343 8563 19344 8 19345 8 19346 8 19347 8564 19348 8565 19349 8 19350 8 19351 8566 19352 8567 19353 8568 19354 8569 19355 8570 19356 8571 19357 8572 19358 8573 19359 8574 19360 8575 19361 8576 19362 8 19363 8577 19364 8578 19365 8579 19366 8580 19367 8581 19368 8582 19369 8583 19370 8584 19371 8585 19372 8586 19373 8587 19374 8588 19375 8589 19376 8590 19377 8591 19378 8592 19379 8593 19380 8594 19381 8595 19382 8 19383 8596 19384 8 19385 8597 19386 8 19387 8 19388 8598 19389 8599 19390 8600 19391 8601 19392 8 19393 8 19394 8602 19395 8603 19396 8604 19397 8605 19398 8606 19399 8 19400 8607 19401 8 19402 8608 19403 8 19404 8 19405 8 19406 8 19407 8609 19408 8 19409 8610 19410 8611 19411 8 19412 8612 19413 8613 19414 8614 19415 8615 19416 8616 19417 8 19418 8617 19419 8618 19420 8 19421 8619 19422 8 19423 8620 19424 8621 19425 8622 19426 8623 19427 8624 19428 8625 19429 8626 19430 8 19431 8 19432 8627 19433 8 19434 8 19435 8628 19436 8629 19437 8630 19438 8 19439 8 19440 8 19441 8 19442 8631 19443 8 19444 8 19445 8 19446 8 19447 8 19448 8632 19449 8633 19450 8634 19451 8635 19452 8636 19453 8637 19454 8638 19455 8639 19456 8640 19457 8641 19458 8642 19459 8643 19460 8644 19461 8645 19462 8646 19463 8647 19464 8648 19465 8649 19466 8650 19467 8651 19468 8652 19469 8653 19470 8654 19471 8655 19472 8656 19473 8657 19474 8658 19475 8659 19476 8660 19477 8661 19478 8662 19479 8663 19480 8664 19481 8665 19482 8635 19483 8635 19484 8635 19485 8635 19486 8635 19487 8635 19488 8635 19489 8635 19490 8635 19491 8635 19492 8635 19493 8635 19494 8635 19495 8635 19496 8635 19497 8666 19498 8667 19499 8 19500 8 19501 8 19502 8 19503 8 19504 8 19505 8668 19506 8 19507 8 19508 8669 19509 8 19510 8670 19511 8671 19512 8672 19513 8673 19514 8674 19515 8 19516 8 19517 8 19518 8 19519 8 19520 8675 19521 8676 19522 8 19523 8 19524 8677 19525 8 19526 8678 19527 8 19528 8 19529 8679 19530 8680 19531 8 19532 8635 19533 8635 19534 8635 19535 8635 19536 8635 19537 8681 19538 8682 19539 8683 19540 8635 19541 8635 19542 8635 19543 8635 19544 8635 19545 8635 19546 8635 19547 8635 19548 8635 19549 8635 19550 8684 19551 8 19552 8 19553 8 19554 8 19555 8 19556 8 19557 8 19558 8 19559 8 19560 8 19561 8 19562 8 19563 8 19564 8 19565 8 19566 8 19567 8 19568 8 19569 8 19570 8 19571 8 19572 8 19573 8 19574 8 19575 8 19576 8 19577 8 19578 8 19579 8 19580 8 19581 8 19582 8 19583 8 19584 8 19585 8 19586 8 19587 8 19588 8 19589 8 19590 8 19591 8 19592 8 19593 8 19594 8 19595 8 19596 8 19597 8 19598 8 19599 8 19600 8 19601 8 19602 8 19603 8 19604 8 19605 8 19606 8 19607 8 19608 8 19609 8 19610 8 19611 8 19612 8 19613 8 19614 8 19615 8 19616 8 19617 8 19618 8 19619 8 19620 8 19621 8 19622 8 19623 8 19624 8 19625 8 19626 8 19627 8 19628 8 19629 8 19630 8 19631 8685 19632 8 19633 8 19634 8 19635 8686 19636 8687 19637 8 19638 8688 19639 8 19640 8 19641 8689 19642 8690 19643 8691 19644 8 19645 8 19646 8692 19647 8693 19648 8694 19649 8695 19650 8 19651 8696 19652 8 19653 8697 19654 8698 19655 8699 19656 8700 19657 8701 19658 8702 19659 8703 19660 8704 19661 8705 19662 8635 19663 8635 19664 8635 19665 8635 19666 8635 19667 8635 19668 8635 19669 8635 19670 8635 19671 8635 19672 8635 19673 8635 19674 8635 19675 8706 19676 8635 19677 8707 19678 8708 19679 8 19680 8709 19681 8710 19682 8 19683 8711 19684 8 19685 8712 19686 8 19687 8 19688 8 19689 8713 19690 8714 19691 8715 19692 8 19693 8716 19694 8 19695 8 19696 8717 19697 8718 19698 8719 19699 8 19700 8 19701 8 19702 8720 19703 8721 19704 8 19705 8 19706 8 19707 8722 19708 8723 19709 8724 19710 8725 19711 8 19712 8 19713 8726 19714 8 19715 8727 19716 8728 19717 8 19718 8 19719 8729 19720 8 19721 8 19722 8 19723 8730 19724 8731 19725 8 19726 8 19727 8732 19728 8733 19729 8734 19730 8735 19731 8736 19732 8737 19733 8738 19734 8739 19735 8740 19736 8741 19737 8742 19738 8 19739 8743 19740 8744 19741 8745 19742 8746 19743 8747 19744 8748 19745 8749 19746 8750 19747 8751 19748 8752 19749 8 19750 8 19751 8753 19752 8 19753 8754 19754 8 19755 8755 19756 8756 19757 8 19758 8 19759 8 19760 8757 19761 8 19762 8 19763 8758 19764 8759 19765 8760 19766 8761 19767 8 19768 8 19769 8762 19770 8 19771 8763 19772 8764 19773 8 19774 8 19775 8765 19776 8 19777 8766 19778 8 19779 8767 19780 8768 19781 8769 19782 8635 19783 8635 19784 8770 19785 8771 19786 8772 19787 8773 19788 8774 19789 8775 19790 8776 19791 8777 19792 8778 19793 8779 19794 8 19795 8780 19796 8 19797 8781 19798 8 19799 8 19800 8782 19801 8783 19802 8784 19803 8785 19804 8 19805 8 19806 8786 19807 8787 19808 8788 19809 8789 19810 8790 19811 8 19812 8791 19813 8 19814 8792 19815 8 19816 8 19817 8 19818 8 19819 8793 19820 8 19821 8794 19822 8795 19823 8 19824 8796 19825 8797 19826 8798 19827 8799 19828 8800 19829 8 19830 8801 19831 8802 19832 8 19833 8803 19834 8 19835 8804 19836 8805 19837 8806 19838 8807 19839 8808 19840 8809 19841 8810 19842 8 19843 8 19844 8811 19845 8 19846 8 19847 8812 19848 8813 19849 8814 19850 8 19851 8 19852 8 19853 8 19854 8815 19855 8 19856 8 19857 8 19858 8 19859 8 19860 8816 19861 8817 19862 8818 19863 8819 19864 8820 19865 8821 19866 8822 19867 8823 19868 8824 19869 8825 19870 8826 19871 8827 19872 8828 19873 8829 19874 8830 19875 8831 19876 8832 19877 8833 19878 8834 19879 8835 19880 8836 19881 8837 19882 8838 19883 8839 19884 8840 19885 8841 19886 8842 19887 8843 19888 8844 19889 8845 19890 8846 19891 8847 19892 8848 19893 8849 19894 8819 19895 8819 19896 8819 19897 8819 19898 8819 19899 8819 19900 8819 19901 8819 19902 8819 19903 8819 19904 8819 19905 8819 19906 8819 19907 8819 19908 8819 19909 8850 19910 8851 19911 8 19912 8 19913 8 19914 8 19915 8 19916 8 19917 8852 19918 8 19919 8 19920 8853 19921 8 19922 8854 19923 8855 19924 8856 19925 8857 19926 8858 19927 8 19928 8 19929 8 19930 8 19931 8 19932 8859 19933 8860 19934 8 19935 8 19936 8861 19937 8 19938 8862 19939 8 19940 8 19941 8863 19942 8864 19943 8 19944 8819 19945 8819 19946 8819 19947 8819 19948 8819 19949 8865 19950 8866 19951 8867 19952 8819 19953 8819 19954 8819 19955 8819 19956 8819 19957 8819 19958 8819 19959 8819 19960 8819 19961 8819 19962 8868 19963 8 19964 8 19965 8 19966 8 19967 8 19968 8 19969 8 19970 8 19971 8 19972 8 19973 8 19974 8 19975 8 19976 8 19977 8 19978 8 19979 8 19980 8 19981 8 19982 8 19983 8 19984 8 19985 8 19986 8 19987 8 19988 8 19989 8 19990 8 19991 8 19992 8 19993 8 19994 8 19995 8 19996 8 19997 8 19998 8 19999 8 20000 8 20001 8 20002 8 20003 8 20004 8 20005 8 20006 8 20007 8 20008 8 20009 8 20010 8 20011 8 20012 8 20013 8 20014 8 20015 8 20016 8 20017 8 20018 8 20019 8 20020 8 20021 8 20022 8 20023 8 20024 8 20025 8 20026 8 20027 8 20028 8 20029 8 20030 8 20031 8 20032 8 20033 8 20034 8 20035 8 20036 8 20037 8 20038 8 20039 8 20040 8 20041 8 20042 8 20043 8869 20044 8 20045 8 20046 8 20047 8870 20048 8871 20049 8 20050 8872 20051 8 20052 8 20053 8873 20054 8874 20055 8875 20056 8 20057 8 20058 8876 20059 8877 20060 8878 20061 8879 20062 8 20063 8880 20064 8 20065 8881 20066 8882 20067 8883 20068 8884 20069 8885 20070 8886 20071 8887 20072 8888 20073 8889 20074 8819 20075 8819 20076 8819 20077 8819 20078 8819 20079 8819 20080 8819 20081 8819 20082 8819 20083 8819 20084 8819 20085 8819 20086 8819 20087 8890 20088 8819 20089 8891 20090 8892 20091 8 20092 8893 20093 8894 20094 8 20095 8895 20096 8 20097 8896 20098 8 20099 8 20100 8 20101 8897 20102 8898 20103 8899 20104 8 20105 8900 20106 8 20107 8 20108 8901 20109 8902 20110 8903 20111 8 20112 8 20113 8 20114 8904 20115 8905 20116 8 20117 8 20118 8 20119 8906 20120 8907 20121 8908 20122 8909 20123 8 20124 8 20125 8910 20126 8 20127 8911 20128 8912 20129 8 20130 8 20131 8913 20132 8 20133 8 20134 8 20135 8914 20136 8915 20137 8 20138 8 20139 8916 20140 8917 20141 8918 20142 8919 20143 8920 20144 8921 20145 8922 20146 8923 20147 8924 20148 8925 20149 8926 20150 8 20151 8927 20152 8928 20153 8929 20154 8930 20155 8931 20156 8932 20157 8933 20158 8934 20159 8935 20160 8936 20161 8 20162 8937 20163 8635 20164 8938 20165 8 20166 8939 20167 8 20168 8940 20169 8941 20170 8 20171 8 20172 8 20173 8942 20174 8 20175 8 20176 8943 20177 8944 20178 8945 20179 8946 20180 8 20181 8 20182 8947 20183 8 20184 8948 20185 8949 20186 8 20187 8 20188 8950 20189 8 20190 8951 20191 8 20192 8952 20193 8953 20194 8954 20195 8819 20196 8819 20197 8955 20198 8956 20199 8819 20200 8819 20201 8819 20202 8819 20203 8 20204 8957 20205 8 20206 8958 20207 8 20208 8 20209 8819 20210 8819 20211 8 20212 8819 20213 8819 20214 8959 20215 8 20216 8 20217 8 20218 8 20219 8 20220 8 20221 8960 20222 8961 20223 8962 20224 8963 20225 8 20226 8964 20227 8 20228 8965 20229 8966 20230 8967 20231 8968 20232 8969 20233 8 20234 8 20235 8 20236 8 20237 8 20238 8970 20239 8 20240 8 20241 8971 20242 8972 20243 8973 20244 8974 20245 8975 20246 8976 20247 8977 20248 8 20249 8978 20250 8 20251 8979 20252 8980 20253 8981 20254 8982 20255 8983 20256 8 20257 8984 20258 8985 20259 8986 20260 8987 20261 8819 20262 8988 20263 8989 20264 8990 20265 8 20266 8819 20267 8819 20268 8819 20269 8 20270 8991 20271 8 20272 8 20273 8 20274 8992 20275 8993 20276 8994 20277 8 20278 8 20279 8995 20280 8996 20281 8 20282 8997 20283 8998 20284 8 20285 8819 20286 8 20287 8819 20288 8819 20289 8819 20290 8999 20291 9000 20292 8 20293 8819 20294 8819 20295 8819 20296 9001 20297 8 20298 8 20299 9002 20300 9003 20301 8 20302 9004 20303 8 20304 8819 20305 8635 20306 8635 20307 8635 20308 8 20309 9005 20310 8 20311 9006 20312 8 20313 8 20314 8635 20315 8635 20316 8 20317 8635 20318 8635 20319 9007 20320 8 20321 8 20322 8 20323 8 20324 8 20325 8 20326 9008 20327 9009 20328 9010 20329 9011 20330 8 20331 9012 20332 8 20333 9013 20334 9014 20335 9015 20336 9016 20337 9017 20338 8 20339 8 20340 8 20341 8 20342 8 20343 9018 20344 8 20345 8 20346 9019 20347 9020 20348 9021 20349 9022 20350 9023 20351 9024 20352 9025 20353 8 20354 9026 20355 8 20356 9027 20357 9028 20358 9029 20359 9030 20360 9031 20361 8 20362 9032 20363 9033 20364 9034 20365 9035 20366 8635 20367 9036 20368 9037 20369 9038 20370 8 20371 8635 20372 8635 20373 8635 20374 8 20375 9039 20376 8 20377 8 20378 8 20379 9040 20380 9041 20381 9042 20382 8 20383 8 20384 9043 20385 9044 20386 8 20387 9045 20388 9046 20389 8 20390 8635 20391 8 20392 8635 20393 8635 20394 8635 20395 9047 20396 9048 20397 8 20398 8635 20399 8635 20400 8635 20401 9049 20402 8 20403 8 20404 9050 20405 9051 20406 8 20407 9052 20408 8 20409 8635 20410 9053 20411 9054 20412 9055 20413 9056 20414 8 20415 9057 20416 9058 20417 9059 20418 9060 20419 8 20420 9061 20421 9062 20422 8 20423 9063 20424 8 20425 8 20426 8 20427 9064 20428 9065 20429 9066 20430 8 20431 9067 20432 8 20433 8 20434 9068 20435 9069 20436 9070 20437 8 20438 8 20439 8 20440 9071 20441 9072 20442 9073 20443 9074 20444 9075 20445 8 20446 9076 20447 9077 20448 9078 20449 9079 20450 8 20451 9080 20452 9081 20453 8 20454 9082 20455 8 20456 8 20457 8 20458 9083 20459 9084 20460 9085 20461 8 20462 9086 20463 8 20464 8 20465 9087 20466 9088 20467 9089 20468 8 20469 8 20470 8 20471 9090 20472 9091 20473 8 20474 8 20475 9092 20476 8 20477 9093 20478 9094 20479 9095 20480 9096 20481 9097 20482 9098 20483 9099 20484 9100 20485 9101 20486 9102 20487 9103 20488 9104 20489 9105 20490 9106 20491 8 20492 9107 20493 9108 20494 9109 20495 9110 20496 9111 20497 9112 20498 9113 20499 9114 20500 9115 20501 8 20502 9116 20503 8 20504 9117 20505 8 20506 9118 20507 8 20508 9119 20509 9120 20510 8 20511 8 20512 9121 20513 8 20514 8 20515 8 20516 8 20517 9122 20518 9123 20519 8 20520 9124 20521 9125 20522 9126 20523 9127 20524 9128 20525 8 20526 9129 20527 9130 20528 9131 20529 9132 20530 9133 20531 9134 20532 9135 20533 9136 20534 9137 20535 8 20536 9138 20537 9139 20538 9140 20539 9141 20540 9142 20541 9143 20542 9144 20543 8 20544 9145 20545 9146 20546 8 20547 9147 20548 8 20549 9148 20550 9149 20551 8 20552 9150 20553 8 20554 8 20555 9151 20556 9152 20557 8 20558 9153 20559 9154 20560 9155 20561 8 20562 9156 20563 9157 20564 9158 20565 9159 20566 8 20567 8 20568 9160 20569 9161 20570 8 20571 9162 20572 9163 20573 9164 20574 9165 20575 8 20576 9166 20577 9167 20578 9168 20579 9169 20580 9170 20581 9171 20582 8 20583 9172 20584 9173 20585 8 20586 9174 20587 9175 20588 9176 20589 9177 20590 9178 20591 9179 20592 8 20593 9180 20594 9181 20595 9182 20596 9183 20597 8 20598 9184 20599 9185 20600 9186 20601 8 20602 9187 20603 8 20604 9188 20605 9189 20606 9190 20607 8 20608 9191 20609 9192 20610 8 20611 8 20612 9193 20613 8 20614 9194 20615 8 20616 9195 20617 8 20618 9196 20619 8 20620 9197 20621 9198 20622 9199 20623 9200 20624 9201 20625 9202 20626 9203 20627 9204 20628 9205 20629 8 20630 8 20631 9206 20632 9207 20633 8 20634 8 20635 9208 20636 7988 20637 9209 20638 9210 20639 9211 20640 7988 20641 9212 20642 9213 20643 9214 20644 1111 20645 9215 20646 9216 20647 9217 20648 1111 20649 9218 20650 9219 20651 8 20652 8 20653 9220 20654 8 20655 9221 20656 9222 20657 9223 20658 9224 20659 9225 20660 9226 20661 9227 20662 9228 20663 9229 20664 9230 20665 9231 20666 9232 20667 9233 20668 9234 20669 8 20670 9235 20671 9236 20672 9237 20673 9238 20674 9239 20675 9240 20676 9241 20677 9242 20678 9243 20679 8 20680 9244 20681 8 20682 9245 20683 8 20684 9246 20685 8 20686 9247 20687 9248 20688 8 20689 8 20690 9249 20691 8 20692 8 20693 8 20694 8 20695 9250 20696 9251 20697 8 20698 9252 20699 9253 20700 9254 20701 9255 20702 9256 20703 8 20704 9257 20705 9258 20706 9259 20707 9260 20708 9261 20709 9262 20710 9263 20711 9264 20712 9265 20713 8 20714 9266 20715 9267 20716 9268 20717 9269 20718 9270 20719 9271 20720 9272 20721 8 20722 9273 20723 9274 20724 8 20725 9275 20726 8 20727 9276 20728 9277 20729 8 20730 9278 20731 8 20732 8 20733 9279 20734 9280 20735 8 20736 9281 20737 9282 20738 9283 20739 8 20740 9284 20741 9285 20742 9286 20743 9287 20744 8 20745 8 20746 9288 20747 9289 20748 8 20749 9290 20750 9291 20751 9292 20752 9293 20753 8 20754 9294 20755 9295 20756 9296 20757 9297 20758 9298 20759 9299 20760 8 20761 9300 20762 9301 20763 8 20764 9302 20765 9303 20766 9304 20767 9305 20768 9306 20769 9307 20770 8 20771 9308 20772 9309 20773 9310 20774 9311 20775 8 20776 9312 20777 9313 20778 9314 20779 8 20780 9315 20781 8 20782 9316 20783 9317 20784 9318 20785 8 20786 9319 20787 9320 20788 8 20789 8 20790 9321 20791 8 20792 9322 20793 8 20794 9323 20795 8 20796 9324 20797 8 20798 9325 20799 9326 20800 9327 20801 9328 20802 9329 20803 9330 20804 9331 20805 9332 20806 9333 20807 8 20808 8 20809 9334 20810 9335 20811 8 20812 8 20813 9336 20814 9337 20815 9338 20816 9339 20817 9340 20818 9341 20819 9342 20820 9343 20821 8 20822 8 20823 9344 20824 9345 20825 8 20826 8 20827 9346 20828 7988 20829 9347 20830 8 20831 8 20832 8 20833 9348 20834 9349 20835 9350 20836 9351 20837 9352 20838 9353 20839 9354 20840 9355 20841 9356 20842 9357 20843 8 20844 9358 20845 9359 20846 9360 20847 9361 20848 9362 20849 9363 20850 9364 20851 8 20852 9365 20853 9366 20854 8 20855 9367 20856 9368 20857 9369 20858 9370 20859 8 20860 9371 20861 9372 20862 9373 20863 9374 20864 8 20865 9375 20866 9376 20867 9377 20868 9378 20869 9379 20870 8 20871 9380 20872 9381 20873 9382 20874 9383 20875 9384 20876 9385 20877 9386 20878 9387 20879 9388 20880 8 20881 9389 20882 9390 20883 9391 20884 9392 20885 9393 20886 9394 20887 9395 20888 8 20889 9396 20890 9397 20891 8 20892 9398 20893 8 20894 9399 20895 9400 20896 8 20897 9401 20898 8 20899 8 20900 9402 20901 9403 20902 8 20903 9404 20904 9405 20905 9406 20906 8 20907 9407 20908 9408 20909 9409 20910 9410 20911 8 20912 8 20913 9411 20914 9412 20915 8 20916 9413 20917 9414 20918 9415 20919 9416 20920 8 20921 9417 20922 9418 20923 9419 20924 9420 20925 9421 20926 9422 20927 8 20928 9423 20929 9424 20930 8 20931 9425 20932 9426 20933 9427 20934 9428 20935 9429 20936 9430 20937 8 20938 9431 20939 9432 20940 9433 20941 9434 20942 8 20943 9435 20944 9436 20945 9437 20946 8 20947 9438 20948 8 20949 9439 20950 9440 20951 9441 20952 8 20953 9442 20954 9443 20955 8 20956 8 20957 9444 20958 8 20959 9445 20960 8 20961 9446 20962 8 20963 9447 20964 8 20965 9448 20966 9449 20967 9450 20968 9451 20969 9452 20970 9453 20971 9454 20972 9455 20973 9456 20974 8 20975 8 20976 9457 20977 9458 20978 8 20979 8 20980 9459 20981 9460 20982 9461 20983 9462 20984 9463 20985 9464 20986 9465 20987 9466 20988 8 20989 8 20990 9467 20991 9468 20992 8 20993 9469 20994 8 20995 9470 20996 9471 20997 9472 20998 9473 20999 9474 21000 8 21001 9475 21002 9476 21003 8 21004 9477 21005 9478 21006 9479 21007 8 21008 9480 21009 8 21010 8 21011 9481 21012 8 21013 9482 21014 9483 21015 9484 21016 9485 21017 9486 21018 9487 21019 9488 21020 9489 21021 9490 21022 9491 21023 9492 21024 9493 21025 9494 21026 9495 21027 8 21028 9496 21029 9497 21030 9498 21031 9499 21032 9500 21033 9501 21034 9502 21035 9503 21036 9504 21037 8 21038 9505 21039 8 21040 9506 21041 8 21042 9507 21043 8 21044 9508 21045 9509 21046 8 21047 8 21048 9510 21049 8 21050 8 21051 8 21052 8 21053 9511 21054 9512 21055 8 21056 9513 21057 9514 21058 9515 21059 9516 21060 9517 21061 8 21062 9518 21063 9519 21064 9520 21065 9521 21066 9522 21067 9523 21068 9524 21069 9525 21070 9526 21071 8 21072 9527 21073 9528 21074 9529 21075 9530 21076 9531 21077 9532 21078 9533 21079 8 21080 9534 21081 9535 21082 8 21083 9536 21084 8 21085 9537 21086 9538 21087 8 21088 9539 21089 8 21090 8 21091 9540 21092 9541 21093 8 21094 9542 21095 9543 21096 9544 21097 8 21098 9545 21099 9546 21100 9547 21101 9548 21102 8 21103 8 21104 9549 21105 9550 21106 8 21107 9551 21108 9552 21109 9553 21110 9554 21111 8 21112 9555 21113 9556 21114 9557 21115 9558 21116 9559 21117 9560 21118 8 21119 9561 21120 9562 21121 8 21122 9563 21123 9564 21124 9565 21125 9566 21126 9567 21127 9568 21128 8 21129 9569 21130 9570 21131 9571 21132 9572 21133 8 21134 9573 21135 9574 21136 9575 21137 8 21138 9576 21139 8 21140 9577 21141 9578 21142 9579 21143 8 21144 9580 21145 9581 21146 8 21147 8 21148 9582 21149 8 21150 9583 21151 8 21152 9584 21153 8 21154 9585 21155 8 21156 9586 21157 9587 21158 9588 21159 9589 21160 9590 21161 9591 21162 9592 21163 9593 21164 9594 21165 8 21166 8 21167 9595 21168 9596 21169 8 21170 8 21171 9597 21172 9598 21173 9599 21174 9600 21175 9601 21176 9602 21177 9603 21178 9604 21179 8 21180 8 21181 9605 21182 9606 21183 8 21184 9607 21185 8 21186 9608 21187 9609 21188 9610 21189 9611 21190 9612 21191 8 21192 9613 21193 9614 21194 9615 21195 9616 21196 8 21197 8 21198 9617 21199 9618 21200 8 21201 9619 21202 8 21203 9620 21204 8 21205 8 21206 9621 21207 9622 21208 9623 21209 9624 21210 9625 21211 9626 21212 9627 21213 9628 21214 9629 21215 9630 21216 9631 21217 8 21218 8 21219 9632 21220 9633 21221 9634 21222 9635 21223 9636 21224 9637 21225 9638 21226 8 21227 8 21228 9639 21229 9640 21230 8 21231 9641 21232 9642 21233 9643 21234 8 21235 9644 21236 9645 21237 9646 21238 9647 21239 8 21240 8 21241 9648 21242 9649 21243 8 21244 9650 21245 9651 21246 9652 21247 9653 21248 8 21249 9654 21250 9655 21251 9656 21252 9657 21253 9658 21254 9659 21255 8 21256 9660 21257 9661 21258 8 21259 9662 21260 9663 21261 9664 21262 9665 21263 9666 21264 9667 21265 8 21266 9668 21267 9669 21268 9670 21269 9671 21270 8 21271 9672 21272 9673 21273 9674 21274 9675 21275 8 21276 9676 21277 9677 21278 9678 21279 8 21280 9679 21281 8 21282 9680 21283 9681 21284 9682 21285 8 21286 9683 21287 9684 21288 8 21289 8 21290 9685 21291 8 21292 9686 21293 8 21294 9687 21295 9688 21296 9689 21297 9690 21298 9691 21299 9692 21300 8 21301 3197 21302 9693 21303 9694 21304 9695 21305 9696 21306 8 21307 7499 21308 9697 21309 9698 21310 8 21311 9699 21312 8 21313 9700 21314 8 21315 9701 21316 9702 21317 9703 21318 9704 21319 9705 21320 9706 21321 9707 21322 9708 21323 9709 21324 8 21325 8 21326 9710 21327 9711 21328 8 21329 8 21330 9712 21331 9713 21332 9714 21333 9715 21334 9716 21335 9717 21336 9718 21337 9719 21338 8 21339 8 21340 9720 21341 9721 21342 8 21343 9722 21344 9723 21345 9724 21346 9725 21347 9726 21348 9727 21349 9728 21350 8 21351 9729 21352 8 21353 8 21354 8 21355 8 21356 8 21357 8 21358 8 21359 9730 21360 8 21361 9731 21362 8 21363 8 21364 9732 21365 9733 21366 9734 21367 9735 21368 8 21369 8 21370 8 21371 8 21372 8 21373 8 21374 8 21375 8 21376 8 21377 8 21378 8 21379 9736 21380 8 21381 9737 21382 8 21383 9738 21384 8 21385 8 21386 8 21387 8 21388 8 21389 8 21390 8 21391 8 21392 8 21393 8 21394 9739 21395 8 21396 8 21397 8 21398 8 21399 8 21400 8 21401 8 21402 8 21403 9740 21404 8 21405 9741 21406 9742 21407 9743 21408 9744 21409 9745 21410 8 21411 9746 21412 9747 21413 9748 21414 9749 21415 8 21416 8 21417 9750 21418 9751 21419 8 21420 9752 21421 8 21422 9753 21423 9754 21424 9755 21425 9756 21426 9757 21427 9758 21428 8 21429 9759 21430 8 21431 9760 21432 9761 21433 8 21434 9762 21435 8 21436 9763 21437 8 21438 8 21439 8 21440 9764 21441 9765 21442 8 21443 9766 21444 9767 21445 9768 21446 8 21447 9769 21448 9770 21449 9771 21450 9772 21451 8 21452 8 21453 9773 21454 9774 21455 8 21456 9775 21457 9776 21458 9777 21459 9778 21460 8 21461 9779 21462 9780 21463 9781 21464 9782 21465 9783 21466 9784 21467 8 21468 9785 21469 9786 21470 8 21471 9787 21472 9788 21473 9789 21474 9790 21475 9791 21476 9792 21477 8 21478 9793 21479 9794 21480 9795 21481 9796 21482 8 21483 9797 21484 9798 21485 9799 21486 8 21487 9800 21488 8 21489 9801 21490 9802 21491 9803 21492 8 21493 9804 21494 9805 21495 8 21496 8 21497 9806 21498 8 21499 9807 21500 8 21501 9808 21502 9809 21503 9810 21504 8 21505 9811 21506 8 21507 9812 21508 8 21509 9813 21510 9814 21511 9815 21512 9816 21513 9817 21514 9818 21515 9819 21516 9820 21517 9821 21518 8 21519 8 21520 9822 21521 9823 21522 8 21523 8 21524 9824 21525 9825 21526 9826 21527 9827 21528 9828 21529 9829 21530 9830 21531 9831 21532 8 21533 8 21534 9832 21535 9833 21536 8 21537 9834 21538 8 21539 9835 21540 9836 21541 9837 21542 9838 21543 9839 21544 8 21545 9840 21546 9841 21547 9842 21548 9843 21549 8 21550 8 21551 9844 21552 9845 21553 8 21554 9846 21555 8 21556 9847 21557 8 21558 9848 21559 9849 21560 9850 21561 9851 21562 8 21563 8 21564 9852 21565 9853 21566 8 21567 9854 21568 8 21569 9855 21570 8 21571 8 21572 9856 21573 9857 21574 9858 21575 9859 21576 9860 21577 9861 21578 9862 21579 9863 21580 9864 21581 9865 21582 9866 21583 8 21584 8 21585 9867 21586 9868 21587 9869 21588 9870 21589 9871 21590 9872 21591 9873 21592 8 21593 8 21594 9874 21595 9875 21596 8 21597 9876 21598 9877 21599 9878 21600 8 21601 9879 21602 9880 21603 9881 21604 9882 21605 8 21606 8 21607 9883 21608 9884 21609 8 21610 9885 21611 9886 21612 9887 21613 9888 21614 8 21615 9889 21616 9890 21617 9891 21618 9892 21619 9893 21620 9894 21621 8 21622 9895 21623 9896 21624 8 21625 9897 21626 9898 21627 9899 21628 9900 21629 9901 21630 9902 21631 8 21632 9903 21633 9904 21634 9905 21635 9906 21636 8 21637 9907 21638 9908 21639 9909 21640 9910 21641 8 21642 9911 21643 9912 21644 9913 21645 8 21646 9914 21647 8 21648 9915 21649 9916 21650 9917 21651 8 21652 9918 21653 9919 21654 8 21655 8 21656 9920 21657 8 21658 9921 21659 8 21660 9922 21661 9923 21662 9924 21663 9925 21664 9926 21665 8 21666 9927 21667 9928 21668 9929 21669 8 21670 9930 21671 9931 21672 8 21673 9932 21674 8 21675 9933 21676 8 21677 9934 21678 9935 21679 9936 21680 9937 21681 9938 21682 9939 21683 9940 21684 9941 21685 9942 21686 8 21687 8 21688 9943 21689 9944 21690 8 21691 8 21692 9945 21693 9946 21694 9947 21695 9948 21696 9949 21697 9950 21698 9951 21699 9952 21700 8 21701 8 21702 9953 21703 9954 21704 8 21705 9955 21706 9956 21707 9957 21708 9958 21709 9959 21710 9960 21711 8 21712 8 21713 9961 21714 8 21715 9962 21716 9963 21717 9964 21718 9965 21719 9966 21720 8 21721 9967 21722 9968 21723 9969 21724 9970 21725 8 21726 8 21727 9971 21728 9972 21729 8 21730 9973 21731 8 21732 9974 21733 9975 21734 9976 21735 9977 21736 9978 21737 9979 21738 8 21739 9980 21740 8 21741 9981 21742 8 21743 9982 21744 9983 21745 9984 21746 9985 21747 9986 21748 8 21749 9987 21750 9988 21751 9989 21752 9990 21753 8 21754 8 21755 9991 21756 9992 21757 8 21758 9993 21759 8 21760 9994 21761 8 21762 8 21763 9995 21764 9996 21765 9997 21766 9998 21767 9999 21768 10000 21769 10001 21770 10002 21771 10003 21772 10004 21773 10005 21774 8 21775 8 21776 10006 21777 10007 21778 10008 21779 10009 21780 10010 21781 10011 21782 10012 21783 8 21784 8 21785 10013 21786 10014 21787 8 21788 10015 21789 10016 21790 10017 21791 8 21792 10018 21793 10019 21794 10020 21795 10021 21796 8 21797 8 21798 10022 21799 10023 21800 8 21801 10024 21802 10025 21803 10026 21804 10027 21805 8 21806 10028 21807 10029 21808 10030 21809 10031 21810 10032 21811 10033 21812 8 21813 10034 21814 10035 21815 8 21816 10036 21817 10037 21818 10038 21819 10039 21820 10040 21821 10041 21822 8 21823 10042 21824 10043 21825 10044 21826 10045 21827 8 21828 10046 21829 10047 21830 10048 21831 10049 21832 8 21833 10050 21834 10051 21835 10052 21836 8 21837 10053 21838 8 21839 10054 21840 10055 21841 10056 21842 8 21843 10057 21844 10058 21845 8 21846 8 21847 10059 21848 8 21849 10060 21850 8 21851 10061 21852 10062 21853 10063 21854 10064 21855 10065 21856 8 21857 10066 21858 10067 21859 10068 21860 8 21861 10069 21862 10070 21863 8 21864 10071 21865 8 21866 10072 21867 8 21868 10073 21869 10074 21870 10075 21871 10076 21872 10077 21873 10078 21874 10079 21875 10080 21876 10081 21877 8 21878 8 21879 10082 21880 10083 21881 8 21882 8 21883 10084 21884 10085 21885 10086 21886 10087 21887 10088 21888 10089 21889 10090 21890 10091 21891 8 21892 8 21893 10092 21894 10093 21895 8 21896 10094 21897 10095 21898 10096 21899 10097 21900 10098 21901 10099 21902 8 21903 8 21904 10100 21905 8 21906 10101 21907 10102 21908 10103 21909 10104 21910 10105 21911 8 21912 10106 21913 10107 21914 10108 21915 10109 21916 8 21917 8 21918 10110 21919 10111 21920 8 21921 10112 21922 8 21923 10113 21924 10114 21925 10115 21926 10116 21927 10117 21928 10118 21929 8 21930 10119 21931 8 21932 10120 21933 10121 21934 8 21935 8 21936 8 21937 10122 21938 10123 21939 10124 21940 10125 21941 8 21942 8 21943 10126 21944 8 21945 10127 21946 10128 21947 8 21948 8 21949 10129 21950 8 21951 8 21952 10130 21953 8 21954 8 21955 10131 21956 10132 21957 8 21958 8 21959 10133 21960 10134 21961 8 21962 8 21963 10135 21964 10136 21965 8 21966 8 21967 8 21968 10137 21969 10138 21970 10139 21971 10140 21972 8 21973 8 21974 10141 21975 8 21976 10142 21977 10143 21978 8 21979 8 21980 10144 21981 8 21982 8 21983 10145 21984 8 21985 8 21986 10146 21987 10147 21988 8 21989 8 21990 10148 21991 10149 21992 8 21993 10150 21994 8 21995 10151 21996 8 21997 10152 21998 10153 21999 8 22000 8 22001 8 22002 10154 22003 8 22004 8 22005 10155 22006 10156 22007 10157 22008 10158 22009 8 22010 8 22011 10159 22012 8 22013 10160 22014 10161 22015 8 22016 8 22017 10162 22018 8 22019 10163 22020 8 22021 10164 22022 10165 22023 10166 22024 8469 22025 8469 22026 10167 22027 10168 22028 10169 22029 10170 22030 10171 22031 10172 22032 10173 22033 10174 22034 10175 22035 10176 22036 8 22037 10177 22038 8 22039 10178 22040 8 22041 8 22042 10179 22043 10180 22044 10181 22045 10182 22046 8 22047 8 22048 10183 22049 10184 22050 10185 22051 10186 22052 10187 22053 8 22054 10188 22055 8 22056 10189 22057 8 22058 8 22059 8 22060 8 22061 10190 22062 8 22063 10191 22064 10192 22065 8 22066 10193 22067 10194 22068 10195 22069 10196 22070 10197 22071 8 22072 10198 22073 10199 22074 8 22075 10200 22076 8 22077 10201 22078 10202 22079 10203 22080 10204 22081 10205 22082 10206 22083 10207 22084 8 22085 8 22086 10208 22087 8 22088 8 22089 10209 22090 10210 22091 10211 22092 8 22093 8 22094 8 22095 8 22096 10212 22097 8 22098 8 22099 8 22100 8 22101 8 22102 10213 22103 10214 22104 10215 22105 10216 22106 10217 22107 10218 22108 10219 22109 10220 22110 10221 22111 10222 22112 10223 22113 10224 22114 10225 22115 10226 22116 10227 22117 10228 22118 10229 22119 10230 22120 10231 22121 10232 22122 10233 22123 10234 22124 10235 22125 10236 22126 10237 22127 10238 22128 10239 22129 10240 22130 10241 22131 10242 22132 10243 22133 10244 22134 10245 22135 10246 22136 10216 22137 10216 22138 10216 22139 10216 22140 10216 22141 10216 22142 10216 22143 10216 22144 10216 22145 10216 22146 10216 22147 10216 22148 10216 22149 10216 22150 10216 22151 10247 22152 10248 22153 8 22154 8 22155 8 22156 8 22157 8 22158 8 22159 10249 22160 8 22161 8 22162 10250 22163 8 22164 10251 22165 10252 22166 10253 22167 10254 22168 10255 22169 8 22170 8 22171 8 22172 8 22173 8 22174 10256 22175 10257 22176 8 22177 8 22178 10258 22179 8 22180 10259 22181 8 22182 8 22183 10260 22184 10261 22185 8 22186 10216 22187 10216 22188 10216 22189 10216 22190 10216 22191 10262 22192 10263 22193 10264 22194 10216 22195 10216 22196 10216 22197 10216 22198 10216 22199 10216 22200 10216 22201 10216 22202 10216 22203 10216 22204 10265 22205 8 22206 8 22207 8 22208 8 22209 8 22210 8 22211 8 22212 8 22213 8 22214 8 22215 8 22216 8 22217 8 22218 8 22219 8 22220 8 22221 8 22222 8 22223 8 22224 8 22225 8 22226 8 22227 8 22228 8 22229 8 22230 8 22231 8 22232 8 22233 8 22234 8 22235 8 22236 8 22237 8 22238 8 22239 8 22240 8 22241 8 22242 8 22243 8 22244 8 22245 8 22246 8 22247 8 22248 8 22249 8 22250 8 22251 8 22252 8 22253 8 22254 8 22255 8 22256 8 22257 8 22258 8 22259 8 22260 8 22261 8 22262 8 22263 8 22264 8 22265 8 22266 8 22267 8 22268 8 22269 8 22270 8 22271 8 22272 8 22273 8 22274 8 22275 8 22276 8 22277 8 22278 8 22279 8 22280 8 22281 8 22282 8 22283 8 22284 8 22285 10266 22286 8 22287 8 22288 8 22289 10267 22290 10268 22291 8 22292 10269 22293 8 22294 8 22295 10270 22296 10271 22297 10272 22298 8 22299 8 22300 10273 22301 10274 22302 10275 22303 10276 22304 8 22305 10277 22306 8 22307 10278 22308 10279 22309 10280 22310 10281 22311 10282 22312 10283 22313 10284 22314 10285 22315 10286 22316 10216 22317 10216 22318 10216 22319 10216 22320 10216 22321 10216 22322 10216 22323 10216 22324 10216 22325 10216 22326 10216 22327 10216 22328 10216 22329 10287 22330 10216 22331 10288 22332 10289 22333 8 22334 10290 22335 10291 22336 8 22337 10292 22338 8 22339 10293 22340 8 22341 8 22342 8 22343 10294 22344 10295 22345 10296 22346 8 22347 10297 22348 8 22349 8 22350 10298 22351 10299 22352 10300 22353 8 22354 8 22355 8 22356 10301 22357 10302 22358 8 22359 8 22360 8 22361 10303 22362 10304 22363 10305 22364 10306 22365 8 22366 8 22367 10307 22368 8 22369 10308 22370 10309 22371 8 22372 8 22373 10310 22374 8 22375 8 22376 8 22377 10311 22378 10312 22379 8 22380 8 22381 10313 22382 10314 22383 10315 22384 10316 22385 10317 22386 10318 22387 10319 22388 10320 22389 10321 22390 10322 22391 10323 22392 8 22393 10324 22394 10325 22395 10326 22396 10327 22397 10328 22398 10329 22399 10330 22400 10331 22401 10332 22402 10333 22403 8 22404 10334 22405 8469 22406 10335 22407 8 22408 10336 22409 8 22410 10337 22411 10338 22412 8 22413 8 22414 8 22415 10339 22416 8 22417 8 22418 10340 22419 10341 22420 10342 22421 10343 22422 8 22423 8 22424 10344 22425 8 22426 10345 22427 10346 22428 8 22429 8 22430 10347 22431 8 22432 10348 22433 8 22434 10349 22435 10350 22436 10351 22437 10216 22438 10216 22439 10352 22440 10353 22441 10216 22442 10216 22443 10216 22444 10216 22445 8 22446 10354 22447 8 22448 10355 22449 8 22450 8 22451 10216 22452 10216 22453 8 22454 10216 22455 10216 22456 10356 22457 8 22458 8 22459 8 22460 8 22461 8 22462 8 22463 10357 22464 10358 22465 10359 22466 10360 22467 8 22468 10361 22469 8 22470 10362 22471 10363 22472 10364 22473 10365 22474 10366 22475 8 22476 8 22477 8 22478 8 22479 8 22480 10367 22481 8 22482 8 22483 10368 22484 10369 22485 10370 22486 10371 22487 10372 22488 10373 22489 10374 22490 8 22491 10375 22492 8 22493 10376 22494 10377 22495 10378 22496 10379 22497 10380 22498 8 22499 10381 22500 10382 22501 10383 22502 10384 22503 10216 22504 10385 22505 10386 22506 10387 22507 8 22508 10216 22509 10216 22510 10216 22511 8 22512 10388 22513 8 22514 8 22515 8 22516 10389 22517 10390 22518 10391 22519 8 22520 8 22521 10392 22522 10393 22523 8 22524 10394 22525 10395 22526 8 22527 10216 22528 8 22529 10216 22530 10216 22531 10216 22532 10396 22533 10397 22534 8 22535 10216 22536 10216 22537 10216 22538 10398 22539 8 22540 8 22541 10399 22542 10400 22543 8 22544 10401 22545 8 22546 10216 22547 8469 22548 8469 22549 8469 22550 8 22551 10402 22552 8 22553 10403 22554 8 22555 8 22556 8469 22557 8469 22558 8 22559 8469 22560 8469 22561 10404 22562 8 22563 8 22564 8 22565 8 22566 8 22567 8 22568 10405 22569 10406 22570 10407 22571 10408 22572 8 22573 10409 22574 8 22575 10410 22576 10411 22577 10412 22578 10413 22579 10414 22580 8 22581 8 22582 8 22583 8 22584 8 22585 10415 22586 8 22587 8 22588 10416 22589 10417 22590 10418 22591 10419 22592 10420 22593 10421 22594 10422 22595 8 22596 10423 22597 8 22598 10424 22599 10425 22600 10426 22601 10427 22602 10428 22603 8 22604 10429 22605 10430 22606 10431 22607 10432 22608 8469 22609 10433 22610 10434 22611 10435 22612 8 22613 8469 22614 8469 22615 8469 22616 8 22617 10436 22618 8 22619 8 22620 8 22621 10437 22622 10438 22623 10439 22624 8 22625 8 22626 10440 22627 10441 22628 8 22629 10442 22630 10443 22631 8 22632 8469 22633 8 22634 8469 22635 8469 22636 8469 22637 10444 22638 10445 22639 8 22640 8469 22641 8469 22642 8469 22643 10446 22644 8 22645 8 22646 10447 22647 10448 22648 8 22649 10449 22650 8 22651 8469 22652 10450 22653 10451 22654 10452 22655 10453 22656 10454 22657 10455 22658 10456 22659 10457 22660 8 22661 8 22662 10458 22663 10459 22664 8 22665 10460 22666 8 22667 10461 22668 10462 22669 10463 22670 10464 22671 10465 22672 8 22673 10466 22674 10467 22675 10468 22676 10469 22677 8 22678 8 22679 10470 22680 10471 22681 8 22682 10472 22683 8 22684 10473 22685 8 22686 8 22687 10474 22688 10475 22689 10476 22690 10477 22691 10478 22692 10479 22693 10480 22694 10481 22695 10482 22696 10483 22697 10484 22698 8 22699 8 22700 10485 22701 10486 22702 10487 22703 10488 22704 10489 22705 10490 22706 10491 22707 8 22708 8 22709 10492 22710 10493 22711 8 22712 10494 22713 10495 22714 10496 22715 8 22716 10497 22717 10498 22718 10499 22719 10500 22720 8 22721 8 22722 10501 22723 10502 22724 8 22725 10503 22726 10504 22727 10505 22728 10506 22729 8 22730 10507 22731 10508 22732 10509 22733 10510 22734 10511 22735 10512 22736 8 22737 10513 22738 10514 22739 8 22740 10515 22741 10516 22742 10517 22743 10518 22744 10519 22745 10520 22746 8 22747 10521 22748 10522 22749 10523 22750 10524 22751 8 22752 10525 22753 10526 22754 10527 22755 10528 22756 8 22757 10529 22758 10530 22759 10531 22760 8 22761 10532 22762 8 22763 10533 22764 10534 22765 10535 22766 8 22767 10536 22768 10537 22769 8 22770 8 22771 10538 22772 8 22773 10539 22774 8 22775 10540 22776 10541 22777 10542 22778 10543 22779 10544 22780 8 22781 10545 22782 10546 22783 10547 22784 8 22785 10548 22786 10549 22787 8 22788 10550 22789 8 22790 10551 22791 8 22792 10552 22793 10553 22794 10554 22795 10555 22796 10556 22797 10557 22798 10558 22799 10559 22800 10560 22801 8 22802 8 22803 10561 22804 10562 22805 8 22806 8 22807 10563 22808 10564 22809 10565 22810 10566 22811 10567 22812 10568 22813 10569 22814 10570 22815 8 22816 8 22817 10571 22818 10572 22819 8 22820 10573 22821 10574 22822 10575 22823 10576 22824 10577 22825 10578 22826 8 22827 10579 22828 8 22829 10580 22830 10581 22831 10582 22832 10583 22833 10584 22834 8 22835 10585 22836 10586 22837 10587 22838 10588 22839 8 22840 8 22841 10589 22842 10590 22843 8 22844 10591 22845 8 22846 10592 22847 10593 22848 10594 22849 10595 22850 10596 22851 10597 22852 8 22853 10598 22854 8 22855 10599 22856 10600 22857 10601 22858 10602 22859 10603 22860 10604 22861 10605 22862 10606 22863 10607 22864 10608 22865 8 22866 10609 22867 8 22868 10610 22869 8 22870 8 22871 10611 22872 10612 22873 10613 22874 10614 22875 8 22876 8 22877 10615 22878 10616 22879 10617 22880 10618 22881 10619 22882 8 22883 10620 22884 8 22885 10621 22886 8 22887 8 22888 8 22889 8 22890 10622 22891 8 22892 10623 22893 10624 22894 8 22895 10625 22896 10626 22897 10627 22898 10628 22899 10629 22900 8 22901 10630 22902 10631 22903 8 22904 10632 22905 8 22906 10633 22907 10634 22908 10635 22909 10636 22910 10637 22911 10638 22912 10639 22913 8 22914 8 22915 10640 22916 8 22917 8 22918 10641 22919 10642 22920 10643 22921 8 22922 8 22923 8 22924 8 22925 10644 22926 8 22927 8 22928 8 22929 8 22930 8 22931 10645 22932 10646 22933 10647 22934 10648 22935 10649 22936 10650 22937 10651 22938 10652 22939 10653 22940 10654 22941 10655 22942 10656 22943 10657 22944 10658 22945 10659 22946 10660 22947 10661 22948 10662 22949 10663 22950 10664 22951 10665 22952 10666 22953 10667 22954 10668 22955 10669 22956 10670 22957 10671 22958 10672 22959 10673 22960 10674 22961 10675 22962 10676 22963 10677 22964 10678 22965 10648 22966 10648 22967 10648 22968 10648 22969 10648 22970 10648 22971 10648 22972 10648 22973 10648 22974 10648 22975 10648 22976 10648 22977 10648 22978 10648 22979 10648 22980 10679 22981 10680 22982 8 22983 8 22984 8 22985 8 22986 8 22987 8 22988 10681 22989 8 22990 8 22991 10682 22992 8 22993 10683 22994 10684 22995 10685 22996 10686 22997 10687 22998 8 22999 8 23000 8 23001 8 23002 8 23003 10688 23004 10689 23005 8 23006 8 23007 10690 23008 8 23009 10691 23010 8 23011 8 23012 10692 23013 10693 23014 8 23015 10648 23016 10648 23017 10648 23018 10648 23019 10648 23020 10694 23021 10695 23022 10696 23023 10648 23024 10648 23025 10648 23026 10648 23027 10648 23028 10648 23029 10648 23030 10648 23031 10648 23032 10648 23033 10697 23034 8 23035 8 23036 8 23037 8 23038 8 23039 8 23040 8 23041 8 23042 8 23043 8 23044 8 23045 8 23046 8 23047 8 23048 8 23049 8 23050 8 23051 8 23052 8 23053 8 23054 8 23055 8 23056 8 23057 8 23058 8 23059 8 23060 8 23061 8 23062 8 23063 8 23064 8 23065 8 23066 8 23067 8 23068 8 23069 8 23070 8 23071 8 23072 8 23073 8 23074 8 23075 8 23076 8 23077 8 23078 8 23079 8 23080 8 23081 8 23082 8 23083 8 23084 8 23085 8 23086 8 23087 8 23088 8 23089 8 23090 8 23091 8 23092 8 23093 8 23094 8 23095 8 23096 8 23097 8 23098 8 23099 8 23100 8 23101 8 23102 8 23103 8 23104 8 23105 8 23106 8 23107 8 23108 8 23109 8 23110 8 23111 8 23112 8 23113 8 23114 10698 23115 8 23116 8 23117 8 23118 10699 23119 10700 23120 8 23121 10701 23122 8 23123 8 23124 10702 23125 10703 23126 10704 23127 8 23128 8 23129 10705 23130 10706 23131 10707 23132 10708 23133 8 23134 10709 23135 8 23136 10710 23137 10711 23138 10712 23139 10713 23140 10714 23141 10715 23142 10716 23143 10717 23144 10718 23145 10648 23146 10648 23147 10648 23148 10648 23149 10648 23150 10648 23151 10648 23152 10648 23153 10648 23154 10648 23155 10648 23156 10648 23157 10648 23158 10719 23159 10648 23160 10720 23161 10721 23162 8 23163 10722 23164 10723 23165 8 23166 10724 23167 8 23168 10725 23169 8 23170 8 23171 8 23172 10726 23173 10727 23174 10728 23175 8 23176 10729 23177 8 23178 8 23179 10730 23180 10731 23181 10732 23182 8 23183 8 23184 8 23185 10733 23186 10734 23187 8 23188 8 23189 8 23190 10735 23191 10736 23192 10737 23193 10738 23194 8 23195 8 23196 10739 23197 8 23198 10740 23199 10741 23200 8 23201 8 23202 10742 23203 8 23204 8 23205 8 23206 10743 23207 10744 23208 8 23209 8 23210 10745 23211 10746 23212 10747 23213 10748 23214 10749 23215 10750 23216 10751 23217 10752 23218 10753 23219 10754 23220 10755 23221 8 23222 10756 23223 10757 23224 10758 23225 10759 23226 10760 23227 10761 23228 10762 23229 10763 23230 10764 23231 10765 23232 10766 23233 10767 23234 10768 23235 10769 23236 10770 23237 10771 23238 10772 23239 10773 23240 10774 23241 8 23242 10775 23243 8 23244 10776 23245 8 23246 8 23247 10777 23248 10778 23249 10779 23250 10780 23251 8 23252 8 23253 10781 23254 10782 23255 10783 23256 10784 23257 10785 23258 8 23259 10786 23260 8 23261 10787 23262 8 23263 8 23264 8 23265 8 23266 10788 23267 8 23268 10789 23269 10790 23270 8 23271 10791 23272 10792 23273 10793 23274 10794 23275 10795 23276 8 23277 10796 23278 10797 23279 8 23280 10798 23281 8 23282 10799 23283 10800 23284 10801 23285 10802 23286 10803 23287 10804 23288 10805 23289 8 23290 8 23291 10806 23292 8 23293 8 23294 10807 23295 10808 23296 10809 23297 8 23298 8 23299 8 23300 8 23301 10810 23302 8 23303 8 23304 8 23305 8 23306 8 23307 10811 23308 10812 23309 10813 23310 10814 23311 10815 23312 10816 23313 10817 23314 10818 23315 10819 23316 10820 23317 10821 23318 10822 23319 10823 23320 10824 23321 10825 23322 10826 23323 10827 23324 10828 23325 10829 23326 10830 23327 10831 23328 10832 23329 10833 23330 10834 23331 10835 23332 10836 23333 10837 23334 10838 23335 10839 23336 10840 23337 10841 23338 10842 23339 10843 23340 10844 23341 10814 23342 10814 23343 10814 23344 10814 23345 10814 23346 10814 23347 10814 23348 10814 23349 10814 23350 10814 23351 10814 23352 10814 23353 10814 23354 10814 23355 10814 23356 10845 23357 10846 23358 8 23359 8 23360 8 23361 8 23362 8 23363 8 23364 10847 23365 8 23366 8 23367 10848 23368 8 23369 10849 23370 10850 23371 10851 23372 10852 23373 10853 23374 8 23375 8 23376 8 23377 8 23378 8 23379 10854 23380 10855 23381 8 23382 8 23383 10856 23384 8 23385 10857 23386 8 23387 8 23388 10858 23389 10859 23390 8 23391 10814 23392 10814 23393 10814 23394 10814 23395 10814 23396 10860 23397 10861 23398 10862 23399 10814 23400 10814 23401 10814 23402 10814 23403 10814 23404 10814 23405 10814 23406 10814 23407 10814 23408 10814 23409 10863 23410 8 23411 8 23412 8 23413 8 23414 8 23415 8 23416 8 23417 8 23418 8 23419 8 23420 8 23421 8 23422 8 23423 8 23424 8 23425 8 23426 8 23427 8 23428 8 23429 8 23430 8 23431 8 23432 8 23433 8 23434 8 23435 8 23436 8 23437 8 23438 8 23439 8 23440 8 23441 8 23442 8 23443 8 23444 8 23445 8 23446 8 23447 8 23448 8 23449 8 23450 8 23451 8 23452 8 23453 8 23454 8 23455 8 23456 8 23457 8 23458 8 23459 8 23460 8 23461 8 23462 8 23463 8 23464 8 23465 8 23466 8 23467 8 23468 8 23469 8 23470 8 23471 8 23472 8 23473 8 23474 8 23475 8 23476 8 23477 8 23478 8 23479 8 23480 8 23481 8 23482 8 23483 8 23484 8 23485 8 23486 8 23487 8 23488 8 23489 8 23490 10864 23491 8 23492 8 23493 8 23494 10865 23495 10866 23496 8 23497 10867 23498 8 23499 8 23500 10868 23501 10869 23502 10870 23503 8 23504 8 23505 10871 23506 10872 23507 10873 23508 10874 23509 8 23510 10875 23511 8 23512 10876 23513 10877 23514 10878 23515 10879 23516 10880 23517 10881 23518 10882 23519 10883 23520 10884 23521 10814 23522 10814 23523 10814 23524 10814 23525 10814 23526 10814 23527 10814 23528 10814 23529 10814 23530 10814 23531 10814 23532 10814 23533 10814 23534 10885 23535 10814 23536 10886 23537 10887 23538 8 23539 10888 23540 10889 23541 8 23542 10890 23543 8 23544 10891 23545 8 23546 8 23547 8 23548 10892 23549 10893 23550 10894 23551 8 23552 10895 23553 8 23554 8 23555 10896 23556 10897 23557 10898 23558 8 23559 8 23560 8 23561 10899 23562 10900 23563 8 23564 8 23565 8 23566 10901 23567 10902 23568 10903 23569 10904 23570 8 23571 8 23572 10905 23573 8 23574 10906 23575 10907 23576 8 23577 8 23578 10908 23579 8 23580 8 23581 8 23582 10909 23583 10910 23584 8 23585 8 23586 10911 23587 10912 23588 10913 23589 10914 23590 10915 23591 10916 23592 10917 23593 10918 23594 10919 23595 10920 23596 10921 23597 8 23598 10922 23599 10923 23600 10924 23601 10925 23602 10926 23603 10927 23604 10928 23605 10929 23606 10930 23607 10931 23608 8 23609 8 23610 10932 23611 8 23612 10933 23613 8 23614 10934 23615 10935 23616 8 23617 8 23618 8 23619 10936 23620 8 23621 8 23622 10937 23623 10938 23624 10939 23625 10940 23626 8 23627 8 23628 10941 23629 8 23630 10942 23631 10943 23632 8 23633 8 23634 10944 23635 8 23636 10945 23637 8 23638 10946 23639 10947 23640 10948 23641 10814 23642 10814 23643 10949 23644 10950 23645 10951 23646 10952 23647 10953 23648 10954 23649 10955 23650 10956 23651 10957 23652 10958 23653 8 23654 10959 23655 8 23656 10960 23657 8 23658 8 23659 10961 23660 10962 23661 10963 23662 10964 23663 8 23664 8 23665 10965 23666 10966 23667 10967 23668 10968 23669 10969 23670 8 23671 10970 23672 8 23673 10971 23674 8 23675 8 23676 8 23677 8 23678 10972 23679 8 23680 10973 23681 10974 23682 8 23683 10975 23684 10976 23685 10977 23686 10978 23687 10979 23688 8 23689 10980 23690 10981 23691 8 23692 10982 23693 8 23694 10983 23695 10984 23696 10985 23697 10986 23698 10987 23699 10988 23700 10989 23701 8 23702 8 23703 10990 23704 8 23705 8 23706 10991 23707 10992 23708 10993 23709 8 23710 8 23711 8 23712 8 23713 10994 23714 8 23715 8 23716 8 23717 8 23718 8 23719 10995 23720 10996 23721 10997 23722 10998 23723 10999 23724 11000 23725 11001 23726 11002 23727 11003 23728 11004 23729 11005 23730 11006 23731 11007 23732 11008 23733 11009 23734 11010 23735 11011 23736 11012 23737 11013 23738 11014 23739 11015 23740 11016 23741 11017 23742 11018 23743 11019 23744 11020 23745 11021 23746 11022 23747 11023 23748 11024 23749 11025 23750 11026 23751 11027 23752 11028 23753 10998 23754 10998 23755 10998 23756 10998 23757 10998 23758 10998 23759 10998 23760 10998 23761 10998 23762 10998 23763 10998 23764 10998 23765 10998 23766 10998 23767 10998 23768 11029 23769 11030 23770 8 23771 8 23772 8 23773 8 23774 8 23775 8 23776 11031 23777 8 23778 8 23779 11032 23780 8 23781 11033 23782 11034 23783 11035 23784 11036 23785 11037 23786 8 23787 8 23788 8 23789 8 23790 8 23791 11038 23792 11039 23793 8 23794 8 23795 11040 23796 8 23797 11041 23798 8 23799 8 23800 11042 23801 11043 23802 8 23803 10998 23804 10998 23805 10998 23806 10998 23807 10998 23808 11044 23809 11045 23810 11046 23811 10998 23812 10998 23813 10998 23814 10998 23815 10998 23816 10998 23817 10998 23818 10998 23819 10998 23820 10998 23821 11047 23822 8 23823 8 23824 8 23825 8 23826 8 23827 8 23828 8 23829 8 23830 8 23831 8 23832 8 23833 8 23834 8 23835 8 23836 8 23837 8 23838 8 23839 8 23840 8 23841 8 23842 8 23843 8 23844 8 23845 8 23846 8 23847 8 23848 8 23849 8 23850 8 23851 8 23852 8 23853 8 23854 8 23855 8 23856 8 23857 8 23858 8 23859 8 23860 8 23861 8 23862 8 23863 8 23864 8 23865 8 23866 8 23867 8 23868 8 23869 8 23870 8 23871 8 23872 8 23873 8 23874 8 23875 8 23876 8 23877 8 23878 8 23879 8 23880 8 23881 8 23882 8 23883 8 23884 8 23885 8 23886 8 23887 8 23888 8 23889 8 23890 8 23891 8 23892 8 23893 8 23894 8 23895 8 23896 8 23897 8 23898 8 23899 8 23900 8 23901 8 23902 11048 23903 8 23904 8 23905 8 23906 11049 23907 11050 23908 8 23909 11051 23910 8 23911 8 23912 11052 23913 11053 23914 11054 23915 8 23916 8 23917 11055 23918 11056 23919 11057 23920 11058 23921 8 23922 11059 23923 8 23924 11060 23925 11061 23926 11062 23927 11063 23928 11064 23929 11065 23930 11066 23931 11067 23932 11068 23933 10998 23934 10998 23935 10998 23936 10998 23937 10998 23938 10998 23939 10998 23940 10998 23941 10998 23942 10998 23943 10998 23944 10998 23945 10998 23946 11069 23947 10998 23948 11070 23949 11071 23950 8 23951 11072 23952 11073 23953 8 23954 11074 23955 8 23956 11075 23957 8 23958 8 23959 8 23960 11076 23961 11077 23962 11078 23963 8 23964 11079 23965 8 23966 8 23967 11080 23968 11081 23969 11082 23970 8 23971 8 23972 8 23973 11083 23974 11084 23975 8 23976 8 23977 8 23978 11085 23979 11086 23980 11087 23981 11088 23982 8 23983 8 23984 11089 23985 8 23986 11090 23987 11091 23988 8 23989 8 23990 11092 23991 8 23992 8 23993 8 23994 11093 23995 11094 23996 8 23997 8 23998 11095 23999 11096 24000 11097 24001 11098 24002 11099 24003 11100 24004 11101 24005 11102 24006 11103 24007 11104 24008 11105 24009 8 24010 11106 24011 11107 24012 11108 24013 11109 24014 11110 24015 11111 24016 11112 24017 11113 24018 11114 24019 11115 24020 8 24021 11116 24022 10814 24023 11117 24024 8 24025 11118 24026 8 24027 11119 24028 11120 24029 8 24030 8 24031 8 24032 11121 24033 8 24034 8 24035 11122 24036 11123 24037 11124 24038 11125 24039 8 24040 8 24041 11126 24042 8 24043 11127 24044 11128 24045 8 24046 8 24047 11129 24048 8 24049 11130 24050 8 24051 11131 24052 11132 24053 11133 24054 10998 24055 10998 24056 11134 24057 11135 24058 10998 24059 10998 24060 10998 24061 10998 24062 8 24063 11136 24064 8 24065 11137 24066 8 24067 8 24068 10998 24069 10998 24070 8 24071 10998 24072 10998 24073 11138 24074 8 24075 8 24076 8 24077 8 24078 8 24079 8 24080 11139 24081 11140 24082 11141 24083 11142 24084 8 24085 11143 24086 8 24087 11144 24088 11145 24089 11146 24090 11147 24091 11148 24092 8 24093 8 24094 8 24095 8 24096 8 24097 11149 24098 8 24099 8 24100 11150 24101 11151 24102 11152 24103 11153 24104 11154 24105 11155 24106 11156 24107 8 24108 11157 24109 8 24110 11158 24111 11159 24112 11160 24113 11161 24114 11162 24115 8 24116 11163 24117 11164 24118 11165 24119 11166 24120 10998 24121 11167 24122 11168 24123 11169 24124 8 24125 10998 24126 10998 24127 10998 24128 8 24129 11170 24130 8 24131 8 24132 8 24133 11171 24134 11172 24135 11173 24136 8 24137 8 24138 11174 24139 11175 24140 8 24141 11176 24142 11177 24143 8 24144 10998 24145 8 24146 10998 24147 10998 24148 10998 24149 11178 24150 11179 24151 8 24152 10998 24153 10998 24154 10998 24155 11180 24156 8 24157 8 24158 11181 24159 11182 24160 8 24161 11183 24162 8 24163 10998 24164 10814 24165 10814 24166 10814 24167 8 24168 11184 24169 8 24170 11185 24171 8 24172 8 24173 10814 24174 10814 24175 8 24176 10814 24177 10814 24178 11186 24179 8 24180 8 24181 8 24182 8 24183 8 24184 8 24185 11187 24186 11188 24187 11189 24188 11190 24189 8 24190 11191 24191 8 24192 11192 24193 11193 24194 11194 24195 11195 24196 11196 24197 8 24198 8 24199 8 24200 8 24201 8 24202 11197 24203 8 24204 8 24205 11198 24206 11199 24207 11200 24208 11201 24209 11202 24210 11203 24211 11204 24212 8 24213 11205 24214 8 24215 11206 24216 11207 24217 11208 24218 11209 24219 11210 24220 8 24221 11211 24222 11212 24223 11213 24224 11214 24225 10814 24226 11215 24227 11216 24228 11217 24229 8 24230 10814 24231 10814 24232 10814 24233 8 24234 11218 24235 8 24236 8 24237 8 24238 11219 24239 11220 24240 11221 24241 8 24242 8 24243 11222 24244 11223 24245 8 24246 11224 24247 11225 24248 8 24249 10814 24250 8 24251 10814 24252 10814 24253 10814 24254 11226 24255 11227 24256 8 24257 10814 24258 10814 24259 10814 24260 11228 24261 8 24262 8 24263 11229 24264 11230 24265 8 24266 11231 24267 8 24268 10814 24269 11232 24270 7842 24271 11233 24272 8 24273 11234 24274 8 24275 11235 24276 11236 24277 8 24278 8 24279 8 24280 11237 24281 8 24282 8 24283 11238 24284 11239 24285 11240 24286 11241 24287 8 24288 8 24289 11242 24290 8 24291 11243 24292 11244 24293 8 24294 8 24295 11245 24296 8 24297 11246 24298 8 24299 11247 24300 11248 24301 11249 24302 10648 24303 10648 24304 11250 24305 11251 24306 11252 24307 11253 24308 11254 24309 11255 24310 11256 24311 11257 24312 11258 24313 11259 24314 8 24315 11260 24316 8 24317 11261 24318 8 24319 8 24320 11262 24321 11263 24322 11264 24323 11265 24324 8 24325 8 24326 11266 24327 11267 24328 11268 24329 11269 24330 11270 24331 8 24332 11271 24333 8 24334 11272 24335 8 24336 8 24337 8 24338 8 24339 11273 24340 8 24341 11274 24342 11275 24343 8 24344 11276 24345 11277 24346 11278 24347 11279 24348 11280 24349 8 24350 11281 24351 11282 24352 8 24353 11283 24354 8 24355 11284 24356 11285 24357 11286 24358 11287 24359 11288 24360 11289 24361 11290 24362 8 24363 8 24364 11291 24365 8 24366 8 24367 11292 24368 11293 24369 11294 24370 8 24371 8 24372 8 24373 8 24374 11295 24375 8 24376 8 24377 8 24378 8 24379 8 24380 11296 24381 11297 24382 11298 24383 11299 24384 11300 24385 11301 24386 11302 24387 11303 24388 11304 24389 11305 24390 11306 24391 11307 24392 11308 24393 11309 24394 11310 24395 11311 24396 11312 24397 11313 24398 11314 24399 11315 24400 11316 24401 11317 24402 11318 24403 11319 24404 11320 24405 11321 24406 11322 24407 11323 24408 11324 24409 11325 24410 11326 24411 11327 24412 11328 24413 11329 24414 11299 24415 11299 24416 11299 24417 11299 24418 11299 24419 11299 24420 11299 24421 11299 24422 11299 24423 11299 24424 11299 24425 11299 24426 11299 24427 11299 24428 11299 24429 11330 24430 11331 24431 8 24432 8 24433 8 24434 8 24435 8 24436 8 24437 11332 24438 8 24439 8 24440 11333 24441 8 24442 11334 24443 11335 24444 11336 24445 11337 24446 11338 24447 8 24448 8 24449 8 24450 8 24451 8 24452 11339 24453 11340 24454 8 24455 8 24456 11341 24457 8 24458 11342 24459 8 24460 8 24461 11343 24462 11344 24463 8 24464 11299 24465 11299 24466 11299 24467 11299 24468 11299 24469 11345 24470 11346 24471 11347 24472 11299 24473 11299 24474 11299 24475 11299 24476 11299 24477 11299 24478 11299 24479 11299 24480 11299 24481 11299 24482 11348 24483 8 24484 8 24485 8 24486 8 24487 8 24488 8 24489 8 24490 8 24491 8 24492 8 24493 8 24494 8 24495 8 24496 8 24497 8 24498 8 24499 8 24500 8 24501 8 24502 8 24503 8 24504 8 24505 8 24506 8 24507 8 24508 8 24509 8 24510 8 24511 8 24512 8 24513 8 24514 8 24515 8 24516 8 24517 8 24518 8 24519 8 24520 8 24521 8 24522 8 24523 8 24524 8 24525 8 24526 8 24527 8 24528 8 24529 8 24530 8 24531 8 24532 8 24533 8 24534 8 24535 8 24536 8 24537 8 24538 8 24539 8 24540 8 24541 8 24542 8 24543 8 24544 8 24545 8 24546 8 24547 8 24548 8 24549 8 24550 8 24551 8 24552 8 24553 8 24554 8 24555 8 24556 8 24557 8 24558 8 24559 8 24560 8 24561 8 24562 8 24563 11349 24564 8 24565 8 24566 8 24567 11350 24568 11351 24569 8 24570 11352 24571 8 24572 8 24573 11353 24574 11354 24575 11355 24576 8 24577 8 24578 11356 24579 11357 24580 11358 24581 11359 24582 8 24583 11360 24584 8 24585 11361 24586 11362 24587 11363 24588 11364 24589 11365 24590 11366 24591 11367 24592 11368 24593 11369 24594 11299 24595 11299 24596 11299 24597 11299 24598 11299 24599 11299 24600 11299 24601 11299 24602 11299 24603 11299 24604 11299 24605 11299 24606 11299 24607 11370 24608 11299 24609 11371 24610 11372 24611 8 24612 11373 24613 11374 24614 8 24615 11375 24616 8 24617 11376 24618 8 24619 8 24620 8 24621 11377 24622 11378 24623 11379 24624 8 24625 11380 24626 8 24627 8 24628 11381 24629 11382 24630 11383 24631 8 24632 8 24633 8 24634 11384 24635 11385 24636 8 24637 8 24638 8 24639 11386 24640 11387 24641 11388 24642 11389 24643 8 24644 8 24645 11390 24646 8 24647 11391 24648 11392 24649 8 24650 8 24651 11393 24652 8 24653 8 24654 8 24655 11394 24656 11395 24657 8 24658 8 24659 11396 24660 11397 24661 11398 24662 11399 24663 11400 24664 11401 24665 11402 24666 11403 24667 11404 24668 11405 24669 11406 24670 8 24671 11407 24672 11408 24673 11409 24674 11410 24675 11411 24676 11412 24677 11413 24678 11414 24679 11415 24680 11416 24681 8 24682 11417 24683 10648 24684 11418 24685 8 24686 11419 24687 8 24688 11420 24689 11421 24690 8 24691 8 24692 8 24693 11422 24694 8 24695 8 24696 11423 24697 11424 24698 11425 24699 11426 24700 8 24701 8 24702 11427 24703 8 24704 11428 24705 11429 24706 8 24707 8 24708 11430 24709 8 24710 11431 24711 8 24712 11432 24713 11433 24714 11434 24715 11299 24716 11299 24717 11435 24718 11436 24719 11299 24720 11299 24721 11299 24722 11299 24723 8 24724 11437 24725 8 24726 11438 24727 8 24728 8 24729 11299 24730 11299 24731 8 24732 11299 24733 11299 24734 11439 24735 8 24736 8 24737 8 24738 8 24739 8 24740 8 24741 11440 24742 11441 24743 11442 24744 11443 24745 8 24746 11444 24747 8 24748 11445 24749 11446 24750 11447 24751 11448 24752 11449 24753 8 24754 8 24755 8 24756 8 24757 8 24758 11450 24759 8 24760 8 24761 11451 24762 11452 24763 11453 24764 11454 24765 11455 24766 11456 24767 11457 24768 8 24769 11458 24770 8 24771 11459 24772 11460 24773 11461 24774 11462 24775 11463 24776 8 24777 11464 24778 11465 24779 11466 24780 11467 24781 11299 24782 11468 24783 11469 24784 11470 24785 8 24786 11299 24787 11299 24788 11299 24789 8 24790 11471 24791 8 24792 8 24793 8 24794 11472 24795 11473 24796 11474 24797 8 24798 8 24799 11475 24800 11476 24801 8 24802 11477 24803 11478 24804 8 24805 11299 24806 8 24807 11299 24808 11299 24809 11299 24810 11479 24811 11480 24812 8 24813 11299 24814 11299 24815 11299 24816 11481 24817 8 24818 8 24819 11482 24820 11483 24821 8 24822 11484 24823 8 24824 11299 24825 10648 24826 10648 24827 10648 24828 8 24829 11485 24830 8 24831 11486 24832 8 24833 8 24834 10648 24835 10648 24836 8 24837 10648 24838 10648 24839 11487 24840 8 24841 8 24842 8 24843 8 24844 8 24845 8 24846 11488 24847 11489 24848 11490 24849 11491 24850 8 24851 11492 24852 8 24853 11493 24854 11494 24855 11495 24856 11496 24857 11497 24858 8 24859 8 24860 8 24861 8 24862 8 24863 11498 24864 8 24865 8 24866 11499 24867 11500 24868 11501 24869 11502 24870 11503 24871 11504 24872 11505 24873 8 24874 11506 24875 8 24876 11507 24877 11508 24878 11509 24879 11510 24880 11511 24881 8 24882 11512 24883 11513 24884 11514 24885 11515 24886 10648 24887 11516 24888 11517 24889 11518 24890 8 24891 10648 24892 10648 24893 10648 24894 8 24895 11519 24896 8 24897 8 24898 8 24899 11520 24900 11521 24901 11522 24902 8 24903 8 24904 11523 24905 11524 24906 8 24907 11525 24908 11526 24909 8 24910 10648 24911 8 24912 10648 24913 10648 24914 10648 24915 11527 24916 11528 24917 8 24918 10648 24919 10648 24920 10648 24921 11529 24922 8 24923 8 24924 11530 24925 11531 24926 8 24927 11532 24928 8 24929 10648 24930 11533 24931 11534 24932 8 24933 8 24934 11535 24935 8 24936 11536 24937 8 24938 11537 24939 8 24940 11538 24941 8 24942 11539 24943 11540 24944 11541 24945 11542 24946 11543 24947 11544 24948 11545 24949 11546 24950 11547 24951 8 24952 8 24953 11548 24954 11549 24955 8 24956 8 24957 11550 24958 11551 24959 11552 24960 11553 24961 11554 24962 11555 24963 11556 24964 11557 24965 8 24966 8 24967 11558 24968 11559 24969 8 24970 11560 24971 8 24972 11561 24973 11562 24974 11563 24975 11564 24976 11565 24977 8 24978 11566 24979 11567 24980 11568 24981 11569 24982 8 24983 8 24984 11570 24985 11571 24986 8 24987 11572 24988 8 24989 11573 24990 8 24991 8 24992 11574 24993 11575 24994 11576 24995 11577 24996 11578 24997 11579 24998 11580 24999 11581 25000 11582 25001 11583 25002 11584 25003 8 25004 8 25005 11585 25006 11586 25007 11587 25008 11588 25009 11589 25010 11590 25011 11591 25012 8 25013 8 25014 11592 25015 11593 25016 8 25017 11594 25018 11595 25019 11596 25020 8 25021 11597 25022 11598 25023 11599 25024 11600 25025 8 25026 8 25027 11601 25028 11602 25029 8 25030 11603 25031 11604 25032 11605 25033 11606 25034 8 25035 11607 25036 11608 25037 11609 25038 11610 25039 11611 25040 11612 25041 8 25042 11613 25043 11614 25044 8 25045 11615 25046 11616 25047 11617 25048 11618 25049 11619 25050 11620 25051 8 25052 11621 25053 11622 25054 11623 25055 11624 25056 8 25057 11625 25058 11626 25059 11627 25060 11628 25061 8 25062 11629 25063 11630 25064 11631 25065 8 25066 11632 25067 8 25068 11633 25069 11634 25070 11635 25071 8 25072 11636 25073 11637 25074 8 25075 8 25076 11638 25077 8 25078 11639 25079 8 25080 11640 25081 11641 25082 11642 25083 11643 25084 11644 25085 8 25086 11645 25087 11646 25088 11647 25089 8 25090 11648 25091 11649 25092 8 25093 11650 25094 8 25095 11651 25096 8 25097 11652 25098 11653 25099 11654 25100 11655 25101 11656 25102 11657 25103 11658 25104 11659 25105 11660 25106 8 25107 8 25108 11661 25109 11662 25110 8 25111 8 25112 11663 25113 11664 25114 11665 25115 11666 25116 11667 25117 11668 25118 11669 25119 11670 25120 8 25121 8 25122 11671 25123 11672 25124 8 25125 11673 25126 11674 25127 11675 25128 11676 25129 11677 25130 11678 25131 8 25132 11679 25133 8 25134 11680 25135 11681 25136 11682 25137 11683 25138 11684 25139 8 25140 11685 25141 11686 25142 11687 25143 11688 25144 8 25145 8 25146 11689 25147 11690 25148 8 25149 11691 25150 8 25151 11692 25152 11693 25153 11694 25154 11695 25155 11696 25156 11697 25157 8 25158 11698 25159 8 25160 11699 25161 11700 25162 7741 25163 11701 25164 8 25165 11702 25166 11703 25167 11704 25168 8 25169 11705 25170 11706 25171 8 25172 8 25173 11707 25174 8 25175 11708 25176 8 25177 11709 25178 8 25179 11710 25180 8 25181 11711 25182 11712 25183 11713 25184 11714 25185 11715 25186 11716 25187 11717 25188 11718 25189 11719 25190 8 25191 8 25192 11720 25193 11721 25194 8 25195 8 25196 11722 25197 11723 25198 11724 25199 11725 25200 11726 25201 11727 25202 11728 25203 11729 25204 8 25205 8 25206 11730 25207 11731 25208 8 25209 11732 25210 8 25211 11733 25212 11734 25213 11735 25214 11736 25215 11737 25216 8 25217 11738 25218 11739 25219 11740 25220 11741 25221 8 25222 8 25223 11742 25224 11743 25225 8 25226 11744 25227 8 25228 11745 25229 8 25230 8 25231 11746 25232 11747 25233 11748 25234 11749 25235 11750 25236 11751 25237 11752 25238 11753 25239 11754 25240 11755 25241 11756 25242 8 25243 8 25244 11757 25245 11758 25246 11759 25247 11760 25248 11761 25249 11762 25250 11763 25251 8 25252 8 25253 11764 25254 11765 25255 8 25256 11766 25257 11767 25258 11768 25259 8 25260 11769 25261 11770 25262 11771 25263 11772 25264 8 25265 8 25266 11773 25267 11774 25268 8 25269 11775 25270 11776 25271 11777 25272 11778 25273 8 25274 11779 25275 11780 25276 11781 25277 11782 25278 11783 25279 11784 25280 8 25281 11785 25282 11786 25283 8 25284 11787 25285 11788 25286 11789 25287 11790 25288 11791 25289 11792 25290 8 25291 11793 25292 11794 25293 11795 25294 11796 25295 8 25296 11797 25297 11798 25298 11799 25299 11800 25300 8 25301 11801 25302 11802 25303 11803 25304 8 25305 11804 25306 8 25307 11805 25308 11806 25309 11807 25310 8 25311 11808 25312 11809 25313 8 25314 8 25315 11810 25316 8 25317 11811 25318 8 25319 11812 25320 11813 25321 11814 25322 11815 25323 11816 25324 8 25325 11817 25326 11818 25327 11819 25328 8 25329 11820 25330 11821 25331 8 25332 11822 25333 8 25334 11823 25335 8 25336 11824 25337 11825 25338 11826 25339 11827 25340 11828 25341 11829 25342 11830 25343 11831 25344 11832 25345 8 25346 8 25347 11833 25348 11834 25349 8 25350 8 25351 11835 25352 11836 25353 11837 25354 11838 25355 11839 25356 11840 25357 11841 25358 11842 25359 8 25360 8 25361 11843 25362 11844 25363 8 25364 11845 25365 11846 25366 11847 25367 11848 25368 11849 25369 11850 25370 8 25371 8 25372 11851 25373 8 25374 11852 25375 11853 25376 11854 25377 11855 25378 11856 25379 8 25380 11857 25381 11858 25382 11859 25383 11860 25384 8 25385 8 25386 11861 25387 11862 25388 8 25389 11863 25390 8 25391 11864 25392 11865 25393 11866 25394 11867 25395 11868 25396 11869 25397 8 25398 11870 25399 8 25400 11871 25401 11872 25402 11873 25403 11874 25404 8 25405 11875 25406 11876 25407 11877 25408 11878 25409 11879 25410 11880 25411 8 25412 11881 25413 11882 25414 8 25415 11883 25416 11884 25417 11885 25418 11886 25419 11887 25420 11888 25421 8 25422 11889 25423 11890 25424 11891 25425 11892 25426 8 25427 11893 25428 11894 25429 11895 25430 11896 25431 8 25432 11897 25433 11898 25434 11899 25435 8 25436 11900 25437 8 25438 11901 25439 11902 25440 11903 25441 8 25442 11904 25443 11905 25444 8 25445 8 25446 11906 25447 8 25448 11907 25449 8 25450 11908 25451 11909 25452 11910 25453 11911 25454 11912 25455 8 25456 11913 25457 11914 25458 8 25459 11915 25460 8 25461 11916 25462 8 25463 11917 25464 11918 25465 11919 25466 11920 25467 11921 25468 11922 25469 11923 25470 11924 25471 11925 25472 8 25473 8 25474 11926 25475 11927 25476 8 25477 8 25478 11928 25479 11929 25480 11930 25481 11931 25482 11932 25483 11933 25484 11934 25485 11935 25486 8 25487 8 25488 11936 25489 11937 25490 8 25491 11938 25492 11939 25493 11940 25494 11941 25495 11942 25496 11943 25497 8 25498 8 25499 11944 25500 8 25501 11945 25502 11946 25503 11947 25504 11948 25505 11949 25506 8 25507 11950 25508 11951 25509 11952 25510 11953 25511 8 25512 8 25513 11954 25514 11955 25515 8 25516 11956 25517 8 25518 11957 25519 11958 25520 11959 25521 11960 25522 11961 25523 11962 25524 8 25525 11963 25526 8 25527 11964 25528 11965 25529 11966 25530 11967 25531 8 25532 8 25533 11968 25534 11969 25535 8 25536 11970 25537 11971 25538 11972 25539 11973 25540 8 25541 11974 25542 11975 25543 11976 25544 11977 25545 11978 25546 11979 25547 8 25548 11980 25549 11981 25550 8 25551 11982 25552 11983 25553 11984 25554 11985 25555 11986 25556 11987 25557 8 25558 11988 25559 11989 25560 11990 25561 11991 25562 8 25563 11992 25564 11993 25565 11994 25566 11995 25567 8 25568 11996 25569 11997 25570 11998 25571 8 25572 11999 25573 8 25574 12000 25575 12001 25576 12002 25577 8 25578 12003 25579 12004 25580 8 25581 8 25582 12005 25583 8 25584 12006 25585 8 25586 12007 25587 12008 25588 12009 25589 12010 25590 12011 25591 8 25592 12012 25593 12013 25594 12014 25595 8 25596 12015 25597 12016 25598 8 25599 12017 25600 8 25601 12018 25602 8 25603 12019 25604 12020 25605 12021 25606 12022 25607 12023 25608 12024 25609 12025 25610 12026 25611 12027 25612 8 25613 8 25614 12028 25615 12029 25616 8 25617 8 25618 12030 25619 12031 25620 12032 25621 12033 25622 12034 25623 12035 25624 12036 25625 12037 25626 8 25627 8 25628 12038 25629 12039 25630 8 25631 12040 25632 12041 25633 12042 25634 12043 25635 12044 25636 12045 25637 8 25638 12046 25639 8 25640 12047 25641 12048 25642 12049 25643 12050 25644 12051 25645 8 25646 12052 25647 12053 25648 12054 25649 12055 25650 8 25651 8 25652 12056 25653 12057 25654 8 25655 12058 25656 8 25657 12059 25658 12060 25659 12061 25660 12062 25661 12063 25662 12064 25663 8 25664 12065 25665 8 25666 12066 25667 12067 25668 12068 25669 8 25670 12069 25671 12070 25672 12071 25673 12072 25674 8 25675 8 25676 12073 25677 12074 25678 8 25679 12075 25680 12076 25681 12077 25682 12078 25683 8 25684 12079 25685 12080 25686 12081 25687 12082 25688 12083 25689 12084 25690 8 25691 12085 25692 12086 25693 8 25694 12087 25695 12088 25696 12089 25697 12090 25698 12091 25699 12092 25700 8 25701 12093 25702 12094 25703 12095 25704 12096 25705 8 25706 12097 25707 12098 25708 12099 25709 8 25710 12100 25711 8 25712 12101 25713 12102 25714 12103 25715 8 25716 12104 25717 12105 25718 8 25719 8 25720 12106 25721 8 25722 12107 25723 8 25724 12108 25725 8 25726 12109 25727 8 25728 12110 25729 12111 25730 12112 25731 12113 25732 12114 25733 12115 25734 12116 25735 12117 25736 12118 25737 8 25738 8 25739 12119 25740 12120 25741 8 25742 8 25743 12121 25744 12122 25745 12123 25746 12124 25747 12125 25748 12126 25749 12127 25750 12128 25751 8 25752 8 25753 12129 25754 12130 25755 8 25756 12131 25757 8 25758 12132 25759 12133 25760 12134 25761 12135 25762 12136 25763 8 25764 12137 25765 12138 25766 12139 25767 12140 25768 8 25769 8 25770 12141 25771 12142 25772 8 25773 12143 25774 8 25775 12144 25776 8 25777 12145 25778 8 25779 12146 25780 8 25781 12147 25782 12148 25783 8 25784 8 25785 8 25786 12149 25787 8 25788 8 25789 12150 25790 12151 25791 12152 25792 12153 25793 8 25794 8 25795 12154 25796 8 25797 12155 25798 12156 25799 8 25800 8 25801 12157 25802 8 25803 12158 25804 8 25805 12159 25806 12160 25807 12161 25808 186 25809 186 25810 12162 25811 12163 25812 12164 25813 12165 25814 12166 25815 12167 25816 12168 25817 12169 25818 12170 25819 12171 25820 8 25821 12172 25822 8 25823 12173 25824 8 25825 8 25826 12174 25827 12175 25828 12176 25829 12177 25830 8 25831 8 25832 12178 25833 12179 25834 12180 25835 12181 25836 12182 25837 8 25838 12183 25839 8 25840 12184 25841 8 25842 8 25843 8 25844 8 25845 12185 25846 8 25847 12186 25848 12187 25849 8 25850 12188 25851 12189 25852 12190 25853 12191 25854 12192 25855 8 25856 12193 25857 12194 25858 8 25859 12195 25860 8 25861 12196 25862 12197 25863 12198 25864 12199 25865 12200 25866 12201 25867 12202 25868 8 25869 8 25870 12203 25871 8 25872 8 25873 12204 25874 12205 25875 12206 25876 8 25877 8 25878 8 25879 8 25880 12207 25881 8 25882 8 25883 8 25884 8 25885 8 25886 12208 25887 12209 25888 12210 25889 12211 25890 12212 25891 12213 25892 12214 25893 12215 25894 12216 25895 12217 25896 12218 25897 12219 25898 12220 25899 12221 25900 12222 25901 12223 25902 12224 25903 12225 25904 12226 25905 12227 25906 12228 25907 12229 25908 12230 25909 12231 25910 12232 25911 12233 25912 12234 25913 12235 25914 12236 25915 12237 25916 12238 25917 12239 25918 12240 25919 12241 25920 12211 25921 12211 25922 12211 25923 12211 25924 12211 25925 12211 25926 12211 25927 12211 25928 12211 25929 12211 25930 12211 25931 12211 25932 12211 25933 12211 25934 12211 25935 12242 25936 12243 25937 8 25938 8 25939 8 25940 8 25941 8 25942 8 25943 12244 25944 8 25945 8 25946 12245 25947 8 25948 12246 25949 12247 25950 12248 25951 12249 25952 12250 25953 8 25954 8 25955 8 25956 8 25957 8 25958 12251 25959 12252 25960 8 25961 8 25962 12253 25963 8 25964 12254 25965 8 25966 8 25967 12255 25968 12256 25969 8 25970 12211 25971 12211 25972 12211 25973 12211 25974 12211 25975 12257 25976 12258 25977 12259 25978 12211 25979 12211 25980 12211 25981 12211 25982 12211 25983 12211 25984 12211 25985 12211 25986 12211 25987 12211 25988 12260 25989 8 25990 8 25991 8 25992 8 25993 8 25994 8 25995 8 25996 8 25997 8 25998 8 25999 8 26000 8 26001 8 26002 8 26003 8 26004 8 26005 8 26006 8 26007 8 26008 8 26009 8 26010 8 26011 8 26012 8 26013 8 26014 8 26015 8 26016 8 26017 8 26018 8 26019 8 26020 8 26021 8 26022 8 26023 8 26024 8 26025 8 26026 8 26027 8 26028 8 26029 8 26030 8 26031 8 26032 8 26033 8 26034 8 26035 8 26036 8 26037 8 26038 8 26039 8 26040 8 26041 8 26042 8 26043 8 26044 8 26045 8 26046 8 26047 8 26048 8 26049 8 26050 8 26051 8 26052 8 26053 8 26054 8 26055 8 26056 8 26057 8 26058 8 26059 8 26060 8 26061 8 26062 8 26063 8 26064 8 26065 8 26066 8 26067 8 26068 8 26069 12261 26070 8 26071 8 26072 8 26073 12262 26074 12263 26075 8 26076 12264 26077 8 26078 8 26079 12265 26080 12266 26081 12267 26082 8 26083 8 26084 12268 26085 12269 26086 12270 26087 12271 26088 8 26089 12272 26090 8 26091 12273 26092 12274 26093 12275 26094 12276 26095 12277 26096 12278 26097 12279 26098 12280 26099 12281 26100 12211 26101 12211 26102 12211 26103 12211 26104 12211 26105 12211 26106 12211 26107 12211 26108 12211 26109 12211 26110 12211 26111 12211 26112 12211 26113 12282 26114 12211 26115 12283 26116 12284 26117 8 26118 12285 26119 12286 26120 8 26121 12287 26122 8 26123 12288 26124 8 26125 8 26126 8 26127 12289 26128 12290 26129 12291 26130 8 26131 12292 26132 8 26133 8 26134 12293 26135 12294 26136 12295 26137 8 26138 8 26139 8 26140 12296 26141 12297 26142 8 26143 8 26144 8 26145 12298 26146 12299 26147 12300 26148 12301 26149 8 26150 8 26151 12302 26152 8 26153 12303 26154 12304 26155 8 26156 8 26157 12305 26158 8 26159 8 26160 8 26161 12306 26162 12307 26163 8 26164 8 26165 12308 26166 12309 26167 12310 26168 12311 26169 12312 26170 12313 26171 12314 26172 12315 26173 12316 26174 12317 26175 12318 26176 8 26177 12319 26178 12320 26179 12321 26180 12322 26181 12323 26182 12324 26183 12325 26184 12326 26185 12327 26186 12328 26187 8 26188 12329 26189 186 26190 12330 26191 8 26192 12331 26193 8 26194 12332 26195 12333 26196 8 26197 8 26198 8 26199 12334 26200 8 26201 8 26202 12335 26203 12336 26204 12337 26205 12338 26206 8 26207 8 26208 12339 26209 8 26210 12340 26211 12341 26212 8 26213 8 26214 12342 26215 8 26216 12343 26217 8 26218 12344 26219 12345 26220 12346 26221 12211 26222 12211 26223 12347 26224 12348 26225 12211 26226 12211 26227 12211 26228 12211 26229 8 26230 12349 26231 8 26232 12350 26233 8 26234 8 26235 12211 26236 12211 26237 8 26238 12211 26239 12211 26240 12351 26241 8 26242 8 26243 8 26244 8 26245 8 26246 8 26247 12352 26248 12353 26249 12354 26250 12355 26251 8 26252 12356 26253 8 26254 12357 26255 12358 26256 12359 26257 12360 26258 12361 26259 8 26260 8 26261 8 26262 8 26263 8 26264 12362 26265 8 26266 8 26267 12363 26268 12364 26269 12365 26270 12366 26271 12367 26272 12368 26273 12369 26274 8 26275 12370 26276 8 26277 12371 26278 12372 26279 12373 26280 12374 26281 12375 26282 8 26283 12376 26284 12377 26285 12378 26286 12379 26287 12211 26288 12380 26289 12381 26290 12382 26291 8 26292 12211 26293 12211 26294 12211 26295 8 26296 12383 26297 8 26298 8 26299 8 26300 12384 26301 12385 26302 12386 26303 8 26304 8 26305 12387 26306 12388 26307 8 26308 12389 26309 12390 26310 8 26311 12211 26312 8 26313 12211 26314 12211 26315 12211 26316 12391 26317 12392 26318 8 26319 12211 26320 12211 26321 12211 26322 12393 26323 8 26324 8 26325 12394 26326 12395 26327 8 26328 12396 26329 8 26330 12211 26331 186 26332 186 26333 186 26334 8 26335 12397 26336 8 26337 12398 26338 8 26339 8 26340 186 26341 186 26342 8 26343 186 26344 186 26345 12399 26346 8 26347 8 26348 8 26349 8 26350 8 26351 8 26352 12400 26353 12401 26354 12402 26355 12403 26356 8 26357 12404 26358 8 26359 12405 26360 12406 26361 12407 26362 12408 26363 12409 26364 8 26365 8 26366 8 26367 8 26368 8 26369 12410 26370 8 26371 8 26372 12411 26373 12412 26374 12413 26375 12414 26376 12415 26377 12416 26378 12417 26379 8 26380 12418 26381 8 26382 12419 26383 12420 26384 12421 26385 12422 26386 12423 26387 8 26388 12424 26389 12425 26390 12426 26391 12427 26392 186 26393 12428 26394 12429 26395 12430 26396 8 26397 186 26398 186 26399 186 26400 8 26401 12431 26402 8 26403 8 26404 8 26405 12432 26406 12433 26407 12434 26408 8 26409 8 26410 12435 26411 12436 26412 8 26413 12437 26414 12438 26415 8 26416 186 26417 8 26418 186 26419 186 26420 186 26421 12439 26422 12440 26423 8 26424 186 26425 186 26426 186 26427 12441 26428 8 26429 8 26430 12442 26431 12443 26432 8 26433 12444 26434 8 26435 186 26436 8 26437 12445 26438 12446 26439 12447 26440 12448 26441 12449 26442 12450 26443 12451 26444 12452 26445 12453 26446 12454 26447 12455 26448 8 26449 8 26450 12456 26451 12457 26452 12458 26453 12459 26454 12460 26455 12461 26456 12462 26457 8 26458 8 26459 12463 26460 12464 26461 8 26462 12465 26463 12466 26464 12467 26465 8 26466 12468 26467 12469 26468 12470 26469 12471 26470 8 26471 8 26472 12472 26473 12473 26474 8 26475 12474 26476 12475 26477 12476 26478 12477 26479 8 26480 12478 26481 12479 26482 12480 26483 12481 26484 12482 26485 12483 26486 8 26487 12484 26488 12485 26489 8 26490 12486 26491 12487 26492 12488 26493 12489 26494 12490 26495 12491 26496 8 26497 12492 26498 12493 26499 12494 26500 12495 26501 8 26502 12496 26503 12497 26504 12498 26505 12499 26506 8 26507 12500 26508 12501 26509 12502 26510 8 26511 12503 26512 8 26513 12504 26514 12505 26515 12506 26516 8 26517 12507 26518 12508 26519 8 26520 8 26521 12509 26522 8 26523 12510 26524 8 26525 12511 26526 12512 26527 12513 26528 12514 26529 12515 26530 8 26531 12516 26532 12517 26533 12518 26534 8 26535 12519 26536 12520 26537 8 26538 12521 26539 8 26540 12522 26541 8 26542 12523 26543 12524 26544 12525 26545 12526 26546 12527 26547 12528 26548 12529 26549 12530 26550 12531 26551 8 26552 8 26553 12532 26554 12533 26555 8 26556 8 26557 12534 26558 12535 26559 12536 26560 12537 26561 12538 26562 12539 26563 12540 26564 12541 26565 8 26566 8 26567 12542 26568 12543 26569 8 26570 12544 26571 12545 26572 12546 26573 12547 26574 12548 26575 12549 26576 8 26577 12550 26578 8 26579 12551 26580 12552 26581 12553 26582 12554 26583 12555 26584 8 26585 12556 26586 12557 26587 12558 26588 12559 26589 8 26590 8 26591 12560 26592 12561 26593 8 26594 12562 26595 8 26596 12563 26597 12564 26598 12565 26599 12566 26600 12567 26601 12568 26602 8 26603 12569 26604 8 26605 12570 26606 8 26607 8 26608 8 26609 8 26610 8 26611 8 26612 8 26613 12571 26614 8 26615 12572 26616 8 26617 8 26618 12573 26619 12574 26620 12575 26621 12576 26622 8 26623 8 26624 8 26625 8 26626 8 26627 8 26628 8 26629 8 26630 8 26631 8 26632 8 26633 12577 26634 8 26635 12578 26636 8 26637 12579 26638 8 26639 8 26640 8 26641 8 26642 8 26643 8 26644 8 26645 8 26646 8 26647 8 26648 12580 26649 8 26650 8 26651 8 26652 8 26653 8 26654 8 26655 8 26656 8 26657 8 26658 8 26659 8 26660 12581 26661 8 26662 8 26663 8 26664 12582 26665 8 26666 12583 26667 12584 26668 12585 26669 8 26670 8 26671 8 26672 12586 26673 8 26674 8 26675 8 26676 8 26677 8 26678 8 26679 12587 26680 8 26681 12588 26682 8 26683 12589 26684 12590 26685 12591 26686 12592 26687 12593 26688 8 26689 8 26690 8 26691 8 26692 8 26693 8 26694 8 26695 8 26696 8 26697 12594 26698 8 26699 12595 26700 8 26701 12596 26702 8 26703 8 26704 8 26705 8 26706 8 26707 8 26708 8 26709 8 26710 8 26711 12597 26712 8 26713 12598 26714 12599 26715 12600 26716 12601 26717 8 26718 12602 26719 12603 26720 8 26721 12604 26722 12605 26723 12606 26724 12607 26725 12608 26726 12609 26727 12610 26728 12611 26729 12612 26730 12613 26731 8 26732 12614 26733 8 26734 12615 26735 8 26736 8 26737 12616 26738 12617 26739 12618 26740 12619 26741 8 26742 8 26743 12620 26744 12621 26745 12622 26746 12623 26747 12624 26748 8 26749 12625 26750 8 26751 12626 26752 8 26753 8 26754 8 26755 8 26756 12627 26757 8 26758 12628 26759 12629 26760 8 26761 12630 26762 12631 26763 12632 26764 12633 26765 12634 26766 8 26767 12635 26768 12636 26769 8 26770 12637 26771 8 26772 12638 26773 12639 26774 12640 26775 12641 26776 12642 26777 12643 26778 12644 26779 8 26780 8 26781 12645 26782 8 26783 8 26784 12646 26785 12647 26786 12648 26787 8 26788 8 26789 8 26790 8 26791 12649 26792 8 26793 8 26794 8 26795 8 26796 8 26797 12650 26798 12651 26799 12652 26800 12653 26801 12654 26802 12655 26803 12656 26804 12657 26805 12658 26806 12659 26807 12660 26808 12661 26809 12662 26810 12663 26811 12664 26812 12665 26813 12666 26814 12667 26815 12668 26816 12669 26817 12670 26818 12671 26819 12672 26820 12673 26821 12674 26822 12675 26823 12676 26824 12677 26825 12678 26826 12679 26827 12680 26828 12681 26829 12682 26830 12683 26831 12653 26832 12653 26833 12653 26834 12653 26835 12653 26836 12653 26837 12653 26838 12653 26839 12653 26840 12653 26841 12653 26842 12653 26843 12653 26844 12653 26845 12653 26846 12684 26847 12685 26848 8 26849 8 26850 8 26851 8 26852 8 26853 8 26854 12686 26855 8 26856 8 26857 12687 26858 8 26859 12688 26860 12689 26861 12690 26862 12691 26863 12692 26864 8 26865 8 26866 8 26867 8 26868 8 26869 12693 26870 12694 26871 8 26872 8 26873 12695 26874 8 26875 12696 26876 8 26877 8 26878 12697 26879 12698 26880 8 26881 12653 26882 12653 26883 12653 26884 12653 26885 12653 26886 12699 26887 12700 26888 12701 26889 12653 26890 12653 26891 12653 26892 12653 26893 12653 26894 12653 26895 12653 26896 12653 26897 12653 26898 12653 26899 12702 26900 8 26901 8 26902 8 26903 8 26904 8 26905 8 26906 8 26907 8 26908 8 26909 8 26910 8 26911 8 26912 8 26913 8 26914 8 26915 8 26916 8 26917 8 26918 8 26919 8 26920 8 26921 8 26922 8 26923 8 26924 8 26925 8 26926 8 26927 8 26928 8 26929 8 26930 8 26931 8 26932 8 26933 8 26934 8 26935 8 26936 8 26937 8 26938 8 26939 8 26940 8 26941 8 26942 8 26943 8 26944 8 26945 8 26946 8 26947 8 26948 8 26949 8 26950 8 26951 8 26952 8 26953 8 26954 8 26955 8 26956 8 26957 8 26958 8 26959 8 26960 8 26961 8 26962 8 26963 8 26964 8 26965 8 26966 8 26967 8 26968 8 26969 8 26970 8 26971 8 26972 8 26973 8 26974 8 26975 8 26976 8 26977 8 26978 8 26979 8 26980 12703 26981 8 26982 8 26983 8 26984 12704 26985 12705 26986 8 26987 12706 26988 8 26989 8 26990 12707 26991 12708 26992 12709 26993 8 26994 8 26995 12710 26996 12711 26997 12712 26998 12713 26999 8 27000 12714 27001 8 27002 12715 27003 12716 27004 12717 27005 12718 27006 12719 27007 12720 27008 12721 27009 12722 27010 12723 27011 12653 27012 12653 27013 12653 27014 12653 27015 12653 27016 12653 27017 12653 27018 12653 27019 12653 27020 12653 27021 12653 27022 12653 27023 12653 27024 12724 27025 12653 27026 12725 27027 12726 27028 8 27029 12727 27030 12728 27031 8 27032 12729 27033 8 27034 12730 27035 8 27036 8 27037 8 27038 12731 27039 12732 27040 12733 27041 8 27042 12734 27043 8 27044 8 27045 12735 27046 12736 27047 12737 27048 8 27049 8 27050 8 27051 12738 27052 12739 27053 8 27054 8 27055 8 27056 12740 27057 12741 27058 12742 27059 12743 27060 8 27061 8 27062 12744 27063 8 27064 12745 27065 12746 27066 8 27067 8 27068 12747 27069 8 27070 8 27071 8 27072 12748 27073 12749 27074 8 27075 8 27076 12750 27077 12751 27078 12752 27079 12753 27080 12754 27081 12755 27082 12756 27083 12757 27084 12758 27085 12759 27086 12760 27087 8 27088 12761 27089 12762 27090 12763 27091 12764 27092 12765 27093 12766 27094 12767 27095 12768 27096 12769 27097 12770 27098 8 27099 8 27100 12771 27101 8 27102 12772 27103 8 27104 12773 27105 12774 27106 8 27107 8 27108 8 27109 12775 27110 8 27111 8 27112 12776 27113 12777 27114 12778 27115 12779 27116 8 27117 8 27118 12780 27119 8 27120 12781 27121 12782 27122 8 27123 8 27124 12783 27125 8 27126 12784 27127 8 27128 12785 27129 12786 27130 12787 27131 12653 27132 12653 27133 12788 27134 12789 27135 12790 27136 12791 27137 12792 27138 12793 27139 12794 27140 12795 27141 12796 27142 12797 27143 8 27144 12798 27145 8 27146 12799 27147 8 27148 8 27149 12800 27150 12801 27151 12802 27152 12803 27153 8 27154 8 27155 12804 27156 12805 27157 12806 27158 12807 27159 12808 27160 8 27161 12809 27162 8 27163 12810 27164 8 27165 8 27166 8 27167 8 27168 12811 27169 8 27170 12812 27171 12813 27172 8 27173 12814 27174 12815 27175 12816 27176 12817 27177 12818 27178 8 27179 12819 27180 12820 27181 8 27182 12821 27183 8 27184 12822 27185 12823 27186 12824 27187 12825 27188 12826 27189 12827 27190 12828 27191 8 27192 8 27193 12829 27194 8 27195 8 27196 12830 27197 12831 27198 12832 27199 8 27200 8 27201 8 27202 8 27203 12833 27204 8 27205 8 27206 8 27207 8 27208 8 27209 12834 27210 12835 27211 12836 27212 12837 27213 12838 27214 12839 27215 12840 27216 12841 27217 12842 27218 12843 27219 12844 27220 12845 27221 12846 27222 12847 27223 12848 27224 12849 27225 12850 27226 12851 27227 12852 27228 12853 27229 12854 27230 12855 27231 12856 27232 12857 27233 12858 27234 12859 27235 12860 27236 12861 27237 12862 27238 12863 27239 12864 27240 12865 27241 12866 27242 12867 27243 12837 27244 12837 27245 12837 27246 12837 27247 12837 27248 12837 27249 12837 27250 12837 27251 12837 27252 12837 27253 12837 27254 12837 27255 12837 27256 12837 27257 12837 27258 12868 27259 12869 27260 8 27261 8 27262 8 27263 8 27264 8 27265 8 27266 12870 27267 8 27268 8 27269 12871 27270 8 27271 12872 27272 12873 27273 12874 27274 12875 27275 12876 27276 8 27277 8 27278 8 27279 8 27280 8 27281 12877 27282 12878 27283 8 27284 8 27285 12879 27286 8 27287 12880 27288 8 27289 8 27290 12881 27291 12882 27292 8 27293 12837 27294 12837 27295 12837 27296 12837 27297 12837 27298 12883 27299 12884 27300 12885 27301 12837 27302 12837 27303 12837 27304 12837 27305 12837 27306 12837 27307 12837 27308 12837 27309 12837 27310 12837 27311 12886 27312 8 27313 8 27314 8 27315 8 27316 8 27317 8 27318 8 27319 8 27320 8 27321 8 27322 8 27323 8 27324 8 27325 8 27326 8 27327 8 27328 8 27329 8 27330 8 27331 8 27332 8 27333 8 27334 8 27335 8 27336 8 27337 8 27338 8 27339 8 27340 8 27341 8 27342 8 27343 8 27344 8 27345 8 27346 8 27347 8 27348 8 27349 8 27350 8 27351 8 27352 8 27353 8 27354 8 27355 8 27356 8 27357 8 27358 8 27359 8 27360 8 27361 8 27362 8 27363 8 27364 8 27365 8 27366 8 27367 8 27368 8 27369 8 27370 8 27371 8 27372 8 27373 8 27374 8 27375 8 27376 8 27377 8 27378 8 27379 8 27380 8 27381 8 27382 8 27383 8 27384 8 27385 8 27386 8 27387 8 27388 8 27389 8 27390 8 27391 8 27392 12887 27393 8 27394 8 27395 8 27396 12888 27397 12889 27398 8 27399 12890 27400 8 27401 8 27402 12891 27403 12892 27404 12893 27405 8 27406 8 27407 12894 27408 12895 27409 12896 27410 12897 27411 8 27412 12898 27413 8 27414 12899 27415 12900 27416 12901 27417 12902 27418 12903 27419 12904 27420 12905 27421 12906 27422 12907 27423 12837 27424 12837 27425 12837 27426 12837 27427 12837 27428 12837 27429 12837 27430 12837 27431 12837 27432 12837 27433 12837 27434 12837 27435 12837 27436 12908 27437 12837 27438 12909 27439 12910 27440 8 27441 12911 27442 12912 27443 8 27444 12913 27445 8 27446 12914 27447 8 27448 8 27449 8 27450 12915 27451 12916 27452 12917 27453 8 27454 12918 27455 8 27456 8 27457 12919 27458 12920 27459 12921 27460 8 27461 8 27462 8 27463 12922 27464 12923 27465 8 27466 8 27467 8 27468 12924 27469 12925 27470 12926 27471 12927 27472 8 27473 8 27474 12928 27475 8 27476 12929 27477 12930 27478 8 27479 8 27480 12931 27481 8 27482 8 27483 8 27484 12932 27485 12933 27486 8 27487 8 27488 12934 27489 12935 27490 12936 27491 12937 27492 12938 27493 12939 27494 12940 27495 12941 27496 12942 27497 12943 27498 12944 27499 8 27500 12945 27501 12946 27502 12947 27503 12948 27504 12949 27505 12950 27506 12951 27507 12952 27508 12953 27509 12954 27510 8 27511 12955 27512 12653 27513 12956 27514 8 27515 12957 27516 8 27517 12958 27518 12959 27519 8 27520 8 27521 8 27522 12960 27523 8 27524 8 27525 12961 27526 12962 27527 12963 27528 12964 27529 8 27530 8 27531 12965 27532 8 27533 12966 27534 12967 27535 8 27536 8 27537 12968 27538 8 27539 12969 27540 8 27541 12970 27542 12971 27543 12972 27544 12837 27545 12837 27546 12973 27547 12974 27548 12837 27549 12837 27550 12837 27551 12837 27552 8 27553 12975 27554 8 27555 12976 27556 8 27557 8 27558 12837 27559 12837 27560 8 27561 12837 27562 12837 27563 12977 27564 8 27565 8 27566 8 27567 8 27568 8 27569 8 27570 12978 27571 12979 27572 12980 27573 12981 27574 8 27575 12982 27576 8 27577 12983 27578 12984 27579 12985 27580 12986 27581 12987 27582 8 27583 8 27584 8 27585 8 27586 8 27587 12988 27588 8 27589 8 27590 12989 27591 12990 27592 12991 27593 12992 27594 12993 27595 12994 27596 12995 27597 8 27598 12996 27599 8 27600 12997 27601 12998 27602 12999 27603 13000 27604 13001 27605 8 27606 13002 27607 13003 27608 13004 27609 13005 27610 12837 27611 13006 27612 13007 27613 13008 27614 8 27615 12837 27616 12837 27617 12837 27618 8 27619 13009 27620 8 27621 8 27622 8 27623 13010 27624 13011 27625 13012 27626 8 27627 8 27628 13013 27629 13014 27630 8 27631 13015 27632 13016 27633 8 27634 12837 27635 8 27636 12837 27637 12837 27638 12837 27639 13017 27640 13018 27641 8 27642 12837 27643 12837 27644 12837 27645 13019 27646 8 27647 8 27648 13020 27649 13021 27650 8 27651 13022 27652 8 27653 12837 27654 12653 27655 12653 27656 12653 27657 8 27658 13023 27659 8 27660 13024 27661 8 27662 8 27663 12653 27664 12653 27665 8 27666 12653 27667 12653 27668 13025 27669 8 27670 8 27671 8 27672 8 27673 8 27674 8 27675 13026 27676 13027 27677 13028 27678 13029 27679 8 27680 13030 27681 8 27682 13031 27683 13032 27684 13033 27685 13034 27686 13035 27687 8 27688 8 27689 8 27690 8 27691 8 27692 13036 27693 8 27694 8 27695 13037 27696 13038 27697 13039 27698 13040 27699 13041 27700 13042 27701 13043 27702 8 27703 13044 27704 8 27705 13045 27706 13046 27707 13047 27708 13048 27709 13049 27710 8 27711 13050 27712 13051 27713 13052 27714 13053 27715 12653 27716 13054 27717 13055 27718 13056 27719 8 27720 12653 27721 12653 27722 12653 27723 8 27724 13057 27725 8 27726 8 27727 8 27728 13058 27729 13059 27730 13060 27731 8 27732 8 27733 13061 27734 13062 27735 8 27736 13063 27737 13064 27738 8 27739 12653 27740 8 27741 12653 27742 12653 27743 12653 27744 13065 27745 13066 27746 8 27747 12653 27748 12653 27749 12653 27750 13067 27751 8 27752 8 27753 13068 27754 13069 27755 8 27756 13070 27757 8 27758 12653 27759 13071 27760 8 27761 8 27762 8 27763 8 27764 8 27765 8 27766 8 27767 8 27768 8 27769 13072 27770 8 27771 8 27772 8 27773 8 27774 8 27775 13073 27776 13074 27777 8 27778 13075 27779 8 27780 8 27781 13076 27782 13077 27783 13078 27784 8 27785 8 27786 13079 27787 13080 27788 8 27789 13081 27790 8 27791 13082 27792 8 27793 8 27794 8 27795 8 27796 8 27797 8 27798 8 27799 8 27800 8 27801 8 27802 8 27803 8 27804 8 27805 8 27806 8 27807 8 27808 8 27809 8 27810 8 27811 8 27812 8 27813 8 27814 8 27815 8 27816 8 27817 8 27818 8 27819 8 27820 8 27821 8 27822 8 27823 8 27824 8 27825 8 27826 8 27827 8 27828 8 27829 8 27830 8 27831 8 27832 8 27833 8 27834 8 27835 8 27836 8 27837 8 27838 8 27839 8 27840 8 27841 8 27842 8 27843 8 27844 8 27845 8 27846 8 27847 8 27848 8 27849 8 27850 8 27851 8 27852 8 27853 8 27854 8 27855 8 27856 8 27857 8 27858 8 27859 8 27860 8 27861 8 27862 8 27863 8 27864 8 27865 8 27866 8 27867 8 27868 8 27869 8 27870 8 27871 8 27872 13083 27873 8 27874 8 27875 8 27876 13084 27877 13085 27878 8 27879 8 27880 13086 27881 13087 27882 13088 27883 13089 27884 8 27885 13090 27886 8 27887 13091 27888 13092 27889 13093 27890 13094 27891 13095 27892 13096 27893 13097 27894 13098 27895 13099 27896 84 27897 84 27898 84 27899 84 27900 84 27901 84 27902 84 27903 84 27904 84 27905 84 27906 84 27907 84 27908 84 27909 13100 27910 84 27911 13101 27912 13102 27913 8 27914 13103 27915 13104 27916 8 27917 13105 27918 8 27919 13106 27920 8 27921 8 27922 8 27923 13107 27924 13108 27925 13109 27926 8 27927 13110 27928 8 27929 8 27930 13111 27931 13112 27932 13113 27933 8 27934 8 27935 8 27936 13114 27937 13115 27938 8 27939 8 27940 8 27941 13116 27942 13117 27943 13118 27944 13119 27945 8 27946 8 27947 13120 27948 8 27949 13121 27950 13122 27951 8 27952 8 27953 13123 27954 8 27955 8 27956 8 27957 13124 27958 13125 27959 8 27960 8 27961 13126 27962 13127 27963 13128 27964 13129 27965 13130 27966 8 27967 13131 27968 13132 27969 13133 27970 13134 27971 13135 27972 13136 27973 13137 27974 13138 27975 13139 27976 8 27977 13140 27978 13141 27979 13142 27980 13143 27981 13144 27982 13145 27983 13146 27984 8 27985 13147 27986 13148 27987 8 27988 13149 27989 8 27990 13150 27991 13151 27992 8 27993 13152 27994 8 27995 8 27996 13153 27997 13154 27998 8 27999 13155 28000 13156 28001 13157 28002 8 28003 13158 28004 13159 28005 13160 28006 13161 28007 8 28008 8 28009 13162 28010 13163 28011 8 28012 13164 28013 13165 28014 13166 28015 13167 28016 8 28017 13168 28018 13169 28019 13170 28020 13171 28021 13172 28022 13173 28023 8 28024 13174 28025 13175 28026 8 28027 13176 28028 13177 28029 13178 28030 13179 28031 13180 28032 13181 28033 8 28034 13182 28035 13183 28036 13184 28037 13185 28038 8 28039 13186 28040 13187 28041 13188 28042 8 28043 13189 28044 8 28045 13190 28046 13191 28047 13192 28048 8 28049 13193 28050 13194 28051 8 28052 8 28053 13195 28054 8 28055 13196 28056 8 28057 13197 28058 8 28059 13198 28060 8 28061 13199 28062 13200 28063 13201 28064 13202 28065 13203 28066 13204 28067 13205 28068 13206 28069 13207 28070 8 28071 8 28072 13208 28073 13209 28074 8 28075 8 28076 13210 28077 13211 28078 13212 28079 13213 28080 13214 28081 13215 28082 13216 28083 13217 28084 8 28085 8 28086 13218 28087 13219 28088 8 28089 13220 28090 8 28091 13221 28092 13222 28093 13223 28094 13224 28095 13225 28096 8 28097 13226 28098 13227 28099 13228 28100 13229 28101 8 28102 8 28103 13230 28104 13231 28105 8 28106 13232 28107 8 28108 13233 28109 8 28110 8 28111 13234 28112 13235 28113 13236 28114 13237 28115 13238 28116 13239 28117 13240 28118 13241 28119 13242 28120 13243 28121 13244 28122 8 28123 13245 28124 13246 28125 13247 28126 13248 28127 13249 28128 13250 28129 13251 28130 13252 28131 13253 28132 13254 28133 13255 28134 13256 28135 13257 28136 13258 28137 13259 28138 13260 28139 13261 28140 13262 28141 13263 28142 8 28143 13264 28144 8 28145 13265 28146 8 28147 8 28148 13266 28149 13267 28150 13268 28151 13269 28152 8 28153 8 28154 13270 28155 13271 28156 13272 28157 13273 28158 13274 28159 8 28160 13275 28161 8 28162 13276 28163 8 28164 8 28165 8 28166 8 28167 13277 28168 8 28169 13278 28170 13279 28171 8 28172 13280 28173 13281 28174 13282 28175 13283 28176 13284 28177 8 28178 13285 28179 13286 28180 8 28181 13287 28182 8 28183 13288 28184 13289 28185 13290 28186 13291 28187 13292 28188 13293 28189 13294 28190 8 28191 8 28192 13295 28193 8 28194 8 28195 13296 28196 13297 28197 13298 28198 8 28199 8 28200 8 28201 8 28202 13299 28203 8 28204 8 28205 8 28206 8 28207 8 28208 13300 28209 13301 28210 13302 28211 13303 28212 13304 28213 13305 28214 13306 28215 13307 28216 13308 28217 13309 28218 13310 28219 13311 28220 13312 28221 13313 28222 13314 28223 13315 28224 13316 28225 13317 28226 13318 28227 13319 28228 13320 28229 13321 28230 13322 28231 13323 28232 13324 28233 13325 28234 13326 28235 13327 28236 13328 28237 13329 28238 13330 28239 13331 28240 13332 28241 13333 28242 13303 28243 13303 28244 13303 28245 13303 28246 13303 28247 13303 28248 13303 28249 13303 28250 13303 28251 13303 28252 13303 28253 13303 28254 13303 28255 13303 28256 13303 28257 13334 28258 13335 28259 8 28260 8 28261 8 28262 8 28263 8 28264 8 28265 13336 28266 8 28267 8 28268 13337 28269 8 28270 13338 28271 13339 28272 13340 28273 13341 28274 13342 28275 8 28276 8 28277 8 28278 8 28279 8 28280 13343 28281 13344 28282 8 28283 8 28284 13345 28285 8 28286 13346 28287 8 28288 8 28289 13347 28290 13348 28291 8 28292 13303 28293 13303 28294 13303 28295 13303 28296 13303 28297 13349 28298 13350 28299 13351 28300 13303 28301 13303 28302 13303 28303 13303 28304 13303 28305 13303 28306 13303 28307 13303 28308 13303 28309 13303 28310 13352 28311 8 28312 8 28313 8 28314 8 28315 8 28316 8 28317 8 28318 8 28319 8 28320 8 28321 8 28322 8 28323 8 28324 8 28325 8 28326 8 28327 8 28328 8 28329 8 28330 8 28331 8 28332 8 28333 8 28334 8 28335 8 28336 8 28337 8 28338 8 28339 8 28340 8 28341 8 28342 8 28343 8 28344 8 28345 8 28346 8 28347 8 28348 8 28349 8 28350 8 28351 8 28352 8 28353 8 28354 8 28355 8 28356 8 28357 8 28358 8 28359 8 28360 8 28361 8 28362 8 28363 8 28364 8 28365 8 28366 8 28367 8 28368 8 28369 8 28370 8 28371 8 28372 8 28373 8 28374 8 28375 8 28376 8 28377 8 28378 8 28379 8 28380 8 28381 8 28382 8 28383 8 28384 8 28385 8 28386 8 28387 8 28388 8 28389 8 28390 8 28391 13353 28392 8 28393 8 28394 8 28395 13354 28396 13355 28397 8 28398 13356 28399 8 28400 8 28401 13357 28402 13358 28403 13359 28404 8 28405 8 28406 13360 28407 13361 28408 13362 28409 13363 28410 8 28411 13364 28412 8 28413 13365 28414 13366 28415 13367 28416 13368 28417 13369 28418 13370 28419 13371 28420 13372 28421 13373 28422 13303 28423 13303 28424 13303 28425 13303 28426 13303 28427 13303 28428 13303 28429 13303 28430 13303 28431 13303 28432 13303 28433 13303 28434 13303 28435 13374 28436 13303 28437 13375 28438 13376 28439 8 28440 13377 28441 13378 28442 8 28443 13379 28444 8 28445 13380 28446 8 28447 8 28448 8 28449 13381 28450 13382 28451 13383 28452 8 28453 13384 28454 8 28455 8 28456 13385 28457 13386 28458 13387 28459 8 28460 8 28461 8 28462 13388 28463 13389 28464 8 28465 8 28466 8 28467 13390 28468 13391 28469 13392 28470 13393 28471 8 28472 8 28473 13394 28474 8 28475 13395 28476 13396 28477 8 28478 8 28479 13397 28480 8 28481 8 28482 8 28483 13398 28484 13399 28485 8 28486 8 28487 13400 28488 13401 28489 13402 28490 13403 28491 13404 28492 13405 28493 13406 28494 13407 28495 13408 28496 13409 28497 13410 28498 8 28499 13411 28500 13412 28501 13413 28502 13414 28503 13415 28504 13416 28505 13417 28506 13418 28507 13419 28508 13420 28509 8 28510 8 28511 13421 28512 8 28513 13422 28514 8 28515 13423 28516 13424 28517 8 28518 8 28519 8 28520 13425 28521 8 28522 8 28523 13426 28524 13427 28525 13428 28526 13429 28527 8 28528 8 28529 13430 28530 8 28531 13431 28532 13432 28533 8 28534 8 28535 13433 28536 8 28537 13434 28538 8 28539 13435 28540 13436 28541 13437 28542 13303 28543 13303 28544 13438 28545 13439 28546 13440 28547 13441 28548 13442 28549 13443 28550 13444 28551 13445 28552 13446 28553 13447 28554 8 28555 13448 28556 8 28557 13449 28558 8 28559 8 28560 13450 28561 13451 28562 13452 28563 13453 28564 8 28565 8 28566 13454 28567 13455 28568 13456 28569 13457 28570 13458 28571 8 28572 13459 28573 8 28574 13460 28575 8 28576 8 28577 8 28578 8 28579 13461 28580 8 28581 13462 28582 13463 28583 8 28584 13464 28585 13465 28586 13466 28587 13467 28588 13468 28589 8 28590 13469 28591 13470 28592 8 28593 13471 28594 8 28595 13472 28596 13473 28597 13474 28598 13475 28599 13476 28600 13477 28601 13478 28602 8 28603 8 28604 13479 28605 8 28606 8 28607 13480 28608 13481 28609 13482 28610 8 28611 8 28612 8 28613 8 28614 13483 28615 8 28616 8 28617 8 28618 8 28619 8 28620 13484 28621 13485 28622 13486 28623 13487 28624 13488 28625 13489 28626 13490 28627 13491 28628 13492 28629 13493 28630 13494 28631 13495 28632 13496 28633 13497 28634 13498 28635 13499 28636 13500 28637 13501 28638 13502 28639 13503 28640 13504 28641 13505 28642 13506 28643 13507 28644 13508 28645 13509 28646 13510 28647 13511 28648 13512 28649 13513 28650 13514 28651 13515 28652 13516 28653 13517 28654 13487 28655 13487 28656 13487 28657 13487 28658 13487 28659 13487 28660 13487 28661 13487 28662 13487 28663 13487 28664 13487 28665 13487 28666 13487 28667 13487 28668 13487 28669 13518 28670 13519 28671 8 28672 8 28673 8 28674 8 28675 8 28676 8 28677 13520 28678 8 28679 8 28680 13521 28681 8 28682 13522 28683 13523 28684 13524 28685 13525 28686 13526 28687 8 28688 8 28689 8 28690 8 28691 8 28692 13527 28693 13528 28694 8 28695 8 28696 13529 28697 8 28698 13530 28699 8 28700 8 28701 13531 28702 13532 28703 8 28704 13487 28705 13487 28706 13487 28707 13487 28708 13487 28709 13533 28710 13534 28711 13535 28712 13487 28713 13487 28714 13487 28715 13487 28716 13487 28717 13487 28718 13487 28719 13487 28720 13487 28721 13487 28722 13536 28723 8 28724 8 28725 8 28726 8 28727 8 28728 8 28729 8 28730 8 28731 8 28732 8 28733 8 28734 8 28735 8 28736 8 28737 8 28738 8 28739 8 28740 8 28741 8 28742 8 28743 8 28744 8 28745 8 28746 8 28747 8 28748 8 28749 8 28750 8 28751 8 28752 8 28753 8 28754 8 28755 8 28756 8 28757 8 28758 8 28759 8 28760 8 28761 8 28762 8 28763 8 28764 8 28765 8 28766 8 28767 8 28768 8 28769 8 28770 8 28771 8 28772 8 28773 8 28774 8 28775 8 28776 8 28777 8 28778 8 28779 8 28780 8 28781 8 28782 8 28783 8 28784 8 28785 8 28786 8 28787 8 28788 8 28789 8 28790 8 28791 8 28792 8 28793 8 28794 8 28795 8 28796 8 28797 8 28798 8 28799 8 28800 8 28801 8 28802 8 28803 13537 28804 8 28805 8 28806 8 28807 13538 28808 13539 28809 8 28810 13540 28811 8 28812 8 28813 13541 28814 13542 28815 13543 28816 8 28817 8 28818 13544 28819 13545 28820 13546 28821 13547 28822 8 28823 13548 28824 8 28825 13549 28826 13550 28827 13551 28828 13552 28829 13553 28830 13554 28831 13555 28832 13556 28833 13557 28834 13487 28835 13487 28836 13487 28837 13487 28838 13487 28839 13487 28840 13487 28841 13487 28842 13487 28843 13487 28844 13487 28845 13487 28846 13487 28847 13558 28848 13487 28849 13559 28850 13560 28851 8 28852 13561 28853 13562 28854 8 28855 13563 28856 8 28857 13564 28858 8 28859 8 28860 8 28861 13565 28862 13566 28863 13567 28864 8 28865 13568 28866 8 28867 8 28868 13569 28869 13570 28870 13571 28871 8 28872 8 28873 8 28874 13572 28875 13573 28876 8 28877 8 28878 8 28879 13574 28880 13575 28881 13576 28882 13577 28883 8 28884 8 28885 13578 28886 8 28887 13579 28888 13580 28889 8 28890 8 28891 13581 28892 8 28893 8 28894 8 28895 13582 28896 13583 28897 8 28898 8 28899 13584 28900 13585 28901 13586 28902 13587 28903 13588 28904 13589 28905 13590 28906 13591 28907 13592 28908 13593 28909 13594 28910 8 28911 13595 28912 13596 28913 13597 28914 13598 28915 13599 28916 13600 28917 13601 28918 13602 28919 13603 28920 13604 28921 8 28922 13605 28923 13303 28924 13606 28925 8 28926 13607 28927 8 28928 13608 28929 13609 28930 8 28931 8 28932 8 28933 13610 28934 8 28935 8 28936 13611 28937 13612 28938 13613 28939 13614 28940 8 28941 8 28942 13615 28943 8 28944 13616 28945 13617 28946 8 28947 8 28948 13618 28949 8 28950 13619 28951 8 28952 13620 28953 13621 28954 13622 28955 13487 28956 13487 28957 13623 28958 13624 28959 13487 28960 13487 28961 13487 28962 13487 28963 8 28964 13625 28965 8 28966 13626 28967 8 28968 8 28969 13487 28970 13487 28971 8 28972 13487 28973 13487 28974 13627 28975 8 28976 8 28977 8 28978 8 28979 8 28980 8 28981 13628 28982 13629 28983 13630 28984 13631 28985 8 28986 13632 28987 8 28988 13633 28989 13634 28990 13635 28991 13636 28992 13637 28993 8 28994 8 28995 8 28996 8 28997 8 28998 13638 28999 8 29000 8 29001 13639 29002 13640 29003 13641 29004 13642 29005 13643 29006 13644 29007 13645 29008 8 29009 13646 29010 8 29011 13647 29012 13648 29013 13649 29014 13650 29015 13651 29016 8 29017 13652 29018 13653 29019 13654 29020 13655 29021 13487 29022 13656 29023 13657 29024 13658 29025 8 29026 13487 29027 13487 29028 13487 29029 8 29030 13659 29031 8 29032 8 29033 8 29034 13660 29035 13661 29036 13662 29037 8 29038 8 29039 13663 29040 13664 29041 8 29042 13665 29043 13666 29044 8 29045 13487 29046 8 29047 13487 29048 13487 29049 13487 29050 13667 29051 13668 29052 8 29053 13487 29054 13487 29055 13487 29056 13669 29057 8 29058 8 29059 13670 29060 13671 29061 8 29062 13672 29063 8 29064 13487 29065 13303 29066 13303 29067 13303 29068 8 29069 13673 29070 8 29071 13674 29072 8 29073 8 29074 13303 29075 13303 29076 8 29077 13303 29078 13303 29079 13675 29080 8 29081 8 29082 8 29083 8 29084 8 29085 8 29086 13676 29087 13677 29088 13678 29089 13679 29090 8 29091 13680 29092 8 29093 13681 29094 13682 29095 13683 29096 13684 29097 13685 29098 8 29099 8 29100 8 29101 8 29102 8 29103 13686 29104 8 29105 8 29106 13687 29107 13688 29108 13689 29109 13690 29110 13691 29111 13692 29112 13693 29113 8 29114 13694 29115 8 29116 13695 29117 13696 29118 13697 29119 13698 29120 13699 29121 8 29122 13700 29123 13701 29124 13702 29125 13703 29126 13303 29127 13704 29128 13705 29129 13706 29130 8 29131 13303 29132 13303 29133 13303 29134 8 29135 13707 29136 8 29137 8 29138 8 29139 13708 29140 13709 29141 13710 29142 8 29143 8 29144 13711 29145 13712 29146 8 29147 13713 29148 13714 29149 8 29150 13303 29151 8 29152 13303 29153 13303 29154 13303 29155 13715 29156 13716 29157 8 29158 13303 29159 13303 29160 13303 29161 13717 29162 8 29163 8 29164 13718 29165 13719 29166 8 29167 13720 29168 8 29169 13303 29170 8 29171 13721 29172 8 29173 13722 29174 8 29175 13723 29176 13724 29177 8 29178 8 29179 13725 29180 13726 29181 13727 29182 13728 29183 13729 29184 13730 29185 13731 29186 8 29187 8 29188 13732 29189 13733 29190 13734 29191 13735 29192 8 29193 8 29194 13736 29195 8 29196 13737 29197 8 29198 8 29199 8 29200 13738 29201 13739 29202 13740 29203 13741 29204 13742 29205 13743 29206 13744 29207 13745 29208 13746 29209 13747 29210 13748 29211 8 29212 13749 29213 13750 29214 13751 29215 13752 29216 13753 29217 13754 29218 13755 29219 8 29220 13756 29221 13757 29222 13758 29223 13759 29224 8 29225 13760 29226 13761 29227 13762 29228 13763 29229 13764 29230 8 29231 13765 29232 13766 29233 13767 29234 13768 29235 13769 29236 13770 29237 8 29238 13771 29239 13772 29240 13773 29241 13774 29242 13775 29243 13776 29244 8 29245 13777 29246 13778 29247 13779 29248 13780 29249 8 29250 8 29251 13781 29252 8 29253 13782 29254 13783 29255 8 29256 8 29257 13784 29258 8 29259 13785 29260 13786 29261 8 29262 13787 29263 13788 29264 13789 29265 8 29266 13790 29267 8 29268 13791 29269 13792 29270 13793 29271 84 29272 84 29273 13794 29274 13795 29275 13796 29276 13797 29277 13798 29278 13799 29279 13800 29280 13801 29281 13802 29282 13803 29283 8 29284 13804 29285 8 29286 13805 29287 8 29288 8 29289 13806 29290 13807 29291 13808 29292 13809 29293 8 29294 8 29295 13810 29296 13811 29297 13812 29298 13813 29299 13814 29300 8 29301 13815 29302 8 29303 13816 29304 8 29305 8 29306 8 29307 8 29308 13817 29309 8 29310 13818 29311 13819 29312 8 29313 13820 29314 13821 29315 13822 29316 13823 29317 13824 29318 8 29319 13825 29320 13826 29321 8 29322 13827 29323 8 29324 13828 29325 13829 29326 13830 29327 13831 29328 13832 29329 13833 29330 13834 29331 8 29332 8 29333 13835 29334 8 29335 8 29336 13836 29337 13837 29338 13838 29339 8 29340 8 29341 8 29342 8 29343 13839 29344 8 29345 8 29346 8 29347 8 29348 8 29349 13840 29350 13841 29351 13842 29352 13843 29353 13844 29354 13845 29355 13846 29356 13847 29357 13848 29358 13849 29359 13850 29360 13851 29361 13852 29362 13853 29363 13854 29364 13855 29365 13856 29366 13857 29367 13858 29368 13859 29369 13860 29370 13861 29371 13862 29372 13863 29373 13864 29374 13865 29375 13866 29376 13867 29377 13868 29378 13869 29379 13870 29380 13871 29381 13872 29382 13873 29383 13843 29384 13843 29385 13843 29386 13843 29387 13843 29388 13843 29389 13843 29390 13843 29391 13843 29392 13843 29393 13843 29394 13843 29395 13843 29396 13843 29397 13843 29398 13874 29399 13875 29400 8 29401 8 29402 8 29403 8 29404 8 29405 8 29406 13876 29407 8 29408 8 29409 13877 29410 8 29411 13878 29412 13879 29413 13880 29414 13881 29415 13882 29416 8 29417 8 29418 8 29419 8 29420 8 29421 13883 29422 13884 29423 8 29424 8 29425 13885 29426 8 29427 13886 29428 8 29429 8 29430 13887 29431 13888 29432 8 29433 13843 29434 13843 29435 13843 29436 13843 29437 13843 29438 13889 29439 13890 29440 13891 29441 13843 29442 13843 29443 13843 29444 13843 29445 13843 29446 13843 29447 13843 29448 13843 29449 13843 29450 13843 29451 13892 29452 8 29453 8 29454 8 29455 8 29456 8 29457 8 29458 8 29459 8 29460 8 29461 8 29462 8 29463 8 29464 8 29465 8 29466 8 29467 8 29468 8 29469 8 29470 8 29471 8 29472 8 29473 8 29474 8 29475 8 29476 8 29477 8 29478 8 29479 8 29480 8 29481 8 29482 8 29483 8 29484 8 29485 8 29486 8 29487 8 29488 8 29489 8 29490 8 29491 8 29492 8 29493 8 29494 8 29495 8 29496 8 29497 8 29498 8 29499 8 29500 8 29501 8 29502 8 29503 8 29504 8 29505 8 29506 8 29507 8 29508 8 29509 8 29510 8 29511 8 29512 8 29513 8 29514 8 29515 8 29516 8 29517 8 29518 8 29519 8 29520 8 29521 8 29522 8 29523 8 29524 8 29525 8 29526 8 29527 8 29528 8 29529 8 29530 8 29531 8 29532 13893 29533 8 29534 8 29535 8 29536 13894 29537 13895 29538 8 29539 13896 29540 8 29541 8 29542 13897 29543 13898 29544 13899 29545 8 29546 8 29547 13900 29548 13901 29549 13902 29550 13903 29551 8 29552 13904 29553 8 29554 13905 29555 13906 29556 13907 29557 13908 29558 13909 29559 13910 29560 13911 29561 13912 29562 13913 29563 13843 29564 13843 29565 13843 29566 13843 29567 13843 29568 13843 29569 13843 29570 13843 29571 13843 29572 13843 29573 13843 29574 13843 29575 13843 29576 13914 29577 13843 29578 13915 29579 13916 29580 8 29581 13917 29582 13918 29583 8 29584 13919 29585 8 29586 13920 29587 8 29588 8 29589 8 29590 13921 29591 13922 29592 13923 29593 8 29594 13924 29595 8 29596 8 29597 13925 29598 13926 29599 13927 29600 8 29601 8 29602 8 29603 13928 29604 13929 29605 8 29606 8 29607 8 29608 13930 29609 13931 29610 13932 29611 13933 29612 8 29613 8 29614 13934 29615 8 29616 13935 29617 13936 29618 8 29619 8 29620 13937 29621 8 29622 8 29623 8 29624 13938 29625 13939 29626 8 29627 8 29628 13940 29629 13941 29630 13942 29631 13943 29632 13944 29633 13945 29634 13946 29635 13947 29636 13948 29637 13949 29638 13950 29639 8 29640 13951 29641 13952 29642 13953 29643 13954 29644 13955 29645 13956 29646 13957 29647 13958 29648 13959 29649 13960 29650 8 29651 13961 29652 84 29653 13962 29654 8 29655 13963 29656 8 29657 13964 29658 13965 29659 8 29660 8 29661 8 29662 13966 29663 8 29664 8 29665 13967 29666 13968 29667 13969 29668 13970 29669 8 29670 8 29671 13971 29672 8 29673 13972 29674 13973 29675 8 29676 8 29677 13974 29678 8 29679 13975 29680 8 29681 13976 29682 13977 29683 13978 29684 13843 29685 13843 29686 13979 29687 13980 29688 13843 29689 13843 29690 13843 29691 13843 29692 8 29693 13981 29694 8 29695 13982 29696 8 29697 8 29698 13843 29699 13843 29700 8 29701 13843 29702 13843 29703 13983 29704 8 29705 8 29706 8 29707 8 29708 8 29709 8 29710 13984 29711 13985 29712 13986 29713 13987 29714 8 29715 13988 29716 8 29717 13989 29718 13990 29719 13991 29720 13992 29721 13993 29722 8 29723 8 29724 8 29725 8 29726 8 29727 13994 29728 8 29729 8 29730 13995 29731 13996 29732 13997 29733 13998 29734 13999 29735 14000 29736 14001 29737 8 29738 14002 29739 8 29740 14003 29741 14004 29742 14005 29743 14006 29744 14007 29745 8 29746 14008 29747 14009 29748 14010 29749 14011 29750 13843 29751 14012 29752 14013 29753 14014 29754 8 29755 13843 29756 13843 29757 13843 29758 8 29759 14015 29760 8 29761 8 29762 8 29763 14016 29764 14017 29765 14018 29766 8 29767 8 29768 14019 29769 14020 29770 8 29771 14021 29772 14022 29773 8 29774 13843 29775 8 29776 13843 29777 13843 29778 13843 29779 14023 29780 14024 29781 8 29782 13843 29783 13843 29784 13843 29785 14025 29786 8 29787 8 29788 14026 29789 14027 29790 8 29791 14028 29792 8 29793 13843 29794 14029 29795 14030 29796 14031 29797 14032 29798 8 29799 8 29800 14033 29801 14034 29802 8 29803 8 29804 8 29805 8 29806 8 29807 8 29808 14035 29809 14036 29810 8 29811 8 29812 14037 29813 14038 29814 14039 29815 14040 29816 14041 29817 14042 29818 8 29819 8 29820 8 29821 8 29822 14043 29823 8 29824 8 29825 14044 29826 8 29827 14045 29828 14046 29829 8 29830 8 29831 14047 29832 8 29833 14048 29834 14049 29835 14050 29836 8 29837 8 29838 14051 29839 8 29840 8 29841 8 29842 8 29843 14052 29844 14053 29845 8 29846 8 29847 8 29848 14054 29849 8 29850 14055 29851 8 29852 14056 29853 14057 29854 8 29855 8 29856 84 29857 14058 29858 14059 29859 14060 29860 14061 29861 8 29862 14062 29863 14063 29864 14064 29865 14065 29866 14066 29867 14067 29868 8 29869 14068 29870 14069 29871 8 29872 14070 29873 14071 29874 14072 29875 14073 29876 14074 29877 14075 29878 8 29879 14076 29880 14077 29881 14078 29882 14079 29883 8 29884 84 29885 14080 29886 14081 29887 14082 29888 14083 29889 8 29890 14084 29891 14085 29892 14086 29893 8 29894 14087 29895 8 29896 14088 29897 14089 29898 14090 29899 8 29900 14091 29901 14092 29902 8 29903 84 29904 8 29905 14093 29906 8 29907 14094 29908 8 29909 14095 29910 8 29911 8 29912 14096 29913 8 29914 84 29915 84 29916 84 29917 84 29918 14097 29919 8 29920 8 29921 8 29922 8 29923 8 29924 8 29925 14098 29926 14099 29927 14100 29928 14101 29929 8 29930 14102 29931 8 29932 14103 29933 14104 29934 14105 29935 14106 29936 14107 29937 8 29938 8 29939 8 29940 8 29941 8 29942 14108 29943 8 29944 8 29945 14109 29946 14110 29947 14111 29948 14112 29949 14113 29950 14114 29951 14115 29952 8 29953 14116 29954 8 29955 8 29956 8 29957 8 29958 8 29959 8 29960 8 29961 8 29962 8 29963 8 29964 8 29965 8 29966 8 29967 8 29968 8 29969 8 29970 8 29971 8 29972 8 29973 8 29974 8 29975 8 29976 8 29977 8 29978 8 29979 8 29980 8 29981 8 29982 8 29983 8 29984 8 29985 8 29986 8 29987 8 29988 8 29989 8 29990 8 29991 8 29992 8 29993 8 29994 8 29995 8 29996 8 29997 8 29998 8 29999 8 30000 8 30001 8 30002 8 30003 8 30004 8 30005 8 30006 8 30007 8 30008 8 30009 8 30010 8 30011 8 30012 8 30013 8 30014 8 30015 8 30016 8 30017 8 30018 8 30019 8 30020 8 30021 8 30022 8 30023 8 30024 8 30025 8 30026 8 30027 8 30028 8 30029 8 30030 8 30031 8 30032 8 30033 8 30034 8 30035 8 30036 14117 30037 8 30038 8 30039 14118 30040 14119 30041 8 30042 14120 30043 14121 30044 14122 30045 14123 30046 14124 30047 14125 30048 14126 30049 14127 30050 14128 30051 14129 30052 14130 30053 14131 30054 14132 30055 14133 30056 8 30057 14134 30058 14135 30059 14136 30060 14137 30061 14138 30062 14139 30063 14140 30064 14141 30065 14142 30066 14143 30067 14144 30068 8 30069 14145 30070 8 30071 14146 30072 8 30073 14147 30074 8 30075 14148 30076 14149 30077 8 30078 8 30079 14150 30080 8 30081 8 30082 8 30083 8 30084 14151 30085 8 30086 14152 30087 8 30088 14153 30089 14154 30090 14155 30091 14156 30092 14157 30093 8 30094 14158 30095 14159 30096 14160 30097 14161 30098 14162 30099 14163 30100 14164 30101 14165 30102 14166 30103 8 30104 14167 30105 14168 30106 14169 30107 14170 30108 14171 30109 14172 30110 14173 30111 8 30112 14174 30113 14175 30114 8 30115 14176 30116 8 30117 14177 30118 14178 30119 8 30120 14179 30121 8 30122 8 30123 14180 30124 14181 30125 8 30126 14182 30127 14183 30128 14184 30129 8 30130 14185 30131 14186 30132 14187 30133 14188 30134 8 30135 8 30136 14189 30137 14190 30138 8 30139 14191 30140 14192 30141 14193 30142 14194 30143 8 30144 14195 30145 14196 30146 14197 30147 14198 30148 14199 30149 14200 30150 8 30151 14201 30152 14202 30153 8 30154 14203 30155 14204 30156 14205 30157 14206 30158 14207 30159 14208 30160 8 30161 14209 30162 14210 30163 14211 30164 14212 30165 8 30166 14213 30167 14214 30168 14215 30169 8 30170 14216 30171 8 30172 14217 30173 14218 30174 14219 30175 8 30176 14220 30177 14221 30178 8 30179 8 30180 14222 30181 8 30182 14223 30183 8 30184 14224 30185 8 30186 14225 30187 8 30188 14226 30189 14227 30190 14228 30191 14229 30192 14230 30193 14231 30194 14232 30195 14233 30196 14234 30197 8 30198 8 30199 14235 30200 14236 30201 8 30202 8 30203 14237 30204 14238 30205 14239 30206 14240 30207 14241 30208 14242 30209 14243 30210 14244 30211 8 30212 8 30213 14245 30214 14246 30215 8 30216 14247 30217 8 30218 14248 30219 14249 30220 14250 30221 14251 30222 14252 30223 8 30224 14253 30225 14254 30226 14255 30227 14256 30228 8 30229 8 30230 14257 30231 14258 30232 8 30233 14259 30234 8 30235 14260 30236 8 30237 8 30238 14261 30239 14262 30240 14263 30241 14264 30242 14265 30243 14266 30244 14267 30245 14268 30246 14269 30247 14270 30248 14271 30249 8 30250 8 30251 14272 30252 14273 30253 14274 30254 14275 30255 14276 30256 14277 30257 14278 30258 8 30259 8 30260 14279 30261 14280 30262 8 30263 14281 30264 14282 30265 14283 30266 8 30267 14284 30268 14285 30269 14286 30270 14287 30271 8 30272 8 30273 14288 30274 14289 30275 8 30276 14290 30277 14291 30278 14292 30279 14293 30280 14294 30281 8 30282 8 30283 8 30284 8 30285 14295 30286 14296 30287 14297 30288 14298 30289 8 30290 14299 30291 14300 30292 14301 30293 14302 30294 14303 30295 14304 30296 8 30297 14305 30298 14306 30299 8 30300 14307 30301 14308 30302 14309 30303 14310 30304 14311 30305 14312 30306 8 30307 14313 30308 14314 30309 14315 30310 14316 30311 8 30312 14317 30313 14318 30314 14319 30315 14320 30316 8 30317 14321 30318 14322 30319 14323 30320 8 30321 14324 30322 8 30323 14325 30324 14326 30325 14327 30326 8 30327 14328 30328 14329 30329 8 30330 8 30331 14330 30332 8 30333 14331 30334 8 30335 14332 30336 14333 30337 14334 30338 8 30339 8 30340 14335 30341 14336 30342 8 30343 14337 30344 14338 30345 14339 30346 14340 30347 14341 30348 8 30349 14342 30350 14343 30351 14344 30352 8 30353 14345 30354 14346 30355 8 30356 14347 30357 8 30358 14348 30359 8 30360 14349 30361 14350 30362 14351 30363 14352 30364 14353 30365 14354 30366 14355 30367 14356 30368 14357 30369 8 30370 8 30371 14358 30372 14359 30373 8 30374 8 30375 14360 30376 14361 30377 14362 30378 14363 30379 14364 30380 14365 30381 14366 30382 14367 30383 8 30384 8 30385 14368 30386 14369 30387 8 30388 14370 30389 14371 30390 14372 30391 14373 30392 14374 30393 14375 30394 8 30395 14376 30396 8 30397 14377 30398 14378 30399 14379 30400 14380 30401 14381 30402 8 30403 14382 30404 14383 30405 14384 30406 14385 30407 8 30408 8 30409 14386 30410 14387 30411 8 30412 14388 30413 8 30414 14389 30415 8 30416 14390 30417 14391 30418 8 30419 8 30420 14392 30421 14393 30422 8 30423 14394 30424 14395 30425 14396 30426 14397 30427 14398 30428 14399 30429 8 30430 14400 30431 8 30432 8 30433 14401 30434 14402 30435 8 30436 14403 30437 14404 30438 14405 30439 14406 30440 14407 30441 8 30442 14408 30443 14409 30444 14410 30445 14411 30446 84 30447 14412 30448 14413 30449 14414 30450 8 30451 84 30452 14415 30453 14416 30454 14417 30455 14418 30456 14419 30457 8 30458 84 30459 84 30460 8 30461 14420 30462 14421 30463 14422 30464 8 30465 14423 30466 14424 30467 14425 30468 14426 30469 14427 30470 14428 30471 14429 30472 14430 30473 14431 30474 14432 30475 8 30476 14433 30477 14434 30478 14435 30479 14436 30480 14437 30481 14438 30482 14439 30483 8 30484 14440 30485 14441 30486 8 30487 14442 30488 8 30489 14443 30490 14444 30491 8 30492 14445 30493 14446 30494 14447 30495 14448 30496 14449 30497 8 30498 8 30499 8 30500 14450 30501 14451 30502 8 30503 14452 30504 14453 30505 14454 30506 8 30507 14455 30508 14456 30509 14457 30510 14458 30511 8 30512 8 30513 14459 30514 14460 30515 8 30516 14461 30517 14462 30518 14463 30519 14464 30520 8 30521 14465 30522 14466 30523 14467 30524 14468 30525 14469 30526 14470 30527 8 30528 14471 30529 14472 30530 8 30531 14473 30532 14474 30533 14475 30534 14476 30535 14477 30536 14478 30537 8 30538 14479 30539 14480 30540 14481 30541 14482 30542 8 30543 14483 30544 14484 30545 14485 30546 8 30547 14486 30548 8 30549 14487 30550 14488 30551 14489 30552 8 30553 14490 30554 14491 30555 8 30556 8 30557 14492 30558 8 30559 14493 30560 8 30561 14494 30562 14495 30563 14496 30564 8 30565 14497 30566 8 30567 14498 30568 8 30569 14499 30570 14500 30571 14501 30572 14502 30573 14503 30574 14504 30575 14505 30576 14506 30577 14507 30578 8 30579 8 30580 14508 30581 14509 30582 8 30583 8 30584 14510 30585 14511 30586 14512 30587 14513 30588 14514 30589 14515 30590 14516 30591 14517 30592 8 30593 8 30594 14518 30595 14519 30596 8 30597 14520 30598 8 30599 14521 30600 14522 30601 14523 30602 14524 30603 14525 30604 8 30605 14526 30606 14527 30607 14528 30608 14529 30609 8 30610 8 30611 14530 30612 14531 30613 8 30614 14532 30615 8 30616 14533 30617 8 30618 14534 30619 14535 30620 8 30621 14536 30622 8 30623 14537 30624 8 30625 14538 30626 14539 30627 14540 30628 14541 30629 14542 30630 14543 30631 14544 30632 14545 30633 14546 30634 8 30635 8 30636 14547 30637 14548 30638 8 30639 8 30640 14549 30641 14550 30642 14551 30643 14552 30644 14553 30645 14554 30646 14555 30647 14556 30648 8 30649 8 30650 14557 30651 14558 30652 8 30653 14559 30654 8 30655 8 30656 8 30657 14560 30658 14561 30659 14562 30660 8 30661 8 30662 14563 30663 14564 30664 8 30665 14565 30666 14566 30667 14567 30668 14568 30669 14569 30670 14570 30671 8 30672 14571 30673 14572 30674 8 30675 14573 30676 8 30677 84 30678 14574 30679 14575 30680 14576 30681 14577 30682 14578 30683 8 30684 14579 30685 14580 30686 14581 30687 14582 30688 8 30689 8 30690 14583 30691 14584 30692 8 30693 8 30694 14585 30695 8 30696 14586 30697 8 30698 84 30699 14587 30700 14588 30701 14589 30702 14590 30703 14591 30704 14592 30705 8 30706 14593 30707 8 30708 84 30709 14594 30710 14595 30711 8 30712 84 30713 14596 30714 14597 30715 8 30716 84 30717 84 30718 84 30719 14598 30720 8 30721 8 30722 14599 30723 14600 30724 8 30725 14601 30726 8 30727 84))
  "The generated goto-tables.")

(defvar
  phps-mode-parser--distinct-goto-tables
  #s(hash-table size 18699 test equal rehash-size 1.5 rehash-threshold 0.8125 data (0 ((top_statement_list 1)) 1 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 8) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS 16) (T_CLASS_C 17) (T_CLONE 18) (T_CONST 19) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_ENUM 30) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_HALT_COMPILER 42) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INTERFACE 48) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAMESPACE 56) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 68) (T_START_HEREDOC 69) (T_STATIC 70) (T_STRING 71) (T_STRING_CAST 72) (T_SWITCH 73) (T_THROW 74) (T_TRAIT 75) (T_TRAIT_C 76) (T_TRY 77) (T_UNSET 78) (T_UNSET_CAST 79) (T_USE 80) (T_VARIABLE 81) (T_WHILE 82) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 87) (alt_if_stmt_without_else 88) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 91) (attributes 92) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 96) (class_modifier 97) (class_modifiers 98) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 102) (expr 103) (fn 104) (fully_dereferenceable 105) #'106 (function_call 107) (function_declaration_statement 108) (if_stmt 109) (if_stmt_without_else 110) (inline_function 111) (interface_declaration_statement 112) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 120) (static_member 121) (top_statement 122) (trait_declaration_statement 123) (variable 124) (variable_class_name 125) ("{" 126) ("~" 127)) 2 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30727) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 3 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 30725) (encaps_var 5237)) 4 (("$" 4) (T_VARIABLE 81) (simple_variable 30721) ("{" 30722)) 5 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30719) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30718) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30717) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 8 nil 9 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30716) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 10 (("(" 30713)) 11 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30712) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 12 ((T_NAME_FULLY_QUALIFIED 5830) (T_NAME_QUALIFIED 5831) (T_NAME_RELATIVE 5832) (T_STATIC 5833) (T_STRING 5834) (attribute_decl 5835) (attribute_group 30709) (class_name 5837) (name 5838)) 13 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30708) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 14 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 30706) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 15 ((T_STRING 30699)) 16 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30698) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 17 ((T_STRING 30414) (const_decl 30415) (const_list 30696)) 18 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 30694) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 19 (("\"" 1520) ("$" 27911) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 27912) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 27913) ("[" 1542) (array_object_dereferenceable 27914) (callable_expr 27915) (callable_variable 27916) (class_constant 1546) (class_name 27917) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 27918) (name 27919) (new_dereferenceable 1552) (simple_variable 27920) (static_member 27921) (variable 30693) (variable_class_name 27923)) 20 (("(" 30684)) 21 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30678) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 22 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30677) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 23 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 30675) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 24 (("(" 30672)) 25 ((T_STRING 30665)) 26 (("(" 30662)) 27 (("(" 30653) (argument_list 30654) (ctor_arguments 30655)) 28 (("(" 30640)) 29 (("(" 30625)) 30 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 30623) (simple_variable 18987)) 31 ((T_STRING 30621)) 32 (("(" 30618)) 33 (("(" 30461)) 34 (("\"" 1520) ("$" 27911) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 27912) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 27913) ("[" 1542) (array_object_dereferenceable 27914) (callable_expr 27915) (callable_variable 27916) (class_constant 1546) (class_name 27917) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 27918) (name 27919) (new_dereferenceable 1552) (simple_variable 27920) (static_member 27921) (variable 30460) (variable_class_name 27923)) 35 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30459) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 36 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30458) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 37 ((T_STRING 30452)) 38 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30451) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 39 (("(" 30447)) 40 (("(" 30442)) 41 (("(" 30436)) 42 ((T_ABSTRACT 29955) (T_ARRAY 29956) (T_AS 29957) (T_BREAK 29958) (T_CALLABLE 29959) (T_CASE 29960) (T_CATCH 29961) (T_CLASS 29962) (T_CLASS_C 29963) (T_CLONE 29964) (T_CONST 29965) (T_CONTINUE 29966) (T_DECLARE 29967) (T_DEFAULT 29968) (T_DIR 29969) (T_DO 29970) (T_ECHO 29971) (T_ELSE 29972) (T_ELSEIF 29973) (T_EMPTY 29974) (T_ENDDECLARE 29975) (T_ENDFOR 29976) (T_ENDFOREACH 29977) (T_ENDIF 29978) (T_ENDSWITCH 29979) (T_ENDWHILE 29980) (T_ENUM 29981) (T_EVAL 29982) (T_EXIT 29983) (T_EXTENDS 29984) (T_FILE 29985) (T_FINAL 29986) (T_FINALLY 29987) (T_FN 29988) (T_FOR 29989) (T_FOREACH 29990) (T_FUNCTION 29991) (T_FUNC_C 29992) (T_GLOBAL 29993) (T_GOTO 29994) (T_IF 29995) (T_IMPLEMENTS 29996) (T_INCLUDE 29997) (T_INCLUDE_ONCE 29998) (T_INSTANCEOF 29999) (T_INSTEADOF 30000) (T_INTERFACE 30001) (T_ISSET 30002) (T_LINE 30003) (T_LIST 30004) (T_LOGICAL_AND 30005) (T_LOGICAL_OR 30006) (T_LOGICAL_XOR 30007) (T_MATCH 30008) (T_METHOD_C 30009) (T_NAMESPACE 30010) (T_NAME_QUALIFIED 30011) (T_NEW 30012) (T_NS_C 30013) (T_PRINT 30014) (T_PRIVATE 30015) (T_PROPERTY_C 30016) (T_PROTECTED 30017) (T_PUBLIC 30018) (T_READONLY 30019) (T_REQUIRE 30020) (T_REQUIRE_ONCE 30021) (T_RETURN 30022) (T_STATIC 30023) (T_STRING 30024) (T_SWITCH 30025) (T_THROW 30026) (T_TRAIT 30027) (T_TRAIT_C 30028) (T_TRY 30029) (T_UNSET 30030) (T_USE 30031) (T_VAR 30032) (T_WHILE 30033) (T_YIELD 30034) (identifier 30035) (namespace_declaration_name 30036) (reserved_non_modifiers 30037) (semi_reserved 30038) ("{" 30039)) 43 (("$" 27911) ("(" 29918) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 29919) (T_NAME_QUALIFIED 29920) (T_NAME_RELATIVE 29921) (T_READONLY 2931) (T_STATIC 29922) (T_STRING 29923) (T_VARIABLE 27913) (anonymous_class 29924) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 29925) (attribute 2937) (attributes 29926) (class_modifier 2939) (class_name 29927) (class_name_reference 29928) (name 29929) (new_variable 29930) (simple_variable 29931)) 44 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 29917) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 45 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 29916) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 46 (("(" 27778) (argument_list 27779)) 47 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 29915) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 48 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 29914) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 49 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 29912) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 50 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 29907) (T_END_HEREDOC 29908) (T_VARIABLE 2896) (encaps_list 29909) (encaps_var 2898)) 51 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 29905)) 52 ((":" 29904)) 53 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 29903) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 54 (("(" 29885)) 55 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 29884) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 56 ((T_STRING 29879)) 57 (("{" 29862)) 58 (("(" 29857)) 59 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 29856) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 60 ((T_CONST 29803) (T_FUNCTION 29804) (T_NAME_FULLY_QUALIFIED 29805) (T_NAME_QUALIFIED 29806) (T_STRING 29807) (legacy_namespace_name 29808) (mixed_group_use_declaration 29809) (namespace_name 29810) (use_declaration 29811) (use_declarations 29812) (use_type 29813)) 61 (("(" 29794)) 62 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29336) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 63 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 29272) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 64 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 29268) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 65 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 29266) (encaps_list 2804) (encaps_var 2805)) 66 ((T_ELSE 29259) (T_ELSEIF 17291) (T_ENDIF 29260)) 67 ((T_NULLSAFE_OBJECT_OPERATOR 29245) (T_OBJECT_OPERATOR 29246) ("[" 29247)) 68 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 29188) (T_ENUM 29189) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 29190) (T_READONLY 2931) (T_STATIC 17376) (T_TRAIT 29191) (attribute 17377) (attributed_statement 29192) (class_declaration_statement 29193) (class_modifier 97) (class_modifiers 29194) (enum_declaration_statement 29195) (fn 104) #'29196 (function_declaration_statement 29197) (inline_function 17379) (interface_declaration_statement 29198) (trait_declaration_statement 29199)) 69 (("(" 27778) (argument_list 29187)) 70 ((T_ABSTRACT 10) (T_CLASS 29179) (T_FINAL 34) (T_READONLY 2931) (class_modifier 17368)) 71 ((T_PAAMAYIM_NEKUDOTAYIM 29171)) 72 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 29170) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 73 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 28123)) 74 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17349) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17350) (ampersand 17351) (returns_ref 28111)) 75 ((T_ELSE 27961) (T_ELSEIF 27962)) 76 (("(" 27778) (argument_list 27960)) 77 (("=" 27880) (T_AND_EQUAL 27881) (T_COALESCE_EQUAL 27882) (T_CONCAT_EQUAL 27883) (T_DEC 27884) (T_DIV_EQUAL 27885) (T_INC 27886) (T_MINUS_EQUAL 27887) (T_MOD_EQUAL 27888) (T_MUL_EQUAL 27889) (T_OR_EQUAL 27890) (T_PLUS_EQUAL 27891) (T_POW_EQUAL 27892) (T_SL_EQUAL 27893) (T_SR_EQUAL 27894) (T_XOR_EQUAL 27895)) 78 ((T_PAAMAYIM_NEKUDOTAYIM 27791)) 79 ((inner_statement_list 27789)) 80 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 135) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 81 ((T_NAME_FULLY_QUALIFIED 5830) (T_NAME_QUALIFIED 5831) (T_NAME_RELATIVE 5832) (T_STATIC 5833) (T_STRING 5834) (attribute_decl 5835) (attribute_group 27786) (class_name 5837) (name 5838)) 82 ((T_FN 35) (T_FUNCTION 129) (fn 104) #'136 (inline_function 17934)) 83 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 17376) (attribute 2733) (fn 104) #'136 (inline_function 17379)) 84 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 85 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 140)) 86 ((backup_doc_comment 141)) 87 (("(" 142)) 88 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 162)) 89 ((T_NAME_FULLY_QUALIFIED 5830) (T_NAME_QUALIFIED 5831) (T_NAME_RELATIVE 5832) (T_STATIC 5833) (T_STRING 5834) (attribute_decl 5835) (attribute_group 27775) (class_name 5837) (name 5838)) 90 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 27773) (member_modifier 157) (non_empty_member_modifiers 158) (optional_cpp_modifiers 160) (parameter 27774)) 91 ((T_ABSTRACT 143) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (member_modifier 27772)) 92 (("," 27769) (possible_comma 27770)) 93 (("(" 2995) ("?" 26672) (T_ARRAY 26673) (T_CALLABLE 26674) (T_NAME_FULLY_QUALIFIED 26675) (T_NAME_QUALIFIED 26676) (T_NAME_RELATIVE 26677) (T_STRING 26678) (intersection_type_without_static 26679) (name 26680) (optional_type_without_static 26681) (type_expr_without_static 26682) (type_without_static 26683) (union_type_without_static 26684) (union_type_without_static_element 26685)) 94 ((")" 163)) 95 ((T_USE 164) (lexical_vars 165)) 96 (("(" 26660)) 97 ((":" 166) (return_type 167)) 98 (("(" 2256) ("?" 26605) (T_ARRAY 26606) (T_CALLABLE 26607) (T_NAME_FULLY_QUALIFIED 26608) (T_NAME_QUALIFIED 26609) (T_NAME_RELATIVE 26610) (T_STATIC 26611) (T_STRING 26612) (intersection_type 26613) (name 26614) (type 26615) (type_expr 26616) (type_without_static 26617) (union_type 26618) (union_type_element 26619)) 99 ((backup_fn_flags 168)) 100 (("{" 169)) 101 ((inner_statement_list 170)) 102 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 212) ("~" 127)) 103 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 26603) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 104 ((T_STRING 26596)) 105 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 26594) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 106 (("(" 26585)) 107 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 26579) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 108 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 26577) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 109 ((T_STRING 26570)) 110 (("(" 26557)) 111 (("(" 26542)) 112 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 26540) (simple_variable 18987)) 113 ((T_STRING 26538)) 114 (("(" 26535)) 115 (("(" 26531)) 116 ((T_STRING 26525)) 117 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 26523) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 118 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 26521)) 119 ((":" 26520)) 120 (("(" 26502)) 121 ((T_STRING 26497)) 122 (("{" 26480)) 123 (("(" 26475)) 124 (("(" 26466)) 125 ((T_ELSE 26459) (T_ELSEIF 17291) (T_ENDIF 26460)) 126 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 173) (T_ENUM 178) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 186) (T_READONLY 2931) (T_STATIC 17376) (T_TRAIT 191) (attribute 17377) (attributed_statement 26458) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (enum_declaration_statement 201) (fn 104) #'203 (function_declaration_statement 204) (inline_function 17379) (interface_declaration_statement 208) (trait_declaration_statement 210)) 127 ((T_ABSTRACT 10) (T_CLASS 26450) (T_FINAL 34) (T_READONLY 2931) (class_modifier 17368)) 128 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 26449) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 129 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17349) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17350) (ampersand 17351) (returns_ref 26437)) 130 ((T_ELSE 216) (T_ELSEIF 217)) 131 ((inner_statement_list 214)) 132 ((backup_fn_flags 213)) 133 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 215) ("~" 127)) 134 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 202) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 26436) (static_member 121) (variable 124) (variable_class_name 125) ("{" 211) ("~" 127)) 135 (("(" 218)) 136 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 282) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 137 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26435) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 138 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 26433) (encaps_var 5237)) 139 (("$" 221) (T_VARIABLE 269) (simple_variable 26429) ("{" 26430)) 140 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26427) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 141 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26426) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 142 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26425) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 143 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26424) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 144 (("(" 26421)) 145 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26420) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 146 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26419) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 147 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26418) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 148 (("\"" 1520) ("$" 16150) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 16151) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 16152) ("[" 1542) (array_object_dereferenceable 16153) (callable_expr 16154) (callable_variable 16155) (class_constant 1546) (class_name 16156) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 16157) (name 16158) (new_dereferenceable 1552) (simple_variable 16159) (static_member 16160) (variable 26417) (variable_class_name 16162)) 149 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26416) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 150 (("(" 26413)) 151 (("(" 26410)) 152 (("(" 26401) (argument_list 26402) (ctor_arguments 26403)) 153 (("\"" 1520) ("$" 16150) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 16151) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 16152) ("[" 1542) (array_object_dereferenceable 16153) (callable_expr 16154) (callable_variable 16155) (class_constant 1546) (class_name 16156) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 16157) (name 16158) (new_dereferenceable 1552) (simple_variable 16159) (static_member 16160) (variable 26400) (variable_class_name 16162)) 154 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26399) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 155 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26398) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 156 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26397) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 157 (("(" 26393)) 158 (("(" 26388)) 159 (("(" 26382)) 160 (("$" 16150) ("(" 26345) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 26346) (T_NAME_QUALIFIED 26347) (T_NAME_RELATIVE 26348) (T_READONLY 2931) (T_STATIC 26349) (T_STRING 26350) (T_VARIABLE 16152) (anonymous_class 26351) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 26352) (attribute 2937) (attributes 26353) (class_modifier 2939) (class_name 26354) (class_name_reference 26355) (name 26356) (new_variable 26357) (simple_variable 26358)) 161 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26344) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 162 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26343) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 163 (("(" 16111) (argument_list 26342)) 164 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26341) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 165 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26340) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 166 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 26335) (T_END_HEREDOC 26336) (T_VARIABLE 2896) (encaps_list 26337) (encaps_var 2898)) 167 ((T_FN 35) (T_FUNCTION 129) (fn 283) #'284 (inline_function 26334)) 168 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26333) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 169 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26332) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 170 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26331) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 171 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25873) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 172 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25809) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 173 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 25805) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 174 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 25803) (encaps_list 2804) (encaps_var 2805)) 175 ((T_NULLSAFE_OBJECT_OPERATOR 25789) (T_OBJECT_OPERATOR 25790) ("[" 25791)) 176 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 25786) (attribute 2733) (fn 283) #'284 (inline_function 25787)) 177 (("(" 16111) (argument_list 25785)) 178 ((T_PAAMAYIM_NEKUDOTAYIM 25777)) 179 (("%" 299) (")" 17259) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 180 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 16212)) 181 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 16200)) 182 (("(" 16111) (argument_list 16199)) 183 (("=" 16119) (T_AND_EQUAL 16120) (T_COALESCE_EQUAL 16121) (T_CONCAT_EQUAL 16122) (T_DEC 16123) (T_DIV_EQUAL 16124) (T_INC 16125) (T_MINUS_EQUAL 16126) (T_MOD_EQUAL 16127) (T_MUL_EQUAL 16128) (T_OR_EQUAL 16129) (T_PLUS_EQUAL 16130) (T_POW_EQUAL 16131) (T_SL_EQUAL 16132) (T_SR_EQUAL 16133) (T_XOR_EQUAL 16134)) 184 ((T_PAAMAYIM_NEKUDOTAYIM 16023)) 185 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 298) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 186 (("%" 299) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 187 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16022) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 188 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16021) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 189 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16020) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 190 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16019) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 191 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16018) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 192 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16017) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 193 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16016) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 194 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16015) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 195 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 16010) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 16011) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 196 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16009) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 197 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16008) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 198 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16007) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 199 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16006) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 200 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16005) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 201 (("$" 344) ("(" 345) (T_NAME_FULLY_QUALIFIED 346) (T_NAME_QUALIFIED 347) (T_NAME_RELATIVE 348) (T_STATIC 349) (T_STRING 350) (T_VARIABLE 351) (class_name 352) (class_name_reference 353) (name 354) (new_variable 355) (simple_variable 356)) 202 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 343) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 203 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 342) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 204 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 341) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 205 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 340) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 206 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 339) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 207 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 338) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 208 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 337) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 209 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 336) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 210 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 335) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 211 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 334) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 212 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 333) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 213 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 332) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 214 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 331) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 215 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 330) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 216 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 329) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 217 (("$" 344) (T_VARIABLE 351) (simple_variable 16001) ("{" 16002)) 218 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15999) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 219 ((T_PAAMAYIM_NEKUDOTAYIM 15997)) 220 ((T_NULLSAFE_OBJECT_OPERATOR 357) (T_OBJECT_OPERATOR 358) (T_PAAMAYIM_NEKUDOTAYIM 359) ("[" 360)) 221 (("$" 344) (T_STRING 15990) (T_VARIABLE 351) (property_name 15996) (simple_variable 15992) ("{" 15993)) 222 (("$" 344) (T_STRING 15990) (T_VARIABLE 351) (property_name 15991) (simple_variable 15992) ("{" 15993)) 223 (("$" 344) (T_VARIABLE 351) (simple_variable 15989)) 224 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 434) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 225 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15988) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 226 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 15986) (encaps_var 5237)) 227 (("$" 363) (T_VARIABLE 411) (simple_variable 15982) ("{" 15983)) 228 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15980) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 229 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15979) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 230 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15978) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 231 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15977) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 232 (("(" 15974)) 233 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15973) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 234 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15972) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 235 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15971) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 236 (("\"" 1520) ("$" 14220) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 14221) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 14222) ("[" 1542) (array_object_dereferenceable 14223) (callable_expr 14224) (callable_variable 14225) (class_constant 1546) (class_name 14226) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 14227) (name 14228) (new_dereferenceable 1552) (simple_variable 14229) (static_member 14230) (variable 15970) (variable_class_name 14232)) 237 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15969) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 238 (("(" 15966)) 239 (("(" 15963)) 240 (("(" 15954) (argument_list 15955) (ctor_arguments 15956)) 241 (("\"" 1520) ("$" 14220) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 14221) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 14222) ("[" 1542) (array_object_dereferenceable 14223) (callable_expr 14224) (callable_variable 14225) (class_constant 1546) (class_name 14226) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 14227) (name 14228) (new_dereferenceable 1552) (simple_variable 14229) (static_member 14230) (variable 15953) (variable_class_name 14232)) 242 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15952) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 243 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15951) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 244 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15950) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 245 (("(" 15946)) 246 (("(" 15941)) 247 (("(" 15935)) 248 (("$" 14220) ("(" 15898) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 15899) (T_NAME_QUALIFIED 15900) (T_NAME_RELATIVE 15901) (T_READONLY 2931) (T_STATIC 15902) (T_STRING 15903) (T_VARIABLE 14222) (anonymous_class 15904) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 15905) (attribute 2937) (attributes 15906) (class_modifier 2939) (class_name 15907) (class_name_reference 15908) (name 15909) (new_variable 15910) (simple_variable 15911)) 249 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15897) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 250 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15896) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 251 (("(" 14181) (argument_list 15895)) 252 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15894) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 253 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15893) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 254 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 15888) (T_END_HEREDOC 15889) (T_VARIABLE 2896) (encaps_list 15890) (encaps_var 2898)) 255 ((T_FN 35) (T_FUNCTION 129) (fn 425) #'426 (inline_function 15887)) 256 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15886) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 257 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15885) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 258 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15884) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 259 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15426) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 260 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15362) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 261 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 15358) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 262 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 15356) (encaps_list 2804) (encaps_var 2805)) 263 ((T_NULLSAFE_OBJECT_OPERATOR 15342) (T_OBJECT_OPERATOR 15343) ("[" 15344)) 264 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 15339) (attribute 2733) (fn 425) #'426 (inline_function 15340)) 265 (("(" 14181) (argument_list 15338)) 266 ((T_PAAMAYIM_NEKUDOTAYIM 15330)) 267 (("%" 442) ("*" 443) ("+" 444) ("-" 445) ("." 446) ("/" 447) ("<" 448) (">" 449) ("?" 450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 452) (T_BOOLEAN_AND 453) (T_BOOLEAN_OR 454) (T_COALESCE 455) (T_INSTANCEOF 456) (T_IS_EQUAL 457) (T_IS_GREATER_OR_EQUAL 458) (T_IS_IDENTICAL 459) (T_IS_NOT_EQUAL 460) (T_IS_NOT_IDENTICAL 461) (T_IS_SMALLER_OR_EQUAL 462) (T_LOGICAL_AND 463) (T_LOGICAL_OR 464) (T_LOGICAL_XOR 465) (T_POW 466) (T_SL 467) (T_SPACESHIP 468) (T_SR 469) ("^" 470) ("|" 471)) 268 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 14283)) 269 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 14271)) 270 (("(" 14181) (argument_list 14270)) 271 (("]" 14269)) 272 (("=" 14189) (T_AND_EQUAL 14190) (T_COALESCE_EQUAL 14191) (T_CONCAT_EQUAL 14192) (T_DEC 14193) (T_DIV_EQUAL 14194) (T_INC 14195) (T_MINUS_EQUAL 14196) (T_MOD_EQUAL 14197) (T_MUL_EQUAL 14198) (T_OR_EQUAL 14199) (T_PLUS_EQUAL 14200) (T_POW_EQUAL 14201) (T_SL_EQUAL 14202) (T_SR_EQUAL 14203) (T_XOR_EQUAL 14204)) 273 ((T_PAAMAYIM_NEKUDOTAYIM 14093)) 274 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 441) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 275 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14092) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 276 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14091) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 277 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14090) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 278 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14089) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 279 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14088) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 280 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14087) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 281 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14086) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 282 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14085) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 283 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 14080) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 14081) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 284 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14079) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 285 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14078) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 286 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14077) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 287 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14076) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 288 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14075) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 289 (("$" 487) ("(" 488) (T_NAME_FULLY_QUALIFIED 489) (T_NAME_QUALIFIED 490) (T_NAME_RELATIVE 491) (T_STATIC 492) (T_STRING 493) (T_VARIABLE 494) (class_name 495) (class_name_reference 496) (name 497) (new_variable 498) (simple_variable 499)) 290 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 486) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 291 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 485) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 292 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 484) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 293 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 483) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 294 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 482) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 295 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 481) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 296 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 480) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 297 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 479) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 298 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 478) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 299 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 477) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 300 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 476) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 301 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 475) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 302 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 474) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 303 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 473) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 304 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 472) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 305 (("$" 487) (T_VARIABLE 494) (simple_variable 14071) ("{" 14072)) 306 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 14069) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 307 ((T_PAAMAYIM_NEKUDOTAYIM 14067)) 308 ((T_NULLSAFE_OBJECT_OPERATOR 500) (T_OBJECT_OPERATOR 501) (T_PAAMAYIM_NEKUDOTAYIM 502) ("[" 503)) 309 (("$" 487) (T_STRING 507) (T_VARIABLE 494) (property_name 14066) (simple_variable 509) ("{" 510)) 310 (("$" 487) (T_STRING 507) (T_VARIABLE 494) (property_name 508) (simple_variable 509) ("{" 510)) 311 (("$" 487) (T_VARIABLE 494) (simple_variable 506)) 312 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 504) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 313 (("]" 505)) 314 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 574) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 315 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14065) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 316 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 14063) (encaps_var 5237)) 317 (("$" 513) (T_VARIABLE 561) (simple_variable 14059) ("{" 14060)) 318 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 14057) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 319 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14056) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 320 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14055) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 321 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14054) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 322 (("(" 14051)) 323 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14050) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 324 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14049) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 325 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14048) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 326 (("\"" 1520) ("$" 12297) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 12298) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 12299) ("[" 1542) (array_object_dereferenceable 12300) (callable_expr 12301) (callable_variable 12302) (class_constant 1546) (class_name 12303) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 12304) (name 12305) (new_dereferenceable 1552) (simple_variable 12306) (static_member 12307) (variable 14047) (variable_class_name 12309)) 327 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14046) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 328 (("(" 14043)) 329 (("(" 14040)) 330 (("(" 14031) (argument_list 14032) (ctor_arguments 14033)) 331 (("\"" 1520) ("$" 12297) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 12298) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 12299) ("[" 1542) (array_object_dereferenceable 12300) (callable_expr 12301) (callable_variable 12302) (class_constant 1546) (class_name 12303) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 12304) (name 12305) (new_dereferenceable 1552) (simple_variable 12306) (static_member 12307) (variable 14030) (variable_class_name 12309)) 332 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14029) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 333 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14028) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 334 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14027) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 335 (("(" 14023)) 336 (("(" 14018)) 337 (("(" 14012)) 338 (("$" 12297) ("(" 13975) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 13976) (T_NAME_QUALIFIED 13977) (T_NAME_RELATIVE 13978) (T_READONLY 2931) (T_STATIC 13979) (T_STRING 13980) (T_VARIABLE 12299) (anonymous_class 13981) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 13982) (attribute 2937) (attributes 13983) (class_modifier 2939) (class_name 13984) (class_name_reference 13985) (name 13986) (new_variable 13987) (simple_variable 13988)) 339 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13974) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 340 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13973) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 341 (("(" 12258) (argument_list 13972)) 342 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13971) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 343 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13970) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 344 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 13965) (T_END_HEREDOC 13966) (T_VARIABLE 2896) (encaps_list 13967) (encaps_var 2898)) 345 ((T_FN 35) (T_FUNCTION 129) (fn 575) #'576 (inline_function 13964)) 346 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13963) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 347 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13962) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 348 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13961) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 349 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13503) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 350 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13439) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 351 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 13435) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 352 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 13433) (encaps_list 2804) (encaps_var 2805)) 353 ((T_NULLSAFE_OBJECT_OPERATOR 13419) (T_OBJECT_OPERATOR 13420) ("[" 13421)) 354 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 13416) (attribute 2733) (fn 575) #'576 (inline_function 13417)) 355 (("(" 12258) (argument_list 13415)) 356 ((T_PAAMAYIM_NEKUDOTAYIM 13407)) 357 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13406)) 358 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 12359)) 359 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 12347)) 360 (("(" 12258) (argument_list 12346)) 361 (("=" 12266) (T_AND_EQUAL 12267) (T_COALESCE_EQUAL 12268) (T_CONCAT_EQUAL 12269) (T_DEC 12270) (T_DIV_EQUAL 12271) (T_INC 12272) (T_MINUS_EQUAL 12273) (T_MOD_EQUAL 12274) (T_MUL_EQUAL 12275) (T_OR_EQUAL 12276) (T_PLUS_EQUAL 12277) (T_POW_EQUAL 12278) (T_SL_EQUAL 12279) (T_SR_EQUAL 12280) (T_XOR_EQUAL 12281)) 362 ((T_PAAMAYIM_NEKUDOTAYIM 12170)) 363 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 590) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 364 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620)) 365 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12169) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 366 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12168) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 367 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12167) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 368 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12166) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 369 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12165) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 370 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12164) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 371 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12163) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 372 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12162) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 373 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 682) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 740) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 374 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 675) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 375 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 674) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 376 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 673) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 377 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 672) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 378 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 671) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 379 (("$" 636) ("(" 637) (T_NAME_FULLY_QUALIFIED 638) (T_NAME_QUALIFIED 639) (T_NAME_RELATIVE 640) (T_STATIC 641) (T_STRING 642) (T_VARIABLE 643) (class_name 644) (class_name_reference 645) (name 646) (new_variable 647) (simple_variable 648)) 380 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 635) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 381 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 634) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 382 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 633) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 383 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 632) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 384 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 631) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 385 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 630) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 386 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 629) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 387 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 628) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 388 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 627) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 389 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 626) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 390 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 625) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 391 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 624) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 392 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 623) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 393 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 622) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 394 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 621) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 395 (("$" 636) (T_VARIABLE 643) (simple_variable 667) ("{" 668)) 396 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 665) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 397 ((T_PAAMAYIM_NEKUDOTAYIM 663)) 398 ((T_NULLSAFE_OBJECT_OPERATOR 649) (T_OBJECT_OPERATOR 650) (T_PAAMAYIM_NEKUDOTAYIM 651) ("[" 652)) 399 (("$" 636) (T_STRING 656) (T_VARIABLE 643) (property_name 662) (simple_variable 658) ("{" 659)) 400 (("$" 636) (T_STRING 656) (T_VARIABLE 643) (property_name 657) (simple_variable 658) ("{" 659)) 401 (("$" 636) (T_VARIABLE 643) (simple_variable 655)) 402 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 653) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 403 (("]" 654)) 404 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 660) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 405 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 661)) 406 (("$" 636) (T_VARIABLE 643) (simple_variable 664)) 407 (("%" 299) (")" 666) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 408 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 669) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 409 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 670)) 410 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12161) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 411 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 12159) (encaps_var 5237)) 412 (("$" 678) (T_VARIABLE 727) (simple_variable 12155) ("{" 12156)) 413 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 12153) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 414 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12152) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 415 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12151) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 416 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12150) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 417 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12149) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 418 (("(" 12146)) 419 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12145) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 420 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12144) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 421 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12143) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 422 (("\"" 1520) ("$" 10391) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 10392) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 10393) ("[" 1542) (array_object_dereferenceable 10394) (callable_expr 10395) (callable_variable 10396) (class_constant 1546) (class_name 10397) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 10398) (name 10399) (new_dereferenceable 1552) (simple_variable 10400) (static_member 10401) (variable 12142) (variable_class_name 10403)) 423 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12141) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 424 (("(" 12138)) 425 (("(" 12135)) 426 (("(" 12126) (argument_list 12127) (ctor_arguments 12128)) 427 (("\"" 1520) ("$" 10391) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 10392) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 10393) ("[" 1542) (array_object_dereferenceable 10394) (callable_expr 10395) (callable_variable 10396) (class_constant 1546) (class_name 10397) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 10398) (name 10399) (new_dereferenceable 1552) (simple_variable 10400) (static_member 10401) (variable 12125) (variable_class_name 10403)) 428 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12124) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 429 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12123) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 430 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12122) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 431 (("(" 12118)) 432 (("(" 12113)) 433 (("(" 12107)) 434 (("$" 10391) ("(" 12070) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 12071) (T_NAME_QUALIFIED 12072) (T_NAME_RELATIVE 12073) (T_READONLY 2931) (T_STATIC 12074) (T_STRING 12075) (T_VARIABLE 10393) (anonymous_class 12076) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 12077) (attribute 2937) (attributes 12078) (class_modifier 2939) (class_name 12079) (class_name_reference 12080) (name 12081) (new_variable 12082) (simple_variable 12083)) 435 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12069) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 436 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12068) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 437 (("(" 943) (argument_list 12067)) 438 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12066) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 439 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12065) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 440 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 12060) (T_END_HEREDOC 12061) (T_VARIABLE 2896) (encaps_list 12062) (encaps_var 2898)) 441 ((T_FN 35) (T_FUNCTION 129) (fn 741) #'742 (inline_function 12059)) 442 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12058) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 443 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12057) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 444 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12056) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 445 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11598) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 446 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 11534) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 447 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 11530) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 448 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 11528) (encaps_list 2804) (encaps_var 2805)) 449 ((T_NULLSAFE_OBJECT_OPERATOR 11514) (T_OBJECT_OPERATOR 11515) ("[" 11516)) 450 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 11511) (attribute 2733) (fn 741) #'742 (inline_function 11512)) 451 (("(" 943) (argument_list 11510)) 452 ((T_PAAMAYIM_NEKUDOTAYIM 11502)) 453 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 11500) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 454 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 10453)) 455 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 10441)) 456 (("(" 943) (argument_list 10440)) 457 (("=" 10360) (T_AND_EQUAL 10361) (T_COALESCE_EQUAL 10362) (T_CONCAT_EQUAL 10363) (T_DEC 10364) (T_DIV_EQUAL 10365) (T_INC 10366) (T_MINUS_EQUAL 10367) (T_MOD_EQUAL 10368) (T_MUL_EQUAL 10369) (T_OR_EQUAL 10370) (T_PLUS_EQUAL 10371) (T_POW_EQUAL 10372) (T_SL_EQUAL 10373) (T_SR_EQUAL 10374) (T_XOR_EQUAL 10375)) 458 ((T_PAAMAYIM_NEKUDOTAYIM 855)) 459 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 756) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 460 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 461 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 854) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 462 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 853) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 463 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 852) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 464 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 851) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 465 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 850) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 466 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 849) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 467 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 848) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 468 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 847) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 469 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 842) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 843) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 470 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 841) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 471 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 840) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 472 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 839) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 473 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 838) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 474 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 837) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 475 (("$" 802) ("(" 803) (T_NAME_FULLY_QUALIFIED 804) (T_NAME_QUALIFIED 805) (T_NAME_RELATIVE 806) (T_STATIC 807) (T_STRING 808) (T_VARIABLE 809) (class_name 810) (class_name_reference 811) (name 812) (new_variable 813) (simple_variable 814)) 476 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 801) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 477 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 800) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 478 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 799) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 479 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 798) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 480 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 797) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 481 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 796) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 482 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 795) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 483 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 794) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 484 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 793) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 485 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 792) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 486 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 791) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 487 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 790) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 488 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 789) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 489 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 788) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 490 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 787) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 491 (("$" 802) (T_VARIABLE 809) (simple_variable 833) ("{" 834)) 492 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 831) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 493 ((T_PAAMAYIM_NEKUDOTAYIM 829)) 494 ((T_NULLSAFE_OBJECT_OPERATOR 815) (T_OBJECT_OPERATOR 816) (T_PAAMAYIM_NEKUDOTAYIM 817) ("[" 818)) 495 (("$" 802) (T_STRING 822) (T_VARIABLE 809) (property_name 828) (simple_variable 824) ("{" 825)) 496 (("$" 802) (T_STRING 822) (T_VARIABLE 809) (property_name 823) (simple_variable 824) ("{" 825)) 497 (("$" 802) (T_VARIABLE 809) (simple_variable 821)) 498 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 819) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 499 (("]" 820)) 500 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 826) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 501 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 827)) 502 (("$" 802) (T_VARIABLE 809) (simple_variable 830)) 503 (("%" 299) (")" 832) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 504 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 835) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 505 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 836)) 506 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 846) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 507 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 844) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 508 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 845) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 509 (("$" 678) (T_ABSTRACT 856) (T_ARRAY 857) (T_AS 858) (T_BREAK 859) (T_CALLABLE 860) (T_CASE 861) (T_CATCH 862) (T_CLASS 863) (T_CLASS_C 864) (T_CLONE 865) (T_CONST 866) (T_CONTINUE 867) (T_DECLARE 868) (T_DEFAULT 869) (T_DIR 870) (T_DO 871) (T_ECHO 872) (T_ELSE 873) (T_ELSEIF 874) (T_EMPTY 875) (T_ENDDECLARE 876) (T_ENDFOR 877) (T_ENDFOREACH 878) (T_ENDIF 879) (T_ENDSWITCH 880) (T_ENDWHILE 881) (T_ENUM 882) (T_EVAL 883) (T_EXIT 884) (T_EXTENDS 885) (T_FILE 886) (T_FINAL 887) (T_FINALLY 888) (T_FN 889) (T_FOR 890) (T_FOREACH 891) (T_FUNCTION 892) (T_FUNC_C 893) (T_GLOBAL 894) (T_GOTO 895) (T_IF 896) (T_IMPLEMENTS 897) (T_INCLUDE 898) (T_INCLUDE_ONCE 899) (T_INSTANCEOF 900) (T_INSTEADOF 901) (T_INTERFACE 902) (T_ISSET 903) (T_LINE 904) (T_LIST 905) (T_LOGICAL_AND 906) (T_LOGICAL_OR 907) (T_LOGICAL_XOR 908) (T_MATCH 909) (T_METHOD_C 910) (T_NAMESPACE 911) (T_NEW 912) (T_NS_C 913) (T_PRINT 914) (T_PRIVATE 915) (T_PROPERTY_C 916) (T_PROTECTED 917) (T_PUBLIC 918) (T_READONLY 919) (T_REQUIRE 920) (T_REQUIRE_ONCE 921) (T_RETURN 922) (T_STATIC 923) (T_STRING 924) (T_SWITCH 925) (T_THROW 926) (T_TRAIT 927) (T_TRAIT_C 928) (T_TRY 929) (T_UNSET 930) (T_USE 931) (T_VAR 932) (T_VARIABLE 727) (T_WHILE 933) (T_YIELD 934) (identifier 935) (member_name 936) (reserved_non_modifiers 937) (semi_reserved 938) (simple_variable 939) ("{" 940)) 510 (("(" 943) (argument_list 944)) 511 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 941) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 512 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 942)) 513 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 949) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 976) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 1073) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 514 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10359) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 515 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 10357) (encaps_var 5237)) 516 (("$" 947) (T_VARIABLE 1047) (simple_variable 10353) ("{" 10354)) 517 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 10351) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 518 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10350) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 519 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10349) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 520 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10348) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 521 (("(" 10118)) 522 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10347) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 523 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10346) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 524 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10117) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 525 (("\"" 1520) ("$" 10030) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 10241) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 10037) ("[" 1542) (array_object_dereferenceable 10242) (callable_expr 10243) (callable_variable 10244) (class_constant 1546) (class_name 10245) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 10246) (name 10247) (new_dereferenceable 1552) (simple_variable 10248) (static_member 10249) (variable 10345) (variable_class_name 10251)) 526 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10344) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 527 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 10343) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 528 (("(" 10114)) 529 (("(" 10111)) 530 (("(" 10102) (argument_list 10103) (ctor_arguments 10104)) 531 (("\"" 1520) ("$" 10030) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 10241) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 10037) ("[" 1542) (array_object_dereferenceable 10242) (callable_expr 10243) (callable_variable 10244) (class_constant 1546) (class_name 10245) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 10246) (name 10247) (new_dereferenceable 1552) (simple_variable 10248) (static_member 10249) (variable 10342) (variable_class_name 10251)) 532 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10101) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 533 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10100) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 534 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10341) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 535 (("(" 10096)) 536 (("(" 10091)) 537 (("(" 10085)) 538 (("$" 10030) ("(" 10031) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 10032) (T_NAME_QUALIFIED 10033) (T_NAME_RELATIVE 10034) (T_READONLY 2931) (T_STATIC 10035) (T_STRING 10036) (T_VARIABLE 10037) (anonymous_class 10038) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 10039) (attribute 2937) (attributes 10040) (class_modifier 2939) (class_name 10041) (class_name_reference 10042) (name 10043) (new_variable 10044) (simple_variable 10045)) 539 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10340) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 540 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10029) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 541 (("(" 10021) (argument_list 10022)) 542 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10020) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 543 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10019) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 544 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 10335) (T_END_HEREDOC 10336) (T_VARIABLE 2896) (encaps_list 10337) (encaps_var 2898)) 545 ((T_FN 35) (T_FUNCTION 129) (fn 1063) #'1064 (inline_function 10018)) 546 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10334) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 547 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10017) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 548 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10333) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 549 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1272) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 550 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10332) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 551 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 10328) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 552 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 10326) (encaps_list 2804) (encaps_var 2805)) 553 ((T_NULLSAFE_OBJECT_OPERATOR 10312) (T_OBJECT_OPERATOR 10313) ("[" 10314)) 554 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 10309) (attribute 2733) (fn 1063) #'1064 (inline_function 10310)) 555 (("(" 10021) (argument_list 10308)) 556 ((T_PAAMAYIM_NEKUDOTAYIM 10300)) 557 (("%" 1111) ("*" 1112) ("+" 1113) ("-" 1114) ("." 1115) ("/" 1116) ("<" 1117) (">" 1118) ("?" 1119) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1120) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1121) (T_BOOLEAN_AND 1122) (T_BOOLEAN_OR 1123) (T_COALESCE 1124) (T_INSTANCEOF 1125) (T_IS_EQUAL 1126) (T_IS_GREATER_OR_EQUAL 1127) (T_IS_IDENTICAL 1128) (T_IS_NOT_EQUAL 1129) (T_IS_NOT_IDENTICAL 1130) (T_IS_SMALLER_OR_EQUAL 1131) (T_LOGICAL_AND 1132) (T_LOGICAL_OR 1133) (T_LOGICAL_XOR 1134) (T_POW 1135) (T_SL 1136) (T_SPACESHIP 1137) (T_SR 1138) ("^" 1139) ("|" 1140)) 558 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 10289)) 559 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 10277)) 560 ((":" 10275)) 561 (("(" 10021) (argument_list 10274)) 562 (("," 1480) (possible_comma 10272)) 563 (("=" 10210) (T_AND_EQUAL 10211) (T_COALESCE_EQUAL 10212) (T_CONCAT_EQUAL 10213) (T_DEC 10214) (T_DIV_EQUAL 10215) (T_INC 10216) (T_MINUS_EQUAL 10217) (T_MOD_EQUAL 10218) (T_MUL_EQUAL 10219) (T_OR_EQUAL 10220) (T_PLUS_EQUAL 10221) (T_POW_EQUAL 10222) (T_SL_EQUAL 10223) (T_SR_EQUAL 10224) (T_XOR_EQUAL 10225)) 564 ((T_PAAMAYIM_NEKUDOTAYIM 10121)) 565 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1110) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 566 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1208) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 567 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1207) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 568 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1206) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 569 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1205) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 570 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1204) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 571 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1203) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 572 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1202) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 573 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1201) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 574 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 1196) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 1197) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 575 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1195) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 576 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1194) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 577 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1193) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 578 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1192) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 579 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1191) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 580 (("$" 1156) ("(" 1157) (T_NAME_FULLY_QUALIFIED 1158) (T_NAME_QUALIFIED 1159) (T_NAME_RELATIVE 1160) (T_STATIC 1161) (T_STRING 1162) (T_VARIABLE 1163) (class_name 1164) (class_name_reference 1165) (name 1166) (new_variable 1167) (simple_variable 1168)) 581 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1155) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 582 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1154) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 583 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1153) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 584 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1152) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 585 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1151) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 586 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1150) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 587 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1149) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 588 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1148) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 589 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1147) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 590 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1146) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 591 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1145) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 592 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1144) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 593 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1143) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 594 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1142) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 595 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1141) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 596 (("$" 1156) (T_VARIABLE 1163) (simple_variable 1187) ("{" 1188)) 597 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 1185) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 598 ((T_PAAMAYIM_NEKUDOTAYIM 1183)) 599 ((T_NULLSAFE_OBJECT_OPERATOR 1169) (T_OBJECT_OPERATOR 1170) (T_PAAMAYIM_NEKUDOTAYIM 1171) ("[" 1172)) 600 (("$" 1156) (T_STRING 1176) (T_VARIABLE 1163) (property_name 1182) (simple_variable 1178) ("{" 1179)) 601 (("$" 1156) (T_STRING 1176) (T_VARIABLE 1163) (property_name 1177) (simple_variable 1178) ("{" 1179)) 602 (("$" 1156) (T_VARIABLE 1163) (simple_variable 1175)) 603 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 1173) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 604 (("]" 1174)) 605 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 1180) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 606 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 1181)) 607 (("$" 1156) (T_VARIABLE 1163) (simple_variable 1184)) 608 (("%" 299) (")" 1186) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 609 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 1189) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 610 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 1190)) 611 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1200) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 612 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 1198) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 613 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1199) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 614 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 10016) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 615 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 10014) (encaps_var 5237)) 616 (("$" 1211) (T_VARIABLE 1259) (simple_variable 10010) ("{" 10011)) 617 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 10008) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 618 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 10007) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 619 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 10006) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 620 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 10005) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 621 (("(" 10002)) 622 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 10001) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 623 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 10000) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 624 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9999) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 625 (("\"" 1520) ("$" 1521) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 1537) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 1541) ("[" 1542) (array_object_dereferenceable 1543) (callable_expr 1544) (callable_variable 1545) (class_constant 1546) (class_name 1547) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 1550) (name 1551) (new_dereferenceable 1552) (simple_variable 1553) (static_member 1554) (variable 9998) (variable_class_name 1556)) 626 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9997) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 627 (("(" 9994)) 628 (("(" 9991)) 629 (("(" 9982) (argument_list 9983) (ctor_arguments 9984)) 630 (("\"" 1520) ("$" 1521) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 1537) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 1541) ("[" 1542) (array_object_dereferenceable 1543) (callable_expr 1544) (callable_variable 1545) (class_constant 1546) (class_name 1547) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 1550) (name 1551) (new_dereferenceable 1552) (simple_variable 1553) (static_member 1554) (variable 9981) (variable_class_name 1556)) 631 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9980) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 632 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9979) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 633 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9978) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 634 (("(" 9974)) 635 (("(" 9969)) 636 (("(" 9963)) 637 (("$" 1521) ("(" 9926) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 9927) (T_NAME_QUALIFIED 9928) (T_NAME_RELATIVE 9929) (T_READONLY 2931) (T_STATIC 9930) (T_STRING 9931) (T_VARIABLE 1541) (anonymous_class 9932) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 9933) (attribute 2937) (attributes 9934) (class_modifier 2939) (class_name 9935) (class_name_reference 9936) (name 9937) (new_variable 9938) (simple_variable 9939)) 638 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9925) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 639 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9924) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 640 (("(" 1475) (argument_list 9923)) 641 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9922) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 642 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9921) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 643 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 9916) (T_END_HEREDOC 9917) (T_VARIABLE 2896) (encaps_list 9918) (encaps_var 2898)) 644 ((T_FN 35) (T_FUNCTION 129) (fn 1273) #'1274 (inline_function 9915)) 645 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9914) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 646 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9913) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 647 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9912) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 648 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9909) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 649 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9908) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 650 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 9904) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 651 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 9902) (encaps_list 2804) (encaps_var 2805)) 652 ((T_NULLSAFE_OBJECT_OPERATOR 9888) (T_OBJECT_OPERATOR 9889) ("[" 9890)) 653 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 9885) (attribute 2733) (fn 1273) #'1274 (inline_function 9886)) 654 (("(" 1475) (argument_list 9884)) 655 ((T_PAAMAYIM_NEKUDOTAYIM 9876)) 656 (("%" 1289) ("*" 1290) ("+" 1291) ("-" 1292) ("." 1293) ("/" 1294) ("<" 1295) (">" 1296) ("?" 1297) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1298) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1299) (T_BOOLEAN_AND 1300) (T_BOOLEAN_OR 1301) (T_COALESCE 1302) (T_DOUBLE_ARROW 9874) (T_INSTANCEOF 1303) (T_IS_EQUAL 1304) (T_IS_GREATER_OR_EQUAL 1305) (T_IS_IDENTICAL 1306) (T_IS_NOT_EQUAL 1307) (T_IS_NOT_IDENTICAL 1308) (T_IS_SMALLER_OR_EQUAL 1309) (T_LOGICAL_AND 1310) (T_LOGICAL_OR 1311) (T_LOGICAL_XOR 1312) (T_POW 1313) (T_SL 1314) (T_SPACESHIP 1315) (T_SR 1316) ("^" 1317) ("|" 1318)) 657 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 9863)) 658 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 9851)) 659 (("(" 1475) (argument_list 9850)) 660 (("=" 1487) (T_AND_EQUAL 1488) (T_COALESCE_EQUAL 1489) (T_CONCAT_EQUAL 1490) (T_DEC 1491) (T_DIV_EQUAL 1492) (T_INC 1493) (T_MINUS_EQUAL 1494) (T_MOD_EQUAL 1495) (T_MUL_EQUAL 1496) (T_OR_EQUAL 1497) (T_PLUS_EQUAL 1498) (T_POW_EQUAL 1499) (T_SL_EQUAL 1500) (T_SR_EQUAL 1501) (T_XOR_EQUAL 1502)) 661 ((T_PAAMAYIM_NEKUDOTAYIM 1387)) 662 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1288) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 663 (("%" 1289) ("*" 1290) ("+" 1291) ("-" 1292) ("." 1293) ("/" 1294) ("<" 1295) (">" 1296) ("?" 1297) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1298) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1299) (T_BOOLEAN_AND 1300) (T_BOOLEAN_OR 1301) (T_COALESCE 1302) (T_INSTANCEOF 1303) (T_IS_EQUAL 1304) (T_IS_GREATER_OR_EQUAL 1305) (T_IS_IDENTICAL 1306) (T_IS_NOT_EQUAL 1307) (T_IS_NOT_IDENTICAL 1308) (T_IS_SMALLER_OR_EQUAL 1309) (T_LOGICAL_AND 1310) (T_LOGICAL_OR 1311) (T_LOGICAL_XOR 1312) (T_POW 1313) (T_SL 1314) (T_SPACESHIP 1315) (T_SR 1316) ("^" 1317) ("|" 1318)) 664 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1386) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 665 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1385) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 666 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1384) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 667 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1383) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 668 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1382) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 669 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1381) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 670 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1380) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 671 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1379) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 672 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 1374) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 1375) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 673 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1373) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 674 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1372) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 675 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1371) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 676 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1370) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 677 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1369) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 678 (("$" 1334) ("(" 1335) (T_NAME_FULLY_QUALIFIED 1336) (T_NAME_QUALIFIED 1337) (T_NAME_RELATIVE 1338) (T_STATIC 1339) (T_STRING 1340) (T_VARIABLE 1341) (class_name 1342) (class_name_reference 1343) (name 1344) (new_variable 1345) (simple_variable 1346)) 679 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1333) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 680 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1332) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 681 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1331) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 682 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1330) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 683 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1329) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 684 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1328) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 685 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1327) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 686 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1326) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 687 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1325) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 688 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1324) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 689 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1323) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 690 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1322) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 691 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1321) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 692 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1320) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 693 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1319) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 694 (("$" 1334) (T_VARIABLE 1341) (simple_variable 1365) ("{" 1366)) 695 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 1363) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 696 ((T_PAAMAYIM_NEKUDOTAYIM 1361)) 697 ((T_NULLSAFE_OBJECT_OPERATOR 1347) (T_OBJECT_OPERATOR 1348) (T_PAAMAYIM_NEKUDOTAYIM 1349) ("[" 1350)) 698 (("$" 1334) (T_STRING 1354) (T_VARIABLE 1341) (property_name 1360) (simple_variable 1356) ("{" 1357)) 699 (("$" 1334) (T_STRING 1354) (T_VARIABLE 1341) (property_name 1355) (simple_variable 1356) ("{" 1357)) 700 (("$" 1334) (T_VARIABLE 1341) (simple_variable 1353)) 701 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 1351) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 702 (("]" 1352)) 703 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 1358) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 704 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 1359)) 705 (("$" 1334) (T_VARIABLE 1341) (simple_variable 1362)) 706 (("%" 299) (")" 1364) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 707 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 1367) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 708 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 1368)) 709 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1378) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 710 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 1376) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 711 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1377) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 712 (("$" 1211) (T_ABSTRACT 1388) (T_ARRAY 1389) (T_AS 1390) (T_BREAK 1391) (T_CALLABLE 1392) (T_CASE 1393) (T_CATCH 1394) (T_CLASS 1395) (T_CLASS_C 1396) (T_CLONE 1397) (T_CONST 1398) (T_CONTINUE 1399) (T_DECLARE 1400) (T_DEFAULT 1401) (T_DIR 1402) (T_DO 1403) (T_ECHO 1404) (T_ELSE 1405) (T_ELSEIF 1406) (T_EMPTY 1407) (T_ENDDECLARE 1408) (T_ENDFOR 1409) (T_ENDFOREACH 1410) (T_ENDIF 1411) (T_ENDSWITCH 1412) (T_ENDWHILE 1413) (T_ENUM 1414) (T_EVAL 1415) (T_EXIT 1416) (T_EXTENDS 1417) (T_FILE 1418) (T_FINAL 1419) (T_FINALLY 1420) (T_FN 1421) (T_FOR 1422) (T_FOREACH 1423) (T_FUNCTION 1424) (T_FUNC_C 1425) (T_GLOBAL 1426) (T_GOTO 1427) (T_IF 1428) (T_IMPLEMENTS 1429) (T_INCLUDE 1430) (T_INCLUDE_ONCE 1431) (T_INSTANCEOF 1432) (T_INSTEADOF 1433) (T_INTERFACE 1434) (T_ISSET 1435) (T_LINE 1436) (T_LIST 1437) (T_LOGICAL_AND 1438) (T_LOGICAL_OR 1439) (T_LOGICAL_XOR 1440) (T_MATCH 1441) (T_METHOD_C 1442) (T_NAMESPACE 1443) (T_NEW 1444) (T_NS_C 1445) (T_PRINT 1446) (T_PRIVATE 1447) (T_PROPERTY_C 1448) (T_PROTECTED 1449) (T_PUBLIC 1450) (T_READONLY 1451) (T_REQUIRE 1452) (T_REQUIRE_ONCE 1453) (T_RETURN 1454) (T_STATIC 1455) (T_STRING 1456) (T_SWITCH 1457) (T_THROW 1458) (T_TRAIT 1459) (T_TRAIT_C 1460) (T_TRY 1461) (T_UNSET 1462) (T_USE 1463) (T_VAR 1464) (T_VARIABLE 1259) (T_WHILE 1465) (T_YIELD 1466) (identifier 1467) (member_name 1468) (reserved_non_modifiers 1469) (semi_reserved 1470) (simple_variable 1471) ("{" 1472)) 713 (("(" 1475) (argument_list 1476)) 714 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 1473) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 715 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 1474)) 716 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 1477) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 1478) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 1479) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 717 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 1486) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 718 (("," 1480) (possible_comma 1481)) 719 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 1483) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1484) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 720 ((")" 1482)) 721 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 722 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (ampersand 1518) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1519) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 723 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1515) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 724 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1514) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 725 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1513) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 726 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1512) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 727 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1511) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 728 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1510) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 729 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1509) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 730 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1508) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 731 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1507) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 732 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1506) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 733 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1505) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 734 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1504) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 735 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 1503) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 736 (("\"" 1520) ("$" 1521) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 1537) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 1541) ("[" 1542) (array_object_dereferenceable 1543) (callable_expr 1544) (callable_variable 1545) (class_constant 1546) (class_name 1547) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 1550) (name 1551) (new_dereferenceable 1552) (simple_variable 1553) (static_member 1554) (variable 1555) (variable_class_name 1556)) 737 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 9848) (encaps_var 5237)) 738 (("$" 1521) (T_VARIABLE 1541) (simple_variable 9844) ("{" 9845)) 739 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9842) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 740 (("(" 9839)) 741 (("$" 9784) ("(" 9785) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 9786) (T_NAME_QUALIFIED 9787) (T_NAME_RELATIVE 9788) (T_READONLY 2931) (T_STATIC 9789) (T_STRING 9790) (T_VARIABLE 9791) (anonymous_class 9792) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 9793) (attribute 2937) (attributes 9794) (class_modifier 2939) (class_name 9795) (class_name_reference 9796) (name 9797) (new_variable 9798) (simple_variable 9799)) 742 (("(" 1645) (argument_list 9783)) 743 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 1736) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 744 ((T_NULLSAFE_OBJECT_OPERATOR 1663) (T_OBJECT_OPERATOR 1664) ("[" 1665)) 745 (("(" 1645) (argument_list 1662)) 746 ((T_PAAMAYIM_NEKUDOTAYIM 1654)) 747 (("(" 1645) (argument_list 1653)) 748 ((T_PAAMAYIM_NEKUDOTAYIM 1557)) 749 (("$" 1521) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 1541) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 1638) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 1641) ("{" 1642)) 750 (("(" 1645) (argument_list 1646)) 751 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 1643) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 752 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 1644)) 753 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 1647) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 1648) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 1649) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 754 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 1652) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 755 (("," 1480) (possible_comma 1650)) 756 ((")" 1651)) 757 (("$" 1521) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 1541) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 1656) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 1657) ("{" 1658)) 758 (("(" 1645) (argument_list 1661)) 759 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 1659) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 760 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 1660)) 761 (("$" 1521) (T_STRING 1668) (T_VARIABLE 1541) (property_name 1675) (simple_variable 1670) ("{" 1671)) 762 (("$" 1521) (T_STRING 1668) (T_VARIABLE 1541) (property_name 1669) (simple_variable 1670) ("{" 1671)) 763 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 1666) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 764 (("]" 1667)) 765 (("(" 1645) (argument_list 1674)) 766 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 1672) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 767 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 1673)) 768 (("(" 1645) (argument_list 1676)) 769 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9782) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 770 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 9780) (encaps_var 5237)) 771 (("$" 1679) (T_VARIABLE 1728) (simple_variable 9776) ("{" 9777)) 772 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9774) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 773 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9773) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 774 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9772) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 775 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9771) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 776 (("(" 9768)) 777 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9767) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 778 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9766) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 779 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9765) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 780 (("\"" 1520) ("$" 7956) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 7957) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 7958) ("[" 1542) (array_object_dereferenceable 7959) (callable_expr 7960) (callable_variable 7961) (class_constant 1546) (class_name 7962) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 7963) (name 7964) (new_dereferenceable 1552) (simple_variable 7965) (static_member 7966) (variable 9764) (variable_class_name 7968)) 781 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9763) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 782 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9762) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 783 (("(" 9759)) 784 (("(" 9756)) 785 (("(" 9747) (argument_list 9748) (ctor_arguments 9749)) 786 (("\"" 1520) ("$" 7956) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 7957) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 7958) ("[" 1542) (array_object_dereferenceable 7959) (callable_expr 7960) (callable_variable 7961) (class_constant 1546) (class_name 7962) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 7963) (name 7964) (new_dereferenceable 1552) (simple_variable 7965) (static_member 7966) (variable 9746) (variable_class_name 7968)) 787 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9745) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 788 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 789 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9743) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 790 (("(" 9739)) 791 (("(" 9736)) 792 (("(" 9730)) 793 (("$" 7956) ("(" 9693) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 9694) (T_NAME_QUALIFIED 9695) (T_NAME_RELATIVE 9696) (T_READONLY 2931) (T_STATIC 9697) (T_STRING 9698) (T_VARIABLE 7958) (anonymous_class 9699) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 9700) (attribute 2937) (attributes 9701) (class_modifier 2939) (class_name 9702) (class_name_reference 9703) (name 9704) (new_variable 9705) (simple_variable 9706)) 794 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9692) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 795 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9691) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 796 (("(" 7917) (argument_list 9690)) 797 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9689) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 798 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9688) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 799 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 9683) (T_END_HEREDOC 9684) (T_VARIABLE 2896) (encaps_list 9685) (encaps_var 2898)) 800 ((T_FN 35) (T_FUNCTION 129) (fn 1745) #'1746 (inline_function 9682)) 801 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9681) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 802 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9680) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 803 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9679) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 804 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9676) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 805 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9675) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 806 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 9671) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 807 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 9669) (encaps_list 2804) (encaps_var 2805)) 808 (("\"" 1520) ("$" 9479) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 9480) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 9481) ("[" 1542) (array_object_dereferenceable 9482) (callable_expr 9483) (callable_variable 9484) (class_constant 1546) (class_name 9485) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 9486) (name 9487) (new_dereferenceable 1552) (simple_variable 9488) (static_member 9489) (variable 9668) (variable_class_name 9491)) 809 ((T_NULLSAFE_OBJECT_OPERATOR 9654) (T_OBJECT_OPERATOR 9655) ("[" 9656)) 810 (("]" 9653)) 811 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 9650) (attribute 2733) (fn 1745) #'1746 (inline_function 9651)) 812 (("(" 7917) (argument_list 9649)) 813 ((T_PAAMAYIM_NEKUDOTAYIM 9641)) 814 (("%" 1764) ("*" 1765) ("+" 1766) ("-" 1767) ("." 1768) ("/" 1769) ("<" 1770) (">" 1771) ("?" 1772) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1773) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1774) (T_BOOLEAN_AND 1775) (T_BOOLEAN_OR 1776) (T_COALESCE 1777) (T_DOUBLE_ARROW 8549) (T_INSTANCEOF 1778) (T_IS_EQUAL 1779) (T_IS_GREATER_OR_EQUAL 1780) (T_IS_IDENTICAL 1781) (T_IS_NOT_EQUAL 1782) (T_IS_NOT_IDENTICAL 1783) (T_IS_SMALLER_OR_EQUAL 1784) (T_LOGICAL_AND 1785) (T_LOGICAL_OR 1786) (T_LOGICAL_XOR 1787) (T_POW 1788) (T_SL 1789) (T_SPACESHIP 1790) (T_SR 1791) ("^" 1792) ("|" 1793)) 815 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 8020)) 816 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 8008)) 817 (("(" 7917) (argument_list 8007)) 818 (("," 8005)) 819 (("=" 7925) (T_AND_EQUAL 7926) (T_COALESCE_EQUAL 7927) (T_CONCAT_EQUAL 7928) (T_DEC 7929) (T_DIV_EQUAL 7930) (T_INC 7931) (T_MINUS_EQUAL 7932) (T_MOD_EQUAL 7933) (T_MUL_EQUAL 7934) (T_OR_EQUAL 7935) (T_PLUS_EQUAL 7936) (T_POW_EQUAL 7937) (T_SL_EQUAL 7938) (T_SR_EQUAL 7939) (T_XOR_EQUAL 7940)) 820 ((T_PAAMAYIM_NEKUDOTAYIM 7829)) 821 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1763) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 822 (("(" 1862)) 823 (("%" 1764) ("*" 1765) ("+" 1766) ("-" 1767) ("." 1768) ("/" 1769) ("<" 1770) (">" 1771) ("?" 1772) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1773) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1774) (T_BOOLEAN_AND 1775) (T_BOOLEAN_OR 1776) (T_COALESCE 1777) (T_INSTANCEOF 1778) (T_IS_EQUAL 1779) (T_IS_GREATER_OR_EQUAL 1780) (T_IS_IDENTICAL 1781) (T_IS_NOT_EQUAL 1782) (T_IS_NOT_IDENTICAL 1783) (T_IS_SMALLER_OR_EQUAL 1784) (T_LOGICAL_AND 1785) (T_LOGICAL_OR 1786) (T_LOGICAL_XOR 1787) (T_POW 1788) (T_SL 1789) (T_SPACESHIP 1790) (T_SR 1791) ("^" 1792) ("|" 1793)) 824 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1861) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 825 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1860) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 826 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1859) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 827 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1858) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 828 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1857) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 829 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1856) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 830 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1855) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 831 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1854) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 832 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 1849) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 1850) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 833 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1848) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 834 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1847) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 835 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1846) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 836 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1845) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 837 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1844) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 838 (("$" 1809) ("(" 1810) (T_NAME_FULLY_QUALIFIED 1811) (T_NAME_QUALIFIED 1812) (T_NAME_RELATIVE 1813) (T_STATIC 1814) (T_STRING 1815) (T_VARIABLE 1816) (class_name 1817) (class_name_reference 1818) (name 1819) (new_variable 1820) (simple_variable 1821)) 839 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1808) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 840 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1807) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 841 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1806) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 842 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1805) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 843 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1804) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 844 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1803) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 845 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1802) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 846 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1801) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 847 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1800) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 848 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1799) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 849 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1798) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 850 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1797) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 851 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1796) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 852 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1795) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 853 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1794) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 854 (("$" 1809) (T_VARIABLE 1816) (simple_variable 1840) ("{" 1841)) 855 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 1838) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 856 ((T_PAAMAYIM_NEKUDOTAYIM 1836)) 857 ((T_NULLSAFE_OBJECT_OPERATOR 1822) (T_OBJECT_OPERATOR 1823) (T_PAAMAYIM_NEKUDOTAYIM 1824) ("[" 1825)) 858 (("$" 1809) (T_STRING 1829) (T_VARIABLE 1816) (property_name 1835) (simple_variable 1831) ("{" 1832)) 859 (("$" 1809) (T_STRING 1829) (T_VARIABLE 1816) (property_name 1830) (simple_variable 1831) ("{" 1832)) 860 (("$" 1809) (T_VARIABLE 1816) (simple_variable 1828)) 861 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 1826) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 862 (("]" 1827)) 863 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 1833) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 864 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 1834)) 865 (("$" 1809) (T_VARIABLE 1816) (simple_variable 1837)) 866 (("%" 299) (")" 1839) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 867 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 1842) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 868 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 1843)) 869 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1853) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 870 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 1851) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 871 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1852) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 872 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 1922) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 873 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7828) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 874 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 7826) (encaps_var 5237)) 875 (("$" 1865) (T_VARIABLE 1914) (simple_variable 7822) ("{" 7823)) 876 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 7820) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 877 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7819) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 878 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7818) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 879 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7817) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 880 (("(" 7814)) 881 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7813) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 882 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7812) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 883 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7811) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 884 (("\"" 1520) ("$" 2180) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 2181) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 2182) ("[" 1542) (array_object_dereferenceable 2183) (callable_expr 2184) (callable_variable 2185) (class_constant 1546) (class_name 2186) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 2187) (name 2188) (new_dereferenceable 1552) (simple_variable 2189) (static_member 2190) (variable 7810) (variable_class_name 2192)) 885 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7809) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 886 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 7808) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 887 (("(" 7805)) 888 (("(" 7802)) 889 (("(" 7793) (argument_list 7794) (ctor_arguments 7795)) 890 (("\"" 1520) ("$" 2180) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 2181) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 2182) ("[" 1542) (array_object_dereferenceable 2183) (callable_expr 2184) (callable_variable 2185) (class_constant 1546) (class_name 2186) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 2187) (name 2188) (new_dereferenceable 1552) (simple_variable 2189) (static_member 2190) (variable 7792) (variable_class_name 2192)) 891 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7791) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 892 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7790) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 893 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7789) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 894 (("(" 7785)) 895 (("(" 7782)) 896 (("(" 7776)) 897 (("$" 2180) ("(" 7739) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 7740) (T_NAME_QUALIFIED 7741) (T_NAME_RELATIVE 7742) (T_READONLY 2931) (T_STATIC 7743) (T_STRING 7744) (T_VARIABLE 2182) (anonymous_class 7745) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 7746) (attribute 2937) (attributes 7747) (class_modifier 2939) (class_name 7748) (class_name_reference 7749) (name 7750) (new_variable 7751) (simple_variable 7752)) 898 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7738) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 899 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7737) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 900 (("(" 2141) (argument_list 7736)) 901 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7735) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 902 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7734) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 903 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 7729) (T_END_HEREDOC 7730) (T_VARIABLE 2896) (encaps_list 7731) (encaps_var 2898)) 904 ((T_FN 35) (T_FUNCTION 129) (fn 1931) #'1932 (inline_function 7728)) 905 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7727) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 906 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7726) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 907 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7725) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 908 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7722) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 909 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7721) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 910 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 7717) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 911 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 7715) (encaps_list 2804) (encaps_var 2805)) 912 (("\"" 1520) ("$" 2750) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 2751) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 2752) ("[" 1542) (array_object_dereferenceable 2753) (callable_expr 2754) (callable_variable 2755) (class_constant 1546) (class_name 2756) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 2757) (name 2758) (new_dereferenceable 1552) (simple_variable 2759) (static_member 2760) (variable 7714) (variable_class_name 2762)) 913 ((T_NULLSAFE_OBJECT_OPERATOR 7700) (T_OBJECT_OPERATOR 7701) ("[" 7702)) 914 ((")" 7697)) 915 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 7694) (attribute 2733) (fn 1931) #'1932 (inline_function 7695)) 916 (("(" 2141) (argument_list 7693)) 917 ((T_PAAMAYIM_NEKUDOTAYIM 7685)) 918 (("%" 1950) ("*" 1951) ("+" 1952) ("-" 1953) ("." 1954) ("/" 1955) ("<" 1956) (">" 1957) ("?" 1958) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1959) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1960) (T_BOOLEAN_AND 1961) (T_BOOLEAN_OR 1962) (T_COALESCE 1963) (T_DOUBLE_ARROW 2337) (T_INSTANCEOF 1964) (T_IS_EQUAL 1965) (T_IS_GREATER_OR_EQUAL 1966) (T_IS_IDENTICAL 1967) (T_IS_NOT_EQUAL 1968) (T_IS_NOT_IDENTICAL 1969) (T_IS_SMALLER_OR_EQUAL 1970) (T_LOGICAL_AND 1971) (T_LOGICAL_OR 1972) (T_LOGICAL_XOR 1973) (T_POW 1974) (T_SL 1975) (T_SPACESHIP 1976) (T_SR 1977) ("^" 1978) ("|" 1979)) 919 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 2244)) 920 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 2232)) 921 (("(" 2141) (argument_list 2231)) 922 (("," 2229)) 923 (("=" 2149) (T_AND_EQUAL 2150) (T_COALESCE_EQUAL 2151) (T_CONCAT_EQUAL 2152) (T_DEC 2153) (T_DIV_EQUAL 2154) (T_INC 2155) (T_MINUS_EQUAL 2156) (T_MOD_EQUAL 2157) (T_MUL_EQUAL 2158) (T_OR_EQUAL 2159) (T_PLUS_EQUAL 2160) (T_POW_EQUAL 2161) (T_SL_EQUAL 2162) (T_SR_EQUAL 2163) (T_XOR_EQUAL 2164)) 924 ((T_PAAMAYIM_NEKUDOTAYIM 2053)) 925 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1949) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 926 (("(" 2048)) 927 (("%" 1950) ("*" 1951) ("+" 1952) ("-" 1953) ("." 1954) ("/" 1955) ("<" 1956) (">" 1957) ("?" 1958) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1959) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1960) (T_BOOLEAN_AND 1961) (T_BOOLEAN_OR 1962) (T_COALESCE 1963) (T_INSTANCEOF 1964) (T_IS_EQUAL 1965) (T_IS_GREATER_OR_EQUAL 1966) (T_IS_IDENTICAL 1967) (T_IS_NOT_EQUAL 1968) (T_IS_NOT_IDENTICAL 1969) (T_IS_SMALLER_OR_EQUAL 1970) (T_LOGICAL_AND 1971) (T_LOGICAL_OR 1972) (T_LOGICAL_XOR 1973) (T_POW 1974) (T_SL 1975) (T_SPACESHIP 1976) (T_SR 1977) ("^" 1978) ("|" 1979)) 928 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2047) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 929 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2046) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 930 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2045) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 931 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2044) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 932 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2043) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 933 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2042) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 934 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2041) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 935 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2040) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 936 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 2035) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 2036) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 937 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2034) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 938 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2033) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 939 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2032) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 940 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2031) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 941 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2030) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 942 (("$" 1995) ("(" 1996) (T_NAME_FULLY_QUALIFIED 1997) (T_NAME_QUALIFIED 1998) (T_NAME_RELATIVE 1999) (T_STATIC 2000) (T_STRING 2001) (T_VARIABLE 2002) (class_name 2003) (class_name_reference 2004) (name 2005) (new_variable 2006) (simple_variable 2007)) 943 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1994) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 944 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1993) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 945 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1992) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 946 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1991) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 947 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1990) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 948 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1989) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 949 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1988) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 950 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1987) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 951 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1986) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 952 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1985) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 953 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1984) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 954 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1983) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 955 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1982) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 956 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1981) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 957 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1980) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 958 (("$" 1995) (T_VARIABLE 2002) (simple_variable 2026) ("{" 2027)) 959 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 2024) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 960 ((T_PAAMAYIM_NEKUDOTAYIM 2022)) 961 ((T_NULLSAFE_OBJECT_OPERATOR 2008) (T_OBJECT_OPERATOR 2009) (T_PAAMAYIM_NEKUDOTAYIM 2010) ("[" 2011)) 962 (("$" 1995) (T_STRING 2015) (T_VARIABLE 2002) (property_name 2021) (simple_variable 2017) ("{" 2018)) 963 (("$" 1995) (T_STRING 2015) (T_VARIABLE 2002) (property_name 2016) (simple_variable 2017) ("{" 2018)) 964 (("$" 1995) (T_VARIABLE 2002) (simple_variable 2014)) 965 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 2012) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 966 (("]" 2013)) 967 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2019) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 968 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2020)) 969 (("$" 1995) (T_VARIABLE 2002) (simple_variable 2023)) 970 (("%" 299) (")" 2025) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 971 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2028) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 972 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2029)) 973 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2039) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 974 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 2037) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 975 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2038) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 976 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 2049) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 977 ((")" 2050)) 978 (("=" 2051)) 979 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2052) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 980 (("$" 1865) (T_ABSTRACT 2054) (T_ARRAY 2055) (T_AS 2056) (T_BREAK 2057) (T_CALLABLE 2058) (T_CASE 2059) (T_CATCH 2060) (T_CLASS 2061) (T_CLASS_C 2062) (T_CLONE 2063) (T_CONST 2064) (T_CONTINUE 2065) (T_DECLARE 2066) (T_DEFAULT 2067) (T_DIR 2068) (T_DO 2069) (T_ECHO 2070) (T_ELSE 2071) (T_ELSEIF 2072) (T_EMPTY 2073) (T_ENDDECLARE 2074) (T_ENDFOR 2075) (T_ENDFOREACH 2076) (T_ENDIF 2077) (T_ENDSWITCH 2078) (T_ENDWHILE 2079) (T_ENUM 2080) (T_EVAL 2081) (T_EXIT 2082) (T_EXTENDS 2083) (T_FILE 2084) (T_FINAL 2085) (T_FINALLY 2086) (T_FN 2087) (T_FOR 2088) (T_FOREACH 2089) (T_FUNCTION 2090) (T_FUNC_C 2091) (T_GLOBAL 2092) (T_GOTO 2093) (T_IF 2094) (T_IMPLEMENTS 2095) (T_INCLUDE 2096) (T_INCLUDE_ONCE 2097) (T_INSTANCEOF 2098) (T_INSTEADOF 2099) (T_INTERFACE 2100) (T_ISSET 2101) (T_LINE 2102) (T_LIST 2103) (T_LOGICAL_AND 2104) (T_LOGICAL_OR 2105) (T_LOGICAL_XOR 2106) (T_MATCH 2107) (T_METHOD_C 2108) (T_NAMESPACE 2109) (T_NEW 2110) (T_NS_C 2111) (T_PRINT 2112) (T_PRIVATE 2113) (T_PROPERTY_C 2114) (T_PROTECTED 2115) (T_PUBLIC 2116) (T_READONLY 2117) (T_REQUIRE 2118) (T_REQUIRE_ONCE 2119) (T_RETURN 2120) (T_STATIC 2121) (T_STRING 2122) (T_SWITCH 2123) (T_THROW 2124) (T_TRAIT 2125) (T_TRAIT_C 2126) (T_TRY 2127) (T_UNSET 2128) (T_USE 2129) (T_VAR 2130) (T_VARIABLE 1914) (T_WHILE 2131) (T_YIELD 2132) (identifier 2133) (member_name 2134) (reserved_non_modifiers 2135) (semi_reserved 2136) (simple_variable 2137) ("{" 2138)) 981 (("(" 2141) (argument_list 2142)) 982 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2139) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 983 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2140)) 984 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 2143) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 2144) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 2145) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 985 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 2148) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 986 (("," 1480) (possible_comma 2146)) 987 ((")" 2147)) 988 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 2178) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2179) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 989 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2177) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 990 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2176) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 991 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2175) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 992 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2174) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 993 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2173) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 994 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2172) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 995 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2171) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 996 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2170) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 997 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2169) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 998 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2168) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 999 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2167) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 1000 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2166) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 1001 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2165) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 1002 (("\"" 1520) ("$" 2180) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 2181) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 2182) ("[" 1542) (array_object_dereferenceable 2183) (callable_expr 2184) (callable_variable 2185) (class_constant 1546) (class_name 2186) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 2187) (name 2188) (new_dereferenceable 1552) (simple_variable 2189) (static_member 2190) (variable 2191) (variable_class_name 2192)) 1003 (("$" 2180) (T_VARIABLE 2182) (simple_variable 2225) ("{" 2226)) 1004 (("(" 2196) (argument_list 2224)) 1005 ((T_NULLSAFE_OBJECT_OPERATOR 2210) (T_OBJECT_OPERATOR 2211) ("[" 2212)) 1006 (("(" 2196) (argument_list 2209)) 1007 ((T_PAAMAYIM_NEKUDOTAYIM 2205)) 1008 (("(" 2196) (argument_list 2204)) 1009 ((T_PAAMAYIM_NEKUDOTAYIM 2193)) 1010 (("$" 2180) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 2182) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 2194) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 2195) ("{" 1642)) 1011 (("(" 2196) (argument_list 2197)) 1012 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 2198) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 2199) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 2200) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1013 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 2203) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1014 (("," 1480) (possible_comma 2201)) 1015 ((")" 2202)) 1016 (("$" 2180) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 2182) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 2206) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 2207) ("{" 1658)) 1017 (("(" 2196) (argument_list 2208)) 1018 (("$" 2180) (T_STRING 2215) (T_VARIABLE 2182) (property_name 2222) (simple_variable 2217) ("{" 2218)) 1019 (("$" 2180) (T_STRING 2215) (T_VARIABLE 2182) (property_name 2216) (simple_variable 2217) ("{" 2218)) 1020 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 2213) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1021 (("]" 2214)) 1022 (("(" 2196) (argument_list 2221)) 1023 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2219) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1024 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2220)) 1025 (("(" 2196) (argument_list 2223)) 1026 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2227) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1027 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2228)) 1028 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (possible_array_pair 2230) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 1029 ((backup_doc_comment 2233)) 1030 (("(" 2234)) 1031 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 2235)) 1032 ((")" 2236)) 1033 ((T_USE 164) (lexical_vars 2237)) 1034 ((":" 166) (return_type 2238)) 1035 ((backup_fn_flags 2239)) 1036 (("{" 2240)) 1037 ((inner_statement_list 2241)) 1038 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 2242) ("~" 127)) 1039 ((backup_fn_flags 2243)) 1040 ((backup_doc_comment 2245)) 1041 (("(" 2246)) 1042 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 2247)) 1043 ((")" 2248)) 1044 ((":" 2249) (return_type 2250)) 1045 (("(" 2256) ("?" 2257) (T_ARRAY 2258) (T_CALLABLE 2259) (T_NAME_FULLY_QUALIFIED 2260) (T_NAME_QUALIFIED 2261) (T_NAME_RELATIVE 2262) (T_STATIC 2263) (T_STRING 2264) (intersection_type 2265) (name 2266) (type 2267) (type_expr 2268) (type_without_static 2269) (union_type 2270) (union_type_element 2271)) 1046 ((T_DOUBLE_ARROW 2251)) 1047 ((backup_fn_flags 2252)) 1048 ((backup_lex_pos 2253)) 1049 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 2254) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 1050 (("%" 1289) ("*" 1290) ("+" 1291) ("-" 1292) ("." 1293) ("/" 1294) ("<" 1295) (">" 1296) ("?" 1297) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1298) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1299) (T_BOOLEAN_AND 1300) (T_BOOLEAN_OR 1301) (T_COALESCE 1302) (T_INSTANCEOF 1303) (T_IS_EQUAL 1304) (T_IS_GREATER_OR_EQUAL 1305) (T_IS_IDENTICAL 1306) (T_IS_NOT_EQUAL 1307) (T_IS_NOT_IDENTICAL 1308) (T_IS_SMALLER_OR_EQUAL 1309) (T_LOGICAL_AND 1310) (T_LOGICAL_OR 1311) (T_LOGICAL_XOR 1312) (T_POW 1313) (T_SL 1314) (T_SPACESHIP 1315) (T_SR 1316) ("^" 1317) (backup_fn_flags 2255) ("|" 1318)) 1051 ((T_ARRAY 2285) (T_CALLABLE 2286) (T_NAME_FULLY_QUALIFIED 2287) (T_NAME_QUALIFIED 2288) (T_NAME_RELATIVE 2289) (T_STATIC 2290) (T_STRING 2291) (intersection_type 2335) (name 2293) (type 2294) (type_without_static 2295)) 1052 ((T_ARRAY 2325) (T_CALLABLE 2326) (T_NAME_FULLY_QUALIFIED 2327) (T_NAME_QUALIFIED 2328) (T_NAME_RELATIVE 2329) (T_STATIC 2330) (T_STRING 2331) (name 2332) (type 2333) (type_without_static 2334)) 1053 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2323)) 1054 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2312)) 1055 (("|" 2310)) 1056 (("|" 2272)) 1057 (("(" 2273) (T_ARRAY 2274) (T_CALLABLE 2275) (T_NAME_FULLY_QUALIFIED 2276) (T_NAME_QUALIFIED 2277) (T_NAME_RELATIVE 2278) (T_STATIC 2279) (T_STRING 2280) (name 2281) (type 2282) (type_without_static 2283) (union_type_element 2284)) 1058 ((T_ARRAY 2285) (T_CALLABLE 2286) (T_NAME_FULLY_QUALIFIED 2287) (T_NAME_QUALIFIED 2288) (T_NAME_RELATIVE 2289) (T_STATIC 2290) (T_STRING 2291) (intersection_type 2292) (name 2293) (type 2294) (type_without_static 2295)) 1059 ((")" 2307) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2308)) 1060 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2296)) 1061 ((T_ARRAY 2297) (T_CALLABLE 2298) (T_NAME_FULLY_QUALIFIED 2299) (T_NAME_QUALIFIED 2300) (T_NAME_RELATIVE 2301) (T_STATIC 2302) (T_STRING 2303) (name 2304) (type 2305) (type_without_static 2306)) 1062 ((T_ARRAY 2297) (T_CALLABLE 2298) (T_NAME_FULLY_QUALIFIED 2299) (T_NAME_QUALIFIED 2300) (T_NAME_RELATIVE 2301) (T_STATIC 2302) (T_STRING 2303) (name 2304) (type 2309) (type_without_static 2306)) 1063 (("(" 2273) (T_ARRAY 2274) (T_CALLABLE 2275) (T_NAME_FULLY_QUALIFIED 2276) (T_NAME_QUALIFIED 2277) (T_NAME_RELATIVE 2278) (T_STATIC 2279) (T_STRING 2280) (name 2281) (type 2282) (type_without_static 2283) (union_type_element 2311)) 1064 ((T_ARRAY 2313) (T_CALLABLE 2314) (T_NAME_FULLY_QUALIFIED 2315) (T_NAME_QUALIFIED 2316) (T_NAME_RELATIVE 2317) (T_STATIC 2318) (T_STRING 2319) (name 2320) (type 2321) (type_without_static 2322)) 1065 ((T_ARRAY 2313) (T_CALLABLE 2314) (T_NAME_FULLY_QUALIFIED 2315) (T_NAME_QUALIFIED 2316) (T_NAME_RELATIVE 2317) (T_STATIC 2318) (T_STRING 2319) (name 2320) (type 2324) (type_without_static 2322)) 1066 ((")" 2336) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2308)) 1067 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2366) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (ampersand 2393) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2402) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1068 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 7684) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1069 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 7682) (encaps_var 5237)) 1070 (("$" 2340) (T_VARIABLE 2388) (simple_variable 7678) ("{" 7679)) 1071 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 7676) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 1072 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 7675) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1073 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 7674) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1074 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 7673) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1075 (("(" 7670)) 1076 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 7669) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1077 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 7668) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1078 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 7667) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1079 (("\"" 1520) ("$" 2650) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 2651) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 2652) ("[" 1542) (array_object_dereferenceable 2653) (callable_expr 2654) (callable_variable 2655) (class_constant 1546) (class_name 2656) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 2657) (name 2658) (new_dereferenceable 1552) (simple_variable 2659) (static_member 2660) (variable 7666) (variable_class_name 2662)) 1080 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 7665) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1081 (("(" 7662)) 1082 (("(" 7659)) 1083 (("(" 7650) (argument_list 7651) (ctor_arguments 7652)) 1084 (("\"" 1520) ("$" 2650) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 2651) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 2652) ("[" 1542) (array_object_dereferenceable 2653) (callable_expr 2654) (callable_variable 2655) (class_constant 1546) (class_name 2656) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 2657) (name 2658) (new_dereferenceable 1552) (simple_variable 2659) (static_member 2660) (variable 7649) (variable_class_name 2662)) 1085 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 7648) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1086 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 7647) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1087 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 7646) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1088 (("(" 7642)) 1089 (("(" 7639)) 1090 (("(" 7633)) 1091 (("$" 2650) ("(" 2926) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 2928) (T_NAME_QUALIFIED 2929) (T_NAME_RELATIVE 2930) (T_READONLY 2931) (T_STATIC 2932) (T_STRING 2933) (T_VARIABLE 2652) (anonymous_class 2934) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 2936) (attribute 2937) (attributes 2938) (class_modifier 2939) (class_name 2940) (class_name_reference 2941) (name 2942) (new_variable 2943) (simple_variable 2944)) 1092 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2925) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1093 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2924) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1094 (("(" 2611) (argument_list 2923)) 1095 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2922) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1096 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2921) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1097 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2894) (T_END_HEREDOC 2895) (T_VARIABLE 2896) (encaps_list 2897) (encaps_var 2898)) 1098 ((T_FN 35) (T_FUNCTION 129) (fn 2403) #'2404 (inline_function 2891)) 1099 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2890) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1100 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2889) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1101 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2888) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1102 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 2885) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 1103 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2884) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1104 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 2880) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 1105 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 2803) (encaps_list 2804) (encaps_var 2805)) 1106 (("\"" 1520) ("$" 2750) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 2751) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 2752) ("[" 1542) (array_object_dereferenceable 2753) (callable_expr 2754) (callable_variable 2755) (class_constant 1546) (class_name 2756) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 2757) (name 2758) (new_dereferenceable 1552) (simple_variable 2759) (static_member 2760) (variable 2761) (variable_class_name 2762)) 1107 ((T_NULLSAFE_OBJECT_OPERATOR 2736) (T_OBJECT_OPERATOR 2737) ("[" 2738)) 1108 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 2732) (attribute 2733) (fn 2403) #'2404 (inline_function 2734)) 1109 (("(" 2611) (argument_list 2731)) 1110 ((T_PAAMAYIM_NEKUDOTAYIM 2723)) 1111 (("%" 2420) ("*" 2421) ("+" 2422) ("-" 2423) ("." 2424) ("/" 2425) ("<" 2426) (">" 2427) ("?" 2428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2430) (T_BOOLEAN_AND 2431) (T_BOOLEAN_OR 2432) (T_COALESCE 2433) (T_INSTANCEOF 2434) (T_IS_EQUAL 2435) (T_IS_GREATER_OR_EQUAL 2436) (T_IS_IDENTICAL 2437) (T_IS_NOT_EQUAL 2438) (T_IS_NOT_IDENTICAL 2439) (T_IS_SMALLER_OR_EQUAL 2440) (T_LOGICAL_AND 2441) (T_LOGICAL_OR 2442) (T_LOGICAL_XOR 2443) (T_POW 2444) (T_SL 2445) (T_SPACESHIP 2446) (T_SR 2447) ("^" 2448) ("|" 2449)) 1112 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 2712)) 1113 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 2700)) 1114 (("(" 2611) (argument_list 2699)) 1115 (("=" 2619) (T_AND_EQUAL 2620) (T_COALESCE_EQUAL 2621) (T_CONCAT_EQUAL 2622) (T_DEC 2623) (T_DIV_EQUAL 2624) (T_INC 2625) (T_MINUS_EQUAL 2626) (T_MOD_EQUAL 2627) (T_MUL_EQUAL 2628) (T_OR_EQUAL 2629) (T_PLUS_EQUAL 2630) (T_POW_EQUAL 2631) (T_SL_EQUAL 2632) (T_SR_EQUAL 2633) (T_XOR_EQUAL 2634)) 1116 ((T_PAAMAYIM_NEKUDOTAYIM 2523)) 1117 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2419) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1118 (("(" 2518)) 1119 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2517) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1120 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2516) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1121 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2515) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1122 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2514) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1123 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2513) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1124 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2512) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1125 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2511) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1126 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2510) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1127 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 2505) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 2506) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 1128 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2504) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1129 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2503) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1130 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2502) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1131 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2501) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1132 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2500) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1133 (("$" 2465) ("(" 2466) (T_NAME_FULLY_QUALIFIED 2467) (T_NAME_QUALIFIED 2468) (T_NAME_RELATIVE 2469) (T_STATIC 2470) (T_STRING 2471) (T_VARIABLE 2472) (class_name 2473) (class_name_reference 2474) (name 2475) (new_variable 2476) (simple_variable 2477)) 1134 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2464) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1135 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2463) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1136 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2462) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1137 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2461) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1138 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2460) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1139 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2459) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1140 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2458) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1141 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2457) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1142 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2456) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1143 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2455) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1144 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2454) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1145 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2453) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1146 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2452) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1147 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2451) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1148 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2450) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1149 (("$" 2465) (T_VARIABLE 2472) (simple_variable 2496) ("{" 2497)) 1150 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 2494) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 1151 ((T_PAAMAYIM_NEKUDOTAYIM 2492)) 1152 ((T_NULLSAFE_OBJECT_OPERATOR 2478) (T_OBJECT_OPERATOR 2479) (T_PAAMAYIM_NEKUDOTAYIM 2480) ("[" 2481)) 1153 (("$" 2465) (T_STRING 2485) (T_VARIABLE 2472) (property_name 2491) (simple_variable 2487) ("{" 2488)) 1154 (("$" 2465) (T_STRING 2485) (T_VARIABLE 2472) (property_name 2486) (simple_variable 2487) ("{" 2488)) 1155 (("$" 2465) (T_VARIABLE 2472) (simple_variable 2484)) 1156 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 2482) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1157 (("]" 2483)) 1158 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2489) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1159 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2490)) 1160 (("$" 2465) (T_VARIABLE 2472) (simple_variable 2493)) 1161 (("%" 299) (")" 2495) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 1162 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2498) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1163 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2499)) 1164 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2509) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1165 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 2507) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 1166 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2508) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1167 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 2519) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 1168 ((")" 2520)) 1169 (("=" 2521)) 1170 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2522) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1171 (("$" 2340) (T_ABSTRACT 2524) (T_ARRAY 2525) (T_AS 2526) (T_BREAK 2527) (T_CALLABLE 2528) (T_CASE 2529) (T_CATCH 2530) (T_CLASS 2531) (T_CLASS_C 2532) (T_CLONE 2533) (T_CONST 2534) (T_CONTINUE 2535) (T_DECLARE 2536) (T_DEFAULT 2537) (T_DIR 2538) (T_DO 2539) (T_ECHO 2540) (T_ELSE 2541) (T_ELSEIF 2542) (T_EMPTY 2543) (T_ENDDECLARE 2544) (T_ENDFOR 2545) (T_ENDFOREACH 2546) (T_ENDIF 2547) (T_ENDSWITCH 2548) (T_ENDWHILE 2549) (T_ENUM 2550) (T_EVAL 2551) (T_EXIT 2552) (T_EXTENDS 2553) (T_FILE 2554) (T_FINAL 2555) (T_FINALLY 2556) (T_FN 2557) (T_FOR 2558) (T_FOREACH 2559) (T_FUNCTION 2560) (T_FUNC_C 2561) (T_GLOBAL 2562) (T_GOTO 2563) (T_IF 2564) (T_IMPLEMENTS 2565) (T_INCLUDE 2566) (T_INCLUDE_ONCE 2567) (T_INSTANCEOF 2568) (T_INSTEADOF 2569) (T_INTERFACE 2570) (T_ISSET 2571) (T_LINE 2572) (T_LIST 2573) (T_LOGICAL_AND 2574) (T_LOGICAL_OR 2575) (T_LOGICAL_XOR 2576) (T_MATCH 2577) (T_METHOD_C 2578) (T_NAMESPACE 2579) (T_NEW 2580) (T_NS_C 2581) (T_PRINT 2582) (T_PRIVATE 2583) (T_PROPERTY_C 2584) (T_PROTECTED 2585) (T_PUBLIC 2586) (T_READONLY 2587) (T_REQUIRE 2588) (T_REQUIRE_ONCE 2589) (T_RETURN 2590) (T_STATIC 2591) (T_STRING 2592) (T_SWITCH 2593) (T_THROW 2594) (T_TRAIT 2595) (T_TRAIT_C 2596) (T_TRY 2597) (T_UNSET 2598) (T_USE 2599) (T_VAR 2600) (T_VARIABLE 2388) (T_WHILE 2601) (T_YIELD 2602) (identifier 2603) (member_name 2604) (reserved_non_modifiers 2605) (semi_reserved 2606) (simple_variable 2607) ("{" 2608)) 1172 (("(" 2611) (argument_list 2612)) 1173 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2609) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1174 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2610)) 1175 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 2613) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 2614) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 2615) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1176 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 2618) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1177 (("," 1480) (possible_comma 2616)) 1178 ((")" 2617)) 1179 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (ampersand 2648) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2649) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1180 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2647) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1181 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2646) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1182 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2645) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1183 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2644) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1184 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2643) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1185 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2642) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1186 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2641) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1187 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2640) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1188 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2639) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1189 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2638) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1190 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2637) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1191 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2636) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1192 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2635) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1193 (("\"" 1520) ("$" 2650) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 2651) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 2652) ("[" 1542) (array_object_dereferenceable 2653) (callable_expr 2654) (callable_variable 2655) (class_constant 1546) (class_name 2656) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 2657) (name 2658) (new_dereferenceable 1552) (simple_variable 2659) (static_member 2660) (variable 2661) (variable_class_name 2662)) 1194 (("$" 2650) (T_VARIABLE 2652) (simple_variable 2695) ("{" 2696)) 1195 (("(" 2666) (argument_list 2694)) 1196 ((T_NULLSAFE_OBJECT_OPERATOR 2680) (T_OBJECT_OPERATOR 2681) ("[" 2682)) 1197 (("(" 2666) (argument_list 2679)) 1198 ((T_PAAMAYIM_NEKUDOTAYIM 2675)) 1199 (("(" 2666) (argument_list 2674)) 1200 ((T_PAAMAYIM_NEKUDOTAYIM 2663)) 1201 (("$" 2650) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 2652) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 2664) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 2665) ("{" 1642)) 1202 (("(" 2666) (argument_list 2667)) 1203 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 2668) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 2669) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 2670) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1204 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 2673) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1205 (("," 1480) (possible_comma 2671)) 1206 ((")" 2672)) 1207 (("$" 2650) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 2652) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 2676) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 2677) ("{" 1658)) 1208 (("(" 2666) (argument_list 2678)) 1209 (("$" 2650) (T_STRING 2685) (T_VARIABLE 2652) (property_name 2692) (simple_variable 2687) ("{" 2688)) 1210 (("$" 2650) (T_STRING 2685) (T_VARIABLE 2652) (property_name 2686) (simple_variable 2687) ("{" 2688)) 1211 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 2683) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1212 (("]" 2684)) 1213 (("(" 2666) (argument_list 2691)) 1214 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2689) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1215 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2690)) 1216 (("(" 2666) (argument_list 2693)) 1217 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2697) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1218 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2698)) 1219 ((backup_doc_comment 2701)) 1220 (("(" 2702)) 1221 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 2703)) 1222 ((")" 2704)) 1223 ((T_USE 164) (lexical_vars 2705)) 1224 ((":" 166) (return_type 2706)) 1225 ((backup_fn_flags 2707)) 1226 (("{" 2708)) 1227 ((inner_statement_list 2709)) 1228 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 2710) ("~" 127)) 1229 ((backup_fn_flags 2711)) 1230 ((backup_doc_comment 2713)) 1231 (("(" 2714)) 1232 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 2715)) 1233 ((")" 2716)) 1234 ((":" 2249) (return_type 2717)) 1235 ((T_DOUBLE_ARROW 2718)) 1236 ((backup_fn_flags 2719)) 1237 ((backup_lex_pos 2720)) 1238 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 2721) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1239 (("%" 1111) ("*" 1112) ("+" 1113) ("-" 1114) ("." 1115) ("/" 1116) ("<" 1117) (">" 1118) ("?" 1119) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1120) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1121) (T_BOOLEAN_AND 1122) (T_BOOLEAN_OR 1123) (T_COALESCE 1124) (T_INSTANCEOF 1125) (T_IS_EQUAL 1126) (T_IS_GREATER_OR_EQUAL 1127) (T_IS_IDENTICAL 1128) (T_IS_NOT_EQUAL 1129) (T_IS_NOT_IDENTICAL 1130) (T_IS_SMALLER_OR_EQUAL 1131) (T_LOGICAL_AND 1132) (T_LOGICAL_OR 1133) (T_LOGICAL_XOR 1134) (T_POW 1135) (T_SL 1136) (T_SPACESHIP 1137) (T_SR 1138) ("^" 1139) (backup_fn_flags 2722) ("|" 1140)) 1240 (("$" 2340) (T_ABSTRACT 2524) (T_ARRAY 2525) (T_AS 2526) (T_BREAK 2527) (T_CALLABLE 2528) (T_CASE 2529) (T_CATCH 2530) (T_CLASS 2531) (T_CLASS_C 2532) (T_CLONE 2533) (T_CONST 2534) (T_CONTINUE 2535) (T_DECLARE 2536) (T_DEFAULT 2537) (T_DIR 2538) (T_DO 2539) (T_ECHO 2540) (T_ELSE 2541) (T_ELSEIF 2542) (T_EMPTY 2543) (T_ENDDECLARE 2544) (T_ENDFOR 2545) (T_ENDFOREACH 2546) (T_ENDIF 2547) (T_ENDSWITCH 2548) (T_ENDWHILE 2549) (T_ENUM 2550) (T_EVAL 2551) (T_EXIT 2552) (T_EXTENDS 2553) (T_FILE 2554) (T_FINAL 2555) (T_FINALLY 2556) (T_FN 2557) (T_FOR 2558) (T_FOREACH 2559) (T_FUNCTION 2560) (T_FUNC_C 2561) (T_GLOBAL 2562) (T_GOTO 2563) (T_IF 2564) (T_IMPLEMENTS 2565) (T_INCLUDE 2566) (T_INCLUDE_ONCE 2567) (T_INSTANCEOF 2568) (T_INSTEADOF 2569) (T_INTERFACE 2570) (T_ISSET 2571) (T_LINE 2572) (T_LIST 2573) (T_LOGICAL_AND 2574) (T_LOGICAL_OR 2575) (T_LOGICAL_XOR 2576) (T_MATCH 2577) (T_METHOD_C 2578) (T_NAMESPACE 2579) (T_NEW 2580) (T_NS_C 2581) (T_PRINT 2582) (T_PRIVATE 2583) (T_PROPERTY_C 2584) (T_PROTECTED 2585) (T_PUBLIC 2586) (T_READONLY 2587) (T_REQUIRE 2588) (T_REQUIRE_ONCE 2589) (T_RETURN 2590) (T_STATIC 2591) (T_STRING 2592) (T_SWITCH 2593) (T_THROW 2594) (T_TRAIT 2595) (T_TRAIT_C 2596) (T_TRY 2597) (T_UNSET 2598) (T_USE 2599) (T_VAR 2600) (T_VARIABLE 2388) (T_WHILE 2601) (T_YIELD 2602) (identifier 2724) (member_name 2725) (reserved_non_modifiers 2605) (semi_reserved 2606) (simple_variable 2726) ("{" 2727)) 1241 (("(" 2611) (argument_list 2730)) 1242 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2728) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1243 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2729)) 1244 ((T_FN 35) (T_FUNCTION 129) (fn 2403) #'2404 (inline_function 2735)) 1245 (("$" 2340) (T_STRING 2741) (T_VARIABLE 2388) (property_name 2748) (simple_variable 2743) ("{" 2744)) 1246 (("$" 2340) (T_STRING 2741) (T_VARIABLE 2388) (property_name 2742) (simple_variable 2743) ("{" 2744)) 1247 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 2739) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1248 (("]" 2740)) 1249 (("(" 2611) (argument_list 2747)) 1250 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2745) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1251 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2746)) 1252 (("(" 2611) (argument_list 2749)) 1253 (("$" 2750) (T_VARIABLE 2752) (simple_variable 2795) ("{" 2796)) 1254 (("(" 2766) (argument_list 2794)) 1255 ((T_NULLSAFE_OBJECT_OPERATOR 2780) (T_OBJECT_OPERATOR 2781) ("[" 2782)) 1256 (("(" 2766) (argument_list 2779)) 1257 ((T_PAAMAYIM_NEKUDOTAYIM 2775)) 1258 (("(" 2766) (argument_list 2774)) 1259 ((T_PAAMAYIM_NEKUDOTAYIM 2763)) 1260 (("$" 2750) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 2752) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 2764) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 2765) ("{" 1642)) 1261 (("(" 2766) (argument_list 2767)) 1262 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 2768) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 2769) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 2770) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1263 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 2773) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1264 (("," 1480) (possible_comma 2771)) 1265 ((")" 2772)) 1266 (("$" 2750) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 2752) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 2776) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 2777) ("{" 1658)) 1267 (("(" 2766) (argument_list 2778)) 1268 (("$" 2750) (T_STRING 2785) (T_VARIABLE 2752) (property_name 2792) (simple_variable 2787) ("{" 2788)) 1269 (("$" 2750) (T_STRING 2785) (T_VARIABLE 2752) (property_name 2786) (simple_variable 2787) ("{" 2788)) 1270 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 2783) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1271 (("]" 2784)) 1272 (("(" 2766) (argument_list 2791)) 1273 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2789) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1274 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2790)) 1275 (("(" 2766) (argument_list 2793)) 1276 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2797) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1277 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2798)) 1278 (("\"" 1520) ("$" 2830) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 2831) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 2832) ("[" 1542) (array_object_dereferenceable 2833) (callable_expr 2834) (callable_variable 2835) (class_constant 1546) (class_name 2836) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 2837) (name 2838) (new_dereferenceable 1552) (simple_variable 2839) (static_member 2840) (variable 2841) (variable_class_name 2842)) 1279 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_STRING_VARNAME 2822) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2823) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1280 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_VARIABLE 2802) (encaps_var 2821)) 1281 ((T_NULLSAFE_OBJECT_OPERATOR 2809) (T_OBJECT_OPERATOR 2810) ("[" 2811)) 1282 (("`" 2808)) 1283 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2806) (T_VARIABLE 2802) (encaps_var 2807)) 1284 ((T_STRING 2820)) 1285 ((T_STRING 2819)) 1286 (("-" 2812) (T_NUM_STRING 2813) (T_STRING 2814) (T_VARIABLE 2815) (encaps_var_offset 2816)) 1287 ((T_NUM_STRING 2818)) 1288 (("]" 2817)) 1289 (("[" 2825) ("}" 2826)) 1290 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2824)) 1291 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 2827) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1292 (("%" 442) ("*" 443) ("+" 444) ("-" 445) ("." 446) ("/" 447) ("<" 448) (">" 449) ("?" 450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 452) (T_BOOLEAN_AND 453) (T_BOOLEAN_OR 454) (T_COALESCE 455) (T_INSTANCEOF 456) (T_IS_EQUAL 457) (T_IS_GREATER_OR_EQUAL 458) (T_IS_IDENTICAL 459) (T_IS_NOT_EQUAL 460) (T_IS_NOT_IDENTICAL 461) (T_IS_SMALLER_OR_EQUAL 462) (T_LOGICAL_AND 463) (T_LOGICAL_OR 464) (T_LOGICAL_XOR 465) (T_POW 466) (T_SL 467) (T_SPACESHIP 468) (T_SR 469) ("]" 2828) ("^" 470) ("|" 471)) 1293 (("}" 2829)) 1294 (("$" 2830) (T_VARIABLE 2832) (simple_variable 2876) ("{" 2877)) 1295 (("(" 2846) (argument_list 2875)) 1296 ((T_NULLSAFE_OBJECT_OPERATOR 2861) (T_OBJECT_OPERATOR 2862) ("[" 2863)) 1297 (("(" 2846) (argument_list 2860)) 1298 ((T_PAAMAYIM_NEKUDOTAYIM 2856)) 1299 (("(" 2846) (argument_list 2855)) 1300 (("}" 2854)) 1301 ((T_PAAMAYIM_NEKUDOTAYIM 2843)) 1302 (("$" 2830) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 2832) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 2844) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 2845) ("{" 1642)) 1303 (("(" 2846) (argument_list 2847)) 1304 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 2848) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 2849) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 2850) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1305 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 2853) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1306 (("," 1480) (possible_comma 2851)) 1307 ((")" 2852)) 1308 (("$" 2830) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 2832) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 2857) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 2858) ("{" 1658)) 1309 (("(" 2846) (argument_list 2859)) 1310 (("$" 2830) (T_STRING 2866) (T_VARIABLE 2832) (property_name 2873) (simple_variable 2868) ("{" 2869)) 1311 (("$" 2830) (T_STRING 2866) (T_VARIABLE 2832) (property_name 2867) (simple_variable 2868) ("{" 2869)) 1312 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 2864) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1313 (("]" 2865)) 1314 (("(" 2846) (argument_list 2872)) 1315 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2870) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1316 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2871)) 1317 (("(" 2846) (argument_list 2874)) 1318 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2878) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1319 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2879)) 1320 (("]" 2881)) 1321 (("=" 2882)) 1322 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2883) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1323 (("%" 1950) ("*" 1951) ("+" 1952) ("-" 1953) ("." 1954) ("/" 1955) ("<" 1956) (">" 1957) ("?" 1958) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1959) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1960) (T_BOOLEAN_AND 1961) (T_BOOLEAN_OR 1962) (T_COALESCE 1963) (T_DOUBLE_ARROW 2886) (T_INSTANCEOF 1964) (T_IS_EQUAL 1965) (T_IS_GREATER_OR_EQUAL 1966) (T_IS_IDENTICAL 1967) (T_IS_NOT_EQUAL 1968) (T_IS_NOT_IDENTICAL 1969) (T_IS_SMALLER_OR_EQUAL 1970) (T_LOGICAL_AND 1971) (T_LOGICAL_OR 1972) (T_LOGICAL_XOR 1973) (T_POW 1974) (T_SL 1975) (T_SPACESHIP 1976) (T_SR 1977) ("^" 1978) ("|" 1979)) 1324 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 2887) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 1325 (("\"" 1520) ("$" 2830) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 2831) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 2832) ("[" 1542) (array_object_dereferenceable 2833) (callable_expr 2834) (callable_variable 2835) (class_constant 1546) (class_name 2836) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 2837) (name 2838) (new_dereferenceable 1552) (simple_variable 2839) (static_member 2840) (variable 2919) (variable_class_name 2842)) 1326 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_STRING_VARNAME 2911) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 2912) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1327 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 2909) (T_VARIABLE 2896) (encaps_var 2910)) 1328 ((T_NULLSAFE_OBJECT_OPERATOR 2902) (T_OBJECT_OPERATOR 2903) ("[" 2904)) 1329 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 2900) (T_VARIABLE 2896) (encaps_var 2901)) 1330 ((T_STRING 2908)) 1331 ((T_STRING 2907)) 1332 (("-" 2812) (T_NUM_STRING 2813) (T_STRING 2814) (T_VARIABLE 2815) (encaps_var_offset 2905)) 1333 (("]" 2906)) 1334 (("[" 2914) ("}" 2915)) 1335 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 2913)) 1336 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 2916) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1337 (("%" 442) ("*" 443) ("+" 444) ("-" 445) ("." 446) ("/" 447) ("<" 448) (">" 449) ("?" 450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 452) (T_BOOLEAN_AND 453) (T_BOOLEAN_OR 454) (T_COALESCE 455) (T_INSTANCEOF 456) (T_IS_EQUAL 457) (T_IS_GREATER_OR_EQUAL 458) (T_IS_IDENTICAL 459) (T_IS_NOT_EQUAL 460) (T_IS_NOT_IDENTICAL 461) (T_IS_SMALLER_OR_EQUAL 462) (T_LOGICAL_AND 463) (T_LOGICAL_OR 464) (T_LOGICAL_XOR 465) (T_POW 466) (T_SL 467) (T_SPACESHIP 468) (T_SR 469) ("]" 2917) ("^" 470) ("|" 471)) 1338 (("}" 2918)) 1339 (("}" 2920)) 1340 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 7631) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 1341 ((T_NAME_FULLY_QUALIFIED 5830) (T_NAME_QUALIFIED 5831) (T_NAME_RELATIVE 5832) (T_STATIC 5833) (T_STRING 5834) (attribute_decl 5835) (attribute_group 7628) (class_name 5837) (name 5838)) 1342 ((T_ABSTRACT 10) (T_FINAL 34) (T_READONLY 2931) (class_modifier 7627)) 1343 ((T_CLASS 2959)) 1344 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 2957) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 2936) (attribute 2958) (class_modifier 2939)) 1345 ((T_PAAMAYIM_NEKUDOTAYIM 2955)) 1346 (("(" 2666) (argument_list 2954)) 1347 ((T_NULLSAFE_OBJECT_OPERATOR 2945) (T_OBJECT_OPERATOR 2946) (T_PAAMAYIM_NEKUDOTAYIM 2947) ("[" 2948)) 1348 (("$" 2650) (T_STRING 2685) (T_VARIABLE 2652) (property_name 2953) (simple_variable 2687) ("{" 2688)) 1349 (("$" 2650) (T_STRING 2685) (T_VARIABLE 2652) (property_name 2952) (simple_variable 2687) ("{" 2688)) 1350 (("$" 2650) (T_VARIABLE 2652) (simple_variable 2951)) 1351 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 2949) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1352 (("]" 2950)) 1353 (("$" 2650) (T_VARIABLE 2652) (simple_variable 2956)) 1354 (("(" 2960) (argument_list 2961) (ctor_arguments 2962)) 1355 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 7621) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 7622) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 7623) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1356 ((T_EXTENDS 2963) (extends_from 2964)) 1357 ((T_NAME_FULLY_QUALIFIED 7614) (T_NAME_QUALIFIED 7615) (T_NAME_RELATIVE 7616) (T_STATIC 7617) (T_STRING 7618) (class_name 7619) (name 7620)) 1358 ((T_IMPLEMENTS 2965) (implements_list 2966)) 1359 ((T_NAME_FULLY_QUALIFIED 7604) (T_NAME_QUALIFIED 7605) (T_NAME_RELATIVE 7606) (T_STATIC 7607) (T_STRING 7608) (class_name 7609) (class_name_list 7610) (name 7611)) 1360 ((backup_doc_comment 2967)) 1361 (("{" 2968)) 1362 ((class_statement_list 2969)) 1363 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 2994)) 1364 ((T_NAME_FULLY_QUALIFIED 5830) (T_NAME_QUALIFIED 5831) (T_NAME_RELATIVE 5832) (T_STATIC 5833) (T_STRING 5834) (attribute_decl 5835) (attribute_group 7601) (class_name 5837) (name 5838)) 1365 ((backup_doc_comment 7514)) 1366 ((T_NAME_FULLY_QUALIFIED 7213) (T_NAME_QUALIFIED 7214) (T_NAME_RELATIVE 7215) (T_STATIC 7216) (T_STRING 7217) (class_name 7218) (class_name_list 7219) (name 7220)) 1367 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_VAR 2983) (attribute 7211) (attributed_class_statement 7212) (class_const_modifiers 2987) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993)) 1368 ((T_CONST 7061)) 1369 ((T_FUNCTION 6905) #'6906) 1370 ((T_ABSTRACT 2970) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (member_modifier 6904)) 1371 (("(" 2995) ("?" 2996) (T_ARRAY 2997) (T_CALLABLE 2998) (T_NAME_FULLY_QUALIFIED 2999) (T_NAME_QUALIFIED 3000) (T_NAME_RELATIVE 3001) (T_STRING 3002) (intersection_type_without_static 3003) (name 3004) (optional_type_without_static 3005) (type_expr_without_static 3006) (type_without_static 3007) (union_type_without_static 3008) (union_type_without_static_element 3009)) 1372 ((T_ARRAY 2285) (T_CALLABLE 2286) (T_NAME_FULLY_QUALIFIED 2287) (T_NAME_QUALIFIED 2288) (T_NAME_RELATIVE 2289) (T_STRING 2291) (intersection_type_without_static 6902) (name 2293) (type_without_static 3022)) 1373 ((T_ARRAY 6894) (T_CALLABLE 6895) (T_NAME_FULLY_QUALIFIED 6896) (T_NAME_QUALIFIED 6897) (T_NAME_RELATIVE 6898) (T_STRING 6899) (name 6900) (type_without_static 6901)) 1374 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6892)) 1375 ((T_VARIABLE 3039) (hooked_property 3040) (property 3041) (property_list 3042)) 1376 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3030)) 1377 (("|" 3028)) 1378 (("|" 3010)) 1379 (("(" 3011) (T_ARRAY 3012) (T_CALLABLE 3013) (T_NAME_FULLY_QUALIFIED 3014) (T_NAME_QUALIFIED 3015) (T_NAME_RELATIVE 3016) (T_STRING 3017) (name 3018) (type_without_static 3019) (union_type_without_static_element 3020)) 1380 ((T_ARRAY 2285) (T_CALLABLE 2286) (T_NAME_FULLY_QUALIFIED 2287) (T_NAME_QUALIFIED 2288) (T_NAME_RELATIVE 2289) (T_STRING 2291) (intersection_type_without_static 3021) (name 2293) (type_without_static 3022)) 1381 ((")" 3025) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3026)) 1382 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3023)) 1383 ((T_ARRAY 2297) (T_CALLABLE 2298) (T_NAME_FULLY_QUALIFIED 2299) (T_NAME_QUALIFIED 2300) (T_NAME_RELATIVE 2301) (T_STRING 2303) (name 2304) (type_without_static 3024)) 1384 ((T_ARRAY 2297) (T_CALLABLE 2298) (T_NAME_FULLY_QUALIFIED 2299) (T_NAME_QUALIFIED 2300) (T_NAME_RELATIVE 2301) (T_STRING 2303) (name 2304) (type_without_static 3027)) 1385 (("(" 3011) (T_ARRAY 3012) (T_CALLABLE 3013) (T_NAME_FULLY_QUALIFIED 3014) (T_NAME_QUALIFIED 3015) (T_NAME_RELATIVE 3016) (T_STRING 3017) (name 3018) (type_without_static 3019) (union_type_without_static_element 3029)) 1386 ((T_ARRAY 3031) (T_CALLABLE 3032) (T_NAME_FULLY_QUALIFIED 3033) (T_NAME_QUALIFIED 3034) (T_NAME_RELATIVE 3035) (T_STRING 3036) (name 3037) (type_without_static 3038)) 1387 (("=" 5687) (backup_doc_comment 5688)) 1388 (("," 3043) (";" 3044)) 1389 ((T_VARIABLE 3045) (property 3046)) 1390 (("=" 3047) (backup_doc_comment 3048)) 1391 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3112) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1392 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5686) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1393 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 5684) (encaps_var 5237)) 1394 (("$" 3051) (T_VARIABLE 3099) (simple_variable 5680) ("{" 5681)) 1395 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5678) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 1396 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5677) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1397 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5676) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1398 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5675) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1399 (("(" 5672)) 1400 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5671) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1401 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5670) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1402 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5669) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1403 (("\"" 1520) ("$" 3354) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 3355) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 3356) ("[" 1542) (array_object_dereferenceable 3357) (callable_expr 3358) (callable_variable 3359) (class_constant 1546) (class_name 3360) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 3361) (name 3362) (new_dereferenceable 1552) (simple_variable 3363) (static_member 3364) (variable 5668) (variable_class_name 3366)) 1404 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5667) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1405 (("(" 5664)) 1406 (("(" 5661)) 1407 (("(" 5652) (argument_list 5653) (ctor_arguments 5654)) 1408 (("\"" 1520) ("$" 3354) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 3355) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 3356) ("[" 1542) (array_object_dereferenceable 3357) (callable_expr 3358) (callable_variable 3359) (class_constant 1546) (class_name 3360) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 3361) (name 3362) (new_dereferenceable 1552) (simple_variable 3363) (static_member 3364) (variable 5651) (variable_class_name 3366)) 1409 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5650) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1410 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5649) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1411 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5648) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1412 (("(" 5644)) 1413 (("(" 5639)) 1414 (("(" 5633)) 1415 (("$" 3354) ("(" 5596) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 5597) (T_NAME_QUALIFIED 5598) (T_NAME_RELATIVE 5599) (T_READONLY 2931) (T_STATIC 5600) (T_STRING 5601) (T_VARIABLE 3356) (anonymous_class 5602) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 5603) (attribute 2937) (attributes 5604) (class_modifier 2939) (class_name 5605) (class_name_reference 5606) (name 5607) (new_variable 5608) (simple_variable 5609)) 1416 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5595) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1417 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5594) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1418 (("(" 3315) (argument_list 5593)) 1419 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5592) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1420 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5591) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1421 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 5586) (T_END_HEREDOC 5587) (T_VARIABLE 2896) (encaps_list 5588) (encaps_var 2898)) 1422 ((T_FN 35) (T_FUNCTION 129) (fn 3113) #'3114 (inline_function 5585)) 1423 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5584) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1424 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5583) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1425 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5582) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1426 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3524) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1427 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3460) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1428 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 3456) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 1429 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 3454) (encaps_list 2804) (encaps_var 2805)) 1430 ((T_NULLSAFE_OBJECT_OPERATOR 3440) (T_OBJECT_OPERATOR 3441) ("[" 3442)) 1431 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 3437) (attribute 2733) (fn 3113) #'3114 (inline_function 3438)) 1432 (("(" 3315) (argument_list 3436)) 1433 ((T_PAAMAYIM_NEKUDOTAYIM 3428)) 1434 (("%" 3129) ("*" 3130) ("+" 3131) ("-" 3132) ("." 3133) ("/" 3134) ("<" 3135) (">" 3136) ("?" 3137) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3138) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3139) (T_BOOLEAN_AND 3140) (T_BOOLEAN_OR 3141) (T_COALESCE 3142) (T_INSTANCEOF 3143) (T_IS_EQUAL 3144) (T_IS_GREATER_OR_EQUAL 3145) (T_IS_IDENTICAL 3146) (T_IS_NOT_EQUAL 3147) (T_IS_NOT_IDENTICAL 3148) (T_IS_SMALLER_OR_EQUAL 3149) (T_LOGICAL_AND 3150) (T_LOGICAL_OR 3151) (T_LOGICAL_XOR 3152) (T_POW 3153) (T_SL 3154) (T_SPACESHIP 3155) (T_SR 3156) ("^" 3157) (backup_doc_comment 3427) ("|" 3158)) 1435 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 3416)) 1436 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 3404)) 1437 (("(" 3315) (argument_list 3403)) 1438 (("=" 3323) (T_AND_EQUAL 3324) (T_COALESCE_EQUAL 3325) (T_CONCAT_EQUAL 3326) (T_DEC 3327) (T_DIV_EQUAL 3328) (T_INC 3329) (T_MINUS_EQUAL 3330) (T_MOD_EQUAL 3331) (T_MUL_EQUAL 3332) (T_OR_EQUAL 3333) (T_PLUS_EQUAL 3334) (T_POW_EQUAL 3335) (T_SL_EQUAL 3336) (T_SR_EQUAL 3337) (T_XOR_EQUAL 3338)) 1439 ((T_PAAMAYIM_NEKUDOTAYIM 3227)) 1440 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3128) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1441 (("%" 3129) ("*" 3130) ("+" 3131) ("-" 3132) ("." 3133) ("/" 3134) ("<" 3135) (">" 3136) ("?" 3137) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3138) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3139) (T_BOOLEAN_AND 3140) (T_BOOLEAN_OR 3141) (T_COALESCE 3142) (T_INSTANCEOF 3143) (T_IS_EQUAL 3144) (T_IS_GREATER_OR_EQUAL 3145) (T_IS_IDENTICAL 3146) (T_IS_NOT_EQUAL 3147) (T_IS_NOT_IDENTICAL 3148) (T_IS_SMALLER_OR_EQUAL 3149) (T_LOGICAL_AND 3150) (T_LOGICAL_OR 3151) (T_LOGICAL_XOR 3152) (T_POW 3153) (T_SL 3154) (T_SPACESHIP 3155) (T_SR 3156) ("^" 3157) ("|" 3158)) 1442 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3226) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1443 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3225) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1444 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3224) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1445 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3223) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1446 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3222) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1447 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3221) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1448 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3220) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1449 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3219) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1450 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 3214) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 3215) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 1451 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3213) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1452 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3212) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1453 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3211) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1454 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3210) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1455 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3209) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1456 (("$" 3174) ("(" 3175) (T_NAME_FULLY_QUALIFIED 3176) (T_NAME_QUALIFIED 3177) (T_NAME_RELATIVE 3178) (T_STATIC 3179) (T_STRING 3180) (T_VARIABLE 3181) (class_name 3182) (class_name_reference 3183) (name 3184) (new_variable 3185) (simple_variable 3186)) 1457 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3173) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1458 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3172) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1459 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3171) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1460 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3170) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1461 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3169) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1462 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3168) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1463 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3167) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1464 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3166) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1465 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3165) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1466 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3164) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1467 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3163) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1468 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3162) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1469 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3161) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1470 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3160) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1471 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3159) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1472 (("$" 3174) (T_VARIABLE 3181) (simple_variable 3205) ("{" 3206)) 1473 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 3203) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 1474 ((T_PAAMAYIM_NEKUDOTAYIM 3201)) 1475 ((T_NULLSAFE_OBJECT_OPERATOR 3187) (T_OBJECT_OPERATOR 3188) (T_PAAMAYIM_NEKUDOTAYIM 3189) ("[" 3190)) 1476 (("$" 3174) (T_STRING 3194) (T_VARIABLE 3181) (property_name 3200) (simple_variable 3196) ("{" 3197)) 1477 (("$" 3174) (T_STRING 3194) (T_VARIABLE 3181) (property_name 3195) (simple_variable 3196) ("{" 3197)) 1478 (("$" 3174) (T_VARIABLE 3181) (simple_variable 3193)) 1479 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 3191) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1480 (("]" 3192)) 1481 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 3198) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1482 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 3199)) 1483 (("$" 3174) (T_VARIABLE 3181) (simple_variable 3202)) 1484 (("%" 299) (")" 3204) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 1485 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 3207) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1486 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 3208)) 1487 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3218) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1488 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 3216) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 1489 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3217) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1490 (("$" 3051) (T_ABSTRACT 3228) (T_ARRAY 3229) (T_AS 3230) (T_BREAK 3231) (T_CALLABLE 3232) (T_CASE 3233) (T_CATCH 3234) (T_CLASS 3235) (T_CLASS_C 3236) (T_CLONE 3237) (T_CONST 3238) (T_CONTINUE 3239) (T_DECLARE 3240) (T_DEFAULT 3241) (T_DIR 3242) (T_DO 3243) (T_ECHO 3244) (T_ELSE 3245) (T_ELSEIF 3246) (T_EMPTY 3247) (T_ENDDECLARE 3248) (T_ENDFOR 3249) (T_ENDFOREACH 3250) (T_ENDIF 3251) (T_ENDSWITCH 3252) (T_ENDWHILE 3253) (T_ENUM 3254) (T_EVAL 3255) (T_EXIT 3256) (T_EXTENDS 3257) (T_FILE 3258) (T_FINAL 3259) (T_FINALLY 3260) (T_FN 3261) (T_FOR 3262) (T_FOREACH 3263) (T_FUNCTION 3264) (T_FUNC_C 3265) (T_GLOBAL 3266) (T_GOTO 3267) (T_IF 3268) (T_IMPLEMENTS 3269) (T_INCLUDE 3270) (T_INCLUDE_ONCE 3271) (T_INSTANCEOF 3272) (T_INSTEADOF 3273) (T_INTERFACE 3274) (T_ISSET 3275) (T_LINE 3276) (T_LIST 3277) (T_LOGICAL_AND 3278) (T_LOGICAL_OR 3279) (T_LOGICAL_XOR 3280) (T_MATCH 3281) (T_METHOD_C 3282) (T_NAMESPACE 3283) (T_NEW 3284) (T_NS_C 3285) (T_PRINT 3286) (T_PRIVATE 3287) (T_PROPERTY_C 3288) (T_PROTECTED 3289) (T_PUBLIC 3290) (T_READONLY 3291) (T_REQUIRE 3292) (T_REQUIRE_ONCE 3293) (T_RETURN 3294) (T_STATIC 3295) (T_STRING 3296) (T_SWITCH 3297) (T_THROW 3298) (T_TRAIT 3299) (T_TRAIT_C 3300) (T_TRY 3301) (T_UNSET 3302) (T_USE 3303) (T_VAR 3304) (T_VARIABLE 3099) (T_WHILE 3305) (T_YIELD 3306) (identifier 3307) (member_name 3308) (reserved_non_modifiers 3309) (semi_reserved 3310) (simple_variable 3311) ("{" 3312)) 1491 (("(" 3315) (argument_list 3316)) 1492 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 3313) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1493 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 3314)) 1494 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 3317) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 3318) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 3319) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1495 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 3322) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1496 (("," 1480) (possible_comma 3320)) 1497 ((")" 3321)) 1498 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (ampersand 3352) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3353) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1499 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3351) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1500 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3350) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1501 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3349) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1502 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3348) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1503 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3347) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1504 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3346) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1505 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3345) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1506 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3344) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1507 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3343) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1508 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3342) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1509 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3341) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1510 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3340) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1511 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3339) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1512 (("\"" 1520) ("$" 3354) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 3355) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 3356) ("[" 1542) (array_object_dereferenceable 3357) (callable_expr 3358) (callable_variable 3359) (class_constant 1546) (class_name 3360) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 3361) (name 3362) (new_dereferenceable 1552) (simple_variable 3363) (static_member 3364) (variable 3365) (variable_class_name 3366)) 1513 (("$" 3354) (T_VARIABLE 3356) (simple_variable 3399) ("{" 3400)) 1514 (("(" 3370) (argument_list 3398)) 1515 ((T_NULLSAFE_OBJECT_OPERATOR 3384) (T_OBJECT_OPERATOR 3385) ("[" 3386)) 1516 (("(" 3370) (argument_list 3383)) 1517 ((T_PAAMAYIM_NEKUDOTAYIM 3379)) 1518 (("(" 3370) (argument_list 3378)) 1519 ((T_PAAMAYIM_NEKUDOTAYIM 3367)) 1520 (("$" 3354) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 3356) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 3368) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 3369) ("{" 1642)) 1521 (("(" 3370) (argument_list 3371)) 1522 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 3372) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 3373) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 3374) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1523 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 3377) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1524 (("," 1480) (possible_comma 3375)) 1525 ((")" 3376)) 1526 (("$" 3354) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 3356) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 3380) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 3381) ("{" 1658)) 1527 (("(" 3370) (argument_list 3382)) 1528 (("$" 3354) (T_STRING 3389) (T_VARIABLE 3356) (property_name 3396) (simple_variable 3391) ("{" 3392)) 1529 (("$" 3354) (T_STRING 3389) (T_VARIABLE 3356) (property_name 3390) (simple_variable 3391) ("{" 3392)) 1530 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 3387) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1531 (("]" 3388)) 1532 (("(" 3370) (argument_list 3395)) 1533 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 3393) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1534 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 3394)) 1535 (("(" 3370) (argument_list 3397)) 1536 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 3401) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1537 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 3402)) 1538 ((backup_doc_comment 3405)) 1539 (("(" 3406)) 1540 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 3407)) 1541 ((")" 3408)) 1542 ((T_USE 164) (lexical_vars 3409)) 1543 ((":" 166) (return_type 3410)) 1544 ((backup_fn_flags 3411)) 1545 (("{" 3412)) 1546 ((inner_statement_list 3413)) 1547 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 3414) ("~" 127)) 1548 ((backup_fn_flags 3415)) 1549 ((backup_doc_comment 3417)) 1550 (("(" 3418)) 1551 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 3419)) 1552 ((")" 3420)) 1553 ((":" 2249) (return_type 3421)) 1554 ((T_DOUBLE_ARROW 3422)) 1555 ((backup_fn_flags 3423)) 1556 ((backup_lex_pos 3424)) 1557 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3425) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1558 (("%" 3129) ("*" 3130) ("+" 3131) ("-" 3132) ("." 3133) ("/" 3134) ("<" 3135) (">" 3136) ("?" 3137) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3138) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3139) (T_BOOLEAN_AND 3140) (T_BOOLEAN_OR 3141) (T_COALESCE 3142) (T_INSTANCEOF 3143) (T_IS_EQUAL 3144) (T_IS_GREATER_OR_EQUAL 3145) (T_IS_IDENTICAL 3146) (T_IS_NOT_EQUAL 3147) (T_IS_NOT_IDENTICAL 3148) (T_IS_SMALLER_OR_EQUAL 3149) (T_LOGICAL_AND 3150) (T_LOGICAL_OR 3151) (T_LOGICAL_XOR 3152) (T_POW 3153) (T_SL 3154) (T_SPACESHIP 3155) (T_SR 3156) ("^" 3157) (backup_fn_flags 3426) ("|" 3158)) 1559 (("$" 3051) (T_ABSTRACT 3228) (T_ARRAY 3229) (T_AS 3230) (T_BREAK 3231) (T_CALLABLE 3232) (T_CASE 3233) (T_CATCH 3234) (T_CLASS 3235) (T_CLASS_C 3236) (T_CLONE 3237) (T_CONST 3238) (T_CONTINUE 3239) (T_DECLARE 3240) (T_DEFAULT 3241) (T_DIR 3242) (T_DO 3243) (T_ECHO 3244) (T_ELSE 3245) (T_ELSEIF 3246) (T_EMPTY 3247) (T_ENDDECLARE 3248) (T_ENDFOR 3249) (T_ENDFOREACH 3250) (T_ENDIF 3251) (T_ENDSWITCH 3252) (T_ENDWHILE 3253) (T_ENUM 3254) (T_EVAL 3255) (T_EXIT 3256) (T_EXTENDS 3257) (T_FILE 3258) (T_FINAL 3259) (T_FINALLY 3260) (T_FN 3261) (T_FOR 3262) (T_FOREACH 3263) (T_FUNCTION 3264) (T_FUNC_C 3265) (T_GLOBAL 3266) (T_GOTO 3267) (T_IF 3268) (T_IMPLEMENTS 3269) (T_INCLUDE 3270) (T_INCLUDE_ONCE 3271) (T_INSTANCEOF 3272) (T_INSTEADOF 3273) (T_INTERFACE 3274) (T_ISSET 3275) (T_LINE 3276) (T_LIST 3277) (T_LOGICAL_AND 3278) (T_LOGICAL_OR 3279) (T_LOGICAL_XOR 3280) (T_MATCH 3281) (T_METHOD_C 3282) (T_NAMESPACE 3283) (T_NEW 3284) (T_NS_C 3285) (T_PRINT 3286) (T_PRIVATE 3287) (T_PROPERTY_C 3288) (T_PROTECTED 3289) (T_PUBLIC 3290) (T_READONLY 3291) (T_REQUIRE 3292) (T_REQUIRE_ONCE 3293) (T_RETURN 3294) (T_STATIC 3295) (T_STRING 3296) (T_SWITCH 3297) (T_THROW 3298) (T_TRAIT 3299) (T_TRAIT_C 3300) (T_TRY 3301) (T_UNSET 3302) (T_USE 3303) (T_VAR 3304) (T_VARIABLE 3099) (T_WHILE 3305) (T_YIELD 3306) (identifier 3429) (member_name 3430) (reserved_non_modifiers 3309) (semi_reserved 3310) (simple_variable 3431) ("{" 3432)) 1560 (("(" 3315) (argument_list 3435)) 1561 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 3433) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1562 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 3434)) 1563 ((T_FN 35) (T_FUNCTION 129) (fn 3113) #'3114 (inline_function 3439)) 1564 (("$" 3051) (T_STRING 3445) (T_VARIABLE 3099) (property_name 3452) (simple_variable 3447) ("{" 3448)) 1565 (("$" 3051) (T_STRING 3445) (T_VARIABLE 3099) (property_name 3446) (simple_variable 3447) ("{" 3448)) 1566 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 3443) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1567 (("]" 3444)) 1568 (("(" 3315) (argument_list 3451)) 1569 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 3449) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1570 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 3450)) 1571 (("(" 3315) (argument_list 3453)) 1572 (("`" 3455)) 1573 (("]" 3457)) 1574 (("=" 3458)) 1575 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3459) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1576 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 5581) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1577 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 5579) (encaps_var 5237)) 1578 (("$" 3463) (T_VARIABLE 3511) (simple_variable 5575) ("{" 5576)) 1579 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5573) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 1580 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 5572) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1581 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 5571) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1582 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 5570) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1583 (("(" 5567)) 1584 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 5566) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1585 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 5565) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1586 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 5564) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1587 (("\"" 1520) ("$" 3766) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 3767) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 3768) ("[" 1542) (array_object_dereferenceable 3769) (callable_expr 3770) (callable_variable 3771) (class_constant 1546) (class_name 3772) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 3773) (name 3774) (new_dereferenceable 1552) (simple_variable 3775) (static_member 3776) (variable 5563) (variable_class_name 3778)) 1588 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 5562) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1589 (("(" 5559)) 1590 (("(" 5556)) 1591 (("(" 5547) (argument_list 5548) (ctor_arguments 5549)) 1592 (("\"" 1520) ("$" 3766) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 3767) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 3768) ("[" 1542) (array_object_dereferenceable 3769) (callable_expr 3770) (callable_variable 3771) (class_constant 1546) (class_name 3772) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 3773) (name 3774) (new_dereferenceable 1552) (simple_variable 3775) (static_member 3776) (variable 5546) (variable_class_name 3778)) 1593 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 5545) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1594 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 5544) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1595 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 5543) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1596 (("(" 5539)) 1597 (("(" 5534)) 1598 (("(" 3928)) 1599 (("$" 3766) ("(" 3891) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 3892) (T_NAME_QUALIFIED 3893) (T_NAME_RELATIVE 3894) (T_READONLY 2931) (T_STATIC 3895) (T_STRING 3896) (T_VARIABLE 3768) (anonymous_class 3897) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 3898) (attribute 2937) (attributes 3899) (class_modifier 2939) (class_name 3900) (class_name_reference 3901) (name 3902) (new_variable 3903) (simple_variable 3904)) 1600 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3890) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1601 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3889) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1602 (("(" 3727) (argument_list 3888)) 1603 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3887) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1604 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3886) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1605 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 3881) (T_END_HEREDOC 3882) (T_VARIABLE 2896) (encaps_list 3883) (encaps_var 2898)) 1606 ((T_FN 35) (T_FUNCTION 129) (fn 3525) #'3526 (inline_function 3880)) 1607 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3879) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1608 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3878) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1609 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3877) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1610 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3874) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1611 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3873) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1612 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 3869) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 1613 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 3867) (encaps_list 2804) (encaps_var 2805)) 1614 ((T_NULLSAFE_OBJECT_OPERATOR 3853) (T_OBJECT_OPERATOR 3854) ("[" 3855)) 1615 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 3850) (attribute 2733) (fn 3525) #'3526 (inline_function 3851)) 1616 (("(" 3727) (argument_list 3849)) 1617 ((T_PAAMAYIM_NEKUDOTAYIM 3841)) 1618 (("%" 3541) ("*" 3542) ("+" 3543) ("-" 3544) ("." 3545) ("/" 3546) ("<" 3547) (">" 3548) ("?" 3549) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3550) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3551) (T_BOOLEAN_AND 3552) (T_BOOLEAN_OR 3553) (T_COALESCE 3554) (T_DOUBLE_ARROW 3839) (T_INSTANCEOF 3555) (T_IS_EQUAL 3556) (T_IS_GREATER_OR_EQUAL 3557) (T_IS_IDENTICAL 3558) (T_IS_NOT_EQUAL 3559) (T_IS_NOT_IDENTICAL 3560) (T_IS_SMALLER_OR_EQUAL 3561) (T_LOGICAL_AND 3562) (T_LOGICAL_OR 3563) (T_LOGICAL_XOR 3564) (T_POW 3565) (T_SL 3566) (T_SPACESHIP 3567) (T_SR 3568) ("^" 3569) ("|" 3570)) 1619 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 3828)) 1620 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 3816)) 1621 (("(" 3727) (argument_list 3815)) 1622 (("=" 3735) (T_AND_EQUAL 3736) (T_COALESCE_EQUAL 3737) (T_CONCAT_EQUAL 3738) (T_DEC 3739) (T_DIV_EQUAL 3740) (T_INC 3741) (T_MINUS_EQUAL 3742) (T_MOD_EQUAL 3743) (T_MUL_EQUAL 3744) (T_OR_EQUAL 3745) (T_PLUS_EQUAL 3746) (T_POW_EQUAL 3747) (T_SL_EQUAL 3748) (T_SR_EQUAL 3749) (T_XOR_EQUAL 3750)) 1623 ((T_PAAMAYIM_NEKUDOTAYIM 3639)) 1624 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3540) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1625 (("%" 3541) ("*" 3542) ("+" 3543) ("-" 3544) ("." 3545) ("/" 3546) ("<" 3547) (">" 3548) ("?" 3549) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3550) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3551) (T_BOOLEAN_AND 3552) (T_BOOLEAN_OR 3553) (T_COALESCE 3554) (T_INSTANCEOF 3555) (T_IS_EQUAL 3556) (T_IS_GREATER_OR_EQUAL 3557) (T_IS_IDENTICAL 3558) (T_IS_NOT_EQUAL 3559) (T_IS_NOT_IDENTICAL 3560) (T_IS_SMALLER_OR_EQUAL 3561) (T_LOGICAL_AND 3562) (T_LOGICAL_OR 3563) (T_LOGICAL_XOR 3564) (T_POW 3565) (T_SL 3566) (T_SPACESHIP 3567) (T_SR 3568) ("^" 3569) ("|" 3570)) 1626 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3638) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1627 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3637) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1628 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3636) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1629 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3635) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1630 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3634) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1631 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3633) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1632 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3632) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1633 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3631) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1634 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 3626) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 3627) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 1635 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3625) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1636 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3624) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1637 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3623) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1638 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3622) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1639 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3621) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1640 (("$" 3586) ("(" 3587) (T_NAME_FULLY_QUALIFIED 3588) (T_NAME_QUALIFIED 3589) (T_NAME_RELATIVE 3590) (T_STATIC 3591) (T_STRING 3592) (T_VARIABLE 3593) (class_name 3594) (class_name_reference 3595) (name 3596) (new_variable 3597) (simple_variable 3598)) 1641 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3585) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1642 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3584) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1643 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3583) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1644 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3582) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1645 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3581) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1646 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3580) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1647 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3579) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1648 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3578) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1649 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3577) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1650 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3576) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1651 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3575) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1652 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3574) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1653 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3573) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1654 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3572) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1655 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3571) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1656 (("$" 3586) (T_VARIABLE 3593) (simple_variable 3617) ("{" 3618)) 1657 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 3615) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 1658 ((T_PAAMAYIM_NEKUDOTAYIM 3613)) 1659 ((T_NULLSAFE_OBJECT_OPERATOR 3599) (T_OBJECT_OPERATOR 3600) (T_PAAMAYIM_NEKUDOTAYIM 3601) ("[" 3602)) 1660 (("$" 3586) (T_STRING 3606) (T_VARIABLE 3593) (property_name 3612) (simple_variable 3608) ("{" 3609)) 1661 (("$" 3586) (T_STRING 3606) (T_VARIABLE 3593) (property_name 3607) (simple_variable 3608) ("{" 3609)) 1662 (("$" 3586) (T_VARIABLE 3593) (simple_variable 3605)) 1663 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 3603) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1664 (("]" 3604)) 1665 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 3610) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1666 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 3611)) 1667 (("$" 3586) (T_VARIABLE 3593) (simple_variable 3614)) 1668 (("%" 299) (")" 3616) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 1669 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 3619) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1670 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 3620)) 1671 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3630) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1672 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 3628) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 1673 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3629) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1674 (("$" 3463) (T_ABSTRACT 3640) (T_ARRAY 3641) (T_AS 3642) (T_BREAK 3643) (T_CALLABLE 3644) (T_CASE 3645) (T_CATCH 3646) (T_CLASS 3647) (T_CLASS_C 3648) (T_CLONE 3649) (T_CONST 3650) (T_CONTINUE 3651) (T_DECLARE 3652) (T_DEFAULT 3653) (T_DIR 3654) (T_DO 3655) (T_ECHO 3656) (T_ELSE 3657) (T_ELSEIF 3658) (T_EMPTY 3659) (T_ENDDECLARE 3660) (T_ENDFOR 3661) (T_ENDFOREACH 3662) (T_ENDIF 3663) (T_ENDSWITCH 3664) (T_ENDWHILE 3665) (T_ENUM 3666) (T_EVAL 3667) (T_EXIT 3668) (T_EXTENDS 3669) (T_FILE 3670) (T_FINAL 3671) (T_FINALLY 3672) (T_FN 3673) (T_FOR 3674) (T_FOREACH 3675) (T_FUNCTION 3676) (T_FUNC_C 3677) (T_GLOBAL 3678) (T_GOTO 3679) (T_IF 3680) (T_IMPLEMENTS 3681) (T_INCLUDE 3682) (T_INCLUDE_ONCE 3683) (T_INSTANCEOF 3684) (T_INSTEADOF 3685) (T_INTERFACE 3686) (T_ISSET 3687) (T_LINE 3688) (T_LIST 3689) (T_LOGICAL_AND 3690) (T_LOGICAL_OR 3691) (T_LOGICAL_XOR 3692) (T_MATCH 3693) (T_METHOD_C 3694) (T_NAMESPACE 3695) (T_NEW 3696) (T_NS_C 3697) (T_PRINT 3698) (T_PRIVATE 3699) (T_PROPERTY_C 3700) (T_PROTECTED 3701) (T_PUBLIC 3702) (T_READONLY 3703) (T_REQUIRE 3704) (T_REQUIRE_ONCE 3705) (T_RETURN 3706) (T_STATIC 3707) (T_STRING 3708) (T_SWITCH 3709) (T_THROW 3710) (T_TRAIT 3711) (T_TRAIT_C 3712) (T_TRY 3713) (T_UNSET 3714) (T_USE 3715) (T_VAR 3716) (T_VARIABLE 3511) (T_WHILE 3717) (T_YIELD 3718) (identifier 3719) (member_name 3720) (reserved_non_modifiers 3721) (semi_reserved 3722) (simple_variable 3723) ("{" 3724)) 1675 (("(" 3727) (argument_list 3728)) 1676 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 3725) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1677 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 3726)) 1678 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 3729) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 3730) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 3731) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1679 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 3734) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1680 (("," 1480) (possible_comma 3732)) 1681 ((")" 3733)) 1682 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (ampersand 3764) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3765) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1683 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3763) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1684 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3762) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1685 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3761) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1686 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3760) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1687 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3759) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1688 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3758) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1689 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3757) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1690 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3756) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1691 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3755) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1692 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3754) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1693 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3753) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1694 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3752) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1695 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3751) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1696 (("\"" 1520) ("$" 3766) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 3767) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 3768) ("[" 1542) (array_object_dereferenceable 3769) (callable_expr 3770) (callable_variable 3771) (class_constant 1546) (class_name 3772) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 3773) (name 3774) (new_dereferenceable 1552) (simple_variable 3775) (static_member 3776) (variable 3777) (variable_class_name 3778)) 1697 (("$" 3766) (T_VARIABLE 3768) (simple_variable 3811) ("{" 3812)) 1698 (("(" 3782) (argument_list 3810)) 1699 ((T_NULLSAFE_OBJECT_OPERATOR 3796) (T_OBJECT_OPERATOR 3797) ("[" 3798)) 1700 (("(" 3782) (argument_list 3795)) 1701 ((T_PAAMAYIM_NEKUDOTAYIM 3791)) 1702 (("(" 3782) (argument_list 3790)) 1703 ((T_PAAMAYIM_NEKUDOTAYIM 3779)) 1704 (("$" 3766) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 3768) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 3780) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 3781) ("{" 1642)) 1705 (("(" 3782) (argument_list 3783)) 1706 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 3784) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 3785) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 3786) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1707 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 3789) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1708 (("," 1480) (possible_comma 3787)) 1709 ((")" 3788)) 1710 (("$" 3766) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 3768) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 3792) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 3793) ("{" 1658)) 1711 (("(" 3782) (argument_list 3794)) 1712 (("$" 3766) (T_STRING 3801) (T_VARIABLE 3768) (property_name 3808) (simple_variable 3803) ("{" 3804)) 1713 (("$" 3766) (T_STRING 3801) (T_VARIABLE 3768) (property_name 3802) (simple_variable 3803) ("{" 3804)) 1714 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 3799) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1715 (("]" 3800)) 1716 (("(" 3782) (argument_list 3807)) 1717 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 3805) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1718 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 3806)) 1719 (("(" 3782) (argument_list 3809)) 1720 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 3813) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1721 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 3814)) 1722 ((backup_doc_comment 3817)) 1723 (("(" 3818)) 1724 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 3819)) 1725 ((")" 3820)) 1726 ((T_USE 164) (lexical_vars 3821)) 1727 ((":" 166) (return_type 3822)) 1728 ((backup_fn_flags 3823)) 1729 (("{" 3824)) 1730 ((inner_statement_list 3825)) 1731 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 3826) ("~" 127)) 1732 ((backup_fn_flags 3827)) 1733 ((backup_doc_comment 3829)) 1734 (("(" 3830)) 1735 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 3831)) 1736 ((")" 3832)) 1737 ((":" 2249) (return_type 3833)) 1738 ((T_DOUBLE_ARROW 3834)) 1739 ((backup_fn_flags 3835)) 1740 ((backup_lex_pos 3836)) 1741 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3837) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1742 (("%" 3541) ("*" 3542) ("+" 3543) ("-" 3544) ("." 3545) ("/" 3546) ("<" 3547) (">" 3548) ("?" 3549) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3550) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3551) (T_BOOLEAN_AND 3552) (T_BOOLEAN_OR 3553) (T_COALESCE 3554) (T_INSTANCEOF 3555) (T_IS_EQUAL 3556) (T_IS_GREATER_OR_EQUAL 3557) (T_IS_IDENTICAL 3558) (T_IS_NOT_EQUAL 3559) (T_IS_NOT_IDENTICAL 3560) (T_IS_SMALLER_OR_EQUAL 3561) (T_LOGICAL_AND 3562) (T_LOGICAL_OR 3563) (T_LOGICAL_XOR 3564) (T_POW 3565) (T_SL 3566) (T_SPACESHIP 3567) (T_SR 3568) ("^" 3569) (backup_fn_flags 3838) ("|" 3570)) 1743 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 3840) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 1744 (("$" 3463) (T_ABSTRACT 3640) (T_ARRAY 3641) (T_AS 3642) (T_BREAK 3643) (T_CALLABLE 3644) (T_CASE 3645) (T_CATCH 3646) (T_CLASS 3647) (T_CLASS_C 3648) (T_CLONE 3649) (T_CONST 3650) (T_CONTINUE 3651) (T_DECLARE 3652) (T_DEFAULT 3653) (T_DIR 3654) (T_DO 3655) (T_ECHO 3656) (T_ELSE 3657) (T_ELSEIF 3658) (T_EMPTY 3659) (T_ENDDECLARE 3660) (T_ENDFOR 3661) (T_ENDFOREACH 3662) (T_ENDIF 3663) (T_ENDSWITCH 3664) (T_ENDWHILE 3665) (T_ENUM 3666) (T_EVAL 3667) (T_EXIT 3668) (T_EXTENDS 3669) (T_FILE 3670) (T_FINAL 3671) (T_FINALLY 3672) (T_FN 3673) (T_FOR 3674) (T_FOREACH 3675) (T_FUNCTION 3676) (T_FUNC_C 3677) (T_GLOBAL 3678) (T_GOTO 3679) (T_IF 3680) (T_IMPLEMENTS 3681) (T_INCLUDE 3682) (T_INCLUDE_ONCE 3683) (T_INSTANCEOF 3684) (T_INSTEADOF 3685) (T_INTERFACE 3686) (T_ISSET 3687) (T_LINE 3688) (T_LIST 3689) (T_LOGICAL_AND 3690) (T_LOGICAL_OR 3691) (T_LOGICAL_XOR 3692) (T_MATCH 3693) (T_METHOD_C 3694) (T_NAMESPACE 3695) (T_NEW 3696) (T_NS_C 3697) (T_PRINT 3698) (T_PRIVATE 3699) (T_PROPERTY_C 3700) (T_PROTECTED 3701) (T_PUBLIC 3702) (T_READONLY 3703) (T_REQUIRE 3704) (T_REQUIRE_ONCE 3705) (T_RETURN 3706) (T_STATIC 3707) (T_STRING 3708) (T_SWITCH 3709) (T_THROW 3710) (T_TRAIT 3711) (T_TRAIT_C 3712) (T_TRY 3713) (T_UNSET 3714) (T_USE 3715) (T_VAR 3716) (T_VARIABLE 3511) (T_WHILE 3717) (T_YIELD 3718) (identifier 3842) (member_name 3843) (reserved_non_modifiers 3721) (semi_reserved 3722) (simple_variable 3844) ("{" 3845)) 1745 (("(" 3727) (argument_list 3848)) 1746 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 3846) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1747 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 3847)) 1748 ((T_FN 35) (T_FUNCTION 129) (fn 3525) #'3526 (inline_function 3852)) 1749 (("$" 3463) (T_STRING 3858) (T_VARIABLE 3511) (property_name 3865) (simple_variable 3860) ("{" 3861)) 1750 (("$" 3463) (T_STRING 3858) (T_VARIABLE 3511) (property_name 3859) (simple_variable 3860) ("{" 3861)) 1751 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 3856) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1752 (("]" 3857)) 1753 (("(" 3727) (argument_list 3864)) 1754 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 3862) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1755 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 3863)) 1756 (("(" 3727) (argument_list 3866)) 1757 (("`" 3868)) 1758 (("]" 3870)) 1759 (("=" 3871)) 1760 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3872) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1761 (("%" 3541) ("*" 3542) ("+" 3543) ("-" 3544) ("." 3545) ("/" 3546) ("<" 3547) (">" 3548) ("?" 3549) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3550) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3551) (T_BOOLEAN_AND 3552) (T_BOOLEAN_OR 3553) (T_COALESCE 3554) (T_DOUBLE_ARROW 3875) (T_INSTANCEOF 3555) (T_IS_EQUAL 3556) (T_IS_GREATER_OR_EQUAL 3557) (T_IS_IDENTICAL 3558) (T_IS_NOT_EQUAL 3559) (T_IS_NOT_IDENTICAL 3560) (T_IS_SMALLER_OR_EQUAL 3561) (T_LOGICAL_AND 3562) (T_LOGICAL_OR 3563) (T_LOGICAL_XOR 3564) (T_POW 3565) (T_SL 3566) (T_SPACESHIP 3567) (T_SR 3568) ("^" 3569) ("|" 3570)) 1762 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 3876) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 1763 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 3885) (T_VARIABLE 2896) (encaps_var 2910)) 1764 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 3884) (T_VARIABLE 2896) (encaps_var 2901)) 1765 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 3926) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 1766 ((T_CLASS 3918)) 1767 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 3917) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 3898) (attribute 2958) (class_modifier 2939)) 1768 ((T_PAAMAYIM_NEKUDOTAYIM 3915)) 1769 (("(" 3782) (argument_list 3914)) 1770 ((T_NULLSAFE_OBJECT_OPERATOR 3905) (T_OBJECT_OPERATOR 3906) (T_PAAMAYIM_NEKUDOTAYIM 3907) ("[" 3908)) 1771 (("$" 3766) (T_STRING 3801) (T_VARIABLE 3768) (property_name 3913) (simple_variable 3803) ("{" 3804)) 1772 (("$" 3766) (T_STRING 3801) (T_VARIABLE 3768) (property_name 3912) (simple_variable 3803) ("{" 3804)) 1773 (("$" 3766) (T_VARIABLE 3768) (simple_variable 3911)) 1774 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 3909) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1775 (("]" 3910)) 1776 (("$" 3766) (T_VARIABLE 3768) (simple_variable 3916)) 1777 (("(" 2960) (argument_list 2961) (ctor_arguments 3919)) 1778 ((T_EXTENDS 2963) (extends_from 3920)) 1779 ((T_IMPLEMENTS 2965) (implements_list 3921)) 1780 ((backup_doc_comment 3922)) 1781 (("{" 3923)) 1782 ((class_statement_list 3924)) 1783 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 3925)) 1784 (("%" 299) (")" 3927) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 1785 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 3929) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 1786 (("%" 299) (")" 3930) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 1787 (("{" 3931)) 1788 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 4005) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1789 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5533) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1790 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 5531) (encaps_var 5237)) 1791 (("$" 3934) (T_VARIABLE 3983) (simple_variable 5527) ("{" 5528)) 1792 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5525) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 1793 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5524) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1794 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5523) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1795 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5522) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1796 (("(" 5519)) 1797 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5518) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1798 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5517) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1799 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5516) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1800 (("\"" 1520) ("$" 4242) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 4243) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 4244) ("[" 1542) (array_object_dereferenceable 4245) (callable_expr 4246) (callable_variable 4247) (class_constant 1546) (class_name 4248) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 4249) (name 4250) (new_dereferenceable 1552) (simple_variable 4251) (static_member 4252) (variable 5515) (variable_class_name 4254)) 1801 (("," 5511) (possible_comma 5512)) 1802 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5510) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1803 (("(" 5507)) 1804 (("(" 5504)) 1805 (("(" 5495) (argument_list 5496) (ctor_arguments 5497)) 1806 (("\"" 1520) ("$" 4242) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 4243) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 4244) ("[" 1542) (array_object_dereferenceable 4245) (callable_expr 4246) (callable_variable 4247) (class_constant 1546) (class_name 4248) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 4249) (name 4250) (new_dereferenceable 1552) (simple_variable 4251) (static_member 4252) (variable 5494) (variable_class_name 4254)) 1807 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5493) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1808 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5492) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1809 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5491) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1810 (("(" 5487)) 1811 (("(" 5482)) 1812 (("(" 5476)) 1813 (("$" 4242) ("(" 5439) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 5440) (T_NAME_QUALIFIED 5441) (T_NAME_RELATIVE 5442) (T_READONLY 2931) (T_STATIC 5443) (T_STRING 5444) (T_VARIABLE 4244) (anonymous_class 5445) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 5446) (attribute 2937) (attributes 5447) (class_modifier 2939) (class_name 5448) (class_name_reference 5449) (name 5450) (new_variable 5451) (simple_variable 5452)) 1814 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5438) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1815 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5437) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1816 (("(" 4203) (argument_list 5436)) 1817 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5435) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1818 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5434) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1819 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 5429) (T_END_HEREDOC 5430) (T_VARIABLE 2896) (encaps_list 5431) (encaps_var 2898)) 1820 ((T_FN 35) (T_FUNCTION 129) (fn 3997) #'3998 (inline_function 5428)) 1821 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5427) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1822 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5426) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1823 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5425) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1824 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5422) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1825 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5421) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1826 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 5417) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 1827 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 5415) (encaps_list 2804) (encaps_var 2805)) 1828 ((T_NULLSAFE_OBJECT_OPERATOR 5401) (T_OBJECT_OPERATOR 5402) ("[" 5403)) 1829 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 5398) (attribute 2733) (fn 3997) #'3998 (inline_function 5399)) 1830 (("(" 4203) (argument_list 5397)) 1831 ((T_PAAMAYIM_NEKUDOTAYIM 5389)) 1832 (("%" 4017) ("*" 4018) ("+" 4019) ("-" 4020) ("." 4021) ("/" 4022) ("<" 4023) (">" 4024) ("?" 4025) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4026) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4027) (T_BOOLEAN_AND 4028) (T_BOOLEAN_OR 4029) (T_COALESCE 4030) (T_INSTANCEOF 4031) (T_IS_EQUAL 4032) (T_IS_GREATER_OR_EQUAL 4033) (T_IS_IDENTICAL 4034) (T_IS_NOT_EQUAL 4035) (T_IS_NOT_IDENTICAL 4036) (T_IS_SMALLER_OR_EQUAL 4037) (T_LOGICAL_AND 4038) (T_LOGICAL_OR 4039) (T_LOGICAL_XOR 4040) (T_POW 4041) (T_SL 4042) (T_SPACESHIP 4043) (T_SR 4044) ("^" 4045) ("|" 4046)) 1833 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 5378)) 1834 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 5366)) 1835 (("," 4296) (possible_comma 4297)) 1836 (("}" 4295)) 1837 (("(" 4203) (argument_list 4294)) 1838 (("," 4291) (possible_comma 4292)) 1839 (("=" 4211) (T_AND_EQUAL 4212) (T_COALESCE_EQUAL 4213) (T_CONCAT_EQUAL 4214) (T_DEC 4215) (T_DIV_EQUAL 4216) (T_INC 4217) (T_MINUS_EQUAL 4218) (T_MOD_EQUAL 4219) (T_MUL_EQUAL 4220) (T_OR_EQUAL 4221) (T_PLUS_EQUAL 4222) (T_POW_EQUAL 4223) (T_SL_EQUAL 4224) (T_SR_EQUAL 4225) (T_XOR_EQUAL 4226)) 1840 ((T_PAAMAYIM_NEKUDOTAYIM 4115)) 1841 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4016) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1842 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4114) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1843 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4113) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1844 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4112) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1845 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4111) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1846 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4110) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1847 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4109) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1848 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4108) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1849 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4107) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1850 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 4102) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 4103) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 1851 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4101) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1852 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4100) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1853 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4099) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1854 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4098) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1855 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4097) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1856 (("$" 4062) ("(" 4063) (T_NAME_FULLY_QUALIFIED 4064) (T_NAME_QUALIFIED 4065) (T_NAME_RELATIVE 4066) (T_STATIC 4067) (T_STRING 4068) (T_VARIABLE 4069) (class_name 4070) (class_name_reference 4071) (name 4072) (new_variable 4073) (simple_variable 4074)) 1857 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4061) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1858 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4060) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1859 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4059) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1860 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4058) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1861 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4057) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1862 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4056) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1863 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4055) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1864 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4054) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1865 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4053) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1866 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4052) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1867 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4051) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1868 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4050) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1869 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4049) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1870 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4048) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1871 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4047) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1872 (("$" 4062) (T_VARIABLE 4069) (simple_variable 4093) ("{" 4094)) 1873 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 4091) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 1874 ((T_PAAMAYIM_NEKUDOTAYIM 4089)) 1875 ((T_NULLSAFE_OBJECT_OPERATOR 4075) (T_OBJECT_OPERATOR 4076) (T_PAAMAYIM_NEKUDOTAYIM 4077) ("[" 4078)) 1876 (("$" 4062) (T_STRING 4082) (T_VARIABLE 4069) (property_name 4088) (simple_variable 4084) ("{" 4085)) 1877 (("$" 4062) (T_STRING 4082) (T_VARIABLE 4069) (property_name 4083) (simple_variable 4084) ("{" 4085)) 1878 (("$" 4062) (T_VARIABLE 4069) (simple_variable 4081)) 1879 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 4079) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1880 (("]" 4080)) 1881 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 4086) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1882 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 4087)) 1883 (("$" 4062) (T_VARIABLE 4069) (simple_variable 4090)) 1884 (("%" 299) (")" 4092) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 1885 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 4095) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1886 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 4096)) 1887 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4106) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1888 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 4104) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 1889 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4105) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1890 (("$" 3934) (T_ABSTRACT 4116) (T_ARRAY 4117) (T_AS 4118) (T_BREAK 4119) (T_CALLABLE 4120) (T_CASE 4121) (T_CATCH 4122) (T_CLASS 4123) (T_CLASS_C 4124) (T_CLONE 4125) (T_CONST 4126) (T_CONTINUE 4127) (T_DECLARE 4128) (T_DEFAULT 4129) (T_DIR 4130) (T_DO 4131) (T_ECHO 4132) (T_ELSE 4133) (T_ELSEIF 4134) (T_EMPTY 4135) (T_ENDDECLARE 4136) (T_ENDFOR 4137) (T_ENDFOREACH 4138) (T_ENDIF 4139) (T_ENDSWITCH 4140) (T_ENDWHILE 4141) (T_ENUM 4142) (T_EVAL 4143) (T_EXIT 4144) (T_EXTENDS 4145) (T_FILE 4146) (T_FINAL 4147) (T_FINALLY 4148) (T_FN 4149) (T_FOR 4150) (T_FOREACH 4151) (T_FUNCTION 4152) (T_FUNC_C 4153) (T_GLOBAL 4154) (T_GOTO 4155) (T_IF 4156) (T_IMPLEMENTS 4157) (T_INCLUDE 4158) (T_INCLUDE_ONCE 4159) (T_INSTANCEOF 4160) (T_INSTEADOF 4161) (T_INTERFACE 4162) (T_ISSET 4163) (T_LINE 4164) (T_LIST 4165) (T_LOGICAL_AND 4166) (T_LOGICAL_OR 4167) (T_LOGICAL_XOR 4168) (T_MATCH 4169) (T_METHOD_C 4170) (T_NAMESPACE 4171) (T_NEW 4172) (T_NS_C 4173) (T_PRINT 4174) (T_PRIVATE 4175) (T_PROPERTY_C 4176) (T_PROTECTED 4177) (T_PUBLIC 4178) (T_READONLY 4179) (T_REQUIRE 4180) (T_REQUIRE_ONCE 4181) (T_RETURN 4182) (T_STATIC 4183) (T_STRING 4184) (T_SWITCH 4185) (T_THROW 4186) (T_TRAIT 4187) (T_TRAIT_C 4188) (T_TRY 4189) (T_UNSET 4190) (T_USE 4191) (T_VAR 4192) (T_VARIABLE 3983) (T_WHILE 4193) (T_YIELD 4194) (identifier 4195) (member_name 4196) (reserved_non_modifiers 4197) (semi_reserved 4198) (simple_variable 4199) ("{" 4200)) 1891 (("(" 4203) (argument_list 4204)) 1892 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 4201) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1893 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 4202)) 1894 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 4205) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 4206) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 4207) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1895 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 4210) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1896 (("," 1480) (possible_comma 4208)) 1897 ((")" 4209)) 1898 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (ampersand 4240) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4241) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1899 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4239) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1900 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4238) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1901 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4237) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1902 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4236) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1903 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4235) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1904 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4234) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1905 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4233) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1906 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4232) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1907 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4231) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1908 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4230) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1909 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4229) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1910 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4228) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1911 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 4227) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1912 (("\"" 1520) ("$" 4242) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 4243) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 4244) ("[" 1542) (array_object_dereferenceable 4245) (callable_expr 4246) (callable_variable 4247) (class_constant 1546) (class_name 4248) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 4249) (name 4250) (new_dereferenceable 1552) (simple_variable 4251) (static_member 4252) (variable 4253) (variable_class_name 4254)) 1913 (("$" 4242) (T_VARIABLE 4244) (simple_variable 4287) ("{" 4288)) 1914 (("(" 4258) (argument_list 4286)) 1915 ((T_NULLSAFE_OBJECT_OPERATOR 4272) (T_OBJECT_OPERATOR 4273) ("[" 4274)) 1916 (("(" 4258) (argument_list 4271)) 1917 ((T_PAAMAYIM_NEKUDOTAYIM 4267)) 1918 (("(" 4258) (argument_list 4266)) 1919 ((T_PAAMAYIM_NEKUDOTAYIM 4255)) 1920 (("$" 4242) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 4244) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 4256) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 4257) ("{" 1642)) 1921 (("(" 4258) (argument_list 4259)) 1922 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 4260) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 4261) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 4262) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1923 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 4265) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 1924 (("," 1480) (possible_comma 4263)) 1925 ((")" 4264)) 1926 (("$" 4242) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 4244) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 4268) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 4269) ("{" 1658)) 1927 (("(" 4258) (argument_list 4270)) 1928 (("$" 4242) (T_STRING 4277) (T_VARIABLE 4244) (property_name 4284) (simple_variable 4279) ("{" 4280)) 1929 (("$" 4242) (T_STRING 4277) (T_VARIABLE 4244) (property_name 4278) (simple_variable 4279) ("{" 4280)) 1930 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 4275) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 1931 (("]" 4276)) 1932 (("(" 4258) (argument_list 4283)) 1933 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 4281) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1934 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 4282)) 1935 (("(" 4258) (argument_list 4285)) 1936 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 4289) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 1937 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 4290)) 1938 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4293) (match_arm_cond_list 4004) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1939 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5365) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 1940 ((T_DOUBLE_ARROW 4298)) 1941 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4362) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1942 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5364) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1943 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 5362) (encaps_var 5237)) 1944 (("$" 4301) (T_VARIABLE 4349) (simple_variable 5358) ("{" 5359)) 1945 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5356) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 1946 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5355) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1947 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5354) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1948 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5353) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1949 (("(" 5350)) 1950 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5349) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1951 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5348) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1952 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5347) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1953 (("\"" 1520) ("$" 4604) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 4605) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 4606) ("[" 1542) (array_object_dereferenceable 4607) (callable_expr 4608) (callable_variable 4609) (class_constant 1546) (class_name 4610) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 4611) (name 4612) (new_dereferenceable 1552) (simple_variable 4613) (static_member 4614) (variable 5346) (variable_class_name 4616)) 1954 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5345) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1955 (("(" 5342)) 1956 (("(" 5339)) 1957 (("(" 5330) (argument_list 5331) (ctor_arguments 5332)) 1958 (("\"" 1520) ("$" 4604) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 4605) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 4606) ("[" 1542) (array_object_dereferenceable 4607) (callable_expr 4608) (callable_variable 4609) (class_constant 1546) (class_name 4610) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 4611) (name 4612) (new_dereferenceable 1552) (simple_variable 4613) (static_member 4614) (variable 5329) (variable_class_name 4616)) 1959 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5328) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1960 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5327) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1961 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5326) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1962 (("(" 5322)) 1963 (("(" 5317)) 1964 (("(" 5311)) 1965 (("$" 4604) ("(" 5274) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 5275) (T_NAME_QUALIFIED 5276) (T_NAME_RELATIVE 5277) (T_READONLY 2931) (T_STATIC 5278) (T_STRING 5279) (T_VARIABLE 4606) (anonymous_class 5280) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 5281) (attribute 2937) (attributes 5282) (class_modifier 2939) (class_name 5283) (class_name_reference 5284) (name 5285) (new_variable 5286) (simple_variable 5287)) 1966 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5273) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1967 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5272) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1968 (("(" 4565) (argument_list 5271)) 1969 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5270) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1970 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5269) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1971 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 5264) (T_END_HEREDOC 5265) (T_VARIABLE 2896) (encaps_list 5266) (encaps_var 2898)) 1972 ((T_FN 35) (T_FUNCTION 129) (fn 4363) #'4364 (inline_function 5263)) 1973 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5262) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1974 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5261) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1975 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5260) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1976 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4773) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 1977 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4709) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1978 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 4705) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 1979 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 4703) (encaps_list 2804) (encaps_var 2805)) 1980 ((T_NULLSAFE_OBJECT_OPERATOR 4689) (T_OBJECT_OPERATOR 4690) ("[" 4691)) 1981 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 4686) (attribute 2733) (fn 4363) #'4364 (inline_function 4687)) 1982 (("(" 4565) (argument_list 4685)) 1983 ((T_PAAMAYIM_NEKUDOTAYIM 4677)) 1984 (("%" 4379) ("*" 4380) ("+" 4381) ("-" 4382) ("." 4383) ("/" 4384) ("<" 4385) (">" 4386) ("?" 4387) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4388) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4389) (T_BOOLEAN_AND 4390) (T_BOOLEAN_OR 4391) (T_COALESCE 4392) (T_INSTANCEOF 4393) (T_IS_EQUAL 4394) (T_IS_GREATER_OR_EQUAL 4395) (T_IS_IDENTICAL 4396) (T_IS_NOT_EQUAL 4397) (T_IS_NOT_IDENTICAL 4398) (T_IS_SMALLER_OR_EQUAL 4399) (T_LOGICAL_AND 4400) (T_LOGICAL_OR 4401) (T_LOGICAL_XOR 4402) (T_POW 4403) (T_SL 4404) (T_SPACESHIP 4405) (T_SR 4406) ("^" 4407) ("|" 4408)) 1985 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 4666)) 1986 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 4654)) 1987 (("(" 4565) (argument_list 4653)) 1988 (("=" 4573) (T_AND_EQUAL 4574) (T_COALESCE_EQUAL 4575) (T_CONCAT_EQUAL 4576) (T_DEC 4577) (T_DIV_EQUAL 4578) (T_INC 4579) (T_MINUS_EQUAL 4580) (T_MOD_EQUAL 4581) (T_MUL_EQUAL 4582) (T_OR_EQUAL 4583) (T_PLUS_EQUAL 4584) (T_POW_EQUAL 4585) (T_SL_EQUAL 4586) (T_SR_EQUAL 4587) (T_XOR_EQUAL 4588)) 1989 ((T_PAAMAYIM_NEKUDOTAYIM 4477)) 1990 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4378) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1991 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4476) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1992 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4475) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1993 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4474) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1994 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4473) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1995 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4472) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1996 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4471) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1997 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4470) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1998 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4469) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 1999 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 4464) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 4465) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 2000 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4463) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2001 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4462) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2002 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4461) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2003 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4460) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2004 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4459) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2005 (("$" 4424) ("(" 4425) (T_NAME_FULLY_QUALIFIED 4426) (T_NAME_QUALIFIED 4427) (T_NAME_RELATIVE 4428) (T_STATIC 4429) (T_STRING 4430) (T_VARIABLE 4431) (class_name 4432) (class_name_reference 4433) (name 4434) (new_variable 4435) (simple_variable 4436)) 2006 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4423) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2007 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4422) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2008 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4421) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2009 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4420) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2010 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4419) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2011 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4418) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2012 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4417) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2013 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4416) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2014 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4415) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2015 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4414) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2016 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4413) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2017 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4412) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2018 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4411) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2019 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4410) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2020 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4409) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2021 (("$" 4424) (T_VARIABLE 4431) (simple_variable 4455) ("{" 4456)) 2022 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 4453) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2023 ((T_PAAMAYIM_NEKUDOTAYIM 4451)) 2024 ((T_NULLSAFE_OBJECT_OPERATOR 4437) (T_OBJECT_OPERATOR 4438) (T_PAAMAYIM_NEKUDOTAYIM 4439) ("[" 4440)) 2025 (("$" 4424) (T_STRING 4444) (T_VARIABLE 4431) (property_name 4450) (simple_variable 4446) ("{" 4447)) 2026 (("$" 4424) (T_STRING 4444) (T_VARIABLE 4431) (property_name 4445) (simple_variable 4446) ("{" 4447)) 2027 (("$" 4424) (T_VARIABLE 4431) (simple_variable 4443)) 2028 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 4441) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2029 (("]" 4442)) 2030 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 4448) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2031 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 4449)) 2032 (("$" 4424) (T_VARIABLE 4431) (simple_variable 4452)) 2033 (("%" 299) (")" 4454) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2034 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 4457) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2035 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 4458)) 2036 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4468) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2037 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 4466) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 2038 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4467) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2039 (("$" 4301) (T_ABSTRACT 4478) (T_ARRAY 4479) (T_AS 4480) (T_BREAK 4481) (T_CALLABLE 4482) (T_CASE 4483) (T_CATCH 4484) (T_CLASS 4485) (T_CLASS_C 4486) (T_CLONE 4487) (T_CONST 4488) (T_CONTINUE 4489) (T_DECLARE 4490) (T_DEFAULT 4491) (T_DIR 4492) (T_DO 4493) (T_ECHO 4494) (T_ELSE 4495) (T_ELSEIF 4496) (T_EMPTY 4497) (T_ENDDECLARE 4498) (T_ENDFOR 4499) (T_ENDFOREACH 4500) (T_ENDIF 4501) (T_ENDSWITCH 4502) (T_ENDWHILE 4503) (T_ENUM 4504) (T_EVAL 4505) (T_EXIT 4506) (T_EXTENDS 4507) (T_FILE 4508) (T_FINAL 4509) (T_FINALLY 4510) (T_FN 4511) (T_FOR 4512) (T_FOREACH 4513) (T_FUNCTION 4514) (T_FUNC_C 4515) (T_GLOBAL 4516) (T_GOTO 4517) (T_IF 4518) (T_IMPLEMENTS 4519) (T_INCLUDE 4520) (T_INCLUDE_ONCE 4521) (T_INSTANCEOF 4522) (T_INSTEADOF 4523) (T_INTERFACE 4524) (T_ISSET 4525) (T_LINE 4526) (T_LIST 4527) (T_LOGICAL_AND 4528) (T_LOGICAL_OR 4529) (T_LOGICAL_XOR 4530) (T_MATCH 4531) (T_METHOD_C 4532) (T_NAMESPACE 4533) (T_NEW 4534) (T_NS_C 4535) (T_PRINT 4536) (T_PRIVATE 4537) (T_PROPERTY_C 4538) (T_PROTECTED 4539) (T_PUBLIC 4540) (T_READONLY 4541) (T_REQUIRE 4542) (T_REQUIRE_ONCE 4543) (T_RETURN 4544) (T_STATIC 4545) (T_STRING 4546) (T_SWITCH 4547) (T_THROW 4548) (T_TRAIT 4549) (T_TRAIT_C 4550) (T_TRY 4551) (T_UNSET 4552) (T_USE 4553) (T_VAR 4554) (T_VARIABLE 4349) (T_WHILE 4555) (T_YIELD 4556) (identifier 4557) (member_name 4558) (reserved_non_modifiers 4559) (semi_reserved 4560) (simple_variable 4561) ("{" 4562)) 2040 (("(" 4565) (argument_list 4566)) 2041 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 4563) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2042 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 4564)) 2043 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 4567) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 4568) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 4569) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2044 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 4572) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2045 (("," 1480) (possible_comma 4570)) 2046 ((")" 4571)) 2047 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (ampersand 4602) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4603) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2048 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4601) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2049 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4600) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2050 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4599) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2051 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4598) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2052 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4597) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2053 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4596) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2054 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4595) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2055 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4594) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2056 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4593) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2057 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4592) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2058 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4591) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2059 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4590) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2060 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4589) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2061 (("\"" 1520) ("$" 4604) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 4605) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 4606) ("[" 1542) (array_object_dereferenceable 4607) (callable_expr 4608) (callable_variable 4609) (class_constant 1546) (class_name 4610) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 4611) (name 4612) (new_dereferenceable 1552) (simple_variable 4613) (static_member 4614) (variable 4615) (variable_class_name 4616)) 2062 (("$" 4604) (T_VARIABLE 4606) (simple_variable 4649) ("{" 4650)) 2063 (("(" 4620) (argument_list 4648)) 2064 ((T_NULLSAFE_OBJECT_OPERATOR 4634) (T_OBJECT_OPERATOR 4635) ("[" 4636)) 2065 (("(" 4620) (argument_list 4633)) 2066 ((T_PAAMAYIM_NEKUDOTAYIM 4629)) 2067 (("(" 4620) (argument_list 4628)) 2068 ((T_PAAMAYIM_NEKUDOTAYIM 4617)) 2069 (("$" 4604) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 4606) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 4618) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 4619) ("{" 1642)) 2070 (("(" 4620) (argument_list 4621)) 2071 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 4622) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 4623) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 4624) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2072 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 4627) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2073 (("," 1480) (possible_comma 4625)) 2074 ((")" 4626)) 2075 (("$" 4604) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 4606) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 4630) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 4631) ("{" 1658)) 2076 (("(" 4620) (argument_list 4632)) 2077 (("$" 4604) (T_STRING 4639) (T_VARIABLE 4606) (property_name 4646) (simple_variable 4641) ("{" 4642)) 2078 (("$" 4604) (T_STRING 4639) (T_VARIABLE 4606) (property_name 4640) (simple_variable 4641) ("{" 4642)) 2079 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 4637) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2080 (("]" 4638)) 2081 (("(" 4620) (argument_list 4645)) 2082 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 4643) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2083 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 4644)) 2084 (("(" 4620) (argument_list 4647)) 2085 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 4651) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2086 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 4652)) 2087 ((backup_doc_comment 4655)) 2088 (("(" 4656)) 2089 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 4657)) 2090 ((")" 4658)) 2091 ((T_USE 164) (lexical_vars 4659)) 2092 ((":" 166) (return_type 4660)) 2093 ((backup_fn_flags 4661)) 2094 (("{" 4662)) 2095 ((inner_statement_list 4663)) 2096 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 4664) ("~" 127)) 2097 ((backup_fn_flags 4665)) 2098 ((backup_doc_comment 4667)) 2099 (("(" 4668)) 2100 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 4669)) 2101 ((")" 4670)) 2102 ((":" 2249) (return_type 4671)) 2103 ((T_DOUBLE_ARROW 4672)) 2104 ((backup_fn_flags 4673)) 2105 ((backup_lex_pos 4674)) 2106 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4675) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2107 (("%" 4379) ("*" 4380) ("+" 4381) ("-" 4382) ("." 4383) ("/" 4384) ("<" 4385) (">" 4386) ("?" 4387) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4388) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4389) (T_BOOLEAN_AND 4390) (T_BOOLEAN_OR 4391) (T_COALESCE 4392) (T_INSTANCEOF 4393) (T_IS_EQUAL 4394) (T_IS_GREATER_OR_EQUAL 4395) (T_IS_IDENTICAL 4396) (T_IS_NOT_EQUAL 4397) (T_IS_NOT_IDENTICAL 4398) (T_IS_SMALLER_OR_EQUAL 4399) (T_LOGICAL_AND 4400) (T_LOGICAL_OR 4401) (T_LOGICAL_XOR 4402) (T_POW 4403) (T_SL 4404) (T_SPACESHIP 4405) (T_SR 4406) ("^" 4407) (backup_fn_flags 4676) ("|" 4408)) 2108 (("$" 4301) (T_ABSTRACT 4478) (T_ARRAY 4479) (T_AS 4480) (T_BREAK 4481) (T_CALLABLE 4482) (T_CASE 4483) (T_CATCH 4484) (T_CLASS 4485) (T_CLASS_C 4486) (T_CLONE 4487) (T_CONST 4488) (T_CONTINUE 4489) (T_DECLARE 4490) (T_DEFAULT 4491) (T_DIR 4492) (T_DO 4493) (T_ECHO 4494) (T_ELSE 4495) (T_ELSEIF 4496) (T_EMPTY 4497) (T_ENDDECLARE 4498) (T_ENDFOR 4499) (T_ENDFOREACH 4500) (T_ENDIF 4501) (T_ENDSWITCH 4502) (T_ENDWHILE 4503) (T_ENUM 4504) (T_EVAL 4505) (T_EXIT 4506) (T_EXTENDS 4507) (T_FILE 4508) (T_FINAL 4509) (T_FINALLY 4510) (T_FN 4511) (T_FOR 4512) (T_FOREACH 4513) (T_FUNCTION 4514) (T_FUNC_C 4515) (T_GLOBAL 4516) (T_GOTO 4517) (T_IF 4518) (T_IMPLEMENTS 4519) (T_INCLUDE 4520) (T_INCLUDE_ONCE 4521) (T_INSTANCEOF 4522) (T_INSTEADOF 4523) (T_INTERFACE 4524) (T_ISSET 4525) (T_LINE 4526) (T_LIST 4527) (T_LOGICAL_AND 4528) (T_LOGICAL_OR 4529) (T_LOGICAL_XOR 4530) (T_MATCH 4531) (T_METHOD_C 4532) (T_NAMESPACE 4533) (T_NEW 4534) (T_NS_C 4535) (T_PRINT 4536) (T_PRIVATE 4537) (T_PROPERTY_C 4538) (T_PROTECTED 4539) (T_PUBLIC 4540) (T_READONLY 4541) (T_REQUIRE 4542) (T_REQUIRE_ONCE 4543) (T_RETURN 4544) (T_STATIC 4545) (T_STRING 4546) (T_SWITCH 4547) (T_THROW 4548) (T_TRAIT 4549) (T_TRAIT_C 4550) (T_TRY 4551) (T_UNSET 4552) (T_USE 4553) (T_VAR 4554) (T_VARIABLE 4349) (T_WHILE 4555) (T_YIELD 4556) (identifier 4678) (member_name 4679) (reserved_non_modifiers 4559) (semi_reserved 4560) (simple_variable 4680) ("{" 4681)) 2109 (("(" 4565) (argument_list 4684)) 2110 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 4682) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2111 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 4683)) 2112 ((T_FN 35) (T_FUNCTION 129) (fn 4363) #'4364 (inline_function 4688)) 2113 (("$" 4301) (T_STRING 4694) (T_VARIABLE 4349) (property_name 4701) (simple_variable 4696) ("{" 4697)) 2114 (("$" 4301) (T_STRING 4694) (T_VARIABLE 4349) (property_name 4695) (simple_variable 4696) ("{" 4697)) 2115 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 4692) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2116 (("]" 4693)) 2117 (("(" 4565) (argument_list 4700)) 2118 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 4698) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2119 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 4699)) 2120 (("(" 4565) (argument_list 4702)) 2121 (("`" 4704)) 2122 (("]" 4706)) 2123 (("=" 4707)) 2124 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 4708) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2125 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5259) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2126 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 5236) (encaps_var 5237)) 2127 (("$" 4712) (T_VARIABLE 4760) (simple_variable 5228) ("{" 5229)) 2128 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5226) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2129 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5225) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2130 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5224) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2131 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5223) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2132 (("(" 5220)) 2133 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5219) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2134 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5218) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2135 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5217) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2136 (("\"" 1520) ("$" 5015) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 5016) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 5017) ("[" 1542) (array_object_dereferenceable 5018) (callable_expr 5019) (callable_variable 5020) (class_constant 1546) (class_name 5021) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 5022) (name 5023) (new_dereferenceable 1552) (simple_variable 5024) (static_member 5025) (variable 5216) (variable_class_name 5027)) 2137 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5215) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2138 (("(" 5212)) 2139 (("(" 5209)) 2140 (("(" 5200) (argument_list 5201) (ctor_arguments 5202)) 2141 (("\"" 1520) ("$" 5015) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 5016) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 5017) ("[" 1542) (array_object_dereferenceable 5018) (callable_expr 5019) (callable_variable 5020) (class_constant 1546) (class_name 5021) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 5022) (name 5023) (new_dereferenceable 1552) (simple_variable 5024) (static_member 5025) (variable 5199) (variable_class_name 5027)) 2142 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5198) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2143 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5197) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2144 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5196) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2145 (("(" 5188)) 2146 (("(" 5183)) 2147 (("(" 5177)) 2148 (("$" 5015) ("(" 5140) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 5141) (T_NAME_QUALIFIED 5142) (T_NAME_RELATIVE 5143) (T_READONLY 2931) (T_STATIC 5144) (T_STRING 5145) (T_VARIABLE 5017) (anonymous_class 5146) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 5147) (attribute 2937) (attributes 5148) (class_modifier 2939) (class_name 5149) (class_name_reference 5150) (name 5151) (new_variable 5152) (simple_variable 5153)) 2149 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5139) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2150 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5138) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2151 (("(" 4976) (argument_list 5137)) 2152 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5136) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2153 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5135) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2154 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 5130) (T_END_HEREDOC 5131) (T_VARIABLE 2896) (encaps_list 5132) (encaps_var 2898)) 2155 ((T_FN 35) (T_FUNCTION 129) (fn 4774) #'4775 (inline_function 5129)) 2156 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5128) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2157 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5127) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2158 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5126) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2159 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5123) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2160 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5122) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2161 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 5118) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 2162 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 5116) (encaps_list 2804) (encaps_var 2805)) 2163 ((T_NULLSAFE_OBJECT_OPERATOR 5102) (T_OBJECT_OPERATOR 5103) ("[" 5104)) 2164 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 5099) (attribute 2733) (fn 4774) #'4775 (inline_function 5100)) 2165 (("(" 4976) (argument_list 5098)) 2166 ((T_PAAMAYIM_NEKUDOTAYIM 5090)) 2167 (("%" 4790) ("*" 4791) ("+" 4792) ("-" 4793) ("." 4794) ("/" 4795) ("<" 4796) (">" 4797) ("?" 4798) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4799) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4800) (T_BOOLEAN_AND 4801) (T_BOOLEAN_OR 4802) (T_COALESCE 4803) (T_DOUBLE_ARROW 5088) (T_INSTANCEOF 4804) (T_IS_EQUAL 4805) (T_IS_GREATER_OR_EQUAL 4806) (T_IS_IDENTICAL 4807) (T_IS_NOT_EQUAL 4808) (T_IS_NOT_IDENTICAL 4809) (T_IS_SMALLER_OR_EQUAL 4810) (T_LOGICAL_AND 4811) (T_LOGICAL_OR 4812) (T_LOGICAL_XOR 4813) (T_POW 4814) (T_SL 4815) (T_SPACESHIP 4816) (T_SR 4817) ("^" 4818) ("|" 4819)) 2168 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 5077)) 2169 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 5065)) 2170 (("(" 4976) (argument_list 5064)) 2171 (("=" 4984) (T_AND_EQUAL 4985) (T_COALESCE_EQUAL 4986) (T_CONCAT_EQUAL 4987) (T_DEC 4988) (T_DIV_EQUAL 4989) (T_INC 4990) (T_MINUS_EQUAL 4991) (T_MOD_EQUAL 4992) (T_MUL_EQUAL 4993) (T_OR_EQUAL 4994) (T_PLUS_EQUAL 4995) (T_POW_EQUAL 4996) (T_SL_EQUAL 4997) (T_SR_EQUAL 4998) (T_XOR_EQUAL 4999)) 2172 ((T_PAAMAYIM_NEKUDOTAYIM 4888)) 2173 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4789) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2174 (("%" 4790) ("*" 4791) ("+" 4792) ("-" 4793) ("." 4794) ("/" 4795) ("<" 4796) (">" 4797) ("?" 4798) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4799) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4800) (T_BOOLEAN_AND 4801) (T_BOOLEAN_OR 4802) (T_COALESCE 4803) (T_INSTANCEOF 4804) (T_IS_EQUAL 4805) (T_IS_GREATER_OR_EQUAL 4806) (T_IS_IDENTICAL 4807) (T_IS_NOT_EQUAL 4808) (T_IS_NOT_IDENTICAL 4809) (T_IS_SMALLER_OR_EQUAL 4810) (T_LOGICAL_AND 4811) (T_LOGICAL_OR 4812) (T_LOGICAL_XOR 4813) (T_POW 4814) (T_SL 4815) (T_SPACESHIP 4816) (T_SR 4817) ("^" 4818) ("|" 4819)) 2175 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4887) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2176 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4886) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2177 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4885) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2178 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4884) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2179 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4883) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2180 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4882) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2181 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4881) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2182 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4880) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2183 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 4875) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 4876) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 2184 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4874) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2185 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4873) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2186 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4872) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2187 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4871) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2188 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4870) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2189 (("$" 4835) ("(" 4836) (T_NAME_FULLY_QUALIFIED 4837) (T_NAME_QUALIFIED 4838) (T_NAME_RELATIVE 4839) (T_STATIC 4840) (T_STRING 4841) (T_VARIABLE 4842) (class_name 4843) (class_name_reference 4844) (name 4845) (new_variable 4846) (simple_variable 4847)) 2190 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4834) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2191 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4833) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2192 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4832) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2193 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4831) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2194 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4830) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2195 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4829) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2196 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4828) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2197 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4827) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2198 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4826) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2199 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4825) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2200 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4824) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2201 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4823) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2202 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4822) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2203 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4821) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2204 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4820) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2205 (("$" 4835) (T_VARIABLE 4842) (simple_variable 4866) ("{" 4867)) 2206 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 4864) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2207 ((T_PAAMAYIM_NEKUDOTAYIM 4862)) 2208 ((T_NULLSAFE_OBJECT_OPERATOR 4848) (T_OBJECT_OPERATOR 4849) (T_PAAMAYIM_NEKUDOTAYIM 4850) ("[" 4851)) 2209 (("$" 4835) (T_STRING 4855) (T_VARIABLE 4842) (property_name 4861) (simple_variable 4857) ("{" 4858)) 2210 (("$" 4835) (T_STRING 4855) (T_VARIABLE 4842) (property_name 4856) (simple_variable 4857) ("{" 4858)) 2211 (("$" 4835) (T_VARIABLE 4842) (simple_variable 4854)) 2212 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 4852) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2213 (("]" 4853)) 2214 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 4859) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2215 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 4860)) 2216 (("$" 4835) (T_VARIABLE 4842) (simple_variable 4863)) 2217 (("%" 299) (")" 4865) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2218 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 4868) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2219 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 4869)) 2220 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4879) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2221 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 4877) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 2222 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 4878) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2223 (("$" 4712) (T_ABSTRACT 4889) (T_ARRAY 4890) (T_AS 4891) (T_BREAK 4892) (T_CALLABLE 4893) (T_CASE 4894) (T_CATCH 4895) (T_CLASS 4896) (T_CLASS_C 4897) (T_CLONE 4898) (T_CONST 4899) (T_CONTINUE 4900) (T_DECLARE 4901) (T_DEFAULT 4902) (T_DIR 4903) (T_DO 4904) (T_ECHO 4905) (T_ELSE 4906) (T_ELSEIF 4907) (T_EMPTY 4908) (T_ENDDECLARE 4909) (T_ENDFOR 4910) (T_ENDFOREACH 4911) (T_ENDIF 4912) (T_ENDSWITCH 4913) (T_ENDWHILE 4914) (T_ENUM 4915) (T_EVAL 4916) (T_EXIT 4917) (T_EXTENDS 4918) (T_FILE 4919) (T_FINAL 4920) (T_FINALLY 4921) (T_FN 4922) (T_FOR 4923) (T_FOREACH 4924) (T_FUNCTION 4925) (T_FUNC_C 4926) (T_GLOBAL 4927) (T_GOTO 4928) (T_IF 4929) (T_IMPLEMENTS 4930) (T_INCLUDE 4931) (T_INCLUDE_ONCE 4932) (T_INSTANCEOF 4933) (T_INSTEADOF 4934) (T_INTERFACE 4935) (T_ISSET 4936) (T_LINE 4937) (T_LIST 4938) (T_LOGICAL_AND 4939) (T_LOGICAL_OR 4940) (T_LOGICAL_XOR 4941) (T_MATCH 4942) (T_METHOD_C 4943) (T_NAMESPACE 4944) (T_NEW 4945) (T_NS_C 4946) (T_PRINT 4947) (T_PRIVATE 4948) (T_PROPERTY_C 4949) (T_PROTECTED 4950) (T_PUBLIC 4951) (T_READONLY 4952) (T_REQUIRE 4953) (T_REQUIRE_ONCE 4954) (T_RETURN 4955) (T_STATIC 4956) (T_STRING 4957) (T_SWITCH 4958) (T_THROW 4959) (T_TRAIT 4960) (T_TRAIT_C 4961) (T_TRY 4962) (T_UNSET 4963) (T_USE 4964) (T_VAR 4965) (T_VARIABLE 4760) (T_WHILE 4966) (T_YIELD 4967) (identifier 4968) (member_name 4969) (reserved_non_modifiers 4970) (semi_reserved 4971) (simple_variable 4972) ("{" 4973)) 2224 (("(" 4976) (argument_list 4977)) 2225 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 4974) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2226 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 4975)) 2227 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 4978) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 4979) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 4980) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2228 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 4983) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2229 (("," 1480) (possible_comma 4981)) 2230 ((")" 4982)) 2231 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (ampersand 5013) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5014) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2232 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5012) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2233 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5011) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2234 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5010) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2235 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5009) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2236 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5008) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2237 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5007) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2238 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5006) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2239 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5005) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2240 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5004) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2241 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5003) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2242 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5002) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2243 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5001) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2244 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5000) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2245 (("\"" 1520) ("$" 5015) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 5016) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 5017) ("[" 1542) (array_object_dereferenceable 5018) (callable_expr 5019) (callable_variable 5020) (class_constant 1546) (class_name 5021) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 5022) (name 5023) (new_dereferenceable 1552) (simple_variable 5024) (static_member 5025) (variable 5026) (variable_class_name 5027)) 2246 (("$" 5015) (T_VARIABLE 5017) (simple_variable 5060) ("{" 5061)) 2247 (("(" 5031) (argument_list 5059)) 2248 ((T_NULLSAFE_OBJECT_OPERATOR 5045) (T_OBJECT_OPERATOR 5046) ("[" 5047)) 2249 (("(" 5031) (argument_list 5044)) 2250 ((T_PAAMAYIM_NEKUDOTAYIM 5040)) 2251 (("(" 5031) (argument_list 5039)) 2252 ((T_PAAMAYIM_NEKUDOTAYIM 5028)) 2253 (("$" 5015) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 5017) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 5029) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 5030) ("{" 1642)) 2254 (("(" 5031) (argument_list 5032)) 2255 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 5033) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 5034) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 5035) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2256 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 5038) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2257 (("," 1480) (possible_comma 5036)) 2258 ((")" 5037)) 2259 (("$" 5015) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 5017) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 5041) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 5042) ("{" 1658)) 2260 (("(" 5031) (argument_list 5043)) 2261 (("$" 5015) (T_STRING 5050) (T_VARIABLE 5017) (property_name 5057) (simple_variable 5052) ("{" 5053)) 2262 (("$" 5015) (T_STRING 5050) (T_VARIABLE 5017) (property_name 5051) (simple_variable 5052) ("{" 5053)) 2263 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 5048) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2264 (("]" 5049)) 2265 (("(" 5031) (argument_list 5056)) 2266 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 5054) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2267 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 5055)) 2268 (("(" 5031) (argument_list 5058)) 2269 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 5062) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2270 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 5063)) 2271 ((backup_doc_comment 5066)) 2272 (("(" 5067)) 2273 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 5068)) 2274 ((")" 5069)) 2275 ((T_USE 164) (lexical_vars 5070)) 2276 ((":" 166) (return_type 5071)) 2277 ((backup_fn_flags 5072)) 2278 (("{" 5073)) 2279 ((inner_statement_list 5074)) 2280 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 5075) ("~" 127)) 2281 ((backup_fn_flags 5076)) 2282 ((backup_doc_comment 5078)) 2283 (("(" 5079)) 2284 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 5080)) 2285 ((")" 5081)) 2286 ((":" 2249) (return_type 5082)) 2287 ((T_DOUBLE_ARROW 5083)) 2288 ((backup_fn_flags 5084)) 2289 ((backup_lex_pos 5085)) 2290 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5086) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2291 (("%" 4790) ("*" 4791) ("+" 4792) ("-" 4793) ("." 4794) ("/" 4795) ("<" 4796) (">" 4797) ("?" 4798) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4799) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4800) (T_BOOLEAN_AND 4801) (T_BOOLEAN_OR 4802) (T_COALESCE 4803) (T_INSTANCEOF 4804) (T_IS_EQUAL 4805) (T_IS_GREATER_OR_EQUAL 4806) (T_IS_IDENTICAL 4807) (T_IS_NOT_EQUAL 4808) (T_IS_NOT_IDENTICAL 4809) (T_IS_SMALLER_OR_EQUAL 4810) (T_LOGICAL_AND 4811) (T_LOGICAL_OR 4812) (T_LOGICAL_XOR 4813) (T_POW 4814) (T_SL 4815) (T_SPACESHIP 4816) (T_SR 4817) ("^" 4818) (backup_fn_flags 5087) ("|" 4819)) 2292 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5089) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2293 (("$" 4712) (T_ABSTRACT 4889) (T_ARRAY 4890) (T_AS 4891) (T_BREAK 4892) (T_CALLABLE 4893) (T_CASE 4894) (T_CATCH 4895) (T_CLASS 4896) (T_CLASS_C 4897) (T_CLONE 4898) (T_CONST 4899) (T_CONTINUE 4900) (T_DECLARE 4901) (T_DEFAULT 4902) (T_DIR 4903) (T_DO 4904) (T_ECHO 4905) (T_ELSE 4906) (T_ELSEIF 4907) (T_EMPTY 4908) (T_ENDDECLARE 4909) (T_ENDFOR 4910) (T_ENDFOREACH 4911) (T_ENDIF 4912) (T_ENDSWITCH 4913) (T_ENDWHILE 4914) (T_ENUM 4915) (T_EVAL 4916) (T_EXIT 4917) (T_EXTENDS 4918) (T_FILE 4919) (T_FINAL 4920) (T_FINALLY 4921) (T_FN 4922) (T_FOR 4923) (T_FOREACH 4924) (T_FUNCTION 4925) (T_FUNC_C 4926) (T_GLOBAL 4927) (T_GOTO 4928) (T_IF 4929) (T_IMPLEMENTS 4930) (T_INCLUDE 4931) (T_INCLUDE_ONCE 4932) (T_INSTANCEOF 4933) (T_INSTEADOF 4934) (T_INTERFACE 4935) (T_ISSET 4936) (T_LINE 4937) (T_LIST 4938) (T_LOGICAL_AND 4939) (T_LOGICAL_OR 4940) (T_LOGICAL_XOR 4941) (T_MATCH 4942) (T_METHOD_C 4943) (T_NAMESPACE 4944) (T_NEW 4945) (T_NS_C 4946) (T_PRINT 4947) (T_PRIVATE 4948) (T_PROPERTY_C 4949) (T_PROTECTED 4950) (T_PUBLIC 4951) (T_READONLY 4952) (T_REQUIRE 4953) (T_REQUIRE_ONCE 4954) (T_RETURN 4955) (T_STATIC 4956) (T_STRING 4957) (T_SWITCH 4958) (T_THROW 4959) (T_TRAIT 4960) (T_TRAIT_C 4961) (T_TRY 4962) (T_UNSET 4963) (T_USE 4964) (T_VAR 4965) (T_VARIABLE 4760) (T_WHILE 4966) (T_YIELD 4967) (identifier 5091) (member_name 5092) (reserved_non_modifiers 4970) (semi_reserved 4971) (simple_variable 5093) ("{" 5094)) 2294 (("(" 4976) (argument_list 5097)) 2295 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 5095) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2296 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 5096)) 2297 ((T_FN 35) (T_FUNCTION 129) (fn 4774) #'4775 (inline_function 5101)) 2298 (("$" 4712) (T_STRING 5107) (T_VARIABLE 4760) (property_name 5114) (simple_variable 5109) ("{" 5110)) 2299 (("$" 4712) (T_STRING 5107) (T_VARIABLE 4760) (property_name 5108) (simple_variable 5109) ("{" 5110)) 2300 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 5105) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2301 (("]" 5106)) 2302 (("(" 4976) (argument_list 5113)) 2303 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 5111) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2304 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 5112)) 2305 (("(" 4976) (argument_list 5115)) 2306 (("`" 5117)) 2307 (("]" 5119)) 2308 (("=" 5120)) 2309 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5121) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2310 (("%" 4790) ("*" 4791) ("+" 4792) ("-" 4793) ("." 4794) ("/" 4795) ("<" 4796) (">" 4797) ("?" 4798) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4799) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4800) (T_BOOLEAN_AND 4801) (T_BOOLEAN_OR 4802) (T_COALESCE 4803) (T_DOUBLE_ARROW 5124) (T_INSTANCEOF 4804) (T_IS_EQUAL 4805) (T_IS_GREATER_OR_EQUAL 4806) (T_IS_IDENTICAL 4807) (T_IS_NOT_EQUAL 4808) (T_IS_NOT_IDENTICAL 4809) (T_IS_SMALLER_OR_EQUAL 4810) (T_LOGICAL_AND 4811) (T_LOGICAL_OR 4812) (T_LOGICAL_XOR 4813) (T_POW 4814) (T_SL 4815) (T_SPACESHIP 4816) (T_SR 4817) ("^" 4818) ("|" 4819)) 2311 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5125) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2312 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 5134) (T_VARIABLE 2896) (encaps_var 2910)) 2313 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 5133) (T_VARIABLE 2896) (encaps_var 2901)) 2314 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5175) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2315 ((T_CLASS 5167)) 2316 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 5166) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 5147) (attribute 2958) (class_modifier 2939)) 2317 ((T_PAAMAYIM_NEKUDOTAYIM 5164)) 2318 (("(" 5031) (argument_list 5163)) 2319 ((T_NULLSAFE_OBJECT_OPERATOR 5154) (T_OBJECT_OPERATOR 5155) (T_PAAMAYIM_NEKUDOTAYIM 5156) ("[" 5157)) 2320 (("$" 5015) (T_STRING 5050) (T_VARIABLE 5017) (property_name 5162) (simple_variable 5052) ("{" 5053)) 2321 (("$" 5015) (T_STRING 5050) (T_VARIABLE 5017) (property_name 5161) (simple_variable 5052) ("{" 5053)) 2322 (("$" 5015) (T_VARIABLE 5017) (simple_variable 5160)) 2323 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 5158) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2324 (("]" 5159)) 2325 (("$" 5015) (T_VARIABLE 5017) (simple_variable 5165)) 2326 (("(" 2960) (argument_list 2961) (ctor_arguments 5168)) 2327 ((T_EXTENDS 2963) (extends_from 5169)) 2328 ((T_IMPLEMENTS 2965) (implements_list 5170)) 2329 ((backup_doc_comment 5171)) 2330 (("{" 5172)) 2331 ((class_statement_list 5173)) 2332 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 5174)) 2333 (("%" 299) (")" 5176) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2334 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5178) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2335 (("%" 299) (")" 5179) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2336 (("{" 5180)) 2337 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 5181) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 2338 (("}" 5182)) 2339 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 5184) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 2340 ((")" 5185)) 2341 (("=" 5186)) 2342 (("!" 4710) ("\"" 4711) ("$" 4712) ("(" 4713) ("+" 4714) ("-" 4715) ("@" 4716) (T_ARRAY 4717) (T_ARRAY_CAST 4718) (T_ATTRIBUTE 128) (T_BOOL_CAST 4719) (T_CLASS_C 4720) (T_CLONE 4721) (T_CONSTANT_ENCAPSED_STRING 4722) (T_DEC 4723) (T_DIR 4724) (T_DNUMBER 4725) (T_DOUBLE_CAST 4726) (T_EMPTY 4727) (T_EVAL 4728) (T_EXIT 4729) (T_FILE 4730) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4731) (T_INC 4732) (T_INCLUDE 4733) (T_INCLUDE_ONCE 4734) (T_INT_CAST 4735) (T_ISSET 4736) (T_LINE 4737) (T_LIST 4738) (T_LNUMBER 4739) (T_MATCH 4740) (T_METHOD_C 4741) (T_NAME_FULLY_QUALIFIED 4742) (T_NAME_QUALIFIED 4743) (T_NAME_RELATIVE 4744) (T_NEW 4745) (T_NS_C 4746) (T_OBJECT_CAST 4747) (T_PRINT 4748) (T_PROPERTY_C 4749) (T_READONLY 4750) (T_REQUIRE 4751) (T_REQUIRE_ONCE 4752) (T_START_HEREDOC 4753) (T_STATIC 4754) (T_STRING 4755) (T_STRING_CAST 4756) (T_THROW 4757) (T_TRAIT_C 4758) (T_UNSET_CAST 4759) (T_VARIABLE 4760) (T_YIELD 4761) (T_YIELD_FROM 4762) ("[" 4763) ("`" 4764) (array_object_dereferenceable 4765) (attribute 133) (attributes 4766) (callable_expr 4767) (callable_variable 4768) (class_constant 4769) (class_name 4770) (constant 4771) (dereferenceable_scalar 4772) (expr 5187) (fn 4774) (fully_dereferenceable 105) #'4775 (function_call 4776) (inline_function 4777) (internal_functions_in_yacc 4778) (match 4779) (name 4780) (new_dereferenceable 4781) (new_non_dereferenceable 4782) (scalar 4783) (simple_variable 4784) (static_member 4785) (variable 4786) (variable_class_name 4787) ("~" 4788)) 2343 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 5191) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2344 (("," 5192) (possible_comma 5193)) 2345 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5195) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2346 ((")" 5194)) 2347 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 5203) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 5204) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 5205) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2348 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 5208) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2349 (("," 1480) (possible_comma 5206)) 2350 ((")" 5207)) 2351 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5210) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2352 (("%" 299) (")" 5211) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2353 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5213) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2354 (("%" 299) (")" 5214) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2355 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 5221) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 2356 ((")" 5222)) 2357 (("%" 299) (")" 5227) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2358 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 5230) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2359 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 5231)) 2360 (("\"" 1520) ("$" 2830) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 2831) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 2832) ("[" 1542) (array_object_dereferenceable 2833) (callable_expr 2834) (callable_variable 2835) (class_constant 1546) (class_name 2836) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 2837) (name 2838) (new_dereferenceable 1552) (simple_variable 2839) (static_member 2840) (variable 5257) (variable_class_name 2842)) 2361 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_STRING_VARNAME 5249) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 5250) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2362 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_VARIABLE 5235) (encaps_var 5248)) 2363 ((T_NULLSAFE_OBJECT_OPERATOR 5241) (T_OBJECT_OPERATOR 5242) ("[" 5243)) 2364 (("\"" 5238) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 2365 ((T_STRING 5247)) 2366 ((T_STRING 5246)) 2367 (("-" 2812) (T_NUM_STRING 2813) (T_STRING 2814) (T_VARIABLE 2815) (encaps_var_offset 5244)) 2368 (("]" 5245)) 2369 (("[" 5252) ("}" 5253)) 2370 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 5251)) 2371 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 5254) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2372 (("%" 442) ("*" 443) ("+" 444) ("-" 445) ("." 446) ("/" 447) ("<" 448) (">" 449) ("?" 450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 452) (T_BOOLEAN_AND 453) (T_BOOLEAN_OR 454) (T_COALESCE 455) (T_INSTANCEOF 456) (T_IS_EQUAL 457) (T_IS_GREATER_OR_EQUAL 458) (T_IS_IDENTICAL 459) (T_IS_NOT_EQUAL 460) (T_IS_NOT_IDENTICAL 461) (T_IS_SMALLER_OR_EQUAL 462) (T_LOGICAL_AND 463) (T_LOGICAL_OR 464) (T_LOGICAL_XOR 465) (T_POW 466) (T_SL 467) (T_SPACESHIP 468) (T_SR 469) ("]" 5255) ("^" 470) ("|" 471)) 2373 (("}" 5256)) 2374 (("}" 5258)) 2375 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 5268) (T_VARIABLE 2896) (encaps_var 2910)) 2376 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 5267) (T_VARIABLE 2896) (encaps_var 2901)) 2377 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5309) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2378 ((T_CLASS 5301)) 2379 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 5300) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 5281) (attribute 2958) (class_modifier 2939)) 2380 ((T_PAAMAYIM_NEKUDOTAYIM 5298)) 2381 (("(" 4620) (argument_list 5297)) 2382 ((T_NULLSAFE_OBJECT_OPERATOR 5288) (T_OBJECT_OPERATOR 5289) (T_PAAMAYIM_NEKUDOTAYIM 5290) ("[" 5291)) 2383 (("$" 4604) (T_STRING 4639) (T_VARIABLE 4606) (property_name 5296) (simple_variable 4641) ("{" 4642)) 2384 (("$" 4604) (T_STRING 4639) (T_VARIABLE 4606) (property_name 5295) (simple_variable 4641) ("{" 4642)) 2385 (("$" 4604) (T_VARIABLE 4606) (simple_variable 5294)) 2386 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 5292) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2387 (("]" 5293)) 2388 (("$" 4604) (T_VARIABLE 4606) (simple_variable 5299)) 2389 (("(" 2960) (argument_list 2961) (ctor_arguments 5302)) 2390 ((T_EXTENDS 2963) (extends_from 5303)) 2391 ((T_IMPLEMENTS 2965) (implements_list 5304)) 2392 ((backup_doc_comment 5305)) 2393 (("{" 5306)) 2394 ((class_statement_list 5307)) 2395 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 5308)) 2396 (("%" 299) (")" 5310) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2397 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5312) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2398 (("%" 299) (")" 5313) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2399 (("{" 5314)) 2400 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 5315) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 2401 (("}" 5316)) 2402 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 5318) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 2403 ((")" 5319)) 2404 (("=" 5320)) 2405 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5321) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2406 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 5323) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2407 (("," 5192) (possible_comma 5324)) 2408 ((")" 5325)) 2409 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 5333) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 5334) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 5335) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2410 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 5338) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2411 (("," 1480) (possible_comma 5336)) 2412 ((")" 5337)) 2413 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5340) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2414 (("%" 299) (")" 5341) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2415 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5343) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2416 (("%" 299) (")" 5344) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2417 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 5351) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 2418 ((")" 5352)) 2419 (("%" 299) (")" 5357) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2420 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 5360) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2421 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 5361)) 2422 (("\"" 5363) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 2423 ((backup_doc_comment 5367)) 2424 (("(" 5368)) 2425 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 5369)) 2426 ((")" 5370)) 2427 ((T_USE 164) (lexical_vars 5371)) 2428 ((":" 166) (return_type 5372)) 2429 ((backup_fn_flags 5373)) 2430 (("{" 5374)) 2431 ((inner_statement_list 5375)) 2432 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 5376) ("~" 127)) 2433 ((backup_fn_flags 5377)) 2434 ((backup_doc_comment 5379)) 2435 (("(" 5380)) 2436 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 5381)) 2437 ((")" 5382)) 2438 ((":" 2249) (return_type 5383)) 2439 ((T_DOUBLE_ARROW 5384)) 2440 ((backup_fn_flags 5385)) 2441 ((backup_lex_pos 5386)) 2442 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5387) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 2443 (("%" 4017) ("*" 4018) ("+" 4019) ("-" 4020) ("." 4021) ("/" 4022) ("<" 4023) (">" 4024) ("?" 4025) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4026) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4027) (T_BOOLEAN_AND 4028) (T_BOOLEAN_OR 4029) (T_COALESCE 4030) (T_INSTANCEOF 4031) (T_IS_EQUAL 4032) (T_IS_GREATER_OR_EQUAL 4033) (T_IS_IDENTICAL 4034) (T_IS_NOT_EQUAL 4035) (T_IS_NOT_IDENTICAL 4036) (T_IS_SMALLER_OR_EQUAL 4037) (T_LOGICAL_AND 4038) (T_LOGICAL_OR 4039) (T_LOGICAL_XOR 4040) (T_POW 4041) (T_SL 4042) (T_SPACESHIP 4043) (T_SR 4044) ("^" 4045) (backup_fn_flags 5388) ("|" 4046)) 2444 (("$" 3934) (T_ABSTRACT 4116) (T_ARRAY 4117) (T_AS 4118) (T_BREAK 4119) (T_CALLABLE 4120) (T_CASE 4121) (T_CATCH 4122) (T_CLASS 4123) (T_CLASS_C 4124) (T_CLONE 4125) (T_CONST 4126) (T_CONTINUE 4127) (T_DECLARE 4128) (T_DEFAULT 4129) (T_DIR 4130) (T_DO 4131) (T_ECHO 4132) (T_ELSE 4133) (T_ELSEIF 4134) (T_EMPTY 4135) (T_ENDDECLARE 4136) (T_ENDFOR 4137) (T_ENDFOREACH 4138) (T_ENDIF 4139) (T_ENDSWITCH 4140) (T_ENDWHILE 4141) (T_ENUM 4142) (T_EVAL 4143) (T_EXIT 4144) (T_EXTENDS 4145) (T_FILE 4146) (T_FINAL 4147) (T_FINALLY 4148) (T_FN 4149) (T_FOR 4150) (T_FOREACH 4151) (T_FUNCTION 4152) (T_FUNC_C 4153) (T_GLOBAL 4154) (T_GOTO 4155) (T_IF 4156) (T_IMPLEMENTS 4157) (T_INCLUDE 4158) (T_INCLUDE_ONCE 4159) (T_INSTANCEOF 4160) (T_INSTEADOF 4161) (T_INTERFACE 4162) (T_ISSET 4163) (T_LINE 4164) (T_LIST 4165) (T_LOGICAL_AND 4166) (T_LOGICAL_OR 4167) (T_LOGICAL_XOR 4168) (T_MATCH 4169) (T_METHOD_C 4170) (T_NAMESPACE 4171) (T_NEW 4172) (T_NS_C 4173) (T_PRINT 4174) (T_PRIVATE 4175) (T_PROPERTY_C 4176) (T_PROTECTED 4177) (T_PUBLIC 4178) (T_READONLY 4179) (T_REQUIRE 4180) (T_REQUIRE_ONCE 4181) (T_RETURN 4182) (T_STATIC 4183) (T_STRING 4184) (T_SWITCH 4185) (T_THROW 4186) (T_TRAIT 4187) (T_TRAIT_C 4188) (T_TRY 4189) (T_UNSET 4190) (T_USE 4191) (T_VAR 4192) (T_VARIABLE 3983) (T_WHILE 4193) (T_YIELD 4194) (identifier 5390) (member_name 5391) (reserved_non_modifiers 4197) (semi_reserved 4198) (simple_variable 5392) ("{" 5393)) 2445 (("(" 4203) (argument_list 5396)) 2446 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 5394) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2447 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 5395)) 2448 ((T_FN 35) (T_FUNCTION 129) (fn 3997) #'3998 (inline_function 5400)) 2449 (("$" 3934) (T_STRING 5406) (T_VARIABLE 3983) (property_name 5413) (simple_variable 5408) ("{" 5409)) 2450 (("$" 3934) (T_STRING 5406) (T_VARIABLE 3983) (property_name 5407) (simple_variable 5408) ("{" 5409)) 2451 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 5404) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2452 (("]" 5405)) 2453 (("(" 4203) (argument_list 5412)) 2454 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 5410) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2455 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 5411)) 2456 (("(" 4203) (argument_list 5414)) 2457 (("`" 5416)) 2458 (("]" 5418)) 2459 (("=" 5419)) 2460 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5420) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 2461 (("%" 4017) ("*" 4018) ("+" 4019) ("-" 4020) ("." 4021) ("/" 4022) ("<" 4023) (">" 4024) ("?" 4025) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4026) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4027) (T_BOOLEAN_AND 4028) (T_BOOLEAN_OR 4029) (T_COALESCE 4030) (T_DOUBLE_ARROW 5423) (T_INSTANCEOF 4031) (T_IS_EQUAL 4032) (T_IS_GREATER_OR_EQUAL 4033) (T_IS_IDENTICAL 4034) (T_IS_NOT_EQUAL 4035) (T_IS_NOT_IDENTICAL 4036) (T_IS_SMALLER_OR_EQUAL 4037) (T_LOGICAL_AND 4038) (T_LOGICAL_OR 4039) (T_LOGICAL_XOR 4040) (T_POW 4041) (T_SL 4042) (T_SPACESHIP 4043) (T_SR 4044) ("^" 4045) ("|" 4046)) 2462 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5424) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 2463 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 5433) (T_VARIABLE 2896) (encaps_var 2910)) 2464 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 5432) (T_VARIABLE 2896) (encaps_var 2901)) 2465 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5474) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2466 ((T_CLASS 5466)) 2467 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 5465) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 5446) (attribute 2958) (class_modifier 2939)) 2468 ((T_PAAMAYIM_NEKUDOTAYIM 5463)) 2469 (("(" 4258) (argument_list 5462)) 2470 ((T_NULLSAFE_OBJECT_OPERATOR 5453) (T_OBJECT_OPERATOR 5454) (T_PAAMAYIM_NEKUDOTAYIM 5455) ("[" 5456)) 2471 (("$" 4242) (T_STRING 4277) (T_VARIABLE 4244) (property_name 5461) (simple_variable 4279) ("{" 4280)) 2472 (("$" 4242) (T_STRING 4277) (T_VARIABLE 4244) (property_name 5460) (simple_variable 4279) ("{" 4280)) 2473 (("$" 4242) (T_VARIABLE 4244) (simple_variable 5459)) 2474 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 5457) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2475 (("]" 5458)) 2476 (("$" 4242) (T_VARIABLE 4244) (simple_variable 5464)) 2477 (("(" 2960) (argument_list 2961) (ctor_arguments 5467)) 2478 ((T_EXTENDS 2963) (extends_from 5468)) 2479 ((T_IMPLEMENTS 2965) (implements_list 5469)) 2480 ((backup_doc_comment 5470)) 2481 (("{" 5471)) 2482 ((class_statement_list 5472)) 2483 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 5473)) 2484 (("%" 299) (")" 5475) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2485 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5477) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2486 (("%" 299) (")" 5478) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2487 (("{" 5479)) 2488 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 5480) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 2489 (("}" 5481)) 2490 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 5483) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 2491 ((")" 5484)) 2492 (("=" 5485)) 2493 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 5486) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 2494 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 5488) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2495 (("," 5192) (possible_comma 5489)) 2496 ((")" 5490)) 2497 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 5498) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 5499) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 5500) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2498 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 5503) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2499 (("," 1480) (possible_comma 5501)) 2500 ((")" 5502)) 2501 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5505) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2502 (("%" 299) (")" 5506) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2503 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5508) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2504 (("%" 299) (")" 5509) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2505 ((T_DOUBLE_ARROW 5513)) 2506 (("!" 4299) ("\"" 4300) ("$" 4301) ("(" 4302) ("+" 4303) ("-" 4304) ("@" 4305) (T_ARRAY 4306) (T_ARRAY_CAST 4307) (T_ATTRIBUTE 128) (T_BOOL_CAST 4308) (T_CLASS_C 4309) (T_CLONE 4310) (T_CONSTANT_ENCAPSED_STRING 4311) (T_DEC 4312) (T_DIR 4313) (T_DNUMBER 4314) (T_DOUBLE_CAST 4315) (T_EMPTY 4316) (T_EVAL 4317) (T_EXIT 4318) (T_FILE 4319) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 4320) (T_INC 4321) (T_INCLUDE 4322) (T_INCLUDE_ONCE 4323) (T_INT_CAST 4324) (T_ISSET 4325) (T_LINE 4326) (T_LIST 4327) (T_LNUMBER 4328) (T_MATCH 4329) (T_METHOD_C 4330) (T_NAME_FULLY_QUALIFIED 4331) (T_NAME_QUALIFIED 4332) (T_NAME_RELATIVE 4333) (T_NEW 4334) (T_NS_C 4335) (T_OBJECT_CAST 4336) (T_PRINT 4337) (T_PROPERTY_C 4338) (T_READONLY 4339) (T_REQUIRE 4340) (T_REQUIRE_ONCE 4341) (T_START_HEREDOC 4342) (T_STATIC 4343) (T_STRING 4344) (T_STRING_CAST 4345) (T_THROW 4346) (T_TRAIT_C 4347) (T_UNSET_CAST 4348) (T_VARIABLE 4349) (T_YIELD 4350) (T_YIELD_FROM 4351) ("[" 4352) ("`" 4353) (array_object_dereferenceable 4354) (attribute 133) (attributes 4355) (callable_expr 4356) (callable_variable 4357) (class_constant 4358) (class_name 4359) (constant 4360) (dereferenceable_scalar 4361) (expr 5514) (fn 4363) (fully_dereferenceable 105) #'4364 (function_call 4365) (inline_function 4366) (internal_functions_in_yacc 4367) (match 4368) (name 4369) (new_dereferenceable 4370) (new_non_dereferenceable 4371) (scalar 4372) (simple_variable 4373) (static_member 4374) (variable 4375) (variable_class_name 4376) ("~" 4377)) 2507 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 5520) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 2508 ((")" 5521)) 2509 (("%" 299) (")" 5526) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2510 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 5529) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2511 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 5530)) 2512 (("\"" 5532) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 2513 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 5535) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 2514 ((")" 5536)) 2515 (("=" 5537)) 2516 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 5538) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 2517 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 5540) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2518 (("," 5192) (possible_comma 5541)) 2519 ((")" 5542)) 2520 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 5550) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 5551) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 5552) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2521 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 5555) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2522 (("," 1480) (possible_comma 5553)) 2523 ((")" 5554)) 2524 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5557) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2525 (("%" 299) (")" 5558) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2526 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5560) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2527 (("%" 299) (")" 5561) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2528 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 5568) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 2529 ((")" 5569)) 2530 (("%" 299) (")" 5574) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2531 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 5577) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2532 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 5578)) 2533 (("\"" 5580) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 2534 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 5590) (T_VARIABLE 2896) (encaps_var 2910)) 2535 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 5589) (T_VARIABLE 2896) (encaps_var 2901)) 2536 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5631) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2537 ((T_CLASS 5623)) 2538 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 5622) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 5603) (attribute 2958) (class_modifier 2939)) 2539 ((T_PAAMAYIM_NEKUDOTAYIM 5620)) 2540 (("(" 3370) (argument_list 5619)) 2541 ((T_NULLSAFE_OBJECT_OPERATOR 5610) (T_OBJECT_OPERATOR 5611) (T_PAAMAYIM_NEKUDOTAYIM 5612) ("[" 5613)) 2542 (("$" 3354) (T_STRING 3389) (T_VARIABLE 3356) (property_name 5618) (simple_variable 3391) ("{" 3392)) 2543 (("$" 3354) (T_STRING 3389) (T_VARIABLE 3356) (property_name 5617) (simple_variable 3391) ("{" 3392)) 2544 (("$" 3354) (T_VARIABLE 3356) (simple_variable 5616)) 2545 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 5614) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2546 (("]" 5615)) 2547 (("$" 3354) (T_VARIABLE 3356) (simple_variable 5621)) 2548 (("(" 2960) (argument_list 2961) (ctor_arguments 5624)) 2549 ((T_EXTENDS 2963) (extends_from 5625)) 2550 ((T_IMPLEMENTS 2965) (implements_list 5626)) 2551 ((backup_doc_comment 5627)) 2552 (("{" 5628)) 2553 ((class_statement_list 5629)) 2554 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 5630)) 2555 (("%" 299) (")" 5632) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2556 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5634) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2557 (("%" 299) (")" 5635) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2558 (("{" 5636)) 2559 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 5637) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 2560 (("}" 5638)) 2561 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 5640) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 2562 ((")" 5641)) 2563 (("=" 5642)) 2564 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 5643) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 2565 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 5645) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2566 (("," 5192) (possible_comma 5646)) 2567 ((")" 5647)) 2568 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 5655) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 5656) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 5657) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2569 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 5660) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2570 (("," 1480) (possible_comma 5658)) 2571 ((")" 5659)) 2572 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5662) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2573 (("%" 299) (")" 5663) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2574 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5665) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2575 (("%" 299) (")" 5666) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2576 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 5673) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 2577 ((")" 5674)) 2578 (("%" 299) (")" 5679) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2579 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 5682) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2580 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 5683)) 2581 (("\"" 5685) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 2582 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5914) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2583 (("{" 5689)) 2584 ((property_hook_list 5690)) 2585 ((T_ABSTRACT 5691) (T_ATTRIBUTE 5692) (T_FINAL 5693) (T_PRIVATE 5694) (T_PRIVATE_SET 5695) (T_PROTECTED 5696) (T_PROTECTED_SET 5697) (T_PUBLIC 5698) (T_PUBLIC_SET 5699) (T_READONLY 5700) (T_STATIC 5701) (attribute 5702) (attributes 5703) (member_modifier 5704) (non_empty_member_modifiers 5705) (property_hook 5706) (property_hook_modifiers 5707) ("}" 5708)) 2586 ((T_NAME_FULLY_QUALIFIED 5830) (T_NAME_QUALIFIED 5831) (T_NAME_RELATIVE 5832) (T_STATIC 5833) (T_STRING 5834) (attribute_decl 5835) (attribute_group 5836) (class_name 5837) (name 5838)) 2587 ((T_ABSTRACT 5691) (T_ATTRIBUTE 5692) (T_FINAL 5693) (T_PRIVATE 5694) (T_PRIVATE_SET 5695) (T_PROTECTED 5696) (T_PROTECTED_SET 5697) (T_PUBLIC 5698) (T_PUBLIC_SET 5699) (T_READONLY 5700) (T_STATIC 5701) (attribute 5828) (member_modifier 5704) (non_empty_member_modifiers 5705) (property_hook 5829) (property_hook_modifiers 5707)) 2588 ((T_ABSTRACT 5691) (T_FINAL 5693) (T_PRIVATE 5694) (T_PRIVATE_SET 5695) (T_PROTECTED 5696) (T_PROTECTED_SET 5697) (T_PUBLIC 5698) (T_PUBLIC_SET 5699) (T_READONLY 5700) (T_STATIC 5701) (member_modifier 5827)) 2589 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5709) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5710) (ampersand 5711) (returns_ref 5712)) 2590 ((T_STRING 5713)) 2591 ((backup_doc_comment 5714)) 2592 (("(" 5715) (optional_parameter_list 5716)) 2593 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 5825)) 2594 ((backup_fn_flags 5717)) 2595 ((";" 5718) (T_DOUBLE_ARROW 5719) (property_hook_body 5720) ("{" 5721)) 2596 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5725) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2597 ((backup_fn_flags 5724)) 2598 ((inner_statement_list 5722)) 2599 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 5723) ("~" 127)) 2600 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 5732) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 2601 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5824) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2602 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5823) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2603 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5822) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2604 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5821) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2605 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5820) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2606 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5819) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2607 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5818) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2608 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5817) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2609 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 5812) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 5813) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 2610 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5811) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2611 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5810) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2612 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5809) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2613 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5808) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2614 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5807) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2615 (("$" 5772) ("(" 5773) (T_NAME_FULLY_QUALIFIED 5774) (T_NAME_QUALIFIED 5775) (T_NAME_RELATIVE 5776) (T_STATIC 5777) (T_STRING 5778) (T_VARIABLE 5779) (class_name 5780) (class_name_reference 5781) (name 5782) (new_variable 5783) (simple_variable 5784)) 2616 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5771) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2617 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5770) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2618 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5769) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2619 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5768) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2620 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5767) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2621 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5766) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2622 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5765) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2623 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5764) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2624 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5763) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2625 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5762) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2626 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5761) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2627 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5760) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2628 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5759) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2629 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5758) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2630 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5757) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2631 (("$" 5772) (T_VARIABLE 5779) (simple_variable 5803) ("{" 5804)) 2632 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 5801) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2633 ((T_PAAMAYIM_NEKUDOTAYIM 5799)) 2634 ((T_NULLSAFE_OBJECT_OPERATOR 5785) (T_OBJECT_OPERATOR 5786) (T_PAAMAYIM_NEKUDOTAYIM 5787) ("[" 5788)) 2635 (("$" 5772) (T_STRING 5792) (T_VARIABLE 5779) (property_name 5798) (simple_variable 5794) ("{" 5795)) 2636 (("$" 5772) (T_STRING 5792) (T_VARIABLE 5779) (property_name 5793) (simple_variable 5794) ("{" 5795)) 2637 (("$" 5772) (T_VARIABLE 5779) (simple_variable 5791)) 2638 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 5789) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2639 (("]" 5790)) 2640 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 5796) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2641 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 5797)) 2642 (("$" 5772) (T_VARIABLE 5779) (simple_variable 5800)) 2643 (("%" 299) (")" 5802) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2644 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 5805) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2645 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 5806)) 2646 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5816) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2647 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 5814) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 2648 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 5815) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 2649 ((")" 5826)) 2650 (("," 5847) (possible_comma 5848)) 2651 (("(" 5839) (argument_list 5840)) 2652 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 5841) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 5842) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 5843) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2653 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 5846) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2654 (("," 1480) (possible_comma 5844)) 2655 ((")" 5845)) 2656 ((T_NAME_FULLY_QUALIFIED 5830) (T_NAME_QUALIFIED 5831) (T_NAME_RELATIVE 5832) (T_STATIC 5833) (T_STRING 5834) (attribute_decl 5850) (class_name 5837) (name 5838)) 2657 (("]" 5849)) 2658 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6891) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2659 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 6889) (encaps_var 5237)) 2660 (("$" 5853) (T_VARIABLE 5901) (simple_variable 6885) ("{" 6886)) 2661 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 6883) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2662 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6882) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2663 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6881) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2664 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6880) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2665 (("(" 6877)) 2666 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6876) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2667 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6875) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2668 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6874) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2669 (("\"" 1520) ("$" 6156) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 6157) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 6158) ("[" 1542) (array_object_dereferenceable 6159) (callable_expr 6160) (callable_variable 6161) (class_constant 1546) (class_name 6162) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 6163) (name 6164) (new_dereferenceable 1552) (simple_variable 6165) (static_member 6166) (variable 6873) (variable_class_name 6168)) 2670 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6872) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2671 (("(" 6869)) 2672 (("(" 6866)) 2673 (("(" 6857) (argument_list 6858) (ctor_arguments 6859)) 2674 (("\"" 1520) ("$" 6156) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 6157) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 6158) ("[" 1542) (array_object_dereferenceable 6159) (callable_expr 6160) (callable_variable 6161) (class_constant 1546) (class_name 6162) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 6163) (name 6164) (new_dereferenceable 1552) (simple_variable 6165) (static_member 6166) (variable 6856) (variable_class_name 6168)) 2675 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6855) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2676 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6854) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2677 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6853) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2678 (("(" 6849)) 2679 (("(" 6844)) 2680 (("(" 6838)) 2681 (("$" 6156) ("(" 6801) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 6802) (T_NAME_QUALIFIED 6803) (T_NAME_RELATIVE 6804) (T_READONLY 2931) (T_STATIC 6805) (T_STRING 6806) (T_VARIABLE 6158) (anonymous_class 6807) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 6808) (attribute 2937) (attributes 6809) (class_modifier 2939) (class_name 6810) (class_name_reference 6811) (name 6812) (new_variable 6813) (simple_variable 6814)) 2682 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6800) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2683 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6799) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2684 (("(" 6117) (argument_list 6798)) 2685 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6797) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2686 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6796) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2687 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 6791) (T_END_HEREDOC 6792) (T_VARIABLE 2896) (encaps_list 6793) (encaps_var 2898)) 2688 ((T_FN 35) (T_FUNCTION 129) (fn 5915) #'5916 (inline_function 6790)) 2689 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6789) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2690 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6788) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2691 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6787) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2692 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6329) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2693 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6265) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2694 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 6261) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 2695 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 6259) (encaps_list 2804) (encaps_var 2805)) 2696 ((T_NULLSAFE_OBJECT_OPERATOR 6245) (T_OBJECT_OPERATOR 6246) ("[" 6247)) 2697 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 6242) (attribute 2733) (fn 5915) #'5916 (inline_function 6243)) 2698 (("(" 6117) (argument_list 6241)) 2699 ((T_PAAMAYIM_NEKUDOTAYIM 6233)) 2700 (("%" 5931) ("*" 5932) ("+" 5933) ("-" 5934) ("." 5935) ("/" 5936) ("<" 5937) (">" 5938) ("?" 5939) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5940) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5941) (T_BOOLEAN_AND 5942) (T_BOOLEAN_OR 5943) (T_COALESCE 5944) (T_INSTANCEOF 5945) (T_IS_EQUAL 5946) (T_IS_GREATER_OR_EQUAL 5947) (T_IS_IDENTICAL 5948) (T_IS_NOT_EQUAL 5949) (T_IS_NOT_IDENTICAL 5950) (T_IS_SMALLER_OR_EQUAL 5951) (T_LOGICAL_AND 5952) (T_LOGICAL_OR 5953) (T_LOGICAL_XOR 5954) (T_POW 5955) (T_SL 5956) (T_SPACESHIP 5957) (T_SR 5958) ("^" 5959) (backup_doc_comment 6229) ("|" 5960)) 2701 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 6218)) 2702 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 6206)) 2703 (("(" 6117) (argument_list 6205)) 2704 (("=" 6125) (T_AND_EQUAL 6126) (T_COALESCE_EQUAL 6127) (T_CONCAT_EQUAL 6128) (T_DEC 6129) (T_DIV_EQUAL 6130) (T_INC 6131) (T_MINUS_EQUAL 6132) (T_MOD_EQUAL 6133) (T_MUL_EQUAL 6134) (T_OR_EQUAL 6135) (T_PLUS_EQUAL 6136) (T_POW_EQUAL 6137) (T_SL_EQUAL 6138) (T_SR_EQUAL 6139) (T_XOR_EQUAL 6140)) 2705 ((T_PAAMAYIM_NEKUDOTAYIM 6029)) 2706 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5930) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2707 (("%" 5931) ("*" 5932) ("+" 5933) ("-" 5934) ("." 5935) ("/" 5936) ("<" 5937) (">" 5938) ("?" 5939) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5940) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5941) (T_BOOLEAN_AND 5942) (T_BOOLEAN_OR 5943) (T_COALESCE 5944) (T_INSTANCEOF 5945) (T_IS_EQUAL 5946) (T_IS_GREATER_OR_EQUAL 5947) (T_IS_IDENTICAL 5948) (T_IS_NOT_EQUAL 5949) (T_IS_NOT_IDENTICAL 5950) (T_IS_SMALLER_OR_EQUAL 5951) (T_LOGICAL_AND 5952) (T_LOGICAL_OR 5953) (T_LOGICAL_XOR 5954) (T_POW 5955) (T_SL 5956) (T_SPACESHIP 5957) (T_SR 5958) ("^" 5959) ("|" 5960)) 2708 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6028) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2709 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6027) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2710 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6026) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2711 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6025) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2712 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6024) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2713 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6023) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2714 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6022) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2715 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6021) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2716 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 6016) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 6017) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 2717 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6015) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2718 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6014) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2719 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6013) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2720 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6012) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2721 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6011) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2722 (("$" 5976) ("(" 5977) (T_NAME_FULLY_QUALIFIED 5978) (T_NAME_QUALIFIED 5979) (T_NAME_RELATIVE 5980) (T_STATIC 5981) (T_STRING 5982) (T_VARIABLE 5983) (class_name 5984) (class_name_reference 5985) (name 5986) (new_variable 5987) (simple_variable 5988)) 2723 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5975) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2724 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5974) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2725 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5973) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2726 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5972) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2727 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5971) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2728 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5970) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2729 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5969) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2730 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5968) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2731 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5967) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2732 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5966) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2733 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5965) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2734 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5964) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2735 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5963) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2736 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5962) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2737 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 5961) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2738 (("$" 5976) (T_VARIABLE 5983) (simple_variable 6007) ("{" 6008)) 2739 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 6005) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2740 ((T_PAAMAYIM_NEKUDOTAYIM 6003)) 2741 ((T_NULLSAFE_OBJECT_OPERATOR 5989) (T_OBJECT_OPERATOR 5990) (T_PAAMAYIM_NEKUDOTAYIM 5991) ("[" 5992)) 2742 (("$" 5976) (T_STRING 5996) (T_VARIABLE 5983) (property_name 6002) (simple_variable 5998) ("{" 5999)) 2743 (("$" 5976) (T_STRING 5996) (T_VARIABLE 5983) (property_name 5997) (simple_variable 5998) ("{" 5999)) 2744 (("$" 5976) (T_VARIABLE 5983) (simple_variable 5995)) 2745 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 5993) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2746 (("]" 5994)) 2747 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6000) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2748 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6001)) 2749 (("$" 5976) (T_VARIABLE 5983) (simple_variable 6004)) 2750 (("%" 299) (")" 6006) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2751 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6009) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2752 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6010)) 2753 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6020) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2754 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 6018) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 2755 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6019) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2756 (("$" 5853) (T_ABSTRACT 6030) (T_ARRAY 6031) (T_AS 6032) (T_BREAK 6033) (T_CALLABLE 6034) (T_CASE 6035) (T_CATCH 6036) (T_CLASS 6037) (T_CLASS_C 6038) (T_CLONE 6039) (T_CONST 6040) (T_CONTINUE 6041) (T_DECLARE 6042) (T_DEFAULT 6043) (T_DIR 6044) (T_DO 6045) (T_ECHO 6046) (T_ELSE 6047) (T_ELSEIF 6048) (T_EMPTY 6049) (T_ENDDECLARE 6050) (T_ENDFOR 6051) (T_ENDFOREACH 6052) (T_ENDIF 6053) (T_ENDSWITCH 6054) (T_ENDWHILE 6055) (T_ENUM 6056) (T_EVAL 6057) (T_EXIT 6058) (T_EXTENDS 6059) (T_FILE 6060) (T_FINAL 6061) (T_FINALLY 6062) (T_FN 6063) (T_FOR 6064) (T_FOREACH 6065) (T_FUNCTION 6066) (T_FUNC_C 6067) (T_GLOBAL 6068) (T_GOTO 6069) (T_IF 6070) (T_IMPLEMENTS 6071) (T_INCLUDE 6072) (T_INCLUDE_ONCE 6073) (T_INSTANCEOF 6074) (T_INSTEADOF 6075) (T_INTERFACE 6076) (T_ISSET 6077) (T_LINE 6078) (T_LIST 6079) (T_LOGICAL_AND 6080) (T_LOGICAL_OR 6081) (T_LOGICAL_XOR 6082) (T_MATCH 6083) (T_METHOD_C 6084) (T_NAMESPACE 6085) (T_NEW 6086) (T_NS_C 6087) (T_PRINT 6088) (T_PRIVATE 6089) (T_PROPERTY_C 6090) (T_PROTECTED 6091) (T_PUBLIC 6092) (T_READONLY 6093) (T_REQUIRE 6094) (T_REQUIRE_ONCE 6095) (T_RETURN 6096) (T_STATIC 6097) (T_STRING 6098) (T_SWITCH 6099) (T_THROW 6100) (T_TRAIT 6101) (T_TRAIT_C 6102) (T_TRY 6103) (T_UNSET 6104) (T_USE 6105) (T_VAR 6106) (T_VARIABLE 5901) (T_WHILE 6107) (T_YIELD 6108) (identifier 6109) (member_name 6110) (reserved_non_modifiers 6111) (semi_reserved 6112) (simple_variable 6113) ("{" 6114)) 2757 (("(" 6117) (argument_list 6118)) 2758 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6115) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2759 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6116)) 2760 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 6119) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 6120) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 6121) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2761 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 6124) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2762 (("," 1480) (possible_comma 6122)) 2763 ((")" 6123)) 2764 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (ampersand 6154) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6155) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2765 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6153) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2766 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6152) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2767 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6151) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2768 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6150) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2769 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6149) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2770 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6148) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2771 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6147) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2772 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6146) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2773 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6145) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2774 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6144) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2775 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6143) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2776 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6142) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2777 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6141) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2778 (("\"" 1520) ("$" 6156) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 6157) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 6158) ("[" 1542) (array_object_dereferenceable 6159) (callable_expr 6160) (callable_variable 6161) (class_constant 1546) (class_name 6162) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 6163) (name 6164) (new_dereferenceable 1552) (simple_variable 6165) (static_member 6166) (variable 6167) (variable_class_name 6168)) 2779 (("$" 6156) (T_VARIABLE 6158) (simple_variable 6201) ("{" 6202)) 2780 (("(" 6172) (argument_list 6200)) 2781 ((T_NULLSAFE_OBJECT_OPERATOR 6186) (T_OBJECT_OPERATOR 6187) ("[" 6188)) 2782 (("(" 6172) (argument_list 6185)) 2783 ((T_PAAMAYIM_NEKUDOTAYIM 6181)) 2784 (("(" 6172) (argument_list 6180)) 2785 ((T_PAAMAYIM_NEKUDOTAYIM 6169)) 2786 (("$" 6156) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 6158) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 6170) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 6171) ("{" 1642)) 2787 (("(" 6172) (argument_list 6173)) 2788 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 6174) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 6175) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 6176) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2789 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 6179) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2790 (("," 1480) (possible_comma 6177)) 2791 ((")" 6178)) 2792 (("$" 6156) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 6158) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 6182) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 6183) ("{" 1658)) 2793 (("(" 6172) (argument_list 6184)) 2794 (("$" 6156) (T_STRING 6191) (T_VARIABLE 6158) (property_name 6198) (simple_variable 6193) ("{" 6194)) 2795 (("$" 6156) (T_STRING 6191) (T_VARIABLE 6158) (property_name 6192) (simple_variable 6193) ("{" 6194)) 2796 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 6189) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2797 (("]" 6190)) 2798 (("(" 6172) (argument_list 6197)) 2799 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6195) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2800 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6196)) 2801 (("(" 6172) (argument_list 6199)) 2802 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6203) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2803 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6204)) 2804 ((backup_doc_comment 6207)) 2805 (("(" 6208)) 2806 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 6209)) 2807 ((")" 6210)) 2808 ((T_USE 164) (lexical_vars 6211)) 2809 ((":" 166) (return_type 6212)) 2810 ((backup_fn_flags 6213)) 2811 (("{" 6214)) 2812 ((inner_statement_list 6215)) 2813 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 6216) ("~" 127)) 2814 ((backup_fn_flags 6217)) 2815 ((backup_doc_comment 6219)) 2816 (("(" 6220)) 2817 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 6221)) 2818 ((")" 6222)) 2819 ((":" 2249) (return_type 6223)) 2820 ((T_DOUBLE_ARROW 6224)) 2821 ((backup_fn_flags 6225)) 2822 ((backup_lex_pos 6226)) 2823 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6227) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2824 (("%" 5931) ("*" 5932) ("+" 5933) ("-" 5934) ("." 5935) ("/" 5936) ("<" 5937) (">" 5938) ("?" 5939) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5940) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5941) (T_BOOLEAN_AND 5942) (T_BOOLEAN_OR 5943) (T_COALESCE 5944) (T_INSTANCEOF 5945) (T_IS_EQUAL 5946) (T_IS_GREATER_OR_EQUAL 5947) (T_IS_IDENTICAL 5948) (T_IS_NOT_EQUAL 5949) (T_IS_NOT_IDENTICAL 5950) (T_IS_SMALLER_OR_EQUAL 5951) (T_LOGICAL_AND 5952) (T_LOGICAL_OR 5953) (T_LOGICAL_XOR 5954) (T_POW 5955) (T_SL 5956) (T_SPACESHIP 5957) (T_SR 5958) ("^" 5959) (backup_fn_flags 6228) ("|" 5960)) 2825 (("{" 6230)) 2826 ((property_hook_list 6231)) 2827 ((T_ABSTRACT 5691) (T_ATTRIBUTE 5692) (T_FINAL 5693) (T_PRIVATE 5694) (T_PRIVATE_SET 5695) (T_PROTECTED 5696) (T_PROTECTED_SET 5697) (T_PUBLIC 5698) (T_PUBLIC_SET 5699) (T_READONLY 5700) (T_STATIC 5701) (attribute 5702) (attributes 5703) (member_modifier 5704) (non_empty_member_modifiers 5705) (property_hook 5706) (property_hook_modifiers 5707) ("}" 6232)) 2828 (("$" 5853) (T_ABSTRACT 6030) (T_ARRAY 6031) (T_AS 6032) (T_BREAK 6033) (T_CALLABLE 6034) (T_CASE 6035) (T_CATCH 6036) (T_CLASS 6037) (T_CLASS_C 6038) (T_CLONE 6039) (T_CONST 6040) (T_CONTINUE 6041) (T_DECLARE 6042) (T_DEFAULT 6043) (T_DIR 6044) (T_DO 6045) (T_ECHO 6046) (T_ELSE 6047) (T_ELSEIF 6048) (T_EMPTY 6049) (T_ENDDECLARE 6050) (T_ENDFOR 6051) (T_ENDFOREACH 6052) (T_ENDIF 6053) (T_ENDSWITCH 6054) (T_ENDWHILE 6055) (T_ENUM 6056) (T_EVAL 6057) (T_EXIT 6058) (T_EXTENDS 6059) (T_FILE 6060) (T_FINAL 6061) (T_FINALLY 6062) (T_FN 6063) (T_FOR 6064) (T_FOREACH 6065) (T_FUNCTION 6066) (T_FUNC_C 6067) (T_GLOBAL 6068) (T_GOTO 6069) (T_IF 6070) (T_IMPLEMENTS 6071) (T_INCLUDE 6072) (T_INCLUDE_ONCE 6073) (T_INSTANCEOF 6074) (T_INSTEADOF 6075) (T_INTERFACE 6076) (T_ISSET 6077) (T_LINE 6078) (T_LIST 6079) (T_LOGICAL_AND 6080) (T_LOGICAL_OR 6081) (T_LOGICAL_XOR 6082) (T_MATCH 6083) (T_METHOD_C 6084) (T_NAMESPACE 6085) (T_NEW 6086) (T_NS_C 6087) (T_PRINT 6088) (T_PRIVATE 6089) (T_PROPERTY_C 6090) (T_PROTECTED 6091) (T_PUBLIC 6092) (T_READONLY 6093) (T_REQUIRE 6094) (T_REQUIRE_ONCE 6095) (T_RETURN 6096) (T_STATIC 6097) (T_STRING 6098) (T_SWITCH 6099) (T_THROW 6100) (T_TRAIT 6101) (T_TRAIT_C 6102) (T_TRY 6103) (T_UNSET 6104) (T_USE 6105) (T_VAR 6106) (T_VARIABLE 5901) (T_WHILE 6107) (T_YIELD 6108) (identifier 6234) (member_name 6235) (reserved_non_modifiers 6111) (semi_reserved 6112) (simple_variable 6236) ("{" 6237)) 2829 (("(" 6117) (argument_list 6240)) 2830 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6238) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2831 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6239)) 2832 ((T_FN 35) (T_FUNCTION 129) (fn 5915) #'5916 (inline_function 6244)) 2833 (("$" 5853) (T_STRING 6250) (T_VARIABLE 5901) (property_name 6257) (simple_variable 6252) ("{" 6253)) 2834 (("$" 5853) (T_STRING 6250) (T_VARIABLE 5901) (property_name 6251) (simple_variable 6252) ("{" 6253)) 2835 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 6248) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2836 (("]" 6249)) 2837 (("(" 6117) (argument_list 6256)) 2838 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6254) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2839 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6255)) 2840 (("(" 6117) (argument_list 6258)) 2841 (("`" 6260)) 2842 (("]" 6262)) 2843 (("=" 6263)) 2844 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6264) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 2845 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6786) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2846 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 6784) (encaps_var 5237)) 2847 (("$" 6268) (T_VARIABLE 6316) (simple_variable 6780) ("{" 6781)) 2848 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 6778) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2849 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6777) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2850 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6776) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2851 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6775) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2852 (("(" 6772)) 2853 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6771) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2854 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6770) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2855 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6769) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2856 (("\"" 1520) ("$" 6571) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 6572) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 6573) ("[" 1542) (array_object_dereferenceable 6574) (callable_expr 6575) (callable_variable 6576) (class_constant 1546) (class_name 6577) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 6578) (name 6579) (new_dereferenceable 1552) (simple_variable 6580) (static_member 6581) (variable 6768) (variable_class_name 6583)) 2857 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6767) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2858 (("(" 6764)) 2859 (("(" 6761)) 2860 (("(" 6752) (argument_list 6753) (ctor_arguments 6754)) 2861 (("\"" 1520) ("$" 6571) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 6572) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 6573) ("[" 1542) (array_object_dereferenceable 6574) (callable_expr 6575) (callable_variable 6576) (class_constant 1546) (class_name 6577) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 6578) (name 6579) (new_dereferenceable 1552) (simple_variable 6580) (static_member 6581) (variable 6751) (variable_class_name 6583)) 2862 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6750) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2863 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6749) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2864 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6748) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2865 (("(" 6744)) 2866 (("(" 6739)) 2867 (("(" 6733)) 2868 (("$" 6571) ("(" 6696) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 6697) (T_NAME_QUALIFIED 6698) (T_NAME_RELATIVE 6699) (T_READONLY 2931) (T_STATIC 6700) (T_STRING 6701) (T_VARIABLE 6573) (anonymous_class 6702) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 6703) (attribute 2937) (attributes 6704) (class_modifier 2939) (class_name 6705) (class_name_reference 6706) (name 6707) (new_variable 6708) (simple_variable 6709)) 2869 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6695) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2870 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6694) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2871 (("(" 6532) (argument_list 6693)) 2872 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6692) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2873 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6691) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2874 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 6686) (T_END_HEREDOC 6687) (T_VARIABLE 2896) (encaps_list 6688) (encaps_var 2898)) 2875 ((T_FN 35) (T_FUNCTION 129) (fn 6330) #'6331 (inline_function 6685)) 2876 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6684) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2877 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6683) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2878 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6682) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2879 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6679) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2880 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6678) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2881 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 6674) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 2882 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 6672) (encaps_list 2804) (encaps_var 2805)) 2883 ((T_NULLSAFE_OBJECT_OPERATOR 6658) (T_OBJECT_OPERATOR 6659) ("[" 6660)) 2884 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 6655) (attribute 2733) (fn 6330) #'6331 (inline_function 6656)) 2885 (("(" 6532) (argument_list 6654)) 2886 ((T_PAAMAYIM_NEKUDOTAYIM 6646)) 2887 (("%" 6346) ("*" 6347) ("+" 6348) ("-" 6349) ("." 6350) ("/" 6351) ("<" 6352) (">" 6353) ("?" 6354) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6355) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6356) (T_BOOLEAN_AND 6357) (T_BOOLEAN_OR 6358) (T_COALESCE 6359) (T_DOUBLE_ARROW 6644) (T_INSTANCEOF 6360) (T_IS_EQUAL 6361) (T_IS_GREATER_OR_EQUAL 6362) (T_IS_IDENTICAL 6363) (T_IS_NOT_EQUAL 6364) (T_IS_NOT_IDENTICAL 6365) (T_IS_SMALLER_OR_EQUAL 6366) (T_LOGICAL_AND 6367) (T_LOGICAL_OR 6368) (T_LOGICAL_XOR 6369) (T_POW 6370) (T_SL 6371) (T_SPACESHIP 6372) (T_SR 6373) ("^" 6374) ("|" 6375)) 2888 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 6633)) 2889 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 6621)) 2890 (("(" 6532) (argument_list 6620)) 2891 (("=" 6540) (T_AND_EQUAL 6541) (T_COALESCE_EQUAL 6542) (T_CONCAT_EQUAL 6543) (T_DEC 6544) (T_DIV_EQUAL 6545) (T_INC 6546) (T_MINUS_EQUAL 6547) (T_MOD_EQUAL 6548) (T_MUL_EQUAL 6549) (T_OR_EQUAL 6550) (T_PLUS_EQUAL 6551) (T_POW_EQUAL 6552) (T_SL_EQUAL 6553) (T_SR_EQUAL 6554) (T_XOR_EQUAL 6555)) 2892 ((T_PAAMAYIM_NEKUDOTAYIM 6444)) 2893 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6345) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2894 (("%" 6346) ("*" 6347) ("+" 6348) ("-" 6349) ("." 6350) ("/" 6351) ("<" 6352) (">" 6353) ("?" 6354) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6355) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6356) (T_BOOLEAN_AND 6357) (T_BOOLEAN_OR 6358) (T_COALESCE 6359) (T_INSTANCEOF 6360) (T_IS_EQUAL 6361) (T_IS_GREATER_OR_EQUAL 6362) (T_IS_IDENTICAL 6363) (T_IS_NOT_EQUAL 6364) (T_IS_NOT_IDENTICAL 6365) (T_IS_SMALLER_OR_EQUAL 6366) (T_LOGICAL_AND 6367) (T_LOGICAL_OR 6368) (T_LOGICAL_XOR 6369) (T_POW 6370) (T_SL 6371) (T_SPACESHIP 6372) (T_SR 6373) ("^" 6374) ("|" 6375)) 2895 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6443) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2896 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6442) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2897 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6441) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2898 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6440) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2899 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6439) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2900 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6438) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2901 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6437) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2902 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6436) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2903 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 6431) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 6432) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 2904 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6430) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2905 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6429) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2906 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6428) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2907 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6427) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2908 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6426) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2909 (("$" 6391) ("(" 6392) (T_NAME_FULLY_QUALIFIED 6393) (T_NAME_QUALIFIED 6394) (T_NAME_RELATIVE 6395) (T_STATIC 6396) (T_STRING 6397) (T_VARIABLE 6398) (class_name 6399) (class_name_reference 6400) (name 6401) (new_variable 6402) (simple_variable 6403)) 2910 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6390) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2911 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6389) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2912 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6388) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2913 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6387) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2914 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6386) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2915 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6385) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2916 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6384) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2917 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6383) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2918 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6382) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2919 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6381) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2920 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6380) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2921 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6379) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2922 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6378) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2923 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6377) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2924 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6376) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2925 (("$" 6391) (T_VARIABLE 6398) (simple_variable 6422) ("{" 6423)) 2926 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 6420) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 2927 ((T_PAAMAYIM_NEKUDOTAYIM 6418)) 2928 ((T_NULLSAFE_OBJECT_OPERATOR 6404) (T_OBJECT_OPERATOR 6405) (T_PAAMAYIM_NEKUDOTAYIM 6406) ("[" 6407)) 2929 (("$" 6391) (T_STRING 6411) (T_VARIABLE 6398) (property_name 6417) (simple_variable 6413) ("{" 6414)) 2930 (("$" 6391) (T_STRING 6411) (T_VARIABLE 6398) (property_name 6412) (simple_variable 6413) ("{" 6414)) 2931 (("$" 6391) (T_VARIABLE 6398) (simple_variable 6410)) 2932 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 6408) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2933 (("]" 6409)) 2934 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6415) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2935 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6416)) 2936 (("$" 6391) (T_VARIABLE 6398) (simple_variable 6419)) 2937 (("%" 299) (")" 6421) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 2938 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6424) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2939 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6425)) 2940 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6435) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2941 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 6433) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 2942 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6434) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2943 (("$" 6268) (T_ABSTRACT 6445) (T_ARRAY 6446) (T_AS 6447) (T_BREAK 6448) (T_CALLABLE 6449) (T_CASE 6450) (T_CATCH 6451) (T_CLASS 6452) (T_CLASS_C 6453) (T_CLONE 6454) (T_CONST 6455) (T_CONTINUE 6456) (T_DECLARE 6457) (T_DEFAULT 6458) (T_DIR 6459) (T_DO 6460) (T_ECHO 6461) (T_ELSE 6462) (T_ELSEIF 6463) (T_EMPTY 6464) (T_ENDDECLARE 6465) (T_ENDFOR 6466) (T_ENDFOREACH 6467) (T_ENDIF 6468) (T_ENDSWITCH 6469) (T_ENDWHILE 6470) (T_ENUM 6471) (T_EVAL 6472) (T_EXIT 6473) (T_EXTENDS 6474) (T_FILE 6475) (T_FINAL 6476) (T_FINALLY 6477) (T_FN 6478) (T_FOR 6479) (T_FOREACH 6480) (T_FUNCTION 6481) (T_FUNC_C 6482) (T_GLOBAL 6483) (T_GOTO 6484) (T_IF 6485) (T_IMPLEMENTS 6486) (T_INCLUDE 6487) (T_INCLUDE_ONCE 6488) (T_INSTANCEOF 6489) (T_INSTEADOF 6490) (T_INTERFACE 6491) (T_ISSET 6492) (T_LINE 6493) (T_LIST 6494) (T_LOGICAL_AND 6495) (T_LOGICAL_OR 6496) (T_LOGICAL_XOR 6497) (T_MATCH 6498) (T_METHOD_C 6499) (T_NAMESPACE 6500) (T_NEW 6501) (T_NS_C 6502) (T_PRINT 6503) (T_PRIVATE 6504) (T_PROPERTY_C 6505) (T_PROTECTED 6506) (T_PUBLIC 6507) (T_READONLY 6508) (T_REQUIRE 6509) (T_REQUIRE_ONCE 6510) (T_RETURN 6511) (T_STATIC 6512) (T_STRING 6513) (T_SWITCH 6514) (T_THROW 6515) (T_TRAIT 6516) (T_TRAIT_C 6517) (T_TRY 6518) (T_UNSET 6519) (T_USE 6520) (T_VAR 6521) (T_VARIABLE 6316) (T_WHILE 6522) (T_YIELD 6523) (identifier 6524) (member_name 6525) (reserved_non_modifiers 6526) (semi_reserved 6527) (simple_variable 6528) ("{" 6529)) 2944 (("(" 6532) (argument_list 6533)) 2945 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6530) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2946 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6531)) 2947 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 6534) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 6535) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 6536) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2948 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 6539) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2949 (("," 1480) (possible_comma 6537)) 2950 ((")" 6538)) 2951 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (ampersand 6569) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6570) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2952 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6568) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2953 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6567) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2954 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6566) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2955 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6565) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2956 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6564) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2957 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6563) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2958 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6562) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2959 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6561) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2960 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6560) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2961 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6559) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2962 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6558) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2963 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6557) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2964 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6556) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 2965 (("\"" 1520) ("$" 6571) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 6572) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 6573) ("[" 1542) (array_object_dereferenceable 6574) (callable_expr 6575) (callable_variable 6576) (class_constant 1546) (class_name 6577) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 6578) (name 6579) (new_dereferenceable 1552) (simple_variable 6580) (static_member 6581) (variable 6582) (variable_class_name 6583)) 2966 (("$" 6571) (T_VARIABLE 6573) (simple_variable 6616) ("{" 6617)) 2967 (("(" 6587) (argument_list 6615)) 2968 ((T_NULLSAFE_OBJECT_OPERATOR 6601) (T_OBJECT_OPERATOR 6602) ("[" 6603)) 2969 (("(" 6587) (argument_list 6600)) 2970 ((T_PAAMAYIM_NEKUDOTAYIM 6596)) 2971 (("(" 6587) (argument_list 6595)) 2972 ((T_PAAMAYIM_NEKUDOTAYIM 6584)) 2973 (("$" 6571) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 6573) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 6585) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 6586) ("{" 1642)) 2974 (("(" 6587) (argument_list 6588)) 2975 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 6589) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 6590) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 6591) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2976 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 6594) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 2977 (("," 1480) (possible_comma 6592)) 2978 ((")" 6593)) 2979 (("$" 6571) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 6573) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 6597) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 6598) ("{" 1658)) 2980 (("(" 6587) (argument_list 6599)) 2981 (("$" 6571) (T_STRING 6606) (T_VARIABLE 6573) (property_name 6613) (simple_variable 6608) ("{" 6609)) 2982 (("$" 6571) (T_STRING 6606) (T_VARIABLE 6573) (property_name 6607) (simple_variable 6608) ("{" 6609)) 2983 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 6604) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 2984 (("]" 6605)) 2985 (("(" 6587) (argument_list 6612)) 2986 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6610) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2987 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6611)) 2988 (("(" 6587) (argument_list 6614)) 2989 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6618) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 2990 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6619)) 2991 ((backup_doc_comment 6622)) 2992 (("(" 6623)) 2993 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 6624)) 2994 ((")" 6625)) 2995 ((T_USE 164) (lexical_vars 6626)) 2996 ((":" 166) (return_type 6627)) 2997 ((backup_fn_flags 6628)) 2998 (("{" 6629)) 2999 ((inner_statement_list 6630)) 3000 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 6631) ("~" 127)) 3001 ((backup_fn_flags 6632)) 3002 ((backup_doc_comment 6634)) 3003 (("(" 6635)) 3004 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 6636)) 3005 ((")" 6637)) 3006 ((":" 2249) (return_type 6638)) 3007 ((T_DOUBLE_ARROW 6639)) 3008 ((backup_fn_flags 6640)) 3009 ((backup_lex_pos 6641)) 3010 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6642) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 3011 (("%" 6346) ("*" 6347) ("+" 6348) ("-" 6349) ("." 6350) ("/" 6351) ("<" 6352) (">" 6353) ("?" 6354) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6355) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6356) (T_BOOLEAN_AND 6357) (T_BOOLEAN_OR 6358) (T_COALESCE 6359) (T_INSTANCEOF 6360) (T_IS_EQUAL 6361) (T_IS_GREATER_OR_EQUAL 6362) (T_IS_IDENTICAL 6363) (T_IS_NOT_EQUAL 6364) (T_IS_NOT_IDENTICAL 6365) (T_IS_SMALLER_OR_EQUAL 6366) (T_LOGICAL_AND 6367) (T_LOGICAL_OR 6368) (T_LOGICAL_XOR 6369) (T_POW 6370) (T_SL 6371) (T_SPACESHIP 6372) (T_SR 6373) ("^" 6374) (backup_fn_flags 6643) ("|" 6375)) 3012 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6645) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 3013 (("$" 6268) (T_ABSTRACT 6445) (T_ARRAY 6446) (T_AS 6447) (T_BREAK 6448) (T_CALLABLE 6449) (T_CASE 6450) (T_CATCH 6451) (T_CLASS 6452) (T_CLASS_C 6453) (T_CLONE 6454) (T_CONST 6455) (T_CONTINUE 6456) (T_DECLARE 6457) (T_DEFAULT 6458) (T_DIR 6459) (T_DO 6460) (T_ECHO 6461) (T_ELSE 6462) (T_ELSEIF 6463) (T_EMPTY 6464) (T_ENDDECLARE 6465) (T_ENDFOR 6466) (T_ENDFOREACH 6467) (T_ENDIF 6468) (T_ENDSWITCH 6469) (T_ENDWHILE 6470) (T_ENUM 6471) (T_EVAL 6472) (T_EXIT 6473) (T_EXTENDS 6474) (T_FILE 6475) (T_FINAL 6476) (T_FINALLY 6477) (T_FN 6478) (T_FOR 6479) (T_FOREACH 6480) (T_FUNCTION 6481) (T_FUNC_C 6482) (T_GLOBAL 6483) (T_GOTO 6484) (T_IF 6485) (T_IMPLEMENTS 6486) (T_INCLUDE 6487) (T_INCLUDE_ONCE 6488) (T_INSTANCEOF 6489) (T_INSTEADOF 6490) (T_INTERFACE 6491) (T_ISSET 6492) (T_LINE 6493) (T_LIST 6494) (T_LOGICAL_AND 6495) (T_LOGICAL_OR 6496) (T_LOGICAL_XOR 6497) (T_MATCH 6498) (T_METHOD_C 6499) (T_NAMESPACE 6500) (T_NEW 6501) (T_NS_C 6502) (T_PRINT 6503) (T_PRIVATE 6504) (T_PROPERTY_C 6505) (T_PROTECTED 6506) (T_PUBLIC 6507) (T_READONLY 6508) (T_REQUIRE 6509) (T_REQUIRE_ONCE 6510) (T_RETURN 6511) (T_STATIC 6512) (T_STRING 6513) (T_SWITCH 6514) (T_THROW 6515) (T_TRAIT 6516) (T_TRAIT_C 6517) (T_TRY 6518) (T_UNSET 6519) (T_USE 6520) (T_VAR 6521) (T_VARIABLE 6316) (T_WHILE 6522) (T_YIELD 6523) (identifier 6647) (member_name 6648) (reserved_non_modifiers 6526) (semi_reserved 6527) (simple_variable 6649) ("{" 6650)) 3014 (("(" 6532) (argument_list 6653)) 3015 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6651) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3016 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6652)) 3017 ((T_FN 35) (T_FUNCTION 129) (fn 6330) #'6331 (inline_function 6657)) 3018 (("$" 6268) (T_STRING 6663) (T_VARIABLE 6316) (property_name 6670) (simple_variable 6665) ("{" 6666)) 3019 (("$" 6268) (T_STRING 6663) (T_VARIABLE 6316) (property_name 6664) (simple_variable 6665) ("{" 6666)) 3020 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 6661) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3021 (("]" 6662)) 3022 (("(" 6532) (argument_list 6669)) 3023 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6667) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3024 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6668)) 3025 (("(" 6532) (argument_list 6671)) 3026 (("`" 6673)) 3027 (("]" 6675)) 3028 (("=" 6676)) 3029 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6677) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 3030 (("%" 6346) ("*" 6347) ("+" 6348) ("-" 6349) ("." 6350) ("/" 6351) ("<" 6352) (">" 6353) ("?" 6354) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6355) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6356) (T_BOOLEAN_AND 6357) (T_BOOLEAN_OR 6358) (T_COALESCE 6359) (T_DOUBLE_ARROW 6680) (T_INSTANCEOF 6360) (T_IS_EQUAL 6361) (T_IS_GREATER_OR_EQUAL 6362) (T_IS_IDENTICAL 6363) (T_IS_NOT_EQUAL 6364) (T_IS_NOT_IDENTICAL 6365) (T_IS_SMALLER_OR_EQUAL 6366) (T_LOGICAL_AND 6367) (T_LOGICAL_OR 6368) (T_LOGICAL_XOR 6369) (T_POW 6370) (T_SL 6371) (T_SPACESHIP 6372) (T_SR 6373) ("^" 6374) ("|" 6375)) 3031 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6681) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 3032 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 6690) (T_VARIABLE 2896) (encaps_var 2910)) 3033 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 6689) (T_VARIABLE 2896) (encaps_var 2901)) 3034 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 6731) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3035 ((T_CLASS 6723)) 3036 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 6722) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 6703) (attribute 2958) (class_modifier 2939)) 3037 ((T_PAAMAYIM_NEKUDOTAYIM 6720)) 3038 (("(" 6587) (argument_list 6719)) 3039 ((T_NULLSAFE_OBJECT_OPERATOR 6710) (T_OBJECT_OPERATOR 6711) (T_PAAMAYIM_NEKUDOTAYIM 6712) ("[" 6713)) 3040 (("$" 6571) (T_STRING 6606) (T_VARIABLE 6573) (property_name 6718) (simple_variable 6608) ("{" 6609)) 3041 (("$" 6571) (T_STRING 6606) (T_VARIABLE 6573) (property_name 6717) (simple_variable 6608) ("{" 6609)) 3042 (("$" 6571) (T_VARIABLE 6573) (simple_variable 6716)) 3043 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 6714) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3044 (("]" 6715)) 3045 (("$" 6571) (T_VARIABLE 6573) (simple_variable 6721)) 3046 (("(" 2960) (argument_list 2961) (ctor_arguments 6724)) 3047 ((T_EXTENDS 2963) (extends_from 6725)) 3048 ((T_IMPLEMENTS 2965) (implements_list 6726)) 3049 ((backup_doc_comment 6727)) 3050 (("{" 6728)) 3051 ((class_statement_list 6729)) 3052 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 6730)) 3053 (("%" 299) (")" 6732) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3054 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 6734) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3055 (("%" 299) (")" 6735) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3056 (("{" 6736)) 3057 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 6737) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 3058 (("}" 6738)) 3059 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 6740) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 3060 ((")" 6741)) 3061 (("=" 6742)) 3062 (("!" 6266) ("\"" 6267) ("$" 6268) ("(" 6269) ("+" 6270) ("-" 6271) ("@" 6272) (T_ARRAY 6273) (T_ARRAY_CAST 6274) (T_ATTRIBUTE 128) (T_BOOL_CAST 6275) (T_CLASS_C 6276) (T_CLONE 6277) (T_CONSTANT_ENCAPSED_STRING 6278) (T_DEC 6279) (T_DIR 6280) (T_DNUMBER 6281) (T_DOUBLE_CAST 6282) (T_EMPTY 6283) (T_EVAL 6284) (T_EXIT 6285) (T_FILE 6286) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 6287) (T_INC 6288) (T_INCLUDE 6289) (T_INCLUDE_ONCE 6290) (T_INT_CAST 6291) (T_ISSET 6292) (T_LINE 6293) (T_LIST 6294) (T_LNUMBER 6295) (T_MATCH 6296) (T_METHOD_C 6297) (T_NAME_FULLY_QUALIFIED 6298) (T_NAME_QUALIFIED 6299) (T_NAME_RELATIVE 6300) (T_NEW 6301) (T_NS_C 6302) (T_OBJECT_CAST 6303) (T_PRINT 6304) (T_PROPERTY_C 6305) (T_READONLY 6306) (T_REQUIRE 6307) (T_REQUIRE_ONCE 6308) (T_START_HEREDOC 6309) (T_STATIC 6310) (T_STRING 6311) (T_STRING_CAST 6312) (T_THROW 6313) (T_TRAIT_C 6314) (T_UNSET_CAST 6315) (T_VARIABLE 6316) (T_YIELD 6317) (T_YIELD_FROM 6318) ("[" 6319) ("`" 6320) (array_object_dereferenceable 6321) (attribute 133) (attributes 6322) (callable_expr 6323) (callable_variable 6324) (class_constant 6325) (class_name 6326) (constant 6327) (dereferenceable_scalar 6328) (expr 6743) (fn 6330) (fully_dereferenceable 105) #'6331 (function_call 6332) (inline_function 6333) (internal_functions_in_yacc 6334) (match 6335) (name 6336) (new_dereferenceable 6337) (new_non_dereferenceable 6338) (scalar 6339) (simple_variable 6340) (static_member 6341) (variable 6342) (variable_class_name 6343) ("~" 6344)) 3063 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 6745) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3064 (("," 5192) (possible_comma 6746)) 3065 ((")" 6747)) 3066 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 6755) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 6756) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 6757) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3067 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 6760) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3068 (("," 1480) (possible_comma 6758)) 3069 ((")" 6759)) 3070 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 6762) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3071 (("%" 299) (")" 6763) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3072 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 6765) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3073 (("%" 299) (")" 6766) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3074 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 6773) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 3075 ((")" 6774)) 3076 (("%" 299) (")" 6779) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3077 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6782) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3078 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6783)) 3079 (("\"" 6785) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 3080 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 6795) (T_VARIABLE 2896) (encaps_var 2910)) 3081 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 6794) (T_VARIABLE 2896) (encaps_var 2901)) 3082 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 6836) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3083 ((T_CLASS 6828)) 3084 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 6827) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 6808) (attribute 2958) (class_modifier 2939)) 3085 ((T_PAAMAYIM_NEKUDOTAYIM 6825)) 3086 (("(" 6172) (argument_list 6824)) 3087 ((T_NULLSAFE_OBJECT_OPERATOR 6815) (T_OBJECT_OPERATOR 6816) (T_PAAMAYIM_NEKUDOTAYIM 6817) ("[" 6818)) 3088 (("$" 6156) (T_STRING 6191) (T_VARIABLE 6158) (property_name 6823) (simple_variable 6193) ("{" 6194)) 3089 (("$" 6156) (T_STRING 6191) (T_VARIABLE 6158) (property_name 6822) (simple_variable 6193) ("{" 6194)) 3090 (("$" 6156) (T_VARIABLE 6158) (simple_variable 6821)) 3091 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 6819) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3092 (("]" 6820)) 3093 (("$" 6156) (T_VARIABLE 6158) (simple_variable 6826)) 3094 (("(" 2960) (argument_list 2961) (ctor_arguments 6829)) 3095 ((T_EXTENDS 2963) (extends_from 6830)) 3096 ((T_IMPLEMENTS 2965) (implements_list 6831)) 3097 ((backup_doc_comment 6832)) 3098 (("{" 6833)) 3099 ((class_statement_list 6834)) 3100 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 6835)) 3101 (("%" 299) (")" 6837) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3102 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 6839) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3103 (("%" 299) (")" 6840) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3104 (("{" 6841)) 3105 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 6842) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 3106 (("}" 6843)) 3107 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 6845) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 3108 ((")" 6846)) 3109 (("=" 6847)) 3110 (("!" 5851) ("\"" 5852) ("$" 5853) ("(" 5854) ("+" 5855) ("-" 5856) ("@" 5857) (T_ARRAY 5858) (T_ARRAY_CAST 5859) (T_ATTRIBUTE 128) (T_BOOL_CAST 5860) (T_CLASS_C 5861) (T_CLONE 5862) (T_CONSTANT_ENCAPSED_STRING 5863) (T_DEC 5864) (T_DIR 5865) (T_DNUMBER 5866) (T_DOUBLE_CAST 5867) (T_EMPTY 5868) (T_EVAL 5869) (T_EXIT 5870) (T_FILE 5871) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 5872) (T_INC 5873) (T_INCLUDE 5874) (T_INCLUDE_ONCE 5875) (T_INT_CAST 5876) (T_ISSET 5877) (T_LINE 5878) (T_LIST 5879) (T_LNUMBER 5880) (T_MATCH 5881) (T_METHOD_C 5882) (T_NAME_FULLY_QUALIFIED 5883) (T_NAME_QUALIFIED 5884) (T_NAME_RELATIVE 5885) (T_NEW 5886) (T_NS_C 5887) (T_OBJECT_CAST 5888) (T_PRINT 5889) (T_PROPERTY_C 5890) (T_READONLY 5891) (T_REQUIRE 5892) (T_REQUIRE_ONCE 5893) (T_START_HEREDOC 5894) (T_STATIC 5895) (T_STRING 5896) (T_STRING_CAST 5897) (T_THROW 5898) (T_TRAIT_C 5899) (T_UNSET_CAST 5900) (T_VARIABLE 5901) (T_YIELD 5902) (T_YIELD_FROM 5903) ("[" 5904) ("`" 5905) (array_object_dereferenceable 5906) (attribute 133) (attributes 5907) (callable_expr 5908) (callable_variable 5909) (class_constant 5910) (class_name 5911) (constant 5912) (dereferenceable_scalar 5913) (expr 6848) (fn 5915) (fully_dereferenceable 105) #'5916 (function_call 5917) (inline_function 5918) (internal_functions_in_yacc 5919) (match 5920) (name 5921) (new_dereferenceable 5922) (new_non_dereferenceable 5923) (scalar 5924) (simple_variable 5925) (static_member 5926) (variable 5927) (variable_class_name 5928) ("~" 5929)) 3111 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 6850) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3112 (("," 5192) (possible_comma 6851)) 3113 ((")" 6852)) 3114 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 6860) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 6861) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 6862) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3115 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 6865) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3116 (("," 1480) (possible_comma 6863)) 3117 ((")" 6864)) 3118 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 6867) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3119 (("%" 299) (")" 6868) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3120 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 6870) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3121 (("%" 299) (")" 6871) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3122 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 6878) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 3123 ((")" 6879)) 3124 (("%" 299) (")" 6884) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3125 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 6887) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3126 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 6888)) 3127 (("\"" 6890) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 3128 ((T_ARRAY 3031) (T_CALLABLE 3032) (T_NAME_FULLY_QUALIFIED 3033) (T_NAME_QUALIFIED 3034) (T_NAME_RELATIVE 3035) (T_STRING 3036) (name 3037) (type_without_static 6893)) 3129 ((")" 6903) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3026)) 3130 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6907) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6908) (ampersand 6909) (returns_ref 6910)) 3131 ((T_ABSTRACT 6911) (T_ARRAY 6912) (T_AS 6913) (T_BREAK 6914) (T_CALLABLE 6915) (T_CASE 6916) (T_CATCH 6917) (T_CLASS 6918) (T_CLASS_C 6919) (T_CLONE 6920) (T_CONST 6921) (T_CONTINUE 6922) (T_DECLARE 6923) (T_DEFAULT 6924) (T_DIR 6925) (T_DO 6926) (T_ECHO 6927) (T_ELSE 6928) (T_ELSEIF 6929) (T_EMPTY 6930) (T_ENDDECLARE 6931) (T_ENDFOR 6932) (T_ENDFOREACH 6933) (T_ENDIF 6934) (T_ENDSWITCH 6935) (T_ENDWHILE 6936) (T_ENUM 6937) (T_EVAL 6938) (T_EXIT 6939) (T_EXTENDS 6940) (T_FILE 6941) (T_FINAL 6942) (T_FINALLY 6943) (T_FN 6944) (T_FOR 6945) (T_FOREACH 6946) (T_FUNCTION 6947) (T_FUNC_C 6948) (T_GLOBAL 6949) (T_GOTO 6950) (T_IF 6951) (T_IMPLEMENTS 6952) (T_INCLUDE 6953) (T_INCLUDE_ONCE 6954) (T_INSTANCEOF 6955) (T_INSTEADOF 6956) (T_INTERFACE 6957) (T_ISSET 6958) (T_LINE 6959) (T_LIST 6960) (T_LOGICAL_AND 6961) (T_LOGICAL_OR 6962) (T_LOGICAL_XOR 6963) (T_MATCH 6964) (T_METHOD_C 6965) (T_NAMESPACE 6966) (T_NEW 6967) (T_NS_C 6968) (T_PRINT 6969) (T_PRIVATE 6970) (T_PROPERTY_C 6971) (T_PROTECTED 6972) (T_PUBLIC 6973) (T_READONLY 6974) (T_REQUIRE 6975) (T_REQUIRE_ONCE 6976) (T_RETURN 6977) (T_STATIC 6978) (T_STRING 6979) (T_SWITCH 6980) (T_THROW 6981) (T_TRAIT 6982) (T_TRAIT_C 6983) (T_TRY 6984) (T_UNSET 6985) (T_USE 6986) (T_VAR 6987) (T_WHILE 6988) (T_YIELD 6989) (identifier 6990) (reserved_non_modifiers 6991) (semi_reserved 6992)) 3132 ((backup_doc_comment 6993)) 3133 (("(" 6994)) 3134 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 6995)) 3135 ((")" 6996)) 3136 ((":" 6997) (return_type 6998)) 3137 (("(" 2256) ("?" 7006) (T_ARRAY 7007) (T_CALLABLE 7008) (T_NAME_FULLY_QUALIFIED 7009) (T_NAME_QUALIFIED 7010) (T_NAME_RELATIVE 7011) (T_STATIC 7012) (T_STRING 7013) (intersection_type 7014) (name 7015) (type 7016) (type_expr 7017) (type_without_static 7018) (union_type 7019) (union_type_element 7020)) 3138 ((backup_fn_flags 6999)) 3139 ((";" 7000) (method_body 7001) ("{" 7002)) 3140 ((backup_fn_flags 7005)) 3141 ((inner_statement_list 7003)) 3142 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 7004) ("~" 127)) 3143 ((T_ARRAY 7051) (T_CALLABLE 7052) (T_NAME_FULLY_QUALIFIED 7053) (T_NAME_QUALIFIED 7054) (T_NAME_RELATIVE 7055) (T_STATIC 7056) (T_STRING 7057) (name 7058) (type 7059) (type_without_static 7060)) 3144 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 7049)) 3145 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 7038)) 3146 (("|" 7036)) 3147 (("|" 7021)) 3148 (("(" 7022) (T_ARRAY 7023) (T_CALLABLE 7024) (T_NAME_FULLY_QUALIFIED 7025) (T_NAME_QUALIFIED 7026) (T_NAME_RELATIVE 7027) (T_STATIC 7028) (T_STRING 7029) (name 7030) (type 7031) (type_without_static 7032) (union_type_element 7033)) 3149 ((T_ARRAY 2285) (T_CALLABLE 2286) (T_NAME_FULLY_QUALIFIED 2287) (T_NAME_QUALIFIED 2288) (T_NAME_RELATIVE 2289) (T_STATIC 2290) (T_STRING 2291) (intersection_type 7034) (name 2293) (type 2294) (type_without_static 2295)) 3150 ((")" 7035) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2308)) 3151 (("(" 7022) (T_ARRAY 7023) (T_CALLABLE 7024) (T_NAME_FULLY_QUALIFIED 7025) (T_NAME_QUALIFIED 7026) (T_NAME_RELATIVE 7027) (T_STATIC 7028) (T_STRING 7029) (name 7030) (type 7031) (type_without_static 7032) (union_type_element 7037)) 3152 ((T_ARRAY 7039) (T_CALLABLE 7040) (T_NAME_FULLY_QUALIFIED 7041) (T_NAME_QUALIFIED 7042) (T_NAME_RELATIVE 7043) (T_STATIC 7044) (T_STRING 7045) (name 7046) (type 7047) (type_without_static 7048)) 3153 ((T_ARRAY 7039) (T_CALLABLE 7040) (T_NAME_FULLY_QUALIFIED 7041) (T_NAME_QUALIFIED 7042) (T_NAME_RELATIVE 7043) (T_STATIC 7044) (T_STRING 7045) (name 7046) (type 7050) (type_without_static 7048)) 3154 (("(" 2256) ("?" 7062) (T_ABSTRACT 7063) (T_ARRAY 7064) (T_AS 7065) (T_BREAK 7066) (T_CALLABLE 7067) (T_CASE 7068) (T_CATCH 7069) (T_CLASS 7070) (T_CLASS_C 7071) (T_CLONE 7072) (T_CONST 7073) (T_CONTINUE 7074) (T_DECLARE 7075) (T_DEFAULT 7076) (T_DIR 7077) (T_DO 7078) (T_ECHO 7079) (T_ELSE 7080) (T_ELSEIF 7081) (T_EMPTY 7082) (T_ENDDECLARE 7083) (T_ENDFOR 7084) (T_ENDFOREACH 7085) (T_ENDIF 7086) (T_ENDSWITCH 7087) (T_ENDWHILE 7088) (T_ENUM 7089) (T_EVAL 7090) (T_EXIT 7091) (T_EXTENDS 7092) (T_FILE 7093) (T_FINAL 7094) (T_FINALLY 7095) (T_FN 7096) (T_FOR 7097) (T_FOREACH 7098) (T_FUNCTION 7099) (T_FUNC_C 7100) (T_GLOBAL 7101) (T_GOTO 7102) (T_IF 7103) (T_IMPLEMENTS 7104) (T_INCLUDE 7105) (T_INCLUDE_ONCE 7106) (T_INSTANCEOF 7107) (T_INSTEADOF 7108) (T_INTERFACE 7109) (T_ISSET 7110) (T_LINE 7111) (T_LIST 7112) (T_LOGICAL_AND 7113) (T_LOGICAL_OR 7114) (T_LOGICAL_XOR 7115) (T_MATCH 7116) (T_METHOD_C 7117) (T_NAMESPACE 7118) (T_NAME_FULLY_QUALIFIED 7119) (T_NAME_QUALIFIED 7120) (T_NAME_RELATIVE 7121) (T_NEW 7122) (T_NS_C 7123) (T_PRINT 7124) (T_PRIVATE 7125) (T_PROPERTY_C 7126) (T_PROTECTED 7127) (T_PUBLIC 7128) (T_READONLY 7129) (T_REQUIRE 7130) (T_REQUIRE_ONCE 7131) (T_RETURN 7132) (T_STATIC 7133) (T_STRING 7134) (T_SWITCH 7135) (T_THROW 7136) (T_TRAIT 7137) (T_TRAIT_C 7138) (T_TRY 7139) (T_UNSET 7140) (T_USE 7141) (T_VAR 7142) (T_WHILE 7143) (T_YIELD 7144) (class_const_decl 7145) (class_const_list 7146) (intersection_type 7147) (name 7148) (reserved_non_modifiers 7149) (semi_reserved 7150) (type 7151) (type_expr 7152) (type_without_static 7153) (union_type 7154) (union_type_element 7155)) 3155 ((T_ARRAY 7201) (T_CALLABLE 7202) (T_NAME_FULLY_QUALIFIED 7203) (T_NAME_QUALIFIED 7204) (T_NAME_RELATIVE 7205) (T_STATIC 7206) (T_STRING 7207) (name 7208) (type 7209) (type_without_static 7210)) 3156 (("=" 7181)) 3157 (("," 7178) (";" 7200)) 3158 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 7198)) 3159 (("=" 7195)) 3160 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 7184)) 3161 ((T_ABSTRACT 7063) (T_ARRAY 7173) (T_AS 7065) (T_BREAK 7066) (T_CALLABLE 7174) (T_CASE 7068) (T_CATCH 7069) (T_CLASS 7070) (T_CLASS_C 7071) (T_CLONE 7072) (T_CONST 7073) (T_CONTINUE 7074) (T_DECLARE 7075) (T_DEFAULT 7076) (T_DIR 7077) (T_DO 7078) (T_ECHO 7079) (T_ELSE 7080) (T_ELSEIF 7081) (T_EMPTY 7082) (T_ENDDECLARE 7083) (T_ENDFOR 7084) (T_ENDFOREACH 7085) (T_ENDIF 7086) (T_ENDSWITCH 7087) (T_ENDWHILE 7088) (T_ENUM 7089) (T_EVAL 7090) (T_EXIT 7091) (T_EXTENDS 7092) (T_FILE 7093) (T_FINAL 7094) (T_FINALLY 7095) (T_FN 7096) (T_FOR 7097) (T_FOREACH 7098) (T_FUNCTION 7099) (T_FUNC_C 7100) (T_GLOBAL 7101) (T_GOTO 7102) (T_IF 7103) (T_IMPLEMENTS 7104) (T_INCLUDE 7105) (T_INCLUDE_ONCE 7106) (T_INSTANCEOF 7107) (T_INSTEADOF 7108) (T_INTERFACE 7109) (T_ISSET 7110) (T_LINE 7111) (T_LIST 7112) (T_LOGICAL_AND 7113) (T_LOGICAL_OR 7114) (T_LOGICAL_XOR 7115) (T_MATCH 7116) (T_METHOD_C 7117) (T_NAMESPACE 7118) (T_NEW 7122) (T_NS_C 7123) (T_PRINT 7124) (T_PRIVATE 7125) (T_PROPERTY_C 7126) (T_PROTECTED 7127) (T_PUBLIC 7128) (T_READONLY 7129) (T_REQUIRE 7130) (T_REQUIRE_ONCE 7131) (T_RETURN 7132) (T_STATIC 7175) (T_STRING 7176) (T_SWITCH 7135) (T_THROW 7136) (T_TRAIT 7137) (T_TRAIT_C 7138) (T_TRY 7139) (T_UNSET 7140) (T_USE 7141) (T_VAR 7142) (T_WHILE 7143) (T_YIELD 7144) (class_const_decl 7145) (class_const_list 7177) (reserved_non_modifiers 7149) (semi_reserved 7150)) 3162 (("|" 7171)) 3163 (("|" 7156)) 3164 (("(" 7157) (T_ARRAY 7158) (T_CALLABLE 7159) (T_NAME_FULLY_QUALIFIED 7160) (T_NAME_QUALIFIED 7161) (T_NAME_RELATIVE 7162) (T_STATIC 7163) (T_STRING 7164) (name 7165) (type 7166) (type_without_static 7167) (union_type_element 7168)) 3165 ((T_ARRAY 2285) (T_CALLABLE 2286) (T_NAME_FULLY_QUALIFIED 2287) (T_NAME_QUALIFIED 2288) (T_NAME_RELATIVE 2289) (T_STATIC 2290) (T_STRING 2291) (intersection_type 7169) (name 2293) (type 2294) (type_without_static 2295)) 3166 ((")" 7170) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2308)) 3167 (("(" 7157) (T_ARRAY 7158) (T_CALLABLE 7159) (T_NAME_FULLY_QUALIFIED 7160) (T_NAME_QUALIFIED 7161) (T_NAME_RELATIVE 7162) (T_STATIC 7163) (T_STRING 7164) (name 7165) (type 7166) (type_without_static 7167) (union_type_element 7172)) 3168 (("," 7178) (";" 7179)) 3169 ((T_ABSTRACT 7063) (T_ARRAY 7173) (T_AS 7065) (T_BREAK 7066) (T_CALLABLE 7174) (T_CASE 7068) (T_CATCH 7069) (T_CLASS 7070) (T_CLASS_C 7071) (T_CLONE 7072) (T_CONST 7073) (T_CONTINUE 7074) (T_DECLARE 7075) (T_DEFAULT 7076) (T_DIR 7077) (T_DO 7078) (T_ECHO 7079) (T_ELSE 7080) (T_ELSEIF 7081) (T_EMPTY 7082) (T_ENDDECLARE 7083) (T_ENDFOR 7084) (T_ENDFOREACH 7085) (T_ENDIF 7086) (T_ENDSWITCH 7087) (T_ENDWHILE 7088) (T_ENUM 7089) (T_EVAL 7090) (T_EXIT 7091) (T_EXTENDS 7092) (T_FILE 7093) (T_FINAL 7094) (T_FINALLY 7095) (T_FN 7096) (T_FOR 7097) (T_FOREACH 7098) (T_FUNCTION 7099) (T_FUNC_C 7100) (T_GLOBAL 7101) (T_GOTO 7102) (T_IF 7103) (T_IMPLEMENTS 7104) (T_INCLUDE 7105) (T_INCLUDE_ONCE 7106) (T_INSTANCEOF 7107) (T_INSTEADOF 7108) (T_INTERFACE 7109) (T_ISSET 7110) (T_LINE 7111) (T_LIST 7112) (T_LOGICAL_AND 7113) (T_LOGICAL_OR 7114) (T_LOGICAL_XOR 7115) (T_MATCH 7116) (T_METHOD_C 7117) (T_NAMESPACE 7118) (T_NEW 7122) (T_NS_C 7123) (T_PRINT 7124) (T_PRIVATE 7125) (T_PROPERTY_C 7126) (T_PROTECTED 7127) (T_PUBLIC 7128) (T_READONLY 7129) (T_REQUIRE 7130) (T_REQUIRE_ONCE 7131) (T_RETURN 7132) (T_STATIC 7175) (T_STRING 7176) (T_SWITCH 7135) (T_THROW 7136) (T_TRAIT 7137) (T_TRAIT_C 7138) (T_TRY 7139) (T_UNSET 7140) (T_USE 7141) (T_VAR 7142) (T_WHILE 7143) (T_YIELD 7144) (class_const_decl 7180) (reserved_non_modifiers 7149) (semi_reserved 7150)) 3170 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 7182) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 3171 (("%" 3129) ("*" 3130) ("+" 3131) ("-" 3132) ("." 3133) ("/" 3134) ("<" 3135) (">" 3136) ("?" 3137) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3138) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3139) (T_BOOLEAN_AND 3140) (T_BOOLEAN_OR 3141) (T_COALESCE 3142) (T_INSTANCEOF 3143) (T_IS_EQUAL 3144) (T_IS_GREATER_OR_EQUAL 3145) (T_IS_IDENTICAL 3146) (T_IS_NOT_EQUAL 3147) (T_IS_NOT_IDENTICAL 3148) (T_IS_SMALLER_OR_EQUAL 3149) (T_LOGICAL_AND 3150) (T_LOGICAL_OR 3151) (T_LOGICAL_XOR 3152) (T_POW 3153) (T_SL 3154) (T_SPACESHIP 3155) (T_SR 3156) ("^" 3157) (backup_doc_comment 7183) ("|" 3158)) 3172 ((T_ARRAY 7185) (T_CALLABLE 7186) (T_NAME_FULLY_QUALIFIED 7187) (T_NAME_QUALIFIED 7188) (T_NAME_RELATIVE 7189) (T_STATIC 7190) (T_STRING 7191) (name 7192) (type 7193) (type_without_static 7194)) 3173 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 7196) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 3174 (("%" 3129) ("*" 3130) ("+" 3131) ("-" 3132) ("." 3133) ("/" 3134) ("<" 3135) (">" 3136) ("?" 3137) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3138) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3139) (T_BOOLEAN_AND 3140) (T_BOOLEAN_OR 3141) (T_COALESCE 3142) (T_INSTANCEOF 3143) (T_IS_EQUAL 3144) (T_IS_GREATER_OR_EQUAL 3145) (T_IS_IDENTICAL 3146) (T_IS_NOT_EQUAL 3147) (T_IS_NOT_IDENTICAL 3148) (T_IS_SMALLER_OR_EQUAL 3149) (T_LOGICAL_AND 3150) (T_LOGICAL_OR 3151) (T_LOGICAL_XOR 3152) (T_POW 3153) (T_SL 3154) (T_SPACESHIP 3155) (T_SR 3156) ("^" 3157) (backup_doc_comment 7197) ("|" 3158)) 3175 ((T_ARRAY 7185) (T_CALLABLE 7186) (T_NAME_FULLY_QUALIFIED 7187) (T_NAME_QUALIFIED 7188) (T_NAME_RELATIVE 7189) (T_STATIC 7190) (T_STRING 7191) (name 7192) (type 7199) (type_without_static 7194)) 3176 (("," 7221) (";" 7222) (trait_adaptations 7223) ("{" 7224)) 3177 ((T_NAME_FULLY_QUALIFIED 7213) (T_NAME_QUALIFIED 7214) (T_NAME_RELATIVE 7215) (T_STATIC 7216) (T_STRING 7217) (class_name 7513) (name 7220)) 3178 ((T_ABSTRACT 7225) (T_ARRAY 7226) (T_AS 7227) (T_BREAK 7228) (T_CALLABLE 7229) (T_CASE 7230) (T_CATCH 7231) (T_CLASS 7232) (T_CLASS_C 7233) (T_CLONE 7234) (T_CONST 7235) (T_CONTINUE 7236) (T_DECLARE 7237) (T_DEFAULT 7238) (T_DIR 7239) (T_DO 7240) (T_ECHO 7241) (T_ELSE 7242) (T_ELSEIF 7243) (T_EMPTY 7244) (T_ENDDECLARE 7245) (T_ENDFOR 7246) (T_ENDFOREACH 7247) (T_ENDIF 7248) (T_ENDSWITCH 7249) (T_ENDWHILE 7250) (T_ENUM 7251) (T_EVAL 7252) (T_EXIT 7253) (T_EXTENDS 7254) (T_FILE 7255) (T_FINAL 7256) (T_FINALLY 7257) (T_FN 7258) (T_FOR 7259) (T_FOREACH 7260) (T_FUNCTION 7261) (T_FUNC_C 7262) (T_GLOBAL 7263) (T_GOTO 7264) (T_IF 7265) (T_IMPLEMENTS 7266) (T_INCLUDE 7267) (T_INCLUDE_ONCE 7268) (T_INSTANCEOF 7269) (T_INSTEADOF 7270) (T_INTERFACE 7271) (T_ISSET 7272) (T_LINE 7273) (T_LIST 7274) (T_LOGICAL_AND 7275) (T_LOGICAL_OR 7276) (T_LOGICAL_XOR 7277) (T_MATCH 7278) (T_METHOD_C 7279) (T_NAMESPACE 7280) (T_NAME_FULLY_QUALIFIED 7281) (T_NAME_QUALIFIED 7282) (T_NAME_RELATIVE 7283) (T_NEW 7284) (T_NS_C 7285) (T_PRINT 7286) (T_PRIVATE 7287) (T_PROPERTY_C 7288) (T_PROTECTED 7289) (T_PUBLIC 7290) (T_READONLY 7291) (T_REQUIRE 7292) (T_REQUIRE_ONCE 7293) (T_RETURN 7294) (T_STATIC 7295) (T_STRING 7296) (T_SWITCH 7297) (T_THROW 7298) (T_TRAIT 7299) (T_TRAIT_C 7300) (T_TRY 7301) (T_UNSET 7302) (T_USE 7303) (T_VAR 7304) (T_WHILE 7305) (T_YIELD 7306) (absolute_trait_method_reference 7307) (class_name 7308) (identifier 7309) (name 7310) (reserved_non_modifiers 7311) (semi_reserved 7312) (trait_adaptation 7313) (trait_adaptation_list 7314) (trait_alias 7315) (trait_method_reference 7316) (trait_precedence 7317) ("}" 7318)) 3179 ((T_INSTEADOF 7502)) 3180 ((T_PAAMAYIM_NEKUDOTAYIM 7419)) 3181 ((T_ABSTRACT 7225) (T_ARRAY 7226) (T_AS 7227) (T_BREAK 7228) (T_CALLABLE 7229) (T_CASE 7230) (T_CATCH 7231) (T_CLASS 7232) (T_CLASS_C 7233) (T_CLONE 7234) (T_CONST 7235) (T_CONTINUE 7236) (T_DECLARE 7237) (T_DEFAULT 7238) (T_DIR 7239) (T_DO 7240) (T_ECHO 7241) (T_ELSE 7242) (T_ELSEIF 7243) (T_EMPTY 7244) (T_ENDDECLARE 7245) (T_ENDFOR 7246) (T_ENDFOREACH 7247) (T_ENDIF 7248) (T_ENDSWITCH 7249) (T_ENDWHILE 7250) (T_ENUM 7251) (T_EVAL 7252) (T_EXIT 7253) (T_EXTENDS 7254) (T_FILE 7255) (T_FINAL 7256) (T_FINALLY 7257) (T_FN 7258) (T_FOR 7259) (T_FOREACH 7260) (T_FUNCTION 7261) (T_FUNC_C 7262) (T_GLOBAL 7263) (T_GOTO 7264) (T_IF 7265) (T_IMPLEMENTS 7266) (T_INCLUDE 7267) (T_INCLUDE_ONCE 7268) (T_INSTANCEOF 7269) (T_INSTEADOF 7270) (T_INTERFACE 7271) (T_ISSET 7272) (T_LINE 7273) (T_LIST 7274) (T_LOGICAL_AND 7275) (T_LOGICAL_OR 7276) (T_LOGICAL_XOR 7277) (T_MATCH 7278) (T_METHOD_C 7279) (T_NAMESPACE 7280) (T_NAME_FULLY_QUALIFIED 7281) (T_NAME_QUALIFIED 7282) (T_NAME_RELATIVE 7283) (T_NEW 7284) (T_NS_C 7285) (T_PRINT 7286) (T_PRIVATE 7287) (T_PROPERTY_C 7288) (T_PROTECTED 7289) (T_PUBLIC 7290) (T_READONLY 7291) (T_REQUIRE 7292) (T_REQUIRE_ONCE 7293) (T_RETURN 7294) (T_STATIC 7295) (T_STRING 7296) (T_SWITCH 7297) (T_THROW 7298) (T_TRAIT 7299) (T_TRAIT_C 7300) (T_TRY 7301) (T_UNSET 7302) (T_USE 7303) (T_VAR 7304) (T_WHILE 7305) (T_YIELD 7306) (absolute_trait_method_reference 7307) (class_name 7308) (identifier 7309) (name 7310) (reserved_non_modifiers 7311) (semi_reserved 7312) (trait_adaptation 7417) (trait_alias 7315) (trait_method_reference 7316) (trait_precedence 7317) ("}" 7418)) 3182 ((";" 7416)) 3183 ((T_AS 7320)) 3184 ((";" 7319)) 3185 ((T_ABSTRACT 7321) (T_ARRAY 7322) (T_AS 7323) (T_BREAK 7324) (T_CALLABLE 7325) (T_CASE 7326) (T_CATCH 7327) (T_CLASS 7328) (T_CLASS_C 7329) (T_CLONE 7330) (T_CONST 7331) (T_CONTINUE 7332) (T_DECLARE 7333) (T_DEFAULT 7334) (T_DIR 7335) (T_DO 7336) (T_ECHO 7337) (T_ELSE 7338) (T_ELSEIF 7339) (T_EMPTY 7340) (T_ENDDECLARE 7341) (T_ENDFOR 7342) (T_ENDFOREACH 7343) (T_ENDIF 7344) (T_ENDSWITCH 7345) (T_ENDWHILE 7346) (T_ENUM 7347) (T_EVAL 7348) (T_EXIT 7349) (T_EXTENDS 7350) (T_FILE 7351) (T_FINAL 7352) (T_FINALLY 7353) (T_FN 7354) (T_FOR 7355) (T_FOREACH 7356) (T_FUNCTION 7357) (T_FUNC_C 7358) (T_GLOBAL 7359) (T_GOTO 7360) (T_IF 7361) (T_IMPLEMENTS 7362) (T_INCLUDE 7363) (T_INCLUDE_ONCE 7364) (T_INSTANCEOF 7365) (T_INSTEADOF 7366) (T_INTERFACE 7367) (T_ISSET 7368) (T_LINE 7369) (T_LIST 7370) (T_LOGICAL_AND 7371) (T_LOGICAL_OR 7372) (T_LOGICAL_XOR 7373) (T_MATCH 7374) (T_METHOD_C 7375) (T_NAMESPACE 7376) (T_NEW 7377) (T_NS_C 7378) (T_PRINT 7379) (T_PRIVATE 7380) (T_PRIVATE_SET 7381) (T_PROPERTY_C 7382) (T_PROTECTED 7383) (T_PROTECTED_SET 7384) (T_PUBLIC 7385) (T_PUBLIC_SET 7386) (T_READONLY 7387) (T_REQUIRE 7388) (T_REQUIRE_ONCE 7389) (T_RETURN 7390) (T_STATIC 7391) (T_STRING 7392) (T_SWITCH 7393) (T_THROW 7394) (T_TRAIT 7395) (T_TRAIT_C 7396) (T_TRY 7397) (T_UNSET 7398) (T_USE 7399) (T_VAR 7400) (T_WHILE 7401) (T_YIELD 7402) (member_modifier 7403) (reserved_non_modifiers 7404)) 3186 ((T_ABSTRACT 7405) (T_ARRAY 7322) (T_AS 7323) (T_BREAK 7324) (T_CALLABLE 7325) (T_CASE 7326) (T_CATCH 7327) (T_CLASS 7328) (T_CLASS_C 7329) (T_CLONE 7330) (T_CONST 7331) (T_CONTINUE 7332) (T_DECLARE 7333) (T_DEFAULT 7334) (T_DIR 7335) (T_DO 7336) (T_ECHO 7337) (T_ELSE 7338) (T_ELSEIF 7339) (T_EMPTY 7340) (T_ENDDECLARE 7341) (T_ENDFOR 7342) (T_ENDFOREACH 7343) (T_ENDIF 7344) (T_ENDSWITCH 7345) (T_ENDWHILE 7346) (T_ENUM 7347) (T_EVAL 7348) (T_EXIT 7349) (T_EXTENDS 7350) (T_FILE 7351) (T_FINAL 7406) (T_FINALLY 7353) (T_FN 7354) (T_FOR 7355) (T_FOREACH 7356) (T_FUNCTION 7357) (T_FUNC_C 7358) (T_GLOBAL 7359) (T_GOTO 7360) (T_IF 7361) (T_IMPLEMENTS 7362) (T_INCLUDE 7363) (T_INCLUDE_ONCE 7364) (T_INSTANCEOF 7365) (T_INSTEADOF 7366) (T_INTERFACE 7367) (T_ISSET 7368) (T_LINE 7369) (T_LIST 7370) (T_LOGICAL_AND 7371) (T_LOGICAL_OR 7372) (T_LOGICAL_XOR 7373) (T_MATCH 7374) (T_METHOD_C 7375) (T_NAMESPACE 7376) (T_NEW 7377) (T_NS_C 7378) (T_PRINT 7379) (T_PRIVATE 7407) (T_PROPERTY_C 7382) (T_PROTECTED 7408) (T_PUBLIC 7409) (T_READONLY 7410) (T_REQUIRE 7388) (T_REQUIRE_ONCE 7389) (T_RETURN 7390) (T_STATIC 7411) (T_STRING 7412) (T_SWITCH 7393) (T_THROW 7394) (T_TRAIT 7395) (T_TRAIT_C 7396) (T_TRY 7397) (T_UNSET 7398) (T_USE 7399) (T_VAR 7400) (T_WHILE 7401) (T_YIELD 7402) (identifier 7413) (reserved_non_modifiers 7414) (semi_reserved 7415)) 3187 ((T_ABSTRACT 7420) (T_ARRAY 7421) (T_AS 7422) (T_BREAK 7423) (T_CALLABLE 7424) (T_CASE 7425) (T_CATCH 7426) (T_CLASS 7427) (T_CLASS_C 7428) (T_CLONE 7429) (T_CONST 7430) (T_CONTINUE 7431) (T_DECLARE 7432) (T_DEFAULT 7433) (T_DIR 7434) (T_DO 7435) (T_ECHO 7436) (T_ELSE 7437) (T_ELSEIF 7438) (T_EMPTY 7439) (T_ENDDECLARE 7440) (T_ENDFOR 7441) (T_ENDFOREACH 7442) (T_ENDIF 7443) (T_ENDSWITCH 7444) (T_ENDWHILE 7445) (T_ENUM 7446) (T_EVAL 7447) (T_EXIT 7448) (T_EXTENDS 7449) (T_FILE 7450) (T_FINAL 7451) (T_FINALLY 7452) (T_FN 7453) (T_FOR 7454) (T_FOREACH 7455) (T_FUNCTION 7456) (T_FUNC_C 7457) (T_GLOBAL 7458) (T_GOTO 7459) (T_IF 7460) (T_IMPLEMENTS 7461) (T_INCLUDE 7462) (T_INCLUDE_ONCE 7463) (T_INSTANCEOF 7464) (T_INSTEADOF 7465) (T_INTERFACE 7466) (T_ISSET 7467) (T_LINE 7468) (T_LIST 7469) (T_LOGICAL_AND 7470) (T_LOGICAL_OR 7471) (T_LOGICAL_XOR 7472) (T_MATCH 7473) (T_METHOD_C 7474) (T_NAMESPACE 7475) (T_NEW 7476) (T_NS_C 7477) (T_PRINT 7478) (T_PRIVATE 7479) (T_PROPERTY_C 7480) (T_PROTECTED 7481) (T_PUBLIC 7482) (T_READONLY 7483) (T_REQUIRE 7484) (T_REQUIRE_ONCE 7485) (T_RETURN 7486) (T_STATIC 7487) (T_STRING 7488) (T_SWITCH 7489) (T_THROW 7490) (T_TRAIT 7491) (T_TRAIT_C 7492) (T_TRY 7493) (T_UNSET 7494) (T_USE 7495) (T_VAR 7496) (T_WHILE 7497) (T_YIELD 7498) (identifier 7499) (reserved_non_modifiers 7500) (semi_reserved 7501)) 3188 ((T_NAME_FULLY_QUALIFIED 7503) (T_NAME_QUALIFIED 7504) (T_NAME_RELATIVE 7505) (T_STATIC 7506) (T_STRING 7507) (class_name 7508) (class_name_list 7509) (name 7510)) 3189 (("," 7511)) 3190 ((T_NAME_FULLY_QUALIFIED 7503) (T_NAME_QUALIFIED 7504) (T_NAME_RELATIVE 7505) (T_STATIC 7506) (T_STRING 7507) (class_name 7512) (name 7510)) 3191 ((T_ABSTRACT 7515) (T_ARRAY 7516) (T_AS 7517) (T_BREAK 7518) (T_CALLABLE 7519) (T_CASE 7520) (T_CATCH 7521) (T_CLASS 7522) (T_CLASS_C 7523) (T_CLONE 7524) (T_CONST 7525) (T_CONTINUE 7526) (T_DECLARE 7527) (T_DEFAULT 7528) (T_DIR 7529) (T_DO 7530) (T_ECHO 7531) (T_ELSE 7532) (T_ELSEIF 7533) (T_EMPTY 7534) (T_ENDDECLARE 7535) (T_ENDFOR 7536) (T_ENDFOREACH 7537) (T_ENDIF 7538) (T_ENDSWITCH 7539) (T_ENDWHILE 7540) (T_ENUM 7541) (T_EVAL 7542) (T_EXIT 7543) (T_EXTENDS 7544) (T_FILE 7545) (T_FINAL 7546) (T_FINALLY 7547) (T_FN 7548) (T_FOR 7549) (T_FOREACH 7550) (T_FUNCTION 7551) (T_FUNC_C 7552) (T_GLOBAL 7553) (T_GOTO 7554) (T_IF 7555) (T_IMPLEMENTS 7556) (T_INCLUDE 7557) (T_INCLUDE_ONCE 7558) (T_INSTANCEOF 7559) (T_INSTEADOF 7560) (T_INTERFACE 7561) (T_ISSET 7562) (T_LINE 7563) (T_LIST 7564) (T_LOGICAL_AND 7565) (T_LOGICAL_OR 7566) (T_LOGICAL_XOR 7567) (T_MATCH 7568) (T_METHOD_C 7569) (T_NAMESPACE 7570) (T_NEW 7571) (T_NS_C 7572) (T_PRINT 7573) (T_PRIVATE 7574) (T_PROPERTY_C 7575) (T_PROTECTED 7576) (T_PUBLIC 7577) (T_READONLY 7578) (T_REQUIRE 7579) (T_REQUIRE_ONCE 7580) (T_RETURN 7581) (T_STATIC 7582) (T_STRING 7583) (T_SWITCH 7584) (T_THROW 7585) (T_TRAIT 7586) (T_TRAIT_C 7587) (T_TRY 7588) (T_UNSET 7589) (T_USE 7590) (T_VAR 7591) (T_WHILE 7592) (T_YIELD 7593) (identifier 7594) (reserved_non_modifiers 7595) (semi_reserved 7596)) 3192 (("=" 7597) (enum_case_expr 7598)) 3193 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 7600) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 3194 ((";" 7599)) 3195 (("," 5847) (possible_comma 7602)) 3196 (("]" 7603)) 3197 (("," 7612)) 3198 ((T_NAME_FULLY_QUALIFIED 7604) (T_NAME_QUALIFIED 7605) (T_NAME_RELATIVE 7606) (T_STATIC 7607) (T_STRING 7608) (class_name 7613) (name 7611)) 3199 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 7626) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3200 (("," 1480) (possible_comma 7624)) 3201 ((")" 7625)) 3202 (("," 5847) (possible_comma 7629)) 3203 (("]" 7630)) 3204 (("%" 299) (")" 7632) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3205 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 7634) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3206 (("%" 299) (")" 7635) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3207 (("{" 7636)) 3208 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 7637) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 3209 (("}" 7638)) 3210 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 7640) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 3211 ((")" 7641)) 3212 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 7643) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3213 (("," 5192) (possible_comma 7644)) 3214 ((")" 7645)) 3215 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 7653) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 7654) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 7655) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3216 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 7658) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3217 (("," 1480) (possible_comma 7656)) 3218 ((")" 7657)) 3219 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 7660) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3220 (("%" 299) (")" 7661) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3221 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 7663) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3222 (("%" 299) (")" 7664) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3223 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 7671) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 3224 ((")" 7672)) 3225 (("%" 299) (")" 7677) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3226 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 7680) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3227 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 7681)) 3228 (("\"" 7683) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 3229 (("$" 1865) (T_ABSTRACT 2054) (T_ARRAY 2055) (T_AS 2056) (T_BREAK 2057) (T_CALLABLE 2058) (T_CASE 2059) (T_CATCH 2060) (T_CLASS 2061) (T_CLASS_C 2062) (T_CLONE 2063) (T_CONST 2064) (T_CONTINUE 2065) (T_DECLARE 2066) (T_DEFAULT 2067) (T_DIR 2068) (T_DO 2069) (T_ECHO 2070) (T_ELSE 2071) (T_ELSEIF 2072) (T_EMPTY 2073) (T_ENDDECLARE 2074) (T_ENDFOR 2075) (T_ENDFOREACH 2076) (T_ENDIF 2077) (T_ENDSWITCH 2078) (T_ENDWHILE 2079) (T_ENUM 2080) (T_EVAL 2081) (T_EXIT 2082) (T_EXTENDS 2083) (T_FILE 2084) (T_FINAL 2085) (T_FINALLY 2086) (T_FN 2087) (T_FOR 2088) (T_FOREACH 2089) (T_FUNCTION 2090) (T_FUNC_C 2091) (T_GLOBAL 2092) (T_GOTO 2093) (T_IF 2094) (T_IMPLEMENTS 2095) (T_INCLUDE 2096) (T_INCLUDE_ONCE 2097) (T_INSTANCEOF 2098) (T_INSTEADOF 2099) (T_INTERFACE 2100) (T_ISSET 2101) (T_LINE 2102) (T_LIST 2103) (T_LOGICAL_AND 2104) (T_LOGICAL_OR 2105) (T_LOGICAL_XOR 2106) (T_MATCH 2107) (T_METHOD_C 2108) (T_NAMESPACE 2109) (T_NEW 2110) (T_NS_C 2111) (T_PRINT 2112) (T_PRIVATE 2113) (T_PROPERTY_C 2114) (T_PROTECTED 2115) (T_PUBLIC 2116) (T_READONLY 2117) (T_REQUIRE 2118) (T_REQUIRE_ONCE 2119) (T_RETURN 2120) (T_STATIC 2121) (T_STRING 2122) (T_SWITCH 2123) (T_THROW 2124) (T_TRAIT 2125) (T_TRAIT_C 2126) (T_TRY 2127) (T_UNSET 2128) (T_USE 2129) (T_VAR 2130) (T_VARIABLE 1914) (T_WHILE 2131) (T_YIELD 2132) (identifier 7686) (member_name 7687) (reserved_non_modifiers 2135) (semi_reserved 2136) (simple_variable 7688) ("{" 7689)) 3230 (("(" 2141) (argument_list 7692)) 3231 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 7690) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3232 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 7691)) 3233 ((T_FN 35) (T_FUNCTION 129) (fn 1931) #'1932 (inline_function 7696)) 3234 (("=" 7698)) 3235 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 7699) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3236 (("$" 1865) (T_STRING 7705) (T_VARIABLE 1914) (property_name 7712) (simple_variable 7707) ("{" 7708)) 3237 (("$" 1865) (T_STRING 7705) (T_VARIABLE 1914) (property_name 7706) (simple_variable 7707) ("{" 7708)) 3238 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 7703) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3239 (("]" 7704)) 3240 (("(" 2141) (argument_list 7711)) 3241 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 7709) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3242 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 7710)) 3243 (("(" 2141) (argument_list 7713)) 3244 (("`" 7716)) 3245 (("]" 7718)) 3246 (("=" 7719)) 3247 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7720) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 3248 (("%" 1950) ("*" 1951) ("+" 1952) ("-" 1953) ("." 1954) ("/" 1955) ("<" 1956) (">" 1957) ("?" 1958) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1959) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1960) (T_BOOLEAN_AND 1961) (T_BOOLEAN_OR 1962) (T_COALESCE 1963) (T_DOUBLE_ARROW 7723) (T_INSTANCEOF 1964) (T_IS_EQUAL 1965) (T_IS_GREATER_OR_EQUAL 1966) (T_IS_IDENTICAL 1967) (T_IS_NOT_EQUAL 1968) (T_IS_NOT_IDENTICAL 1969) (T_IS_SMALLER_OR_EQUAL 1970) (T_LOGICAL_AND 1971) (T_LOGICAL_OR 1972) (T_LOGICAL_XOR 1973) (T_POW 1974) (T_SL 1975) (T_SPACESHIP 1976) (T_SR 1977) ("^" 1978) ("|" 1979)) 3249 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1948) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (array_object_dereferenceable 1920) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 7724) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 3250 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 7733) (T_VARIABLE 2896) (encaps_var 2910)) 3251 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 7732) (T_VARIABLE 2896) (encaps_var 2901)) 3252 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 7774) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3253 ((T_CLASS 7766)) 3254 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 7765) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 7746) (attribute 2958) (class_modifier 2939)) 3255 ((T_PAAMAYIM_NEKUDOTAYIM 7763)) 3256 (("(" 2196) (argument_list 7762)) 3257 ((T_NULLSAFE_OBJECT_OPERATOR 7753) (T_OBJECT_OPERATOR 7754) (T_PAAMAYIM_NEKUDOTAYIM 7755) ("[" 7756)) 3258 (("$" 2180) (T_STRING 2215) (T_VARIABLE 2182) (property_name 7761) (simple_variable 2217) ("{" 2218)) 3259 (("$" 2180) (T_STRING 2215) (T_VARIABLE 2182) (property_name 7760) (simple_variable 2217) ("{" 2218)) 3260 (("$" 2180) (T_VARIABLE 2182) (simple_variable 7759)) 3261 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 7757) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3262 (("]" 7758)) 3263 (("$" 2180) (T_VARIABLE 2182) (simple_variable 7764)) 3264 (("(" 2960) (argument_list 2961) (ctor_arguments 7767)) 3265 ((T_EXTENDS 2963) (extends_from 7768)) 3266 ((T_IMPLEMENTS 2965) (implements_list 7769)) 3267 ((backup_doc_comment 7770)) 3268 (("{" 7771)) 3269 ((class_statement_list 7772)) 3270 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 7773)) 3271 (("%" 299) (")" 7775) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3272 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 7777) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3273 (("%" 299) (")" 7778) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3274 (("{" 7779)) 3275 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 7780) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 3276 (("}" 7781)) 3277 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 7783) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 3278 ((")" 7784)) 3279 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 7786) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3280 (("," 5192) (possible_comma 7787)) 3281 ((")" 7788)) 3282 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 7796) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 7797) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 7798) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3283 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 7801) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3284 (("," 1480) (possible_comma 7799)) 3285 ((")" 7800)) 3286 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 7803) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3287 (("%" 299) (")" 7804) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3288 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 7806) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3289 (("%" 299) (")" 7807) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3290 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 7815) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 3291 ((")" 7816)) 3292 (("%" 299) (")" 7821) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3293 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 7824) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3294 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 7825)) 3295 (("\"" 7827) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 3296 (("$" 1679) (T_ABSTRACT 7830) (T_ARRAY 7831) (T_AS 7832) (T_BREAK 7833) (T_CALLABLE 7834) (T_CASE 7835) (T_CATCH 7836) (T_CLASS 7837) (T_CLASS_C 7838) (T_CLONE 7839) (T_CONST 7840) (T_CONTINUE 7841) (T_DECLARE 7842) (T_DEFAULT 7843) (T_DIR 7844) (T_DO 7845) (T_ECHO 7846) (T_ELSE 7847) (T_ELSEIF 7848) (T_EMPTY 7849) (T_ENDDECLARE 7850) (T_ENDFOR 7851) (T_ENDFOREACH 7852) (T_ENDIF 7853) (T_ENDSWITCH 7854) (T_ENDWHILE 7855) (T_ENUM 7856) (T_EVAL 7857) (T_EXIT 7858) (T_EXTENDS 7859) (T_FILE 7860) (T_FINAL 7861) (T_FINALLY 7862) (T_FN 7863) (T_FOR 7864) (T_FOREACH 7865) (T_FUNCTION 7866) (T_FUNC_C 7867) (T_GLOBAL 7868) (T_GOTO 7869) (T_IF 7870) (T_IMPLEMENTS 7871) (T_INCLUDE 7872) (T_INCLUDE_ONCE 7873) (T_INSTANCEOF 7874) (T_INSTEADOF 7875) (T_INTERFACE 7876) (T_ISSET 7877) (T_LINE 7878) (T_LIST 7879) (T_LOGICAL_AND 7880) (T_LOGICAL_OR 7881) (T_LOGICAL_XOR 7882) (T_MATCH 7883) (T_METHOD_C 7884) (T_NAMESPACE 7885) (T_NEW 7886) (T_NS_C 7887) (T_PRINT 7888) (T_PRIVATE 7889) (T_PROPERTY_C 7890) (T_PROTECTED 7891) (T_PUBLIC 7892) (T_READONLY 7893) (T_REQUIRE 7894) (T_REQUIRE_ONCE 7895) (T_RETURN 7896) (T_STATIC 7897) (T_STRING 7898) (T_SWITCH 7899) (T_THROW 7900) (T_TRAIT 7901) (T_TRAIT_C 7902) (T_TRY 7903) (T_UNSET 7904) (T_USE 7905) (T_VAR 7906) (T_VARIABLE 1728) (T_WHILE 7907) (T_YIELD 7908) (identifier 7909) (member_name 7910) (reserved_non_modifiers 7911) (semi_reserved 7912) (simple_variable 7913) ("{" 7914)) 3297 (("(" 7917) (argument_list 7918)) 3298 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 7915) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3299 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 7916)) 3300 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 7919) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 7920) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 7921) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3301 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 7924) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3302 (("," 1480) (possible_comma 7922)) 3303 ((")" 7923)) 3304 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 7954) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 7955) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3305 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 7953) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3306 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 7952) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3307 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 7951) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3308 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 7950) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3309 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 7949) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3310 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 7948) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3311 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 7947) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3312 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 7946) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3313 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 7945) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3314 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 7944) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3315 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 7943) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3316 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 7942) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3317 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 7941) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3318 (("\"" 1520) ("$" 7956) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 7957) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 7958) ("[" 1542) (array_object_dereferenceable 7959) (callable_expr 7960) (callable_variable 7961) (class_constant 1546) (class_name 7962) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 7963) (name 7964) (new_dereferenceable 1552) (simple_variable 7965) (static_member 7966) (variable 7967) (variable_class_name 7968)) 3319 (("$" 7956) (T_VARIABLE 7958) (simple_variable 8001) ("{" 8002)) 3320 (("(" 7972) (argument_list 8000)) 3321 ((T_NULLSAFE_OBJECT_OPERATOR 7986) (T_OBJECT_OPERATOR 7987) ("[" 7988)) 3322 (("(" 7972) (argument_list 7985)) 3323 ((T_PAAMAYIM_NEKUDOTAYIM 7981)) 3324 (("(" 7972) (argument_list 7980)) 3325 ((T_PAAMAYIM_NEKUDOTAYIM 7969)) 3326 (("$" 7956) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 7958) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 7970) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 7971) ("{" 1642)) 3327 (("(" 7972) (argument_list 7973)) 3328 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 7974) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 7975) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 7976) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3329 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 7979) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3330 (("," 1480) (possible_comma 7977)) 3331 ((")" 7978)) 3332 (("$" 7956) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 7958) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 7982) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 7983) ("{" 1658)) 3333 (("(" 7972) (argument_list 7984)) 3334 (("$" 7956) (T_STRING 7991) (T_VARIABLE 7958) (property_name 7998) (simple_variable 7993) ("{" 7994)) 3335 (("$" 7956) (T_STRING 7991) (T_VARIABLE 7958) (property_name 7992) (simple_variable 7993) ("{" 7994)) 3336 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 7989) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3337 (("]" 7990)) 3338 (("(" 7972) (argument_list 7997)) 3339 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 7995) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3340 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 7996)) 3341 (("(" 7972) (argument_list 7999)) 3342 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 8003) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3343 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 8004)) 3344 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (possible_array_pair 8006) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3345 ((backup_doc_comment 8009)) 3346 (("(" 8010)) 3347 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 8011)) 3348 ((")" 8012)) 3349 ((T_USE 164) (lexical_vars 8013)) 3350 ((":" 166) (return_type 8014)) 3351 ((backup_fn_flags 8015)) 3352 (("{" 8016)) 3353 ((inner_statement_list 8017)) 3354 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 8018) ("~" 127)) 3355 ((backup_fn_flags 8019)) 3356 ((backup_doc_comment 8021)) 3357 (("(" 8022)) 3358 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 8023)) 3359 ((")" 8024)) 3360 ((":" 2249) (return_type 8025)) 3361 ((T_DOUBLE_ARROW 8026)) 3362 ((backup_fn_flags 8027)) 3363 ((backup_lex_pos 8028)) 3364 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8092) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3365 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8548) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3366 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 8546) (encaps_var 5237)) 3367 (("$" 8031) (T_VARIABLE 8079) (simple_variable 8542) ("{" 8543)) 3368 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 8540) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3369 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8539) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3370 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8538) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3371 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8537) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3372 (("(" 8534)) 3373 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8533) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3374 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8532) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3375 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8531) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3376 (("\"" 1520) ("$" 8334) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 8335) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 8336) ("[" 1542) (array_object_dereferenceable 8337) (callable_expr 8338) (callable_variable 8339) (class_constant 1546) (class_name 8340) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 8341) (name 8342) (new_dereferenceable 1552) (simple_variable 8343) (static_member 8344) (variable 8530) (variable_class_name 8346)) 3377 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8529) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3378 (("(" 8526)) 3379 (("(" 8523)) 3380 (("(" 8514) (argument_list 8515) (ctor_arguments 8516)) 3381 (("\"" 1520) ("$" 8334) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 8335) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 8336) ("[" 1542) (array_object_dereferenceable 8337) (callable_expr 8338) (callable_variable 8339) (class_constant 1546) (class_name 8340) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 8341) (name 8342) (new_dereferenceable 1552) (simple_variable 8343) (static_member 8344) (variable 8513) (variable_class_name 8346)) 3382 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8512) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3383 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8511) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3384 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8510) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3385 (("(" 8506)) 3386 (("(" 8501)) 3387 (("(" 8495)) 3388 (("$" 8334) ("(" 8458) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 8459) (T_NAME_QUALIFIED 8460) (T_NAME_RELATIVE 8461) (T_READONLY 2931) (T_STATIC 8462) (T_STRING 8463) (T_VARIABLE 8336) (anonymous_class 8464) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 8465) (attribute 2937) (attributes 8466) (class_modifier 2939) (class_name 8467) (class_name_reference 8468) (name 8469) (new_variable 8470) (simple_variable 8471)) 3389 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8457) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3390 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8456) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3391 (("(" 8295) (argument_list 8455)) 3392 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8454) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3393 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8453) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3394 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 8448) (T_END_HEREDOC 8449) (T_VARIABLE 2896) (encaps_list 8450) (encaps_var 2898)) 3395 ((T_FN 35) (T_FUNCTION 129) (fn 8093) #'8094 (inline_function 8447)) 3396 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8446) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3397 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8445) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3398 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8444) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3399 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8441) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3400 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8440) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3401 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 8436) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3402 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 8434) (encaps_list 2804) (encaps_var 2805)) 3403 ((T_NULLSAFE_OBJECT_OPERATOR 8420) (T_OBJECT_OPERATOR 8421) ("[" 8422)) 3404 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 8417) (attribute 2733) (fn 8093) #'8094 (inline_function 8418)) 3405 (("(" 8295) (argument_list 8416)) 3406 ((T_PAAMAYIM_NEKUDOTAYIM 8408)) 3407 (("%" 8109) ("*" 8110) ("+" 8111) ("-" 8112) ("." 8113) ("/" 8114) ("<" 8115) (">" 8116) ("?" 8117) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 8118) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 8119) (T_BOOLEAN_AND 8120) (T_BOOLEAN_OR 8121) (T_COALESCE 8122) (T_INSTANCEOF 8123) (T_IS_EQUAL 8124) (T_IS_GREATER_OR_EQUAL 8125) (T_IS_IDENTICAL 8126) (T_IS_NOT_EQUAL 8127) (T_IS_NOT_IDENTICAL 8128) (T_IS_SMALLER_OR_EQUAL 8129) (T_LOGICAL_AND 8130) (T_LOGICAL_OR 8131) (T_LOGICAL_XOR 8132) (T_POW 8133) (T_SL 8134) (T_SPACESHIP 8135) (T_SR 8136) ("^" 8137) (backup_fn_flags 8407) ("|" 8138)) 3408 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 8396)) 3409 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 8384)) 3410 (("(" 8295) (argument_list 8383)) 3411 (("=" 8303) (T_AND_EQUAL 8304) (T_COALESCE_EQUAL 8305) (T_CONCAT_EQUAL 8306) (T_DEC 8307) (T_DIV_EQUAL 8308) (T_INC 8309) (T_MINUS_EQUAL 8310) (T_MOD_EQUAL 8311) (T_MUL_EQUAL 8312) (T_OR_EQUAL 8313) (T_PLUS_EQUAL 8314) (T_POW_EQUAL 8315) (T_SL_EQUAL 8316) (T_SR_EQUAL 8317) (T_XOR_EQUAL 8318)) 3412 ((T_PAAMAYIM_NEKUDOTAYIM 8207)) 3413 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8108) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3414 (("%" 8109) ("*" 8110) ("+" 8111) ("-" 8112) ("." 8113) ("/" 8114) ("<" 8115) (">" 8116) ("?" 8117) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 8118) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 8119) (T_BOOLEAN_AND 8120) (T_BOOLEAN_OR 8121) (T_COALESCE 8122) (T_INSTANCEOF 8123) (T_IS_EQUAL 8124) (T_IS_GREATER_OR_EQUAL 8125) (T_IS_IDENTICAL 8126) (T_IS_NOT_EQUAL 8127) (T_IS_NOT_IDENTICAL 8128) (T_IS_SMALLER_OR_EQUAL 8129) (T_LOGICAL_AND 8130) (T_LOGICAL_OR 8131) (T_LOGICAL_XOR 8132) (T_POW 8133) (T_SL 8134) (T_SPACESHIP 8135) (T_SR 8136) ("^" 8137) ("|" 8138)) 3415 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8206) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3416 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8205) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3417 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8204) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3418 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8203) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3419 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8202) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3420 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8201) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3421 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8200) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3422 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8199) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3423 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 8194) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 8195) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 3424 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8193) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3425 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8192) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3426 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8191) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3427 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8190) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3428 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8189) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3429 (("$" 8154) ("(" 8155) (T_NAME_FULLY_QUALIFIED 8156) (T_NAME_QUALIFIED 8157) (T_NAME_RELATIVE 8158) (T_STATIC 8159) (T_STRING 8160) (T_VARIABLE 8161) (class_name 8162) (class_name_reference 8163) (name 8164) (new_variable 8165) (simple_variable 8166)) 3430 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8153) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3431 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8152) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3432 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8151) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3433 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8150) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3434 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8149) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3435 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8148) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3436 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8147) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3437 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8146) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3438 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8145) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3439 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8144) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3440 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8143) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3441 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8142) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3442 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8141) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3443 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8140) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3444 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8139) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3445 (("$" 8154) (T_VARIABLE 8161) (simple_variable 8185) ("{" 8186)) 3446 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 8183) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3447 ((T_PAAMAYIM_NEKUDOTAYIM 8181)) 3448 ((T_NULLSAFE_OBJECT_OPERATOR 8167) (T_OBJECT_OPERATOR 8168) (T_PAAMAYIM_NEKUDOTAYIM 8169) ("[" 8170)) 3449 (("$" 8154) (T_STRING 8174) (T_VARIABLE 8161) (property_name 8180) (simple_variable 8176) ("{" 8177)) 3450 (("$" 8154) (T_STRING 8174) (T_VARIABLE 8161) (property_name 8175) (simple_variable 8176) ("{" 8177)) 3451 (("$" 8154) (T_VARIABLE 8161) (simple_variable 8173)) 3452 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 8171) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3453 (("]" 8172)) 3454 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 8178) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3455 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 8179)) 3456 (("$" 8154) (T_VARIABLE 8161) (simple_variable 8182)) 3457 (("%" 299) (")" 8184) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3458 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 8187) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3459 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 8188)) 3460 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8198) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3461 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 8196) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 3462 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8197) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3463 (("$" 8031) (T_ABSTRACT 8208) (T_ARRAY 8209) (T_AS 8210) (T_BREAK 8211) (T_CALLABLE 8212) (T_CASE 8213) (T_CATCH 8214) (T_CLASS 8215) (T_CLASS_C 8216) (T_CLONE 8217) (T_CONST 8218) (T_CONTINUE 8219) (T_DECLARE 8220) (T_DEFAULT 8221) (T_DIR 8222) (T_DO 8223) (T_ECHO 8224) (T_ELSE 8225) (T_ELSEIF 8226) (T_EMPTY 8227) (T_ENDDECLARE 8228) (T_ENDFOR 8229) (T_ENDFOREACH 8230) (T_ENDIF 8231) (T_ENDSWITCH 8232) (T_ENDWHILE 8233) (T_ENUM 8234) (T_EVAL 8235) (T_EXIT 8236) (T_EXTENDS 8237) (T_FILE 8238) (T_FINAL 8239) (T_FINALLY 8240) (T_FN 8241) (T_FOR 8242) (T_FOREACH 8243) (T_FUNCTION 8244) (T_FUNC_C 8245) (T_GLOBAL 8246) (T_GOTO 8247) (T_IF 8248) (T_IMPLEMENTS 8249) (T_INCLUDE 8250) (T_INCLUDE_ONCE 8251) (T_INSTANCEOF 8252) (T_INSTEADOF 8253) (T_INTERFACE 8254) (T_ISSET 8255) (T_LINE 8256) (T_LIST 8257) (T_LOGICAL_AND 8258) (T_LOGICAL_OR 8259) (T_LOGICAL_XOR 8260) (T_MATCH 8261) (T_METHOD_C 8262) (T_NAMESPACE 8263) (T_NEW 8264) (T_NS_C 8265) (T_PRINT 8266) (T_PRIVATE 8267) (T_PROPERTY_C 8268) (T_PROTECTED 8269) (T_PUBLIC 8270) (T_READONLY 8271) (T_REQUIRE 8272) (T_REQUIRE_ONCE 8273) (T_RETURN 8274) (T_STATIC 8275) (T_STRING 8276) (T_SWITCH 8277) (T_THROW 8278) (T_TRAIT 8279) (T_TRAIT_C 8280) (T_TRY 8281) (T_UNSET 8282) (T_USE 8283) (T_VAR 8284) (T_VARIABLE 8079) (T_WHILE 8285) (T_YIELD 8286) (identifier 8287) (member_name 8288) (reserved_non_modifiers 8289) (semi_reserved 8290) (simple_variable 8291) ("{" 8292)) 3464 (("(" 8295) (argument_list 8296)) 3465 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 8293) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3466 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 8294)) 3467 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 8297) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 8298) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 8299) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3468 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 8302) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3469 (("," 1480) (possible_comma 8300)) 3470 ((")" 8301)) 3471 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (ampersand 8332) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8333) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3472 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8331) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3473 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8330) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3474 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8329) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3475 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8328) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3476 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8327) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3477 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8326) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3478 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8325) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3479 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8324) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3480 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8323) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3481 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8322) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3482 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8321) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3483 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8320) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3484 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8319) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3485 (("\"" 1520) ("$" 8334) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 8335) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 8336) ("[" 1542) (array_object_dereferenceable 8337) (callable_expr 8338) (callable_variable 8339) (class_constant 1546) (class_name 8340) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 8341) (name 8342) (new_dereferenceable 1552) (simple_variable 8343) (static_member 8344) (variable 8345) (variable_class_name 8346)) 3486 (("$" 8334) (T_VARIABLE 8336) (simple_variable 8379) ("{" 8380)) 3487 (("(" 8350) (argument_list 8378)) 3488 ((T_NULLSAFE_OBJECT_OPERATOR 8364) (T_OBJECT_OPERATOR 8365) ("[" 8366)) 3489 (("(" 8350) (argument_list 8363)) 3490 ((T_PAAMAYIM_NEKUDOTAYIM 8359)) 3491 (("(" 8350) (argument_list 8358)) 3492 ((T_PAAMAYIM_NEKUDOTAYIM 8347)) 3493 (("$" 8334) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 8336) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 8348) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 8349) ("{" 1642)) 3494 (("(" 8350) (argument_list 8351)) 3495 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 8352) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 8353) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 8354) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3496 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 8357) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3497 (("," 1480) (possible_comma 8355)) 3498 ((")" 8356)) 3499 (("$" 8334) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 8336) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 8360) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 8361) ("{" 1658)) 3500 (("(" 8350) (argument_list 8362)) 3501 (("$" 8334) (T_STRING 8369) (T_VARIABLE 8336) (property_name 8376) (simple_variable 8371) ("{" 8372)) 3502 (("$" 8334) (T_STRING 8369) (T_VARIABLE 8336) (property_name 8370) (simple_variable 8371) ("{" 8372)) 3503 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 8367) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3504 (("]" 8368)) 3505 (("(" 8350) (argument_list 8375)) 3506 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 8373) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3507 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 8374)) 3508 (("(" 8350) (argument_list 8377)) 3509 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 8381) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3510 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 8382)) 3511 ((backup_doc_comment 8385)) 3512 (("(" 8386)) 3513 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 8387)) 3514 ((")" 8388)) 3515 ((T_USE 164) (lexical_vars 8389)) 3516 ((":" 166) (return_type 8390)) 3517 ((backup_fn_flags 8391)) 3518 (("{" 8392)) 3519 ((inner_statement_list 8393)) 3520 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 8394) ("~" 127)) 3521 ((backup_fn_flags 8395)) 3522 ((backup_doc_comment 8397)) 3523 (("(" 8398)) 3524 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 8399)) 3525 ((")" 8400)) 3526 ((":" 2249) (return_type 8401)) 3527 ((T_DOUBLE_ARROW 8402)) 3528 ((backup_fn_flags 8403)) 3529 ((backup_lex_pos 8404)) 3530 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8405) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3531 (("%" 8109) ("*" 8110) ("+" 8111) ("-" 8112) ("." 8113) ("/" 8114) ("<" 8115) (">" 8116) ("?" 8117) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 8118) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 8119) (T_BOOLEAN_AND 8120) (T_BOOLEAN_OR 8121) (T_COALESCE 8122) (T_INSTANCEOF 8123) (T_IS_EQUAL 8124) (T_IS_GREATER_OR_EQUAL 8125) (T_IS_IDENTICAL 8126) (T_IS_NOT_EQUAL 8127) (T_IS_NOT_IDENTICAL 8128) (T_IS_SMALLER_OR_EQUAL 8129) (T_LOGICAL_AND 8130) (T_LOGICAL_OR 8131) (T_LOGICAL_XOR 8132) (T_POW 8133) (T_SL 8134) (T_SPACESHIP 8135) (T_SR 8136) ("^" 8137) (backup_fn_flags 8406) ("|" 8138)) 3532 (("$" 8031) (T_ABSTRACT 8208) (T_ARRAY 8209) (T_AS 8210) (T_BREAK 8211) (T_CALLABLE 8212) (T_CASE 8213) (T_CATCH 8214) (T_CLASS 8215) (T_CLASS_C 8216) (T_CLONE 8217) (T_CONST 8218) (T_CONTINUE 8219) (T_DECLARE 8220) (T_DEFAULT 8221) (T_DIR 8222) (T_DO 8223) (T_ECHO 8224) (T_ELSE 8225) (T_ELSEIF 8226) (T_EMPTY 8227) (T_ENDDECLARE 8228) (T_ENDFOR 8229) (T_ENDFOREACH 8230) (T_ENDIF 8231) (T_ENDSWITCH 8232) (T_ENDWHILE 8233) (T_ENUM 8234) (T_EVAL 8235) (T_EXIT 8236) (T_EXTENDS 8237) (T_FILE 8238) (T_FINAL 8239) (T_FINALLY 8240) (T_FN 8241) (T_FOR 8242) (T_FOREACH 8243) (T_FUNCTION 8244) (T_FUNC_C 8245) (T_GLOBAL 8246) (T_GOTO 8247) (T_IF 8248) (T_IMPLEMENTS 8249) (T_INCLUDE 8250) (T_INCLUDE_ONCE 8251) (T_INSTANCEOF 8252) (T_INSTEADOF 8253) (T_INTERFACE 8254) (T_ISSET 8255) (T_LINE 8256) (T_LIST 8257) (T_LOGICAL_AND 8258) (T_LOGICAL_OR 8259) (T_LOGICAL_XOR 8260) (T_MATCH 8261) (T_METHOD_C 8262) (T_NAMESPACE 8263) (T_NEW 8264) (T_NS_C 8265) (T_PRINT 8266) (T_PRIVATE 8267) (T_PROPERTY_C 8268) (T_PROTECTED 8269) (T_PUBLIC 8270) (T_READONLY 8271) (T_REQUIRE 8272) (T_REQUIRE_ONCE 8273) (T_RETURN 8274) (T_STATIC 8275) (T_STRING 8276) (T_SWITCH 8277) (T_THROW 8278) (T_TRAIT 8279) (T_TRAIT_C 8280) (T_TRY 8281) (T_UNSET 8282) (T_USE 8283) (T_VAR 8284) (T_VARIABLE 8079) (T_WHILE 8285) (T_YIELD 8286) (identifier 8409) (member_name 8410) (reserved_non_modifiers 8289) (semi_reserved 8290) (simple_variable 8411) ("{" 8412)) 3533 (("(" 8295) (argument_list 8415)) 3534 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 8413) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3535 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 8414)) 3536 ((T_FN 35) (T_FUNCTION 129) (fn 8093) #'8094 (inline_function 8419)) 3537 (("$" 8031) (T_STRING 8425) (T_VARIABLE 8079) (property_name 8432) (simple_variable 8427) ("{" 8428)) 3538 (("$" 8031) (T_STRING 8425) (T_VARIABLE 8079) (property_name 8426) (simple_variable 8427) ("{" 8428)) 3539 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 8423) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3540 (("]" 8424)) 3541 (("(" 8295) (argument_list 8431)) 3542 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 8429) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3543 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 8430)) 3544 (("(" 8295) (argument_list 8433)) 3545 (("`" 8435)) 3546 (("]" 8437)) 3547 (("=" 8438)) 3548 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8439) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3549 (("%" 8109) ("*" 8110) ("+" 8111) ("-" 8112) ("." 8113) ("/" 8114) ("<" 8115) (">" 8116) ("?" 8117) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 8118) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 8119) (T_BOOLEAN_AND 8120) (T_BOOLEAN_OR 8121) (T_COALESCE 8122) (T_DOUBLE_ARROW 8442) (T_INSTANCEOF 8123) (T_IS_EQUAL 8124) (T_IS_GREATER_OR_EQUAL 8125) (T_IS_IDENTICAL 8126) (T_IS_NOT_EQUAL 8127) (T_IS_NOT_IDENTICAL 8128) (T_IS_SMALLER_OR_EQUAL 8129) (T_LOGICAL_AND 8130) (T_LOGICAL_OR 8131) (T_LOGICAL_XOR 8132) (T_POW 8133) (T_SL 8134) (T_SPACESHIP 8135) (T_SR 8136) ("^" 8137) ("|" 8138)) 3550 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8443) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3551 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 8452) (T_VARIABLE 2896) (encaps_var 2910)) 3552 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 8451) (T_VARIABLE 2896) (encaps_var 2901)) 3553 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 8493) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3554 ((T_CLASS 8485)) 3555 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 8484) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 8465) (attribute 2958) (class_modifier 2939)) 3556 ((T_PAAMAYIM_NEKUDOTAYIM 8482)) 3557 (("(" 8350) (argument_list 8481)) 3558 ((T_NULLSAFE_OBJECT_OPERATOR 8472) (T_OBJECT_OPERATOR 8473) (T_PAAMAYIM_NEKUDOTAYIM 8474) ("[" 8475)) 3559 (("$" 8334) (T_STRING 8369) (T_VARIABLE 8336) (property_name 8480) (simple_variable 8371) ("{" 8372)) 3560 (("$" 8334) (T_STRING 8369) (T_VARIABLE 8336) (property_name 8479) (simple_variable 8371) ("{" 8372)) 3561 (("$" 8334) (T_VARIABLE 8336) (simple_variable 8478)) 3562 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 8476) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3563 (("]" 8477)) 3564 (("$" 8334) (T_VARIABLE 8336) (simple_variable 8483)) 3565 (("(" 2960) (argument_list 2961) (ctor_arguments 8486)) 3566 ((T_EXTENDS 2963) (extends_from 8487)) 3567 ((T_IMPLEMENTS 2965) (implements_list 8488)) 3568 ((backup_doc_comment 8489)) 3569 (("{" 8490)) 3570 ((class_statement_list 8491)) 3571 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 8492)) 3572 (("%" 299) (")" 8494) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3573 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 8496) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3574 (("%" 299) (")" 8497) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3575 (("{" 8498)) 3576 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 8499) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 3577 (("}" 8500)) 3578 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 8502) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 3579 ((")" 8503)) 3580 (("=" 8504)) 3581 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 8505) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3582 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 8507) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3583 (("," 5192) (possible_comma 8508)) 3584 ((")" 8509)) 3585 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 8517) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 8518) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 8519) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3586 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 8522) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3587 (("," 1480) (possible_comma 8520)) 3588 ((")" 8521)) 3589 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 8524) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3590 (("%" 299) (")" 8525) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3591 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 8527) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3592 (("%" 299) (")" 8528) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3593 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 8535) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 3594 ((")" 8536)) 3595 (("%" 299) (")" 8541) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3596 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 8544) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3597 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 8545)) 3598 (("\"" 8547) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 3599 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8578) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (ampersand 8605) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8614) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3600 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9640) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3601 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 9638) (encaps_var 5237)) 3602 (("$" 8552) (T_VARIABLE 8600) (simple_variable 9634) ("{" 9635)) 3603 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9632) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3604 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9631) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3605 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9630) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3606 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9629) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3607 (("(" 9626)) 3608 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9625) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3609 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9624) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3610 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9623) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3611 (("\"" 1520) ("$" 8862) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 8863) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 8864) ("[" 1542) (array_object_dereferenceable 8865) (callable_expr 8866) (callable_variable 8867) (class_constant 1546) (class_name 8868) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 8869) (name 8870) (new_dereferenceable 1552) (simple_variable 8871) (static_member 8872) (variable 9622) (variable_class_name 8874)) 3612 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9621) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3613 (("(" 9618)) 3614 (("(" 9615)) 3615 (("(" 9606) (argument_list 9607) (ctor_arguments 9608)) 3616 (("\"" 1520) ("$" 8862) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 8863) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 8864) ("[" 1542) (array_object_dereferenceable 8865) (callable_expr 8866) (callable_variable 8867) (class_constant 1546) (class_name 8868) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 8869) (name 8870) (new_dereferenceable 1552) (simple_variable 8871) (static_member 8872) (variable 9605) (variable_class_name 8874)) 3617 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9604) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3618 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9603) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3619 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9602) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3620 (("(" 9598)) 3621 (("(" 9595)) 3622 (("(" 9589)) 3623 (("$" 8862) ("(" 9552) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 9553) (T_NAME_QUALIFIED 9554) (T_NAME_RELATIVE 9555) (T_READONLY 2931) (T_STATIC 9556) (T_STRING 9557) (T_VARIABLE 8864) (anonymous_class 9558) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 9559) (attribute 2937) (attributes 9560) (class_modifier 2939) (class_name 9561) (class_name_reference 9562) (name 9563) (new_variable 9564) (simple_variable 9565)) 3624 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9551) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3625 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9550) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3626 (("(" 8823) (argument_list 9549)) 3627 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9548) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3628 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9547) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3629 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 9542) (T_END_HEREDOC 9543) (T_VARIABLE 2896) (encaps_list 9544) (encaps_var 2898)) 3630 ((T_FN 35) (T_FUNCTION 129) (fn 8615) #'8616 (inline_function 9541)) 3631 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9540) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3632 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9539) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3633 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9538) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3634 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9535) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3635 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9534) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3636 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 9530) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3637 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 9528) (encaps_list 2804) (encaps_var 2805)) 3638 (("\"" 1520) ("$" 9479) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 9480) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 9481) ("[" 1542) (array_object_dereferenceable 9482) (callable_expr 9483) (callable_variable 9484) (class_constant 1546) (class_name 9485) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 9486) (name 9487) (new_dereferenceable 1552) (simple_variable 9488) (static_member 9489) (variable 9490) (variable_class_name 9491)) 3639 ((T_NULLSAFE_OBJECT_OPERATOR 9465) (T_OBJECT_OPERATOR 9466) ("[" 9467)) 3640 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 9462) (attribute 2733) (fn 8615) #'8616 (inline_function 9463)) 3641 (("(" 8823) (argument_list 9461)) 3642 ((T_PAAMAYIM_NEKUDOTAYIM 9453)) 3643 (("%" 8632) ("*" 8633) ("+" 8634) ("-" 8635) ("." 8636) ("/" 8637) ("<" 8638) (">" 8639) ("?" 8640) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 8641) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 8642) (T_BOOLEAN_AND 8643) (T_BOOLEAN_OR 8644) (T_COALESCE 8645) (T_INSTANCEOF 8646) (T_IS_EQUAL 8647) (T_IS_GREATER_OR_EQUAL 8648) (T_IS_IDENTICAL 8649) (T_IS_NOT_EQUAL 8650) (T_IS_NOT_IDENTICAL 8651) (T_IS_SMALLER_OR_EQUAL 8652) (T_LOGICAL_AND 8653) (T_LOGICAL_OR 8654) (T_LOGICAL_XOR 8655) (T_POW 8656) (T_SL 8657) (T_SPACESHIP 8658) (T_SR 8659) ("^" 8660) ("|" 8661)) 3644 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 8924)) 3645 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 8912)) 3646 (("(" 8823) (argument_list 8911)) 3647 (("=" 8831) (T_AND_EQUAL 8832) (T_COALESCE_EQUAL 8833) (T_CONCAT_EQUAL 8834) (T_DEC 8835) (T_DIV_EQUAL 8836) (T_INC 8837) (T_MINUS_EQUAL 8838) (T_MOD_EQUAL 8839) (T_MUL_EQUAL 8840) (T_OR_EQUAL 8841) (T_PLUS_EQUAL 8842) (T_POW_EQUAL 8843) (T_SL_EQUAL 8844) (T_SR_EQUAL 8845) (T_XOR_EQUAL 8846)) 3648 ((T_PAAMAYIM_NEKUDOTAYIM 8735)) 3649 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8631) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3650 (("(" 8730)) 3651 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8729) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3652 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8728) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3653 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8727) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3654 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8726) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3655 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8725) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3656 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8724) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3657 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8723) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3658 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8722) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3659 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 8717) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 8718) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 3660 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8716) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3661 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8715) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3662 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8714) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3663 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8713) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3664 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8712) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3665 (("$" 8677) ("(" 8678) (T_NAME_FULLY_QUALIFIED 8679) (T_NAME_QUALIFIED 8680) (T_NAME_RELATIVE 8681) (T_STATIC 8682) (T_STRING 8683) (T_VARIABLE 8684) (class_name 8685) (class_name_reference 8686) (name 8687) (new_variable 8688) (simple_variable 8689)) 3666 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8676) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3667 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8675) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3668 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8674) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3669 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8673) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3670 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8672) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3671 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8671) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3672 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8670) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3673 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8669) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3674 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8668) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3675 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8667) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3676 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8666) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3677 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8665) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3678 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8664) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3679 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8663) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3680 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8662) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3681 (("$" 8677) (T_VARIABLE 8684) (simple_variable 8708) ("{" 8709)) 3682 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 8706) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3683 ((T_PAAMAYIM_NEKUDOTAYIM 8704)) 3684 ((T_NULLSAFE_OBJECT_OPERATOR 8690) (T_OBJECT_OPERATOR 8691) (T_PAAMAYIM_NEKUDOTAYIM 8692) ("[" 8693)) 3685 (("$" 8677) (T_STRING 8697) (T_VARIABLE 8684) (property_name 8703) (simple_variable 8699) ("{" 8700)) 3686 (("$" 8677) (T_STRING 8697) (T_VARIABLE 8684) (property_name 8698) (simple_variable 8699) ("{" 8700)) 3687 (("$" 8677) (T_VARIABLE 8684) (simple_variable 8696)) 3688 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 8694) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3689 (("]" 8695)) 3690 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 8701) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3691 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 8702)) 3692 (("$" 8677) (T_VARIABLE 8684) (simple_variable 8705)) 3693 (("%" 299) (")" 8707) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3694 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 8710) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3695 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 8711)) 3696 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8721) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3697 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 8719) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 3698 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8720) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3699 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 8731) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 3700 ((")" 8732)) 3701 (("=" 8733)) 3702 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8734) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3703 (("$" 8552) (T_ABSTRACT 8736) (T_ARRAY 8737) (T_AS 8738) (T_BREAK 8739) (T_CALLABLE 8740) (T_CASE 8741) (T_CATCH 8742) (T_CLASS 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONST 8746) (T_CONTINUE 8747) (T_DECLARE 8748) (T_DEFAULT 8749) (T_DIR 8750) (T_DO 8751) (T_ECHO 8752) (T_ELSE 8753) (T_ELSEIF 8754) (T_EMPTY 8755) (T_ENDDECLARE 8756) (T_ENDFOR 8757) (T_ENDFOREACH 8758) (T_ENDIF 8759) (T_ENDSWITCH 8760) (T_ENDWHILE 8761) (T_ENUM 8762) (T_EVAL 8763) (T_EXIT 8764) (T_EXTENDS 8765) (T_FILE 8766) (T_FINAL 8767) (T_FINALLY 8768) (T_FN 8769) (T_FOR 8770) (T_FOREACH 8771) (T_FUNCTION 8772) (T_FUNC_C 8773) (T_GLOBAL 8774) (T_GOTO 8775) (T_IF 8776) (T_IMPLEMENTS 8777) (T_INCLUDE 8778) (T_INCLUDE_ONCE 8779) (T_INSTANCEOF 8780) (T_INSTEADOF 8781) (T_INTERFACE 8782) (T_ISSET 8783) (T_LINE 8784) (T_LIST 8785) (T_LOGICAL_AND 8786) (T_LOGICAL_OR 8787) (T_LOGICAL_XOR 8788) (T_MATCH 8789) (T_METHOD_C 8790) (T_NAMESPACE 8791) (T_NEW 8792) (T_NS_C 8793) (T_PRINT 8794) (T_PRIVATE 8795) (T_PROPERTY_C 8796) (T_PROTECTED 8797) (T_PUBLIC 8798) (T_READONLY 8799) (T_REQUIRE 8800) (T_REQUIRE_ONCE 8801) (T_RETURN 8802) (T_STATIC 8803) (T_STRING 8804) (T_SWITCH 8805) (T_THROW 8806) (T_TRAIT 8807) (T_TRAIT_C 8808) (T_TRY 8809) (T_UNSET 8810) (T_USE 8811) (T_VAR 8812) (T_VARIABLE 8600) (T_WHILE 8813) (T_YIELD 8814) (identifier 8815) (member_name 8816) (reserved_non_modifiers 8817) (semi_reserved 8818) (simple_variable 8819) ("{" 8820)) 3704 (("(" 8823) (argument_list 8824)) 3705 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 8821) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3706 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 8822)) 3707 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 8825) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 8826) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 8827) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3708 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 8830) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3709 (("," 1480) (possible_comma 8828)) 3710 ((")" 8829)) 3711 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (ampersand 8860) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8861) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3712 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8859) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3713 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8858) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3714 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8857) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3715 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8856) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3716 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8855) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3717 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8854) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3718 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8853) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3719 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8852) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3720 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8851) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3721 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8850) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3722 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8849) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3723 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8848) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3724 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 8847) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 3725 (("\"" 1520) ("$" 8862) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 8863) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 8864) ("[" 1542) (array_object_dereferenceable 8865) (callable_expr 8866) (callable_variable 8867) (class_constant 1546) (class_name 8868) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 8869) (name 8870) (new_dereferenceable 1552) (simple_variable 8871) (static_member 8872) (variable 8873) (variable_class_name 8874)) 3726 (("$" 8862) (T_VARIABLE 8864) (simple_variable 8907) ("{" 8908)) 3727 (("(" 8878) (argument_list 8906)) 3728 ((T_NULLSAFE_OBJECT_OPERATOR 8892) (T_OBJECT_OPERATOR 8893) ("[" 8894)) 3729 (("(" 8878) (argument_list 8891)) 3730 ((T_PAAMAYIM_NEKUDOTAYIM 8887)) 3731 (("(" 8878) (argument_list 8886)) 3732 ((T_PAAMAYIM_NEKUDOTAYIM 8875)) 3733 (("$" 8862) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 8864) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 8876) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 8877) ("{" 1642)) 3734 (("(" 8878) (argument_list 8879)) 3735 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 8880) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 8881) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 8882) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3736 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 8885) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3737 (("," 1480) (possible_comma 8883)) 3738 ((")" 8884)) 3739 (("$" 8862) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 8864) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 8888) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 8889) ("{" 1658)) 3740 (("(" 8878) (argument_list 8890)) 3741 (("$" 8862) (T_STRING 8897) (T_VARIABLE 8864) (property_name 8904) (simple_variable 8899) ("{" 8900)) 3742 (("$" 8862) (T_STRING 8897) (T_VARIABLE 8864) (property_name 8898) (simple_variable 8899) ("{" 8900)) 3743 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 8895) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3744 (("]" 8896)) 3745 (("(" 8878) (argument_list 8903)) 3746 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 8901) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3747 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 8902)) 3748 (("(" 8878) (argument_list 8905)) 3749 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 8909) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3750 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 8910)) 3751 ((backup_doc_comment 8913)) 3752 (("(" 8914)) 3753 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 8915)) 3754 ((")" 8916)) 3755 ((T_USE 164) (lexical_vars 8917)) 3756 ((":" 166) (return_type 8918)) 3757 ((backup_fn_flags 8919)) 3758 (("{" 8920)) 3759 ((inner_statement_list 8921)) 3760 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 8922) ("~" 127)) 3761 ((backup_fn_flags 8923)) 3762 ((backup_doc_comment 8925)) 3763 (("(" 8926)) 3764 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 8927)) 3765 ((")" 8928)) 3766 ((":" 2249) (return_type 8929)) 3767 ((T_DOUBLE_ARROW 8930)) 3768 ((backup_fn_flags 8931)) 3769 ((backup_lex_pos 8932)) 3770 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 8996) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3771 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9452) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3772 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 9450) (encaps_var 5237)) 3773 (("$" 8935) (T_VARIABLE 8983) (simple_variable 9446) ("{" 9447)) 3774 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9444) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3775 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9443) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3776 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9442) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3777 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9441) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3778 (("(" 9438)) 3779 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9437) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3780 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9436) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3781 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9435) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3782 (("\"" 1520) ("$" 9238) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 9239) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 9240) ("[" 1542) (array_object_dereferenceable 9241) (callable_expr 9242) (callable_variable 9243) (class_constant 1546) (class_name 9244) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 9245) (name 9246) (new_dereferenceable 1552) (simple_variable 9247) (static_member 9248) (variable 9434) (variable_class_name 9250)) 3783 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9433) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3784 (("(" 9430)) 3785 (("(" 9427)) 3786 (("(" 9418) (argument_list 9419) (ctor_arguments 9420)) 3787 (("\"" 1520) ("$" 9238) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 9239) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 9240) ("[" 1542) (array_object_dereferenceable 9241) (callable_expr 9242) (callable_variable 9243) (class_constant 1546) (class_name 9244) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 9245) (name 9246) (new_dereferenceable 1552) (simple_variable 9247) (static_member 9248) (variable 9417) (variable_class_name 9250)) 3788 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9416) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3789 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9415) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3790 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9414) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3791 (("(" 9410)) 3792 (("(" 9405)) 3793 (("(" 9399)) 3794 (("$" 9238) ("(" 9362) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 9363) (T_NAME_QUALIFIED 9364) (T_NAME_RELATIVE 9365) (T_READONLY 2931) (T_STATIC 9366) (T_STRING 9367) (T_VARIABLE 9240) (anonymous_class 9368) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 9369) (attribute 2937) (attributes 9370) (class_modifier 2939) (class_name 9371) (class_name_reference 9372) (name 9373) (new_variable 9374) (simple_variable 9375)) 3795 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9361) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3796 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9360) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3797 (("(" 9199) (argument_list 9359)) 3798 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9358) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3799 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9357) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3800 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 9352) (T_END_HEREDOC 9353) (T_VARIABLE 2896) (encaps_list 9354) (encaps_var 2898)) 3801 ((T_FN 35) (T_FUNCTION 129) (fn 8997) #'8998 (inline_function 9351)) 3802 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9350) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3803 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9349) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3804 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9348) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3805 (("!" 8029) ("\"" 8030) ("$" 8031) ("(" 8032) ("+" 8033) ("-" 8034) ("@" 8035) (T_ARRAY 8036) (T_ARRAY_CAST 8037) (T_ATTRIBUTE 128) (T_BOOL_CAST 8038) (T_CLASS_C 8039) (T_CLONE 8040) (T_CONSTANT_ENCAPSED_STRING 8041) (T_DEC 8042) (T_DIR 8043) (T_DNUMBER 8044) (T_DOUBLE_CAST 8045) (T_EMPTY 8046) (T_EVAL 8047) (T_EXIT 8048) (T_FILE 8049) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8050) (T_INC 8051) (T_INCLUDE 8052) (T_INCLUDE_ONCE 8053) (T_INT_CAST 8054) (T_ISSET 8055) (T_LINE 8056) (T_LIST 8057) (T_LNUMBER 8058) (T_MATCH 8059) (T_METHOD_C 8060) (T_NAME_FULLY_QUALIFIED 8061) (T_NAME_QUALIFIED 8062) (T_NAME_RELATIVE 8063) (T_NEW 8064) (T_NS_C 8065) (T_OBJECT_CAST 8066) (T_PRINT 8067) (T_PROPERTY_C 8068) (T_READONLY 8069) (T_REQUIRE 8070) (T_REQUIRE_ONCE 8071) (T_START_HEREDOC 8072) (T_STATIC 8073) (T_STRING 8074) (T_STRING_CAST 8075) (T_THROW 8076) (T_TRAIT_C 8077) (T_UNSET_CAST 8078) (T_VARIABLE 8079) (T_YIELD 8080) (T_YIELD_FROM 8081) ("[" 8082) ("`" 8083) (array_object_dereferenceable 8084) (attribute 133) (attributes 8085) (callable_expr 8086) (callable_variable 8087) (class_constant 8088) (class_name 8089) (constant 8090) (dereferenceable_scalar 8091) (expr 9345) (fn 8093) (fully_dereferenceable 105) #'8094 (function_call 8095) (inline_function 8096) (internal_functions_in_yacc 8097) (match 8098) (name 8099) (new_dereferenceable 8100) (new_non_dereferenceable 8101) (scalar 8102) (simple_variable 8103) (static_member 8104) (variable 8105) (variable_class_name 8106) ("~" 8107)) 3806 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9344) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3807 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 9340) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 3808 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 9338) (encaps_list 2804) (encaps_var 2805)) 3809 ((T_NULLSAFE_OBJECT_OPERATOR 9324) (T_OBJECT_OPERATOR 9325) ("[" 9326)) 3810 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 9321) (attribute 2733) (fn 8997) #'8998 (inline_function 9322)) 3811 (("(" 9199) (argument_list 9320)) 3812 ((T_PAAMAYIM_NEKUDOTAYIM 9312)) 3813 (("%" 9013) ("*" 9014) ("+" 9015) ("-" 9016) ("." 9017) ("/" 9018) ("<" 9019) (">" 9020) ("?" 9021) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 9022) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 9023) (T_BOOLEAN_AND 9024) (T_BOOLEAN_OR 9025) (T_COALESCE 9026) (T_INSTANCEOF 9027) (T_IS_EQUAL 9028) (T_IS_GREATER_OR_EQUAL 9029) (T_IS_IDENTICAL 9030) (T_IS_NOT_EQUAL 9031) (T_IS_NOT_IDENTICAL 9032) (T_IS_SMALLER_OR_EQUAL 9033) (T_LOGICAL_AND 9034) (T_LOGICAL_OR 9035) (T_LOGICAL_XOR 9036) (T_POW 9037) (T_SL 9038) (T_SPACESHIP 9039) (T_SR 9040) ("^" 9041) (backup_fn_flags 9311) ("|" 9042)) 3814 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 9300)) 3815 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 9288)) 3816 (("(" 9199) (argument_list 9287)) 3817 (("=" 9207) (T_AND_EQUAL 9208) (T_COALESCE_EQUAL 9209) (T_CONCAT_EQUAL 9210) (T_DEC 9211) (T_DIV_EQUAL 9212) (T_INC 9213) (T_MINUS_EQUAL 9214) (T_MOD_EQUAL 9215) (T_MUL_EQUAL 9216) (T_OR_EQUAL 9217) (T_PLUS_EQUAL 9218) (T_POW_EQUAL 9219) (T_SL_EQUAL 9220) (T_SR_EQUAL 9221) (T_XOR_EQUAL 9222)) 3818 ((T_PAAMAYIM_NEKUDOTAYIM 9111)) 3819 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9012) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3820 (("%" 9013) ("*" 9014) ("+" 9015) ("-" 9016) ("." 9017) ("/" 9018) ("<" 9019) (">" 9020) ("?" 9021) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 9022) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 9023) (T_BOOLEAN_AND 9024) (T_BOOLEAN_OR 9025) (T_COALESCE 9026) (T_INSTANCEOF 9027) (T_IS_EQUAL 9028) (T_IS_GREATER_OR_EQUAL 9029) (T_IS_IDENTICAL 9030) (T_IS_NOT_EQUAL 9031) (T_IS_NOT_IDENTICAL 9032) (T_IS_SMALLER_OR_EQUAL 9033) (T_LOGICAL_AND 9034) (T_LOGICAL_OR 9035) (T_LOGICAL_XOR 9036) (T_POW 9037) (T_SL 9038) (T_SPACESHIP 9039) (T_SR 9040) ("^" 9041) ("|" 9042)) 3821 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9110) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3822 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9109) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3823 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9108) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3824 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9107) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3825 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9106) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3826 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9105) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3827 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9104) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3828 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9103) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3829 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 9098) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 9099) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 3830 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9097) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3831 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9096) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3832 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9095) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3833 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9094) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3834 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9093) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3835 (("$" 9058) ("(" 9059) (T_NAME_FULLY_QUALIFIED 9060) (T_NAME_QUALIFIED 9061) (T_NAME_RELATIVE 9062) (T_STATIC 9063) (T_STRING 9064) (T_VARIABLE 9065) (class_name 9066) (class_name_reference 9067) (name 9068) (new_variable 9069) (simple_variable 9070)) 3836 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9057) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3837 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9056) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3838 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9055) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3839 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9054) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3840 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9053) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3841 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9052) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3842 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9051) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3843 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9050) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3844 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9049) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3845 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9048) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3846 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9047) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3847 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9046) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3848 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9045) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3849 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9044) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3850 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9043) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3851 (("$" 9058) (T_VARIABLE 9065) (simple_variable 9089) ("{" 9090)) 3852 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9087) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3853 ((T_PAAMAYIM_NEKUDOTAYIM 9085)) 3854 ((T_NULLSAFE_OBJECT_OPERATOR 9071) (T_OBJECT_OPERATOR 9072) (T_PAAMAYIM_NEKUDOTAYIM 9073) ("[" 9074)) 3855 (("$" 9058) (T_STRING 9078) (T_VARIABLE 9065) (property_name 9084) (simple_variable 9080) ("{" 9081)) 3856 (("$" 9058) (T_STRING 9078) (T_VARIABLE 9065) (property_name 9079) (simple_variable 9080) ("{" 9081)) 3857 (("$" 9058) (T_VARIABLE 9065) (simple_variable 9077)) 3858 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 9075) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3859 (("]" 9076)) 3860 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9082) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3861 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9083)) 3862 (("$" 9058) (T_VARIABLE 9065) (simple_variable 9086)) 3863 (("%" 299) (")" 9088) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3864 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9091) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3865 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9092)) 3866 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9102) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3867 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 9100) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 3868 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9101) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3869 (("$" 8935) (T_ABSTRACT 9112) (T_ARRAY 9113) (T_AS 9114) (T_BREAK 9115) (T_CALLABLE 9116) (T_CASE 9117) (T_CATCH 9118) (T_CLASS 9119) (T_CLASS_C 9120) (T_CLONE 9121) (T_CONST 9122) (T_CONTINUE 9123) (T_DECLARE 9124) (T_DEFAULT 9125) (T_DIR 9126) (T_DO 9127) (T_ECHO 9128) (T_ELSE 9129) (T_ELSEIF 9130) (T_EMPTY 9131) (T_ENDDECLARE 9132) (T_ENDFOR 9133) (T_ENDFOREACH 9134) (T_ENDIF 9135) (T_ENDSWITCH 9136) (T_ENDWHILE 9137) (T_ENUM 9138) (T_EVAL 9139) (T_EXIT 9140) (T_EXTENDS 9141) (T_FILE 9142) (T_FINAL 9143) (T_FINALLY 9144) (T_FN 9145) (T_FOR 9146) (T_FOREACH 9147) (T_FUNCTION 9148) (T_FUNC_C 9149) (T_GLOBAL 9150) (T_GOTO 9151) (T_IF 9152) (T_IMPLEMENTS 9153) (T_INCLUDE 9154) (T_INCLUDE_ONCE 9155) (T_INSTANCEOF 9156) (T_INSTEADOF 9157) (T_INTERFACE 9158) (T_ISSET 9159) (T_LINE 9160) (T_LIST 9161) (T_LOGICAL_AND 9162) (T_LOGICAL_OR 9163) (T_LOGICAL_XOR 9164) (T_MATCH 9165) (T_METHOD_C 9166) (T_NAMESPACE 9167) (T_NEW 9168) (T_NS_C 9169) (T_PRINT 9170) (T_PRIVATE 9171) (T_PROPERTY_C 9172) (T_PROTECTED 9173) (T_PUBLIC 9174) (T_READONLY 9175) (T_REQUIRE 9176) (T_REQUIRE_ONCE 9177) (T_RETURN 9178) (T_STATIC 9179) (T_STRING 9180) (T_SWITCH 9181) (T_THROW 9182) (T_TRAIT 9183) (T_TRAIT_C 9184) (T_TRY 9185) (T_UNSET 9186) (T_USE 9187) (T_VAR 9188) (T_VARIABLE 8983) (T_WHILE 9189) (T_YIELD 9190) (identifier 9191) (member_name 9192) (reserved_non_modifiers 9193) (semi_reserved 9194) (simple_variable 9195) ("{" 9196)) 3870 (("(" 9199) (argument_list 9200)) 3871 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9197) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3872 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9198)) 3873 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9201) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 9202) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 9203) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3874 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9206) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3875 (("," 1480) (possible_comma 9204)) 3876 ((")" 9205)) 3877 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (ampersand 9236) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9237) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3878 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9235) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3879 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9234) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3880 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9233) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3881 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9232) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3882 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9231) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3883 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9230) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3884 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9229) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3885 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9228) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3886 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9227) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3887 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9226) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3888 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9225) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3889 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9224) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3890 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9223) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3891 (("\"" 1520) ("$" 9238) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 9239) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 9240) ("[" 1542) (array_object_dereferenceable 9241) (callable_expr 9242) (callable_variable 9243) (class_constant 1546) (class_name 9244) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 9245) (name 9246) (new_dereferenceable 1552) (simple_variable 9247) (static_member 9248) (variable 9249) (variable_class_name 9250)) 3892 (("$" 9238) (T_VARIABLE 9240) (simple_variable 9283) ("{" 9284)) 3893 (("(" 9254) (argument_list 9282)) 3894 ((T_NULLSAFE_OBJECT_OPERATOR 9268) (T_OBJECT_OPERATOR 9269) ("[" 9270)) 3895 (("(" 9254) (argument_list 9267)) 3896 ((T_PAAMAYIM_NEKUDOTAYIM 9263)) 3897 (("(" 9254) (argument_list 9262)) 3898 ((T_PAAMAYIM_NEKUDOTAYIM 9251)) 3899 (("$" 9238) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 9240) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 9252) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 9253) ("{" 1642)) 3900 (("(" 9254) (argument_list 9255)) 3901 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9256) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 9257) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 9258) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3902 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9261) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3903 (("," 1480) (possible_comma 9259)) 3904 ((")" 9260)) 3905 (("$" 9238) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 9240) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 9264) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 9265) ("{" 1658)) 3906 (("(" 9254) (argument_list 9266)) 3907 (("$" 9238) (T_STRING 9273) (T_VARIABLE 9240) (property_name 9280) (simple_variable 9275) ("{" 9276)) 3908 (("$" 9238) (T_STRING 9273) (T_VARIABLE 9240) (property_name 9274) (simple_variable 9275) ("{" 9276)) 3909 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 9271) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3910 (("]" 9272)) 3911 (("(" 9254) (argument_list 9279)) 3912 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9277) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3913 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9278)) 3914 (("(" 9254) (argument_list 9281)) 3915 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9285) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3916 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9286)) 3917 ((backup_doc_comment 9289)) 3918 (("(" 9290)) 3919 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 9291)) 3920 ((")" 9292)) 3921 ((T_USE 164) (lexical_vars 9293)) 3922 ((":" 166) (return_type 9294)) 3923 ((backup_fn_flags 9295)) 3924 (("{" 9296)) 3925 ((inner_statement_list 9297)) 3926 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 9298) ("~" 127)) 3927 ((backup_fn_flags 9299)) 3928 ((backup_doc_comment 9301)) 3929 (("(" 9302)) 3930 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 9303)) 3931 ((")" 9304)) 3932 ((":" 2249) (return_type 9305)) 3933 ((T_DOUBLE_ARROW 9306)) 3934 ((backup_fn_flags 9307)) 3935 ((backup_lex_pos 9308)) 3936 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9309) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3937 (("%" 9013) ("*" 9014) ("+" 9015) ("-" 9016) ("." 9017) ("/" 9018) ("<" 9019) (">" 9020) ("?" 9021) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 9022) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 9023) (T_BOOLEAN_AND 9024) (T_BOOLEAN_OR 9025) (T_COALESCE 9026) (T_INSTANCEOF 9027) (T_IS_EQUAL 9028) (T_IS_GREATER_OR_EQUAL 9029) (T_IS_IDENTICAL 9030) (T_IS_NOT_EQUAL 9031) (T_IS_NOT_IDENTICAL 9032) (T_IS_SMALLER_OR_EQUAL 9033) (T_LOGICAL_AND 9034) (T_LOGICAL_OR 9035) (T_LOGICAL_XOR 9036) (T_POW 9037) (T_SL 9038) (T_SPACESHIP 9039) (T_SR 9040) ("^" 9041) (backup_fn_flags 9310) ("|" 9042)) 3938 (("$" 8935) (T_ABSTRACT 9112) (T_ARRAY 9113) (T_AS 9114) (T_BREAK 9115) (T_CALLABLE 9116) (T_CASE 9117) (T_CATCH 9118) (T_CLASS 9119) (T_CLASS_C 9120) (T_CLONE 9121) (T_CONST 9122) (T_CONTINUE 9123) (T_DECLARE 9124) (T_DEFAULT 9125) (T_DIR 9126) (T_DO 9127) (T_ECHO 9128) (T_ELSE 9129) (T_ELSEIF 9130) (T_EMPTY 9131) (T_ENDDECLARE 9132) (T_ENDFOR 9133) (T_ENDFOREACH 9134) (T_ENDIF 9135) (T_ENDSWITCH 9136) (T_ENDWHILE 9137) (T_ENUM 9138) (T_EVAL 9139) (T_EXIT 9140) (T_EXTENDS 9141) (T_FILE 9142) (T_FINAL 9143) (T_FINALLY 9144) (T_FN 9145) (T_FOR 9146) (T_FOREACH 9147) (T_FUNCTION 9148) (T_FUNC_C 9149) (T_GLOBAL 9150) (T_GOTO 9151) (T_IF 9152) (T_IMPLEMENTS 9153) (T_INCLUDE 9154) (T_INCLUDE_ONCE 9155) (T_INSTANCEOF 9156) (T_INSTEADOF 9157) (T_INTERFACE 9158) (T_ISSET 9159) (T_LINE 9160) (T_LIST 9161) (T_LOGICAL_AND 9162) (T_LOGICAL_OR 9163) (T_LOGICAL_XOR 9164) (T_MATCH 9165) (T_METHOD_C 9166) (T_NAMESPACE 9167) (T_NEW 9168) (T_NS_C 9169) (T_PRINT 9170) (T_PRIVATE 9171) (T_PROPERTY_C 9172) (T_PROTECTED 9173) (T_PUBLIC 9174) (T_READONLY 9175) (T_REQUIRE 9176) (T_REQUIRE_ONCE 9177) (T_RETURN 9178) (T_STATIC 9179) (T_STRING 9180) (T_SWITCH 9181) (T_THROW 9182) (T_TRAIT 9183) (T_TRAIT_C 9184) (T_TRY 9185) (T_UNSET 9186) (T_USE 9187) (T_VAR 9188) (T_VARIABLE 8983) (T_WHILE 9189) (T_YIELD 9190) (identifier 9313) (member_name 9314) (reserved_non_modifiers 9193) (semi_reserved 9194) (simple_variable 9315) ("{" 9316)) 3939 (("(" 9199) (argument_list 9319)) 3940 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9317) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3941 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9318)) 3942 ((T_FN 35) (T_FUNCTION 129) (fn 8997) #'8998 (inline_function 9323)) 3943 (("$" 8935) (T_STRING 9329) (T_VARIABLE 8983) (property_name 9336) (simple_variable 9331) ("{" 9332)) 3944 (("$" 8935) (T_STRING 9329) (T_VARIABLE 8983) (property_name 9330) (simple_variable 9331) ("{" 9332)) 3945 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 9327) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3946 (("]" 9328)) 3947 (("(" 9199) (argument_list 9335)) 3948 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9333) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 3949 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9334)) 3950 (("(" 9199) (argument_list 9337)) 3951 (("`" 9339)) 3952 (("]" 9341)) 3953 (("=" 9342)) 3954 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9343) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3955 (("%" 8109) ("*" 8110) ("+" 8111) ("-" 8112) ("." 8113) ("/" 8114) ("<" 8115) (">" 8116) ("?" 8117) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 8118) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 8119) (T_BOOLEAN_AND 8120) (T_BOOLEAN_OR 8121) (T_COALESCE 8122) (T_DOUBLE_ARROW 9346) (T_INSTANCEOF 8123) (T_IS_EQUAL 8124) (T_IS_GREATER_OR_EQUAL 8125) (T_IS_IDENTICAL 8126) (T_IS_NOT_EQUAL 8127) (T_IS_NOT_IDENTICAL 8128) (T_IS_SMALLER_OR_EQUAL 8129) (T_LOGICAL_AND 8130) (T_LOGICAL_OR 8131) (T_LOGICAL_XOR 8132) (T_POW 8133) (T_SL 8134) (T_SPACESHIP 8135) (T_SR 8136) ("^" 8137) ("|" 8138)) 3956 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9347) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3957 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 9356) (T_VARIABLE 2896) (encaps_var 2910)) 3958 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 9355) (T_VARIABLE 2896) (encaps_var 2901)) 3959 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9397) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3960 ((T_CLASS 9389)) 3961 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 9388) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 9369) (attribute 2958) (class_modifier 2939)) 3962 ((T_PAAMAYIM_NEKUDOTAYIM 9386)) 3963 (("(" 9254) (argument_list 9385)) 3964 ((T_NULLSAFE_OBJECT_OPERATOR 9376) (T_OBJECT_OPERATOR 9377) (T_PAAMAYIM_NEKUDOTAYIM 9378) ("[" 9379)) 3965 (("$" 9238) (T_STRING 9273) (T_VARIABLE 9240) (property_name 9384) (simple_variable 9275) ("{" 9276)) 3966 (("$" 9238) (T_STRING 9273) (T_VARIABLE 9240) (property_name 9383) (simple_variable 9275) ("{" 9276)) 3967 (("$" 9238) (T_VARIABLE 9240) (simple_variable 9382)) 3968 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 9380) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 3969 (("]" 9381)) 3970 (("$" 9238) (T_VARIABLE 9240) (simple_variable 9387)) 3971 (("(" 2960) (argument_list 2961) (ctor_arguments 9390)) 3972 ((T_EXTENDS 2963) (extends_from 9391)) 3973 ((T_IMPLEMENTS 2965) (implements_list 9392)) 3974 ((backup_doc_comment 9393)) 3975 (("{" 9394)) 3976 ((class_statement_list 9395)) 3977 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 9396)) 3978 (("%" 299) (")" 9398) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3979 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9400) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3980 (("%" 299) (")" 9401) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3981 (("{" 9402)) 3982 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 9403) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 3983 (("}" 9404)) 3984 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 9406) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 3985 ((")" 9407)) 3986 (("=" 9408)) 3987 (("!" 8933) ("\"" 8934) ("$" 8935) ("(" 8936) ("+" 8937) ("-" 8938) ("@" 8939) (T_ARRAY 8940) (T_ARRAY_CAST 8941) (T_ATTRIBUTE 128) (T_BOOL_CAST 8942) (T_CLASS_C 8943) (T_CLONE 8944) (T_CONSTANT_ENCAPSED_STRING 8945) (T_DEC 8946) (T_DIR 8947) (T_DNUMBER 8948) (T_DOUBLE_CAST 8949) (T_EMPTY 8950) (T_EVAL 8951) (T_EXIT 8952) (T_FILE 8953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8954) (T_INC 8955) (T_INCLUDE 8956) (T_INCLUDE_ONCE 8957) (T_INT_CAST 8958) (T_ISSET 8959) (T_LINE 8960) (T_LIST 8961) (T_LNUMBER 8962) (T_MATCH 8963) (T_METHOD_C 8964) (T_NAME_FULLY_QUALIFIED 8965) (T_NAME_QUALIFIED 8966) (T_NAME_RELATIVE 8967) (T_NEW 8968) (T_NS_C 8969) (T_OBJECT_CAST 8970) (T_PRINT 8971) (T_PROPERTY_C 8972) (T_READONLY 8973) (T_REQUIRE 8974) (T_REQUIRE_ONCE 8975) (T_START_HEREDOC 8976) (T_STATIC 8977) (T_STRING 8978) (T_STRING_CAST 8979) (T_THROW 8980) (T_TRAIT_C 8981) (T_UNSET_CAST 8982) (T_VARIABLE 8983) (T_YIELD 8984) (T_YIELD_FROM 8985) ("[" 8986) ("`" 8987) (array_object_dereferenceable 8988) (attribute 133) (attributes 8989) (callable_expr 8990) (callable_variable 8991) (class_constant 8992) (class_name 8993) (constant 8994) (dereferenceable_scalar 8995) (expr 9409) (fn 8997) (fully_dereferenceable 105) #'8998 (function_call 8999) (inline_function 9000) (internal_functions_in_yacc 9001) (match 9002) (name 9003) (new_dereferenceable 9004) (new_non_dereferenceable 9005) (scalar 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010) ("~" 9011)) 3988 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 9411) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3989 (("," 5192) (possible_comma 9412)) 3990 ((")" 9413)) 3991 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9421) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 9422) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 9423) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3992 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9426) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 3993 (("," 1480) (possible_comma 9424)) 3994 ((")" 9425)) 3995 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9428) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3996 (("%" 299) (")" 9429) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3997 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9431) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 3998 (("%" 299) (")" 9432) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 3999 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 9439) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 4000 ((")" 9440)) 4001 (("%" 299) (")" 9445) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4002 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9448) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4003 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9449)) 4004 (("\"" 9451) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 4005 (("$" 8552) (T_ABSTRACT 8736) (T_ARRAY 8737) (T_AS 8738) (T_BREAK 8739) (T_CALLABLE 8740) (T_CASE 8741) (T_CATCH 8742) (T_CLASS 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONST 8746) (T_CONTINUE 8747) (T_DECLARE 8748) (T_DEFAULT 8749) (T_DIR 8750) (T_DO 8751) (T_ECHO 8752) (T_ELSE 8753) (T_ELSEIF 8754) (T_EMPTY 8755) (T_ENDDECLARE 8756) (T_ENDFOR 8757) (T_ENDFOREACH 8758) (T_ENDIF 8759) (T_ENDSWITCH 8760) (T_ENDWHILE 8761) (T_ENUM 8762) (T_EVAL 8763) (T_EXIT 8764) (T_EXTENDS 8765) (T_FILE 8766) (T_FINAL 8767) (T_FINALLY 8768) (T_FN 8769) (T_FOR 8770) (T_FOREACH 8771) (T_FUNCTION 8772) (T_FUNC_C 8773) (T_GLOBAL 8774) (T_GOTO 8775) (T_IF 8776) (T_IMPLEMENTS 8777) (T_INCLUDE 8778) (T_INCLUDE_ONCE 8779) (T_INSTANCEOF 8780) (T_INSTEADOF 8781) (T_INTERFACE 8782) (T_ISSET 8783) (T_LINE 8784) (T_LIST 8785) (T_LOGICAL_AND 8786) (T_LOGICAL_OR 8787) (T_LOGICAL_XOR 8788) (T_MATCH 8789) (T_METHOD_C 8790) (T_NAMESPACE 8791) (T_NEW 8792) (T_NS_C 8793) (T_PRINT 8794) (T_PRIVATE 8795) (T_PROPERTY_C 8796) (T_PROTECTED 8797) (T_PUBLIC 8798) (T_READONLY 8799) (T_REQUIRE 8800) (T_REQUIRE_ONCE 8801) (T_RETURN 8802) (T_STATIC 8803) (T_STRING 8804) (T_SWITCH 8805) (T_THROW 8806) (T_TRAIT 8807) (T_TRAIT_C 8808) (T_TRY 8809) (T_UNSET 8810) (T_USE 8811) (T_VAR 8812) (T_VARIABLE 8600) (T_WHILE 8813) (T_YIELD 8814) (identifier 9454) (member_name 9455) (reserved_non_modifiers 8817) (semi_reserved 8818) (simple_variable 9456) ("{" 9457)) 4006 (("(" 8823) (argument_list 9460)) 4007 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9458) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4008 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9459)) 4009 ((T_FN 35) (T_FUNCTION 129) (fn 8615) #'8616 (inline_function 9464)) 4010 (("$" 8552) (T_STRING 9470) (T_VARIABLE 8600) (property_name 9477) (simple_variable 9472) ("{" 9473)) 4011 (("$" 8552) (T_STRING 9470) (T_VARIABLE 8600) (property_name 9471) (simple_variable 9472) ("{" 9473)) 4012 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 9468) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4013 (("]" 9469)) 4014 (("(" 8823) (argument_list 9476)) 4015 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9474) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4016 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9475)) 4017 (("(" 8823) (argument_list 9478)) 4018 (("$" 9479) (T_VARIABLE 9481) (simple_variable 9524) ("{" 9525)) 4019 (("(" 9495) (argument_list 9523)) 4020 ((T_NULLSAFE_OBJECT_OPERATOR 9509) (T_OBJECT_OPERATOR 9510) ("[" 9511)) 4021 (("(" 9495) (argument_list 9508)) 4022 ((T_PAAMAYIM_NEKUDOTAYIM 9504)) 4023 (("(" 9495) (argument_list 9503)) 4024 ((T_PAAMAYIM_NEKUDOTAYIM 9492)) 4025 (("$" 9479) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 9481) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 9493) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 9494) ("{" 1642)) 4026 (("(" 9495) (argument_list 9496)) 4027 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9497) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 9498) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 9499) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4028 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9502) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4029 (("," 1480) (possible_comma 9500)) 4030 ((")" 9501)) 4031 (("$" 9479) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 9481) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 9505) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 9506) ("{" 1658)) 4032 (("(" 9495) (argument_list 9507)) 4033 (("$" 9479) (T_STRING 9514) (T_VARIABLE 9481) (property_name 9521) (simple_variable 9516) ("{" 9517)) 4034 (("$" 9479) (T_STRING 9514) (T_VARIABLE 9481) (property_name 9515) (simple_variable 9516) ("{" 9517)) 4035 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 9512) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4036 (("]" 9513)) 4037 (("(" 9495) (argument_list 9520)) 4038 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9518) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4039 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9519)) 4040 (("(" 9495) (argument_list 9522)) 4041 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9526) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4042 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9527)) 4043 (("`" 9529)) 4044 (("]" 9531)) 4045 (("=" 9532)) 4046 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9533) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 4047 (("%" 1764) ("*" 1765) ("+" 1766) ("-" 1767) ("." 1768) ("/" 1769) ("<" 1770) (">" 1771) ("?" 1772) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1773) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1774) (T_BOOLEAN_AND 1775) (T_BOOLEAN_OR 1776) (T_COALESCE 1777) (T_DOUBLE_ARROW 9536) (T_INSTANCEOF 1778) (T_IS_EQUAL 1779) (T_IS_GREATER_OR_EQUAL 1780) (T_IS_IDENTICAL 1781) (T_IS_NOT_EQUAL 1782) (T_IS_NOT_IDENTICAL 1783) (T_IS_SMALLER_OR_EQUAL 1784) (T_LOGICAL_AND 1785) (T_LOGICAL_OR 1786) (T_LOGICAL_XOR 1787) (T_POW 1788) (T_SL 1789) (T_SPACESHIP 1790) (T_SR 1791) ("^" 1792) ("|" 1793)) 4048 (("!" 8550) ("\"" 8551) ("$" 8552) ("(" 8553) ("+" 8554) ("-" 8555) ("@" 8556) (T_ARRAY 8557) (T_ARRAY_CAST 8558) (T_ATTRIBUTE 128) (T_BOOL_CAST 8559) (T_CLASS_C 8560) (T_CLONE 8561) (T_CONSTANT_ENCAPSED_STRING 8562) (T_DEC 8563) (T_DIR 8564) (T_DNUMBER 8565) (T_DOUBLE_CAST 8566) (T_EMPTY 8567) (T_EVAL 8568) (T_EXIT 8569) (T_FILE 8570) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 8571) (T_INC 8572) (T_INCLUDE 8573) (T_INCLUDE_ONCE 8574) (T_INT_CAST 8575) (T_ISSET 8576) (T_LINE 8577) (T_LIST 8630) (T_LNUMBER 8579) (T_MATCH 8580) (T_METHOD_C 8581) (T_NAME_FULLY_QUALIFIED 8582) (T_NAME_QUALIFIED 8583) (T_NAME_RELATIVE 8584) (T_NEW 8585) (T_NS_C 8586) (T_OBJECT_CAST 8587) (T_PRINT 8588) (T_PROPERTY_C 8589) (T_READONLY 8590) (T_REQUIRE 8591) (T_REQUIRE_ONCE 8592) (T_START_HEREDOC 8593) (T_STATIC 8594) (T_STRING 8595) (T_STRING_CAST 8596) (T_THROW 8597) (T_TRAIT_C 8598) (T_UNSET_CAST 8599) (T_VARIABLE 8600) (T_YIELD 8601) (T_YIELD_FROM 8602) ("[" 8603) ("`" 8604) (array_object_dereferenceable 8606) (attribute 133) (attributes 8607) (callable_expr 8608) (callable_variable 8609) (class_constant 8610) (class_name 8611) (constant 8612) (dereferenceable_scalar 8613) (expr 9537) (fn 8615) (fully_dereferenceable 105) #'8616 (function_call 8617) (inline_function 8618) (internal_functions_in_yacc 8619) (match 8620) (name 8621) (new_dereferenceable 8622) (new_non_dereferenceable 8623) (scalar 8624) (simple_variable 8625) (static_member 8626) (variable 8627) (variable_class_name 8628) ("~" 8629)) 4049 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 9546) (T_VARIABLE 2896) (encaps_var 2910)) 4050 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 9545) (T_VARIABLE 2896) (encaps_var 2901)) 4051 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9587) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4052 ((T_CLASS 9579)) 4053 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 9578) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 9559) (attribute 2958) (class_modifier 2939)) 4054 ((T_PAAMAYIM_NEKUDOTAYIM 9576)) 4055 (("(" 8878) (argument_list 9575)) 4056 ((T_NULLSAFE_OBJECT_OPERATOR 9566) (T_OBJECT_OPERATOR 9567) (T_PAAMAYIM_NEKUDOTAYIM 9568) ("[" 9569)) 4057 (("$" 8862) (T_STRING 8897) (T_VARIABLE 8864) (property_name 9574) (simple_variable 8899) ("{" 8900)) 4058 (("$" 8862) (T_STRING 8897) (T_VARIABLE 8864) (property_name 9573) (simple_variable 8899) ("{" 8900)) 4059 (("$" 8862) (T_VARIABLE 8864) (simple_variable 9572)) 4060 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 9570) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4061 (("]" 9571)) 4062 (("$" 8862) (T_VARIABLE 8864) (simple_variable 9577)) 4063 (("(" 2960) (argument_list 2961) (ctor_arguments 9580)) 4064 ((T_EXTENDS 2963) (extends_from 9581)) 4065 ((T_IMPLEMENTS 2965) (implements_list 9582)) 4066 ((backup_doc_comment 9583)) 4067 (("{" 9584)) 4068 ((class_statement_list 9585)) 4069 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 9586)) 4070 (("%" 299) (")" 9588) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4071 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9590) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4072 (("%" 299) (")" 9591) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4073 (("{" 9592)) 4074 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 9593) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 4075 (("}" 9594)) 4076 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 9596) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 4077 ((")" 9597)) 4078 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 9599) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4079 (("," 5192) (possible_comma 9600)) 4080 ((")" 9601)) 4081 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9609) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 9610) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 9611) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4082 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9614) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4083 (("," 1480) (possible_comma 9612)) 4084 ((")" 9613)) 4085 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9616) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4086 (("%" 299) (")" 9617) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4087 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9619) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4088 (("%" 299) (")" 9620) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4089 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 9627) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 4090 ((")" 9628)) 4091 (("%" 299) (")" 9633) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4092 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9636) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4093 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9637)) 4094 (("\"" 9639) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 4095 (("$" 1679) (T_ABSTRACT 7830) (T_ARRAY 7831) (T_AS 7832) (T_BREAK 7833) (T_CALLABLE 7834) (T_CASE 7835) (T_CATCH 7836) (T_CLASS 7837) (T_CLASS_C 7838) (T_CLONE 7839) (T_CONST 7840) (T_CONTINUE 7841) (T_DECLARE 7842) (T_DEFAULT 7843) (T_DIR 7844) (T_DO 7845) (T_ECHO 7846) (T_ELSE 7847) (T_ELSEIF 7848) (T_EMPTY 7849) (T_ENDDECLARE 7850) (T_ENDFOR 7851) (T_ENDFOREACH 7852) (T_ENDIF 7853) (T_ENDSWITCH 7854) (T_ENDWHILE 7855) (T_ENUM 7856) (T_EVAL 7857) (T_EXIT 7858) (T_EXTENDS 7859) (T_FILE 7860) (T_FINAL 7861) (T_FINALLY 7862) (T_FN 7863) (T_FOR 7864) (T_FOREACH 7865) (T_FUNCTION 7866) (T_FUNC_C 7867) (T_GLOBAL 7868) (T_GOTO 7869) (T_IF 7870) (T_IMPLEMENTS 7871) (T_INCLUDE 7872) (T_INCLUDE_ONCE 7873) (T_INSTANCEOF 7874) (T_INSTEADOF 7875) (T_INTERFACE 7876) (T_ISSET 7877) (T_LINE 7878) (T_LIST 7879) (T_LOGICAL_AND 7880) (T_LOGICAL_OR 7881) (T_LOGICAL_XOR 7882) (T_MATCH 7883) (T_METHOD_C 7884) (T_NAMESPACE 7885) (T_NEW 7886) (T_NS_C 7887) (T_PRINT 7888) (T_PRIVATE 7889) (T_PROPERTY_C 7890) (T_PROTECTED 7891) (T_PUBLIC 7892) (T_READONLY 7893) (T_REQUIRE 7894) (T_REQUIRE_ONCE 7895) (T_RETURN 7896) (T_STATIC 7897) (T_STRING 7898) (T_SWITCH 7899) (T_THROW 7900) (T_TRAIT 7901) (T_TRAIT_C 7902) (T_TRY 7903) (T_UNSET 7904) (T_USE 7905) (T_VAR 7906) (T_VARIABLE 1728) (T_WHILE 7907) (T_YIELD 7908) (identifier 9642) (member_name 9643) (reserved_non_modifiers 7911) (semi_reserved 7912) (simple_variable 9644) ("{" 9645)) 4096 (("(" 7917) (argument_list 9648)) 4097 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9646) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4098 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9647)) 4099 ((T_FN 35) (T_FUNCTION 129) (fn 1745) #'1746 (inline_function 9652)) 4100 (("$" 1679) (T_STRING 9659) (T_VARIABLE 1728) (property_name 9666) (simple_variable 9661) ("{" 9662)) 4101 (("$" 1679) (T_STRING 9659) (T_VARIABLE 1728) (property_name 9660) (simple_variable 9661) ("{" 9662)) 4102 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 9657) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4103 (("]" 9658)) 4104 (("(" 7917) (argument_list 9665)) 4105 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9663) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4106 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9664)) 4107 (("(" 7917) (argument_list 9667)) 4108 (("`" 9670)) 4109 (("]" 9672)) 4110 (("=" 9673)) 4111 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9674) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 4112 (("%" 1764) ("*" 1765) ("+" 1766) ("-" 1767) ("." 1768) ("/" 1769) ("<" 1770) (">" 1771) ("?" 1772) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1773) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1774) (T_BOOLEAN_AND 1775) (T_BOOLEAN_OR 1776) (T_COALESCE 1777) (T_DOUBLE_ARROW 9677) (T_INSTANCEOF 1778) (T_IS_EQUAL 1779) (T_IS_GREATER_OR_EQUAL 1780) (T_IS_IDENTICAL 1781) (T_IS_NOT_EQUAL 1782) (T_IS_NOT_IDENTICAL 1783) (T_IS_SMALLER_OR_EQUAL 1784) (T_LOGICAL_AND 1785) (T_LOGICAL_OR 1786) (T_LOGICAL_XOR 1787) (T_POW 1788) (T_SL 1789) (T_SPACESHIP 1790) (T_SR 1791) ("^" 1792) ("|" 1793)) 4113 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1762) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (array_object_dereferenceable 1734) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 9678) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 4114 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 9687) (T_VARIABLE 2896) (encaps_var 2910)) 4115 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 9686) (T_VARIABLE 2896) (encaps_var 2901)) 4116 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9728) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4117 ((T_CLASS 9720)) 4118 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 9719) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 9700) (attribute 2958) (class_modifier 2939)) 4119 ((T_PAAMAYIM_NEKUDOTAYIM 9717)) 4120 (("(" 7972) (argument_list 9716)) 4121 ((T_NULLSAFE_OBJECT_OPERATOR 9707) (T_OBJECT_OPERATOR 9708) (T_PAAMAYIM_NEKUDOTAYIM 9709) ("[" 9710)) 4122 (("$" 7956) (T_STRING 7991) (T_VARIABLE 7958) (property_name 9715) (simple_variable 7993) ("{" 7994)) 4123 (("$" 7956) (T_STRING 7991) (T_VARIABLE 7958) (property_name 9714) (simple_variable 7993) ("{" 7994)) 4124 (("$" 7956) (T_VARIABLE 7958) (simple_variable 9713)) 4125 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 9711) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4126 (("]" 9712)) 4127 (("$" 7956) (T_VARIABLE 7958) (simple_variable 9718)) 4128 (("(" 2960) (argument_list 2961) (ctor_arguments 9721)) 4129 ((T_EXTENDS 2963) (extends_from 9722)) 4130 ((T_IMPLEMENTS 2965) (implements_list 9723)) 4131 ((backup_doc_comment 9724)) 4132 (("{" 9725)) 4133 ((class_statement_list 9726)) 4134 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 9727)) 4135 (("%" 299) (")" 9729) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4136 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9731) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4137 (("%" 299) (")" 9732) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4138 (("{" 9733)) 4139 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 9734) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 4140 (("}" 9735)) 4141 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 9737) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 4142 ((")" 9738)) 4143 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 9740) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4144 (("," 5192) (possible_comma 9741)) 4145 ((")" 9742)) 4146 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9750) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 9751) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 9752) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4147 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9755) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4148 (("," 1480) (possible_comma 9753)) 4149 ((")" 9754)) 4150 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9757) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4151 (("%" 299) (")" 9758) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4152 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9760) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4153 (("%" 299) (")" 9761) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4154 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 9769) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 4155 ((")" 9770)) 4156 (("%" 299) (")" 9775) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4157 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9778) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4158 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9779)) 4159 (("\"" 9781) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 4160 (("$" 9784) (T_VARIABLE 9791) (simple_variable 9835) ("{" 9836)) 4161 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9833) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4162 ((T_CLASS 9825)) 4163 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 9824) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 9793) (attribute 2958) (class_modifier 2939)) 4164 ((T_PAAMAYIM_NEKUDOTAYIM 9822)) 4165 (("(" 9814) (argument_list 9815)) 4166 ((T_NULLSAFE_OBJECT_OPERATOR 9800) (T_OBJECT_OPERATOR 9801) (T_PAAMAYIM_NEKUDOTAYIM 9802) ("[" 9803)) 4167 (("$" 9784) (T_STRING 9807) (T_VARIABLE 9791) (property_name 9813) (simple_variable 9809) ("{" 9810)) 4168 (("$" 9784) (T_STRING 9807) (T_VARIABLE 9791) (property_name 9808) (simple_variable 9809) ("{" 9810)) 4169 (("$" 9784) (T_VARIABLE 9791) (simple_variable 9806)) 4170 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 9804) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4171 (("]" 9805)) 4172 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9811) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4173 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9812)) 4174 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9816) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 9817) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 9818) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4175 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9821) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4176 (("," 1480) (possible_comma 9819)) 4177 ((")" 9820)) 4178 (("$" 9784) (T_VARIABLE 9791) (simple_variable 9823)) 4179 (("(" 2960) (argument_list 2961) (ctor_arguments 9826)) 4180 ((T_EXTENDS 2963) (extends_from 9827)) 4181 ((T_IMPLEMENTS 2965) (implements_list 9828)) 4182 ((backup_doc_comment 9829)) 4183 (("{" 9830)) 4184 ((class_statement_list 9831)) 4185 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 9832)) 4186 (("%" 299) (")" 9834) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4187 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9837) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4188 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9838)) 4189 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 9840) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 4190 ((")" 9841)) 4191 (("%" 299) (")" 9843) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4192 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9846) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4193 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9847)) 4194 (("\"" 9849) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 4195 ((backup_doc_comment 9852)) 4196 (("(" 9853)) 4197 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 9854)) 4198 ((")" 9855)) 4199 ((T_USE 164) (lexical_vars 9856)) 4200 ((":" 166) (return_type 9857)) 4201 ((backup_fn_flags 9858)) 4202 (("{" 9859)) 4203 ((inner_statement_list 9860)) 4204 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 9861) ("~" 127)) 4205 ((backup_fn_flags 9862)) 4206 ((backup_doc_comment 9864)) 4207 (("(" 9865)) 4208 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 9866)) 4209 ((")" 9867)) 4210 ((":" 2249) (return_type 9868)) 4211 ((T_DOUBLE_ARROW 9869)) 4212 ((backup_fn_flags 9870)) 4213 ((backup_lex_pos 9871)) 4214 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9872) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 4215 (("%" 1289) ("*" 1290) ("+" 1291) ("-" 1292) ("." 1293) ("/" 1294) ("<" 1295) (">" 1296) ("?" 1297) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1298) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1299) (T_BOOLEAN_AND 1300) (T_BOOLEAN_OR 1301) (T_COALESCE 1302) (T_INSTANCEOF 1303) (T_IS_EQUAL 1304) (T_IS_GREATER_OR_EQUAL 1305) (T_IS_IDENTICAL 1306) (T_IS_NOT_EQUAL 1307) (T_IS_NOT_IDENTICAL 1308) (T_IS_SMALLER_OR_EQUAL 1309) (T_LOGICAL_AND 1310) (T_LOGICAL_OR 1311) (T_LOGICAL_XOR 1312) (T_POW 1313) (T_SL 1314) (T_SPACESHIP 1315) (T_SR 1316) ("^" 1317) (backup_fn_flags 9873) ("|" 1318)) 4216 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 9875) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4217 (("$" 1211) (T_ABSTRACT 1388) (T_ARRAY 1389) (T_AS 1390) (T_BREAK 1391) (T_CALLABLE 1392) (T_CASE 1393) (T_CATCH 1394) (T_CLASS 1395) (T_CLASS_C 1396) (T_CLONE 1397) (T_CONST 1398) (T_CONTINUE 1399) (T_DECLARE 1400) (T_DEFAULT 1401) (T_DIR 1402) (T_DO 1403) (T_ECHO 1404) (T_ELSE 1405) (T_ELSEIF 1406) (T_EMPTY 1407) (T_ENDDECLARE 1408) (T_ENDFOR 1409) (T_ENDFOREACH 1410) (T_ENDIF 1411) (T_ENDSWITCH 1412) (T_ENDWHILE 1413) (T_ENUM 1414) (T_EVAL 1415) (T_EXIT 1416) (T_EXTENDS 1417) (T_FILE 1418) (T_FINAL 1419) (T_FINALLY 1420) (T_FN 1421) (T_FOR 1422) (T_FOREACH 1423) (T_FUNCTION 1424) (T_FUNC_C 1425) (T_GLOBAL 1426) (T_GOTO 1427) (T_IF 1428) (T_IMPLEMENTS 1429) (T_INCLUDE 1430) (T_INCLUDE_ONCE 1431) (T_INSTANCEOF 1432) (T_INSTEADOF 1433) (T_INTERFACE 1434) (T_ISSET 1435) (T_LINE 1436) (T_LIST 1437) (T_LOGICAL_AND 1438) (T_LOGICAL_OR 1439) (T_LOGICAL_XOR 1440) (T_MATCH 1441) (T_METHOD_C 1442) (T_NAMESPACE 1443) (T_NEW 1444) (T_NS_C 1445) (T_PRINT 1446) (T_PRIVATE 1447) (T_PROPERTY_C 1448) (T_PROTECTED 1449) (T_PUBLIC 1450) (T_READONLY 1451) (T_REQUIRE 1452) (T_REQUIRE_ONCE 1453) (T_RETURN 1454) (T_STATIC 1455) (T_STRING 1456) (T_SWITCH 1457) (T_THROW 1458) (T_TRAIT 1459) (T_TRAIT_C 1460) (T_TRY 1461) (T_UNSET 1462) (T_USE 1463) (T_VAR 1464) (T_VARIABLE 1259) (T_WHILE 1465) (T_YIELD 1466) (identifier 9877) (member_name 9878) (reserved_non_modifiers 1469) (semi_reserved 1470) (simple_variable 9879) ("{" 9880)) 4218 (("(" 1475) (argument_list 9883)) 4219 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9881) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4220 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9882)) 4221 ((T_FN 35) (T_FUNCTION 129) (fn 1273) #'1274 (inline_function 9887)) 4222 (("$" 1211) (T_STRING 9893) (T_VARIABLE 1259) (property_name 9900) (simple_variable 9895) ("{" 9896)) 4223 (("$" 1211) (T_STRING 9893) (T_VARIABLE 1259) (property_name 9894) (simple_variable 9895) ("{" 9896)) 4224 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 9891) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4225 (("]" 9892)) 4226 (("(" 1475) (argument_list 9899)) 4227 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 9897) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4228 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 9898)) 4229 (("(" 1475) (argument_list 9901)) 4230 (("`" 9903)) 4231 (("]" 9905)) 4232 (("=" 9906)) 4233 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9907) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 4234 (("%" 1289) ("*" 1290) ("+" 1291) ("-" 1292) ("." 1293) ("/" 1294) ("<" 1295) (">" 1296) ("?" 1297) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1298) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1299) (T_BOOLEAN_AND 1300) (T_BOOLEAN_OR 1301) (T_COALESCE 1302) (T_DOUBLE_ARROW 9910) (T_INSTANCEOF 1303) (T_IS_EQUAL 1304) (T_IS_GREATER_OR_EQUAL 1305) (T_IS_IDENTICAL 1306) (T_IS_NOT_EQUAL 1307) (T_IS_NOT_IDENTICAL 1308) (T_IS_SMALLER_OR_EQUAL 1309) (T_LOGICAL_AND 1310) (T_LOGICAL_OR 1311) (T_LOGICAL_XOR 1312) (T_POW 1313) (T_SL 1314) (T_SPACESHIP 1315) (T_SR 1316) ("^" 1317) ("|" 1318)) 4235 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9911) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 4236 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 9920) (T_VARIABLE 2896) (encaps_var 2910)) 4237 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 9919) (T_VARIABLE 2896) (encaps_var 2901)) 4238 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9961) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4239 ((T_CLASS 9953)) 4240 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 9952) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 9933) (attribute 2958) (class_modifier 2939)) 4241 ((T_PAAMAYIM_NEKUDOTAYIM 9950)) 4242 (("(" 1645) (argument_list 9949)) 4243 ((T_NULLSAFE_OBJECT_OPERATOR 9940) (T_OBJECT_OPERATOR 9941) (T_PAAMAYIM_NEKUDOTAYIM 9942) ("[" 9943)) 4244 (("$" 1521) (T_STRING 1668) (T_VARIABLE 1541) (property_name 9948) (simple_variable 1670) ("{" 1671)) 4245 (("$" 1521) (T_STRING 1668) (T_VARIABLE 1541) (property_name 9947) (simple_variable 1670) ("{" 1671)) 4246 (("$" 1521) (T_VARIABLE 1541) (simple_variable 9946)) 4247 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 9944) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4248 (("]" 9945)) 4249 (("$" 1521) (T_VARIABLE 1541) (simple_variable 9951)) 4250 (("(" 2960) (argument_list 2961) (ctor_arguments 9954)) 4251 ((T_EXTENDS 2963) (extends_from 9955)) 4252 ((T_IMPLEMENTS 2965) (implements_list 9956)) 4253 ((backup_doc_comment 9957)) 4254 (("{" 9958)) 4255 ((class_statement_list 9959)) 4256 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 9960)) 4257 (("%" 299) (")" 9962) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4258 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9964) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4259 (("%" 299) (")" 9965) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4260 (("{" 9966)) 4261 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 9967) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 4262 (("}" 9968)) 4263 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 9970) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 4264 ((")" 9971)) 4265 (("=" 9972)) 4266 (("!" 1209) ("\"" 1210) ("$" 1211) ("(" 1212) ("+" 1213) ("-" 1214) ("@" 1215) (T_ARRAY 1216) (T_ARRAY_CAST 1217) (T_ATTRIBUTE 128) (T_BOOL_CAST 1218) (T_CLASS_C 1219) (T_CLONE 1220) (T_CONSTANT_ENCAPSED_STRING 1221) (T_DEC 1222) (T_DIR 1223) (T_DNUMBER 1224) (T_DOUBLE_CAST 1225) (T_EMPTY 1226) (T_EVAL 1227) (T_EXIT 1228) (T_FILE 1229) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1230) (T_INC 1231) (T_INCLUDE 1232) (T_INCLUDE_ONCE 1233) (T_INT_CAST 1234) (T_ISSET 1235) (T_LINE 1236) (T_LIST 1237) (T_LNUMBER 1238) (T_MATCH 1239) (T_METHOD_C 1240) (T_NAME_FULLY_QUALIFIED 1241) (T_NAME_QUALIFIED 1242) (T_NAME_RELATIVE 1243) (T_NEW 1244) (T_NS_C 1245) (T_OBJECT_CAST 1246) (T_PRINT 1247) (T_PROPERTY_C 1248) (T_READONLY 1249) (T_REQUIRE 1250) (T_REQUIRE_ONCE 1251) (T_START_HEREDOC 1252) (T_STATIC 1253) (T_STRING 1254) (T_STRING_CAST 1255) (T_THROW 1256) (T_TRAIT_C 1257) (T_UNSET_CAST 1258) (T_VARIABLE 1259) (T_YIELD 1260) (T_YIELD_FROM 1261) ("[" 1262) ("`" 1263) (array_object_dereferenceable 1264) (attribute 133) (attributes 1265) (callable_expr 1266) (callable_variable 1267) (class_constant 1268) (class_name 1269) (constant 1270) (dereferenceable_scalar 1271) (expr 9973) (fn 1273) (fully_dereferenceable 105) #'1274 (function_call 1275) (inline_function 1276) (internal_functions_in_yacc 1277) (match 1278) (name 1279) (new_dereferenceable 1280) (new_non_dereferenceable 1281) (scalar 1282) (simple_variable 1283) (static_member 1284) (variable 1285) (variable_class_name 1286) ("~" 1287)) 4267 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 9975) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4268 (("," 5192) (possible_comma 9976)) 4269 ((")" 9977)) 4270 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9985) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 9986) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 9987) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4271 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 9990) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4272 (("," 1480) (possible_comma 9988)) 4273 ((")" 9989)) 4274 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9992) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4275 (("%" 299) (")" 9993) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4276 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 9995) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4277 (("%" 299) (")" 9996) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4278 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 10003) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 4279 ((")" 10004)) 4280 (("%" 299) (")" 10009) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4281 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10012) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4282 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10013)) 4283 (("\"" 10015) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 4284 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 10023) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 10024) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 10025) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4285 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 10028) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4286 (("," 1480) (possible_comma 10026)) 4287 ((")" 10027)) 4288 (("$" 10030) (T_VARIABLE 10037) (simple_variable 10081) ("{" 10082)) 4289 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 10079) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4290 ((T_CLASS 10071)) 4291 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 10070) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 10039) (attribute 2958) (class_modifier 2939)) 4292 ((T_PAAMAYIM_NEKUDOTAYIM 10068)) 4293 (("(" 10060) (argument_list 10061)) 4294 ((T_NULLSAFE_OBJECT_OPERATOR 10046) (T_OBJECT_OPERATOR 10047) (T_PAAMAYIM_NEKUDOTAYIM 10048) ("[" 10049)) 4295 (("$" 10030) (T_STRING 10053) (T_VARIABLE 10037) (property_name 10059) (simple_variable 10055) ("{" 10056)) 4296 (("$" 10030) (T_STRING 10053) (T_VARIABLE 10037) (property_name 10054) (simple_variable 10055) ("{" 10056)) 4297 (("$" 10030) (T_VARIABLE 10037) (simple_variable 10052)) 4298 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 10050) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4299 (("]" 10051)) 4300 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10057) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4301 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10058)) 4302 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 10062) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 10063) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 10064) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4303 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 10067) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4304 (("," 1480) (possible_comma 10065)) 4305 ((")" 10066)) 4306 (("$" 10030) (T_VARIABLE 10037) (simple_variable 10069)) 4307 (("(" 2960) (argument_list 2961) (ctor_arguments 10072)) 4308 ((T_EXTENDS 2963) (extends_from 10073)) 4309 ((T_IMPLEMENTS 2965) (implements_list 10074)) 4310 ((backup_doc_comment 10075)) 4311 (("{" 10076)) 4312 ((class_statement_list 10077)) 4313 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 10078)) 4314 (("%" 299) (")" 10080) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4315 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10083) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4316 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10084)) 4317 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 10086) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4318 (("%" 299) (")" 10087) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4319 (("{" 10088)) 4320 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 10089) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 4321 (("}" 10090)) 4322 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 10092) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 4323 ((")" 10093)) 4324 (("=" 10094)) 4325 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10095) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4326 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 10097) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4327 (("," 5192) (possible_comma 10098)) 4328 ((")" 10099)) 4329 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 10105) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 10106) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 10107) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4330 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 10110) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4331 (("," 1480) (possible_comma 10108)) 4332 ((")" 10109)) 4333 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 10112) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4334 (("%" 299) (")" 10113) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4335 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 10115) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4336 (("%" 299) (")" 10116) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4337 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 10119) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 4338 ((")" 10120)) 4339 (("$" 947) (T_ABSTRACT 10122) (T_ARRAY 10123) (T_AS 10124) (T_BREAK 10125) (T_CALLABLE 10126) (T_CASE 10127) (T_CATCH 10128) (T_CLASS 10129) (T_CLASS_C 10130) (T_CLONE 10131) (T_CONST 10132) (T_CONTINUE 10133) (T_DECLARE 10134) (T_DEFAULT 10135) (T_DIR 10136) (T_DO 10137) (T_ECHO 10138) (T_ELSE 10139) (T_ELSEIF 10140) (T_EMPTY 10141) (T_ENDDECLARE 10142) (T_ENDFOR 10143) (T_ENDFOREACH 10144) (T_ENDIF 10145) (T_ENDSWITCH 10146) (T_ENDWHILE 10147) (T_ENUM 10148) (T_EVAL 10149) (T_EXIT 10150) (T_EXTENDS 10151) (T_FILE 10152) (T_FINAL 10153) (T_FINALLY 10154) (T_FN 10155) (T_FOR 10156) (T_FOREACH 10157) (T_FUNCTION 10158) (T_FUNC_C 10159) (T_GLOBAL 10160) (T_GOTO 10161) (T_IF 10162) (T_IMPLEMENTS 10163) (T_INCLUDE 10164) (T_INCLUDE_ONCE 10165) (T_INSTANCEOF 10166) (T_INSTEADOF 10167) (T_INTERFACE 10168) (T_ISSET 10169) (T_LINE 10170) (T_LIST 10171) (T_LOGICAL_AND 10172) (T_LOGICAL_OR 10173) (T_LOGICAL_XOR 10174) (T_MATCH 10175) (T_METHOD_C 10176) (T_NAMESPACE 10177) (T_NEW 10178) (T_NS_C 10179) (T_PRINT 10180) (T_PRIVATE 10181) (T_PROPERTY_C 10182) (T_PROTECTED 10183) (T_PUBLIC 10184) (T_READONLY 10185) (T_REQUIRE 10186) (T_REQUIRE_ONCE 10187) (T_RETURN 10188) (T_STATIC 10189) (T_STRING 10190) (T_SWITCH 10191) (T_THROW 10192) (T_TRAIT 10193) (T_TRAIT_C 10194) (T_TRY 10195) (T_UNSET 10196) (T_USE 10197) (T_VAR 10198) (T_VARIABLE 1047) (T_WHILE 10199) (T_YIELD 10200) (identifier 10201) (member_name 10202) (reserved_non_modifiers 10203) (semi_reserved 10204) (simple_variable 10205) ("{" 10206)) 4340 (("(" 10021) (argument_list 10209)) 4341 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10207) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4342 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10208)) 4343 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (ampersand 10239) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10240) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4344 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10238) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4345 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10237) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4346 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10236) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4347 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10235) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4348 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10234) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4349 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10233) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4350 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10232) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4351 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10231) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4352 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10230) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4353 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10229) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4354 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10228) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4355 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10227) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4356 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10226) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4357 (("\"" 1520) ("$" 10030) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 10241) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 10037) ("[" 1542) (array_object_dereferenceable 10242) (callable_expr 10243) (callable_variable 10244) (class_constant 1546) (class_name 10245) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 10246) (name 10247) (new_dereferenceable 1552) (simple_variable 10248) (static_member 10249) (variable 10250) (variable_class_name 10251)) 4358 (("(" 10060) (argument_list 10271)) 4359 ((T_NULLSAFE_OBJECT_OPERATOR 10262) (T_OBJECT_OPERATOR 10263) ("[" 10264)) 4360 (("(" 10060) (argument_list 10261)) 4361 ((T_PAAMAYIM_NEKUDOTAYIM 10257)) 4362 (("(" 10060) (argument_list 10256)) 4363 ((T_PAAMAYIM_NEKUDOTAYIM 10252)) 4364 (("$" 10030) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 10037) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 10253) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 10254) ("{" 1642)) 4365 (("(" 10060) (argument_list 10255)) 4366 (("$" 10030) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 10037) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 10258) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 10259) ("{" 1658)) 4367 (("(" 10060) (argument_list 10260)) 4368 (("$" 10030) (T_STRING 10053) (T_VARIABLE 10037) (property_name 10269) (simple_variable 10055) ("{" 10056)) 4369 (("$" 10030) (T_STRING 10053) (T_VARIABLE 10037) (property_name 10267) (simple_variable 10055) ("{" 10056)) 4370 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 10265) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4371 (("]" 10266)) 4372 (("(" 10060) (argument_list 10268)) 4373 (("(" 10060) (argument_list 10270)) 4374 ((")" 10273)) 4375 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10276) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4376 ((backup_doc_comment 10278)) 4377 (("(" 10279)) 4378 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 10280)) 4379 ((")" 10281)) 4380 ((T_USE 164) (lexical_vars 10282)) 4381 ((":" 166) (return_type 10283)) 4382 ((backup_fn_flags 10284)) 4383 (("{" 10285)) 4384 ((inner_statement_list 10286)) 4385 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 10287) ("~" 127)) 4386 ((backup_fn_flags 10288)) 4387 ((backup_doc_comment 10290)) 4388 (("(" 10291)) 4389 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 10292)) 4390 ((")" 10293)) 4391 ((":" 2249) (return_type 10294)) 4392 ((T_DOUBLE_ARROW 10295)) 4393 ((backup_fn_flags 10296)) 4394 ((backup_lex_pos 10297)) 4395 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10298) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4396 (("%" 1111) ("*" 1112) ("+" 1113) ("-" 1114) ("." 1115) ("/" 1116) ("<" 1117) (">" 1118) ("?" 1119) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1120) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1121) (T_BOOLEAN_AND 1122) (T_BOOLEAN_OR 1123) (T_COALESCE 1124) (T_INSTANCEOF 1125) (T_IS_EQUAL 1126) (T_IS_GREATER_OR_EQUAL 1127) (T_IS_IDENTICAL 1128) (T_IS_NOT_EQUAL 1129) (T_IS_NOT_IDENTICAL 1130) (T_IS_SMALLER_OR_EQUAL 1131) (T_LOGICAL_AND 1132) (T_LOGICAL_OR 1133) (T_LOGICAL_XOR 1134) (T_POW 1135) (T_SL 1136) (T_SPACESHIP 1137) (T_SR 1138) ("^" 1139) (backup_fn_flags 10299) ("|" 1140)) 4397 (("$" 947) (T_ABSTRACT 10122) (T_ARRAY 10123) (T_AS 10124) (T_BREAK 10125) (T_CALLABLE 10126) (T_CASE 10127) (T_CATCH 10128) (T_CLASS 10129) (T_CLASS_C 10130) (T_CLONE 10131) (T_CONST 10132) (T_CONTINUE 10133) (T_DECLARE 10134) (T_DEFAULT 10135) (T_DIR 10136) (T_DO 10137) (T_ECHO 10138) (T_ELSE 10139) (T_ELSEIF 10140) (T_EMPTY 10141) (T_ENDDECLARE 10142) (T_ENDFOR 10143) (T_ENDFOREACH 10144) (T_ENDIF 10145) (T_ENDSWITCH 10146) (T_ENDWHILE 10147) (T_ENUM 10148) (T_EVAL 10149) (T_EXIT 10150) (T_EXTENDS 10151) (T_FILE 10152) (T_FINAL 10153) (T_FINALLY 10154) (T_FN 10155) (T_FOR 10156) (T_FOREACH 10157) (T_FUNCTION 10158) (T_FUNC_C 10159) (T_GLOBAL 10160) (T_GOTO 10161) (T_IF 10162) (T_IMPLEMENTS 10163) (T_INCLUDE 10164) (T_INCLUDE_ONCE 10165) (T_INSTANCEOF 10166) (T_INSTEADOF 10167) (T_INTERFACE 10168) (T_ISSET 10169) (T_LINE 10170) (T_LIST 10171) (T_LOGICAL_AND 10172) (T_LOGICAL_OR 10173) (T_LOGICAL_XOR 10174) (T_MATCH 10175) (T_METHOD_C 10176) (T_NAMESPACE 10177) (T_NEW 10178) (T_NS_C 10179) (T_PRINT 10180) (T_PRIVATE 10181) (T_PROPERTY_C 10182) (T_PROTECTED 10183) (T_PUBLIC 10184) (T_READONLY 10185) (T_REQUIRE 10186) (T_REQUIRE_ONCE 10187) (T_RETURN 10188) (T_STATIC 10189) (T_STRING 10190) (T_SWITCH 10191) (T_THROW 10192) (T_TRAIT 10193) (T_TRAIT_C 10194) (T_TRY 10195) (T_UNSET 10196) (T_USE 10197) (T_VAR 10198) (T_VARIABLE 1047) (T_WHILE 10199) (T_YIELD 10200) (identifier 10301) (member_name 10302) (reserved_non_modifiers 10203) (semi_reserved 10204) (simple_variable 10303) ("{" 10304)) 4398 (("(" 10021) (argument_list 10307)) 4399 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10305) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4400 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10306)) 4401 ((T_FN 35) (T_FUNCTION 129) (fn 1063) #'1064 (inline_function 10311)) 4402 (("$" 947) (T_STRING 10317) (T_VARIABLE 1047) (property_name 10324) (simple_variable 10319) ("{" 10320)) 4403 (("$" 947) (T_STRING 10317) (T_VARIABLE 1047) (property_name 10318) (simple_variable 10319) ("{" 10320)) 4404 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 10315) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4405 (("]" 10316)) 4406 (("(" 10021) (argument_list 10323)) 4407 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10321) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4408 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10322)) 4409 (("(" 10021) (argument_list 10325)) 4410 (("`" 10327)) 4411 (("]" 10329)) 4412 (("=" 10330)) 4413 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 10331) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4414 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 10339) (T_VARIABLE 2896) (encaps_var 2910)) 4415 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 10338) (T_VARIABLE 2896) (encaps_var 2901)) 4416 (("%" 299) (")" 10352) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4417 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10355) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4418 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10356)) 4419 (("\"" 10358) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 4420 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (ampersand 10389) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 10390) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4421 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 10388) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4422 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 10387) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4423 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 10386) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4424 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 10385) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4425 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 10384) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4426 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 10383) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4427 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 10382) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4428 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 10381) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4429 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 10380) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4430 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 10379) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4431 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 10378) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4432 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 10377) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4433 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 10376) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4434 (("\"" 1520) ("$" 10391) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 10392) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 10393) ("[" 1542) (array_object_dereferenceable 10394) (callable_expr 10395) (callable_variable 10396) (class_constant 1546) (class_name 10397) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 10398) (name 10399) (new_dereferenceable 1552) (simple_variable 10400) (static_member 10401) (variable 10402) (variable_class_name 10403)) 4435 (("$" 10391) (T_VARIABLE 10393) (simple_variable 10436) ("{" 10437)) 4436 (("(" 10407) (argument_list 10435)) 4437 ((T_NULLSAFE_OBJECT_OPERATOR 10421) (T_OBJECT_OPERATOR 10422) ("[" 10423)) 4438 (("(" 10407) (argument_list 10420)) 4439 ((T_PAAMAYIM_NEKUDOTAYIM 10416)) 4440 (("(" 10407) (argument_list 10415)) 4441 ((T_PAAMAYIM_NEKUDOTAYIM 10404)) 4442 (("$" 10391) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 10393) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 10405) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 10406) ("{" 1642)) 4443 (("(" 10407) (argument_list 10408)) 4444 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 10409) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 10410) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 10411) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4445 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 10414) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4446 (("," 1480) (possible_comma 10412)) 4447 ((")" 10413)) 4448 (("$" 10391) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 10393) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 10417) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 10418) ("{" 1658)) 4449 (("(" 10407) (argument_list 10419)) 4450 (("$" 10391) (T_STRING 10426) (T_VARIABLE 10393) (property_name 10433) (simple_variable 10428) ("{" 10429)) 4451 (("$" 10391) (T_STRING 10426) (T_VARIABLE 10393) (property_name 10427) (simple_variable 10428) ("{" 10429)) 4452 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 10424) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4453 (("]" 10425)) 4454 (("(" 10407) (argument_list 10432)) 4455 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10430) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4456 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10431)) 4457 (("(" 10407) (argument_list 10434)) 4458 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10438) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4459 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10439)) 4460 ((backup_doc_comment 10442)) 4461 (("(" 10443)) 4462 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 10444)) 4463 ((")" 10445)) 4464 ((T_USE 164) (lexical_vars 10446)) 4465 ((":" 166) (return_type 10447)) 4466 ((backup_fn_flags 10448)) 4467 (("{" 10449)) 4468 ((inner_statement_list 10450)) 4469 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 10451) ("~" 127)) 4470 ((backup_fn_flags 10452)) 4471 ((backup_doc_comment 10454)) 4472 (("(" 10455)) 4473 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 10456)) 4474 ((")" 10457)) 4475 ((":" 2249) (return_type 10458)) 4476 ((T_DOUBLE_ARROW 10459)) 4477 ((backup_fn_flags 10460)) 4478 ((backup_lex_pos 10461)) 4479 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10525) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4480 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11499) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4481 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 11497) (encaps_var 5237)) 4482 (("$" 10464) (T_VARIABLE 10512) (simple_variable 11493) ("{" 11494)) 4483 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 11491) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4484 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11490) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4485 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11489) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4486 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11488) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4487 (("(" 11485)) 4488 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11484) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4489 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11483) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4490 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11482) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4491 (("\"" 1520) ("$" 10767) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 10768) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 10769) ("[" 1542) (array_object_dereferenceable 10770) (callable_expr 10771) (callable_variable 10772) (class_constant 1546) (class_name 10773) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 10774) (name 10775) (new_dereferenceable 1552) (simple_variable 10776) (static_member 10777) (variable 11481) (variable_class_name 10779)) 4492 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11480) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4493 (("(" 11477)) 4494 (("(" 11474)) 4495 (("(" 11465) (argument_list 11466) (ctor_arguments 11467)) 4496 (("\"" 1520) ("$" 10767) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 10768) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 10769) ("[" 1542) (array_object_dereferenceable 10770) (callable_expr 10771) (callable_variable 10772) (class_constant 1546) (class_name 10773) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 10774) (name 10775) (new_dereferenceable 1552) (simple_variable 10776) (static_member 10777) (variable 11464) (variable_class_name 10779)) 4497 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11463) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4498 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11462) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4499 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11461) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4500 (("(" 11457)) 4501 (("(" 11452)) 4502 (("(" 11446)) 4503 (("$" 10767) ("(" 11409) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 11410) (T_NAME_QUALIFIED 11411) (T_NAME_RELATIVE 11412) (T_READONLY 2931) (T_STATIC 11413) (T_STRING 11414) (T_VARIABLE 10769) (anonymous_class 11415) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 11416) (attribute 2937) (attributes 11417) (class_modifier 2939) (class_name 11418) (class_name_reference 11419) (name 11420) (new_variable 11421) (simple_variable 11422)) 4504 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11408) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4505 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11407) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4506 (("(" 10728) (argument_list 11406)) 4507 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11405) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4508 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11404) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4509 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 11399) (T_END_HEREDOC 11400) (T_VARIABLE 2896) (encaps_list 11401) (encaps_var 2898)) 4510 ((T_FN 35) (T_FUNCTION 129) (fn 10526) #'10527 (inline_function 11398)) 4511 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11397) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4512 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11396) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4513 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11395) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4514 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10937) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4515 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10873) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4516 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 10869) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 4517 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 10867) (encaps_list 2804) (encaps_var 2805)) 4518 ((T_NULLSAFE_OBJECT_OPERATOR 10853) (T_OBJECT_OPERATOR 10854) ("[" 10855)) 4519 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 10850) (attribute 2733) (fn 10526) #'10527 (inline_function 10851)) 4520 (("(" 10728) (argument_list 10849)) 4521 ((T_PAAMAYIM_NEKUDOTAYIM 10841)) 4522 (("%" 10542) ("*" 10543) ("+" 10544) ("-" 10545) ("." 10546) ("/" 10547) ("<" 10548) (">" 10549) ("?" 10550) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 10551) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 10552) (T_BOOLEAN_AND 10553) (T_BOOLEAN_OR 10554) (T_COALESCE 10555) (T_INSTANCEOF 10556) (T_IS_EQUAL 10557) (T_IS_GREATER_OR_EQUAL 10558) (T_IS_IDENTICAL 10559) (T_IS_NOT_EQUAL 10560) (T_IS_NOT_IDENTICAL 10561) (T_IS_SMALLER_OR_EQUAL 10562) (T_LOGICAL_AND 10563) (T_LOGICAL_OR 10564) (T_LOGICAL_XOR 10565) (T_POW 10566) (T_SL 10567) (T_SPACESHIP 10568) (T_SR 10569) ("^" 10570) (backup_fn_flags 10840) ("|" 10571)) 4523 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 10829)) 4524 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 10817)) 4525 (("(" 10728) (argument_list 10816)) 4526 (("=" 10736) (T_AND_EQUAL 10737) (T_COALESCE_EQUAL 10738) (T_CONCAT_EQUAL 10739) (T_DEC 10740) (T_DIV_EQUAL 10741) (T_INC 10742) (T_MINUS_EQUAL 10743) (T_MOD_EQUAL 10744) (T_MUL_EQUAL 10745) (T_OR_EQUAL 10746) (T_PLUS_EQUAL 10747) (T_POW_EQUAL 10748) (T_SL_EQUAL 10749) (T_SR_EQUAL 10750) (T_XOR_EQUAL 10751)) 4527 ((T_PAAMAYIM_NEKUDOTAYIM 10640)) 4528 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10541) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4529 (("%" 10542) ("*" 10543) ("+" 10544) ("-" 10545) ("." 10546) ("/" 10547) ("<" 10548) (">" 10549) ("?" 10550) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 10551) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 10552) (T_BOOLEAN_AND 10553) (T_BOOLEAN_OR 10554) (T_COALESCE 10555) (T_INSTANCEOF 10556) (T_IS_EQUAL 10557) (T_IS_GREATER_OR_EQUAL 10558) (T_IS_IDENTICAL 10559) (T_IS_NOT_EQUAL 10560) (T_IS_NOT_IDENTICAL 10561) (T_IS_SMALLER_OR_EQUAL 10562) (T_LOGICAL_AND 10563) (T_LOGICAL_OR 10564) (T_LOGICAL_XOR 10565) (T_POW 10566) (T_SL 10567) (T_SPACESHIP 10568) (T_SR 10569) ("^" 10570) ("|" 10571)) 4530 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10639) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4531 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10638) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4532 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10637) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4533 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10636) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4534 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10635) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4535 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10634) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4536 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10633) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4537 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10632) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4538 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 10627) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 10628) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4539 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10626) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4540 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10625) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4541 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10624) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4542 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10623) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4543 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10622) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4544 (("$" 10587) ("(" 10588) (T_NAME_FULLY_QUALIFIED 10589) (T_NAME_QUALIFIED 10590) (T_NAME_RELATIVE 10591) (T_STATIC 10592) (T_STRING 10593) (T_VARIABLE 10594) (class_name 10595) (class_name_reference 10596) (name 10597) (new_variable 10598) (simple_variable 10599)) 4545 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10586) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4546 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10585) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4547 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10584) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4548 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10583) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4549 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10582) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4550 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10581) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4551 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10580) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4552 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10579) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4553 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10578) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4554 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10577) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4555 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10576) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4556 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10575) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4557 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10574) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4558 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10573) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4559 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10572) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4560 (("$" 10587) (T_VARIABLE 10594) (simple_variable 10618) ("{" 10619)) 4561 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 10616) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4562 ((T_PAAMAYIM_NEKUDOTAYIM 10614)) 4563 ((T_NULLSAFE_OBJECT_OPERATOR 10600) (T_OBJECT_OPERATOR 10601) (T_PAAMAYIM_NEKUDOTAYIM 10602) ("[" 10603)) 4564 (("$" 10587) (T_STRING 10607) (T_VARIABLE 10594) (property_name 10613) (simple_variable 10609) ("{" 10610)) 4565 (("$" 10587) (T_STRING 10607) (T_VARIABLE 10594) (property_name 10608) (simple_variable 10609) ("{" 10610)) 4566 (("$" 10587) (T_VARIABLE 10594) (simple_variable 10606)) 4567 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 10604) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4568 (("]" 10605)) 4569 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10611) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4570 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10612)) 4571 (("$" 10587) (T_VARIABLE 10594) (simple_variable 10615)) 4572 (("%" 299) (")" 10617) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4573 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10620) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4574 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10621)) 4575 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10631) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4576 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 10629) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 4577 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10630) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4578 (("$" 10464) (T_ABSTRACT 10641) (T_ARRAY 10642) (T_AS 10643) (T_BREAK 10644) (T_CALLABLE 10645) (T_CASE 10646) (T_CATCH 10647) (T_CLASS 10648) (T_CLASS_C 10649) (T_CLONE 10650) (T_CONST 10651) (T_CONTINUE 10652) (T_DECLARE 10653) (T_DEFAULT 10654) (T_DIR 10655) (T_DO 10656) (T_ECHO 10657) (T_ELSE 10658) (T_ELSEIF 10659) (T_EMPTY 10660) (T_ENDDECLARE 10661) (T_ENDFOR 10662) (T_ENDFOREACH 10663) (T_ENDIF 10664) (T_ENDSWITCH 10665) (T_ENDWHILE 10666) (T_ENUM 10667) (T_EVAL 10668) (T_EXIT 10669) (T_EXTENDS 10670) (T_FILE 10671) (T_FINAL 10672) (T_FINALLY 10673) (T_FN 10674) (T_FOR 10675) (T_FOREACH 10676) (T_FUNCTION 10677) (T_FUNC_C 10678) (T_GLOBAL 10679) (T_GOTO 10680) (T_IF 10681) (T_IMPLEMENTS 10682) (T_INCLUDE 10683) (T_INCLUDE_ONCE 10684) (T_INSTANCEOF 10685) (T_INSTEADOF 10686) (T_INTERFACE 10687) (T_ISSET 10688) (T_LINE 10689) (T_LIST 10690) (T_LOGICAL_AND 10691) (T_LOGICAL_OR 10692) (T_LOGICAL_XOR 10693) (T_MATCH 10694) (T_METHOD_C 10695) (T_NAMESPACE 10696) (T_NEW 10697) (T_NS_C 10698) (T_PRINT 10699) (T_PRIVATE 10700) (T_PROPERTY_C 10701) (T_PROTECTED 10702) (T_PUBLIC 10703) (T_READONLY 10704) (T_REQUIRE 10705) (T_REQUIRE_ONCE 10706) (T_RETURN 10707) (T_STATIC 10708) (T_STRING 10709) (T_SWITCH 10710) (T_THROW 10711) (T_TRAIT 10712) (T_TRAIT_C 10713) (T_TRY 10714) (T_UNSET 10715) (T_USE 10716) (T_VAR 10717) (T_VARIABLE 10512) (T_WHILE 10718) (T_YIELD 10719) (identifier 10720) (member_name 10721) (reserved_non_modifiers 10722) (semi_reserved 10723) (simple_variable 10724) ("{" 10725)) 4579 (("(" 10728) (argument_list 10729)) 4580 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10726) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4581 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10727)) 4582 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 10730) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 10731) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 10732) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4583 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 10735) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4584 (("," 1480) (possible_comma 10733)) 4585 ((")" 10734)) 4586 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (ampersand 10765) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10766) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4587 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10764) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4588 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10763) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4589 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10762) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4590 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10761) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4591 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10760) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4592 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10759) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4593 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10758) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4594 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10757) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4595 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10756) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4596 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10755) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4597 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10754) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4598 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10753) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4599 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10752) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4600 (("\"" 1520) ("$" 10767) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 10768) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 10769) ("[" 1542) (array_object_dereferenceable 10770) (callable_expr 10771) (callable_variable 10772) (class_constant 1546) (class_name 10773) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 10774) (name 10775) (new_dereferenceable 1552) (simple_variable 10776) (static_member 10777) (variable 10778) (variable_class_name 10779)) 4601 (("$" 10767) (T_VARIABLE 10769) (simple_variable 10812) ("{" 10813)) 4602 (("(" 10783) (argument_list 10811)) 4603 ((T_NULLSAFE_OBJECT_OPERATOR 10797) (T_OBJECT_OPERATOR 10798) ("[" 10799)) 4604 (("(" 10783) (argument_list 10796)) 4605 ((T_PAAMAYIM_NEKUDOTAYIM 10792)) 4606 (("(" 10783) (argument_list 10791)) 4607 ((T_PAAMAYIM_NEKUDOTAYIM 10780)) 4608 (("$" 10767) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 10769) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 10781) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 10782) ("{" 1642)) 4609 (("(" 10783) (argument_list 10784)) 4610 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 10785) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 10786) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 10787) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4611 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 10790) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4612 (("," 1480) (possible_comma 10788)) 4613 ((")" 10789)) 4614 (("$" 10767) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 10769) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 10793) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 10794) ("{" 1658)) 4615 (("(" 10783) (argument_list 10795)) 4616 (("$" 10767) (T_STRING 10802) (T_VARIABLE 10769) (property_name 10809) (simple_variable 10804) ("{" 10805)) 4617 (("$" 10767) (T_STRING 10802) (T_VARIABLE 10769) (property_name 10803) (simple_variable 10804) ("{" 10805)) 4618 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 10800) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4619 (("]" 10801)) 4620 (("(" 10783) (argument_list 10808)) 4621 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10806) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4622 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10807)) 4623 (("(" 10783) (argument_list 10810)) 4624 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10814) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4625 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10815)) 4626 ((backup_doc_comment 10818)) 4627 (("(" 10819)) 4628 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 10820)) 4629 ((")" 10821)) 4630 ((T_USE 164) (lexical_vars 10822)) 4631 ((":" 166) (return_type 10823)) 4632 ((backup_fn_flags 10824)) 4633 (("{" 10825)) 4634 ((inner_statement_list 10826)) 4635 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 10827) ("~" 127)) 4636 ((backup_fn_flags 10828)) 4637 ((backup_doc_comment 10830)) 4638 (("(" 10831)) 4639 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 10832)) 4640 ((")" 10833)) 4641 ((":" 2249) (return_type 10834)) 4642 ((T_DOUBLE_ARROW 10835)) 4643 ((backup_fn_flags 10836)) 4644 ((backup_lex_pos 10837)) 4645 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10838) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4646 (("%" 10542) ("*" 10543) ("+" 10544) ("-" 10545) ("." 10546) ("/" 10547) ("<" 10548) (">" 10549) ("?" 10550) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 10551) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 10552) (T_BOOLEAN_AND 10553) (T_BOOLEAN_OR 10554) (T_COALESCE 10555) (T_INSTANCEOF 10556) (T_IS_EQUAL 10557) (T_IS_GREATER_OR_EQUAL 10558) (T_IS_IDENTICAL 10559) (T_IS_NOT_EQUAL 10560) (T_IS_NOT_IDENTICAL 10561) (T_IS_SMALLER_OR_EQUAL 10562) (T_LOGICAL_AND 10563) (T_LOGICAL_OR 10564) (T_LOGICAL_XOR 10565) (T_POW 10566) (T_SL 10567) (T_SPACESHIP 10568) (T_SR 10569) ("^" 10570) (backup_fn_flags 10839) ("|" 10571)) 4647 (("$" 10464) (T_ABSTRACT 10641) (T_ARRAY 10642) (T_AS 10643) (T_BREAK 10644) (T_CALLABLE 10645) (T_CASE 10646) (T_CATCH 10647) (T_CLASS 10648) (T_CLASS_C 10649) (T_CLONE 10650) (T_CONST 10651) (T_CONTINUE 10652) (T_DECLARE 10653) (T_DEFAULT 10654) (T_DIR 10655) (T_DO 10656) (T_ECHO 10657) (T_ELSE 10658) (T_ELSEIF 10659) (T_EMPTY 10660) (T_ENDDECLARE 10661) (T_ENDFOR 10662) (T_ENDFOREACH 10663) (T_ENDIF 10664) (T_ENDSWITCH 10665) (T_ENDWHILE 10666) (T_ENUM 10667) (T_EVAL 10668) (T_EXIT 10669) (T_EXTENDS 10670) (T_FILE 10671) (T_FINAL 10672) (T_FINALLY 10673) (T_FN 10674) (T_FOR 10675) (T_FOREACH 10676) (T_FUNCTION 10677) (T_FUNC_C 10678) (T_GLOBAL 10679) (T_GOTO 10680) (T_IF 10681) (T_IMPLEMENTS 10682) (T_INCLUDE 10683) (T_INCLUDE_ONCE 10684) (T_INSTANCEOF 10685) (T_INSTEADOF 10686) (T_INTERFACE 10687) (T_ISSET 10688) (T_LINE 10689) (T_LIST 10690) (T_LOGICAL_AND 10691) (T_LOGICAL_OR 10692) (T_LOGICAL_XOR 10693) (T_MATCH 10694) (T_METHOD_C 10695) (T_NAMESPACE 10696) (T_NEW 10697) (T_NS_C 10698) (T_PRINT 10699) (T_PRIVATE 10700) (T_PROPERTY_C 10701) (T_PROTECTED 10702) (T_PUBLIC 10703) (T_READONLY 10704) (T_REQUIRE 10705) (T_REQUIRE_ONCE 10706) (T_RETURN 10707) (T_STATIC 10708) (T_STRING 10709) (T_SWITCH 10710) (T_THROW 10711) (T_TRAIT 10712) (T_TRAIT_C 10713) (T_TRY 10714) (T_UNSET 10715) (T_USE 10716) (T_VAR 10717) (T_VARIABLE 10512) (T_WHILE 10718) (T_YIELD 10719) (identifier 10842) (member_name 10843) (reserved_non_modifiers 10722) (semi_reserved 10723) (simple_variable 10844) ("{" 10845)) 4648 (("(" 10728) (argument_list 10848)) 4649 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10846) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4650 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10847)) 4651 ((T_FN 35) (T_FUNCTION 129) (fn 10526) #'10527 (inline_function 10852)) 4652 (("$" 10464) (T_STRING 10858) (T_VARIABLE 10512) (property_name 10865) (simple_variable 10860) ("{" 10861)) 4653 (("$" 10464) (T_STRING 10858) (T_VARIABLE 10512) (property_name 10859) (simple_variable 10860) ("{" 10861)) 4654 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 10856) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4655 (("]" 10857)) 4656 (("(" 10728) (argument_list 10864)) 4657 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 10862) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4658 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 10863)) 4659 (("(" 10728) (argument_list 10866)) 4660 (("`" 10868)) 4661 (("]" 10870)) 4662 (("=" 10871)) 4663 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 10872) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4664 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11394) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4665 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 11392) (encaps_var 5237)) 4666 (("$" 10876) (T_VARIABLE 10924) (simple_variable 11388) ("{" 11389)) 4667 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 11386) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4668 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11385) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4669 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11384) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4670 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11383) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4671 (("(" 11380)) 4672 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11379) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4673 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11378) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4674 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11377) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4675 (("\"" 1520) ("$" 11179) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 11180) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 11181) ("[" 1542) (array_object_dereferenceable 11182) (callable_expr 11183) (callable_variable 11184) (class_constant 1546) (class_name 11185) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 11186) (name 11187) (new_dereferenceable 1552) (simple_variable 11188) (static_member 11189) (variable 11376) (variable_class_name 11191)) 4676 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11375) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4677 (("(" 11372)) 4678 (("(" 11369)) 4679 (("(" 11360) (argument_list 11361) (ctor_arguments 11362)) 4680 (("\"" 1520) ("$" 11179) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 11180) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 11181) ("[" 1542) (array_object_dereferenceable 11182) (callable_expr 11183) (callable_variable 11184) (class_constant 1546) (class_name 11185) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 11186) (name 11187) (new_dereferenceable 1552) (simple_variable 11188) (static_member 11189) (variable 11359) (variable_class_name 11191)) 4681 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11358) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4682 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11357) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4683 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11356) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4684 (("(" 11352)) 4685 (("(" 11347)) 4686 (("(" 11341)) 4687 (("$" 11179) ("(" 11304) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 11305) (T_NAME_QUALIFIED 11306) (T_NAME_RELATIVE 11307) (T_READONLY 2931) (T_STATIC 11308) (T_STRING 11309) (T_VARIABLE 11181) (anonymous_class 11310) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 11311) (attribute 2937) (attributes 11312) (class_modifier 2939) (class_name 11313) (class_name_reference 11314) (name 11315) (new_variable 11316) (simple_variable 11317)) 4688 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11303) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4689 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11302) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4690 (("(" 11140) (argument_list 11301)) 4691 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11300) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4692 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11299) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4693 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 11294) (T_END_HEREDOC 11295) (T_VARIABLE 2896) (encaps_list 11296) (encaps_var 2898)) 4694 ((T_FN 35) (T_FUNCTION 129) (fn 10938) #'10939 (inline_function 11293)) 4695 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11292) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4696 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11291) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4697 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11290) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4698 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11287) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4699 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11286) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4700 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 11282) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 4701 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 11280) (encaps_list 2804) (encaps_var 2805)) 4702 ((T_NULLSAFE_OBJECT_OPERATOR 11266) (T_OBJECT_OPERATOR 11267) ("[" 11268)) 4703 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 11263) (attribute 2733) (fn 10938) #'10939 (inline_function 11264)) 4704 (("(" 11140) (argument_list 11262)) 4705 ((T_PAAMAYIM_NEKUDOTAYIM 11254)) 4706 (("%" 10954) ("*" 10955) ("+" 10956) ("-" 10957) ("." 10958) ("/" 10959) ("<" 10960) (">" 10961) ("?" 10962) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 10963) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 10964) (T_BOOLEAN_AND 10965) (T_BOOLEAN_OR 10966) (T_COALESCE 10967) (T_DOUBLE_ARROW 11252) (T_INSTANCEOF 10968) (T_IS_EQUAL 10969) (T_IS_GREATER_OR_EQUAL 10970) (T_IS_IDENTICAL 10971) (T_IS_NOT_EQUAL 10972) (T_IS_NOT_IDENTICAL 10973) (T_IS_SMALLER_OR_EQUAL 10974) (T_LOGICAL_AND 10975) (T_LOGICAL_OR 10976) (T_LOGICAL_XOR 10977) (T_POW 10978) (T_SL 10979) (T_SPACESHIP 10980) (T_SR 10981) ("^" 10982) ("|" 10983)) 4707 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 11241)) 4708 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 11229)) 4709 (("(" 11140) (argument_list 11228)) 4710 (("=" 11148) (T_AND_EQUAL 11149) (T_COALESCE_EQUAL 11150) (T_CONCAT_EQUAL 11151) (T_DEC 11152) (T_DIV_EQUAL 11153) (T_INC 11154) (T_MINUS_EQUAL 11155) (T_MOD_EQUAL 11156) (T_MUL_EQUAL 11157) (T_OR_EQUAL 11158) (T_PLUS_EQUAL 11159) (T_POW_EQUAL 11160) (T_SL_EQUAL 11161) (T_SR_EQUAL 11162) (T_XOR_EQUAL 11163)) 4711 ((T_PAAMAYIM_NEKUDOTAYIM 11052)) 4712 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10953) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4713 (("%" 10954) ("*" 10955) ("+" 10956) ("-" 10957) ("." 10958) ("/" 10959) ("<" 10960) (">" 10961) ("?" 10962) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 10963) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 10964) (T_BOOLEAN_AND 10965) (T_BOOLEAN_OR 10966) (T_COALESCE 10967) (T_INSTANCEOF 10968) (T_IS_EQUAL 10969) (T_IS_GREATER_OR_EQUAL 10970) (T_IS_IDENTICAL 10971) (T_IS_NOT_EQUAL 10972) (T_IS_NOT_IDENTICAL 10973) (T_IS_SMALLER_OR_EQUAL 10974) (T_LOGICAL_AND 10975) (T_LOGICAL_OR 10976) (T_LOGICAL_XOR 10977) (T_POW 10978) (T_SL 10979) (T_SPACESHIP 10980) (T_SR 10981) ("^" 10982) ("|" 10983)) 4714 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11051) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4715 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11050) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4716 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11049) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4717 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11048) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4718 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11047) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4719 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11046) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4720 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11045) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4721 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11044) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4722 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 11039) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 11040) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4723 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11038) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4724 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11037) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4725 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11036) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4726 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11035) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4727 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11034) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4728 (("$" 10999) ("(" 11000) (T_NAME_FULLY_QUALIFIED 11001) (T_NAME_QUALIFIED 11002) (T_NAME_RELATIVE 11003) (T_STATIC 11004) (T_STRING 11005) (T_VARIABLE 11006) (class_name 11007) (class_name_reference 11008) (name 11009) (new_variable 11010) (simple_variable 11011)) 4729 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10998) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4730 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10997) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4731 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10996) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4732 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10995) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4733 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10994) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4734 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10993) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4735 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10992) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4736 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10991) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4737 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10990) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4738 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10989) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4739 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10988) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4740 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10987) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4741 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10986) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4742 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10985) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4743 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 10984) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4744 (("$" 10999) (T_VARIABLE 11006) (simple_variable 11030) ("{" 11031)) 4745 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 11028) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4746 ((T_PAAMAYIM_NEKUDOTAYIM 11026)) 4747 ((T_NULLSAFE_OBJECT_OPERATOR 11012) (T_OBJECT_OPERATOR 11013) (T_PAAMAYIM_NEKUDOTAYIM 11014) ("[" 11015)) 4748 (("$" 10999) (T_STRING 11019) (T_VARIABLE 11006) (property_name 11025) (simple_variable 11021) ("{" 11022)) 4749 (("$" 10999) (T_STRING 11019) (T_VARIABLE 11006) (property_name 11020) (simple_variable 11021) ("{" 11022)) 4750 (("$" 10999) (T_VARIABLE 11006) (simple_variable 11018)) 4751 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 11016) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4752 (("]" 11017)) 4753 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11023) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4754 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11024)) 4755 (("$" 10999) (T_VARIABLE 11006) (simple_variable 11027)) 4756 (("%" 299) (")" 11029) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4757 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11032) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4758 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11033)) 4759 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11043) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4760 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 11041) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 4761 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11042) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4762 (("$" 10876) (T_ABSTRACT 11053) (T_ARRAY 11054) (T_AS 11055) (T_BREAK 11056) (T_CALLABLE 11057) (T_CASE 11058) (T_CATCH 11059) (T_CLASS 11060) (T_CLASS_C 11061) (T_CLONE 11062) (T_CONST 11063) (T_CONTINUE 11064) (T_DECLARE 11065) (T_DEFAULT 11066) (T_DIR 11067) (T_DO 11068) (T_ECHO 11069) (T_ELSE 11070) (T_ELSEIF 11071) (T_EMPTY 11072) (T_ENDDECLARE 11073) (T_ENDFOR 11074) (T_ENDFOREACH 11075) (T_ENDIF 11076) (T_ENDSWITCH 11077) (T_ENDWHILE 11078) (T_ENUM 11079) (T_EVAL 11080) (T_EXIT 11081) (T_EXTENDS 11082) (T_FILE 11083) (T_FINAL 11084) (T_FINALLY 11085) (T_FN 11086) (T_FOR 11087) (T_FOREACH 11088) (T_FUNCTION 11089) (T_FUNC_C 11090) (T_GLOBAL 11091) (T_GOTO 11092) (T_IF 11093) (T_IMPLEMENTS 11094) (T_INCLUDE 11095) (T_INCLUDE_ONCE 11096) (T_INSTANCEOF 11097) (T_INSTEADOF 11098) (T_INTERFACE 11099) (T_ISSET 11100) (T_LINE 11101) (T_LIST 11102) (T_LOGICAL_AND 11103) (T_LOGICAL_OR 11104) (T_LOGICAL_XOR 11105) (T_MATCH 11106) (T_METHOD_C 11107) (T_NAMESPACE 11108) (T_NEW 11109) (T_NS_C 11110) (T_PRINT 11111) (T_PRIVATE 11112) (T_PROPERTY_C 11113) (T_PROTECTED 11114) (T_PUBLIC 11115) (T_READONLY 11116) (T_REQUIRE 11117) (T_REQUIRE_ONCE 11118) (T_RETURN 11119) (T_STATIC 11120) (T_STRING 11121) (T_SWITCH 11122) (T_THROW 11123) (T_TRAIT 11124) (T_TRAIT_C 11125) (T_TRY 11126) (T_UNSET 11127) (T_USE 11128) (T_VAR 11129) (T_VARIABLE 10924) (T_WHILE 11130) (T_YIELD 11131) (identifier 11132) (member_name 11133) (reserved_non_modifiers 11134) (semi_reserved 11135) (simple_variable 11136) ("{" 11137)) 4763 (("(" 11140) (argument_list 11141)) 4764 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11138) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4765 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11139)) 4766 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 11142) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 11143) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 11144) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4767 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 11147) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4768 (("," 1480) (possible_comma 11145)) 4769 ((")" 11146)) 4770 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (ampersand 11177) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11178) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4771 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11176) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4772 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11175) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4773 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11174) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4774 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11173) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4775 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11172) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4776 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11171) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4777 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11170) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4778 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11169) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4779 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11168) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4780 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11167) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4781 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11166) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4782 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11165) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4783 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11164) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4784 (("\"" 1520) ("$" 11179) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 11180) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 11181) ("[" 1542) (array_object_dereferenceable 11182) (callable_expr 11183) (callable_variable 11184) (class_constant 1546) (class_name 11185) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 11186) (name 11187) (new_dereferenceable 1552) (simple_variable 11188) (static_member 11189) (variable 11190) (variable_class_name 11191)) 4785 (("$" 11179) (T_VARIABLE 11181) (simple_variable 11224) ("{" 11225)) 4786 (("(" 11195) (argument_list 11223)) 4787 ((T_NULLSAFE_OBJECT_OPERATOR 11209) (T_OBJECT_OPERATOR 11210) ("[" 11211)) 4788 (("(" 11195) (argument_list 11208)) 4789 ((T_PAAMAYIM_NEKUDOTAYIM 11204)) 4790 (("(" 11195) (argument_list 11203)) 4791 ((T_PAAMAYIM_NEKUDOTAYIM 11192)) 4792 (("$" 11179) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 11181) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 11193) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 11194) ("{" 1642)) 4793 (("(" 11195) (argument_list 11196)) 4794 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 11197) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 11198) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 11199) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4795 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 11202) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4796 (("," 1480) (possible_comma 11200)) 4797 ((")" 11201)) 4798 (("$" 11179) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 11181) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 11205) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 11206) ("{" 1658)) 4799 (("(" 11195) (argument_list 11207)) 4800 (("$" 11179) (T_STRING 11214) (T_VARIABLE 11181) (property_name 11221) (simple_variable 11216) ("{" 11217)) 4801 (("$" 11179) (T_STRING 11214) (T_VARIABLE 11181) (property_name 11215) (simple_variable 11216) ("{" 11217)) 4802 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 11212) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4803 (("]" 11213)) 4804 (("(" 11195) (argument_list 11220)) 4805 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11218) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4806 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11219)) 4807 (("(" 11195) (argument_list 11222)) 4808 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11226) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4809 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11227)) 4810 ((backup_doc_comment 11230)) 4811 (("(" 11231)) 4812 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 11232)) 4813 ((")" 11233)) 4814 ((T_USE 164) (lexical_vars 11234)) 4815 ((":" 166) (return_type 11235)) 4816 ((backup_fn_flags 11236)) 4817 (("{" 11237)) 4818 ((inner_statement_list 11238)) 4819 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 11239) ("~" 127)) 4820 ((backup_fn_flags 11240)) 4821 ((backup_doc_comment 11242)) 4822 (("(" 11243)) 4823 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 11244)) 4824 ((")" 11245)) 4825 ((":" 2249) (return_type 11246)) 4826 ((T_DOUBLE_ARROW 11247)) 4827 ((backup_fn_flags 11248)) 4828 ((backup_lex_pos 11249)) 4829 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11250) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4830 (("%" 10954) ("*" 10955) ("+" 10956) ("-" 10957) ("." 10958) ("/" 10959) ("<" 10960) (">" 10961) ("?" 10962) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 10963) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 10964) (T_BOOLEAN_AND 10965) (T_BOOLEAN_OR 10966) (T_COALESCE 10967) (T_INSTANCEOF 10968) (T_IS_EQUAL 10969) (T_IS_GREATER_OR_EQUAL 10970) (T_IS_IDENTICAL 10971) (T_IS_NOT_EQUAL 10972) (T_IS_NOT_IDENTICAL 10973) (T_IS_SMALLER_OR_EQUAL 10974) (T_LOGICAL_AND 10975) (T_LOGICAL_OR 10976) (T_LOGICAL_XOR 10977) (T_POW 10978) (T_SL 10979) (T_SPACESHIP 10980) (T_SR 10981) ("^" 10982) (backup_fn_flags 11251) ("|" 10983)) 4831 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11253) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4832 (("$" 10876) (T_ABSTRACT 11053) (T_ARRAY 11054) (T_AS 11055) (T_BREAK 11056) (T_CALLABLE 11057) (T_CASE 11058) (T_CATCH 11059) (T_CLASS 11060) (T_CLASS_C 11061) (T_CLONE 11062) (T_CONST 11063) (T_CONTINUE 11064) (T_DECLARE 11065) (T_DEFAULT 11066) (T_DIR 11067) (T_DO 11068) (T_ECHO 11069) (T_ELSE 11070) (T_ELSEIF 11071) (T_EMPTY 11072) (T_ENDDECLARE 11073) (T_ENDFOR 11074) (T_ENDFOREACH 11075) (T_ENDIF 11076) (T_ENDSWITCH 11077) (T_ENDWHILE 11078) (T_ENUM 11079) (T_EVAL 11080) (T_EXIT 11081) (T_EXTENDS 11082) (T_FILE 11083) (T_FINAL 11084) (T_FINALLY 11085) (T_FN 11086) (T_FOR 11087) (T_FOREACH 11088) (T_FUNCTION 11089) (T_FUNC_C 11090) (T_GLOBAL 11091) (T_GOTO 11092) (T_IF 11093) (T_IMPLEMENTS 11094) (T_INCLUDE 11095) (T_INCLUDE_ONCE 11096) (T_INSTANCEOF 11097) (T_INSTEADOF 11098) (T_INTERFACE 11099) (T_ISSET 11100) (T_LINE 11101) (T_LIST 11102) (T_LOGICAL_AND 11103) (T_LOGICAL_OR 11104) (T_LOGICAL_XOR 11105) (T_MATCH 11106) (T_METHOD_C 11107) (T_NAMESPACE 11108) (T_NEW 11109) (T_NS_C 11110) (T_PRINT 11111) (T_PRIVATE 11112) (T_PROPERTY_C 11113) (T_PROTECTED 11114) (T_PUBLIC 11115) (T_READONLY 11116) (T_REQUIRE 11117) (T_REQUIRE_ONCE 11118) (T_RETURN 11119) (T_STATIC 11120) (T_STRING 11121) (T_SWITCH 11122) (T_THROW 11123) (T_TRAIT 11124) (T_TRAIT_C 11125) (T_TRY 11126) (T_UNSET 11127) (T_USE 11128) (T_VAR 11129) (T_VARIABLE 10924) (T_WHILE 11130) (T_YIELD 11131) (identifier 11255) (member_name 11256) (reserved_non_modifiers 11134) (semi_reserved 11135) (simple_variable 11257) ("{" 11258)) 4833 (("(" 11140) (argument_list 11261)) 4834 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11259) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4835 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11260)) 4836 ((T_FN 35) (T_FUNCTION 129) (fn 10938) #'10939 (inline_function 11265)) 4837 (("$" 10876) (T_STRING 11271) (T_VARIABLE 10924) (property_name 11278) (simple_variable 11273) ("{" 11274)) 4838 (("$" 10876) (T_STRING 11271) (T_VARIABLE 10924) (property_name 11272) (simple_variable 11273) ("{" 11274)) 4839 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 11269) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4840 (("]" 11270)) 4841 (("(" 11140) (argument_list 11277)) 4842 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11275) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4843 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11276)) 4844 (("(" 11140) (argument_list 11279)) 4845 (("`" 11281)) 4846 (("]" 11283)) 4847 (("=" 11284)) 4848 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11285) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4849 (("%" 10954) ("*" 10955) ("+" 10956) ("-" 10957) ("." 10958) ("/" 10959) ("<" 10960) (">" 10961) ("?" 10962) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 10963) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 10964) (T_BOOLEAN_AND 10965) (T_BOOLEAN_OR 10966) (T_COALESCE 10967) (T_DOUBLE_ARROW 11288) (T_INSTANCEOF 10968) (T_IS_EQUAL 10969) (T_IS_GREATER_OR_EQUAL 10970) (T_IS_IDENTICAL 10971) (T_IS_NOT_EQUAL 10972) (T_IS_NOT_IDENTICAL 10973) (T_IS_SMALLER_OR_EQUAL 10974) (T_LOGICAL_AND 10975) (T_LOGICAL_OR 10976) (T_LOGICAL_XOR 10977) (T_POW 10978) (T_SL 10979) (T_SPACESHIP 10980) (T_SR 10981) ("^" 10982) ("|" 10983)) 4850 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11289) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4851 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 11298) (T_VARIABLE 2896) (encaps_var 2910)) 4852 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 11297) (T_VARIABLE 2896) (encaps_var 2901)) 4853 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 11339) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4854 ((T_CLASS 11331)) 4855 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 11330) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 11311) (attribute 2958) (class_modifier 2939)) 4856 ((T_PAAMAYIM_NEKUDOTAYIM 11328)) 4857 (("(" 11195) (argument_list 11327)) 4858 ((T_NULLSAFE_OBJECT_OPERATOR 11318) (T_OBJECT_OPERATOR 11319) (T_PAAMAYIM_NEKUDOTAYIM 11320) ("[" 11321)) 4859 (("$" 11179) (T_STRING 11214) (T_VARIABLE 11181) (property_name 11326) (simple_variable 11216) ("{" 11217)) 4860 (("$" 11179) (T_STRING 11214) (T_VARIABLE 11181) (property_name 11325) (simple_variable 11216) ("{" 11217)) 4861 (("$" 11179) (T_VARIABLE 11181) (simple_variable 11324)) 4862 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 11322) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4863 (("]" 11323)) 4864 (("$" 11179) (T_VARIABLE 11181) (simple_variable 11329)) 4865 (("(" 2960) (argument_list 2961) (ctor_arguments 11332)) 4866 ((T_EXTENDS 2963) (extends_from 11333)) 4867 ((T_IMPLEMENTS 2965) (implements_list 11334)) 4868 ((backup_doc_comment 11335)) 4869 (("{" 11336)) 4870 ((class_statement_list 11337)) 4871 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 11338)) 4872 (("%" 299) (")" 11340) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4873 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 11342) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4874 (("%" 299) (")" 11343) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4875 (("{" 11344)) 4876 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 11345) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 4877 (("}" 11346)) 4878 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 11348) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 4879 ((")" 11349)) 4880 (("=" 11350)) 4881 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11351) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 4882 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 11353) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4883 (("," 5192) (possible_comma 11354)) 4884 ((")" 11355)) 4885 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 11363) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 11364) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 11365) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4886 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 11368) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4887 (("," 1480) (possible_comma 11366)) 4888 ((")" 11367)) 4889 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 11370) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4890 (("%" 299) (")" 11371) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4891 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 11373) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4892 (("%" 299) (")" 11374) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4893 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 11381) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 4894 ((")" 11382)) 4895 (("%" 299) (")" 11387) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4896 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11390) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4897 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11391)) 4898 (("\"" 11393) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 4899 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 11403) (T_VARIABLE 2896) (encaps_var 2910)) 4900 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 11402) (T_VARIABLE 2896) (encaps_var 2901)) 4901 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 11444) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4902 ((T_CLASS 11436)) 4903 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 11435) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 11416) (attribute 2958) (class_modifier 2939)) 4904 ((T_PAAMAYIM_NEKUDOTAYIM 11433)) 4905 (("(" 10783) (argument_list 11432)) 4906 ((T_NULLSAFE_OBJECT_OPERATOR 11423) (T_OBJECT_OPERATOR 11424) (T_PAAMAYIM_NEKUDOTAYIM 11425) ("[" 11426)) 4907 (("$" 10767) (T_STRING 10802) (T_VARIABLE 10769) (property_name 11431) (simple_variable 10804) ("{" 10805)) 4908 (("$" 10767) (T_STRING 10802) (T_VARIABLE 10769) (property_name 11430) (simple_variable 10804) ("{" 10805)) 4909 (("$" 10767) (T_VARIABLE 10769) (simple_variable 11429)) 4910 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 11427) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4911 (("]" 11428)) 4912 (("$" 10767) (T_VARIABLE 10769) (simple_variable 11434)) 4913 (("(" 2960) (argument_list 2961) (ctor_arguments 11437)) 4914 ((T_EXTENDS 2963) (extends_from 11438)) 4915 ((T_IMPLEMENTS 2965) (implements_list 11439)) 4916 ((backup_doc_comment 11440)) 4917 (("{" 11441)) 4918 ((class_statement_list 11442)) 4919 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 11443)) 4920 (("%" 299) (")" 11445) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4921 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 11447) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4922 (("%" 299) (")" 11448) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4923 (("{" 11449)) 4924 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 11450) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 4925 (("}" 11451)) 4926 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 11453) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 4927 ((")" 11454)) 4928 (("=" 11455)) 4929 (("!" 10462) ("\"" 10463) ("$" 10464) ("(" 10465) ("+" 10466) ("-" 10467) ("@" 10468) (T_ARRAY 10469) (T_ARRAY_CAST 10470) (T_ATTRIBUTE 128) (T_BOOL_CAST 10471) (T_CLASS_C 10472) (T_CLONE 10473) (T_CONSTANT_ENCAPSED_STRING 10474) (T_DEC 10475) (T_DIR 10476) (T_DNUMBER 10477) (T_DOUBLE_CAST 10478) (T_EMPTY 10479) (T_EVAL 10480) (T_EXIT 10481) (T_FILE 10482) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10483) (T_INC 10484) (T_INCLUDE 10485) (T_INCLUDE_ONCE 10486) (T_INT_CAST 10487) (T_ISSET 10488) (T_LINE 10489) (T_LIST 10490) (T_LNUMBER 10491) (T_MATCH 10492) (T_METHOD_C 10493) (T_NAME_FULLY_QUALIFIED 10494) (T_NAME_QUALIFIED 10495) (T_NAME_RELATIVE 10496) (T_NEW 10497) (T_NS_C 10498) (T_OBJECT_CAST 10499) (T_PRINT 10500) (T_PROPERTY_C 10501) (T_READONLY 10502) (T_REQUIRE 10503) (T_REQUIRE_ONCE 10504) (T_START_HEREDOC 10505) (T_STATIC 10506) (T_STRING 10507) (T_STRING_CAST 10508) (T_THROW 10509) (T_TRAIT_C 10510) (T_UNSET_CAST 10511) (T_VARIABLE 10512) (T_YIELD 10513) (T_YIELD_FROM 10514) ("[" 10515) ("`" 10516) (array_object_dereferenceable 10517) (attribute 133) (attributes 10518) (callable_expr 10519) (callable_variable 10520) (class_constant 10521) (class_name 10522) (constant 10523) (dereferenceable_scalar 10524) (expr 11456) (fn 10526) (fully_dereferenceable 105) #'10527 (function_call 10528) (inline_function 10529) (internal_functions_in_yacc 10530) (match 10531) (name 10532) (new_dereferenceable 10533) (new_non_dereferenceable 10534) (scalar 10535) (simple_variable 10536) (static_member 10537) (variable 10538) (variable_class_name 10539) ("~" 10540)) 4930 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 11458) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4931 (("," 5192) (possible_comma 11459)) 4932 ((")" 11460)) 4933 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 11468) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 11469) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 11470) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4934 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 11473) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 4935 (("," 1480) (possible_comma 11471)) 4936 ((")" 11472)) 4937 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 11475) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4938 (("%" 299) (")" 11476) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4939 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 11478) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4940 (("%" 299) (")" 11479) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4941 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 11486) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 4942 ((")" 11487)) 4943 (("%" 299) (")" 11492) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 4944 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11495) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4945 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11496)) 4946 (("\"" 11498) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 4947 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11501) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4948 (("$" 678) (T_ABSTRACT 856) (T_ARRAY 857) (T_AS 858) (T_BREAK 859) (T_CALLABLE 860) (T_CASE 861) (T_CATCH 862) (T_CLASS 863) (T_CLASS_C 864) (T_CLONE 865) (T_CONST 866) (T_CONTINUE 867) (T_DECLARE 868) (T_DEFAULT 869) (T_DIR 870) (T_DO 871) (T_ECHO 872) (T_ELSE 873) (T_ELSEIF 874) (T_EMPTY 875) (T_ENDDECLARE 876) (T_ENDFOR 877) (T_ENDFOREACH 878) (T_ENDIF 879) (T_ENDSWITCH 880) (T_ENDWHILE 881) (T_ENUM 882) (T_EVAL 883) (T_EXIT 884) (T_EXTENDS 885) (T_FILE 886) (T_FINAL 887) (T_FINALLY 888) (T_FN 889) (T_FOR 890) (T_FOREACH 891) (T_FUNCTION 892) (T_FUNC_C 893) (T_GLOBAL 894) (T_GOTO 895) (T_IF 896) (T_IMPLEMENTS 897) (T_INCLUDE 898) (T_INCLUDE_ONCE 899) (T_INSTANCEOF 900) (T_INSTEADOF 901) (T_INTERFACE 902) (T_ISSET 903) (T_LINE 904) (T_LIST 905) (T_LOGICAL_AND 906) (T_LOGICAL_OR 907) (T_LOGICAL_XOR 908) (T_MATCH 909) (T_METHOD_C 910) (T_NAMESPACE 911) (T_NEW 912) (T_NS_C 913) (T_PRINT 914) (T_PRIVATE 915) (T_PROPERTY_C 916) (T_PROTECTED 917) (T_PUBLIC 918) (T_READONLY 919) (T_REQUIRE 920) (T_REQUIRE_ONCE 921) (T_RETURN 922) (T_STATIC 923) (T_STRING 924) (T_SWITCH 925) (T_THROW 926) (T_TRAIT 927) (T_TRAIT_C 928) (T_TRY 929) (T_UNSET 930) (T_USE 931) (T_VAR 932) (T_VARIABLE 727) (T_WHILE 933) (T_YIELD 934) (identifier 11503) (member_name 11504) (reserved_non_modifiers 937) (semi_reserved 938) (simple_variable 11505) ("{" 11506)) 4949 (("(" 943) (argument_list 11509)) 4950 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11507) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4951 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11508)) 4952 ((T_FN 35) (T_FUNCTION 129) (fn 741) #'742 (inline_function 11513)) 4953 (("$" 678) (T_STRING 11519) (T_VARIABLE 727) (property_name 11526) (simple_variable 11521) ("{" 11522)) 4954 (("$" 678) (T_STRING 11519) (T_VARIABLE 727) (property_name 11520) (simple_variable 11521) ("{" 11522)) 4955 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 11517) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 4956 (("]" 11518)) 4957 (("(" 943) (argument_list 11525)) 4958 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11523) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 4959 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11524)) 4960 (("(" 943) (argument_list 11527)) 4961 (("`" 11529)) 4962 (("]" 11531)) 4963 (("=" 11532)) 4964 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 11533) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 4965 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 12055) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4966 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 12053) (encaps_var 5237)) 4967 (("$" 11537) (T_VARIABLE 11585) (simple_variable 12049) ("{" 12050)) 4968 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 12047) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 4969 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 12046) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4970 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 12045) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4971 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 12044) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4972 (("(" 12041)) 4973 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 12040) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4974 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 12039) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4975 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 12038) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4976 (("\"" 1520) ("$" 11840) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 11841) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 11842) ("[" 1542) (array_object_dereferenceable 11843) (callable_expr 11844) (callable_variable 11845) (class_constant 1546) (class_name 11846) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 11847) (name 11848) (new_dereferenceable 1552) (simple_variable 11849) (static_member 11850) (variable 12037) (variable_class_name 11852)) 4977 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 12036) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4978 (("(" 12033)) 4979 (("(" 12030)) 4980 (("(" 12021) (argument_list 12022) (ctor_arguments 12023)) 4981 (("\"" 1520) ("$" 11840) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 11841) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 11842) ("[" 1542) (array_object_dereferenceable 11843) (callable_expr 11844) (callable_variable 11845) (class_constant 1546) (class_name 11846) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 11847) (name 11848) (new_dereferenceable 1552) (simple_variable 11849) (static_member 11850) (variable 12020) (variable_class_name 11852)) 4982 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 12019) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4983 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 12018) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4984 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 12017) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4985 (("(" 12013)) 4986 (("(" 12008)) 4987 (("(" 12002)) 4988 (("$" 11840) ("(" 11965) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 11966) (T_NAME_QUALIFIED 11967) (T_NAME_RELATIVE 11968) (T_READONLY 2931) (T_STATIC 11969) (T_STRING 11970) (T_VARIABLE 11842) (anonymous_class 11971) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 11972) (attribute 2937) (attributes 11973) (class_modifier 2939) (class_name 11974) (class_name_reference 11975) (name 11976) (new_variable 11977) (simple_variable 11978)) 4989 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11964) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4990 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11963) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4991 (("(" 11801) (argument_list 11962)) 4992 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11961) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4993 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11960) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4994 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 11955) (T_END_HEREDOC 11956) (T_VARIABLE 2896) (encaps_list 11957) (encaps_var 2898)) 4995 ((T_FN 35) (T_FUNCTION 129) (fn 11599) #'11600 (inline_function 11954)) 4996 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11953) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4997 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11952) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4998 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11951) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 4999 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11948) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5000 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11947) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5001 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 11943) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 5002 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 11941) (encaps_list 2804) (encaps_var 2805)) 5003 ((T_NULLSAFE_OBJECT_OPERATOR 11927) (T_OBJECT_OPERATOR 11928) ("[" 11929)) 5004 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 11924) (attribute 2733) (fn 11599) #'11600 (inline_function 11925)) 5005 (("(" 11801) (argument_list 11923)) 5006 ((T_PAAMAYIM_NEKUDOTAYIM 11915)) 5007 (("%" 11615) ("*" 11616) ("+" 11617) ("-" 11618) ("." 11619) ("/" 11620) ("<" 11621) (">" 11622) ("?" 11623) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11624) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11625) (T_BOOLEAN_AND 11626) (T_BOOLEAN_OR 11627) (T_COALESCE 11628) (T_DOUBLE_ARROW 11913) (T_INSTANCEOF 11629) (T_IS_EQUAL 11630) (T_IS_GREATER_OR_EQUAL 11631) (T_IS_IDENTICAL 11632) (T_IS_NOT_EQUAL 11633) (T_IS_NOT_IDENTICAL 11634) (T_IS_SMALLER_OR_EQUAL 11635) (T_LOGICAL_AND 11636) (T_LOGICAL_OR 11637) (T_LOGICAL_XOR 11638) (T_POW 11639) (T_SL 11640) (T_SPACESHIP 11641) (T_SR 11642) ("^" 11643) ("|" 11644)) 5008 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 11902)) 5009 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 11890)) 5010 (("(" 11801) (argument_list 11889)) 5011 (("=" 11809) (T_AND_EQUAL 11810) (T_COALESCE_EQUAL 11811) (T_CONCAT_EQUAL 11812) (T_DEC 11813) (T_DIV_EQUAL 11814) (T_INC 11815) (T_MINUS_EQUAL 11816) (T_MOD_EQUAL 11817) (T_MUL_EQUAL 11818) (T_OR_EQUAL 11819) (T_PLUS_EQUAL 11820) (T_POW_EQUAL 11821) (T_SL_EQUAL 11822) (T_SR_EQUAL 11823) (T_XOR_EQUAL 11824)) 5012 ((T_PAAMAYIM_NEKUDOTAYIM 11713)) 5013 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11614) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5014 (("%" 11615) ("*" 11616) ("+" 11617) ("-" 11618) ("." 11619) ("/" 11620) ("<" 11621) (">" 11622) ("?" 11623) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11624) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11625) (T_BOOLEAN_AND 11626) (T_BOOLEAN_OR 11627) (T_COALESCE 11628) (T_INSTANCEOF 11629) (T_IS_EQUAL 11630) (T_IS_GREATER_OR_EQUAL 11631) (T_IS_IDENTICAL 11632) (T_IS_NOT_EQUAL 11633) (T_IS_NOT_IDENTICAL 11634) (T_IS_SMALLER_OR_EQUAL 11635) (T_LOGICAL_AND 11636) (T_LOGICAL_OR 11637) (T_LOGICAL_XOR 11638) (T_POW 11639) (T_SL 11640) (T_SPACESHIP 11641) (T_SR 11642) ("^" 11643) ("|" 11644)) 5015 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11712) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5016 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11711) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5017 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11710) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5018 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11709) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5019 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11708) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5020 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11707) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5021 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11706) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5022 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11705) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5023 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 11700) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 11701) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 5024 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11699) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5025 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11698) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5026 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11697) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5027 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11696) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5028 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11695) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5029 (("$" 11660) ("(" 11661) (T_NAME_FULLY_QUALIFIED 11662) (T_NAME_QUALIFIED 11663) (T_NAME_RELATIVE 11664) (T_STATIC 11665) (T_STRING 11666) (T_VARIABLE 11667) (class_name 11668) (class_name_reference 11669) (name 11670) (new_variable 11671) (simple_variable 11672)) 5030 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11659) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5031 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11658) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5032 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11657) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5033 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11656) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5034 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11655) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5035 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11654) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5036 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11653) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5037 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11652) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5038 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11651) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5039 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11650) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5040 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11649) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5041 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11648) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5042 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11647) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5043 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11646) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5044 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11645) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5045 (("$" 11660) (T_VARIABLE 11667) (simple_variable 11691) ("{" 11692)) 5046 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 11689) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5047 ((T_PAAMAYIM_NEKUDOTAYIM 11687)) 5048 ((T_NULLSAFE_OBJECT_OPERATOR 11673) (T_OBJECT_OPERATOR 11674) (T_PAAMAYIM_NEKUDOTAYIM 11675) ("[" 11676)) 5049 (("$" 11660) (T_STRING 11680) (T_VARIABLE 11667) (property_name 11686) (simple_variable 11682) ("{" 11683)) 5050 (("$" 11660) (T_STRING 11680) (T_VARIABLE 11667) (property_name 11681) (simple_variable 11682) ("{" 11683)) 5051 (("$" 11660) (T_VARIABLE 11667) (simple_variable 11679)) 5052 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 11677) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5053 (("]" 11678)) 5054 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11684) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5055 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11685)) 5056 (("$" 11660) (T_VARIABLE 11667) (simple_variable 11688)) 5057 (("%" 299) (")" 11690) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5058 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11693) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5059 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11694)) 5060 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11704) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5061 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 11702) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 5062 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11703) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5063 (("$" 11537) (T_ABSTRACT 11714) (T_ARRAY 11715) (T_AS 11716) (T_BREAK 11717) (T_CALLABLE 11718) (T_CASE 11719) (T_CATCH 11720) (T_CLASS 11721) (T_CLASS_C 11722) (T_CLONE 11723) (T_CONST 11724) (T_CONTINUE 11725) (T_DECLARE 11726) (T_DEFAULT 11727) (T_DIR 11728) (T_DO 11729) (T_ECHO 11730) (T_ELSE 11731) (T_ELSEIF 11732) (T_EMPTY 11733) (T_ENDDECLARE 11734) (T_ENDFOR 11735) (T_ENDFOREACH 11736) (T_ENDIF 11737) (T_ENDSWITCH 11738) (T_ENDWHILE 11739) (T_ENUM 11740) (T_EVAL 11741) (T_EXIT 11742) (T_EXTENDS 11743) (T_FILE 11744) (T_FINAL 11745) (T_FINALLY 11746) (T_FN 11747) (T_FOR 11748) (T_FOREACH 11749) (T_FUNCTION 11750) (T_FUNC_C 11751) (T_GLOBAL 11752) (T_GOTO 11753) (T_IF 11754) (T_IMPLEMENTS 11755) (T_INCLUDE 11756) (T_INCLUDE_ONCE 11757) (T_INSTANCEOF 11758) (T_INSTEADOF 11759) (T_INTERFACE 11760) (T_ISSET 11761) (T_LINE 11762) (T_LIST 11763) (T_LOGICAL_AND 11764) (T_LOGICAL_OR 11765) (T_LOGICAL_XOR 11766) (T_MATCH 11767) (T_METHOD_C 11768) (T_NAMESPACE 11769) (T_NEW 11770) (T_NS_C 11771) (T_PRINT 11772) (T_PRIVATE 11773) (T_PROPERTY_C 11774) (T_PROTECTED 11775) (T_PUBLIC 11776) (T_READONLY 11777) (T_REQUIRE 11778) (T_REQUIRE_ONCE 11779) (T_RETURN 11780) (T_STATIC 11781) (T_STRING 11782) (T_SWITCH 11783) (T_THROW 11784) (T_TRAIT 11785) (T_TRAIT_C 11786) (T_TRY 11787) (T_UNSET 11788) (T_USE 11789) (T_VAR 11790) (T_VARIABLE 11585) (T_WHILE 11791) (T_YIELD 11792) (identifier 11793) (member_name 11794) (reserved_non_modifiers 11795) (semi_reserved 11796) (simple_variable 11797) ("{" 11798)) 5064 (("(" 11801) (argument_list 11802)) 5065 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11799) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5066 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11800)) 5067 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 11803) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 11804) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 11805) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5068 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 11808) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5069 (("," 1480) (possible_comma 11806)) 5070 ((")" 11807)) 5071 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (ampersand 11838) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11839) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5072 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11837) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5073 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11836) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5074 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11835) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5075 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11834) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5076 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11833) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5077 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11832) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5078 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11831) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5079 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11830) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5080 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11829) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5081 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11828) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5082 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11827) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5083 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11826) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5084 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11825) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5085 (("\"" 1520) ("$" 11840) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 11841) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 11842) ("[" 1542) (array_object_dereferenceable 11843) (callable_expr 11844) (callable_variable 11845) (class_constant 1546) (class_name 11846) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 11847) (name 11848) (new_dereferenceable 1552) (simple_variable 11849) (static_member 11850) (variable 11851) (variable_class_name 11852)) 5086 (("$" 11840) (T_VARIABLE 11842) (simple_variable 11885) ("{" 11886)) 5087 (("(" 11856) (argument_list 11884)) 5088 ((T_NULLSAFE_OBJECT_OPERATOR 11870) (T_OBJECT_OPERATOR 11871) ("[" 11872)) 5089 (("(" 11856) (argument_list 11869)) 5090 ((T_PAAMAYIM_NEKUDOTAYIM 11865)) 5091 (("(" 11856) (argument_list 11864)) 5092 ((T_PAAMAYIM_NEKUDOTAYIM 11853)) 5093 (("$" 11840) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 11842) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 11854) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 11855) ("{" 1642)) 5094 (("(" 11856) (argument_list 11857)) 5095 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 11858) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 11859) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 11860) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5096 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 11863) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5097 (("," 1480) (possible_comma 11861)) 5098 ((")" 11862)) 5099 (("$" 11840) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 11842) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 11866) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 11867) ("{" 1658)) 5100 (("(" 11856) (argument_list 11868)) 5101 (("$" 11840) (T_STRING 11875) (T_VARIABLE 11842) (property_name 11882) (simple_variable 11877) ("{" 11878)) 5102 (("$" 11840) (T_STRING 11875) (T_VARIABLE 11842) (property_name 11876) (simple_variable 11877) ("{" 11878)) 5103 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 11873) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5104 (("]" 11874)) 5105 (("(" 11856) (argument_list 11881)) 5106 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11879) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5107 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11880)) 5108 (("(" 11856) (argument_list 11883)) 5109 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11887) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5110 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11888)) 5111 ((backup_doc_comment 11891)) 5112 (("(" 11892)) 5113 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 11893)) 5114 ((")" 11894)) 5115 ((T_USE 164) (lexical_vars 11895)) 5116 ((":" 166) (return_type 11896)) 5117 ((backup_fn_flags 11897)) 5118 (("{" 11898)) 5119 ((inner_statement_list 11899)) 5120 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 11900) ("~" 127)) 5121 ((backup_fn_flags 11901)) 5122 ((backup_doc_comment 11903)) 5123 (("(" 11904)) 5124 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 11905)) 5125 ((")" 11906)) 5126 ((":" 2249) (return_type 11907)) 5127 ((T_DOUBLE_ARROW 11908)) 5128 ((backup_fn_flags 11909)) 5129 ((backup_lex_pos 11910)) 5130 (("!" 10874) ("\"" 10875) ("$" 10876) ("(" 10877) ("+" 10878) ("-" 10879) ("@" 10880) (T_ARRAY 10881) (T_ARRAY_CAST 10882) (T_ATTRIBUTE 128) (T_BOOL_CAST 10883) (T_CLASS_C 10884) (T_CLONE 10885) (T_CONSTANT_ENCAPSED_STRING 10886) (T_DEC 10887) (T_DIR 10888) (T_DNUMBER 10889) (T_DOUBLE_CAST 10890) (T_EMPTY 10891) (T_EVAL 10892) (T_EXIT 10893) (T_FILE 10894) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 10895) (T_INC 10896) (T_INCLUDE 10897) (T_INCLUDE_ONCE 10898) (T_INT_CAST 10899) (T_ISSET 10900) (T_LINE 10901) (T_LIST 10902) (T_LNUMBER 10903) (T_MATCH 10904) (T_METHOD_C 10905) (T_NAME_FULLY_QUALIFIED 10906) (T_NAME_QUALIFIED 10907) (T_NAME_RELATIVE 10908) (T_NEW 10909) (T_NS_C 10910) (T_OBJECT_CAST 10911) (T_PRINT 10912) (T_PROPERTY_C 10913) (T_READONLY 10914) (T_REQUIRE 10915) (T_REQUIRE_ONCE 10916) (T_START_HEREDOC 10917) (T_STATIC 10918) (T_STRING 10919) (T_STRING_CAST 10920) (T_THROW 10921) (T_TRAIT_C 10922) (T_UNSET_CAST 10923) (T_VARIABLE 10924) (T_YIELD 10925) (T_YIELD_FROM 10926) ("[" 10927) ("`" 10928) (array_object_dereferenceable 10929) (attribute 133) (attributes 10930) (callable_expr 10931) (callable_variable 10932) (class_constant 10933) (class_name 10934) (constant 10935) (dereferenceable_scalar 10936) (expr 11911) (fn 10938) (fully_dereferenceable 105) #'10939 (function_call 10940) (inline_function 10941) (internal_functions_in_yacc 10942) (match 10943) (name 10944) (new_dereferenceable 10945) (new_non_dereferenceable 10946) (scalar 10947) (simple_variable 10948) (static_member 10949) (variable 10950) (variable_class_name 10951) ("~" 10952)) 5131 (("%" 10954) ("*" 10955) ("+" 10956) ("-" 10957) ("." 10958) ("/" 10959) ("<" 10960) (">" 10961) ("?" 10962) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 10963) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 10964) (T_BOOLEAN_AND 10965) (T_BOOLEAN_OR 10966) (T_COALESCE 10967) (T_INSTANCEOF 10968) (T_IS_EQUAL 10969) (T_IS_GREATER_OR_EQUAL 10970) (T_IS_IDENTICAL 10971) (T_IS_NOT_EQUAL 10972) (T_IS_NOT_IDENTICAL 10973) (T_IS_SMALLER_OR_EQUAL 10974) (T_LOGICAL_AND 10975) (T_LOGICAL_OR 10976) (T_LOGICAL_XOR 10977) (T_POW 10978) (T_SL 10979) (T_SPACESHIP 10980) (T_SR 10981) ("^" 10982) (backup_fn_flags 11912) ("|" 10983)) 5132 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 11914) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 5133 (("$" 11537) (T_ABSTRACT 11714) (T_ARRAY 11715) (T_AS 11716) (T_BREAK 11717) (T_CALLABLE 11718) (T_CASE 11719) (T_CATCH 11720) (T_CLASS 11721) (T_CLASS_C 11722) (T_CLONE 11723) (T_CONST 11724) (T_CONTINUE 11725) (T_DECLARE 11726) (T_DEFAULT 11727) (T_DIR 11728) (T_DO 11729) (T_ECHO 11730) (T_ELSE 11731) (T_ELSEIF 11732) (T_EMPTY 11733) (T_ENDDECLARE 11734) (T_ENDFOR 11735) (T_ENDFOREACH 11736) (T_ENDIF 11737) (T_ENDSWITCH 11738) (T_ENDWHILE 11739) (T_ENUM 11740) (T_EVAL 11741) (T_EXIT 11742) (T_EXTENDS 11743) (T_FILE 11744) (T_FINAL 11745) (T_FINALLY 11746) (T_FN 11747) (T_FOR 11748) (T_FOREACH 11749) (T_FUNCTION 11750) (T_FUNC_C 11751) (T_GLOBAL 11752) (T_GOTO 11753) (T_IF 11754) (T_IMPLEMENTS 11755) (T_INCLUDE 11756) (T_INCLUDE_ONCE 11757) (T_INSTANCEOF 11758) (T_INSTEADOF 11759) (T_INTERFACE 11760) (T_ISSET 11761) (T_LINE 11762) (T_LIST 11763) (T_LOGICAL_AND 11764) (T_LOGICAL_OR 11765) (T_LOGICAL_XOR 11766) (T_MATCH 11767) (T_METHOD_C 11768) (T_NAMESPACE 11769) (T_NEW 11770) (T_NS_C 11771) (T_PRINT 11772) (T_PRIVATE 11773) (T_PROPERTY_C 11774) (T_PROTECTED 11775) (T_PUBLIC 11776) (T_READONLY 11777) (T_REQUIRE 11778) (T_REQUIRE_ONCE 11779) (T_RETURN 11780) (T_STATIC 11781) (T_STRING 11782) (T_SWITCH 11783) (T_THROW 11784) (T_TRAIT 11785) (T_TRAIT_C 11786) (T_TRY 11787) (T_UNSET 11788) (T_USE 11789) (T_VAR 11790) (T_VARIABLE 11585) (T_WHILE 11791) (T_YIELD 11792) (identifier 11916) (member_name 11917) (reserved_non_modifiers 11795) (semi_reserved 11796) (simple_variable 11918) ("{" 11919)) 5134 (("(" 11801) (argument_list 11922)) 5135 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11920) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5136 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11921)) 5137 ((T_FN 35) (T_FUNCTION 129) (fn 11599) #'11600 (inline_function 11926)) 5138 (("$" 11537) (T_STRING 11932) (T_VARIABLE 11585) (property_name 11939) (simple_variable 11934) ("{" 11935)) 5139 (("$" 11537) (T_STRING 11932) (T_VARIABLE 11585) (property_name 11933) (simple_variable 11934) ("{" 11935)) 5140 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 11930) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5141 (("]" 11931)) 5142 (("(" 11801) (argument_list 11938)) 5143 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 11936) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5144 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 11937)) 5145 (("(" 11801) (argument_list 11940)) 5146 (("`" 11942)) 5147 (("]" 11944)) 5148 (("=" 11945)) 5149 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11946) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5150 (("%" 11615) ("*" 11616) ("+" 11617) ("-" 11618) ("." 11619) ("/" 11620) ("<" 11621) (">" 11622) ("?" 11623) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11624) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11625) (T_BOOLEAN_AND 11626) (T_BOOLEAN_OR 11627) (T_COALESCE 11628) (T_DOUBLE_ARROW 11949) (T_INSTANCEOF 11629) (T_IS_EQUAL 11630) (T_IS_GREATER_OR_EQUAL 11631) (T_IS_IDENTICAL 11632) (T_IS_NOT_EQUAL 11633) (T_IS_NOT_IDENTICAL 11634) (T_IS_SMALLER_OR_EQUAL 11635) (T_LOGICAL_AND 11636) (T_LOGICAL_OR 11637) (T_LOGICAL_XOR 11638) (T_POW 11639) (T_SL 11640) (T_SPACESHIP 11641) (T_SR 11642) ("^" 11643) ("|" 11644)) 5151 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 11950) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5152 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 11959) (T_VARIABLE 2896) (encaps_var 2910)) 5153 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 11958) (T_VARIABLE 2896) (encaps_var 2901)) 5154 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 12000) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5155 ((T_CLASS 11992)) 5156 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 11991) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 11972) (attribute 2958) (class_modifier 2939)) 5157 ((T_PAAMAYIM_NEKUDOTAYIM 11989)) 5158 (("(" 11856) (argument_list 11988)) 5159 ((T_NULLSAFE_OBJECT_OPERATOR 11979) (T_OBJECT_OPERATOR 11980) (T_PAAMAYIM_NEKUDOTAYIM 11981) ("[" 11982)) 5160 (("$" 11840) (T_STRING 11875) (T_VARIABLE 11842) (property_name 11987) (simple_variable 11877) ("{" 11878)) 5161 (("$" 11840) (T_STRING 11875) (T_VARIABLE 11842) (property_name 11986) (simple_variable 11877) ("{" 11878)) 5162 (("$" 11840) (T_VARIABLE 11842) (simple_variable 11985)) 5163 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 11983) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5164 (("]" 11984)) 5165 (("$" 11840) (T_VARIABLE 11842) (simple_variable 11990)) 5166 (("(" 2960) (argument_list 2961) (ctor_arguments 11993)) 5167 ((T_EXTENDS 2963) (extends_from 11994)) 5168 ((T_IMPLEMENTS 2965) (implements_list 11995)) 5169 ((backup_doc_comment 11996)) 5170 (("{" 11997)) 5171 ((class_statement_list 11998)) 5172 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 11999)) 5173 (("%" 299) (")" 12001) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5174 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 12003) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5175 (("%" 299) (")" 12004) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5176 (("{" 12005)) 5177 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 12006) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 5178 (("}" 12007)) 5179 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 12009) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 5180 ((")" 12010)) 5181 (("=" 12011)) 5182 (("!" 11535) ("\"" 11536) ("$" 11537) ("(" 11538) ("+" 11539) ("-" 11540) ("@" 11541) (T_ARRAY 11542) (T_ARRAY_CAST 11543) (T_ATTRIBUTE 128) (T_BOOL_CAST 11544) (T_CLASS_C 11545) (T_CLONE 11546) (T_CONSTANT_ENCAPSED_STRING 11547) (T_DEC 11548) (T_DIR 11549) (T_DNUMBER 11550) (T_DOUBLE_CAST 11551) (T_EMPTY 11552) (T_EVAL 11553) (T_EXIT 11554) (T_FILE 11555) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 11556) (T_INC 11557) (T_INCLUDE 11558) (T_INCLUDE_ONCE 11559) (T_INT_CAST 11560) (T_ISSET 11561) (T_LINE 11562) (T_LIST 11563) (T_LNUMBER 11564) (T_MATCH 11565) (T_METHOD_C 11566) (T_NAME_FULLY_QUALIFIED 11567) (T_NAME_QUALIFIED 11568) (T_NAME_RELATIVE 11569) (T_NEW 11570) (T_NS_C 11571) (T_OBJECT_CAST 11572) (T_PRINT 11573) (T_PROPERTY_C 11574) (T_READONLY 11575) (T_REQUIRE 11576) (T_REQUIRE_ONCE 11577) (T_START_HEREDOC 11578) (T_STATIC 11579) (T_STRING 11580) (T_STRING_CAST 11581) (T_THROW 11582) (T_TRAIT_C 11583) (T_UNSET_CAST 11584) (T_VARIABLE 11585) (T_YIELD 11586) (T_YIELD_FROM 11587) ("[" 11588) ("`" 11589) (array_object_dereferenceable 11590) (attribute 133) (attributes 11591) (callable_expr 11592) (callable_variable 11593) (class_constant 11594) (class_name 11595) (constant 11596) (dereferenceable_scalar 11597) (expr 12012) (fn 11599) (fully_dereferenceable 105) #'11600 (function_call 11601) (inline_function 11602) (internal_functions_in_yacc 11603) (match 11604) (name 11605) (new_dereferenceable 11606) (new_non_dereferenceable 11607) (scalar 11608) (simple_variable 11609) (static_member 11610) (variable 11611) (variable_class_name 11612) ("~" 11613)) 5183 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 12014) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5184 (("," 5192) (possible_comma 12015)) 5185 ((")" 12016)) 5186 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 12024) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 12025) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 12026) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5187 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 12029) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5188 (("," 1480) (possible_comma 12027)) 5189 ((")" 12028)) 5190 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 12031) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5191 (("%" 299) (")" 12032) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5192 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 12034) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5193 (("%" 299) (")" 12035) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5194 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 12042) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 5195 ((")" 12043)) 5196 (("%" 299) (")" 12048) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5197 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12051) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5198 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 12052)) 5199 (("\"" 12054) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 5200 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 12064) (T_VARIABLE 2896) (encaps_var 2910)) 5201 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 12063) (T_VARIABLE 2896) (encaps_var 2901)) 5202 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 12105) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5203 ((T_CLASS 12097)) 5204 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 12096) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 12077) (attribute 2958) (class_modifier 2939)) 5205 ((T_PAAMAYIM_NEKUDOTAYIM 12094)) 5206 (("(" 10407) (argument_list 12093)) 5207 ((T_NULLSAFE_OBJECT_OPERATOR 12084) (T_OBJECT_OPERATOR 12085) (T_PAAMAYIM_NEKUDOTAYIM 12086) ("[" 12087)) 5208 (("$" 10391) (T_STRING 10426) (T_VARIABLE 10393) (property_name 12092) (simple_variable 10428) ("{" 10429)) 5209 (("$" 10391) (T_STRING 10426) (T_VARIABLE 10393) (property_name 12091) (simple_variable 10428) ("{" 10429)) 5210 (("$" 10391) (T_VARIABLE 10393) (simple_variable 12090)) 5211 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 12088) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5212 (("]" 12089)) 5213 (("$" 10391) (T_VARIABLE 10393) (simple_variable 12095)) 5214 (("(" 2960) (argument_list 2961) (ctor_arguments 12098)) 5215 ((T_EXTENDS 2963) (extends_from 12099)) 5216 ((T_IMPLEMENTS 2965) (implements_list 12100)) 5217 ((backup_doc_comment 12101)) 5218 (("{" 12102)) 5219 ((class_statement_list 12103)) 5220 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 12104)) 5221 (("%" 299) (")" 12106) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5222 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 12108) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5223 (("%" 299) (")" 12109) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5224 (("{" 12110)) 5225 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 12111) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 5226 (("}" 12112)) 5227 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 12114) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 5228 ((")" 12115)) 5229 (("=" 12116)) 5230 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12117) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 5231 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 12119) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5232 (("," 5192) (possible_comma 12120)) 5233 ((")" 12121)) 5234 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 12129) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 12130) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 12131) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5235 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 12134) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5236 (("," 1480) (possible_comma 12132)) 5237 ((")" 12133)) 5238 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 12136) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5239 (("%" 299) (")" 12137) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5240 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 12139) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5241 (("%" 299) (")" 12140) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5242 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 12147) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 5243 ((")" 12148)) 5244 (("%" 299) (")" 12154) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5245 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12157) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5246 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 12158)) 5247 (("\"" 12160) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 5248 (("$" 513) (T_ABSTRACT 12171) (T_ARRAY 12172) (T_AS 12173) (T_BREAK 12174) (T_CALLABLE 12175) (T_CASE 12176) (T_CATCH 12177) (T_CLASS 12178) (T_CLASS_C 12179) (T_CLONE 12180) (T_CONST 12181) (T_CONTINUE 12182) (T_DECLARE 12183) (T_DEFAULT 12184) (T_DIR 12185) (T_DO 12186) (T_ECHO 12187) (T_ELSE 12188) (T_ELSEIF 12189) (T_EMPTY 12190) (T_ENDDECLARE 12191) (T_ENDFOR 12192) (T_ENDFOREACH 12193) (T_ENDIF 12194) (T_ENDSWITCH 12195) (T_ENDWHILE 12196) (T_ENUM 12197) (T_EVAL 12198) (T_EXIT 12199) (T_EXTENDS 12200) (T_FILE 12201) (T_FINAL 12202) (T_FINALLY 12203) (T_FN 12204) (T_FOR 12205) (T_FOREACH 12206) (T_FUNCTION 12207) (T_FUNC_C 12208) (T_GLOBAL 12209) (T_GOTO 12210) (T_IF 12211) (T_IMPLEMENTS 12212) (T_INCLUDE 12213) (T_INCLUDE_ONCE 12214) (T_INSTANCEOF 12215) (T_INSTEADOF 12216) (T_INTERFACE 12217) (T_ISSET 12218) (T_LINE 12219) (T_LIST 12220) (T_LOGICAL_AND 12221) (T_LOGICAL_OR 12222) (T_LOGICAL_XOR 12223) (T_MATCH 12224) (T_METHOD_C 12225) (T_NAMESPACE 12226) (T_NEW 12227) (T_NS_C 12228) (T_PRINT 12229) (T_PRIVATE 12230) (T_PROPERTY_C 12231) (T_PROTECTED 12232) (T_PUBLIC 12233) (T_READONLY 12234) (T_REQUIRE 12235) (T_REQUIRE_ONCE 12236) (T_RETURN 12237) (T_STATIC 12238) (T_STRING 12239) (T_SWITCH 12240) (T_THROW 12241) (T_TRAIT 12242) (T_TRAIT_C 12243) (T_TRY 12244) (T_UNSET 12245) (T_USE 12246) (T_VAR 12247) (T_VARIABLE 561) (T_WHILE 12248) (T_YIELD 12249) (identifier 12250) (member_name 12251) (reserved_non_modifiers 12252) (semi_reserved 12253) (simple_variable 12254) ("{" 12255)) 5249 (("(" 12258) (argument_list 12259)) 5250 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12256) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5251 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 12257)) 5252 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 12260) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 12261) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 12262) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5253 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 12265) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5254 (("," 1480) (possible_comma 12263)) 5255 ((")" 12264)) 5256 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (ampersand 12295) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12296) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5257 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12294) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5258 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12293) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5259 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12292) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5260 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12291) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5261 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12290) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5262 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12289) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5263 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12288) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5264 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12287) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5265 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12286) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5266 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12285) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5267 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12284) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5268 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12283) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5269 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12282) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5270 (("\"" 1520) ("$" 12297) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 12298) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 12299) ("[" 1542) (array_object_dereferenceable 12300) (callable_expr 12301) (callable_variable 12302) (class_constant 1546) (class_name 12303) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 12304) (name 12305) (new_dereferenceable 1552) (simple_variable 12306) (static_member 12307) (variable 12308) (variable_class_name 12309)) 5271 (("$" 12297) (T_VARIABLE 12299) (simple_variable 12342) ("{" 12343)) 5272 (("(" 12313) (argument_list 12341)) 5273 ((T_NULLSAFE_OBJECT_OPERATOR 12327) (T_OBJECT_OPERATOR 12328) ("[" 12329)) 5274 (("(" 12313) (argument_list 12326)) 5275 ((T_PAAMAYIM_NEKUDOTAYIM 12322)) 5276 (("(" 12313) (argument_list 12321)) 5277 ((T_PAAMAYIM_NEKUDOTAYIM 12310)) 5278 (("$" 12297) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 12299) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 12311) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 12312) ("{" 1642)) 5279 (("(" 12313) (argument_list 12314)) 5280 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 12315) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 12316) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 12317) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5281 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 12320) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5282 (("," 1480) (possible_comma 12318)) 5283 ((")" 12319)) 5284 (("$" 12297) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 12299) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 12323) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 12324) ("{" 1658)) 5285 (("(" 12313) (argument_list 12325)) 5286 (("$" 12297) (T_STRING 12332) (T_VARIABLE 12299) (property_name 12339) (simple_variable 12334) ("{" 12335)) 5287 (("$" 12297) (T_STRING 12332) (T_VARIABLE 12299) (property_name 12333) (simple_variable 12334) ("{" 12335)) 5288 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 12330) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5289 (("]" 12331)) 5290 (("(" 12313) (argument_list 12338)) 5291 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12336) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5292 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 12337)) 5293 (("(" 12313) (argument_list 12340)) 5294 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12344) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5295 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 12345)) 5296 ((backup_doc_comment 12348)) 5297 (("(" 12349)) 5298 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 12350)) 5299 ((")" 12351)) 5300 ((T_USE 164) (lexical_vars 12352)) 5301 ((":" 166) (return_type 12353)) 5302 ((backup_fn_flags 12354)) 5303 (("{" 12355)) 5304 ((inner_statement_list 12356)) 5305 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 12357) ("~" 127)) 5306 ((backup_fn_flags 12358)) 5307 ((backup_doc_comment 12360)) 5308 (("(" 12361)) 5309 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 12362)) 5310 ((")" 12363)) 5311 ((":" 2249) (return_type 12364)) 5312 ((T_DOUBLE_ARROW 12365)) 5313 ((backup_fn_flags 12366)) 5314 ((backup_lex_pos 12367)) 5315 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12431) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5316 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13405) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5317 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 13403) (encaps_var 5237)) 5318 (("$" 12370) (T_VARIABLE 12418) (simple_variable 13399) ("{" 13400)) 5319 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13397) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5320 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13396) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5321 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13395) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5322 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13394) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5323 (("(" 13391)) 5324 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13390) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5325 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13389) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5326 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13388) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5327 (("\"" 1520) ("$" 12673) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 12674) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 12675) ("[" 1542) (array_object_dereferenceable 12676) (callable_expr 12677) (callable_variable 12678) (class_constant 1546) (class_name 12679) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 12680) (name 12681) (new_dereferenceable 1552) (simple_variable 12682) (static_member 12683) (variable 13387) (variable_class_name 12685)) 5328 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13386) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5329 (("(" 13383)) 5330 (("(" 13380)) 5331 (("(" 13371) (argument_list 13372) (ctor_arguments 13373)) 5332 (("\"" 1520) ("$" 12673) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 12674) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 12675) ("[" 1542) (array_object_dereferenceable 12676) (callable_expr 12677) (callable_variable 12678) (class_constant 1546) (class_name 12679) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 12680) (name 12681) (new_dereferenceable 1552) (simple_variable 12682) (static_member 12683) (variable 13370) (variable_class_name 12685)) 5333 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13369) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5334 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13368) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5335 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13367) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5336 (("(" 13363)) 5337 (("(" 13358)) 5338 (("(" 13352)) 5339 (("$" 12673) ("(" 13315) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 13316) (T_NAME_QUALIFIED 13317) (T_NAME_RELATIVE 13318) (T_READONLY 2931) (T_STATIC 13319) (T_STRING 13320) (T_VARIABLE 12675) (anonymous_class 13321) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 13322) (attribute 2937) (attributes 13323) (class_modifier 2939) (class_name 13324) (class_name_reference 13325) (name 13326) (new_variable 13327) (simple_variable 13328)) 5340 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13314) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5341 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13313) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5342 (("(" 12634) (argument_list 13312)) 5343 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13311) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5344 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13310) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5345 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 13305) (T_END_HEREDOC 13306) (T_VARIABLE 2896) (encaps_list 13307) (encaps_var 2898)) 5346 ((T_FN 35) (T_FUNCTION 129) (fn 12432) #'12433 (inline_function 13304)) 5347 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13303) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5348 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13302) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5349 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13301) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5350 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12843) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5351 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12779) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5352 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 12775) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 5353 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 12773) (encaps_list 2804) (encaps_var 2805)) 5354 ((T_NULLSAFE_OBJECT_OPERATOR 12759) (T_OBJECT_OPERATOR 12760) ("[" 12761)) 5355 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 12756) (attribute 2733) (fn 12432) #'12433 (inline_function 12757)) 5356 (("(" 12634) (argument_list 12755)) 5357 ((T_PAAMAYIM_NEKUDOTAYIM 12747)) 5358 (("%" 12448) ("*" 12449) ("+" 12450) ("-" 12451) ("." 12452) ("/" 12453) ("<" 12454) (">" 12455) ("?" 12456) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12457) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12458) (T_BOOLEAN_AND 12459) (T_BOOLEAN_OR 12460) (T_COALESCE 12461) (T_INSTANCEOF 12462) (T_IS_EQUAL 12463) (T_IS_GREATER_OR_EQUAL 12464) (T_IS_IDENTICAL 12465) (T_IS_NOT_EQUAL 12466) (T_IS_NOT_IDENTICAL 12467) (T_IS_SMALLER_OR_EQUAL 12468) (T_LOGICAL_AND 12469) (T_LOGICAL_OR 12470) (T_LOGICAL_XOR 12471) (T_POW 12472) (T_SL 12473) (T_SPACESHIP 12474) (T_SR 12475) ("^" 12476) (backup_fn_flags 12746) ("|" 12477)) 5359 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 12735)) 5360 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 12723)) 5361 (("(" 12634) (argument_list 12722)) 5362 (("=" 12642) (T_AND_EQUAL 12643) (T_COALESCE_EQUAL 12644) (T_CONCAT_EQUAL 12645) (T_DEC 12646) (T_DIV_EQUAL 12647) (T_INC 12648) (T_MINUS_EQUAL 12649) (T_MOD_EQUAL 12650) (T_MUL_EQUAL 12651) (T_OR_EQUAL 12652) (T_PLUS_EQUAL 12653) (T_POW_EQUAL 12654) (T_SL_EQUAL 12655) (T_SR_EQUAL 12656) (T_XOR_EQUAL 12657)) 5363 ((T_PAAMAYIM_NEKUDOTAYIM 12546)) 5364 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12447) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5365 (("%" 12448) ("*" 12449) ("+" 12450) ("-" 12451) ("." 12452) ("/" 12453) ("<" 12454) (">" 12455) ("?" 12456) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12457) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12458) (T_BOOLEAN_AND 12459) (T_BOOLEAN_OR 12460) (T_COALESCE 12461) (T_INSTANCEOF 12462) (T_IS_EQUAL 12463) (T_IS_GREATER_OR_EQUAL 12464) (T_IS_IDENTICAL 12465) (T_IS_NOT_EQUAL 12466) (T_IS_NOT_IDENTICAL 12467) (T_IS_SMALLER_OR_EQUAL 12468) (T_LOGICAL_AND 12469) (T_LOGICAL_OR 12470) (T_LOGICAL_XOR 12471) (T_POW 12472) (T_SL 12473) (T_SPACESHIP 12474) (T_SR 12475) ("^" 12476) ("|" 12477)) 5366 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12545) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5367 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12544) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5368 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12543) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5369 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12542) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5370 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12541) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5371 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12540) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5372 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12539) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5373 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12538) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5374 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 12533) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12534) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 5375 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12532) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5376 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12531) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5377 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12530) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5378 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12529) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5379 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12528) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5380 (("$" 12493) ("(" 12494) (T_NAME_FULLY_QUALIFIED 12495) (T_NAME_QUALIFIED 12496) (T_NAME_RELATIVE 12497) (T_STATIC 12498) (T_STRING 12499) (T_VARIABLE 12500) (class_name 12501) (class_name_reference 12502) (name 12503) (new_variable 12504) (simple_variable 12505)) 5381 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12492) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5382 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12491) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5383 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12490) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5384 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12489) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5385 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12488) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5386 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12487) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5387 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12486) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5388 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12485) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5389 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12484) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5390 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12483) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5391 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12482) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5392 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12481) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5393 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12480) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5394 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12479) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5395 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12478) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5396 (("$" 12493) (T_VARIABLE 12500) (simple_variable 12524) ("{" 12525)) 5397 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 12522) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5398 ((T_PAAMAYIM_NEKUDOTAYIM 12520)) 5399 ((T_NULLSAFE_OBJECT_OPERATOR 12506) (T_OBJECT_OPERATOR 12507) (T_PAAMAYIM_NEKUDOTAYIM 12508) ("[" 12509)) 5400 (("$" 12493) (T_STRING 12513) (T_VARIABLE 12500) (property_name 12519) (simple_variable 12515) ("{" 12516)) 5401 (("$" 12493) (T_STRING 12513) (T_VARIABLE 12500) (property_name 12514) (simple_variable 12515) ("{" 12516)) 5402 (("$" 12493) (T_VARIABLE 12500) (simple_variable 12512)) 5403 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 12510) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5404 (("]" 12511)) 5405 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12517) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5406 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 12518)) 5407 (("$" 12493) (T_VARIABLE 12500) (simple_variable 12521)) 5408 (("%" 299) (")" 12523) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5409 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12526) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5410 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 12527)) 5411 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12537) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5412 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 12535) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 5413 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12536) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5414 (("$" 12370) (T_ABSTRACT 12547) (T_ARRAY 12548) (T_AS 12549) (T_BREAK 12550) (T_CALLABLE 12551) (T_CASE 12552) (T_CATCH 12553) (T_CLASS 12554) (T_CLASS_C 12555) (T_CLONE 12556) (T_CONST 12557) (T_CONTINUE 12558) (T_DECLARE 12559) (T_DEFAULT 12560) (T_DIR 12561) (T_DO 12562) (T_ECHO 12563) (T_ELSE 12564) (T_ELSEIF 12565) (T_EMPTY 12566) (T_ENDDECLARE 12567) (T_ENDFOR 12568) (T_ENDFOREACH 12569) (T_ENDIF 12570) (T_ENDSWITCH 12571) (T_ENDWHILE 12572) (T_ENUM 12573) (T_EVAL 12574) (T_EXIT 12575) (T_EXTENDS 12576) (T_FILE 12577) (T_FINAL 12578) (T_FINALLY 12579) (T_FN 12580) (T_FOR 12581) (T_FOREACH 12582) (T_FUNCTION 12583) (T_FUNC_C 12584) (T_GLOBAL 12585) (T_GOTO 12586) (T_IF 12587) (T_IMPLEMENTS 12588) (T_INCLUDE 12589) (T_INCLUDE_ONCE 12590) (T_INSTANCEOF 12591) (T_INSTEADOF 12592) (T_INTERFACE 12593) (T_ISSET 12594) (T_LINE 12595) (T_LIST 12596) (T_LOGICAL_AND 12597) (T_LOGICAL_OR 12598) (T_LOGICAL_XOR 12599) (T_MATCH 12600) (T_METHOD_C 12601) (T_NAMESPACE 12602) (T_NEW 12603) (T_NS_C 12604) (T_PRINT 12605) (T_PRIVATE 12606) (T_PROPERTY_C 12607) (T_PROTECTED 12608) (T_PUBLIC 12609) (T_READONLY 12610) (T_REQUIRE 12611) (T_REQUIRE_ONCE 12612) (T_RETURN 12613) (T_STATIC 12614) (T_STRING 12615) (T_SWITCH 12616) (T_THROW 12617) (T_TRAIT 12618) (T_TRAIT_C 12619) (T_TRY 12620) (T_UNSET 12621) (T_USE 12622) (T_VAR 12623) (T_VARIABLE 12418) (T_WHILE 12624) (T_YIELD 12625) (identifier 12626) (member_name 12627) (reserved_non_modifiers 12628) (semi_reserved 12629) (simple_variable 12630) ("{" 12631)) 5415 (("(" 12634) (argument_list 12635)) 5416 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12632) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5417 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 12633)) 5418 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 12636) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 12637) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 12638) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5419 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 12641) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5420 (("," 1480) (possible_comma 12639)) 5421 ((")" 12640)) 5422 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (ampersand 12671) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12672) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5423 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12670) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5424 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12669) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5425 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12668) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5426 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12667) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5427 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12666) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5428 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12665) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5429 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12664) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5430 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12663) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5431 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12662) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5432 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12661) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5433 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12660) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5434 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12659) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5435 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12658) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5436 (("\"" 1520) ("$" 12673) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 12674) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 12675) ("[" 1542) (array_object_dereferenceable 12676) (callable_expr 12677) (callable_variable 12678) (class_constant 1546) (class_name 12679) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 12680) (name 12681) (new_dereferenceable 1552) (simple_variable 12682) (static_member 12683) (variable 12684) (variable_class_name 12685)) 5437 (("$" 12673) (T_VARIABLE 12675) (simple_variable 12718) ("{" 12719)) 5438 (("(" 12689) (argument_list 12717)) 5439 ((T_NULLSAFE_OBJECT_OPERATOR 12703) (T_OBJECT_OPERATOR 12704) ("[" 12705)) 5440 (("(" 12689) (argument_list 12702)) 5441 ((T_PAAMAYIM_NEKUDOTAYIM 12698)) 5442 (("(" 12689) (argument_list 12697)) 5443 ((T_PAAMAYIM_NEKUDOTAYIM 12686)) 5444 (("$" 12673) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 12675) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 12687) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 12688) ("{" 1642)) 5445 (("(" 12689) (argument_list 12690)) 5446 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 12691) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 12692) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 12693) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5447 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 12696) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5448 (("," 1480) (possible_comma 12694)) 5449 ((")" 12695)) 5450 (("$" 12673) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 12675) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 12699) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 12700) ("{" 1658)) 5451 (("(" 12689) (argument_list 12701)) 5452 (("$" 12673) (T_STRING 12708) (T_VARIABLE 12675) (property_name 12715) (simple_variable 12710) ("{" 12711)) 5453 (("$" 12673) (T_STRING 12708) (T_VARIABLE 12675) (property_name 12709) (simple_variable 12710) ("{" 12711)) 5454 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 12706) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5455 (("]" 12707)) 5456 (("(" 12689) (argument_list 12714)) 5457 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12712) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5458 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 12713)) 5459 (("(" 12689) (argument_list 12716)) 5460 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12720) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5461 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 12721)) 5462 ((backup_doc_comment 12724)) 5463 (("(" 12725)) 5464 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 12726)) 5465 ((")" 12727)) 5466 ((T_USE 164) (lexical_vars 12728)) 5467 ((":" 166) (return_type 12729)) 5468 ((backup_fn_flags 12730)) 5469 (("{" 12731)) 5470 ((inner_statement_list 12732)) 5471 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 12733) ("~" 127)) 5472 ((backup_fn_flags 12734)) 5473 ((backup_doc_comment 12736)) 5474 (("(" 12737)) 5475 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 12738)) 5476 ((")" 12739)) 5477 ((":" 2249) (return_type 12740)) 5478 ((T_DOUBLE_ARROW 12741)) 5479 ((backup_fn_flags 12742)) 5480 ((backup_lex_pos 12743)) 5481 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12744) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5482 (("%" 12448) ("*" 12449) ("+" 12450) ("-" 12451) ("." 12452) ("/" 12453) ("<" 12454) (">" 12455) ("?" 12456) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12457) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12458) (T_BOOLEAN_AND 12459) (T_BOOLEAN_OR 12460) (T_COALESCE 12461) (T_INSTANCEOF 12462) (T_IS_EQUAL 12463) (T_IS_GREATER_OR_EQUAL 12464) (T_IS_IDENTICAL 12465) (T_IS_NOT_EQUAL 12466) (T_IS_NOT_IDENTICAL 12467) (T_IS_SMALLER_OR_EQUAL 12468) (T_LOGICAL_AND 12469) (T_LOGICAL_OR 12470) (T_LOGICAL_XOR 12471) (T_POW 12472) (T_SL 12473) (T_SPACESHIP 12474) (T_SR 12475) ("^" 12476) (backup_fn_flags 12745) ("|" 12477)) 5483 (("$" 12370) (T_ABSTRACT 12547) (T_ARRAY 12548) (T_AS 12549) (T_BREAK 12550) (T_CALLABLE 12551) (T_CASE 12552) (T_CATCH 12553) (T_CLASS 12554) (T_CLASS_C 12555) (T_CLONE 12556) (T_CONST 12557) (T_CONTINUE 12558) (T_DECLARE 12559) (T_DEFAULT 12560) (T_DIR 12561) (T_DO 12562) (T_ECHO 12563) (T_ELSE 12564) (T_ELSEIF 12565) (T_EMPTY 12566) (T_ENDDECLARE 12567) (T_ENDFOR 12568) (T_ENDFOREACH 12569) (T_ENDIF 12570) (T_ENDSWITCH 12571) (T_ENDWHILE 12572) (T_ENUM 12573) (T_EVAL 12574) (T_EXIT 12575) (T_EXTENDS 12576) (T_FILE 12577) (T_FINAL 12578) (T_FINALLY 12579) (T_FN 12580) (T_FOR 12581) (T_FOREACH 12582) (T_FUNCTION 12583) (T_FUNC_C 12584) (T_GLOBAL 12585) (T_GOTO 12586) (T_IF 12587) (T_IMPLEMENTS 12588) (T_INCLUDE 12589) (T_INCLUDE_ONCE 12590) (T_INSTANCEOF 12591) (T_INSTEADOF 12592) (T_INTERFACE 12593) (T_ISSET 12594) (T_LINE 12595) (T_LIST 12596) (T_LOGICAL_AND 12597) (T_LOGICAL_OR 12598) (T_LOGICAL_XOR 12599) (T_MATCH 12600) (T_METHOD_C 12601) (T_NAMESPACE 12602) (T_NEW 12603) (T_NS_C 12604) (T_PRINT 12605) (T_PRIVATE 12606) (T_PROPERTY_C 12607) (T_PROTECTED 12608) (T_PUBLIC 12609) (T_READONLY 12610) (T_REQUIRE 12611) (T_REQUIRE_ONCE 12612) (T_RETURN 12613) (T_STATIC 12614) (T_STRING 12615) (T_SWITCH 12616) (T_THROW 12617) (T_TRAIT 12618) (T_TRAIT_C 12619) (T_TRY 12620) (T_UNSET 12621) (T_USE 12622) (T_VAR 12623) (T_VARIABLE 12418) (T_WHILE 12624) (T_YIELD 12625) (identifier 12748) (member_name 12749) (reserved_non_modifiers 12628) (semi_reserved 12629) (simple_variable 12750) ("{" 12751)) 5484 (("(" 12634) (argument_list 12754)) 5485 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12752) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5486 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 12753)) 5487 ((T_FN 35) (T_FUNCTION 129) (fn 12432) #'12433 (inline_function 12758)) 5488 (("$" 12370) (T_STRING 12764) (T_VARIABLE 12418) (property_name 12771) (simple_variable 12766) ("{" 12767)) 5489 (("$" 12370) (T_STRING 12764) (T_VARIABLE 12418) (property_name 12765) (simple_variable 12766) ("{" 12767)) 5490 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 12762) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5491 (("]" 12763)) 5492 (("(" 12634) (argument_list 12770)) 5493 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12768) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5494 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 12769)) 5495 (("(" 12634) (argument_list 12772)) 5496 (("`" 12774)) 5497 (("]" 12776)) 5498 (("=" 12777)) 5499 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 12778) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5500 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13300) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5501 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 13298) (encaps_var 5237)) 5502 (("$" 12782) (T_VARIABLE 12830) (simple_variable 13294) ("{" 13295)) 5503 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13292) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5504 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13291) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5505 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13290) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5506 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13289) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5507 (("(" 13286)) 5508 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13285) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5509 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13284) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5510 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13283) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5511 (("\"" 1520) ("$" 13085) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 13086) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 13087) ("[" 1542) (array_object_dereferenceable 13088) (callable_expr 13089) (callable_variable 13090) (class_constant 1546) (class_name 13091) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 13092) (name 13093) (new_dereferenceable 1552) (simple_variable 13094) (static_member 13095) (variable 13282) (variable_class_name 13097)) 5512 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13281) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5513 (("(" 13278)) 5514 (("(" 13275)) 5515 (("(" 13266) (argument_list 13267) (ctor_arguments 13268)) 5516 (("\"" 1520) ("$" 13085) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 13086) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 13087) ("[" 1542) (array_object_dereferenceable 13088) (callable_expr 13089) (callable_variable 13090) (class_constant 1546) (class_name 13091) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 13092) (name 13093) (new_dereferenceable 1552) (simple_variable 13094) (static_member 13095) (variable 13265) (variable_class_name 13097)) 5517 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13264) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5518 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13263) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5519 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13262) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5520 (("(" 13258)) 5521 (("(" 13253)) 5522 (("(" 13247)) 5523 (("$" 13085) ("(" 13210) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 13211) (T_NAME_QUALIFIED 13212) (T_NAME_RELATIVE 13213) (T_READONLY 2931) (T_STATIC 13214) (T_STRING 13215) (T_VARIABLE 13087) (anonymous_class 13216) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 13217) (attribute 2937) (attributes 13218) (class_modifier 2939) (class_name 13219) (class_name_reference 13220) (name 13221) (new_variable 13222) (simple_variable 13223)) 5524 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13209) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5525 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13208) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5526 (("(" 13046) (argument_list 13207)) 5527 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13206) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5528 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13205) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5529 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 13200) (T_END_HEREDOC 13201) (T_VARIABLE 2896) (encaps_list 13202) (encaps_var 2898)) 5530 ((T_FN 35) (T_FUNCTION 129) (fn 12844) #'12845 (inline_function 13199)) 5531 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13198) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5532 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13197) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5533 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13196) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5534 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13193) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5535 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13192) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5536 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 13188) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 5537 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 13186) (encaps_list 2804) (encaps_var 2805)) 5538 ((T_NULLSAFE_OBJECT_OPERATOR 13172) (T_OBJECT_OPERATOR 13173) ("[" 13174)) 5539 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 13169) (attribute 2733) (fn 12844) #'12845 (inline_function 13170)) 5540 (("(" 13046) (argument_list 13168)) 5541 ((T_PAAMAYIM_NEKUDOTAYIM 13160)) 5542 (("%" 12860) ("*" 12861) ("+" 12862) ("-" 12863) ("." 12864) ("/" 12865) ("<" 12866) (">" 12867) ("?" 12868) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12869) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12870) (T_BOOLEAN_AND 12871) (T_BOOLEAN_OR 12872) (T_COALESCE 12873) (T_DOUBLE_ARROW 13158) (T_INSTANCEOF 12874) (T_IS_EQUAL 12875) (T_IS_GREATER_OR_EQUAL 12876) (T_IS_IDENTICAL 12877) (T_IS_NOT_EQUAL 12878) (T_IS_NOT_IDENTICAL 12879) (T_IS_SMALLER_OR_EQUAL 12880) (T_LOGICAL_AND 12881) (T_LOGICAL_OR 12882) (T_LOGICAL_XOR 12883) (T_POW 12884) (T_SL 12885) (T_SPACESHIP 12886) (T_SR 12887) ("^" 12888) ("|" 12889)) 5543 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 13147)) 5544 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 13135)) 5545 (("(" 13046) (argument_list 13134)) 5546 (("=" 13054) (T_AND_EQUAL 13055) (T_COALESCE_EQUAL 13056) (T_CONCAT_EQUAL 13057) (T_DEC 13058) (T_DIV_EQUAL 13059) (T_INC 13060) (T_MINUS_EQUAL 13061) (T_MOD_EQUAL 13062) (T_MUL_EQUAL 13063) (T_OR_EQUAL 13064) (T_PLUS_EQUAL 13065) (T_POW_EQUAL 13066) (T_SL_EQUAL 13067) (T_SR_EQUAL 13068) (T_XOR_EQUAL 13069)) 5547 ((T_PAAMAYIM_NEKUDOTAYIM 12958)) 5548 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12859) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5549 (("%" 12860) ("*" 12861) ("+" 12862) ("-" 12863) ("." 12864) ("/" 12865) ("<" 12866) (">" 12867) ("?" 12868) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12869) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12870) (T_BOOLEAN_AND 12871) (T_BOOLEAN_OR 12872) (T_COALESCE 12873) (T_INSTANCEOF 12874) (T_IS_EQUAL 12875) (T_IS_GREATER_OR_EQUAL 12876) (T_IS_IDENTICAL 12877) (T_IS_NOT_EQUAL 12878) (T_IS_NOT_IDENTICAL 12879) (T_IS_SMALLER_OR_EQUAL 12880) (T_LOGICAL_AND 12881) (T_LOGICAL_OR 12882) (T_LOGICAL_XOR 12883) (T_POW 12884) (T_SL 12885) (T_SPACESHIP 12886) (T_SR 12887) ("^" 12888) ("|" 12889)) 5550 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12957) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5551 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12956) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5552 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12955) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5553 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12954) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5554 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12953) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5555 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12952) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5556 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12951) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5557 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12950) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5558 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 12945) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 12946) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 5559 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12944) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5560 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12943) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5561 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12942) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5562 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12941) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5563 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12940) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5564 (("$" 12905) ("(" 12906) (T_NAME_FULLY_QUALIFIED 12907) (T_NAME_QUALIFIED 12908) (T_NAME_RELATIVE 12909) (T_STATIC 12910) (T_STRING 12911) (T_VARIABLE 12912) (class_name 12913) (class_name_reference 12914) (name 12915) (new_variable 12916) (simple_variable 12917)) 5565 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12904) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5566 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12903) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5567 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12902) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5568 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12901) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5569 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12900) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5570 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12899) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5571 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12898) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5572 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12897) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5573 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12896) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5574 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12895) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5575 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12894) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5576 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12893) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5577 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12892) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5578 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12891) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5579 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12890) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5580 (("$" 12905) (T_VARIABLE 12912) (simple_variable 12936) ("{" 12937)) 5581 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 12934) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5582 ((T_PAAMAYIM_NEKUDOTAYIM 12932)) 5583 ((T_NULLSAFE_OBJECT_OPERATOR 12918) (T_OBJECT_OPERATOR 12919) (T_PAAMAYIM_NEKUDOTAYIM 12920) ("[" 12921)) 5584 (("$" 12905) (T_STRING 12925) (T_VARIABLE 12912) (property_name 12931) (simple_variable 12927) ("{" 12928)) 5585 (("$" 12905) (T_STRING 12925) (T_VARIABLE 12912) (property_name 12926) (simple_variable 12927) ("{" 12928)) 5586 (("$" 12905) (T_VARIABLE 12912) (simple_variable 12924)) 5587 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 12922) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5588 (("]" 12923)) 5589 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12929) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5590 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 12930)) 5591 (("$" 12905) (T_VARIABLE 12912) (simple_variable 12933)) 5592 (("%" 299) (")" 12935) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5593 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 12938) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5594 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 12939)) 5595 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12949) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5596 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 12947) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 5597 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 12948) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5598 (("$" 12782) (T_ABSTRACT 12959) (T_ARRAY 12960) (T_AS 12961) (T_BREAK 12962) (T_CALLABLE 12963) (T_CASE 12964) (T_CATCH 12965) (T_CLASS 12966) (T_CLASS_C 12967) (T_CLONE 12968) (T_CONST 12969) (T_CONTINUE 12970) (T_DECLARE 12971) (T_DEFAULT 12972) (T_DIR 12973) (T_DO 12974) (T_ECHO 12975) (T_ELSE 12976) (T_ELSEIF 12977) (T_EMPTY 12978) (T_ENDDECLARE 12979) (T_ENDFOR 12980) (T_ENDFOREACH 12981) (T_ENDIF 12982) (T_ENDSWITCH 12983) (T_ENDWHILE 12984) (T_ENUM 12985) (T_EVAL 12986) (T_EXIT 12987) (T_EXTENDS 12988) (T_FILE 12989) (T_FINAL 12990) (T_FINALLY 12991) (T_FN 12992) (T_FOR 12993) (T_FOREACH 12994) (T_FUNCTION 12995) (T_FUNC_C 12996) (T_GLOBAL 12997) (T_GOTO 12998) (T_IF 12999) (T_IMPLEMENTS 13000) (T_INCLUDE 13001) (T_INCLUDE_ONCE 13002) (T_INSTANCEOF 13003) (T_INSTEADOF 13004) (T_INTERFACE 13005) (T_ISSET 13006) (T_LINE 13007) (T_LIST 13008) (T_LOGICAL_AND 13009) (T_LOGICAL_OR 13010) (T_LOGICAL_XOR 13011) (T_MATCH 13012) (T_METHOD_C 13013) (T_NAMESPACE 13014) (T_NEW 13015) (T_NS_C 13016) (T_PRINT 13017) (T_PRIVATE 13018) (T_PROPERTY_C 13019) (T_PROTECTED 13020) (T_PUBLIC 13021) (T_READONLY 13022) (T_REQUIRE 13023) (T_REQUIRE_ONCE 13024) (T_RETURN 13025) (T_STATIC 13026) (T_STRING 13027) (T_SWITCH 13028) (T_THROW 13029) (T_TRAIT 13030) (T_TRAIT_C 13031) (T_TRY 13032) (T_UNSET 13033) (T_USE 13034) (T_VAR 13035) (T_VARIABLE 12830) (T_WHILE 13036) (T_YIELD 13037) (identifier 13038) (member_name 13039) (reserved_non_modifiers 13040) (semi_reserved 13041) (simple_variable 13042) ("{" 13043)) 5599 (("(" 13046) (argument_list 13047)) 5600 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13044) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5601 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13045)) 5602 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 13048) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 13049) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 13050) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5603 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 13053) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5604 (("," 1480) (possible_comma 13051)) 5605 ((")" 13052)) 5606 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (ampersand 13083) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13084) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5607 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13082) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5608 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13081) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5609 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13080) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5610 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13079) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5611 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13078) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5612 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13077) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5613 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13076) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5614 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13075) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5615 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13074) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5616 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13073) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5617 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13072) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5618 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13071) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5619 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13070) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5620 (("\"" 1520) ("$" 13085) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 13086) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 13087) ("[" 1542) (array_object_dereferenceable 13088) (callable_expr 13089) (callable_variable 13090) (class_constant 1546) (class_name 13091) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 13092) (name 13093) (new_dereferenceable 1552) (simple_variable 13094) (static_member 13095) (variable 13096) (variable_class_name 13097)) 5621 (("$" 13085) (T_VARIABLE 13087) (simple_variable 13130) ("{" 13131)) 5622 (("(" 13101) (argument_list 13129)) 5623 ((T_NULLSAFE_OBJECT_OPERATOR 13115) (T_OBJECT_OPERATOR 13116) ("[" 13117)) 5624 (("(" 13101) (argument_list 13114)) 5625 ((T_PAAMAYIM_NEKUDOTAYIM 13110)) 5626 (("(" 13101) (argument_list 13109)) 5627 ((T_PAAMAYIM_NEKUDOTAYIM 13098)) 5628 (("$" 13085) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 13087) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 13099) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 13100) ("{" 1642)) 5629 (("(" 13101) (argument_list 13102)) 5630 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 13103) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 13104) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 13105) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5631 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 13108) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5632 (("," 1480) (possible_comma 13106)) 5633 ((")" 13107)) 5634 (("$" 13085) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 13087) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 13111) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 13112) ("{" 1658)) 5635 (("(" 13101) (argument_list 13113)) 5636 (("$" 13085) (T_STRING 13120) (T_VARIABLE 13087) (property_name 13127) (simple_variable 13122) ("{" 13123)) 5637 (("$" 13085) (T_STRING 13120) (T_VARIABLE 13087) (property_name 13121) (simple_variable 13122) ("{" 13123)) 5638 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 13118) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5639 (("]" 13119)) 5640 (("(" 13101) (argument_list 13126)) 5641 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13124) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5642 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13125)) 5643 (("(" 13101) (argument_list 13128)) 5644 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13132) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5645 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13133)) 5646 ((backup_doc_comment 13136)) 5647 (("(" 13137)) 5648 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 13138)) 5649 ((")" 13139)) 5650 ((T_USE 164) (lexical_vars 13140)) 5651 ((":" 166) (return_type 13141)) 5652 ((backup_fn_flags 13142)) 5653 (("{" 13143)) 5654 ((inner_statement_list 13144)) 5655 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 13145) ("~" 127)) 5656 ((backup_fn_flags 13146)) 5657 ((backup_doc_comment 13148)) 5658 (("(" 13149)) 5659 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 13150)) 5660 ((")" 13151)) 5661 ((":" 2249) (return_type 13152)) 5662 ((T_DOUBLE_ARROW 13153)) 5663 ((backup_fn_flags 13154)) 5664 ((backup_lex_pos 13155)) 5665 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13156) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5666 (("%" 12860) ("*" 12861) ("+" 12862) ("-" 12863) ("." 12864) ("/" 12865) ("<" 12866) (">" 12867) ("?" 12868) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12869) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12870) (T_BOOLEAN_AND 12871) (T_BOOLEAN_OR 12872) (T_COALESCE 12873) (T_INSTANCEOF 12874) (T_IS_EQUAL 12875) (T_IS_GREATER_OR_EQUAL 12876) (T_IS_IDENTICAL 12877) (T_IS_NOT_EQUAL 12878) (T_IS_NOT_IDENTICAL 12879) (T_IS_SMALLER_OR_EQUAL 12880) (T_LOGICAL_AND 12881) (T_LOGICAL_OR 12882) (T_LOGICAL_XOR 12883) (T_POW 12884) (T_SL 12885) (T_SPACESHIP 12886) (T_SR 12887) ("^" 12888) (backup_fn_flags 13157) ("|" 12889)) 5667 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13159) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5668 (("$" 12782) (T_ABSTRACT 12959) (T_ARRAY 12960) (T_AS 12961) (T_BREAK 12962) (T_CALLABLE 12963) (T_CASE 12964) (T_CATCH 12965) (T_CLASS 12966) (T_CLASS_C 12967) (T_CLONE 12968) (T_CONST 12969) (T_CONTINUE 12970) (T_DECLARE 12971) (T_DEFAULT 12972) (T_DIR 12973) (T_DO 12974) (T_ECHO 12975) (T_ELSE 12976) (T_ELSEIF 12977) (T_EMPTY 12978) (T_ENDDECLARE 12979) (T_ENDFOR 12980) (T_ENDFOREACH 12981) (T_ENDIF 12982) (T_ENDSWITCH 12983) (T_ENDWHILE 12984) (T_ENUM 12985) (T_EVAL 12986) (T_EXIT 12987) (T_EXTENDS 12988) (T_FILE 12989) (T_FINAL 12990) (T_FINALLY 12991) (T_FN 12992) (T_FOR 12993) (T_FOREACH 12994) (T_FUNCTION 12995) (T_FUNC_C 12996) (T_GLOBAL 12997) (T_GOTO 12998) (T_IF 12999) (T_IMPLEMENTS 13000) (T_INCLUDE 13001) (T_INCLUDE_ONCE 13002) (T_INSTANCEOF 13003) (T_INSTEADOF 13004) (T_INTERFACE 13005) (T_ISSET 13006) (T_LINE 13007) (T_LIST 13008) (T_LOGICAL_AND 13009) (T_LOGICAL_OR 13010) (T_LOGICAL_XOR 13011) (T_MATCH 13012) (T_METHOD_C 13013) (T_NAMESPACE 13014) (T_NEW 13015) (T_NS_C 13016) (T_PRINT 13017) (T_PRIVATE 13018) (T_PROPERTY_C 13019) (T_PROTECTED 13020) (T_PUBLIC 13021) (T_READONLY 13022) (T_REQUIRE 13023) (T_REQUIRE_ONCE 13024) (T_RETURN 13025) (T_STATIC 13026) (T_STRING 13027) (T_SWITCH 13028) (T_THROW 13029) (T_TRAIT 13030) (T_TRAIT_C 13031) (T_TRY 13032) (T_UNSET 13033) (T_USE 13034) (T_VAR 13035) (T_VARIABLE 12830) (T_WHILE 13036) (T_YIELD 13037) (identifier 13161) (member_name 13162) (reserved_non_modifiers 13040) (semi_reserved 13041) (simple_variable 13163) ("{" 13164)) 5669 (("(" 13046) (argument_list 13167)) 5670 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13165) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5671 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13166)) 5672 ((T_FN 35) (T_FUNCTION 129) (fn 12844) #'12845 (inline_function 13171)) 5673 (("$" 12782) (T_STRING 13177) (T_VARIABLE 12830) (property_name 13184) (simple_variable 13179) ("{" 13180)) 5674 (("$" 12782) (T_STRING 13177) (T_VARIABLE 12830) (property_name 13178) (simple_variable 13179) ("{" 13180)) 5675 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 13175) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5676 (("]" 13176)) 5677 (("(" 13046) (argument_list 13183)) 5678 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13181) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5679 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13182)) 5680 (("(" 13046) (argument_list 13185)) 5681 (("`" 13187)) 5682 (("]" 13189)) 5683 (("=" 13190)) 5684 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13191) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5685 (("%" 12860) ("*" 12861) ("+" 12862) ("-" 12863) ("." 12864) ("/" 12865) ("<" 12866) (">" 12867) ("?" 12868) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12869) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12870) (T_BOOLEAN_AND 12871) (T_BOOLEAN_OR 12872) (T_COALESCE 12873) (T_DOUBLE_ARROW 13194) (T_INSTANCEOF 12874) (T_IS_EQUAL 12875) (T_IS_GREATER_OR_EQUAL 12876) (T_IS_IDENTICAL 12877) (T_IS_NOT_EQUAL 12878) (T_IS_NOT_IDENTICAL 12879) (T_IS_SMALLER_OR_EQUAL 12880) (T_LOGICAL_AND 12881) (T_LOGICAL_OR 12882) (T_LOGICAL_XOR 12883) (T_POW 12884) (T_SL 12885) (T_SPACESHIP 12886) (T_SR 12887) ("^" 12888) ("|" 12889)) 5686 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13195) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5687 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 13204) (T_VARIABLE 2896) (encaps_var 2910)) 5688 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 13203) (T_VARIABLE 2896) (encaps_var 2901)) 5689 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13245) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5690 ((T_CLASS 13237)) 5691 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 13236) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 13217) (attribute 2958) (class_modifier 2939)) 5692 ((T_PAAMAYIM_NEKUDOTAYIM 13234)) 5693 (("(" 13101) (argument_list 13233)) 5694 ((T_NULLSAFE_OBJECT_OPERATOR 13224) (T_OBJECT_OPERATOR 13225) (T_PAAMAYIM_NEKUDOTAYIM 13226) ("[" 13227)) 5695 (("$" 13085) (T_STRING 13120) (T_VARIABLE 13087) (property_name 13232) (simple_variable 13122) ("{" 13123)) 5696 (("$" 13085) (T_STRING 13120) (T_VARIABLE 13087) (property_name 13231) (simple_variable 13122) ("{" 13123)) 5697 (("$" 13085) (T_VARIABLE 13087) (simple_variable 13230)) 5698 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 13228) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5699 (("]" 13229)) 5700 (("$" 13085) (T_VARIABLE 13087) (simple_variable 13235)) 5701 (("(" 2960) (argument_list 2961) (ctor_arguments 13238)) 5702 ((T_EXTENDS 2963) (extends_from 13239)) 5703 ((T_IMPLEMENTS 2965) (implements_list 13240)) 5704 ((backup_doc_comment 13241)) 5705 (("{" 13242)) 5706 ((class_statement_list 13243)) 5707 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 13244)) 5708 (("%" 299) (")" 13246) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5709 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13248) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5710 (("%" 299) (")" 13249) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5711 (("{" 13250)) 5712 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 13251) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 5713 (("}" 13252)) 5714 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 13254) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 5715 ((")" 13255)) 5716 (("=" 13256)) 5717 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13257) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5718 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 13259) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5719 (("," 5192) (possible_comma 13260)) 5720 ((")" 13261)) 5721 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 13269) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 13270) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 13271) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5722 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 13274) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5723 (("," 1480) (possible_comma 13272)) 5724 ((")" 13273)) 5725 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13276) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5726 (("%" 299) (")" 13277) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5727 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13279) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5728 (("%" 299) (")" 13280) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5729 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 13287) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 5730 ((")" 13288)) 5731 (("%" 299) (")" 13293) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5732 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13296) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5733 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13297)) 5734 (("\"" 13299) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 5735 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 13309) (T_VARIABLE 2896) (encaps_var 2910)) 5736 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 13308) (T_VARIABLE 2896) (encaps_var 2901)) 5737 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13350) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5738 ((T_CLASS 13342)) 5739 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 13341) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 13322) (attribute 2958) (class_modifier 2939)) 5740 ((T_PAAMAYIM_NEKUDOTAYIM 13339)) 5741 (("(" 12689) (argument_list 13338)) 5742 ((T_NULLSAFE_OBJECT_OPERATOR 13329) (T_OBJECT_OPERATOR 13330) (T_PAAMAYIM_NEKUDOTAYIM 13331) ("[" 13332)) 5743 (("$" 12673) (T_STRING 12708) (T_VARIABLE 12675) (property_name 13337) (simple_variable 12710) ("{" 12711)) 5744 (("$" 12673) (T_STRING 12708) (T_VARIABLE 12675) (property_name 13336) (simple_variable 12710) ("{" 12711)) 5745 (("$" 12673) (T_VARIABLE 12675) (simple_variable 13335)) 5746 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 13333) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5747 (("]" 13334)) 5748 (("$" 12673) (T_VARIABLE 12675) (simple_variable 13340)) 5749 (("(" 2960) (argument_list 2961) (ctor_arguments 13343)) 5750 ((T_EXTENDS 2963) (extends_from 13344)) 5751 ((T_IMPLEMENTS 2965) (implements_list 13345)) 5752 ((backup_doc_comment 13346)) 5753 (("{" 13347)) 5754 ((class_statement_list 13348)) 5755 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 13349)) 5756 (("%" 299) (")" 13351) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5757 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13353) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5758 (("%" 299) (")" 13354) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5759 (("{" 13355)) 5760 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 13356) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 5761 (("}" 13357)) 5762 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 13359) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 5763 ((")" 13360)) 5764 (("=" 13361)) 5765 (("!" 12368) ("\"" 12369) ("$" 12370) ("(" 12371) ("+" 12372) ("-" 12373) ("@" 12374) (T_ARRAY 12375) (T_ARRAY_CAST 12376) (T_ATTRIBUTE 128) (T_BOOL_CAST 12377) (T_CLASS_C 12378) (T_CLONE 12379) (T_CONSTANT_ENCAPSED_STRING 12380) (T_DEC 12381) (T_DIR 12382) (T_DNUMBER 12383) (T_DOUBLE_CAST 12384) (T_EMPTY 12385) (T_EVAL 12386) (T_EXIT 12387) (T_FILE 12388) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12389) (T_INC 12390) (T_INCLUDE 12391) (T_INCLUDE_ONCE 12392) (T_INT_CAST 12393) (T_ISSET 12394) (T_LINE 12395) (T_LIST 12396) (T_LNUMBER 12397) (T_MATCH 12398) (T_METHOD_C 12399) (T_NAME_FULLY_QUALIFIED 12400) (T_NAME_QUALIFIED 12401) (T_NAME_RELATIVE 12402) (T_NEW 12403) (T_NS_C 12404) (T_OBJECT_CAST 12405) (T_PRINT 12406) (T_PROPERTY_C 12407) (T_READONLY 12408) (T_REQUIRE 12409) (T_REQUIRE_ONCE 12410) (T_START_HEREDOC 12411) (T_STATIC 12412) (T_STRING 12413) (T_STRING_CAST 12414) (T_THROW 12415) (T_TRAIT_C 12416) (T_UNSET_CAST 12417) (T_VARIABLE 12418) (T_YIELD 12419) (T_YIELD_FROM 12420) ("[" 12421) ("`" 12422) (array_object_dereferenceable 12423) (attribute 133) (attributes 12424) (callable_expr 12425) (callable_variable 12426) (class_constant 12427) (class_name 12428) (constant 12429) (dereferenceable_scalar 12430) (expr 13362) (fn 12432) (fully_dereferenceable 105) #'12433 (function_call 12434) (inline_function 12435) (internal_functions_in_yacc 12436) (match 12437) (name 12438) (new_dereferenceable 12439) (new_non_dereferenceable 12440) (scalar 12441) (simple_variable 12442) (static_member 12443) (variable 12444) (variable_class_name 12445) ("~" 12446)) 5766 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 13364) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5767 (("," 5192) (possible_comma 13365)) 5768 ((")" 13366)) 5769 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 13374) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 13375) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 13376) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5770 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 13379) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5771 (("," 1480) (possible_comma 13377)) 5772 ((")" 13378)) 5773 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13381) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5774 (("%" 299) (")" 13382) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5775 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13384) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5776 (("%" 299) (")" 13385) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5777 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 13392) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 5778 ((")" 13393)) 5779 (("%" 299) (")" 13398) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5780 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13401) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5781 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13402)) 5782 (("\"" 13404) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 5783 (("$" 513) (T_ABSTRACT 12171) (T_ARRAY 12172) (T_AS 12173) (T_BREAK 12174) (T_CALLABLE 12175) (T_CASE 12176) (T_CATCH 12177) (T_CLASS 12178) (T_CLASS_C 12179) (T_CLONE 12180) (T_CONST 12181) (T_CONTINUE 12182) (T_DECLARE 12183) (T_DEFAULT 12184) (T_DIR 12185) (T_DO 12186) (T_ECHO 12187) (T_ELSE 12188) (T_ELSEIF 12189) (T_EMPTY 12190) (T_ENDDECLARE 12191) (T_ENDFOR 12192) (T_ENDFOREACH 12193) (T_ENDIF 12194) (T_ENDSWITCH 12195) (T_ENDWHILE 12196) (T_ENUM 12197) (T_EVAL 12198) (T_EXIT 12199) (T_EXTENDS 12200) (T_FILE 12201) (T_FINAL 12202) (T_FINALLY 12203) (T_FN 12204) (T_FOR 12205) (T_FOREACH 12206) (T_FUNCTION 12207) (T_FUNC_C 12208) (T_GLOBAL 12209) (T_GOTO 12210) (T_IF 12211) (T_IMPLEMENTS 12212) (T_INCLUDE 12213) (T_INCLUDE_ONCE 12214) (T_INSTANCEOF 12215) (T_INSTEADOF 12216) (T_INTERFACE 12217) (T_ISSET 12218) (T_LINE 12219) (T_LIST 12220) (T_LOGICAL_AND 12221) (T_LOGICAL_OR 12222) (T_LOGICAL_XOR 12223) (T_MATCH 12224) (T_METHOD_C 12225) (T_NAMESPACE 12226) (T_NEW 12227) (T_NS_C 12228) (T_PRINT 12229) (T_PRIVATE 12230) (T_PROPERTY_C 12231) (T_PROTECTED 12232) (T_PUBLIC 12233) (T_READONLY 12234) (T_REQUIRE 12235) (T_REQUIRE_ONCE 12236) (T_RETURN 12237) (T_STATIC 12238) (T_STRING 12239) (T_SWITCH 12240) (T_THROW 12241) (T_TRAIT 12242) (T_TRAIT_C 12243) (T_TRY 12244) (T_UNSET 12245) (T_USE 12246) (T_VAR 12247) (T_VARIABLE 561) (T_WHILE 12248) (T_YIELD 12249) (identifier 13408) (member_name 13409) (reserved_non_modifiers 12252) (semi_reserved 12253) (simple_variable 13410) ("{" 13411)) 5784 (("(" 12258) (argument_list 13414)) 5785 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13412) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5786 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13413)) 5787 ((T_FN 35) (T_FUNCTION 129) (fn 575) #'576 (inline_function 13418)) 5788 (("$" 513) (T_STRING 13424) (T_VARIABLE 561) (property_name 13431) (simple_variable 13426) ("{" 13427)) 5789 (("$" 513) (T_STRING 13424) (T_VARIABLE 561) (property_name 13425) (simple_variable 13426) ("{" 13427)) 5790 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 13422) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5791 (("]" 13423)) 5792 (("(" 12258) (argument_list 13430)) 5793 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13428) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5794 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13429)) 5795 (("(" 12258) (argument_list 13432)) 5796 (("`" 13434)) 5797 (("]" 13436)) 5798 (("=" 13437)) 5799 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13438) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5800 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13960) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5801 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 13958) (encaps_var 5237)) 5802 (("$" 13442) (T_VARIABLE 13490) (simple_variable 13954) ("{" 13955)) 5803 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13952) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5804 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13951) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5805 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13950) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5806 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13949) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5807 (("(" 13946)) 5808 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13945) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5809 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13944) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5810 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13943) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5811 (("\"" 1520) ("$" 13745) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 13746) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 13747) ("[" 1542) (array_object_dereferenceable 13748) (callable_expr 13749) (callable_variable 13750) (class_constant 1546) (class_name 13751) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 13752) (name 13753) (new_dereferenceable 1552) (simple_variable 13754) (static_member 13755) (variable 13942) (variable_class_name 13757)) 5812 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13941) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5813 (("(" 13938)) 5814 (("(" 13935)) 5815 (("(" 13926) (argument_list 13927) (ctor_arguments 13928)) 5816 (("\"" 1520) ("$" 13745) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 13746) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 13747) ("[" 1542) (array_object_dereferenceable 13748) (callable_expr 13749) (callable_variable 13750) (class_constant 1546) (class_name 13751) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 13752) (name 13753) (new_dereferenceable 1552) (simple_variable 13754) (static_member 13755) (variable 13925) (variable_class_name 13757)) 5817 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13924) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5818 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13923) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5819 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13922) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5820 (("(" 13918)) 5821 (("(" 13913)) 5822 (("(" 13907)) 5823 (("$" 13745) ("(" 13870) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 13871) (T_NAME_QUALIFIED 13872) (T_NAME_RELATIVE 13873) (T_READONLY 2931) (T_STATIC 13874) (T_STRING 13875) (T_VARIABLE 13747) (anonymous_class 13876) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 13877) (attribute 2937) (attributes 13878) (class_modifier 2939) (class_name 13879) (class_name_reference 13880) (name 13881) (new_variable 13882) (simple_variable 13883)) 5824 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13869) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5825 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13868) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5826 (("(" 13706) (argument_list 13867)) 5827 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13866) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5828 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13865) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5829 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 13860) (T_END_HEREDOC 13861) (T_VARIABLE 2896) (encaps_list 13862) (encaps_var 2898)) 5830 ((T_FN 35) (T_FUNCTION 129) (fn 13504) #'13505 (inline_function 13859)) 5831 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13858) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5832 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13857) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5833 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13856) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5834 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13853) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5835 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13852) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5836 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 13848) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 5837 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 13846) (encaps_list 2804) (encaps_var 2805)) 5838 ((T_NULLSAFE_OBJECT_OPERATOR 13832) (T_OBJECT_OPERATOR 13833) ("[" 13834)) 5839 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 13829) (attribute 2733) (fn 13504) #'13505 (inline_function 13830)) 5840 (("(" 13706) (argument_list 13828)) 5841 ((T_PAAMAYIM_NEKUDOTAYIM 13820)) 5842 (("%" 13520) ("*" 13521) ("+" 13522) ("-" 13523) ("." 13524) ("/" 13525) ("<" 13526) (">" 13527) ("?" 13528) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13529) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13530) (T_BOOLEAN_AND 13531) (T_BOOLEAN_OR 13532) (T_COALESCE 13533) (T_DOUBLE_ARROW 13818) (T_INSTANCEOF 13534) (T_IS_EQUAL 13535) (T_IS_GREATER_OR_EQUAL 13536) (T_IS_IDENTICAL 13537) (T_IS_NOT_EQUAL 13538) (T_IS_NOT_IDENTICAL 13539) (T_IS_SMALLER_OR_EQUAL 13540) (T_LOGICAL_AND 13541) (T_LOGICAL_OR 13542) (T_LOGICAL_XOR 13543) (T_POW 13544) (T_SL 13545) (T_SPACESHIP 13546) (T_SR 13547) ("^" 13548) ("|" 13549)) 5843 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 13807)) 5844 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 13795)) 5845 (("(" 13706) (argument_list 13794)) 5846 (("=" 13714) (T_AND_EQUAL 13715) (T_COALESCE_EQUAL 13716) (T_CONCAT_EQUAL 13717) (T_DEC 13718) (T_DIV_EQUAL 13719) (T_INC 13720) (T_MINUS_EQUAL 13721) (T_MOD_EQUAL 13722) (T_MUL_EQUAL 13723) (T_OR_EQUAL 13724) (T_PLUS_EQUAL 13725) (T_POW_EQUAL 13726) (T_SL_EQUAL 13727) (T_SR_EQUAL 13728) (T_XOR_EQUAL 13729)) 5847 ((T_PAAMAYIM_NEKUDOTAYIM 13618)) 5848 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13519) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5849 (("%" 13520) ("*" 13521) ("+" 13522) ("-" 13523) ("." 13524) ("/" 13525) ("<" 13526) (">" 13527) ("?" 13528) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13529) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13530) (T_BOOLEAN_AND 13531) (T_BOOLEAN_OR 13532) (T_COALESCE 13533) (T_INSTANCEOF 13534) (T_IS_EQUAL 13535) (T_IS_GREATER_OR_EQUAL 13536) (T_IS_IDENTICAL 13537) (T_IS_NOT_EQUAL 13538) (T_IS_NOT_IDENTICAL 13539) (T_IS_SMALLER_OR_EQUAL 13540) (T_LOGICAL_AND 13541) (T_LOGICAL_OR 13542) (T_LOGICAL_XOR 13543) (T_POW 13544) (T_SL 13545) (T_SPACESHIP 13546) (T_SR 13547) ("^" 13548) ("|" 13549)) 5850 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13617) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5851 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13616) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5852 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13615) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5853 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13614) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5854 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13613) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5855 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13612) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5856 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13611) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5857 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13610) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5858 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 13605) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 13606) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 5859 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13604) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5860 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13603) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5861 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13602) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5862 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13601) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5863 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13600) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5864 (("$" 13565) ("(" 13566) (T_NAME_FULLY_QUALIFIED 13567) (T_NAME_QUALIFIED 13568) (T_NAME_RELATIVE 13569) (T_STATIC 13570) (T_STRING 13571) (T_VARIABLE 13572) (class_name 13573) (class_name_reference 13574) (name 13575) (new_variable 13576) (simple_variable 13577)) 5865 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13564) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5866 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13563) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5867 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13562) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5868 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13561) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5869 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13560) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5870 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13559) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5871 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13558) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5872 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13557) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5873 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13556) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5874 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13555) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5875 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13554) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5876 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13553) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5877 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13552) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5878 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13551) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5879 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13550) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5880 (("$" 13565) (T_VARIABLE 13572) (simple_variable 13596) ("{" 13597)) 5881 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13594) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5882 ((T_PAAMAYIM_NEKUDOTAYIM 13592)) 5883 ((T_NULLSAFE_OBJECT_OPERATOR 13578) (T_OBJECT_OPERATOR 13579) (T_PAAMAYIM_NEKUDOTAYIM 13580) ("[" 13581)) 5884 (("$" 13565) (T_STRING 13585) (T_VARIABLE 13572) (property_name 13591) (simple_variable 13587) ("{" 13588)) 5885 (("$" 13565) (T_STRING 13585) (T_VARIABLE 13572) (property_name 13586) (simple_variable 13587) ("{" 13588)) 5886 (("$" 13565) (T_VARIABLE 13572) (simple_variable 13584)) 5887 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 13582) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5888 (("]" 13583)) 5889 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13589) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5890 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13590)) 5891 (("$" 13565) (T_VARIABLE 13572) (simple_variable 13593)) 5892 (("%" 299) (")" 13595) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 5893 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13598) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5894 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13599)) 5895 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13609) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5896 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 13607) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 5897 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13608) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5898 (("$" 13442) (T_ABSTRACT 13619) (T_ARRAY 13620) (T_AS 13621) (T_BREAK 13622) (T_CALLABLE 13623) (T_CASE 13624) (T_CATCH 13625) (T_CLASS 13626) (T_CLASS_C 13627) (T_CLONE 13628) (T_CONST 13629) (T_CONTINUE 13630) (T_DECLARE 13631) (T_DEFAULT 13632) (T_DIR 13633) (T_DO 13634) (T_ECHO 13635) (T_ELSE 13636) (T_ELSEIF 13637) (T_EMPTY 13638) (T_ENDDECLARE 13639) (T_ENDFOR 13640) (T_ENDFOREACH 13641) (T_ENDIF 13642) (T_ENDSWITCH 13643) (T_ENDWHILE 13644) (T_ENUM 13645) (T_EVAL 13646) (T_EXIT 13647) (T_EXTENDS 13648) (T_FILE 13649) (T_FINAL 13650) (T_FINALLY 13651) (T_FN 13652) (T_FOR 13653) (T_FOREACH 13654) (T_FUNCTION 13655) (T_FUNC_C 13656) (T_GLOBAL 13657) (T_GOTO 13658) (T_IF 13659) (T_IMPLEMENTS 13660) (T_INCLUDE 13661) (T_INCLUDE_ONCE 13662) (T_INSTANCEOF 13663) (T_INSTEADOF 13664) (T_INTERFACE 13665) (T_ISSET 13666) (T_LINE 13667) (T_LIST 13668) (T_LOGICAL_AND 13669) (T_LOGICAL_OR 13670) (T_LOGICAL_XOR 13671) (T_MATCH 13672) (T_METHOD_C 13673) (T_NAMESPACE 13674) (T_NEW 13675) (T_NS_C 13676) (T_PRINT 13677) (T_PRIVATE 13678) (T_PROPERTY_C 13679) (T_PROTECTED 13680) (T_PUBLIC 13681) (T_READONLY 13682) (T_REQUIRE 13683) (T_REQUIRE_ONCE 13684) (T_RETURN 13685) (T_STATIC 13686) (T_STRING 13687) (T_SWITCH 13688) (T_THROW 13689) (T_TRAIT 13690) (T_TRAIT_C 13691) (T_TRY 13692) (T_UNSET 13693) (T_USE 13694) (T_VAR 13695) (T_VARIABLE 13490) (T_WHILE 13696) (T_YIELD 13697) (identifier 13698) (member_name 13699) (reserved_non_modifiers 13700) (semi_reserved 13701) (simple_variable 13702) ("{" 13703)) 5899 (("(" 13706) (argument_list 13707)) 5900 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13704) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5901 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13705)) 5902 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 13708) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 13709) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 13710) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5903 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 13713) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5904 (("," 1480) (possible_comma 13711)) 5905 ((")" 13712)) 5906 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (ampersand 13743) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13744) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5907 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13742) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5908 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13741) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5909 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13740) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5910 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13739) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5911 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13738) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5912 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13737) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5913 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13736) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5914 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13735) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5915 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13734) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5916 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13733) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5917 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13732) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5918 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13731) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5919 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13730) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5920 (("\"" 1520) ("$" 13745) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 13746) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 13747) ("[" 1542) (array_object_dereferenceable 13748) (callable_expr 13749) (callable_variable 13750) (class_constant 1546) (class_name 13751) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 13752) (name 13753) (new_dereferenceable 1552) (simple_variable 13754) (static_member 13755) (variable 13756) (variable_class_name 13757)) 5921 (("$" 13745) (T_VARIABLE 13747) (simple_variable 13790) ("{" 13791)) 5922 (("(" 13761) (argument_list 13789)) 5923 ((T_NULLSAFE_OBJECT_OPERATOR 13775) (T_OBJECT_OPERATOR 13776) ("[" 13777)) 5924 (("(" 13761) (argument_list 13774)) 5925 ((T_PAAMAYIM_NEKUDOTAYIM 13770)) 5926 (("(" 13761) (argument_list 13769)) 5927 ((T_PAAMAYIM_NEKUDOTAYIM 13758)) 5928 (("$" 13745) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 13747) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 13759) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 13760) ("{" 1642)) 5929 (("(" 13761) (argument_list 13762)) 5930 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 13763) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 13764) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 13765) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5931 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 13768) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 5932 (("," 1480) (possible_comma 13766)) 5933 ((")" 13767)) 5934 (("$" 13745) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 13747) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 13771) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 13772) ("{" 1658)) 5935 (("(" 13761) (argument_list 13773)) 5936 (("$" 13745) (T_STRING 13780) (T_VARIABLE 13747) (property_name 13787) (simple_variable 13782) ("{" 13783)) 5937 (("$" 13745) (T_STRING 13780) (T_VARIABLE 13747) (property_name 13781) (simple_variable 13782) ("{" 13783)) 5938 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 13778) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5939 (("]" 13779)) 5940 (("(" 13761) (argument_list 13786)) 5941 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13784) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5942 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13785)) 5943 (("(" 13761) (argument_list 13788)) 5944 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13792) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5945 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13793)) 5946 ((backup_doc_comment 13796)) 5947 (("(" 13797)) 5948 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 13798)) 5949 ((")" 13799)) 5950 ((T_USE 164) (lexical_vars 13800)) 5951 ((":" 166) (return_type 13801)) 5952 ((backup_fn_flags 13802)) 5953 (("{" 13803)) 5954 ((inner_statement_list 13804)) 5955 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 13805) ("~" 127)) 5956 ((backup_fn_flags 13806)) 5957 ((backup_doc_comment 13808)) 5958 (("(" 13809)) 5959 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 13810)) 5960 ((")" 13811)) 5961 ((":" 2249) (return_type 13812)) 5962 ((T_DOUBLE_ARROW 13813)) 5963 ((backup_fn_flags 13814)) 5964 ((backup_lex_pos 13815)) 5965 (("!" 12780) ("\"" 12781) ("$" 12782) ("(" 12783) ("+" 12784) ("-" 12785) ("@" 12786) (T_ARRAY 12787) (T_ARRAY_CAST 12788) (T_ATTRIBUTE 128) (T_BOOL_CAST 12789) (T_CLASS_C 12790) (T_CLONE 12791) (T_CONSTANT_ENCAPSED_STRING 12792) (T_DEC 12793) (T_DIR 12794) (T_DNUMBER 12795) (T_DOUBLE_CAST 12796) (T_EMPTY 12797) (T_EVAL 12798) (T_EXIT 12799) (T_FILE 12800) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 12801) (T_INC 12802) (T_INCLUDE 12803) (T_INCLUDE_ONCE 12804) (T_INT_CAST 12805) (T_ISSET 12806) (T_LINE 12807) (T_LIST 12808) (T_LNUMBER 12809) (T_MATCH 12810) (T_METHOD_C 12811) (T_NAME_FULLY_QUALIFIED 12812) (T_NAME_QUALIFIED 12813) (T_NAME_RELATIVE 12814) (T_NEW 12815) (T_NS_C 12816) (T_OBJECT_CAST 12817) (T_PRINT 12818) (T_PROPERTY_C 12819) (T_READONLY 12820) (T_REQUIRE 12821) (T_REQUIRE_ONCE 12822) (T_START_HEREDOC 12823) (T_STATIC 12824) (T_STRING 12825) (T_STRING_CAST 12826) (T_THROW 12827) (T_TRAIT_C 12828) (T_UNSET_CAST 12829) (T_VARIABLE 12830) (T_YIELD 12831) (T_YIELD_FROM 12832) ("[" 12833) ("`" 12834) (array_object_dereferenceable 12835) (attribute 133) (attributes 12836) (callable_expr 12837) (callable_variable 12838) (class_constant 12839) (class_name 12840) (constant 12841) (dereferenceable_scalar 12842) (expr 13816) (fn 12844) (fully_dereferenceable 105) #'12845 (function_call 12846) (inline_function 12847) (internal_functions_in_yacc 12848) (match 12849) (name 12850) (new_dereferenceable 12851) (new_non_dereferenceable 12852) (scalar 12853) (simple_variable 12854) (static_member 12855) (variable 12856) (variable_class_name 12857) ("~" 12858)) 5966 (("%" 12860) ("*" 12861) ("+" 12862) ("-" 12863) ("." 12864) ("/" 12865) ("<" 12866) (">" 12867) ("?" 12868) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12869) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12870) (T_BOOLEAN_AND 12871) (T_BOOLEAN_OR 12872) (T_COALESCE 12873) (T_INSTANCEOF 12874) (T_IS_EQUAL 12875) (T_IS_GREATER_OR_EQUAL 12876) (T_IS_IDENTICAL 12877) (T_IS_NOT_EQUAL 12878) (T_IS_NOT_IDENTICAL 12879) (T_IS_SMALLER_OR_EQUAL 12880) (T_LOGICAL_AND 12881) (T_LOGICAL_OR 12882) (T_LOGICAL_XOR 12883) (T_POW 12884) (T_SL 12885) (T_SPACESHIP 12886) (T_SR 12887) ("^" 12888) (backup_fn_flags 13817) ("|" 12889)) 5967 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13819) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5968 (("$" 13442) (T_ABSTRACT 13619) (T_ARRAY 13620) (T_AS 13621) (T_BREAK 13622) (T_CALLABLE 13623) (T_CASE 13624) (T_CATCH 13625) (T_CLASS 13626) (T_CLASS_C 13627) (T_CLONE 13628) (T_CONST 13629) (T_CONTINUE 13630) (T_DECLARE 13631) (T_DEFAULT 13632) (T_DIR 13633) (T_DO 13634) (T_ECHO 13635) (T_ELSE 13636) (T_ELSEIF 13637) (T_EMPTY 13638) (T_ENDDECLARE 13639) (T_ENDFOR 13640) (T_ENDFOREACH 13641) (T_ENDIF 13642) (T_ENDSWITCH 13643) (T_ENDWHILE 13644) (T_ENUM 13645) (T_EVAL 13646) (T_EXIT 13647) (T_EXTENDS 13648) (T_FILE 13649) (T_FINAL 13650) (T_FINALLY 13651) (T_FN 13652) (T_FOR 13653) (T_FOREACH 13654) (T_FUNCTION 13655) (T_FUNC_C 13656) (T_GLOBAL 13657) (T_GOTO 13658) (T_IF 13659) (T_IMPLEMENTS 13660) (T_INCLUDE 13661) (T_INCLUDE_ONCE 13662) (T_INSTANCEOF 13663) (T_INSTEADOF 13664) (T_INTERFACE 13665) (T_ISSET 13666) (T_LINE 13667) (T_LIST 13668) (T_LOGICAL_AND 13669) (T_LOGICAL_OR 13670) (T_LOGICAL_XOR 13671) (T_MATCH 13672) (T_METHOD_C 13673) (T_NAMESPACE 13674) (T_NEW 13675) (T_NS_C 13676) (T_PRINT 13677) (T_PRIVATE 13678) (T_PROPERTY_C 13679) (T_PROTECTED 13680) (T_PUBLIC 13681) (T_READONLY 13682) (T_REQUIRE 13683) (T_REQUIRE_ONCE 13684) (T_RETURN 13685) (T_STATIC 13686) (T_STRING 13687) (T_SWITCH 13688) (T_THROW 13689) (T_TRAIT 13690) (T_TRAIT_C 13691) (T_TRY 13692) (T_UNSET 13693) (T_USE 13694) (T_VAR 13695) (T_VARIABLE 13490) (T_WHILE 13696) (T_YIELD 13697) (identifier 13821) (member_name 13822) (reserved_non_modifiers 13700) (semi_reserved 13701) (simple_variable 13823) ("{" 13824)) 5969 (("(" 13706) (argument_list 13827)) 5970 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13825) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5971 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13826)) 5972 ((T_FN 35) (T_FUNCTION 129) (fn 13504) #'13505 (inline_function 13831)) 5973 (("$" 13442) (T_STRING 13837) (T_VARIABLE 13490) (property_name 13844) (simple_variable 13839) ("{" 13840)) 5974 (("$" 13442) (T_STRING 13837) (T_VARIABLE 13490) (property_name 13838) (simple_variable 13839) ("{" 13840)) 5975 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 13835) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5976 (("]" 13836)) 5977 (("(" 13706) (argument_list 13843)) 5978 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13841) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 5979 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13842)) 5980 (("(" 13706) (argument_list 13845)) 5981 (("`" 13847)) 5982 (("]" 13849)) 5983 (("=" 13850)) 5984 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13851) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5985 (("%" 13520) ("*" 13521) ("+" 13522) ("-" 13523) ("." 13524) ("/" 13525) ("<" 13526) (">" 13527) ("?" 13528) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13529) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13530) (T_BOOLEAN_AND 13531) (T_BOOLEAN_OR 13532) (T_COALESCE 13533) (T_DOUBLE_ARROW 13854) (T_INSTANCEOF 13534) (T_IS_EQUAL 13535) (T_IS_GREATER_OR_EQUAL 13536) (T_IS_IDENTICAL 13537) (T_IS_NOT_EQUAL 13538) (T_IS_NOT_IDENTICAL 13539) (T_IS_SMALLER_OR_EQUAL 13540) (T_LOGICAL_AND 13541) (T_LOGICAL_OR 13542) (T_LOGICAL_XOR 13543) (T_POW 13544) (T_SL 13545) (T_SPACESHIP 13546) (T_SR 13547) ("^" 13548) ("|" 13549)) 5986 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13855) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 5987 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 13864) (T_VARIABLE 2896) (encaps_var 2910)) 5988 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 13863) (T_VARIABLE 2896) (encaps_var 2901)) 5989 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13905) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 5990 ((T_CLASS 13897)) 5991 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 13896) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 13877) (attribute 2958) (class_modifier 2939)) 5992 ((T_PAAMAYIM_NEKUDOTAYIM 13894)) 5993 (("(" 13761) (argument_list 13893)) 5994 ((T_NULLSAFE_OBJECT_OPERATOR 13884) (T_OBJECT_OPERATOR 13885) (T_PAAMAYIM_NEKUDOTAYIM 13886) ("[" 13887)) 5995 (("$" 13745) (T_STRING 13780) (T_VARIABLE 13747) (property_name 13892) (simple_variable 13782) ("{" 13783)) 5996 (("$" 13745) (T_STRING 13780) (T_VARIABLE 13747) (property_name 13891) (simple_variable 13782) ("{" 13783)) 5997 (("$" 13745) (T_VARIABLE 13747) (simple_variable 13890)) 5998 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 13888) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 5999 (("]" 13889)) 6000 (("$" 13745) (T_VARIABLE 13747) (simple_variable 13895)) 6001 (("(" 2960) (argument_list 2961) (ctor_arguments 13898)) 6002 ((T_EXTENDS 2963) (extends_from 13899)) 6003 ((T_IMPLEMENTS 2965) (implements_list 13900)) 6004 ((backup_doc_comment 13901)) 6005 (("{" 13902)) 6006 ((class_statement_list 13903)) 6007 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 13904)) 6008 (("%" 299) (")" 13906) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6009 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13908) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6010 (("%" 299) (")" 13909) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6011 (("{" 13910)) 6012 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 13911) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 6013 (("}" 13912)) 6014 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 13914) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 6015 ((")" 13915)) 6016 (("=" 13916)) 6017 (("!" 13440) ("\"" 13441) ("$" 13442) ("(" 13443) ("+" 13444) ("-" 13445) ("@" 13446) (T_ARRAY 13447) (T_ARRAY_CAST 13448) (T_ATTRIBUTE 128) (T_BOOL_CAST 13449) (T_CLASS_C 13450) (T_CLONE 13451) (T_CONSTANT_ENCAPSED_STRING 13452) (T_DEC 13453) (T_DIR 13454) (T_DNUMBER 13455) (T_DOUBLE_CAST 13456) (T_EMPTY 13457) (T_EVAL 13458) (T_EXIT 13459) (T_FILE 13460) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 13461) (T_INC 13462) (T_INCLUDE 13463) (T_INCLUDE_ONCE 13464) (T_INT_CAST 13465) (T_ISSET 13466) (T_LINE 13467) (T_LIST 13468) (T_LNUMBER 13469) (T_MATCH 13470) (T_METHOD_C 13471) (T_NAME_FULLY_QUALIFIED 13472) (T_NAME_QUALIFIED 13473) (T_NAME_RELATIVE 13474) (T_NEW 13475) (T_NS_C 13476) (T_OBJECT_CAST 13477) (T_PRINT 13478) (T_PROPERTY_C 13479) (T_READONLY 13480) (T_REQUIRE 13481) (T_REQUIRE_ONCE 13482) (T_START_HEREDOC 13483) (T_STATIC 13484) (T_STRING 13485) (T_STRING_CAST 13486) (T_THROW 13487) (T_TRAIT_C 13488) (T_UNSET_CAST 13489) (T_VARIABLE 13490) (T_YIELD 13491) (T_YIELD_FROM 13492) ("[" 13493) ("`" 13494) (array_object_dereferenceable 13495) (attribute 133) (attributes 13496) (callable_expr 13497) (callable_variable 13498) (class_constant 13499) (class_name 13500) (constant 13501) (dereferenceable_scalar 13502) (expr 13917) (fn 13504) (fully_dereferenceable 105) #'13505 (function_call 13506) (inline_function 13507) (internal_functions_in_yacc 13508) (match 13509) (name 13510) (new_dereferenceable 13511) (new_non_dereferenceable 13512) (scalar 13513) (simple_variable 13514) (static_member 13515) (variable 13516) (variable_class_name 13517) ("~" 13518)) 6018 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 13919) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6019 (("," 5192) (possible_comma 13920)) 6020 ((")" 13921)) 6021 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 13929) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 13930) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 13931) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6022 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 13934) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6023 (("," 1480) (possible_comma 13932)) 6024 ((")" 13933)) 6025 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13936) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6026 (("%" 299) (")" 13937) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6027 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 13939) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6028 (("%" 299) (")" 13940) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6029 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 13947) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 6030 ((")" 13948)) 6031 (("%" 299) (")" 13953) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6032 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 13956) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6033 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 13957)) 6034 (("\"" 13959) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 6035 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 13969) (T_VARIABLE 2896) (encaps_var 2910)) 6036 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 13968) (T_VARIABLE 2896) (encaps_var 2901)) 6037 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 14010) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6038 ((T_CLASS 14002)) 6039 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 14001) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 13982) (attribute 2958) (class_modifier 2939)) 6040 ((T_PAAMAYIM_NEKUDOTAYIM 13999)) 6041 (("(" 12313) (argument_list 13998)) 6042 ((T_NULLSAFE_OBJECT_OPERATOR 13989) (T_OBJECT_OPERATOR 13990) (T_PAAMAYIM_NEKUDOTAYIM 13991) ("[" 13992)) 6043 (("$" 12297) (T_STRING 12332) (T_VARIABLE 12299) (property_name 13997) (simple_variable 12334) ("{" 12335)) 6044 (("$" 12297) (T_STRING 12332) (T_VARIABLE 12299) (property_name 13996) (simple_variable 12334) ("{" 12335)) 6045 (("$" 12297) (T_VARIABLE 12299) (simple_variable 13995)) 6046 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 13993) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6047 (("]" 13994)) 6048 (("$" 12297) (T_VARIABLE 12299) (simple_variable 14000)) 6049 (("(" 2960) (argument_list 2961) (ctor_arguments 14003)) 6050 ((T_EXTENDS 2963) (extends_from 14004)) 6051 ((T_IMPLEMENTS 2965) (implements_list 14005)) 6052 ((backup_doc_comment 14006)) 6053 (("{" 14007)) 6054 ((class_statement_list 14008)) 6055 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 14009)) 6056 (("%" 299) (")" 14011) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6057 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 14013) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6058 (("%" 299) (")" 14014) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6059 (("{" 14015)) 6060 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 14016) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 6061 (("}" 14017)) 6062 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 14019) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 6063 ((")" 14020)) 6064 (("=" 14021)) 6065 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14022) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6066 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 14024) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6067 (("," 5192) (possible_comma 14025)) 6068 ((")" 14026)) 6069 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 14034) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 14035) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 14036) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6070 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 14039) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6071 (("," 1480) (possible_comma 14037)) 6072 ((")" 14038)) 6073 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 14041) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6074 (("%" 299) (")" 14042) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6075 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 14044) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6076 (("%" 299) (")" 14045) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6077 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 14052) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 6078 ((")" 14053)) 6079 (("%" 299) (")" 14058) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6080 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14061) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6081 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 14062)) 6082 (("\"" 14064) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 6083 (("$" 487) (T_VARIABLE 494) (simple_variable 14068)) 6084 (("%" 299) (")" 14070) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6085 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14073) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6086 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 14074)) 6087 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14084) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6088 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 14082) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 6089 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14083) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6090 (("$" 363) (T_ABSTRACT 14094) (T_ARRAY 14095) (T_AS 14096) (T_BREAK 14097) (T_CALLABLE 14098) (T_CASE 14099) (T_CATCH 14100) (T_CLASS 14101) (T_CLASS_C 14102) (T_CLONE 14103) (T_CONST 14104) (T_CONTINUE 14105) (T_DECLARE 14106) (T_DEFAULT 14107) (T_DIR 14108) (T_DO 14109) (T_ECHO 14110) (T_ELSE 14111) (T_ELSEIF 14112) (T_EMPTY 14113) (T_ENDDECLARE 14114) (T_ENDFOR 14115) (T_ENDFOREACH 14116) (T_ENDIF 14117) (T_ENDSWITCH 14118) (T_ENDWHILE 14119) (T_ENUM 14120) (T_EVAL 14121) (T_EXIT 14122) (T_EXTENDS 14123) (T_FILE 14124) (T_FINAL 14125) (T_FINALLY 14126) (T_FN 14127) (T_FOR 14128) (T_FOREACH 14129) (T_FUNCTION 14130) (T_FUNC_C 14131) (T_GLOBAL 14132) (T_GOTO 14133) (T_IF 14134) (T_IMPLEMENTS 14135) (T_INCLUDE 14136) (T_INCLUDE_ONCE 14137) (T_INSTANCEOF 14138) (T_INSTEADOF 14139) (T_INTERFACE 14140) (T_ISSET 14141) (T_LINE 14142) (T_LIST 14143) (T_LOGICAL_AND 14144) (T_LOGICAL_OR 14145) (T_LOGICAL_XOR 14146) (T_MATCH 14147) (T_METHOD_C 14148) (T_NAMESPACE 14149) (T_NEW 14150) (T_NS_C 14151) (T_PRINT 14152) (T_PRIVATE 14153) (T_PROPERTY_C 14154) (T_PROTECTED 14155) (T_PUBLIC 14156) (T_READONLY 14157) (T_REQUIRE 14158) (T_REQUIRE_ONCE 14159) (T_RETURN 14160) (T_STATIC 14161) (T_STRING 14162) (T_SWITCH 14163) (T_THROW 14164) (T_TRAIT 14165) (T_TRAIT_C 14166) (T_TRY 14167) (T_UNSET 14168) (T_USE 14169) (T_VAR 14170) (T_VARIABLE 411) (T_WHILE 14171) (T_YIELD 14172) (identifier 14173) (member_name 14174) (reserved_non_modifiers 14175) (semi_reserved 14176) (simple_variable 14177) ("{" 14178)) 6091 (("(" 14181) (argument_list 14182)) 6092 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14179) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6093 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 14180)) 6094 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 14183) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 14184) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 14185) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6095 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 14188) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6096 (("," 1480) (possible_comma 14186)) 6097 ((")" 14187)) 6098 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (ampersand 14218) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14219) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6099 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14217) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6100 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14216) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6101 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14215) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6102 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14214) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6103 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14213) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6104 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14212) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6105 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14211) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6106 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14210) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6107 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14209) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6108 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14208) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6109 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14207) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6110 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14206) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6111 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 14205) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6112 (("\"" 1520) ("$" 14220) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 14221) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 14222) ("[" 1542) (array_object_dereferenceable 14223) (callable_expr 14224) (callable_variable 14225) (class_constant 1546) (class_name 14226) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 14227) (name 14228) (new_dereferenceable 1552) (simple_variable 14229) (static_member 14230) (variable 14231) (variable_class_name 14232)) 6113 (("$" 14220) (T_VARIABLE 14222) (simple_variable 14265) ("{" 14266)) 6114 (("(" 14236) (argument_list 14264)) 6115 ((T_NULLSAFE_OBJECT_OPERATOR 14250) (T_OBJECT_OPERATOR 14251) ("[" 14252)) 6116 (("(" 14236) (argument_list 14249)) 6117 ((T_PAAMAYIM_NEKUDOTAYIM 14245)) 6118 (("(" 14236) (argument_list 14244)) 6119 ((T_PAAMAYIM_NEKUDOTAYIM 14233)) 6120 (("$" 14220) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 14222) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 14234) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 14235) ("{" 1642)) 6121 (("(" 14236) (argument_list 14237)) 6122 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 14238) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 14239) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 14240) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6123 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 14243) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6124 (("," 1480) (possible_comma 14241)) 6125 ((")" 14242)) 6126 (("$" 14220) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 14222) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 14246) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 14247) ("{" 1658)) 6127 (("(" 14236) (argument_list 14248)) 6128 (("$" 14220) (T_STRING 14255) (T_VARIABLE 14222) (property_name 14262) (simple_variable 14257) ("{" 14258)) 6129 (("$" 14220) (T_STRING 14255) (T_VARIABLE 14222) (property_name 14256) (simple_variable 14257) ("{" 14258)) 6130 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 14253) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6131 (("]" 14254)) 6132 (("(" 14236) (argument_list 14261)) 6133 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14259) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6134 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 14260)) 6135 (("(" 14236) (argument_list 14263)) 6136 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14267) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6137 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 14268)) 6138 ((backup_doc_comment 14272)) 6139 (("(" 14273)) 6140 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 14274)) 6141 ((")" 14275)) 6142 ((T_USE 164) (lexical_vars 14276)) 6143 ((":" 166) (return_type 14277)) 6144 ((backup_fn_flags 14278)) 6145 (("{" 14279)) 6146 ((inner_statement_list 14280)) 6147 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 14281) ("~" 127)) 6148 ((backup_fn_flags 14282)) 6149 ((backup_doc_comment 14284)) 6150 (("(" 14285)) 6151 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 14286)) 6152 ((")" 14287)) 6153 ((":" 2249) (return_type 14288)) 6154 ((T_DOUBLE_ARROW 14289)) 6155 ((backup_fn_flags 14290)) 6156 ((backup_lex_pos 14291)) 6157 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14355) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6158 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15329) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6159 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 15327) (encaps_var 5237)) 6160 (("$" 14294) (T_VARIABLE 14342) (simple_variable 15323) ("{" 15324)) 6161 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15321) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6162 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15320) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6163 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15319) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6164 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15318) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6165 (("(" 15315)) 6166 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15314) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6167 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15313) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6168 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15312) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6169 (("\"" 1520) ("$" 14597) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 14598) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 14599) ("[" 1542) (array_object_dereferenceable 14600) (callable_expr 14601) (callable_variable 14602) (class_constant 1546) (class_name 14603) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 14604) (name 14605) (new_dereferenceable 1552) (simple_variable 14606) (static_member 14607) (variable 15311) (variable_class_name 14609)) 6170 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15310) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6171 (("(" 15307)) 6172 (("(" 15304)) 6173 (("(" 15295) (argument_list 15296) (ctor_arguments 15297)) 6174 (("\"" 1520) ("$" 14597) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 14598) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 14599) ("[" 1542) (array_object_dereferenceable 14600) (callable_expr 14601) (callable_variable 14602) (class_constant 1546) (class_name 14603) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 14604) (name 14605) (new_dereferenceable 1552) (simple_variable 14606) (static_member 14607) (variable 15294) (variable_class_name 14609)) 6175 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15293) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6176 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15292) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6177 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15291) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6178 (("(" 15287)) 6179 (("(" 15282)) 6180 (("(" 15276)) 6181 (("$" 14597) ("(" 15239) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 15240) (T_NAME_QUALIFIED 15241) (T_NAME_RELATIVE 15242) (T_READONLY 2931) (T_STATIC 15243) (T_STRING 15244) (T_VARIABLE 14599) (anonymous_class 15245) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 15246) (attribute 2937) (attributes 15247) (class_modifier 2939) (class_name 15248) (class_name_reference 15249) (name 15250) (new_variable 15251) (simple_variable 15252)) 6182 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15238) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6183 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15237) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6184 (("(" 14558) (argument_list 15236)) 6185 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15235) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6186 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15234) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6187 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 15229) (T_END_HEREDOC 15230) (T_VARIABLE 2896) (encaps_list 15231) (encaps_var 2898)) 6188 ((T_FN 35) (T_FUNCTION 129) (fn 14356) #'14357 (inline_function 15228)) 6189 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15227) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6190 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15226) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6191 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15225) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6192 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14767) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6193 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14703) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6194 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 14699) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 6195 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 14697) (encaps_list 2804) (encaps_var 2805)) 6196 ((T_NULLSAFE_OBJECT_OPERATOR 14683) (T_OBJECT_OPERATOR 14684) ("[" 14685)) 6197 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 14680) (attribute 2733) (fn 14356) #'14357 (inline_function 14681)) 6198 (("(" 14558) (argument_list 14679)) 6199 ((T_PAAMAYIM_NEKUDOTAYIM 14671)) 6200 (("%" 14372) ("*" 14373) ("+" 14374) ("-" 14375) ("." 14376) ("/" 14377) ("<" 14378) (">" 14379) ("?" 14380) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 14381) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 14382) (T_BOOLEAN_AND 14383) (T_BOOLEAN_OR 14384) (T_COALESCE 14385) (T_INSTANCEOF 14386) (T_IS_EQUAL 14387) (T_IS_GREATER_OR_EQUAL 14388) (T_IS_IDENTICAL 14389) (T_IS_NOT_EQUAL 14390) (T_IS_NOT_IDENTICAL 14391) (T_IS_SMALLER_OR_EQUAL 14392) (T_LOGICAL_AND 14393) (T_LOGICAL_OR 14394) (T_LOGICAL_XOR 14395) (T_POW 14396) (T_SL 14397) (T_SPACESHIP 14398) (T_SR 14399) ("^" 14400) (backup_fn_flags 14670) ("|" 14401)) 6201 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 14659)) 6202 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 14647)) 6203 (("(" 14558) (argument_list 14646)) 6204 (("=" 14566) (T_AND_EQUAL 14567) (T_COALESCE_EQUAL 14568) (T_CONCAT_EQUAL 14569) (T_DEC 14570) (T_DIV_EQUAL 14571) (T_INC 14572) (T_MINUS_EQUAL 14573) (T_MOD_EQUAL 14574) (T_MUL_EQUAL 14575) (T_OR_EQUAL 14576) (T_PLUS_EQUAL 14577) (T_POW_EQUAL 14578) (T_SL_EQUAL 14579) (T_SR_EQUAL 14580) (T_XOR_EQUAL 14581)) 6205 ((T_PAAMAYIM_NEKUDOTAYIM 14470)) 6206 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14371) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6207 (("%" 14372) ("*" 14373) ("+" 14374) ("-" 14375) ("." 14376) ("/" 14377) ("<" 14378) (">" 14379) ("?" 14380) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 14381) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 14382) (T_BOOLEAN_AND 14383) (T_BOOLEAN_OR 14384) (T_COALESCE 14385) (T_INSTANCEOF 14386) (T_IS_EQUAL 14387) (T_IS_GREATER_OR_EQUAL 14388) (T_IS_IDENTICAL 14389) (T_IS_NOT_EQUAL 14390) (T_IS_NOT_IDENTICAL 14391) (T_IS_SMALLER_OR_EQUAL 14392) (T_LOGICAL_AND 14393) (T_LOGICAL_OR 14394) (T_LOGICAL_XOR 14395) (T_POW 14396) (T_SL 14397) (T_SPACESHIP 14398) (T_SR 14399) ("^" 14400) ("|" 14401)) 6208 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14469) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6209 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14468) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6210 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14467) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6211 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14466) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6212 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14465) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6213 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14464) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6214 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14463) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6215 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14462) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6216 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 14457) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 14458) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 6217 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14456) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6218 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14455) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6219 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14454) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6220 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14453) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6221 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14452) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6222 (("$" 14417) ("(" 14418) (T_NAME_FULLY_QUALIFIED 14419) (T_NAME_QUALIFIED 14420) (T_NAME_RELATIVE 14421) (T_STATIC 14422) (T_STRING 14423) (T_VARIABLE 14424) (class_name 14425) (class_name_reference 14426) (name 14427) (new_variable 14428) (simple_variable 14429)) 6223 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14416) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6224 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14415) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6225 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14414) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6226 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14413) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6227 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14412) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6228 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14411) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6229 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14410) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6230 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14409) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6231 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14408) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6232 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14407) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6233 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14406) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6234 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14405) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6235 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14404) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6236 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14403) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6237 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14402) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6238 (("$" 14417) (T_VARIABLE 14424) (simple_variable 14448) ("{" 14449)) 6239 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 14446) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6240 ((T_PAAMAYIM_NEKUDOTAYIM 14444)) 6241 ((T_NULLSAFE_OBJECT_OPERATOR 14430) (T_OBJECT_OPERATOR 14431) (T_PAAMAYIM_NEKUDOTAYIM 14432) ("[" 14433)) 6242 (("$" 14417) (T_STRING 14437) (T_VARIABLE 14424) (property_name 14443) (simple_variable 14439) ("{" 14440)) 6243 (("$" 14417) (T_STRING 14437) (T_VARIABLE 14424) (property_name 14438) (simple_variable 14439) ("{" 14440)) 6244 (("$" 14417) (T_VARIABLE 14424) (simple_variable 14436)) 6245 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 14434) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6246 (("]" 14435)) 6247 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14441) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6248 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 14442)) 6249 (("$" 14417) (T_VARIABLE 14424) (simple_variable 14445)) 6250 (("%" 299) (")" 14447) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6251 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14450) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6252 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 14451)) 6253 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14461) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6254 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 14459) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 6255 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14460) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6256 (("$" 14294) (T_ABSTRACT 14471) (T_ARRAY 14472) (T_AS 14473) (T_BREAK 14474) (T_CALLABLE 14475) (T_CASE 14476) (T_CATCH 14477) (T_CLASS 14478) (T_CLASS_C 14479) (T_CLONE 14480) (T_CONST 14481) (T_CONTINUE 14482) (T_DECLARE 14483) (T_DEFAULT 14484) (T_DIR 14485) (T_DO 14486) (T_ECHO 14487) (T_ELSE 14488) (T_ELSEIF 14489) (T_EMPTY 14490) (T_ENDDECLARE 14491) (T_ENDFOR 14492) (T_ENDFOREACH 14493) (T_ENDIF 14494) (T_ENDSWITCH 14495) (T_ENDWHILE 14496) (T_ENUM 14497) (T_EVAL 14498) (T_EXIT 14499) (T_EXTENDS 14500) (T_FILE 14501) (T_FINAL 14502) (T_FINALLY 14503) (T_FN 14504) (T_FOR 14505) (T_FOREACH 14506) (T_FUNCTION 14507) (T_FUNC_C 14508) (T_GLOBAL 14509) (T_GOTO 14510) (T_IF 14511) (T_IMPLEMENTS 14512) (T_INCLUDE 14513) (T_INCLUDE_ONCE 14514) (T_INSTANCEOF 14515) (T_INSTEADOF 14516) (T_INTERFACE 14517) (T_ISSET 14518) (T_LINE 14519) (T_LIST 14520) (T_LOGICAL_AND 14521) (T_LOGICAL_OR 14522) (T_LOGICAL_XOR 14523) (T_MATCH 14524) (T_METHOD_C 14525) (T_NAMESPACE 14526) (T_NEW 14527) (T_NS_C 14528) (T_PRINT 14529) (T_PRIVATE 14530) (T_PROPERTY_C 14531) (T_PROTECTED 14532) (T_PUBLIC 14533) (T_READONLY 14534) (T_REQUIRE 14535) (T_REQUIRE_ONCE 14536) (T_RETURN 14537) (T_STATIC 14538) (T_STRING 14539) (T_SWITCH 14540) (T_THROW 14541) (T_TRAIT 14542) (T_TRAIT_C 14543) (T_TRY 14544) (T_UNSET 14545) (T_USE 14546) (T_VAR 14547) (T_VARIABLE 14342) (T_WHILE 14548) (T_YIELD 14549) (identifier 14550) (member_name 14551) (reserved_non_modifiers 14552) (semi_reserved 14553) (simple_variable 14554) ("{" 14555)) 6257 (("(" 14558) (argument_list 14559)) 6258 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14556) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6259 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 14557)) 6260 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 14560) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 14561) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 14562) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6261 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 14565) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6262 (("," 1480) (possible_comma 14563)) 6263 ((")" 14564)) 6264 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (ampersand 14595) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14596) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6265 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14594) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6266 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14593) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6267 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14592) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6268 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14591) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6269 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14590) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6270 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14589) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6271 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14588) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6272 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14587) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6273 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14586) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6274 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14585) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6275 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14584) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6276 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14583) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6277 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14582) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6278 (("\"" 1520) ("$" 14597) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 14598) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 14599) ("[" 1542) (array_object_dereferenceable 14600) (callable_expr 14601) (callable_variable 14602) (class_constant 1546) (class_name 14603) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 14604) (name 14605) (new_dereferenceable 1552) (simple_variable 14606) (static_member 14607) (variable 14608) (variable_class_name 14609)) 6279 (("$" 14597) (T_VARIABLE 14599) (simple_variable 14642) ("{" 14643)) 6280 (("(" 14613) (argument_list 14641)) 6281 ((T_NULLSAFE_OBJECT_OPERATOR 14627) (T_OBJECT_OPERATOR 14628) ("[" 14629)) 6282 (("(" 14613) (argument_list 14626)) 6283 ((T_PAAMAYIM_NEKUDOTAYIM 14622)) 6284 (("(" 14613) (argument_list 14621)) 6285 ((T_PAAMAYIM_NEKUDOTAYIM 14610)) 6286 (("$" 14597) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 14599) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 14611) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 14612) ("{" 1642)) 6287 (("(" 14613) (argument_list 14614)) 6288 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 14615) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 14616) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 14617) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6289 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 14620) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6290 (("," 1480) (possible_comma 14618)) 6291 ((")" 14619)) 6292 (("$" 14597) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 14599) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 14623) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 14624) ("{" 1658)) 6293 (("(" 14613) (argument_list 14625)) 6294 (("$" 14597) (T_STRING 14632) (T_VARIABLE 14599) (property_name 14639) (simple_variable 14634) ("{" 14635)) 6295 (("$" 14597) (T_STRING 14632) (T_VARIABLE 14599) (property_name 14633) (simple_variable 14634) ("{" 14635)) 6296 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 14630) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6297 (("]" 14631)) 6298 (("(" 14613) (argument_list 14638)) 6299 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14636) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6300 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 14637)) 6301 (("(" 14613) (argument_list 14640)) 6302 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14644) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6303 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 14645)) 6304 ((backup_doc_comment 14648)) 6305 (("(" 14649)) 6306 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 14650)) 6307 ((")" 14651)) 6308 ((T_USE 164) (lexical_vars 14652)) 6309 ((":" 166) (return_type 14653)) 6310 ((backup_fn_flags 14654)) 6311 (("{" 14655)) 6312 ((inner_statement_list 14656)) 6313 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 14657) ("~" 127)) 6314 ((backup_fn_flags 14658)) 6315 ((backup_doc_comment 14660)) 6316 (("(" 14661)) 6317 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 14662)) 6318 ((")" 14663)) 6319 ((":" 2249) (return_type 14664)) 6320 ((T_DOUBLE_ARROW 14665)) 6321 ((backup_fn_flags 14666)) 6322 ((backup_lex_pos 14667)) 6323 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14668) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6324 (("%" 14372) ("*" 14373) ("+" 14374) ("-" 14375) ("." 14376) ("/" 14377) ("<" 14378) (">" 14379) ("?" 14380) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 14381) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 14382) (T_BOOLEAN_AND 14383) (T_BOOLEAN_OR 14384) (T_COALESCE 14385) (T_INSTANCEOF 14386) (T_IS_EQUAL 14387) (T_IS_GREATER_OR_EQUAL 14388) (T_IS_IDENTICAL 14389) (T_IS_NOT_EQUAL 14390) (T_IS_NOT_IDENTICAL 14391) (T_IS_SMALLER_OR_EQUAL 14392) (T_LOGICAL_AND 14393) (T_LOGICAL_OR 14394) (T_LOGICAL_XOR 14395) (T_POW 14396) (T_SL 14397) (T_SPACESHIP 14398) (T_SR 14399) ("^" 14400) (backup_fn_flags 14669) ("|" 14401)) 6325 (("$" 14294) (T_ABSTRACT 14471) (T_ARRAY 14472) (T_AS 14473) (T_BREAK 14474) (T_CALLABLE 14475) (T_CASE 14476) (T_CATCH 14477) (T_CLASS 14478) (T_CLASS_C 14479) (T_CLONE 14480) (T_CONST 14481) (T_CONTINUE 14482) (T_DECLARE 14483) (T_DEFAULT 14484) (T_DIR 14485) (T_DO 14486) (T_ECHO 14487) (T_ELSE 14488) (T_ELSEIF 14489) (T_EMPTY 14490) (T_ENDDECLARE 14491) (T_ENDFOR 14492) (T_ENDFOREACH 14493) (T_ENDIF 14494) (T_ENDSWITCH 14495) (T_ENDWHILE 14496) (T_ENUM 14497) (T_EVAL 14498) (T_EXIT 14499) (T_EXTENDS 14500) (T_FILE 14501) (T_FINAL 14502) (T_FINALLY 14503) (T_FN 14504) (T_FOR 14505) (T_FOREACH 14506) (T_FUNCTION 14507) (T_FUNC_C 14508) (T_GLOBAL 14509) (T_GOTO 14510) (T_IF 14511) (T_IMPLEMENTS 14512) (T_INCLUDE 14513) (T_INCLUDE_ONCE 14514) (T_INSTANCEOF 14515) (T_INSTEADOF 14516) (T_INTERFACE 14517) (T_ISSET 14518) (T_LINE 14519) (T_LIST 14520) (T_LOGICAL_AND 14521) (T_LOGICAL_OR 14522) (T_LOGICAL_XOR 14523) (T_MATCH 14524) (T_METHOD_C 14525) (T_NAMESPACE 14526) (T_NEW 14527) (T_NS_C 14528) (T_PRINT 14529) (T_PRIVATE 14530) (T_PROPERTY_C 14531) (T_PROTECTED 14532) (T_PUBLIC 14533) (T_READONLY 14534) (T_REQUIRE 14535) (T_REQUIRE_ONCE 14536) (T_RETURN 14537) (T_STATIC 14538) (T_STRING 14539) (T_SWITCH 14540) (T_THROW 14541) (T_TRAIT 14542) (T_TRAIT_C 14543) (T_TRY 14544) (T_UNSET 14545) (T_USE 14546) (T_VAR 14547) (T_VARIABLE 14342) (T_WHILE 14548) (T_YIELD 14549) (identifier 14672) (member_name 14673) (reserved_non_modifiers 14552) (semi_reserved 14553) (simple_variable 14674) ("{" 14675)) 6326 (("(" 14558) (argument_list 14678)) 6327 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14676) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6328 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 14677)) 6329 ((T_FN 35) (T_FUNCTION 129) (fn 14356) #'14357 (inline_function 14682)) 6330 (("$" 14294) (T_STRING 14688) (T_VARIABLE 14342) (property_name 14695) (simple_variable 14690) ("{" 14691)) 6331 (("$" 14294) (T_STRING 14688) (T_VARIABLE 14342) (property_name 14689) (simple_variable 14690) ("{" 14691)) 6332 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 14686) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6333 (("]" 14687)) 6334 (("(" 14558) (argument_list 14694)) 6335 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14692) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6336 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 14693)) 6337 (("(" 14558) (argument_list 14696)) 6338 (("`" 14698)) 6339 (("]" 14700)) 6340 (("=" 14701)) 6341 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 14702) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6342 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15224) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6343 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 15222) (encaps_var 5237)) 6344 (("$" 14706) (T_VARIABLE 14754) (simple_variable 15218) ("{" 15219)) 6345 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15216) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6346 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15215) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6347 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15214) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6348 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15213) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6349 (("(" 15210)) 6350 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15209) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6351 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15208) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6352 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15207) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6353 (("\"" 1520) ("$" 15009) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 15010) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 15011) ("[" 1542) (array_object_dereferenceable 15012) (callable_expr 15013) (callable_variable 15014) (class_constant 1546) (class_name 15015) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 15016) (name 15017) (new_dereferenceable 1552) (simple_variable 15018) (static_member 15019) (variable 15206) (variable_class_name 15021)) 6354 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15205) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6355 (("(" 15202)) 6356 (("(" 15199)) 6357 (("(" 15190) (argument_list 15191) (ctor_arguments 15192)) 6358 (("\"" 1520) ("$" 15009) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 15010) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 15011) ("[" 1542) (array_object_dereferenceable 15012) (callable_expr 15013) (callable_variable 15014) (class_constant 1546) (class_name 15015) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 15016) (name 15017) (new_dereferenceable 1552) (simple_variable 15018) (static_member 15019) (variable 15189) (variable_class_name 15021)) 6359 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15188) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6360 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15187) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6361 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15186) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6362 (("(" 15182)) 6363 (("(" 15177)) 6364 (("(" 15171)) 6365 (("$" 15009) ("(" 15134) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 15135) (T_NAME_QUALIFIED 15136) (T_NAME_RELATIVE 15137) (T_READONLY 2931) (T_STATIC 15138) (T_STRING 15139) (T_VARIABLE 15011) (anonymous_class 15140) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 15141) (attribute 2937) (attributes 15142) (class_modifier 2939) (class_name 15143) (class_name_reference 15144) (name 15145) (new_variable 15146) (simple_variable 15147)) 6366 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15133) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6367 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15132) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6368 (("(" 14970) (argument_list 15131)) 6369 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15130) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6370 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15129) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6371 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 15124) (T_END_HEREDOC 15125) (T_VARIABLE 2896) (encaps_list 15126) (encaps_var 2898)) 6372 ((T_FN 35) (T_FUNCTION 129) (fn 14768) #'14769 (inline_function 15123)) 6373 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15122) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6374 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15121) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6375 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15120) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6376 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15117) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6377 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15116) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6378 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 15112) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 6379 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 15110) (encaps_list 2804) (encaps_var 2805)) 6380 ((T_NULLSAFE_OBJECT_OPERATOR 15096) (T_OBJECT_OPERATOR 15097) ("[" 15098)) 6381 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 15093) (attribute 2733) (fn 14768) #'14769 (inline_function 15094)) 6382 (("(" 14970) (argument_list 15092)) 6383 ((T_PAAMAYIM_NEKUDOTAYIM 15084)) 6384 (("%" 14784) ("*" 14785) ("+" 14786) ("-" 14787) ("." 14788) ("/" 14789) ("<" 14790) (">" 14791) ("?" 14792) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 14793) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 14794) (T_BOOLEAN_AND 14795) (T_BOOLEAN_OR 14796) (T_COALESCE 14797) (T_DOUBLE_ARROW 15082) (T_INSTANCEOF 14798) (T_IS_EQUAL 14799) (T_IS_GREATER_OR_EQUAL 14800) (T_IS_IDENTICAL 14801) (T_IS_NOT_EQUAL 14802) (T_IS_NOT_IDENTICAL 14803) (T_IS_SMALLER_OR_EQUAL 14804) (T_LOGICAL_AND 14805) (T_LOGICAL_OR 14806) (T_LOGICAL_XOR 14807) (T_POW 14808) (T_SL 14809) (T_SPACESHIP 14810) (T_SR 14811) ("^" 14812) ("|" 14813)) 6385 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 15071)) 6386 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 15059)) 6387 (("(" 14970) (argument_list 15058)) 6388 (("=" 14978) (T_AND_EQUAL 14979) (T_COALESCE_EQUAL 14980) (T_CONCAT_EQUAL 14981) (T_DEC 14982) (T_DIV_EQUAL 14983) (T_INC 14984) (T_MINUS_EQUAL 14985) (T_MOD_EQUAL 14986) (T_MUL_EQUAL 14987) (T_OR_EQUAL 14988) (T_PLUS_EQUAL 14989) (T_POW_EQUAL 14990) (T_SL_EQUAL 14991) (T_SR_EQUAL 14992) (T_XOR_EQUAL 14993)) 6389 ((T_PAAMAYIM_NEKUDOTAYIM 14882)) 6390 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14783) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6391 (("%" 14784) ("*" 14785) ("+" 14786) ("-" 14787) ("." 14788) ("/" 14789) ("<" 14790) (">" 14791) ("?" 14792) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 14793) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 14794) (T_BOOLEAN_AND 14795) (T_BOOLEAN_OR 14796) (T_COALESCE 14797) (T_INSTANCEOF 14798) (T_IS_EQUAL 14799) (T_IS_GREATER_OR_EQUAL 14800) (T_IS_IDENTICAL 14801) (T_IS_NOT_EQUAL 14802) (T_IS_NOT_IDENTICAL 14803) (T_IS_SMALLER_OR_EQUAL 14804) (T_LOGICAL_AND 14805) (T_LOGICAL_OR 14806) (T_LOGICAL_XOR 14807) (T_POW 14808) (T_SL 14809) (T_SPACESHIP 14810) (T_SR 14811) ("^" 14812) ("|" 14813)) 6392 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14881) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6393 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14880) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6394 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14879) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6395 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14878) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6396 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14877) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6397 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14876) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6398 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14875) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6399 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14874) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6400 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 14869) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 14870) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 6401 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14868) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6402 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14867) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6403 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14866) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6404 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14865) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6405 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14864) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6406 (("$" 14829) ("(" 14830) (T_NAME_FULLY_QUALIFIED 14831) (T_NAME_QUALIFIED 14832) (T_NAME_RELATIVE 14833) (T_STATIC 14834) (T_STRING 14835) (T_VARIABLE 14836) (class_name 14837) (class_name_reference 14838) (name 14839) (new_variable 14840) (simple_variable 14841)) 6407 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14828) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6408 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14827) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6409 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14826) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6410 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14825) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6411 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14824) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6412 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14823) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6413 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14822) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6414 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14821) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6415 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14820) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6416 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14819) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6417 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14818) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6418 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14817) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6419 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14816) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6420 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14815) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6421 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14814) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6422 (("$" 14829) (T_VARIABLE 14836) (simple_variable 14860) ("{" 14861)) 6423 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 14858) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6424 ((T_PAAMAYIM_NEKUDOTAYIM 14856)) 6425 ((T_NULLSAFE_OBJECT_OPERATOR 14842) (T_OBJECT_OPERATOR 14843) (T_PAAMAYIM_NEKUDOTAYIM 14844) ("[" 14845)) 6426 (("$" 14829) (T_STRING 14849) (T_VARIABLE 14836) (property_name 14855) (simple_variable 14851) ("{" 14852)) 6427 (("$" 14829) (T_STRING 14849) (T_VARIABLE 14836) (property_name 14850) (simple_variable 14851) ("{" 14852)) 6428 (("$" 14829) (T_VARIABLE 14836) (simple_variable 14848)) 6429 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 14846) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6430 (("]" 14847)) 6431 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14853) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6432 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 14854)) 6433 (("$" 14829) (T_VARIABLE 14836) (simple_variable 14857)) 6434 (("%" 299) (")" 14859) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6435 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14862) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6436 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 14863)) 6437 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14873) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6438 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 14871) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 6439 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14872) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6440 (("$" 14706) (T_ABSTRACT 14883) (T_ARRAY 14884) (T_AS 14885) (T_BREAK 14886) (T_CALLABLE 14887) (T_CASE 14888) (T_CATCH 14889) (T_CLASS 14890) (T_CLASS_C 14891) (T_CLONE 14892) (T_CONST 14893) (T_CONTINUE 14894) (T_DECLARE 14895) (T_DEFAULT 14896) (T_DIR 14897) (T_DO 14898) (T_ECHO 14899) (T_ELSE 14900) (T_ELSEIF 14901) (T_EMPTY 14902) (T_ENDDECLARE 14903) (T_ENDFOR 14904) (T_ENDFOREACH 14905) (T_ENDIF 14906) (T_ENDSWITCH 14907) (T_ENDWHILE 14908) (T_ENUM 14909) (T_EVAL 14910) (T_EXIT 14911) (T_EXTENDS 14912) (T_FILE 14913) (T_FINAL 14914) (T_FINALLY 14915) (T_FN 14916) (T_FOR 14917) (T_FOREACH 14918) (T_FUNCTION 14919) (T_FUNC_C 14920) (T_GLOBAL 14921) (T_GOTO 14922) (T_IF 14923) (T_IMPLEMENTS 14924) (T_INCLUDE 14925) (T_INCLUDE_ONCE 14926) (T_INSTANCEOF 14927) (T_INSTEADOF 14928) (T_INTERFACE 14929) (T_ISSET 14930) (T_LINE 14931) (T_LIST 14932) (T_LOGICAL_AND 14933) (T_LOGICAL_OR 14934) (T_LOGICAL_XOR 14935) (T_MATCH 14936) (T_METHOD_C 14937) (T_NAMESPACE 14938) (T_NEW 14939) (T_NS_C 14940) (T_PRINT 14941) (T_PRIVATE 14942) (T_PROPERTY_C 14943) (T_PROTECTED 14944) (T_PUBLIC 14945) (T_READONLY 14946) (T_REQUIRE 14947) (T_REQUIRE_ONCE 14948) (T_RETURN 14949) (T_STATIC 14950) (T_STRING 14951) (T_SWITCH 14952) (T_THROW 14953) (T_TRAIT 14954) (T_TRAIT_C 14955) (T_TRY 14956) (T_UNSET 14957) (T_USE 14958) (T_VAR 14959) (T_VARIABLE 14754) (T_WHILE 14960) (T_YIELD 14961) (identifier 14962) (member_name 14963) (reserved_non_modifiers 14964) (semi_reserved 14965) (simple_variable 14966) ("{" 14967)) 6441 (("(" 14970) (argument_list 14971)) 6442 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 14968) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6443 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 14969)) 6444 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 14972) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 14973) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 14974) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6445 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 14977) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6446 (("," 1480) (possible_comma 14975)) 6447 ((")" 14976)) 6448 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (ampersand 15007) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15008) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6449 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15006) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6450 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15005) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6451 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15004) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6452 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15003) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6453 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15002) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6454 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15001) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6455 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15000) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6456 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14999) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6457 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14998) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6458 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14997) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6459 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14996) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6460 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14995) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6461 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 14994) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6462 (("\"" 1520) ("$" 15009) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 15010) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 15011) ("[" 1542) (array_object_dereferenceable 15012) (callable_expr 15013) (callable_variable 15014) (class_constant 1546) (class_name 15015) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 15016) (name 15017) (new_dereferenceable 1552) (simple_variable 15018) (static_member 15019) (variable 15020) (variable_class_name 15021)) 6463 (("$" 15009) (T_VARIABLE 15011) (simple_variable 15054) ("{" 15055)) 6464 (("(" 15025) (argument_list 15053)) 6465 ((T_NULLSAFE_OBJECT_OPERATOR 15039) (T_OBJECT_OPERATOR 15040) ("[" 15041)) 6466 (("(" 15025) (argument_list 15038)) 6467 ((T_PAAMAYIM_NEKUDOTAYIM 15034)) 6468 (("(" 15025) (argument_list 15033)) 6469 ((T_PAAMAYIM_NEKUDOTAYIM 15022)) 6470 (("$" 15009) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 15011) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 15023) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 15024) ("{" 1642)) 6471 (("(" 15025) (argument_list 15026)) 6472 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 15027) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 15028) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 15029) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6473 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 15032) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6474 (("," 1480) (possible_comma 15030)) 6475 ((")" 15031)) 6476 (("$" 15009) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 15011) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 15035) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 15036) ("{" 1658)) 6477 (("(" 15025) (argument_list 15037)) 6478 (("$" 15009) (T_STRING 15044) (T_VARIABLE 15011) (property_name 15051) (simple_variable 15046) ("{" 15047)) 6479 (("$" 15009) (T_STRING 15044) (T_VARIABLE 15011) (property_name 15045) (simple_variable 15046) ("{" 15047)) 6480 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 15042) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6481 (("]" 15043)) 6482 (("(" 15025) (argument_list 15050)) 6483 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15048) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6484 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15049)) 6485 (("(" 15025) (argument_list 15052)) 6486 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15056) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6487 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15057)) 6488 ((backup_doc_comment 15060)) 6489 (("(" 15061)) 6490 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 15062)) 6491 ((")" 15063)) 6492 ((T_USE 164) (lexical_vars 15064)) 6493 ((":" 166) (return_type 15065)) 6494 ((backup_fn_flags 15066)) 6495 (("{" 15067)) 6496 ((inner_statement_list 15068)) 6497 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 15069) ("~" 127)) 6498 ((backup_fn_flags 15070)) 6499 ((backup_doc_comment 15072)) 6500 (("(" 15073)) 6501 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 15074)) 6502 ((")" 15075)) 6503 ((":" 2249) (return_type 15076)) 6504 ((T_DOUBLE_ARROW 15077)) 6505 ((backup_fn_flags 15078)) 6506 ((backup_lex_pos 15079)) 6507 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15080) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6508 (("%" 14784) ("*" 14785) ("+" 14786) ("-" 14787) ("." 14788) ("/" 14789) ("<" 14790) (">" 14791) ("?" 14792) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 14793) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 14794) (T_BOOLEAN_AND 14795) (T_BOOLEAN_OR 14796) (T_COALESCE 14797) (T_INSTANCEOF 14798) (T_IS_EQUAL 14799) (T_IS_GREATER_OR_EQUAL 14800) (T_IS_IDENTICAL 14801) (T_IS_NOT_EQUAL 14802) (T_IS_NOT_IDENTICAL 14803) (T_IS_SMALLER_OR_EQUAL 14804) (T_LOGICAL_AND 14805) (T_LOGICAL_OR 14806) (T_LOGICAL_XOR 14807) (T_POW 14808) (T_SL 14809) (T_SPACESHIP 14810) (T_SR 14811) ("^" 14812) (backup_fn_flags 15081) ("|" 14813)) 6509 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15083) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6510 (("$" 14706) (T_ABSTRACT 14883) (T_ARRAY 14884) (T_AS 14885) (T_BREAK 14886) (T_CALLABLE 14887) (T_CASE 14888) (T_CATCH 14889) (T_CLASS 14890) (T_CLASS_C 14891) (T_CLONE 14892) (T_CONST 14893) (T_CONTINUE 14894) (T_DECLARE 14895) (T_DEFAULT 14896) (T_DIR 14897) (T_DO 14898) (T_ECHO 14899) (T_ELSE 14900) (T_ELSEIF 14901) (T_EMPTY 14902) (T_ENDDECLARE 14903) (T_ENDFOR 14904) (T_ENDFOREACH 14905) (T_ENDIF 14906) (T_ENDSWITCH 14907) (T_ENDWHILE 14908) (T_ENUM 14909) (T_EVAL 14910) (T_EXIT 14911) (T_EXTENDS 14912) (T_FILE 14913) (T_FINAL 14914) (T_FINALLY 14915) (T_FN 14916) (T_FOR 14917) (T_FOREACH 14918) (T_FUNCTION 14919) (T_FUNC_C 14920) (T_GLOBAL 14921) (T_GOTO 14922) (T_IF 14923) (T_IMPLEMENTS 14924) (T_INCLUDE 14925) (T_INCLUDE_ONCE 14926) (T_INSTANCEOF 14927) (T_INSTEADOF 14928) (T_INTERFACE 14929) (T_ISSET 14930) (T_LINE 14931) (T_LIST 14932) (T_LOGICAL_AND 14933) (T_LOGICAL_OR 14934) (T_LOGICAL_XOR 14935) (T_MATCH 14936) (T_METHOD_C 14937) (T_NAMESPACE 14938) (T_NEW 14939) (T_NS_C 14940) (T_PRINT 14941) (T_PRIVATE 14942) (T_PROPERTY_C 14943) (T_PROTECTED 14944) (T_PUBLIC 14945) (T_READONLY 14946) (T_REQUIRE 14947) (T_REQUIRE_ONCE 14948) (T_RETURN 14949) (T_STATIC 14950) (T_STRING 14951) (T_SWITCH 14952) (T_THROW 14953) (T_TRAIT 14954) (T_TRAIT_C 14955) (T_TRY 14956) (T_UNSET 14957) (T_USE 14958) (T_VAR 14959) (T_VARIABLE 14754) (T_WHILE 14960) (T_YIELD 14961) (identifier 15085) (member_name 15086) (reserved_non_modifiers 14964) (semi_reserved 14965) (simple_variable 15087) ("{" 15088)) 6511 (("(" 14970) (argument_list 15091)) 6512 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15089) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6513 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15090)) 6514 ((T_FN 35) (T_FUNCTION 129) (fn 14768) #'14769 (inline_function 15095)) 6515 (("$" 14706) (T_STRING 15101) (T_VARIABLE 14754) (property_name 15108) (simple_variable 15103) ("{" 15104)) 6516 (("$" 14706) (T_STRING 15101) (T_VARIABLE 14754) (property_name 15102) (simple_variable 15103) ("{" 15104)) 6517 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 15099) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6518 (("]" 15100)) 6519 (("(" 14970) (argument_list 15107)) 6520 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15105) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6521 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15106)) 6522 (("(" 14970) (argument_list 15109)) 6523 (("`" 15111)) 6524 (("]" 15113)) 6525 (("=" 15114)) 6526 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15115) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6527 (("%" 14784) ("*" 14785) ("+" 14786) ("-" 14787) ("." 14788) ("/" 14789) ("<" 14790) (">" 14791) ("?" 14792) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 14793) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 14794) (T_BOOLEAN_AND 14795) (T_BOOLEAN_OR 14796) (T_COALESCE 14797) (T_DOUBLE_ARROW 15118) (T_INSTANCEOF 14798) (T_IS_EQUAL 14799) (T_IS_GREATER_OR_EQUAL 14800) (T_IS_IDENTICAL 14801) (T_IS_NOT_EQUAL 14802) (T_IS_NOT_IDENTICAL 14803) (T_IS_SMALLER_OR_EQUAL 14804) (T_LOGICAL_AND 14805) (T_LOGICAL_OR 14806) (T_LOGICAL_XOR 14807) (T_POW 14808) (T_SL 14809) (T_SPACESHIP 14810) (T_SR 14811) ("^" 14812) ("|" 14813)) 6528 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15119) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6529 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 15128) (T_VARIABLE 2896) (encaps_var 2910)) 6530 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 15127) (T_VARIABLE 2896) (encaps_var 2901)) 6531 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15169) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6532 ((T_CLASS 15161)) 6533 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 15160) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 15141) (attribute 2958) (class_modifier 2939)) 6534 ((T_PAAMAYIM_NEKUDOTAYIM 15158)) 6535 (("(" 15025) (argument_list 15157)) 6536 ((T_NULLSAFE_OBJECT_OPERATOR 15148) (T_OBJECT_OPERATOR 15149) (T_PAAMAYIM_NEKUDOTAYIM 15150) ("[" 15151)) 6537 (("$" 15009) (T_STRING 15044) (T_VARIABLE 15011) (property_name 15156) (simple_variable 15046) ("{" 15047)) 6538 (("$" 15009) (T_STRING 15044) (T_VARIABLE 15011) (property_name 15155) (simple_variable 15046) ("{" 15047)) 6539 (("$" 15009) (T_VARIABLE 15011) (simple_variable 15154)) 6540 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 15152) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6541 (("]" 15153)) 6542 (("$" 15009) (T_VARIABLE 15011) (simple_variable 15159)) 6543 (("(" 2960) (argument_list 2961) (ctor_arguments 15162)) 6544 ((T_EXTENDS 2963) (extends_from 15163)) 6545 ((T_IMPLEMENTS 2965) (implements_list 15164)) 6546 ((backup_doc_comment 15165)) 6547 (("{" 15166)) 6548 ((class_statement_list 15167)) 6549 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 15168)) 6550 (("%" 299) (")" 15170) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6551 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15172) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6552 (("%" 299) (")" 15173) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6553 (("{" 15174)) 6554 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 15175) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 6555 (("}" 15176)) 6556 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 15178) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 6557 ((")" 15179)) 6558 (("=" 15180)) 6559 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15181) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6560 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 15183) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6561 (("," 5192) (possible_comma 15184)) 6562 ((")" 15185)) 6563 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 15193) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 15194) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 15195) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6564 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 15198) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6565 (("," 1480) (possible_comma 15196)) 6566 ((")" 15197)) 6567 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15200) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6568 (("%" 299) (")" 15201) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6569 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15203) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6570 (("%" 299) (")" 15204) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6571 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 15211) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 6572 ((")" 15212)) 6573 (("%" 299) (")" 15217) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6574 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15220) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6575 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15221)) 6576 (("\"" 15223) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 6577 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 15233) (T_VARIABLE 2896) (encaps_var 2910)) 6578 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 15232) (T_VARIABLE 2896) (encaps_var 2901)) 6579 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15274) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6580 ((T_CLASS 15266)) 6581 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 15265) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 15246) (attribute 2958) (class_modifier 2939)) 6582 ((T_PAAMAYIM_NEKUDOTAYIM 15263)) 6583 (("(" 14613) (argument_list 15262)) 6584 ((T_NULLSAFE_OBJECT_OPERATOR 15253) (T_OBJECT_OPERATOR 15254) (T_PAAMAYIM_NEKUDOTAYIM 15255) ("[" 15256)) 6585 (("$" 14597) (T_STRING 14632) (T_VARIABLE 14599) (property_name 15261) (simple_variable 14634) ("{" 14635)) 6586 (("$" 14597) (T_STRING 14632) (T_VARIABLE 14599) (property_name 15260) (simple_variable 14634) ("{" 14635)) 6587 (("$" 14597) (T_VARIABLE 14599) (simple_variable 15259)) 6588 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 15257) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6589 (("]" 15258)) 6590 (("$" 14597) (T_VARIABLE 14599) (simple_variable 15264)) 6591 (("(" 2960) (argument_list 2961) (ctor_arguments 15267)) 6592 ((T_EXTENDS 2963) (extends_from 15268)) 6593 ((T_IMPLEMENTS 2965) (implements_list 15269)) 6594 ((backup_doc_comment 15270)) 6595 (("{" 15271)) 6596 ((class_statement_list 15272)) 6597 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 15273)) 6598 (("%" 299) (")" 15275) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6599 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15277) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6600 (("%" 299) (")" 15278) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6601 (("{" 15279)) 6602 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 15280) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 6603 (("}" 15281)) 6604 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 15283) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 6605 ((")" 15284)) 6606 (("=" 15285)) 6607 (("!" 14292) ("\"" 14293) ("$" 14294) ("(" 14295) ("+" 14296) ("-" 14297) ("@" 14298) (T_ARRAY 14299) (T_ARRAY_CAST 14300) (T_ATTRIBUTE 128) (T_BOOL_CAST 14301) (T_CLASS_C 14302) (T_CLONE 14303) (T_CONSTANT_ENCAPSED_STRING 14304) (T_DEC 14305) (T_DIR 14306) (T_DNUMBER 14307) (T_DOUBLE_CAST 14308) (T_EMPTY 14309) (T_EVAL 14310) (T_EXIT 14311) (T_FILE 14312) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14313) (T_INC 14314) (T_INCLUDE 14315) (T_INCLUDE_ONCE 14316) (T_INT_CAST 14317) (T_ISSET 14318) (T_LINE 14319) (T_LIST 14320) (T_LNUMBER 14321) (T_MATCH 14322) (T_METHOD_C 14323) (T_NAME_FULLY_QUALIFIED 14324) (T_NAME_QUALIFIED 14325) (T_NAME_RELATIVE 14326) (T_NEW 14327) (T_NS_C 14328) (T_OBJECT_CAST 14329) (T_PRINT 14330) (T_PROPERTY_C 14331) (T_READONLY 14332) (T_REQUIRE 14333) (T_REQUIRE_ONCE 14334) (T_START_HEREDOC 14335) (T_STATIC 14336) (T_STRING 14337) (T_STRING_CAST 14338) (T_THROW 14339) (T_TRAIT_C 14340) (T_UNSET_CAST 14341) (T_VARIABLE 14342) (T_YIELD 14343) (T_YIELD_FROM 14344) ("[" 14345) ("`" 14346) (array_object_dereferenceable 14347) (attribute 133) (attributes 14348) (callable_expr 14349) (callable_variable 14350) (class_constant 14351) (class_name 14352) (constant 14353) (dereferenceable_scalar 14354) (expr 15286) (fn 14356) (fully_dereferenceable 105) #'14357 (function_call 14358) (inline_function 14359) (internal_functions_in_yacc 14360) (match 14361) (name 14362) (new_dereferenceable 14363) (new_non_dereferenceable 14364) (scalar 14365) (simple_variable 14366) (static_member 14367) (variable 14368) (variable_class_name 14369) ("~" 14370)) 6608 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 15288) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6609 (("," 5192) (possible_comma 15289)) 6610 ((")" 15290)) 6611 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 15298) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 15299) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 15300) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6612 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 15303) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6613 (("," 1480) (possible_comma 15301)) 6614 ((")" 15302)) 6615 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15305) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6616 (("%" 299) (")" 15306) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6617 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15308) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6618 (("%" 299) (")" 15309) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6619 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 15316) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 6620 ((")" 15317)) 6621 (("%" 299) (")" 15322) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6622 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15325) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6623 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15326)) 6624 (("\"" 15328) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 6625 (("$" 363) (T_ABSTRACT 14094) (T_ARRAY 14095) (T_AS 14096) (T_BREAK 14097) (T_CALLABLE 14098) (T_CASE 14099) (T_CATCH 14100) (T_CLASS 14101) (T_CLASS_C 14102) (T_CLONE 14103) (T_CONST 14104) (T_CONTINUE 14105) (T_DECLARE 14106) (T_DEFAULT 14107) (T_DIR 14108) (T_DO 14109) (T_ECHO 14110) (T_ELSE 14111) (T_ELSEIF 14112) (T_EMPTY 14113) (T_ENDDECLARE 14114) (T_ENDFOR 14115) (T_ENDFOREACH 14116) (T_ENDIF 14117) (T_ENDSWITCH 14118) (T_ENDWHILE 14119) (T_ENUM 14120) (T_EVAL 14121) (T_EXIT 14122) (T_EXTENDS 14123) (T_FILE 14124) (T_FINAL 14125) (T_FINALLY 14126) (T_FN 14127) (T_FOR 14128) (T_FOREACH 14129) (T_FUNCTION 14130) (T_FUNC_C 14131) (T_GLOBAL 14132) (T_GOTO 14133) (T_IF 14134) (T_IMPLEMENTS 14135) (T_INCLUDE 14136) (T_INCLUDE_ONCE 14137) (T_INSTANCEOF 14138) (T_INSTEADOF 14139) (T_INTERFACE 14140) (T_ISSET 14141) (T_LINE 14142) (T_LIST 14143) (T_LOGICAL_AND 14144) (T_LOGICAL_OR 14145) (T_LOGICAL_XOR 14146) (T_MATCH 14147) (T_METHOD_C 14148) (T_NAMESPACE 14149) (T_NEW 14150) (T_NS_C 14151) (T_PRINT 14152) (T_PRIVATE 14153) (T_PROPERTY_C 14154) (T_PROTECTED 14155) (T_PUBLIC 14156) (T_READONLY 14157) (T_REQUIRE 14158) (T_REQUIRE_ONCE 14159) (T_RETURN 14160) (T_STATIC 14161) (T_STRING 14162) (T_SWITCH 14163) (T_THROW 14164) (T_TRAIT 14165) (T_TRAIT_C 14166) (T_TRY 14167) (T_UNSET 14168) (T_USE 14169) (T_VAR 14170) (T_VARIABLE 411) (T_WHILE 14171) (T_YIELD 14172) (identifier 15331) (member_name 15332) (reserved_non_modifiers 14175) (semi_reserved 14176) (simple_variable 15333) ("{" 15334)) 6626 (("(" 14181) (argument_list 15337)) 6627 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15335) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6628 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15336)) 6629 ((T_FN 35) (T_FUNCTION 129) (fn 425) #'426 (inline_function 15341)) 6630 (("$" 363) (T_STRING 15347) (T_VARIABLE 411) (property_name 15354) (simple_variable 15349) ("{" 15350)) 6631 (("$" 363) (T_STRING 15347) (T_VARIABLE 411) (property_name 15348) (simple_variable 15349) ("{" 15350)) 6632 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 15345) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6633 (("]" 15346)) 6634 (("(" 14181) (argument_list 15353)) 6635 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15351) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6636 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15352)) 6637 (("(" 14181) (argument_list 15355)) 6638 (("`" 15357)) 6639 (("]" 15359)) 6640 (("=" 15360)) 6641 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15361) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6642 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15883) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6643 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 15881) (encaps_var 5237)) 6644 (("$" 15365) (T_VARIABLE 15413) (simple_variable 15877) ("{" 15878)) 6645 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15875) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6646 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15874) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6647 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15873) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6648 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15872) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6649 (("(" 15869)) 6650 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15868) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6651 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15867) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6652 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15866) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6653 (("\"" 1520) ("$" 15668) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 15669) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 15670) ("[" 1542) (array_object_dereferenceable 15671) (callable_expr 15672) (callable_variable 15673) (class_constant 1546) (class_name 15674) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 15675) (name 15676) (new_dereferenceable 1552) (simple_variable 15677) (static_member 15678) (variable 15865) (variable_class_name 15680)) 6654 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15864) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6655 (("(" 15861)) 6656 (("(" 15858)) 6657 (("(" 15849) (argument_list 15850) (ctor_arguments 15851)) 6658 (("\"" 1520) ("$" 15668) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 15669) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 15670) ("[" 1542) (array_object_dereferenceable 15671) (callable_expr 15672) (callable_variable 15673) (class_constant 1546) (class_name 15674) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 15675) (name 15676) (new_dereferenceable 1552) (simple_variable 15677) (static_member 15678) (variable 15848) (variable_class_name 15680)) 6659 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15847) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6660 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15846) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6661 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15845) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6662 (("(" 15841)) 6663 (("(" 15836)) 6664 (("(" 15830)) 6665 (("$" 15668) ("(" 15793) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 15794) (T_NAME_QUALIFIED 15795) (T_NAME_RELATIVE 15796) (T_READONLY 2931) (T_STATIC 15797) (T_STRING 15798) (T_VARIABLE 15670) (anonymous_class 15799) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 15800) (attribute 2937) (attributes 15801) (class_modifier 2939) (class_name 15802) (class_name_reference 15803) (name 15804) (new_variable 15805) (simple_variable 15806)) 6666 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15792) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6667 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15791) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6668 (("(" 15629) (argument_list 15790)) 6669 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15789) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6670 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15788) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6671 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 15783) (T_END_HEREDOC 15784) (T_VARIABLE 2896) (encaps_list 15785) (encaps_var 2898)) 6672 ((T_FN 35) (T_FUNCTION 129) (fn 15427) #'15428 (inline_function 15782)) 6673 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15781) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6674 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15780) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6675 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15779) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6676 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15776) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6677 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15775) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6678 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 15771) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 6679 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 15769) (encaps_list 2804) (encaps_var 2805)) 6680 ((T_NULLSAFE_OBJECT_OPERATOR 15755) (T_OBJECT_OPERATOR 15756) ("[" 15757)) 6681 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 15752) (attribute 2733) (fn 15427) #'15428 (inline_function 15753)) 6682 (("(" 15629) (argument_list 15751)) 6683 ((T_PAAMAYIM_NEKUDOTAYIM 15743)) 6684 (("%" 15443) ("*" 15444) ("+" 15445) ("-" 15446) ("." 15447) ("/" 15448) ("<" 15449) (">" 15450) ("?" 15451) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15452) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15453) (T_BOOLEAN_AND 15454) (T_BOOLEAN_OR 15455) (T_COALESCE 15456) (T_DOUBLE_ARROW 15741) (T_INSTANCEOF 15457) (T_IS_EQUAL 15458) (T_IS_GREATER_OR_EQUAL 15459) (T_IS_IDENTICAL 15460) (T_IS_NOT_EQUAL 15461) (T_IS_NOT_IDENTICAL 15462) (T_IS_SMALLER_OR_EQUAL 15463) (T_LOGICAL_AND 15464) (T_LOGICAL_OR 15465) (T_LOGICAL_XOR 15466) (T_POW 15467) (T_SL 15468) (T_SPACESHIP 15469) (T_SR 15470) ("^" 15471) ("|" 15472)) 6685 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 15730)) 6686 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 15718)) 6687 (("(" 15629) (argument_list 15717)) 6688 (("=" 15637) (T_AND_EQUAL 15638) (T_COALESCE_EQUAL 15639) (T_CONCAT_EQUAL 15640) (T_DEC 15641) (T_DIV_EQUAL 15642) (T_INC 15643) (T_MINUS_EQUAL 15644) (T_MOD_EQUAL 15645) (T_MUL_EQUAL 15646) (T_OR_EQUAL 15647) (T_PLUS_EQUAL 15648) (T_POW_EQUAL 15649) (T_SL_EQUAL 15650) (T_SR_EQUAL 15651) (T_XOR_EQUAL 15652)) 6689 ((T_PAAMAYIM_NEKUDOTAYIM 15541)) 6690 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15442) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6691 (("%" 15443) ("*" 15444) ("+" 15445) ("-" 15446) ("." 15447) ("/" 15448) ("<" 15449) (">" 15450) ("?" 15451) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15452) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15453) (T_BOOLEAN_AND 15454) (T_BOOLEAN_OR 15455) (T_COALESCE 15456) (T_INSTANCEOF 15457) (T_IS_EQUAL 15458) (T_IS_GREATER_OR_EQUAL 15459) (T_IS_IDENTICAL 15460) (T_IS_NOT_EQUAL 15461) (T_IS_NOT_IDENTICAL 15462) (T_IS_SMALLER_OR_EQUAL 15463) (T_LOGICAL_AND 15464) (T_LOGICAL_OR 15465) (T_LOGICAL_XOR 15466) (T_POW 15467) (T_SL 15468) (T_SPACESHIP 15469) (T_SR 15470) ("^" 15471) ("|" 15472)) 6692 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15540) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6693 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15539) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6694 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15538) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6695 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15537) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6696 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15536) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6697 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15535) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6698 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15534) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6699 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15533) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6700 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 15528) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 15529) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 6701 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15527) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6702 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15526) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6703 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15525) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6704 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15524) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6705 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15523) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6706 (("$" 15488) ("(" 15489) (T_NAME_FULLY_QUALIFIED 15490) (T_NAME_QUALIFIED 15491) (T_NAME_RELATIVE 15492) (T_STATIC 15493) (T_STRING 15494) (T_VARIABLE 15495) (class_name 15496) (class_name_reference 15497) (name 15498) (new_variable 15499) (simple_variable 15500)) 6707 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15487) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6708 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15486) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6709 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15485) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6710 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15484) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6711 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15483) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6712 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15482) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6713 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15481) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6714 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15480) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6715 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15479) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6716 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15478) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6717 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15477) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6718 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15476) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6719 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15475) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6720 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15474) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6721 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15473) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6722 (("$" 15488) (T_VARIABLE 15495) (simple_variable 15519) ("{" 15520)) 6723 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15517) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6724 ((T_PAAMAYIM_NEKUDOTAYIM 15515)) 6725 ((T_NULLSAFE_OBJECT_OPERATOR 15501) (T_OBJECT_OPERATOR 15502) (T_PAAMAYIM_NEKUDOTAYIM 15503) ("[" 15504)) 6726 (("$" 15488) (T_STRING 15508) (T_VARIABLE 15495) (property_name 15514) (simple_variable 15510) ("{" 15511)) 6727 (("$" 15488) (T_STRING 15508) (T_VARIABLE 15495) (property_name 15509) (simple_variable 15510) ("{" 15511)) 6728 (("$" 15488) (T_VARIABLE 15495) (simple_variable 15507)) 6729 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 15505) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6730 (("]" 15506)) 6731 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15512) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6732 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15513)) 6733 (("$" 15488) (T_VARIABLE 15495) (simple_variable 15516)) 6734 (("%" 299) (")" 15518) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6735 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15521) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6736 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15522)) 6737 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15532) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6738 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 15530) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 6739 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15531) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6740 (("$" 15365) (T_ABSTRACT 15542) (T_ARRAY 15543) (T_AS 15544) (T_BREAK 15545) (T_CALLABLE 15546) (T_CASE 15547) (T_CATCH 15548) (T_CLASS 15549) (T_CLASS_C 15550) (T_CLONE 15551) (T_CONST 15552) (T_CONTINUE 15553) (T_DECLARE 15554) (T_DEFAULT 15555) (T_DIR 15556) (T_DO 15557) (T_ECHO 15558) (T_ELSE 15559) (T_ELSEIF 15560) (T_EMPTY 15561) (T_ENDDECLARE 15562) (T_ENDFOR 15563) (T_ENDFOREACH 15564) (T_ENDIF 15565) (T_ENDSWITCH 15566) (T_ENDWHILE 15567) (T_ENUM 15568) (T_EVAL 15569) (T_EXIT 15570) (T_EXTENDS 15571) (T_FILE 15572) (T_FINAL 15573) (T_FINALLY 15574) (T_FN 15575) (T_FOR 15576) (T_FOREACH 15577) (T_FUNCTION 15578) (T_FUNC_C 15579) (T_GLOBAL 15580) (T_GOTO 15581) (T_IF 15582) (T_IMPLEMENTS 15583) (T_INCLUDE 15584) (T_INCLUDE_ONCE 15585) (T_INSTANCEOF 15586) (T_INSTEADOF 15587) (T_INTERFACE 15588) (T_ISSET 15589) (T_LINE 15590) (T_LIST 15591) (T_LOGICAL_AND 15592) (T_LOGICAL_OR 15593) (T_LOGICAL_XOR 15594) (T_MATCH 15595) (T_METHOD_C 15596) (T_NAMESPACE 15597) (T_NEW 15598) (T_NS_C 15599) (T_PRINT 15600) (T_PRIVATE 15601) (T_PROPERTY_C 15602) (T_PROTECTED 15603) (T_PUBLIC 15604) (T_READONLY 15605) (T_REQUIRE 15606) (T_REQUIRE_ONCE 15607) (T_RETURN 15608) (T_STATIC 15609) (T_STRING 15610) (T_SWITCH 15611) (T_THROW 15612) (T_TRAIT 15613) (T_TRAIT_C 15614) (T_TRY 15615) (T_UNSET 15616) (T_USE 15617) (T_VAR 15618) (T_VARIABLE 15413) (T_WHILE 15619) (T_YIELD 15620) (identifier 15621) (member_name 15622) (reserved_non_modifiers 15623) (semi_reserved 15624) (simple_variable 15625) ("{" 15626)) 6741 (("(" 15629) (argument_list 15630)) 6742 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15627) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6743 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15628)) 6744 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 15631) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 15632) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 15633) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6745 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 15636) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6746 (("," 1480) (possible_comma 15634)) 6747 ((")" 15635)) 6748 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (ampersand 15666) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15667) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6749 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15665) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6750 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15664) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6751 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15663) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6752 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15662) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6753 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15661) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6754 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15660) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6755 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15659) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6756 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15658) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6757 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15657) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6758 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15656) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6759 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15655) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6760 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15654) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6761 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15653) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6762 (("\"" 1520) ("$" 15668) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 15669) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 15670) ("[" 1542) (array_object_dereferenceable 15671) (callable_expr 15672) (callable_variable 15673) (class_constant 1546) (class_name 15674) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 15675) (name 15676) (new_dereferenceable 1552) (simple_variable 15677) (static_member 15678) (variable 15679) (variable_class_name 15680)) 6763 (("$" 15668) (T_VARIABLE 15670) (simple_variable 15713) ("{" 15714)) 6764 (("(" 15684) (argument_list 15712)) 6765 ((T_NULLSAFE_OBJECT_OPERATOR 15698) (T_OBJECT_OPERATOR 15699) ("[" 15700)) 6766 (("(" 15684) (argument_list 15697)) 6767 ((T_PAAMAYIM_NEKUDOTAYIM 15693)) 6768 (("(" 15684) (argument_list 15692)) 6769 ((T_PAAMAYIM_NEKUDOTAYIM 15681)) 6770 (("$" 15668) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 15670) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 15682) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 15683) ("{" 1642)) 6771 (("(" 15684) (argument_list 15685)) 6772 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 15686) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 15687) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 15688) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6773 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 15691) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6774 (("," 1480) (possible_comma 15689)) 6775 ((")" 15690)) 6776 (("$" 15668) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 15670) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 15694) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 15695) ("{" 1658)) 6777 (("(" 15684) (argument_list 15696)) 6778 (("$" 15668) (T_STRING 15703) (T_VARIABLE 15670) (property_name 15710) (simple_variable 15705) ("{" 15706)) 6779 (("$" 15668) (T_STRING 15703) (T_VARIABLE 15670) (property_name 15704) (simple_variable 15705) ("{" 15706)) 6780 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 15701) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6781 (("]" 15702)) 6782 (("(" 15684) (argument_list 15709)) 6783 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15707) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6784 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15708)) 6785 (("(" 15684) (argument_list 15711)) 6786 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15715) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6787 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15716)) 6788 ((backup_doc_comment 15719)) 6789 (("(" 15720)) 6790 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 15721)) 6791 ((")" 15722)) 6792 ((T_USE 164) (lexical_vars 15723)) 6793 ((":" 166) (return_type 15724)) 6794 ((backup_fn_flags 15725)) 6795 (("{" 15726)) 6796 ((inner_statement_list 15727)) 6797 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 15728) ("~" 127)) 6798 ((backup_fn_flags 15729)) 6799 ((backup_doc_comment 15731)) 6800 (("(" 15732)) 6801 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 15733)) 6802 ((")" 15734)) 6803 ((":" 2249) (return_type 15735)) 6804 ((T_DOUBLE_ARROW 15736)) 6805 ((backup_fn_flags 15737)) 6806 ((backup_lex_pos 15738)) 6807 (("!" 14704) ("\"" 14705) ("$" 14706) ("(" 14707) ("+" 14708) ("-" 14709) ("@" 14710) (T_ARRAY 14711) (T_ARRAY_CAST 14712) (T_ATTRIBUTE 128) (T_BOOL_CAST 14713) (T_CLASS_C 14714) (T_CLONE 14715) (T_CONSTANT_ENCAPSED_STRING 14716) (T_DEC 14717) (T_DIR 14718) (T_DNUMBER 14719) (T_DOUBLE_CAST 14720) (T_EMPTY 14721) (T_EVAL 14722) (T_EXIT 14723) (T_FILE 14724) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 14725) (T_INC 14726) (T_INCLUDE 14727) (T_INCLUDE_ONCE 14728) (T_INT_CAST 14729) (T_ISSET 14730) (T_LINE 14731) (T_LIST 14732) (T_LNUMBER 14733) (T_MATCH 14734) (T_METHOD_C 14735) (T_NAME_FULLY_QUALIFIED 14736) (T_NAME_QUALIFIED 14737) (T_NAME_RELATIVE 14738) (T_NEW 14739) (T_NS_C 14740) (T_OBJECT_CAST 14741) (T_PRINT 14742) (T_PROPERTY_C 14743) (T_READONLY 14744) (T_REQUIRE 14745) (T_REQUIRE_ONCE 14746) (T_START_HEREDOC 14747) (T_STATIC 14748) (T_STRING 14749) (T_STRING_CAST 14750) (T_THROW 14751) (T_TRAIT_C 14752) (T_UNSET_CAST 14753) (T_VARIABLE 14754) (T_YIELD 14755) (T_YIELD_FROM 14756) ("[" 14757) ("`" 14758) (array_object_dereferenceable 14759) (attribute 133) (attributes 14760) (callable_expr 14761) (callable_variable 14762) (class_constant 14763) (class_name 14764) (constant 14765) (dereferenceable_scalar 14766) (expr 15739) (fn 14768) (fully_dereferenceable 105) #'14769 (function_call 14770) (inline_function 14771) (internal_functions_in_yacc 14772) (match 14773) (name 14774) (new_dereferenceable 14775) (new_non_dereferenceable 14776) (scalar 14777) (simple_variable 14778) (static_member 14779) (variable 14780) (variable_class_name 14781) ("~" 14782)) 6808 (("%" 14784) ("*" 14785) ("+" 14786) ("-" 14787) ("." 14788) ("/" 14789) ("<" 14790) (">" 14791) ("?" 14792) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 14793) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 14794) (T_BOOLEAN_AND 14795) (T_BOOLEAN_OR 14796) (T_COALESCE 14797) (T_INSTANCEOF 14798) (T_IS_EQUAL 14799) (T_IS_GREATER_OR_EQUAL 14800) (T_IS_IDENTICAL 14801) (T_IS_NOT_EQUAL 14802) (T_IS_NOT_IDENTICAL 14803) (T_IS_SMALLER_OR_EQUAL 14804) (T_LOGICAL_AND 14805) (T_LOGICAL_OR 14806) (T_LOGICAL_XOR 14807) (T_POW 14808) (T_SL 14809) (T_SPACESHIP 14810) (T_SR 14811) ("^" 14812) (backup_fn_flags 15740) ("|" 14813)) 6809 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15742) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6810 (("$" 15365) (T_ABSTRACT 15542) (T_ARRAY 15543) (T_AS 15544) (T_BREAK 15545) (T_CALLABLE 15546) (T_CASE 15547) (T_CATCH 15548) (T_CLASS 15549) (T_CLASS_C 15550) (T_CLONE 15551) (T_CONST 15552) (T_CONTINUE 15553) (T_DECLARE 15554) (T_DEFAULT 15555) (T_DIR 15556) (T_DO 15557) (T_ECHO 15558) (T_ELSE 15559) (T_ELSEIF 15560) (T_EMPTY 15561) (T_ENDDECLARE 15562) (T_ENDFOR 15563) (T_ENDFOREACH 15564) (T_ENDIF 15565) (T_ENDSWITCH 15566) (T_ENDWHILE 15567) (T_ENUM 15568) (T_EVAL 15569) (T_EXIT 15570) (T_EXTENDS 15571) (T_FILE 15572) (T_FINAL 15573) (T_FINALLY 15574) (T_FN 15575) (T_FOR 15576) (T_FOREACH 15577) (T_FUNCTION 15578) (T_FUNC_C 15579) (T_GLOBAL 15580) (T_GOTO 15581) (T_IF 15582) (T_IMPLEMENTS 15583) (T_INCLUDE 15584) (T_INCLUDE_ONCE 15585) (T_INSTANCEOF 15586) (T_INSTEADOF 15587) (T_INTERFACE 15588) (T_ISSET 15589) (T_LINE 15590) (T_LIST 15591) (T_LOGICAL_AND 15592) (T_LOGICAL_OR 15593) (T_LOGICAL_XOR 15594) (T_MATCH 15595) (T_METHOD_C 15596) (T_NAMESPACE 15597) (T_NEW 15598) (T_NS_C 15599) (T_PRINT 15600) (T_PRIVATE 15601) (T_PROPERTY_C 15602) (T_PROTECTED 15603) (T_PUBLIC 15604) (T_READONLY 15605) (T_REQUIRE 15606) (T_REQUIRE_ONCE 15607) (T_RETURN 15608) (T_STATIC 15609) (T_STRING 15610) (T_SWITCH 15611) (T_THROW 15612) (T_TRAIT 15613) (T_TRAIT_C 15614) (T_TRY 15615) (T_UNSET 15616) (T_USE 15617) (T_VAR 15618) (T_VARIABLE 15413) (T_WHILE 15619) (T_YIELD 15620) (identifier 15744) (member_name 15745) (reserved_non_modifiers 15623) (semi_reserved 15624) (simple_variable 15746) ("{" 15747)) 6811 (("(" 15629) (argument_list 15750)) 6812 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15748) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6813 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15749)) 6814 ((T_FN 35) (T_FUNCTION 129) (fn 15427) #'15428 (inline_function 15754)) 6815 (("$" 15365) (T_STRING 15760) (T_VARIABLE 15413) (property_name 15767) (simple_variable 15762) ("{" 15763)) 6816 (("$" 15365) (T_STRING 15760) (T_VARIABLE 15413) (property_name 15761) (simple_variable 15762) ("{" 15763)) 6817 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 15758) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6818 (("]" 15759)) 6819 (("(" 15629) (argument_list 15766)) 6820 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15764) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6821 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15765)) 6822 (("(" 15629) (argument_list 15768)) 6823 (("`" 15770)) 6824 (("]" 15772)) 6825 (("=" 15773)) 6826 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15774) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6827 (("%" 15443) ("*" 15444) ("+" 15445) ("-" 15446) ("." 15447) ("/" 15448) ("<" 15449) (">" 15450) ("?" 15451) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15452) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15453) (T_BOOLEAN_AND 15454) (T_BOOLEAN_OR 15455) (T_COALESCE 15456) (T_DOUBLE_ARROW 15777) (T_INSTANCEOF 15457) (T_IS_EQUAL 15458) (T_IS_GREATER_OR_EQUAL 15459) (T_IS_IDENTICAL 15460) (T_IS_NOT_EQUAL 15461) (T_IS_NOT_IDENTICAL 15462) (T_IS_SMALLER_OR_EQUAL 15463) (T_LOGICAL_AND 15464) (T_LOGICAL_OR 15465) (T_LOGICAL_XOR 15466) (T_POW 15467) (T_SL 15468) (T_SPACESHIP 15469) (T_SR 15470) ("^" 15471) ("|" 15472)) 6828 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15778) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6829 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 15787) (T_VARIABLE 2896) (encaps_var 2910)) 6830 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 15786) (T_VARIABLE 2896) (encaps_var 2901)) 6831 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15828) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6832 ((T_CLASS 15820)) 6833 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 15819) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 15800) (attribute 2958) (class_modifier 2939)) 6834 ((T_PAAMAYIM_NEKUDOTAYIM 15817)) 6835 (("(" 15684) (argument_list 15816)) 6836 ((T_NULLSAFE_OBJECT_OPERATOR 15807) (T_OBJECT_OPERATOR 15808) (T_PAAMAYIM_NEKUDOTAYIM 15809) ("[" 15810)) 6837 (("$" 15668) (T_STRING 15703) (T_VARIABLE 15670) (property_name 15815) (simple_variable 15705) ("{" 15706)) 6838 (("$" 15668) (T_STRING 15703) (T_VARIABLE 15670) (property_name 15814) (simple_variable 15705) ("{" 15706)) 6839 (("$" 15668) (T_VARIABLE 15670) (simple_variable 15813)) 6840 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 15811) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6841 (("]" 15812)) 6842 (("$" 15668) (T_VARIABLE 15670) (simple_variable 15818)) 6843 (("(" 2960) (argument_list 2961) (ctor_arguments 15821)) 6844 ((T_EXTENDS 2963) (extends_from 15822)) 6845 ((T_IMPLEMENTS 2965) (implements_list 15823)) 6846 ((backup_doc_comment 15824)) 6847 (("{" 15825)) 6848 ((class_statement_list 15826)) 6849 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 15827)) 6850 (("%" 299) (")" 15829) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6851 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15831) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6852 (("%" 299) (")" 15832) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6853 (("{" 15833)) 6854 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 15834) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 6855 (("}" 15835)) 6856 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 15837) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 6857 ((")" 15838)) 6858 (("=" 15839)) 6859 (("!" 15363) ("\"" 15364) ("$" 15365) ("(" 15366) ("+" 15367) ("-" 15368) ("@" 15369) (T_ARRAY 15370) (T_ARRAY_CAST 15371) (T_ATTRIBUTE 128) (T_BOOL_CAST 15372) (T_CLASS_C 15373) (T_CLONE 15374) (T_CONSTANT_ENCAPSED_STRING 15375) (T_DEC 15376) (T_DIR 15377) (T_DNUMBER 15378) (T_DOUBLE_CAST 15379) (T_EMPTY 15380) (T_EVAL 15381) (T_EXIT 15382) (T_FILE 15383) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 15384) (T_INC 15385) (T_INCLUDE 15386) (T_INCLUDE_ONCE 15387) (T_INT_CAST 15388) (T_ISSET 15389) (T_LINE 15390) (T_LIST 15391) (T_LNUMBER 15392) (T_MATCH 15393) (T_METHOD_C 15394) (T_NAME_FULLY_QUALIFIED 15395) (T_NAME_QUALIFIED 15396) (T_NAME_RELATIVE 15397) (T_NEW 15398) (T_NS_C 15399) (T_OBJECT_CAST 15400) (T_PRINT 15401) (T_PROPERTY_C 15402) (T_READONLY 15403) (T_REQUIRE 15404) (T_REQUIRE_ONCE 15405) (T_START_HEREDOC 15406) (T_STATIC 15407) (T_STRING 15408) (T_STRING_CAST 15409) (T_THROW 15410) (T_TRAIT_C 15411) (T_UNSET_CAST 15412) (T_VARIABLE 15413) (T_YIELD 15414) (T_YIELD_FROM 15415) ("[" 15416) ("`" 15417) (array_object_dereferenceable 15418) (attribute 133) (attributes 15419) (callable_expr 15420) (callable_variable 15421) (class_constant 15422) (class_name 15423) (constant 15424) (dereferenceable_scalar 15425) (expr 15840) (fn 15427) (fully_dereferenceable 105) #'15428 (function_call 15429) (inline_function 15430) (internal_functions_in_yacc 15431) (match 15432) (name 15433) (new_dereferenceable 15434) (new_non_dereferenceable 15435) (scalar 15436) (simple_variable 15437) (static_member 15438) (variable 15439) (variable_class_name 15440) ("~" 15441)) 6860 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 15842) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6861 (("," 5192) (possible_comma 15843)) 6862 ((")" 15844)) 6863 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 15852) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 15853) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 15854) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6864 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 15857) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6865 (("," 1480) (possible_comma 15855)) 6866 ((")" 15856)) 6867 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15859) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6868 (("%" 299) (")" 15860) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6869 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15862) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6870 (("%" 299) (")" 15863) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6871 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 15870) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 6872 ((")" 15871)) 6873 (("%" 299) (")" 15876) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6874 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15879) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6875 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15880)) 6876 (("\"" 15882) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 6877 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 15892) (T_VARIABLE 2896) (encaps_var 2910)) 6878 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 15891) (T_VARIABLE 2896) (encaps_var 2901)) 6879 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15933) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6880 ((T_CLASS 15925)) 6881 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 15924) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 15905) (attribute 2958) (class_modifier 2939)) 6882 ((T_PAAMAYIM_NEKUDOTAYIM 15922)) 6883 (("(" 14236) (argument_list 15921)) 6884 ((T_NULLSAFE_OBJECT_OPERATOR 15912) (T_OBJECT_OPERATOR 15913) (T_PAAMAYIM_NEKUDOTAYIM 15914) ("[" 15915)) 6885 (("$" 14220) (T_STRING 14255) (T_VARIABLE 14222) (property_name 15920) (simple_variable 14257) ("{" 14258)) 6886 (("$" 14220) (T_STRING 14255) (T_VARIABLE 14222) (property_name 15919) (simple_variable 14257) ("{" 14258)) 6887 (("$" 14220) (T_VARIABLE 14222) (simple_variable 15918)) 6888 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 15916) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6889 (("]" 15917)) 6890 (("$" 14220) (T_VARIABLE 14222) (simple_variable 15923)) 6891 (("(" 2960) (argument_list 2961) (ctor_arguments 15926)) 6892 ((T_EXTENDS 2963) (extends_from 15927)) 6893 ((T_IMPLEMENTS 2965) (implements_list 15928)) 6894 ((backup_doc_comment 15929)) 6895 (("{" 15930)) 6896 ((class_statement_list 15931)) 6897 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 15932)) 6898 (("%" 299) (")" 15934) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6899 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15936) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6900 (("%" 299) (")" 15937) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6901 (("{" 15938)) 6902 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 15939) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 6903 (("}" 15940)) 6904 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 15942) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 6905 ((")" 15943)) 6906 (("=" 15944)) 6907 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 15945) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6908 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 15947) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6909 (("," 5192) (possible_comma 15948)) 6910 ((")" 15949)) 6911 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 15957) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 15958) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 15959) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6912 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 15962) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6913 (("," 1480) (possible_comma 15960)) 6914 ((")" 15961)) 6915 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15964) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6916 (("%" 299) (")" 15965) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6917 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 15967) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6918 (("%" 299) (")" 15968) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6919 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 15975) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 6920 ((")" 15976)) 6921 (("%" 299) (")" 15981) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6922 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15984) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6923 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15985)) 6924 (("\"" 15987) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 6925 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 15994) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6926 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 15995)) 6927 (("$" 344) (T_VARIABLE 351) (simple_variable 15998)) 6928 (("%" 299) (")" 16000) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 6929 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16003) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6930 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16004)) 6931 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16014) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6932 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 16012) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 6933 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16013) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6934 (("$" 221) (T_ABSTRACT 16024) (T_ARRAY 16025) (T_AS 16026) (T_BREAK 16027) (T_CALLABLE 16028) (T_CASE 16029) (T_CATCH 16030) (T_CLASS 16031) (T_CLASS_C 16032) (T_CLONE 16033) (T_CONST 16034) (T_CONTINUE 16035) (T_DECLARE 16036) (T_DEFAULT 16037) (T_DIR 16038) (T_DO 16039) (T_ECHO 16040) (T_ELSE 16041) (T_ELSEIF 16042) (T_EMPTY 16043) (T_ENDDECLARE 16044) (T_ENDFOR 16045) (T_ENDFOREACH 16046) (T_ENDIF 16047) (T_ENDSWITCH 16048) (T_ENDWHILE 16049) (T_ENUM 16050) (T_EVAL 16051) (T_EXIT 16052) (T_EXTENDS 16053) (T_FILE 16054) (T_FINAL 16055) (T_FINALLY 16056) (T_FN 16057) (T_FOR 16058) (T_FOREACH 16059) (T_FUNCTION 16060) (T_FUNC_C 16061) (T_GLOBAL 16062) (T_GOTO 16063) (T_IF 16064) (T_IMPLEMENTS 16065) (T_INCLUDE 16066) (T_INCLUDE_ONCE 16067) (T_INSTANCEOF 16068) (T_INSTEADOF 16069) (T_INTERFACE 16070) (T_ISSET 16071) (T_LINE 16072) (T_LIST 16073) (T_LOGICAL_AND 16074) (T_LOGICAL_OR 16075) (T_LOGICAL_XOR 16076) (T_MATCH 16077) (T_METHOD_C 16078) (T_NAMESPACE 16079) (T_NEW 16080) (T_NS_C 16081) (T_PRINT 16082) (T_PRIVATE 16083) (T_PROPERTY_C 16084) (T_PROTECTED 16085) (T_PUBLIC 16086) (T_READONLY 16087) (T_REQUIRE 16088) (T_REQUIRE_ONCE 16089) (T_RETURN 16090) (T_STATIC 16091) (T_STRING 16092) (T_SWITCH 16093) (T_THROW 16094) (T_TRAIT 16095) (T_TRAIT_C 16096) (T_TRY 16097) (T_UNSET 16098) (T_USE 16099) (T_VAR 16100) (T_VARIABLE 269) (T_WHILE 16101) (T_YIELD 16102) (identifier 16103) (member_name 16104) (reserved_non_modifiers 16105) (semi_reserved 16106) (simple_variable 16107) ("{" 16108)) 6935 (("(" 16111) (argument_list 16112)) 6936 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16109) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6937 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16110)) 6938 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 16113) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 16114) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 16115) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6939 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 16118) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6940 (("," 1480) (possible_comma 16116)) 6941 ((")" 16117)) 6942 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (ampersand 16148) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16149) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6943 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16147) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6944 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16146) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6945 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16145) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6946 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16144) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6947 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16143) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6948 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16142) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6949 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16141) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6950 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16140) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6951 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16139) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6952 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16138) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6953 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16137) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6954 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16136) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6955 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16135) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 6956 (("\"" 1520) ("$" 16150) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 16151) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 16152) ("[" 1542) (array_object_dereferenceable 16153) (callable_expr 16154) (callable_variable 16155) (class_constant 1546) (class_name 16156) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 16157) (name 16158) (new_dereferenceable 1552) (simple_variable 16159) (static_member 16160) (variable 16161) (variable_class_name 16162)) 6957 (("$" 16150) (T_VARIABLE 16152) (simple_variable 16195) ("{" 16196)) 6958 (("(" 16166) (argument_list 16194)) 6959 ((T_NULLSAFE_OBJECT_OPERATOR 16180) (T_OBJECT_OPERATOR 16181) ("[" 16182)) 6960 (("(" 16166) (argument_list 16179)) 6961 ((T_PAAMAYIM_NEKUDOTAYIM 16175)) 6962 (("(" 16166) (argument_list 16174)) 6963 ((T_PAAMAYIM_NEKUDOTAYIM 16163)) 6964 (("$" 16150) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 16152) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 16164) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 16165) ("{" 1642)) 6965 (("(" 16166) (argument_list 16167)) 6966 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 16168) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 16169) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 16170) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6967 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 16173) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 6968 (("," 1480) (possible_comma 16171)) 6969 ((")" 16172)) 6970 (("$" 16150) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 16152) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 16176) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 16177) ("{" 1658)) 6971 (("(" 16166) (argument_list 16178)) 6972 (("$" 16150) (T_STRING 16185) (T_VARIABLE 16152) (property_name 16192) (simple_variable 16187) ("{" 16188)) 6973 (("$" 16150) (T_STRING 16185) (T_VARIABLE 16152) (property_name 16186) (simple_variable 16187) ("{" 16188)) 6974 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 16183) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 6975 (("]" 16184)) 6976 (("(" 16166) (argument_list 16191)) 6977 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16189) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6978 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16190)) 6979 (("(" 16166) (argument_list 16193)) 6980 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16197) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 6981 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16198)) 6982 ((backup_doc_comment 16201)) 6983 (("(" 16202)) 6984 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 16203)) 6985 ((")" 16204)) 6986 ((T_USE 164) (lexical_vars 16205)) 6987 ((":" 166) (return_type 16206)) 6988 ((backup_fn_flags 16207)) 6989 (("{" 16208)) 6990 ((inner_statement_list 16209)) 6991 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 16210) ("~" 127)) 6992 ((backup_fn_flags 16211)) 6993 ((backup_doc_comment 16213)) 6994 (("(" 16214)) 6995 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 16215)) 6996 ((")" 16216)) 6997 ((":" 2249) (return_type 16217)) 6998 ((T_DOUBLE_ARROW 16218)) 6999 ((backup_fn_flags 16219)) 7000 ((backup_lex_pos 16220)) 7001 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16284) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7002 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17258) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7003 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 17256) (encaps_var 5237)) 7004 (("$" 16223) (T_VARIABLE 16271) (simple_variable 17252) ("{" 17253)) 7005 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17250) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7006 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17249) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7007 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17248) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7008 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17247) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7009 (("(" 17244)) 7010 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17243) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7011 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17242) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7012 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17241) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7013 (("\"" 1520) ("$" 16526) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 16527) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 16528) ("[" 1542) (array_object_dereferenceable 16529) (callable_expr 16530) (callable_variable 16531) (class_constant 1546) (class_name 16532) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 16533) (name 16534) (new_dereferenceable 1552) (simple_variable 16535) (static_member 16536) (variable 17240) (variable_class_name 16538)) 7014 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17239) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7015 (("(" 17236)) 7016 (("(" 17233)) 7017 (("(" 17224) (argument_list 17225) (ctor_arguments 17226)) 7018 (("\"" 1520) ("$" 16526) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 16527) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 16528) ("[" 1542) (array_object_dereferenceable 16529) (callable_expr 16530) (callable_variable 16531) (class_constant 1546) (class_name 16532) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 16533) (name 16534) (new_dereferenceable 1552) (simple_variable 16535) (static_member 16536) (variable 17223) (variable_class_name 16538)) 7019 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17222) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7020 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17221) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7021 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17220) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7022 (("(" 17216)) 7023 (("(" 17211)) 7024 (("(" 17205)) 7025 (("$" 16526) ("(" 17168) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 17169) (T_NAME_QUALIFIED 17170) (T_NAME_RELATIVE 17171) (T_READONLY 2931) (T_STATIC 17172) (T_STRING 17173) (T_VARIABLE 16528) (anonymous_class 17174) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 17175) (attribute 2937) (attributes 17176) (class_modifier 2939) (class_name 17177) (class_name_reference 17178) (name 17179) (new_variable 17180) (simple_variable 17181)) 7026 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17167) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7027 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17166) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7028 (("(" 16487) (argument_list 17165)) 7029 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17164) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7030 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17163) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7031 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 17158) (T_END_HEREDOC 17159) (T_VARIABLE 2896) (encaps_list 17160) (encaps_var 2898)) 7032 ((T_FN 35) (T_FUNCTION 129) (fn 16285) #'16286 (inline_function 17157)) 7033 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17156) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7034 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17155) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7035 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17154) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7036 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16696) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7037 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16632) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7038 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 16628) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 7039 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 16626) (encaps_list 2804) (encaps_var 2805)) 7040 ((T_NULLSAFE_OBJECT_OPERATOR 16612) (T_OBJECT_OPERATOR 16613) ("[" 16614)) 7041 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 16609) (attribute 2733) (fn 16285) #'16286 (inline_function 16610)) 7042 (("(" 16487) (argument_list 16608)) 7043 ((T_PAAMAYIM_NEKUDOTAYIM 16600)) 7044 (("%" 16301) ("*" 16302) ("+" 16303) ("-" 16304) ("." 16305) ("/" 16306) ("<" 16307) (">" 16308) ("?" 16309) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 16310) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 16311) (T_BOOLEAN_AND 16312) (T_BOOLEAN_OR 16313) (T_COALESCE 16314) (T_INSTANCEOF 16315) (T_IS_EQUAL 16316) (T_IS_GREATER_OR_EQUAL 16317) (T_IS_IDENTICAL 16318) (T_IS_NOT_EQUAL 16319) (T_IS_NOT_IDENTICAL 16320) (T_IS_SMALLER_OR_EQUAL 16321) (T_LOGICAL_AND 16322) (T_LOGICAL_OR 16323) (T_LOGICAL_XOR 16324) (T_POW 16325) (T_SL 16326) (T_SPACESHIP 16327) (T_SR 16328) ("^" 16329) (backup_fn_flags 16599) ("|" 16330)) 7045 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 16588)) 7046 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 16576)) 7047 (("(" 16487) (argument_list 16575)) 7048 (("=" 16495) (T_AND_EQUAL 16496) (T_COALESCE_EQUAL 16497) (T_CONCAT_EQUAL 16498) (T_DEC 16499) (T_DIV_EQUAL 16500) (T_INC 16501) (T_MINUS_EQUAL 16502) (T_MOD_EQUAL 16503) (T_MUL_EQUAL 16504) (T_OR_EQUAL 16505) (T_PLUS_EQUAL 16506) (T_POW_EQUAL 16507) (T_SL_EQUAL 16508) (T_SR_EQUAL 16509) (T_XOR_EQUAL 16510)) 7049 ((T_PAAMAYIM_NEKUDOTAYIM 16399)) 7050 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16300) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7051 (("%" 16301) ("*" 16302) ("+" 16303) ("-" 16304) ("." 16305) ("/" 16306) ("<" 16307) (">" 16308) ("?" 16309) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 16310) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 16311) (T_BOOLEAN_AND 16312) (T_BOOLEAN_OR 16313) (T_COALESCE 16314) (T_INSTANCEOF 16315) (T_IS_EQUAL 16316) (T_IS_GREATER_OR_EQUAL 16317) (T_IS_IDENTICAL 16318) (T_IS_NOT_EQUAL 16319) (T_IS_NOT_IDENTICAL 16320) (T_IS_SMALLER_OR_EQUAL 16321) (T_LOGICAL_AND 16322) (T_LOGICAL_OR 16323) (T_LOGICAL_XOR 16324) (T_POW 16325) (T_SL 16326) (T_SPACESHIP 16327) (T_SR 16328) ("^" 16329) ("|" 16330)) 7052 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16398) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7053 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16397) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7054 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16396) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7055 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16395) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7056 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16394) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7057 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16393) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7058 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16392) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7059 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16391) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7060 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 16386) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 16387) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 7061 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16385) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7062 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16384) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7063 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16383) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7064 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16382) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7065 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16381) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7066 (("$" 16346) ("(" 16347) (T_NAME_FULLY_QUALIFIED 16348) (T_NAME_QUALIFIED 16349) (T_NAME_RELATIVE 16350) (T_STATIC 16351) (T_STRING 16352) (T_VARIABLE 16353) (class_name 16354) (class_name_reference 16355) (name 16356) (new_variable 16357) (simple_variable 16358)) 7067 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16345) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7068 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16344) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7069 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16343) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7070 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16342) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7071 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16341) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7072 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16340) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7073 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16339) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7074 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16338) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7075 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16337) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7076 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16336) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7077 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16335) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7078 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16334) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7079 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16333) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7080 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16332) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7081 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16331) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7082 (("$" 16346) (T_VARIABLE 16353) (simple_variable 16377) ("{" 16378)) 7083 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16375) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7084 ((T_PAAMAYIM_NEKUDOTAYIM 16373)) 7085 ((T_NULLSAFE_OBJECT_OPERATOR 16359) (T_OBJECT_OPERATOR 16360) (T_PAAMAYIM_NEKUDOTAYIM 16361) ("[" 16362)) 7086 (("$" 16346) (T_STRING 16366) (T_VARIABLE 16353) (property_name 16372) (simple_variable 16368) ("{" 16369)) 7087 (("$" 16346) (T_STRING 16366) (T_VARIABLE 16353) (property_name 16367) (simple_variable 16368) ("{" 16369)) 7088 (("$" 16346) (T_VARIABLE 16353) (simple_variable 16365)) 7089 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 16363) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 7090 (("]" 16364)) 7091 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16370) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7092 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16371)) 7093 (("$" 16346) (T_VARIABLE 16353) (simple_variable 16374)) 7094 (("%" 299) (")" 16376) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7095 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16379) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7096 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16380)) 7097 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16390) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7098 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 16388) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 7099 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16389) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7100 (("$" 16223) (T_ABSTRACT 16400) (T_ARRAY 16401) (T_AS 16402) (T_BREAK 16403) (T_CALLABLE 16404) (T_CASE 16405) (T_CATCH 16406) (T_CLASS 16407) (T_CLASS_C 16408) (T_CLONE 16409) (T_CONST 16410) (T_CONTINUE 16411) (T_DECLARE 16412) (T_DEFAULT 16413) (T_DIR 16414) (T_DO 16415) (T_ECHO 16416) (T_ELSE 16417) (T_ELSEIF 16418) (T_EMPTY 16419) (T_ENDDECLARE 16420) (T_ENDFOR 16421) (T_ENDFOREACH 16422) (T_ENDIF 16423) (T_ENDSWITCH 16424) (T_ENDWHILE 16425) (T_ENUM 16426) (T_EVAL 16427) (T_EXIT 16428) (T_EXTENDS 16429) (T_FILE 16430) (T_FINAL 16431) (T_FINALLY 16432) (T_FN 16433) (T_FOR 16434) (T_FOREACH 16435) (T_FUNCTION 16436) (T_FUNC_C 16437) (T_GLOBAL 16438) (T_GOTO 16439) (T_IF 16440) (T_IMPLEMENTS 16441) (T_INCLUDE 16442) (T_INCLUDE_ONCE 16443) (T_INSTANCEOF 16444) (T_INSTEADOF 16445) (T_INTERFACE 16446) (T_ISSET 16447) (T_LINE 16448) (T_LIST 16449) (T_LOGICAL_AND 16450) (T_LOGICAL_OR 16451) (T_LOGICAL_XOR 16452) (T_MATCH 16453) (T_METHOD_C 16454) (T_NAMESPACE 16455) (T_NEW 16456) (T_NS_C 16457) (T_PRINT 16458) (T_PRIVATE 16459) (T_PROPERTY_C 16460) (T_PROTECTED 16461) (T_PUBLIC 16462) (T_READONLY 16463) (T_REQUIRE 16464) (T_REQUIRE_ONCE 16465) (T_RETURN 16466) (T_STATIC 16467) (T_STRING 16468) (T_SWITCH 16469) (T_THROW 16470) (T_TRAIT 16471) (T_TRAIT_C 16472) (T_TRY 16473) (T_UNSET 16474) (T_USE 16475) (T_VAR 16476) (T_VARIABLE 16271) (T_WHILE 16477) (T_YIELD 16478) (identifier 16479) (member_name 16480) (reserved_non_modifiers 16481) (semi_reserved 16482) (simple_variable 16483) ("{" 16484)) 7101 (("(" 16487) (argument_list 16488)) 7102 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16485) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7103 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16486)) 7104 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 16489) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 16490) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 16491) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7105 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 16494) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7106 (("," 1480) (possible_comma 16492)) 7107 ((")" 16493)) 7108 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (ampersand 16524) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16525) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7109 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16523) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7110 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16522) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7111 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16521) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7112 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16520) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7113 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16519) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7114 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16518) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7115 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16517) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7116 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16516) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7117 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16515) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7118 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16514) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7119 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16513) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7120 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16512) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7121 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16511) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7122 (("\"" 1520) ("$" 16526) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 16527) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 16528) ("[" 1542) (array_object_dereferenceable 16529) (callable_expr 16530) (callable_variable 16531) (class_constant 1546) (class_name 16532) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 16533) (name 16534) (new_dereferenceable 1552) (simple_variable 16535) (static_member 16536) (variable 16537) (variable_class_name 16538)) 7123 (("$" 16526) (T_VARIABLE 16528) (simple_variable 16571) ("{" 16572)) 7124 (("(" 16542) (argument_list 16570)) 7125 ((T_NULLSAFE_OBJECT_OPERATOR 16556) (T_OBJECT_OPERATOR 16557) ("[" 16558)) 7126 (("(" 16542) (argument_list 16555)) 7127 ((T_PAAMAYIM_NEKUDOTAYIM 16551)) 7128 (("(" 16542) (argument_list 16550)) 7129 ((T_PAAMAYIM_NEKUDOTAYIM 16539)) 7130 (("$" 16526) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 16528) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 16540) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 16541) ("{" 1642)) 7131 (("(" 16542) (argument_list 16543)) 7132 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 16544) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 16545) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 16546) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7133 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 16549) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7134 (("," 1480) (possible_comma 16547)) 7135 ((")" 16548)) 7136 (("$" 16526) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 16528) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 16552) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 16553) ("{" 1658)) 7137 (("(" 16542) (argument_list 16554)) 7138 (("$" 16526) (T_STRING 16561) (T_VARIABLE 16528) (property_name 16568) (simple_variable 16563) ("{" 16564)) 7139 (("$" 16526) (T_STRING 16561) (T_VARIABLE 16528) (property_name 16562) (simple_variable 16563) ("{" 16564)) 7140 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 16559) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 7141 (("]" 16560)) 7142 (("(" 16542) (argument_list 16567)) 7143 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16565) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7144 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16566)) 7145 (("(" 16542) (argument_list 16569)) 7146 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16573) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7147 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16574)) 7148 ((backup_doc_comment 16577)) 7149 (("(" 16578)) 7150 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 16579)) 7151 ((")" 16580)) 7152 ((T_USE 164) (lexical_vars 16581)) 7153 ((":" 166) (return_type 16582)) 7154 ((backup_fn_flags 16583)) 7155 (("{" 16584)) 7156 ((inner_statement_list 16585)) 7157 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 16586) ("~" 127)) 7158 ((backup_fn_flags 16587)) 7159 ((backup_doc_comment 16589)) 7160 (("(" 16590)) 7161 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 16591)) 7162 ((")" 16592)) 7163 ((":" 2249) (return_type 16593)) 7164 ((T_DOUBLE_ARROW 16594)) 7165 ((backup_fn_flags 16595)) 7166 ((backup_lex_pos 16596)) 7167 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16597) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7168 (("%" 16301) ("*" 16302) ("+" 16303) ("-" 16304) ("." 16305) ("/" 16306) ("<" 16307) (">" 16308) ("?" 16309) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 16310) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 16311) (T_BOOLEAN_AND 16312) (T_BOOLEAN_OR 16313) (T_COALESCE 16314) (T_INSTANCEOF 16315) (T_IS_EQUAL 16316) (T_IS_GREATER_OR_EQUAL 16317) (T_IS_IDENTICAL 16318) (T_IS_NOT_EQUAL 16319) (T_IS_NOT_IDENTICAL 16320) (T_IS_SMALLER_OR_EQUAL 16321) (T_LOGICAL_AND 16322) (T_LOGICAL_OR 16323) (T_LOGICAL_XOR 16324) (T_POW 16325) (T_SL 16326) (T_SPACESHIP 16327) (T_SR 16328) ("^" 16329) (backup_fn_flags 16598) ("|" 16330)) 7169 (("$" 16223) (T_ABSTRACT 16400) (T_ARRAY 16401) (T_AS 16402) (T_BREAK 16403) (T_CALLABLE 16404) (T_CASE 16405) (T_CATCH 16406) (T_CLASS 16407) (T_CLASS_C 16408) (T_CLONE 16409) (T_CONST 16410) (T_CONTINUE 16411) (T_DECLARE 16412) (T_DEFAULT 16413) (T_DIR 16414) (T_DO 16415) (T_ECHO 16416) (T_ELSE 16417) (T_ELSEIF 16418) (T_EMPTY 16419) (T_ENDDECLARE 16420) (T_ENDFOR 16421) (T_ENDFOREACH 16422) (T_ENDIF 16423) (T_ENDSWITCH 16424) (T_ENDWHILE 16425) (T_ENUM 16426) (T_EVAL 16427) (T_EXIT 16428) (T_EXTENDS 16429) (T_FILE 16430) (T_FINAL 16431) (T_FINALLY 16432) (T_FN 16433) (T_FOR 16434) (T_FOREACH 16435) (T_FUNCTION 16436) (T_FUNC_C 16437) (T_GLOBAL 16438) (T_GOTO 16439) (T_IF 16440) (T_IMPLEMENTS 16441) (T_INCLUDE 16442) (T_INCLUDE_ONCE 16443) (T_INSTANCEOF 16444) (T_INSTEADOF 16445) (T_INTERFACE 16446) (T_ISSET 16447) (T_LINE 16448) (T_LIST 16449) (T_LOGICAL_AND 16450) (T_LOGICAL_OR 16451) (T_LOGICAL_XOR 16452) (T_MATCH 16453) (T_METHOD_C 16454) (T_NAMESPACE 16455) (T_NEW 16456) (T_NS_C 16457) (T_PRINT 16458) (T_PRIVATE 16459) (T_PROPERTY_C 16460) (T_PROTECTED 16461) (T_PUBLIC 16462) (T_READONLY 16463) (T_REQUIRE 16464) (T_REQUIRE_ONCE 16465) (T_RETURN 16466) (T_STATIC 16467) (T_STRING 16468) (T_SWITCH 16469) (T_THROW 16470) (T_TRAIT 16471) (T_TRAIT_C 16472) (T_TRY 16473) (T_UNSET 16474) (T_USE 16475) (T_VAR 16476) (T_VARIABLE 16271) (T_WHILE 16477) (T_YIELD 16478) (identifier 16601) (member_name 16602) (reserved_non_modifiers 16481) (semi_reserved 16482) (simple_variable 16603) ("{" 16604)) 7170 (("(" 16487) (argument_list 16607)) 7171 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16605) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7172 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16606)) 7173 ((T_FN 35) (T_FUNCTION 129) (fn 16285) #'16286 (inline_function 16611)) 7174 (("$" 16223) (T_STRING 16617) (T_VARIABLE 16271) (property_name 16624) (simple_variable 16619) ("{" 16620)) 7175 (("$" 16223) (T_STRING 16617) (T_VARIABLE 16271) (property_name 16618) (simple_variable 16619) ("{" 16620)) 7176 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 16615) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 7177 (("]" 16616)) 7178 (("(" 16487) (argument_list 16623)) 7179 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16621) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7180 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16622)) 7181 (("(" 16487) (argument_list 16625)) 7182 (("`" 16627)) 7183 (("]" 16629)) 7184 (("=" 16630)) 7185 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 16631) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7186 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17153) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7187 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 17151) (encaps_var 5237)) 7188 (("$" 16635) (T_VARIABLE 16683) (simple_variable 17147) ("{" 17148)) 7189 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17145) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7190 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17144) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7191 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17143) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7192 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17142) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7193 (("(" 17139)) 7194 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17138) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7195 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17137) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7196 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17136) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7197 (("\"" 1520) ("$" 16938) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 16939) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 16940) ("[" 1542) (array_object_dereferenceable 16941) (callable_expr 16942) (callable_variable 16943) (class_constant 1546) (class_name 16944) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 16945) (name 16946) (new_dereferenceable 1552) (simple_variable 16947) (static_member 16948) (variable 17135) (variable_class_name 16950)) 7198 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17134) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7199 (("(" 17131)) 7200 (("(" 17128)) 7201 (("(" 17119) (argument_list 17120) (ctor_arguments 17121)) 7202 (("\"" 1520) ("$" 16938) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 16939) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 16940) ("[" 1542) (array_object_dereferenceable 16941) (callable_expr 16942) (callable_variable 16943) (class_constant 1546) (class_name 16944) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 16945) (name 16946) (new_dereferenceable 1552) (simple_variable 16947) (static_member 16948) (variable 17118) (variable_class_name 16950)) 7203 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17117) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7204 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17116) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7205 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17115) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7206 (("(" 17111)) 7207 (("(" 17106)) 7208 (("(" 17100)) 7209 (("$" 16938) ("(" 17063) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 17064) (T_NAME_QUALIFIED 17065) (T_NAME_RELATIVE 17066) (T_READONLY 2931) (T_STATIC 17067) (T_STRING 17068) (T_VARIABLE 16940) (anonymous_class 17069) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 17070) (attribute 2937) (attributes 17071) (class_modifier 2939) (class_name 17072) (class_name_reference 17073) (name 17074) (new_variable 17075) (simple_variable 17076)) 7210 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17062) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7211 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17061) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7212 (("(" 16899) (argument_list 17060)) 7213 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17059) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7214 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17058) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7215 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 17053) (T_END_HEREDOC 17054) (T_VARIABLE 2896) (encaps_list 17055) (encaps_var 2898)) 7216 ((T_FN 35) (T_FUNCTION 129) (fn 16697) #'16698 (inline_function 17052)) 7217 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17051) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7218 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17050) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7219 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17049) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7220 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17046) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7221 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17045) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7222 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 17041) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 7223 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 17039) (encaps_list 2804) (encaps_var 2805)) 7224 ((T_NULLSAFE_OBJECT_OPERATOR 17025) (T_OBJECT_OPERATOR 17026) ("[" 17027)) 7225 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 17022) (attribute 2733) (fn 16697) #'16698 (inline_function 17023)) 7226 (("(" 16899) (argument_list 17021)) 7227 ((T_PAAMAYIM_NEKUDOTAYIM 17013)) 7228 (("%" 16713) ("*" 16714) ("+" 16715) ("-" 16716) ("." 16717) ("/" 16718) ("<" 16719) (">" 16720) ("?" 16721) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 16722) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 16723) (T_BOOLEAN_AND 16724) (T_BOOLEAN_OR 16725) (T_COALESCE 16726) (T_DOUBLE_ARROW 17011) (T_INSTANCEOF 16727) (T_IS_EQUAL 16728) (T_IS_GREATER_OR_EQUAL 16729) (T_IS_IDENTICAL 16730) (T_IS_NOT_EQUAL 16731) (T_IS_NOT_IDENTICAL 16732) (T_IS_SMALLER_OR_EQUAL 16733) (T_LOGICAL_AND 16734) (T_LOGICAL_OR 16735) (T_LOGICAL_XOR 16736) (T_POW 16737) (T_SL 16738) (T_SPACESHIP 16739) (T_SR 16740) ("^" 16741) ("|" 16742)) 7229 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 17000)) 7230 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 16988)) 7231 (("(" 16899) (argument_list 16987)) 7232 (("=" 16907) (T_AND_EQUAL 16908) (T_COALESCE_EQUAL 16909) (T_CONCAT_EQUAL 16910) (T_DEC 16911) (T_DIV_EQUAL 16912) (T_INC 16913) (T_MINUS_EQUAL 16914) (T_MOD_EQUAL 16915) (T_MUL_EQUAL 16916) (T_OR_EQUAL 16917) (T_PLUS_EQUAL 16918) (T_POW_EQUAL 16919) (T_SL_EQUAL 16920) (T_SR_EQUAL 16921) (T_XOR_EQUAL 16922)) 7233 ((T_PAAMAYIM_NEKUDOTAYIM 16811)) 7234 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16712) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7235 (("%" 16713) ("*" 16714) ("+" 16715) ("-" 16716) ("." 16717) ("/" 16718) ("<" 16719) (">" 16720) ("?" 16721) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 16722) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 16723) (T_BOOLEAN_AND 16724) (T_BOOLEAN_OR 16725) (T_COALESCE 16726) (T_INSTANCEOF 16727) (T_IS_EQUAL 16728) (T_IS_GREATER_OR_EQUAL 16729) (T_IS_IDENTICAL 16730) (T_IS_NOT_EQUAL 16731) (T_IS_NOT_IDENTICAL 16732) (T_IS_SMALLER_OR_EQUAL 16733) (T_LOGICAL_AND 16734) (T_LOGICAL_OR 16735) (T_LOGICAL_XOR 16736) (T_POW 16737) (T_SL 16738) (T_SPACESHIP 16739) (T_SR 16740) ("^" 16741) ("|" 16742)) 7236 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16810) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7237 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16809) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7238 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16808) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7239 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16807) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7240 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16806) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7241 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16805) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7242 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16804) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7243 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16803) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7244 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 16798) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 16799) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 7245 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16797) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7246 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16796) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7247 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16795) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7248 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16794) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7249 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16793) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7250 (("$" 16758) ("(" 16759) (T_NAME_FULLY_QUALIFIED 16760) (T_NAME_QUALIFIED 16761) (T_NAME_RELATIVE 16762) (T_STATIC 16763) (T_STRING 16764) (T_VARIABLE 16765) (class_name 16766) (class_name_reference 16767) (name 16768) (new_variable 16769) (simple_variable 16770)) 7251 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16757) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7252 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16756) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7253 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16755) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7254 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16754) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7255 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16753) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7256 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16752) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7257 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16751) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7258 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16750) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7259 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16749) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7260 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16748) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7261 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16747) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7262 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16746) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7263 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16745) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7264 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16744) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7265 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16743) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7266 (("$" 16758) (T_VARIABLE 16765) (simple_variable 16789) ("{" 16790)) 7267 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 16787) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7268 ((T_PAAMAYIM_NEKUDOTAYIM 16785)) 7269 ((T_NULLSAFE_OBJECT_OPERATOR 16771) (T_OBJECT_OPERATOR 16772) (T_PAAMAYIM_NEKUDOTAYIM 16773) ("[" 16774)) 7270 (("$" 16758) (T_STRING 16778) (T_VARIABLE 16765) (property_name 16784) (simple_variable 16780) ("{" 16781)) 7271 (("$" 16758) (T_STRING 16778) (T_VARIABLE 16765) (property_name 16779) (simple_variable 16780) ("{" 16781)) 7272 (("$" 16758) (T_VARIABLE 16765) (simple_variable 16777)) 7273 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 16775) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 7274 (("]" 16776)) 7275 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16782) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7276 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16783)) 7277 (("$" 16758) (T_VARIABLE 16765) (simple_variable 16786)) 7278 (("%" 299) (")" 16788) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7279 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16791) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7280 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16792)) 7281 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16802) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7282 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 16800) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 7283 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16801) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7284 (("$" 16635) (T_ABSTRACT 16812) (T_ARRAY 16813) (T_AS 16814) (T_BREAK 16815) (T_CALLABLE 16816) (T_CASE 16817) (T_CATCH 16818) (T_CLASS 16819) (T_CLASS_C 16820) (T_CLONE 16821) (T_CONST 16822) (T_CONTINUE 16823) (T_DECLARE 16824) (T_DEFAULT 16825) (T_DIR 16826) (T_DO 16827) (T_ECHO 16828) (T_ELSE 16829) (T_ELSEIF 16830) (T_EMPTY 16831) (T_ENDDECLARE 16832) (T_ENDFOR 16833) (T_ENDFOREACH 16834) (T_ENDIF 16835) (T_ENDSWITCH 16836) (T_ENDWHILE 16837) (T_ENUM 16838) (T_EVAL 16839) (T_EXIT 16840) (T_EXTENDS 16841) (T_FILE 16842) (T_FINAL 16843) (T_FINALLY 16844) (T_FN 16845) (T_FOR 16846) (T_FOREACH 16847) (T_FUNCTION 16848) (T_FUNC_C 16849) (T_GLOBAL 16850) (T_GOTO 16851) (T_IF 16852) (T_IMPLEMENTS 16853) (T_INCLUDE 16854) (T_INCLUDE_ONCE 16855) (T_INSTANCEOF 16856) (T_INSTEADOF 16857) (T_INTERFACE 16858) (T_ISSET 16859) (T_LINE 16860) (T_LIST 16861) (T_LOGICAL_AND 16862) (T_LOGICAL_OR 16863) (T_LOGICAL_XOR 16864) (T_MATCH 16865) (T_METHOD_C 16866) (T_NAMESPACE 16867) (T_NEW 16868) (T_NS_C 16869) (T_PRINT 16870) (T_PRIVATE 16871) (T_PROPERTY_C 16872) (T_PROTECTED 16873) (T_PUBLIC 16874) (T_READONLY 16875) (T_REQUIRE 16876) (T_REQUIRE_ONCE 16877) (T_RETURN 16878) (T_STATIC 16879) (T_STRING 16880) (T_SWITCH 16881) (T_THROW 16882) (T_TRAIT 16883) (T_TRAIT_C 16884) (T_TRY 16885) (T_UNSET 16886) (T_USE 16887) (T_VAR 16888) (T_VARIABLE 16683) (T_WHILE 16889) (T_YIELD 16890) (identifier 16891) (member_name 16892) (reserved_non_modifiers 16893) (semi_reserved 16894) (simple_variable 16895) ("{" 16896)) 7285 (("(" 16899) (argument_list 16900)) 7286 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16897) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7287 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16898)) 7288 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 16901) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 16902) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 16903) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7289 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 16906) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7290 (("," 1480) (possible_comma 16904)) 7291 ((")" 16905)) 7292 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (ampersand 16936) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16937) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7293 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16935) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7294 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16934) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7295 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16933) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7296 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16932) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7297 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16931) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7298 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16930) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7299 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16929) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7300 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16928) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7301 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16927) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7302 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16926) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7303 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16925) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7304 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16924) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7305 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 16923) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7306 (("\"" 1520) ("$" 16938) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 16939) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 16940) ("[" 1542) (array_object_dereferenceable 16941) (callable_expr 16942) (callable_variable 16943) (class_constant 1546) (class_name 16944) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 16945) (name 16946) (new_dereferenceable 1552) (simple_variable 16947) (static_member 16948) (variable 16949) (variable_class_name 16950)) 7307 (("$" 16938) (T_VARIABLE 16940) (simple_variable 16983) ("{" 16984)) 7308 (("(" 16954) (argument_list 16982)) 7309 ((T_NULLSAFE_OBJECT_OPERATOR 16968) (T_OBJECT_OPERATOR 16969) ("[" 16970)) 7310 (("(" 16954) (argument_list 16967)) 7311 ((T_PAAMAYIM_NEKUDOTAYIM 16963)) 7312 (("(" 16954) (argument_list 16962)) 7313 ((T_PAAMAYIM_NEKUDOTAYIM 16951)) 7314 (("$" 16938) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 16940) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 16952) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 16953) ("{" 1642)) 7315 (("(" 16954) (argument_list 16955)) 7316 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 16956) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 16957) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 16958) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7317 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 16961) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7318 (("," 1480) (possible_comma 16959)) 7319 ((")" 16960)) 7320 (("$" 16938) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 16940) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 16964) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 16965) ("{" 1658)) 7321 (("(" 16954) (argument_list 16966)) 7322 (("$" 16938) (T_STRING 16973) (T_VARIABLE 16940) (property_name 16980) (simple_variable 16975) ("{" 16976)) 7323 (("$" 16938) (T_STRING 16973) (T_VARIABLE 16940) (property_name 16974) (simple_variable 16975) ("{" 16976)) 7324 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 16971) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 7325 (("]" 16972)) 7326 (("(" 16954) (argument_list 16979)) 7327 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16977) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7328 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16978)) 7329 (("(" 16954) (argument_list 16981)) 7330 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 16985) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7331 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 16986)) 7332 ((backup_doc_comment 16989)) 7333 (("(" 16990)) 7334 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 16991)) 7335 ((")" 16992)) 7336 ((T_USE 164) (lexical_vars 16993)) 7337 ((":" 166) (return_type 16994)) 7338 ((backup_fn_flags 16995)) 7339 (("{" 16996)) 7340 ((inner_statement_list 16997)) 7341 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 16998) ("~" 127)) 7342 ((backup_fn_flags 16999)) 7343 ((backup_doc_comment 17001)) 7344 (("(" 17002)) 7345 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 17003)) 7346 ((")" 17004)) 7347 ((":" 2249) (return_type 17005)) 7348 ((T_DOUBLE_ARROW 17006)) 7349 ((backup_fn_flags 17007)) 7350 ((backup_lex_pos 17008)) 7351 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17009) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7352 (("%" 16713) ("*" 16714) ("+" 16715) ("-" 16716) ("." 16717) ("/" 16718) ("<" 16719) (">" 16720) ("?" 16721) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 16722) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 16723) (T_BOOLEAN_AND 16724) (T_BOOLEAN_OR 16725) (T_COALESCE 16726) (T_INSTANCEOF 16727) (T_IS_EQUAL 16728) (T_IS_GREATER_OR_EQUAL 16729) (T_IS_IDENTICAL 16730) (T_IS_NOT_EQUAL 16731) (T_IS_NOT_IDENTICAL 16732) (T_IS_SMALLER_OR_EQUAL 16733) (T_LOGICAL_AND 16734) (T_LOGICAL_OR 16735) (T_LOGICAL_XOR 16736) (T_POW 16737) (T_SL 16738) (T_SPACESHIP 16739) (T_SR 16740) ("^" 16741) (backup_fn_flags 17010) ("|" 16742)) 7353 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17012) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7354 (("$" 16635) (T_ABSTRACT 16812) (T_ARRAY 16813) (T_AS 16814) (T_BREAK 16815) (T_CALLABLE 16816) (T_CASE 16817) (T_CATCH 16818) (T_CLASS 16819) (T_CLASS_C 16820) (T_CLONE 16821) (T_CONST 16822) (T_CONTINUE 16823) (T_DECLARE 16824) (T_DEFAULT 16825) (T_DIR 16826) (T_DO 16827) (T_ECHO 16828) (T_ELSE 16829) (T_ELSEIF 16830) (T_EMPTY 16831) (T_ENDDECLARE 16832) (T_ENDFOR 16833) (T_ENDFOREACH 16834) (T_ENDIF 16835) (T_ENDSWITCH 16836) (T_ENDWHILE 16837) (T_ENUM 16838) (T_EVAL 16839) (T_EXIT 16840) (T_EXTENDS 16841) (T_FILE 16842) (T_FINAL 16843) (T_FINALLY 16844) (T_FN 16845) (T_FOR 16846) (T_FOREACH 16847) (T_FUNCTION 16848) (T_FUNC_C 16849) (T_GLOBAL 16850) (T_GOTO 16851) (T_IF 16852) (T_IMPLEMENTS 16853) (T_INCLUDE 16854) (T_INCLUDE_ONCE 16855) (T_INSTANCEOF 16856) (T_INSTEADOF 16857) (T_INTERFACE 16858) (T_ISSET 16859) (T_LINE 16860) (T_LIST 16861) (T_LOGICAL_AND 16862) (T_LOGICAL_OR 16863) (T_LOGICAL_XOR 16864) (T_MATCH 16865) (T_METHOD_C 16866) (T_NAMESPACE 16867) (T_NEW 16868) (T_NS_C 16869) (T_PRINT 16870) (T_PRIVATE 16871) (T_PROPERTY_C 16872) (T_PROTECTED 16873) (T_PUBLIC 16874) (T_READONLY 16875) (T_REQUIRE 16876) (T_REQUIRE_ONCE 16877) (T_RETURN 16878) (T_STATIC 16879) (T_STRING 16880) (T_SWITCH 16881) (T_THROW 16882) (T_TRAIT 16883) (T_TRAIT_C 16884) (T_TRY 16885) (T_UNSET 16886) (T_USE 16887) (T_VAR 16888) (T_VARIABLE 16683) (T_WHILE 16889) (T_YIELD 16890) (identifier 17014) (member_name 17015) (reserved_non_modifiers 16893) (semi_reserved 16894) (simple_variable 17016) ("{" 17017)) 7355 (("(" 16899) (argument_list 17020)) 7356 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 17018) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7357 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 17019)) 7358 ((T_FN 35) (T_FUNCTION 129) (fn 16697) #'16698 (inline_function 17024)) 7359 (("$" 16635) (T_STRING 17030) (T_VARIABLE 16683) (property_name 17037) (simple_variable 17032) ("{" 17033)) 7360 (("$" 16635) (T_STRING 17030) (T_VARIABLE 16683) (property_name 17031) (simple_variable 17032) ("{" 17033)) 7361 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 17028) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 7362 (("]" 17029)) 7363 (("(" 16899) (argument_list 17036)) 7364 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 17034) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7365 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 17035)) 7366 (("(" 16899) (argument_list 17038)) 7367 (("`" 17040)) 7368 (("]" 17042)) 7369 (("=" 17043)) 7370 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17044) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7371 (("%" 16713) ("*" 16714) ("+" 16715) ("-" 16716) ("." 16717) ("/" 16718) ("<" 16719) (">" 16720) ("?" 16721) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 16722) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 16723) (T_BOOLEAN_AND 16724) (T_BOOLEAN_OR 16725) (T_COALESCE 16726) (T_DOUBLE_ARROW 17047) (T_INSTANCEOF 16727) (T_IS_EQUAL 16728) (T_IS_GREATER_OR_EQUAL 16729) (T_IS_IDENTICAL 16730) (T_IS_NOT_EQUAL 16731) (T_IS_NOT_IDENTICAL 16732) (T_IS_SMALLER_OR_EQUAL 16733) (T_LOGICAL_AND 16734) (T_LOGICAL_OR 16735) (T_LOGICAL_XOR 16736) (T_POW 16737) (T_SL 16738) (T_SPACESHIP 16739) (T_SR 16740) ("^" 16741) ("|" 16742)) 7372 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17048) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7373 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 17057) (T_VARIABLE 2896) (encaps_var 2910)) 7374 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 17056) (T_VARIABLE 2896) (encaps_var 2901)) 7375 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17098) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7376 ((T_CLASS 17090)) 7377 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 17089) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 17070) (attribute 2958) (class_modifier 2939)) 7378 ((T_PAAMAYIM_NEKUDOTAYIM 17087)) 7379 (("(" 16954) (argument_list 17086)) 7380 ((T_NULLSAFE_OBJECT_OPERATOR 17077) (T_OBJECT_OPERATOR 17078) (T_PAAMAYIM_NEKUDOTAYIM 17079) ("[" 17080)) 7381 (("$" 16938) (T_STRING 16973) (T_VARIABLE 16940) (property_name 17085) (simple_variable 16975) ("{" 16976)) 7382 (("$" 16938) (T_STRING 16973) (T_VARIABLE 16940) (property_name 17084) (simple_variable 16975) ("{" 16976)) 7383 (("$" 16938) (T_VARIABLE 16940) (simple_variable 17083)) 7384 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 17081) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 7385 (("]" 17082)) 7386 (("$" 16938) (T_VARIABLE 16940) (simple_variable 17088)) 7387 (("(" 2960) (argument_list 2961) (ctor_arguments 17091)) 7388 ((T_EXTENDS 2963) (extends_from 17092)) 7389 ((T_IMPLEMENTS 2965) (implements_list 17093)) 7390 ((backup_doc_comment 17094)) 7391 (("{" 17095)) 7392 ((class_statement_list 17096)) 7393 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 17097)) 7394 (("%" 299) (")" 17099) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7395 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17101) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7396 (("%" 299) (")" 17102) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7397 (("{" 17103)) 7398 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 17104) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 7399 (("}" 17105)) 7400 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 17107) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 7401 ((")" 17108)) 7402 (("=" 17109)) 7403 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 17110) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 7404 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 17112) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7405 (("," 5192) (possible_comma 17113)) 7406 ((")" 17114)) 7407 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 17122) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 17123) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 17124) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7408 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 17127) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7409 (("," 1480) (possible_comma 17125)) 7410 ((")" 17126)) 7411 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17129) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7412 (("%" 299) (")" 17130) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7413 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17132) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7414 (("%" 299) (")" 17133) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7415 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 17140) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 7416 ((")" 17141)) 7417 (("%" 299) (")" 17146) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7418 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 17149) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7419 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 17150)) 7420 (("\"" 17152) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 7421 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 17162) (T_VARIABLE 2896) (encaps_var 2910)) 7422 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 17161) (T_VARIABLE 2896) (encaps_var 2901)) 7423 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17203) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7424 ((T_CLASS 17195)) 7425 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 17194) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 17175) (attribute 2958) (class_modifier 2939)) 7426 ((T_PAAMAYIM_NEKUDOTAYIM 17192)) 7427 (("(" 16542) (argument_list 17191)) 7428 ((T_NULLSAFE_OBJECT_OPERATOR 17182) (T_OBJECT_OPERATOR 17183) (T_PAAMAYIM_NEKUDOTAYIM 17184) ("[" 17185)) 7429 (("$" 16526) (T_STRING 16561) (T_VARIABLE 16528) (property_name 17190) (simple_variable 16563) ("{" 16564)) 7430 (("$" 16526) (T_STRING 16561) (T_VARIABLE 16528) (property_name 17189) (simple_variable 16563) ("{" 16564)) 7431 (("$" 16526) (T_VARIABLE 16528) (simple_variable 17188)) 7432 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 17186) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 7433 (("]" 17187)) 7434 (("$" 16526) (T_VARIABLE 16528) (simple_variable 17193)) 7435 (("(" 2960) (argument_list 2961) (ctor_arguments 17196)) 7436 ((T_EXTENDS 2963) (extends_from 17197)) 7437 ((T_IMPLEMENTS 2965) (implements_list 17198)) 7438 ((backup_doc_comment 17199)) 7439 (("{" 17200)) 7440 ((class_statement_list 17201)) 7441 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 17202)) 7442 (("%" 299) (")" 17204) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7443 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17206) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7444 (("%" 299) (")" 17207) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7445 (("{" 17208)) 7446 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 17209) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 7447 (("}" 17210)) 7448 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 17212) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 7449 ((")" 17213)) 7450 (("=" 17214)) 7451 (("!" 16221) ("\"" 16222) ("$" 16223) ("(" 16224) ("+" 16225) ("-" 16226) ("@" 16227) (T_ARRAY 16228) (T_ARRAY_CAST 16229) (T_ATTRIBUTE 128) (T_BOOL_CAST 16230) (T_CLASS_C 16231) (T_CLONE 16232) (T_CONSTANT_ENCAPSED_STRING 16233) (T_DEC 16234) (T_DIR 16235) (T_DNUMBER 16236) (T_DOUBLE_CAST 16237) (T_EMPTY 16238) (T_EVAL 16239) (T_EXIT 16240) (T_FILE 16241) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16242) (T_INC 16243) (T_INCLUDE 16244) (T_INCLUDE_ONCE 16245) (T_INT_CAST 16246) (T_ISSET 16247) (T_LINE 16248) (T_LIST 16249) (T_LNUMBER 16250) (T_MATCH 16251) (T_METHOD_C 16252) (T_NAME_FULLY_QUALIFIED 16253) (T_NAME_QUALIFIED 16254) (T_NAME_RELATIVE 16255) (T_NEW 16256) (T_NS_C 16257) (T_OBJECT_CAST 16258) (T_PRINT 16259) (T_PROPERTY_C 16260) (T_READONLY 16261) (T_REQUIRE 16262) (T_REQUIRE_ONCE 16263) (T_START_HEREDOC 16264) (T_STATIC 16265) (T_STRING 16266) (T_STRING_CAST 16267) (T_THROW 16268) (T_TRAIT_C 16269) (T_UNSET_CAST 16270) (T_VARIABLE 16271) (T_YIELD 16272) (T_YIELD_FROM 16273) ("[" 16274) ("`" 16275) (array_object_dereferenceable 16276) (attribute 133) (attributes 16277) (callable_expr 16278) (callable_variable 16279) (class_constant 16280) (class_name 16281) (constant 16282) (dereferenceable_scalar 16283) (expr 17215) (fn 16285) (fully_dereferenceable 105) #'16286 (function_call 16287) (inline_function 16288) (internal_functions_in_yacc 16289) (match 16290) (name 16291) (new_dereferenceable 16292) (new_non_dereferenceable 16293) (scalar 16294) (simple_variable 16295) (static_member 16296) (variable 16297) (variable_class_name 16298) ("~" 16299)) 7452 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 17217) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7453 (("," 5192) (possible_comma 17218)) 7454 ((")" 17219)) 7455 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 17227) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 17228) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 17229) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7456 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 17232) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7457 (("," 1480) (possible_comma 17230)) 7458 ((")" 17231)) 7459 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17234) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7460 (("%" 299) (")" 17235) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7461 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17237) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7462 (("%" 299) (")" 17238) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7463 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 17245) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 7464 ((")" 17246)) 7465 (("%" 299) (")" 17251) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7466 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 17254) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7467 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 17255)) 7468 (("\"" 17257) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 7469 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17260) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17261) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17262) (T_DEC 22) (T_DECLARE 17263) (T_DIR 24) (T_DNUMBER 25) (T_DO 17264) (T_DOUBLE_CAST 27) (T_ECHO 17265) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17266) (T_FOREACH 17267) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17268) (T_GOTO 17269) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17270) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17271) (T_START_HEREDOC 69) (T_STATIC 17272) (T_STRING 17273) (T_STRING_CAST 72) (T_SWITCH 17274) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17275) (T_UNSET 17276) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17277) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17278) (alt_if_stmt_without_else 17279) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17280) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17281) (if_stmt_without_else 17282) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17283) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17284) ("~" 127)) 7470 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25775) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7471 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25773) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7472 (("(" 25764)) 7473 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25758) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 7474 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 25756) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7475 (("(" 25743)) 7476 (("(" 25728)) 7477 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 25726) (simple_variable 18987)) 7478 ((T_STRING 25724)) 7479 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25722) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7480 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 25720)) 7481 ((":" 25719)) 7482 (("(" 25701)) 7483 (("{" 25684)) 7484 (("(" 25679)) 7485 (("(" 25670)) 7486 ((T_ELSE 17290) (T_ELSEIF 17291) (T_ENDIF 17292)) 7487 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 17289) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 7488 ((T_ELSE 17287) (T_ELSEIF 217)) 7489 ((inner_statement_list 17285)) 7490 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17286) ("~" 127)) 7491 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17260) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17261) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17262) (T_DEC 22) (T_DECLARE 17263) (T_DIR 24) (T_DNUMBER 25) (T_DO 17264) (T_DOUBLE_CAST 27) (T_ECHO 17265) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17266) (T_FOREACH 17267) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17268) (T_GOTO 17269) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17270) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17271) (T_START_HEREDOC 69) (T_STATIC 17272) (T_STRING 17273) (T_STRING_CAST 72) (T_SWITCH 17274) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17275) (T_UNSET 17276) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17277) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17278) (alt_if_stmt_without_else 17279) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17280) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17281) (if_stmt_without_else 17282) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17288) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17284) ("~" 127)) 7492 ((":" 25666)) 7493 (("(" 17294)) 7494 ((";" 17293)) 7495 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17295) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7496 (("%" 299) (")" 17296) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7497 ((":" 17297)) 7498 ((inner_statement_list 17298)) 7499 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17299) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17300) (T_CLASS 17301) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17302) (T_DEC 22) (T_DECLARE 17303) (T_DIR 24) (T_DNUMBER 25) (T_DO 17304) (T_DOUBLE_CAST 27) (T_ECHO 17305) (T_EMPTY 29) (T_ENUM 17306) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17307) (T_FOREACH 17308) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17309) (T_GOTO 17310) (T_HALT_COMPILER 17311) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17313) (T_INTERFACE 17314) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17315) (T_START_HEREDOC 69) (T_STATIC 17316) (T_STRING 17317) (T_STRING_CAST 72) (T_SWITCH 17318) (T_THROW 74) (T_TRAIT 17319) (T_TRAIT_C 76) (T_TRY 17320) (T_UNSET 17321) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17322) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17323) (alt_if_stmt_without_else 17324) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17325) (attributes 17326) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17327) (class_modifier 97) (class_modifiers 17328) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17329) (expr 17330) (fn 104) (fully_dereferenceable 105) #'17331 (function_call 107) (function_declaration_statement 17332) (if_stmt 17333) (if_stmt_without_else 17334) (inline_function 111) (inner_statement 17335) (interface_declaration_statement 17336) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17337) (static_member 121) (trait_declaration_statement 17338) (variable 124) (variable_class_name 125) ("{" 17339) ("~" 127)) 7500 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25664) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7501 ((T_STRING 25657)) 7502 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25655) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7503 (("(" 25646)) 7504 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25640) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 7505 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 25638) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7506 ((T_STRING 25631)) 7507 (("(" 25618)) 7508 (("(" 25603)) 7509 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 25601) (simple_variable 18987)) 7510 ((T_STRING 25599)) 7511 (("(" 25596)) 7512 (("(" 25592)) 7513 ((T_STRING 25586)) 7514 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25584) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7515 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 25582)) 7516 ((":" 25581)) 7517 (("(" 25563)) 7518 ((T_STRING 25558)) 7519 (("{" 25541)) 7520 (("(" 25536)) 7521 (("(" 25527)) 7522 ((T_ELSE 17381) (T_ELSEIF 17291) (T_ENDIF 17382)) 7523 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 17301) (T_ENUM 17306) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 17314) (T_READONLY 2931) (T_STATIC 17376) (T_TRAIT 17319) (attribute 17377) (attributed_statement 17378) (class_declaration_statement 17327) (class_modifier 97) (class_modifiers 17328) (enum_declaration_statement 17329) (fn 104) #'17331 (function_declaration_statement 17332) (inline_function 17379) (interface_declaration_statement 17336) (trait_declaration_statement 17338)) 7524 ((T_ABSTRACT 10) (T_CLASS 17367) (T_FINAL 34) (T_READONLY 2931) (class_modifier 17368)) 7525 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 17366) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 7526 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17349) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17350) (ampersand 17351) (returns_ref 17352)) 7527 ((T_ELSE 17342) (T_ELSEIF 17343)) 7528 ((inner_statement_list 17340)) 7529 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17341) ("~" 127)) 7530 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17299) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17300) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17302) (T_DEC 22) (T_DECLARE 17303) (T_DIR 24) (T_DNUMBER 25) (T_DO 17304) (T_DOUBLE_CAST 27) (T_ECHO 17305) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17307) (T_FOREACH 17308) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17309) (T_GOTO 17310) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17313) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17315) (T_START_HEREDOC 69) (T_STATIC 17316) (T_STRING 17317) (T_STRING_CAST 72) (T_SWITCH 17318) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17320) (T_UNSET 17321) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17322) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17323) (alt_if_stmt_without_else 17324) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17330) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17333) (if_stmt_without_else 17334) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17348) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17339) ("~" 127)) 7531 (("(" 17344)) 7532 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17345) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7533 (("%" 299) (")" 17346) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7534 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17299) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17300) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17302) (T_DEC 22) (T_DECLARE 17303) (T_DIR 24) (T_DNUMBER 25) (T_DO 17304) (T_DOUBLE_CAST 27) (T_ECHO 17305) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17307) (T_FOREACH 17308) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17309) (T_GOTO 17310) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17313) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17315) (T_START_HEREDOC 69) (T_STATIC 17316) (T_STRING 17317) (T_STRING_CAST 72) (T_SWITCH 17318) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17320) (T_UNSET 17321) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17322) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17323) (alt_if_stmt_without_else 17324) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17330) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17333) (if_stmt_without_else 17334) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17347) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17339) ("~" 127)) 7535 ((T_READONLY 17353) (T_STRING 17354) (backup_doc_comment 141) (function_name 17355)) 7536 ((backup_doc_comment 17356)) 7537 (("(" 17357)) 7538 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 17358)) 7539 ((")" 17359)) 7540 ((":" 166) (return_type 17360)) 7541 ((backup_fn_flags 17361)) 7542 (("{" 17362)) 7543 ((inner_statement_list 17363)) 7544 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17364) ("~" 127)) 7545 ((backup_fn_flags 17365)) 7546 ((T_STRING 17369)) 7547 ((T_EXTENDS 2963) (extends_from 17370)) 7548 ((T_IMPLEMENTS 2965) (implements_list 17371)) 7549 ((backup_doc_comment 17372)) 7550 (("{" 17373)) 7551 ((class_statement_list 17374)) 7552 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 17375)) 7553 ((T_FN 35) (T_FUNCTION 129) (fn 104) #'136 (inline_function 17380)) 7554 ((":" 17384)) 7555 ((";" 17383)) 7556 ((inner_statement_list 17385)) 7557 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 17393) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 7558 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25525) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7559 ((T_STRING 25518)) 7560 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25516) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7561 (("(" 25507)) 7562 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25501) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 7563 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 25499) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7564 ((";" 25498)) 7565 ((T_STRING 25491)) 7566 (("(" 25478)) 7567 (("(" 25463)) 7568 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 25461) (simple_variable 18987)) 7569 ((T_STRING 25459)) 7570 (("(" 25456)) 7571 ((T_STRING 25450)) 7572 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25448) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7573 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 25446)) 7574 ((":" 25445)) 7575 (("(" 25427)) 7576 ((T_STRING 25422)) 7577 (("{" 25405)) 7578 (("(" 25400)) 7579 (("(" 17460)) 7580 ((T_ELSE 17453) (T_ELSEIF 17291) (T_ENDIF 17454)) 7581 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 17388) (T_ENUM 17394) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 17401) (T_READONLY 2931) (T_STATIC 17376) (T_TRAIT 17406) (attribute 17377) (attributed_statement 17452) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (enum_declaration_statement 17416) (fn 104) #'17418 (function_declaration_statement 17419) (inline_function 17379) (interface_declaration_statement 17423) (trait_declaration_statement 17425)) 7582 ((T_ABSTRACT 10) (T_CLASS 17444) (T_FINAL 34) (T_READONLY 2931) (class_modifier 17368)) 7583 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 17443) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 7584 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17349) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17350) (ampersand 17351) (returns_ref 17431)) 7585 ((T_ELSE 17429) (T_ELSEIF 17343)) 7586 ((inner_statement_list 17427)) 7587 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17428) ("~" 127)) 7588 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17417) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17430) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 7589 ((T_READONLY 17353) (T_STRING 17354) (backup_doc_comment 141) (function_name 17432)) 7590 ((backup_doc_comment 17433)) 7591 (("(" 17434)) 7592 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 17435)) 7593 ((")" 17436)) 7594 ((":" 166) (return_type 17437)) 7595 ((backup_fn_flags 17438)) 7596 (("{" 17439)) 7597 ((inner_statement_list 17440)) 7598 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17441) ("~" 127)) 7599 ((backup_fn_flags 17442)) 7600 ((T_STRING 17445)) 7601 ((T_EXTENDS 2963) (extends_from 17446)) 7602 ((T_IMPLEMENTS 2965) (implements_list 17447)) 7603 ((backup_doc_comment 17448)) 7604 (("{" 17449)) 7605 ((class_statement_list 17450)) 7606 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 17451)) 7607 ((":" 17456)) 7608 ((";" 17455)) 7609 ((inner_statement_list 17457)) 7610 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 17458) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 7611 ((";" 17459)) 7612 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17461) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7613 (("%" 299) (")" 17462) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7614 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 17463) (";" 17386) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17417) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17464) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 17465) ("{" 17426) ("~" 127)) 7615 ((inner_statement_list 17466)) 7616 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 17474) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 7617 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25398) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7618 ((T_STRING 25391)) 7619 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25389) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7620 (("(" 25380)) 7621 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25374) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 7622 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 25372) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7623 ((";" 25371)) 7624 ((T_STRING 25364)) 7625 (("(" 25351)) 7626 (("(" 25336)) 7627 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 25334) (simple_variable 18987)) 7628 ((T_STRING 25332)) 7629 (("(" 25329)) 7630 (("(" 25325)) 7631 ((T_STRING 25319)) 7632 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25317) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7633 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 25315)) 7634 ((":" 25314)) 7635 (("(" 25296)) 7636 ((T_STRING 25291)) 7637 (("{" 25274)) 7638 (("(" 25269)) 7639 (("(" 25260)) 7640 ((T_ELSE 25253) (T_ELSEIF 17291) (T_ENDIF 25254)) 7641 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 17469) (T_ENUM 17475) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 17483) (T_READONLY 2931) (T_STATIC 17376) (T_TRAIT 17488) (attribute 17377) (attributed_statement 25252) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (enum_declaration_statement 17498) (fn 104) #'17500 (function_declaration_statement 17501) (inline_function 17379) (interface_declaration_statement 17505) (trait_declaration_statement 17507)) 7642 ((T_ABSTRACT 10) (T_CLASS 25244) (T_FINAL 34) (T_READONLY 2931) (class_modifier 17368)) 7643 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 25243) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 7644 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17349) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17350) (ampersand 17351) (returns_ref 25231)) 7645 ((T_ELSE 17511) (T_ELSEIF 17512)) 7646 ((inner_statement_list 17509)) 7647 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17510) ("~" 127)) 7648 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17499) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25230) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 7649 (("(" 17513)) 7650 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17514) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7651 (("%" 299) (")" 17515) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7652 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17516) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17517) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17518) (T_DEC 22) (T_DECLARE 17519) (T_DIR 24) (T_DNUMBER 25) (T_DO 17520) (T_DOUBLE_CAST 27) (T_ECHO 17521) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17522) (T_FOREACH 17523) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17524) (T_GOTO 17525) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17526) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17527) (T_START_HEREDOC 69) (T_STATIC 17528) (T_STRING 17529) (T_STRING_CAST 72) (T_SWITCH 17530) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17531) (T_UNSET 17532) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17533) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17534) (alt_if_stmt_without_else 17535) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17536) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17537) (if_stmt_without_else 17538) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17539) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17540) ("~" 127)) 7653 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25228) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7654 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25226) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7655 (("(" 25217)) 7656 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25211) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 7657 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 25209) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7658 (("(" 25196)) 7659 (("(" 25181)) 7660 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 25179) (simple_variable 18987)) 7661 ((T_STRING 25177)) 7662 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25175) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7663 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 25173)) 7664 ((":" 25172)) 7665 (("(" 17646)) 7666 (("{" 17619)) 7667 (("(" 17562)) 7668 (("(" 17553)) 7669 ((T_ELSE 17546) (T_ELSEIF 17291) (T_ENDIF 17547)) 7670 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 17545) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 7671 ((T_ELSE 17543) (T_ELSEIF 17512)) 7672 ((inner_statement_list 17541)) 7673 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17542) ("~" 127)) 7674 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17516) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17517) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17518) (T_DEC 22) (T_DECLARE 17519) (T_DIR 24) (T_DNUMBER 25) (T_DO 17520) (T_DOUBLE_CAST 27) (T_ECHO 17521) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17522) (T_FOREACH 17523) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17524) (T_GOTO 17525) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17526) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17527) (T_START_HEREDOC 69) (T_STATIC 17528) (T_STRING 17529) (T_STRING_CAST 72) (T_SWITCH 17530) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17531) (T_UNSET 17532) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17533) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17534) (alt_if_stmt_without_else 17535) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17536) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17537) (if_stmt_without_else 17538) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17544) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17540) ("~" 127)) 7675 ((":" 17549)) 7676 ((";" 17548)) 7677 ((inner_statement_list 17550)) 7678 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 17551) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 7679 ((";" 17552)) 7680 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17554) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7681 (("%" 299) (")" 17555) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7682 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 17556) (";" 17516) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17517) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17518) (T_DEC 22) (T_DECLARE 17519) (T_DIR 24) (T_DNUMBER 25) (T_DO 17520) (T_DOUBLE_CAST 27) (T_ECHO 17521) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17522) (T_FOREACH 17523) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17524) (T_GOTO 17525) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17526) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17527) (T_START_HEREDOC 69) (T_STATIC 17528) (T_STRING 17529) (T_STRING_CAST 72) (T_SWITCH 17530) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17531) (T_UNSET 17532) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17533) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17534) (alt_if_stmt_without_else 17535) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17536) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17537) (if_stmt_without_else 17538) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17557) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 17558) ("{" 17540) ("~" 127)) 7683 ((inner_statement_list 17559)) 7684 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 17560) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 7685 ((";" 17561)) 7686 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 17575) (variable 17576) (variable_class_name 17577)) 7687 (("$" 17563) (T_VARIABLE 17565) (simple_variable 17615) ("{" 17616)) 7688 (("(" 17581) (argument_list 17614)) 7689 ((T_NULLSAFE_OBJECT_OPERATOR 17600) (T_OBJECT_OPERATOR 17601) ("[" 17602)) 7690 (("(" 17581) (argument_list 17599)) 7691 ((T_PAAMAYIM_NEKUDOTAYIM 17595)) 7692 (("(" 17581) (argument_list 17594)) 7693 (("," 17589) (possible_comma 17590)) 7694 ((T_PAAMAYIM_NEKUDOTAYIM 17578)) 7695 (("$" 17563) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 17565) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 17579) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 17580) ("{" 1642)) 7696 (("(" 17581) (argument_list 17582)) 7697 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 17583) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 17584) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 17585) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7698 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 17588) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 7699 (("," 1480) (possible_comma 17586)) 7700 ((")" 17587)) 7701 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17593) (variable 17576) (variable_class_name 17577)) 7702 ((")" 17591)) 7703 ((";" 17592)) 7704 (("$" 17563) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 17565) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 17596) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 17597) ("{" 1658)) 7705 (("(" 17581) (argument_list 17598)) 7706 (("$" 17563) (T_STRING 17605) (T_VARIABLE 17565) (property_name 17612) (simple_variable 17607) ("{" 17608)) 7707 (("$" 17563) (T_STRING 17605) (T_VARIABLE 17565) (property_name 17606) (simple_variable 17607) ("{" 17608)) 7708 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 17603) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 7709 (("]" 17604)) 7710 (("(" 17581) (argument_list 17611)) 7711 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 17609) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7712 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 17610)) 7713 (("(" 17581) (argument_list 17613)) 7714 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 17617) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 7715 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 17618)) 7716 ((inner_statement_list 17620)) 7717 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17621) ("~" 127)) 7718 ((catch_list 17622)) 7719 ((T_CATCH 17623) (T_FINALLY 17624) (finally_statement 17625)) 7720 (("(" 17629)) 7721 (("{" 17626)) 7722 ((inner_statement_list 17627)) 7723 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17628) ("~" 127)) 7724 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 17635) (class_name 17636) (name 17637)) 7725 ((T_VARIABLE 17638) (optional_variable 17639) ("|" 17640)) 7726 ((")" 17642)) 7727 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (class_name 17641) (name 17637)) 7728 (("{" 17643)) 7729 ((inner_statement_list 17644)) 7730 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17645) ("~" 127)) 7731 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17647) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7732 (("%" 299) (")" 17648) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7733 ((":" 17649) (switch_case_list 17650) ("{" 17651)) 7734 ((";" 25165) (case_list 25166)) 7735 ((";" 17652) (case_list 17653)) 7736 ((case_list 25163)) 7737 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 17656)) 7738 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 25160) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 7739 ((":" 17657) (";" 17658) (case_separator 17659)) 7740 ((inner_statement_list 17660)) 7741 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17661) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17662) (T_CLASS 17663) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17664) (T_DEC 22) (T_DECLARE 17665) (T_DIR 24) (T_DNUMBER 25) (T_DO 17666) (T_DOUBLE_CAST 27) (T_ECHO 17667) (T_EMPTY 29) (T_ENUM 17668) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17669) (T_FOREACH 17670) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17671) (T_GOTO 17672) (T_HALT_COMPILER 17673) (T_IF 17674) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17675) (T_INTERFACE 17676) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17677) (T_START_HEREDOC 69) (T_STATIC 17678) (T_STRING 17679) (T_STRING_CAST 72) (T_SWITCH 17680) (T_THROW 74) (T_TRAIT 17681) (T_TRAIT_C 76) (T_TRY 17682) (T_UNSET 17683) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17684) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17685) (alt_if_stmt_without_else 17686) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17687) (attributes 17688) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17689) (class_modifier 97) (class_modifiers 17690) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17691) (expr 17692) (fn 104) (fully_dereferenceable 105) #'17693 (function_call 107) (function_declaration_statement 17694) (if_stmt 17695) (if_stmt_without_else 17696) (inline_function 111) (inner_statement 17697) (interface_declaration_statement 17698) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17699) (static_member 121) (trait_declaration_statement 17700) (variable 124) (variable_class_name 125) ("{" 17701) ("~" 127)) 7742 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25158) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7743 ((T_STRING 25151)) 7744 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25149) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7745 (("(" 25140)) 7746 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25134) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 7747 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 25132) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7748 ((T_STRING 25125)) 7749 (("(" 25112)) 7750 (("(" 25097)) 7751 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 25095) (simple_variable 18987)) 7752 ((T_STRING 25093)) 7753 (("(" 25090)) 7754 (("(" 25086)) 7755 ((T_STRING 25080)) 7756 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 25078) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7757 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 25076)) 7758 ((":" 25075)) 7759 (("(" 25057)) 7760 ((T_STRING 25052)) 7761 (("{" 25035)) 7762 (("(" 25030)) 7763 (("(" 25021)) 7764 ((T_ELSE 25014) (T_ELSEIF 17291) (T_ENDIF 25015)) 7765 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 17663) (T_ENUM 17668) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 17676) (T_READONLY 2931) (T_STATIC 17376) (T_TRAIT 17681) (attribute 17377) (attributed_statement 25013) (class_declaration_statement 17689) (class_modifier 97) (class_modifiers 17690) (enum_declaration_statement 17691) (fn 104) #'17693 (function_declaration_statement 17694) (inline_function 17379) (interface_declaration_statement 17698) (trait_declaration_statement 17700)) 7766 ((T_ABSTRACT 10) (T_CLASS 25005) (T_FINAL 34) (T_READONLY 2931) (class_modifier 17368)) 7767 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 25004) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 7768 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17349) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17350) (ampersand 17351) (returns_ref 24992)) 7769 ((T_ELSE 17704) (T_ELSEIF 17705)) 7770 ((inner_statement_list 17702)) 7771 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17703) ("~" 127)) 7772 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17661) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17662) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17664) (T_DEC 22) (T_DECLARE 17665) (T_DIR 24) (T_DNUMBER 25) (T_DO 17666) (T_DOUBLE_CAST 27) (T_ECHO 17667) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17669) (T_FOREACH 17670) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17671) (T_GOTO 17672) (T_IF 17674) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17675) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17677) (T_START_HEREDOC 69) (T_STATIC 17678) (T_STRING 17679) (T_STRING_CAST 72) (T_SWITCH 17680) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17682) (T_UNSET 17683) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17684) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17685) (alt_if_stmt_without_else 17686) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17692) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17695) (if_stmt_without_else 17696) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 24991) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17701) ("~" 127)) 7773 (("(" 17706)) 7774 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17707) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7775 (("%" 299) (")" 17708) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7776 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17709) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17710) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17711) (T_DEC 22) (T_DECLARE 17712) (T_DIR 24) (T_DNUMBER 25) (T_DO 17713) (T_DOUBLE_CAST 27) (T_ECHO 17714) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17715) (T_FOREACH 17716) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17717) (T_GOTO 17718) (T_IF 17674) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17719) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17720) (T_START_HEREDOC 69) (T_STATIC 17721) (T_STRING 17722) (T_STRING_CAST 72) (T_SWITCH 17723) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17724) (T_UNSET 17725) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17726) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17727) (alt_if_stmt_without_else 17728) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17729) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17730) (if_stmt_without_else 17731) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17732) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17733) ("~" 127)) 7777 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 24989) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7778 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 24987) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7779 (("(" 24978)) 7780 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 24972) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 7781 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 24970) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7782 (("(" 24957)) 7783 (("(" 24942)) 7784 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 24940) (simple_variable 18987)) 7785 ((T_STRING 24938)) 7786 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 24936) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7787 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 24934)) 7788 ((":" 24933)) 7789 (("(" 17777)) 7790 (("{" 17760)) 7791 (("(" 17755)) 7792 (("(" 17746)) 7793 ((T_ELSE 17739) (T_ELSEIF 17291) (T_ENDIF 17740)) 7794 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 17738) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 7795 ((T_ELSE 17736) (T_ELSEIF 17705)) 7796 ((inner_statement_list 17734)) 7797 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17735) ("~" 127)) 7798 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17709) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17710) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17711) (T_DEC 22) (T_DECLARE 17712) (T_DIR 24) (T_DNUMBER 25) (T_DO 17713) (T_DOUBLE_CAST 27) (T_ECHO 17714) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17715) (T_FOREACH 17716) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17717) (T_GOTO 17718) (T_IF 17674) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17719) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17720) (T_START_HEREDOC 69) (T_STATIC 17721) (T_STRING 17722) (T_STRING_CAST 72) (T_SWITCH 17723) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17724) (T_UNSET 17725) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17726) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17727) (alt_if_stmt_without_else 17728) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17729) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17730) (if_stmt_without_else 17731) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17737) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17733) ("~" 127)) 7799 ((":" 17742)) 7800 ((";" 17741)) 7801 ((inner_statement_list 17743)) 7802 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 17744) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 7803 ((";" 17745)) 7804 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17747) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7805 (("%" 299) (")" 17748) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7806 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 17749) (";" 17709) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17710) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17711) (T_DEC 22) (T_DECLARE 17712) (T_DIR 24) (T_DNUMBER 25) (T_DO 17713) (T_DOUBLE_CAST 27) (T_ECHO 17714) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17715) (T_FOREACH 17716) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17717) (T_GOTO 17718) (T_IF 17674) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17719) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17720) (T_START_HEREDOC 69) (T_STATIC 17721) (T_STRING 17722) (T_STRING_CAST 72) (T_SWITCH 17723) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17724) (T_UNSET 17725) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17726) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17727) (alt_if_stmt_without_else 17728) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17729) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17730) (if_stmt_without_else 17731) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17750) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 17751) ("{" 17733) ("~" 127)) 7807 ((inner_statement_list 17752)) 7808 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 17753) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 7809 ((";" 17754)) 7810 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 17756) (variable 17576) (variable_class_name 17577)) 7811 (("," 17589) (possible_comma 17757)) 7812 ((")" 17758)) 7813 ((";" 17759)) 7814 ((inner_statement_list 17761)) 7815 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17762) ("~" 127)) 7816 ((catch_list 17763)) 7817 ((T_CATCH 17764) (T_FINALLY 17765) (finally_statement 17766)) 7818 (("(" 17770)) 7819 (("{" 17767)) 7820 ((inner_statement_list 17768)) 7821 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17769) ("~" 127)) 7822 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 17771) (class_name 17636) (name 17637)) 7823 ((T_VARIABLE 17638) (optional_variable 17772) ("|" 17640)) 7824 ((")" 17773)) 7825 (("{" 17774)) 7826 ((inner_statement_list 17775)) 7827 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17776) ("~" 127)) 7828 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17778) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7829 (("%" 299) (")" 17779) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7830 ((":" 17780) (switch_case_list 17781) ("{" 17782)) 7831 ((";" 17788) (case_list 17789)) 7832 ((";" 17783) (case_list 17784)) 7833 ((case_list 17786)) 7834 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 17785)) 7835 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 17787)) 7836 ((case_list 24930)) 7837 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 17792)) 7838 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22918) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 7839 ((":" 17794) (";" 17795) (case_separator 17796)) 7840 ((";" 17793)) 7841 ((inner_statement_list 17797)) 7842 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17798) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17799) (T_CLASS 17800) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17801) (T_DEC 22) (T_DECLARE 17802) (T_DIR 24) (T_DNUMBER 25) (T_DO 17803) (T_DOUBLE_CAST 27) (T_ECHO 17804) (T_EMPTY 29) (T_ENUM 17805) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17806) (T_FOREACH 17807) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17808) (T_GOTO 17809) (T_HALT_COMPILER 17810) (T_IF 17811) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17812) (T_INTERFACE 17813) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17814) (T_START_HEREDOC 69) (T_STATIC 17815) (T_STRING 17816) (T_STRING_CAST 72) (T_SWITCH 17817) (T_THROW 74) (T_TRAIT 17818) (T_TRAIT_C 76) (T_TRY 17819) (T_UNSET 17820) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17821) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17822) (alt_if_stmt_without_else 17823) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17824) (attributes 17825) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17826) (class_modifier 97) (class_modifiers 17827) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17828) (expr 17829) (fn 104) (fully_dereferenceable 105) #'17830 (function_call 107) (function_declaration_statement 17831) (if_stmt 17832) (if_stmt_without_else 17833) (inline_function 111) (inner_statement 17834) (interface_declaration_statement 17835) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17836) (static_member 121) (trait_declaration_statement 17837) (variable 124) (variable_class_name 125) ("{" 17838) ("~" 127)) 7843 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 22853) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7844 ((T_STRING 22846)) 7845 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 22844) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7846 (("(" 22835)) 7847 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 22829) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 7848 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 22827) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7849 ((T_STRING 22820)) 7850 (("(" 22807)) 7851 (("(" 22792)) 7852 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 22790) (simple_variable 18987)) 7853 ((T_STRING 22788)) 7854 (("(" 22785)) 7855 (("(" 22781)) 7856 ((T_STRING 22775)) 7857 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 22773) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7858 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 22771)) 7859 ((":" 22770)) 7860 (("(" 22752)) 7861 ((T_STRING 22747)) 7862 (("{" 22730)) 7863 (("(" 22725)) 7864 (("(" 22716)) 7865 ((T_ELSE 22709) (T_ELSEIF 17291) (T_ENDIF 22710)) 7866 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 17800) (T_ENUM 17805) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 17813) (T_READONLY 2931) (T_STATIC 17376) (T_TRAIT 17818) (attribute 17377) (attributed_statement 22708) (class_declaration_statement 17826) (class_modifier 97) (class_modifiers 17827) (enum_declaration_statement 17828) (fn 104) #'17830 (function_declaration_statement 17831) (inline_function 17379) (interface_declaration_statement 17835) (trait_declaration_statement 17837)) 7867 ((T_ABSTRACT 10) (T_CLASS 22700) (T_FINAL 34) (T_READONLY 2931) (class_modifier 17368)) 7868 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 22699) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 7869 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17349) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17350) (ampersand 17351) (returns_ref 22687)) 7870 ((T_ELSE 17841) (T_ELSEIF 17842)) 7871 ((inner_statement_list 17839)) 7872 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17840) ("~" 127)) 7873 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17798) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17799) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17801) (T_DEC 22) (T_DECLARE 17802) (T_DIR 24) (T_DNUMBER 25) (T_DO 17803) (T_DOUBLE_CAST 27) (T_ECHO 17804) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17806) (T_FOREACH 17807) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17808) (T_GOTO 17809) (T_IF 17811) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17812) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17814) (T_START_HEREDOC 69) (T_STATIC 17815) (T_STRING 17816) (T_STRING_CAST 72) (T_SWITCH 17817) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17819) (T_UNSET 17820) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17821) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17822) (alt_if_stmt_without_else 17823) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17829) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17832) (if_stmt_without_else 17833) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 22686) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17838) ("~" 127)) 7874 (("(" 17843)) 7875 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17844) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7876 (("%" 299) (")" 17845) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7877 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17846) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17847) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17848) (T_DEC 22) (T_DECLARE 17849) (T_DIR 24) (T_DNUMBER 25) (T_DO 17850) (T_DOUBLE_CAST 27) (T_ECHO 17851) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17852) (T_FOREACH 17853) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17854) (T_GOTO 17855) (T_IF 17811) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17856) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17857) (T_START_HEREDOC 69) (T_STATIC 17858) (T_STRING 17859) (T_STRING_CAST 72) (T_SWITCH 17860) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17861) (T_UNSET 17862) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17863) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17864) (alt_if_stmt_without_else 17865) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17866) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17867) (if_stmt_without_else 17868) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17869) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17870) ("~" 127)) 7878 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 22684) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7879 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 22682) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7880 (("(" 22673)) 7881 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 22667) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 7882 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 22665) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7883 (("(" 22652)) 7884 (("(" 18995)) 7885 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 18986) (simple_variable 18987)) 7886 ((T_STRING 18981)) 7887 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 18979) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 7888 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 17936)) 7889 ((":" 17932)) 7890 (("(" 17914)) 7891 (("{" 17897)) 7892 (("(" 17892)) 7893 (("(" 17883)) 7894 ((T_ELSE 17876) (T_ELSEIF 17291) (T_ENDIF 17877)) 7895 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 17875) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 7896 ((T_ELSE 17873) (T_ELSEIF 17842)) 7897 ((inner_statement_list 17871)) 7898 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17872) ("~" 127)) 7899 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17846) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17847) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17848) (T_DEC 22) (T_DECLARE 17849) (T_DIR 24) (T_DNUMBER 25) (T_DO 17850) (T_DOUBLE_CAST 27) (T_ECHO 17851) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17852) (T_FOREACH 17853) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17854) (T_GOTO 17855) (T_IF 17811) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17856) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17857) (T_START_HEREDOC 69) (T_STATIC 17858) (T_STRING 17859) (T_STRING_CAST 72) (T_SWITCH 17860) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17861) (T_UNSET 17862) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17863) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17864) (alt_if_stmt_without_else 17865) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17866) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17867) (if_stmt_without_else 17868) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17874) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17870) ("~" 127)) 7900 ((":" 17879)) 7901 ((";" 17878)) 7902 ((inner_statement_list 17880)) 7903 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 17881) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 7904 ((";" 17882)) 7905 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17884) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7906 (("%" 299) (")" 17885) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7907 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 17886) (";" 17846) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17847) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17848) (T_DEC 22) (T_DECLARE 17849) (T_DIR 24) (T_DNUMBER 25) (T_DO 17850) (T_DOUBLE_CAST 27) (T_ECHO 17851) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17852) (T_FOREACH 17853) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17854) (T_GOTO 17855) (T_IF 17811) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17856) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17857) (T_START_HEREDOC 69) (T_STATIC 17858) (T_STRING 17859) (T_STRING_CAST 72) (T_SWITCH 17860) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17861) (T_UNSET 17862) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17863) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17864) (alt_if_stmt_without_else 17865) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17866) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17867) (if_stmt_without_else 17868) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17887) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 17888) ("{" 17870) ("~" 127)) 7908 ((inner_statement_list 17889)) 7909 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 17890) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 7910 ((";" 17891)) 7911 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 17893) (variable 17576) (variable_class_name 17577)) 7912 (("," 17589) (possible_comma 17894)) 7913 ((")" 17895)) 7914 ((";" 17896)) 7915 ((inner_statement_list 17898)) 7916 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17899) ("~" 127)) 7917 ((catch_list 17900)) 7918 ((T_CATCH 17901) (T_FINALLY 17902) (finally_statement 17903)) 7919 (("(" 17907)) 7920 (("{" 17904)) 7921 ((inner_statement_list 17905)) 7922 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17906) ("~" 127)) 7923 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 17908) (class_name 17636) (name 17637)) 7924 ((T_VARIABLE 17638) (optional_variable 17909) ("|" 17640)) 7925 ((")" 17910)) 7926 (("{" 17911)) 7927 ((inner_statement_list 17912)) 7928 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 17913) ("~" 127)) 7929 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 17915) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7930 (("%" 299) (")" 17916) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 7931 ((":" 17917) (switch_case_list 17918) ("{" 17919)) 7932 ((";" 17925) (case_list 17926)) 7933 ((";" 17920) (case_list 17921)) 7934 ((case_list 17923)) 7935 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 17922)) 7936 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 17924)) 7937 ((case_list 17929)) 7938 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 17927)) 7939 ((";" 17928)) 7940 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 17930)) 7941 ((";" 17931)) 7942 (("=" 17940)) 7943 (("," 17937) (";" 17938)) 7944 ((T_VARIABLE 17933) (static_var 17939)) 7945 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18004) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7946 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18977) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7947 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 18975) (encaps_var 5237)) 7948 (("$" 17943) (T_VARIABLE 17991) (simple_variable 18971) ("{" 18972)) 7949 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 18969) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 7950 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18968) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7951 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18967) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7952 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18966) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7953 (("(" 18963)) 7954 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18962) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7955 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18961) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7956 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18960) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7957 (("\"" 1520) ("$" 18246) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 18247) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 18248) ("[" 1542) (array_object_dereferenceable 18249) (callable_expr 18250) (callable_variable 18251) (class_constant 1546) (class_name 18252) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 18253) (name 18254) (new_dereferenceable 1552) (simple_variable 18255) (static_member 18256) (variable 18959) (variable_class_name 18258)) 7958 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18958) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7959 (("(" 18955)) 7960 (("(" 18952)) 7961 (("(" 18943) (argument_list 18944) (ctor_arguments 18945)) 7962 (("\"" 1520) ("$" 18246) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 18247) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 18248) ("[" 1542) (array_object_dereferenceable 18249) (callable_expr 18250) (callable_variable 18251) (class_constant 1546) (class_name 18252) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 18253) (name 18254) (new_dereferenceable 1552) (simple_variable 18255) (static_member 18256) (variable 18942) (variable_class_name 18258)) 7963 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18941) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7964 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18940) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7965 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18939) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7966 (("(" 18935)) 7967 (("(" 18930)) 7968 (("(" 18924)) 7969 (("$" 18246) ("(" 18887) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 18888) (T_NAME_QUALIFIED 18889) (T_NAME_RELATIVE 18890) (T_READONLY 2931) (T_STATIC 18891) (T_STRING 18892) (T_VARIABLE 18248) (anonymous_class 18893) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 18894) (attribute 2937) (attributes 18895) (class_modifier 2939) (class_name 18896) (class_name_reference 18897) (name 18898) (new_variable 18899) (simple_variable 18900)) 7970 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18886) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7971 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18885) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7972 (("(" 18207) (argument_list 18884)) 7973 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18883) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7974 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18882) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7975 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 18877) (T_END_HEREDOC 18878) (T_VARIABLE 2896) (encaps_list 18879) (encaps_var 2898)) 7976 ((T_FN 35) (T_FUNCTION 129) (fn 18005) #'18006 (inline_function 18876)) 7977 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18875) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7978 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18874) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7979 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18873) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7980 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18415) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 7981 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18351) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7982 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 18347) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 7983 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 18345) (encaps_list 2804) (encaps_var 2805)) 7984 ((T_NULLSAFE_OBJECT_OPERATOR 18331) (T_OBJECT_OPERATOR 18332) ("[" 18333)) 7985 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 18328) (attribute 2733) (fn 18005) #'18006 (inline_function 18329)) 7986 (("(" 18207) (argument_list 18327)) 7987 ((T_PAAMAYIM_NEKUDOTAYIM 18319)) 7988 (("%" 18021) ("*" 18022) ("+" 18023) ("-" 18024) ("." 18025) ("/" 18026) ("<" 18027) (">" 18028) ("?" 18029) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 18030) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 18031) (T_BOOLEAN_AND 18032) (T_BOOLEAN_OR 18033) (T_COALESCE 18034) (T_INSTANCEOF 18035) (T_IS_EQUAL 18036) (T_IS_GREATER_OR_EQUAL 18037) (T_IS_IDENTICAL 18038) (T_IS_NOT_EQUAL 18039) (T_IS_NOT_IDENTICAL 18040) (T_IS_SMALLER_OR_EQUAL 18041) (T_LOGICAL_AND 18042) (T_LOGICAL_OR 18043) (T_LOGICAL_XOR 18044) (T_POW 18045) (T_SL 18046) (T_SPACESHIP 18047) (T_SR 18048) ("^" 18049) ("|" 18050)) 7989 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 18308)) 7990 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 18296)) 7991 (("(" 18207) (argument_list 18295)) 7992 (("=" 18215) (T_AND_EQUAL 18216) (T_COALESCE_EQUAL 18217) (T_CONCAT_EQUAL 18218) (T_DEC 18219) (T_DIV_EQUAL 18220) (T_INC 18221) (T_MINUS_EQUAL 18222) (T_MOD_EQUAL 18223) (T_MUL_EQUAL 18224) (T_OR_EQUAL 18225) (T_PLUS_EQUAL 18226) (T_POW_EQUAL 18227) (T_SL_EQUAL 18228) (T_SR_EQUAL 18229) (T_XOR_EQUAL 18230)) 7993 ((T_PAAMAYIM_NEKUDOTAYIM 18119)) 7994 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18020) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7995 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18118) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7996 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18117) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7997 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18116) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7998 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18115) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 7999 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18114) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8000 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18113) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8001 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18112) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8002 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18111) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8003 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 18106) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 18107) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 8004 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18105) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8005 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18104) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8006 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18103) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8007 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18102) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8008 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18101) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8009 (("$" 18066) ("(" 18067) (T_NAME_FULLY_QUALIFIED 18068) (T_NAME_QUALIFIED 18069) (T_NAME_RELATIVE 18070) (T_STATIC 18071) (T_STRING 18072) (T_VARIABLE 18073) (class_name 18074) (class_name_reference 18075) (name 18076) (new_variable 18077) (simple_variable 18078)) 8010 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18065) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8011 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18064) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8012 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18063) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8013 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18062) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8014 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18061) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8015 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18060) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8016 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18059) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8017 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18058) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8018 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18057) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8019 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18056) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8020 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18055) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8021 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18054) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8022 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18053) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8023 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18052) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8024 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18051) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8025 (("$" 18066) (T_VARIABLE 18073) (simple_variable 18097) ("{" 18098)) 8026 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 18095) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8027 ((T_PAAMAYIM_NEKUDOTAYIM 18093)) 8028 ((T_NULLSAFE_OBJECT_OPERATOR 18079) (T_OBJECT_OPERATOR 18080) (T_PAAMAYIM_NEKUDOTAYIM 18081) ("[" 18082)) 8029 (("$" 18066) (T_STRING 18086) (T_VARIABLE 18073) (property_name 18092) (simple_variable 18088) ("{" 18089)) 8030 (("$" 18066) (T_STRING 18086) (T_VARIABLE 18073) (property_name 18087) (simple_variable 18088) ("{" 18089)) 8031 (("$" 18066) (T_VARIABLE 18073) (simple_variable 18085)) 8032 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 18083) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8033 (("]" 18084)) 8034 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18090) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8035 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18091)) 8036 (("$" 18066) (T_VARIABLE 18073) (simple_variable 18094)) 8037 (("%" 299) (")" 18096) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8038 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18099) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8039 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18100)) 8040 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18110) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8041 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 18108) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 8042 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18109) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8043 (("$" 17943) (T_ABSTRACT 18120) (T_ARRAY 18121) (T_AS 18122) (T_BREAK 18123) (T_CALLABLE 18124) (T_CASE 18125) (T_CATCH 18126) (T_CLASS 18127) (T_CLASS_C 18128) (T_CLONE 18129) (T_CONST 18130) (T_CONTINUE 18131) (T_DECLARE 18132) (T_DEFAULT 18133) (T_DIR 18134) (T_DO 18135) (T_ECHO 18136) (T_ELSE 18137) (T_ELSEIF 18138) (T_EMPTY 18139) (T_ENDDECLARE 18140) (T_ENDFOR 18141) (T_ENDFOREACH 18142) (T_ENDIF 18143) (T_ENDSWITCH 18144) (T_ENDWHILE 18145) (T_ENUM 18146) (T_EVAL 18147) (T_EXIT 18148) (T_EXTENDS 18149) (T_FILE 18150) (T_FINAL 18151) (T_FINALLY 18152) (T_FN 18153) (T_FOR 18154) (T_FOREACH 18155) (T_FUNCTION 18156) (T_FUNC_C 18157) (T_GLOBAL 18158) (T_GOTO 18159) (T_IF 18160) (T_IMPLEMENTS 18161) (T_INCLUDE 18162) (T_INCLUDE_ONCE 18163) (T_INSTANCEOF 18164) (T_INSTEADOF 18165) (T_INTERFACE 18166) (T_ISSET 18167) (T_LINE 18168) (T_LIST 18169) (T_LOGICAL_AND 18170) (T_LOGICAL_OR 18171) (T_LOGICAL_XOR 18172) (T_MATCH 18173) (T_METHOD_C 18174) (T_NAMESPACE 18175) (T_NEW 18176) (T_NS_C 18177) (T_PRINT 18178) (T_PRIVATE 18179) (T_PROPERTY_C 18180) (T_PROTECTED 18181) (T_PUBLIC 18182) (T_READONLY 18183) (T_REQUIRE 18184) (T_REQUIRE_ONCE 18185) (T_RETURN 18186) (T_STATIC 18187) (T_STRING 18188) (T_SWITCH 18189) (T_THROW 18190) (T_TRAIT 18191) (T_TRAIT_C 18192) (T_TRY 18193) (T_UNSET 18194) (T_USE 18195) (T_VAR 18196) (T_VARIABLE 17991) (T_WHILE 18197) (T_YIELD 18198) (identifier 18199) (member_name 18200) (reserved_non_modifiers 18201) (semi_reserved 18202) (simple_variable 18203) ("{" 18204)) 8044 (("(" 18207) (argument_list 18208)) 8045 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18205) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8046 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18206)) 8047 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 18209) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 18210) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 18211) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8048 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 18214) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8049 (("," 1480) (possible_comma 18212)) 8050 ((")" 18213)) 8051 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (ampersand 18244) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18245) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8052 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18243) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8053 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18242) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8054 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18241) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8055 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18240) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8056 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18239) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8057 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18238) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8058 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18237) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8059 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18236) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8060 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18235) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8061 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18234) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8062 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18233) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8063 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18232) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8064 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18231) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8065 (("\"" 1520) ("$" 18246) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 18247) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 18248) ("[" 1542) (array_object_dereferenceable 18249) (callable_expr 18250) (callable_variable 18251) (class_constant 1546) (class_name 18252) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 18253) (name 18254) (new_dereferenceable 1552) (simple_variable 18255) (static_member 18256) (variable 18257) (variable_class_name 18258)) 8066 (("$" 18246) (T_VARIABLE 18248) (simple_variable 18291) ("{" 18292)) 8067 (("(" 18262) (argument_list 18290)) 8068 ((T_NULLSAFE_OBJECT_OPERATOR 18276) (T_OBJECT_OPERATOR 18277) ("[" 18278)) 8069 (("(" 18262) (argument_list 18275)) 8070 ((T_PAAMAYIM_NEKUDOTAYIM 18271)) 8071 (("(" 18262) (argument_list 18270)) 8072 ((T_PAAMAYIM_NEKUDOTAYIM 18259)) 8073 (("$" 18246) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 18248) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 18260) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 18261) ("{" 1642)) 8074 (("(" 18262) (argument_list 18263)) 8075 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 18264) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 18265) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 18266) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8076 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 18269) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8077 (("," 1480) (possible_comma 18267)) 8078 ((")" 18268)) 8079 (("$" 18246) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 18248) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 18272) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 18273) ("{" 1658)) 8080 (("(" 18262) (argument_list 18274)) 8081 (("$" 18246) (T_STRING 18281) (T_VARIABLE 18248) (property_name 18288) (simple_variable 18283) ("{" 18284)) 8082 (("$" 18246) (T_STRING 18281) (T_VARIABLE 18248) (property_name 18282) (simple_variable 18283) ("{" 18284)) 8083 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 18279) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8084 (("]" 18280)) 8085 (("(" 18262) (argument_list 18287)) 8086 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18285) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8087 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18286)) 8088 (("(" 18262) (argument_list 18289)) 8089 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18293) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8090 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18294)) 8091 ((backup_doc_comment 18297)) 8092 (("(" 18298)) 8093 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 18299)) 8094 ((")" 18300)) 8095 ((T_USE 164) (lexical_vars 18301)) 8096 ((":" 166) (return_type 18302)) 8097 ((backup_fn_flags 18303)) 8098 (("{" 18304)) 8099 ((inner_statement_list 18305)) 8100 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 18306) ("~" 127)) 8101 ((backup_fn_flags 18307)) 8102 ((backup_doc_comment 18309)) 8103 (("(" 18310)) 8104 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 18311)) 8105 ((")" 18312)) 8106 ((":" 2249) (return_type 18313)) 8107 ((T_DOUBLE_ARROW 18314)) 8108 ((backup_fn_flags 18315)) 8109 ((backup_lex_pos 18316)) 8110 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 18317) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 8111 (("%" 3129) ("*" 3130) ("+" 3131) ("-" 3132) ("." 3133) ("/" 3134) ("<" 3135) (">" 3136) ("?" 3137) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3138) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3139) (T_BOOLEAN_AND 3140) (T_BOOLEAN_OR 3141) (T_COALESCE 3142) (T_INSTANCEOF 3143) (T_IS_EQUAL 3144) (T_IS_GREATER_OR_EQUAL 3145) (T_IS_IDENTICAL 3146) (T_IS_NOT_EQUAL 3147) (T_IS_NOT_IDENTICAL 3148) (T_IS_SMALLER_OR_EQUAL 3149) (T_LOGICAL_AND 3150) (T_LOGICAL_OR 3151) (T_LOGICAL_XOR 3152) (T_POW 3153) (T_SL 3154) (T_SPACESHIP 3155) (T_SR 3156) ("^" 3157) (backup_fn_flags 18318) ("|" 3158)) 8112 (("$" 17943) (T_ABSTRACT 18120) (T_ARRAY 18121) (T_AS 18122) (T_BREAK 18123) (T_CALLABLE 18124) (T_CASE 18125) (T_CATCH 18126) (T_CLASS 18127) (T_CLASS_C 18128) (T_CLONE 18129) (T_CONST 18130) (T_CONTINUE 18131) (T_DECLARE 18132) (T_DEFAULT 18133) (T_DIR 18134) (T_DO 18135) (T_ECHO 18136) (T_ELSE 18137) (T_ELSEIF 18138) (T_EMPTY 18139) (T_ENDDECLARE 18140) (T_ENDFOR 18141) (T_ENDFOREACH 18142) (T_ENDIF 18143) (T_ENDSWITCH 18144) (T_ENDWHILE 18145) (T_ENUM 18146) (T_EVAL 18147) (T_EXIT 18148) (T_EXTENDS 18149) (T_FILE 18150) (T_FINAL 18151) (T_FINALLY 18152) (T_FN 18153) (T_FOR 18154) (T_FOREACH 18155) (T_FUNCTION 18156) (T_FUNC_C 18157) (T_GLOBAL 18158) (T_GOTO 18159) (T_IF 18160) (T_IMPLEMENTS 18161) (T_INCLUDE 18162) (T_INCLUDE_ONCE 18163) (T_INSTANCEOF 18164) (T_INSTEADOF 18165) (T_INTERFACE 18166) (T_ISSET 18167) (T_LINE 18168) (T_LIST 18169) (T_LOGICAL_AND 18170) (T_LOGICAL_OR 18171) (T_LOGICAL_XOR 18172) (T_MATCH 18173) (T_METHOD_C 18174) (T_NAMESPACE 18175) (T_NEW 18176) (T_NS_C 18177) (T_PRINT 18178) (T_PRIVATE 18179) (T_PROPERTY_C 18180) (T_PROTECTED 18181) (T_PUBLIC 18182) (T_READONLY 18183) (T_REQUIRE 18184) (T_REQUIRE_ONCE 18185) (T_RETURN 18186) (T_STATIC 18187) (T_STRING 18188) (T_SWITCH 18189) (T_THROW 18190) (T_TRAIT 18191) (T_TRAIT_C 18192) (T_TRY 18193) (T_UNSET 18194) (T_USE 18195) (T_VAR 18196) (T_VARIABLE 17991) (T_WHILE 18197) (T_YIELD 18198) (identifier 18320) (member_name 18321) (reserved_non_modifiers 18201) (semi_reserved 18202) (simple_variable 18322) ("{" 18323)) 8113 (("(" 18207) (argument_list 18326)) 8114 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18324) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8115 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18325)) 8116 ((T_FN 35) (T_FUNCTION 129) (fn 18005) #'18006 (inline_function 18330)) 8117 (("$" 17943) (T_STRING 18336) (T_VARIABLE 17991) (property_name 18343) (simple_variable 18338) ("{" 18339)) 8118 (("$" 17943) (T_STRING 18336) (T_VARIABLE 17991) (property_name 18337) (simple_variable 18338) ("{" 18339)) 8119 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 18334) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8120 (("]" 18335)) 8121 (("(" 18207) (argument_list 18342)) 8122 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18340) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8123 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18341)) 8124 (("(" 18207) (argument_list 18344)) 8125 (("`" 18346)) 8126 (("]" 18348)) 8127 (("=" 18349)) 8128 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18350) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8129 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18872) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8130 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 18870) (encaps_var 5237)) 8131 (("$" 18354) (T_VARIABLE 18402) (simple_variable 18866) ("{" 18867)) 8132 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 18864) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8133 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18863) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8134 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18862) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8135 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18861) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8136 (("(" 18858)) 8137 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18857) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8138 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18856) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8139 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18855) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8140 (("\"" 1520) ("$" 18657) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 18658) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 18659) ("[" 1542) (array_object_dereferenceable 18660) (callable_expr 18661) (callable_variable 18662) (class_constant 1546) (class_name 18663) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 18664) (name 18665) (new_dereferenceable 1552) (simple_variable 18666) (static_member 18667) (variable 18854) (variable_class_name 18669)) 8141 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18853) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8142 (("(" 18850)) 8143 (("(" 18847)) 8144 (("(" 18838) (argument_list 18839) (ctor_arguments 18840)) 8145 (("\"" 1520) ("$" 18657) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 18658) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 18659) ("[" 1542) (array_object_dereferenceable 18660) (callable_expr 18661) (callable_variable 18662) (class_constant 1546) (class_name 18663) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 18664) (name 18665) (new_dereferenceable 1552) (simple_variable 18666) (static_member 18667) (variable 18837) (variable_class_name 18669)) 8146 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18836) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8147 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18835) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8148 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18834) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8149 (("(" 18830)) 8150 (("(" 18825)) 8151 (("(" 18819)) 8152 (("$" 18657) ("(" 18782) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 18783) (T_NAME_QUALIFIED 18784) (T_NAME_RELATIVE 18785) (T_READONLY 2931) (T_STATIC 18786) (T_STRING 18787) (T_VARIABLE 18659) (anonymous_class 18788) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 18789) (attribute 2937) (attributes 18790) (class_modifier 2939) (class_name 18791) (class_name_reference 18792) (name 18793) (new_variable 18794) (simple_variable 18795)) 8153 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18781) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8154 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18780) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8155 (("(" 18618) (argument_list 18779)) 8156 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18778) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8157 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18777) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8158 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 18772) (T_END_HEREDOC 18773) (T_VARIABLE 2896) (encaps_list 18774) (encaps_var 2898)) 8159 ((T_FN 35) (T_FUNCTION 129) (fn 18416) #'18417 (inline_function 18771)) 8160 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18770) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8161 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18769) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8162 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18768) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8163 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18765) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8164 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18764) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8165 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 18760) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 8166 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 18758) (encaps_list 2804) (encaps_var 2805)) 8167 ((T_NULLSAFE_OBJECT_OPERATOR 18744) (T_OBJECT_OPERATOR 18745) ("[" 18746)) 8168 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 18741) (attribute 2733) (fn 18416) #'18417 (inline_function 18742)) 8169 (("(" 18618) (argument_list 18740)) 8170 ((T_PAAMAYIM_NEKUDOTAYIM 18732)) 8171 (("%" 18432) ("*" 18433) ("+" 18434) ("-" 18435) ("." 18436) ("/" 18437) ("<" 18438) (">" 18439) ("?" 18440) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 18441) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 18442) (T_BOOLEAN_AND 18443) (T_BOOLEAN_OR 18444) (T_COALESCE 18445) (T_DOUBLE_ARROW 18730) (T_INSTANCEOF 18446) (T_IS_EQUAL 18447) (T_IS_GREATER_OR_EQUAL 18448) (T_IS_IDENTICAL 18449) (T_IS_NOT_EQUAL 18450) (T_IS_NOT_IDENTICAL 18451) (T_IS_SMALLER_OR_EQUAL 18452) (T_LOGICAL_AND 18453) (T_LOGICAL_OR 18454) (T_LOGICAL_XOR 18455) (T_POW 18456) (T_SL 18457) (T_SPACESHIP 18458) (T_SR 18459) ("^" 18460) ("|" 18461)) 8172 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 18719)) 8173 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 18707)) 8174 (("(" 18618) (argument_list 18706)) 8175 (("=" 18626) (T_AND_EQUAL 18627) (T_COALESCE_EQUAL 18628) (T_CONCAT_EQUAL 18629) (T_DEC 18630) (T_DIV_EQUAL 18631) (T_INC 18632) (T_MINUS_EQUAL 18633) (T_MOD_EQUAL 18634) (T_MUL_EQUAL 18635) (T_OR_EQUAL 18636) (T_PLUS_EQUAL 18637) (T_POW_EQUAL 18638) (T_SL_EQUAL 18639) (T_SR_EQUAL 18640) (T_XOR_EQUAL 18641)) 8176 ((T_PAAMAYIM_NEKUDOTAYIM 18530)) 8177 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18431) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8178 (("%" 18432) ("*" 18433) ("+" 18434) ("-" 18435) ("." 18436) ("/" 18437) ("<" 18438) (">" 18439) ("?" 18440) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 18441) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 18442) (T_BOOLEAN_AND 18443) (T_BOOLEAN_OR 18444) (T_COALESCE 18445) (T_INSTANCEOF 18446) (T_IS_EQUAL 18447) (T_IS_GREATER_OR_EQUAL 18448) (T_IS_IDENTICAL 18449) (T_IS_NOT_EQUAL 18450) (T_IS_NOT_IDENTICAL 18451) (T_IS_SMALLER_OR_EQUAL 18452) (T_LOGICAL_AND 18453) (T_LOGICAL_OR 18454) (T_LOGICAL_XOR 18455) (T_POW 18456) (T_SL 18457) (T_SPACESHIP 18458) (T_SR 18459) ("^" 18460) ("|" 18461)) 8179 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18529) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8180 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18528) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8181 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18527) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8182 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18526) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8183 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18525) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8184 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18524) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8185 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18523) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8186 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18522) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8187 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 18517) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 18518) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 8188 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18516) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8189 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18515) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8190 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18514) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8191 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18513) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8192 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18512) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8193 (("$" 18477) ("(" 18478) (T_NAME_FULLY_QUALIFIED 18479) (T_NAME_QUALIFIED 18480) (T_NAME_RELATIVE 18481) (T_STATIC 18482) (T_STRING 18483) (T_VARIABLE 18484) (class_name 18485) (class_name_reference 18486) (name 18487) (new_variable 18488) (simple_variable 18489)) 8194 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18476) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8195 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18475) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8196 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18474) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8197 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18473) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8198 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18472) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8199 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18471) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8200 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18470) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8201 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18469) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8202 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18468) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8203 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18467) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8204 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18466) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8205 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18465) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8206 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18464) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8207 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18463) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8208 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18462) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8209 (("$" 18477) (T_VARIABLE 18484) (simple_variable 18508) ("{" 18509)) 8210 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 18506) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8211 ((T_PAAMAYIM_NEKUDOTAYIM 18504)) 8212 ((T_NULLSAFE_OBJECT_OPERATOR 18490) (T_OBJECT_OPERATOR 18491) (T_PAAMAYIM_NEKUDOTAYIM 18492) ("[" 18493)) 8213 (("$" 18477) (T_STRING 18497) (T_VARIABLE 18484) (property_name 18503) (simple_variable 18499) ("{" 18500)) 8214 (("$" 18477) (T_STRING 18497) (T_VARIABLE 18484) (property_name 18498) (simple_variable 18499) ("{" 18500)) 8215 (("$" 18477) (T_VARIABLE 18484) (simple_variable 18496)) 8216 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 18494) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8217 (("]" 18495)) 8218 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18501) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8219 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18502)) 8220 (("$" 18477) (T_VARIABLE 18484) (simple_variable 18505)) 8221 (("%" 299) (")" 18507) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8222 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18510) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8223 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18511)) 8224 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18521) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8225 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 18519) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 8226 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18520) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8227 (("$" 18354) (T_ABSTRACT 18531) (T_ARRAY 18532) (T_AS 18533) (T_BREAK 18534) (T_CALLABLE 18535) (T_CASE 18536) (T_CATCH 18537) (T_CLASS 18538) (T_CLASS_C 18539) (T_CLONE 18540) (T_CONST 18541) (T_CONTINUE 18542) (T_DECLARE 18543) (T_DEFAULT 18544) (T_DIR 18545) (T_DO 18546) (T_ECHO 18547) (T_ELSE 18548) (T_ELSEIF 18549) (T_EMPTY 18550) (T_ENDDECLARE 18551) (T_ENDFOR 18552) (T_ENDFOREACH 18553) (T_ENDIF 18554) (T_ENDSWITCH 18555) (T_ENDWHILE 18556) (T_ENUM 18557) (T_EVAL 18558) (T_EXIT 18559) (T_EXTENDS 18560) (T_FILE 18561) (T_FINAL 18562) (T_FINALLY 18563) (T_FN 18564) (T_FOR 18565) (T_FOREACH 18566) (T_FUNCTION 18567) (T_FUNC_C 18568) (T_GLOBAL 18569) (T_GOTO 18570) (T_IF 18571) (T_IMPLEMENTS 18572) (T_INCLUDE 18573) (T_INCLUDE_ONCE 18574) (T_INSTANCEOF 18575) (T_INSTEADOF 18576) (T_INTERFACE 18577) (T_ISSET 18578) (T_LINE 18579) (T_LIST 18580) (T_LOGICAL_AND 18581) (T_LOGICAL_OR 18582) (T_LOGICAL_XOR 18583) (T_MATCH 18584) (T_METHOD_C 18585) (T_NAMESPACE 18586) (T_NEW 18587) (T_NS_C 18588) (T_PRINT 18589) (T_PRIVATE 18590) (T_PROPERTY_C 18591) (T_PROTECTED 18592) (T_PUBLIC 18593) (T_READONLY 18594) (T_REQUIRE 18595) (T_REQUIRE_ONCE 18596) (T_RETURN 18597) (T_STATIC 18598) (T_STRING 18599) (T_SWITCH 18600) (T_THROW 18601) (T_TRAIT 18602) (T_TRAIT_C 18603) (T_TRY 18604) (T_UNSET 18605) (T_USE 18606) (T_VAR 18607) (T_VARIABLE 18402) (T_WHILE 18608) (T_YIELD 18609) (identifier 18610) (member_name 18611) (reserved_non_modifiers 18612) (semi_reserved 18613) (simple_variable 18614) ("{" 18615)) 8228 (("(" 18618) (argument_list 18619)) 8229 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18616) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8230 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18617)) 8231 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 18620) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 18621) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 18622) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8232 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 18625) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8233 (("," 1480) (possible_comma 18623)) 8234 ((")" 18624)) 8235 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (ampersand 18655) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18656) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8236 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18654) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8237 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18653) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8238 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18652) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8239 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18651) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8240 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18650) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8241 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18649) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8242 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18648) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8243 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18647) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8244 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18646) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8245 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18645) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8246 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18644) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8247 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18643) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8248 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18642) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8249 (("\"" 1520) ("$" 18657) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 18658) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 18659) ("[" 1542) (array_object_dereferenceable 18660) (callable_expr 18661) (callable_variable 18662) (class_constant 1546) (class_name 18663) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 18664) (name 18665) (new_dereferenceable 1552) (simple_variable 18666) (static_member 18667) (variable 18668) (variable_class_name 18669)) 8250 (("$" 18657) (T_VARIABLE 18659) (simple_variable 18702) ("{" 18703)) 8251 (("(" 18673) (argument_list 18701)) 8252 ((T_NULLSAFE_OBJECT_OPERATOR 18687) (T_OBJECT_OPERATOR 18688) ("[" 18689)) 8253 (("(" 18673) (argument_list 18686)) 8254 ((T_PAAMAYIM_NEKUDOTAYIM 18682)) 8255 (("(" 18673) (argument_list 18681)) 8256 ((T_PAAMAYIM_NEKUDOTAYIM 18670)) 8257 (("$" 18657) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 18659) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 18671) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 18672) ("{" 1642)) 8258 (("(" 18673) (argument_list 18674)) 8259 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 18675) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 18676) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 18677) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8260 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 18680) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8261 (("," 1480) (possible_comma 18678)) 8262 ((")" 18679)) 8263 (("$" 18657) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 18659) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 18683) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 18684) ("{" 1658)) 8264 (("(" 18673) (argument_list 18685)) 8265 (("$" 18657) (T_STRING 18692) (T_VARIABLE 18659) (property_name 18699) (simple_variable 18694) ("{" 18695)) 8266 (("$" 18657) (T_STRING 18692) (T_VARIABLE 18659) (property_name 18693) (simple_variable 18694) ("{" 18695)) 8267 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 18690) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8268 (("]" 18691)) 8269 (("(" 18673) (argument_list 18698)) 8270 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18696) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8271 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18697)) 8272 (("(" 18673) (argument_list 18700)) 8273 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18704) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8274 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18705)) 8275 ((backup_doc_comment 18708)) 8276 (("(" 18709)) 8277 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 18710)) 8278 ((")" 18711)) 8279 ((T_USE 164) (lexical_vars 18712)) 8280 ((":" 166) (return_type 18713)) 8281 ((backup_fn_flags 18714)) 8282 (("{" 18715)) 8283 ((inner_statement_list 18716)) 8284 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 18717) ("~" 127)) 8285 ((backup_fn_flags 18718)) 8286 ((backup_doc_comment 18720)) 8287 (("(" 18721)) 8288 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 18722)) 8289 ((")" 18723)) 8290 ((":" 2249) (return_type 18724)) 8291 ((T_DOUBLE_ARROW 18725)) 8292 ((backup_fn_flags 18726)) 8293 ((backup_lex_pos 18727)) 8294 (("!" 3461) ("\"" 3462) ("$" 3463) ("(" 3464) ("+" 3465) ("-" 3466) ("@" 3467) (T_ARRAY 3468) (T_ARRAY_CAST 3469) (T_ATTRIBUTE 128) (T_BOOL_CAST 3470) (T_CLASS_C 3471) (T_CLONE 3472) (T_CONSTANT_ENCAPSED_STRING 3473) (T_DEC 3474) (T_DIR 3475) (T_DNUMBER 3476) (T_DOUBLE_CAST 3477) (T_EMPTY 3478) (T_EVAL 3479) (T_EXIT 3480) (T_FILE 3481) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3482) (T_INC 3483) (T_INCLUDE 3484) (T_INCLUDE_ONCE 3485) (T_INT_CAST 3486) (T_ISSET 3487) (T_LINE 3488) (T_LIST 3489) (T_LNUMBER 3490) (T_MATCH 3491) (T_METHOD_C 3492) (T_NAME_FULLY_QUALIFIED 3493) (T_NAME_QUALIFIED 3494) (T_NAME_RELATIVE 3495) (T_NEW 3496) (T_NS_C 3497) (T_OBJECT_CAST 3498) (T_PRINT 3499) (T_PROPERTY_C 3500) (T_READONLY 3501) (T_REQUIRE 3502) (T_REQUIRE_ONCE 3503) (T_START_HEREDOC 3504) (T_STATIC 3505) (T_STRING 3506) (T_STRING_CAST 3507) (T_THROW 3508) (T_TRAIT_C 3509) (T_UNSET_CAST 3510) (T_VARIABLE 3511) (T_YIELD 3512) (T_YIELD_FROM 3513) ("[" 3514) ("`" 3515) (array_object_dereferenceable 3516) (attribute 133) (attributes 3517) (callable_expr 3518) (callable_variable 3519) (class_constant 3520) (class_name 3521) (constant 3522) (dereferenceable_scalar 3523) (expr 18728) (fn 3525) (fully_dereferenceable 105) #'3526 (function_call 3527) (inline_function 3528) (internal_functions_in_yacc 3529) (match 3530) (name 3531) (new_dereferenceable 3532) (new_non_dereferenceable 3533) (scalar 3534) (simple_variable 3535) (static_member 3536) (variable 3537) (variable_class_name 3538) ("~" 3539)) 8295 (("%" 3541) ("*" 3542) ("+" 3543) ("-" 3544) ("." 3545) ("/" 3546) ("<" 3547) (">" 3548) ("?" 3549) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3550) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3551) (T_BOOLEAN_AND 3552) (T_BOOLEAN_OR 3553) (T_COALESCE 3554) (T_INSTANCEOF 3555) (T_IS_EQUAL 3556) (T_IS_GREATER_OR_EQUAL 3557) (T_IS_IDENTICAL 3558) (T_IS_NOT_EQUAL 3559) (T_IS_NOT_IDENTICAL 3560) (T_IS_SMALLER_OR_EQUAL 3561) (T_LOGICAL_AND 3562) (T_LOGICAL_OR 3563) (T_LOGICAL_XOR 3564) (T_POW 3565) (T_SL 3566) (T_SPACESHIP 3567) (T_SR 3568) ("^" 3569) (backup_fn_flags 18729) ("|" 3570)) 8296 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18731) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8297 (("$" 18354) (T_ABSTRACT 18531) (T_ARRAY 18532) (T_AS 18533) (T_BREAK 18534) (T_CALLABLE 18535) (T_CASE 18536) (T_CATCH 18537) (T_CLASS 18538) (T_CLASS_C 18539) (T_CLONE 18540) (T_CONST 18541) (T_CONTINUE 18542) (T_DECLARE 18543) (T_DEFAULT 18544) (T_DIR 18545) (T_DO 18546) (T_ECHO 18547) (T_ELSE 18548) (T_ELSEIF 18549) (T_EMPTY 18550) (T_ENDDECLARE 18551) (T_ENDFOR 18552) (T_ENDFOREACH 18553) (T_ENDIF 18554) (T_ENDSWITCH 18555) (T_ENDWHILE 18556) (T_ENUM 18557) (T_EVAL 18558) (T_EXIT 18559) (T_EXTENDS 18560) (T_FILE 18561) (T_FINAL 18562) (T_FINALLY 18563) (T_FN 18564) (T_FOR 18565) (T_FOREACH 18566) (T_FUNCTION 18567) (T_FUNC_C 18568) (T_GLOBAL 18569) (T_GOTO 18570) (T_IF 18571) (T_IMPLEMENTS 18572) (T_INCLUDE 18573) (T_INCLUDE_ONCE 18574) (T_INSTANCEOF 18575) (T_INSTEADOF 18576) (T_INTERFACE 18577) (T_ISSET 18578) (T_LINE 18579) (T_LIST 18580) (T_LOGICAL_AND 18581) (T_LOGICAL_OR 18582) (T_LOGICAL_XOR 18583) (T_MATCH 18584) (T_METHOD_C 18585) (T_NAMESPACE 18586) (T_NEW 18587) (T_NS_C 18588) (T_PRINT 18589) (T_PRIVATE 18590) (T_PROPERTY_C 18591) (T_PROTECTED 18592) (T_PUBLIC 18593) (T_READONLY 18594) (T_REQUIRE 18595) (T_REQUIRE_ONCE 18596) (T_RETURN 18597) (T_STATIC 18598) (T_STRING 18599) (T_SWITCH 18600) (T_THROW 18601) (T_TRAIT 18602) (T_TRAIT_C 18603) (T_TRY 18604) (T_UNSET 18605) (T_USE 18606) (T_VAR 18607) (T_VARIABLE 18402) (T_WHILE 18608) (T_YIELD 18609) (identifier 18733) (member_name 18734) (reserved_non_modifiers 18612) (semi_reserved 18613) (simple_variable 18735) ("{" 18736)) 8298 (("(" 18618) (argument_list 18739)) 8299 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18737) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8300 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18738)) 8301 ((T_FN 35) (T_FUNCTION 129) (fn 18416) #'18417 (inline_function 18743)) 8302 (("$" 18354) (T_STRING 18749) (T_VARIABLE 18402) (property_name 18756) (simple_variable 18751) ("{" 18752)) 8303 (("$" 18354) (T_STRING 18749) (T_VARIABLE 18402) (property_name 18750) (simple_variable 18751) ("{" 18752)) 8304 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 18747) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8305 (("]" 18748)) 8306 (("(" 18618) (argument_list 18755)) 8307 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18753) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8308 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18754)) 8309 (("(" 18618) (argument_list 18757)) 8310 (("`" 18759)) 8311 (("]" 18761)) 8312 (("=" 18762)) 8313 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18763) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8314 (("%" 18432) ("*" 18433) ("+" 18434) ("-" 18435) ("." 18436) ("/" 18437) ("<" 18438) (">" 18439) ("?" 18440) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 18441) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 18442) (T_BOOLEAN_AND 18443) (T_BOOLEAN_OR 18444) (T_COALESCE 18445) (T_DOUBLE_ARROW 18766) (T_INSTANCEOF 18446) (T_IS_EQUAL 18447) (T_IS_GREATER_OR_EQUAL 18448) (T_IS_IDENTICAL 18449) (T_IS_NOT_EQUAL 18450) (T_IS_NOT_IDENTICAL 18451) (T_IS_SMALLER_OR_EQUAL 18452) (T_LOGICAL_AND 18453) (T_LOGICAL_OR 18454) (T_LOGICAL_XOR 18455) (T_POW 18456) (T_SL 18457) (T_SPACESHIP 18458) (T_SR 18459) ("^" 18460) ("|" 18461)) 8315 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18767) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8316 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 18776) (T_VARIABLE 2896) (encaps_var 2910)) 8317 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 18775) (T_VARIABLE 2896) (encaps_var 2901)) 8318 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 18817) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8319 ((T_CLASS 18809)) 8320 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 18808) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 18789) (attribute 2958) (class_modifier 2939)) 8321 ((T_PAAMAYIM_NEKUDOTAYIM 18806)) 8322 (("(" 18673) (argument_list 18805)) 8323 ((T_NULLSAFE_OBJECT_OPERATOR 18796) (T_OBJECT_OPERATOR 18797) (T_PAAMAYIM_NEKUDOTAYIM 18798) ("[" 18799)) 8324 (("$" 18657) (T_STRING 18692) (T_VARIABLE 18659) (property_name 18804) (simple_variable 18694) ("{" 18695)) 8325 (("$" 18657) (T_STRING 18692) (T_VARIABLE 18659) (property_name 18803) (simple_variable 18694) ("{" 18695)) 8326 (("$" 18657) (T_VARIABLE 18659) (simple_variable 18802)) 8327 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 18800) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8328 (("]" 18801)) 8329 (("$" 18657) (T_VARIABLE 18659) (simple_variable 18807)) 8330 (("(" 2960) (argument_list 2961) (ctor_arguments 18810)) 8331 ((T_EXTENDS 2963) (extends_from 18811)) 8332 ((T_IMPLEMENTS 2965) (implements_list 18812)) 8333 ((backup_doc_comment 18813)) 8334 (("{" 18814)) 8335 ((class_statement_list 18815)) 8336 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 18816)) 8337 (("%" 299) (")" 18818) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8338 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 18820) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8339 (("%" 299) (")" 18821) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8340 (("{" 18822)) 8341 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 18823) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 8342 (("}" 18824)) 8343 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 18826) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 8344 ((")" 18827)) 8345 (("=" 18828)) 8346 (("!" 18352) ("\"" 18353) ("$" 18354) ("(" 18355) ("+" 18356) ("-" 18357) ("@" 18358) (T_ARRAY 18359) (T_ARRAY_CAST 18360) (T_ATTRIBUTE 128) (T_BOOL_CAST 18361) (T_CLASS_C 18362) (T_CLONE 18363) (T_CONSTANT_ENCAPSED_STRING 18364) (T_DEC 18365) (T_DIR 18366) (T_DNUMBER 18367) (T_DOUBLE_CAST 18368) (T_EMPTY 18369) (T_EVAL 18370) (T_EXIT 18371) (T_FILE 18372) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 18373) (T_INC 18374) (T_INCLUDE 18375) (T_INCLUDE_ONCE 18376) (T_INT_CAST 18377) (T_ISSET 18378) (T_LINE 18379) (T_LIST 18380) (T_LNUMBER 18381) (T_MATCH 18382) (T_METHOD_C 18383) (T_NAME_FULLY_QUALIFIED 18384) (T_NAME_QUALIFIED 18385) (T_NAME_RELATIVE 18386) (T_NEW 18387) (T_NS_C 18388) (T_OBJECT_CAST 18389) (T_PRINT 18390) (T_PROPERTY_C 18391) (T_READONLY 18392) (T_REQUIRE 18393) (T_REQUIRE_ONCE 18394) (T_START_HEREDOC 18395) (T_STATIC 18396) (T_STRING 18397) (T_STRING_CAST 18398) (T_THROW 18399) (T_TRAIT_C 18400) (T_UNSET_CAST 18401) (T_VARIABLE 18402) (T_YIELD 18403) (T_YIELD_FROM 18404) ("[" 18405) ("`" 18406) (array_object_dereferenceable 18407) (attribute 133) (attributes 18408) (callable_expr 18409) (callable_variable 18410) (class_constant 18411) (class_name 18412) (constant 18413) (dereferenceable_scalar 18414) (expr 18829) (fn 18416) (fully_dereferenceable 105) #'18417 (function_call 18418) (inline_function 18419) (internal_functions_in_yacc 18420) (match 18421) (name 18422) (new_dereferenceable 18423) (new_non_dereferenceable 18424) (scalar 18425) (simple_variable 18426) (static_member 18427) (variable 18428) (variable_class_name 18429) ("~" 18430)) 8347 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 18831) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8348 (("," 5192) (possible_comma 18832)) 8349 ((")" 18833)) 8350 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 18841) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 18842) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 18843) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8351 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 18846) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8352 (("," 1480) (possible_comma 18844)) 8353 ((")" 18845)) 8354 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 18848) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8355 (("%" 299) (")" 18849) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8356 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 18851) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8357 (("%" 299) (")" 18852) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8358 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 18859) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 8359 ((")" 18860)) 8360 (("%" 299) (")" 18865) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8361 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18868) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8362 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18869)) 8363 (("\"" 18871) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 8364 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 18881) (T_VARIABLE 2896) (encaps_var 2910)) 8365 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 18880) (T_VARIABLE 2896) (encaps_var 2901)) 8366 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 18922) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8367 ((T_CLASS 18914)) 8368 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 18913) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 18894) (attribute 2958) (class_modifier 2939)) 8369 ((T_PAAMAYIM_NEKUDOTAYIM 18911)) 8370 (("(" 18262) (argument_list 18910)) 8371 ((T_NULLSAFE_OBJECT_OPERATOR 18901) (T_OBJECT_OPERATOR 18902) (T_PAAMAYIM_NEKUDOTAYIM 18903) ("[" 18904)) 8372 (("$" 18246) (T_STRING 18281) (T_VARIABLE 18248) (property_name 18909) (simple_variable 18283) ("{" 18284)) 8373 (("$" 18246) (T_STRING 18281) (T_VARIABLE 18248) (property_name 18908) (simple_variable 18283) ("{" 18284)) 8374 (("$" 18246) (T_VARIABLE 18248) (simple_variable 18907)) 8375 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 18905) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8376 (("]" 18906)) 8377 (("$" 18246) (T_VARIABLE 18248) (simple_variable 18912)) 8378 (("(" 2960) (argument_list 2961) (ctor_arguments 18915)) 8379 ((T_EXTENDS 2963) (extends_from 18916)) 8380 ((T_IMPLEMENTS 2965) (implements_list 18917)) 8381 ((backup_doc_comment 18918)) 8382 (("{" 18919)) 8383 ((class_statement_list 18920)) 8384 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 18921)) 8385 (("%" 299) (")" 18923) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8386 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 18925) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8387 (("%" 299) (")" 18926) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8388 (("{" 18927)) 8389 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 18928) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 8390 (("}" 18929)) 8391 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 18931) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 8392 ((")" 18932)) 8393 (("=" 18933)) 8394 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 18934) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 8395 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 18936) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8396 (("," 5192) (possible_comma 18937)) 8397 ((")" 18938)) 8398 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 18946) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 18947) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 18948) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8399 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 18951) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8400 (("," 1480) (possible_comma 18949)) 8401 ((")" 18950)) 8402 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 18953) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8403 (("%" 299) (")" 18954) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8404 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 18956) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8405 (("%" 299) (")" 18957) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8406 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 18964) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 8407 ((")" 18965)) 8408 (("%" 299) (")" 18970) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8409 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18973) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8410 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18974)) 8411 (("\"" 18976) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 8412 ((";" 18980)) 8413 ((";" 18982)) 8414 (("$" 18983) (T_VARIABLE 18984) (simple_variable 18991) ("{" 18992)) 8415 (("," 18988) (";" 18989)) 8416 (("$" 18983) (T_VARIABLE 18984) (global_var 18990) (simple_variable 18987)) 8417 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 18993) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8418 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 18994)) 8419 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19059) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8420 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22651) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8421 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 22649) (encaps_var 5237)) 8422 (("$" 18998) (T_VARIABLE 19046) (simple_variable 22645) ("{" 22646)) 8423 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22643) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8424 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22642) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8425 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22641) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8426 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22640) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8427 (("(" 22637)) 8428 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22636) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8429 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22635) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8430 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22634) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8431 (("\"" 1520) ("$" 19301) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 19302) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 19303) ("[" 1542) (array_object_dereferenceable 19304) (callable_expr 19305) (callable_variable 19306) (class_constant 1546) (class_name 19307) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 19308) (name 19309) (new_dereferenceable 1552) (simple_variable 19310) (static_member 19311) (variable 22633) (variable_class_name 19313)) 8432 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22632) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8433 (("(" 22629)) 8434 (("(" 22626)) 8435 (("(" 22617) (argument_list 22618) (ctor_arguments 22619)) 8436 (("\"" 1520) ("$" 19301) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 19302) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 19303) ("[" 1542) (array_object_dereferenceable 19304) (callable_expr 19305) (callable_variable 19306) (class_constant 1546) (class_name 19307) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 19308) (name 19309) (new_dereferenceable 1552) (simple_variable 19310) (static_member 19311) (variable 22616) (variable_class_name 19313)) 8437 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22615) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8438 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22614) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8439 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22613) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8440 (("(" 22609)) 8441 (("(" 22604)) 8442 (("(" 22598)) 8443 (("$" 19301) ("(" 22561) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 22562) (T_NAME_QUALIFIED 22563) (T_NAME_RELATIVE 22564) (T_READONLY 2931) (T_STATIC 22565) (T_STRING 22566) (T_VARIABLE 19303) (anonymous_class 22567) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 22568) (attribute 2937) (attributes 22569) (class_modifier 2939) (class_name 22570) (class_name_reference 22571) (name 22572) (new_variable 22573) (simple_variable 22574)) 8444 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22560) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8445 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22559) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8446 (("(" 19262) (argument_list 22558)) 8447 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22557) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8448 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22556) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8449 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 22551) (T_END_HEREDOC 22552) (T_VARIABLE 2896) (encaps_list 22553) (encaps_var 2898)) 8450 ((T_FN 35) (T_FUNCTION 129) (fn 19060) #'19061 (inline_function 22550)) 8451 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22549) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8452 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22548) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8453 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22547) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8454 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22089) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 8455 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22025) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8456 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 22021) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 8457 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 22019) (encaps_list 2804) (encaps_var 2805)) 8458 ((T_NULLSAFE_OBJECT_OPERATOR 22005) (T_OBJECT_OPERATOR 22006) ("[" 22007)) 8459 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 22002) (attribute 2733) (fn 19060) #'19061 (inline_function 22003)) 8460 (("(" 19262) (argument_list 22001)) 8461 ((T_PAAMAYIM_NEKUDOTAYIM 21993)) 8462 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 20410) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 8463 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 19363)) 8464 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 19351)) 8465 (("(" 19262) (argument_list 19350)) 8466 (("=" 19270) (T_AND_EQUAL 19271) (T_COALESCE_EQUAL 19272) (T_CONCAT_EQUAL 19273) (T_DEC 19274) (T_DIV_EQUAL 19275) (T_INC 19276) (T_MINUS_EQUAL 19277) (T_MOD_EQUAL 19278) (T_MUL_EQUAL 19279) (T_OR_EQUAL 19280) (T_PLUS_EQUAL 19281) (T_POW_EQUAL 19282) (T_SL_EQUAL 19283) (T_SR_EQUAL 19284) (T_XOR_EQUAL 19285)) 8467 ((T_PAAMAYIM_NEKUDOTAYIM 19174)) 8468 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19075) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8469 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 8470 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19173) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8471 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19172) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8472 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19171) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8473 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19170) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8474 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19169) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8475 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19168) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8476 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19167) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8477 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19166) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8478 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 19161) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 19162) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 8479 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19160) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8480 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19159) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8481 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19158) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8482 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19157) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8483 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19156) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8484 (("$" 19121) ("(" 19122) (T_NAME_FULLY_QUALIFIED 19123) (T_NAME_QUALIFIED 19124) (T_NAME_RELATIVE 19125) (T_STATIC 19126) (T_STRING 19127) (T_VARIABLE 19128) (class_name 19129) (class_name_reference 19130) (name 19131) (new_variable 19132) (simple_variable 19133)) 8485 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19120) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8486 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19119) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8487 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19118) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8488 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19117) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8489 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19116) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8490 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19115) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8491 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19114) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8492 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19113) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8493 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19112) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8494 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19111) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8495 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19110) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8496 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19109) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8497 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19108) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8498 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19107) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8499 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19106) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8500 (("$" 19121) (T_VARIABLE 19128) (simple_variable 19152) ("{" 19153)) 8501 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 19150) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8502 ((T_PAAMAYIM_NEKUDOTAYIM 19148)) 8503 ((T_NULLSAFE_OBJECT_OPERATOR 19134) (T_OBJECT_OPERATOR 19135) (T_PAAMAYIM_NEKUDOTAYIM 19136) ("[" 19137)) 8504 (("$" 19121) (T_STRING 19141) (T_VARIABLE 19128) (property_name 19147) (simple_variable 19143) ("{" 19144)) 8505 (("$" 19121) (T_STRING 19141) (T_VARIABLE 19128) (property_name 19142) (simple_variable 19143) ("{" 19144)) 8506 (("$" 19121) (T_VARIABLE 19128) (simple_variable 19140)) 8507 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 19138) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8508 (("]" 19139)) 8509 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 19145) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8510 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 19146)) 8511 (("$" 19121) (T_VARIABLE 19128) (simple_variable 19149)) 8512 (("%" 299) (")" 19151) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8513 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 19154) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8514 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 19155)) 8515 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19165) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8516 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 19163) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 8517 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19164) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8518 (("$" 18998) (T_ABSTRACT 19175) (T_ARRAY 19176) (T_AS 19177) (T_BREAK 19178) (T_CALLABLE 19179) (T_CASE 19180) (T_CATCH 19181) (T_CLASS 19182) (T_CLASS_C 19183) (T_CLONE 19184) (T_CONST 19185) (T_CONTINUE 19186) (T_DECLARE 19187) (T_DEFAULT 19188) (T_DIR 19189) (T_DO 19190) (T_ECHO 19191) (T_ELSE 19192) (T_ELSEIF 19193) (T_EMPTY 19194) (T_ENDDECLARE 19195) (T_ENDFOR 19196) (T_ENDFOREACH 19197) (T_ENDIF 19198) (T_ENDSWITCH 19199) (T_ENDWHILE 19200) (T_ENUM 19201) (T_EVAL 19202) (T_EXIT 19203) (T_EXTENDS 19204) (T_FILE 19205) (T_FINAL 19206) (T_FINALLY 19207) (T_FN 19208) (T_FOR 19209) (T_FOREACH 19210) (T_FUNCTION 19211) (T_FUNC_C 19212) (T_GLOBAL 19213) (T_GOTO 19214) (T_IF 19215) (T_IMPLEMENTS 19216) (T_INCLUDE 19217) (T_INCLUDE_ONCE 19218) (T_INSTANCEOF 19219) (T_INSTEADOF 19220) (T_INTERFACE 19221) (T_ISSET 19222) (T_LINE 19223) (T_LIST 19224) (T_LOGICAL_AND 19225) (T_LOGICAL_OR 19226) (T_LOGICAL_XOR 19227) (T_MATCH 19228) (T_METHOD_C 19229) (T_NAMESPACE 19230) (T_NEW 19231) (T_NS_C 19232) (T_PRINT 19233) (T_PRIVATE 19234) (T_PROPERTY_C 19235) (T_PROTECTED 19236) (T_PUBLIC 19237) (T_READONLY 19238) (T_REQUIRE 19239) (T_REQUIRE_ONCE 19240) (T_RETURN 19241) (T_STATIC 19242) (T_STRING 19243) (T_SWITCH 19244) (T_THROW 19245) (T_TRAIT 19246) (T_TRAIT_C 19247) (T_TRY 19248) (T_UNSET 19249) (T_USE 19250) (T_VAR 19251) (T_VARIABLE 19046) (T_WHILE 19252) (T_YIELD 19253) (identifier 19254) (member_name 19255) (reserved_non_modifiers 19256) (semi_reserved 19257) (simple_variable 19258) ("{" 19259)) 8519 (("(" 19262) (argument_list 19263)) 8520 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 19260) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8521 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 19261)) 8522 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 19264) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 19265) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 19266) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8523 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 19269) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8524 (("," 1480) (possible_comma 19267)) 8525 ((")" 19268)) 8526 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (ampersand 19299) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19300) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8527 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19298) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8528 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19297) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8529 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19296) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8530 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19295) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8531 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19294) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8532 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19293) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8533 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19292) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8534 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19291) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8535 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19290) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8536 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19289) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8537 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19288) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8538 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19287) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8539 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 19286) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 8540 (("\"" 1520) ("$" 19301) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 19302) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 19303) ("[" 1542) (array_object_dereferenceable 19304) (callable_expr 19305) (callable_variable 19306) (class_constant 1546) (class_name 19307) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 19308) (name 19309) (new_dereferenceable 1552) (simple_variable 19310) (static_member 19311) (variable 19312) (variable_class_name 19313)) 8541 (("$" 19301) (T_VARIABLE 19303) (simple_variable 19346) ("{" 19347)) 8542 (("(" 19317) (argument_list 19345)) 8543 ((T_NULLSAFE_OBJECT_OPERATOR 19331) (T_OBJECT_OPERATOR 19332) ("[" 19333)) 8544 (("(" 19317) (argument_list 19330)) 8545 ((T_PAAMAYIM_NEKUDOTAYIM 19326)) 8546 (("(" 19317) (argument_list 19325)) 8547 ((T_PAAMAYIM_NEKUDOTAYIM 19314)) 8548 (("$" 19301) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 19303) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 19315) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 19316) ("{" 1642)) 8549 (("(" 19317) (argument_list 19318)) 8550 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 19319) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 19320) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 19321) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8551 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 19324) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8552 (("," 1480) (possible_comma 19322)) 8553 ((")" 19323)) 8554 (("$" 19301) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 19303) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 19327) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 19328) ("{" 1658)) 8555 (("(" 19317) (argument_list 19329)) 8556 (("$" 19301) (T_STRING 19336) (T_VARIABLE 19303) (property_name 19343) (simple_variable 19338) ("{" 19339)) 8557 (("$" 19301) (T_STRING 19336) (T_VARIABLE 19303) (property_name 19337) (simple_variable 19338) ("{" 19339)) 8558 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 19334) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8559 (("]" 19335)) 8560 (("(" 19317) (argument_list 19342)) 8561 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 19340) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8562 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 19341)) 8563 (("(" 19317) (argument_list 19344)) 8564 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 19348) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8565 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 19349)) 8566 ((backup_doc_comment 19352)) 8567 (("(" 19353)) 8568 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 19354)) 8569 ((")" 19355)) 8570 ((T_USE 164) (lexical_vars 19356)) 8571 ((":" 166) (return_type 19357)) 8572 ((backup_fn_flags 19358)) 8573 (("{" 19359)) 8574 ((inner_statement_list 19360)) 8575 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 19361) ("~" 127)) 8576 ((backup_fn_flags 19362)) 8577 ((backup_doc_comment 19364)) 8578 (("(" 19365)) 8579 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 19366)) 8580 ((")" 19367)) 8581 ((":" 2249) (return_type 19368)) 8582 ((T_DOUBLE_ARROW 19369)) 8583 ((backup_fn_flags 19370)) 8584 ((backup_lex_pos 19371)) 8585 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19435) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8586 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20409) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8587 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 20407) (encaps_var 5237)) 8588 (("$" 19374) (T_VARIABLE 19422) (simple_variable 20403) ("{" 20404)) 8589 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20401) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8590 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20400) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8591 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20399) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8592 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20398) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8593 (("(" 20395)) 8594 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20394) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8595 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20393) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8596 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20392) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8597 (("\"" 1520) ("$" 19677) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 19678) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 19679) ("[" 1542) (array_object_dereferenceable 19680) (callable_expr 19681) (callable_variable 19682) (class_constant 1546) (class_name 19683) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 19684) (name 19685) (new_dereferenceable 1552) (simple_variable 19686) (static_member 19687) (variable 20391) (variable_class_name 19689)) 8598 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20390) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8599 (("(" 20387)) 8600 (("(" 20384)) 8601 (("(" 20375) (argument_list 20376) (ctor_arguments 20377)) 8602 (("\"" 1520) ("$" 19677) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 19678) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 19679) ("[" 1542) (array_object_dereferenceable 19680) (callable_expr 19681) (callable_variable 19682) (class_constant 1546) (class_name 19683) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 19684) (name 19685) (new_dereferenceable 1552) (simple_variable 19686) (static_member 19687) (variable 20374) (variable_class_name 19689)) 8603 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20373) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8604 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20372) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8605 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20371) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8606 (("(" 20367)) 8607 (("(" 20362)) 8608 (("(" 20356)) 8609 (("$" 19677) ("(" 20319) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 20320) (T_NAME_QUALIFIED 20321) (T_NAME_RELATIVE 20322) (T_READONLY 2931) (T_STATIC 20323) (T_STRING 20324) (T_VARIABLE 19679) (anonymous_class 20325) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 20326) (attribute 2937) (attributes 20327) (class_modifier 2939) (class_name 20328) (class_name_reference 20329) (name 20330) (new_variable 20331) (simple_variable 20332)) 8610 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20318) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8611 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20317) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8612 (("(" 19638) (argument_list 20316)) 8613 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20315) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8614 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20314) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8615 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 20309) (T_END_HEREDOC 20310) (T_VARIABLE 2896) (encaps_list 20311) (encaps_var 2898)) 8616 ((T_FN 35) (T_FUNCTION 129) (fn 19436) #'19437 (inline_function 20308)) 8617 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20307) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8618 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20306) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8619 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20305) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8620 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19847) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8621 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19783) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8622 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 19779) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 8623 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 19777) (encaps_list 2804) (encaps_var 2805)) 8624 ((T_NULLSAFE_OBJECT_OPERATOR 19763) (T_OBJECT_OPERATOR 19764) ("[" 19765)) 8625 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 19760) (attribute 2733) (fn 19436) #'19437 (inline_function 19761)) 8626 (("(" 19638) (argument_list 19759)) 8627 ((T_PAAMAYIM_NEKUDOTAYIM 19751)) 8628 (("%" 19452) ("*" 19453) ("+" 19454) ("-" 19455) ("." 19456) ("/" 19457) ("<" 19458) (">" 19459) ("?" 19460) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19461) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19462) (T_BOOLEAN_AND 19463) (T_BOOLEAN_OR 19464) (T_COALESCE 19465) (T_INSTANCEOF 19466) (T_IS_EQUAL 19467) (T_IS_GREATER_OR_EQUAL 19468) (T_IS_IDENTICAL 19469) (T_IS_NOT_EQUAL 19470) (T_IS_NOT_IDENTICAL 19471) (T_IS_SMALLER_OR_EQUAL 19472) (T_LOGICAL_AND 19473) (T_LOGICAL_OR 19474) (T_LOGICAL_XOR 19475) (T_POW 19476) (T_SL 19477) (T_SPACESHIP 19478) (T_SR 19479) ("^" 19480) (backup_fn_flags 19750) ("|" 19481)) 8629 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 19739)) 8630 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 19727)) 8631 (("(" 19638) (argument_list 19726)) 8632 (("=" 19646) (T_AND_EQUAL 19647) (T_COALESCE_EQUAL 19648) (T_CONCAT_EQUAL 19649) (T_DEC 19650) (T_DIV_EQUAL 19651) (T_INC 19652) (T_MINUS_EQUAL 19653) (T_MOD_EQUAL 19654) (T_MUL_EQUAL 19655) (T_OR_EQUAL 19656) (T_PLUS_EQUAL 19657) (T_POW_EQUAL 19658) (T_SL_EQUAL 19659) (T_SR_EQUAL 19660) (T_XOR_EQUAL 19661)) 8633 ((T_PAAMAYIM_NEKUDOTAYIM 19550)) 8634 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19451) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8635 (("%" 19452) ("*" 19453) ("+" 19454) ("-" 19455) ("." 19456) ("/" 19457) ("<" 19458) (">" 19459) ("?" 19460) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19461) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19462) (T_BOOLEAN_AND 19463) (T_BOOLEAN_OR 19464) (T_COALESCE 19465) (T_INSTANCEOF 19466) (T_IS_EQUAL 19467) (T_IS_GREATER_OR_EQUAL 19468) (T_IS_IDENTICAL 19469) (T_IS_NOT_EQUAL 19470) (T_IS_NOT_IDENTICAL 19471) (T_IS_SMALLER_OR_EQUAL 19472) (T_LOGICAL_AND 19473) (T_LOGICAL_OR 19474) (T_LOGICAL_XOR 19475) (T_POW 19476) (T_SL 19477) (T_SPACESHIP 19478) (T_SR 19479) ("^" 19480) ("|" 19481)) 8636 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19549) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8637 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19548) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8638 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19547) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8639 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19546) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8640 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19545) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8641 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19544) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8642 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19543) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8643 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19542) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8644 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 19537) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 19538) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 8645 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19536) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8646 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19535) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8647 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19534) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8648 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19533) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8649 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19532) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8650 (("$" 19497) ("(" 19498) (T_NAME_FULLY_QUALIFIED 19499) (T_NAME_QUALIFIED 19500) (T_NAME_RELATIVE 19501) (T_STATIC 19502) (T_STRING 19503) (T_VARIABLE 19504) (class_name 19505) (class_name_reference 19506) (name 19507) (new_variable 19508) (simple_variable 19509)) 8651 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19496) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8652 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19495) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8653 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19494) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8654 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19493) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8655 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19492) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8656 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19491) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8657 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19490) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8658 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19489) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8659 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19488) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8660 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19487) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8661 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19486) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8662 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19485) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8663 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19484) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8664 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19483) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8665 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19482) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8666 (("$" 19497) (T_VARIABLE 19504) (simple_variable 19528) ("{" 19529)) 8667 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 19526) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8668 ((T_PAAMAYIM_NEKUDOTAYIM 19524)) 8669 ((T_NULLSAFE_OBJECT_OPERATOR 19510) (T_OBJECT_OPERATOR 19511) (T_PAAMAYIM_NEKUDOTAYIM 19512) ("[" 19513)) 8670 (("$" 19497) (T_STRING 19517) (T_VARIABLE 19504) (property_name 19523) (simple_variable 19519) ("{" 19520)) 8671 (("$" 19497) (T_STRING 19517) (T_VARIABLE 19504) (property_name 19518) (simple_variable 19519) ("{" 19520)) 8672 (("$" 19497) (T_VARIABLE 19504) (simple_variable 19516)) 8673 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 19514) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8674 (("]" 19515)) 8675 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 19521) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8676 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 19522)) 8677 (("$" 19497) (T_VARIABLE 19504) (simple_variable 19525)) 8678 (("%" 299) (")" 19527) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8679 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 19530) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8680 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 19531)) 8681 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19541) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8682 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 19539) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 8683 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19540) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8684 (("$" 19374) (T_ABSTRACT 19551) (T_ARRAY 19552) (T_AS 19553) (T_BREAK 19554) (T_CALLABLE 19555) (T_CASE 19556) (T_CATCH 19557) (T_CLASS 19558) (T_CLASS_C 19559) (T_CLONE 19560) (T_CONST 19561) (T_CONTINUE 19562) (T_DECLARE 19563) (T_DEFAULT 19564) (T_DIR 19565) (T_DO 19566) (T_ECHO 19567) (T_ELSE 19568) (T_ELSEIF 19569) (T_EMPTY 19570) (T_ENDDECLARE 19571) (T_ENDFOR 19572) (T_ENDFOREACH 19573) (T_ENDIF 19574) (T_ENDSWITCH 19575) (T_ENDWHILE 19576) (T_ENUM 19577) (T_EVAL 19578) (T_EXIT 19579) (T_EXTENDS 19580) (T_FILE 19581) (T_FINAL 19582) (T_FINALLY 19583) (T_FN 19584) (T_FOR 19585) (T_FOREACH 19586) (T_FUNCTION 19587) (T_FUNC_C 19588) (T_GLOBAL 19589) (T_GOTO 19590) (T_IF 19591) (T_IMPLEMENTS 19592) (T_INCLUDE 19593) (T_INCLUDE_ONCE 19594) (T_INSTANCEOF 19595) (T_INSTEADOF 19596) (T_INTERFACE 19597) (T_ISSET 19598) (T_LINE 19599) (T_LIST 19600) (T_LOGICAL_AND 19601) (T_LOGICAL_OR 19602) (T_LOGICAL_XOR 19603) (T_MATCH 19604) (T_METHOD_C 19605) (T_NAMESPACE 19606) (T_NEW 19607) (T_NS_C 19608) (T_PRINT 19609) (T_PRIVATE 19610) (T_PROPERTY_C 19611) (T_PROTECTED 19612) (T_PUBLIC 19613) (T_READONLY 19614) (T_REQUIRE 19615) (T_REQUIRE_ONCE 19616) (T_RETURN 19617) (T_STATIC 19618) (T_STRING 19619) (T_SWITCH 19620) (T_THROW 19621) (T_TRAIT 19622) (T_TRAIT_C 19623) (T_TRY 19624) (T_UNSET 19625) (T_USE 19626) (T_VAR 19627) (T_VARIABLE 19422) (T_WHILE 19628) (T_YIELD 19629) (identifier 19630) (member_name 19631) (reserved_non_modifiers 19632) (semi_reserved 19633) (simple_variable 19634) ("{" 19635)) 8685 (("(" 19638) (argument_list 19639)) 8686 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 19636) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8687 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 19637)) 8688 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 19640) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 19641) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 19642) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8689 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 19645) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8690 (("," 1480) (possible_comma 19643)) 8691 ((")" 19644)) 8692 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (ampersand 19675) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19676) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8693 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19674) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8694 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19673) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8695 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19672) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8696 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19671) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8697 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19670) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8698 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19669) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8699 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19668) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8700 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19667) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8701 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19666) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8702 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19665) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8703 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19664) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8704 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19663) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8705 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19662) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8706 (("\"" 1520) ("$" 19677) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 19678) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 19679) ("[" 1542) (array_object_dereferenceable 19680) (callable_expr 19681) (callable_variable 19682) (class_constant 1546) (class_name 19683) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 19684) (name 19685) (new_dereferenceable 1552) (simple_variable 19686) (static_member 19687) (variable 19688) (variable_class_name 19689)) 8707 (("$" 19677) (T_VARIABLE 19679) (simple_variable 19722) ("{" 19723)) 8708 (("(" 19693) (argument_list 19721)) 8709 ((T_NULLSAFE_OBJECT_OPERATOR 19707) (T_OBJECT_OPERATOR 19708) ("[" 19709)) 8710 (("(" 19693) (argument_list 19706)) 8711 ((T_PAAMAYIM_NEKUDOTAYIM 19702)) 8712 (("(" 19693) (argument_list 19701)) 8713 ((T_PAAMAYIM_NEKUDOTAYIM 19690)) 8714 (("$" 19677) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 19679) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 19691) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 19692) ("{" 1642)) 8715 (("(" 19693) (argument_list 19694)) 8716 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 19695) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 19696) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 19697) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8717 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 19700) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8718 (("," 1480) (possible_comma 19698)) 8719 ((")" 19699)) 8720 (("$" 19677) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 19679) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 19703) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 19704) ("{" 1658)) 8721 (("(" 19693) (argument_list 19705)) 8722 (("$" 19677) (T_STRING 19712) (T_VARIABLE 19679) (property_name 19719) (simple_variable 19714) ("{" 19715)) 8723 (("$" 19677) (T_STRING 19712) (T_VARIABLE 19679) (property_name 19713) (simple_variable 19714) ("{" 19715)) 8724 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 19710) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8725 (("]" 19711)) 8726 (("(" 19693) (argument_list 19718)) 8727 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 19716) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8728 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 19717)) 8729 (("(" 19693) (argument_list 19720)) 8730 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 19724) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8731 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 19725)) 8732 ((backup_doc_comment 19728)) 8733 (("(" 19729)) 8734 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 19730)) 8735 ((")" 19731)) 8736 ((T_USE 164) (lexical_vars 19732)) 8737 ((":" 166) (return_type 19733)) 8738 ((backup_fn_flags 19734)) 8739 (("{" 19735)) 8740 ((inner_statement_list 19736)) 8741 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 19737) ("~" 127)) 8742 ((backup_fn_flags 19738)) 8743 ((backup_doc_comment 19740)) 8744 (("(" 19741)) 8745 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 19742)) 8746 ((")" 19743)) 8747 ((":" 2249) (return_type 19744)) 8748 ((T_DOUBLE_ARROW 19745)) 8749 ((backup_fn_flags 19746)) 8750 ((backup_lex_pos 19747)) 8751 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19748) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8752 (("%" 19452) ("*" 19453) ("+" 19454) ("-" 19455) ("." 19456) ("/" 19457) ("<" 19458) (">" 19459) ("?" 19460) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19461) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19462) (T_BOOLEAN_AND 19463) (T_BOOLEAN_OR 19464) (T_COALESCE 19465) (T_INSTANCEOF 19466) (T_IS_EQUAL 19467) (T_IS_GREATER_OR_EQUAL 19468) (T_IS_IDENTICAL 19469) (T_IS_NOT_EQUAL 19470) (T_IS_NOT_IDENTICAL 19471) (T_IS_SMALLER_OR_EQUAL 19472) (T_LOGICAL_AND 19473) (T_LOGICAL_OR 19474) (T_LOGICAL_XOR 19475) (T_POW 19476) (T_SL 19477) (T_SPACESHIP 19478) (T_SR 19479) ("^" 19480) (backup_fn_flags 19749) ("|" 19481)) 8753 (("$" 19374) (T_ABSTRACT 19551) (T_ARRAY 19552) (T_AS 19553) (T_BREAK 19554) (T_CALLABLE 19555) (T_CASE 19556) (T_CATCH 19557) (T_CLASS 19558) (T_CLASS_C 19559) (T_CLONE 19560) (T_CONST 19561) (T_CONTINUE 19562) (T_DECLARE 19563) (T_DEFAULT 19564) (T_DIR 19565) (T_DO 19566) (T_ECHO 19567) (T_ELSE 19568) (T_ELSEIF 19569) (T_EMPTY 19570) (T_ENDDECLARE 19571) (T_ENDFOR 19572) (T_ENDFOREACH 19573) (T_ENDIF 19574) (T_ENDSWITCH 19575) (T_ENDWHILE 19576) (T_ENUM 19577) (T_EVAL 19578) (T_EXIT 19579) (T_EXTENDS 19580) (T_FILE 19581) (T_FINAL 19582) (T_FINALLY 19583) (T_FN 19584) (T_FOR 19585) (T_FOREACH 19586) (T_FUNCTION 19587) (T_FUNC_C 19588) (T_GLOBAL 19589) (T_GOTO 19590) (T_IF 19591) (T_IMPLEMENTS 19592) (T_INCLUDE 19593) (T_INCLUDE_ONCE 19594) (T_INSTANCEOF 19595) (T_INSTEADOF 19596) (T_INTERFACE 19597) (T_ISSET 19598) (T_LINE 19599) (T_LIST 19600) (T_LOGICAL_AND 19601) (T_LOGICAL_OR 19602) (T_LOGICAL_XOR 19603) (T_MATCH 19604) (T_METHOD_C 19605) (T_NAMESPACE 19606) (T_NEW 19607) (T_NS_C 19608) (T_PRINT 19609) (T_PRIVATE 19610) (T_PROPERTY_C 19611) (T_PROTECTED 19612) (T_PUBLIC 19613) (T_READONLY 19614) (T_REQUIRE 19615) (T_REQUIRE_ONCE 19616) (T_RETURN 19617) (T_STATIC 19618) (T_STRING 19619) (T_SWITCH 19620) (T_THROW 19621) (T_TRAIT 19622) (T_TRAIT_C 19623) (T_TRY 19624) (T_UNSET 19625) (T_USE 19626) (T_VAR 19627) (T_VARIABLE 19422) (T_WHILE 19628) (T_YIELD 19629) (identifier 19752) (member_name 19753) (reserved_non_modifiers 19632) (semi_reserved 19633) (simple_variable 19754) ("{" 19755)) 8754 (("(" 19638) (argument_list 19758)) 8755 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 19756) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8756 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 19757)) 8757 ((T_FN 35) (T_FUNCTION 129) (fn 19436) #'19437 (inline_function 19762)) 8758 (("$" 19374) (T_STRING 19768) (T_VARIABLE 19422) (property_name 19775) (simple_variable 19770) ("{" 19771)) 8759 (("$" 19374) (T_STRING 19768) (T_VARIABLE 19422) (property_name 19769) (simple_variable 19770) ("{" 19771)) 8760 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 19766) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8761 (("]" 19767)) 8762 (("(" 19638) (argument_list 19774)) 8763 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 19772) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8764 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 19773)) 8765 (("(" 19638) (argument_list 19776)) 8766 (("`" 19778)) 8767 (("]" 19780)) 8768 (("=" 19781)) 8769 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 19782) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8770 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20304) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8771 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 20302) (encaps_var 5237)) 8772 (("$" 19786) (T_VARIABLE 19834) (simple_variable 20298) ("{" 20299)) 8773 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20296) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8774 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20295) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8775 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20294) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8776 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20293) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8777 (("(" 20290)) 8778 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20289) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8779 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20288) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8780 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20287) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8781 (("\"" 1520) ("$" 20089) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20090) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20091) ("[" 1542) (array_object_dereferenceable 20092) (callable_expr 20093) (callable_variable 20094) (class_constant 1546) (class_name 20095) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 20096) (name 20097) (new_dereferenceable 1552) (simple_variable 20098) (static_member 20099) (variable 20286) (variable_class_name 20101)) 8782 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20285) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8783 (("(" 20282)) 8784 (("(" 20279)) 8785 (("(" 20270) (argument_list 20271) (ctor_arguments 20272)) 8786 (("\"" 1520) ("$" 20089) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20090) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20091) ("[" 1542) (array_object_dereferenceable 20092) (callable_expr 20093) (callable_variable 20094) (class_constant 1546) (class_name 20095) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 20096) (name 20097) (new_dereferenceable 1552) (simple_variable 20098) (static_member 20099) (variable 20269) (variable_class_name 20101)) 8787 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20268) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8788 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20267) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8789 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20266) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8790 (("(" 20262)) 8791 (("(" 20257)) 8792 (("(" 20251)) 8793 (("$" 20089) ("(" 20214) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 20215) (T_NAME_QUALIFIED 20216) (T_NAME_RELATIVE 20217) (T_READONLY 2931) (T_STATIC 20218) (T_STRING 20219) (T_VARIABLE 20091) (anonymous_class 20220) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 20221) (attribute 2937) (attributes 20222) (class_modifier 2939) (class_name 20223) (class_name_reference 20224) (name 20225) (new_variable 20226) (simple_variable 20227)) 8794 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20213) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8795 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20212) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8796 (("(" 20050) (argument_list 20211)) 8797 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20210) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8798 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20209) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8799 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 20204) (T_END_HEREDOC 20205) (T_VARIABLE 2896) (encaps_list 20206) (encaps_var 2898)) 8800 ((T_FN 35) (T_FUNCTION 129) (fn 19848) #'19849 (inline_function 20203)) 8801 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20202) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8802 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20201) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8803 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20200) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8804 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20197) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8805 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20196) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8806 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 20192) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 8807 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 20190) (encaps_list 2804) (encaps_var 2805)) 8808 ((T_NULLSAFE_OBJECT_OPERATOR 20176) (T_OBJECT_OPERATOR 20177) ("[" 20178)) 8809 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 20173) (attribute 2733) (fn 19848) #'19849 (inline_function 20174)) 8810 (("(" 20050) (argument_list 20172)) 8811 ((T_PAAMAYIM_NEKUDOTAYIM 20164)) 8812 (("%" 19864) ("*" 19865) ("+" 19866) ("-" 19867) ("." 19868) ("/" 19869) ("<" 19870) (">" 19871) ("?" 19872) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19873) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19874) (T_BOOLEAN_AND 19875) (T_BOOLEAN_OR 19876) (T_COALESCE 19877) (T_DOUBLE_ARROW 20162) (T_INSTANCEOF 19878) (T_IS_EQUAL 19879) (T_IS_GREATER_OR_EQUAL 19880) (T_IS_IDENTICAL 19881) (T_IS_NOT_EQUAL 19882) (T_IS_NOT_IDENTICAL 19883) (T_IS_SMALLER_OR_EQUAL 19884) (T_LOGICAL_AND 19885) (T_LOGICAL_OR 19886) (T_LOGICAL_XOR 19887) (T_POW 19888) (T_SL 19889) (T_SPACESHIP 19890) (T_SR 19891) ("^" 19892) ("|" 19893)) 8813 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 20151)) 8814 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 20139)) 8815 (("(" 20050) (argument_list 20138)) 8816 (("=" 20058) (T_AND_EQUAL 20059) (T_COALESCE_EQUAL 20060) (T_CONCAT_EQUAL 20061) (T_DEC 20062) (T_DIV_EQUAL 20063) (T_INC 20064) (T_MINUS_EQUAL 20065) (T_MOD_EQUAL 20066) (T_MUL_EQUAL 20067) (T_OR_EQUAL 20068) (T_PLUS_EQUAL 20069) (T_POW_EQUAL 20070) (T_SL_EQUAL 20071) (T_SR_EQUAL 20072) (T_XOR_EQUAL 20073)) 8817 ((T_PAAMAYIM_NEKUDOTAYIM 19962)) 8818 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19863) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8819 (("%" 19864) ("*" 19865) ("+" 19866) ("-" 19867) ("." 19868) ("/" 19869) ("<" 19870) (">" 19871) ("?" 19872) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19873) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19874) (T_BOOLEAN_AND 19875) (T_BOOLEAN_OR 19876) (T_COALESCE 19877) (T_INSTANCEOF 19878) (T_IS_EQUAL 19879) (T_IS_GREATER_OR_EQUAL 19880) (T_IS_IDENTICAL 19881) (T_IS_NOT_EQUAL 19882) (T_IS_NOT_IDENTICAL 19883) (T_IS_SMALLER_OR_EQUAL 19884) (T_LOGICAL_AND 19885) (T_LOGICAL_OR 19886) (T_LOGICAL_XOR 19887) (T_POW 19888) (T_SL 19889) (T_SPACESHIP 19890) (T_SR 19891) ("^" 19892) ("|" 19893)) 8820 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19961) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8821 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19960) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8822 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19959) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8823 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19958) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8824 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19957) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8825 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19956) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8826 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19955) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8827 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19954) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8828 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 19949) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 19950) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 8829 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19948) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8830 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19947) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8831 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19946) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8832 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19945) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8833 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19944) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8834 (("$" 19909) ("(" 19910) (T_NAME_FULLY_QUALIFIED 19911) (T_NAME_QUALIFIED 19912) (T_NAME_RELATIVE 19913) (T_STATIC 19914) (T_STRING 19915) (T_VARIABLE 19916) (class_name 19917) (class_name_reference 19918) (name 19919) (new_variable 19920) (simple_variable 19921)) 8835 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19908) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8836 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19907) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8837 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19906) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8838 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19905) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8839 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19904) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8840 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19903) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8841 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19902) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8842 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19901) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8843 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19900) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8844 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19899) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8845 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19898) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8846 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19897) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8847 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19896) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8848 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19895) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8849 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19894) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8850 (("$" 19909) (T_VARIABLE 19916) (simple_variable 19940) ("{" 19941)) 8851 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 19938) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8852 ((T_PAAMAYIM_NEKUDOTAYIM 19936)) 8853 ((T_NULLSAFE_OBJECT_OPERATOR 19922) (T_OBJECT_OPERATOR 19923) (T_PAAMAYIM_NEKUDOTAYIM 19924) ("[" 19925)) 8854 (("$" 19909) (T_STRING 19929) (T_VARIABLE 19916) (property_name 19935) (simple_variable 19931) ("{" 19932)) 8855 (("$" 19909) (T_STRING 19929) (T_VARIABLE 19916) (property_name 19930) (simple_variable 19931) ("{" 19932)) 8856 (("$" 19909) (T_VARIABLE 19916) (simple_variable 19928)) 8857 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 19926) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8858 (("]" 19927)) 8859 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 19933) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8860 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 19934)) 8861 (("$" 19909) (T_VARIABLE 19916) (simple_variable 19937)) 8862 (("%" 299) (")" 19939) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8863 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 19942) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8864 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 19943)) 8865 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19953) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8866 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 19951) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 8867 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 19952) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8868 (("$" 19786) (T_ABSTRACT 19963) (T_ARRAY 19964) (T_AS 19965) (T_BREAK 19966) (T_CALLABLE 19967) (T_CASE 19968) (T_CATCH 19969) (T_CLASS 19970) (T_CLASS_C 19971) (T_CLONE 19972) (T_CONST 19973) (T_CONTINUE 19974) (T_DECLARE 19975) (T_DEFAULT 19976) (T_DIR 19977) (T_DO 19978) (T_ECHO 19979) (T_ELSE 19980) (T_ELSEIF 19981) (T_EMPTY 19982) (T_ENDDECLARE 19983) (T_ENDFOR 19984) (T_ENDFOREACH 19985) (T_ENDIF 19986) (T_ENDSWITCH 19987) (T_ENDWHILE 19988) (T_ENUM 19989) (T_EVAL 19990) (T_EXIT 19991) (T_EXTENDS 19992) (T_FILE 19993) (T_FINAL 19994) (T_FINALLY 19995) (T_FN 19996) (T_FOR 19997) (T_FOREACH 19998) (T_FUNCTION 19999) (T_FUNC_C 20000) (T_GLOBAL 20001) (T_GOTO 20002) (T_IF 20003) (T_IMPLEMENTS 20004) (T_INCLUDE 20005) (T_INCLUDE_ONCE 20006) (T_INSTANCEOF 20007) (T_INSTEADOF 20008) (T_INTERFACE 20009) (T_ISSET 20010) (T_LINE 20011) (T_LIST 20012) (T_LOGICAL_AND 20013) (T_LOGICAL_OR 20014) (T_LOGICAL_XOR 20015) (T_MATCH 20016) (T_METHOD_C 20017) (T_NAMESPACE 20018) (T_NEW 20019) (T_NS_C 20020) (T_PRINT 20021) (T_PRIVATE 20022) (T_PROPERTY_C 20023) (T_PROTECTED 20024) (T_PUBLIC 20025) (T_READONLY 20026) (T_REQUIRE 20027) (T_REQUIRE_ONCE 20028) (T_RETURN 20029) (T_STATIC 20030) (T_STRING 20031) (T_SWITCH 20032) (T_THROW 20033) (T_TRAIT 20034) (T_TRAIT_C 20035) (T_TRY 20036) (T_UNSET 20037) (T_USE 20038) (T_VAR 20039) (T_VARIABLE 19834) (T_WHILE 20040) (T_YIELD 20041) (identifier 20042) (member_name 20043) (reserved_non_modifiers 20044) (semi_reserved 20045) (simple_variable 20046) ("{" 20047)) 8869 (("(" 20050) (argument_list 20051)) 8870 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 20048) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8871 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 20049)) 8872 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 20052) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 20053) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 20054) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8873 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 20057) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8874 (("," 1480) (possible_comma 20055)) 8875 ((")" 20056)) 8876 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (ampersand 20087) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20088) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8877 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20086) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8878 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20085) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8879 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20084) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8880 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20083) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8881 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20082) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8882 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20081) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8883 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20080) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8884 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20079) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8885 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20078) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8886 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20077) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8887 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20076) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8888 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20075) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8889 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20074) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8890 (("\"" 1520) ("$" 20089) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20090) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20091) ("[" 1542) (array_object_dereferenceable 20092) (callable_expr 20093) (callable_variable 20094) (class_constant 1546) (class_name 20095) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 20096) (name 20097) (new_dereferenceable 1552) (simple_variable 20098) (static_member 20099) (variable 20100) (variable_class_name 20101)) 8891 (("$" 20089) (T_VARIABLE 20091) (simple_variable 20134) ("{" 20135)) 8892 (("(" 20105) (argument_list 20133)) 8893 ((T_NULLSAFE_OBJECT_OPERATOR 20119) (T_OBJECT_OPERATOR 20120) ("[" 20121)) 8894 (("(" 20105) (argument_list 20118)) 8895 ((T_PAAMAYIM_NEKUDOTAYIM 20114)) 8896 (("(" 20105) (argument_list 20113)) 8897 ((T_PAAMAYIM_NEKUDOTAYIM 20102)) 8898 (("$" 20089) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 20091) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 20103) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 20104) ("{" 1642)) 8899 (("(" 20105) (argument_list 20106)) 8900 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 20107) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 20108) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 20109) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8901 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 20112) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8902 (("," 1480) (possible_comma 20110)) 8903 ((")" 20111)) 8904 (("$" 20089) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 20091) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 20115) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 20116) ("{" 1658)) 8905 (("(" 20105) (argument_list 20117)) 8906 (("$" 20089) (T_STRING 20124) (T_VARIABLE 20091) (property_name 20131) (simple_variable 20126) ("{" 20127)) 8907 (("$" 20089) (T_STRING 20124) (T_VARIABLE 20091) (property_name 20125) (simple_variable 20126) ("{" 20127)) 8908 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 20122) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8909 (("]" 20123)) 8910 (("(" 20105) (argument_list 20130)) 8911 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 20128) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8912 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 20129)) 8913 (("(" 20105) (argument_list 20132)) 8914 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 20136) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8915 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 20137)) 8916 ((backup_doc_comment 20140)) 8917 (("(" 20141)) 8918 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 20142)) 8919 ((")" 20143)) 8920 ((T_USE 164) (lexical_vars 20144)) 8921 ((":" 166) (return_type 20145)) 8922 ((backup_fn_flags 20146)) 8923 (("{" 20147)) 8924 ((inner_statement_list 20148)) 8925 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20149) ("~" 127)) 8926 ((backup_fn_flags 20150)) 8927 ((backup_doc_comment 20152)) 8928 (("(" 20153)) 8929 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 20154)) 8930 ((")" 20155)) 8931 ((":" 2249) (return_type 20156)) 8932 ((T_DOUBLE_ARROW 20157)) 8933 ((backup_fn_flags 20158)) 8934 ((backup_lex_pos 20159)) 8935 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20160) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8936 (("%" 19864) ("*" 19865) ("+" 19866) ("-" 19867) ("." 19868) ("/" 19869) ("<" 19870) (">" 19871) ("?" 19872) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19873) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19874) (T_BOOLEAN_AND 19875) (T_BOOLEAN_OR 19876) (T_COALESCE 19877) (T_INSTANCEOF 19878) (T_IS_EQUAL 19879) (T_IS_GREATER_OR_EQUAL 19880) (T_IS_IDENTICAL 19881) (T_IS_NOT_EQUAL 19882) (T_IS_NOT_IDENTICAL 19883) (T_IS_SMALLER_OR_EQUAL 19884) (T_LOGICAL_AND 19885) (T_LOGICAL_OR 19886) (T_LOGICAL_XOR 19887) (T_POW 19888) (T_SL 19889) (T_SPACESHIP 19890) (T_SR 19891) ("^" 19892) (backup_fn_flags 20161) ("|" 19893)) 8937 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20163) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 8938 (("$" 19786) (T_ABSTRACT 19963) (T_ARRAY 19964) (T_AS 19965) (T_BREAK 19966) (T_CALLABLE 19967) (T_CASE 19968) (T_CATCH 19969) (T_CLASS 19970) (T_CLASS_C 19971) (T_CLONE 19972) (T_CONST 19973) (T_CONTINUE 19974) (T_DECLARE 19975) (T_DEFAULT 19976) (T_DIR 19977) (T_DO 19978) (T_ECHO 19979) (T_ELSE 19980) (T_ELSEIF 19981) (T_EMPTY 19982) (T_ENDDECLARE 19983) (T_ENDFOR 19984) (T_ENDFOREACH 19985) (T_ENDIF 19986) (T_ENDSWITCH 19987) (T_ENDWHILE 19988) (T_ENUM 19989) (T_EVAL 19990) (T_EXIT 19991) (T_EXTENDS 19992) (T_FILE 19993) (T_FINAL 19994) (T_FINALLY 19995) (T_FN 19996) (T_FOR 19997) (T_FOREACH 19998) (T_FUNCTION 19999) (T_FUNC_C 20000) (T_GLOBAL 20001) (T_GOTO 20002) (T_IF 20003) (T_IMPLEMENTS 20004) (T_INCLUDE 20005) (T_INCLUDE_ONCE 20006) (T_INSTANCEOF 20007) (T_INSTEADOF 20008) (T_INTERFACE 20009) (T_ISSET 20010) (T_LINE 20011) (T_LIST 20012) (T_LOGICAL_AND 20013) (T_LOGICAL_OR 20014) (T_LOGICAL_XOR 20015) (T_MATCH 20016) (T_METHOD_C 20017) (T_NAMESPACE 20018) (T_NEW 20019) (T_NS_C 20020) (T_PRINT 20021) (T_PRIVATE 20022) (T_PROPERTY_C 20023) (T_PROTECTED 20024) (T_PUBLIC 20025) (T_READONLY 20026) (T_REQUIRE 20027) (T_REQUIRE_ONCE 20028) (T_RETURN 20029) (T_STATIC 20030) (T_STRING 20031) (T_SWITCH 20032) (T_THROW 20033) (T_TRAIT 20034) (T_TRAIT_C 20035) (T_TRY 20036) (T_UNSET 20037) (T_USE 20038) (T_VAR 20039) (T_VARIABLE 19834) (T_WHILE 20040) (T_YIELD 20041) (identifier 20165) (member_name 20166) (reserved_non_modifiers 20044) (semi_reserved 20045) (simple_variable 20167) ("{" 20168)) 8939 (("(" 20050) (argument_list 20171)) 8940 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 20169) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8941 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 20170)) 8942 ((T_FN 35) (T_FUNCTION 129) (fn 19848) #'19849 (inline_function 20175)) 8943 (("$" 19786) (T_STRING 20181) (T_VARIABLE 19834) (property_name 20188) (simple_variable 20183) ("{" 20184)) 8944 (("$" 19786) (T_STRING 20181) (T_VARIABLE 19834) (property_name 20182) (simple_variable 20183) ("{" 20184)) 8945 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 20179) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8946 (("]" 20180)) 8947 (("(" 20050) (argument_list 20187)) 8948 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 20185) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 8949 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 20186)) 8950 (("(" 20050) (argument_list 20189)) 8951 (("`" 20191)) 8952 (("]" 20193)) 8953 (("=" 20194)) 8954 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20195) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8955 (("%" 19864) ("*" 19865) ("+" 19866) ("-" 19867) ("." 19868) ("/" 19869) ("<" 19870) (">" 19871) ("?" 19872) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19873) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19874) (T_BOOLEAN_AND 19875) (T_BOOLEAN_OR 19876) (T_COALESCE 19877) (T_DOUBLE_ARROW 20198) (T_INSTANCEOF 19878) (T_IS_EQUAL 19879) (T_IS_GREATER_OR_EQUAL 19880) (T_IS_IDENTICAL 19881) (T_IS_NOT_EQUAL 19882) (T_IS_NOT_IDENTICAL 19883) (T_IS_SMALLER_OR_EQUAL 19884) (T_LOGICAL_AND 19885) (T_LOGICAL_OR 19886) (T_LOGICAL_XOR 19887) (T_POW 19888) (T_SL 19889) (T_SPACESHIP 19890) (T_SR 19891) ("^" 19892) ("|" 19893)) 8956 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20199) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8957 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 20208) (T_VARIABLE 2896) (encaps_var 2910)) 8958 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 20207) (T_VARIABLE 2896) (encaps_var 2901)) 8959 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20249) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8960 ((T_CLASS 20241)) 8961 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 20240) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 20221) (attribute 2958) (class_modifier 2939)) 8962 ((T_PAAMAYIM_NEKUDOTAYIM 20238)) 8963 (("(" 20105) (argument_list 20237)) 8964 ((T_NULLSAFE_OBJECT_OPERATOR 20228) (T_OBJECT_OPERATOR 20229) (T_PAAMAYIM_NEKUDOTAYIM 20230) ("[" 20231)) 8965 (("$" 20089) (T_STRING 20124) (T_VARIABLE 20091) (property_name 20236) (simple_variable 20126) ("{" 20127)) 8966 (("$" 20089) (T_STRING 20124) (T_VARIABLE 20091) (property_name 20235) (simple_variable 20126) ("{" 20127)) 8967 (("$" 20089) (T_VARIABLE 20091) (simple_variable 20234)) 8968 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 20232) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 8969 (("]" 20233)) 8970 (("$" 20089) (T_VARIABLE 20091) (simple_variable 20239)) 8971 (("(" 2960) (argument_list 2961) (ctor_arguments 20242)) 8972 ((T_EXTENDS 2963) (extends_from 20243)) 8973 ((T_IMPLEMENTS 2965) (implements_list 20244)) 8974 ((backup_doc_comment 20245)) 8975 (("{" 20246)) 8976 ((class_statement_list 20247)) 8977 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 20248)) 8978 (("%" 299) (")" 20250) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8979 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20252) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8980 (("%" 299) (")" 20253) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8981 (("{" 20254)) 8982 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 20255) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 8983 (("}" 20256)) 8984 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 20258) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 8985 ((")" 20259)) 8986 (("=" 20260)) 8987 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 20261) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 8988 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 20263) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8989 (("," 5192) (possible_comma 20264)) 8990 ((")" 20265)) 8991 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 20273) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 20274) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 20275) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8992 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 20278) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 8993 (("," 1480) (possible_comma 20276)) 8994 ((")" 20277)) 8995 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20280) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8996 (("%" 299) (")" 20281) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8997 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20283) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 8998 (("%" 299) (")" 20284) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 8999 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 20291) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 9000 ((")" 20292)) 9001 (("%" 299) (")" 20297) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9002 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 20300) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 9003 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 20301)) 9004 (("\"" 20303) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 9005 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 20313) (T_VARIABLE 2896) (encaps_var 2910)) 9006 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 20312) (T_VARIABLE 2896) (encaps_var 2901)) 9007 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20354) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9008 ((T_CLASS 20346)) 9009 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 20345) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 20326) (attribute 2958) (class_modifier 2939)) 9010 ((T_PAAMAYIM_NEKUDOTAYIM 20343)) 9011 (("(" 19693) (argument_list 20342)) 9012 ((T_NULLSAFE_OBJECT_OPERATOR 20333) (T_OBJECT_OPERATOR 20334) (T_PAAMAYIM_NEKUDOTAYIM 20335) ("[" 20336)) 9013 (("$" 19677) (T_STRING 19712) (T_VARIABLE 19679) (property_name 20341) (simple_variable 19714) ("{" 19715)) 9014 (("$" 19677) (T_STRING 19712) (T_VARIABLE 19679) (property_name 20340) (simple_variable 19714) ("{" 19715)) 9015 (("$" 19677) (T_VARIABLE 19679) (simple_variable 20339)) 9016 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 20337) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 9017 (("]" 20338)) 9018 (("$" 19677) (T_VARIABLE 19679) (simple_variable 20344)) 9019 (("(" 2960) (argument_list 2961) (ctor_arguments 20347)) 9020 ((T_EXTENDS 2963) (extends_from 20348)) 9021 ((T_IMPLEMENTS 2965) (implements_list 20349)) 9022 ((backup_doc_comment 20350)) 9023 (("{" 20351)) 9024 ((class_statement_list 20352)) 9025 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 20353)) 9026 (("%" 299) (")" 20355) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9027 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20357) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9028 (("%" 299) (")" 20358) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9029 (("{" 20359)) 9030 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 20360) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 9031 (("}" 20361)) 9032 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 20363) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 9033 ((")" 20364)) 9034 (("=" 20365)) 9035 (("!" 19372) ("\"" 19373) ("$" 19374) ("(" 19375) ("+" 19376) ("-" 19377) ("@" 19378) (T_ARRAY 19379) (T_ARRAY_CAST 19380) (T_ATTRIBUTE 128) (T_BOOL_CAST 19381) (T_CLASS_C 19382) (T_CLONE 19383) (T_CONSTANT_ENCAPSED_STRING 19384) (T_DEC 19385) (T_DIR 19386) (T_DNUMBER 19387) (T_DOUBLE_CAST 19388) (T_EMPTY 19389) (T_EVAL 19390) (T_EXIT 19391) (T_FILE 19392) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19393) (T_INC 19394) (T_INCLUDE 19395) (T_INCLUDE_ONCE 19396) (T_INT_CAST 19397) (T_ISSET 19398) (T_LINE 19399) (T_LIST 19400) (T_LNUMBER 19401) (T_MATCH 19402) (T_METHOD_C 19403) (T_NAME_FULLY_QUALIFIED 19404) (T_NAME_QUALIFIED 19405) (T_NAME_RELATIVE 19406) (T_NEW 19407) (T_NS_C 19408) (T_OBJECT_CAST 19409) (T_PRINT 19410) (T_PROPERTY_C 19411) (T_READONLY 19412) (T_REQUIRE 19413) (T_REQUIRE_ONCE 19414) (T_START_HEREDOC 19415) (T_STATIC 19416) (T_STRING 19417) (T_STRING_CAST 19418) (T_THROW 19419) (T_TRAIT_C 19420) (T_UNSET_CAST 19421) (T_VARIABLE 19422) (T_YIELD 19423) (T_YIELD_FROM 19424) ("[" 19425) ("`" 19426) (array_object_dereferenceable 19427) (attribute 133) (attributes 19428) (callable_expr 19429) (callable_variable 19430) (class_constant 19431) (class_name 19432) (constant 19433) (dereferenceable_scalar 19434) (expr 20366) (fn 19436) (fully_dereferenceable 105) #'19437 (function_call 19438) (inline_function 19439) (internal_functions_in_yacc 19440) (match 19441) (name 19442) (new_dereferenceable 19443) (new_non_dereferenceable 19444) (scalar 19445) (simple_variable 19446) (static_member 19447) (variable 19448) (variable_class_name 19449) ("~" 19450)) 9036 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 20368) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 9037 (("," 5192) (possible_comma 20369)) 9038 ((")" 20370)) 9039 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 20378) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 20379) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 20380) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 9040 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 20383) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 9041 (("," 1480) (possible_comma 20381)) 9042 ((")" 20382)) 9043 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20385) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9044 (("%" 299) (")" 20386) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9045 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20388) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9046 (("%" 299) (")" 20389) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9047 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 20396) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 9048 ((")" 20397)) 9049 (("%" 299) (")" 20402) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9050 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 20405) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 9051 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 20406)) 9052 (("\"" 20408) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 9053 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 20421) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 9054 (("$" 20411) (T_VARIABLE 20414) (simple_variable 21989) ("{" 21990)) 9055 (("(" 21986)) 9056 (("(" 20431) (argument_list 21985)) 9057 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 21983) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 9058 (("\"" 1520) ("$" 20411) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 1542) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 21982) (variable_class_name 20427)) 9059 ((T_NULLSAFE_OBJECT_OPERATOR 21968) (T_OBJECT_OPERATOR 21969) ("[" 21970)) 9060 (("(" 20431) (argument_list 21967)) 9061 ((T_PAAMAYIM_NEKUDOTAYIM 21963)) 9062 ((")" 20440) (T_DOUBLE_ARROW 20441)) 9063 (("(" 20431) (argument_list 20439)) 9064 ((T_PAAMAYIM_NEKUDOTAYIM 20428)) 9065 (("$" 20411) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 20414) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 20429) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 20430) ("{" 1642)) 9066 (("(" 20431) (argument_list 20432)) 9067 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 20433) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 20434) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 20435) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 9068 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 20438) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 9069 (("," 1480) (possible_comma 20436)) 9070 ((")" 20437)) 9071 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20472) (";" 17846) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17847) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17848) (T_DEC 22) (T_DECLARE 17849) (T_DIR 24) (T_DNUMBER 25) (T_DO 17850) (T_DOUBLE_CAST 27) (T_ECHO 17851) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17852) (T_FOREACH 17853) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17854) (T_GOTO 17855) (T_IF 17811) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17856) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17857) (T_START_HEREDOC 69) (T_STATIC 17858) (T_STRING 17859) (T_STRING_CAST 72) (T_SWITCH 17860) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17861) (T_UNSET 17862) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17863) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17864) (alt_if_stmt_without_else 17865) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17866) (fn 104) (foreach_statement 21962) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17867) (if_stmt_without_else 17868) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20474) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17870) ("~" 127)) 9072 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 20452) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 9073 (("$" 20442) (T_VARIABLE 20445) (simple_variable 21958) ("{" 21959)) 9074 (("(" 21955)) 9075 (("(" 20462) (argument_list 21954)) 9076 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 21952) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 9077 (("\"" 1520) ("$" 20442) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 1542) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 21951) (variable_class_name 20458)) 9078 ((T_NULLSAFE_OBJECT_OPERATOR 21937) (T_OBJECT_OPERATOR 21938) ("[" 21939)) 9079 (("(" 20462) (argument_list 21936)) 9080 ((T_PAAMAYIM_NEKUDOTAYIM 21932)) 9081 ((")" 20471)) 9082 (("(" 20462) (argument_list 20470)) 9083 ((T_PAAMAYIM_NEKUDOTAYIM 20459)) 9084 (("$" 20442) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 20445) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 20460) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 20461) ("{" 1642)) 9085 (("(" 20462) (argument_list 20463)) 9086 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 20464) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 20465) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 20466) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 9087 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 20469) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 9088 (("," 1480) (possible_comma 20467)) 9089 ((")" 20468)) 9090 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20472) (";" 17846) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17847) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17848) (T_DEC 22) (T_DECLARE 17849) (T_DIR 24) (T_DNUMBER 25) (T_DO 17850) (T_DOUBLE_CAST 27) (T_ECHO 17851) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17852) (T_FOREACH 17853) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17854) (T_GOTO 17855) (T_IF 17811) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17856) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17857) (T_START_HEREDOC 69) (T_STATIC 17858) (T_STRING 17859) (T_STRING_CAST 72) (T_SWITCH 17860) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17861) (T_UNSET 17862) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17863) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17864) (alt_if_stmt_without_else 17865) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17866) (fn 104) (foreach_statement 20473) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17867) (if_stmt_without_else 17868) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20474) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17870) ("~" 127)) 9091 ((inner_statement_list 20475)) 9092 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 20483) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 9093 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21930) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9094 ((T_STRING 21923)) 9095 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21921) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9096 (("(" 21912)) 9097 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21906) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 9098 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 21904) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9099 ((";" 21903)) 9100 ((T_STRING 21896)) 9101 (("(" 21883)) 9102 (("(" 21868)) 9103 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 21866) (simple_variable 18987)) 9104 ((T_STRING 21864)) 9105 (("(" 21861)) 9106 (("(" 21857)) 9107 ((T_STRING 21851)) 9108 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21849) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9109 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 21847)) 9110 ((":" 21846)) 9111 (("(" 21828)) 9112 ((T_STRING 21823)) 9113 (("{" 21806)) 9114 (("(" 21801)) 9115 (("(" 21792)) 9116 ((T_ELSE 21785) (T_ELSEIF 17291) (T_ENDIF 21786)) 9117 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 20478) (T_ENUM 20484) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 20492) (T_READONLY 2931) (T_STATIC 17376) (T_TRAIT 20497) (attribute 17377) (attributed_statement 21784) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (enum_declaration_statement 20507) (fn 104) #'20509 (function_declaration_statement 20510) (inline_function 17379) (interface_declaration_statement 20514) (trait_declaration_statement 20516)) 9118 ((T_ABSTRACT 10) (T_CLASS 21776) (T_FINAL 34) (T_READONLY 2931) (class_modifier 17368)) 9119 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 21775) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 9120 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17349) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17350) (ampersand 17351) (returns_ref 21763)) 9121 ((T_ELSE 20520) (T_ELSEIF 20521)) 9122 ((inner_statement_list 20518)) 9123 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20519) ("~" 127)) 9124 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20508) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21762) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 9125 (("(" 20522)) 9126 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20523) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9127 (("%" 299) (")" 20524) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9128 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20525) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20526) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20527) (T_DEC 22) (T_DECLARE 20528) (T_DIR 24) (T_DNUMBER 25) (T_DO 20529) (T_DOUBLE_CAST 27) (T_ECHO 20530) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20531) (T_FOREACH 20532) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20533) (T_GOTO 20534) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20535) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20536) (T_START_HEREDOC 69) (T_STATIC 20537) (T_STRING 20538) (T_STRING_CAST 72) (T_SWITCH 20539) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20540) (T_UNSET 20541) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20542) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20543) (alt_if_stmt_without_else 20544) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20545) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20546) (if_stmt_without_else 20547) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20548) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20549) ("~" 127)) 9129 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21760) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9130 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21758) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9131 (("(" 21749)) 9132 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21743) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 9133 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 21741) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9134 (("(" 20635)) 9135 (("(" 20620)) 9136 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 20618) (simple_variable 18987)) 9137 ((T_STRING 20616)) 9138 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 20614) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9139 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 20612)) 9140 ((":" 20611)) 9141 (("(" 20593)) 9142 (("{" 20576)) 9143 (("(" 20571)) 9144 (("(" 20562)) 9145 ((T_ELSE 20555) (T_ELSEIF 17291) (T_ENDIF 20556)) 9146 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 20554) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 9147 ((T_ELSE 20552) (T_ELSEIF 20521)) 9148 ((inner_statement_list 20550)) 9149 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20551) ("~" 127)) 9150 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20525) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20526) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20527) (T_DEC 22) (T_DECLARE 20528) (T_DIR 24) (T_DNUMBER 25) (T_DO 20529) (T_DOUBLE_CAST 27) (T_ECHO 20530) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20531) (T_FOREACH 20532) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20533) (T_GOTO 20534) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20535) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20536) (T_START_HEREDOC 69) (T_STATIC 20537) (T_STRING 20538) (T_STRING_CAST 72) (T_SWITCH 20539) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20540) (T_UNSET 20541) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20542) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20543) (alt_if_stmt_without_else 20544) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20545) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20546) (if_stmt_without_else 20547) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20553) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20549) ("~" 127)) 9151 ((":" 20558)) 9152 ((";" 20557)) 9153 ((inner_statement_list 20559)) 9154 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 20560) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 9155 ((";" 20561)) 9156 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20563) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9157 (("%" 299) (")" 20564) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9158 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20565) (";" 20525) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20526) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20527) (T_DEC 22) (T_DECLARE 20528) (T_DIR 24) (T_DNUMBER 25) (T_DO 20529) (T_DOUBLE_CAST 27) (T_ECHO 20530) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20531) (T_FOREACH 20532) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20533) (T_GOTO 20534) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20535) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20536) (T_START_HEREDOC 69) (T_STATIC 20537) (T_STRING 20538) (T_STRING_CAST 72) (T_SWITCH 20539) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20540) (T_UNSET 20541) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20542) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20543) (alt_if_stmt_without_else 20544) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20545) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20546) (if_stmt_without_else 20547) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20566) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 20567) ("{" 20549) ("~" 127)) 9159 ((inner_statement_list 20568)) 9160 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 20569) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 9161 ((";" 20570)) 9162 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 20572) (variable 17576) (variable_class_name 17577)) 9163 (("," 17589) (possible_comma 20573)) 9164 ((")" 20574)) 9165 ((";" 20575)) 9166 ((inner_statement_list 20577)) 9167 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20578) ("~" 127)) 9168 ((catch_list 20579)) 9169 ((T_CATCH 20580) (T_FINALLY 20581) (finally_statement 20582)) 9170 (("(" 20586)) 9171 (("{" 20583)) 9172 ((inner_statement_list 20584)) 9173 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20585) ("~" 127)) 9174 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 20587) (class_name 17636) (name 17637)) 9175 ((T_VARIABLE 17638) (optional_variable 20588) ("|" 17640)) 9176 ((")" 20589)) 9177 (("{" 20590)) 9178 ((inner_statement_list 20591)) 9179 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20592) ("~" 127)) 9180 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20594) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9181 (("%" 299) (")" 20595) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9182 ((":" 20596) (switch_case_list 20597) ("{" 20598)) 9183 ((";" 20604) (case_list 20605)) 9184 ((";" 20599) (case_list 20600)) 9185 ((case_list 20602)) 9186 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 20601)) 9187 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 20603)) 9188 ((case_list 20608)) 9189 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 20606)) 9190 ((";" 20607)) 9191 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 20609)) 9192 ((";" 20610)) 9193 (("," 17937) (";" 20613)) 9194 ((";" 20615)) 9195 ((";" 20617)) 9196 (("," 18988) (";" 20619)) 9197 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 20621) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 9198 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 20622) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 9199 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 20623) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 9200 ((")" 20624) (T_DOUBLE_ARROW 20625)) 9201 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20628) (";" 20525) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20526) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20527) (T_DEC 22) (T_DECLARE 20528) (T_DIR 24) (T_DNUMBER 25) (T_DO 20529) (T_DOUBLE_CAST 27) (T_ECHO 20530) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20531) (T_FOREACH 20532) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20533) (T_GOTO 20534) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20535) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20536) (T_START_HEREDOC 69) (T_STATIC 20537) (T_STRING 20538) (T_STRING_CAST 72) (T_SWITCH 20539) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20540) (T_UNSET 20541) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20542) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20543) (alt_if_stmt_without_else 20544) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20545) (fn 104) (foreach_statement 20634) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20546) (if_stmt_without_else 20547) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20630) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20549) ("~" 127)) 9202 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 20626) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 9203 ((")" 20627)) 9204 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20628) (";" 20525) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20526) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20527) (T_DEC 22) (T_DECLARE 20528) (T_DIR 24) (T_DNUMBER 25) (T_DO 20529) (T_DOUBLE_CAST 27) (T_ECHO 20530) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20531) (T_FOREACH 20532) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20533) (T_GOTO 20534) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20535) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20536) (T_START_HEREDOC 69) (T_STATIC 20537) (T_STRING 20538) (T_STRING_CAST 72) (T_SWITCH 20539) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20540) (T_UNSET 20541) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20542) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20543) (alt_if_stmt_without_else 20544) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20545) (fn 104) (foreach_statement 20629) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20546) (if_stmt_without_else 20547) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20630) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20549) ("~" 127)) 9205 ((inner_statement_list 20631)) 9206 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 20632) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 9207 ((";" 20633)) 9208 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 20637) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9209 ((";" 20641)) 9210 (("," 20639)) 9211 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20640) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9212 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 20642) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9213 ((";" 20643)) 9214 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 20645) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 9215 ((")" 20649)) 9216 (("," 20647)) 9217 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20648) (fn 2403) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 9218 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20650) (";" 20525) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20526) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20527) (T_DEC 22) (T_DECLARE 20528) (T_DIR 24) (T_DNUMBER 25) (T_DO 20529) (T_DOUBLE_CAST 27) (T_ECHO 20530) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20531) (T_FOREACH 20532) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20533) (T_GOTO 20534) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20535) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20536) (T_START_HEREDOC 69) (T_STATIC 20537) (T_STRING 20538) (T_STRING_CAST 72) (T_SWITCH 20539) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20540) (T_UNSET 20541) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20542) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20543) (alt_if_stmt_without_else 20544) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20545) (fn 104) (for_statement 20651) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20546) (if_stmt_without_else 20547) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20652) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20549) ("~" 127)) 9219 ((inner_statement_list 20653)) 9220 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 20661) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 9221 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21739) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9222 ((T_STRING 21732)) 9223 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21730) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9224 (("(" 21721)) 9225 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21715) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 9226 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 21713) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9227 ((";" 21712)) 9228 ((T_STRING 21705)) 9229 (("(" 21692)) 9230 (("(" 21677)) 9231 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 21675) (simple_variable 18987)) 9232 ((T_STRING 21673)) 9233 (("(" 21670)) 9234 (("(" 21666)) 9235 ((T_STRING 21660)) 9236 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21658) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9237 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 21656)) 9238 ((":" 21655)) 9239 (("(" 21637)) 9240 ((T_STRING 21632)) 9241 (("{" 21615)) 9242 (("(" 21610)) 9243 (("(" 21601)) 9244 ((T_ELSE 21594) (T_ELSEIF 17291) (T_ENDIF 21595)) 9245 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 20656) (T_ENUM 20662) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 20670) (T_READONLY 2931) (T_STATIC 17376) (T_TRAIT 20675) (attribute 17377) (attributed_statement 21593) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (enum_declaration_statement 20685) (fn 104) #'20687 (function_declaration_statement 20688) (inline_function 17379) (interface_declaration_statement 20692) (trait_declaration_statement 20694)) 9246 ((T_ABSTRACT 10) (T_CLASS 21585) (T_FINAL 34) (T_READONLY 2931) (class_modifier 17368)) 9247 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 21584) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 9248 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17349) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17350) (ampersand 17351) (returns_ref 21572)) 9249 ((T_ELSE 20698) (T_ELSEIF 20699)) 9250 ((inner_statement_list 20696)) 9251 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20697) ("~" 127)) 9252 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20686) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21571) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 9253 (("(" 20700)) 9254 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20701) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9255 (("%" 299) (")" 20702) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9256 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20703) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20704) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20705) (T_DEC 22) (T_DECLARE 20706) (T_DIR 24) (T_DNUMBER 25) (T_DO 20707) (T_DOUBLE_CAST 27) (T_ECHO 20708) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20709) (T_FOREACH 20710) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20711) (T_GOTO 20712) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20713) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20714) (T_START_HEREDOC 69) (T_STATIC 20715) (T_STRING 20716) (T_STRING_CAST 72) (T_SWITCH 20717) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20718) (T_UNSET 20719) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20720) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20721) (alt_if_stmt_without_else 20722) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20723) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20724) (if_stmt_without_else 20725) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20726) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20727) ("~" 127)) 9257 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21569) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9258 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21567) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9259 (("(" 21558)) 9260 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20856) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 9261 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 20827) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9262 (("(" 20813)) 9263 (("(" 20798)) 9264 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 20796) (simple_variable 18987)) 9265 ((T_STRING 20794)) 9266 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 20792) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9267 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 20790)) 9268 ((":" 20789)) 9269 (("(" 20771)) 9270 (("{" 20754)) 9271 (("(" 20749)) 9272 (("(" 20740)) 9273 ((T_ELSE 20733) (T_ELSEIF 17291) (T_ENDIF 20734)) 9274 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 20732) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 9275 ((T_ELSE 20730) (T_ELSEIF 20699)) 9276 ((inner_statement_list 20728)) 9277 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20729) ("~" 127)) 9278 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20703) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20704) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20705) (T_DEC 22) (T_DECLARE 20706) (T_DIR 24) (T_DNUMBER 25) (T_DO 20707) (T_DOUBLE_CAST 27) (T_ECHO 20708) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20709) (T_FOREACH 20710) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20711) (T_GOTO 20712) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20713) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20714) (T_START_HEREDOC 69) (T_STATIC 20715) (T_STRING 20716) (T_STRING_CAST 72) (T_SWITCH 20717) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20718) (T_UNSET 20719) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20720) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20721) (alt_if_stmt_without_else 20722) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20723) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20724) (if_stmt_without_else 20725) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20731) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20727) ("~" 127)) 9279 ((":" 20736)) 9280 ((";" 20735)) 9281 ((inner_statement_list 20737)) 9282 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 20738) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 9283 ((";" 20739)) 9284 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20741) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9285 (("%" 299) (")" 20742) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9286 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20743) (";" 20703) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20704) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20705) (T_DEC 22) (T_DECLARE 20706) (T_DIR 24) (T_DNUMBER 25) (T_DO 20707) (T_DOUBLE_CAST 27) (T_ECHO 20708) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20709) (T_FOREACH 20710) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20711) (T_GOTO 20712) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20713) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20714) (T_START_HEREDOC 69) (T_STATIC 20715) (T_STRING 20716) (T_STRING_CAST 72) (T_SWITCH 20717) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20718) (T_UNSET 20719) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20720) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20721) (alt_if_stmt_without_else 20722) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20723) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20724) (if_stmt_without_else 20725) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20744) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 20745) ("{" 20727) ("~" 127)) 9287 ((inner_statement_list 20746)) 9288 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 20747) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 9289 ((";" 20748)) 9290 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 20750) (variable 17576) (variable_class_name 17577)) 9291 (("," 17589) (possible_comma 20751)) 9292 ((")" 20752)) 9293 ((";" 20753)) 9294 ((inner_statement_list 20755)) 9295 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20756) ("~" 127)) 9296 ((catch_list 20757)) 9297 ((T_CATCH 20758) (T_FINALLY 20759) (finally_statement 20760)) 9298 (("(" 20764)) 9299 (("{" 20761)) 9300 ((inner_statement_list 20762)) 9301 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20763) ("~" 127)) 9302 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 20765) (class_name 17636) (name 17637)) 9303 ((T_VARIABLE 17638) (optional_variable 20766) ("|" 17640)) 9304 ((")" 20767)) 9305 (("{" 20768)) 9306 ((inner_statement_list 20769)) 9307 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20770) ("~" 127)) 9308 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20772) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9309 (("%" 299) (")" 20773) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9310 ((":" 20774) (switch_case_list 20775) ("{" 20776)) 9311 ((";" 20782) (case_list 20783)) 9312 ((";" 20777) (case_list 20778)) 9313 ((case_list 20780)) 9314 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 20779)) 9315 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 20781)) 9316 ((case_list 20786)) 9317 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 20784)) 9318 ((";" 20785)) 9319 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 20787)) 9320 ((";" 20788)) 9321 (("," 17937) (";" 20791)) 9322 ((";" 20793)) 9323 ((";" 20795)) 9324 (("," 18988) (";" 20797)) 9325 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 20799) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 9326 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 20800) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 9327 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 20801) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 9328 ((")" 20802) (T_DOUBLE_ARROW 20803)) 9329 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20806) (";" 20703) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20704) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20705) (T_DEC 22) (T_DECLARE 20706) (T_DIR 24) (T_DNUMBER 25) (T_DO 20707) (T_DOUBLE_CAST 27) (T_ECHO 20708) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20709) (T_FOREACH 20710) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20711) (T_GOTO 20712) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20713) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20714) (T_START_HEREDOC 69) (T_STATIC 20715) (T_STRING 20716) (T_STRING_CAST 72) (T_SWITCH 20717) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20718) (T_UNSET 20719) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20720) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20721) (alt_if_stmt_without_else 20722) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20723) (fn 104) (foreach_statement 20812) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20724) (if_stmt_without_else 20725) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20808) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20727) ("~" 127)) 9330 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 20804) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 9331 ((")" 20805)) 9332 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20806) (";" 20703) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20704) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20705) (T_DEC 22) (T_DECLARE 20706) (T_DIR 24) (T_DNUMBER 25) (T_DO 20707) (T_DOUBLE_CAST 27) (T_ECHO 20708) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20709) (T_FOREACH 20710) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20711) (T_GOTO 20712) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20713) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20714) (T_START_HEREDOC 69) (T_STATIC 20715) (T_STRING 20716) (T_STRING_CAST 72) (T_SWITCH 20717) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20718) (T_UNSET 20719) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20720) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20721) (alt_if_stmt_without_else 20722) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20723) (fn 104) (foreach_statement 20807) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20724) (if_stmt_without_else 20725) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20808) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20727) ("~" 127)) 9333 ((inner_statement_list 20809)) 9334 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 20810) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 9335 ((";" 20811)) 9336 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 20814) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9337 ((";" 20815)) 9338 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 20816) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9339 ((";" 20817)) 9340 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 20818) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 9341 ((")" 20819)) 9342 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20820) (";" 20703) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20704) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20705) (T_DEC 22) (T_DECLARE 20706) (T_DIR 24) (T_DNUMBER 25) (T_DO 20707) (T_DOUBLE_CAST 27) (T_ECHO 20708) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20709) (T_FOREACH 20710) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20711) (T_GOTO 20712) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20713) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20714) (T_START_HEREDOC 69) (T_STATIC 20715) (T_STRING 20716) (T_STRING_CAST 72) (T_SWITCH 20717) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20718) (T_UNSET 20719) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20720) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20721) (alt_if_stmt_without_else 20722) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20723) (fn 104) (for_statement 20821) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20724) (if_stmt_without_else 20725) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20822) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20727) ("~" 127)) 9343 ((inner_statement_list 20823)) 9344 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 20824) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 9345 ((";" 20825)) 9346 (("," 20829) (";" 20830)) 9347 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20831) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9348 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21556) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9349 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21554) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9350 (("(" 21545)) 9351 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21539) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 9352 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 21537) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9353 (("(" 21524)) 9354 (("(" 21509)) 9355 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 21507) (simple_variable 18987)) 9356 ((T_STRING 21505)) 9357 (("(" 21501)) 9358 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21499) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9359 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 21497)) 9360 ((":" 21496)) 9361 (("(" 21478)) 9362 (("{" 21461)) 9363 (("(" 21456)) 9364 (("(" 21447)) 9365 ((T_ELSE 21440) (T_ELSEIF 17291) (T_ENDIF 21441)) 9366 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 21439) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 9367 ((T_ELSE 20865) (T_ELSEIF 20866)) 9368 ((T_WHILE 20860)) 9369 ((inner_statement_list 20858)) 9370 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20859) ("~" 127)) 9371 (("(" 20861)) 9372 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20862) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9373 (("%" 299) (")" 20863) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9374 ((";" 20864)) 9375 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21438) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 9376 (("(" 20867)) 9377 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20868) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9378 (("%" 299) (")" 20869) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9379 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20870) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20871) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20872) (T_DEC 22) (T_DECLARE 20873) (T_DIR 24) (T_DNUMBER 25) (T_DO 20874) (T_DOUBLE_CAST 27) (T_ECHO 20875) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20876) (T_FOREACH 20877) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20878) (T_GOTO 20879) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20880) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20881) (T_START_HEREDOC 69) (T_STATIC 20882) (T_STRING 20883) (T_STRING_CAST 72) (T_SWITCH 20884) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20885) (T_UNSET 20886) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20887) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20888) (alt_if_stmt_without_else 20889) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20890) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20891) (if_stmt_without_else 20892) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20893) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20894) ("~" 127)) 9380 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21436) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9381 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21434) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9382 (("(" 21001)) 9383 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20995) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 9384 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 20993) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9385 (("(" 20980)) 9386 (("(" 20965)) 9387 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 20963) (simple_variable 18987)) 9388 ((T_STRING 20961)) 9389 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 20959) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9390 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 20957)) 9391 ((":" 20956)) 9392 (("(" 20938)) 9393 (("{" 20921)) 9394 (("(" 20916)) 9395 (("(" 20907)) 9396 ((T_ELSE 20900) (T_ELSEIF 17291) (T_ENDIF 20901)) 9397 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 20899) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 9398 ((T_ELSE 20897) (T_ELSEIF 20866)) 9399 ((inner_statement_list 20895)) 9400 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20896) ("~" 127)) 9401 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20870) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20871) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20872) (T_DEC 22) (T_DECLARE 20873) (T_DIR 24) (T_DNUMBER 25) (T_DO 20874) (T_DOUBLE_CAST 27) (T_ECHO 20875) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20876) (T_FOREACH 20877) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20878) (T_GOTO 20879) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20880) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20881) (T_START_HEREDOC 69) (T_STATIC 20882) (T_STRING 20883) (T_STRING_CAST 72) (T_SWITCH 20884) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20885) (T_UNSET 20886) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20887) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20888) (alt_if_stmt_without_else 20889) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20890) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20891) (if_stmt_without_else 20892) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20898) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20894) ("~" 127)) 9402 ((":" 20903)) 9403 ((";" 20902)) 9404 ((inner_statement_list 20904)) 9405 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 20905) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 9406 ((";" 20906)) 9407 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20908) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9408 (("%" 299) (")" 20909) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9409 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20910) (";" 20870) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20871) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20872) (T_DEC 22) (T_DECLARE 20873) (T_DIR 24) (T_DNUMBER 25) (T_DO 20874) (T_DOUBLE_CAST 27) (T_ECHO 20875) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20876) (T_FOREACH 20877) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20878) (T_GOTO 20879) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20880) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20881) (T_START_HEREDOC 69) (T_STATIC 20882) (T_STRING 20883) (T_STRING_CAST 72) (T_SWITCH 20884) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20885) (T_UNSET 20886) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20887) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20888) (alt_if_stmt_without_else 20889) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20890) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20891) (if_stmt_without_else 20892) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20911) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 20912) ("{" 20894) ("~" 127)) 9410 ((inner_statement_list 20913)) 9411 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 20914) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 9412 ((";" 20915)) 9413 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 20917) (variable 17576) (variable_class_name 17577)) 9414 (("," 17589) (possible_comma 20918)) 9415 ((")" 20919)) 9416 ((";" 20920)) 9417 ((inner_statement_list 20922)) 9418 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20923) ("~" 127)) 9419 ((catch_list 20924)) 9420 ((T_CATCH 20925) (T_FINALLY 20926) (finally_statement 20927)) 9421 (("(" 20931)) 9422 (("{" 20928)) 9423 ((inner_statement_list 20929)) 9424 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20930) ("~" 127)) 9425 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 20932) (class_name 17636) (name 17637)) 9426 ((T_VARIABLE 17638) (optional_variable 20933) ("|" 17640)) 9427 ((")" 20934)) 9428 (("{" 20935)) 9429 ((inner_statement_list 20936)) 9430 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 20937) ("~" 127)) 9431 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20939) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9432 (("%" 299) (")" 20940) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9433 ((":" 20941) (switch_case_list 20942) ("{" 20943)) 9434 ((";" 20949) (case_list 20950)) 9435 ((";" 20944) (case_list 20945)) 9436 ((case_list 20947)) 9437 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 20946)) 9438 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 20948)) 9439 ((case_list 20953)) 9440 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 20951)) 9441 ((";" 20952)) 9442 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 20954)) 9443 ((";" 20955)) 9444 (("," 17937) (";" 20958)) 9445 ((";" 20960)) 9446 ((";" 20962)) 9447 (("," 18988) (";" 20964)) 9448 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 20966) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 9449 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 20967) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 9450 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 20968) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 9451 ((")" 20969) (T_DOUBLE_ARROW 20970)) 9452 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20973) (";" 20870) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20871) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20872) (T_DEC 22) (T_DECLARE 20873) (T_DIR 24) (T_DNUMBER 25) (T_DO 20874) (T_DOUBLE_CAST 27) (T_ECHO 20875) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20876) (T_FOREACH 20877) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20878) (T_GOTO 20879) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20880) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20881) (T_START_HEREDOC 69) (T_STATIC 20882) (T_STRING 20883) (T_STRING_CAST 72) (T_SWITCH 20884) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20885) (T_UNSET 20886) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20887) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20888) (alt_if_stmt_without_else 20889) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20890) (fn 104) (foreach_statement 20979) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20891) (if_stmt_without_else 20892) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20975) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20894) ("~" 127)) 9453 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 20971) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 9454 ((")" 20972)) 9455 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20973) (";" 20870) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20871) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20872) (T_DEC 22) (T_DECLARE 20873) (T_DIR 24) (T_DNUMBER 25) (T_DO 20874) (T_DOUBLE_CAST 27) (T_ECHO 20875) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20876) (T_FOREACH 20877) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20878) (T_GOTO 20879) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20880) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20881) (T_START_HEREDOC 69) (T_STATIC 20882) (T_STRING 20883) (T_STRING_CAST 72) (T_SWITCH 20884) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20885) (T_UNSET 20886) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20887) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20888) (alt_if_stmt_without_else 20889) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20890) (fn 104) (foreach_statement 20974) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20891) (if_stmt_without_else 20892) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20975) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20894) ("~" 127)) 9456 ((inner_statement_list 20976)) 9457 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 20977) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 9458 ((";" 20978)) 9459 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 20981) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9460 ((";" 20982)) 9461 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 20983) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9462 ((";" 20984)) 9463 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 20985) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 9464 ((")" 20986)) 9465 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20987) (";" 20870) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20871) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20872) (T_DEC 22) (T_DECLARE 20873) (T_DIR 24) (T_DNUMBER 25) (T_DO 20874) (T_DOUBLE_CAST 27) (T_ECHO 20875) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20876) (T_FOREACH 20877) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20878) (T_GOTO 20879) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20880) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20881) (T_START_HEREDOC 69) (T_STATIC 20882) (T_STRING 20883) (T_STRING_CAST 72) (T_SWITCH 20884) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20885) (T_UNSET 20886) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20887) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20888) (alt_if_stmt_without_else 20889) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20890) (fn 104) (for_statement 20988) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20891) (if_stmt_without_else 20892) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20989) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20894) ("~" 127)) 9466 ((inner_statement_list 20990)) 9467 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 20991) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 9468 ((";" 20992)) 9469 (("," 20829) (";" 20994)) 9470 ((T_WHILE 20996)) 9471 (("(" 20997)) 9472 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 20998) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9473 (("%" 299) (")" 20999) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9474 ((";" 21000)) 9475 ((T_STRING 21002) (const_decl 21003) (const_list 21004)) 9476 (("=" 21431)) 9477 ((")" 21005) ("," 21006)) 9478 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21008) (";" 20870) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20871) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20872) (T_DEC 22) (T_DECLARE 20873) (T_DIR 24) (T_DNUMBER 25) (T_DO 20874) (T_DOUBLE_CAST 27) (T_ECHO 20875) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20876) (T_FOREACH 20877) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20878) (T_GOTO 20879) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20880) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20881) (T_START_HEREDOC 69) (T_STATIC 20882) (T_STRING 20883) (T_STRING_CAST 72) (T_SWITCH 20884) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20885) (T_UNSET 20886) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20887) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20888) (alt_if_stmt_without_else 20889) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 21009) (dereferenceable_scalar 101) (expr 20890) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20891) (if_stmt_without_else 20892) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21010) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20894) ("~" 127)) 9479 ((T_STRING 21002) (const_decl 21007)) 9480 ((inner_statement_list 21011)) 9481 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 21019) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 9482 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21429) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9483 ((T_STRING 21422)) 9484 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21420) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9485 (("(" 21411)) 9486 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21405) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 9487 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 21403) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9488 ((";" 21402)) 9489 ((T_STRING 21343)) 9490 (("(" 21330)) 9491 (("(" 21315)) 9492 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 21313) (simple_variable 18987)) 9493 ((T_STRING 21311)) 9494 (("(" 21308)) 9495 (("(" 21302)) 9496 ((T_STRING 21294)) 9497 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21292) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9498 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 21290)) 9499 ((":" 21289)) 9500 (("(" 21271)) 9501 ((T_STRING 21266)) 9502 (("{" 21249)) 9503 (("(" 21244)) 9504 (("(" 21235)) 9505 ((T_ELSE 21228) (T_ELSEIF 17291) (T_ENDIF 21229)) 9506 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 21014) (T_ENUM 21020) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 21028) (T_READONLY 2931) (T_STATIC 17376) (T_TRAIT 21033) (attribute 17377) (attributed_statement 21227) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (enum_declaration_statement 21043) (fn 104) #'21045 (function_declaration_statement 21046) (inline_function 17379) (interface_declaration_statement 21050) (trait_declaration_statement 21052)) 9507 ((T_ABSTRACT 10) (T_CLASS 21219) (T_FINAL 34) (T_READONLY 2931) (class_modifier 17368)) 9508 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 21218) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 9509 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17349) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17350) (ampersand 17351) (returns_ref 21206)) 9510 ((T_ELSE 21056) (T_ELSEIF 21057)) 9511 ((inner_statement_list 21054)) 9512 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21055) ("~" 127)) 9513 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 21044) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21205) (static_member 121) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 9514 (("(" 21058)) 9515 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21059) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9516 (("%" 299) (")" 21060) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9517 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21061) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 21062) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21063) (T_DEC 22) (T_DECLARE 21064) (T_DIR 24) (T_DNUMBER 25) (T_DO 21065) (T_DOUBLE_CAST 27) (T_ECHO 21066) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21067) (T_FOREACH 21068) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 21069) (T_GOTO 21070) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21071) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21072) (T_START_HEREDOC 69) (T_STATIC 21073) (T_STRING 21074) (T_STRING_CAST 72) (T_SWITCH 21075) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21076) (T_UNSET 21077) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21078) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21079) (alt_if_stmt_without_else 21080) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 21081) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 21082) (if_stmt_without_else 21083) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21084) (static_member 121) (variable 124) (variable_class_name 125) ("{" 21085) ("~" 127)) 9518 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21203) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9519 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21201) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9520 (("(" 21192)) 9521 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21186) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 9522 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 21184) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9523 (("(" 21171)) 9524 (("(" 21156)) 9525 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 21154) (simple_variable 18987)) 9526 ((T_STRING 21152)) 9527 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 21150) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 9528 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 21148)) 9529 ((":" 21147)) 9530 (("(" 21129)) 9531 (("{" 21112)) 9532 (("(" 21107)) 9533 (("(" 21098)) 9534 ((T_ELSE 21091) (T_ELSEIF 17291) (T_ENDIF 21092)) 9535 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 21090) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 9536 ((T_ELSE 21088) (T_ELSEIF 21057)) 9537 ((inner_statement_list 21086)) 9538 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21087) ("~" 127)) 9539 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21061) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 21062) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21063) (T_DEC 22) (T_DECLARE 21064) (T_DIR 24) (T_DNUMBER 25) (T_DO 21065) (T_DOUBLE_CAST 27) (T_ECHO 21066) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21067) (T_FOREACH 21068) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 21069) (T_GOTO 21070) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21071) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21072) (T_START_HEREDOC 69) (T_STATIC 21073) (T_STRING 21074) (T_STRING_CAST 72) (T_SWITCH 21075) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21076) (T_UNSET 21077) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21078) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21079) (alt_if_stmt_without_else 21080) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 21081) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 21082) (if_stmt_without_else 21083) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21089) (static_member 121) (variable 124) (variable_class_name 125) ("{" 21085) ("~" 127)) 9540 ((":" 21094)) 9541 ((";" 21093)) 9542 ((inner_statement_list 21095)) 9543 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 21096) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 9544 ((";" 21097)) 9545 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21099) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9546 (("%" 299) (")" 21100) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9547 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21101) (";" 21061) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 21062) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21063) (T_DEC 22) (T_DECLARE 21064) (T_DIR 24) (T_DNUMBER 25) (T_DO 21065) (T_DOUBLE_CAST 27) (T_ECHO 21066) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21067) (T_FOREACH 21068) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 21069) (T_GOTO 21070) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21071) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21072) (T_START_HEREDOC 69) (T_STATIC 21073) (T_STRING 21074) (T_STRING_CAST 72) (T_SWITCH 21075) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21076) (T_UNSET 21077) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21078) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21079) (alt_if_stmt_without_else 21080) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 21081) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 21082) (if_stmt_without_else 21083) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21102) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 21103) ("{" 21085) ("~" 127)) 9548 ((inner_statement_list 21104)) 9549 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 21105) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 9550 ((";" 21106)) 9551 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 21108) (variable 17576) (variable_class_name 17577)) 9552 (("," 17589) (possible_comma 21109)) 9553 ((")" 21110)) 9554 ((";" 21111)) 9555 ((inner_statement_list 21113)) 9556 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21114) ("~" 127)) 9557 ((catch_list 21115)) 9558 ((T_CATCH 21116) (T_FINALLY 21117) (finally_statement 21118)) 9559 (("(" 21122)) 9560 (("{" 21119)) 9561 ((inner_statement_list 21120)) 9562 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21121) ("~" 127)) 9563 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 21123) (class_name 17636) (name 17637)) 9564 ((T_VARIABLE 17638) (optional_variable 21124) ("|" 17640)) 9565 ((")" 21125)) 9566 (("{" 21126)) 9567 ((inner_statement_list 21127)) 9568 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21128) ("~" 127)) 9569 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21130) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9570 (("%" 299) (")" 21131) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9571 ((":" 21132) (switch_case_list 21133) ("{" 21134)) 9572 ((";" 21140) (case_list 21141)) 9573 ((";" 21135) (case_list 21136)) 9574 ((case_list 21138)) 9575 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 21137)) 9576 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 21139)) 9577 ((case_list 21144)) 9578 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 21142)) 9579 ((";" 21143)) 9580 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 21145)) 9581 ((";" 21146)) 9582 (("," 17937) (";" 21149)) 9583 ((";" 21151)) 9584 ((";" 21153)) 9585 (("," 18988) (";" 21155)) 9586 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 21157) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 9587 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 21158) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 9588 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 21159) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 9589 ((")" 21160) (T_DOUBLE_ARROW 21161)) 9590 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21164) (";" 21061) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 21062) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21063) (T_DEC 22) (T_DECLARE 21064) (T_DIR 24) (T_DNUMBER 25) (T_DO 21065) (T_DOUBLE_CAST 27) (T_ECHO 21066) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21067) (T_FOREACH 21068) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 21069) (T_GOTO 21070) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21071) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21072) (T_START_HEREDOC 69) (T_STATIC 21073) (T_STRING 21074) (T_STRING_CAST 72) (T_SWITCH 21075) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21076) (T_UNSET 21077) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21078) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21079) (alt_if_stmt_without_else 21080) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 21081) (fn 104) (foreach_statement 21170) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 21082) (if_stmt_without_else 21083) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21166) (static_member 121) (variable 124) (variable_class_name 125) ("{" 21085) ("~" 127)) 9591 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 21162) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 9592 ((")" 21163)) 9593 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21164) (";" 21061) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 21062) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21063) (T_DEC 22) (T_DECLARE 21064) (T_DIR 24) (T_DNUMBER 25) (T_DO 21065) (T_DOUBLE_CAST 27) (T_ECHO 21066) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21067) (T_FOREACH 21068) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 21069) (T_GOTO 21070) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21071) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21072) (T_START_HEREDOC 69) (T_STATIC 21073) (T_STRING 21074) (T_STRING_CAST 72) (T_SWITCH 21075) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21076) (T_UNSET 21077) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21078) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21079) (alt_if_stmt_without_else 21080) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 21081) (fn 104) (foreach_statement 21165) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 21082) (if_stmt_without_else 21083) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21166) (static_member 121) (variable 124) (variable_class_name 125) ("{" 21085) ("~" 127)) 9594 ((inner_statement_list 21167)) 9595 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 21168) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 9596 ((";" 21169)) 9597 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 21172) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9598 ((";" 21173)) 9599 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 21174) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9600 ((";" 21175)) 9601 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 21176) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 9602 ((")" 21177)) 9603 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21178) (";" 21061) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 21062) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21063) (T_DEC 22) (T_DECLARE 21064) (T_DIR 24) (T_DNUMBER 25) (T_DO 21065) (T_DOUBLE_CAST 27) (T_ECHO 21066) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21067) (T_FOREACH 21068) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 21069) (T_GOTO 21070) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21071) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21072) (T_START_HEREDOC 69) (T_STATIC 21073) (T_STRING 21074) (T_STRING_CAST 72) (T_SWITCH 21075) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21076) (T_UNSET 21077) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21078) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21079) (alt_if_stmt_without_else 21080) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 21081) (fn 104) (for_statement 21179) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 21082) (if_stmt_without_else 21083) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21180) (static_member 121) (variable 124) (variable_class_name 125) ("{" 21085) ("~" 127)) 9604 ((inner_statement_list 21181)) 9605 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 21182) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 9606 ((";" 21183)) 9607 (("," 20829) (";" 21185)) 9608 ((T_WHILE 21187)) 9609 (("(" 21188)) 9610 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21189) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9611 (("%" 299) (")" 21190) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9612 ((";" 21191)) 9613 ((T_STRING 21002) (const_decl 21003) (const_list 21193)) 9614 ((")" 21194) ("," 21006)) 9615 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21195) (";" 21061) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 21062) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21063) (T_DEC 22) (T_DECLARE 21064) (T_DIR 24) (T_DNUMBER 25) (T_DO 21065) (T_DOUBLE_CAST 27) (T_ECHO 21066) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21067) (T_FOREACH 21068) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 21069) (T_GOTO 21070) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21071) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21072) (T_START_HEREDOC 69) (T_STATIC 21073) (T_STRING 21074) (T_STRING_CAST 72) (T_SWITCH 21075) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21076) (T_UNSET 21077) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21078) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21079) (alt_if_stmt_without_else 21080) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 21196) (dereferenceable_scalar 101) (expr 21081) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 21082) (if_stmt_without_else 21083) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21197) (static_member 121) (variable 124) (variable_class_name 125) ("{" 21085) ("~" 127)) 9616 ((inner_statement_list 21198)) 9617 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 21199) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 9618 ((";" 21200)) 9619 ((";" 21202)) 9620 ((";" 21204)) 9621 ((T_READONLY 17353) (T_STRING 17354) (backup_doc_comment 141) (function_name 21207)) 9622 ((backup_doc_comment 21208)) 9623 (("(" 21209)) 9624 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 21210)) 9625 ((")" 21211)) 9626 ((":" 166) (return_type 21212)) 9627 ((backup_fn_flags 21213)) 9628 (("{" 21214)) 9629 ((inner_statement_list 21215)) 9630 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21216) ("~" 127)) 9631 ((backup_fn_flags 21217)) 9632 ((T_STRING 21220)) 9633 ((T_EXTENDS 2963) (extends_from 21221)) 9634 ((T_IMPLEMENTS 2965) (implements_list 21222)) 9635 ((backup_doc_comment 21223)) 9636 (("{" 21224)) 9637 ((class_statement_list 21225)) 9638 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 21226)) 9639 ((":" 21231)) 9640 ((";" 21230)) 9641 ((inner_statement_list 21232)) 9642 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 21233) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 9643 ((";" 21234)) 9644 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21236) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9645 (("%" 299) (")" 21237) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9646 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21238) (";" 21012) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 21044) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21239) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 21240) ("{" 21053) ("~" 127)) 9647 ((inner_statement_list 21241)) 9648 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 21242) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 9649 ((";" 21243)) 9650 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 21245) (variable 17576) (variable_class_name 17577)) 9651 (("," 17589) (possible_comma 21246)) 9652 ((")" 21247)) 9653 ((";" 21248)) 9654 ((inner_statement_list 21250)) 9655 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21251) ("~" 127)) 9656 ((catch_list 21252)) 9657 ((T_CATCH 21253) (T_FINALLY 21254) (finally_statement 21255)) 9658 (("(" 21259)) 9659 (("{" 21256)) 9660 ((inner_statement_list 21257)) 9661 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21258) ("~" 127)) 9662 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 21260) (class_name 17636) (name 17637)) 9663 ((T_VARIABLE 17638) (optional_variable 21261) ("|" 17640)) 9664 ((")" 21262)) 9665 (("{" 21263)) 9666 ((inner_statement_list 21264)) 9667 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21265) ("~" 127)) 9668 ((backup_doc_comment 21267)) 9669 (("{" 21268)) 9670 ((class_statement_list 21269)) 9671 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 21270)) 9672 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21272) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9673 (("%" 299) (")" 21273) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9674 ((":" 21274) (switch_case_list 21275) ("{" 21276)) 9675 ((";" 21282) (case_list 21283)) 9676 ((";" 21277) (case_list 21278)) 9677 ((case_list 21280)) 9678 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 21279)) 9679 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 21281)) 9680 ((case_list 21286)) 9681 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 21284)) 9682 ((";" 21285)) 9683 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 21287)) 9684 ((";" 21288)) 9685 (("," 17937) (";" 21291)) 9686 ((";" 21293)) 9687 ((T_EXTENDS 21295) (interface_extends_list 21296)) 9688 ((T_NAME_FULLY_QUALIFIED 7604) (T_NAME_QUALIFIED 7605) (T_NAME_RELATIVE 7606) (T_STATIC 7607) (T_STRING 7608) (class_name 7609) (class_name_list 21301) (name 7611)) 9689 ((backup_doc_comment 21297)) 9690 (("{" 21298)) 9691 ((class_statement_list 21299)) 9692 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 21300)) 9693 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21303) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9694 (("%" 299) (")" 21304) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9695 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21305) (";" 21061) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 21062) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21063) (T_DEC 22) (T_DECLARE 21064) (T_DIR 24) (T_DNUMBER 25) (T_DO 21065) (T_DOUBLE_CAST 27) (T_ECHO 21066) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21067) (T_FOREACH 21068) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 21069) (T_GOTO 21070) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21071) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21072) (T_START_HEREDOC 69) (T_STATIC 21073) (T_STRING 21074) (T_STRING_CAST 72) (T_SWITCH 21075) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21076) (T_UNSET 21077) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21078) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21079) (alt_if_stmt_without_else 21080) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 21081) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 21082) (if_stmt_without_else 21083) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21306) (static_member 121) (variable 124) (variable_class_name 125) ("{" 21085) ("~" 127)) 9696 ((inner_statement_list 21307)) 9697 ((")" 21309)) 9698 ((";" 21310)) 9699 ((";" 21312)) 9700 (("," 18988) (";" 21314)) 9701 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 21316) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 9702 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 21317) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 9703 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 21318) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 9704 ((")" 21319) (T_DOUBLE_ARROW 21320)) 9705 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21323) (";" 21012) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 21044) (fn 104) (foreach_statement 21329) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21325) (static_member 121) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 9706 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 21321) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 9707 ((")" 21322)) 9708 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21323) (";" 21012) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 21044) (fn 104) (foreach_statement 21324) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21325) (static_member 121) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 9709 ((inner_statement_list 21326)) 9710 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 21327) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 9711 ((";" 21328)) 9712 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 21331) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9713 ((";" 21332)) 9714 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 21333) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9715 ((";" 21334)) 9716 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 21335) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 9717 ((")" 21336)) 9718 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21337) (";" 21012) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 21044) (fn 104) (for_statement 21338) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21339) (static_member 121) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 9719 ((inner_statement_list 21340)) 9720 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 21341) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 9721 ((";" 21342)) 9722 ((":" 21344) (enum_backing_type 21345)) 9723 (("(" 2256) ("?" 21351) (T_ARRAY 21352) (T_CALLABLE 21353) (T_NAME_FULLY_QUALIFIED 21354) (T_NAME_QUALIFIED 21355) (T_NAME_RELATIVE 21356) (T_STATIC 21357) (T_STRING 21358) (intersection_type 21359) (name 21360) (type 21361) (type_expr 21362) (type_without_static 21363) (union_type 21364) (union_type_element 21365)) 9724 ((T_IMPLEMENTS 2965) (implements_list 21346)) 9725 ((backup_doc_comment 21347)) 9726 (("{" 21348)) 9727 ((class_statement_list 21349)) 9728 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 21350)) 9729 ((T_ARRAY 21396) (T_CALLABLE 21397) (T_NAME_FULLY_QUALIFIED 7614) (T_NAME_QUALIFIED 7615) (T_NAME_RELATIVE 7616) (T_STATIC 21398) (T_STRING 7618) (name 21399) (type 21400) (type_without_static 21401)) 9730 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 21394)) 9731 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 21383)) 9732 (("|" 21381)) 9733 (("|" 21366)) 9734 (("(" 21367) (T_ARRAY 21368) (T_CALLABLE 21369) (T_NAME_FULLY_QUALIFIED 21370) (T_NAME_QUALIFIED 21371) (T_NAME_RELATIVE 21372) (T_STATIC 21373) (T_STRING 21374) (name 21375) (type 21376) (type_without_static 21377) (union_type_element 21378)) 9735 ((T_ARRAY 2285) (T_CALLABLE 2286) (T_NAME_FULLY_QUALIFIED 2287) (T_NAME_QUALIFIED 2288) (T_NAME_RELATIVE 2289) (T_STATIC 2290) (T_STRING 2291) (intersection_type 21379) (name 2293) (type 2294) (type_without_static 2295)) 9736 ((")" 21380) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2308)) 9737 (("(" 21367) (T_ARRAY 21368) (T_CALLABLE 21369) (T_NAME_FULLY_QUALIFIED 21370) (T_NAME_QUALIFIED 21371) (T_NAME_RELATIVE 21372) (T_STATIC 21373) (T_STRING 21374) (name 21375) (type 21376) (type_without_static 21377) (union_type_element 21382)) 9738 ((T_ARRAY 21384) (T_CALLABLE 21385) (T_NAME_FULLY_QUALIFIED 21386) (T_NAME_QUALIFIED 21387) (T_NAME_RELATIVE 21388) (T_STATIC 21389) (T_STRING 21390) (name 21391) (type 21392) (type_without_static 21393)) 9739 ((T_ARRAY 21384) (T_CALLABLE 21385) (T_NAME_FULLY_QUALIFIED 21386) (T_NAME_QUALIFIED 21387) (T_NAME_RELATIVE 21388) (T_STATIC 21389) (T_STRING 21390) (name 21391) (type 21395) (type_without_static 21393)) 9740 (("," 20829) (";" 21404)) 9741 ((T_WHILE 21406)) 9742 (("(" 21407)) 9743 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21408) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9744 (("%" 299) (")" 21409) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9745 ((";" 21410)) 9746 ((T_STRING 21002) (const_decl 21003) (const_list 21412)) 9747 ((")" 21413) ("," 21006)) 9748 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21414) (";" 21012) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 21415) (dereferenceable_scalar 101) (expr 21044) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21416) (static_member 121) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 9749 ((inner_statement_list 21417)) 9750 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 21418) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 9751 ((";" 21419)) 9752 ((";" 21421)) 9753 ((T_EXTENDS 2963) (extends_from 21423)) 9754 ((T_IMPLEMENTS 2965) (implements_list 21424)) 9755 ((backup_doc_comment 21425)) 9756 (("{" 21426)) 9757 ((class_statement_list 21427)) 9758 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 21428)) 9759 ((";" 21430)) 9760 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 21432) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 9761 (("%" 1111) ("*" 1112) ("+" 1113) ("-" 1114) ("." 1115) ("/" 1116) ("<" 1117) (">" 1118) ("?" 1119) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1120) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1121) (T_BOOLEAN_AND 1122) (T_BOOLEAN_OR 1123) (T_COALESCE 1124) (T_INSTANCEOF 1125) (T_IS_EQUAL 1126) (T_IS_GREATER_OR_EQUAL 1127) (T_IS_IDENTICAL 1128) (T_IS_NOT_EQUAL 1129) (T_IS_NOT_IDENTICAL 1130) (T_IS_SMALLER_OR_EQUAL 1131) (T_LOGICAL_AND 1132) (T_LOGICAL_OR 1133) (T_LOGICAL_XOR 1134) (T_POW 1135) (T_SL 1136) (T_SPACESHIP 1137) (T_SR 1138) ("^" 1139) (backup_doc_comment 21433) ("|" 1140)) 9762 ((";" 21435)) 9763 ((";" 21437)) 9764 ((":" 21443)) 9765 ((";" 21442)) 9766 ((inner_statement_list 21444)) 9767 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 21445) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 9768 ((";" 21446)) 9769 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21448) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9770 (("%" 299) (")" 21449) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9771 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21450) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21451) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 21452) ("{" 20857) ("~" 127)) 9772 ((inner_statement_list 21453)) 9773 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 21454) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 9774 ((";" 21455)) 9775 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 21457) (variable 17576) (variable_class_name 17577)) 9776 (("," 17589) (possible_comma 21458)) 9777 ((")" 21459)) 9778 ((";" 21460)) 9779 ((inner_statement_list 21462)) 9780 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21463) ("~" 127)) 9781 ((catch_list 21464)) 9782 ((T_CATCH 21465) (T_FINALLY 21466) (finally_statement 21467)) 9783 (("(" 21471)) 9784 (("{" 21468)) 9785 ((inner_statement_list 21469)) 9786 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21470) ("~" 127)) 9787 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 21472) (class_name 17636) (name 17637)) 9788 ((T_VARIABLE 17638) (optional_variable 21473) ("|" 17640)) 9789 ((")" 21474)) 9790 (("{" 21475)) 9791 ((inner_statement_list 21476)) 9792 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21477) ("~" 127)) 9793 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21479) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9794 (("%" 299) (")" 21480) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9795 ((":" 21481) (switch_case_list 21482) ("{" 21483)) 9796 ((";" 21489) (case_list 21490)) 9797 ((";" 21484) (case_list 21485)) 9798 ((case_list 21487)) 9799 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 21486)) 9800 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 21488)) 9801 ((case_list 21493)) 9802 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 21491)) 9803 ((";" 21492)) 9804 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 21494)) 9805 ((";" 21495)) 9806 (("," 17937) (";" 21498)) 9807 ((";" 21500)) 9808 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21502) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9809 (("%" 299) (")" 21503) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9810 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21305) (";" 20870) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20871) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20872) (T_DEC 22) (T_DECLARE 20873) (T_DIR 24) (T_DNUMBER 25) (T_DO 20874) (T_DOUBLE_CAST 27) (T_ECHO 20875) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20876) (T_FOREACH 20877) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20878) (T_GOTO 20879) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20880) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20881) (T_START_HEREDOC 69) (T_STATIC 20882) (T_STRING 20883) (T_STRING_CAST 72) (T_SWITCH 20884) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20885) (T_UNSET 20886) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20887) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20888) (alt_if_stmt_without_else 20889) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20890) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20891) (if_stmt_without_else 20892) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21504) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20894) ("~" 127)) 9811 ((";" 21506)) 9812 (("," 18988) (";" 21508)) 9813 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 21510) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 9814 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 21511) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 9815 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 21512) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 9816 ((")" 21513) (T_DOUBLE_ARROW 21514)) 9817 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21517) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (foreach_statement 21523) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21519) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 9818 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 21515) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 9819 ((")" 21516)) 9820 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21517) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (foreach_statement 21518) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21519) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 9821 ((inner_statement_list 21520)) 9822 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 21521) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 9823 ((";" 21522)) 9824 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 21525) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9825 ((";" 21526)) 9826 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 21527) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9827 ((";" 21528)) 9828 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 21529) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 9829 ((")" 21530)) 9830 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21531) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (for_statement 21532) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21533) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 9831 ((inner_statement_list 21534)) 9832 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 21535) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 9833 ((";" 21536)) 9834 (("," 20829) (";" 21538)) 9835 ((T_WHILE 21540)) 9836 (("(" 21541)) 9837 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21542) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9838 (("%" 299) (")" 21543) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9839 ((";" 21544)) 9840 ((T_STRING 21002) (const_decl 21003) (const_list 21546)) 9841 ((")" 21547) ("," 21006)) 9842 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21548) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 21549) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21550) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 9843 ((inner_statement_list 21551)) 9844 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 21552) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 9845 ((";" 21553)) 9846 ((";" 21555)) 9847 ((";" 21557)) 9848 ((T_STRING 21002) (const_decl 21003) (const_list 21559)) 9849 ((")" 21560) ("," 21006)) 9850 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21561) (";" 20703) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20704) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20705) (T_DEC 22) (T_DECLARE 20706) (T_DIR 24) (T_DNUMBER 25) (T_DO 20707) (T_DOUBLE_CAST 27) (T_ECHO 20708) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20709) (T_FOREACH 20710) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20711) (T_GOTO 20712) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20713) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20714) (T_START_HEREDOC 69) (T_STATIC 20715) (T_STRING 20716) (T_STRING_CAST 72) (T_SWITCH 20717) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20718) (T_UNSET 20719) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20720) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20721) (alt_if_stmt_without_else 20722) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 21562) (dereferenceable_scalar 101) (expr 20723) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20724) (if_stmt_without_else 20725) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21563) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20727) ("~" 127)) 9851 ((inner_statement_list 21564)) 9852 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 21565) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 9853 ((";" 21566)) 9854 ((";" 21568)) 9855 ((";" 21570)) 9856 ((T_READONLY 17353) (T_STRING 17354) (backup_doc_comment 141) (function_name 21573)) 9857 ((backup_doc_comment 21574)) 9858 (("(" 21575)) 9859 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 21576)) 9860 ((")" 21577)) 9861 ((":" 166) (return_type 21578)) 9862 ((backup_fn_flags 21579)) 9863 (("{" 21580)) 9864 ((inner_statement_list 21581)) 9865 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21582) ("~" 127)) 9866 ((backup_fn_flags 21583)) 9867 ((T_STRING 21586)) 9868 ((T_EXTENDS 2963) (extends_from 21587)) 9869 ((T_IMPLEMENTS 2965) (implements_list 21588)) 9870 ((backup_doc_comment 21589)) 9871 (("{" 21590)) 9872 ((class_statement_list 21591)) 9873 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 21592)) 9874 ((":" 21597)) 9875 ((";" 21596)) 9876 ((inner_statement_list 21598)) 9877 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 21599) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 9878 ((";" 21600)) 9879 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21602) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9880 (("%" 299) (")" 21603) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9881 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21604) (";" 20654) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20686) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21605) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 21606) ("{" 20695) ("~" 127)) 9882 ((inner_statement_list 21607)) 9883 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 21608) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 9884 ((";" 21609)) 9885 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 21611) (variable 17576) (variable_class_name 17577)) 9886 (("," 17589) (possible_comma 21612)) 9887 ((")" 21613)) 9888 ((";" 21614)) 9889 ((inner_statement_list 21616)) 9890 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21617) ("~" 127)) 9891 ((catch_list 21618)) 9892 ((T_CATCH 21619) (T_FINALLY 21620) (finally_statement 21621)) 9893 (("(" 21625)) 9894 (("{" 21622)) 9895 ((inner_statement_list 21623)) 9896 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21624) ("~" 127)) 9897 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 21626) (class_name 17636) (name 17637)) 9898 ((T_VARIABLE 17638) (optional_variable 21627) ("|" 17640)) 9899 ((")" 21628)) 9900 (("{" 21629)) 9901 ((inner_statement_list 21630)) 9902 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21631) ("~" 127)) 9903 ((backup_doc_comment 21633)) 9904 (("{" 21634)) 9905 ((class_statement_list 21635)) 9906 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 21636)) 9907 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21638) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9908 (("%" 299) (")" 21639) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9909 ((":" 21640) (switch_case_list 21641) ("{" 21642)) 9910 ((";" 21648) (case_list 21649)) 9911 ((";" 21643) (case_list 21644)) 9912 ((case_list 21646)) 9913 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 21645)) 9914 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 21647)) 9915 ((case_list 21652)) 9916 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 21650)) 9917 ((";" 21651)) 9918 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 21653)) 9919 ((";" 21654)) 9920 (("," 17937) (";" 21657)) 9921 ((";" 21659)) 9922 ((T_EXTENDS 21295) (interface_extends_list 21661)) 9923 ((backup_doc_comment 21662)) 9924 (("{" 21663)) 9925 ((class_statement_list 21664)) 9926 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 21665)) 9927 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21667) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9928 (("%" 299) (")" 21668) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9929 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21305) (";" 20703) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20704) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20705) (T_DEC 22) (T_DECLARE 20706) (T_DIR 24) (T_DNUMBER 25) (T_DO 20707) (T_DOUBLE_CAST 27) (T_ECHO 20708) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20709) (T_FOREACH 20710) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20711) (T_GOTO 20712) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20713) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20714) (T_START_HEREDOC 69) (T_STATIC 20715) (T_STRING 20716) (T_STRING_CAST 72) (T_SWITCH 20717) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20718) (T_UNSET 20719) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20720) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20721) (alt_if_stmt_without_else 20722) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20723) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20724) (if_stmt_without_else 20725) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21669) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20727) ("~" 127)) 9930 ((")" 21671)) 9931 ((";" 21672)) 9932 ((";" 21674)) 9933 (("," 18988) (";" 21676)) 9934 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 21678) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 9935 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 21679) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 9936 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 21680) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 9937 ((")" 21681) (T_DOUBLE_ARROW 21682)) 9938 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21685) (";" 20654) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20686) (fn 104) (foreach_statement 21691) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21687) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 9939 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 21683) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 9940 ((")" 21684)) 9941 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21685) (";" 20654) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20686) (fn 104) (foreach_statement 21686) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21687) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 9942 ((inner_statement_list 21688)) 9943 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 21689) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 9944 ((";" 21690)) 9945 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 21693) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9946 ((";" 21694)) 9947 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 21695) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 9948 ((";" 21696)) 9949 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 21697) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 9950 ((")" 21698)) 9951 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21699) (";" 20654) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20686) (fn 104) (for_statement 21700) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21701) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 9952 ((inner_statement_list 21702)) 9953 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 21703) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 9954 ((";" 21704)) 9955 ((":" 21344) (enum_backing_type 21706)) 9956 ((T_IMPLEMENTS 2965) (implements_list 21707)) 9957 ((backup_doc_comment 21708)) 9958 (("{" 21709)) 9959 ((class_statement_list 21710)) 9960 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 21711)) 9961 (("," 20829) (";" 21714)) 9962 ((T_WHILE 21716)) 9963 (("(" 21717)) 9964 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21718) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9965 (("%" 299) (")" 21719) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9966 ((";" 21720)) 9967 ((T_STRING 21002) (const_decl 21003) (const_list 21722)) 9968 ((")" 21723) ("," 21006)) 9969 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21724) (";" 20654) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 21725) (dereferenceable_scalar 101) (expr 20686) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21726) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 9970 ((inner_statement_list 21727)) 9971 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 21728) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 9972 ((";" 21729)) 9973 ((";" 21731)) 9974 ((T_EXTENDS 2963) (extends_from 21733)) 9975 ((T_IMPLEMENTS 2965) (implements_list 21734)) 9976 ((backup_doc_comment 21735)) 9977 (("{" 21736)) 9978 ((class_statement_list 21737)) 9979 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 21738)) 9980 ((";" 21740)) 9981 (("," 20829) (";" 21742)) 9982 ((T_WHILE 21744)) 9983 (("(" 21745)) 9984 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21746) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 9985 (("%" 299) (")" 21747) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 9986 ((";" 21748)) 9987 ((T_STRING 21002) (const_decl 21003) (const_list 21750)) 9988 ((")" 21751) ("," 21006)) 9989 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21752) (";" 20525) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20526) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20527) (T_DEC 22) (T_DECLARE 20528) (T_DIR 24) (T_DNUMBER 25) (T_DO 20529) (T_DOUBLE_CAST 27) (T_ECHO 20530) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20531) (T_FOREACH 20532) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20533) (T_GOTO 20534) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20535) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20536) (T_START_HEREDOC 69) (T_STATIC 20537) (T_STRING 20538) (T_STRING_CAST 72) (T_SWITCH 20539) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20540) (T_UNSET 20541) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20542) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20543) (alt_if_stmt_without_else 20544) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 21753) (dereferenceable_scalar 101) (expr 20545) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20546) (if_stmt_without_else 20547) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21754) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20549) ("~" 127)) 9990 ((inner_statement_list 21755)) 9991 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 21756) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 9992 ((";" 21757)) 9993 ((";" 21759)) 9994 ((";" 21761)) 9995 ((T_READONLY 17353) (T_STRING 17354) (backup_doc_comment 141) (function_name 21764)) 9996 ((backup_doc_comment 21765)) 9997 (("(" 21766)) 9998 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 21767)) 9999 ((")" 21768)) 10000 ((":" 166) (return_type 21769)) 10001 ((backup_fn_flags 21770)) 10002 (("{" 21771)) 10003 ((inner_statement_list 21772)) 10004 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21773) ("~" 127)) 10005 ((backup_fn_flags 21774)) 10006 ((T_STRING 21777)) 10007 ((T_EXTENDS 2963) (extends_from 21778)) 10008 ((T_IMPLEMENTS 2965) (implements_list 21779)) 10009 ((backup_doc_comment 21780)) 10010 (("{" 21781)) 10011 ((class_statement_list 21782)) 10012 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 21783)) 10013 ((":" 21788)) 10014 ((";" 21787)) 10015 ((inner_statement_list 21789)) 10016 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 21790) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 10017 ((";" 21791)) 10018 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21793) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10019 (("%" 299) (")" 21794) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10020 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21795) (";" 20476) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20508) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21796) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 21797) ("{" 20517) ("~" 127)) 10021 ((inner_statement_list 21798)) 10022 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 21799) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 10023 ((";" 21800)) 10024 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 21802) (variable 17576) (variable_class_name 17577)) 10025 (("," 17589) (possible_comma 21803)) 10026 ((")" 21804)) 10027 ((";" 21805)) 10028 ((inner_statement_list 21807)) 10029 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21808) ("~" 127)) 10030 ((catch_list 21809)) 10031 ((T_CATCH 21810) (T_FINALLY 21811) (finally_statement 21812)) 10032 (("(" 21816)) 10033 (("{" 21813)) 10034 ((inner_statement_list 21814)) 10035 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21815) ("~" 127)) 10036 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 21817) (class_name 17636) (name 17637)) 10037 ((T_VARIABLE 17638) (optional_variable 21818) ("|" 17640)) 10038 ((")" 21819)) 10039 (("{" 21820)) 10040 ((inner_statement_list 21821)) 10041 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 21822) ("~" 127)) 10042 ((backup_doc_comment 21824)) 10043 (("{" 21825)) 10044 ((class_statement_list 21826)) 10045 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 21827)) 10046 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21829) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10047 (("%" 299) (")" 21830) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10048 ((":" 21831) (switch_case_list 21832) ("{" 21833)) 10049 ((";" 21839) (case_list 21840)) 10050 ((";" 21834) (case_list 21835)) 10051 ((case_list 21837)) 10052 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 21836)) 10053 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 21838)) 10054 ((case_list 21843)) 10055 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 21841)) 10056 ((";" 21842)) 10057 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 21844)) 10058 ((";" 21845)) 10059 (("," 17937) (";" 21848)) 10060 ((";" 21850)) 10061 ((T_EXTENDS 21295) (interface_extends_list 21852)) 10062 ((backup_doc_comment 21853)) 10063 (("{" 21854)) 10064 ((class_statement_list 21855)) 10065 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 21856)) 10066 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21858) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10067 (("%" 299) (")" 21859) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10068 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21305) (";" 20525) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20526) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20527) (T_DEC 22) (T_DECLARE 20528) (T_DIR 24) (T_DNUMBER 25) (T_DO 20529) (T_DOUBLE_CAST 27) (T_ECHO 20530) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20531) (T_FOREACH 20532) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20533) (T_GOTO 20534) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20535) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20536) (T_START_HEREDOC 69) (T_STATIC 20537) (T_STRING 20538) (T_STRING_CAST 72) (T_SWITCH 20539) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20540) (T_UNSET 20541) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20542) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20543) (alt_if_stmt_without_else 20544) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20545) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20546) (if_stmt_without_else 20547) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21860) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20549) ("~" 127)) 10069 ((")" 21862)) 10070 ((";" 21863)) 10071 ((";" 21865)) 10072 (("," 18988) (";" 21867)) 10073 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 21869) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 10074 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 21870) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 10075 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 21871) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 10076 ((")" 21872) (T_DOUBLE_ARROW 21873)) 10077 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21876) (";" 20476) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20508) (fn 104) (foreach_statement 21882) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21878) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 10078 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 21874) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 10079 ((")" 21875)) 10080 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21876) (";" 20476) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20508) (fn 104) (foreach_statement 21877) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21878) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 10081 ((inner_statement_list 21879)) 10082 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 21880) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 10083 ((";" 21881)) 10084 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 21884) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 10085 ((";" 21885)) 10086 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 21886) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 10087 ((";" 21887)) 10088 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 21888) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 10089 ((")" 21889)) 10090 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21890) (";" 20476) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20508) (fn 104) (for_statement 21891) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21892) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 10091 ((inner_statement_list 21893)) 10092 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 21894) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 10093 ((";" 21895)) 10094 ((":" 21344) (enum_backing_type 21897)) 10095 ((T_IMPLEMENTS 2965) (implements_list 21898)) 10096 ((backup_doc_comment 21899)) 10097 (("{" 21900)) 10098 ((class_statement_list 21901)) 10099 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 21902)) 10100 (("," 20829) (";" 21905)) 10101 ((T_WHILE 21907)) 10102 (("(" 21908)) 10103 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 21909) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10104 (("%" 299) (")" 21910) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10105 ((";" 21911)) 10106 ((T_STRING 21002) (const_decl 21003) (const_list 21913)) 10107 ((")" 21914) ("," 21006)) 10108 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21915) (";" 20476) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 21916) (dereferenceable_scalar 101) (expr 20508) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21917) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 10109 ((inner_statement_list 21918)) 10110 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 21919) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 10111 ((";" 21920)) 10112 ((";" 21922)) 10113 ((T_EXTENDS 2963) (extends_from 21924)) 10114 ((T_IMPLEMENTS 2965) (implements_list 21925)) 10115 ((backup_doc_comment 21926)) 10116 (("{" 21927)) 10117 ((class_statement_list 21928)) 10118 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 21929)) 10119 ((";" 21931)) 10120 (("$" 20442) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 20445) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 21933) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 21934) ("{" 1658)) 10121 (("(" 20462) (argument_list 21935)) 10122 (("$" 20442) (T_STRING 21942) (T_VARIABLE 20445) (property_name 21949) (simple_variable 21944) ("{" 21945)) 10123 (("$" 20442) (T_STRING 21942) (T_VARIABLE 20445) (property_name 21943) (simple_variable 21944) ("{" 21945)) 10124 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 21940) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 10125 (("]" 21941)) 10126 (("(" 20462) (argument_list 21948)) 10127 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 21946) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10128 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 21947)) 10129 (("(" 20462) (argument_list 21950)) 10130 (("]" 21953)) 10131 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 21956) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 10132 ((")" 21957)) 10133 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 21960) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10134 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 21961)) 10135 (("$" 20411) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 20414) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 21964) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 21965) ("{" 1658)) 10136 (("(" 20431) (argument_list 21966)) 10137 (("$" 20411) (T_STRING 21973) (T_VARIABLE 20414) (property_name 21980) (simple_variable 21975) ("{" 21976)) 10138 (("$" 20411) (T_STRING 21973) (T_VARIABLE 20414) (property_name 21974) (simple_variable 21975) ("{" 21976)) 10139 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 21971) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 10140 (("]" 21972)) 10141 (("(" 20431) (argument_list 21979)) 10142 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 21977) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10143 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 21978)) 10144 (("(" 20431) (argument_list 21981)) 10145 (("]" 21984)) 10146 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 21987) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 10147 ((")" 21988)) 10148 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 21991) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10149 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 21992)) 10150 (("$" 18998) (T_ABSTRACT 19175) (T_ARRAY 19176) (T_AS 19177) (T_BREAK 19178) (T_CALLABLE 19179) (T_CASE 19180) (T_CATCH 19181) (T_CLASS 19182) (T_CLASS_C 19183) (T_CLONE 19184) (T_CONST 19185) (T_CONTINUE 19186) (T_DECLARE 19187) (T_DEFAULT 19188) (T_DIR 19189) (T_DO 19190) (T_ECHO 19191) (T_ELSE 19192) (T_ELSEIF 19193) (T_EMPTY 19194) (T_ENDDECLARE 19195) (T_ENDFOR 19196) (T_ENDFOREACH 19197) (T_ENDIF 19198) (T_ENDSWITCH 19199) (T_ENDWHILE 19200) (T_ENUM 19201) (T_EVAL 19202) (T_EXIT 19203) (T_EXTENDS 19204) (T_FILE 19205) (T_FINAL 19206) (T_FINALLY 19207) (T_FN 19208) (T_FOR 19209) (T_FOREACH 19210) (T_FUNCTION 19211) (T_FUNC_C 19212) (T_GLOBAL 19213) (T_GOTO 19214) (T_IF 19215) (T_IMPLEMENTS 19216) (T_INCLUDE 19217) (T_INCLUDE_ONCE 19218) (T_INSTANCEOF 19219) (T_INSTEADOF 19220) (T_INTERFACE 19221) (T_ISSET 19222) (T_LINE 19223) (T_LIST 19224) (T_LOGICAL_AND 19225) (T_LOGICAL_OR 19226) (T_LOGICAL_XOR 19227) (T_MATCH 19228) (T_METHOD_C 19229) (T_NAMESPACE 19230) (T_NEW 19231) (T_NS_C 19232) (T_PRINT 19233) (T_PRIVATE 19234) (T_PROPERTY_C 19235) (T_PROTECTED 19236) (T_PUBLIC 19237) (T_READONLY 19238) (T_REQUIRE 19239) (T_REQUIRE_ONCE 19240) (T_RETURN 19241) (T_STATIC 19242) (T_STRING 19243) (T_SWITCH 19244) (T_THROW 19245) (T_TRAIT 19246) (T_TRAIT_C 19247) (T_TRY 19248) (T_UNSET 19249) (T_USE 19250) (T_VAR 19251) (T_VARIABLE 19046) (T_WHILE 19252) (T_YIELD 19253) (identifier 21994) (member_name 21995) (reserved_non_modifiers 19256) (semi_reserved 19257) (simple_variable 21996) ("{" 21997)) 10151 (("(" 19262) (argument_list 22000)) 10152 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 21998) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10153 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 21999)) 10154 ((T_FN 35) (T_FUNCTION 129) (fn 19060) #'19061 (inline_function 22004)) 10155 (("$" 18998) (T_STRING 22010) (T_VARIABLE 19046) (property_name 22017) (simple_variable 22012) ("{" 22013)) 10156 (("$" 18998) (T_STRING 22010) (T_VARIABLE 19046) (property_name 22011) (simple_variable 22012) ("{" 22013)) 10157 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 22008) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 10158 (("]" 22009)) 10159 (("(" 19262) (argument_list 22016)) 10160 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 22014) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10161 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 22015)) 10162 (("(" 19262) (argument_list 22018)) 10163 (("`" 22020)) 10164 (("]" 22022)) 10165 (("=" 22023)) 10166 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22024) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 10167 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22546) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10168 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 22544) (encaps_var 5237)) 10169 (("$" 22028) (T_VARIABLE 22076) (simple_variable 22540) ("{" 22541)) 10170 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22538) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10171 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22537) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10172 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22536) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10173 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22535) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10174 (("(" 22532)) 10175 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22531) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10176 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22530) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10177 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22529) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10178 (("\"" 1520) ("$" 22331) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 22332) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 22333) ("[" 1542) (array_object_dereferenceable 22334) (callable_expr 22335) (callable_variable 22336) (class_constant 1546) (class_name 22337) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 22338) (name 22339) (new_dereferenceable 1552) (simple_variable 22340) (static_member 22341) (variable 22528) (variable_class_name 22343)) 10179 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22527) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10180 (("(" 22524)) 10181 (("(" 22521)) 10182 (("(" 22512) (argument_list 22513) (ctor_arguments 22514)) 10183 (("\"" 1520) ("$" 22331) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 22332) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 22333) ("[" 1542) (array_object_dereferenceable 22334) (callable_expr 22335) (callable_variable 22336) (class_constant 1546) (class_name 22337) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 22338) (name 22339) (new_dereferenceable 1552) (simple_variable 22340) (static_member 22341) (variable 22511) (variable_class_name 22343)) 10184 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22510) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10185 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22509) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10186 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22508) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10187 (("(" 22504)) 10188 (("(" 22499)) 10189 (("(" 22493)) 10190 (("$" 22331) ("(" 22456) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 22457) (T_NAME_QUALIFIED 22458) (T_NAME_RELATIVE 22459) (T_READONLY 2931) (T_STATIC 22460) (T_STRING 22461) (T_VARIABLE 22333) (anonymous_class 22462) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 22463) (attribute 2937) (attributes 22464) (class_modifier 2939) (class_name 22465) (class_name_reference 22466) (name 22467) (new_variable 22468) (simple_variable 22469)) 10191 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22455) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10192 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22454) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10193 (("(" 22292) (argument_list 22453)) 10194 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22452) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10195 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22451) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10196 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 22446) (T_END_HEREDOC 22447) (T_VARIABLE 2896) (encaps_list 22448) (encaps_var 2898)) 10197 ((T_FN 35) (T_FUNCTION 129) (fn 22090) #'22091 (inline_function 22445)) 10198 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22444) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10199 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22443) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10200 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22442) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10201 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22439) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10202 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22438) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10203 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 22434) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 10204 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 22432) (encaps_list 2804) (encaps_var 2805)) 10205 ((T_NULLSAFE_OBJECT_OPERATOR 22418) (T_OBJECT_OPERATOR 22419) ("[" 22420)) 10206 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 22415) (attribute 2733) (fn 22090) #'22091 (inline_function 22416)) 10207 (("(" 22292) (argument_list 22414)) 10208 ((T_PAAMAYIM_NEKUDOTAYIM 22406)) 10209 (("%" 22106) ("*" 22107) ("+" 22108) ("-" 22109) ("." 22110) ("/" 22111) ("<" 22112) (">" 22113) ("?" 22114) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22115) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22116) (T_BOOLEAN_AND 22117) (T_BOOLEAN_OR 22118) (T_COALESCE 22119) (T_DOUBLE_ARROW 22404) (T_INSTANCEOF 22120) (T_IS_EQUAL 22121) (T_IS_GREATER_OR_EQUAL 22122) (T_IS_IDENTICAL 22123) (T_IS_NOT_EQUAL 22124) (T_IS_NOT_IDENTICAL 22125) (T_IS_SMALLER_OR_EQUAL 22126) (T_LOGICAL_AND 22127) (T_LOGICAL_OR 22128) (T_LOGICAL_XOR 22129) (T_POW 22130) (T_SL 22131) (T_SPACESHIP 22132) (T_SR 22133) ("^" 22134) ("|" 22135)) 10210 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 22393)) 10211 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 22381)) 10212 (("(" 22292) (argument_list 22380)) 10213 (("=" 22300) (T_AND_EQUAL 22301) (T_COALESCE_EQUAL 22302) (T_CONCAT_EQUAL 22303) (T_DEC 22304) (T_DIV_EQUAL 22305) (T_INC 22306) (T_MINUS_EQUAL 22307) (T_MOD_EQUAL 22308) (T_MUL_EQUAL 22309) (T_OR_EQUAL 22310) (T_PLUS_EQUAL 22311) (T_POW_EQUAL 22312) (T_SL_EQUAL 22313) (T_SR_EQUAL 22314) (T_XOR_EQUAL 22315)) 10214 ((T_PAAMAYIM_NEKUDOTAYIM 22204)) 10215 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22105) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10216 (("%" 22106) ("*" 22107) ("+" 22108) ("-" 22109) ("." 22110) ("/" 22111) ("<" 22112) (">" 22113) ("?" 22114) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22115) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22116) (T_BOOLEAN_AND 22117) (T_BOOLEAN_OR 22118) (T_COALESCE 22119) (T_INSTANCEOF 22120) (T_IS_EQUAL 22121) (T_IS_GREATER_OR_EQUAL 22122) (T_IS_IDENTICAL 22123) (T_IS_NOT_EQUAL 22124) (T_IS_NOT_IDENTICAL 22125) (T_IS_SMALLER_OR_EQUAL 22126) (T_LOGICAL_AND 22127) (T_LOGICAL_OR 22128) (T_LOGICAL_XOR 22129) (T_POW 22130) (T_SL 22131) (T_SPACESHIP 22132) (T_SR 22133) ("^" 22134) ("|" 22135)) 10217 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22203) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10218 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22202) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10219 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22201) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10220 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22200) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10221 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22199) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10222 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22198) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10223 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22197) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10224 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22196) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10225 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 22191) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 22192) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 10226 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22190) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10227 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22189) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10228 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22188) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10229 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22187) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10230 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22186) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10231 (("$" 22151) ("(" 22152) (T_NAME_FULLY_QUALIFIED 22153) (T_NAME_QUALIFIED 22154) (T_NAME_RELATIVE 22155) (T_STATIC 22156) (T_STRING 22157) (T_VARIABLE 22158) (class_name 22159) (class_name_reference 22160) (name 22161) (new_variable 22162) (simple_variable 22163)) 10232 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22150) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10233 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22149) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10234 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22148) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10235 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22147) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10236 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22146) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10237 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22145) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10238 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22144) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10239 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22143) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10240 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22142) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10241 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22141) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10242 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22140) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10243 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22139) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10244 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22138) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10245 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22137) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10246 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22136) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10247 (("$" 22151) (T_VARIABLE 22158) (simple_variable 22182) ("{" 22183)) 10248 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22180) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10249 ((T_PAAMAYIM_NEKUDOTAYIM 22178)) 10250 ((T_NULLSAFE_OBJECT_OPERATOR 22164) (T_OBJECT_OPERATOR 22165) (T_PAAMAYIM_NEKUDOTAYIM 22166) ("[" 22167)) 10251 (("$" 22151) (T_STRING 22171) (T_VARIABLE 22158) (property_name 22177) (simple_variable 22173) ("{" 22174)) 10252 (("$" 22151) (T_STRING 22171) (T_VARIABLE 22158) (property_name 22172) (simple_variable 22173) ("{" 22174)) 10253 (("$" 22151) (T_VARIABLE 22158) (simple_variable 22170)) 10254 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 22168) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 10255 (("]" 22169)) 10256 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 22175) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10257 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 22176)) 10258 (("$" 22151) (T_VARIABLE 22158) (simple_variable 22179)) 10259 (("%" 299) (")" 22181) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10260 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 22184) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10261 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 22185)) 10262 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22195) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10263 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 22193) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 10264 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22194) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10265 (("$" 22028) (T_ABSTRACT 22205) (T_ARRAY 22206) (T_AS 22207) (T_BREAK 22208) (T_CALLABLE 22209) (T_CASE 22210) (T_CATCH 22211) (T_CLASS 22212) (T_CLASS_C 22213) (T_CLONE 22214) (T_CONST 22215) (T_CONTINUE 22216) (T_DECLARE 22217) (T_DEFAULT 22218) (T_DIR 22219) (T_DO 22220) (T_ECHO 22221) (T_ELSE 22222) (T_ELSEIF 22223) (T_EMPTY 22224) (T_ENDDECLARE 22225) (T_ENDFOR 22226) (T_ENDFOREACH 22227) (T_ENDIF 22228) (T_ENDSWITCH 22229) (T_ENDWHILE 22230) (T_ENUM 22231) (T_EVAL 22232) (T_EXIT 22233) (T_EXTENDS 22234) (T_FILE 22235) (T_FINAL 22236) (T_FINALLY 22237) (T_FN 22238) (T_FOR 22239) (T_FOREACH 22240) (T_FUNCTION 22241) (T_FUNC_C 22242) (T_GLOBAL 22243) (T_GOTO 22244) (T_IF 22245) (T_IMPLEMENTS 22246) (T_INCLUDE 22247) (T_INCLUDE_ONCE 22248) (T_INSTANCEOF 22249) (T_INSTEADOF 22250) (T_INTERFACE 22251) (T_ISSET 22252) (T_LINE 22253) (T_LIST 22254) (T_LOGICAL_AND 22255) (T_LOGICAL_OR 22256) (T_LOGICAL_XOR 22257) (T_MATCH 22258) (T_METHOD_C 22259) (T_NAMESPACE 22260) (T_NEW 22261) (T_NS_C 22262) (T_PRINT 22263) (T_PRIVATE 22264) (T_PROPERTY_C 22265) (T_PROTECTED 22266) (T_PUBLIC 22267) (T_READONLY 22268) (T_REQUIRE 22269) (T_REQUIRE_ONCE 22270) (T_RETURN 22271) (T_STATIC 22272) (T_STRING 22273) (T_SWITCH 22274) (T_THROW 22275) (T_TRAIT 22276) (T_TRAIT_C 22277) (T_TRY 22278) (T_UNSET 22279) (T_USE 22280) (T_VAR 22281) (T_VARIABLE 22076) (T_WHILE 22282) (T_YIELD 22283) (identifier 22284) (member_name 22285) (reserved_non_modifiers 22286) (semi_reserved 22287) (simple_variable 22288) ("{" 22289)) 10266 (("(" 22292) (argument_list 22293)) 10267 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 22290) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10268 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 22291)) 10269 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 22294) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 22295) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 22296) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10270 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 22299) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10271 (("," 1480) (possible_comma 22297)) 10272 ((")" 22298)) 10273 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (ampersand 22329) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22330) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10274 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22328) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10275 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22327) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10276 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22326) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10277 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22325) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10278 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22324) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10279 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22323) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10280 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22322) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10281 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22321) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10282 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22320) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10283 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22319) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10284 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22318) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10285 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22317) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10286 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22316) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10287 (("\"" 1520) ("$" 22331) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 22332) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 22333) ("[" 1542) (array_object_dereferenceable 22334) (callable_expr 22335) (callable_variable 22336) (class_constant 1546) (class_name 22337) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 22338) (name 22339) (new_dereferenceable 1552) (simple_variable 22340) (static_member 22341) (variable 22342) (variable_class_name 22343)) 10288 (("$" 22331) (T_VARIABLE 22333) (simple_variable 22376) ("{" 22377)) 10289 (("(" 22347) (argument_list 22375)) 10290 ((T_NULLSAFE_OBJECT_OPERATOR 22361) (T_OBJECT_OPERATOR 22362) ("[" 22363)) 10291 (("(" 22347) (argument_list 22360)) 10292 ((T_PAAMAYIM_NEKUDOTAYIM 22356)) 10293 (("(" 22347) (argument_list 22355)) 10294 ((T_PAAMAYIM_NEKUDOTAYIM 22344)) 10295 (("$" 22331) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 22333) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 22345) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 22346) ("{" 1642)) 10296 (("(" 22347) (argument_list 22348)) 10297 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 22349) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 22350) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 22351) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10298 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 22354) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10299 (("," 1480) (possible_comma 22352)) 10300 ((")" 22353)) 10301 (("$" 22331) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 22333) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 22357) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 22358) ("{" 1658)) 10302 (("(" 22347) (argument_list 22359)) 10303 (("$" 22331) (T_STRING 22366) (T_VARIABLE 22333) (property_name 22373) (simple_variable 22368) ("{" 22369)) 10304 (("$" 22331) (T_STRING 22366) (T_VARIABLE 22333) (property_name 22367) (simple_variable 22368) ("{" 22369)) 10305 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 22364) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 10306 (("]" 22365)) 10307 (("(" 22347) (argument_list 22372)) 10308 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 22370) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10309 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 22371)) 10310 (("(" 22347) (argument_list 22374)) 10311 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 22378) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10312 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 22379)) 10313 ((backup_doc_comment 22382)) 10314 (("(" 22383)) 10315 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 22384)) 10316 ((")" 22385)) 10317 ((T_USE 164) (lexical_vars 22386)) 10318 ((":" 166) (return_type 22387)) 10319 ((backup_fn_flags 22388)) 10320 (("{" 22389)) 10321 ((inner_statement_list 22390)) 10322 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 22391) ("~" 127)) 10323 ((backup_fn_flags 22392)) 10324 ((backup_doc_comment 22394)) 10325 (("(" 22395)) 10326 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 22396)) 10327 ((")" 22397)) 10328 ((":" 2249) (return_type 22398)) 10329 ((T_DOUBLE_ARROW 22399)) 10330 ((backup_fn_flags 22400)) 10331 ((backup_lex_pos 22401)) 10332 (("!" 19784) ("\"" 19785) ("$" 19786) ("(" 19787) ("+" 19788) ("-" 19789) ("@" 19790) (T_ARRAY 19791) (T_ARRAY_CAST 19792) (T_ATTRIBUTE 128) (T_BOOL_CAST 19793) (T_CLASS_C 19794) (T_CLONE 19795) (T_CONSTANT_ENCAPSED_STRING 19796) (T_DEC 19797) (T_DIR 19798) (T_DNUMBER 19799) (T_DOUBLE_CAST 19800) (T_EMPTY 19801) (T_EVAL 19802) (T_EXIT 19803) (T_FILE 19804) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19805) (T_INC 19806) (T_INCLUDE 19807) (T_INCLUDE_ONCE 19808) (T_INT_CAST 19809) (T_ISSET 19810) (T_LINE 19811) (T_LIST 19812) (T_LNUMBER 19813) (T_MATCH 19814) (T_METHOD_C 19815) (T_NAME_FULLY_QUALIFIED 19816) (T_NAME_QUALIFIED 19817) (T_NAME_RELATIVE 19818) (T_NEW 19819) (T_NS_C 19820) (T_OBJECT_CAST 19821) (T_PRINT 19822) (T_PROPERTY_C 19823) (T_READONLY 19824) (T_REQUIRE 19825) (T_REQUIRE_ONCE 19826) (T_START_HEREDOC 19827) (T_STATIC 19828) (T_STRING 19829) (T_STRING_CAST 19830) (T_THROW 19831) (T_TRAIT_C 19832) (T_UNSET_CAST 19833) (T_VARIABLE 19834) (T_YIELD 19835) (T_YIELD_FROM 19836) ("[" 19837) ("`" 19838) (array_object_dereferenceable 19839) (attribute 133) (attributes 19840) (callable_expr 19841) (callable_variable 19842) (class_constant 19843) (class_name 19844) (constant 19845) (dereferenceable_scalar 19846) (expr 22402) (fn 19848) (fully_dereferenceable 105) #'19849 (function_call 19850) (inline_function 19851) (internal_functions_in_yacc 19852) (match 19853) (name 19854) (new_dereferenceable 19855) (new_non_dereferenceable 19856) (scalar 19857) (simple_variable 19858) (static_member 19859) (variable 19860) (variable_class_name 19861) ("~" 19862)) 10333 (("%" 19864) ("*" 19865) ("+" 19866) ("-" 19867) ("." 19868) ("/" 19869) ("<" 19870) (">" 19871) ("?" 19872) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19873) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19874) (T_BOOLEAN_AND 19875) (T_BOOLEAN_OR 19876) (T_COALESCE 19877) (T_INSTANCEOF 19878) (T_IS_EQUAL 19879) (T_IS_GREATER_OR_EQUAL 19880) (T_IS_IDENTICAL 19881) (T_IS_NOT_EQUAL 19882) (T_IS_NOT_IDENTICAL 19883) (T_IS_SMALLER_OR_EQUAL 19884) (T_LOGICAL_AND 19885) (T_LOGICAL_OR 19886) (T_LOGICAL_XOR 19887) (T_POW 19888) (T_SL 19889) (T_SPACESHIP 19890) (T_SR 19891) ("^" 19892) (backup_fn_flags 22403) ("|" 19893)) 10334 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22405) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 10335 (("$" 22028) (T_ABSTRACT 22205) (T_ARRAY 22206) (T_AS 22207) (T_BREAK 22208) (T_CALLABLE 22209) (T_CASE 22210) (T_CATCH 22211) (T_CLASS 22212) (T_CLASS_C 22213) (T_CLONE 22214) (T_CONST 22215) (T_CONTINUE 22216) (T_DECLARE 22217) (T_DEFAULT 22218) (T_DIR 22219) (T_DO 22220) (T_ECHO 22221) (T_ELSE 22222) (T_ELSEIF 22223) (T_EMPTY 22224) (T_ENDDECLARE 22225) (T_ENDFOR 22226) (T_ENDFOREACH 22227) (T_ENDIF 22228) (T_ENDSWITCH 22229) (T_ENDWHILE 22230) (T_ENUM 22231) (T_EVAL 22232) (T_EXIT 22233) (T_EXTENDS 22234) (T_FILE 22235) (T_FINAL 22236) (T_FINALLY 22237) (T_FN 22238) (T_FOR 22239) (T_FOREACH 22240) (T_FUNCTION 22241) (T_FUNC_C 22242) (T_GLOBAL 22243) (T_GOTO 22244) (T_IF 22245) (T_IMPLEMENTS 22246) (T_INCLUDE 22247) (T_INCLUDE_ONCE 22248) (T_INSTANCEOF 22249) (T_INSTEADOF 22250) (T_INTERFACE 22251) (T_ISSET 22252) (T_LINE 22253) (T_LIST 22254) (T_LOGICAL_AND 22255) (T_LOGICAL_OR 22256) (T_LOGICAL_XOR 22257) (T_MATCH 22258) (T_METHOD_C 22259) (T_NAMESPACE 22260) (T_NEW 22261) (T_NS_C 22262) (T_PRINT 22263) (T_PRIVATE 22264) (T_PROPERTY_C 22265) (T_PROTECTED 22266) (T_PUBLIC 22267) (T_READONLY 22268) (T_REQUIRE 22269) (T_REQUIRE_ONCE 22270) (T_RETURN 22271) (T_STATIC 22272) (T_STRING 22273) (T_SWITCH 22274) (T_THROW 22275) (T_TRAIT 22276) (T_TRAIT_C 22277) (T_TRY 22278) (T_UNSET 22279) (T_USE 22280) (T_VAR 22281) (T_VARIABLE 22076) (T_WHILE 22282) (T_YIELD 22283) (identifier 22407) (member_name 22408) (reserved_non_modifiers 22286) (semi_reserved 22287) (simple_variable 22409) ("{" 22410)) 10336 (("(" 22292) (argument_list 22413)) 10337 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 22411) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10338 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 22412)) 10339 ((T_FN 35) (T_FUNCTION 129) (fn 22090) #'22091 (inline_function 22417)) 10340 (("$" 22028) (T_STRING 22423) (T_VARIABLE 22076) (property_name 22430) (simple_variable 22425) ("{" 22426)) 10341 (("$" 22028) (T_STRING 22423) (T_VARIABLE 22076) (property_name 22424) (simple_variable 22425) ("{" 22426)) 10342 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 22421) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 10343 (("]" 22422)) 10344 (("(" 22292) (argument_list 22429)) 10345 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 22427) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10346 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 22428)) 10347 (("(" 22292) (argument_list 22431)) 10348 (("`" 22433)) 10349 (("]" 22435)) 10350 (("=" 22436)) 10351 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22437) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10352 (("%" 22106) ("*" 22107) ("+" 22108) ("-" 22109) ("." 22110) ("/" 22111) ("<" 22112) (">" 22113) ("?" 22114) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22115) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22116) (T_BOOLEAN_AND 22117) (T_BOOLEAN_OR 22118) (T_COALESCE 22119) (T_DOUBLE_ARROW 22440) (T_INSTANCEOF 22120) (T_IS_EQUAL 22121) (T_IS_GREATER_OR_EQUAL 22122) (T_IS_IDENTICAL 22123) (T_IS_NOT_EQUAL 22124) (T_IS_NOT_IDENTICAL 22125) (T_IS_SMALLER_OR_EQUAL 22126) (T_LOGICAL_AND 22127) (T_LOGICAL_OR 22128) (T_LOGICAL_XOR 22129) (T_POW 22130) (T_SL 22131) (T_SPACESHIP 22132) (T_SR 22133) ("^" 22134) ("|" 22135)) 10353 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22441) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10354 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 22450) (T_VARIABLE 2896) (encaps_var 2910)) 10355 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 22449) (T_VARIABLE 2896) (encaps_var 2901)) 10356 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22491) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10357 ((T_CLASS 22483)) 10358 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 22482) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 22463) (attribute 2958) (class_modifier 2939)) 10359 ((T_PAAMAYIM_NEKUDOTAYIM 22480)) 10360 (("(" 22347) (argument_list 22479)) 10361 ((T_NULLSAFE_OBJECT_OPERATOR 22470) (T_OBJECT_OPERATOR 22471) (T_PAAMAYIM_NEKUDOTAYIM 22472) ("[" 22473)) 10362 (("$" 22331) (T_STRING 22366) (T_VARIABLE 22333) (property_name 22478) (simple_variable 22368) ("{" 22369)) 10363 (("$" 22331) (T_STRING 22366) (T_VARIABLE 22333) (property_name 22477) (simple_variable 22368) ("{" 22369)) 10364 (("$" 22331) (T_VARIABLE 22333) (simple_variable 22476)) 10365 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 22474) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 10366 (("]" 22475)) 10367 (("$" 22331) (T_VARIABLE 22333) (simple_variable 22481)) 10368 (("(" 2960) (argument_list 2961) (ctor_arguments 22484)) 10369 ((T_EXTENDS 2963) (extends_from 22485)) 10370 ((T_IMPLEMENTS 2965) (implements_list 22486)) 10371 ((backup_doc_comment 22487)) 10372 (("{" 22488)) 10373 ((class_statement_list 22489)) 10374 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 22490)) 10375 (("%" 299) (")" 22492) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10376 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22494) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10377 (("%" 299) (")" 22495) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10378 (("{" 22496)) 10379 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 22497) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 10380 (("}" 22498)) 10381 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 22500) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 10382 ((")" 22501)) 10383 (("=" 22502)) 10384 (("!" 22026) ("\"" 22027) ("$" 22028) ("(" 22029) ("+" 22030) ("-" 22031) ("@" 22032) (T_ARRAY 22033) (T_ARRAY_CAST 22034) (T_ATTRIBUTE 128) (T_BOOL_CAST 22035) (T_CLASS_C 22036) (T_CLONE 22037) (T_CONSTANT_ENCAPSED_STRING 22038) (T_DEC 22039) (T_DIR 22040) (T_DNUMBER 22041) (T_DOUBLE_CAST 22042) (T_EMPTY 22043) (T_EVAL 22044) (T_EXIT 22045) (T_FILE 22046) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22047) (T_INC 22048) (T_INCLUDE 22049) (T_INCLUDE_ONCE 22050) (T_INT_CAST 22051) (T_ISSET 22052) (T_LINE 22053) (T_LIST 22054) (T_LNUMBER 22055) (T_MATCH 22056) (T_METHOD_C 22057) (T_NAME_FULLY_QUALIFIED 22058) (T_NAME_QUALIFIED 22059) (T_NAME_RELATIVE 22060) (T_NEW 22061) (T_NS_C 22062) (T_OBJECT_CAST 22063) (T_PRINT 22064) (T_PROPERTY_C 22065) (T_READONLY 22066) (T_REQUIRE 22067) (T_REQUIRE_ONCE 22068) (T_START_HEREDOC 22069) (T_STATIC 22070) (T_STRING 22071) (T_STRING_CAST 22072) (T_THROW 22073) (T_TRAIT_C 22074) (T_UNSET_CAST 22075) (T_VARIABLE 22076) (T_YIELD 22077) (T_YIELD_FROM 22078) ("[" 22079) ("`" 22080) (array_object_dereferenceable 22081) (attribute 133) (attributes 22082) (callable_expr 22083) (callable_variable 22084) (class_constant 22085) (class_name 22086) (constant 22087) (dereferenceable_scalar 22088) (expr 22503) (fn 22090) (fully_dereferenceable 105) #'22091 (function_call 22092) (inline_function 22093) (internal_functions_in_yacc 22094) (match 22095) (name 22096) (new_dereferenceable 22097) (new_non_dereferenceable 22098) (scalar 22099) (simple_variable 22100) (static_member 22101) (variable 22102) (variable_class_name 22103) ("~" 22104)) 10385 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 22505) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10386 (("," 5192) (possible_comma 22506)) 10387 ((")" 22507)) 10388 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 22515) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 22516) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 22517) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10389 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 22520) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10390 (("," 1480) (possible_comma 22518)) 10391 ((")" 22519)) 10392 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22522) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10393 (("%" 299) (")" 22523) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10394 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22525) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10395 (("%" 299) (")" 22526) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10396 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 22533) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 10397 ((")" 22534)) 10398 (("%" 299) (")" 22539) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10399 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 22542) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10400 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 22543)) 10401 (("\"" 22545) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 10402 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 22555) (T_VARIABLE 2896) (encaps_var 2910)) 10403 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 22554) (T_VARIABLE 2896) (encaps_var 2901)) 10404 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22596) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10405 ((T_CLASS 22588)) 10406 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 22587) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 22568) (attribute 2958) (class_modifier 2939)) 10407 ((T_PAAMAYIM_NEKUDOTAYIM 22585)) 10408 (("(" 19317) (argument_list 22584)) 10409 ((T_NULLSAFE_OBJECT_OPERATOR 22575) (T_OBJECT_OPERATOR 22576) (T_PAAMAYIM_NEKUDOTAYIM 22577) ("[" 22578)) 10410 (("$" 19301) (T_STRING 19336) (T_VARIABLE 19303) (property_name 22583) (simple_variable 19338) ("{" 19339)) 10411 (("$" 19301) (T_STRING 19336) (T_VARIABLE 19303) (property_name 22582) (simple_variable 19338) ("{" 19339)) 10412 (("$" 19301) (T_VARIABLE 19303) (simple_variable 22581)) 10413 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 22579) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 10414 (("]" 22580)) 10415 (("$" 19301) (T_VARIABLE 19303) (simple_variable 22586)) 10416 (("(" 2960) (argument_list 2961) (ctor_arguments 22589)) 10417 ((T_EXTENDS 2963) (extends_from 22590)) 10418 ((T_IMPLEMENTS 2965) (implements_list 22591)) 10419 ((backup_doc_comment 22592)) 10420 (("{" 22593)) 10421 ((class_statement_list 22594)) 10422 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 22595)) 10423 (("%" 299) (")" 22597) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10424 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22599) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10425 (("%" 299) (")" 22600) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10426 (("{" 22601)) 10427 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 22602) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 10428 (("}" 22603)) 10429 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 22605) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 10430 ((")" 22606)) 10431 (("=" 22607)) 10432 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22608) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 10433 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 22610) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10434 (("," 5192) (possible_comma 22611)) 10435 ((")" 22612)) 10436 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 22620) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 22621) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 22622) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10437 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 22625) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10438 (("," 1480) (possible_comma 22623)) 10439 ((")" 22624)) 10440 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22627) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10441 (("%" 299) (")" 22628) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10442 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22630) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10443 (("%" 299) (")" 22631) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10444 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 22638) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 10445 ((")" 22639)) 10446 (("%" 299) (")" 22644) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10447 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 22647) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10448 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 22648)) 10449 (("\"" 22650) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 10450 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 22653) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 10451 ((";" 22654)) 10452 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 22655) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 10453 ((";" 22656)) 10454 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 22657) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 10455 ((")" 22658)) 10456 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 22659) (";" 17846) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17847) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17848) (T_DEC 22) (T_DECLARE 17849) (T_DIR 24) (T_DNUMBER 25) (T_DO 17850) (T_DOUBLE_CAST 27) (T_ECHO 17851) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17852) (T_FOREACH 17853) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17854) (T_GOTO 17855) (T_IF 17811) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17856) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17857) (T_START_HEREDOC 69) (T_STATIC 17858) (T_STRING 17859) (T_STRING_CAST 72) (T_SWITCH 17860) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17861) (T_UNSET 17862) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17863) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17864) (alt_if_stmt_without_else 17865) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17866) (fn 104) (for_statement 22660) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17867) (if_stmt_without_else 17868) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 22661) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17870) ("~" 127)) 10457 ((inner_statement_list 22662)) 10458 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 22663) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 10459 ((";" 22664)) 10460 (("," 20829) (";" 22666)) 10461 ((T_WHILE 22668)) 10462 (("(" 22669)) 10463 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22670) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10464 (("%" 299) (")" 22671) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10465 ((";" 22672)) 10466 ((T_STRING 21002) (const_decl 21003) (const_list 22674)) 10467 ((")" 22675) ("," 21006)) 10468 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 22676) (";" 17846) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17847) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17848) (T_DEC 22) (T_DECLARE 17849) (T_DIR 24) (T_DNUMBER 25) (T_DO 17850) (T_DOUBLE_CAST 27) (T_ECHO 17851) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17852) (T_FOREACH 17853) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17854) (T_GOTO 17855) (T_IF 17811) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17856) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17857) (T_START_HEREDOC 69) (T_STATIC 17858) (T_STRING 17859) (T_STRING_CAST 72) (T_SWITCH 17860) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17861) (T_UNSET 17862) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17863) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17864) (alt_if_stmt_without_else 17865) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 22677) (dereferenceable_scalar 101) (expr 17866) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17867) (if_stmt_without_else 17868) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 22678) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17870) ("~" 127)) 10469 ((inner_statement_list 22679)) 10470 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 22680) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 10471 ((";" 22681)) 10472 ((";" 22683)) 10473 ((";" 22685)) 10474 ((T_READONLY 17353) (T_STRING 17354) (backup_doc_comment 141) (function_name 22688)) 10475 ((backup_doc_comment 22689)) 10476 (("(" 22690)) 10477 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 22691)) 10478 ((")" 22692)) 10479 ((":" 166) (return_type 22693)) 10480 ((backup_fn_flags 22694)) 10481 (("{" 22695)) 10482 ((inner_statement_list 22696)) 10483 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 22697) ("~" 127)) 10484 ((backup_fn_flags 22698)) 10485 ((T_STRING 22701)) 10486 ((T_EXTENDS 2963) (extends_from 22702)) 10487 ((T_IMPLEMENTS 2965) (implements_list 22703)) 10488 ((backup_doc_comment 22704)) 10489 (("{" 22705)) 10490 ((class_statement_list 22706)) 10491 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 22707)) 10492 ((":" 22712)) 10493 ((";" 22711)) 10494 ((inner_statement_list 22713)) 10495 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 22714) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 10496 ((";" 22715)) 10497 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22717) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10498 (("%" 299) (")" 22718) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10499 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 22719) (";" 17798) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17799) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17801) (T_DEC 22) (T_DECLARE 17802) (T_DIR 24) (T_DNUMBER 25) (T_DO 17803) (T_DOUBLE_CAST 27) (T_ECHO 17804) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17806) (T_FOREACH 17807) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17808) (T_GOTO 17809) (T_IF 17811) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17812) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17814) (T_START_HEREDOC 69) (T_STATIC 17815) (T_STRING 17816) (T_STRING_CAST 72) (T_SWITCH 17817) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17819) (T_UNSET 17820) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17821) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17822) (alt_if_stmt_without_else 17823) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17829) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17832) (if_stmt_without_else 17833) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 22720) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 22721) ("{" 17838) ("~" 127)) 10500 ((inner_statement_list 22722)) 10501 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 22723) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 10502 ((";" 22724)) 10503 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 22726) (variable 17576) (variable_class_name 17577)) 10504 (("," 17589) (possible_comma 22727)) 10505 ((")" 22728)) 10506 ((";" 22729)) 10507 ((inner_statement_list 22731)) 10508 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 22732) ("~" 127)) 10509 ((catch_list 22733)) 10510 ((T_CATCH 22734) (T_FINALLY 22735) (finally_statement 22736)) 10511 (("(" 22740)) 10512 (("{" 22737)) 10513 ((inner_statement_list 22738)) 10514 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 22739) ("~" 127)) 10515 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 22741) (class_name 17636) (name 17637)) 10516 ((T_VARIABLE 17638) (optional_variable 22742) ("|" 17640)) 10517 ((")" 22743)) 10518 (("{" 22744)) 10519 ((inner_statement_list 22745)) 10520 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 22746) ("~" 127)) 10521 ((backup_doc_comment 22748)) 10522 (("{" 22749)) 10523 ((class_statement_list 22750)) 10524 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 22751)) 10525 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22753) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10526 (("%" 299) (")" 22754) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10527 ((":" 22755) (switch_case_list 22756) ("{" 22757)) 10528 ((";" 22763) (case_list 22764)) 10529 ((";" 22758) (case_list 22759)) 10530 ((case_list 22761)) 10531 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 22760)) 10532 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 22762)) 10533 ((case_list 22767)) 10534 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 22765)) 10535 ((";" 22766)) 10536 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 22768)) 10537 ((";" 22769)) 10538 (("," 17937) (";" 22772)) 10539 ((";" 22774)) 10540 ((T_EXTENDS 21295) (interface_extends_list 22776)) 10541 ((backup_doc_comment 22777)) 10542 (("{" 22778)) 10543 ((class_statement_list 22779)) 10544 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 22780)) 10545 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22782) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10546 (("%" 299) (")" 22783) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10547 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21305) (";" 17846) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17847) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17848) (T_DEC 22) (T_DECLARE 17849) (T_DIR 24) (T_DNUMBER 25) (T_DO 17850) (T_DOUBLE_CAST 27) (T_ECHO 17851) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17852) (T_FOREACH 17853) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17854) (T_GOTO 17855) (T_IF 17811) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17856) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17857) (T_START_HEREDOC 69) (T_STATIC 17858) (T_STRING 17859) (T_STRING_CAST 72) (T_SWITCH 17860) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17861) (T_UNSET 17862) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17863) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17864) (alt_if_stmt_without_else 17865) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17866) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17867) (if_stmt_without_else 17868) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 22784) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17870) ("~" 127)) 10548 ((")" 22786)) 10549 ((";" 22787)) 10550 ((";" 22789)) 10551 (("," 18988) (";" 22791)) 10552 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 22793) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 10553 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 22794) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 10554 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 22795) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 10555 ((")" 22796) (T_DOUBLE_ARROW 22797)) 10556 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 22800) (";" 17798) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17799) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17801) (T_DEC 22) (T_DECLARE 17802) (T_DIR 24) (T_DNUMBER 25) (T_DO 17803) (T_DOUBLE_CAST 27) (T_ECHO 17804) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17806) (T_FOREACH 17807) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17808) (T_GOTO 17809) (T_IF 17811) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17812) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17814) (T_START_HEREDOC 69) (T_STATIC 17815) (T_STRING 17816) (T_STRING_CAST 72) (T_SWITCH 17817) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17819) (T_UNSET 17820) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17821) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17822) (alt_if_stmt_without_else 17823) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17829) (fn 104) (foreach_statement 22806) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17832) (if_stmt_without_else 17833) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 22802) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17838) ("~" 127)) 10557 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 22798) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 10558 ((")" 22799)) 10559 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 22800) (";" 17798) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17799) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17801) (T_DEC 22) (T_DECLARE 17802) (T_DIR 24) (T_DNUMBER 25) (T_DO 17803) (T_DOUBLE_CAST 27) (T_ECHO 17804) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17806) (T_FOREACH 17807) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17808) (T_GOTO 17809) (T_IF 17811) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17812) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17814) (T_START_HEREDOC 69) (T_STATIC 17815) (T_STRING 17816) (T_STRING_CAST 72) (T_SWITCH 17817) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17819) (T_UNSET 17820) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17821) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17822) (alt_if_stmt_without_else 17823) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17829) (fn 104) (foreach_statement 22801) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17832) (if_stmt_without_else 17833) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 22802) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17838) ("~" 127)) 10560 ((inner_statement_list 22803)) 10561 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 22804) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 10562 ((";" 22805)) 10563 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 22808) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 10564 ((";" 22809)) 10565 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 22810) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 10566 ((";" 22811)) 10567 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 22812) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 10568 ((")" 22813)) 10569 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 22814) (";" 17798) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17799) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17801) (T_DEC 22) (T_DECLARE 17802) (T_DIR 24) (T_DNUMBER 25) (T_DO 17803) (T_DOUBLE_CAST 27) (T_ECHO 17804) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17806) (T_FOREACH 17807) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17808) (T_GOTO 17809) (T_IF 17811) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17812) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17814) (T_START_HEREDOC 69) (T_STATIC 17815) (T_STRING 17816) (T_STRING_CAST 72) (T_SWITCH 17817) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17819) (T_UNSET 17820) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17821) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17822) (alt_if_stmt_without_else 17823) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17829) (fn 104) (for_statement 22815) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17832) (if_stmt_without_else 17833) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 22816) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17838) ("~" 127)) 10570 ((inner_statement_list 22817)) 10571 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 22818) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 10572 ((";" 22819)) 10573 ((":" 21344) (enum_backing_type 22821)) 10574 ((T_IMPLEMENTS 2965) (implements_list 22822)) 10575 ((backup_doc_comment 22823)) 10576 (("{" 22824)) 10577 ((class_statement_list 22825)) 10578 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 22826)) 10579 (("," 20829) (";" 22828)) 10580 ((T_WHILE 22830)) 10581 (("(" 22831)) 10582 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 22832) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10583 (("%" 299) (")" 22833) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10584 ((";" 22834)) 10585 ((T_STRING 21002) (const_decl 21003) (const_list 22836)) 10586 ((")" 22837) ("," 21006)) 10587 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 22838) (";" 17798) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17799) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17801) (T_DEC 22) (T_DECLARE 17802) (T_DIR 24) (T_DNUMBER 25) (T_DO 17803) (T_DOUBLE_CAST 27) (T_ECHO 17804) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17806) (T_FOREACH 17807) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17808) (T_GOTO 17809) (T_IF 17811) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17812) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17814) (T_START_HEREDOC 69) (T_STATIC 17815) (T_STRING 17816) (T_STRING_CAST 72) (T_SWITCH 17817) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17819) (T_UNSET 17820) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17821) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17822) (alt_if_stmt_without_else 17823) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 22839) (dereferenceable_scalar 101) (expr 17829) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17832) (if_stmt_without_else 17833) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 22840) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17838) ("~" 127)) 10588 ((inner_statement_list 22841)) 10589 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 22842) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 10590 ((";" 22843)) 10591 ((";" 22845)) 10592 ((T_EXTENDS 2963) (extends_from 22847)) 10593 ((T_IMPLEMENTS 2965) (implements_list 22848)) 10594 ((backup_doc_comment 22849)) 10595 (("{" 22850)) 10596 ((class_statement_list 22851)) 10597 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 22852)) 10598 ((";" 22854)) 10599 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24929) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10600 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 24927) (encaps_var 5237)) 10601 (("$" 22857) (T_VARIABLE 22905) (simple_variable 24923) ("{" 24924)) 10602 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24921) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10603 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24920) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10604 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24919) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10605 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24918) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10606 (("(" 24915)) 10607 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24914) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10608 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24913) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10609 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24912) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10610 (("\"" 1520) ("$" 23160) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 23161) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 23162) ("[" 1542) (array_object_dereferenceable 23163) (callable_expr 23164) (callable_variable 23165) (class_constant 1546) (class_name 23166) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 23167) (name 23168) (new_dereferenceable 1552) (simple_variable 23169) (static_member 23170) (variable 24911) (variable_class_name 23172)) 10611 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24910) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10612 (("(" 24907)) 10613 (("(" 24904)) 10614 (("(" 24895) (argument_list 24896) (ctor_arguments 24897)) 10615 (("\"" 1520) ("$" 23160) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 23161) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 23162) ("[" 1542) (array_object_dereferenceable 23163) (callable_expr 23164) (callable_variable 23165) (class_constant 1546) (class_name 23166) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 23167) (name 23168) (new_dereferenceable 1552) (simple_variable 23169) (static_member 23170) (variable 24894) (variable_class_name 23172)) 10616 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24893) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10617 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24892) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10618 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24891) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10619 (("(" 24887)) 10620 (("(" 24882)) 10621 (("(" 24876)) 10622 (("$" 23160) ("(" 24839) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 24840) (T_NAME_QUALIFIED 24841) (T_NAME_RELATIVE 24842) (T_READONLY 2931) (T_STATIC 24843) (T_STRING 24844) (T_VARIABLE 23162) (anonymous_class 24845) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 24846) (attribute 2937) (attributes 24847) (class_modifier 2939) (class_name 24848) (class_name_reference 24849) (name 24850) (new_variable 24851) (simple_variable 24852)) 10623 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24838) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10624 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24837) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10625 (("(" 23121) (argument_list 24836)) 10626 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24835) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10627 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24834) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10628 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 24829) (T_END_HEREDOC 24830) (T_VARIABLE 2896) (encaps_list 24831) (encaps_var 2898)) 10629 ((T_FN 35) (T_FUNCTION 129) (fn 22919) #'22920 (inline_function 24828)) 10630 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24827) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10631 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24826) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10632 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24825) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10633 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24367) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 10634 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24303) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10635 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 24299) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 10636 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 24297) (encaps_list 2804) (encaps_var 2805)) 10637 ((T_NULLSAFE_OBJECT_OPERATOR 24283) (T_OBJECT_OPERATOR 24284) ("[" 24285)) 10638 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 24280) (attribute 2733) (fn 22919) #'22920 (inline_function 24281)) 10639 (("(" 23121) (argument_list 24279)) 10640 ((T_PAAMAYIM_NEKUDOTAYIM 24271)) 10641 (("%" 22935) ("*" 22936) ("+" 22937) ("-" 22938) ("." 22939) ("/" 22940) (":" 17794) (";" 17795) ("<" 22941) (">" 22942) ("?" 22943) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22944) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22945) (T_BOOLEAN_AND 22946) (T_BOOLEAN_OR 22947) (T_COALESCE 22948) (T_INSTANCEOF 22949) (T_IS_EQUAL 22950) (T_IS_GREATER_OR_EQUAL 22951) (T_IS_IDENTICAL 22952) (T_IS_NOT_EQUAL 22953) (T_IS_NOT_IDENTICAL 22954) (T_IS_SMALLER_OR_EQUAL 22955) (T_LOGICAL_AND 22956) (T_LOGICAL_OR 22957) (T_LOGICAL_XOR 22958) (T_POW 22959) (T_SL 22960) (T_SPACESHIP 22961) (T_SR 22962) ("^" 22963) (case_separator 24269) ("|" 22964)) 10642 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 23222)) 10643 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 23210)) 10644 (("(" 23121) (argument_list 23209)) 10645 (("=" 23129) (T_AND_EQUAL 23130) (T_COALESCE_EQUAL 23131) (T_CONCAT_EQUAL 23132) (T_DEC 23133) (T_DIV_EQUAL 23134) (T_INC 23135) (T_MINUS_EQUAL 23136) (T_MOD_EQUAL 23137) (T_MUL_EQUAL 23138) (T_OR_EQUAL 23139) (T_PLUS_EQUAL 23140) (T_POW_EQUAL 23141) (T_SL_EQUAL 23142) (T_SR_EQUAL 23143) (T_XOR_EQUAL 23144)) 10646 ((T_PAAMAYIM_NEKUDOTAYIM 23033)) 10647 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22934) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10648 (("%" 22935) ("*" 22936) ("+" 22937) ("-" 22938) ("." 22939) ("/" 22940) ("<" 22941) (">" 22942) ("?" 22943) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22944) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22945) (T_BOOLEAN_AND 22946) (T_BOOLEAN_OR 22947) (T_COALESCE 22948) (T_INSTANCEOF 22949) (T_IS_EQUAL 22950) (T_IS_GREATER_OR_EQUAL 22951) (T_IS_IDENTICAL 22952) (T_IS_NOT_EQUAL 22953) (T_IS_NOT_IDENTICAL 22954) (T_IS_SMALLER_OR_EQUAL 22955) (T_LOGICAL_AND 22956) (T_LOGICAL_OR 22957) (T_LOGICAL_XOR 22958) (T_POW 22959) (T_SL 22960) (T_SPACESHIP 22961) (T_SR 22962) ("^" 22963) ("|" 22964)) 10649 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23032) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10650 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23031) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10651 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23030) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10652 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23029) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10653 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23028) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10654 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23027) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10655 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23026) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10656 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23025) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10657 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 23020) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 23021) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 10658 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23019) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10659 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23018) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10660 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23017) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10661 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23016) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10662 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23015) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10663 (("$" 22980) ("(" 22981) (T_NAME_FULLY_QUALIFIED 22982) (T_NAME_QUALIFIED 22983) (T_NAME_RELATIVE 22984) (T_STATIC 22985) (T_STRING 22986) (T_VARIABLE 22987) (class_name 22988) (class_name_reference 22989) (name 22990) (new_variable 22991) (simple_variable 22992)) 10664 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22979) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10665 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22978) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10666 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22977) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10667 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22976) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10668 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22975) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10669 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22974) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10670 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22973) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10671 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22972) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10672 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22971) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10673 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22970) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10674 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22969) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10675 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22968) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10676 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22967) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10677 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22966) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10678 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 22965) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10679 (("$" 22980) (T_VARIABLE 22987) (simple_variable 23011) ("{" 23012)) 10680 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 23009) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10681 ((T_PAAMAYIM_NEKUDOTAYIM 23007)) 10682 ((T_NULLSAFE_OBJECT_OPERATOR 22993) (T_OBJECT_OPERATOR 22994) (T_PAAMAYIM_NEKUDOTAYIM 22995) ("[" 22996)) 10683 (("$" 22980) (T_STRING 23000) (T_VARIABLE 22987) (property_name 23006) (simple_variable 23002) ("{" 23003)) 10684 (("$" 22980) (T_STRING 23000) (T_VARIABLE 22987) (property_name 23001) (simple_variable 23002) ("{" 23003)) 10685 (("$" 22980) (T_VARIABLE 22987) (simple_variable 22999)) 10686 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 22997) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 10687 (("]" 22998)) 10688 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23004) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10689 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23005)) 10690 (("$" 22980) (T_VARIABLE 22987) (simple_variable 23008)) 10691 (("%" 299) (")" 23010) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10692 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23013) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10693 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23014)) 10694 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23024) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10695 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 23022) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 10696 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23023) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10697 (("$" 22857) (T_ABSTRACT 23034) (T_ARRAY 23035) (T_AS 23036) (T_BREAK 23037) (T_CALLABLE 23038) (T_CASE 23039) (T_CATCH 23040) (T_CLASS 23041) (T_CLASS_C 23042) (T_CLONE 23043) (T_CONST 23044) (T_CONTINUE 23045) (T_DECLARE 23046) (T_DEFAULT 23047) (T_DIR 23048) (T_DO 23049) (T_ECHO 23050) (T_ELSE 23051) (T_ELSEIF 23052) (T_EMPTY 23053) (T_ENDDECLARE 23054) (T_ENDFOR 23055) (T_ENDFOREACH 23056) (T_ENDIF 23057) (T_ENDSWITCH 23058) (T_ENDWHILE 23059) (T_ENUM 23060) (T_EVAL 23061) (T_EXIT 23062) (T_EXTENDS 23063) (T_FILE 23064) (T_FINAL 23065) (T_FINALLY 23066) (T_FN 23067) (T_FOR 23068) (T_FOREACH 23069) (T_FUNCTION 23070) (T_FUNC_C 23071) (T_GLOBAL 23072) (T_GOTO 23073) (T_IF 23074) (T_IMPLEMENTS 23075) (T_INCLUDE 23076) (T_INCLUDE_ONCE 23077) (T_INSTANCEOF 23078) (T_INSTEADOF 23079) (T_INTERFACE 23080) (T_ISSET 23081) (T_LINE 23082) (T_LIST 23083) (T_LOGICAL_AND 23084) (T_LOGICAL_OR 23085) (T_LOGICAL_XOR 23086) (T_MATCH 23087) (T_METHOD_C 23088) (T_NAMESPACE 23089) (T_NEW 23090) (T_NS_C 23091) (T_PRINT 23092) (T_PRIVATE 23093) (T_PROPERTY_C 23094) (T_PROTECTED 23095) (T_PUBLIC 23096) (T_READONLY 23097) (T_REQUIRE 23098) (T_REQUIRE_ONCE 23099) (T_RETURN 23100) (T_STATIC 23101) (T_STRING 23102) (T_SWITCH 23103) (T_THROW 23104) (T_TRAIT 23105) (T_TRAIT_C 23106) (T_TRY 23107) (T_UNSET 23108) (T_USE 23109) (T_VAR 23110) (T_VARIABLE 22905) (T_WHILE 23111) (T_YIELD 23112) (identifier 23113) (member_name 23114) (reserved_non_modifiers 23115) (semi_reserved 23116) (simple_variable 23117) ("{" 23118)) 10698 (("(" 23121) (argument_list 23122)) 10699 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23119) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10700 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23120)) 10701 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 23123) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 23124) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 23125) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10702 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 23128) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10703 (("," 1480) (possible_comma 23126)) 10704 ((")" 23127)) 10705 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (ampersand 23158) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23159) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10706 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23157) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10707 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23156) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10708 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23155) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10709 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23154) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10710 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23153) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10711 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23152) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10712 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23151) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10713 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23150) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10714 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23149) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10715 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23148) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10716 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23147) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10717 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23146) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10718 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 23145) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 10719 (("\"" 1520) ("$" 23160) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 23161) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 23162) ("[" 1542) (array_object_dereferenceable 23163) (callable_expr 23164) (callable_variable 23165) (class_constant 1546) (class_name 23166) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 23167) (name 23168) (new_dereferenceable 1552) (simple_variable 23169) (static_member 23170) (variable 23171) (variable_class_name 23172)) 10720 (("$" 23160) (T_VARIABLE 23162) (simple_variable 23205) ("{" 23206)) 10721 (("(" 23176) (argument_list 23204)) 10722 ((T_NULLSAFE_OBJECT_OPERATOR 23190) (T_OBJECT_OPERATOR 23191) ("[" 23192)) 10723 (("(" 23176) (argument_list 23189)) 10724 ((T_PAAMAYIM_NEKUDOTAYIM 23185)) 10725 (("(" 23176) (argument_list 23184)) 10726 ((T_PAAMAYIM_NEKUDOTAYIM 23173)) 10727 (("$" 23160) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 23162) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 23174) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 23175) ("{" 1642)) 10728 (("(" 23176) (argument_list 23177)) 10729 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 23178) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 23179) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 23180) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10730 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 23183) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10731 (("," 1480) (possible_comma 23181)) 10732 ((")" 23182)) 10733 (("$" 23160) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 23162) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 23186) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 23187) ("{" 1658)) 10734 (("(" 23176) (argument_list 23188)) 10735 (("$" 23160) (T_STRING 23195) (T_VARIABLE 23162) (property_name 23202) (simple_variable 23197) ("{" 23198)) 10736 (("$" 23160) (T_STRING 23195) (T_VARIABLE 23162) (property_name 23196) (simple_variable 23197) ("{" 23198)) 10737 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 23193) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 10738 (("]" 23194)) 10739 (("(" 23176) (argument_list 23201)) 10740 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23199) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10741 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23200)) 10742 (("(" 23176) (argument_list 23203)) 10743 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23207) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10744 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23208)) 10745 ((backup_doc_comment 23211)) 10746 (("(" 23212)) 10747 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 23213)) 10748 ((")" 23214)) 10749 ((T_USE 164) (lexical_vars 23215)) 10750 ((":" 166) (return_type 23216)) 10751 ((backup_fn_flags 23217)) 10752 (("{" 23218)) 10753 ((inner_statement_list 23219)) 10754 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 23220) ("~" 127)) 10755 ((backup_fn_flags 23221)) 10756 ((backup_doc_comment 23223)) 10757 (("(" 23224)) 10758 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 23225)) 10759 ((")" 23226)) 10760 ((":" 2249) (return_type 23227)) 10761 ((T_DOUBLE_ARROW 23228)) 10762 ((backup_fn_flags 23229)) 10763 ((backup_lex_pos 23230)) 10764 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23294) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10765 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24268) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10766 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 24266) (encaps_var 5237)) 10767 (("$" 23233) (T_VARIABLE 23281) (simple_variable 24262) ("{" 24263)) 10768 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24260) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10769 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24259) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10770 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24258) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10771 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24257) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10772 (("(" 24254)) 10773 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24253) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10774 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24252) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10775 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24251) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10776 (("\"" 1520) ("$" 23536) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 23537) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 23538) ("[" 1542) (array_object_dereferenceable 23539) (callable_expr 23540) (callable_variable 23541) (class_constant 1546) (class_name 23542) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 23543) (name 23544) (new_dereferenceable 1552) (simple_variable 23545) (static_member 23546) (variable 24250) (variable_class_name 23548)) 10777 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24249) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10778 (("(" 24246)) 10779 (("(" 24243)) 10780 (("(" 24234) (argument_list 24235) (ctor_arguments 24236)) 10781 (("\"" 1520) ("$" 23536) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 23537) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 23538) ("[" 1542) (array_object_dereferenceable 23539) (callable_expr 23540) (callable_variable 23541) (class_constant 1546) (class_name 23542) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 23543) (name 23544) (new_dereferenceable 1552) (simple_variable 23545) (static_member 23546) (variable 24233) (variable_class_name 23548)) 10782 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24232) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10783 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24231) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10784 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24230) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10785 (("(" 24226)) 10786 (("(" 24221)) 10787 (("(" 24215)) 10788 (("$" 23536) ("(" 24178) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 24179) (T_NAME_QUALIFIED 24180) (T_NAME_RELATIVE 24181) (T_READONLY 2931) (T_STATIC 24182) (T_STRING 24183) (T_VARIABLE 23538) (anonymous_class 24184) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 24185) (attribute 2937) (attributes 24186) (class_modifier 2939) (class_name 24187) (class_name_reference 24188) (name 24189) (new_variable 24190) (simple_variable 24191)) 10789 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24177) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10790 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24176) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10791 (("(" 23497) (argument_list 24175)) 10792 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24174) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10793 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24173) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10794 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 24168) (T_END_HEREDOC 24169) (T_VARIABLE 2896) (encaps_list 24170) (encaps_var 2898)) 10795 ((T_FN 35) (T_FUNCTION 129) (fn 23295) #'23296 (inline_function 24167)) 10796 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24166) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10797 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24165) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10798 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24164) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10799 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23706) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10800 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23642) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10801 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 23638) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 10802 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 23636) (encaps_list 2804) (encaps_var 2805)) 10803 ((T_NULLSAFE_OBJECT_OPERATOR 23622) (T_OBJECT_OPERATOR 23623) ("[" 23624)) 10804 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 23619) (attribute 2733) (fn 23295) #'23296 (inline_function 23620)) 10805 (("(" 23497) (argument_list 23618)) 10806 ((T_PAAMAYIM_NEKUDOTAYIM 23610)) 10807 (("%" 23311) ("*" 23312) ("+" 23313) ("-" 23314) ("." 23315) ("/" 23316) ("<" 23317) (">" 23318) ("?" 23319) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 23320) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 23321) (T_BOOLEAN_AND 23322) (T_BOOLEAN_OR 23323) (T_COALESCE 23324) (T_INSTANCEOF 23325) (T_IS_EQUAL 23326) (T_IS_GREATER_OR_EQUAL 23327) (T_IS_IDENTICAL 23328) (T_IS_NOT_EQUAL 23329) (T_IS_NOT_IDENTICAL 23330) (T_IS_SMALLER_OR_EQUAL 23331) (T_LOGICAL_AND 23332) (T_LOGICAL_OR 23333) (T_LOGICAL_XOR 23334) (T_POW 23335) (T_SL 23336) (T_SPACESHIP 23337) (T_SR 23338) ("^" 23339) (backup_fn_flags 23609) ("|" 23340)) 10808 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 23598)) 10809 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 23586)) 10810 (("(" 23497) (argument_list 23585)) 10811 (("=" 23505) (T_AND_EQUAL 23506) (T_COALESCE_EQUAL 23507) (T_CONCAT_EQUAL 23508) (T_DEC 23509) (T_DIV_EQUAL 23510) (T_INC 23511) (T_MINUS_EQUAL 23512) (T_MOD_EQUAL 23513) (T_MUL_EQUAL 23514) (T_OR_EQUAL 23515) (T_PLUS_EQUAL 23516) (T_POW_EQUAL 23517) (T_SL_EQUAL 23518) (T_SR_EQUAL 23519) (T_XOR_EQUAL 23520)) 10812 ((T_PAAMAYIM_NEKUDOTAYIM 23409)) 10813 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23310) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10814 (("%" 23311) ("*" 23312) ("+" 23313) ("-" 23314) ("." 23315) ("/" 23316) ("<" 23317) (">" 23318) ("?" 23319) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 23320) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 23321) (T_BOOLEAN_AND 23322) (T_BOOLEAN_OR 23323) (T_COALESCE 23324) (T_INSTANCEOF 23325) (T_IS_EQUAL 23326) (T_IS_GREATER_OR_EQUAL 23327) (T_IS_IDENTICAL 23328) (T_IS_NOT_EQUAL 23329) (T_IS_NOT_IDENTICAL 23330) (T_IS_SMALLER_OR_EQUAL 23331) (T_LOGICAL_AND 23332) (T_LOGICAL_OR 23333) (T_LOGICAL_XOR 23334) (T_POW 23335) (T_SL 23336) (T_SPACESHIP 23337) (T_SR 23338) ("^" 23339) ("|" 23340)) 10815 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23408) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10816 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23407) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10817 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23406) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10818 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23405) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10819 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23404) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10820 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23403) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10821 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23402) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10822 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23401) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10823 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 23396) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 23397) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 10824 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23395) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10825 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23394) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10826 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23393) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10827 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23392) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10828 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23391) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10829 (("$" 23356) ("(" 23357) (T_NAME_FULLY_QUALIFIED 23358) (T_NAME_QUALIFIED 23359) (T_NAME_RELATIVE 23360) (T_STATIC 23361) (T_STRING 23362) (T_VARIABLE 23363) (class_name 23364) (class_name_reference 23365) (name 23366) (new_variable 23367) (simple_variable 23368)) 10830 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23355) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10831 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23354) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10832 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23353) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10833 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23352) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10834 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23351) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10835 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23350) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10836 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23349) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10837 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23348) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10838 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23347) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10839 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23346) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10840 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23345) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10841 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23344) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10842 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23343) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10843 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23342) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10844 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23341) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10845 (("$" 23356) (T_VARIABLE 23363) (simple_variable 23387) ("{" 23388)) 10846 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 23385) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10847 ((T_PAAMAYIM_NEKUDOTAYIM 23383)) 10848 ((T_NULLSAFE_OBJECT_OPERATOR 23369) (T_OBJECT_OPERATOR 23370) (T_PAAMAYIM_NEKUDOTAYIM 23371) ("[" 23372)) 10849 (("$" 23356) (T_STRING 23376) (T_VARIABLE 23363) (property_name 23382) (simple_variable 23378) ("{" 23379)) 10850 (("$" 23356) (T_STRING 23376) (T_VARIABLE 23363) (property_name 23377) (simple_variable 23378) ("{" 23379)) 10851 (("$" 23356) (T_VARIABLE 23363) (simple_variable 23375)) 10852 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 23373) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 10853 (("]" 23374)) 10854 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23380) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10855 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23381)) 10856 (("$" 23356) (T_VARIABLE 23363) (simple_variable 23384)) 10857 (("%" 299) (")" 23386) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 10858 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23389) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10859 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23390)) 10860 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23400) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10861 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 23398) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 10862 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23399) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10863 (("$" 23233) (T_ABSTRACT 23410) (T_ARRAY 23411) (T_AS 23412) (T_BREAK 23413) (T_CALLABLE 23414) (T_CASE 23415) (T_CATCH 23416) (T_CLASS 23417) (T_CLASS_C 23418) (T_CLONE 23419) (T_CONST 23420) (T_CONTINUE 23421) (T_DECLARE 23422) (T_DEFAULT 23423) (T_DIR 23424) (T_DO 23425) (T_ECHO 23426) (T_ELSE 23427) (T_ELSEIF 23428) (T_EMPTY 23429) (T_ENDDECLARE 23430) (T_ENDFOR 23431) (T_ENDFOREACH 23432) (T_ENDIF 23433) (T_ENDSWITCH 23434) (T_ENDWHILE 23435) (T_ENUM 23436) (T_EVAL 23437) (T_EXIT 23438) (T_EXTENDS 23439) (T_FILE 23440) (T_FINAL 23441) (T_FINALLY 23442) (T_FN 23443) (T_FOR 23444) (T_FOREACH 23445) (T_FUNCTION 23446) (T_FUNC_C 23447) (T_GLOBAL 23448) (T_GOTO 23449) (T_IF 23450) (T_IMPLEMENTS 23451) (T_INCLUDE 23452) (T_INCLUDE_ONCE 23453) (T_INSTANCEOF 23454) (T_INSTEADOF 23455) (T_INTERFACE 23456) (T_ISSET 23457) (T_LINE 23458) (T_LIST 23459) (T_LOGICAL_AND 23460) (T_LOGICAL_OR 23461) (T_LOGICAL_XOR 23462) (T_MATCH 23463) (T_METHOD_C 23464) (T_NAMESPACE 23465) (T_NEW 23466) (T_NS_C 23467) (T_PRINT 23468) (T_PRIVATE 23469) (T_PROPERTY_C 23470) (T_PROTECTED 23471) (T_PUBLIC 23472) (T_READONLY 23473) (T_REQUIRE 23474) (T_REQUIRE_ONCE 23475) (T_RETURN 23476) (T_STATIC 23477) (T_STRING 23478) (T_SWITCH 23479) (T_THROW 23480) (T_TRAIT 23481) (T_TRAIT_C 23482) (T_TRY 23483) (T_UNSET 23484) (T_USE 23485) (T_VAR 23486) (T_VARIABLE 23281) (T_WHILE 23487) (T_YIELD 23488) (identifier 23489) (member_name 23490) (reserved_non_modifiers 23491) (semi_reserved 23492) (simple_variable 23493) ("{" 23494)) 10864 (("(" 23497) (argument_list 23498)) 10865 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23495) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10866 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23496)) 10867 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 23499) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 23500) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 23501) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10868 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 23504) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10869 (("," 1480) (possible_comma 23502)) 10870 ((")" 23503)) 10871 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (ampersand 23534) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23535) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10872 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23533) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10873 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23532) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10874 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23531) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10875 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23530) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10876 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23529) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10877 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23528) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10878 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23527) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10879 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23526) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10880 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23525) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10881 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23524) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10882 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23523) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10883 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23522) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10884 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23521) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10885 (("\"" 1520) ("$" 23536) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 23537) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 23538) ("[" 1542) (array_object_dereferenceable 23539) (callable_expr 23540) (callable_variable 23541) (class_constant 1546) (class_name 23542) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 23543) (name 23544) (new_dereferenceable 1552) (simple_variable 23545) (static_member 23546) (variable 23547) (variable_class_name 23548)) 10886 (("$" 23536) (T_VARIABLE 23538) (simple_variable 23581) ("{" 23582)) 10887 (("(" 23552) (argument_list 23580)) 10888 ((T_NULLSAFE_OBJECT_OPERATOR 23566) (T_OBJECT_OPERATOR 23567) ("[" 23568)) 10889 (("(" 23552) (argument_list 23565)) 10890 ((T_PAAMAYIM_NEKUDOTAYIM 23561)) 10891 (("(" 23552) (argument_list 23560)) 10892 ((T_PAAMAYIM_NEKUDOTAYIM 23549)) 10893 (("$" 23536) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 23538) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 23550) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 23551) ("{" 1642)) 10894 (("(" 23552) (argument_list 23553)) 10895 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 23554) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 23555) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 23556) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10896 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 23559) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 10897 (("," 1480) (possible_comma 23557)) 10898 ((")" 23558)) 10899 (("$" 23536) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 23538) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 23562) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 23563) ("{" 1658)) 10900 (("(" 23552) (argument_list 23564)) 10901 (("$" 23536) (T_STRING 23571) (T_VARIABLE 23538) (property_name 23578) (simple_variable 23573) ("{" 23574)) 10902 (("$" 23536) (T_STRING 23571) (T_VARIABLE 23538) (property_name 23572) (simple_variable 23573) ("{" 23574)) 10903 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 23569) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 10904 (("]" 23570)) 10905 (("(" 23552) (argument_list 23577)) 10906 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23575) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10907 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23576)) 10908 (("(" 23552) (argument_list 23579)) 10909 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23583) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10910 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23584)) 10911 ((backup_doc_comment 23587)) 10912 (("(" 23588)) 10913 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 23589)) 10914 ((")" 23590)) 10915 ((T_USE 164) (lexical_vars 23591)) 10916 ((":" 166) (return_type 23592)) 10917 ((backup_fn_flags 23593)) 10918 (("{" 23594)) 10919 ((inner_statement_list 23595)) 10920 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 23596) ("~" 127)) 10921 ((backup_fn_flags 23597)) 10922 ((backup_doc_comment 23599)) 10923 (("(" 23600)) 10924 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 23601)) 10925 ((")" 23602)) 10926 ((":" 2249) (return_type 23603)) 10927 ((T_DOUBLE_ARROW 23604)) 10928 ((backup_fn_flags 23605)) 10929 ((backup_lex_pos 23606)) 10930 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23607) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10931 (("%" 23311) ("*" 23312) ("+" 23313) ("-" 23314) ("." 23315) ("/" 23316) ("<" 23317) (">" 23318) ("?" 23319) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 23320) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 23321) (T_BOOLEAN_AND 23322) (T_BOOLEAN_OR 23323) (T_COALESCE 23324) (T_INSTANCEOF 23325) (T_IS_EQUAL 23326) (T_IS_GREATER_OR_EQUAL 23327) (T_IS_IDENTICAL 23328) (T_IS_NOT_EQUAL 23329) (T_IS_NOT_IDENTICAL 23330) (T_IS_SMALLER_OR_EQUAL 23331) (T_LOGICAL_AND 23332) (T_LOGICAL_OR 23333) (T_LOGICAL_XOR 23334) (T_POW 23335) (T_SL 23336) (T_SPACESHIP 23337) (T_SR 23338) ("^" 23339) (backup_fn_flags 23608) ("|" 23340)) 10932 (("$" 23233) (T_ABSTRACT 23410) (T_ARRAY 23411) (T_AS 23412) (T_BREAK 23413) (T_CALLABLE 23414) (T_CASE 23415) (T_CATCH 23416) (T_CLASS 23417) (T_CLASS_C 23418) (T_CLONE 23419) (T_CONST 23420) (T_CONTINUE 23421) (T_DECLARE 23422) (T_DEFAULT 23423) (T_DIR 23424) (T_DO 23425) (T_ECHO 23426) (T_ELSE 23427) (T_ELSEIF 23428) (T_EMPTY 23429) (T_ENDDECLARE 23430) (T_ENDFOR 23431) (T_ENDFOREACH 23432) (T_ENDIF 23433) (T_ENDSWITCH 23434) (T_ENDWHILE 23435) (T_ENUM 23436) (T_EVAL 23437) (T_EXIT 23438) (T_EXTENDS 23439) (T_FILE 23440) (T_FINAL 23441) (T_FINALLY 23442) (T_FN 23443) (T_FOR 23444) (T_FOREACH 23445) (T_FUNCTION 23446) (T_FUNC_C 23447) (T_GLOBAL 23448) (T_GOTO 23449) (T_IF 23450) (T_IMPLEMENTS 23451) (T_INCLUDE 23452) (T_INCLUDE_ONCE 23453) (T_INSTANCEOF 23454) (T_INSTEADOF 23455) (T_INTERFACE 23456) (T_ISSET 23457) (T_LINE 23458) (T_LIST 23459) (T_LOGICAL_AND 23460) (T_LOGICAL_OR 23461) (T_LOGICAL_XOR 23462) (T_MATCH 23463) (T_METHOD_C 23464) (T_NAMESPACE 23465) (T_NEW 23466) (T_NS_C 23467) (T_PRINT 23468) (T_PRIVATE 23469) (T_PROPERTY_C 23470) (T_PROTECTED 23471) (T_PUBLIC 23472) (T_READONLY 23473) (T_REQUIRE 23474) (T_REQUIRE_ONCE 23475) (T_RETURN 23476) (T_STATIC 23477) (T_STRING 23478) (T_SWITCH 23479) (T_THROW 23480) (T_TRAIT 23481) (T_TRAIT_C 23482) (T_TRY 23483) (T_UNSET 23484) (T_USE 23485) (T_VAR 23486) (T_VARIABLE 23281) (T_WHILE 23487) (T_YIELD 23488) (identifier 23611) (member_name 23612) (reserved_non_modifiers 23491) (semi_reserved 23492) (simple_variable 23613) ("{" 23614)) 10933 (("(" 23497) (argument_list 23617)) 10934 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23615) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10935 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23616)) 10936 ((T_FN 35) (T_FUNCTION 129) (fn 23295) #'23296 (inline_function 23621)) 10937 (("$" 23233) (T_STRING 23627) (T_VARIABLE 23281) (property_name 23634) (simple_variable 23629) ("{" 23630)) 10938 (("$" 23233) (T_STRING 23627) (T_VARIABLE 23281) (property_name 23628) (simple_variable 23629) ("{" 23630)) 10939 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 23625) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 10940 (("]" 23626)) 10941 (("(" 23497) (argument_list 23633)) 10942 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23631) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 10943 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23632)) 10944 (("(" 23497) (argument_list 23635)) 10945 (("`" 23637)) 10946 (("]" 23639)) 10947 (("=" 23640)) 10948 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 23641) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 10949 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24163) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10950 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 24161) (encaps_var 5237)) 10951 (("$" 23645) (T_VARIABLE 23693) (simple_variable 24157) ("{" 24158)) 10952 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24155) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 10953 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24154) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10954 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24153) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10955 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24152) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10956 (("(" 24149)) 10957 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24148) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10958 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24147) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10959 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24146) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10960 (("\"" 1520) ("$" 23948) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 23949) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 23950) ("[" 1542) (array_object_dereferenceable 23951) (callable_expr 23952) (callable_variable 23953) (class_constant 1546) (class_name 23954) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 23955) (name 23956) (new_dereferenceable 1552) (simple_variable 23957) (static_member 23958) (variable 24145) (variable_class_name 23960)) 10961 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24144) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10962 (("(" 24141)) 10963 (("(" 24138)) 10964 (("(" 24129) (argument_list 24130) (ctor_arguments 24131)) 10965 (("\"" 1520) ("$" 23948) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 23949) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 23950) ("[" 1542) (array_object_dereferenceable 23951) (callable_expr 23952) (callable_variable 23953) (class_constant 1546) (class_name 23954) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 23955) (name 23956) (new_dereferenceable 1552) (simple_variable 23957) (static_member 23958) (variable 24128) (variable_class_name 23960)) 10966 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24127) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10967 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24126) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10968 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24125) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10969 (("(" 24121)) 10970 (("(" 24116)) 10971 (("(" 24110)) 10972 (("$" 23948) ("(" 24073) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 24074) (T_NAME_QUALIFIED 24075) (T_NAME_RELATIVE 24076) (T_READONLY 2931) (T_STATIC 24077) (T_STRING 24078) (T_VARIABLE 23950) (anonymous_class 24079) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 24080) (attribute 2937) (attributes 24081) (class_modifier 2939) (class_name 24082) (class_name_reference 24083) (name 24084) (new_variable 24085) (simple_variable 24086)) 10973 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24072) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10974 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24071) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10975 (("(" 23909) (argument_list 24070)) 10976 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24069) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10977 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24068) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10978 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 24063) (T_END_HEREDOC 24064) (T_VARIABLE 2896) (encaps_list 24065) (encaps_var 2898)) 10979 ((T_FN 35) (T_FUNCTION 129) (fn 23707) #'23708 (inline_function 24062)) 10980 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24061) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10981 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24060) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10982 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24059) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10983 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24056) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10984 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24055) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10985 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 24051) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 10986 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 24049) (encaps_list 2804) (encaps_var 2805)) 10987 ((T_NULLSAFE_OBJECT_OPERATOR 24035) (T_OBJECT_OPERATOR 24036) ("[" 24037)) 10988 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 24032) (attribute 2733) (fn 23707) #'23708 (inline_function 24033)) 10989 (("(" 23909) (argument_list 24031)) 10990 ((T_PAAMAYIM_NEKUDOTAYIM 24023)) 10991 (("%" 23723) ("*" 23724) ("+" 23725) ("-" 23726) ("." 23727) ("/" 23728) ("<" 23729) (">" 23730) ("?" 23731) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 23732) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 23733) (T_BOOLEAN_AND 23734) (T_BOOLEAN_OR 23735) (T_COALESCE 23736) (T_DOUBLE_ARROW 24021) (T_INSTANCEOF 23737) (T_IS_EQUAL 23738) (T_IS_GREATER_OR_EQUAL 23739) (T_IS_IDENTICAL 23740) (T_IS_NOT_EQUAL 23741) (T_IS_NOT_IDENTICAL 23742) (T_IS_SMALLER_OR_EQUAL 23743) (T_LOGICAL_AND 23744) (T_LOGICAL_OR 23745) (T_LOGICAL_XOR 23746) (T_POW 23747) (T_SL 23748) (T_SPACESHIP 23749) (T_SR 23750) ("^" 23751) ("|" 23752)) 10992 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 24010)) 10993 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 23998)) 10994 (("(" 23909) (argument_list 23997)) 10995 (("=" 23917) (T_AND_EQUAL 23918) (T_COALESCE_EQUAL 23919) (T_CONCAT_EQUAL 23920) (T_DEC 23921) (T_DIV_EQUAL 23922) (T_INC 23923) (T_MINUS_EQUAL 23924) (T_MOD_EQUAL 23925) (T_MUL_EQUAL 23926) (T_OR_EQUAL 23927) (T_PLUS_EQUAL 23928) (T_POW_EQUAL 23929) (T_SL_EQUAL 23930) (T_SR_EQUAL 23931) (T_XOR_EQUAL 23932)) 10996 ((T_PAAMAYIM_NEKUDOTAYIM 23821)) 10997 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23722) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 10998 (("%" 23723) ("*" 23724) ("+" 23725) ("-" 23726) ("." 23727) ("/" 23728) ("<" 23729) (">" 23730) ("?" 23731) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 23732) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 23733) (T_BOOLEAN_AND 23734) (T_BOOLEAN_OR 23735) (T_COALESCE 23736) (T_INSTANCEOF 23737) (T_IS_EQUAL 23738) (T_IS_GREATER_OR_EQUAL 23739) (T_IS_IDENTICAL 23740) (T_IS_NOT_EQUAL 23741) (T_IS_NOT_IDENTICAL 23742) (T_IS_SMALLER_OR_EQUAL 23743) (T_LOGICAL_AND 23744) (T_LOGICAL_OR 23745) (T_LOGICAL_XOR 23746) (T_POW 23747) (T_SL 23748) (T_SPACESHIP 23749) (T_SR 23750) ("^" 23751) ("|" 23752)) 10999 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23820) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11000 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23819) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11001 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23818) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11002 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23817) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11003 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23816) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11004 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23815) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11005 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23814) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11006 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23813) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11007 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 23808) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 23809) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 11008 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23807) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11009 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23806) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11010 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23805) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11011 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23804) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11012 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23803) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11013 (("$" 23768) ("(" 23769) (T_NAME_FULLY_QUALIFIED 23770) (T_NAME_QUALIFIED 23771) (T_NAME_RELATIVE 23772) (T_STATIC 23773) (T_STRING 23774) (T_VARIABLE 23775) (class_name 23776) (class_name_reference 23777) (name 23778) (new_variable 23779) (simple_variable 23780)) 11014 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23767) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11015 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23766) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11016 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23765) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11017 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23764) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11018 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23763) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11019 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23762) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11020 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23761) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11021 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23760) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11022 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23759) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11023 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23758) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11024 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23757) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11025 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23756) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11026 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23755) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11027 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23754) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11028 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23753) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11029 (("$" 23768) (T_VARIABLE 23775) (simple_variable 23799) ("{" 23800)) 11030 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 23797) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11031 ((T_PAAMAYIM_NEKUDOTAYIM 23795)) 11032 ((T_NULLSAFE_OBJECT_OPERATOR 23781) (T_OBJECT_OPERATOR 23782) (T_PAAMAYIM_NEKUDOTAYIM 23783) ("[" 23784)) 11033 (("$" 23768) (T_STRING 23788) (T_VARIABLE 23775) (property_name 23794) (simple_variable 23790) ("{" 23791)) 11034 (("$" 23768) (T_STRING 23788) (T_VARIABLE 23775) (property_name 23789) (simple_variable 23790) ("{" 23791)) 11035 (("$" 23768) (T_VARIABLE 23775) (simple_variable 23787)) 11036 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 23785) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 11037 (("]" 23786)) 11038 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23792) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11039 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23793)) 11040 (("$" 23768) (T_VARIABLE 23775) (simple_variable 23796)) 11041 (("%" 299) (")" 23798) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11042 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23801) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11043 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23802)) 11044 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23812) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11045 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 23810) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 11046 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23811) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11047 (("$" 23645) (T_ABSTRACT 23822) (T_ARRAY 23823) (T_AS 23824) (T_BREAK 23825) (T_CALLABLE 23826) (T_CASE 23827) (T_CATCH 23828) (T_CLASS 23829) (T_CLASS_C 23830) (T_CLONE 23831) (T_CONST 23832) (T_CONTINUE 23833) (T_DECLARE 23834) (T_DEFAULT 23835) (T_DIR 23836) (T_DO 23837) (T_ECHO 23838) (T_ELSE 23839) (T_ELSEIF 23840) (T_EMPTY 23841) (T_ENDDECLARE 23842) (T_ENDFOR 23843) (T_ENDFOREACH 23844) (T_ENDIF 23845) (T_ENDSWITCH 23846) (T_ENDWHILE 23847) (T_ENUM 23848) (T_EVAL 23849) (T_EXIT 23850) (T_EXTENDS 23851) (T_FILE 23852) (T_FINAL 23853) (T_FINALLY 23854) (T_FN 23855) (T_FOR 23856) (T_FOREACH 23857) (T_FUNCTION 23858) (T_FUNC_C 23859) (T_GLOBAL 23860) (T_GOTO 23861) (T_IF 23862) (T_IMPLEMENTS 23863) (T_INCLUDE 23864) (T_INCLUDE_ONCE 23865) (T_INSTANCEOF 23866) (T_INSTEADOF 23867) (T_INTERFACE 23868) (T_ISSET 23869) (T_LINE 23870) (T_LIST 23871) (T_LOGICAL_AND 23872) (T_LOGICAL_OR 23873) (T_LOGICAL_XOR 23874) (T_MATCH 23875) (T_METHOD_C 23876) (T_NAMESPACE 23877) (T_NEW 23878) (T_NS_C 23879) (T_PRINT 23880) (T_PRIVATE 23881) (T_PROPERTY_C 23882) (T_PROTECTED 23883) (T_PUBLIC 23884) (T_READONLY 23885) (T_REQUIRE 23886) (T_REQUIRE_ONCE 23887) (T_RETURN 23888) (T_STATIC 23889) (T_STRING 23890) (T_SWITCH 23891) (T_THROW 23892) (T_TRAIT 23893) (T_TRAIT_C 23894) (T_TRY 23895) (T_UNSET 23896) (T_USE 23897) (T_VAR 23898) (T_VARIABLE 23693) (T_WHILE 23899) (T_YIELD 23900) (identifier 23901) (member_name 23902) (reserved_non_modifiers 23903) (semi_reserved 23904) (simple_variable 23905) ("{" 23906)) 11048 (("(" 23909) (argument_list 23910)) 11049 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23907) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11050 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23908)) 11051 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 23911) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 23912) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 23913) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11052 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 23916) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11053 (("," 1480) (possible_comma 23914)) 11054 ((")" 23915)) 11055 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (ampersand 23946) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23947) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11056 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23945) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11057 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23944) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11058 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23943) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11059 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23942) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11060 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23941) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11061 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23940) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11062 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23939) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11063 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23938) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11064 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23937) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11065 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23936) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11066 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23935) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11067 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23934) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11068 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 23933) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11069 (("\"" 1520) ("$" 23948) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 23949) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 23950) ("[" 1542) (array_object_dereferenceable 23951) (callable_expr 23952) (callable_variable 23953) (class_constant 1546) (class_name 23954) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 23955) (name 23956) (new_dereferenceable 1552) (simple_variable 23957) (static_member 23958) (variable 23959) (variable_class_name 23960)) 11070 (("$" 23948) (T_VARIABLE 23950) (simple_variable 23993) ("{" 23994)) 11071 (("(" 23964) (argument_list 23992)) 11072 ((T_NULLSAFE_OBJECT_OPERATOR 23978) (T_OBJECT_OPERATOR 23979) ("[" 23980)) 11073 (("(" 23964) (argument_list 23977)) 11074 ((T_PAAMAYIM_NEKUDOTAYIM 23973)) 11075 (("(" 23964) (argument_list 23972)) 11076 ((T_PAAMAYIM_NEKUDOTAYIM 23961)) 11077 (("$" 23948) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 23950) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 23962) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 23963) ("{" 1642)) 11078 (("(" 23964) (argument_list 23965)) 11079 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 23966) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 23967) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 23968) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11080 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 23971) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11081 (("," 1480) (possible_comma 23969)) 11082 ((")" 23970)) 11083 (("$" 23948) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 23950) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 23974) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 23975) ("{" 1658)) 11084 (("(" 23964) (argument_list 23976)) 11085 (("$" 23948) (T_STRING 23983) (T_VARIABLE 23950) (property_name 23990) (simple_variable 23985) ("{" 23986)) 11086 (("$" 23948) (T_STRING 23983) (T_VARIABLE 23950) (property_name 23984) (simple_variable 23985) ("{" 23986)) 11087 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 23981) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 11088 (("]" 23982)) 11089 (("(" 23964) (argument_list 23989)) 11090 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23987) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11091 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23988)) 11092 (("(" 23964) (argument_list 23991)) 11093 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 23995) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11094 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 23996)) 11095 ((backup_doc_comment 23999)) 11096 (("(" 24000)) 11097 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 24001)) 11098 ((")" 24002)) 11099 ((T_USE 164) (lexical_vars 24003)) 11100 ((":" 166) (return_type 24004)) 11101 ((backup_fn_flags 24005)) 11102 (("{" 24006)) 11103 ((inner_statement_list 24007)) 11104 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 24008) ("~" 127)) 11105 ((backup_fn_flags 24009)) 11106 ((backup_doc_comment 24011)) 11107 (("(" 24012)) 11108 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 24013)) 11109 ((")" 24014)) 11110 ((":" 2249) (return_type 24015)) 11111 ((T_DOUBLE_ARROW 24016)) 11112 ((backup_fn_flags 24017)) 11113 ((backup_lex_pos 24018)) 11114 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24019) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11115 (("%" 23723) ("*" 23724) ("+" 23725) ("-" 23726) ("." 23727) ("/" 23728) ("<" 23729) (">" 23730) ("?" 23731) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 23732) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 23733) (T_BOOLEAN_AND 23734) (T_BOOLEAN_OR 23735) (T_COALESCE 23736) (T_INSTANCEOF 23737) (T_IS_EQUAL 23738) (T_IS_GREATER_OR_EQUAL 23739) (T_IS_IDENTICAL 23740) (T_IS_NOT_EQUAL 23741) (T_IS_NOT_IDENTICAL 23742) (T_IS_SMALLER_OR_EQUAL 23743) (T_LOGICAL_AND 23744) (T_LOGICAL_OR 23745) (T_LOGICAL_XOR 23746) (T_POW 23747) (T_SL 23748) (T_SPACESHIP 23749) (T_SR 23750) ("^" 23751) (backup_fn_flags 24020) ("|" 23752)) 11116 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24022) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 11117 (("$" 23645) (T_ABSTRACT 23822) (T_ARRAY 23823) (T_AS 23824) (T_BREAK 23825) (T_CALLABLE 23826) (T_CASE 23827) (T_CATCH 23828) (T_CLASS 23829) (T_CLASS_C 23830) (T_CLONE 23831) (T_CONST 23832) (T_CONTINUE 23833) (T_DECLARE 23834) (T_DEFAULT 23835) (T_DIR 23836) (T_DO 23837) (T_ECHO 23838) (T_ELSE 23839) (T_ELSEIF 23840) (T_EMPTY 23841) (T_ENDDECLARE 23842) (T_ENDFOR 23843) (T_ENDFOREACH 23844) (T_ENDIF 23845) (T_ENDSWITCH 23846) (T_ENDWHILE 23847) (T_ENUM 23848) (T_EVAL 23849) (T_EXIT 23850) (T_EXTENDS 23851) (T_FILE 23852) (T_FINAL 23853) (T_FINALLY 23854) (T_FN 23855) (T_FOR 23856) (T_FOREACH 23857) (T_FUNCTION 23858) (T_FUNC_C 23859) (T_GLOBAL 23860) (T_GOTO 23861) (T_IF 23862) (T_IMPLEMENTS 23863) (T_INCLUDE 23864) (T_INCLUDE_ONCE 23865) (T_INSTANCEOF 23866) (T_INSTEADOF 23867) (T_INTERFACE 23868) (T_ISSET 23869) (T_LINE 23870) (T_LIST 23871) (T_LOGICAL_AND 23872) (T_LOGICAL_OR 23873) (T_LOGICAL_XOR 23874) (T_MATCH 23875) (T_METHOD_C 23876) (T_NAMESPACE 23877) (T_NEW 23878) (T_NS_C 23879) (T_PRINT 23880) (T_PRIVATE 23881) (T_PROPERTY_C 23882) (T_PROTECTED 23883) (T_PUBLIC 23884) (T_READONLY 23885) (T_REQUIRE 23886) (T_REQUIRE_ONCE 23887) (T_RETURN 23888) (T_STATIC 23889) (T_STRING 23890) (T_SWITCH 23891) (T_THROW 23892) (T_TRAIT 23893) (T_TRAIT_C 23894) (T_TRY 23895) (T_UNSET 23896) (T_USE 23897) (T_VAR 23898) (T_VARIABLE 23693) (T_WHILE 23899) (T_YIELD 23900) (identifier 24024) (member_name 24025) (reserved_non_modifiers 23903) (semi_reserved 23904) (simple_variable 24026) ("{" 24027)) 11118 (("(" 23909) (argument_list 24030)) 11119 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 24028) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11120 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 24029)) 11121 ((T_FN 35) (T_FUNCTION 129) (fn 23707) #'23708 (inline_function 24034)) 11122 (("$" 23645) (T_STRING 24040) (T_VARIABLE 23693) (property_name 24047) (simple_variable 24042) ("{" 24043)) 11123 (("$" 23645) (T_STRING 24040) (T_VARIABLE 23693) (property_name 24041) (simple_variable 24042) ("{" 24043)) 11124 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 24038) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 11125 (("]" 24039)) 11126 (("(" 23909) (argument_list 24046)) 11127 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 24044) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11128 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 24045)) 11129 (("(" 23909) (argument_list 24048)) 11130 (("`" 24050)) 11131 (("]" 24052)) 11132 (("=" 24053)) 11133 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24054) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11134 (("%" 23723) ("*" 23724) ("+" 23725) ("-" 23726) ("." 23727) ("/" 23728) ("<" 23729) (">" 23730) ("?" 23731) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 23732) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 23733) (T_BOOLEAN_AND 23734) (T_BOOLEAN_OR 23735) (T_COALESCE 23736) (T_DOUBLE_ARROW 24057) (T_INSTANCEOF 23737) (T_IS_EQUAL 23738) (T_IS_GREATER_OR_EQUAL 23739) (T_IS_IDENTICAL 23740) (T_IS_NOT_EQUAL 23741) (T_IS_NOT_IDENTICAL 23742) (T_IS_SMALLER_OR_EQUAL 23743) (T_LOGICAL_AND 23744) (T_LOGICAL_OR 23745) (T_LOGICAL_XOR 23746) (T_POW 23747) (T_SL 23748) (T_SPACESHIP 23749) (T_SR 23750) ("^" 23751) ("|" 23752)) 11135 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24058) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11136 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 24067) (T_VARIABLE 2896) (encaps_var 2910)) 11137 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 24066) (T_VARIABLE 2896) (encaps_var 2901)) 11138 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24108) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11139 ((T_CLASS 24100)) 11140 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 24099) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 24080) (attribute 2958) (class_modifier 2939)) 11141 ((T_PAAMAYIM_NEKUDOTAYIM 24097)) 11142 (("(" 23964) (argument_list 24096)) 11143 ((T_NULLSAFE_OBJECT_OPERATOR 24087) (T_OBJECT_OPERATOR 24088) (T_PAAMAYIM_NEKUDOTAYIM 24089) ("[" 24090)) 11144 (("$" 23948) (T_STRING 23983) (T_VARIABLE 23950) (property_name 24095) (simple_variable 23985) ("{" 23986)) 11145 (("$" 23948) (T_STRING 23983) (T_VARIABLE 23950) (property_name 24094) (simple_variable 23985) ("{" 23986)) 11146 (("$" 23948) (T_VARIABLE 23950) (simple_variable 24093)) 11147 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 24091) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 11148 (("]" 24092)) 11149 (("$" 23948) (T_VARIABLE 23950) (simple_variable 24098)) 11150 (("(" 2960) (argument_list 2961) (ctor_arguments 24101)) 11151 ((T_EXTENDS 2963) (extends_from 24102)) 11152 ((T_IMPLEMENTS 2965) (implements_list 24103)) 11153 ((backup_doc_comment 24104)) 11154 (("{" 24105)) 11155 ((class_statement_list 24106)) 11156 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 24107)) 11157 (("%" 299) (")" 24109) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11158 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24111) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11159 (("%" 299) (")" 24112) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11160 (("{" 24113)) 11161 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 24114) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 11162 (("}" 24115)) 11163 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 24117) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 11164 ((")" 24118)) 11165 (("=" 24119)) 11166 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24120) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11167 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 24122) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11168 (("," 5192) (possible_comma 24123)) 11169 ((")" 24124)) 11170 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 24132) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 24133) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 24134) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11171 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 24137) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11172 (("," 1480) (possible_comma 24135)) 11173 ((")" 24136)) 11174 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24139) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11175 (("%" 299) (")" 24140) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11176 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24142) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11177 (("%" 299) (")" 24143) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11178 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 24150) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 11179 ((")" 24151)) 11180 (("%" 299) (")" 24156) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11181 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 24159) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11182 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 24160)) 11183 (("\"" 24162) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 11184 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 24172) (T_VARIABLE 2896) (encaps_var 2910)) 11185 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 24171) (T_VARIABLE 2896) (encaps_var 2901)) 11186 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24213) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11187 ((T_CLASS 24205)) 11188 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 24204) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 24185) (attribute 2958) (class_modifier 2939)) 11189 ((T_PAAMAYIM_NEKUDOTAYIM 24202)) 11190 (("(" 23552) (argument_list 24201)) 11191 ((T_NULLSAFE_OBJECT_OPERATOR 24192) (T_OBJECT_OPERATOR 24193) (T_PAAMAYIM_NEKUDOTAYIM 24194) ("[" 24195)) 11192 (("$" 23536) (T_STRING 23571) (T_VARIABLE 23538) (property_name 24200) (simple_variable 23573) ("{" 23574)) 11193 (("$" 23536) (T_STRING 23571) (T_VARIABLE 23538) (property_name 24199) (simple_variable 23573) ("{" 23574)) 11194 (("$" 23536) (T_VARIABLE 23538) (simple_variable 24198)) 11195 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 24196) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 11196 (("]" 24197)) 11197 (("$" 23536) (T_VARIABLE 23538) (simple_variable 24203)) 11198 (("(" 2960) (argument_list 2961) (ctor_arguments 24206)) 11199 ((T_EXTENDS 2963) (extends_from 24207)) 11200 ((T_IMPLEMENTS 2965) (implements_list 24208)) 11201 ((backup_doc_comment 24209)) 11202 (("{" 24210)) 11203 ((class_statement_list 24211)) 11204 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 24212)) 11205 (("%" 299) (")" 24214) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11206 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24216) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11207 (("%" 299) (")" 24217) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11208 (("{" 24218)) 11209 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 24219) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 11210 (("}" 24220)) 11211 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 24222) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 11212 ((")" 24223)) 11213 (("=" 24224)) 11214 (("!" 23231) ("\"" 23232) ("$" 23233) ("(" 23234) ("+" 23235) ("-" 23236) ("@" 23237) (T_ARRAY 23238) (T_ARRAY_CAST 23239) (T_ATTRIBUTE 128) (T_BOOL_CAST 23240) (T_CLASS_C 23241) (T_CLONE 23242) (T_CONSTANT_ENCAPSED_STRING 23243) (T_DEC 23244) (T_DIR 23245) (T_DNUMBER 23246) (T_DOUBLE_CAST 23247) (T_EMPTY 23248) (T_EVAL 23249) (T_EXIT 23250) (T_FILE 23251) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23252) (T_INC 23253) (T_INCLUDE 23254) (T_INCLUDE_ONCE 23255) (T_INT_CAST 23256) (T_ISSET 23257) (T_LINE 23258) (T_LIST 23259) (T_LNUMBER 23260) (T_MATCH 23261) (T_METHOD_C 23262) (T_NAME_FULLY_QUALIFIED 23263) (T_NAME_QUALIFIED 23264) (T_NAME_RELATIVE 23265) (T_NEW 23266) (T_NS_C 23267) (T_OBJECT_CAST 23268) (T_PRINT 23269) (T_PROPERTY_C 23270) (T_READONLY 23271) (T_REQUIRE 23272) (T_REQUIRE_ONCE 23273) (T_START_HEREDOC 23274) (T_STATIC 23275) (T_STRING 23276) (T_STRING_CAST 23277) (T_THROW 23278) (T_TRAIT_C 23279) (T_UNSET_CAST 23280) (T_VARIABLE 23281) (T_YIELD 23282) (T_YIELD_FROM 23283) ("[" 23284) ("`" 23285) (array_object_dereferenceable 23286) (attribute 133) (attributes 23287) (callable_expr 23288) (callable_variable 23289) (class_constant 23290) (class_name 23291) (constant 23292) (dereferenceable_scalar 23293) (expr 24225) (fn 23295) (fully_dereferenceable 105) #'23296 (function_call 23297) (inline_function 23298) (internal_functions_in_yacc 23299) (match 23300) (name 23301) (new_dereferenceable 23302) (new_non_dereferenceable 23303) (scalar 23304) (simple_variable 23305) (static_member 23306) (variable 23307) (variable_class_name 23308) ("~" 23309)) 11215 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 24227) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11216 (("," 5192) (possible_comma 24228)) 11217 ((")" 24229)) 11218 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 24237) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 24238) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 24239) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11219 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 24242) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11220 (("," 1480) (possible_comma 24240)) 11221 ((")" 24241)) 11222 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24244) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11223 (("%" 299) (")" 24245) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11224 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24247) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11225 (("%" 299) (")" 24248) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11226 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 24255) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 11227 ((")" 24256)) 11228 (("%" 299) (")" 24261) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11229 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 24264) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11230 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 24265)) 11231 (("\"" 24267) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 11232 ((inner_statement_list 24270)) 11233 (("$" 22857) (T_ABSTRACT 23034) (T_ARRAY 23035) (T_AS 23036) (T_BREAK 23037) (T_CALLABLE 23038) (T_CASE 23039) (T_CATCH 23040) (T_CLASS 23041) (T_CLASS_C 23042) (T_CLONE 23043) (T_CONST 23044) (T_CONTINUE 23045) (T_DECLARE 23046) (T_DEFAULT 23047) (T_DIR 23048) (T_DO 23049) (T_ECHO 23050) (T_ELSE 23051) (T_ELSEIF 23052) (T_EMPTY 23053) (T_ENDDECLARE 23054) (T_ENDFOR 23055) (T_ENDFOREACH 23056) (T_ENDIF 23057) (T_ENDSWITCH 23058) (T_ENDWHILE 23059) (T_ENUM 23060) (T_EVAL 23061) (T_EXIT 23062) (T_EXTENDS 23063) (T_FILE 23064) (T_FINAL 23065) (T_FINALLY 23066) (T_FN 23067) (T_FOR 23068) (T_FOREACH 23069) (T_FUNCTION 23070) (T_FUNC_C 23071) (T_GLOBAL 23072) (T_GOTO 23073) (T_IF 23074) (T_IMPLEMENTS 23075) (T_INCLUDE 23076) (T_INCLUDE_ONCE 23077) (T_INSTANCEOF 23078) (T_INSTEADOF 23079) (T_INTERFACE 23080) (T_ISSET 23081) (T_LINE 23082) (T_LIST 23083) (T_LOGICAL_AND 23084) (T_LOGICAL_OR 23085) (T_LOGICAL_XOR 23086) (T_MATCH 23087) (T_METHOD_C 23088) (T_NAMESPACE 23089) (T_NEW 23090) (T_NS_C 23091) (T_PRINT 23092) (T_PRIVATE 23093) (T_PROPERTY_C 23094) (T_PROTECTED 23095) (T_PUBLIC 23096) (T_READONLY 23097) (T_REQUIRE 23098) (T_REQUIRE_ONCE 23099) (T_RETURN 23100) (T_STATIC 23101) (T_STRING 23102) (T_SWITCH 23103) (T_THROW 23104) (T_TRAIT 23105) (T_TRAIT_C 23106) (T_TRY 23107) (T_UNSET 23108) (T_USE 23109) (T_VAR 23110) (T_VARIABLE 22905) (T_WHILE 23111) (T_YIELD 23112) (identifier 24272) (member_name 24273) (reserved_non_modifiers 23115) (semi_reserved 23116) (simple_variable 24274) ("{" 24275)) 11234 (("(" 23121) (argument_list 24278)) 11235 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 24276) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11236 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 24277)) 11237 ((T_FN 35) (T_FUNCTION 129) (fn 22919) #'22920 (inline_function 24282)) 11238 (("$" 22857) (T_STRING 24288) (T_VARIABLE 22905) (property_name 24295) (simple_variable 24290) ("{" 24291)) 11239 (("$" 22857) (T_STRING 24288) (T_VARIABLE 22905) (property_name 24289) (simple_variable 24290) ("{" 24291)) 11240 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 24286) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 11241 (("]" 24287)) 11242 (("(" 23121) (argument_list 24294)) 11243 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 24292) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11244 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 24293)) 11245 (("(" 23121) (argument_list 24296)) 11246 (("`" 24298)) 11247 (("]" 24300)) 11248 (("=" 24301)) 11249 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24302) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 11250 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24824) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11251 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 24822) (encaps_var 5237)) 11252 (("$" 24306) (T_VARIABLE 24354) (simple_variable 24818) ("{" 24819)) 11253 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24816) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11254 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24815) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11255 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24814) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11256 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24813) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11257 (("(" 24810)) 11258 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24809) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11259 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24808) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11260 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24807) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11261 (("\"" 1520) ("$" 24609) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 24610) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 24611) ("[" 1542) (array_object_dereferenceable 24612) (callable_expr 24613) (callable_variable 24614) (class_constant 1546) (class_name 24615) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 24616) (name 24617) (new_dereferenceable 1552) (simple_variable 24618) (static_member 24619) (variable 24806) (variable_class_name 24621)) 11262 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24805) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11263 (("(" 24802)) 11264 (("(" 24799)) 11265 (("(" 24790) (argument_list 24791) (ctor_arguments 24792)) 11266 (("\"" 1520) ("$" 24609) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 24610) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 24611) ("[" 1542) (array_object_dereferenceable 24612) (callable_expr 24613) (callable_variable 24614) (class_constant 1546) (class_name 24615) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 24616) (name 24617) (new_dereferenceable 1552) (simple_variable 24618) (static_member 24619) (variable 24789) (variable_class_name 24621)) 11267 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24788) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11268 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24787) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11269 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24786) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11270 (("(" 24782)) 11271 (("(" 24777)) 11272 (("(" 24771)) 11273 (("$" 24609) ("(" 24734) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 24735) (T_NAME_QUALIFIED 24736) (T_NAME_RELATIVE 24737) (T_READONLY 2931) (T_STATIC 24738) (T_STRING 24739) (T_VARIABLE 24611) (anonymous_class 24740) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 24741) (attribute 2937) (attributes 24742) (class_modifier 2939) (class_name 24743) (class_name_reference 24744) (name 24745) (new_variable 24746) (simple_variable 24747)) 11274 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24733) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11275 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24732) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11276 (("(" 24570) (argument_list 24731)) 11277 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24730) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11278 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24729) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11279 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 24724) (T_END_HEREDOC 24725) (T_VARIABLE 2896) (encaps_list 24726) (encaps_var 2898)) 11280 ((T_FN 35) (T_FUNCTION 129) (fn 24368) #'24369 (inline_function 24723)) 11281 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24722) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11282 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24721) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11283 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24720) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11284 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24717) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11285 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24716) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11286 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 24712) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 11287 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 24710) (encaps_list 2804) (encaps_var 2805)) 11288 ((T_NULLSAFE_OBJECT_OPERATOR 24696) (T_OBJECT_OPERATOR 24697) ("[" 24698)) 11289 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 24693) (attribute 2733) (fn 24368) #'24369 (inline_function 24694)) 11290 (("(" 24570) (argument_list 24692)) 11291 ((T_PAAMAYIM_NEKUDOTAYIM 24684)) 11292 (("%" 24384) ("*" 24385) ("+" 24386) ("-" 24387) ("." 24388) ("/" 24389) ("<" 24390) (">" 24391) ("?" 24392) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 24393) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 24394) (T_BOOLEAN_AND 24395) (T_BOOLEAN_OR 24396) (T_COALESCE 24397) (T_DOUBLE_ARROW 24682) (T_INSTANCEOF 24398) (T_IS_EQUAL 24399) (T_IS_GREATER_OR_EQUAL 24400) (T_IS_IDENTICAL 24401) (T_IS_NOT_EQUAL 24402) (T_IS_NOT_IDENTICAL 24403) (T_IS_SMALLER_OR_EQUAL 24404) (T_LOGICAL_AND 24405) (T_LOGICAL_OR 24406) (T_LOGICAL_XOR 24407) (T_POW 24408) (T_SL 24409) (T_SPACESHIP 24410) (T_SR 24411) ("^" 24412) ("|" 24413)) 11293 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 24671)) 11294 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 24659)) 11295 (("(" 24570) (argument_list 24658)) 11296 (("=" 24578) (T_AND_EQUAL 24579) (T_COALESCE_EQUAL 24580) (T_CONCAT_EQUAL 24581) (T_DEC 24582) (T_DIV_EQUAL 24583) (T_INC 24584) (T_MINUS_EQUAL 24585) (T_MOD_EQUAL 24586) (T_MUL_EQUAL 24587) (T_OR_EQUAL 24588) (T_PLUS_EQUAL 24589) (T_POW_EQUAL 24590) (T_SL_EQUAL 24591) (T_SR_EQUAL 24592) (T_XOR_EQUAL 24593)) 11297 ((T_PAAMAYIM_NEKUDOTAYIM 24482)) 11298 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24383) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11299 (("%" 24384) ("*" 24385) ("+" 24386) ("-" 24387) ("." 24388) ("/" 24389) ("<" 24390) (">" 24391) ("?" 24392) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 24393) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 24394) (T_BOOLEAN_AND 24395) (T_BOOLEAN_OR 24396) (T_COALESCE 24397) (T_INSTANCEOF 24398) (T_IS_EQUAL 24399) (T_IS_GREATER_OR_EQUAL 24400) (T_IS_IDENTICAL 24401) (T_IS_NOT_EQUAL 24402) (T_IS_NOT_IDENTICAL 24403) (T_IS_SMALLER_OR_EQUAL 24404) (T_LOGICAL_AND 24405) (T_LOGICAL_OR 24406) (T_LOGICAL_XOR 24407) (T_POW 24408) (T_SL 24409) (T_SPACESHIP 24410) (T_SR 24411) ("^" 24412) ("|" 24413)) 11300 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24481) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11301 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24480) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11302 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24479) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11303 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24478) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11304 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24477) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11305 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24476) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11306 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24475) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11307 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24474) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11308 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 24469) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 24470) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 11309 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24468) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11310 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24467) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11311 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24466) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11312 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24465) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11313 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24464) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11314 (("$" 24429) ("(" 24430) (T_NAME_FULLY_QUALIFIED 24431) (T_NAME_QUALIFIED 24432) (T_NAME_RELATIVE 24433) (T_STATIC 24434) (T_STRING 24435) (T_VARIABLE 24436) (class_name 24437) (class_name_reference 24438) (name 24439) (new_variable 24440) (simple_variable 24441)) 11315 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24428) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11316 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24427) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11317 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24426) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11318 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24425) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11319 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24424) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11320 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24423) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11321 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24422) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11322 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24421) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11323 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24420) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11324 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24419) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11325 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24418) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11326 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24417) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11327 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24416) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11328 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24415) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11329 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24414) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11330 (("$" 24429) (T_VARIABLE 24436) (simple_variable 24460) ("{" 24461)) 11331 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24458) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11332 ((T_PAAMAYIM_NEKUDOTAYIM 24456)) 11333 ((T_NULLSAFE_OBJECT_OPERATOR 24442) (T_OBJECT_OPERATOR 24443) (T_PAAMAYIM_NEKUDOTAYIM 24444) ("[" 24445)) 11334 (("$" 24429) (T_STRING 24449) (T_VARIABLE 24436) (property_name 24455) (simple_variable 24451) ("{" 24452)) 11335 (("$" 24429) (T_STRING 24449) (T_VARIABLE 24436) (property_name 24450) (simple_variable 24451) ("{" 24452)) 11336 (("$" 24429) (T_VARIABLE 24436) (simple_variable 24448)) 11337 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 24446) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 11338 (("]" 24447)) 11339 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 24453) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11340 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 24454)) 11341 (("$" 24429) (T_VARIABLE 24436) (simple_variable 24457)) 11342 (("%" 299) (")" 24459) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11343 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 24462) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11344 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 24463)) 11345 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24473) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11346 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 24471) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 11347 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24472) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11348 (("$" 24306) (T_ABSTRACT 24483) (T_ARRAY 24484) (T_AS 24485) (T_BREAK 24486) (T_CALLABLE 24487) (T_CASE 24488) (T_CATCH 24489) (T_CLASS 24490) (T_CLASS_C 24491) (T_CLONE 24492) (T_CONST 24493) (T_CONTINUE 24494) (T_DECLARE 24495) (T_DEFAULT 24496) (T_DIR 24497) (T_DO 24498) (T_ECHO 24499) (T_ELSE 24500) (T_ELSEIF 24501) (T_EMPTY 24502) (T_ENDDECLARE 24503) (T_ENDFOR 24504) (T_ENDFOREACH 24505) (T_ENDIF 24506) (T_ENDSWITCH 24507) (T_ENDWHILE 24508) (T_ENUM 24509) (T_EVAL 24510) (T_EXIT 24511) (T_EXTENDS 24512) (T_FILE 24513) (T_FINAL 24514) (T_FINALLY 24515) (T_FN 24516) (T_FOR 24517) (T_FOREACH 24518) (T_FUNCTION 24519) (T_FUNC_C 24520) (T_GLOBAL 24521) (T_GOTO 24522) (T_IF 24523) (T_IMPLEMENTS 24524) (T_INCLUDE 24525) (T_INCLUDE_ONCE 24526) (T_INSTANCEOF 24527) (T_INSTEADOF 24528) (T_INTERFACE 24529) (T_ISSET 24530) (T_LINE 24531) (T_LIST 24532) (T_LOGICAL_AND 24533) (T_LOGICAL_OR 24534) (T_LOGICAL_XOR 24535) (T_MATCH 24536) (T_METHOD_C 24537) (T_NAMESPACE 24538) (T_NEW 24539) (T_NS_C 24540) (T_PRINT 24541) (T_PRIVATE 24542) (T_PROPERTY_C 24543) (T_PROTECTED 24544) (T_PUBLIC 24545) (T_READONLY 24546) (T_REQUIRE 24547) (T_REQUIRE_ONCE 24548) (T_RETURN 24549) (T_STATIC 24550) (T_STRING 24551) (T_SWITCH 24552) (T_THROW 24553) (T_TRAIT 24554) (T_TRAIT_C 24555) (T_TRY 24556) (T_UNSET 24557) (T_USE 24558) (T_VAR 24559) (T_VARIABLE 24354) (T_WHILE 24560) (T_YIELD 24561) (identifier 24562) (member_name 24563) (reserved_non_modifiers 24564) (semi_reserved 24565) (simple_variable 24566) ("{" 24567)) 11349 (("(" 24570) (argument_list 24571)) 11350 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 24568) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11351 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 24569)) 11352 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 24572) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 24573) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 24574) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11353 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 24577) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11354 (("," 1480) (possible_comma 24575)) 11355 ((")" 24576)) 11356 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (ampersand 24607) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24608) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11357 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24606) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11358 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24605) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11359 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24604) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11360 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24603) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11361 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24602) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11362 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24601) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11363 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24600) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11364 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24599) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11365 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24598) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11366 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24597) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11367 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24596) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11368 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24595) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11369 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24594) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11370 (("\"" 1520) ("$" 24609) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 24610) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 24611) ("[" 1542) (array_object_dereferenceable 24612) (callable_expr 24613) (callable_variable 24614) (class_constant 1546) (class_name 24615) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 24616) (name 24617) (new_dereferenceable 1552) (simple_variable 24618) (static_member 24619) (variable 24620) (variable_class_name 24621)) 11371 (("$" 24609) (T_VARIABLE 24611) (simple_variable 24654) ("{" 24655)) 11372 (("(" 24625) (argument_list 24653)) 11373 ((T_NULLSAFE_OBJECT_OPERATOR 24639) (T_OBJECT_OPERATOR 24640) ("[" 24641)) 11374 (("(" 24625) (argument_list 24638)) 11375 ((T_PAAMAYIM_NEKUDOTAYIM 24634)) 11376 (("(" 24625) (argument_list 24633)) 11377 ((T_PAAMAYIM_NEKUDOTAYIM 24622)) 11378 (("$" 24609) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 24611) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 24623) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 24624) ("{" 1642)) 11379 (("(" 24625) (argument_list 24626)) 11380 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 24627) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 24628) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 24629) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11381 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 24632) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11382 (("," 1480) (possible_comma 24630)) 11383 ((")" 24631)) 11384 (("$" 24609) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 24611) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 24635) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 24636) ("{" 1658)) 11385 (("(" 24625) (argument_list 24637)) 11386 (("$" 24609) (T_STRING 24644) (T_VARIABLE 24611) (property_name 24651) (simple_variable 24646) ("{" 24647)) 11387 (("$" 24609) (T_STRING 24644) (T_VARIABLE 24611) (property_name 24645) (simple_variable 24646) ("{" 24647)) 11388 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 24642) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 11389 (("]" 24643)) 11390 (("(" 24625) (argument_list 24650)) 11391 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 24648) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11392 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 24649)) 11393 (("(" 24625) (argument_list 24652)) 11394 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 24656) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11395 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 24657)) 11396 ((backup_doc_comment 24660)) 11397 (("(" 24661)) 11398 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 24662)) 11399 ((")" 24663)) 11400 ((T_USE 164) (lexical_vars 24664)) 11401 ((":" 166) (return_type 24665)) 11402 ((backup_fn_flags 24666)) 11403 (("{" 24667)) 11404 ((inner_statement_list 24668)) 11405 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 24669) ("~" 127)) 11406 ((backup_fn_flags 24670)) 11407 ((backup_doc_comment 24672)) 11408 (("(" 24673)) 11409 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 24674)) 11410 ((")" 24675)) 11411 ((":" 2249) (return_type 24676)) 11412 ((T_DOUBLE_ARROW 24677)) 11413 ((backup_fn_flags 24678)) 11414 ((backup_lex_pos 24679)) 11415 (("!" 23643) ("\"" 23644) ("$" 23645) ("(" 23646) ("+" 23647) ("-" 23648) ("@" 23649) (T_ARRAY 23650) (T_ARRAY_CAST 23651) (T_ATTRIBUTE 128) (T_BOOL_CAST 23652) (T_CLASS_C 23653) (T_CLONE 23654) (T_CONSTANT_ENCAPSED_STRING 23655) (T_DEC 23656) (T_DIR 23657) (T_DNUMBER 23658) (T_DOUBLE_CAST 23659) (T_EMPTY 23660) (T_EVAL 23661) (T_EXIT 23662) (T_FILE 23663) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 23664) (T_INC 23665) (T_INCLUDE 23666) (T_INCLUDE_ONCE 23667) (T_INT_CAST 23668) (T_ISSET 23669) (T_LINE 23670) (T_LIST 23671) (T_LNUMBER 23672) (T_MATCH 23673) (T_METHOD_C 23674) (T_NAME_FULLY_QUALIFIED 23675) (T_NAME_QUALIFIED 23676) (T_NAME_RELATIVE 23677) (T_NEW 23678) (T_NS_C 23679) (T_OBJECT_CAST 23680) (T_PRINT 23681) (T_PROPERTY_C 23682) (T_READONLY 23683) (T_REQUIRE 23684) (T_REQUIRE_ONCE 23685) (T_START_HEREDOC 23686) (T_STATIC 23687) (T_STRING 23688) (T_STRING_CAST 23689) (T_THROW 23690) (T_TRAIT_C 23691) (T_UNSET_CAST 23692) (T_VARIABLE 23693) (T_YIELD 23694) (T_YIELD_FROM 23695) ("[" 23696) ("`" 23697) (array_object_dereferenceable 23698) (attribute 133) (attributes 23699) (callable_expr 23700) (callable_variable 23701) (class_constant 23702) (class_name 23703) (constant 23704) (dereferenceable_scalar 23705) (expr 24680) (fn 23707) (fully_dereferenceable 105) #'23708 (function_call 23709) (inline_function 23710) (internal_functions_in_yacc 23711) (match 23712) (name 23713) (new_dereferenceable 23714) (new_non_dereferenceable 23715) (scalar 23716) (simple_variable 23717) (static_member 23718) (variable 23719) (variable_class_name 23720) ("~" 23721)) 11416 (("%" 23723) ("*" 23724) ("+" 23725) ("-" 23726) ("." 23727) ("/" 23728) ("<" 23729) (">" 23730) ("?" 23731) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 23732) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 23733) (T_BOOLEAN_AND 23734) (T_BOOLEAN_OR 23735) (T_COALESCE 23736) (T_INSTANCEOF 23737) (T_IS_EQUAL 23738) (T_IS_GREATER_OR_EQUAL 23739) (T_IS_IDENTICAL 23740) (T_IS_NOT_EQUAL 23741) (T_IS_NOT_IDENTICAL 23742) (T_IS_SMALLER_OR_EQUAL 23743) (T_LOGICAL_AND 23744) (T_LOGICAL_OR 23745) (T_LOGICAL_XOR 23746) (T_POW 23747) (T_SL 23748) (T_SPACESHIP 23749) (T_SR 23750) ("^" 23751) (backup_fn_flags 24681) ("|" 23752)) 11417 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24683) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 11418 (("$" 24306) (T_ABSTRACT 24483) (T_ARRAY 24484) (T_AS 24485) (T_BREAK 24486) (T_CALLABLE 24487) (T_CASE 24488) (T_CATCH 24489) (T_CLASS 24490) (T_CLASS_C 24491) (T_CLONE 24492) (T_CONST 24493) (T_CONTINUE 24494) (T_DECLARE 24495) (T_DEFAULT 24496) (T_DIR 24497) (T_DO 24498) (T_ECHO 24499) (T_ELSE 24500) (T_ELSEIF 24501) (T_EMPTY 24502) (T_ENDDECLARE 24503) (T_ENDFOR 24504) (T_ENDFOREACH 24505) (T_ENDIF 24506) (T_ENDSWITCH 24507) (T_ENDWHILE 24508) (T_ENUM 24509) (T_EVAL 24510) (T_EXIT 24511) (T_EXTENDS 24512) (T_FILE 24513) (T_FINAL 24514) (T_FINALLY 24515) (T_FN 24516) (T_FOR 24517) (T_FOREACH 24518) (T_FUNCTION 24519) (T_FUNC_C 24520) (T_GLOBAL 24521) (T_GOTO 24522) (T_IF 24523) (T_IMPLEMENTS 24524) (T_INCLUDE 24525) (T_INCLUDE_ONCE 24526) (T_INSTANCEOF 24527) (T_INSTEADOF 24528) (T_INTERFACE 24529) (T_ISSET 24530) (T_LINE 24531) (T_LIST 24532) (T_LOGICAL_AND 24533) (T_LOGICAL_OR 24534) (T_LOGICAL_XOR 24535) (T_MATCH 24536) (T_METHOD_C 24537) (T_NAMESPACE 24538) (T_NEW 24539) (T_NS_C 24540) (T_PRINT 24541) (T_PRIVATE 24542) (T_PROPERTY_C 24543) (T_PROTECTED 24544) (T_PUBLIC 24545) (T_READONLY 24546) (T_REQUIRE 24547) (T_REQUIRE_ONCE 24548) (T_RETURN 24549) (T_STATIC 24550) (T_STRING 24551) (T_SWITCH 24552) (T_THROW 24553) (T_TRAIT 24554) (T_TRAIT_C 24555) (T_TRY 24556) (T_UNSET 24557) (T_USE 24558) (T_VAR 24559) (T_VARIABLE 24354) (T_WHILE 24560) (T_YIELD 24561) (identifier 24685) (member_name 24686) (reserved_non_modifiers 24564) (semi_reserved 24565) (simple_variable 24687) ("{" 24688)) 11419 (("(" 24570) (argument_list 24691)) 11420 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 24689) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11421 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 24690)) 11422 ((T_FN 35) (T_FUNCTION 129) (fn 24368) #'24369 (inline_function 24695)) 11423 (("$" 24306) (T_STRING 24701) (T_VARIABLE 24354) (property_name 24708) (simple_variable 24703) ("{" 24704)) 11424 (("$" 24306) (T_STRING 24701) (T_VARIABLE 24354) (property_name 24702) (simple_variable 24703) ("{" 24704)) 11425 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 24699) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 11426 (("]" 24700)) 11427 (("(" 24570) (argument_list 24707)) 11428 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 24705) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11429 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 24706)) 11430 (("(" 24570) (argument_list 24709)) 11431 (("`" 24711)) 11432 (("]" 24713)) 11433 (("=" 24714)) 11434 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24715) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11435 (("%" 24384) ("*" 24385) ("+" 24386) ("-" 24387) ("." 24388) ("/" 24389) ("<" 24390) (">" 24391) ("?" 24392) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 24393) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 24394) (T_BOOLEAN_AND 24395) (T_BOOLEAN_OR 24396) (T_COALESCE 24397) (T_DOUBLE_ARROW 24718) (T_INSTANCEOF 24398) (T_IS_EQUAL 24399) (T_IS_GREATER_OR_EQUAL 24400) (T_IS_IDENTICAL 24401) (T_IS_NOT_EQUAL 24402) (T_IS_NOT_IDENTICAL 24403) (T_IS_SMALLER_OR_EQUAL 24404) (T_LOGICAL_AND 24405) (T_LOGICAL_OR 24406) (T_LOGICAL_XOR 24407) (T_POW 24408) (T_SL 24409) (T_SPACESHIP 24410) (T_SR 24411) ("^" 24412) ("|" 24413)) 11436 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24719) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11437 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 24728) (T_VARIABLE 2896) (encaps_var 2910)) 11438 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 24727) (T_VARIABLE 2896) (encaps_var 2901)) 11439 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24769) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11440 ((T_CLASS 24761)) 11441 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 24760) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 24741) (attribute 2958) (class_modifier 2939)) 11442 ((T_PAAMAYIM_NEKUDOTAYIM 24758)) 11443 (("(" 24625) (argument_list 24757)) 11444 ((T_NULLSAFE_OBJECT_OPERATOR 24748) (T_OBJECT_OPERATOR 24749) (T_PAAMAYIM_NEKUDOTAYIM 24750) ("[" 24751)) 11445 (("$" 24609) (T_STRING 24644) (T_VARIABLE 24611) (property_name 24756) (simple_variable 24646) ("{" 24647)) 11446 (("$" 24609) (T_STRING 24644) (T_VARIABLE 24611) (property_name 24755) (simple_variable 24646) ("{" 24647)) 11447 (("$" 24609) (T_VARIABLE 24611) (simple_variable 24754)) 11448 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 24752) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 11449 (("]" 24753)) 11450 (("$" 24609) (T_VARIABLE 24611) (simple_variable 24759)) 11451 (("(" 2960) (argument_list 2961) (ctor_arguments 24762)) 11452 ((T_EXTENDS 2963) (extends_from 24763)) 11453 ((T_IMPLEMENTS 2965) (implements_list 24764)) 11454 ((backup_doc_comment 24765)) 11455 (("{" 24766)) 11456 ((class_statement_list 24767)) 11457 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 24768)) 11458 (("%" 299) (")" 24770) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11459 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24772) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11460 (("%" 299) (")" 24773) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11461 (("{" 24774)) 11462 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 24775) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 11463 (("}" 24776)) 11464 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 24778) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 11465 ((")" 24779)) 11466 (("=" 24780)) 11467 (("!" 24304) ("\"" 24305) ("$" 24306) ("(" 24307) ("+" 24308) ("-" 24309) ("@" 24310) (T_ARRAY 24311) (T_ARRAY_CAST 24312) (T_ATTRIBUTE 128) (T_BOOL_CAST 24313) (T_CLASS_C 24314) (T_CLONE 24315) (T_CONSTANT_ENCAPSED_STRING 24316) (T_DEC 24317) (T_DIR 24318) (T_DNUMBER 24319) (T_DOUBLE_CAST 24320) (T_EMPTY 24321) (T_EVAL 24322) (T_EXIT 24323) (T_FILE 24324) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 24325) (T_INC 24326) (T_INCLUDE 24327) (T_INCLUDE_ONCE 24328) (T_INT_CAST 24329) (T_ISSET 24330) (T_LINE 24331) (T_LIST 24332) (T_LNUMBER 24333) (T_MATCH 24334) (T_METHOD_C 24335) (T_NAME_FULLY_QUALIFIED 24336) (T_NAME_QUALIFIED 24337) (T_NAME_RELATIVE 24338) (T_NEW 24339) (T_NS_C 24340) (T_OBJECT_CAST 24341) (T_PRINT 24342) (T_PROPERTY_C 24343) (T_READONLY 24344) (T_REQUIRE 24345) (T_REQUIRE_ONCE 24346) (T_START_HEREDOC 24347) (T_STATIC 24348) (T_STRING 24349) (T_STRING_CAST 24350) (T_THROW 24351) (T_TRAIT_C 24352) (T_UNSET_CAST 24353) (T_VARIABLE 24354) (T_YIELD 24355) (T_YIELD_FROM 24356) ("[" 24357) ("`" 24358) (array_object_dereferenceable 24359) (attribute 133) (attributes 24360) (callable_expr 24361) (callable_variable 24362) (class_constant 24363) (class_name 24364) (constant 24365) (dereferenceable_scalar 24366) (expr 24781) (fn 24368) (fully_dereferenceable 105) #'24369 (function_call 24370) (inline_function 24371) (internal_functions_in_yacc 24372) (match 24373) (name 24374) (new_dereferenceable 24375) (new_non_dereferenceable 24376) (scalar 24377) (simple_variable 24378) (static_member 24379) (variable 24380) (variable_class_name 24381) ("~" 24382)) 11468 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 24783) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11469 (("," 5192) (possible_comma 24784)) 11470 ((")" 24785)) 11471 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 24793) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 24794) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 24795) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11472 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 24798) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11473 (("," 1480) (possible_comma 24796)) 11474 ((")" 24797)) 11475 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24800) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11476 (("%" 299) (")" 24801) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11477 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24803) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11478 (("%" 299) (")" 24804) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11479 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 24811) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 11480 ((")" 24812)) 11481 (("%" 299) (")" 24817) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11482 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 24820) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11483 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 24821)) 11484 (("\"" 24823) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 11485 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 24833) (T_VARIABLE 2896) (encaps_var 2910)) 11486 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 24832) (T_VARIABLE 2896) (encaps_var 2901)) 11487 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24874) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11488 ((T_CLASS 24866)) 11489 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 24865) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 24846) (attribute 2958) (class_modifier 2939)) 11490 ((T_PAAMAYIM_NEKUDOTAYIM 24863)) 11491 (("(" 23176) (argument_list 24862)) 11492 ((T_NULLSAFE_OBJECT_OPERATOR 24853) (T_OBJECT_OPERATOR 24854) (T_PAAMAYIM_NEKUDOTAYIM 24855) ("[" 24856)) 11493 (("$" 23160) (T_STRING 23195) (T_VARIABLE 23162) (property_name 24861) (simple_variable 23197) ("{" 23198)) 11494 (("$" 23160) (T_STRING 23195) (T_VARIABLE 23162) (property_name 24860) (simple_variable 23197) ("{" 23198)) 11495 (("$" 23160) (T_VARIABLE 23162) (simple_variable 24859)) 11496 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 24857) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 11497 (("]" 24858)) 11498 (("$" 23160) (T_VARIABLE 23162) (simple_variable 24864)) 11499 (("(" 2960) (argument_list 2961) (ctor_arguments 24867)) 11500 ((T_EXTENDS 2963) (extends_from 24868)) 11501 ((T_IMPLEMENTS 2965) (implements_list 24869)) 11502 ((backup_doc_comment 24870)) 11503 (("{" 24871)) 11504 ((class_statement_list 24872)) 11505 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 24873)) 11506 (("%" 299) (")" 24875) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11507 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24877) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11508 (("%" 299) (")" 24878) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11509 (("{" 24879)) 11510 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 24880) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 11511 (("}" 24881)) 11512 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 24883) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 11513 ((")" 24884)) 11514 (("=" 24885)) 11515 (("!" 22855) ("\"" 22856) ("$" 22857) ("(" 22858) ("+" 22859) ("-" 22860) ("@" 22861) (T_ARRAY 22862) (T_ARRAY_CAST 22863) (T_ATTRIBUTE 128) (T_BOOL_CAST 22864) (T_CLASS_C 22865) (T_CLONE 22866) (T_CONSTANT_ENCAPSED_STRING 22867) (T_DEC 22868) (T_DIR 22869) (T_DNUMBER 22870) (T_DOUBLE_CAST 22871) (T_EMPTY 22872) (T_EVAL 22873) (T_EXIT 22874) (T_FILE 22875) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 22876) (T_INC 22877) (T_INCLUDE 22878) (T_INCLUDE_ONCE 22879) (T_INT_CAST 22880) (T_ISSET 22881) (T_LINE 22882) (T_LIST 22883) (T_LNUMBER 22884) (T_MATCH 22885) (T_METHOD_C 22886) (T_NAME_FULLY_QUALIFIED 22887) (T_NAME_QUALIFIED 22888) (T_NAME_RELATIVE 22889) (T_NEW 22890) (T_NS_C 22891) (T_OBJECT_CAST 22892) (T_PRINT 22893) (T_PROPERTY_C 22894) (T_READONLY 22895) (T_REQUIRE 22896) (T_REQUIRE_ONCE 22897) (T_START_HEREDOC 22898) (T_STATIC 22899) (T_STRING 22900) (T_STRING_CAST 22901) (T_THROW 22902) (T_TRAIT_C 22903) (T_UNSET_CAST 22904) (T_VARIABLE 22905) (T_YIELD 22906) (T_YIELD_FROM 22907) ("[" 22908) ("`" 22909) (array_object_dereferenceable 22910) (attribute 133) (attributes 22911) (callable_expr 22912) (callable_variable 22913) (class_constant 22914) (class_name 22915) (constant 22916) (dereferenceable_scalar 22917) (expr 24886) (fn 22919) (fully_dereferenceable 105) #'22920 (function_call 22921) (inline_function 22922) (internal_functions_in_yacc 22923) (match 22924) (name 22925) (new_dereferenceable 22926) (new_non_dereferenceable 22927) (scalar 22928) (simple_variable 22929) (static_member 22930) (variable 22931) (variable_class_name 22932) ("~" 22933)) 11516 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 24888) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11517 (("," 5192) (possible_comma 24889)) 11518 ((")" 24890)) 11519 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 24898) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 24899) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 24900) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11520 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 24903) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 11521 (("," 1480) (possible_comma 24901)) 11522 ((")" 24902)) 11523 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24905) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11524 (("%" 299) (")" 24906) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11525 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24908) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11526 (("%" 299) (")" 24909) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11527 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 24916) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 11528 ((")" 24917)) 11529 (("%" 299) (")" 24922) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11530 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 24925) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 11531 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 24926)) 11532 (("\"" 24928) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 11533 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 24931)) 11534 ((";" 24932)) 11535 (("," 17937) (";" 24935)) 11536 ((";" 24937)) 11537 ((";" 24939)) 11538 (("," 18988) (";" 24941)) 11539 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 24943) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 11540 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 24944) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 11541 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 24945) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 11542 ((")" 24946) (T_DOUBLE_ARROW 24947)) 11543 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24950) (";" 17709) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17710) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17711) (T_DEC 22) (T_DECLARE 17712) (T_DIR 24) (T_DNUMBER 25) (T_DO 17713) (T_DOUBLE_CAST 27) (T_ECHO 17714) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17715) (T_FOREACH 17716) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17717) (T_GOTO 17718) (T_IF 17674) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17719) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17720) (T_START_HEREDOC 69) (T_STATIC 17721) (T_STRING 17722) (T_STRING_CAST 72) (T_SWITCH 17723) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17724) (T_UNSET 17725) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17726) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17727) (alt_if_stmt_without_else 17728) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17729) (fn 104) (foreach_statement 24956) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17730) (if_stmt_without_else 17731) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 24952) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17733) ("~" 127)) 11544 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 24948) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 11545 ((")" 24949)) 11546 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24950) (";" 17709) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17710) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17711) (T_DEC 22) (T_DECLARE 17712) (T_DIR 24) (T_DNUMBER 25) (T_DO 17713) (T_DOUBLE_CAST 27) (T_ECHO 17714) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17715) (T_FOREACH 17716) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17717) (T_GOTO 17718) (T_IF 17674) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17719) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17720) (T_START_HEREDOC 69) (T_STATIC 17721) (T_STRING 17722) (T_STRING_CAST 72) (T_SWITCH 17723) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17724) (T_UNSET 17725) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17726) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17727) (alt_if_stmt_without_else 17728) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17729) (fn 104) (foreach_statement 24951) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17730) (if_stmt_without_else 17731) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 24952) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17733) ("~" 127)) 11547 ((inner_statement_list 24953)) 11548 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 24954) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 11549 ((";" 24955)) 11550 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 24958) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 11551 ((";" 24959)) 11552 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 24960) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 11553 ((";" 24961)) 11554 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 24962) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 11555 ((")" 24963)) 11556 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24964) (";" 17709) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17710) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17711) (T_DEC 22) (T_DECLARE 17712) (T_DIR 24) (T_DNUMBER 25) (T_DO 17713) (T_DOUBLE_CAST 27) (T_ECHO 17714) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17715) (T_FOREACH 17716) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17717) (T_GOTO 17718) (T_IF 17674) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17719) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17720) (T_START_HEREDOC 69) (T_STATIC 17721) (T_STRING 17722) (T_STRING_CAST 72) (T_SWITCH 17723) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17724) (T_UNSET 17725) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17726) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17727) (alt_if_stmt_without_else 17728) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17729) (fn 104) (for_statement 24965) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17730) (if_stmt_without_else 17731) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 24966) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17733) ("~" 127)) 11557 ((inner_statement_list 24967)) 11558 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 24968) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 11559 ((";" 24969)) 11560 (("," 20829) (";" 24971)) 11561 ((T_WHILE 24973)) 11562 (("(" 24974)) 11563 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 24975) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11564 (("%" 299) (")" 24976) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11565 ((";" 24977)) 11566 ((T_STRING 21002) (const_decl 21003) (const_list 24979)) 11567 ((")" 24980) ("," 21006)) 11568 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24981) (";" 17709) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17710) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17711) (T_DEC 22) (T_DECLARE 17712) (T_DIR 24) (T_DNUMBER 25) (T_DO 17713) (T_DOUBLE_CAST 27) (T_ECHO 17714) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17715) (T_FOREACH 17716) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17717) (T_GOTO 17718) (T_IF 17674) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17719) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17720) (T_START_HEREDOC 69) (T_STATIC 17721) (T_STRING 17722) (T_STRING_CAST 72) (T_SWITCH 17723) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17724) (T_UNSET 17725) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17726) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17727) (alt_if_stmt_without_else 17728) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 24982) (dereferenceable_scalar 101) (expr 17729) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17730) (if_stmt_without_else 17731) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 24983) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17733) ("~" 127)) 11569 ((inner_statement_list 24984)) 11570 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 24985) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 11571 ((";" 24986)) 11572 ((";" 24988)) 11573 ((";" 24990)) 11574 ((T_READONLY 17353) (T_STRING 17354) (backup_doc_comment 141) (function_name 24993)) 11575 ((backup_doc_comment 24994)) 11576 (("(" 24995)) 11577 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 24996)) 11578 ((")" 24997)) 11579 ((":" 166) (return_type 24998)) 11580 ((backup_fn_flags 24999)) 11581 (("{" 25000)) 11582 ((inner_statement_list 25001)) 11583 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25002) ("~" 127)) 11584 ((backup_fn_flags 25003)) 11585 ((T_STRING 25006)) 11586 ((T_EXTENDS 2963) (extends_from 25007)) 11587 ((T_IMPLEMENTS 2965) (implements_list 25008)) 11588 ((backup_doc_comment 25009)) 11589 (("{" 25010)) 11590 ((class_statement_list 25011)) 11591 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25012)) 11592 ((":" 25017)) 11593 ((";" 25016)) 11594 ((inner_statement_list 25018)) 11595 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 25019) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 11596 ((";" 25020)) 11597 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25022) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11598 (("%" 299) (")" 25023) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11599 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25024) (";" 17661) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17662) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17664) (T_DEC 22) (T_DECLARE 17665) (T_DIR 24) (T_DNUMBER 25) (T_DO 17666) (T_DOUBLE_CAST 27) (T_ECHO 17667) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17669) (T_FOREACH 17670) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17671) (T_GOTO 17672) (T_IF 17674) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17675) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17677) (T_START_HEREDOC 69) (T_STATIC 17678) (T_STRING 17679) (T_STRING_CAST 72) (T_SWITCH 17680) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17682) (T_UNSET 17683) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17684) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17685) (alt_if_stmt_without_else 17686) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17692) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17695) (if_stmt_without_else 17696) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25025) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 25026) ("{" 17701) ("~" 127)) 11600 ((inner_statement_list 25027)) 11601 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 25028) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 11602 ((";" 25029)) 11603 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 25031) (variable 17576) (variable_class_name 17577)) 11604 (("," 17589) (possible_comma 25032)) 11605 ((")" 25033)) 11606 ((";" 25034)) 11607 ((inner_statement_list 25036)) 11608 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25037) ("~" 127)) 11609 ((catch_list 25038)) 11610 ((T_CATCH 25039) (T_FINALLY 25040) (finally_statement 25041)) 11611 (("(" 25045)) 11612 (("{" 25042)) 11613 ((inner_statement_list 25043)) 11614 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25044) ("~" 127)) 11615 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 25046) (class_name 17636) (name 17637)) 11616 ((T_VARIABLE 17638) (optional_variable 25047) ("|" 17640)) 11617 ((")" 25048)) 11618 (("{" 25049)) 11619 ((inner_statement_list 25050)) 11620 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25051) ("~" 127)) 11621 ((backup_doc_comment 25053)) 11622 (("{" 25054)) 11623 ((class_statement_list 25055)) 11624 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25056)) 11625 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25058) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11626 (("%" 299) (")" 25059) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11627 ((":" 25060) (switch_case_list 25061) ("{" 25062)) 11628 ((";" 25068) (case_list 25069)) 11629 ((";" 25063) (case_list 25064)) 11630 ((case_list 25066)) 11631 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 25065)) 11632 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 25067)) 11633 ((case_list 25072)) 11634 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 25070)) 11635 ((";" 25071)) 11636 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 25073)) 11637 ((";" 25074)) 11638 (("," 17937) (";" 25077)) 11639 ((";" 25079)) 11640 ((T_EXTENDS 21295) (interface_extends_list 25081)) 11641 ((backup_doc_comment 25082)) 11642 (("{" 25083)) 11643 ((class_statement_list 25084)) 11644 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25085)) 11645 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25087) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11646 (("%" 299) (")" 25088) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11647 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21305) (";" 17709) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17710) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17711) (T_DEC 22) (T_DECLARE 17712) (T_DIR 24) (T_DNUMBER 25) (T_DO 17713) (T_DOUBLE_CAST 27) (T_ECHO 17714) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17715) (T_FOREACH 17716) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17717) (T_GOTO 17718) (T_IF 17674) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17719) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17720) (T_START_HEREDOC 69) (T_STATIC 17721) (T_STRING 17722) (T_STRING_CAST 72) (T_SWITCH 17723) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17724) (T_UNSET 17725) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17726) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17727) (alt_if_stmt_without_else 17728) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17729) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17730) (if_stmt_without_else 17731) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25089) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17733) ("~" 127)) 11648 ((")" 25091)) 11649 ((";" 25092)) 11650 ((";" 25094)) 11651 (("," 18988) (";" 25096)) 11652 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 25098) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 11653 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 25099) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 11654 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 25100) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 11655 ((")" 25101) (T_DOUBLE_ARROW 25102)) 11656 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25105) (";" 17661) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17662) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17664) (T_DEC 22) (T_DECLARE 17665) (T_DIR 24) (T_DNUMBER 25) (T_DO 17666) (T_DOUBLE_CAST 27) (T_ECHO 17667) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17669) (T_FOREACH 17670) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17671) (T_GOTO 17672) (T_IF 17674) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17675) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17677) (T_START_HEREDOC 69) (T_STATIC 17678) (T_STRING 17679) (T_STRING_CAST 72) (T_SWITCH 17680) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17682) (T_UNSET 17683) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17684) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17685) (alt_if_stmt_without_else 17686) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17692) (fn 104) (foreach_statement 25111) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17695) (if_stmt_without_else 17696) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25107) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17701) ("~" 127)) 11657 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 25103) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 11658 ((")" 25104)) 11659 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25105) (";" 17661) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17662) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17664) (T_DEC 22) (T_DECLARE 17665) (T_DIR 24) (T_DNUMBER 25) (T_DO 17666) (T_DOUBLE_CAST 27) (T_ECHO 17667) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17669) (T_FOREACH 17670) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17671) (T_GOTO 17672) (T_IF 17674) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17675) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17677) (T_START_HEREDOC 69) (T_STATIC 17678) (T_STRING 17679) (T_STRING_CAST 72) (T_SWITCH 17680) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17682) (T_UNSET 17683) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17684) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17685) (alt_if_stmt_without_else 17686) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17692) (fn 104) (foreach_statement 25106) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17695) (if_stmt_without_else 17696) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25107) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17701) ("~" 127)) 11660 ((inner_statement_list 25108)) 11661 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 25109) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 11662 ((";" 25110)) 11663 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 25113) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 11664 ((";" 25114)) 11665 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 25115) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 11666 ((";" 25116)) 11667 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 25117) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 11668 ((")" 25118)) 11669 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25119) (";" 17661) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17662) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17664) (T_DEC 22) (T_DECLARE 17665) (T_DIR 24) (T_DNUMBER 25) (T_DO 17666) (T_DOUBLE_CAST 27) (T_ECHO 17667) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17669) (T_FOREACH 17670) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17671) (T_GOTO 17672) (T_IF 17674) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17675) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17677) (T_START_HEREDOC 69) (T_STATIC 17678) (T_STRING 17679) (T_STRING_CAST 72) (T_SWITCH 17680) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17682) (T_UNSET 17683) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17684) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17685) (alt_if_stmt_without_else 17686) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17692) (fn 104) (for_statement 25120) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17695) (if_stmt_without_else 17696) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25121) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17701) ("~" 127)) 11670 ((inner_statement_list 25122)) 11671 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 25123) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 11672 ((";" 25124)) 11673 ((":" 21344) (enum_backing_type 25126)) 11674 ((T_IMPLEMENTS 2965) (implements_list 25127)) 11675 ((backup_doc_comment 25128)) 11676 (("{" 25129)) 11677 ((class_statement_list 25130)) 11678 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25131)) 11679 (("," 20829) (";" 25133)) 11680 ((T_WHILE 25135)) 11681 (("(" 25136)) 11682 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25137) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11683 (("%" 299) (")" 25138) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11684 ((";" 25139)) 11685 ((T_STRING 21002) (const_decl 21003) (const_list 25141)) 11686 ((")" 25142) ("," 21006)) 11687 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25143) (";" 17661) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17662) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17664) (T_DEC 22) (T_DECLARE 17665) (T_DIR 24) (T_DNUMBER 25) (T_DO 17666) (T_DOUBLE_CAST 27) (T_ECHO 17667) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17669) (T_FOREACH 17670) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17671) (T_GOTO 17672) (T_IF 17674) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17675) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17677) (T_START_HEREDOC 69) (T_STATIC 17678) (T_STRING 17679) (T_STRING_CAST 72) (T_SWITCH 17680) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17682) (T_UNSET 17683) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17684) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17685) (alt_if_stmt_without_else 17686) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 25144) (dereferenceable_scalar 101) (expr 17692) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17695) (if_stmt_without_else 17696) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25145) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17701) ("~" 127)) 11688 ((inner_statement_list 25146)) 11689 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 25147) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 11690 ((";" 25148)) 11691 ((";" 25150)) 11692 ((T_EXTENDS 2963) (extends_from 25152)) 11693 ((T_IMPLEMENTS 2965) (implements_list 25153)) 11694 ((backup_doc_comment 25154)) 11695 (("{" 25155)) 11696 ((class_statement_list 25156)) 11697 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25157)) 11698 ((";" 25159)) 11699 (("%" 22935) ("*" 22936) ("+" 22937) ("-" 22938) ("." 22939) ("/" 22940) (":" 17657) (";" 17658) ("<" 22941) (">" 22942) ("?" 22943) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22944) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22945) (T_BOOLEAN_AND 22946) (T_BOOLEAN_OR 22947) (T_COALESCE 22948) (T_INSTANCEOF 22949) (T_IS_EQUAL 22950) (T_IS_GREATER_OR_EQUAL 22951) (T_IS_IDENTICAL 22952) (T_IS_NOT_EQUAL 22953) (T_IS_NOT_IDENTICAL 22954) (T_IS_SMALLER_OR_EQUAL 22955) (T_LOGICAL_AND 22956) (T_LOGICAL_OR 22957) (T_LOGICAL_XOR 22958) (T_POW 22959) (T_SL 22960) (T_SPACESHIP 22961) (T_SR 22962) ("^" 22963) (case_separator 25161) ("|" 22964)) 11700 ((inner_statement_list 25162)) 11701 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 25164)) 11702 ((case_list 25169)) 11703 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 25167)) 11704 ((";" 25168)) 11705 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 25170)) 11706 ((";" 25171)) 11707 (("," 17937) (";" 25174)) 11708 ((";" 25176)) 11709 ((";" 25178)) 11710 (("," 18988) (";" 25180)) 11711 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 25182) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 11712 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 25183) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 11713 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 25184) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 11714 ((")" 25185) (T_DOUBLE_ARROW 25186)) 11715 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25189) (";" 17516) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17517) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17518) (T_DEC 22) (T_DECLARE 17519) (T_DIR 24) (T_DNUMBER 25) (T_DO 17520) (T_DOUBLE_CAST 27) (T_ECHO 17521) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17522) (T_FOREACH 17523) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17524) (T_GOTO 17525) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17526) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17527) (T_START_HEREDOC 69) (T_STATIC 17528) (T_STRING 17529) (T_STRING_CAST 72) (T_SWITCH 17530) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17531) (T_UNSET 17532) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17533) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17534) (alt_if_stmt_without_else 17535) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17536) (fn 104) (foreach_statement 25195) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17537) (if_stmt_without_else 17538) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25191) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17540) ("~" 127)) 11716 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 25187) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 11717 ((")" 25188)) 11718 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25189) (";" 17516) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17517) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17518) (T_DEC 22) (T_DECLARE 17519) (T_DIR 24) (T_DNUMBER 25) (T_DO 17520) (T_DOUBLE_CAST 27) (T_ECHO 17521) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17522) (T_FOREACH 17523) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17524) (T_GOTO 17525) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17526) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17527) (T_START_HEREDOC 69) (T_STATIC 17528) (T_STRING 17529) (T_STRING_CAST 72) (T_SWITCH 17530) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17531) (T_UNSET 17532) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17533) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17534) (alt_if_stmt_without_else 17535) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17536) (fn 104) (foreach_statement 25190) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17537) (if_stmt_without_else 17538) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25191) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17540) ("~" 127)) 11719 ((inner_statement_list 25192)) 11720 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 25193) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 11721 ((";" 25194)) 11722 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 25197) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 11723 ((";" 25198)) 11724 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 25199) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 11725 ((";" 25200)) 11726 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 25201) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 11727 ((")" 25202)) 11728 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25203) (";" 17516) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17517) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17518) (T_DEC 22) (T_DECLARE 17519) (T_DIR 24) (T_DNUMBER 25) (T_DO 17520) (T_DOUBLE_CAST 27) (T_ECHO 17521) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17522) (T_FOREACH 17523) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17524) (T_GOTO 17525) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17526) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17527) (T_START_HEREDOC 69) (T_STATIC 17528) (T_STRING 17529) (T_STRING_CAST 72) (T_SWITCH 17530) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17531) (T_UNSET 17532) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17533) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17534) (alt_if_stmt_without_else 17535) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17536) (fn 104) (for_statement 25204) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17537) (if_stmt_without_else 17538) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25205) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17540) ("~" 127)) 11729 ((inner_statement_list 25206)) 11730 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 25207) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 11731 ((";" 25208)) 11732 (("," 20829) (";" 25210)) 11733 ((T_WHILE 25212)) 11734 (("(" 25213)) 11735 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25214) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11736 (("%" 299) (")" 25215) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11737 ((";" 25216)) 11738 ((T_STRING 21002) (const_decl 21003) (const_list 25218)) 11739 ((")" 25219) ("," 21006)) 11740 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25220) (";" 17516) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17517) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17518) (T_DEC 22) (T_DECLARE 17519) (T_DIR 24) (T_DNUMBER 25) (T_DO 17520) (T_DOUBLE_CAST 27) (T_ECHO 17521) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17522) (T_FOREACH 17523) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17524) (T_GOTO 17525) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17526) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17527) (T_START_HEREDOC 69) (T_STATIC 17528) (T_STRING 17529) (T_STRING_CAST 72) (T_SWITCH 17530) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17531) (T_UNSET 17532) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17533) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17534) (alt_if_stmt_without_else 17535) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 25221) (dereferenceable_scalar 101) (expr 17536) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17537) (if_stmt_without_else 17538) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25222) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17540) ("~" 127)) 11741 ((inner_statement_list 25223)) 11742 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 25224) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 11743 ((";" 25225)) 11744 ((";" 25227)) 11745 ((";" 25229)) 11746 ((T_READONLY 17353) (T_STRING 17354) (backup_doc_comment 141) (function_name 25232)) 11747 ((backup_doc_comment 25233)) 11748 (("(" 25234)) 11749 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 25235)) 11750 ((")" 25236)) 11751 ((":" 166) (return_type 25237)) 11752 ((backup_fn_flags 25238)) 11753 (("{" 25239)) 11754 ((inner_statement_list 25240)) 11755 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25241) ("~" 127)) 11756 ((backup_fn_flags 25242)) 11757 ((T_STRING 25245)) 11758 ((T_EXTENDS 2963) (extends_from 25246)) 11759 ((T_IMPLEMENTS 2965) (implements_list 25247)) 11760 ((backup_doc_comment 25248)) 11761 (("{" 25249)) 11762 ((class_statement_list 25250)) 11763 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25251)) 11764 ((":" 25256)) 11765 ((";" 25255)) 11766 ((inner_statement_list 25257)) 11767 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 25258) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 11768 ((";" 25259)) 11769 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25261) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11770 (("%" 299) (")" 25262) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11771 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25263) (";" 17467) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17499) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25264) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 25265) ("{" 17508) ("~" 127)) 11772 ((inner_statement_list 25266)) 11773 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 25267) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 11774 ((";" 25268)) 11775 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 25270) (variable 17576) (variable_class_name 17577)) 11776 (("," 17589) (possible_comma 25271)) 11777 ((")" 25272)) 11778 ((";" 25273)) 11779 ((inner_statement_list 25275)) 11780 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25276) ("~" 127)) 11781 ((catch_list 25277)) 11782 ((T_CATCH 25278) (T_FINALLY 25279) (finally_statement 25280)) 11783 (("(" 25284)) 11784 (("{" 25281)) 11785 ((inner_statement_list 25282)) 11786 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25283) ("~" 127)) 11787 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 25285) (class_name 17636) (name 17637)) 11788 ((T_VARIABLE 17638) (optional_variable 25286) ("|" 17640)) 11789 ((")" 25287)) 11790 (("{" 25288)) 11791 ((inner_statement_list 25289)) 11792 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25290) ("~" 127)) 11793 ((backup_doc_comment 25292)) 11794 (("{" 25293)) 11795 ((class_statement_list 25294)) 11796 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25295)) 11797 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25297) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11798 (("%" 299) (")" 25298) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11799 ((":" 25299) (switch_case_list 25300) ("{" 25301)) 11800 ((";" 25307) (case_list 25308)) 11801 ((";" 25302) (case_list 25303)) 11802 ((case_list 25305)) 11803 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 25304)) 11804 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 25306)) 11805 ((case_list 25311)) 11806 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 25309)) 11807 ((";" 25310)) 11808 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 25312)) 11809 ((";" 25313)) 11810 (("," 17937) (";" 25316)) 11811 ((";" 25318)) 11812 ((T_EXTENDS 21295) (interface_extends_list 25320)) 11813 ((backup_doc_comment 25321)) 11814 (("{" 25322)) 11815 ((class_statement_list 25323)) 11816 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25324)) 11817 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25326) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11818 (("%" 299) (")" 25327) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11819 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21305) (";" 17516) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17517) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17518) (T_DEC 22) (T_DECLARE 17519) (T_DIR 24) (T_DNUMBER 25) (T_DO 17520) (T_DOUBLE_CAST 27) (T_ECHO 17521) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17522) (T_FOREACH 17523) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17524) (T_GOTO 17525) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17526) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17527) (T_START_HEREDOC 69) (T_STATIC 17528) (T_STRING 17529) (T_STRING_CAST 72) (T_SWITCH 17530) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17531) (T_UNSET 17532) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17533) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17534) (alt_if_stmt_without_else 17535) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17536) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17537) (if_stmt_without_else 17538) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25328) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17540) ("~" 127)) 11820 ((")" 25330)) 11821 ((";" 25331)) 11822 ((";" 25333)) 11823 (("," 18988) (";" 25335)) 11824 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 25337) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 11825 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 25338) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 11826 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 25339) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 11827 ((")" 25340) (T_DOUBLE_ARROW 25341)) 11828 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25344) (";" 17467) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17499) (fn 104) (foreach_statement 25350) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25346) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 11829 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 25342) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 11830 ((")" 25343)) 11831 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25344) (";" 17467) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17499) (fn 104) (foreach_statement 25345) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25346) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 11832 ((inner_statement_list 25347)) 11833 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 25348) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 11834 ((";" 25349)) 11835 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 25352) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 11836 ((";" 25353)) 11837 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 25354) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 11838 ((";" 25355)) 11839 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 25356) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 11840 ((")" 25357)) 11841 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25358) (";" 17467) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17499) (fn 104) (for_statement 25359) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25360) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 11842 ((inner_statement_list 25361)) 11843 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 25362) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 11844 ((";" 25363)) 11845 ((":" 21344) (enum_backing_type 25365)) 11846 ((T_IMPLEMENTS 2965) (implements_list 25366)) 11847 ((backup_doc_comment 25367)) 11848 (("{" 25368)) 11849 ((class_statement_list 25369)) 11850 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25370)) 11851 (("," 20829) (";" 25373)) 11852 ((T_WHILE 25375)) 11853 (("(" 25376)) 11854 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25377) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11855 (("%" 299) (")" 25378) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11856 ((";" 25379)) 11857 ((T_STRING 21002) (const_decl 21003) (const_list 25381)) 11858 ((")" 25382) ("," 21006)) 11859 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25383) (";" 17467) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 25384) (dereferenceable_scalar 101) (expr 17499) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25385) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 11860 ((inner_statement_list 25386)) 11861 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 25387) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 11862 ((";" 25388)) 11863 ((";" 25390)) 11864 ((T_EXTENDS 2963) (extends_from 25392)) 11865 ((T_IMPLEMENTS 2965) (implements_list 25393)) 11866 ((backup_doc_comment 25394)) 11867 (("{" 25395)) 11868 ((class_statement_list 25396)) 11869 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25397)) 11870 ((";" 25399)) 11871 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 25401) (variable 17576) (variable_class_name 17577)) 11872 (("," 17589) (possible_comma 25402)) 11873 ((")" 25403)) 11874 ((";" 25404)) 11875 ((inner_statement_list 25406)) 11876 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25407) ("~" 127)) 11877 ((catch_list 25408)) 11878 ((T_CATCH 25409) (T_FINALLY 25410) (finally_statement 25411)) 11879 (("(" 25415)) 11880 (("{" 25412)) 11881 ((inner_statement_list 25413)) 11882 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25414) ("~" 127)) 11883 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 25416) (class_name 17636) (name 17637)) 11884 ((T_VARIABLE 17638) (optional_variable 25417) ("|" 17640)) 11885 ((")" 25418)) 11886 (("{" 25419)) 11887 ((inner_statement_list 25420)) 11888 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25421) ("~" 127)) 11889 ((backup_doc_comment 25423)) 11890 (("{" 25424)) 11891 ((class_statement_list 25425)) 11892 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25426)) 11893 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25428) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11894 (("%" 299) (")" 25429) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11895 ((":" 25430) (switch_case_list 25431) ("{" 25432)) 11896 ((";" 25438) (case_list 25439)) 11897 ((";" 25433) (case_list 25434)) 11898 ((case_list 25436)) 11899 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 25435)) 11900 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 25437)) 11901 ((case_list 25442)) 11902 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 25440)) 11903 ((";" 25441)) 11904 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 25443)) 11905 ((";" 25444)) 11906 (("," 17937) (";" 25447)) 11907 ((";" 25449)) 11908 ((T_EXTENDS 21295) (interface_extends_list 25451)) 11909 ((backup_doc_comment 25452)) 11910 (("{" 25453)) 11911 ((class_statement_list 25454)) 11912 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25455)) 11913 ((")" 25457)) 11914 ((";" 25458)) 11915 ((";" 25460)) 11916 (("," 18988) (";" 25462)) 11917 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 25464) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 11918 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 25465) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 11919 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 25466) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 11920 ((")" 25467) (T_DOUBLE_ARROW 25468)) 11921 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25471) (";" 17386) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17417) (fn 104) (foreach_statement 25477) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25473) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 11922 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 25469) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 11923 ((")" 25470)) 11924 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25471) (";" 17386) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17417) (fn 104) (foreach_statement 25472) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25473) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 11925 ((inner_statement_list 25474)) 11926 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 25475) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 11927 ((";" 25476)) 11928 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 25479) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 11929 ((";" 25480)) 11930 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 25481) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 11931 ((";" 25482)) 11932 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 25483) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 11933 ((")" 25484)) 11934 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25485) (";" 17386) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17417) (fn 104) (for_statement 25486) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25487) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 11935 ((inner_statement_list 25488)) 11936 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 25489) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 11937 ((";" 25490)) 11938 ((":" 21344) (enum_backing_type 25492)) 11939 ((T_IMPLEMENTS 2965) (implements_list 25493)) 11940 ((backup_doc_comment 25494)) 11941 (("{" 25495)) 11942 ((class_statement_list 25496)) 11943 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25497)) 11944 (("," 20829) (";" 25500)) 11945 ((T_WHILE 25502)) 11946 (("(" 25503)) 11947 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25504) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11948 (("%" 299) (")" 25505) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11949 ((";" 25506)) 11950 ((T_STRING 21002) (const_decl 21003) (const_list 25508)) 11951 ((")" 25509) ("," 21006)) 11952 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25510) (";" 17386) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 25511) (dereferenceable_scalar 101) (expr 17417) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25512) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 11953 ((inner_statement_list 25513)) 11954 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 25514) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 11955 ((";" 25515)) 11956 ((";" 25517)) 11957 ((T_EXTENDS 2963) (extends_from 25519)) 11958 ((T_IMPLEMENTS 2965) (implements_list 25520)) 11959 ((backup_doc_comment 25521)) 11960 (("{" 25522)) 11961 ((class_statement_list 25523)) 11962 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25524)) 11963 ((";" 25526)) 11964 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25528) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11965 (("%" 299) (")" 25529) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11966 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25530) (";" 17299) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17300) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17302) (T_DEC 22) (T_DECLARE 17303) (T_DIR 24) (T_DNUMBER 25) (T_DO 17304) (T_DOUBLE_CAST 27) (T_ECHO 17305) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17307) (T_FOREACH 17308) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17309) (T_GOTO 17310) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17313) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17315) (T_START_HEREDOC 69) (T_STATIC 17316) (T_STRING 17317) (T_STRING_CAST 72) (T_SWITCH 17318) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17320) (T_UNSET 17321) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17322) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17323) (alt_if_stmt_without_else 17324) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17330) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17333) (if_stmt_without_else 17334) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25531) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 25532) ("{" 17339) ("~" 127)) 11967 ((inner_statement_list 25533)) 11968 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 25534) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 11969 ((";" 25535)) 11970 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 25537) (variable 17576) (variable_class_name 17577)) 11971 (("," 17589) (possible_comma 25538)) 11972 ((")" 25539)) 11973 ((";" 25540)) 11974 ((inner_statement_list 25542)) 11975 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25543) ("~" 127)) 11976 ((catch_list 25544)) 11977 ((T_CATCH 25545) (T_FINALLY 25546) (finally_statement 25547)) 11978 (("(" 25551)) 11979 (("{" 25548)) 11980 ((inner_statement_list 25549)) 11981 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25550) ("~" 127)) 11982 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 25552) (class_name 17636) (name 17637)) 11983 ((T_VARIABLE 17638) (optional_variable 25553) ("|" 17640)) 11984 ((")" 25554)) 11985 (("{" 25555)) 11986 ((inner_statement_list 25556)) 11987 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25557) ("~" 127)) 11988 ((backup_doc_comment 25559)) 11989 (("{" 25560)) 11990 ((class_statement_list 25561)) 11991 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25562)) 11992 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25564) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 11993 (("%" 299) (")" 25565) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 11994 ((":" 25566) (switch_case_list 25567) ("{" 25568)) 11995 ((";" 25574) (case_list 25575)) 11996 ((";" 25569) (case_list 25570)) 11997 ((case_list 25572)) 11998 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 25571)) 11999 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 25573)) 12000 ((case_list 25578)) 12001 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 25576)) 12002 ((";" 25577)) 12003 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 25579)) 12004 ((";" 25580)) 12005 (("," 17937) (";" 25583)) 12006 ((";" 25585)) 12007 ((T_EXTENDS 21295) (interface_extends_list 25587)) 12008 ((backup_doc_comment 25588)) 12009 (("{" 25589)) 12010 ((class_statement_list 25590)) 12011 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25591)) 12012 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25593) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12013 (("%" 299) (")" 25594) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12014 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21305) (";" 17299) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17300) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17302) (T_DEC 22) (T_DECLARE 17303) (T_DIR 24) (T_DNUMBER 25) (T_DO 17304) (T_DOUBLE_CAST 27) (T_ECHO 17305) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17307) (T_FOREACH 17308) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17309) (T_GOTO 17310) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17313) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17315) (T_START_HEREDOC 69) (T_STATIC 17316) (T_STRING 17317) (T_STRING_CAST 72) (T_SWITCH 17318) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17320) (T_UNSET 17321) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17322) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17323) (alt_if_stmt_without_else 17324) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17330) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17333) (if_stmt_without_else 17334) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25595) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17339) ("~" 127)) 12015 ((")" 25597)) 12016 ((";" 25598)) 12017 ((";" 25600)) 12018 (("," 18988) (";" 25602)) 12019 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 25604) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 12020 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 25605) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 12021 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 25606) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 12022 ((")" 25607) (T_DOUBLE_ARROW 25608)) 12023 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25611) (";" 17299) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17300) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17302) (T_DEC 22) (T_DECLARE 17303) (T_DIR 24) (T_DNUMBER 25) (T_DO 17304) (T_DOUBLE_CAST 27) (T_ECHO 17305) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17307) (T_FOREACH 17308) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17309) (T_GOTO 17310) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17313) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17315) (T_START_HEREDOC 69) (T_STATIC 17316) (T_STRING 17317) (T_STRING_CAST 72) (T_SWITCH 17318) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17320) (T_UNSET 17321) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17322) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17323) (alt_if_stmt_without_else 17324) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17330) (fn 104) (foreach_statement 25617) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17333) (if_stmt_without_else 17334) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25613) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17339) ("~" 127)) 12024 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 25609) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 12025 ((")" 25610)) 12026 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25611) (";" 17299) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17300) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17302) (T_DEC 22) (T_DECLARE 17303) (T_DIR 24) (T_DNUMBER 25) (T_DO 17304) (T_DOUBLE_CAST 27) (T_ECHO 17305) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17307) (T_FOREACH 17308) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17309) (T_GOTO 17310) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17313) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17315) (T_START_HEREDOC 69) (T_STATIC 17316) (T_STRING 17317) (T_STRING_CAST 72) (T_SWITCH 17318) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17320) (T_UNSET 17321) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17322) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17323) (alt_if_stmt_without_else 17324) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17330) (fn 104) (foreach_statement 25612) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17333) (if_stmt_without_else 17334) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25613) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17339) ("~" 127)) 12027 ((inner_statement_list 25614)) 12028 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 25615) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 12029 ((";" 25616)) 12030 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 25619) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 12031 ((";" 25620)) 12032 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 25621) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 12033 ((";" 25622)) 12034 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 25623) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 12035 ((")" 25624)) 12036 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25625) (";" 17299) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17300) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17302) (T_DEC 22) (T_DECLARE 17303) (T_DIR 24) (T_DNUMBER 25) (T_DO 17304) (T_DOUBLE_CAST 27) (T_ECHO 17305) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17307) (T_FOREACH 17308) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17309) (T_GOTO 17310) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17313) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17315) (T_START_HEREDOC 69) (T_STATIC 17316) (T_STRING 17317) (T_STRING_CAST 72) (T_SWITCH 17318) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17320) (T_UNSET 17321) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17322) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17323) (alt_if_stmt_without_else 17324) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17330) (fn 104) (for_statement 25626) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17333) (if_stmt_without_else 17334) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25627) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17339) ("~" 127)) 12037 ((inner_statement_list 25628)) 12038 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 25629) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 12039 ((";" 25630)) 12040 ((":" 21344) (enum_backing_type 25632)) 12041 ((T_IMPLEMENTS 2965) (implements_list 25633)) 12042 ((backup_doc_comment 25634)) 12043 (("{" 25635)) 12044 ((class_statement_list 25636)) 12045 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25637)) 12046 (("," 20829) (";" 25639)) 12047 ((T_WHILE 25641)) 12048 (("(" 25642)) 12049 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25643) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12050 (("%" 299) (")" 25644) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12051 ((";" 25645)) 12052 ((T_STRING 21002) (const_decl 21003) (const_list 25647)) 12053 ((")" 25648) ("," 21006)) 12054 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25649) (";" 17299) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17300) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17302) (T_DEC 22) (T_DECLARE 17303) (T_DIR 24) (T_DNUMBER 25) (T_DO 17304) (T_DOUBLE_CAST 27) (T_ECHO 17305) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17307) (T_FOREACH 17308) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17309) (T_GOTO 17310) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17313) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17315) (T_START_HEREDOC 69) (T_STATIC 17316) (T_STRING 17317) (T_STRING_CAST 72) (T_SWITCH 17318) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17320) (T_UNSET 17321) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17322) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17323) (alt_if_stmt_without_else 17324) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 25650) (dereferenceable_scalar 101) (expr 17330) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17333) (if_stmt_without_else 17334) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25651) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17339) ("~" 127)) 12055 ((inner_statement_list 25652)) 12056 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 25653) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 12057 ((";" 25654)) 12058 ((";" 25656)) 12059 ((T_EXTENDS 2963) (extends_from 25658)) 12060 ((T_IMPLEMENTS 2965) (implements_list 25659)) 12061 ((backup_doc_comment 25660)) 12062 (("{" 25661)) 12063 ((class_statement_list 25662)) 12064 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 25663)) 12065 ((";" 25665)) 12066 ((inner_statement_list 25667)) 12067 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 25668) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 12068 ((";" 25669)) 12069 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25671) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12070 (("%" 299) (")" 25672) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12071 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25673) (";" 17260) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17261) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17262) (T_DEC 22) (T_DECLARE 17263) (T_DIR 24) (T_DNUMBER 25) (T_DO 17264) (T_DOUBLE_CAST 27) (T_ECHO 17265) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17266) (T_FOREACH 17267) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17268) (T_GOTO 17269) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17270) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17271) (T_START_HEREDOC 69) (T_STATIC 17272) (T_STRING 17273) (T_STRING_CAST 72) (T_SWITCH 17274) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17275) (T_UNSET 17276) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17277) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17278) (alt_if_stmt_without_else 17279) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17280) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17281) (if_stmt_without_else 17282) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25674) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 25675) ("{" 17284) ("~" 127)) 12072 ((inner_statement_list 25676)) 12073 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 25677) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 12074 ((";" 25678)) 12075 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 25680) (variable 17576) (variable_class_name 17577)) 12076 (("," 17589) (possible_comma 25681)) 12077 ((")" 25682)) 12078 ((";" 25683)) 12079 ((inner_statement_list 25685)) 12080 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25686) ("~" 127)) 12081 ((catch_list 25687)) 12082 ((T_CATCH 25688) (T_FINALLY 25689) (finally_statement 25690)) 12083 (("(" 25694)) 12084 (("{" 25691)) 12085 ((inner_statement_list 25692)) 12086 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25693) ("~" 127)) 12087 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 25695) (class_name 17636) (name 17637)) 12088 ((T_VARIABLE 17638) (optional_variable 25696) ("|" 17640)) 12089 ((")" 25697)) 12090 (("{" 25698)) 12091 ((inner_statement_list 25699)) 12092 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 25700) ("~" 127)) 12093 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25702) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12094 (("%" 299) (")" 25703) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12095 ((":" 25704) (switch_case_list 25705) ("{" 25706)) 12096 ((";" 25712) (case_list 25713)) 12097 ((";" 25707) (case_list 25708)) 12098 ((case_list 25710)) 12099 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 25709)) 12100 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 25711)) 12101 ((case_list 25716)) 12102 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 25714)) 12103 ((";" 25715)) 12104 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 25717)) 12105 ((";" 25718)) 12106 (("," 17937) (";" 25721)) 12107 ((";" 25723)) 12108 ((";" 25725)) 12109 (("," 18988) (";" 25727)) 12110 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 25729) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 12111 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 25730) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 12112 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 25731) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 12113 ((")" 25732) (T_DOUBLE_ARROW 25733)) 12114 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25736) (";" 17260) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17261) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17262) (T_DEC 22) (T_DECLARE 17263) (T_DIR 24) (T_DNUMBER 25) (T_DO 17264) (T_DOUBLE_CAST 27) (T_ECHO 17265) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17266) (T_FOREACH 17267) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17268) (T_GOTO 17269) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17270) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17271) (T_START_HEREDOC 69) (T_STATIC 17272) (T_STRING 17273) (T_STRING_CAST 72) (T_SWITCH 17274) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17275) (T_UNSET 17276) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17277) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17278) (alt_if_stmt_without_else 17279) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17280) (fn 104) (foreach_statement 25742) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17281) (if_stmt_without_else 17282) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25738) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17284) ("~" 127)) 12115 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 25734) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 12116 ((")" 25735)) 12117 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25736) (";" 17260) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17261) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17262) (T_DEC 22) (T_DECLARE 17263) (T_DIR 24) (T_DNUMBER 25) (T_DO 17264) (T_DOUBLE_CAST 27) (T_ECHO 17265) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17266) (T_FOREACH 17267) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17268) (T_GOTO 17269) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17270) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17271) (T_START_HEREDOC 69) (T_STATIC 17272) (T_STRING 17273) (T_STRING_CAST 72) (T_SWITCH 17274) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17275) (T_UNSET 17276) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17277) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17278) (alt_if_stmt_without_else 17279) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17280) (fn 104) (foreach_statement 25737) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17281) (if_stmt_without_else 17282) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25738) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17284) ("~" 127)) 12118 ((inner_statement_list 25739)) 12119 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 25740) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 12120 ((";" 25741)) 12121 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 25744) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 12122 ((";" 25745)) 12123 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 25746) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 12124 ((";" 25747)) 12125 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 25748) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 12126 ((")" 25749)) 12127 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25750) (";" 17260) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17261) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17262) (T_DEC 22) (T_DECLARE 17263) (T_DIR 24) (T_DNUMBER 25) (T_DO 17264) (T_DOUBLE_CAST 27) (T_ECHO 17265) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17266) (T_FOREACH 17267) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17268) (T_GOTO 17269) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17270) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17271) (T_START_HEREDOC 69) (T_STATIC 17272) (T_STRING 17273) (T_STRING_CAST 72) (T_SWITCH 17274) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17275) (T_UNSET 17276) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17277) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17278) (alt_if_stmt_without_else 17279) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17280) (fn 104) (for_statement 25751) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17281) (if_stmt_without_else 17282) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25752) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17284) ("~" 127)) 12128 ((inner_statement_list 25753)) 12129 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 25754) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 12130 ((";" 25755)) 12131 (("," 20829) (";" 25757)) 12132 ((T_WHILE 25759)) 12133 (("(" 25760)) 12134 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25761) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12135 (("%" 299) (")" 25762) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12136 ((";" 25763)) 12137 ((T_STRING 21002) (const_decl 21003) (const_list 25765)) 12138 ((")" 25766) ("," 21006)) 12139 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25767) (";" 17260) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17261) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17262) (T_DEC 22) (T_DECLARE 17263) (T_DIR 24) (T_DNUMBER 25) (T_DO 17264) (T_DOUBLE_CAST 27) (T_ECHO 17265) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17266) (T_FOREACH 17267) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17268) (T_GOTO 17269) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17270) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17271) (T_START_HEREDOC 69) (T_STATIC 17272) (T_STRING 17273) (T_STRING_CAST 72) (T_SWITCH 17274) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17275) (T_UNSET 17276) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17277) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17278) (alt_if_stmt_without_else 17279) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 25768) (dereferenceable_scalar 101) (expr 17280) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17281) (if_stmt_without_else 17282) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 25769) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17284) ("~" 127)) 12140 ((inner_statement_list 25770)) 12141 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 25771) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 12142 ((";" 25772)) 12143 ((";" 25774)) 12144 ((";" 25776)) 12145 (("$" 221) (T_ABSTRACT 16024) (T_ARRAY 16025) (T_AS 16026) (T_BREAK 16027) (T_CALLABLE 16028) (T_CASE 16029) (T_CATCH 16030) (T_CLASS 16031) (T_CLASS_C 16032) (T_CLONE 16033) (T_CONST 16034) (T_CONTINUE 16035) (T_DECLARE 16036) (T_DEFAULT 16037) (T_DIR 16038) (T_DO 16039) (T_ECHO 16040) (T_ELSE 16041) (T_ELSEIF 16042) (T_EMPTY 16043) (T_ENDDECLARE 16044) (T_ENDFOR 16045) (T_ENDFOREACH 16046) (T_ENDIF 16047) (T_ENDSWITCH 16048) (T_ENDWHILE 16049) (T_ENUM 16050) (T_EVAL 16051) (T_EXIT 16052) (T_EXTENDS 16053) (T_FILE 16054) (T_FINAL 16055) (T_FINALLY 16056) (T_FN 16057) (T_FOR 16058) (T_FOREACH 16059) (T_FUNCTION 16060) (T_FUNC_C 16061) (T_GLOBAL 16062) (T_GOTO 16063) (T_IF 16064) (T_IMPLEMENTS 16065) (T_INCLUDE 16066) (T_INCLUDE_ONCE 16067) (T_INSTANCEOF 16068) (T_INSTEADOF 16069) (T_INTERFACE 16070) (T_ISSET 16071) (T_LINE 16072) (T_LIST 16073) (T_LOGICAL_AND 16074) (T_LOGICAL_OR 16075) (T_LOGICAL_XOR 16076) (T_MATCH 16077) (T_METHOD_C 16078) (T_NAMESPACE 16079) (T_NEW 16080) (T_NS_C 16081) (T_PRINT 16082) (T_PRIVATE 16083) (T_PROPERTY_C 16084) (T_PROTECTED 16085) (T_PUBLIC 16086) (T_READONLY 16087) (T_REQUIRE 16088) (T_REQUIRE_ONCE 16089) (T_RETURN 16090) (T_STATIC 16091) (T_STRING 16092) (T_SWITCH 16093) (T_THROW 16094) (T_TRAIT 16095) (T_TRAIT_C 16096) (T_TRY 16097) (T_UNSET 16098) (T_USE 16099) (T_VAR 16100) (T_VARIABLE 269) (T_WHILE 16101) (T_YIELD 16102) (identifier 25778) (member_name 25779) (reserved_non_modifiers 16105) (semi_reserved 16106) (simple_variable 25780) ("{" 25781)) 12146 (("(" 16111) (argument_list 25784)) 12147 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 25782) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12148 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 25783)) 12149 ((T_FN 35) (T_FUNCTION 129) (fn 283) #'284 (inline_function 25788)) 12150 (("$" 221) (T_STRING 25794) (T_VARIABLE 269) (property_name 25801) (simple_variable 25796) ("{" 25797)) 12151 (("$" 221) (T_STRING 25794) (T_VARIABLE 269) (property_name 25795) (simple_variable 25796) ("{" 25797)) 12152 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 25792) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 12153 (("]" 25793)) 12154 (("(" 16111) (argument_list 25800)) 12155 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 25798) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12156 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 25799)) 12157 (("(" 16111) (argument_list 25802)) 12158 (("`" 25804)) 12159 (("]" 25806)) 12160 (("=" 25807)) 12161 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25808) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12162 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26330) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12163 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 26328) (encaps_var 5237)) 12164 (("$" 25812) (T_VARIABLE 25860) (simple_variable 26324) ("{" 26325)) 12165 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26322) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12166 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26321) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12167 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26320) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12168 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26319) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12169 (("(" 26316)) 12170 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26315) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12171 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26314) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12172 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26313) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12173 (("\"" 1520) ("$" 26115) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 26116) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 26117) ("[" 1542) (array_object_dereferenceable 26118) (callable_expr 26119) (callable_variable 26120) (class_constant 1546) (class_name 26121) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 26122) (name 26123) (new_dereferenceable 1552) (simple_variable 26124) (static_member 26125) (variable 26312) (variable_class_name 26127)) 12174 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26311) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12175 (("(" 26308)) 12176 (("(" 26305)) 12177 (("(" 26296) (argument_list 26297) (ctor_arguments 26298)) 12178 (("\"" 1520) ("$" 26115) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 26116) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 26117) ("[" 1542) (array_object_dereferenceable 26118) (callable_expr 26119) (callable_variable 26120) (class_constant 1546) (class_name 26121) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 26122) (name 26123) (new_dereferenceable 1552) (simple_variable 26124) (static_member 26125) (variable 26295) (variable_class_name 26127)) 12179 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26294) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12180 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26293) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12181 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26292) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12182 (("(" 26288)) 12183 (("(" 26283)) 12184 (("(" 26277)) 12185 (("$" 26115) ("(" 26240) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 26241) (T_NAME_QUALIFIED 26242) (T_NAME_RELATIVE 26243) (T_READONLY 2931) (T_STATIC 26244) (T_STRING 26245) (T_VARIABLE 26117) (anonymous_class 26246) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 26247) (attribute 2937) (attributes 26248) (class_modifier 2939) (class_name 26249) (class_name_reference 26250) (name 26251) (new_variable 26252) (simple_variable 26253)) 12186 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26239) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12187 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26238) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12188 (("(" 26076) (argument_list 26237)) 12189 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26236) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12190 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26235) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12191 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 26230) (T_END_HEREDOC 26231) (T_VARIABLE 2896) (encaps_list 26232) (encaps_var 2898)) 12192 ((T_FN 35) (T_FUNCTION 129) (fn 25874) #'25875 (inline_function 26229)) 12193 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26228) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12194 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26227) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12195 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26226) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12196 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26223) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12197 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26222) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12198 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 26218) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 12199 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 26216) (encaps_list 2804) (encaps_var 2805)) 12200 ((T_NULLSAFE_OBJECT_OPERATOR 26202) (T_OBJECT_OPERATOR 26203) ("[" 26204)) 12201 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 26199) (attribute 2733) (fn 25874) #'25875 (inline_function 26200)) 12202 (("(" 26076) (argument_list 26198)) 12203 ((T_PAAMAYIM_NEKUDOTAYIM 26190)) 12204 (("%" 25890) ("*" 25891) ("+" 25892) ("-" 25893) ("." 25894) ("/" 25895) ("<" 25896) (">" 25897) ("?" 25898) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25899) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25900) (T_BOOLEAN_AND 25901) (T_BOOLEAN_OR 25902) (T_COALESCE 25903) (T_DOUBLE_ARROW 26188) (T_INSTANCEOF 25904) (T_IS_EQUAL 25905) (T_IS_GREATER_OR_EQUAL 25906) (T_IS_IDENTICAL 25907) (T_IS_NOT_EQUAL 25908) (T_IS_NOT_IDENTICAL 25909) (T_IS_SMALLER_OR_EQUAL 25910) (T_LOGICAL_AND 25911) (T_LOGICAL_OR 25912) (T_LOGICAL_XOR 25913) (T_POW 25914) (T_SL 25915) (T_SPACESHIP 25916) (T_SR 25917) ("^" 25918) ("|" 25919)) 12205 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 26177)) 12206 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 26165)) 12207 (("(" 26076) (argument_list 26164)) 12208 (("=" 26084) (T_AND_EQUAL 26085) (T_COALESCE_EQUAL 26086) (T_CONCAT_EQUAL 26087) (T_DEC 26088) (T_DIV_EQUAL 26089) (T_INC 26090) (T_MINUS_EQUAL 26091) (T_MOD_EQUAL 26092) (T_MUL_EQUAL 26093) (T_OR_EQUAL 26094) (T_PLUS_EQUAL 26095) (T_POW_EQUAL 26096) (T_SL_EQUAL 26097) (T_SR_EQUAL 26098) (T_XOR_EQUAL 26099)) 12209 ((T_PAAMAYIM_NEKUDOTAYIM 25988)) 12210 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25889) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12211 (("%" 25890) ("*" 25891) ("+" 25892) ("-" 25893) ("." 25894) ("/" 25895) ("<" 25896) (">" 25897) ("?" 25898) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25899) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25900) (T_BOOLEAN_AND 25901) (T_BOOLEAN_OR 25902) (T_COALESCE 25903) (T_INSTANCEOF 25904) (T_IS_EQUAL 25905) (T_IS_GREATER_OR_EQUAL 25906) (T_IS_IDENTICAL 25907) (T_IS_NOT_EQUAL 25908) (T_IS_NOT_IDENTICAL 25909) (T_IS_SMALLER_OR_EQUAL 25910) (T_LOGICAL_AND 25911) (T_LOGICAL_OR 25912) (T_LOGICAL_XOR 25913) (T_POW 25914) (T_SL 25915) (T_SPACESHIP 25916) (T_SR 25917) ("^" 25918) ("|" 25919)) 12212 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25987) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12213 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25986) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12214 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25985) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12215 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25984) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12216 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25983) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12217 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25982) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12218 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25981) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12219 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25980) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12220 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 25975) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 25976) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 12221 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25974) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12222 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25973) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12223 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25972) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12224 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25971) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12225 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25970) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12226 (("$" 25935) ("(" 25936) (T_NAME_FULLY_QUALIFIED 25937) (T_NAME_QUALIFIED 25938) (T_NAME_RELATIVE 25939) (T_STATIC 25940) (T_STRING 25941) (T_VARIABLE 25942) (class_name 25943) (class_name_reference 25944) (name 25945) (new_variable 25946) (simple_variable 25947)) 12227 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25934) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12228 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25933) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12229 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25932) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12230 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25931) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12231 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25930) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12232 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25929) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12233 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25928) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12234 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25927) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12235 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25926) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12236 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25925) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12237 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25924) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12238 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25923) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12239 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25922) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12240 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25921) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12241 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25920) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12242 (("$" 25935) (T_VARIABLE 25942) (simple_variable 25966) ("{" 25967)) 12243 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 25964) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12244 ((T_PAAMAYIM_NEKUDOTAYIM 25962)) 12245 ((T_NULLSAFE_OBJECT_OPERATOR 25948) (T_OBJECT_OPERATOR 25949) (T_PAAMAYIM_NEKUDOTAYIM 25950) ("[" 25951)) 12246 (("$" 25935) (T_STRING 25955) (T_VARIABLE 25942) (property_name 25961) (simple_variable 25957) ("{" 25958)) 12247 (("$" 25935) (T_STRING 25955) (T_VARIABLE 25942) (property_name 25956) (simple_variable 25957) ("{" 25958)) 12248 (("$" 25935) (T_VARIABLE 25942) (simple_variable 25954)) 12249 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 25952) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 12250 (("]" 25953)) 12251 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 25959) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12252 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 25960)) 12253 (("$" 25935) (T_VARIABLE 25942) (simple_variable 25963)) 12254 (("%" 299) (")" 25965) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12255 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 25968) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12256 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 25969)) 12257 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25979) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12258 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 25977) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 12259 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 25978) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12260 (("$" 25812) (T_ABSTRACT 25989) (T_ARRAY 25990) (T_AS 25991) (T_BREAK 25992) (T_CALLABLE 25993) (T_CASE 25994) (T_CATCH 25995) (T_CLASS 25996) (T_CLASS_C 25997) (T_CLONE 25998) (T_CONST 25999) (T_CONTINUE 26000) (T_DECLARE 26001) (T_DEFAULT 26002) (T_DIR 26003) (T_DO 26004) (T_ECHO 26005) (T_ELSE 26006) (T_ELSEIF 26007) (T_EMPTY 26008) (T_ENDDECLARE 26009) (T_ENDFOR 26010) (T_ENDFOREACH 26011) (T_ENDIF 26012) (T_ENDSWITCH 26013) (T_ENDWHILE 26014) (T_ENUM 26015) (T_EVAL 26016) (T_EXIT 26017) (T_EXTENDS 26018) (T_FILE 26019) (T_FINAL 26020) (T_FINALLY 26021) (T_FN 26022) (T_FOR 26023) (T_FOREACH 26024) (T_FUNCTION 26025) (T_FUNC_C 26026) (T_GLOBAL 26027) (T_GOTO 26028) (T_IF 26029) (T_IMPLEMENTS 26030) (T_INCLUDE 26031) (T_INCLUDE_ONCE 26032) (T_INSTANCEOF 26033) (T_INSTEADOF 26034) (T_INTERFACE 26035) (T_ISSET 26036) (T_LINE 26037) (T_LIST 26038) (T_LOGICAL_AND 26039) (T_LOGICAL_OR 26040) (T_LOGICAL_XOR 26041) (T_MATCH 26042) (T_METHOD_C 26043) (T_NAMESPACE 26044) (T_NEW 26045) (T_NS_C 26046) (T_PRINT 26047) (T_PRIVATE 26048) (T_PROPERTY_C 26049) (T_PROTECTED 26050) (T_PUBLIC 26051) (T_READONLY 26052) (T_REQUIRE 26053) (T_REQUIRE_ONCE 26054) (T_RETURN 26055) (T_STATIC 26056) (T_STRING 26057) (T_SWITCH 26058) (T_THROW 26059) (T_TRAIT 26060) (T_TRAIT_C 26061) (T_TRY 26062) (T_UNSET 26063) (T_USE 26064) (T_VAR 26065) (T_VARIABLE 25860) (T_WHILE 26066) (T_YIELD 26067) (identifier 26068) (member_name 26069) (reserved_non_modifiers 26070) (semi_reserved 26071) (simple_variable 26072) ("{" 26073)) 12261 (("(" 26076) (argument_list 26077)) 12262 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 26074) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12263 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 26075)) 12264 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 26078) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 26079) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 26080) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12265 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 26083) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12266 (("," 1480) (possible_comma 26081)) 12267 ((")" 26082)) 12268 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (ampersand 26113) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26114) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12269 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26112) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12270 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26111) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12271 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26110) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12272 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26109) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12273 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26108) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12274 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26107) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12275 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26106) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12276 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26105) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12277 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26104) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12278 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26103) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12279 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26102) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12280 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26101) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12281 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26100) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12282 (("\"" 1520) ("$" 26115) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 26116) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 26117) ("[" 1542) (array_object_dereferenceable 26118) (callable_expr 26119) (callable_variable 26120) (class_constant 1546) (class_name 26121) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 26122) (name 26123) (new_dereferenceable 1552) (simple_variable 26124) (static_member 26125) (variable 26126) (variable_class_name 26127)) 12283 (("$" 26115) (T_VARIABLE 26117) (simple_variable 26160) ("{" 26161)) 12284 (("(" 26131) (argument_list 26159)) 12285 ((T_NULLSAFE_OBJECT_OPERATOR 26145) (T_OBJECT_OPERATOR 26146) ("[" 26147)) 12286 (("(" 26131) (argument_list 26144)) 12287 ((T_PAAMAYIM_NEKUDOTAYIM 26140)) 12288 (("(" 26131) (argument_list 26139)) 12289 ((T_PAAMAYIM_NEKUDOTAYIM 26128)) 12290 (("$" 26115) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 26117) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 26129) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 26130) ("{" 1642)) 12291 (("(" 26131) (argument_list 26132)) 12292 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 26133) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 26134) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 26135) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12293 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 26138) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12294 (("," 1480) (possible_comma 26136)) 12295 ((")" 26137)) 12296 (("$" 26115) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 26117) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 26141) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 26142) ("{" 1658)) 12297 (("(" 26131) (argument_list 26143)) 12298 (("$" 26115) (T_STRING 26150) (T_VARIABLE 26117) (property_name 26157) (simple_variable 26152) ("{" 26153)) 12299 (("$" 26115) (T_STRING 26150) (T_VARIABLE 26117) (property_name 26151) (simple_variable 26152) ("{" 26153)) 12300 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 26148) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 12301 (("]" 26149)) 12302 (("(" 26131) (argument_list 26156)) 12303 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 26154) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12304 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 26155)) 12305 (("(" 26131) (argument_list 26158)) 12306 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 26162) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12307 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 26163)) 12308 ((backup_doc_comment 26166)) 12309 (("(" 26167)) 12310 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 26168)) 12311 ((")" 26169)) 12312 ((T_USE 164) (lexical_vars 26170)) 12313 ((":" 166) (return_type 26171)) 12314 ((backup_fn_flags 26172)) 12315 (("{" 26173)) 12316 ((inner_statement_list 26174)) 12317 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 26175) ("~" 127)) 12318 ((backup_fn_flags 26176)) 12319 ((backup_doc_comment 26178)) 12320 (("(" 26179)) 12321 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 26180)) 12322 ((")" 26181)) 12323 ((":" 2249) (return_type 26182)) 12324 ((T_DOUBLE_ARROW 26183)) 12325 ((backup_fn_flags 26184)) 12326 ((backup_lex_pos 26185)) 12327 (("!" 16633) ("\"" 16634) ("$" 16635) ("(" 16636) ("+" 16637) ("-" 16638) ("@" 16639) (T_ARRAY 16640) (T_ARRAY_CAST 16641) (T_ATTRIBUTE 128) (T_BOOL_CAST 16642) (T_CLASS_C 16643) (T_CLONE 16644) (T_CONSTANT_ENCAPSED_STRING 16645) (T_DEC 16646) (T_DIR 16647) (T_DNUMBER 16648) (T_DOUBLE_CAST 16649) (T_EMPTY 16650) (T_EVAL 16651) (T_EXIT 16652) (T_FILE 16653) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 16654) (T_INC 16655) (T_INCLUDE 16656) (T_INCLUDE_ONCE 16657) (T_INT_CAST 16658) (T_ISSET 16659) (T_LINE 16660) (T_LIST 16661) (T_LNUMBER 16662) (T_MATCH 16663) (T_METHOD_C 16664) (T_NAME_FULLY_QUALIFIED 16665) (T_NAME_QUALIFIED 16666) (T_NAME_RELATIVE 16667) (T_NEW 16668) (T_NS_C 16669) (T_OBJECT_CAST 16670) (T_PRINT 16671) (T_PROPERTY_C 16672) (T_READONLY 16673) (T_REQUIRE 16674) (T_REQUIRE_ONCE 16675) (T_START_HEREDOC 16676) (T_STATIC 16677) (T_STRING 16678) (T_STRING_CAST 16679) (T_THROW 16680) (T_TRAIT_C 16681) (T_UNSET_CAST 16682) (T_VARIABLE 16683) (T_YIELD 16684) (T_YIELD_FROM 16685) ("[" 16686) ("`" 16687) (array_object_dereferenceable 16688) (attribute 133) (attributes 16689) (callable_expr 16690) (callable_variable 16691) (class_constant 16692) (class_name 16693) (constant 16694) (dereferenceable_scalar 16695) (expr 26186) (fn 16697) (fully_dereferenceable 105) #'16698 (function_call 16699) (inline_function 16700) (internal_functions_in_yacc 16701) (match 16702) (name 16703) (new_dereferenceable 16704) (new_non_dereferenceable 16705) (scalar 16706) (simple_variable 16707) (static_member 16708) (variable 16709) (variable_class_name 16710) ("~" 16711)) 12328 (("%" 16713) ("*" 16714) ("+" 16715) ("-" 16716) ("." 16717) ("/" 16718) ("<" 16719) (">" 16720) ("?" 16721) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 16722) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 16723) (T_BOOLEAN_AND 16724) (T_BOOLEAN_OR 16725) (T_COALESCE 16726) (T_INSTANCEOF 16727) (T_IS_EQUAL 16728) (T_IS_GREATER_OR_EQUAL 16729) (T_IS_IDENTICAL 16730) (T_IS_NOT_EQUAL 16731) (T_IS_NOT_IDENTICAL 16732) (T_IS_SMALLER_OR_EQUAL 16733) (T_LOGICAL_AND 16734) (T_LOGICAL_OR 16735) (T_LOGICAL_XOR 16736) (T_POW 16737) (T_SL 16738) (T_SPACESHIP 16739) (T_SR 16740) ("^" 16741) (backup_fn_flags 26187) ("|" 16742)) 12329 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26189) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12330 (("$" 25812) (T_ABSTRACT 25989) (T_ARRAY 25990) (T_AS 25991) (T_BREAK 25992) (T_CALLABLE 25993) (T_CASE 25994) (T_CATCH 25995) (T_CLASS 25996) (T_CLASS_C 25997) (T_CLONE 25998) (T_CONST 25999) (T_CONTINUE 26000) (T_DECLARE 26001) (T_DEFAULT 26002) (T_DIR 26003) (T_DO 26004) (T_ECHO 26005) (T_ELSE 26006) (T_ELSEIF 26007) (T_EMPTY 26008) (T_ENDDECLARE 26009) (T_ENDFOR 26010) (T_ENDFOREACH 26011) (T_ENDIF 26012) (T_ENDSWITCH 26013) (T_ENDWHILE 26014) (T_ENUM 26015) (T_EVAL 26016) (T_EXIT 26017) (T_EXTENDS 26018) (T_FILE 26019) (T_FINAL 26020) (T_FINALLY 26021) (T_FN 26022) (T_FOR 26023) (T_FOREACH 26024) (T_FUNCTION 26025) (T_FUNC_C 26026) (T_GLOBAL 26027) (T_GOTO 26028) (T_IF 26029) (T_IMPLEMENTS 26030) (T_INCLUDE 26031) (T_INCLUDE_ONCE 26032) (T_INSTANCEOF 26033) (T_INSTEADOF 26034) (T_INTERFACE 26035) (T_ISSET 26036) (T_LINE 26037) (T_LIST 26038) (T_LOGICAL_AND 26039) (T_LOGICAL_OR 26040) (T_LOGICAL_XOR 26041) (T_MATCH 26042) (T_METHOD_C 26043) (T_NAMESPACE 26044) (T_NEW 26045) (T_NS_C 26046) (T_PRINT 26047) (T_PRIVATE 26048) (T_PROPERTY_C 26049) (T_PROTECTED 26050) (T_PUBLIC 26051) (T_READONLY 26052) (T_REQUIRE 26053) (T_REQUIRE_ONCE 26054) (T_RETURN 26055) (T_STATIC 26056) (T_STRING 26057) (T_SWITCH 26058) (T_THROW 26059) (T_TRAIT 26060) (T_TRAIT_C 26061) (T_TRY 26062) (T_UNSET 26063) (T_USE 26064) (T_VAR 26065) (T_VARIABLE 25860) (T_WHILE 26066) (T_YIELD 26067) (identifier 26191) (member_name 26192) (reserved_non_modifiers 26070) (semi_reserved 26071) (simple_variable 26193) ("{" 26194)) 12331 (("(" 26076) (argument_list 26197)) 12332 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 26195) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12333 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 26196)) 12334 ((T_FN 35) (T_FUNCTION 129) (fn 25874) #'25875 (inline_function 26201)) 12335 (("$" 25812) (T_STRING 26207) (T_VARIABLE 25860) (property_name 26214) (simple_variable 26209) ("{" 26210)) 12336 (("$" 25812) (T_STRING 26207) (T_VARIABLE 25860) (property_name 26208) (simple_variable 26209) ("{" 26210)) 12337 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 26205) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 12338 (("]" 26206)) 12339 (("(" 26076) (argument_list 26213)) 12340 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 26211) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12341 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 26212)) 12342 (("(" 26076) (argument_list 26215)) 12343 (("`" 26217)) 12344 (("]" 26219)) 12345 (("=" 26220)) 12346 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26221) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12347 (("%" 25890) ("*" 25891) ("+" 25892) ("-" 25893) ("." 25894) ("/" 25895) ("<" 25896) (">" 25897) ("?" 25898) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25899) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25900) (T_BOOLEAN_AND 25901) (T_BOOLEAN_OR 25902) (T_COALESCE 25903) (T_DOUBLE_ARROW 26224) (T_INSTANCEOF 25904) (T_IS_EQUAL 25905) (T_IS_GREATER_OR_EQUAL 25906) (T_IS_IDENTICAL 25907) (T_IS_NOT_EQUAL 25908) (T_IS_NOT_IDENTICAL 25909) (T_IS_SMALLER_OR_EQUAL 25910) (T_LOGICAL_AND 25911) (T_LOGICAL_OR 25912) (T_LOGICAL_XOR 25913) (T_POW 25914) (T_SL 25915) (T_SPACESHIP 25916) (T_SR 25917) ("^" 25918) ("|" 25919)) 12348 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26225) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12349 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 26234) (T_VARIABLE 2896) (encaps_var 2910)) 12350 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 26233) (T_VARIABLE 2896) (encaps_var 2901)) 12351 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26275) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12352 ((T_CLASS 26267)) 12353 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 26266) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 26247) (attribute 2958) (class_modifier 2939)) 12354 ((T_PAAMAYIM_NEKUDOTAYIM 26264)) 12355 (("(" 26131) (argument_list 26263)) 12356 ((T_NULLSAFE_OBJECT_OPERATOR 26254) (T_OBJECT_OPERATOR 26255) (T_PAAMAYIM_NEKUDOTAYIM 26256) ("[" 26257)) 12357 (("$" 26115) (T_STRING 26150) (T_VARIABLE 26117) (property_name 26262) (simple_variable 26152) ("{" 26153)) 12358 (("$" 26115) (T_STRING 26150) (T_VARIABLE 26117) (property_name 26261) (simple_variable 26152) ("{" 26153)) 12359 (("$" 26115) (T_VARIABLE 26117) (simple_variable 26260)) 12360 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 26258) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 12361 (("]" 26259)) 12362 (("$" 26115) (T_VARIABLE 26117) (simple_variable 26265)) 12363 (("(" 2960) (argument_list 2961) (ctor_arguments 26268)) 12364 ((T_EXTENDS 2963) (extends_from 26269)) 12365 ((T_IMPLEMENTS 2965) (implements_list 26270)) 12366 ((backup_doc_comment 26271)) 12367 (("{" 26272)) 12368 ((class_statement_list 26273)) 12369 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 26274)) 12370 (("%" 299) (")" 26276) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12371 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26278) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12372 (("%" 299) (")" 26279) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12373 (("{" 26280)) 12374 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 26281) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 12375 (("}" 26282)) 12376 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 26284) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 12377 ((")" 26285)) 12378 (("=" 26286)) 12379 (("!" 25810) ("\"" 25811) ("$" 25812) ("(" 25813) ("+" 25814) ("-" 25815) ("@" 25816) (T_ARRAY 25817) (T_ARRAY_CAST 25818) (T_ATTRIBUTE 128) (T_BOOL_CAST 25819) (T_CLASS_C 25820) (T_CLONE 25821) (T_CONSTANT_ENCAPSED_STRING 25822) (T_DEC 25823) (T_DIR 25824) (T_DNUMBER 25825) (T_DOUBLE_CAST 25826) (T_EMPTY 25827) (T_EVAL 25828) (T_EXIT 25829) (T_FILE 25830) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 25831) (T_INC 25832) (T_INCLUDE 25833) (T_INCLUDE_ONCE 25834) (T_INT_CAST 25835) (T_ISSET 25836) (T_LINE 25837) (T_LIST 25838) (T_LNUMBER 25839) (T_MATCH 25840) (T_METHOD_C 25841) (T_NAME_FULLY_QUALIFIED 25842) (T_NAME_QUALIFIED 25843) (T_NAME_RELATIVE 25844) (T_NEW 25845) (T_NS_C 25846) (T_OBJECT_CAST 25847) (T_PRINT 25848) (T_PROPERTY_C 25849) (T_READONLY 25850) (T_REQUIRE 25851) (T_REQUIRE_ONCE 25852) (T_START_HEREDOC 25853) (T_STATIC 25854) (T_STRING 25855) (T_STRING_CAST 25856) (T_THROW 25857) (T_TRAIT_C 25858) (T_UNSET_CAST 25859) (T_VARIABLE 25860) (T_YIELD 25861) (T_YIELD_FROM 25862) ("[" 25863) ("`" 25864) (array_object_dereferenceable 25865) (attribute 133) (attributes 25866) (callable_expr 25867) (callable_variable 25868) (class_constant 25869) (class_name 25870) (constant 25871) (dereferenceable_scalar 25872) (expr 26287) (fn 25874) (fully_dereferenceable 105) #'25875 (function_call 25876) (inline_function 25877) (internal_functions_in_yacc 25878) (match 25879) (name 25880) (new_dereferenceable 25881) (new_non_dereferenceable 25882) (scalar 25883) (simple_variable 25884) (static_member 25885) (variable 25886) (variable_class_name 25887) ("~" 25888)) 12380 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 26289) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12381 (("," 5192) (possible_comma 26290)) 12382 ((")" 26291)) 12383 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 26299) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 26300) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 26301) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12384 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 26304) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12385 (("," 1480) (possible_comma 26302)) 12386 ((")" 26303)) 12387 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26306) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12388 (("%" 299) (")" 26307) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12389 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26309) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12390 (("%" 299) (")" 26310) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12391 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 26317) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 12392 ((")" 26318)) 12393 (("%" 299) (")" 26323) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12394 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 26326) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12395 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 26327)) 12396 (("\"" 26329) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 12397 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 26339) (T_VARIABLE 2896) (encaps_var 2910)) 12398 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 26338) (T_VARIABLE 2896) (encaps_var 2901)) 12399 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26380) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12400 ((T_CLASS 26372)) 12401 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 26371) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 26352) (attribute 2958) (class_modifier 2939)) 12402 ((T_PAAMAYIM_NEKUDOTAYIM 26369)) 12403 (("(" 16166) (argument_list 26368)) 12404 ((T_NULLSAFE_OBJECT_OPERATOR 26359) (T_OBJECT_OPERATOR 26360) (T_PAAMAYIM_NEKUDOTAYIM 26361) ("[" 26362)) 12405 (("$" 16150) (T_STRING 16185) (T_VARIABLE 16152) (property_name 26367) (simple_variable 16187) ("{" 16188)) 12406 (("$" 16150) (T_STRING 16185) (T_VARIABLE 16152) (property_name 26366) (simple_variable 16187) ("{" 16188)) 12407 (("$" 16150) (T_VARIABLE 16152) (simple_variable 26365)) 12408 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 26363) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 12409 (("]" 26364)) 12410 (("$" 16150) (T_VARIABLE 16152) (simple_variable 26370)) 12411 (("(" 2960) (argument_list 2961) (ctor_arguments 26373)) 12412 ((T_EXTENDS 2963) (extends_from 26374)) 12413 ((T_IMPLEMENTS 2965) (implements_list 26375)) 12414 ((backup_doc_comment 26376)) 12415 (("{" 26377)) 12416 ((class_statement_list 26378)) 12417 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 26379)) 12418 (("%" 299) (")" 26381) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12419 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26383) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12420 (("%" 299) (")" 26384) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12421 (("{" 26385)) 12422 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 26386) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 12423 (("}" 26387)) 12424 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 26389) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 12425 ((")" 26390)) 12426 (("=" 26391)) 12427 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26392) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12428 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 26394) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12429 (("," 5192) (possible_comma 26395)) 12430 ((")" 26396)) 12431 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 26404) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 26405) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 26406) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12432 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 26409) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12433 (("," 1480) (possible_comma 26407)) 12434 ((")" 26408)) 12435 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26411) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12436 (("%" 299) (")" 26412) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12437 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26414) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12438 (("%" 299) (")" 26415) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12439 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 26422) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 12440 ((")" 26423)) 12441 (("%" 299) (")" 26428) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12442 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 26431) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12443 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 26432)) 12444 (("\"" 26434) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 12445 ((T_READONLY 17353) (T_STRING 17354) (backup_doc_comment 141) (function_name 26438)) 12446 ((backup_doc_comment 26439)) 12447 (("(" 26440)) 12448 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 26441)) 12449 ((")" 26442)) 12450 ((":" 166) (return_type 26443)) 12451 ((backup_fn_flags 26444)) 12452 (("{" 26445)) 12453 ((inner_statement_list 26446)) 12454 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 26447) ("~" 127)) 12455 ((backup_fn_flags 26448)) 12456 ((T_STRING 26451)) 12457 ((T_EXTENDS 2963) (extends_from 26452)) 12458 ((T_IMPLEMENTS 2965) (implements_list 26453)) 12459 ((backup_doc_comment 26454)) 12460 (("{" 26455)) 12461 ((class_statement_list 26456)) 12462 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 26457)) 12463 ((":" 26462)) 12464 ((";" 26461)) 12465 ((inner_statement_list 26463)) 12466 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 26464) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 12467 ((";" 26465)) 12468 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26467) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12469 (("%" 299) (")" 26468) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12470 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 26469) (";" 171) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 202) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 26470) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 26471) ("{" 211) ("~" 127)) 12471 ((inner_statement_list 26472)) 12472 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 26473) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 12473 ((";" 26474)) 12474 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 26476) (variable 17576) (variable_class_name 17577)) 12475 (("," 17589) (possible_comma 26477)) 12476 ((")" 26478)) 12477 ((";" 26479)) 12478 ((inner_statement_list 26481)) 12479 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 26482) ("~" 127)) 12480 ((catch_list 26483)) 12481 ((T_CATCH 26484) (T_FINALLY 26485) (finally_statement 26486)) 12482 (("(" 26490)) 12483 (("{" 26487)) 12484 ((inner_statement_list 26488)) 12485 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 26489) ("~" 127)) 12486 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 26491) (class_name 17636) (name 17637)) 12487 ((T_VARIABLE 17638) (optional_variable 26492) ("|" 17640)) 12488 ((")" 26493)) 12489 (("{" 26494)) 12490 ((inner_statement_list 26495)) 12491 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 26496) ("~" 127)) 12492 ((backup_doc_comment 26498)) 12493 (("{" 26499)) 12494 ((class_statement_list 26500)) 12495 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 26501)) 12496 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26503) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12497 (("%" 299) (")" 26504) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12498 ((":" 26505) (switch_case_list 26506) ("{" 26507)) 12499 ((";" 26513) (case_list 26514)) 12500 ((";" 26508) (case_list 26509)) 12501 ((case_list 26511)) 12502 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 26510)) 12503 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 26512)) 12504 ((case_list 26517)) 12505 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 26515)) 12506 ((";" 26516)) 12507 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 26518)) 12508 ((";" 26519)) 12509 (("," 17937) (";" 26522)) 12510 ((";" 26524)) 12511 ((T_EXTENDS 21295) (interface_extends_list 26526)) 12512 ((backup_doc_comment 26527)) 12513 (("{" 26528)) 12514 ((class_statement_list 26529)) 12515 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 26530)) 12516 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26532) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12517 (("%" 299) (")" 26533) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12518 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21305) (";" 17260) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 17261) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17262) (T_DEC 22) (T_DECLARE 17263) (T_DIR 24) (T_DNUMBER 25) (T_DO 17264) (T_DOUBLE_CAST 27) (T_ECHO 17265) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17266) (T_FOREACH 17267) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 17268) (T_GOTO 17269) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17270) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17271) (T_START_HEREDOC 69) (T_STATIC 17272) (T_STRING 17273) (T_STRING_CAST 72) (T_SWITCH 17274) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17275) (T_UNSET 17276) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17277) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17278) (alt_if_stmt_without_else 17279) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 17280) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 17281) (if_stmt_without_else 17282) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 26534) (static_member 121) (variable 124) (variable_class_name 125) ("{" 17284) ("~" 127)) 12519 ((")" 26536)) 12520 ((";" 26537)) 12521 ((";" 26539)) 12522 (("," 18988) (";" 26541)) 12523 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 26543) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 12524 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 26544) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 12525 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 26545) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 12526 ((")" 26546) (T_DOUBLE_ARROW 26547)) 12527 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 26550) (";" 171) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 202) (fn 104) (foreach_statement 26556) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 26552) (static_member 121) (variable 124) (variable_class_name 125) ("{" 211) ("~" 127)) 12528 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 26548) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 12529 ((")" 26549)) 12530 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 26550) (";" 171) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 202) (fn 104) (foreach_statement 26551) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 26552) (static_member 121) (variable 124) (variable_class_name 125) ("{" 211) ("~" 127)) 12531 ((inner_statement_list 26553)) 12532 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 26554) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 12533 ((";" 26555)) 12534 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 26558) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 12535 ((";" 26559)) 12536 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 26560) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 12537 ((";" 26561)) 12538 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 26562) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 12539 ((")" 26563)) 12540 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 26564) (";" 171) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 202) (fn 104) (for_statement 26565) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 26566) (static_member 121) (variable 124) (variable_class_name 125) ("{" 211) ("~" 127)) 12541 ((inner_statement_list 26567)) 12542 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 26568) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 12543 ((";" 26569)) 12544 ((":" 21344) (enum_backing_type 26571)) 12545 ((T_IMPLEMENTS 2965) (implements_list 26572)) 12546 ((backup_doc_comment 26573)) 12547 (("{" 26574)) 12548 ((class_statement_list 26575)) 12549 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 26576)) 12550 (("," 20829) (";" 26578)) 12551 ((T_WHILE 26580)) 12552 (("(" 26581)) 12553 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26582) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12554 (("%" 299) (")" 26583) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12555 ((";" 26584)) 12556 ((T_STRING 21002) (const_decl 21003) (const_list 26586)) 12557 ((")" 26587) ("," 21006)) 12558 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 26588) (";" 171) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 26589) (dereferenceable_scalar 101) (expr 202) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 26590) (static_member 121) (variable 124) (variable_class_name 125) ("{" 211) ("~" 127)) 12559 ((inner_statement_list 26591)) 12560 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 26592) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 12561 ((";" 26593)) 12562 ((";" 26595)) 12563 ((T_EXTENDS 2963) (extends_from 26597)) 12564 ((T_IMPLEMENTS 2965) (implements_list 26598)) 12565 ((backup_doc_comment 26599)) 12566 (("{" 26600)) 12567 ((class_statement_list 26601)) 12568 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 26602)) 12569 ((";" 26604)) 12570 ((T_ARRAY 26650) (T_CALLABLE 26651) (T_NAME_FULLY_QUALIFIED 26652) (T_NAME_QUALIFIED 26653) (T_NAME_RELATIVE 26654) (T_STATIC 26655) (T_STRING 26656) (name 26657) (type 26658) (type_without_static 26659)) 12571 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 26648)) 12572 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 26637)) 12573 (("|" 26635)) 12574 (("|" 26620)) 12575 (("(" 26621) (T_ARRAY 26622) (T_CALLABLE 26623) (T_NAME_FULLY_QUALIFIED 26624) (T_NAME_QUALIFIED 26625) (T_NAME_RELATIVE 26626) (T_STATIC 26627) (T_STRING 26628) (name 26629) (type 26630) (type_without_static 26631) (union_type_element 26632)) 12576 ((T_ARRAY 2285) (T_CALLABLE 2286) (T_NAME_FULLY_QUALIFIED 2287) (T_NAME_QUALIFIED 2288) (T_NAME_RELATIVE 2289) (T_STATIC 2290) (T_STRING 2291) (intersection_type 26633) (name 2293) (type 2294) (type_without_static 2295)) 12577 ((")" 26634) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2308)) 12578 (("(" 26621) (T_ARRAY 26622) (T_CALLABLE 26623) (T_NAME_FULLY_QUALIFIED 26624) (T_NAME_QUALIFIED 26625) (T_NAME_RELATIVE 26626) (T_STATIC 26627) (T_STRING 26628) (name 26629) (type 26630) (type_without_static 26631) (union_type_element 26636)) 12579 ((T_ARRAY 26638) (T_CALLABLE 26639) (T_NAME_FULLY_QUALIFIED 26640) (T_NAME_QUALIFIED 26641) (T_NAME_RELATIVE 26642) (T_STATIC 26643) (T_STRING 26644) (name 26645) (type 26646) (type_without_static 26647)) 12580 ((T_ARRAY 26638) (T_CALLABLE 26639) (T_NAME_FULLY_QUALIFIED 26640) (T_NAME_QUALIFIED 26641) (T_NAME_RELATIVE 26642) (T_STATIC 26643) (T_STRING 26644) (name 26645) (type 26649) (type_without_static 26647)) 12581 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 26661) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 26662) (T_VARIABLE 26663) (ampersand 26664) (lexical_var 26665) (lexical_var_list 26666)) 12582 ((T_VARIABLE 26671)) 12583 (("," 26667) (possible_comma 26668)) 12584 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 26661) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 26662) (T_VARIABLE 26663) (ampersand 26664) (lexical_var 26670)) 12585 ((")" 26669)) 12586 ((T_ARRAY 27761) (T_CALLABLE 27762) (T_NAME_FULLY_QUALIFIED 27763) (T_NAME_QUALIFIED 27764) (T_NAME_RELATIVE 27765) (T_STRING 27766) (name 27767) (type_without_static 27768)) 12587 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 27759)) 12588 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 26710) (is_reference 26711)) 12589 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 26701)) 12590 (("|" 26699)) 12591 (("|" 26686)) 12592 (("(" 26687) (T_ARRAY 26688) (T_CALLABLE 26689) (T_NAME_FULLY_QUALIFIED 26690) (T_NAME_QUALIFIED 26691) (T_NAME_RELATIVE 26692) (T_STRING 26693) (name 26694) (type_without_static 26695) (union_type_without_static_element 26696)) 12593 ((T_ARRAY 2285) (T_CALLABLE 2286) (T_NAME_FULLY_QUALIFIED 2287) (T_NAME_QUALIFIED 2288) (T_NAME_RELATIVE 2289) (T_STRING 2291) (intersection_type_without_static 26697) (name 2293) (type_without_static 3022)) 12594 ((")" 26698) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3026)) 12595 (("(" 26687) (T_ARRAY 26688) (T_CALLABLE 26689) (T_NAME_FULLY_QUALIFIED 26690) (T_NAME_QUALIFIED 26691) (T_NAME_RELATIVE 26692) (T_STRING 26693) (name 26694) (type_without_static 26695) (union_type_without_static_element 26700)) 12596 ((T_ARRAY 26702) (T_CALLABLE 26703) (T_NAME_FULLY_QUALIFIED 26704) (T_NAME_QUALIFIED 26705) (T_NAME_RELATIVE 26706) (T_STRING 26707) (name 26708) (type_without_static 26709)) 12597 ((T_ELLIPSIS 26712) (is_variadic 26713)) 12598 ((T_VARIABLE 26714)) 12599 ((backup_doc_comment 26715)) 12600 (("=" 26716) (optional_property_hook_list 26717) ("{" 26718)) 12601 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26784) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12602 ((property_hook_list 26719)) 12603 ((T_ABSTRACT 5691) (T_ATTRIBUTE 5692) (T_FINAL 5693) (T_PRIVATE 5694) (T_PRIVATE_SET 5695) (T_PROTECTED 5696) (T_PROTECTED_SET 5697) (T_PUBLIC 5698) (T_PUBLIC_SET 5699) (T_READONLY 5700) (T_STATIC 5701) (attribute 5702) (attributes 5703) (member_modifier 5704) (non_empty_member_modifiers 5705) (property_hook 5706) (property_hook_modifiers 5707) ("}" 26720)) 12604 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27758) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12605 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 27756) (encaps_var 5237)) 12606 (("$" 26723) (T_VARIABLE 26771) (simple_variable 27752) ("{" 27753)) 12607 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 27750) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12608 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27749) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12609 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27748) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12610 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27747) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12611 (("(" 27744)) 12612 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27743) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12613 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27742) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12614 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27741) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12615 (("\"" 1520) ("$" 27026) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 27027) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 27028) ("[" 1542) (array_object_dereferenceable 27029) (callable_expr 27030) (callable_variable 27031) (class_constant 1546) (class_name 27032) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 27033) (name 27034) (new_dereferenceable 1552) (simple_variable 27035) (static_member 27036) (variable 27740) (variable_class_name 27038)) 12616 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27739) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12617 (("(" 27736)) 12618 (("(" 27733)) 12619 (("(" 27724) (argument_list 27725) (ctor_arguments 27726)) 12620 (("\"" 1520) ("$" 27026) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 27027) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 27028) ("[" 1542) (array_object_dereferenceable 27029) (callable_expr 27030) (callable_variable 27031) (class_constant 1546) (class_name 27032) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 27033) (name 27034) (new_dereferenceable 1552) (simple_variable 27035) (static_member 27036) (variable 27723) (variable_class_name 27038)) 12621 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27722) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12622 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27721) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12623 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27720) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12624 (("(" 27716)) 12625 (("(" 27711)) 12626 (("(" 27705)) 12627 (("$" 27026) ("(" 27668) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 27669) (T_NAME_QUALIFIED 27670) (T_NAME_RELATIVE 27671) (T_READONLY 2931) (T_STATIC 27672) (T_STRING 27673) (T_VARIABLE 27028) (anonymous_class 27674) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 27675) (attribute 2937) (attributes 27676) (class_modifier 2939) (class_name 27677) (class_name_reference 27678) (name 27679) (new_variable 27680) (simple_variable 27681)) 12628 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27667) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12629 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27666) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12630 (("(" 26987) (argument_list 27665)) 12631 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27664) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12632 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27663) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12633 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 27658) (T_END_HEREDOC 27659) (T_VARIABLE 2896) (encaps_list 27660) (encaps_var 2898)) 12634 ((T_FN 35) (T_FUNCTION 129) (fn 26785) #'26786 (inline_function 27657)) 12635 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27656) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12636 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27655) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12637 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27654) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12638 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27196) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12639 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27132) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12640 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 27128) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 12641 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 27126) (encaps_list 2804) (encaps_var 2805)) 12642 ((T_NULLSAFE_OBJECT_OPERATOR 27112) (T_OBJECT_OPERATOR 27113) ("[" 27114)) 12643 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 27109) (attribute 2733) (fn 26785) #'26786 (inline_function 27110)) 12644 (("(" 26987) (argument_list 27108)) 12645 ((T_PAAMAYIM_NEKUDOTAYIM 27100)) 12646 (("%" 26801) ("*" 26802) ("+" 26803) ("-" 26804) ("." 26805) ("/" 26806) ("<" 26807) (">" 26808) ("?" 26809) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 26810) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 26811) (T_BOOLEAN_AND 26812) (T_BOOLEAN_OR 26813) (T_COALESCE 26814) (T_INSTANCEOF 26815) (T_IS_EQUAL 26816) (T_IS_GREATER_OR_EQUAL 26817) (T_IS_IDENTICAL 26818) (T_IS_NOT_EQUAL 26819) (T_IS_NOT_IDENTICAL 26820) (T_IS_SMALLER_OR_EQUAL 26821) (T_LOGICAL_AND 26822) (T_LOGICAL_OR 26823) (T_LOGICAL_XOR 26824) (T_POW 26825) (T_SL 26826) (T_SPACESHIP 26827) (T_SR 26828) ("^" 26829) (optional_property_hook_list 27099) ("{" 26718) ("|" 26830)) 12647 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 27088)) 12648 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 27076)) 12649 (("(" 26987) (argument_list 27075)) 12650 (("=" 26995) (T_AND_EQUAL 26996) (T_COALESCE_EQUAL 26997) (T_CONCAT_EQUAL 26998) (T_DEC 26999) (T_DIV_EQUAL 27000) (T_INC 27001) (T_MINUS_EQUAL 27002) (T_MOD_EQUAL 27003) (T_MUL_EQUAL 27004) (T_OR_EQUAL 27005) (T_PLUS_EQUAL 27006) (T_POW_EQUAL 27007) (T_SL_EQUAL 27008) (T_SR_EQUAL 27009) (T_XOR_EQUAL 27010)) 12651 ((T_PAAMAYIM_NEKUDOTAYIM 26899)) 12652 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26800) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12653 (("%" 26801) ("*" 26802) ("+" 26803) ("-" 26804) ("." 26805) ("/" 26806) ("<" 26807) (">" 26808) ("?" 26809) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 26810) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 26811) (T_BOOLEAN_AND 26812) (T_BOOLEAN_OR 26813) (T_COALESCE 26814) (T_INSTANCEOF 26815) (T_IS_EQUAL 26816) (T_IS_GREATER_OR_EQUAL 26817) (T_IS_IDENTICAL 26818) (T_IS_NOT_EQUAL 26819) (T_IS_NOT_IDENTICAL 26820) (T_IS_SMALLER_OR_EQUAL 26821) (T_LOGICAL_AND 26822) (T_LOGICAL_OR 26823) (T_LOGICAL_XOR 26824) (T_POW 26825) (T_SL 26826) (T_SPACESHIP 26827) (T_SR 26828) ("^" 26829) ("|" 26830)) 12654 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26898) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12655 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26897) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12656 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26896) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12657 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26895) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12658 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26894) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12659 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26893) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12660 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26892) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12661 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26891) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12662 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 26886) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 26887) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 12663 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26885) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12664 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26884) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12665 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26883) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12666 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26882) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12667 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26881) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12668 (("$" 26846) ("(" 26847) (T_NAME_FULLY_QUALIFIED 26848) (T_NAME_QUALIFIED 26849) (T_NAME_RELATIVE 26850) (T_STATIC 26851) (T_STRING 26852) (T_VARIABLE 26853) (class_name 26854) (class_name_reference 26855) (name 26856) (new_variable 26857) (simple_variable 26858)) 12669 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26845) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12670 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26844) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12671 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26843) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12672 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26842) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12673 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26841) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12674 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26840) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12675 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26839) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12676 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26838) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12677 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26837) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12678 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26836) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12679 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26835) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12680 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26834) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12681 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26833) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12682 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26832) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12683 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26831) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12684 (("$" 26846) (T_VARIABLE 26853) (simple_variable 26877) ("{" 26878)) 12685 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 26875) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12686 ((T_PAAMAYIM_NEKUDOTAYIM 26873)) 12687 ((T_NULLSAFE_OBJECT_OPERATOR 26859) (T_OBJECT_OPERATOR 26860) (T_PAAMAYIM_NEKUDOTAYIM 26861) ("[" 26862)) 12688 (("$" 26846) (T_STRING 26866) (T_VARIABLE 26853) (property_name 26872) (simple_variable 26868) ("{" 26869)) 12689 (("$" 26846) (T_STRING 26866) (T_VARIABLE 26853) (property_name 26867) (simple_variable 26868) ("{" 26869)) 12690 (("$" 26846) (T_VARIABLE 26853) (simple_variable 26865)) 12691 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 26863) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 12692 (("]" 26864)) 12693 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 26870) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12694 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 26871)) 12695 (("$" 26846) (T_VARIABLE 26853) (simple_variable 26874)) 12696 (("%" 299) (")" 26876) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12697 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 26879) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12698 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 26880)) 12699 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26890) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12700 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 26888) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 12701 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 26889) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12702 (("$" 26723) (T_ABSTRACT 26900) (T_ARRAY 26901) (T_AS 26902) (T_BREAK 26903) (T_CALLABLE 26904) (T_CASE 26905) (T_CATCH 26906) (T_CLASS 26907) (T_CLASS_C 26908) (T_CLONE 26909) (T_CONST 26910) (T_CONTINUE 26911) (T_DECLARE 26912) (T_DEFAULT 26913) (T_DIR 26914) (T_DO 26915) (T_ECHO 26916) (T_ELSE 26917) (T_ELSEIF 26918) (T_EMPTY 26919) (T_ENDDECLARE 26920) (T_ENDFOR 26921) (T_ENDFOREACH 26922) (T_ENDIF 26923) (T_ENDSWITCH 26924) (T_ENDWHILE 26925) (T_ENUM 26926) (T_EVAL 26927) (T_EXIT 26928) (T_EXTENDS 26929) (T_FILE 26930) (T_FINAL 26931) (T_FINALLY 26932) (T_FN 26933) (T_FOR 26934) (T_FOREACH 26935) (T_FUNCTION 26936) (T_FUNC_C 26937) (T_GLOBAL 26938) (T_GOTO 26939) (T_IF 26940) (T_IMPLEMENTS 26941) (T_INCLUDE 26942) (T_INCLUDE_ONCE 26943) (T_INSTANCEOF 26944) (T_INSTEADOF 26945) (T_INTERFACE 26946) (T_ISSET 26947) (T_LINE 26948) (T_LIST 26949) (T_LOGICAL_AND 26950) (T_LOGICAL_OR 26951) (T_LOGICAL_XOR 26952) (T_MATCH 26953) (T_METHOD_C 26954) (T_NAMESPACE 26955) (T_NEW 26956) (T_NS_C 26957) (T_PRINT 26958) (T_PRIVATE 26959) (T_PROPERTY_C 26960) (T_PROTECTED 26961) (T_PUBLIC 26962) (T_READONLY 26963) (T_REQUIRE 26964) (T_REQUIRE_ONCE 26965) (T_RETURN 26966) (T_STATIC 26967) (T_STRING 26968) (T_SWITCH 26969) (T_THROW 26970) (T_TRAIT 26971) (T_TRAIT_C 26972) (T_TRY 26973) (T_UNSET 26974) (T_USE 26975) (T_VAR 26976) (T_VARIABLE 26771) (T_WHILE 26977) (T_YIELD 26978) (identifier 26979) (member_name 26980) (reserved_non_modifiers 26981) (semi_reserved 26982) (simple_variable 26983) ("{" 26984)) 12703 (("(" 26987) (argument_list 26988)) 12704 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 26985) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12705 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 26986)) 12706 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 26989) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 26990) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 26991) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12707 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 26994) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12708 (("," 1480) (possible_comma 26992)) 12709 ((")" 26993)) 12710 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (ampersand 27024) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27025) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12711 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27023) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12712 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27022) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12713 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27021) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12714 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27020) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12715 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27019) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12716 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27018) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12717 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27017) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12718 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27016) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12719 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27015) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12720 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27014) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12721 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27013) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12722 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27012) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12723 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27011) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12724 (("\"" 1520) ("$" 27026) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 27027) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 27028) ("[" 1542) (array_object_dereferenceable 27029) (callable_expr 27030) (callable_variable 27031) (class_constant 1546) (class_name 27032) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 27033) (name 27034) (new_dereferenceable 1552) (simple_variable 27035) (static_member 27036) (variable 27037) (variable_class_name 27038)) 12725 (("$" 27026) (T_VARIABLE 27028) (simple_variable 27071) ("{" 27072)) 12726 (("(" 27042) (argument_list 27070)) 12727 ((T_NULLSAFE_OBJECT_OPERATOR 27056) (T_OBJECT_OPERATOR 27057) ("[" 27058)) 12728 (("(" 27042) (argument_list 27055)) 12729 ((T_PAAMAYIM_NEKUDOTAYIM 27051)) 12730 (("(" 27042) (argument_list 27050)) 12731 ((T_PAAMAYIM_NEKUDOTAYIM 27039)) 12732 (("$" 27026) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 27028) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 27040) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 27041) ("{" 1642)) 12733 (("(" 27042) (argument_list 27043)) 12734 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 27044) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 27045) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 27046) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12735 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 27049) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12736 (("," 1480) (possible_comma 27047)) 12737 ((")" 27048)) 12738 (("$" 27026) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 27028) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 27052) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 27053) ("{" 1658)) 12739 (("(" 27042) (argument_list 27054)) 12740 (("$" 27026) (T_STRING 27061) (T_VARIABLE 27028) (property_name 27068) (simple_variable 27063) ("{" 27064)) 12741 (("$" 27026) (T_STRING 27061) (T_VARIABLE 27028) (property_name 27062) (simple_variable 27063) ("{" 27064)) 12742 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 27059) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 12743 (("]" 27060)) 12744 (("(" 27042) (argument_list 27067)) 12745 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27065) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12746 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27066)) 12747 (("(" 27042) (argument_list 27069)) 12748 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27073) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12749 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27074)) 12750 ((backup_doc_comment 27077)) 12751 (("(" 27078)) 12752 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 27079)) 12753 ((")" 27080)) 12754 ((T_USE 164) (lexical_vars 27081)) 12755 ((":" 166) (return_type 27082)) 12756 ((backup_fn_flags 27083)) 12757 (("{" 27084)) 12758 ((inner_statement_list 27085)) 12759 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 27086) ("~" 127)) 12760 ((backup_fn_flags 27087)) 12761 ((backup_doc_comment 27089)) 12762 (("(" 27090)) 12763 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 27091)) 12764 ((")" 27092)) 12765 ((":" 2249) (return_type 27093)) 12766 ((T_DOUBLE_ARROW 27094)) 12767 ((backup_fn_flags 27095)) 12768 ((backup_lex_pos 27096)) 12769 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27097) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12770 (("%" 26801) ("*" 26802) ("+" 26803) ("-" 26804) ("." 26805) ("/" 26806) ("<" 26807) (">" 26808) ("?" 26809) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 26810) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 26811) (T_BOOLEAN_AND 26812) (T_BOOLEAN_OR 26813) (T_COALESCE 26814) (T_INSTANCEOF 26815) (T_IS_EQUAL 26816) (T_IS_GREATER_OR_EQUAL 26817) (T_IS_IDENTICAL 26818) (T_IS_NOT_EQUAL 26819) (T_IS_NOT_IDENTICAL 26820) (T_IS_SMALLER_OR_EQUAL 26821) (T_LOGICAL_AND 26822) (T_LOGICAL_OR 26823) (T_LOGICAL_XOR 26824) (T_POW 26825) (T_SL 26826) (T_SPACESHIP 26827) (T_SR 26828) ("^" 26829) (backup_fn_flags 27098) ("|" 26830)) 12771 (("$" 26723) (T_ABSTRACT 26900) (T_ARRAY 26901) (T_AS 26902) (T_BREAK 26903) (T_CALLABLE 26904) (T_CASE 26905) (T_CATCH 26906) (T_CLASS 26907) (T_CLASS_C 26908) (T_CLONE 26909) (T_CONST 26910) (T_CONTINUE 26911) (T_DECLARE 26912) (T_DEFAULT 26913) (T_DIR 26914) (T_DO 26915) (T_ECHO 26916) (T_ELSE 26917) (T_ELSEIF 26918) (T_EMPTY 26919) (T_ENDDECLARE 26920) (T_ENDFOR 26921) (T_ENDFOREACH 26922) (T_ENDIF 26923) (T_ENDSWITCH 26924) (T_ENDWHILE 26925) (T_ENUM 26926) (T_EVAL 26927) (T_EXIT 26928) (T_EXTENDS 26929) (T_FILE 26930) (T_FINAL 26931) (T_FINALLY 26932) (T_FN 26933) (T_FOR 26934) (T_FOREACH 26935) (T_FUNCTION 26936) (T_FUNC_C 26937) (T_GLOBAL 26938) (T_GOTO 26939) (T_IF 26940) (T_IMPLEMENTS 26941) (T_INCLUDE 26942) (T_INCLUDE_ONCE 26943) (T_INSTANCEOF 26944) (T_INSTEADOF 26945) (T_INTERFACE 26946) (T_ISSET 26947) (T_LINE 26948) (T_LIST 26949) (T_LOGICAL_AND 26950) (T_LOGICAL_OR 26951) (T_LOGICAL_XOR 26952) (T_MATCH 26953) (T_METHOD_C 26954) (T_NAMESPACE 26955) (T_NEW 26956) (T_NS_C 26957) (T_PRINT 26958) (T_PRIVATE 26959) (T_PROPERTY_C 26960) (T_PROTECTED 26961) (T_PUBLIC 26962) (T_READONLY 26963) (T_REQUIRE 26964) (T_REQUIRE_ONCE 26965) (T_RETURN 26966) (T_STATIC 26967) (T_STRING 26968) (T_SWITCH 26969) (T_THROW 26970) (T_TRAIT 26971) (T_TRAIT_C 26972) (T_TRY 26973) (T_UNSET 26974) (T_USE 26975) (T_VAR 26976) (T_VARIABLE 26771) (T_WHILE 26977) (T_YIELD 26978) (identifier 27101) (member_name 27102) (reserved_non_modifiers 26981) (semi_reserved 26982) (simple_variable 27103) ("{" 27104)) 12772 (("(" 26987) (argument_list 27107)) 12773 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27105) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12774 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27106)) 12775 ((T_FN 35) (T_FUNCTION 129) (fn 26785) #'26786 (inline_function 27111)) 12776 (("$" 26723) (T_STRING 27117) (T_VARIABLE 26771) (property_name 27124) (simple_variable 27119) ("{" 27120)) 12777 (("$" 26723) (T_STRING 27117) (T_VARIABLE 26771) (property_name 27118) (simple_variable 27119) ("{" 27120)) 12778 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 27115) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 12779 (("]" 27116)) 12780 (("(" 26987) (argument_list 27123)) 12781 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27121) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12782 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27122)) 12783 (("(" 26987) (argument_list 27125)) 12784 (("`" 27127)) 12785 (("]" 27129)) 12786 (("=" 27130)) 12787 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27131) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12788 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27653) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12789 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 27651) (encaps_var 5237)) 12790 (("$" 27135) (T_VARIABLE 27183) (simple_variable 27647) ("{" 27648)) 12791 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 27645) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12792 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27644) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12793 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27643) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12794 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27642) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12795 (("(" 27639)) 12796 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27638) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12797 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27637) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12798 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27636) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12799 (("\"" 1520) ("$" 27438) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 27439) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 27440) ("[" 1542) (array_object_dereferenceable 27441) (callable_expr 27442) (callable_variable 27443) (class_constant 1546) (class_name 27444) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 27445) (name 27446) (new_dereferenceable 1552) (simple_variable 27447) (static_member 27448) (variable 27635) (variable_class_name 27450)) 12800 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27634) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12801 (("(" 27631)) 12802 (("(" 27628)) 12803 (("(" 27619) (argument_list 27620) (ctor_arguments 27621)) 12804 (("\"" 1520) ("$" 27438) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 27439) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 27440) ("[" 1542) (array_object_dereferenceable 27441) (callable_expr 27442) (callable_variable 27443) (class_constant 1546) (class_name 27444) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 27445) (name 27446) (new_dereferenceable 1552) (simple_variable 27447) (static_member 27448) (variable 27618) (variable_class_name 27450)) 12805 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27617) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12806 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27616) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12807 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27615) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12808 (("(" 27611)) 12809 (("(" 27606)) 12810 (("(" 27600)) 12811 (("$" 27438) ("(" 27563) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 27564) (T_NAME_QUALIFIED 27565) (T_NAME_RELATIVE 27566) (T_READONLY 2931) (T_STATIC 27567) (T_STRING 27568) (T_VARIABLE 27440) (anonymous_class 27569) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 27570) (attribute 2937) (attributes 27571) (class_modifier 2939) (class_name 27572) (class_name_reference 27573) (name 27574) (new_variable 27575) (simple_variable 27576)) 12812 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27562) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12813 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27561) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12814 (("(" 27399) (argument_list 27560)) 12815 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27559) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12816 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27558) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12817 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 27553) (T_END_HEREDOC 27554) (T_VARIABLE 2896) (encaps_list 27555) (encaps_var 2898)) 12818 ((T_FN 35) (T_FUNCTION 129) (fn 27197) #'27198 (inline_function 27552)) 12819 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27551) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12820 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27550) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12821 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27549) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12822 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27546) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12823 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27545) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12824 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 27541) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 12825 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 27539) (encaps_list 2804) (encaps_var 2805)) 12826 ((T_NULLSAFE_OBJECT_OPERATOR 27525) (T_OBJECT_OPERATOR 27526) ("[" 27527)) 12827 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 27522) (attribute 2733) (fn 27197) #'27198 (inline_function 27523)) 12828 (("(" 27399) (argument_list 27521)) 12829 ((T_PAAMAYIM_NEKUDOTAYIM 27513)) 12830 (("%" 27213) ("*" 27214) ("+" 27215) ("-" 27216) ("." 27217) ("/" 27218) ("<" 27219) (">" 27220) ("?" 27221) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 27222) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 27223) (T_BOOLEAN_AND 27224) (T_BOOLEAN_OR 27225) (T_COALESCE 27226) (T_DOUBLE_ARROW 27511) (T_INSTANCEOF 27227) (T_IS_EQUAL 27228) (T_IS_GREATER_OR_EQUAL 27229) (T_IS_IDENTICAL 27230) (T_IS_NOT_EQUAL 27231) (T_IS_NOT_IDENTICAL 27232) (T_IS_SMALLER_OR_EQUAL 27233) (T_LOGICAL_AND 27234) (T_LOGICAL_OR 27235) (T_LOGICAL_XOR 27236) (T_POW 27237) (T_SL 27238) (T_SPACESHIP 27239) (T_SR 27240) ("^" 27241) ("|" 27242)) 12831 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 27500)) 12832 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 27488)) 12833 (("(" 27399) (argument_list 27487)) 12834 (("=" 27407) (T_AND_EQUAL 27408) (T_COALESCE_EQUAL 27409) (T_CONCAT_EQUAL 27410) (T_DEC 27411) (T_DIV_EQUAL 27412) (T_INC 27413) (T_MINUS_EQUAL 27414) (T_MOD_EQUAL 27415) (T_MUL_EQUAL 27416) (T_OR_EQUAL 27417) (T_PLUS_EQUAL 27418) (T_POW_EQUAL 27419) (T_SL_EQUAL 27420) (T_SR_EQUAL 27421) (T_XOR_EQUAL 27422)) 12835 ((T_PAAMAYIM_NEKUDOTAYIM 27311)) 12836 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27212) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12837 (("%" 27213) ("*" 27214) ("+" 27215) ("-" 27216) ("." 27217) ("/" 27218) ("<" 27219) (">" 27220) ("?" 27221) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 27222) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 27223) (T_BOOLEAN_AND 27224) (T_BOOLEAN_OR 27225) (T_COALESCE 27226) (T_INSTANCEOF 27227) (T_IS_EQUAL 27228) (T_IS_GREATER_OR_EQUAL 27229) (T_IS_IDENTICAL 27230) (T_IS_NOT_EQUAL 27231) (T_IS_NOT_IDENTICAL 27232) (T_IS_SMALLER_OR_EQUAL 27233) (T_LOGICAL_AND 27234) (T_LOGICAL_OR 27235) (T_LOGICAL_XOR 27236) (T_POW 27237) (T_SL 27238) (T_SPACESHIP 27239) (T_SR 27240) ("^" 27241) ("|" 27242)) 12838 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27310) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12839 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27309) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12840 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27308) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12841 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27307) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12842 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27306) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12843 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27305) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12844 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27304) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12845 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27303) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12846 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 27298) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 27299) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 12847 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27297) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12848 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27296) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12849 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27295) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12850 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27294) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12851 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27293) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12852 (("$" 27258) ("(" 27259) (T_NAME_FULLY_QUALIFIED 27260) (T_NAME_QUALIFIED 27261) (T_NAME_RELATIVE 27262) (T_STATIC 27263) (T_STRING 27264) (T_VARIABLE 27265) (class_name 27266) (class_name_reference 27267) (name 27268) (new_variable 27269) (simple_variable 27270)) 12853 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27257) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12854 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27256) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12855 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27255) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12856 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27254) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12857 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27253) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12858 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27252) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12859 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27251) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12860 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27250) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12861 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27249) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12862 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27248) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12863 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27247) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12864 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27246) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12865 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27245) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12866 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27244) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12867 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27243) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12868 (("$" 27258) (T_VARIABLE 27265) (simple_variable 27289) ("{" 27290)) 12869 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 27287) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12870 ((T_PAAMAYIM_NEKUDOTAYIM 27285)) 12871 ((T_NULLSAFE_OBJECT_OPERATOR 27271) (T_OBJECT_OPERATOR 27272) (T_PAAMAYIM_NEKUDOTAYIM 27273) ("[" 27274)) 12872 (("$" 27258) (T_STRING 27278) (T_VARIABLE 27265) (property_name 27284) (simple_variable 27280) ("{" 27281)) 12873 (("$" 27258) (T_STRING 27278) (T_VARIABLE 27265) (property_name 27279) (simple_variable 27280) ("{" 27281)) 12874 (("$" 27258) (T_VARIABLE 27265) (simple_variable 27277)) 12875 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 27275) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 12876 (("]" 27276)) 12877 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27282) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12878 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27283)) 12879 (("$" 27258) (T_VARIABLE 27265) (simple_variable 27286)) 12880 (("%" 299) (")" 27288) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12881 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27291) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12882 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27292)) 12883 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27302) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12884 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 27300) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 12885 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27301) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12886 (("$" 27135) (T_ABSTRACT 27312) (T_ARRAY 27313) (T_AS 27314) (T_BREAK 27315) (T_CALLABLE 27316) (T_CASE 27317) (T_CATCH 27318) (T_CLASS 27319) (T_CLASS_C 27320) (T_CLONE 27321) (T_CONST 27322) (T_CONTINUE 27323) (T_DECLARE 27324) (T_DEFAULT 27325) (T_DIR 27326) (T_DO 27327) (T_ECHO 27328) (T_ELSE 27329) (T_ELSEIF 27330) (T_EMPTY 27331) (T_ENDDECLARE 27332) (T_ENDFOR 27333) (T_ENDFOREACH 27334) (T_ENDIF 27335) (T_ENDSWITCH 27336) (T_ENDWHILE 27337) (T_ENUM 27338) (T_EVAL 27339) (T_EXIT 27340) (T_EXTENDS 27341) (T_FILE 27342) (T_FINAL 27343) (T_FINALLY 27344) (T_FN 27345) (T_FOR 27346) (T_FOREACH 27347) (T_FUNCTION 27348) (T_FUNC_C 27349) (T_GLOBAL 27350) (T_GOTO 27351) (T_IF 27352) (T_IMPLEMENTS 27353) (T_INCLUDE 27354) (T_INCLUDE_ONCE 27355) (T_INSTANCEOF 27356) (T_INSTEADOF 27357) (T_INTERFACE 27358) (T_ISSET 27359) (T_LINE 27360) (T_LIST 27361) (T_LOGICAL_AND 27362) (T_LOGICAL_OR 27363) (T_LOGICAL_XOR 27364) (T_MATCH 27365) (T_METHOD_C 27366) (T_NAMESPACE 27367) (T_NEW 27368) (T_NS_C 27369) (T_PRINT 27370) (T_PRIVATE 27371) (T_PROPERTY_C 27372) (T_PROTECTED 27373) (T_PUBLIC 27374) (T_READONLY 27375) (T_REQUIRE 27376) (T_REQUIRE_ONCE 27377) (T_RETURN 27378) (T_STATIC 27379) (T_STRING 27380) (T_SWITCH 27381) (T_THROW 27382) (T_TRAIT 27383) (T_TRAIT_C 27384) (T_TRY 27385) (T_UNSET 27386) (T_USE 27387) (T_VAR 27388) (T_VARIABLE 27183) (T_WHILE 27389) (T_YIELD 27390) (identifier 27391) (member_name 27392) (reserved_non_modifiers 27393) (semi_reserved 27394) (simple_variable 27395) ("{" 27396)) 12887 (("(" 27399) (argument_list 27400)) 12888 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27397) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12889 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27398)) 12890 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 27401) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 27402) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 27403) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12891 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 27406) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12892 (("," 1480) (possible_comma 27404)) 12893 ((")" 27405)) 12894 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (ampersand 27436) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27437) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12895 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27435) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12896 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27434) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12897 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27433) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12898 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27432) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12899 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27431) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12900 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27430) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12901 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27429) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12902 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27428) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12903 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27427) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12904 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27426) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12905 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27425) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12906 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27424) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12907 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27423) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12908 (("\"" 1520) ("$" 27438) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 27439) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 27440) ("[" 1542) (array_object_dereferenceable 27441) (callable_expr 27442) (callable_variable 27443) (class_constant 1546) (class_name 27444) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 27445) (name 27446) (new_dereferenceable 1552) (simple_variable 27447) (static_member 27448) (variable 27449) (variable_class_name 27450)) 12909 (("$" 27438) (T_VARIABLE 27440) (simple_variable 27483) ("{" 27484)) 12910 (("(" 27454) (argument_list 27482)) 12911 ((T_NULLSAFE_OBJECT_OPERATOR 27468) (T_OBJECT_OPERATOR 27469) ("[" 27470)) 12912 (("(" 27454) (argument_list 27467)) 12913 ((T_PAAMAYIM_NEKUDOTAYIM 27463)) 12914 (("(" 27454) (argument_list 27462)) 12915 ((T_PAAMAYIM_NEKUDOTAYIM 27451)) 12916 (("$" 27438) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 27440) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 27452) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 27453) ("{" 1642)) 12917 (("(" 27454) (argument_list 27455)) 12918 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 27456) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 27457) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 27458) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12919 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 27461) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 12920 (("," 1480) (possible_comma 27459)) 12921 ((")" 27460)) 12922 (("$" 27438) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 27440) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 27464) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 27465) ("{" 1658)) 12923 (("(" 27454) (argument_list 27466)) 12924 (("$" 27438) (T_STRING 27473) (T_VARIABLE 27440) (property_name 27480) (simple_variable 27475) ("{" 27476)) 12925 (("$" 27438) (T_STRING 27473) (T_VARIABLE 27440) (property_name 27474) (simple_variable 27475) ("{" 27476)) 12926 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 27471) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 12927 (("]" 27472)) 12928 (("(" 27454) (argument_list 27479)) 12929 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27477) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12930 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27478)) 12931 (("(" 27454) (argument_list 27481)) 12932 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27485) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12933 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27486)) 12934 ((backup_doc_comment 27489)) 12935 (("(" 27490)) 12936 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 27491)) 12937 ((")" 27492)) 12938 ((T_USE 164) (lexical_vars 27493)) 12939 ((":" 166) (return_type 27494)) 12940 ((backup_fn_flags 27495)) 12941 (("{" 27496)) 12942 ((inner_statement_list 27497)) 12943 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 27498) ("~" 127)) 12944 ((backup_fn_flags 27499)) 12945 ((backup_doc_comment 27501)) 12946 (("(" 27502)) 12947 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 27503)) 12948 ((")" 27504)) 12949 ((":" 2249) (return_type 27505)) 12950 ((T_DOUBLE_ARROW 27506)) 12951 ((backup_fn_flags 27507)) 12952 ((backup_lex_pos 27508)) 12953 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27509) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12954 (("%" 27213) ("*" 27214) ("+" 27215) ("-" 27216) ("." 27217) ("/" 27218) ("<" 27219) (">" 27220) ("?" 27221) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 27222) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 27223) (T_BOOLEAN_AND 27224) (T_BOOLEAN_OR 27225) (T_COALESCE 27226) (T_INSTANCEOF 27227) (T_IS_EQUAL 27228) (T_IS_GREATER_OR_EQUAL 27229) (T_IS_IDENTICAL 27230) (T_IS_NOT_EQUAL 27231) (T_IS_NOT_IDENTICAL 27232) (T_IS_SMALLER_OR_EQUAL 27233) (T_LOGICAL_AND 27234) (T_LOGICAL_OR 27235) (T_LOGICAL_XOR 27236) (T_POW 27237) (T_SL 27238) (T_SPACESHIP 27239) (T_SR 27240) ("^" 27241) (backup_fn_flags 27510) ("|" 27242)) 12955 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27512) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 12956 (("$" 27135) (T_ABSTRACT 27312) (T_ARRAY 27313) (T_AS 27314) (T_BREAK 27315) (T_CALLABLE 27316) (T_CASE 27317) (T_CATCH 27318) (T_CLASS 27319) (T_CLASS_C 27320) (T_CLONE 27321) (T_CONST 27322) (T_CONTINUE 27323) (T_DECLARE 27324) (T_DEFAULT 27325) (T_DIR 27326) (T_DO 27327) (T_ECHO 27328) (T_ELSE 27329) (T_ELSEIF 27330) (T_EMPTY 27331) (T_ENDDECLARE 27332) (T_ENDFOR 27333) (T_ENDFOREACH 27334) (T_ENDIF 27335) (T_ENDSWITCH 27336) (T_ENDWHILE 27337) (T_ENUM 27338) (T_EVAL 27339) (T_EXIT 27340) (T_EXTENDS 27341) (T_FILE 27342) (T_FINAL 27343) (T_FINALLY 27344) (T_FN 27345) (T_FOR 27346) (T_FOREACH 27347) (T_FUNCTION 27348) (T_FUNC_C 27349) (T_GLOBAL 27350) (T_GOTO 27351) (T_IF 27352) (T_IMPLEMENTS 27353) (T_INCLUDE 27354) (T_INCLUDE_ONCE 27355) (T_INSTANCEOF 27356) (T_INSTEADOF 27357) (T_INTERFACE 27358) (T_ISSET 27359) (T_LINE 27360) (T_LIST 27361) (T_LOGICAL_AND 27362) (T_LOGICAL_OR 27363) (T_LOGICAL_XOR 27364) (T_MATCH 27365) (T_METHOD_C 27366) (T_NAMESPACE 27367) (T_NEW 27368) (T_NS_C 27369) (T_PRINT 27370) (T_PRIVATE 27371) (T_PROPERTY_C 27372) (T_PROTECTED 27373) (T_PUBLIC 27374) (T_READONLY 27375) (T_REQUIRE 27376) (T_REQUIRE_ONCE 27377) (T_RETURN 27378) (T_STATIC 27379) (T_STRING 27380) (T_SWITCH 27381) (T_THROW 27382) (T_TRAIT 27383) (T_TRAIT_C 27384) (T_TRY 27385) (T_UNSET 27386) (T_USE 27387) (T_VAR 27388) (T_VARIABLE 27183) (T_WHILE 27389) (T_YIELD 27390) (identifier 27514) (member_name 27515) (reserved_non_modifiers 27393) (semi_reserved 27394) (simple_variable 27516) ("{" 27517)) 12957 (("(" 27399) (argument_list 27520)) 12958 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27518) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12959 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27519)) 12960 ((T_FN 35) (T_FUNCTION 129) (fn 27197) #'27198 (inline_function 27524)) 12961 (("$" 27135) (T_STRING 27530) (T_VARIABLE 27183) (property_name 27537) (simple_variable 27532) ("{" 27533)) 12962 (("$" 27135) (T_STRING 27530) (T_VARIABLE 27183) (property_name 27531) (simple_variable 27532) ("{" 27533)) 12963 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 27528) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 12964 (("]" 27529)) 12965 (("(" 27399) (argument_list 27536)) 12966 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27534) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 12967 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27535)) 12968 (("(" 27399) (argument_list 27538)) 12969 (("`" 27540)) 12970 (("]" 27542)) 12971 (("=" 27543)) 12972 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27544) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12973 (("%" 27213) ("*" 27214) ("+" 27215) ("-" 27216) ("." 27217) ("/" 27218) ("<" 27219) (">" 27220) ("?" 27221) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 27222) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 27223) (T_BOOLEAN_AND 27224) (T_BOOLEAN_OR 27225) (T_COALESCE 27226) (T_DOUBLE_ARROW 27547) (T_INSTANCEOF 27227) (T_IS_EQUAL 27228) (T_IS_GREATER_OR_EQUAL 27229) (T_IS_IDENTICAL 27230) (T_IS_NOT_EQUAL 27231) (T_IS_NOT_IDENTICAL 27232) (T_IS_SMALLER_OR_EQUAL 27233) (T_LOGICAL_AND 27234) (T_LOGICAL_OR 27235) (T_LOGICAL_XOR 27236) (T_POW 27237) (T_SL 27238) (T_SPACESHIP 27239) (T_SR 27240) ("^" 27241) ("|" 27242)) 12974 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27548) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 12975 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 27557) (T_VARIABLE 2896) (encaps_var 2910)) 12976 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 27556) (T_VARIABLE 2896) (encaps_var 2901)) 12977 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 27598) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12978 ((T_CLASS 27590)) 12979 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 27589) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 27570) (attribute 2958) (class_modifier 2939)) 12980 ((T_PAAMAYIM_NEKUDOTAYIM 27587)) 12981 (("(" 27454) (argument_list 27586)) 12982 ((T_NULLSAFE_OBJECT_OPERATOR 27577) (T_OBJECT_OPERATOR 27578) (T_PAAMAYIM_NEKUDOTAYIM 27579) ("[" 27580)) 12983 (("$" 27438) (T_STRING 27473) (T_VARIABLE 27440) (property_name 27585) (simple_variable 27475) ("{" 27476)) 12984 (("$" 27438) (T_STRING 27473) (T_VARIABLE 27440) (property_name 27584) (simple_variable 27475) ("{" 27476)) 12985 (("$" 27438) (T_VARIABLE 27440) (simple_variable 27583)) 12986 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 27581) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 12987 (("]" 27582)) 12988 (("$" 27438) (T_VARIABLE 27440) (simple_variable 27588)) 12989 (("(" 2960) (argument_list 2961) (ctor_arguments 27591)) 12990 ((T_EXTENDS 2963) (extends_from 27592)) 12991 ((T_IMPLEMENTS 2965) (implements_list 27593)) 12992 ((backup_doc_comment 27594)) 12993 (("{" 27595)) 12994 ((class_statement_list 27596)) 12995 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 27597)) 12996 (("%" 299) (")" 27599) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12997 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 27601) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 12998 (("%" 299) (")" 27602) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 12999 (("{" 27603)) 13000 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 27604) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 13001 (("}" 27605)) 13002 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 27607) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 13003 ((")" 27608)) 13004 (("=" 27609)) 13005 (("!" 27133) ("\"" 27134) ("$" 27135) ("(" 27136) ("+" 27137) ("-" 27138) ("@" 27139) (T_ARRAY 27140) (T_ARRAY_CAST 27141) (T_ATTRIBUTE 128) (T_BOOL_CAST 27142) (T_CLASS_C 27143) (T_CLONE 27144) (T_CONSTANT_ENCAPSED_STRING 27145) (T_DEC 27146) (T_DIR 27147) (T_DNUMBER 27148) (T_DOUBLE_CAST 27149) (T_EMPTY 27150) (T_EVAL 27151) (T_EXIT 27152) (T_FILE 27153) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 27154) (T_INC 27155) (T_INCLUDE 27156) (T_INCLUDE_ONCE 27157) (T_INT_CAST 27158) (T_ISSET 27159) (T_LINE 27160) (T_LIST 27161) (T_LNUMBER 27162) (T_MATCH 27163) (T_METHOD_C 27164) (T_NAME_FULLY_QUALIFIED 27165) (T_NAME_QUALIFIED 27166) (T_NAME_RELATIVE 27167) (T_NEW 27168) (T_NS_C 27169) (T_OBJECT_CAST 27170) (T_PRINT 27171) (T_PROPERTY_C 27172) (T_READONLY 27173) (T_REQUIRE 27174) (T_REQUIRE_ONCE 27175) (T_START_HEREDOC 27176) (T_STATIC 27177) (T_STRING 27178) (T_STRING_CAST 27179) (T_THROW 27180) (T_TRAIT_C 27181) (T_UNSET_CAST 27182) (T_VARIABLE 27183) (T_YIELD 27184) (T_YIELD_FROM 27185) ("[" 27186) ("`" 27187) (array_object_dereferenceable 27188) (attribute 133) (attributes 27189) (callable_expr 27190) (callable_variable 27191) (class_constant 27192) (class_name 27193) (constant 27194) (dereferenceable_scalar 27195) (expr 27610) (fn 27197) (fully_dereferenceable 105) #'27198 (function_call 27199) (inline_function 27200) (internal_functions_in_yacc 27201) (match 27202) (name 27203) (new_dereferenceable 27204) (new_non_dereferenceable 27205) (scalar 27206) (simple_variable 27207) (static_member 27208) (variable 27209) (variable_class_name 27210) ("~" 27211)) 13006 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 27612) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13007 (("," 5192) (possible_comma 27613)) 13008 ((")" 27614)) 13009 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 27622) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 27623) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 27624) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13010 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 27627) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13011 (("," 1480) (possible_comma 27625)) 13012 ((")" 27626)) 13013 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 27629) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13014 (("%" 299) (")" 27630) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13015 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 27632) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13016 (("%" 299) (")" 27633) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13017 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 27640) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 13018 ((")" 27641)) 13019 (("%" 299) (")" 27646) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13020 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27649) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13021 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27650)) 13022 (("\"" 27652) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 13023 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 27662) (T_VARIABLE 2896) (encaps_var 2910)) 13024 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 27661) (T_VARIABLE 2896) (encaps_var 2901)) 13025 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 27703) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13026 ((T_CLASS 27695)) 13027 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 27694) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 27675) (attribute 2958) (class_modifier 2939)) 13028 ((T_PAAMAYIM_NEKUDOTAYIM 27692)) 13029 (("(" 27042) (argument_list 27691)) 13030 ((T_NULLSAFE_OBJECT_OPERATOR 27682) (T_OBJECT_OPERATOR 27683) (T_PAAMAYIM_NEKUDOTAYIM 27684) ("[" 27685)) 13031 (("$" 27026) (T_STRING 27061) (T_VARIABLE 27028) (property_name 27690) (simple_variable 27063) ("{" 27064)) 13032 (("$" 27026) (T_STRING 27061) (T_VARIABLE 27028) (property_name 27689) (simple_variable 27063) ("{" 27064)) 13033 (("$" 27026) (T_VARIABLE 27028) (simple_variable 27688)) 13034 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 27686) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 13035 (("]" 27687)) 13036 (("$" 27026) (T_VARIABLE 27028) (simple_variable 27693)) 13037 (("(" 2960) (argument_list 2961) (ctor_arguments 27696)) 13038 ((T_EXTENDS 2963) (extends_from 27697)) 13039 ((T_IMPLEMENTS 2965) (implements_list 27698)) 13040 ((backup_doc_comment 27699)) 13041 (("{" 27700)) 13042 ((class_statement_list 27701)) 13043 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 27702)) 13044 (("%" 299) (")" 27704) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13045 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 27706) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13046 (("%" 299) (")" 27707) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13047 (("{" 27708)) 13048 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 27709) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 13049 (("}" 27710)) 13050 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 27712) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 13051 ((")" 27713)) 13052 (("=" 27714)) 13053 (("!" 26721) ("\"" 26722) ("$" 26723) ("(" 26724) ("+" 26725) ("-" 26726) ("@" 26727) (T_ARRAY 26728) (T_ARRAY_CAST 26729) (T_ATTRIBUTE 128) (T_BOOL_CAST 26730) (T_CLASS_C 26731) (T_CLONE 26732) (T_CONSTANT_ENCAPSED_STRING 26733) (T_DEC 26734) (T_DIR 26735) (T_DNUMBER 26736) (T_DOUBLE_CAST 26737) (T_EMPTY 26738) (T_EVAL 26739) (T_EXIT 26740) (T_FILE 26741) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 26742) (T_INC 26743) (T_INCLUDE 26744) (T_INCLUDE_ONCE 26745) (T_INT_CAST 26746) (T_ISSET 26747) (T_LINE 26748) (T_LIST 26749) (T_LNUMBER 26750) (T_MATCH 26751) (T_METHOD_C 26752) (T_NAME_FULLY_QUALIFIED 26753) (T_NAME_QUALIFIED 26754) (T_NAME_RELATIVE 26755) (T_NEW 26756) (T_NS_C 26757) (T_OBJECT_CAST 26758) (T_PRINT 26759) (T_PROPERTY_C 26760) (T_READONLY 26761) (T_REQUIRE 26762) (T_REQUIRE_ONCE 26763) (T_START_HEREDOC 26764) (T_STATIC 26765) (T_STRING 26766) (T_STRING_CAST 26767) (T_THROW 26768) (T_TRAIT_C 26769) (T_UNSET_CAST 26770) (T_VARIABLE 26771) (T_YIELD 26772) (T_YIELD_FROM 26773) ("[" 26774) ("`" 26775) (array_object_dereferenceable 26776) (attribute 133) (attributes 26777) (callable_expr 26778) (callable_variable 26779) (class_constant 26780) (class_name 26781) (constant 26782) (dereferenceable_scalar 26783) (expr 27715) (fn 26785) (fully_dereferenceable 105) #'26786 (function_call 26787) (inline_function 26788) (internal_functions_in_yacc 26789) (match 26790) (name 26791) (new_dereferenceable 26792) (new_non_dereferenceable 26793) (scalar 26794) (simple_variable 26795) (static_member 26796) (variable 26797) (variable_class_name 26798) ("~" 26799)) 13054 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 27717) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13055 (("," 5192) (possible_comma 27718)) 13056 ((")" 27719)) 13057 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 27727) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 27728) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 27729) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13058 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 27732) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13059 (("," 1480) (possible_comma 27730)) 13060 ((")" 27731)) 13061 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 27734) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13062 (("%" 299) (")" 27735) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13063 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 27737) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13064 (("%" 299) (")" 27738) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13065 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 27745) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 13066 ((")" 27746)) 13067 (("%" 299) (")" 27751) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13068 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27754) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13069 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27755)) 13070 (("\"" 27757) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 13071 ((T_ARRAY 26702) (T_CALLABLE 26703) (T_NAME_FULLY_QUALIFIED 26704) (T_NAME_QUALIFIED 26705) (T_NAME_RELATIVE 26706) (T_STRING 26707) (name 26708) (type_without_static 27760)) 13072 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 27771) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (optional_cpp_modifiers 160) (parameter 161)) 13073 (("," 5847) (possible_comma 27776)) 13074 (("]" 27777)) 13075 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 27780) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 27781) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 27782) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13076 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 27785) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13077 (("," 1480) (possible_comma 27783)) 13078 ((")" 27784)) 13079 (("," 5847) (possible_comma 27787)) 13080 (("]" 27788)) 13081 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 27790) ("~" 127)) 13082 (("$" 4) (T_ABSTRACT 27792) (T_ARRAY 27793) (T_AS 27794) (T_BREAK 27795) (T_CALLABLE 27796) (T_CASE 27797) (T_CATCH 27798) (T_CLASS 27799) (T_CLASS_C 27800) (T_CLONE 27801) (T_CONST 27802) (T_CONTINUE 27803) (T_DECLARE 27804) (T_DEFAULT 27805) (T_DIR 27806) (T_DO 27807) (T_ECHO 27808) (T_ELSE 27809) (T_ELSEIF 27810) (T_EMPTY 27811) (T_ENDDECLARE 27812) (T_ENDFOR 27813) (T_ENDFOREACH 27814) (T_ENDIF 27815) (T_ENDSWITCH 27816) (T_ENDWHILE 27817) (T_ENUM 27818) (T_EVAL 27819) (T_EXIT 27820) (T_EXTENDS 27821) (T_FILE 27822) (T_FINAL 27823) (T_FINALLY 27824) (T_FN 27825) (T_FOR 27826) (T_FOREACH 27827) (T_FUNCTION 27828) (T_FUNC_C 27829) (T_GLOBAL 27830) (T_GOTO 27831) (T_IF 27832) (T_IMPLEMENTS 27833) (T_INCLUDE 27834) (T_INCLUDE_ONCE 27835) (T_INSTANCEOF 27836) (T_INSTEADOF 27837) (T_INTERFACE 27838) (T_ISSET 27839) (T_LINE 27840) (T_LIST 27841) (T_LOGICAL_AND 27842) (T_LOGICAL_OR 27843) (T_LOGICAL_XOR 27844) (T_MATCH 27845) (T_METHOD_C 27846) (T_NAMESPACE 27847) (T_NEW 27848) (T_NS_C 27849) (T_PRINT 27850) (T_PRIVATE 27851) (T_PROPERTY_C 27852) (T_PROTECTED 27853) (T_PUBLIC 27854) (T_READONLY 27855) (T_REQUIRE 27856) (T_REQUIRE_ONCE 27857) (T_RETURN 27858) (T_STATIC 27859) (T_STRING 27860) (T_SWITCH 27861) (T_THROW 27862) (T_TRAIT 27863) (T_TRAIT_C 27864) (T_TRY 27865) (T_UNSET 27866) (T_USE 27867) (T_VAR 27868) (T_VARIABLE 81) (T_WHILE 27869) (T_YIELD 27870) (identifier 27871) (member_name 27872) (reserved_non_modifiers 27873) (semi_reserved 27874) (simple_variable 27875) ("{" 27876)) 13083 (("(" 27778) (argument_list 27879)) 13084 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27877) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13085 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27878)) 13086 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (ampersand 27909) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27910) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13087 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27908) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13088 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27907) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13089 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27906) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13090 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27905) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13091 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27904) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13092 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27903) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13093 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27902) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13094 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27901) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13095 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27900) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13096 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27899) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13097 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27898) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13098 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27897) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13099 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27896) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13100 (("\"" 1520) ("$" 27911) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 27912) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 27913) ("[" 1542) (array_object_dereferenceable 27914) (callable_expr 27915) (callable_variable 27916) (class_constant 1546) (class_name 27917) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 27918) (name 27919) (new_dereferenceable 1552) (simple_variable 27920) (static_member 27921) (variable 27922) (variable_class_name 27923)) 13101 (("$" 27911) (T_VARIABLE 27913) (simple_variable 27956) ("{" 27957)) 13102 (("(" 27927) (argument_list 27955)) 13103 ((T_NULLSAFE_OBJECT_OPERATOR 27941) (T_OBJECT_OPERATOR 27942) ("[" 27943)) 13104 (("(" 27927) (argument_list 27940)) 13105 ((T_PAAMAYIM_NEKUDOTAYIM 27936)) 13106 (("(" 27927) (argument_list 27935)) 13107 ((T_PAAMAYIM_NEKUDOTAYIM 27924)) 13108 (("$" 27911) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 27913) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 27925) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 27926) ("{" 1642)) 13109 (("(" 27927) (argument_list 27928)) 13110 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 27929) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 27930) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 27931) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13111 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 27934) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13112 (("," 1480) (possible_comma 27932)) 13113 ((")" 27933)) 13114 (("$" 27911) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 27913) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 27937) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 27938) ("{" 1658)) 13115 (("(" 27927) (argument_list 27939)) 13116 (("$" 27911) (T_STRING 27946) (T_VARIABLE 27913) (property_name 27953) (simple_variable 27948) ("{" 27949)) 13117 (("$" 27911) (T_STRING 27946) (T_VARIABLE 27913) (property_name 27947) (simple_variable 27948) ("{" 27949)) 13118 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 27944) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 13119 (("]" 27945)) 13120 (("(" 27927) (argument_list 27952)) 13121 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27950) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13122 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27951)) 13123 (("(" 27927) (argument_list 27954)) 13124 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 27958) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13125 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 27959)) 13126 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 68) (T_START_HEREDOC 69) (T_STATIC 70) (T_STRING 71) (T_STRING_CAST 72) (T_SWITCH 73) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 77) (T_UNSET 78) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 82) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 87) (alt_if_stmt_without_else 88) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 103) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 109) (if_stmt_without_else 110) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 28110) (static_member 121) (variable 124) (variable_class_name 125) ("{" 126) ("~" 127)) 13127 (("(" 27963)) 13128 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 27964) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13129 (("%" 299) (")" 27965) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13130 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 27966) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 27967) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27968) (T_DEC 22) (T_DECLARE 27969) (T_DIR 24) (T_DNUMBER 25) (T_DO 27970) (T_DOUBLE_CAST 27) (T_ECHO 27971) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27972) (T_FOREACH 27973) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 27974) (T_GOTO 27975) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27976) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 27977) (T_START_HEREDOC 69) (T_STATIC 27978) (T_STRING 27979) (T_STRING_CAST 72) (T_SWITCH 27980) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 27981) (T_UNSET 27982) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 27983) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 27984) (alt_if_stmt_without_else 27985) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27986) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 27987) (if_stmt_without_else 27988) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 27989) (static_member 121) (variable 124) (variable_class_name 125) ("{" 27990) ("~" 127)) 13131 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 28108) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13132 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 28106) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13133 (("(" 28097)) 13134 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 28091) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 13135 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 28089) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 13136 (("(" 28076)) 13137 (("(" 28061)) 13138 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 28059) (simple_variable 18987)) 13139 ((T_STRING 28057)) 13140 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 28055) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13141 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 28053)) 13142 ((":" 28052)) 13143 (("(" 28034)) 13144 (("{" 28017)) 13145 (("(" 28012)) 13146 (("(" 28003)) 13147 ((T_ELSE 27996) (T_ELSEIF 17291) (T_ENDIF 27997)) 13148 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 27995) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 13149 ((T_ELSE 27993) (T_ELSEIF 27962)) 13150 ((inner_statement_list 27991)) 13151 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 27992) ("~" 127)) 13152 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 27966) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 27967) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27968) (T_DEC 22) (T_DECLARE 27969) (T_DIR 24) (T_DNUMBER 25) (T_DO 27970) (T_DOUBLE_CAST 27) (T_ECHO 27971) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27972) (T_FOREACH 27973) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 27974) (T_GOTO 27975) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27976) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 27977) (T_START_HEREDOC 69) (T_STATIC 27978) (T_STRING 27979) (T_STRING_CAST 72) (T_SWITCH 27980) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 27981) (T_UNSET 27982) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 27983) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 27984) (alt_if_stmt_without_else 27985) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27986) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 27987) (if_stmt_without_else 27988) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 27994) (static_member 121) (variable 124) (variable_class_name 125) ("{" 27990) ("~" 127)) 13153 ((":" 27999)) 13154 ((";" 27998)) 13155 ((inner_statement_list 28000)) 13156 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 28001) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 13157 ((";" 28002)) 13158 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 28004) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13159 (("%" 299) (")" 28005) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13160 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 28006) (";" 27966) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 27967) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27968) (T_DEC 22) (T_DECLARE 27969) (T_DIR 24) (T_DNUMBER 25) (T_DO 27970) (T_DOUBLE_CAST 27) (T_ECHO 27971) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27972) (T_FOREACH 27973) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 27974) (T_GOTO 27975) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27976) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 27977) (T_START_HEREDOC 69) (T_STATIC 27978) (T_STRING 27979) (T_STRING_CAST 72) (T_SWITCH 27980) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 27981) (T_UNSET 27982) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 27983) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 27984) (alt_if_stmt_without_else 27985) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27986) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 27987) (if_stmt_without_else 27988) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 28007) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 28008) ("{" 27990) ("~" 127)) 13161 ((inner_statement_list 28009)) 13162 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 28010) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 13163 ((";" 28011)) 13164 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 28013) (variable 17576) (variable_class_name 17577)) 13165 (("," 17589) (possible_comma 28014)) 13166 ((")" 28015)) 13167 ((";" 28016)) 13168 ((inner_statement_list 28018)) 13169 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 28019) ("~" 127)) 13170 ((catch_list 28020)) 13171 ((T_CATCH 28021) (T_FINALLY 28022) (finally_statement 28023)) 13172 (("(" 28027)) 13173 (("{" 28024)) 13174 ((inner_statement_list 28025)) 13175 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 28026) ("~" 127)) 13176 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 28028) (class_name 17636) (name 17637)) 13177 ((T_VARIABLE 17638) (optional_variable 28029) ("|" 17640)) 13178 ((")" 28030)) 13179 (("{" 28031)) 13180 ((inner_statement_list 28032)) 13181 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 28033) ("~" 127)) 13182 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 28035) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13183 (("%" 299) (")" 28036) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13184 ((":" 28037) (switch_case_list 28038) ("{" 28039)) 13185 ((";" 28045) (case_list 28046)) 13186 ((";" 28040) (case_list 28041)) 13187 ((case_list 28043)) 13188 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 28042)) 13189 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 28044)) 13190 ((case_list 28049)) 13191 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 28047)) 13192 ((";" 28048)) 13193 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 28050)) 13194 ((";" 28051)) 13195 (("," 17937) (";" 28054)) 13196 ((";" 28056)) 13197 ((";" 28058)) 13198 (("," 18988) (";" 28060)) 13199 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 28062) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 13200 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 28063) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 13201 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 28064) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 13202 ((")" 28065) (T_DOUBLE_ARROW 28066)) 13203 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 28069) (";" 27966) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 27967) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27968) (T_DEC 22) (T_DECLARE 27969) (T_DIR 24) (T_DNUMBER 25) (T_DO 27970) (T_DOUBLE_CAST 27) (T_ECHO 27971) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27972) (T_FOREACH 27973) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 27974) (T_GOTO 27975) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27976) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 27977) (T_START_HEREDOC 69) (T_STATIC 27978) (T_STRING 27979) (T_STRING_CAST 72) (T_SWITCH 27980) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 27981) (T_UNSET 27982) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 27983) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 27984) (alt_if_stmt_without_else 27985) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27986) (fn 104) (foreach_statement 28075) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 27987) (if_stmt_without_else 27988) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 28071) (static_member 121) (variable 124) (variable_class_name 125) ("{" 27990) ("~" 127)) 13204 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 28067) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 13205 ((")" 28068)) 13206 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 28069) (";" 27966) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 27967) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27968) (T_DEC 22) (T_DECLARE 27969) (T_DIR 24) (T_DNUMBER 25) (T_DO 27970) (T_DOUBLE_CAST 27) (T_ECHO 27971) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27972) (T_FOREACH 27973) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 27974) (T_GOTO 27975) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27976) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 27977) (T_START_HEREDOC 69) (T_STATIC 27978) (T_STRING 27979) (T_STRING_CAST 72) (T_SWITCH 27980) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 27981) (T_UNSET 27982) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 27983) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 27984) (alt_if_stmt_without_else 27985) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27986) (fn 104) (foreach_statement 28070) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 27987) (if_stmt_without_else 27988) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 28071) (static_member 121) (variable 124) (variable_class_name 125) ("{" 27990) ("~" 127)) 13207 ((inner_statement_list 28072)) 13208 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 28073) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 13209 ((";" 28074)) 13210 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 28077) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 13211 ((";" 28078)) 13212 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 28079) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 13213 ((";" 28080)) 13214 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 28081) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 13215 ((")" 28082)) 13216 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 28083) (";" 27966) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 27967) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27968) (T_DEC 22) (T_DECLARE 27969) (T_DIR 24) (T_DNUMBER 25) (T_DO 27970) (T_DOUBLE_CAST 27) (T_ECHO 27971) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27972) (T_FOREACH 27973) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 27974) (T_GOTO 27975) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27976) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 27977) (T_START_HEREDOC 69) (T_STATIC 27978) (T_STRING 27979) (T_STRING_CAST 72) (T_SWITCH 27980) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 27981) (T_UNSET 27982) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 27983) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 27984) (alt_if_stmt_without_else 27985) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27986) (fn 104) (for_statement 28084) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 27987) (if_stmt_without_else 27988) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 28085) (static_member 121) (variable 124) (variable_class_name 125) ("{" 27990) ("~" 127)) 13217 ((inner_statement_list 28086)) 13218 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 28087) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 13219 ((";" 28088)) 13220 (("," 20829) (";" 28090)) 13221 ((T_WHILE 28092)) 13222 (("(" 28093)) 13223 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 28094) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13224 (("%" 299) (")" 28095) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13225 ((";" 28096)) 13226 ((T_STRING 21002) (const_decl 21003) (const_list 28098)) 13227 ((")" 28099) ("," 21006)) 13228 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 28100) (";" 27966) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 27967) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27968) (T_DEC 22) (T_DECLARE 27969) (T_DIR 24) (T_DNUMBER 25) (T_DO 27970) (T_DOUBLE_CAST 27) (T_ECHO 27971) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27972) (T_FOREACH 27973) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 27974) (T_GOTO 27975) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27976) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 27977) (T_START_HEREDOC 69) (T_STATIC 27978) (T_STRING 27979) (T_STRING_CAST 72) (T_SWITCH 27980) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 27981) (T_UNSET 27982) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 27983) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 27984) (alt_if_stmt_without_else 27985) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 28101) (dereferenceable_scalar 101) (expr 27986) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 27987) (if_stmt_without_else 27988) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 28102) (static_member 121) (variable 124) (variable_class_name 125) ("{" 27990) ("~" 127)) 13229 ((inner_statement_list 28103)) 13230 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 28104) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 13231 ((";" 28105)) 13232 ((";" 28107)) 13233 ((";" 28109)) 13234 ((T_READONLY 17353) (T_STRING 17354) (backup_doc_comment 141) (function_name 28112)) 13235 ((backup_doc_comment 28113)) 13236 (("(" 28114)) 13237 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 28115)) 13238 ((")" 28116)) 13239 ((":" 166) (return_type 28117)) 13240 ((backup_fn_flags 28118)) 13241 (("{" 28119)) 13242 ((inner_statement_list 28120)) 13243 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 28121) ("~" 127)) 13244 ((backup_fn_flags 28122)) 13245 ((backup_doc_comment 28124)) 13246 (("(" 28125)) 13247 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 28126)) 13248 ((")" 28127)) 13249 ((":" 2249) (return_type 28128)) 13250 ((T_DOUBLE_ARROW 28129)) 13251 ((backup_fn_flags 28130)) 13252 ((backup_lex_pos 28131)) 13253 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28195) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13254 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29169) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13255 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 29167) (encaps_var 5237)) 13256 (("$" 28134) (T_VARIABLE 28182) (simple_variable 29163) ("{" 29164)) 13257 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29161) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13258 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29160) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13259 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29159) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13260 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29158) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13261 (("(" 29155)) 13262 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29154) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13263 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29153) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13264 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29152) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13265 (("\"" 1520) ("$" 28437) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 28438) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 28439) ("[" 1542) (array_object_dereferenceable 28440) (callable_expr 28441) (callable_variable 28442) (class_constant 1546) (class_name 28443) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 28444) (name 28445) (new_dereferenceable 1552) (simple_variable 28446) (static_member 28447) (variable 29151) (variable_class_name 28449)) 13266 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29150) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13267 (("(" 29147)) 13268 (("(" 29144)) 13269 (("(" 29135) (argument_list 29136) (ctor_arguments 29137)) 13270 (("\"" 1520) ("$" 28437) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 28438) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 28439) ("[" 1542) (array_object_dereferenceable 28440) (callable_expr 28441) (callable_variable 28442) (class_constant 1546) (class_name 28443) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 28444) (name 28445) (new_dereferenceable 1552) (simple_variable 28446) (static_member 28447) (variable 29134) (variable_class_name 28449)) 13271 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29133) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13272 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29132) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13273 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29131) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13274 (("(" 29127)) 13275 (("(" 29122)) 13276 (("(" 29116)) 13277 (("$" 28437) ("(" 29079) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 29080) (T_NAME_QUALIFIED 29081) (T_NAME_RELATIVE 29082) (T_READONLY 2931) (T_STATIC 29083) (T_STRING 29084) (T_VARIABLE 28439) (anonymous_class 29085) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 29086) (attribute 2937) (attributes 29087) (class_modifier 2939) (class_name 29088) (class_name_reference 29089) (name 29090) (new_variable 29091) (simple_variable 29092)) 13278 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29078) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13279 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29077) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13280 (("(" 28398) (argument_list 29076)) 13281 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29075) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13282 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29074) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13283 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 29069) (T_END_HEREDOC 29070) (T_VARIABLE 2896) (encaps_list 29071) (encaps_var 2898)) 13284 ((T_FN 35) (T_FUNCTION 129) (fn 28196) #'28197 (inline_function 29068)) 13285 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29067) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13286 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29066) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13287 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29065) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13288 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28607) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13289 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28543) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13290 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 28539) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 13291 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 28537) (encaps_list 2804) (encaps_var 2805)) 13292 ((T_NULLSAFE_OBJECT_OPERATOR 28523) (T_OBJECT_OPERATOR 28524) ("[" 28525)) 13293 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 28520) (attribute 2733) (fn 28196) #'28197 (inline_function 28521)) 13294 (("(" 28398) (argument_list 28519)) 13295 ((T_PAAMAYIM_NEKUDOTAYIM 28511)) 13296 (("%" 28212) ("*" 28213) ("+" 28214) ("-" 28215) ("." 28216) ("/" 28217) ("<" 28218) (">" 28219) ("?" 28220) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 28221) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28222) (T_BOOLEAN_AND 28223) (T_BOOLEAN_OR 28224) (T_COALESCE 28225) (T_INSTANCEOF 28226) (T_IS_EQUAL 28227) (T_IS_GREATER_OR_EQUAL 28228) (T_IS_IDENTICAL 28229) (T_IS_NOT_EQUAL 28230) (T_IS_NOT_IDENTICAL 28231) (T_IS_SMALLER_OR_EQUAL 28232) (T_LOGICAL_AND 28233) (T_LOGICAL_OR 28234) (T_LOGICAL_XOR 28235) (T_POW 28236) (T_SL 28237) (T_SPACESHIP 28238) (T_SR 28239) ("^" 28240) (backup_fn_flags 28510) ("|" 28241)) 13297 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 28499)) 13298 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 28487)) 13299 (("(" 28398) (argument_list 28486)) 13300 (("=" 28406) (T_AND_EQUAL 28407) (T_COALESCE_EQUAL 28408) (T_CONCAT_EQUAL 28409) (T_DEC 28410) (T_DIV_EQUAL 28411) (T_INC 28412) (T_MINUS_EQUAL 28413) (T_MOD_EQUAL 28414) (T_MUL_EQUAL 28415) (T_OR_EQUAL 28416) (T_PLUS_EQUAL 28417) (T_POW_EQUAL 28418) (T_SL_EQUAL 28419) (T_SR_EQUAL 28420) (T_XOR_EQUAL 28421)) 13301 ((T_PAAMAYIM_NEKUDOTAYIM 28310)) 13302 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28211) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13303 (("%" 28212) ("*" 28213) ("+" 28214) ("-" 28215) ("." 28216) ("/" 28217) ("<" 28218) (">" 28219) ("?" 28220) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 28221) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28222) (T_BOOLEAN_AND 28223) (T_BOOLEAN_OR 28224) (T_COALESCE 28225) (T_INSTANCEOF 28226) (T_IS_EQUAL 28227) (T_IS_GREATER_OR_EQUAL 28228) (T_IS_IDENTICAL 28229) (T_IS_NOT_EQUAL 28230) (T_IS_NOT_IDENTICAL 28231) (T_IS_SMALLER_OR_EQUAL 28232) (T_LOGICAL_AND 28233) (T_LOGICAL_OR 28234) (T_LOGICAL_XOR 28235) (T_POW 28236) (T_SL 28237) (T_SPACESHIP 28238) (T_SR 28239) ("^" 28240) ("|" 28241)) 13304 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28309) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13305 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28308) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13306 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28307) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13307 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28306) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13308 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28305) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13309 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28304) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13310 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28303) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13311 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28302) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13312 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 28297) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 28298) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 13313 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28296) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13314 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28295) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13315 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28294) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13316 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28293) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13317 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28292) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13318 (("$" 28257) ("(" 28258) (T_NAME_FULLY_QUALIFIED 28259) (T_NAME_QUALIFIED 28260) (T_NAME_RELATIVE 28261) (T_STATIC 28262) (T_STRING 28263) (T_VARIABLE 28264) (class_name 28265) (class_name_reference 28266) (name 28267) (new_variable 28268) (simple_variable 28269)) 13319 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28256) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13320 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28255) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13321 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28254) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13322 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28253) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13323 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28252) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13324 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28251) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13325 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28250) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13326 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28249) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13327 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28248) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13328 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28247) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13329 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28246) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13330 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28245) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13331 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28244) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13332 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28243) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13333 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28242) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13334 (("$" 28257) (T_VARIABLE 28264) (simple_variable 28288) ("{" 28289)) 13335 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 28286) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13336 ((T_PAAMAYIM_NEKUDOTAYIM 28284)) 13337 ((T_NULLSAFE_OBJECT_OPERATOR 28270) (T_OBJECT_OPERATOR 28271) (T_PAAMAYIM_NEKUDOTAYIM 28272) ("[" 28273)) 13338 (("$" 28257) (T_STRING 28277) (T_VARIABLE 28264) (property_name 28283) (simple_variable 28279) ("{" 28280)) 13339 (("$" 28257) (T_STRING 28277) (T_VARIABLE 28264) (property_name 28278) (simple_variable 28279) ("{" 28280)) 13340 (("$" 28257) (T_VARIABLE 28264) (simple_variable 28276)) 13341 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 28274) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 13342 (("]" 28275)) 13343 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 28281) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13344 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 28282)) 13345 (("$" 28257) (T_VARIABLE 28264) (simple_variable 28285)) 13346 (("%" 299) (")" 28287) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13347 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 28290) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13348 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 28291)) 13349 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28301) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13350 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 28299) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 13351 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28300) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13352 (("$" 28134) (T_ABSTRACT 28311) (T_ARRAY 28312) (T_AS 28313) (T_BREAK 28314) (T_CALLABLE 28315) (T_CASE 28316) (T_CATCH 28317) (T_CLASS 28318) (T_CLASS_C 28319) (T_CLONE 28320) (T_CONST 28321) (T_CONTINUE 28322) (T_DECLARE 28323) (T_DEFAULT 28324) (T_DIR 28325) (T_DO 28326) (T_ECHO 28327) (T_ELSE 28328) (T_ELSEIF 28329) (T_EMPTY 28330) (T_ENDDECLARE 28331) (T_ENDFOR 28332) (T_ENDFOREACH 28333) (T_ENDIF 28334) (T_ENDSWITCH 28335) (T_ENDWHILE 28336) (T_ENUM 28337) (T_EVAL 28338) (T_EXIT 28339) (T_EXTENDS 28340) (T_FILE 28341) (T_FINAL 28342) (T_FINALLY 28343) (T_FN 28344) (T_FOR 28345) (T_FOREACH 28346) (T_FUNCTION 28347) (T_FUNC_C 28348) (T_GLOBAL 28349) (T_GOTO 28350) (T_IF 28351) (T_IMPLEMENTS 28352) (T_INCLUDE 28353) (T_INCLUDE_ONCE 28354) (T_INSTANCEOF 28355) (T_INSTEADOF 28356) (T_INTERFACE 28357) (T_ISSET 28358) (T_LINE 28359) (T_LIST 28360) (T_LOGICAL_AND 28361) (T_LOGICAL_OR 28362) (T_LOGICAL_XOR 28363) (T_MATCH 28364) (T_METHOD_C 28365) (T_NAMESPACE 28366) (T_NEW 28367) (T_NS_C 28368) (T_PRINT 28369) (T_PRIVATE 28370) (T_PROPERTY_C 28371) (T_PROTECTED 28372) (T_PUBLIC 28373) (T_READONLY 28374) (T_REQUIRE 28375) (T_REQUIRE_ONCE 28376) (T_RETURN 28377) (T_STATIC 28378) (T_STRING 28379) (T_SWITCH 28380) (T_THROW 28381) (T_TRAIT 28382) (T_TRAIT_C 28383) (T_TRY 28384) (T_UNSET 28385) (T_USE 28386) (T_VAR 28387) (T_VARIABLE 28182) (T_WHILE 28388) (T_YIELD 28389) (identifier 28390) (member_name 28391) (reserved_non_modifiers 28392) (semi_reserved 28393) (simple_variable 28394) ("{" 28395)) 13353 (("(" 28398) (argument_list 28399)) 13354 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 28396) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13355 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 28397)) 13356 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 28400) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 28401) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 28402) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13357 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 28405) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13358 (("," 1480) (possible_comma 28403)) 13359 ((")" 28404)) 13360 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (ampersand 28435) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28436) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13361 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28434) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13362 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28433) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13363 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28432) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13364 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28431) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13365 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28430) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13366 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28429) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13367 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28428) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13368 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28427) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13369 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28426) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13370 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28425) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13371 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28424) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13372 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28423) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13373 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28422) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13374 (("\"" 1520) ("$" 28437) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 28438) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 28439) ("[" 1542) (array_object_dereferenceable 28440) (callable_expr 28441) (callable_variable 28442) (class_constant 1546) (class_name 28443) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 28444) (name 28445) (new_dereferenceable 1552) (simple_variable 28446) (static_member 28447) (variable 28448) (variable_class_name 28449)) 13375 (("$" 28437) (T_VARIABLE 28439) (simple_variable 28482) ("{" 28483)) 13376 (("(" 28453) (argument_list 28481)) 13377 ((T_NULLSAFE_OBJECT_OPERATOR 28467) (T_OBJECT_OPERATOR 28468) ("[" 28469)) 13378 (("(" 28453) (argument_list 28466)) 13379 ((T_PAAMAYIM_NEKUDOTAYIM 28462)) 13380 (("(" 28453) (argument_list 28461)) 13381 ((T_PAAMAYIM_NEKUDOTAYIM 28450)) 13382 (("$" 28437) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 28439) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 28451) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 28452) ("{" 1642)) 13383 (("(" 28453) (argument_list 28454)) 13384 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 28455) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 28456) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 28457) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13385 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 28460) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13386 (("," 1480) (possible_comma 28458)) 13387 ((")" 28459)) 13388 (("$" 28437) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 28439) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 28463) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 28464) ("{" 1658)) 13389 (("(" 28453) (argument_list 28465)) 13390 (("$" 28437) (T_STRING 28472) (T_VARIABLE 28439) (property_name 28479) (simple_variable 28474) ("{" 28475)) 13391 (("$" 28437) (T_STRING 28472) (T_VARIABLE 28439) (property_name 28473) (simple_variable 28474) ("{" 28475)) 13392 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 28470) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 13393 (("]" 28471)) 13394 (("(" 28453) (argument_list 28478)) 13395 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 28476) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13396 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 28477)) 13397 (("(" 28453) (argument_list 28480)) 13398 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 28484) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13399 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 28485)) 13400 ((backup_doc_comment 28488)) 13401 (("(" 28489)) 13402 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 28490)) 13403 ((")" 28491)) 13404 ((T_USE 164) (lexical_vars 28492)) 13405 ((":" 166) (return_type 28493)) 13406 ((backup_fn_flags 28494)) 13407 (("{" 28495)) 13408 ((inner_statement_list 28496)) 13409 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 28497) ("~" 127)) 13410 ((backup_fn_flags 28498)) 13411 ((backup_doc_comment 28500)) 13412 (("(" 28501)) 13413 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 28502)) 13414 ((")" 28503)) 13415 ((":" 2249) (return_type 28504)) 13416 ((T_DOUBLE_ARROW 28505)) 13417 ((backup_fn_flags 28506)) 13418 ((backup_lex_pos 28507)) 13419 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28508) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13420 (("%" 28212) ("*" 28213) ("+" 28214) ("-" 28215) ("." 28216) ("/" 28217) ("<" 28218) (">" 28219) ("?" 28220) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 28221) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28222) (T_BOOLEAN_AND 28223) (T_BOOLEAN_OR 28224) (T_COALESCE 28225) (T_INSTANCEOF 28226) (T_IS_EQUAL 28227) (T_IS_GREATER_OR_EQUAL 28228) (T_IS_IDENTICAL 28229) (T_IS_NOT_EQUAL 28230) (T_IS_NOT_IDENTICAL 28231) (T_IS_SMALLER_OR_EQUAL 28232) (T_LOGICAL_AND 28233) (T_LOGICAL_OR 28234) (T_LOGICAL_XOR 28235) (T_POW 28236) (T_SL 28237) (T_SPACESHIP 28238) (T_SR 28239) ("^" 28240) (backup_fn_flags 28509) ("|" 28241)) 13421 (("$" 28134) (T_ABSTRACT 28311) (T_ARRAY 28312) (T_AS 28313) (T_BREAK 28314) (T_CALLABLE 28315) (T_CASE 28316) (T_CATCH 28317) (T_CLASS 28318) (T_CLASS_C 28319) (T_CLONE 28320) (T_CONST 28321) (T_CONTINUE 28322) (T_DECLARE 28323) (T_DEFAULT 28324) (T_DIR 28325) (T_DO 28326) (T_ECHO 28327) (T_ELSE 28328) (T_ELSEIF 28329) (T_EMPTY 28330) (T_ENDDECLARE 28331) (T_ENDFOR 28332) (T_ENDFOREACH 28333) (T_ENDIF 28334) (T_ENDSWITCH 28335) (T_ENDWHILE 28336) (T_ENUM 28337) (T_EVAL 28338) (T_EXIT 28339) (T_EXTENDS 28340) (T_FILE 28341) (T_FINAL 28342) (T_FINALLY 28343) (T_FN 28344) (T_FOR 28345) (T_FOREACH 28346) (T_FUNCTION 28347) (T_FUNC_C 28348) (T_GLOBAL 28349) (T_GOTO 28350) (T_IF 28351) (T_IMPLEMENTS 28352) (T_INCLUDE 28353) (T_INCLUDE_ONCE 28354) (T_INSTANCEOF 28355) (T_INSTEADOF 28356) (T_INTERFACE 28357) (T_ISSET 28358) (T_LINE 28359) (T_LIST 28360) (T_LOGICAL_AND 28361) (T_LOGICAL_OR 28362) (T_LOGICAL_XOR 28363) (T_MATCH 28364) (T_METHOD_C 28365) (T_NAMESPACE 28366) (T_NEW 28367) (T_NS_C 28368) (T_PRINT 28369) (T_PRIVATE 28370) (T_PROPERTY_C 28371) (T_PROTECTED 28372) (T_PUBLIC 28373) (T_READONLY 28374) (T_REQUIRE 28375) (T_REQUIRE_ONCE 28376) (T_RETURN 28377) (T_STATIC 28378) (T_STRING 28379) (T_SWITCH 28380) (T_THROW 28381) (T_TRAIT 28382) (T_TRAIT_C 28383) (T_TRY 28384) (T_UNSET 28385) (T_USE 28386) (T_VAR 28387) (T_VARIABLE 28182) (T_WHILE 28388) (T_YIELD 28389) (identifier 28512) (member_name 28513) (reserved_non_modifiers 28392) (semi_reserved 28393) (simple_variable 28514) ("{" 28515)) 13422 (("(" 28398) (argument_list 28518)) 13423 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 28516) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13424 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 28517)) 13425 ((T_FN 35) (T_FUNCTION 129) (fn 28196) #'28197 (inline_function 28522)) 13426 (("$" 28134) (T_STRING 28528) (T_VARIABLE 28182) (property_name 28535) (simple_variable 28530) ("{" 28531)) 13427 (("$" 28134) (T_STRING 28528) (T_VARIABLE 28182) (property_name 28529) (simple_variable 28530) ("{" 28531)) 13428 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 28526) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 13429 (("]" 28527)) 13430 (("(" 28398) (argument_list 28534)) 13431 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 28532) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13432 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 28533)) 13433 (("(" 28398) (argument_list 28536)) 13434 (("`" 28538)) 13435 (("]" 28540)) 13436 (("=" 28541)) 13437 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28542) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13438 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 29064) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13439 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 29062) (encaps_var 5237)) 13440 (("$" 28546) (T_VARIABLE 28594) (simple_variable 29058) ("{" 29059)) 13441 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29056) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13442 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 29055) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13443 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 29054) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13444 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 29053) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13445 (("(" 29050)) 13446 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 29049) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13447 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 29048) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13448 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 29047) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13449 (("\"" 1520) ("$" 28849) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 28850) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 28851) ("[" 1542) (array_object_dereferenceable 28852) (callable_expr 28853) (callable_variable 28854) (class_constant 1546) (class_name 28855) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 28856) (name 28857) (new_dereferenceable 1552) (simple_variable 28858) (static_member 28859) (variable 29046) (variable_class_name 28861)) 13450 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 29045) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13451 (("(" 29042)) 13452 (("(" 29039)) 13453 (("(" 29030) (argument_list 29031) (ctor_arguments 29032)) 13454 (("\"" 1520) ("$" 28849) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 28850) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 28851) ("[" 1542) (array_object_dereferenceable 28852) (callable_expr 28853) (callable_variable 28854) (class_constant 1546) (class_name 28855) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 28856) (name 28857) (new_dereferenceable 1552) (simple_variable 28858) (static_member 28859) (variable 29029) (variable_class_name 28861)) 13455 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 29028) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13456 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 29027) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13457 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 29026) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13458 (("(" 29022)) 13459 (("(" 29017)) 13460 (("(" 29011)) 13461 (("$" 28849) ("(" 28974) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 28975) (T_NAME_QUALIFIED 28976) (T_NAME_RELATIVE 28977) (T_READONLY 2931) (T_STATIC 28978) (T_STRING 28979) (T_VARIABLE 28851) (anonymous_class 28980) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 28981) (attribute 2937) (attributes 28982) (class_modifier 2939) (class_name 28983) (class_name_reference 28984) (name 28985) (new_variable 28986) (simple_variable 28987)) 13462 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28973) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13463 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28972) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13464 (("(" 28810) (argument_list 28971)) 13465 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28970) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13466 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28969) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13467 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 28964) (T_END_HEREDOC 28965) (T_VARIABLE 2896) (encaps_list 28966) (encaps_var 2898)) 13468 ((T_FN 35) (T_FUNCTION 129) (fn 28608) #'28609 (inline_function 28963)) 13469 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28962) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13470 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28961) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13471 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28960) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13472 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28957) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13473 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28956) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13474 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 28952) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 13475 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 28950) (encaps_list 2804) (encaps_var 2805)) 13476 ((T_NULLSAFE_OBJECT_OPERATOR 28936) (T_OBJECT_OPERATOR 28937) ("[" 28938)) 13477 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 28933) (attribute 2733) (fn 28608) #'28609 (inline_function 28934)) 13478 (("(" 28810) (argument_list 28932)) 13479 ((T_PAAMAYIM_NEKUDOTAYIM 28924)) 13480 (("%" 28624) ("*" 28625) ("+" 28626) ("-" 28627) ("." 28628) ("/" 28629) ("<" 28630) (">" 28631) ("?" 28632) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 28633) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28634) (T_BOOLEAN_AND 28635) (T_BOOLEAN_OR 28636) (T_COALESCE 28637) (T_DOUBLE_ARROW 28922) (T_INSTANCEOF 28638) (T_IS_EQUAL 28639) (T_IS_GREATER_OR_EQUAL 28640) (T_IS_IDENTICAL 28641) (T_IS_NOT_EQUAL 28642) (T_IS_NOT_IDENTICAL 28643) (T_IS_SMALLER_OR_EQUAL 28644) (T_LOGICAL_AND 28645) (T_LOGICAL_OR 28646) (T_LOGICAL_XOR 28647) (T_POW 28648) (T_SL 28649) (T_SPACESHIP 28650) (T_SR 28651) ("^" 28652) ("|" 28653)) 13481 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 28911)) 13482 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 28899)) 13483 (("(" 28810) (argument_list 28898)) 13484 (("=" 28818) (T_AND_EQUAL 28819) (T_COALESCE_EQUAL 28820) (T_CONCAT_EQUAL 28821) (T_DEC 28822) (T_DIV_EQUAL 28823) (T_INC 28824) (T_MINUS_EQUAL 28825) (T_MOD_EQUAL 28826) (T_MUL_EQUAL 28827) (T_OR_EQUAL 28828) (T_PLUS_EQUAL 28829) (T_POW_EQUAL 28830) (T_SL_EQUAL 28831) (T_SR_EQUAL 28832) (T_XOR_EQUAL 28833)) 13485 ((T_PAAMAYIM_NEKUDOTAYIM 28722)) 13486 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28623) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13487 (("%" 28624) ("*" 28625) ("+" 28626) ("-" 28627) ("." 28628) ("/" 28629) ("<" 28630) (">" 28631) ("?" 28632) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 28633) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28634) (T_BOOLEAN_AND 28635) (T_BOOLEAN_OR 28636) (T_COALESCE 28637) (T_INSTANCEOF 28638) (T_IS_EQUAL 28639) (T_IS_GREATER_OR_EQUAL 28640) (T_IS_IDENTICAL 28641) (T_IS_NOT_EQUAL 28642) (T_IS_NOT_IDENTICAL 28643) (T_IS_SMALLER_OR_EQUAL 28644) (T_LOGICAL_AND 28645) (T_LOGICAL_OR 28646) (T_LOGICAL_XOR 28647) (T_POW 28648) (T_SL 28649) (T_SPACESHIP 28650) (T_SR 28651) ("^" 28652) ("|" 28653)) 13488 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28721) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13489 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28720) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13490 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28719) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13491 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28718) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13492 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28717) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13493 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28716) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13494 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28715) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13495 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28714) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13496 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 28709) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 28710) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 13497 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28708) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13498 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28707) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13499 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28706) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13500 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28705) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13501 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28704) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13502 (("$" 28669) ("(" 28670) (T_NAME_FULLY_QUALIFIED 28671) (T_NAME_QUALIFIED 28672) (T_NAME_RELATIVE 28673) (T_STATIC 28674) (T_STRING 28675) (T_VARIABLE 28676) (class_name 28677) (class_name_reference 28678) (name 28679) (new_variable 28680) (simple_variable 28681)) 13503 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28668) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13504 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28667) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13505 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28666) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13506 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28665) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13507 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28664) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13508 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28663) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13509 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28662) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13510 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28661) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13511 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28660) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13512 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28659) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13513 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28658) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13514 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28657) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13515 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28656) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13516 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28655) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13517 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28654) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13518 (("$" 28669) (T_VARIABLE 28676) (simple_variable 28700) ("{" 28701)) 13519 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 28698) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13520 ((T_PAAMAYIM_NEKUDOTAYIM 28696)) 13521 ((T_NULLSAFE_OBJECT_OPERATOR 28682) (T_OBJECT_OPERATOR 28683) (T_PAAMAYIM_NEKUDOTAYIM 28684) ("[" 28685)) 13522 (("$" 28669) (T_STRING 28689) (T_VARIABLE 28676) (property_name 28695) (simple_variable 28691) ("{" 28692)) 13523 (("$" 28669) (T_STRING 28689) (T_VARIABLE 28676) (property_name 28690) (simple_variable 28691) ("{" 28692)) 13524 (("$" 28669) (T_VARIABLE 28676) (simple_variable 28688)) 13525 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 28686) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 13526 (("]" 28687)) 13527 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 28693) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13528 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 28694)) 13529 (("$" 28669) (T_VARIABLE 28676) (simple_variable 28697)) 13530 (("%" 299) (")" 28699) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13531 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 28702) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13532 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 28703)) 13533 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28713) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13534 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 28711) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 13535 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28712) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13536 (("$" 28546) (T_ABSTRACT 28723) (T_ARRAY 28724) (T_AS 28725) (T_BREAK 28726) (T_CALLABLE 28727) (T_CASE 28728) (T_CATCH 28729) (T_CLASS 28730) (T_CLASS_C 28731) (T_CLONE 28732) (T_CONST 28733) (T_CONTINUE 28734) (T_DECLARE 28735) (T_DEFAULT 28736) (T_DIR 28737) (T_DO 28738) (T_ECHO 28739) (T_ELSE 28740) (T_ELSEIF 28741) (T_EMPTY 28742) (T_ENDDECLARE 28743) (T_ENDFOR 28744) (T_ENDFOREACH 28745) (T_ENDIF 28746) (T_ENDSWITCH 28747) (T_ENDWHILE 28748) (T_ENUM 28749) (T_EVAL 28750) (T_EXIT 28751) (T_EXTENDS 28752) (T_FILE 28753) (T_FINAL 28754) (T_FINALLY 28755) (T_FN 28756) (T_FOR 28757) (T_FOREACH 28758) (T_FUNCTION 28759) (T_FUNC_C 28760) (T_GLOBAL 28761) (T_GOTO 28762) (T_IF 28763) (T_IMPLEMENTS 28764) (T_INCLUDE 28765) (T_INCLUDE_ONCE 28766) (T_INSTANCEOF 28767) (T_INSTEADOF 28768) (T_INTERFACE 28769) (T_ISSET 28770) (T_LINE 28771) (T_LIST 28772) (T_LOGICAL_AND 28773) (T_LOGICAL_OR 28774) (T_LOGICAL_XOR 28775) (T_MATCH 28776) (T_METHOD_C 28777) (T_NAMESPACE 28778) (T_NEW 28779) (T_NS_C 28780) (T_PRINT 28781) (T_PRIVATE 28782) (T_PROPERTY_C 28783) (T_PROTECTED 28784) (T_PUBLIC 28785) (T_READONLY 28786) (T_REQUIRE 28787) (T_REQUIRE_ONCE 28788) (T_RETURN 28789) (T_STATIC 28790) (T_STRING 28791) (T_SWITCH 28792) (T_THROW 28793) (T_TRAIT 28794) (T_TRAIT_C 28795) (T_TRY 28796) (T_UNSET 28797) (T_USE 28798) (T_VAR 28799) (T_VARIABLE 28594) (T_WHILE 28800) (T_YIELD 28801) (identifier 28802) (member_name 28803) (reserved_non_modifiers 28804) (semi_reserved 28805) (simple_variable 28806) ("{" 28807)) 13537 (("(" 28810) (argument_list 28811)) 13538 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 28808) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13539 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 28809)) 13540 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 28812) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 28813) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 28814) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13541 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 28817) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13542 (("," 1480) (possible_comma 28815)) 13543 ((")" 28816)) 13544 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (ampersand 28847) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28848) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13545 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28846) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13546 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28845) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13547 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28844) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13548 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28843) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13549 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28842) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13550 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28841) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13551 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28840) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13552 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28839) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13553 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28838) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13554 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28837) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13555 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28836) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13556 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28835) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13557 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28834) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13558 (("\"" 1520) ("$" 28849) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 28850) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 28851) ("[" 1542) (array_object_dereferenceable 28852) (callable_expr 28853) (callable_variable 28854) (class_constant 1546) (class_name 28855) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 28856) (name 28857) (new_dereferenceable 1552) (simple_variable 28858) (static_member 28859) (variable 28860) (variable_class_name 28861)) 13559 (("$" 28849) (T_VARIABLE 28851) (simple_variable 28894) ("{" 28895)) 13560 (("(" 28865) (argument_list 28893)) 13561 ((T_NULLSAFE_OBJECT_OPERATOR 28879) (T_OBJECT_OPERATOR 28880) ("[" 28881)) 13562 (("(" 28865) (argument_list 28878)) 13563 ((T_PAAMAYIM_NEKUDOTAYIM 28874)) 13564 (("(" 28865) (argument_list 28873)) 13565 ((T_PAAMAYIM_NEKUDOTAYIM 28862)) 13566 (("$" 28849) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 28851) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 28863) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 28864) ("{" 1642)) 13567 (("(" 28865) (argument_list 28866)) 13568 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 28867) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 28868) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 28869) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13569 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 28872) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13570 (("," 1480) (possible_comma 28870)) 13571 ((")" 28871)) 13572 (("$" 28849) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 28851) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 28875) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 28876) ("{" 1658)) 13573 (("(" 28865) (argument_list 28877)) 13574 (("$" 28849) (T_STRING 28884) (T_VARIABLE 28851) (property_name 28891) (simple_variable 28886) ("{" 28887)) 13575 (("$" 28849) (T_STRING 28884) (T_VARIABLE 28851) (property_name 28885) (simple_variable 28886) ("{" 28887)) 13576 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 28882) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 13577 (("]" 28883)) 13578 (("(" 28865) (argument_list 28890)) 13579 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 28888) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13580 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 28889)) 13581 (("(" 28865) (argument_list 28892)) 13582 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 28896) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13583 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 28897)) 13584 ((backup_doc_comment 28900)) 13585 (("(" 28901)) 13586 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 28902)) 13587 ((")" 28903)) 13588 ((T_USE 164) (lexical_vars 28904)) 13589 ((":" 166) (return_type 28905)) 13590 ((backup_fn_flags 28906)) 13591 (("{" 28907)) 13592 ((inner_statement_list 28908)) 13593 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 28909) ("~" 127)) 13594 ((backup_fn_flags 28910)) 13595 ((backup_doc_comment 28912)) 13596 (("(" 28913)) 13597 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 28914)) 13598 ((")" 28915)) 13599 ((":" 2249) (return_type 28916)) 13600 ((T_DOUBLE_ARROW 28917)) 13601 ((backup_fn_flags 28918)) 13602 ((backup_lex_pos 28919)) 13603 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28920) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13604 (("%" 28624) ("*" 28625) ("+" 28626) ("-" 28627) ("." 28628) ("/" 28629) ("<" 28630) (">" 28631) ("?" 28632) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 28633) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28634) (T_BOOLEAN_AND 28635) (T_BOOLEAN_OR 28636) (T_COALESCE 28637) (T_INSTANCEOF 28638) (T_IS_EQUAL 28639) (T_IS_GREATER_OR_EQUAL 28640) (T_IS_IDENTICAL 28641) (T_IS_NOT_EQUAL 28642) (T_IS_NOT_IDENTICAL 28643) (T_IS_SMALLER_OR_EQUAL 28644) (T_LOGICAL_AND 28645) (T_LOGICAL_OR 28646) (T_LOGICAL_XOR 28647) (T_POW 28648) (T_SL 28649) (T_SPACESHIP 28650) (T_SR 28651) ("^" 28652) (backup_fn_flags 28921) ("|" 28653)) 13605 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 28923) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13606 (("$" 28546) (T_ABSTRACT 28723) (T_ARRAY 28724) (T_AS 28725) (T_BREAK 28726) (T_CALLABLE 28727) (T_CASE 28728) (T_CATCH 28729) (T_CLASS 28730) (T_CLASS_C 28731) (T_CLONE 28732) (T_CONST 28733) (T_CONTINUE 28734) (T_DECLARE 28735) (T_DEFAULT 28736) (T_DIR 28737) (T_DO 28738) (T_ECHO 28739) (T_ELSE 28740) (T_ELSEIF 28741) (T_EMPTY 28742) (T_ENDDECLARE 28743) (T_ENDFOR 28744) (T_ENDFOREACH 28745) (T_ENDIF 28746) (T_ENDSWITCH 28747) (T_ENDWHILE 28748) (T_ENUM 28749) (T_EVAL 28750) (T_EXIT 28751) (T_EXTENDS 28752) (T_FILE 28753) (T_FINAL 28754) (T_FINALLY 28755) (T_FN 28756) (T_FOR 28757) (T_FOREACH 28758) (T_FUNCTION 28759) (T_FUNC_C 28760) (T_GLOBAL 28761) (T_GOTO 28762) (T_IF 28763) (T_IMPLEMENTS 28764) (T_INCLUDE 28765) (T_INCLUDE_ONCE 28766) (T_INSTANCEOF 28767) (T_INSTEADOF 28768) (T_INTERFACE 28769) (T_ISSET 28770) (T_LINE 28771) (T_LIST 28772) (T_LOGICAL_AND 28773) (T_LOGICAL_OR 28774) (T_LOGICAL_XOR 28775) (T_MATCH 28776) (T_METHOD_C 28777) (T_NAMESPACE 28778) (T_NEW 28779) (T_NS_C 28780) (T_PRINT 28781) (T_PRIVATE 28782) (T_PROPERTY_C 28783) (T_PROTECTED 28784) (T_PUBLIC 28785) (T_READONLY 28786) (T_REQUIRE 28787) (T_REQUIRE_ONCE 28788) (T_RETURN 28789) (T_STATIC 28790) (T_STRING 28791) (T_SWITCH 28792) (T_THROW 28793) (T_TRAIT 28794) (T_TRAIT_C 28795) (T_TRY 28796) (T_UNSET 28797) (T_USE 28798) (T_VAR 28799) (T_VARIABLE 28594) (T_WHILE 28800) (T_YIELD 28801) (identifier 28925) (member_name 28926) (reserved_non_modifiers 28804) (semi_reserved 28805) (simple_variable 28927) ("{" 28928)) 13607 (("(" 28810) (argument_list 28931)) 13608 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 28929) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13609 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 28930)) 13610 ((T_FN 35) (T_FUNCTION 129) (fn 28608) #'28609 (inline_function 28935)) 13611 (("$" 28546) (T_STRING 28941) (T_VARIABLE 28594) (property_name 28948) (simple_variable 28943) ("{" 28944)) 13612 (("$" 28546) (T_STRING 28941) (T_VARIABLE 28594) (property_name 28942) (simple_variable 28943) ("{" 28944)) 13613 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 28939) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 13614 (("]" 28940)) 13615 (("(" 28810) (argument_list 28947)) 13616 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 28945) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13617 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 28946)) 13618 (("(" 28810) (argument_list 28949)) 13619 (("`" 28951)) 13620 (("]" 28953)) 13621 (("=" 28954)) 13622 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28955) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13623 (("%" 28624) ("*" 28625) ("+" 28626) ("-" 28627) ("." 28628) ("/" 28629) ("<" 28630) (">" 28631) ("?" 28632) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 28633) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28634) (T_BOOLEAN_AND 28635) (T_BOOLEAN_OR 28636) (T_COALESCE 28637) (T_DOUBLE_ARROW 28958) (T_INSTANCEOF 28638) (T_IS_EQUAL 28639) (T_IS_GREATER_OR_EQUAL 28640) (T_IS_IDENTICAL 28641) (T_IS_NOT_EQUAL 28642) (T_IS_NOT_IDENTICAL 28643) (T_IS_SMALLER_OR_EQUAL 28644) (T_LOGICAL_AND 28645) (T_LOGICAL_OR 28646) (T_LOGICAL_XOR 28647) (T_POW 28648) (T_SL 28649) (T_SPACESHIP 28650) (T_SR 28651) ("^" 28652) ("|" 28653)) 13624 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 28959) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13625 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 28968) (T_VARIABLE 2896) (encaps_var 2910)) 13626 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 28967) (T_VARIABLE 2896) (encaps_var 2901)) 13627 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29009) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13628 ((T_CLASS 29001)) 13629 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 29000) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 28981) (attribute 2958) (class_modifier 2939)) 13630 ((T_PAAMAYIM_NEKUDOTAYIM 28998)) 13631 (("(" 28865) (argument_list 28997)) 13632 ((T_NULLSAFE_OBJECT_OPERATOR 28988) (T_OBJECT_OPERATOR 28989) (T_PAAMAYIM_NEKUDOTAYIM 28990) ("[" 28991)) 13633 (("$" 28849) (T_STRING 28884) (T_VARIABLE 28851) (property_name 28996) (simple_variable 28886) ("{" 28887)) 13634 (("$" 28849) (T_STRING 28884) (T_VARIABLE 28851) (property_name 28995) (simple_variable 28886) ("{" 28887)) 13635 (("$" 28849) (T_VARIABLE 28851) (simple_variable 28994)) 13636 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 28992) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 13637 (("]" 28993)) 13638 (("$" 28849) (T_VARIABLE 28851) (simple_variable 28999)) 13639 (("(" 2960) (argument_list 2961) (ctor_arguments 29002)) 13640 ((T_EXTENDS 2963) (extends_from 29003)) 13641 ((T_IMPLEMENTS 2965) (implements_list 29004)) 13642 ((backup_doc_comment 29005)) 13643 (("{" 29006)) 13644 ((class_statement_list 29007)) 13645 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 29008)) 13646 (("%" 299) (")" 29010) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13647 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29012) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13648 (("%" 299) (")" 29013) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13649 (("{" 29014)) 13650 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 29015) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 13651 (("}" 29016)) 13652 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 29018) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 13653 ((")" 29019)) 13654 (("=" 29020)) 13655 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 29021) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13656 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 29023) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13657 (("," 5192) (possible_comma 29024)) 13658 ((")" 29025)) 13659 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 29033) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 29034) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 29035) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13660 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 29038) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13661 (("," 1480) (possible_comma 29036)) 13662 ((")" 29037)) 13663 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29040) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13664 (("%" 299) (")" 29041) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13665 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29043) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13666 (("%" 299) (")" 29044) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13667 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 29051) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 13668 ((")" 29052)) 13669 (("%" 299) (")" 29057) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13670 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 29060) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13671 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 29061)) 13672 (("\"" 29063) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 13673 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 29073) (T_VARIABLE 2896) (encaps_var 2910)) 13674 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 29072) (T_VARIABLE 2896) (encaps_var 2901)) 13675 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29114) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13676 ((T_CLASS 29106)) 13677 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 29105) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 29086) (attribute 2958) (class_modifier 2939)) 13678 ((T_PAAMAYIM_NEKUDOTAYIM 29103)) 13679 (("(" 28453) (argument_list 29102)) 13680 ((T_NULLSAFE_OBJECT_OPERATOR 29093) (T_OBJECT_OPERATOR 29094) (T_PAAMAYIM_NEKUDOTAYIM 29095) ("[" 29096)) 13681 (("$" 28437) (T_STRING 28472) (T_VARIABLE 28439) (property_name 29101) (simple_variable 28474) ("{" 28475)) 13682 (("$" 28437) (T_STRING 28472) (T_VARIABLE 28439) (property_name 29100) (simple_variable 28474) ("{" 28475)) 13683 (("$" 28437) (T_VARIABLE 28439) (simple_variable 29099)) 13684 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 29097) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 13685 (("]" 29098)) 13686 (("$" 28437) (T_VARIABLE 28439) (simple_variable 29104)) 13687 (("(" 2960) (argument_list 2961) (ctor_arguments 29107)) 13688 ((T_EXTENDS 2963) (extends_from 29108)) 13689 ((T_IMPLEMENTS 2965) (implements_list 29109)) 13690 ((backup_doc_comment 29110)) 13691 (("{" 29111)) 13692 ((class_statement_list 29112)) 13693 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 29113)) 13694 (("%" 299) (")" 29115) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13695 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29117) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13696 (("%" 299) (")" 29118) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13697 (("{" 29119)) 13698 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 29120) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 13699 (("}" 29121)) 13700 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 29123) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 13701 ((")" 29124)) 13702 (("=" 29125)) 13703 (("!" 28132) ("\"" 28133) ("$" 28134) ("(" 28135) ("+" 28136) ("-" 28137) ("@" 28138) (T_ARRAY 28139) (T_ARRAY_CAST 28140) (T_ATTRIBUTE 128) (T_BOOL_CAST 28141) (T_CLASS_C 28142) (T_CLONE 28143) (T_CONSTANT_ENCAPSED_STRING 28144) (T_DEC 28145) (T_DIR 28146) (T_DNUMBER 28147) (T_DOUBLE_CAST 28148) (T_EMPTY 28149) (T_EVAL 28150) (T_EXIT 28151) (T_FILE 28152) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28153) (T_INC 28154) (T_INCLUDE 28155) (T_INCLUDE_ONCE 28156) (T_INT_CAST 28157) (T_ISSET 28158) (T_LINE 28159) (T_LIST 28160) (T_LNUMBER 28161) (T_MATCH 28162) (T_METHOD_C 28163) (T_NAME_FULLY_QUALIFIED 28164) (T_NAME_QUALIFIED 28165) (T_NAME_RELATIVE 28166) (T_NEW 28167) (T_NS_C 28168) (T_OBJECT_CAST 28169) (T_PRINT 28170) (T_PROPERTY_C 28171) (T_READONLY 28172) (T_REQUIRE 28173) (T_REQUIRE_ONCE 28174) (T_START_HEREDOC 28175) (T_STATIC 28176) (T_STRING 28177) (T_STRING_CAST 28178) (T_THROW 28179) (T_TRAIT_C 28180) (T_UNSET_CAST 28181) (T_VARIABLE 28182) (T_YIELD 28183) (T_YIELD_FROM 28184) ("[" 28185) ("`" 28186) (array_object_dereferenceable 28187) (attribute 133) (attributes 28188) (callable_expr 28189) (callable_variable 28190) (class_constant 28191) (class_name 28192) (constant 28193) (dereferenceable_scalar 28194) (expr 29126) (fn 28196) (fully_dereferenceable 105) #'28197 (function_call 28198) (inline_function 28199) (internal_functions_in_yacc 28200) (match 28201) (name 28202) (new_dereferenceable 28203) (new_non_dereferenceable 28204) (scalar 28205) (simple_variable 28206) (static_member 28207) (variable 28208) (variable_class_name 28209) ("~" 28210)) 13704 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 29128) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13705 (("," 5192) (possible_comma 29129)) 13706 ((")" 29130)) 13707 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 29138) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 29139) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 29140) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13708 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 29143) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13709 (("," 1480) (possible_comma 29141)) 13710 ((")" 29142)) 13711 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29145) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13712 (("%" 299) (")" 29146) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13713 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29148) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13714 (("%" 299) (")" 29149) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13715 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 29156) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 13716 ((")" 29157)) 13717 (("%" 299) (")" 29162) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13718 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 29165) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13719 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 29166)) 13720 (("\"" 29168) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 13721 (("$" 4) (T_ABSTRACT 27792) (T_ARRAY 27793) (T_AS 27794) (T_BREAK 27795) (T_CALLABLE 27796) (T_CASE 27797) (T_CATCH 27798) (T_CLASS 27799) (T_CLASS_C 27800) (T_CLONE 27801) (T_CONST 27802) (T_CONTINUE 27803) (T_DECLARE 27804) (T_DEFAULT 27805) (T_DIR 27806) (T_DO 27807) (T_ECHO 27808) (T_ELSE 27809) (T_ELSEIF 27810) (T_EMPTY 27811) (T_ENDDECLARE 27812) (T_ENDFOR 27813) (T_ENDFOREACH 27814) (T_ENDIF 27815) (T_ENDSWITCH 27816) (T_ENDWHILE 27817) (T_ENUM 27818) (T_EVAL 27819) (T_EXIT 27820) (T_EXTENDS 27821) (T_FILE 27822) (T_FINAL 27823) (T_FINALLY 27824) (T_FN 27825) (T_FOR 27826) (T_FOREACH 27827) (T_FUNCTION 27828) (T_FUNC_C 27829) (T_GLOBAL 27830) (T_GOTO 27831) (T_IF 27832) (T_IMPLEMENTS 27833) (T_INCLUDE 27834) (T_INCLUDE_ONCE 27835) (T_INSTANCEOF 27836) (T_INSTEADOF 27837) (T_INTERFACE 27838) (T_ISSET 27839) (T_LINE 27840) (T_LIST 27841) (T_LOGICAL_AND 27842) (T_LOGICAL_OR 27843) (T_LOGICAL_XOR 27844) (T_MATCH 27845) (T_METHOD_C 27846) (T_NAMESPACE 27847) (T_NEW 27848) (T_NS_C 27849) (T_PRINT 27850) (T_PRIVATE 27851) (T_PROPERTY_C 27852) (T_PROTECTED 27853) (T_PUBLIC 27854) (T_READONLY 27855) (T_REQUIRE 27856) (T_REQUIRE_ONCE 27857) (T_RETURN 27858) (T_STATIC 27859) (T_STRING 27860) (T_SWITCH 27861) (T_THROW 27862) (T_TRAIT 27863) (T_TRAIT_C 27864) (T_TRY 27865) (T_UNSET 27866) (T_USE 27867) (T_VAR 27868) (T_VARIABLE 81) (T_WHILE 27869) (T_YIELD 27870) (identifier 29172) (member_name 29173) (reserved_non_modifiers 27873) (semi_reserved 27874) (simple_variable 29174) ("{" 29175)) 13722 (("(" 27778) (argument_list 29178)) 13723 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 29176) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13724 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 29177)) 13725 ((T_STRING 29180)) 13726 ((T_EXTENDS 2963) (extends_from 29181)) 13727 ((T_IMPLEMENTS 2965) (implements_list 29182)) 13728 ((backup_doc_comment 29183)) 13729 (("{" 29184)) 13730 ((class_statement_list 29185)) 13731 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 29186)) 13732 ((T_STRING 29238)) 13733 ((T_STRING 29231)) 13734 ((T_STRING 29225)) 13735 ((T_STRING 29220)) 13736 ((T_ABSTRACT 10) (T_CLASS 29212) (T_FINAL 34) (T_READONLY 2931) (class_modifier 17368)) 13737 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17349) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17350) (ampersand 17351) (returns_ref 29200)) 13738 ((T_READONLY 17353) (T_STRING 17354) (backup_doc_comment 141) (function_name 29201)) 13739 ((backup_doc_comment 29202)) 13740 (("(" 29203)) 13741 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 29204)) 13742 ((")" 29205)) 13743 ((":" 166) (return_type 29206)) 13744 ((backup_fn_flags 29207)) 13745 (("{" 29208)) 13746 ((inner_statement_list 29209)) 13747 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 29210) ("~" 127)) 13748 ((backup_fn_flags 29211)) 13749 ((T_STRING 29213)) 13750 ((T_EXTENDS 2963) (extends_from 29214)) 13751 ((T_IMPLEMENTS 2965) (implements_list 29215)) 13752 ((backup_doc_comment 29216)) 13753 (("{" 29217)) 13754 ((class_statement_list 29218)) 13755 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 29219)) 13756 ((backup_doc_comment 29221)) 13757 (("{" 29222)) 13758 ((class_statement_list 29223)) 13759 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 29224)) 13760 ((T_EXTENDS 21295) (interface_extends_list 29226)) 13761 ((backup_doc_comment 29227)) 13762 (("{" 29228)) 13763 ((class_statement_list 29229)) 13764 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 29230)) 13765 ((":" 21344) (enum_backing_type 29232)) 13766 ((T_IMPLEMENTS 2965) (implements_list 29233)) 13767 ((backup_doc_comment 29234)) 13768 (("{" 29235)) 13769 ((class_statement_list 29236)) 13770 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 29237)) 13771 ((T_EXTENDS 2963) (extends_from 29239)) 13772 ((T_IMPLEMENTS 2965) (implements_list 29240)) 13773 ((backup_doc_comment 29241)) 13774 (("{" 29242)) 13775 ((class_statement_list 29243)) 13776 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 29244)) 13777 (("$" 4) (T_STRING 29250) (T_VARIABLE 81) (property_name 29257) (simple_variable 29252) ("{" 29253)) 13778 (("$" 4) (T_STRING 29250) (T_VARIABLE 81) (property_name 29251) (simple_variable 29252) ("{" 29253)) 13779 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 29248) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 13780 (("]" 29249)) 13781 (("(" 27778) (argument_list 29256)) 13782 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 29254) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13783 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 29255)) 13784 (("(" 27778) (argument_list 29258)) 13785 ((":" 29262)) 13786 ((";" 29261)) 13787 ((inner_statement_list 29263)) 13788 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 29264) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 13789 ((";" 29265)) 13790 (("`" 29267)) 13791 (("]" 29269)) 13792 (("=" 29270)) 13793 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 29271) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13794 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29793) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13795 ((T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5234) (T_VARIABLE 5235) (encaps_list 29791) (encaps_var 5237)) 13796 (("$" 29275) (T_VARIABLE 29323) (simple_variable 29787) ("{" 29788)) 13797 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29785) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13798 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29784) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13799 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29783) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13800 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29782) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13801 (("(" 29779)) 13802 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29778) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13803 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29777) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13804 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29776) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13805 (("\"" 1520) ("$" 29578) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 29579) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 29580) ("[" 1542) (array_object_dereferenceable 29581) (callable_expr 29582) (callable_variable 29583) (class_constant 1546) (class_name 29584) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 29585) (name 29586) (new_dereferenceable 1552) (simple_variable 29587) (static_member 29588) (variable 29775) (variable_class_name 29590)) 13806 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29774) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13807 (("(" 29771)) 13808 (("(" 29768)) 13809 (("(" 29759) (argument_list 29760) (ctor_arguments 29761)) 13810 (("\"" 1520) ("$" 29578) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 29579) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 29580) ("[" 1542) (array_object_dereferenceable 29581) (callable_expr 29582) (callable_variable 29583) (class_constant 1546) (class_name 29584) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 29585) (name 29586) (new_dereferenceable 1552) (simple_variable 29587) (static_member 29588) (variable 29758) (variable_class_name 29590)) 13811 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29757) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13812 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29756) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13813 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29755) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13814 (("(" 29751)) 13815 (("(" 29746)) 13816 (("(" 29740)) 13817 (("$" 29578) ("(" 29703) (T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 29704) (T_NAME_QUALIFIED 29705) (T_NAME_RELATIVE 29706) (T_READONLY 2931) (T_STATIC 29707) (T_STRING 29708) (T_VARIABLE 29580) (anonymous_class 29709) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 29710) (attribute 2937) (attributes 29711) (class_modifier 2939) (class_name 29712) (class_name_reference 29713) (name 29714) (new_variable 29715) (simple_variable 29716)) 13818 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29702) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13819 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29701) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13820 (("(" 29539) (argument_list 29700)) 13821 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29699) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13822 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29698) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13823 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 29693) (T_END_HEREDOC 29694) (T_VARIABLE 2896) (encaps_list 29695) (encaps_var 2898)) 13824 ((T_FN 35) (T_FUNCTION 129) (fn 29337) #'29338 (inline_function 29692)) 13825 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29691) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13826 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29690) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13827 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29689) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13828 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29686) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13829 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29685) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13830 (("!" 1677) ("\"" 1678) ("$" 1679) ("(" 1680) ("+" 1681) ("-" 1682) ("@" 1683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1684) (T_ARRAY_CAST 1685) (T_ATTRIBUTE 128) (T_BOOL_CAST 1686) (T_CLASS_C 1687) (T_CLONE 1688) (T_CONSTANT_ENCAPSED_STRING 1689) (T_DEC 1690) (T_DIR 1691) (T_DNUMBER 1692) (T_DOUBLE_CAST 1693) (T_ELLIPSIS 1694) (T_EMPTY 1695) (T_EVAL 1696) (T_EXIT 1697) (T_FILE 1698) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1699) (T_INC 1700) (T_INCLUDE 1701) (T_INCLUDE_ONCE 1702) (T_INT_CAST 1703) (T_ISSET 1704) (T_LINE 1705) (T_LIST 1706) (T_LNUMBER 1707) (T_MATCH 1708) (T_METHOD_C 1709) (T_NAME_FULLY_QUALIFIED 1710) (T_NAME_QUALIFIED 1711) (T_NAME_RELATIVE 1712) (T_NEW 1713) (T_NS_C 1714) (T_OBJECT_CAST 1715) (T_PRINT 1716) (T_PROPERTY_C 1717) (T_READONLY 1718) (T_REQUIRE 1719) (T_REQUIRE_ONCE 1720) (T_START_HEREDOC 1721) (T_STATIC 1722) (T_STRING 1723) (T_STRING_CAST 1724) (T_THROW 1725) (T_TRAIT_C 1726) (T_UNSET_CAST 1727) (T_VARIABLE 1728) (T_YIELD 1729) (T_YIELD_FROM 1730) ("[" 1731) ("`" 1732) (ampersand 1733) (array_object_dereferenceable 1734) (array_pair 1735) (array_pair_list 29681) (attribute 133) (attributes 1737) (callable_expr 1738) (callable_variable 1739) (class_constant 1740) (class_name 1741) (constant 1742) (dereferenceable_scalar 1743) (expr 1744) (fn 1745) (fully_dereferenceable 105) #'1746 (function_call 1747) (inline_function 1748) (internal_functions_in_yacc 1749) (match 1750) (name 1751) (new_dereferenceable 1752) (new_non_dereferenceable 1753) (non_empty_array_pair_list 1754) (possible_array_pair 1755) (scalar 1756) (simple_variable 1757) (static_member 1758) (variable 1759) (variable_class_name 1760) ("~" 1761)) 13831 ((T_CURLY_OPEN 2799) (T_DOLLAR_OPEN_CURLY_BRACES 2800) (T_ENCAPSED_AND_WHITESPACE 2801) (T_VARIABLE 2802) (backticks_expr 29679) (encaps_list 2804) (encaps_var 2805)) 13832 ((T_NULLSAFE_OBJECT_OPERATOR 29665) (T_OBJECT_OPERATOR 29666) ("[" 29667)) 13833 ((T_ATTRIBUTE 128) (T_FN 35) (T_FUNCTION 129) (T_STATIC 29662) (attribute 2733) (fn 29337) #'29338 (inline_function 29663)) 13834 (("(" 29539) (argument_list 29661)) 13835 ((T_PAAMAYIM_NEKUDOTAYIM 29653)) 13836 (("%" 29353) ("*" 29354) ("+" 29355) ("-" 29356) ("." 29357) ("/" 29358) ("<" 29359) (">" 29360) ("?" 29361) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 29362) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 29363) (T_BOOLEAN_AND 29364) (T_BOOLEAN_OR 29365) (T_COALESCE 29366) (T_DOUBLE_ARROW 29651) (T_INSTANCEOF 29367) (T_IS_EQUAL 29368) (T_IS_GREATER_OR_EQUAL 29369) (T_IS_IDENTICAL 29370) (T_IS_NOT_EQUAL 29371) (T_IS_NOT_IDENTICAL 29372) (T_IS_SMALLER_OR_EQUAL 29373) (T_LOGICAL_AND 29374) (T_LOGICAL_OR 29375) (T_LOGICAL_XOR 29376) (T_POW 29377) (T_SL 29378) (T_SPACESHIP 29379) (T_SR 29380) ("^" 29381) ("|" 29382)) 13837 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 29640)) 13838 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 137) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 138) (ampersand 139) (returns_ref 29628)) 13839 (("(" 29539) (argument_list 29627)) 13840 (("=" 29547) (T_AND_EQUAL 29548) (T_COALESCE_EQUAL 29549) (T_CONCAT_EQUAL 29550) (T_DEC 29551) (T_DIV_EQUAL 29552) (T_INC 29553) (T_MINUS_EQUAL 29554) (T_MOD_EQUAL 29555) (T_MUL_EQUAL 29556) (T_OR_EQUAL 29557) (T_PLUS_EQUAL 29558) (T_POW_EQUAL 29559) (T_SL_EQUAL 29560) (T_SR_EQUAL 29561) (T_XOR_EQUAL 29562)) 13841 ((T_PAAMAYIM_NEKUDOTAYIM 29451)) 13842 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29352) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13843 (("%" 29353) ("*" 29354) ("+" 29355) ("-" 29356) ("." 29357) ("/" 29358) ("<" 29359) (">" 29360) ("?" 29361) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 29362) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 29363) (T_BOOLEAN_AND 29364) (T_BOOLEAN_OR 29365) (T_COALESCE 29366) (T_INSTANCEOF 29367) (T_IS_EQUAL 29368) (T_IS_GREATER_OR_EQUAL 29369) (T_IS_IDENTICAL 29370) (T_IS_NOT_EQUAL 29371) (T_IS_NOT_IDENTICAL 29372) (T_IS_SMALLER_OR_EQUAL 29373) (T_LOGICAL_AND 29374) (T_LOGICAL_OR 29375) (T_LOGICAL_XOR 29376) (T_POW 29377) (T_SL 29378) (T_SPACESHIP 29379) (T_SR 29380) ("^" 29381) ("|" 29382)) 13844 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29450) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13845 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29449) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13846 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29448) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13847 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29447) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13848 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29446) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13849 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29445) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13850 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29444) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13851 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29443) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13852 (("!" 676) ("\"" 677) ("$" 678) ("(" 679) ("+" 680) ("-" 681) (":" 29438) ("@" 683) (T_ARRAY 684) (T_ARRAY_CAST 685) (T_ATTRIBUTE 128) (T_BOOL_CAST 686) (T_CLASS_C 687) (T_CLONE 688) (T_CONSTANT_ENCAPSED_STRING 689) (T_DEC 690) (T_DIR 691) (T_DNUMBER 692) (T_DOUBLE_CAST 693) (T_EMPTY 694) (T_EVAL 695) (T_EXIT 696) (T_FILE 697) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 698) (T_INC 699) (T_INCLUDE 700) (T_INCLUDE_ONCE 701) (T_INT_CAST 702) (T_ISSET 703) (T_LINE 704) (T_LIST 705) (T_LNUMBER 706) (T_MATCH 707) (T_METHOD_C 708) (T_NAME_FULLY_QUALIFIED 709) (T_NAME_QUALIFIED 710) (T_NAME_RELATIVE 711) (T_NEW 712) (T_NS_C 713) (T_OBJECT_CAST 714) (T_PRINT 715) (T_PROPERTY_C 716) (T_READONLY 717) (T_REQUIRE 718) (T_REQUIRE_ONCE 719) (T_START_HEREDOC 720) (T_STATIC 721) (T_STRING 722) (T_STRING_CAST 723) (T_THROW 724) (T_TRAIT_C 725) (T_UNSET_CAST 726) (T_VARIABLE 727) (T_YIELD 728) (T_YIELD_FROM 729) ("[" 730) ("`" 731) (array_object_dereferenceable 732) (attribute 133) (attributes 733) (callable_expr 734) (callable_variable 735) (class_constant 736) (class_name 737) (constant 738) (dereferenceable_scalar 739) (expr 29439) (fn 741) (fully_dereferenceable 105) #'742 (function_call 743) (inline_function 744) (internal_functions_in_yacc 745) (match 746) (name 747) (new_dereferenceable 748) (new_non_dereferenceable 749) (scalar 750) (simple_variable 751) (static_member 752) (variable 753) (variable_class_name 754) ("~" 755)) 13853 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29437) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13854 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29436) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13855 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29435) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13856 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29434) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13857 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29433) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13858 (("$" 29398) ("(" 29399) (T_NAME_FULLY_QUALIFIED 29400) (T_NAME_QUALIFIED 29401) (T_NAME_RELATIVE 29402) (T_STATIC 29403) (T_STRING 29404) (T_VARIABLE 29405) (class_name 29406) (class_name_reference 29407) (name 29408) (new_variable 29409) (simple_variable 29410)) 13859 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29397) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13860 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29396) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13861 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29395) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13862 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29394) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13863 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29393) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13864 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29392) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13865 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29391) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13866 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29390) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13867 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29389) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13868 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29388) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13869 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29387) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13870 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29386) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13871 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29385) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13872 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29384) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13873 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29383) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13874 (("$" 29398) (T_VARIABLE 29405) (simple_variable 29429) ("{" 29430)) 13875 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29427) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13876 ((T_PAAMAYIM_NEKUDOTAYIM 29425)) 13877 ((T_NULLSAFE_OBJECT_OPERATOR 29411) (T_OBJECT_OPERATOR 29412) (T_PAAMAYIM_NEKUDOTAYIM 29413) ("[" 29414)) 13878 (("$" 29398) (T_STRING 29418) (T_VARIABLE 29405) (property_name 29424) (simple_variable 29420) ("{" 29421)) 13879 (("$" 29398) (T_STRING 29418) (T_VARIABLE 29405) (property_name 29419) (simple_variable 29420) ("{" 29421)) 13880 (("$" 29398) (T_VARIABLE 29405) (simple_variable 29417)) 13881 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 29415) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 13882 (("]" 29416)) 13883 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 29422) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13884 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 29423)) 13885 (("$" 29398) (T_VARIABLE 29405) (simple_variable 29426)) 13886 (("%" 299) (")" 29428) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 13887 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 29431) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13888 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 29432)) 13889 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29442) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13890 (("%" 757) ("*" 758) ("+" 759) ("-" 760) ("." 761) ("/" 762) (":" 29440) ("<" 763) (">" 764) ("?" 765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 766) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 767) (T_BOOLEAN_AND 768) (T_BOOLEAN_OR 769) (T_COALESCE 770) (T_INSTANCEOF 771) (T_IS_EQUAL 772) (T_IS_GREATER_OR_EQUAL 773) (T_IS_IDENTICAL 774) (T_IS_NOT_EQUAL 775) (T_IS_NOT_IDENTICAL 776) (T_IS_SMALLER_OR_EQUAL 777) (T_LOGICAL_AND 778) (T_LOGICAL_OR 779) (T_LOGICAL_XOR 780) (T_POW 781) (T_SL 782) (T_SPACESHIP 783) (T_SR 784) ("^" 785) ("|" 786)) 13891 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29441) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13892 (("$" 29275) (T_ABSTRACT 29452) (T_ARRAY 29453) (T_AS 29454) (T_BREAK 29455) (T_CALLABLE 29456) (T_CASE 29457) (T_CATCH 29458) (T_CLASS 29459) (T_CLASS_C 29460) (T_CLONE 29461) (T_CONST 29462) (T_CONTINUE 29463) (T_DECLARE 29464) (T_DEFAULT 29465) (T_DIR 29466) (T_DO 29467) (T_ECHO 29468) (T_ELSE 29469) (T_ELSEIF 29470) (T_EMPTY 29471) (T_ENDDECLARE 29472) (T_ENDFOR 29473) (T_ENDFOREACH 29474) (T_ENDIF 29475) (T_ENDSWITCH 29476) (T_ENDWHILE 29477) (T_ENUM 29478) (T_EVAL 29479) (T_EXIT 29480) (T_EXTENDS 29481) (T_FILE 29482) (T_FINAL 29483) (T_FINALLY 29484) (T_FN 29485) (T_FOR 29486) (T_FOREACH 29487) (T_FUNCTION 29488) (T_FUNC_C 29489) (T_GLOBAL 29490) (T_GOTO 29491) (T_IF 29492) (T_IMPLEMENTS 29493) (T_INCLUDE 29494) (T_INCLUDE_ONCE 29495) (T_INSTANCEOF 29496) (T_INSTEADOF 29497) (T_INTERFACE 29498) (T_ISSET 29499) (T_LINE 29500) (T_LIST 29501) (T_LOGICAL_AND 29502) (T_LOGICAL_OR 29503) (T_LOGICAL_XOR 29504) (T_MATCH 29505) (T_METHOD_C 29506) (T_NAMESPACE 29507) (T_NEW 29508) (T_NS_C 29509) (T_PRINT 29510) (T_PRIVATE 29511) (T_PROPERTY_C 29512) (T_PROTECTED 29513) (T_PUBLIC 29514) (T_READONLY 29515) (T_REQUIRE 29516) (T_REQUIRE_ONCE 29517) (T_RETURN 29518) (T_STATIC 29519) (T_STRING 29520) (T_SWITCH 29521) (T_THROW 29522) (T_TRAIT 29523) (T_TRAIT_C 29524) (T_TRY 29525) (T_UNSET 29526) (T_USE 29527) (T_VAR 29528) (T_VARIABLE 29323) (T_WHILE 29529) (T_YIELD 29530) (identifier 29531) (member_name 29532) (reserved_non_modifiers 29533) (semi_reserved 29534) (simple_variable 29535) ("{" 29536)) 13893 (("(" 29539) (argument_list 29540)) 13894 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 29537) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13895 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 29538)) 13896 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 29541) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 29542) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 29543) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13897 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 29546) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13898 (("," 1480) (possible_comma 29544)) 13899 ((")" 29545)) 13900 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (ampersand 29576) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29577) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13901 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29575) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13902 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29574) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13903 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29573) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13904 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29572) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13905 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29571) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13906 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29570) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13907 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29569) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13908 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29568) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13909 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29567) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13910 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29566) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13911 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29565) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13912 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29564) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13913 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29563) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13914 (("\"" 1520) ("$" 29578) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 29579) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 29580) ("[" 1542) (array_object_dereferenceable 29581) (callable_expr 29582) (callable_variable 29583) (class_constant 1546) (class_name 29584) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 29585) (name 29586) (new_dereferenceable 1552) (simple_variable 29587) (static_member 29588) (variable 29589) (variable_class_name 29590)) 13915 (("$" 29578) (T_VARIABLE 29580) (simple_variable 29623) ("{" 29624)) 13916 (("(" 29594) (argument_list 29622)) 13917 ((T_NULLSAFE_OBJECT_OPERATOR 29608) (T_OBJECT_OPERATOR 29609) ("[" 29610)) 13918 (("(" 29594) (argument_list 29607)) 13919 ((T_PAAMAYIM_NEKUDOTAYIM 29603)) 13920 (("(" 29594) (argument_list 29602)) 13921 ((T_PAAMAYIM_NEKUDOTAYIM 29591)) 13922 (("$" 29578) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 29580) (T_WHILE 1635) (T_YIELD 1636) (identifier 1637) (member_name 29592) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 29593) ("{" 1642)) 13923 (("(" 29594) (argument_list 29595)) 13924 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 29596) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 29597) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 29598) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13925 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 29601) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 13926 (("," 1480) (possible_comma 29599)) 13927 ((")" 29600)) 13928 (("$" 29578) (T_ABSTRACT 1558) (T_ARRAY 1559) (T_AS 1560) (T_BREAK 1561) (T_CALLABLE 1562) (T_CASE 1563) (T_CATCH 1564) (T_CLASS 1565) (T_CLASS_C 1566) (T_CLONE 1567) (T_CONST 1568) (T_CONTINUE 1569) (T_DECLARE 1570) (T_DEFAULT 1571) (T_DIR 1572) (T_DO 1573) (T_ECHO 1574) (T_ELSE 1575) (T_ELSEIF 1576) (T_EMPTY 1577) (T_ENDDECLARE 1578) (T_ENDFOR 1579) (T_ENDFOREACH 1580) (T_ENDIF 1581) (T_ENDSWITCH 1582) (T_ENDWHILE 1583) (T_ENUM 1584) (T_EVAL 1585) (T_EXIT 1586) (T_EXTENDS 1587) (T_FILE 1588) (T_FINAL 1589) (T_FINALLY 1590) (T_FN 1591) (T_FOR 1592) (T_FOREACH 1593) (T_FUNCTION 1594) (T_FUNC_C 1595) (T_GLOBAL 1596) (T_GOTO 1597) (T_IF 1598) (T_IMPLEMENTS 1599) (T_INCLUDE 1600) (T_INCLUDE_ONCE 1601) (T_INSTANCEOF 1602) (T_INSTEADOF 1603) (T_INTERFACE 1604) (T_ISSET 1605) (T_LINE 1606) (T_LIST 1607) (T_LOGICAL_AND 1608) (T_LOGICAL_OR 1609) (T_LOGICAL_XOR 1610) (T_MATCH 1611) (T_METHOD_C 1612) (T_NAMESPACE 1613) (T_NEW 1614) (T_NS_C 1615) (T_PRINT 1616) (T_PRIVATE 1617) (T_PROPERTY_C 1618) (T_PROTECTED 1619) (T_PUBLIC 1620) (T_READONLY 1621) (T_REQUIRE 1622) (T_REQUIRE_ONCE 1623) (T_RETURN 1624) (T_STATIC 1625) (T_STRING 1626) (T_SWITCH 1627) (T_THROW 1628) (T_TRAIT 1629) (T_TRAIT_C 1630) (T_TRY 1631) (T_UNSET 1632) (T_USE 1633) (T_VAR 1634) (T_VARIABLE 29580) (T_WHILE 1635) (T_YIELD 1636) (identifier 1655) (member_name 29604) (reserved_non_modifiers 1639) (semi_reserved 1640) (simple_variable 29605) ("{" 1658)) 13929 (("(" 29594) (argument_list 29606)) 13930 (("$" 29578) (T_STRING 29613) (T_VARIABLE 29580) (property_name 29620) (simple_variable 29615) ("{" 29616)) 13931 (("$" 29578) (T_STRING 29613) (T_VARIABLE 29580) (property_name 29614) (simple_variable 29615) ("{" 29616)) 13932 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 29611) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 13933 (("]" 29612)) 13934 (("(" 29594) (argument_list 29619)) 13935 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 29617) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13936 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 29618)) 13937 (("(" 29594) (argument_list 29621)) 13938 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 29625) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13939 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 29626)) 13940 ((backup_doc_comment 29629)) 13941 (("(" 29630)) 13942 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 29631)) 13943 ((")" 29632)) 13944 ((T_USE 164) (lexical_vars 29633)) 13945 ((":" 166) (return_type 29634)) 13946 ((backup_fn_flags 29635)) 13947 (("{" 29636)) 13948 ((inner_statement_list 29637)) 13949 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 29638) ("~" 127)) 13950 ((backup_fn_flags 29639)) 13951 ((backup_doc_comment 29641)) 13952 (("(" 29642)) 13953 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 29643)) 13954 ((")" 29644)) 13955 ((":" 2249) (return_type 29645)) 13956 ((T_DOUBLE_ARROW 29646)) 13957 ((backup_fn_flags 29647)) 13958 ((backup_lex_pos 29648)) 13959 (("!" 28544) ("\"" 28545) ("$" 28546) ("(" 28547) ("+" 28548) ("-" 28549) ("@" 28550) (T_ARRAY 28551) (T_ARRAY_CAST 28552) (T_ATTRIBUTE 128) (T_BOOL_CAST 28553) (T_CLASS_C 28554) (T_CLONE 28555) (T_CONSTANT_ENCAPSED_STRING 28556) (T_DEC 28557) (T_DIR 28558) (T_DNUMBER 28559) (T_DOUBLE_CAST 28560) (T_EMPTY 28561) (T_EVAL 28562) (T_EXIT 28563) (T_FILE 28564) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 28565) (T_INC 28566) (T_INCLUDE 28567) (T_INCLUDE_ONCE 28568) (T_INT_CAST 28569) (T_ISSET 28570) (T_LINE 28571) (T_LIST 28572) (T_LNUMBER 28573) (T_MATCH 28574) (T_METHOD_C 28575) (T_NAME_FULLY_QUALIFIED 28576) (T_NAME_QUALIFIED 28577) (T_NAME_RELATIVE 28578) (T_NEW 28579) (T_NS_C 28580) (T_OBJECT_CAST 28581) (T_PRINT 28582) (T_PROPERTY_C 28583) (T_READONLY 28584) (T_REQUIRE 28585) (T_REQUIRE_ONCE 28586) (T_START_HEREDOC 28587) (T_STATIC 28588) (T_STRING 28589) (T_STRING_CAST 28590) (T_THROW 28591) (T_TRAIT_C 28592) (T_UNSET_CAST 28593) (T_VARIABLE 28594) (T_YIELD 28595) (T_YIELD_FROM 28596) ("[" 28597) ("`" 28598) (array_object_dereferenceable 28599) (attribute 133) (attributes 28600) (callable_expr 28601) (callable_variable 28602) (class_constant 28603) (class_name 28604) (constant 28605) (dereferenceable_scalar 28606) (expr 29649) (fn 28608) (fully_dereferenceable 105) #'28609 (function_call 28610) (inline_function 28611) (internal_functions_in_yacc 28612) (match 28613) (name 28614) (new_dereferenceable 28615) (new_non_dereferenceable 28616) (scalar 28617) (simple_variable 28618) (static_member 28619) (variable 28620) (variable_class_name 28621) ("~" 28622)) 13960 (("%" 28624) ("*" 28625) ("+" 28626) ("-" 28627) ("." 28628) ("/" 28629) ("<" 28630) (">" 28631) ("?" 28632) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 28633) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28634) (T_BOOLEAN_AND 28635) (T_BOOLEAN_OR 28636) (T_COALESCE 28637) (T_INSTANCEOF 28638) (T_IS_EQUAL 28639) (T_IS_GREATER_OR_EQUAL 28640) (T_IS_IDENTICAL 28641) (T_IS_NOT_EQUAL 28642) (T_IS_NOT_IDENTICAL 28643) (T_IS_SMALLER_OR_EQUAL 28644) (T_LOGICAL_AND 28645) (T_LOGICAL_OR 28646) (T_LOGICAL_XOR 28647) (T_POW 28648) (T_SL 28649) (T_SPACESHIP 28650) (T_SR 28651) ("^" 28652) (backup_fn_flags 29650) ("|" 28653)) 13961 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 29652) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 13962 (("$" 29275) (T_ABSTRACT 29452) (T_ARRAY 29453) (T_AS 29454) (T_BREAK 29455) (T_CALLABLE 29456) (T_CASE 29457) (T_CATCH 29458) (T_CLASS 29459) (T_CLASS_C 29460) (T_CLONE 29461) (T_CONST 29462) (T_CONTINUE 29463) (T_DECLARE 29464) (T_DEFAULT 29465) (T_DIR 29466) (T_DO 29467) (T_ECHO 29468) (T_ELSE 29469) (T_ELSEIF 29470) (T_EMPTY 29471) (T_ENDDECLARE 29472) (T_ENDFOR 29473) (T_ENDFOREACH 29474) (T_ENDIF 29475) (T_ENDSWITCH 29476) (T_ENDWHILE 29477) (T_ENUM 29478) (T_EVAL 29479) (T_EXIT 29480) (T_EXTENDS 29481) (T_FILE 29482) (T_FINAL 29483) (T_FINALLY 29484) (T_FN 29485) (T_FOR 29486) (T_FOREACH 29487) (T_FUNCTION 29488) (T_FUNC_C 29489) (T_GLOBAL 29490) (T_GOTO 29491) (T_IF 29492) (T_IMPLEMENTS 29493) (T_INCLUDE 29494) (T_INCLUDE_ONCE 29495) (T_INSTANCEOF 29496) (T_INSTEADOF 29497) (T_INTERFACE 29498) (T_ISSET 29499) (T_LINE 29500) (T_LIST 29501) (T_LOGICAL_AND 29502) (T_LOGICAL_OR 29503) (T_LOGICAL_XOR 29504) (T_MATCH 29505) (T_METHOD_C 29506) (T_NAMESPACE 29507) (T_NEW 29508) (T_NS_C 29509) (T_PRINT 29510) (T_PRIVATE 29511) (T_PROPERTY_C 29512) (T_PROTECTED 29513) (T_PUBLIC 29514) (T_READONLY 29515) (T_REQUIRE 29516) (T_REQUIRE_ONCE 29517) (T_RETURN 29518) (T_STATIC 29519) (T_STRING 29520) (T_SWITCH 29521) (T_THROW 29522) (T_TRAIT 29523) (T_TRAIT_C 29524) (T_TRY 29525) (T_UNSET 29526) (T_USE 29527) (T_VAR 29528) (T_VARIABLE 29323) (T_WHILE 29529) (T_YIELD 29530) (identifier 29654) (member_name 29655) (reserved_non_modifiers 29533) (semi_reserved 29534) (simple_variable 29656) ("{" 29657)) 13963 (("(" 29539) (argument_list 29660)) 13964 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 29658) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13965 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 29659)) 13966 ((T_FN 35) (T_FUNCTION 129) (fn 29337) #'29338 (inline_function 29664)) 13967 (("$" 29275) (T_STRING 29670) (T_VARIABLE 29323) (property_name 29677) (simple_variable 29672) ("{" 29673)) 13968 (("$" 29275) (T_STRING 29670) (T_VARIABLE 29323) (property_name 29671) (simple_variable 29672) ("{" 29673)) 13969 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 29668) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 13970 (("]" 29669)) 13971 (("(" 29539) (argument_list 29676)) 13972 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 29674) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 13973 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 29675)) 13974 (("(" 29539) (argument_list 29678)) 13975 (("`" 29680)) 13976 (("]" 29682)) 13977 (("=" 29683)) 13978 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29684) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13979 (("%" 29353) ("*" 29354) ("+" 29355) ("-" 29356) ("." 29357) ("/" 29358) ("<" 29359) (">" 29360) ("?" 29361) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 29362) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 29363) (T_BOOLEAN_AND 29364) (T_BOOLEAN_OR 29365) (T_COALESCE 29366) (T_DOUBLE_ARROW 29687) (T_INSTANCEOF 29367) (T_IS_EQUAL 29368) (T_IS_GREATER_OR_EQUAL 29369) (T_IS_IDENTICAL 29370) (T_IS_NOT_EQUAL 29371) (T_IS_NOT_IDENTICAL 29372) (T_IS_SMALLER_OR_EQUAL 29373) (T_LOGICAL_AND 29374) (T_LOGICAL_OR 29375) (T_LOGICAL_XOR 29376) (T_POW 29377) (T_SL 29378) (T_SPACESHIP 29379) (T_SR 29380) ("^" 29381) ("|" 29382)) 13980 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29688) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 13981 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 29697) (T_VARIABLE 2896) (encaps_var 2910)) 13982 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 29696) (T_VARIABLE 2896) (encaps_var 2901)) 13983 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29738) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 13984 ((T_CLASS 29730)) 13985 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 29729) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 29710) (attribute 2958) (class_modifier 2939)) 13986 ((T_PAAMAYIM_NEKUDOTAYIM 29727)) 13987 (("(" 29594) (argument_list 29726)) 13988 ((T_NULLSAFE_OBJECT_OPERATOR 29717) (T_OBJECT_OPERATOR 29718) (T_PAAMAYIM_NEKUDOTAYIM 29719) ("[" 29720)) 13989 (("$" 29578) (T_STRING 29613) (T_VARIABLE 29580) (property_name 29725) (simple_variable 29615) ("{" 29616)) 13990 (("$" 29578) (T_STRING 29613) (T_VARIABLE 29580) (property_name 29724) (simple_variable 29615) ("{" 29616)) 13991 (("$" 29578) (T_VARIABLE 29580) (simple_variable 29723)) 13992 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 29721) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 13993 (("]" 29722)) 13994 (("$" 29578) (T_VARIABLE 29580) (simple_variable 29728)) 13995 (("(" 2960) (argument_list 2961) (ctor_arguments 29731)) 13996 ((T_EXTENDS 2963) (extends_from 29732)) 13997 ((T_IMPLEMENTS 2965) (implements_list 29733)) 13998 ((backup_doc_comment 29734)) 13999 (("{" 29735)) 14000 ((class_statement_list 29736)) 14001 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 29737)) 14002 (("%" 299) (")" 29739) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14003 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29741) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14004 (("%" 299) (")" 29742) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14005 (("{" 29743)) 14006 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 29744) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 14007 (("}" 29745)) 14008 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 29747) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 14009 ((")" 29748)) 14010 (("=" 29749)) 14011 (("!" 29273) ("\"" 29274) ("$" 29275) ("(" 29276) ("+" 29277) ("-" 29278) ("@" 29279) (T_ARRAY 29280) (T_ARRAY_CAST 29281) (T_ATTRIBUTE 128) (T_BOOL_CAST 29282) (T_CLASS_C 29283) (T_CLONE 29284) (T_CONSTANT_ENCAPSED_STRING 29285) (T_DEC 29286) (T_DIR 29287) (T_DNUMBER 29288) (T_DOUBLE_CAST 29289) (T_EMPTY 29290) (T_EVAL 29291) (T_EXIT 29292) (T_FILE 29293) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 29294) (T_INC 29295) (T_INCLUDE 29296) (T_INCLUDE_ONCE 29297) (T_INT_CAST 29298) (T_ISSET 29299) (T_LINE 29300) (T_LIST 29301) (T_LNUMBER 29302) (T_MATCH 29303) (T_METHOD_C 29304) (T_NAME_FULLY_QUALIFIED 29305) (T_NAME_QUALIFIED 29306) (T_NAME_RELATIVE 29307) (T_NEW 29308) (T_NS_C 29309) (T_OBJECT_CAST 29310) (T_PRINT 29311) (T_PROPERTY_C 29312) (T_READONLY 29313) (T_REQUIRE 29314) (T_REQUIRE_ONCE 29315) (T_START_HEREDOC 29316) (T_STATIC 29317) (T_STRING 29318) (T_STRING_CAST 29319) (T_THROW 29320) (T_TRAIT_C 29321) (T_UNSET_CAST 29322) (T_VARIABLE 29323) (T_YIELD 29324) (T_YIELD_FROM 29325) ("[" 29326) ("`" 29327) (array_object_dereferenceable 29328) (attribute 133) (attributes 29329) (callable_expr 29330) (callable_variable 29331) (class_constant 29332) (class_name 29333) (constant 29334) (dereferenceable_scalar 29335) (expr 29750) (fn 29337) (fully_dereferenceable 105) #'29338 (function_call 29339) (inline_function 29340) (internal_functions_in_yacc 29341) (match 29342) (name 29343) (new_dereferenceable 29344) (new_non_dereferenceable 29345) (scalar 29346) (simple_variable 29347) (static_member 29348) (variable 29349) (variable_class_name 29350) ("~" 29351)) 14012 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 29752) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 14013 (("," 5192) (possible_comma 29753)) 14014 ((")" 29754)) 14015 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 29762) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 29763) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 29764) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 14016 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 29767) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 14017 (("," 1480) (possible_comma 29765)) 14018 ((")" 29766)) 14019 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29769) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14020 (("%" 299) (")" 29770) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14021 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29772) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14022 (("%" 299) (")" 29773) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14023 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 29780) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 14024 ((")" 29781)) 14025 (("%" 299) (")" 29786) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14026 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 29789) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 14027 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 29790)) 14028 (("\"" 29792) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240)) 14029 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29795) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14030 (("%" 299) (")" 29796) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14031 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 29797) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 68) (T_START_HEREDOC 69) (T_STATIC 70) (T_STRING 71) (T_STRING_CAST 72) (T_SWITCH 73) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 77) (T_UNSET 78) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 82) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 87) (alt_if_stmt_without_else 88) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 103) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 109) (if_stmt_without_else 110) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 29798) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 29799) ("{" 126) ("~" 127)) 14032 ((inner_statement_list 29800)) 14033 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 29801) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 14034 ((";" 29802)) 14035 ((T_AS 29825) (T_NS_SEPARATOR 29843)) 14036 ((";" 29842)) 14037 (("," 29817) (";" 29841)) 14038 ((T_NAME_FULLY_QUALIFIED 29805) (T_NAME_QUALIFIED 29806) (T_STRING 29807) (group_use_declaration 29814) (legacy_namespace_name 29815) (namespace_name 29810) (use_declaration 29811) (use_declarations 29816)) 14039 ((";" 29840)) 14040 ((T_AS 29825) (T_NS_SEPARATOR 29827)) 14041 (("," 29817) (";" 29818)) 14042 ((T_NAME_FULLY_QUALIFIED 29819) (T_NAME_QUALIFIED 29820) (T_STRING 29821) (legacy_namespace_name 29822) (namespace_name 29823) (use_declaration 29824)) 14043 ((T_AS 29825)) 14044 ((T_STRING 29826)) 14045 (("{" 29828)) 14046 ((T_NAME_QUALIFIED 29829) (T_STRING 29830) (namespace_name 29831) (unprefixed_use_declaration 29832) (unprefixed_use_declarations 29833)) 14047 ((T_AS 29838)) 14048 (("," 29834) (possible_comma 29835)) 14049 ((T_NAME_QUALIFIED 29829) (T_STRING 29830) (namespace_name 29831) (unprefixed_use_declaration 29837)) 14050 (("}" 29836)) 14051 ((T_STRING 29839)) 14052 (("{" 29844)) 14053 ((T_CONST 29845) (T_FUNCTION 29846) (T_NAME_QUALIFIED 29829) (T_STRING 29830) (inline_use_declaration 29847) (inline_use_declarations 29848) (namespace_name 29831) (unprefixed_use_declaration 29849) (use_type 29850)) 14054 (("," 29852) (possible_comma 29853)) 14055 ((T_NAME_QUALIFIED 29829) (T_STRING 29830) (namespace_name 29831) (unprefixed_use_declaration 29851)) 14056 ((T_CONST 29845) (T_FUNCTION 29846) (T_NAME_QUALIFIED 29829) (T_STRING 29830) (inline_use_declaration 29855) (namespace_name 29831) (unprefixed_use_declaration 29849) (use_type 29850)) 14057 (("}" 29854)) 14058 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 29858) (variable 17576) (variable_class_name 17577)) 14059 (("," 17589) (possible_comma 29859)) 14060 ((")" 29860)) 14061 ((";" 29861)) 14062 ((inner_statement_list 29863)) 14063 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 29864) ("~" 127)) 14064 ((catch_list 29865)) 14065 ((T_CATCH 29866) (T_FINALLY 29867) (finally_statement 29868)) 14066 (("(" 29872)) 14067 (("{" 29869)) 14068 ((inner_statement_list 29870)) 14069 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 29871) ("~" 127)) 14070 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 29873) (class_name 17636) (name 17637)) 14071 ((T_VARIABLE 17638) (optional_variable 29874) ("|" 17640)) 14072 ((")" 29875)) 14073 (("{" 29876)) 14074 ((inner_statement_list 29877)) 14075 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 29878) ("~" 127)) 14076 ((backup_doc_comment 29880)) 14077 (("{" 29881)) 14078 ((class_statement_list 29882)) 14079 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 29883)) 14080 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29886) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14081 (("%" 299) (")" 29887) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14082 ((":" 29888) (switch_case_list 29889) ("{" 29890)) 14083 ((";" 29896) (case_list 29897)) 14084 ((";" 29891) (case_list 29892)) 14085 ((case_list 29894)) 14086 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 29893)) 14087 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 29895)) 14088 ((case_list 29900)) 14089 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 29898)) 14090 ((";" 29899)) 14091 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 29901)) 14092 ((";" 29902)) 14093 (("," 17937) (";" 29906)) 14094 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_END_HEREDOC 29911) (T_VARIABLE 2896) (encaps_var 2910)) 14095 ((T_CURLY_OPEN 2892) (T_DOLLAR_OPEN_CURLY_BRACES 2893) (T_ENCAPSED_AND_WHITESPACE 2899) (T_END_HEREDOC 29910) (T_VARIABLE 2896) (encaps_var 2901)) 14096 ((";" 29913)) 14097 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 29953) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14098 ((T_CLASS 29945)) 14099 ((T_ABSTRACT 10) (T_ATTRIBUTE 2927) (T_FINAL 34) (T_READONLY 2931) (anonymous_class 29944) (anonymous_class_modifiers 2935) (anonymous_class_modifiers_optional 29925) (attribute 2958) (class_modifier 2939)) 14100 ((T_PAAMAYIM_NEKUDOTAYIM 29942)) 14101 (("(" 27927) (argument_list 29941)) 14102 ((T_NULLSAFE_OBJECT_OPERATOR 29932) (T_OBJECT_OPERATOR 29933) (T_PAAMAYIM_NEKUDOTAYIM 29934) ("[" 29935)) 14103 (("$" 27911) (T_STRING 27946) (T_VARIABLE 27913) (property_name 29940) (simple_variable 27948) ("{" 27949)) 14104 (("$" 27911) (T_STRING 27946) (T_VARIABLE 27913) (property_name 29939) (simple_variable 27948) ("{" 27949)) 14105 (("$" 27911) (T_VARIABLE 27913) (simple_variable 29938)) 14106 (("!" 361) ("\"" 362) ("$" 363) ("(" 364) ("+" 365) ("-" 366) ("@" 367) (T_ARRAY 368) (T_ARRAY_CAST 369) (T_ATTRIBUTE 128) (T_BOOL_CAST 370) (T_CLASS_C 371) (T_CLONE 372) (T_CONSTANT_ENCAPSED_STRING 373) (T_DEC 374) (T_DIR 375) (T_DNUMBER 376) (T_DOUBLE_CAST 377) (T_EMPTY 378) (T_EVAL 379) (T_EXIT 380) (T_FILE 381) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 382) (T_INC 383) (T_INCLUDE 384) (T_INCLUDE_ONCE 385) (T_INT_CAST 386) (T_ISSET 387) (T_LINE 388) (T_LIST 389) (T_LNUMBER 390) (T_MATCH 391) (T_METHOD_C 392) (T_NAME_FULLY_QUALIFIED 393) (T_NAME_QUALIFIED 394) (T_NAME_RELATIVE 395) (T_NEW 396) (T_NS_C 397) (T_OBJECT_CAST 398) (T_PRINT 399) (T_PROPERTY_C 400) (T_READONLY 401) (T_REQUIRE 402) (T_REQUIRE_ONCE 403) (T_START_HEREDOC 404) (T_STATIC 405) (T_STRING 406) (T_STRING_CAST 407) (T_THROW 408) (T_TRAIT_C 409) (T_UNSET_CAST 410) (T_VARIABLE 411) (T_YIELD 412) (T_YIELD_FROM 413) ("[" 414) ("`" 415) (array_object_dereferenceable 416) (attribute 133) (attributes 417) (callable_expr 418) (callable_variable 419) (class_constant 420) (class_name 421) (constant 422) (dereferenceable_scalar 423) (expr 424) (fn 425) (fully_dereferenceable 105) #'426 (function_call 427) (inline_function 428) (internal_functions_in_yacc 429) (match 430) (name 431) (new_dereferenceable 432) (new_non_dereferenceable 433) (optional_expr 29936) (scalar 435) (simple_variable 436) (static_member 437) (variable 438) (variable_class_name 439) ("~" 440)) 14107 (("]" 29937)) 14108 (("$" 27911) (T_VARIABLE 27913) (simple_variable 29943)) 14109 (("(" 2960) (argument_list 2961) (ctor_arguments 29946)) 14110 ((T_EXTENDS 2963) (extends_from 29947)) 14111 ((T_IMPLEMENTS 2965) (implements_list 29948)) 14112 ((backup_doc_comment 29949)) 14113 (("{" 29950)) 14114 ((class_statement_list 29951)) 14115 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 29952)) 14116 (("%" 299) (")" 29954) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14117 ((";" 30432) ("{" 30433)) 14118 ((top_statement_list 30040)) 14119 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 30041) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 30042) (T_CLASS 30043) (T_CLASS_C 17) (T_CLONE 18) (T_CONST 30044) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30045) (T_DEC 22) (T_DECLARE 30046) (T_DIR 24) (T_DNUMBER 25) (T_DO 30047) (T_DOUBLE_CAST 27) (T_ECHO 30048) (T_EMPTY 29) (T_ENUM 30049) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 30050) (T_FOREACH 30051) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 30052) (T_GOTO 30053) (T_HALT_COMPILER 30054) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30056) (T_INTERFACE 30057) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAMESPACE 30058) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30059) (T_START_HEREDOC 69) (T_STATIC 30060) (T_STRING 30061) (T_STRING_CAST 72) (T_SWITCH 30062) (T_THROW 74) (T_TRAIT 30063) (T_TRAIT_C 76) (T_TRY 30064) (T_UNSET 30065) (T_UNSET_CAST 79) (T_USE 30066) (T_VARIABLE 81) (T_WHILE 30067) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30068) (alt_if_stmt_without_else 30069) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 30070) (attributes 30071) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 30072) (class_modifier 97) (class_modifiers 30073) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 30074) (expr 30075) (fn 104) (fully_dereferenceable 105) #'30076 (function_call 107) (function_declaration_statement 30077) (if_stmt 30078) (if_stmt_without_else 30079) (inline_function 111) (interface_declaration_statement 30080) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30081) (static_member 121) (top_statement 30082) (trait_declaration_statement 30083) (variable 124) (variable_class_name 125) ("{" 30084) ("}" 30085) ("~" 127)) 14120 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 30430) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 14121 ((T_STRING 30423)) 14122 ((T_STRING 30414) (const_decl 30415) (const_list 30416)) 14123 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 30412) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 14124 (("(" 30403)) 14125 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30397) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 14126 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 30395) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 14127 ((T_STRING 30388)) 14128 (("(" 30375)) 14129 (("(" 30360)) 14130 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 30358) (simple_variable 18987)) 14131 ((T_STRING 30356)) 14132 (("(" 30353)) 14133 (("(" 30349)) 14134 ((T_STRING 30343)) 14135 ((T_ABSTRACT 29955) (T_ARRAY 29956) (T_AS 29957) (T_BREAK 29958) (T_CALLABLE 29959) (T_CASE 29960) (T_CATCH 29961) (T_CLASS 29962) (T_CLASS_C 29963) (T_CLONE 29964) (T_CONST 29965) (T_CONTINUE 29966) (T_DECLARE 29967) (T_DEFAULT 29968) (T_DIR 29969) (T_DO 29970) (T_ECHO 29971) (T_ELSE 29972) (T_ELSEIF 29973) (T_EMPTY 29974) (T_ENDDECLARE 29975) (T_ENDFOR 29976) (T_ENDFOREACH 29977) (T_ENDIF 29978) (T_ENDSWITCH 29979) (T_ENDWHILE 29980) (T_ENUM 29981) (T_EVAL 29982) (T_EXIT 29983) (T_EXTENDS 29984) (T_FILE 29985) (T_FINAL 29986) (T_FINALLY 29987) (T_FN 29988) (T_FOR 29989) (T_FOREACH 29990) (T_FUNCTION 29991) (T_FUNC_C 29992) (T_GLOBAL 29993) (T_GOTO 29994) (T_IF 29995) (T_IMPLEMENTS 29996) (T_INCLUDE 29997) (T_INCLUDE_ONCE 29998) (T_INSTANCEOF 29999) (T_INSTEADOF 30000) (T_INTERFACE 30001) (T_ISSET 30002) (T_LINE 30003) (T_LIST 30004) (T_LOGICAL_AND 30005) (T_LOGICAL_OR 30006) (T_LOGICAL_XOR 30007) (T_MATCH 30008) (T_METHOD_C 30009) (T_NAMESPACE 30010) (T_NAME_QUALIFIED 30011) (T_NEW 30012) (T_NS_C 30013) (T_PRINT 30014) (T_PRIVATE 30015) (T_PROPERTY_C 30016) (T_PROTECTED 30017) (T_PUBLIC 30018) (T_READONLY 30019) (T_REQUIRE 30020) (T_REQUIRE_ONCE 30021) (T_RETURN 30022) (T_STATIC 30023) (T_STRING 30024) (T_SWITCH 30025) (T_THROW 30026) (T_TRAIT 30027) (T_TRAIT_C 30028) (T_TRY 30029) (T_UNSET 30030) (T_USE 30031) (T_VAR 30032) (T_WHILE 30033) (T_YIELD 30034) (identifier 30035) (namespace_declaration_name 30335) (reserved_non_modifiers 30037) (semi_reserved 30038) ("{" 30336)) 14136 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 30333) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 14137 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 30331)) 14138 ((":" 30330)) 14139 (("(" 30312)) 14140 ((T_STRING 30307)) 14141 (("{" 30290)) 14142 (("(" 30285)) 14143 ((T_CONST 29803) (T_FUNCTION 29804) (T_NAME_FULLY_QUALIFIED 29805) (T_NAME_QUALIFIED 29806) (T_STRING 29807) (legacy_namespace_name 29808) (mixed_group_use_declaration 30276) (namespace_name 29810) (use_declaration 29811) (use_declarations 30277) (use_type 30278)) 14144 (("(" 30267)) 14145 ((T_ELSE 30260) (T_ELSEIF 17291) (T_ENDIF 30261)) 14146 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 30043) (T_ENUM 30049) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 30057) (T_READONLY 2931) (T_STATIC 17376) (T_TRAIT 30063) (attribute 17377) (attributed_statement 30259) (class_declaration_statement 30072) (class_modifier 97) (class_modifiers 30073) (enum_declaration_statement 30074) (fn 104) #'30076 (function_declaration_statement 30077) (inline_function 17379) (interface_declaration_statement 30080) (trait_declaration_statement 30083)) 14147 ((T_ABSTRACT 10) (T_CLASS 30251) (T_FINAL 34) (T_READONLY 2931) (class_modifier 17368)) 14148 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 30250) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 14149 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17349) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17350) (ampersand 17351) (returns_ref 30238)) 14150 ((T_ELSE 30088) (T_ELSEIF 30089)) 14151 ((inner_statement_list 30086)) 14152 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 30087) ("~" 127)) 14153 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 30041) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30042) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30045) (T_DEC 22) (T_DECLARE 30046) (T_DIR 24) (T_DNUMBER 25) (T_DO 30047) (T_DOUBLE_CAST 27) (T_ECHO 30048) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30050) (T_FOREACH 30051) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30052) (T_GOTO 30053) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30056) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30059) (T_START_HEREDOC 69) (T_STATIC 30060) (T_STRING 30061) (T_STRING_CAST 72) (T_SWITCH 30062) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30064) (T_UNSET 30065) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30067) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30068) (alt_if_stmt_without_else 30069) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30075) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30078) (if_stmt_without_else 30079) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30237) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30084) ("~" 127)) 14154 (("(" 30090)) 14155 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30091) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14156 (("%" 299) (")" 30092) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14157 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 30093) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30094) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30095) (T_DEC 22) (T_DECLARE 30096) (T_DIR 24) (T_DNUMBER 25) (T_DO 30097) (T_DOUBLE_CAST 27) (T_ECHO 30098) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30099) (T_FOREACH 30100) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30101) (T_GOTO 30102) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30103) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30104) (T_START_HEREDOC 69) (T_STATIC 30105) (T_STRING 30106) (T_STRING_CAST 72) (T_SWITCH 30107) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30108) (T_UNSET 30109) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30110) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30111) (alt_if_stmt_without_else 30112) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30113) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30114) (if_stmt_without_else 30115) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30116) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30117) ("~" 127)) 14158 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 30235) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 14159 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 30233) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 14160 (("(" 30224)) 14161 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30218) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 14162 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 30216) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 14163 (("(" 30203)) 14164 (("(" 30188)) 14165 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 30186) (simple_variable 18987)) 14166 ((T_STRING 30184)) 14167 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 30182) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 14168 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 30180)) 14169 ((":" 30179)) 14170 (("(" 30161)) 14171 (("{" 30144)) 14172 (("(" 30139)) 14173 (("(" 30130)) 14174 ((T_ELSE 30123) (T_ELSEIF 17291) (T_ENDIF 30124)) 14175 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 30122) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 14176 ((T_ELSE 30120) (T_ELSEIF 30089)) 14177 ((inner_statement_list 30118)) 14178 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 30119) ("~" 127)) 14179 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 30093) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30094) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30095) (T_DEC 22) (T_DECLARE 30096) (T_DIR 24) (T_DNUMBER 25) (T_DO 30097) (T_DOUBLE_CAST 27) (T_ECHO 30098) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30099) (T_FOREACH 30100) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30101) (T_GOTO 30102) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30103) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30104) (T_START_HEREDOC 69) (T_STATIC 30105) (T_STRING 30106) (T_STRING_CAST 72) (T_SWITCH 30107) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30108) (T_UNSET 30109) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30110) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30111) (alt_if_stmt_without_else 30112) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30113) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30114) (if_stmt_without_else 30115) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30121) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30117) ("~" 127)) 14180 ((":" 30126)) 14181 ((";" 30125)) 14182 ((inner_statement_list 30127)) 14183 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 30128) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 14184 ((";" 30129)) 14185 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30131) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14186 (("%" 299) (")" 30132) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14187 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30133) (";" 30093) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30094) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30095) (T_DEC 22) (T_DECLARE 30096) (T_DIR 24) (T_DNUMBER 25) (T_DO 30097) (T_DOUBLE_CAST 27) (T_ECHO 30098) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30099) (T_FOREACH 30100) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30101) (T_GOTO 30102) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30103) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30104) (T_START_HEREDOC 69) (T_STATIC 30105) (T_STRING 30106) (T_STRING_CAST 72) (T_SWITCH 30107) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30108) (T_UNSET 30109) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30110) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30111) (alt_if_stmt_without_else 30112) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30113) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30114) (if_stmt_without_else 30115) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30134) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 30135) ("{" 30117) ("~" 127)) 14188 ((inner_statement_list 30136)) 14189 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 30137) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 14190 ((";" 30138)) 14191 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 30140) (variable 17576) (variable_class_name 17577)) 14192 (("," 17589) (possible_comma 30141)) 14193 ((")" 30142)) 14194 ((";" 30143)) 14195 ((inner_statement_list 30145)) 14196 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 30146) ("~" 127)) 14197 ((catch_list 30147)) 14198 ((T_CATCH 30148) (T_FINALLY 30149) (finally_statement 30150)) 14199 (("(" 30154)) 14200 (("{" 30151)) 14201 ((inner_statement_list 30152)) 14202 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 30153) ("~" 127)) 14203 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 30155) (class_name 17636) (name 17637)) 14204 ((T_VARIABLE 17638) (optional_variable 30156) ("|" 17640)) 14205 ((")" 30157)) 14206 (("{" 30158)) 14207 ((inner_statement_list 30159)) 14208 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 30160) ("~" 127)) 14209 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30162) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14210 (("%" 299) (")" 30163) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14211 ((":" 30164) (switch_case_list 30165) ("{" 30166)) 14212 ((";" 30172) (case_list 30173)) 14213 ((";" 30167) (case_list 30168)) 14214 ((case_list 30170)) 14215 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 30169)) 14216 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 30171)) 14217 ((case_list 30176)) 14218 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 30174)) 14219 ((";" 30175)) 14220 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 30177)) 14221 ((";" 30178)) 14222 (("," 17937) (";" 30181)) 14223 ((";" 30183)) 14224 ((";" 30185)) 14225 (("," 18988) (";" 30187)) 14226 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 30189) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 14227 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 30190) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 14228 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 30191) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 14229 ((")" 30192) (T_DOUBLE_ARROW 30193)) 14230 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30196) (";" 30093) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30094) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30095) (T_DEC 22) (T_DECLARE 30096) (T_DIR 24) (T_DNUMBER 25) (T_DO 30097) (T_DOUBLE_CAST 27) (T_ECHO 30098) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30099) (T_FOREACH 30100) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30101) (T_GOTO 30102) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30103) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30104) (T_START_HEREDOC 69) (T_STATIC 30105) (T_STRING 30106) (T_STRING_CAST 72) (T_SWITCH 30107) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30108) (T_UNSET 30109) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30110) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30111) (alt_if_stmt_without_else 30112) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30113) (fn 104) (foreach_statement 30202) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30114) (if_stmt_without_else 30115) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30198) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30117) ("~" 127)) 14231 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 30194) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 14232 ((")" 30195)) 14233 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30196) (";" 30093) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30094) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30095) (T_DEC 22) (T_DECLARE 30096) (T_DIR 24) (T_DNUMBER 25) (T_DO 30097) (T_DOUBLE_CAST 27) (T_ECHO 30098) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30099) (T_FOREACH 30100) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30101) (T_GOTO 30102) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30103) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30104) (T_START_HEREDOC 69) (T_STATIC 30105) (T_STRING 30106) (T_STRING_CAST 72) (T_SWITCH 30107) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30108) (T_UNSET 30109) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30110) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30111) (alt_if_stmt_without_else 30112) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30113) (fn 104) (foreach_statement 30197) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30114) (if_stmt_without_else 30115) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30198) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30117) ("~" 127)) 14234 ((inner_statement_list 30199)) 14235 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 30200) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 14236 ((";" 30201)) 14237 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 30204) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 14238 ((";" 30205)) 14239 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 30206) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 14240 ((";" 30207)) 14241 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 30208) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 14242 ((")" 30209)) 14243 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30210) (";" 30093) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30094) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30095) (T_DEC 22) (T_DECLARE 30096) (T_DIR 24) (T_DNUMBER 25) (T_DO 30097) (T_DOUBLE_CAST 27) (T_ECHO 30098) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30099) (T_FOREACH 30100) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30101) (T_GOTO 30102) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30103) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30104) (T_START_HEREDOC 69) (T_STATIC 30105) (T_STRING 30106) (T_STRING_CAST 72) (T_SWITCH 30107) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30108) (T_UNSET 30109) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30110) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30111) (alt_if_stmt_without_else 30112) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30113) (fn 104) (for_statement 30211) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30114) (if_stmt_without_else 30115) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30212) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30117) ("~" 127)) 14244 ((inner_statement_list 30213)) 14245 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 30214) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 14246 ((";" 30215)) 14247 (("," 20829) (";" 30217)) 14248 ((T_WHILE 30219)) 14249 (("(" 30220)) 14250 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30221) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14251 (("%" 299) (")" 30222) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14252 ((";" 30223)) 14253 ((T_STRING 21002) (const_decl 21003) (const_list 30225)) 14254 ((")" 30226) ("," 21006)) 14255 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30227) (";" 30093) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30094) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30095) (T_DEC 22) (T_DECLARE 30096) (T_DIR 24) (T_DNUMBER 25) (T_DO 30097) (T_DOUBLE_CAST 27) (T_ECHO 30098) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30099) (T_FOREACH 30100) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30101) (T_GOTO 30102) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30103) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30104) (T_START_HEREDOC 69) (T_STATIC 30105) (T_STRING 30106) (T_STRING_CAST 72) (T_SWITCH 30107) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30108) (T_UNSET 30109) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30110) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30111) (alt_if_stmt_without_else 30112) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 30228) (dereferenceable_scalar 101) (expr 30113) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30114) (if_stmt_without_else 30115) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30229) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30117) ("~" 127)) 14256 ((inner_statement_list 30230)) 14257 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 30231) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 14258 ((";" 30232)) 14259 ((";" 30234)) 14260 ((";" 30236)) 14261 ((T_READONLY 17353) (T_STRING 17354) (backup_doc_comment 141) (function_name 30239)) 14262 ((backup_doc_comment 30240)) 14263 (("(" 30241)) 14264 ((T_ABSTRACT 143) (T_ATTRIBUTE 144) (T_FINAL 145) (T_PRIVATE 146) (T_PRIVATE_SET 147) (T_PROTECTED 148) (T_PROTECTED_SET 149) (T_PUBLIC 150) (T_PUBLIC_SET 151) (T_READONLY 152) (T_STATIC 153) (attribute 154) (attributed_parameter 155) (attributes 156) (member_modifier 157) (non_empty_member_modifiers 158) (non_empty_parameter_list 159) (optional_cpp_modifiers 160) (parameter 161) (parameter_list 30242)) 14265 ((")" 30243)) 14266 ((":" 166) (return_type 30244)) 14267 ((backup_fn_flags 30245)) 14268 (("{" 30246)) 14269 ((inner_statement_list 30247)) 14270 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 30248) ("~" 127)) 14271 ((backup_fn_flags 30249)) 14272 ((T_STRING 30252)) 14273 ((T_EXTENDS 2963) (extends_from 30253)) 14274 ((T_IMPLEMENTS 2965) (implements_list 30254)) 14275 ((backup_doc_comment 30255)) 14276 (("{" 30256)) 14277 ((class_statement_list 30257)) 14278 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 30258)) 14279 ((":" 30263)) 14280 ((";" 30262)) 14281 ((inner_statement_list 30264)) 14282 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 30265) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 14283 ((";" 30266)) 14284 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30268) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14285 (("%" 299) (")" 30269) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14286 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30270) (";" 30041) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30042) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30045) (T_DEC 22) (T_DECLARE 30046) (T_DIR 24) (T_DNUMBER 25) (T_DO 30047) (T_DOUBLE_CAST 27) (T_ECHO 30048) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30050) (T_FOREACH 30051) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30052) (T_GOTO 30053) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30056) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30059) (T_START_HEREDOC 69) (T_STATIC 30060) (T_STRING 30061) (T_STRING_CAST 72) (T_SWITCH 30062) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30064) (T_UNSET 30065) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30067) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30068) (alt_if_stmt_without_else 30069) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30075) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30078) (if_stmt_without_else 30079) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30271) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 30272) ("{" 30084) ("~" 127)) 14287 ((inner_statement_list 30273)) 14288 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 30274) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 14289 ((";" 30275)) 14290 ((";" 30284)) 14291 (("," 29817) (";" 30283)) 14292 ((T_NAME_FULLY_QUALIFIED 29805) (T_NAME_QUALIFIED 29806) (T_STRING 29807) (group_use_declaration 30279) (legacy_namespace_name 29815) (namespace_name 29810) (use_declaration 29811) (use_declarations 30280)) 14293 ((";" 30282)) 14294 (("," 29817) (";" 30281)) 14295 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 30286) (variable 17576) (variable_class_name 17577)) 14296 (("," 17589) (possible_comma 30287)) 14297 ((")" 30288)) 14298 ((";" 30289)) 14299 ((inner_statement_list 30291)) 14300 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 30292) ("~" 127)) 14301 ((catch_list 30293)) 14302 ((T_CATCH 30294) (T_FINALLY 30295) (finally_statement 30296)) 14303 (("(" 30300)) 14304 (("{" 30297)) 14305 ((inner_statement_list 30298)) 14306 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 30299) ("~" 127)) 14307 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 30301) (class_name 17636) (name 17637)) 14308 ((T_VARIABLE 17638) (optional_variable 30302) ("|" 17640)) 14309 ((")" 30303)) 14310 (("{" 30304)) 14311 ((inner_statement_list 30305)) 14312 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 30306) ("~" 127)) 14313 ((backup_doc_comment 30308)) 14314 (("{" 30309)) 14315 ((class_statement_list 30310)) 14316 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 30311)) 14317 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30313) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14318 (("%" 299) (")" 30314) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14319 ((":" 30315) (switch_case_list 30316) ("{" 30317)) 14320 ((";" 30323) (case_list 30324)) 14321 ((";" 30318) (case_list 30319)) 14322 ((case_list 30321)) 14323 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 30320)) 14324 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 30322)) 14325 ((case_list 30327)) 14326 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 30325)) 14327 ((";" 30326)) 14328 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 30328)) 14329 ((";" 30329)) 14330 (("," 17937) (";" 30332)) 14331 ((";" 30334)) 14332 ((";" 30339) ("{" 30340)) 14333 ((top_statement_list 30337)) 14334 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 30041) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 30042) (T_CLASS 30043) (T_CLASS_C 17) (T_CLONE 18) (T_CONST 30044) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30045) (T_DEC 22) (T_DECLARE 30046) (T_DIR 24) (T_DNUMBER 25) (T_DO 30047) (T_DOUBLE_CAST 27) (T_ECHO 30048) (T_EMPTY 29) (T_ENUM 30049) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 30050) (T_FOREACH 30051) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 30052) (T_GOTO 30053) (T_HALT_COMPILER 30054) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30056) (T_INTERFACE 30057) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAMESPACE 30058) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30059) (T_START_HEREDOC 69) (T_STATIC 30060) (T_STRING 30061) (T_STRING_CAST 72) (T_SWITCH 30062) (T_THROW 74) (T_TRAIT 30063) (T_TRAIT_C 76) (T_TRY 30064) (T_UNSET 30065) (T_UNSET_CAST 79) (T_USE 30066) (T_VARIABLE 81) (T_WHILE 30067) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30068) (alt_if_stmt_without_else 30069) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 30070) (attributes 30071) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 30072) (class_modifier 97) (class_modifiers 30073) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 30074) (expr 30075) (fn 104) (fully_dereferenceable 105) #'30076 (function_call 107) (function_declaration_statement 30077) (if_stmt 30078) (if_stmt_without_else 30079) (inline_function 111) (interface_declaration_statement 30080) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30081) (static_member 121) (top_statement 30082) (trait_declaration_statement 30083) (variable 124) (variable_class_name 125) ("{" 30084) ("}" 30338) ("~" 127)) 14335 ((top_statement_list 30341)) 14336 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 30041) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 30042) (T_CLASS 30043) (T_CLASS_C 17) (T_CLONE 18) (T_CONST 30044) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30045) (T_DEC 22) (T_DECLARE 30046) (T_DIR 24) (T_DNUMBER 25) (T_DO 30047) (T_DOUBLE_CAST 27) (T_ECHO 30048) (T_EMPTY 29) (T_ENUM 30049) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 30050) (T_FOREACH 30051) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 30052) (T_GOTO 30053) (T_HALT_COMPILER 30054) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30056) (T_INTERFACE 30057) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAMESPACE 30058) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30059) (T_START_HEREDOC 69) (T_STATIC 30060) (T_STRING 30061) (T_STRING_CAST 72) (T_SWITCH 30062) (T_THROW 74) (T_TRAIT 30063) (T_TRAIT_C 76) (T_TRY 30064) (T_UNSET 30065) (T_UNSET_CAST 79) (T_USE 30066) (T_VARIABLE 81) (T_WHILE 30067) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30068) (alt_if_stmt_without_else 30069) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 30070) (attributes 30071) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 30072) (class_modifier 97) (class_modifiers 30073) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 30074) (expr 30075) (fn 104) (fully_dereferenceable 105) #'30076 (function_call 107) (function_declaration_statement 30077) (if_stmt 30078) (if_stmt_without_else 30079) (inline_function 111) (interface_declaration_statement 30080) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30081) (static_member 121) (top_statement 30082) (trait_declaration_statement 30083) (variable 124) (variable_class_name 125) ("{" 30084) ("}" 30342) ("~" 127)) 14337 ((T_EXTENDS 21295) (interface_extends_list 30344)) 14338 ((backup_doc_comment 30345)) 14339 (("{" 30346)) 14340 ((class_statement_list 30347)) 14341 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 30348)) 14342 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30350) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14343 (("%" 299) (")" 30351) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14344 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21305) (";" 30093) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30094) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30095) (T_DEC 22) (T_DECLARE 30096) (T_DIR 24) (T_DNUMBER 25) (T_DO 30097) (T_DOUBLE_CAST 27) (T_ECHO 30098) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30099) (T_FOREACH 30100) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30101) (T_GOTO 30102) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30103) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30104) (T_START_HEREDOC 69) (T_STATIC 30105) (T_STRING 30106) (T_STRING_CAST 72) (T_SWITCH 30107) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30108) (T_UNSET 30109) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30110) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30111) (alt_if_stmt_without_else 30112) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30113) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30114) (if_stmt_without_else 30115) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30352) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30117) ("~" 127)) 14345 ((")" 30354)) 14346 ((";" 30355)) 14347 ((";" 30357)) 14348 (("," 18988) (";" 30359)) 14349 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 30361) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 14350 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 30362) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 14351 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 30363) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 14352 ((")" 30364) (T_DOUBLE_ARROW 30365)) 14353 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30368) (";" 30041) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30042) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30045) (T_DEC 22) (T_DECLARE 30046) (T_DIR 24) (T_DNUMBER 25) (T_DO 30047) (T_DOUBLE_CAST 27) (T_ECHO 30048) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30050) (T_FOREACH 30051) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30052) (T_GOTO 30053) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30056) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30059) (T_START_HEREDOC 69) (T_STATIC 30060) (T_STRING 30061) (T_STRING_CAST 72) (T_SWITCH 30062) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30064) (T_UNSET 30065) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30067) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30068) (alt_if_stmt_without_else 30069) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30075) (fn 104) (foreach_statement 30374) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30078) (if_stmt_without_else 30079) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30370) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30084) ("~" 127)) 14354 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 30366) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 14355 ((")" 30367)) 14356 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30368) (";" 30041) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30042) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30045) (T_DEC 22) (T_DECLARE 30046) (T_DIR 24) (T_DNUMBER 25) (T_DO 30047) (T_DOUBLE_CAST 27) (T_ECHO 30048) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30050) (T_FOREACH 30051) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30052) (T_GOTO 30053) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30056) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30059) (T_START_HEREDOC 69) (T_STATIC 30060) (T_STRING 30061) (T_STRING_CAST 72) (T_SWITCH 30062) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30064) (T_UNSET 30065) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30067) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30068) (alt_if_stmt_without_else 30069) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30075) (fn 104) (foreach_statement 30369) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30078) (if_stmt_without_else 30079) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30370) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30084) ("~" 127)) 14357 ((inner_statement_list 30371)) 14358 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 30372) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 14359 ((";" 30373)) 14360 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 30376) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 14361 ((";" 30377)) 14362 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 30378) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 14363 ((";" 30379)) 14364 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 30380) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 14365 ((")" 30381)) 14366 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30382) (";" 30041) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30042) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30045) (T_DEC 22) (T_DECLARE 30046) (T_DIR 24) (T_DNUMBER 25) (T_DO 30047) (T_DOUBLE_CAST 27) (T_ECHO 30048) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30050) (T_FOREACH 30051) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30052) (T_GOTO 30053) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30056) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30059) (T_START_HEREDOC 69) (T_STATIC 30060) (T_STRING 30061) (T_STRING_CAST 72) (T_SWITCH 30062) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30064) (T_UNSET 30065) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30067) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30068) (alt_if_stmt_without_else 30069) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30075) (fn 104) (for_statement 30383) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30078) (if_stmt_without_else 30079) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30384) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30084) ("~" 127)) 14367 ((inner_statement_list 30385)) 14368 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 30386) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 14369 ((";" 30387)) 14370 ((":" 21344) (enum_backing_type 30389)) 14371 ((T_IMPLEMENTS 2965) (implements_list 30390)) 14372 ((backup_doc_comment 30391)) 14373 (("{" 30392)) 14374 ((class_statement_list 30393)) 14375 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 30394)) 14376 (("," 20829) (";" 30396)) 14377 ((T_WHILE 30398)) 14378 (("(" 30399)) 14379 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30400) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14380 (("%" 299) (")" 30401) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14381 ((";" 30402)) 14382 ((T_STRING 21002) (const_decl 21003) (const_list 30404)) 14383 ((")" 30405) ("," 21006)) 14384 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30406) (";" 30041) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30042) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30045) (T_DEC 22) (T_DECLARE 30046) (T_DIR 24) (T_DNUMBER 25) (T_DO 30047) (T_DOUBLE_CAST 27) (T_ECHO 30048) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30050) (T_FOREACH 30051) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30052) (T_GOTO 30053) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30056) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30059) (T_START_HEREDOC 69) (T_STATIC 30060) (T_STRING 30061) (T_STRING_CAST 72) (T_SWITCH 30062) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30064) (T_UNSET 30065) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30067) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30068) (alt_if_stmt_without_else 30069) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 30407) (dereferenceable_scalar 101) (expr 30075) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30078) (if_stmt_without_else 30079) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30408) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30084) ("~" 127)) 14385 ((inner_statement_list 30409)) 14386 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 30410) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 14387 ((";" 30411)) 14388 ((";" 30413)) 14389 (("=" 30420)) 14390 (("," 30417) (";" 30418)) 14391 ((T_STRING 30414) (const_decl 30419)) 14392 (("!" 3049) ("\"" 3050) ("$" 3051) ("(" 3052) ("+" 3053) ("-" 3054) ("@" 3055) (T_ARRAY 3056) (T_ARRAY_CAST 3057) (T_ATTRIBUTE 128) (T_BOOL_CAST 3058) (T_CLASS_C 3059) (T_CLONE 3060) (T_CONSTANT_ENCAPSED_STRING 3061) (T_DEC 3062) (T_DIR 3063) (T_DNUMBER 3064) (T_DOUBLE_CAST 3065) (T_EMPTY 3066) (T_EVAL 3067) (T_EXIT 3068) (T_FILE 3069) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3070) (T_INC 3071) (T_INCLUDE 3072) (T_INCLUDE_ONCE 3073) (T_INT_CAST 3074) (T_ISSET 3075) (T_LINE 3076) (T_LIST 3077) (T_LNUMBER 3078) (T_MATCH 3079) (T_METHOD_C 3080) (T_NAME_FULLY_QUALIFIED 3081) (T_NAME_QUALIFIED 3082) (T_NAME_RELATIVE 3083) (T_NEW 3084) (T_NS_C 3085) (T_OBJECT_CAST 3086) (T_PRINT 3087) (T_PROPERTY_C 3088) (T_READONLY 3089) (T_REQUIRE 3090) (T_REQUIRE_ONCE 3091) (T_START_HEREDOC 3092) (T_STATIC 3093) (T_STRING 3094) (T_STRING_CAST 3095) (T_THROW 3096) (T_TRAIT_C 3097) (T_UNSET_CAST 3098) (T_VARIABLE 3099) (T_YIELD 3100) (T_YIELD_FROM 3101) ("[" 3102) ("`" 3103) (array_object_dereferenceable 3104) (attribute 133) (attributes 3105) (callable_expr 3106) (callable_variable 3107) (class_constant 3108) (class_name 3109) (constant 3110) (dereferenceable_scalar 3111) (expr 30421) (fn 3113) (fully_dereferenceable 105) #'3114 (function_call 3115) (inline_function 3116) (internal_functions_in_yacc 3117) (match 3118) (name 3119) (new_dereferenceable 3120) (new_non_dereferenceable 3121) (scalar 3122) (simple_variable 3123) (static_member 3124) (variable 3125) (variable_class_name 3126) ("~" 3127)) 14393 (("%" 3129) ("*" 3130) ("+" 3131) ("-" 3132) ("." 3133) ("/" 3134) ("<" 3135) (">" 3136) ("?" 3137) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3138) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3139) (T_BOOLEAN_AND 3140) (T_BOOLEAN_OR 3141) (T_COALESCE 3142) (T_INSTANCEOF 3143) (T_IS_EQUAL 3144) (T_IS_GREATER_OR_EQUAL 3145) (T_IS_IDENTICAL 3146) (T_IS_NOT_EQUAL 3147) (T_IS_NOT_IDENTICAL 3148) (T_IS_SMALLER_OR_EQUAL 3149) (T_LOGICAL_AND 3150) (T_LOGICAL_OR 3151) (T_LOGICAL_XOR 3152) (T_POW 3153) (T_SL 3154) (T_SPACESHIP 3155) (T_SR 3156) ("^" 3157) (backup_doc_comment 30422) ("|" 3158)) 14394 ((T_EXTENDS 2963) (extends_from 30424)) 14395 ((T_IMPLEMENTS 2965) (implements_list 30425)) 14396 ((backup_doc_comment 30426)) 14397 (("{" 30427)) 14398 ((class_statement_list 30428)) 14399 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 30429)) 14400 ((";" 30431)) 14401 ((top_statement_list 30434)) 14402 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 30041) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 30042) (T_CLASS 30043) (T_CLASS_C 17) (T_CLONE 18) (T_CONST 30044) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30045) (T_DEC 22) (T_DECLARE 30046) (T_DIR 24) (T_DNUMBER 25) (T_DO 30047) (T_DOUBLE_CAST 27) (T_ECHO 30048) (T_EMPTY 29) (T_ENUM 30049) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 30050) (T_FOREACH 30051) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 30052) (T_GOTO 30053) (T_HALT_COMPILER 30054) (T_IF 30055) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30056) (T_INTERFACE 30057) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAMESPACE 30058) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30059) (T_START_HEREDOC 69) (T_STATIC 30060) (T_STRING 30061) (T_STRING_CAST 72) (T_SWITCH 30062) (T_THROW 74) (T_TRAIT 30063) (T_TRAIT_C 76) (T_TRY 30064) (T_UNSET 30065) (T_UNSET_CAST 79) (T_USE 30066) (T_VARIABLE 81) (T_WHILE 30067) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30068) (alt_if_stmt_without_else 30069) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 30070) (attributes 30071) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 30072) (class_modifier 97) (class_modifiers 30073) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 30074) (expr 30075) (fn 104) (fully_dereferenceable 105) #'30076 (function_call 107) (function_declaration_statement 30077) (if_stmt 30078) (if_stmt_without_else 30079) (inline_function 111) (interface_declaration_statement 30080) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30081) (static_member 121) (top_statement 30082) (trait_declaration_statement 30083) (variable 124) (variable_class_name 125) ("{" 30084) ("}" 30435) ("~" 127)) 14403 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30437) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14404 (("%" 299) (")" 30438) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14405 (("{" 30439)) 14406 (("!" 3932) ("\"" 3933) ("$" 3934) ("(" 3935) ("+" 3936) ("-" 3937) ("@" 3938) (T_ARRAY 3939) (T_ARRAY_CAST 3940) (T_ATTRIBUTE 128) (T_BOOL_CAST 3941) (T_CLASS_C 3942) (T_CLONE 3943) (T_CONSTANT_ENCAPSED_STRING 3944) (T_DEC 3945) (T_DEFAULT 3946) (T_DIR 3947) (T_DNUMBER 3948) (T_DOUBLE_CAST 3949) (T_EMPTY 3950) (T_EVAL 3951) (T_EXIT 3952) (T_FILE 3953) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 3954) (T_INC 3955) (T_INCLUDE 3956) (T_INCLUDE_ONCE 3957) (T_INT_CAST 3958) (T_ISSET 3959) (T_LINE 3960) (T_LIST 3961) (T_LNUMBER 3962) (T_MATCH 3963) (T_METHOD_C 3964) (T_NAME_FULLY_QUALIFIED 3965) (T_NAME_QUALIFIED 3966) (T_NAME_RELATIVE 3967) (T_NEW 3968) (T_NS_C 3969) (T_OBJECT_CAST 3970) (T_PRINT 3971) (T_PROPERTY_C 3972) (T_READONLY 3973) (T_REQUIRE 3974) (T_REQUIRE_ONCE 3975) (T_START_HEREDOC 3976) (T_STATIC 3977) (T_STRING 3978) (T_STRING_CAST 3979) (T_THROW 3980) (T_TRAIT_C 3981) (T_UNSET_CAST 3982) (T_VARIABLE 3983) (T_YIELD 3984) (T_YIELD_FROM 3985) ("[" 3986) ("`" 3987) (array_object_dereferenceable 3988) (attribute 133) (attributes 3989) (callable_expr 3990) (callable_variable 3991) (class_constant 3992) (class_name 3993) (constant 3994) (dereferenceable_scalar 3995) (expr 3996) (fn 3997) (fully_dereferenceable 105) #'3998 (function_call 3999) (inline_function 4000) (internal_functions_in_yacc 4001) (match 4002) (match_arm 4003) (match_arm_cond_list 4004) (match_arm_list 30440) (name 4006) (new_dereferenceable 4007) (new_non_dereferenceable 4008) (non_empty_match_arm_list 4009) (scalar 4010) (simple_variable 4011) (static_member 4012) (variable 4013) (variable_class_name 4014) ("~" 4015)) 14407 (("}" 30441)) 14408 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 30443) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 14409 ((")" 30444)) 14410 (("=" 30445)) 14411 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30446) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 14412 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 5189) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (isset_variable 5190) (isset_variables 30448) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 14413 (("," 5192) (possible_comma 30449)) 14414 ((")" 30450)) 14415 ((T_EXTENDS 21295) (interface_extends_list 30453)) 14416 ((backup_doc_comment 30454)) 14417 (("{" 30455)) 14418 ((class_statement_list 30456)) 14419 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 30457)) 14420 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30462) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14421 (("%" 299) (")" 30463) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14422 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21305) (";" 30464) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30465) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30466) (T_DEC 22) (T_DECLARE 30467) (T_DIR 24) (T_DNUMBER 25) (T_DO 30468) (T_DOUBLE_CAST 27) (T_ECHO 30469) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30470) (T_FOREACH 30471) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30472) (T_GOTO 30473) (T_IF 30474) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30475) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30476) (T_START_HEREDOC 69) (T_STATIC 30477) (T_STRING 30478) (T_STRING_CAST 72) (T_SWITCH 30479) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30480) (T_UNSET 30481) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30482) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30483) (alt_if_stmt_without_else 30484) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30485) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30486) (if_stmt_without_else 30487) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30488) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30489) ("~" 127)) 14423 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 30616) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 14424 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 30614) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 14425 (("(" 30605)) 14426 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20832) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 20833) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20834) (T_DEC 22) (T_DECLARE 20835) (T_DIR 24) (T_DNUMBER 25) (T_DO 20836) (T_DOUBLE_CAST 27) (T_ECHO 20837) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20838) (T_FOREACH 20839) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 20840) (T_GOTO 20841) (T_IF 20842) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20843) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20844) (T_START_HEREDOC 69) (T_STATIC 20845) (T_STRING 20846) (T_STRING_CAST 72) (T_SWITCH 20847) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20848) (T_UNSET 20849) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20850) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20851) (alt_if_stmt_without_else 20852) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 20853) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 20854) (if_stmt_without_else 20855) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30599) (static_member 121) (variable 124) (variable_class_name 125) ("{" 20857) ("~" 127)) 14427 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (echo_expr 20826) (echo_expr_list 30597) (expr 20828) (fn 18005) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 14428 (("(" 30584)) 14429 (("(" 30569)) 14430 (("$" 18983) (T_VARIABLE 18984) (global_var 18985) (global_var_list 30567) (simple_variable 18987)) 14431 ((T_STRING 30565)) 14432 (("(" 30561)) 14433 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 131) (T_STRING 132) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 18978) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (optional_expr 30559) (scalar 118) (simple_variable 119) (static_member 121) (variable 124) (variable_class_name 125) ("~" 127)) 14434 ((T_FN 35) (T_FUNCTION 129) (T_VARIABLE 17933) (fn 104) #'136 (inline_function 17934) (static_var 17935) (static_var_list 30557)) 14435 ((":" 30556)) 14436 (("(" 30538)) 14437 (("{" 30521)) 14438 (("(" 30516)) 14439 (("(" 30507)) 14440 ((T_ELSE 30500) (T_ELSEIF 17291) (T_ENDIF 30501)) 14441 (("%" 5726) ("*" 5727) ("+" 5728) ("-" 5729) ("." 5730) ("/" 5731) (";" 30499) ("<" 5733) (">" 5734) ("?" 5735) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5736) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5737) (T_BOOLEAN_AND 5738) (T_BOOLEAN_OR 5739) (T_COALESCE 5740) (T_INSTANCEOF 5741) (T_IS_EQUAL 5742) (T_IS_GREATER_OR_EQUAL 5743) (T_IS_IDENTICAL 5744) (T_IS_NOT_EQUAL 5745) (T_IS_NOT_IDENTICAL 5746) (T_IS_SMALLER_OR_EQUAL 5747) (T_LOGICAL_AND 5748) (T_LOGICAL_OR 5749) (T_LOGICAL_XOR 5750) (T_POW 5751) (T_SL 5752) (T_SPACESHIP 5753) (T_SR 5754) ("^" 5755) ("|" 5756)) 14442 ((T_ELSE 30492) (T_ELSEIF 30493)) 14443 ((inner_statement_list 30490)) 14444 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 30491) ("~" 127)) 14445 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 30464) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30465) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30466) (T_DEC 22) (T_DECLARE 30467) (T_DIR 24) (T_DNUMBER 25) (T_DO 30468) (T_DOUBLE_CAST 27) (T_ECHO 30469) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30470) (T_FOREACH 30471) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30472) (T_GOTO 30473) (T_IF 30474) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30475) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30476) (T_START_HEREDOC 69) (T_STATIC 30477) (T_STRING 30478) (T_STRING_CAST 72) (T_SWITCH 30479) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30480) (T_UNSET 30481) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30482) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30483) (alt_if_stmt_without_else 30484) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30485) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30486) (if_stmt_without_else 30487) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30498) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30489) ("~" 127)) 14446 (("(" 30494)) 14447 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30495) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14448 (("%" 299) (")" 30496) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14449 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 30464) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30465) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30466) (T_DEC 22) (T_DECLARE 30467) (T_DIR 24) (T_DNUMBER 25) (T_DO 30468) (T_DOUBLE_CAST 27) (T_ECHO 30469) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30470) (T_FOREACH 30471) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30472) (T_GOTO 30473) (T_IF 30474) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30475) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30476) (T_START_HEREDOC 69) (T_STATIC 30477) (T_STRING 30478) (T_STRING_CAST 72) (T_SWITCH 30479) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30480) (T_UNSET 30481) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30482) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30483) (alt_if_stmt_without_else 30484) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30485) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30486) (if_stmt_without_else 30487) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30497) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30489) ("~" 127)) 14450 ((":" 30503)) 14451 ((";" 30502)) 14452 ((inner_statement_list 30504)) 14453 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17386) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17387) (T_CLASS 17388) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17389) (T_DEC 22) (T_DECLARE 17390) (T_DIR 24) (T_DNUMBER 25) (T_DO 17391) (T_DOUBLE_CAST 27) (T_ECHO 17392) (T_EMPTY 29) (T_ENDIF 30505) (T_ENUM 17394) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17395) (T_FOREACH 17396) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17397) (T_GOTO 17398) (T_HALT_COMPILER 17399) (T_IF 17312) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17400) (T_INTERFACE 17401) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17402) (T_START_HEREDOC 69) (T_STATIC 17403) (T_STRING 17404) (T_STRING_CAST 72) (T_SWITCH 17405) (T_THROW 74) (T_TRAIT 17406) (T_TRAIT_C 76) (T_TRY 17407) (T_UNSET 17408) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17409) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17410) (alt_if_stmt_without_else 17411) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17412) (attributes 17413) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17414) (class_modifier 97) (class_modifiers 17415) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17416) (expr 17417) (fn 104) (fully_dereferenceable 105) #'17418 (function_call 107) (function_declaration_statement 17419) (if_stmt 17420) (if_stmt_without_else 17421) (inline_function 111) (inner_statement 17422) (interface_declaration_statement 17423) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17424) (static_member 121) (trait_declaration_statement 17425) (variable 124) (variable_class_name 125) ("{" 17426) ("~" 127)) 14454 ((";" 30506)) 14455 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30508) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14456 (("%" 299) (")" 30509) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14457 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30510) (";" 30464) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30465) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30466) (T_DEC 22) (T_DECLARE 30467) (T_DIR 24) (T_DNUMBER 25) (T_DO 30468) (T_DOUBLE_CAST 27) (T_ECHO 30469) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30470) (T_FOREACH 30471) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30472) (T_GOTO 30473) (T_IF 30474) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30475) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30476) (T_START_HEREDOC 69) (T_STATIC 30477) (T_STRING 30478) (T_STRING_CAST 72) (T_SWITCH 30479) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30480) (T_UNSET 30481) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30482) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30483) (alt_if_stmt_without_else 30484) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30485) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30486) (if_stmt_without_else 30487) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30511) (static_member 121) (variable 124) (variable_class_name 125) (while_statement 30512) ("{" 30489) ("~" 127)) 14458 ((inner_statement_list 30513)) 14459 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17467) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 17468) (T_CLASS 17469) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17470) (T_DEC 22) (T_DECLARE 17471) (T_DIR 24) (T_DNUMBER 25) (T_DO 17472) (T_DOUBLE_CAST 27) (T_ECHO 17473) (T_EMPTY 29) (T_ENDWHILE 30514) (T_ENUM 17475) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17476) (T_FOREACH 17477) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17478) (T_GOTO 17479) (T_HALT_COMPILER 17480) (T_IF 17481) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17482) (T_INTERFACE 17483) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17484) (T_START_HEREDOC 69) (T_STATIC 17485) (T_STRING 17486) (T_STRING_CAST 72) (T_SWITCH 17487) (T_THROW 74) (T_TRAIT 17488) (T_TRAIT_C 76) (T_TRY 17489) (T_UNSET 17490) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 17491) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 17492) (alt_if_stmt_without_else 17493) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 17494) (attributes 17495) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 17496) (class_modifier 97) (class_modifiers 17497) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 17498) (expr 17499) (fn 104) (fully_dereferenceable 105) #'17500 (function_call 107) (function_declaration_statement 17501) (if_stmt 17502) (if_stmt_without_else 17503) (inline_function 111) (inner_statement 17504) (interface_declaration_statement 17505) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 17506) (static_member 121) (trait_declaration_statement 17507) (variable 124) (variable_class_name 125) ("{" 17508) ("~" 127)) 14460 ((";" 30515)) 14461 (("\"" 1520) ("$" 17563) ("(" 1522) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 17564) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 17565) ("[" 1542) (array_object_dereferenceable 17566) (callable_expr 17567) (callable_variable 17568) (class_constant 1546) (class_name 17569) (constant 1548) (dereferenceable_scalar 1549) (fully_dereferenceable 105) (function_call 17570) (name 17571) (new_dereferenceable 1552) (simple_variable 17572) (static_member 17573) (unset_variable 17574) (unset_variables 30517) (variable 17576) (variable_class_name 17577)) 14462 (("," 17589) (possible_comma 30518)) 14463 ((")" 30519)) 14464 ((";" 30520)) 14465 ((inner_statement_list 30522)) 14466 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 30523) ("~" 127)) 14467 ((catch_list 30524)) 14468 ((T_CATCH 30525) (T_FINALLY 30526) (finally_statement 30527)) 14469 (("(" 30531)) 14470 (("{" 30528)) 14471 ((inner_statement_list 30529)) 14472 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 30530) ("~" 127)) 14473 ((T_NAME_FULLY_QUALIFIED 17630) (T_NAME_QUALIFIED 17631) (T_NAME_RELATIVE 17632) (T_STATIC 17633) (T_STRING 17634) (catch_name_list 30532) (class_name 17636) (name 17637)) 14474 ((T_VARIABLE 17638) (optional_variable 30533) ("|" 17640)) 14475 ((")" 30534)) 14476 (("{" 30535)) 14477 ((inner_statement_list 30536)) 14478 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 171) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 172) (T_CLASS 173) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 174) (T_DEC 22) (T_DECLARE 175) (T_DIR 24) (T_DNUMBER 25) (T_DO 176) (T_DOUBLE_CAST 27) (T_ECHO 177) (T_EMPTY 29) (T_ENUM 178) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 179) (T_FOREACH 180) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 181) (T_GOTO 182) (T_HALT_COMPILER 183) (T_IF 184) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 185) (T_INTERFACE 186) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 187) (T_START_HEREDOC 69) (T_STATIC 188) (T_STRING 189) (T_STRING_CAST 72) (T_SWITCH 190) (T_THROW 74) (T_TRAIT 191) (T_TRAIT_C 76) (T_TRY 192) (T_UNSET 193) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 194) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 195) (alt_if_stmt_without_else 196) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 197) (attributes 198) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 199) (class_modifier 97) (class_modifiers 200) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 201) (expr 202) (fn 104) (fully_dereferenceable 105) #'203 (function_call 107) (function_declaration_statement 204) (if_stmt 205) (if_stmt_without_else 206) (inline_function 111) (inner_statement 207) (interface_declaration_statement 208) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 209) (static_member 121) (trait_declaration_statement 210) (variable 124) (variable_class_name 125) ("{" 211) ("}" 30537) ("~" 127)) 14479 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30539) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14480 (("%" 299) (")" 30540) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14481 ((":" 30541) (switch_case_list 30542) ("{" 30543)) 14482 ((";" 30549) (case_list 30550)) 14483 ((";" 30544) (case_list 30545)) 14484 ((case_list 30547)) 14485 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 30546)) 14486 ((T_CASE 17654) (T_DEFAULT 17655) ("}" 30548)) 14487 ((case_list 30553)) 14488 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 30551)) 14489 ((";" 30552)) 14490 ((T_CASE 17790) (T_DEFAULT 17791) (T_ENDSWITCH 30554)) 14491 ((";" 30555)) 14492 (("," 17937) (";" 30558)) 14493 ((";" 30560)) 14494 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30562) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14495 (("%" 299) (")" 30563) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14496 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21305) (";" 27966) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 27967) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27968) (T_DEC 22) (T_DECLARE 27969) (T_DIR 24) (T_DNUMBER 25) (T_DO 27970) (T_DOUBLE_CAST 27) (T_ECHO 27971) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27972) (T_FOREACH 27973) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 27974) (T_GOTO 27975) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27976) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 27977) (T_START_HEREDOC 69) (T_STATIC 27978) (T_STRING 27979) (T_STRING_CAST 72) (T_SWITCH 27980) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 27981) (T_UNSET 27982) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 27983) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 27984) (alt_if_stmt_without_else 27985) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 27986) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 27987) (if_stmt_without_else 27988) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30564) (static_member 121) (variable 124) (variable_class_name 125) ("{" 27990) ("~" 127)) 14497 ((";" 30566)) 14498 (("," 18988) (";" 30568)) 14499 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 30570) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 14500 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 30571) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 14501 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 30572) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 14502 ((")" 30573) (T_DOUBLE_ARROW 30574)) 14503 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30577) (";" 30464) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30465) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30466) (T_DEC 22) (T_DECLARE 30467) (T_DIR 24) (T_DNUMBER 25) (T_DO 30468) (T_DOUBLE_CAST 27) (T_ECHO 30469) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30470) (T_FOREACH 30471) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30472) (T_GOTO 30473) (T_IF 30474) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30475) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30476) (T_START_HEREDOC 69) (T_STATIC 30477) (T_STRING 30478) (T_STRING_CAST 72) (T_SWITCH 30479) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30480) (T_UNSET 30481) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30482) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30483) (alt_if_stmt_without_else 30484) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30485) (fn 104) (foreach_statement 30583) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30486) (if_stmt_without_else 30487) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30579) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30489) ("~" 127)) 14504 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 30575) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 14505 ((")" 30576)) 14506 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30577) (";" 30464) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30465) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30466) (T_DEC 22) (T_DECLARE 30467) (T_DIR 24) (T_DNUMBER 25) (T_DO 30468) (T_DOUBLE_CAST 27) (T_ECHO 30469) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30470) (T_FOREACH 30471) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30472) (T_GOTO 30473) (T_IF 30474) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30475) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30476) (T_START_HEREDOC 69) (T_STATIC 30477) (T_STRING 30478) (T_STRING_CAST 72) (T_SWITCH 30479) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30480) (T_UNSET 30481) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30482) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30483) (alt_if_stmt_without_else 30484) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30485) (fn 104) (foreach_statement 30578) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30486) (if_stmt_without_else 30487) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30579) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30489) ("~" 127)) 14507 ((inner_statement_list 30580)) 14508 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 30581) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 14509 ((";" 30582)) 14510 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 30585) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 14511 ((";" 30586)) 14512 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 30587) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 14513 ((";" 30588)) 14514 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 30589) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 14515 ((")" 30590)) 14516 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30591) (";" 30464) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30465) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30466) (T_DEC 22) (T_DECLARE 30467) (T_DIR 24) (T_DNUMBER 25) (T_DO 30468) (T_DOUBLE_CAST 27) (T_ECHO 30469) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30470) (T_FOREACH 30471) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30472) (T_GOTO 30473) (T_IF 30474) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30475) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30476) (T_START_HEREDOC 69) (T_STATIC 30477) (T_STRING 30478) (T_STRING_CAST 72) (T_SWITCH 30479) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30480) (T_UNSET 30481) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30482) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30483) (alt_if_stmt_without_else 30484) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 30485) (fn 104) (for_statement 30592) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30486) (if_stmt_without_else 30487) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30593) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30489) ("~" 127)) 14517 ((inner_statement_list 30594)) 14518 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 30595) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 14519 ((";" 30596)) 14520 (("," 20829) (";" 30598)) 14521 ((T_WHILE 30600)) 14522 (("(" 30601)) 14523 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30602) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14524 (("%" 299) (")" 30603) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14525 ((";" 30604)) 14526 ((T_STRING 21002) (const_decl 21003) (const_list 30606)) 14527 ((")" 30607) ("," 21006)) 14528 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30608) (";" 30464) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 30465) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30466) (T_DEC 22) (T_DECLARE 30467) (T_DIR 24) (T_DNUMBER 25) (T_DO 30468) (T_DOUBLE_CAST 27) (T_ECHO 30469) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30470) (T_FOREACH 30471) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 30472) (T_GOTO 30473) (T_IF 30474) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30475) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30476) (T_START_HEREDOC 69) (T_STATIC 30477) (T_STRING 30478) (T_STRING_CAST 72) (T_SWITCH 30479) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30480) (T_UNSET 30481) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 30482) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 30483) (alt_if_stmt_without_else 30484) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 30609) (dereferenceable_scalar 101) (expr 30485) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 30486) (if_stmt_without_else 30487) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30610) (static_member 121) (variable 124) (variable_class_name 125) ("{" 30489) ("~" 127)) 14529 ((inner_statement_list 30611)) 14530 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 30612) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 14531 ((";" 30613)) 14532 ((";" 30615)) 14533 ((";" 30617)) 14534 ((")" 30619)) 14535 ((";" 30620)) 14536 ((";" 30622)) 14537 (("," 18988) (";" 30624)) 14538 (("!" 18996) ("\"" 18997) ("$" 18998) ("(" 18999) ("+" 19000) ("-" 19001) ("@" 19002) (T_ARRAY 19003) (T_ARRAY_CAST 19004) (T_ATTRIBUTE 128) (T_BOOL_CAST 19005) (T_CLASS_C 19006) (T_CLONE 19007) (T_CONSTANT_ENCAPSED_STRING 19008) (T_DEC 19009) (T_DIR 19010) (T_DNUMBER 19011) (T_DOUBLE_CAST 19012) (T_EMPTY 19013) (T_EVAL 19014) (T_EXIT 19015) (T_FILE 19016) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 19017) (T_INC 19018) (T_INCLUDE 19019) (T_INCLUDE_ONCE 19020) (T_INT_CAST 19021) (T_ISSET 19022) (T_LINE 19023) (T_LIST 19024) (T_LNUMBER 19025) (T_MATCH 19026) (T_METHOD_C 19027) (T_NAME_FULLY_QUALIFIED 19028) (T_NAME_QUALIFIED 19029) (T_NAME_RELATIVE 19030) (T_NEW 19031) (T_NS_C 19032) (T_OBJECT_CAST 19033) (T_PRINT 19034) (T_PROPERTY_C 19035) (T_READONLY 19036) (T_REQUIRE 19037) (T_REQUIRE_ONCE 19038) (T_START_HEREDOC 19039) (T_STATIC 19040) (T_STRING 19041) (T_STRING_CAST 19042) (T_THROW 19043) (T_TRAIT_C 19044) (T_UNSET_CAST 19045) (T_VARIABLE 19046) (T_YIELD 19047) (T_YIELD_FROM 19048) ("[" 19049) ("`" 19050) (array_object_dereferenceable 19051) (attribute 133) (attributes 19052) (callable_expr 19053) (callable_variable 19054) (class_constant 19055) (class_name 19056) (constant 19057) (dereferenceable_scalar 19058) (expr 30626) (fn 19060) (fully_dereferenceable 105) #'19061 (function_call 19062) (inline_function 19063) (internal_functions_in_yacc 19064) (match 19065) (name 19066) (new_dereferenceable 19067) (new_non_dereferenceable 19068) (scalar 19069) (simple_variable 19070) (static_member 19071) (variable 19072) (variable_class_name 19073) ("~" 19074)) 14539 (("%" 19076) ("*" 19077) ("+" 19078) ("-" 19079) ("." 19080) ("/" 19081) ("<" 19082) (">" 19083) ("?" 19084) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19085) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19086) (T_AS 30627) (T_BOOLEAN_AND 19087) (T_BOOLEAN_OR 19088) (T_COALESCE 19089) (T_INSTANCEOF 19090) (T_IS_EQUAL 19091) (T_IS_GREATER_OR_EQUAL 19092) (T_IS_IDENTICAL 19093) (T_IS_NOT_EQUAL 19094) (T_IS_NOT_IDENTICAL 19095) (T_IS_SMALLER_OR_EQUAL 19096) (T_LOGICAL_AND 19097) (T_LOGICAL_OR 19098) (T_LOGICAL_XOR 19099) (T_POW 19100) (T_SL 19101) (T_SPACESHIP 19102) (T_SR 19103) ("^" 19104) ("|" 19105)) 14540 (("\"" 1520) ("$" 20411) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20412) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20413) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20414) ("[" 20415) (ampersand 20416) (array_object_dereferenceable 20417) (callable_expr 20418) (callable_variable 20419) (class_constant 1546) (class_name 20420) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 30628) (fully_dereferenceable 105) (function_call 20422) (name 20423) (new_dereferenceable 1552) (simple_variable 20424) (static_member 20425) (variable 20426) (variable_class_name 20427)) 14541 ((")" 30629) (T_DOUBLE_ARROW 30630)) 14542 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30633) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 68) (T_START_HEREDOC 69) (T_STATIC 70) (T_STRING 71) (T_STRING_CAST 72) (T_SWITCH 73) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 77) (T_UNSET 78) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 82) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 87) (alt_if_stmt_without_else 88) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 103) (fn 104) (foreach_statement 30639) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 109) (if_stmt_without_else 110) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30635) (static_member 121) (variable 124) (variable_class_name 125) ("{" 126) ("~" 127)) 14543 (("\"" 1520) ("$" 20442) ("(" 1522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1523) (T_CLASS_C 1524) (T_CONSTANT_ENCAPSED_STRING 1525) (T_DIR 1526) (T_FILE 1527) (T_FUNC_C 1528) (T_LINE 1529) (T_LIST 20443) (T_METHOD_C 1530) (T_NAME_FULLY_QUALIFIED 1531) (T_NAME_QUALIFIED 1532) (T_NAME_RELATIVE 1533) (T_NEW 1534) (T_NS_C 1535) (T_PROPERTY_C 1536) (T_READONLY 20444) (T_STATIC 1538) (T_STRING 1539) (T_TRAIT_C 1540) (T_VARIABLE 20445) ("[" 20446) (ampersand 20447) (array_object_dereferenceable 20448) (callable_expr 20449) (callable_variable 20450) (class_constant 1546) (class_name 20451) (constant 1548) (dereferenceable_scalar 1549) (foreach_variable 30631) (fully_dereferenceable 105) (function_call 20453) (name 20454) (new_dereferenceable 1552) (simple_variable 20455) (static_member 20456) (variable 20457) (variable_class_name 20458)) 14544 ((")" 30632)) 14545 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30633) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 68) (T_START_HEREDOC 69) (T_STATIC 70) (T_STRING 71) (T_STRING_CAST 72) (T_SWITCH 73) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 77) (T_UNSET 78) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 82) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 87) (alt_if_stmt_without_else 88) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 103) (fn 104) (foreach_statement 30634) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 109) (if_stmt_without_else 110) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30635) (static_member 121) (variable 124) (variable_class_name 125) ("{" 126) ("~" 127)) 14546 ((inner_statement_list 30636)) 14547 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20476) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20477) (T_CLASS 20478) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20479) (T_DEC 22) (T_DECLARE 20480) (T_DIR 24) (T_DNUMBER 25) (T_DO 20481) (T_DOUBLE_CAST 27) (T_ECHO 20482) (T_EMPTY 29) (T_ENDFOREACH 30637) (T_ENUM 20484) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20485) (T_FOREACH 20486) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20487) (T_GOTO 20488) (T_HALT_COMPILER 20489) (T_IF 20490) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20491) (T_INTERFACE 20492) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20493) (T_START_HEREDOC 69) (T_STATIC 20494) (T_STRING 20495) (T_STRING_CAST 72) (T_SWITCH 20496) (T_THROW 74) (T_TRAIT 20497) (T_TRAIT_C 76) (T_TRY 20498) (T_UNSET 20499) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20500) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20501) (alt_if_stmt_without_else 20502) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20503) (attributes 20504) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20505) (class_modifier 97) (class_modifiers 20506) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20507) (expr 20508) (fn 104) (fully_dereferenceable 105) #'20509 (function_call 107) (function_declaration_statement 20510) (if_stmt 20511) (if_stmt_without_else 20512) (inline_function 111) (inner_statement 20513) (interface_declaration_statement 20514) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20515) (static_member 121) (trait_declaration_statement 20516) (variable 124) (variable_class_name 125) ("{" 20517) ("~" 127)) 14548 ((";" 30638)) 14549 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 30641) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 14550 ((";" 30642)) 14551 (("!" 17941) ("\"" 17942) ("$" 17943) ("(" 17944) ("+" 17945) ("-" 17946) ("@" 17947) (T_ARRAY 17948) (T_ARRAY_CAST 17949) (T_ATTRIBUTE 128) (T_BOOL_CAST 17950) (T_CLASS_C 17951) (T_CLONE 17952) (T_CONSTANT_ENCAPSED_STRING 17953) (T_DEC 17954) (T_DIR 17955) (T_DNUMBER 17956) (T_DOUBLE_CAST 17957) (T_EMPTY 17958) (T_EVAL 17959) (T_EXIT 17960) (T_FILE 17961) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 17962) (T_INC 17963) (T_INCLUDE 17964) (T_INCLUDE_ONCE 17965) (T_INT_CAST 17966) (T_ISSET 17967) (T_LINE 17968) (T_LIST 17969) (T_LNUMBER 17970) (T_MATCH 17971) (T_METHOD_C 17972) (T_NAME_FULLY_QUALIFIED 17973) (T_NAME_QUALIFIED 17974) (T_NAME_RELATIVE 17975) (T_NEW 17976) (T_NS_C 17977) (T_OBJECT_CAST 17978) (T_PRINT 17979) (T_PROPERTY_C 17980) (T_READONLY 17981) (T_REQUIRE 17982) (T_REQUIRE_ONCE 17983) (T_START_HEREDOC 17984) (T_STATIC 17985) (T_STRING 17986) (T_STRING_CAST 17987) (T_THROW 17988) (T_TRAIT_C 17989) (T_UNSET_CAST 17990) (T_VARIABLE 17991) (T_YIELD 17992) (T_YIELD_FROM 17993) ("[" 17994) ("`" 17995) (array_object_dereferenceable 17996) (attribute 133) (attributes 17997) (callable_expr 17998) (callable_variable 17999) (class_constant 18000) (class_name 18001) (constant 18002) (dereferenceable_scalar 18003) (expr 20636) (fn 18005) (for_exprs 30643) (fully_dereferenceable 105) #'18006 (function_call 18007) (inline_function 18008) (internal_functions_in_yacc 18009) (match 18010) (name 18011) (new_dereferenceable 18012) (new_non_dereferenceable 18013) (non_empty_for_exprs 20638) (scalar 18014) (simple_variable 18015) (static_member 18016) (variable 18017) (variable_class_name 18018) ("~" 18019)) 14552 ((";" 30644)) 14553 (("!" 2338) ("\"" 2339) ("$" 2340) ("(" 2341) ("+" 2342) ("-" 2343) ("@" 2344) (T_ARRAY 2345) (T_ARRAY_CAST 2346) (T_ATTRIBUTE 128) (T_BOOL_CAST 2347) (T_CLASS_C 2348) (T_CLONE 2349) (T_CONSTANT_ENCAPSED_STRING 2350) (T_DEC 2351) (T_DIR 2352) (T_DNUMBER 2353) (T_DOUBLE_CAST 2354) (T_EMPTY 2355) (T_EVAL 2356) (T_EXIT 2357) (T_FILE 2358) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 2359) (T_INC 2360) (T_INCLUDE 2361) (T_INCLUDE_ONCE 2362) (T_INT_CAST 2363) (T_ISSET 2364) (T_LINE 2365) (T_LIST 2418) (T_LNUMBER 2367) (T_MATCH 2368) (T_METHOD_C 2369) (T_NAME_FULLY_QUALIFIED 2370) (T_NAME_QUALIFIED 2371) (T_NAME_RELATIVE 2372) (T_NEW 2373) (T_NS_C 2374) (T_OBJECT_CAST 2375) (T_PRINT 2376) (T_PROPERTY_C 2377) (T_READONLY 2378) (T_REQUIRE 2379) (T_REQUIRE_ONCE 2380) (T_START_HEREDOC 2381) (T_STATIC 2382) (T_STRING 2383) (T_STRING_CAST 2384) (T_THROW 2385) (T_TRAIT_C 2386) (T_UNSET_CAST 2387) (T_VARIABLE 2388) (T_YIELD 2389) (T_YIELD_FROM 2390) ("[" 2391) ("`" 2392) (array_object_dereferenceable 2394) (attribute 133) (attributes 2395) (callable_expr 2396) (callable_variable 2397) (class_constant 2398) (class_name 2399) (constant 2400) (dereferenceable_scalar 2401) (expr 20644) (fn 2403) (for_exprs 30645) (fully_dereferenceable 105) #'2404 (function_call 2405) (inline_function 2406) (internal_functions_in_yacc 2407) (match 2408) (name 2409) (new_dereferenceable 2410) (new_non_dereferenceable 2411) (non_empty_for_exprs 20646) (scalar 2412) (simple_variable 2413) (static_member 2414) (variable 2415) (variable_class_name 2416) ("~" 2417)) 14554 ((")" 30646)) 14555 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30647) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 68) (T_START_HEREDOC 69) (T_STATIC 70) (T_STRING 71) (T_STRING_CAST 72) (T_SWITCH 73) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 77) (T_UNSET 78) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 82) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 87) (alt_if_stmt_without_else 88) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (dereferenceable_scalar 101) (expr 103) (fn 104) (for_statement 30648) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 109) (if_stmt_without_else 110) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30649) (static_member 121) (variable 124) (variable_class_name 125) ("{" 126) ("~" 127)) 14556 ((inner_statement_list 30650)) 14557 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20654) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 20655) (T_CLASS 20656) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20657) (T_DEC 22) (T_DECLARE 20658) (T_DIR 24) (T_DNUMBER 25) (T_DO 20659) (T_DOUBLE_CAST 27) (T_ECHO 20660) (T_EMPTY 29) (T_ENDFOR 30651) (T_ENUM 20662) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20663) (T_FOREACH 20664) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20665) (T_GOTO 20666) (T_HALT_COMPILER 20667) (T_IF 20668) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20669) (T_INTERFACE 20670) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20671) (T_START_HEREDOC 69) (T_STATIC 20672) (T_STRING 20673) (T_STRING_CAST 72) (T_SWITCH 20674) (T_THROW 74) (T_TRAIT 20675) (T_TRAIT_C 76) (T_TRY 20676) (T_UNSET 20677) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 20678) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 20679) (alt_if_stmt_without_else 20680) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 20681) (attributes 20682) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 20683) (class_modifier 97) (class_modifiers 20684) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 20685) (expr 20686) (fn 104) (fully_dereferenceable 105) #'20687 (function_call 107) (function_declaration_statement 20688) (if_stmt 20689) (if_stmt_without_else 20690) (inline_function 111) (inner_statement 20691) (interface_declaration_statement 20692) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 20693) (static_member 121) (trait_declaration_statement 20694) (variable 124) (variable_class_name 125) ("{" 20695) ("~" 127)) 14558 ((";" 30652)) 14559 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 30656) ("+" 950) ("-" 951) ("@" 952) (T_ABSTRACT 953) (T_ARRAY 954) (T_ARRAY_CAST 955) (T_AS 956) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_BREAK 958) (T_CALLABLE 959) (T_CASE 960) (T_CATCH 961) (T_CLASS 962) (T_CLASS_C 963) (T_CLONE 964) (T_CONST 965) (T_CONSTANT_ENCAPSED_STRING 966) (T_CONTINUE 967) (T_DEC 968) (T_DECLARE 969) (T_DEFAULT 970) (T_DIR 971) (T_DNUMBER 972) (T_DO 973) (T_DOUBLE_CAST 974) (T_ECHO 975) (T_ELLIPSIS 30657) (T_ELSE 977) (T_ELSEIF 978) (T_EMPTY 979) (T_ENDDECLARE 980) (T_ENDFOR 981) (T_ENDFOREACH 982) (T_ENDIF 983) (T_ENDSWITCH 984) (T_ENDWHILE 985) (T_ENUM 986) (T_EVAL 987) (T_EXIT 988) (T_EXTENDS 989) (T_FILE 990) (T_FINAL 991) (T_FINALLY 992) (T_FN 993) (T_FOR 994) (T_FOREACH 995) (T_FUNCTION 996) (T_FUNC_C 997) (T_GLOBAL 998) (T_GOTO 999) (T_IF 1000) (T_IMPLEMENTS 1001) (T_INC 1002) (T_INCLUDE 1003) (T_INCLUDE_ONCE 1004) (T_INSTANCEOF 1005) (T_INSTEADOF 1006) (T_INTERFACE 1007) (T_INT_CAST 1008) (T_ISSET 1009) (T_LINE 1010) (T_LIST 1011) (T_LNUMBER 1012) (T_LOGICAL_AND 1013) (T_LOGICAL_OR 1014) (T_LOGICAL_XOR 1015) (T_MATCH 1016) (T_METHOD_C 1017) (T_NAMESPACE 1018) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1022) (T_NS_C 1023) (T_OBJECT_CAST 1024) (T_PRINT 1025) (T_PRIVATE 1026) (T_PROPERTY_C 1027) (T_PROTECTED 1028) (T_PUBLIC 1029) (T_READONLY 1030) (T_REQUIRE 1031) (T_REQUIRE_ONCE 1032) (T_RETURN 1033) (T_START_HEREDOC 1034) (T_STATIC 1035) (T_STRING 1036) (T_STRING_CAST 1037) (T_SWITCH 1038) (T_THROW 1039) (T_TRAIT 1040) (T_TRAIT_C 1041) (T_TRY 1042) (T_UNSET 1043) (T_UNSET_CAST 1044) (T_USE 1045) (T_VAR 1046) (T_VARIABLE 1047) (T_WHILE 1048) (T_YIELD 1049) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (argument 1053) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1062) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (identifier 1066) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (non_empty_argument_list 30658) (reserved_non_modifiers 1074) (scalar 1075) (semi_reserved 1076) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 14560 (("!" 945) ("\"" 946) ("$" 947) ("(" 948) (")" 30661) ("+" 950) ("-" 951) ("@" 952) (T_ARRAY 1082) (T_ARRAY_CAST 955) (T_ATTRIBUTE 128) (T_BOOL_CAST 957) (T_CLASS_C 1083) (T_CLONE 1084) (T_CONSTANT_ENCAPSED_STRING 966) (T_DEC 968) (T_DIR 1085) (T_DNUMBER 972) (T_DOUBLE_CAST 974) (T_EMPTY 1086) (T_EVAL 1087) (T_EXIT 1088) (T_FILE 1089) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1090) (T_INC 1002) (T_INCLUDE 1091) (T_INCLUDE_ONCE 1092) (T_INT_CAST 1008) (T_ISSET 1093) (T_LINE 1094) (T_LIST 1095) (T_LNUMBER 1012) (T_MATCH 1096) (T_METHOD_C 1097) (T_NAME_FULLY_QUALIFIED 1019) (T_NAME_QUALIFIED 1020) (T_NAME_RELATIVE 1021) (T_NEW 1098) (T_NS_C 1099) (T_OBJECT_CAST 1024) (T_PRINT 1100) (T_PROPERTY_C 1101) (T_READONLY 1102) (T_REQUIRE 1103) (T_REQUIRE_ONCE 1104) (T_START_HEREDOC 1034) (T_STATIC 1105) (T_STRING 1106) (T_STRING_CAST 1037) (T_THROW 1107) (T_TRAIT_C 1108) (T_UNSET_CAST 1044) (T_VARIABLE 1047) (T_YIELD 1109) (T_YIELD_FROM 1050) ("[" 1051) ("`" 1052) (array_object_dereferenceable 1054) (attribute 133) (attributes 1055) (callable_expr 1056) (callable_variable 1057) (class_constant 1058) (class_name 1059) (constant 1060) (dereferenceable_scalar 1061) (expr 1485) (fn 1063) (fully_dereferenceable 105) #'1064 (function_call 1065) (inline_function 1067) (internal_functions_in_yacc 1068) (match 1069) (name 1070) (new_dereferenceable 1071) (new_non_dereferenceable 1072) (scalar 1075) (simple_variable 1077) (static_member 1078) (variable 1079) (variable_class_name 1080) ("~" 1081)) 14561 (("," 1480) (possible_comma 30659)) 14562 ((")" 30660)) 14563 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30663) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14564 (("%" 299) (")" 30664) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14565 ((":" 21344) (enum_backing_type 30666)) 14566 ((T_IMPLEMENTS 2965) (implements_list 30667)) 14567 ((backup_doc_comment 30668)) 14568 (("{" 30669)) 14569 ((class_statement_list 30670)) 14570 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 30671)) 14571 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30673) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14572 (("%" 299) (")" 30674) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14573 (("," 20829) (";" 30676)) 14574 ((T_WHILE 30679)) 14575 (("(" 30680)) 14576 (("!" 219) ("\"" 220) ("$" 221) ("(" 222) ("+" 223) ("-" 224) ("@" 225) (T_ARRAY 226) (T_ARRAY_CAST 227) (T_ATTRIBUTE 128) (T_BOOL_CAST 228) (T_CLASS_C 229) (T_CLONE 230) (T_CONSTANT_ENCAPSED_STRING 231) (T_DEC 232) (T_DIR 233) (T_DNUMBER 234) (T_DOUBLE_CAST 235) (T_EMPTY 236) (T_EVAL 237) (T_EXIT 238) (T_FILE 239) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 240) (T_INC 241) (T_INCLUDE 242) (T_INCLUDE_ONCE 243) (T_INT_CAST 244) (T_ISSET 245) (T_LINE 246) (T_LIST 247) (T_LNUMBER 248) (T_MATCH 249) (T_METHOD_C 250) (T_NAME_FULLY_QUALIFIED 251) (T_NAME_QUALIFIED 252) (T_NAME_RELATIVE 253) (T_NEW 254) (T_NS_C 255) (T_OBJECT_CAST 256) (T_PRINT 257) (T_PROPERTY_C 258) (T_READONLY 259) (T_REQUIRE 260) (T_REQUIRE_ONCE 261) (T_START_HEREDOC 262) (T_STATIC 263) (T_STRING 264) (T_STRING_CAST 265) (T_THROW 266) (T_TRAIT_C 267) (T_UNSET_CAST 268) (T_VARIABLE 269) (T_YIELD 270) (T_YIELD_FROM 271) ("[" 272) ("`" 273) (array_object_dereferenceable 274) (attribute 133) (attributes 275) (callable_expr 276) (callable_variable 277) (class_constant 278) (class_name 279) (constant 280) (dereferenceable_scalar 281) (expr 30681) (fn 283) (fully_dereferenceable 105) #'284 (function_call 285) (inline_function 286) (internal_functions_in_yacc 287) (match 288) (name 289) (new_dereferenceable 290) (new_non_dereferenceable 291) (scalar 292) (simple_variable 293) (static_member 294) (variable 295) (variable_class_name 296) ("~" 297)) 14577 (("%" 299) (")" 30682) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14578 ((";" 30683)) 14579 ((T_STRING 21002) (const_decl 21003) (const_list 30685)) 14580 ((")" 30686) ("," 21006)) 14581 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30687) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 128) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 129) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 130) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 68) (T_START_HEREDOC 69) (T_STATIC 70) (T_STRING 71) (T_STRING_CAST 72) (T_SWITCH 73) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 77) (T_UNSET 78) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 82) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 87) (alt_if_stmt_without_else 88) (array_object_dereferenceable 89) (attribute 133) (attributes 134) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_name 99) (constant 100) (declare_statement 30688) (dereferenceable_scalar 101) (expr 103) (fn 104) (fully_dereferenceable 105) #'136 (function_call 107) (if_stmt 109) (if_stmt_without_else 110) (inline_function 111) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 30689) (static_member 121) (variable 124) (variable_class_name 125) ("{" 126) ("~" 127)) 14582 ((inner_statement_list 30690)) 14583 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS 21014) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21015) (T_DEC 22) (T_DECLARE 21016) (T_DIR 24) (T_DNUMBER 25) (T_DO 21017) (T_DOUBLE_CAST 27) (T_ECHO 21018) (T_EMPTY 29) (T_ENDDECLARE 30691) (T_ENUM 21020) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 21021) (T_FOREACH 21022) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 21023) (T_GOTO 21024) (T_HALT_COMPILER 21025) (T_IF 21026) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21027) (T_INTERFACE 21028) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21029) (T_START_HEREDOC 69) (T_STATIC 21030) (T_STRING 21031) (T_STRING_CAST 72) (T_SWITCH 21032) (T_THROW 74) (T_TRAIT 21033) (T_TRAIT_C 76) (T_TRY 21034) (T_UNSET 21035) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_WHILE 21036) (T_YIELD 83) (T_YIELD_FROM 84) ("[" 85) ("`" 86) (alt_if_stmt 21037) (alt_if_stmt_without_else 21038) (array_object_dereferenceable 89) (attribute 90) (attributed_statement 21039) (attributes 21040) (callable_expr 93) (callable_variable 94) (class_constant 95) (class_declaration_statement 21041) (class_modifier 97) (class_modifiers 21042) (class_name 99) (constant 100) (dereferenceable_scalar 101) (enum_declaration_statement 21043) (expr 21044) (fn 104) (fully_dereferenceable 105) #'21045 (function_call 107) (function_declaration_statement 21046) (if_stmt 21047) (if_stmt_without_else 21048) (inline_function 111) (inner_statement 21049) (interface_declaration_statement 21050) (internal_functions_in_yacc 113) (match 114) (name 115) (new_dereferenceable 116) (new_non_dereferenceable 117) (scalar 118) (simple_variable 119) (statement 21051) (static_member 121) (trait_declaration_statement 21052) (variable 124) (variable_class_name 125) ("{" 21053) ("~" 127)) 14584 ((";" 30692)) 14585 ((";" 30695)) 14586 (("," 30417) (";" 30697)) 14587 ((T_EXTENDS 2963) (extends_from 30700)) 14588 ((T_IMPLEMENTS 2965) (implements_list 30701)) 14589 ((backup_doc_comment 30702)) 14590 (("{" 30703)) 14591 ((class_statement_list 30704)) 14592 ((T_ABSTRACT 2970) (T_ATTRIBUTE 2971) (T_CASE 2972) (T_FINAL 2973) (T_PRIVATE 2974) (T_PRIVATE_SET 2975) (T_PROTECTED 2976) (T_PROTECTED_SET 2977) (T_PUBLIC 2978) (T_PUBLIC_SET 2979) (T_READONLY 2980) (T_STATIC 2981) (T_USE 2982) (T_VAR 2983) (attribute 2984) (attributed_class_statement 2985) (attributes 2986) (class_const_modifiers 2987) (class_statement 2988) (enum_case 2989) (member_modifier 2990) (method_modifiers 2991) (non_empty_member_modifiers 2992) (property_modifiers 2993) ("}" 30705)) 14593 ((";" 30707)) 14594 (("," 5847) (possible_comma 30710)) 14595 (("]" 30711)) 14596 (("!" 1863) ("\"" 1864) ("$" 1865) ("(" 1866) ("+" 1867) ("-" 1868) ("@" 1869) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1516) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1517) (T_ARRAY 1870) (T_ARRAY_CAST 1871) (T_ATTRIBUTE 128) (T_BOOL_CAST 1872) (T_CLASS_C 1873) (T_CLONE 1874) (T_CONSTANT_ENCAPSED_STRING 1875) (T_DEC 1876) (T_DIR 1877) (T_DNUMBER 1878) (T_DOUBLE_CAST 1879) (T_ELLIPSIS 1880) (T_EMPTY 1881) (T_EVAL 1882) (T_EXIT 1883) (T_FILE 1884) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 1885) (T_INC 1886) (T_INCLUDE 1887) (T_INCLUDE_ONCE 1888) (T_INT_CAST 1889) (T_ISSET 1890) (T_LINE 1891) (T_LIST 1892) (T_LNUMBER 1893) (T_MATCH 1894) (T_METHOD_C 1895) (T_NAME_FULLY_QUALIFIED 1896) (T_NAME_QUALIFIED 1897) (T_NAME_RELATIVE 1898) (T_NEW 1899) (T_NS_C 1900) (T_OBJECT_CAST 1901) (T_PRINT 1902) (T_PROPERTY_C 1903) (T_READONLY 1904) (T_REQUIRE 1905) (T_REQUIRE_ONCE 1906) (T_START_HEREDOC 1907) (T_STATIC 1908) (T_STRING 1909) (T_STRING_CAST 1910) (T_THROW 1911) (T_TRAIT_C 1912) (T_UNSET_CAST 1913) (T_VARIABLE 1914) (T_YIELD 1915) (T_YIELD_FROM 1916) ("[" 1917) ("`" 1918) (ampersand 1919) (array_object_dereferenceable 1920) (array_pair 1921) (array_pair_list 30714) (attribute 133) (attributes 1923) (callable_expr 1924) (callable_variable 1925) (class_constant 1926) (class_name 1927) (constant 1928) (dereferenceable_scalar 1929) (expr 1930) (fn 1931) (fully_dereferenceable 105) #'1932 (function_call 1933) (inline_function 1934) (internal_functions_in_yacc 1935) (match 1936) (name 1937) (new_dereferenceable 1938) (new_non_dereferenceable 1939) (non_empty_array_pair_list 1940) (possible_array_pair 1941) (scalar 1942) (simple_variable 1943) (static_member 1944) (variable 1945) (variable_class_name 1946) ("~" 1947)) 14597 ((")" 30715)) 14598 (("%" 299) (")" 30720) ("*" 300) ("+" 301) ("-" 302) ("." 303) ("/" 304) ("<" 305) (">" 306) ("?" 307) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 308) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 309) (T_BOOLEAN_AND 310) (T_BOOLEAN_OR 311) (T_COALESCE 312) (T_INSTANCEOF 313) (T_IS_EQUAL 314) (T_IS_GREATER_OR_EQUAL 315) (T_IS_IDENTICAL 316) (T_IS_NOT_EQUAL 317) (T_IS_NOT_IDENTICAL 318) (T_IS_SMALLER_OR_EQUAL 319) (T_LOGICAL_AND 320) (T_LOGICAL_OR 321) (T_LOGICAL_XOR 322) (T_POW 323) (T_SL 324) (T_SPACESHIP 325) (T_SR 326) ("^" 327) ("|" 328)) 14599 (("!" 511) ("\"" 512) ("$" 513) ("(" 514) ("+" 515) ("-" 516) ("@" 517) (T_ARRAY 518) (T_ARRAY_CAST 519) (T_ATTRIBUTE 128) (T_BOOL_CAST 520) (T_CLASS_C 521) (T_CLONE 522) (T_CONSTANT_ENCAPSED_STRING 523) (T_DEC 524) (T_DIR 525) (T_DNUMBER 526) (T_DOUBLE_CAST 527) (T_EMPTY 528) (T_EVAL 529) (T_EXIT 530) (T_FILE 531) (T_FN 35) (T_FUNCTION 129) (T_FUNC_C 532) (T_INC 533) (T_INCLUDE 534) (T_INCLUDE_ONCE 535) (T_INT_CAST 536) (T_ISSET 537) (T_LINE 538) (T_LIST 539) (T_LNUMBER 540) (T_MATCH 541) (T_METHOD_C 542) (T_NAME_FULLY_QUALIFIED 543) (T_NAME_QUALIFIED 544) (T_NAME_RELATIVE 545) (T_NEW 546) (T_NS_C 547) (T_OBJECT_CAST 548) (T_PRINT 549) (T_PROPERTY_C 550) (T_READONLY 551) (T_REQUIRE 552) (T_REQUIRE_ONCE 553) (T_START_HEREDOC 554) (T_STATIC 555) (T_STRING 556) (T_STRING_CAST 557) (T_THROW 558) (T_TRAIT_C 559) (T_UNSET_CAST 560) (T_VARIABLE 561) (T_YIELD 562) (T_YIELD_FROM 563) ("[" 564) ("`" 565) (array_object_dereferenceable 566) (attribute 133) (attributes 567) (callable_expr 568) (callable_variable 569) (class_constant 570) (class_name 571) (constant 572) (dereferenceable_scalar 573) (expr 30723) (fn 575) (fully_dereferenceable 105) #'576 (function_call 577) (inline_function 578) (internal_functions_in_yacc 579) (match 580) (name 581) (new_dereferenceable 582) (new_non_dereferenceable 583) (scalar 584) (simple_variable 585) (static_member 586) (variable 587) (variable_class_name 588) ("~" 589)) 14600 (("%" 591) ("*" 592) ("+" 593) ("-" 594) ("." 595) ("/" 596) ("<" 597) (">" 598) ("?" 599) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 600) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 601) (T_BOOLEAN_AND 602) (T_BOOLEAN_OR 603) (T_COALESCE 604) (T_INSTANCEOF 605) (T_IS_EQUAL 606) (T_IS_GREATER_OR_EQUAL 607) (T_IS_IDENTICAL 608) (T_IS_NOT_EQUAL 609) (T_IS_NOT_IDENTICAL 610) (T_IS_SMALLER_OR_EQUAL 611) (T_LOGICAL_AND 612) (T_LOGICAL_OR 613) (T_LOGICAL_XOR 614) (T_POW 615) (T_SL 616) (T_SPACESHIP 617) (T_SR 618) ("^" 619) ("|" 620) ("}" 30724)) 14601 (("\"" 30726) (T_CURLY_OPEN 5232) (T_DOLLAR_OPEN_CURLY_BRACES 5233) (T_ENCAPSED_AND_WHITESPACE 5239) (T_VARIABLE 5235) (encaps_var 5240))))
  "The generated distinct goto-tables.")

(defvar
  phps-mode-parser--table-productions-number-reverse
  #s(hash-table size 730 test equal rehash-size 1.5 rehash-threshold 0.8125 data (0 ((start) (top_statement_list)) 1 ((reserved_non_modifiers) (T_INCLUDE)) 2 ((reserved_non_modifiers) (T_INCLUDE_ONCE)) 3 ((reserved_non_modifiers) (T_EVAL)) 4 ((reserved_non_modifiers) (T_REQUIRE)) 5 ((reserved_non_modifiers) (T_REQUIRE_ONCE)) 6 ((reserved_non_modifiers) (T_LOGICAL_OR)) 7 ((reserved_non_modifiers) (T_LOGICAL_XOR)) 8 ((reserved_non_modifiers) (T_LOGICAL_AND)) 9 ((reserved_non_modifiers) (T_INSTANCEOF)) 10 ((reserved_non_modifiers) (T_NEW)) 11 ((reserved_non_modifiers) (T_CLONE)) 12 ((reserved_non_modifiers) (T_EXIT)) 13 ((reserved_non_modifiers) (T_IF)) 14 ((reserved_non_modifiers) (T_ELSEIF)) 15 ((reserved_non_modifiers) (T_ELSE)) 16 ((reserved_non_modifiers) (T_ENDIF)) 17 ((reserved_non_modifiers) (T_ECHO)) 18 ((reserved_non_modifiers) (T_DO)) 19 ((reserved_non_modifiers) (T_WHILE)) 20 ((reserved_non_modifiers) (T_ENDWHILE)) 21 ((reserved_non_modifiers) (T_FOR)) 22 ((reserved_non_modifiers) (T_ENDFOR)) 23 ((reserved_non_modifiers) (T_FOREACH)) 24 ((reserved_non_modifiers) (T_ENDFOREACH)) 25 ((reserved_non_modifiers) (T_DECLARE)) 26 ((reserved_non_modifiers) (T_ENDDECLARE)) 27 ((reserved_non_modifiers) (T_AS)) 28 ((reserved_non_modifiers) (T_TRY)) 29 ((reserved_non_modifiers) (T_CATCH)) 30 ((reserved_non_modifiers) (T_FINALLY)) 31 ((reserved_non_modifiers) (T_THROW)) 32 ((reserved_non_modifiers) (T_USE)) 33 ((reserved_non_modifiers) (T_INSTEADOF)) 34 ((reserved_non_modifiers) (T_GLOBAL)) 35 ((reserved_non_modifiers) (T_VAR)) 36 ((reserved_non_modifiers) (T_UNSET)) 37 ((reserved_non_modifiers) (T_ISSET)) 38 ((reserved_non_modifiers) (T_EMPTY)) 39 ((reserved_non_modifiers) (T_CONTINUE)) 40 ((reserved_non_modifiers) (T_GOTO)) 41 ((reserved_non_modifiers) (T_FUNCTION)) 42 ((reserved_non_modifiers) (T_CONST)) 43 ((reserved_non_modifiers) (T_RETURN)) 44 ((reserved_non_modifiers) (T_PRINT)) 45 ((reserved_non_modifiers) (T_YIELD)) 46 ((reserved_non_modifiers) (T_LIST)) 47 ((reserved_non_modifiers) (T_SWITCH)) 48 ((reserved_non_modifiers) (T_ENDSWITCH)) 49 ((reserved_non_modifiers) (T_CASE)) 50 ((reserved_non_modifiers) (T_DEFAULT)) 51 ((reserved_non_modifiers) (T_BREAK)) 52 ((reserved_non_modifiers) (T_ARRAY)) 53 ((reserved_non_modifiers) (T_CALLABLE)) 54 ((reserved_non_modifiers) (T_EXTENDS)) 55 ((reserved_non_modifiers) (T_IMPLEMENTS)) 56 ((reserved_non_modifiers) (T_NAMESPACE)) 57 ((reserved_non_modifiers) (T_TRAIT)) 58 ((reserved_non_modifiers) (T_INTERFACE)) 59 ((reserved_non_modifiers) (T_CLASS)) 60 ((reserved_non_modifiers) (T_CLASS_C)) 61 ((reserved_non_modifiers) (T_TRAIT_C)) 62 ((reserved_non_modifiers) (T_FUNC_C)) 63 ((reserved_non_modifiers) (T_METHOD_C)) 64 ((reserved_non_modifiers) (T_LINE)) 65 ((reserved_non_modifiers) (T_FILE)) 66 ((reserved_non_modifiers) (T_DIR)) 67 ((reserved_non_modifiers) (T_NS_C)) 68 ((reserved_non_modifiers) (T_FN)) 69 ((reserved_non_modifiers) (T_MATCH)) 70 ((reserved_non_modifiers) (T_ENUM)) 71 ((reserved_non_modifiers) (T_PROPERTY_C)) 72 ((semi_reserved) (reserved_non_modifiers)) 73 ((semi_reserved) (T_STATIC)) 74 ((semi_reserved) (T_ABSTRACT)) 75 ((semi_reserved) (T_FINAL)) 76 ((semi_reserved) (T_PRIVATE)) 77 ((semi_reserved) (T_PROTECTED)) 78 ((semi_reserved) (T_PUBLIC)) 79 ((semi_reserved) (T_READONLY)) 80 ((ampersand) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG)) 81 ((ampersand) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG)) 82 ((identifier) (T_STRING)) 83 ((identifier) (semi_reserved)) 84 ((top_statement_list) (top_statement_list top_statement)) 85 ((top_statement_list) (%empty)) 86 ((namespace_declaration_name) (identifier)) 87 ((namespace_declaration_name) (T_NAME_QUALIFIED)) 88 ((namespace_name) (T_STRING)) 89 ((namespace_name) (T_NAME_QUALIFIED)) 90 ((legacy_namespace_name) (namespace_name)) 91 ((legacy_namespace_name) (T_NAME_FULLY_QUALIFIED)) 92 ((name) (T_STRING)) 93 ((name) (T_NAME_QUALIFIED)) 94 ((name) (T_NAME_FULLY_QUALIFIED)) 95 ((name) (T_NAME_RELATIVE)) 96 ((attribute_decl) (class_name)) 97 ((attribute_decl) (class_name argument_list)) 98 ((attribute_group) (attribute_decl)) 99 ((attribute_group) (attribute_group "," attribute_decl)) 100 ((attribute) (T_ATTRIBUTE attribute_group possible_comma "]")) 101 ((attributes) (attribute)) 102 ((attributes) (attributes attribute)) 103 ((attributed_statement) (function_declaration_statement)) 104 ((attributed_statement) (class_declaration_statement)) 105 ((attributed_statement) (trait_declaration_statement)) 106 ((attributed_statement) (interface_declaration_statement)) 107 ((attributed_statement) (enum_declaration_statement)) 108 ((top_statement) (statement)) 109 ((top_statement) (attributed_statement)) 110 ((top_statement) (attributes attributed_statement)) 111 ((top_statement) (T_HALT_COMPILER "(" ")" ";")) 112 ((top_statement) (T_NAMESPACE namespace_declaration_name ";")) 113 ((top_statement) (T_NAMESPACE namespace_declaration_name "{" top_statement_list "}")) 114 ((top_statement) (T_NAMESPACE "{" top_statement_list "}")) 115 ((top_statement) (T_USE mixed_group_use_declaration ";")) 116 ((top_statement) (T_USE use_type group_use_declaration ";")) 117 ((top_statement) (T_USE use_declarations ";")) 118 ((top_statement) (T_USE use_type use_declarations ";")) 119 ((top_statement) (T_CONST const_list ";")) 120 ((use_type) (T_FUNCTION)) 121 ((use_type) (T_CONST)) 122 ((group_use_declaration) (legacy_namespace_name T_NS_SEPARATOR "{" unprefixed_use_declarations possible_comma "}")) 123 ((mixed_group_use_declaration) (legacy_namespace_name T_NS_SEPARATOR "{" inline_use_declarations possible_comma "}")) 124 ((possible_comma) (%empty)) 125 ((possible_comma) (",")) 126 ((inline_use_declarations) (inline_use_declarations "," inline_use_declaration)) 127 ((inline_use_declarations) (inline_use_declaration)) 128 ((unprefixed_use_declarations) (unprefixed_use_declarations "," unprefixed_use_declaration)) 129 ((unprefixed_use_declarations) (unprefixed_use_declaration)) 130 ((use_declarations) (use_declarations "," use_declaration)) 131 ((use_declarations) (use_declaration)) 132 ((inline_use_declaration) (unprefixed_use_declaration)) 133 ((inline_use_declaration) (use_type unprefixed_use_declaration)) 134 ((unprefixed_use_declaration) (namespace_name)) 135 ((unprefixed_use_declaration) (namespace_name T_AS T_STRING)) 136 ((use_declaration) (legacy_namespace_name)) 137 ((use_declaration) (legacy_namespace_name T_AS T_STRING)) 138 ((const_list) (const_list "," const_decl)) 139 ((const_list) (const_decl)) 140 ((inner_statement_list) (inner_statement_list inner_statement)) 141 ((inner_statement_list) (%empty)) 142 ((inner_statement) (statement)) 143 ((inner_statement) (attributed_statement)) 144 ((inner_statement) (attributes attributed_statement)) 145 ((inner_statement) (T_HALT_COMPILER "(" ")" ";")) 146 ((statement) ("{" inner_statement_list "}")) 147 ((statement) (if_stmt)) 148 ((statement) (alt_if_stmt)) 149 ((statement) (T_WHILE "(" expr ")" while_statement)) 150 ((statement) (T_DO statement T_WHILE "(" expr ")" ";")) 151 ((statement) (T_FOR "(" for_exprs ";" for_exprs ";" for_exprs ")" for_statement)) 152 ((statement) (T_SWITCH "(" expr ")" switch_case_list)) 153 ((statement) (T_BREAK optional_expr ";")) 154 ((statement) (T_CONTINUE optional_expr ";")) 155 ((statement) (T_RETURN optional_expr ";")) 156 ((statement) (T_GLOBAL global_var_list ";")) 157 ((statement) (T_STATIC static_var_list ";")) 158 ((statement) (T_ECHO echo_expr_list ";")) 159 ((statement) (T_INLINE_HTML)) 160 ((statement) (expr ";")) 161 ((statement) (T_UNSET "(" unset_variables possible_comma ")" ";")) 162 ((statement) (T_FOREACH "(" expr T_AS foreach_variable ")" foreach_statement)) 163 ((statement) (T_FOREACH "(" expr T_AS foreach_variable T_DOUBLE_ARROW foreach_variable ")" foreach_statement)) 164 ((statement) (T_DECLARE "(" const_list ")" declare_statement)) 165 ((statement) (";")) 166 ((statement) (T_TRY "{" inner_statement_list "}" catch_list finally_statement)) 167 ((statement) (T_GOTO T_STRING ";")) 168 ((statement) (T_STRING ":")) 169 ((catch_list) (%empty)) 170 ((catch_list) (catch_list T_CATCH "(" catch_name_list optional_variable ")" "{" inner_statement_list "}")) 171 ((catch_name_list) (class_name)) 172 ((catch_name_list) (catch_name_list "|" class_name)) 173 ((optional_variable) (%empty)) 174 ((optional_variable) (T_VARIABLE)) 175 ((finally_statement) (%empty)) 176 ((finally_statement) (T_FINALLY "{" inner_statement_list "}")) 177 ((unset_variables) (unset_variable)) 178 ((unset_variables) (unset_variables "," unset_variable)) 179 ((unset_variable) (variable)) 180 ((function_name) (T_STRING)) 181 ((function_name) (T_READONLY)) 182 ((function_declaration_statement) (function returns_ref function_name backup_doc_comment "(" parameter_list ")" return_type backup_fn_flags "{" inner_statement_list "}" backup_fn_flags)) 183 ((is_reference) (%empty)) 184 ((is_reference) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG)) 185 ((is_variadic) (%empty)) 186 ((is_variadic) (T_ELLIPSIS)) 187 ((class_declaration_statement) (class_modifiers T_CLASS T_STRING extends_from implements_list backup_doc_comment "{" class_statement_list "}")) 188 ((class_declaration_statement) (T_CLASS T_STRING extends_from implements_list backup_doc_comment "{" class_statement_list "}")) 189 ((class_modifiers) (class_modifier)) 190 ((class_modifiers) (class_modifiers class_modifier)) 191 ((anonymous_class_modifiers) (class_modifier)) 192 ((anonymous_class_modifiers) (anonymous_class_modifiers class_modifier)) 193 ((anonymous_class_modifiers_optional) (%empty)) 194 ((anonymous_class_modifiers_optional) (anonymous_class_modifiers)) 195 ((class_modifier) (T_ABSTRACT)) 196 ((class_modifier) (T_FINAL)) 197 ((class_modifier) (T_READONLY)) 198 ((trait_declaration_statement) (T_TRAIT T_STRING backup_doc_comment "{" class_statement_list "}")) 199 ((interface_declaration_statement) (T_INTERFACE T_STRING interface_extends_list backup_doc_comment "{" class_statement_list "}")) 200 ((enum_declaration_statement) (T_ENUM T_STRING enum_backing_type implements_list backup_doc_comment "{" class_statement_list "}")) 201 ((enum_backing_type) (%empty)) 202 ((enum_backing_type) (":" type_expr)) 203 ((enum_case) (T_CASE backup_doc_comment identifier enum_case_expr ";")) 204 ((enum_case_expr) (%empty)) 205 ((enum_case_expr) ("=" expr)) 206 ((extends_from) (%empty)) 207 ((extends_from) (T_EXTENDS class_name)) 208 ((interface_extends_list) (%empty)) 209 ((interface_extends_list) (T_EXTENDS class_name_list)) 210 ((implements_list) (%empty)) 211 ((implements_list) (T_IMPLEMENTS class_name_list)) 212 ((foreach_variable) (variable)) 213 ((foreach_variable) (ampersand variable)) 214 ((foreach_variable) (T_LIST "(" array_pair_list ")")) 215 ((foreach_variable) ("[" array_pair_list "]")) 216 ((for_statement) (statement)) 217 ((for_statement) (":" inner_statement_list T_ENDFOR ";")) 218 ((foreach_statement) (statement)) 219 ((foreach_statement) (":" inner_statement_list T_ENDFOREACH ";")) 220 ((declare_statement) (statement)) 221 ((declare_statement) (":" inner_statement_list T_ENDDECLARE ";")) 222 ((switch_case_list) ("{" case_list "}")) 223 ((switch_case_list) ("{" ";" case_list "}")) 224 ((switch_case_list) (":" case_list T_ENDSWITCH ";")) 225 ((switch_case_list) (":" ";" case_list T_ENDSWITCH ";")) 226 ((case_list) (%empty)) 227 ((case_list) (case_list T_CASE expr case_separator inner_statement_list)) 228 ((case_list) (case_list T_DEFAULT case_separator inner_statement_list)) 229 ((case_separator) (":")) 230 ((case_separator) (";")) 231 ((match) (T_MATCH "(" expr ")" "{" match_arm_list "}")) 232 ((match_arm_list) (%empty)) 233 ((match_arm_list) (non_empty_match_arm_list possible_comma)) 234 ((non_empty_match_arm_list) (match_arm)) 235 ((non_empty_match_arm_list) (non_empty_match_arm_list "," match_arm)) 236 ((match_arm) (match_arm_cond_list possible_comma T_DOUBLE_ARROW expr)) 237 ((match_arm) (T_DEFAULT possible_comma T_DOUBLE_ARROW expr)) 238 ((match_arm_cond_list) (expr)) 239 ((match_arm_cond_list) (match_arm_cond_list "," expr)) 240 ((while_statement) (statement)) 241 ((while_statement) (":" inner_statement_list T_ENDWHILE ";")) 242 ((if_stmt_without_else) (T_IF "(" expr ")" statement)) 243 ((if_stmt_without_else) (if_stmt_without_else T_ELSEIF "(" expr ")" statement)) 244 ((if_stmt) (if_stmt_without_else)) 245 ((if_stmt) (if_stmt_without_else T_ELSE statement)) 246 ((alt_if_stmt_without_else) (T_IF "(" expr ")" ":" inner_statement_list)) 247 ((alt_if_stmt_without_else) (alt_if_stmt_without_else T_ELSEIF "(" expr ")" ":" inner_statement_list)) 248 ((alt_if_stmt) (alt_if_stmt_without_else T_ENDIF ";")) 249 ((alt_if_stmt) (alt_if_stmt_without_else T_ELSE ":" inner_statement_list T_ENDIF ";")) 250 ((parameter_list) (non_empty_parameter_list possible_comma)) 251 ((parameter_list) (%empty)) 252 ((non_empty_parameter_list) (attributed_parameter)) 253 ((non_empty_parameter_list) (non_empty_parameter_list "," attributed_parameter)) 254 ((attributed_parameter) (attributes parameter)) 255 ((attributed_parameter) (parameter)) 256 ((optional_cpp_modifiers) (%empty)) 257 ((optional_cpp_modifiers) (non_empty_member_modifiers)) 258 ((parameter) (optional_cpp_modifiers optional_type_without_static is_reference is_variadic T_VARIABLE backup_doc_comment optional_property_hook_list)) 259 ((parameter) (optional_cpp_modifiers optional_type_without_static is_reference is_variadic T_VARIABLE backup_doc_comment "=" expr optional_property_hook_list)) 260 ((optional_type_without_static) (%empty)) 261 ((optional_type_without_static) (type_expr_without_static)) 262 ((type_expr) (type)) 263 ((type_expr) ("?" type)) 264 ((type_expr) (union_type)) 265 ((type_expr) (intersection_type)) 266 ((type) (type_without_static)) 267 ((type) (T_STATIC)) 268 ((union_type_element) (type)) 269 ((union_type_element) ("(" intersection_type ")")) 270 ((union_type) (union_type_element "|" union_type_element)) 271 ((union_type) (union_type "|" union_type_element)) 272 ((intersection_type) (type T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG type)) 273 ((intersection_type) (intersection_type T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG type)) 274 ((type_expr_without_static) (type_without_static)) 275 ((type_expr_without_static) ("?" type_without_static)) 276 ((type_expr_without_static) (union_type_without_static)) 277 ((type_expr_without_static) (intersection_type_without_static)) 278 ((type_without_static) (T_ARRAY)) 279 ((type_without_static) (T_CALLABLE)) 280 ((type_without_static) (name)) 281 ((union_type_without_static_element) (type_without_static)) 282 ((union_type_without_static_element) ("(" intersection_type_without_static ")")) 283 ((union_type_without_static) (union_type_without_static_element "|" union_type_without_static_element)) 284 ((union_type_without_static) (union_type_without_static "|" union_type_without_static_element)) 285 ((intersection_type_without_static) (type_without_static T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG type_without_static)) 286 ((intersection_type_without_static) (intersection_type_without_static T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG type_without_static)) 287 ((return_type) (%empty)) 288 ((return_type) (":" type_expr)) 289 ((argument_list) ("(" ")")) 290 ((argument_list) ("(" non_empty_argument_list possible_comma ")")) 291 ((argument_list) ("(" T_ELLIPSIS ")")) 292 ((non_empty_argument_list) (argument)) 293 ((non_empty_argument_list) (non_empty_argument_list "," argument)) 294 ((argument) (expr)) 295 ((argument) (identifier ":" expr)) 296 ((argument) (T_ELLIPSIS expr)) 297 ((global_var_list) (global_var_list "," global_var)) 298 ((global_var_list) (global_var)) 299 ((global_var) (simple_variable)) 300 ((static_var_list) (static_var_list "," static_var)) 301 ((static_var_list) (static_var)) 302 ((static_var) (T_VARIABLE)) 303 ((static_var) (T_VARIABLE "=" expr)) 304 ((class_statement_list) (class_statement_list class_statement)) 305 ((class_statement_list) (%empty)) 306 ((attributed_class_statement) (property_modifiers optional_type_without_static property_list ";")) 307 ((attributed_class_statement) (property_modifiers optional_type_without_static hooked_property)) 308 ((attributed_class_statement) (class_const_modifiers T_CONST class_const_list ";")) 309 ((attributed_class_statement) (class_const_modifiers T_CONST type_expr class_const_list ";")) 310 ((attributed_class_statement) (method_modifiers function returns_ref identifier backup_doc_comment "(" parameter_list ")" return_type backup_fn_flags method_body backup_fn_flags)) 311 ((attributed_class_statement) (enum_case)) 312 ((class_statement) (attributed_class_statement)) 313 ((class_statement) (attributes attributed_class_statement)) 314 ((class_statement) (T_USE class_name_list trait_adaptations)) 315 ((class_name_list) (class_name)) 316 ((class_name_list) (class_name_list "," class_name)) 317 ((trait_adaptations) (";")) 318 ((trait_adaptations) ("{" "}")) 319 ((trait_adaptations) ("{" trait_adaptation_list "}")) 320 ((trait_adaptation_list) (trait_adaptation)) 321 ((trait_adaptation_list) (trait_adaptation_list trait_adaptation)) 322 ((trait_adaptation) (trait_precedence ";")) 323 ((trait_adaptation) (trait_alias ";")) 324 ((trait_precedence) (absolute_trait_method_reference T_INSTEADOF class_name_list)) 325 ((trait_alias) (trait_method_reference T_AS T_STRING)) 326 ((trait_alias) (trait_method_reference T_AS reserved_non_modifiers)) 327 ((trait_alias) (trait_method_reference T_AS member_modifier identifier)) 328 ((trait_alias) (trait_method_reference T_AS member_modifier)) 329 ((trait_method_reference) (identifier)) 330 ((trait_method_reference) (absolute_trait_method_reference)) 331 ((absolute_trait_method_reference) (class_name T_PAAMAYIM_NEKUDOTAYIM identifier)) 332 ((method_body) (";")) 333 ((method_body) ("{" inner_statement_list "}")) 334 ((property_modifiers) (non_empty_member_modifiers)) 335 ((property_modifiers) (T_VAR)) 336 ((method_modifiers) (%empty)) 337 ((method_modifiers) (non_empty_member_modifiers)) 338 ((class_const_modifiers) (%empty)) 339 ((class_const_modifiers) (non_empty_member_modifiers)) 340 ((non_empty_member_modifiers) (member_modifier)) 341 ((non_empty_member_modifiers) (non_empty_member_modifiers member_modifier)) 342 ((member_modifier) (T_PUBLIC)) 343 ((member_modifier) (T_PROTECTED)) 344 ((member_modifier) (T_PRIVATE)) 345 ((member_modifier) (T_PUBLIC_SET)) 346 ((member_modifier) (T_PROTECTED_SET)) 347 ((member_modifier) (T_PRIVATE_SET)) 348 ((member_modifier) (T_STATIC)) 349 ((member_modifier) (T_ABSTRACT)) 350 ((member_modifier) (T_FINAL)) 351 ((member_modifier) (T_READONLY)) 352 ((property_list) (property_list "," property)) 353 ((property_list) (property)) 354 ((property) (T_VARIABLE backup_doc_comment)) 355 ((property) (T_VARIABLE "=" expr backup_doc_comment)) 356 ((hooked_property) (T_VARIABLE backup_doc_comment "{" property_hook_list "}")) 357 ((hooked_property) (T_VARIABLE "=" expr backup_doc_comment "{" property_hook_list "}")) 358 ((property_hook_list) (%empty)) 359 ((property_hook_list) (property_hook_list property_hook)) 360 ((property_hook_list) (property_hook_list attributes property_hook)) 361 ((optional_property_hook_list) (%empty)) 362 ((optional_property_hook_list) ("{" property_hook_list "}")) 363 ((property_hook_modifiers) (%empty)) 364 ((property_hook_modifiers) (non_empty_member_modifiers)) 365 ((property_hook) (property_hook_modifiers returns_ref T_STRING backup_doc_comment optional_parameter_list backup_fn_flags property_hook_body backup_fn_flags)) 366 ((property_hook_body) (";")) 367 ((property_hook_body) ("{" inner_statement_list "}")) 368 ((property_hook_body) (T_DOUBLE_ARROW expr ";")) 369 ((optional_parameter_list) (%empty)) 370 ((optional_parameter_list) ("(" parameter_list ")")) 371 ((class_const_list) (class_const_list "," class_const_decl)) 372 ((class_const_list) (class_const_decl)) 373 ((class_const_decl) (T_STRING "=" expr backup_doc_comment)) 374 ((class_const_decl) (semi_reserved "=" expr backup_doc_comment)) 375 ((const_decl) (T_STRING "=" expr backup_doc_comment)) 376 ((echo_expr_list) (echo_expr_list "," echo_expr)) 377 ((echo_expr_list) (echo_expr)) 378 ((echo_expr) (expr)) 379 ((for_exprs) (%empty)) 380 ((for_exprs) (non_empty_for_exprs)) 381 ((non_empty_for_exprs) (non_empty_for_exprs "," expr)) 382 ((non_empty_for_exprs) (expr)) 383 ((anonymous_class) (anonymous_class_modifiers_optional T_CLASS ctor_arguments extends_from implements_list backup_doc_comment "{" class_statement_list "}")) 384 ((new_dereferenceable) (T_NEW class_name_reference argument_list)) 385 ((new_dereferenceable) (T_NEW anonymous_class)) 386 ((new_dereferenceable) (T_NEW attributes anonymous_class)) 387 ((new_non_dereferenceable) (T_NEW class_name_reference)) 388 ((expr) (variable)) 389 ((expr) (T_LIST "(" array_pair_list ")" "=" expr)) 390 ((expr) ("[" array_pair_list "]" "=" expr)) 391 ((expr) (variable "=" expr)) 392 ((expr) (variable "=" ampersand variable)) 393 ((expr) (T_CLONE expr)) 394 ((expr) (variable T_PLUS_EQUAL expr)) 395 ((expr) (variable T_MINUS_EQUAL expr)) 396 ((expr) (variable T_MUL_EQUAL expr)) 397 ((expr) (variable T_POW_EQUAL expr)) 398 ((expr) (variable T_DIV_EQUAL expr)) 399 ((expr) (variable T_CONCAT_EQUAL expr)) 400 ((expr) (variable T_MOD_EQUAL expr)) 401 ((expr) (variable T_AND_EQUAL expr)) 402 ((expr) (variable T_OR_EQUAL expr)) 403 ((expr) (variable T_XOR_EQUAL expr)) 404 ((expr) (variable T_SL_EQUAL expr)) 405 ((expr) (variable T_SR_EQUAL expr)) 406 ((expr) (variable T_COALESCE_EQUAL expr)) 407 ((expr) (variable T_INC)) 408 ((expr) (T_INC variable)) 409 ((expr) (variable T_DEC)) 410 ((expr) (T_DEC variable)) 411 ((expr) (expr T_BOOLEAN_OR expr)) 412 ((expr) (expr T_BOOLEAN_AND expr)) 413 ((expr) (expr T_LOGICAL_OR expr)) 414 ((expr) (expr T_LOGICAL_AND expr)) 415 ((expr) (expr T_LOGICAL_XOR expr)) 416 ((expr) (expr "|" expr)) 417 ((expr) (expr T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG expr)) 418 ((expr) (expr T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG expr)) 419 ((expr) (expr "^" expr)) 420 ((expr) (expr "." expr)) 421 ((expr) (expr "+" expr)) 422 ((expr) (expr "-" expr)) 423 ((expr) (expr "*" expr)) 424 ((expr) (expr T_POW expr)) 425 ((expr) (expr "/" expr)) 426 ((expr) (expr "%" expr)) 427 ((expr) (expr T_SL expr)) 428 ((expr) (expr T_SR expr)) 429 ((expr) ("+" expr)) 430 ((expr) ("-" expr)) 431 ((expr) ("!" expr)) 432 ((expr) ("~" expr)) 433 ((expr) (expr T_IS_IDENTICAL expr)) 434 ((expr) (expr T_IS_NOT_IDENTICAL expr)) 435 ((expr) (expr T_IS_EQUAL expr)) 436 ((expr) (expr T_IS_NOT_EQUAL expr)) 437 ((expr) (expr "<" expr)) 438 ((expr) (expr T_IS_SMALLER_OR_EQUAL expr)) 439 ((expr) (expr ">" expr)) 440 ((expr) (expr T_IS_GREATER_OR_EQUAL expr)) 441 ((expr) (expr T_SPACESHIP expr)) 442 ((expr) (expr T_INSTANCEOF class_name_reference)) 443 ((expr) ("(" expr ")")) 444 ((expr) (new_dereferenceable)) 445 ((expr) (new_non_dereferenceable)) 446 ((expr) (expr "?" expr ":" expr)) 447 ((expr) (expr "?" ":" expr)) 448 ((expr) (expr T_COALESCE expr)) 449 ((expr) (internal_functions_in_yacc)) 450 ((expr) (T_INT_CAST expr)) 451 ((expr) (T_DOUBLE_CAST expr)) 452 ((expr) (T_STRING_CAST expr)) 453 ((expr) (T_ARRAY_CAST expr)) 454 ((expr) (T_OBJECT_CAST expr)) 455 ((expr) (T_BOOL_CAST expr)) 456 ((expr) (T_UNSET_CAST expr)) 457 ((expr) (T_EXIT ctor_arguments)) 458 ((expr) ("@" expr)) 459 ((expr) (scalar)) 460 ((expr) ("`" backticks_expr "`")) 461 ((expr) (T_PRINT expr)) 462 ((expr) (T_YIELD)) 463 ((expr) (T_YIELD expr)) 464 ((expr) (T_YIELD expr T_DOUBLE_ARROW expr)) 465 ((expr) (T_YIELD_FROM expr)) 466 ((expr) (T_THROW expr)) 467 ((expr) (inline_function)) 468 ((expr) (attributes inline_function)) 469 ((expr) (T_STATIC inline_function)) 470 ((expr) (attributes T_STATIC inline_function)) 471 ((expr) (match)) 472 ((inline_function) (function returns_ref backup_doc_comment "(" parameter_list ")" lexical_vars return_type backup_fn_flags "{" inner_statement_list "}" backup_fn_flags)) 473 ((inline_function) (fn returns_ref backup_doc_comment "(" parameter_list ")" return_type T_DOUBLE_ARROW backup_fn_flags backup_lex_pos expr backup_fn_flags)) 474 ((fn) (T_FN)) 475 ((function) (T_FUNCTION)) 476 ((backup_doc_comment) (%empty)) 477 ((backup_fn_flags) (%empty)) 478 ((backup_lex_pos) (%empty)) 479 ((returns_ref) (%empty)) 480 ((returns_ref) (ampersand)) 481 ((lexical_vars) (%empty)) 482 ((lexical_vars) (T_USE "(" lexical_var_list possible_comma ")")) 483 ((lexical_var_list) (lexical_var_list "," lexical_var)) 484 ((lexical_var_list) (lexical_var)) 485 ((lexical_var) (T_VARIABLE)) 486 ((lexical_var) (ampersand T_VARIABLE)) 487 ((function_call) (name argument_list)) 488 ((function_call) (T_READONLY argument_list)) 489 ((function_call) (class_name T_PAAMAYIM_NEKUDOTAYIM member_name argument_list)) 490 ((function_call) (variable_class_name T_PAAMAYIM_NEKUDOTAYIM member_name argument_list)) 491 ((function_call) (callable_expr argument_list)) 492 ((class_name) (T_STATIC)) 493 ((class_name) (name)) 494 ((class_name_reference) (class_name)) 495 ((class_name_reference) (new_variable)) 496 ((class_name_reference) ("(" expr ")")) 497 ((backticks_expr) (%empty)) 498 ((backticks_expr) (T_ENCAPSED_AND_WHITESPACE)) 499 ((backticks_expr) (encaps_list)) 500 ((ctor_arguments) (%empty)) 501 ((ctor_arguments) (argument_list)) 502 ((dereferenceable_scalar) (T_ARRAY "(" array_pair_list ")")) 503 ((dereferenceable_scalar) ("[" array_pair_list "]")) 504 ((dereferenceable_scalar) (T_CONSTANT_ENCAPSED_STRING)) 505 ((dereferenceable_scalar) ("\"" encaps_list "\"")) 506 ((scalar) (T_LNUMBER)) 507 ((scalar) (T_DNUMBER)) 508 ((scalar) (T_START_HEREDOC T_ENCAPSED_AND_WHITESPACE T_END_HEREDOC)) 509 ((scalar) (T_START_HEREDOC T_END_HEREDOC)) 510 ((scalar) (T_START_HEREDOC encaps_list T_END_HEREDOC)) 511 ((scalar) (dereferenceable_scalar)) 512 ((scalar) (constant)) 513 ((scalar) (class_constant)) 514 ((constant) (name)) 515 ((constant) (T_LINE)) 516 ((constant) (T_FILE)) 517 ((constant) (T_DIR)) 518 ((constant) (T_TRAIT_C)) 519 ((constant) (T_METHOD_C)) 520 ((constant) (T_FUNC_C)) 521 ((constant) (T_PROPERTY_C)) 522 ((constant) (T_NS_C)) 523 ((constant) (T_CLASS_C)) 524 ((class_constant) (class_name T_PAAMAYIM_NEKUDOTAYIM identifier)) 525 ((class_constant) (variable_class_name T_PAAMAYIM_NEKUDOTAYIM identifier)) 526 ((class_constant) (class_name T_PAAMAYIM_NEKUDOTAYIM "{" expr "}")) 527 ((class_constant) (variable_class_name T_PAAMAYIM_NEKUDOTAYIM "{" expr "}")) 528 ((optional_expr) (%empty)) 529 ((optional_expr) (expr)) 530 ((variable_class_name) (fully_dereferenceable)) 531 ((fully_dereferenceable) (variable)) 532 ((fully_dereferenceable) ("(" expr ")")) 533 ((fully_dereferenceable) (dereferenceable_scalar)) 534 ((fully_dereferenceable) (class_constant)) 535 ((fully_dereferenceable) (new_dereferenceable)) 536 ((array_object_dereferenceable) (fully_dereferenceable)) 537 ((array_object_dereferenceable) (constant)) 538 ((callable_expr) (callable_variable)) 539 ((callable_expr) ("(" expr ")")) 540 ((callable_expr) (dereferenceable_scalar)) 541 ((callable_expr) (new_dereferenceable)) 542 ((callable_variable) (simple_variable)) 543 ((callable_variable) (array_object_dereferenceable "[" optional_expr "]")) 544 ((callable_variable) (array_object_dereferenceable T_OBJECT_OPERATOR property_name argument_list)) 545 ((callable_variable) (array_object_dereferenceable T_NULLSAFE_OBJECT_OPERATOR property_name argument_list)) 546 ((callable_variable) (function_call)) 547 ((variable) (callable_variable)) 548 ((variable) (static_member)) 549 ((variable) (array_object_dereferenceable T_OBJECT_OPERATOR property_name)) 550 ((variable) (array_object_dereferenceable T_NULLSAFE_OBJECT_OPERATOR property_name)) 551 ((simple_variable) (T_VARIABLE)) 552 ((simple_variable) ("$" "{" expr "}")) 553 ((simple_variable) ("$" simple_variable)) 554 ((static_member) (class_name T_PAAMAYIM_NEKUDOTAYIM simple_variable)) 555 ((static_member) (variable_class_name T_PAAMAYIM_NEKUDOTAYIM simple_variable)) 556 ((new_variable) (simple_variable)) 557 ((new_variable) (new_variable "[" optional_expr "]")) 558 ((new_variable) (new_variable T_OBJECT_OPERATOR property_name)) 559 ((new_variable) (new_variable T_NULLSAFE_OBJECT_OPERATOR property_name)) 560 ((new_variable) (class_name T_PAAMAYIM_NEKUDOTAYIM simple_variable)) 561 ((new_variable) (new_variable T_PAAMAYIM_NEKUDOTAYIM simple_variable)) 562 ((member_name) (identifier)) 563 ((member_name) ("{" expr "}")) 564 ((member_name) (simple_variable)) 565 ((property_name) (T_STRING)) 566 ((property_name) ("{" expr "}")) 567 ((property_name) (simple_variable)) 568 ((array_pair_list) (non_empty_array_pair_list)) 569 ((possible_array_pair) (%empty)) 570 ((possible_array_pair) (array_pair)) 571 ((non_empty_array_pair_list) (non_empty_array_pair_list "," possible_array_pair)) 572 ((non_empty_array_pair_list) (possible_array_pair)) 573 ((array_pair) (expr T_DOUBLE_ARROW expr)) 574 ((array_pair) (expr)) 575 ((array_pair) (expr T_DOUBLE_ARROW ampersand variable)) 576 ((array_pair) (ampersand variable)) 577 ((array_pair) (T_ELLIPSIS expr)) 578 ((array_pair) (expr T_DOUBLE_ARROW T_LIST "(" array_pair_list ")")) 579 ((array_pair) (T_LIST "(" array_pair_list ")")) 580 ((encaps_list) (encaps_list encaps_var)) 581 ((encaps_list) (encaps_list T_ENCAPSED_AND_WHITESPACE)) 582 ((encaps_list) (encaps_var)) 583 ((encaps_list) (T_ENCAPSED_AND_WHITESPACE encaps_var)) 584 ((encaps_var) (T_VARIABLE)) 585 ((encaps_var) (T_VARIABLE "[" encaps_var_offset "]")) 586 ((encaps_var) (T_VARIABLE T_OBJECT_OPERATOR T_STRING)) 587 ((encaps_var) (T_VARIABLE T_NULLSAFE_OBJECT_OPERATOR T_STRING)) 588 ((encaps_var) (T_DOLLAR_OPEN_CURLY_BRACES expr "}")) 589 ((encaps_var) (T_DOLLAR_OPEN_CURLY_BRACES T_STRING_VARNAME "}")) 590 ((encaps_var) (T_DOLLAR_OPEN_CURLY_BRACES T_STRING_VARNAME "[" expr "]" "}")) 591 ((encaps_var) (T_CURLY_OPEN variable "}")) 592 ((encaps_var_offset) (T_STRING)) 593 ((encaps_var_offset) (T_NUM_STRING)) 594 ((encaps_var_offset) ("-" T_NUM_STRING)) 595 ((encaps_var_offset) (T_VARIABLE)) 596 ((internal_functions_in_yacc) (T_ISSET "(" isset_variables possible_comma ")")) 597 ((internal_functions_in_yacc) (T_EMPTY "(" expr ")")) 598 ((internal_functions_in_yacc) (T_INCLUDE expr)) 599 ((internal_functions_in_yacc) (T_INCLUDE_ONCE expr)) 600 ((internal_functions_in_yacc) (T_EVAL "(" expr ")")) 601 ((internal_functions_in_yacc) (T_REQUIRE expr)) 602 ((internal_functions_in_yacc) (T_REQUIRE_ONCE expr)) 603 ((isset_variables) (isset_variable)) 604 ((isset_variables) (isset_variables "," isset_variable)) 605 ((isset_variable) (expr))))
  "The hash-table indexed by production-number and value is production.")

(defvar
  phps-mode-parser--table-look-aheads
  #s(hash-table size 217 test equal rehash-size 1.5 rehash-threshold 0.8125 data (("!") t ("\"") t ($) t ("$") t ("%") t ("(") t (")") t ("*") t ("+") t (",") t ("-") t (".") t ("/") t (":") t (";") t ("<") t ("=") t (">") t ("?") t ("@") t (PREC_ARROW_FUNCTION) t (T_ABSTRACT) t (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) t (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) t (T_AND_EQUAL) t (T_ARRAY) t (T_ARRAY_CAST) t (T_AS) t (T_ATTRIBUTE) t (T_BOOLEAN_AND) t (T_BOOLEAN_OR) t (T_BOOL_CAST) t (T_BREAK) t (T_CALLABLE) t (T_CASE) t (T_CATCH) t (T_CLASS) t (T_CLASS_C) t (T_CLONE) t (T_COALESCE) t (T_COALESCE_EQUAL) t (T_CONCAT_EQUAL) t (T_CONST) t (T_CONSTANT_ENCAPSED_STRING) t (T_CONTINUE) t (T_CURLY_OPEN) t (T_DEC) t (T_DECLARE) t (T_DEFAULT) t (T_DIR) t (T_DIV_EQUAL) t (T_DNUMBER) t (T_DO) t (T_DOLLAR_OPEN_CURLY_BRACES) t (T_DOUBLE_ARROW) t (T_DOUBLE_CAST) t (T_ECHO) t (T_ELLIPSIS) t (T_ELSE) t (T_ELSEIF) t (T_EMPTY) t (T_ENCAPSED_AND_WHITESPACE) t (T_ENDDECLARE) t (T_ENDFOR) t (T_ENDFOREACH) t (T_ENDIF) t (T_ENDSWITCH) t (T_ENDWHILE) t (T_END_HEREDOC) t (T_ENUM) t (T_EVAL) t (T_EXIT) t (T_EXTENDS) t (T_FILE) t (T_FINAL) t (T_FINALLY) t (T_FN) t (T_FOR) t (T_FOREACH) t (T_FUNCTION) t (T_FUNC_C) t (T_GLOBAL) t (T_GOTO) t (T_HALT_COMPILER) t (T_IF) t (T_IMPLEMENTS) t (T_INC) t (T_INCLUDE) t (T_INCLUDE_ONCE) t (T_INLINE_HTML) t (T_INSTANCEOF) t (T_INSTEADOF) t (T_INTERFACE) t (T_INT_CAST) t (T_ISSET) t (T_IS_EQUAL) t (T_IS_GREATER_OR_EQUAL) t (T_IS_IDENTICAL) t (T_IS_NOT_EQUAL) t (T_IS_NOT_IDENTICAL) t (T_IS_SMALLER_OR_EQUAL) t (T_LINE) t (T_LIST) t (T_LNUMBER) t (T_LOGICAL_AND) t (T_LOGICAL_OR) t (T_LOGICAL_XOR) t (T_MATCH) t (T_METHOD_C) t (T_MINUS_EQUAL) t (T_MOD_EQUAL) t (T_MUL_EQUAL) t (T_NAMESPACE) t (T_NAME_FULLY_QUALIFIED) t (T_NAME_QUALIFIED) t (T_NAME_RELATIVE) t (T_NEW) t (T_NOELSE) t (T_NS_C) t (T_NS_SEPARATOR) t (T_NULLSAFE_OBJECT_OPERATOR) t (T_NUM_STRING) t (T_OBJECT_CAST) t (T_OBJECT_OPERATOR) t (T_OR_EQUAL) t (T_PAAMAYIM_NEKUDOTAYIM) t (T_PLUS_EQUAL) t (T_POW) t (T_POW_EQUAL) t (T_PRINT) t (T_PRIVATE) t (T_PRIVATE_SET) t (T_PROPERTY_C) t (T_PROTECTED) t (T_PROTECTED_SET) t (T_PUBLIC) t (T_PUBLIC_SET) t (T_READONLY) t (T_REQUIRE) t (T_REQUIRE_ONCE) t (T_RETURN) t (T_SL) t (T_SL_EQUAL) t (T_SPACESHIP) t (T_SR) t (T_SR_EQUAL) t (T_START_HEREDOC) t (T_STATIC) t (T_STRING) t (T_STRING_CAST) t (T_STRING_VARNAME) t (T_SWITCH) t (T_THROW) t (T_TRAIT) t (T_TRAIT_C) t (T_TRY) t (T_UNSET) t (T_UNSET_CAST) t (T_USE) t (T_VAR) t (T_VARIABLE) t (T_WHILE) t (T_XOR_EQUAL) t (T_YIELD) t (T_YIELD_FROM) t ("[") t ("]") t ("^") t ("`") t ("{") t ("|") t ("}") t ("~") t))
  "The hash-table of valid look-aheads.")

(defvar
  phps-mode-parser--table-terminal-p
  #s(hash-table size 217 test equal rehash-size 1.5 rehash-threshold 0.8125 data (T_NUM_STRING t T_CURLY_OPEN t T_STRING_VARNAME t T_DOLLAR_OPEN_CURLY_BRACES t "$" t T_NULLSAFE_OBJECT_OPERATOR t T_OBJECT_OPERATOR t T_END_HEREDOC t T_START_HEREDOC t T_DNUMBER t T_LNUMBER t "\"" t T_CONSTANT_ENCAPSED_STRING t T_ENCAPSED_AND_WHITESPACE t PREC_ARROW_FUNCTION t T_YIELD_FROM t "`" t "@" t T_UNSET_CAST t T_BOOL_CAST t T_OBJECT_CAST t T_ARRAY_CAST t T_STRING_CAST t T_DOUBLE_CAST t T_INT_CAST t T_COALESCE t T_SPACESHIP t T_IS_GREATER_OR_EQUAL t ">" t T_IS_SMALLER_OR_EQUAL t "<" t T_IS_NOT_EQUAL t T_IS_EQUAL t T_IS_NOT_IDENTICAL t T_IS_IDENTICAL t "!" t "~" t T_SR t T_SL t "%" t "/" t T_POW t "*" t "-" t "+" t "." t "^" t T_BOOLEAN_AND t T_BOOLEAN_OR t T_DEC t T_INC t T_COALESCE_EQUAL t T_SR_EQUAL t T_SL_EQUAL t T_XOR_EQUAL t T_OR_EQUAL t T_AND_EQUAL t T_MOD_EQUAL t T_CONCAT_EQUAL t T_DIV_EQUAL t T_POW_EQUAL t T_MUL_EQUAL t T_MINUS_EQUAL t T_PLUS_EQUAL t T_PRIVATE_SET t T_PROTECTED_SET t T_PUBLIC_SET t T_PAAMAYIM_NEKUDOTAYIM t "?" t T_NOELSE t "[" t "=" t T_ELLIPSIS t T_VARIABLE t "|" t ":" t T_DOUBLE_ARROW t T_INLINE_HTML t T_NS_SEPARATOR t "}" t "{" t ";" t ")" t "(" t T_HALT_COMPILER t "]" t T_ATTRIBUTE t "," t T_NAME_RELATIVE t T_NAME_FULLY_QUALIFIED t T_NAME_QUALIFIED t T_STRING t T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG t T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG t T_READONLY t T_PUBLIC t T_PROTECTED t T_PRIVATE t T_FINAL t T_ABSTRACT t T_STATIC t T_PROPERTY_C t T_ENUM t T_MATCH t T_FN t T_NS_C t T_DIR t T_FILE t T_LINE t T_METHOD_C t T_FUNC_C t T_TRAIT_C t T_CLASS_C t T_CLASS t T_INTERFACE t T_TRAIT t T_NAMESPACE t T_IMPLEMENTS t T_EXTENDS t T_CALLABLE t T_ARRAY t T_BREAK t T_DEFAULT t T_CASE t T_ENDSWITCH t T_SWITCH t T_LIST t T_YIELD t T_PRINT t T_RETURN t T_CONST t T_FUNCTION t T_GOTO t T_CONTINUE t T_EMPTY t T_ISSET t T_UNSET t T_VAR t T_GLOBAL t T_INSTEADOF t T_USE t T_THROW t T_FINALLY t T_CATCH t T_TRY t T_AS t T_ENDDECLARE t T_DECLARE t T_ENDFOREACH t T_FOREACH t T_ENDFOR t T_FOR t T_ENDWHILE t T_WHILE t T_DO t T_ECHO t T_ENDIF t T_ELSE t T_ELSEIF t T_IF t T_EXIT t T_CLONE t T_NEW t T_INSTANCEOF t T_LOGICAL_AND t T_LOGICAL_XOR t T_LOGICAL_OR t T_REQUIRE_ONCE t T_REQUIRE t T_EVAL t T_INCLUDE_ONCE t T_INCLUDE t))
  "The hash-table of valid terminals.")

(defvar
  phps-mode-parser--table-non-terminal-p
  #s(hash-table size 217 test equal rehash-size 1.5 rehash-threshold 0.8125 data (isset_variable t isset_variables t encaps_var_offset t encaps_var t possible_array_pair t array_pair t non_empty_array_pair_list t static_member t function_call t property_name t array_object_dereferenceable t callable_variable t fully_dereferenceable t class_constant t constant t dereferenceable_scalar t encaps_list t new_variable t callable_expr t variable_class_name t member_name t lexical_var t lexical_var_list t backup_lex_pos t fn t lexical_vars t match t inline_function t backticks_expr t scalar t internal_functions_in_yacc t new_non_dereferenceable t new_dereferenceable t anonymous_class t class_name_reference t ctor_arguments t anonymous_class_modifiers_optional t non_empty_for_exprs t echo_expr t class_const_decl t property_hook_body t optional_parameter_list t property_hook_modifiers t property_hook t property_hook_list t property t member_modifier t trait_method_reference t absolute_trait_method_reference t trait_alias t trait_precedence t trait_adaptation t trait_adaptation_list t trait_adaptations t attributed_class_statement t enum_case t method_body t method_modifiers t class_const_list t class_const_modifiers t hooked_property t property_list t property_modifiers t class_statement t static_var t simple_variable t global_var t argument t non_empty_argument_list t union_type_without_static_element t name t intersection_type_without_static t union_type_without_static t union_type_element t type_without_static t intersection_type t union_type t type t type_expr_without_static t optional_property_hook_list t is_variadic t is_reference t optional_type_without_static t optional_cpp_modifiers t non_empty_member_modifiers t parameter t attributed_parameter t non_empty_parameter_list t alt_if_stmt_without_else t if_stmt_without_else t match_arm_cond_list t match_arm t non_empty_match_arm_list t match_arm_list t case_separator t case_list t array_pair_list t ampersand t class_name_list t enum_case_expr t type_expr t enum_backing_type t interface_extends_list t anonymous_class_modifiers t class_modifier t class_statement_list t implements_list t extends_from t class_modifiers t backup_fn_flags t return_type t parameter_list t backup_doc_comment t function_name t returns_ref t function t variable t unset_variable t optional_variable t catch_name_list t finally_statement t catch_list t declare_statement t foreach_statement t foreach_variable t unset_variables t echo_expr_list t static_var_list t global_var_list t optional_expr t switch_case_list t for_statement t for_exprs t while_statement t expr t alt_if_stmt t if_stmt t inner_statement t inner_statement_list t const_decl t use_declaration t unprefixed_use_declaration t inline_use_declaration t inline_use_declarations t unprefixed_use_declarations t legacy_namespace_name t const_list t use_declarations t group_use_declaration t use_type t mixed_group_use_declaration t namespace_declaration_name t attributed_statement t statement t enum_declaration_statement t interface_declaration_statement t trait_declaration_statement t class_declaration_statement t function_declaration_statement t attributes t attribute t possible_comma t attribute_group t attribute_decl t argument_list t class_name t namespace_name t identifier t top_statement t semi_reserved t reserved_non_modifiers t top_statement_list t))
  "The hash-table of valid non-terminals.")

(defvar
  phps-mode-parser--table-translations
  #s(hash-table size 65 test equal rehash-size 1.5 rehash-threshold 0.8125 data ())
  "The hash-table of translations.")

(defvar
  phps-mode-parser--e-identifier
  '%empty
  "The e-identifier.")

(defvar
  phps-mode-parser--eof-identifier
  '$
  "The end-of-file-identifier.")

(defvar
  phps-mode-parser--look-ahead-number
  1
  "The look-ahead number.")

(defvar
  phps-mode-parser-lex-analyzer--index-init
  1
  "Initial value of index.")

(defvar
  phps-mode-parser-lex-analyzer--state-init
  (list 'ST_INITIAL)
  "Initial value of state.")


;;; Buffer-Local Variables:


(defvar-local
  phps-mode-parser-lex-analyzer--buffered-response
  nil
  "Buffered tokens of lex-analyzer.")

(defvar-local
  phps-mode-parser-lex-analyzer--index
  0
  "The current index of the lex-analyzer.")

(defvar-local
  phps-mode-parser-lex-analyzer--state
  nil
  "The current state of the lex-analyzer.")


;;; Variable Functions:


(defvar
  phps-mode-parser-lex-analyzer--get-function
  (lambda (token) (let ((start (car (cdr token))) (end (cdr (cdr token)))) (if (<= end (point-max)) (progn (buffer-substring-no-properties start end)))))
  "The lex-analyzer get function.")

(defvar
  phps-mode-parser-lex-analyzer--function
  (lambda (index old-state) (let* ((lexer-response (phps-mode-lexer--re2c index old-state)) (tokens (nth 0 lexer-response)) (move-to-index (nth 1 lexer-response)) (new-state (nth 2 lexer-response))) (if move-to-index nil (let ((token-type (car (car tokens)))) (cond ((or (equal token-type 'T_OPEN_TAG) (equal token-type 'T_COMMENT) (equal token-type 'T_DOC_COMMENT)) (setq move-to-index (cdr (cdr (car tokens))))) ((equal token-type 'T_OPEN_TAG_WITH_ECHO) (let* ((v (car tokens))) (setcar v 'T_ECHO))) ((equal token-type 'T_CLOSE_TAG) (let* ((v (car tokens))) (setcar v ";")))))) (list tokens move-to-index new-state)))
  "The lex-analyzer function.")

(defvar
  phps-mode-parser-lex-analyzer--reset-function
  (lambda nil (progn))
  "The lex-analyzer reset function.")


;;; Functions:


;;; Functions for Lex-Analyzer:


(defun
  phps-mode-parser-lex-analyzer--get-function (token)
  "Get information about TOKEN."
  (let ((meta-information))
    (condition-case
      error
      (progn
        (setq
          meta-information
          (funcall
            phps-mode-parser-lex-analyzer--get-function
            token)))
      (error (error
        "Lex-analyze failed to get token meta-data of %s, error: %s"
        token
        (car (cdr error)))))
    meta-information))

(defun
  phps-mode-parser-lex-analyzer--reset
  ()
  "Reset Lex-Analyzer."
  (setq
    phps-mode-parser-lex-analyzer--buffered-response
    (make-hash-table :test 'equal))
  (setq
    phps-mode-parser-lex-analyzer--index
    phps-mode-parser-lex-analyzer--index-init)
  (setq
    phps-mode-parser-lex-analyzer--state
    phps-mode-parser-lex-analyzer--state-init)
  (when
    phps-mode-parser-lex-analyzer--reset-function
    (funcall
      phps-mode-parser-lex-analyzer--reset-function)))

(defun
  phps-mode-parser-lex-analyzer--peek-next-look-ahead
  ()
  "Peek next look-ahead number of tokens via lex-analyzer."
  (let ((look-ahead)
        (look-ahead-length
          0)
        (index
          phps-mode-parser-lex-analyzer--index)
        (state
          phps-mode-parser-lex-analyzer--state)
        (k (max
            1
            phps-mode-parser--look-ahead-number)))
    (while (<
            look-ahead-length
            k)
      (let* ((result-list
              (phps-mode-parser-lex-analyzer--get-buffered-lex
               index
               state))
             (token
              (nth 0 result-list))
             (new-index
              (nth 2 result-list))
             (new-state
              (nth 3 result-list)))
        (push
         token
         look-ahead)
        (setq
         look-ahead-length
         (1+ look-ahead-length))
        (setq
         index
         new-index)
        (setq
         state
         new-state)))

    (nreverse look-ahead)))

(defun
  phps-mode-parser-lex-analyzer--pop-token ()
  "Pop next token via lex-analyzer."
  (let* ((result-list
          (phps-mode-parser-lex-analyzer--get-buffered-lex
           phps-mode-parser-lex-analyzer--index
           phps-mode-parser-lex-analyzer--state))
         (token
          (nth 0 result-list))
         (new-index
          (nth 2 result-list))
         (new-state
          (nth 3 result-list)))
    (setq-local
     phps-mode-parser-lex-analyzer--index
     new-index)
    (setq-local
     phps-mode-parser-lex-analyzer--state
     new-state)
    (list token)))

(defun phps-mode-parser-lex-analyzer--get-buffered-lex (index state)
  "Get next token in stream, use buffer to only call function when needed."
  (unless (gethash
           index
           phps-mode-parser-lex-analyzer--buffered-response)
    (let ((continue t)
          (tmp-index index)
          (tmp-state state))
      (while continue
        (condition-case error
            (progn
              (let* ((result-list
                      (funcall
                       phps-mode-parser-lex-analyzer--function
                       tmp-index
                       tmp-state))
                     (tokens
                      (nth 0 result-list))
                     (move-to-index-flag
                      (nth 1 result-list))
                     (new-state
                      (nth 2 result-list)))
                (if move-to-index-flag
                    (progn
                      (setq
                       tmp-index
                       move-to-index-flag)
                      (setq
                       tmp-state
                       new-state))

                  (if tokens

                      (progn
                        (unless (listp (car tokens))
                          (setq tokens (list tokens)))

                        (let* ((first-token (car tokens))
                               (first-token-start (car (cdr first-token)))
                               (first-token-end (cdr (cdr first-token))))
                          (when (< index first-token-start)
                            (let ((token-start index))
                              (while (< token-start first-token-start)
                                (puthash
                                 token-start
                                 (list
                                  first-token
                                  nil
                                  first-token-end
                                  new-state)
                                 phps-mode-parser-lex-analyzer--buffered-response)
                                (setq
                                 token-start
                                 (1+ token-start))))))

                        (dolist (token tokens)
                          (let ((token-start (car (cdr token)))
                                (token-end (cdr (cdr token))))
                            (puthash
                             token-start
                             (list token nil token-end new-state)
                             phps-mode-parser-lex-analyzer--buffered-response))))

                    ;; Fill up look-ahead with EOF-identifier if we found nothing
                    (puthash
                     index
                     (list (list phps-mode-parser--eof-identifier) nil (1+ index) nil)
                     phps-mode-parser-lex-analyzer--buffered-response))

                  (setq
                   continue
                   nil))))
          (error
           (error
            "Lex-analyze failed to get next token at: %s in state: %s, error: %s"
            index
            state
            error))))))
  (gethash
   index
   phps-mode-parser-lex-analyzer--buffered-response))


;;; Functions for Syntax-Analyzer / Parser:


(defun
  phps-mode-parser--get-grammar-production-by-number
  (production-number)
  "If PRODUCTION-NUMBER exist, return it's production."
  (gethash
   production-number
   phps-mode-parser--table-productions-number-reverse))

(defun
  phps-mode-parser--valid-symbol-p
  (symbol)
  "Return whether SYMBOL is valid or not."
  (let ((is-valid t))
    (unless (or
             (phps-mode-parser--valid-e-p symbol)
             (phps-mode-parser--valid-eof-p symbol)
             (phps-mode-parser--valid-non-terminal-p symbol)
             (phps-mode-parser--valid-terminal-p symbol))
      (setq is-valid nil))
    is-valid))

(defun
  phps-mode-parser--valid-e-p
  (symbol)
  "Return whether SYMBOL is the e identifier or not."
  (eq
   symbol
   phps-mode-parser--e-identifier))

(defun
  phps-mode-parser--valid-eof-p
  (symbol)
  "Return whether SYMBOL is the EOF identifier or not."
  (eq
    symbol
    phps-mode-parser--eof-identifier))

(defun
  phps-mode-parser--valid-non-terminal-p (symbol)
  "Return whether SYMBOL is a non-terminal in grammar or not."
  (gethash
   symbol
   phps-mode-parser--table-non-terminal-p))

(defun
  phps-mode-parser--valid-terminal-p (symbol)
  "Return whether SYMBOL is a terminal in grammar or not."
  (gethash
   symbol
   phps-mode-parser--table-terminal-p))

(defun
  phps-mode-parser--get-grammar-translation-by-number
  (production-number)
  "If translation for PRODUCTION-NUMBER exist, return it."
  (gethash
    production-number
    phps-mode-parser--table-translations))

(defun
  phps-mode-parser--parse
  (&optional
    perform-sdt
    input-tape-index
    pushdown-list
    output
    translation
    translation-symbol-table-list
    history)
  "Perform a LR-parse via lex-analyzer, optionally PERFORM-SDT means to perform
  syntax-directed translation and optioanlly start at INPUT-TAPE-INDEX with
  PUSHDOWN-LIST, OUTPUT, TRANSLATION, TRANSLATION-SYMBOL-TABLE-LIST and HISTORY."
  (unless input-tape-index
    (setq input-tape-index 1))
  (unless pushdown-list
    (push 0 pushdown-list))
  (let ((translation-symbol-table
         (make-hash-table :test 'equal)))
    (when translation-symbol-table-list
      (dolist
          (item translation-symbol-table-list)
        (puthash
         (nth 0 item)
         (nth 1 item)
         translation-symbol-table)))

    (if (and
         input-tape-index
         (> input-tape-index 1))
        (setq
         phps-mode-parser-lex-analyzer--index
         input-tape-index)
      (phps-mode-parser-lex-analyzer--reset))

    ;; Make sure tables exists
    (unless phps-mode-parser--action-tables
      (error "Missing action-tables for grammar!"))
    (unless phps-mode-parser--distinct-action-tables
      (error "Missing distinct GOTO-tables for grammar!"))
    (unless phps-mode-parser--goto-tables
      (error "Missing GOTO-tables for grammar!"))
    (unless phps-mode-parser--distinct-goto-tables
      (error "Missing distinct GOTO-tables for grammar!"))

    (let ((accept))
      (while (not accept)

        ;; (1) The look-ahead string u, consisting of the next k input symbols, is determined.
        (let ((look-ahead
               (phps-mode-parser-lex-analyzer--peek-next-look-ahead))
              (look-ahead-full))

          ;; Save token stream indexes in separate variable if needed later
          (setq look-ahead-full look-ahead)

          ;; Create simplified look-ahead for logic below
          (setq look-ahead nil)
          (dolist (look-ahead-item look-ahead-full)
            (if (listp look-ahead-item)
                (push (car look-ahead-item) look-ahead)
              (push look-ahead-item look-ahead)))
          (setq look-ahead (nreverse look-ahead))

          (let ((table-index
                 (car pushdown-list)))
            (let ((action-table-distinct-index
                   (gethash
                    table-index
                    phps-mode-parser--action-tables)))
              (let ((action-table
                     (gethash
                      action-table-distinct-index
                      phps-mode-parser--distinct-action-tables)))
              (unless action-table
                (error
                 "Action-table with index %s is empty! Push-down-list: %s"
                 table-index
                 pushdown-list))
              (let ((action-match nil)
                    (action-table-length (length action-table))
                    (action-index 0)
                    (possible-look-aheads))

                ;; (2) The parsing action f of the table on top of the pushdown list is applied to the lookahead string u.
                (while (and
                        (not action-match)
                        (< action-index action-table-length))
                  (let ((action (nth action-index action-table)))
                    (let ((action-look-ahead (car action)))
                      (push
                       action-look-ahead
                       possible-look-aheads)
                      (when
                          (equal
                           action-look-ahead
                           look-ahead)
                        (setq
                         action-match
                         (cdr action)))
                      (when
                          (and
                           (=
                            phps-mode-parser--look-ahead-number
                            0)
                           (not
                            action-look-ahead))
                        ;; LR(0) reduce actions occupy entire row
                        ;; and is applied regardless of look-ahead
                        (setq
                         action-match
                         (cdr action))))
                    (setq
                     action-index
                     (1+ action-index))))

                (unless action-match
                  ;; (c) If f(u) = error, we halt parsing (and, in practice
                  ;; transfer to an error recovery routine).
                  (signal
                    'error
                    (list
                      (format
                        "Invalid syntax! Expected one of %s found %s at %s"

                   possible-look-aheads
                   look-ahead
                   phps-mode-parser-lex-analyzer--index)
                   possible-look-aheads
                   look-ahead
                   phps-mode-parser-lex-analyzer--index)))

                (cond

                 ((equal action-match '(shift))
                  ;; (a) If f(u) = shift, then the next input symbol, say a
                  ;; is removed from the input and shifted onto the pushdown list.
                  ;; The goto function g of the table on top of the pushdown list
                  ;; is applied to a to determine the new table to be placed on
                  ;; top of the pushdown list. We then return to step(1). If
                  ;; there is no next input symbol or g(a) is undefined, halt
                  ;; and declare error.

                  (let ((a (list (car look-ahead)))
                        (a-full (list (car look-ahead-full))))
                      (let ((goto-table-distinct-index
                             (gethash
                              table-index
                              phps-mode-parser--goto-tables)))
                        (let ((goto-table
                               (gethash
                                goto-table-distinct-index
                                phps-mode-parser--distinct-goto-tables)))
                      (let ((goto-table-length (length goto-table))
                            (goto-index 0)
                            (searching-match t)
                            (next-index)
                            (possible-look-aheads))

                        (while (and
                                searching-match
                                (< goto-index goto-table-length))
                          (let ((goto-item (nth goto-index goto-table)))
                            (let ((goto-item-symbol (list (car goto-item)))
                                  (goto-item-next-index (car (cdr goto-item))))
                              (push goto-item-symbol possible-look-aheads)

                              (when (equal
                                     goto-item-symbol
                                     a)
                                (setq next-index goto-item-next-index)
                                (setq searching-match nil))))

                          (setq goto-index (1+ goto-index)))
                        (unless next-index
                          (error
                           "In shift, found no GOTO-item for %s at %s, expected one of %s"
                           a
                           phps-mode-parser-lex-analyzer--index
                           possible-look-aheads))

                        (push (car a-full) pushdown-list)
                        (push next-index pushdown-list)
                        (phps-mode-parser-lex-analyzer--pop-token))))))

                 ((equal (car action-match) 'reduce)
                  ;; (b) If f(u) = reduce i and production i is A -> a,
                  ;; then 2|a| symbols are removed from the top of the pushdown
                  ;; list, and production number i is placed in the output
                  ;; buffer. A new table T' is then exposed as the top table
                  ;; of the pushdown list, and the goto function of T' is applied
                  ;; to A to determine the next table to be placed on top of the
                  ;; pushdown list. We place A and this new table on top of the
                  ;; the pushdown list and return to step (1)

                  (let ((production-number (car (cdr action-match))))

                    (let ((production
                           (phps-mode-parser--get-grammar-production-by-number
                            production-number)))
                      (let ((production-lhs (car production))
                            (production-rhs (car (cdr production)))
                            (popped-items-contents))
                        (unless (equal
                                 production-rhs
                                 (list phps-mode-parser--e-identifier))
                          (let ((pop-items (* 2 (length production-rhs)))
                                (popped-items 0)
                                (popped-item))
                            (while (< popped-items pop-items)
                              (setq popped-item (pop pushdown-list))
                              (when (and
                                     (listp popped-item)
                                     (phps-mode-parser--valid-symbol-p
                                      (car popped-item)))
                                (push
                                 popped-item
                                 popped-items-contents))
                              (setq popped-items (1+ popped-items)))))
                        (push production-number output)

                        (when perform-sdt
                          (let ((popped-items-meta-contents)
                                (popped-items-terminals))
                            (setq
                             popped-items-contents
                             (reverse popped-items-contents))
                            ;; Collect arguments for translation
                            (dolist (popped-item popped-items-contents)
                              (if (and
                                   (listp popped-item)
                                   (cdr popped-item))
                                  ;; If item is a terminal, use it's literal value
                                  (progn
                                    (push
                                      (phps-mode-parser-lex-analyzer--get-function
                                       popped-item)
                                       popped-items-meta-contents)
                                      (push
                                       popped-item
                                       popped-items-terminals))

                                ;; If item is a non-terminal
                                (let ((temp-hash-key
                                       (format
                                        "%S"
                                         popped-item)))

                                ;; If we have a translation for symbol, pop one
                                ;; otherwise push nil on translation argument stack
                                (if (gethash
                                         temp-hash-key
                                         translation-symbol-table)
                                        (let ((symbol-translations
                                               (gethash
                                                temp-hash-key
                                                translation-symbol-table)))
                                          (let ((symbol-translation
                                                 (pop symbol-translations)))
                                            (push
                                             (car symbol-translation)
                                             popped-items-meta-contents)
                                            (push
                                             (car (cdr symbol-translation))
                                             popped-items-terminals)
                                            (puthash
                                             temp-hash-key
                                             symbol-translations
                                             translation-symbol-table)))
                                      (push
                                       nil
                                       popped-items-meta-contents)
                                      (push
                                       nil
                                       popped-items-terminals)))))

                              ;; If we just have one argument, pass it as a instead of a list
                              (when (= (length popped-items-meta-contents) 1)
                                (setq
                                 popped-items-meta-contents
                                 (car popped-items-meta-contents))
                                (setq
                                 popped-items-terminals
                                 (car popped-items-terminals)))

                              ;; Perform translation at reduction if specified
                              (if
                                  (phps-mode-parser--get-grammar-translation-by-number
                                   production-number)
                                  (progn
                                    (condition-case conditions
                                        (let ((partial-translation
                                               (funcall
                                                (phps-mode-parser--get-grammar-translation-by-number
                                                 production-number)
                                                popped-items-meta-contents
                                                popped-items-terminals)))
                                          (let ((temp-hash-key
                                                 (format
                                                  "%S"
                                                  production-lhs)))
                                            (let ((symbol-translations
                                                   (gethash
                                                    temp-hash-key
                                                    translation-symbol-table)))
                                              (push
                                               (list
                                                partial-translation
                                                popped-items-terminals)
                                               symbol-translations)
                                              (puthash
                                               temp-hash-key
                                               symbol-translations
                                               translation-symbol-table)
                                              (setq
                                               translation
                                               partial-translation))))
                                      (error
                                       (signal
                                        'error
                                        (list
                                         (format
                                          "Failed AST translation for production %S with error: %S"
                                          production-number
                                          conditions))))))

                                ;; When no translation is specified just use popped contents as translation
                                (let ((partial-translation
                                       (list
                                        popped-items-meta-contents
                                        popped-items-terminals)))
                                  (let ((temp-hash-key
                                         (format
                                          "%S"
                                          production-lhs)))
                                    (let ((symbol-translations
                                           (gethash
                                            temp-hash-key
                                            translation-symbol-table)))
                                      (push
                                       partial-translation
                                       symbol-translations)
                                      (puthash
                                       temp-hash-key
                                       symbol-translations
                                       translation-symbol-table)
                                      (setq
                                       translation
                                       (car partial-translation))))))))

                          (let ((new-table-index (car pushdown-list)))
                            (let ((goto-table-distinct-index
                                   (gethash
                                    new-table-index
                                    phps-mode-parser--goto-tables)))
                              (let ((goto-table
                                     (gethash
                                      goto-table-distinct-index
                                      phps-mode-parser--distinct-goto-tables)))
                                (let ((goto-table-length
                                       (length goto-table))
                                      (goto-index 0)
                                      (searching-match t)
                                      (next-index))

                                  (while (and
                                          searching-match
                                          (< goto-index goto-table-length))
                                    (let ((goto-item (nth goto-index goto-table)))
                                      (let ((goto-item-symbol (list (car goto-item)))
                                            (goto-item-next-index (car (cdr goto-item))))

                                        (when (equal
                                               goto-item-symbol
                                               production-lhs)
                                          (setq next-index goto-item-next-index)
                                          (setq searching-match nil))))

                                    (setq goto-index (1+ goto-index)))

                                  (when next-index
                                    (push production-lhs pushdown-list)
                                    (push next-index pushdown-list))))))))))

                   ((equal action-match '(accept))
                    ;;    (d) If f(u) = accept, we halt and declare the string
                    ;;    in the output buffer to be the right parse of the original
                    ;;    input string.

                    (setq accept t))
                   (t (error
                       "Invalid action-match: %s!"
                       action-match)))))))))
      (unless accept
        (error
         "Parsed entire string without getting accepting! Output: %s"
         (reverse output)))
      (when history
        (setq history (reverse history)))
      (when output
        (setq output (reverse output)))
      (let ((translation-symbol-table-list))
        (when translation-symbol-table
          (maphash
           (lambda (key value)
             (push
              `(,key ,value)
              translation-symbol-table-list))
           translation-symbol-table))
        (list
         output
         translation
         translation-symbol-table-list
         history)))))

(defun phps-mode-parser-parse
    (&optional
     input-tape-index
     pushdown-list
     output
     translation
     history)
  "Perform a LR-parse via lex-analyzer, optionally at INPUT-TAPE-INDEX with
  PUSHDOWN-LIST, OUTPUT, TRANSLATION and HISTORY."
  (let ((result
         (phps-mode-parser--parse
          nil
          input-tape-index
          pushdown-list
          output
          translation
          history)))
    (nth 0 result)))

(defun phps-mode-parser-translate
    (&optional
     input-tape-index
     pushdown-list
     output
     translation
     history)
  "Perform a LR-parse via lex-analyzer, optionally at INPUT-TAPE-INDEX with
  PUSHDOWN-LIST, OUTPUT, TRANSLATION and HISTORY."
  (let ((result
         (phps-mode-parser--parse
          t
          input-tape-index
          pushdown-list
          output
          translation
          history)))
    (nth 1 result)))

(provide 'phps-mode-parser)

;;; phps-mode-parser.el ends here
